/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include <folly/Benchmark.h>
#include <folly/init/Init.h>
#include <folly/json.h>

#include "mcrouter/lib/carbon/example/gen/HelloGoodbyeRouteHandleIf.h"
#include "mcrouter/lib/carbon/example/gen/HelloGoodbyeRouterInfo.h"
#include "mcrouter/lib/config/RouteHandleFactory.h"
#include "mcrouter/lib/test/RouteHandleTestUtil.h"
#include "mcrouter/routes/ShardSelectionRouteFactory.h"

using facebook::memcache::RouteHandleFactory;
using facebook::memcache::SimpleRouteHandleProvider;
using hellogoodbye::HelloGoodbyeRouteHandleIf;
using hellogoodbye::HelloGoodbyeRouterInfo;

namespace {

class BasicShardSelector {
 public:
  explicit BasicShardSelector(std::unordered_map<uint32_t, uint32_t> shardsMap)
      : shardsMap_(std::move(shardsMap)) {}

  std::string type() const {
    return "basic-shard-selector";
  }

  template <class Request>
  size_t select(const Request& req, size_t /* size */) const {
    auto it = shardsMap_.find(*req.shardId_ref());
    if (it == shardsMap_.end()) {
      return std::numeric_limits<size_t>::max();
    }
    return it->second;
  }

 private:
  const std::unordered_map<uint32_t, uint32_t> shardsMap_;
};

constexpr folly::StringPiece kShardSelectionSmall = R"(
{
  "pool": {
    "type": "Pool",
    "name": "kShardSelectionSmall",
    "servers": [
      "localhost:12312",
      "localhost:12312",
      "localhost:12312"
    ],
    "shards": [
      "0,1,2,3,4,5,6,7,8,9,10",
      "6,7,8,9,10,11,12,13,14,15,16",
      "10,11,12,13,14,15,16,17,18,19,20"
    ]
  }
}
)";
const folly::dynamic kShardSelectionSmallJson{
    folly::parseJson(kShardSelectionSmall)};

constexpr folly::StringPiece kShardSelectionMedium = R"(
{
  "pool": {
    "type": "Pool",
    "name": "kShardSelectionMedium",
    "servers": [
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312"
    ],
    "shards": [
      "10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39",
      "20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49",
      "30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59",
      "40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69",
      "50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79",
      "60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89",
      "70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99"
    ]
  }
}
)";
const folly::dynamic kShardSelectionMediumJson{
    folly::parseJson(kShardSelectionMedium)};

constexpr folly::StringPiece kShardSelectionLarge = R"(
{
  "pool": {
    "type": "Pool",
    "name": "kShardSelectionLarge",
    "servers": [
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312",
      "localhost:12312"
    ],
    "shards": [
      "509,610,640,644,648,659,820,907,972,1096,1148,1199,1377,1417,1420,1529,1565,1576,1587,1588,1589,1632,1713,1761,1830,1832,1860,1882,1942,2075,2093,2142,2246,2249,2253,2276,2361,2542,2596,2739,2812,2841,2865,2868,2916,2950,2973,3125,3334,3338,3379,3732,3752,3800,3816,4115,4124,4131,4360,4494,4606,4626,4772,4821,4822,4827,4877,5015,5046,5154,5177,5229,5296,5426,5467,5510,5536,5540,5542,5546,5647,5664,5665,5681,5757,5793,5800,5814,5828,5847,5853,5954,5976,6021,6023,6028,6064,6238,6370,6422,6510,6675,6745,7018,7148,7248,7261,7287,7309,7360,7401,7466,7472,7478,7499,7512,7937,7959,7991,8043,8204,8210,8214,8395,8399,8421,8490,8516,8627,8647,9017,9123,9297,9317,9319,9363,9401,9424,9433,9505,9541,9786,9897,9937,10128,10167,10169,10240,10243,10306,10491,10540,10692,10729,10798,10803,10881,11097,11160,11164,11344,11506,11536,11625,11693,11909,11994,12030,12049,12201,12277,12279,12321,12347,12545,12554,12634,12748,12773,12866,13159,13324,13417,13457,13462,13551,13683,13704,13832,13876,14025,14047,14160,14209,14219,14371,14405,14551,14614,14626,14678,15093,15122,15175,15226,15242,15500,15549,15579,15591,16364,16400,16536,16552,16571,16799,16843,16976,17071,17155,17310,17490,17622,17642,17664,17700,17903,17956,17981,18162,18416,18538,18782,18787,18888,18908,19119,19154,19161,19263,19488,19533,19715,20049,20094,20168,20344,20354,20559,20658,20707,20808,20827,21017,21169,21243,21306,21413,21581,21620,21649,21797,21845,22085,22251,22306,22493,22625,22674,22767,22822,22864,22908,22922,23201,23213,23449,23490,23766,24057,24108,24325,24363,24589,24646,24936,25028,25085,25089,25185,25216,25217,25262,25308,25406,25592,25605,25710,25797,26071,26139,26199,26242,26267,26394,26441,26563,26630,26718,26867,26882,27047,27073,27102,27315,27375,27464,27700,27808,27883,27984,28164,28459,28673,28747,28791,28814,28841,28847,28915,29030,29065,29110,29246,29249,29571,29847,29925,29928,29965,29971,30264,30960,31018,31453,31656,31783,31861,31948,32091,32102,32787,33262,33309,33572,33599,33842,34208,34326,34846,34991,35177,35689,35697,35708,35902,35950,36007,36136,36440,36447,36804,36823,37014,37111,37324,37357,37386,37562,37699,37940,38109,38116,38233,38504,38576,38806,38836,38898,39041,39058,39062,39253,39258,39264,39310,39321,39630,39667,39924,39938,39984,39987,39991,40317,40720,40781,40865,40984,41111,42092,42262,42344,43000,43169,43189,43251,43355,43387,43406,43454,43631,43948,43949,44016,44284,45323,45364,45386,45440,45818,45943,46219,46365,46801,46809,47178,47955,47979,48087,48112,48463,48858,49002,49558,49976,50005,50027,50371,50418,50437,50458,50556,50562,50734,50741,50872,50895,50925,50948,50988,51018,51134,51168,51460,51468,51543,51634,51649,51670,51688,51806,51822,51844,52056,52073,52174,52282,52433,52555,52600,52616,52677,52812,52877,52900,52983,53183,53252,53388,53463,53587,53721,53772,53915,53962,53991,54307,54345,54436,54466,54569,54881,54958,54974,55057,55258,55302,55345,55415,55496,55503,55659,55751,55820,55841,55880,55911,55935,56027,56047,56122,56223,56327,56366,56405,56635,56642,56756,56948,57002,57104,57112,57427,57566,57584,57589,57643,57669,57874,57980,58117,58157,58196,58308,58325,58503,58575,58829,58853,58876,58933,58947,59138,59145,59260,59275,59311,59346,59367,59390,59435,59505,59511,59527,59683,59743,59749,59794,59854,59879,59891,59977,59999,60017,60102,60106,60108,60169,60334,60379,60526,60847,60885,60959,60969,60988,60991,60996,61102,61174,61199,61252,61262,61383,61394,61408,61430,61442,61528,61664,61681,61815,61943,62055,62098,62129,62191,62351,62353,62362,62371,62422,62511,62516,62578,62652,62775,62877,62879,62935,62940,63010,63139,63150,63233,63315,63353,63377,63384,63418,63447,63490,63606,63705,63710,63913,63920,64109,64140,64188,64214,64240,64252,64288,64317,64444,64476,64553,64572,64595,64687,64760,64808,64887,64930,64933,65029,65210,65247,65450,65531,65575,65674,65695,65930,66051,66128,66223,66230,66377,66444,66527,66682,66784,66877,66980,67158,67250,67540,67579,67795,67828,67993,68030,68075,68209,68321,68397,68609,68645,68745,68784,68816,68930,69105,69140,69160,69208,69222,69441,69574,69594,69641,69663,69773,69798,69886,69955,69967,70027,70054,70133,70301,70351,70539,70662,70726,70844,70996,71039,71126,71209,71353,71406,71474,71501,71734,71754,71798,71804,72100,72117,72134,72159,72183,72206,72253,72275,72324,72335,72384,72573,72652,72691,72725,72884,72887,72941,73048,73120,73281,73372,73494,73533,73802,73855,73896,74309,74327,74447,74528,74583,74593,74699,74786,74890,74954,74972,74994,75015,75179,75183,75190,75537,75733,75760,75809,76002,76179,76337,76349,76361,76382,76475,76625,76771,76853,76922,76992,77059,77120,77165,77289,77401,77498,77583,77600,77617,77630,77648,77750,77847,77892,77898,77981,78048,78256,78290,78328,78336,78350,78407,78518,78546,78714,78734,78929,78963,78975,79112,79211,79246,79249,79263,79267,79434,79490,79491,79568,79585,79601,79618,79709,79871,79981,80184,80223,80271,80411,80584,80678,80844,80887,80914,80964,81071,81349,81358,81393,81434,81467,81521,81574,81611,81841,81928,81936,82038,82237,82254,82267,82311,82330,82356,82387,82479,82508,82542,82694,82707,82851,82885,82925,82940,83087,83391,83546,83642,83661,83676,83744,83824,83910,83922,84062,84172,84229,84234,84251,84533,84539,84553,84640,84699,84880,84946,85014,85047,85050,85056,85099,85213,85237,85243,85387,85486,85659,85687,85830,86055,86264,86272,86304,86333,86499,86552,86579,86894,86903,86927,86954,87112,87156,87237,87486,87572,87659,87684,87695,87736,87775,87776,87851,87882,88044,88102,88178,88219,88338,88612,88662,88802,88836,88847,88908,88938,88985,89183,89197,89262,89306,89418,89453,89549,89566,89591,89614,89707,89722,89744,89746,89756,89808,90095,90391,90423,90505,90538,90699,90729,90779,90825,90874,90876,91118,91370,91392,91521,91618,91679,91739,91877,91890,91892,91914,91923,92124,92210,92294,92373,92414,92619,92633,92822,92930,93014,93100,93193,93273,93338,93397,93408,93805,94083,94090,94258,94295,94371,94552,94604,94684,94745,94804,94915,94941,94981,95028,95067,95106,95130,95252,95301,95351,95352,95354,95489,95555,95667,95686,95722,95846,96089,96165,96258,96490,96554,96563,96660,96888,96919,96931,97039,97070,97171,97269,97320,97355,97358,97368,97371,97416,97638,97665,98065,98135,98225,98242,98267,98291,98488,98527,98545,98656,98712,98716,98851,98875,98907,98963,99003,99131,99146,99221,99249,99346,99482,99540,99561,99573,99667,99804,99823,99851,99949",
      "168,210,1972,2029,2043,2162,2179,2428,2439,2539,2579,2647,2664,2713,2736,3198,3277,3281,3301,3358,3377,3492,3530,3549,3563,3589,3659,3665,3722,3738,4005,4177,4189,4438,4619,4636,4648,4712,4763,4838,5719,6395,7340,7454,7516,9477,9478,9496,9553,9572,9591,9629,9648,9667,9686,9705,9743,9781,9800,9833,9857,10016,10028,10047,10123,10142,10522,10598,10681,10684,10702,10712,10731,10750,10769,10807,10826,10845,10901,10902,10978,11000,11035,11111,11130,11225,11263,11320,11358,11415,11421,11491,11586,11624,11643,11662,11852,11871,11928,11966,12023,12099,12118,12132,12137,12175,12213,12232,12251,12308,12327,12346,12418,12441,12492,12555,12579,12612,12631,12652,12691,12707,12729,12731,12745,12764,12818,12851,12859,12897,12934,12954,13011,13049,13068,13071,13125,13161,13163,13193,13220,13258,13277,13296,13329,13340,13448,13569,13581,13657,13676,13714,13733,13752,13771,13790,13809,13847,13866,13923,13942,14018,14037,14093,14094,14113,14148,14170,14208,14246,14284,14360,14369,14379,14464,14508,14516,14531,14561,14584,14647,14664,14687,14702,14721,14730,14753,14767,14785,14797,14816,14851,14892,14930,14949,14987,15044,15063,15101,15120,15177,15196,15234,15253,15310,15348,15367,15386,15424,15443,15462,15481,15519,15538,15557,15563,15633,15652,15665,15671,15709,15770,15823,15842,15876,15899,15918,15937,15956,15994,16027,16032,16042,16051,16070,16089,16146,16165,16222,16241,16279,16298,16317,16336,16355,16374,16393,16423,16431,16469,16488,16526,16545,16559,16583,16697,16716,16735,16761,16777,16792,16849,16906,16908,16944,17001,17020,17058,17077,17084,17115,17153,17172,17210,17248,17267,17286,17343,17362,17381,17400,17453,17456,17457,17495,17533,17552,17566,17609,17638,17645,17647,17704,17742,17761,17799,17837,17894,17913,17951,17970,18008,18027,18074,18103,18141,18160,18217,18236,18255,18312,18326,18331,18369,18388,18407,18426,18445,18464,18475,18502,18537,18540,18559,18578,18589,18597,18616,18673,18692,18730,18804,18806,18844,18875,18901,18920,18977,19015,19072,19091,19110,19129,19159,19162,19168,19205,19224,19262,19300,19338,19346,19357,19395,19414,19433,19452,19490,19509,19566,19585,19623,19642,19661,19680,19684,19699,19737,19764,19775,19794,19832,19851,19870,19889,19908,19927,19946,19965,19984,20079,20098,20117,20162,20212,20269,20307,20326,20345,20383,20402,20421,20440,20459,20478,20516,20534,20535,20554,20573,20592,20611,20630,20638,20687,20725,20782,20801,20805,20820,20839,20872,20877,20915,20934,20953,20955,20982,20991,21048,21105,21110,21124,21143,21162,21181,21238,21254,21276,21295,21324,21333,21352,21390,21428,21433,21447,21466,21485,21504,21523,21542,21580,21618,21637,21675,21686,21713,21732,21751,21770,21789,21846,21865,21884,21903,21922,21941,21960,21979,21998,22036,22055,22112,22169,22188,22245,22264,22283,22321,22359,22378,22397,22416,22419,22435,22473,22511,22530,22549,22587,22606,22644,22663,22673,22682,22696,22701,22720,22730,22739,22777,22806,22834,22842,22853,22891,22910,22948,22967,22994,23005,23024,23081,23119,23157,23214,23218,23233,23271,23290,23319,23328,23365,23385,23404,23442,23480,23518,23537,23594,23613,23651,23670,23708,23727,23746,23784,23822,23841,23879,23917,23936,23955,23974,23993,24012,24031,24050,24069,24084,24088,24148,24183,24202,24278,24297,24316,24373,24449,24468,24487,24506,24525,24544,24563,24582,24594,24620,24639,24640,24658,24732,24734,24753,24772,24810,24829,24867,24924,24943,24948,24981,25019,25038,25067,25080,25095,25114,25133,25179,25209,25247,25266,25274,25285,25304,25323,25399,25402,25456,25475,25494,25513,25532,25570,25623,25629,25671,25684,25722,25741,25760,25798,25834,25836,25912,25918,25931,25950,25969,25977,26007,26026,26045,26064,26083,26102,26121,26140,26178,26197,26209,26216,26273,26292,26368,26378,26425,26501,26539,26596,26653,26786,26805,26836,26848,26900,26976,27004,27017,27090,27099,27120,27166,27204,27223,27231,27261,27280,27451,27546,27648,27698,27701,27777,27784,27823,27891,27926,28059,28066,28154,28173,28192,28230,28240,28249,28268,28281,28317,28325,28363,28382,28458,28477,28515,28686,28800,28857,28866,28876,28895,28933,28952,29012,29142,29199,29247,29256,29275,29294,29370,29408,29427,29446,29454,29474,29522,29525,29582,29597,29696,29731,29738,29769,29777,29871,29883,29921,29941,29968,29993,29997,30025,30054,30083,30112,30141,30170,30199,30228,30257,30286,30315,30344,30373,30402,30431,30460,30489,30518,30547,30576,30605,30634,30663,30692,30721,30750,30779,30808,30837,30866,30895,30924,30953,30982,31011,31040,31069,31098,31127,31156,31185,31214,31243,31272,31301,31330,31359,31388,31417,31446,31475,31504,31533,31562,31591,31620,31649,31678,31707,31736,31765,31794,31823,31852,31881,31910,31939,31968,31997,32026,32055,32084,32113,32142,32171,32200,32229,32258,32287,32316,32345,32374,32403,32432,32461,32490,32519,32548,32577,32606,32635,32664,32693,32722,32751,32780,32809,32838,32867,32896,32925,32954,32983,33012,33041,33070,33099,33128,33157,33186,33215,33244,33273,33302,33331,33360,33389,33418,33447,33476,33505,33534,33563,33592,33621,33650,33679,33708,33737,33766,33795,33824,33853,33882,33911,33940,33969,33998,34027,34056,34085,34114,34143,34172,34201,34230,34259,34288,34317,34346,34375,34404,34433,34462,34491,34520,34549,34578,34607,34636,34665,34694,34723,34752,34781,34810,34839,34868,34897,34926,34955,34984,35025,35058,35114,35145,35188,35209,35222,35249,35277,35351,35384,35385,35392,35399,35425,35496,35544,35574,35603,35662,35669,35681,35735,35794,35853,35860,35924,35983,36042,36055,36105,36125,36165,36224,36248,36295,36354,36413,36485,36544,36603,36609,36667,36726,36737,36785,36790,36799,36856,36915,36930,36974,36985,37006,37046,37065,37105,37164,37171,37228,37287,37311,37346,37375,37417,37476,37535,37607,37666,37682,37700,37725,37740,37743,37789,37848,37907,37920,37978,37998,38037,38052,38096,38105,38168,38227,38249,38286,38291,38350,38369,38409,38415,38468,38539,38598,38657,38683,38729,38739,38788,38847,38854,38911,38930,38970,39029,39035,39044,39052,39100,39159,39218,39240,39290,39349,39408,39421,39472,39506,39531,39553,39590,39661,39720,39739,39779,39826,39851,39862,39910,39969,39997,40119,40148,40177,40206,40235,40264,40293,40322,40351,40380,40409,40438,40467,40496,40525,40554,40583,40612,40641,40670,40699,40728,40757,40786,40815,40844,40873,40902,40931,40960,40989,41018,41047,41076,41105,41134,41163,41192,41221,41250,41279,41308,41337,41366,41395,41424,41453,41482,41511,41540,42010,42039,42068,42097,42126,42186,42235,42249,42276,42314,42315,42351,42395,42426,42513,42570,42588,42608,42621,42663,42690,42750,42770,42825,42869,42900,42920,42963,42987,43007,43046,43062,43081,43114,43137,43224,43281,43299,43320,43374,43408,43426,43427,43461,43491,43512,43536,43571,43579,43611,43635,43649,43698,43721,43773,43805,43828,43848,43866,43915,43935,43992,43994,44010,44085,44106,44172,44191,44199,44231,44247,44270,44279,44287,44290,44322,44368,44389,44409,44484,44542,44559,44571,44600,44646,44697,44704,44721,44740,44752,44796,44833,44859,44864,44883,44924,44937,44958,44969,45033,45046,45069,45106,45119,45154,45165,45215,45273,45285,45327,45346,45371,45403,45461,45533,45590,45648,45701,45720,45777,45799,45804,45835,45907,45964,46019,46087,46137,46170,46242,46249,46254,46261,46267,46313,46326,46347,46355,46411,46462,46479,46495,46543,46579,46602,46622,46668,46673,46675,46682,46687,46765,46781,46851,46920,46966,46972,47014,47032,47058,47060,47064,47136,47154,47193,47196,47229,47246,47272,47278,47297,47333,47335,47422,47439,47450,47462,47473,47502,47505,47577,47578,47612,47651,47724,47820,47825,47831,47835,47853,47877,47882,47892,47973,48021,48031,48079,48081,48093,48143,48162,48167,48220,48243,48248,48251,48271,48281,48287,48343,48353,48390,48400,48418,48419,48422,48438,48461,48487,48549,48616,48694,48717,48758,48773,48815,48859,48861,48889,48893,48985,49041,49055,49107,49183,49234,49268,49283,49300,49336,49345,49355,49413,49476,49482,49496,49621,49624,49644,49658,49671,49721,49775,49778,49870,49883,49902,49925,49933,49977,49986,50019,50161,50167,50299,50300,50342,50365,50374,50471,50512,50520,50643,50662,50758,50930,51015,51055,51160,51221,51357,51360,51492,51631,51825,51867,51909,51910,51929,52061,52255,52358,52496,52505,52578,52766,52928,52942,52966,53000,53051,53134,53136,53139,53262,53323,53354,53394,53404,53543,53737,53758,53844,53955,54041,54108,54357,54396,54558,54622,54732,54741,54845,54950,55040,55199,55275,55320,55334,55360,55400,55527,55586,55613,55656,55711,55774,55830,55895,55920,55985,56282,56309,56610,56733,56771,56776,57118,57189,57198,57277,57313,57317,57387,57394,57486,57548,57602,57640,57707,57833,58035,58088,58118,58147,58343,58355,58360,58400,58449,58507,58597,58612,58641,58704,58779,58790,58888,59161,59165,59172,59216,59310,59429,59785,59793,59819,60067,60181,60616,60824,60831,60841,60990,61043,61348,61558,61698,61772,61883,61905,61922,61964,62030,62056,62152,62161,62397,62402,62421,62459,62477,62490,62604,62698,62802,62806,62830,62975,63050,63064,63079,63080,63236,63348,63378,63498,63609,63746,63836,64008,64169,64184,64187,64469,64504,64598,64770,64798,64804,64941,65079,65096,65115,65144,65149,65211,65272,65283,65287,65332,65536,65615,65627,65891,65939,65950,65951,66024,66127,66198,66218,66231,66318,66352,66364,66379,66478,66610,66632,66640,66691,66808,66884,66906,66915,66953,66977,66979,66984,67032,67128,67132,67140,67141,67297,67298,67310,67322,67530,67716,67839,67925,67962,67981,68245,68509,68539,68734,68804,68809,68835,68838,68854,68896,68903,68951,68974,69031,69097,69169,69175,69424,69436,69478,69616,69664,69704,69841,69851,69910,69971,69995,70029,70248,70270,70360,70456,70494,70495,70823,70866,70981,70985,71051,71225,71250,71266,71333,71493,71532,71633,71698,71704,71735,71773,71818,71902,71928,71999,72185,72219,72249,72366,72549,72597,72609,72626,72660,72727,72809,72896,72961,73033,73063,73525,73579,73778,73867,73876,74054,74171,74224,74283,74288,74370,74420,74501,74669,74821,74901,74910,74953,74975,74987,75024,75068,75107,75495,75587,75653,75686,75742,75793,75796,76005,76203,76270,76305,76331,76385,76537,76538,76618,76630,76707,76842,76857,76894,77045,77205,77222,77309,77340,77373,77385,77395,77417,77566,77604,77607,77709,77739,77798,77824,77838,77845,77871,77884,77897,78016,78037,78141,78227,78228,78565,78735,78785,78955,78985,79151,79185,79192,79264,79401,79444,79565,79733,79976,80029,80086,80217,80232,80402,80445,80464,80511,80530,80660,80776,80778,80837,81091,81168,81727,81735,81948,81996,82076,82165,82214,82299,82322,82651,82703,82767,82801,82815,83026,83071,83185,83197,83255,83334,83431,83442,83524,83617,83707,83723,83907,83925,83957,83964,83968,84038,84066,84119,84174,84264,84265,84369,84400,84467,84471,84572,84631,84651,84693,84852,84866,84877,84908,84937,85088,85272,85311,85371,85404,85411,85493,85503,85608,85625,85635,85678,85697,85725,85752,85794,85835,85927,86461,86580,86773,86910,86930,87019,87271,87312,87360,87373,87425,87563,87605,87645,87661,88002,88255,88291,88407,88508,88535,88701,88731,88767,88827,88831,88896,88918,88927,88995,89193,89203,89216,89249,89452,89463,89496,89759,89792,89810,89887,89975,90051,90142,90328,90350,90464,90602,90638,90661,91092,91096,91143,91321,91363,91536,91695,91868,91918,92148,92208,92383,92463,92475,92511,92629,92658,92689,92737,92821,92861,92907,92963,93119,93146,93229,93260,93288,93563,93597,93607,93770,93788,93799,93820,93885,93950,94066,94098,94129,94289,94316,94388,94678,94797,94844,94903,95064,95065,95100,95132,95218,95391,95432,95499,95517,95566,95679,95855,95891,95898,95902,95950,96026,96040,96043,96099,96147,96194,96265,96267,96385,96436,96516,96535,96695,96730,96849,96956,97102,97302,97357,97390,97543,97587,97672,97885,97968,98235,98493,98508,98549,98559,98584,98634,98643,98879,98919,98953,99090,99241,99264,99342,99426,99434,99589,99668,99674,99862,99905,99945",
      "22,170,384,620,904,1022,1070,1107,1426,1498,1618,1651,1701,1707,1820,1872,2064,2215,2221,2237,2271,2343,2518,2560,2779,2819,2848,3053,3123,3384,3428,3542,3561,3607,3860,3942,4331,4336,4371,4762,4938,5050,5149,5197,5210,5223,5252,5380,5445,5474,5495,5506,5600,5743,5753,5807,5916,5943,6002,6192,6305,6373,6409,6441,6485,6511,6542,6645,6665,6667,6683,6729,6758,6804,6808,6951,6996,7002,7162,7186,7199,7207,7424,7528,7538,7574,7606,7717,7950,7979,8066,8201,8495,8496,8507,8583,8596,8604,8814,8894,8953,8963,9033,9082,9093,9112,9129,9190,9283,9386,9463,9717,9751,9762,9841,9994,10000,10027,10199,10203,10327,10338,10353,10383,10432,10438,10484,10514,10515,10535,10602,10668,10766,10811,10868,10949,11583,11603,11666,11689,11706,11717,11777,11825,11935,12017,12192,12235,12372,12447,12567,12727,12867,12913,13070,13087,13104,13184,13257,13410,13453,13585,13586,13656,13684,13694,13751,13878,13879,13959,13991,14014,14046,14114,14154,14240,14244,14254,14321,14358,14688,14876,15128,15165,15181,15261,15351,15406,15521,15587,15717,15732,15785,15898,15908,15935,15941,15977,16074,16188,16192,16262,16287,16523,16666,16679,16687,16705,16772,16788,16979,17003,17027,17291,17369,17384,17535,17665,17752,18085,18143,18197,18371,18583,18728,18814,18853,18933,18969,18976,19017,19216,19220,19523,19549,19608,19647,20308,20480,20486,20634,20787,20838,20997,21034,21153,21161,21240,21369,21487,21690,21929,21989,22165,22329,22452,22483,22551,22578,22608,22623,22632,22660,22732,22833,22874,22905,22983,23098,23174,23399,23405,23459,23563,23622,23714,23785,23810,23952,23953,24020,24106,24121,24228,24314,24390,24552,24585,24795,24814,24849,24909,25001,25009,25042,25100,25155,25200,25244,25340,25648,25664,25669,25731,25761,25855,25987,26180,26476,26582,26625,26659,26725,26736,26751,26770,26784,26880,26933,27092,27098,27271,27328,27336,27398,27401,27416,27437,27653,27667,27798,27845,27939,28315,28328,28384,28399,28578,28597,28627,28688,28690,28759,28809,28870,28878,28946,29096,29123,29161,29178,29311,29328,29348,29432,29440,29464,29495,29518,29720,29750,29844,29866,30116,30191,30546,30554,30945,30952,30984,31424,31622,31880,31883,31908,32344,32440,32662,33107,33159,33277,33449,34199,34512,34808,34867,34899,35129,35190,35198,35505,35559,35640,35700,35743,35897,35916,35956,36011,36059,36104,36256,36265,36412,36601,36613,36630,36671,36692,36818,37009,37010,37024,37318,37358,37388,37448,37534,37611,37614,37680,37688,37818,37840,37847,37980,38008,38046,38147,38170,38195,38295,38321,38322,38387,38864,38868,38876,38885,38932,38933,38945,39116,39297,39317,39355,39446,39551,39812,39914,40075,40158,40377,41058,41278,41289,41996,42063,42437,43209,43271,43398,43410,43675,43764,43784,43847,43916,43917,44187,44212,44496,44772,44835,44943,45248,45306,45374,45383,45460,45501,45503,45557,45789,45806,45949,46180,46217,46421,46521,46780,47053,47071,47076,47162,47434,47566,47641,47676,47739,47744,47959,48253,48523,48574,48592,48687,48834,48997,49290,49423,49483,50006,50059,50101,50190,50704,50789,50822,50846,50888,51017,51050,51115,51131,51132,51135,51167,51225,51232,51241,51261,51374,51416,51475,51556,51610,51669,51922,51924,51959,51977,52075,52117,52127,52215,52429,52498,52587,52592,52598,52605,52708,52984,53034,53222,53244,53280,53407,53455,53469,53724,54067,54099,54203,54210,54411,54596,54675,54690,54731,54894,54913,54962,54990,54998,55048,55128,55182,55234,55252,55386,55408,55421,55490,55500,55599,55667,55816,55924,56001,56042,56050,56142,56153,56180,56206,56243,56297,56311,56535,56585,56660,56680,56690,57067,57115,57176,57197,57284,57302,57348,57402,57404,57506,57617,57704,57851,57942,58009,58052,58062,58064,58145,58190,58385,58630,58658,58711,58782,58930,58963,58986,59000,59032,59053,59201,59229,59233,59360,59430,59454,59516,59647,59662,59874,59880,60078,60081,60374,60415,60421,60457,60459,60495,60546,60611,60759,60871,60882,61044,61086,61093,61170,61213,61268,61592,61631,61684,61713,61754,61765,61875,61944,62022,62275,62304,62316,62447,62476,62480,62502,62554,62597,62749,62752,62807,62852,63023,63031,63052,63154,63174,63368,63497,63661,63697,63714,63767,63801,63831,63903,63919,64079,64318,64481,64574,64613,64782,64832,64888,64921,65127,65130,65330,65420,65471,65512,65528,65638,65649,65686,65739,65787,65871,66036,66150,66157,66180,66264,66404,66501,66536,66609,66645,67212,67346,67365,67372,67556,67571,67668,67775,67805,67827,67851,67967,68205,68225,68337,68343,68424,68494,68706,68748,68794,68937,68940,69055,69127,69246,69262,69268,69351,69352,69414,69446,69474,69532,69659,69670,69740,69843,69954,69977,70066,70070,70178,70355,70383,70572,70645,70680,70850,70922,71114,71208,71369,71434,71457,71514,71618,71701,71900,71906,71926,72086,72226,72235,72238,72298,72446,72610,72702,72907,73042,73143,73183,73288,73305,73469,73543,73684,73741,73844,73900,73943,73979,74006,74027,74113,74118,74148,74244,74328,74428,74545,74691,74803,74827,74861,75148,75163,75188,75240,75279,75309,75365,75440,75931,75993,76048,76252,76274,76350,76451,76487,76823,76914,76934,77021,77087,77113,77145,77237,77311,77494,77511,77517,77608,77615,77702,77780,77791,77864,77870,77887,77987,78031,78084,78091,78211,78293,78421,78438,78560,78570,78579,78607,78648,78711,78727,78877,79022,79031,79336,79344,79415,79450,79556,79575,79874,79970,80023,80227,80228,80240,80611,80630,80857,80904,80957,81055,81236,81258,81299,81462,81658,81664,81676,81733,81788,81802,82040,82081,82139,82280,82312,82336,82399,82615,82637,82775,82784,82896,82923,82938,83005,83011,83052,83240,83271,83361,83395,83463,83482,83492,83530,83705,83747,83905,84027,84101,84137,84339,84343,84584,84709,84773,84801,84970,85039,85152,85207,85234,85241,85258,85309,85314,85331,85400,85434,85533,85542,85557,85660,85743,85772,85783,85792,85840,85909,86062,86170,86424,86471,86600,86607,86850,86866,87075,87275,87380,87383,87419,87461,87477,87524,87584,87586,87598,87760,87817,87863,87887,87902,87995,88208,88235,88317,88549,88638,88642,88865,88890,88926,88957,89075,89166,89194,89238,89273,89384,89412,89826,89842,89865,89963,90072,90109,90133,90249,90371,90410,90583,90629,90657,90701,90758,90782,90793,90844,90853,91068,91131,91283,91297,91388,91405,91465,91581,91597,91797,91809,91907,92008,92197,92324,92403,92521,92548,92597,92607,92641,92645,92748,92800,92808,92890,92895,92975,93055,93075,93106,93224,93238,93314,93355,93356,93402,93591,93611,93637,93765,93835,94014,94082,94099,94140,94170,94260,94321,94334,94528,94820,94834,94880,95058,95268,95322,95337,95424,95553,95658,95793,95966,96034,96087,96115,96283,96545,96612,96677,96678,96819,97056,97063,97101,97127,97181,97213,97254,97278,97483,97595,97646,97662,97823,97832,97847,97868,97904,97921,97995,98127,98183,98197,98415,98459,98490,98546,98570,98596,98598,98707,98765,98889,98900,99044,99059,99064,99108,99127,99274,99501,99531,99835,99897,99923,99953",
      "17,19,30,520,548,552,556,571,636,645,664,677,681,689,706,750,785,816,818,859,920,981,1046,1120,1147,1155,1260,1274,1277,1305,1336,1338,1428,1469,1492,1539,1582,1585,1591,1696,1828,1842,1875,2392,4073,5038,5123,5173,5180,5192,5200,5222,5226,5242,5244,5280,5286,5291,5348,5364,5366,5406,5407,5419,5442,5473,5475,5500,5537,5624,5652,5688,5689,5709,5713,5728,5730,5810,5864,5922,5925,5979,6049,6063,6150,6155,6180,6214,6309,6319,6343,6349,6391,6453,6491,6577,6717,6726,6734,6748,6777,7097,7149,7179,7192,7335,7346,7364,7369,7408,7590,7685,7702,7885,7892,7961,7970,7987,8150,8170,8419,8449,8538,8576,8633,8685,8703,8719,8766,8863,8880,8912,9127,9187,9204,9287,9335,9941,10040,10163,10253,10305,10307,10334,10352,10408,10420,10460,10525,10613,10849,10974,10995,11177,11206,11690,11732,11754,11837,12125,14950,15352,16857,19493,22839,23086,23176,24275,24508,24514,24613,24641,24698,24808,25447,26802,28120,28232,29148,29239,29746,30720,31145,31606,35578,36490,38560,41667,42259,42670,42791,45164,45381,46505,48990,50017,50171,50200,50334,50481,50529,50571,50582,50721,50776,50783,50870,50899,50924,51187,51212,51273,51339,51439,51586,51592,51639,51750,51811,51820,51838,51878,52143,52275,52329,52393,52458,52593,52686,52721,52871,53019,53023,53127,53176,53197,53290,53304,53423,53541,53581,53769,53828,53993,54093,54188,54196,54266,54280,54779,55035,55044,55084,55148,55171,55276,55291,55387,55481,55548,55760,55770,55780,55821,55845,56049,56073,56077,56140,56257,56495,56534,56712,56719,56751,56769,56791,56929,56949,57077,57081,57083,57364,57717,57943,57954,57967,58002,58100,58109,58168,58234,58376,58468,58499,58500,58809,58813,58890,58959,59124,59180,59468,59479,59507,59526,59559,59681,60180,60206,60294,60344,60370,60496,60508,60529,60550,60629,60644,60873,61041,61047,61165,61185,61236,61247,61258,61275,61332,61358,61362,61366,61559,61683,61802,61819,61874,61885,61991,62113,62336,62377,62521,62558,62579,62612,62659,62673,62686,62815,62866,62874,62930,63034,63231,63360,63426,63579,63669,63706,63725,63816,63890,64010,64013,64076,64161,64176,64346,64408,64441,64500,64606,64662,64733,64769,64904,64926,64984,65175,65207,65356,65404,65414,65433,65560,65593,65658,65740,65901,65986,66053,66072,66089,66111,66147,66185,66189,66194,66208,66229,66315,66351,66485,66487,66597,66760,66922,66960,66989,67060,67087,67168,67305,67394,67538,67615,67706,67744,67754,67847,67982,68022,68081,68108,68111,68290,68466,68473,68607,68663,68704,68714,68963,68978,68979,69091,69302,69326,69330,69363,69430,69445,69477,69805,69861,69923,69929,70114,70267,70349,70419,70510,70586,70725,70790,70838,70966,71091,71107,71206,71326,71328,71431,71620,71730,71788,71937,71938,71970,72063,72187,72201,72221,72262,72271,72513,72571,72580,72748,72915,72916,72925,73113,73239,73329,73441,73442,73568,73631,73703,73794,73809,74008,74114,74172,74180,74193,74362,74382,74411,74643,74717,74754,74850,74868,74891,74900,74970,75041,75136,75153,75161,75357,75613,75896,76021,76022,76114,76396,76611,76646,76731,76746,76787,76861,76895,76904,77011,77041,77166,77198,77229,77384,77644,77831,78039,78111,78143,78152,78242,78303,78331,78419,78446,78465,78567,78619,78685,78708,78779,78813,78948,79147,79166,79207,79239,79286,79328,79353,79424,79446,79591,79623,79648,79765,79877,80138,80140,80146,80398,80440,80559,80694,80784,80792,81044,81089,81118,81167,81342,81402,81470,81504,81591,81661,81674,81745,81849,81901,81962,82027,82168,82279,82567,82601,82619,82731,82853,82933,83009,83051,83070,83073,83090,83273,83627,83636,83683,83774,83819,83870,83900,84042,84344,84398,84521,84767,84807,84943,84948,85036,85043,85101,85257,85306,85401,85663,85705,85706,85726,85785,85883,86028,86037,86202,86228,86324,86352,86371,86488,86542,86649,86733,86815,86821,86855,86990,87047,87359,87400,87404,87470,87512,87619,87666,87788,87807,87899,88053,88345,88365,88383,88393,88450,88693,88726,88951,88992,89021,89201,89204,89285,89390,89415,89489,89582,89653,89676,89749,89788,89834,89896,89998,90007,90037,90045,90102,90117,90155,90157,90220,90326,90457,90558,90577,90679,90714,90790,90830,91013,91052,91138,91237,91248,91346,91471,91547,91559,91656,91787,91887,91962,92318,92368,92618,92663,92675,92771,92847,92874,93095,93214,93218,93475,93697,93771,93790,93833,93900,94004,94041,94296,94331,94353,94393,94490,94532,94563,94979,95074,95269,95274,95276,95385,95490,95523,95631,95672,95796,95988,96107,96226,96245,96390,96428,96506,96513,96557,96665,96754,96851,96944,96959,96968,97041,97098,97227,97504,97586,97780,98091,98124,98134,98412,98444,98772,98890,99196,99240,99260,99324,99366,99479,99514,99519,99566,99756,99825,99960",
      "277,334,386,406,423,444,465,480,515,553,568,595,621,643,670,723,734,792,817,847,873,989,1021,1043,1050,1068,1141,1175,1327,1353,1363,1386,1404,1432,1437,1448,1477,1563,1636,1791,1833,1845,1903,1954,2943,4639,5144,5234,5254,5330,5338,5700,5849,5867,5918,6040,6135,6139,6154,6158,6197,6265,6410,6533,6813,6941,7121,7202,7216,7222,7229,7238,7339,7453,7571,7579,7644,7687,7743,7783,7787,7914,7992,8036,8072,8140,8190,8392,8519,8546,8635,8653,8682,8742,8885,9040,9065,9111,9121,9122,9275,9284,9351,9453,9862,9874,10184,10205,10241,10417,10424,10623,10657,10805,10907,10931,10948,11073,11090,11126,11336,11566,11632,11714,11769,11835,13724,13865,14089,14096,14372,14535,14611,14705,14802,14915,15330,15442,16130,16263,16599,17264,17436,17568,17588,17667,17966,17975,18122,18337,18594,19494,19681,20266,20296,20674,21050,21639,22612,22830,22850,22932,22984,24280,25132,25485,26422,26919,27025,27242,27491,27497,27791,28158,28273,29371,29405,29485,29733,29737,29996,30511,33930,33985,35844,36673,38746,39820,39905,39979,40677,41300,42336,43120,44121,45936,46999,47254,49604,49874,50058,50117,50128,50180,50193,50288,50317,50443,50497,50534,50631,50632,50744,50796,50863,50889,51047,51361,51421,51432,51444,51514,51551,51612,51623,51630,51636,51694,51754,51796,51845,52009,52119,52150,52269,52319,52347,52360,52404,52448,52461,52485,52535,52601,52606,52657,52730,52888,52913,52972,53146,53164,53232,53328,53334,53381,53466,53750,53814,54010,54018,54052,54255,54296,54420,54478,54719,54834,54877,54952,54985,55007,55394,55513,55525,55718,55746,55928,56230,56320,56321,56369,56436,56517,56553,56558,56602,56707,56740,56919,56966,57042,57072,57095,57173,57694,57748,57935,58108,58379,58406,58473,58631,58895,58919,58937,58970,59023,59295,59306,59319,59745,59832,59885,59902,60062,60083,60103,60171,60311,60351,60509,60582,60760,60846,60931,60934,60981,61057,61127,61266,61303,61305,61405,61586,61620,61668,61702,61882,61952,62261,62322,62324,62436,62539,62557,62573,62586,62646,62729,62750,62758,62810,62882,62911,63122,63151,63187,63209,63311,63355,63364,63591,64049,64245,64268,64340,64732,64751,64791,65034,65110,65163,65256,65359,65372,65384,65496,65561,65694,65782,65791,65879,65983,66011,66118,66263,66354,66445,66474,66504,66607,66652,66778,66933,66961,67133,67181,67232,67234,67237,67542,67577,67724,68002,68138,68301,68323,68367,68411,68485,68516,68580,68616,68634,68702,68860,68871,68874,68880,68881,68908,68946,68975,69032,69053,69061,69073,69178,69364,69565,69808,69907,69921,70019,70093,70161,70325,70531,70575,70635,70872,70882,70883,70970,71011,71087,71271,71286,71293,71443,71516,71577,71631,71669,71770,71882,71899,71953,72024,72193,72195,72211,72407,72444,72541,72592,72621,72731,72743,72905,72987,73116,73162,73196,73423,73725,73726,74392,74554,74680,74761,74928,74960,75000,75037,75144,75171,75175,75303,75395,75460,75499,75604,75691,75696,75713,75774,75821,75903,76088,76113,76143,76530,76580,76606,76676,76683,76754,76803,76863,76949,76960,77002,77016,77134,77238,77274,77338,77355,77433,77489,77528,77688,77785,77908,77946,78014,78186,78229,78259,78292,78422,78669,78763,78861,79016,79063,79089,79247,79339,79370,79406,79468,79504,79539,79580,79644,79645,79662,79784,79915,80020,80121,80176,80214,80296,80321,80415,80526,80573,80635,80692,80799,80851,81104,81175,81625,81635,81688,81721,81899,81913,81998,82006,82058,82162,82248,82323,82466,82505,82560,82723,82900,82965,83008,83110,83237,83330,83396,83404,83559,83593,83655,83679,83728,83783,83840,83986,84079,84197,84290,84291,84388,84390,84430,84603,84758,84925,85004,85073,85086,85300,85323,85394,85407,85455,85497,85674,85707,85781,85850,85863,85999,86022,86268,86490,86500,86591,86623,86712,86748,86798,86824,86870,86999,87045,87062,87102,87511,87580,87646,87748,87782,87870,87967,87986,88009,88159,88214,88318,88440,88518,88527,88637,88765,88833,88867,88964,88996,89054,89084,89161,89357,89367,89395,89417,89550,89587,89638,89986,89994,90017,90080,90233,90310,90432,90465,90531,90715,90849,91161,91250,91312,91323,91337,91418,91601,91749,91853,91871,91888,92021,92084,92103,92193,92229,92367,92429,92434,92455,92491,92500,92638,92721,92738,92829,92854,92926,92966,93056,93097,93213,93281,93301,93382,93423,93500,93506,93570,93696,93704,93755,93813,93920,94241,94343,94557,94573,94837,94882,94969,95005,95009,95039,95071,95089,95109,95137,95161,95190,95407,95497,95506,95619,95627,95719,95729,95778,95808,95899,95973,95976,95994,95997,96071,96130,96244,96263,96284,96301,96348,96402,96406,96434,96671,96817,96909,97179,97239,97353,97386,97414,97477,97618,97679,97950,98043,98051,98054,98092,98130,98168,98253,98358,98457,98469,98703,98777,98869,98884,99097,99099,99306,99329,99350,99352,99374,99448,99657,99706,99820,99879,99900",
      "297,338,341,356,364,486,559,584,585,615,638,656,711,811,890,957,998,1083,1105,1111,1156,1224,1262,1290,1470,1521,1616,1727,1756,1856,1861,2095,2111,2193,2228,2441,2559,2797,2801,2835,2939,2949,3017,3182,3276,3312,3340,3343,3495,3602,3620,3733,3766,3774,4123,4151,4183,4341,4411,4692,4693,4727,4739,4880,4896,5098,5189,5270,5340,5375,5440,5453,5494,5543,5573,5590,5643,5673,5698,5802,5830,5939,5985,6075,6102,6137,6165,6276,6474,6608,6669,6732,6847,7001,7064,7070,7083,7240,7241,7282,7418,7490,7605,7695,8068,8142,8469,8603,8733,8850,8883,8977,9053,9126,9138,9144,9173,9265,9370,9411,9455,9773,10087,10156,10182,10335,10463,10592,10651,10700,10705,10748,10815,10848,10927,10962,10971,11076,11115,11149,11165,11345,11349,11434,11493,11503,11510,11579,11608,11684,11809,11830,12009,12047,12067,12165,12292,12306,12382,12622,12736,12812,12937,13031,13133,13355,13409,13450,13472,13497,13542,13544,13673,13871,13920,14060,14117,14261,14686,14696,14856,14864,15243,15256,15331,15531,16049,16113,16172,16419,16437,16491,16625,16764,16846,16891,17160,17189,17494,17583,17707,17708,18608,18643,18648,18675,18680,18740,18750,18856,19007,19122,19214,19505,19619,19639,19771,19843,20201,20249,20490,20751,20841,20996,21038,21361,21540,21658,21683,21711,21782,21890,22140,22258,22613,22927,22968,22991,23087,23189,23247,23376,23561,23566,23593,23649,23666,23689,23776,23805,24109,24134,24313,24338,24532,24608,24733,25321,25326,25600,25801,25881,25940,26333,26684,26690,26729,26798,26874,26981,27046,27195,27323,27340,27355,27441,27644,27688,27767,27923,28221,28294,28470,28555,28765,28842,28846,28889,28919,28982,29128,29165,29171,29185,29211,29255,29312,29316,29349,29382,29603,29776,29810,29840,29846,30032,30130,30488,30913,30990,32315,32434,32450,32566,33628,34103,34121,34480,34832,35092,35556,35618,35705,35709,35740,35755,35997,36012,36259,36267,36383,36686,36888,37019,37035,37066,37114,37127,37138,37305,37564,37636,37802,37807,37939,38009,38085,38190,38229,38669,38699,38812,38877,38931,38986,39119,39501,39687,39841,39876,39989,39994,40362,40547,40589,40618,40734,40799,40826,40944,41350,41401,42454,42463,42529,42631,42788,42798,43230,43711,43827,43962,44108,44123,44606,44619,44749,44976,45188,45258,45435,45633,45731,45802,46268,46546,47200,47325,47927,47954,48166,48569,48649,49062,49844,50329,50396,50425,50453,50671,50786,50855,50912,50918,50978,50993,51025,51029,51222,51239,51253,51348,51372,51384,51399,51456,51495,51545,51562,51580,51676,52102,52158,52214,52244,52264,52268,52308,52337,52359,52371,52424,52736,52826,52965,53017,53097,53283,53338,53346,53413,53467,53529,53679,53725,53726,53761,53817,53825,53918,54043,54045,54124,54140,54211,54538,54539,54700,54792,54892,54942,55011,55013,55067,55112,55122,55227,55235,55529,55583,55652,55698,55861,55870,55888,55913,55919,55936,55949,55993,56002,56165,56240,56315,56328,56437,56497,56626,56732,56816,56893,56956,56983,57205,57294,57389,57441,57521,57539,57545,57592,57621,57651,57680,57765,57780,57810,57873,58067,58169,58229,58365,58458,58506,58530,58566,58582,58615,58623,58701,58713,58728,58737,58819,58828,58881,59052,59182,59264,59407,59426,59545,59748,59862,59864,60047,60052,60146,60159,60260,60304,60327,60362,60397,60488,60558,60614,60729,60867,61053,61132,61216,61250,61338,61718,61814,61854,61856,61865,61911,62010,62025,62196,62220,62233,62533,62607,62654,62705,62751,62754,62755,62764,62895,62907,62909,62947,63128,63313,63370,63386,63441,63520,63539,63701,63736,63834,64017,64061,64062,64131,64235,64244,64254,64330,64352,64362,64390,64431,64462,64483,64524,64679,64924,65145,65225,65241,65281,65502,65542,65601,65752,65803,65828,65853,65971,66007,66066,66203,66240,66341,66385,66456,66493,66642,66705,66758,66822,66823,66951,67121,67143,67186,67230,67267,67303,67361,67371,67402,67440,67523,67748,67787,67840,67920,67927,68062,68104,68235,68360,68381,68387,68489,68524,68534,68637,68728,68740,68762,68795,68861,68914,69089,69101,69207,69238,69275,69340,69525,69682,69739,69902,70025,70031,70095,70516,70617,70626,70658,70676,70751,70852,70962,71294,71343,71371,71401,71417,71446,71536,71625,71636,71708,71869,71883,71909,71936,71965,72484,72649,72699,72774,72836,72876,72885,73000,73142,73235,73243,73247,73375,73381,73399,73408,73475,73508,73649,73659,73688,73811,73854,73980,73994,74026,74121,74123,74200,74210,74270,74478,74510,74540,74623,74721,74737,74854,74993,75043,75139,75243,75434,75620,75762,75807,75870,76151,76200,76479,76490,76634,76715,77008,77061,77114,77250,77292,77334,77484,77485,77556,77802,77807,77809,77839,77895,77939,77963,78051,78183,78200,78214,78276,78309,78383,78435,78706,78906,78912,78944,79051,79152,79510,79517,79583,79602,79613,79689,79807,79819,80017,80194,80333,80367,80386,80442,80534,80621,80689,80810,80865,80947,80980,81014,81145,81238,81251,81281,81380,81405,81495,81573,81742,81777,81823,81857,82125,82297,82306,82320,82409,82500,82634,82735,82893,82998,83109,83363,83586,83597,83623,83668,83684,83702,84085,84113,84342,84454,84462,84581,84706,84849,84896,84969,85029,85084,85092,85111,85177,85424,85682,85691,85754,85902,86070,86094,86213,86246,86310,86464,86593,86677,86781,86922,86952,86970,86971,86973,87142,87211,87239,87314,87394,87440,87631,87657,87686,87691,87802,87840,87952,88052,88187,88197,88207,88333,88409,88422,88588,88589,88848,89085,89172,89296,89446,89486,89493,89568,89571,89584,89617,89679,89680,89712,89753,89978,90001,90022,90047,90090,90093,90295,90477,90574,90576,90621,90668,90732,90902,90942,91029,91051,91080,91244,91532,91610,91703,91796,91898,91925,91968,91977,92094,92138,92211,92254,92289,92290,92315,92352,92381,92546,92614,92665,92692,92757,92825,92948,92959,92976,92990,93019,93021,93175,93257,93521,93546,93573,93613,93758,93776,93817,93843,93883,93951,93991,94015,94058,94120,94162,94164,94193,94338,94379,94442,94598,94786,94873,94960,95066,95097,95201,95309,95364,95453,95469,95824,96046,96158,96242,96248,96360,96373,96504,96526,96614,96738,96739,96771,96789,96903,97109,97114,97132,97280,97375,97403,97404,97444,97585,97598,97699,97709,97735,97790,97802,97946,97963,97980,97996,98034,98172,98316,98341,98353,98387,98399,98528,98586,98727,98733,98763,98791,98818,98856,98871,98968,99035,99080,99102,99107,99245,99372,99403,99477,99550,99571,99651,99662,99712,99772,99867,99876",
      "87,320,339,374,414,428,641,698,733,903,965,984,995,1072,1100,1110,1149,1166,1198,1211,1236,1239,1275,1291,1310,1345,1410,1513,1542,1548,1572,1590,1593,1608,1698,1867,1868,1923,2087,2092,2608,2699,2719,2800,3107,3223,3355,3370,3701,3888,4087,4145,4217,4359,4530,4716,4806,5018,5023,5043,5045,5065,5079,5083,5145,5181,5196,5215,5224,5231,5265,5273,5274,5283,5311,5314,5395,5464,5496,5523,5538,5687,5741,5778,5801,5845,5899,5903,5949,6051,6080,6173,6202,6277,6285,6331,6393,6433,6442,6494,6498,6507,6553,6607,6648,6769,6817,6920,6982,7063,7169,7396,7423,7457,7495,7522,7535,7779,7808,7907,7929,8001,8197,8237,8311,8336,8427,8494,8642,8767,9140,9327,9441,9462,9473,9669,9700,9866,9963,9968,10038,10285,10529,10619,10952,11103,11221,11464,11524,11542,11634,11844,11872,12139,12505,12874,12931,12936,13156,13165,13176,13648,13729,14658,14736,14835,14848,14904,15150,15483,15545,16059,16804,17274,17983,18086,18630,18645,18779,19034,19200,19405,19483,19824,20746,20789,20928,21359,21561,22010,22226,23058,23486,23511,23596,23672,23767,23957,24356,24901,25021,25249,25300,25775,25895,26453,26562,26744,26861,27252,27341,27372,27425,27879,27975,28376,28573,28675,28723,28741,29078,29442,29486,29663,29929,30694,31477,31709,33739,34087,34319,34406,35083,35605,35682,35788,36184,36744,36867,37001,37070,37130,37687,37754,37944,38250,38648,38689,39120,39179,39180,39627,39632,39663,39853,39931,41029,41145,41232,41435,42683,42841,42921,43809,44229,44347,44921,45500,45692,45752,45931,45941,46336,47088,47775,48314,48416,48664,49469,49669,50032,50044,50048,50349,50381,50426,50461,50463,50487,50605,50665,50726,50917,51158,51195,51214,51282,51284,51303,51377,51427,51431,51540,51710,51877,51912,51998,52000,52085,52200,52256,52569,52695,52808,52833,52989,53072,53078,53208,53294,53536,53602,53630,53665,53843,53909,54050,54084,54094,54237,54256,54692,54794,54806,54831,54955,55012,55062,55069,55143,55155,55201,55246,55343,55351,55361,55399,55414,55547,55550,55557,55639,55701,55772,55963,55976,56012,56135,56340,56666,56671,56899,56942,57141,57164,57204,57250,57261,57307,57569,57587,57593,57664,57776,57797,57857,57971,57986,58151,58344,58601,58846,59083,59088,59131,59380,59381,59629,59724,59740,59744,59777,59861,59876,59928,59959,59994,60116,60122,60150,60272,60290,60376,60387,60568,60792,60812,60875,60998,61018,61049,61054,61300,61306,61397,61447,61477,61793,62037,62210,62265,62326,62626,62794,62859,62905,62996,63057,63096,63100,63106,63234,63244,63299,63307,63319,63406,63449,63494,63495,63600,63616,63689,63729,63747,63768,63879,63915,63994,64582,64728,64755,64857,65039,65077,65100,65118,65296,65328,65435,65570,65799,65815,65827,65848,65883,66048,66202,66255,66261,66482,66532,66564,66588,66647,66650,66862,66869,66879,66907,67007,67056,67086,67190,67268,67325,67495,67497,67581,67667,67756,67873,67912,68256,68279,68300,68345,68413,68519,68713,68736,68780,68856,68870,68919,69015,69115,69148,69229,69256,69258,69403,69470,69753,69771,69847,70106,70120,70197,70453,70544,70555,70556,70557,70667,70717,70778,70814,70892,71007,71015,71024,71031,71133,71140,71221,71226,71248,71264,71291,71339,71365,71426,71433,71510,71615,71649,71688,71713,71815,71908,71919,71940,72039,72406,72414,72477,72594,72648,72696,72700,72718,72753,72783,72797,72804,72850,72993,73160,73191,73240,73265,73267,73385,73635,73643,73664,73869,73965,73987,74052,74340,74468,74496,74614,74703,74808,75047,75152,75212,75262,75382,75441,75448,75539,75564,75608,75658,75772,75932,76050,76109,76134,76196,76209,76225,76280,76315,76637,76694,76744,76751,76802,77154,77231,77264,77312,77553,77659,77677,77685,77968,77977,78201,78206,78238,78265,78284,78525,78638,78720,78850,78858,78938,79020,79024,79095,79176,79179,79258,79287,79400,79593,79675,79686,79702,79748,79803,79858,79938,79946,79954,80030,80249,80283,80325,80430,80468,80513,80529,80588,80634,80701,80741,80789,80832,80951,80992,81020,81028,81038,81083,81150,81185,81243,81274,81316,81427,81478,81499,81513,81633,81670,81719,81766,81934,82160,82195,82474,82507,82512,82762,82803,82804,82941,83142,83164,83299,83388,83522,83536,83552,83738,83838,84154,84231,84261,84321,84346,84441,84586,84594,84701,84707,84795,84813,84826,84883,84922,85080,85121,85232,85444,85460,85535,85565,85791,85859,85954,86318,86448,86522,86550,86751,86774,87057,87104,87157,87169,87238,87266,87269,87282,87315,87502,87637,87638,87762,87832,87848,87906,87908,87917,87979,88094,88307,88550,88668,88691,88706,88785,88882,88883,88949,88960,89016,89088,89150,89179,89301,89326,89333,89382,89429,89454,89474,89576,89656,89675,89708,89797,89840,89854,89944,89988,90078,90122,90149,90245,90251,90348,90375,90468,90723,90820,90892,90905,90931,90961,90968,91025,91302,91313,91329,91440,91454,91464,91589,91622,91623,91658,91746,91866,91927,92048,92130,92502,92515,92539,92583,92622,92685,92955,92973,92993,93010,93222,93287,93379,93394,93436,93507,93528,93633,93643,93665,93725,93750,93830,94168,94256,94265,94789,94956,95003,95083,95202,95242,95317,95410,95463,95547,95582,95629,95689,95701,95774,95838,95987,95995,96001,96060,96155,96228,96372,96415,96443,96572,96619,96624,96682,96689,96838,96913,97152,97261,97418,97493,97510,97567,97589,97940,98230,98314,98393,98406,98461,98472,98603,98648,98815,98840,98876,98886,98961,98995,99065,99096,99155,99158,99208,99257,99412,99669,99670,99743,99749,99780",
      "40,201,206,226,244,296,315,521,626,729,883,888,1029,1128,1318,1325,1348,1503,1652,1678,1749,1846,1859,1866,1937,1961,1970,1977,1980,1985,1991,1994,2002,2017,2036,2051,2052,2053,2068,2070,2106,2157,2166,2174,2185,2213,2238,2252,2273,2281,2328,2331,2340,2345,2370,2376,2383,2384,2385,2401,2416,2431,2471,2473,2510,2540,2545,2601,2618,2625,2652,2655,2663,2682,2731,2756,2847,2859,2866,2874,2878,2904,2920,2931,2946,2988,3027,3055,3116,3122,3124,3212,3284,3293,3316,3385,3386,3391,3392,3393,3435,3438,3456,3469,3486,3501,3502,3506,3511,3521,3531,3533,3540,3547,3644,3648,3654,3670,3708,3762,3778,3821,3871,3912,3958,4011,4028,4029,4068,4072,4146,4165,4178,4181,4229,4261,4286,4364,4381,4439,4447,4451,4471,4481,4486,4499,4525,4531,4534,4536,4542,4544,4549,4568,4578,4583,4586,4596,4621,4622,4694,4713,4714,4721,4736,4748,4750,4798,4828,4836,4839,4855,4865,4891,4917,4954,4978,4985,4993,5058,5370,5427,5535,5551,5690,5716,5756,5764,5782,5857,5927,5970,6065,6081,6120,6132,6206,6221,6234,6311,6358,6412,6530,6575,6595,6682,6697,6747,6752,6768,6825,6828,7022,7042,7098,7175,7237,7348,7353,7430,7459,7469,7518,7534,7560,7705,7797,7803,7836,7839,7900,7911,7963,8024,8167,8273,8284,8293,8320,8322,8337,8353,8394,8465,8585,8799,8898,8904,8925,8927,8959,8966,8986,9205,9356,9373,9377,9399,9738,10102,10204,10272,10313,10387,10561,10575,10685,10727,10783,10802,10976,11005,11346,11665,11696,11789,12207,12312,12389,12437,12454,12951,13023,13336,13778,13869,13884,13891,14053,14300,14306,14367,14477,14489,14623,15080,15248,15485,15503,15562,16115,16141,16154,16246,16494,16852,16866,16910,16925,17152,17230,17322,17328,17397,17806,17856,17947,18026,18098,18269,18423,18642,18706,18735,18864,19190,19312,19607,19618,20151,20181,20261,20475,20599,20826,20994,21040,21101,21141,21261,21311,21366,21459,21594,21681,21948,22174,22328,22490,22722,23060,23250,23333,23455,23762,23909,23925,24014,24484,24586,24986,25030,25184,25234,25282,25291,25505,25638,25932,25998,26013,26061,26062,26215,26544,26553,26556,26575,26577,26591,26599,26611,26620,26679,26747,26789,26954,27035,27183,27358,27453,27476,27505,27543,27569,27589,27611,27685,27729,27931,27964,27982,27994,27995,28005,28007,28025,28032,28043,28049,28051,28052,28079,28105,28108,28129,28142,28145,28147,28227,28405,28420,28450,28469,28472,28497,28500,28502,28514,28577,28589,28604,28608,28610,28625,28632,28637,28669,28807,28970,29011,29050,29085,29207,29308,29327,29358,29389,29523,29526,29530,29534,29572,29591,29608,29617,29625,29652,29671,29677,29678,29721,29772,29832,29957,30020,30049,30078,30107,30136,30165,30194,30223,30252,30281,30310,30339,30368,30397,30426,30455,30484,30513,30542,30571,30600,30629,30658,30687,30716,30745,30774,30803,30832,30861,30890,30919,30948,30977,31006,31035,31064,31093,31122,31151,31180,31209,31238,31267,31296,31325,31354,31370,31383,31412,31441,31470,31499,31528,31557,31586,31615,31644,31673,31702,31731,31760,31789,31818,31847,31876,31905,31934,31963,31992,32021,32050,32079,32108,32137,32166,32195,32224,32282,32311,32340,32369,32398,32427,32456,32485,32514,32543,32572,32601,32630,32659,32688,32717,32746,32775,32804,32833,32862,32891,32920,32949,32978,33007,33036,33065,33094,33123,33152,33181,33210,33239,33268,33297,33326,33355,33384,33413,33442,33471,33500,33529,33558,33587,33616,33645,33674,33703,33732,33761,33790,33819,33848,33877,33906,33935,33964,33993,34022,34051,34080,34109,34138,34167,34196,34225,34254,34283,34312,34341,34370,34399,34428,34457,34486,34544,34573,34602,34631,34660,34689,34718,34747,34776,34805,34834,34863,34892,34921,34950,34979,35007,35055,35065,35079,35081,35109,35158,35183,35211,35216,35272,35299,35321,35346,35383,35420,35462,35491,35539,35598,35657,35730,35746,35789,35848,35873,35919,35978,36037,36064,36072,36100,36160,36219,36290,36319,36349,36425,36480,36539,36598,36662,36721,36738,36780,36803,36851,36880,36910,36920,36969,36989,37041,37100,37116,37159,37223,37241,37282,37341,37377,37412,37471,37530,37602,37661,37689,37720,37745,37747,37784,37843,37857,37902,37930,37973,38000,38032,38091,38163,38222,38281,38345,38404,38423,38427,38463,38534,38593,38619,38652,38724,38783,38842,38858,38906,38916,38965,38975,39024,39061,39095,39154,39213,39285,39305,39344,39403,39482,39526,39585,39656,39715,39736,39774,39846,39964,39985,40124,40153,40182,40240,40269,40298,40327,40356,40414,40472,40501,40530,40559,40588,40617,40646,40675,40704,40733,40762,40791,40820,40849,40878,40936,40965,40994,41023,41052,41081,41110,41139,41168,41197,41226,41255,41284,41313,41342,41371,41400,41429,41458,41487,41516,41545,42015,42044,42073,42102,42131,42191,42221,42240,42281,42335,42356,42431,42446,42448,42518,42593,42639,42668,42703,42755,42801,42830,42842,42856,42871,42905,42922,42928,42935,42962,42992,43020,43067,43103,43107,43142,43180,43229,43261,43304,43330,43342,43352,43379,43391,43466,43474,43484,43541,43556,43566,43616,43659,43677,43703,43778,43853,43881,43897,43952,43983,43984,44015,44024,44053,44090,44127,44128,44146,44147,44177,44206,44252,44275,44327,44414,44438,44467,44489,44564,44577,44602,44624,44625,44651,44678,44689,44726,44732,44739,44770,44801,44840,44888,44923,44963,45003,45038,45064,45111,45121,45125,45159,45191,45220,45278,45351,45408,45426,45443,45466,45482,45538,45595,45653,45665,45674,45725,45743,45782,45797,45840,45912,45969,46004,46014,46031,46048,46062,46072,46093,46124,46138,46141,46148,46228,46234,46264,46276,46319,46367,46408,46426,46429,46436,46452,46550,46563,46612,46664,46670,46693,46696,46712,46728,46730,46739,46770,46799,46834,46858,46900,46911,46937,47016,47044,47056,47066,47079,47099,47110,47135,47151,47156,47158,47168,47269,47288,47322,47328,47338,47353,47410,47429,47522,47548,47558,47569,47582,47586,47635,47666,47685,47696,47716,47733,47769,47791,47801,47817,47818,47861,47869,47928,47964,47975,47999,48015,48080,48149,48192,48205,48215,48221,48235,48244,48278,48283,48320,48322,48389,48444,48474,48502,48516,48627,48701,48723,48726,48744,48754,48760,48785,48822,48846,48871,48928,48968,49007,49118,49141,49255,49263,49264,49309,49320,49331,49383,49441,49544,49557,49664,49673,49678,49719,49771,49841,49842,49848,49891,49899,49901,49939,49960,49972,49995,50146,50235,50392,50399,50491,50689,50730,50762,50847,50896,51008,51049,51153,51341,51693,51791,51854,51923,51971,52211,52281,52379,52439,52463,52480,52532,52759,52813,52829,52834,52953,53150,53206,53287,53648,53697,53742,53752,53754,53800,53884,54016,54078,54272,54367,54499,54684,54805,55034,55080,55237,55471,55663,55704,55716,55720,55762,55979,55984,56236,56549,56725,56758,56814,56885,56957,56985,57223,57275,57420,57449,57527,57604,57697,57786,57836,57982,58014,58146,58298,58327,58455,58521,58592,58638,58730,58740,58951,58999,59132,59141,59156,59177,59291,59292,59318,59334,59376,59452,59484,59576,59609,59657,59663,59701,59811,59882,59905,59910,59969,59973,60048,60085,60128,60149,60195,60230,60259,60367,60400,60696,60725,60732,60761,60796,60975,61001,61145,61196,61313,61320,61360,61396,61407,61497,61519,61627,61694,61734,61925,61929,61961,62135,62241,62481,62520,62535,62581,62712,62768,62868,62922,63161,63371,63452,63455,63496,63642,63796,63825,63962,63980,63998,64233,64506,64535,64831,64863,64877,64889,65105,65106,65195,65201,65337,65398,65545,65584,65667,65849,65851,65954,65982,66052,66113,66135,66155,66215,66268,66348,66408,66544,66766,66774,66832,66909,66998,67192,67242,67397,67407,67520,67590,67614,67645,67679,67713,67737,67831,67954,67988,68098,68131,68253,68273,68289,68308,68455,68470,68502,68517,68530,68685,68732,68741,68754,68813,68844,68952,69029,69071,69293,69428,69689,69748,70273,70291,70324,70371,70473,70613,70621,70685,70824,70957,70993,71311,71508,71641,71700,71835,71873,71877,71984,72006,72015,72105,72124,72152,72260,72308,72360,72412,72853,72870,73117,73200,73208,73222,73254,73284,73307,73648,73752,73777,73804,73808,73817,73843,73931,73942,74065,74145,74275,74330,74536,74551,74556,74576,74662,74681,74783,74894,74918,75010,75039,75082,75093,75151,75227,75263,75277,75287,75307,75326,75394,75450,75630,75670,75693,75704,75839,75899,75948,75953,75960,75999,76000,76074,76201,76228,76245,76290,76338,76344,76371,76408,76454,76480,76572,76759,76871,76903,76911,76930,76938,76990,77049,77075,77127,77256,77275,77374,77402,77403,77602,77634,77646,77724,77779,77796,77914,77955,78088,78208,78244,78387,78389,78420,78503,78627,78654,78737,78798,78807,78873,78981,78982,79023,79274,79307,79548,79606,79630,79710,79930,80125,80155,80199,80346,80361,80544,80545,80549,80777,80816,80889,80978,81275,81293,81463,81473,81503,81556,81575,81594,81656,81799,82023,82042,82117,82176,82294,82418,82533,82544,82574,82691,82716,82857,82866,82883,83023,83068,83170,83381,83470,83545,83612,83685,83746,83908,83987,84116,84380,84455,84458,84527,84537,84551,84711,84936,84982,84994,85019,85020,85060,85100,85200,85282,85319,85461,85478,85583,85650,85749,85949,85978,86119,86129,86515,86533,86555,86573,86621,86683,86752,86767,86814,86845,86908,86942,87061,87133,87361,87591,87656,87854,87869,87898,87900,88036,88076,88085,88106,88260,88273,88533,88630,88707,88764,88944,88965,88981,88994,89059,89122,89248,89250,89294,89338,89383,89408,89428,89556,89662,89849,89861,89907,90126,90145,90178,90260,90339,90445,90471,90532,90571,90781,90817,90855,90960,91124,91247,91264,91357,91408,91455,91482,91587,91594,91768,91778,91835,92026,92063,92239,92374,92392,92399,92402,92422,92580,92587,92651,92666,92860,92940,93009,93180,93272,93319,93894,93911,94055,94071,94104,94115,94269,94318,94367,94736,94748,94806,94864,94970,94991,95053,95068,95114,95246,95280,95372,95419,95576,95621,95882,95971,95981,96004,96079,96084,96343,96371,96429,96555,96666,96759,96760,96915,97096,97115,97118,97382,97495,97538,97581,97688,97728,97730,97825,97840,97883,97937,97957,98059,98177,98214,98377,98407,98512,98693,98931,98991,99063,99170,99200,99229,99383,99457,99520,99594,99639,99791,99814,99848,99938",
      "156,234,259,292,303,319,407,420,446,469,470,600,627,654,709,721,804,809,895,906,1042,1087,1183,1306,1387,1390,1433,1483,1511,1514,1595,1599,1643,1695,2448,2516,2759,3234,3240,3388,3398,4081,4380,4956,5010,5051,5060,5078,5104,5199,5404,5405,5409,5435,5596,5674,5786,5876,5902,6006,6114,6162,6315,6344,6384,6401,6438,6562,6646,6706,6912,7101,7250,7276,7380,7393,7517,7580,7626,7686,7760,7784,7813,8094,8335,8408,8456,8499,8567,8580,8656,8749,8841,8914,8945,9097,9264,9338,9409,9575,9905,9961,9980,10062,10091,10143,10223,10316,10412,10446,10490,10757,10776,10850,10876,11039,11162,11244,11363,11377,11581,11588,11804,11812,11887,11953,11996,12016,12060,12122,12167,12422,12647,12879,13333,13372,13610,14035,14350,14632,14634,14737,14798,14900,14977,15198,15298,15459,15764,15786,15990,16637,16794,17190,17220,17356,17431,17551,17633,17741,17927,18116,18495,18520,18549,18628,18690,18893,18965,19351,19556,19766,20069,20624,21406,22132,22546,22664,22738,22754,23020,23111,23239,23530,23722,23754,23826,24241,24366,24654,24667,24686,24790,24797,24861,25039,25043,25728,26233,26727,26907,26957,27060,27234,27889,27919,28530,28770,28782,28984,29130,29451,29514,29527,29732,29827,29856,29901,35875,36177,36242,36444,36933,37244,37615,37671,37760,38053,38054,38118,38245,38319,38428,38490,38550,38760,38762,39563,39693,39917,39927,40211,40778,40907,42378,43457,45105,45183,45304,45315,46075,47540,50071,50116,50137,50142,50149,50221,50226,50231,50291,50308,50356,50389,50687,50770,50817,51028,51119,51146,51288,51409,51423,51428,51464,51511,51524,51619,51689,51861,51932,52239,52445,52484,52504,52637,52660,52693,52745,52932,52964,52992,53068,53105,53143,53531,53547,53661,53706,53710,53975,54035,54164,54200,54232,54554,54627,54734,54855,54859,54902,54994,55039,55041,55051,55136,55196,55295,55369,55509,55707,55713,55743,55789,55965,56030,56031,56084,56085,56241,56374,56421,56466,56487,56492,56511,56515,56636,56705,56731,56774,56781,56788,56970,56993,57020,57079,57187,57407,57540,57761,57886,58030,58121,58236,58247,58281,58413,58469,58553,58784,58797,58818,58962,58982,59028,59125,59223,59280,59337,59349,59418,59460,59668,59847,59978,60056,60071,60201,60241,60299,60312,60355,60404,60460,60469,60557,60768,60840,60923,61096,61144,61278,61290,61686,61706,61741,61799,61812,61894,62076,62102,62347,62451,62547,62744,62849,62946,63136,63408,63422,63429,63489,63534,63698,63934,63944,64112,64166,64349,64501,64619,64657,64722,64785,64925,65164,65180,65362,65516,65647,65654,65684,65747,65864,65967,66062,66112,66161,66165,66242,66282,66398,66669,66711,66714,66956,67015,67058,67120,67129,67151,67264,67454,67491,67493,67507,67636,67802,67818,68001,68007,68270,68388,68417,68451,68503,68520,68553,68599,68744,68776,68818,68837,68970,69040,69206,69295,69301,69372,69375,69449,69457,69460,69606,69729,69737,69815,69970,70014,70134,70137,70146,70304,70328,70352,70439,70444,70475,70524,70771,70843,71073,71150,71193,71265,71276,71350,71428,71436,71458,71559,71591,71617,71664,71682,71758,71831,71858,72114,72190,72321,72392,72604,72629,72656,72762,72788,72869,72958,72981,72995,73083,73153,73422,73654,73746,73780,73911,73919,73959,73960,73966,74119,74166,74213,74316,74350,74427,74479,74547,74750,74794,74809,75025,75106,75126,75562,75622,75759,75855,75907,75940,75987,76028,76049,76160,76272,76297,76424,76442,76478,76591,76714,76834,76881,76888,76954,76965,77029,77096,77152,77168,77191,77201,77228,77268,77367,77469,77470,77488,77563,77601,77725,77837,77998,78099,78149,78181,78330,78608,78694,78702,78809,78889,79028,79109,79229,79253,79316,79319,79364,79382,79597,79626,79688,79833,80004,80021,80192,80280,80385,80524,80830,80874,80924,81000,81007,81043,81109,81192,81329,81397,81472,81507,81702,81716,81768,81937,82082,82257,82338,82357,82451,82622,82850,83095,83195,83306,83374,83432,83785,83830,83904,83923,83930,83966,83998,84008,84029,84031,84135,84303,84314,84529,84617,84634,84703,84789,84798,84819,84828,84898,84980,85027,85046,85067,85093,85098,85156,85303,85332,85373,85438,85469,85544,85690,85757,85808,85827,85852,85898,85966,86057,86074,86092,86118,86152,86221,86259,86271,86303,86373,86569,86578,86655,86667,86777,86878,87037,87038,87039,87044,87054,87137,87279,87378,87382,87489,87507,87521,87675,87689,87969,88061,88114,88132,88198,88203,88272,88335,88350,88366,88487,88573,88850,88904,88911,89071,89078,89169,89290,89339,89388,89391,89475,89508,89607,89661,89725,89727,89732,89899,89933,89980,89990,90077,90187,90255,90265,90282,90438,90622,90651,90733,90797,90861,90908,90972,91028,91069,91187,91273,91304,91376,91431,91616,91711,91815,91837,92029,92044,92380,92535,92568,92592,92730,92762,92792,92837,93023,93061,93104,93140,93141,93145,93158,93204,93307,93344,93370,93413,93445,93481,93488,93530,93625,93777,93821,93861,93992,94016,94139,94151,94380,94564,94619,94631,94672,94707,94768,94796,94817,94922,94968,94978,95076,95146,95223,95308,95353,95588,95617,95656,95784,95803,95865,95889,95915,95926,96006,96085,96105,96211,96335,96457,96827,96867,96954,97176,97259,97338,97455,97575,97668,97993,98082,98117,98122,98196,98245,98247,98325,98405,98418,98429,98448,98552,98615,98654,98779,98821,98829,98839,98914,98946,99014,99085,99095,99277,99283,99357,99363,99422,99450,99543,99546,99629,99732,99830,99965",
      "83,84,124,195,282,419,540,580,763,771,773,898,912,1064,1097,1174,1184,1244,1248,1250,1450,1516,1557,1570,1614,1781,1901,2269,2280,2557,2570,2701,2883,3131,3194,3374,3416,3609,3728,3775,3789,3855,3939,4199,4348,4358,4429,4490,4521,4538,4920,5094,5136,5165,5184,5343,5444,5468,5539,5581,5752,5787,5792,5794,5839,5863,5934,6000,6024,6170,6245,6405,6480,6555,6690,6715,6810,6876,6877,6919,6969,7037,7085,7174,7223,7272,7344,7473,7483,7500,7583,7591,7668,7820,7903,7958,7974,8209,8239,8269,8287,8331,8457,8544,8632,8765,8773,8782,8906,9003,9046,9125,9203,9211,9334,9449,9489,9742,9768,9825,9827,9889,9911,9979,10127,10145,10238,10404,10423,10578,10661,10770,10782,10817,10836,10918,10973,10981,10989,10996,11043,11144,11329,11435,11437,11466,11469,11546,11561,11578,11604,11683,11705,11707,11788,11792,11796,11917,12054,12467,12515,12565,12611,12682,12752,12822,12864,13033,13079,13487,13499,13631,13941,14140,14331,14348,14498,14583,14926,15172,15252,15271,15334,15336,15428,15489,15555,15632,16126,16381,16492,16530,16557,16563,16685,16753,16905,16917,16923,17019,17340,17775,17787,18099,18121,18126,18300,18330,18781,18833,18870,19101,19225,19305,19489,19540,19558,19573,19801,19980,19985,19998,20002,20029,20110,20199,20282,20288,20395,20401,20521,20531,20581,20618,20851,21134,21244,21245,21321,21344,21389,21513,21525,21706,21771,22337,22347,22350,22529,22575,22868,22964,23012,23100,23211,23364,23402,23443,23536,23653,23665,23872,23910,24122,24166,24247,24322,24327,24401,24619,24779,24864,24896,24942,24984,25358,25520,25531,25550,25786,25835,26384,26406,26440,26451,26766,26872,27114,27318,27675,27716,27759,27786,27792,27807,27863,27916,27951,28034,28259,28296,28336,28345,28359,28435,28718,28812,28818,28910,28999,29089,29276,29392,29426,29447,29453,29491,29692,29694,29807,29888,30017,30024,30206,30420,30510,30815,31126,31931,32286,32424,32569,32598,32605,32917,33396,33860,33976,34295,34483,34628,34933,35231,35360,35551,35571,35601,35812,35821,35886,35895,36014,36041,36111,36216,36368,36496,36536,36550,36595,36750,37076,37249,37300,37312,37381,37508,37553,37618,37812,38101,38119,38130,38383,38814,38825,38874,38880,38883,38977,39323,39426,39441,39479,39509,39582,39609,40253,40654,41002,41216,41274,42053,42165,42228,42237,42363,42535,42545,42654,42840,42895,43354,43549,43810,43843,43867,44581,44599,45187,45237,45240,45302,45337,45358,45550,45566,45662,45676,45981,45982,46162,46274,46435,46869,46921,46961,47617,47661,47712,47991,48004,48023,48124,48191,48749,48845,49655,49785,50451,50588,50594,50596,50667,50829,50853,50939,51005,51033,51040,51059,51094,51099,51149,51190,51247,51305,51309,51334,51376,51499,51919,51986,51991,52092,52093,52302,52311,52316,52742,52850,52853,52884,52912,52935,52976,53137,53247,53286,53454,53644,53715,53733,53849,53855,53979,53997,54098,54229,54456,54693,54981,55606,55657,55665,55702,55848,56065,56091,56225,56614,56687,56837,56866,56923,56968,57105,57165,57391,57417,57444,57568,57597,57758,57845,57905,57908,57947,57989,58086,58497,58574,58577,58765,58775,58905,59005,59336,59386,59493,59673,59762,59804,59945,60208,60220,60464,60519,60525,60789,61138,61347,61392,61449,61520,61548,61655,61760,61984,62049,62078,62158,62286,62501,62523,62541,62594,62776,62933,63018,63068,63117,63149,63168,63190,63314,63318,63458,63475,63479,63523,63615,63694,63818,63823,64075,64160,64168,64311,64326,64387,64530,64533,64818,64970,65070,65078,65173,65264,65365,65416,65612,65675,65700,65713,65728,65788,65811,65991,66133,66172,66237,66283,66473,66476,66495,66630,66680,66690,66706,66866,66895,66923,66972,66986,67029,67073,67179,67208,67227,67248,67251,67327,67431,67567,67633,67796,67990,68017,68091,68154,68298,68352,68356,68581,68597,68687,68770,68815,68885,69176,69289,69333,69490,69569,69617,69705,69747,69779,69858,69884,70040,70064,70069,70173,70215,70222,70274,70289,70402,70417,70430,70526,70581,70595,70598,70654,70761,70798,70900,70951,70953,71035,71115,71301,71340,71465,71500,71515,71771,71811,71945,72192,72255,72522,72614,72643,72659,72711,72803,72927,72956,73034,73073,73148,73170,73181,73251,73278,73304,73449,73464,73610,73673,73861,73865,73875,74051,74061,74287,74552,74572,74581,74629,74674,74877,74908,74930,75206,75236,75250,75253,75257,75408,75420,75437,75526,75614,75676,75770,75783,75815,76004,76051,76126,76130,76189,76311,76549,76588,76651,76794,76868,76902,77353,77394,77481,77521,77527,77606,77662,77747,77986,77991,78033,78034,78049,78104,78112,78269,78291,78385,78426,78741,78758,78789,78817,78863,78872,78995,79156,79248,79261,79408,79501,79577,79627,79628,79638,79697,79706,79779,79831,79862,79995,80067,80102,80119,80188,80266,80339,80355,80426,80523,80728,80809,80836,80885,80897,80997,81325,81431,81487,81551,81617,81673,81678,81709,81773,81839,81939,81986,82187,82194,82209,82270,82295,82470,82492,82497,82556,82612,82736,82854,82926,82971,82975,82987,83029,83072,83085,83105,83124,83163,83177,83199,83228,83229,83262,83263,83286,83355,83448,83452,83641,83739,83757,83790,83802,83804,83812,83886,83906,83920,83924,83943,83946,84050,84160,84199,84555,84679,84734,84742,84780,84814,84855,84869,85011,85022,85032,85048,85196,85215,85226,85429,85560,85671,85679,85774,85832,85847,85926,86036,86089,86138,86243,86440,86453,86474,86544,86705,86708,86846,86965,87051,87121,87143,87146,87216,87230,87323,87485,87497,87571,87595,87662,87668,87693,87696,87747,87749,87960,88025,88048,88133,88184,88205,88252,88444,88513,88524,88571,88620,88697,88709,88812,88943,88952,89012,89060,89061,89310,89440,89492,89547,89585,89770,89800,89832,89883,89941,90049,90118,90312,90858,90907,91164,91174,91199,91292,91490,91590,91629,91788,91814,92056,92072,92145,92300,92311,92520,92627,92840,92850,92919,92922,92934,93102,93111,93219,93308,93391,93657,93762,93896,93910,93915,93983,93995,94153,94175,94317,94336,94386,94716,94809,94833,95015,95057,95080,95116,95169,95176,95209,95226,95405,95439,95574,95810,96122,96441,96448,96667,96690,96761,96792,96866,96892,97053,97175,97210,97294,97516,97613,97617,97623,97864,97956,98001,98015,98116,98327,98349,98391,98423,98463,98517,98619,98624,98722,98738,98764,98802,99009,99086,99446,99512,99537,99547,99558,99598,99609,99695,99837",
      "43,132,254,488,542,544,550,658,722,833,967,1205,1467,1518,1625,1682,1687,1835,1862,1884,1886,2144,2196,2241,2388,2402,2475,2502,3178,3225,3403,3600,3764,3813,3822,3879,4159,4270,4278,4389,4434,4514,4588,4820,4922,5186,5220,5277,5448,5449,5483,5508,5548,5591,5726,5734,5908,5920,5948,6036,6112,6201,6264,6290,6388,6398,6418,6545,6616,6678,6704,6764,6788,6908,6958,6970,7096,7260,7389,7414,7431,7448,7642,7662,7748,7755,7762,7777,7809,7869,7933,8093,8132,8328,8500,8540,8615,8619,8705,8715,8721,8731,8860,8921,9171,9278,9359,9360,9429,9486,9506,9527,9536,9600,9607,9756,10017,10289,10418,10430,10452,10628,10671,10751,10796,10806,11050,11063,11121,11186,11215,11307,11387,11497,11526,11567,11676,11727,11748,11782,11959,12029,12051,12088,12164,12315,12472,12526,12549,12556,12559,12665,12690,12751,12894,12922,12925,13007,13019,13027,13144,13311,13313,13319,13390,13473,13531,13622,13666,13894,14042,14106,14132,14293,14349,14529,14643,14711,14761,14836,14843,14925,14966,14988,15092,15238,15422,15508,15667,15670,15850,15895,15978,16011,16269,16395,16429,16476,16480,16550,16641,16693,17013,17319,17371,17390,17404,17418,17587,17653,17797,17854,17949,17960,18112,18150,18167,18183,18228,18310,18414,18454,18500,18506,18528,18551,18592,18799,18890,18985,19146,19175,19180,19289,19328,19575,19594,20040,20088,20144,20196,20204,20278,20553,20569,20607,20620,20659,20723,20771,20799,20938,20973,21160,21209,21246,21517,21624,21740,21758,21768,22007,22141,22197,22206,22363,22387,22444,22668,22836,22873,23003,23134,23307,23335,23470,23602,23720,23736,23763,23830,23842,24154,24229,24279,24371,24458,24553,24923,25313,25539,25601,25693,25748,25779,25812,25853,25872,25957,26114,26168,26276,26349,26416,26613,26652,26657,26688,26712,26720,26730,26757,26765,26792,26985,27054,27077,27101,27108,27186,27240,27297,27316,27519,27585,27658,27665,27790,28167,28200,28265,28330,28404,28440,28674,28679,28724,28779,28794,28858,28959,28974,28977,29140,29202,29235,29236,29278,29435,29493,29681,29686,29758,29760,29802,29961,29974,29986,29989,30139,30149,30498,30759,30817,31368,31425,32064,32122,32209,32324,32498,32933,33456,33561,33734,33793,33920,34036,34837,35151,35405,35562,35660,35679,35684,35703,35744,35880,35898,35981,36003,36009,36052,36170,36258,36321,36363,36372,36495,36573,36665,36677,36754,36836,36879,37000,37016,37134,37247,37295,37447,37497,37499,37555,37634,37759,37942,37952,38055,38075,38129,38140,38189,38367,38502,38578,38711,38795,38818,38820,38991,38992,39043,39064,39185,39259,39329,39435,39718,39747,39998,40174,40662,41046,41265,41421,41479,42453,42526,42599,42694,42785,42865,42943,42980,43134,43154,43263,43445,43607,43629,43715,43804,44339,44423,44714,44811,45318,45366,45484,45666,45672,45695,45809,45866,45870,46685,46826,47637,47662,48095,48427,48657,48899,49109,49214,49427,49465,49761,50033,50050,50064,50100,50111,50332,50400,50484,50494,50589,50693,50716,50794,50901,50958,50974,50998,51201,51211,51271,51310,51373,51471,51477,51713,51762,51790,51834,51851,51907,51911,51958,51963,52023,52084,52171,52363,52365,52533,52548,52638,52705,52947,53082,53203,53284,53353,53401,53443,53594,53860,53873,53986,54026,54058,54240,54275,54293,54301,54327,54355,54398,54408,54409,54424,54427,54475,54506,54549,54555,54575,54634,54695,54710,54759,54773,54808,54840,54951,54972,55005,55137,55326,55342,55450,55748,55781,55802,55926,56120,56179,56262,56363,56380,56537,56783,56796,56876,56905,56926,56955,57018,57059,57132,57251,57466,57474,57581,57585,57638,57682,57730,57759,57853,57956,58053,58080,58174,58347,58520,58537,58659,58761,58817,58948,59200,59293,59469,59485,59560,59675,59775,59802,60010,60130,60217,60291,60356,60446,60531,60674,60754,60815,60920,61003,61007,61033,61081,61092,61337,61373,61429,61459,61467,61498,61601,61743,61776,61980,62060,62071,62126,62380,62435,62486,62538,62664,62792,62914,63058,63213,63230,63296,63331,63402,63535,63553,63607,63643,63647,63678,63824,64025,64102,64164,64216,64270,64313,64345,64358,64370,64497,64502,64640,64644,64682,64709,64860,64894,65162,65188,65360,65369,65377,65399,65520,65645,65795,65822,65895,66004,66196,66278,66357,66457,66471,66524,66560,66624,66655,66845,66873,66897,67008,67081,67113,67335,67603,68129,68171,68371,68421,68613,68724,68767,68846,68886,68967,69017,69234,69642,69678,69949,69969,70001,70022,70038,70228,70263,70332,70365,70392,70483,70615,70641,70741,70756,70937,71025,71170,71315,71432,71470,71483,71488,71529,71601,71656,71689,71767,71870,71911,71963,71991,72171,72208,72370,72404,72431,72467,72511,72551,72670,72934,72976,73124,73245,73298,73666,73709,73791,73803,73920,73986,74029,74240,74257,74356,74555,74589,74624,74646,74893,75001,75031,75050,75740,75860,75862,76027,76101,76115,76153,76166,76239,76505,76779,76933,77044,77221,77259,77332,77495,77562,77640,77694,77707,77717,77740,77753,77778,77872,78018,78052,78077,78314,78321,78481,78568,78661,78683,78876,78892,78904,78922,79048,79056,79174,79188,79493,79696,79701,79801,79864,79962,80044,80048,80173,80187,80207,80293,80328,80434,80461,80489,80569,80610,80618,80753,80834,80913,80983,81426,81553,81563,81640,81663,81744,82033,82071,82145,82172,82192,82317,82545,82647,82793,82842,82846,82863,82880,83006,83117,83146,83270,83323,83372,83418,83473,83634,83759,83798,83848,84178,84202,84498,84565,84766,84815,84821,84823,84917,84938,85095,85119,85125,85172,85298,85328,85482,85527,85594,85620,85698,85733,85961,85991,86372,86396,86514,86571,86604,86732,86737,86750,86812,86828,87078,87256,87307,87338,87648,87650,87746,87810,87880,87912,87984,88004,88008,88034,88062,88068,88099,88240,88298,88405,88534,88541,88631,88768,88866,88876,88899,89128,89431,89443,89532,89677,89695,89740,89813,90036,90107,90177,90179,90200,90235,90277,90344,90466,90585,90672,90836,90837,90974,90979,90994,91006,91024,91043,91046,91195,91197,91386,91481,91549,91651,91817,91880,91906,92002,92157,92207,92274,92277,92280,92345,92507,92555,92560,92576,93029,93082,93144,93192,93194,93250,93275,93278,93375,93589,93642,93663,93668,93716,93865,93974,94026,94073,94125,94187,94197,94214,94433,94434,94576,94659,94828,94958,94961,95244,95288,95302,95319,95342,95452,95502,95546,95624,95662,95710,95923,95945,95960,96008,96098,96216,96223,96287,96317,96630,96705,96721,96756,96829,96991,97089,97148,97242,97301,97462,97506,97599,97634,97658,97705,97706,97716,97764,97941,98063,98352,98492,98541,98635,98637,98644,98855,98956,98977,99252,99356,99388,99398,99676,99741,99881,99884",
      "75,227,239,512,1368,1667,1694,1705,1740,2009,2333,2685,2860,2953,3080,3200,3248,3270,3303,3396,3529,3618,3726,3730,3947,4004,4042,4061,4102,4234,4325,4347,4433,4625,4657,4687,4846,4884,5381,6113,6603,6623,6854,6949,6960,8051,8127,8133,8146,8165,8556,8640,8659,8683,8697,9178,9305,9343,9362,9438,9533,9552,9628,9647,9666,9685,9737,9799,9881,10061,10065,10070,10083,10118,10141,10198,10298,10350,10374,10379,10388,10597,10616,10659,10679,10723,10768,10825,10827,10842,10844,10960,10970,11064,11110,11148,11205,11243,11262,11281,11319,11395,11401,11441,11490,11507,11509,11547,11635,11685,11737,11973,12003,12065,12117,12150,12181,12193,12212,12250,12269,12273,12345,12408,12440,12535,12552,12573,12593,12604,12649,12680,12706,12730,12744,12826,12858,12896,12901,12906,12991,13018,13063,13094,13111,13124,13143,13162,13200,13238,13295,13369,13382,13486,13580,13675,13770,13789,13808,13903,13922,13998,14013,14112,14169,14226,14264,14302,14397,14454,14473,14492,14511,14568,14606,14701,14720,14758,14815,14853,14891,14929,14967,15005,15036,15062,15125,15138,15195,15199,15201,15290,15309,15363,15423,15499,15537,15575,15651,15727,15765,15841,15917,15924,15936,16046,16050,16069,16088,16145,16202,16240,16259,16260,16278,16297,16316,16335,16354,16392,16468,16487,16525,16582,16603,16604,16696,16702,16715,16791,16796,16825,16829,16841,16864,16886,16981,17000,17057,17209,17228,17247,17304,17361,17380,17432,17475,17532,17585,17655,17760,17779,17798,17817,17874,18064,18108,18207,18234,18349,18368,18387,18406,18463,18476,18485,18501,18563,18634,18710,18712,18748,18767,18786,18824,18862,18900,18919,18926,19090,19128,19147,19223,19261,19280,19292,19299,19318,19394,19413,19565,19596,19622,19641,19660,19679,19812,19822,19831,19964,20059,20071,20097,20116,20135,20173,20185,20192,20230,20268,20306,20363,20382,20458,20477,20515,20591,20629,20686,20743,20895,20914,20933,20971,20979,20988,21028,21047,21104,21180,21256,21294,21351,21446,21465,21484,21522,21579,21617,21636,21674,21693,21769,21788,21864,21883,21902,21921,21940,21997,22016,22065,22111,22149,22175,22187,22240,22263,22282,22301,22358,22396,22434,22491,22510,22514,22605,22624,22643,22681,22700,22757,22778,22795,22871,22883,22909,22915,22928,22946,22978,23023,23052,23194,23232,23289,23339,23373,23384,23419,23460,23498,23559,23574,23631,23669,23707,23726,23878,23973,23985,23992,24011,24030,24049,24220,24239,24249,24258,24296,24334,24448,24486,24543,24657,24676,24695,24735,24741,24752,24771,24783,24809,24847,24885,24906,24961,24999,25037,25056,25170,25208,25303,25341,25398,25455,25469,25512,25569,25588,25645,25683,25778,25816,25873,25904,25911,25949,25968,26006,26025,26044,26082,26101,26158,26177,26196,26253,26291,26329,26367,26386,26424,26443,26462,26481,26500,26519,26538,26728,26819,27147,27165,27222,27255,27392,27625,27849,27851,28058,28262,28370,28552,28590,28647,28793,28799,28808,28831,28854,29059,29229,29268,29388,29436,29503,29619,29654,29833,29841,29863,29910,29940,30082,30111,30140,30227,30297,30314,30372,30430,30459,30517,30575,30662,30691,30749,30807,30981,31068,31155,31184,31213,31271,31300,31387,31416,31503,31561,31590,31619,31648,31677,31706,31793,31851,31909,31996,32025,32054,32112,32141,32170,32199,32228,32402,32431,32472,32489,32518,32547,32576,32663,32721,32779,32837,32924,32953,33011,33069,33098,33156,33214,33243,33272,33301,33359,33388,33475,33504,33591,33620,33649,33678,33765,33794,33823,33852,33939,33997,34009,34026,34113,34171,34229,34241,34258,34316,34345,34403,34461,34490,34519,34548,34577,34606,34635,34693,34780,34838,34925,34954,34983,35024,35038,35141,35156,35159,35161,35243,35319,35350,35387,35424,35446,35495,35543,35602,35661,35769,35793,35852,35869,35881,35923,35982,35995,36058,36060,36164,36223,36353,36426,36543,36602,36611,36666,36725,36784,36815,36855,36914,36973,37104,37227,37286,37416,37475,37560,37606,37665,37724,37732,37736,37738,37788,37906,37977,38036,38048,38049,38095,38167,38226,38325,38408,38467,38478,38538,38563,38597,38656,38728,38846,38860,38910,38937,38969,39028,39158,39229,39348,39364,39407,39471,39530,39589,39719,39778,39850,39871,39909,39968,40132,40161,40190,40248,40277,40306,40335,40364,40393,40625,40683,40712,40741,40770,40828,40857,40886,40915,40973,41031,41060,41118,41147,41176,41205,41263,41292,41321,41379,41408,41437,41495,41524,41961,41969,41973,41977,41990,41995,42024,42082,42171,42246,42252,42385,42402,42411,42481,42491,42537,42569,42613,42648,42662,42761,42767,42780,42810,42868,42885,42930,42959,42966,43035,43041,43047,43122,43184,43277,43359,43437,43521,43596,43639,43758,43803,43808,43995,44232,44307,44366,44381,44524,44544,44597,44664,44771,44810,44855,44900,45018,45117,45139,45200,45227,45236,45331,45384,45388,45518,45575,45679,45705,45738,45788,45820,45863,45892,45921,46050,46052,46079,46085,46114,46220,46238,46271,46277,46284,46332,46374,46448,46502,46514,46556,46601,46711,46746,46812,46833,46845,46894,46947,47024,47025,47048,47059,47130,47189,47284,47312,47324,47397,47408,47432,47457,47524,47538,47560,47728,47748,47794,47798,47926,47958,47967,47987,48001,48006,48027,48035,48194,48325,48363,48402,48412,48476,48511,48705,48770,48774,48811,48932,48954,48957,49102,49169,49253,49301,49312,49328,49498,49527,49545,49590,49600,49679,49682,49809,49821,49890,49907,49912,49949,49954,50041,50049,50078,50110,50112,50113,50115,50121,50205,50460,50523,50579,50625,50633,50645,50661,50731,50757,50784,51038,51044,51142,51191,51248,51294,51311,51447,51523,51526,51536,51539,51597,51614,51664,51767,51850,51862,51951,51987,52207,52335,52473,52477,52510,52770,52909,52996,53205,53350,53397,53406,53435,53486,53548,53572,53858,53917,54086,54116,54146,54257,54259,54309,54341,54393,54507,54543,54592,54653,54658,54680,54718,54790,54799,54843,54922,54971,55024,55221,55265,55268,55463,55480,55838,55893,55937,56060,56128,56188,56197,56281,56285,56308,56377,56564,56576,56727,56836,56842,56853,56868,56894,56994,57021,57024,57188,57267,57268,57356,57499,57528,57775,57843,57856,57983,58020,58049,58066,58142,58184,58212,58310,58446,58478,58621,58668,58681,58685,58967,59235,59379,59567,59736,59906,59981,59986,59997,60035,60165,60183,60215,60258,60280,60309,60383,60595,60635,60921,60960,61166,61410,61513,61833,61927,62189,62270,62475,62545,62657,62727,62910,62985,63035,63265,63322,63335,63415,63573,63630,63703,63713,63723,63897,64043,64054,64117,64141,64339,64509,64534,64551,64642,65087,65088,65261,65571,65625,65681,65865,66034,66074,66119,66121,66311,66363,66631,66649,66827,66880,67184,67261,67349,67514,67618,67673,67694,67868,67905,67953,68087,68132,68193,68372,68623,68653,68657,68665,68670,68737,68779,68852,68988,69020,69119,69141,69284,69318,69483,69585,69607,69644,69686,69783,70132,70210,70354,70373,70602,70678,70744,70924,71070,71075,71180,71288,71329,71357,71581,71663,71855,71959,72092,72130,72169,72207,72229,72317,72355,72487,72569,72771,72800,72845,72889,73016,73210,73215,73217,73257,73283,73458,73468,73503,73519,73531,73674,73683,73842,73923,73924,73925,74098,74238,74296,74321,74333,74351,74452,74466,74509,74550,74582,74652,74687,74702,74875,74905,74921,75006,75028,75051,75080,75150,75199,75311,75392,75432,75471,75491,75555,75567,75668,75729,75801,76034,76163,76170,76195,76253,76615,76743,76776,76847,76900,77037,77326,77328,77555,77596,77623,77626,77661,77760,77774,77813,77853,78002,78063,78108,78320,78394,78414,78473,78495,78601,78640,78645,78757,78816,78827,78919,78965,78973,78991,79000,79070,79157,79160,79164,79215,79320,79332,79380,79417,79479,79941,80060,80142,80211,80226,80236,80308,80441,80518,80519,80547,80645,80670,80683,81080,81170,81214,81231,81399,81477,81549,81683,81690,81694,81780,81854,81896,81929,82273,82419,82447,82531,82585,82589,82700,82799,83014,83152,83154,83321,83420,83510,83618,83659,83740,83741,83826,83834,83857,83881,83885,83977,83979,84006,84078,84100,84103,84110,84145,84198,84215,84272,84323,84325,84623,84650,84673,84825,84888,84935,84967,85068,85078,85204,85421,85439,85496,85618,85720,85735,85817,85889,85955,86008,86166,86266,86270,86358,86413,86523,86574,86608,86697,86698,86873,86902,87111,87117,87130,87167,87178,87261,87545,87673,87778,87809,88020,88024,88075,88228,88247,88340,88401,88655,88739,88858,88914,89036,89125,89324,89377,89478,89533,89723,89891,89995,90004,90053,90079,90206,90218,90321,90356,90390,90430,90549,90674,90739,90829,90856,91014,91089,91101,91260,91272,91295,91338,91341,91706,91828,91944,91950,92030,92252,92557,92603,92657,92807,92827,92932,93022,93045,93086,93205,93245,93322,93404,93418,93429,93522,93535,93603,93604,93824,93829,93849,93869,93892,94021,94122,94298,94485,94615,94666,94771,94845,94869,94886,94950,94974,94998,95085,95172,95267,95277,95395,95457,95567,95597,95606,95677,95715,95728,96069,96081,96104,96117,96141,96180,96262,96280,96289,96291,96307,96382,96418,96447,96510,96539,96550,96588,96744,96768,96859,97073,97093,97205,97434,97522,97712,97925,98000,98056,98094,98161,98485,98573,98589,98739,98754,98934,98993,99066,99076,99122,99243,99248,99272,99348,99359,99371,99375,99406,99445,99470,99821,99940,99955",
      "6,15,157,317,326,427,558,728,742,764,839,900,1036,1116,1119,1150,1317,1401,1423,1474,1552,1581,1607,1800,1808,1816,1840,1865,2065,2108,2133,2201,2349,2398,2409,2457,2591,2658,2703,2770,2954,3082,3100,3227,3238,3258,3551,3601,3606,3635,3684,3704,3913,3914,4083,4317,4753,4786,5009,5225,5228,5260,5281,5490,5504,5648,5679,5685,5799,5889,5894,5896,5898,5900,5923,5928,5946,6053,6157,6171,6242,6348,6359,6394,6445,6451,6459,6477,6516,6519,6532,6618,6692,6759,6821,6831,6892,6899,6923,6964,7050,7235,7258,7322,7358,7520,7553,7559,7650,7744,7923,7951,8010,8048,8129,8202,8219,8259,8313,8316,8423,8461,8481,8796,8892,9024,9061,9092,9158,9509,9713,9769,9854,9879,9934,10052,10085,10296,10505,10510,10520,10523,10534,10548,10773,10830,10863,10906,11044,11106,11487,11522,11548,11692,11704,11795,11992,12052,12446,12590,12595,12642,12670,12782,12791,12865,12871,12918,12956,13154,13166,13175,13188,13226,13348,13352,13391,13471,13510,13595,13627,13716,14065,14490,14493,14826,14844,14980,15043,15152,15371,15398,15610,15860,16615,16644,16755,17200,17395,17569,17658,17706,17794,17921,17941,17982,18327,18514,18545,18699,18707,18709,18809,18834,18873,18928,19397,19512,19574,20265,21074,21223,21250,21354,21378,21384,21397,21458,21492,21615,21781,22051,22103,22171,22207,22295,22377,22719,22736,23106,23303,23586,24068,24214,24413,24421,24751,24955,25086,25139,25228,25254,25377,25459,25637,25813,25882,26040,26043,26301,26344,26579,26750,26752,26753,26776,26790,26818,26878,26969,27062,27126,27153,27206,27235,27301,27309,27457,27641,27693,27830,27839,27872,27911,28409,28473,28748,28935,28975,29036,29107,29144,29169,29324,29429,29455,29516,29565,29861,30201,30275,31841,32057,32218,32439,32845,33349,33436,33533,33871,34145,34664,34712,35048,35374,35509,35518,35570,35748,35766,35825,36001,36094,36261,36262,36379,36506,36569,36774,36808,37056,37117,37196,37296,37301,37309,37406,37626,37632,37778,37881,37943,38015,38110,38133,38194,38201,38317,38605,39176,39234,39249,39322,39358,39559,39634,39804,39918,40172,40415,40444,40937,41377,42016,42646,43175,43663,43670,43831,43912,43960,44070,44317,44501,44774,44932,45422,45686,47093,47220,47266,47367,47496,47692,47701,47761,48060,48297,48442,49329,49663,49732,50026,50103,50124,50204,50261,50366,50379,50538,50549,50622,50678,50708,50771,50834,50844,50960,51087,51089,51147,51179,51198,51356,51402,51895,51981,52012,52047,52118,52183,52262,52621,52930,52931,53043,53169,53319,53351,53377,53453,53477,53816,53903,54101,54114,54127,54244,54324,54350,54429,54516,54633,54745,54754,54807,54812,54849,54866,54880,54898,54954,55019,55028,55210,55212,55254,55322,55395,55476,55572,55579,55600,55619,55620,55744,55858,55954,56306,56472,56524,56716,56722,56768,56778,56840,56856,56950,57060,57127,57448,57675,57676,57740,57773,57985,57992,58187,58319,58330,58398,58423,58428,58433,58494,58505,58874,59057,59066,59074,59106,59192,59247,59251,59326,59357,59483,59513,59546,59580,59699,59728,59731,59750,59949,59962,60004,60069,60072,60095,60104,60111,60188,60395,60456,60494,60512,60588,60641,60714,60918,60951,61193,61402,61415,61437,61489,61570,61794,61975,62003,62087,62111,62370,62489,62587,62704,62796,62801,62838,62908,62926,62998,63087,63210,63317,63343,63435,63471,63515,63664,63763,63799,63838,63889,63918,63921,63931,63968,63997,64097,64101,64177,64238,64421,64470,64471,64545,64565,64633,64695,64702,64712,64764,64827,64957,65017,65074,65083,65097,65131,65138,65156,65214,65403,65486,65521,65562,65800,65805,65903,65948,65963,66201,66442,66518,66671,66713,66719,66830,66958,66995,67040,67045,67202,67216,67313,67320,67370,67400,67503,67544,67548,67757,67792,67880,67885,67951,68039,68042,68165,68261,68318,68399,68430,68432,68532,68654,68814,68821,69012,69168,69193,69421,69473,69590,69630,69633,69650,69685,69694,69724,69756,69875,70007,70071,70107,70150,70290,70372,70470,70551,70560,70608,70651,70867,71071,71163,71190,71234,71268,71393,71398,71461,71659,71743,71759,71805,71891,71942,71987,72002,72011,72022,72150,72181,72287,72430,72489,72525,72536,72693,72697,72773,72838,72860,72882,72913,72959,73049,73341,73459,73596,73740,73895,73933,73939,74046,74084,74190,74256,74282,74311,74360,74480,74539,74568,74574,74606,74757,74797,74835,74892,74990,75023,75084,75102,75143,75172,75512,75541,75609,75610,75656,75689,75884,75962,75984,76071,76111,76187,76231,76399,76471,76484,76524,76665,76673,76720,76732,76806,76818,76843,76939,76997,77073,77148,77171,77247,77345,77354,77378,77390,77409,77509,77674,77763,77929,78024,78075,78158,78286,78332,78397,78401,78519,78614,78672,78792,78920,78939,78958,78964,79052,79099,79103,79104,79137,79191,79195,79259,79273,79395,79403,79405,79492,79564,79657,79670,79767,79809,79889,79935,79960,79988,80107,80122,80141,80147,80320,80341,80360,80462,80508,80532,80602,80622,80686,80866,80942,80946,80963,81024,81030,81033,81294,81390,81566,81972,82002,82126,82199,82262,82296,82313,82504,82562,82583,82643,82777,82877,82886,82909,82981,83057,83145,83166,83176,83280,83441,83484,83573,83578,83708,83742,83917,84134,84191,84257,84328,84399,84431,84491,84656,84657,84667,84669,84844,84981,84998,85033,85147,85183,85399,85520,85619,85760,85885,85929,86026,86298,86309,86487,86504,86526,86539,86562,86567,86584,86632,86635,86644,86716,86735,86756,86849,86958,87131,87187,87191,87198,87242,87283,87352,87522,87523,87677,87705,87881,87888,87982,88179,88180,88190,88201,88257,88263,88331,88453,88461,88546,88658,88753,89102,89149,89195,89237,89448,89461,89481,89497,89500,89613,89713,89809,89814,90027,90182,90291,90325,90472,90510,90520,90608,90615,90862,90877,90904,91133,91444,91468,91493,91613,91694,92086,92101,92116,92140,92149,92163,92271,92344,92377,92378,92417,92464,92468,92481,92536,92654,92688,92758,92870,92902,92913,92986,93078,93124,93167,93246,93268,93324,93325,93353,93478,93529,93534,93542,93552,93561,93662,93938,94096,94124,94207,94220,94223,94271,94390,94412,94450,94547,94655,94675,94685,94715,94752,94939,94947,94948,95019,95105,95112,95192,95247,95292,95314,95375,95384,95415,95444,95516,95540,95643,95651,95777,96047,96109,96146,96164,96366,96389,96392,96462,96704,96797,96832,96833,96845,97020,97161,97165,97313,97316,97432,97476,97559,97643,97708,97763,97775,97886,97890,97930,98086,98173,98304,98337,98366,98561,98582,98655,98740,98808,98942,98947,99041,99162,99206,99227,99238,99303,99369,99405,99549,99587,99630,99693,99863,99943,99971",
      "218,327,343,350,408,487,549,649,653,795,799,918,1037,1051,1117,1133,1216,1218,1309,1342,1527,1843,1853,2077,2088,2147,2217,2449,2522,2952,3032,3078,3136,3288,3633,3785,3818,3920,3994,3997,4168,4233,4633,4656,4790,5146,5216,5309,5321,5368,5420,5469,5501,5530,5567,5653,5704,5749,5754,5950,5988,6213,6227,6237,6272,6332,6374,6437,6513,6662,6693,6766,6784,6801,6939,6994,7014,7259,7268,7305,7723,7768,7863,7868,7880,7904,8021,8174,8301,8369,8455,8631,8671,8689,8713,8748,8812,8866,8944,8970,9013,9045,9055,9182,9259,9262,9267,9402,9448,9676,9779,9807,9843,9852,9912,9959,10029,10033,10059,10129,10252,10488,10501,10676,10857,10869,10963,10984,10999,11033,11088,11268,11321,11326,11353,11368,11502,11711,11728,11756,11807,11857,11878,11888,12046,12200,12242,12268,12288,12429,12480,12564,13008,13058,13215,13243,13254,13429,13521,13541,13797,13833,13915,14248,14390,14503,14587,14723,14726,14794,14834,15061,15068,15494,15551,15792,16007,16282,16386,16611,16642,16898,16982,17103,17191,17219,17234,17262,17474,17598,18081,18107,18191,18263,18650,18729,18993,19058,19136,19195,19438,19461,19604,19643,19728,20523,20691,20692,20695,20729,20735,20795,20876,21071,21079,21326,21332,21502,21509,21549,21587,21606,21695,21696,21800,22088,22167,22189,22285,22439,22824,22949,23013,23048,23099,23136,23222,23468,23597,23617,23716,24149,24182,24206,24264,24477,24781,24989,25140,25151,25370,25418,25429,25551,25655,25717,25910,26080,26104,26296,26357,26366,26505,26676,26689,26716,26726,26737,26895,26906,26960,27018,27026,27033,27129,27159,27208,27224,27319,27343,27458,27588,27631,27677,27710,27812,27900,28045,28071,28111,28170,28195,28211,28246,28313,28379,28455,28523,28534,28689,28701,28764,28785,29027,29048,29070,29092,29262,29390,29404,29422,29583,29709,29722,29766,29800,29850,29975,30043,30409,30612,30614,31039,31293,31310,32088,32584,32905,33059,33149,34657,34692,34741,35023,35123,35555,35762,35832,35952,36013,36051,36243,36274,36533,36577,36812,36817,36827,36987,37062,37153,37180,37369,37391,37478,37574,37577,37621,37816,37821,37856,38051,38135,38138,38200,38234,38260,38285,38426,38616,38630,38737,38761,38797,38982,39139,39175,39303,39316,39373,39474,39487,39510,39624,39641,39794,39817,40073,40274,40478,40988,41985,42248,42710,42937,43130,43578,43644,43738,44259,44738,44957,45166,45246,45264,45573,45985,46477,46501,46681,46870,46913,47001,47743,47880,48380,48459,48585,48995,49129,49446,49487,49765,50045,50139,50292,50351,50450,50496,50566,50670,50694,50774,50807,51081,51116,51140,51182,51279,51434,51506,51537,51698,51744,51749,51773,52007,52141,52202,52218,52241,52283,52326,52385,52394,52958,53054,53061,53131,53220,53234,53266,53298,53539,53610,53717,53735,53781,53898,54292,54494,54514,54528,54673,54720,54782,54829,54966,55015,55023,55339,55340,55634,55728,55810,55844,56406,56482,56565,56606,56803,56830,56851,56896,56953,56965,56997,57038,57039,57244,57321,57340,57350,57460,57509,57530,57763,57812,57934,58111,58154,58166,58203,58283,58366,58440,58447,58570,58613,58777,58796,58927,58969,59017,59033,59111,59152,59155,59197,59384,59434,59437,59461,59648,59691,59738,60127,60161,60213,60317,60320,60433,60484,60491,60564,60730,60772,60820,60892,60929,60966,61119,61175,61253,61264,61307,61324,61629,61733,61801,61826,61860,62064,62205,62217,62298,62355,62386,62391,62442,62524,62632,62753,62799,62819,62823,62901,63006,63186,63249,63375,63399,63413,63557,63599,63613,63627,63655,63685,63750,63792,63985,64000,64048,64272,64284,64338,64493,64542,64567,64590,64639,64705,65174,65313,65483,65529,65691,65734,65777,65896,65981,66190,66232,66406,66423,66484,66726,66957,66970,67047,67100,67210,67416,67420,67504,67780,67817,67860,67924,67966,68037,68186,68304,68307,68353,68380,68639,68650,68834,68868,68957,68960,69082,69215,69217,69221,69251,69378,69422,69455,69578,69852,69887,69953,69979,70285,70350,70413,70418,70579,70638,70661,70670,70703,70738,70820,71122,71134,71165,71166,71202,71279,71391,71395,71396,71440,71481,71555,71587,71590,71878,72020,72126,72299,72377,72426,72433,72535,72574,72587,72785,72817,72835,72938,72969,72972,73007,73136,73315,73354,73355,73363,73393,73556,73679,73760,74107,74205,74206,74487,74494,74538,74756,74906,74985,75005,75013,75124,75141,75245,75265,75296,75351,75404,75486,75487,75516,75577,75579,75660,75716,75775,75914,75958,76036,76061,76140,76211,76249,76345,76508,76640,76679,76681,76733,76966,77004,77054,77121,77190,77226,77279,77315,77368,77430,77443,77487,77678,78006,78138,78145,78147,78275,78297,78430,78458,78531,78539,78806,78821,78840,79378,79399,79521,79716,79737,79740,79822,79992,80079,80198,80278,80324,80347,80655,80737,80760,80765,80779,80788,80928,80941,80956,81100,81155,81252,81255,81256,81300,81331,81445,81457,81465,81582,81588,81612,81855,82055,82092,82179,82196,82463,82532,82536,82676,82772,82785,82813,82837,82964,83002,83097,83274,83316,83405,83633,83844,83860,83990,84024,84076,84123,84184,84222,84227,84248,84260,84270,84410,84446,84778,84831,85126,85151,85163,85494,85502,85566,85730,85734,85778,85937,85983,86041,86054,86060,86107,86256,86393,86420,86641,86818,86832,86857,86884,87021,87108,87414,87473,87501,87513,87536,87608,87621,87676,87729,87903,87943,88077,88224,88306,88313,88385,88403,88596,88623,88728,88734,89157,89160,89207,89234,89528,89535,89574,89586,89606,89622,89689,89711,89789,89803,89825,89852,89886,90084,90313,90329,90489,90493,90519,90564,90636,90697,90935,90941,90969,91017,91147,91389,91446,91602,91719,91913,92012,92060,92088,92122,92128,92185,92192,92233,92242,92406,92454,92509,92559,92637,92640,92903,92939,92978,93004,93031,93070,93211,93290,93363,93503,93515,93855,93870,93904,93945,93963,93984,94024,94113,94230,94415,94471,94515,94628,94799,94853,94966,95044,95047,95236,95250,95399,95983,96157,96377,96514,96515,96578,96584,96776,96843,96852,97012,97086,97108,97160,97215,97271,97305,97306,97514,97523,97533,97588,97739,97784,98005,98011,98193,98471,98560,98623,98669,98814,99094,99192,99211,99215,99330,99577,99605,99774,99899,99966",
      "48,73,85,131,504,527,702,727,1145,1170,1409,1497,1574,1657,1720,1771,1775,1837,2023,2248,2279,2293,2307,2316,2547,2566,2686,2689,2708,2774,2962,2967,3010,3063,3165,3181,3345,3461,3545,3729,3823,3838,3881,3952,4118,4153,4166,4211,4291,4295,4655,4731,4841,4848,4859,4933,4934,4949,5001,5217,5371,5382,5477,5513,5519,5527,5710,5819,5835,5972,6022,6152,6224,6252,6286,6308,6336,6402,6515,6698,6782,6829,6848,6959,7020,7033,7109,7158,7189,7284,7301,7313,7378,7383,7449,7475,7494,7558,7596,7598,7643,7658,7770,7793,7798,7865,7874,7876,7889,7916,7924,8057,8073,8101,8166,8264,8348,8349,8356,8432,8440,8478,8586,8667,8694,8724,8741,8752,8824,8949,9030,9064,9071,9104,9218,9294,9309,9318,9554,9704,9846,9925,9938,10101,10117,10132,10237,10409,10533,10583,10639,10762,10788,10872,10880,10977,10987,11116,11129,11233,11331,11333,11348,11414,11556,11619,11633,11669,11805,11821,12005,12111,12156,12194,12462,12507,12623,12666,12702,12726,12850,12907,12919,13134,13217,13219,13225,13248,13383,13399,13418,13470,13526,13602,13607,13702,13997,14027,14049,14075,14099,14122,14129,14149,14409,14453,14597,14666,14928,14937,14986,15025,15048,15106,15173,15647,15933,16144,16179,16267,16286,16396,16674,16812,16816,16887,16941,16971,17028,17046,17086,17107,17195,17323,17405,17426,17440,17500,17553,17578,17590,17612,17654,17893,18012,18022,18045,18051,18140,18168,18282,18435,18492,18715,18822,18871,19189,19191,19244,19375,19378,19381,19508,19593,19625,19632,19708,19717,19945,19972,20033,20277,20297,20442,20527,20582,20704,20705,20759,20809,20903,20943,21080,21158,21310,21643,21644,21667,21677,21682,21741,21777,21779,21824,21835,22096,22190,22346,22443,22576,22985,22988,23009,23027,23097,23164,23188,23430,23690,23697,23795,23905,23907,24123,24168,24293,24402,24610,24656,24700,24723,24806,24911,25007,25014,25045,25173,25255,25416,25446,25597,25647,25762,25899,25928,26015,26149,26270,26318,26348,26446,26509,26602,26781,26944,26964,26987,27005,27021,27048,27239,27339,28318,28360,28367,28606,28769,28836,28839,28913,29054,29152,29240,29457,29553,29567,29579,29641,29742,29825,29881,29946,30401,30433,30568,30873,30875,30923,31067,31332,31390,31426,31670,31735,31738,31743,31825,32006,32083,32131,32179,32207,32289,32627,32692,32956,32961,32985,33039,33062,33079,33233,33494,33562,33797,34000,34770,34809,35034,35166,35361,35368,35381,35617,35692,36118,36172,36260,36373,36380,36443,36477,36624,36659,36682,36756,36822,36945,37094,37144,37307,37321,37338,37521,37572,37573,37658,37828,38202,38265,38310,38388,38398,38431,38590,38611,38631,38708,38787,38867,39031,39063,39113,39484,39494,39540,39552,39712,39840,40169,40224,40491,40908,41203,41326,41372,41728,41981,42030,42185,42201,42601,42623,42681,42702,42723,42728,42741,42786,42816,42818,42955,43043,43168,43365,43456,43475,43628,43704,43755,43833,43890,44037,44404,44446,44469,44812,44863,45228,45241,45319,45423,45508,45555,45688,45734,45762,45805,45849,46009,46164,46229,46307,46331,46437,46560,46566,46741,46793,46905,47706,47756,47961,48195,48245,48288,48636,48803,48839,49197,49323,49330,49688,49999,50072,50114,50145,50232,50438,50532,50628,50637,50749,50795,50833,50861,50867,50957,51022,51023,51062,51080,51264,51297,51390,51525,51533,51558,51650,51703,51765,51915,52022,52153,52157,52206,52494,52563,52582,52675,52722,52791,52920,52939,52962,53089,53092,53177,53272,53278,53462,53494,53500,53711,53959,54051,54168,54299,54315,54356,54501,54556,54628,54691,54699,54727,54744,54766,54908,54925,55029,55058,55085,55285,55502,55645,55682,55693,55768,55788,55879,55909,55931,56006,56007,56036,56064,56067,56211,56239,56275,56424,56478,56548,56577,56583,56702,56865,56908,56954,57012,57013,57041,57146,57255,57290,57349,57735,57854,57987,58197,58337,58345,58532,58558,58651,58671,58756,59027,59157,59244,59269,59588,59734,59796,59815,59826,59866,59920,59924,59956,60053,60058,60097,60123,60207,60431,60678,60978,61050,61146,61156,61178,61412,61502,61530,61539,61552,61603,61667,62019,62243,62267,62287,62358,62592,62938,63014,63019,63152,63185,63198,63212,63272,63575,63684,63743,63762,64027,64269,64430,64494,64612,64717,64737,64856,64862,64951,64994,65000,65370,65485,65549,65605,65680,65759,65776,65820,65909,65920,65947,65965,65970,66013,66108,66115,66267,66270,66317,66448,66569,66637,66673,66794,66837,67001,67115,67125,67136,67193,67224,67487,67588,67785,67893,67961,68077,68113,68178,68189,68284,68348,68420,68542,68586,68753,68817,68824,68968,69181,69213,69343,69426,69807,69934,69998,70004,70048,70136,70336,70441,70537,70583,70671,70711,70946,70965,71037,71065,71081,71231,71355,71495,71537,71711,71714,71839,71876,71879,71971,72070,72093,72257,72403,72488,72493,72499,72542,72640,72901,72960,72965,72998,73002,73035,73157,73159,73182,73201,73218,73345,73406,73418,73419,73455,73510,73528,73544,73657,73711,73723,73934,74069,74100,74120,74124,74227,74389,74415,74429,74543,74618,74707,74795,74804,74826,74934,75239,75254,75330,75345,75370,75430,75517,75528,75554,75576,75582,75635,75707,75910,76026,76064,76110,76180,76263,76301,76326,76328,76368,76426,76456,76492,76521,76568,76798,76870,76878,77024,77437,77518,77526,77631,77664,77711,77776,77915,78038,78132,78179,78274,78461,78566,78634,78721,78722,78953,79003,79058,79066,79150,79375,79442,79663,79678,80005,80088,80493,80496,80506,80512,80570,80625,80804,81149,81247,81268,81317,81389,81406,81715,81783,81844,81850,81880,82037,82142,82178,82201,82293,82324,82515,82570,82600,82679,82968,83022,83033,83096,83103,83108,83309,83346,83369,83462,83465,83766,83810,83912,84141,84164,84182,84235,84316,84317,84338,84505,84536,84560,84615,84840,84847,84951,85013,85094,85155,85189,85194,85302,85342,85372,85484,85513,85530,85747,85907,86004,86157,86175,86179,86263,86282,86323,86351,86401,86509,86546,86779,86972,86975,87068,87140,87159,87232,87250,87304,87413,87460,87735,87955,88049,88092,88194,88295,88722,88775,88895,88898,88936,89048,89079,89136,89278,89314,89319,89354,89358,89465,89909,89991,90075,90076,90120,90167,90223,90250,90543,90569,90625,90725,90736,90752,90807,90871,90878,90887,91057,91146,91345,91375,91526,91685,91757,91784,91919,92054,92136,92146,92268,92323,92382,92553,92596,92630,92778,92868,92880,92983,93005,93121,93321,93417,93519,93532,93616,93686,93699,93846,93871,93882,93924,93937,93954,93964,94224,94276,94339,94355,94404,94437,94501,94572,94681,94730,94847,94855,94872,94877,94934,94955,94977,95115,95177,95239,95240,95254,95320,95371,95387,95482,95504,95527,95577,95613,95788,95900,96051,96074,96152,96153,96202,96214,96395,96483,96544,96546,96594,96700,96811,96814,96898,96940,97009,97016,97035,97091,97207,97214,97321,97331,97391,97448,97525,97720,97725,97740,97973,98018,98040,98355,98408,98417,98543,98590,98641,98834,98850,98883,98903,98924,99060,99123,99126,99166,99327,99435,99597,99652,99769,99779,99861",
      "24,28,53,199,232,295,311,528,606,1013,1626,1692,1755,1789,1906,2012,2083,2099,2113,2167,2261,2294,2344,2394,2486,2614,2763,2768,2821,3087,3148,3199,3314,3346,3356,3375,3491,3539,3741,3786,3923,3938,3943,3999,4080,4106,4114,4218,4308,4373,4409,4512,4557,4688,4699,4726,4751,4757,4771,4833,4915,5108,5230,5253,5272,5354,5439,5454,5635,5678,5769,5791,5931,6015,6033,6047,6062,6067,6160,6183,6243,6278,6414,6427,6436,6455,6626,6672,6742,6771,6773,6832,6868,6900,6946,7044,7095,7114,7294,7477,7479,7523,7614,7628,7821,7826,7902,8148,8252,8297,8343,8361,8444,8475,8545,8569,8755,8770,8790,8844,8941,8946,9063,9103,9217,9227,9261,9299,9341,9417,9483,9550,9590,9877,9992,10139,10140,10152,10231,10254,10317,10326,10347,10375,10376,10426,10434,10622,10663,10801,10866,10895,10921,10938,11138,11171,11223,11275,11290,11595,11600,11642,11729,11794,11827,11889,12040,12185,12245,12263,12297,12352,12364,12434,12609,12756,12796,12856,12903,12958,12971,13198,13203,13328,13408,13446,13504,13561,13611,13711,13719,13776,13841,13854,13864,14002,14055,14115,14151,14164,14353,14416,14469,14481,14500,14512,14523,14588,14676,14766,15001,15046,15113,15135,15142,15240,15258,15268,15403,15417,15543,15594,15599,15801,15822,15966,15993,16041,16156,16576,16639,16675,16730,16786,16926,17076,17089,17171,17185,17205,17260,17443,17606,17656,17855,17984,17987,18029,18229,18425,18526,18652,18656,18739,18867,19009,19043,19242,19350,19559,19609,19848,19931,19997,20095,20106,20320,20331,20512,20519,20622,20762,20874,20936,21284,21300,21436,21558,21661,22146,22221,22231,22248,22409,22467,22548,22598,22645,22689,22756,22847,23041,23126,23170,23193,23245,23281,23457,23897,24090,24138,24163,24283,24340,24391,24500,24557,24562,24575,24600,24710,24749,24907,24965,24967,25207,25651,25653,25678,25795,25829,25857,26063,26066,26153,26184,26294,26499,26536,26565,26674,26695,26706,26711,26734,26783,26823,26864,26980,27010,27105,27139,27168,27229,27365,27417,27430,27438,27477,27499,27521,27537,27557,27709,27720,27842,27869,27945,28148,28171,28214,28350,28352,28361,28481,28528,28662,28717,28780,28801,28820,28993,28998,29057,29066,29181,29223,29251,29253,29302,29303,29400,29423,29424,29466,29684,29716,29728,29755,29788,29889,29930,30000,30053,30104,30270,30285,30291,30343,30452,30467,30604,30623,30894,31081,31110,31319,31358,31598,31690,31873,31941,31956,31967,32018,32178,32305,32474,32497,32550,32618,33004,33555,33707,33784,34190,34219,34329,34353,34493,34773,34857,34896,35036,35126,35251,35269,35335,35400,35448,35463,35511,35591,35609,35629,35701,35998,36020,36063,36131,36132,36182,36266,36273,36387,36469,36615,36674,36687,36759,36777,36831,36866,36886,36889,36928,37132,37176,37194,37201,37275,37316,37405,37432,37435,37439,37454,37561,37579,37983,38181,38246,38355,38378,38544,38554,38645,38679,38807,38809,38962,38980,39001,39128,39217,39257,39628,39743,39767,39821,39839,39893,39919,40103,40143,40219,40375,40420,40422,40502,40594,40652,40684,40862,40890,41089,41116,41138,41322,41807,41849,41899,41901,41925,41956,41965,42014,42167,42202,42254,42362,42396,42417,42467,42485,42679,42685,42784,42936,43044,43068,43079,43218,43317,43392,43452,43650,43674,43854,43883,43907,43941,43958,44362,44550,44727,44820,44896,45124,45300,45310,45352,45572,45580,45596,45613,45683,45847,45927,45934,45955,46057,46094,46147,46221,46247,46302,46506,46571,46590,46701,46945,47504,47528,47699,47779,47886,48057,48138,48222,48262,48521,48660,48703,48738,48851,48856,48989,49034,49177,49492,49652,49654,49889,49931,50028,50169,50260,50280,50348,50691,50787,50798,50850,50868,51032,51043,51061,51100,51155,51255,51388,51394,51440,51515,51658,51857,51858,51978,52138,52155,52169,52205,52223,52466,52617,52622,52968,52985,53018,53029,53047,53219,53225,53288,53390,53391,53418,53617,53694,53740,53972,54081,54106,54131,54151,54407,54448,54468,54657,54696,54703,54707,54918,55104,55204,55280,55584,55648,55671,55708,56070,56086,56246,56372,56388,56398,56404,56621,56622,56630,56658,56674,56798,56951,57016,57068,57213,57220,57298,57428,57507,57945,58001,58012,58024,58039,58138,58189,58214,58221,58409,58470,58525,58697,58747,58785,58804,58957,58993,59009,59060,59203,59447,59582,59586,59825,59840,59890,59939,59954,59976,60003,60029,60141,60214,60242,60281,60310,60314,60321,60453,60753,60852,60891,60913,60917,61183,61349,61614,61658,61846,61909,62044,62069,62086,62110,62202,62293,62328,62350,62462,62532,62608,62759,62763,62999,63002,63039,63081,63176,63201,63211,63216,63218,63221,63281,63400,63487,63550,63622,63634,63734,63845,63881,63932,64018,64087,64144,64178,64246,64304,64328,64400,64438,64585,64809,64932,64937,64945,64961,65112,65129,65199,65216,65226,65238,65292,65366,65406,65412,65413,65479,65490,65498,65510,65525,65585,65671,65699,65753,65978,65988,66012,66084,66238,66274,66399,66521,66585,66602,66772,66925,67049,67059,67069,67238,67391,67517,67686,67711,67765,67798,68082,68086,68092,68156,68325,68377,68461,68468,68471,68491,68563,68682,68707,68855,69171,69173,69226,69306,69842,69880,70085,70194,70327,70385,70543,70578,71155,71397,71421,71480,71504,71527,71579,71654,71826,71932,72031,72212,72233,72315,72325,72339,72389,72394,72466,72607,72733,72736,72812,72873,73004,73044,73090,73337,73379,73415,73527,73624,73663,73704,73796,73814,73825,73897,74127,74135,74187,74211,74284,74403,74421,74524,74557,74697,74763,74814,74902,74951,74956,74988,75061,75092,75157,75160,75533,75556,75563,75621,75629,75655,75748,75766,75773,75859,75867,75869,75945,75947,76070,76127,76273,76556,76585,76633,76641,76655,76815,76840,76874,76899,77046,77077,77098,77101,77472,77475,77723,77743,77787,77789,77930,77973,78212,78522,78621,78647,78749,78805,78819,78976,79010,79061,79116,79117,79454,79635,79844,79920,79927,79972,80003,80010,80117,80315,80319,80330,80383,80505,80528,80556,80613,80658,80675,80698,80750,80790,80850,81102,81146,81196,81265,81287,81417,81443,81466,81619,81634,81693,81706,81945,81970,82091,82247,82298,82367,82369,82403,82406,82453,82571,82616,82738,82962,83082,83101,83120,83132,83284,83291,83296,83490,83541,83564,83598,83699,83745,83861,83955,83961,83983,84020,84060,84088,84165,84208,84391,84413,84435,84463,84492,84589,85238,85435,85445,85464,85501,85523,85611,85685,85763,85819,85823,85828,85864,85868,85917,85969,86145,86346,86377,86498,86639,86679,86688,86718,86747,86755,86762,86842,86844,86893,86909,86948,87107,87160,87356,87367,87415,87437,87442,87551,87628,87633,87799,87927,87944,88058,88060,88125,88226,88256,88261,88388,88448,88523,88544,88563,88647,88667,88679,88752,88878,89004,89081,89083,89334,89368,89504,89525,89767,89772,89822,89853,90043,90105,90115,90161,90225,90259,90335,90370,90377,90433,90605,90721,90742,90776,90812,90867,90947,91056,91107,91216,91426,91475,91557,91568,91612,91653,91758,91804,91847,91864,91924,92188,92390,92425,92426,92430,92542,92549,92577,92594,92647,92701,92843,92951,92960,93001,93038,93234,93249,93303,93368,93369,93415,93510,93538,93636,93732,93782,93786,94022,94032,94060,94101,94268,94463,94481,94650,94701,94747,94825,94851,94884,94899,94900,94993,94997,95077,95119,95120,95163,95228,95355,95382,95495,95541,95560,95580,95663,95782,95974,95977,96039,96080,96195,96344,96370,96421,96495,96496,96533,96632,96742,96743,96790,96810,96834,96863,96900,96950,96955,96969,97004,97006,97011,97083,97348,97389,97439,97446,97512,97518,97615,97690,97750,97947,98012,98248,98318,98333,98339,98359,98525,98538,98673,98692,98724,98736,98743,98817,98838,98873,98933,99016,99036,99115,99152,99191,99230,99275,99278,99321,99453,99740,99784,99890,99932",
      "79,222,287,309,358,365,373,678,688,802,940,948,1122,1323,1326,1380,1452,1528,1666,1697,1725,1834,2169,2290,2506,2555,2665,2674,2695,2730,2807,2817,2851,2876,2998,3168,3513,3566,3638,3834,4190,4296,4343,4369,4387,4513,4645,4684,4765,4899,4904,4921,4925,5031,5116,5130,5131,5391,5529,5545,5601,5626,5656,5821,5932,5955,6076,6116,6296,6297,6363,6429,6456,6496,6524,6547,6686,6753,6824,6834,6872,6881,6938,6999,7089,7155,7219,7323,7415,7420,7438,7474,7498,7789,7849,7896,7953,7981,7996,8149,8305,8473,8520,8571,8595,8700,8701,8745,8822,8852,8939,9035,9050,9074,9149,9150,9153,9288,9310,9388,9431,9569,9872,9884,9929,10392,10486,10507,10558,10674,10884,10887,10932,11054,11122,11127,11340,11419,11559,11589,11631,11672,11779,11840,11861,11886,11916,11947,12041,12195,12356,12438,12453,12632,12734,12885,13322,13338,13507,13560,13562,13620,13818,13875,13905,13947,13968,13996,14391,14462,14505,14515,14549,14564,14690,14717,14791,14813,14969,14996,15040,15072,15160,15376,15460,15565,15624,15669,15686,15957,16031,16108,16142,16264,16357,16387,16406,16546,16632,16838,17072,17202,17311,17408,17483,17662,17997,18046,18055,18076,18146,18504,18535,18620,18756,18760,18913,18916,19035,19056,19079,19183,19291,19503,19522,20039,20615,20855,20862,20875,20932,21046,21065,21166,21375,21408,21623,21664,21731,21775,21988,22018,22114,22131,22208,22290,22312,22372,22374,22496,22629,22651,22675,22678,22838,22863,22869,22935,23004,23121,23137,23204,23447,23493,23821,24581,24635,24960,24962,24964,25027,25225,25257,25335,25408,25618,25643,25840,26128,26483,26651,26671,26810,26920,26922,26974,27127,27196,27277,27347,27379,27427,27551,27556,27575,27882,27907,27934,27946,28235,28339,28365,28386,28422,28496,28712,28912,28983,28989,29231,29259,29501,29524,29596,29734,29873,29977,30081,30090,30150,30255,30324,30865,30904,31232,31251,31339,32608,32691,32836,33175,34193,34287,34760,35014,35017,35256,35622,35693,35715,35955,36301,36474,36637,36647,36986,37107,37124,37183,37250,37289,37490,37623,37633,37690,38139,38254,38257,38503,38548,38602,38617,38693,39066,39067,39184,39376,39439,39742,39932,40398,40705,40959,41010,42022,42483,42831,42854,43118,43402,43420,43460,43547,43667,43985,44041,44202,44357,45238,45247,45288,45415,45446,45600,45607,45630,45693,45698,45749,45858,45865,46318,46369,47527,48024,48373,48873,49510,49614,50023,50263,50305,50368,50408,50543,50739,50765,50830,50904,50991,51041,51126,51269,51415,51596,51648,51712,51815,51984,52103,52180,52224,52253,52272,52295,52313,52646,52697,52712,52805,52903,52911,52950,53011,53014,53024,53110,53274,53343,53606,53623,53640,53791,53839,53930,53980,54311,54317,54363,54454,54508,54584,54640,54642,54722,54832,54876,54946,54949,55177,55356,55526,55558,55847,56075,56092,56096,56292,56329,56349,56410,56459,56574,56575,56648,56772,56790,56874,56928,56931,57286,57421,57467,57492,57493,57630,57696,57723,57844,57878,57974,58000,58003,58078,58249,58256,58314,58331,58438,58580,58626,58772,58979,59008,59015,59047,59148,59234,59285,59298,59325,59391,59399,59594,59756,59896,59990,60270,60406,60443,60536,60567,60660,60798,60968,61005,61017,61129,61192,61215,61343,61390,61448,61458,61535,61566,61593,61606,61640,61766,61900,61979,62008,62119,62268,62296,62321,62381,62399,62443,62444,62522,62542,62575,62602,62785,62795,63061,63092,63167,63635,63771,63833,63952,64146,64222,64327,64389,64401,64626,64965,64995,64999,65086,65262,65449,65578,65661,65724,65880,65916,65961,66166,66452,66595,66622,66762,66929,67103,67137,67249,67312,67417,67456,67526,67551,67587,67628,67664,67824,68142,68229,68358,68422,68667,68672,68760,68966,69022,69045,69227,69300,69391,69444,69555,69583,69599,69620,69794,69864,69882,69906,70151,70185,70294,70490,70527,70567,70573,70601,70656,70690,70694,70753,70835,70876,71056,71141,71161,71313,71366,71419,71430,71452,71453,71592,71630,71644,71665,71672,71775,71790,71874,71992,72012,72045,72276,72371,72417,72517,72570,72740,72746,72757,72947,72954,73064,73135,73155,73166,73176,73194,73249,73286,73293,73325,73481,73569,73570,73662,73669,73707,73751,73757,73799,73916,73932,74012,74251,74312,74315,74423,74504,74511,74515,74516,74532,74751,74787,75255,75565,75637,75695,75916,75917,75939,75970,76097,76184,76419,76482,76500,76584,76703,76724,76764,76775,76781,76924,76937,76942,77050,77083,77142,77419,77465,77523,77675,77732,77738,77744,77762,77891,77901,77921,77941,77975,77994,78135,78159,78204,78251,78306,78325,78408,78434,78616,78618,78628,78738,78828,78885,78926,78937,79111,79272,79308,79326,79362,79376,79396,79435,79438,79472,79523,79550,79588,79673,79700,79855,79929,80073,80074,80166,80189,80190,80233,80292,80348,80531,80578,80668,80803,80934,81160,81208,81391,81392,81449,81475,81528,81578,81598,81627,81739,81761,81829,82057,82105,82217,82389,82411,82421,82440,82514,82594,82599,82625,82774,82831,82902,82931,82983,83056,83269,83319,83493,83600,83877,83972,84059,84221,84258,84329,84331,84404,84543,84626,84710,84755,84799,84816,84817,84906,84930,84952,84957,85037,85076,85141,85285,85459,85468,85499,85571,85593,85629,85798,85904,85921,85950,86038,86177,86278,86283,86288,86315,86356,86466,86645,86811,86838,86887,86916,86936,87071,87204,87231,87369,87547,87593,87741,87792,88014,88165,88168,88193,88250,88325,88374,88566,88576,88578,88591,88720,88723,88809,89006,89050,89095,89176,89230,89356,89517,89578,89625,89760,89821,89862,89982,90061,90129,90272,90274,90304,90333,90360,90482,90686,90704,90726,90764,90767,90799,90841,90921,90967,91132,91262,91286,91387,91463,91496,91577,91674,91680,91747,91785,91805,91902,91956,92014,92028,92041,92096,92234,92236,92262,92265,92573,92714,92728,92732,92809,92832,92834,92872,92873,93018,93425,93944,94136,94173,94202,94242,94292,94405,94454,94476,94479,94514,94587,94699,94700,94929,94982,95078,95128,95340,95431,95458,95531,95586,95611,95614,95673,95809,95843,96062,96097,96148,96207,96268,96327,96350,96477,96486,96499,96507,96551,96552,96586,96652,96668,96683,96720,96883,97048,97228,97251,97329,97687,97701,97704,97879,98106,98174,98289,98303,98484,98563,98571,98660,98677,98731,98734,98820,98941,99075,99175,99218,99287,99391,99698,99713,99903,99930",
      "5,9,23,26,250,313,506,518,624,628,667,950,1266,1278,1288,1289,1292,1302,1354,1370,1458,1463,1510,1522,1586,1809,1811,1813,1836,1844,1848,1899,2081,2242,2247,2250,2291,2296,2420,2478,2517,2692,3183,3389,3481,3528,3696,3779,3812,3832,3836,3869,4377,4691,4717,4724,4814,4847,4911,4947,5115,5124,5141,5282,5329,5384,5390,5398,5415,5423,5456,5479,5547,5658,5718,5808,5952,5989,6032,6056,6110,6189,6194,6215,6223,6232,6255,6279,6288,6353,6376,6420,6421,6443,6450,6465,6473,6538,6558,6560,6574,6602,6661,6707,6727,6737,6860,7145,7206,7365,7370,7426,7432,7492,7576,7595,7597,7612,7749,7763,7782,7816,7855,7890,8033,8062,8074,8180,8249,8251,8255,8332,8483,8530,8561,8629,8684,8956,8985,9085,9326,9337,9369,9408,9470,9771,9868,9902,9982,9999,10323,10330,10344,10415,10497,10553,10567,10662,10694,10696,10841,10942,11067,11297,11312,11351,11584,11675,11785,11968,11979,12439,12676,12749,12821,12824,12857,12941,13004,13831,13984,14120,14133,14225,14399,14430,14537,14683,14887,14942,14960,15059,15225,15308,15546,15590,15783,15794,15971,16060,16766,16856,17242,17468,17643,17668,17688,17698,18110,18139,18180,18569,18747,18994,19518,19796,19865,19967,20037,20167,20803,21018,21093,21108,21479,21531,21548,21648,22319,22601,22718,22829,22865,23191,23288,23311,23630,23854,24077,24725,24748,24846,25032,25169,25359,25493,25666,26244,26360,26629,26632,26685,26723,26733,26768,26779,26839,26862,26876,26889,27015,27081,27137,27175,27322,27371,27393,27620,27702,27819,28095,28269,28394,28437,28486,28715,28851,28874,29245,29340,29342,29472,29533,29535,29588,29662,29759,29775,29864,30659,31009,31647,31790,31879,32343,32401,33068,33909,35338,35466,35634,35807,36137,36309,36318,36619,36679,37012,37128,37187,37193,37308,37310,37384,37411,37450,37763,37868,37941,38059,38312,38315,38364,38372,38466,38559,38748,39065,39254,39347,39542,39733,39738,39777,39988,40551,40783,42143,42546,42691,42911,42916,42984,43151,43253,43315,43336,43495,43498,43627,43643,43734,43749,43812,44421,44872,45007,45030,45123,45270,45316,45510,45511,45563,45608,45685,45790,46600,47354,47377,49089,49225,50069,50198,50256,50277,50394,50568,50586,50626,50745,51024,51026,51184,51398,51403,51503,51527,51553,51848,51904,52035,52139,52219,52266,52368,52407,52455,52502,52509,52632,52696,52794,52876,52940,53003,53081,53121,53182,53324,53398,53509,53527,53571,53574,53586,53611,53658,53875,54102,54170,54253,54303,54381,54414,54451,54535,54602,54793,54890,54904,54907,55097,55222,55260,55262,55324,55419,55437,55516,55596,55807,56003,56009,56022,56054,56339,56501,56729,56826,56902,57045,57056,57249,57304,57379,57388,57411,57462,57684,57751,57794,57839,58220,58222,58255,58387,58481,59035,59089,59395,59518,59726,59792,59838,59989,60100,60232,60239,60243,60303,60435,60487,60548,60572,60574,60675,60962,61024,61182,61246,61276,61317,61478,61538,61672,61783,61839,61884,61890,62050,62103,62122,62167,62390,62543,62549,62621,62668,62723,62760,62870,62886,63067,63111,63170,63237,63257,63306,63358,63365,63416,63528,63540,63638,63820,63839,63852,63902,63938,64066,64080,64197,64261,64276,64334,64417,64517,64526,64554,64575,64658,64714,64788,64906,64942,64989,65081,65190,65244,65325,65594,65635,65764,65893,65924,65956,66162,66373,66414,66425,66446,66554,66850,67014,67172,67315,67355,67453,67516,67521,67609,67781,67857,67867,68099,68130,68136,68329,68330,68355,68369,68575,68591,69058,69077,69151,69203,69297,69349,69380,69468,69550,69611,69715,69745,69800,69897,69909,69925,69975,70033,70097,70159,70169,70204,70300,70364,70468,70480,70568,70577,70633,70804,70810,70870,71079,71106,71120,71235,71303,71405,71414,71420,71439,71451,71556,71706,71834,71977,71982,72003,72217,72272,72546,72630,72714,72719,72805,73014,73138,73199,73300,73344,73367,73403,73561,73565,73706,73801,74072,74093,74140,74160,74178,74230,74358,74476,74695,74836,74995,75026,75071,75079,75238,75337,75359,75377,75451,75527,75624,75709,75723,75787,75858,75918,76194,76238,76285,76291,76343,76384,76459,76464,76631,76836,76890,76901,76913,76921,76948,76951,77051,77118,77197,77217,77236,77341,77344,77503,77575,77591,77641,77653,77683,77734,77745,77749,77916,78065,78106,78210,78453,78530,78576,78637,78641,78643,78646,78719,78729,78736,78941,78956,79107,79178,79216,79333,79350,79693,79798,79799,79838,79964,79974,79999,80011,80012,80072,80158,80200,80213,80303,80313,80399,80460,80667,80700,80767,80773,80892,80938,81117,81136,81216,81263,81276,81372,81453,81592,81647,81732,82068,82283,82345,82465,82577,82741,83182,83212,83305,83353,83557,83574,83770,83779,83842,83865,84061,84106,84285,84315,84348,84453,84676,84685,84722,84730,84744,84885,84954,85003,85057,85120,85127,85199,85255,85341,85458,85462,85599,85640,85941,86078,86135,86199,86236,86261,86291,86394,86548,86557,86614,86650,86664,86680,86738,86881,86962,87056,87165,87207,87446,87506,87546,87669,87721,87853,87862,87867,87896,87966,88011,88066,88070,88079,88116,88181,88339,88458,88492,88558,88601,88677,88758,88902,88966,88973,89053,89062,89090,89297,89393,89411,89456,89709,89719,89741,89778,89878,89930,89954,89967,90031,90124,90158,90243,90256,90286,90447,90526,90541,90616,90658,90833,90925,90933,91079,91157,91160,91210,91257,91517,91640,91672,91689,91777,91801,91995,92111,92112,92114,92115,92152,92351,92360,92375,92612,92729,92763,92810,92831,92893,92924,93103,93134,93163,93168,93182,93248,93424,93715,93764,93766,93769,93779,93901,93934,93936,94023,94144,94194,94272,94280,94293,94323,94400,94449,94455,94464,94862,94870,95185,95199,95260,95341,95348,95521,95529,95570,95630,95691,95934,95972,96103,96137,96241,96269,96270,96303,96411,96583,96664,96749,96854,97253,97349,97379,97387,97393,97569,97577,97715,97989,98019,98067,98083,98262,98392,98479,98486,98524,98537,98539,98631,98638,98706,98719,98842,98904,99042,99106,99251,99415,99521,99599,99711,99715,99785,99792,99796,99850,99933,99984",
      "123,183,193,202,537,598,623,707,872,891,1077,1135,1447,1601,1690,1739,1806,2067,2131,2194,2274,2372,2434,2450,2459,2575,2781,2820,2875,2936,3217,3309,3703,3905,4164,4205,4256,4284,4385,4402,4431,4448,4515,4573,4637,4734,5004,5057,5068,5243,5271,5294,5503,5521,5611,5696,5759,5772,5804,5861,5941,6019,6061,6073,6176,6187,6365,6399,6431,6434,6444,6463,6552,6576,6580,6739,6840,6869,6918,7046,7198,7283,7407,7427,7434,7533,7664,7678,7692,7811,7891,7920,7930,7999,8032,8040,8067,8080,8089,8122,8223,8278,8324,8413,8450,8486,8608,8630,8735,8792,8801,8859,8890,8907,8924,8980,9052,9199,9231,9248,9301,9368,9375,9398,9422,9426,9444,9504,9529,9585,9731,9797,9923,10242,10355,10370,10377,10413,10433,10595,10660,10672,10724,10775,10818,10847,10862,11016,11062,11105,11181,11194,11235,11287,11300,11314,11389,11440,11565,11682,11700,11745,11873,11954,11975,12032,12074,12084,12145,12177,12216,12404,12520,12563,12723,12742,12809,12840,12846,12876,12932,13001,13009,13010,13055,13072,13075,13147,13160,13218,13287,13396,13426,13439,13498,13511,13836,13896,13952,13970,14009,14084,14224,14555,14662,14681,14704,14714,14773,14880,14978,15294,15426,15611,15774,15779,15803,15973,16055,16537,16711,16724,16876,16977,16983,17104,17144,17165,17445,17597,17723,17749,17829,17950,18144,18466,18541,18546,18696,18784,18790,18795,18797,18805,19016,19080,19177,19185,19250,19279,19337,19670,19689,19784,19853,19993,20004,20176,20200,20253,20544,20563,20575,20734,20781,20831,20868,20974,21057,21225,21398,21418,21421,21512,21603,21609,21616,21699,21776,21853,21962,22053,22776,22823,22852,22870,23021,23031,23209,23249,23257,23269,23422,23527,23677,23678,23692,23850,24104,24136,24417,24663,24720,24886,24900,24966,25036,25066,25076,25176,25187,25193,25330,25389,25412,25612,25724,25767,25820,25845,26052,26111,26156,26289,26353,26696,26700,26903,26965,27052,27055,27286,27532,27762,27782,27860,27903,27927,28234,28277,28297,28334,28372,28474,28509,28676,28790,28792,28914,29067,29213,29273,29282,29415,29484,29647,29710,29798,29820,29911,29995,30592,31032,31329,31764,32265,32753,33246,33968,34324,35088,35301,35380,35501,35550,35565,35630,35632,35702,35801,35879,35888,35988,36002,36047,36185,36188,36246,36449,36570,36634,36640,36811,36816,36819,36828,36934,36939,37004,37057,37129,37351,37379,37387,37551,37568,37575,37668,37691,37761,37867,38064,38186,38255,38362,38629,38632,38690,38936,38940,38985,39056,39070,39121,39168,39245,39246,39255,39427,39433,39443,39536,39549,39565,39606,39740,39751,39809,39814,39815,39879,42150,42210,42303,42459,42475,42490,42495,42549,42559,42617,42789,42879,42944,42960,43150,43270,43279,43284,43478,43560,43563,43676,43908,44021,44135,44221,44286,44294,44340,44350,44386,44512,44521,44536,44609,44685,44746,44756,44816,44849,44850,44866,45081,45091,45505,45528,45547,45561,45661,45663,45845,45888,45890,46534,46625,46908,47533,47925,48014,48233,49466,49546,50009,50051,50070,50155,50182,50238,50294,50498,50819,50915,51104,51238,51257,51263,51298,51517,51530,51591,51609,51655,51725,51823,51893,52028,52043,52306,52515,52524,52545,52684,52859,52864,52915,53038,53062,53117,53125,53157,53230,53408,53415,53448,53508,53656,53776,53793,53795,53799,53862,53942,54053,54273,54302,54374,54422,54457,54503,54597,54663,54721,55025,55145,55332,55436,55468,55482,55813,55837,55867,55901,55903,55999,56131,56195,56218,56226,56429,56604,56670,56862,56995,57029,57094,57149,57211,57221,57288,57610,57612,57754,57826,58339,58342,58407,58608,58618,58699,58780,58960,58971,59107,59143,59400,59463,59475,59690,59790,59821,59833,60023,60049,60132,60364,60498,60503,60580,60617,60653,60749,60762,60834,60869,60999,61002,61012,61042,61259,61411,61511,61560,61638,61693,61886,62128,62211,62251,62403,62611,62617,62624,62694,62695,63113,63215,63323,63338,63476,63533,63561,63699,63841,63847,64175,64307,64322,64568,64632,64758,65028,65057,65102,65111,65246,65538,65572,65829,66212,66325,66392,66507,66559,66613,66644,66707,66780,66831,66997,67083,67102,67259,67373,67799,67897,67907,67941,68090,68135,68214,68215,68227,68297,68394,68436,68568,68859,68925,68944,68953,69110,69260,69272,69367,69397,69541,69668,69721,69755,69839,69894,69963,70125,70233,70429,70432,70433,70434,70479,70523,70769,70781,70812,70831,70851,70864,70884,70948,70961,71062,71083,71101,71425,71455,71568,71648,71715,71736,71889,71912,71978,71995,72064,72144,72158,72303,72309,72346,72540,72713,72801,72903,72933,73022,73051,73068,73163,73280,73290,73390,73622,73630,73691,73785,73789,73882,73902,73952,74007,74062,74154,74307,74512,74530,74693,74701,74714,74722,74734,74775,74829,74955,75083,75105,75310,75346,75368,75618,75657,75662,75735,75950,75964,75966,76053,76131,76405,76457,76554,76574,76608,76609,76659,76685,76790,76801,76813,76928,76959,76983,77000,77028,77065,77080,77085,77116,77377,77441,77450,77671,77842,78015,78107,78114,78170,78299,78577,78596,78633,78664,78782,79015,79053,79091,79251,79255,79373,79617,79654,79656,79681,79718,79849,79903,79934,79955,80009,80116,80273,80382,80435,80455,80475,80533,80907,80912,81082,81130,81428,81436,81446,81512,81597,81667,81730,81779,81807,81956,81988,82123,82157,82228,82252,82286,82382,82629,82633,82669,82685,82711,82724,82737,82833,82872,82919,82946,83066,83179,83221,83227,83241,83385,83709,83852,83947,84025,84111,84159,84362,84438,84578,84791,84891,85263,85476,85539,85801,85820,85893,86020,86050,86067,86155,86225,86322,86350,86435,86443,86467,86556,86695,86754,86790,86823,86917,87059,87065,87101,87185,87194,87228,87251,87297,87337,87377,87499,87588,87768,88199,88296,88302,88352,88442,88543,88590,88703,88821,88863,88963,88979,89143,89256,89464,89579,89694,89817,90068,90172,90712,90795,90800,90890,90894,91148,91167,91192,91240,91242,91245,91251,91270,91289,91359,91380,91427,91484,91578,91579,91628,91821,91938,91957,92190,92203,92329,92413,92458,92471,92954,92992,93051,93069,93165,93200,93298,93300,93581,93644,93653,93837,93841,93852,93887,93946,94030,94176,94178,94180,94348,94358,94402,94475,94536,94577,94634,94938,95056,95081,95110,95216,95235,95286,95289,95454,95480,95559,95575,95605,95644,95654,95820,95872,95883,95970,95989,96011,96090,96131,96337,96359,96404,96534,96645,96746,96775,96777,96786,96848,96911,96943,97113,97174,97362,97503,97553,97561,97614,97659,97855,97893,97926,97927,97953,98024,98129,98166,98258,98265,98287,98324,98342,98507,98909,98996,99029,99195,99205,99207,99539,99613,99696,99752,99827",
      "209,318,366,397,415,417,441,519,564,578,581,607,815,860,869,893,975,990,996,999,1038,1185,1249,1269,1346,1361,1422,1446,1459,1515,1578,1879,1890,3567,4676,5040,5090,5105,5202,5206,5247,5360,5447,5729,5826,5919,6082,6369,6406,6432,6486,6505,6509,6550,6565,6627,6652,6673,6674,6689,6724,6735,6826,7016,7053,7066,7073,7075,7193,7273,7291,7310,7345,7356,7461,7541,7593,7689,7709,7835,7898,7921,8081,8199,8281,8302,8401,8409,8468,8477,8547,8675,8747,8967,8990,9001,9058,9070,9088,9154,9185,9188,9380,9454,9785,9976,10002,10247,10279,10450,10461,10494,10572,10576,10632,10735,10790,10792,10917,10939,10979,11055,11142,11199,11286,11659,11687,11817,11820,11890,11974,11987,12063,12128,12197,12374,12521,12561,12689,12718,12870,12933,13127,13401,13525,13874,14406,14952,15051,15359,15679,15753,16139,16385,16718,16739,16858,16985,17010,17207,17326,17571,17684,18018,18034,18483,18685,18701,18942,19547,20011,20538,20727,21070,21260,21285,21543,21598,21634,21981,22038,22119,22192,23340,23416,23564,24766,24818,24895,25172,25796,26413,26796,26822,26835,26851,26961,26977,27074,27216,27690,28412,28475,28697,28835,29113,29748,29845,36114,36193,36195,36196,36385,36439,36571,36734,37558,37637,37639,37672,37698,37799,38005,38297,38509,38610,38677,38921,39068,39111,39440,39442,39882,45379,45488,45615,45855,45878,49943,50092,50165,50175,50219,50775,50878,50897,51120,51220,51518,51557,51564,51607,51620,51653,51753,51787,51831,51856,52021,52106,52156,52243,52249,52260,52324,52331,52411,52518,52551,52574,52710,52755,52844,53009,53013,53071,53094,53098,53130,53241,53457,53544,53593,53731,53753,54013,54316,54332,54565,54595,54833,54991,55020,55081,55321,55327,55341,55434,55651,55653,55658,55697,55783,55818,55850,55900,56011,56044,56059,56296,56358,56444,56498,56519,56543,56584,56591,56611,56640,56662,56765,56936,57093,57178,57183,57252,57330,57395,57626,57729,57811,57859,57909,58034,58181,58296,58341,58485,58639,58753,58800,58860,58975,59038,59043,59055,59068,59105,59277,59286,59411,59555,59605,59607,59721,59853,59855,60079,60135,60152,60164,60211,60227,60462,60527,60537,60599,60605,60619,60663,60664,60694,60787,61256,61284,61387,61454,61515,61563,61590,61617,61628,61636,61722,61745,61811,61813,61853,62014,62109,62297,62309,62342,62361,62364,62467,62479,62584,62707,62743,62747,62748,62887,63007,63073,63158,63204,63232,63369,63425,63493,63560,63571,63593,63644,63695,63731,63764,63770,63775,63849,63851,64040,64083,64093,64118,64413,64704,64768,64905,64979,64980,65001,65032,65103,65151,65202,65410,65493,65523,65637,65702,65767,65798,65804,65812,66000,66010,66145,66146,66235,66548,66676,66983,67098,67147,67279,67448,67502,67642,67820,67871,67915,68012,68033,68049,68133,68296,68354,68376,68393,68398,68433,68463,68651,68697,68718,68763,68913,68958,68999,69044,69132,69276,69374,69385,69463,69498,69530,69531,69614,69672,69765,69855,70010,70023,70105,70192,70193,70546,70604,70666,70693,70784,70913,70945,70971,70987,71066,71172,71200,71259,71380,71394,71424,71525,72400,72575,72677,72793,72875,72881,73237,73466,73513,73566,73607,73633,73650,73675,73856,73917,73997,74074,74225,74233,74249,74367,74394,74448,74663,74747,74871,74887,74980,75027,75125,75228,75334,75389,75447,75452,75459,75580,75590,75667,75776,75802,75849,75850,75868,75898,76006,76068,76141,76244,76251,76324,76390,76421,76518,76535,76555,76859,76910,76987,77064,77183,77212,77468,77496,77627,77652,77680,77829,77858,78262,78278,78359,78471,78486,78506,78513,78540,78723,79034,79055,79085,79124,79514,79569,79610,79622,79757,80036,80126,80230,80272,80309,80334,80463,80501,80771,80835,80922,81001,81053,81188,81278,81312,81351,81435,81586,81759,81778,81881,81892,81894,82050,82059,82101,82234,82363,82488,82491,82624,82631,83046,83190,83242,83384,83430,83487,83520,83534,83675,83823,83854,83941,84069,84086,84122,84288,84371,84456,84488,84490,84544,84616,84643,84666,84827,84863,84882,85187,85326,85409,85516,85575,85577,85637,85712,85894,86006,86059,86211,86269,86285,86292,86301,86460,86547,86594,86736,86829,86896,86950,87018,87182,87253,87320,87464,87559,87569,87720,87754,87774,87798,87918,87987,87999,88122,88136,88231,88243,88319,88391,88551,88711,89235,89300,89322,89400,89580,89596,89598,89610,89615,89642,89648,89776,89807,89920,89959,90008,90063,90136,90196,90342,90420,90458,90481,90484,90637,90828,90840,90937,91012,91221,91296,91351,91356,91384,91531,91546,91576,91769,91812,91836,91879,91882,92036,92095,92198,92238,92279,92281,92334,92340,92397,92489,92554,92710,92784,92823,92869,92894,93028,93129,93304,93365,93381,93435,93540,93578,93605,93698,93775,93811,93860,93953,94034,94063,94266,94300,94313,94376,94394,94425,94495,94545,94568,94694,94816,94835,94953,95203,95299,95441,95505,95642,95650,95670,95885,95892,95906,95952,96024,96077,96236,96256,96311,96319,96378,96383,96410,96581,96600,96698,96799,96824,96835,96930,97058,97071,97140,97178,97212,97225,97298,97457,97674,97727,97738,97756,97814,97824,97887,97906,97914,97936,97939,98087,98210,98299,98374,98395,98449,98583,98592,98597,98666,98783,98789,98949,98955,98962,99000,99023,99313,99354,99458,99590,99644,99645,99656,99692,99718,99781,99795,99811",
      "4,269,314,394,401,410,484,491,524,574,696,712,720,748,761,765,887,899,925,933,985,1018,1071,1104,1212,1328,1466,1487,1526,1624,1724,1794,1807,1857,1909,1915,2925,2930,2966,3186,3413,3507,3587,3608,3917,4237,4247,5135,5214,5239,5240,5258,5307,5322,5410,5450,5562,5586,5645,5724,5770,5829,5870,5886,5993,6037,6099,6108,6141,6151,6228,6320,6339,6820,6893,6998,7117,7127,7134,7164,7220,7326,7338,7451,7501,7554,7573,7577,7611,7615,7638,7998,8039,8046,8131,8182,8189,8191,8198,8211,8227,8325,8391,8425,8439,8542,8577,8588,8663,8798,8987,9072,9179,9214,9236,9336,9345,9636,9904,10001,10171,10195,10224,10282,10333,10354,10389,10395,10402,10421,10427,10477,10479,10511,10528,10571,10585,10709,10879,10897,11107,11207,11288,11591,11710,11713,11715,11757,11819,11929,12732,12838,12953,13626,15145,15478,15790,15865,16315,17731,17857,18843,19192,19989,20066,21275,22740,22831,22855,22937,23089,24259,24286,24881,25130,25443,25743,26194,26405,27057,27184,27374,27847,28322,28349,28553,28721,30273,32019,32226,34139,35020,35561,36073,36171,36382,36700,37184,37613,39423,40385,42476,44693,45631,48228,48269,49417,49481,49580,50020,50109,50160,50295,50369,50557,50595,50752,50761,50799,50800,50824,50920,50984,51072,51085,51367,51497,51595,51715,51732,51751,51786,51846,51965,51979,51985,52011,52116,52154,52193,52300,52334,52356,52718,52778,52796,52843,52926,52929,52967,53031,53114,53128,53144,53149,53578,53654,53949,54069,54089,54185,54481,54505,54525,54629,54678,54682,55018,55070,55075,55103,55146,55370,55442,55478,55568,55604,55691,55801,56166,56265,56469,56542,56579,56620,56755,56845,57064,57309,57353,57453,57579,57607,57627,57632,57721,57732,57889,57991,58018,58128,58486,58538,58552,58885,59072,59102,59213,59340,59341,59351,59600,59630,59765,59904,59931,59936,59941,59951,60012,60131,60177,60308,60511,60610,60724,60874,60912,61181,61222,61299,61326,61356,61409,61424,61501,61581,61597,61607,61660,61712,61823,61914,61971,62134,62184,62344,62359,62387,62452,62526,62625,62633,62681,62714,62734,62892,62973,63033,63147,63264,63345,63423,63551,63637,63641,63658,63832,63843,63864,63887,63892,63975,64132,64209,64249,64323,64331,64461,64684,64778,64871,64912,64939,65022,65140,65458,65476,65527,65546,65583,65591,65913,65945,65987,66009,66080,66220,66241,66265,66279,66297,66384,66433,66516,66572,66611,66695,66846,67239,67332,67376,67378,67398,67447,67536,67566,67627,67637,67833,67834,67906,67940,67942,67996,68275,68292,68492,68537,68605,68773,68825,69035,69060,69085,69130,69266,69274,69447,69493,69511,69540,69570,69703,69719,69776,69850,69877,69883,69966,69981,70158,70163,70389,70440,70471,70592,70698,70737,70755,70827,70893,70952,70997,71135,71224,71242,71256,71263,71323,71437,71463,71509,71551,71572,71583,71642,71705,71777,71881,72178,72277,72386,72529,72577,72589,72703,72716,72777,72784,72920,73045,73102,73303,73491,73552,73584,73620,73645,73698,73732,73921,74015,74116,74138,74229,74239,74816,74817,74858,75053,75060,75182,75194,75200,75216,75278,75320,75426,75458,75497,75572,75654,75817,75929,76072,76082,76106,76182,76205,76214,76321,76322,76332,76643,76768,77066,77388,77467,77492,77543,77621,77622,77660,77773,77821,77833,77902,78068,78090,78131,78353,78423,78489,78564,78626,78667,78686,78921,79027,79276,79325,79527,79763,79766,79770,79800,79817,79890,79991,80259,80467,80643,80812,80878,80881,80906,80916,80987,81143,81407,81488,81554,81666,81689,81741,81815,81957,82009,82053,82070,82111,82377,82464,82597,82602,82635,82646,82653,82690,83075,83343,83620,83632,83637,83706,83777,83913,84040,84126,84207,84230,84279,84361,84387,84501,84534,84556,84568,84663,84861,84904,84933,84947,85079,85130,85186,85297,85431,85463,85473,85540,85553,85595,85825,85851,86042,86091,86112,86257,86276,86387,86399,86428,86439,86518,86529,86561,86565,86720,86813,86852,87009,87309,87439,87520,87543,87587,87815,87860,87928,88177,88343,88495,88609,88690,88692,88700,88800,88893,89007,89023,89034,89035,89082,89279,89320,89404,89490,89540,89543,89779,89787,89844,89869,89874,89911,90106,90293,90387,90575,90626,90694,91144,91308,91325,91497,91560,91670,91741,91949,91954,92104,92142,92226,92317,92662,92767,92838,92849,92878,93047,93156,93291,93374,93431,93592,93614,93691,93721,93798,93986,94010,94127,94285,94342,94364,94451,94617,94836,94858,95006,95092,95104,95126,95129,95148,95284,95310,95394,95416,95423,95434,95466,95484,95635,95647,95718,95755,95834,95908,95924,96059,96133,96322,96342,96469,96518,96607,96713,96729,96887,96905,96981,96987,97149,97241,97263,97307,97334,97360,97373,97383,97473,97478,97501,97601,97753,97794,97967,98066,98266,98373,98501,98574,98621,98787,98992,99072,99119,99177,99179,99293,99325,99336,99442,99489,99492,99493,99499,99510,99637,99640,99728,99797",
      "92,283,508,517,531,569,629,746,753,755,840,958,1132,1144,1223,1235,1351,1418,1439,1508,1517,1577,1847,1864,1943,2298,2494,2497,2563,2765,2803,2853,2937,2977,3083,3129,3161,3195,3229,3300,3390,3579,3687,4157,4264,4281,4430,4658,4700,4755,4775,4824,4842,5032,5089,5101,5109,5167,5188,5249,5297,5303,5318,5319,5362,5363,5413,5421,5431,5433,5436,5452,5460,5470,5482,5511,5625,5693,5714,5715,5738,5758,5773,5781,5790,5875,5926,5935,5953,5956,5984,5995,6106,6121,6200,6203,6216,6226,6250,6316,6366,6378,6408,6489,6504,6651,6710,6783,6842,6867,6880,6888,6931,6942,7004,7017,7052,7133,7184,7269,7312,7385,7398,7440,7455,7727,7734,7754,7852,7860,7918,8099,8106,8109,8233,8263,8283,8430,8523,8568,8575,8601,8625,8739,9209,9263,9355,9446,9589,9679,9908,9960,10213,10361,10458,10459,10480,10546,10608,10631,10923,10959,11014,11407,11582,11749,11802,11834,12206,12309,12344,12416,12491,12508,12585,12586,12967,13066,13148,13349,13447,14467,14472,14735,15301,15388,15570,15601,15612,15802,16084,16410,17034,17226,17512,17548,17560,17911,18025,18519,18604,18874,19182,19184,19428,19710,19751,20109,20533,20700,20798,21130,22081,22115,23054,23604,23794,23934,25625,25719,25846,25851,26190,26254,26310,26686,27024,27167,27282,27288,27452,27765,27800,27831,28213,28275,28443,28682,28716,28735,28896,29482,29805,29951,30574,30828,31123,31444,32053,32807,33354,33619,33648,35379,35613,35792,35820,35889,36308,36678,37060,37570,37752,38810,38863,39183,39256,39370,39477,39734,39805,39811,39990,40638,41450,41478,41507,42464,42528,42705,42873,43371,43493,43589,43681,44262,44497,44621,45307,45625,45923,46283,46724,46872,47411,49566,50105,50269,50398,50422,50448,50609,50610,50634,50699,50803,50809,50877,50900,50910,51004,51060,51113,51459,51513,51550,51659,51674,51685,51900,52017,52049,52152,52387,52441,52471,52672,52856,52945,53279,53315,53492,53516,53649,53664,53783,53932,53939,54059,54216,54235,54320,54614,54639,54670,54701,54763,54828,54836,54862,54927,55083,55175,55304,55411,55520,55569,55633,55714,55827,55835,55872,56033,56191,56294,56431,56509,56679,56683,56747,56835,56938,57033,57085,57131,57291,57372,57438,57456,57533,57563,57565,57652,57779,57781,57798,57842,57911,57966,58155,58226,58239,58269,58453,58461,58502,58559,58571,58578,58794,58840,58882,58941,59044,59070,59114,59136,59158,59236,59296,59377,59553,59624,59641,59653,59769,59857,59858,59955,60105,60109,60204,60237,60331,60479,60575,60584,60603,60735,60980,61008,61015,61254,61421,61492,61504,61651,61837,61926,62271,62300,62388,62396,62427,62440,62627,62631,62890,62981,63001,63005,63030,63182,63308,63312,63363,63650,63744,63842,63855,63925,64145,64242,64434,64525,64529,64569,64629,64693,64726,64727,64736,64802,64916,65041,65155,65259,65482,65499,65568,65592,65663,65814,65842,65897,65914,65944,65972,66015,66071,66073,66103,66183,66252,66413,66470,66589,66789,67142,67225,67256,67496,67519,67561,67692,67803,67816,67989,68141,68157,68222,68291,68375,68385,68406,68452,68464,68589,68644,68696,68722,68781,68830,68843,68864,68989,68994,68996,69039,69043,69059,69093,69243,69298,69395,69476,69537,69588,69643,69683,69774,70026,70042,70253,70450,70459,70484,70508,70709,70740,70868,70929,70943,71105,71211,71332,71359,71367,71403,71460,71619,71666,71787,71810,71894,72078,72135,72293,72311,72550,72590,72695,72741,72749,72794,72918,72949,73105,73386,73410,73502,73529,73545,73715,73840,73870,74047,74083,74094,74097,74237,74259,74280,74300,74342,74354,74456,74533,74567,74641,74678,74709,74728,74800,74945,75066,75193,75469,75485,75927,75973,76009,76054,76055,76123,76165,76488,76517,76575,76653,76793,76822,76879,76925,77100,77243,77271,77843,77861,78081,78395,78480,78527,78699,78718,78767,78775,78820,78874,78895,78901,79026,79214,79269,79511,79596,79611,79774,79901,80113,80178,80216,80261,80276,80332,80491,80498,80632,80662,80674,80815,80858,80877,80920,80965,81099,81212,81303,81318,81458,81545,81697,81708,81749,81822,81824,81984,82001,82003,82014,82215,82263,82271,82415,82606,82870,82910,82951,83042,83060,83127,83171,83290,83342,83453,83726,83763,83776,83801,83864,83866,83945,84074,84309,84345,84354,84365,84424,84633,84718,84752,84793,84830,84845,84916,85146,85158,85182,85353,85430,85572,85628,85657,85665,85686,85787,85793,85876,86103,86184,86329,86462,86493,86502,86619,86631,86840,86858,86941,87053,87083,87109,87217,87226,87296,87300,87405,87583,87596,87702,87797,87919,87938,87989,88078,88112,88413,88433,88542,88618,88793,88819,88978,89047,89049,89144,89376,89467,89483,89494,89512,89513,89558,89716,89748,90003,90165,90209,90269,90281,90376,90395,90422,90431,90512,90563,90587,90673,90708,90745,90811,90966,91034,91202,91236,91309,91400,91434,91572,91819,92032,92098,92110,92113,92120,92121,92161,92312,92355,92391,92424,92433,92480,92485,92929,92942,92964,93017,93025,93040,93297,93333,93361,93364,93385,93433,93555,93632,93648,93842,93941,94002,94079,94145,94204,94219,94297,94395,94491,94500,94510,94560,94569,94644,94661,94719,94772,94787,95091,95147,95191,95411,95637,95639,95669,95705,95721,95804,95812,95863,95897,96192,96351,96471,96497,96564,96590,96596,96701,96747,96773,96914,97220,97290,97297,97406,97484,97572,97898,97928,97942,98095,98198,98212,98217,98413,98437,98454,98551,98622,98651,98841,98843,98848,98888,98911,98984,99079,99159,99216,99281,99317,99332,99339,99409,99515,99641",
      "63,325,353,429,476,554,588,717,788,857,876,881,930,942,1039,1091,1154,1252,1279,1295,1489,1530,1579,1602,1604,1637,1686,1758,1863,1896,2225,4214,5006,5055,5059,5063,5187,5266,5325,5418,5576,5614,5667,5703,5817,5837,5911,5945,5959,5981,6074,6284,6299,6304,6317,6323,6326,6334,6362,6447,6461,6590,6730,6780,6803,6875,6978,7049,7081,7139,7270,7336,7433,7443,7468,7656,7665,7726,7815,7823,7842,7866,8017,8061,8119,8124,8134,8161,8181,8215,8243,8277,8358,8380,8453,8460,8539,8657,8709,8716,8722,8727,8738,8753,8874,8922,8926,8954,8989,8997,9174,9183,9221,9350,9765,9885,10138,10159,10160,10164,10320,10336,10468,10489,10566,10587,10740,10781,10878,11371,11383,11436,11467,11609,11678,11751,11859,11962,12070,12136,12173,12373,12489,12610,12719,13116,13221,13268,13371,13413,13481,13495,13529,13911,14083,14097,14595,14599,14699,14764,14861,14881,14906,14968,15163,15239,15315,15408,15453,15488,15523,15766,15887,16111,16306,16646,16690,16733,17344,17346,17416,17521,17613,17666,18238,18865,19403,19470,19517,20100,20221,20542,20549,20588,20628,21033,21127,21192,21528,21672,21745,22107,22707,22805,22923,23112,23139,23263,23272,24146,24159,24176,24605,25145,25205,25219,25458,26281,26799,27061,27541,27834,28252,28362,28439,28819,28938,29001,29006,29134,29159,29209,29615,29793,29938,34203,35752,35870,35894,35941,36575,36576,36636,36695,36824,36923,36948,37017,37431,38010,38236,38278,38571,38939,38997,39069,39117,39132,39488,39607,39617,39674,39940,39983,41282,45421,45621,45807,45990,50037,50099,50147,50196,50201,50264,50297,50313,50393,50546,50653,50710,50805,50882,50976,50999,51493,51665,51666,51696,51745,51936,51964,52079,52089,52114,52322,52333,52543,52596,52599,52627,52683,52729,52765,52772,52774,52823,52974,52990,53002,53141,53195,53468,53493,53562,53683,53722,53788,53801,53893,53900,53985,54009,54024,54037,54120,54218,54228,54242,54509,54557,54601,54659,54669,54771,55096,55100,55114,55125,55131,55138,55425,55546,55562,55593,56053,56057,56132,56170,56256,56355,56506,56623,56664,56701,56797,57034,57102,57452,57483,57547,57634,57665,57667,57782,57852,57951,58044,58179,58215,58270,58273,58278,58397,58401,58444,58556,58596,58767,58848,59073,59179,59187,59301,59347,59615,59737,59787,59940,60098,60110,60196,60289,60328,60343,60385,60538,60698,60853,60883,60994,61030,61139,61177,61223,61323,61357,61359,61541,61619,61697,61747,61764,61768,61891,61908,62024,62046,62080,62085,62204,62284,62314,62437,62492,62517,62595,62614,62809,62812,62831,62880,62931,62958,63074,63277,63303,63372,63409,63542,63663,63702,63745,63756,63757,63758,63844,64165,64298,64374,64540,64581,64596,64685,64835,64848,64866,64915,64985,65002,65045,65072,65186,65213,65324,65395,65497,65518,65672,65726,65786,65817,65824,65868,66177,66207,66368,66382,66429,66499,66643,66688,66728,66765,66816,66844,66847,66899,66942,67101,67138,67276,67329,67443,67459,67476,67534,67654,67790,67854,67901,67972,67985,68019,68128,68146,68357,68359,68456,68484,68498,68543,68546,68583,68604,68680,68761,68768,68777,68787,68882,68917,68920,69066,69102,69146,69165,69186,69218,69451,69486,69506,69533,69547,69619,69742,70002,70081,70256,70265,70353,70374,70425,70493,70593,70647,70695,70734,70779,70822,70989,71021,71168,71547,71552,71569,71604,71632,72033,72331,72332,72351,72436,72514,72720,72739,72829,72912,73003,73024,73065,73089,73430,73461,73480,73517,73557,73599,73639,73728,73766,73812,73953,73954,74050,74219,74458,74464,74492,74505,74586,74617,74651,74777,74815,74916,74931,75174,75264,75282,75305,75317,75383,75489,75583,75641,75853,75876,75878,76269,76278,76283,76358,76392,76545,76581,76623,76753,76769,76967,77149,77253,77263,77300,77343,77445,77524,77658,77751,77997,78122,78125,78219,78253,78445,78451,78470,78523,78597,78600,78612,78676,78704,78728,78750,78888,79014,79221,79301,79315,79513,79910,80136,80234,80318,80472,80558,80724,80783,80882,81046,81065,81138,81182,81218,81357,81581,81660,81684,81859,81912,81985,82077,82090,82141,82342,82429,82441,82546,82673,82674,82759,83031,83088,83130,83300,83628,83692,83719,83754,83784,83871,83887,83985,84023,84241,84308,84324,84360,84412,84519,84542,84579,84590,84732,84775,84905,84940,85170,85201,85247,85260,85274,85307,85327,85488,85602,85639,85666,85702,85812,85814,85948,85990,86209,86517,86534,86757,86786,86875,86901,86912,86974,87035,87086,87106,87241,87263,87308,87318,87324,87435,87475,87483,87487,87505,87617,87636,87789,87847,87866,87889,87939,88064,88105,88109,88135,88237,88359,88474,88547,88606,88607,88628,88886,88889,88941,89189,89433,89468,89530,89829,89860,89952,90026,90057,90058,90217,90234,90283,90369,90388,90514,90561,90787,91008,91059,91379,91461,91541,91624,91648,91827,91844,91942,91975,91998,92164,92273,92332,92347,92394,92469,92547,92705,92744,92798,92864,93093,93197,93438,93440,93525,93568,93606,93658,93669,93705,93763,93889,93966,94008,94064,94154,94347,94392,94411,94423,94520,94606,94695,94698,94739,94919,94989,95103,95152,95281,95545,95584,95610,95700,95875,95888,95903,96093,96094,96363,96475,96493,96556,96592,96631,96822,96861,97068,97230,97244,97314,97315,97590,97766,97806,97808,97836,97958,98036,98153,98203,98286,98336,98468,98526,98606,98662,98792,98797,98913,99001,99054,99088,99149,99270,99551,99681,99723,99730,99809,99816,99898,99968,99989",
      "262,592,651,685,769,780,797,821,870,1258,1286,1343,1349,1407,1852,1877,1911,2165,2550,2554,2572,2605,2629,2666,3097,3149,3372,3581,3634,3769,4170,4276,4282,4388,4903,4988,5052,5064,5134,5158,5194,5331,5517,5555,5662,5820,5840,5892,5961,6018,6093,6149,6207,6340,6404,6470,6503,6508,6521,6611,6636,6711,6736,6775,6792,6797,6921,6954,6972,7176,7195,7368,7391,7399,7421,7572,7627,7824,7825,7841,7844,7976,8050,8083,8090,8102,8184,8280,8288,8355,8479,8555,8593,8599,8600,8658,8726,8740,8843,8900,9119,9270,9312,9342,9357,9367,9420,9458,9501,9540,9577,9608,9615,9638,9719,9758,9767,9817,9864,9875,9888,9895,9975,9989,10005,10039,10051,10189,10245,10288,10378,10483,10883,10955,11020,11070,11081,11085,11092,11252,11277,11339,11367,11671,11753,11772,11774,11793,11853,11964,11997,12013,12043,12127,12176,12199,12336,12400,12468,12582,12640,13013,13030,13216,13424,13431,13533,13545,13664,13685,13696,13890,13902,14001,14010,14011,14016,14062,14377,14445,14463,14491,14673,14679,14774,14871,14893,14957,15023,15052,15099,15166,15276,15347,15418,15573,15603,15613,15928,15946,16009,16082,16117,16375,16798,16847,17037,17062,17095,17249,17504,17565,17717,17770,17808,18069,18329,18340,18462,18567,18596,18660,18703,18841,18842,18877,18953,18975,19149,19308,19319,19584,19676,19687,19739,19821,19982,20050,20130,20155,20188,20335,20524,20532,20632,20647,20685,20732,20790,20894,20905,20972,21026,21152,21350,21396,21583,21593,21626,21703,21833,21931,22135,22182,22252,22501,22584,22768,23148,23242,23274,23391,23796,23812,24211,24344,24577,24651,24690,24754,24760,24820,24898,24903,24922,25088,25113,25522,25549,25571,25865,25966,26248,26278,26322,26419,26428,26743,26803,26886,26927,27136,27173,27194,27226,27247,27262,27283,27335,27395,27463,27496,27610,27664,27896,27954,27967,28258,28284,28355,28803,28869,29039,29068,29187,29204,29257,29378,29391,29489,29574,29635,29701,29778,29857,29867,30845,30874,30980,31078,31164,31328,31716,32314,32353,32354,32633,32760,32789,32865,33368,33398,33746,33851,34267,34876,35093,35162,35233,35542,35564,35582,35805,35885,35893,35944,35948,36010,36065,36183,36204,36552,36578,36622,36646,36798,37002,37069,37177,37189,37236,37436,37509,37862,37864,38003,38065,38068,38191,38225,38238,38253,38284,38348,38421,38448,38501,38694,38727,38743,38753,38756,38870,38881,38944,38950,38968,38994,39177,39304,39660,39746,39759,39935,40401,40553,40582,41247,41336,42224,42269,42585,42636,42644,42730,42880,42919,42939,42999,43006,43076,43090,43292,43343,43440,43444,43548,43591,43817,44068,44192,44222,44321,44552,44556,44847,44990,45126,45474,45614,45624,45629,45675,45690,45694,45776,45957,46592,46637,46718,46749,46803,47217,47270,47499,47638,47804,47895,48039,48094,48116,48164,48559,48591,48679,48911,48965,49098,49505,49610,50075,50131,50229,50320,50322,50363,50387,50411,50467,50482,50728,51009,51010,51143,51144,51163,51171,51174,51199,51229,51275,51319,51386,51453,51656,51849,51870,51889,52018,52052,52065,52109,52112,52148,52291,52327,52381,52408,52446,52449,52586,52743,52846,52925,52960,52993,53214,53223,53281,53399,53432,53465,53503,53651,53698,53914,53968,54029,54032,54039,54142,54340,54434,54439,54567,54589,54617,54803,54817,54920,54984,55086,55116,55186,55205,55242,55259,55282,55489,55575,55778,55842,55876,56017,56210,56235,56307,56400,56523,56528,56808,56847,56967,57096,57116,57144,57191,57234,57263,57280,57285,57310,57405,57454,57606,57713,57827,57829,57976,57988,58006,58164,58182,58338,58445,58489,58535,58560,58564,58645,58932,59030,59056,59080,59104,59171,59250,59302,59393,59606,59613,59786,59845,59971,60057,60066,60117,60286,60467,60684,60699,60788,60993,61038,61148,61384,61476,61495,61695,61740,61995,62026,62282,62337,62735,62742,62834,62896,62968,63130,63156,63289,63332,63396,63432,63511,63674,63693,64078,64084,64129,64192,64264,64271,64280,64399,64418,64446,64482,64577,64744,64822,64960,64966,65085,65101,65339,65596,65611,65634,65721,65793,65866,65869,66064,66239,66365,66391,66549,66677,66887,66941,66987,67011,67031,67071,67091,67157,67203,67296,67368,67409,67499,67606,67877,68035,68234,68240,68368,68533,68538,68551,68559,68602,68782,69118,69348,69411,69505,69581,69635,69673,69730,69820,69848,69895,70057,70164,70170,70175,70258,70400,70412,70420,70525,70535,70558,70594,70648,70707,70713,70743,70999,71058,71094,71337,71534,71667,71710,71785,71939,72026,72083,72099,72143,72416,72568,72645,72671,72728,72763,72792,72982,73029,73075,73080,73211,73273,73324,73330,73534,73642,73793,73837,73937,74242,74271,74375,74385,74396,74399,74430,74485,74491,74495,74499,74654,74671,74755,74810,74848,74986,75138,75327,75482,75568,75617,75708,75784,76093,76118,76136,76383,76444,76514,76898,76940,76975,77048,77057,77112,77131,77143,77314,77352,77435,77456,77460,77520,77588,77729,77765,77835,77900,78061,78472,78559,78823,78847,78909,78952,78980,78998,79110,79217,79278,79440,79474,79554,79721,79723,79802,79940,79961,80098,80130,80263,80288,80349,80356,80482,80495,80661,80796,80937,80961,81152,81229,81350,81371,81452,81461,81527,81540,81543,81561,81565,81604,81618,81630,81637,81657,81782,81994,82167,82355,82368,82372,82462,82477,82493,82523,82537,82655,82659,82809,82819,82927,83067,83168,83257,83332,83392,83498,83629,83666,83901,84073,84233,84267,84277,84280,84302,84351,84421,84427,84580,84698,84750,84777,84931,85096,85104,85116,85153,85164,85242,85246,85248,85362,85389,85426,85609,85616,85692,85710,85713,85837,85858,85992,86104,86156,86347,86402,86458,86696,86782,86816,86859,86991,87015,87151,87177,87196,87255,87614,87755,87777,87835,87925,87929,87934,88012,88267,88278,88288,88424,88501,88586,88729,88790,88808,89111,89242,89414,89480,89501,89505,89621,89623,89666,89678,89688,89841,89965,90018,90204,90207,90261,90300,90389,90517,90548,90556,90598,90675,90713,90727,90983,91045,91050,91085,91150,91179,91191,91234,91378,91412,91435,91443,91501,91502,91520,91525,91611,91662,91766,91946,92042,92151,92159,92194,92201,92205,92245,92296,92565,92615,92820,93044,93091,93164,93216,93225,93259,93266,93342,93608,93676,93717,93768,93965,93982,94056,94332,94586,94712,94758,94760,94832,94892,95123,95251,95285,95318,95350,95356,95429,95443,95563,95601,95628,95676,95746,95841,95937,95956,96068,96183,96186,96218,96233,96323,96326,96349,96455,96521,96540,96582,96591,96634,96809,97036,97124,97129,97399,97426,97542,97693,97711,97881,97959,97983,98200,98481,98816,98929,99187,99261,99418,99467,99475,99487,99658,99826,99859,99969,99999",
      "205,264,405,451,561,575,622,661,662,774,790,924,1114,1159,1160,1219,1253,1270,1283,1367,1403,1415,1575,1640,1719,1734,1751,1824,1883,1919,2044,2101,2181,2239,2553,2599,2604,2643,2688,2715,2854,2921,3091,3366,3611,3630,3868,4279,4309,4393,4417,4535,4555,4593,4725,4834,4918,4953,5039,5157,5160,5275,5341,5396,5528,5597,5727,5751,5822,5825,5967,6007,6039,6070,6078,6097,6205,6283,6454,6457,6520,6588,6723,6756,6757,6986,7012,7099,7254,7257,7362,7386,7422,7462,7485,7655,7700,7806,7829,7843,7853,7941,8194,8212,8321,8327,8406,8462,8712,8768,8819,8855,8857,8870,8955,8988,9020,9142,9195,9220,9224,9226,9228,9432,9499,9512,9901,10024,10311,10385,10502,10556,10563,10655,10691,10771,10780,10786,10800,10860,10985,10991,11037,11084,11163,11379,11380,11449,11667,11712,11806,11822,11867,11883,12021,12069,12174,12188,12191,12377,12406,12432,12637,12784,12962,12999,13173,13385,13556,13747,13811,13937,13975,14039,14359,14421,14436,14478,14712,14790,14849,14946,14976,15156,15170,15176,15558,15576,15585,15695,15767,15768,15784,15787,15797,15847,16099,16373,16435,16566,16623,16680,16720,16884,16953,16963,17002,17038,17066,17114,17168,17187,17318,17627,17690,17703,17793,17820,17858,17988,18147,18155,18277,18615,18713,18827,18970,19074,19158,19187,19198,19510,19535,19536,19644,19944,20067,20150,20290,20325,20374,20546,20689,20708,20719,20765,20913,21024,21029,21082,21316,21772,21826,21862,22156,22299,22593,22622,22642,23251,23517,23954,24125,24336,24429,24483,24660,24884,25178,25192,25214,25355,25417,25420,25593,25686,25718,25721,25885,25943,26009,26059,26085,26616,26677,26746,26748,26921,26967,26993,27002,27209,27260,27300,27306,27408,27429,27584,27772,27856,27858,27885,28069,28101,28180,28184,28193,28217,28220,28253,28292,28327,28389,28654,28761,28802,28806,28848,29173,29188,29376,29545,29744,29747,29771,29826,29836,29970,30085,30539,31593,31938,32202,32366,32714,32750,32808,32859,32895,32982,33420,33613,33652,33910,34058,34084,34432,34580,34751,34904,35128,35187,35699,35761,35802,35871,35949,35954,36452,36565,36623,36864,36940,37005,37188,37484,37563,37581,37628,37631,37693,37800,37858,38001,38188,38342,38349,38505,38557,38621,38692,38702,38759,38913,38976,39072,39223,39231,39558,39753,39816,39902,39923,40538,40775,40804,41071,41348,41772,41957,42001,42218,42238,42442,43573,43975,44183,44462,44545,44572,44637,44894,44964,45182,45250,45259,45430,45622,45794,45800,45882,46011,46223,46310,46489,47476,48047,48051,48099,48296,49124,49230,49837,49879,50152,50243,50258,50270,50303,50397,50406,50429,50434,50526,50533,50560,50624,50644,50684,50751,50856,50923,50934,50989,51003,51375,51422,51437,51531,51563,51576,51579,51678,51727,51737,51785,51797,52064,52221,52228,52285,52352,52453,52459,52552,52561,52655,52739,52776,52781,52815,52870,52902,53233,53389,53598,53671,53718,53984,54088,54180,54202,54249,54276,54380,54419,54442,54482,54544,54571,54577,54636,54648,54649,54797,55141,55224,55251,55273,55349,55377,55484,55522,55532,55559,55640,55806,55824,55829,56035,56192,56330,56416,56455,56518,56612,56737,56869,57005,57009,57126,57199,57208,57281,57337,57338,57369,57553,57590,57670,58301,58313,58434,58821,58835,58843,58870,58908,59041,59282,59320,59654,59669,59706,59771,59817,59823,59846,59915,60014,60042,60070,60210,60427,60432,60448,60499,60665,60928,61381,61393,61452,61572,61700,61792,61818,61972,62062,62148,62165,62334,62487,62496,62525,62565,62574,62638,62655,62709,62718,62774,62818,62861,62878,63094,63124,63145,63271,63603,63632,63848,63858,64002,64019,64042,64098,64174,64185,64286,64356,64488,64556,64620,64631,64665,64807,65012,65117,65212,65310,65388,65417,65429,65434,65553,65636,65806,65819,65959,66006,66047,66106,66285,66343,66443,66506,66513,66539,66727,66889,66894,66917,66945,66985,67278,67345,67396,67427,67480,67484,67558,67563,67644,67712,67736,67791,67808,67819,67826,67830,67845,67853,68014,68166,68565,68572,68797,68811,68891,68922,69007,69065,69137,69174,69210,69290,69368,69522,69601,69618,69621,69629,69690,69760,69768,69806,69838,70013,70028,70101,70144,70201,70266,70344,70435,70443,70449,70723,70797,70813,70930,70991,71045,71157,71305,71321,71351,71363,71629,71699,71817,71861,71927,71983,72047,72138,72216,72352,72415,72491,72494,72501,72750,72940,73209,73285,73551,73638,73681,73826,74017,74075,74203,74207,74245,74261,74371,74433,74449,74661,74683,74685,74785,74830,74874,74950,74977,74996,75003,75032,75059,75221,75225,75234,75272,75391,75399,75519,75529,75543,75702,75882,75891,75913,75935,76030,76220,76257,76281,76470,76511,76562,76590,76727,76821,76953,76977,76991,77031,77137,77140,77153,77371,77474,77532,77632,77695,77932,78064,78289,78376,78524,78551,78572,78623,78682,78730,78762,78786,78923,78950,78994,79097,79218,79388,79445,79590,79932,80104,80118,80164,80255,80345,80543,80555,80721,80733,80787,80795,80819,80944,80984,81009,81023,81073,81095,81123,81200,81302,81420,81492,81520,81621,81655,81760,81926,81927,81940,81964,81995,82191,82319,82325,82428,82485,82498,82701,82867,82963,82974,82976,83064,83084,83150,83184,83287,83569,83572,83591,83631,83732,83751,83903,83965,84444,84480,84876,84918,85066,85087,85252,85262,85299,85322,85355,85433,85507,85579,85680,85736,85870,86024,86084,86289,86624,86682,86700,86888,87002,87004,87017,87073,87123,87138,87183,87257,87334,87354,87372,87411,87611,87708,87733,87767,87897,87970,88084,88100,88113,88192,88323,88347,88506,88569,88633,88656,88666,88714,88782,88930,88983,89030,89043,89153,89175,89220,89406,89437,89506,89546,89575,89823,89866,90070,90071,90307,90581,90669,90835,90946,91005,91193,91200,91322,91373,91394,91433,91551,91569,91684,91873,91881,91947,92001,92024,92040,92081,92162,92209,92288,92398,92483,92697,92755,92785,92991,93083,93125,93176,93360,93455,93487,93520,93565,93584,93688,93711,93923,93929,94029,94123,94221,94233,94255,94291,94310,94391,94421,94424,94474,94867,94984,95012,95156,95376,95633,95775,95877,95938,96305,96474,96731,96860,96862,96961,96984,97078,97139,97281,97288,97351,97470,97492,97583,97719,97803,98016,98074,98147,98207,98238,98410,98529,98585,98588,98691,98717,98745,98785,98958,99296,99299,99355,99429,99451,99496,99585,99586,99631,99682,99975,99990,100000",
      "27,118,138,331,625,666,739,743,826,986,1197,1273,1313,1436,1462,1506,1594,2164,2295,2310,2323,2421,2445,2480,2616,2872,3224,3286,3305,3351,3399,3412,3520,3538,3661,3700,3763,4090,4105,4225,4241,4418,4443,4554,4610,4627,4682,4695,4737,4779,4815,5013,5077,5155,5195,5264,5268,5306,5317,5485,5663,5779,5806,5881,5905,5933,5978,6025,6026,6041,6161,6225,6247,6313,6322,6385,6403,6598,6716,6722,6770,6790,6815,6873,6885,6917,6953,7006,7157,7256,7267,7280,7281,7295,7563,7570,7663,7707,7728,7752,7767,7822,7906,7964,8007,8042,8079,8103,8193,8231,8362,8363,8527,8536,8628,8743,8758,8775,8853,8919,9105,9361,9390,9562,9571,9712,9732,9780,9839,9850,9915,9967,10044,10318,10342,10851,10861,10888,10905,10915,10929,10993,11013,11087,11102,11120,11145,11146,11295,11479,11620,11766,11824,11826,12260,12325,12445,12477,12601,12658,12662,12695,12762,12831,12839,12982,12996,13109,13208,13304,13337,13403,13451,13479,13567,13594,13740,13742,13849,13951,13971,14017,14051,14126,14172,14186,14312,14351,14617,14656,14689,14716,14729,14732,14748,14873,14939,15129,15185,15384,15414,15547,15664,15673,15680,15879,16002,16066,16086,16127,16128,16388,16421,16478,16631,16636,16722,16762,16913,16978,16996,17016,17359,17407,17409,17427,17482,17562,17685,17724,17768,17967,18070,18187,18203,18295,18845,18955,19005,19024,19027,19036,19041,19059,19083,19100,19194,19248,19316,19369,19376,19498,19706,19735,19763,19968,20352,20366,20718,20807,20904,21121,21299,21520,21854,21930,21950,22082,22125,22159,22177,22271,22315,22557,22596,22662,22939,22980,23159,23184,23280,23304,23331,23412,23465,23482,23508,23544,23575,23606,23787,23793,23848,23900,24192,24457,24476,24647,24661,24717,24859,24950,25065,25103,25230,25237,25275,25338,25483,25496,25541,25620,25681,25690,25733,25883,25959,25978,26172,26224,26239,26339,26393,26432,26702,26834,27031,27039,27212,27215,27225,27346,27349,27424,27668,27696,27758,27817,27886,27928,28030,28074,28123,28271,28302,28306,28398,28526,28543,28563,28740,28776,28845,28875,28906,28939,28966,28969,28971,29010,29034,29162,29234,29265,29295,29335,29381,29413,29528,29622,29757,29786,29843,29853,29943,29952,29959,29998,30120,30178,30288,30641,30711,30942,31042,31076,31125,31250,31337,31366,31455,31734,31890,32073,32238,32382,32469,32526,32972,32990,33251,33282,33310,33716,33929,33958,34007,34180,34248,34325,34464,34469,34614,34616,34644,34667,34992,35215,35563,35675,35714,35757,35865,36021,36213,36255,36257,36326,36359,36448,36464,36518,36645,36684,36697,36825,36863,36895,36952,37067,37121,37191,37208,37256,37397,37426,37438,37514,37524,37540,37635,37742,37798,37814,37875,37886,38074,38076,38266,38425,38481,38506,38852,39079,39080,39115,39338,39425,39451,39543,39672,39792,39877,39929,40230,40379,40459,40636,40752,40981,41133,41242,41249,42003,42005,42090,42096,42152,42179,42200,42271,42412,42684,42803,42953,43179,43318,43367,43413,43428,43575,43651,43789,44009,44165,44171,44291,44354,44495,44641,44662,44777,44895,44940,45131,45190,45201,45314,45365,45398,45456,45606,45611,45736,45869,45884,45902,46073,46457,46532,46539,46661,46789,46827,46877,47087,47419,47919,48096,48207,48252,48848,48952,48963,49004,49093,49112,49205,49351,49391,49392,49460,49493,49681,49928,50025,50144,50224,50250,50253,50298,50325,50565,50615,50736,51209,51246,51338,51355,51393,51433,51467,51469,51470,51541,51603,51769,51809,51952,52105,52120,52128,52222,52288,52400,52460,52476,52492,52579,52607,52620,52762,52922,53070,53073,53104,53226,53271,53348,53410,53459,53506,53577,53583,53625,53639,53653,53720,53837,53872,53882,53926,53957,54021,54159,54172,54186,54206,54252,54260,54418,54462,54467,54484,54534,54635,54730,54772,54837,54844,54884,54891,55185,55296,55563,55740,55798,55832,55871,55892,55955,56008,56023,56051,56108,56177,56185,56219,56242,56505,56613,56708,56821,56975,57182,57295,57327,57383,57400,57465,57490,57496,57538,57573,57708,57749,57770,57802,57841,57907,58023,58028,58081,58195,58225,58254,58268,58323,58511,58512,58518,58547,58916,58931,58958,59084,59329,59365,59417,59510,59589,59618,59710,59729,59735,59806,59901,60044,60093,60170,60173,60174,60223,60248,60274,60333,60335,60445,60477,60540,60639,60963,60971,61078,61111,61112,61128,61191,61207,61251,61353,61439,61556,61565,61723,61898,61933,62027,62031,62047,62213,62239,62369,62478,62623,62662,62833,62954,63053,63082,63119,63140,63177,63188,63279,63387,63558,63578,63660,63794,63874,63933,63966,63986,64001,64065,64186,64189,64208,64355,64381,64503,64694,64919,65023,65142,65192,65303,65453,65589,65622,65854,65863,65922,65932,65995,66070,66144,66199,66383,66396,66428,66525,66636,66777,66992,67000,67003,67111,67162,67286,67362,67492,67533,67669,67690,67695,67783,67794,67948,68182,68299,68319,68334,68445,68493,68497,68570,68708,68764,68771,68774,68793,68826,68877,69010,69025,69062,69244,69328,69932,69933,70032,70179,70302,70340,70348,70388,70395,70409,70448,70672,70789,70908,70920,70941,71005,71063,71069,71289,71318,71400,71498,71511,71626,71670,71738,71865,72198,72263,72264,72304,72342,72383,72419,72564,72603,72686,72704,72865,72957,73087,73092,73144,73320,73336,73339,73356,73380,73420,73439,73518,73564,73768,73771,73820,73859,73863,73879,73967,73985,74009,74266,74408,74409,74413,74486,74497,74595,74679,74882,75103,75162,75209,75288,75414,75597,75647,75698,76081,76090,76137,76235,76341,76435,76440,76497,76532,76557,76571,76596,76755,76848,76976,77026,77094,77141,77215,77434,77436,77449,77479,77534,77539,77582,77764,77767,77851,77911,77950,77952,78080,78232,78404,78409,78431,78460,78496,78656,78732,78902,78905,78930,78992,79106,79139,79181,79200,79201,79331,79365,79432,79821,79843,79852,79953,79987,80053,80054,80105,80134,80168,80295,80432,80564,80603,80627,80659,80801,80825,80884,80966,81045,81156,81159,81328,81394,81418,81450,81479,81659,81707,81796,81837,81879,81888,81903,81906,81967,81992,82029,82095,82255,82365,82720,82773,82814,82912,83061,83140,83248,83382,83714,83731,83853,83874,84014,84015,84055,84167,84364,84461,84495,84522,84638,84687,84837,84843,84871,84910,84984,85044,85134,85143,85526,85573,85582,85589,85596,85607,85613,85771,85877,85905,85964,85994,86015,86023,86071,86073,86167,86223,86302,86340,86362,86489,86810,86856,86867,86938,87069,87072,87236,87276,87321,87525,87585,87604,87769,87808,87965,87971,87988,88200,88253,88265,88382,88554,88598,88664,88715,88862,88871,88915,89196,89303,89348,89389,89498,89510,89552,89554,89581,89783,89843,89961,90189,90227,90399,90513,90580,90611,90628,90654,90741,90754,90786,90953,91074,91142,91153,91233,91507,91818,92073,92196,92206,92336,92393,92415,92466,92581,92664,92671,92684,92686,92797,92815,92886,93107,93116,93161,93179,93269,93279,93384,93446,93449,93513,93628,93713,93761,93772,93898,93958,94007,94246,94311,94535,94538,94643,94753,94781,94871,95011,95073,95181,95245,95450,95822,95829,95866,95933,95948,95957,95965,95978,96054,96092,96144,96177,96197,96553,96641,96707,96734,96769,96938,96993,97025,97050,97072,97138,97141,97332,97438,97443,97531,97564,97568,97969,97971,98136,98194,98282,98354,98515,98575,98741,98747,98810,98881,98887,98935,98980,99018,99033,99148,99285,99430,99491,99533,99564,99684,99836,99948",
      "45,50,64,106,115,159,169,221,1793,1922,1936,2079,2127,2251,2255,2270,2335,2341,2396,2477,2590,2662,2680,2706,2711,2767,2818,2832,2836,2882,2909,2956,3152,3206,3446,3522,3714,4098,4099,4191,4204,4390,4458,4601,4611,4747,4853,4866,4948,5544,6980,7949,7973,9404,9412,9476,9517,9545,9546,9621,9640,9678,9697,9716,9721,9808,9824,9944,10020,10077,10086,10090,10096,10210,10381,10590,10630,10647,10666,10689,10745,10837,10856,10951,11038,11046,11179,11197,11198,11203,11217,11236,11255,11293,11369,11445,11477,11540,11597,11654,11730,11882,11920,12015,12036,12072,12129,12186,12224,12243,12262,12338,12395,12414,12419,12471,12528,12547,12654,12661,12699,12710,12737,12758,12775,12813,12823,12889,12892,12916,12930,12965,12973,12984,13003,13022,13041,13060,13112,13117,13204,13223,13231,13250,13326,13363,13370,13381,13398,13402,13404,13421,13459,13516,13573,13599,13601,13606,13630,13649,13654,13687,13703,13706,13725,13736,13744,13763,13782,13820,13858,13885,13906,13934,13953,13978,14029,14048,14100,14105,14124,14162,14181,14200,14221,14238,14245,14276,14295,14308,14314,14333,14352,14435,14444,14448,14509,14519,14525,14534,14544,14580,14618,14620,14637,14675,14694,14747,14770,14827,14839,14846,14884,14909,14922,14935,14972,14979,15017,15102,15134,15147,15149,15169,15188,15207,15208,15264,15283,15302,15318,15321,15397,15416,15432,15479,15480,15509,15511,15530,15625,15644,15653,15666,15682,15701,15720,15740,15758,15760,15791,15815,15834,15853,15872,15891,15948,15967,15986,16043,16081,16124,16138,16157,16214,16225,16233,16252,16253,16268,16290,16309,16333,16347,16366,16404,16461,16499,16518,16575,16581,16613,16640,16651,16662,16670,16704,16746,16765,16770,16775,16784,16822,16947,16974,16993,17048,17088,17126,17164,17183,17221,17240,17275,17297,17312,17373,17392,17430,17449,17465,17506,17525,17561,17601,17608,17620,17715,17767,17772,17791,17810,17848,17886,17898,17924,17976,18038,18054,18057,18114,18171,18190,18209,18266,18285,18293,18311,18335,18342,18350,18361,18380,18437,18456,18469,18494,18525,18532,18646,18665,18684,18724,18817,18931,18943,18950,18988,19022,19032,19045,19064,19088,19121,19140,19178,19197,19233,19254,19259,19303,19311,19330,19364,19366,19387,19406,19444,19475,19481,19482,19501,19520,19539,19615,19663,19672,19691,19707,19729,19767,19786,19838,19862,19881,19888,19900,19938,19957,19995,20003,20014,20052,20128,20147,20223,20226,20299,20318,20333,20337,20356,20375,20388,20390,20394,20432,20439,20470,20481,20489,20498,20508,20587,20603,20641,20648,20679,20736,20755,20772,20793,20812,20850,20869,20909,20983,21002,21025,21045,21078,21090,21097,21116,21154,21173,21198,21211,21249,21268,21287,21325,21363,21382,21420,21439,21472,21477,21519,21534,21553,21566,21610,21629,21724,21743,21773,21838,21876,21893,21895,21914,21933,21952,21971,22009,22047,22054,22060,22066,22102,22104,22123,22142,22153,22199,22218,22274,22275,22294,22302,22332,22351,22364,22370,22389,22408,22410,22427,22446,22458,22465,22541,22560,22579,22619,22636,22639,22655,22693,22712,22731,22750,22753,22788,22826,22890,22898,22902,22997,23029,23073,23079,23092,23130,23146,23149,23187,23200,23206,23219,23225,23244,23282,23301,23358,23377,23386,23392,23396,23415,23434,23491,23510,23529,23567,23624,23643,23667,23681,23700,23719,23738,23750,23753,23757,23814,23845,23852,23871,23890,23902,23928,23947,23966,24023,24042,24061,24080,24118,24156,24175,24213,24222,24232,24270,24308,24346,24365,24384,24403,24410,24441,24460,24479,24494,24498,24517,24534,24536,24555,24574,24631,24669,24688,24707,24726,24764,24799,24840,24878,24897,24916,24927,24983,24992,25006,25011,25075,25125,25144,25189,25201,25220,25296,25315,25356,25360,25384,25391,25410,25448,25502,25543,25555,25562,25581,25619,25633,25657,25676,25695,25752,25771,25790,25793,25809,25847,25859,25866,25874,25942,25961,25980,25999,26020,26036,26037,26056,26075,26094,26113,26120,26170,26183,26246,26265,26303,26341,26379,26417,26436,26474,26493,26512,26623,26626,26645,26664,26740,26755,26892,26911,26949,26986,27042,27063,27246,27265,27272,27291,27310,27389,27436,27465,27671,27766,27804,27874,27877,28006,28070,28089,28127,28182,28196,28218,28241,28326,28338,28380,28447,28451,28462,28504,28507,28583,28611,28678,28694,28881,28962,29077,29225,29271,29323,29350,29433,29475,29558,29667,29670,29780,29851,29892,29894,29913,30007,30036,30065,30094,30123,30152,30210,30239,30268,30355,30384,30413,30442,30471,30500,30521,30529,30558,30587,30590,30616,30645,30674,30703,30725,30732,30761,30782,30790,30819,30848,30877,30906,30935,30964,30993,31022,31051,31080,31109,31138,31160,31167,31196,31225,31242,31254,31283,31312,31331,31341,31399,31428,31457,31486,31541,31544,31573,31602,31631,31660,31683,31689,31718,31747,31776,31805,31834,31863,31892,31921,31950,31966,31979,32008,32037,32058,32066,32095,32124,32149,32153,32182,32211,32217,32240,32269,32298,32327,32356,32385,32414,32443,32501,32530,32559,32588,32617,32646,32675,32704,32733,32762,32791,32820,32849,32878,32907,32936,32959,32965,32994,33023,33052,33081,33110,33139,33168,33185,33197,33226,33255,33284,33342,33371,33400,33429,33458,33487,33516,33545,33574,33603,33632,33661,33690,33719,33748,33777,33806,33835,33864,33881,33893,33922,33951,33980,34025,34038,34067,34096,34125,34154,34183,34200,34212,34270,34286,34299,34328,34357,34374,34386,34415,34444,34446,34465,34473,34502,34531,34560,34589,34618,34676,34705,34722,34734,34763,34792,34821,34850,34879,34908,34937,34966,34995,35050,35096,35140,35153,35170,35201,35228,35242,35259,35326,35333,35373,35397,35407,35443,35478,35512,35526,35585,35619,35624,35628,35644,35717,35738,35750,35776,35835,35906,35931,35965,36005,36024,36049,36050,36087,36147,36206,36277,36336,36358,36361,36395,36418,36427,36467,36526,36555,36585,36612,36649,36708,36731,36767,36838,36878,36897,36956,37028,37087,37146,37210,37257,37269,37328,37356,37359,37399,37458,37517,37541,37552,37589,37609,37648,37707,37771,37830,37889,37916,37919,37960,38019,38078,38112,38150,38209,38268,38292,38332,38391,38450,38521,38580,38626,38639,38770,38829,38893,38929,38934,38952,39011,39082,39141,39171,39200,39272,39331,39354,39390,39454,39513,39572,39643,39702,39761,39788,39833,39892,39951,40000,40001,40002,40003,40004,40005,40006,40007,40008,40009,40010,40011,40012,40013,40014,40015,40016,40017,40018,40019,40020,40021,40022,40023,40024,40025,40026,40027,40028,40029,40030,40031,40032,40033,40034,40035,40036,40037,40038,40039,40040,40042,40043,40044,40045,40046,40047,40048,40049,40050,40051,40052,40053,40054,40055,40056,40057,40058,40059,40060,40061,40062,40063,40064,40065,40066,40067,40068,40069,40070,40071,40072,40074,40076,40077,40078,40079,40081,40082,40083,40084,40085,40086,40087,40089,40090,40091,40093,40094,40095,40096,40097,40098,40099,40100,40120,40149,40178,40207,40236,40265,40294,40323,40352,40381,40410,40439,40468,40497,40526,40555,40584,40613,40642,40671,40700,40729,40758,40787,40816,40845,40874,40903,40932,40961,40990,41019,41048,41077,41106,41135,41164,41193,41251,41280,41309,41338,41367,41396,41425,41454,41483,41512,41541,42011,42040,42069,42098,42127,42146,42149,42187,42247,42277,42352,42403,42409,42427,42482,42514,42561,42563,42567,42589,42628,42664,42697,42751,42765,42796,42808,42826,42847,42901,42942,42947,42988,43042,43063,43138,43149,43153,43181,43187,43225,43236,43257,43259,43296,43300,43321,43375,43411,43421,43462,43490,43515,43537,43581,43612,43699,43712,43752,43757,43774,43806,43807,43849,43872,43896,43936,43969,44011,44042,44051,44086,44173,44219,44248,44268,44281,44293,44297,44323,44342,44377,44410,44440,44442,44485,44533,44560,44574,44629,44647,44698,44708,44722,44754,44775,44797,44817,44857,44884,44906,44918,44920,44959,45014,45034,45107,45155,45175,45216,45274,45295,45347,45357,45367,45404,45413,45462,45534,45562,45591,45602,45649,45699,45721,45747,45778,45801,45836,45908,45933,45965,46043,46068,46091,46102,46153,46167,46171,46186,46212,46218,46225,46235,46241,46296,46359,46416,46422,46427,46507,46610,46611,46614,46623,46645,46656,46677,46678,46706,46717,46731,46794,46798,46817,46847,46866,46917,47036,47041,47069,47084,47089,47119,47161,47195,47257,47260,47277,47306,47309,47330,47345,47444,47475,47477,47487,47494,47530,47576,47611,47616,47618,47626,47628,47652,47668,47679,47750,47755,47758,47759,47762,47813,47822,47855,47857,47860,47863,47962,47983,47992,48005,48008,48091,48107,48160,48187,48264,48318,48406,48430,48473,48490,48491,48495,48568,48602,48609,48615,48680,48786,48791,48793,48797,48823,48833,48841,48886,48891,48907,48910,48913,48925,48929,48953,48984,48999,49051,49059,49079,49117,49142,49175,49191,49192,49233,49239,49289,49291,49298,49386,49405,49451,49475,49520,49522,49534,49576,49661,49713,49731,49757,49759,49760,49789,49869,49916,49918,49942,49962,49967,49988,50013,50174,50194,50318,50336,50388,50468,50489,50502,50528,50611,50650,50723,50763,50791,50922,50941,51007,51125,51258,51281,51418,51452,51507,51572,51582,51635,51748,51947,51957,51967,51994,52002,52042,52172,52236,52376,52402,52421,52588,52702,52737,52758,52790,52878,52885,52904,53039,53155,53228,53452,53460,53461,53464,53514,53560,53672,53686,53713,53847,53973,54001,54057,54135,54136,54147,54217,54271,54284,54450,54453,54463,54562,54739,54765,54860,55283,55427,55438,55469,55536,55540,55543,55595,55791,55930,55961,55972,56083,56094,56154,56499,56586,56723,56843,56873,57007,57058,57082,57212,57227,57256,57296,57326,57418,57567,57650,57666,57720,57724,57838,57850,57861,57880,57997,58257,58353,58501,58743,58783,58816,58861,58926,59116,59210,59243,59509,59537,59602,59633,59751,59799,59801,59834,59889,60073,60094,60166,60189,60278,60296,60347,60409,60520,60532,60681,60711,60822,61056,61124,61239,61298,61426,61475,61775,61910,61976,61982,62292,62341,62384,62504,62515,62603,62605,62696,62717,62767,62841,63195,63339,63391,63715,63815,63835,63871,63912,64088,64121,64218,64263,64354,64397,64414,64521,64536,64579,64630,64783,64841,64885,64893,65040,65109,65205,65248,65280,65385,65415,65474,65643,65809,65830,65882,65929,66101,66109,66214,66271,66277,66295,66333,66426,66455,66498,66665,66696,66715,66775,66790,66806,66836,66896,66918,67253,67385,67509,67511,67624,67776,67865,67958,68610,68664,68731,68738,68786,68863,68921,69049,69079,69189,69194,69202,69225,69231,69322,69469,69563,69589,69717,69781,69823,69940,69965,70084,70088,70092,70234,70326,70361,70446,70477,70622,70625,70627,70644,71119,71383,71441,71477,71539,71546,71657,71739,71769,72164,72336,72441,72483,72504,72505,72524,72650,72665,72861,73112,73158,73174,73241,73402,73483,73512,73765,73767,73783,73790,73807,73910,74136,74305,74343,74349,74459,74473,74482,74564,74659,74682,74770,74935,74946,75205,75456,75665,75679,75700,75734,75826,75934,75971,76308,76403,76522,76567,76639,76719,76721,76767,76893,76981,76996,77090,77111,77246,77281,77283,77363,77567,77625,77961,77979,78250,78312,78369,78388,78487,78505,78512,78542,78586,78759,78857,78887,78927,79045,79134,79268,79297,79357,79431,79457,79483,79484,79535,79624,79671,79698,79703,79839,79882,79923,79968,80139,80238,80252,80340,80390,80416,80535,80595,80601,80652,80730,80742,80743,80808,80895,80899,80903,80952,80986,81012,81144,81234,81267,81347,81356,81456,81506,81544,81583,81641,81728,81797,81816,81851,81925,82031,82102,82138,82186,82227,82390,82407,82414,82538,82553,82558,82668,82671,82770,82869,83111,83114,83219,83222,83344,83351,83433,83491,83527,83547,83553,83566,83592,83643,83644,83662,83794,83944,84081,84142,84289,84312,84340,84411,84528,84839,84964,84972,84989,85018,85024,85041,85190,85217,85345,85347,85391,85412,85487,85489,85498,85549,85554,85638,85677,85790,85873,85888,85890,86012,86153,86188,86220,86380,86414,86558,86665,86879,86911,86949,86951,87074,87136,87209,87358,87490,87561,87715,87717,87744,87770,87844,87895,88071,88087,88121,88169,88305,88371,88389,88603,88678,88710,88713,88743,88766,89070,89107,89119,89124,89298,89305,89378,89427,89435,89438,89441,89594,89665,89672,89925,90013,90048,90092,90119,90144,90199,90212,90319,90327,90461,90562,90660,90696,90706,91003,91033,91203,91301,91307,91328,91335,91360,91476,91483,91709,91752,91840,92013,92025,92067,92100,92256,92319,92606,92626,92644,92668,92722,92724,92736,92770,92796,93252,93264,93411,93420,93480,93511,93527,93558,93723,93731,93734,93791,93809,93826,93828,93848,93868,93874,93876,93933,94335,94384,94385,94575,94682,94722,94775,94906,94972,95158,95160,95243,95428,95492,95549,95731,95756,95845,95859,95893,95943,95985,96022,96055,96232,96251,96278,96309,96328,96466,96571,96629,96638,96836,96902,96949,97013,97067,97104,97170,97183,97203,97206,97218,97347,97517,97524,97757,97811,97843,97876,98009,98105,98152,98155,98404,98676,98689,98748,98751,98767,98891,99050,99182,99416,99506,99604,99648,99677,99742,99750,99789,99802,99803,99885,99909,100001",
      "46,65,103,105,122,160,182,186,191,2054,2143,2272,2303,2350,2355,2463,2498,2592,2626,2636,3075,3137,3794,3820,3861,3875,3922,4403,4406,4504,4528,4703,4764,4804,6140,6621,7688,8387,9280,9394,9451,9555,9565,9566,9584,9641,9660,9698,9755,9793,9812,9831,9926,10021,10037,10078,10097,10131,10192,10363,10648,10819,10914,10933,11009,11066,11104,11161,11180,11218,11237,11256,11313,11408,11468,11598,11617,11636,11750,11930,11978,12092,12095,12121,12130,12149,12187,12225,12244,12282,12301,12317,12358,12360,12396,12415,12469,12510,12529,12624,12738,12776,12814,12833,12928,12947,12957,12985,13061,13099,13118,13194,13213,13251,13270,13289,13310,13327,13384,13441,13555,13593,13612,13650,13688,13707,13726,13764,13783,13802,13821,13859,13916,13992,14030,14087,14139,14143,14144,14163,14182,14220,14239,14258,14277,14296,14315,14334,14410,14429,14484,14524,14543,14638,14657,14752,14809,14810,14847,14863,14866,14886,14923,14951,14961,15018,15075,15094,15104,15151,15246,15265,15284,15322,15341,15360,15455,15493,15512,15588,15626,15645,15690,15702,15721,15835,15854,15873,15892,15947,15949,16006,16015,16025,16044,16078,16101,16120,16136,16158,16177,16207,16215,16234,16272,16291,16310,16329,16348,16367,16405,16424,16443,16481,16484,16500,16520,16538,16549,16587,16614,16633,16635,16652,16671,16728,16747,16785,16823,16842,16851,16861,16899,16909,16956,16975,16986,16994,17070,17108,17146,17203,17222,17298,17317,17393,17433,17450,17469,17507,17545,17564,17640,17678,17695,17697,17716,17735,17754,17773,17792,17811,17868,17887,17925,17944,18001,18058,18077,18093,18115,18153,18163,18172,18248,18267,18286,18305,18324,18343,18362,18381,18400,18419,18438,18533,18552,18571,18586,18609,18647,18666,18738,18761,18780,18818,18894,18932,18951,18989,19023,19057,19061,19065,19103,19112,19123,19160,19179,19236,19253,19274,19331,19333,19388,19407,19426,19445,19464,19502,19578,19599,19673,19692,19703,19711,19730,19749,19768,19806,19825,19863,19882,19920,19939,19958,19996,20007,20015,20043,20064,20091,20129,20148,20186,20205,20281,20300,20301,20338,20357,20376,20414,20452,20471,20509,20528,20566,20585,20604,20642,20661,20678,20699,20737,20756,20811,20813,20870,20889,20908,20946,20962,20965,21003,21019,21089,21117,21131,21136,21155,21212,21226,21231,21269,21288,21307,21309,21345,21364,21440,21478,21497,21535,21554,21573,21592,21611,21668,21702,21725,21744,21762,21763,21766,21820,21858,21874,21877,21896,21915,21934,21991,22048,22064,22105,22124,22162,22200,22215,22219,22238,22276,22333,22352,22371,22390,22428,22447,22466,22485,22504,22523,22542,22572,22618,22637,22652,22656,22694,22713,22789,22808,22827,22846,22884,22903,22960,22998,23017,23036,23093,23150,23169,23302,23321,23359,23378,23435,23473,23492,23549,23587,23625,23644,23682,23701,23758,23777,23815,23834,23929,23967,23986,24028,24043,24062,24081,24100,24157,24195,24233,24252,24290,24309,24347,24348,24385,24404,24423,24442,24461,24480,24499,24520,24537,24556,24632,24708,24727,24746,24762,24765,24784,24831,24879,24917,24974,24993,25012,25050,25069,25137,25164,25183,25202,25278,25280,25373,25392,25430,25449,25509,25525,25563,25582,25639,25658,25677,25734,25753,25792,25810,25827,25867,25886,25905,25923,25924,25962,25981,26000,26019,26038,26057,26076,26095,26133,26163,26171,26228,26247,26302,26304,26342,26361,26399,26437,26456,26475,26478,26494,26513,26520,26532,26589,26608,26627,26654,26722,26741,26759,26782,26821,26855,26912,26937,26982,27169,27178,27308,27317,27330,27348,27400,27523,27539,27577,27608,27615,27630,27672,27711,27821,27862,27878,27938,27976,28166,28204,28251,28403,28434,28527,28660,28667,28698,28789,28907,28926,28945,28976,28994,29040,29053,29094,29192,29219,29230,29287,29363,29401,29414,29477,29502,29515,29599,29629,29666,29773,29781,29869,29882,30008,30037,30066,30095,30124,30153,30182,30211,30240,30269,30298,30327,30356,30385,30414,30443,30472,30501,30530,30559,30588,30617,30646,30675,30704,30733,30762,30791,30820,30849,30878,30907,30936,30965,30994,31023,31052,31139,31168,31197,31226,31255,31284,31313,31342,31371,31400,31429,31458,31487,31516,31545,31574,31603,31632,31661,31719,31748,31777,31806,31835,31864,31893,31922,31951,31980,32009,32038,32067,32096,32125,32154,32183,32212,32241,32270,32299,32328,32357,32386,32415,32444,32473,32502,32531,32560,32589,32647,32676,32705,32734,32763,32792,32821,32850,32879,32908,32937,32966,32995,33024,33053,33082,33111,33140,33169,33198,33227,33256,33285,33314,33343,33372,33401,33430,33459,33488,33517,33546,33575,33604,33633,33662,33691,33720,33749,33778,33807,33836,33865,33894,33923,33952,33981,34010,34039,34068,34097,34126,34155,34184,34213,34242,34271,34300,34358,34387,34416,34445,34474,34503,34532,34561,34590,34619,34648,34677,34706,34735,34764,34793,34822,34851,34880,34909,34938,34967,34996,35008,35061,35064,35069,35097,35171,35196,35260,35317,35334,35390,35408,35439,35471,35479,35527,35586,35645,35673,35718,35777,35836,35907,35932,35966,36025,36061,36088,36124,36130,36148,36207,36234,36278,36317,36337,36396,36468,36527,36586,36650,36685,36689,36709,36768,36793,36839,36898,36957,37029,37088,37115,37147,37211,37270,37302,37329,37400,37459,37518,37590,37649,37708,37772,37810,37831,37890,37961,37996,38020,38061,38071,38079,38114,38151,38210,38269,38333,38363,38392,38439,38451,38487,38522,38581,38640,38665,38668,38712,38771,38800,38830,38894,38953,39012,39036,39039,39083,39142,39201,39273,39332,39391,39455,39514,39550,39573,39644,39684,39703,39762,39799,39834,39952,40107,40136,40165,40194,40223,40252,40281,40310,40339,40368,40397,40426,40455,40484,40513,40542,40571,40600,40629,40658,40687,40716,40745,40774,40803,40832,40861,40919,40948,40977,41006,41035,41064,41093,41122,41151,41180,41209,41238,41267,41296,41325,41354,41383,41412,41441,41470,41499,41528,42000,42029,42058,42087,42116,42151,42176,42257,42266,42322,42341,42391,42416,42503,42552,42578,42606,42627,42653,42716,42740,42815,42890,42927,42977,43052,43082,43101,43112,43127,43173,43214,43289,43326,43364,43417,43451,43506,43526,43557,43601,43688,43733,43737,43763,43796,43798,43838,43918,43925,43989,44000,44046,44075,44162,44237,44261,44312,44349,44399,44454,44474,44549,44598,44614,44636,44711,44786,44824,44873,44941,44948,44974,44979,45023,45084,45096,45144,45171,45205,45263,45286,45336,45393,45434,45451,45490,45523,45569,45638,45669,45710,45767,45825,45897,45954,45977,45987,46095,46113,46120,46174,46301,46304,46306,46357,46412,46417,46490,46524,46567,46597,46604,46631,46676,46700,46734,46784,46906,46948,46965,46992,47022,47029,47055,47078,47149,47169,47173,47177,47190,47274,47293,47321,47391,47401,47423,47443,47458,47469,47495,47510,47690,47742,47751,47754,47764,47802,47841,47846,47883,47908,47914,47915,47921,47933,48063,48072,48088,48097,48119,48131,48140,48168,48169,48201,48208,48217,48272,48279,48300,48312,48370,48371,48413,48414,48454,48486,48510,48524,48541,48557,48583,48597,48614,48646,48654,48685,48720,48730,48748,48804,48826,48854,48894,48900,48917,49018,49092,49137,49220,49244,49280,49282,49315,49357,49416,49424,49425,49444,49472,49474,49484,49488,49513,49619,49636,49637,49645,49667,49683,49743,49758,49780,49800,49871,49968,49990,50034,50087,50090,50129,50220,50428,50441,50503,50539,50569,50593,50606,50719,50961,51014,51082,51347,51382,51488,51679,51723,51839,51903,51927,51937,51941,52006,52024,52122,52146,52160,52185,52386,52467,52520,52534,52537,52591,52640,53087,53102,53172,53251,53264,53371,53763,53777,53784,53790,53827,53892,53938,53948,54115,54227,54258,54491,54541,54581,54582,54705,54841,55001,55008,55088,55130,55163,55215,55519,55533,55692,55775,55777,55786,55826,55970,55995,56015,56104,56187,56212,56244,56350,56351,56403,56448,56485,56581,56654,56757,56823,56959,56984,57097,57109,57121,57408,57442,57596,57603,57716,57741,57796,57806,57885,57968,58005,58232,58424,58430,58492,58526,58586,58798,58808,58826,59085,59191,59274,59287,59289,59299,59307,59343,59358,59519,59584,59610,59697,59746,60184,60229,60542,60559,60625,60650,60776,61070,61087,61088,61201,61289,61518,61550,61602,61641,61730,61763,61950,61973,62117,62176,62209,62313,62375,62505,62518,62519,62567,62572,62645,62703,62766,62876,62904,62982,63041,63193,63207,63220,63329,63405,63521,63567,63619,63654,63686,63781,63936,64016,64125,64320,64376,64465,64557,64559,64615,64673,64740,64753,65200,65284,65295,65383,65472,65509,65644,65677,65744,65934,66002,66088,66137,66143,66256,66321,66402,66505,66612,66779,66927,66939,67089,67182,67319,67358,67386,67430,67557,67574,67597,67626,67631,67678,67703,67746,67815,67870,68139,68167,68262,68386,68431,68438,68439,68547,68775,68785,68833,68867,68869,69253,69270,69324,69331,69631,69679,69888,69903,69982,70009,70100,70117,70142,70165,70168,70260,70277,70511,70585,70702,70878,70963,71175,71247,71472,71722,71749,71791,71793,71803,71843,71890,71946,71951,71973,72018,72065,72076,72129,72174,72334,72345,72479,72516,72559,72588,72625,72680,72685,72751,72830,72935,73169,73271,73432,73465,73472,73699,73759,73761,73813,73927,73978,73991,74023,74142,74164,74331,74374,74436,74500,74575,74684,74752,74857,75021,75033,75154,75235,75421,75431,75483,75525,75636,75680,75730,75771,75808,75824,75856,75933,76007,76171,76363,76397,76525,76569,76587,76628,76680,76726,76814,76884,77207,77210,77267,77404,77557,77635,77708,77790,77846,77856,77970,77980,78007,78249,78337,78356,78367,78396,78855,78917,79154,79194,79238,79290,79304,79391,79413,79425,79478,79494,79551,79586,79683,79758,79836,79944,79975,80124,80132,80162,80251,80358,80417,80425,80629,80697,80761,80933,81068,81282,81403,81430,81451,81481,81510,81534,81541,81600,81650,81653,81758,81831,81910,82158,82265,82301,82559,82603,82609,82706,82787,82855,82907,82985,83058,83094,83158,83475,83576,83588,83619,83806,84030,84149,84259,84319,84513,84532,84635,84720,84735,84851,85105,85144,85230,85290,85324,85336,85348,85450,85466,85467,85474,85543,85551,85622,85684,85932,85944,85962,85972,86018,86102,86146,86197,86284,86378,86395,86429,86491,86524,86758,86804,86830,86882,86997,87050,87077,87534,87535,87553,87773,87975,88088,88143,88402,88562,88868,88923,88939,89140,89350,89364,89551,89636,89681,89731,89743,89915,89976,90030,90252,90288,90511,90530,90572,90607,90822,90915,90952,91048,91094,91113,91318,91342,91397,91398,91401,91460,91488,91529,91565,91643,91718,91810,91958,91991,92075,92079,92176,92177,92250,92302,92327,92371,92412,92595,92989,93085,93132,93188,93241,93280,93299,93357,93386,93401,93562,93577,93594,93639,93672,93685,93722,93744,93801,93836,93879,93917,93968,93978,94012,94061,94093,94210,94236,94279,94340,94486,94613,94676,94692,94757,94808,94857,94952,95048,95086,95099,95208,95213,95233,95283,95297,95726,95750,95816,95826,95914,96189,96238,96292,96333,96401,96603,96644,96766,96828,96893,96947,97029,97236,97284,97385,97541,97565,97742,97798,97827,97835,98060,98139,98192,98261,98273,98315,98381,98389,98453,98483,98518,98670,98758,98833,98917,98926,99058,99077,99081,99234,99297,99716,99841,99865,99895,99915,99918,99937,99939",
      "71,129,185,207,224,1735,1778,1949,1963,1982,2039,2151,2226,2244,2275,2314,2339,2617,2811,2913,2989,3016,3051,3139,3140,3177,3202,3247,3325,3339,3352,3407,3662,3790,3867,3972,4008,4034,4139,4243,4334,4612,4630,4647,4919,5120,5526,6955,9474,9563,9601,9620,9639,9658,9674,9675,9677,9696,9772,9810,9829,9924,9962,10095,10114,10722,10741,10760,10912,10950,11254,11292,11406,11442,11482,11539,11558,11615,11657,11881,11900,11957,11980,11995,12014,12033,12090,12147,12166,12204,12223,12261,12280,12318,12365,12413,12451,12588,12603,12717,12755,12888,12945,13021,13059,13078,13098,13102,13138,13230,13249,13263,13325,13344,13415,13420,13477,13496,13515,13536,13574,13591,13600,13667,13686,13705,13743,13762,13781,13819,13857,13887,13933,14085,14104,14180,14199,14218,14237,14256,14275,14294,14313,14389,14408,14427,14522,14541,14655,14659,14667,14674,14677,14693,14788,14860,14902,14921,14959,14990,14997,15016,15054,15073,15111,15130,15206,15215,15244,15263,15272,15282,15287,15320,15339,15358,15415,15434,15510,15515,15529,15539,15561,15567,15586,15643,15662,15700,15738,15776,15795,15833,15852,15871,15939,16004,16023,16061,16064,16080,16118,16175,16194,16199,16213,16232,16251,16289,16327,16346,16365,16403,16422,16455,16460,16479,16489,16498,16555,16574,16593,16612,16650,16707,16726,16745,16802,16821,16850,16859,16869,16878,16935,16954,17011,17030,17068,17087,17106,17119,17163,17182,17201,17258,17277,17279,17296,17315,17353,17410,17448,17467,17517,17524,17543,17547,17563,17600,17619,17623,17657,17676,17714,17771,17781,17790,17809,17828,17847,17885,17904,17923,17943,17961,17994,17999,18037,18090,18132,18189,18208,18227,18257,18259,18265,18284,18294,18322,18341,18360,18379,18398,18417,18436,18455,18512,18531,18550,18570,18584,18626,18664,18702,18721,18733,18759,18766,18778,18816,18835,18892,18911,18917,18930,18968,19006,19025,19082,19120,19139,19164,19196,19215,19234,19256,19268,19272,19276,19348,19349,19367,19372,19386,19424,19432,19443,19462,19476,19519,19538,19557,19587,19614,19652,19690,19709,19747,19785,19804,19807,19823,19842,19861,19880,19899,19918,19937,19956,19975,19994,20032,20070,20127,20165,20184,20208,20213,20241,20279,20298,20317,20336,20393,20412,20431,20450,20469,20507,20526,20545,20570,20583,20602,20621,20640,20682,20697,20716,20754,20773,20792,20849,20883,20887,20925,20944,20963,21020,21096,21115,21138,21172,21191,21229,21267,21286,21305,21327,21343,21419,21438,21457,21464,21476,21533,21552,21571,21590,21628,21666,21704,21723,21761,21780,21799,21818,21875,21894,21913,21932,21951,21970,22008,22027,22046,22084,22126,22198,22223,22236,22293,22303,22331,22369,22388,22407,22422,22426,22464,22502,22540,22559,22654,22692,22711,22717,22749,22764,22787,22800,22825,22844,22882,22901,22911,22920,22977,23015,23034,23053,23110,23129,23205,23224,23243,23262,23283,23300,23309,23357,23362,23395,23414,23452,23509,23528,23580,23591,23642,23661,23696,23706,23737,23756,23813,23832,23853,23863,23870,23889,23908,23927,23965,23984,24003,24022,24041,24060,24079,24098,24117,24174,24201,24212,24231,24288,24326,24364,24383,24440,24478,24497,24516,24554,24573,24611,24630,24687,24689,24706,24744,24763,24782,24801,24807,24839,24858,24877,24953,24976,25018,25029,25048,25143,25181,25276,25295,25314,25337,25352,25371,25425,25428,25466,25499,25504,25533,25542,25580,25599,25659,25675,25679,25694,25713,25740,25751,25770,25808,25831,25869,25903,25922,25941,25944,25960,26017,26055,26074,26093,26125,26150,26188,26207,26264,26283,26359,26397,26435,26473,26492,26511,26568,26587,26644,26663,26682,26761,26807,26854,26887,26948,27138,27141,27176,27197,27249,27290,27312,27366,27404,27445,27518,27694,27708,27803,27841,27857,27921,28050,28088,28107,28202,28324,28335,28446,28449,28454,28487,28582,28677,28754,28880,28943,28951,28987,29002,29354,29399,29421,29537,29551,29570,29627,29665,29705,29725,29741,29817,29862,29893,29912,29988,30006,30035,30064,30093,30122,30151,30180,30209,30238,30267,30296,30325,30354,30383,30412,30441,30470,30499,30528,30557,30586,30615,30644,30673,30702,30731,30760,30789,30818,30847,30876,30905,30934,30963,30992,31021,31050,31079,31108,31137,31166,31195,31224,31253,31282,31311,31340,31369,31398,31427,31456,31485,31514,31543,31572,31601,31630,31659,31688,31717,31746,31775,31804,31833,31862,31891,31920,31949,31978,32007,32036,32065,32094,32123,32152,32181,32210,32239,32268,32297,32326,32355,32384,32413,32442,32471,32500,32529,32558,32587,32616,32645,32674,32703,32732,32761,32790,32819,32848,32877,32906,32935,32964,32993,33022,33051,33080,33109,33138,33167,33196,33225,33254,33283,33312,33341,33370,33399,33428,33457,33486,33515,33544,33573,33602,33631,33660,33689,33718,33747,33776,33805,33834,33863,33892,33921,33950,33979,34008,34037,34066,34095,34124,34153,34182,34211,34240,34269,34298,34327,34356,34385,34414,34443,34472,34501,34530,34559,34588,34617,34646,34675,34704,34733,34762,34791,34820,34849,34878,34907,34936,34965,34994,35006,35095,35157,35169,35199,35245,35258,35290,35292,35322,35332,35386,35398,35406,35477,35525,35554,35584,35637,35643,35676,35716,35775,35834,35866,35905,35940,35964,36023,36057,36066,36086,36146,36199,36205,36240,36247,36276,36335,36394,36466,36525,36584,36648,36707,36766,36837,36870,36896,36955,37027,37054,37086,37145,37209,37268,37327,37398,37437,37457,37516,37588,37647,37706,37755,37756,37770,37829,37888,37923,37959,37995,38018,38077,38149,38176,38208,38267,38331,38390,38420,38449,38494,38520,38579,38638,38710,38769,38828,38892,38951,39010,39081,39131,39140,39187,39199,39237,39271,39312,39314,39330,39389,39453,39498,39512,39571,39622,39642,39701,39725,39760,39832,39891,39933,39944,39950,40110,40139,40168,40197,40226,40255,40284,40313,40342,40371,40400,40429,40458,40487,40516,40545,40574,40603,40632,40661,40690,40719,40748,40777,40806,40835,40864,40893,40922,40951,40980,41009,41038,41067,41096,41125,41154,41183,41212,41241,41270,41299,41328,41357,41386,41415,41444,41473,41502,41531,42002,42031,42060,42089,42118,42166,42178,42251,42268,42293,42328,42343,42418,42450,42466,42496,42505,42554,42571,42580,42655,42693,42711,42742,42772,42775,42802,42806,42817,42849,42892,42979,43019,43054,43093,43110,43129,43161,43162,43201,43216,43241,43274,43291,43366,43409,43453,43482,43492,43528,43552,43580,43588,43603,43622,43642,43690,43748,43765,43822,43825,43840,43860,43894,43927,43951,43981,43987,44002,44023,44029,44038,44077,44097,44100,44126,44138,44149,44154,44164,44200,44226,44239,44272,44300,44303,44314,44343,44353,44401,44476,44505,44509,44551,44582,44588,44638,44663,44675,44682,44713,44788,44832,44842,44875,44950,45025,45098,45130,45146,45169,45207,45230,45265,45338,45395,45433,45453,45525,45582,45640,45712,45769,45827,45862,45899,45956,45976,45989,46020,46034,46071,46133,46146,46150,46227,46251,46269,46272,46273,46320,46341,46346,46386,46392,46399,46400,46468,46472,46474,46494,46517,46523,46548,46564,46569,46650,46776,46779,46783,46786,46796,46818,46820,46836,46893,46935,46936,46939,46942,46954,47000,47008,47065,47094,47112,47157,47197,47227,47280,47343,47359,47394,47427,47448,47466,47467,47501,47503,47561,47583,47631,47700,47738,47763,47816,47821,47837,47851,47867,47872,47897,47923,47924,47960,47974,47976,47996,47998,48002,48042,48061,48066,48101,48135,48171,48204,48224,48261,48273,48336,48344,48378,48397,48417,48425,48462,48470,48485,48567,48594,48617,48622,48632,48659,48663,48672,48673,48681,48732,48747,48755,48801,48852,48860,48879,48905,48949,48982,49003,49073,49135,49185,49202,49215,49216,49232,49247,49256,49333,49366,49388,49397,49409,49421,49437,49502,49507,49529,49540,49564,49583,49591,49720,49727,49755,49766,49779,49817,49885,49888,49892,49948,49987,49997,50127,50177,50211,50344,50372,50476,50483,50493,50504,50505,50578,50740,50782,50927,50949,50992,50994,51054,51064,51291,51349,51380,51554,51708,51736,51747,51778,51783,51931,51966,52069,52192,52235,52271,52314,52366,52787,52789,52955,53005,53016,53049,53057,53186,53194,53239,53253,53285,53326,53341,53444,53604,53689,53792,53919,53927,53946,53961,53994,54141,54291,54412,54446,54531,54574,54725,54762,54906,54996,55223,55228,55337,55354,55374,55392,55454,55589,55641,55729,55831,55852,55905,55933,56048,56102,56401,56673,56746,56760,56811,56976,57015,57091,57142,57463,57555,57690,57767,57832,57862,57863,58036,58288,58316,58392,58394,58448,58541,58633,58672,58696,58911,58965,58974,58977,58983,59014,59082,59222,59259,59323,59451,59564,59752,59869,59912,60045,60276,60369,60407,60535,60598,60601,60774,60926,60945,61068,61241,61615,61728,61803,61821,62058,62093,62151,62234,62469,62498,62601,62691,62731,62797,62820,62827,62847,62893,62903,63181,63206,63292,63297,63453,63465,63543,63802,63865,64044,64091,64137,64259,64505,64555,64561,65007,65013,65052,65354,65407,65444,65475,65508,65530,65577,65597,65705,65825,65858,65900,65955,66020,66179,66280,66293,66300,66335,66337,66358,66441,66508,66546,66557,66586,66641,66709,66718,66720,66722,66736,66786,66954,67062,67178,67204,67299,67573,67648,67684,67718,67731,67764,67804,67850,67876,67918,67968,68000,68026,68029,68044,68048,68112,68127,68162,68333,68336,68390,68400,68535,68600,68655,68739,68931,68932,68981,69001,69002,69094,69205,69482,69527,69602,69661,69757,69764,69976,70000,70008,70167,70207,70293,70322,70386,70403,70416,70668,70801,70802,70861,71019,71154,71177,71191,71203,71218,71249,71404,71597,71623,71635,71746,71764,71807,71809,71859,72027,72133,72142,72147,72184,72186,72232,72241,72254,72341,72442,72449,72473,72886,72922,73017,73189,73310,73318,73362,73366,73440,73578,73580,73603,73652,73708,73721,73819,73885,73886,73926,74110,74183,74216,74235,74380,74388,74404,74442,74463,74525,74765,75070,75184,75211,75217,75256,75301,75416,75557,75749,75810,75857,75997,76016,76043,76057,76167,76372,76377,76411,76416,76548,76657,76739,76783,76858,76866,76946,76957,77160,77248,77254,77261,77262,77412,77656,77840,77855,77903,78217,78463,78624,78799,78842,78903,78951,78987,79453,79522,79552,79865,79904,79997,80156,80247,80250,80343,80509,80521,80539,80596,80623,80680,80707,80717,80971,81042,81079,81290,81359,81369,81415,81440,81606,81642,81643,81649,81803,81846,81883,82004,82021,82190,82221,82391,82435,82458,82541,82551,82640,82725,82751,82810,82841,82916,83303,83386,83474,83665,83713,83832,84028,84090,84121,84158,84195,84218,84236,84282,84392,84473,84507,84644,84714,84740,84757,84805,85188,85214,85420,85479,85731,85789,85987,86063,86095,86131,86150,86169,86191,86230,86281,86296,86412,86482,86554,86590,86647,86651,86759,86768,86776,86780,86787,86923,87036,87046,87088,87176,87246,87375,87386,87441,87577,87855,87961,88057,88059,88434,88532,88538,88716,88732,88825,88830,89022,89191,89240,89371,89570,89601,89682,89758,89764,89868,89908,89913,89926,89938,90152,90215,90230,90322,90386,90446,90476,90765,90796,90885,90950,91134,91238,91310,91349,91453,91605,91696,91771,91775,91843,91895,91967,92015,92107,92237,92341,92510,92525,92529,92621,92673,92733,92740,92802,92803,92937,93054,93064,93071,93239,93471,93504,93537,93541,93567,93634,93729,93808,93810,93960,93997,94009,94069,94103,94105,94309,94566,94597,94795,95050,95134,95173,95196,95311,95359,95397,95445,95478,95491,95530,95615,95757,95776,95792,95871,95873,96007,96145,96188,96205,96281,96367,96470,96708,96726,96762,96916,96927,96935,96942,97061,97097,97126,97195,97249,97472,97537,97562,97639,97673,97697,97801,97805,97815,97875,97901,97908,97948,98229,98396,98397,98425,98462,98612,98663,98793,98853,98915,98951,98981,99025,99304,99315,99319,99523,99596,99685,99694,99758,99993,100002",
      "203,256,286,340,379,400,421,424,425,450,461,468,499,586,590,693,705,724,768,848,862,867,1005,1016,1078,1125,1152,1206,1237,1245,1271,1395,1494,1535,1540,1801,5302,5424,5465,5549,5574,5642,5750,5915,6035,6045,6087,6091,6127,6177,6356,6392,6430,6517,6518,6609,6906,6976,6985,7108,7123,7204,7243,7311,7319,7361,7406,7439,7445,7539,7669,7731,7735,7737,7901,7940,7944,8006,8020,8071,8221,8368,8378,8565,8664,8774,8864,9025,9078,9110,9364,9594,9856,9899,9917,10004,10079,10162,10214,10369,10384,10500,10612,10677,10690,10821,10867,10871,10908,10916,10983,10998,11781,11836,12348,12426,13157,13925,14057,14161,14401,14476,14776,14872,15312,15677,16096,16452,16512,16931,17133,17630,17646,18670,18749,18936,20520,20556,22241,24016,24073,24262,24702,25044,26313,26762,26787,26837,26863,26973,27201,27397,27681,28109,28176,28844,28949,29244,29296,29508,30599,31241,35707,36138,36197,36269,36371,36453,36459,36757,37003,37186,37260,37373,37449,37697,37727,37758,37822,37931,37948,38256,38682,38999,39377,39419,39489,39943,40792,41239,44374,45167,46684,48595,49966,50079,50233,50416,50486,50519,50575,50646,50854,50865,50908,51217,51369,51407,51608,51642,51652,51989,51995,52161,52177,52187,52231,52250,52478,52541,52780,52798,52809,52890,52914,53079,53138,53162,53199,53217,53291,53316,53835,53877,54019,54072,54162,54331,54339,54416,54515,54521,54740,54875,54886,54959,54978,55017,55030,55033,55055,55149,55153,55253,55272,55422,55618,55686,55709,55721,56056,56136,56145,56263,56289,56317,56375,56391,56479,56595,56785,56800,56812,56825,56848,56907,57035,57170,57238,57319,57375,57406,57519,57605,57659,57673,57801,57820,57952,57965,58043,58102,58176,58297,58326,58431,58567,58652,58676,58934,58952,59316,59405,59413,59457,59713,59723,59733,59816,59836,59918,59929,60009,60033,60236,60246,60266,60497,60502,60555,60576,60583,60620,60626,60770,60786,60954,61034,61164,61427,61443,61533,61598,61604,61613,61677,61750,61752,61779,61830,61992,62143,62174,62307,62482,62570,62658,62713,62813,62839,62891,62972,63025,63131,63226,63351,63628,63639,63687,63784,63947,64225,64498,64513,64636,64729,65036,65148,65187,65222,65228,65255,65285,65389,65396,65445,65580,65640,65676,65678,65701,65722,65845,65910,65941,65966,65980,66057,66171,66234,66302,66329,66336,66359,66410,66436,66494,66531,66538,66551,66767,66817,66854,66875,66921,67126,67363,67467,67470,67478,67479,67575,67758,67767,67779,67788,67823,67866,67872,67882,67947,67991,68027,68143,68221,68244,68247,68328,68341,68447,68480,68626,68669,68695,68711,68819,68987,69158,69386,69553,69691,69710,69948,70052,70110,70236,70536,70596,70763,70775,70819,70821,70865,70928,70974,71195,71331,71338,71387,71560,71594,71602,71610,71653,71695,71717,71896,71903,71933,71981,72161,72168,72177,72297,72343,72359,72561,72633,72682,72754,72782,72825,72872,73041,73132,73172,73322,73338,73447,73582,73661,73720,73834,73839,73845,73866,73929,74031,74058,74108,74163,74441,74766,74768,74841,74872,74969,75016,75062,75110,75149,75158,75313,75398,75417,75514,76213,76266,76388,76407,76430,76461,76520,76527,76558,76695,76758,76805,76831,77139,77200,77278,77285,77291,77427,77478,77548,77603,77737,77815,77983,78050,78240,78501,78644,78650,78882,78949,79038,79233,79257,79397,79410,79412,79426,79448,79464,79486,79503,79660,79668,79687,79704,79713,79728,79804,79853,79856,79947,79950,80028,80043,80057,80160,80443,80525,80576,80702,80833,80853,80856,80867,80908,80930,80931,80948,80998,81015,81047,81127,81217,81264,81460,81502,81632,81671,81710,81743,81746,81765,81872,81875,81958,81971,82034,82063,82066,82253,82343,82347,82487,82516,82521,82592,82618,82764,82808,83040,83102,83202,83335,83378,83409,83424,83518,83526,83710,83850,84004,84124,84306,84335,84401,84526,84557,84716,84765,84785,85249,85374,85378,85425,85634,85670,85699,85756,85855,85986,86056,86083,86147,86203,86237,86505,86528,86570,86597,86684,86711,86713,86796,86890,86981,87026,87449,87532,87557,87565,88037,88043,88072,88134,88166,88328,88357,88392,88552,88585,88619,88622,88712,89001,89173,89252,89425,89630,89632,89635,89806,89905,90101,90137,90440,90703,90716,90748,90770,90827,90850,90929,90944,90964,91231,91239,91269,91303,91407,91519,91639,91650,91682,91760,91773,91860,91909,92003,92006,92033,92066,92139,92263,92275,92420,92766,92788,92795,92839,92947,93050,93063,93232,93242,93292,93317,93330,93345,93380,93421,93484,93703,93834,93998,94086,94111,94239,94363,94401,94579,94591,94636,94838,94852,95102,95121,95133,95189,95197,95343,95430,95607,95692,95759,96025,96313,96488,96567,96669,96670,96795,96901,96917,96972,96994,96996,97066,97198,97226,97256,97395,97402,97530,97557,97620,97633,97669,97717,97718,97796,98025,98132,98182,98246,98308,98452,98667,98698,98823,98908,99034,99186,99292,99444,99469,99767,99831,99834,99907,99936",
      "137,208,219,434,931,977,1158,1182,1272,1431,1641,1732,1765,1799,1815,1829,1839,2047,2076,2359,2564,2690,2748,2798,2857,3042,3221,3367,3421,3500,3613,3655,3702,3795,3804,4127,4171,4185,4266,4383,4550,4572,4575,4741,5029,5073,5096,5156,5163,5219,5262,5312,5326,5372,5488,5515,5524,5533,5571,5747,5809,5816,5887,5901,5940,6101,6167,6293,6383,6400,6528,6551,6564,6578,6634,6702,6940,6945,6947,6992,7013,7047,7163,7308,7350,7412,7546,7575,7649,7710,7719,7796,7814,7861,7909,7954,7995,8143,8164,8216,8229,8289,8396,8414,8492,8624,8674,8756,8760,8788,8821,8831,8842,8931,9235,9277,9464,9581,9599,9730,9794,9865,9916,9933,9983,10058,10068,10126,10230,10276,10322,10325,10341,10530,10550,10607,10624,10719,10733,10765,10813,10834,10944,10986,11025,11082,11143,11232,11323,11365,11416,11557,11560,11660,11697,11838,11906,11937,11956,11990,12114,12146,12278,12335,12578,12739,12811,12891,12911,13137,13171,13276,13380,13407,13427,13557,13584,13665,13799,13842,13860,13943,13990,14198,14329,14402,14460,14763,14803,14804,14806,14894,14956,14982,15014,15214,15224,15254,15280,15413,15436,15505,15728,15736,15868,15905,16039,16095,16211,16212,16401,16801,16818,16937,17050,17118,17266,17589,17602,17635,17812,17814,17883,17920,17928,18457,18471,18539,18542,18672,18785,19138,19156,19288,19427,19528,19634,19659,19698,19736,19896,19934,20087,20124,20183,20270,20497,20580,20714,20744,20822,21149,21174,21470,21526,21678,21688,21748,21849,21873,22138,22185,22233,22246,22286,22310,22314,22425,22482,22595,22597,22657,22659,22862,22976,23019,23032,23033,23076,23298,23547,23621,23729,23824,23833,24097,24128,24245,24287,24496,24593,24724,24825,24844,24874,24894,24912,25083,25142,25333,25521,25617,25692,25783,25901,25989,26217,26232,26340,26491,26710,26788,26806,26877,26975,27083,27091,27146,27236,27327,27439,27448,27461,27545,27666,27745,27754,28092,28187,28203,28311,28348,28364,28416,28465,28467,28617,28685,28766,28860,28903,28932,28942,29017,29151,29319,29430,29462,29598,29834,29933,29966,29978,30005,30034,30092,30177,30266,30293,30903,31367,32004,32180,32325,32527,32615,32702,32729,32816,33367,33717,33948,34122,34152,34500,34528,34558,34645,34818,35147,35208,35583,35685,35819,35878,36079,36139,36202,36450,36465,36502,36514,36691,36932,36944,37011,37053,37071,37073,37077,37383,37395,37483,37506,37675,37704,38043,38062,38122,38126,38142,38329,38330,38577,38603,39071,39126,39262,39623,39635,39680,39731,39750,39857,39878,39930,40669,41097,41184,41234,41358,42117,42261,42405,42506,42709,42982,43080,43170,43219,43403,43404,43434,43497,43693,43888,44167,44526,44558,44670,44720,44807,44933,45067,45233,45320,45617,45671,45697,46154,46338,46382,46979,47017,47139,47732,47815,47936,48102,48142,48242,48991,49042,49224,49803,49836,50030,50074,50140,50283,50337,50445,50462,50479,50513,50530,50563,50584,50590,50651,50838,50891,50892,50932,50956,51088,51111,51161,51280,51391,51474,51500,51502,51508,51587,51624,51640,51646,51779,51808,51868,51955,51983,52053,52055,52238,52419,52422,52440,52475,52540,52611,52711,53001,53076,53108,53133,53159,53181,53303,53384,53564,53700,53815,53859,53883,53923,53924,54012,54038,54100,54143,54288,54289,54385,54397,54406,54428,54489,54529,54551,54757,54865,54871,54910,54923,55016,55046,55076,55390,55553,55636,55771,55914,55941,55990,56020,56061,56147,56160,56346,56458,56490,56541,56810,56883,56895,56900,57027,57106,57129,57172,57262,57324,57344,57367,57489,57513,57687,57710,57743,57744,57848,57959,57990,58057,58171,58219,58305,58318,58442,58472,58549,58554,58565,58852,58864,59061,59193,59258,59317,59374,59462,59498,59563,59634,59652,59754,59807,59852,59958,60253,60384,60485,60556,60589,60654,60823,60890,60899,61094,61203,61214,61240,61315,61321,61346,61374,61594,61759,61978,62012,62034,62139,62320,62338,62471,62725,62773,62902,63093,63141,63229,63266,63295,63366,63531,63665,63737,63803,63868,63906,63928,64035,64104,64159,64347,64548,64605,64622,64683,64697,64703,64766,64850,64949,64962,64993,65027,65080,65094,65113,65124,65167,65277,65613,65633,65704,65708,65770,65943,66005,66069,66192,66209,66491,66578,66692,66768,66776,66910,66928,67006,67088,67323,67367,67390,67450,67547,67549,67576,67582,67680,67773,67793,67855,67939,68195,68259,68265,68285,68365,68510,68612,68681,68890,68977,69042,69054,69172,69282,69410,69564,69706,69716,70208,70245,70478,70629,70776,70791,70885,70917,70977,71076,71257,71358,71761,71778,71840,71851,72162,72282,72379,72534,72556,72689,72904,73225,73252,73261,73294,73353,73672,73788,73904,74010,74101,74139,74303,74322,74377,74431,74633,74696,74779,74790,74938,74963,75004,75011,75019,75156,75248,75312,75319,75331,75361,75439,75474,75794,75798,75831,76080,76210,76229,76296,76325,76660,76690,76932,77022,77032,77297,77418,77426,77431,77525,77741,77879,77967,77976,78004,78011,78180,78273,78382,78494,78558,78642,78745,78829,78879,78881,78898,78899,79004,79122,79159,79384,79729,79749,79788,79842,79897,79945,80085,80111,80135,80153,80260,80353,80373,80477,80553,80589,80590,80840,80862,81005,81165,81180,81184,81266,81363,81429,81548,81557,81713,81747,81756,81866,81959,82143,82155,82229,82242,82292,82332,82648,82792,82967,82989,83028,83083,83091,83169,83174,83243,83261,83279,83283,83345,83373,83450,83466,83537,83554,83872,83892,83893,84127,84275,84333,84396,84518,84549,84583,84601,84723,84762,84770,84808,84892,84900,84915,84920,84942,84955,85115,85261,85335,85506,85610,85759,85805,85911,85936,85977,86090,86097,86124,86128,86195,86200,86233,86239,86305,86326,86432,86457,86661,87429,87438,87451,87462,87573,87601,87751,87957,88028,88355,88473,88496,88579,88649,88704,88804,88851,88872,88906,88922,88975,88976,88982,89037,89069,89073,89117,89200,89224,89255,89267,89381,89745,89902,90038,90067,90268,90536,90566,90826,90875,91015,91129,91170,91198,91219,91320,91361,91429,91452,91543,91603,91704,91850,91889,91921,92125,92181,92200,92227,92267,92556,92836,92943,92956,92971,92981,92984,93006,93008,93101,93138,93150,93173,93289,93313,93332,93351,93467,93473,93495,93626,93629,93687,93793,94031,94037,94043,94107,94118,94505,94567,94618,94649,94856,94891,94896,94962,95043,95222,95303,95360,95476,95550,95554,95779,95844,95922,95969,96138,96231,96295,96330,96502,96855,96925,96995,97137,97246,97442,97529,97547,97579,97644,97691,97736,97755,97767,97779,97888,97954,98007,98039,98169,98256,98260,98272,98332,98424,98433,98531,98564,98659,98672,99071,99129,99214,99262,99314,99379,99462,99478,99610,99717,99734,99747,99799,99857,99866,99922,99957,99972",
      "49,68,125,144,241,2072,2325,2556,2609,2630,2649,2908,3044,3059,3074,3090,3173,3873,3900,4174,4404,4461,4579,4617,4641,4746,4845,6879,7676,9568,9587,9644,9663,9682,9701,9715,9739,9796,9834,9840,10043,10708,10746,11240,11373,11392,11430,11554,11563,11601,11639,11658,11848,11921,11924,11943,12057,12076,12152,12209,12228,12247,12266,12285,12298,12361,12380,12399,12456,12465,12475,12551,12570,12627,12683,12684,12703,12770,12779,12798,12807,12836,12893,12912,12988,12993,13026,13064,13083,13121,13140,13178,13197,13232,13235,13255,13273,13292,13330,13358,13365,13368,13406,13444,13463,13482,13494,13503,13520,13539,13558,13575,13577,13596,13613,13615,13634,13653,13672,13710,13741,13767,13786,13805,13843,13862,13900,13904,13938,13957,13973,13976,13995,14007,14033,14052,14067,14109,14128,14166,14185,14204,14223,14280,14299,14318,14337,14356,14394,14413,14431,14432,14451,14470,14501,14527,14546,14565,14603,14622,14660,14698,14755,14789,14793,14808,14817,14831,14850,14868,14869,14888,14898,14945,14964,15002,15021,15097,15116,15154,15158,15192,15212,15230,15249,15306,15325,15382,15420,15439,15477,15534,15553,15629,15705,15743,15819,15838,15857,15914,15952,16028,16040,16047,16104,16159,16161,16180,16182,16218,16237,16239,16244,16256,16275,16294,16313,16351,16389,16408,16411,16465,16522,16560,16598,16602,16617,16655,16677,16689,16750,16769,16778,16845,16883,16902,16997,17007,17035,17073,17092,17130,17206,17225,17263,17282,17301,17320,17339,17377,17415,17434,17510,17529,17567,17605,17661,17702,17719,17738,17753,17757,17776,17795,17833,17852,17890,17909,17985,17989,18004,18020,18039,18042,18061,18080,18102,18118,18134,18137,18156,18175,18270,18308,18346,18365,18399,18403,18418,18422,18441,18498,18517,18536,18555,18605,18621,18631,18669,18726,18821,18897,18954,18992,18996,19011,19014,19030,19049,19068,19087,19108,19125,19163,19277,19315,19391,19429,19467,19486,19524,19543,19562,19580,19600,19602,19638,19695,19714,19752,19790,19809,19847,19866,19885,19904,19923,19942,19961,19999,20018,20056,20075,20170,20189,20227,20228,20246,20280,20303,20322,20324,20359,20360,20364,20379,20398,20417,20436,20451,20493,20547,20550,20623,20626,20645,20664,20683,20702,20717,20721,20740,20816,20835,20873,20911,20930,20949,20968,21006,21010,21032,21063,21120,21122,21139,21177,21215,21234,21235,21253,21272,21315,21329,21348,21367,21386,21424,21443,21462,21481,21595,21614,21633,21671,21709,21728,21747,21839,21861,21880,21882,21911,21918,21937,21956,21975,21994,22013,22032,22070,22089,22108,22127,22150,22168,22203,22222,22257,22279,22300,22317,22336,22355,22393,22412,22431,22450,22469,22488,22500,22507,22526,22561,22564,22602,22621,22640,22661,22697,22716,22773,22792,22811,22817,22849,22906,22925,22963,22982,22986,23039,23096,23115,23131,23138,23172,23210,23248,23286,23324,23343,23381,23400,23438,23476,23495,23514,23524,23571,23590,23628,23647,23662,23685,23704,23745,23761,23799,23818,23894,23913,23932,23951,23989,24004,24024,24027,24046,24065,24103,24114,24141,24164,24200,24217,24236,24274,24312,24350,24369,24388,24407,24411,24426,24428,24430,24445,24464,24502,24503,24509,24521,24540,24559,24578,24616,24670,24673,24680,24692,24711,24730,24768,24863,24920,24958,25015,25034,25051,25053,25057,25060,25091,25106,25110,25111,25148,25167,25190,25224,25243,25281,25319,25342,25357,25361,25411,25414,25433,25490,25528,25547,25587,25598,25604,25642,25661,25699,25720,25737,25756,25772,25794,25832,25870,25889,25908,25920,25927,25946,25965,26041,26060,26079,26098,26117,26119,26136,26155,26174,26193,26269,26288,26345,26383,26438,26588,26846,26870,26894,26915,26930,26972,27095,27124,27181,27190,27281,27415,27428,27542,27580,27599,27606,27627,27683,27713,27732,27864,27943,27971,27998,28000,28014,28036,28093,28169,28283,28321,28378,28397,28466,28511,28537,28549,28587,28644,28704,28739,28744,28758,28774,28872,28929,28941,28986,29005,29062,29081,29138,29154,29176,29233,29309,29347,29366,29395,29480,29556,29610,29613,29693,29698,29727,29784,29822,29914,29955,30011,30040,30069,30098,30127,30156,30181,30185,30214,30243,30256,30272,30301,30330,30359,30417,30475,30504,30533,30562,30591,30620,30649,30678,30707,30736,30765,30794,30823,30852,30881,30910,30939,30968,30997,31026,31055,31084,31113,31142,31171,31200,31229,31287,31316,31345,31374,31403,31432,31461,31490,31519,31548,31577,31635,31664,31693,31722,31751,31780,31809,31838,31867,31896,31925,31954,31983,32012,32041,32070,32099,32128,32157,32186,32215,32244,32273,32302,32331,32360,32389,32418,32447,32476,32505,32534,32563,32592,32621,32650,32679,32708,32737,32766,32795,32824,32853,32882,32911,32940,32969,32998,33027,33056,33085,33114,33143,33172,33201,33230,33259,33288,33313,33317,33346,33375,33404,33433,33462,33491,33520,33549,33578,33607,33636,33665,33694,33723,33752,33781,33810,33839,33868,33897,33926,33955,33984,34013,34042,34071,34100,34112,34129,34158,34187,34216,34245,34274,34303,34332,34361,34390,34412,34419,34448,34477,34506,34535,34564,34593,34622,34651,34680,34709,34738,34767,34796,34825,34854,34883,34912,34941,34970,34999,35011,35100,35124,35174,35224,35263,35298,35337,35369,35371,35375,35411,35435,35482,35504,35510,35530,35568,35589,35648,35721,35780,35803,35839,35872,35910,35969,35999,36028,36070,36091,36151,36210,36281,36320,36333,36340,36399,36471,36501,36530,36589,36653,36712,36771,36842,36881,36901,36925,36960,36990,36998,36999,37021,37032,37091,37150,37190,37214,37251,37273,37332,37371,37374,37403,37462,37487,37501,37593,37624,37652,37711,37751,37753,37775,37811,37834,37859,37893,37913,37918,37921,37933,37964,38006,38023,38082,38154,38213,38261,38272,38314,38336,38361,38395,38433,38435,38454,38480,38525,38562,38584,38636,38643,38680,38715,38774,38805,38833,38875,38897,38956,39015,39057,39086,39145,39170,39178,39204,39244,39276,39308,39335,39363,39394,39458,39492,39517,39576,39618,39647,39673,39689,39706,39765,39793,39795,39837,39896,39955,39981,40080,40088,40092,40105,40134,40163,40192,40221,40250,40279,40308,40337,40366,40395,40424,40453,40482,40511,40540,40569,40598,40627,40656,40685,40714,40743,40772,40801,40830,40859,40888,40917,40946,40975,41004,41033,41062,41091,41120,41149,41178,41207,41222,41236,41294,41323,41352,41381,41410,41439,41468,41497,41526,41998,42027,42056,42085,42114,42174,42199,42203,42207,42225,42242,42264,42290,42309,42339,42367,42392,42414,42501,42525,42531,42576,42600,42605,42632,42651,42738,42764,42793,42813,42836,42855,42888,42925,42975,43050,43092,43104,43125,43157,43182,43212,43235,43287,43329,43341,43362,43407,43435,43436,43449,43473,43481,43500,43524,43554,43599,43653,43682,43686,43730,43761,43785,43795,43816,43836,43923,43950,43972,43973,43998,44073,44114,44160,44186,44196,44235,44283,44310,44346,44356,44397,44445,44472,44503,44547,44573,44634,44674,44691,44699,44709,44759,44766,44784,44844,44871,44907,44946,44971,45000,45021,45074,45094,45122,45142,45186,45203,45255,45261,45334,45391,45449,45471,45480,45494,45521,45564,45578,45605,45623,45636,45681,45682,45708,45758,45765,45810,45817,45823,45868,45895,45952,45991,45997,46006,46058,46081,46086,46169,46179,46197,46207,46232,46237,46244,46252,46262,46265,46285,46322,46356,46361,46366,46432,46458,46480,46509,46526,46542,46562,46626,46643,46660,46662,46704,46760,46788,46816,46821,46862,46864,46902,46986,47006,47072,47085,47109,47131,47132,47148,47179,47188,47248,47249,47263,47265,47295,47298,47311,47375,47425,47468,47471,47472,47483,47529,47537,47614,47619,47621,47647,47648,47653,47657,47674,47697,47702,47734,47776,47832,47874,47875,47889,47891,47939,47952,47966,47969,47972,48018,48037,48038,48082,48083,48103,48120,48157,48179,48209,48259,48293,48333,48379,48391,48396,48399,48456,48468,48483,48500,48534,48545,48561,48565,48642,48666,48709,48713,48731,48740,48865,48868,48883,48919,48942,48947,48967,49013,49043,49104,49168,49174,49188,49227,49248,49252,49254,49279,49316,49325,49340,49342,49396,49418,49438,49458,49468,49526,49556,49595,49625,49634,49773,49829,49846,49981,50148,50153,50154,50202,50216,50340,50343,50346,50384,50452,50540,50547,50642,50748,50813,50828,50898,50936,51036,51078,51095,51107,51192,51260,51262,51522,51552,51566,51589,51716,51740,51816,51920,52026,52067,52077,52098,52123,52595,53028,53045,53053,53086,53152,53209,53218,53238,53518,53546,53615,53637,53729,53823,53833,53941,53995,54011,54017,54064,54234,54342,54353,54365,54545,54747,54780,54787,54804,54842,54867,54878,55105,55184,55552,55745,55764,55917,55982,56076,56103,56193,56198,56228,56754,56833,57044,57113,57320,57390,57414,57475,57502,57552,57633,57635,57736,57835,57891,58098,58132,58160,58204,58208,58391,58404,58457,58519,58581,58589,58640,58657,58851,58878,58897,58936,58953,59031,59048,59137,59239,59240,59249,59397,59409,59470,59477,59535,59711,59714,59716,60021,60121,60176,60190,60226,60275,60341,60366,60423,60493,60507,60562,60748,60862,60893,60897,60910,60948,61051,61079,61115,61126,61205,61312,61444,61735,61797,61867,61868,61999,62041,62138,62187,62299,62332,62438,62465,62503,62536,62639,62873,62900,62963,63103,63239,63414,63457,63513,63612,63712,63716,63717,63742,63878,64024,64082,64092,64108,64332,64364,64472,64701,64977,65018,65042,65116,65307,65342,65582,65600,65832,66090,66092,66097,66098,66110,66310,66533,66596,66920,66964,67092,67122,67175,67266,67421,67472,67568,67649,67658,67707,67766,67768,67810,67863,68046,68103,68158,68181,68269,68374,68482,68487,68506,68635,68647,68649,68671,68683,68689,68717,68831,69026,69033,69201,69311,69433,69700,69761,69803,69827,69896,69950,69974,70017,70118,70152,70303,70379,70590,70630,70806,70875,70916,71012,71030,71036,71348,71444,71612,71650,71721,71742,71863,71915,72108,72123,72146,72296,72363,72789,72848,73027,73060,73084,73226,73234,73270,73275,73369,73495,73504,73595,73677,73781,73878,74132,74253,74348,74418,74454,74563,74801,74813,74878,74911,74917,74979,74982,75207,75275,75353,75464,75542,75645,75915,75944,76094,76188,76191,76267,76320,76417,76602,76662,76906,76923,76964,76973,77035,77086,77133,77146,77188,77223,77399,77416,77500,77544,77546,77576,77696,77795,77943,77999,78109,78134,78189,78225,78226,78363,78493,78535,78553,78752,78774,78795,79033,79394,79461,79542,79600,79643,79707,79912,80051,80056,80065,80143,80492,80583,80807,80859,80985,81048,81058,81106,81126,81158,81176,81242,81291,81332,81672,81819,81884,81909,82036,82185,82224,82238,82241,82341,82455,82596,82598,82882,82977,83129,83147,83157,83181,83191,83217,83256,83260,83268,83289,83611,83653,83712,83781,83805,83822,83971,83978,84102,84115,84177,84294,84322,84535,84619,84700,84708,84736,84884,84909,84927,84949,84973,85009,85108,85270,85398,85491,85742,85782,85842,86007,86080,86141,86185,86205,86300,86535,86648,86656,86690,86729,86772,86920,86960,87081,87155,87181,87267,87272,87370,87403,87431,87517,87622,87680,87711,87726,87871,87893,87953,88001,88174,88175,88346,88390,88421,88545,88597,88671,88675,88684,88738,88755,88798,88829,88869,89152,89177,89198,89223,89266,89272,89344,89375,89397,89488,89511,89700,89880,89984,90184,90331,90355,90434,90435,90618,90677,90778,91218,91314,91367,91438,91595,91647,91772,91783,91848,91940,92007,92039,92127,92153,92222,92357,92593,92599,92620,92631,92635,92656,92676,92678,92683,92707,92898,92952,93133,93160,93198,93233,93326,93456,93726,93789,93847,93863,93907,93914,93922,93971,94062,94128,94389,94410,94420,94439,94754,94761,94890,94911,94924,94943,94980,95150,95293,95513,95542,95590,95785,95799,95801,95839,95931,95944,95949,96044,96139,96151,96213,96234,96324,96413,96547,96575,96616,96622,96654,96741,96945,96948,97094,97119,97143,97188,97309,97346,97412,97551,97676,97746,97776,97785,97821,98093,98102,98191,98211,98213,98364,98580,98599,98602,98979,99142,99231,99440,99562,99591,99619,99623,99628,99661,99702,99714,99727,99763,99788,99888",
      "70,74,143,175,194,242,1737,1934,1999,2001,2112,2190,2377,2427,2476,2602,2694,2915,3189,3275,3468,3557,3576,3671,3751,3890,3926,3990,4013,4032,4070,4136,4163,4212,4330,4370,4462,4600,4796,5093,6103,6287,6635,6731,6977,7547,7661,8117,8136,8155,8554,8573,8649,8668,8673,8687,8808,8972,9107,9143,9147,9181,9293,9430,9580,9618,9637,9642,9656,9694,9867,9903,9998,10066,10074,10093,10155,10188,10207,10226,10321,10397,10549,10586,10644,10777,10894,10967,10990,11024,11083,11100,11118,11147,11157,11195,11214,11280,11301,11309,11347,11356,11384,11423,11461,11499,11564,11594,11708,11765,11803,11879,11955,12012,12031,12035,12098,12107,12126,12183,12221,12240,12316,12394,12411,12421,12449,12506,12536,12542,12587,12620,12625,12677,12696,12701,12789,12905,12908,12915,12924,13038,13095,13152,13168,13209,13222,13247,13266,13285,13323,13456,13475,13532,13578,13589,13646,13722,13760,13779,13798,13817,13931,13950,13988,14064,14178,14197,14235,14273,14292,14330,14368,14400,14425,14449,14480,14504,14520,14682,14805,14862,14901,14938,14995,15011,15066,15071,15204,15299,15327,15333,15337,15356,15394,15451,15470,15527,15554,15622,15698,15755,15831,15869,15888,15907,15945,15964,16097,16119,16249,16325,16344,16363,16420,16439,16496,16515,16553,16554,16564,16568,16572,16610,16648,16667,16734,16800,16811,16819,16895,16914,16933,16990,17047,17085,17180,17218,17237,17284,17313,17370,17389,17522,17541,17617,17634,17636,17674,17677,17712,17750,17807,17845,17864,17902,17940,17959,17978,18092,18111,18149,18206,18244,18301,18320,18339,18358,18396,18472,18491,18510,18562,18574,18624,18662,18700,18757,18852,18947,19042,19099,19118,19137,19251,19287,19297,19327,19384,19422,19448,19460,19515,19555,19612,19631,19669,19725,19745,19783,19802,19859,19878,19916,19935,19954,19992,20030,20086,20125,20140,20163,20193,20239,20315,20372,20391,20410,20429,20448,20503,20505,20562,20600,20619,20733,20752,20866,20871,20901,20923,20942,21075,21113,21132,21170,21227,21265,21379,21417,21474,21493,21550,21588,21721,21750,21759,21816,21892,22044,22120,22139,22158,22196,22234,22253,22272,22348,22386,22424,22462,22481,22519,22538,22633,22671,22747,22785,22804,22861,22899,22918,23051,23070,23127,23203,23220,23260,23336,23355,23393,23431,23437,23488,23556,23583,23640,23646,23659,23786,23790,23792,23819,23944,24001,24039,24058,24096,24115,24172,24210,24248,24267,24343,24362,24400,24495,24571,24609,24628,24666,24761,24780,24791,24837,24913,24970,25160,25198,25236,25293,25312,25350,25388,25426,25445,25464,25523,25524,25540,25578,25654,25730,25749,25825,25844,25939,25996,26034,26072,26129,26164,26186,26195,26205,26243,26319,26326,26338,26376,26395,26433,26452,26471,26604,26642,26660,26693,26699,26756,26760,26832,26984,27152,27174,27304,27307,27342,27350,27390,27412,27421,27725,27763,27764,27991,28029,28390,28429,28545,28561,28618,28752,29018,29047,29055,29093,29279,29487,29511,29756,29872,29948,29967,30004,30062,30091,30207,30236,30265,30294,30323,30326,30410,30439,30497,30555,30584,30613,30671,30700,30729,30758,30816,30932,30961,31019,31048,31077,31106,31193,31222,31309,31338,31396,31454,31483,31512,31599,31628,31657,31686,31715,31744,31773,31802,31860,31889,31918,31947,31976,32005,32034,32063,32092,32121,32150,32208,32237,32266,32295,32411,32556,32585,32614,32643,32672,32701,32730,32759,32788,32817,32846,32875,32904,32991,33020,33049,33078,33136,33165,33194,33223,33252,33281,33339,33397,33426,33455,33542,33571,33600,33629,33658,33745,33774,33803,33832,33861,33890,33919,33977,34006,34035,34064,34151,34238,34354,34383,34441,34470,34499,34586,34615,34673,34702,34789,34905,34934,34963,35052,35080,35167,35250,35303,35330,35388,35404,35475,35523,35549,35641,35747,35773,35903,35938,35962,36084,36144,36203,36392,36508,36523,36582,36632,36705,36760,36764,36794,36835,36876,36894,36931,36953,36982,37025,37080,37084,37143,37207,37246,37262,37266,37315,37325,37396,37455,37586,37645,37684,37768,37827,37872,37878,37917,37926,37957,38016,38113,38206,38242,38429,38434,38447,38497,38514,38518,38570,38767,38826,38890,38949,39008,39138,39164,39197,39265,39328,39361,39372,39387,39569,39636,39640,39699,39758,39787,39889,40176,40205,40234,40263,40321,40350,40408,40437,40466,40495,40524,40611,40640,40698,40727,40756,40785,40843,40901,40930,41017,41075,41104,41162,41220,41307,41365,41394,41423,41440,41452,41510,41539,42009,42038,42067,42125,42142,42208,42301,42331,42350,42425,42462,42479,42512,42587,42641,42749,42899,42934,42986,43005,43008,43021,43036,43074,43099,43136,43156,43200,43223,43298,43331,43373,43393,43516,43535,43550,43565,43610,43697,43934,44050,44084,44131,44152,44193,44246,44265,44385,44408,44481,44483,44528,44589,44645,44666,44676,44686,44735,44795,44821,44852,45013,45045,45075,45076,45214,45272,45345,45402,45485,45532,45589,45647,45719,45787,45834,45853,45906,45925,45963,46041,46044,46181,46182,46184,46196,46205,46390,46446,46531,46580,46609,46663,46692,46722,46755,46785,46787,46843,46993,47002,47010,47023,47030,47095,47116,47171,47172,47235,47239,47308,47323,47402,47415,47493,47592,47695,47830,47865,47899,47920,47990,47994,48064,48106,48125,48258,48265,48331,48435,48445,48700,48719,48796,48824,48842,49054,49111,49178,49257,49281,49292,49302,49310,49399,49455,49463,49539,49562,49649,49662,49689,49691,49711,49712,49807,49887,50015,50083,50098,50284,50326,50697,50862,50902,50996,51118,51370,51482,51605,51734,51814,51879,51897,52252,52303,52369,52522,52529,52615,52648,52727,52732,52746,52881,53216,53306,53312,53382,53439,53441,53632,53797,54148,54233,54247,54387,54460,54483,54487,54492,54517,54616,54708,54716,54756,54914,54943,55043,55134,55213,55219,55362,55435,55444,55685,55737,55797,55916,55946,55988,56063,56126,56134,56194,56200,56205,56269,56279,56476,56512,56665,56718,56741,56777,57006,57048,57050,57076,57114,57192,57202,57215,57331,57434,57551,57681,57685,57777,58047,58110,58152,58364,58412,58462,58542,58543,58627,58635,58646,58654,58661,58663,58675,58760,58812,59021,59036,59134,59140,59149,59162,59163,59202,59482,59496,59671,59672,59742,59892,59921,60040,60134,60636,60779,60785,60838,60877,60886,60938,61013,61063,61099,61108,61136,61157,61197,61319,61334,61509,61544,61666,61699,61725,61738,61844,61919,62038,62057,62089,62224,62227,62278,62345,62357,62531,62568,62676,62740,62843,63015,63029,63037,63105,63381,63393,63472,63492,63545,63585,63608,63682,63927,63957,63981,63995,63999,64015,64126,64127,64171,64337,64439,64520,64653,64690,64810,64821,64884,64891,64903,64975,64997,65019,65191,65294,65300,65336,65358,65411,65423,65431,65478,65586,65607,65614,65626,65642,65738,65769,66246,66296,66535,66604,66745,66809,66841,66903,66914,66932,66962,67289,67374,67572,67611,67945,67960,67983,68061,68177,68185,68200,68808,69212,69334,69388,69400,69401,69427,69549,69577,69654,69732,69758,69790,69825,69994,70056,70080,70181,70268,70292,70296,70367,70458,70487,70603,70688,70696,70772,70857,70859,70960,71207,71342,71360,71459,71542,71684,71685,71768,71774,71948,72036,72050,72090,72176,72191,72329,72372,72418,72567,72586,72735,72775,72892,72924,73107,73119,73349,73428,73485,73549,73775,73821,73888,73969,73976,74001,74085,74091,74212,74241,74267,74344,74445,74460,74462,74514,74615,74626,74764,74824,74876,75012,75185,75299,75386,75410,75446,75570,75589,75905,75954,76133,76144,76148,76186,76222,76380,76635,76749,76809,76873,77070,77156,77163,77175,77386,77464,77580,77697,77731,77904,78335,78341,78432,78462,78475,78497,78583,78591,78693,79006,79018,79275,79341,79381,79427,79460,79642,79650,79785,79894,79908,79990,80013,80274,80466,80469,80487,80536,80572,80736,81004,81019,81097,81125,81306,81396,81413,81422,82149,82468,82478,82519,82580,82667,82672,82718,82740,82807,82903,82982,83017,83112,83272,83389,83401,83403,83478,83694,83718,83725,83856,83988,84043,84089,84451,84466,84577,84628,84652,84686,84694,85010,85058,85154,85286,85294,85406,85669,85845,85910,85970,86226,86258,86260,86364,86477,86492,86652,86658,86681,86797,86961,87001,87027,87060,87258,87331,87387,87407,87510,87576,87581,87737,87739,87794,87843,87852,88082,88464,88583,88602,88839,88947,88958,88962,89003,89040,89056,89167,89170,89307,89308,89469,89747,89805,89882,89894,90098,90134,90176,90248,90345,90372,90381,90436,90818,90854,90981,90993,91039,91227,91267,91330,91425,91442,91528,91574,91620,91677,91733,91764,91829,91859,91901,91970,92074,92117,92187,92272,92276,92278,92350,92516,92602,92604,92617,92806,92828,92856,92950,92994,93057,93105,93112,93118,93120,93166,93270,93406,93476,93678,93724,93752,93760,93884,93902,94081,94166,94245,94253,94303,94307,94377,94407,94472,94487,94504,94543,94646,94759,94821,94921,94925,94937,94983,95174,95183,95256,95374,95459,95519,95618,95695,95733,95805,95821,95876,96015,96102,96128,96405,96420,96422,96656,96728,96791,96806,96850,96865,96889,96982,97459,97556,97747,97932,97975,97981,97986,98017,98028,98062,98108,98142,98165,98195,98249,98252,98379,98446,98475,98511,98744,98766,98897,99199,99239,99250,99265,99267,99370,99423,99466,99557,99568,99572,99575,99643,99775,99942,99991,99997",
      "126,162,167,1647,1663,1986,2078,2100,2172,2484,2899,2986,3171,3209,3259,3645,3664,3782,3811,4038,4057,4093,4111,4184,4202,4274,4350,4353,4362,4395,4718,4923,4983,6199,6324,6916,8370,9472,9524,9548,9567,9586,9605,9624,9643,9662,9681,9757,9774,9795,9985,10042,10080,10673,10688,10699,10707,10742,10764,10968,11011,11220,11258,11372,11391,11458,11486,11505,11549,11552,11562,11638,11784,11847,11866,11885,11910,11933,11942,11961,11965,11999,12006,12037,12085,12113,12148,12151,12170,12198,12208,12214,12227,12246,12265,12284,12398,12405,12417,12476,12493,12512,12550,12553,12569,12607,12626,12664,12740,12757,12778,12854,12868,12873,12886,12940,12944,12949,12968,12987,13025,13062,13101,13106,13139,13177,13196,13233,13234,13253,13386,13443,13474,13553,13576,13614,13652,13671,13690,13728,13766,13785,13823,13895,13899,13979,13982,14127,14146,14184,14203,14222,14241,14260,14279,14298,14317,14355,14364,14374,14428,14507,14526,14602,14697,14811,14944,15024,15035,15077,15096,15103,15119,15191,15229,15267,15305,15362,15400,15419,15437,15457,15461,15476,15495,15514,15564,15571,15609,15628,15685,15704,15723,15742,15745,15762,15780,15837,15875,15894,15951,15970,15989,16008,16010,16033,16065,16083,16103,16116,16122,16155,16160,16198,16217,16236,16255,16274,16312,16331,16350,16369,16407,16426,16464,16517,16521,16556,16578,16597,16600,16673,16692,16712,16749,16768,16787,16806,16879,16901,16907,16939,16951,16959,17015,17053,17129,17167,17186,17224,17281,17300,17338,17342,17396,17398,17403,17414,17509,17528,17540,17604,17651,17670,17680,17682,17693,17718,17736,17737,17756,17830,17832,17851,17870,17889,17908,17932,17935,17946,17965,18003,18010,18060,18073,18078,18083,18117,18136,18193,18199,18212,18231,18253,18288,18307,18345,18364,18383,18402,18421,18434,18459,18478,18516,18529,18544,18548,18554,18573,18595,18649,18668,18687,18744,18763,18801,18820,18832,18858,18863,18896,18906,18934,18948,18972,18991,19008,19010,19013,19026,19048,19067,19078,19086,19105,19143,19144,19169,19181,19219,19257,19314,19358,19371,19409,19466,19485,19504,19542,19633,19637,19675,19694,19789,19808,19827,19864,19884,19903,19922,19960,19969,19979,20017,20036,20055,20093,20131,20169,20207,20243,20245,20283,20321,20340,20353,20378,20416,20435,20454,20473,20492,20511,20530,20568,20606,20610,20625,20701,20720,20739,20834,20853,20891,20910,20948,20967,21005,21043,21062,21068,21102,21119,21148,21156,21157,21176,21196,21214,21233,21252,21271,21322,21328,21347,21385,21404,21423,21442,21461,21480,21537,21572,21575,21651,21669,21670,21689,21708,21746,21803,21822,21841,21860,21879,21898,21917,21936,21955,21974,21993,22012,22031,22063,22145,22183,22202,22230,22259,22278,22292,22316,22335,22345,22354,22373,22392,22411,22430,22487,22506,22544,22550,22563,22582,22614,22620,22677,22734,22772,22791,22848,22867,22880,22924,22929,22941,22962,22981,22992,23000,23057,23061,23095,23114,23152,23167,23171,23190,23228,23285,23342,23361,23370,23380,23418,23475,23494,23532,23551,23570,23608,23684,23691,23703,23741,23760,23764,23772,23798,23817,23855,23860,23874,23893,23931,23950,23969,23988,24007,24026,24045,24051,24083,24099,24102,24140,24191,24193,24197,24216,24234,24235,24253,24254,24261,24266,24273,24292,24311,24330,24368,24387,24406,24425,24463,24482,24501,24539,24558,24596,24615,24653,24672,24691,24729,24767,24786,24802,24843,24862,24919,24938,24957,24995,25017,25033,25071,25090,25109,25128,25166,25204,25242,25261,25299,25327,25353,25375,25394,25413,25432,25451,25470,25527,25545,25584,25603,25607,25622,25660,25698,25736,25755,25774,25850,25888,25907,25926,25964,26002,26003,26021,26078,26097,26116,26154,26192,26230,26249,26268,26306,26325,26364,26382,26398,26401,26420,26439,26496,26515,26637,26672,26692,26724,26795,26858,26914,26971,26990,27038,27066,27085,27104,27123,27135,27172,27237,27274,27351,27360,27368,27384,27423,27435,27485,27579,27728,27789,27826,27852,27897,27933,27959,27970,28016,28054,28073,28130,28149,28243,28263,28278,28320,28358,28377,28392,28396,28415,28468,28567,28624,28643,28777,28795,28833,28852,28909,28947,28985,29037,29118,29174,29194,29214,29232,29270,29437,29460,29517,29552,29594,29612,29631,29633,29650,29745,29783,29795,29935,29937,30010,30039,30068,30097,30126,30155,30184,30213,30242,30271,30300,30322,30329,30358,30387,30416,30445,30474,30503,30532,30561,30619,30648,30677,30706,30735,30764,30793,30822,30851,30880,30909,30938,30967,30996,31025,31054,31083,31112,31141,31170,31199,31228,31257,31286,31315,31344,31373,31402,31431,31460,31489,31518,31547,31576,31605,31634,31663,31692,31721,31750,31779,31808,31837,31866,31895,31924,31953,31982,32011,32040,32069,32098,32127,32156,32185,32214,32243,32272,32301,32330,32359,32388,32417,32446,32475,32504,32533,32562,32591,32620,32649,32678,32707,32736,32765,32794,32823,32852,32881,32910,32939,32968,32997,33026,33055,33084,33113,33142,33171,33200,33229,33258,33287,33316,33345,33374,33403,33432,33461,33490,33519,33548,33577,33606,33635,33664,33693,33722,33751,33780,33809,33838,33867,33896,33925,33954,33983,34012,34041,34070,34099,34128,34157,34186,34215,34244,34273,34302,34331,34360,34389,34418,34447,34476,34505,34534,34563,34592,34621,34650,34679,34708,34737,34766,34795,34824,34853,34882,34911,34940,34969,34998,35010,35035,35057,35072,35084,35089,35099,35133,35173,35219,35220,35244,35262,35304,35336,35410,35437,35481,35529,35588,35615,35647,35672,35720,35779,35838,35859,35909,35935,35968,36027,36053,36090,36150,36187,36209,36253,36270,36280,36339,36398,36430,36470,36505,36529,36588,36625,36652,36711,36746,36770,36841,36874,36877,36900,36959,37031,37090,37149,37213,37272,37331,37361,37402,37424,37461,37520,37592,37651,37710,37774,37833,37860,37892,37963,38022,38081,38153,38178,38212,38271,38294,38318,38335,38359,38394,38414,38453,38499,38524,38546,38583,38642,38714,38741,38773,38832,38857,38896,38955,38979,39014,39085,39144,39181,39182,39203,39275,39334,39393,39413,39416,39457,39516,39538,39575,39646,39685,39705,39728,39764,39790,39796,39836,39895,39954,40121,40150,40179,40208,40237,40266,40295,40324,40353,40382,40411,40440,40469,40498,40527,40556,40585,40614,40643,40672,40701,40730,40759,40788,40817,40846,40875,40904,40933,40962,40991,41020,41049,41078,41107,41136,41165,41194,41223,41252,41281,41310,41339,41368,41397,41426,41455,41484,41513,41542,42012,42041,42070,42099,42128,42147,42188,42213,42220,42278,42305,42353,42388,42428,42515,42548,42568,42590,42615,42619,42665,42688,42696,42752,42797,42827,42839,42902,42912,42917,42951,42970,42989,43003,43010,43064,43096,43139,43178,43192,43226,43252,43301,43314,43376,43463,43509,43511,43538,43592,43613,43700,43775,43794,43850,43900,43937,44012,44033,44052,44055,44087,44113,44141,44174,44249,44324,44361,44375,44411,44431,44486,44529,44561,44648,44683,44723,44765,44798,44860,44885,44960,44972,45035,45108,45136,45156,45217,45239,45275,45290,45298,45348,45405,45463,45489,45535,45565,45567,45592,45627,45650,45659,45667,45722,45779,45837,45909,45966,45979,46012,46013,46021,46036,46040,46070,46103,46134,46139,46140,46145,46215,46236,46255,46260,46297,46353,46360,46368,46373,46413,46425,46431,46461,46478,46498,46504,46510,46513,46552,46557,46593,46596,46608,46630,46669,46726,46782,46800,46807,46829,46889,46892,46899,46907,46927,46944,46950,46997,47012,47028,47038,47042,47081,47082,47102,47104,47120,47187,47203,47225,47237,47247,47279,47313,47318,47334,47344,47365,47373,47385,47412,47431,47447,47449,47490,47584,47596,47640,47658,47673,47746,47757,47760,47786,47800,47806,47852,47858,47905,47922,47947,47950,47953,48100,48129,48210,48275,48291,48299,48323,48332,48366,48388,48407,48448,48450,48522,48538,48553,48555,48620,48661,48708,48728,48743,48838,48940,48962,48973,48974,48992,49031,49047,49099,49149,49201,49210,49295,49360,49378,49381,49401,49426,49445,49464,49490,49581,49587,49613,49641,49684,49794,49812,49893,49910,49952,50061,50126,50604,50660,50746,50840,50852,51021,51034,51066,51123,51141,51152,51197,51200,51202,51276,51462,51618,51706,51768,51813,51859,52030,52031,52032,52167,52278,52353,52530,52559,52728,52777,52838,52869,52897,52971,53077,53154,53158,53267,53373,53374,53517,53526,53533,53618,53626,53641,53736,53744,53771,53879,54122,54285,54287,54305,54313,54334,54338,54443,54472,54474,54827,54897,55167,55287,55298,55336,55338,55355,55592,55617,55649,55877,55981,56151,56480,56493,56502,56563,56644,56649,56736,56828,56861,56879,56904,56916,56943,56977,57439,57450,57642,57649,57662,57671,57718,57768,57823,57926,58031,58056,58083,58105,58113,58178,58213,58238,58287,58375,58396,58402,58425,58548,58550,58617,58719,58731,58745,58746,58899,58945,59037,59268,59324,59490,59497,59585,59612,59620,59626,59687,59767,59789,59888,60212,60249,60269,60339,60420,60523,60640,60682,60741,60857,60863,60986,61114,61242,61249,61288,61569,61576,61726,61727,61737,61739,61770,61785,61822,61859,61893,61959,61977,62029,62157,62200,62257,62308,62418,62455,62563,62697,62779,62871,63085,63223,63235,63269,63276,63451,63460,63467,63470,63526,63549,63576,63605,63656,63970,63990,64004,64047,64113,64395,64443,64745,64776,64777,65026,65090,65091,65165,65267,65279,65327,65352,65467,65468,65698,65703,65852,65876,66025,66126,66153,66164,66286,66360,66389,66405,66570,66598,66749,66770,66860,66943,66947,66971,67344,67382,67439,67498,67501,67508,67625,67661,67759,67789,67811,67846,67879,67916,68078,68088,68172,68184,68201,68263,68266,68287,68412,68419,68474,68531,68556,68561,68676,68730,68791,68812,68887,68901,68907,68941,69014,69023,69120,69261,69280,69358,69450,69487,69512,69702,69853,69866,69920,70198,70205,70224,70506,70691,70705,70710,70830,70934,71052,71238,71244,71347,71390,71492,71609,71819,71847,71848,71856,72017,72049,72054,72091,72256,72376,72396,72456,72512,72666,72675,72988,73047,73096,73111,73198,73297,73376,73416,73474,73653,73660,73852,73906,74067,74199,74390,74548,75131,75363,75409,75462,75501,75571,75574,75581,75592,75626,75732,75736,75800,75818,75830,75845,75848,75989,76045,76159,76206,76271,76702,76835,77072,77303,77499,77564,77672,77689,77722,77927,77956,78057,78191,78223,78234,78247,78282,78492,78550,78632,78674,78700,78843,78967,79177,79347,79389,79541,79579,79735,79830,79861,79887,79939,80275,80336,80368,80438,80451,80718,80813,80883,80960,80972,81018,81122,81201,81315,81326,81437,81455,81537,81593,81629,81648,81840,81847,82099,82107,82222,82240,82528,82719,82838,82848,82972,83048,83055,83086,83128,83231,83265,83500,83528,83647,83657,83749,83755,83902,83931,83940,84112,84114,84183,84220,84225,84274,84377,84433,84478,84637,84726,85002,85135,85251,85377,85382,85442,85524,85803,85833,86087,86173,86251,86360,86363,86437,86495,86507,86687,86715,87012,87040,87270,87340,87344,87556,87624,87627,87641,87838,87911,87963,87980,88091,88130,88172,88377,88379,88384,88702,88717,88747,88892,88967,89008,89011,89134,89323,89405,89462,89687,89692,89766,89796,89837,89957,90024,90131,90132,90413,90485,90609,90617,90645,90663,90955,90975,91136,91166,91208,91225,91228,91339,91385,91390,91571,91614,91646,91678,91736,91886,91978,91994,92027,92037,92077,92186,92354,92409,92459,92474,92527,93126,93243,93348,93494,93550,93649,93854,93881,93949,93961,94040,94108,94213,94262,94302,94315,94431,94460,94508,94512,94594,94599,95046,95170,95198,95261,95345,95460,95538,95608,95764,95806,95840,95853,95955,95982,95998,95999,96030,96184,96290,96403,96532,96558,96837,96989,97135,97264,97318,97325,97363,97424,97436,97453,97467,97743,97800,97804,97839,97892,97903,97952,98014,98022,98079,98088,98378,98441,98544,98548,98671,98769,98854,98894,98910,98912,98939,98940,99017,99038,99082,99084,99134,99140,99167,99171,99204,99255,99381,99672,99787,99833,99883,99904,99925",
      "18,51,55,99,121,231,272,565,836,1165,1664,1703,1880,1900,1902,1920,1945,1951,1965,1971,2003,2014,2015,2038,2049,2066,2118,2146,2203,2223,2311,2327,2346,2365,2371,2442,2454,2467,2468,2470,2481,2488,2523,2537,2541,2600,2603,2610,2623,2669,2724,2734,2743,2849,2871,2895,2955,2969,2970,2972,2975,2990,2999,3000,3003,3030,3036,3056,3058,3061,3071,3133,3142,3144,3169,3201,3203,3231,3257,3262,3265,3272,3279,3313,3337,3354,3387,3420,3448,3450,3457,3467,3475,3480,3482,3512,3553,3573,3627,3643,3675,3699,3709,3803,3854,3889,3895,3902,3910,3931,3935,3945,3963,3971,3976,3977,3980,3991,4003,4006,4009,4016,4022,4030,4033,4035,4053,4060,4062,4074,4075,4086,4117,4126,4129,4147,4154,4244,4245,4285,4310,4327,4337,4394,4407,4413,4435,4453,4492,4551,4589,4635,4677,4701,4766,4807,4860,4872,4879,4914,4930,4935,4944,4957,4958,4959,4965,4967,4973,5036,5112,5256,5619,5649,5980,6027,6077,6124,6129,6156,6321,6338,6345,6364,6367,6529,6582,6668,6688,6750,6846,6855,6864,6883,7032,7040,7146,7154,7292,7352,7444,7552,7564,7585,7599,7623,7670,7680,7733,7758,7795,7804,7833,7834,7912,7925,7936,7983,8078,8123,8160,8205,8206,8230,8242,8256,8340,8365,8437,8464,8543,9011,9086,9155,9162,9276,9314,9414,9491,9654,9655,9723,9787,9909,10098,10186,10219,10324,10373,10541,10555,10609,10706,10726,10846,10865,10870,10882,10926,11068,11071,11140,11246,11602,11761,11780,11972,12083,12091,12142,12172,12210,12378,12482,12638,12678,12692,12759,12981,13128,13461,13543,13792,14012,14175,14202,14257,14419,14455,14578,14648,14792,14919,14984,15055,15216,15266,15338,15438,15568,15617,15739,15812,15877,15880,15901,15912,16079,16162,16302,16382,16444,16624,16629,16826,17060,17149,17159,17161,17174,17374,17508,17542,17992,18044,18246,18333,18336,18497,18741,19003,19404,19421,19423,19577,19606,19674,19718,19849,19897,20013,20060,20453,20474,20724,20920,21053,21067,21297,21400,21401,21608,21652,21694,21727,21792,21819,21916,22003,22021,22155,22308,22356,22766,22887,23363,23444,23526,23595,23627,23823,23851,23885,23942,23982,23994,24075,24321,24386,24392,24655,24792,24933,25010,25084,25141,25147,25229,25235,25294,25434,25480,25589,25979,25995,26134,26266,26282,26480,26540,26542,26557,26559,26580,26603,26615,26681,26769,26813,26815,26929,26989,27075,27084,27211,27259,27326,27422,27426,27481,27506,27510,27526,27528,27529,27535,27570,27638,27936,27974,27989,28001,28009,28084,28086,28124,28132,28144,28156,28264,28418,28430,28480,28544,28576,28580,28640,28651,28661,28750,29258,29338,29438,29494,29532,29560,29577,29664,29676,29723,29806,29849,29859,29972,30016,30045,30074,30103,30132,30161,30190,30219,30248,30277,30306,30335,30364,30393,30422,30451,30480,30509,30538,30567,30625,30654,30683,30712,30741,30770,30799,30857,30886,30915,30944,30973,31002,31031,31060,31089,31118,31147,31176,31205,31234,31263,31292,31321,31350,31379,31408,31437,31466,31495,31524,31553,31582,31611,31640,31669,31698,31727,31756,31785,31814,31843,31872,31901,31930,31959,31988,32017,32046,32075,32104,32133,32162,32191,32220,32249,32278,32307,32336,32365,32394,32423,32452,32481,32510,32539,32568,32597,32626,32655,32684,32713,32742,32771,32800,32829,32858,32887,32916,32945,32974,33003,33032,33061,33090,33119,33148,33177,33206,33235,33264,33293,33322,33351,33380,33409,33438,33467,33496,33554,33583,33612,33641,33670,33699,33728,33786,33815,33844,33873,33902,33931,33960,33989,34018,34047,34076,34105,34134,34163,34192,34250,34279,34308,34337,34366,34395,34424,34453,34482,34511,34540,34569,34598,34627,34656,34685,34714,34743,34772,34801,34830,34859,34888,34917,34946,34975,35066,35105,35144,35179,35195,35268,35309,35342,35356,35367,35389,35416,35431,35461,35487,35507,35535,35566,35594,35653,35726,35785,35810,35884,35915,35933,35974,35991,36033,36096,36115,36156,36215,36235,36244,36249,36286,36345,36370,36388,36404,36476,36535,36594,36658,36717,36776,36847,36861,36906,36965,37037,37055,37096,37155,37219,37278,37337,37408,37467,37503,37510,37526,37598,37641,37657,37716,37750,37780,37839,37898,37912,37915,37969,37989,38028,38087,38159,38218,38248,38277,38306,38341,38400,38438,38459,38476,38530,38589,38623,38720,38749,38779,38838,38879,38902,38961,39020,39049,39091,39150,39209,39225,39281,39340,39399,39417,39463,39481,39497,39522,39581,39621,39652,39683,39686,39711,39770,39842,39901,39960,39976,40127,40156,40185,40214,40243,40272,40301,40330,40359,40388,40417,40446,40475,40504,40533,40562,40591,40620,40649,40678,40707,40736,40765,40794,40823,40852,40881,40910,40939,40968,40997,41026,41055,41084,41113,41142,41171,41200,41229,41258,41287,41316,41345,41374,41403,41432,41461,41490,41519,41548,42018,42047,42076,42105,42134,42194,42217,42284,42313,42359,42380,42434,42441,42474,42486,42521,42533,42596,42642,42671,42675,42700,42758,42766,42769,42773,42777,42833,42878,42908,42945,42949,42961,42995,43027,43070,43073,43087,43097,43145,43148,43185,43203,43205,43232,43258,43260,43307,43322,43382,43395,43419,43429,43469,43544,43594,43619,43626,43645,43660,43706,43754,43781,43797,43856,43863,43882,43893,43943,44018,44027,44093,44180,44209,44255,44278,44282,44330,44390,44391,44417,44426,44436,44444,44492,44567,44604,44610,44622,44654,44673,44729,44751,44763,44804,44815,44891,44902,44966,44992,44997,45001,45060,45072,45088,45114,45162,45180,45181,45223,45229,45281,45354,45375,45411,45469,45541,45598,45656,45728,45785,45803,45843,45851,45886,45915,45938,45972,46026,46078,46080,46106,46116,46143,46149,46195,46203,46209,46233,46250,46253,46266,46329,46342,46354,46456,46499,46512,46516,46559,46595,46599,46624,46638,46646,46648,46653,46698,46738,46777,46791,46823,46879,46952,47013,47031,47034,47039,47047,47057,47061,47127,47129,47134,47144,47163,47167,47219,47221,47222,47264,47271,47302,47303,47307,47310,47314,47361,47374,47386,47392,47428,47484,47579,47627,47634,47649,47655,47708,47715,47736,47782,47812,47834,47849,47873,47911,47938,47948,47970,47997,48017,48029,48050,48053,48132,48158,48239,48241,48246,48260,48289,48309,48360,48377,48395,48465,48501,48540,48577,48613,48637,48645,48669,48677,48710,48737,48771,48779,48784,48813,48816,48828,48850,48875,48898,48904,48916,48922,48972,49005,49009,49011,49035,49037,49050,49080,49085,49106,49113,49133,49143,49152,49170,49222,49231,49243,49259,49288,49297,49324,49334,49339,49364,49393,49404,49420,49434,49456,49549,49593,49686,49728,49744,49770,49795,49808,49850,49877,49900,49904,49914,49937,49961,49998,50065,50355,50423,50516,50541,50767,50884,50968,50970,50986,51046,51128,51130,51157,51180,51189,51322,51342,51458,51481,51560,51638,51667,51700,51717,51721,51970,52016,52048,52066,52137,52140,52145,52321,52481,52597,52613,52689,52814,53360,53383,53387,53393,53431,53530,53599,53834,53857,53990,54005,54119,54254,54297,54300,54359,54404,54608,54624,54689,54770,54800,54858,54879,55170,55194,55230,55424,55470,55492,55507,55508,55514,55614,55630,55642,55662,55677,55680,55694,55712,55747,55773,55840,55910,56000,56118,56138,56159,56445,56561,56700,56762,56779,56782,57031,57133,57185,57253,57260,57374,57409,57522,57656,57700,57887,57901,57927,58013,58072,58124,58170,58299,58334,58335,58384,58390,58456,58480,58587,58707,58708,58712,58744,58795,58836,58972,58988,59039,59194,59396,59406,59445,59503,59540,59741,59747,59900,60007,60034,60065,60205,60348,60358,60380,60381,60473,60517,60773,60835,61037,61103,61131,61135,61285,61302,61327,61441,61505,61517,61600,61690,61888,61967,62042,62081,62252,62319,62405,62456,62457,62461,62494,62590,62598,62720,62777,62798,62803,62943,62957,63022,63047,63290,63342,63361,63463,63522,63574,63690,63880,63894,63909,63923,63942,64089,64309,64440,64515,64516,64539,64668,64680,64725,64820,64851,64852,64864,64874,64899,64914,64986,64988,65010,65037,65082,65098,65099,65291,65338,65537,65616,65685,65768,65904,66087,66160,66206,66339,66340,66420,66459,66558,66712,66743,66746,66828,66871,66872,66905,67018,67020,67191,67196,67244,67383,67532,67599,67762,67778,67856,68028,68085,68117,68119,68196,68198,68242,68344,68405,68507,68558,68631,68690,68822,68899,68947,68982,69009,69048,69150,69376,69437,69622,69854,69922,70062,70129,70231,70250,70391,70502,70542,70632,71003,71020,71095,71117,71145,71409,71512,71687,71725,71757,71849,71892,71955,71980,72051,72102,72111,72153,72170,72295,72455,72481,72724,72759,72778,72900,72919,72926,73021,73077,73095,73103,73129,73228,73331,73368,73634,73689,74109,74152,74159,74162,74432,74744,74760,74818,74976,74981,74997,75030,75038,75111,75202,75249,75424,75540,75619,75625,75672,75744,75747,75769,75880,75922,75959,76155,76234,76284,76304,76375,76391,76422,76670,76675,76698,76795,76819,76883,76918,76969,76978,77010,77122,77124,77155,77284,77295,77408,77497,77719,77797,77960,78203,78283,78345,78482,78520,78636,78717,78742,78849,78960,78966,78978,79077,79126,79356,79496,79794,79815,79829,79982,80031,80064,80163,80237,80372,80486,80504,80562,80581,80628,80672,80731,80757,81008,81017,81050,81116,81128,81137,81221,81289,81490,81614,81675,81887,81919,81920,81989,82017,82088,82116,82266,82450,82539,82677,82835,82947,82970,83045,83165,83226,83259,83398,83460,83515,83608,83621,83835,83895,83921,84139,84293,84341,84359,84428,84496,84538,84899,84976,85005,85117,85124,85175,85449,85561,85606,85626,85645,85653,85658,85694,85737,85769,85788,85813,85943,85947,85982,85984,86035,86051,86082,86196,86210,86389,86431,86446,86459,86560,86760,86769,86833,86898,87113,87294,87618,87651,87740,87841,87885,87935,88018,88055,88103,88259,88329,88471,88493,88512,88531,88540,88640,88794,88838,88984,89113,89192,89211,89410,89495,89527,89557,89563,89565,89892,89971,90123,90238,90338,90351,90353,90405,90728,90737,90766,90930,91075,91083,91154,91207,91220,91355,91491,91588,91627,91691,91702,91762,91937,92076,92099,92416,92437,92446,92723,92743,92805,92855,93041,93042,93065,93128,93276,93575,93601,93684,93794,93806,93931,94192,94237,94277,94341,94406,94493,94523,94861,94916,94923,94926,94959,95034,95155,95165,95166,95272,95298,95300,95401,95435,95568,95727,95946,96049,96096,96106,96150,96161,96217,96340,96375,96435,96523,96561,96653,96663,96714,96758,97081,97106,97388,97430,97449,97625,97748,97769,97818,97902,98033,98077,98098,98126,98128,98140,98150,98154,98226,98322,98532,98609,98613,98630,98770,99043,99087,99110,99165,99198,99320,99401,99508,99552,99555,99620,99679,99686,99689,99764,99870,99886,99894,99952",
      "47,66,267,321,335,392,431,452,459,462,475,511,551,646,878,946,1075,1086,1140,1187,1391,1427,1612,1630,1642,1710,2656,2755,3330,4609,4976,5041,5162,5491,5509,5525,5608,5622,5666,5692,5695,5858,6072,6125,6256,6263,6295,6302,6469,6610,6676,6856,6928,7045,7236,7334,7377,7397,7550,7694,7703,7775,7939,7993,8111,8144,8153,8192,8359,8373,8524,8636,8695,8699,8702,8803,8847,8867,8933,8947,8974,9007,9128,9168,9169,9213,9219,9281,9353,9467,9853,9886,10023,10045,10084,10135,10286,10399,10406,10457,10492,10547,10584,10611,10635,10693,10873,10892,11007,11012,11045,11069,11259,11302,11330,11733,11740,11808,11813,11934,12094,12119,12178,12407,12720,12799,12927,12980,13051,13052,13086,13092,13141,13414,13508,13537,13618,13748,14088,14370,14633,14778,14795,14936,15019,15027,15157,15162,15180,15324,15435,15627,15909,16018,16539,16567,16579,16595,16659,17018,17246,17663,17733,17813,17952,18049,18094,18232,18627,18725,18772,18783,18895,18921,19153,19353,19363,19732,19926,20076,20089,20722,20796,20900,21036,21088,21277,21283,21405,21697,22304,23006,23234,23264,23387,23456,23479,23585,23779,23836,24145,24304,24349,24926,24951,25136,25174,25489,25515,26210,26327,26396,26458,26555,26668,26884,26941,26943,27013,27030,27385,27781,28172,28353,28374,28423,28837,28865,28894,28996,29190,29452,29492,29600,29700,29848,30864,30946,32253,33525,36431,36434,36635,36681,37018,37237,37319,37880,37994,38302,38552,38609,38614,38815,38882,39114,39300,39382,39690,39880,42524,44140,44986,45243,45444,45499,45601,45603,50168,50183,50206,50237,50246,50330,50542,50759,50836,50843,50937,50990,51001,51102,51350,51501,51529,51565,51654,51686,51699,51758,51759,51819,52090,52344,52348,52523,52547,52570,52625,52707,52760,52797,52841,52895,52927,52999,53063,53115,53120,53142,53309,53425,53510,53521,53568,53766,53826,53841,53897,53931,53936,53958,53978,54033,54065,54097,54184,54550,54591,54605,54717,54760,54801,55126,55129,55187,55208,55289,55371,55373,55416,55452,55610,55661,55878,55950,56081,56224,56231,56273,56368,56393,56503,56504,56691,56699,56704,56815,56937,56972,57397,57485,57504,57644,57648,57678,57800,58055,58073,58149,58242,58362,58545,58634,58665,58689,59022,59178,59331,59512,59514,59591,59638,59755,59827,59831,59930,60005,60019,60028,60074,60292,60480,60481,60545,60659,60668,60727,60800,60887,61110,61120,61123,61304,61325,61379,61420,61634,61957,62166,62373,62400,62458,62485,62552,62583,62790,62918,62967,62980,62991,63016,63054,63142,63173,63197,63243,63388,63394,63450,63480,63556,63582,63610,63625,63640,63752,63772,63795,63967,63974,63984,64005,64007,64050,64265,64300,64407,64410,64427,64463,64617,64660,64678,64710,64750,64796,64875,64934,64947,65060,65172,65236,65260,65375,65543,65792,66086,66114,66221,66284,66290,66328,66397,66462,66509,66620,66672,66716,66795,66810,67072,67145,67219,67273,67318,67364,67401,67481,67555,67569,67596,67688,67742,67973,68224,68254,68342,68382,68443,68643,68802,68902,68997,69126,69230,69232,69309,69636,69826,69918,69931,69964,70584,70606,70829,70842,70979,71041,71269,71407,71468,71553,71562,71651,71744,71929,72194,72218,72357,72385,72461,72617,72709,72813,72863,73140,73150,73263,73279,73365,73450,73506,73509,73538,73597,73640,73670,73739,74030,74041,74126,74151,74231,74260,74395,74402,74520,74565,74591,74720,74919,74973,75055,75165,75218,75246,75290,75322,75598,76092,76139,76360,76563,76652,76663,76674,76734,76782,76796,76944,76988,76993,77107,77244,77339,77510,77512,77574,77666,77721,77771,77799,77816,77860,77876,78000,78043,78044,78045,78164,78173,78235,78379,78452,78517,78533,78544,78575,78691,78716,78793,78924,78935,79088,79296,79349,79476,79558,79732,79789,79832,79846,79905,79937,79965,80038,80172,80423,80685,80754,80756,80772,80839,80855,80977,81063,81367,81511,81692,81704,81774,81811,82000,82175,82193,82206,82251,82380,82397,82400,82529,82552,82554,82686,82729,82749,82797,83078,83149,83352,83414,83445,83454,83471,83486,83681,83750,83811,83868,83989,84107,84363,84370,84381,84592,84647,84727,84756,84759,84800,84879,85150,85165,85253,85310,85360,85368,85408,85576,85643,85860,86121,86127,86354,86992,87032,87116,87145,87164,87319,87447,87476,87555,87699,87742,87756,87780,87793,87878,88040,88111,88229,88238,88315,88326,88341,88397,88466,88472,88478,88745,88806,88953,89241,89304,89409,89459,89669,89686,89702,89734,89818,89867,89901,89919,89928,89989,90039,90202,90253,90270,90396,90521,90594,90600,90707,90842,90843,91007,91021,91077,91188,91261,91350,91368,91457,91738,91807,91858,91932,92219,92428,92435,92450,92451,92479,92526,92720,92739,92887,93077,93092,93202,93203,93235,93236,93283,93320,93372,93526,93596,93650,93660,93692,93780,93825,93862,93973,94074,94092,94179,94248,94436,94446,94570,94704,94763,94831,94848,94965,95164,95210,95270,95290,95379,95420,95481,95571,95706,95794,95835,96010,96169,96204,96310,96369,96459,96463,96537,96635,96767,96813,96820,96896,97080,97131,97312,97339,97374,97431,97582,97602,97722,97826,97880,98004,98010,98037,98114,98205,98223,98432,98610,98618,98775,98784,98827,99004,99039,99061,99139,99153,99164,99368,99378,99600,99687,99759,99916,99928,99962,99978",
      "14,29,89,111,116,150,215,233,310,479,905,961,992,1011,1079,1171,1202,1281,1441,1491,1502,1623,1693,1773,1854,1888,1927,1950,1959,1962,1998,2010,2027,2031,2134,2153,2177,2191,2197,2199,2205,2214,2286,2289,2292,2334,2366,2387,2395,2399,2433,2446,2455,2483,2525,2568,2577,2578,2622,2627,2648,2683,2700,2704,2785,2794,2802,2813,2842,2877,2907,2942,2947,2959,2961,2965,2971,2995,3023,3039,3068,3098,3121,3143,3154,3211,3220,3244,3252,3260,3327,3401,3402,3431,3443,3471,3478,3621,3625,3632,3660,3688,3737,3739,3756,3765,3770,3829,3901,3903,3928,3937,3948,3950,3953,4000,4031,4041,4047,4067,4135,4144,4156,4175,4206,4251,4289,4329,4340,4351,4357,4399,4416,4424,4445,4464,4532,4563,4567,4604,4643,4654,4679,4706,4773,4837,4840,4908,4929,4936,4941,4955,4971,4984,5030,5208,5248,5295,5336,5569,5621,5630,5644,5671,5699,5705,5737,5788,5856,5893,5944,5982,6016,6044,6079,6273,6460,6664,6677,6743,6744,6776,6778,6791,6857,6865,7048,7054,7214,7224,7317,7384,7489,7497,7504,7588,7674,7757,8012,8022,8058,8098,8115,8310,8559,8678,8704,8734,8736,8810,8836,8837,9018,9026,9081,9135,9163,9196,9233,9285,9300,9307,9339,9396,9456,9459,9485,9634,9766,9830,9849,9971,10104,10105,10166,10274,10391,10396,10431,10448,10637,10640,10646,10743,10772,10816,11182,11273,11315,11317,11332,11350,11376,11388,11404,11425,11483,11504,11695,11764,11981,12010,12138,12431,12629,12767,12942,12946,13006,13040,13135,13308,13757,13801,13848,14123,14188,14303,14363,14443,14474,14554,14558,14608,14974,15232,15440,15552,15693,15789,16058,16270,16454,16506,16551,16688,16783,16998,17127,17505,17849,17875,18075,18145,18205,18235,18268,18344,18637,18742,18769,18830,18839,19290,19646,19713,19723,19756,19770,19953,20289,20455,20987,21230,21538,21798,22017,22298,22588,22809,22858,23045,23161,23195,23223,23420,23680,23751,24009,24255,24850,24892,25453,25610,25656,25672,25742,26208,26308,26567,26569,26574,26605,26622,26670,26820,26849,26983,27078,27088,27200,27241,27388,27471,27509,27527,27550,27637,27659,27669,27687,27952,27988,27990,28002,28023,28044,28083,28125,28126,28137,28153,28168,28185,28242,28307,28457,28517,28541,28571,28585,28588,28605,28622,28631,28642,28646,28664,28772,28973,29180,29314,29318,29346,29357,29379,29397,29428,29539,29554,29614,29653,29811,29835,29915,30046,30075,30133,30220,30249,30278,30336,30365,30423,30481,30597,30626,30655,30684,30713,30742,30771,30800,30858,30887,30916,30974,31003,31061,31090,31119,31148,31177,31206,31235,31264,31351,31380,31409,31467,31474,31496,31525,31554,31583,31612,31641,31699,31728,31757,31786,31815,31822,31844,31902,31960,31989,32076,32105,32134,32163,32192,32221,32250,32279,32308,32337,32395,32453,32482,32511,32540,32656,32685,32743,32772,32801,32830,32888,32946,32975,33033,33091,33120,33178,33207,33236,33265,33294,33323,33352,33381,33410,33439,33468,33497,33526,33584,33642,33671,33700,33729,33758,33787,33816,33845,33874,33903,33932,33961,33990,34019,34048,34077,34106,34135,34164,34222,34251,34280,34309,34338,34367,34396,34425,34454,34541,34570,34599,34686,34715,34744,34802,34831,34889,34918,34947,34976,35049,35180,35206,35213,35234,35282,35291,35343,35394,35417,35453,35488,35536,35558,35595,35654,35686,35727,35786,35822,35845,35861,35975,36034,36097,36126,36135,36179,36230,36287,36346,36384,36405,36690,36802,36848,36885,36966,36983,36996,37038,37097,37156,37279,37355,37409,37468,37527,37557,37599,37717,37781,37806,37899,37937,37970,38012,38029,38088,38160,38219,38401,38460,38531,38556,38627,38649,38672,38721,38747,38839,38903,38989,39021,39059,39092,39151,39210,39239,39282,39341,39400,39432,39464,39523,39653,39771,39791,39806,39843,39863,39961,39980,39999,40108,40137,40166,40195,40282,40311,40340,40369,40456,40485,40514,40543,40572,40601,40630,40659,40688,40717,40746,40833,40891,40920,40949,40978,41007,41036,41065,41094,41123,41152,41181,41210,41217,41268,41384,41406,41413,41471,41529,42059,42088,42177,42267,42342,42383,42393,42451,42504,42579,42774,42805,42891,42964,42978,43053,43083,43086,43089,43128,43166,43215,43245,43255,43280,43290,43446,43527,43662,43679,43689,43732,43815,43839,43873,43926,44001,44040,44076,44118,44163,44224,44230,44238,44266,44299,44313,44355,44363,44370,44400,44427,44475,44507,44586,44659,44667,44712,44787,44901,44928,44949,44993,44999,45024,45145,45206,45225,45226,45394,45524,45581,45735,45768,45826,45898,46032,46061,46112,46166,46188,46270,46290,46324,46350,46418,46484,46525,46530,46555,46565,46598,46647,46649,46683,46691,46694,46716,46729,46761,46831,46852,46915,46959,46964,47074,47176,47191,47212,47406,47413,47515,47535,47575,47609,47623,47630,47718,47719,47785,47929,47993,48034,48036,48070,48085,48177,48182,48216,48257,48280,48298,48337,48449,48481,48482,48492,48526,48618,48634,48688,48724,48832,48895,48931,48941,48956,49105,49212,49275,49308,49412,49431,49473,49532,49563,49568,49572,49726,49748,49810,49819,49843,49878,49947,50210,50215,50358,50499,50548,50647,50664,50816,50835,50977,51012,51030,51101,51231,51352,51410,51637,51847,51876,51976,51980,52111,52159,52277,52431,52513,52656,52773,52963,53083,53163,53166,53211,53356,53479,53676,53723,53905,53929,54015,54132,54241,54298,54337,54346,54382,54469,54540,54681,54709,54724,54751,54882,54885,55150,55195,55200,55676,55853,55964,55971,56016,56024,56037,56146,56174,56260,56261,56278,56415,56468,56488,56496,56552,56555,56570,56593,56594,56601,56607,56609,56750,56759,56770,56915,57119,57179,57214,57224,57272,57312,57332,57359,57380,57601,57677,57807,58016,58071,58292,58421,58466,58534,58606,58610,58644,58807,58954,59050,59081,59103,59151,59173,59218,59230,59248,59288,59441,59486,59550,59570,59593,59623,59678,59707,59848,60020,60037,60228,60252,60297,60322,60353,60439,60553,60646,60676,60795,60833,60839,60864,61023,61279,61352,61708,61717,61736,61872,61906,61989,61996,62036,62163,62182,62193,62203,62215,62232,62263,62441,62491,62512,62540,62669,62672,62771,62867,62932,63078,63171,63259,63316,63380,63397,63438,63568,63631,63633,63730,63941,63971,63977,64028,64138,64158,64230,64267,64279,64402,64420,64428,64437,64447,64474,64519,64549,64550,64588,64670,64718,64950,65015,65089,65242,65286,65367,65425,65665,65870,66078,66186,66304,66369,66387,66409,66411,66419,66468,66769,66926,66940,67160,67413,67455,67515,67586,67601,67623,67745,67900,67934,68187,68199,68230,68239,68249,68282,68418,68476,68504,68686,68688,68807,68915,68962,69153,69188,69314,69366,69439,69584,69676,69727,69735,69736,69784,69817,69818,69836,69930,70005,70063,70119,70139,70254,70306,70319,70334,70345,70369,70488,70576,70663,70760,70848,70863,71014,71018,71033,71103,71304,71335,71418,71429,71435,71478,71521,71538,71570,71675,71728,71914,71966,71989,71996,72038,72107,72140,72188,72285,72318,72327,72399,72409,72572,72624,72636,72760,72818,72831,72858,72921,72942,72979,73058,73126,73378,73407,73426,73487,73586,73694,73745,73805,73823,73862,74021,74059,74096,74106,74234,74410,74425,74472,74660,74716,74723,74745,74833,75076,75112,75169,75187,75247,75260,75297,75438,75478,75681,75743,75847,76125,76158,76250,76513,76765,76812,76947,77078,77325,77429,77448,77462,77486,77578,77673,77817,78012,78082,78140,78488,78490,78554,78589,78684,78866,78971,79011,79131,79570,79640,79652,79690,79705,79725,79980,80075,80108,80152,80196,80300,80421,80433,80457,80499,80551,80593,80654,80681,80747,80794,80823,80888,80900,80918,81112,81210,81335,81344,81424,81558,81580,81631,81905,81908,81914,82106,82120,82151,82189,82404,82548,82696,82714,82757,82758,82795,82852,82944,82995,83080,83267,83313,83768,83786,83938,84105,84156,84206,84298,84318,84358,84417,84420,84449,84472,84599,84721,84870,84929,85075,85077,85222,85239,85254,85259,85305,85454,85591,85673,85928,85952,86000,86003,86075,86133,86137,86192,86247,86255,86295,86311,86382,86411,86508,86603,86630,86749,86831,86836,87049,87179,87188,87277,87284,87349,87353,87408,87427,87443,87519,87544,87615,87731,87763,87790,87796,87915,87947,87993,88148,88212,88287,88367,88369,88381,88695,88759,88760,89028,89313,89351,89599,89629,89685,89704,89750,89897,89955,90005,90164,90292,90397,90415,90452,90491,90494,90497,90528,90832,91065,91091,91171,91336,91391,91506,91765,91811,91963,92090,92154,92313,92438,92460,92503,92540,92564,92571,92608,92694,92708,92747,92852,92881,92889,92897,93295,93377,93549,93571,93674,93795,93935,93948,94019,94156,94247,94270,94286,94287,94319,94378,94387,94435,94516,94527,94589,94605,94632,94653,94688,94738,94881,94920,94954,94994,95008,95042,95101,95265,95315,95477,95479,95507,95511,95551,95565,95579,95595,95623,95636,95751,95753,95819,95920,95932,95953,96032,96185,96235,96288,96296,96334,96341,96379,96431,96464,96500,96770,96980,97032,97130,97384,97772,97774,97913,97944,98052,98072,98121,98216,98251,98274,98298,98357,98367,98380,98477,98658,98723,98861,98918,98938,99010,99193,99395,99396,99465,99476,99616,99725,99963",
      "200,263,324,344,383,416,438,500,526,757,813,829,932,1040,1052,1082,1093,1163,1192,1229,1280,1296,1297,1314,1560,1606,1613,1804,1827,1889,1974,1988,2288,2299,2386,2493,2546,2606,2644,2661,2766,2831,2864,2940,2951,3005,3006,3322,3394,3408,3463,3623,3624,3740,3780,3787,3842,3863,3940,4048,4352,4728,4812,4852,4890,5025,5027,5125,5159,5164,5198,5287,5299,5350,5352,5373,5425,5629,5634,5722,5815,5841,5913,5965,5975,6126,6164,6188,6301,6439,6483,6497,6536,6763,6798,6935,7092,7197,7297,7333,7482,7604,7616,7618,7867,7887,7946,7978,8009,8082,8188,8360,8482,8518,8711,8725,8964,8971,9136,9242,9257,9286,9484,9492,9535,9576,9625,9951,9955,10036,10100,10200,10228,10244,10394,10441,10487,10493,10617,10718,10980,11042,11072,11125,11183,11424,11724,11949,12339,12424,12524,12597,12621,12645,12952,13076,13149,13320,13347,13378,13425,13501,13524,13659,13669,13929,13954,14080,14340,14375,14819,14907,14958,14991,15082,15409,15602,15706,15793,15861,15915,15930,16013,16134,16412,16427,16654,16916,17006,17102,17251,17337,17472,17530,17581,17696,17721,17969,18210,18802,18859,18878,18966,19098,19148,19245,19296,19437,19511,19755,19810,19850,19869,20202,20219,20220,20316,20657,20670,20892,20940,20959,21189,21218,22092,22122,22217,22532,22995,23104,23151,23179,23254,23270,23520,23699,23827,24078,24269,24489,24671,24855,24880,24939,25121,25163,25240,25279,25284,25322,25380,25436,25585,25673,25716,25811,25817,26173,26226,26293,26337,26444,26497,26633,26640,26833,26883,26905,26945,27133,27193,27268,27743,27785,27961,28161,28197,28225,28298,28408,28482,28521,28601,28615,28804,28950,29325,29590,29679,29823,29899,29954,30101,30114,30169,30188,30198,30507,30536,30652,30884,31029,32903,32943,33088,33391,33407,33581,33657,33802,35226,35312,35391,35592,35783,35842,35939,36008,36044,36315,36316,36454,36551,36917,36921,36936,37007,37072,37131,37137,37181,37242,37389,37465,37544,37619,37695,37877,37929,37991,38057,38240,38243,38252,38311,38416,38440,38568,38612,38766,38811,38866,39192,39318,39375,39500,39562,39616,39625,39629,39730,39921,39937,40270,40304,40480,40596,40647,41227,41261,42025,42111,42282,42706,42734,42863,43085,43123,43479,43786,44026,44344,44706,45177,45317,45687,45746,45798,45974,46714,46748,46871,47664,47843,47968,48012,48347,48621,48712,48878,49008,49272,49470,49820,50040,50056,50186,50278,50537,50679,50820,50885,50971,51177,51204,51230,51363,51520,51604,51616,51668,51714,51880,51881,52062,52071,52126,52142,52307,52428,52536,52550,52560,52571,52609,52631,52639,52682,52691,52709,52771,52785,52861,52879,53080,53192,53242,53475,53476,53487,53538,53540,53601,53609,53674,53675,53746,53756,53832,53867,53876,53956,54034,54104,54191,54279,54283,54351,54533,54572,54768,54775,54810,54846,54956,55031,55092,55217,55451,55504,55539,55541,55717,55906,56026,56028,56055,56082,56158,56173,56237,56325,56387,56417,56432,56527,56637,56689,56721,56831,56849,56982,57065,57195,57218,57242,57303,57526,57537,57625,57692,57808,57864,57876,57919,57922,57957,57984,58017,58125,58172,58324,58369,58436,58459,58465,58491,58516,58774,58820,59018,59051,59120,59246,59263,59328,59387,59404,59489,59494,59504,59533,59547,59554,59565,59617,59650,59684,59791,59830,59919,59938,59946,60075,60137,60139,60148,60277,60371,60405,60410,60630,60721,60790,60803,60844,61180,61184,61187,61206,61749,61948,62007,62018,62168,62183,62214,62228,62276,62303,62327,62352,62383,62393,62445,62529,62836,62920,62965,62971,62976,63283,63337,63462,63482,63709,63740,63827,63885,63922,63991,64046,64052,64219,64274,64433,64547,64812,64897,65031,65063,65133,65143,65350,65355,65378,65391,65604,65697,65745,65857,66178,66181,66576,66629,66687,66798,66864,66969,67009,67107,67167,67291,67294,67295,67483,67546,67584,67594,67683,67685,67782,67835,67913,67943,67974,67986,68025,68326,68350,68395,68472,68549,68678,68799,68865,69034,69037,69163,69192,69296,69398,69443,69985,70006,70024,70186,70239,70384,70501,70715,70793,70837,70858,70968,71046,71068,71278,71327,71593,71985,72025,72068,72098,72163,72223,72224,72490,72632,72667,72707,72723,72758,72795,72807,72808,72868,72964,73018,73026,73040,73059,73072,73147,73246,73260,73282,73309,73340,73601,73627,73686,73970,74060,74122,74177,74294,74301,74347,74366,74434,74636,74712,74731,74845,74999,75335,75640,75643,75692,75763,75788,75816,75930,76119,76216,76254,76294,76354,76365,76539,76671,76718,76735,76738,76740,76752,76887,76935,76972,77052,77093,77379,77536,77569,77668,77783,77935,77937,77984,78151,78153,78163,78194,78213,78243,78344,78400,78448,78476,78574,78602,78657,78739,78743,78784,78788,78846,78856,78962,78990,79079,79226,79265,79367,79473,79573,79734,79810,79966,79983,80092,80177,80254,80291,80311,80317,80364,80388,80520,80522,80582,80586,80605,80651,80734,80744,80939,80990,81034,81140,81157,81187,81314,81327,81532,81607,81628,81776,81842,81893,81898,81918,82047,82135,82269,82277,82467,82550,82569,82575,82591,82683,82765,82782,82928,82949,83079,83118,83192,83496,83635,84175,84266,84385,84389,84588,84745,84975,84977,84979,85042,85054,85055,85118,85145,85386,85440,85556,85605,85655,85696,85732,85908,85925,85956,86034,86077,86293,86317,86381,86598,86739,86851,86979,87022,87163,87292,87333,87397,87457,87468,87515,87562,87567,87606,87616,87639,87643,87710,87820,87824,88007,88127,88292,88368,88425,88494,88581,88681,88754,88771,88787,88834,88860,88945,89063,89087,89093,89131,89159,89232,89236,89284,89287,89471,89507,89521,89541,89572,89628,89657,89690,89763,89871,89898,89927,90183,90273,90416,90534,90644,90649,90700,90751,90896,90951,90991,91149,91280,91294,91332,91369,91445,91487,91751,92047,92241,92299,92342,92472,92495,92727,92817,92846,92961,93035,93060,93169,93311,93463,93930,94203,94396,94465,94468,94502,94554,94580,94671,94713,94737,94818,94866,94927,95025,95026,95035,95195,95304,95366,95440,95471,95473,95533,95655,95739,95766,95769,95878,95890,95921,95935,95958,96072,96076,96095,96125,96149,96156,96200,96224,96237,96300,96426,96430,96458,96640,96688,96692,96821,96967,97077,97162,97196,97356,97545,97635,97656,97682,97695,97793,97828,97850,97905,98021,98064,98078,98101,98118,98133,98143,98164,98292,98326,98502,98506,98893,98989,99141,99163,99203,99349,99530,99633,99768,99954,99977",
      "31,354,376,377,381,387,422,602,650,683,735,747,807,849,889,952,979,994,1004,1024,1048,1081,1112,1142,1194,1227,1232,1243,1263,1264,1371,1419,1504,1677,1772,1878,1912,3073,3692,5022,5037,5047,5069,5081,5099,5107,5111,5176,5315,5332,5394,5432,5459,5480,5486,5577,5582,5592,5595,5615,5675,5784,5798,5852,5868,5888,5910,5994,5998,6014,6069,6143,6210,6219,6231,6244,6270,6289,6335,6354,6372,6407,6543,6600,6604,6628,6649,6699,6779,7019,7026,7036,7051,7410,7417,7488,7625,7667,7729,7915,8016,8038,8063,8441,8445,8541,8618,8634,8641,8661,8789,8839,8899,8903,9008,9109,9113,9130,9139,9215,9225,9246,9289,9946,9947,10056,10227,10499,10625,10838,10935,11114,11768,11790,12158,12525,14214,14878,15716,16449,16984,17082,17709,18176,18848,19329,20111,20256,20576,21455,23035,23313,23483,24353,25077,25329,25354,25487,26465,26709,27221,27538,28900,29163,35896,36310,36432,36639,37638,38296,38703,39825,42556,43098,44473,45032,45173,46281,46476,50018,50062,50089,50136,50239,50359,50430,50447,50465,50488,50524,50603,50686,50880,50909,50945,51019,51156,51218,51249,51336,51371,51445,51581,51599,51660,51726,51810,51865,52037,52054,52100,52343,52450,52495,52629,52664,52784,53153,53200,53202,53372,53497,53501,53554,53621,53659,53730,53782,53871,53874,54062,54073,54123,54167,54358,54360,54366,54425,54430,54447,54488,54513,54546,54586,54752,54883,55063,55154,55159,55209,55250,55290,55350,55364,55383,55431,55577,55582,55678,55696,55784,55814,55969,55989,56115,56119,56386,56449,56508,56652,56912,56971,57057,57071,57099,57117,57254,57461,57561,57564,57646,57737,57830,57900,57937,57996,58058,58084,58264,58312,58322,58333,58386,58410,58441,58443,58452,58539,58585,58605,58873,58884,58922,59045,59129,59252,59330,59333,59416,59421,59506,59571,59829,59860,59911,60147,60203,60316,60390,60450,60515,60570,60642,60647,60657,60766,60783,60818,60826,60837,61020,61097,61162,61255,61369,61431,61464,61482,61542,61585,61679,62009,62073,62091,62159,62188,62218,62305,62398,62473,62566,62786,62915,62921,62927,63101,63104,63284,63285,63298,63321,63733,63774,63811,63817,63899,64030,64325,64576,64618,64649,64715,64735,64754,64767,64811,64824,64920,64990,65128,65161,65257,65401,65554,65746,65750,65790,65794,65797,65836,66033,66061,66093,66132,66156,66169,66222,66327,66421,66464,66553,66575,66615,66697,66733,66740,66771,66819,66833,66886,66891,66913,66996,67025,67080,67117,67166,67221,67436,67589,67591,67647,67726,67732,67752,67761,67887,67892,68034,68084,68217,68228,68528,68630,68636,68725,68766,68942,69003,69100,69136,69162,69211,69239,69316,69488,69546,69763,69824,69857,69859,69893,70043,70058,70116,70156,70183,70202,70394,70704,70752,70795,70871,70877,70942,70980,71064,71285,71334,71413,71565,71575,71595,72001,72075,72120,72131,72246,72319,72338,72356,72429,72476,72822,72851,72895,73213,73230,73351,73401,73436,73437,73516,73548,73676,73682,73702,73733,73795,73816,73827,74090,74105,74247,74352,74364,74406,74553,74566,74580,74650,74762,74796,74865,74907,74962,75014,75252,75266,75280,75286,75306,75375,75465,75623,75677,75737,75981,75994,76232,76303,76312,76412,76565,76589,76700,76855,76882,77117,77257,77424,77516,77565,77589,77619,77686,77772,77896,77933,77971,78093,78193,78199,78329,78368,78443,78499,78548,78593,78671,78825,78830,78915,78932,78997,79005,79035,79138,79142,79162,79231,79993,80035,80137,80209,80337,80377,80465,80470,80478,80653,80677,80824,80845,81037,81039,81166,81171,81348,81377,81438,81560,81646,81785,81793,81806,81860,81902,81954,81961,81975,81982,82136,82205,82226,82268,82522,82547,82568,82590,82726,82734,82779,82796,82864,82969,83239,83375,83390,83504,83560,83607,83695,83756,83796,83851,84204,84247,84394,84452,84474,84609,84728,85149,85365,85509,85547,85612,85708,85715,85779,85796,85880,85895,85924,86053,86114,86232,86319,86334,86342,86379,86417,86587,86634,86761,86862,86871,86915,86995,87310,87549,87734,87905,87931,87933,87946,87948,88027,88213,88354,88439,88559,88584,88625,88680,88880,89019,89068,89139,89151,89221,89247,89258,89281,89737,89889,89940,89964,89987,90032,90033,90111,90112,90154,90285,90454,90459,90614,90667,90893,90903,91104,91121,91201,91406,91533,91715,91722,92131,92169,92182,92322,92349,92362,92364,92453,92497,92690,92794,92915,93048,93073,93335,93419,93498,93612,93774,93792,93827,93975,93990,94046,94049,94227,94278,94301,94345,94438,94458,94494,94503,94578,94608,94640,94670,94709,94721,94729,95030,95326,95333,95593,95704,95717,95732,95910,96003,96135,96179,96221,96453,96489,96503,96573,96831,96842,96897,96963,96970,97022,97075,97128,97262,97447,97482,97534,97636,97647,97724,97734,97760,97792,97807,97919,97976,98035,98103,98131,98293,98305,98311,98416,98542,98562,98650,98668,98715,98718,98774,98819,98895,99056,99083,99098,99273,99376,99380,99394,99471,99581,99636,99663,99735,99778",
      "91,1365,1716,1717,1779,1873,1979,2048,2211,2240,2319,2368,2458,3029,3110,3243,3328,3442,3537,3723,3955,4050,4069,4419,4440,4489,4810,4873,5044,6145,6596,6615,6653,6963,6990,7128,7850,8053,8059,8135,8154,8314,8333,8553,8648,8762,8778,8917,9060,9117,9161,9180,9237,9308,9316,9406,9493,9503,9510,9522,9560,9598,9617,9693,9750,9788,9823,9845,9883,9978,9997,10035,10111,10130,10149,10168,10187,10191,10206,10225,10292,10339,10605,10738,10833,10890,10893,10909,10928,10966,11023,11137,11156,11172,11175,11213,11251,11265,11296,11308,11327,11403,11422,11517,11574,11611,11612,11726,11897,12068,12087,12144,12163,12239,12258,12391,12543,12619,12700,12721,12733,12746,12825,12828,12904,12926,12955,13056,13132,13151,13185,13189,13227,13246,13262,13272,13284,13303,13360,13379,13455,13493,13502,13588,13735,13759,13816,13835,13873,13886,13913,13930,13949,13974,14044,14101,14158,14177,14215,14253,14272,14291,14310,14386,14456,14538,14559,14671,14725,14728,14734,14842,14899,14975,15032,15033,15108,15127,15146,15211,15222,15241,15279,15317,15393,15450,15469,15507,15559,15621,15659,15678,15688,15697,15798,15811,15849,15944,15963,15982,16020,16077,16109,16210,16229,16305,16324,16343,16362,16495,16514,16586,16628,16647,16678,16713,16742,16875,16894,16932,16970,16989,17116,17122,17198,17217,17255,17261,17331,17350,17354,17464,17502,17559,17628,17673,17825,17863,17882,17939,17980,18016,18072,18091,18129,18186,18224,18243,18281,18319,18338,18395,18433,18452,18490,18509,18585,18635,18661,18737,18813,18825,18851,18927,18946,19033,19117,19155,19172,19174,19212,19231,19345,19383,19402,19417,19459,19478,19497,19500,19513,19516,19592,19613,19616,19630,19649,19668,19726,19782,19820,19839,19915,19974,19991,20022,20143,20276,20295,20302,20314,20355,20371,20409,20447,20466,20485,20504,20561,20567,20637,20656,20694,20770,20867,20884,20917,21107,21112,21150,21262,21302,21435,21473,21559,21568,21625,21720,21739,21834,21891,21910,21967,21986,22024,22043,22100,22176,22214,22323,22385,22404,22442,22461,22480,22518,22537,22556,22568,22583,22594,22670,22765,22784,22803,22841,22875,22917,22936,22947,22955,23050,23069,23145,23175,23183,23202,23259,23278,23316,23354,23487,23525,23620,23639,23658,23715,23734,23791,23867,23886,23924,23962,23981,24019,24076,24095,24133,24178,24190,24263,24285,24342,24354,24361,24380,24399,24418,24437,24456,24513,24570,24598,24684,24703,24722,24798,24836,24893,24969,25026,25102,25159,25197,25273,25297,25349,25368,25444,25463,25501,25558,25577,25596,25615,25691,25729,25805,25824,25843,25919,25938,26014,26033,26106,26109,26147,26151,26223,26261,26280,26299,26414,26463,26470,26527,26641,26717,26845,27154,27157,27187,27325,27382,27406,27629,27844,27876,27935,28199,28351,28522,28731,28826,28921,29020,29155,29269,29288,29301,29377,29461,29768,29818,30061,30119,30148,30235,30351,30380,30438,30525,30583,30670,30699,30728,30757,30786,30844,30902,30931,30989,31105,31134,31192,31221,31279,31308,31395,31482,31511,31540,31569,31627,31685,31714,31772,31801,31830,31859,31888,31917,31946,31975,32033,32062,32120,32236,32294,32323,32352,32410,32468,32555,32613,32642,32671,32700,32758,32874,32932,33019,33106,33135,33164,33193,33222,33280,33338,33425,33483,33512,33541,33570,33686,33715,33773,33831,33889,33918,33947,34005,34034,34063,34092,34150,34179,34237,34266,34382,34411,34440,34527,34556,34585,34643,34672,34701,34730,34788,34817,34875,34962,35003,35087,35132,35210,35218,35255,35325,35329,35396,35403,35432,35451,35469,35474,35506,35522,35581,35667,35678,35772,35831,35877,35961,35994,36083,36143,36176,36325,36332,36391,36429,36519,36522,36581,36610,36675,36704,36715,36735,36763,36834,36893,37083,37142,37206,37372,37451,37502,37513,37585,37644,37703,37767,37826,37885,37956,38146,38205,38264,38328,38370,38446,38517,38635,38707,38754,38889,38948,39078,39106,39137,39196,39268,39327,39368,39386,39450,39639,39676,39698,39726,39757,39785,39797,39829,39856,39859,39947,40201,40259,40288,40346,40404,40433,40462,40520,40549,40578,40607,40665,40694,40723,40810,40839,40868,40897,40926,40955,41013,41042,41100,41129,41158,41187,41332,41361,41390,41419,41448,41477,41506,41535,42121,42214,42256,42346,42366,42421,42444,42447,42508,42547,42566,42583,42658,42704,42787,42820,42938,42969,43038,43057,43106,43132,43159,43262,43272,43294,43369,43531,43606,43632,43647,43656,43657,43714,43739,43901,43911,43930,43966,44005,44030,44034,44048,44080,44185,44242,44269,44388,44450,44479,44534,44554,44716,44736,44791,44813,44819,44828,44846,44848,44878,44911,44934,44953,45028,45101,45149,45210,45268,45309,45341,45380,45418,45585,45658,45715,45772,45830,45860,45959,46035,46053,46084,46128,46199,46384,46605,46680,46702,46725,46740,46756,46841,46856,46874,46912,46958,47009,47027,47068,47075,47107,47125,47145,47146,47152,47170,47210,47230,47296,47304,47347,47456,47460,47526,47536,47593,47603,47629,47707,47795,47985,47986,48003,48011,48052,48071,48117,48173,48219,48236,48250,48266,48316,48338,48358,48598,48611,48625,48827,48924,48948,48961,49015,49025,49179,49190,49270,49387,49400,49503,49511,49524,49531,49553,49575,49676,49696,49797,49969,50164,50223,50268,50475,50567,50722,50737,50778,50929,51228,51254,51283,51760,51793,52044,52080,52125,52258,52299,52336,52395,52470,52584,52642,52883,53180,53187,53213,53258,53260,53414,53472,53496,53690,53865,53999,54090,54095,54111,54214,54330,54344,54623,54814,54975,54992,55071,55106,55119,55367,55689,55732,55758,55828,55873,55923,56005,56318,56418,56632,56641,56685,56766,57148,57169,57322,57376,57470,57557,57653,57660,57727,57822,57960,58040,58050,58096,58199,58317,58372,58509,58515,58572,58666,58705,58748,58844,58896,58946,59059,59108,59168,59199,59225,59245,59267,59354,59541,59702,59803,60031,60101,60394,60441,60458,60637,60757,60982,60989,61052,61072,61179,61286,61287,61318,61609,61626,61778,62002,62066,62070,62106,62279,62346,62636,62670,62772,62828,62937,62949,62993,63202,63225,63228,63301,63305,63469,63484,63559,63749,63788,63830,63875,63888,63900,64034,64142,64155,64220,64257,64422,64464,64477,64558,64584,64624,64646,64723,64790,65071,65075,65154,65189,65227,65347,65495,65540,65662,65690,65758,65773,65783,65796,65826,66027,66043,66130,66191,66361,66401,66451,66520,66547,66633,66710,66912,66919,67002,67016,67076,67093,67119,67339,67347,67426,67471,67602,67848,67862,67883,67889,67956,68210,68233,68305,68346,68475,68526,68573,68625,68658,68661,68693,68905,68954,69250,69292,69431,69596,69598,69681,69712,69751,69762,69787,69788,69863,69928,69961,69962,69983,70011,70091,70099,70226,70242,70243,70311,70338,70532,70545,70746,70747,70825,70921,70983,71027,71097,71109,71162,71176,71222,71408,71445,71683,71910,71986,72074,72151,72244,72333,72434,72553,72563,72799,72849,72883,72923,72945,72948,73052,73173,73192,73334,73409,73515,73536,73585,73592,73671,73685,73872,73908,73977,74068,74078,74087,74197,74202,74306,74422,74439,74502,74656,74719,74741,74933,75127,75229,75381,75425,75476,75508,75546,75551,76023,76099,76135,76190,76215,76237,76256,76351,76512,76531,76598,76610,76701,76745,76772,76773,76838,76929,76945,77058,77084,77258,77290,77375,77482,77483,77490,77531,77756,77931,78001,78066,78120,78128,78177,78207,78224,78295,78405,78547,78561,78933,79161,79209,79279,79285,79372,79481,79561,79625,79848,80041,80181,80193,80222,80449,80479,80516,80769,80821,80863,80973,81273,81439,81441,81576,81590,81804,82072,82128,82156,82249,82300,82321,82361,82416,82438,82444,82518,82578,82614,82689,82818,82825,82826,82862,82887,83162,83194,83201,83408,83561,83624,83654,83687,83855,83875,83891,83967,84180,84250,84379,84447,84448,84514,84763,84771,84820,84874,85081,85090,85244,85275,85359,85380,85395,85537,85558,85642,85649,85695,85723,85750,85834,85878,86241,86262,86367,86451,86611,86612,86629,86740,86744,86964,86993,87034,87189,87203,87288,87302,87374,87713,87725,87837,87936,88010,88096,88120,88290,88308,88361,88469,88744,88807,88861,88894,88990,89015,89024,89086,89104,89141,89155,89185,89218,89261,89263,89292,89487,89499,89595,89603,89611,89640,89903,89918,89983,90103,90159,90191,90208,90216,90258,90316,90448,90552,90631,90632,90634,90670,90734,90773,90846,90868,90901,90927,91116,91212,91224,91315,91899,91908,91945,92058,92082,92123,92160,92170,92269,92326,92361,92365,92432,92514,92532,92569,92613,92636,92734,92749,92833,92871,93015,93036,93115,93139,93142,93149,93412,93444,93464,93497,93502,93690,93707,93873,93891,93905,93926,94143,94208,94282,94397,94467,94521,94525,94827,94878,94949,94975,94987,95139,95168,95259,95386,95685,95730,95864,95879,95895,96031,96045,96325,96412,96529,96602,96844,97111,97147,97169,97293,97337,97480,97552,97628,97812,97982,98206,98440,98550,98591,98746,98761,98844,99160,99328,99454,99532,99554,99569,99574,99736,99843",
            "32,127,146,1655,1659,1685,1712,1747,1780,1784,1932,2312,2318,2352,2391,2598,2676,2705,2733,3069,3172,3205,3261,3509,3646,3965,3984,4192,4484,4803,6096,6587,6625,7168,7544,7926,8126,8145,8183,8563,8639,8677,8696,8804,8872,9036,9095,9152,9247,9266,9378,9418,9421,9513,9532,9570,9627,9646,9665,9684,9703,9722,9741,9798,9803,9921,9950,10064,10106,10216,10232,10235,10331,10471,10473,10577,10627,10634,10653,10779,10843,10859,10900,11026,11052,11109,11185,11261,11360,11432,11470,11489,11535,11622,11641,11736,11755,11850,11851,11907,11976,12059,12078,12135,12287,12367,12486,12495,12496,12546,12572,12667,12705,12754,12781,12844,12895,12914,12964,13142,13237,13342,13465,13467,13484,13522,13528,13540,13598,13636,13655,13674,13701,13712,13731,13750,13769,13807,13828,13914,14031,14056,14073,14111,14130,14135,14168,14187,14282,14301,14320,14415,14567,14586,14641,14738,14757,14814,14947,15004,15006,15042,15137,15194,15202,15251,15270,15289,15343,15346,15365,15405,15473,15517,15533,15536,15631,15634,15648,15707,15708,15726,15821,15859,15878,15897,15954,16068,16106,16163,16201,16220,16258,16277,16296,16334,16353,16391,16448,16467,16524,16562,16657,16771,16790,16807,16809,16862,16885,16920,16961,16972,17005,17026,17056,17094,17100,17134,17151,17170,17235,17303,17341,17360,17379,17417,17421,17484,17486,17531,17626,17639,17652,17740,17759,17816,17873,17892,17930,17968,18006,18063,18082,18101,18158,18196,18367,18386,18405,18439,18443,18449,18481,18547,18557,18614,18633,18655,18720,18727,18823,18861,18880,18918,18937,18967,19012,19051,19089,19127,19203,19239,19247,19298,19317,19374,19393,19412,19469,19507,19545,19564,19621,19697,19701,19716,19773,19811,19830,19868,19906,19925,19963,19973,20001,20020,20077,20172,20210,20250,20267,20286,20305,20343,20362,20381,20400,20419,20457,20476,20495,20609,20666,20780,20837,20856,20888,20970,21027,21084,21179,21217,21236,21255,21331,21388,21407,21422,21426,21445,21483,21521,21578,21597,21635,21653,21654,21730,21749,21787,21801,21806,21844,21856,21863,21901,21939,21958,21977,21996,22015,22072,22091,22129,22195,22205,22224,22243,22262,22357,22376,22414,22433,22471,22509,22528,22547,22604,22680,22724,22737,22775,22813,22851,22965,22970,23022,23063,23117,23155,23231,23327,23440,23478,23516,23535,23548,23573,23687,23725,23739,23811,23839,23849,23868,23877,23915,23972,24010,24029,24086,24181,24219,24257,24291,24352,24409,24446,24447,24485,24504,24523,24542,24561,24580,24618,24675,24713,24740,24770,24865,24891,24998,25055,25131,25150,25226,25245,25264,25283,25301,25397,25401,25435,25457,25511,25568,25644,25682,25738,25815,25856,25897,25909,25948,25967,25986,26024,26081,26100,26138,26214,26252,26262,26290,26300,26309,26328,26371,26385,26390,26423,26461,26537,26594,26643,26707,26860,27045,27069,27107,27143,27254,27278,27320,27354,27386,27544,27597,27718,27734,27753,27843,27981,28133,28226,28228,28233,28304,28442,28494,28551,28738,28760,28867,28893,29026,29043,29197,29332,29445,29481,29506,29735,29767,29973,30023,30110,30197,30226,30284,30313,30342,30371,30429,30458,30516,30603,30632,30661,30690,30719,30748,30777,30806,30893,30922,30951,31038,31096,31212,31270,31299,31357,31415,31502,31531,31560,31589,31676,31705,31792,31821,31850,31937,31995,32082,32111,32140,32169,32227,32256,32285,32372,32430,32459,32488,32546,32604,32720,32749,32778,32894,32923,32952,32981,33010,33097,33126,33155,33184,33213,33242,33271,33300,33358,33387,33416,33445,33474,33503,33532,33590,33706,33735,33764,33822,33880,33938,33967,33996,34054,34083,34141,34170,34257,34315,34344,34402,34489,34547,34605,34634,34663,34750,34779,34895,34924,34953,34982,35030,35112,35186,35247,35248,35275,35302,35349,35358,35423,35449,35457,35460,35494,35687,35733,35742,35851,35922,36040,36103,36163,36186,36293,36352,36411,36483,36556,36724,36783,36869,36887,36913,36995,37058,37103,37162,37203,37226,37238,37285,37344,37474,37533,37605,37664,37723,37787,37846,37905,37976,38035,38166,38366,38407,38475,38537,38596,38655,38662,38750,38786,38861,38909,39027,39055,39098,39130,39157,39288,39406,39529,39588,39659,39692,39849,39967,39996,40116,40145,40203,40232,40261,40290,40319,40348,40406,40435,40464,40493,40522,40580,40609,40667,40696,40725,40754,40870,40899,40928,40986,41015,41044,41073,41102,41131,41160,41189,41218,41276,41305,41363,41392,41508,41537,42007,42036,42065,42094,42123,42183,42273,42291,42299,42321,42348,42376,42407,42423,42489,42510,42607,42638,42660,42747,42867,42897,42956,43009,43059,43345,43458,43477,43518,43533,43555,43608,43625,43695,43720,43728,43792,43845,43876,43891,43932,43980,44007,44082,44195,44203,44204,44244,44319,44364,44382,44383,44406,44425,44455,44628,44643,44744,44755,44793,44837,44856,44880,44936,44955,44978,44989,44995,45004,45055,45078,45103,45151,45172,45212,45249,45343,45400,45458,45486,45530,45587,45645,45717,45751,45753,45832,45904,45961,46022,46160,46258,46316,46375,46396,46488,46545,46575,46627,46844,46848,46881,46896,46916,46926,46976,47045,47049,47133,47142,47153,47301,47316,47351,47398,47541,47557,47591,47788,47807,47811,47896,47949,47977,48033,48073,48108,48176,48197,48453,48460,48528,48563,48590,48635,48670,48718,48721,48756,48768,48783,48903,48921,48960,48969,49033,49061,49127,49161,49173,49207,49235,49238,49245,49262,49332,49348,49414,49462,49500,49506,49547,49588,49605,49640,49687,49763,49832,49872,49934,49941,49984,50053,50055,50093,50156,50427,50614,50806,50826,51071,51073,51313,51340,51344,51346,51396,51404,51704,51729,51757,51803,51968,52013,52039,52096,52209,52232,52338,52456,52527,52612,52690,52706,52716,52937,52961,53046,53084,53174,53327,53357,53436,53440,53551,53699,53830,54092,54245,54294,54304,54335,54361,54654,54664,54811,54921,55102,55113,55211,55226,55301,55359,55396,55413,55439,55501,55570,55578,55730,55742,55882,55883,55952,56018,56074,56079,56182,56184,56216,56255,56283,56287,56353,56397,56618,56655,56738,56784,56855,56933,57168,57196,57228,57265,57534,57541,57586,57733,57747,57785,57940,57975,58032,58045,58061,58076,58192,58244,58285,58625,58694,58854,58867,58886,58893,58904,59096,59097,59135,59154,59170,59362,59370,59403,59520,59599,59649,59689,59696,59780,59795,60013,60114,60155,60185,60257,60505,60597,60607,60613,60808,60866,60967,60997,61060,61076,61121,61160,61291,61322,61450,61470,61524,61529,61688,61850,61881,61889,61917,62043,62145,62198,62330,62374,62446,62824,63008,63155,63162,63163,63178,63268,63501,63519,63673,63718,64173,64377,64593,64600,64663,64876,64908,64969,65147,65231,65321,65379,65443,65454,65457,65511,65517,65558,65619,65917,66040,66044,66081,66099,66138,66140,66187,66195,66247,66303,66370,66497,66725,67085,67284,67399,67418,67488,67629,67749,67841,67965,67998,68040,68066,68120,68204,68218,68252,68258,68274,68303,68351,68392,68434,68467,68477,68715,69069,69074,69144,69471,69624,69978,69997,70082,70131,70209,70247,70318,70329,70405,70474,70642,70730,70855,70879,70894,71077,71300,71643,71680,71806,71844,71868,71943,72061,72300,72312,72457,72560,72566,72717,72939,73032,73115,73151,73308,73313,73391,73404,73417,73425,73484,73526,73567,73774,73891,74002,74040,74189,74292,74446,74534,74560,74628,74708,74735,74856,74898,74926,74989,75231,75237,75468,75593,75752,75836,75863,75924,75983,76173,76221,76299,76300,76510,76621,76645,76658,77067,77203,77230,77234,77296,77358,77361,77397,77542,77572,77782,77803,77852,77859,77909,77945,78056,78121,78146,78154,78245,78362,78425,78450,78769,78834,78936,78972,79002,79225,79322,79329,79359,79369,79489,79545,79566,79608,79672,79869,80131,80145,80179,80446,80485,80497,80510,80568,80712,80735,80869,80970,81026,81121,81215,81219,81223,81240,81298,81374,81505,81771,81890,81900,81943,81966,82025,82044,82048,82132,82150,82219,82349,82350,82448,82461,82484,82499,82513,82666,82698,82739,83054,83122,83367,83417,83481,83550,83625,83793,83815,83963,84071,84200,84238,84376,84423,84443,84713,84753,84784,84873,84985,85012,85045,85071,85192,85304,85452,85481,85512,85603,85662,85901,85939,85960,86031,86098,86158,86249,86409,86572,86694,86728,86745,86771,86868,86967,87153,87335,87343,87347,87399,87450,87503,87732,87830,87990,88276,88348,88356,88360,88386,88428,88516,88608,88615,88661,88750,88791,88961,88972,89044,89190,89336,89447,89477,89539,89612,89736,89754,89794,89816,89830,89847,89851,89859,89942,89943,90128,90222,90424,90664,90808,90852,90864,90912,90945,90963,91372,91421,91467,91552,91645,91716,91721,91728,91750,91774,91851,91854,91973,91981,91992,92035,92046,92070,92105,92172,92251,92369,92545,92600,92742,92891,93059,93072,93084,93109,93162,93190,93237,93337,93341,93410,93712,93767,93989,94027,94036,94047,94072,94132,94198,94211,94365,94381,94399,94403,94443,94480,94506,94581,94641,94648,94656,94691,94693,94932,95125,95136,95188,95398,95437,95496,95535,95696,95752,95787,95830,95884,95951,96013,96021,96038,96423,96676,96823,96974,97065,97158,97163,97185,97211,97245,97266,97401,97454,97497,97526,97648,97949,98057,98073,98080,98175,98190,98237,98257,98411,98625,98665,98729,98762,98798,98831,98870,98901,98930,99093,99222,99244,99345,99481,99603,99765,99858",
      "20,82,213,359,362,464,589,605,635,719,754,767,782,783,805,830,843,882,917,921,1027,1085,1088,1126,1143,1176,1191,1209,1214,1247,1282,1284,1434,1460,1538,1541,1547,1556,1584,1688,1814,1818,1874,2170,2218,2235,2707,2825,3119,3237,3296,3707,3874,4294,4356,4382,4400,4437,4580,4585,5007,5014,5114,5133,5151,5171,5263,5397,5414,5441,5455,5534,5564,5677,5739,5748,5762,5768,5780,5811,5827,5842,5958,5990,6220,6266,6360,6440,6541,6554,6581,6754,6811,6819,6927,6957,7005,7007,7024,7055,7094,7137,7138,7226,7315,7343,7354,7413,7441,7617,7693,7720,7725,7740,7818,7870,7872,8056,8064,8100,8158,8240,8258,8272,8307,8587,8730,8845,8950,9042,9252,9295,9303,9436,9623,10007,10054,10116,10137,10267,10364,10476,11031,11065,11094,11191,11396,11417,11448,11519,11572,11810,11984,12073,12082,12133,12337,12376,12589,12819,13167,13387,13423,13570,13918,14373,14434,14865,14963,14994,15090,15203,15344,15381,15719,15992,16092,16248,16541,16729,16900,17142,17208,17491,17607,17958,18041,18348,18513,18719,18723,18745,19201,19211,19355,19450,19591,19653,19792,20068,20113,20134,20255,20488,20806,20945,21001,21008,21103,21207,21281,21383,21796,21825,21920,22058,22068,22184,22201,22889,23165,23584,23896,24238,24394,24419,24535,24704,24803,25492,25746,26204,26375,26827,26917,26988,27109,27185,27253,27264,27403,27444,27741,27750,27835,27850,28189,28763,29031,29878,29936,30835,31154,31183,31386,31473,31618,32198,33677,34228,34373,35137,35695,36113,36129,36542,36563,36568,37243,37442,37571,37813,37869,37874,37945,38094,38251,38430,38496,38802,38808,38988,38990,39252,39367,39381,39418,39936,39982,42304,42333,42611,43075,43199,43441,43678,43770,43820,44898,45311,45312,45742,45885,45920,45995,45999,46775,47046,47595,47799,48163,48648,50107,50125,50143,50159,50214,50316,50473,50659,50715,50781,50812,50982,51103,51245,51252,51278,51308,51383,51419,51466,51542,51559,51575,51645,51766,51824,51833,51835,51905,51950,52267,52301,52305,52339,52454,52506,52517,52666,52674,52860,52865,52934,52957,52986,53064,53067,53101,53170,53318,53370,53411,53716,53767,53773,53829,53870,53951,54087,54230,54321,54473,54552,54593,54641,54668,54685,54733,54861,55036,55132,55160,55172,55180,55244,55294,55323,55333,55381,55429,55446,55460,55495,55566,55591,55736,55765,55785,55866,55943,56196,56215,56336,56361,56392,56409,56420,56443,56471,56571,56795,56802,56846,56875,56887,56944,56991,57054,57092,57190,57264,57546,57616,57618,57645,57809,57910,57950,57994,58077,58153,58175,58237,58352,58354,58416,58450,58483,58579,58649,58653,58709,58741,58776,58842,58902,59147,59262,59270,59271,59352,59415,59423,59448,59456,59604,59645,59808,60063,60091,60107,60125,60218,60472,60530,60563,60628,60643,60742,60807,60825,60843,60878,60927,61014,61025,61055,61098,61231,61555,61583,61616,61711,61774,61809,61841,61892,61915,61941,62000,62035,62121,62146,62160,62245,62559,62591,63017,63036,63066,63070,63118,63179,63310,63359,63589,63649,63840,63961,64058,64074,64133,64139,64281,64316,64324,64342,64405,64455,64478,64570,64604,64648,64795,64814,64819,64922,65046,65076,65171,65223,65230,65251,65266,65346,65608,65762,65779,65802,65821,66035,66134,66174,66224,66228,66320,66458,66583,66639,66759,66936,67024,67034,67065,67068,67074,67124,67333,67340,67592,67677,67699,67705,67813,67898,67931,68058,68100,68147,68264,68458,68459,68603,68622,68648,68839,68897,68924,68926,68965,69083,69095,69161,69177,69267,69413,69539,69576,69615,69660,69695,69738,69952,70076,70102,70122,70123,70588,70624,70699,70724,70766,70807,70886,70890,70897,70954,70986,71116,71132,71138,71158,71171,71189,71237,71262,71361,71368,71379,71482,71507,71566,71634,71748,71760,71920,71956,72004,72019,72046,72196,72239,72283,72448,72601,72651,72661,72694,72712,72871,72962,72970,73242,73352,73443,73452,73478,73489,73522,73535,73559,73647,73695,73730,73831,73945,73972,73981,74215,74236,74248,74298,74379,74579,74598,74600,74748,74793,74828,74859,74869,74884,74912,74915,75098,75118,75135,75166,75355,75419,75523,75595,75786,75825,75888,75941,76046,76132,76292,76316,76340,76346,76434,76462,76577,76648,76736,76750,76766,76828,76833,77001,77038,77047,77513,77573,77577,77639,77687,77811,77814,77912,77913,77951,78047,78305,78343,78534,78815,78894,78942,79030,79062,79069,79090,79127,79202,79240,79528,79619,79632,79684,79720,79746,79792,79841,80022,80034,80061,80106,80329,80357,80413,80444,80500,80552,80606,80841,80868,80917,80940,81284,81297,81378,81442,81501,81530,81546,81602,81731,81740,81852,81976,82112,82245,82291,82395,82432,82527,82712,82823,82836,82873,82891,82917,82918,82991,83063,83115,83156,83210,83251,83253,83318,83338,83671,84019,84039,84049,84108,84212,84244,84299,84300,84305,84415,84459,84504,84515,84562,84614,84630,84636,84654,84729,84769,84810,85031,85167,85203,85418,85436,85748,85879,85899,85996,86045,86064,86154,86212,86277,86336,86400,86410,86415,86438,86615,86674,86778,86783,87029,87127,87345,87381,87514,87579,87632,87642,87719,87856,87859,87864,87894,88101,88141,88176,88404,88613,88654,88733,88840,88928,88954,89009,89039,89228,89396,89402,89567,89637,89667,89693,89836,90141,90201,90332,90428,90443,90547,90683,90744,90769,90774,90798,90986,91182,91255,91276,91362,91402,91473,91518,91535,91538,91683,91803,91990,92050,92137,92217,92508,92570,92574,92642,92799,92814,93033,93039,93547,93576,93640,93651,93797,93818,93919,93940,93959,94020,94045,94112,94244,94373,94429,94497,94559,94616,94645,94647,94756,94778,94790,94854,94901,95031,95087,95140,95151,95175,95258,95263,95307,95421,95562,95680,95703,95723,95770,95772,95817,95861,95980,96020,96129,96240,96261,96347,96414,96479,96482,96569,96620,96636,96658,96661,96921,97122,97145,97392,97409,97423,97499,97666,97759,97813,97924,97960,98027,98137,98264,98421,98636,98645,98865,99111,99132,99176,99197,99210,99291,99312,99334,99402,99518,99583,99635,99807,99842,99845,99913,99914",
      "214,268,634,690,812,936,951,1106,1108,1173,1188,1360,1394,1413,1430,1438,1455,1465,1468,2105,2107,2129,2259,2326,2671,2696,2780,3065,3294,3311,3584,3852,4051,4239,4316,4392,4497,4592,4690,4711,4829,4869,4942,5017,5024,5278,5279,5359,5386,5434,5437,5498,5668,5746,5763,5824,5844,5937,6012,6043,6046,6123,6144,6172,6186,6212,6240,6341,6413,6425,6452,6492,6535,6663,6714,6725,6746,6772,6818,7056,7062,7159,7160,7171,7228,7233,7387,7608,7750,7776,7781,7893,7955,8008,8031,8107,8235,8377,8379,8463,8472,8503,8532,8710,8868,8905,8965,8983,9114,9124,9372,9381,9606,9714,9910,9918,9981,10009,10031,10067,10110,10251,10280,10283,10386,10444,10453,10470,10474,10704,10720,10797,10947,10957,10988,11139,11159,11176,11253,11402,11530,11670,11786,11841,12096,12120,12171,12211,12354,12366,12392,12498,12532,12583,12591,12618,12660,12810,12929,13085,13210,13332,13339,13359,13454,13488,13788,13837,13969,13983,13985,13994,14145,14159,14263,14311,14344,14398,14465,14581,14652,14740,14885,15074,15132,15187,15210,15395,15467,15526,15578,15614,15615,15777,15800,15925,16001,16123,16173,16307,16372,16399,16441,16477,16638,16803,16948,17090,17105,17135,17166,17324,17333,17659,17751,18035,18453,18460,18576,18695,18998,19062,19093,19145,19157,19213,19293,19440,19526,19532,19576,19962,20132,20191,20258,20506,20590,20601,20608,20785,20861,21037,21056,21076,21114,21151,21312,21362,21437,21507,21855,22006,22086,22180,22339,22367,22566,22794,23001,23026,23077,23088,23123,23207,23297,23450,23451,23655,23831,24105,24324,24420,24515,24590,24591,24677,24822,24838,25074,25122,25129,25218,25223,25238,25256,25663,25864,25988,26035,26176,26263,26285,26307,26320,26358,26418,26490,26812,26843,26918,26963,26998,27056,27163,27313,27813,27873,27905,27912,27913,28266,28316,28425,28433,28503,28705,28719,28768,28798,28815,28888,28940,28961,29023,29112,29137,29175,29291,29298,29307,29359,29449,29496,29569,29812,29900,29903,29906,30033,30168,30304,30400,30526,30545,30642,31774,32024,32296,32441,32575,33108,33329,33659,34093,34209,34296,34460,34721,34819,34866,34877,35135,35257,35305,35311,35465,35552,35749,35760,36328,36445,36512,36524,36972,37044,37074,37112,37122,37258,37352,37456,37494,37620,37627,37817,37928,38131,38134,38375,38380,38519,38674,38701,38736,38755,38801,38819,38878,38943,39127,39169,39216,39365,39366,39371,39422,39470,39480,39495,39564,39818,39822,39830,39858,39867,39948,40118,40140,40198,40292,40814,40836,40923,40957,41191,42232,42258,42439,42630,42762,42800,42822,42914,42946,43237,43529,43772,43946,44125,44336,44422,44461,44623,44701,44718,44874,45197,45427,45452,45612,45795,45812,46066,46340,46387,46500,46561,47864,47894,47956,48007,48086,48385,48552,48727,48745,48964,49024,49182,49523,49722,49786,50002,50011,50021,50304,50386,50725,50866,51145,51169,51226,51491,51722,51738,51841,52019,52166,52265,52406,52432,52486,52576,52673,52680,52800,52802,52894,53030,53249,53299,53400,53416,53428,53446,53592,53607,53629,53819,53921,53937,53965,54007,54049,54149,54209,54559,54620,54626,54702,54749,54864,54901,55124,55156,55166,55174,55229,55263,55281,55303,55358,55464,55510,55538,55624,55675,55808,56066,56300,56310,56425,56669,56696,56763,56775,56817,56877,56980,57019,57022,57074,57156,57161,57225,57392,57429,57512,57559,57611,57778,57815,57871,57936,57963,58114,58119,58122,58265,58303,58336,58374,58403,58437,58517,58531,58662,58729,58733,58803,58839,58938,58985,58990,59024,59026,59071,59077,59091,59190,59238,59428,59443,59534,59720,60006,60008,60022,60025,60051,60064,60179,60197,60325,60372,60492,60533,60686,60942,60974,61039,61080,61152,61376,61385,61487,61514,61545,61623,61675,61682,61716,61987,61990,62001,62023,62120,62170,62216,62360,62365,62379,62415,62548,62978,62997,63102,63110,63123,63143,63194,63250,63336,63350,63374,63485,63510,63537,63617,63623,63626,63910,64009,64068,64072,64090,64114,64181,64193,64411,64468,64479,64485,64564,64652,64909,64910,65030,65069,65271,65463,65494,65565,65651,65706,65835,65877,65892,65894,65985,66049,66077,66188,66217,66288,66338,66427,66490,66737,66764,66842,67252,67352,67432,67485,67486,67650,67662,67717,67738,67859,67935,67937,68053,68097,68281,68293,68339,68389,68490,68627,68640,68700,68772,68840,68939,68950,68993,69068,69123,69131,69344,69359,69559,69655,69680,69731,69992,70046,70053,70229,70262,70362,70370,70437,70454,70482,70528,70574,70750,70782,70785,71059,71104,71204,71227,71241,71308,71317,71354,71469,71524,71531,71605,71660,71752,72289,72378,72598,72612,72779,72841,72844,72856,72985,72992,72999,73236,73266,73482,73571,73575,73898,73909,73975,73998,74025,74155,74184,74209,74222,74314,74329,74424,74457,74710,74726,74811,74831,74838,74903,75087,75116,75292,75371,75521,75612,75648,75651,75675,75724,75738,75751,75765,75904,75990,76017,76138,76428,76448,76455,76582,76682,76706,76741,76761,76820,76830,76860,76867,76971,77006,77106,77301,77493,77540,77713,77804,77823,77873,77874,77942,78041,78166,78241,78266,78298,78315,78316,78477,78532,78552,78896,79012,79050,79074,79075,79280,79300,79402,79409,79429,79459,79532,79543,79637,79658,79715,79730,79747,79872,79878,80161,80299,80427,80453,80471,80585,80676,80861,80870,80995,81051,81174,81307,81324,81538,81572,81685,81751,81775,81836,82020,82069,82074,82096,82144,82261,82379,82392,82426,82449,82588,82636,82638,82843,82878,82939,82990,83155,83285,83324,83506,83626,83773,83799,83808,84077,84143,84196,84237,84271,84311,84375,84425,84470,84484,84559,84567,84597,84797,84894,84932,84958,85040,85063,85185,85206,85417,85448,85574,85795,85802,85971,86019,86204,86215,86343,86403,86405,86496,86506,86589,86642,86657,86727,86800,86874,86944,86946,86953,86969,86976,87180,87219,87332,87388,87402,87612,87613,87795,87833,87845,87883,87992,88006,88045,88173,88183,88189,88435,88454,88491,88577,88604,88616,88687,88801,88855,88980,88991,89014,89029,89097,89171,89180,89219,89271,89654,89715,89726,89793,89795,89798,89839,89848,89960,90055,90059,90203,90241,90280,90289,90341,90365,90411,90525,90550,90584,90682,90690,90847,90873,90898,91032,91063,91073,91122,91229,91291,91371,91470,91495,91500,91522,91699,91861,91936,92102,92247,92253,92605,92698,92853,92865,92974,93367,93470,93579,93621,93683,93783,93816,93888,93908,94006,94147,94228,94531,94550,94679,94697,94708,94830,95029,95082,95108,95262,95589,95596,95720,95795,95880,95939,96033,96091,96136,96193,96274,96650,96662,96679,96697,96899,96998,96999,97047,97190,97255,97361,97369,97419,97420,97421,97450,97532,97536,97578,97603,97810,97822,98120,98356,98369,98513,98554,98565,98605,98725,98849,98970,99008,99040,99290,99443,99622,99626,99875,99931,99959,99970,99994",
      "251,293,328,388,403,432,435,523,687,716,777,796,825,1006,1017,1026,1062,1139,1189,1200,1207,1230,1234,1256,1276,1298,1334,1374,1378,1388,1411,1464,1476,1485,1525,1555,1633,1653,1783,1891,2815,3008,3235,3241,3376,4754,4864,5061,5082,5092,5127,5147,5168,5169,5193,5233,5310,5324,5393,5408,5430,5560,5584,5604,5691,5720,5785,5846,5874,5917,5936,5951,5974,6001,6004,6088,6153,6506,6785,6903,6937,7023,7031,7074,7130,7153,7253,7357,7403,7646,7683,7765,7969,8028,8095,8234,8247,8374,8400,8493,8514,8529,8551,8566,8598,8613,8660,8672,8780,8826,8918,8938,8994,8999,9069,9131,9170,9189,9302,9304,9330,9435,9914,10170,10175,10190,10211,10295,10380,10440,10465,10519,10526,10569,11136,11189,11271,11272,11324,11545,11719,11720,11760,12064,15379,15387,15902,15932,15981,17516,20693,23382,23552,24816,24915,25344,26334,26947,28020,28299,28584,28823,28873,28981,29082,29090,29463,29609,30962,30976,32106,34024,34362,34980,35929,36631,36865,37298,37925,38174,39311,39995,41257,41315,42472,42883,45516,45551,45880,46189,46231,46617,48546,49752,50014,50162,50245,50311,50319,50338,50391,50401,50477,50705,50837,50972,51057,51117,51150,51166,51188,51267,51287,51302,51429,51442,51567,51629,51677,51684,51741,51916,51918,52010,52059,52129,52132,52181,52190,52254,52388,52420,52521,52526,52544,52546,52603,52608,52630,52650,52719,52752,52818,52933,52948,52973,52979,53074,53135,53151,53185,53201,53259,53273,53296,53321,53337,53355,53365,53437,53458,53535,53556,53579,53796,53836,53869,54000,54056,54121,54243,54251,54637,54652,54672,54679,54683,54853,54903,54915,54929,54969,54987,55152,55179,55266,55573,55635,55706,55719,55864,55973,56019,56251,56298,56314,56526,56557,56559,56587,56588,56897,56963,57032,57051,57209,57271,57333,57580,57609,57624,57855,57904,58075,58123,58304,58686,58688,58831,58862,58866,58944,58961,59020,59049,59095,59101,59160,59255,59257,59290,59705,59788,59797,59849,59964,60027,60194,60295,60373,60386,60463,60549,60554,60569,60737,60868,60902,61089,61153,61238,61406,61453,61491,61596,61644,61701,61773,61832,61932,61969,62123,62144,62354,62432,62682,62884,63012,63046,63115,63153,63180,63294,63356,63398,63464,63474,63691,63787,63808,63904,63926,63960,63979,64111,64157,64170,64237,64366,64380,64448,64531,64637,64816,64839,65158,65243,65361,65566,65576,65603,65670,65712,65754,65765,65844,66184,66390,66453,66481,66530,66600,66638,66812,66843,66874,66946,66965,67050,67110,67211,67222,67241,67246,67280,67414,67428,67652,67702,67710,67722,67878,67936,68121,68150,68243,68314,68340,68410,68450,68709,68735,68895,68909,69000,69138,69245,69252,69303,69315,69438,69472,69605,69699,69743,69816,69844,69867,70072,70155,70257,70414,70498,70561,70849,70860,70909,71004,71044,71060,71274,71275,71411,71505,71821,72023,72062,72251,72428,72937,73006,73076,73086,73274,73438,73523,73540,73542,73576,73581,73729,73874,73912,73956,74038,74039,74095,74167,74176,74337,74397,74513,74521,74559,74692,74776,74913,75008,75065,75117,75142,75378,75453,75496,75575,75728,75785,75844,75846,75873,76025,76117,76406,76452,76661,76666,76729,77151,77216,77225,77252,77265,77287,77356,77432,77507,77586,77594,77598,77645,77670,77863,77877,77878,77882,78046,78085,78101,78150,78161,78195,78197,78272,78361,78377,78386,78665,78705,78835,79037,79042,79072,79084,79291,79443,79475,79525,79659,79840,79917,79922,79926,79967,80068,80076,80087,80129,80375,80548,80591,80752,80755,80805,80919,81035,81061,81078,81105,81153,81178,81189,81244,81245,81447,81562,81599,81651,81680,81769,81818,81897,81911,82064,82212,82264,82287,82303,82328,82353,82358,82388,82608,82642,82687,82732,82748,82789,82892,82948,82960,82996,83236,83339,83362,83393,83587,83645,83735,83882,83951,83970,84033,84136,84169,84239,84276,84347,84367,84418,84554,84641,84731,84768,84921,84928,84944,85072,85173,85265,85273,85281,85346,85587,85632,85811,85973,86005,86009,86161,86252,86308,86338,86404,86479,86519,86742,86919,86977,87103,87161,87173,87234,87278,87341,87417,87452,87655,87690,87977,87981,88149,88152,88158,88204,88264,88275,88289,88417,88455,88479,88490,88689,88742,88778,88813,88845,89033,89076,89080,89094,89156,89168,89231,89276,89291,89349,89394,89782,89974,89999,90085,90094,90151,90297,90334,90354,90414,90623,90633,91027,91030,91060,91067,91099,91102,91140,91311,91479,91505,91510,91586,91723,92011,92061,92143,92309,92359,92427,92442,92524,92650,92813,93131,93265,93496,93551,93655,93681,93710,93748,93773,93819,93823,93875,93916,94089,94116,94149,94226,94231,94250,94254,94328,94357,94426,94441,94473,94518,94802,94893,95118,95455,95468,95522,95638,95738,95851,95862,95874,95929,96064,96065,96285,96345,96352,96380,96560,96719,96722,96737,96960,97216,97345,97596,97605,97640,97653,97677,97789,97833,97934,97935,97955,98031,98089,98344,98428,98431,98465,98482,98491,98500,98710,99006,99161,99323,99362,99688,99929,99947,99964,99992",
      "62,101,133,257,266,274,285,291,294,466,474,493,522,534,617,793,831,844,1059,1195,1471,1507,1551,1691,1711,1746,2751,3718,4101,4268,4781,5100,5361,5399,5471,5499,5522,5553,5565,5589,5620,5636,5721,5987,6136,6191,6294,6306,6502,6557,6654,6740,6751,6861,7058,7067,7107,7135,7327,7395,7463,7581,7589,7708,7772,8114,8139,8169,8246,8274,8329,8390,8692,8786,8937,9116,9239,9240,9315,9450,9457,9490,9519,9836,9892,9919,9995,10030,10089,10180,10315,10358,10368,10400,10518,10568,10574,10582,10652,10670,10787,10913,10965,10972,10997,11048,11099,11366,11418,11451,11668,11677,11913,11931,11977,12004,12007,12038,12425,12502,12517,12533,12674,12687,12800,12869,12877,13088,13136,13146,13153,13549,13572,13645,13668,13888,13912,13919,13980,14450,14640,14692,14751,14823,14858,15037,15045,15105,15118,15174,15178,15209,15274,15441,15447,15958,16057,16094,16434,16462,16528,16547,16676,16700,16752,16868,16992,17029,17099,17419,17649,17955,18015,18056,18068,18113,18159,18174,18599,18974,18979,19000,19238,19410,19457,19521,19758,20218,20238,20684,20698,20984,21266,21399,21557,21815,22291,22297,22590,22592,22648,22944,23016,23633,23663,23778,23923,24048,24173,24284,24300,24375,24382,25454,25933,26167,26256,26315,26816,26857,26873,27111,27125,27161,27284,27287,27411,27706,27715,27757,27779,27787,27801,27824,27914,27941,28274,28368,28461,28978,29121,29129,29297,29540,35763,35764,36075,36435,36516,36698,37015,37744,37997,38358,38379,38488,38513,38687,38816,38917,39073,39298,39302,39620,39694,39807,45872,50038,50052,50080,50097,50132,50192,50208,50212,50551,50599,50864,50954,51076,51277,51328,51351,51387,51789,51836,51969,52113,52210,52248,52367,52397,52508,52676,52714,52803,52907,52918,53129,53178,53188,53198,53313,53430,53522,53620,53701,53743,53789,53906,53981,53982,54048,54415,54421,54524,54530,54600,54909,54916,55054,55456,55625,55626,55629,55695,55749,55796,55823,55863,55912,55945,55974,56071,56088,56203,56290,56313,56344,56399,56433,56540,56832,56964,57043,57086,57123,57125,57194,57203,57233,57241,57436,57510,57529,57582,57691,57731,57789,57813,57869,57998,58010,58307,58311,58315,58408,58414,58527,58725,58736,58788,58880,58998,59174,59552,59700,59783,59812,59926,60041,60124,60126,60156,60234,60352,60690,60704,60720,60755,60953,61090,61130,61218,61433,61521,61608,61761,61788,61939,61983,62079,62094,62178,62472,62648,62708,62728,62765,62864,63060,63077,63238,63242,63288,63505,63692,63861,63955,64014,64153,64162,64190,64253,64383,64386,64538,64543,64847,64865,64880,65058,65169,65304,65363,65426,65725,65927,65949,66019,66253,66356,66460,66510,66523,66545,66590,66666,66731,66747,66748,66911,66924,67019,67023,67293,67301,67357,67375,67392,68153,68194,68525,68540,68550,68719,68757,68810,68998,69004,69075,69133,69145,69286,69337,69383,69406,69423,69434,69480,69494,69701,69707,69856,69904,69939,69946,70108,70115,70223,70333,70343,70467,70565,70600,70664,70701,70792,70915,70939,71090,71372,71389,71820,71872,71974,71990,71997,72008,72016,72071,72119,72274,72279,72438,72454,72463,72520,72615,72705,72767,72798,72802,72806,72854,72906,72917,72984,73015,73019,73020,73031,73093,73106,73130,73165,73373,73598,73605,73716,73764,73832,73846,73860,73871,74055,74079,74086,74137,74273,74346,74381,74585,74742,74767,74864,74936,74944,75147,75155,75325,75369,75380,75467,75510,75532,75550,75674,75827,75843,75938,76015,76223,76381,76473,76704,76908,76984,77266,77270,77305,77347,77365,77405,77459,77665,77822,77869,77893,77996,78017,78209,78327,78360,78428,78447,78469,78526,78658,78679,78688,78746,78760,78812,78824,78838,78908,79041,79108,79121,79123,79132,79172,79305,79310,79423,79437,79547,79629,79646,79744,79755,79823,79866,80062,80081,80094,80114,80202,80253,80326,80458,80494,80820,80848,80932,81006,81132,81205,81206,81286,81398,81409,81459,81518,81622,81681,81784,81820,81826,81830,81862,81917,81930,82032,82061,82085,82114,82310,82315,82402,82408,82422,82430,82623,82752,82876,82956,83007,83013,83038,83148,83151,83512,83549,83568,83803,83833,83909,84118,84133,84217,84240,84304,84352,84355,84393,84468,84508,84593,84649,84712,84897,84974,85181,85195,85361,85396,85505,85518,85598,85751,85764,85765,85775,85916,85933,86025,86299,86444,86511,86957,87025,87336,87351,87421,87495,87518,87530,87610,87714,87724,88123,88145,88170,88215,88266,88279,88537,88634,88751,88859,88912,88971,88997,89027,89038,89042,89121,89186,89202,89245,89251,89257,89299,89329,89342,89353,89385,89562,89583,89651,89728,89820,89956,90089,90140,90181,90287,90362,90401,90490,90522,90579,90870,90911,90976,90988,91055,91062,91084,91152,91222,91353,91544,91567,91591,91630,91690,91748,91781,91841,91845,91874,91933,91961,91969,91979,92023,92240,92264,92270,92305,92343,92498,92566,92667,92679,92713,92718,92726,92750,92835,92841,92879,93080,93098,93174,93221,93468,93524,93544,93850,93858,93877,94003,94119,94163,94174,94232,94325,94352,94574,94633,94654,94973,95022,95051,95093,95194,95215,95217,95231,95271,95287,95558,95927,96124,96159,96225,96230,96396,96424,96492,96696,96802,96874,96876,97084,97090,97154,97223,97238,97250,97319,97410,97471,97515,97550,97558,97609,97654,97778,97852,97912,97945,98026,98068,98125,98204,98208,98221,98306,98350,98401,98427,98480,98540,98750,98757,98799,98800,98877,98988,99045,99078,99130,99151,99367,99414,99459,99529,99556,99675,99871",
      "7,21,98,216,302,545,560,593,597,614,663,704,714,730,732,751,778,787,794,828,919,938,953,968,980,987,1063,1168,1316,1375,1408,1429,1443,1481,1501,1524,1603,1803,1895,1907,1993,2020,2063,2085,2086,2192,2320,2321,2324,2423,2521,2640,2737,3092,3282,3324,3653,3773,3884,4110,4273,4366,4495,4802,4862,4969,5026,5153,5369,5383,5612,5646,5711,5717,5731,5774,5862,5942,6009,6169,6357,6526,6531,6650,6679,6850,6971,6973,7009,7065,7141,7178,7306,7382,7400,7484,7532,7624,7652,7706,7715,7774,7888,7972,7975,8177,8298,8341,8411,8436,8447,8470,8526,8602,8708,8759,8783,8795,8879,8889,8943,8973,8981,9133,9186,9229,9365,9609,9631,9814,9953,9970,10122,10150,10179,10348,10349,10543,10596,10736,10954,11006,11010,11018,11053,11150,11343,11397,11427,11428,11438,11452,11537,11569,11593,11626,11770,11797,11816,11902,12218,12443,12571,12606,12608,12659,12725,12950,13119,13294,13331,13512,13637,14422,14483,14630,14854,15007,15167,15402,15616,15882,15996,16021,16030,16062,16112,16585,16848,16973,17031,17124,17241,17283,17402,17422,17614,17618,17689,18430,18543,18558,18580,18657,18768,19055,19193,19226,19356,19361,19640,19815,20404,20518,20675,20760,20784,20832,20924,21031,21041,21133,21208,21449,21514,21532,21899,22074,22161,22225,22391,22646,22799,22931,23080,23090,23374,23403,23441,23545,23723,23749,24087,24320,24333,24397,24524,24546,24787,24841,24868,24932,24979,25046,25336,25366,25439,25486,25641,25732,25750,25884,26047,26251,26321,26323,26332,26459,26661,26703,26715,26797,26814,27214,27227,27279,27352,27364,27447,27649,27662,27678,27735,28290,28323,28366,28426,28432,28524,28700,28734,28882,28931,28972,29016,29042,29100,29198,29280,29345,29386,29792,29803,29824,29852,29984,30578,31097,32373,32634,32782,32811,33040,33127,33417,33446,34142,35378,35572,35800,35926,35943,36076,36282,36300,36313,36442,36500,36510,36620,36638,36641,36814,36941,37045,37261,37314,37366,37380,37498,37549,37819,37950,38067,38258,38324,38368,38437,38441,38500,38675,38696,38757,39000,39351,39362,39610,39682,39729,39752,39803,39872,40217,40509,41198,41290,42323,42488,42550,42573,42722,42781,42860,43111,43196,43394,43433,43885,44302,44430,44463,45185,45189,45235,45252,45329,45881,45950,45993,46118,46455,46574,46589,46830,46873,47555,48477,49057,49198,49582,49675,50060,50172,50187,50378,50380,50442,50490,50525,50544,50663,50788,50883,50916,50921,51365,51436,51528,51532,51663,51755,51946,51993,52027,52068,52082,52168,52237,52251,52304,52423,52491,52499,52887,52889,52969,53050,53224,53256,53473,53576,53662,53824,53848,53880,53996,54023,54036,54083,54096,54107,54194,54265,54323,54395,54580,54610,54748,54838,55056,55068,55074,55108,55151,55315,55335,55366,55461,55505,55621,55759,55908,56025,56093,56101,56234,56249,56258,56302,56383,56390,56507,56514,56551,56753,56767,56787,56807,56809,56952,57101,57134,57239,57306,57433,57620,57719,57915,57928,57961,57962,57981,58007,58070,58150,58328,58348,58487,58562,58648,58739,58754,58792,58845,59006,59184,59221,59281,59314,59408,59698,59768,59813,59851,59868,59899,59925,59987,60089,60151,60209,60251,60338,60437,60438,60586,60594,60623,60764,60922,60933,61048,61243,61345,61351,61485,61532,61551,61704,61847,61985,62061,62074,62171,62172,62238,62325,62367,62588,62634,62643,62793,62885,62906,62941,63000,63129,63362,63376,63548,63569,63570,63596,63651,63707,63869,63872,63901,64039,64206,64221,64241,64371,64546,64724,64747,64772,65064,65233,65305,65374,65438,65455,65460,65489,65653,65751,65862,65940,66037,66259,66334,66526,66605,66700,66732,66753,66967,67096,67131,67304,67475,67524,67598,67612,67635,67655,67721,67734,67838,67923,67944,68004,68206,68212,68286,68347,68518,68729,68778,68883,68956,68985,69087,69113,69117,69125,69134,69454,69485,69623,69667,69830,69892,69960,70225,70240,70272,70287,70321,70366,70442,70520,70564,70618,70727,71067,71085,71198,71228,71287,71345,71412,71502,71530,71783,71833,71867,71871,71923,72103,72104,72259,72443,72518,72531,72654,72755,72769,72847,72953,73097,73100,73141,73244,73312,73394,73473,73593,73724,73887,73957,74081,74129,74156,74217,74398,74508,74578,74630,74655,74729,74840,74897,74943,75034,75091,75100,75137,75195,75204,75219,75300,75360,75524,75600,75601,75607,75644,75671,75699,75965,76029,76056,76096,76208,76217,76230,76260,76400,76458,76516,76536,76559,76684,76797,77034,77187,77219,77232,77329,77461,77552,77679,77819,77828,77867,77978,78030,78040,78096,78136,78277,78357,78509,78678,78916,78969,79136,79167,79270,79343,79455,79458,79634,79711,79756,79772,79777,79902,80032,80100,80109,80110,80148,80169,80231,80301,80587,80711,80846,80945,80953,81021,81052,81195,81199,81237,81250,81254,81295,81346,81500,81605,81610,81770,81786,81843,81874,81923,82475,82503,82510,82704,82730,82769,82881,82904,82906,82953,82973,83018,83235,83266,83281,83356,83380,83394,83464,83577,83582,83703,83820,83911,83976,84012,84095,84096,84203,84216,84287,84313,84327,84457,84606,84761,84853,84960,85122,85209,85264,85330,85397,85437,85470,85500,85614,85636,85661,85739,85797,85861,85934,86111,86139,86181,86253,86280,86297,86473,86586,86628,86817,86904,86935,87098,87166,87195,87265,87328,87342,87623,87697,87787,87877,87973,88021,88119,88155,88188,88351,88651,88725,88736,88781,88853,88881,88887,88931,88942,89188,89416,89449,89450,89643,89668,89833,90232,90262,90303,90340,90380,90392,90400,90478,90685,90884,90923,91038,91081,91206,91347,91494,91504,91562,91710,91712,91823,91910,91955,92135,92228,92395,92488,92709,92917,92941,92985,93049,93184,93201,93256,93396,93398,93403,93454,93804,93970,94205,94215,94498,94593,94600,94623,94629,94638,94687,94766,94791,94801,94849,94879,95157,95171,95296,95328,95389,95400,95418,95433,95474,95765,95773,95828,96187,96229,96336,96381,96386,96439,96481,96647,96736,96750,96839,97142,97150,97275,97286,97328,97344,97452,97481,97845,98113,98288,98365,98385,98447,98510,98536,98657,98713,98780,98804,98822,98932,99007,99030,99135,99173,99225,99280,99472,99484,99647,99746,99794,99838,99856,99982",
      "136,184,230,1620,1676,1745,1764,1992,2011,2061,2122,2149,2173,2209,2438,2440,2464,2582,2595,2691,2839,2985,3015,3050,3285,3329,3474,3695,4025,4063,4271,4306,4391,4488,4558,4581,4597,4618,4642,4849,4892,5075,7120,7373,7773,8393,9184,9497,9611,9630,9649,9668,9706,9720,9725,9782,9801,10048,10599,10618,10675,10713,10725,10808,10903,10941,11017,11093,11112,11131,11226,11227,11245,11359,11444,11454,11511,11525,11528,11550,11587,11606,11644,11663,11891,11948,11967,11986,12024,12157,12233,12252,12271,12290,12328,12329,12385,12403,12423,12478,12499,12537,12596,12613,12639,12668,12679,12708,12765,12803,12847,12860,12883,12898,12923,12974,13012,13048,13050,13069,13114,13145,13164,13240,13297,13345,13354,13362,13373,13430,13449,13563,13617,13639,13658,13677,13715,13734,13772,13791,13829,13838,13855,13870,13889,13962,13981,14019,14045,14076,14082,14095,14147,14152,14171,14190,14201,14228,14247,14266,14285,14304,14323,14332,14380,14418,14446,14475,14502,14570,14571,14646,14665,14672,14722,14741,14742,14759,14859,14874,14931,15026,15064,15083,15100,15140,15197,15235,15273,15292,15311,15349,15368,15425,15444,15501,15577,15595,15596,15605,15640,15672,15691,15710,15748,15805,15824,15843,15862,15900,15919,15938,15942,15976,15988,16052,16071,16090,16100,16147,16166,16183,16185,16204,16221,16242,16261,16273,16280,16293,16299,16318,16337,16356,16413,16428,16432,16451,16458,16470,16508,16527,16565,16584,16622,16698,16717,16793,16831,16888,16929,16962,16964,17021,17040,17059,17079,17097,17154,17173,17192,17211,17239,17268,17306,17325,17363,17382,17401,17458,17496,17515,17534,17591,17624,17648,17686,17705,17743,17762,17800,17819,17838,17876,17895,17914,17971,17990,18028,18066,18104,18123,18142,18161,18170,18216,18218,18237,18250,18256,18275,18313,18332,18351,18370,18389,18408,18427,18484,18503,18523,18579,18587,18598,18617,18674,18693,18731,18734,18807,18826,18854,18883,18902,18940,18941,18959,18978,18997,19021,19028,19044,19054,19073,19092,19130,19206,19232,19282,19320,19377,19396,19415,19434,19468,19472,19529,19544,19548,19586,19624,19662,19700,19724,19738,19757,19776,19795,19814,19833,19852,19871,19890,19928,19966,20023,20061,20080,20099,20108,20156,20182,20194,20203,20232,20251,20319,20327,20346,20365,20384,20403,20422,20441,20460,20517,20536,20555,20574,20612,20631,20669,20688,20726,20745,20764,20775,20840,20859,20878,20916,20935,20954,20992,21011,21049,21087,21106,21159,21163,21201,21220,21239,21296,21303,21353,21372,21410,21429,21505,21524,21562,21591,21600,21607,21619,21638,21647,21657,21676,21714,21752,21790,21809,21828,21866,21885,21904,21923,21942,21961,21980,21999,22037,22056,22075,22094,22113,22151,22170,22227,22247,22256,22265,22284,22322,22330,22341,22360,22398,22417,22436,22449,22455,22512,22522,22531,22569,22580,22607,22626,22683,22702,22721,22723,22759,22797,22816,22820,22835,22930,22987,23025,23044,23082,23101,23120,23177,23196,23291,23310,23367,23394,23424,23453,23472,23481,23500,23538,23557,23576,23582,23614,23652,23671,23688,23709,23718,23728,23747,23782,23804,23861,23880,23899,23918,23937,23956,23975,24013,24032,24070,24101,24127,24144,24165,24184,24203,24260,24276,24298,24317,24355,24374,24393,24412,24433,24450,24469,24507,24526,24545,24583,24602,24612,24621,24659,24678,24697,24716,24773,24811,24830,24883,24887,24925,24952,24963,24982,24988,25020,25058,25064,25096,25115,25134,25191,25248,25267,25268,25305,25318,25324,25343,25345,25419,25438,25476,25495,25552,25590,25609,25627,25780,25799,25818,25863,25875,25913,25929,25951,25970,26008,26027,26046,26065,26084,26103,26107,26122,26131,26141,26179,26185,26198,26227,26236,26255,26274,26331,26350,26369,26388,26407,26426,26445,26464,26502,26521,26578,26635,26683,26758,26958,27110,27140,27205,27414,27547,27561,27574,27604,27680,27699,27865,27884,27908,28003,28155,28174,28478,28491,28516,28535,28554,28592,28649,28703,28737,28822,28877,28988,29111,29124,29200,29216,29286,29352,29504,29549,29580,29618,29656,29703,29979,30026,30055,30084,30113,30142,30171,30200,30229,30258,30287,30316,30345,30374,30403,30432,30461,30490,30519,30548,30577,30606,30635,30664,30693,30722,30751,30780,30809,30838,30867,30896,30925,30954,30983,31012,31041,31070,31099,31128,31157,31186,31215,31244,31273,31302,31360,31389,31418,31447,31476,31505,31534,31563,31592,31621,31650,31679,31708,31737,31766,31795,31824,31853,31882,31911,31940,31969,31998,32027,32056,32085,32114,32143,32172,32201,32230,32259,32288,32317,32346,32375,32404,32433,32462,32491,32520,32549,32578,32607,32636,32665,32694,32723,32752,32781,32810,32839,32868,32897,32926,32955,32984,33013,33042,33071,33100,33129,33158,33187,33216,33245,33274,33303,33332,33361,33390,33419,33448,33477,33506,33535,33564,33593,33622,33651,33680,33709,33738,33767,33796,33825,33854,33883,33912,33941,33970,33999,34028,34057,34086,34115,34144,34173,34202,34231,34260,34289,34318,34347,34376,34405,34434,34463,34492,34521,34550,34579,34608,34637,34666,34695,34724,34753,34782,34811,34840,34869,34898,34927,34956,34985,35026,35068,35115,35131,35146,35189,35203,35207,35212,35225,35227,35252,35278,35286,35316,35352,35426,35438,35442,35454,35497,35502,35545,35604,35610,35663,35690,35736,35795,35818,35854,35925,35984,36043,36106,36166,36225,36231,36233,36245,36296,36355,36360,36377,36378,36414,36486,36545,36557,36604,36668,36727,36786,36857,36916,36975,37047,37106,37120,37165,37229,37252,37288,37297,37347,37418,37477,37536,37566,37608,37667,37726,37731,37790,37808,37849,37908,37979,38007,38038,38097,38106,38169,38180,38185,38228,38287,38308,38351,38410,38469,38540,38564,38599,38620,38658,38730,38740,38789,38848,38912,38927,38971,39030,39040,39045,39101,39110,39160,39219,39228,39291,39350,39409,39473,39532,39591,39600,39662,39721,39727,39780,39852,39911,39970,40102,40131,40160,40189,40218,40247,40276,40305,40334,40363,40392,40421,40450,40479,40508,40537,40566,40595,40624,40653,40682,40711,40740,40769,40798,40827,40856,40885,40914,40943,40972,41001,41030,41059,41088,41117,41146,41175,41204,41233,41262,41291,41320,41349,41378,41407,41436,41465,41494,41523,41997,42026,42055,42084,42113,42139,42141,42173,42236,42239,42263,42288,42306,42338,42413,42457,42498,42500,42538,42575,42618,42650,42695,42737,42782,42812,42861,42887,42974,43011,43026,43032,43049,43124,43188,43193,43211,43240,43256,43265,43286,43313,43323,43335,43361,43386,43388,43448,43523,43598,43633,43638,43671,43685,43713,43760,43835,43898,43922,43964,43997,44028,44072,44159,44215,44234,44309,44338,44371,44376,44396,44452,44456,44457,44471,44523,44546,44605,44607,44633,44700,44783,44870,44897,44899,44945,45020,45048,45050,45065,45093,45141,45202,45260,45291,45333,45385,45390,45420,45424,45448,45520,45577,45635,45660,45707,45739,45764,45811,45822,45894,45951,46076,46125,46151,46210,46224,46230,46321,46334,46370,46402,46453,46493,46503,46518,46537,46549,46551,46620,46707,46743,46754,46774,46837,46853,46885,46904,46922,46938,46956,46978,46995,47026,47035,47037,47051,47140,47150,47215,47232,47245,47255,47259,47332,47350,47362,47420,47454,47465,47470,47506,47507,47508,47511,47549,47553,47563,47585,47599,47604,47606,47624,47625,47654,47671,47684,47711,47737,47752,47774,47778,47797,47810,47827,47904,48020,48055,48156,48238,48255,48263,48285,48303,48305,48352,48381,48382,48429,48471,48472,48514,48529,48548,48578,48652,48668,48729,48769,48787,48807,48810,48892,48927,48933,48970,49014,49067,49074,49083,49091,49119,49160,49206,49218,49267,49293,49304,49322,49347,49432,49442,49449,49536,49550,49569,49601,49631,49647,49653,49657,49709,49714,49717,49733,49769,49798,49833,49851,49895,49903,49929,49932,49993,50063,50081,50203,50213,50457,50601,50648,50860,50940,50947,51058,51110,51154,51185,51234,51286,51354,51489,51622,51775,51837,51883,51896,51925,51939,52074,52134,52149,52182,52186,52197,52370,52392,52687,52783,52811,52852,52857,52919,53008,53021,53075,53095,53119,53161,53175,53190,53292,53310,53403,53427,53442,53552,53553,53624,53714,53734,53774,53778,53922,53933,54025,54060,54145,54158,54193,54281,54479,54498,54542,54561,54604,54655,54697,54961,54980,54993,55026,55047,55135,55236,55245,55375,55404,55433,55467,55616,55731,55757,55836,55855,55865,55894,55922,55944,56274,56286,56364,56379,56491,56633,56793,56794,56891,56978,57180,57201,57276,57385,57437,57516,57549,57734,57762,57884,58290,58295,58629,58692,58717,58824,58894,58940,58984,59093,59144,59146,59284,59543,59573,59595,59603,59967,59984,60036,60345,60377,60398,60468,60475,60706,60710,60811,60832,60876,60915,61011,61029,61062,61095,61106,61117,61150,61161,61331,61391,61461,61484,61523,61637,61729,61902,61907,61946,61968,62017,62072,62104,62236,62258,62318,62385,62428,62448,62474,62683,62687,62699,62770,62822,62872,62953,62983,63116,63121,63309,63340,63440,63442,63500,63797,63898,63951,64022,64026,64051,64134,64260,64426,64445,64645,64676,64692,64881,64923,65166,65198,65209,65220,65249,65265,65297,65314,65351,65382,65397,65504,65673,65719,65763,65993,66031,66117,66149,66372,66571,66616,66627,66634,66656,66684,66738,66785,66813,66870,66901,66990,67037,67046,67066,67127,67183,67265,67308,67343,67406,67482,67651,67656,67672,67772,67809,67881,68163,68169,68306,68478,68594,68800,68806,68889,68929,69157,69164,69283,69440,69772,69915,70060,70160,70312,70376,70500,70503,70597,70708,70735,70799,70803,70881,70912,70931,71178,71252,71270,71349,71494,71662,71732,72010,72037,72057,72109,72115,72240,72247,72307,72432,72435,72450,72578,72620,72676,72688,72816,72855,72874,73046,73081,73109,73139,73193,73350,73382,73429,73602,73604,73625,73763,73858,73962,73973,73983,74043,74170,74173,74174,74281,74483,74493,74562,74590,74597,74603,74616,74639,74686,74736,75045,75090,75115,75224,75289,75329,75411,75652,75952,75961,76040,76084,76086,76129,76152,76181,76218,76336,76347,76398,76402,76431,76472,76624,76778,76788,76872,76880,76920,77159,77466,77522,77636,77701,77917,77944,77965,77990,78026,78098,78378,78543,78581,78662,78715,78783,78836,78910,78986,79017,79021,79047,79054,79096,79163,79230,79254,79256,79303,79313,79470,79507,79531,79719,79921,79956,80063,80277,80400,80473,80490,80762,81092,81111,81131,81134,81141,81147,81227,81233,81283,81568,81570,81577,81589,81645,81712,81798,82083,82374,82423,82452,82496,82613,82827,82889,82958,83019,83173,83508,83516,83762,83927,83933,84007,84146,84151,84181,84194,84269,84450,84704,84806,85059,85131,85137,85198,85212,85218,85280,85293,85443,85453,85546,85588,85604,85681,85722,85951,86048,86072,86117,86149,86172,86186,86189,86193,86265,86275,86416,86468,86486,86566,86582,86685,86785,86861,86872,86889,86921,87048,87148,87213,87244,87289,87420,87454,87466,87504,87527,87548,87976,87996,88019,88124,88233,88342,88362,88378,88420,88432,88459,88489,88530,88572,88635,88735,88762,88772,88815,88864,88917,88929,88932,89041,89135,89269,89424,89442,89555,89721,89724,90002,90066,90470,90659,90709,90720,90775,90891,90900,91035,91078,91241,91306,91326,91424,91437,91583,91609,91668,91697,91705,91830,92022,92295,92333,92585,92717,92772,92848,92967,93172,93251,93305,93336,93371,93652,94126,94217,94312,94372,94382,94546,94549,94635,94667,94819,94824,94885,94930,95153,95279,95295,95383,95641,95648,95653,95678,95687,95854,95911,95990,96073,96264,96277,96467,96715,96783,96815,96856,96971,97136,97146,97209,97548,97608,97642,97809,97829,97837,97972,98008,98042,98227,98277,98388,98419,98476,98487,98519,98521,98572,98581,98616,98627,98686,98702,98752,98760,98857,99019,99037,99223,99271,99289,99417,99818,99860",
      "57,95,114,152,171,223,1679,1702,1721,2302,2356,2358,2536,2709,2884,3299,3310,3336,3458,3628,3746,3801,3898,4066,4130,4398,4420,4460,4502,4591,4674,4867,4887,4997,4999,5103,5417,5672,6303,6592,7173,7724,8405,8426,9495,9500,9508,9538,9559,9595,9614,9633,9652,9671,9690,9709,9764,9822,9838,10092,10621,10678,10697,10734,10759,11134,11210,11229,11381,11394,11410,11495,11571,11613,11628,11647,11856,11862,11901,11932,11970,12008,12103,12106,12141,12154,12160,12189,12217,12236,12255,12274,12323,12350,12369,12435,12464,12511,12616,12635,12673,12768,12787,12806,12815,12861,12863,12882,12921,12939,13034,13053,13091,13110,13186,13205,13224,13229,13281,13300,13357,13366,13376,13412,13438,13452,13490,13564,13604,13608,13642,13661,13680,13699,13718,13756,13775,13813,13872,13908,13927,13946,13964,14022,14079,14102,14136,14142,14155,14193,14206,14212,14231,14250,14269,14307,14326,14345,14403,14459,14497,14560,14573,14592,14649,14706,14715,14719,14731,14744,14768,14769,14786,14801,14820,14877,14911,14934,15010,15067,15086,15143,15200,15257,15295,15314,15390,15466,15496,15532,15618,15668,15675,15808,15827,15846,15884,15913,15927,15960,15979,15995,15998,16017,16024,16093,16131,16169,16226,16245,16283,16321,16340,16359,16378,16397,16473,16511,16534,16606,16699,16701,16743,16758,16815,16834,16853,16872,16882,16921,16967,17024,17043,17074,17096,17125,17157,17176,17178,17179,17214,17252,17271,17309,17347,17366,17385,17480,17499,17537,17544,17557,17575,17621,17632,17727,17746,17765,17784,17803,17822,17841,17860,17879,17974,18050,18164,18179,18202,18215,18221,18240,18278,18297,18316,18354,18373,18392,18440,18468,18487,18493,18566,18582,18601,18639,18677,18775,18808,18810,18829,18866,18886,18905,18924,18981,19029,19076,19081,19133,19152,19171,19209,19228,19285,19304,19323,19342,19380,19399,19418,19456,19551,19552,19570,19589,19650,19665,19678,19719,19722,19750,19779,19798,19817,19836,19855,19893,19912,19950,19988,20045,20083,20102,20121,20197,20216,20222,20235,20254,20292,20311,20330,20349,20368,20387,20406,20425,20444,20463,20482,20539,20558,20596,20650,20680,20710,20767,20824,20854,20881,20995,21014,21052,21109,21204,21242,21273,21280,21314,21318,21356,21394,21432,21451,21489,21495,21508,21565,21584,21622,21641,21646,21660,21679,21717,21742,21755,21774,21812,21831,21850,21869,21888,21907,21945,21964,21983,22002,22040,22078,22097,22116,22154,22186,22211,22249,22268,22287,22344,22382,22401,22420,22477,22494,22515,22534,22553,22591,22610,22667,22691,22705,22727,22743,22762,22781,22810,22821,22876,22934,22952,22971,22990,23066,23085,23140,23142,23143,23156,23180,23237,23256,23275,23292,23294,23320,23332,23371,23389,23406,23427,23446,23484,23503,23522,23560,23579,23598,23636,23693,23712,23731,23788,23807,23859,23864,23883,23921,23940,23959,23978,23997,24054,24111,24189,24225,24282,24301,24339,24358,24377,24396,24415,24434,24444,24453,24510,24529,24548,24624,24627,24662,24681,24682,24719,24833,24852,24871,24890,24904,24947,24985,25004,25023,25025,25099,25118,25152,25156,25175,25186,25194,25203,25227,25232,25270,25289,25346,25365,25393,25403,25479,25498,25517,25536,25574,25631,25650,25701,25707,25726,25764,25768,25821,25878,25916,25935,25954,25973,25992,26011,26030,26068,26087,26144,26182,26201,26258,26277,26287,26372,26391,26429,26448,26467,26486,26619,26828,26847,26885,26904,26923,26934,26999,27170,27359,27402,27531,27591,27816,27818,27920,27922,27930,27937,27949,28055,28082,28183,28329,28419,28479,28557,28633,28652,28695,28708,28905,28918,28956,29025,29032,29061,29080,29097,29108,29114,29203,29248,29380,29479,29488,29513,29611,29763,29830,29874,29922,29923,29944,30029,30058,30087,30145,30174,30203,30232,30261,30290,30319,30348,30377,30406,30435,30464,30493,30522,30551,30580,30609,30638,30667,30696,30754,30783,30812,30841,30870,30899,30928,30957,31015,31044,31073,31102,31131,31189,31218,31247,31276,31305,31334,31363,31392,31421,31450,31479,31508,31537,31566,31595,31624,31682,31711,31740,31769,31798,31827,31856,31885,31914,31943,31972,32001,32030,32059,32117,32146,32204,32233,32262,32291,32320,32349,32378,32407,32436,32465,32494,32523,32552,32581,32610,32639,32668,32697,32726,32755,32784,32813,32842,32871,32900,32929,32958,32987,33016,33045,33074,33103,33132,33161,33190,33219,33248,33306,33335,33393,33422,33451,33480,33509,33538,33567,33596,33625,33654,33683,33712,33741,33770,33799,33828,33857,33886,33915,33944,33973,34002,34031,34060,34089,34118,34147,34176,34205,34234,34263,34292,34321,34350,34379,34408,34437,34466,34495,34524,34553,34582,34611,34640,34669,34698,34727,34756,34785,34814,34843,34872,34901,34930,34959,34988,35029,35118,35154,35192,35246,35281,35288,35307,35355,35362,35366,35429,35500,35515,35548,35607,35611,35666,35671,35739,35798,35858,35868,35899,35928,35987,36074,36109,36169,36228,36229,36250,36299,36314,36417,36422,36489,36507,36548,36607,36696,36730,36749,36789,36860,36919,36922,36927,36937,36946,36949,36978,37050,37109,37168,37232,37291,37317,37350,37354,37421,37480,37539,37670,37683,37729,37793,37820,37852,37911,37924,37982,38041,38100,38103,38124,38172,38231,38354,38413,38472,38543,38606,38661,38663,38691,38704,38733,38745,38792,38872,38873,38915,38974,39033,39038,39104,39163,39222,39236,39294,39353,39412,39447,39476,39535,39594,39665,39724,39745,39855,39874,39973,40187,40216,40245,40303,40332,40361,40390,40419,40448,40477,40506,40535,40564,40593,40622,40651,40680,40709,40738,40767,40796,40825,40854,40883,40912,40941,40970,40999,41028,41057,41086,41115,41144,41173,41202,41231,41260,41318,41347,41376,41405,41434,41463,41492,41521,41988,41993,42051,42080,42109,42168,42227,42243,42300,42318,42408,42536,42543,42555,42612,42633,42643,42645,42720,42725,42792,42882,42957,42967,43029,43033,43039,43194,43266,43324,43356,43357,43401,43415,43431,43499,43503,43593,43668,43717,43729,43740,43745,43826,43830,43864,43877,43905,43959,43968,43977,43990,44060,44067,44104,44137,44142,44145,44184,44197,44214,44263,44304,44341,44359,44373,44379,44392,44437,44459,44465,44508,44516,44522,44540,44541,44575,44587,44616,44630,44658,44665,44671,44688,44778,44826,44851,44853,44904,44925,44982,45015,45047,45086,45128,45134,45170,45195,45244,45253,45292,45326,45441,45513,45570,45628,45700,45757,45815,45887,45944,45946,46027,46055,46056,46067,46082,46083,46119,46129,46131,46152,46245,46330,46371,46385,46439,46443,46449,46469,46470,46519,46633,46708,46732,46792,46876,46886,46890,46970,46987,46991,47128,47236,47390,47395,47416,47426,47498,47513,47518,47532,47552,47610,47656,47659,47663,47682,47723,47729,47840,47870,47942,48147,48234,48277,48361,48384,48403,48404,48411,48439,48452,48475,48505,48509,48608,48630,48643,48674,48693,48695,48715,48763,48802,48808,48890,48978,48987,49023,49029,49045,49048,49070,49087,49110,49115,49122,49131,49139,49229,49286,49314,49343,49358,49373,49374,49398,49508,49514,49517,49528,49543,49585,49586,49609,49616,49648,49708,49710,49782,49787,49868,49911,49936,49979,50082,50424,50456,50472,50627,50729,50875,50975,51035,51065,51272,51335,51412,51632,51719,51780,51827,51830,51938,52083,52121,52176,52293,52341,52345,52354,52390,52507,52837,52956,53040,53116,53265,53269,53317,53386,53392,53751,53854,53899,54006,54150,54236,54389,54417,54470,54480,54560,54594,54644,54736,54738,54776,54821,54896,55032,55202,55207,55240,55352,55378,55515,55800,55805,55898,55977,56137,56144,56284,56335,56367,56484,56639,56715,56909,56924,57000,57066,57229,57459,57478,57639,57764,57766,58060,58218,58346,58435,58513,58588,58667,58732,58766,58914,58935,58978,59075,59112,59121,59217,59261,59344,59368,59476,59500,59581,59732,59800,59898,59935,59992,60199,60256,60306,60440,60474,60687,60688,60707,60771,60814,60830,60879,61274,61308,61494,61547,61562,61758,61786,61855,61887,61963,62100,62133,62186,62340,62528,62609,62651,62677,62808,62821,62856,62928,62961,63133,63164,63326,63688,63739,63850,63939,63978,64033,64067,64096,64115,64194,64215,64289,64302,64480,64562,64594,64610,64621,64627,64677,64721,64952,65050,65185,65268,65311,65335,65757,65780,65992,66083,66102,66116,66233,66260,66262,66307,66313,66316,66407,66574,66587,66606,66702,66783,66930,66991,67005,67055,67090,67180,67187,67254,67260,67316,67360,67369,67564,67570,67600,68231,68331,68423,68747,68750,68820,68980,69092,69111,69220,69288,69339,69379,69475,69479,69632,69688,69697,69720,69746,69796,69845,69849,70049,70251,70295,70320,70465,70486,70509,70522,70722,70731,70832,70840,70990,71043,71049,71130,71147,71214,71292,71296,71415,71471,71486,71549,71607,71622,71686,71781,71864,71885,71913,71934,72034,72154,72156,72267,72420,72585,72611,72823,72952,72980,73025,73175,73292,73357,73388,73470,73471,73608,73632,73797,74048,74324,74334,74345,74613,74853,74909,74966,75017,75020,75069,75498,75522,75544,75710,75854,75919,75967,76003,76073,76183,76197,76433,76564,77007,77036,77055,77095,77115,77147,77169,77185,77204,77209,77213,77239,77298,77428,77560,77587,77590,77616,77693,77757,77770,77834,77926,78032,78086,78110,78285,78557,78585,78587,78751,78787,78804,78954,79007,79049,79098,79141,79193,79260,79314,79404,79574,79578,79620,79691,79724,79787,79808,79860,79886,79942,80115,80174,80282,80307,80359,80459,80517,80542,80561,80650,80663,80684,80745,80817,80954,80975,81027,81056,81308,81686,81865,81877,81987,82030,82086,82334,82351,82383,82413,82584,82593,83121,83180,83244,83348,83505,83580,83599,83648,83682,83788,83950,84021,84052,84075,84219,84226,84296,84310,84422,84475,84483,84523,84570,84571,84573,84596,84620,84629,84675,84690,84850,84924,84953,85008,85017,85139,85210,85223,85278,85301,85475,85528,85570,85728,85824,85838,85841,86148,86208,86250,86294,86332,86335,86359,86374,86390,86421,86456,86793,86794,86925,87339,87390,87538,87539,87629,87688,87718,87764,87865,87932,87962,88142,88150,88211,88239,88258,88321,88449,88650,88698,89013,89025,89064,89109,89115,89137,89214,89253,89335,89341,89363,89470,89518,89524,89536,89548,89559,89752,89873,89937,90011,90168,90174,90229,90308,90311,90336,90359,90363,90374,90378,90403,90498,90518,90542,90557,90619,90692,90710,90731,90784,90851,90899,90932,90956,90962,91119,91278,91331,91354,91383,91625,91693,91730,91849,91904,91905,91912,91976,92049,92291,92325,92462,92735,92819,92925,92927,92957,92962,92980,93094,93159,93392,93422,93434,93623,93670,93694,93845,94183,94507,94585,94603,94639,94728,94749,94868,94999,95010,95040,95059,95368,95406,95534,95544,95585,95661,95683,95745,95881,95917,95986,96016,96029,96140,96166,96206,96250,96312,96394,96407,96585,96604,96702,96875,96957,97021,97272,97451,97528,97554,97600,97657,97721,97749,97854,97863,97882,97899,97998,98071,98167,98269,98436,98470,98494,98522,98595,98680,98776,98858,99157,99235,99237,99242,99425,99464,99505,99576,99593,99624,99709,99720,99828,99872,99874,99944,99996",
      "1,279,342,372,411,442,495,529,536,570,604,619,671,682,819,838,935,941,976,1001,1329,1350,1356,1399,1400,1580,1596,1708,1871,2277,2607,3117,3271,3831,3919,4158,5033,5095,5142,5170,5261,5342,5355,5357,5389,5478,5616,5803,5882,6020,6055,6105,6119,6333,6488,6537,6539,6612,6656,6670,6695,6718,6796,6823,6839,6853,6995,7039,7196,7211,7242,7342,7371,7464,7481,7515,7545,7566,7962,7985,8052,8138,8330,8389,8531,8698,8720,8813,8825,9012,9021,9059,9165,9200,9241,9256,9321,9392,9439,9465,9557,9564,9863,9932,9956,9991,10015,10103,10249,10263,10293,10527,10669,10701,10799,10824,10940,11028,11041,11057,11405,11433,11501,11771,11829,11864,11919,12311,12322,12409,12455,12539,12576,12600,12643,12688,12848,12972,12992,13047,13107,13181,13195,13275,13554,13579,13623,13880,14347,14437,14488,14552,14594,14596,14644,14830,14889,14983,14999,15159,15189,15366,16105,16219,16457,16725,17162,17269,17411,17931,18007,18053,18169,18465,18777,18794,19134,19683,19947,20081,20096,20653,20676,20709,20768,20783,21454,21496,21663,22026,22232,22255,22790,22951,23118,23141,23192,23702,24481,24714,24794,24817,24856,24946,24996,25049,25068,25250,25608,25635,26206,26624,26850,26866,26938,27218,27257,27479,27639,27761,27805,27871,27880,28175,28224,28291,28452,28510,28696,28699,28849,28892,29019,29252,29585,29601,29669,29837,29880,29963,31532,32028,32074,35827,36016,36457,36493,36806,36810,37136,37199,37434,37578,37617,37679,37803,38175,38422,38436,38510,38681,38928,39046,39263,39545,39666,39866,39883,45369,45431,45497,45783,45873,45876,45928,45992,50173,50197,50275,50347,50350,50364,50421,50510,50550,50581,50600,50760,50818,50946,51027,51097,51306,51317,51337,51465,51770,51805,51917,52014,52070,52072,52095,52270,52434,52575,52626,52703,52717,52825,52830,52848,52849,52872,52951,52987,53056,53243,53429,53445,53542,53636,53739,53747,53812,53913,53964,54079,54125,54183,54379,54493,54603,54632,54661,54711,54928,54931,54957,54963,55003,55139,55144,55271,55459,55479,55660,55753,55886,55915,56141,56326,56481,56486,56510,56589,57078,57243,57297,57435,57458,57636,57655,57788,57834,58079,58133,58136,58143,58207,58223,58263,58309,58510,58568,58628,58763,59012,59019,59353,59375,59382,59414,59561,59859,60024,60046,60129,60160,60250,60388,60470,60551,60573,60658,60743,60794,60908,60932,60979,61071,61085,61091,61122,61137,61149,61311,61316,61339,61486,61579,61647,61661,61678,61755,61816,61836,62142,62175,62180,62244,62283,62513,62628,62732,62761,62969,63059,63208,63214,63261,63341,63344,63478,63583,63590,63680,63732,64041,64119,64163,64203,64255,64491,64601,64650,64691,64698,64730,64974,65095,65470,65541,65931,65974,66017,66038,66068,66091,66243,66269,66287,66305,66322,66324,66415,66434,66469,66517,66599,66661,66664,66787,66796,66876,67258,67311,67337,67341,67449,67468,67541,67562,67727,67874,67908,67922,68021,68024,68067,68203,68316,68415,68522,68562,68579,68716,68878,68923,68943,69154,69170,69197,69209,69214,69240,69242,69291,69320,69350,69357,69361,69464,69580,69627,69647,69812,69822,69840,69878,69947,70079,70113,70140,70154,70176,70188,70431,70455,70553,70566,70611,70655,70853,70911,70978,71010,71213,71217,71230,71253,71255,71449,71690,71750,72055,72058,72073,72077,72084,72210,72228,72323,72405,72616,72619,72681,72734,72765,72911,72989,73118,73186,73289,73347,73444,73687,73713,73736,73849,73951,73982,73996,74042,74286,74407,74443,74453,74490,74542,74594,74700,74774,74846,75002,75078,75186,75198,75553,75714,75886,75890,75968,75986,76019,76033,76059,76105,76175,76243,76314,76414,76493,76566,76604,76664,76696,76786,76865,76962,76968,77030,77068,77233,77240,77282,77286,77318,77439,77529,77599,77715,77755,77825,77830,78054,78126,78220,78246,78302,78333,78347,78373,78380,78498,78592,78617,78701,78841,78851,78875,78886,79013,79140,79321,79447,79502,79731,79818,79900,79994,80175,80352,80404,80502,80577,80624,81036,81086,81088,81154,81536,81579,81695,81696,81795,81871,81965,82015,82109,82161,82169,82174,82250,82482,82489,82502,82587,82639,82692,82695,82742,82754,82894,83010,83200,83328,83456,83476,83479,83494,83517,83529,83548,83551,83667,83697,83748,83849,83984,84017,84148,84161,84210,84253,84366,84674,84751,84868,84903,85052,85205,85531,85548,85647,85776,85821,85914,86086,86134,86376,86581,86746,86933,86943,87010,87016,87135,87222,87355,87453,87603,87644,87682,87752,87761,87873,87886,88016,88098,88162,88209,88375,88431,88484,88510,88582,88593,88599,88646,88874,89130,89142,89343,89379,89426,89523,89619,89631,89738,89757,89929,90044,90214,90425,90473,90488,90554,90578,90588,90640,90802,90860,90990,91000,91111,91159,91178,91263,91462,91527,91550,91708,91740,91753,91855,91862,91983,92004,92141,92175,92257,92260,92304,92473,92567,92655,92691,92712,92716,92776,92862,92968,92972,93178,93328,93373,93443,93583,93598,93701,94167,94267,94326,94440,94664,94822,94904,95038,95049,95098,95557,95591,95660,95711,95735,95747,95786,95814,95837,95852,95857,95860,96009,96222,96368,96374,96522,96605,96681,96687,96740,96784,96798,96805,96816,96871,96886,96983,97192,97292,97324,97549,97680,97700,97751,97895,98013,98044,98107,98263,98279,98283,98284,98334,98443,98568,98640,98642,98664,98684,98755,98937,99184,99194,99302,99358,99386,99488,99595,99832",
      "102,149,178,217,460,496,1656,1669,1706,1728,1759,1792,1904,1917,1933,1935,1944,1983,1987,1995,2028,2062,2119,2139,2150,2189,2198,2224,2243,2266,2308,2330,2348,2393,2408,2425,2426,2500,2511,2530,2534,2667,2702,2720,2741,2793,2804,2823,2890,2968,3004,3009,3012,3022,3025,3026,3037,3052,3057,3060,3062,3064,3094,3101,3104,3111,3132,3174,3197,3208,3245,3295,3302,3318,3353,3411,3427,3441,3447,3451,3453,3454,3472,3499,3508,3516,3523,3525,3556,3558,3568,3578,3591,3604,3636,3640,3651,3657,3676,3748,3799,3865,3876,3885,3916,3929,3964,3968,3969,3970,3996,4012,4019,4037,4044,4059,4076,4088,4095,4113,4133,4138,4155,4162,4179,4186,4195,4209,4248,4250,4302,4315,4318,4338,4355,4361,4372,4397,4452,4537,4539,4587,4640,4665,4708,4722,4752,4780,4793,4805,4868,4883,4893,4924,4926,4943,4950,4951,4961,4964,4968,5076,5313,5580,5605,5755,5766,6159,6275,6327,6342,6352,6476,6617,6647,6767,6805,6890,7088,7093,7102,7136,7147,7221,7320,7374,7425,7568,7600,7601,7657,7684,7691,7790,7831,7968,8011,8121,8218,8222,8285,8317,8350,8404,8422,8458,8467,8506,8537,8622,8729,8769,8784,8820,8823,8835,8848,8862,8865,8896,8920,8951,8962,8995,9034,9083,9166,9191,9269,9325,9393,9466,9613,9659,9687,9692,9851,9869,9952,9957,9972,10034,10134,10255,10309,10319,10422,10428,10445,10554,11004,11056,11123,11334,11355,11362,11374,11375,11544,11698,12196,12249,12296,12326,12359,12370,12516,12633,12685,12880,12994,13291,13592,13745,14050,14210,14232,14286,14319,14821,15117,15303,15329,15487,15714,15781,15813,15814,15903,15921,16067,16150,16196,16486,16533,16543,16694,16780,17244,17250,17293,17336,17355,17372,17399,17487,17514,17526,17550,17683,17734,17780,17867,17901,17905,17963,18100,18274,18415,18560,18565,18612,18632,18688,18837,18956,19053,19109,19222,19474,19554,19774,19837,19844,20054,20105,20242,20287,20341,20706,20937,20941,20951,21098,21100,21289,21370,21373,21409,21503,21650,21793,21804,21881,22067,22340,22432,22463,22472,22484,22779,22872,22879,22959,22999,23055,23252,23253,23344,23454,23461,23497,23632,23802,23891,23898,23948,24126,24160,24372,24472,24551,24564,24601,24633,24905,24945,25031,25286,25385,25510,25715,25791,25945,25984,26018,26067,26091,26468,26488,26534,26543,26560,26570,26585,26597,26610,26618,26665,26830,26852,26888,26942,26955,27012,27034,27070,27128,27164,27207,27363,27466,27474,27475,27492,27530,27564,27571,27576,27582,27586,27596,27607,27612,27642,27724,27731,27740,27770,27778,27794,27795,27811,27868,27960,27987,28038,28062,28078,28085,28094,28139,28141,28207,28286,28308,28341,28347,28414,28427,28464,28488,28489,28493,28531,28558,28568,28603,28607,28626,28653,28805,28817,28840,28856,28964,28980,28992,29009,29084,29119,29126,29143,29153,29156,29179,29208,29224,29264,29284,29289,29290,29372,29512,29529,29550,29563,29607,29628,29661,29682,29685,29689,29790,29819,29876,29884,29947,30021,30050,30079,30108,30137,30166,30195,30224,30253,30282,30311,30340,30369,30398,30427,30446,30456,30485,30514,30543,30601,30630,30688,30717,30746,30775,30804,30833,30862,30891,30920,30949,30978,31005,31007,31036,31065,31094,31152,31181,31210,31239,31268,31297,31326,31353,31355,31384,31413,31442,31471,31500,31515,31529,31558,31587,31616,31645,31674,31703,31761,31819,31848,31877,31906,31935,31964,31993,32022,32051,32080,32109,32138,32167,32196,32225,32254,32283,32312,32341,32370,32399,32428,32457,32486,32515,32544,32573,32602,32631,32660,32689,32718,32747,32776,32805,32834,32863,32892,32921,32950,32979,33008,33037,33066,33095,33124,33153,33182,33211,33240,33269,33298,33327,33356,33385,33414,33443,33472,33501,33530,33559,33588,33617,33646,33675,33704,33733,33762,33791,33820,33878,33907,33936,33965,33994,34023,34052,34081,34110,34168,34197,34226,34255,34284,34313,34342,34371,34400,34429,34458,34487,34516,34545,34574,34603,34632,34661,34690,34719,34748,34777,34806,34835,34864,34893,34922,34951,35021,35040,35042,35044,35075,35085,35110,35122,35152,35184,35194,35229,35235,35273,35297,35347,35421,35441,35492,35540,35599,35658,35731,35745,35759,35790,35804,35817,35849,35882,35920,35947,35958,35979,36038,36054,36101,36161,36189,36220,36291,36323,36327,36350,36367,36409,36481,36540,36562,36599,36628,36629,36663,36722,36752,36781,36852,36911,36929,36943,36970,37042,37063,37101,37160,37182,37192,37224,37342,37413,37472,37491,37493,37496,37500,37531,37603,37662,37686,37721,37737,37785,37823,37844,37871,37903,37974,37990,37999,38004,38033,38092,38164,38223,38247,38282,38298,38346,38376,38405,38464,38495,38535,38551,38594,38625,38653,38725,38784,38822,38843,38907,38925,38966,39025,39034,39051,39096,39112,39155,39186,39214,39286,39345,39404,39436,39468,39499,39527,39586,39608,39657,39681,39716,39775,39808,39847,39861,39869,39885,39906,39965,40126,40155,40184,40213,40242,40271,40300,40329,40358,40387,40416,40445,40474,40503,40515,40532,40561,40590,40619,40648,40706,40735,40764,40793,40822,40851,40880,40909,40938,40967,40996,41025,41054,41083,41112,41141,41170,41199,41228,41286,41344,41373,41402,41431,41460,41489,41518,41547,41802,42017,42046,42075,42104,42133,42193,42283,42310,42316,42358,42371,42381,42433,42438,42445,42497,42520,42572,42595,42757,42832,42843,42866,42907,42915,42933,42994,43024,43069,43117,43144,43171,43231,43239,43249,43306,43310,43340,43381,43397,43414,43425,43439,43468,43502,43618,43624,43705,43741,43780,43855,43862,43887,43904,43942,44017,44092,44136,44179,44189,44254,44329,44416,44491,44566,44580,44584,44620,44653,44668,44696,44728,44803,44814,44861,44890,44914,44916,44965,44996,45008,45040,45053,45113,45161,45222,45280,45353,45363,45410,45429,45442,45468,45472,45540,45597,45655,45702,45727,45784,45842,45856,45914,45971,45996,46042,46054,46065,46104,46117,46127,46157,46172,46183,46191,46206,46222,46240,46294,46308,46317,46328,46337,46406,46433,46438,46464,46475,46485,46497,46522,46536,46572,46578,46652,46665,46686,46688,46751,46767,46850,46854,46865,46868,46880,46897,47011,47033,47054,47086,47092,47118,47137,47182,47185,47186,47281,47287,47305,47315,47387,47388,47418,47424,47442,47463,47489,47562,47590,47598,47613,47669,47836,47845,47900,47935,47944,47945,47989,48019,48026,48030,48041,48044,48056,48074,48090,48139,48165,48183,48189,48213,48226,48310,48321,48330,48515,48517,48571,48573,48639,48647,48675,48683,48684,48686,48735,48757,48778,48798,48799,48805,48812,48825,48837,48864,48869,48876,48897,48901,48935,48937,48943,48950,48966,48975,48976,49019,49053,49123,49145,49194,49344,49368,49382,49489,49495,49512,49608,49630,49685,49692,49695,49701,49724,49747,49772,49781,49823,49824,49830,49839,49860,49908,49915,49924,49927,49950,50024,50118,50158,50259,50279,50301,50382,50412,50415,50454,50531,50641,50675,50717,50766,51096,51173,51320,51420,51601,51633,51651,51709,51784,51935,52034,52078,52135,52188,52191,52217,52227,52247,52261,52273,52539,52558,52645,52726,52782,52906,53212,53297,53311,53330,53336,53395,53566,53612,53707,54126,54157,54173,54198,54222,54371,54376,54510,54523,54694,54835,54854,55059,55064,55140,55189,55249,55277,55318,55443,55544,55571,55611,55752,55794,55839,55962,56072,56164,56190,56238,56305,56324,56385,56395,56407,56462,56475,56615,56692,56761,56839,56974,56996,57040,57139,57143,57147,57150,57347,57357,57451,57469,57473,57484,57699,57923,57972,58274,58482,58498,58523,58590,58673,58718,58781,58789,58802,58887,58900,58906,58996,59109,59364,59508,59758,59863,60080,60082,60186,60268,60382,60506,60585,60645,60661,60728,60851,61006,61107,61143,61198,61333,61367,61456,61584,61827,62006,62053,62077,62095,62230,62250,62253,62295,62372,62497,62500,62576,62622,62685,62689,62701,62778,62857,62894,62899,63148,63258,63275,63389,63417,63503,63595,63653,63805,63821,64128,64205,64226,64256,64343,64353,64591,64757,64883,64959,65120,65193,65196,65197,65273,65298,65340,65400,65424,65513,65557,65574,65689,65785,65838,65859,65874,65886,65889,65936,66075,66096,66168,66193,66200,66301,66463,66488,66512,66593,66693,66935,66950,66973,67017,67036,67079,67109,67169,67185,67281,67445,67537,67593,67622,67743,67747,67760,67774,67797,67806,67832,67911,68059,68271,68469,68845,68906,69128,69195,69257,69356,69409,69613,69750,70067,70481,70530,70547,70607,70706,70733,70749,70787,70841,70903,70927,70944,70988,71001,71113,71146,71277,71344,71447,71533,71794,71829,71830,71898,72035,72180,72353,72362,72367,72474,72506,72824,72898,72994,73152,73205,73400,73413,73414,73434,73454,73456,73490,73553,73619,73628,73646,73678,73714,73750,73877,73889,74035,74143,74221,74252,74279,74339,74531,74549,74605,74622,74632,74649,74759,74899,74914,74983,75044,75048,75054,75285,75308,75344,75352,75387,75535,75649,75721,75756,75768,75879,75894,76240,76306,76348,76415,76438,76443,76446,76474,76485,76543,77076,77130,77383,77389,77410,77506,77612,77613,77699,77705,77868,78021,78097,78142,78370,78417,78549,78653,78655,78677,78681,79169,79203,79292,79387,79487,79534,79727,79753,79881,79933,80327,80405,80408,80448,80560,80665,80713,80719,80726,80758,80793,80797,80926,80950,80967,80969,80996,81013,81041,81115,81198,81277,81337,81360,81529,81569,81682,81750,81762,81848,81915,81924,81931,81955,82164,82259,82354,82582,82786,83016,83025,83044,83161,83198,83234,83254,83327,83333,83419,83423,83425,83513,83650,83677,83792,83847,83936,84131,84186,84382,84402,84403,84493,84600,84923,85110,85180,85211,85339,85370,85485,85630,85689,85740,85766,85777,85822,85903,85919,85938,85997,86046,86066,86108,86132,86314,86341,86344,86472,86592,86686,87126,87134,87205,87214,87424,87428,87481,87574,87685,87694,88131,88137,88248,88249,88380,88395,88502,88665,88719,88877,88901,88925,89091,89118,89162,89295,89647,89786,89931,89979,89996,90025,90121,90246,90357,90382,90394,90496,90515,90652,90711,90848,90879,90987,90992,91018,91211,91217,91223,91266,91415,91598,91637,91780,91791,91798,91960,92053,92204,92232,92301,92308,92330,92519,92648,92703,92804,92918,93108,93209,93210,93263,93477,93737,93778,93864,93886,93893,93913,93925,94131,94172,94240,94427,94470,94558,94677,94683,94689,94829,94874,94931,95253,95485,95520,95532,95652,95749,95791,95849,96063,96086,96491,96528,96580,96673,96757,96879,96923,96964,96979,97003,97121,97173,97303,97343,97405,97407,97607,97645,97703,97910,97985,98047,98188,98243,98348,98370,98460,98533,98594,98705,98944,98966,99057,99326,99461,99468,99632,99753,99839,99921",
      "41,60,298,301,308,316,369,380,398,454,481,489,539,546,572,573,596,749,762,823,865,880,1067,1099,1102,1118,1151,1301,1303,1396,1486,1553,1774,1897,2024,2206,3743,4576,4970,5072,5085,5182,5461,5561,5628,5680,5843,5866,5884,5890,5897,6090,6130,6236,6314,6448,6569,6606,6613,6749,6774,6781,6897,6930,6989,6991,7003,7011,7015,7034,7144,7262,7402,7450,7527,7567,7635,7647,7653,7679,7712,7886,7935,7943,8013,8086,8108,8152,8207,8236,8338,8339,8670,8838,8846,8936,8948,9051,9101,9151,9234,9245,9282,9320,9379,9468,9809,9945,9948,10025,10032,10154,10158,10277,10345,10362,10593,10853,10919,10922,11008,11124,11178,11224,11316,11357,11573,11787,11832,11833,11845,11849,11939,11940,11945,12020,12022,12351,12772,13113,13433,14026,14189,14589,14600,14713,14953,15089,15446,16626,16871,17051,17232,17536,17783,18062,18252,18521,18855,18899,18912,19283,20174,20198,20293,20597,20668,20769,20794,20802,20879,20907,21015,21290,21338,21448,21630,21655,21736,21778,22635,22714,22845,23299,23351,23366,23568,23601,23705,24067,25153,25559,25566,25636,26356,26410,26777,26794,26879,26962,27094,27329,27460,28303,28702,28934,29120,29407,29543,29576,29868,29983,35041,35608,35636,36241,36491,37175,37234,37363,37696,37739,37855,37909,38137,38303,38320,38549,38758,39503,42160,42998,43051,43834,45871,45947,50042,50120,50255,50286,50289,50492,50518,50598,50672,50702,50706,50754,50755,50780,50785,50804,50845,51002,51039,51105,51368,51395,51424,51594,52003,52130,52144,52170,52198,52332,52447,52761,52804,53044,53085,53088,53107,53132,53302,53447,53502,53504,53570,53678,53745,53757,53810,53878,54134,54377,54426,54490,54573,54728,54758,54798,54874,55093,55158,55176,55183,55198,55329,55398,55687,55688,55862,55978,55983,56043,56068,56169,56319,56402,56452,56525,56617,56645,56663,56799,56838,56864,56917,57089,57226,57360,57361,57501,57505,57837,57924,57933,58116,58148,58201,58418,58420,58475,58544,58551,58674,58715,58834,58872,58917,58956,59078,59110,59127,59453,59501,59544,59614,59676,59774,59932,59961,60077,60096,60247,60301,60336,60500,60600,60672,60680,60750,60801,61173,61220,61378,61382,61471,61516,61649,61656,61665,61724,61835,61842,61904,62223,62235,62255,62317,62394,62449,62649,62715,62736,62782,62814,62898,62916,63009,63021,63038,63097,63135,63530,63581,63624,63720,63804,63946,64202,64423,64592,64616,64913,64940,65084,65132,65462,65544,65774,65975,66029,66054,66124,66258,66331,66519,66561,66667,66679,66804,66938,66988,67021,67217,67283,67605,67786,67814,67903,67914,68079,68134,68173,68176,68220,68267,68295,68363,68454,68499,68501,68521,68548,68564,68587,68618,68642,68646,68758,68823,68857,68872,68875,68945,68959,68992,69114,69139,69285,69517,69536,69741,69770,70055,70213,70230,70408,70460,70679,70748,70898,71023,71050,71053,71112,71124,71184,71192,71261,71298,71519,71528,71543,71576,71645,71716,71745,71812,71998,72148,72149,72215,72270,72273,72678,72721,72756,72955,73070,73078,73088,73156,73190,73258,73277,73541,73829,74011,74255,74268,74461,74484,74577,74675,74698,74738,74789,74883,74964,74967,74992,75213,75230,75294,75885,75909,75926,75988,76318,76460,76552,76553,76576,76599,76600,76627,76654,76722,76980,77174,77413,77519,77633,77638,77972,78053,78073,78079,78100,78167,78202,78205,78255,78364,78562,78609,78687,78698,78761,78790,79065,79170,79212,79250,79323,79745,79764,79806,79868,80008,80018,80019,80091,80220,80306,80323,80639,80706,80909,81032,81135,81401,81595,81687,81858,81921,81938,82198,82246,82359,82384,82385,82471,82561,82664,82697,82753,82871,82937,83175,83178,83278,83459,83495,83579,83609,83678,83743,83795,83949,83999,84044,84246,84503,84661,84764,84856,84913,85035,85064,85103,85129,85148,85228,85315,85343,85384,85422,85550,85590,85644,85856,85920,86242,86313,86427,86463,86521,86940,86959,87055,87064,87093,87125,87139,87154,87186,87285,87371,87423,87472,87474,87640,87703,87723,87743,87825,87876,88023,88139,88330,88373,88438,88457,88553,88718,88761,88888,88910,88919,89000,89398,89420,89451,89482,89529,89845,89948,89968,90014,90021,90056,90083,90096,90113,90127,90231,90368,90504,90535,90570,90750,90984,91162,91253,91451,91593,91676,91776,91966,91993,92166,92224,92283,92419,92461,92465,92499,92591,92672,92896,93032,93223,93271,93447,93545,93588,93624,93635,93709,93831,93912,93972,94018,94070,94185,94261,94349,94351,94368,94419,94428,94660,94731,94894,94914,95127,95144,95221,95227,95294,95347,95396,95413,95417,95472,95501,95556,95599,95659,95807,95831,95847,95858,95912,95913,95964,96036,96167,96376,96543,96576,96674,96870,97112,97166,97177,97270,97326,97330,97352,97521,97675,97713,97737,97744,97768,97773,97862,98020,98115,98123,98148,98390,98398,98414,98438,98547,98704,98768,98847,98862,98863,98880,98987,98994,99180,99542,99678,99846,99854,99979",
      "52,58,120,135,181,322,708,874,1181,1321,1384,1484,1509,1670,1838,1931,1938,1947,1948,1952,2006,2018,2019,2055,2121,2138,2184,2268,2373,2374,2418,2443,2451,2495,2512,2528,2531,2533,2576,2645,2735,2738,2740,2744,2790,2858,2885,2893,2896,2900,2911,2918,2934,2984,3020,3033,3043,3067,3108,3115,3188,3278,3404,3405,3417,3439,3464,3473,3490,3498,3524,3527,3546,3570,3577,3663,3685,3691,3711,3721,3772,3791,3815,3825,3826,3934,3936,3959,3975,3979,3981,3998,4036,4040,4049,4058,4084,4094,4108,4112,4152,4194,4230,4236,4246,4269,4283,4290,4297,4379,4474,4482,4487,4491,4496,4498,4501,4507,4540,4552,4553,4577,4590,4595,4629,4646,4659,4686,4745,4784,4830,4854,4857,4928,4940,5152,5289,5402,5493,5585,5594,5623,5833,5848,5873,5960,5983,5986,6066,6085,6092,6178,6196,6208,6261,6291,6300,6325,6347,6375,6416,6419,6467,6501,6567,6639,6733,6761,6898,6948,6984,7029,7059,7076,7110,7129,7177,7213,7225,7246,7251,7298,7303,7409,7442,7460,7465,7632,7677,7716,7786,7792,7840,7938,7997,8015,8045,8157,8265,8431,8438,8476,8533,8557,8564,8574,8591,8761,8764,8776,8817,8861,9044,9079,9094,9193,9255,9534,9635,9673,9727,9818,9894,9969,10109,10229,10259,10270,10390,10451,10560,11168,11219,11260,11305,11453,11484,11702,11925,11958,12000,12190,12342,12357,12656,13183,13314,13691,13732,13804,13863,14032,14070,14207,14287,14461,14510,14572,14822,14852,14855,15008,15056,15328,15342,15654,15816,16125,16339,16370,16416,16504,16682,16836,16892,16936,17061,17083,17138,17177,17278,17420,17442,17471,17481,17577,17641,17844,17877,17888,18133,18292,18321,18375,18575,18667,18973,19018,19069,19603,19654,19657,19741,19754,19813,19874,19907,20024,20367,20433,20522,20651,20758,20990,21086,21427,21501,21506,21784,21870,21889,21968,21985,22260,22280,22361,22380,22468,22525,23296,23323,23421,23919,23963,24147,24198,24488,24775,25239,25258,25400,25424,25503,25576,25725,26211,26352,26487,26498,26507,26533,26549,26601,26606,26607,26655,26738,26767,26771,26778,26831,26844,26992,27006,27011,27022,27036,27044,27097,27103,27130,27142,27151,27180,27230,27263,27331,27399,27446,27507,27512,27515,27533,27562,27590,27613,27614,27624,27632,27640,27697,27726,27755,27769,27853,27887,27924,27962,27963,27972,27997,28042,28057,28072,28098,28116,28151,28209,28222,28223,28236,28254,28255,28260,28301,28314,28343,28344,28407,28428,28438,28495,28498,28506,28513,28539,28564,28586,28595,28599,28629,28634,28636,28645,28663,28665,28722,28783,28797,28957,28968,29015,29083,29105,29116,29146,29166,29184,29196,29217,29218,29237,29238,29313,29336,29387,29448,29470,29507,29520,29536,29546,29557,29584,29587,29644,29645,29658,29659,29691,29695,29706,29726,29740,29754,29762,29764,29887,29895,29920,30018,30047,30076,30105,30134,30163,30192,30221,30250,30279,30308,30337,30366,30395,30424,30453,30482,30540,30569,30598,30627,30656,30685,30714,30743,30772,30801,30830,30859,30888,30917,30975,31004,31033,31091,31120,31149,31178,31207,31236,31265,31294,31323,31352,31381,31410,31439,31468,31497,31526,31555,31584,31613,31642,31671,31700,31729,31758,31787,31816,31845,31874,31903,31932,31961,31990,32048,32077,32135,32164,32193,32222,32251,32280,32309,32338,32396,32425,32454,32483,32512,32541,32570,32599,32628,32657,32686,32715,32744,32773,32802,32831,32860,32889,32918,32947,32976,33005,33034,33063,33092,33121,33150,33179,33208,33237,33266,33295,33324,33353,33382,33411,33440,33469,33498,33527,33556,33585,33614,33643,33672,33701,33730,33759,33788,33817,33846,33875,33904,33933,33962,33991,34020,34049,34078,34107,34136,34165,34194,34223,34252,34281,34310,34339,34368,34397,34426,34455,34484,34513,34542,34571,34600,34629,34658,34687,34716,34745,34774,34803,34861,34890,34919,34948,34977,35018,35032,35051,35070,35107,35138,35181,35230,35270,35344,35363,35418,35468,35489,35503,35537,35557,35560,35596,35655,35710,35728,35787,35809,35846,35917,35976,36035,36098,36158,36217,36238,36288,36311,36347,36406,36433,36438,36478,36537,36596,36660,36719,36740,36778,36849,36908,36967,37039,37098,37157,37169,37221,37255,37280,37339,37360,37410,37428,37469,37489,37528,37556,37600,37659,37718,37782,37841,37900,37971,38030,38060,38089,38161,38220,38279,38343,38373,38402,38461,38532,38591,38650,38686,38722,38781,38840,38904,38963,39022,39042,39060,39093,39152,39211,39283,39342,39401,39465,39524,39583,39654,39713,39754,39772,39844,39903,39962,40115,40144,40173,40202,40231,40260,40289,40318,40347,40376,40405,40434,40463,40492,40521,40550,40579,40608,40637,40666,40695,40724,40753,40782,40811,40840,40869,40898,40927,40956,40985,41014,41043,41072,41101,41130,41159,41188,41246,41275,41304,41333,41362,41391,41420,41449,41536,42006,42035,42064,42093,42122,42138,42153,42161,42182,42216,42241,42272,42347,42382,42399,42422,42469,42509,42558,42584,42626,42659,42682,42692,42707,42712,42746,42821,42896,42923,42983,43012,43015,43058,43133,43160,43165,43220,43295,43316,43332,43337,43370,43396,43472,43532,43570,43648,43694,43750,43769,43844,43878,43886,43931,43991,44006,44049,44063,44081,44133,44168,44223,44227,44243,44318,44352,44405,44480,44518,44555,44595,44613,44642,44717,44758,44792,44879,44954,45002,45029,45102,45150,45193,45211,45269,45342,45399,45436,45457,45492,45512,45529,45552,45586,45644,45668,45716,45740,45756,45773,45831,45852,45903,45960,46122,46192,46279,46293,46358,46454,46471,46541,46603,46709,46737,46804,46815,46860,46918,46923,46928,46943,46957,46960,47003,47043,47062,47073,47208,47231,47286,47289,47290,47294,47404,47407,47482,47632,47646,47672,47677,47686,47689,47741,47745,47753,47768,47777,47784,47823,47826,47833,47898,47918,47988,48048,48098,48122,48127,48190,48202,48237,48270,48284,48295,48301,48307,48341,48345,48354,48375,48415,48421,48464,48480,48494,48496,48508,48619,48633,48650,48849,48874,48918,48944,49158,49184,49203,49221,49306,49318,49352,49376,49410,49447,49479,49533,49561,49573,49579,49611,49704,49738,49751,49774,49806,49822,49840,49845,49866,49876,49898,49920,49938,49970,49980,50039,50141,50227,50464,50522,50750,50810,50841,50919,50963,51244,51300,51448,51498,51578,51602,51690,51724,51974,52136,52189,52195,52351,52426,52442,52679,52685,52725,52753,52788,52792,52801,52970,53091,53293,53362,53363,53515,53663,53666,53850,53888,53895,53988,54030,54055,54075,54177,54208,54225,54282,54706,54774,54839,54932,54935,54997,55066,55087,55231,55299,55348,55428,55462,55627,55734,55833,55854,55891,55948,55997,56069,56100,56207,56268,56312,56463,56477,56516,56545,56638,56643,56667,56694,56714,56921,56930,57023,57135,57137,57186,57258,57318,57622,57654,57701,57772,57816,57819,57875,57970,58065,58089,58351,58529,58561,58636,58734,58806,59098,59119,59186,59212,59220,59254,59363,59378,59492,59542,59598,59642,59643,59664,59670,59688,59764,59867,59916,59947,59972,60113,60154,60158,60313,60392,60561,60578,60592,60606,60782,60797,60911,60957,61031,61113,61398,61546,61568,61574,61589,61633,61787,61808,61810,61838,61848,61920,62114,62147,62222,62331,62404,62412,62596,62702,62844,62845,62951,62989,63032,63043,63065,63199,63240,63251,63286,63738,63776,63807,64011,64081,64120,64201,64266,64341,64563,64647,64742,64748,64792,64800,64878,64895,64954,64971,64976,65043,65051,65357,65386,65466,65532,65567,65639,65660,65756,65771,65878,65984,66131,66151,66344,66349,66543,66584,66626,66730,66803,66834,66840,66883,67097,67195,67201,67207,67245,67282,67300,67366,67763,67890,68031,68160,68191,68202,68257,68391,68593,68662,68673,68884,68910,69038,69271,69299,69371,69466,69515,69548,69722,69791,69879,69913,70018,70036,70039,70112,70141,70220,70307,70375,70427,70492,70562,70669,70809,70873,70907,71159,71302,71402,71548,71726,71853,71968,72139,72294,72344,72375,72408,72469,72528,72538,72548,72583,72599,72635,72639,72726,72894,73131,73276,73374,73479,73500,73521,73644,73737,73810,73822,73883,73903,73915,74299,74419,74537,74922,74998,75120,75130,75203,75276,75304,75347,75364,75384,75405,75442,75445,75481,75494,75530,75659,75682,75875,76091,76095,76172,76265,76268,76329,76357,76445,76498,76540,76544,76791,76800,76810,76811,76839,76907,76989,77249,77251,77381,77393,77514,77728,77748,77801,77827,77848,77881,77899,78196,78231,78399,78449,78613,78733,78794,78802,78868,78883,78891,78925,78999,79237,79294,79452,79509,79636,79653,79736,79850,79851,79859,79919,79996,80338,80350,80374,80381,80389,80406,80565,80636,80740,80911,80991,80999,81076,81124,81133,81169,81280,81385,81496,81585,81705,81722,81723,81752,81861,82220,82309,82373,82469,82524,82621,82755,82830,82895,83037,83188,83370,83652,83753,83782,83828,83843,83845,83867,83942,84018,84072,84179,84378,84660,84682,84739,84802,85245,85350,85358,85534,85727,85848,85871,85886,86002,86014,86032,86085,86110,86320,86408,86441,86606,86610,86637,86640,86895,86988,87082,87095,87118,87147,87150,87224,87357,87445,87558,87667,87813,87921,87930,88032,88054,88284,88300,88324,88429,88456,88483,88485,88556,88670,88805,88955,88986,89057,89065,89145,89260,89270,89421,89509,89542,89785,89828,89885,89934,89970,90054,90100,90150,90185,90221,90450,90500,90719,90948,91061,91110,91115,91123,91319,91403,91411,91558,91619,91714,91720,91761,91875,91935,91939,91997,92045,92119,92452,92563,92579,92588,92693,92753,92782,92783,92793,92812,92883,92923,92949,93000,93148,93186,93253,93277,93426,93432,93585,93615,93631,93675,93728,93899,94146,94229,94290,94299,94409,94607,94621,94735,94917,94933,95001,95007,95095,95159,95323,95377,95442,95494,95524,95573,95697,95709,95714,95725,95940,96042,96078,96172,96196,96203,96219,96297,96438,96598,96648,96787,96847,96986,97000,97055,97120,97123,97197,97335,97400,97425,97437,97469,97490,97584,97624,97873,97915,98070,98146,98180,98496,98514,98523,98608,98795,98811,98923,99011,99055,99118,99335,99486,99617,99719,99808,99847,99892,99967",
      "37,59,94,164,187,1629,1709,1785,1975,2013,2125,2132,2233,2380,2465,2632,2637,2996,3035,3216,3233,3361,3371,3422,3616,3727,3754,3833,4027,4046,4096,4134,4160,4216,4335,4368,4376,4782,4882,4979,4995,4996,6248,6981,8047,9442,9528,9537,9556,9632,9651,9670,9689,9708,9763,9776,9993,10012,10069,10107,10601,10620,10753,10829,10924,11095,11133,11152,11209,11228,11247,11304,11361,11370,11478,11494,11513,11532,11627,11630,11646,11741,11854,11855,11874,11884,11912,11950,11951,11969,11983,12026,12042,12045,12062,12102,12105,12140,12159,12254,12330,12368,12427,12444,12463,12474,12501,12534,12558,12577,12615,12653,12672,12681,12786,12843,12900,12938,12969,12995,13014,13090,13242,13261,13280,13356,13388,13432,13506,13527,13546,13565,13603,13625,13641,13660,13679,13698,13717,13755,13774,13793,13812,13850,13907,13926,13945,13986,14021,14059,14116,14173,14192,14211,14230,14249,14268,14382,14439,14553,14569,14629,14743,14800,14838,14895,14914,14971,14992,15009,15047,15123,15139,15161,15184,15218,15237,15275,15313,15332,15427,15465,15484,15498,15560,15598,15636,15712,15731,15750,15788,15807,15826,15845,15864,15883,15959,15997,16016,16035,16054,16073,16168,16187,16206,16301,16320,16358,16377,16415,16453,16472,16503,16529,16548,16605,16681,16708,16731,16736,16738,16757,16795,16814,16833,16835,16890,16928,16946,17004,17042,17080,17137,17156,17169,17175,17194,17270,17308,17327,17365,17441,17460,17479,17498,17539,17555,17593,17650,17726,17745,17802,17821,17840,17859,17878,17897,18011,18023,18030,18087,18106,18125,18182,18220,18239,18247,18258,18296,18315,18334,18353,18372,18391,18410,18429,18448,18467,18470,18505,18515,18581,18600,18676,18752,18828,18847,18885,18914,18923,18958,18961,18980,19037,19075,19094,19132,19151,19170,19207,19208,19227,19265,19284,19307,19341,19360,19379,19398,19436,19455,19463,19484,19550,19560,19569,19645,19664,19702,19721,19759,19797,19835,19854,19873,19892,19911,19930,19949,19978,19987,20006,20025,20042,20063,20082,20139,20158,20166,20177,20215,20234,20291,20310,20329,20348,20386,20405,20424,20443,20462,20500,20557,20595,20614,20652,20671,20728,20747,20766,20804,20823,20880,20918,20950,20956,21013,21051,21073,21146,21165,21184,21199,21222,21241,21291,21317,21336,21374,21393,21412,21431,21450,21469,21488,21545,21564,21602,21613,21621,21659,21707,21716,21735,21754,21811,21868,21906,21925,21963,21982,22000,22001,22020,22077,22128,22134,22172,22210,22229,22237,22267,22305,22343,22362,22381,22400,22438,22457,22476,22495,22570,22571,22609,22666,22685,22704,22715,22742,22761,22780,22856,22894,22913,22933,22989,23008,23046,23084,23103,23160,23198,23208,23217,23236,23266,23312,23350,23388,23426,23445,23464,23502,23521,23540,23578,23616,23635,23654,23673,23675,23711,23730,23768,23825,23844,23882,23920,23935,23939,23958,23977,23996,24015,24034,24053,24091,24110,24129,24167,24186,24224,24243,24281,24319,24395,24414,24490,24528,24547,24566,24604,24623,24642,24696,24699,24718,24737,24756,24813,24815,24832,24851,24870,24908,24928,25003,25016,25022,25041,25098,25117,25177,25212,25231,25269,25288,25307,25364,25383,25421,25440,25482,25497,25500,25516,25535,25554,25630,25649,25685,25687,25706,25744,25763,25839,25854,25858,25877,25896,25915,25934,25953,25991,26010,26048,26086,26124,26160,26175,26181,26200,26238,26257,26421,26447,26466,26485,26504,26523,26656,26713,26731,26732,26808,26826,26865,26868,26979,27112,27121,27150,27188,27199,27238,27275,27289,27321,27333,27369,27376,27462,27483,27520,27568,27673,27682,27691,27722,27796,27814,27929,27948,28024,28112,28138,28157,28256,28356,28385,28518,28575,28670,28746,28781,28879,28930,28948,28955,28960,29069,29079,29088,29135,29145,29331,29337,29365,29373,29403,29483,29592,29639,29704,29715,29753,29904,29919,29939,29962,29982,30028,30057,30086,30115,30144,30173,30202,30231,30260,30289,30318,30347,30376,30405,30434,30463,30492,30550,30608,30637,30666,30695,30724,30753,30811,30840,30869,30898,30927,30956,30985,31014,31043,31072,31101,31130,31159,31188,31217,31246,31275,31304,31333,31362,31391,31420,31449,31478,31507,31536,31565,31594,31623,31652,31681,31710,31739,31768,31797,31826,31855,31884,31913,31942,31971,32000,32029,32087,32116,32145,32174,32203,32232,32261,32290,32319,32348,32377,32406,32435,32464,32493,32522,32551,32580,32609,32638,32667,32696,32725,32754,32783,32812,32841,32870,32899,32928,32957,32986,33015,33044,33073,33102,33131,33160,33189,33218,33247,33276,33305,33334,33363,33392,33421,33450,33479,33508,33537,33566,33595,33624,33653,33682,33711,33740,33769,33798,33827,33856,33885,33914,33943,33972,34001,34030,34059,34088,34117,34146,34175,34204,34233,34262,34291,34320,34349,34378,34407,34436,34494,34523,34552,34581,34610,34639,34668,34697,34726,34755,34784,34813,34842,34871,34900,34929,34958,34987,35028,35071,35078,35117,35191,35280,35354,35428,35444,35499,35547,35606,35631,35665,35797,35828,35856,35883,35927,35986,36045,36108,36123,36168,36227,36237,36298,36357,36416,36488,36547,36606,36614,36670,36683,36729,36788,36797,36805,36859,36918,36977,36991,37049,37108,37119,37167,37231,37290,37349,37376,37420,37440,37445,37479,37492,37538,37610,37669,37728,37792,37851,37910,37932,37981,38040,38099,38107,38171,38230,38289,38309,38353,38412,38471,38489,38498,38542,38545,38601,38615,38660,38732,38752,38791,38850,38865,38914,38926,38973,38987,39032,39103,39105,39162,39221,39293,39352,39383,39411,39475,39534,39593,39664,39723,39748,39782,39854,39913,39928,39972,40122,40151,40180,40209,40238,40267,40296,40325,40354,40383,40412,40441,40470,40499,40528,40557,40586,40615,40644,40673,40702,40731,40760,40789,40818,40847,40876,40905,40934,40963,40992,41021,41050,41079,41108,41137,41166,41195,41224,41253,41311,41340,41369,41398,41427,41456,41485,41514,41543,42013,42042,42071,42100,42129,42145,42158,42189,42206,42211,42226,42279,42287,42317,42354,42368,42397,42401,42429,42465,42516,42591,42666,42753,42828,42858,42874,42903,42924,42990,43028,43065,43140,43191,43227,43302,43346,43349,43377,43399,43416,43438,43464,43517,43539,43587,43614,43672,43701,43743,43776,43802,43819,43821,43851,43879,43938,43957,44013,44044,44088,44151,44156,44175,44210,44250,44274,44285,44289,44325,44412,44487,44506,44562,44626,44649,44724,44799,44867,44886,44910,44961,44987,44991,44998,45036,45068,45109,45157,45218,45276,45296,45349,45377,45406,45464,45515,45536,45593,45651,45723,45780,45838,45910,45930,45967,46015,46045,46047,46074,46099,46100,46101,46130,46158,46176,46204,46239,46311,46335,46351,46379,46391,46403,46428,46445,46450,46515,46538,46558,46629,46655,46699,46705,46713,46750,46752,46766,46768,46778,46901,46940,46977,46985,47007,47052,47067,47205,47206,47214,47243,47261,47268,47376,47380,47382,47414,47488,47497,47500,47520,47542,47545,47565,47571,47601,47665,47691,47747,47803,47847,47866,47887,47902,47946,48104,48111,48114,48121,48145,48174,48199,48206,48212,48214,48247,48249,48282,48286,48311,48376,48386,48410,48431,48497,48530,48536,48575,48587,48662,48667,48762,48767,48800,48806,48896,48909,48979,48993,49022,49069,49081,49082,49108,49147,49171,49204,49249,49265,49269,49276,49284,49303,49317,49338,49341,49379,49406,49452,49453,49552,49584,49642,49660,49670,49674,49694,49697,49753,49792,49793,49834,49858,49897,49909,49935,49956,50007,50084,50122,50166,50360,50361,50375,50509,50680,50695,50753,50869,50890,50967,51106,51274,51293,51321,51401,51473,51521,51590,51643,51982,52246,52398,52565,52623,52821,52835,52868,52873,52882,52975,53147,53196,53248,53255,53277,53308,53364,53584,53614,53633,53646,53657,53688,53738,53748,53831,53881,54002,54155,54274,54452,54476,54564,54588,54613,54704,54726,54893,54960,55091,55115,55123,55188,55255,55417,55432,55472,55512,55531,55607,55738,55766,55776,55792,55843,55897,55939,55968,56121,56176,56303,56381,56556,56568,56625,56710,56713,56734,56818,56863,56880,56986,57103,57120,57140,57158,57219,57365,57495,57511,57558,57683,57728,57790,57814,57817,57846,57917,57930,57946,57958,57993,58021,58097,58161,58180,58272,58383,58467,58655,58660,58693,58700,58720,58727,59011,59283,59369,59383,59419,59433,59487,59679,59712,59820,59839,59943,59948,59950,59993,60038,60175,60267,60283,60700,60709,60726,60751,60758,60858,60861,60961,60976,60995,61118,61176,61194,61200,61212,61295,61336,61363,61389,61436,61473,61611,61618,61646,61781,61903,61937,61960,61981,61993,62033,62116,62149,62606,62781,62835,62865,63004,63280,63325,63547,63594,63597,63648,63675,63727,63765,63786,63924,64021,64059,64064,64086,64179,64195,64247,64378,64475,64541,64713,64858,65033,65056,65157,65345,65492,65550,65623,65714,65813,65840,65942,66058,66105,66170,66326,66332,66400,66435,66614,66617,66724,66742,66892,67054,67106,67156,67353,67529,67608,67715,67753,67769,67919,67926,68144,68159,68237,68429,68481,68486,68566,68571,68836,68918,68969,69241,69265,69341,69347,69467,69524,69572,69646,69648,69752,69801,69945,70034,70037,70401,70540,70550,70774,70808,70811,70906,70919,71088,71099,71223,71312,71382,71639,71703,71753,71799,71824,71887,72009,72030,72040,72113,72245,72258,72261,72301,72365,72393,72437,72465,72576,72641,72761,72840,72899,72931,72944,72946,72977,73050,73101,73154,73204,73250,73421,73558,73577,73747,73772,73835,73857,73988,74013,74049,74134,74277,74363,74369,74386,74477,74571,74604,74637,74718,74732,74743,74806,74819,74822,74885,75064,75074,75113,75119,75283,75372,75685,75690,75757,75895,75978,76041,76060,76076,76164,76288,76364,76370,76692,76712,76717,76808,76829,76886,76950,77017,77082,77144,77313,77333,77369,77380,77392,77710,77712,77716,77781,77875,77888,78139,78160,78175,78176,78198,78252,78324,78355,78444,78508,78639,78670,78772,78862,78864,78867,78931,79125,79145,79368,79398,79441,79463,79477,79480,79768,79931,79958,80040,80045,80112,80133,80159,80182,80281,80351,80370,80480,80515,80647,80723,80921,80923,81087,81285,81352,81419,81421,81491,81547,81652,81699,81772,81781,81932,81933,81946,81973,82104,82118,82121,82424,82443,82627,82632,82750,82840,82844,82922,82986,83021,83288,83294,83320,83387,83434,83730,84037,84046,84284,84326,84416,84561,84563,84733,84743,84774,84926,85065,85138,85171,85227,85277,85767,85809,85946,86101,86227,86321,86449,86659,86730,86791,86865,86891,86892,87120,87172,87268,87410,87448,87560,87658,87698,87727,87801,87804,87849,87904,88029,88081,88186,88232,88314,88344,88436,88514,88636,88673,88683,88749,88795,88822,88879,88969,89092,89099,89138,89146,89288,89545,89634,89639,89768,89904,89947,90015,90086,90087,90163,90226,90263,90330,90591,90756,90763,90866,90883,90888,90959,91098,91189,91209,91282,91348,91382,91654,91667,91675,91755,91903,92057,92089,92106,92155,92165,92259,92314,92321,92331,92492,92544,92610,92674,92700,92745,92858,93110,93171,93354,93387,93430,93451,93536,93554,93569,93600,93730,93739,93759,93927,94005,94234,94238,94333,94513,94522,94626,94742,94888,95061,95306,95325,95487,95626,95736,95737,95780,95813,95825,95856,95941,96014,96057,96119,96171,96174,96293,96388,96399,96460,96484,96524,96625,96894,96973,96988,97044,97062,97200,97240,97342,97364,97367,97460,97496,97859,97907,97979,98186,98271,98347,98681,98682,98695,98825,98864,98945,99031,99185,99263,99311,99473,99606,99703,99737",
      "16,253,278,367,371,402,426,455,490,492,532,582,691,697,699,744,756,808,852,886,910,939,945,962,1008,1028,1044,1061,1090,1186,1213,1215,1217,1242,1339,1372,1425,1451,1545,1546,1566,1600,1881,1918,2681,4064,5028,5048,5119,5505,5579,5607,5725,5736,5789,5812,5834,5871,6146,6424,6466,6493,6499,6525,6641,6642,6644,6802,6836,6894,6929,7105,7188,7200,7217,7271,7331,7493,7505,7530,7578,7630,7636,7742,7828,7965,8034,8044,8091,8159,8173,8220,8352,8410,8417,8443,8474,8578,8646,8717,8809,8834,8952,8979,9403,9419,9783,9973,10121,10157,10161,10260,10299,10308,10371,10456,10503,10992,11061,11248,11738,11814,11927,12071,12448,13259,13505,13883,14043,14910,15656,15724,15916,16756,17036,17725,19052,19070,19102,19439,20593,20919,20975,20989,21197,21490,21830,22585,24828,25670,25766,25976,26402,26804,27132,27145,27391,27733,27771,27774,27944,28381,28456,28460,29141,29351,29642,29927,29956,30468,32866,34515,36492,37259,37795,37935,38508,39916,40199,42250,43617,43940,44593,45041,45301,45703,47327,50047,50163,50170,50225,50501,50714,50747,50769,50874,50903,50913,50935,50951,50980,51042,51083,51098,51193,51233,51266,51364,51385,51534,52005,52050,52163,52229,52373,52430,52444,52604,52634,53035,53036,53096,53204,53339,53359,53433,53434,53470,53482,53779,53890,53894,53947,54326,54329,54394,54403,54848,55384,55465,55574,55605,55666,55681,55725,55804,55874,55959,56106,56123,56130,56133,56245,56264,56423,56460,56532,56544,56695,56935,57008,57028,57098,57166,57384,57457,57583,57756,57787,57858,57877,57882,58091,58101,58131,58186,58271,58321,58373,58380,58495,58656,58724,58742,58769,58805,58827,58909,58921,59142,59169,59189,59278,59480,59539,59590,59596,59660,59665,59757,59877,59883,59893,59894,59963,59965,60016,60039,60167,60202,60240,60255,60414,60447,60466,60478,60528,60651,60697,60715,60747,60778,60860,60972,61217,61257,61265,61292,61372,61432,61457,61512,61648,61719,61789,61940,61945,62207,62264,62311,62470,62660,62739,62800,62851,62869,62950,62960,62990,63013,63160,63263,63273,63481,63544,63563,63598,63789,63814,63982,64057,64415,64507,64512,64514,64731,65005,65301,65306,65343,65418,65477,65500,65552,65556,65609,65624,65664,65718,65755,65860,65884,65990,66023,66046,66123,66565,66825,67057,67144,67240,67403,67506,67553,67554,67585,67674,67729,67741,67825,67861,68009,68124,68168,68448,68453,68619,68790,69027,69084,69247,69370,69462,69507,69534,69734,69810,69813,69835,69951,69968,70111,70200,70238,70278,70341,70396,70499,70514,70609,70631,70692,70736,70783,70874,70938,71080,71139,71210,71233,71280,71450,71517,71588,71652,71661,71718,72042,72066,72209,72243,72281,72369,72373,72395,72397,72502,72593,72605,72618,72796,72810,72842,72877,72967,73161,73164,73177,73195,73268,73311,73342,73384,73389,73501,73629,73705,73744,73748,73776,74036,74131,74146,74185,74196,74295,74318,74341,74602,74635,74769,75072,75097,75176,75178,75214,75463,75479,75518,75627,75666,75694,75805,75949,76079,76116,76120,76124,76149,76258,76310,76418,76427,76468,76501,76519,76534,76579,76612,76649,76656,76705,76710,76784,76799,76919,77015,77091,77097,77182,77454,77508,77533,77545,77610,77628,77792,77880,77940,77993,78010,78020,78083,78133,78233,78491,78502,78569,78595,78660,78690,78747,78748,78897,78914,79025,79130,79187,79190,79241,79288,79449,79505,79571,79717,79743,79782,79820,79827,80103,80183,80185,80191,80264,80378,80452,80679,80828,80843,80854,81031,81110,81400,81474,81587,81603,81613,81677,81790,81791,81825,82010,82080,82223,82231,82236,82278,82344,82454,82563,82658,82665,82791,82806,82824,82897,82978,83131,83143,83223,83245,83307,83347,83558,83571,83610,83688,83894,84001,84016,84035,84190,84332,84547,84574,84677,84680,84696,84746,84956,85049,85074,85136,85221,85291,85318,85338,85492,85826,85844,86001,86017,86081,86216,86286,86530,86540,86770,86827,86931,86966,87052,87202,87208,87247,87366,87430,87465,87602,87635,87663,87757,87811,87942,88015,88046,88065,88144,88154,88310,88336,88406,88443,88498,88503,88763,88856,88921,89365,89460,89514,89544,89561,89780,89872,89888,89972,90034,90040,90046,90116,90139,90302,90309,90349,90367,90406,90407,90474,90565,90604,90665,90757,90789,90863,90918,91042,91127,91181,91258,91396,91459,91508,91514,91537,91698,91717,91744,91789,92092,92144,92258,92518,92561,92628,92639,92746,92824,92844,93227,93309,93400,93453,93462,93700,94044,94076,94152,94416,94571,94705,94717,94726,94751,94755,94839,94865,95023,95060,95063,95179,95234,95427,95449,95510,95515,95536,95594,95612,95674,95842,95887,95959,96017,96061,96199,96308,96425,96617,96710,96745,96748,96779,96780,96908,96936,96990,97002,97034,97037,97046,97092,97116,97199,97222,97237,97308,97422,97466,97474,97486,97663,97694,97777,97858,97861,97897,98006,98320,98375,98382,98422,98576,98649,98696,98803,98830,98832,98860,98943,98969,99015,99133,99217,99276,99279,99305,99351,99390,99480,99578,99760,99946",
      "8,11,280,300,336,347,440,449,457,471,547,555,594,603,609,611,639,710,810,824,863,934,955,969,988,1000,1015,1023,1127,1254,1265,1312,1362,1369,1449,1479,1500,1505,1512,1558,1561,1573,1631,2265,2548,3363,3369,3598,3862,4562,4885,4939,4977,5002,5074,5139,5183,5209,5429,5451,5507,5557,5563,5587,5702,5744,5883,5914,5968,6008,6048,6050,6094,6095,6128,6190,6209,6495,6637,6703,6708,6713,6721,6728,6794,6886,6887,6913,6924,7115,7122,7290,7307,7429,7511,7682,7794,7817,7932,7934,8076,8077,8179,8268,8342,8418,8433,8521,8579,8714,8827,8884,8901,8982,8998,9004,9148,9207,9311,9400,9416,9996,10124,10148,10202,10478,10794,10886,10911,11030,11174,11201,11202,11722,11743,11799,11818,15683,17973,19246,19335,21072,23458,23803,25696,25841,25900,26090,26978,27058,27378,27810,27859,28208,28346,28778,28853,29266,29473,31258,33499,34021,35016,35826,38356,38482,40111,41575,41606,41766,41946,42492,42564,43543,43646,45059,45198,45256,47115,49165,49362,49486,50130,50150,50176,50244,50321,50527,50576,50607,50914,50933,51122,51397,51451,51457,51733,51739,51792,51826,51855,51860,51874,52004,52038,52107,52274,52279,52310,52320,52754,52769,52994,53126,53300,53385,53519,53555,53565,53645,53650,53728,53760,53806,53822,53887,53943,53945,54054,54250,54267,54308,54372,54459,54650,54847,54973,55270,55453,55499,55506,55587,55601,55899,55942,56029,56062,56220,56250,56316,56345,56446,56521,56599,56804,56962,57046,57090,57231,57315,57355,57431,57487,57498,57518,57599,57623,57742,57753,57755,57903,58037,58240,58267,58291,58454,58637,58778,59007,59122,59185,59232,59348,59401,59455,59517,59708,59719,59824,60112,60298,60402,60434,60452,60612,60671,60685,60716,60717,60738,60855,60895,60925,60940,61028,61210,61244,61270,61277,61350,61399,61413,61455,61462,61503,61543,61680,61721,61806,61840,61871,61895,61956,62048,62090,62097,62107,62132,62155,62190,62294,62571,62671,62730,62788,62862,62917,63071,63086,63099,63112,63134,63172,63411,63468,63486,63509,63524,63587,63618,63721,63728,63766,63810,63854,64012,64020,64038,64232,64368,64384,64432,64435,64458,64738,64787,64918,65126,65218,65309,65368,65464,65507,65519,65526,65579,65652,65996,66014,66104,66148,66175,66350,66393,66437,66651,66721,66741,66754,66807,66849,66900,66981,66994,67038,67077,67105,67159,67213,67271,67309,67462,67494,67595,67687,67700,67739,67750,67969,67971,68076,68093,68114,68268,68370,68446,68632,68691,68699,68705,68765,69051,69070,69159,69167,69196,69308,69319,69353,69365,69402,69412,69429,69503,69529,69562,69628,69669,69718,69777,69814,69832,69900,69901,69937,70035,70075,70094,70096,70206,70232,70271,70447,70591,70794,70816,70847,70933,71102,71182,71297,71319,71381,71526,71540,71557,71608,71627,71658,71797,71808,71857,71860,71921,71944,71954,72234,72265,72306,72411,72637,72646,72687,72706,72708,72752,72951,72986,73028,73137,73223,73335,73457,73546,73613,73710,73864,73868,73963,73974,74014,74037,74089,74169,74192,74250,74465,74469,74544,74570,74607,74642,74705,74971,75052,75057,75215,75222,75284,75373,75397,75461,75477,75488,75561,75585,75588,75726,75906,75977,76067,76102,76246,76279,76469,76594,76642,76723,76725,76877,76961,77005,77013,77063,77123,77136,77294,77501,77547,77571,77642,77925,77959,77988,78013,78127,78155,78168,78187,78354,78375,78416,78418,78433,78478,78528,78611,78707,78822,78832,78848,78860,79040,79289,79348,79379,79467,79515,79560,79605,79783,79837,80167,80212,80290,80297,80298,80380,80420,80527,80609,80798,80925,80994,81054,81060,81069,81077,81098,81319,81379,81416,81464,81515,81571,81626,81703,81734,81800,81856,81876,81949,81951,82039,82100,82148,82204,82230,82256,82509,82543,82565,82573,82645,82702,82778,82798,82901,82936,82955,82994,83004,83032,83035,83050,83247,83258,83337,83360,83397,83680,83733,83734,83737,83919,83937,84022,84245,84281,84386,84414,84437,84482,84566,84646,84741,84822,84872,84889,84893,84978,85025,85051,85160,85235,85403,85627,85651,85829,85891,85967,86238,86345,86384,86454,86646,86826,86913,87020,87105,87409,87592,87594,87872,87950,88118,88146,88160,88244,88293,88462,88737,88776,88854,89002,89120,89147,89229,89259,89372,89434,89588,89775,89951,90009,90035,90082,90097,90170,90188,90205,90257,90266,90294,90373,90439,90533,90539,90540,90589,90662,90676,90717,90724,90761,90805,90815,90845,90913,91082,91126,91499,91615,91770,91779,91813,91831,91917,92000,92235,92385,92386,92490,92609,92632,92702,92719,92791,93003,93012,93081,93088,93155,93258,93512,93667,93718,93957,93976,93981,93988,94186,94200,94212,94218,94281,94324,94344,94398,94668,94720,95107,95232,95255,95447,95603,95640,95645,95734,95744,95771,95815,95925,95993,96005,96056,96198,96331,96364,96427,96465,96672,97069,97134,97191,97282,97489,97491,97684,97762,97877,97991,97994,98184,98301,98307,98458,98466,98509,98579,98714,98790,98828,98836,98874,98898,99288,99298,99385,99700,99724,99986",
      "3,258,333,352,355,361,363,385,393,412,458,467,494,562,566,579,686,725,779,806,834,914,923,956,1060,1080,1115,1201,1225,1268,1322,1324,1381,1414,1442,1482,1488,1554,1869,2229,2584,3014,3415,4991,5005,5056,5086,5088,5102,5132,5175,5251,5328,5335,5392,5484,5497,5516,5632,5708,5760,5776,5777,5880,5906,5999,6011,6115,6134,6258,6310,6386,6630,6849,6878,6934,6975,7078,7079,7082,7103,7140,7436,7631,7634,7673,7810,7812,7942,8113,8185,8262,8286,8376,8384,8459,8488,8508,8549,8693,8732,8737,8800,8829,8913,9084,9108,9192,9198,9322,9329,9354,9842,9848,9871,9936,10151,10173,10218,10357,10711,10943,11783,12334,12504,13130,14651,14993,15028,15245,17009,17181,18014,18052,18444,18636,18764,19401,19477,19530,19597,20057,20078,20380,20571,20748,21279,21334,21511,21719,22041,22543,22943,23618,24209,24736,24738,25107,26924,27717,28862,29749,29799,29808,29958,31058,31062,32247,34050,35120,36561,37020,39631,39881,40454,42198,44915,45711,45760,45901,45984,49010,50076,50240,50383,50449,50552,50553,50602,50621,50709,50802,50848,50928,51020,51159,51164,51176,51219,51235,51299,51343,51479,51510,51512,51606,51829,51873,51875,52287,52372,52375,52417,52516,52590,52723,52734,52822,52840,53010,53254,53275,53276,53342,53405,53498,53811,53852,54044,54070,54105,54110,54112,54128,54153,54370,54423,54937,55049,55065,55173,55331,55418,55466,55521,55523,55623,55699,55811,55994,56013,56038,56129,56163,56209,56334,56560,56597,56653,56726,56914,57017,57087,57328,57399,57482,57578,57647,57897,57916,57953,58126,58135,58185,58200,58368,58463,58474,58758,58855,59069,59206,59227,59265,59272,59279,59303,59339,59465,59474,59499,59601,59703,59773,59781,59933,60086,60168,60182,60187,60360,60430,60552,60571,60602,60621,60627,60745,60775,60799,60813,60817,60845,60904,60924,60992,61140,61169,61228,61245,61335,61355,61422,61438,61451,61463,61474,61522,61599,61742,61970,62005,62028,62125,62173,62208,62348,62580,62613,62644,62726,62825,63024,63166,63219,63241,63385,63433,63448,63508,63751,63785,63812,63958,63973,63976,64053,64105,64116,64148,64152,64167,64234,64248,64391,64654,64667,64752,64762,64780,64805,64890,64892,64898,64964,64973,65014,65068,65139,65315,65323,65655,65659,65687,65856,66045,66141,66289,66292,66380,66403,66412,66416,66658,66668,66744,66751,66851,66861,66976,67030,67173,67269,67277,67617,67638,67640,67659,67663,67899,67921,67963,67978,68094,68155,68161,68219,68255,68283,68313,68428,68460,68723,68851,68873,68936,68990,69013,69041,69099,69204,69219,69255,69571,69597,69604,69626,69649,69652,69662,69744,69775,69876,69936,70147,70153,70282,70299,70451,70614,70677,70826,70889,70949,70975,71196,71201,71245,71346,71364,71733,71841,71952,72222,72387,72510,72584,72647,72683,72781,72897,73011,73038,73184,73187,73229,73332,73499,73656,73753,73779,73913,74033,74313,74400,74475,74481,74519,74608,74688,74780,74820,74832,74889,74896,74959,74965,75177,75401,75515,75599,75687,75697,75717,75777,75797,75901,75979,75991,76032,76242,76353,76373,76413,76423,76638,76668,77157,77179,77227,77342,77453,77759,77777,77826,77836,77883,78087,78218,78271,78280,78317,78464,78630,78709,78710,78744,78765,78781,78831,79101,79143,79228,79271,79324,79499,79647,79759,79775,79884,79895,80180,80242,80305,80371,80414,80419,80422,80431,80507,80633,80708,80766,80814,80873,80902,81057,81186,81194,81366,81468,81662,81700,81789,81941,81981,82012,82173,82288,82376,82495,82576,82581,82586,82620,82626,82652,82663,82713,82790,82920,82932,82943,83030,83036,83041,83053,83069,83160,83189,83204,83282,83297,83326,83458,83639,83789,83813,83975,84002,84099,84109,84201,84263,84295,84301,84485,84582,84683,84811,84842,84846,84854,84860,84865,84990,84999,85007,85236,85256,85334,85483,85495,85538,85567,85623,85683,85701,85704,85738,85744,85753,85773,85940,86027,86386,86426,86442,86563,86636,86643,86676,86820,86869,87058,87099,87170,87290,87469,87482,87494,87552,87665,87772,87839,87926,87958,88035,88051,88108,88138,88157,88196,88269,88285,88332,88460,88470,88505,88567,88757,88777,88779,88818,88909,89112,89126,89158,89210,89239,89286,89309,89318,89658,89673,89699,89739,89804,89914,89985,90006,90041,90073,90239,90247,90527,90595,90689,90958,91022,91040,91204,91275,91365,91492,91582,91599,91617,91621,91652,91742,91763,91857,91941,91948,92016,92018,92091,92129,92184,92202,92225,92231,92266,92286,92307,92338,92358,92379,92504,92541,92598,92695,92764,92884,92904,92911,92933,93037,93068,93113,93127,93207,93312,93389,93437,93499,93590,93656,93659,93671,93689,93708,93745,93796,94135,94171,94209,94259,94492,94511,94590,94696,94740,94765,94810,94889,95117,95141,95162,95338,95518,95666,95740,95789,95802,95811,95836,96272,96286,96511,96621,96684,96929,96946,96951,96966,97007,97049,97257,97268,97365,97398,97440,97544,97621,97655,97681,97696,97698,97723,97754,97977,97990,98151,98158,98201,98209,98244,98268,98297,98360,98464,98587,98604,98620,98805,98835,98867,98927,99032,99052,99070,99138,99420,99522,99638,99819,99919",
      "141,273,289,348,389,409,413,430,433,472,473,514,530,613,630,700,784,814,832,861,866,868,1010,1069,1074,1121,1167,1193,1222,1331,1333,1496,1520,1617,1621,1671,1894,2090,2635,2778,2826,3242,3650,4788,4910,5049,5067,5213,5218,5257,5292,5293,5334,5353,5374,5401,5518,5613,5650,5651,5660,5712,5818,5823,5912,6017,6071,6117,6138,6239,6471,6479,6755,6902,6936,7008,7087,7091,7143,7170,7230,7275,7277,7289,7302,7347,7367,7584,7698,7718,7722,7761,7910,7988,7994,8105,8253,8292,8617,8654,8777,8815,8869,8886,8887,8968,9029,9160,9222,9760,9811,9882,10271,10291,10462,10538,10570,10591,10638,10730,10823,10936,10937,11169,11393,11429,11744,13826,14442,14496,16620,16958,17438,17747,18009,18223,19451,19601,20231,21085,21247,21560,22035,22093,23346,23462,23611,23668,23717,23901,24008,24072,25290,25668,25826,26143,26311,26404,26691,26714,26825,27870,27890,28424,28868,29044,29310,29538,29616,29789,30496,33434,35204,36549,36826,37283,37984,38066,38198,38651,38676,38794,38884,39768,41442,41756,42455,43652,45376,45414,45673,46576,47262,47551,48128,49791,50031,50178,50242,50335,50353,50459,50572,50688,50773,50825,50905,50907,50962,51109,51148,51331,51426,51487,51573,51627,51673,51764,51902,51930,51933,51953,52041,52519,52580,52662,52704,52744,52756,52810,52874,52959,52997,53004,53033,53065,53122,53257,53325,53367,53402,53484,53558,53600,53642,53684,53691,53925,53950,54068,54076,54139,54165,54199,54221,54246,54471,54477,54566,54666,54676,54723,54781,54786,54819,54820,54830,55098,55203,55313,55458,55528,55603,55631,55755,55763,55957,56152,56199,56217,56277,56323,56389,56422,56451,56573,56697,56743,56829,56920,56987,57111,57430,57479,57725,57929,57955,58107,58320,58471,58488,58563,58584,58751,58877,58892,58925,58929,58943,59064,59153,59196,59373,59402,59538,59631,59727,59766,59805,59913,59991,60153,60254,60265,60271,60442,60514,60566,60701,60836,60854,60898,61101,61248,61263,61283,61328,61368,61403,61419,61691,61715,61771,61790,61852,61861,61876,61879,61947,62040,62101,62254,62280,62406,62419,62468,62642,62684,62690,62719,63048,63075,63089,63347,63403,63424,63461,63819,63856,63886,64095,64198,64199,64213,64363,64382,64394,64454,64459,64518,64833,64907,64938,64992,65003,65004,65009,65044,65065,65067,65153,65269,65290,65316,65317,65318,65322,65387,65442,65628,65732,65742,65816,65911,65979,65998,66039,66082,66125,66182,66197,66562,66579,66659,66670,66723,66852,67161,67199,67235,67274,67334,67441,67446,67550,67630,67740,67842,67946,67984,67987,68074,68102,68248,68701,68727,68866,68912,68949,69086,69104,69198,69355,69396,69510,69543,69804,69811,69860,69865,69914,70148,70149,70174,70190,70221,70428,70452,70463,70529,70714,70721,70765,70815,70992,71042,71118,71462,71499,71578,71600,71681,71838,71880,71931,72079,72085,72145,72398,72460,72482,72485,72537,72653,72679,72701,72738,72834,72943,72950,72973,73054,73104,73125,73145,73203,73255,73262,73314,73364,73370,73398,73435,73591,73717,73727,73743,73769,73833,73851,74019,74263,74264,74467,74647,74690,74812,74925,74939,74991,75191,75268,75339,75341,75433,75505,75559,75578,75661,75701,75711,75755,76276,76282,76367,76410,76436,76449,76503,76756,76943,77110,77150,77173,77181,77319,77502,77554,77618,77629,77663,77754,77775,77794,77857,77920,77934,78263,78301,78358,78424,78668,78697,78756,78768,78808,78946,79064,79078,79080,79119,79180,79411,79599,79726,79751,79891,79898,79914,80007,80203,80215,80481,80703,80714,80738,80831,80915,81084,81211,81226,81246,81248,81262,81321,81340,81508,81517,81533,81812,81863,81895,82013,82065,82517,82610,82828,82845,83250,83301,83325,83399,83421,83444,83469,83764,83969,84138,84209,84262,84464,84576,84692,84859,84907,84965,84983,85325,85423,85432,85646,85746,86069,86126,86130,86178,86503,86545,86653,86707,86743,86803,86929,86968,86987,86996,87114,87281,87291,87363,87412,87516,87533,87652,87679,87753,87803,87829,88140,88164,88294,88301,88476,88575,88674,88748,88797,88820,88873,88916,88970,88998,89359,89369,89491,89600,89799,89924,90153,90171,90186,90190,90306,90383,90516,90567,90592,90593,90624,90656,90678,90783,90839,90865,90909,90978,91095,91274,91316,91395,91423,91447,91449,91511,91608,91634,91745,91863,91911,91982,91988,92174,92216,92284,92366,92396,92449,92486,92534,92550,92781,92818,93046,93147,93154,93255,93310,93383,93531,93580,93647,93746,93753,93815,93943,94109,94110,94134,94444,94614,94624,94651,94746,94762,94788,94895,94918,95041,95186,95257,95334,95362,95634,95699,95798,95867,95896,96023,96052,96110,96160,96252,96315,96332,96398,96417,96487,96626,96680,96764,96904,96937,96958,97005,97231,97247,97260,97304,97458,97487,97566,97683,97758,97816,97846,98046,98163,98319,98321,98330,98473,98626,98688,98801,98902,98922,98928,98978,98986,98990,99147,99156,99213,99308,99427,99432,99503,99536,99582,99690,99815,99849,99873,99927,100003",
      "196,197,235,271,290,396,505,510,525,567,576,675,766,841,846,864,966,973,1041,1123,1157,1203,1251,1315,1347,1352,1357,1364,1366,1385,1389,1392,1393,1445,1475,1490,1610,1645,1790,1805,1817,1826,1858,1926,1964,1968,1997,2033,2045,2114,2116,2158,2283,2287,2337,2338,2347,2367,2397,2424,2453,2472,2529,2532,2549,2551,2561,2594,2641,2678,2684,2772,2808,2833,2873,2894,2901,2910,2926,2932,2935,2958,2960,2974,2980,2981,2997,3001,3038,3127,3135,3145,3151,3230,3251,3268,3306,3331,3350,3395,3410,3414,3429,3440,3504,3565,3582,3586,3588,3614,3658,3679,3680,3710,3713,3759,3768,3771,3846,3856,3859,3877,3887,3908,3921,3933,3946,3960,4017,4043,4128,4142,4173,4207,4249,4272,4405,4426,4450,4454,4466,4480,4493,4522,4527,4541,4545,4546,4547,4560,4561,4653,4661,4662,4669,4678,4697,4698,4705,4730,4742,4817,4863,4876,4881,4895,4946,4960,4962,4994,5021,5118,5129,5250,5438,5631,5775,5805,5832,5851,5878,5885,5929,5957,6030,6042,6089,6174,6195,6198,6328,6330,6396,6462,6500,6624,6660,6685,6944,7116,7452,7471,7640,7747,7771,7847,7899,7952,7966,8023,8187,8224,8300,8312,8364,8420,8480,8489,8754,8818,8875,8910,8916,8935,8969,9041,9102,9106,9258,9268,9324,9348,9413,9520,9525,9531,9645,9726,9873,10046,10498,11027,11267,11385,11758,11860,11863,11923,12412,12630,12650,12910,13067,13346,13489,13514,13628,13682,13845,13967,13972,14420,14447,14486,14532,14625,14746,14779,15098,15458,15502,15541,15544,15713,15775,15881,15906,15934,16022,16203,16338,16510,16942,17556,17572,17936,17962,18303,18446,18591,18939,19310,19334,19546,20153,20159,20175,20272,20420,20912,20964,21044,21140,21576,21599,21760,21837,21857,21976,22405,22453,22567,22649,22699,22770,22798,22807,22893,22938,22996,23135,23229,23466,23612,23789,24124,24177,24318,24323,24471,24750,24940,24978,25059,25215,25390,25473,25508,25526,25573,25777,25828,26088,26132,26169,26225,26530,26547,26548,26550,26590,26662,26890,26996,27072,27122,27256,27420,27454,27489,27583,27617,27628,27893,27966,27983,28012,28021,28031,28047,28076,28077,28104,28118,28121,28146,28188,28319,28453,28570,28621,28638,28641,28650,28672,28767,28810,28827,28850,29029,29133,29334,29417,29632,29657,29712,30014,30072,30159,30217,30246,30309,30362,30388,30391,30449,30478,30565,30594,30710,30739,30768,30797,30826,30855,30971,31000,31087,31116,31174,31203,31216,31261,31348,31377,31406,31435,31464,31493,31551,31580,31609,31638,31667,31754,31812,31870,31899,31928,31957,31986,32015,32044,32160,32189,32276,32334,32421,32479,32508,32595,32624,32653,32682,32711,32740,32769,32798,32827,32856,32885,32914,33030,33117,33146,33204,33378,33465,33523,33552,33610,33639,33668,33697,33726,33755,33813,33900,33987,34045,34132,34161,34277,34306,34335,34364,34393,34431,34451,34509,34538,34567,34596,34625,34654,34683,34799,34828,34860,34886,34915,34944,34973,35062,35103,35193,35205,35217,35266,35295,35414,35485,35519,35533,35633,35651,35694,35724,35799,35813,35863,35913,35972,36031,36154,36178,36343,36364,36402,36423,36592,36656,36845,36883,36890,36904,36963,37217,37239,37276,37335,37362,37596,37655,37714,37896,37922,37967,38026,38216,38275,38339,38457,38483,38555,38587,38685,38718,38777,38900,39018,39047,39074,39089,39148,39207,39279,39397,39461,39483,39520,39579,39595,39601,39650,39709,39899,39958,40041,40125,40154,40183,40212,40241,40299,40328,40357,40386,40473,40531,40560,40763,40821,40850,40879,40966,40995,41024,41053,41082,41140,41169,41256,41285,41314,41343,41430,41459,41488,41546,42043,42045,42074,42132,42140,42192,42292,42311,42312,42332,42357,42432,42470,42519,42527,42530,42594,42669,42726,42756,42862,42906,42948,42993,43022,43143,43183,43254,43305,43380,43542,43569,43658,43661,43735,43779,43861,43869,43947,44057,44058,44091,44102,44178,44205,44253,44280,44328,44415,44490,44565,44601,44612,44652,44660,44680,44773,44802,44839,44889,44912,44942,44970,44981,45039,45112,45160,45221,45234,45279,45362,45409,45467,45487,45654,45726,45741,45841,45913,45970,46003,46046,46069,46088,46089,46115,46142,46156,46175,46214,46259,46275,46299,46363,46401,46533,46535,46570,46573,46577,46607,46639,46689,46710,46727,46797,46805,46857,46925,46973,46982,47070,47080,47155,47174,47192,47455,47459,47486,47554,47681,47688,47749,47787,47907,47932,47971,48188,48267,48292,48324,48326,48359,48387,48437,48443,48533,48593,48696,48775,48789,48884,48914,48915,49148,49163,49186,49195,49250,49277,49307,49419,49485,49578,49626,49665,49729,49735,49736,49746,49750,49814,49894,49905,49978,50046,50272,50307,50395,50500,50742,50871,50876,50952,50959,51016,51070,51151,51170,51296,51315,51702,51752,51962,51973,51992,52040,52063,52099,52220,52233,52315,52438,52487,52490,52512,52566,52628,52644,52647,52659,52667,53329,53673,53765,53889,53920,53954,54046,54061,54171,54368,54388,54413,54495,54537,54606,54618,54671,54677,54698,54746,54818,54863,54941,55042,55168,55239,55261,55267,55594,55790,55812,55819,55921,55975,56004,56034,56113,56161,56295,56337,56352,56384,56439,56453,56457,56494,56608,56627,56629,56646,56742,56745,56749,56806,56834,56884,56886,56960,56979,57003,57073,57138,57163,57167,57266,57363,57378,57517,57554,57574,57658,57672,57738,57818,57821,57825,57888,57898,58027,58141,58165,58191,58194,58217,58302,58484,58528,58533,58600,58624,58710,58822,58825,58830,58928,58980,59076,59090,59345,59389,59422,59577,59632,59680,59850,59887,59966,59974,60118,60318,60330,60332,60389,60425,60591,60604,60618,60632,60666,60670,60692,60756,60804,61032,61082,61141,61155,61267,61310,61370,61490,61534,61582,61610,61746,61791,61795,61800,61928,61938,61942,61958,62032,62181,62206,62225,62259,62409,62564,62700,62811,62888,62936,63072,63098,63120,63346,63390,63491,63517,63577,63667,63700,63708,63754,63755,63790,63791,63965,64123,64283,64335,64396,64403,64573,64696,64789,64803,64815,64828,64859,64861,64946,64948,65184,65224,65232,65348,65353,65569,65595,65729,65887,65958,65964,65989,66067,66107,66211,66306,66346,66386,66394,66431,66479,66483,66582,66685,66761,66763,66858,66959,66968,67263,67350,67354,67442,67518,67522,67604,67621,67670,67720,67770,67821,67869,67909,67964,68018,68110,68125,68278,68302,68309,68442,68633,68720,69307,69393,69528,69872,69999,70217,70337,70387,70406,70649,70675,70745,70757,70770,70972,70976,71047,71048,71136,71137,71143,71183,71216,71314,71388,71545,71979,72028,72080,72292,72471,72526,72627,72662,73009,73013,73110,73133,73178,73180,73253,73272,73445,73451,73477,73514,73680,73722,73755,73806,73848,74053,74153,74220,74223,74297,74725,74772,75094,75146,75201,75273,75295,75315,75385,75586,75646,75673,75754,75819,75872,75877,75998,76020,76098,76122,76420,76467,76780,76970,77089,77099,77387,77442,77550,77605,77667,77714,77808,77849,77886,78005,78157,78323,78381,78384,78393,78555,78573,78580,78588,78652,78675,78712,78731,78853,78890,78940,78957,79019,79242,79295,79312,79639,79682,79741,79769,79834,80033,80042,80084,80150,80221,80289,80331,80616,80710,80727,80775,81164,81257,81259,81305,81310,81333,81361,81386,81480,81493,81832,81944,81980,82005,82046,82060,82094,82235,82282,82378,82457,82506,82817,82879,82908,83081,83100,83230,83246,83507,83514,83519,83581,83622,83752,83831,83980,84005,84026,84041,84056,84093,84128,84132,84144,84188,84223,84283,84476,84500,84510,84605,84653,84655,84878,84919,84950,85082,85480,85700,85745,85786,85867,85872,85912,85942,85981,86093,86217,86325,86672,86822,86899,86937,87091,87119,87215,87259,87301,87306,87392,87488,87526,87785,87806,87816,87823,87910,87923,87941,87997,88026,88218,88270,88282,88299,88427,88557,88600,88617,88741,88773,88792,89098,89215,89274,89407,89439,89466,89479,89655,89659,89729,89801,90169,90244,90271,90296,90320,90324,90603,90620,90777,90785,90869,90920,90924,91001,91070,91108,91112,91117,91175,91256,91489,91671,91688,91713,91799,91802,91824,91876,91915,92180,92213,92223,92448,92470,92513,92528,92582,92649,92687,92789,92875,92946,93226,93318,93362,93474,93593,93706,93787,93856,93859,94078,94085,94283,94304,94329,94360,94553,94584,94782,94875,94883,94940,94990,95032,95219,95329,95392,95402,95404,95498,95512,95525,95561,95632,95763,95832,96118,96120,96302,96318,96393,96397,96449,96530,96559,96570,96606,96659,96853,96933,96965,96992,97164,97235,97267,97322,97397,97612,97630,97652,97671,97714,97788,97830,97865,97929,98023,98085,98100,98138,98402,98445,98455,98498,98503,98505,98614,98726,98773,98782,98866,99069,99105,99169,99258,99294,99601,99754,99810",
      "307,337,443,463,482,507,535,587,642,713,760,772,822,842,845,911,915,922,937,944,970,978,993,1002,1007,1034,1035,1057,1098,1101,1124,1137,1169,1226,1285,1293,1311,1344,1405,1495,1519,1537,1549,1870,1916,2260,2491,4323,4457,5034,5070,5143,5150,5166,5245,5323,5347,5376,5377,5403,5481,5489,5492,5606,5654,5686,5796,5872,5895,5924,6084,6222,6368,6428,6458,6514,6594,6666,6691,6765,6870,6966,7010,7021,7068,7119,7124,7131,7166,7239,7279,7296,7328,7404,7651,7778,7832,7877,7971,8096,8254,8397,8415,8471,8623,8797,8828,8830,8932,8942,8976,9019,9047,9137,9273,9306,9744,9887,10176,10266,10300,10304,10346,10398,10545,10629,10840,11034,11040,11074,11190,11299,11457,11616,11725,11776,12797,13437,13476,13513,13723,14131,14342,14897,14973,15593,15619,15773,17081,17193,17254,17933,17991,18887,19971,19981,20361,20791,21320,21848,22212,22760,23383,24431,25024,25210,26824,27029,27829,28310,28683,29353,33495,36458,36503,36515,37576,37794,38141,38698,39002,39309,39678,39824,39939,41087,43634,45419,45495,45544,47238,49478,50181,50262,50310,50357,50431,50564,50592,50636,50674,50700,50724,50893,51075,51165,51438,51478,51516,51570,51735,51771,51772,51799,51869,51913,51944,51960,52033,52124,52131,52230,52289,52349,52355,52361,52389,52768,52842,52875,52893,52923,53066,53099,53124,53160,53246,53263,53282,53331,53417,53426,53499,53512,53668,53727,53732,53804,53845,53976,53977,54205,54318,54400,54519,54536,54647,54656,54743,54936,55053,55165,55169,55206,55243,55305,55376,55441,55545,55551,55622,55787,55889,55927,56172,56232,56247,56347,56533,56600,56650,56813,56969,57036,57062,57175,57532,57562,57613,57619,57629,57695,57703,57774,57793,58183,58210,58233,58357,58907,58973,59237,59241,59392,59466,59531,59619,59635,59717,59730,60285,60337,60413,60631,60656,60949,61171,61260,61269,61301,61417,61479,61481,61557,61561,61605,61652,61720,61762,61769,61831,61962,62063,62088,62162,62301,62302,62395,62411,62495,62616,62756,62832,63055,63262,63293,63352,63436,63541,63629,63652,63748,63759,63870,63943,64023,64037,64183,64294,64297,64321,64360,64508,64743,64749,64797,64806,64837,64838,64840,64869,64870,64935,64983,64991,65021,65125,65239,65252,65539,65590,65631,65731,65741,65749,65781,66050,66176,66395,66467,66500,66563,66646,66660,66717,66781,66848,66856,66859,66963,67012,67022,67053,67170,67171,67176,67236,67451,67460,67676,67681,67849,68005,68072,68118,68140,68145,68148,68238,68277,68511,68578,68598,68659,68668,68888,68898,68986,69098,69418,69544,69560,69586,69651,69809,69829,69916,70012,70166,70227,70496,70519,70905,70910,71026,71142,71179,71246,71283,71422,71467,71544,71606,71709,71740,71795,71845,71893,71925,72007,72053,72094,72213,72305,72402,72439,72447,72496,72600,72655,72766,72857,72867,72893,73039,73167,73188,73302,73377,73448,73574,73637,73815,74005,74147,74191,74278,74391,74414,74488,74609,74706,74739,74791,74881,74940,74949,75104,75180,75181,75233,75244,75291,75367,75390,75457,75507,75639,75664,75782,75883,75893,75920,75943,75974,76012,76121,76142,76156,76174,76219,76339,76466,76489,76542,76644,76672,76737,76763,76852,76875,77273,77335,77372,77620,77726,77730,77922,77989,78070,78092,78221,78248,78307,78334,78352,78372,78437,78439,78455,78584,78590,78631,78814,78818,78839,78844,78996,79076,79224,79282,79358,79377,79390,79428,79466,79495,79540,79680,79776,79813,79885,79963,80006,80225,80316,80365,80392,80541,80592,80594,80847,80876,80968,81320,81343,81355,81469,81471,81616,81828,81922,81993,82007,82129,82170,82302,82393,82417,82425,82442,82705,82805,82839,82915,83065,83099,83106,83107,83186,83264,83402,83446,83461,83509,83787,83953,83958,84057,84064,84192,84273,84426,84429,84487,84506,84610,84689,84934,85030,85083,85091,85166,85402,85456,85559,85592,85664,85761,85882,86096,86254,86274,86328,86660,86666,86673,86701,87000,87011,87171,87206,87274,87316,87471,87589,87626,87706,87722,87766,87783,87826,87831,87861,87972,88110,88210,88297,88364,88372,88415,88515,88605,88676,88699,88740,88774,88837,88968,89293,89430,89445,89502,89531,89646,89870,89881,89936,89973,90337,90379,90408,90441,90551,90553,90643,90780,90791,90809,90989,91009,91053,91169,91176,91243,91252,91279,91343,91416,91439,91554,91852,91986,92080,92212,92244,92389,92421,92493,92543,92661,92760,93087,93282,93316,93329,93407,93416,93486,93517,93586,93736,93802,93807,93822,93832,93839,93895,93993,94305,94374,94499,94541,94542,94612,94673,94812,95013,95045,95184,95193,95275,95278,95380,95583,95668,95907,95961,96066,96127,96220,96259,96271,96329,96440,96468,96494,96536,96577,96611,96841,96895,97001,97076,97194,97291,97333,97350,97394,97485,97580,97616,97650,97771,97853,97911,97933,98331,98338,98426,98450,98593,98679,98759,98806,98807,99022,99073,99190,99201,99212,99220,99253,99318,99382,99387,99570,99659,99665,99701,99751,99783,99813,99824,99902,99935,99976",
      "34,86,220,247,382,563,591,684,715,781,791,856,916,926,928,929,1164,1220,1299,1337,1341,1376,1379,1398,1435,1444,1454,1531,1569,1605,1689,1767,1768,1822,1825,1849,1850,1913,1930,2026,2059,2137,2612,2613,2659,2673,2880,3084,3095,3112,3210,3280,3383,3629,3681,3776,3805,3837,3941,4188,4242,4321,4401,4410,4414,4623,4723,4770,4778,4878,5016,5020,5053,5113,5126,5221,5333,5358,5379,5457,5541,5559,5568,5683,5771,5813,5854,5891,5964,6054,6346,6397,6423,6484,6572,6658,6680,6760,6952,6987,7028,7165,7210,7215,7355,7363,7508,7648,7838,8084,8208,8354,8388,8398,8498,8509,8534,8550,8605,8771,8791,8961,9009,9068,9172,9197,9391,9427,9530,9734,9770,10041,10050,10108,10401,10436,10443,10447,10481,10482,10504,10562,10810,11080,11128,11279,11284,11325,11328,11443,11481,11541,11543,11674,11681,11739,11828,11875,11898,12116,12300,12548,12686,12793,12852,13123,13126,13155,13180,13201,13305,13307,13316,13367,13375,13468,13500,13547,13662,13827,13893,14000,14054,14068,14165,14196,14242,14283,14361,14407,14417,14495,14548,14627,14777,14796,15689,15703,15863,15910,16140,16164,16170,16174,16430,16505,16570,16609,16686,16817,16863,17012,17123,17132,17145,17285,17383,17492,17699,17945,17977,17996,18120,18130,18148,18233,18254,18424,18482,18577,18840,18909,18915,18944,19038,19085,19131,19186,19202,19235,19629,19671,19840,19877,19898,19983,20041,20229,20397,20552,20577,20644,20672,20742,20749,20847,20858,20860,21167,21360,21371,21546,21847,22061,22164,22679,22746,22814,23074,23369,23411,23781,23906,24038,24130,24221,24422,24505,24587,24622,24644,24712,24991,25116,25363,25468,25519,25628,25990,26157,26159,26166,26219,26631,26673,26780,26800,26893,27040,27059,27119,27302,27345,27394,27502,27689,27738,27746,27773,27822,27828,27838,27855,27925,27956,28250,28279,28342,28373,28406,28413,28566,28572,28745,28991,29024,29049,29221,29621,29634,29707,29714,29739,29779,29785,29829,30307,30585,30788,32047,32257,32876,33978,34074,35033,35113,35393,35467,35569,35575,35620,35767,35768,35808,35857,35930,35989,36046,36062,36175,36268,36375,36386,36567,36742,36872,36907,37068,37163,37200,37303,37304,37370,37385,37415,37612,37630,37748,37762,37854,37863,38108,38193,38232,38290,38485,38569,38608,38817,38821,38998,39102,39124,39374,39556,39615,39819,39865,39873,39920,39934,39941,40427,40841,40872,41721,42169,42181,42364,42389,42656,42715,42727,42763,42824,42853,42859,42870,42940,43014,43078,43424,43756,43813,44157,44169,44296,44451,44578,44705,45092,45118,45153,45174,45321,45378,45496,45548,45620,45643,45774,45926,45937,45939,46018,46466,46861,46983,47181,47727,47930,48000,48181,48447,48840,49114,50157,50273,50276,50315,50323,50345,50444,50507,50508,50608,50658,50738,50821,50873,50938,50964,51329,51358,51443,51476,51583,51756,51763,51777,51788,51802,51898,52020,52058,52151,52342,52350,52364,52380,52549,52554,52594,52661,52671,52738,52839,52845,52854,53027,53055,53106,53148,53229,53322,53335,53369,53537,53597,53680,53682,53759,53802,53821,53901,54160,54169,54220,54612,54964,54989,55010,55021,55061,55089,55133,55197,55220,55264,55286,55293,55385,55493,55655,55668,55669,55815,55849,55885,55896,55918,55940,56127,56150,56360,56371,56456,56520,56657,56735,56858,56918,57014,57061,57342,57412,57698,57906,58042,58085,58130,58158,58250,58275,58286,58359,58439,58460,58514,58569,58664,58679,58755,58901,59034,59065,59115,59118,59205,59361,59372,59449,59524,59579,59616,59656,59960,59970,60068,60300,60319,60396,60408,60411,60422,60428,60429,60489,60513,60702,60736,60767,60903,60984,61026,61154,61364,61575,61587,61621,61662,61692,61974,62150,62289,62310,62329,62417,62647,62675,62762,62784,62804,62860,63063,63248,63407,63439,63446,63536,63555,63672,63760,64073,64258,64303,64308,64348,64385,64456,64669,65293,65302,65480,65581,65688,65727,65730,65847,65960,66026,66032,66152,66204,66276,66319,66342,66496,66555,66683,66811,66893,66898,66978,67200,67395,67419,67543,67578,67580,67697,67714,67801,67822,67858,68020,68050,68095,68260,68312,68315,68373,68576,68615,68755,68783,68788,68832,68983,69052,69081,69106,69143,69263,69279,69305,69317,69496,69554,69671,69714,69874,69991,69996,70189,70359,70469,70504,70571,70612,70967,70995,71038,71148,71164,71169,71473,71490,71582,71693,71712,71814,71854,71905,72041,72060,72097,72157,72166,72202,72268,72288,72340,72368,72495,72565,72582,72668,72764,72770,72909,73030,73043,73149,73227,73256,73462,73655,73914,73947,74066,74076,74111,74112,74218,74254,74276,74291,74471,74529,74599,74672,74839,74942,74968,75232,75258,75393,75407,75531,75534,75605,75631,75791,75795,75812,75889,75937,75972,75996,76087,76146,76227,76233,76247,76379,76409,76507,76632,76713,76777,76869,77039,77069,77138,77189,77280,77366,77476,77530,77537,77570,77682,77805,77889,78022,78025,78027,78184,78261,78268,78484,78620,78984,79032,79198,79204,79330,79366,79456,79581,79674,79893,79971,79979,80050,80066,80224,80241,80284,80566,80580,80620,80666,80673,80732,80785,80838,80880,80898,81022,81119,81177,81272,81341,81383,81498,81531,81679,81736,81738,81886,81977,81978,81999,82008,82073,82445,82480,82555,82605,82654,82660,82709,82768,82849,82868,82954,83012,83292,83457,83606,83758,83765,83772,83859,83888,83899,83939,83956,83991,84036,84297,84531,84546,84591,84781,84988,84995,85053,85296,85357,85393,85415,85564,85581,85758,85839,85846,85869,85930,85959,86029,86160,86183,86190,86273,86430,86447,86510,86706,86721,86864,86877,87080,87240,87262,87295,87350,87364,87396,87406,87418,87422,87607,87670,87678,87791,87913,88206,88312,88334,88370,88497,88561,88565,88595,88783,88828,88935,89254,89280,89352,89683,89875,89912,89923,90175,90264,90463,90545,90601,90684,90688,90834,90897,90906,90910,90916,90940,91041,91190,91287,91344,91469,91477,91498,91539,91575,91920,91922,91952,92052,92109,92282,92423,92575,92601,92680,92699,93143,93230,93261,93466,93492,93514,93599,93610,93664,93738,93918,93962,94017,94025,94035,94057,94077,94158,94182,94190,94199,94243,94330,94448,94551,94769,94774,94840,94843,94910,95000,95142,95143,95220,95486,95600,95625,95646,95702,95741,95762,95916,95919,95991,95996,96126,96282,96473,96549,96609,96727,96830,97274,97359,97513,97610,97667,97786,97909,98178,98607,98611,98960,98967,98972,99013,99026,99047,99256,99322,99397,99449,99463,99490,99608,99611,99722,99729,99829,99988",
      "36,56,112,139,147,192,1715,2141,2148,2245,2364,2620,2728,2777,2791,3011,3045,3163,3239,3673,4213,4232,4485,4509,4974,4975,5877,6013,6591,6643,6852,6863,6932,7161,7247,8019,8130,8168,8244,8510,8643,8681,8928,8930,9023,9049,9080,9099,9141,9175,9194,9346,9384,9460,9469,9498,9593,9612,9650,9688,9707,9819,9820,9821,9878,9935,9943,10049,10057,10165,10212,10220,10239,10258,10269,10328,10429,10467,10469,10512,10581,10600,10636,10654,10664,10752,10809,10820,10885,10961,11015,11021,11047,11113,11170,11188,11208,11303,11455,11512,11518,11521,11551,11576,11599,11645,11721,11759,11877,11892,11899,11911,11993,12025,12109,12215,12234,12253,12291,12310,12386,12481,12500,12557,12566,12671,12747,12804,12899,13089,13211,13214,13260,13279,13298,13374,13469,13492,13621,13640,13678,13697,13754,13773,13921,13944,13963,14020,14058,14061,14077,14153,14191,14229,14267,14305,14381,14457,14533,14590,14628,14663,14685,14818,14896,14913,14918,14985,14989,14998,15034,15141,15179,15227,15236,15255,15350,15377,15407,15464,15597,15635,15692,15730,15749,15806,15825,15844,15920,16034,16053,16063,16110,16148,16167,16186,16205,16224,16243,16281,16300,16319,16376,16433,16483,16490,16509,16661,16737,16813,16870,16927,16940,16943,16965,17022,17091,17117,17212,17231,17307,17345,17364,17368,17437,17459,17478,17497,17592,17610,17611,17687,17720,17744,17782,17801,17839,17896,17934,17953,17954,17972,17979,18048,18105,18124,18181,18200,18276,18314,18352,18409,18447,18479,18593,18618,18694,18765,18770,18789,18846,18903,18960,19084,19150,19188,19230,19264,19302,19340,19359,19416,19419,19435,19453,19682,19740,19834,19872,19891,19910,19948,20005,20062,20074,20119,20138,20157,20178,20195,20233,20271,20309,20347,20423,20438,20461,20594,20613,20898,20993,21021,21069,21077,21145,21164,21188,21202,21219,21274,21278,21335,21380,21411,21430,21468,21482,21563,21582,21715,21734,21753,21791,21810,21829,21867,21905,21924,21943,22019,22073,22076,22228,22266,22342,22399,22418,22437,22456,22513,22533,22589,22647,22684,22703,22735,22832,22840,22945,22950,23010,23028,23083,23178,23197,23235,23255,23273,23329,23330,23347,23368,23425,23463,23499,23501,23539,23558,23577,23605,23623,23634,23710,23748,23862,23881,23946,23976,23995,24052,24150,24194,24204,24223,24242,24250,24337,24432,24451,24470,24527,24565,24584,24755,24857,24869,24888,24941,25002,25040,25073,25078,25094,25097,25135,25211,25287,25302,25325,25381,25382,25477,25534,25553,25565,25667,25714,25781,25819,25838,25876,25914,25952,25971,26123,26142,26161,26218,26237,26351,26408,26427,26484,26503,26636,26687,26705,26763,26875,26902,26997,27116,27162,27203,27244,27250,27295,27377,27383,27418,27434,27593,27618,27670,27719,27736,27909,28080,28099,28113,28194,28295,28536,28574,28593,28762,28897,28954,29201,29226,29227,29292,29439,29562,29638,29752,29809,29860,29879,29934,29985,30027,30056,30143,30172,30230,30259,30317,30346,30375,30404,30462,30491,30520,30607,30636,30665,30723,30752,30781,30810,30839,30897,30926,30955,31013,31071,31100,31129,31158,31187,31245,31274,31290,31361,31448,31506,31535,31651,31680,31725,31767,31796,31854,31912,31970,31999,32086,32115,32144,32173,32231,32260,32318,32347,32376,32392,32405,32463,32492,32579,32666,32695,32724,32840,32869,32927,33014,33043,33072,33101,33130,33188,33217,33304,33333,33362,33478,33507,33536,33594,33623,33681,33710,33855,33884,33913,33971,34029,34116,34174,34232,34261,34290,34348,34522,34609,34638,34696,34725,34754,34783,34812,34870,34928,34957,35027,35116,35142,35279,35324,35353,35359,35427,35546,35664,35796,35855,35867,35985,35992,36004,36080,36107,36112,36167,36191,36226,36284,36297,36356,36415,36605,36669,36728,36787,36858,36976,36980,36988,37048,37166,37172,37230,37293,37299,37348,37392,37419,37488,37537,37548,37674,37791,37850,37914,38039,38288,38352,38411,38470,38528,38541,38600,38659,38731,38981,38993,39161,39226,39248,39292,39533,39592,39781,39912,39971,40101,40130,40188,40246,40275,40333,40391,40507,40536,40565,40623,40676,40681,40710,40739,40768,40797,40884,40913,40942,40971,41000,41174,41319,41464,41493,41522,42054,42083,42112,42172,42231,42260,42337,42406,42499,42574,42603,42620,42649,42736,42771,42811,42845,42886,42973,43016,43048,43116,43155,43210,43248,43275,43278,43283,43285,43347,43360,43447,43488,43574,43597,43641,43655,43684,43759,43814,43899,43909,43921,43996,44056,44071,44119,44129,44134,44158,44190,44308,44334,44360,44395,44470,44500,44525,44530,44583,44603,44632,44661,44669,44684,44707,44743,44782,44823,44841,44944,44975,45010,45019,45140,45254,45332,45389,45425,45519,45576,45634,45706,45763,45821,45848,45854,45893,46030,46098,46105,46111,46161,46364,46381,46444,46508,46591,46619,46735,46742,46762,46969,47103,47175,47180,47216,47234,47250,47317,47340,47346,47378,47539,47567,47587,47605,47660,47765,47850,47940,48069,48130,48136,48306,48329,48409,48499,48519,48543,48566,48589,48610,48624,48644,48704,48739,48880,48945,49026,49028,49071,49228,49311,49361,49477,49535,49574,49646,49706,49811,49813,49828,49831,49855,49856,49940,49945,49973,50004,50088,50108,50138,50199,50266,50271,50287,50629,50649,50995,51037,51053,51077,51243,51251,51256,51332,51381,51450,51691,51728,51852,51886,51943,51945,51988,52060,52101,52357,52382,52589,52763,52764,52862,52886,52905,53171,53184,53438,53511,53622,53631,53660,54118,54178,54263,54383,54461,54512,54587,54625,54630,54737,54868,54939,55000,55107,55157,55311,55330,55368,55380,55397,55517,55554,55638,55672,55724,55754,55956,56039,56111,56114,56248,56291,56450,56474,56582,56603,56684,56693,56901,56911,57136,57237,57308,57416,57503,57514,57570,57572,57688,58019,58059,58095,58231,58300,58399,58422,58536,58771,58786,58787,58918,58968,58997,59004,59040,59054,59308,59321,59371,59424,59578,59597,59608,59784,59814,59828,59835,59884,59923,59982,60001,60030,60136,60140,60142,60162,60221,60235,60615,60719,60777,60944,60958,60964,61069,61211,61342,61435,61496,61687,61857,61912,61921,62131,62169,62231,62248,62274,62363,62561,62769,62855,62974,63084,63137,63278,63412,63566,63809,63914,63969,64055,64056,64135,64172,64204,64273,64310,64357,64367,64375,64404,64666,64775,64779,65119,65215,65219,65270,65421,65559,65599,65650,65693,65850,65915,65923,65976,66094,66159,66254,66281,66477,66540,66628,66681,66729,66750,66756,67035,67063,67084,67153,67155,67275,67393,67411,67458,67465,67565,67917,67930,68006,68008,68011,68065,68080,68211,68213,68216,68488,68500,68514,68536,69019,69024,69142,69233,69273,69345,69514,69657,69684,69693,69723,69749,69833,69834,69862,69899,69927,69958,70050,70237,70275,70423,70461,70599,70845,71034,71078,71084,71129,71131,71320,71356,71496,71564,71585,71613,71624,71671,71692,71694,71772,71901,71907,71957,71988,71993,72052,72088,72121,72313,72390,72391,72509,72622,72742,72744,72880,72910,73023,73460,73498,73547,73572,73719,73735,73800,73955,74022,74056,74226,74265,74320,74338,74451,74518,74619,74620,74625,74658,74715,74792,74844,74849,74851,74961,75029,75075,75088,75099,75340,75374,75813,75874,76010,76042,76107,76150,76177,76193,76439,76453,76748,77012,77071,77195,77306,77310,77396,77477,77551,77568,77581,77700,77704,77928,77954,77995,78172,78340,78468,78859,78945,79043,79205,79383,79430,79433,79712,79793,79847,79916,80037,80437,80456,80574,80774,80981,80988,81139,81163,81179,81253,81364,81404,81809,81845,81889,81904,81907,81947,81963,82045,82052,82084,82097,82137,82364,82401,82459,82473,82816,83059,83207,83295,83366,83468,83521,83663,83673,83686,83701,83821,83915,83928,83973,83996,84129,84187,84189,84268,84320,84356,84373,84395,84719,84812,84835,84890,84941,84971,85016,85069,85102,85157,85216,85219,85288,85383,85586,85652,85656,85804,85945,85988,86100,86187,86218,86307,86337,86368,86480,86483,86627,86784,87030,87085,87090,87144,87317,87436,87493,87529,87590,87800,87812,87822,87916,87945,88042,88309,88337,88482,88521,88555,88724,89077,89227,89321,89340,89386,89387,89403,89526,89616,89626,89652,89781,89815,89893,89922,90019,90267,90346,90467,90639,90804,90816,90857,90928,91037,91066,91087,91285,91305,91327,91604,91632,91642,91692,91725,91808,92062,92220,92306,92410,92440,92484,92562,93428,93442,93566,93803,93812,93897,94130,94138,94249,94445,94478,94620,94807,94815,94826,94946,95020,95178,95282,95369,95412,95461,95526,95548,95827,96067,96531,96703,96781,96782,96881,96884,97038,97151,97186,97219,97234,97366,97475,97505,97848,97849,97867,98239,98497,98982,99027,99092,99150,99172,99224,99331,99353,99438,99495,99498,99504,99511,99513,99528,99541,99544,99625,99773,99805,99869,99882",
      "88,107,174,180,252,718,1627,1665,1752,1763,1910,1921,1928,1946,1956,1960,1973,1989,2000,2005,2021,2032,2040,2071,2073,2084,2097,2098,2102,2104,2115,2124,2130,2145,2171,2175,2187,2220,2234,2236,2254,2262,2285,2315,2403,2407,2444,2460,2489,2503,2520,2552,2558,2597,2615,2639,2650,2675,2687,2718,2726,2729,2750,2752,2760,2762,2776,2787,2796,2840,2856,2881,2888,2898,2906,2917,2945,2963,2964,2978,2991,2993,2994,3002,3019,3024,3041,3047,3048,3076,3079,3128,3138,3167,3170,3176,3187,3191,3236,3254,3263,3266,3267,3269,3283,3292,3307,3365,3397,3423,3426,3436,3444,3445,3449,3465,3488,3489,3494,3505,3514,3555,3559,3583,3590,3599,3626,3652,3666,3669,3678,3697,3705,3712,3725,3788,3806,3808,3809,3819,3844,3847,3848,3850,3866,3872,3892,3894,3904,3909,3957,3962,3982,3992,3993,3995,4002,4021,4039,4045,4065,4082,4085,4103,4120,4143,4150,4193,4196,4219,4222,4224,4228,4238,4254,4288,4292,4298,4301,4305,4307,4313,4326,4332,4339,4345,4365,4421,4446,4463,4468,4478,4503,4510,4518,4520,4543,4556,4564,4598,4615,4632,4638,4664,4673,4689,4740,4744,4756,4758,4776,4792,4794,4823,4825,4832,4856,4861,4871,4874,4889,4909,4932,4937,5463,5583,5598,5655,5676,5921,6005,6104,6298,6472,6481,6487,6568,6586,6593,6605,6657,6701,6720,6741,6789,6838,6904,6909,6965,7090,7118,7181,7232,7234,7265,7293,7324,7349,7359,7510,7551,7582,7619,7699,7802,7854,7881,7913,8041,8060,8147,8163,8200,8250,8375,8616,8686,8750,8909,8992,9014,9031,9089,9146,9210,9244,9292,9551,9729,9733,9754,9813,9928,9954,10008,10010,10115,10256,10290,10301,10302,10419,10437,10454,10589,10594,10784,10904,10994,11086,11089,11141,11187,11289,11390,11420,11610,11680,11716,11865,11926,11936,12027,12080,12179,12362,12401,12402,12544,12675,12714,13000,13097,13288,13538,13552,13583,13721,13753,13761,13806,13815,13861,13877,14015,14125,14174,14281,14392,14393,14482,14550,14562,14576,14582,14825,14903,14948,15065,15115,15288,15345,15399,15718,15735,16087,16195,16308,16384,16544,16723,16810,16860,16968,17265,17316,17376,17429,17558,17675,17730,17915,18019,18067,18151,18178,18271,18328,18474,18623,18638,18641,18671,18691,18711,18714,18732,18776,18788,18798,18836,19113,19165,19295,19425,19491,19655,19712,19778,19857,19901,19943,20136,20211,20399,20426,20479,20491,20514,20543,20584,20667,20800,20931,20947,20976,21035,21123,21147,21258,21416,21500,21530,21556,21840,21959,22005,22069,22118,22281,22379,22492,22695,22857,22940,22958,22979,23047,23122,23215,23216,23401,23408,23436,23609,23835,23846,23847,23980,24005,24120,24179,24244,24634,24742,25376,25704,25860,25891,25892,25985,26005,26054,26105,26135,26165,26362,26365,26571,26572,26592,26646,26869,26953,27071,27117,27177,27179,27311,27334,27338,27494,27540,27560,27600,27633,27651,27744,27899,27992,28010,28035,28053,28063,28090,28096,28117,28119,28135,28136,28159,28300,28463,28476,28499,28508,28525,28546,28548,28579,28609,28639,28648,28655,28656,28838,28899,29182,29541,29593,29605,29674,29675,29690,29854,29896,29917,29942,29992,30015,30044,30073,30102,30131,30160,30189,30218,30247,30276,30305,30334,30363,30392,30421,30450,30479,30508,30537,30566,30595,30624,30653,30682,30740,30769,30798,30827,30856,30885,30914,30943,30972,31001,31030,31059,31088,31117,31146,31175,31204,31233,31262,31291,31320,31349,31378,31407,31436,31465,31494,31523,31552,31581,31610,31639,31668,31697,31726,31755,31784,31813,31842,31871,31900,31929,31958,31987,32016,32045,32103,32132,32161,32190,32219,32248,32277,32306,32335,32364,32393,32422,32451,32480,32509,32538,32567,32596,32625,32654,32683,32712,32741,32770,32799,32828,32857,32886,32915,32944,32973,33002,33031,33060,33089,33118,33147,33176,33205,33234,33263,33292,33321,33350,33379,33408,33437,33466,33524,33553,33582,33611,33640,33669,33698,33727,33756,33785,33814,33843,33872,33901,33959,33988,34017,34046,34075,34104,34133,34162,34191,34220,34249,34278,34307,34336,34365,34394,34423,34452,34481,34510,34539,34568,34597,34626,34655,34684,34713,34742,34771,34800,34829,34858,34887,34916,34945,34974,35015,35046,35059,35073,35104,35127,35155,35178,35214,35240,35267,35341,35415,35440,35486,35534,35593,35614,35652,35725,35784,35843,35864,35914,35973,36017,36032,36056,36069,36095,36155,36190,36214,36251,36285,36344,36376,36403,36475,36497,36534,36558,36593,36621,36626,36657,36716,36732,36733,36739,36775,36846,36905,36964,36979,36984,36997,37036,37095,37126,37154,37218,37235,37245,37277,37336,37407,37446,37466,37505,37525,37550,37582,37656,37677,37685,37715,37733,37779,37797,37838,37897,37938,37968,38027,38086,38102,38158,38182,38217,38237,38276,38340,38360,38399,38443,38458,38477,38529,38588,38622,38647,38719,38738,38778,38796,38799,38837,38901,38935,38960,39019,39090,39149,39208,39280,39339,39398,39424,39429,39430,39462,39502,39521,39537,39544,39580,39604,39651,39677,39710,39769,39784,39786,39900,39959,39977,40112,40141,40170,40228,40257,40286,40315,40344,40373,40402,40431,40489,40518,40576,40605,40634,40663,40692,40721,40750,40779,40808,40837,40866,40895,40924,40953,40982,41011,41040,41069,41098,41127,41156,41185,41214,41243,41272,41301,41330,41359,41388,41417,41446,41475,41504,41533,42004,42033,42062,42091,42120,42156,42180,42270,42298,42324,42345,42386,42420,42477,42507,42534,42551,42562,42582,42657,42744,42819,42837,42894,42931,42954,42965,42981,43056,43077,43094,43095,43108,43109,43131,43243,43244,43293,43368,43455,43530,43605,43692,43744,43767,43818,43842,43929,44004,44065,44079,44096,44098,44117,44130,44150,44166,44218,44241,44264,44316,44333,44358,44403,44429,44432,44441,44478,44498,44527,44553,44640,44715,44776,44790,44838,44868,44877,44931,44938,44952,44983,44985,45006,45027,45051,45063,45066,45087,45100,45133,45148,45209,45267,45289,45294,45303,45340,45397,45455,45527,45560,45584,45642,45670,45714,45771,45829,45922,45940,45958,45983,46028,46090,46123,46173,46202,46278,46287,46288,46325,46352,46372,46380,46383,46397,46483,46511,46544,46582,46654,46658,46758,46769,46808,46825,46828,46933,46981,47105,47194,47204,47223,47337,47358,47409,47451,47485,47521,47531,47570,47572,47580,47600,47620,47687,47710,47805,47856,47876,47878,47890,47934,47943,48032,48059,48068,48110,48133,48137,48144,48175,48231,48240,48328,48357,48362,48405,48424,48428,48432,48441,48532,48537,48539,48556,48582,48606,48640,48641,48651,48671,48678,48722,48742,48831,48930,48938,48996,49016,49052,49076,49096,49097,49138,49140,49153,49155,49162,49167,49193,49209,49260,49273,49294,49327,49335,49359,49375,49385,49433,49454,49459,49504,49537,49551,49559,49565,49567,49594,49602,49618,49659,49666,49677,49690,49734,49764,49802,49826,49906,49922,49964,49975,50022,50066,50290,50324,50385,50402,50478,50558,50570,50574,50703,50839,50911,50950,50953,51006,51069,51079,51108,51114,51121,51268,51353,51449,51454,51547,51574,51644,51661,51707,51795,51817,52091,52097,52162,52276,52399,52413,52465,52482,52681,52735,52747,52936,52995,53006,53012,53250,53270,53333,53550,53559,53582,53591,53652,53805,53838,54020,54156,54212,54231,54306,54638,54761,54777,54802,54815,55060,55079,55120,55284,55556,55608,55733,55735,55741,55782,55793,55947,55951,55986,55991,56010,56021,56041,56090,56148,56227,56288,56411,56647,56675,56805,56854,56871,56882,56932,56958,57283,57343,57352,57401,57472,57520,57883,57978,58008,58051,58068,58074,58090,58162,58419,58716,58764,58791,58833,58949,58989,59016,59062,59113,59181,59198,59338,59366,59528,59763,59903,60076,60090,60238,60263,60315,60342,60436,60483,60510,60543,60596,60624,60765,60809,60884,60906,60943,61104,61147,61159,61188,61209,61221,61229,61230,61281,61536,61573,61748,61784,61804,61834,61997,62343,62420,62509,62544,62629,62667,62789,62924,62984,63045,63107,63253,63255,63477,63527,63546,63552,63565,63704,63813,63859,63996,64045,64278,64409,64416,64638,64674,64873,64931,64958,64987,65035,65049,65168,65253,65341,65419,65451,65524,65818,65831,65861,65867,65908,65921,66085,66120,66173,66440,66511,66608,66625,66653,66654,66793,66818,66908,66955,67043,67139,67223,67226,67257,67408,67607,67634,67657,67698,67902,68057,68126,68444,68513,68629,68656,68674,68694,68733,68759,68876,68911,68964,69021,69080,69090,69129,69155,69249,69281,69304,69415,69416,69432,69582,69637,69709,69797,69941,69942,70089,70103,70172,70255,70297,70331,70335,70421,70732,70786,70914,70932,71061,71236,71243,71258,71454,71535,71558,71586,71640,71737,71816,71852,71897,71904,71964,71976,72137,72280,72348,72425,72521,72638,72786,72820,72932,72966,73127,73295,73321,73343,73621,73641,73697,73782,74003,74115,74157,74426,74470,74517,74546,74645,74664,74802,74805,74863,74879,74920,74958,75042,75281,75328,75336,75423,75449,75484,75741,75822,75861,76085,76289,76486,76686,76693,76709,76804,76912,76955,76958,77023,77056,77104,77108,77208,77308,77398,77455,77558,77595,77651,77736,77786,77894,78003,78009,78029,78055,78058,78060,78095,78296,78313,78346,78673,78773,78778,78947,79081,79082,79100,79113,79277,79317,79354,79363,79462,79518,79529,79594,79676,79722,79738,79786,79906,79948,80014,80046,80096,80256,80342,80394,80401,80644,80764,80849,80896,80993,81220,81225,81230,81353,81381,81382,81516,81638,81764,82016,82041,82051,82098,82371,82420,82436,82511,82526,82649,82794,82874,82992,83119,83172,83224,83349,83503,83523,83596,83658,83809,83817,83873,83914,83992,84032,84048,84372,84550,84611,84613,84715,84832,84902,84963,85038,85097,85133,85225,85317,85552,85854,85968,86076,86143,86520,86564,86609,86618,86633,86638,86654,86897,86998,87031,87096,87225,87245,87286,87434,87455,87480,87709,87818,88080,88271,88283,88353,88418,88441,88511,88648,88708,88756,88816,88959,88987,89163,89212,89213,89222,89243,89265,89374,89423,89520,89590,89592,89649,89650,89674,89684,89696,89701,89742,89884,90000,90091,90099,90156,90224,90254,90284,90544,90642,90693,90698,90806,90919,91036,91120,91139,91156,91246,91284,91299,91358,91422,91524,91657,91700,91872,91885,91965,91984,92179,92215,92328,92348,92444,92558,92643,92646,92787,92857,92938,92988,93122,93170,93185,93306,93340,93347,93472,93618,93682,93751,93857,93985,94011,94160,94595,94630,94734,94902,94928,95017,95094,95248,95312,95313,95336,95543,95592,95609,95620,95886,96088,96123,96143,96408,96446,96568,96599,96639,96716,96772,96785,96818,96868,96869,96877,96975,96977,97232,97381,97511,97632,97797,97820,97841,97884,98097,98250,98302,98335,98386,98499,98653,98675,98711,98749,98781,98973,99048,99136,99143,99178,99282,99393,99447,99509,99707,99974,99995",
      "39,166,173,245,249,1672,1730,1787,1958,1996,2110,2152,2278,2846,3113,3150,3497,3641,3757,3814,3817,3845,4010,4020,4089,4275,4304,4311,4479,4483,4519,4783,4819,4992,4998,5602,5659,8611,9366,9382,9387,9482,9521,9539,9558,9596,9653,9672,9691,9710,9805,9930,10014,10063,10071,10119,10185,10261,10603,10641,10698,10717,10754,10755,10774,10793,10812,10925,10945,10969,11059,11078,11135,11154,11173,11192,11230,11249,11270,11282,11306,11382,11399,11439,11456,11496,11515,11534,11553,11629,11648,11686,11876,11880,11895,11896,11952,11971,12028,12039,12077,12104,12123,12161,12180,12237,12256,12275,12294,12484,12494,12503,12560,12584,12598,12617,12636,12655,12693,12712,12750,12769,12771,12774,12834,12845,12902,12959,12997,13035,13054,13073,13187,13244,13269,13282,13301,13377,13422,13434,13466,13509,13548,13550,13566,13571,13605,13624,13643,13681,13700,13738,13795,13814,13846,13852,13928,13966,14004,14023,14137,14138,14156,14194,14213,14251,14270,14289,14327,14384,14426,14468,14479,14517,14536,14574,14593,14612,14631,14650,14654,14707,14783,14807,14824,14840,14845,14875,15030,15049,15144,15217,15220,15223,15277,15296,15353,15372,15391,15404,15410,15429,15433,15486,15556,15581,15600,15638,15657,15676,15733,15752,15771,15799,15809,15828,15856,15866,15885,15911,15923,15961,15965,15980,15983,16056,16075,16132,16137,16189,16208,16223,16227,16265,16284,16303,16322,16341,16360,16379,16398,16417,16436,16474,16531,16569,16607,16645,16721,16740,16759,16854,16911,16930,16945,16949,16987,17025,17044,17063,17101,17110,17120,17148,17158,17196,17215,17273,17288,17289,17294,17329,17348,17367,17386,17424,17462,17489,17519,17538,17576,17595,17669,17671,17728,17766,17769,17785,17804,17823,17842,17861,17899,17906,17918,17937,18013,18032,18127,18165,18184,18198,18241,18279,18298,18317,18355,18374,18393,18412,18431,18450,18488,18507,18564,18602,18678,18697,18754,18773,18792,18811,18868,18925,18949,18963,18987,19001,19020,19039,19077,19096,19111,19115,19210,19229,19267,19286,19343,19362,19365,19390,19400,19495,19514,19571,19628,19685,19704,19742,19761,19780,19803,19856,19875,19894,19932,19951,19970,20008,20027,20065,20084,20103,20118,20141,20179,20217,20262,20274,20312,20350,20369,20407,20445,20464,20483,20502,20540,20578,20635,20654,20673,20711,20730,20777,20825,20842,20844,20863,20865,20882,20958,20977,20999,21058,21059,21091,21094,21129,21194,21205,21224,21319,21357,21376,21414,21452,21471,21547,21585,21604,21680,21718,21737,21756,21785,21794,21813,21832,21851,21908,21927,21946,21965,21984,22022,22079,22095,22117,22136,22193,22250,22269,22307,22326,22383,22402,22421,22440,22459,22478,22497,22516,22535,22554,22565,22573,22611,22628,22630,22687,22706,22725,22744,22782,22801,22818,22819,22877,22886,22896,22953,22972,23067,23105,23109,23124,23162,23181,23238,23276,23279,23295,23305,23314,23352,23390,23407,23409,23504,23523,23542,23565,23599,23626,23637,23656,23713,23732,23770,23780,23808,23865,23875,23884,23903,23922,23941,23960,23979,23998,24017,24055,24112,24131,24169,24207,24226,24240,24302,24357,24359,24376,24378,24416,24435,24454,24473,24492,24511,24530,24549,24606,24625,24649,24701,24739,24758,24777,24796,24834,24853,24910,25005,25062,25072,25081,25138,25157,25161,25195,25233,25252,25271,25309,25328,25347,25423,25442,25461,25537,25544,25556,25594,25613,25702,25708,25727,25757,25765,25784,25803,25848,25861,25879,25898,25917,25936,25955,25974,25993,26012,26031,26050,26069,26145,26212,26221,26240,26241,26297,26316,26317,26354,26363,26373,26380,26392,26411,26430,26506,26639,26658,26675,26772,27000,27019,27028,27113,27228,27344,27357,27361,27380,27440,27456,27478,27513,27524,27563,27646,27783,27836,27848,27898,27902,27958,27969,28015,28026,28102,28216,28293,28501,28596,28691,28710,28729,28756,28816,28824,28924,29004,29033,29052,29058,29075,29095,29158,29172,29299,29375,29418,29419,29510,29575,29660,29699,29717,29719,29774,29831,29870,29950,30001,30030,30059,30088,30117,30146,30175,30204,30233,30262,30320,30349,30378,30407,30436,30465,30494,30523,30552,30572,30581,30610,30639,30668,30697,30726,30755,30784,30813,30842,30871,30900,30929,30958,30987,31016,31045,31074,31103,31132,31161,31190,31219,31248,31277,31306,31335,31364,31393,31422,31451,31480,31509,31538,31567,31596,31625,31654,31712,31741,31770,31799,31828,31857,31886,31915,31944,31973,32002,32031,32060,32089,32118,32147,32176,32205,32234,32263,32292,32321,32350,32379,32408,32437,32466,32495,32524,32553,32582,32611,32640,32669,32698,32727,32756,32785,32814,32843,32872,32901,32930,32988,33017,33046,33075,33104,33133,33162,33191,33220,33249,33278,33307,33336,33365,33394,33423,33452,33481,33510,33539,33568,33597,33626,33655,33684,33713,33742,33771,33800,33829,33858,33887,33916,33945,33974,34003,34032,34061,34090,34119,34148,34177,34206,34235,34264,34293,34322,34351,34380,34409,34435,34438,34467,34496,34525,34554,34583,34612,34641,34670,34699,34728,34757,34786,34815,34844,34873,34902,34931,34960,34989,35001,35076,35082,35090,35148,35164,35239,35253,35300,35327,35376,35395,35401,35472,35513,35520,35553,35573,35579,35616,35638,35711,35741,35770,35829,35900,35942,35959,36018,36071,36081,36141,36200,36271,36324,36330,36362,36365,36389,36424,36436,36441,36461,36520,36579,36643,36672,36702,36743,36761,36801,36832,36891,36950,37022,37059,37081,37140,37204,37263,37292,37306,37322,37393,37452,37511,37583,37642,37701,37765,37824,37853,37883,37954,38013,38042,38047,38058,38072,38144,38203,38262,38293,38326,38374,38385,38444,38515,38574,38618,38633,38684,38705,38751,38764,38823,38887,38920,38946,38978,39005,39076,39135,39166,39167,39190,39193,39194,39266,39313,39325,39357,39384,39434,39448,39507,39566,39599,39637,39696,39755,39798,39810,39827,39886,39945,39974,40113,40142,40171,40200,40229,40258,40287,40316,40345,40374,40403,40432,40461,40490,40519,40548,40577,40606,40635,40664,40693,40722,40751,40780,40809,40838,40867,40896,40925,40954,40983,41012,41041,41070,41099,41128,41157,41186,41215,41244,41273,41302,41331,41360,41389,41418,41447,41476,41505,41534,41700,41789,41791,41793,41795,41797,41799,41801,41803,41805,41809,41811,41813,41815,41817,41819,41821,41823,41825,41827,41829,41831,41833,41835,41837,41839,41841,41843,41845,41847,41851,41853,41855,41857,41859,41861,41863,41865,41867,41869,41871,41873,41875,41877,41879,41881,41883,41885,41887,41889,41891,41893,41895,41897,41903,41905,41907,41909,41911,41913,41915,41917,41919,41921,41923,41927,41930,41933,41936,41939,41942,41945,41948,41951,41954,41958,41962,41966,41970,41974,41978,41982,41986,41991,42020,42049,42078,42107,42136,42155,42164,42196,42205,42209,42286,42327,42361,42374,42375,42436,42468,42523,42541,42542,42598,42673,42674,42699,42701,42719,42732,42760,42835,42910,42952,42971,43037,43072,43105,43113,43147,43152,43167,43174,43186,43204,43234,43309,43334,43338,43384,43405,43471,43485,43507,43546,43561,43576,43621,43664,43708,43718,43719,43751,43783,43800,43858,43865,43919,43945,43953,43982,44020,44035,44045,44095,44105,44182,44201,44225,44257,44292,44301,44332,44351,44384,44393,44419,44466,44494,44511,44569,44591,44656,44692,44731,44737,44750,44806,44808,44818,44843,44845,44893,44929,44935,44939,44968,45043,45052,45054,45056,45061,45070,45073,45116,45176,45232,45284,45305,45356,45361,45416,45417,45473,45479,45493,45543,45604,45609,45680,45730,45750,45791,45850,45864,45867,45917,45929,45978,46001,46007,46008,46023,46077,46144,46177,46243,46300,46348,46394,46398,46404,46410,46482,46487,46520,46527,46657,46736,46745,46806,46863,46878,46887,46914,46963,46967,46988,47050,47063,47091,47111,47121,47122,47143,47159,47228,47319,47320,47329,47339,47356,47366,47405,47421,47436,47437,47480,47481,47517,47564,47568,47574,47581,47594,47643,47678,47694,47713,47773,47783,47790,47792,47793,47819,47828,47829,47885,47917,47937,47941,47981,48010,48025,48028,48062,48113,48123,48126,48134,48150,48154,48161,48229,48230,48232,48304,48313,48317,48367,48393,48398,48401,48451,48489,48513,48584,48588,48605,48658,48665,48682,48706,48716,48746,48776,48790,48794,48809,48814,48817,48819,48836,48863,48920,49172,49189,49196,49219,49246,49274,49296,49346,49395,49403,49443,49501,49519,49530,49599,49627,49707,49737,49742,49767,49847,49852,49873,49882,49953,49957,49991,49992,50029,50188,50195,50314,50373,50410,50417,50455,50470,50583,50735,50858,50894,50981,51213,51290,51405,51414,51441,51544,51571,51647,51672,51701,51761,51782,51798,52108,52175,52194,52196,52346,52425,52437,52469,52497,52618,52793,52847,52880,52943,52944,52980,53123,53379,53424,53449,53573,53616,53768,53794,53798,53861,53969,54022,54027,54066,54174,54179,54226,54375,54511,54615,54795,54850,54919,54934,54967,55216,55274,55309,55316,55406,55477,55518,55534,55564,55580,55615,55664,55723,55769,55846,55887,55958,56078,56125,56252,56280,56331,56338,56427,56461,56522,56773,57037,57088,57107,57128,57162,57230,57287,57300,57323,57334,57358,57382,57524,57542,57560,57679,57726,57745,57804,57914,58349,58350,58393,58476,58611,58616,58698,58793,58841,58942,59130,59188,59294,59525,59575,59686,59715,59772,59776,59822,59886,59957,59980,60011,60061,60087,60378,60516,60579,60593,60673,60691,60722,60849,61142,61341,61375,61377,61428,61567,61588,61674,61703,61843,61870,62197,62219,62335,62641,62706,62757,62948,63040,63056,63076,63109,63184,63392,63395,63427,63696,63724,63895,64359,64369,64544,64614,64681,64781,64842,64846,64854,64868,65008,65194,65203,65208,65263,65469,65488,65564,65789,65823,65841,65906,65937,65968,65977,66008,66063,66095,66142,66154,66163,66465,66492,66552,66573,66678,66757,66791,66800,66885,67287,67405,67464,67539,67696,67837,67895,68107,68115,68361,68364,68378,68457,68515,68569,68601,68611,68641,68684,68862,68935,69335,69523,69766,69782,69792,70044,70269,70309,70339,70363,70634,70720,70918,70958,70973,71074,71098,71153,71215,71273,71290,71322,71476,71567,71574,71691,71696,71697,71729,71784,71825,71827,71960,71994,72081,72167,72349,72745,72833,72862,72975,73005,73061,73071,73207,73212,73486,73507,73614,73618,73841,73880,74080,74141,74144,74269,74308,74393,74417,74455,74506,74558,74561,74746,74978,75095,75173,75333,75413,75509,75628,75758,75790,75811,75969,75980,76333,76386,76429,76432,76506,76547,76586,76667,76689,76742,76841,76896,77053,77350,77480,77614,77643,77649,77832,77862,77905,78008,78074,78102,78148,78190,78258,78287,78318,78351,78392,78563,78689,78791,78796,78803,78845,78870,79044,79175,79236,79414,79469,79557,79559,79664,79685,79797,79925,79989,80055,80097,80144,80165,80248,80302,80410,80454,80682,80705,80720,80893,80910,80959,81101,81161,81444,81550,81623,81639,81698,81737,81767,81834,81869,81870,81873,81991,82026,82079,82103,82119,82163,82181,82281,82476,82534,82557,82579,82595,82604,82628,82675,82761,82924,83276,83502,83543,83567,83601,83630,83729,83797,83862,83890,83974,83981,84092,84224,84397,84419,84489,84494,84530,84575,84585,84587,84648,84754,84875,85006,85648,85693,85799,85831,85853,85913,85922,86058,86353,86397,86436,86531,86605,86717,86825,87249,87287,87384,87391,87712,88022,88067,88227,88242,88245,88539,88587,88769,88843,88974,89074,89123,89132,89208,89289,89503,89714,89733,89751,89765,89895,90104,90819,91026,91125,91128,91130,91334,91374,91417,91659,91724,91731,91743,91767,91793,91832,91839,91985,92031,92087,92248,92297,92303,92786,93027,93339,93399,93533,93553,93560,93602,93609,93747,93994,94091,94117,94195,94252,94418,94784,94887,94909,95036,95182,95264,95578,95664,95748,95942,95967,96048,96132,96162,96273,96306,96442,96505,96675,96793,96803,96882,97155,97167,97299,97433,97435,97509,97629,97637,97710,97741,97770,97917,97984,98104,98160,98219,98220,98280,98403,98474,98566,98632,98721,98753,98882,98885,98920,98965,99021,99114,99144,99168,99365,99377,99474,99525,99545,99950,99983",
      "12,42,72,145,188,237,248,498,947,1646,1648,1661,1733,1754,1796,1797,1810,1924,1955,1984,2016,2025,2034,2042,2046,2056,2060,2103,2159,2163,2182,2282,2305,2332,2351,2357,2363,2375,2379,2381,2390,2400,2404,2474,2490,2499,2504,2507,2508,2519,2573,2580,2593,2624,2633,2646,2660,2714,2721,2747,2753,2758,2764,2788,2837,2844,2845,2867,2891,2892,2922,2923,2927,2938,2979,2992,3018,3031,3040,3077,3088,3096,3157,3162,3185,3207,3219,3246,3255,3320,3323,3326,3335,3359,3364,3400,3418,3419,3424,3430,3437,3455,3459,3466,3470,3483,3534,3535,3541,3548,3569,3585,3592,3594,3610,3631,3639,3649,3689,3716,3717,3719,3736,3745,3758,3760,3761,3767,3781,3783,3796,3807,3835,3839,3840,3851,3857,3858,3864,3880,3893,3927,3944,3966,3983,3985,3986,4001,4014,4024,4054,4055,4091,4116,4119,4169,4201,4223,4231,4255,4258,4262,4265,4287,4303,4322,4349,4375,4408,4449,4456,4470,4500,4523,4529,4565,4569,4614,4650,4675,4683,4685,4710,4719,4799,4826,4875,4898,4901,4905,4916,4972,4980,4989,5304,5610,5633,5638,6490,6527,6534,6561,6563,6601,6619,6629,6696,6806,6841,6858,6862,6871,6882,6891,6915,6922,6925,6950,6968,6993,7150,7183,7209,7249,7316,7332,7351,7394,7416,7428,7456,7586,7602,7681,7696,7732,7766,7805,7846,7848,7864,7919,7931,7956,7989,8075,8203,8261,8266,8304,8347,8386,8402,8407,8522,8706,8802,8882,8996,9066,9091,9115,9212,9223,9250,9271,9274,9405,9434,9487,9526,9543,9582,9622,9761,9880,10072,10136,10183,10193,10257,10559,10579,10715,10855,10958,11019,11167,11322,11398,11508,11651,11656,11688,11762,11868,11985,12205,12270,12531,12594,12646,12722,12763,12790,12795,12975,13065,13306,13392,13442,13709,13881,13987,14036,14103,14227,14325,14376,14604,14837,14890,14924,15029,15078,15110,15112,15286,15355,15482,15491,15535,15572,15751,16014,16151,16425,16643,16683,16695,16748,16754,16827,16938,17136,17213,17272,17461,17579,17778,17835,17872,18005,18088,18089,18291,18377,18607,18654,18717,19301,19531,19611,19743,19805,19829,19858,19919,20010,20236,20385,20392,20428,20467,20690,20703,20761,20818,20927,21012,21210,21259,21381,21673,21953,21972,22034,22062,22157,22309,22395,22445,22503,22577,22638,22843,22956,23133,23448,23467,23555,23674,23695,23806,23820,23856,23912,23943,24035,24036,24071,24152,24599,24665,24866,24929,25409,25441,25478,25506,25646,25674,25804,25894,26077,26234,26541,26614,26638,26719,26721,26840,26940,27041,27051,27053,27065,27076,27118,27149,27266,27299,27373,27381,27396,27410,27442,27468,27469,27472,27473,27488,27493,27508,27511,27517,27559,27565,27566,27573,27595,27609,27619,27661,27674,27679,27760,27775,27917,27955,27978,27993,27999,28011,28013,28018,28022,28048,28065,28067,28081,28106,28114,28115,28143,28150,28206,28245,28261,28340,28354,28444,28445,28485,28519,28540,28556,28559,28565,28591,28619,28658,28671,28687,28713,28786,28788,28811,28825,28859,28884,28908,28917,28923,28944,28967,28990,28997,29007,29022,29109,29115,29131,29157,29167,29168,29177,29205,29210,29242,29254,29272,29321,29330,29343,29367,29368,29385,29412,29456,29547,29561,29602,29606,29626,29646,29672,29683,29708,29729,29730,29743,29761,29770,29791,29804,29813,29838,29842,29931,29932,29945,29960,29969,29980,30012,30041,30070,30099,30128,30157,30186,30215,30244,30331,30360,30389,30418,30447,30476,30505,30534,30563,30621,30650,30679,30708,30737,30766,30795,30824,30853,30882,30911,30940,30969,30998,31027,31056,31085,31114,31143,31172,31201,31230,31259,31288,31317,31346,31375,31404,31433,31462,31491,31520,31549,31578,31607,31636,31665,31694,31723,31752,31781,31810,31839,31868,31897,31926,31955,31984,32013,32042,32071,32100,32129,32158,32187,32216,32245,32274,32303,32332,32361,32390,32419,32448,32477,32506,32535,32564,32593,32622,32651,32680,32709,32738,32767,32796,32825,32854,32883,32912,32941,32970,32999,33028,33057,33086,33115,33144,33173,33202,33231,33260,33289,33318,33347,33376,33405,33463,33492,33521,33550,33579,33608,33637,33666,33695,33724,33753,33782,33811,33840,33869,33898,33927,33956,34014,34043,34072,34101,34130,34159,34188,34217,34246,34275,34304,34333,34391,34420,34449,34478,34507,34536,34565,34594,34623,34647,34652,34681,34710,34739,34768,34797,34826,34855,34884,34913,34942,34971,35000,35012,35074,35101,35175,35197,35202,35236,35264,35315,35372,35412,35483,35531,35590,35612,35649,35722,35781,35840,35911,35970,35996,36029,36092,36110,36120,36140,36152,36181,36211,36304,36341,36400,36420,36421,36472,36531,36590,36654,36688,36713,36748,36753,36772,36809,36843,36902,36961,37033,37092,37118,37151,37195,37215,37253,37254,37274,37333,37404,37423,37463,37485,37507,37522,37542,37594,37653,37712,37776,37835,37894,37936,37953,37965,38024,38083,38155,38177,38214,38235,38273,38313,38337,38396,38419,38455,38491,38526,38561,38585,38644,38664,38670,38716,38735,38775,38803,38834,38855,38924,38957,39016,39037,39087,39146,39205,39277,39336,39395,39431,39438,39459,39518,39548,39555,39561,39577,39648,39707,39766,39789,39838,39897,39956,40106,40135,40164,40193,40222,40251,40280,40309,40338,40367,40396,40425,40483,40512,40541,40570,40599,40628,40657,40686,40715,40744,40773,40802,40831,40860,40889,40918,40947,40976,41005,41034,41063,41092,41121,41150,41172,41179,41208,41237,41266,41295,41324,41353,41382,41411,41469,41498,41527,41999,42028,42057,42086,42115,42175,42255,42265,42334,42340,42387,42415,42443,42461,42502,42577,42622,42624,42635,42652,42687,42739,42768,42814,42864,42881,42889,42918,42941,42976,43001,43126,43158,43176,43190,43213,43238,43288,43363,43389,43418,43422,43450,43508,43525,43567,43584,43600,43687,43762,43837,43889,43895,43924,43955,43963,43965,43999,44032,44074,44161,44213,44236,44311,44337,44345,44369,44378,44398,44449,44460,44468,44504,44513,44535,44548,44576,44579,44596,44611,44627,44635,44679,44687,44694,44710,44757,44767,44785,44829,44947,45005,45022,45058,45095,45135,45143,45204,45262,45328,45335,45368,45372,45392,45432,45450,45475,45522,45545,45559,45579,45637,45664,45678,45709,45748,45759,45766,45824,45889,45896,45945,45953,45980,46029,46121,46201,46216,46289,46312,46323,46327,46339,46344,46376,46407,46430,46440,46492,46553,46554,46583,46613,46621,46651,46667,46720,46747,46763,46764,46771,46810,46875,46909,46931,46932,46951,46980,46989,46994,47077,47090,47100,47207,47209,47256,47292,47299,47331,47393,47441,47461,47525,47559,47615,47667,47693,47698,47705,47731,47808,47888,47910,47957,47978,47980,47995,48022,48040,48045,48115,48118,48141,48172,48200,48256,48268,48276,48356,48368,48369,48372,48383,48408,48466,48488,48507,48550,48570,48599,48638,48676,48736,48753,48759,48821,48887,48923,48958,49000,49001,49101,49103,49128,49132,49154,49164,49236,49369,49370,49384,49408,49430,49494,49497,49515,49525,49555,49570,49607,49633,49638,49703,49723,49741,49749,49815,49862,50008,50207,50222,50236,50265,50612,50764,50768,50777,50792,50842,51208,51215,51304,51379,51519,51843,52179,52240,52750,52751,52863,52978,53032,53210,53320,53490,53557,53762,53808,53963,54085,54113,54166,54187,54314,54688,54755,54809,54930,54976,55192,55344,55365,55379,55410,55473,55567,55602,55628,55644,55809,55860,55925,55967,56099,56157,56178,56202,56267,56376,56513,56682,56844,56903,56939,56941,57001,57110,57130,57145,57157,57245,57274,57282,57299,57351,57594,57631,57705,57895,58025,58033,58093,58094,58104,58106,58206,58289,58340,58370,58405,58426,58576,58595,58647,58691,58757,58811,58883,58995,59002,59079,59117,59209,59304,59309,59355,59356,59439,59444,59536,59557,59625,59627,59809,59995,60018,60144,60279,60350,60416,60426,60539,60590,60655,60695,60723,60896,60937,61195,61330,61386,61418,61499,61506,61540,61630,61916,61934,61935,61949,62004,62039,62084,62112,62179,62247,62407,62414,62493,62527,62585,62692,62721,62787,62883,62919,62923,62956,62977,63191,63270,63300,63604,63636,63711,63940,63972,64070,64224,64490,64511,64655,64771,64801,64826,64843,64929,65066,65122,65136,65250,65373,65381,65533,65563,65696,65902,66042,66245,66251,66353,66432,66480,66489,66502,66567,66601,66618,66752,66826,66931,67026,67052,67061,67146,67247,67292,67412,67422,67477,67510,67525,67583,67610,67693,67733,67777,67959,68023,68440,68496,68749,68904,69030,69235,69373,69399,69425,69458,69519,69521,69561,69634,69639,69786,69924,69987,70127,70162,70283,70390,70569,70657,70719,70739,70764,70805,70891,70902,70935,71006,71016,71028,71110,71181,71187,71284,71295,71310,71427,71448,71497,71589,71731,71930,72044,72248,72316,72328,72364,72374,72440,72634,72642,72837,72866,72914,72968,72974,73056,73221,73259,73359,73405,73493,73530,73539,73555,73606,73609,73611,73773,73792,73853,73930,73999,74032,74186,74232,74243,74435,74640,74657,74758,74798,74870,74873,74923,75293,75473,75569,75634,75642,75669,75727,75820,76176,76287,76302,76327,76342,76404,76504,76526,76622,76716,76770,76927,77162,77346,77364,77425,77505,77637,77691,77885,77969,78028,78239,78294,78604,78651,78696,78724,78766,78918,78983,79060,79144,79146,79210,79334,79355,79422,79498,79519,79537,79544,79669,79795,79828,79973,80077,80095,80204,80205,80294,80387,80642,80646,80695,80748,80751,80829,80901,80943,80974,81011,81066,81162,81197,81354,81395,81519,81636,81794,82056,82113,82127,82159,82284,82331,82381,82446,82662,82715,82771,82829,82979,83104,83123,83138,83252,83407,83413,83426,83535,83613,83727,84034,84104,84254,84337,84349,84374,84384,84540,84725,84776,84833,84962,85174,85276,85379,85532,85585,85617,85675,85780,85887,86044,86099,86164,86198,86224,86306,86355,86423,86478,86525,86532,86543,86595,86689,86691,86714,86719,86734,86802,86886,86963,86986,87094,87128,87192,87395,87653,87683,87827,87879,87891,87956,88171,88262,88268,88277,88416,88468,88480,88626,88629,88669,88789,88803,88835,89031,89100,89154,89174,89206,89226,89325,89327,89331,89573,89660,89762,89774,89777,90050,90064,90385,90402,90573,90691,90740,90914,91031,91044,91103,91205,91226,91420,91561,91606,91631,91759,91930,91951,91953,92034,92038,92118,92372,92439,92441,92512,92578,92670,92681,92830,92909,93030,93244,93458,93595,93727,94050,94222,94263,94284,94346,94362,94432,94602,94905,94964,94976,94995,95072,95079,95167,95291,95367,95422,95581,95698,95707,95901,96178,96201,96227,96276,96299,96356,96456,96501,96685,96724,96801,96825,96857,96924,97045,97054,97085,97107,97461,97479,97502,97571,97838,97970,98278,98313,98394,98495,98567,98646,98700,98701,98730,98771,98786,98796,98845,98846,99049,99128,99181,99309,99316,99592,99739,99757,99844,99906,99934,99958",
      "176,276,477,608,637,652,665,668,726,738,752,827,897,943,974,1045,1136,1153,1162,1208,1255,1267,1335,1424,1457,1473,1480,1559,1592,1597,1611,1615,1635,1644,1674,1748,1841,2030,2188,2195,2589,2638,2693,2742,2773,2792,2799,2902,2928,3046,3103,3126,3160,3179,3226,3333,3647,3667,3706,3792,3828,4198,4210,4319,4333,4354,4559,4582,4667,4671,4720,4729,4749,4945,5008,5035,5084,5110,5122,5138,5140,5178,5190,5201,5203,5205,5241,5255,5259,5298,5301,5327,5339,5388,5400,5512,5514,5578,5641,5706,5742,5745,5865,5938,5969,6003,6031,6038,6058,6148,6182,6185,6218,6229,6251,6468,6522,6548,6597,6822,7077,7080,7187,7286,7288,7299,7405,7411,7476,7486,7525,7561,7666,7714,7741,7780,7879,7894,7905,7927,8005,8029,8120,8245,8382,8501,8679,8728,8816,9077,9230,9313,9344,9374,9383,9385,9423,9428,9736,9815,9898,10003,10287,10337,10425,10439,10464,10475,10532,10687,10874,10934,11274,11378,11411,11463,11673,11746,11870,12383,12433,12470,12527,13037,13105,13361,13440,13935,13955,14063,14339,14530,14607,14828,15340,15401,15472,15804,15890,16005,16616,16714,16867,16955,17032,17054,17064,17223,17518,17871,18881,19281,19370,19479,19720,20034,20264,20408,20663,20757,20896,21499,21544,21692,21842,22521,22796,23783,23829,24452,24812,25070,25093,25171,26162,26739,26774,26901,26909,27219,27432,27525,27854,28190,28210,28229,28244,28289,28709,28728,28732,29277,29991,30333,30596,30681,31522,31696,32363,32537,33001,33291,33320,33849,34016,34422,35039,35450,35621,35993,36446,36499,37248,37382,37444,37692,37837,38128,38132,38157,38187,38304,38492,38646,38744,38959,39004,39118,39173,39315,39360,39557,39626,39679,39732,39875,39992,41517,42103,42493,42610,42932,43164,43467,43519,43520,43654,44260,44585,45231,45370,45437,45556,45558,45732,45737,45874,45879,45919,45998,46126,46984,47982,48084,49754,49849,49996,50104,50106,50119,50251,50252,50341,50433,50511,50673,50718,50733,50814,50827,50985,51090,51138,51178,51206,51242,51265,51359,51705,51942,51990,52203,52257,52323,52325,52412,52427,52531,52567,52636,52663,52799,53025,53191,53349,53595,53677,53702,54261,54455,54464,54496,54660,54715,54889,54988,55095,55117,55121,55147,55300,55393,55498,55549,55647,55690,55703,55715,55795,56089,56105,56253,56293,56396,56412,56470,56550,56789,57010,57030,57047,57070,57080,57200,57206,57293,57373,57422,57481,57494,57535,57571,57893,57896,57969,57977,58092,58115,58361,58496,58524,58557,58799,58859,58879,58898,58915,58924,59042,59058,59204,59226,59446,59549,59566,59798,59837,59865,59872,59979,60002,60219,60399,60403,60454,60476,60634,60733,60744,60848,60914,61004,61066,61077,61100,61134,61224,61235,61296,61361,61425,61445,61578,61796,61858,62067,62164,62246,62277,62288,62366,62368,62401,62408,62555,62656,62962,62995,63026,63028,63042,63090,63205,63282,63349,63401,63410,63507,63554,63611,63621,63683,63735,63761,63826,63860,63907,63988,64003,64077,64106,64143,64227,64277,64523,64528,64580,64786,64829,64849,64968,64978,64981,65054,65061,65121,65237,65437,65452,65656,65709,65784,65925,66055,66167,66216,66226,66312,66347,66430,66515,66577,66580,66603,66662,66805,66888,67116,67215,67262,67415,67435,67751,67888,67896,67950,68015,68071,68137,68197,68483,68527,68608,68628,68789,68805,68842,68892,68916,69182,69185,69216,69264,69481,69495,69567,69568,69789,69821,69881,69959,70068,70078,70087,70126,70138,70211,70276,70281,70310,70315,70378,70415,70497,70505,70552,70616,70665,70846,70936,70984,70994,70998,71029,71108,71156,71373,71378,71423,71513,71561,71580,71668,71678,71822,71828,71895,72014,72043,72116,72128,72141,72204,72252,72486,72497,72552,72555,72579,72657,72732,72890,73001,73291,73301,73326,73333,73496,73497,73756,73941,73964,74272,74357,74638,74730,74929,75085,75096,75134,75140,75170,75192,75274,75302,75332,75427,75435,75472,75584,75606,75718,75792,75866,75912,75976,76014,76077,76185,76264,76286,76366,76450,76502,76561,76578,76650,76849,76850,76892,76915,77170,77177,77199,77220,77242,77302,77370,77400,77452,77473,77650,77684,77733,77769,78402,78515,78545,78603,78740,78754,78988,79008,79059,79128,79133,79158,79196,79284,79338,79421,79439,79488,79533,79752,79791,79867,79936,79969,80026,80154,80171,80197,80208,80210,80229,80270,80369,80699,80886,80890,81003,81010,81025,81093,81309,81412,81423,81483,81813,81960,82152,82171,82183,82289,82333,82396,82405,82684,82693,82766,82781,82911,82921,82959,83092,83233,83358,83364,83400,83422,83562,83646,83700,83711,83807,83878,83952,84009,84047,84067,84157,84292,84368,84552,84621,84794,84796,84864,85109,85279,85376,85381,85472,85515,85529,85716,85729,85784,85807,85875,85935,86120,86123,86162,86331,86388,86391,86407,86723,86805,86837,86863,86883,86905,86924,86989,87376,87379,87478,87498,87575,87634,87681,87937,88039,88063,88202,88220,88254,88311,88399,88408,88411,88430,88526,88614,88639,88652,88659,88696,88870,89178,89553,89564,89641,89663,89863,90012,90114,90192,90279,90449,90560,90627,90760,90788,90880,90886,91002,91168,91364,91563,91564,91600,91626,91641,91673,91856,91883,91896,91971,92019,92020,92183,92243,92445,92616,92752,92777,92979,93013,93062,93114,93151,93262,93680,93838,94051,94106,94114,94273,94422,94456,94529,94540,94596,94718,94863,95018,95070,95438,95616,95768,95797,95869,95947,95979,95984,96035,96083,96163,96190,96208,96253,96339,96355,96384,96416,96450,96480,96498,96723,96880,97033,97059,97110,97180,97193,97273,97377,97429,97619,97626,97842,97851,98029,98061,98157,98236,98329,98346,98372,98430,98516,98699,98788,98892,98998,99051,99112,99137,99226,99347,99389,99439,99535,99548,99588,99646,99673,99755,99762,99812,99901,99926",
      "13,284,288,351,360,445,453,541,557,612,680,695,737,758,770,775,798,801,855,902,963,964,971,983,1049,1129,1300,1332,1453,1461,1472,1536,1543,1564,1753,1802,1812,1831,1855,1898,1905,1908,2816,3158,3949,4759,5071,5148,5236,5411,5588,5627,5723,5831,5963,6131,6411,6566,6620,6712,6845,6962,7000,7060,7071,7086,7112,7126,7190,7203,7278,7321,7341,7376,7390,7435,7507,7514,7526,7531,7557,7569,7654,7660,7672,7701,7711,7745,7756,7785,7800,7858,7967,8027,8037,8065,8070,8110,8213,8257,8267,8303,8306,8334,8344,8372,8484,8513,8637,8746,8807,8856,8908,8934,9043,9073,9415,9440,9603,9890,9958,10026,10125,10144,10209,10284,10382,10506,10650,10658,10732,10920,11132,11431,11500,11605,11773,11823,11842,11893,12075,12110,12230,12379,12540,12614,12728,12832,12878,12976,13299,13582,13825,13948,14118,14387,14609,14624,14708,15136,15231,15987,16091,16837,17121,17287,17388,18473,18561,18659,19266,20090,20137,20696,20857,20952,21248,21337,21527,21687,22039,22143,22616,22854,22975,23317,23397,23562,23638,24033,24132,24161,24685,24848,25182,25362,25575,25611,25806,26617,26701,26791,26842,26853,27243,27647,27705,27742,27756,27861,28248,28267,28371,28431,29341,29361,29431,29578,29953,32367,34221,35516,36116,36173,36264,36694,36830,36981,36993,37052,37198,37320,37429,37504,38136,38196,38244,38700,39444,39560,39671,39884,42034,44367,45127,45297,45360,45924,45975,48882,50003,50016,50077,50218,50282,50309,50367,50666,50669,50696,50727,50857,50943,50966,50997,51000,51086,51093,51181,51237,51333,51389,51400,51455,51561,51626,51671,51711,51774,51794,51842,51853,51866,51891,52199,52553,52573,52670,52748,52827,52855,52946,52952,53058,53189,53305,53347,53375,53378,53380,53409,53507,53669,53911,54269,54352,54362,54364,54433,54449,54486,54563,54583,54665,54764,54948,55002,55022,55037,55233,55247,55445,55447,55485,55670,55799,55857,55868,55875,56032,56112,56124,56189,56348,56356,56365,56378,56730,56870,56961,56981,57100,57152,57216,57259,57316,57339,57366,57468,57525,57536,57575,57608,57686,57783,57840,57849,57879,57892,57902,58046,58054,58137,58156,58216,58224,58246,58251,58294,58417,58522,58594,58599,58678,58682,58735,58801,58810,58903,59164,59350,59472,59521,59810,59871,59927,59985,60245,60305,60340,60354,60544,60638,60648,60791,60919,60973,61272,61297,61480,61483,61493,61642,61659,61780,61897,61913,61953,62426,62433,62483,62674,62679,62693,62964,63108,63428,63499,63726,63741,63769,63983,63992,64085,64212,64229,64282,64306,64496,64597,64700,64813,64872,64902,64917,65048,65245,65254,65308,65487,65506,65534,65598,65621,65646,65648,65666,65682,65735,65808,65873,65881,66129,66439,66466,66550,66821,66829,66835,66853,66934,67070,67218,67336,67342,67377,67410,67425,67613,67976,67994,68232,68408,68495,68523,68710,68743,68928,69006,69057,69108,69135,69199,69277,69332,69461,69484,69502,69535,69538,69558,69566,69573,69625,69769,69785,69819,69911,69956,70030,70045,70086,70219,70249,70347,70517,70628,70673,70729,70818,71009,71022,71173,71205,71316,71341,71362,71376,71628,71763,72021,72048,72072,72136,72214,72227,72286,72291,72507,72772,72839,72859,72878,72928,72930,72936,73185,73563,73718,73905,73935,73938,73984,74024,74034,74045,74082,74088,74289,74373,74665,74713,74924,74927,74984,75086,75108,75167,75196,75366,75429,75470,75513,75688,75833,75951,76001,76018,76145,76241,76323,76369,76483,76495,76605,76616,76708,76807,76832,76856,76956,76986,77025,77033,77126,77176,77255,77293,77307,77423,77549,77609,77647,77949,78042,78076,78169,78188,78222,78230,78264,78281,78300,78398,78456,78934,79046,79189,79219,79318,79337,79361,79500,79538,79563,79576,79582,79592,79615,79892,79957,79959,79985,80016,80039,80047,80049,80101,80157,80262,80335,80617,80637,80852,80891,81049,81064,81074,81114,81181,81296,81375,81388,81408,81448,81584,81624,81969,82108,82110,82133,82202,82232,82239,82243,82337,82670,82681,82699,82744,82760,82899,83208,83449,83488,83511,83590,83660,83769,83827,83916,83960,84051,84080,84232,84350,84405,84442,84479,84486,84509,84604,84618,84632,84738,84779,84809,84829,84836,84886,85140,85142,85269,85329,85363,85369,85504,85667,85672,85714,85816,85865,85965,86047,86052,86109,86214,86235,86248,86455,86536,86671,86792,86918,86985,87005,87006,87084,87141,87158,87200,87243,87305,87463,87528,87654,87805,87857,87940,88000,88013,88286,88304,88358,88423,88500,88504,88632,88663,88682,88784,88849,88933,89055,89127,89148,89205,89569,89605,89620,89671,89697,89735,89771,89773,89802,89838,89910,89993,90081,90305,90409,90444,90479,90537,90555,90559,90666,90695,90718,90859,91086,91173,91298,91410,91542,91555,91825,91865,91869,91870,91916,91959,92017,92059,92126,92298,92353,92477,92623,92652,92669,92790,92866,92882,92928,93016,93024,93052,93135,93231,93409,93459,93693,93714,93996,93999,94059,94100,94177,94583,94609,94686,94723,94725,94733,94770,94776,94859,94898,94912,94942,95090,95131,95145,95388,95426,95483,95761,95918,95975,95992,96012,96478,96542,96623,96733,96755,96858,97018,97057,97103,97133,97248,97354,97441,97463,97570,97594,97611,97622,97781,97795,97817,97931,97964,98179,98189,98231,98296,98556,98647,98906,98976,99002,99074,99120,99154,99236,99404,99408,99441,99460,99485,99697,99705,99738,99782,99981",
      "61,69,76,90,128,130,1660,1684,1722,1760,1821,1925,2160,2216,2306,2513,2657,2722,2789,2834,2861,2957,2976,3072,3093,3147,3215,3232,3287,3308,3332,3612,3619,3749,3824,4078,4161,4293,4367,4374,4427,4455,4465,4469,4506,4603,4634,4649,4738,4769,4785,4787,5000,5161,6377,6659,7156,7977,8030,8049,8125,8151,8315,8429,8505,8562,8581,8638,8676,9037,9056,9075,9588,9602,9626,9664,9683,9702,9724,9778,9816,9837,9987,9988,10006,10082,10120,10177,10405,10610,10614,10633,10710,10747,10804,10828,10896,10956,10975,11032,11051,11222,11241,11298,11338,11409,11412,11465,11473,11474,11514,11523,11596,11621,11640,11735,11963,12001,12050,12058,12066,12089,12115,12153,12168,12202,12229,12248,12259,12267,12286,12343,12349,12381,12430,12457,12459,12538,12628,12697,12704,12761,12802,12875,12977,12979,12989,13028,13084,13100,13120,13170,13174,13190,13236,13241,13274,13293,13312,13350,13464,13483,13597,13616,13635,13692,13730,13749,13768,13787,13844,13882,13939,13958,13977,14034,14041,14110,14167,14205,14243,14262,14338,14346,14357,14395,14414,14433,14452,14471,14528,14557,14566,14613,14642,14680,14718,14756,14762,14775,14832,14870,14908,14912,14916,14927,14965,15003,15022,15041,15057,15060,15079,15081,15155,15193,15250,15269,15307,15326,15364,15383,15389,15421,15497,15516,15592,15646,15649,15687,15744,15746,15763,15782,15820,15839,15858,15896,15929,15953,15974,15991,16029,16048,16143,16149,16176,16181,16191,16197,16200,16238,16257,16276,16295,16314,16352,16390,16409,16438,16447,16466,16507,16535,16542,16561,16580,16619,16656,16727,16751,16789,16808,16865,16903,16960,17017,17055,17112,17150,17188,17245,17295,17302,17305,17321,17378,17435,17473,17511,17549,17574,17599,17644,17672,17739,17796,17815,17834,17853,17891,17910,17929,17986,18024,18043,18119,18138,18157,18213,18214,18290,18309,18347,18366,18385,18404,18480,18518,18556,18613,18651,18689,18716,18746,18774,18791,18860,18879,18898,18904,18999,19031,19050,19063,19107,19126,19141,19221,19240,19278,19322,19339,19354,19368,19373,19392,19430,19441,19449,19506,19525,19537,19553,19561,19563,19582,19590,19620,19658,19734,19748,19753,19772,19791,19867,19886,19905,19924,20000,20019,20026,20038,20058,20101,20114,20133,20190,20247,20285,20304,20323,20342,20418,20456,20494,20551,20564,20589,20627,20633,20646,20649,20665,20741,20779,20817,20836,20893,20961,20969,21007,21064,21083,21178,21183,21216,21257,21270,21292,21301,21313,21330,21340,21349,21387,21425,21444,21463,21518,21539,21577,21596,21640,21656,21691,21710,21729,21767,21786,21805,21817,21843,21900,21919,21938,21957,21990,21995,22014,22033,22052,22057,22090,22109,22130,22147,22152,22166,22204,22242,22261,22318,22324,22394,22413,22470,22489,22508,22527,22603,22641,22698,22755,22774,22793,22812,22888,22907,22912,22926,22954,23002,23040,23059,23062,23078,23116,23128,23154,23158,23173,23267,23268,23287,23306,23325,23439,23469,23477,23485,23496,23513,23515,23533,23534,23553,23572,23607,23619,23629,23650,23686,23698,23721,23724,23735,23743,23800,23838,23857,23876,23887,23895,23914,23933,23971,23990,24047,24066,24085,24142,24171,24180,24187,24188,24199,24218,24256,24289,24295,24332,24351,24389,24427,24465,24522,24541,24560,24579,24617,24636,24674,24693,24731,24769,24788,24845,24860,24876,24902,24921,24959,24977,24997,25013,25035,25047,25054,25092,25149,25168,25206,25263,25277,25320,25339,25396,25404,25407,25415,25460,25472,25491,25529,25548,25586,25616,25662,25700,25739,25758,25776,25814,25833,25852,25890,25947,26004,26023,26042,26118,26137,26460,26477,26479,26517,26593,26745,26775,26859,26896,26916,26928,26935,26946,27079,27087,27106,27144,27158,27182,27220,27258,27276,27332,27353,27486,27553,27554,27581,27594,27656,27695,27714,27727,27809,27942,27980,28037,28075,28162,28238,28375,28391,28417,28436,28490,28547,28550,28569,28714,28751,28911,29063,29099,29228,29285,29320,29329,29416,29469,29500,29519,29858,29886,29907,29994,30013,30042,30071,30100,30129,30158,30187,30216,30245,30274,30303,30332,30361,30390,30419,30448,30477,30506,30535,30564,30593,30622,30651,30680,30709,30738,30767,30796,30825,30854,30883,30912,30941,30970,30999,31028,31057,31086,31115,31144,31173,31202,31231,31260,31289,31318,31347,31376,31405,31434,31463,31492,31521,31550,31579,31608,31637,31666,31695,31724,31753,31782,31811,31840,31869,31898,31927,31985,32014,32043,32072,32101,32130,32159,32188,32246,32275,32304,32333,32362,32391,32420,32449,32478,32507,32536,32565,32594,32623,32652,32681,32710,32739,32768,32797,32826,32855,32884,32913,32942,32971,33000,33029,33058,33087,33116,33145,33174,33203,33232,33261,33290,33319,33348,33377,33406,33435,33464,33493,33522,33551,33580,33609,33638,33667,33696,33725,33754,33783,33812,33841,33870,33899,33928,33957,33986,34015,34044,34073,34102,34131,34160,34189,34218,34247,34276,34305,34334,34363,34392,34421,34450,34479,34508,34537,34566,34595,34624,34653,34682,34711,34740,34769,34798,34827,34856,34885,34914,34943,34972,35013,35102,35121,35163,35176,35237,35265,35284,35306,35320,35323,35339,35364,35413,35434,35459,35484,35514,35532,35626,35650,35668,35674,35723,35758,35782,35841,35912,35971,36030,36093,36128,36153,36192,36212,36283,36329,36342,36369,36401,36473,36504,36532,36591,36655,36714,36751,36773,36844,36903,36962,36992,37034,37093,37113,37125,37152,37173,37179,37185,37216,37334,37368,37433,37464,37523,37595,37654,37713,37741,37777,37796,37836,37895,37966,38002,38025,38063,38084,38111,38143,38156,38215,38274,38338,38397,38456,38474,38527,38586,38667,38717,38734,38763,38776,38835,38899,38958,38996,39017,39075,39088,39125,39147,39172,39206,39278,39307,39324,39337,39396,39437,39460,39496,39519,39578,39597,39605,39613,39649,39708,39860,39898,39925,39957,39986,40123,40152,40181,40210,40239,40268,40297,40326,40355,40384,40413,40442,40471,40500,40529,40558,40587,40616,40645,40674,40703,40732,40761,40790,40819,40848,40877,40906,40935,40964,40993,41022,41051,41080,41109,41167,41196,41225,41254,41283,41312,41341,41370,41399,41428,41457,41486,41515,41544,42072,42101,42130,42190,42219,42223,42229,42280,42297,42355,42390,42394,42430,42480,42487,42517,42540,42592,42604,42667,42733,42754,42807,42829,42848,42857,42904,42991,43040,43066,43091,43141,43228,43303,43350,43378,43465,43480,43540,43551,43553,43558,43564,43582,43615,43683,43702,43716,43727,43777,43852,43871,43939,43970,43986,44014,44059,44066,44089,44103,44112,44115,44132,44148,44176,44194,44211,44251,44258,44288,44326,44394,44413,44424,44488,44538,44563,44608,44650,44725,44733,44745,44769,44800,44831,44834,44887,44905,44908,44962,45037,45079,45080,45110,45158,45219,45277,45350,45382,45407,45465,45476,45537,45549,45594,45652,45724,45781,45839,45861,45911,45968,45994,46051,46060,46256,46263,46315,46333,46345,46362,46393,46405,46409,46414,46419,46442,46459,46460,46481,46496,46540,46587,46690,46757,46790,46802,46814,46824,46839,46840,46910,46934,46990,46996,47020,47021,47083,47097,47098,47123,47126,47138,47201,47244,47348,47352,47438,47440,47464,47478,47479,47492,47516,47589,47602,47642,47670,47704,47770,47842,47848,47909,47963,47965,47984,48046,48065,48211,48290,48308,48346,48365,48467,48525,48527,48535,48558,48560,48564,48601,48603,48698,48699,48734,48741,48750,48777,48782,48820,48853,48867,48872,48912,48939,48955,48959,48977,48986,49012,49020,49036,49063,49075,49146,49151,49166,49258,49266,49278,49321,49337,49349,49350,49367,49407,49411,49435,49448,49596,49612,49629,49632,49668,49672,49715,49756,49788,49790,49816,49859,49930,49944,49951,49974,49985,50151,50209,50228,50230,50247,50302,50339,50409,50474,50587,50656,50790,51013,51056,51067,51223,51307,51504,51718,51742,52015,52087,52208,52242,52292,52294,52403,52474,52501,52610,52858,52899,52991,53048,53215,53237,53245,53314,53352,53368,53451,53520,53589,53605,53619,53647,53809,53863,53864,53966,54040,54077,54082,54189,54192,54223,54295,54373,54410,54445,54485,54784,54912,54940,55142,55307,55310,55346,55388,55409,55412,55423,55449,55929,55938,55966,56110,56272,56359,56413,56473,56578,56688,56872,56881,56888,56913,56927,56989,57445,57464,57497,57500,57803,57860,57890,57949,58038,58063,58099,58129,58205,58235,58248,58266,58277,58282,58329,58389,58451,58622,58632,58670,58721,58920,58992,59094,59099,59183,59195,59266,59300,59440,59502,59685,59759,59818,59917,60032,60055,60198,60225,60307,60324,60349,60541,60652,60689,60734,60905,60956,60965,61271,61371,61488,61527,61537,61553,61663,61705,61756,61767,62016,62410,62425,62661,62688,62711,62850,62889,62925,63183,63287,63357,63659,63783,63828,63862,63950,63953,64060,64100,64107,64136,64236,64239,64290,64299,64361,64450,64467,64532,64586,64641,64671,64746,64774,64830,64928,64998,65093,65108,65159,65364,65422,65446,65547,65733,65775,65997,66266,66355,66374,66581,66701,66801,66815,66890,66948,67067,67330,67359,67388,67423,67643,67665,67708,67709,67884,67910,67938,67977,67992,68052,68236,68250,68272,68435,68554,68620,68624,68751,68756,68827,68849,68853,68938,68961,69112,69124,69223,69236,69342,69390,69392,69489,69666,69711,69837,69873,69919,69986,70059,70135,70259,70280,70358,70368,70377,70399,70538,70620,70636,70650,70660,70754,70869,70888,70950,70969,71055,71092,71282,71352,71599,71603,71707,71720,71741,71766,71780,71918,71967,72056,72165,72182,72382,72421,72458,72475,72527,72547,72631,72692,72787,72821,72826,72929,72963,73168,73532,73550,73616,73668,73830,73884,73936,74168,74325,74587,74601,74611,74667,74668,74689,74727,74753,74771,74781,74847,74888,75133,75422,75428,75549,75615,75632,75663,75678,75683,75778,75829,75851,75897,75908,75928,76038,76262,76359,76425,76494,76496,76523,76614,76669,76687,76905,76952,76974,77062,77193,77360,77446,77538,77561,77584,77585,77841,77850,77918,77923,77936,78254,78454,78459,78598,78663,78911,79067,79092,79105,79155,79351,79553,79692,79695,79771,79778,79805,79857,79875,79880,80090,80099,80151,80206,80474,80607,80782,80786,80822,81235,81241,81260,81292,81304,81311,81336,81608,81609,81615,81814,81821,81942,82019,82062,82124,82131,82166,82203,82244,82260,82285,82375,82434,82520,82617,82746,82756,82763,82847,82945,82952,82957,82993,83076,83089,83213,83216,83225,83249,83308,83314,83379,83440,83451,83539,83540,83691,83858,84003,84010,84063,84176,84242,84436,84502,84747,84749,84760,84838,84857,84993,85015,85089,85193,85364,85465,85578,85957,86049,86229,86290,86425,86549,86617,86625,86704,86726,86764,86819,86983,86984,87201,87330,87362,87368,87496,87540,87599,87609,87704,87909,88056,88074,88147,88394,88437,88446,88570,88580,88643,88657,88688,88721,88832,88913,88937,88940,88948,89164,89302,89476,89515,89522,89602,89755,89761,89877,89939,89962,89966,89977,90028,90029,90052,90060,90194,90366,90393,90398,90429,90442,90524,90596,90655,90680,90747,90838,90934,90973,90995,91155,91293,91377,91513,91540,91548,91607,91727,91820,91838,91931,92005,92285,92370,92401,92443,92505,92589,92625,92768,92773,92780,92811,92901,92945,92987,93123,93152,93212,93215,93267,93393,93556,93617,93645,93666,93741,93932,94052,94067,94075,94150,94191,94257,94354,94447,94517,94562,94642,94663,94727,94743,94785,95016,95149,95349,95724,95928,95930,96134,96191,96298,96452,96476,96520,96601,96657,96699,96778,96873,96962,97010,97064,97159,97189,97221,97277,97327,97465,97498,97520,97707,97726,97916,98030,98055,98075,98171,98202,98240,98255,98400,98639,98674,98697,98735,98999,99068,99113,99209,99219,99233,99247,99341,99361,99507,99517,99579,99634,99649,99650,99793,99878",
      "35,177,304,312,391,439,448,483,577,601,633,672,674,800,803,835,851,853,871,877,879,884,885,909,913,991,997,1003,1009,1012,1019,1025,1094,1134,1190,1231,1238,1240,1287,1294,1319,1320,1373,1382,1406,1412,1532,1851,1876,1957,2161,2219,2227,2231,2414,2447,2526,2745,2828,2887,2905,2941,3114,3155,3196,4800,5091,5097,5117,5212,5532,5570,5617,5618,5637,5640,5682,5740,5783,5850,5966,5991,5992,6029,6098,6100,6168,6193,6312,6379,6482,6687,6800,7038,7041,7152,7314,7330,7372,7521,7556,7620,7629,7639,7641,7827,8092,8171,8178,8238,8299,8326,8385,8485,8497,8570,8651,8833,9090,9098,9120,9206,9445,9891,9990,10055,10222,10250,10265,10329,10340,10416,10466,10496,10551,10564,10761,10778,10785,10852,10964,11096,11278,11778,11922,12081,12101,12320,13029,13839,13840,13897,14378,14585,14610,14733,15069,15070,15085,15131,15316,15431,15522,15999,16107,16471,16519,16664,16732,16744,16773,16776,17582,18188,18522,18683,18884,18889,19480,19627,20122,20248,20496,21039,21060,21125,21298,22179,23043,23773,24277,24679,24789,24826,24973,25124,25632,25871,25972,26112,26272,26347,26680,26694,27082,27156,27324,27572,28272,28749,28787,29191,29542,31732,33757,35706,35953,36263,36408,36676,36868,37616,37640,37951,38050,38197,38323,38511,38938,38941,39003,39191,39295,39359,39445,39691,39749,39800,39823,42275,42997,45120,45792,45808,45918,47364,50102,50134,50135,50435,50446,50580,50585,50619,50654,50676,50698,50832,50944,51031,51063,51207,51227,51327,51366,51411,51548,51593,51730,51807,51832,51884,51901,51954,51961,51996,51997,52001,52378,52415,52416,52443,52614,52624,52654,52699,52757,52816,52820,52832,52892,52921,53093,53173,53227,53635,53638,53708,53787,53807,53842,53907,53910,53971,54047,54152,54213,54277,54286,54328,54343,54347,54348,54399,54440,54526,54735,54742,54796,54826,54926,54965,55094,55190,55292,55297,55448,55609,55673,55710,55722,55727,55750,55904,55996,56087,56156,56342,56447,56598,56748,56752,56857,56999,57011,57063,57160,57311,57325,57419,57446,57447,57543,57615,57663,57784,57847,57866,57894,57932,57941,58144,58198,58228,58293,58367,58415,58540,58546,58684,58738,58762,58773,58814,58837,58994,59123,59224,59231,59315,59335,59410,59412,59425,59495,59611,59661,59709,59844,60504,60712,60816,60941,61084,61219,61293,61340,61416,61466,61591,61595,61782,61863,61866,61878,61965,61986,62015,62021,62059,62185,62195,62413,62454,62499,62530,62665,62791,62875,62912,62934,62952,63095,63169,63224,63246,63304,63437,63454,63504,64130,64182,64292,64351,64436,64453,64484,64489,64603,64688,64689,64716,64761,64825,64867,65146,65183,65204,65240,65275,65289,65390,65440,65459,65548,65587,65618,65620,65669,65679,65715,65720,65946,66205,66210,66236,66275,66371,66376,66418,66486,66704,66982,67033,67082,67135,67164,67177,67270,67306,67338,67351,67513,67682,67725,67812,67829,67932,67970,68043,68047,68051,68054,68208,68362,68383,68425,68529,68560,68660,68828,68858,68955,68991,69122,69179,69184,69237,69327,69360,69491,69526,69653,69677,69728,69869,69871,69890,69908,69993,70235,70316,70346,70404,70485,70623,70828,70899,71017,71054,71082,71111,71596,71813,71837,71972,72231,72310,72424,72452,72468,72470,72519,72533,72596,72669,72672,72684,72791,72814,72971,73066,73114,73411,73505,73520,73524,73693,73754,73838,73850,73892,73907,73949,74000,74028,74044,74208,74326,74612,74694,74733,75123,75145,75226,75379,75415,75520,75573,75602,75715,75764,75823,75837,75975,76128,76204,76207,76335,76387,76395,76477,76529,76792,76844,76909,76926,76995,77158,77164,77241,77336,77337,77447,77624,77657,77812,78123,78129,78165,78216,78260,78270,78415,78466,78511,78594,78610,78666,78725,78764,78871,78959,79086,79114,79165,79182,79184,79374,79471,79567,79598,79633,79655,79665,79762,79780,79811,79816,79879,79907,79918,79928,79943,79952,79986,80071,80201,80239,80268,80310,80362,80395,80439,80450,80615,80640,80688,80818,80871,80955,81081,81142,81148,81202,81207,81271,81368,81552,81720,81833,81868,81885,81953,82049,82054,82208,82290,82304,82360,82472,82501,82717,82856,82914,83049,83113,83144,83357,83411,83415,83439,83501,83594,83602,83640,83696,83839,83918,83934,84054,84097,84150,84166,84255,84408,84432,84481,84499,84516,84525,84607,84622,84804,84914,84991,84996,85034,85295,85584,85633,85724,85762,85874,85923,85931,85993,86222,86240,86327,86675,86880,87014,87122,87129,87218,87348,87630,87924,88104,88191,88217,88234,88327,88349,88507,88525,88528,88746,88810,88823,88841,88988,89058,89129,89282,89317,89337,89577,89633,89876,89879,89969,90108,90197,90198,90278,90358,90483,90502,90599,90612,90722,90753,90759,90810,90954,91047,91071,91271,91340,91466,91503,91584,91638,91665,91795,91834,91894,91897,91943,91972,92133,92150,92221,92388,92405,92456,92476,92660,92696,92725,92751,92759,92779,92801,92816,92842,92859,92920,92982,93079,93349,93395,93457,93461,93620,93622,93735,93928,93947,94048,94088,94159,94206,94264,94306,94361,94375,94430,94539,94565,94690,94777,94813,94876,94945,95088,95135,95204,95211,95273,95331,95357,95425,95446,95552,95572,95694,95716,95767,95868,95870,96260,96433,96541,96589,96709,96712,96725,96752,96912,96920,97378,97408,97540,97593,97664,97689,97870,97872,98081,98109,98145,98181,98224,98317,98489,98504,98813,98925,99373,99421,99553,99708,99721,99800,99806,99852,99868",
      "78,148,228,229,1649,1731,1736,1769,2035,2094,2155,2156,2230,2415,2429,2628,2679,2727,2746,2829,2852,2870,2983,3193,3214,3425,3479,3517,3574,3793,4314,4328,4342,4517,4797,4851,4858,4986,5080,6052,7524,9479,9488,9597,9711,9749,9806,9920,9977,10262,10604,10642,10680,10756,10767,10832,10889,11003,11022,11117,11193,11212,11231,11242,11250,11264,11342,11364,11459,11498,11516,11585,11592,11649,11801,11858,11915,11941,12048,12086,12124,12143,12162,12219,12238,12257,12276,12295,12313,12314,12341,12371,12390,12428,12485,12514,12523,12580,12599,12694,12713,12741,12827,12881,12960,12998,13017,13036,13043,13045,13074,13093,13131,13169,13206,13207,13239,13245,13264,13278,13302,13321,13393,13394,13397,13416,13428,13435,13518,13530,13568,13587,13644,13663,13689,13739,13758,13777,13796,13830,13910,14005,14024,14078,14081,14092,14119,14157,14176,14195,14233,14252,14271,14290,14309,14328,14343,14366,14385,14404,14441,14499,14556,14575,14661,14669,14670,14727,14784,14841,14879,14917,14943,14955,15012,15031,15050,15088,15107,15126,15164,15183,15219,15221,15297,15335,15354,15373,15378,15411,15430,15468,15506,15525,15582,15620,15639,15658,15696,15711,15734,15772,15810,15829,15848,15867,15886,15943,15962,16000,16019,16037,16076,16133,16152,16171,16209,16228,16247,16266,16285,16304,16323,16342,16361,16380,16418,16456,16475,16513,16532,16589,16601,16608,16627,16665,16703,16741,16760,16763,16779,16830,16855,16874,16893,16950,16969,16988,17039,17041,17045,17093,17140,17197,17216,17290,17292,17330,17349,17387,17406,17425,17444,17463,17501,17520,17596,17615,17691,17710,17729,17748,17805,17824,17843,17862,17881,17900,17919,17926,17957,17995,18033,18071,18128,18185,18195,18204,18242,18260,18261,18299,18318,18356,18394,18413,18432,18489,18508,18527,18603,18679,18736,18751,18755,18793,18803,18812,18850,18869,18907,18945,18964,18983,19002,19040,19097,19116,19135,19204,19249,19306,19325,19344,19382,19420,19458,19496,19534,19567,19572,19610,19648,19667,19686,19705,19762,19781,19800,19819,19876,19895,19914,19933,19952,19990,20047,20085,20104,20123,20142,20161,20171,20180,20237,20275,20294,20313,20332,20351,20370,20389,20427,20446,20465,20484,20499,20529,20541,20548,20560,20572,20598,20616,20617,20636,20655,20660,20712,20731,20750,20786,20788,20864,20902,20921,20978,21016,21054,21092,21111,21168,21187,21206,21221,21263,21264,21282,21339,21358,21395,21415,21434,21453,21491,21510,21516,21529,21567,21586,21662,21700,21738,21757,21795,21814,21852,21871,21909,21928,21947,21966,22004,22023,22042,22050,22080,22099,22110,22194,22209,22213,22270,22289,22327,22365,22375,22384,22403,22441,22460,22479,22498,22536,22555,22574,22631,22650,22688,22726,22745,22802,22859,22878,22897,22916,22969,22973,23030,23049,23068,23125,23144,23163,23199,23240,23258,23277,23315,23334,23349,23353,23372,23410,23423,23429,23505,23541,23543,23581,23600,23610,23648,23657,23676,23733,23752,23771,23809,23828,23904,23961,23999,24018,24037,24056,24094,24107,24113,24151,24170,24208,24227,24237,24246,24303,24315,24328,24331,24341,24360,24398,24436,24455,24474,24493,24512,24531,24550,24569,24588,24607,24626,24645,24664,24683,24721,24757,24759,24774,24778,24835,24854,24873,24949,24968,24987,25000,25061,25063,25082,25120,25158,25196,25221,25253,25272,25310,25348,25367,25386,25405,25462,25481,25538,25557,25595,25614,25652,25688,25709,25747,25785,25823,25842,25880,25893,25937,25956,25975,25994,26032,26051,26070,26089,26108,26127,26146,26203,26222,26260,26279,26298,26355,26374,26387,26412,26431,26450,26469,26526,26545,26554,26621,26678,26697,26735,26754,26773,26811,26925,26926,27020,27068,27134,27198,27210,27248,27267,27362,27387,27431,27552,27652,27704,27780,27799,27806,27837,27894,27977,28017,28103,28122,28181,28198,28312,28331,28616,28635,28871,29117,29241,29471,29498,29499,29604,29620,29630,29697,29711,29724,29765,29794,29908,30002,30031,30060,30089,30118,30147,30176,30205,30234,30263,30292,30321,30350,30379,30408,30437,30466,30495,30524,30553,30582,30611,30640,30669,30698,30727,30756,30785,30814,30843,30872,30901,30930,30959,30988,31017,31046,31075,31104,31133,31162,31191,31220,31249,31278,31307,31336,31365,31394,31423,31452,31481,31510,31539,31568,31597,31626,31655,31684,31713,31742,31771,31800,31829,31858,31887,31916,31945,31974,32003,32032,32061,32090,32119,32148,32177,32206,32235,32264,32293,32322,32351,32380,32409,32438,32467,32496,32525,32554,32583,32612,32641,32670,32699,32728,32757,32786,32815,32844,32873,32902,32931,32960,32989,33018,33047,33076,33105,33134,33163,33192,33221,33250,33279,33308,33337,33366,33395,33424,33453,33482,33511,33540,33569,33598,33627,33656,33685,33714,33743,33772,33801,33830,33859,33888,33917,33946,33975,34004,34033,34062,34091,34120,34149,34178,34207,34236,34265,34294,34323,34352,34381,34410,34439,34468,34497,34526,34555,34584,34613,34642,34671,34700,34729,34758,34787,34816,34845,34874,34903,34932,34961,34990,35002,35060,35063,35091,35119,35143,35160,35165,35232,35254,35310,35328,35382,35402,35445,35452,35458,35473,35521,35580,35625,35639,35680,35688,35712,35771,35824,35830,35876,35901,35960,36019,36048,36082,36127,36142,36201,36239,36272,36302,36331,36374,36390,36462,36521,36580,36633,36644,36703,36762,36791,36792,36833,36892,36951,37023,37082,37141,37205,37264,37323,37353,37394,37453,37481,37486,37512,37546,37584,37643,37702,37746,37766,37825,37884,37955,37993,38014,38056,38073,38145,38173,38179,38204,38263,38299,38327,38386,38445,38516,38575,38624,38634,38706,38765,38793,38824,38886,38888,38947,39006,39053,39077,39107,39136,39174,39195,39235,39242,39267,39326,39385,39449,39508,39567,39596,39602,39638,39668,39695,39697,39756,39828,39887,39926,39946,40117,40146,40175,40204,40233,40262,40291,40320,40349,40378,40407,40436,40465,40494,40523,40552,40581,40610,40639,40668,40697,40726,40755,40784,40813,40842,40871,40900,40929,40958,40987,41016,41045,41074,41103,41132,41161,41190,41219,41248,41277,41306,41335,41364,41393,41422,41451,41480,41509,41538,42008,42037,42066,42095,42124,42163,42184,42274,42307,42349,42365,42372,42424,42440,42449,42511,42586,42629,42661,42698,42714,42717,42729,42748,42779,42823,42898,42913,42985,43060,43135,43222,43297,43372,43390,43443,43459,43486,43504,43534,43568,43609,43636,43666,43696,43722,43771,43787,43829,43846,43859,43880,43914,43933,43967,43971,43974,43993,44008,44064,44083,44107,44109,44110,44170,44228,44245,44277,44320,44335,44372,44407,44482,44557,44615,44644,44681,44719,44753,44760,44794,44865,44881,44922,44930,44956,44977,45031,45082,45085,45104,45152,45213,45271,45293,45308,45344,45401,45459,45531,45553,45588,45618,45646,45718,45744,45775,45816,45833,45905,45962,46037,46096,46165,46168,46190,46194,46200,46211,46226,46280,46282,46286,46291,46295,46309,46424,46568,46618,46641,46644,46671,46672,46674,46719,46753,46835,46859,46883,46895,46924,46929,46949,47019,47101,47108,47114,47141,47164,47211,47233,47240,47241,47251,47273,47275,47300,47355,47360,47381,47399,47453,47474,47512,47519,47534,47544,47550,47683,47730,47766,47772,47780,47838,47859,47868,47871,47881,47916,47951,48009,48067,48092,48109,48186,48218,48223,48319,48334,48342,48392,48446,48469,48478,48479,48503,48518,48547,48554,48572,48581,48604,48612,48655,48691,48711,48725,48733,48843,48847,48855,48866,48902,48936,48998,49021,49077,49094,49100,49157,49200,49211,49251,49261,49287,49354,49372,49394,49429,49471,49541,49554,49571,49606,49615,49617,49622,49651,49680,49700,49739,49762,49783,49857,49863,49865,49880,49971,49982,49983,49989,50010,50043,50312,50405,50485,50517,50591,50811,50979,51048,51139,51183,51301,51413,51555,51568,51600,51683,51776,51812,51906,51928,52115,52340,52410,52488,52525,52577,52643,52649,52749,52891,52938,53022,53231,53340,53483,53513,53603,53712,53741,53891,53902,53952,53974,54129,54133,54201,54402,54437,54441,54458,54500,54547,54548,54621,54667,54687,54825,55052,55181,55191,55278,55306,55700,55851,55856,56109,56301,56322,56554,56566,56590,56616,56634,56656,56681,56780,56852,56988,57069,57576,57598,57600,57693,57939,57948,58015,58029,58069,58082,58193,58243,58262,58371,58378,58388,58427,58493,58598,58602,58690,58706,58723,58832,58857,58987,58991,59001,59176,59471,59478,59568,59587,59637,59644,59842,59908,59922,59942,59983,60015,60092,60120,60133,60157,60178,60363,60412,60471,60608,60679,60780,60784,60907,60977,61046,61059,61061,61168,61186,61202,61344,61434,61507,61635,61685,61798,61817,61824,61825,61828,61845,61851,61899,62177,62306,62312,62423,62453,62460,62466,62599,62600,62618,62640,62826,62979,62994,63011,63049,63144,63227,63302,63330,63373,63419,63514,63671,63782,63873,63989,64122,64275,64495,64793,64936,65053,65160,65176,65276,65299,65331,65436,65447,65514,65717,65872,66021,66450,66522,66694,66974,67039,67108,67118,67194,67272,67285,67324,67434,67461,67675,67875,67886,67933,67995,67999,68038,68207,68246,68310,68414,68465,68505,68574,68577,68582,68698,68703,69078,69096,69310,69321,69362,69405,69456,69595,69608,69696,69708,69754,69793,69891,70083,70090,70098,70199,70246,70252,70261,70424,70457,70489,70549,70605,70653,70682,70684,70686,70833,70880,70940,71040,71127,71160,71456,71485,71487,71646,71702,71727,71747,71762,71875,71888,71950,71958,71962,72173,72330,72361,72523,72602,72606,72690,72698,72891,72990,73220,73224,73287,73319,73573,73588,73770,73787,73798,73873,73881,73971,74103,74104,74285,74302,74383,74523,74541,74627,74677,74778,74834,74867,74937,74957,75063,75128,75132,75208,75269,75455,75506,75835,75957,75963,76162,76199,76393,76528,76573,76697,76730,76774,76889,77020,77081,77129,77269,77348,77376,77421,77471,77592,77611,77681,77820,77844,77906,77907,78118,78119,78130,78304,78406,78440,78510,78578,78599,78776,78993,79039,79222,79327,79346,79436,79508,79530,79562,79589,79612,79631,79641,79699,79773,79978,80058,80314,80322,80403,80476,80484,80638,80687,80691,80696,80746,80879,80905,81040,81085,81193,81203,81269,81339,81384,81717,81725,81835,82035,82184,82200,82272,82329,82394,82564,82644,82680,82858,82859,82935,82961,83034,83135,83136,83139,83159,83193,83315,83412,83656,84091,84098,84147,84162,84211,84407,84524,84639,84659,84772,84787,84824,85106,85202,85233,85320,85375,85427,85545,85562,85563,85755,85915,86065,86180,86194,86201,86207,86330,86370,86375,86450,86475,86568,86602,86670,86799,86809,86835,86839,86876,87023,87092,87168,87248,87264,87385,87537,87647,87700,87728,87730,87758,87779,87850,87868,87901,87954,87968,87978,87994,88003,88038,88093,88129,88221,88447,88486,88594,88611,88770,88796,88814,88817,88852,89045,89067,89311,89432,89608,89609,89706,89824,89917,90062,90074,90088,90242,90343,90437,90957,90999,91064,91158,91268,91419,91430,91509,91592,91735,91934,91964,92064,92068,92071,92147,92363,92754,92845,92996,93076,93189,93199,93284,93388,93491,93505,93523,93539,93572,93785,93814,94028,94095,94251,94308,94489,94519,94524,94556,94622,94674,94703,94794,94814,94841,94907,94957,94971,95054,95237,95363,95436,95488,95503,95587,95665,95684,95688,95904,95909,96002,96018,96108,96215,96316,96485,96646,96864,96906,96932,96978,97060,97204,97287,97428,97519,97731,97869,97900,97962,98058,98156,98218,98233,98254,98420,98439,98687,98709,98812,98954,98975,99091,99125,99174,99269,99295,99410,99419,99494,99699,99744,99853,99896,99908,99951",
      "2,44,119,261,299,368,370,436,447,456,502,741,745,789,837,908,982,1033,1053,1131,1177,1304,1583,1673,1750,1762,1819,1941,2004,2588,2749,2869,3021,3460,3930,4320,5172,5351,5428,5502,5552,5575,5603,5697,6133,6241,6271,6380,6446,6449,6559,6571,6585,6816,6830,6874,6889,6979,7027,7035,7180,7191,7266,7366,7470,7503,7509,7519,7540,7549,7594,7621,7851,7862,7882,7908,7917,8000,8018,8055,8172,8195,8226,8279,8282,8323,8434,8512,8515,8517,8528,8597,8612,8691,8744,8751,8772,8781,8793,8858,8876,8881,8895,8902,8929,9005,9028,9132,9249,9358,9443,9561,9747,9835,9861,9896,10053,10112,10221,10312,10314,10366,10393,10435,10539,10744,11060,11079,11166,11460,11527,11653,11731,11831,11944,11982,12056,12355,12648,12792,12841,12862,13042,13228,13265,13559,13834,13940,14396,14506,14598,14724,14954,14981,15133,15186,15650,15904,16190,16573,16594,16873,16922,17229,17412,17570,17916,17942,18280,18588,19166,19167,19321,19656,19696,20160,20713,20763,20774,20899,21368,21685,21823,21886,22025,22665,23428,23801,24089,24153,25378,25689,25782,26275,26403,26454,26932,26939,26991,27096,27501,27636,27660,27892,28178,28337,28743,28890,28902,29087,29136,29263,29281,29305,29555,29559,29573,29718,29885,29926,29964,29981,30302,34576,35754,35892,35937,36121,36157,36862,37064,37078,37123,37133,37365,37430,37597,37749,37801,37861,37927,38045,38121,38486,38507,38567,38804,38862,38918,39108,39261,39467,39505,39802,39922,39942,41500,43102,45168,45610,45684,45733,45859,47336,47384,48152,50185,50274,50281,50403,50635,50711,50881,51295,51406,51490,51509,51577,51657,51682,51804,52045,52173,52201,52216,52263,52330,52362,52457,52489,52511,52568,52724,52740,52786,52998,53020,53069,53113,53221,53240,53332,53421,53481,53488,53495,53569,53590,53820,53908,54014,54063,54109,54117,54278,54598,54607,54767,54769,54856,54938,54945,54999,55006,55009,55090,55288,55312,55325,55426,55674,55684,55767,55902,55960,56162,56186,56208,56213,56341,56343,56435,56539,56619,56728,56822,56867,57210,57236,57246,57346,57370,57377,57386,57415,57476,57523,57689,57944,58041,58173,58230,58261,58332,58377,58432,58555,58726,58750,58752,58815,58868,58913,59126,59276,59305,59427,59436,59459,59467,59760,60088,60282,60375,60419,60482,60577,60708,60740,60746,60805,60881,60909,60985,61021,61027,61035,61045,61204,61549,61732,61901,62099,62194,62201,62229,62273,62349,62507,62510,62610,62615,62881,62986,62987,63200,63488,63502,63518,63532,63681,63954,64031,64094,64200,64223,64251,64319,64336,64451,64466,64634,64784,64799,64845,64901,65011,65024,65062,65137,65141,65150,65288,65408,65409,65522,65551,65723,65898,65905,65918,65962,66059,66079,66257,66273,66314,66345,66449,66568,66675,66799,66902,67028,67041,67134,67243,67404,67512,67666,67689,67949,67980,68069,68083,68151,68179,68320,68416,68441,68449,68512,68592,68638,68726,68798,69011,69018,69036,69046,69047,69116,69228,69254,69323,69325,69377,69452,69640,69692,70047,70077,70121,70157,70216,70438,70462,70507,70587,70643,70796,70834,70854,70896,71128,71197,71229,71281,71386,71466,71584,71776,71786,72000,72013,72106,72110,72118,72237,72278,72451,72498,72768,72828,73053,73055,73248,73264,73397,73696,73712,73901,73948,73961,74070,74071,74128,74133,74175,74317,74450,74527,74666,74704,74711,74788,74852,74932,75018,75046,75220,75223,75354,75443,75492,75560,75591,75745,75814,75842,75865,75881,75902,75936,76401,76491,76550,76570,76595,76619,76817,76824,76854,76891,77009,77060,77128,77167,77180,77186,77206,77245,77322,77327,77362,77407,77451,77457,77720,77784,77788,78019,78067,78089,78182,78429,78606,78625,78659,78726,78869,78878,79083,79186,79281,79418,79420,79465,79482,79549,79595,79607,79814,79854,79863,79977,79998,80078,80149,80279,80363,80409,80429,80488,80554,80597,80664,80763,80768,80802,80894,81183,81222,81239,81270,81288,81334,81433,81484,81526,81542,81711,81714,81810,81817,81891,81983,81997,82078,82307,82318,82327,82340,82439,82486,82607,82780,83024,83304,83443,83533,83589,83638,83717,83771,83816,83836,83863,83982,84083,84434,84517,84520,84558,84670,84671,84672,84702,84783,85026,85028,85062,85123,85184,85287,85525,85624,85654,85719,85741,85770,85836,85906,85989,86030,86168,86287,86312,86365,86383,86398,86418,86669,86724,86788,86807,86848,86860,86934,86947,87028,87174,87190,87212,87273,87491,87542,87620,87664,87784,87884,88031,88086,88090,88161,88182,88195,88216,88223,88241,88410,88548,88574,88624,88645,88875,88900,88920,88956,89026,89096,89184,89328,89347,89366,89399,89473,89485,89560,89644,89811,89846,89890,90010,90162,90180,90211,90219,90301,90456,90503,90506,90597,90606,90630,90650,90735,90738,90949,91010,91072,91177,91249,91381,91512,91516,91534,91754,91800,91822,91826,92043,92171,92214,92230,92387,92400,92501,92711,92863,92914,92965,93089,93117,93137,93208,93254,93315,93366,93501,93516,93564,93743,93872,93878,93942,93980,94102,94181,94414,94459,94496,94732,94780,94793,94798,94803,94823,94908,94996,95002,95027,95187,95241,95266,95381,95403,95408,95848,95936,95968,96028,96170,96173,96254,96353,96419,96451,96643,96763,96840,97008,97168,97370,97372,97380,97417,97651,97678,97685,97791,97874,97961,97997,98045,98096,98144,98162,98309,98328,98794,99103,99117,99188,99228,99301,99310,99340,99433,99483,99534,99565,99602,99618,99660,99664,99704,99731,99790,99956,99973",
      "54,108,109,140,155,163,179,189,190,238,240,759,1788,1929,1939,1966,1967,1969,1990,2037,2109,2123,2183,2207,2212,2232,2258,2264,2284,2297,2301,2353,2360,2369,2382,2389,2410,2413,2422,2456,2461,2479,2482,2492,2496,2501,2515,2524,2535,2538,2543,2544,2565,2567,2574,2581,2586,2611,2619,2653,2697,2710,2716,2717,2723,2725,2761,2769,2771,2775,2795,2810,2814,2843,2855,2863,2889,2897,2903,2912,2919,2924,2933,2987,3070,3081,3085,3089,3099,3106,3109,3141,3153,3213,3222,3256,3304,3315,3317,3319,3344,3347,3349,3362,3373,3378,3381,3382,3409,3432,3476,3484,3515,3518,3519,3536,3543,3544,3550,3552,3554,3564,3571,3572,3593,3595,3605,3622,3642,3668,3672,3682,3693,3698,3731,3742,3750,3755,3798,3802,3810,3830,3841,3849,3870,3882,3897,3899,3911,3915,3925,3973,3974,3988,3989,4015,4026,4056,4071,4077,4097,4104,4121,4137,4141,4167,4200,4208,4220,4227,4252,4260,4263,4277,4280,4324,4344,4346,4363,4384,4415,4422,4425,4428,4436,4444,4472,4473,4475,4477,4505,4508,4516,4526,4533,4566,4594,4602,4613,4628,4631,4651,4652,4666,4680,4681,4704,4707,4709,4761,4768,4789,4791,4795,4808,4813,4816,4831,4843,4888,4897,4900,4906,4907,4927,4931,4966,4987,4990,5054,5066,5572,6111,6260,6292,6307,6337,6371,6584,6622,6632,6812,6866,6910,6914,6926,6974,6988,7057,7245,7263,7337,7381,7437,7458,7587,7637,7675,7736,7807,7819,7857,7875,7884,7922,7948,7957,8014,8026,8104,8116,8162,8217,8290,8309,8345,8504,8535,8552,8606,8609,8620,8652,8787,8854,8878,8911,9022,9159,9208,9260,9290,9328,9332,9333,9407,9461,9516,9573,9735,9745,9748,9939,10194,10196,10248,10297,10351,10407,10524,10573,10606,10643,10791,10875,10899,10982,11211,11266,11283,11623,11699,11904,11946,12002,12011,12169,12272,12363,12375,12387,12410,12483,12490,12574,12816,12820,12884,12983,13046,13108,13182,13353,13800,13810,13960,14028,14090,14518,14547,14645,14745,14933,15013,15259,15319,15370,15449,15520,15681,15757,15796,16135,16184,16271,16328,16540,16618,16621,16658,16797,16820,17033,17233,17243,17428,17476,17488,17631,17948,17993,18031,18095,18096,18177,18192,18451,18477,18590,18611,18625,18962,18986,19269,19352,19411,19828,19941,19976,20112,20513,20778,20846,20939,20957,20986,21022,21182,21185,21203,21213,21341,21355,21541,21551,21569,21601,21605,21827,21944,22451,22499,22709,22892,23072,23519,23554,23837,23858,23930,24064,24408,24745,24776,24824,24872,24882,24972,25251,25334,25452,25474,25530,25567,26028,26049,26099,26110,26126,26187,26189,26202,26220,26305,26442,26455,26528,26531,26546,26561,26566,26581,26595,26598,26600,26669,26841,26899,27001,27007,27014,27064,27080,27202,27296,27303,27367,27409,27459,27484,27487,27500,27514,27516,27558,27616,27623,27643,27655,27657,27751,27776,27815,28004,28033,28040,28056,28100,28128,28140,28152,28177,28201,28237,28239,28395,28484,28512,28542,28594,28612,28623,28630,28659,28666,28753,28885,28886,28920,28928,29008,29014,29035,29056,29071,29076,29164,29402,29409,29441,29444,29531,29581,29589,29636,29680,29688,29801,29815,29839,29855,29898,29924,29990,30019,30048,30077,30106,30135,30164,30193,30222,30251,30280,30338,30367,30396,30425,30454,30483,30512,30541,30570,30579,30628,30657,30686,30715,30744,30773,30802,30831,30860,30889,30918,30947,31034,31063,31092,31121,31150,31179,31208,31237,31266,31295,31324,31382,31411,31440,31469,31498,31527,31556,31585,31614,31643,31672,31701,31730,31759,31788,31817,31846,31875,31904,31933,31962,31991,32020,32049,32078,32107,32136,32165,32194,32223,32252,32281,32310,32339,32368,32397,32426,32455,32484,32513,32542,32571,32600,32629,32658,32687,32716,32745,32774,32803,32832,32861,32890,32919,32948,32977,33006,33035,33064,33093,33122,33151,33180,33209,33238,33267,33296,33325,33383,33412,33441,33470,33528,33557,33586,33615,33644,33673,33702,33731,33760,33789,33818,33847,33876,33905,33934,33963,33992,34079,34108,34137,34166,34195,34224,34253,34282,34311,34340,34369,34398,34427,34456,34485,34514,34543,34572,34601,34630,34659,34688,34717,34746,34775,34804,34833,34862,34891,34920,34949,34978,35019,35037,35077,35086,35108,35125,35182,35200,35271,35289,35293,35308,35318,35345,35370,35377,35419,35470,35490,35538,35567,35597,35656,35691,35729,35756,35847,35862,35874,35918,35977,35990,36036,36067,36099,36159,36218,36232,36289,36306,36348,36407,36460,36479,36538,36597,36661,36720,36779,36795,36796,36807,36850,36909,36924,36968,37040,37099,37158,37174,37222,37233,37281,37340,37470,37529,37545,37601,37629,37660,37719,37734,37783,37809,37842,37865,37901,37934,37972,38031,38044,38090,38104,38162,38221,38280,38305,38307,38344,38365,38381,38403,38462,38484,38533,38547,38573,38592,38613,38666,38678,38723,38782,38841,38856,38869,38905,38919,38923,38964,38995,39023,39048,39094,39109,39123,39153,39212,39224,39230,39232,39284,39296,39343,39402,39466,39491,39525,39539,39584,39598,39603,39655,39675,39714,39773,39845,39904,39915,39963,40109,40138,40167,40196,40225,40254,40283,40312,40341,40370,40399,40428,40457,40486,40544,40573,40602,40631,40660,40689,40718,40747,40776,40805,40834,40863,40892,40921,40950,40979,41008,41037,41066,41095,41124,41153,41182,41211,41240,41269,41298,41327,41356,41385,41414,41443,41472,41501,41530,41550,41551,41552,41553,41554,41555,41556,41557,41558,41559,41560,41561,41562,41563,41564,41565,41566,41567,41568,41569,41570,41571,41572,41573,41574,41576,41577,41578,41579,41580,41581,41582,41583,41584,41585,41586,41587,41588,41589,41590,41591,41592,41593,41594,41595,41596,41597,41598,41599,41600,41601,41602,41603,41604,41605,41607,41608,41609,41610,41611,41612,41613,41614,41615,41616,41617,41618,41619,41620,41621,41622,41623,41624,41625,41626,41627,41628,41629,41630,41631,41632,41633,41634,41635,41636,41637,41638,41639,41640,41641,41642,41643,41644,41645,41646,41647,41648,41649,41650,41651,41652,41653,41654,41655,41656,41657,41658,41659,41660,41661,41662,41663,41664,41665,41666,41668,41669,41670,41671,41672,41673,41674,41675,41676,41677,41678,41679,41680,41681,41682,41683,41684,41685,41686,41687,41688,41689,41690,41691,41692,41693,41694,41695,41696,41697,41698,41699,41701,41702,41703,41704,41705,41706,41707,41708,41709,41710,41711,41712,41713,41714,41715,41716,41717,41718,41719,41720,41722,41723,41724,41725,41726,41727,41729,41730,41731,41732,41733,41734,41735,41736,41737,41738,41739,41741,41742,41743,41744,41745,41746,41747,41748,41749,41750,41751,41752,41753,41754,41755,41757,41758,41759,41760,41761,41762,41763,41764,41765,41767,41768,41769,41770,41771,41773,41774,41775,41776,41777,41778,41779,41780,41781,41782,41783,41784,41785,41786,41787,41788,41790,41792,41794,41796,41798,41800,41804,41806,41808,41810,41812,41814,41816,41818,41820,41822,41824,41826,41828,41830,41832,41834,41836,41838,41840,41842,41844,41848,41850,41852,41854,41856,41858,41860,41862,41864,41866,41868,41870,41872,41874,41876,41878,41880,41882,41884,41886,41888,41890,41892,41894,41896,41898,41900,41902,41904,41906,41908,41910,41912,41914,41916,41918,41920,41922,41924,41926,41928,41931,41934,41937,41940,41943,41949,41952,41955,41959,41963,41967,41971,41975,41979,41983,41987,41992,42021,42050,42079,42108,42137,42204,42212,42222,42302,42370,42377,42452,42471,42473,42478,42539,42544,42560,42602,42609,42614,42686,42689,42724,42731,42776,42851,42926,43013,43025,43045,43088,43163,43202,43242,43246,43250,43269,43273,43276,43325,43327,43333,43339,43400,43423,43442,43487,43513,43562,43572,43577,43585,43590,43637,43640,43665,43724,43725,43731,43747,43799,43801,43823,43874,43903,43910,43961,43976,44031,44036,44043,44111,44120,44188,44198,44271,44273,44348,44433,44434,44435,44443,44510,44531,44539,44590,44594,44672,44741,44747,44822,44836,44909,44984,45012,45049,45062,45083,45132,45179,45184,45192,45196,45251,45287,45322,45359,45439,45481,45509,45568,45571,45619,45626,45696,45755,45793,45796,45813,45883,45932,45942,46000,46002,46025,46063,46132,46178,46257,46303,46314,46343,46388,46434,46447,46451,46465,46528,46581,46586,46640,46703,46822,46882,46888,46891,46941,46946,46953,46971,47005,47015,47018,47096,47113,47184,47199,47213,47267,47283,47285,47368,47371,47400,47435,47445,47597,47607,47633,47639,47721,47726,47767,47771,47789,47796,47862,47879,48016,48054,48075,48105,48159,48184,48193,48196,48254,48294,48327,48340,48350,48374,48394,48493,48506,48520,48531,48579,48586,48600,48607,48761,48772,48788,48792,48857,48877,48888,48906,48946,48951,48983,49017,49046,49049,49058,49072,49078,49086,49088,49121,49130,49134,49144,49156,49240,49242,49271,49299,49313,49326,49365,49380,49415,49436,49440,49450,49480,49521,49538,49560,49656,49698,49716,49725,49740,49745,49796,49861,49875,49896,49921,49963,50035,50095,50096,50133,50285,50377,50419,50420,50657,50772,50886,50887,51129,51162,51210,51216,51425,51461,51485,51569,51625,51687,51720,51840,52029,52164,52184,52309,52312,52318,52396,52414,52452,52493,52665,52669,52715,52733,52767,52795,52824,52828,52867,53015,53140,53261,53358,53366,53480,53588,53608,53687,53770,53840,53868,53896,54333,54336,54386,54444,54729,54983,54986,54995,55077,55078,55099,55238,55257,55347,55382,55389,55420,55491,55565,55581,55612,55705,55859,55932,55934,55987,56107,56201,56259,56266,56270,56299,56362,56414,56483,56820,56910,56992,57341,57371,57396,57531,57715,57824,57865,57921,57925,57995,58087,58103,58134,58202,58209,58211,58284,58356,58358,58508,58591,58683,58687,58858,58869,58923,58966,59100,59150,59159,59175,59297,59359,59432,59491,59522,59556,59569,59622,59659,59667,59674,59677,59718,59944,59988,60026,60043,60115,60326,60368,60393,60455,60524,60683,60872,60894,60952,61065,61234,61237,61280,61625,61643,61753,61777,61862,61918,61998,62065,62082,62118,62199,62221,62416,62484,62551,62619,62678,62733,62780,62805,62945,62966,63069,63222,63245,63379,63538,63586,63677,63896,63930,64069,64103,64110,64243,64250,64611,64656,64659,64707,64900,64953,65020,65107,65177,65274,65393,65501,65505,65692,65761,65875,65952,66001,66028,66056,66076,66136,66225,66250,66362,66375,66388,66424,66438,66534,66623,66708,66839,66952,66993,67150,67174,67220,67233,67331,67381,67505,67531,67632,67646,67704,67844,67864,67894,68073,68149,68164,68170,68175,68251,68379,68396,68508,68545,68595,68742,68746,68848,69088,69109,69312,69448,69497,69520,69612,69656,69802,70020,70041,70065,70128,70171,70184,70203,70218,70241,70284,70286,70317,70410,70411,70554,70570,70580,70780,70923,70947,71089,71121,71174,71185,71188,71325,71370,71374,71384,71392,71518,71638,71655,71677,71796,71947,71969,72132,72155,72172,72199,72225,72236,72358,72423,72539,72554,72591,72864,72983,73099,73134,73202,73232,73360,73396,73587,73992,73993,74077,74125,74179,74365,74437,74438,74535,74648,74904,75040,75114,75323,75349,75402,75412,75475,75596,75684,75746,75761,75832,76039,76063,76192,76226,76307,76313,76319,76376,76728,76916,77103,77172,77260,77304,77391,77593,77727,77758,77947,77966,77974,78023,78059,78071,78103,78116,78171,78237,78311,78319,78326,78371,78500,78571,78837,78893,78913,78961,79068,79183,79197,79227,79235,79298,79393,79584,79825,80001,80059,80089,80195,80218,80245,80384,80391,80412,80540,80550,80619,80725,80729,80770,80864,80927,80958,81094,81107,81173,81213,81454,81654,81691,81757,81787,81801,81827,81979,82122,82207,82211,82346,82352,82456,82535,82708,82860,82934,83015,83043,83047,83116,83125,83205,83376,83416,83483,83499,83575,83674,83690,83716,83721,83775,83837,83889,83954,84045,84070,84125,84171,84185,84228,84658,84697,84782,84961,84986,85085,85162,85178,85197,85224,85271,85283,85344,85352,85536,85676,85976,86125,86151,86392,86445,86470,86622,86692,86702,86703,86722,86763,86994,87079,87087,87210,87235,87293,87298,87531,87600,87781,87858,88163,88363,88414,88426,88451,88509,88610,88891,88946,88977,89032,89133,89182,89217,89315,89380,89401,89422,89472,89593,89597,89720,89784,89906,90065,90299,90318,90418,90419,90746,90943,90977,90997,91058,91076,91090,91163,91194,91254,91288,91436,91472,91478,91530,91580,91585,91644,91655,91726,91792,91806,91891,91974,92078,92408,92457,92467,92715,92769,92888,92892,92935,92997,92999,93011,93067,93136,93191,93195,93240,93327,93359,93543,93587,93627,93661,93754,93867,93967,94039,94065,94068,94087,94121,94274,94320,94337,94369,94417,94526,94530,94548,94561,94582,94669,95004,95033,95069,95084,95207,95324,95467,95564,95671,95690,96113,96114,96209,96279,96444,96461,96508,96579,96649,96711,96717,96732,96753,96890,96928,97028,97095,97125,97153,97201,97202,97252,97265,97539,97563,97782,97819,97922,97965,97974,98002,98003,98170,98259,98295,98368,98442,98451,98628,98737,99053,99062,99343,99344,99364,99502,99524,99584,99941",
      "305,345,346,357,418,478,497,503,599,631,669,673,854,858,892,1014,1065,1084,1138,1210,1307,1456,1499,1622,1634,1638,1668,1680,1729,1742,1744,2080,2322,3146,3891,3987,4140,4624,5042,5062,5191,5284,5367,5556,5669,5733,5838,5907,6010,6068,6086,6107,6142,6235,6254,6361,6523,6549,6556,6570,6709,6738,6843,6844,7084,7100,7104,7113,7142,7227,7231,7252,7285,7304,7447,7513,7543,7659,7704,7721,7769,8069,8176,8196,8275,8291,8366,8371,8428,8525,8560,8592,8614,8757,8794,8811,8915,9010,9157,9167,9216,9232,9243,9291,9331,9352,9425,9452,9494,9511,9518,9574,9789,9790,9804,9922,9927,10146,10310,10359,10365,10513,10517,10531,10667,10898,11075,11276,11446,11492,11529,11531,11570,11590,11679,11691,11703,11752,11798,11800,11846,11989,12093,12452,12760,12808,12909,13212,13335,13651,13713,14086,14423,14485,14487,14579,14700,15124,15374,15445,15454,15492,15528,15583,15641,15663,15747,15778,15922,16446,16590,16840,17139,17236,17439,17493,17594,17758,17777,18059,18152,18262,18622,18891,18938,19336,19595,19841,20209,20738,20897,20929,21066,21081,22098,22173,22658,22815,23042,23107,23293,23769,24092,24137,24307,24370,25101,25126,25246,25665,25703,25958,26259,26271,26389,26409,26434,26482,26648,26667,26809,27067,27093,27100,27233,27245,27269,27747,27947,28727,28784,28829,28843,28936,29106,29467,29897,32093,35498,35704,35934,36133,36180,36517,36618,37061,37110,37390,37482,37495,37580,37681,37804,38184,38259,38300,38377,38697,39165,39188,39410,39619,39737,39801,39868,40443,43978,45089,45324,45507,49285,50179,50293,50352,50506,50655,50668,50743,50801,50973,50983,51175,51463,51641,51743,51871,51890,51949,52165,52259,52435,52528,52538,52557,52602,52668,52779,52817,52949,52981,53168,53193,53236,53361,53532,53818,53885,53916,54176,54207,54270,54384,54497,54522,54662,54873,54888,54899,54977,55082,55162,55308,55314,55319,55353,55457,55488,56014,56143,56167,56168,56171,56175,56221,56438,56440,56441,56500,56530,56605,56824,56878,56892,56922,56973,57122,57154,57279,57301,57556,57641,57712,57714,57722,57746,57870,57913,58004,58163,58167,58227,58259,58276,58306,58363,58381,58411,58464,58573,58838,58863,58865,59166,59253,59256,59529,59530,59558,59658,59666,59694,59695,59870,60000,60231,60449,60677,60739,60763,60821,60829,60870,60900,60935,61075,61163,61465,61510,61577,61612,61653,61670,61673,61709,61714,61757,61864,61873,61936,62013,62051,62083,62192,62333,62376,62534,62553,62589,62593,62630,62783,62854,62955,63459,63512,63529,63564,63620,63666,63905,63908,63917,64029,64150,64262,64295,64296,64312,64372,64412,64424,64643,64651,64711,64719,64763,64896,64996,65104,65135,65181,65221,65344,65380,65394,65606,65668,65736,65737,65855,65919,65994,66308,66663,66739,66855,67064,67452,67527,67535,67552,67807,67891,68013,68055,68064,68123,68288,68322,68544,68557,68585,68679,68792,69072,69149,69259,69499,69501,69592,69638,69665,69726,69935,69988,70003,70016,70074,70288,70380,70381,70512,70683,70697,70777,70788,70817,70925,70959,71008,71232,71336,71503,71506,71550,71616,71647,71886,72101,72189,72200,72445,72492,72558,72888,73082,73108,73146,73233,73306,73488,73658,73894,73989,74158,74201,74335,74355,74401,74412,74584,74837,74886,74974,75101,75109,75129,75314,75316,75388,75396,75418,75511,75552,75594,75603,75712,75720,75722,75725,75799,75803,75985,76178,76441,76533,76897,76999,77196,77320,77655,77746,78072,78137,78144,78348,78457,78479,78521,79029,79057,79102,79149,79171,79342,79416,79661,79739,79812,79824,79826,79899,80070,80080,80082,80170,80235,80285,80393,80397,80407,80483,80563,80608,80612,80671,80709,80860,81075,81151,81410,81525,81567,81805,82024,82067,82180,82182,82188,82308,82460,82549,82966,83027,83074,83209,83211,83311,83317,83341,83350,83538,83542,83555,83563,83669,83736,83767,83780,83814,83825,83993,84068,84336,84691,84737,84788,84911,84966,85001,85240,85321,85410,85569,85601,85615,85621,85718,85768,85806,85849,85892,85953,85975,86013,86142,86219,86366,86841,86945,87089,87115,87229,87303,87458,87759,87828,87834,87920,88156,88167,88230,88274,88281,88467,88522,88786,88884,88907,89052,89101,89114,89209,89264,89277,89458,89534,89730,89857,89945,89946,90042,90228,90240,90298,90352,90475,90499,90582,90590,90646,90702,90730,90794,90881,90939,90985,91011,91023,91097,91165,91184,91213,91317,91352,91428,91782,91794,91867,92168,92335,92482,92517,92624,92634,92653,92682,92704,92741,92761,92765,92908,92969,93034,93053,93090,93220,93286,93294,93427,93460,93508,93781,93844,93921,94080,94137,94488,94610,94741,94846,94913,94935,94985,95024,95200,95205,95238,95316,95393,95500,95537,95539,95742,95781,96058,96175,96243,96320,96472,96565,96627,96807,96953,97052,97082,97088,97100,97172,97182,97233,97295,97592,97660,97702,97856,97866,97923,97978,98041,98048,98110,98185,98222,98232,98281,98456,98557,98661,98683,98685,98809,98824,98852,98896,98905,98959,99012,99046,99360,99436,99456,99563,99580,99615,99776,99777,99864,99887,99980",
      "77,104,142,151,161,211,225,1662,1681,1683,1700,1738,1741,1757,1776,1782,1795,1798,2074,2082,2091,2487,2634,2654,2712,2782,2830,2886,3253,3406,3753,4018,4132,4148,4235,4257,4476,4644,4696,4767,6166,7208,8225,8806,9376,9395,9481,9523,9547,9604,9661,9680,9699,9718,9740,9775,9791,9832,9984,10022,10060,10649,10703,10749,10858,10877,10953,11002,11029,11119,11200,11238,11257,11352,11447,11485,11580,11607,11618,11637,11694,11903,11960,11998,12055,12061,12112,12131,12226,12264,12283,12299,12302,12333,12340,12353,12397,12436,12442,12473,12519,12530,12568,12592,12663,12709,12766,12777,12801,12830,12835,12853,12872,12920,12978,12986,13005,13015,13044,13081,13082,13096,13103,13158,13271,13290,13309,13480,13485,13491,13632,13638,13670,13693,13708,13746,13765,13784,13803,13822,13898,13909,13917,13956,13993,14069,14091,14098,14107,14183,14259,14278,14297,14316,14335,14354,14411,14563,14601,14615,14636,14639,14760,14772,14829,14857,14867,14883,14905,14962,15000,15038,15076,15114,15171,15182,15190,15213,15228,15247,15285,15304,15323,15361,15380,15456,15513,15574,15589,15608,15630,15637,15684,15722,15741,15817,15836,15855,15874,15893,15926,15931,15950,15969,16026,16045,16102,16121,16178,16216,16235,16254,16292,16311,16330,16349,16368,16463,16482,16485,16501,16558,16577,16634,16653,16669,16672,16691,16710,16805,16824,16881,16919,16957,16995,17014,17049,17052,17109,17111,17128,17147,17204,17280,17299,17375,17394,17413,17451,17477,17527,17546,17584,17603,17616,17629,17660,17679,17701,17722,17755,17764,17774,17788,17831,17850,17869,17907,17912,17964,17998,18002,18021,18040,18047,18065,18097,18135,18154,18166,18211,18230,18249,18306,18325,18363,18382,18401,18420,18458,18486,18496,18534,18553,18610,18629,18686,18705,18743,18762,18800,18819,18838,18857,18876,18952,18990,19019,19047,19066,19104,19106,19142,19199,19218,19275,19294,19313,19332,19408,19446,19465,19473,19541,19579,19598,19617,19636,19688,19693,19731,19769,19788,19826,19845,19883,19902,19921,19940,19959,20016,20028,20035,20072,20073,20092,20146,20149,20187,20206,20214,20225,20263,20339,20358,20377,20396,20415,20434,20472,20510,20565,20586,20605,20643,20662,20681,20776,20814,20833,20845,20852,20890,20926,20966,20985,21000,21004,21009,21023,21042,21055,21061,21099,21128,21135,21175,21186,21232,21251,21308,21346,21365,21377,21403,21441,21460,21467,21498,21515,21536,21555,21574,21612,21631,21632,21701,21712,21726,21764,21765,21783,21802,21821,21859,21878,21897,21935,21954,21973,22011,22030,22049,22087,22101,22106,22144,22163,22181,22220,22239,22277,22296,22334,22353,22429,22448,22505,22524,22562,22581,22600,22676,22690,22733,22752,22771,22828,22866,22885,22904,22942,22961,22966,23018,23037,23056,23065,23075,23091,23094,23113,23132,23227,23265,23284,23322,23338,23341,23360,23398,23417,23474,23507,23512,23550,23569,23588,23645,23664,23683,23740,23744,23759,23775,23797,23816,23873,23892,23911,23949,23968,23987,24006,24025,24044,24063,24074,24082,24139,24158,24196,24215,24272,24299,24310,24329,24367,24424,24443,24459,24462,24519,24538,24576,24595,24614,24652,24728,24747,24785,24804,24823,24842,24899,24918,24931,24937,24956,24975,24994,25079,25108,25112,25119,25146,25165,25222,25241,25259,25260,25298,25317,25374,25379,25431,25437,25450,25488,25507,25561,25564,25602,25621,25640,25697,25735,25754,25773,25787,25830,25849,25868,25887,25906,25925,25930,25963,25982,26001,26039,26058,26096,26115,26191,26229,26245,26286,26324,26343,26400,26457,26495,26514,26552,26634,26647,26666,26704,26742,26793,26856,26913,26952,26968,26970,27008,27027,27217,27293,27407,27419,27504,27555,27578,27635,27703,27730,27748,27749,27768,27788,27901,27918,27932,27965,27996,28028,28087,28091,28110,28186,28205,28257,28332,28357,28369,28400,28471,28692,28775,28813,28828,28832,28834,28965,29003,29041,29060,29064,29098,29139,29193,29212,29250,29261,29300,29326,29364,29383,29411,29459,29478,29497,29668,29687,29702,29782,29814,29891,29916,30009,30038,30067,30096,30125,30154,30183,30212,30241,30299,30328,30357,30386,30415,30444,30473,30502,30531,30560,30589,30618,30647,30676,30705,30734,30763,30792,30821,30850,30879,30908,30937,30966,30995,31024,31053,31082,31111,31140,31169,31198,31227,31256,31285,31314,31343,31372,31401,31430,31459,31488,31517,31546,31575,31604,31633,31662,31691,31720,31749,31778,31807,31836,31865,31894,31923,31952,31981,32010,32039,32068,32097,32126,32155,32184,32213,32242,32271,32300,32329,32358,32387,32416,32445,32503,32532,32561,32590,32619,32648,32677,32706,32735,32764,32793,32822,32851,32880,32909,32938,32967,32996,33025,33054,33083,33112,33141,33170,33199,33228,33257,33286,33315,33344,33373,33402,33431,33460,33489,33518,33547,33576,33605,33634,33663,33692,33721,33750,33779,33808,33837,33866,33895,33924,33953,33982,34011,34040,34069,34098,34127,34156,34185,34214,34243,34272,34301,34330,34359,34388,34417,34475,34504,34533,34562,34591,34620,34649,34678,34707,34736,34765,34794,34823,34852,34881,34910,34939,34968,34997,35009,35054,35098,35136,35172,35261,35314,35357,35409,35447,35455,35456,35480,35528,35587,35646,35670,35683,35719,35751,35753,35765,35778,35811,35816,35837,35908,35946,35967,36026,36089,36149,36174,36208,36279,36303,36307,36338,36397,36437,36528,36564,36587,36642,36651,36710,36769,36813,36840,36871,36882,36899,36935,36958,37030,37089,37139,37148,37212,37271,37313,37330,37367,37401,37460,37519,37591,37650,37673,37709,37730,37757,37764,37773,37832,37866,37882,37891,37962,37985,38021,38080,38115,38152,38211,38270,38334,38384,38393,38452,38473,38523,38582,38641,38713,38772,38831,38895,38954,38983,39013,39050,39084,39143,39202,39241,39243,39274,39333,39356,39392,39414,39415,39456,39493,39515,39554,39574,39645,39669,39670,39688,39704,39763,39835,39894,39953,40104,40133,40162,40191,40220,40249,40278,40307,40336,40365,40394,40423,40452,40481,40510,40539,40568,40597,40626,40655,40713,40742,40771,40800,40829,40858,40887,40916,40945,40974,41003,41032,41061,41090,41119,41148,41177,41206,41235,41264,41293,41351,41380,41409,41438,41467,41496,41525,41929,41932,41935,41938,41941,41944,41947,41950,41953,41960,41964,41968,41972,41976,41980,41984,41989,41994,42023,42052,42081,42110,42162,42170,42215,42233,42245,42294,42320,42329,42400,42410,42484,42557,42625,42647,42677,42678,42713,42721,42794,42876,42877,42884,42929,42950,42958,43002,43017,43031,43121,43195,43197,43247,43268,43312,43328,43344,43351,43353,43358,43432,43483,43496,43505,43583,43595,43669,43680,43723,43736,43742,43746,43788,43791,43793,43811,43832,43868,43906,43913,43920,43979,44047,44054,44061,44069,44101,44124,44139,44143,44216,44306,44380,44387,44428,44447,44453,44514,44532,44537,44543,44617,44657,44690,44695,44764,44780,44854,44862,44903,44913,44917,44919,44926,44927,44973,44988,45009,45011,45017,45057,45077,45090,45129,45138,45199,45257,45283,45330,45373,45387,45445,45477,45483,45491,45517,45546,45574,45632,45704,45754,45761,45819,45846,45891,45948,46005,46033,46064,46097,46108,46155,46159,46298,46305,46349,46377,46378,46389,46415,46420,46634,46636,46642,46666,46679,46744,46773,46795,46842,46846,46898,46903,46919,46974,46975,46998,47040,47124,47160,47183,47198,47202,47252,47291,47342,47357,47363,47403,47417,47430,47446,47452,47491,47514,47547,47556,47675,47714,47717,47720,47725,47735,47814,47824,47839,47884,47893,47906,48146,48153,48203,48351,48355,48426,48434,48436,48580,48626,48631,48689,48690,48697,48714,48764,48766,48781,48795,48818,48829,48862,48870,48908,48926,48988,49027,49039,49044,49056,49065,49066,49150,49176,49180,49181,49213,49241,49305,49353,49363,49371,49377,49422,49439,49467,49499,49542,49598,49623,49628,49643,49702,49784,49801,49827,49835,49838,49917,50001,50085,50248,50254,50370,50390,50407,50466,50554,50617,50756,50823,50942,51203,51205,51270,51314,51316,51323,51378,51538,51585,51598,51617,51872,52036,52212,52213,52280,52383,52562,52581,52619,52652,52653,52924,53103,53109,53456,53563,53567,53613,53634,53692,53785,53846,53853,53934,53953,53970,53983,54028,54074,54144,54161,54204,54322,54392,54504,54532,54570,54579,54590,54646,54686,54750,54783,54822,54970,55072,55111,55279,55407,55455,55475,55537,55679,55739,55803,55817,55980,56139,56229,56373,56408,56434,56442,56464,56536,56651,56668,56672,56706,56711,56744,56841,57269,57278,57398,57410,57413,57432,57595,57702,57706,57711,57899,57964,58253,58490,58504,58614,58714,58749,58856,58981,59067,59133,59208,59214,59228,59327,59420,59562,59583,59646,59761,59897,59975,60119,60192,60323,60391,60490,60521,60633,60713,60718,60842,60865,60916,60936,60970,61000,61073,61189,61190,61225,61233,61261,61309,61354,61395,61414,61468,61639,61650,61710,61751,61877,61880,61924,61966,62092,62136,62154,62156,62269,62537,62897,62929,62992,63165,63192,63260,63525,63719,63773,63793,63829,63882,63911,64124,64154,64180,64207,64211,64293,64333,64449,64452,64486,64587,64589,64599,64759,64886,64911,64944,65055,65059,65326,65441,65461,65484,65503,65610,65641,65657,65843,65885,65912,65935,65938,66060,66227,66291,66514,66529,66556,66657,66882,66949,66975,67044,67051,67095,67104,67112,67152,67290,67389,67463,67500,67641,67660,67701,67728,67975,67979,67997,68063,68101,68152,68192,68324,68349,68541,68614,68752,68796,68971,69008,69016,69063,69156,69287,69329,69336,69387,69394,69435,69442,69591,69593,69889,69898,69943,70143,70177,70182,70398,70541,70559,70582,70637,70652,70659,70687,70716,70728,70926,71057,71096,71152,71194,71212,71324,71377,71554,71573,71724,71756,71779,71823,71832,71961,72122,72220,72410,72480,72530,72557,72780,72832,72852,72996,73079,73085,73123,73171,73431,73492,73562,73667,73690,73738,73828,73893,73928,74198,74214,74361,74596,74610,74631,74634,74724,74825,75036,75077,75121,75259,75480,75490,75703,75706,75804,75925,76024,76065,76066,76157,76236,76362,76394,76620,76762,76827,76837,76846,76864,76979,77027,77074,77105,77109,77211,77235,77323,77351,77440,77458,77818,77985,78156,78342,78365,78390,78436,78442,78514,78529,78713,78970,79073,79093,79148,79208,79213,79220,79245,79252,79262,79283,79302,79512,79520,79526,79536,79694,79750,79870,79883,79896,79924,80015,80027,80069,80083,80127,80219,80379,80428,80503,80626,80648,80749,80781,80875,80976,81016,81072,81129,81322,81370,81414,81485,81486,81509,81535,81555,81729,81754,81792,81838,81950,82022,82153,82216,82276,82316,82326,82335,82339,82410,82481,82530,82678,82688,82745,82800,83077,83126,83196,83218,83238,83298,83329,83340,83371,83435,83532,83704,83760,83791,83997,84094,84163,84213,84278,84307,84440,84569,84688,84803,84818,84887,84968,85023,85107,85250,85268,85289,85405,85413,85457,85703,85918,85979,86116,86182,86231,86267,86279,86419,86476,86497,86527,86585,86626,86753,86765,86766,86956,86982,87041,87162,87197,87322,87365,87398,87401,87416,87649,87674,87771,87890,87907,87914,88041,88050,88089,88126,88322,88419,88445,88463,88481,88517,88621,88824,88846,88999,89010,89017,89116,89244,89355,89361,89373,89516,89624,89705,89858,89864,90130,90347,90361,90421,90462,90480,90486,90529,90568,90681,90687,90772,90814,91277,91393,91573,91633,91666,91734,92097,92173,92191,92249,92506,92523,92537,92584,92775,92851,92876,92877,93002,93043,93323,93334,93376,93390,93441,93509,93557,93641,93749,93977,94094,94133,94141,94350,94466,94601,94627,95113,95305,95358,95361,95409,95414,95509,96019,96176,96249,96321,96512,96826,96918,97023,97024,97031,97043,97087,97229,97285,97445,97456,97500,97597,97627,97670,97951,98099,98187,98228,98345,98384,98534,98633,98690,99232,99307,99413,99428,99527,99538,99653,99655,99761,99877,99910",
      "80,172,332,378,395,404,516,543,731,927,1020,1030,1076,1180,1196,1246,1261,1397,1440,1523,1628,1699,1704,1770,1823,2089,2168,2256,2309,2784,2838,3013,3028,3175,3273,3341,3827,3878,3896,3907,4100,4386,4574,4774,4818,4902,5003,5011,5246,5285,5300,5308,5349,5385,5387,5412,5422,5443,5694,5765,5859,5904,5962,5971,6118,6267,6318,6351,6417,6540,6544,6719,6793,6835,6896,6907,6961,6983,7106,7132,7151,7194,7201,7329,7506,7536,7537,7555,7565,7622,7671,7746,7764,7801,7928,8088,8241,8448,8491,8607,8957,9177,9251,9272,9323,9349,9437,9542,9592,9610,9792,9826,9844,9893,9906,10197,10246,10273,10275,10485,10626,10822,10854,11001,11184,11196,11204,11341,11472,11475,11488,11520,11723,11742,11908,12220,12231,12293,12332,12420,12458,12487,12509,12518,12562,12605,12644,12753,12788,12837,12842,12890,12917,13002,13122,13191,13286,13318,13334,13460,13517,13727,13794,13851,14003,14074,14141,14234,14288,14362,14539,14653,14695,14765,14799,14941,15153,15291,15490,15518,15542,15548,15607,15729,15840,15940,15955,15968,15985,16012,16036,16129,16230,16394,16828,16832,16880,16889,16897,16904,16966,16980,17008,17023,17199,17256,17351,17423,17452,17454,17470,17513,17917,17938,18084,18109,18264,18304,18461,18644,18718,18984,18995,19095,19241,19255,19258,19324,19431,19760,19816,19818,19846,19986,20053,20120,20257,21200,21228,21391,21642,21645,21733,21926,21987,22028,22071,22148,22313,22325,22475,22552,22751,22758,22860,22895,22914,23038,23168,23186,23212,23221,23241,23345,23379,23506,23592,23991,24000,24059,24119,24143,24162,24230,24294,24381,24467,24491,24567,24568,24668,24709,24889,24980,25154,25162,25387,25514,25624,25802,25837,26022,26152,26231,26335,26524,26897,26936,26950,26956,26994,26995,27016,27155,27189,27305,27356,27490,27686,27739,27832,27904,27953,28163,28215,28411,28492,28533,28736,29102,29127,29186,29206,29215,29283,29293,29396,29465,29490,29821,29890,29902,29918,29976,30052,30487,31653,31763,32175,32898,33048,33077,33364,33744,33942,34518,35149,35150,35238,35276,35734,35887,36015,36068,36077,36222,36294,36451,36455,36456,36484,36511,36546,36553,36554,36699,36747,36821,36854,36873,36942,37135,37197,37547,37567,37569,37678,37992,38098,38123,38127,38199,38239,38417,38512,38565,38607,38845,38851,38942,39007,39306,39319,39379,39380,39485,39486,39490,39614,39722,39783,39888,39908,40114,40129,40812,41334,42319,42494,42565,43034,43115,43119,43198,43221,43476,43522,44144,44458,45428,45438,45502,45691,45875,46198,47326,47608,48151,48198,48364,48765,49040,49428,49592,49639,50054,50068,50091,50094,50184,50296,50327,50432,50515,50616,50623,50630,50639,50640,50652,50690,50713,50720,50859,51068,51196,51224,51285,51362,51417,51484,51505,51546,51615,51628,51692,51885,51899,51908,52088,52178,52284,52401,52564,52658,52694,52701,52836,52896,53042,53100,53145,53419,53505,53523,53670,53719,53764,53786,53813,53940,54042,54181,54354,54369,54378,54401,54432,54502,54578,54643,54645,54788,54824,54887,54933,54982,55110,55118,55241,55391,55483,55561,56098,56370,56467,56592,56628,56720,56898,56940,56947,57026,57049,57075,57171,57174,57181,57247,57273,57305,57455,57471,57477,57488,57544,57588,57657,57771,57791,57918,57920,58159,58252,58593,58702,58770,58976,59025,59211,59450,59515,59574,59651,59692,59778,59952,60059,60060,60216,60262,60329,60346,60465,60662,60669,60752,60781,60810,60819,60888,61009,61067,61083,61109,61151,61158,61404,61622,61807,61849,61869,62240,62260,62266,62323,62439,62506,62508,62637,62737,62817,62863,63132,63175,63196,63203,63256,63320,63333,63383,63404,63420,63562,63645,63670,63676,63806,63866,63876,63891,63945,63949,63964,64099,64301,64315,64398,64419,64429,64442,64527,64608,64623,64664,64686,64720,64817,64927,65047,65092,65152,65182,65235,65329,65392,65402,65439,65481,65573,65711,65846,65907,65973,66323,66330,66447,66461,66503,66591,66782,66802,66838,66857,66904,67165,67209,67214,67424,67437,67438,67444,67653,67719,67735,67771,67904,68041,68096,68276,68407,68462,68567,68588,68692,68927,68948,69005,69107,69404,69408,69492,69508,69518,69600,69645,69698,69778,69828,69984,70195,70298,70342,70356,70357,70407,70681,70689,70700,70762,71032,71186,71267,71571,71765,71792,71801,71862,71949,72032,72069,72269,72459,72464,72503,72608,72623,72658,72790,72811,72827,72991,73067,73074,73091,73238,73383,73583,73824,73836,74063,74117,74161,74262,74378,74405,74862,75164,75210,75251,75261,75324,75400,75503,75545,75616,75633,75719,75739,75767,75921,76037,76154,76202,76317,76334,76352,76617,76699,76816,77079,77092,77178,77357,77411,77535,77735,77793,77890,77938,78035,78215,78257,78308,78366,78413,78427,78474,78516,78615,78771,78797,79001,79153,79206,79243,79299,79385,79516,79609,79708,79714,79742,79873,79909,79911,79949,80120,80123,80258,80304,80376,80538,80546,80631,80842,81172,81313,81323,81373,81387,81665,81718,81753,81878,82018,82043,82075,82412,82437,82490,82572,82630,82682,82722,82728,82747,82811,82820,82821,82832,82888,82898,82942,82988,83062,83436,83556,83616,83664,83689,83720,83761,83800,83869,83883,83896,83897,84065,84120,84173,84205,84512,84541,84717,84858,84862,84895,85070,85113,85128,85168,85208,85229,85267,85292,85337,85511,85519,85541,85709,85866,85881,86010,86122,86171,86406,86538,86553,86575,86663,86693,86789,86795,86806,86808,86853,86906,87042,87254,87299,87326,87327,87459,87568,87570,87625,87671,87672,87716,87745,87765,87819,87836,87949,87974,87985,88005,88097,88185,88225,88236,88499,88560,88730,88780,88799,88993,89018,89105,89283,89316,89330,89362,89819,89827,89855,89900,89953,89958,90275,90427,90453,90460,90613,90648,90889,90970,90982,91054,91093,91151,91186,91259,91281,91290,91409,91458,91485,91523,91545,91701,91707,91816,91878,92009,92065,92083,92134,92199,92246,92293,92337,92339,92384,92522,92590,92774,92906,92916,92936,92944,93096,93153,93296,93343,93346,93378,93630,93733,93906,93939,94084,94157,94366,94484,94680,94764,94783,94811,94963,94992,95055,95180,95344,95378,95462,95464,95493,95514,95602,95708,95754,95783,95823,95962,96082,96100,96111,96454,96517,96613,96628,96655,96788,96804,96872,96985,97015,97040,97187,97279,97310,97376,97396,97415,97546,97555,97732,97745,97752,97857,97860,97896,97918,98141,98361,98708,98878,98899,98936,99028,99254,99266,99333,99497,99621,99680,99798,99893,99998",
      "10,38,97,100,158,165,375,583,647,655,657,786,875,1032,1056,1058,1073,1095,1103,1109,1161,1221,1228,1257,1259,1308,1330,1493,1544,1568,1571,1639,1650,1658,1718,1723,1743,2041,2117,2120,2200,2222,2485,2668,2670,2732,2806,2824,3159,3348,3357,3734,3777,3797,4122,4259,4300,4467,4511,4672,4733,4850,4912,4952,5019,5137,5179,5238,5267,5316,5320,5520,5531,5566,5707,5761,5836,5947,6057,6059,6179,6181,6204,6246,6259,6415,6464,6694,6795,6809,6884,7072,7111,7212,7392,7446,7491,7496,7856,7871,7980,7986,7990,8035,8260,8276,8367,8454,8502,8584,8707,8723,8785,9002,9067,9298,9410,9447,9502,9514,9579,9616,9802,9855,10073,10147,10172,10174,10233,10234,10264,10332,10343,10356,10449,10495,10509,10537,10542,10557,10580,10721,10814,10839,11049,11077,11108,11310,11400,11462,11533,11555,11655,11664,11747,11843,12182,12184,12319,12651,12669,12715,12783,12829,12961,13077,13129,13150,13341,13445,13619,13824,13892,13961,14038,14108,14412,14458,14494,14545,14577,14591,14605,14882,14932,15020,15084,15095,15205,15260,15278,15392,15550,15699,15715,15725,15759,15830,16038,16153,16445,16516,16588,16596,16663,16719,16781,16912,16934,17075,17131,17143,17184,17238,17314,17332,17446,17523,17692,17786,17826,17865,18000,18173,18194,18251,18323,18376,18524,18698,18708,18831,18935,18957,18982,19004,19060,19114,19326,19447,19471,19581,19635,19677,19744,19799,19917,20009,20021,20044,20046,20048,20051,20677,20715,20922,20980,21030,21144,21193,21705,21808,21836,21872,21887,21949,21992,22235,22338,22366,22423,22454,22520,22710,23153,23246,23531,23765,24305,24306,24438,24475,24533,24592,24637,24643,24793,24875,24930,25008,25052,25292,25311,25351,25422,25583,25680,25862,26148,26295,26312,26330,26489,26749,26891,26931,27050,27191,27549,27645,27654,27663,27684,27712,27721,27875,27906,27940,27973,28280,28383,28393,28441,28681,28693,28711,28863,28883,28958,29021,29045,29195,29243,29260,29267,29306,29322,29333,29339,29425,29566,29736,29877,30003,30352,30381,30787,31135,31163,31280,31322,31397,31570,31687,31831,32381,32499,32586,32962,32963,33050,33454,33484,33485,33513,33687,34498,34557,34674,34731,34759,34847,34993,35004,35294,35635,35713,35814,36078,36198,36254,36312,36322,36463,36509,36559,36560,36701,36745,36765,36820,36829,37008,37079,37265,37441,37443,37554,37565,37805,37870,37879,37946,37947,38011,38125,38207,38316,38382,38442,38742,39251,39269,39320,39428,39504,39568,39611,39949,39993,40147,40256,40517,41039,41245,41303,41481,42061,42197,42234,42289,42398,42616,42676,42680,42745,42804,42838,42846,42875,42968,43061,43172,43206,43264,43348,43385,43768,43824,44208,44305,44448,44477,44515,44570,44677,44779,44809,44825,44830,44882,45071,45325,45498,45504,45616,45689,45713,45814,45877,45986,45988,46059,46441,46491,46529,46594,46616,46772,46838,47226,48089,48178,48420,48498,48542,49032,49319,49603,49853,50354,50413,50436,50685,50851,51289,51324,51326,51446,51535,51613,51681,51828,51894,51940,52110,52133,52245,52405,52451,52483,52572,52633,52688,52713,52720,52731,52775,52831,52977,53007,53112,53167,53295,53301,53396,53450,53485,53524,53575,53580,53643,53705,53886,53987,53989,53992,53998,54130,54137,54195,54215,54224,54238,54290,54325,54527,54599,54611,54619,54631,54712,54713,54714,54753,54851,54872,54911,55101,55109,55164,55363,55402,55403,55486,55535,55555,55585,55992,56095,56117,56181,56233,56304,56354,56394,56465,56567,56572,56596,56624,56709,56724,56860,57025,57053,57084,57153,57217,57329,57345,57508,57674,57739,57760,57938,57973,58026,58048,58127,58139,58177,58245,58258,58429,58619,58642,59010,59086,59207,59312,59481,59551,59655,59693,59739,59895,59937,60099,60244,60287,60359,60451,60560,60705,60987,61016,61074,61116,61226,61401,61696,61707,61829,61896,62011,62054,62075,62140,62237,62242,62291,62434,62514,62556,62577,62635,62724,62741,62746,62837,62939,63044,63088,63146,63354,63382,63431,63580,63592,63753,63777,63863,63883,63935,63963,64291,64425,64473,64492,64499,64571,64583,64607,64635,64963,64982,65114,65229,65515,65707,65834,65890,65899,65969,66018,66122,66139,66272,66294,66454,66537,66541,66594,66648,66735,66863,66881,67004,67010,67114,67154,67198,67205,67231,67288,67321,67457,67473,67560,67691,67723,67784,67928,68003,68045,68070,68105,68106,68116,68122,68183,68317,68335,68384,68401,68404,68409,68437,68606,68803,68893,68933,68995,69056,69152,69191,69278,69382,69384,69420,69453,69500,69610,69674,69687,69795,69926,69973,70051,70145,70196,70264,70279,70305,70330,70393,70397,70436,70515,70718,70956,71330,71475,71479,71489,71614,71719,71751,71866,71941,71975,72175,72230,72250,72284,72326,72422,72500,72515,72543,72715,72737,72843,72902,72908,72978,73057,73098,73323,73348,73433,73537,73554,73594,73612,73700,73734,73758,73818,73940,74004,74016,74018,74130,74228,74319,74323,74332,74336,74416,74503,74526,74670,74673,74784,74807,75022,75242,75348,75350,75406,75493,75500,75779,75789,75840,75871,75892,76008,76013,76044,76075,76083,76212,76378,76463,76465,76476,76509,76560,76601,76936,76941,77042,77194,77272,77299,77414,77420,77422,77463,77491,77504,77698,77800,77865,77924,77964,78185,78288,78339,78374,78403,78582,78695,78755,78770,78777,78780,78801,78854,78928,79118,79135,79199,79371,79555,79616,79666,79667,79679,79888,79913,80000,80186,80267,80396,80418,80436,80575,80649,80780,80935,80962,80982,81249,81345,81564,81701,81724,81748,81882,81952,81968,82154,82210,82213,82274,82275,82386,82494,82783,82861,82950,82997,83039,83336,83406,83427,83497,83584,83585,83879,83935,83948,84058,84152,84330,84353,84409,84645,84881,85428,85451,85471,85597,85815,85862,86033,86079,86088,86163,86174,86206,86339,86348,86433,86469,86576,86616,86801,86914,86926,86955,87003,87033,87063,87066,87076,87220,87393,87444,87508,87541,87554,87786,87842,87983,88033,88083,88107,88115,88222,88280,88303,88387,88400,88452,88475,88641,88685,89046,89066,89072,89312,89345,89413,89455,89519,89618,89645,89791,89835,89850,89856,89932,89935,90125,90193,90213,90451,90455,90487,90495,90546,90653,90705,90926,90938,91004,91109,91185,91265,91450,91556,91566,91570,91649,91732,91737,91893,92069,92108,92158,92167,92292,92310,92356,92404,92407,92431,92533,92538,92551,92731,92995,93285,93302,93331,93448,93619,93756,93956,93987,94097,94148,94169,94275,94413,94453,94462,94482,94533,94652,94657,94658,94750,94850,94951,95138,95346,95390,95528,95569,95675,95681,95713,96027,96053,96101,96121,96154,96182,96212,96255,96304,96354,96432,96538,96562,96595,96608,96642,96686,96706,96765,96885,97027,97051,97079,97105,97144,97341,97508,97576,97692,97799,97943,97992,97999,98076,98119,98234,98276,98323,98343,98435,98678,98756,98826,98916,98948,98964,99020,99024,99101,99300,99384,99392,99399,99407,99455,99614,99666,99748,99880,99912,99987",
      "25,93,117,154,212,265,281,618,660,776,960,1047,1066,1113,1130,1204,1241,1340,1383,1416,1478,1562,1598,1619,1714,1726,1766,1777,1914,2128,2263,2304,2317,2509,2527,2569,2948,3066,3580,3617,3674,3715,4378,4620,4777,5174,5207,5211,5227,5232,5269,5276,5288,5337,5345,5416,5446,5458,5462,5472,5476,5487,5593,5661,5670,5701,5732,5735,5795,5869,5930,5973,5977,5996,6060,6083,6175,6217,6233,6268,6269,6281,6387,6390,6589,6681,6684,6705,6786,6827,6851,6911,6943,7030,7043,7061,7069,7300,7318,7467,7562,7592,7610,7751,7753,7759,7791,7799,7845,7859,7878,7883,7897,7947,8308,8442,8452,8548,8582,8680,8871,8960,8991,9000,9015,9032,9038,9118,9176,9784,9870,9964,9966,10011,10019,10133,10201,10367,10737,10758,10795,10910,11294,11337,11354,11577,11650,11767,11775,11791,11869,12305,12307,12641,12657,12743,12785,12943,12948,12966,12970,13057,13115,13364,13400,13867,13868,14066,14336,14466,14514,14621,14750,14771,14940,15058,15109,15293,15448,15474,15584,15660,15754,16072,16709,17069,17259,17357,17455,17503,17554,17586,17711,17763,18079,18428,18572,18640,18722,18922,18971,19124,19260,19273,19626,19777,20885,21126,21142,21402,22160,22669,22708,22741,23007,23064,23318,23589,23615,23938,24271,24597,24638,24650,25087,25395,25705,25983,26377,26415,26573,26628,26838,26898,27003,27213,27273,27314,27825,27881,27968,28061,28270,28388,28757,28864,28979,29132,29420,29796,29828,31047,31419,31564,32521,32637,33275,33768,33826,34377,34551,35508,35627,35696,35698,35737,35891,36122,36494,36574,36680,36693,36926,37425,38357,38424,38695,38790,38849,39122,39247,39301,39870,40460,42154,42308,42325,42373,42456,42640,42708,42718,42795,42809,43100,43710,43884,43892,43954,44420,45178,45313,45447,45539,45935,46855,46955,47147,48185,48457,48835,49390,49518,50057,50067,50073,50362,50439,50480,50613,50620,50879,50906,51011,51051,51074,51084,51124,51186,51236,51408,51472,51480,51611,51887,51892,52081,52104,52204,52225,52296,52384,52391,52436,52542,52583,52819,52901,52910,52941,53289,53474,53935,54080,54138,54175,54182,54219,54239,54264,54312,54319,54391,54431,54553,54778,54816,54823,54917,54944,54947,54968,55004,55248,55269,55401,55494,55590,55683,55726,55825,55834,55881,55953,56080,56183,56222,56426,56531,56538,56546,56686,56827,56906,56945,56946,57052,57151,57177,57193,57222,57232,57292,57314,57335,57381,57403,57515,57614,57709,57752,57757,57805,57867,58120,58280,58609,58620,58849,58850,59063,59215,59273,59313,59473,59548,59628,59725,59770,59998,60050,60264,60418,60518,60547,60587,60731,60806,61010,61058,61273,61294,61314,61365,61388,61423,61440,61446,61460,61526,61645,62105,62124,62127,62153,62212,62226,62378,62450,62463,62464,62562,62663,62710,62745,62829,62840,62858,62913,62944,62988,63091,63126,63328,63506,63601,63602,63646,63780,63877,63884,63929,63959,63987,64032,64210,64228,64314,64457,64522,64609,64706,64708,64756,64765,64955,64967,65038,65073,65282,65312,65320,65427,65456,65473,65630,65926,66030,66213,66366,66367,66422,66689,66824,66865,66999,67013,67099,67148,67189,67302,67317,67639,67957,68056,68089,68190,68226,68241,68294,68596,68617,68675,68801,68841,68973,69064,69187,69248,69346,69389,69504,69542,69545,69587,69603,69725,69831,69912,69938,69972,69989,70187,70382,70426,70466,70518,70534,70563,70639,70640,70758,70768,70773,70955,70982,71000,71086,71144,71167,71199,71240,71410,71416,71484,71520,71563,71679,71850,72029,72089,72125,72205,72242,72322,72380,72453,72472,72544,72545,72613,72628,72663,72674,72729,72776,72819,72997,73008,73010,73012,73062,73069,73122,73231,73299,73358,73463,73511,73600,73623,73651,73665,73749,74057,74149,74150,74188,74372,74489,74740,74773,74782,74843,74880,74947,74948,75007,75056,75058,75067,75089,75271,75342,75538,75548,75750,75956,76047,76052,76147,76169,76389,76583,76613,76636,76677,76711,76760,76785,76825,76851,76876,76963,76982,76998,77018,77125,77202,77288,77321,77359,77406,77415,77438,77541,77559,77597,77742,77768,77806,77810,77866,77958,77962,78062,78069,78192,78322,78410,78541,78649,78826,78900,78907,78968,78974,79087,79168,79173,79223,79352,79497,79546,79572,79587,79604,79621,79761,79781,79796,79845,79876,80002,80424,80447,80514,80571,80599,80600,80604,80669,80690,80693,80722,80759,81059,81113,81209,81224,81261,81365,81432,81497,82087,82140,82218,82225,82483,82525,82540,82743,82802,82884,82890,82905,82930,82984,82999,83001,83020,83203,83232,83277,83293,83354,83368,83437,83447,83455,83467,83565,83603,83649,83651,83722,83724,83829,83846,83880,83884,83898,83929,83932,83995,84117,84155,84170,84214,84406,84445,84469,84477,84497,84608,84664,84665,84681,84684,84724,84841,84901,84939,84959,85000,85132,85266,85333,85354,85390,85419,85441,85477,85580,85631,85641,85688,85711,85900,86349,86361,86481,86494,86599,86709,86843,86854,86932,87067,87149,87227,87233,87252,87311,87492,87578,87582,87738,87750,87821,87874,87875,87951,87998,88128,88151,88246,88376,88529,88592,88653,88672,88897,88903,88934,89089,89225,89268,89332,89444,89484,89537,89589,89981,90110,90173,90384,90404,90412,90417,90509,90755,90768,90831,90882,91016,91019,91183,91635,91790,91842,91884,91980,92316,92447,92706,92867,92885,93183,93196,93274,93405,93439,93483,93485,93582,93757,93890,93903,93955,93979,94000,94013,94042,94184,94189,94196,94235,94288,94370,94408,94452,94477,94555,94706,94744,94773,94805,94860,94936,95021,95124,95206,95212,95365,95373,95451,95456,95604,95657,95693,95833,95963,96070,96116,96257,96266,96338,96346,96437,96615,96735,96774,96796,96812,96997,97208,97224,97323,97427,97464,97494,97535,97560,97661,97891,97987,98038,98176,98290,98383,98434,98478,98535,98578,98617,98720,98778,98859,98950,98997,99100,99104,99246,99411,99560,99671,99691,99766,99817,99920",
      "67,255,270,275,306,329,330,349,390,399,485,513,616,676,692,701,736,740,850,1031,1054,1146,1358,1359,1421,1550,1609,1893,2850,3130,3166,3615,3747,4007,4423,4432,4835,4894,5237,5305,5466,5558,5609,5639,5657,5860,5997,6109,6122,6211,6249,6274,6282,6329,6355,6573,6614,6633,6671,6762,6799,6905,6956,7025,7244,7264,7325,7375,7379,7480,7542,7603,7609,7613,7730,7738,7739,7837,7982,8141,8271,8318,8357,8381,8403,8435,8466,8558,8589,8644,8832,8849,8984,9048,9087,9096,9202,9340,9389,9475,9746,9900,10076,10153,10178,10215,10281,10472,10508,10544,10565,10615,10665,10686,10695,10714,10716,10728,11091,11098,11311,11318,11335,11568,11661,11815,11839,11894,11988,12053,12100,12108,12303,12460,12466,12497,12541,12581,12711,13032,13080,13256,13315,13633,14542,14668,14754,14833,15375,15385,15412,15540,15674,15975,16231,16414,16630,16660,16952,17065,17276,17391,17681,18272,18384,18511,18658,18704,18753,18882,19270,19487,19583,19733,19787,19977,20126,20224,20260,20284,20501,20797,20828,20830,21137,22137,22191,22728,22769,22974,24155,24185,24694,24914,25518,25634,25711,25822,26130,26314,26817,26908,27009,27037,27160,27467,27470,27707,27737,27802,28287,28505,28898,28904,28925,29000,29104,29344,29362,29406,29509,29586,29640,29999,30868,30986,35340,35576,35577,35806,35951,35957,36117,36381,36736,36741,36947,37051,37202,37240,37427,37622,37625,37676,37694,37735,37949,37987,38069,38070,38120,38604,38688,38984,39054,39133,39238,39250,39260,39378,39478,39612,40567,41740,42148,43267,43602,45245,45506,48349,50555,50559,50561,50618,50682,50797,50926,50931,50969,51091,51112,51136,51172,51250,51312,51318,51345,51494,51588,51731,51818,51882,51926,51948,52057,52076,52086,52094,52147,52290,52297,52409,52500,52514,52635,52698,52807,52866,52982,52988,53026,53491,53528,53534,53545,53596,53780,53960,54003,54031,54103,54190,54197,54248,54262,54310,54390,54651,54789,54791,54857,54953,55372,55524,55560,55588,55597,55637,55646,55869,55884,56040,56045,56149,56271,56276,56489,56580,56661,56786,56792,56819,56850,56859,57424,57425,57440,57591,57792,57799,57831,57868,57931,58011,58241,58260,58583,58603,58607,58650,58695,58703,58847,59128,59219,59322,59458,59488,59592,59621,59636,59753,59782,59856,59909,59996,60054,60084,60143,60163,60172,60191,60200,60288,60401,60424,60486,60501,60667,60793,60828,60901,60983,61019,61125,61508,61571,61669,61671,61988,62068,62096,62115,62272,62281,62290,62569,62620,62722,62816,62959,63125,63157,63267,63324,63367,63444,63483,63722,63779,63800,63822,63916,63937,64071,64191,64305,64329,64460,64566,64675,64794,64844,64972,65217,65258,65278,65319,65333,65428,65535,65555,65588,65716,65743,65778,66016,66041,66100,66219,66248,66299,66309,66542,66699,66755,66916,66944,67048,67123,67206,67229,67379,67380,67429,67469,67474,67619,67620,67852,68032,68060,68180,68188,68366,68479,68552,68555,68584,68590,68934,68984,69076,69200,69224,69381,69407,69417,69658,69767,70180,70422,70464,70476,70513,70533,70589,70646,70674,70767,70836,71002,71013,71149,71220,71260,71309,71385,71438,71491,71673,71755,71782,71789,71800,71842,71922,71935,72127,72197,72266,72320,72337,72388,72401,72413,73037,73121,73219,73317,73346,73371,73446,73476,73560,73636,73701,73786,73946,73990,74064,74073,74102,74246,74474,74569,74644,74823,74952,75009,75049,75122,75197,75338,75356,75502,75558,75611,75705,75828,75864,75946,76198,76261,76309,76437,76481,76541,76546,76551,76607,76626,76647,76678,77019,77040,77043,77102,77192,77224,77277,77324,77330,77331,77515,77676,77766,77854,77992,78078,78113,78117,78279,78310,78485,78537,78605,78629,78635,78680,78703,78753,78833,78943,78977,79009,79036,79120,79232,79234,79340,79614,79677,79754,79790,79835,79951,79984,80025,80093,80287,80312,80366,80598,80806,80826,81029,81204,81279,81301,81338,81425,81476,81522,81524,81559,81601,82089,82233,82314,82427,82641,82733,82776,82929,82980,83003,83183,83214,83312,83331,83438,83472,83480,83531,83570,83583,83672,83698,84013,84084,84243,84252,84256,84460,84545,84548,84595,84695,84705,84867,84987,85112,85176,85312,85316,85351,85366,85446,85447,85508,85514,85522,85568,85668,85721,85800,85818,85897,85980,86043,86068,86105,86113,86144,86244,86357,86434,86452,86501,86537,86541,86551,86559,86588,86620,86678,86699,86834,86900,86907,86939,86980,87024,87097,87199,87432,87456,87892,87959,88069,88117,88488,88519,88520,88694,88705,88811,88857,89051,89110,89165,89275,89457,89664,89703,89710,89769,89812,90148,90210,90323,90426,90492,90523,90647,90762,90771,90801,90813,90821,90823,90824,90872,90895,90917,90922,90980,91105,91106,91145,91230,91399,91474,91480,91669,91681,91687,91756,91786,91833,91900,91928,91996,92010,92132,92255,92376,92487,92531,92552,92572,92586,92659,92899,92900,93026,93157,93177,93482,93489,93493,93574,93952,94054,94142,94201,94216,94225,94356,94359,94383,94509,94537,94625,94702,94724,95014,95062,95225,95230,95370,95448,95622,95712,95760,95790,95800,95894,96168,96210,96314,96357,96358,96400,96519,96527,96574,96593,96610,96718,96878,96922,96934,97026,97258,97296,97336,97468,97527,97574,97604,97606,97641,97649,97686,97894,97920,98032,98049,98050,98053,98084,98149,98199,98215,98270,98351,98362,98363,98376,98530,98555,98569,98601,98728,98974,99124,99189,99202,99286,99338,99786,99911,99924",
      "33,243,260,437,501,533,538,632,679,694,703,894,896,901,949,954,959,1055,1089,1092,1172,1178,1179,1355,1402,1533,1534,1567,1885,1887,1892,2202,2208,2406,2417,2419,3218,3289,3290,3297,3298,3360,3434,3575,3596,3735,3886,4180,4240,4608,4732,4913,5012,5087,5106,5121,5128,5185,5204,5235,5290,5344,5346,5356,5365,5378,5550,5554,5599,5767,5797,5855,5879,5909,6034,6147,6163,6184,6230,6280,6426,6475,6478,6512,6631,6700,6807,6895,6933,7182,7419,7713,7873,7895,8025,8054,8085,8186,8228,8351,8383,8451,8487,8594,8610,8718,8763,8805,8851,8877,8888,8978,8993,9057,9156,9164,9254,9347,9371,9515,9544,9549,9876,9907,9913,9931,9940,9974,10013,10018,10268,10278,10294,10410,10414,10442,10682,10683,10739,10891,10946,11151,11153,11239,11471,11811,11905,12019,12079,12155,12304,12461,12513,12805,12817,12935,12990,13252,13351,13458,13523,13629,13853,13999,14006,14040,14072,14134,14150,14324,14365,14438,14513,14684,14709,14710,14749,14780,14812,14970,15087,15452,15463,15504,15580,15606,15655,15769,15972,16085,16332,16493,16591,16684,16999,17098,17227,17358,17836,18201,18222,18273,18287,18390,18619,18653,18681,18771,18849,19071,19173,19217,19454,19492,19527,19568,19746,19793,19929,20154,20244,20252,20259,20328,20334,20579,20819,20906,20960,21118,21698,21722,21807,22029,22133,22415,22545,22586,22627,22763,22837,23011,23226,23308,23326,23433,23694,23742,23840,23843,24335,24345,24379,24405,24715,24934,24944,24954,25265,25316,25331,25372,25467,25572,25723,25759,25800,26346,26650,26708,26959,27032,27043,27131,27171,27251,27298,27413,27450,27482,27592,27622,27752,27793,27888,27915,27985,28276,28387,28402,28730,28742,28755,28830,28891,28927,28963,29046,29122,29183,29222,29274,29450,29505,29544,29548,29623,29651,29751,30162,30394,30549,30633,30778,30829,30836,31010,31303,31438,31445,32460,32517,33330,33565,33736,33862,34055,34841,34986,35043,35053,35106,35623,35823,35890,35936,36000,36006,36119,36305,36487,36513,36572,36627,36718,36758,36875,36994,37013,37075,37220,37345,37543,37559,37873,37986,38117,38192,38301,38371,38572,38628,38780,38798,38813,38922,38972,39099,39189,39220,39289,39299,39420,39547,39633,39735,39744,39813,39864,40159,40449,40451,40855,41297,41355,41466,42553,42844,42850,43023,43489,43586,43709,43726,43790,44022,44062,44116,44233,44592,44618,44734,44762,44781,44869,45016,45044,45097,45137,45242,45554,45639,45745,46016,46109,46632,46695,46968,47117,47224,47242,47523,48274,48302,48512,48656,49060,49223,49237,49577,49777,50000,50012,50086,50189,50333,50414,50440,50514,50521,50536,50597,50681,50692,50707,50815,50831,51045,51133,51259,51292,51330,51483,51549,51662,51864,51999,52046,52226,52286,52317,52374,52377,52556,52641,52806,52851,52908,52916,52954,53052,53059,53156,53165,53268,53307,53420,53549,53627,53628,53681,53693,53695,53703,53866,54008,54163,54349,54405,54438,54518,54674,54852,54869,54905,55027,55038,55050,55127,55225,55232,55317,55328,55474,55487,55511,55530,55542,55761,55779,55822,55890,56046,56332,56562,56676,56677,56739,56764,56889,56990,57055,57108,57155,57235,57257,57270,57368,57550,57661,57828,57881,58140,58477,58768,58823,58871,58875,58891,58912,58950,59139,59394,59438,59532,59572,59639,59704,59873,59907,59934,60138,60145,60222,60233,60284,60361,60417,60461,60534,60565,60609,60622,60802,60827,60856,60859,60880,60947,60955,61036,61040,61208,61227,61282,61329,61472,61500,61531,61554,61564,61676,61731,61820,61930,61955,62137,62285,62339,62382,62392,62429,62431,62550,62582,62853,62970,63027,63051,63062,63083,63159,63189,63217,63247,63252,63254,63327,63430,63434,63466,63516,63584,63657,63668,63679,63778,63837,63853,63867,63948,63956,64006,64147,64151,64287,64373,64487,64537,64560,64823,64879,64943,65123,65206,65430,65632,65801,65807,65810,65933,65999,66022,66065,66249,66381,66528,66621,66814,66820,66868,66937,67027,67228,67307,67326,67387,67528,67559,67616,67671,67755,67800,67929,67952,68016,68036,68068,68338,68402,68426,68621,68652,68666,68712,68769,68879,68894,68900,69028,69103,69180,69190,69269,69419,69465,69509,69516,69556,69557,69609,69675,69780,69799,69846,69868,69957,69980,70015,70124,70472,70759,70839,70862,70901,71123,71151,71306,71399,71464,71676,71916,71917,72059,72179,72314,72347,72354,72427,72508,72532,72664,72710,72730,73216,73296,73327,73392,73412,73424,73427,73590,73918,73922,73968,74092,74195,74204,74274,74290,74353,74498,74588,74592,74676,74855,74866,75189,75267,75343,75358,75403,75436,75466,75547,75731,75838,75900,75923,75982,76011,76062,76069,76103,76104,76248,76255,76275,76293,76295,76356,76499,76515,76597,76757,76885,76985,76994,77088,77161,77214,77218,77276,77317,77349,77752,77761,77982,78094,78115,78391,78467,78483,78536,78538,78800,78811,78852,78880,79244,79266,79311,79451,79485,79649,80024,80052,80128,80246,80257,80557,80579,80641,80657,80715,80739,80811,80872,80979,80989,81002,81096,81108,81120,81228,81494,81596,81755,81763,81935,82011,82028,82197,82305,82348,82433,82611,82650,82657,82710,82788,83134,83137,83187,83220,83410,83428,83489,83525,83595,83778,83876,83959,83962,83994,84011,84087,84140,84249,84334,84357,84598,84602,84625,84668,84678,84790,84848,84997,85061,85114,85159,85161,85179,85231,85284,85392,85490,85600,85810,85958,85963,86159,86165,86234,86316,86422,86465,86484,86485,86516,86596,86731,86775,86847,87013,87100,87124,87175,87223,87346,87389,87484,87597,87707,87814,87964,87991,88017,88047,88153,88398,88412,88564,88568,88627,88644,89020,89103,89108,89199,89346,89360,89419,89436,89604,89627,89717,89790,89921,89949,89992,90016,90143,90147,90195,90276,90317,90364,90469,90586,90610,90635,90792,90803,91020,91088,91100,91114,91137,91214,91215,91333,91366,91515,91553,91663,91929,92195,92261,92287,92320,92411,92436,92494,92611,92677,92905,92910,92931,92958,92970,92998,93066,93181,93206,93217,93352,93358,93414,93450,93452,93490,93559,93646,93695,93702,93719,93720,93740,93742,93800,93853,93866,93909,94001,94053,94161,94165,94314,94457,94461,94483,94792,94944,95075,95154,95229,95327,95330,95475,95649,95905,95954,96037,96239,96246,96361,96409,96509,96597,96618,96637,96691,96693,96694,96794,96910,96952,97014,97017,97157,97184,97311,97413,97488,97507,97591,97733,97761,97878,97889,98090,98111,98112,98275,98285,98371,98520,98553,98600,98629,98652,98694,98732,98868,98872,98952,98985,99145,99259,99337,99400,99424,99431,99559,99771,99801,99855,99889,99891,99917,99985",
      "81,113,153,236,246,1233,1981,2058,2096,2135,2136,2204,2300,2412,2452,2587,2809,2879,3007,3054,3190,3204,3228,3291,3683,4052,4215,4441,4524,4548,4715,4801,4809,5684,6350,6638,6655,6814,6997,7185,7205,7255,7529,7548,7830,7945,8002,8004,8087,8118,8128,8137,8156,8175,8232,8270,8294,8319,8346,8416,8424,8650,8662,8669,8840,8891,8897,8975,9039,9100,9134,9201,9238,9296,9397,9480,9619,9657,9695,9728,9753,9759,9828,9847,9942,9965,9986,10075,10081,10094,10113,10208,10236,10303,10360,10411,10455,10521,10588,10645,10656,10763,10835,10930,11036,11158,11216,11234,11285,11291,11386,11413,11480,11575,11614,11652,11709,11918,11938,12034,12097,12203,12222,12241,12384,12388,12393,12450,12479,12488,12602,12698,12716,12724,12735,12794,12849,12887,12963,13016,13020,13039,13172,13192,13267,13343,13419,13436,13534,13590,13609,13647,13856,13924,13932,13989,14008,14071,14121,14179,14216,14217,14236,14255,14274,14383,14388,14424,14440,14521,14540,14616,14619,14635,14691,14703,14787,14920,15015,15039,15053,15091,15121,15148,15262,15281,15300,15357,15369,15396,15471,15475,15566,15604,15623,15642,15661,15694,15737,15756,15832,15851,15870,15889,15984,16003,16098,16193,16250,16288,16326,16345,16383,16402,16440,16442,16450,16459,16497,16592,16649,16668,16706,16782,16839,16877,16896,16915,16991,17067,17141,17253,17257,17334,17352,17447,17466,17485,17580,17637,17694,17713,17732,17789,17827,17846,17884,17922,18017,18036,18131,18219,18225,18226,18245,18283,18302,18359,18378,18397,18530,18568,18606,18663,18682,18758,18796,18815,18872,18910,18929,19176,19237,19252,19271,19309,19347,19385,19442,19499,19588,19605,19651,19727,19765,19860,19879,19936,19955,20012,20031,20107,20115,20145,20152,20164,20240,20273,20373,20411,20430,20449,20468,20487,20525,20639,20753,20810,20821,20829,20843,20848,20886,20981,20998,21095,21171,21190,21293,21304,21323,21342,21456,21475,21494,21570,21589,21627,21665,21684,21912,21969,22045,22083,22121,22178,22216,22254,22273,22311,22349,22368,22406,22539,22558,22599,22615,22634,22653,22672,22729,22748,22783,22786,22881,22900,22919,22921,22957,22993,23014,23071,23108,23147,23166,23185,23261,23337,23348,23356,23375,23413,23432,23489,23546,23603,23641,23660,23679,23755,23774,23869,23888,23926,23945,23964,23970,23983,24002,24021,24040,24116,24135,24265,24268,24439,24572,24629,24648,24705,24743,24800,24805,24819,24821,24935,24971,24990,25104,25105,25123,25180,25199,25332,25369,25427,25465,25471,25484,25560,25579,25606,25626,25712,25769,25788,25789,25807,25902,25921,25997,26016,26073,26092,26213,26235,26370,26472,26510,26529,26535,26649,26871,27023,27270,27480,27840,27846,28068,28179,28285,28309,28333,28410,28448,28562,28600,28733,28771,28773,28901,28922,29051,29073,29170,29189,29398,29458,29648,29816,29875,29949,29987,30063,30121,30179,30208,30237,30295,30353,30382,30411,30440,30469,30527,30556,30643,30672,30701,30730,30846,30933,30991,31020,31049,31107,31136,31165,31194,31223,31252,31281,31484,31513,31542,31571,31600,31629,31658,31745,31803,31832,31919,31977,32035,32151,32267,32383,32412,32470,32528,32557,32644,32673,32731,32818,32847,32934,32992,33021,33137,33166,33195,33224,33253,33311,33340,33369,33427,33514,33543,33601,33630,33688,33775,33804,33833,33891,33949,34065,34094,34123,34181,34210,34239,34268,34297,34355,34384,34413,34442,34471,34529,34587,34703,34732,34761,34790,34848,34906,34935,34964,35005,35031,35047,35056,35067,35094,35168,35241,35331,35365,35430,35436,35476,35517,35524,35642,35774,35815,35833,35904,35963,36022,36085,36145,36236,36275,36334,36393,36428,36498,36583,36706,36755,36800,36954,37026,37085,37267,37294,37326,37364,37515,37587,37646,37705,37769,37887,37958,37988,38017,38148,38183,38241,38389,38432,38479,38566,38637,38709,38768,38827,38891,39009,39129,39134,39198,39227,39233,39270,39369,39388,39452,39511,39570,39700,39831,39890,39978,40227,40285,40314,40343,40372,40430,40488,40546,40575,40604,40633,40691,40749,40807,40894,40952,41068,41126,41155,41213,41271,41329,41387,41416,41445,41474,41503,41532,41846,42032,42119,42159,42244,42253,42295,42379,42419,42458,42532,42581,42735,42743,42790,42799,42852,42893,43004,43055,43084,43177,43207,43208,43217,43282,43311,43319,43494,43501,43604,43623,43673,43691,43766,43841,43870,43902,43928,44003,44078,44122,44207,44217,44220,44240,44267,44276,44315,44402,44464,44639,44768,44789,44858,44876,44951,44980,44994,45026,45099,45147,45208,45266,45339,45396,45454,45478,45526,45583,45641,45770,45828,45900,46017,46039,46092,46136,46163,46185,46193,46208,46292,46423,46588,46697,46715,46733,46832,46884,46930,47106,47253,47258,47276,47370,47383,47389,47396,47433,47509,47546,47573,47644,47740,47781,47854,47901,48013,48043,48077,48148,48155,48335,48423,48504,48596,48623,48692,48702,48707,48752,48844,48885,48971,49038,49064,49068,49116,49136,49159,49187,49199,49217,49402,49457,49491,49509,49516,49548,49699,49718,49804,49805,49825,49867,49884,49886,49913,49926,49965,49994,50123,50217,50234,50249,50257,50267,50328,50469,50495,50545,50638,50683,50701,50712,50793,50808,50849,50987,51092,51127,51240,51325,51496,51621,51697,51746,51781,51800,51863,51914,51956,51975,52025,52234,52298,52418,52468,52472,52479,52898,52917,53041,53090,53179,53344,53376,53412,53422,53471,53478,53525,53585,53655,53696,53709,53749,53851,53856,53944,53967,54004,54091,54268,54435,54465,54520,54568,54576,54813,54870,54900,54924,55073,55178,55218,55430,55497,55598,55756,56052,56097,56116,56155,56204,56214,56254,56333,56357,56419,56430,56454,56547,56659,56678,56698,56717,56934,56998,57124,57159,57240,57248,57289,57354,57480,57577,57668,57750,57769,57872,57979,58112,58279,58643,58669,58722,58910,59013,59029,59046,59087,59167,59242,59342,59385,59442,59464,59640,59682,59779,59841,59843,59878,59914,59953,60193,60261,60302,60444,60581,60693,60703,60939,60946,60950,61064,61133,61167,61172,61400,61469,61525,61624,61689,62045,62130,62389,62424,62430,62546,62560,62666,62680,62846,62848,63003,63421,63443,63572,63588,63614,63662,63798,63846,63893,63993,64036,64063,64149,64156,64217,64365,64392,64406,64510,64552,64628,64661,64672,64699,64734,64739,64834,64855,64882,65006,65025,65170,65178,65179,65234,65334,65349,65432,65617,65629,65710,65833,65837,65839,65888,65928,65957,66003,66378,66592,66674,66703,66788,67042,67075,67163,67188,67197,67314,67348,67433,67466,67489,67730,67955,68010,68109,68327,68403,68427,68850,69050,69121,69338,69551,69575,69733,69885,70061,70130,70214,70308,70313,70323,70445,70491,70610,70619,70712,70742,70800,70887,70895,70904,71093,71219,71299,71522,71523,71637,71674,71802,71836,71924,72005,72082,72087,72095,72096,72203,72290,72302,72381,72462,72478,72562,72595,72644,72673,72722,72747,72846,73036,73128,73179,73197,73269,73316,73328,73387,73395,73453,73467,73589,73615,73847,73890,73950,73958,73995,74165,74181,74194,74293,74304,74310,74376,74384,74507,74522,74573,74653,74799,74842,75073,75159,75241,75318,75321,75362,75444,75454,75504,75638,75780,75806,75992,75995,76031,76035,76058,76100,76108,76161,76224,76298,76330,76355,76447,76629,76691,76826,76845,76862,76917,76931,77014,77119,77184,77382,77444,77579,77654,77669,77718,77953,77957,78105,78124,78162,78178,78236,78349,78411,78441,78504,78507,78556,78884,78979,78989,79094,79115,79129,79306,79309,79335,79360,79407,79524,79603,79760,80243,80244,80269,80286,80354,80537,80567,80656,80704,80716,80791,80827,81062,81067,81090,81103,81232,81620,81668,81726,81853,81974,82115,82130,82177,82362,82398,82566,82661,82721,82812,82834,82865,82913,83000,83093,83133,83141,83359,83377,83477,83544,83614,83615,83715,83818,84153,84168,84193,84383,84465,84511,84564,84612,84624,84627,84945,85169,85191,85220,85340,85356,85367,85385,85388,85510,85517,85555,85717,85843,85857,85896,85985,85995,86011,86040,86061,86106,86136,86176,86369,86385,86512,86577,86583,86741,86928,86978,87043,87110,87132,87152,87260,87280,87325,87329,87433,87467,87479,87500,87566,87660,87687,87692,87701,87846,87922,88095,88316,88396,88465,88477,88536,88788,88826,88885,88989,89005,89187,89370,89538,89670,89691,89916,90023,90138,90146,90166,90237,90290,90314,90315,90501,90507,90508,90749,90965,90971,90996,91049,91135,91300,91414,91456,91596,91636,91661,91686,91729,91846,91926,91987,91989,91999,92085,92156,92189,92218,92346,92418,92496,92756,92826,92921,92953,93020,93074,93099,93187,93247,93293,93350,93479,93548,93679,93840,93851,94588,94592,94611,94662,94710,94711,94767,94779,94967,95037,95249,95335,95508,95598,95682,95743,95758,95818,96000,96050,96075,96112,96142,96247,96362,96365,96387,96445,96525,96548,96566,96587,96651,96800,96891,96907,96939,96941,97030,97042,97074,97117,97156,97217,97276,97283,97300,97340,97411,97729,97765,97871,98069,98241,98294,98300,98310,98577,98837,98957,98983,99005,99067,99116,99121,99183,99437,99452,99500,99526,99607,99642,99683,99710,99733,99745,99770,99822",
      "96,110,134,198,204,323,1654,1675,1786,1940,1953,1976,1978,2007,2008,2022,2050,2057,2069,2126,2140,2154,2176,2178,2180,2186,2210,2257,2267,2313,2329,2336,2342,2354,2362,2378,2405,2411,2430,2432,2435,2436,2437,2462,2466,2469,2505,2514,2562,2571,2583,2585,2621,2631,2642,2651,2672,2677,2698,2754,2757,2783,2786,2805,2822,2827,2862,2914,2929,2944,2982,3034,3049,3086,3102,3105,3118,3120,3134,3156,3164,3180,3184,3192,3249,3250,3264,3274,3321,3342,3368,3380,3433,3452,3462,3477,3485,3487,3493,3496,3503,3510,3526,3532,3560,3562,3597,3603,3637,3656,3677,3686,3690,3694,3720,3724,3744,3784,3843,3853,3883,3906,3918,3924,3932,3951,3954,3956,3961,3967,3978,4023,4079,4092,4107,4109,4125,4149,4172,4176,4182,4187,4197,4203,4221,4226,4253,4267,4299,4312,4396,4412,4442,4459,4570,4571,4584,4599,4605,4607,4616,4660,4663,4668,4670,4702,4735,4743,4760,4811,4844,4870,4886,4963,4981,4982,6253,6257,6262,6381,6382,6389,6435,6546,6579,6583,6599,6640,6787,6833,6837,6859,6901,6967,7125,7167,7172,7218,7274,7388,7487,7502,7607,7633,7645,7690,7697,7788,7960,7984,8003,8097,8112,8248,8295,8296,8412,8446,8511,8572,8590,8621,8626,8645,8655,8665,8666,8688,8690,8779,8873,8893,8923,8940,8958,9006,9016,9027,9054,9062,9076,9145,9253,9279,9471,9507,9578,9583,9752,9777,9858,9859,9860,9949,10088,10099,10181,10217,10372,10403,10516,10536,10552,10789,10831,10864,11058,11101,11155,11269,11426,11450,11476,11538,11701,11718,11734,11763,11914,11991,12018,12044,12134,12281,12289,12324,12331,12522,12575,12780,12855,13024,13179,13199,13202,13283,13317,13389,13395,13405,13411,13478,13519,13535,13695,13720,13737,13780,13901,13936,13965,14265,14322,14341,14739,14781,14782,15168,15233,15524,15569,15761,15818,16114,16371,16502,16767,16774,16844,16918,16924,17078,17113,17335,17573,17625,17818,17866,17880,18289,18357,18411,18442,18499,19046,19243,19389,19666,19887,19909,19913,20413,20437,20537,20815,21195,21237,21392,21486,21978,22059,22244,22288,22320,22474,22486,22517,22617,22686,23102,23182,23230,23471,23866,23916,24093,24205,24251,24466,24518,24603,24827,25127,25188,25213,25306,25546,25591,25745,26029,26053,26250,26284,26336,26381,26449,26508,26516,26518,26522,26525,26551,26558,26564,26576,26583,26584,26586,26609,26612,26698,26764,26785,26801,26829,26881,26910,26951,26966,27049,27086,27089,27115,27148,27192,27232,27285,27292,27294,27337,27370,27405,27433,27443,27449,27455,27495,27498,27503,27522,27534,27536,27548,27567,27587,27598,27601,27602,27603,27605,27621,27626,27634,27650,27676,27692,27723,27797,27820,27827,27833,27866,27867,27895,27910,27950,27957,27979,27986,28008,28019,28027,28039,28041,28046,28060,28064,28097,28131,28134,28160,28165,28191,28212,28219,28231,28247,28282,28288,28305,28401,28421,28483,28520,28529,28532,28538,28560,28581,28598,28602,28613,28614,28620,28628,28657,28668,28680,28684,28706,28707,28720,28725,28726,28796,28821,28855,28861,28887,28916,28937,28953,28995,29013,29028,29038,29072,29074,29086,29091,29101,29103,29125,29147,29149,29150,29160,29220,29304,29315,29317,29355,29356,29360,29369,29374,29384,29393,29394,29410,29434,29443,29468,29476,29521,29564,29568,29595,29624,29637,29643,29649,29655,29673,29713,29787,29797,29865,29905,29909,30022,30051,30080,30109,30138,30167,30196,30225,30254,30283,30312,30341,30370,30399,30428,30457,30486,30515,30544,30573,30602,30631,30660,30689,30718,30747,30776,30805,30834,30863,30892,30921,30950,30979,31008,31037,31066,31095,31124,31153,31182,31211,31240,31269,31298,31327,31356,31385,31414,31443,31472,31501,31530,31559,31588,31617,31646,31675,31704,31733,31762,31791,31820,31849,31878,31907,31936,31965,31994,32023,32052,32081,32110,32139,32168,32197,32255,32284,32313,32342,32371,32400,32429,32458,32487,32516,32545,32574,32603,32632,32661,32690,32719,32748,32777,32806,32835,32864,32893,32922,32951,32980,33009,33038,33067,33096,33125,33154,33183,33212,33241,33270,33299,33328,33357,33386,33415,33444,33473,33502,33531,33560,33589,33618,33647,33676,33705,33763,33792,33821,33850,33879,33908,33937,33966,33995,34053,34082,34111,34140,34169,34198,34227,34256,34285,34314,34343,34372,34401,34430,34459,34488,34517,34546,34575,34604,34633,34662,34691,34720,34749,34778,34807,34836,34865,34894,34923,34952,34981,35022,35045,35111,35130,35134,35139,35185,35221,35223,35274,35283,35285,35287,35296,35313,35348,35422,35433,35464,35493,35541,35600,35659,35677,35732,35791,35850,35921,35945,35980,36039,36102,36134,36162,36194,36221,36252,36292,36351,36366,36410,36419,36482,36541,36566,36600,36608,36616,36617,36664,36723,36782,36853,36884,36912,36938,36971,37043,37102,37161,37170,37178,37225,37284,37343,37378,37414,37422,37473,37532,37604,37663,37722,37786,37815,37845,37876,37904,37975,38034,38093,38165,38224,38283,38347,38406,38418,38465,38493,38536,38553,38558,38595,38654,38671,38673,38726,38785,38844,38853,38859,38871,38908,38967,39026,39097,39156,39215,39287,39346,39405,39469,39528,39541,39546,39587,39658,39717,39741,39776,39848,39907,39966,39975,40128,40157,40186,40215,40244,40273,40302,40331,40360,40389,40418,40447,40476,40505,40534,40563,40592,40621,40650,40679,40708,40737,40766,40795,40824,40853,40882,40911,40940,40969,40998,41027,41056,41085,41114,41143,41201,41230,41259,41288,41317,41346,41375,41404,41433,41462,41491,41520,41549,42019,42048,42077,42106,42135,42144,42157,42195,42230,42285,42296,42326,42330,42360,42369,42384,42404,42435,42460,42522,42597,42634,42637,42672,42759,42778,42783,42834,42872,42909,42972,42996,43018,43030,43071,43146,43233,43308,43383,43412,43430,43470,43510,43514,43545,43559,43620,43630,43707,43753,43782,43857,43875,43944,43956,43988,44019,44025,44039,44094,44099,44153,44155,44181,44256,44295,44298,44331,44365,44418,44439,44493,44499,44502,44517,44519,44520,44568,44631,44655,44702,44703,44730,44742,44748,44761,44805,44827,44892,44967,45042,45115,45163,45194,45224,45282,45299,45355,45412,45470,45514,45542,45599,45657,45677,45729,45786,45844,45857,45916,45973,46010,46024,46038,46049,46107,46110,46135,46187,46213,46246,46248,46395,46463,46467,46473,46486,46547,46584,46585,46606,46615,46628,46635,46659,46721,46723,46759,46811,46813,46819,46849,46867,46962,47004,47165,47166,47218,47282,47341,47349,47369,47372,47379,47543,47588,47622,47636,47645,47650,47680,47703,47709,47722,47809,47844,47903,47912,47913,47931,48049,48058,48076,48078,48170,48180,48225,48227,48315,48339,48348,48433,48440,48455,48458,48484,48544,48551,48562,48576,48628,48629,48653,48751,48780,48830,48881,48934,48980,48981,48994,49006,49030,49084,49090,49095,49120,49125,49126,49208,49226,49356,49389,49461,49589,49597,49620,49635,49650,49693,49705,49730,49768,49776,49799,49818,49854,49864,49881,49919,49923,49946,49955,49958,49959,50036,50191,50241,50306,50331,50376,50404,50535,50573,50577,50677,50732,50779,50955,50965,51052,51137,51194,51392,51430,51435,51486,51584,51675,51680,51695,51801,51821,51888,51921,51934,51972,52008,52051,52328,52462,52464,52503,52585,52651,52678,52692,52700,52741,53037,53060,53111,53118,53207,53235,53345,53489,53561,53667,53685,53704,53755,53775,53803,53904,53912,53928,54071,54154,54585,54609,54785,54895,54979,55014,55045,55161,55193,55214,55256,55357,55405,55440,55576,55632,55643,55650,55654,55907,55998,56058,56382,56428,56529,56569,56631,56703,56801,56890,56925,57004,57184,57207,57336,57362,57393,57423,57426,57443,57491,57628,57637,57795,57912,57999,58022,58188,58382,58395,58479,58604,58677,58680,58759,58889,58939,58955,58964,59003,59092,59332,59388,59398,59431,59523,59722,59875,59881,59968,60224,60273,60293,60357,60365,60522,60649,60769,60850,60889,60930,61022,61105,61232,61380,61580,61632,61654,61657,61744,61805,61923,61931,61951,61954,61994,62020,62052,62108,62141,62249,62256,62262,62315,62356,62488,62650,62653,62716,62738,62842,62942,63020,63114,63127,63138,63274,63291,63334,63445,63456,63473,63857,64196,64231,64285,64344,64350,64379,64388,64393,64578,64602,64625,64741,64773,64836,64853,64956,65016,65134,65371,65376,65405,65448,65465,65491,65602,65683,65748,65760,65766,65772,65953,66158,66244,66298,66417,66472,66475,66566,66619,66635,66686,66698,66734,66773,66792,66797,66867,66878,66966,67078,67094,67130,67149,67255,67328,67356,67384,67490,67545,67836,67843,68174,68223,68280,68311,68332,68677,68721,68829,68847,68972,68976,69067,69147,69166,69183,69294,69313,69354,69369,69459,69513,69552,69579,69713,69759,69870,69905,69917,69944,69990,70021,70073,70104,70109,70191,70212,70244,70314,70521,70548,70856,70964,71072,71100,71125,71239,71251,71254,71272,71307,71375,71442,71541,71598,71611,71621,71723,71846,71884,72067,72112,72160,72350,72581,72815,72879,73094,73206,73214,73361,73617,73626,73692,73731,73742,73762,73784,73899,73944,74020,74099,74182,74258,74359,74368,74387,74440,74444,74621,74749,74860,74895,74941,75035,75081,75168,75270,75298,75376,75536,75566,75650,75753,75781,75834,75841,75852,75887,75911,75942,75955,76078,76089,76112,76168,76259,76277,76374,76592,76593,76603,76688,76747,76789,77003,77132,77135,77316,77690,77692,77703,77706,77910,77919,77948,78036,78174,78267,78338,78412,78622,78692,78810,78865,79071,79293,79345,79386,79392,79419,79506,79651,80265,80344,80614,80800,80929,80936,80949,81070,81190,81191,81330,81362,81376,81411,81482,81489,81514,81523,81539,81644,81669,81808,81864,81867,81916,81990,82093,82134,82146,82147,82258,82366,82370,82431,82656,82727,82822,82875,83098,83153,83167,83206,83215,83275,83302,83310,83322,83365,83383,83429,83485,83604,83605,83670,83693,83841,83926,84000,84053,84082,84130,84286,84439,84642,84662,84748,84786,84792,84834,84912,84992,85021,85308,85313,85349,85414,85416,85521,85884,85974,85998,86016,86021,86039,86115,86140,86245,86513,86601,86613,86662,86668,86710,86725,86885,87007,87008,87070,87184,87193,87221,87313,87426,87509,87550,87564,88030,88073,88251,88320,88660,88686,88727,88842,88844,88905,88924,88950,89106,89181,89233,89246,89392,89698,89718,89831,89950,89997,90020,90069,90135,90160,90236,90641,90671,90743,90936,90998,91141,91172,91180,91196,91232,91235,91324,91404,91413,91432,91441,91448,91486,91660,91664,92051,92055,92093,92178,92478,92530,92912,92977,93007,93058,93130,93228,93465,93469,93518,93638,93654,93673,93677,93784,93880,93969,94033,94038,94155,94188,94294,94322,94327,94469,94534,94544,94637,94665,94714,94800,94842,94897,94986,94988,95052,95096,95111,95122,95214,95224,95321,95332,95339,95465,95470,95850,96041,96181,96275,96294,96391,96633,96751,96808,96846,96926,96976,97019,97099,97243,97289,97317,97573,97631,97783,97787,97831,97834,97844,97938,97966,97988,98159,98312,98340,98409,98467,98558,98742,98921,98971,99089,99109,99268,99284,99516,99567,99612,99627,99654,99726,99840,99961"
    ]
  }
}
)";
const folly::dynamic kShardSelectionLargeJson{
    folly::parseJson(kShardSelectionLarge)};

constexpr folly::StringPiece kEagerShardSelectionHuge = R"(
{
  "children_type": "LatestRoute",
  "children_settings" : {},
  "pools": [
    {
      "pool": {
        "type": "Pool",
        "name": "kShardSelectionHuge.01",
        "servers": [
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312"
        ],
        "shards": [
          "509,610,640,644,648,659,820,907,972,1096,1148,1199,1377,1417,1420,1529,1565,1576,1587,1588,1589,1632,1713,1761,1830,1832,1860,1882,1942,2075,2093,2142,2246,2249,2253,2276,2361,2542,2596,2739,2812,2841,2865,2868,2916,2950,2973,3125,3334,3338,3379,3732,3752,3800,3816,4115,4124,4131,4360,4494,4606,4626,4772,4821,4822,4827,4877,5015,5046,5154,5177,5229,5296,5426,5467,5510,5536,5540,5542,5546,5647,5664,5665,5681,5757,5793,5800,5814,5828,5847,5853,5954,5976,6021,6023,6028,6064,6238,6370,6422,6510,6675,6745,7018,7148,7248,7261,7287,7309,7360,7401,7466,7472,7478,7499,7512,7937,7959,7991,8043,8204,8210,8214,8395,8399,8421,8490,8516,8627,8647,9017,9123,9297,9317,9319,9363,9401,9424,9433,9505,9541,9786,9897,9937,10128,10167,10169,10240,10243,10306,10491,10540,10692,10729,10798,10803,10881,11097,11160,11164,11344,11506,11536,11625,11693,11909,11994,12030,12049,12201,12277,12279,12321,12347,12545,12554,12634,12748,12773,12866,13159,13324,13417,13457,13462,13551,13683,13704,13832,13876,14025,14047,14160,14209,14219,14371,14405,14551,14614,14626,14678,15093,15122,15175,15226,15242,15500,15549,15579,15591,16364,16400,16536,16552,16571,16799,16843,16976,17071,17155,17310,17490,17622,17642,17664,17700,17903,17956,17981,18162,18416,18538,18782,18787,18888,18908,19119,19154,19161,19263,19488,19533,19715,20049,20094,20168,20344,20354,20559,20658,20707,20808,20827,21017,21169,21243,21306,21413,21581,21620,21649,21797,21845,22085,22251,22306,22493,22625,22674,22767,22822,22864,22908,22922,23201,23213,23449,23490,23766,24057,24108,24325,24363,24589,24646,24936,25028,25085,25089,25185,25216,25217,25262,25308,25406,25592,25605,25710,25797,26071,26139,26199,26242,26267,26394,26441,26563,26630,26718,26867,26882,27047,27073,27102,27315,27375,27464,27700,27808,27883,27984,28164,28459,28673,28747,28791,28814,28841,28847,28915,29030,29065,29110,29246,29249,29571,29847,29925,29928,29965,29971,30264,30960,31018,31453,31656,31783,31861,31948,32091,32102,32787,33262,33309,33572,33599,33842,34208,34326,34846,34991,35177,35689,35697,35708,35902,35950,36007,36136,36440,36447,36804,36823,37014,37111,37324,37357,37386,37562,37699,37940,38109,38116,38233,38504,38576,38806,38836,38898,39041,39058,39062,39253,39258,39264,39310,39321,39630,39667,39924,39938,39984,39987,39991,40317,40720,40781,40865,40984,41111,42092,42262,42344,43000,43169,43189,43251,43355,43387,43406,43454,43631,43948,43949,44016,44284,45323,45364,45386,45440,45818,45943,46219,46365,46801,46809,47178,47955,47979,48087,48112,48463,48858,49002,49558,49976,50005,50027,50371,50418,50437,50458,50556,50562,50734,50741,50872,50895,50925,50948,50988,51018,51134,51168,51460,51468,51543,51634,51649,51670,51688,51806,51822,51844,52056,52073,52174,52282,52433,52555,52600,52616,52677,52812,52877,52900,52983,53183,53252,53388,53463,53587,53721,53772,53915,53962,53991,54307,54345,54436,54466,54569,54881,54958,54974,55057,55258,55302,55345,55415,55496,55503,55659,55751,55820,55841,55880,55911,55935,56027,56047,56122,56223,56327,56366,56405,56635,56642,56756,56948,57002,57104,57112,57427,57566,57584,57589,57643,57669,57874,57980,58117,58157,58196,58308,58325,58503,58575,58829,58853,58876,58933,58947,59138,59145,59260,59275,59311,59346,59367,59390,59435,59505,59511,59527,59683,59743,59749,59794,59854,59879,59891,59977,59999,60017,60102,60106,60108,60169,60334,60379,60526,60847,60885,60959,60969,60988,60991,60996,61102,61174,61199,61252,61262,61383,61394,61408,61430,61442,61528,61664,61681,61815,61943,62055,62098,62129,62191,62351,62353,62362,62371,62422,62511,62516,62578,62652,62775,62877,62879,62935,62940,63010,63139,63150,63233,63315,63353,63377,63384,63418,63447,63490,63606,63705,63710,63913,63920,64109,64140,64188,64214,64240,64252,64288,64317,64444,64476,64553,64572,64595,64687,64760,64808,64887,64930,64933,65029,65210,65247,65450,65531,65575,65674,65695,65930,66051,66128,66223,66230,66377,66444,66527,66682,66784,66877,66980,67158,67250,67540,67579,67795,67828,67993,68030,68075,68209,68321,68397,68609,68645,68745,68784,68816,68930,69105,69140,69160,69208,69222,69441,69574,69594,69641,69663,69773,69798,69886,69955,69967,70027,70054,70133,70301,70351,70539,70662,70726,70844,70996,71039,71126,71209,71353,71406,71474,71501,71734,71754,71798,71804,72100,72117,72134,72159,72183,72206,72253,72275,72324,72335,72384,72573,72652,72691,72725,72884,72887,72941,73048,73120,73281,73372,73494,73533,73802,73855,73896,74309,74327,74447,74528,74583,74593,74699,74786,74890,74954,74972,74994,75015,75179,75183,75190,75537,75733,75760,75809,76002,76179,76337,76349,76361,76382,76475,76625,76771,76853,76922,76992,77059,77120,77165,77289,77401,77498,77583,77600,77617,77630,77648,77750,77847,77892,77898,77981,78048,78256,78290,78328,78336,78350,78407,78518,78546,78714,78734,78929,78963,78975,79112,79211,79246,79249,79263,79267,79434,79490,79491,79568,79585,79601,79618,79709,79871,79981,80184,80223,80271,80411,80584,80678,80844,80887,80914,80964,81071,81349,81358,81393,81434,81467,81521,81574,81611,81841,81928,81936,82038,82237,82254,82267,82311,82330,82356,82387,82479,82508,82542,82694,82707,82851,82885,82925,82940,83087,83391,83546,83642,83661,83676,83744,83824,83910,83922,84062,84172,84229,84234,84251,84533,84539,84553,84640,84699,84880,84946,85014,85047,85050,85056,85099,85213,85237,85243,85387,85486,85659,85687,85830,86055,86264,86272,86304,86333,86499,86552,86579,86894,86903,86927,86954,87112,87156,87237,87486,87572,87659,87684,87695,87736,87775,87776,87851,87882,88044,88102,88178,88219,88338,88612,88662,88802,88836,88847,88908,88938,88985,89183,89197,89262,89306,89418,89453,89549,89566,89591,89614,89707,89722,89744,89746,89756,89808,90095,90391,90423,90505,90538,90699,90729,90779,90825,90874,90876,91118,91370,91392,91521,91618,91679,91739,91877,91890,91892,91914,91923,92124,92210,92294,92373,92414,92619,92633,92822,92930,93014,93100,93193,93273,93338,93397,93408,93805,94083,94090,94258,94295,94371,94552,94604,94684,94745,94804,94915,94941,94981,95028,95067,95106,95130,95252,95301,95351,95352,95354,95489,95555,95667,95686,95722,95846,96089,96165,96258,96490,96554,96563,96660,96888,96919,96931,97039,97070,97171,97269,97320,97355,97358,97368,97371,97416,97638,97665,98065,98135,98225,98242,98267,98291,98488,98527,98545,98656,98712,98716,98851,98875,98907,98963,99003,99131,99146,99221,99249,99346,99482,99540,99561,99573,99667,99804,99823,99851,99949",
          "168,210,1972,2029,2043,2162,2179,2428,2439,2539,2579,2647,2664,2713,2736,3198,3277,3281,3301,3358,3377,3492,3530,3549,3563,3589,3659,3665,3722,3738,4005,4177,4189,4438,4619,4636,4648,4712,4763,4838,5719,6395,7340,7454,7516,9477,9478,9496,9553,9572,9591,9629,9648,9667,9686,9705,9743,9781,9800,9833,9857,10016,10028,10047,10123,10142,10522,10598,10681,10684,10702,10712,10731,10750,10769,10807,10826,10845,10901,10902,10978,11000,11035,11111,11130,11225,11263,11320,11358,11415,11421,11491,11586,11624,11643,11662,11852,11871,11928,11966,12023,12099,12118,12132,12137,12175,12213,12232,12251,12308,12327,12346,12418,12441,12492,12555,12579,12612,12631,12652,12691,12707,12729,12731,12745,12764,12818,12851,12859,12897,12934,12954,13011,13049,13068,13071,13125,13161,13163,13193,13220,13258,13277,13296,13329,13340,13448,13569,13581,13657,13676,13714,13733,13752,13771,13790,13809,13847,13866,13923,13942,14018,14037,14093,14094,14113,14148,14170,14208,14246,14284,14360,14369,14379,14464,14508,14516,14531,14561,14584,14647,14664,14687,14702,14721,14730,14753,14767,14785,14797,14816,14851,14892,14930,14949,14987,15044,15063,15101,15120,15177,15196,15234,15253,15310,15348,15367,15386,15424,15443,15462,15481,15519,15538,15557,15563,15633,15652,15665,15671,15709,15770,15823,15842,15876,15899,15918,15937,15956,15994,16027,16032,16042,16051,16070,16089,16146,16165,16222,16241,16279,16298,16317,16336,16355,16374,16393,16423,16431,16469,16488,16526,16545,16559,16583,16697,16716,16735,16761,16777,16792,16849,16906,16908,16944,17001,17020,17058,17077,17084,17115,17153,17172,17210,17248,17267,17286,17343,17362,17381,17400,17453,17456,17457,17495,17533,17552,17566,17609,17638,17645,17647,17704,17742,17761,17799,17837,17894,17913,17951,17970,18008,18027,18074,18103,18141,18160,18217,18236,18255,18312,18326,18331,18369,18388,18407,18426,18445,18464,18475,18502,18537,18540,18559,18578,18589,18597,18616,18673,18692,18730,18804,18806,18844,18875,18901,18920,18977,19015,19072,19091,19110,19129,19159,19162,19168,19205,19224,19262,19300,19338,19346,19357,19395,19414,19433,19452,19490,19509,19566,19585,19623,19642,19661,19680,19684,19699,19737,19764,19775,19794,19832,19851,19870,19889,19908,19927,19946,19965,19984,20079,20098,20117,20162,20212,20269,20307,20326,20345,20383,20402,20421,20440,20459,20478,20516,20534,20535,20554,20573,20592,20611,20630,20638,20687,20725,20782,20801,20805,20820,20839,20872,20877,20915,20934,20953,20955,20982,20991,21048,21105,21110,21124,21143,21162,21181,21238,21254,21276,21295,21324,21333,21352,21390,21428,21433,21447,21466,21485,21504,21523,21542,21580,21618,21637,21675,21686,21713,21732,21751,21770,21789,21846,21865,21884,21903,21922,21941,21960,21979,21998,22036,22055,22112,22169,22188,22245,22264,22283,22321,22359,22378,22397,22416,22419,22435,22473,22511,22530,22549,22587,22606,22644,22663,22673,22682,22696,22701,22720,22730,22739,22777,22806,22834,22842,22853,22891,22910,22948,22967,22994,23005,23024,23081,23119,23157,23214,23218,23233,23271,23290,23319,23328,23365,23385,23404,23442,23480,23518,23537,23594,23613,23651,23670,23708,23727,23746,23784,23822,23841,23879,23917,23936,23955,23974,23993,24012,24031,24050,24069,24084,24088,24148,24183,24202,24278,24297,24316,24373,24449,24468,24487,24506,24525,24544,24563,24582,24594,24620,24639,24640,24658,24732,24734,24753,24772,24810,24829,24867,24924,24943,24948,24981,25019,25038,25067,25080,25095,25114,25133,25179,25209,25247,25266,25274,25285,25304,25323,25399,25402,25456,25475,25494,25513,25532,25570,25623,25629,25671,25684,25722,25741,25760,25798,25834,25836,25912,25918,25931,25950,25969,25977,26007,26026,26045,26064,26083,26102,26121,26140,26178,26197,26209,26216,26273,26292,26368,26378,26425,26501,26539,26596,26653,26786,26805,26836,26848,26900,26976,27004,27017,27090,27099,27120,27166,27204,27223,27231,27261,27280,27451,27546,27648,27698,27701,27777,27784,27823,27891,27926,28059,28066,28154,28173,28192,28230,28240,28249,28268,28281,28317,28325,28363,28382,28458,28477,28515,28686,28800,28857,28866,28876,28895,28933,28952,29012,29142,29199,29247,29256,29275,29294,29370,29408,29427,29446,29454,29474,29522,29525,29582,29597,29696,29731,29738,29769,29777,29871,29883,29921,29941,29968,29993,29997,30025,30054,30083,30112,30141,30170,30199,30228,30257,30286,30315,30344,30373,30402,30431,30460,30489,30518,30547,30576,30605,30634,30663,30692,30721,30750,30779,30808,30837,30866,30895,30924,30953,30982,31011,31040,31069,31098,31127,31156,31185,31214,31243,31272,31301,31330,31359,31388,31417,31446,31475,31504,31533,31562,31591,31620,31649,31678,31707,31736,31765,31794,31823,31852,31881,31910,31939,31968,31997,32026,32055,32084,32113,32142,32171,32200,32229,32258,32287,32316,32345,32374,32403,32432,32461,32490,32519,32548,32577,32606,32635,32664,32693,32722,32751,32780,32809,32838,32867,32896,32925,32954,32983,33012,33041,33070,33099,33128,33157,33186,33215,33244,33273,33302,33331,33360,33389,33418,33447,33476,33505,33534,33563,33592,33621,33650,33679,33708,33737,33766,33795,33824,33853,33882,33911,33940,33969,33998,34027,34056,34085,34114,34143,34172,34201,34230,34259,34288,34317,34346,34375,34404,34433,34462,34491,34520,34549,34578,34607,34636,34665,34694,34723,34752,34781,34810,34839,34868,34897,34926,34955,34984,35025,35058,35114,35145,35188,35209,35222,35249,35277,35351,35384,35385,35392,35399,35425,35496,35544,35574,35603,35662,35669,35681,35735,35794,35853,35860,35924,35983,36042,36055,36105,36125,36165,36224,36248,36295,36354,36413,36485,36544,36603,36609,36667,36726,36737,36785,36790,36799,36856,36915,36930,36974,36985,37006,37046,37065,37105,37164,37171,37228,37287,37311,37346,37375,37417,37476,37535,37607,37666,37682,37700,37725,37740,37743,37789,37848,37907,37920,37978,37998,38037,38052,38096,38105,38168,38227,38249,38286,38291,38350,38369,38409,38415,38468,38539,38598,38657,38683,38729,38739,38788,38847,38854,38911,38930,38970,39029,39035,39044,39052,39100,39159,39218,39240,39290,39349,39408,39421,39472,39506,39531,39553,39590,39661,39720,39739,39779,39826,39851,39862,39910,39969,39997,40119,40148,40177,40206,40235,40264,40293,40322,40351,40380,40409,40438,40467,40496,40525,40554,40583,40612,40641,40670,40699,40728,40757,40786,40815,40844,40873,40902,40931,40960,40989,41018,41047,41076,41105,41134,41163,41192,41221,41250,41279,41308,41337,41366,41395,41424,41453,41482,41511,41540,42010,42039,42068,42097,42126,42186,42235,42249,42276,42314,42315,42351,42395,42426,42513,42570,42588,42608,42621,42663,42690,42750,42770,42825,42869,42900,42920,42963,42987,43007,43046,43062,43081,43114,43137,43224,43281,43299,43320,43374,43408,43426,43427,43461,43491,43512,43536,43571,43579,43611,43635,43649,43698,43721,43773,43805,43828,43848,43866,43915,43935,43992,43994,44010,44085,44106,44172,44191,44199,44231,44247,44270,44279,44287,44290,44322,44368,44389,44409,44484,44542,44559,44571,44600,44646,44697,44704,44721,44740,44752,44796,44833,44859,44864,44883,44924,44937,44958,44969,45033,45046,45069,45106,45119,45154,45165,45215,45273,45285,45327,45346,45371,45403,45461,45533,45590,45648,45701,45720,45777,45799,45804,45835,45907,45964,46019,46087,46137,46170,46242,46249,46254,46261,46267,46313,46326,46347,46355,46411,46462,46479,46495,46543,46579,46602,46622,46668,46673,46675,46682,46687,46765,46781,46851,46920,46966,46972,47014,47032,47058,47060,47064,47136,47154,47193,47196,47229,47246,47272,47278,47297,47333,47335,47422,47439,47450,47462,47473,47502,47505,47577,47578,47612,47651,47724,47820,47825,47831,47835,47853,47877,47882,47892,47973,48021,48031,48079,48081,48093,48143,48162,48167,48220,48243,48248,48251,48271,48281,48287,48343,48353,48390,48400,48418,48419,48422,48438,48461,48487,48549,48616,48694,48717,48758,48773,48815,48859,48861,48889,48893,48985,49041,49055,49107,49183,49234,49268,49283,49300,49336,49345,49355,49413,49476,49482,49496,49621,49624,49644,49658,49671,49721,49775,49778,49870,49883,49902,49925,49933,49977,49986,50019,50161,50167,50299,50300,50342,50365,50374,50471,50512,50520,50643,50662,50758,50930,51015,51055,51160,51221,51357,51360,51492,51631,51825,51867,51909,51910,51929,52061,52255,52358,52496,52505,52578,52766,52928,52942,52966,53000,53051,53134,53136,53139,53262,53323,53354,53394,53404,53543,53737,53758,53844,53955,54041,54108,54357,54396,54558,54622,54732,54741,54845,54950,55040,55199,55275,55320,55334,55360,55400,55527,55586,55613,55656,55711,55774,55830,55895,55920,55985,56282,56309,56610,56733,56771,56776,57118,57189,57198,57277,57313,57317,57387,57394,57486,57548,57602,57640,57707,57833,58035,58088,58118,58147,58343,58355,58360,58400,58449,58507,58597,58612,58641,58704,58779,58790,58888,59161,59165,59172,59216,59310,59429,59785,59793,59819,60067,60181,60616,60824,60831,60841,60990,61043,61348,61558,61698,61772,61883,61905,61922,61964,62030,62056,62152,62161,62397,62402,62421,62459,62477,62490,62604,62698,62802,62806,62830,62975,63050,63064,63079,63080,63236,63348,63378,63498,63609,63746,63836,64008,64169,64184,64187,64469,64504,64598,64770,64798,64804,64941,65079,65096,65115,65144,65149,65211,65272,65283,65287,65332,65536,65615,65627,65891,65939,65950,65951,66024,66127,66198,66218,66231,66318,66352,66364,66379,66478,66610,66632,66640,66691,66808,66884,66906,66915,66953,66977,66979,66984,67032,67128,67132,67140,67141,67297,67298,67310,67322,67530,67716,67839,67925,67962,67981,68245,68509,68539,68734,68804,68809,68835,68838,68854,68896,68903,68951,68974,69031,69097,69169,69175,69424,69436,69478,69616,69664,69704,69841,69851,69910,69971,69995,70029,70248,70270,70360,70456,70494,70495,70823,70866,70981,70985,71051,71225,71250,71266,71333,71493,71532,71633,71698,71704,71735,71773,71818,71902,71928,71999,72185,72219,72249,72366,72549,72597,72609,72626,72660,72727,72809,72896,72961,73033,73063,73525,73579,73778,73867,73876,74054,74171,74224,74283,74288,74370,74420,74501,74669,74821,74901,74910,74953,74975,74987,75024,75068,75107,75495,75587,75653,75686,75742,75793,75796,76005,76203,76270,76305,76331,76385,76537,76538,76618,76630,76707,76842,76857,76894,77045,77205,77222,77309,77340,77373,77385,77395,77417,77566,77604,77607,77709,77739,77798,77824,77838,77845,77871,77884,77897,78016,78037,78141,78227,78228,78565,78735,78785,78955,78985,79151,79185,79192,79264,79401,79444,79565,79733,79976,80029,80086,80217,80232,80402,80445,80464,80511,80530,80660,80776,80778,80837,81091,81168,81727,81735,81948,81996,82076,82165,82214,82299,82322,82651,82703,82767,82801,82815,83026,83071,83185,83197,83255,83334,83431,83442,83524,83617,83707,83723,83907,83925,83957,83964,83968,84038,84066,84119,84174,84264,84265,84369,84400,84467,84471,84572,84631,84651,84693,84852,84866,84877,84908,84937,85088,85272,85311,85371,85404,85411,85493,85503,85608,85625,85635,85678,85697,85725,85752,85794,85835,85927,86461,86580,86773,86910,86930,87019,87271,87312,87360,87373,87425,87563,87605,87645,87661,88002,88255,88291,88407,88508,88535,88701,88731,88767,88827,88831,88896,88918,88927,88995,89193,89203,89216,89249,89452,89463,89496,89759,89792,89810,89887,89975,90051,90142,90328,90350,90464,90602,90638,90661,91092,91096,91143,91321,91363,91536,91695,91868,91918,92148,92208,92383,92463,92475,92511,92629,92658,92689,92737,92821,92861,92907,92963,93119,93146,93229,93260,93288,93563,93597,93607,93770,93788,93799,93820,93885,93950,94066,94098,94129,94289,94316,94388,94678,94797,94844,94903,95064,95065,95100,95132,95218,95391,95432,95499,95517,95566,95679,95855,95891,95898,95902,95950,96026,96040,96043,96099,96147,96194,96265,96267,96385,96436,96516,96535,96695,96730,96849,96956,97102,97302,97357,97390,97543,97587,97672,97885,97968,98235,98493,98508,98549,98559,98584,98634,98643,98879,98919,98953,99090,99241,99264,99342,99426,99434,99589,99668,99674,99862,99905,99945",
          "22,170,384,620,904,1022,1070,1107,1426,1498,1618,1651,1701,1707,1820,1872,2064,2215,2221,2237,2271,2343,2518,2560,2779,2819,2848,3053,3123,3384,3428,3542,3561,3607,3860,3942,4331,4336,4371,4762,4938,5050,5149,5197,5210,5223,5252,5380,5445,5474,5495,5506,5600,5743,5753,5807,5916,5943,6002,6192,6305,6373,6409,6441,6485,6511,6542,6645,6665,6667,6683,6729,6758,6804,6808,6951,6996,7002,7162,7186,7199,7207,7424,7528,7538,7574,7606,7717,7950,7979,8066,8201,8495,8496,8507,8583,8596,8604,8814,8894,8953,8963,9033,9082,9093,9112,9129,9190,9283,9386,9463,9717,9751,9762,9841,9994,10000,10027,10199,10203,10327,10338,10353,10383,10432,10438,10484,10514,10515,10535,10602,10668,10766,10811,10868,10949,11583,11603,11666,11689,11706,11717,11777,11825,11935,12017,12192,12235,12372,12447,12567,12727,12867,12913,13070,13087,13104,13184,13257,13410,13453,13585,13586,13656,13684,13694,13751,13878,13879,13959,13991,14014,14046,14114,14154,14240,14244,14254,14321,14358,14688,14876,15128,15165,15181,15261,15351,15406,15521,15587,15717,15732,15785,15898,15908,15935,15941,15977,16074,16188,16192,16262,16287,16523,16666,16679,16687,16705,16772,16788,16979,17003,17027,17291,17369,17384,17535,17665,17752,18085,18143,18197,18371,18583,18728,18814,18853,18933,18969,18976,19017,19216,19220,19523,19549,19608,19647,20308,20480,20486,20634,20787,20838,20997,21034,21153,21161,21240,21369,21487,21690,21929,21989,22165,22329,22452,22483,22551,22578,22608,22623,22632,22660,22732,22833,22874,22905,22983,23098,23174,23399,23405,23459,23563,23622,23714,23785,23810,23952,23953,24020,24106,24121,24228,24314,24390,24552,24585,24795,24814,24849,24909,25001,25009,25042,25100,25155,25200,25244,25340,25648,25664,25669,25731,25761,25855,25987,26180,26476,26582,26625,26659,26725,26736,26751,26770,26784,26880,26933,27092,27098,27271,27328,27336,27398,27401,27416,27437,27653,27667,27798,27845,27939,28315,28328,28384,28399,28578,28597,28627,28688,28690,28759,28809,28870,28878,28946,29096,29123,29161,29178,29311,29328,29348,29432,29440,29464,29495,29518,29720,29750,29844,29866,30116,30191,30546,30554,30945,30952,30984,31424,31622,31880,31883,31908,32344,32440,32662,33107,33159,33277,33449,34199,34512,34808,34867,34899,35129,35190,35198,35505,35559,35640,35700,35743,35897,35916,35956,36011,36059,36104,36256,36265,36412,36601,36613,36630,36671,36692,36818,37009,37010,37024,37318,37358,37388,37448,37534,37611,37614,37680,37688,37818,37840,37847,37980,38008,38046,38147,38170,38195,38295,38321,38322,38387,38864,38868,38876,38885,38932,38933,38945,39116,39297,39317,39355,39446,39551,39812,39914,40075,40158,40377,41058,41278,41289,41996,42063,42437,43209,43271,43398,43410,43675,43764,43784,43847,43916,43917,44187,44212,44496,44772,44835,44943,45248,45306,45374,45383,45460,45501,45503,45557,45789,45806,45949,46180,46217,46421,46521,46780,47053,47071,47076,47162,47434,47566,47641,47676,47739,47744,47959,48253,48523,48574,48592,48687,48834,48997,49290,49423,49483,50006,50059,50101,50190,50704,50789,50822,50846,50888,51017,51050,51115,51131,51132,51135,51167,51225,51232,51241,51261,51374,51416,51475,51556,51610,51669,51922,51924,51959,51977,52075,52117,52127,52215,52429,52498,52587,52592,52598,52605,52708,52984,53034,53222,53244,53280,53407,53455,53469,53724,54067,54099,54203,54210,54411,54596,54675,54690,54731,54894,54913,54962,54990,54998,55048,55128,55182,55234,55252,55386,55408,55421,55490,55500,55599,55667,55816,55924,56001,56042,56050,56142,56153,56180,56206,56243,56297,56311,56535,56585,56660,56680,56690,57067,57115,57176,57197,57284,57302,57348,57402,57404,57506,57617,57704,57851,57942,58009,58052,58062,58064,58145,58190,58385,58630,58658,58711,58782,58930,58963,58986,59000,59032,59053,59201,59229,59233,59360,59430,59454,59516,59647,59662,59874,59880,60078,60081,60374,60415,60421,60457,60459,60495,60546,60611,60759,60871,60882,61044,61086,61093,61170,61213,61268,61592,61631,61684,61713,61754,61765,61875,61944,62022,62275,62304,62316,62447,62476,62480,62502,62554,62597,62749,62752,62807,62852,63023,63031,63052,63154,63174,63368,63497,63661,63697,63714,63767,63801,63831,63903,63919,64079,64318,64481,64574,64613,64782,64832,64888,64921,65127,65130,65330,65420,65471,65512,65528,65638,65649,65686,65739,65787,65871,66036,66150,66157,66180,66264,66404,66501,66536,66609,66645,67212,67346,67365,67372,67556,67571,67668,67775,67805,67827,67851,67967,68205,68225,68337,68343,68424,68494,68706,68748,68794,68937,68940,69055,69127,69246,69262,69268,69351,69352,69414,69446,69474,69532,69659,69670,69740,69843,69954,69977,70066,70070,70178,70355,70383,70572,70645,70680,70850,70922,71114,71208,71369,71434,71457,71514,71618,71701,71900,71906,71926,72086,72226,72235,72238,72298,72446,72610,72702,72907,73042,73143,73183,73288,73305,73469,73543,73684,73741,73844,73900,73943,73979,74006,74027,74113,74118,74148,74244,74328,74428,74545,74691,74803,74827,74861,75148,75163,75188,75240,75279,75309,75365,75440,75931,75993,76048,76252,76274,76350,76451,76487,76823,76914,76934,77021,77087,77113,77145,77237,77311,77494,77511,77517,77608,77615,77702,77780,77791,77864,77870,77887,77987,78031,78084,78091,78211,78293,78421,78438,78560,78570,78579,78607,78648,78711,78727,78877,79022,79031,79336,79344,79415,79450,79556,79575,79874,79970,80023,80227,80228,80240,80611,80630,80857,80904,80957,81055,81236,81258,81299,81462,81658,81664,81676,81733,81788,81802,82040,82081,82139,82280,82312,82336,82399,82615,82637,82775,82784,82896,82923,82938,83005,83011,83052,83240,83271,83361,83395,83463,83482,83492,83530,83705,83747,83905,84027,84101,84137,84339,84343,84584,84709,84773,84801,84970,85039,85152,85207,85234,85241,85258,85309,85314,85331,85400,85434,85533,85542,85557,85660,85743,85772,85783,85792,85840,85909,86062,86170,86424,86471,86600,86607,86850,86866,87075,87275,87380,87383,87419,87461,87477,87524,87584,87586,87598,87760,87817,87863,87887,87902,87995,88208,88235,88317,88549,88638,88642,88865,88890,88926,88957,89075,89166,89194,89238,89273,89384,89412,89826,89842,89865,89963,90072,90109,90133,90249,90371,90410,90583,90629,90657,90701,90758,90782,90793,90844,90853,91068,91131,91283,91297,91388,91405,91465,91581,91597,91797,91809,91907,92008,92197,92324,92403,92521,92548,92597,92607,92641,92645,92748,92800,92808,92890,92895,92975,93055,93075,93106,93224,93238,93314,93355,93356,93402,93591,93611,93637,93765,93835,94014,94082,94099,94140,94170,94260,94321,94334,94528,94820,94834,94880,95058,95268,95322,95337,95424,95553,95658,95793,95966,96034,96087,96115,96283,96545,96612,96677,96678,96819,97056,97063,97101,97127,97181,97213,97254,97278,97483,97595,97646,97662,97823,97832,97847,97868,97904,97921,97995,98127,98183,98197,98415,98459,98490,98546,98570,98596,98598,98707,98765,98889,98900,99044,99059,99064,99108,99127,99274,99501,99531,99835,99897,99923,99953",
          "17,19,30,520,548,552,556,571,636,645,664,677,681,689,706,750,785,816,818,859,920,981,1046,1120,1147,1155,1260,1274,1277,1305,1336,1338,1428,1469,1492,1539,1582,1585,1591,1696,1828,1842,1875,2392,4073,5038,5123,5173,5180,5192,5200,5222,5226,5242,5244,5280,5286,5291,5348,5364,5366,5406,5407,5419,5442,5473,5475,5500,5537,5624,5652,5688,5689,5709,5713,5728,5730,5810,5864,5922,5925,5979,6049,6063,6150,6155,6180,6214,6309,6319,6343,6349,6391,6453,6491,6577,6717,6726,6734,6748,6777,7097,7149,7179,7192,7335,7346,7364,7369,7408,7590,7685,7702,7885,7892,7961,7970,7987,8150,8170,8419,8449,8538,8576,8633,8685,8703,8719,8766,8863,8880,8912,9127,9187,9204,9287,9335,9941,10040,10163,10253,10305,10307,10334,10352,10408,10420,10460,10525,10613,10849,10974,10995,11177,11206,11690,11732,11754,11837,12125,14950,15352,16857,19493,22839,23086,23176,24275,24508,24514,24613,24641,24698,24808,25447,26802,28120,28232,29148,29239,29746,30720,31145,31606,35578,36490,38560,41667,42259,42670,42791,45164,45381,46505,48990,50017,50171,50200,50334,50481,50529,50571,50582,50721,50776,50783,50870,50899,50924,51187,51212,51273,51339,51439,51586,51592,51639,51750,51811,51820,51838,51878,52143,52275,52329,52393,52458,52593,52686,52721,52871,53019,53023,53127,53176,53197,53290,53304,53423,53541,53581,53769,53828,53993,54093,54188,54196,54266,54280,54779,55035,55044,55084,55148,55171,55276,55291,55387,55481,55548,55760,55770,55780,55821,55845,56049,56073,56077,56140,56257,56495,56534,56712,56719,56751,56769,56791,56929,56949,57077,57081,57083,57364,57717,57943,57954,57967,58002,58100,58109,58168,58234,58376,58468,58499,58500,58809,58813,58890,58959,59124,59180,59468,59479,59507,59526,59559,59681,60180,60206,60294,60344,60370,60496,60508,60529,60550,60629,60644,60873,61041,61047,61165,61185,61236,61247,61258,61275,61332,61358,61362,61366,61559,61683,61802,61819,61874,61885,61991,62113,62336,62377,62521,62558,62579,62612,62659,62673,62686,62815,62866,62874,62930,63034,63231,63360,63426,63579,63669,63706,63725,63816,63890,64010,64013,64076,64161,64176,64346,64408,64441,64500,64606,64662,64733,64769,64904,64926,64984,65175,65207,65356,65404,65414,65433,65560,65593,65658,65740,65901,65986,66053,66072,66089,66111,66147,66185,66189,66194,66208,66229,66315,66351,66485,66487,66597,66760,66922,66960,66989,67060,67087,67168,67305,67394,67538,67615,67706,67744,67754,67847,67982,68022,68081,68108,68111,68290,68466,68473,68607,68663,68704,68714,68963,68978,68979,69091,69302,69326,69330,69363,69430,69445,69477,69805,69861,69923,69929,70114,70267,70349,70419,70510,70586,70725,70790,70838,70966,71091,71107,71206,71326,71328,71431,71620,71730,71788,71937,71938,71970,72063,72187,72201,72221,72262,72271,72513,72571,72580,72748,72915,72916,72925,73113,73239,73329,73441,73442,73568,73631,73703,73794,73809,74008,74114,74172,74180,74193,74362,74382,74411,74643,74717,74754,74850,74868,74891,74900,74970,75041,75136,75153,75161,75357,75613,75896,76021,76022,76114,76396,76611,76646,76731,76746,76787,76861,76895,76904,77011,77041,77166,77198,77229,77384,77644,77831,78039,78111,78143,78152,78242,78303,78331,78419,78446,78465,78567,78619,78685,78708,78779,78813,78948,79147,79166,79207,79239,79286,79328,79353,79424,79446,79591,79623,79648,79765,79877,80138,80140,80146,80398,80440,80559,80694,80784,80792,81044,81089,81118,81167,81342,81402,81470,81504,81591,81661,81674,81745,81849,81901,81962,82027,82168,82279,82567,82601,82619,82731,82853,82933,83009,83051,83070,83073,83090,83273,83627,83636,83683,83774,83819,83870,83900,84042,84344,84398,84521,84767,84807,84943,84948,85036,85043,85101,85257,85306,85401,85663,85705,85706,85726,85785,85883,86028,86037,86202,86228,86324,86352,86371,86488,86542,86649,86733,86815,86821,86855,86990,87047,87359,87400,87404,87470,87512,87619,87666,87788,87807,87899,88053,88345,88365,88383,88393,88450,88693,88726,88951,88992,89021,89201,89204,89285,89390,89415,89489,89582,89653,89676,89749,89788,89834,89896,89998,90007,90037,90045,90102,90117,90155,90157,90220,90326,90457,90558,90577,90679,90714,90790,90830,91013,91052,91138,91237,91248,91346,91471,91547,91559,91656,91787,91887,91962,92318,92368,92618,92663,92675,92771,92847,92874,93095,93214,93218,93475,93697,93771,93790,93833,93900,94004,94041,94296,94331,94353,94393,94490,94532,94563,94979,95074,95269,95274,95276,95385,95490,95523,95631,95672,95796,95988,96107,96226,96245,96390,96428,96506,96513,96557,96665,96754,96851,96944,96959,96968,97041,97098,97227,97504,97586,97780,98091,98124,98134,98412,98444,98772,98890,99196,99240,99260,99324,99366,99479,99514,99519,99566,99756,99825,99960",
          "277,334,386,406,423,444,465,480,515,553,568,595,621,643,670,723,734,792,817,847,873,989,1021,1043,1050,1068,1141,1175,1327,1353,1363,1386,1404,1432,1437,1448,1477,1563,1636,1791,1833,1845,1903,1954,2943,4639,5144,5234,5254,5330,5338,5700,5849,5867,5918,6040,6135,6139,6154,6158,6197,6265,6410,6533,6813,6941,7121,7202,7216,7222,7229,7238,7339,7453,7571,7579,7644,7687,7743,7783,7787,7914,7992,8036,8072,8140,8190,8392,8519,8546,8635,8653,8682,8742,8885,9040,9065,9111,9121,9122,9275,9284,9351,9453,9862,9874,10184,10205,10241,10417,10424,10623,10657,10805,10907,10931,10948,11073,11090,11126,11336,11566,11632,11714,11769,11835,13724,13865,14089,14096,14372,14535,14611,14705,14802,14915,15330,15442,16130,16263,16599,17264,17436,17568,17588,17667,17966,17975,18122,18337,18594,19494,19681,20266,20296,20674,21050,21639,22612,22830,22850,22932,22984,24280,25132,25485,26422,26919,27025,27242,27491,27497,27791,28158,28273,29371,29405,29485,29733,29737,29996,30511,33930,33985,35844,36673,38746,39820,39905,39979,40677,41300,42336,43120,44121,45936,46999,47254,49604,49874,50058,50117,50128,50180,50193,50288,50317,50443,50497,50534,50631,50632,50744,50796,50863,50889,51047,51361,51421,51432,51444,51514,51551,51612,51623,51630,51636,51694,51754,51796,51845,52009,52119,52150,52269,52319,52347,52360,52404,52448,52461,52485,52535,52601,52606,52657,52730,52888,52913,52972,53146,53164,53232,53328,53334,53381,53466,53750,53814,54010,54018,54052,54255,54296,54420,54478,54719,54834,54877,54952,54985,55007,55394,55513,55525,55718,55746,55928,56230,56320,56321,56369,56436,56517,56553,56558,56602,56707,56740,56919,56966,57042,57072,57095,57173,57694,57748,57935,58108,58379,58406,58473,58631,58895,58919,58937,58970,59023,59295,59306,59319,59745,59832,59885,59902,60062,60083,60103,60171,60311,60351,60509,60582,60760,60846,60931,60934,60981,61057,61127,61266,61303,61305,61405,61586,61620,61668,61702,61882,61952,62261,62322,62324,62436,62539,62557,62573,62586,62646,62729,62750,62758,62810,62882,62911,63122,63151,63187,63209,63311,63355,63364,63591,64049,64245,64268,64340,64732,64751,64791,65034,65110,65163,65256,65359,65372,65384,65496,65561,65694,65782,65791,65879,65983,66011,66118,66263,66354,66445,66474,66504,66607,66652,66778,66933,66961,67133,67181,67232,67234,67237,67542,67577,67724,68002,68138,68301,68323,68367,68411,68485,68516,68580,68616,68634,68702,68860,68871,68874,68880,68881,68908,68946,68975,69032,69053,69061,69073,69178,69364,69565,69808,69907,69921,70019,70093,70161,70325,70531,70575,70635,70872,70882,70883,70970,71011,71087,71271,71286,71293,71443,71516,71577,71631,71669,71770,71882,71899,71953,72024,72193,72195,72211,72407,72444,72541,72592,72621,72731,72743,72905,72987,73116,73162,73196,73423,73725,73726,74392,74554,74680,74761,74928,74960,75000,75037,75144,75171,75175,75303,75395,75460,75499,75604,75691,75696,75713,75774,75821,75903,76088,76113,76143,76530,76580,76606,76676,76683,76754,76803,76863,76949,76960,77002,77016,77134,77238,77274,77338,77355,77433,77489,77528,77688,77785,77908,77946,78014,78186,78229,78259,78292,78422,78669,78763,78861,79016,79063,79089,79247,79339,79370,79406,79468,79504,79539,79580,79644,79645,79662,79784,79915,80020,80121,80176,80214,80296,80321,80415,80526,80573,80635,80692,80799,80851,81104,81175,81625,81635,81688,81721,81899,81913,81998,82006,82058,82162,82248,82323,82466,82505,82560,82723,82900,82965,83008,83110,83237,83330,83396,83404,83559,83593,83655,83679,83728,83783,83840,83986,84079,84197,84290,84291,84388,84390,84430,84603,84758,84925,85004,85073,85086,85300,85323,85394,85407,85455,85497,85674,85707,85781,85850,85863,85999,86022,86268,86490,86500,86591,86623,86712,86748,86798,86824,86870,86999,87045,87062,87102,87511,87580,87646,87748,87782,87870,87967,87986,88009,88159,88214,88318,88440,88518,88527,88637,88765,88833,88867,88964,88996,89054,89084,89161,89357,89367,89395,89417,89550,89587,89638,89986,89994,90017,90080,90233,90310,90432,90465,90531,90715,90849,91161,91250,91312,91323,91337,91418,91601,91749,91853,91871,91888,92021,92084,92103,92193,92229,92367,92429,92434,92455,92491,92500,92638,92721,92738,92829,92854,92926,92966,93056,93097,93213,93281,93301,93382,93423,93500,93506,93570,93696,93704,93755,93813,93920,94241,94343,94557,94573,94837,94882,94969,95005,95009,95039,95071,95089,95109,95137,95161,95190,95407,95497,95506,95619,95627,95719,95729,95778,95808,95899,95973,95976,95994,95997,96071,96130,96244,96263,96284,96301,96348,96402,96406,96434,96671,96817,96909,97179,97239,97353,97386,97414,97477,97618,97679,97950,98043,98051,98054,98092,98130,98168,98253,98358,98457,98469,98703,98777,98869,98884,99097,99099,99306,99329,99350,99352,99374,99448,99657,99706,99820,99879,99900",
          "297,338,341,356,364,486,559,584,585,615,638,656,711,811,890,957,998,1083,1105,1111,1156,1224,1262,1290,1470,1521,1616,1727,1756,1856,1861,2095,2111,2193,2228,2441,2559,2797,2801,2835,2939,2949,3017,3182,3276,3312,3340,3343,3495,3602,3620,3733,3766,3774,4123,4151,4183,4341,4411,4692,4693,4727,4739,4880,4896,5098,5189,5270,5340,5375,5440,5453,5494,5543,5573,5590,5643,5673,5698,5802,5830,5939,5985,6075,6102,6137,6165,6276,6474,6608,6669,6732,6847,7001,7064,7070,7083,7240,7241,7282,7418,7490,7605,7695,8068,8142,8469,8603,8733,8850,8883,8977,9053,9126,9138,9144,9173,9265,9370,9411,9455,9773,10087,10156,10182,10335,10463,10592,10651,10700,10705,10748,10815,10848,10927,10962,10971,11076,11115,11149,11165,11345,11349,11434,11493,11503,11510,11579,11608,11684,11809,11830,12009,12047,12067,12165,12292,12306,12382,12622,12736,12812,12937,13031,13133,13355,13409,13450,13472,13497,13542,13544,13673,13871,13920,14060,14117,14261,14686,14696,14856,14864,15243,15256,15331,15531,16049,16113,16172,16419,16437,16491,16625,16764,16846,16891,17160,17189,17494,17583,17707,17708,18608,18643,18648,18675,18680,18740,18750,18856,19007,19122,19214,19505,19619,19639,19771,19843,20201,20249,20490,20751,20841,20996,21038,21361,21540,21658,21683,21711,21782,21890,22140,22258,22613,22927,22968,22991,23087,23189,23247,23376,23561,23566,23593,23649,23666,23689,23776,23805,24109,24134,24313,24338,24532,24608,24733,25321,25326,25600,25801,25881,25940,26333,26684,26690,26729,26798,26874,26981,27046,27195,27323,27340,27355,27441,27644,27688,27767,27923,28221,28294,28470,28555,28765,28842,28846,28889,28919,28982,29128,29165,29171,29185,29211,29255,29312,29316,29349,29382,29603,29776,29810,29840,29846,30032,30130,30488,30913,30990,32315,32434,32450,32566,33628,34103,34121,34480,34832,35092,35556,35618,35705,35709,35740,35755,35997,36012,36259,36267,36383,36686,36888,37019,37035,37066,37114,37127,37138,37305,37564,37636,37802,37807,37939,38009,38085,38190,38229,38669,38699,38812,38877,38931,38986,39119,39501,39687,39841,39876,39989,39994,40362,40547,40589,40618,40734,40799,40826,40944,41350,41401,42454,42463,42529,42631,42788,42798,43230,43711,43827,43962,44108,44123,44606,44619,44749,44976,45188,45258,45435,45633,45731,45802,46268,46546,47200,47325,47927,47954,48166,48569,48649,49062,49844,50329,50396,50425,50453,50671,50786,50855,50912,50918,50978,50993,51025,51029,51222,51239,51253,51348,51372,51384,51399,51456,51495,51545,51562,51580,51676,52102,52158,52214,52244,52264,52268,52308,52337,52359,52371,52424,52736,52826,52965,53017,53097,53283,53338,53346,53413,53467,53529,53679,53725,53726,53761,53817,53825,53918,54043,54045,54124,54140,54211,54538,54539,54700,54792,54892,54942,55011,55013,55067,55112,55122,55227,55235,55529,55583,55652,55698,55861,55870,55888,55913,55919,55936,55949,55993,56002,56165,56240,56315,56328,56437,56497,56626,56732,56816,56893,56956,56983,57205,57294,57389,57441,57521,57539,57545,57592,57621,57651,57680,57765,57780,57810,57873,58067,58169,58229,58365,58458,58506,58530,58566,58582,58615,58623,58701,58713,58728,58737,58819,58828,58881,59052,59182,59264,59407,59426,59545,59748,59862,59864,60047,60052,60146,60159,60260,60304,60327,60362,60397,60488,60558,60614,60729,60867,61053,61132,61216,61250,61338,61718,61814,61854,61856,61865,61911,62010,62025,62196,62220,62233,62533,62607,62654,62705,62751,62754,62755,62764,62895,62907,62909,62947,63128,63313,63370,63386,63441,63520,63539,63701,63736,63834,64017,64061,64062,64131,64235,64244,64254,64330,64352,64362,64390,64431,64462,64483,64524,64679,64924,65145,65225,65241,65281,65502,65542,65601,65752,65803,65828,65853,65971,66007,66066,66203,66240,66341,66385,66456,66493,66642,66705,66758,66822,66823,66951,67121,67143,67186,67230,67267,67303,67361,67371,67402,67440,67523,67748,67787,67840,67920,67927,68062,68104,68235,68360,68381,68387,68489,68524,68534,68637,68728,68740,68762,68795,68861,68914,69089,69101,69207,69238,69275,69340,69525,69682,69739,69902,70025,70031,70095,70516,70617,70626,70658,70676,70751,70852,70962,71294,71343,71371,71401,71417,71446,71536,71625,71636,71708,71869,71883,71909,71936,71965,72484,72649,72699,72774,72836,72876,72885,73000,73142,73235,73243,73247,73375,73381,73399,73408,73475,73508,73649,73659,73688,73811,73854,73980,73994,74026,74121,74123,74200,74210,74270,74478,74510,74540,74623,74721,74737,74854,74993,75043,75139,75243,75434,75620,75762,75807,75870,76151,76200,76479,76490,76634,76715,77008,77061,77114,77250,77292,77334,77484,77485,77556,77802,77807,77809,77839,77895,77939,77963,78051,78183,78200,78214,78276,78309,78383,78435,78706,78906,78912,78944,79051,79152,79510,79517,79583,79602,79613,79689,79807,79819,80017,80194,80333,80367,80386,80442,80534,80621,80689,80810,80865,80947,80980,81014,81145,81238,81251,81281,81380,81405,81495,81573,81742,81777,81823,81857,82125,82297,82306,82320,82409,82500,82634,82735,82893,82998,83109,83363,83586,83597,83623,83668,83684,83702,84085,84113,84342,84454,84462,84581,84706,84849,84896,84969,85029,85084,85092,85111,85177,85424,85682,85691,85754,85902,86070,86094,86213,86246,86310,86464,86593,86677,86781,86922,86952,86970,86971,86973,87142,87211,87239,87314,87394,87440,87631,87657,87686,87691,87802,87840,87952,88052,88187,88197,88207,88333,88409,88422,88588,88589,88848,89085,89172,89296,89446,89486,89493,89568,89571,89584,89617,89679,89680,89712,89753,89978,90001,90022,90047,90090,90093,90295,90477,90574,90576,90621,90668,90732,90902,90942,91029,91051,91080,91244,91532,91610,91703,91796,91898,91925,91968,91977,92094,92138,92211,92254,92289,92290,92315,92352,92381,92546,92614,92665,92692,92757,92825,92948,92959,92976,92990,93019,93021,93175,93257,93521,93546,93573,93613,93758,93776,93817,93843,93883,93951,93991,94015,94058,94120,94162,94164,94193,94338,94379,94442,94598,94786,94873,94960,95066,95097,95201,95309,95364,95453,95469,95824,96046,96158,96242,96248,96360,96373,96504,96526,96614,96738,96739,96771,96789,96903,97109,97114,97132,97280,97375,97403,97404,97444,97585,97598,97699,97709,97735,97790,97802,97946,97963,97980,97996,98034,98172,98316,98341,98353,98387,98399,98528,98586,98727,98733,98763,98791,98818,98856,98871,98968,99035,99080,99102,99107,99245,99372,99403,99477,99550,99571,99651,99662,99712,99772,99867,99876",
          "87,320,339,374,414,428,641,698,733,903,965,984,995,1072,1100,1110,1149,1166,1198,1211,1236,1239,1275,1291,1310,1345,1410,1513,1542,1548,1572,1590,1593,1608,1698,1867,1868,1923,2087,2092,2608,2699,2719,2800,3107,3223,3355,3370,3701,3888,4087,4145,4217,4359,4530,4716,4806,5018,5023,5043,5045,5065,5079,5083,5145,5181,5196,5215,5224,5231,5265,5273,5274,5283,5311,5314,5395,5464,5496,5523,5538,5687,5741,5778,5801,5845,5899,5903,5949,6051,6080,6173,6202,6277,6285,6331,6393,6433,6442,6494,6498,6507,6553,6607,6648,6769,6817,6920,6982,7063,7169,7396,7423,7457,7495,7522,7535,7779,7808,7907,7929,8001,8197,8237,8311,8336,8427,8494,8642,8767,9140,9327,9441,9462,9473,9669,9700,9866,9963,9968,10038,10285,10529,10619,10952,11103,11221,11464,11524,11542,11634,11844,11872,12139,12505,12874,12931,12936,13156,13165,13176,13648,13729,14658,14736,14835,14848,14904,15150,15483,15545,16059,16804,17274,17983,18086,18630,18645,18779,19034,19200,19405,19483,19824,20746,20789,20928,21359,21561,22010,22226,23058,23486,23511,23596,23672,23767,23957,24356,24901,25021,25249,25300,25775,25895,26453,26562,26744,26861,27252,27341,27372,27425,27879,27975,28376,28573,28675,28723,28741,29078,29442,29486,29663,29929,30694,31477,31709,33739,34087,34319,34406,35083,35605,35682,35788,36184,36744,36867,37001,37070,37130,37687,37754,37944,38250,38648,38689,39120,39179,39180,39627,39632,39663,39853,39931,41029,41145,41232,41435,42683,42841,42921,43809,44229,44347,44921,45500,45692,45752,45931,45941,46336,47088,47775,48314,48416,48664,49469,49669,50032,50044,50048,50349,50381,50426,50461,50463,50487,50605,50665,50726,50917,51158,51195,51214,51282,51284,51303,51377,51427,51431,51540,51710,51877,51912,51998,52000,52085,52200,52256,52569,52695,52808,52833,52989,53072,53078,53208,53294,53536,53602,53630,53665,53843,53909,54050,54084,54094,54237,54256,54692,54794,54806,54831,54955,55012,55062,55069,55143,55155,55201,55246,55343,55351,55361,55399,55414,55547,55550,55557,55639,55701,55772,55963,55976,56012,56135,56340,56666,56671,56899,56942,57141,57164,57204,57250,57261,57307,57569,57587,57593,57664,57776,57797,57857,57971,57986,58151,58344,58601,58846,59083,59088,59131,59380,59381,59629,59724,59740,59744,59777,59861,59876,59928,59959,59994,60116,60122,60150,60272,60290,60376,60387,60568,60792,60812,60875,60998,61018,61049,61054,61300,61306,61397,61447,61477,61793,62037,62210,62265,62326,62626,62794,62859,62905,62996,63057,63096,63100,63106,63234,63244,63299,63307,63319,63406,63449,63494,63495,63600,63616,63689,63729,63747,63768,63879,63915,63994,64582,64728,64755,64857,65039,65077,65100,65118,65296,65328,65435,65570,65799,65815,65827,65848,65883,66048,66202,66255,66261,66482,66532,66564,66588,66647,66650,66862,66869,66879,66907,67007,67056,67086,67190,67268,67325,67495,67497,67581,67667,67756,67873,67912,68256,68279,68300,68345,68413,68519,68713,68736,68780,68856,68870,68919,69015,69115,69148,69229,69256,69258,69403,69470,69753,69771,69847,70106,70120,70197,70453,70544,70555,70556,70557,70667,70717,70778,70814,70892,71007,71015,71024,71031,71133,71140,71221,71226,71248,71264,71291,71339,71365,71426,71433,71510,71615,71649,71688,71713,71815,71908,71919,71940,72039,72406,72414,72477,72594,72648,72696,72700,72718,72753,72783,72797,72804,72850,72993,73160,73191,73240,73265,73267,73385,73635,73643,73664,73869,73965,73987,74052,74340,74468,74496,74614,74703,74808,75047,75152,75212,75262,75382,75441,75448,75539,75564,75608,75658,75772,75932,76050,76109,76134,76196,76209,76225,76280,76315,76637,76694,76744,76751,76802,77154,77231,77264,77312,77553,77659,77677,77685,77968,77977,78201,78206,78238,78265,78284,78525,78638,78720,78850,78858,78938,79020,79024,79095,79176,79179,79258,79287,79400,79593,79675,79686,79702,79748,79803,79858,79938,79946,79954,80030,80249,80283,80325,80430,80468,80513,80529,80588,80634,80701,80741,80789,80832,80951,80992,81020,81028,81038,81083,81150,81185,81243,81274,81316,81427,81478,81499,81513,81633,81670,81719,81766,81934,82160,82195,82474,82507,82512,82762,82803,82804,82941,83142,83164,83299,83388,83522,83536,83552,83738,83838,84154,84231,84261,84321,84346,84441,84586,84594,84701,84707,84795,84813,84826,84883,84922,85080,85121,85232,85444,85460,85535,85565,85791,85859,85954,86318,86448,86522,86550,86751,86774,87057,87104,87157,87169,87238,87266,87269,87282,87315,87502,87637,87638,87762,87832,87848,87906,87908,87917,87979,88094,88307,88550,88668,88691,88706,88785,88882,88883,88949,88960,89016,89088,89150,89179,89301,89326,89333,89382,89429,89454,89474,89576,89656,89675,89708,89797,89840,89854,89944,89988,90078,90122,90149,90245,90251,90348,90375,90468,90723,90820,90892,90905,90931,90961,90968,91025,91302,91313,91329,91440,91454,91464,91589,91622,91623,91658,91746,91866,91927,92048,92130,92502,92515,92539,92583,92622,92685,92955,92973,92993,93010,93222,93287,93379,93394,93436,93507,93528,93633,93643,93665,93725,93750,93830,94168,94256,94265,94789,94956,95003,95083,95202,95242,95317,95410,95463,95547,95582,95629,95689,95701,95774,95838,95987,95995,96001,96060,96155,96228,96372,96415,96443,96572,96619,96624,96682,96689,96838,96913,97152,97261,97418,97493,97510,97567,97589,97940,98230,98314,98393,98406,98461,98472,98603,98648,98815,98840,98876,98886,98961,98995,99065,99096,99155,99158,99208,99257,99412,99669,99670,99743,99749,99780",
          "40,201,206,226,244,296,315,521,626,729,883,888,1029,1128,1318,1325,1348,1503,1652,1678,1749,1846,1859,1866,1937,1961,1970,1977,1980,1985,1991,1994,2002,2017,2036,2051,2052,2053,2068,2070,2106,2157,2166,2174,2185,2213,2238,2252,2273,2281,2328,2331,2340,2345,2370,2376,2383,2384,2385,2401,2416,2431,2471,2473,2510,2540,2545,2601,2618,2625,2652,2655,2663,2682,2731,2756,2847,2859,2866,2874,2878,2904,2920,2931,2946,2988,3027,3055,3116,3122,3124,3212,3284,3293,3316,3385,3386,3391,3392,3393,3435,3438,3456,3469,3486,3501,3502,3506,3511,3521,3531,3533,3540,3547,3644,3648,3654,3670,3708,3762,3778,3821,3871,3912,3958,4011,4028,4029,4068,4072,4146,4165,4178,4181,4229,4261,4286,4364,4381,4439,4447,4451,4471,4481,4486,4499,4525,4531,4534,4536,4542,4544,4549,4568,4578,4583,4586,4596,4621,4622,4694,4713,4714,4721,4736,4748,4750,4798,4828,4836,4839,4855,4865,4891,4917,4954,4978,4985,4993,5058,5370,5427,5535,5551,5690,5716,5756,5764,5782,5857,5927,5970,6065,6081,6120,6132,6206,6221,6234,6311,6358,6412,6530,6575,6595,6682,6697,6747,6752,6768,6825,6828,7022,7042,7098,7175,7237,7348,7353,7430,7459,7469,7518,7534,7560,7705,7797,7803,7836,7839,7900,7911,7963,8024,8167,8273,8284,8293,8320,8322,8337,8353,8394,8465,8585,8799,8898,8904,8925,8927,8959,8966,8986,9205,9356,9373,9377,9399,9738,10102,10204,10272,10313,10387,10561,10575,10685,10727,10783,10802,10976,11005,11346,11665,11696,11789,12207,12312,12389,12437,12454,12951,13023,13336,13778,13869,13884,13891,14053,14300,14306,14367,14477,14489,14623,15080,15248,15485,15503,15562,16115,16141,16154,16246,16494,16852,16866,16910,16925,17152,17230,17322,17328,17397,17806,17856,17947,18026,18098,18269,18423,18642,18706,18735,18864,19190,19312,19607,19618,20151,20181,20261,20475,20599,20826,20994,21040,21101,21141,21261,21311,21366,21459,21594,21681,21948,22174,22328,22490,22722,23060,23250,23333,23455,23762,23909,23925,24014,24484,24586,24986,25030,25184,25234,25282,25291,25505,25638,25932,25998,26013,26061,26062,26215,26544,26553,26556,26575,26577,26591,26599,26611,26620,26679,26747,26789,26954,27035,27183,27358,27453,27476,27505,27543,27569,27589,27611,27685,27729,27931,27964,27982,27994,27995,28005,28007,28025,28032,28043,28049,28051,28052,28079,28105,28108,28129,28142,28145,28147,28227,28405,28420,28450,28469,28472,28497,28500,28502,28514,28577,28589,28604,28608,28610,28625,28632,28637,28669,28807,28970,29011,29050,29085,29207,29308,29327,29358,29389,29523,29526,29530,29534,29572,29591,29608,29617,29625,29652,29671,29677,29678,29721,29772,29832,29957,30020,30049,30078,30107,30136,30165,30194,30223,30252,30281,30310,30339,30368,30397,30426,30455,30484,30513,30542,30571,30600,30629,30658,30687,30716,30745,30774,30803,30832,30861,30890,30919,30948,30977,31006,31035,31064,31093,31122,31151,31180,31209,31238,31267,31296,31325,31354,31370,31383,31412,31441,31470,31499,31528,31557,31586,31615,31644,31673,31702,31731,31760,31789,31818,31847,31876,31905,31934,31963,31992,32021,32050,32079,32108,32137,32166,32195,32224,32282,32311,32340,32369,32398,32427,32456,32485,32514,32543,32572,32601,32630,32659,32688,32717,32746,32775,32804,32833,32862,32891,32920,32949,32978,33007,33036,33065,33094,33123,33152,33181,33210,33239,33268,33297,33326,33355,33384,33413,33442,33471,33500,33529,33558,33587,33616,33645,33674,33703,33732,33761,33790,33819,33848,33877,33906,33935,33964,33993,34022,34051,34080,34109,34138,34167,34196,34225,34254,34283,34312,34341,34370,34399,34428,34457,34486,34544,34573,34602,34631,34660,34689,34718,34747,34776,34805,34834,34863,34892,34921,34950,34979,35007,35055,35065,35079,35081,35109,35158,35183,35211,35216,35272,35299,35321,35346,35383,35420,35462,35491,35539,35598,35657,35730,35746,35789,35848,35873,35919,35978,36037,36064,36072,36100,36160,36219,36290,36319,36349,36425,36480,36539,36598,36662,36721,36738,36780,36803,36851,36880,36910,36920,36969,36989,37041,37100,37116,37159,37223,37241,37282,37341,37377,37412,37471,37530,37602,37661,37689,37720,37745,37747,37784,37843,37857,37902,37930,37973,38000,38032,38091,38163,38222,38281,38345,38404,38423,38427,38463,38534,38593,38619,38652,38724,38783,38842,38858,38906,38916,38965,38975,39024,39061,39095,39154,39213,39285,39305,39344,39403,39482,39526,39585,39656,39715,39736,39774,39846,39964,39985,40124,40153,40182,40240,40269,40298,40327,40356,40414,40472,40501,40530,40559,40588,40617,40646,40675,40704,40733,40762,40791,40820,40849,40878,40936,40965,40994,41023,41052,41081,41110,41139,41168,41197,41226,41255,41284,41313,41342,41371,41400,41429,41458,41487,41516,41545,42015,42044,42073,42102,42131,42191,42221,42240,42281,42335,42356,42431,42446,42448,42518,42593,42639,42668,42703,42755,42801,42830,42842,42856,42871,42905,42922,42928,42935,42962,42992,43020,43067,43103,43107,43142,43180,43229,43261,43304,43330,43342,43352,43379,43391,43466,43474,43484,43541,43556,43566,43616,43659,43677,43703,43778,43853,43881,43897,43952,43983,43984,44015,44024,44053,44090,44127,44128,44146,44147,44177,44206,44252,44275,44327,44414,44438,44467,44489,44564,44577,44602,44624,44625,44651,44678,44689,44726,44732,44739,44770,44801,44840,44888,44923,44963,45003,45038,45064,45111,45121,45125,45159,45191,45220,45278,45351,45408,45426,45443,45466,45482,45538,45595,45653,45665,45674,45725,45743,45782,45797,45840,45912,45969,46004,46014,46031,46048,46062,46072,46093,46124,46138,46141,46148,46228,46234,46264,46276,46319,46367,46408,46426,46429,46436,46452,46550,46563,46612,46664,46670,46693,46696,46712,46728,46730,46739,46770,46799,46834,46858,46900,46911,46937,47016,47044,47056,47066,47079,47099,47110,47135,47151,47156,47158,47168,47269,47288,47322,47328,47338,47353,47410,47429,47522,47548,47558,47569,47582,47586,47635,47666,47685,47696,47716,47733,47769,47791,47801,47817,47818,47861,47869,47928,47964,47975,47999,48015,48080,48149,48192,48205,48215,48221,48235,48244,48278,48283,48320,48322,48389,48444,48474,48502,48516,48627,48701,48723,48726,48744,48754,48760,48785,48822,48846,48871,48928,48968,49007,49118,49141,49255,49263,49264,49309,49320,49331,49383,49441,49544,49557,49664,49673,49678,49719,49771,49841,49842,49848,49891,49899,49901,49939,49960,49972,49995,50146,50235,50392,50399,50491,50689,50730,50762,50847,50896,51008,51049,51153,51341,51693,51791,51854,51923,51971,52211,52281,52379,52439,52463,52480,52532,52759,52813,52829,52834,52953,53150,53206,53287,53648,53697,53742,53752,53754,53800,53884,54016,54078,54272,54367,54499,54684,54805,55034,55080,55237,55471,55663,55704,55716,55720,55762,55979,55984,56236,56549,56725,56758,56814,56885,56957,56985,57223,57275,57420,57449,57527,57604,57697,57786,57836,57982,58014,58146,58298,58327,58455,58521,58592,58638,58730,58740,58951,58999,59132,59141,59156,59177,59291,59292,59318,59334,59376,59452,59484,59576,59609,59657,59663,59701,59811,59882,59905,59910,59969,59973,60048,60085,60128,60149,60195,60230,60259,60367,60400,60696,60725,60732,60761,60796,60975,61001,61145,61196,61313,61320,61360,61396,61407,61497,61519,61627,61694,61734,61925,61929,61961,62135,62241,62481,62520,62535,62581,62712,62768,62868,62922,63161,63371,63452,63455,63496,63642,63796,63825,63962,63980,63998,64233,64506,64535,64831,64863,64877,64889,65105,65106,65195,65201,65337,65398,65545,65584,65667,65849,65851,65954,65982,66052,66113,66135,66155,66215,66268,66348,66408,66544,66766,66774,66832,66909,66998,67192,67242,67397,67407,67520,67590,67614,67645,67679,67713,67737,67831,67954,67988,68098,68131,68253,68273,68289,68308,68455,68470,68502,68517,68530,68685,68732,68741,68754,68813,68844,68952,69029,69071,69293,69428,69689,69748,70273,70291,70324,70371,70473,70613,70621,70685,70824,70957,70993,71311,71508,71641,71700,71835,71873,71877,71984,72006,72015,72105,72124,72152,72260,72308,72360,72412,72853,72870,73117,73200,73208,73222,73254,73284,73307,73648,73752,73777,73804,73808,73817,73843,73931,73942,74065,74145,74275,74330,74536,74551,74556,74576,74662,74681,74783,74894,74918,75010,75039,75082,75093,75151,75227,75263,75277,75287,75307,75326,75394,75450,75630,75670,75693,75704,75839,75899,75948,75953,75960,75999,76000,76074,76201,76228,76245,76290,76338,76344,76371,76408,76454,76480,76572,76759,76871,76903,76911,76930,76938,76990,77049,77075,77127,77256,77275,77374,77402,77403,77602,77634,77646,77724,77779,77796,77914,77955,78088,78208,78244,78387,78389,78420,78503,78627,78654,78737,78798,78807,78873,78981,78982,79023,79274,79307,79548,79606,79630,79710,79930,80125,80155,80199,80346,80361,80544,80545,80549,80777,80816,80889,80978,81275,81293,81463,81473,81503,81556,81575,81594,81656,81799,82023,82042,82117,82176,82294,82418,82533,82544,82574,82691,82716,82857,82866,82883,83023,83068,83170,83381,83470,83545,83612,83685,83746,83908,83987,84116,84380,84455,84458,84527,84537,84551,84711,84936,84982,84994,85019,85020,85060,85100,85200,85282,85319,85461,85478,85583,85650,85749,85949,85978,86119,86129,86515,86533,86555,86573,86621,86683,86752,86767,86814,86845,86908,86942,87061,87133,87361,87591,87656,87854,87869,87898,87900,88036,88076,88085,88106,88260,88273,88533,88630,88707,88764,88944,88965,88981,88994,89059,89122,89248,89250,89294,89338,89383,89408,89428,89556,89662,89849,89861,89907,90126,90145,90178,90260,90339,90445,90471,90532,90571,90781,90817,90855,90960,91124,91247,91264,91357,91408,91455,91482,91587,91594,91768,91778,91835,92026,92063,92239,92374,92392,92399,92402,92422,92580,92587,92651,92666,92860,92940,93009,93180,93272,93319,93894,93911,94055,94071,94104,94115,94269,94318,94367,94736,94748,94806,94864,94970,94991,95053,95068,95114,95246,95280,95372,95419,95576,95621,95882,95971,95981,96004,96079,96084,96343,96371,96429,96555,96666,96759,96760,96915,97096,97115,97118,97382,97495,97538,97581,97688,97728,97730,97825,97840,97883,97937,97957,98059,98177,98214,98377,98407,98512,98693,98931,98991,99063,99170,99200,99229,99383,99457,99520,99594,99639,99791,99814,99848,99938",
          "156,234,259,292,303,319,407,420,446,469,470,600,627,654,709,721,804,809,895,906,1042,1087,1183,1306,1387,1390,1433,1483,1511,1514,1595,1599,1643,1695,2448,2516,2759,3234,3240,3388,3398,4081,4380,4956,5010,5051,5060,5078,5104,5199,5404,5405,5409,5435,5596,5674,5786,5876,5902,6006,6114,6162,6315,6344,6384,6401,6438,6562,6646,6706,6912,7101,7250,7276,7380,7393,7517,7580,7626,7686,7760,7784,7813,8094,8335,8408,8456,8499,8567,8580,8656,8749,8841,8914,8945,9097,9264,9338,9409,9575,9905,9961,9980,10062,10091,10143,10223,10316,10412,10446,10490,10757,10776,10850,10876,11039,11162,11244,11363,11377,11581,11588,11804,11812,11887,11953,11996,12016,12060,12122,12167,12422,12647,12879,13333,13372,13610,14035,14350,14632,14634,14737,14798,14900,14977,15198,15298,15459,15764,15786,15990,16637,16794,17190,17220,17356,17431,17551,17633,17741,17927,18116,18495,18520,18549,18628,18690,18893,18965,19351,19556,19766,20069,20624,21406,22132,22546,22664,22738,22754,23020,23111,23239,23530,23722,23754,23826,24241,24366,24654,24667,24686,24790,24797,24861,25039,25043,25728,26233,26727,26907,26957,27060,27234,27889,27919,28530,28770,28782,28984,29130,29451,29514,29527,29732,29827,29856,29901,35875,36177,36242,36444,36933,37244,37615,37671,37760,38053,38054,38118,38245,38319,38428,38490,38550,38760,38762,39563,39693,39917,39927,40211,40778,40907,42378,43457,45105,45183,45304,45315,46075,47540,50071,50116,50137,50142,50149,50221,50226,50231,50291,50308,50356,50389,50687,50770,50817,51028,51119,51146,51288,51409,51423,51428,51464,51511,51524,51619,51689,51861,51932,52239,52445,52484,52504,52637,52660,52693,52745,52932,52964,52992,53068,53105,53143,53531,53547,53661,53706,53710,53975,54035,54164,54200,54232,54554,54627,54734,54855,54859,54902,54994,55039,55041,55051,55136,55196,55295,55369,55509,55707,55713,55743,55789,55965,56030,56031,56084,56085,56241,56374,56421,56466,56487,56492,56511,56515,56636,56705,56731,56774,56781,56788,56970,56993,57020,57079,57187,57407,57540,57761,57886,58030,58121,58236,58247,58281,58413,58469,58553,58784,58797,58818,58962,58982,59028,59125,59223,59280,59337,59349,59418,59460,59668,59847,59978,60056,60071,60201,60241,60299,60312,60355,60404,60460,60469,60557,60768,60840,60923,61096,61144,61278,61290,61686,61706,61741,61799,61812,61894,62076,62102,62347,62451,62547,62744,62849,62946,63136,63408,63422,63429,63489,63534,63698,63934,63944,64112,64166,64349,64501,64619,64657,64722,64785,64925,65164,65180,65362,65516,65647,65654,65684,65747,65864,65967,66062,66112,66161,66165,66242,66282,66398,66669,66711,66714,66956,67015,67058,67120,67129,67151,67264,67454,67491,67493,67507,67636,67802,67818,68001,68007,68270,68388,68417,68451,68503,68520,68553,68599,68744,68776,68818,68837,68970,69040,69206,69295,69301,69372,69375,69449,69457,69460,69606,69729,69737,69815,69970,70014,70134,70137,70146,70304,70328,70352,70439,70444,70475,70524,70771,70843,71073,71150,71193,71265,71276,71350,71428,71436,71458,71559,71591,71617,71664,71682,71758,71831,71858,72114,72190,72321,72392,72604,72629,72656,72762,72788,72869,72958,72981,72995,73083,73153,73422,73654,73746,73780,73911,73919,73959,73960,73966,74119,74166,74213,74316,74350,74427,74479,74547,74750,74794,74809,75025,75106,75126,75562,75622,75759,75855,75907,75940,75987,76028,76049,76160,76272,76297,76424,76442,76478,76591,76714,76834,76881,76888,76954,76965,77029,77096,77152,77168,77191,77201,77228,77268,77367,77469,77470,77488,77563,77601,77725,77837,77998,78099,78149,78181,78330,78608,78694,78702,78809,78889,79028,79109,79229,79253,79316,79319,79364,79382,79597,79626,79688,79833,80004,80021,80192,80280,80385,80524,80830,80874,80924,81000,81007,81043,81109,81192,81329,81397,81472,81507,81702,81716,81768,81937,82082,82257,82338,82357,82451,82622,82850,83095,83195,83306,83374,83432,83785,83830,83904,83923,83930,83966,83998,84008,84029,84031,84135,84303,84314,84529,84617,84634,84703,84789,84798,84819,84828,84898,84980,85027,85046,85067,85093,85098,85156,85303,85332,85373,85438,85469,85544,85690,85757,85808,85827,85852,85898,85966,86057,86074,86092,86118,86152,86221,86259,86271,86303,86373,86569,86578,86655,86667,86777,86878,87037,87038,87039,87044,87054,87137,87279,87378,87382,87489,87507,87521,87675,87689,87969,88061,88114,88132,88198,88203,88272,88335,88350,88366,88487,88573,88850,88904,88911,89071,89078,89169,89290,89339,89388,89391,89475,89508,89607,89661,89725,89727,89732,89899,89933,89980,89990,90077,90187,90255,90265,90282,90438,90622,90651,90733,90797,90861,90908,90972,91028,91069,91187,91273,91304,91376,91431,91616,91711,91815,91837,92029,92044,92380,92535,92568,92592,92730,92762,92792,92837,93023,93061,93104,93140,93141,93145,93158,93204,93307,93344,93370,93413,93445,93481,93488,93530,93625,93777,93821,93861,93992,94016,94139,94151,94380,94564,94619,94631,94672,94707,94768,94796,94817,94922,94968,94978,95076,95146,95223,95308,95353,95588,95617,95656,95784,95803,95865,95889,95915,95926,96006,96085,96105,96211,96335,96457,96827,96867,96954,97176,97259,97338,97455,97575,97668,97993,98082,98117,98122,98196,98245,98247,98325,98405,98418,98429,98448,98552,98615,98654,98779,98821,98829,98839,98914,98946,99014,99085,99095,99277,99283,99357,99363,99422,99450,99543,99546,99629,99732,99830,99965",
          "83,84,124,195,282,419,540,580,763,771,773,898,912,1064,1097,1174,1184,1244,1248,1250,1450,1516,1557,1570,1614,1781,1901,2269,2280,2557,2570,2701,2883,3131,3194,3374,3416,3609,3728,3775,3789,3855,3939,4199,4348,4358,4429,4490,4521,4538,4920,5094,5136,5165,5184,5343,5444,5468,5539,5581,5752,5787,5792,5794,5839,5863,5934,6000,6024,6170,6245,6405,6480,6555,6690,6715,6810,6876,6877,6919,6969,7037,7085,7174,7223,7272,7344,7473,7483,7500,7583,7591,7668,7820,7903,7958,7974,8209,8239,8269,8287,8331,8457,8544,8632,8765,8773,8782,8906,9003,9046,9125,9203,9211,9334,9449,9489,9742,9768,9825,9827,9889,9911,9979,10127,10145,10238,10404,10423,10578,10661,10770,10782,10817,10836,10918,10973,10981,10989,10996,11043,11144,11329,11435,11437,11466,11469,11546,11561,11578,11604,11683,11705,11707,11788,11792,11796,11917,12054,12467,12515,12565,12611,12682,12752,12822,12864,13033,13079,13487,13499,13631,13941,14140,14331,14348,14498,14583,14926,15172,15252,15271,15334,15336,15428,15489,15555,15632,16126,16381,16492,16530,16557,16563,16685,16753,16905,16917,16923,17019,17340,17775,17787,18099,18121,18126,18300,18330,18781,18833,18870,19101,19225,19305,19489,19540,19558,19573,19801,19980,19985,19998,20002,20029,20110,20199,20282,20288,20395,20401,20521,20531,20581,20618,20851,21134,21244,21245,21321,21344,21389,21513,21525,21706,21771,22337,22347,22350,22529,22575,22868,22964,23012,23100,23211,23364,23402,23443,23536,23653,23665,23872,23910,24122,24166,24247,24322,24327,24401,24619,24779,24864,24896,24942,24984,25358,25520,25531,25550,25786,25835,26384,26406,26440,26451,26766,26872,27114,27318,27675,27716,27759,27786,27792,27807,27863,27916,27951,28034,28259,28296,28336,28345,28359,28435,28718,28812,28818,28910,28999,29089,29276,29392,29426,29447,29453,29491,29692,29694,29807,29888,30017,30024,30206,30420,30510,30815,31126,31931,32286,32424,32569,32598,32605,32917,33396,33860,33976,34295,34483,34628,34933,35231,35360,35551,35571,35601,35812,35821,35886,35895,36014,36041,36111,36216,36368,36496,36536,36550,36595,36750,37076,37249,37300,37312,37381,37508,37553,37618,37812,38101,38119,38130,38383,38814,38825,38874,38880,38883,38977,39323,39426,39441,39479,39509,39582,39609,40253,40654,41002,41216,41274,42053,42165,42228,42237,42363,42535,42545,42654,42840,42895,43354,43549,43810,43843,43867,44581,44599,45187,45237,45240,45302,45337,45358,45550,45566,45662,45676,45981,45982,46162,46274,46435,46869,46921,46961,47617,47661,47712,47991,48004,48023,48124,48191,48749,48845,49655,49785,50451,50588,50594,50596,50667,50829,50853,50939,51005,51033,51040,51059,51094,51099,51149,51190,51247,51305,51309,51334,51376,51499,51919,51986,51991,52092,52093,52302,52311,52316,52742,52850,52853,52884,52912,52935,52976,53137,53247,53286,53454,53644,53715,53733,53849,53855,53979,53997,54098,54229,54456,54693,54981,55606,55657,55665,55702,55848,56065,56091,56225,56614,56687,56837,56866,56923,56968,57105,57165,57391,57417,57444,57568,57597,57758,57845,57905,57908,57947,57989,58086,58497,58574,58577,58765,58775,58905,59005,59336,59386,59493,59673,59762,59804,59945,60208,60220,60464,60519,60525,60789,61138,61347,61392,61449,61520,61548,61655,61760,61984,62049,62078,62158,62286,62501,62523,62541,62594,62776,62933,63018,63068,63117,63149,63168,63190,63314,63318,63458,63475,63479,63523,63615,63694,63818,63823,64075,64160,64168,64311,64326,64387,64530,64533,64818,64970,65070,65078,65173,65264,65365,65416,65612,65675,65700,65713,65728,65788,65811,65991,66133,66172,66237,66283,66473,66476,66495,66630,66680,66690,66706,66866,66895,66923,66972,66986,67029,67073,67179,67208,67227,67248,67251,67327,67431,67567,67633,67796,67990,68017,68091,68154,68298,68352,68356,68581,68597,68687,68770,68815,68885,69176,69289,69333,69490,69569,69617,69705,69747,69779,69858,69884,70040,70064,70069,70173,70215,70222,70274,70289,70402,70417,70430,70526,70581,70595,70598,70654,70761,70798,70900,70951,70953,71035,71115,71301,71340,71465,71500,71515,71771,71811,71945,72192,72255,72522,72614,72643,72659,72711,72803,72927,72956,73034,73073,73148,73170,73181,73251,73278,73304,73449,73464,73610,73673,73861,73865,73875,74051,74061,74287,74552,74572,74581,74629,74674,74877,74908,74930,75206,75236,75250,75253,75257,75408,75420,75437,75526,75614,75676,75770,75783,75815,76004,76051,76126,76130,76189,76311,76549,76588,76651,76794,76868,76902,77353,77394,77481,77521,77527,77606,77662,77747,77986,77991,78033,78034,78049,78104,78112,78269,78291,78385,78426,78741,78758,78789,78817,78863,78872,78995,79156,79248,79261,79408,79501,79577,79627,79628,79638,79697,79706,79779,79831,79862,79995,80067,80102,80119,80188,80266,80339,80355,80426,80523,80728,80809,80836,80885,80897,80997,81325,81431,81487,81551,81617,81673,81678,81709,81773,81839,81939,81986,82187,82194,82209,82270,82295,82470,82492,82497,82556,82612,82736,82854,82926,82971,82975,82987,83029,83072,83085,83105,83124,83163,83177,83199,83228,83229,83262,83263,83286,83355,83448,83452,83641,83739,83757,83790,83802,83804,83812,83886,83906,83920,83924,83943,83946,84050,84160,84199,84555,84679,84734,84742,84780,84814,84855,84869,85011,85022,85032,85048,85196,85215,85226,85429,85560,85671,85679,85774,85832,85847,85926,86036,86089,86138,86243,86440,86453,86474,86544,86705,86708,86846,86965,87051,87121,87143,87146,87216,87230,87323,87485,87497,87571,87595,87662,87668,87693,87696,87747,87749,87960,88025,88048,88133,88184,88205,88252,88444,88513,88524,88571,88620,88697,88709,88812,88943,88952,89012,89060,89061,89310,89440,89492,89547,89585,89770,89800,89832,89883,89941,90049,90118,90312,90858,90907,91164,91174,91199,91292,91490,91590,91629,91788,91814,92056,92072,92145,92300,92311,92520,92627,92840,92850,92919,92922,92934,93102,93111,93219,93308,93391,93657,93762,93896,93910,93915,93983,93995,94153,94175,94317,94336,94386,94716,94809,94833,95015,95057,95080,95116,95169,95176,95209,95226,95405,95439,95574,95810,96122,96441,96448,96667,96690,96761,96792,96866,96892,97053,97175,97210,97294,97516,97613,97617,97623,97864,97956,98001,98015,98116,98327,98349,98391,98423,98463,98517,98619,98624,98722,98738,98764,98802,99009,99086,99446,99512,99537,99547,99558,99598,99609,99695,99837",
          "43,132,254,488,542,544,550,658,722,833,967,1205,1467,1518,1625,1682,1687,1835,1862,1884,1886,2144,2196,2241,2388,2402,2475,2502,3178,3225,3403,3600,3764,3813,3822,3879,4159,4270,4278,4389,4434,4514,4588,4820,4922,5186,5220,5277,5448,5449,5483,5508,5548,5591,5726,5734,5908,5920,5948,6036,6112,6201,6264,6290,6388,6398,6418,6545,6616,6678,6704,6764,6788,6908,6958,6970,7096,7260,7389,7414,7431,7448,7642,7662,7748,7755,7762,7777,7809,7869,7933,8093,8132,8328,8500,8540,8615,8619,8705,8715,8721,8731,8860,8921,9171,9278,9359,9360,9429,9486,9506,9527,9536,9600,9607,9756,10017,10289,10418,10430,10452,10628,10671,10751,10796,10806,11050,11063,11121,11186,11215,11307,11387,11497,11526,11567,11676,11727,11748,11782,11959,12029,12051,12088,12164,12315,12472,12526,12549,12556,12559,12665,12690,12751,12894,12922,12925,13007,13019,13027,13144,13311,13313,13319,13390,13473,13531,13622,13666,13894,14042,14106,14132,14293,14349,14529,14643,14711,14761,14836,14843,14925,14966,14988,15092,15238,15422,15508,15667,15670,15850,15895,15978,16011,16269,16395,16429,16476,16480,16550,16641,16693,17013,17319,17371,17390,17404,17418,17587,17653,17797,17854,17949,17960,18112,18150,18167,18183,18228,18310,18414,18454,18500,18506,18528,18551,18592,18799,18890,18985,19146,19175,19180,19289,19328,19575,19594,20040,20088,20144,20196,20204,20278,20553,20569,20607,20620,20659,20723,20771,20799,20938,20973,21160,21209,21246,21517,21624,21740,21758,21768,22007,22141,22197,22206,22363,22387,22444,22668,22836,22873,23003,23134,23307,23335,23470,23602,23720,23736,23763,23830,23842,24154,24229,24279,24371,24458,24553,24923,25313,25539,25601,25693,25748,25779,25812,25853,25872,25957,26114,26168,26276,26349,26416,26613,26652,26657,26688,26712,26720,26730,26757,26765,26792,26985,27054,27077,27101,27108,27186,27240,27297,27316,27519,27585,27658,27665,27790,28167,28200,28265,28330,28404,28440,28674,28679,28724,28779,28794,28858,28959,28974,28977,29140,29202,29235,29236,29278,29435,29493,29681,29686,29758,29760,29802,29961,29974,29986,29989,30139,30149,30498,30759,30817,31368,31425,32064,32122,32209,32324,32498,32933,33456,33561,33734,33793,33920,34036,34837,35151,35405,35562,35660,35679,35684,35703,35744,35880,35898,35981,36003,36009,36052,36170,36258,36321,36363,36372,36495,36573,36665,36677,36754,36836,36879,37000,37016,37134,37247,37295,37447,37497,37499,37555,37634,37759,37942,37952,38055,38075,38129,38140,38189,38367,38502,38578,38711,38795,38818,38820,38991,38992,39043,39064,39185,39259,39329,39435,39718,39747,39998,40174,40662,41046,41265,41421,41479,42453,42526,42599,42694,42785,42865,42943,42980,43134,43154,43263,43445,43607,43629,43715,43804,44339,44423,44714,44811,45318,45366,45484,45666,45672,45695,45809,45866,45870,46685,46826,47637,47662,48095,48427,48657,48899,49109,49214,49427,49465,49761,50033,50050,50064,50100,50111,50332,50400,50484,50494,50589,50693,50716,50794,50901,50958,50974,50998,51201,51211,51271,51310,51373,51471,51477,51713,51762,51790,51834,51851,51907,51911,51958,51963,52023,52084,52171,52363,52365,52533,52548,52638,52705,52947,53082,53203,53284,53353,53401,53443,53594,53860,53873,53986,54026,54058,54240,54275,54293,54301,54327,54355,54398,54408,54409,54424,54427,54475,54506,54549,54555,54575,54634,54695,54710,54759,54773,54808,54840,54951,54972,55005,55137,55326,55342,55450,55748,55781,55802,55926,56120,56179,56262,56363,56380,56537,56783,56796,56876,56905,56926,56955,57018,57059,57132,57251,57466,57474,57581,57585,57638,57682,57730,57759,57853,57956,58053,58080,58174,58347,58520,58537,58659,58761,58817,58948,59200,59293,59469,59485,59560,59675,59775,59802,60010,60130,60217,60291,60356,60446,60531,60674,60754,60815,60920,61003,61007,61033,61081,61092,61337,61373,61429,61459,61467,61498,61601,61743,61776,61980,62060,62071,62126,62380,62435,62486,62538,62664,62792,62914,63058,63213,63230,63296,63331,63402,63535,63553,63607,63643,63647,63678,63824,64025,64102,64164,64216,64270,64313,64345,64358,64370,64497,64502,64640,64644,64682,64709,64860,64894,65162,65188,65360,65369,65377,65399,65520,65645,65795,65822,65895,66004,66196,66278,66357,66457,66471,66524,66560,66624,66655,66845,66873,66897,67008,67081,67113,67335,67603,68129,68171,68371,68421,68613,68724,68767,68846,68886,68967,69017,69234,69642,69678,69949,69969,70001,70022,70038,70228,70263,70332,70365,70392,70483,70615,70641,70741,70756,70937,71025,71170,71315,71432,71470,71483,71488,71529,71601,71656,71689,71767,71870,71911,71963,71991,72171,72208,72370,72404,72431,72467,72511,72551,72670,72934,72976,73124,73245,73298,73666,73709,73791,73803,73920,73986,74029,74240,74257,74356,74555,74589,74624,74646,74893,75001,75031,75050,75740,75860,75862,76027,76101,76115,76153,76166,76239,76505,76779,76933,77044,77221,77259,77332,77495,77562,77640,77694,77707,77717,77740,77753,77778,77872,78018,78052,78077,78314,78321,78481,78568,78661,78683,78876,78892,78904,78922,79048,79056,79174,79188,79493,79696,79701,79801,79864,79962,80044,80048,80173,80187,80207,80293,80328,80434,80461,80489,80569,80610,80618,80753,80834,80913,80983,81426,81553,81563,81640,81663,81744,82033,82071,82145,82172,82192,82317,82545,82647,82793,82842,82846,82863,82880,83006,83117,83146,83270,83323,83372,83418,83473,83634,83759,83798,83848,84178,84202,84498,84565,84766,84815,84821,84823,84917,84938,85095,85119,85125,85172,85298,85328,85482,85527,85594,85620,85698,85733,85961,85991,86372,86396,86514,86571,86604,86732,86737,86750,86812,86828,87078,87256,87307,87338,87648,87650,87746,87810,87880,87912,87984,88004,88008,88034,88062,88068,88099,88240,88298,88405,88534,88541,88631,88768,88866,88876,88899,89128,89431,89443,89532,89677,89695,89740,89813,90036,90107,90177,90179,90200,90235,90277,90344,90466,90585,90672,90836,90837,90974,90979,90994,91006,91024,91043,91046,91195,91197,91386,91481,91549,91651,91817,91880,91906,92002,92157,92207,92274,92277,92280,92345,92507,92555,92560,92576,93029,93082,93144,93192,93194,93250,93275,93278,93375,93589,93642,93663,93668,93716,93865,93974,94026,94073,94125,94187,94197,94214,94433,94434,94576,94659,94828,94958,94961,95244,95288,95302,95319,95342,95452,95502,95546,95624,95662,95710,95923,95945,95960,96008,96098,96216,96223,96287,96317,96630,96705,96721,96756,96829,96991,97089,97148,97242,97301,97462,97506,97599,97634,97658,97705,97706,97716,97764,97941,98063,98352,98492,98541,98635,98637,98644,98855,98956,98977,99252,99356,99388,99398,99676,99741,99881,99884",
          "75,227,239,512,1368,1667,1694,1705,1740,2009,2333,2685,2860,2953,3080,3200,3248,3270,3303,3396,3529,3618,3726,3730,3947,4004,4042,4061,4102,4234,4325,4347,4433,4625,4657,4687,4846,4884,5381,6113,6603,6623,6854,6949,6960,8051,8127,8133,8146,8165,8556,8640,8659,8683,8697,9178,9305,9343,9362,9438,9533,9552,9628,9647,9666,9685,9737,9799,9881,10061,10065,10070,10083,10118,10141,10198,10298,10350,10374,10379,10388,10597,10616,10659,10679,10723,10768,10825,10827,10842,10844,10960,10970,11064,11110,11148,11205,11243,11262,11281,11319,11395,11401,11441,11490,11507,11509,11547,11635,11685,11737,11973,12003,12065,12117,12150,12181,12193,12212,12250,12269,12273,12345,12408,12440,12535,12552,12573,12593,12604,12649,12680,12706,12730,12744,12826,12858,12896,12901,12906,12991,13018,13063,13094,13111,13124,13143,13162,13200,13238,13295,13369,13382,13486,13580,13675,13770,13789,13808,13903,13922,13998,14013,14112,14169,14226,14264,14302,14397,14454,14473,14492,14511,14568,14606,14701,14720,14758,14815,14853,14891,14929,14967,15005,15036,15062,15125,15138,15195,15199,15201,15290,15309,15363,15423,15499,15537,15575,15651,15727,15765,15841,15917,15924,15936,16046,16050,16069,16088,16145,16202,16240,16259,16260,16278,16297,16316,16335,16354,16392,16468,16487,16525,16582,16603,16604,16696,16702,16715,16791,16796,16825,16829,16841,16864,16886,16981,17000,17057,17209,17228,17247,17304,17361,17380,17432,17475,17532,17585,17655,17760,17779,17798,17817,17874,18064,18108,18207,18234,18349,18368,18387,18406,18463,18476,18485,18501,18563,18634,18710,18712,18748,18767,18786,18824,18862,18900,18919,18926,19090,19128,19147,19223,19261,19280,19292,19299,19318,19394,19413,19565,19596,19622,19641,19660,19679,19812,19822,19831,19964,20059,20071,20097,20116,20135,20173,20185,20192,20230,20268,20306,20363,20382,20458,20477,20515,20591,20629,20686,20743,20895,20914,20933,20971,20979,20988,21028,21047,21104,21180,21256,21294,21351,21446,21465,21484,21522,21579,21617,21636,21674,21693,21769,21788,21864,21883,21902,21921,21940,21997,22016,22065,22111,22149,22175,22187,22240,22263,22282,22301,22358,22396,22434,22491,22510,22514,22605,22624,22643,22681,22700,22757,22778,22795,22871,22883,22909,22915,22928,22946,22978,23023,23052,23194,23232,23289,23339,23373,23384,23419,23460,23498,23559,23574,23631,23669,23707,23726,23878,23973,23985,23992,24011,24030,24049,24220,24239,24249,24258,24296,24334,24448,24486,24543,24657,24676,24695,24735,24741,24752,24771,24783,24809,24847,24885,24906,24961,24999,25037,25056,25170,25208,25303,25341,25398,25455,25469,25512,25569,25588,25645,25683,25778,25816,25873,25904,25911,25949,25968,26006,26025,26044,26082,26101,26158,26177,26196,26253,26291,26329,26367,26386,26424,26443,26462,26481,26500,26519,26538,26728,26819,27147,27165,27222,27255,27392,27625,27849,27851,28058,28262,28370,28552,28590,28647,28793,28799,28808,28831,28854,29059,29229,29268,29388,29436,29503,29619,29654,29833,29841,29863,29910,29940,30082,30111,30140,30227,30297,30314,30372,30430,30459,30517,30575,30662,30691,30749,30807,30981,31068,31155,31184,31213,31271,31300,31387,31416,31503,31561,31590,31619,31648,31677,31706,31793,31851,31909,31996,32025,32054,32112,32141,32170,32199,32228,32402,32431,32472,32489,32518,32547,32576,32663,32721,32779,32837,32924,32953,33011,33069,33098,33156,33214,33243,33272,33301,33359,33388,33475,33504,33591,33620,33649,33678,33765,33794,33823,33852,33939,33997,34009,34026,34113,34171,34229,34241,34258,34316,34345,34403,34461,34490,34519,34548,34577,34606,34635,34693,34780,34838,34925,34954,34983,35024,35038,35141,35156,35159,35161,35243,35319,35350,35387,35424,35446,35495,35543,35602,35661,35769,35793,35852,35869,35881,35923,35982,35995,36058,36060,36164,36223,36353,36426,36543,36602,36611,36666,36725,36784,36815,36855,36914,36973,37104,37227,37286,37416,37475,37560,37606,37665,37724,37732,37736,37738,37788,37906,37977,38036,38048,38049,38095,38167,38226,38325,38408,38467,38478,38538,38563,38597,38656,38728,38846,38860,38910,38937,38969,39028,39158,39229,39348,39364,39407,39471,39530,39589,39719,39778,39850,39871,39909,39968,40132,40161,40190,40248,40277,40306,40335,40364,40393,40625,40683,40712,40741,40770,40828,40857,40886,40915,40973,41031,41060,41118,41147,41176,41205,41263,41292,41321,41379,41408,41437,41495,41524,41961,41969,41973,41977,41990,41995,42024,42082,42171,42246,42252,42385,42402,42411,42481,42491,42537,42569,42613,42648,42662,42761,42767,42780,42810,42868,42885,42930,42959,42966,43035,43041,43047,43122,43184,43277,43359,43437,43521,43596,43639,43758,43803,43808,43995,44232,44307,44366,44381,44524,44544,44597,44664,44771,44810,44855,44900,45018,45117,45139,45200,45227,45236,45331,45384,45388,45518,45575,45679,45705,45738,45788,45820,45863,45892,45921,46050,46052,46079,46085,46114,46220,46238,46271,46277,46284,46332,46374,46448,46502,46514,46556,46601,46711,46746,46812,46833,46845,46894,46947,47024,47025,47048,47059,47130,47189,47284,47312,47324,47397,47408,47432,47457,47524,47538,47560,47728,47748,47794,47798,47926,47958,47967,47987,48001,48006,48027,48035,48194,48325,48363,48402,48412,48476,48511,48705,48770,48774,48811,48932,48954,48957,49102,49169,49253,49301,49312,49328,49498,49527,49545,49590,49600,49679,49682,49809,49821,49890,49907,49912,49949,49954,50041,50049,50078,50110,50112,50113,50115,50121,50205,50460,50523,50579,50625,50633,50645,50661,50731,50757,50784,51038,51044,51142,51191,51248,51294,51311,51447,51523,51526,51536,51539,51597,51614,51664,51767,51850,51862,51951,51987,52207,52335,52473,52477,52510,52770,52909,52996,53205,53350,53397,53406,53435,53486,53548,53572,53858,53917,54086,54116,54146,54257,54259,54309,54341,54393,54507,54543,54592,54653,54658,54680,54718,54790,54799,54843,54922,54971,55024,55221,55265,55268,55463,55480,55838,55893,55937,56060,56128,56188,56197,56281,56285,56308,56377,56564,56576,56727,56836,56842,56853,56868,56894,56994,57021,57024,57188,57267,57268,57356,57499,57528,57775,57843,57856,57983,58020,58049,58066,58142,58184,58212,58310,58446,58478,58621,58668,58681,58685,58967,59235,59379,59567,59736,59906,59981,59986,59997,60035,60165,60183,60215,60258,60280,60309,60383,60595,60635,60921,60960,61166,61410,61513,61833,61927,62189,62270,62475,62545,62657,62727,62910,62985,63035,63265,63322,63335,63415,63573,63630,63703,63713,63723,63897,64043,64054,64117,64141,64339,64509,64534,64551,64642,65087,65088,65261,65571,65625,65681,65865,66034,66074,66119,66121,66311,66363,66631,66649,66827,66880,67184,67261,67349,67514,67618,67673,67694,67868,67905,67953,68087,68132,68193,68372,68623,68653,68657,68665,68670,68737,68779,68852,68988,69020,69119,69141,69284,69318,69483,69585,69607,69644,69686,69783,70132,70210,70354,70373,70602,70678,70744,70924,71070,71075,71180,71288,71329,71357,71581,71663,71855,71959,72092,72130,72169,72207,72229,72317,72355,72487,72569,72771,72800,72845,72889,73016,73210,73215,73217,73257,73283,73458,73468,73503,73519,73531,73674,73683,73842,73923,73924,73925,74098,74238,74296,74321,74333,74351,74452,74466,74509,74550,74582,74652,74687,74702,74875,74905,74921,75006,75028,75051,75080,75150,75199,75311,75392,75432,75471,75491,75555,75567,75668,75729,75801,76034,76163,76170,76195,76253,76615,76743,76776,76847,76900,77037,77326,77328,77555,77596,77623,77626,77661,77760,77774,77813,77853,78002,78063,78108,78320,78394,78414,78473,78495,78601,78640,78645,78757,78816,78827,78919,78965,78973,78991,79000,79070,79157,79160,79164,79215,79320,79332,79380,79417,79479,79941,80060,80142,80211,80226,80236,80308,80441,80518,80519,80547,80645,80670,80683,81080,81170,81214,81231,81399,81477,81549,81683,81690,81694,81780,81854,81896,81929,82273,82419,82447,82531,82585,82589,82700,82799,83014,83152,83154,83321,83420,83510,83618,83659,83740,83741,83826,83834,83857,83881,83885,83977,83979,84006,84078,84100,84103,84110,84145,84198,84215,84272,84323,84325,84623,84650,84673,84825,84888,84935,84967,85068,85078,85204,85421,85439,85496,85618,85720,85735,85817,85889,85955,86008,86166,86266,86270,86358,86413,86523,86574,86608,86697,86698,86873,86902,87111,87117,87130,87167,87178,87261,87545,87673,87778,87809,88020,88024,88075,88228,88247,88340,88401,88655,88739,88858,88914,89036,89125,89324,89377,89478,89533,89723,89891,89995,90004,90053,90079,90206,90218,90321,90356,90390,90430,90549,90674,90739,90829,90856,91014,91089,91101,91260,91272,91295,91338,91341,91706,91828,91944,91950,92030,92252,92557,92603,92657,92807,92827,92932,93022,93045,93086,93205,93245,93322,93404,93418,93429,93522,93535,93603,93604,93824,93829,93849,93869,93892,94021,94122,94298,94485,94615,94666,94771,94845,94869,94886,94950,94974,94998,95085,95172,95267,95277,95395,95457,95567,95597,95606,95677,95715,95728,96069,96081,96104,96117,96141,96180,96262,96280,96289,96291,96307,96382,96418,96447,96510,96539,96550,96588,96744,96768,96859,97073,97093,97205,97434,97522,97712,97925,98000,98056,98094,98161,98485,98573,98589,98739,98754,98934,98993,99066,99076,99122,99243,99248,99272,99348,99359,99371,99375,99406,99445,99470,99821,99940,99955",
          "6,15,157,317,326,427,558,728,742,764,839,900,1036,1116,1119,1150,1317,1401,1423,1474,1552,1581,1607,1800,1808,1816,1840,1865,2065,2108,2133,2201,2349,2398,2409,2457,2591,2658,2703,2770,2954,3082,3100,3227,3238,3258,3551,3601,3606,3635,3684,3704,3913,3914,4083,4317,4753,4786,5009,5225,5228,5260,5281,5490,5504,5648,5679,5685,5799,5889,5894,5896,5898,5900,5923,5928,5946,6053,6157,6171,6242,6348,6359,6394,6445,6451,6459,6477,6516,6519,6532,6618,6692,6759,6821,6831,6892,6899,6923,6964,7050,7235,7258,7322,7358,7520,7553,7559,7650,7744,7923,7951,8010,8048,8129,8202,8219,8259,8313,8316,8423,8461,8481,8796,8892,9024,9061,9092,9158,9509,9713,9769,9854,9879,9934,10052,10085,10296,10505,10510,10520,10523,10534,10548,10773,10830,10863,10906,11044,11106,11487,11522,11548,11692,11704,11795,11992,12052,12446,12590,12595,12642,12670,12782,12791,12865,12871,12918,12956,13154,13166,13175,13188,13226,13348,13352,13391,13471,13510,13595,13627,13716,14065,14490,14493,14826,14844,14980,15043,15152,15371,15398,15610,15860,16615,16644,16755,17200,17395,17569,17658,17706,17794,17921,17941,17982,18327,18514,18545,18699,18707,18709,18809,18834,18873,18928,19397,19512,19574,20265,21074,21223,21250,21354,21378,21384,21397,21458,21492,21615,21781,22051,22103,22171,22207,22295,22377,22719,22736,23106,23303,23586,24068,24214,24413,24421,24751,24955,25086,25139,25228,25254,25377,25459,25637,25813,25882,26040,26043,26301,26344,26579,26750,26752,26753,26776,26790,26818,26878,26969,27062,27126,27153,27206,27235,27301,27309,27457,27641,27693,27830,27839,27872,27911,28409,28473,28748,28935,28975,29036,29107,29144,29169,29324,29429,29455,29516,29565,29861,30201,30275,31841,32057,32218,32439,32845,33349,33436,33533,33871,34145,34664,34712,35048,35374,35509,35518,35570,35748,35766,35825,36001,36094,36261,36262,36379,36506,36569,36774,36808,37056,37117,37196,37296,37301,37309,37406,37626,37632,37778,37881,37943,38015,38110,38133,38194,38201,38317,38605,39176,39234,39249,39322,39358,39559,39634,39804,39918,40172,40415,40444,40937,41377,42016,42646,43175,43663,43670,43831,43912,43960,44070,44317,44501,44774,44932,45422,45686,47093,47220,47266,47367,47496,47692,47701,47761,48060,48297,48442,49329,49663,49732,50026,50103,50124,50204,50261,50366,50379,50538,50549,50622,50678,50708,50771,50834,50844,50960,51087,51089,51147,51179,51198,51356,51402,51895,51981,52012,52047,52118,52183,52262,52621,52930,52931,53043,53169,53319,53351,53377,53453,53477,53816,53903,54101,54114,54127,54244,54324,54350,54429,54516,54633,54745,54754,54807,54812,54849,54866,54880,54898,54954,55019,55028,55210,55212,55254,55322,55395,55476,55572,55579,55600,55619,55620,55744,55858,55954,56306,56472,56524,56716,56722,56768,56778,56840,56856,56950,57060,57127,57448,57675,57676,57740,57773,57985,57992,58187,58319,58330,58398,58423,58428,58433,58494,58505,58874,59057,59066,59074,59106,59192,59247,59251,59326,59357,59483,59513,59546,59580,59699,59728,59731,59750,59949,59962,60004,60069,60072,60095,60104,60111,60188,60395,60456,60494,60512,60588,60641,60714,60918,60951,61193,61402,61415,61437,61489,61570,61794,61975,62003,62087,62111,62370,62489,62587,62704,62796,62801,62838,62908,62926,62998,63087,63210,63317,63343,63435,63471,63515,63664,63763,63799,63838,63889,63918,63921,63931,63968,63997,64097,64101,64177,64238,64421,64470,64471,64545,64565,64633,64695,64702,64712,64764,64827,64957,65017,65074,65083,65097,65131,65138,65156,65214,65403,65486,65521,65562,65800,65805,65903,65948,65963,66201,66442,66518,66671,66713,66719,66830,66958,66995,67040,67045,67202,67216,67313,67320,67370,67400,67503,67544,67548,67757,67792,67880,67885,67951,68039,68042,68165,68261,68318,68399,68430,68432,68532,68654,68814,68821,69012,69168,69193,69421,69473,69590,69630,69633,69650,69685,69694,69724,69756,69875,70007,70071,70107,70150,70290,70372,70470,70551,70560,70608,70651,70867,71071,71163,71190,71234,71268,71393,71398,71461,71659,71743,71759,71805,71891,71942,71987,72002,72011,72022,72150,72181,72287,72430,72489,72525,72536,72693,72697,72773,72838,72860,72882,72913,72959,73049,73341,73459,73596,73740,73895,73933,73939,74046,74084,74190,74256,74282,74311,74360,74480,74539,74568,74574,74606,74757,74797,74835,74892,74990,75023,75084,75102,75143,75172,75512,75541,75609,75610,75656,75689,75884,75962,75984,76071,76111,76187,76231,76399,76471,76484,76524,76665,76673,76720,76732,76806,76818,76843,76939,76997,77073,77148,77171,77247,77345,77354,77378,77390,77409,77509,77674,77763,77929,78024,78075,78158,78286,78332,78397,78401,78519,78614,78672,78792,78920,78939,78958,78964,79052,79099,79103,79104,79137,79191,79195,79259,79273,79395,79403,79405,79492,79564,79657,79670,79767,79809,79889,79935,79960,79988,80107,80122,80141,80147,80320,80341,80360,80462,80508,80532,80602,80622,80686,80866,80942,80946,80963,81024,81030,81033,81294,81390,81566,81972,82002,82126,82199,82262,82296,82313,82504,82562,82583,82643,82777,82877,82886,82909,82981,83057,83145,83166,83176,83280,83441,83484,83573,83578,83708,83742,83917,84134,84191,84257,84328,84399,84431,84491,84656,84657,84667,84669,84844,84981,84998,85033,85147,85183,85399,85520,85619,85760,85885,85929,86026,86298,86309,86487,86504,86526,86539,86562,86567,86584,86632,86635,86644,86716,86735,86756,86849,86958,87131,87187,87191,87198,87242,87283,87352,87522,87523,87677,87705,87881,87888,87982,88179,88180,88190,88201,88257,88263,88331,88453,88461,88546,88658,88753,89102,89149,89195,89237,89448,89461,89481,89497,89500,89613,89713,89809,89814,90027,90182,90291,90325,90472,90510,90520,90608,90615,90862,90877,90904,91133,91444,91468,91493,91613,91694,92086,92101,92116,92140,92149,92163,92271,92344,92377,92378,92417,92464,92468,92481,92536,92654,92688,92758,92870,92902,92913,92986,93078,93124,93167,93246,93268,93324,93325,93353,93478,93529,93534,93542,93552,93561,93662,93938,94096,94124,94207,94220,94223,94271,94390,94412,94450,94547,94655,94675,94685,94715,94752,94939,94947,94948,95019,95105,95112,95192,95247,95292,95314,95375,95384,95415,95444,95516,95540,95643,95651,95777,96047,96109,96146,96164,96366,96389,96392,96462,96704,96797,96832,96833,96845,97020,97161,97165,97313,97316,97432,97476,97559,97643,97708,97763,97775,97886,97890,97930,98086,98173,98304,98337,98366,98561,98582,98655,98740,98808,98942,98947,99041,99162,99206,99227,99238,99303,99369,99405,99549,99587,99630,99693,99863,99943,99971",
          "218,327,343,350,408,487,549,649,653,795,799,918,1037,1051,1117,1133,1216,1218,1309,1342,1527,1843,1853,2077,2088,2147,2217,2449,2522,2952,3032,3078,3136,3288,3633,3785,3818,3920,3994,3997,4168,4233,4633,4656,4790,5146,5216,5309,5321,5368,5420,5469,5501,5530,5567,5653,5704,5749,5754,5950,5988,6213,6227,6237,6272,6332,6374,6437,6513,6662,6693,6766,6784,6801,6939,6994,7014,7259,7268,7305,7723,7768,7863,7868,7880,7904,8021,8174,8301,8369,8455,8631,8671,8689,8713,8748,8812,8866,8944,8970,9013,9045,9055,9182,9259,9262,9267,9402,9448,9676,9779,9807,9843,9852,9912,9959,10029,10033,10059,10129,10252,10488,10501,10676,10857,10869,10963,10984,10999,11033,11088,11268,11321,11326,11353,11368,11502,11711,11728,11756,11807,11857,11878,11888,12046,12200,12242,12268,12288,12429,12480,12564,13008,13058,13215,13243,13254,13429,13521,13541,13797,13833,13915,14248,14390,14503,14587,14723,14726,14794,14834,15061,15068,15494,15551,15792,16007,16282,16386,16611,16642,16898,16982,17103,17191,17219,17234,17262,17474,17598,18081,18107,18191,18263,18650,18729,18993,19058,19136,19195,19438,19461,19604,19643,19728,20523,20691,20692,20695,20729,20735,20795,20876,21071,21079,21326,21332,21502,21509,21549,21587,21606,21695,21696,21800,22088,22167,22189,22285,22439,22824,22949,23013,23048,23099,23136,23222,23468,23597,23617,23716,24149,24182,24206,24264,24477,24781,24989,25140,25151,25370,25418,25429,25551,25655,25717,25910,26080,26104,26296,26357,26366,26505,26676,26689,26716,26726,26737,26895,26906,26960,27018,27026,27033,27129,27159,27208,27224,27319,27343,27458,27588,27631,27677,27710,27812,27900,28045,28071,28111,28170,28195,28211,28246,28313,28379,28455,28523,28534,28689,28701,28764,28785,29027,29048,29070,29092,29262,29390,29404,29422,29583,29709,29722,29766,29800,29850,29975,30043,30409,30612,30614,31039,31293,31310,32088,32584,32905,33059,33149,34657,34692,34741,35023,35123,35555,35762,35832,35952,36013,36051,36243,36274,36533,36577,36812,36817,36827,36987,37062,37153,37180,37369,37391,37478,37574,37577,37621,37816,37821,37856,38051,38135,38138,38200,38234,38260,38285,38426,38616,38630,38737,38761,38797,38982,39139,39175,39303,39316,39373,39474,39487,39510,39624,39641,39794,39817,40073,40274,40478,40988,41985,42248,42710,42937,43130,43578,43644,43738,44259,44738,44957,45166,45246,45264,45573,45985,46477,46501,46681,46870,46913,47001,47743,47880,48380,48459,48585,48995,49129,49446,49487,49765,50045,50139,50292,50351,50450,50496,50566,50670,50694,50774,50807,51081,51116,51140,51182,51279,51434,51506,51537,51698,51744,51749,51773,52007,52141,52202,52218,52241,52283,52326,52385,52394,52958,53054,53061,53131,53220,53234,53266,53298,53539,53610,53717,53735,53781,53898,54292,54494,54514,54528,54673,54720,54782,54829,54966,55015,55023,55339,55340,55634,55728,55810,55844,56406,56482,56565,56606,56803,56830,56851,56896,56953,56965,56997,57038,57039,57244,57321,57340,57350,57460,57509,57530,57763,57812,57934,58111,58154,58166,58203,58283,58366,58440,58447,58570,58613,58777,58796,58927,58969,59017,59033,59111,59152,59155,59197,59384,59434,59437,59461,59648,59691,59738,60127,60161,60213,60317,60320,60433,60484,60491,60564,60730,60772,60820,60892,60929,60966,61119,61175,61253,61264,61307,61324,61629,61733,61801,61826,61860,62064,62205,62217,62298,62355,62386,62391,62442,62524,62632,62753,62799,62819,62823,62901,63006,63186,63249,63375,63399,63413,63557,63599,63613,63627,63655,63685,63750,63792,63985,64000,64048,64272,64284,64338,64493,64542,64567,64590,64639,64705,65174,65313,65483,65529,65691,65734,65777,65896,65981,66190,66232,66406,66423,66484,66726,66957,66970,67047,67100,67210,67416,67420,67504,67780,67817,67860,67924,67966,68037,68186,68304,68307,68353,68380,68639,68650,68834,68868,68957,68960,69082,69215,69217,69221,69251,69378,69422,69455,69578,69852,69887,69953,69979,70285,70350,70413,70418,70579,70638,70661,70670,70703,70738,70820,71122,71134,71165,71166,71202,71279,71391,71395,71396,71440,71481,71555,71587,71590,71878,72020,72126,72299,72377,72426,72433,72535,72574,72587,72785,72817,72835,72938,72969,72972,73007,73136,73315,73354,73355,73363,73393,73556,73679,73760,74107,74205,74206,74487,74494,74538,74756,74906,74985,75005,75013,75124,75141,75245,75265,75296,75351,75404,75486,75487,75516,75577,75579,75660,75716,75775,75914,75958,76036,76061,76140,76211,76249,76345,76508,76640,76679,76681,76733,76966,77004,77054,77121,77190,77226,77279,77315,77368,77430,77443,77487,77678,78006,78138,78145,78147,78275,78297,78430,78458,78531,78539,78806,78821,78840,79378,79399,79521,79716,79737,79740,79822,79992,80079,80198,80278,80324,80347,80655,80737,80760,80765,80779,80788,80928,80941,80956,81100,81155,81252,81255,81256,81300,81331,81445,81457,81465,81582,81588,81612,81855,82055,82092,82179,82196,82463,82532,82536,82676,82772,82785,82813,82837,82964,83002,83097,83274,83316,83405,83633,83844,83860,83990,84024,84076,84123,84184,84222,84227,84248,84260,84270,84410,84446,84778,84831,85126,85151,85163,85494,85502,85566,85730,85734,85778,85937,85983,86041,86054,86060,86107,86256,86393,86420,86641,86818,86832,86857,86884,87021,87108,87414,87473,87501,87513,87536,87608,87621,87676,87729,87903,87943,88077,88224,88306,88313,88385,88403,88596,88623,88728,88734,89157,89160,89207,89234,89528,89535,89574,89586,89606,89622,89689,89711,89789,89803,89825,89852,89886,90084,90313,90329,90489,90493,90519,90564,90636,90697,90935,90941,90969,91017,91147,91389,91446,91602,91719,91913,92012,92060,92088,92122,92128,92185,92192,92233,92242,92406,92454,92509,92559,92637,92640,92903,92939,92978,93004,93031,93070,93211,93290,93363,93503,93515,93855,93870,93904,93945,93963,93984,94024,94113,94230,94415,94471,94515,94628,94799,94853,94966,95044,95047,95236,95250,95399,95983,96157,96377,96514,96515,96578,96584,96776,96843,96852,97012,97086,97108,97160,97215,97271,97305,97306,97514,97523,97533,97588,97739,97784,98005,98011,98193,98471,98560,98623,98669,98814,99094,99192,99211,99215,99330,99577,99605,99774,99899,99966",
          "48,73,85,131,504,527,702,727,1145,1170,1409,1497,1574,1657,1720,1771,1775,1837,2023,2248,2279,2293,2307,2316,2547,2566,2686,2689,2708,2774,2962,2967,3010,3063,3165,3181,3345,3461,3545,3729,3823,3838,3881,3952,4118,4153,4166,4211,4291,4295,4655,4731,4841,4848,4859,4933,4934,4949,5001,5217,5371,5382,5477,5513,5519,5527,5710,5819,5835,5972,6022,6152,6224,6252,6286,6308,6336,6402,6515,6698,6782,6829,6848,6959,7020,7033,7109,7158,7189,7284,7301,7313,7378,7383,7449,7475,7494,7558,7596,7598,7643,7658,7770,7793,7798,7865,7874,7876,7889,7916,7924,8057,8073,8101,8166,8264,8348,8349,8356,8432,8440,8478,8586,8667,8694,8724,8741,8752,8824,8949,9030,9064,9071,9104,9218,9294,9309,9318,9554,9704,9846,9925,9938,10101,10117,10132,10237,10409,10533,10583,10639,10762,10788,10872,10880,10977,10987,11116,11129,11233,11331,11333,11348,11414,11556,11619,11633,11669,11805,11821,12005,12111,12156,12194,12462,12507,12623,12666,12702,12726,12850,12907,12919,13134,13217,13219,13225,13248,13383,13399,13418,13470,13526,13602,13607,13702,13997,14027,14049,14075,14099,14122,14129,14149,14409,14453,14597,14666,14928,14937,14986,15025,15048,15106,15173,15647,15933,16144,16179,16267,16286,16396,16674,16812,16816,16887,16941,16971,17028,17046,17086,17107,17195,17323,17405,17426,17440,17500,17553,17578,17590,17612,17654,17893,18012,18022,18045,18051,18140,18168,18282,18435,18492,18715,18822,18871,19189,19191,19244,19375,19378,19381,19508,19593,19625,19632,19708,19717,19945,19972,20033,20277,20297,20442,20527,20582,20704,20705,20759,20809,20903,20943,21080,21158,21310,21643,21644,21667,21677,21682,21741,21777,21779,21824,21835,22096,22190,22346,22443,22576,22985,22988,23009,23027,23097,23164,23188,23430,23690,23697,23795,23905,23907,24123,24168,24293,24402,24610,24656,24700,24723,24806,24911,25007,25014,25045,25173,25255,25416,25446,25597,25647,25762,25899,25928,26015,26149,26270,26318,26348,26446,26509,26602,26781,26944,26964,26987,27005,27021,27048,27239,27339,28318,28360,28367,28606,28769,28836,28839,28913,29054,29152,29240,29457,29553,29567,29579,29641,29742,29825,29881,29946,30401,30433,30568,30873,30875,30923,31067,31332,31390,31426,31670,31735,31738,31743,31825,32006,32083,32131,32179,32207,32289,32627,32692,32956,32961,32985,33039,33062,33079,33233,33494,33562,33797,34000,34770,34809,35034,35166,35361,35368,35381,35617,35692,36118,36172,36260,36373,36380,36443,36477,36624,36659,36682,36756,36822,36945,37094,37144,37307,37321,37338,37521,37572,37573,37658,37828,38202,38265,38310,38388,38398,38431,38590,38611,38631,38708,38787,38867,39031,39063,39113,39484,39494,39540,39552,39712,39840,40169,40224,40491,40908,41203,41326,41372,41728,41981,42030,42185,42201,42601,42623,42681,42702,42723,42728,42741,42786,42816,42818,42955,43043,43168,43365,43456,43475,43628,43704,43755,43833,43890,44037,44404,44446,44469,44812,44863,45228,45241,45319,45423,45508,45555,45688,45734,45762,45805,45849,46009,46164,46229,46307,46331,46437,46560,46566,46741,46793,46905,47706,47756,47961,48195,48245,48288,48636,48803,48839,49197,49323,49330,49688,49999,50072,50114,50145,50232,50438,50532,50628,50637,50749,50795,50833,50861,50867,50957,51022,51023,51062,51080,51264,51297,51390,51525,51533,51558,51650,51703,51765,51915,52022,52153,52157,52206,52494,52563,52582,52675,52722,52791,52920,52939,52962,53089,53092,53177,53272,53278,53462,53494,53500,53711,53959,54051,54168,54299,54315,54356,54501,54556,54628,54691,54699,54727,54744,54766,54908,54925,55029,55058,55085,55285,55502,55645,55682,55693,55768,55788,55879,55909,55931,56006,56007,56036,56064,56067,56211,56239,56275,56424,56478,56548,56577,56583,56702,56865,56908,56954,57012,57013,57041,57146,57255,57290,57349,57735,57854,57987,58197,58337,58345,58532,58558,58651,58671,58756,59027,59157,59244,59269,59588,59734,59796,59815,59826,59866,59920,59924,59956,60053,60058,60097,60123,60207,60431,60678,60978,61050,61146,61156,61178,61412,61502,61530,61539,61552,61603,61667,62019,62243,62267,62287,62358,62592,62938,63014,63019,63152,63185,63198,63212,63272,63575,63684,63743,63762,64027,64269,64430,64494,64612,64717,64737,64856,64862,64951,64994,65000,65370,65485,65549,65605,65680,65759,65776,65820,65909,65920,65947,65965,65970,66013,66108,66115,66267,66270,66317,66448,66569,66637,66673,66794,66837,67001,67115,67125,67136,67193,67224,67487,67588,67785,67893,67961,68077,68113,68178,68189,68284,68348,68420,68542,68586,68753,68817,68824,68968,69181,69213,69343,69426,69807,69934,69998,70004,70048,70136,70336,70441,70537,70583,70671,70711,70946,70965,71037,71065,71081,71231,71355,71495,71537,71711,71714,71839,71876,71879,71971,72070,72093,72257,72403,72488,72493,72499,72542,72640,72901,72960,72965,72998,73002,73035,73157,73159,73182,73201,73218,73345,73406,73418,73419,73455,73510,73528,73544,73657,73711,73723,73934,74069,74100,74120,74124,74227,74389,74415,74429,74543,74618,74707,74795,74804,74826,74934,75239,75254,75330,75345,75370,75430,75517,75528,75554,75576,75582,75635,75707,75910,76026,76064,76110,76180,76263,76301,76326,76328,76368,76426,76456,76492,76521,76568,76798,76870,76878,77024,77437,77518,77526,77631,77664,77711,77776,77915,78038,78132,78179,78274,78461,78566,78634,78721,78722,78953,79003,79058,79066,79150,79375,79442,79663,79678,80005,80088,80493,80496,80506,80512,80570,80625,80804,81149,81247,81268,81317,81389,81406,81715,81783,81844,81850,81880,82037,82142,82178,82201,82293,82324,82515,82570,82600,82679,82968,83022,83033,83096,83103,83108,83309,83346,83369,83462,83465,83766,83810,83912,84141,84164,84182,84235,84316,84317,84338,84505,84536,84560,84615,84840,84847,84951,85013,85094,85155,85189,85194,85302,85342,85372,85484,85513,85530,85747,85907,86004,86157,86175,86179,86263,86282,86323,86351,86401,86509,86546,86779,86972,86975,87068,87140,87159,87232,87250,87304,87413,87460,87735,87955,88049,88092,88194,88295,88722,88775,88895,88898,88936,89048,89079,89136,89278,89314,89319,89354,89358,89465,89909,89991,90075,90076,90120,90167,90223,90250,90543,90569,90625,90725,90736,90752,90807,90871,90878,90887,91057,91146,91345,91375,91526,91685,91757,91784,91919,92054,92136,92146,92268,92323,92382,92553,92596,92630,92778,92868,92880,92983,93005,93121,93321,93417,93519,93532,93616,93686,93699,93846,93871,93882,93924,93937,93954,93964,94224,94276,94339,94355,94404,94437,94501,94572,94681,94730,94847,94855,94872,94877,94934,94955,94977,95115,95177,95239,95240,95254,95320,95371,95387,95482,95504,95527,95577,95613,95788,95900,96051,96074,96152,96153,96202,96214,96395,96483,96544,96546,96594,96700,96811,96814,96898,96940,97009,97016,97035,97091,97207,97214,97321,97331,97391,97448,97525,97720,97725,97740,97973,98018,98040,98355,98408,98417,98543,98590,98641,98834,98850,98883,98903,98924,99060,99123,99126,99166,99327,99435,99597,99652,99769,99779,99861",
          "24,28,53,199,232,295,311,528,606,1013,1626,1692,1755,1789,1906,2012,2083,2099,2113,2167,2261,2294,2344,2394,2486,2614,2763,2768,2821,3087,3148,3199,3314,3346,3356,3375,3491,3539,3741,3786,3923,3938,3943,3999,4080,4106,4114,4218,4308,4373,4409,4512,4557,4688,4699,4726,4751,4757,4771,4833,4915,5108,5230,5253,5272,5354,5439,5454,5635,5678,5769,5791,5931,6015,6033,6047,6062,6067,6160,6183,6243,6278,6414,6427,6436,6455,6626,6672,6742,6771,6773,6832,6868,6900,6946,7044,7095,7114,7294,7477,7479,7523,7614,7628,7821,7826,7902,8148,8252,8297,8343,8361,8444,8475,8545,8569,8755,8770,8790,8844,8941,8946,9063,9103,9217,9227,9261,9299,9341,9417,9483,9550,9590,9877,9992,10139,10140,10152,10231,10254,10317,10326,10347,10375,10376,10426,10434,10622,10663,10801,10866,10895,10921,10938,11138,11171,11223,11275,11290,11595,11600,11642,11729,11794,11827,11889,12040,12185,12245,12263,12297,12352,12364,12434,12609,12756,12796,12856,12903,12958,12971,13198,13203,13328,13408,13446,13504,13561,13611,13711,13719,13776,13841,13854,13864,14002,14055,14115,14151,14164,14353,14416,14469,14481,14500,14512,14523,14588,14676,14766,15001,15046,15113,15135,15142,15240,15258,15268,15403,15417,15543,15594,15599,15801,15822,15966,15993,16041,16156,16576,16639,16675,16730,16786,16926,17076,17089,17171,17185,17205,17260,17443,17606,17656,17855,17984,17987,18029,18229,18425,18526,18652,18656,18739,18867,19009,19043,19242,19350,19559,19609,19848,19931,19997,20095,20106,20320,20331,20512,20519,20622,20762,20874,20936,21284,21300,21436,21558,21661,22146,22221,22231,22248,22409,22467,22548,22598,22645,22689,22756,22847,23041,23126,23170,23193,23245,23281,23457,23897,24090,24138,24163,24283,24340,24391,24500,24557,24562,24575,24600,24710,24749,24907,24965,24967,25207,25651,25653,25678,25795,25829,25857,26063,26066,26153,26184,26294,26499,26536,26565,26674,26695,26706,26711,26734,26783,26823,26864,26980,27010,27105,27139,27168,27229,27365,27417,27430,27438,27477,27499,27521,27537,27557,27709,27720,27842,27869,27945,28148,28171,28214,28350,28352,28361,28481,28528,28662,28717,28780,28801,28820,28993,28998,29057,29066,29181,29223,29251,29253,29302,29303,29400,29423,29424,29466,29684,29716,29728,29755,29788,29889,29930,30000,30053,30104,30270,30285,30291,30343,30452,30467,30604,30623,30894,31081,31110,31319,31358,31598,31690,31873,31941,31956,31967,32018,32178,32305,32474,32497,32550,32618,33004,33555,33707,33784,34190,34219,34329,34353,34493,34773,34857,34896,35036,35126,35251,35269,35335,35400,35448,35463,35511,35591,35609,35629,35701,35998,36020,36063,36131,36132,36182,36266,36273,36387,36469,36615,36674,36687,36759,36777,36831,36866,36886,36889,36928,37132,37176,37194,37201,37275,37316,37405,37432,37435,37439,37454,37561,37579,37983,38181,38246,38355,38378,38544,38554,38645,38679,38807,38809,38962,38980,39001,39128,39217,39257,39628,39743,39767,39821,39839,39893,39919,40103,40143,40219,40375,40420,40422,40502,40594,40652,40684,40862,40890,41089,41116,41138,41322,41807,41849,41899,41901,41925,41956,41965,42014,42167,42202,42254,42362,42396,42417,42467,42485,42679,42685,42784,42936,43044,43068,43079,43218,43317,43392,43452,43650,43674,43854,43883,43907,43941,43958,44362,44550,44727,44820,44896,45124,45300,45310,45352,45572,45580,45596,45613,45683,45847,45927,45934,45955,46057,46094,46147,46221,46247,46302,46506,46571,46590,46701,46945,47504,47528,47699,47779,47886,48057,48138,48222,48262,48521,48660,48703,48738,48851,48856,48989,49034,49177,49492,49652,49654,49889,49931,50028,50169,50260,50280,50348,50691,50787,50798,50850,50868,51032,51043,51061,51100,51155,51255,51388,51394,51440,51515,51658,51857,51858,51978,52138,52155,52169,52205,52223,52466,52617,52622,52968,52985,53018,53029,53047,53219,53225,53288,53390,53391,53418,53617,53694,53740,53972,54081,54106,54131,54151,54407,54448,54468,54657,54696,54703,54707,54918,55104,55204,55280,55584,55648,55671,55708,56070,56086,56246,56372,56388,56398,56404,56621,56622,56630,56658,56674,56798,56951,57016,57068,57213,57220,57298,57428,57507,57945,58001,58012,58024,58039,58138,58189,58214,58221,58409,58470,58525,58697,58747,58785,58804,58957,58993,59009,59060,59203,59447,59582,59586,59825,59840,59890,59939,59954,59976,60003,60029,60141,60214,60242,60281,60310,60314,60321,60453,60753,60852,60891,60913,60917,61183,61349,61614,61658,61846,61909,62044,62069,62086,62110,62202,62293,62328,62350,62462,62532,62608,62759,62763,62999,63002,63039,63081,63176,63201,63211,63216,63218,63221,63281,63400,63487,63550,63622,63634,63734,63845,63881,63932,64018,64087,64144,64178,64246,64304,64328,64400,64438,64585,64809,64932,64937,64945,64961,65112,65129,65199,65216,65226,65238,65292,65366,65406,65412,65413,65479,65490,65498,65510,65525,65585,65671,65699,65753,65978,65988,66012,66084,66238,66274,66399,66521,66585,66602,66772,66925,67049,67059,67069,67238,67391,67517,67686,67711,67765,67798,68082,68086,68092,68156,68325,68377,68461,68468,68471,68491,68563,68682,68707,68855,69171,69173,69226,69306,69842,69880,70085,70194,70327,70385,70543,70578,71155,71397,71421,71480,71504,71527,71579,71654,71826,71932,72031,72212,72233,72315,72325,72339,72389,72394,72466,72607,72733,72736,72812,72873,73004,73044,73090,73337,73379,73415,73527,73624,73663,73704,73796,73814,73825,73897,74127,74135,74187,74211,74284,74403,74421,74524,74557,74697,74763,74814,74902,74951,74956,74988,75061,75092,75157,75160,75533,75556,75563,75621,75629,75655,75748,75766,75773,75859,75867,75869,75945,75947,76070,76127,76273,76556,76585,76633,76641,76655,76815,76840,76874,76899,77046,77077,77098,77101,77472,77475,77723,77743,77787,77789,77930,77973,78212,78522,78621,78647,78749,78805,78819,78976,79010,79061,79116,79117,79454,79635,79844,79920,79927,79972,80003,80010,80117,80315,80319,80330,80383,80505,80528,80556,80613,80658,80675,80698,80750,80790,80850,81102,81146,81196,81265,81287,81417,81443,81466,81619,81634,81693,81706,81945,81970,82091,82247,82298,82367,82369,82403,82406,82453,82571,82616,82738,82962,83082,83101,83120,83132,83284,83291,83296,83490,83541,83564,83598,83699,83745,83861,83955,83961,83983,84020,84060,84088,84165,84208,84391,84413,84435,84463,84492,84589,85238,85435,85445,85464,85501,85523,85611,85685,85763,85819,85823,85828,85864,85868,85917,85969,86145,86346,86377,86498,86639,86679,86688,86718,86747,86755,86762,86842,86844,86893,86909,86948,87107,87160,87356,87367,87415,87437,87442,87551,87628,87633,87799,87927,87944,88058,88060,88125,88226,88256,88261,88388,88448,88523,88544,88563,88647,88667,88679,88752,88878,89004,89081,89083,89334,89368,89504,89525,89767,89772,89822,89853,90043,90105,90115,90161,90225,90259,90335,90370,90377,90433,90605,90721,90742,90776,90812,90867,90947,91056,91107,91216,91426,91475,91557,91568,91612,91653,91758,91804,91847,91864,91924,92188,92390,92425,92426,92430,92542,92549,92577,92594,92647,92701,92843,92951,92960,93001,93038,93234,93249,93303,93368,93369,93415,93510,93538,93636,93732,93782,93786,94022,94032,94060,94101,94268,94463,94481,94650,94701,94747,94825,94851,94884,94899,94900,94993,94997,95077,95119,95120,95163,95228,95355,95382,95495,95541,95560,95580,95663,95782,95974,95977,96039,96080,96195,96344,96370,96421,96495,96496,96533,96632,96742,96743,96790,96810,96834,96863,96900,96950,96955,96969,97004,97006,97011,97083,97348,97389,97439,97446,97512,97518,97615,97690,97750,97947,98012,98248,98318,98333,98339,98359,98525,98538,98673,98692,98724,98736,98743,98817,98838,98873,98933,99016,99036,99115,99152,99191,99230,99275,99278,99321,99453,99740,99784,99890,99932",
          "79,222,287,309,358,365,373,678,688,802,940,948,1122,1323,1326,1380,1452,1528,1666,1697,1725,1834,2169,2290,2506,2555,2665,2674,2695,2730,2807,2817,2851,2876,2998,3168,3513,3566,3638,3834,4190,4296,4343,4369,4387,4513,4645,4684,4765,4899,4904,4921,4925,5031,5116,5130,5131,5391,5529,5545,5601,5626,5656,5821,5932,5955,6076,6116,6296,6297,6363,6429,6456,6496,6524,6547,6686,6753,6824,6834,6872,6881,6938,6999,7089,7155,7219,7323,7415,7420,7438,7474,7498,7789,7849,7896,7953,7981,7996,8149,8305,8473,8520,8571,8595,8700,8701,8745,8822,8852,8939,9035,9050,9074,9149,9150,9153,9288,9310,9388,9431,9569,9872,9884,9929,10392,10486,10507,10558,10674,10884,10887,10932,11054,11122,11127,11340,11419,11559,11589,11631,11672,11779,11840,11861,11886,11916,11947,12041,12195,12356,12438,12453,12632,12734,12885,13322,13338,13507,13560,13562,13620,13818,13875,13905,13947,13968,13996,14391,14462,14505,14515,14549,14564,14690,14717,14791,14813,14969,14996,15040,15072,15160,15376,15460,15565,15624,15669,15686,15957,16031,16108,16142,16264,16357,16387,16406,16546,16632,16838,17072,17202,17311,17408,17483,17662,17997,18046,18055,18076,18146,18504,18535,18620,18756,18760,18913,18916,19035,19056,19079,19183,19291,19503,19522,20039,20615,20855,20862,20875,20932,21046,21065,21166,21375,21408,21623,21664,21731,21775,21988,22018,22114,22131,22208,22290,22312,22372,22374,22496,22629,22651,22675,22678,22838,22863,22869,22935,23004,23121,23137,23204,23447,23493,23821,24581,24635,24960,24962,24964,25027,25225,25257,25335,25408,25618,25643,25840,26128,26483,26651,26671,26810,26920,26922,26974,27127,27196,27277,27347,27379,27427,27551,27556,27575,27882,27907,27934,27946,28235,28339,28365,28386,28422,28496,28712,28912,28983,28989,29231,29259,29501,29524,29596,29734,29873,29977,30081,30090,30150,30255,30324,30865,30904,31232,31251,31339,32608,32691,32836,33175,34193,34287,34760,35014,35017,35256,35622,35693,35715,35955,36301,36474,36637,36647,36986,37107,37124,37183,37250,37289,37490,37623,37633,37690,38139,38254,38257,38503,38548,38602,38617,38693,39066,39067,39184,39376,39439,39742,39932,40398,40705,40959,41010,42022,42483,42831,42854,43118,43402,43420,43460,43547,43667,43985,44041,44202,44357,45238,45247,45288,45415,45446,45600,45607,45630,45693,45698,45749,45858,45865,46318,46369,47527,48024,48373,48873,49510,49614,50023,50263,50305,50368,50408,50543,50739,50765,50830,50904,50991,51041,51126,51269,51415,51596,51648,51712,51815,51984,52103,52180,52224,52253,52272,52295,52313,52646,52697,52712,52805,52903,52911,52950,53011,53014,53024,53110,53274,53343,53606,53623,53640,53791,53839,53930,53980,54311,54317,54363,54454,54508,54584,54640,54642,54722,54832,54876,54946,54949,55177,55356,55526,55558,55847,56075,56092,56096,56292,56329,56349,56410,56459,56574,56575,56648,56772,56790,56874,56928,56931,57286,57421,57467,57492,57493,57630,57696,57723,57844,57878,57974,58000,58003,58078,58249,58256,58314,58331,58438,58580,58626,58772,58979,59008,59015,59047,59148,59234,59285,59298,59325,59391,59399,59594,59756,59896,59990,60270,60406,60443,60536,60567,60660,60798,60968,61005,61017,61129,61192,61215,61343,61390,61448,61458,61535,61566,61593,61606,61640,61766,61900,61979,62008,62119,62268,62296,62321,62381,62399,62443,62444,62522,62542,62575,62602,62785,62795,63061,63092,63167,63635,63771,63833,63952,64146,64222,64327,64389,64401,64626,64965,64995,64999,65086,65262,65449,65578,65661,65724,65880,65916,65961,66166,66452,66595,66622,66762,66929,67103,67137,67249,67312,67417,67456,67526,67551,67587,67628,67664,67824,68142,68229,68358,68422,68667,68672,68760,68966,69022,69045,69227,69300,69391,69444,69555,69583,69599,69620,69794,69864,69882,69906,70151,70185,70294,70490,70527,70567,70573,70601,70656,70690,70694,70753,70835,70876,71056,71141,71161,71313,71366,71419,71430,71452,71453,71592,71630,71644,71665,71672,71775,71790,71874,71992,72012,72045,72276,72371,72417,72517,72570,72740,72746,72757,72947,72954,73064,73135,73155,73166,73176,73194,73249,73286,73293,73325,73481,73569,73570,73662,73669,73707,73751,73757,73799,73916,73932,74012,74251,74312,74315,74423,74504,74511,74515,74516,74532,74751,74787,75255,75565,75637,75695,75916,75917,75939,75970,76097,76184,76419,76482,76500,76584,76703,76724,76764,76775,76781,76924,76937,76942,77050,77083,77142,77419,77465,77523,77675,77732,77738,77744,77762,77891,77901,77921,77941,77975,77994,78135,78159,78204,78251,78306,78325,78408,78434,78616,78618,78628,78738,78828,78885,78926,78937,79111,79272,79308,79326,79362,79376,79396,79435,79438,79472,79523,79550,79588,79673,79700,79855,79929,80073,80074,80166,80189,80190,80233,80292,80348,80531,80578,80668,80803,80934,81160,81208,81391,81392,81449,81475,81528,81578,81598,81627,81739,81761,81829,82057,82105,82217,82389,82411,82421,82440,82514,82594,82599,82625,82774,82831,82902,82931,82983,83056,83269,83319,83493,83600,83877,83972,84059,84221,84258,84329,84331,84404,84543,84626,84710,84755,84799,84816,84817,84906,84930,84952,84957,85037,85076,85141,85285,85459,85468,85499,85571,85593,85629,85798,85904,85921,85950,86038,86177,86278,86283,86288,86315,86356,86466,86645,86811,86838,86887,86916,86936,87071,87204,87231,87369,87547,87593,87741,87792,88014,88165,88168,88193,88250,88325,88374,88566,88576,88578,88591,88720,88723,88809,89006,89050,89095,89176,89230,89356,89517,89578,89625,89760,89821,89862,89982,90061,90129,90272,90274,90304,90333,90360,90482,90686,90704,90726,90764,90767,90799,90841,90921,90967,91132,91262,91286,91387,91463,91496,91577,91674,91680,91747,91785,91805,91902,91956,92014,92028,92041,92096,92234,92236,92262,92265,92573,92714,92728,92732,92809,92832,92834,92872,92873,93018,93425,93944,94136,94173,94202,94242,94292,94405,94454,94476,94479,94514,94587,94699,94700,94929,94982,95078,95128,95340,95431,95458,95531,95586,95611,95614,95673,95809,95843,96062,96097,96148,96207,96268,96327,96350,96477,96486,96499,96507,96551,96552,96586,96652,96668,96683,96720,96883,97048,97228,97251,97329,97687,97701,97704,97879,98106,98174,98289,98303,98484,98563,98571,98660,98677,98731,98734,98820,98941,99075,99175,99218,99287,99391,99698,99713,99903,99930",
          "5,9,23,26,250,313,506,518,624,628,667,950,1266,1278,1288,1289,1292,1302,1354,1370,1458,1463,1510,1522,1586,1809,1811,1813,1836,1844,1848,1899,2081,2242,2247,2250,2291,2296,2420,2478,2517,2692,3183,3389,3481,3528,3696,3779,3812,3832,3836,3869,4377,4691,4717,4724,4814,4847,4911,4947,5115,5124,5141,5282,5329,5384,5390,5398,5415,5423,5456,5479,5547,5658,5718,5808,5952,5989,6032,6056,6110,6189,6194,6215,6223,6232,6255,6279,6288,6353,6376,6420,6421,6443,6450,6465,6473,6538,6558,6560,6574,6602,6661,6707,6727,6737,6860,7145,7206,7365,7370,7426,7432,7492,7576,7595,7597,7612,7749,7763,7782,7816,7855,7890,8033,8062,8074,8180,8249,8251,8255,8332,8483,8530,8561,8629,8684,8956,8985,9085,9326,9337,9369,9408,9470,9771,9868,9902,9982,9999,10323,10330,10344,10415,10497,10553,10567,10662,10694,10696,10841,10942,11067,11297,11312,11351,11584,11675,11785,11968,11979,12439,12676,12749,12821,12824,12857,12941,13004,13831,13984,14120,14133,14225,14399,14430,14537,14683,14887,14942,14960,15059,15225,15308,15546,15590,15783,15794,15971,16060,16766,16856,17242,17468,17643,17668,17688,17698,18110,18139,18180,18569,18747,18994,19518,19796,19865,19967,20037,20167,20803,21018,21093,21108,21479,21531,21548,21648,22319,22601,22718,22829,22865,23191,23288,23311,23630,23854,24077,24725,24748,24846,25032,25169,25359,25493,25666,26244,26360,26629,26632,26685,26723,26733,26768,26779,26839,26862,26876,26889,27015,27081,27137,27175,27322,27371,27393,27620,27702,27819,28095,28269,28394,28437,28486,28715,28851,28874,29245,29340,29342,29472,29533,29535,29588,29662,29759,29775,29864,30659,31009,31647,31790,31879,32343,32401,33068,33909,35338,35466,35634,35807,36137,36309,36318,36619,36679,37012,37128,37187,37193,37308,37310,37384,37411,37450,37763,37868,37941,38059,38312,38315,38364,38372,38466,38559,38748,39065,39254,39347,39542,39733,39738,39777,39988,40551,40783,42143,42546,42691,42911,42916,42984,43151,43253,43315,43336,43495,43498,43627,43643,43734,43749,43812,44421,44872,45007,45030,45123,45270,45316,45510,45511,45563,45608,45685,45790,46600,47354,47377,49089,49225,50069,50198,50256,50277,50394,50568,50586,50626,50745,51024,51026,51184,51398,51403,51503,51527,51553,51848,51904,52035,52139,52219,52266,52368,52407,52455,52502,52509,52632,52696,52794,52876,52940,53003,53081,53121,53182,53324,53398,53509,53527,53571,53574,53586,53611,53658,53875,54102,54170,54253,54303,54381,54414,54451,54535,54602,54793,54890,54904,54907,55097,55222,55260,55262,55324,55419,55437,55516,55596,55807,56003,56009,56022,56054,56339,56501,56729,56826,56902,57045,57056,57249,57304,57379,57388,57411,57462,57684,57751,57794,57839,58220,58222,58255,58387,58481,59035,59089,59395,59518,59726,59792,59838,59989,60100,60232,60239,60243,60303,60435,60487,60548,60572,60574,60675,60962,61024,61182,61246,61276,61317,61478,61538,61672,61783,61839,61884,61890,62050,62103,62122,62167,62390,62543,62549,62621,62668,62723,62760,62870,62886,63067,63111,63170,63237,63257,63306,63358,63365,63416,63528,63540,63638,63820,63839,63852,63902,63938,64066,64080,64197,64261,64276,64334,64417,64517,64526,64554,64575,64658,64714,64788,64906,64942,64989,65081,65190,65244,65325,65594,65635,65764,65893,65924,65956,66162,66373,66414,66425,66446,66554,66850,67014,67172,67315,67355,67453,67516,67521,67609,67781,67857,67867,68099,68130,68136,68329,68330,68355,68369,68575,68591,69058,69077,69151,69203,69297,69349,69380,69468,69550,69611,69715,69745,69800,69897,69909,69925,69975,70033,70097,70159,70169,70204,70300,70364,70468,70480,70568,70577,70633,70804,70810,70870,71079,71106,71120,71235,71303,71405,71414,71420,71439,71451,71556,71706,71834,71977,71982,72003,72217,72272,72546,72630,72714,72719,72805,73014,73138,73199,73300,73344,73367,73403,73561,73565,73706,73801,74072,74093,74140,74160,74178,74230,74358,74476,74695,74836,74995,75026,75071,75079,75238,75337,75359,75377,75451,75527,75624,75709,75723,75787,75858,75918,76194,76238,76285,76291,76343,76384,76459,76464,76631,76836,76890,76901,76913,76921,76948,76951,77051,77118,77197,77217,77236,77341,77344,77503,77575,77591,77641,77653,77683,77734,77745,77749,77916,78065,78106,78210,78453,78530,78576,78637,78641,78643,78646,78719,78729,78736,78941,78956,79107,79178,79216,79333,79350,79693,79798,79799,79838,79964,79974,79999,80011,80012,80072,80158,80200,80213,80303,80313,80399,80460,80667,80700,80767,80773,80892,80938,81117,81136,81216,81263,81276,81372,81453,81592,81647,81732,82068,82283,82345,82465,82577,82741,83182,83212,83305,83353,83557,83574,83770,83779,83842,83865,84061,84106,84285,84315,84348,84453,84676,84685,84722,84730,84744,84885,84954,85003,85057,85120,85127,85199,85255,85341,85458,85462,85599,85640,85941,86078,86135,86199,86236,86261,86291,86394,86548,86557,86614,86650,86664,86680,86738,86881,86962,87056,87165,87207,87446,87506,87546,87669,87721,87853,87862,87867,87896,87966,88011,88066,88070,88079,88116,88181,88339,88458,88492,88558,88601,88677,88758,88902,88966,88973,89053,89062,89090,89297,89393,89411,89456,89709,89719,89741,89778,89878,89930,89954,89967,90031,90124,90158,90243,90256,90286,90447,90526,90541,90616,90658,90833,90925,90933,91079,91157,91160,91210,91257,91517,91640,91672,91689,91777,91801,91995,92111,92112,92114,92115,92152,92351,92360,92375,92612,92729,92763,92810,92831,92893,92924,93103,93134,93163,93168,93182,93248,93424,93715,93764,93766,93769,93779,93901,93934,93936,94023,94144,94194,94272,94280,94293,94323,94400,94449,94455,94464,94862,94870,95185,95199,95260,95341,95348,95521,95529,95570,95630,95691,95934,95972,96103,96137,96241,96269,96270,96303,96411,96583,96664,96749,96854,97253,97349,97379,97387,97393,97569,97577,97715,97989,98019,98067,98083,98262,98392,98479,98486,98524,98537,98539,98631,98638,98706,98719,98842,98904,99042,99106,99251,99415,99521,99599,99711,99715,99785,99792,99796,99850,99933,99984",
          "123,183,193,202,537,598,623,707,872,891,1077,1135,1447,1601,1690,1739,1806,2067,2131,2194,2274,2372,2434,2450,2459,2575,2781,2820,2875,2936,3217,3309,3703,3905,4164,4205,4256,4284,4385,4402,4431,4448,4515,4573,4637,4734,5004,5057,5068,5243,5271,5294,5503,5521,5611,5696,5759,5772,5804,5861,5941,6019,6061,6073,6176,6187,6365,6399,6431,6434,6444,6463,6552,6576,6580,6739,6840,6869,6918,7046,7198,7283,7407,7427,7434,7533,7664,7678,7692,7811,7891,7920,7930,7999,8032,8040,8067,8080,8089,8122,8223,8278,8324,8413,8450,8486,8608,8630,8735,8792,8801,8859,8890,8907,8924,8980,9052,9199,9231,9248,9301,9368,9375,9398,9422,9426,9444,9504,9529,9585,9731,9797,9923,10242,10355,10370,10377,10413,10433,10595,10660,10672,10724,10775,10818,10847,10862,11016,11062,11105,11181,11194,11235,11287,11300,11314,11389,11440,11565,11682,11700,11745,11873,11954,11975,12032,12074,12084,12145,12177,12216,12404,12520,12563,12723,12742,12809,12840,12846,12876,12932,13001,13009,13010,13055,13072,13075,13147,13160,13218,13287,13396,13426,13439,13498,13511,13836,13896,13952,13970,14009,14084,14224,14555,14662,14681,14704,14714,14773,14880,14978,15294,15426,15611,15774,15779,15803,15973,16055,16537,16711,16724,16876,16977,16983,17104,17144,17165,17445,17597,17723,17749,17829,17950,18144,18466,18541,18546,18696,18784,18790,18795,18797,18805,19016,19080,19177,19185,19250,19279,19337,19670,19689,19784,19853,19993,20004,20176,20200,20253,20544,20563,20575,20734,20781,20831,20868,20974,21057,21225,21398,21418,21421,21512,21603,21609,21616,21699,21776,21853,21962,22053,22776,22823,22852,22870,23021,23031,23209,23249,23257,23269,23422,23527,23677,23678,23692,23850,24104,24136,24417,24663,24720,24886,24900,24966,25036,25066,25076,25176,25187,25193,25330,25389,25412,25612,25724,25767,25820,25845,26052,26111,26156,26289,26353,26696,26700,26903,26965,27052,27055,27286,27532,27762,27782,27860,27903,27927,28234,28277,28297,28334,28372,28474,28509,28676,28790,28792,28914,29067,29213,29273,29282,29415,29484,29647,29710,29798,29820,29911,29995,30592,31032,31329,31764,32265,32753,33246,33968,34324,35088,35301,35380,35501,35550,35565,35630,35632,35702,35801,35879,35888,35988,36002,36047,36185,36188,36246,36449,36570,36634,36640,36811,36816,36819,36828,36934,36939,37004,37057,37129,37351,37379,37387,37551,37568,37575,37668,37691,37761,37867,38064,38186,38255,38362,38629,38632,38690,38936,38940,38985,39056,39070,39121,39168,39245,39246,39255,39427,39433,39443,39536,39549,39565,39606,39740,39751,39809,39814,39815,39879,42150,42210,42303,42459,42475,42490,42495,42549,42559,42617,42789,42879,42944,42960,43150,43270,43279,43284,43478,43560,43563,43676,43908,44021,44135,44221,44286,44294,44340,44350,44386,44512,44521,44536,44609,44685,44746,44756,44816,44849,44850,44866,45081,45091,45505,45528,45547,45561,45661,45663,45845,45888,45890,46534,46625,46908,47533,47925,48014,48233,49466,49546,50009,50051,50070,50155,50182,50238,50294,50498,50819,50915,51104,51238,51257,51263,51298,51517,51530,51591,51609,51655,51725,51823,51893,52028,52043,52306,52515,52524,52545,52684,52859,52864,52915,53038,53062,53117,53125,53157,53230,53408,53415,53448,53508,53656,53776,53793,53795,53799,53862,53942,54053,54273,54302,54374,54422,54457,54503,54597,54663,54721,55025,55145,55332,55436,55468,55482,55813,55837,55867,55901,55903,55999,56131,56195,56218,56226,56429,56604,56670,56862,56995,57029,57094,57149,57211,57221,57288,57610,57612,57754,57826,58339,58342,58407,58608,58618,58699,58780,58960,58971,59107,59143,59400,59463,59475,59690,59790,59821,59833,60023,60049,60132,60364,60498,60503,60580,60617,60653,60749,60762,60834,60869,60999,61002,61012,61042,61259,61411,61511,61560,61638,61693,61886,62128,62211,62251,62403,62611,62617,62624,62694,62695,63113,63215,63323,63338,63476,63533,63561,63699,63841,63847,64175,64307,64322,64568,64632,64758,65028,65057,65102,65111,65246,65538,65572,65829,66212,66325,66392,66507,66559,66613,66644,66707,66780,66831,66997,67083,67102,67259,67373,67799,67897,67907,67941,68090,68135,68214,68215,68227,68297,68394,68436,68568,68859,68925,68944,68953,69110,69260,69272,69367,69397,69541,69668,69721,69755,69839,69894,69963,70125,70233,70429,70432,70433,70434,70479,70523,70769,70781,70812,70831,70851,70864,70884,70948,70961,71062,71083,71101,71425,71455,71568,71648,71715,71736,71889,71912,71978,71995,72064,72144,72158,72303,72309,72346,72540,72713,72801,72903,72933,73022,73051,73068,73163,73280,73290,73390,73622,73630,73691,73785,73789,73882,73902,73952,74007,74062,74154,74307,74512,74530,74693,74701,74714,74722,74734,74775,74829,74955,75083,75105,75310,75346,75368,75618,75657,75662,75735,75950,75964,75966,76053,76131,76405,76457,76554,76574,76608,76609,76659,76685,76790,76801,76813,76928,76959,76983,77000,77028,77065,77080,77085,77116,77377,77441,77450,77671,77842,78015,78107,78114,78170,78299,78577,78596,78633,78664,78782,79015,79053,79091,79251,79255,79373,79617,79654,79656,79681,79718,79849,79903,79934,79955,80009,80116,80273,80382,80435,80455,80475,80533,80907,80912,81082,81130,81428,81436,81446,81512,81597,81667,81730,81779,81807,81956,81988,82123,82157,82228,82252,82286,82382,82629,82633,82669,82685,82711,82724,82737,82833,82872,82919,82946,83066,83179,83221,83227,83241,83385,83709,83852,83947,84025,84111,84159,84362,84438,84578,84791,84891,85263,85476,85539,85801,85820,85893,86020,86050,86067,86155,86225,86322,86350,86435,86443,86467,86556,86695,86754,86790,86823,86917,87059,87065,87101,87185,87194,87228,87251,87297,87337,87377,87499,87588,87768,88199,88296,88302,88352,88442,88543,88590,88703,88821,88863,88963,88979,89143,89256,89464,89579,89694,89817,90068,90172,90712,90795,90800,90890,90894,91148,91167,91192,91240,91242,91245,91251,91270,91289,91359,91380,91427,91484,91578,91579,91628,91821,91938,91957,92190,92203,92329,92413,92458,92471,92954,92992,93051,93069,93165,93200,93298,93300,93581,93644,93653,93837,93841,93852,93887,93946,94030,94176,94178,94180,94348,94358,94402,94475,94536,94577,94634,94938,95056,95081,95110,95216,95235,95286,95289,95454,95480,95559,95575,95605,95644,95654,95820,95872,95883,95970,95989,96011,96090,96131,96337,96359,96404,96534,96645,96746,96775,96777,96786,96848,96911,96943,97113,97174,97362,97503,97553,97561,97614,97659,97855,97893,97926,97927,97953,98024,98129,98166,98258,98265,98287,98324,98342,98507,98909,98996,99029,99195,99205,99207,99539,99613,99696,99752,99827",
          "209,318,366,397,415,417,441,519,564,578,581,607,815,860,869,893,975,990,996,999,1038,1185,1249,1269,1346,1361,1422,1446,1459,1515,1578,1879,1890,3567,4676,5040,5090,5105,5202,5206,5247,5360,5447,5729,5826,5919,6082,6369,6406,6432,6486,6505,6509,6550,6565,6627,6652,6673,6674,6689,6724,6735,6826,7016,7053,7066,7073,7075,7193,7273,7291,7310,7345,7356,7461,7541,7593,7689,7709,7835,7898,7921,8081,8199,8281,8302,8401,8409,8468,8477,8547,8675,8747,8967,8990,9001,9058,9070,9088,9154,9185,9188,9380,9454,9785,9976,10002,10247,10279,10450,10461,10494,10572,10576,10632,10735,10790,10792,10917,10939,10979,11055,11142,11199,11286,11659,11687,11817,11820,11890,11974,11987,12063,12128,12197,12374,12521,12561,12689,12718,12870,12933,13127,13401,13525,13874,14406,14952,15051,15359,15679,15753,16139,16385,16718,16739,16858,16985,17010,17207,17326,17571,17684,18018,18034,18483,18685,18701,18942,19547,20011,20538,20727,21070,21260,21285,21543,21598,21634,21981,22038,22119,22192,23340,23416,23564,24766,24818,24895,25172,25796,26413,26796,26822,26835,26851,26961,26977,27074,27216,27690,28412,28475,28697,28835,29113,29748,29845,36114,36193,36195,36196,36385,36439,36571,36734,37558,37637,37639,37672,37698,37799,38005,38297,38509,38610,38677,38921,39068,39111,39440,39442,39882,45379,45488,45615,45855,45878,49943,50092,50165,50175,50219,50775,50878,50897,51120,51220,51518,51557,51564,51607,51620,51653,51753,51787,51831,51856,52021,52106,52156,52243,52249,52260,52324,52331,52411,52518,52551,52574,52710,52755,52844,53009,53013,53071,53094,53098,53130,53241,53457,53544,53593,53731,53753,54013,54316,54332,54565,54595,54833,54991,55020,55081,55321,55327,55341,55434,55651,55653,55658,55697,55783,55818,55850,55900,56011,56044,56059,56296,56358,56444,56498,56519,56543,56584,56591,56611,56640,56662,56765,56936,57093,57178,57183,57252,57330,57395,57626,57729,57811,57859,57909,58034,58181,58296,58341,58485,58639,58753,58800,58860,58975,59038,59043,59055,59068,59105,59277,59286,59411,59555,59605,59607,59721,59853,59855,60079,60135,60152,60164,60211,60227,60462,60527,60537,60599,60605,60619,60663,60664,60694,60787,61256,61284,61387,61454,61515,61563,61590,61617,61628,61636,61722,61745,61811,61813,61853,62014,62109,62297,62309,62342,62361,62364,62467,62479,62584,62707,62743,62747,62748,62887,63007,63073,63158,63204,63232,63369,63425,63493,63560,63571,63593,63644,63695,63731,63764,63770,63775,63849,63851,64040,64083,64093,64118,64413,64704,64768,64905,64979,64980,65001,65032,65103,65151,65202,65410,65493,65523,65637,65702,65767,65798,65804,65812,66000,66010,66145,66146,66235,66548,66676,66983,67098,67147,67279,67448,67502,67642,67820,67871,67915,68012,68033,68049,68133,68296,68354,68376,68393,68398,68433,68463,68651,68697,68718,68763,68913,68958,68999,69044,69132,69276,69374,69385,69463,69498,69530,69531,69614,69672,69765,69855,70010,70023,70105,70192,70193,70546,70604,70666,70693,70784,70913,70945,70971,70987,71066,71172,71200,71259,71380,71394,71424,71525,72400,72575,72677,72793,72875,72881,73237,73466,73513,73566,73607,73633,73650,73675,73856,73917,73997,74074,74225,74233,74249,74367,74394,74448,74663,74747,74871,74887,74980,75027,75125,75228,75334,75389,75447,75452,75459,75580,75590,75667,75776,75802,75849,75850,75868,75898,76006,76068,76141,76244,76251,76324,76390,76421,76518,76535,76555,76859,76910,76987,77064,77183,77212,77468,77496,77627,77652,77680,77829,77858,78262,78278,78359,78471,78486,78506,78513,78540,78723,79034,79055,79085,79124,79514,79569,79610,79622,79757,80036,80126,80230,80272,80309,80334,80463,80501,80771,80835,80922,81001,81053,81188,81278,81312,81351,81435,81586,81759,81778,81881,81892,81894,82050,82059,82101,82234,82363,82488,82491,82624,82631,83046,83190,83242,83384,83430,83487,83520,83534,83675,83823,83854,83941,84069,84086,84122,84288,84371,84456,84488,84490,84544,84616,84643,84666,84827,84863,84882,85187,85326,85409,85516,85575,85577,85637,85712,85894,86006,86059,86211,86269,86285,86292,86301,86460,86547,86594,86736,86829,86896,86950,87018,87182,87253,87320,87464,87559,87569,87720,87754,87774,87798,87918,87987,87999,88122,88136,88231,88243,88319,88391,88551,88711,89235,89300,89322,89400,89580,89596,89598,89610,89615,89642,89648,89776,89807,89920,89959,90008,90063,90136,90196,90342,90420,90458,90481,90484,90637,90828,90840,90937,91012,91221,91296,91351,91356,91384,91531,91546,91576,91769,91812,91836,91879,91882,92036,92095,92198,92238,92279,92281,92334,92340,92397,92489,92554,92710,92784,92823,92869,92894,93028,93129,93304,93365,93381,93435,93540,93578,93605,93698,93775,93811,93860,93953,94034,94063,94266,94300,94313,94376,94394,94425,94495,94545,94568,94694,94816,94835,94953,95203,95299,95441,95505,95642,95650,95670,95885,95892,95906,95952,96024,96077,96236,96256,96311,96319,96378,96383,96410,96581,96600,96698,96799,96824,96835,96930,97058,97071,97140,97178,97212,97225,97298,97457,97674,97727,97738,97756,97814,97824,97887,97906,97914,97936,97939,98087,98210,98299,98374,98395,98449,98583,98592,98597,98666,98783,98789,98949,98955,98962,99000,99023,99313,99354,99458,99590,99644,99645,99656,99692,99718,99781,99795,99811",
          "4,269,314,394,401,410,484,491,524,574,696,712,720,748,761,765,887,899,925,933,985,1018,1071,1104,1212,1328,1466,1487,1526,1624,1724,1794,1807,1857,1909,1915,2925,2930,2966,3186,3413,3507,3587,3608,3917,4237,4247,5135,5214,5239,5240,5258,5307,5322,5410,5450,5562,5586,5645,5724,5770,5829,5870,5886,5993,6037,6099,6108,6141,6151,6228,6320,6339,6820,6893,6998,7117,7127,7134,7164,7220,7326,7338,7451,7501,7554,7573,7577,7611,7615,7638,7998,8039,8046,8131,8182,8189,8191,8198,8211,8227,8325,8391,8425,8439,8542,8577,8588,8663,8798,8987,9072,9179,9214,9236,9336,9345,9636,9904,10001,10171,10195,10224,10282,10333,10354,10389,10395,10402,10421,10427,10477,10479,10511,10528,10571,10585,10709,10879,10897,11107,11207,11288,11591,11710,11713,11715,11757,11819,11929,12732,12838,12953,13626,15145,15478,15790,15865,16315,17731,17857,18843,19192,19989,20066,21275,22740,22831,22855,22937,23089,24259,24286,24881,25130,25443,25743,26194,26405,27057,27184,27374,27847,28322,28349,28553,28721,30273,32019,32226,34139,35020,35561,36073,36171,36382,36700,37184,37613,39423,40385,42476,44693,45631,48228,48269,49417,49481,49580,50020,50109,50160,50295,50369,50557,50595,50752,50761,50799,50800,50824,50920,50984,51072,51085,51367,51497,51595,51715,51732,51751,51786,51846,51965,51979,51985,52011,52116,52154,52193,52300,52334,52356,52718,52778,52796,52843,52926,52929,52967,53031,53114,53128,53144,53149,53578,53654,53949,54069,54089,54185,54481,54505,54525,54629,54678,54682,55018,55070,55075,55103,55146,55370,55442,55478,55568,55604,55691,55801,56166,56265,56469,56542,56579,56620,56755,56845,57064,57309,57353,57453,57579,57607,57627,57632,57721,57732,57889,57991,58018,58128,58486,58538,58552,58885,59072,59102,59213,59340,59341,59351,59600,59630,59765,59904,59931,59936,59941,59951,60012,60131,60177,60308,60511,60610,60724,60874,60912,61181,61222,61299,61326,61356,61409,61424,61501,61581,61597,61607,61660,61712,61823,61914,61971,62134,62184,62344,62359,62387,62452,62526,62625,62633,62681,62714,62734,62892,62973,63033,63147,63264,63345,63423,63551,63637,63641,63658,63832,63843,63864,63887,63892,63975,64132,64209,64249,64323,64331,64461,64684,64778,64871,64912,64939,65022,65140,65458,65476,65527,65546,65583,65591,65913,65945,65987,66009,66080,66220,66241,66265,66279,66297,66384,66433,66516,66572,66611,66695,66846,67239,67332,67376,67378,67398,67447,67536,67566,67627,67637,67833,67834,67906,67940,67942,67996,68275,68292,68492,68537,68605,68773,68825,69035,69060,69085,69130,69266,69274,69447,69493,69511,69540,69570,69703,69719,69776,69850,69877,69883,69966,69981,70158,70163,70389,70440,70471,70592,70698,70737,70755,70827,70893,70952,70997,71135,71224,71242,71256,71263,71323,71437,71463,71509,71551,71572,71583,71642,71705,71777,71881,72178,72277,72386,72529,72577,72589,72703,72716,72777,72784,72920,73045,73102,73303,73491,73552,73584,73620,73645,73698,73732,73921,74015,74116,74138,74229,74239,74816,74817,74858,75053,75060,75182,75194,75200,75216,75278,75320,75426,75458,75497,75572,75654,75817,75929,76072,76082,76106,76182,76205,76214,76321,76322,76332,76643,76768,77066,77388,77467,77492,77543,77621,77622,77660,77773,77821,77833,77902,78068,78090,78131,78353,78423,78489,78564,78626,78667,78686,78921,79027,79276,79325,79527,79763,79766,79770,79800,79817,79890,79991,80259,80467,80643,80812,80878,80881,80906,80916,80987,81143,81407,81488,81554,81666,81689,81741,81815,81957,82009,82053,82070,82111,82377,82464,82597,82602,82635,82646,82653,82690,83075,83343,83620,83632,83637,83706,83777,83913,84040,84126,84207,84230,84279,84361,84387,84501,84534,84556,84568,84663,84861,84904,84933,84947,85079,85130,85186,85297,85431,85463,85473,85540,85553,85595,85825,85851,86042,86091,86112,86257,86276,86387,86399,86428,86439,86518,86529,86561,86565,86720,86813,86852,87009,87309,87439,87520,87543,87587,87815,87860,87928,88177,88343,88495,88609,88690,88692,88700,88800,88893,89007,89023,89034,89035,89082,89279,89320,89404,89490,89540,89543,89779,89787,89844,89869,89874,89911,90106,90293,90387,90575,90626,90694,91144,91308,91325,91497,91560,91670,91741,91949,91954,92104,92142,92226,92317,92662,92767,92838,92849,92878,93047,93156,93291,93374,93431,93592,93614,93691,93721,93798,93986,94010,94127,94285,94342,94364,94451,94617,94836,94858,95006,95092,95104,95126,95129,95148,95284,95310,95394,95416,95423,95434,95466,95484,95635,95647,95718,95755,95834,95908,95924,96059,96133,96322,96342,96469,96518,96607,96713,96729,96887,96905,96981,96987,97149,97241,97263,97307,97334,97360,97373,97383,97473,97478,97501,97601,97753,97794,97967,98066,98266,98373,98501,98574,98621,98787,98992,99072,99119,99177,99179,99293,99325,99336,99442,99489,99492,99493,99499,99510,99637,99640,99728,99797",
          "92,283,508,517,531,569,629,746,753,755,840,958,1132,1144,1223,1235,1351,1418,1439,1508,1517,1577,1847,1864,1943,2298,2494,2497,2563,2765,2803,2853,2937,2977,3083,3129,3161,3195,3229,3300,3390,3579,3687,4157,4264,4281,4430,4658,4700,4755,4775,4824,4842,5032,5089,5101,5109,5167,5188,5249,5297,5303,5318,5319,5362,5363,5413,5421,5431,5433,5436,5452,5460,5470,5482,5511,5625,5693,5714,5715,5738,5758,5773,5781,5790,5875,5926,5935,5953,5956,5984,5995,6106,6121,6200,6203,6216,6226,6250,6316,6366,6378,6408,6489,6504,6651,6710,6783,6842,6867,6880,6888,6931,6942,7004,7017,7052,7133,7184,7269,7312,7385,7398,7440,7455,7727,7734,7754,7852,7860,7918,8099,8106,8109,8233,8263,8283,8430,8523,8568,8575,8601,8625,8739,9209,9263,9355,9446,9589,9679,9908,9960,10213,10361,10458,10459,10480,10546,10608,10631,10923,10959,11014,11407,11582,11749,11802,11834,12206,12309,12344,12416,12491,12508,12585,12586,12967,13066,13148,13349,13447,14467,14472,14735,15301,15388,15570,15601,15612,15802,16084,16410,17034,17226,17512,17548,17560,17911,18025,18519,18604,18874,19182,19184,19428,19710,19751,20109,20533,20700,20798,21130,22081,22115,23054,23604,23794,23934,25625,25719,25846,25851,26190,26254,26310,26686,27024,27167,27282,27288,27452,27765,27800,27831,28213,28275,28443,28682,28716,28735,28896,29482,29805,29951,30574,30828,31123,31444,32053,32807,33354,33619,33648,35379,35613,35792,35820,35889,36308,36678,37060,37570,37752,38810,38863,39183,39256,39370,39477,39734,39805,39811,39990,40638,41450,41478,41507,42464,42528,42705,42873,43371,43493,43589,43681,44262,44497,44621,45307,45625,45923,46283,46724,46872,47411,49566,50105,50269,50398,50422,50448,50609,50610,50634,50699,50803,50809,50877,50900,50910,51004,51060,51113,51459,51513,51550,51659,51674,51685,51900,52017,52049,52152,52387,52441,52471,52672,52856,52945,53279,53315,53492,53516,53649,53664,53783,53932,53939,54059,54216,54235,54320,54614,54639,54670,54701,54763,54828,54836,54862,54927,55083,55175,55304,55411,55520,55569,55633,55714,55827,55835,55872,56033,56191,56294,56431,56509,56679,56683,56747,56835,56938,57033,57085,57131,57291,57372,57438,57456,57533,57563,57565,57652,57779,57781,57798,57842,57911,57966,58155,58226,58239,58269,58453,58461,58502,58559,58571,58578,58794,58840,58882,58941,59044,59070,59114,59136,59158,59236,59296,59377,59553,59624,59641,59653,59769,59857,59858,59955,60105,60109,60204,60237,60331,60479,60575,60584,60603,60735,60980,61008,61015,61254,61421,61492,61504,61651,61837,61926,62271,62300,62388,62396,62427,62440,62627,62631,62890,62981,63001,63005,63030,63182,63308,63312,63363,63650,63744,63842,63855,63925,64145,64242,64434,64525,64529,64569,64629,64693,64726,64727,64736,64802,64916,65041,65155,65259,65482,65499,65568,65592,65663,65814,65842,65897,65914,65944,65972,66015,66071,66073,66103,66183,66252,66413,66470,66589,66789,67142,67225,67256,67496,67519,67561,67692,67803,67816,67989,68141,68157,68222,68291,68375,68385,68406,68452,68464,68589,68644,68696,68722,68781,68830,68843,68864,68989,68994,68996,69039,69043,69059,69093,69243,69298,69395,69476,69537,69588,69643,69683,69774,70026,70042,70253,70450,70459,70484,70508,70709,70740,70868,70929,70943,71105,71211,71332,71359,71367,71403,71460,71619,71666,71787,71810,71894,72078,72135,72293,72311,72550,72590,72695,72741,72749,72794,72918,72949,73105,73386,73410,73502,73529,73545,73715,73840,73870,74047,74083,74094,74097,74237,74259,74280,74300,74342,74354,74456,74533,74567,74641,74678,74709,74728,74800,74945,75066,75193,75469,75485,75927,75973,76009,76054,76055,76123,76165,76488,76517,76575,76653,76793,76822,76879,76925,77100,77243,77271,77843,77861,78081,78395,78480,78527,78699,78718,78767,78775,78820,78874,78895,78901,79026,79214,79269,79511,79596,79611,79774,79901,80113,80178,80216,80261,80276,80332,80491,80498,80632,80662,80674,80815,80858,80877,80920,80965,81099,81212,81303,81318,81458,81545,81697,81708,81749,81822,81824,81984,82001,82003,82014,82215,82263,82271,82415,82606,82870,82910,82951,83042,83060,83127,83171,83290,83342,83453,83726,83763,83776,83801,83864,83866,83945,84074,84309,84345,84354,84365,84424,84633,84718,84752,84793,84830,84845,84916,85146,85158,85182,85353,85430,85572,85628,85657,85665,85686,85787,85793,85876,86103,86184,86329,86462,86493,86502,86619,86631,86840,86858,86941,87053,87083,87109,87217,87226,87296,87300,87405,87583,87596,87702,87797,87919,87938,87989,88078,88112,88413,88433,88542,88618,88793,88819,88978,89047,89049,89144,89376,89467,89483,89494,89512,89513,89558,89716,89748,90003,90165,90209,90269,90281,90376,90395,90422,90431,90512,90563,90587,90673,90708,90745,90811,90966,91034,91202,91236,91309,91400,91434,91572,91819,92032,92098,92110,92113,92120,92121,92161,92312,92355,92391,92424,92433,92480,92485,92929,92942,92964,93017,93025,93040,93297,93333,93361,93364,93385,93433,93555,93632,93648,93842,93941,94002,94079,94145,94204,94219,94297,94395,94491,94500,94510,94560,94569,94644,94661,94719,94772,94787,95091,95147,95191,95411,95637,95639,95669,95705,95721,95804,95812,95863,95897,96192,96351,96471,96497,96564,96590,96596,96701,96747,96773,96914,97220,97290,97297,97406,97484,97572,97898,97928,97942,98095,98198,98212,98217,98413,98437,98454,98551,98622,98651,98841,98843,98848,98888,98911,98984,99079,99159,99216,99281,99317,99332,99339,99409,99515,99641",
          "63,325,353,429,476,554,588,717,788,857,876,881,930,942,1039,1091,1154,1252,1279,1295,1489,1530,1579,1602,1604,1637,1686,1758,1863,1896,2225,4214,5006,5055,5059,5063,5187,5266,5325,5418,5576,5614,5667,5703,5817,5837,5911,5945,5959,5981,6074,6284,6299,6304,6317,6323,6326,6334,6362,6447,6461,6590,6730,6780,6803,6875,6978,7049,7081,7139,7270,7336,7433,7443,7468,7656,7665,7726,7815,7823,7842,7866,8017,8061,8119,8124,8134,8161,8181,8215,8243,8277,8358,8380,8453,8460,8539,8657,8709,8716,8722,8727,8738,8753,8874,8922,8926,8954,8989,8997,9174,9183,9221,9350,9765,9885,10138,10159,10160,10164,10320,10336,10468,10489,10566,10587,10740,10781,10878,11371,11383,11436,11467,11609,11678,11751,11859,11962,12070,12136,12173,12373,12489,12610,12719,13116,13221,13268,13371,13413,13481,13495,13529,13911,14083,14097,14595,14599,14699,14764,14861,14881,14906,14968,15163,15239,15315,15408,15453,15488,15523,15766,15887,16111,16306,16646,16690,16733,17344,17346,17416,17521,17613,17666,18238,18865,19403,19470,19517,20100,20221,20542,20549,20588,20628,21033,21127,21192,21528,21672,21745,22107,22707,22805,22923,23112,23139,23263,23272,24146,24159,24176,24605,25145,25205,25219,25458,26281,26799,27061,27541,27834,28252,28362,28439,28819,28938,29001,29006,29134,29159,29209,29615,29793,29938,34203,35752,35870,35894,35941,36575,36576,36636,36695,36824,36923,36948,37017,37431,38010,38236,38278,38571,38939,38997,39069,39117,39132,39488,39607,39617,39674,39940,39983,41282,45421,45621,45807,45990,50037,50099,50147,50196,50201,50264,50297,50313,50393,50546,50653,50710,50805,50882,50976,50999,51493,51665,51666,51696,51745,51936,51964,52079,52089,52114,52322,52333,52543,52596,52599,52627,52683,52729,52765,52772,52774,52823,52974,52990,53002,53141,53195,53468,53493,53562,53683,53722,53788,53801,53893,53900,53985,54009,54024,54037,54120,54218,54228,54242,54509,54557,54601,54659,54669,54771,55096,55100,55114,55125,55131,55138,55425,55546,55562,55593,56053,56057,56132,56170,56256,56355,56506,56623,56664,56701,56797,57034,57102,57452,57483,57547,57634,57665,57667,57782,57852,57951,58044,58179,58215,58270,58273,58278,58397,58401,58444,58556,58596,58767,58848,59073,59179,59187,59301,59347,59615,59737,59787,59940,60098,60110,60196,60289,60328,60343,60385,60538,60698,60853,60883,60994,61030,61139,61177,61223,61323,61357,61359,61541,61619,61697,61747,61764,61768,61891,61908,62024,62046,62080,62085,62204,62284,62314,62437,62492,62517,62595,62614,62809,62812,62831,62880,62931,62958,63074,63277,63303,63372,63409,63542,63663,63702,63745,63756,63757,63758,63844,64165,64298,64374,64540,64581,64596,64685,64835,64848,64866,64915,64985,65002,65045,65072,65186,65213,65324,65395,65497,65518,65672,65726,65786,65817,65824,65868,66177,66207,66368,66382,66429,66499,66643,66688,66728,66765,66816,66844,66847,66899,66942,67101,67138,67276,67329,67443,67459,67476,67534,67654,67790,67854,67901,67972,67985,68019,68128,68146,68357,68359,68456,68484,68498,68543,68546,68583,68604,68680,68761,68768,68777,68787,68882,68917,68920,69066,69102,69146,69165,69186,69218,69451,69486,69506,69533,69547,69619,69742,70002,70081,70256,70265,70353,70374,70425,70493,70593,70647,70695,70734,70779,70822,70989,71021,71168,71547,71552,71569,71604,71632,72033,72331,72332,72351,72436,72514,72720,72739,72829,72912,73003,73024,73065,73089,73430,73461,73480,73517,73557,73599,73639,73728,73766,73812,73953,73954,74050,74219,74458,74464,74492,74505,74586,74617,74651,74777,74815,74916,74931,75174,75264,75282,75305,75317,75383,75489,75583,75641,75853,75876,75878,76269,76278,76283,76358,76392,76545,76581,76623,76753,76769,76967,77149,77253,77263,77300,77343,77445,77524,77658,77751,77997,78122,78125,78219,78253,78445,78451,78470,78523,78597,78600,78612,78676,78704,78728,78750,78888,79014,79221,79301,79315,79513,79910,80136,80234,80318,80472,80558,80724,80783,80882,81046,81065,81138,81182,81218,81357,81581,81660,81684,81859,81912,81985,82077,82090,82141,82342,82429,82441,82546,82673,82674,82759,83031,83088,83130,83300,83628,83692,83719,83754,83784,83871,83887,83985,84023,84241,84308,84324,84360,84412,84519,84542,84579,84590,84732,84775,84905,84940,85170,85201,85247,85260,85274,85307,85327,85488,85602,85639,85666,85702,85812,85814,85948,85990,86209,86517,86534,86757,86786,86875,86901,86912,86974,87035,87086,87106,87241,87263,87308,87318,87324,87435,87475,87483,87487,87505,87617,87636,87789,87847,87866,87889,87939,88064,88105,88109,88135,88237,88359,88474,88547,88606,88607,88628,88886,88889,88941,89189,89433,89468,89530,89829,89860,89952,90026,90057,90058,90217,90234,90283,90369,90388,90514,90561,90787,91008,91059,91379,91461,91541,91624,91648,91827,91844,91942,91975,91998,92164,92273,92332,92347,92394,92469,92547,92705,92744,92798,92864,93093,93197,93438,93440,93525,93568,93606,93658,93669,93705,93763,93889,93966,94008,94064,94154,94347,94392,94411,94423,94520,94606,94695,94698,94739,94919,94989,95103,95152,95281,95545,95584,95610,95700,95875,95888,95903,96093,96094,96363,96475,96493,96556,96592,96631,96822,96861,97068,97230,97244,97314,97315,97590,97766,97806,97808,97836,97958,98036,98153,98203,98286,98336,98468,98526,98606,98662,98792,98797,98913,99001,99054,99088,99149,99270,99551,99681,99723,99730,99809,99816,99898,99968,99989",
          "262,592,651,685,769,780,797,821,870,1258,1286,1343,1349,1407,1852,1877,1911,2165,2550,2554,2572,2605,2629,2666,3097,3149,3372,3581,3634,3769,4170,4276,4282,4388,4903,4988,5052,5064,5134,5158,5194,5331,5517,5555,5662,5820,5840,5892,5961,6018,6093,6149,6207,6340,6404,6470,6503,6508,6521,6611,6636,6711,6736,6775,6792,6797,6921,6954,6972,7176,7195,7368,7391,7399,7421,7572,7627,7824,7825,7841,7844,7976,8050,8083,8090,8102,8184,8280,8288,8355,8479,8555,8593,8599,8600,8658,8726,8740,8843,8900,9119,9270,9312,9342,9357,9367,9420,9458,9501,9540,9577,9608,9615,9638,9719,9758,9767,9817,9864,9875,9888,9895,9975,9989,10005,10039,10051,10189,10245,10288,10378,10483,10883,10955,11020,11070,11081,11085,11092,11252,11277,11339,11367,11671,11753,11772,11774,11793,11853,11964,11997,12013,12043,12127,12176,12199,12336,12400,12468,12582,12640,13013,13030,13216,13424,13431,13533,13545,13664,13685,13696,13890,13902,14001,14010,14011,14016,14062,14377,14445,14463,14491,14673,14679,14774,14871,14893,14957,15023,15052,15099,15166,15276,15347,15418,15573,15603,15613,15928,15946,16009,16082,16117,16375,16798,16847,17037,17062,17095,17249,17504,17565,17717,17770,17808,18069,18329,18340,18462,18567,18596,18660,18703,18841,18842,18877,18953,18975,19149,19308,19319,19584,19676,19687,19739,19821,19982,20050,20130,20155,20188,20335,20524,20532,20632,20647,20685,20732,20790,20894,20905,20972,21026,21152,21350,21396,21583,21593,21626,21703,21833,21931,22135,22182,22252,22501,22584,22768,23148,23242,23274,23391,23796,23812,24211,24344,24577,24651,24690,24754,24760,24820,24898,24903,24922,25088,25113,25522,25549,25571,25865,25966,26248,26278,26322,26419,26428,26743,26803,26886,26927,27136,27173,27194,27226,27247,27262,27283,27335,27395,27463,27496,27610,27664,27896,27954,27967,28258,28284,28355,28803,28869,29039,29068,29187,29204,29257,29378,29391,29489,29574,29635,29701,29778,29857,29867,30845,30874,30980,31078,31164,31328,31716,32314,32353,32354,32633,32760,32789,32865,33368,33398,33746,33851,34267,34876,35093,35162,35233,35542,35564,35582,35805,35885,35893,35944,35948,36010,36065,36183,36204,36552,36578,36622,36646,36798,37002,37069,37177,37189,37236,37436,37509,37862,37864,38003,38065,38068,38191,38225,38238,38253,38284,38348,38421,38448,38501,38694,38727,38743,38753,38756,38870,38881,38944,38950,38968,38994,39177,39304,39660,39746,39759,39935,40401,40553,40582,41247,41336,42224,42269,42585,42636,42644,42730,42880,42919,42939,42999,43006,43076,43090,43292,43343,43440,43444,43548,43591,43817,44068,44192,44222,44321,44552,44556,44847,44990,45126,45474,45614,45624,45629,45675,45690,45694,45776,45957,46592,46637,46718,46749,46803,47217,47270,47499,47638,47804,47895,48039,48094,48116,48164,48559,48591,48679,48911,48965,49098,49505,49610,50075,50131,50229,50320,50322,50363,50387,50411,50467,50482,50728,51009,51010,51143,51144,51163,51171,51174,51199,51229,51275,51319,51386,51453,51656,51849,51870,51889,52018,52052,52065,52109,52112,52148,52291,52327,52381,52408,52446,52449,52586,52743,52846,52925,52960,52993,53214,53223,53281,53399,53432,53465,53503,53651,53698,53914,53968,54029,54032,54039,54142,54340,54434,54439,54567,54589,54617,54803,54817,54920,54984,55086,55116,55186,55205,55242,55259,55282,55489,55575,55778,55842,55876,56017,56210,56235,56307,56400,56523,56528,56808,56847,56967,57096,57116,57144,57191,57234,57263,57280,57285,57310,57405,57454,57606,57713,57827,57829,57976,57988,58006,58164,58182,58338,58445,58489,58535,58560,58564,58645,58932,59030,59056,59080,59104,59171,59250,59302,59393,59606,59613,59786,59845,59971,60057,60066,60117,60286,60467,60684,60699,60788,60993,61038,61148,61384,61476,61495,61695,61740,61995,62026,62282,62337,62735,62742,62834,62896,62968,63130,63156,63289,63332,63396,63432,63511,63674,63693,64078,64084,64129,64192,64264,64271,64280,64399,64418,64446,64482,64577,64744,64822,64960,64966,65085,65101,65339,65596,65611,65634,65721,65793,65866,65869,66064,66239,66365,66391,66549,66677,66887,66941,66987,67011,67031,67071,67091,67157,67203,67296,67368,67409,67499,67606,67877,68035,68234,68240,68368,68533,68538,68551,68559,68602,68782,69118,69348,69411,69505,69581,69635,69673,69730,69820,69848,69895,70057,70164,70170,70175,70258,70400,70412,70420,70525,70535,70558,70594,70648,70707,70713,70743,70999,71058,71094,71337,71534,71667,71710,71785,71939,72026,72083,72099,72143,72416,72568,72645,72671,72728,72763,72792,72982,73029,73075,73080,73211,73273,73324,73330,73534,73642,73793,73837,73937,74242,74271,74375,74385,74396,74399,74430,74485,74491,74495,74499,74654,74671,74755,74810,74848,74986,75138,75327,75482,75568,75617,75708,75784,76093,76118,76136,76383,76444,76514,76898,76940,76975,77048,77057,77112,77131,77143,77314,77352,77435,77456,77460,77520,77588,77729,77765,77835,77900,78061,78472,78559,78823,78847,78909,78952,78980,78998,79110,79217,79278,79440,79474,79554,79721,79723,79802,79940,79961,80098,80130,80263,80288,80349,80356,80482,80495,80661,80796,80937,80961,81152,81229,81350,81371,81452,81461,81527,81540,81543,81561,81565,81604,81618,81630,81637,81657,81782,81994,82167,82355,82368,82372,82462,82477,82493,82523,82537,82655,82659,82809,82819,82927,83067,83168,83257,83332,83392,83498,83629,83666,83901,84073,84233,84267,84277,84280,84302,84351,84421,84427,84580,84698,84750,84777,84931,85096,85104,85116,85153,85164,85242,85246,85248,85362,85389,85426,85609,85616,85692,85710,85713,85837,85858,85992,86104,86156,86347,86402,86458,86696,86782,86816,86859,86991,87015,87151,87177,87196,87255,87614,87755,87777,87835,87925,87929,87934,88012,88267,88278,88288,88424,88501,88586,88729,88790,88808,89111,89242,89414,89480,89501,89505,89621,89623,89666,89678,89688,89841,89965,90018,90204,90207,90261,90300,90389,90517,90548,90556,90598,90675,90713,90727,90983,91045,91050,91085,91150,91179,91191,91234,91378,91412,91435,91443,91501,91502,91520,91525,91611,91662,91766,91946,92042,92151,92159,92194,92201,92205,92245,92296,92565,92615,92820,93044,93091,93164,93216,93225,93259,93266,93342,93608,93676,93717,93768,93965,93982,94056,94332,94586,94712,94758,94760,94832,94892,95123,95251,95285,95318,95350,95356,95429,95443,95563,95601,95628,95676,95746,95841,95937,95956,96068,96183,96186,96218,96233,96323,96326,96349,96455,96521,96540,96582,96591,96634,96809,97036,97124,97129,97399,97426,97542,97693,97711,97881,97959,97983,98200,98481,98816,98929,99187,99261,99418,99467,99475,99487,99658,99826,99859,99969,99999",
          "205,264,405,451,561,575,622,661,662,774,790,924,1114,1159,1160,1219,1253,1270,1283,1367,1403,1415,1575,1640,1719,1734,1751,1824,1883,1919,2044,2101,2181,2239,2553,2599,2604,2643,2688,2715,2854,2921,3091,3366,3611,3630,3868,4279,4309,4393,4417,4535,4555,4593,4725,4834,4918,4953,5039,5157,5160,5275,5341,5396,5528,5597,5727,5751,5822,5825,5967,6007,6039,6070,6078,6097,6205,6283,6454,6457,6520,6588,6723,6756,6757,6986,7012,7099,7254,7257,7362,7386,7422,7462,7485,7655,7700,7806,7829,7843,7853,7941,8194,8212,8321,8327,8406,8462,8712,8768,8819,8855,8857,8870,8955,8988,9020,9142,9195,9220,9224,9226,9228,9432,9499,9512,9901,10024,10311,10385,10502,10556,10563,10655,10691,10771,10780,10786,10800,10860,10985,10991,11037,11084,11163,11379,11380,11449,11667,11712,11806,11822,11867,11883,12021,12069,12174,12188,12191,12377,12406,12432,12637,12784,12962,12999,13173,13385,13556,13747,13811,13937,13975,14039,14359,14421,14436,14478,14712,14790,14849,14946,14976,15156,15170,15176,15558,15576,15585,15695,15767,15768,15784,15787,15797,15847,16099,16373,16435,16566,16623,16680,16720,16884,16953,16963,17002,17038,17066,17114,17168,17187,17318,17627,17690,17703,17793,17820,17858,17988,18147,18155,18277,18615,18713,18827,18970,19074,19158,19187,19198,19510,19535,19536,19644,19944,20067,20150,20290,20325,20374,20546,20689,20708,20719,20765,20913,21024,21029,21082,21316,21772,21826,21862,22156,22299,22593,22622,22642,23251,23517,23954,24125,24336,24429,24483,24660,24884,25178,25192,25214,25355,25417,25420,25593,25686,25718,25721,25885,25943,26009,26059,26085,26616,26677,26746,26748,26921,26967,26993,27002,27209,27260,27300,27306,27408,27429,27584,27772,27856,27858,27885,28069,28101,28180,28184,28193,28217,28220,28253,28292,28327,28389,28654,28761,28802,28806,28848,29173,29188,29376,29545,29744,29747,29771,29826,29836,29970,30085,30539,31593,31938,32202,32366,32714,32750,32808,32859,32895,32982,33420,33613,33652,33910,34058,34084,34432,34580,34751,34904,35128,35187,35699,35761,35802,35871,35949,35954,36452,36565,36623,36864,36940,37005,37188,37484,37563,37581,37628,37631,37693,37800,37858,38001,38188,38342,38349,38505,38557,38621,38692,38702,38759,38913,38976,39072,39223,39231,39558,39753,39816,39902,39923,40538,40775,40804,41071,41348,41772,41957,42001,42218,42238,42442,43573,43975,44183,44462,44545,44572,44637,44894,44964,45182,45250,45259,45430,45622,45794,45800,45882,46011,46223,46310,46489,47476,48047,48051,48099,48296,49124,49230,49837,49879,50152,50243,50258,50270,50303,50397,50406,50429,50434,50526,50533,50560,50624,50644,50684,50751,50856,50923,50934,50989,51003,51375,51422,51437,51531,51563,51576,51579,51678,51727,51737,51785,51797,52064,52221,52228,52285,52352,52453,52459,52552,52561,52655,52739,52776,52781,52815,52870,52902,53233,53389,53598,53671,53718,53984,54088,54180,54202,54249,54276,54380,54419,54442,54482,54544,54571,54577,54636,54648,54649,54797,55141,55224,55251,55273,55349,55377,55484,55522,55532,55559,55640,55806,55824,55829,56035,56192,56330,56416,56455,56518,56612,56737,56869,57005,57009,57126,57199,57208,57281,57337,57338,57369,57553,57590,57670,58301,58313,58434,58821,58835,58843,58870,58908,59041,59282,59320,59654,59669,59706,59771,59817,59823,59846,59915,60014,60042,60070,60210,60427,60432,60448,60499,60665,60928,61381,61393,61452,61572,61700,61792,61818,61972,62062,62148,62165,62334,62487,62496,62525,62565,62574,62638,62655,62709,62718,62774,62818,62861,62878,63094,63124,63145,63271,63603,63632,63848,63858,64002,64019,64042,64098,64174,64185,64286,64356,64488,64556,64620,64631,64665,64807,65012,65117,65212,65310,65388,65417,65429,65434,65553,65636,65806,65819,65959,66006,66047,66106,66285,66343,66443,66506,66513,66539,66727,66889,66894,66917,66945,66985,67278,67345,67396,67427,67480,67484,67558,67563,67644,67712,67736,67791,67808,67819,67826,67830,67845,67853,68014,68166,68565,68572,68797,68811,68891,68922,69007,69065,69137,69174,69210,69290,69368,69522,69601,69618,69621,69629,69690,69760,69768,69806,69838,70013,70028,70101,70144,70201,70266,70344,70435,70443,70449,70723,70797,70813,70930,70991,71045,71157,71305,71321,71351,71363,71629,71699,71817,71861,71927,71983,72047,72138,72216,72352,72415,72491,72494,72501,72750,72940,73209,73285,73551,73638,73681,73826,74017,74075,74203,74207,74245,74261,74371,74433,74449,74661,74683,74685,74785,74830,74874,74950,74977,74996,75003,75032,75059,75221,75225,75234,75272,75391,75399,75519,75529,75543,75702,75882,75891,75913,75935,76030,76220,76257,76281,76470,76511,76562,76590,76727,76821,76953,76977,76991,77031,77137,77140,77153,77371,77474,77532,77632,77695,77932,78064,78289,78376,78524,78551,78572,78623,78682,78730,78762,78786,78923,78950,78994,79097,79218,79388,79445,79590,79932,80104,80118,80164,80255,80345,80543,80555,80721,80733,80787,80795,80819,80944,80984,81009,81023,81073,81095,81123,81200,81302,81420,81492,81520,81621,81655,81760,81926,81927,81940,81964,81995,82191,82319,82325,82428,82485,82498,82701,82867,82963,82974,82976,83064,83084,83150,83184,83287,83569,83572,83591,83631,83732,83751,83903,83965,84444,84480,84876,84918,85066,85087,85252,85262,85299,85322,85355,85433,85507,85579,85680,85736,85870,86024,86084,86289,86624,86682,86700,86888,87002,87004,87017,87073,87123,87138,87183,87257,87334,87354,87372,87411,87611,87708,87733,87767,87897,87970,88084,88100,88113,88192,88323,88347,88506,88569,88633,88656,88666,88714,88782,88930,88983,89030,89043,89153,89175,89220,89406,89437,89506,89546,89575,89823,89866,90070,90071,90307,90581,90669,90835,90946,91005,91193,91200,91322,91373,91394,91433,91551,91569,91684,91873,91881,91947,92001,92024,92040,92081,92162,92209,92288,92398,92483,92697,92755,92785,92991,93083,93125,93176,93360,93455,93487,93520,93565,93584,93688,93711,93923,93929,94029,94123,94221,94233,94255,94291,94310,94391,94421,94424,94474,94867,94984,95012,95156,95376,95633,95775,95877,95938,96305,96474,96731,96860,96862,96961,96984,97078,97139,97281,97288,97351,97470,97492,97583,97719,97803,98016,98074,98147,98207,98238,98410,98529,98585,98588,98691,98717,98745,98785,98958,99296,99299,99355,99429,99451,99496,99585,99586,99631,99682,99975,99990,100000",
          "27,118,138,331,625,666,739,743,826,986,1197,1273,1313,1436,1462,1506,1594,2164,2295,2310,2323,2421,2445,2480,2616,2872,3224,3286,3305,3351,3399,3412,3520,3538,3661,3700,3763,4090,4105,4225,4241,4418,4443,4554,4610,4627,4682,4695,4737,4779,4815,5013,5077,5155,5195,5264,5268,5306,5317,5485,5663,5779,5806,5881,5905,5933,5978,6025,6026,6041,6161,6225,6247,6313,6322,6385,6403,6598,6716,6722,6770,6790,6815,6873,6885,6917,6953,7006,7157,7256,7267,7280,7281,7295,7563,7570,7663,7707,7728,7752,7767,7822,7906,7964,8007,8042,8079,8103,8193,8231,8362,8363,8527,8536,8628,8743,8758,8775,8853,8919,9105,9361,9390,9562,9571,9712,9732,9780,9839,9850,9915,9967,10044,10318,10342,10851,10861,10888,10905,10915,10929,10993,11013,11087,11102,11120,11145,11146,11295,11479,11620,11766,11824,11826,12260,12325,12445,12477,12601,12658,12662,12695,12762,12831,12839,12982,12996,13109,13208,13304,13337,13403,13451,13479,13567,13594,13740,13742,13849,13951,13971,14017,14051,14126,14172,14186,14312,14351,14617,14656,14689,14716,14729,14732,14748,14873,14939,15129,15185,15384,15414,15547,15664,15673,15680,15879,16002,16066,16086,16127,16128,16388,16421,16478,16631,16636,16722,16762,16913,16978,16996,17016,17359,17407,17409,17427,17482,17562,17685,17724,17768,17967,18070,18187,18203,18295,18845,18955,19005,19024,19027,19036,19041,19059,19083,19100,19194,19248,19316,19369,19376,19498,19706,19735,19763,19968,20352,20366,20718,20807,20904,21121,21299,21520,21854,21930,21950,22082,22125,22159,22177,22271,22315,22557,22596,22662,22939,22980,23159,23184,23280,23304,23331,23412,23465,23482,23508,23544,23575,23606,23787,23793,23848,23900,24192,24457,24476,24647,24661,24717,24859,24950,25065,25103,25230,25237,25275,25338,25483,25496,25541,25620,25681,25690,25733,25883,25959,25978,26172,26224,26239,26339,26393,26432,26702,26834,27031,27039,27212,27215,27225,27346,27349,27424,27668,27696,27758,27817,27886,27928,28030,28074,28123,28271,28302,28306,28398,28526,28543,28563,28740,28776,28845,28875,28906,28939,28966,28969,28971,29010,29034,29162,29234,29265,29295,29335,29381,29413,29528,29622,29757,29786,29843,29853,29943,29952,29959,29998,30120,30178,30288,30641,30711,30942,31042,31076,31125,31250,31337,31366,31455,31734,31890,32073,32238,32382,32469,32526,32972,32990,33251,33282,33310,33716,33929,33958,34007,34180,34248,34325,34464,34469,34614,34616,34644,34667,34992,35215,35563,35675,35714,35757,35865,36021,36213,36255,36257,36326,36359,36448,36464,36518,36645,36684,36697,36825,36863,36895,36952,37067,37121,37191,37208,37256,37397,37426,37438,37514,37524,37540,37635,37742,37798,37814,37875,37886,38074,38076,38266,38425,38481,38506,38852,39079,39080,39115,39338,39425,39451,39543,39672,39792,39877,39929,40230,40379,40459,40636,40752,40981,41133,41242,41249,42003,42005,42090,42096,42152,42179,42200,42271,42412,42684,42803,42953,43179,43318,43367,43413,43428,43575,43651,43789,44009,44165,44171,44291,44354,44495,44641,44662,44777,44895,44940,45131,45190,45201,45314,45365,45398,45456,45606,45611,45736,45869,45884,45902,46073,46457,46532,46539,46661,46789,46827,46877,47087,47419,47919,48096,48207,48252,48848,48952,48963,49004,49093,49112,49205,49351,49391,49392,49460,49493,49681,49928,50025,50144,50224,50250,50253,50298,50325,50565,50615,50736,51209,51246,51338,51355,51393,51433,51467,51469,51470,51541,51603,51769,51809,51952,52105,52120,52128,52222,52288,52400,52460,52476,52492,52579,52607,52620,52762,52922,53070,53073,53104,53226,53271,53348,53410,53459,53506,53577,53583,53625,53639,53653,53720,53837,53872,53882,53926,53957,54021,54159,54172,54186,54206,54252,54260,54418,54462,54467,54484,54534,54635,54730,54772,54837,54844,54884,54891,55185,55296,55563,55740,55798,55832,55871,55892,55955,56008,56023,56051,56108,56177,56185,56219,56242,56505,56613,56708,56821,56975,57182,57295,57327,57383,57400,57465,57490,57496,57538,57573,57708,57749,57770,57802,57841,57907,58023,58028,58081,58195,58225,58254,58268,58323,58511,58512,58518,58547,58916,58931,58958,59084,59329,59365,59417,59510,59589,59618,59710,59729,59735,59806,59901,60044,60093,60170,60173,60174,60223,60248,60274,60333,60335,60445,60477,60540,60639,60963,60971,61078,61111,61112,61128,61191,61207,61251,61353,61439,61556,61565,61723,61898,61933,62027,62031,62047,62213,62239,62369,62478,62623,62662,62833,62954,63053,63082,63119,63140,63177,63188,63279,63387,63558,63578,63660,63794,63874,63933,63966,63986,64001,64065,64186,64189,64208,64355,64381,64503,64694,64919,65023,65142,65192,65303,65453,65589,65622,65854,65863,65922,65932,65995,66070,66144,66199,66383,66396,66428,66525,66636,66777,66992,67000,67003,67111,67162,67286,67362,67492,67533,67669,67690,67695,67783,67794,67948,68182,68299,68319,68334,68445,68493,68497,68570,68708,68764,68771,68774,68793,68826,68877,69010,69025,69062,69244,69328,69932,69933,70032,70179,70302,70340,70348,70388,70395,70409,70448,70672,70789,70908,70920,70941,71005,71063,71069,71289,71318,71400,71498,71511,71626,71670,71738,71865,72198,72263,72264,72304,72342,72383,72419,72564,72603,72686,72704,72865,72957,73087,73092,73144,73320,73336,73339,73356,73380,73420,73439,73518,73564,73768,73771,73820,73859,73863,73879,73967,73985,74009,74266,74408,74409,74413,74486,74497,74595,74679,74882,75103,75162,75209,75288,75414,75597,75647,75698,76081,76090,76137,76235,76341,76435,76440,76497,76532,76557,76571,76596,76755,76848,76976,77026,77094,77141,77215,77434,77436,77449,77479,77534,77539,77582,77764,77767,77851,77911,77950,77952,78080,78232,78404,78409,78431,78460,78496,78656,78732,78902,78905,78930,78992,79106,79139,79181,79200,79201,79331,79365,79432,79821,79843,79852,79953,79987,80053,80054,80105,80134,80168,80295,80432,80564,80603,80627,80659,80801,80825,80884,80966,81045,81156,81159,81328,81394,81418,81450,81479,81659,81707,81796,81837,81879,81888,81903,81906,81967,81992,82029,82095,82255,82365,82720,82773,82814,82912,83061,83140,83248,83382,83714,83731,83853,83874,84014,84015,84055,84167,84364,84461,84495,84522,84638,84687,84837,84843,84871,84910,84984,85044,85134,85143,85526,85573,85582,85589,85596,85607,85613,85771,85877,85905,85964,85994,86015,86023,86071,86073,86167,86223,86302,86340,86362,86489,86810,86856,86867,86938,87069,87072,87236,87276,87321,87525,87585,87604,87769,87808,87965,87971,87988,88200,88253,88265,88382,88554,88598,88664,88715,88862,88871,88915,89196,89303,89348,89389,89498,89510,89552,89554,89581,89783,89843,89961,90189,90227,90399,90513,90580,90611,90628,90654,90741,90754,90786,90953,91074,91142,91153,91233,91507,91818,92073,92196,92206,92336,92393,92415,92466,92581,92664,92671,92684,92686,92797,92815,92886,93107,93116,93161,93179,93269,93279,93384,93446,93449,93513,93628,93713,93761,93772,93898,93958,94007,94246,94311,94535,94538,94643,94753,94781,94871,95011,95073,95181,95245,95450,95822,95829,95866,95933,95948,95957,95965,95978,96054,96092,96144,96177,96197,96553,96641,96707,96734,96769,96938,96993,97025,97050,97072,97138,97141,97332,97438,97443,97531,97564,97568,97969,97971,98136,98194,98282,98354,98515,98575,98741,98747,98810,98881,98887,98935,98980,99018,99033,99148,99285,99430,99491,99533,99564,99684,99836,99948",
          "45,50,64,106,115,159,169,221,1793,1922,1936,2079,2127,2251,2255,2270,2335,2341,2396,2477,2590,2662,2680,2706,2711,2767,2818,2832,2836,2882,2909,2956,3152,3206,3446,3522,3714,4098,4099,4191,4204,4390,4458,4601,4611,4747,4853,4866,4948,5544,6980,7949,7973,9404,9412,9476,9517,9545,9546,9621,9640,9678,9697,9716,9721,9808,9824,9944,10020,10077,10086,10090,10096,10210,10381,10590,10630,10647,10666,10689,10745,10837,10856,10951,11038,11046,11179,11197,11198,11203,11217,11236,11255,11293,11369,11445,11477,11540,11597,11654,11730,11882,11920,12015,12036,12072,12129,12186,12224,12243,12262,12338,12395,12414,12419,12471,12528,12547,12654,12661,12699,12710,12737,12758,12775,12813,12823,12889,12892,12916,12930,12965,12973,12984,13003,13022,13041,13060,13112,13117,13204,13223,13231,13250,13326,13363,13370,13381,13398,13402,13404,13421,13459,13516,13573,13599,13601,13606,13630,13649,13654,13687,13703,13706,13725,13736,13744,13763,13782,13820,13858,13885,13906,13934,13953,13978,14029,14048,14100,14105,14124,14162,14181,14200,14221,14238,14245,14276,14295,14308,14314,14333,14352,14435,14444,14448,14509,14519,14525,14534,14544,14580,14618,14620,14637,14675,14694,14747,14770,14827,14839,14846,14884,14909,14922,14935,14972,14979,15017,15102,15134,15147,15149,15169,15188,15207,15208,15264,15283,15302,15318,15321,15397,15416,15432,15479,15480,15509,15511,15530,15625,15644,15653,15666,15682,15701,15720,15740,15758,15760,15791,15815,15834,15853,15872,15891,15948,15967,15986,16043,16081,16124,16138,16157,16214,16225,16233,16252,16253,16268,16290,16309,16333,16347,16366,16404,16461,16499,16518,16575,16581,16613,16640,16651,16662,16670,16704,16746,16765,16770,16775,16784,16822,16947,16974,16993,17048,17088,17126,17164,17183,17221,17240,17275,17297,17312,17373,17392,17430,17449,17465,17506,17525,17561,17601,17608,17620,17715,17767,17772,17791,17810,17848,17886,17898,17924,17976,18038,18054,18057,18114,18171,18190,18209,18266,18285,18293,18311,18335,18342,18350,18361,18380,18437,18456,18469,18494,18525,18532,18646,18665,18684,18724,18817,18931,18943,18950,18988,19022,19032,19045,19064,19088,19121,19140,19178,19197,19233,19254,19259,19303,19311,19330,19364,19366,19387,19406,19444,19475,19481,19482,19501,19520,19539,19615,19663,19672,19691,19707,19729,19767,19786,19838,19862,19881,19888,19900,19938,19957,19995,20003,20014,20052,20128,20147,20223,20226,20299,20318,20333,20337,20356,20375,20388,20390,20394,20432,20439,20470,20481,20489,20498,20508,20587,20603,20641,20648,20679,20736,20755,20772,20793,20812,20850,20869,20909,20983,21002,21025,21045,21078,21090,21097,21116,21154,21173,21198,21211,21249,21268,21287,21325,21363,21382,21420,21439,21472,21477,21519,21534,21553,21566,21610,21629,21724,21743,21773,21838,21876,21893,21895,21914,21933,21952,21971,22009,22047,22054,22060,22066,22102,22104,22123,22142,22153,22199,22218,22274,22275,22294,22302,22332,22351,22364,22370,22389,22408,22410,22427,22446,22458,22465,22541,22560,22579,22619,22636,22639,22655,22693,22712,22731,22750,22753,22788,22826,22890,22898,22902,22997,23029,23073,23079,23092,23130,23146,23149,23187,23200,23206,23219,23225,23244,23282,23301,23358,23377,23386,23392,23396,23415,23434,23491,23510,23529,23567,23624,23643,23667,23681,23700,23719,23738,23750,23753,23757,23814,23845,23852,23871,23890,23902,23928,23947,23966,24023,24042,24061,24080,24118,24156,24175,24213,24222,24232,24270,24308,24346,24365,24384,24403,24410,24441,24460,24479,24494,24498,24517,24534,24536,24555,24574,24631,24669,24688,24707,24726,24764,24799,24840,24878,24897,24916,24927,24983,24992,25006,25011,25075,25125,25144,25189,25201,25220,25296,25315,25356,25360,25384,25391,25410,25448,25502,25543,25555,25562,25581,25619,25633,25657,25676,25695,25752,25771,25790,25793,25809,25847,25859,25866,25874,25942,25961,25980,25999,26020,26036,26037,26056,26075,26094,26113,26120,26170,26183,26246,26265,26303,26341,26379,26417,26436,26474,26493,26512,26623,26626,26645,26664,26740,26755,26892,26911,26949,26986,27042,27063,27246,27265,27272,27291,27310,27389,27436,27465,27671,27766,27804,27874,27877,28006,28070,28089,28127,28182,28196,28218,28241,28326,28338,28380,28447,28451,28462,28504,28507,28583,28611,28678,28694,28881,28962,29077,29225,29271,29323,29350,29433,29475,29558,29667,29670,29780,29851,29892,29894,29913,30007,30036,30065,30094,30123,30152,30210,30239,30268,30355,30384,30413,30442,30471,30500,30521,30529,30558,30587,30590,30616,30645,30674,30703,30725,30732,30761,30782,30790,30819,30848,30877,30906,30935,30964,30993,31022,31051,31080,31109,31138,31160,31167,31196,31225,31242,31254,31283,31312,31331,31341,31399,31428,31457,31486,31541,31544,31573,31602,31631,31660,31683,31689,31718,31747,31776,31805,31834,31863,31892,31921,31950,31966,31979,32008,32037,32058,32066,32095,32124,32149,32153,32182,32211,32217,32240,32269,32298,32327,32356,32385,32414,32443,32501,32530,32559,32588,32617,32646,32675,32704,32733,32762,32791,32820,32849,32878,32907,32936,32959,32965,32994,33023,33052,33081,33110,33139,33168,33185,33197,33226,33255,33284,33342,33371,33400,33429,33458,33487,33516,33545,33574,33603,33632,33661,33690,33719,33748,33777,33806,33835,33864,33881,33893,33922,33951,33980,34025,34038,34067,34096,34125,34154,34183,34200,34212,34270,34286,34299,34328,34357,34374,34386,34415,34444,34446,34465,34473,34502,34531,34560,34589,34618,34676,34705,34722,34734,34763,34792,34821,34850,34879,34908,34937,34966,34995,35050,35096,35140,35153,35170,35201,35228,35242,35259,35326,35333,35373,35397,35407,35443,35478,35512,35526,35585,35619,35624,35628,35644,35717,35738,35750,35776,35835,35906,35931,35965,36005,36024,36049,36050,36087,36147,36206,36277,36336,36358,36361,36395,36418,36427,36467,36526,36555,36585,36612,36649,36708,36731,36767,36838,36878,36897,36956,37028,37087,37146,37210,37257,37269,37328,37356,37359,37399,37458,37517,37541,37552,37589,37609,37648,37707,37771,37830,37889,37916,37919,37960,38019,38078,38112,38150,38209,38268,38292,38332,38391,38450,38521,38580,38626,38639,38770,38829,38893,38929,38934,38952,39011,39082,39141,39171,39200,39272,39331,39354,39390,39454,39513,39572,39643,39702,39761,39788,39833,39892,39951,40000,40001,40002,40003,40004,40005,40006,40007,40008,40009,40010,40011,40012,40013,40014,40015,40016,40017,40018,40019,40020,40021,40022,40023,40024,40025,40026,40027,40028,40029,40030,40031,40032,40033,40034,40035,40036,40037,40038,40039,40040,40042,40043,40044,40045,40046,40047,40048,40049,40050,40051,40052,40053,40054,40055,40056,40057,40058,40059,40060,40061,40062,40063,40064,40065,40066,40067,40068,40069,40070,40071,40072,40074,40076,40077,40078,40079,40081,40082,40083,40084,40085,40086,40087,40089,40090,40091,40093,40094,40095,40096,40097,40098,40099,40100,40120,40149,40178,40207,40236,40265,40294,40323,40352,40381,40410,40439,40468,40497,40526,40555,40584,40613,40642,40671,40700,40729,40758,40787,40816,40845,40874,40903,40932,40961,40990,41019,41048,41077,41106,41135,41164,41193,41251,41280,41309,41338,41367,41396,41425,41454,41483,41512,41541,42011,42040,42069,42098,42127,42146,42149,42187,42247,42277,42352,42403,42409,42427,42482,42514,42561,42563,42567,42589,42628,42664,42697,42751,42765,42796,42808,42826,42847,42901,42942,42947,42988,43042,43063,43138,43149,43153,43181,43187,43225,43236,43257,43259,43296,43300,43321,43375,43411,43421,43462,43490,43515,43537,43581,43612,43699,43712,43752,43757,43774,43806,43807,43849,43872,43896,43936,43969,44011,44042,44051,44086,44173,44219,44248,44268,44281,44293,44297,44323,44342,44377,44410,44440,44442,44485,44533,44560,44574,44629,44647,44698,44708,44722,44754,44775,44797,44817,44857,44884,44906,44918,44920,44959,45014,45034,45107,45155,45175,45216,45274,45295,45347,45357,45367,45404,45413,45462,45534,45562,45591,45602,45649,45699,45721,45747,45778,45801,45836,45908,45933,45965,46043,46068,46091,46102,46153,46167,46171,46186,46212,46218,46225,46235,46241,46296,46359,46416,46422,46427,46507,46610,46611,46614,46623,46645,46656,46677,46678,46706,46717,46731,46794,46798,46817,46847,46866,46917,47036,47041,47069,47084,47089,47119,47161,47195,47257,47260,47277,47306,47309,47330,47345,47444,47475,47477,47487,47494,47530,47576,47611,47616,47618,47626,47628,47652,47668,47679,47750,47755,47758,47759,47762,47813,47822,47855,47857,47860,47863,47962,47983,47992,48005,48008,48091,48107,48160,48187,48264,48318,48406,48430,48473,48490,48491,48495,48568,48602,48609,48615,48680,48786,48791,48793,48797,48823,48833,48841,48886,48891,48907,48910,48913,48925,48929,48953,48984,48999,49051,49059,49079,49117,49142,49175,49191,49192,49233,49239,49289,49291,49298,49386,49405,49451,49475,49520,49522,49534,49576,49661,49713,49731,49757,49759,49760,49789,49869,49916,49918,49942,49962,49967,49988,50013,50174,50194,50318,50336,50388,50468,50489,50502,50528,50611,50650,50723,50763,50791,50922,50941,51007,51125,51258,51281,51418,51452,51507,51572,51582,51635,51748,51947,51957,51967,51994,52002,52042,52172,52236,52376,52402,52421,52588,52702,52737,52758,52790,52878,52885,52904,53039,53155,53228,53452,53460,53461,53464,53514,53560,53672,53686,53713,53847,53973,54001,54057,54135,54136,54147,54217,54271,54284,54450,54453,54463,54562,54739,54765,54860,55283,55427,55438,55469,55536,55540,55543,55595,55791,55930,55961,55972,56083,56094,56154,56499,56586,56723,56843,56873,57007,57058,57082,57212,57227,57256,57296,57326,57418,57567,57650,57666,57720,57724,57838,57850,57861,57880,57997,58257,58353,58501,58743,58783,58816,58861,58926,59116,59210,59243,59509,59537,59602,59633,59751,59799,59801,59834,59889,60073,60094,60166,60189,60278,60296,60347,60409,60520,60532,60681,60711,60822,61056,61124,61239,61298,61426,61475,61775,61910,61976,61982,62292,62341,62384,62504,62515,62603,62605,62696,62717,62767,62841,63195,63339,63391,63715,63815,63835,63871,63912,64088,64121,64218,64263,64354,64397,64414,64521,64536,64579,64630,64783,64841,64885,64893,65040,65109,65205,65248,65280,65385,65415,65474,65643,65809,65830,65882,65929,66101,66109,66214,66271,66277,66295,66333,66426,66455,66498,66665,66696,66715,66775,66790,66806,66836,66896,66918,67253,67385,67509,67511,67624,67776,67865,67958,68610,68664,68731,68738,68786,68863,68921,69049,69079,69189,69194,69202,69225,69231,69322,69469,69563,69589,69717,69781,69823,69940,69965,70084,70088,70092,70234,70326,70361,70446,70477,70622,70625,70627,70644,71119,71383,71441,71477,71539,71546,71657,71739,71769,72164,72336,72441,72483,72504,72505,72524,72650,72665,72861,73112,73158,73174,73241,73402,73483,73512,73765,73767,73783,73790,73807,73910,74136,74305,74343,74349,74459,74473,74482,74564,74659,74682,74770,74935,74946,75205,75456,75665,75679,75700,75734,75826,75934,75971,76308,76403,76522,76567,76639,76719,76721,76767,76893,76981,76996,77090,77111,77246,77281,77283,77363,77567,77625,77961,77979,78250,78312,78369,78388,78487,78505,78512,78542,78586,78759,78857,78887,78927,79045,79134,79268,79297,79357,79431,79457,79483,79484,79535,79624,79671,79698,79703,79839,79882,79923,79968,80139,80238,80252,80340,80390,80416,80535,80595,80601,80652,80730,80742,80743,80808,80895,80899,80903,80952,80986,81012,81144,81234,81267,81347,81356,81456,81506,81544,81583,81641,81728,81797,81816,81851,81925,82031,82102,82138,82186,82227,82390,82407,82414,82538,82553,82558,82668,82671,82770,82869,83111,83114,83219,83222,83344,83351,83433,83491,83527,83547,83553,83566,83592,83643,83644,83662,83794,83944,84081,84142,84289,84312,84340,84411,84528,84839,84964,84972,84989,85018,85024,85041,85190,85217,85345,85347,85391,85412,85487,85489,85498,85549,85554,85638,85677,85790,85873,85888,85890,86012,86153,86188,86220,86380,86414,86558,86665,86879,86911,86949,86951,87074,87136,87209,87358,87490,87561,87715,87717,87744,87770,87844,87895,88071,88087,88121,88169,88305,88371,88389,88603,88678,88710,88713,88743,88766,89070,89107,89119,89124,89298,89305,89378,89427,89435,89438,89441,89594,89665,89672,89925,90013,90048,90092,90119,90144,90199,90212,90319,90327,90461,90562,90660,90696,90706,91003,91033,91203,91301,91307,91328,91335,91360,91476,91483,91709,91752,91840,92013,92025,92067,92100,92256,92319,92606,92626,92644,92668,92722,92724,92736,92770,92796,93252,93264,93411,93420,93480,93511,93527,93558,93723,93731,93734,93791,93809,93826,93828,93848,93868,93874,93876,93933,94335,94384,94385,94575,94682,94722,94775,94906,94972,95158,95160,95243,95428,95492,95549,95731,95756,95845,95859,95893,95943,95985,96022,96055,96232,96251,96278,96309,96328,96466,96571,96629,96638,96836,96902,96949,97013,97067,97104,97170,97183,97203,97206,97218,97347,97517,97524,97757,97811,97843,97876,98009,98105,98152,98155,98404,98676,98689,98748,98751,98767,98891,99050,99182,99416,99506,99604,99648,99677,99742,99750,99789,99802,99803,99885,99909,100001",
          "46,65,103,105,122,160,182,186,191,2054,2143,2272,2303,2350,2355,2463,2498,2592,2626,2636,3075,3137,3794,3820,3861,3875,3922,4403,4406,4504,4528,4703,4764,4804,6140,6621,7688,8387,9280,9394,9451,9555,9565,9566,9584,9641,9660,9698,9755,9793,9812,9831,9926,10021,10037,10078,10097,10131,10192,10363,10648,10819,10914,10933,11009,11066,11104,11161,11180,11218,11237,11256,11313,11408,11468,11598,11617,11636,11750,11930,11978,12092,12095,12121,12130,12149,12187,12225,12244,12282,12301,12317,12358,12360,12396,12415,12469,12510,12529,12624,12738,12776,12814,12833,12928,12947,12957,12985,13061,13099,13118,13194,13213,13251,13270,13289,13310,13327,13384,13441,13555,13593,13612,13650,13688,13707,13726,13764,13783,13802,13821,13859,13916,13992,14030,14087,14139,14143,14144,14163,14182,14220,14239,14258,14277,14296,14315,14334,14410,14429,14484,14524,14543,14638,14657,14752,14809,14810,14847,14863,14866,14886,14923,14951,14961,15018,15075,15094,15104,15151,15246,15265,15284,15322,15341,15360,15455,15493,15512,15588,15626,15645,15690,15702,15721,15835,15854,15873,15892,15947,15949,16006,16015,16025,16044,16078,16101,16120,16136,16158,16177,16207,16215,16234,16272,16291,16310,16329,16348,16367,16405,16424,16443,16481,16484,16500,16520,16538,16549,16587,16614,16633,16635,16652,16671,16728,16747,16785,16823,16842,16851,16861,16899,16909,16956,16975,16986,16994,17070,17108,17146,17203,17222,17298,17317,17393,17433,17450,17469,17507,17545,17564,17640,17678,17695,17697,17716,17735,17754,17773,17792,17811,17868,17887,17925,17944,18001,18058,18077,18093,18115,18153,18163,18172,18248,18267,18286,18305,18324,18343,18362,18381,18400,18419,18438,18533,18552,18571,18586,18609,18647,18666,18738,18761,18780,18818,18894,18932,18951,18989,19023,19057,19061,19065,19103,19112,19123,19160,19179,19236,19253,19274,19331,19333,19388,19407,19426,19445,19464,19502,19578,19599,19673,19692,19703,19711,19730,19749,19768,19806,19825,19863,19882,19920,19939,19958,19996,20007,20015,20043,20064,20091,20129,20148,20186,20205,20281,20300,20301,20338,20357,20376,20414,20452,20471,20509,20528,20566,20585,20604,20642,20661,20678,20699,20737,20756,20811,20813,20870,20889,20908,20946,20962,20965,21003,21019,21089,21117,21131,21136,21155,21212,21226,21231,21269,21288,21307,21309,21345,21364,21440,21478,21497,21535,21554,21573,21592,21611,21668,21702,21725,21744,21762,21763,21766,21820,21858,21874,21877,21896,21915,21934,21991,22048,22064,22105,22124,22162,22200,22215,22219,22238,22276,22333,22352,22371,22390,22428,22447,22466,22485,22504,22523,22542,22572,22618,22637,22652,22656,22694,22713,22789,22808,22827,22846,22884,22903,22960,22998,23017,23036,23093,23150,23169,23302,23321,23359,23378,23435,23473,23492,23549,23587,23625,23644,23682,23701,23758,23777,23815,23834,23929,23967,23986,24028,24043,24062,24081,24100,24157,24195,24233,24252,24290,24309,24347,24348,24385,24404,24423,24442,24461,24480,24499,24520,24537,24556,24632,24708,24727,24746,24762,24765,24784,24831,24879,24917,24974,24993,25012,25050,25069,25137,25164,25183,25202,25278,25280,25373,25392,25430,25449,25509,25525,25563,25582,25639,25658,25677,25734,25753,25792,25810,25827,25867,25886,25905,25923,25924,25962,25981,26000,26019,26038,26057,26076,26095,26133,26163,26171,26228,26247,26302,26304,26342,26361,26399,26437,26456,26475,26478,26494,26513,26520,26532,26589,26608,26627,26654,26722,26741,26759,26782,26821,26855,26912,26937,26982,27169,27178,27308,27317,27330,27348,27400,27523,27539,27577,27608,27615,27630,27672,27711,27821,27862,27878,27938,27976,28166,28204,28251,28403,28434,28527,28660,28667,28698,28789,28907,28926,28945,28976,28994,29040,29053,29094,29192,29219,29230,29287,29363,29401,29414,29477,29502,29515,29599,29629,29666,29773,29781,29869,29882,30008,30037,30066,30095,30124,30153,30182,30211,30240,30269,30298,30327,30356,30385,30414,30443,30472,30501,30530,30559,30588,30617,30646,30675,30704,30733,30762,30791,30820,30849,30878,30907,30936,30965,30994,31023,31052,31139,31168,31197,31226,31255,31284,31313,31342,31371,31400,31429,31458,31487,31516,31545,31574,31603,31632,31661,31719,31748,31777,31806,31835,31864,31893,31922,31951,31980,32009,32038,32067,32096,32125,32154,32183,32212,32241,32270,32299,32328,32357,32386,32415,32444,32473,32502,32531,32560,32589,32647,32676,32705,32734,32763,32792,32821,32850,32879,32908,32937,32966,32995,33024,33053,33082,33111,33140,33169,33198,33227,33256,33285,33314,33343,33372,33401,33430,33459,33488,33517,33546,33575,33604,33633,33662,33691,33720,33749,33778,33807,33836,33865,33894,33923,33952,33981,34010,34039,34068,34097,34126,34155,34184,34213,34242,34271,34300,34358,34387,34416,34445,34474,34503,34532,34561,34590,34619,34648,34677,34706,34735,34764,34793,34822,34851,34880,34909,34938,34967,34996,35008,35061,35064,35069,35097,35171,35196,35260,35317,35334,35390,35408,35439,35471,35479,35527,35586,35645,35673,35718,35777,35836,35907,35932,35966,36025,36061,36088,36124,36130,36148,36207,36234,36278,36317,36337,36396,36468,36527,36586,36650,36685,36689,36709,36768,36793,36839,36898,36957,37029,37088,37115,37147,37211,37270,37302,37329,37400,37459,37518,37590,37649,37708,37772,37810,37831,37890,37961,37996,38020,38061,38071,38079,38114,38151,38210,38269,38333,38363,38392,38439,38451,38487,38522,38581,38640,38665,38668,38712,38771,38800,38830,38894,38953,39012,39036,39039,39083,39142,39201,39273,39332,39391,39455,39514,39550,39573,39644,39684,39703,39762,39799,39834,39952,40107,40136,40165,40194,40223,40252,40281,40310,40339,40368,40397,40426,40455,40484,40513,40542,40571,40600,40629,40658,40687,40716,40745,40774,40803,40832,40861,40919,40948,40977,41006,41035,41064,41093,41122,41151,41180,41209,41238,41267,41296,41325,41354,41383,41412,41441,41470,41499,41528,42000,42029,42058,42087,42116,42151,42176,42257,42266,42322,42341,42391,42416,42503,42552,42578,42606,42627,42653,42716,42740,42815,42890,42927,42977,43052,43082,43101,43112,43127,43173,43214,43289,43326,43364,43417,43451,43506,43526,43557,43601,43688,43733,43737,43763,43796,43798,43838,43918,43925,43989,44000,44046,44075,44162,44237,44261,44312,44349,44399,44454,44474,44549,44598,44614,44636,44711,44786,44824,44873,44941,44948,44974,44979,45023,45084,45096,45144,45171,45205,45263,45286,45336,45393,45434,45451,45490,45523,45569,45638,45669,45710,45767,45825,45897,45954,45977,45987,46095,46113,46120,46174,46301,46304,46306,46357,46412,46417,46490,46524,46567,46597,46604,46631,46676,46700,46734,46784,46906,46948,46965,46992,47022,47029,47055,47078,47149,47169,47173,47177,47190,47274,47293,47321,47391,47401,47423,47443,47458,47469,47495,47510,47690,47742,47751,47754,47764,47802,47841,47846,47883,47908,47914,47915,47921,47933,48063,48072,48088,48097,48119,48131,48140,48168,48169,48201,48208,48217,48272,48279,48300,48312,48370,48371,48413,48414,48454,48486,48510,48524,48541,48557,48583,48597,48614,48646,48654,48685,48720,48730,48748,48804,48826,48854,48894,48900,48917,49018,49092,49137,49220,49244,49280,49282,49315,49357,49416,49424,49425,49444,49472,49474,49484,49488,49513,49619,49636,49637,49645,49667,49683,49743,49758,49780,49800,49871,49968,49990,50034,50087,50090,50129,50220,50428,50441,50503,50539,50569,50593,50606,50719,50961,51014,51082,51347,51382,51488,51679,51723,51839,51903,51927,51937,51941,52006,52024,52122,52146,52160,52185,52386,52467,52520,52534,52537,52591,52640,53087,53102,53172,53251,53264,53371,53763,53777,53784,53790,53827,53892,53938,53948,54115,54227,54258,54491,54541,54581,54582,54705,54841,55001,55008,55088,55130,55163,55215,55519,55533,55692,55775,55777,55786,55826,55970,55995,56015,56104,56187,56212,56244,56350,56351,56403,56448,56485,56581,56654,56757,56823,56959,56984,57097,57109,57121,57408,57442,57596,57603,57716,57741,57796,57806,57885,57968,58005,58232,58424,58430,58492,58526,58586,58798,58808,58826,59085,59191,59274,59287,59289,59299,59307,59343,59358,59519,59584,59610,59697,59746,60184,60229,60542,60559,60625,60650,60776,61070,61087,61088,61201,61289,61518,61550,61602,61641,61730,61763,61950,61973,62117,62176,62209,62313,62375,62505,62518,62519,62567,62572,62645,62703,62766,62876,62904,62982,63041,63193,63207,63220,63329,63405,63521,63567,63619,63654,63686,63781,63936,64016,64125,64320,64376,64465,64557,64559,64615,64673,64740,64753,65200,65284,65295,65383,65472,65509,65644,65677,65744,65934,66002,66088,66137,66143,66256,66321,66402,66505,66612,66779,66927,66939,67089,67182,67319,67358,67386,67430,67557,67574,67597,67626,67631,67678,67703,67746,67815,67870,68139,68167,68262,68386,68431,68438,68439,68547,68775,68785,68833,68867,68869,69253,69270,69324,69331,69631,69679,69888,69903,69982,70009,70100,70117,70142,70165,70168,70260,70277,70511,70585,70702,70878,70963,71175,71247,71472,71722,71749,71791,71793,71803,71843,71890,71946,71951,71973,72018,72065,72076,72129,72174,72334,72345,72479,72516,72559,72588,72625,72680,72685,72751,72830,72935,73169,73271,73432,73465,73472,73699,73759,73761,73813,73927,73978,73991,74023,74142,74164,74331,74374,74436,74500,74575,74684,74752,74857,75021,75033,75154,75235,75421,75431,75483,75525,75636,75680,75730,75771,75808,75824,75856,75933,76007,76171,76363,76397,76525,76569,76587,76628,76680,76726,76814,76884,77207,77210,77267,77404,77557,77635,77708,77790,77846,77856,77970,77980,78007,78249,78337,78356,78367,78396,78855,78917,79154,79194,79238,79290,79304,79391,79413,79425,79478,79494,79551,79586,79683,79758,79836,79944,79975,80124,80132,80162,80251,80358,80417,80425,80629,80697,80761,80933,81068,81282,81403,81430,81451,81481,81510,81534,81541,81600,81650,81653,81758,81831,81910,82158,82265,82301,82559,82603,82609,82706,82787,82855,82907,82985,83058,83094,83158,83475,83576,83588,83619,83806,84030,84149,84259,84319,84513,84532,84635,84720,84735,84851,85105,85144,85230,85290,85324,85336,85348,85450,85466,85467,85474,85543,85551,85622,85684,85932,85944,85962,85972,86018,86102,86146,86197,86284,86378,86395,86429,86491,86524,86758,86804,86830,86882,86997,87050,87077,87534,87535,87553,87773,87975,88088,88143,88402,88562,88868,88923,88939,89140,89350,89364,89551,89636,89681,89731,89743,89915,89976,90030,90252,90288,90511,90530,90572,90607,90822,90915,90952,91048,91094,91113,91318,91342,91397,91398,91401,91460,91488,91529,91565,91643,91718,91810,91958,91991,92075,92079,92176,92177,92250,92302,92327,92371,92412,92595,92989,93085,93132,93188,93241,93280,93299,93357,93386,93401,93562,93577,93594,93639,93672,93685,93722,93744,93801,93836,93879,93917,93968,93978,94012,94061,94093,94210,94236,94279,94340,94486,94613,94676,94692,94757,94808,94857,94952,95048,95086,95099,95208,95213,95233,95283,95297,95726,95750,95816,95826,95914,96189,96238,96292,96333,96401,96603,96644,96766,96828,96893,96947,97029,97236,97284,97385,97541,97565,97742,97798,97827,97835,98060,98139,98192,98261,98273,98315,98381,98389,98453,98483,98518,98670,98758,98833,98917,98926,99058,99077,99081,99234,99297,99716,99841,99865,99895,99915,99918,99937,99939",
          "71,129,185,207,224,1735,1778,1949,1963,1982,2039,2151,2226,2244,2275,2314,2339,2617,2811,2913,2989,3016,3051,3139,3140,3177,3202,3247,3325,3339,3352,3407,3662,3790,3867,3972,4008,4034,4139,4243,4334,4612,4630,4647,4919,5120,5526,6955,9474,9563,9601,9620,9639,9658,9674,9675,9677,9696,9772,9810,9829,9924,9962,10095,10114,10722,10741,10760,10912,10950,11254,11292,11406,11442,11482,11539,11558,11615,11657,11881,11900,11957,11980,11995,12014,12033,12090,12147,12166,12204,12223,12261,12280,12318,12365,12413,12451,12588,12603,12717,12755,12888,12945,13021,13059,13078,13098,13102,13138,13230,13249,13263,13325,13344,13415,13420,13477,13496,13515,13536,13574,13591,13600,13667,13686,13705,13743,13762,13781,13819,13857,13887,13933,14085,14104,14180,14199,14218,14237,14256,14275,14294,14313,14389,14408,14427,14522,14541,14655,14659,14667,14674,14677,14693,14788,14860,14902,14921,14959,14990,14997,15016,15054,15073,15111,15130,15206,15215,15244,15263,15272,15282,15287,15320,15339,15358,15415,15434,15510,15515,15529,15539,15561,15567,15586,15643,15662,15700,15738,15776,15795,15833,15852,15871,15939,16004,16023,16061,16064,16080,16118,16175,16194,16199,16213,16232,16251,16289,16327,16346,16365,16403,16422,16455,16460,16479,16489,16498,16555,16574,16593,16612,16650,16707,16726,16745,16802,16821,16850,16859,16869,16878,16935,16954,17011,17030,17068,17087,17106,17119,17163,17182,17201,17258,17277,17279,17296,17315,17353,17410,17448,17467,17517,17524,17543,17547,17563,17600,17619,17623,17657,17676,17714,17771,17781,17790,17809,17828,17847,17885,17904,17923,17943,17961,17994,17999,18037,18090,18132,18189,18208,18227,18257,18259,18265,18284,18294,18322,18341,18360,18379,18398,18417,18436,18455,18512,18531,18550,18570,18584,18626,18664,18702,18721,18733,18759,18766,18778,18816,18835,18892,18911,18917,18930,18968,19006,19025,19082,19120,19139,19164,19196,19215,19234,19256,19268,19272,19276,19348,19349,19367,19372,19386,19424,19432,19443,19462,19476,19519,19538,19557,19587,19614,19652,19690,19709,19747,19785,19804,19807,19823,19842,19861,19880,19899,19918,19937,19956,19975,19994,20032,20070,20127,20165,20184,20208,20213,20241,20279,20298,20317,20336,20393,20412,20431,20450,20469,20507,20526,20545,20570,20583,20602,20621,20640,20682,20697,20716,20754,20773,20792,20849,20883,20887,20925,20944,20963,21020,21096,21115,21138,21172,21191,21229,21267,21286,21305,21327,21343,21419,21438,21457,21464,21476,21533,21552,21571,21590,21628,21666,21704,21723,21761,21780,21799,21818,21875,21894,21913,21932,21951,21970,22008,22027,22046,22084,22126,22198,22223,22236,22293,22303,22331,22369,22388,22407,22422,22426,22464,22502,22540,22559,22654,22692,22711,22717,22749,22764,22787,22800,22825,22844,22882,22901,22911,22920,22977,23015,23034,23053,23110,23129,23205,23224,23243,23262,23283,23300,23309,23357,23362,23395,23414,23452,23509,23528,23580,23591,23642,23661,23696,23706,23737,23756,23813,23832,23853,23863,23870,23889,23908,23927,23965,23984,24003,24022,24041,24060,24079,24098,24117,24174,24201,24212,24231,24288,24326,24364,24383,24440,24478,24497,24516,24554,24573,24611,24630,24687,24689,24706,24744,24763,24782,24801,24807,24839,24858,24877,24953,24976,25018,25029,25048,25143,25181,25276,25295,25314,25337,25352,25371,25425,25428,25466,25499,25504,25533,25542,25580,25599,25659,25675,25679,25694,25713,25740,25751,25770,25808,25831,25869,25903,25922,25941,25944,25960,26017,26055,26074,26093,26125,26150,26188,26207,26264,26283,26359,26397,26435,26473,26492,26511,26568,26587,26644,26663,26682,26761,26807,26854,26887,26948,27138,27141,27176,27197,27249,27290,27312,27366,27404,27445,27518,27694,27708,27803,27841,27857,27921,28050,28088,28107,28202,28324,28335,28446,28449,28454,28487,28582,28677,28754,28880,28943,28951,28987,29002,29354,29399,29421,29537,29551,29570,29627,29665,29705,29725,29741,29817,29862,29893,29912,29988,30006,30035,30064,30093,30122,30151,30180,30209,30238,30267,30296,30325,30354,30383,30412,30441,30470,30499,30528,30557,30586,30615,30644,30673,30702,30731,30760,30789,30818,30847,30876,30905,30934,30963,30992,31021,31050,31079,31108,31137,31166,31195,31224,31253,31282,31311,31340,31369,31398,31427,31456,31485,31514,31543,31572,31601,31630,31659,31688,31717,31746,31775,31804,31833,31862,31891,31920,31949,31978,32007,32036,32065,32094,32123,32152,32181,32210,32239,32268,32297,32326,32355,32384,32413,32442,32471,32500,32529,32558,32587,32616,32645,32674,32703,32732,32761,32790,32819,32848,32877,32906,32935,32964,32993,33022,33051,33080,33109,33138,33167,33196,33225,33254,33283,33312,33341,33370,33399,33428,33457,33486,33515,33544,33573,33602,33631,33660,33689,33718,33747,33776,33805,33834,33863,33892,33921,33950,33979,34008,34037,34066,34095,34124,34153,34182,34211,34240,34269,34298,34327,34356,34385,34414,34443,34472,34501,34530,34559,34588,34617,34646,34675,34704,34733,34762,34791,34820,34849,34878,34907,34936,34965,34994,35006,35095,35157,35169,35199,35245,35258,35290,35292,35322,35332,35386,35398,35406,35477,35525,35554,35584,35637,35643,35676,35716,35775,35834,35866,35905,35940,35964,36023,36057,36066,36086,36146,36199,36205,36240,36247,36276,36335,36394,36466,36525,36584,36648,36707,36766,36837,36870,36896,36955,37027,37054,37086,37145,37209,37268,37327,37398,37437,37457,37516,37588,37647,37706,37755,37756,37770,37829,37888,37923,37959,37995,38018,38077,38149,38176,38208,38267,38331,38390,38420,38449,38494,38520,38579,38638,38710,38769,38828,38892,38951,39010,39081,39131,39140,39187,39199,39237,39271,39312,39314,39330,39389,39453,39498,39512,39571,39622,39642,39701,39725,39760,39832,39891,39933,39944,39950,40110,40139,40168,40197,40226,40255,40284,40313,40342,40371,40400,40429,40458,40487,40516,40545,40574,40603,40632,40661,40690,40719,40748,40777,40806,40835,40864,40893,40922,40951,40980,41009,41038,41067,41096,41125,41154,41183,41212,41241,41270,41299,41328,41357,41386,41415,41444,41473,41502,41531,42002,42031,42060,42089,42118,42166,42178,42251,42268,42293,42328,42343,42418,42450,42466,42496,42505,42554,42571,42580,42655,42693,42711,42742,42772,42775,42802,42806,42817,42849,42892,42979,43019,43054,43093,43110,43129,43161,43162,43201,43216,43241,43274,43291,43366,43409,43453,43482,43492,43528,43552,43580,43588,43603,43622,43642,43690,43748,43765,43822,43825,43840,43860,43894,43927,43951,43981,43987,44002,44023,44029,44038,44077,44097,44100,44126,44138,44149,44154,44164,44200,44226,44239,44272,44300,44303,44314,44343,44353,44401,44476,44505,44509,44551,44582,44588,44638,44663,44675,44682,44713,44788,44832,44842,44875,44950,45025,45098,45130,45146,45169,45207,45230,45265,45338,45395,45433,45453,45525,45582,45640,45712,45769,45827,45862,45899,45956,45976,45989,46020,46034,46071,46133,46146,46150,46227,46251,46269,46272,46273,46320,46341,46346,46386,46392,46399,46400,46468,46472,46474,46494,46517,46523,46548,46564,46569,46650,46776,46779,46783,46786,46796,46818,46820,46836,46893,46935,46936,46939,46942,46954,47000,47008,47065,47094,47112,47157,47197,47227,47280,47343,47359,47394,47427,47448,47466,47467,47501,47503,47561,47583,47631,47700,47738,47763,47816,47821,47837,47851,47867,47872,47897,47923,47924,47960,47974,47976,47996,47998,48002,48042,48061,48066,48101,48135,48171,48204,48224,48261,48273,48336,48344,48378,48397,48417,48425,48462,48470,48485,48567,48594,48617,48622,48632,48659,48663,48672,48673,48681,48732,48747,48755,48801,48852,48860,48879,48905,48949,48982,49003,49073,49135,49185,49202,49215,49216,49232,49247,49256,49333,49366,49388,49397,49409,49421,49437,49502,49507,49529,49540,49564,49583,49591,49720,49727,49755,49766,49779,49817,49885,49888,49892,49948,49987,49997,50127,50177,50211,50344,50372,50476,50483,50493,50504,50505,50578,50740,50782,50927,50949,50992,50994,51054,51064,51291,51349,51380,51554,51708,51736,51747,51778,51783,51931,51966,52069,52192,52235,52271,52314,52366,52787,52789,52955,53005,53016,53049,53057,53186,53194,53239,53253,53285,53326,53341,53444,53604,53689,53792,53919,53927,53946,53961,53994,54141,54291,54412,54446,54531,54574,54725,54762,54906,54996,55223,55228,55337,55354,55374,55392,55454,55589,55641,55729,55831,55852,55905,55933,56048,56102,56401,56673,56746,56760,56811,56976,57015,57091,57142,57463,57555,57690,57767,57832,57862,57863,58036,58288,58316,58392,58394,58448,58541,58633,58672,58696,58911,58965,58974,58977,58983,59014,59082,59222,59259,59323,59451,59564,59752,59869,59912,60045,60276,60369,60407,60535,60598,60601,60774,60926,60945,61068,61241,61615,61728,61803,61821,62058,62093,62151,62234,62469,62498,62601,62691,62731,62797,62820,62827,62847,62893,62903,63181,63206,63292,63297,63453,63465,63543,63802,63865,64044,64091,64137,64259,64505,64555,64561,65007,65013,65052,65354,65407,65444,65475,65508,65530,65577,65597,65705,65825,65858,65900,65955,66020,66179,66280,66293,66300,66335,66337,66358,66441,66508,66546,66557,66586,66641,66709,66718,66720,66722,66736,66786,66954,67062,67178,67204,67299,67573,67648,67684,67718,67731,67764,67804,67850,67876,67918,67968,68000,68026,68029,68044,68048,68112,68127,68162,68333,68336,68390,68400,68535,68600,68655,68739,68931,68932,68981,69001,69002,69094,69205,69482,69527,69602,69661,69757,69764,69976,70000,70008,70167,70207,70293,70322,70386,70403,70416,70668,70801,70802,70861,71019,71154,71177,71191,71203,71218,71249,71404,71597,71623,71635,71746,71764,71807,71809,71859,72027,72133,72142,72147,72184,72186,72232,72241,72254,72341,72442,72449,72473,72886,72922,73017,73189,73310,73318,73362,73366,73440,73578,73580,73603,73652,73708,73721,73819,73885,73886,73926,74110,74183,74216,74235,74380,74388,74404,74442,74463,74525,74765,75070,75184,75211,75217,75256,75301,75416,75557,75749,75810,75857,75997,76016,76043,76057,76167,76372,76377,76411,76416,76548,76657,76739,76783,76858,76866,76946,76957,77160,77248,77254,77261,77262,77412,77656,77840,77855,77903,78217,78463,78624,78799,78842,78903,78951,78987,79453,79522,79552,79865,79904,79997,80156,80247,80250,80343,80509,80521,80539,80596,80623,80680,80707,80717,80971,81042,81079,81290,81359,81369,81415,81440,81606,81642,81643,81649,81803,81846,81883,82004,82021,82190,82221,82391,82435,82458,82541,82551,82640,82725,82751,82810,82841,82916,83303,83386,83474,83665,83713,83832,84028,84090,84121,84158,84195,84218,84236,84282,84392,84473,84507,84644,84714,84740,84757,84805,85188,85214,85420,85479,85731,85789,85987,86063,86095,86131,86150,86169,86191,86230,86281,86296,86412,86482,86554,86590,86647,86651,86759,86768,86776,86780,86787,86923,87036,87046,87088,87176,87246,87375,87386,87441,87577,87855,87961,88057,88059,88434,88532,88538,88716,88732,88825,88830,89022,89191,89240,89371,89570,89601,89682,89758,89764,89868,89908,89913,89926,89938,90152,90215,90230,90322,90386,90446,90476,90765,90796,90885,90950,91134,91238,91310,91349,91453,91605,91696,91771,91775,91843,91895,91967,92015,92107,92237,92341,92510,92525,92529,92621,92673,92733,92740,92802,92803,92937,93054,93064,93071,93239,93471,93504,93537,93541,93567,93634,93729,93808,93810,93960,93997,94009,94069,94103,94105,94309,94566,94597,94795,95050,95134,95173,95196,95311,95359,95397,95445,95478,95491,95530,95615,95757,95776,95792,95871,95873,96007,96145,96188,96205,96281,96367,96470,96708,96726,96762,96916,96927,96935,96942,97061,97097,97126,97195,97249,97472,97537,97562,97639,97673,97697,97801,97805,97815,97875,97901,97908,97948,98229,98396,98397,98425,98462,98612,98663,98793,98853,98915,98951,98981,99025,99304,99315,99319,99523,99596,99685,99694,99758,99993,100002",
          "203,256,286,340,379,400,421,424,425,450,461,468,499,586,590,693,705,724,768,848,862,867,1005,1016,1078,1125,1152,1206,1237,1245,1271,1395,1494,1535,1540,1801,5302,5424,5465,5549,5574,5642,5750,5915,6035,6045,6087,6091,6127,6177,6356,6392,6430,6517,6518,6609,6906,6976,6985,7108,7123,7204,7243,7311,7319,7361,7406,7439,7445,7539,7669,7731,7735,7737,7901,7940,7944,8006,8020,8071,8221,8368,8378,8565,8664,8774,8864,9025,9078,9110,9364,9594,9856,9899,9917,10004,10079,10162,10214,10369,10384,10500,10612,10677,10690,10821,10867,10871,10908,10916,10983,10998,11781,11836,12348,12426,13157,13925,14057,14161,14401,14476,14776,14872,15312,15677,16096,16452,16512,16931,17133,17630,17646,18670,18749,18936,20520,20556,22241,24016,24073,24262,24702,25044,26313,26762,26787,26837,26863,26973,27201,27397,27681,28109,28176,28844,28949,29244,29296,29508,30599,31241,35707,36138,36197,36269,36371,36453,36459,36757,37003,37186,37260,37373,37449,37697,37727,37758,37822,37931,37948,38256,38682,38999,39377,39419,39489,39943,40792,41239,44374,45167,46684,48595,49966,50079,50233,50416,50486,50519,50575,50646,50854,50865,50908,51217,51369,51407,51608,51642,51652,51989,51995,52161,52177,52187,52231,52250,52478,52541,52780,52798,52809,52890,52914,53079,53138,53162,53199,53217,53291,53316,53835,53877,54019,54072,54162,54331,54339,54416,54515,54521,54740,54875,54886,54959,54978,55017,55030,55033,55055,55149,55153,55253,55272,55422,55618,55686,55709,55721,56056,56136,56145,56263,56289,56317,56375,56391,56479,56595,56785,56800,56812,56825,56848,56907,57035,57170,57238,57319,57375,57406,57519,57605,57659,57673,57801,57820,57952,57965,58043,58102,58176,58297,58326,58431,58567,58652,58676,58934,58952,59316,59405,59413,59457,59713,59723,59733,59816,59836,59918,59929,60009,60033,60236,60246,60266,60497,60502,60555,60576,60583,60620,60626,60770,60786,60954,61034,61164,61427,61443,61533,61598,61604,61613,61677,61750,61752,61779,61830,61992,62143,62174,62307,62482,62570,62658,62713,62813,62839,62891,62972,63025,63131,63226,63351,63628,63639,63687,63784,63947,64225,64498,64513,64636,64729,65036,65148,65187,65222,65228,65255,65285,65389,65396,65445,65580,65640,65676,65678,65701,65722,65845,65910,65941,65966,65980,66057,66171,66234,66302,66329,66336,66359,66410,66436,66494,66531,66538,66551,66767,66817,66854,66875,66921,67126,67363,67467,67470,67478,67479,67575,67758,67767,67779,67788,67823,67866,67872,67882,67947,67991,68027,68143,68221,68244,68247,68328,68341,68447,68480,68626,68669,68695,68711,68819,68987,69158,69386,69553,69691,69710,69948,70052,70110,70236,70536,70596,70763,70775,70819,70821,70865,70928,70974,71195,71331,71338,71387,71560,71594,71602,71610,71653,71695,71717,71896,71903,71933,71981,72161,72168,72177,72297,72343,72359,72561,72633,72682,72754,72782,72825,72872,73041,73132,73172,73322,73338,73447,73582,73661,73720,73834,73839,73845,73866,73929,74031,74058,74108,74163,74441,74766,74768,74841,74872,74969,75016,75062,75110,75149,75158,75313,75398,75417,75514,76213,76266,76388,76407,76430,76461,76520,76527,76558,76695,76758,76805,76831,77139,77200,77278,77285,77291,77427,77478,77548,77603,77737,77815,77983,78050,78240,78501,78644,78650,78882,78949,79038,79233,79257,79397,79410,79412,79426,79448,79464,79486,79503,79660,79668,79687,79704,79713,79728,79804,79853,79856,79947,79950,80028,80043,80057,80160,80443,80525,80576,80702,80833,80853,80856,80867,80908,80930,80931,80948,80998,81015,81047,81127,81217,81264,81460,81502,81632,81671,81710,81743,81746,81765,81872,81875,81958,81971,82034,82063,82066,82253,82343,82347,82487,82516,82521,82592,82618,82764,82808,83040,83102,83202,83335,83378,83409,83424,83518,83526,83710,83850,84004,84124,84306,84335,84401,84526,84557,84716,84765,84785,85249,85374,85378,85425,85634,85670,85699,85756,85855,85986,86056,86083,86147,86203,86237,86505,86528,86570,86597,86684,86711,86713,86796,86890,86981,87026,87449,87532,87557,87565,88037,88043,88072,88134,88166,88328,88357,88392,88552,88585,88619,88622,88712,89001,89173,89252,89425,89630,89632,89635,89806,89905,90101,90137,90440,90703,90716,90748,90770,90827,90850,90929,90944,90964,91231,91239,91269,91303,91407,91519,91639,91650,91682,91760,91773,91860,91909,92003,92006,92033,92066,92139,92263,92275,92420,92766,92788,92795,92839,92947,93050,93063,93232,93242,93292,93317,93330,93345,93380,93421,93484,93703,93834,93998,94086,94111,94239,94363,94401,94579,94591,94636,94838,94852,95102,95121,95133,95189,95197,95343,95430,95607,95692,95759,96025,96313,96488,96567,96669,96670,96795,96901,96917,96972,96994,96996,97066,97198,97226,97256,97395,97402,97530,97557,97620,97633,97669,97717,97718,97796,98025,98132,98182,98246,98308,98452,98667,98698,98823,98908,99034,99186,99292,99444,99469,99767,99831,99834,99907,99936",
          "137,208,219,434,931,977,1158,1182,1272,1431,1641,1732,1765,1799,1815,1829,1839,2047,2076,2359,2564,2690,2748,2798,2857,3042,3221,3367,3421,3500,3613,3655,3702,3795,3804,4127,4171,4185,4266,4383,4550,4572,4575,4741,5029,5073,5096,5156,5163,5219,5262,5312,5326,5372,5488,5515,5524,5533,5571,5747,5809,5816,5887,5901,5940,6101,6167,6293,6383,6400,6528,6551,6564,6578,6634,6702,6940,6945,6947,6992,7013,7047,7163,7308,7350,7412,7546,7575,7649,7710,7719,7796,7814,7861,7909,7954,7995,8143,8164,8216,8229,8289,8396,8414,8492,8624,8674,8756,8760,8788,8821,8831,8842,8931,9235,9277,9464,9581,9599,9730,9794,9865,9916,9933,9983,10058,10068,10126,10230,10276,10322,10325,10341,10530,10550,10607,10624,10719,10733,10765,10813,10834,10944,10986,11025,11082,11143,11232,11323,11365,11416,11557,11560,11660,11697,11838,11906,11937,11956,11990,12114,12146,12278,12335,12578,12739,12811,12891,12911,13137,13171,13276,13380,13407,13427,13557,13584,13665,13799,13842,13860,13943,13990,14198,14329,14402,14460,14763,14803,14804,14806,14894,14956,14982,15014,15214,15224,15254,15280,15413,15436,15505,15728,15736,15868,15905,16039,16095,16211,16212,16401,16801,16818,16937,17050,17118,17266,17589,17602,17635,17812,17814,17883,17920,17928,18457,18471,18539,18542,18672,18785,19138,19156,19288,19427,19528,19634,19659,19698,19736,19896,19934,20087,20124,20183,20270,20497,20580,20714,20744,20822,21149,21174,21470,21526,21678,21688,21748,21849,21873,22138,22185,22233,22246,22286,22310,22314,22425,22482,22595,22597,22657,22659,22862,22976,23019,23032,23033,23076,23298,23547,23621,23729,23824,23833,24097,24128,24245,24287,24496,24593,24724,24825,24844,24874,24894,24912,25083,25142,25333,25521,25617,25692,25783,25901,25989,26217,26232,26340,26491,26710,26788,26806,26877,26975,27083,27091,27146,27236,27327,27439,27448,27461,27545,27666,27745,27754,28092,28187,28203,28311,28348,28364,28416,28465,28467,28617,28685,28766,28860,28903,28932,28942,29017,29151,29319,29430,29462,29598,29834,29933,29966,29978,30005,30034,30092,30177,30266,30293,30903,31367,32004,32180,32325,32527,32615,32702,32729,32816,33367,33717,33948,34122,34152,34500,34528,34558,34645,34818,35147,35208,35583,35685,35819,35878,36079,36139,36202,36450,36465,36502,36514,36691,36932,36944,37011,37053,37071,37073,37077,37383,37395,37483,37506,37675,37704,38043,38062,38122,38126,38142,38329,38330,38577,38603,39071,39126,39262,39623,39635,39680,39731,39750,39857,39878,39930,40669,41097,41184,41234,41358,42117,42261,42405,42506,42709,42982,43080,43170,43219,43403,43404,43434,43497,43693,43888,44167,44526,44558,44670,44720,44807,44933,45067,45233,45320,45617,45671,45697,46154,46338,46382,46979,47017,47139,47732,47815,47936,48102,48142,48242,48991,49042,49224,49803,49836,50030,50074,50140,50283,50337,50445,50462,50479,50513,50530,50563,50584,50590,50651,50838,50891,50892,50932,50956,51088,51111,51161,51280,51391,51474,51500,51502,51508,51587,51624,51640,51646,51779,51808,51868,51955,51983,52053,52055,52238,52419,52422,52440,52475,52540,52611,52711,53001,53076,53108,53133,53159,53181,53303,53384,53564,53700,53815,53859,53883,53923,53924,54012,54038,54100,54143,54288,54289,54385,54397,54406,54428,54489,54529,54551,54757,54865,54871,54910,54923,55016,55046,55076,55390,55553,55636,55771,55914,55941,55990,56020,56061,56147,56160,56346,56458,56490,56541,56810,56883,56895,56900,57027,57106,57129,57172,57262,57324,57344,57367,57489,57513,57687,57710,57743,57744,57848,57959,57990,58057,58171,58219,58305,58318,58442,58472,58549,58554,58565,58852,58864,59061,59193,59258,59317,59374,59462,59498,59563,59634,59652,59754,59807,59852,59958,60253,60384,60485,60556,60589,60654,60823,60890,60899,61094,61203,61214,61240,61315,61321,61346,61374,61594,61759,61978,62012,62034,62139,62320,62338,62471,62725,62773,62902,63093,63141,63229,63266,63295,63366,63531,63665,63737,63803,63868,63906,63928,64035,64104,64159,64347,64548,64605,64622,64683,64697,64703,64766,64850,64949,64962,64993,65027,65080,65094,65113,65124,65167,65277,65613,65633,65704,65708,65770,65943,66005,66069,66192,66209,66491,66578,66692,66768,66776,66910,66928,67006,67088,67323,67367,67390,67450,67547,67549,67576,67582,67680,67773,67793,67855,67939,68195,68259,68265,68285,68365,68510,68612,68681,68890,68977,69042,69054,69172,69282,69410,69564,69706,69716,70208,70245,70478,70629,70776,70791,70885,70917,70977,71076,71257,71358,71761,71778,71840,71851,72162,72282,72379,72534,72556,72689,72904,73225,73252,73261,73294,73353,73672,73788,73904,74010,74101,74139,74303,74322,74377,74431,74633,74696,74779,74790,74938,74963,75004,75011,75019,75156,75248,75312,75319,75331,75361,75439,75474,75794,75798,75831,76080,76210,76229,76296,76325,76660,76690,76932,77022,77032,77297,77418,77426,77431,77525,77741,77879,77967,77976,78004,78011,78180,78273,78382,78494,78558,78642,78745,78829,78879,78881,78898,78899,79004,79122,79159,79384,79729,79749,79788,79842,79897,79945,80085,80111,80135,80153,80260,80353,80373,80477,80553,80589,80590,80840,80862,81005,81165,81180,81184,81266,81363,81429,81548,81557,81713,81747,81756,81866,81959,82143,82155,82229,82242,82292,82332,82648,82792,82967,82989,83028,83083,83091,83169,83174,83243,83261,83279,83283,83345,83373,83450,83466,83537,83554,83872,83892,83893,84127,84275,84333,84396,84518,84549,84583,84601,84723,84762,84770,84808,84892,84900,84915,84920,84942,84955,85115,85261,85335,85506,85610,85759,85805,85911,85936,85977,86090,86097,86124,86128,86195,86200,86233,86239,86305,86326,86432,86457,86661,87429,87438,87451,87462,87573,87601,87751,87957,88028,88355,88473,88496,88579,88649,88704,88804,88851,88872,88906,88922,88975,88976,88982,89037,89069,89073,89117,89200,89224,89255,89267,89381,89745,89902,90038,90067,90268,90536,90566,90826,90875,91015,91129,91170,91198,91219,91320,91361,91429,91452,91543,91603,91704,91850,91889,91921,92125,92181,92200,92227,92267,92556,92836,92943,92956,92971,92981,92984,93006,93008,93101,93138,93150,93173,93289,93313,93332,93351,93467,93473,93495,93626,93629,93687,93793,94031,94037,94043,94107,94118,94505,94567,94618,94649,94856,94891,94896,94962,95043,95222,95303,95360,95476,95550,95554,95779,95844,95922,95969,96138,96231,96295,96330,96502,96855,96925,96995,97137,97246,97442,97529,97547,97579,97644,97691,97736,97755,97767,97779,97888,97954,98007,98039,98169,98256,98260,98272,98332,98424,98433,98531,98564,98659,98672,99071,99129,99214,99262,99314,99379,99462,99478,99610,99717,99734,99747,99799,99857,99866,99922,99957,99972",
          "49,68,125,144,241,2072,2325,2556,2609,2630,2649,2908,3044,3059,3074,3090,3173,3873,3900,4174,4404,4461,4579,4617,4641,4746,4845,6879,7676,9568,9587,9644,9663,9682,9701,9715,9739,9796,9834,9840,10043,10708,10746,11240,11373,11392,11430,11554,11563,11601,11639,11658,11848,11921,11924,11943,12057,12076,12152,12209,12228,12247,12266,12285,12298,12361,12380,12399,12456,12465,12475,12551,12570,12627,12683,12684,12703,12770,12779,12798,12807,12836,12893,12912,12988,12993,13026,13064,13083,13121,13140,13178,13197,13232,13235,13255,13273,13292,13330,13358,13365,13368,13406,13444,13463,13482,13494,13503,13520,13539,13558,13575,13577,13596,13613,13615,13634,13653,13672,13710,13741,13767,13786,13805,13843,13862,13900,13904,13938,13957,13973,13976,13995,14007,14033,14052,14067,14109,14128,14166,14185,14204,14223,14280,14299,14318,14337,14356,14394,14413,14431,14432,14451,14470,14501,14527,14546,14565,14603,14622,14660,14698,14755,14789,14793,14808,14817,14831,14850,14868,14869,14888,14898,14945,14964,15002,15021,15097,15116,15154,15158,15192,15212,15230,15249,15306,15325,15382,15420,15439,15477,15534,15553,15629,15705,15743,15819,15838,15857,15914,15952,16028,16040,16047,16104,16159,16161,16180,16182,16218,16237,16239,16244,16256,16275,16294,16313,16351,16389,16408,16411,16465,16522,16560,16598,16602,16617,16655,16677,16689,16750,16769,16778,16845,16883,16902,16997,17007,17035,17073,17092,17130,17206,17225,17263,17282,17301,17320,17339,17377,17415,17434,17510,17529,17567,17605,17661,17702,17719,17738,17753,17757,17776,17795,17833,17852,17890,17909,17985,17989,18004,18020,18039,18042,18061,18080,18102,18118,18134,18137,18156,18175,18270,18308,18346,18365,18399,18403,18418,18422,18441,18498,18517,18536,18555,18605,18621,18631,18669,18726,18821,18897,18954,18992,18996,19011,19014,19030,19049,19068,19087,19108,19125,19163,19277,19315,19391,19429,19467,19486,19524,19543,19562,19580,19600,19602,19638,19695,19714,19752,19790,19809,19847,19866,19885,19904,19923,19942,19961,19999,20018,20056,20075,20170,20189,20227,20228,20246,20280,20303,20322,20324,20359,20360,20364,20379,20398,20417,20436,20451,20493,20547,20550,20623,20626,20645,20664,20683,20702,20717,20721,20740,20816,20835,20873,20911,20930,20949,20968,21006,21010,21032,21063,21120,21122,21139,21177,21215,21234,21235,21253,21272,21315,21329,21348,21367,21386,21424,21443,21462,21481,21595,21614,21633,21671,21709,21728,21747,21839,21861,21880,21882,21911,21918,21937,21956,21975,21994,22013,22032,22070,22089,22108,22127,22150,22168,22203,22222,22257,22279,22300,22317,22336,22355,22393,22412,22431,22450,22469,22488,22500,22507,22526,22561,22564,22602,22621,22640,22661,22697,22716,22773,22792,22811,22817,22849,22906,22925,22963,22982,22986,23039,23096,23115,23131,23138,23172,23210,23248,23286,23324,23343,23381,23400,23438,23476,23495,23514,23524,23571,23590,23628,23647,23662,23685,23704,23745,23761,23799,23818,23894,23913,23932,23951,23989,24004,24024,24027,24046,24065,24103,24114,24141,24164,24200,24217,24236,24274,24312,24350,24369,24388,24407,24411,24426,24428,24430,24445,24464,24502,24503,24509,24521,24540,24559,24578,24616,24670,24673,24680,24692,24711,24730,24768,24863,24920,24958,25015,25034,25051,25053,25057,25060,25091,25106,25110,25111,25148,25167,25190,25224,25243,25281,25319,25342,25357,25361,25411,25414,25433,25490,25528,25547,25587,25598,25604,25642,25661,25699,25720,25737,25756,25772,25794,25832,25870,25889,25908,25920,25927,25946,25965,26041,26060,26079,26098,26117,26119,26136,26155,26174,26193,26269,26288,26345,26383,26438,26588,26846,26870,26894,26915,26930,26972,27095,27124,27181,27190,27281,27415,27428,27542,27580,27599,27606,27627,27683,27713,27732,27864,27943,27971,27998,28000,28014,28036,28093,28169,28283,28321,28378,28397,28466,28511,28537,28549,28587,28644,28704,28739,28744,28758,28774,28872,28929,28941,28986,29005,29062,29081,29138,29154,29176,29233,29309,29347,29366,29395,29480,29556,29610,29613,29693,29698,29727,29784,29822,29914,29955,30011,30040,30069,30098,30127,30156,30181,30185,30214,30243,30256,30272,30301,30330,30359,30417,30475,30504,30533,30562,30591,30620,30649,30678,30707,30736,30765,30794,30823,30852,30881,30910,30939,30968,30997,31026,31055,31084,31113,31142,31171,31200,31229,31287,31316,31345,31374,31403,31432,31461,31490,31519,31548,31577,31635,31664,31693,31722,31751,31780,31809,31838,31867,31896,31925,31954,31983,32012,32041,32070,32099,32128,32157,32186,32215,32244,32273,32302,32331,32360,32389,32418,32447,32476,32505,32534,32563,32592,32621,32650,32679,32708,32737,32766,32795,32824,32853,32882,32911,32940,32969,32998,33027,33056,33085,33114,33143,33172,33201,33230,33259,33288,33313,33317,33346,33375,33404,33433,33462,33491,33520,33549,33578,33607,33636,33665,33694,33723,33752,33781,33810,33839,33868,33897,33926,33955,33984,34013,34042,34071,34100,34112,34129,34158,34187,34216,34245,34274,34303,34332,34361,34390,34412,34419,34448,34477,34506,34535,34564,34593,34622,34651,34680,34709,34738,34767,34796,34825,34854,34883,34912,34941,34970,34999,35011,35100,35124,35174,35224,35263,35298,35337,35369,35371,35375,35411,35435,35482,35504,35510,35530,35568,35589,35648,35721,35780,35803,35839,35872,35910,35969,35999,36028,36070,36091,36151,36210,36281,36320,36333,36340,36399,36471,36501,36530,36589,36653,36712,36771,36842,36881,36901,36925,36960,36990,36998,36999,37021,37032,37091,37150,37190,37214,37251,37273,37332,37371,37374,37403,37462,37487,37501,37593,37624,37652,37711,37751,37753,37775,37811,37834,37859,37893,37913,37918,37921,37933,37964,38006,38023,38082,38154,38213,38261,38272,38314,38336,38361,38395,38433,38435,38454,38480,38525,38562,38584,38636,38643,38680,38715,38774,38805,38833,38875,38897,38956,39015,39057,39086,39145,39170,39178,39204,39244,39276,39308,39335,39363,39394,39458,39492,39517,39576,39618,39647,39673,39689,39706,39765,39793,39795,39837,39896,39955,39981,40080,40088,40092,40105,40134,40163,40192,40221,40250,40279,40308,40337,40366,40395,40424,40453,40482,40511,40540,40569,40598,40627,40656,40685,40714,40743,40772,40801,40830,40859,40888,40917,40946,40975,41004,41033,41062,41091,41120,41149,41178,41207,41222,41236,41294,41323,41352,41381,41410,41439,41468,41497,41526,41998,42027,42056,42085,42114,42174,42199,42203,42207,42225,42242,42264,42290,42309,42339,42367,42392,42414,42501,42525,42531,42576,42600,42605,42632,42651,42738,42764,42793,42813,42836,42855,42888,42925,42975,43050,43092,43104,43125,43157,43182,43212,43235,43287,43329,43341,43362,43407,43435,43436,43449,43473,43481,43500,43524,43554,43599,43653,43682,43686,43730,43761,43785,43795,43816,43836,43923,43950,43972,43973,43998,44073,44114,44160,44186,44196,44235,44283,44310,44346,44356,44397,44445,44472,44503,44547,44573,44634,44674,44691,44699,44709,44759,44766,44784,44844,44871,44907,44946,44971,45000,45021,45074,45094,45122,45142,45186,45203,45255,45261,45334,45391,45449,45471,45480,45494,45521,45564,45578,45605,45623,45636,45681,45682,45708,45758,45765,45810,45817,45823,45868,45895,45952,45991,45997,46006,46058,46081,46086,46169,46179,46197,46207,46232,46237,46244,46252,46262,46265,46285,46322,46356,46361,46366,46432,46458,46480,46509,46526,46542,46562,46626,46643,46660,46662,46704,46760,46788,46816,46821,46862,46864,46902,46986,47006,47072,47085,47109,47131,47132,47148,47179,47188,47248,47249,47263,47265,47295,47298,47311,47375,47425,47468,47471,47472,47483,47529,47537,47614,47619,47621,47647,47648,47653,47657,47674,47697,47702,47734,47776,47832,47874,47875,47889,47891,47939,47952,47966,47969,47972,48018,48037,48038,48082,48083,48103,48120,48157,48179,48209,48259,48293,48333,48379,48391,48396,48399,48456,48468,48483,48500,48534,48545,48561,48565,48642,48666,48709,48713,48731,48740,48865,48868,48883,48919,48942,48947,48967,49013,49043,49104,49168,49174,49188,49227,49248,49252,49254,49279,49316,49325,49340,49342,49396,49418,49438,49458,49468,49526,49556,49595,49625,49634,49773,49829,49846,49981,50148,50153,50154,50202,50216,50340,50343,50346,50384,50452,50540,50547,50642,50748,50813,50828,50898,50936,51036,51078,51095,51107,51192,51260,51262,51522,51552,51566,51589,51716,51740,51816,51920,52026,52067,52077,52098,52123,52595,53028,53045,53053,53086,53152,53209,53218,53238,53518,53546,53615,53637,53729,53823,53833,53941,53995,54011,54017,54064,54234,54342,54353,54365,54545,54747,54780,54787,54804,54842,54867,54878,55105,55184,55552,55745,55764,55917,55982,56076,56103,56193,56198,56228,56754,56833,57044,57113,57320,57390,57414,57475,57502,57552,57633,57635,57736,57835,57891,58098,58132,58160,58204,58208,58391,58404,58457,58519,58581,58589,58640,58657,58851,58878,58897,58936,58953,59031,59048,59137,59239,59240,59249,59397,59409,59470,59477,59535,59711,59714,59716,60021,60121,60176,60190,60226,60275,60341,60366,60423,60493,60507,60562,60748,60862,60893,60897,60910,60948,61051,61079,61115,61126,61205,61312,61444,61735,61797,61867,61868,61999,62041,62138,62187,62299,62332,62438,62465,62503,62536,62639,62873,62900,62963,63103,63239,63414,63457,63513,63612,63712,63716,63717,63742,63878,64024,64082,64092,64108,64332,64364,64472,64701,64977,65018,65042,65116,65307,65342,65582,65600,65832,66090,66092,66097,66098,66110,66310,66533,66596,66920,66964,67092,67122,67175,67266,67421,67472,67568,67649,67658,67707,67766,67768,67810,67863,68046,68103,68158,68181,68269,68374,68482,68487,68506,68635,68647,68649,68671,68683,68689,68717,68831,69026,69033,69201,69311,69433,69700,69761,69803,69827,69896,69950,69974,70017,70118,70152,70303,70379,70590,70630,70806,70875,70916,71012,71030,71036,71348,71444,71612,71650,71721,71742,71863,71915,72108,72123,72146,72296,72363,72789,72848,73027,73060,73084,73226,73234,73270,73275,73369,73495,73504,73595,73677,73781,73878,74132,74253,74348,74418,74454,74563,74801,74813,74878,74911,74917,74979,74982,75207,75275,75353,75464,75542,75645,75915,75944,76094,76188,76191,76267,76320,76417,76602,76662,76906,76923,76964,76973,77035,77086,77133,77146,77188,77223,77399,77416,77500,77544,77546,77576,77696,77795,77943,77999,78109,78134,78189,78225,78226,78363,78493,78535,78553,78752,78774,78795,79033,79394,79461,79542,79600,79643,79707,79912,80051,80056,80065,80143,80492,80583,80807,80859,80985,81048,81058,81106,81126,81158,81176,81242,81291,81332,81672,81819,81884,81909,82036,82185,82224,82238,82241,82341,82455,82596,82598,82882,82977,83129,83147,83157,83181,83191,83217,83256,83260,83268,83289,83611,83653,83712,83781,83805,83822,83971,83978,84102,84115,84177,84294,84322,84535,84619,84700,84708,84736,84884,84909,84927,84949,84973,85009,85108,85270,85398,85491,85742,85782,85842,86007,86080,86141,86185,86205,86300,86535,86648,86656,86690,86729,86772,86920,86960,87081,87155,87181,87267,87272,87370,87403,87431,87517,87622,87680,87711,87726,87871,87893,87953,88001,88174,88175,88346,88390,88421,88545,88597,88671,88675,88684,88738,88755,88798,88829,88869,89152,89177,89198,89223,89266,89272,89344,89375,89397,89488,89511,89700,89880,89984,90184,90331,90355,90434,90435,90618,90677,90778,91218,91314,91367,91438,91595,91647,91772,91783,91848,91940,92007,92039,92127,92153,92222,92357,92593,92599,92620,92631,92635,92656,92676,92678,92683,92707,92898,92952,93133,93160,93198,93233,93326,93456,93726,93789,93847,93863,93907,93914,93922,93971,94062,94128,94389,94410,94420,94439,94754,94761,94890,94911,94924,94943,94980,95150,95293,95513,95542,95590,95785,95799,95801,95839,95931,95944,95949,96044,96139,96151,96213,96234,96324,96413,96547,96575,96616,96622,96654,96741,96945,96948,97094,97119,97143,97188,97309,97346,97412,97551,97676,97746,97776,97785,97821,98093,98102,98191,98211,98213,98364,98580,98599,98602,98979,99142,99231,99440,99562,99591,99619,99623,99628,99661,99702,99714,99727,99763,99788,99888",
          "70,74,143,175,194,242,1737,1934,1999,2001,2112,2190,2377,2427,2476,2602,2694,2915,3189,3275,3468,3557,3576,3671,3751,3890,3926,3990,4013,4032,4070,4136,4163,4212,4330,4370,4462,4600,4796,5093,6103,6287,6635,6731,6977,7547,7661,8117,8136,8155,8554,8573,8649,8668,8673,8687,8808,8972,9107,9143,9147,9181,9293,9430,9580,9618,9637,9642,9656,9694,9867,9903,9998,10066,10074,10093,10155,10188,10207,10226,10321,10397,10549,10586,10644,10777,10894,10967,10990,11024,11083,11100,11118,11147,11157,11195,11214,11280,11301,11309,11347,11356,11384,11423,11461,11499,11564,11594,11708,11765,11803,11879,11955,12012,12031,12035,12098,12107,12126,12183,12221,12240,12316,12394,12411,12421,12449,12506,12536,12542,12587,12620,12625,12677,12696,12701,12789,12905,12908,12915,12924,13038,13095,13152,13168,13209,13222,13247,13266,13285,13323,13456,13475,13532,13578,13589,13646,13722,13760,13779,13798,13817,13931,13950,13988,14064,14178,14197,14235,14273,14292,14330,14368,14400,14425,14449,14480,14504,14520,14682,14805,14862,14901,14938,14995,15011,15066,15071,15204,15299,15327,15333,15337,15356,15394,15451,15470,15527,15554,15622,15698,15755,15831,15869,15888,15907,15945,15964,16097,16119,16249,16325,16344,16363,16420,16439,16496,16515,16553,16554,16564,16568,16572,16610,16648,16667,16734,16800,16811,16819,16895,16914,16933,16990,17047,17085,17180,17218,17237,17284,17313,17370,17389,17522,17541,17617,17634,17636,17674,17677,17712,17750,17807,17845,17864,17902,17940,17959,17978,18092,18111,18149,18206,18244,18301,18320,18339,18358,18396,18472,18491,18510,18562,18574,18624,18662,18700,18757,18852,18947,19042,19099,19118,19137,19251,19287,19297,19327,19384,19422,19448,19460,19515,19555,19612,19631,19669,19725,19745,19783,19802,19859,19878,19916,19935,19954,19992,20030,20086,20125,20140,20163,20193,20239,20315,20372,20391,20410,20429,20448,20503,20505,20562,20600,20619,20733,20752,20866,20871,20901,20923,20942,21075,21113,21132,21170,21227,21265,21379,21417,21474,21493,21550,21588,21721,21750,21759,21816,21892,22044,22120,22139,22158,22196,22234,22253,22272,22348,22386,22424,22462,22481,22519,22538,22633,22671,22747,22785,22804,22861,22899,22918,23051,23070,23127,23203,23220,23260,23336,23355,23393,23431,23437,23488,23556,23583,23640,23646,23659,23786,23790,23792,23819,23944,24001,24039,24058,24096,24115,24172,24210,24248,24267,24343,24362,24400,24495,24571,24609,24628,24666,24761,24780,24791,24837,24913,24970,25160,25198,25236,25293,25312,25350,25388,25426,25445,25464,25523,25524,25540,25578,25654,25730,25749,25825,25844,25939,25996,26034,26072,26129,26164,26186,26195,26205,26243,26319,26326,26338,26376,26395,26433,26452,26471,26604,26642,26660,26693,26699,26756,26760,26832,26984,27152,27174,27304,27307,27342,27350,27390,27412,27421,27725,27763,27764,27991,28029,28390,28429,28545,28561,28618,28752,29018,29047,29055,29093,29279,29487,29511,29756,29872,29948,29967,30004,30062,30091,30207,30236,30265,30294,30323,30326,30410,30439,30497,30555,30584,30613,30671,30700,30729,30758,30816,30932,30961,31019,31048,31077,31106,31193,31222,31309,31338,31396,31454,31483,31512,31599,31628,31657,31686,31715,31744,31773,31802,31860,31889,31918,31947,31976,32005,32034,32063,32092,32121,32150,32208,32237,32266,32295,32411,32556,32585,32614,32643,32672,32701,32730,32759,32788,32817,32846,32875,32904,32991,33020,33049,33078,33136,33165,33194,33223,33252,33281,33339,33397,33426,33455,33542,33571,33600,33629,33658,33745,33774,33803,33832,33861,33890,33919,33977,34006,34035,34064,34151,34238,34354,34383,34441,34470,34499,34586,34615,34673,34702,34789,34905,34934,34963,35052,35080,35167,35250,35303,35330,35388,35404,35475,35523,35549,35641,35747,35773,35903,35938,35962,36084,36144,36203,36392,36508,36523,36582,36632,36705,36760,36764,36794,36835,36876,36894,36931,36953,36982,37025,37080,37084,37143,37207,37246,37262,37266,37315,37325,37396,37455,37586,37645,37684,37768,37827,37872,37878,37917,37926,37957,38016,38113,38206,38242,38429,38434,38447,38497,38514,38518,38570,38767,38826,38890,38949,39008,39138,39164,39197,39265,39328,39361,39372,39387,39569,39636,39640,39699,39758,39787,39889,40176,40205,40234,40263,40321,40350,40408,40437,40466,40495,40524,40611,40640,40698,40727,40756,40785,40843,40901,40930,41017,41075,41104,41162,41220,41307,41365,41394,41423,41440,41452,41510,41539,42009,42038,42067,42125,42142,42208,42301,42331,42350,42425,42462,42479,42512,42587,42641,42749,42899,42934,42986,43005,43008,43021,43036,43074,43099,43136,43156,43200,43223,43298,43331,43373,43393,43516,43535,43550,43565,43610,43697,43934,44050,44084,44131,44152,44193,44246,44265,44385,44408,44481,44483,44528,44589,44645,44666,44676,44686,44735,44795,44821,44852,45013,45045,45075,45076,45214,45272,45345,45402,45485,45532,45589,45647,45719,45787,45834,45853,45906,45925,45963,46041,46044,46181,46182,46184,46196,46205,46390,46446,46531,46580,46609,46663,46692,46722,46755,46785,46787,46843,46993,47002,47010,47023,47030,47095,47116,47171,47172,47235,47239,47308,47323,47402,47415,47493,47592,47695,47830,47865,47899,47920,47990,47994,48064,48106,48125,48258,48265,48331,48435,48445,48700,48719,48796,48824,48842,49054,49111,49178,49257,49281,49292,49302,49310,49399,49455,49463,49539,49562,49649,49662,49689,49691,49711,49712,49807,49887,50015,50083,50098,50284,50326,50697,50862,50902,50996,51118,51370,51482,51605,51734,51814,51879,51897,52252,52303,52369,52522,52529,52615,52648,52727,52732,52746,52881,53216,53306,53312,53382,53439,53441,53632,53797,54148,54233,54247,54387,54460,54483,54487,54492,54517,54616,54708,54716,54756,54914,54943,55043,55134,55213,55219,55362,55435,55444,55685,55737,55797,55916,55946,55988,56063,56126,56134,56194,56200,56205,56269,56279,56476,56512,56665,56718,56741,56777,57006,57048,57050,57076,57114,57192,57202,57215,57331,57434,57551,57681,57685,57777,58047,58110,58152,58364,58412,58462,58542,58543,58627,58635,58646,58654,58661,58663,58675,58760,58812,59021,59036,59134,59140,59149,59162,59163,59202,59482,59496,59671,59672,59742,59892,59921,60040,60134,60636,60779,60785,60838,60877,60886,60938,61013,61063,61099,61108,61136,61157,61197,61319,61334,61509,61544,61666,61699,61725,61738,61844,61919,62038,62057,62089,62224,62227,62278,62345,62357,62531,62568,62676,62740,62843,63015,63029,63037,63105,63381,63393,63472,63492,63545,63585,63608,63682,63927,63957,63981,63995,63999,64015,64126,64127,64171,64337,64439,64520,64653,64690,64810,64821,64884,64891,64903,64975,64997,65019,65191,65294,65300,65336,65358,65411,65423,65431,65478,65586,65607,65614,65626,65642,65738,65769,66246,66296,66535,66604,66745,66809,66841,66903,66914,66932,66962,67289,67374,67572,67611,67945,67960,67983,68061,68177,68185,68200,68808,69212,69334,69388,69400,69401,69427,69549,69577,69654,69732,69758,69790,69825,69994,70056,70080,70181,70268,70292,70296,70367,70458,70487,70603,70688,70696,70772,70857,70859,70960,71207,71342,71360,71459,71542,71684,71685,71768,71774,71948,72036,72050,72090,72176,72191,72329,72372,72418,72567,72586,72735,72775,72892,72924,73107,73119,73349,73428,73485,73549,73775,73821,73888,73969,73976,74001,74085,74091,74212,74241,74267,74344,74445,74460,74462,74514,74615,74626,74764,74824,74876,75012,75185,75299,75386,75410,75446,75570,75589,75905,75954,76133,76144,76148,76186,76222,76380,76635,76749,76809,76873,77070,77156,77163,77175,77386,77464,77580,77697,77731,77904,78335,78341,78432,78462,78475,78497,78583,78591,78693,79006,79018,79275,79341,79381,79427,79460,79642,79650,79785,79894,79908,79990,80013,80274,80466,80469,80487,80536,80572,80736,81004,81019,81097,81125,81306,81396,81413,81422,82149,82468,82478,82519,82580,82667,82672,82718,82740,82807,82903,82982,83017,83112,83272,83389,83401,83403,83478,83694,83718,83725,83856,83988,84043,84089,84451,84466,84577,84628,84652,84686,84694,85010,85058,85154,85286,85294,85406,85669,85845,85910,85970,86226,86258,86260,86364,86477,86492,86652,86658,86681,86797,86961,87001,87027,87060,87258,87331,87387,87407,87510,87576,87581,87737,87739,87794,87843,87852,88082,88464,88583,88602,88839,88947,88958,88962,89003,89040,89056,89167,89170,89307,89308,89469,89747,89805,89882,89894,90098,90134,90176,90248,90345,90372,90381,90436,90818,90854,90981,90993,91039,91227,91267,91330,91425,91442,91528,91574,91620,91677,91733,91764,91829,91859,91901,91970,92074,92117,92187,92272,92276,92278,92350,92516,92602,92604,92617,92806,92828,92856,92950,92994,93057,93105,93112,93118,93120,93166,93270,93406,93476,93678,93724,93752,93760,93884,93902,94081,94166,94245,94253,94303,94307,94377,94407,94472,94487,94504,94543,94646,94759,94821,94921,94925,94937,94983,95174,95183,95256,95374,95459,95519,95618,95695,95733,95805,95821,95876,96015,96102,96128,96405,96420,96422,96656,96728,96791,96806,96850,96865,96889,96982,97459,97556,97747,97932,97975,97981,97986,98017,98028,98062,98108,98142,98165,98195,98249,98252,98379,98446,98475,98511,98744,98766,98897,99199,99239,99250,99265,99267,99370,99423,99466,99557,99568,99572,99575,99643,99775,99942,99991,99997",
          "126,162,167,1647,1663,1986,2078,2100,2172,2484,2899,2986,3171,3209,3259,3645,3664,3782,3811,4038,4057,4093,4111,4184,4202,4274,4350,4353,4362,4395,4718,4923,4983,6199,6324,6916,8370,9472,9524,9548,9567,9586,9605,9624,9643,9662,9681,9757,9774,9795,9985,10042,10080,10673,10688,10699,10707,10742,10764,10968,11011,11220,11258,11372,11391,11458,11486,11505,11549,11552,11562,11638,11784,11847,11866,11885,11910,11933,11942,11961,11965,11999,12006,12037,12085,12113,12148,12151,12170,12198,12208,12214,12227,12246,12265,12284,12398,12405,12417,12476,12493,12512,12550,12553,12569,12607,12626,12664,12740,12757,12778,12854,12868,12873,12886,12940,12944,12949,12968,12987,13025,13062,13101,13106,13139,13177,13196,13233,13234,13253,13386,13443,13474,13553,13576,13614,13652,13671,13690,13728,13766,13785,13823,13895,13899,13979,13982,14127,14146,14184,14203,14222,14241,14260,14279,14298,14317,14355,14364,14374,14428,14507,14526,14602,14697,14811,14944,15024,15035,15077,15096,15103,15119,15191,15229,15267,15305,15362,15400,15419,15437,15457,15461,15476,15495,15514,15564,15571,15609,15628,15685,15704,15723,15742,15745,15762,15780,15837,15875,15894,15951,15970,15989,16008,16010,16033,16065,16083,16103,16116,16122,16155,16160,16198,16217,16236,16255,16274,16312,16331,16350,16369,16407,16426,16464,16517,16521,16556,16578,16597,16600,16673,16692,16712,16749,16768,16787,16806,16879,16901,16907,16939,16951,16959,17015,17053,17129,17167,17186,17224,17281,17300,17338,17342,17396,17398,17403,17414,17509,17528,17540,17604,17651,17670,17680,17682,17693,17718,17736,17737,17756,17830,17832,17851,17870,17889,17908,17932,17935,17946,17965,18003,18010,18060,18073,18078,18083,18117,18136,18193,18199,18212,18231,18253,18288,18307,18345,18364,18383,18402,18421,18434,18459,18478,18516,18529,18544,18548,18554,18573,18595,18649,18668,18687,18744,18763,18801,18820,18832,18858,18863,18896,18906,18934,18948,18972,18991,19008,19010,19013,19026,19048,19067,19078,19086,19105,19143,19144,19169,19181,19219,19257,19314,19358,19371,19409,19466,19485,19504,19542,19633,19637,19675,19694,19789,19808,19827,19864,19884,19903,19922,19960,19969,19979,20017,20036,20055,20093,20131,20169,20207,20243,20245,20283,20321,20340,20353,20378,20416,20435,20454,20473,20492,20511,20530,20568,20606,20610,20625,20701,20720,20739,20834,20853,20891,20910,20948,20967,21005,21043,21062,21068,21102,21119,21148,21156,21157,21176,21196,21214,21233,21252,21271,21322,21328,21347,21385,21404,21423,21442,21461,21480,21537,21572,21575,21651,21669,21670,21689,21708,21746,21803,21822,21841,21860,21879,21898,21917,21936,21955,21974,21993,22012,22031,22063,22145,22183,22202,22230,22259,22278,22292,22316,22335,22345,22354,22373,22392,22411,22430,22487,22506,22544,22550,22563,22582,22614,22620,22677,22734,22772,22791,22848,22867,22880,22924,22929,22941,22962,22981,22992,23000,23057,23061,23095,23114,23152,23167,23171,23190,23228,23285,23342,23361,23370,23380,23418,23475,23494,23532,23551,23570,23608,23684,23691,23703,23741,23760,23764,23772,23798,23817,23855,23860,23874,23893,23931,23950,23969,23988,24007,24026,24045,24051,24083,24099,24102,24140,24191,24193,24197,24216,24234,24235,24253,24254,24261,24266,24273,24292,24311,24330,24368,24387,24406,24425,24463,24482,24501,24539,24558,24596,24615,24653,24672,24691,24729,24767,24786,24802,24843,24862,24919,24938,24957,24995,25017,25033,25071,25090,25109,25128,25166,25204,25242,25261,25299,25327,25353,25375,25394,25413,25432,25451,25470,25527,25545,25584,25603,25607,25622,25660,25698,25736,25755,25774,25850,25888,25907,25926,25964,26002,26003,26021,26078,26097,26116,26154,26192,26230,26249,26268,26306,26325,26364,26382,26398,26401,26420,26439,26496,26515,26637,26672,26692,26724,26795,26858,26914,26971,26990,27038,27066,27085,27104,27123,27135,27172,27237,27274,27351,27360,27368,27384,27423,27435,27485,27579,27728,27789,27826,27852,27897,27933,27959,27970,28016,28054,28073,28130,28149,28243,28263,28278,28320,28358,28377,28392,28396,28415,28468,28567,28624,28643,28777,28795,28833,28852,28909,28947,28985,29037,29118,29174,29194,29214,29232,29270,29437,29460,29517,29552,29594,29612,29631,29633,29650,29745,29783,29795,29935,29937,30010,30039,30068,30097,30126,30155,30184,30213,30242,30271,30300,30322,30329,30358,30387,30416,30445,30474,30503,30532,30561,30619,30648,30677,30706,30735,30764,30793,30822,30851,30880,30909,30938,30967,30996,31025,31054,31083,31112,31141,31170,31199,31228,31257,31286,31315,31344,31373,31402,31431,31460,31489,31518,31547,31576,31605,31634,31663,31692,31721,31750,31779,31808,31837,31866,31895,31924,31953,31982,32011,32040,32069,32098,32127,32156,32185,32214,32243,32272,32301,32330,32359,32388,32417,32446,32475,32504,32533,32562,32591,32620,32649,32678,32707,32736,32765,32794,32823,32852,32881,32910,32939,32968,32997,33026,33055,33084,33113,33142,33171,33200,33229,33258,33287,33316,33345,33374,33403,33432,33461,33490,33519,33548,33577,33606,33635,33664,33693,33722,33751,33780,33809,33838,33867,33896,33925,33954,33983,34012,34041,34070,34099,34128,34157,34186,34215,34244,34273,34302,34331,34360,34389,34418,34447,34476,34505,34534,34563,34592,34621,34650,34679,34708,34737,34766,34795,34824,34853,34882,34911,34940,34969,34998,35010,35035,35057,35072,35084,35089,35099,35133,35173,35219,35220,35244,35262,35304,35336,35410,35437,35481,35529,35588,35615,35647,35672,35720,35779,35838,35859,35909,35935,35968,36027,36053,36090,36150,36187,36209,36253,36270,36280,36339,36398,36430,36470,36505,36529,36588,36625,36652,36711,36746,36770,36841,36874,36877,36900,36959,37031,37090,37149,37213,37272,37331,37361,37402,37424,37461,37520,37592,37651,37710,37774,37833,37860,37892,37963,38022,38081,38153,38178,38212,38271,38294,38318,38335,38359,38394,38414,38453,38499,38524,38546,38583,38642,38714,38741,38773,38832,38857,38896,38955,38979,39014,39085,39144,39181,39182,39203,39275,39334,39393,39413,39416,39457,39516,39538,39575,39646,39685,39705,39728,39764,39790,39796,39836,39895,39954,40121,40150,40179,40208,40237,40266,40295,40324,40353,40382,40411,40440,40469,40498,40527,40556,40585,40614,40643,40672,40701,40730,40759,40788,40817,40846,40875,40904,40933,40962,40991,41020,41049,41078,41107,41136,41165,41194,41223,41252,41281,41310,41339,41368,41397,41426,41455,41484,41513,41542,42012,42041,42070,42099,42128,42147,42188,42213,42220,42278,42305,42353,42388,42428,42515,42548,42568,42590,42615,42619,42665,42688,42696,42752,42797,42827,42839,42902,42912,42917,42951,42970,42989,43003,43010,43064,43096,43139,43178,43192,43226,43252,43301,43314,43376,43463,43509,43511,43538,43592,43613,43700,43775,43794,43850,43900,43937,44012,44033,44052,44055,44087,44113,44141,44174,44249,44324,44361,44375,44411,44431,44486,44529,44561,44648,44683,44723,44765,44798,44860,44885,44960,44972,45035,45108,45136,45156,45217,45239,45275,45290,45298,45348,45405,45463,45489,45535,45565,45567,45592,45627,45650,45659,45667,45722,45779,45837,45909,45966,45979,46012,46013,46021,46036,46040,46070,46103,46134,46139,46140,46145,46215,46236,46255,46260,46297,46353,46360,46368,46373,46413,46425,46431,46461,46478,46498,46504,46510,46513,46552,46557,46593,46596,46608,46630,46669,46726,46782,46800,46807,46829,46889,46892,46899,46907,46927,46944,46950,46997,47012,47028,47038,47042,47081,47082,47102,47104,47120,47187,47203,47225,47237,47247,47279,47313,47318,47334,47344,47365,47373,47385,47412,47431,47447,47449,47490,47584,47596,47640,47658,47673,47746,47757,47760,47786,47800,47806,47852,47858,47905,47922,47947,47950,47953,48100,48129,48210,48275,48291,48299,48323,48332,48366,48388,48407,48448,48450,48522,48538,48553,48555,48620,48661,48708,48728,48743,48838,48940,48962,48973,48974,48992,49031,49047,49099,49149,49201,49210,49295,49360,49378,49381,49401,49426,49445,49464,49490,49581,49587,49613,49641,49684,49794,49812,49893,49910,49952,50061,50126,50604,50660,50746,50840,50852,51021,51034,51066,51123,51141,51152,51197,51200,51202,51276,51462,51618,51706,51768,51813,51859,52030,52031,52032,52167,52278,52353,52530,52559,52728,52777,52838,52869,52897,52971,53077,53154,53158,53267,53373,53374,53517,53526,53533,53618,53626,53641,53736,53744,53771,53879,54122,54285,54287,54305,54313,54334,54338,54443,54472,54474,54827,54897,55167,55287,55298,55336,55338,55355,55592,55617,55649,55877,55981,56151,56480,56493,56502,56563,56644,56649,56736,56828,56861,56879,56904,56916,56943,56977,57439,57450,57642,57649,57662,57671,57718,57768,57823,57926,58031,58056,58083,58105,58113,58178,58213,58238,58287,58375,58396,58402,58425,58548,58550,58617,58719,58731,58745,58746,58899,58945,59037,59268,59324,59490,59497,59585,59612,59620,59626,59687,59767,59789,59888,60212,60249,60269,60339,60420,60523,60640,60682,60741,60857,60863,60986,61114,61242,61249,61288,61569,61576,61726,61727,61737,61739,61770,61785,61822,61859,61893,61959,61977,62029,62157,62200,62257,62308,62418,62455,62563,62697,62779,62871,63085,63223,63235,63269,63276,63451,63460,63467,63470,63526,63549,63576,63605,63656,63970,63990,64004,64047,64113,64395,64443,64745,64776,64777,65026,65090,65091,65165,65267,65279,65327,65352,65467,65468,65698,65703,65852,65876,66025,66126,66153,66164,66286,66360,66389,66405,66570,66598,66749,66770,66860,66943,66947,66971,67344,67382,67439,67498,67501,67508,67625,67661,67759,67789,67811,67846,67879,67916,68078,68088,68172,68184,68201,68263,68266,68287,68412,68419,68474,68531,68556,68561,68676,68730,68791,68812,68887,68901,68907,68941,69014,69023,69120,69261,69280,69358,69450,69487,69512,69702,69853,69866,69920,70198,70205,70224,70506,70691,70705,70710,70830,70934,71052,71238,71244,71347,71390,71492,71609,71819,71847,71848,71856,72017,72049,72054,72091,72256,72376,72396,72456,72512,72666,72675,72988,73047,73096,73111,73198,73297,73376,73416,73474,73653,73660,73852,73906,74067,74199,74390,74548,75131,75363,75409,75462,75501,75571,75574,75581,75592,75626,75732,75736,75800,75818,75830,75845,75848,75989,76045,76159,76206,76271,76702,76835,77072,77303,77499,77564,77672,77689,77722,77927,77956,78057,78191,78223,78234,78247,78282,78492,78550,78632,78674,78700,78843,78967,79177,79347,79389,79541,79579,79735,79830,79861,79887,79939,80275,80336,80368,80438,80451,80718,80813,80883,80960,80972,81018,81122,81201,81315,81326,81437,81455,81537,81593,81629,81648,81840,81847,82099,82107,82222,82240,82528,82719,82838,82848,82972,83048,83055,83086,83128,83231,83265,83500,83528,83647,83657,83749,83755,83902,83931,83940,84112,84114,84183,84220,84225,84274,84377,84433,84478,84637,84726,85002,85135,85251,85377,85382,85442,85524,85803,85833,86087,86173,86251,86360,86363,86437,86495,86507,86687,86715,87012,87040,87270,87340,87344,87556,87624,87627,87641,87838,87911,87963,87980,88091,88130,88172,88377,88379,88384,88702,88717,88747,88892,88967,89008,89011,89134,89323,89405,89462,89687,89692,89766,89796,89837,89957,90024,90131,90132,90413,90485,90609,90617,90645,90663,90955,90975,91136,91166,91208,91225,91228,91339,91385,91390,91571,91614,91646,91678,91736,91886,91978,91994,92027,92037,92077,92186,92354,92409,92459,92474,92527,93126,93243,93348,93494,93550,93649,93854,93881,93949,93961,94040,94108,94213,94262,94302,94315,94431,94460,94508,94512,94594,94599,95046,95170,95198,95261,95345,95460,95538,95608,95764,95806,95840,95853,95955,95982,95998,95999,96030,96184,96290,96403,96532,96558,96837,96989,97135,97264,97318,97325,97363,97424,97436,97453,97467,97743,97800,97804,97839,97892,97903,97952,98014,98022,98079,98088,98378,98441,98544,98548,98671,98769,98854,98894,98910,98912,98939,98940,99017,99038,99082,99084,99134,99140,99167,99171,99204,99255,99381,99672,99787,99833,99883,99904,99925",
          "18,51,55,99,121,231,272,565,836,1165,1664,1703,1880,1900,1902,1920,1945,1951,1965,1971,2003,2014,2015,2038,2049,2066,2118,2146,2203,2223,2311,2327,2346,2365,2371,2442,2454,2467,2468,2470,2481,2488,2523,2537,2541,2600,2603,2610,2623,2669,2724,2734,2743,2849,2871,2895,2955,2969,2970,2972,2975,2990,2999,3000,3003,3030,3036,3056,3058,3061,3071,3133,3142,3144,3169,3201,3203,3231,3257,3262,3265,3272,3279,3313,3337,3354,3387,3420,3448,3450,3457,3467,3475,3480,3482,3512,3553,3573,3627,3643,3675,3699,3709,3803,3854,3889,3895,3902,3910,3931,3935,3945,3963,3971,3976,3977,3980,3991,4003,4006,4009,4016,4022,4030,4033,4035,4053,4060,4062,4074,4075,4086,4117,4126,4129,4147,4154,4244,4245,4285,4310,4327,4337,4394,4407,4413,4435,4453,4492,4551,4589,4635,4677,4701,4766,4807,4860,4872,4879,4914,4930,4935,4944,4957,4958,4959,4965,4967,4973,5036,5112,5256,5619,5649,5980,6027,6077,6124,6129,6156,6321,6338,6345,6364,6367,6529,6582,6668,6688,6750,6846,6855,6864,6883,7032,7040,7146,7154,7292,7352,7444,7552,7564,7585,7599,7623,7670,7680,7733,7758,7795,7804,7833,7834,7912,7925,7936,7983,8078,8123,8160,8205,8206,8230,8242,8256,8340,8365,8437,8464,8543,9011,9086,9155,9162,9276,9314,9414,9491,9654,9655,9723,9787,9909,10098,10186,10219,10324,10373,10541,10555,10609,10706,10726,10846,10865,10870,10882,10926,11068,11071,11140,11246,11602,11761,11780,11972,12083,12091,12142,12172,12210,12378,12482,12638,12678,12692,12759,12981,13128,13461,13543,13792,14012,14175,14202,14257,14419,14455,14578,14648,14792,14919,14984,15055,15216,15266,15338,15438,15568,15617,15739,15812,15877,15880,15901,15912,16079,16162,16302,16382,16444,16624,16629,16826,17060,17149,17159,17161,17174,17374,17508,17542,17992,18044,18246,18333,18336,18497,18741,19003,19404,19421,19423,19577,19606,19674,19718,19849,19897,20013,20060,20453,20474,20724,20920,21053,21067,21297,21400,21401,21608,21652,21694,21727,21792,21819,21916,22003,22021,22155,22308,22356,22766,22887,23363,23444,23526,23595,23627,23823,23851,23885,23942,23982,23994,24075,24321,24386,24392,24655,24792,24933,25010,25084,25141,25147,25229,25235,25294,25434,25480,25589,25979,25995,26134,26266,26282,26480,26540,26542,26557,26559,26580,26603,26615,26681,26769,26813,26815,26929,26989,27075,27084,27211,27259,27326,27422,27426,27481,27506,27510,27526,27528,27529,27535,27570,27638,27936,27974,27989,28001,28009,28084,28086,28124,28132,28144,28156,28264,28418,28430,28480,28544,28576,28580,28640,28651,28661,28750,29258,29338,29438,29494,29532,29560,29577,29664,29676,29723,29806,29849,29859,29972,30016,30045,30074,30103,30132,30161,30190,30219,30248,30277,30306,30335,30364,30393,30422,30451,30480,30509,30538,30567,30625,30654,30683,30712,30741,30770,30799,30857,30886,30915,30944,30973,31002,31031,31060,31089,31118,31147,31176,31205,31234,31263,31292,31321,31350,31379,31408,31437,31466,31495,31524,31553,31582,31611,31640,31669,31698,31727,31756,31785,31814,31843,31872,31901,31930,31959,31988,32017,32046,32075,32104,32133,32162,32191,32220,32249,32278,32307,32336,32365,32394,32423,32452,32481,32510,32539,32568,32597,32626,32655,32684,32713,32742,32771,32800,32829,32858,32887,32916,32945,32974,33003,33032,33061,33090,33119,33148,33177,33206,33235,33264,33293,33322,33351,33380,33409,33438,33467,33496,33554,33583,33612,33641,33670,33699,33728,33786,33815,33844,33873,33902,33931,33960,33989,34018,34047,34076,34105,34134,34163,34192,34250,34279,34308,34337,34366,34395,34424,34453,34482,34511,34540,34569,34598,34627,34656,34685,34714,34743,34772,34801,34830,34859,34888,34917,34946,34975,35066,35105,35144,35179,35195,35268,35309,35342,35356,35367,35389,35416,35431,35461,35487,35507,35535,35566,35594,35653,35726,35785,35810,35884,35915,35933,35974,35991,36033,36096,36115,36156,36215,36235,36244,36249,36286,36345,36370,36388,36404,36476,36535,36594,36658,36717,36776,36847,36861,36906,36965,37037,37055,37096,37155,37219,37278,37337,37408,37467,37503,37510,37526,37598,37641,37657,37716,37750,37780,37839,37898,37912,37915,37969,37989,38028,38087,38159,38218,38248,38277,38306,38341,38400,38438,38459,38476,38530,38589,38623,38720,38749,38779,38838,38879,38902,38961,39020,39049,39091,39150,39209,39225,39281,39340,39399,39417,39463,39481,39497,39522,39581,39621,39652,39683,39686,39711,39770,39842,39901,39960,39976,40127,40156,40185,40214,40243,40272,40301,40330,40359,40388,40417,40446,40475,40504,40533,40562,40591,40620,40649,40678,40707,40736,40765,40794,40823,40852,40881,40910,40939,40968,40997,41026,41055,41084,41113,41142,41171,41200,41229,41258,41287,41316,41345,41374,41403,41432,41461,41490,41519,41548,42018,42047,42076,42105,42134,42194,42217,42284,42313,42359,42380,42434,42441,42474,42486,42521,42533,42596,42642,42671,42675,42700,42758,42766,42769,42773,42777,42833,42878,42908,42945,42949,42961,42995,43027,43070,43073,43087,43097,43145,43148,43185,43203,43205,43232,43258,43260,43307,43322,43382,43395,43419,43429,43469,43544,43594,43619,43626,43645,43660,43706,43754,43781,43797,43856,43863,43882,43893,43943,44018,44027,44093,44180,44209,44255,44278,44282,44330,44390,44391,44417,44426,44436,44444,44492,44567,44604,44610,44622,44654,44673,44729,44751,44763,44804,44815,44891,44902,44966,44992,44997,45001,45060,45072,45088,45114,45162,45180,45181,45223,45229,45281,45354,45375,45411,45469,45541,45598,45656,45728,45785,45803,45843,45851,45886,45915,45938,45972,46026,46078,46080,46106,46116,46143,46149,46195,46203,46209,46233,46250,46253,46266,46329,46342,46354,46456,46499,46512,46516,46559,46595,46599,46624,46638,46646,46648,46653,46698,46738,46777,46791,46823,46879,46952,47013,47031,47034,47039,47047,47057,47061,47127,47129,47134,47144,47163,47167,47219,47221,47222,47264,47271,47302,47303,47307,47310,47314,47361,47374,47386,47392,47428,47484,47579,47627,47634,47649,47655,47708,47715,47736,47782,47812,47834,47849,47873,47911,47938,47948,47970,47997,48017,48029,48050,48053,48132,48158,48239,48241,48246,48260,48289,48309,48360,48377,48395,48465,48501,48540,48577,48613,48637,48645,48669,48677,48710,48737,48771,48779,48784,48813,48816,48828,48850,48875,48898,48904,48916,48922,48972,49005,49009,49011,49035,49037,49050,49080,49085,49106,49113,49133,49143,49152,49170,49222,49231,49243,49259,49288,49297,49324,49334,49339,49364,49393,49404,49420,49434,49456,49549,49593,49686,49728,49744,49770,49795,49808,49850,49877,49900,49904,49914,49937,49961,49998,50065,50355,50423,50516,50541,50767,50884,50968,50970,50986,51046,51128,51130,51157,51180,51189,51322,51342,51458,51481,51560,51638,51667,51700,51717,51721,51970,52016,52048,52066,52137,52140,52145,52321,52481,52597,52613,52689,52814,53360,53383,53387,53393,53431,53530,53599,53834,53857,53990,54005,54119,54254,54297,54300,54359,54404,54608,54624,54689,54770,54800,54858,54879,55170,55194,55230,55424,55470,55492,55507,55508,55514,55614,55630,55642,55662,55677,55680,55694,55712,55747,55773,55840,55910,56000,56118,56138,56159,56445,56561,56700,56762,56779,56782,57031,57133,57185,57253,57260,57374,57409,57522,57656,57700,57887,57901,57927,58013,58072,58124,58170,58299,58334,58335,58384,58390,58456,58480,58587,58707,58708,58712,58744,58795,58836,58972,58988,59039,59194,59396,59406,59445,59503,59540,59741,59747,59900,60007,60034,60065,60205,60348,60358,60380,60381,60473,60517,60773,60835,61037,61103,61131,61135,61285,61302,61327,61441,61505,61517,61600,61690,61888,61967,62042,62081,62252,62319,62405,62456,62457,62461,62494,62590,62598,62720,62777,62798,62803,62943,62957,63022,63047,63290,63342,63361,63463,63522,63574,63690,63880,63894,63909,63923,63942,64089,64309,64440,64515,64516,64539,64668,64680,64725,64820,64851,64852,64864,64874,64899,64914,64986,64988,65010,65037,65082,65098,65099,65291,65338,65537,65616,65685,65768,65904,66087,66160,66206,66339,66340,66420,66459,66558,66712,66743,66746,66828,66871,66872,66905,67018,67020,67191,67196,67244,67383,67532,67599,67762,67778,67856,68028,68085,68117,68119,68196,68198,68242,68344,68405,68507,68558,68631,68690,68822,68899,68947,68982,69009,69048,69150,69376,69437,69622,69854,69922,70062,70129,70231,70250,70391,70502,70542,70632,71003,71020,71095,71117,71145,71409,71512,71687,71725,71757,71849,71892,71955,71980,72051,72102,72111,72153,72170,72295,72455,72481,72724,72759,72778,72900,72919,72926,73021,73077,73095,73103,73129,73228,73331,73368,73634,73689,74109,74152,74159,74162,74432,74744,74760,74818,74976,74981,74997,75030,75038,75111,75202,75249,75424,75540,75619,75625,75672,75744,75747,75769,75880,75922,75959,76155,76234,76284,76304,76375,76391,76422,76670,76675,76698,76795,76819,76883,76918,76969,76978,77010,77122,77124,77155,77284,77295,77408,77497,77719,77797,77960,78203,78283,78345,78482,78520,78636,78717,78742,78849,78960,78966,78978,79077,79126,79356,79496,79794,79815,79829,79982,80031,80064,80163,80237,80372,80486,80504,80562,80581,80628,80672,80731,80757,81008,81017,81050,81116,81128,81137,81221,81289,81490,81614,81675,81887,81919,81920,81989,82017,82088,82116,82266,82450,82539,82677,82835,82947,82970,83045,83165,83226,83259,83398,83460,83515,83608,83621,83835,83895,83921,84139,84293,84341,84359,84428,84496,84538,84899,84976,85005,85117,85124,85175,85449,85561,85606,85626,85645,85653,85658,85694,85737,85769,85788,85813,85943,85947,85982,85984,86035,86051,86082,86196,86210,86389,86431,86446,86459,86560,86760,86769,86833,86898,87113,87294,87618,87651,87740,87841,87885,87935,88018,88055,88103,88259,88329,88471,88493,88512,88531,88540,88640,88794,88838,88984,89113,89192,89211,89410,89495,89527,89557,89563,89565,89892,89971,90123,90238,90338,90351,90353,90405,90728,90737,90766,90930,91075,91083,91154,91207,91220,91355,91491,91588,91627,91691,91702,91762,91937,92076,92099,92416,92437,92446,92723,92743,92805,92855,93041,93042,93065,93128,93276,93575,93601,93684,93794,93806,93931,94192,94237,94277,94341,94406,94493,94523,94861,94916,94923,94926,94959,95034,95155,95165,95166,95272,95298,95300,95401,95435,95568,95727,95946,96049,96096,96106,96150,96161,96217,96340,96375,96435,96523,96561,96653,96663,96714,96758,97081,97106,97388,97430,97449,97625,97748,97769,97818,97902,98033,98077,98098,98126,98128,98140,98150,98154,98226,98322,98532,98609,98613,98630,98770,99043,99087,99110,99165,99198,99320,99401,99508,99552,99555,99620,99679,99686,99689,99764,99870,99886,99894,99952",
          "47,66,267,321,335,392,431,452,459,462,475,511,551,646,878,946,1075,1086,1140,1187,1391,1427,1612,1630,1642,1710,2656,2755,3330,4609,4976,5041,5162,5491,5509,5525,5608,5622,5666,5692,5695,5858,6072,6125,6256,6263,6295,6302,6469,6610,6676,6856,6928,7045,7236,7334,7377,7397,7550,7694,7703,7775,7939,7993,8111,8144,8153,8192,8359,8373,8524,8636,8695,8699,8702,8803,8847,8867,8933,8947,8974,9007,9128,9168,9169,9213,9219,9281,9353,9467,9853,9886,10023,10045,10084,10135,10286,10399,10406,10457,10492,10547,10584,10611,10635,10693,10873,10892,11007,11012,11045,11069,11259,11302,11330,11733,11740,11808,11813,11934,12094,12119,12178,12407,12720,12799,12927,12980,13051,13052,13086,13092,13141,13414,13508,13537,13618,13748,14088,14370,14633,14778,14795,14936,15019,15027,15157,15162,15180,15324,15435,15627,15909,16018,16539,16567,16579,16595,16659,17018,17246,17663,17733,17813,17952,18049,18094,18232,18627,18725,18772,18783,18895,18921,19153,19353,19363,19732,19926,20076,20089,20722,20796,20900,21036,21088,21277,21283,21405,21697,22304,23006,23234,23264,23387,23456,23479,23585,23779,23836,24145,24304,24349,24926,24951,25136,25174,25489,25515,26210,26327,26396,26458,26555,26668,26884,26941,26943,27013,27030,27385,27781,28172,28353,28374,28423,28837,28865,28894,28996,29190,29452,29492,29600,29700,29848,30864,30946,32253,33525,36431,36434,36635,36681,37018,37237,37319,37880,37994,38302,38552,38609,38614,38815,38882,39114,39300,39382,39690,39880,42524,44140,44986,45243,45444,45499,45601,45603,50168,50183,50206,50237,50246,50330,50542,50759,50836,50843,50937,50990,51001,51102,51350,51501,51529,51565,51654,51686,51699,51758,51759,51819,52090,52344,52348,52523,52547,52570,52625,52707,52760,52797,52841,52895,52927,52999,53063,53115,53120,53142,53309,53425,53510,53521,53568,53766,53826,53841,53897,53931,53936,53958,53978,54033,54065,54097,54184,54550,54591,54605,54717,54760,54801,55126,55129,55187,55208,55289,55371,55373,55416,55452,55610,55661,55878,55950,56081,56224,56231,56273,56368,56393,56503,56504,56691,56699,56704,56815,56937,56972,57397,57485,57504,57644,57648,57678,57800,58055,58073,58149,58242,58362,58545,58634,58665,58689,59022,59178,59331,59512,59514,59591,59638,59755,59827,59831,59930,60005,60019,60028,60074,60292,60480,60481,60545,60659,60668,60727,60800,60887,61110,61120,61123,61304,61325,61379,61420,61634,61957,62166,62373,62400,62458,62485,62552,62583,62790,62918,62967,62980,62991,63016,63054,63142,63173,63197,63243,63388,63394,63450,63480,63556,63582,63610,63625,63640,63752,63772,63795,63967,63974,63984,64005,64007,64050,64265,64300,64407,64410,64427,64463,64617,64660,64678,64710,64750,64796,64875,64934,64947,65060,65172,65236,65260,65375,65543,65792,66086,66114,66221,66284,66290,66328,66397,66462,66509,66620,66672,66716,66795,66810,67072,67145,67219,67273,67318,67364,67401,67481,67555,67569,67596,67688,67742,67973,68224,68254,68342,68382,68443,68643,68802,68902,68997,69126,69230,69232,69309,69636,69826,69918,69931,69964,70584,70606,70829,70842,70979,71041,71269,71407,71468,71553,71562,71651,71744,71929,72194,72218,72357,72385,72461,72617,72709,72813,72863,73140,73150,73263,73279,73365,73450,73506,73509,73538,73597,73640,73670,73739,74030,74041,74126,74151,74231,74260,74395,74402,74520,74565,74591,74720,74919,74973,75055,75165,75218,75246,75290,75322,75598,76092,76139,76360,76563,76652,76663,76674,76734,76782,76796,76944,76988,76993,77107,77244,77339,77510,77512,77574,77666,77721,77771,77799,77816,77860,77876,78000,78043,78044,78045,78164,78173,78235,78379,78452,78517,78533,78544,78575,78691,78716,78793,78924,78935,79088,79296,79349,79476,79558,79732,79789,79832,79846,79905,79937,79965,80038,80172,80423,80685,80754,80756,80772,80839,80855,80977,81063,81367,81511,81692,81704,81774,81811,82000,82175,82193,82206,82251,82380,82397,82400,82529,82552,82554,82686,82729,82749,82797,83078,83149,83352,83414,83445,83454,83471,83486,83681,83750,83811,83868,83989,84107,84363,84370,84381,84592,84647,84727,84756,84759,84800,84879,85150,85165,85253,85310,85360,85368,85408,85576,85643,85860,86121,86127,86354,86992,87032,87116,87145,87164,87319,87447,87476,87555,87699,87742,87756,87780,87793,87878,88040,88111,88229,88238,88315,88326,88341,88397,88466,88472,88478,88745,88806,88953,89241,89304,89409,89459,89669,89686,89702,89734,89818,89867,89901,89919,89928,89989,90039,90202,90253,90270,90396,90521,90594,90600,90707,90842,90843,91007,91021,91077,91188,91261,91350,91368,91457,91738,91807,91858,91932,92219,92428,92435,92450,92451,92479,92526,92720,92739,92887,93077,93092,93202,93203,93235,93236,93283,93320,93372,93526,93596,93650,93660,93692,93780,93825,93862,93973,94074,94092,94179,94248,94436,94446,94570,94704,94763,94831,94848,94965,95164,95210,95270,95290,95379,95420,95481,95571,95706,95794,95835,96010,96169,96204,96310,96369,96459,96463,96537,96635,96767,96813,96820,96896,97080,97131,97312,97339,97374,97431,97582,97602,97722,97826,97880,98004,98010,98037,98114,98205,98223,98432,98610,98618,98775,98784,98827,99004,99039,99061,99139,99153,99164,99368,99378,99600,99687,99759,99916,99928,99962,99978",
          "14,29,89,111,116,150,215,233,310,479,905,961,992,1011,1079,1171,1202,1281,1441,1491,1502,1623,1693,1773,1854,1888,1927,1950,1959,1962,1998,2010,2027,2031,2134,2153,2177,2191,2197,2199,2205,2214,2286,2289,2292,2334,2366,2387,2395,2399,2433,2446,2455,2483,2525,2568,2577,2578,2622,2627,2648,2683,2700,2704,2785,2794,2802,2813,2842,2877,2907,2942,2947,2959,2961,2965,2971,2995,3023,3039,3068,3098,3121,3143,3154,3211,3220,3244,3252,3260,3327,3401,3402,3431,3443,3471,3478,3621,3625,3632,3660,3688,3737,3739,3756,3765,3770,3829,3901,3903,3928,3937,3948,3950,3953,4000,4031,4041,4047,4067,4135,4144,4156,4175,4206,4251,4289,4329,4340,4351,4357,4399,4416,4424,4445,4464,4532,4563,4567,4604,4643,4654,4679,4706,4773,4837,4840,4908,4929,4936,4941,4955,4971,4984,5030,5208,5248,5295,5336,5569,5621,5630,5644,5671,5699,5705,5737,5788,5856,5893,5944,5982,6016,6044,6079,6273,6460,6664,6677,6743,6744,6776,6778,6791,6857,6865,7048,7054,7214,7224,7317,7384,7489,7497,7504,7588,7674,7757,8012,8022,8058,8098,8115,8310,8559,8678,8704,8734,8736,8810,8836,8837,9018,9026,9081,9135,9163,9196,9233,9285,9300,9307,9339,9396,9456,9459,9485,9634,9766,9830,9849,9971,10104,10105,10166,10274,10391,10396,10431,10448,10637,10640,10646,10743,10772,10816,11182,11273,11315,11317,11332,11350,11376,11388,11404,11425,11483,11504,11695,11764,11981,12010,12138,12431,12629,12767,12942,12946,13006,13040,13135,13308,13757,13801,13848,14123,14188,14303,14363,14443,14474,14554,14558,14608,14974,15232,15440,15552,15693,15789,16058,16270,16454,16506,16551,16688,16783,16998,17127,17505,17849,17875,18075,18145,18205,18235,18268,18344,18637,18742,18769,18830,18839,19290,19646,19713,19723,19756,19770,19953,20289,20455,20987,21230,21538,21798,22017,22298,22588,22809,22858,23045,23161,23195,23223,23420,23680,23751,24009,24255,24850,24892,25453,25610,25656,25672,25742,26208,26308,26567,26569,26574,26605,26622,26670,26820,26849,26983,27078,27088,27200,27241,27388,27471,27509,27527,27550,27637,27659,27669,27687,27952,27988,27990,28002,28023,28044,28083,28125,28126,28137,28153,28168,28185,28242,28307,28457,28517,28541,28571,28585,28588,28605,28622,28631,28642,28646,28664,28772,28973,29180,29314,29318,29346,29357,29379,29397,29428,29539,29554,29614,29653,29811,29835,29915,30046,30075,30133,30220,30249,30278,30336,30365,30423,30481,30597,30626,30655,30684,30713,30742,30771,30800,30858,30887,30916,30974,31003,31061,31090,31119,31148,31177,31206,31235,31264,31351,31380,31409,31467,31474,31496,31525,31554,31583,31612,31641,31699,31728,31757,31786,31815,31822,31844,31902,31960,31989,32076,32105,32134,32163,32192,32221,32250,32279,32308,32337,32395,32453,32482,32511,32540,32656,32685,32743,32772,32801,32830,32888,32946,32975,33033,33091,33120,33178,33207,33236,33265,33294,33323,33352,33381,33410,33439,33468,33497,33526,33584,33642,33671,33700,33729,33758,33787,33816,33845,33874,33903,33932,33961,33990,34019,34048,34077,34106,34135,34164,34222,34251,34280,34309,34338,34367,34396,34425,34454,34541,34570,34599,34686,34715,34744,34802,34831,34889,34918,34947,34976,35049,35180,35206,35213,35234,35282,35291,35343,35394,35417,35453,35488,35536,35558,35595,35654,35686,35727,35786,35822,35845,35861,35975,36034,36097,36126,36135,36179,36230,36287,36346,36384,36405,36690,36802,36848,36885,36966,36983,36996,37038,37097,37156,37279,37355,37409,37468,37527,37557,37599,37717,37781,37806,37899,37937,37970,38012,38029,38088,38160,38219,38401,38460,38531,38556,38627,38649,38672,38721,38747,38839,38903,38989,39021,39059,39092,39151,39210,39239,39282,39341,39400,39432,39464,39523,39653,39771,39791,39806,39843,39863,39961,39980,39999,40108,40137,40166,40195,40282,40311,40340,40369,40456,40485,40514,40543,40572,40601,40630,40659,40688,40717,40746,40833,40891,40920,40949,40978,41007,41036,41065,41094,41123,41152,41181,41210,41217,41268,41384,41406,41413,41471,41529,42059,42088,42177,42267,42342,42383,42393,42451,42504,42579,42774,42805,42891,42964,42978,43053,43083,43086,43089,43128,43166,43215,43245,43255,43280,43290,43446,43527,43662,43679,43689,43732,43815,43839,43873,43926,44001,44040,44076,44118,44163,44224,44230,44238,44266,44299,44313,44355,44363,44370,44400,44427,44475,44507,44586,44659,44667,44712,44787,44901,44928,44949,44993,44999,45024,45145,45206,45225,45226,45394,45524,45581,45735,45768,45826,45898,46032,46061,46112,46166,46188,46270,46290,46324,46350,46418,46484,46525,46530,46555,46565,46598,46647,46649,46683,46691,46694,46716,46729,46761,46831,46852,46915,46959,46964,47074,47176,47191,47212,47406,47413,47515,47535,47575,47609,47623,47630,47718,47719,47785,47929,47993,48034,48036,48070,48085,48177,48182,48216,48257,48280,48298,48337,48449,48481,48482,48492,48526,48618,48634,48688,48724,48832,48895,48931,48941,48956,49105,49212,49275,49308,49412,49431,49473,49532,49563,49568,49572,49726,49748,49810,49819,49843,49878,49947,50210,50215,50358,50499,50548,50647,50664,50816,50835,50977,51012,51030,51101,51231,51352,51410,51637,51847,51876,51976,51980,52111,52159,52277,52431,52513,52656,52773,52963,53083,53163,53166,53211,53356,53479,53676,53723,53905,53929,54015,54132,54241,54298,54337,54346,54382,54469,54540,54681,54709,54724,54751,54882,54885,55150,55195,55200,55676,55853,55964,55971,56016,56024,56037,56146,56174,56260,56261,56278,56415,56468,56488,56496,56552,56555,56570,56593,56594,56601,56607,56609,56750,56759,56770,56915,57119,57179,57214,57224,57272,57312,57332,57359,57380,57601,57677,57807,58016,58071,58292,58421,58466,58534,58606,58610,58644,58807,58954,59050,59081,59103,59151,59173,59218,59230,59248,59288,59441,59486,59550,59570,59593,59623,59678,59707,59848,60020,60037,60228,60252,60297,60322,60353,60439,60553,60646,60676,60795,60833,60839,60864,61023,61279,61352,61708,61717,61736,61872,61906,61989,61996,62036,62163,62182,62193,62203,62215,62232,62263,62441,62491,62512,62540,62669,62672,62771,62867,62932,63078,63171,63259,63316,63380,63397,63438,63568,63631,63633,63730,63941,63971,63977,64028,64138,64158,64230,64267,64279,64402,64420,64428,64437,64447,64474,64519,64549,64550,64588,64670,64718,64950,65015,65089,65242,65286,65367,65425,65665,65870,66078,66186,66304,66369,66387,66409,66411,66419,66468,66769,66926,66940,67160,67413,67455,67515,67586,67601,67623,67745,67900,67934,68187,68199,68230,68239,68249,68282,68418,68476,68504,68686,68688,68807,68915,68962,69153,69188,69314,69366,69439,69584,69676,69727,69735,69736,69784,69817,69818,69836,69930,70005,70063,70119,70139,70254,70306,70319,70334,70345,70369,70488,70576,70663,70760,70848,70863,71014,71018,71033,71103,71304,71335,71418,71429,71435,71478,71521,71538,71570,71675,71728,71914,71966,71989,71996,72038,72107,72140,72188,72285,72318,72327,72399,72409,72572,72624,72636,72760,72818,72831,72858,72921,72942,72979,73058,73126,73378,73407,73426,73487,73586,73694,73745,73805,73823,73862,74021,74059,74096,74106,74234,74410,74425,74472,74660,74716,74723,74745,74833,75076,75112,75169,75187,75247,75260,75297,75438,75478,75681,75743,75847,76125,76158,76250,76513,76765,76812,76947,77078,77325,77429,77448,77462,77486,77578,77673,77817,78012,78082,78140,78488,78490,78554,78589,78684,78866,78971,79011,79131,79570,79640,79652,79690,79705,79725,79980,80075,80108,80152,80196,80300,80421,80433,80457,80499,80551,80593,80654,80681,80747,80794,80823,80888,80900,80918,81112,81210,81335,81344,81424,81558,81580,81631,81905,81908,81914,82106,82120,82151,82189,82404,82548,82696,82714,82757,82758,82795,82852,82944,82995,83080,83267,83313,83768,83786,83938,84105,84156,84206,84298,84318,84358,84417,84420,84449,84472,84599,84721,84870,84929,85075,85077,85222,85239,85254,85259,85305,85454,85591,85673,85928,85952,86000,86003,86075,86133,86137,86192,86247,86255,86295,86311,86382,86411,86508,86603,86630,86749,86831,86836,87049,87179,87188,87277,87284,87349,87353,87408,87427,87443,87519,87544,87615,87731,87763,87790,87796,87915,87947,87993,88148,88212,88287,88367,88369,88381,88695,88759,88760,89028,89313,89351,89599,89629,89685,89704,89750,89897,89955,90005,90164,90292,90397,90415,90452,90491,90494,90497,90528,90832,91065,91091,91171,91336,91391,91506,91765,91811,91963,92090,92154,92313,92438,92460,92503,92540,92564,92571,92608,92694,92708,92747,92852,92881,92889,92897,93295,93377,93549,93571,93674,93795,93935,93948,94019,94156,94247,94270,94286,94287,94319,94378,94387,94435,94516,94527,94589,94605,94632,94653,94688,94738,94881,94920,94954,94994,95008,95042,95101,95265,95315,95477,95479,95507,95511,95551,95565,95579,95595,95623,95636,95751,95753,95819,95920,95932,95953,96032,96185,96235,96288,96296,96334,96341,96379,96431,96464,96500,96770,96980,97032,97130,97384,97772,97774,97913,97944,98052,98072,98121,98216,98251,98274,98298,98357,98367,98380,98477,98658,98723,98861,98918,98938,99010,99193,99395,99396,99465,99476,99616,99725,99963",
          "200,263,324,344,383,416,438,500,526,757,813,829,932,1040,1052,1082,1093,1163,1192,1229,1280,1296,1297,1314,1560,1606,1613,1804,1827,1889,1974,1988,2288,2299,2386,2493,2546,2606,2644,2661,2766,2831,2864,2940,2951,3005,3006,3322,3394,3408,3463,3623,3624,3740,3780,3787,3842,3863,3940,4048,4352,4728,4812,4852,4890,5025,5027,5125,5159,5164,5198,5287,5299,5350,5352,5373,5425,5629,5634,5722,5815,5841,5913,5965,5975,6126,6164,6188,6301,6439,6483,6497,6536,6763,6798,6935,7092,7197,7297,7333,7482,7604,7616,7618,7867,7887,7946,7978,8009,8082,8188,8360,8482,8518,8711,8725,8964,8971,9136,9242,9257,9286,9484,9492,9535,9576,9625,9951,9955,10036,10100,10200,10228,10244,10394,10441,10487,10493,10617,10718,10980,11042,11072,11125,11183,11424,11724,11949,12339,12424,12524,12597,12621,12645,12952,13076,13149,13320,13347,13378,13425,13501,13524,13659,13669,13929,13954,14080,14340,14375,14819,14907,14958,14991,15082,15409,15602,15706,15793,15861,15915,15930,16013,16134,16412,16427,16654,16916,17006,17102,17251,17337,17472,17530,17581,17696,17721,17969,18210,18802,18859,18878,18966,19098,19148,19245,19296,19437,19511,19755,19810,19850,19869,20202,20219,20220,20316,20657,20670,20892,20940,20959,21189,21218,22092,22122,22217,22532,22995,23104,23151,23179,23254,23270,23520,23699,23827,24078,24269,24489,24671,24855,24880,24939,25121,25163,25240,25279,25284,25322,25380,25436,25585,25673,25716,25811,25817,26173,26226,26293,26337,26444,26497,26633,26640,26833,26883,26905,26945,27133,27193,27268,27743,27785,27961,28161,28197,28225,28298,28408,28482,28521,28601,28615,28804,28950,29325,29590,29679,29823,29899,29954,30101,30114,30169,30188,30198,30507,30536,30652,30884,31029,32903,32943,33088,33391,33407,33581,33657,33802,35226,35312,35391,35592,35783,35842,35939,36008,36044,36315,36316,36454,36551,36917,36921,36936,37007,37072,37131,37137,37181,37242,37389,37465,37544,37619,37695,37877,37929,37991,38057,38240,38243,38252,38311,38416,38440,38568,38612,38766,38811,38866,39192,39318,39375,39500,39562,39616,39625,39629,39730,39921,39937,40270,40304,40480,40596,40647,41227,41261,42025,42111,42282,42706,42734,42863,43085,43123,43479,43786,44026,44344,44706,45177,45317,45687,45746,45798,45974,46714,46748,46871,47664,47843,47968,48012,48347,48621,48712,48878,49008,49272,49470,49820,50040,50056,50186,50278,50537,50679,50820,50885,50971,51177,51204,51230,51363,51520,51604,51616,51668,51714,51880,51881,52062,52071,52126,52142,52307,52428,52536,52550,52560,52571,52609,52631,52639,52682,52691,52709,52771,52785,52861,52879,53080,53192,53242,53475,53476,53487,53538,53540,53601,53609,53674,53675,53746,53756,53832,53867,53876,53956,54034,54104,54191,54279,54283,54351,54533,54572,54768,54775,54810,54846,54956,55031,55092,55217,55451,55504,55539,55541,55717,55906,56026,56028,56055,56082,56158,56173,56237,56325,56387,56417,56432,56527,56637,56689,56721,56831,56849,56982,57065,57195,57218,57242,57303,57526,57537,57625,57692,57808,57864,57876,57919,57922,57957,57984,58017,58125,58172,58324,58369,58436,58459,58465,58491,58516,58774,58820,59018,59051,59120,59246,59263,59328,59387,59404,59489,59494,59504,59533,59547,59554,59565,59617,59650,59684,59791,59830,59919,59938,59946,60075,60137,60139,60148,60277,60371,60405,60410,60630,60721,60790,60803,60844,61180,61184,61187,61206,61749,61948,62007,62018,62168,62183,62214,62228,62276,62303,62327,62352,62383,62393,62445,62529,62836,62920,62965,62971,62976,63283,63337,63462,63482,63709,63740,63827,63885,63922,63991,64046,64052,64219,64274,64433,64547,64812,64897,65031,65063,65133,65143,65350,65355,65378,65391,65604,65697,65745,65857,66178,66181,66576,66629,66687,66798,66864,66969,67009,67107,67167,67291,67294,67295,67483,67546,67584,67594,67683,67685,67782,67835,67913,67943,67974,67986,68025,68326,68350,68395,68472,68549,68678,68799,68865,69034,69037,69163,69192,69296,69398,69443,69985,70006,70024,70186,70239,70384,70501,70715,70793,70837,70858,70968,71046,71068,71278,71327,71593,71985,72025,72068,72098,72163,72223,72224,72490,72632,72667,72707,72723,72758,72795,72807,72808,72868,72964,73018,73026,73040,73059,73072,73147,73246,73260,73282,73309,73340,73601,73627,73686,73970,74060,74122,74177,74294,74301,74347,74366,74434,74636,74712,74731,74845,74999,75335,75640,75643,75692,75763,75788,75816,75930,76119,76216,76254,76294,76354,76365,76539,76671,76718,76735,76738,76740,76752,76887,76935,76972,77052,77093,77379,77536,77569,77668,77783,77935,77937,77984,78151,78153,78163,78194,78213,78243,78344,78400,78448,78476,78574,78602,78657,78739,78743,78784,78788,78846,78856,78962,78990,79079,79226,79265,79367,79473,79573,79734,79810,79966,79983,80092,80177,80254,80291,80311,80317,80364,80388,80520,80522,80582,80586,80605,80651,80734,80744,80939,80990,81034,81140,81157,81187,81314,81327,81532,81607,81628,81776,81842,81893,81898,81918,82047,82135,82269,82277,82467,82550,82569,82575,82591,82683,82765,82782,82928,82949,83079,83118,83192,83496,83635,84175,84266,84385,84389,84588,84745,84975,84977,84979,85042,85054,85055,85118,85145,85386,85440,85556,85605,85655,85696,85732,85908,85925,85956,86034,86077,86293,86317,86381,86598,86739,86851,86979,87022,87163,87292,87333,87397,87457,87468,87515,87562,87567,87606,87616,87639,87643,87710,87820,87824,88007,88127,88292,88368,88425,88494,88581,88681,88754,88771,88787,88834,88860,88945,89063,89087,89093,89131,89159,89232,89236,89284,89287,89471,89507,89521,89541,89572,89628,89657,89690,89763,89871,89898,89927,90183,90273,90416,90534,90644,90649,90700,90751,90896,90951,90991,91149,91280,91294,91332,91369,91445,91487,91751,92047,92241,92299,92342,92472,92495,92727,92817,92846,92961,93035,93060,93169,93311,93463,93930,94203,94396,94465,94468,94502,94554,94580,94671,94713,94737,94818,94866,94927,95025,95026,95035,95195,95304,95366,95440,95471,95473,95533,95655,95739,95766,95769,95878,95890,95921,95935,95958,96072,96076,96095,96125,96149,96156,96200,96224,96237,96300,96426,96430,96458,96640,96688,96692,96821,96967,97077,97162,97196,97356,97545,97635,97656,97682,97695,97793,97828,97850,97905,98021,98064,98078,98101,98118,98133,98143,98164,98292,98326,98502,98506,98893,98989,99141,99163,99203,99349,99530,99633,99768,99954,99977",
          "31,354,376,377,381,387,422,602,650,683,735,747,807,849,889,952,979,994,1004,1024,1048,1081,1112,1142,1194,1227,1232,1243,1263,1264,1371,1419,1504,1677,1772,1878,1912,3073,3692,5022,5037,5047,5069,5081,5099,5107,5111,5176,5315,5332,5394,5432,5459,5480,5486,5577,5582,5592,5595,5615,5675,5784,5798,5852,5868,5888,5910,5994,5998,6014,6069,6143,6210,6219,6231,6244,6270,6289,6335,6354,6372,6407,6543,6600,6604,6628,6649,6699,6779,7019,7026,7036,7051,7410,7417,7488,7625,7667,7729,7915,8016,8038,8063,8441,8445,8541,8618,8634,8641,8661,8789,8839,8899,8903,9008,9109,9113,9130,9139,9215,9225,9246,9289,9946,9947,10056,10227,10499,10625,10838,10935,11114,11768,11790,12158,12525,14214,14878,15716,16449,16984,17082,17709,18176,18848,19329,20111,20256,20576,21455,23035,23313,23483,24353,25077,25329,25354,25487,26465,26709,27221,27538,28900,29163,35896,36310,36432,36639,37638,38296,38703,39825,42556,43098,44473,45032,45173,46281,46476,50018,50062,50089,50136,50239,50359,50430,50447,50465,50488,50524,50603,50686,50880,50909,50945,51019,51156,51218,51249,51336,51371,51445,51581,51599,51660,51726,51810,51865,52037,52054,52100,52343,52450,52495,52629,52664,52784,53153,53200,53202,53372,53497,53501,53554,53621,53659,53730,53782,53871,53874,54062,54073,54123,54167,54358,54360,54366,54425,54430,54447,54488,54513,54546,54586,54752,54883,55063,55154,55159,55209,55250,55290,55350,55364,55383,55431,55577,55582,55678,55696,55784,55814,55969,55989,56115,56119,56386,56449,56508,56652,56912,56971,57057,57071,57099,57117,57254,57461,57561,57564,57646,57737,57830,57900,57937,57996,58058,58084,58264,58312,58322,58333,58386,58410,58441,58443,58452,58539,58585,58605,58873,58884,58922,59045,59129,59252,59330,59333,59416,59421,59506,59571,59829,59860,59911,60147,60203,60316,60390,60450,60515,60570,60642,60647,60657,60766,60783,60818,60826,60837,61020,61097,61162,61255,61369,61431,61464,61482,61542,61585,61679,62009,62073,62091,62159,62188,62218,62305,62398,62473,62566,62786,62915,62921,62927,63101,63104,63284,63285,63298,63321,63733,63774,63811,63817,63899,64030,64325,64576,64618,64649,64715,64735,64754,64767,64811,64824,64920,64990,65128,65161,65257,65401,65554,65746,65750,65790,65794,65797,65836,66033,66061,66093,66132,66156,66169,66222,66327,66421,66464,66553,66575,66615,66697,66733,66740,66771,66819,66833,66886,66891,66913,66996,67025,67080,67117,67166,67221,67436,67589,67591,67647,67726,67732,67752,67761,67887,67892,68034,68084,68217,68228,68528,68630,68636,68725,68766,68942,69003,69100,69136,69162,69211,69239,69316,69488,69546,69763,69824,69857,69859,69893,70043,70058,70116,70156,70183,70202,70394,70704,70752,70795,70871,70877,70942,70980,71064,71285,71334,71413,71565,71575,71595,72001,72075,72120,72131,72246,72319,72338,72356,72429,72476,72822,72851,72895,73213,73230,73351,73401,73436,73437,73516,73548,73676,73682,73702,73733,73795,73816,73827,74090,74105,74247,74352,74364,74406,74553,74566,74580,74650,74762,74796,74865,74907,74962,75014,75252,75266,75280,75286,75306,75375,75465,75623,75677,75737,75981,75994,76232,76303,76312,76412,76565,76589,76700,76855,76882,77117,77257,77424,77516,77565,77589,77619,77686,77772,77896,77933,77971,78093,78193,78199,78329,78368,78443,78499,78548,78593,78671,78825,78830,78915,78932,78997,79005,79035,79138,79142,79162,79231,79993,80035,80137,80209,80337,80377,80465,80470,80478,80653,80677,80824,80845,81037,81039,81166,81171,81348,81377,81438,81560,81646,81785,81793,81806,81860,81902,81954,81961,81975,81982,82136,82205,82226,82268,82522,82547,82568,82590,82726,82734,82779,82796,82864,82969,83239,83375,83390,83504,83560,83607,83695,83756,83796,83851,84204,84247,84394,84452,84474,84609,84728,85149,85365,85509,85547,85612,85708,85715,85779,85796,85880,85895,85924,86053,86114,86232,86319,86334,86342,86379,86417,86587,86634,86761,86862,86871,86915,86995,87310,87549,87734,87905,87931,87933,87946,87948,88027,88213,88354,88439,88559,88584,88625,88680,88880,89019,89068,89139,89151,89221,89247,89258,89281,89737,89889,89940,89964,89987,90032,90033,90111,90112,90154,90285,90454,90459,90614,90667,90893,90903,91104,91121,91201,91406,91533,91715,91722,92131,92169,92182,92322,92349,92362,92364,92453,92497,92690,92794,92915,93048,93073,93335,93419,93498,93612,93774,93792,93827,93975,93990,94046,94049,94227,94278,94301,94345,94438,94458,94494,94503,94578,94608,94640,94670,94709,94721,94729,95030,95326,95333,95593,95704,95717,95732,95910,96003,96135,96179,96221,96453,96489,96503,96573,96831,96842,96897,96963,96970,97022,97075,97128,97262,97447,97482,97534,97636,97647,97724,97734,97760,97792,97807,97919,97976,98035,98103,98131,98293,98305,98311,98416,98542,98562,98650,98668,98715,98718,98774,98819,98895,99056,99083,99098,99273,99376,99380,99394,99471,99581,99636,99663,99735,99778",
          "91,1365,1716,1717,1779,1873,1979,2048,2211,2240,2319,2368,2458,3029,3110,3243,3328,3442,3537,3723,3955,4050,4069,4419,4440,4489,4810,4873,5044,6145,6596,6615,6653,6963,6990,7128,7850,8053,8059,8135,8154,8314,8333,8553,8648,8762,8778,8917,9060,9117,9161,9180,9237,9308,9316,9406,9493,9503,9510,9522,9560,9598,9617,9693,9750,9788,9823,9845,9883,9978,9997,10035,10111,10130,10149,10168,10187,10191,10206,10225,10292,10339,10605,10738,10833,10890,10893,10909,10928,10966,11023,11137,11156,11172,11175,11213,11251,11265,11296,11308,11327,11403,11422,11517,11574,11611,11612,11726,11897,12068,12087,12144,12163,12239,12258,12391,12543,12619,12700,12721,12733,12746,12825,12828,12904,12926,12955,13056,13132,13151,13185,13189,13227,13246,13262,13272,13284,13303,13360,13379,13455,13493,13502,13588,13735,13759,13816,13835,13873,13886,13913,13930,13949,13974,14044,14101,14158,14177,14215,14253,14272,14291,14310,14386,14456,14538,14559,14671,14725,14728,14734,14842,14899,14975,15032,15033,15108,15127,15146,15211,15222,15241,15279,15317,15393,15450,15469,15507,15559,15621,15659,15678,15688,15697,15798,15811,15849,15944,15963,15982,16020,16077,16109,16210,16229,16305,16324,16343,16362,16495,16514,16586,16628,16647,16678,16713,16742,16875,16894,16932,16970,16989,17116,17122,17198,17217,17255,17261,17331,17350,17354,17464,17502,17559,17628,17673,17825,17863,17882,17939,17980,18016,18072,18091,18129,18186,18224,18243,18281,18319,18338,18395,18433,18452,18490,18509,18585,18635,18661,18737,18813,18825,18851,18927,18946,19033,19117,19155,19172,19174,19212,19231,19345,19383,19402,19417,19459,19478,19497,19500,19513,19516,19592,19613,19616,19630,19649,19668,19726,19782,19820,19839,19915,19974,19991,20022,20143,20276,20295,20302,20314,20355,20371,20409,20447,20466,20485,20504,20561,20567,20637,20656,20694,20770,20867,20884,20917,21107,21112,21150,21262,21302,21435,21473,21559,21568,21625,21720,21739,21834,21891,21910,21967,21986,22024,22043,22100,22176,22214,22323,22385,22404,22442,22461,22480,22518,22537,22556,22568,22583,22594,22670,22765,22784,22803,22841,22875,22917,22936,22947,22955,23050,23069,23145,23175,23183,23202,23259,23278,23316,23354,23487,23525,23620,23639,23658,23715,23734,23791,23867,23886,23924,23962,23981,24019,24076,24095,24133,24178,24190,24263,24285,24342,24354,24361,24380,24399,24418,24437,24456,24513,24570,24598,24684,24703,24722,24798,24836,24893,24969,25026,25102,25159,25197,25273,25297,25349,25368,25444,25463,25501,25558,25577,25596,25615,25691,25729,25805,25824,25843,25919,25938,26014,26033,26106,26109,26147,26151,26223,26261,26280,26299,26414,26463,26470,26527,26641,26717,26845,27154,27157,27187,27325,27382,27406,27629,27844,27876,27935,28199,28351,28522,28731,28826,28921,29020,29155,29269,29288,29301,29377,29461,29768,29818,30061,30119,30148,30235,30351,30380,30438,30525,30583,30670,30699,30728,30757,30786,30844,30902,30931,30989,31105,31134,31192,31221,31279,31308,31395,31482,31511,31540,31569,31627,31685,31714,31772,31801,31830,31859,31888,31917,31946,31975,32033,32062,32120,32236,32294,32323,32352,32410,32468,32555,32613,32642,32671,32700,32758,32874,32932,33019,33106,33135,33164,33193,33222,33280,33338,33425,33483,33512,33541,33570,33686,33715,33773,33831,33889,33918,33947,34005,34034,34063,34092,34150,34179,34237,34266,34382,34411,34440,34527,34556,34585,34643,34672,34701,34730,34788,34817,34875,34962,35003,35087,35132,35210,35218,35255,35325,35329,35396,35403,35432,35451,35469,35474,35506,35522,35581,35667,35678,35772,35831,35877,35961,35994,36083,36143,36176,36325,36332,36391,36429,36519,36522,36581,36610,36675,36704,36715,36735,36763,36834,36893,37083,37142,37206,37372,37451,37502,37513,37585,37644,37703,37767,37826,37885,37956,38146,38205,38264,38328,38370,38446,38517,38635,38707,38754,38889,38948,39078,39106,39137,39196,39268,39327,39368,39386,39450,39639,39676,39698,39726,39757,39785,39797,39829,39856,39859,39947,40201,40259,40288,40346,40404,40433,40462,40520,40549,40578,40607,40665,40694,40723,40810,40839,40868,40897,40926,40955,41013,41042,41100,41129,41158,41187,41332,41361,41390,41419,41448,41477,41506,41535,42121,42214,42256,42346,42366,42421,42444,42447,42508,42547,42566,42583,42658,42704,42787,42820,42938,42969,43038,43057,43106,43132,43159,43262,43272,43294,43369,43531,43606,43632,43647,43656,43657,43714,43739,43901,43911,43930,43966,44005,44030,44034,44048,44080,44185,44242,44269,44388,44450,44479,44534,44554,44716,44736,44791,44813,44819,44828,44846,44848,44878,44911,44934,44953,45028,45101,45149,45210,45268,45309,45341,45380,45418,45585,45658,45715,45772,45830,45860,45959,46035,46053,46084,46128,46199,46384,46605,46680,46702,46725,46740,46756,46841,46856,46874,46912,46958,47009,47027,47068,47075,47107,47125,47145,47146,47152,47170,47210,47230,47296,47304,47347,47456,47460,47526,47536,47593,47603,47629,47707,47795,47985,47986,48003,48011,48052,48071,48117,48173,48219,48236,48250,48266,48316,48338,48358,48598,48611,48625,48827,48924,48948,48961,49015,49025,49179,49190,49270,49387,49400,49503,49511,49524,49531,49553,49575,49676,49696,49797,49969,50164,50223,50268,50475,50567,50722,50737,50778,50929,51228,51254,51283,51760,51793,52044,52080,52125,52258,52299,52336,52395,52470,52584,52642,52883,53180,53187,53213,53258,53260,53414,53472,53496,53690,53865,53999,54090,54095,54111,54214,54330,54344,54623,54814,54975,54992,55071,55106,55119,55367,55689,55732,55758,55828,55873,55923,56005,56318,56418,56632,56641,56685,56766,57148,57169,57322,57376,57470,57557,57653,57660,57727,57822,57960,58040,58050,58096,58199,58317,58372,58509,58515,58572,58666,58705,58748,58844,58896,58946,59059,59108,59168,59199,59225,59245,59267,59354,59541,59702,59803,60031,60101,60394,60441,60458,60637,60757,60982,60989,61052,61072,61179,61286,61287,61318,61609,61626,61778,62002,62066,62070,62106,62279,62346,62636,62670,62772,62828,62937,62949,62993,63202,63225,63228,63301,63305,63469,63484,63559,63749,63788,63830,63875,63888,63900,64034,64142,64155,64220,64257,64422,64464,64477,64558,64584,64624,64646,64723,64790,65071,65075,65154,65189,65227,65347,65495,65540,65662,65690,65758,65773,65783,65796,65826,66027,66043,66130,66191,66361,66401,66451,66520,66547,66633,66710,66912,66919,67002,67016,67076,67093,67119,67339,67347,67426,67471,67602,67848,67862,67883,67889,67956,68210,68233,68305,68346,68475,68526,68573,68625,68658,68661,68693,68905,68954,69250,69292,69431,69596,69598,69681,69712,69751,69762,69787,69788,69863,69928,69961,69962,69983,70011,70091,70099,70226,70242,70243,70311,70338,70532,70545,70746,70747,70825,70921,70983,71027,71097,71109,71162,71176,71222,71408,71445,71683,71910,71986,72074,72151,72244,72333,72434,72553,72563,72799,72849,72883,72923,72945,72948,73052,73173,73192,73334,73409,73515,73536,73585,73592,73671,73685,73872,73908,73977,74068,74078,74087,74197,74202,74306,74422,74439,74502,74656,74719,74741,74933,75127,75229,75381,75425,75476,75508,75546,75551,76023,76099,76135,76190,76215,76237,76256,76351,76512,76531,76598,76610,76701,76745,76772,76773,76838,76929,76945,77058,77084,77258,77290,77375,77482,77483,77490,77531,77756,77931,78001,78066,78120,78128,78177,78207,78224,78295,78405,78547,78561,78933,79161,79209,79279,79285,79372,79481,79561,79625,79848,80041,80181,80193,80222,80449,80479,80516,80769,80821,80863,80973,81273,81439,81441,81576,81590,81804,82072,82128,82156,82249,82300,82321,82361,82416,82438,82444,82518,82578,82614,82689,82818,82825,82826,82862,82887,83162,83194,83201,83408,83561,83624,83654,83687,83855,83875,83891,83967,84180,84250,84379,84447,84448,84514,84763,84771,84820,84874,85081,85090,85244,85275,85359,85380,85395,85537,85558,85642,85649,85695,85723,85750,85834,85878,86241,86262,86367,86451,86611,86612,86629,86740,86744,86964,86993,87034,87189,87203,87288,87302,87374,87713,87725,87837,87936,88010,88096,88120,88290,88308,88361,88469,88744,88807,88861,88894,88990,89015,89024,89086,89104,89141,89155,89185,89218,89261,89263,89292,89487,89499,89595,89603,89611,89640,89903,89918,89983,90103,90159,90191,90208,90216,90258,90316,90448,90552,90631,90632,90634,90670,90734,90773,90846,90868,90901,90927,91116,91212,91224,91315,91899,91908,91945,92058,92082,92123,92160,92170,92269,92326,92361,92365,92432,92514,92532,92569,92613,92636,92734,92749,92833,92871,93015,93036,93115,93139,93142,93149,93412,93444,93464,93497,93502,93690,93707,93873,93891,93905,93926,94143,94208,94282,94397,94467,94521,94525,94827,94878,94949,94975,94987,95139,95168,95259,95386,95685,95730,95864,95879,95895,96031,96045,96325,96412,96529,96602,96844,97111,97147,97169,97293,97337,97480,97552,97628,97812,97982,98206,98440,98550,98591,98746,98761,98844,99160,99328,99454,99532,99554,99569,99574,99736,99843",
                "32,127,146,1655,1659,1685,1712,1747,1780,1784,1932,2312,2318,2352,2391,2598,2676,2705,2733,3069,3172,3205,3261,3509,3646,3965,3984,4192,4484,4803,6096,6587,6625,7168,7544,7926,8126,8145,8183,8563,8639,8677,8696,8804,8872,9036,9095,9152,9247,9266,9378,9418,9421,9513,9532,9570,9627,9646,9665,9684,9703,9722,9741,9798,9803,9921,9950,10064,10106,10216,10232,10235,10331,10471,10473,10577,10627,10634,10653,10779,10843,10859,10900,11026,11052,11109,11185,11261,11360,11432,11470,11489,11535,11622,11641,11736,11755,11850,11851,11907,11976,12059,12078,12135,12287,12367,12486,12495,12496,12546,12572,12667,12705,12754,12781,12844,12895,12914,12964,13142,13237,13342,13465,13467,13484,13522,13528,13540,13598,13636,13655,13674,13701,13712,13731,13750,13769,13807,13828,13914,14031,14056,14073,14111,14130,14135,14168,14187,14282,14301,14320,14415,14567,14586,14641,14738,14757,14814,14947,15004,15006,15042,15137,15194,15202,15251,15270,15289,15343,15346,15365,15405,15473,15517,15533,15536,15631,15634,15648,15707,15708,15726,15821,15859,15878,15897,15954,16068,16106,16163,16201,16220,16258,16277,16296,16334,16353,16391,16448,16467,16524,16562,16657,16771,16790,16807,16809,16862,16885,16920,16961,16972,17005,17026,17056,17094,17100,17134,17151,17170,17235,17303,17341,17360,17379,17417,17421,17484,17486,17531,17626,17639,17652,17740,17759,17816,17873,17892,17930,17968,18006,18063,18082,18101,18158,18196,18367,18386,18405,18439,18443,18449,18481,18547,18557,18614,18633,18655,18720,18727,18823,18861,18880,18918,18937,18967,19012,19051,19089,19127,19203,19239,19247,19298,19317,19374,19393,19412,19469,19507,19545,19564,19621,19697,19701,19716,19773,19811,19830,19868,19906,19925,19963,19973,20001,20020,20077,20172,20210,20250,20267,20286,20305,20343,20362,20381,20400,20419,20457,20476,20495,20609,20666,20780,20837,20856,20888,20970,21027,21084,21179,21217,21236,21255,21331,21388,21407,21422,21426,21445,21483,21521,21578,21597,21635,21653,21654,21730,21749,21787,21801,21806,21844,21856,21863,21901,21939,21958,21977,21996,22015,22072,22091,22129,22195,22205,22224,22243,22262,22357,22376,22414,22433,22471,22509,22528,22547,22604,22680,22724,22737,22775,22813,22851,22965,22970,23022,23063,23117,23155,23231,23327,23440,23478,23516,23535,23548,23573,23687,23725,23739,23811,23839,23849,23868,23877,23915,23972,24010,24029,24086,24181,24219,24257,24291,24352,24409,24446,24447,24485,24504,24523,24542,24561,24580,24618,24675,24713,24740,24770,24865,24891,24998,25055,25131,25150,25226,25245,25264,25283,25301,25397,25401,25435,25457,25511,25568,25644,25682,25738,25815,25856,25897,25909,25948,25967,25986,26024,26081,26100,26138,26214,26252,26262,26290,26300,26309,26328,26371,26385,26390,26423,26461,26537,26594,26643,26707,26860,27045,27069,27107,27143,27254,27278,27320,27354,27386,27544,27597,27718,27734,27753,27843,27981,28133,28226,28228,28233,28304,28442,28494,28551,28738,28760,28867,28893,29026,29043,29197,29332,29445,29481,29506,29735,29767,29973,30023,30110,30197,30226,30284,30313,30342,30371,30429,30458,30516,30603,30632,30661,30690,30719,30748,30777,30806,30893,30922,30951,31038,31096,31212,31270,31299,31357,31415,31502,31531,31560,31589,31676,31705,31792,31821,31850,31937,31995,32082,32111,32140,32169,32227,32256,32285,32372,32430,32459,32488,32546,32604,32720,32749,32778,32894,32923,32952,32981,33010,33097,33126,33155,33184,33213,33242,33271,33300,33358,33387,33416,33445,33474,33503,33532,33590,33706,33735,33764,33822,33880,33938,33967,33996,34054,34083,34141,34170,34257,34315,34344,34402,34489,34547,34605,34634,34663,34750,34779,34895,34924,34953,34982,35030,35112,35186,35247,35248,35275,35302,35349,35358,35423,35449,35457,35460,35494,35687,35733,35742,35851,35922,36040,36103,36163,36186,36293,36352,36411,36483,36556,36724,36783,36869,36887,36913,36995,37058,37103,37162,37203,37226,37238,37285,37344,37474,37533,37605,37664,37723,37787,37846,37905,37976,38035,38166,38366,38407,38475,38537,38596,38655,38662,38750,38786,38861,38909,39027,39055,39098,39130,39157,39288,39406,39529,39588,39659,39692,39849,39967,39996,40116,40145,40203,40232,40261,40290,40319,40348,40406,40435,40464,40493,40522,40580,40609,40667,40696,40725,40754,40870,40899,40928,40986,41015,41044,41073,41102,41131,41160,41189,41218,41276,41305,41363,41392,41508,41537,42007,42036,42065,42094,42123,42183,42273,42291,42299,42321,42348,42376,42407,42423,42489,42510,42607,42638,42660,42747,42867,42897,42956,43009,43059,43345,43458,43477,43518,43533,43555,43608,43625,43695,43720,43728,43792,43845,43876,43891,43932,43980,44007,44082,44195,44203,44204,44244,44319,44364,44382,44383,44406,44425,44455,44628,44643,44744,44755,44793,44837,44856,44880,44936,44955,44978,44989,44995,45004,45055,45078,45103,45151,45172,45212,45249,45343,45400,45458,45486,45530,45587,45645,45717,45751,45753,45832,45904,45961,46022,46160,46258,46316,46375,46396,46488,46545,46575,46627,46844,46848,46881,46896,46916,46926,46976,47045,47049,47133,47142,47153,47301,47316,47351,47398,47541,47557,47591,47788,47807,47811,47896,47949,47977,48033,48073,48108,48176,48197,48453,48460,48528,48563,48590,48635,48670,48718,48721,48756,48768,48783,48903,48921,48960,48969,49033,49061,49127,49161,49173,49207,49235,49238,49245,49262,49332,49348,49414,49462,49500,49506,49547,49588,49605,49640,49687,49763,49832,49872,49934,49941,49984,50053,50055,50093,50156,50427,50614,50806,50826,51071,51073,51313,51340,51344,51346,51396,51404,51704,51729,51757,51803,51968,52013,52039,52096,52209,52232,52338,52456,52527,52612,52690,52706,52716,52937,52961,53046,53084,53174,53327,53357,53436,53440,53551,53699,53830,54092,54245,54294,54304,54335,54361,54654,54664,54811,54921,55102,55113,55211,55226,55301,55359,55396,55413,55439,55501,55570,55578,55730,55742,55882,55883,55952,56018,56074,56079,56182,56184,56216,56255,56283,56287,56353,56397,56618,56655,56738,56784,56855,56933,57168,57196,57228,57265,57534,57541,57586,57733,57747,57785,57940,57975,58032,58045,58061,58076,58192,58244,58285,58625,58694,58854,58867,58886,58893,58904,59096,59097,59135,59154,59170,59362,59370,59403,59520,59599,59649,59689,59696,59780,59795,60013,60114,60155,60185,60257,60505,60597,60607,60613,60808,60866,60967,60997,61060,61076,61121,61160,61291,61322,61450,61470,61524,61529,61688,61850,61881,61889,61917,62043,62145,62198,62330,62374,62446,62824,63008,63155,63162,63163,63178,63268,63501,63519,63673,63718,64173,64377,64593,64600,64663,64876,64908,64969,65147,65231,65321,65379,65443,65454,65457,65511,65517,65558,65619,65917,66040,66044,66081,66099,66138,66140,66187,66195,66247,66303,66370,66497,66725,67085,67284,67399,67418,67488,67629,67749,67841,67965,67998,68040,68066,68120,68204,68218,68252,68258,68274,68303,68351,68392,68434,68467,68477,68715,69069,69074,69144,69471,69624,69978,69997,70082,70131,70209,70247,70318,70329,70405,70474,70642,70730,70855,70879,70894,71077,71300,71643,71680,71806,71844,71868,71943,72061,72300,72312,72457,72560,72566,72717,72939,73032,73115,73151,73308,73313,73391,73404,73417,73425,73484,73526,73567,73774,73891,74002,74040,74189,74292,74446,74534,74560,74628,74708,74735,74856,74898,74926,74989,75231,75237,75468,75593,75752,75836,75863,75924,75983,76173,76221,76299,76300,76510,76621,76645,76658,77067,77203,77230,77234,77296,77358,77361,77397,77542,77572,77782,77803,77852,77859,77909,77945,78056,78121,78146,78154,78245,78362,78425,78450,78769,78834,78936,78972,79002,79225,79322,79329,79359,79369,79489,79545,79566,79608,79672,79869,80131,80145,80179,80446,80485,80497,80510,80568,80712,80735,80869,80970,81026,81121,81215,81219,81223,81240,81298,81374,81505,81771,81890,81900,81943,81966,82025,82044,82048,82132,82150,82219,82349,82350,82448,82461,82484,82499,82513,82666,82698,82739,83054,83122,83367,83417,83481,83550,83625,83793,83815,83963,84071,84200,84238,84376,84423,84443,84713,84753,84784,84873,84985,85012,85045,85071,85192,85304,85452,85481,85512,85603,85662,85901,85939,85960,86031,86098,86158,86249,86409,86572,86694,86728,86745,86771,86868,86967,87153,87335,87343,87347,87399,87450,87503,87732,87830,87990,88276,88348,88356,88360,88386,88428,88516,88608,88615,88661,88750,88791,88961,88972,89044,89190,89336,89447,89477,89539,89612,89736,89754,89794,89816,89830,89847,89851,89859,89942,89943,90128,90222,90424,90664,90808,90852,90864,90912,90945,90963,91372,91421,91467,91552,91645,91716,91721,91728,91750,91774,91851,91854,91973,91981,91992,92035,92046,92070,92105,92172,92251,92369,92545,92600,92742,92891,93059,93072,93084,93109,93162,93190,93237,93337,93341,93410,93712,93767,93989,94027,94036,94047,94072,94132,94198,94211,94365,94381,94399,94403,94443,94480,94506,94581,94641,94648,94656,94691,94693,94932,95125,95136,95188,95398,95437,95496,95535,95696,95752,95787,95830,95884,95951,96013,96021,96038,96423,96676,96823,96974,97065,97158,97163,97185,97211,97245,97266,97401,97454,97497,97526,97648,97949,98057,98073,98080,98175,98190,98237,98257,98411,98625,98665,98729,98762,98798,98831,98870,98901,98930,99093,99222,99244,99345,99481,99603,99765,99858",
          "20,82,213,359,362,464,589,605,635,719,754,767,782,783,805,830,843,882,917,921,1027,1085,1088,1126,1143,1176,1191,1209,1214,1247,1282,1284,1434,1460,1538,1541,1547,1556,1584,1688,1814,1818,1874,2170,2218,2235,2707,2825,3119,3237,3296,3707,3874,4294,4356,4382,4400,4437,4580,4585,5007,5014,5114,5133,5151,5171,5263,5397,5414,5441,5455,5534,5564,5677,5739,5748,5762,5768,5780,5811,5827,5842,5958,5990,6220,6266,6360,6440,6541,6554,6581,6754,6811,6819,6927,6957,7005,7007,7024,7055,7094,7137,7138,7226,7315,7343,7354,7413,7441,7617,7693,7720,7725,7740,7818,7870,7872,8056,8064,8100,8158,8240,8258,8272,8307,8587,8730,8845,8950,9042,9252,9295,9303,9436,9623,10007,10054,10116,10137,10267,10364,10476,11031,11065,11094,11191,11396,11417,11448,11519,11572,11810,11984,12073,12082,12133,12337,12376,12589,12819,13167,13387,13423,13570,13918,14373,14434,14865,14963,14994,15090,15203,15344,15381,15719,15992,16092,16248,16541,16729,16900,17142,17208,17491,17607,17958,18041,18348,18513,18719,18723,18745,19201,19211,19355,19450,19591,19653,19792,20068,20113,20134,20255,20488,20806,20945,21001,21008,21103,21207,21281,21383,21796,21825,21920,22058,22068,22184,22201,22889,23165,23584,23896,24238,24394,24419,24535,24704,24803,25492,25746,26204,26375,26827,26917,26988,27109,27185,27253,27264,27403,27444,27741,27750,27835,27850,28189,28763,29031,29878,29936,30835,31154,31183,31386,31473,31618,32198,33677,34228,34373,35137,35695,36113,36129,36542,36563,36568,37243,37442,37571,37813,37869,37874,37945,38094,38251,38430,38496,38802,38808,38988,38990,39252,39367,39381,39418,39936,39982,42304,42333,42611,43075,43199,43441,43678,43770,43820,44898,45311,45312,45742,45885,45920,45995,45999,46775,47046,47595,47799,48163,48648,50107,50125,50143,50159,50214,50316,50473,50659,50715,50781,50812,50982,51103,51245,51252,51278,51308,51383,51419,51466,51542,51559,51575,51645,51766,51824,51833,51835,51905,51950,52267,52301,52305,52339,52454,52506,52517,52666,52674,52860,52865,52934,52957,52986,53064,53067,53101,53170,53318,53370,53411,53716,53767,53773,53829,53870,53951,54087,54230,54321,54473,54552,54593,54641,54668,54685,54733,54861,55036,55132,55160,55172,55180,55244,55294,55323,55333,55381,55429,55446,55460,55495,55566,55591,55736,55765,55785,55866,55943,56196,56215,56336,56361,56392,56409,56420,56443,56471,56571,56795,56802,56846,56875,56887,56944,56991,57054,57092,57190,57264,57546,57616,57618,57645,57809,57910,57950,57994,58077,58153,58175,58237,58352,58354,58416,58450,58483,58579,58649,58653,58709,58741,58776,58842,58902,59147,59262,59270,59271,59352,59415,59423,59448,59456,59604,59645,59808,60063,60091,60107,60125,60218,60472,60530,60563,60628,60643,60742,60807,60825,60843,60878,60927,61014,61025,61055,61098,61231,61555,61583,61616,61711,61774,61809,61841,61892,61915,61941,62000,62035,62121,62146,62160,62245,62559,62591,63017,63036,63066,63070,63118,63179,63310,63359,63589,63649,63840,63961,64058,64074,64133,64139,64281,64316,64324,64342,64405,64455,64478,64570,64604,64648,64795,64814,64819,64922,65046,65076,65171,65223,65230,65251,65266,65346,65608,65762,65779,65802,65821,66035,66134,66174,66224,66228,66320,66458,66583,66639,66759,66936,67024,67034,67065,67068,67074,67124,67333,67340,67592,67677,67699,67705,67813,67898,67931,68058,68100,68147,68264,68458,68459,68603,68622,68648,68839,68897,68924,68926,68965,69083,69095,69161,69177,69267,69413,69539,69576,69615,69660,69695,69738,69952,70076,70102,70122,70123,70588,70624,70699,70724,70766,70807,70886,70890,70897,70954,70986,71116,71132,71138,71158,71171,71189,71237,71262,71361,71368,71379,71482,71507,71566,71634,71748,71760,71920,71956,72004,72019,72046,72196,72239,72283,72448,72601,72651,72661,72694,72712,72871,72962,72970,73242,73352,73443,73452,73478,73489,73522,73535,73559,73647,73695,73730,73831,73945,73972,73981,74215,74236,74248,74298,74379,74579,74598,74600,74748,74793,74828,74859,74869,74884,74912,74915,75098,75118,75135,75166,75355,75419,75523,75595,75786,75825,75888,75941,76046,76132,76292,76316,76340,76346,76434,76462,76577,76648,76736,76750,76766,76828,76833,77001,77038,77047,77513,77573,77577,77639,77687,77811,77814,77912,77913,77951,78047,78305,78343,78534,78815,78894,78942,79030,79062,79069,79090,79127,79202,79240,79528,79619,79632,79684,79720,79746,79792,79841,80022,80034,80061,80106,80329,80357,80413,80444,80500,80552,80606,80841,80868,80917,80940,81284,81297,81378,81442,81501,81530,81546,81602,81731,81740,81852,81976,82112,82245,82291,82395,82432,82527,82712,82823,82836,82873,82891,82917,82918,82991,83063,83115,83156,83210,83251,83253,83318,83338,83671,84019,84039,84049,84108,84212,84244,84299,84300,84305,84415,84459,84504,84515,84562,84614,84630,84636,84654,84729,84769,84810,85031,85167,85203,85418,85436,85748,85879,85899,85996,86045,86064,86154,86212,86277,86336,86400,86410,86415,86438,86615,86674,86778,86783,87029,87127,87345,87381,87514,87579,87632,87642,87719,87856,87859,87864,87894,88101,88141,88176,88404,88613,88654,88733,88840,88928,88954,89009,89039,89228,89396,89402,89567,89637,89667,89693,89836,90141,90201,90332,90428,90443,90547,90683,90744,90769,90774,90798,90986,91182,91255,91276,91362,91402,91473,91518,91535,91538,91683,91803,91990,92050,92137,92217,92508,92570,92574,92642,92799,92814,93033,93039,93547,93576,93640,93651,93797,93818,93919,93940,93959,94020,94045,94112,94244,94373,94429,94497,94559,94616,94645,94647,94756,94778,94790,94854,94901,95031,95087,95140,95151,95175,95258,95263,95307,95421,95562,95680,95703,95723,95770,95772,95817,95861,95980,96020,96129,96240,96261,96347,96414,96479,96482,96569,96620,96636,96658,96661,96921,97122,97145,97392,97409,97423,97499,97666,97759,97813,97924,97960,98027,98137,98264,98421,98636,98645,98865,99111,99132,99176,99197,99210,99291,99312,99334,99402,99518,99583,99635,99807,99842,99845,99913,99914",
          "214,268,634,690,812,936,951,1106,1108,1173,1188,1360,1394,1413,1430,1438,1455,1465,1468,2105,2107,2129,2259,2326,2671,2696,2780,3065,3294,3311,3584,3852,4051,4239,4316,4392,4497,4592,4690,4711,4829,4869,4942,5017,5024,5278,5279,5359,5386,5434,5437,5498,5668,5746,5763,5824,5844,5937,6012,6043,6046,6123,6144,6172,6186,6212,6240,6341,6413,6425,6452,6492,6535,6663,6714,6725,6746,6772,6818,7056,7062,7159,7160,7171,7228,7233,7387,7608,7750,7776,7781,7893,7955,8008,8031,8107,8235,8377,8379,8463,8472,8503,8532,8710,8868,8905,8965,8983,9114,9124,9372,9381,9606,9714,9910,9918,9981,10009,10031,10067,10110,10251,10280,10283,10386,10444,10453,10470,10474,10704,10720,10797,10947,10957,10988,11139,11159,11176,11253,11402,11530,11670,11786,11841,12096,12120,12171,12211,12354,12366,12392,12498,12532,12583,12591,12618,12660,12810,12929,13085,13210,13332,13339,13359,13454,13488,13788,13837,13969,13983,13985,13994,14145,14159,14263,14311,14344,14398,14465,14581,14652,14740,14885,15074,15132,15187,15210,15395,15467,15526,15578,15614,15615,15777,15800,15925,16001,16123,16173,16307,16372,16399,16441,16477,16638,16803,16948,17090,17105,17135,17166,17324,17333,17659,17751,18035,18453,18460,18576,18695,18998,19062,19093,19145,19157,19213,19293,19440,19526,19532,19576,19962,20132,20191,20258,20506,20590,20601,20608,20785,20861,21037,21056,21076,21114,21151,21312,21362,21437,21507,21855,22006,22086,22180,22339,22367,22566,22794,23001,23026,23077,23088,23123,23207,23297,23450,23451,23655,23831,24105,24324,24420,24515,24590,24591,24677,24822,24838,25074,25122,25129,25218,25223,25238,25256,25663,25864,25988,26035,26176,26263,26285,26307,26320,26358,26418,26490,26812,26843,26918,26963,26998,27056,27163,27313,27813,27873,27905,27912,27913,28266,28316,28425,28433,28503,28705,28719,28768,28798,28815,28888,28940,28961,29023,29112,29137,29175,29291,29298,29307,29359,29449,29496,29569,29812,29900,29903,29906,30033,30168,30304,30400,30526,30545,30642,31774,32024,32296,32441,32575,33108,33329,33659,34093,34209,34296,34460,34721,34819,34866,34877,35135,35257,35305,35311,35465,35552,35749,35760,36328,36445,36512,36524,36972,37044,37074,37112,37122,37258,37352,37456,37494,37620,37627,37817,37928,38131,38134,38375,38380,38519,38674,38701,38736,38755,38801,38819,38878,38943,39127,39169,39216,39365,39366,39371,39422,39470,39480,39495,39564,39818,39822,39830,39858,39867,39948,40118,40140,40198,40292,40814,40836,40923,40957,41191,42232,42258,42439,42630,42762,42800,42822,42914,42946,43237,43529,43772,43946,44125,44336,44422,44461,44623,44701,44718,44874,45197,45427,45452,45612,45795,45812,46066,46340,46387,46500,46561,47864,47894,47956,48007,48086,48385,48552,48727,48745,48964,49024,49182,49523,49722,49786,50002,50011,50021,50304,50386,50725,50866,51145,51169,51226,51491,51722,51738,51841,52019,52166,52265,52406,52432,52486,52576,52673,52680,52800,52802,52894,53030,53249,53299,53400,53416,53428,53446,53592,53607,53629,53819,53921,53937,53965,54007,54049,54149,54209,54559,54620,54626,54702,54749,54864,54901,55124,55156,55166,55174,55229,55263,55281,55303,55358,55464,55510,55538,55624,55675,55808,56066,56300,56310,56425,56669,56696,56763,56775,56817,56877,56980,57019,57022,57074,57156,57161,57225,57392,57429,57512,57559,57611,57778,57815,57871,57936,57963,58114,58119,58122,58265,58303,58336,58374,58403,58437,58517,58531,58662,58729,58733,58803,58839,58938,58985,58990,59024,59026,59071,59077,59091,59190,59238,59428,59443,59534,59720,60006,60008,60022,60025,60051,60064,60179,60197,60325,60372,60492,60533,60686,60942,60974,61039,61080,61152,61376,61385,61487,61514,61545,61623,61675,61682,61716,61987,61990,62001,62023,62120,62170,62216,62360,62365,62379,62415,62548,62978,62997,63102,63110,63123,63143,63194,63250,63336,63350,63374,63485,63510,63537,63617,63623,63626,63910,64009,64068,64072,64090,64114,64181,64193,64411,64468,64479,64485,64564,64652,64909,64910,65030,65069,65271,65463,65494,65565,65651,65706,65835,65877,65892,65894,65985,66049,66077,66188,66217,66288,66338,66427,66490,66737,66764,66842,67252,67352,67432,67485,67486,67650,67662,67717,67738,67859,67935,67937,68053,68097,68281,68293,68339,68389,68490,68627,68640,68700,68772,68840,68939,68950,68993,69068,69123,69131,69344,69359,69559,69655,69680,69731,69992,70046,70053,70229,70262,70362,70370,70437,70454,70482,70528,70574,70750,70782,70785,71059,71104,71204,71227,71241,71308,71317,71354,71469,71524,71531,71605,71660,71752,72289,72378,72598,72612,72779,72841,72844,72856,72985,72992,72999,73236,73266,73482,73571,73575,73898,73909,73975,73998,74025,74155,74184,74209,74222,74314,74329,74424,74457,74710,74726,74811,74831,74838,74903,75087,75116,75292,75371,75521,75612,75648,75651,75675,75724,75738,75751,75765,75904,75990,76017,76138,76428,76448,76455,76582,76682,76706,76741,76761,76820,76830,76860,76867,76971,77006,77106,77301,77493,77540,77713,77804,77823,77873,77874,77942,78041,78166,78241,78266,78298,78315,78316,78477,78532,78552,78896,79012,79050,79074,79075,79280,79300,79402,79409,79429,79459,79532,79543,79637,79658,79715,79730,79747,79872,79878,80161,80299,80427,80453,80471,80585,80676,80861,80870,80995,81051,81174,81307,81324,81538,81572,81685,81751,81775,81836,82020,82069,82074,82096,82144,82261,82379,82392,82426,82449,82588,82636,82638,82843,82878,82939,82990,83155,83285,83324,83506,83626,83773,83799,83808,84077,84143,84196,84237,84271,84311,84375,84425,84470,84484,84559,84567,84597,84797,84894,84932,84958,85040,85063,85185,85206,85417,85448,85574,85795,85802,85971,86019,86204,86215,86343,86403,86405,86496,86506,86589,86642,86657,86727,86800,86874,86944,86946,86953,86969,86976,87180,87219,87332,87388,87402,87612,87613,87795,87833,87845,87883,87992,88006,88045,88173,88183,88189,88435,88454,88491,88577,88604,88616,88687,88801,88855,88980,88991,89014,89029,89097,89171,89180,89219,89271,89654,89715,89726,89793,89795,89798,89839,89848,89960,90055,90059,90203,90241,90280,90289,90341,90365,90411,90525,90550,90584,90682,90690,90847,90873,90898,91032,91063,91073,91122,91229,91291,91371,91470,91495,91500,91522,91699,91861,91936,92102,92247,92253,92605,92698,92853,92865,92974,93367,93470,93579,93621,93683,93783,93816,93888,93908,94006,94147,94228,94531,94550,94679,94697,94708,94830,95029,95082,95108,95262,95589,95596,95720,95795,95880,95939,96033,96091,96136,96193,96274,96650,96662,96679,96697,96899,96998,96999,97047,97190,97255,97361,97369,97419,97420,97421,97450,97532,97536,97578,97603,97810,97822,98120,98356,98369,98513,98554,98565,98605,98725,98849,98970,99008,99040,99290,99443,99622,99626,99875,99931,99959,99970,99994",
          "251,293,328,388,403,432,435,523,687,716,777,796,825,1006,1017,1026,1062,1139,1189,1200,1207,1230,1234,1256,1276,1298,1334,1374,1378,1388,1411,1464,1476,1485,1525,1555,1633,1653,1783,1891,2815,3008,3235,3241,3376,4754,4864,5061,5082,5092,5127,5147,5168,5169,5193,5233,5310,5324,5393,5408,5430,5560,5584,5604,5691,5720,5785,5846,5874,5917,5936,5951,5974,6001,6004,6088,6153,6506,6785,6903,6937,7023,7031,7074,7130,7153,7253,7357,7403,7646,7683,7765,7969,8028,8095,8234,8247,8374,8400,8493,8514,8529,8551,8566,8598,8613,8660,8672,8780,8826,8918,8938,8994,8999,9069,9131,9170,9189,9302,9304,9330,9435,9914,10170,10175,10190,10211,10295,10380,10440,10465,10519,10526,10569,11136,11189,11271,11272,11324,11545,11719,11720,11760,12064,15379,15387,15902,15932,15981,17516,20693,23382,23552,24816,24915,25344,26334,26947,28020,28299,28584,28823,28873,28981,29082,29090,29463,29609,30962,30976,32106,34024,34362,34980,35929,36631,36865,37298,37925,38174,39311,39995,41257,41315,42472,42883,45516,45551,45880,46189,46231,46617,48546,49752,50014,50162,50245,50311,50319,50338,50391,50401,50477,50705,50837,50972,51057,51117,51150,51166,51188,51267,51287,51302,51429,51442,51567,51629,51677,51684,51741,51916,51918,52010,52059,52129,52132,52181,52190,52254,52388,52420,52521,52526,52544,52546,52603,52608,52630,52650,52719,52752,52818,52933,52948,52973,52979,53074,53135,53151,53185,53201,53259,53273,53296,53321,53337,53355,53365,53437,53458,53535,53556,53579,53796,53836,53869,54000,54056,54121,54243,54251,54637,54652,54672,54679,54683,54853,54903,54915,54929,54969,54987,55152,55179,55266,55573,55635,55706,55719,55864,55973,56019,56251,56298,56314,56526,56557,56559,56587,56588,56897,56963,57032,57051,57209,57271,57333,57580,57609,57624,57855,57904,58075,58123,58304,58686,58688,58831,58862,58866,58944,58961,59020,59049,59095,59101,59160,59255,59257,59290,59705,59788,59797,59849,59964,60027,60194,60295,60373,60386,60463,60549,60554,60569,60737,60868,60902,61089,61153,61238,61406,61453,61491,61596,61644,61701,61773,61832,61932,61969,62123,62144,62354,62432,62682,62884,63012,63046,63115,63153,63180,63294,63356,63398,63464,63474,63691,63787,63808,63904,63926,63960,63979,64111,64157,64170,64237,64366,64380,64448,64531,64637,64816,64839,65158,65243,65361,65566,65576,65603,65670,65712,65754,65765,65844,66184,66390,66453,66481,66530,66600,66638,66812,66843,66874,66946,66965,67050,67110,67211,67222,67241,67246,67280,67414,67428,67652,67702,67710,67722,67878,67936,68121,68150,68243,68314,68340,68410,68450,68709,68735,68895,68909,69000,69138,69245,69252,69303,69315,69438,69472,69605,69699,69743,69816,69844,69867,70072,70155,70257,70414,70498,70561,70849,70860,70909,71004,71044,71060,71274,71275,71411,71505,71821,72023,72062,72251,72428,72937,73006,73076,73086,73274,73438,73523,73540,73542,73576,73581,73729,73874,73912,73956,74038,74039,74095,74167,74176,74337,74397,74513,74521,74559,74692,74776,74913,75008,75065,75117,75142,75378,75453,75496,75575,75728,75785,75844,75846,75873,76025,76117,76406,76452,76661,76666,76729,77151,77216,77225,77252,77265,77287,77356,77432,77507,77586,77594,77598,77645,77670,77863,77877,77878,77882,78046,78085,78101,78150,78161,78195,78197,78272,78361,78377,78386,78665,78705,78835,79037,79042,79072,79084,79291,79443,79475,79525,79659,79840,79917,79922,79926,79967,80068,80076,80087,80129,80375,80548,80591,80752,80755,80805,80919,81035,81061,81078,81105,81153,81178,81189,81244,81245,81447,81562,81599,81651,81680,81769,81818,81897,81911,82064,82212,82264,82287,82303,82328,82353,82358,82388,82608,82642,82687,82732,82748,82789,82892,82948,82960,82996,83236,83339,83362,83393,83587,83645,83735,83882,83951,83970,84033,84136,84169,84239,84276,84347,84367,84418,84554,84641,84731,84768,84921,84928,84944,85072,85173,85265,85273,85281,85346,85587,85632,85811,85973,86005,86009,86161,86252,86308,86338,86404,86479,86519,86742,86919,86977,87103,87161,87173,87234,87278,87341,87417,87452,87655,87690,87977,87981,88149,88152,88158,88204,88264,88275,88289,88417,88455,88479,88490,88689,88742,88778,88813,88845,89033,89076,89080,89094,89156,89168,89231,89276,89291,89349,89394,89782,89974,89999,90085,90094,90151,90297,90334,90354,90414,90623,90633,91027,91030,91060,91067,91099,91102,91140,91311,91479,91505,91510,91586,91723,92011,92061,92143,92309,92359,92427,92442,92524,92650,92813,93131,93265,93496,93551,93655,93681,93710,93748,93773,93819,93823,93875,93916,94089,94116,94149,94226,94231,94250,94254,94328,94357,94426,94441,94473,94518,94802,94893,95118,95455,95468,95522,95638,95738,95851,95862,95874,95929,96064,96065,96285,96345,96352,96380,96560,96719,96722,96737,96960,97216,97345,97596,97605,97640,97653,97677,97789,97833,97934,97935,97955,98031,98089,98344,98428,98431,98465,98482,98491,98500,98710,99006,99161,99323,99362,99688,99929,99947,99964,99992",
          "62,101,133,257,266,274,285,291,294,466,474,493,522,534,617,793,831,844,1059,1195,1471,1507,1551,1691,1711,1746,2751,3718,4101,4268,4781,5100,5361,5399,5471,5499,5522,5553,5565,5589,5620,5636,5721,5987,6136,6191,6294,6306,6502,6557,6654,6740,6751,6861,7058,7067,7107,7135,7327,7395,7463,7581,7589,7708,7772,8114,8139,8169,8246,8274,8329,8390,8692,8786,8937,9116,9239,9240,9315,9450,9457,9490,9519,9836,9892,9919,9995,10030,10089,10180,10315,10358,10368,10400,10518,10568,10574,10582,10652,10670,10787,10913,10965,10972,10997,11048,11099,11366,11418,11451,11668,11677,11913,11931,11977,12004,12007,12038,12425,12502,12517,12533,12674,12687,12800,12869,12877,13088,13136,13146,13153,13549,13572,13645,13668,13888,13912,13919,13980,14450,14640,14692,14751,14823,14858,15037,15045,15105,15118,15174,15178,15209,15274,15441,15447,15958,16057,16094,16434,16462,16528,16547,16676,16700,16752,16868,16992,17029,17099,17419,17649,17955,18015,18056,18068,18113,18159,18174,18599,18974,18979,19000,19238,19410,19457,19521,19758,20218,20238,20684,20698,20984,21266,21399,21557,21815,22291,22297,22590,22592,22648,22944,23016,23633,23663,23778,23923,24048,24173,24284,24300,24375,24382,25454,25933,26167,26256,26315,26816,26857,26873,27111,27125,27161,27284,27287,27411,27706,27715,27757,27779,27787,27801,27824,27914,27941,28274,28368,28461,28978,29121,29129,29297,29540,35763,35764,36075,36435,36516,36698,37015,37744,37997,38358,38379,38488,38513,38687,38816,38917,39073,39298,39302,39620,39694,39807,45872,50038,50052,50080,50097,50132,50192,50208,50212,50551,50599,50864,50954,51076,51277,51328,51351,51387,51789,51836,51969,52113,52210,52248,52367,52397,52508,52676,52714,52803,52907,52918,53129,53178,53188,53198,53313,53430,53522,53620,53701,53743,53789,53906,53981,53982,54048,54415,54421,54524,54530,54600,54909,54916,55054,55456,55625,55626,55629,55695,55749,55796,55823,55863,55912,55945,55974,56071,56088,56203,56290,56313,56344,56399,56433,56540,56832,56964,57043,57086,57123,57125,57194,57203,57233,57241,57436,57510,57529,57582,57691,57731,57789,57813,57869,57998,58010,58307,58311,58315,58408,58414,58527,58725,58736,58788,58880,58998,59174,59552,59700,59783,59812,59926,60041,60124,60126,60156,60234,60352,60690,60704,60720,60755,60953,61090,61130,61218,61433,61521,61608,61761,61788,61939,61983,62079,62094,62178,62472,62648,62708,62728,62765,62864,63060,63077,63238,63242,63288,63505,63692,63861,63955,64014,64153,64162,64190,64253,64383,64386,64538,64543,64847,64865,64880,65058,65169,65304,65363,65426,65725,65927,65949,66019,66253,66356,66460,66510,66523,66545,66590,66666,66731,66747,66748,66911,66924,67019,67023,67293,67301,67357,67375,67392,68153,68194,68525,68540,68550,68719,68757,68810,68998,69004,69075,69133,69145,69286,69337,69383,69406,69423,69434,69480,69494,69701,69707,69856,69904,69939,69946,70108,70115,70223,70333,70343,70467,70565,70600,70664,70701,70792,70915,70939,71090,71372,71389,71820,71872,71974,71990,71997,72008,72016,72071,72119,72274,72279,72438,72454,72463,72520,72615,72705,72767,72798,72802,72806,72854,72906,72917,72984,73015,73019,73020,73031,73093,73106,73130,73165,73373,73598,73605,73716,73764,73832,73846,73860,73871,74055,74079,74086,74137,74273,74346,74381,74585,74742,74767,74864,74936,74944,75147,75155,75325,75369,75380,75467,75510,75532,75550,75674,75827,75843,75938,76015,76223,76381,76473,76704,76908,76984,77266,77270,77305,77347,77365,77405,77459,77665,77822,77869,77893,77996,78017,78209,78327,78360,78428,78447,78469,78526,78658,78679,78688,78746,78760,78812,78824,78838,78908,79041,79108,79121,79123,79132,79172,79305,79310,79423,79437,79547,79629,79646,79744,79755,79823,79866,80062,80081,80094,80114,80202,80253,80326,80458,80494,80820,80848,80932,81006,81132,81205,81206,81286,81398,81409,81459,81518,81622,81681,81784,81820,81826,81830,81862,81917,81930,82032,82061,82085,82114,82310,82315,82402,82408,82422,82430,82623,82752,82876,82956,83007,83013,83038,83148,83151,83512,83549,83568,83803,83833,83909,84118,84133,84217,84240,84304,84352,84355,84393,84468,84508,84593,84649,84712,84897,84974,85181,85195,85361,85396,85505,85518,85598,85751,85764,85765,85775,85916,85933,86025,86299,86444,86511,86957,87025,87336,87351,87421,87495,87518,87530,87610,87714,87724,88123,88145,88170,88215,88266,88279,88537,88634,88751,88859,88912,88971,88997,89027,89038,89042,89121,89186,89202,89245,89251,89257,89299,89329,89342,89353,89385,89562,89583,89651,89728,89820,89956,90089,90140,90181,90287,90362,90401,90490,90522,90579,90870,90911,90976,90988,91055,91062,91084,91152,91222,91353,91544,91567,91591,91630,91690,91748,91781,91841,91845,91874,91933,91961,91969,91979,92023,92240,92264,92270,92305,92343,92498,92566,92667,92679,92713,92718,92726,92750,92835,92841,92879,93080,93098,93174,93221,93468,93524,93544,93850,93858,93877,94003,94119,94163,94174,94232,94325,94352,94574,94633,94654,94973,95022,95051,95093,95194,95215,95217,95231,95271,95287,95558,95927,96124,96159,96225,96230,96396,96424,96492,96696,96802,96874,96876,97084,97090,97154,97223,97238,97250,97319,97410,97471,97515,97550,97558,97609,97654,97778,97852,97912,97945,98026,98068,98125,98204,98208,98221,98306,98350,98401,98427,98480,98540,98750,98757,98799,98800,98877,98988,99045,99078,99130,99151,99367,99414,99459,99529,99556,99675,99871",
          "7,21,98,216,302,545,560,593,597,614,663,704,714,730,732,751,778,787,794,828,919,938,953,968,980,987,1063,1168,1316,1375,1408,1429,1443,1481,1501,1524,1603,1803,1895,1907,1993,2020,2063,2085,2086,2192,2320,2321,2324,2423,2521,2640,2737,3092,3282,3324,3653,3773,3884,4110,4273,4366,4495,4802,4862,4969,5026,5153,5369,5383,5612,5646,5711,5717,5731,5774,5862,5942,6009,6169,6357,6526,6531,6650,6679,6850,6971,6973,7009,7065,7141,7178,7306,7382,7400,7484,7532,7624,7652,7706,7715,7774,7888,7972,7975,8177,8298,8341,8411,8436,8447,8470,8526,8602,8708,8759,8783,8795,8879,8889,8943,8973,8981,9133,9186,9229,9365,9609,9631,9814,9953,9970,10122,10150,10179,10348,10349,10543,10596,10736,10954,11006,11010,11018,11053,11150,11343,11397,11427,11428,11438,11452,11537,11569,11593,11626,11770,11797,11816,11902,12218,12443,12571,12606,12608,12659,12725,12950,13119,13294,13331,13512,13637,14422,14483,14630,14854,15007,15167,15402,15616,15882,15996,16021,16030,16062,16112,16585,16848,16973,17031,17124,17241,17283,17402,17422,17614,17618,17689,18430,18543,18558,18580,18657,18768,19055,19193,19226,19356,19361,19640,19815,20404,20518,20675,20760,20784,20832,20924,21031,21041,21133,21208,21449,21514,21532,21899,22074,22161,22225,22391,22646,22799,22931,23080,23090,23374,23403,23441,23545,23723,23749,24087,24320,24333,24397,24524,24546,24787,24841,24868,24932,24979,25046,25336,25366,25439,25486,25641,25732,25750,25884,26047,26251,26321,26323,26332,26459,26661,26703,26715,26797,26814,27214,27227,27279,27352,27364,27447,27649,27662,27678,27735,28290,28323,28366,28426,28432,28524,28700,28734,28882,28931,28972,29016,29042,29100,29198,29280,29345,29386,29792,29803,29824,29852,29984,30578,31097,32373,32634,32782,32811,33040,33127,33417,33446,34142,35378,35572,35800,35926,35943,36076,36282,36300,36313,36442,36500,36510,36620,36638,36641,36814,36941,37045,37261,37314,37366,37380,37498,37549,37819,37950,38067,38258,38324,38368,38437,38441,38500,38675,38696,38757,39000,39351,39362,39610,39682,39729,39752,39803,39872,40217,40509,41198,41290,42323,42488,42550,42573,42722,42781,42860,43111,43196,43394,43433,43885,44302,44430,44463,45185,45189,45235,45252,45329,45881,45950,45993,46118,46455,46574,46589,46830,46873,47555,48477,49057,49198,49582,49675,50060,50172,50187,50378,50380,50442,50490,50525,50544,50663,50788,50883,50916,50921,51365,51436,51528,51532,51663,51755,51946,51993,52027,52068,52082,52168,52237,52251,52304,52423,52491,52499,52887,52889,52969,53050,53224,53256,53473,53576,53662,53824,53848,53880,53996,54023,54036,54083,54096,54107,54194,54265,54323,54395,54580,54610,54748,54838,55056,55068,55074,55108,55151,55315,55335,55366,55461,55505,55621,55759,55908,56025,56093,56101,56234,56249,56258,56302,56383,56390,56507,56514,56551,56753,56767,56787,56807,56809,56952,57101,57134,57239,57306,57433,57620,57719,57915,57928,57961,57962,57981,58007,58070,58150,58328,58348,58487,58562,58648,58739,58754,58792,58845,59006,59184,59221,59281,59314,59408,59698,59768,59813,59851,59868,59899,59925,59987,60089,60151,60209,60251,60338,60437,60438,60586,60594,60623,60764,60922,60933,61048,61243,61345,61351,61485,61532,61551,61704,61847,61985,62061,62074,62171,62172,62238,62325,62367,62588,62634,62643,62793,62885,62906,62941,63000,63129,63362,63376,63548,63569,63570,63596,63651,63707,63869,63872,63901,64039,64206,64221,64241,64371,64546,64724,64747,64772,65064,65233,65305,65374,65438,65455,65460,65489,65653,65751,65862,65940,66037,66259,66334,66526,66605,66700,66732,66753,66967,67096,67131,67304,67475,67524,67598,67612,67635,67655,67721,67734,67838,67923,67944,68004,68206,68212,68286,68347,68518,68729,68778,68883,68956,68985,69087,69113,69117,69125,69134,69454,69485,69623,69667,69830,69892,69960,70225,70240,70272,70287,70321,70366,70442,70520,70564,70618,70727,71067,71085,71198,71228,71287,71345,71412,71502,71530,71783,71833,71867,71871,71923,72103,72104,72259,72443,72518,72531,72654,72755,72769,72847,72953,73097,73100,73141,73244,73312,73394,73473,73593,73724,73887,73957,74081,74129,74156,74217,74398,74508,74578,74630,74655,74729,74840,74897,74943,75034,75091,75100,75137,75195,75204,75219,75300,75360,75524,75600,75601,75607,75644,75671,75699,75965,76029,76056,76096,76208,76217,76230,76260,76400,76458,76516,76536,76559,76684,76797,77034,77187,77219,77232,77329,77461,77552,77679,77819,77828,77867,77978,78030,78040,78096,78136,78277,78357,78509,78678,78916,78969,79136,79167,79270,79343,79455,79458,79634,79711,79756,79772,79777,79902,80032,80100,80109,80110,80148,80169,80231,80301,80587,80711,80846,80945,80953,81021,81052,81195,81199,81237,81250,81254,81295,81346,81500,81605,81610,81770,81786,81843,81874,81923,82475,82503,82510,82704,82730,82769,82881,82904,82906,82953,82973,83018,83235,83266,83281,83356,83380,83394,83464,83577,83582,83703,83820,83911,83976,84012,84095,84096,84203,84216,84287,84313,84327,84457,84606,84761,84853,84960,85122,85209,85264,85330,85397,85437,85470,85500,85614,85636,85661,85739,85797,85861,85934,86111,86139,86181,86253,86280,86297,86473,86586,86628,86817,86904,86935,87098,87166,87195,87265,87328,87342,87623,87697,87787,87877,87973,88021,88119,88155,88188,88351,88651,88725,88736,88781,88853,88881,88887,88931,88942,89188,89416,89449,89450,89643,89668,89833,90232,90262,90303,90340,90380,90392,90400,90478,90685,90884,90923,91038,91081,91206,91347,91494,91504,91562,91710,91712,91823,91910,91955,92135,92228,92395,92488,92709,92917,92941,92985,93049,93184,93201,93256,93396,93398,93403,93454,93804,93970,94205,94215,94498,94593,94600,94623,94629,94638,94687,94766,94791,94801,94849,94879,95157,95171,95296,95328,95389,95400,95418,95433,95474,95765,95773,95828,96187,96229,96336,96381,96386,96439,96481,96647,96736,96750,96839,97142,97150,97275,97286,97328,97344,97452,97481,97845,98113,98288,98365,98385,98447,98510,98536,98657,98713,98780,98804,98822,98932,99007,99030,99135,99173,99225,99280,99472,99484,99647,99746,99794,99838,99856,99982",
          "136,184,230,1620,1676,1745,1764,1992,2011,2061,2122,2149,2173,2209,2438,2440,2464,2582,2595,2691,2839,2985,3015,3050,3285,3329,3474,3695,4025,4063,4271,4306,4391,4488,4558,4581,4597,4618,4642,4849,4892,5075,7120,7373,7773,8393,9184,9497,9611,9630,9649,9668,9706,9720,9725,9782,9801,10048,10599,10618,10675,10713,10725,10808,10903,10941,11017,11093,11112,11131,11226,11227,11245,11359,11444,11454,11511,11525,11528,11550,11587,11606,11644,11663,11891,11948,11967,11986,12024,12157,12233,12252,12271,12290,12328,12329,12385,12403,12423,12478,12499,12537,12596,12613,12639,12668,12679,12708,12765,12803,12847,12860,12883,12898,12923,12974,13012,13048,13050,13069,13114,13145,13164,13240,13297,13345,13354,13362,13373,13430,13449,13563,13617,13639,13658,13677,13715,13734,13772,13791,13829,13838,13855,13870,13889,13962,13981,14019,14045,14076,14082,14095,14147,14152,14171,14190,14201,14228,14247,14266,14285,14304,14323,14332,14380,14418,14446,14475,14502,14570,14571,14646,14665,14672,14722,14741,14742,14759,14859,14874,14931,15026,15064,15083,15100,15140,15197,15235,15273,15292,15311,15349,15368,15425,15444,15501,15577,15595,15596,15605,15640,15672,15691,15710,15748,15805,15824,15843,15862,15900,15919,15938,15942,15976,15988,16052,16071,16090,16100,16147,16166,16183,16185,16204,16221,16242,16261,16273,16280,16293,16299,16318,16337,16356,16413,16428,16432,16451,16458,16470,16508,16527,16565,16584,16622,16698,16717,16793,16831,16888,16929,16962,16964,17021,17040,17059,17079,17097,17154,17173,17192,17211,17239,17268,17306,17325,17363,17382,17401,17458,17496,17515,17534,17591,17624,17648,17686,17705,17743,17762,17800,17819,17838,17876,17895,17914,17971,17990,18028,18066,18104,18123,18142,18161,18170,18216,18218,18237,18250,18256,18275,18313,18332,18351,18370,18389,18408,18427,18484,18503,18523,18579,18587,18598,18617,18674,18693,18731,18734,18807,18826,18854,18883,18902,18940,18941,18959,18978,18997,19021,19028,19044,19054,19073,19092,19130,19206,19232,19282,19320,19377,19396,19415,19434,19468,19472,19529,19544,19548,19586,19624,19662,19700,19724,19738,19757,19776,19795,19814,19833,19852,19871,19890,19928,19966,20023,20061,20080,20099,20108,20156,20182,20194,20203,20232,20251,20319,20327,20346,20365,20384,20403,20422,20441,20460,20517,20536,20555,20574,20612,20631,20669,20688,20726,20745,20764,20775,20840,20859,20878,20916,20935,20954,20992,21011,21049,21087,21106,21159,21163,21201,21220,21239,21296,21303,21353,21372,21410,21429,21505,21524,21562,21591,21600,21607,21619,21638,21647,21657,21676,21714,21752,21790,21809,21828,21866,21885,21904,21923,21942,21961,21980,21999,22037,22056,22075,22094,22113,22151,22170,22227,22247,22256,22265,22284,22322,22330,22341,22360,22398,22417,22436,22449,22455,22512,22522,22531,22569,22580,22607,22626,22683,22702,22721,22723,22759,22797,22816,22820,22835,22930,22987,23025,23044,23082,23101,23120,23177,23196,23291,23310,23367,23394,23424,23453,23472,23481,23500,23538,23557,23576,23582,23614,23652,23671,23688,23709,23718,23728,23747,23782,23804,23861,23880,23899,23918,23937,23956,23975,24013,24032,24070,24101,24127,24144,24165,24184,24203,24260,24276,24298,24317,24355,24374,24393,24412,24433,24450,24469,24507,24526,24545,24583,24602,24612,24621,24659,24678,24697,24716,24773,24811,24830,24883,24887,24925,24952,24963,24982,24988,25020,25058,25064,25096,25115,25134,25191,25248,25267,25268,25305,25318,25324,25343,25345,25419,25438,25476,25495,25552,25590,25609,25627,25780,25799,25818,25863,25875,25913,25929,25951,25970,26008,26027,26046,26065,26084,26103,26107,26122,26131,26141,26179,26185,26198,26227,26236,26255,26274,26331,26350,26369,26388,26407,26426,26445,26464,26502,26521,26578,26635,26683,26758,26958,27110,27140,27205,27414,27547,27561,27574,27604,27680,27699,27865,27884,27908,28003,28155,28174,28478,28491,28516,28535,28554,28592,28649,28703,28737,28822,28877,28988,29111,29124,29200,29216,29286,29352,29504,29549,29580,29618,29656,29703,29979,30026,30055,30084,30113,30142,30171,30200,30229,30258,30287,30316,30345,30374,30403,30432,30461,30490,30519,30548,30577,30606,30635,30664,30693,30722,30751,30780,30809,30838,30867,30896,30925,30954,30983,31012,31041,31070,31099,31128,31157,31186,31215,31244,31273,31302,31360,31389,31418,31447,31476,31505,31534,31563,31592,31621,31650,31679,31708,31737,31766,31795,31824,31853,31882,31911,31940,31969,31998,32027,32056,32085,32114,32143,32172,32201,32230,32259,32288,32317,32346,32375,32404,32433,32462,32491,32520,32549,32578,32607,32636,32665,32694,32723,32752,32781,32810,32839,32868,32897,32926,32955,32984,33013,33042,33071,33100,33129,33158,33187,33216,33245,33274,33303,33332,33361,33390,33419,33448,33477,33506,33535,33564,33593,33622,33651,33680,33709,33738,33767,33796,33825,33854,33883,33912,33941,33970,33999,34028,34057,34086,34115,34144,34173,34202,34231,34260,34289,34318,34347,34376,34405,34434,34463,34492,34521,34550,34579,34608,34637,34666,34695,34724,34753,34782,34811,34840,34869,34898,34927,34956,34985,35026,35068,35115,35131,35146,35189,35203,35207,35212,35225,35227,35252,35278,35286,35316,35352,35426,35438,35442,35454,35497,35502,35545,35604,35610,35663,35690,35736,35795,35818,35854,35925,35984,36043,36106,36166,36225,36231,36233,36245,36296,36355,36360,36377,36378,36414,36486,36545,36557,36604,36668,36727,36786,36857,36916,36975,37047,37106,37120,37165,37229,37252,37288,37297,37347,37418,37477,37536,37566,37608,37667,37726,37731,37790,37808,37849,37908,37979,38007,38038,38097,38106,38169,38180,38185,38228,38287,38308,38351,38410,38469,38540,38564,38599,38620,38658,38730,38740,38789,38848,38912,38927,38971,39030,39040,39045,39101,39110,39160,39219,39228,39291,39350,39409,39473,39532,39591,39600,39662,39721,39727,39780,39852,39911,39970,40102,40131,40160,40189,40218,40247,40276,40305,40334,40363,40392,40421,40450,40479,40508,40537,40566,40595,40624,40653,40682,40711,40740,40769,40798,40827,40856,40885,40914,40943,40972,41001,41030,41059,41088,41117,41146,41175,41204,41233,41262,41291,41320,41349,41378,41407,41436,41465,41494,41523,41997,42026,42055,42084,42113,42139,42141,42173,42236,42239,42263,42288,42306,42338,42413,42457,42498,42500,42538,42575,42618,42650,42695,42737,42782,42812,42861,42887,42974,43011,43026,43032,43049,43124,43188,43193,43211,43240,43256,43265,43286,43313,43323,43335,43361,43386,43388,43448,43523,43598,43633,43638,43671,43685,43713,43760,43835,43898,43922,43964,43997,44028,44072,44159,44215,44234,44309,44338,44371,44376,44396,44452,44456,44457,44471,44523,44546,44605,44607,44633,44700,44783,44870,44897,44899,44945,45020,45048,45050,45065,45093,45141,45202,45260,45291,45333,45385,45390,45420,45424,45448,45520,45577,45635,45660,45707,45739,45764,45811,45822,45894,45951,46076,46125,46151,46210,46224,46230,46321,46334,46370,46402,46453,46493,46503,46518,46537,46549,46551,46620,46707,46743,46754,46774,46837,46853,46885,46904,46922,46938,46956,46978,46995,47026,47035,47037,47051,47140,47150,47215,47232,47245,47255,47259,47332,47350,47362,47420,47454,47465,47470,47506,47507,47508,47511,47549,47553,47563,47585,47599,47604,47606,47624,47625,47654,47671,47684,47711,47737,47752,47774,47778,47797,47810,47827,47904,48020,48055,48156,48238,48255,48263,48285,48303,48305,48352,48381,48382,48429,48471,48472,48514,48529,48548,48578,48652,48668,48729,48769,48787,48807,48810,48892,48927,48933,48970,49014,49067,49074,49083,49091,49119,49160,49206,49218,49267,49293,49304,49322,49347,49432,49442,49449,49536,49550,49569,49601,49631,49647,49653,49657,49709,49714,49717,49733,49769,49798,49833,49851,49895,49903,49929,49932,49993,50063,50081,50203,50213,50457,50601,50648,50860,50940,50947,51058,51110,51154,51185,51234,51286,51354,51489,51622,51775,51837,51883,51896,51925,51939,52074,52134,52149,52182,52186,52197,52370,52392,52687,52783,52811,52852,52857,52919,53008,53021,53075,53095,53119,53161,53175,53190,53292,53310,53403,53427,53442,53552,53553,53624,53714,53734,53774,53778,53922,53933,54025,54060,54145,54158,54193,54281,54479,54498,54542,54561,54604,54655,54697,54961,54980,54993,55026,55047,55135,55236,55245,55375,55404,55433,55467,55616,55731,55757,55836,55855,55865,55894,55922,55944,56274,56286,56364,56379,56491,56633,56793,56794,56891,56978,57180,57201,57276,57385,57437,57516,57549,57734,57762,57884,58290,58295,58629,58692,58717,58824,58894,58940,58984,59093,59144,59146,59284,59543,59573,59595,59603,59967,59984,60036,60345,60377,60398,60468,60475,60706,60710,60811,60832,60876,60915,61011,61029,61062,61095,61106,61117,61150,61161,61331,61391,61461,61484,61523,61637,61729,61902,61907,61946,61968,62017,62072,62104,62236,62258,62318,62385,62428,62448,62474,62683,62687,62699,62770,62822,62872,62953,62983,63116,63121,63309,63340,63440,63442,63500,63797,63898,63951,64022,64026,64051,64134,64260,64426,64445,64645,64676,64692,64881,64923,65166,65198,65209,65220,65249,65265,65297,65314,65351,65382,65397,65504,65673,65719,65763,65993,66031,66117,66149,66372,66571,66616,66627,66634,66656,66684,66738,66785,66813,66870,66901,66990,67037,67046,67066,67127,67183,67265,67308,67343,67406,67482,67651,67656,67672,67772,67809,67881,68163,68169,68306,68478,68594,68800,68806,68889,68929,69157,69164,69283,69440,69772,69915,70060,70160,70312,70376,70500,70503,70597,70708,70735,70799,70803,70881,70912,70931,71178,71252,71270,71349,71494,71662,71732,72010,72037,72057,72109,72115,72240,72247,72307,72432,72435,72450,72578,72620,72676,72688,72816,72855,72874,73046,73081,73109,73139,73193,73350,73382,73429,73602,73604,73625,73763,73858,73962,73973,73983,74043,74170,74173,74174,74281,74483,74493,74562,74590,74597,74603,74616,74639,74686,74736,75045,75090,75115,75224,75289,75329,75411,75652,75952,75961,76040,76084,76086,76129,76152,76181,76218,76336,76347,76398,76402,76431,76472,76624,76778,76788,76872,76880,76920,77159,77466,77522,77636,77701,77917,77944,77965,77990,78026,78098,78378,78543,78581,78662,78715,78783,78836,78910,78986,79017,79021,79047,79054,79096,79163,79230,79254,79256,79303,79313,79470,79507,79531,79719,79921,79956,80063,80277,80400,80473,80490,80762,81092,81111,81131,81134,81141,81147,81227,81233,81283,81568,81570,81577,81589,81645,81712,81798,82083,82374,82423,82452,82496,82613,82827,82889,82958,83019,83173,83508,83516,83762,83927,83933,84007,84146,84151,84181,84194,84269,84450,84704,84806,85059,85131,85137,85198,85212,85218,85280,85293,85443,85453,85546,85588,85604,85681,85722,85951,86048,86072,86117,86149,86172,86186,86189,86193,86265,86275,86416,86468,86486,86566,86582,86685,86785,86861,86872,86889,86921,87048,87148,87213,87244,87289,87420,87454,87466,87504,87527,87548,87976,87996,88019,88124,88233,88342,88362,88378,88420,88432,88459,88489,88530,88572,88635,88735,88762,88772,88815,88864,88917,88929,88932,89041,89135,89269,89424,89442,89555,89721,89724,90002,90066,90470,90659,90709,90720,90775,90891,90900,91035,91078,91241,91306,91326,91424,91437,91583,91609,91668,91697,91705,91830,92022,92295,92333,92585,92717,92772,92848,92967,93172,93251,93305,93336,93371,93652,94126,94217,94312,94372,94382,94546,94549,94635,94667,94819,94824,94885,94930,95153,95279,95295,95383,95641,95648,95653,95678,95687,95854,95911,95990,96073,96264,96277,96467,96715,96783,96815,96856,96971,97136,97146,97209,97548,97608,97642,97809,97829,97837,97972,98008,98042,98227,98277,98388,98419,98476,98487,98519,98521,98572,98581,98616,98627,98686,98702,98752,98760,98857,99019,99037,99223,99271,99289,99417,99818,99860",
          "57,95,114,152,171,223,1679,1702,1721,2302,2356,2358,2536,2709,2884,3299,3310,3336,3458,3628,3746,3801,3898,4066,4130,4398,4420,4460,4502,4591,4674,4867,4887,4997,4999,5103,5417,5672,6303,6592,7173,7724,8405,8426,9495,9500,9508,9538,9559,9595,9614,9633,9652,9671,9690,9709,9764,9822,9838,10092,10621,10678,10697,10734,10759,11134,11210,11229,11381,11394,11410,11495,11571,11613,11628,11647,11856,11862,11901,11932,11970,12008,12103,12106,12141,12154,12160,12189,12217,12236,12255,12274,12323,12350,12369,12435,12464,12511,12616,12635,12673,12768,12787,12806,12815,12861,12863,12882,12921,12939,13034,13053,13091,13110,13186,13205,13224,13229,13281,13300,13357,13366,13376,13412,13438,13452,13490,13564,13604,13608,13642,13661,13680,13699,13718,13756,13775,13813,13872,13908,13927,13946,13964,14022,14079,14102,14136,14142,14155,14193,14206,14212,14231,14250,14269,14307,14326,14345,14403,14459,14497,14560,14573,14592,14649,14706,14715,14719,14731,14744,14768,14769,14786,14801,14820,14877,14911,14934,15010,15067,15086,15143,15200,15257,15295,15314,15390,15466,15496,15532,15618,15668,15675,15808,15827,15846,15884,15913,15927,15960,15979,15995,15998,16017,16024,16093,16131,16169,16226,16245,16283,16321,16340,16359,16378,16397,16473,16511,16534,16606,16699,16701,16743,16758,16815,16834,16853,16872,16882,16921,16967,17024,17043,17074,17096,17125,17157,17176,17178,17179,17214,17252,17271,17309,17347,17366,17385,17480,17499,17537,17544,17557,17575,17621,17632,17727,17746,17765,17784,17803,17822,17841,17860,17879,17974,18050,18164,18179,18202,18215,18221,18240,18278,18297,18316,18354,18373,18392,18440,18468,18487,18493,18566,18582,18601,18639,18677,18775,18808,18810,18829,18866,18886,18905,18924,18981,19029,19076,19081,19133,19152,19171,19209,19228,19285,19304,19323,19342,19380,19399,19418,19456,19551,19552,19570,19589,19650,19665,19678,19719,19722,19750,19779,19798,19817,19836,19855,19893,19912,19950,19988,20045,20083,20102,20121,20197,20216,20222,20235,20254,20292,20311,20330,20349,20368,20387,20406,20425,20444,20463,20482,20539,20558,20596,20650,20680,20710,20767,20824,20854,20881,20995,21014,21052,21109,21204,21242,21273,21280,21314,21318,21356,21394,21432,21451,21489,21495,21508,21565,21584,21622,21641,21646,21660,21679,21717,21742,21755,21774,21812,21831,21850,21869,21888,21907,21945,21964,21983,22002,22040,22078,22097,22116,22154,22186,22211,22249,22268,22287,22344,22382,22401,22420,22477,22494,22515,22534,22553,22591,22610,22667,22691,22705,22727,22743,22762,22781,22810,22821,22876,22934,22952,22971,22990,23066,23085,23140,23142,23143,23156,23180,23237,23256,23275,23292,23294,23320,23332,23371,23389,23406,23427,23446,23484,23503,23522,23560,23579,23598,23636,23693,23712,23731,23788,23807,23859,23864,23883,23921,23940,23959,23978,23997,24054,24111,24189,24225,24282,24301,24339,24358,24377,24396,24415,24434,24444,24453,24510,24529,24548,24624,24627,24662,24681,24682,24719,24833,24852,24871,24890,24904,24947,24985,25004,25023,25025,25099,25118,25152,25156,25175,25186,25194,25203,25227,25232,25270,25289,25346,25365,25393,25403,25479,25498,25517,25536,25574,25631,25650,25701,25707,25726,25764,25768,25821,25878,25916,25935,25954,25973,25992,26011,26030,26068,26087,26144,26182,26201,26258,26277,26287,26372,26391,26429,26448,26467,26486,26619,26828,26847,26885,26904,26923,26934,26999,27170,27359,27402,27531,27591,27816,27818,27920,27922,27930,27937,27949,28055,28082,28183,28329,28419,28479,28557,28633,28652,28695,28708,28905,28918,28956,29025,29032,29061,29080,29097,29108,29114,29203,29248,29380,29479,29488,29513,29611,29763,29830,29874,29922,29923,29944,30029,30058,30087,30145,30174,30203,30232,30261,30290,30319,30348,30377,30406,30435,30464,30493,30522,30551,30580,30609,30638,30667,30696,30754,30783,30812,30841,30870,30899,30928,30957,31015,31044,31073,31102,31131,31189,31218,31247,31276,31305,31334,31363,31392,31421,31450,31479,31508,31537,31566,31595,31624,31682,31711,31740,31769,31798,31827,31856,31885,31914,31943,31972,32001,32030,32059,32117,32146,32204,32233,32262,32291,32320,32349,32378,32407,32436,32465,32494,32523,32552,32581,32610,32639,32668,32697,32726,32755,32784,32813,32842,32871,32900,32929,32958,32987,33016,33045,33074,33103,33132,33161,33190,33219,33248,33306,33335,33393,33422,33451,33480,33509,33538,33567,33596,33625,33654,33683,33712,33741,33770,33799,33828,33857,33886,33915,33944,33973,34002,34031,34060,34089,34118,34147,34176,34205,34234,34263,34292,34321,34350,34379,34408,34437,34466,34495,34524,34553,34582,34611,34640,34669,34698,34727,34756,34785,34814,34843,34872,34901,34930,34959,34988,35029,35118,35154,35192,35246,35281,35288,35307,35355,35362,35366,35429,35500,35515,35548,35607,35611,35666,35671,35739,35798,35858,35868,35899,35928,35987,36074,36109,36169,36228,36229,36250,36299,36314,36417,36422,36489,36507,36548,36607,36696,36730,36749,36789,36860,36919,36922,36927,36937,36946,36949,36978,37050,37109,37168,37232,37291,37317,37350,37354,37421,37480,37539,37670,37683,37729,37793,37820,37852,37911,37924,37982,38041,38100,38103,38124,38172,38231,38354,38413,38472,38543,38606,38661,38663,38691,38704,38733,38745,38792,38872,38873,38915,38974,39033,39038,39104,39163,39222,39236,39294,39353,39412,39447,39476,39535,39594,39665,39724,39745,39855,39874,39973,40187,40216,40245,40303,40332,40361,40390,40419,40448,40477,40506,40535,40564,40593,40622,40651,40680,40709,40738,40767,40796,40825,40854,40883,40912,40941,40970,40999,41028,41057,41086,41115,41144,41173,41202,41231,41260,41318,41347,41376,41405,41434,41463,41492,41521,41988,41993,42051,42080,42109,42168,42227,42243,42300,42318,42408,42536,42543,42555,42612,42633,42643,42645,42720,42725,42792,42882,42957,42967,43029,43033,43039,43194,43266,43324,43356,43357,43401,43415,43431,43499,43503,43593,43668,43717,43729,43740,43745,43826,43830,43864,43877,43905,43959,43968,43977,43990,44060,44067,44104,44137,44142,44145,44184,44197,44214,44263,44304,44341,44359,44373,44379,44392,44437,44459,44465,44508,44516,44522,44540,44541,44575,44587,44616,44630,44658,44665,44671,44688,44778,44826,44851,44853,44904,44925,44982,45015,45047,45086,45128,45134,45170,45195,45244,45253,45292,45326,45441,45513,45570,45628,45700,45757,45815,45887,45944,45946,46027,46055,46056,46067,46082,46083,46119,46129,46131,46152,46245,46330,46371,46385,46439,46443,46449,46469,46470,46519,46633,46708,46732,46792,46876,46886,46890,46970,46987,46991,47128,47236,47390,47395,47416,47426,47498,47513,47518,47532,47552,47610,47656,47659,47663,47682,47723,47729,47840,47870,47942,48147,48234,48277,48361,48384,48403,48404,48411,48439,48452,48475,48505,48509,48608,48630,48643,48674,48693,48695,48715,48763,48802,48808,48890,48978,48987,49023,49029,49045,49048,49070,49087,49110,49115,49122,49131,49139,49229,49286,49314,49343,49358,49373,49374,49398,49508,49514,49517,49528,49543,49585,49586,49609,49616,49648,49708,49710,49782,49787,49868,49911,49936,49979,50082,50424,50456,50472,50627,50729,50875,50975,51035,51065,51272,51335,51412,51632,51719,51780,51827,51830,51938,52083,52121,52176,52293,52341,52345,52354,52390,52507,52837,52956,53040,53116,53265,53269,53317,53386,53392,53751,53854,53899,54006,54150,54236,54389,54417,54470,54480,54560,54594,54644,54736,54738,54776,54821,54896,55032,55202,55207,55240,55352,55378,55515,55800,55805,55898,55977,56137,56144,56284,56335,56367,56484,56639,56715,56909,56924,57000,57066,57229,57459,57478,57639,57764,57766,58060,58218,58346,58435,58513,58588,58667,58732,58766,58914,58935,58978,59075,59112,59121,59217,59261,59344,59368,59476,59500,59581,59732,59800,59898,59935,59992,60199,60256,60306,60440,60474,60687,60688,60707,60771,60814,60830,60879,61274,61308,61494,61547,61562,61758,61786,61855,61887,61963,62100,62133,62186,62340,62528,62609,62651,62677,62808,62821,62856,62928,62961,63133,63164,63326,63688,63739,63850,63939,63978,64033,64067,64096,64115,64194,64215,64289,64302,64480,64562,64594,64610,64621,64627,64677,64721,64952,65050,65185,65268,65311,65335,65757,65780,65992,66083,66102,66116,66233,66260,66262,66307,66313,66316,66407,66574,66587,66606,66702,66783,66930,66991,67005,67055,67090,67180,67187,67254,67260,67316,67360,67369,67564,67570,67600,68231,68331,68423,68747,68750,68820,68980,69092,69111,69220,69288,69339,69379,69475,69479,69632,69688,69697,69720,69746,69796,69845,69849,70049,70251,70295,70320,70465,70486,70509,70522,70722,70731,70832,70840,70990,71043,71049,71130,71147,71214,71292,71296,71415,71471,71486,71549,71607,71622,71686,71781,71864,71885,71913,71934,72034,72154,72156,72267,72420,72585,72611,72823,72952,72980,73025,73175,73292,73357,73388,73470,73471,73608,73632,73797,74048,74324,74334,74345,74613,74853,74909,74966,75017,75020,75069,75498,75522,75544,75710,75854,75919,75967,76003,76073,76183,76197,76433,76564,77007,77036,77055,77095,77115,77147,77169,77185,77204,77209,77213,77239,77298,77428,77560,77587,77590,77616,77693,77757,77770,77834,77926,78032,78086,78110,78285,78557,78585,78587,78751,78787,78804,78954,79007,79049,79098,79141,79193,79260,79314,79404,79574,79578,79620,79691,79724,79787,79808,79860,79886,79942,80115,80174,80282,80307,80359,80459,80517,80542,80561,80650,80663,80684,80745,80817,80954,80975,81027,81056,81308,81686,81865,81877,81987,82030,82086,82334,82351,82383,82413,82584,82593,83121,83180,83244,83348,83505,83580,83599,83648,83682,83788,83950,84021,84052,84075,84219,84226,84296,84310,84422,84475,84483,84523,84570,84571,84573,84596,84620,84629,84675,84690,84850,84924,84953,85008,85017,85139,85210,85223,85278,85301,85475,85528,85570,85728,85824,85838,85841,86148,86208,86250,86294,86332,86335,86359,86374,86390,86421,86456,86793,86794,86925,87339,87390,87538,87539,87629,87688,87718,87764,87865,87932,87962,88142,88150,88211,88239,88258,88321,88449,88650,88698,89013,89025,89064,89109,89115,89137,89214,89253,89335,89341,89363,89470,89518,89524,89536,89548,89559,89752,89873,89937,90011,90168,90174,90229,90308,90311,90336,90359,90363,90374,90378,90403,90498,90518,90542,90557,90619,90692,90710,90731,90784,90851,90899,90932,90956,90962,91119,91278,91331,91354,91383,91625,91693,91730,91849,91904,91905,91912,91976,92049,92291,92325,92462,92735,92819,92925,92927,92957,92962,92980,93094,93159,93392,93422,93434,93623,93670,93694,93845,94183,94507,94585,94603,94639,94728,94749,94868,94999,95010,95040,95059,95368,95406,95534,95544,95585,95661,95683,95745,95881,95917,95986,96016,96029,96140,96166,96206,96250,96312,96394,96407,96585,96604,96702,96875,96957,97021,97272,97451,97528,97554,97600,97657,97721,97749,97854,97863,97882,97899,97998,98071,98167,98269,98436,98470,98494,98522,98595,98680,98776,98858,99157,99235,99237,99242,99425,99464,99505,99576,99593,99624,99709,99720,99828,99872,99874,99944,99996",
          "1,279,342,372,411,442,495,529,536,570,604,619,671,682,819,838,935,941,976,1001,1329,1350,1356,1399,1400,1580,1596,1708,1871,2277,2607,3117,3271,3831,3919,4158,5033,5095,5142,5170,5261,5342,5355,5357,5389,5478,5616,5803,5882,6020,6055,6105,6119,6333,6488,6537,6539,6612,6656,6670,6695,6718,6796,6823,6839,6853,6995,7039,7196,7211,7242,7342,7371,7464,7481,7515,7545,7566,7962,7985,8052,8138,8330,8389,8531,8698,8720,8813,8825,9012,9021,9059,9165,9200,9241,9256,9321,9392,9439,9465,9557,9564,9863,9932,9956,9991,10015,10103,10249,10263,10293,10527,10669,10701,10799,10824,10940,11028,11041,11057,11405,11433,11501,11771,11829,11864,11919,12311,12322,12409,12455,12539,12576,12600,12643,12688,12848,12972,12992,13047,13107,13181,13195,13275,13554,13579,13623,13880,14347,14437,14488,14552,14594,14596,14644,14830,14889,14983,14999,15159,15189,15366,16105,16219,16457,16725,17162,17269,17411,17931,18007,18053,18169,18465,18777,18794,19134,19683,19947,20081,20096,20653,20676,20709,20768,20783,21454,21496,21663,22026,22232,22255,22790,22951,23118,23141,23192,23702,24481,24714,24794,24817,24856,24946,24996,25049,25068,25250,25608,25635,26206,26624,26850,26866,26938,27218,27257,27479,27639,27761,27805,27871,27880,28175,28224,28291,28452,28510,28696,28699,28849,28892,29019,29252,29585,29601,29669,29837,29880,29963,31532,32028,32074,35827,36016,36457,36493,36806,36810,37136,37199,37434,37578,37617,37679,37803,38175,38422,38436,38510,38681,38928,39046,39263,39545,39666,39866,39883,45369,45431,45497,45783,45873,45876,45928,45992,50173,50197,50275,50347,50350,50364,50421,50510,50550,50581,50600,50760,50818,50946,51027,51097,51306,51317,51337,51465,51770,51805,51917,52014,52070,52072,52095,52270,52434,52575,52626,52703,52717,52825,52830,52848,52849,52872,52951,52987,53056,53243,53429,53445,53542,53636,53739,53747,53812,53913,53964,54079,54125,54183,54379,54493,54603,54632,54661,54711,54928,54931,54957,54963,55003,55139,55144,55271,55459,55479,55660,55753,55886,55915,56141,56326,56481,56486,56510,56589,57078,57243,57297,57435,57458,57636,57655,57788,57834,58079,58133,58136,58143,58207,58223,58263,58309,58510,58568,58628,58763,59012,59019,59353,59375,59382,59414,59561,59859,60024,60046,60129,60160,60250,60388,60470,60551,60573,60658,60743,60794,60908,60932,60979,61071,61085,61091,61122,61137,61149,61311,61316,61339,61486,61579,61647,61661,61678,61755,61816,61836,62142,62175,62180,62244,62283,62513,62628,62732,62761,62969,63059,63208,63214,63261,63341,63344,63478,63583,63590,63680,63732,64041,64119,64163,64203,64255,64491,64601,64650,64691,64698,64730,64974,65095,65470,65541,65931,65974,66017,66038,66068,66091,66243,66269,66287,66305,66322,66324,66415,66434,66469,66517,66599,66661,66664,66787,66796,66876,67258,67311,67337,67341,67449,67468,67541,67562,67727,67874,67908,67922,68021,68024,68067,68203,68316,68415,68522,68562,68579,68716,68878,68923,68943,69154,69170,69197,69209,69214,69240,69242,69291,69320,69350,69357,69361,69464,69580,69627,69647,69812,69822,69840,69878,69947,70079,70113,70140,70154,70176,70188,70431,70455,70553,70566,70611,70655,70853,70911,70978,71010,71213,71217,71230,71253,71255,71449,71690,71750,72055,72058,72073,72077,72084,72210,72228,72323,72405,72616,72619,72681,72734,72765,72911,72989,73118,73186,73289,73347,73444,73687,73713,73736,73849,73951,73982,73996,74042,74286,74407,74443,74453,74490,74542,74594,74700,74774,74846,75002,75078,75186,75198,75553,75714,75886,75890,75968,75986,76019,76033,76059,76105,76175,76243,76314,76414,76493,76566,76604,76664,76696,76786,76865,76962,76968,77030,77068,77233,77240,77282,77286,77318,77439,77529,77599,77715,77755,77825,77830,78054,78126,78220,78246,78302,78333,78347,78373,78380,78498,78592,78617,78701,78841,78851,78875,78886,79013,79140,79321,79447,79502,79731,79818,79900,79994,80175,80352,80404,80502,80577,80624,81036,81086,81088,81154,81536,81579,81695,81696,81795,81871,81965,82015,82109,82161,82169,82174,82250,82482,82489,82502,82587,82639,82692,82695,82742,82754,82894,83010,83200,83328,83456,83476,83479,83494,83517,83529,83548,83551,83667,83697,83748,83849,83984,84017,84148,84161,84210,84253,84366,84674,84751,84868,84903,85052,85205,85531,85548,85647,85776,85821,85914,86086,86134,86376,86581,86746,86933,86943,87010,87016,87135,87222,87355,87453,87603,87644,87682,87752,87761,87873,87886,88016,88098,88162,88209,88375,88431,88484,88510,88582,88593,88599,88646,88874,89130,89142,89343,89379,89426,89523,89619,89631,89738,89757,89929,90044,90214,90425,90473,90488,90554,90578,90588,90640,90802,90860,90990,91000,91111,91159,91178,91263,91462,91527,91550,91708,91740,91753,91855,91862,91983,92004,92141,92175,92257,92260,92304,92473,92567,92655,92691,92712,92716,92776,92862,92968,92972,93178,93328,93373,93443,93583,93598,93701,94167,94267,94326,94440,94664,94822,94904,95038,95049,95098,95557,95591,95660,95711,95735,95747,95786,95814,95837,95852,95857,95860,96009,96222,96368,96374,96522,96605,96681,96687,96740,96784,96798,96805,96816,96871,96886,96983,97192,97292,97324,97549,97680,97700,97751,97895,98013,98044,98107,98263,98279,98283,98284,98334,98443,98568,98640,98642,98664,98684,98755,98937,99184,99194,99302,99358,99386,99488,99595,99832",
          "102,149,178,217,460,496,1656,1669,1706,1728,1759,1792,1904,1917,1933,1935,1944,1983,1987,1995,2028,2062,2119,2139,2150,2189,2198,2224,2243,2266,2308,2330,2348,2393,2408,2425,2426,2500,2511,2530,2534,2667,2702,2720,2741,2793,2804,2823,2890,2968,3004,3009,3012,3022,3025,3026,3037,3052,3057,3060,3062,3064,3094,3101,3104,3111,3132,3174,3197,3208,3245,3295,3302,3318,3353,3411,3427,3441,3447,3451,3453,3454,3472,3499,3508,3516,3523,3525,3556,3558,3568,3578,3591,3604,3636,3640,3651,3657,3676,3748,3799,3865,3876,3885,3916,3929,3964,3968,3969,3970,3996,4012,4019,4037,4044,4059,4076,4088,4095,4113,4133,4138,4155,4162,4179,4186,4195,4209,4248,4250,4302,4315,4318,4338,4355,4361,4372,4397,4452,4537,4539,4587,4640,4665,4708,4722,4752,4780,4793,4805,4868,4883,4893,4924,4926,4943,4950,4951,4961,4964,4968,5076,5313,5580,5605,5755,5766,6159,6275,6327,6342,6352,6476,6617,6647,6767,6805,6890,7088,7093,7102,7136,7147,7221,7320,7374,7425,7568,7600,7601,7657,7684,7691,7790,7831,7968,8011,8121,8218,8222,8285,8317,8350,8404,8422,8458,8467,8506,8537,8622,8729,8769,8784,8820,8823,8835,8848,8862,8865,8896,8920,8951,8962,8995,9034,9083,9166,9191,9269,9325,9393,9466,9613,9659,9687,9692,9851,9869,9952,9957,9972,10034,10134,10255,10309,10319,10422,10428,10445,10554,11004,11056,11123,11334,11355,11362,11374,11375,11544,11698,12196,12249,12296,12326,12359,12370,12516,12633,12685,12880,12994,13291,13592,13745,14050,14210,14232,14286,14319,14821,15117,15303,15329,15487,15714,15781,15813,15814,15903,15921,16067,16150,16196,16486,16533,16543,16694,16780,17244,17250,17293,17336,17355,17372,17399,17487,17514,17526,17550,17683,17734,17780,17867,17901,17905,17963,18100,18274,18415,18560,18565,18612,18632,18688,18837,18956,19053,19109,19222,19474,19554,19774,19837,19844,20054,20105,20242,20287,20341,20706,20937,20941,20951,21098,21100,21289,21370,21373,21409,21503,21650,21793,21804,21881,22067,22340,22432,22463,22472,22484,22779,22872,22879,22959,22999,23055,23252,23253,23344,23454,23461,23497,23632,23802,23891,23898,23948,24126,24160,24372,24472,24551,24564,24601,24633,24905,24945,25031,25286,25385,25510,25715,25791,25945,25984,26018,26067,26091,26468,26488,26534,26543,26560,26570,26585,26597,26610,26618,26665,26830,26852,26888,26942,26955,27012,27034,27070,27128,27164,27207,27363,27466,27474,27475,27492,27530,27564,27571,27576,27582,27586,27596,27607,27612,27642,27724,27731,27740,27770,27778,27794,27795,27811,27868,27960,27987,28038,28062,28078,28085,28094,28139,28141,28207,28286,28308,28341,28347,28414,28427,28464,28488,28489,28493,28531,28558,28568,28603,28607,28626,28653,28805,28817,28840,28856,28964,28980,28992,29009,29084,29119,29126,29143,29153,29156,29179,29208,29224,29264,29284,29289,29290,29372,29512,29529,29550,29563,29607,29628,29661,29682,29685,29689,29790,29819,29876,29884,29947,30021,30050,30079,30108,30137,30166,30195,30224,30253,30282,30311,30340,30369,30398,30427,30446,30456,30485,30514,30543,30601,30630,30688,30717,30746,30775,30804,30833,30862,30891,30920,30949,30978,31005,31007,31036,31065,31094,31152,31181,31210,31239,31268,31297,31326,31353,31355,31384,31413,31442,31471,31500,31515,31529,31558,31587,31616,31645,31674,31703,31761,31819,31848,31877,31906,31935,31964,31993,32022,32051,32080,32109,32138,32167,32196,32225,32254,32283,32312,32341,32370,32399,32428,32457,32486,32515,32544,32573,32602,32631,32660,32689,32718,32747,32776,32805,32834,32863,32892,32921,32950,32979,33008,33037,33066,33095,33124,33153,33182,33211,33240,33269,33298,33327,33356,33385,33414,33443,33472,33501,33530,33559,33588,33617,33646,33675,33704,33733,33762,33791,33820,33878,33907,33936,33965,33994,34023,34052,34081,34110,34168,34197,34226,34255,34284,34313,34342,34371,34400,34429,34458,34487,34516,34545,34574,34603,34632,34661,34690,34719,34748,34777,34806,34835,34864,34893,34922,34951,35021,35040,35042,35044,35075,35085,35110,35122,35152,35184,35194,35229,35235,35273,35297,35347,35421,35441,35492,35540,35599,35658,35731,35745,35759,35790,35804,35817,35849,35882,35920,35947,35958,35979,36038,36054,36101,36161,36189,36220,36291,36323,36327,36350,36367,36409,36481,36540,36562,36599,36628,36629,36663,36722,36752,36781,36852,36911,36929,36943,36970,37042,37063,37101,37160,37182,37192,37224,37342,37413,37472,37491,37493,37496,37500,37531,37603,37662,37686,37721,37737,37785,37823,37844,37871,37903,37974,37990,37999,38004,38033,38092,38164,38223,38247,38282,38298,38346,38376,38405,38464,38495,38535,38551,38594,38625,38653,38725,38784,38822,38843,38907,38925,38966,39025,39034,39051,39096,39112,39155,39186,39214,39286,39345,39404,39436,39468,39499,39527,39586,39608,39657,39681,39716,39775,39808,39847,39861,39869,39885,39906,39965,40126,40155,40184,40213,40242,40271,40300,40329,40358,40387,40416,40445,40474,40503,40515,40532,40561,40590,40619,40648,40706,40735,40764,40793,40822,40851,40880,40909,40938,40967,40996,41025,41054,41083,41112,41141,41170,41199,41228,41286,41344,41373,41402,41431,41460,41489,41518,41547,41802,42017,42046,42075,42104,42133,42193,42283,42310,42316,42358,42371,42381,42433,42438,42445,42497,42520,42572,42595,42757,42832,42843,42866,42907,42915,42933,42994,43024,43069,43117,43144,43171,43231,43239,43249,43306,43310,43340,43381,43397,43414,43425,43439,43468,43502,43618,43624,43705,43741,43780,43855,43862,43887,43904,43942,44017,44092,44136,44179,44189,44254,44329,44416,44491,44566,44580,44584,44620,44653,44668,44696,44728,44803,44814,44861,44890,44914,44916,44965,44996,45008,45040,45053,45113,45161,45222,45280,45353,45363,45410,45429,45442,45468,45472,45540,45597,45655,45702,45727,45784,45842,45856,45914,45971,45996,46042,46054,46065,46104,46117,46127,46157,46172,46183,46191,46206,46222,46240,46294,46308,46317,46328,46337,46406,46433,46438,46464,46475,46485,46497,46522,46536,46572,46578,46652,46665,46686,46688,46751,46767,46850,46854,46865,46868,46880,46897,47011,47033,47054,47086,47092,47118,47137,47182,47185,47186,47281,47287,47305,47315,47387,47388,47418,47424,47442,47463,47489,47562,47590,47598,47613,47669,47836,47845,47900,47935,47944,47945,47989,48019,48026,48030,48041,48044,48056,48074,48090,48139,48165,48183,48189,48213,48226,48310,48321,48330,48515,48517,48571,48573,48639,48647,48675,48683,48684,48686,48735,48757,48778,48798,48799,48805,48812,48825,48837,48864,48869,48876,48897,48901,48935,48937,48943,48950,48966,48975,48976,49019,49053,49123,49145,49194,49344,49368,49382,49489,49495,49512,49608,49630,49685,49692,49695,49701,49724,49747,49772,49781,49823,49824,49830,49839,49860,49908,49915,49924,49927,49950,50024,50118,50158,50259,50279,50301,50382,50412,50415,50454,50531,50641,50675,50717,50766,51096,51173,51320,51420,51601,51633,51651,51709,51784,51935,52034,52078,52135,52188,52191,52217,52227,52247,52261,52273,52539,52558,52645,52726,52782,52906,53212,53297,53311,53330,53336,53395,53566,53612,53707,54126,54157,54173,54198,54222,54371,54376,54510,54523,54694,54835,54854,55059,55064,55140,55189,55249,55277,55318,55443,55544,55571,55611,55752,55794,55839,55962,56072,56164,56190,56238,56305,56324,56385,56395,56407,56462,56475,56615,56692,56761,56839,56974,56996,57040,57139,57143,57147,57150,57347,57357,57451,57469,57473,57484,57699,57923,57972,58274,58482,58498,58523,58590,58673,58718,58781,58789,58802,58887,58900,58906,58996,59109,59364,59508,59758,59863,60080,60082,60186,60268,60382,60506,60585,60645,60661,60728,60851,61006,61107,61143,61198,61333,61367,61456,61584,61827,62006,62053,62077,62095,62230,62250,62253,62295,62372,62497,62500,62576,62622,62685,62689,62701,62778,62857,62894,62899,63148,63258,63275,63389,63417,63503,63595,63653,63805,63821,64128,64205,64226,64256,64343,64353,64591,64757,64883,64959,65120,65193,65196,65197,65273,65298,65340,65400,65424,65513,65557,65574,65689,65785,65838,65859,65874,65886,65889,65936,66075,66096,66168,66193,66200,66301,66463,66488,66512,66593,66693,66935,66950,66973,67017,67036,67079,67109,67169,67185,67281,67445,67537,67593,67622,67743,67747,67760,67774,67797,67806,67832,67911,68059,68271,68469,68845,68906,69128,69195,69257,69356,69409,69613,69750,70067,70481,70530,70547,70607,70706,70733,70749,70787,70841,70903,70927,70944,70988,71001,71113,71146,71277,71344,71447,71533,71794,71829,71830,71898,72035,72180,72353,72362,72367,72474,72506,72824,72898,72994,73152,73205,73400,73413,73414,73434,73454,73456,73490,73553,73619,73628,73646,73678,73714,73750,73877,73889,74035,74143,74221,74252,74279,74339,74531,74549,74605,74622,74632,74649,74759,74899,74914,74983,75044,75048,75054,75285,75308,75344,75352,75387,75535,75649,75721,75756,75768,75879,75894,76240,76306,76348,76415,76438,76443,76446,76474,76485,76543,77076,77130,77383,77389,77410,77506,77612,77613,77699,77705,77868,78021,78097,78142,78370,78417,78549,78653,78655,78677,78681,79169,79203,79292,79387,79487,79534,79727,79753,79881,79933,80327,80405,80408,80448,80560,80665,80713,80719,80726,80758,80793,80797,80926,80950,80967,80969,80996,81013,81041,81115,81198,81277,81337,81360,81529,81569,81682,81750,81762,81848,81915,81924,81931,81955,82164,82259,82354,82582,82786,83016,83025,83044,83161,83198,83234,83254,83327,83333,83419,83423,83425,83513,83650,83677,83792,83847,83936,84131,84186,84382,84402,84403,84493,84600,84923,85110,85180,85211,85339,85370,85485,85630,85689,85740,85766,85777,85822,85903,85919,85938,85997,86046,86066,86108,86132,86314,86341,86344,86472,86592,86686,87126,87134,87205,87214,87424,87428,87481,87574,87685,87694,88131,88137,88248,88249,88380,88395,88502,88665,88719,88877,88901,88925,89091,89118,89162,89295,89647,89786,89931,89979,89996,90025,90121,90246,90357,90382,90394,90496,90515,90652,90711,90848,90879,90987,90992,91018,91211,91217,91223,91266,91415,91598,91637,91780,91791,91798,91960,92053,92204,92232,92301,92308,92330,92519,92648,92703,92804,92918,93108,93209,93210,93263,93477,93737,93778,93864,93886,93893,93913,93925,94131,94172,94240,94427,94470,94558,94677,94683,94689,94829,94874,94931,95253,95485,95520,95532,95652,95749,95791,95849,96063,96086,96491,96528,96580,96673,96757,96879,96923,96964,96979,97003,97121,97173,97303,97343,97405,97407,97607,97645,97703,97910,97985,98047,98188,98243,98348,98370,98460,98533,98594,98705,98944,98966,99057,99326,99461,99468,99632,99753,99839,99921",
          "41,60,298,301,308,316,369,380,398,454,481,489,539,546,572,573,596,749,762,823,865,880,1067,1099,1102,1118,1151,1301,1303,1396,1486,1553,1774,1897,2024,2206,3743,4576,4970,5072,5085,5182,5461,5561,5628,5680,5843,5866,5884,5890,5897,6090,6130,6236,6314,6448,6569,6606,6613,6749,6774,6781,6897,6930,6989,6991,7003,7011,7015,7034,7144,7262,7402,7450,7527,7567,7635,7647,7653,7679,7712,7886,7935,7943,8013,8086,8108,8152,8207,8236,8338,8339,8670,8838,8846,8936,8948,9051,9101,9151,9234,9245,9282,9320,9379,9468,9809,9945,9948,10025,10032,10154,10158,10277,10345,10362,10593,10853,10919,10922,11008,11124,11178,11224,11316,11357,11573,11787,11832,11833,11845,11849,11939,11940,11945,12020,12022,12351,12772,13113,13433,14026,14189,14589,14600,14713,14953,15089,15446,16626,16871,17051,17232,17536,17783,18062,18252,18521,18855,18899,18912,19283,20174,20198,20293,20597,20668,20769,20794,20802,20879,20907,21015,21290,21338,21448,21630,21655,21736,21778,22635,22714,22845,23299,23351,23366,23568,23601,23705,24067,25153,25559,25566,25636,26356,26410,26777,26794,26879,26962,27094,27329,27460,28303,28702,28934,29120,29407,29543,29576,29868,29983,35041,35608,35636,36241,36491,37175,37234,37363,37696,37739,37855,37909,38137,38303,38320,38549,38758,39503,42160,42998,43051,43834,45871,45947,50042,50120,50255,50286,50289,50492,50518,50598,50672,50702,50706,50754,50755,50780,50785,50804,50845,51002,51039,51105,51368,51395,51424,51594,52003,52130,52144,52170,52198,52332,52447,52761,52804,53044,53085,53088,53107,53132,53302,53447,53502,53504,53570,53678,53745,53757,53810,53878,54134,54377,54426,54490,54573,54728,54758,54798,54874,55093,55158,55176,55183,55198,55329,55398,55687,55688,55862,55978,55983,56043,56068,56169,56319,56402,56452,56525,56617,56645,56663,56799,56838,56864,56917,57089,57226,57360,57361,57501,57505,57837,57924,57933,58116,58148,58201,58418,58420,58475,58544,58551,58674,58715,58834,58872,58917,58956,59078,59110,59127,59453,59501,59544,59614,59676,59774,59932,59961,60077,60096,60247,60301,60336,60500,60600,60672,60680,60750,60801,61173,61220,61378,61382,61471,61516,61649,61656,61665,61724,61835,61842,61904,62223,62235,62255,62317,62394,62449,62649,62715,62736,62782,62814,62898,62916,63009,63021,63038,63097,63135,63530,63581,63624,63720,63804,63946,64202,64423,64592,64616,64913,64940,65084,65132,65462,65544,65774,65975,66029,66054,66124,66258,66331,66519,66561,66667,66679,66804,66938,66988,67021,67217,67283,67605,67786,67814,67903,67914,68079,68134,68173,68176,68220,68267,68295,68363,68454,68499,68501,68521,68548,68564,68587,68618,68642,68646,68758,68823,68857,68872,68875,68945,68959,68992,69114,69139,69285,69517,69536,69741,69770,70055,70213,70230,70408,70460,70679,70748,70898,71023,71050,71053,71112,71124,71184,71192,71261,71298,71519,71528,71543,71576,71645,71716,71745,71812,71998,72148,72149,72215,72270,72273,72678,72721,72756,72955,73070,73078,73088,73156,73190,73258,73277,73541,73829,74011,74255,74268,74461,74484,74577,74675,74698,74738,74789,74883,74964,74967,74992,75213,75230,75294,75885,75909,75926,75988,76318,76460,76552,76553,76576,76599,76600,76627,76654,76722,76980,77174,77413,77519,77633,77638,77972,78053,78073,78079,78100,78167,78202,78205,78255,78364,78562,78609,78687,78698,78761,78790,79065,79170,79212,79250,79323,79745,79764,79806,79868,80008,80018,80019,80091,80220,80306,80323,80639,80706,80909,81032,81135,81401,81595,81687,81858,81921,81938,82198,82246,82359,82384,82385,82471,82561,82664,82697,82753,82871,82937,83175,83178,83278,83459,83495,83579,83609,83678,83743,83795,83949,83999,84044,84246,84503,84661,84764,84856,84913,85035,85064,85103,85129,85148,85228,85315,85343,85384,85422,85550,85590,85644,85856,85920,86242,86313,86427,86463,86521,86940,86959,87055,87064,87093,87125,87139,87154,87186,87285,87371,87423,87472,87474,87640,87703,87723,87743,87825,87876,88023,88139,88330,88373,88438,88457,88553,88718,88761,88888,88910,88919,89000,89398,89420,89451,89482,89529,89845,89948,89968,90014,90021,90056,90083,90096,90113,90127,90231,90368,90504,90535,90570,90750,90984,91162,91253,91451,91593,91676,91776,91966,91993,92166,92224,92283,92419,92461,92465,92499,92591,92672,92896,93032,93223,93271,93447,93545,93588,93624,93635,93709,93831,93912,93972,94018,94070,94185,94261,94349,94351,94368,94419,94428,94660,94731,94894,94914,95127,95144,95221,95227,95294,95347,95396,95413,95417,95472,95501,95556,95599,95659,95807,95831,95847,95858,95912,95913,95964,96036,96167,96376,96543,96576,96674,96870,97112,97166,97177,97270,97326,97330,97352,97521,97675,97713,97737,97744,97768,97773,97862,98020,98115,98123,98148,98390,98398,98414,98438,98547,98704,98768,98847,98862,98863,98880,98987,98994,99180,99542,99678,99846,99854,99979",
          "52,58,120,135,181,322,708,874,1181,1321,1384,1484,1509,1670,1838,1931,1938,1947,1948,1952,2006,2018,2019,2055,2121,2138,2184,2268,2373,2374,2418,2443,2451,2495,2512,2528,2531,2533,2576,2645,2735,2738,2740,2744,2790,2858,2885,2893,2896,2900,2911,2918,2934,2984,3020,3033,3043,3067,3108,3115,3188,3278,3404,3405,3417,3439,3464,3473,3490,3498,3524,3527,3546,3570,3577,3663,3685,3691,3711,3721,3772,3791,3815,3825,3826,3934,3936,3959,3975,3979,3981,3998,4036,4040,4049,4058,4084,4094,4108,4112,4152,4194,4230,4236,4246,4269,4283,4290,4297,4379,4474,4482,4487,4491,4496,4498,4501,4507,4540,4552,4553,4577,4590,4595,4629,4646,4659,4686,4745,4784,4830,4854,4857,4928,4940,5152,5289,5402,5493,5585,5594,5623,5833,5848,5873,5960,5983,5986,6066,6085,6092,6178,6196,6208,6261,6291,6300,6325,6347,6375,6416,6419,6467,6501,6567,6639,6733,6761,6898,6948,6984,7029,7059,7076,7110,7129,7177,7213,7225,7246,7251,7298,7303,7409,7442,7460,7465,7632,7677,7716,7786,7792,7840,7938,7997,8015,8045,8157,8265,8431,8438,8476,8533,8557,8564,8574,8591,8761,8764,8776,8817,8861,9044,9079,9094,9193,9255,9534,9635,9673,9727,9818,9894,9969,10109,10229,10259,10270,10390,10451,10560,11168,11219,11260,11305,11453,11484,11702,11925,11958,12000,12190,12342,12357,12656,13183,13314,13691,13732,13804,13863,14032,14070,14207,14287,14461,14510,14572,14822,14852,14855,15008,15056,15328,15342,15654,15816,16125,16339,16370,16416,16504,16682,16836,16892,16936,17061,17083,17138,17177,17278,17420,17442,17471,17481,17577,17641,17844,17877,17888,18133,18292,18321,18375,18575,18667,18973,19018,19069,19603,19654,19657,19741,19754,19813,19874,19907,20024,20367,20433,20522,20651,20758,20990,21086,21427,21501,21506,21784,21870,21889,21968,21985,22260,22280,22361,22380,22468,22525,23296,23323,23421,23919,23963,24147,24198,24488,24775,25239,25258,25400,25424,25503,25576,25725,26211,26352,26487,26498,26507,26533,26549,26601,26606,26607,26655,26738,26767,26771,26778,26831,26844,26992,27006,27011,27022,27036,27044,27097,27103,27130,27142,27151,27180,27230,27263,27331,27399,27446,27507,27512,27515,27533,27562,27590,27613,27614,27624,27632,27640,27697,27726,27755,27769,27853,27887,27924,27962,27963,27972,27997,28042,28057,28072,28098,28116,28151,28209,28222,28223,28236,28254,28255,28260,28301,28314,28343,28344,28407,28428,28438,28495,28498,28506,28513,28539,28564,28586,28595,28599,28629,28634,28636,28645,28663,28665,28722,28783,28797,28957,28968,29015,29083,29105,29116,29146,29166,29184,29196,29217,29218,29237,29238,29313,29336,29387,29448,29470,29507,29520,29536,29546,29557,29584,29587,29644,29645,29658,29659,29691,29695,29706,29726,29740,29754,29762,29764,29887,29895,29920,30018,30047,30076,30105,30134,30163,30192,30221,30250,30279,30308,30337,30366,30395,30424,30453,30482,30540,30569,30598,30627,30656,30685,30714,30743,30772,30801,30830,30859,30888,30917,30975,31004,31033,31091,31120,31149,31178,31207,31236,31265,31294,31323,31352,31381,31410,31439,31468,31497,31526,31555,31584,31613,31642,31671,31700,31729,31758,31787,31816,31845,31874,31903,31932,31961,31990,32048,32077,32135,32164,32193,32222,32251,32280,32309,32338,32396,32425,32454,32483,32512,32541,32570,32599,32628,32657,32686,32715,32744,32773,32802,32831,32860,32889,32918,32947,32976,33005,33034,33063,33092,33121,33150,33179,33208,33237,33266,33295,33324,33353,33382,33411,33440,33469,33498,33527,33556,33585,33614,33643,33672,33701,33730,33759,33788,33817,33846,33875,33904,33933,33962,33991,34020,34049,34078,34107,34136,34165,34194,34223,34252,34281,34310,34339,34368,34397,34426,34455,34484,34513,34542,34571,34600,34629,34658,34687,34716,34745,34774,34803,34861,34890,34919,34948,34977,35018,35032,35051,35070,35107,35138,35181,35230,35270,35344,35363,35418,35468,35489,35503,35537,35557,35560,35596,35655,35710,35728,35787,35809,35846,35917,35976,36035,36098,36158,36217,36238,36288,36311,36347,36406,36433,36438,36478,36537,36596,36660,36719,36740,36778,36849,36908,36967,37039,37098,37157,37169,37221,37255,37280,37339,37360,37410,37428,37469,37489,37528,37556,37600,37659,37718,37782,37841,37900,37971,38030,38060,38089,38161,38220,38279,38343,38373,38402,38461,38532,38591,38650,38686,38722,38781,38840,38904,38963,39022,39042,39060,39093,39152,39211,39283,39342,39401,39465,39524,39583,39654,39713,39754,39772,39844,39903,39962,40115,40144,40173,40202,40231,40260,40289,40318,40347,40376,40405,40434,40463,40492,40521,40550,40579,40608,40637,40666,40695,40724,40753,40782,40811,40840,40869,40898,40927,40956,40985,41014,41043,41072,41101,41130,41159,41188,41246,41275,41304,41333,41362,41391,41420,41449,41536,42006,42035,42064,42093,42122,42138,42153,42161,42182,42216,42241,42272,42347,42382,42399,42422,42469,42509,42558,42584,42626,42659,42682,42692,42707,42712,42746,42821,42896,42923,42983,43012,43015,43058,43133,43160,43165,43220,43295,43316,43332,43337,43370,43396,43472,43532,43570,43648,43694,43750,43769,43844,43878,43886,43931,43991,44006,44049,44063,44081,44133,44168,44223,44227,44243,44318,44352,44405,44480,44518,44555,44595,44613,44642,44717,44758,44792,44879,44954,45002,45029,45102,45150,45193,45211,45269,45342,45399,45436,45457,45492,45512,45529,45552,45586,45644,45668,45716,45740,45756,45773,45831,45852,45903,45960,46122,46192,46279,46293,46358,46454,46471,46541,46603,46709,46737,46804,46815,46860,46918,46923,46928,46943,46957,46960,47003,47043,47062,47073,47208,47231,47286,47289,47290,47294,47404,47407,47482,47632,47646,47672,47677,47686,47689,47741,47745,47753,47768,47777,47784,47823,47826,47833,47898,47918,47988,48048,48098,48122,48127,48190,48202,48237,48270,48284,48295,48301,48307,48341,48345,48354,48375,48415,48421,48464,48480,48494,48496,48508,48619,48633,48650,48849,48874,48918,48944,49158,49184,49203,49221,49306,49318,49352,49376,49410,49447,49479,49533,49561,49573,49579,49611,49704,49738,49751,49774,49806,49822,49840,49845,49866,49876,49898,49920,49938,49970,49980,50039,50141,50227,50464,50522,50750,50810,50841,50919,50963,51244,51300,51448,51498,51578,51602,51690,51724,51974,52136,52189,52195,52351,52426,52442,52679,52685,52725,52753,52788,52792,52801,52970,53091,53293,53362,53363,53515,53663,53666,53850,53888,53895,53988,54030,54055,54075,54177,54208,54225,54282,54706,54774,54839,54932,54935,54997,55066,55087,55231,55299,55348,55428,55462,55627,55734,55833,55854,55891,55948,55997,56069,56100,56207,56268,56312,56463,56477,56516,56545,56638,56643,56667,56694,56714,56921,56930,57023,57135,57137,57186,57258,57318,57622,57654,57701,57772,57816,57819,57875,57970,58065,58089,58351,58529,58561,58636,58734,58806,59098,59119,59186,59212,59220,59254,59363,59378,59492,59542,59598,59642,59643,59664,59670,59688,59764,59867,59916,59947,59972,60113,60154,60158,60313,60392,60561,60578,60592,60606,60782,60797,60911,60957,61031,61113,61398,61546,61568,61574,61589,61633,61787,61808,61810,61838,61848,61920,62114,62147,62222,62331,62404,62412,62596,62702,62844,62845,62951,62989,63032,63043,63065,63199,63240,63251,63286,63738,63776,63807,64011,64081,64120,64201,64266,64341,64563,64647,64742,64748,64792,64800,64878,64895,64954,64971,64976,65043,65051,65357,65386,65466,65532,65567,65639,65660,65756,65771,65878,65984,66131,66151,66344,66349,66543,66584,66626,66730,66803,66834,66840,66883,67097,67195,67201,67207,67245,67282,67300,67366,67763,67890,68031,68160,68191,68202,68257,68391,68593,68662,68673,68884,68910,69038,69271,69299,69371,69466,69515,69548,69722,69791,69879,69913,70018,70036,70039,70112,70141,70220,70307,70375,70427,70492,70562,70669,70809,70873,70907,71159,71302,71402,71548,71726,71853,71968,72139,72294,72344,72375,72408,72469,72528,72538,72548,72583,72599,72635,72639,72726,72894,73131,73276,73374,73479,73500,73521,73644,73737,73810,73822,73883,73903,73915,74299,74419,74537,74922,74998,75120,75130,75203,75276,75304,75347,75364,75384,75405,75442,75445,75481,75494,75530,75659,75682,75875,76091,76095,76172,76265,76268,76329,76357,76445,76498,76540,76544,76791,76800,76810,76811,76839,76907,76989,77249,77251,77381,77393,77514,77728,77748,77801,77827,77848,77881,77899,78196,78231,78399,78449,78613,78733,78794,78802,78868,78883,78891,78925,78999,79237,79294,79452,79509,79636,79653,79736,79850,79851,79859,79919,79996,80338,80350,80374,80381,80389,80406,80565,80636,80740,80911,80991,80999,81076,81124,81133,81169,81280,81385,81496,81585,81705,81722,81723,81752,81861,82220,82309,82373,82469,82524,82621,82755,82830,82895,83037,83188,83370,83652,83753,83782,83828,83843,83845,83867,83942,84018,84072,84179,84378,84660,84682,84739,84802,85245,85350,85358,85534,85727,85848,85871,85886,86002,86014,86032,86085,86110,86320,86408,86441,86606,86610,86637,86640,86895,86988,87082,87095,87118,87147,87150,87224,87357,87445,87558,87667,87813,87921,87930,88032,88054,88284,88300,88324,88429,88456,88483,88485,88556,88670,88805,88955,88986,89057,89065,89145,89260,89270,89421,89509,89542,89785,89828,89885,89934,89970,90054,90100,90150,90185,90221,90450,90500,90719,90948,91061,91110,91115,91123,91319,91403,91411,91558,91619,91714,91720,91761,91875,91935,91939,91997,92045,92119,92452,92563,92579,92588,92693,92753,92782,92783,92793,92812,92883,92923,92949,93000,93148,93186,93253,93277,93426,93432,93585,93615,93631,93675,93728,93899,94146,94229,94290,94299,94409,94607,94621,94735,94917,94933,95001,95007,95095,95159,95323,95377,95442,95494,95524,95573,95697,95709,95714,95725,95940,96042,96078,96172,96196,96203,96219,96297,96438,96598,96648,96787,96847,96986,97000,97055,97120,97123,97197,97335,97400,97425,97437,97469,97490,97584,97624,97873,97915,98070,98146,98180,98496,98514,98523,98608,98795,98811,98923,99011,99055,99118,99335,99486,99617,99719,99808,99847,99892,99967",
          "37,59,94,164,187,1629,1709,1785,1975,2013,2125,2132,2233,2380,2465,2632,2637,2996,3035,3216,3233,3361,3371,3422,3616,3727,3754,3833,4027,4046,4096,4134,4160,4216,4335,4368,4376,4782,4882,4979,4995,4996,6248,6981,8047,9442,9528,9537,9556,9632,9651,9670,9689,9708,9763,9776,9993,10012,10069,10107,10601,10620,10753,10829,10924,11095,11133,11152,11209,11228,11247,11304,11361,11370,11478,11494,11513,11532,11627,11630,11646,11741,11854,11855,11874,11884,11912,11950,11951,11969,11983,12026,12042,12045,12062,12102,12105,12140,12159,12254,12330,12368,12427,12444,12463,12474,12501,12534,12558,12577,12615,12653,12672,12681,12786,12843,12900,12938,12969,12995,13014,13090,13242,13261,13280,13356,13388,13432,13506,13527,13546,13565,13603,13625,13641,13660,13679,13698,13717,13755,13774,13793,13812,13850,13907,13926,13945,13986,14021,14059,14116,14173,14192,14211,14230,14249,14268,14382,14439,14553,14569,14629,14743,14800,14838,14895,14914,14971,14992,15009,15047,15123,15139,15161,15184,15218,15237,15275,15313,15332,15427,15465,15484,15498,15560,15598,15636,15712,15731,15750,15788,15807,15826,15845,15864,15883,15959,15997,16016,16035,16054,16073,16168,16187,16206,16301,16320,16358,16377,16415,16453,16472,16503,16529,16548,16605,16681,16708,16731,16736,16738,16757,16795,16814,16833,16835,16890,16928,16946,17004,17042,17080,17137,17156,17169,17175,17194,17270,17308,17327,17365,17441,17460,17479,17498,17539,17555,17593,17650,17726,17745,17802,17821,17840,17859,17878,17897,18011,18023,18030,18087,18106,18125,18182,18220,18239,18247,18258,18296,18315,18334,18353,18372,18391,18410,18429,18448,18467,18470,18505,18515,18581,18600,18676,18752,18828,18847,18885,18914,18923,18958,18961,18980,19037,19075,19094,19132,19151,19170,19207,19208,19227,19265,19284,19307,19341,19360,19379,19398,19436,19455,19463,19484,19550,19560,19569,19645,19664,19702,19721,19759,19797,19835,19854,19873,19892,19911,19930,19949,19978,19987,20006,20025,20042,20063,20082,20139,20158,20166,20177,20215,20234,20291,20310,20329,20348,20386,20405,20424,20443,20462,20500,20557,20595,20614,20652,20671,20728,20747,20766,20804,20823,20880,20918,20950,20956,21013,21051,21073,21146,21165,21184,21199,21222,21241,21291,21317,21336,21374,21393,21412,21431,21450,21469,21488,21545,21564,21602,21613,21621,21659,21707,21716,21735,21754,21811,21868,21906,21925,21963,21982,22000,22001,22020,22077,22128,22134,22172,22210,22229,22237,22267,22305,22343,22362,22381,22400,22438,22457,22476,22495,22570,22571,22609,22666,22685,22704,22715,22742,22761,22780,22856,22894,22913,22933,22989,23008,23046,23084,23103,23160,23198,23208,23217,23236,23266,23312,23350,23388,23426,23445,23464,23502,23521,23540,23578,23616,23635,23654,23673,23675,23711,23730,23768,23825,23844,23882,23920,23935,23939,23958,23977,23996,24015,24034,24053,24091,24110,24129,24167,24186,24224,24243,24281,24319,24395,24414,24490,24528,24547,24566,24604,24623,24642,24696,24699,24718,24737,24756,24813,24815,24832,24851,24870,24908,24928,25003,25016,25022,25041,25098,25117,25177,25212,25231,25269,25288,25307,25364,25383,25421,25440,25482,25497,25500,25516,25535,25554,25630,25649,25685,25687,25706,25744,25763,25839,25854,25858,25877,25896,25915,25934,25953,25991,26010,26048,26086,26124,26160,26175,26181,26200,26238,26257,26421,26447,26466,26485,26504,26523,26656,26713,26731,26732,26808,26826,26865,26868,26979,27112,27121,27150,27188,27199,27238,27275,27289,27321,27333,27369,27376,27462,27483,27520,27568,27673,27682,27691,27722,27796,27814,27929,27948,28024,28112,28138,28157,28256,28356,28385,28518,28575,28670,28746,28781,28879,28930,28948,28955,28960,29069,29079,29088,29135,29145,29331,29337,29365,29373,29403,29483,29592,29639,29704,29715,29753,29904,29919,29939,29962,29982,30028,30057,30086,30115,30144,30173,30202,30231,30260,30289,30318,30347,30376,30405,30434,30463,30492,30550,30608,30637,30666,30695,30724,30753,30811,30840,30869,30898,30927,30956,30985,31014,31043,31072,31101,31130,31159,31188,31217,31246,31275,31304,31333,31362,31391,31420,31449,31478,31507,31536,31565,31594,31623,31652,31681,31710,31739,31768,31797,31826,31855,31884,31913,31942,31971,32000,32029,32087,32116,32145,32174,32203,32232,32261,32290,32319,32348,32377,32406,32435,32464,32493,32522,32551,32580,32609,32638,32667,32696,32725,32754,32783,32812,32841,32870,32899,32928,32957,32986,33015,33044,33073,33102,33131,33160,33189,33218,33247,33276,33305,33334,33363,33392,33421,33450,33479,33508,33537,33566,33595,33624,33653,33682,33711,33740,33769,33798,33827,33856,33885,33914,33943,33972,34001,34030,34059,34088,34117,34146,34175,34204,34233,34262,34291,34320,34349,34378,34407,34436,34494,34523,34552,34581,34610,34639,34668,34697,34726,34755,34784,34813,34842,34871,34900,34929,34958,34987,35028,35071,35078,35117,35191,35280,35354,35428,35444,35499,35547,35606,35631,35665,35797,35828,35856,35883,35927,35986,36045,36108,36123,36168,36227,36237,36298,36357,36416,36488,36547,36606,36614,36670,36683,36729,36788,36797,36805,36859,36918,36977,36991,37049,37108,37119,37167,37231,37290,37349,37376,37420,37440,37445,37479,37492,37538,37610,37669,37728,37792,37851,37910,37932,37981,38040,38099,38107,38171,38230,38289,38309,38353,38412,38471,38489,38498,38542,38545,38601,38615,38660,38732,38752,38791,38850,38865,38914,38926,38973,38987,39032,39103,39105,39162,39221,39293,39352,39383,39411,39475,39534,39593,39664,39723,39748,39782,39854,39913,39928,39972,40122,40151,40180,40209,40238,40267,40296,40325,40354,40383,40412,40441,40470,40499,40528,40557,40586,40615,40644,40673,40702,40731,40760,40789,40818,40847,40876,40905,40934,40963,40992,41021,41050,41079,41108,41137,41166,41195,41224,41253,41311,41340,41369,41398,41427,41456,41485,41514,41543,42013,42042,42071,42100,42129,42145,42158,42189,42206,42211,42226,42279,42287,42317,42354,42368,42397,42401,42429,42465,42516,42591,42666,42753,42828,42858,42874,42903,42924,42990,43028,43065,43140,43191,43227,43302,43346,43349,43377,43399,43416,43438,43464,43517,43539,43587,43614,43672,43701,43743,43776,43802,43819,43821,43851,43879,43938,43957,44013,44044,44088,44151,44156,44175,44210,44250,44274,44285,44289,44325,44412,44487,44506,44562,44626,44649,44724,44799,44867,44886,44910,44961,44987,44991,44998,45036,45068,45109,45157,45218,45276,45296,45349,45377,45406,45464,45515,45536,45593,45651,45723,45780,45838,45910,45930,45967,46015,46045,46047,46074,46099,46100,46101,46130,46158,46176,46204,46239,46311,46335,46351,46379,46391,46403,46428,46445,46450,46515,46538,46558,46629,46655,46699,46705,46713,46750,46752,46766,46768,46778,46901,46940,46977,46985,47007,47052,47067,47205,47206,47214,47243,47261,47268,47376,47380,47382,47414,47488,47497,47500,47520,47542,47545,47565,47571,47601,47665,47691,47747,47803,47847,47866,47887,47902,47946,48104,48111,48114,48121,48145,48174,48199,48206,48212,48214,48247,48249,48282,48286,48311,48376,48386,48410,48431,48497,48530,48536,48575,48587,48662,48667,48762,48767,48800,48806,48896,48909,48979,48993,49022,49069,49081,49082,49108,49147,49171,49204,49249,49265,49269,49276,49284,49303,49317,49338,49341,49379,49406,49452,49453,49552,49584,49642,49660,49670,49674,49694,49697,49753,49792,49793,49834,49858,49897,49909,49935,49956,50007,50084,50122,50166,50360,50361,50375,50509,50680,50695,50753,50869,50890,50967,51106,51274,51293,51321,51401,51473,51521,51590,51643,51982,52246,52398,52565,52623,52821,52835,52868,52873,52882,52975,53147,53196,53248,53255,53277,53308,53364,53584,53614,53633,53646,53657,53688,53738,53748,53831,53881,54002,54155,54274,54452,54476,54564,54588,54613,54704,54726,54893,54960,55091,55115,55123,55188,55255,55417,55432,55472,55512,55531,55607,55738,55766,55776,55792,55843,55897,55939,55968,56121,56176,56303,56381,56556,56568,56625,56710,56713,56734,56818,56863,56880,56986,57103,57120,57140,57158,57219,57365,57495,57511,57558,57683,57728,57790,57814,57817,57846,57917,57930,57946,57958,57993,58021,58097,58161,58180,58272,58383,58467,58655,58660,58693,58700,58720,58727,59011,59283,59369,59383,59419,59433,59487,59679,59712,59820,59839,59943,59948,59950,59993,60038,60175,60267,60283,60700,60709,60726,60751,60758,60858,60861,60961,60976,60995,61118,61176,61194,61200,61212,61295,61336,61363,61389,61436,61473,61611,61618,61646,61781,61903,61937,61960,61981,61993,62033,62116,62149,62606,62781,62835,62865,63004,63280,63325,63547,63594,63597,63648,63675,63727,63765,63786,63924,64021,64059,64064,64086,64179,64195,64247,64378,64475,64541,64713,64858,65033,65056,65157,65345,65492,65550,65623,65714,65813,65840,65942,66058,66105,66170,66326,66332,66400,66435,66614,66617,66724,66742,66892,67054,67106,67156,67353,67529,67608,67715,67753,67769,67919,67926,68144,68159,68237,68429,68481,68486,68566,68571,68836,68918,68969,69241,69265,69341,69347,69467,69524,69572,69646,69648,69752,69801,69945,70034,70037,70401,70540,70550,70774,70808,70811,70906,70919,71088,71099,71223,71312,71382,71639,71703,71753,71799,71824,71887,72009,72030,72040,72113,72245,72258,72261,72301,72365,72393,72437,72465,72576,72641,72761,72840,72899,72931,72944,72946,72977,73050,73101,73154,73204,73250,73421,73558,73577,73747,73772,73835,73857,73988,74013,74049,74134,74277,74363,74369,74386,74477,74571,74604,74637,74718,74732,74743,74806,74819,74822,74885,75064,75074,75113,75119,75283,75372,75685,75690,75757,75895,75978,76041,76060,76076,76164,76288,76364,76370,76692,76712,76717,76808,76829,76886,76950,77017,77082,77144,77313,77333,77369,77380,77392,77710,77712,77716,77781,77875,77888,78139,78160,78175,78176,78198,78252,78324,78355,78444,78508,78639,78670,78772,78862,78864,78867,78931,79125,79145,79368,79398,79441,79463,79477,79480,79768,79931,79958,80040,80045,80112,80133,80159,80182,80281,80351,80370,80480,80515,80647,80723,80921,80923,81087,81285,81352,81419,81421,81491,81547,81652,81699,81772,81781,81932,81933,81946,81973,82104,82118,82121,82424,82443,82627,82632,82750,82840,82844,82922,82986,83021,83288,83294,83320,83387,83434,83730,84037,84046,84284,84326,84416,84561,84563,84733,84743,84774,84926,85065,85138,85171,85227,85277,85767,85809,85946,86101,86227,86321,86449,86659,86730,86791,86865,86891,86892,87120,87172,87268,87410,87448,87560,87658,87698,87727,87801,87804,87849,87904,88029,88081,88186,88232,88314,88344,88436,88514,88636,88673,88683,88749,88795,88822,88879,88969,89092,89099,89138,89146,89288,89545,89634,89639,89768,89904,89947,90015,90086,90087,90163,90226,90263,90330,90591,90756,90763,90866,90883,90888,90959,91098,91189,91209,91282,91348,91382,91654,91667,91675,91755,91903,92057,92089,92106,92155,92165,92259,92314,92321,92331,92492,92544,92610,92674,92700,92745,92858,93110,93171,93354,93387,93430,93451,93536,93554,93569,93600,93730,93739,93759,93927,94005,94234,94238,94333,94513,94522,94626,94742,94888,95061,95306,95325,95487,95626,95736,95737,95780,95813,95825,95856,95941,96014,96057,96119,96171,96174,96293,96388,96399,96460,96484,96524,96625,96894,96973,96988,97044,97062,97200,97240,97342,97364,97367,97460,97496,97859,97907,97979,98186,98271,98347,98681,98682,98695,98825,98864,98945,99031,99185,99263,99311,99473,99606,99703,99737",
          "16,253,278,367,371,402,426,455,490,492,532,582,691,697,699,744,756,808,852,886,910,939,945,962,1008,1028,1044,1061,1090,1186,1213,1215,1217,1242,1339,1372,1425,1451,1545,1546,1566,1600,1881,1918,2681,4064,5028,5048,5119,5505,5579,5607,5725,5736,5789,5812,5834,5871,6146,6424,6466,6493,6499,6525,6641,6642,6644,6802,6836,6894,6929,7105,7188,7200,7217,7271,7331,7493,7505,7530,7578,7630,7636,7742,7828,7965,8034,8044,8091,8159,8173,8220,8352,8410,8417,8443,8474,8578,8646,8717,8809,8834,8952,8979,9403,9419,9783,9973,10121,10157,10161,10260,10299,10308,10371,10456,10503,10992,11061,11248,11738,11814,11927,12071,12448,13259,13505,13883,14043,14910,15656,15724,15916,16756,17036,17725,19052,19070,19102,19439,20593,20919,20975,20989,21197,21490,21830,22585,24828,25670,25766,25976,26402,26804,27132,27145,27391,27733,27771,27774,27944,28381,28456,28460,29141,29351,29642,29927,29956,30468,32866,34515,36492,37259,37795,37935,38508,39916,40199,42250,43617,43940,44593,45041,45301,45703,47327,50047,50163,50170,50225,50501,50714,50747,50769,50874,50903,50913,50935,50951,50980,51042,51083,51098,51193,51233,51266,51364,51385,51534,52005,52050,52163,52229,52373,52430,52444,52604,52634,53035,53036,53096,53204,53339,53359,53433,53434,53470,53482,53779,53890,53894,53947,54326,54329,54394,54403,54848,55384,55465,55574,55605,55666,55681,55725,55804,55874,55959,56106,56123,56130,56133,56245,56264,56423,56460,56532,56544,56695,56935,57008,57028,57098,57166,57384,57457,57583,57756,57787,57858,57877,57882,58091,58101,58131,58186,58271,58321,58373,58380,58495,58656,58724,58742,58769,58805,58827,58909,58921,59142,59169,59189,59278,59480,59539,59590,59596,59660,59665,59757,59877,59883,59893,59894,59963,59965,60016,60039,60167,60202,60240,60255,60414,60447,60466,60478,60528,60651,60697,60715,60747,60778,60860,60972,61217,61257,61265,61292,61372,61432,61457,61512,61648,61719,61789,61940,61945,62207,62264,62311,62470,62660,62739,62800,62851,62869,62950,62960,62990,63013,63160,63263,63273,63481,63544,63563,63598,63789,63814,63982,64057,64415,64507,64512,64514,64731,65005,65301,65306,65343,65418,65477,65500,65552,65556,65609,65624,65664,65718,65755,65860,65884,65990,66023,66046,66123,66565,66825,67057,67144,67240,67403,67506,67553,67554,67585,67674,67729,67741,67825,67861,68009,68124,68168,68448,68453,68619,68790,69027,69084,69247,69370,69462,69507,69534,69734,69810,69813,69835,69951,69968,70111,70200,70238,70278,70341,70396,70499,70514,70609,70631,70692,70736,70783,70874,70938,71080,71139,71210,71233,71280,71450,71517,71588,71652,71661,71718,72042,72066,72209,72243,72281,72369,72373,72395,72397,72502,72593,72605,72618,72796,72810,72842,72877,72967,73161,73164,73177,73195,73268,73311,73342,73384,73389,73501,73629,73705,73744,73748,73776,74036,74131,74146,74185,74196,74295,74318,74341,74602,74635,74769,75072,75097,75176,75178,75214,75463,75479,75518,75627,75666,75694,75805,75949,76079,76116,76120,76124,76149,76258,76310,76418,76427,76468,76501,76519,76534,76579,76612,76649,76656,76705,76710,76784,76799,76919,77015,77091,77097,77182,77454,77508,77533,77545,77610,77628,77792,77880,77940,77993,78010,78020,78083,78133,78233,78491,78502,78569,78595,78660,78690,78747,78748,78897,78914,79025,79130,79187,79190,79241,79288,79449,79505,79571,79717,79743,79782,79820,79827,80103,80183,80185,80191,80264,80378,80452,80679,80828,80843,80854,81031,81110,81400,81474,81587,81603,81613,81677,81790,81791,81825,82010,82080,82223,82231,82236,82278,82344,82454,82563,82658,82665,82791,82806,82824,82897,82978,83131,83143,83223,83245,83307,83347,83558,83571,83610,83688,83894,84001,84016,84035,84190,84332,84547,84574,84677,84680,84696,84746,84956,85049,85074,85136,85221,85291,85318,85338,85492,85826,85844,86001,86017,86081,86216,86286,86530,86540,86770,86827,86931,86966,87052,87202,87208,87247,87366,87430,87465,87602,87635,87663,87757,87811,87942,88015,88046,88065,88144,88154,88310,88336,88406,88443,88498,88503,88763,88856,88921,89365,89460,89514,89544,89561,89780,89872,89888,89972,90034,90040,90046,90116,90139,90302,90309,90349,90367,90406,90407,90474,90565,90604,90665,90757,90789,90863,90918,91042,91127,91181,91258,91396,91459,91508,91514,91537,91698,91717,91744,91789,92092,92144,92258,92518,92561,92628,92639,92746,92824,92844,93227,93309,93400,93453,93462,93700,94044,94076,94152,94416,94571,94705,94717,94726,94751,94755,94839,94865,95023,95060,95063,95179,95234,95427,95449,95510,95515,95536,95594,95612,95674,95842,95887,95959,96017,96061,96199,96308,96425,96617,96710,96745,96748,96779,96780,96908,96936,96990,97002,97034,97037,97046,97092,97116,97199,97222,97237,97308,97422,97466,97474,97486,97663,97694,97777,97858,97861,97897,98006,98320,98375,98382,98422,98576,98649,98696,98803,98830,98832,98860,98943,98969,99015,99133,99217,99276,99279,99305,99351,99390,99480,99578,99760,99946",
          "8,11,280,300,336,347,440,449,457,471,547,555,594,603,609,611,639,710,810,824,863,934,955,969,988,1000,1015,1023,1127,1254,1265,1312,1362,1369,1449,1479,1500,1505,1512,1558,1561,1573,1631,2265,2548,3363,3369,3598,3862,4562,4885,4939,4977,5002,5074,5139,5183,5209,5429,5451,5507,5557,5563,5587,5702,5744,5883,5914,5968,6008,6048,6050,6094,6095,6128,6190,6209,6495,6637,6703,6708,6713,6721,6728,6794,6886,6887,6913,6924,7115,7122,7290,7307,7429,7511,7682,7794,7817,7932,7934,8076,8077,8179,8268,8342,8418,8433,8521,8579,8714,8827,8884,8901,8982,8998,9004,9148,9207,9311,9400,9416,9996,10124,10148,10202,10478,10794,10886,10911,11030,11174,11201,11202,11722,11743,11799,11818,15683,17973,19246,19335,21072,23458,23803,25696,25841,25900,26090,26978,27058,27378,27810,27859,28208,28346,28778,28853,29266,29473,31258,33499,34021,35016,35826,38356,38482,40111,41575,41606,41766,41946,42492,42564,43543,43646,45059,45198,45256,47115,49165,49362,49486,50130,50150,50176,50244,50321,50527,50576,50607,50914,50933,51122,51397,51451,51457,51733,51739,51792,51826,51855,51860,51874,52004,52038,52107,52274,52279,52310,52320,52754,52769,52994,53126,53300,53385,53519,53555,53565,53645,53650,53728,53760,53806,53822,53887,53943,53945,54054,54250,54267,54308,54372,54459,54650,54847,54973,55270,55453,55499,55506,55587,55601,55899,55942,56029,56062,56220,56250,56316,56345,56446,56521,56599,56804,56962,57046,57090,57231,57315,57355,57431,57487,57498,57518,57599,57623,57742,57753,57755,57903,58037,58240,58267,58291,58454,58637,58778,59007,59122,59185,59232,59348,59401,59455,59517,59708,59719,59824,60112,60298,60402,60434,60452,60612,60671,60685,60716,60717,60738,60855,60895,60925,60940,61028,61210,61244,61270,61277,61350,61399,61413,61455,61462,61503,61543,61680,61721,61806,61840,61871,61895,61956,62048,62090,62097,62107,62132,62155,62190,62294,62571,62671,62730,62788,62862,62917,63071,63086,63099,63112,63134,63172,63411,63468,63486,63509,63524,63587,63618,63721,63728,63766,63810,63854,64012,64020,64038,64232,64368,64384,64432,64435,64458,64738,64787,64918,65126,65218,65309,65368,65464,65507,65519,65526,65579,65652,65996,66014,66104,66148,66175,66350,66393,66437,66651,66721,66741,66754,66807,66849,66900,66981,66994,67038,67077,67105,67159,67213,67271,67309,67462,67494,67595,67687,67700,67739,67750,67969,67971,68076,68093,68114,68268,68370,68446,68632,68691,68699,68705,68765,69051,69070,69159,69167,69196,69308,69319,69353,69365,69402,69412,69429,69503,69529,69562,69628,69669,69718,69777,69814,69832,69900,69901,69937,70035,70075,70094,70096,70206,70232,70271,70447,70591,70794,70816,70847,70933,71102,71182,71297,71319,71381,71526,71540,71557,71608,71627,71658,71797,71808,71857,71860,71921,71944,71954,72234,72265,72306,72411,72637,72646,72687,72706,72708,72752,72951,72986,73028,73137,73223,73335,73457,73546,73613,73710,73864,73868,73963,73974,74014,74037,74089,74169,74192,74250,74465,74469,74544,74570,74607,74642,74705,74971,75052,75057,75215,75222,75284,75373,75397,75461,75477,75488,75561,75585,75588,75726,75906,75977,76067,76102,76246,76279,76469,76594,76642,76723,76725,76877,76961,77005,77013,77063,77123,77136,77294,77501,77547,77571,77642,77925,77959,77988,78013,78127,78155,78168,78187,78354,78375,78416,78418,78433,78478,78528,78611,78707,78822,78832,78848,78860,79040,79289,79348,79379,79467,79515,79560,79605,79783,79837,80167,80212,80290,80297,80298,80380,80420,80527,80609,80798,80925,80994,81054,81060,81069,81077,81098,81319,81379,81416,81464,81515,81571,81626,81703,81734,81800,81856,81876,81949,81951,82039,82100,82148,82204,82230,82256,82509,82543,82565,82573,82645,82702,82778,82798,82901,82936,82955,82994,83004,83032,83035,83050,83247,83258,83337,83360,83397,83680,83733,83734,83737,83919,83937,84022,84245,84281,84386,84414,84437,84482,84566,84646,84741,84822,84872,84889,84893,84978,85025,85051,85160,85235,85403,85627,85651,85829,85891,85967,86238,86345,86384,86454,86646,86826,86913,87020,87105,87409,87592,87594,87872,87950,88118,88146,88160,88244,88293,88462,88737,88776,88854,89002,89120,89147,89229,89259,89372,89434,89588,89775,89951,90009,90035,90082,90097,90170,90188,90205,90257,90266,90294,90373,90439,90533,90539,90540,90589,90662,90676,90717,90724,90761,90805,90815,90845,90913,91082,91126,91499,91615,91770,91779,91813,91831,91917,92000,92235,92385,92386,92490,92609,92632,92702,92719,92791,93003,93012,93081,93088,93155,93258,93512,93667,93718,93957,93976,93981,93988,94186,94200,94212,94218,94281,94324,94344,94398,94668,94720,95107,95232,95255,95447,95603,95640,95645,95734,95744,95771,95815,95925,95993,96005,96056,96198,96331,96364,96427,96465,96672,97069,97134,97191,97282,97489,97491,97684,97762,97877,97991,97994,98184,98301,98307,98458,98466,98509,98579,98714,98790,98828,98836,98874,98898,99288,99298,99385,99700,99724,99986",
          "3,258,333,352,355,361,363,385,393,412,458,467,494,562,566,579,686,725,779,806,834,914,923,956,1060,1080,1115,1201,1225,1268,1322,1324,1381,1414,1442,1482,1488,1554,1869,2229,2584,3014,3415,4991,5005,5056,5086,5088,5102,5132,5175,5251,5328,5335,5392,5484,5497,5516,5632,5708,5760,5776,5777,5880,5906,5999,6011,6115,6134,6258,6310,6386,6630,6849,6878,6934,6975,7078,7079,7082,7103,7140,7436,7631,7634,7673,7810,7812,7942,8113,8185,8262,8286,8376,8384,8459,8488,8508,8549,8693,8732,8737,8800,8829,8913,9084,9108,9192,9198,9322,9329,9354,9842,9848,9871,9936,10151,10173,10218,10357,10711,10943,11783,12334,12504,13130,14651,14993,15028,15245,17009,17181,18014,18052,18444,18636,18764,19401,19477,19530,19597,20057,20078,20380,20571,20748,21279,21334,21511,21719,22041,22543,22943,23618,24209,24736,24738,25107,26924,27717,28862,29749,29799,29808,29958,31058,31062,32247,34050,35120,36561,37020,39631,39881,40454,42198,44915,45711,45760,45901,45984,49010,50076,50240,50383,50449,50552,50553,50602,50621,50709,50802,50848,50928,51020,51159,51164,51176,51219,51235,51299,51343,51479,51510,51512,51606,51829,51873,51875,52287,52372,52375,52417,52516,52590,52723,52734,52822,52840,53010,53254,53275,53276,53342,53405,53498,53811,53852,54044,54070,54105,54110,54112,54128,54153,54370,54423,54937,55049,55065,55173,55331,55418,55466,55521,55523,55623,55699,55811,55994,56013,56038,56129,56163,56209,56334,56560,56597,56653,56726,56914,57017,57087,57328,57399,57482,57578,57647,57897,57916,57953,58126,58135,58185,58200,58368,58463,58474,58758,58855,59069,59206,59227,59265,59272,59279,59303,59339,59465,59474,59499,59601,59703,59773,59781,59933,60086,60168,60182,60187,60360,60430,60552,60571,60602,60621,60627,60745,60775,60799,60813,60817,60845,60904,60924,60992,61140,61169,61228,61245,61335,61355,61422,61438,61451,61463,61474,61522,61599,61742,61970,62005,62028,62125,62173,62208,62348,62580,62613,62644,62726,62825,63024,63166,63219,63241,63385,63433,63448,63508,63751,63785,63812,63958,63973,63976,64053,64105,64116,64148,64152,64167,64234,64248,64391,64654,64667,64752,64762,64780,64805,64890,64892,64898,64964,64973,65014,65068,65139,65315,65323,65655,65659,65687,65856,66045,66141,66289,66292,66380,66403,66412,66416,66658,66668,66744,66751,66851,66861,66976,67030,67173,67269,67277,67617,67638,67640,67659,67663,67899,67921,67963,67978,68094,68155,68161,68219,68255,68283,68313,68428,68460,68723,68851,68873,68936,68990,69013,69041,69099,69204,69219,69255,69571,69597,69604,69626,69649,69652,69662,69744,69775,69876,69936,70147,70153,70282,70299,70451,70614,70677,70826,70889,70949,70975,71196,71201,71245,71346,71364,71733,71841,71952,72222,72387,72510,72584,72647,72683,72781,72897,73011,73038,73184,73187,73229,73332,73499,73656,73753,73779,73913,74033,74313,74400,74475,74481,74519,74608,74688,74780,74820,74832,74889,74896,74959,74965,75177,75401,75515,75599,75687,75697,75717,75777,75797,75901,75979,75991,76032,76242,76353,76373,76413,76423,76638,76668,77157,77179,77227,77342,77453,77759,77777,77826,77836,77883,78087,78218,78271,78280,78317,78464,78630,78709,78710,78744,78765,78781,78831,79101,79143,79228,79271,79324,79499,79647,79759,79775,79884,79895,80180,80242,80305,80371,80414,80419,80422,80431,80507,80633,80708,80766,80814,80873,80902,81057,81186,81194,81366,81468,81662,81700,81789,81941,81981,82012,82173,82288,82376,82495,82576,82581,82586,82620,82626,82652,82663,82713,82790,82920,82932,82943,83030,83036,83041,83053,83069,83160,83189,83204,83282,83297,83326,83458,83639,83789,83813,83975,84002,84099,84109,84201,84263,84295,84301,84485,84582,84683,84811,84842,84846,84854,84860,84865,84990,84999,85007,85236,85256,85334,85483,85495,85538,85567,85623,85683,85701,85704,85738,85744,85753,85773,85940,86027,86386,86426,86442,86563,86636,86643,86676,86820,86869,87058,87099,87170,87290,87469,87482,87494,87552,87665,87772,87839,87926,87958,88035,88051,88108,88138,88157,88196,88269,88285,88332,88460,88470,88505,88567,88757,88777,88779,88818,88909,89112,89126,89158,89210,89239,89286,89309,89318,89658,89673,89699,89739,89804,89914,89985,90006,90041,90073,90239,90247,90527,90595,90689,90958,91022,91040,91204,91275,91365,91492,91582,91599,91617,91621,91652,91742,91763,91857,91941,91948,92016,92018,92091,92129,92184,92202,92225,92231,92266,92286,92307,92338,92358,92379,92504,92541,92598,92695,92764,92884,92904,92911,92933,93037,93068,93113,93127,93207,93312,93389,93437,93499,93590,93656,93659,93671,93689,93708,93745,93796,94135,94171,94209,94259,94492,94511,94590,94696,94740,94765,94810,94889,95117,95141,95162,95338,95518,95666,95740,95789,95802,95811,95836,96272,96286,96511,96621,96684,96929,96946,96951,96966,97007,97049,97257,97268,97365,97398,97440,97544,97621,97655,97681,97696,97698,97723,97754,97977,97990,98151,98158,98201,98209,98244,98268,98297,98360,98464,98587,98604,98620,98805,98835,98867,98927,99032,99052,99070,99138,99420,99522,99638,99819,99919",
          "141,273,289,348,389,409,413,430,433,472,473,514,530,613,630,700,784,814,832,861,866,868,1010,1069,1074,1121,1167,1193,1222,1331,1333,1496,1520,1617,1621,1671,1894,2090,2635,2778,2826,3242,3650,4788,4910,5049,5067,5213,5218,5257,5292,5293,5334,5353,5374,5401,5518,5613,5650,5651,5660,5712,5818,5823,5912,6017,6071,6117,6138,6239,6471,6479,6755,6902,6936,7008,7087,7091,7143,7170,7230,7275,7277,7289,7302,7347,7367,7584,7698,7718,7722,7761,7910,7988,7994,8105,8253,8292,8617,8654,8777,8815,8869,8886,8887,8968,9029,9160,9222,9760,9811,9882,10271,10291,10462,10538,10570,10591,10638,10730,10823,10936,10937,11169,11393,11429,11744,13826,14442,14496,16620,16958,17438,17747,18009,18223,19451,19601,20231,21085,21247,21560,22035,22093,23346,23462,23611,23668,23717,23901,24008,24072,25290,25668,25826,26143,26311,26404,26691,26714,26825,27870,27890,28424,28868,29044,29310,29538,29616,29789,30496,33434,35204,36549,36826,37283,37984,38066,38198,38651,38676,38794,38884,39768,41442,41756,42455,43652,45376,45414,45673,46576,47262,47551,48128,49791,50031,50178,50242,50335,50353,50459,50572,50688,50773,50825,50905,50907,50962,51109,51148,51331,51426,51487,51573,51627,51673,51764,51902,51930,51933,51953,52041,52519,52580,52662,52704,52744,52756,52810,52874,52959,52997,53004,53033,53065,53122,53257,53325,53367,53402,53484,53558,53600,53642,53684,53691,53925,53950,54068,54076,54139,54165,54199,54221,54246,54471,54477,54566,54666,54676,54723,54781,54786,54819,54820,54830,55098,55203,55313,55458,55528,55603,55631,55755,55763,55957,56152,56199,56217,56277,56323,56389,56422,56451,56573,56697,56743,56829,56920,56987,57111,57430,57479,57725,57929,57955,58107,58320,58471,58488,58563,58584,58751,58877,58892,58925,58929,58943,59064,59153,59196,59373,59402,59538,59631,59727,59766,59805,59913,59991,60153,60254,60265,60271,60442,60514,60566,60701,60836,60854,60898,61101,61248,61263,61283,61328,61368,61403,61419,61691,61715,61771,61790,61852,61861,61876,61879,61947,62040,62101,62254,62280,62406,62419,62468,62642,62684,62690,62719,63048,63075,63089,63347,63403,63424,63461,63819,63856,63886,64095,64198,64199,64213,64363,64382,64394,64454,64459,64518,64833,64907,64938,64992,65003,65004,65009,65044,65065,65067,65153,65269,65290,65316,65317,65318,65322,65387,65442,65628,65732,65742,65816,65911,65979,65998,66039,66082,66125,66182,66197,66562,66579,66659,66670,66723,66852,67161,67199,67235,67274,67334,67441,67446,67550,67630,67740,67842,67946,67984,67987,68074,68102,68248,68701,68727,68866,68912,68949,69086,69104,69198,69355,69396,69510,69543,69804,69811,69860,69865,69914,70148,70149,70174,70190,70221,70428,70452,70463,70529,70714,70721,70765,70815,70992,71042,71118,71462,71499,71578,71600,71681,71838,71880,71931,72079,72085,72145,72398,72460,72482,72485,72537,72653,72679,72701,72738,72834,72943,72950,72973,73054,73104,73125,73145,73203,73255,73262,73314,73364,73370,73398,73435,73591,73717,73727,73743,73769,73833,73851,74019,74263,74264,74467,74647,74690,74812,74925,74939,74991,75191,75268,75339,75341,75433,75505,75559,75578,75661,75701,75711,75755,76276,76282,76367,76410,76436,76449,76503,76756,76943,77110,77150,77173,77181,77319,77502,77554,77618,77629,77663,77754,77775,77794,77857,77920,77934,78263,78301,78358,78424,78668,78697,78756,78768,78808,78946,79064,79078,79080,79119,79180,79411,79599,79726,79751,79891,79898,79914,80007,80203,80215,80481,80703,80714,80738,80831,80915,81084,81211,81226,81246,81248,81262,81321,81340,81508,81517,81533,81812,81863,81895,82013,82065,82517,82610,82828,82845,83250,83301,83325,83399,83421,83444,83469,83764,83969,84138,84209,84262,84464,84576,84692,84859,84907,84965,84983,85325,85423,85432,85646,85746,86069,86126,86130,86178,86503,86545,86653,86707,86743,86803,86929,86968,86987,86996,87114,87281,87291,87363,87412,87516,87533,87652,87679,87753,87803,87829,88140,88164,88294,88301,88476,88575,88674,88748,88797,88820,88873,88916,88970,88998,89359,89369,89491,89600,89799,89924,90153,90171,90186,90190,90306,90383,90516,90567,90592,90593,90624,90656,90678,90783,90839,90865,90909,90978,91095,91274,91316,91395,91423,91447,91449,91511,91608,91634,91745,91863,91911,91982,91988,92174,92216,92284,92366,92396,92449,92486,92534,92550,92781,92818,93046,93147,93154,93255,93310,93383,93531,93580,93647,93746,93753,93815,93943,94109,94110,94134,94444,94614,94624,94651,94746,94762,94788,94895,94918,95041,95186,95257,95334,95362,95634,95699,95798,95867,95896,96023,96052,96110,96160,96252,96315,96332,96398,96417,96487,96626,96680,96764,96904,96937,96958,97005,97231,97247,97260,97304,97458,97487,97566,97683,97758,97816,97846,98046,98163,98319,98321,98330,98473,98626,98688,98801,98902,98922,98928,98978,98986,98990,99147,99156,99213,99308,99427,99432,99503,99536,99582,99690,99815,99849,99873,99927,100003",
          "196,197,235,271,290,396,505,510,525,567,576,675,766,841,846,864,966,973,1041,1123,1157,1203,1251,1315,1347,1352,1357,1364,1366,1385,1389,1392,1393,1445,1475,1490,1610,1645,1790,1805,1817,1826,1858,1926,1964,1968,1997,2033,2045,2114,2116,2158,2283,2287,2337,2338,2347,2367,2397,2424,2453,2472,2529,2532,2549,2551,2561,2594,2641,2678,2684,2772,2808,2833,2873,2894,2901,2910,2926,2932,2935,2958,2960,2974,2980,2981,2997,3001,3038,3127,3135,3145,3151,3230,3251,3268,3306,3331,3350,3395,3410,3414,3429,3440,3504,3565,3582,3586,3588,3614,3658,3679,3680,3710,3713,3759,3768,3771,3846,3856,3859,3877,3887,3908,3921,3933,3946,3960,4017,4043,4128,4142,4173,4207,4249,4272,4405,4426,4450,4454,4466,4480,4493,4522,4527,4541,4545,4546,4547,4560,4561,4653,4661,4662,4669,4678,4697,4698,4705,4730,4742,4817,4863,4876,4881,4895,4946,4960,4962,4994,5021,5118,5129,5250,5438,5631,5775,5805,5832,5851,5878,5885,5929,5957,6030,6042,6089,6174,6195,6198,6328,6330,6396,6462,6500,6624,6660,6685,6944,7116,7452,7471,7640,7747,7771,7847,7899,7952,7966,8023,8187,8224,8300,8312,8364,8420,8480,8489,8754,8818,8875,8910,8916,8935,8969,9041,9102,9106,9258,9268,9324,9348,9413,9520,9525,9531,9645,9726,9873,10046,10498,11027,11267,11385,11758,11860,11863,11923,12412,12630,12650,12910,13067,13346,13489,13514,13628,13682,13845,13967,13972,14420,14447,14486,14532,14625,14746,14779,15098,15458,15502,15541,15544,15713,15775,15881,15906,15934,16022,16203,16338,16510,16942,17556,17572,17936,17962,18303,18446,18591,18939,19310,19334,19546,20153,20159,20175,20272,20420,20912,20964,21044,21140,21576,21599,21760,21837,21857,21976,22405,22453,22567,22649,22699,22770,22798,22807,22893,22938,22996,23135,23229,23466,23612,23789,24124,24177,24318,24323,24471,24750,24940,24978,25059,25215,25390,25473,25508,25526,25573,25777,25828,26088,26132,26169,26225,26530,26547,26548,26550,26590,26662,26890,26996,27072,27122,27256,27420,27454,27489,27583,27617,27628,27893,27966,27983,28012,28021,28031,28047,28076,28077,28104,28118,28121,28146,28188,28319,28453,28570,28621,28638,28641,28650,28672,28767,28810,28827,28850,29029,29133,29334,29417,29632,29657,29712,30014,30072,30159,30217,30246,30309,30362,30388,30391,30449,30478,30565,30594,30710,30739,30768,30797,30826,30855,30971,31000,31087,31116,31174,31203,31216,31261,31348,31377,31406,31435,31464,31493,31551,31580,31609,31638,31667,31754,31812,31870,31899,31928,31957,31986,32015,32044,32160,32189,32276,32334,32421,32479,32508,32595,32624,32653,32682,32711,32740,32769,32798,32827,32856,32885,32914,33030,33117,33146,33204,33378,33465,33523,33552,33610,33639,33668,33697,33726,33755,33813,33900,33987,34045,34132,34161,34277,34306,34335,34364,34393,34431,34451,34509,34538,34567,34596,34625,34654,34683,34799,34828,34860,34886,34915,34944,34973,35062,35103,35193,35205,35217,35266,35295,35414,35485,35519,35533,35633,35651,35694,35724,35799,35813,35863,35913,35972,36031,36154,36178,36343,36364,36402,36423,36592,36656,36845,36883,36890,36904,36963,37217,37239,37276,37335,37362,37596,37655,37714,37896,37922,37967,38026,38216,38275,38339,38457,38483,38555,38587,38685,38718,38777,38900,39018,39047,39074,39089,39148,39207,39279,39397,39461,39483,39520,39579,39595,39601,39650,39709,39899,39958,40041,40125,40154,40183,40212,40241,40299,40328,40357,40386,40473,40531,40560,40763,40821,40850,40879,40966,40995,41024,41053,41082,41140,41169,41256,41285,41314,41343,41430,41459,41488,41546,42043,42045,42074,42132,42140,42192,42292,42311,42312,42332,42357,42432,42470,42519,42527,42530,42594,42669,42726,42756,42862,42906,42948,42993,43022,43143,43183,43254,43305,43380,43542,43569,43658,43661,43735,43779,43861,43869,43947,44057,44058,44091,44102,44178,44205,44253,44280,44328,44415,44490,44565,44601,44612,44652,44660,44680,44773,44802,44839,44889,44912,44942,44970,44981,45039,45112,45160,45221,45234,45279,45362,45409,45467,45487,45654,45726,45741,45841,45913,45970,46003,46046,46069,46088,46089,46115,46142,46156,46175,46214,46259,46275,46299,46363,46401,46533,46535,46570,46573,46577,46607,46639,46689,46710,46727,46797,46805,46857,46925,46973,46982,47070,47080,47155,47174,47192,47455,47459,47486,47554,47681,47688,47749,47787,47907,47932,47971,48188,48267,48292,48324,48326,48359,48387,48437,48443,48533,48593,48696,48775,48789,48884,48914,48915,49148,49163,49186,49195,49250,49277,49307,49419,49485,49578,49626,49665,49729,49735,49736,49746,49750,49814,49894,49905,49978,50046,50272,50307,50395,50500,50742,50871,50876,50952,50959,51016,51070,51151,51170,51296,51315,51702,51752,51962,51973,51992,52040,52063,52099,52220,52233,52315,52438,52487,52490,52512,52566,52628,52644,52647,52659,52667,53329,53673,53765,53889,53920,53954,54046,54061,54171,54368,54388,54413,54495,54537,54606,54618,54671,54677,54698,54746,54818,54863,54941,55042,55168,55239,55261,55267,55594,55790,55812,55819,55921,55975,56004,56034,56113,56161,56295,56337,56352,56384,56439,56453,56457,56494,56608,56627,56629,56646,56742,56745,56749,56806,56834,56884,56886,56960,56979,57003,57073,57138,57163,57167,57266,57363,57378,57517,57554,57574,57658,57672,57738,57818,57821,57825,57888,57898,58027,58141,58165,58191,58194,58217,58302,58484,58528,58533,58600,58624,58710,58822,58825,58830,58928,58980,59076,59090,59345,59389,59422,59577,59632,59680,59850,59887,59966,59974,60118,60318,60330,60332,60389,60425,60591,60604,60618,60632,60666,60670,60692,60756,60804,61032,61082,61141,61155,61267,61310,61370,61490,61534,61582,61610,61746,61791,61795,61800,61928,61938,61942,61958,62032,62181,62206,62225,62259,62409,62564,62700,62811,62888,62936,63072,63098,63120,63346,63390,63491,63517,63577,63667,63700,63708,63754,63755,63790,63791,63965,64123,64283,64335,64396,64403,64573,64696,64789,64803,64815,64828,64859,64861,64946,64948,65184,65224,65232,65348,65353,65569,65595,65729,65887,65958,65964,65989,66067,66107,66211,66306,66346,66386,66394,66431,66479,66483,66582,66685,66761,66763,66858,66959,66968,67263,67350,67354,67442,67518,67522,67604,67621,67670,67720,67770,67821,67869,67909,67964,68018,68110,68125,68278,68302,68309,68442,68633,68720,69307,69393,69528,69872,69999,70217,70337,70387,70406,70649,70675,70745,70757,70770,70972,70976,71047,71048,71136,71137,71143,71183,71216,71314,71388,71545,71979,72028,72080,72292,72471,72526,72627,72662,73009,73013,73110,73133,73178,73180,73253,73272,73445,73451,73477,73514,73680,73722,73755,73806,73848,74053,74153,74220,74223,74297,74725,74772,75094,75146,75201,75273,75295,75315,75385,75586,75646,75673,75754,75819,75872,75877,75998,76020,76098,76122,76420,76467,76780,76970,77089,77099,77387,77442,77550,77605,77667,77714,77808,77849,77886,78005,78157,78323,78381,78384,78393,78555,78573,78580,78588,78652,78675,78712,78731,78853,78890,78940,78957,79019,79242,79295,79312,79639,79682,79741,79769,79834,80033,80042,80084,80150,80221,80289,80331,80616,80710,80727,80775,81164,81257,81259,81305,81310,81333,81361,81386,81480,81493,81832,81944,81980,82005,82046,82060,82094,82235,82282,82378,82457,82506,82817,82879,82908,83081,83100,83230,83246,83507,83514,83519,83581,83622,83752,83831,83980,84005,84026,84041,84056,84093,84128,84132,84144,84188,84223,84283,84476,84500,84510,84605,84653,84655,84878,84919,84950,85082,85480,85700,85745,85786,85867,85872,85912,85942,85981,86093,86217,86325,86672,86822,86899,86937,87091,87119,87215,87259,87301,87306,87392,87488,87526,87785,87806,87816,87823,87910,87923,87941,87997,88026,88218,88270,88282,88299,88427,88557,88600,88617,88741,88773,88792,89098,89215,89274,89407,89439,89466,89479,89655,89659,89729,89801,90169,90244,90271,90296,90320,90324,90603,90620,90777,90785,90869,90920,90924,91001,91070,91108,91112,91117,91175,91256,91489,91671,91688,91713,91799,91802,91824,91876,91915,92180,92213,92223,92448,92470,92513,92528,92582,92649,92687,92789,92875,92946,93226,93318,93362,93474,93593,93706,93787,93856,93859,94078,94085,94283,94304,94329,94360,94553,94584,94782,94875,94883,94940,94990,95032,95219,95329,95392,95402,95404,95498,95512,95525,95561,95632,95763,95832,96118,96120,96302,96318,96393,96397,96449,96530,96559,96570,96606,96659,96853,96933,96965,96992,97164,97235,97267,97322,97397,97612,97630,97652,97671,97714,97788,97830,97865,97929,98023,98085,98100,98138,98402,98445,98455,98498,98503,98505,98614,98726,98773,98782,98866,99069,99105,99169,99258,99294,99601,99754,99810",
          "307,337,443,463,482,507,535,587,642,713,760,772,822,842,845,911,915,922,937,944,970,978,993,1002,1007,1034,1035,1057,1098,1101,1124,1137,1169,1226,1285,1293,1311,1344,1405,1495,1519,1537,1549,1870,1916,2260,2491,4323,4457,5034,5070,5143,5150,5166,5245,5323,5347,5376,5377,5403,5481,5489,5492,5606,5654,5686,5796,5872,5895,5924,6084,6222,6368,6428,6458,6514,6594,6666,6691,6765,6870,6966,7010,7021,7068,7119,7124,7131,7166,7239,7279,7296,7328,7404,7651,7778,7832,7877,7971,8096,8254,8397,8415,8471,8623,8797,8828,8830,8932,8942,8976,9019,9047,9137,9273,9306,9744,9887,10176,10266,10300,10304,10346,10398,10545,10629,10840,11034,11040,11074,11190,11299,11457,11616,11725,11776,12797,13437,13476,13513,13723,14131,14342,14897,14973,15593,15619,15773,17081,17193,17254,17933,17991,18887,19971,19981,20361,20791,21320,21848,22212,22760,23383,24431,25024,25210,26824,27029,27829,28310,28683,29353,33495,36458,36503,36515,37576,37794,38141,38698,39002,39309,39678,39824,39939,41087,43634,45419,45495,45544,47238,49478,50181,50262,50310,50357,50431,50564,50592,50636,50674,50700,50724,50893,51075,51165,51438,51478,51516,51570,51735,51771,51772,51799,51869,51913,51944,51960,52033,52124,52131,52230,52289,52349,52355,52361,52389,52768,52842,52875,52893,52923,53066,53099,53124,53160,53246,53263,53282,53331,53417,53426,53499,53512,53668,53727,53732,53804,53845,53976,53977,54205,54318,54400,54519,54536,54647,54656,54743,54936,55053,55165,55169,55206,55243,55305,55376,55441,55545,55551,55622,55787,55889,55927,56172,56232,56247,56347,56533,56600,56650,56813,56969,57036,57062,57175,57532,57562,57613,57619,57629,57695,57703,57774,57793,58183,58210,58233,58357,58907,58973,59237,59241,59392,59466,59531,59619,59635,59717,59730,60285,60337,60413,60631,60656,60949,61171,61260,61269,61301,61417,61479,61481,61557,61561,61605,61652,61720,61762,61769,61831,61962,62063,62088,62162,62301,62302,62395,62411,62495,62616,62756,62832,63055,63262,63293,63352,63436,63541,63629,63652,63748,63759,63870,63943,64023,64037,64183,64294,64297,64321,64360,64508,64743,64749,64797,64806,64837,64838,64840,64869,64870,64935,64983,64991,65021,65125,65239,65252,65539,65590,65631,65731,65741,65749,65781,66050,66176,66395,66467,66500,66563,66646,66660,66717,66781,66848,66856,66859,66963,67012,67022,67053,67170,67171,67176,67236,67451,67460,67676,67681,67849,68005,68072,68118,68140,68145,68148,68238,68277,68511,68578,68598,68659,68668,68888,68898,68986,69098,69418,69544,69560,69586,69651,69809,69829,69916,70012,70166,70227,70496,70519,70905,70910,71026,71142,71179,71246,71283,71422,71467,71544,71606,71709,71740,71795,71845,71893,71925,72007,72053,72094,72213,72305,72402,72439,72447,72496,72600,72655,72766,72857,72867,72893,73039,73167,73188,73302,73377,73448,73574,73637,73815,74005,74147,74191,74278,74391,74414,74488,74609,74706,74739,74791,74881,74940,74949,75104,75180,75181,75233,75244,75291,75367,75390,75457,75507,75639,75664,75782,75883,75893,75920,75943,75974,76012,76121,76142,76156,76174,76219,76339,76466,76489,76542,76644,76672,76737,76763,76852,76875,77273,77335,77372,77620,77726,77730,77922,77989,78070,78092,78221,78248,78307,78334,78352,78372,78437,78439,78455,78584,78590,78631,78814,78818,78839,78844,78996,79076,79224,79282,79358,79377,79390,79428,79466,79495,79540,79680,79776,79813,79885,79963,80006,80225,80316,80365,80392,80541,80592,80594,80847,80876,80968,81320,81343,81355,81469,81471,81616,81828,81922,81993,82007,82129,82170,82302,82393,82417,82425,82442,82705,82805,82839,82915,83065,83099,83106,83107,83186,83264,83402,83446,83461,83509,83787,83953,83958,84057,84064,84192,84273,84426,84429,84487,84506,84610,84689,84934,85030,85083,85091,85166,85402,85456,85559,85592,85664,85761,85882,86096,86254,86274,86328,86660,86666,86673,86701,87000,87011,87171,87206,87274,87316,87471,87589,87626,87706,87722,87766,87783,87826,87831,87861,87972,88110,88210,88297,88364,88372,88415,88515,88605,88676,88699,88740,88774,88837,88968,89293,89430,89445,89502,89531,89646,89870,89881,89936,89973,90337,90379,90408,90441,90551,90553,90643,90780,90791,90809,90989,91009,91053,91169,91176,91243,91252,91279,91343,91416,91439,91554,91852,91986,92080,92212,92244,92389,92421,92493,92543,92661,92760,93087,93282,93316,93329,93407,93416,93486,93517,93586,93736,93802,93807,93822,93832,93839,93895,93993,94305,94374,94499,94541,94542,94612,94673,94812,95013,95045,95184,95193,95275,95278,95380,95583,95668,95907,95961,96066,96127,96220,96259,96271,96329,96440,96468,96494,96536,96577,96611,96841,96895,97001,97076,97194,97291,97333,97350,97394,97485,97580,97616,97650,97771,97853,97911,97933,98331,98338,98426,98450,98593,98679,98759,98806,98807,99022,99073,99190,99201,99212,99220,99253,99318,99382,99387,99570,99659,99665,99701,99751,99783,99813,99824,99902,99935,99976",
          "34,86,220,247,382,563,591,684,715,781,791,856,916,926,928,929,1164,1220,1299,1337,1341,1376,1379,1398,1435,1444,1454,1531,1569,1605,1689,1767,1768,1822,1825,1849,1850,1913,1930,2026,2059,2137,2612,2613,2659,2673,2880,3084,3095,3112,3210,3280,3383,3629,3681,3776,3805,3837,3941,4188,4242,4321,4401,4410,4414,4623,4723,4770,4778,4878,5016,5020,5053,5113,5126,5221,5333,5358,5379,5457,5541,5559,5568,5683,5771,5813,5854,5891,5964,6054,6346,6397,6423,6484,6572,6658,6680,6760,6952,6987,7028,7165,7210,7215,7355,7363,7508,7648,7838,8084,8208,8354,8388,8398,8498,8509,8534,8550,8605,8771,8791,8961,9009,9068,9172,9197,9391,9427,9530,9734,9770,10041,10050,10108,10401,10436,10443,10447,10481,10482,10504,10562,10810,11080,11128,11279,11284,11325,11328,11443,11481,11541,11543,11674,11681,11739,11828,11875,11898,12116,12300,12548,12686,12793,12852,13123,13126,13155,13180,13201,13305,13307,13316,13367,13375,13468,13500,13547,13662,13827,13893,14000,14054,14068,14165,14196,14242,14283,14361,14407,14417,14495,14548,14627,14777,14796,15689,15703,15863,15910,16140,16164,16170,16174,16430,16505,16570,16609,16686,16817,16863,17012,17123,17132,17145,17285,17383,17492,17699,17945,17977,17996,18120,18130,18148,18233,18254,18424,18482,18577,18840,18909,18915,18944,19038,19085,19131,19186,19202,19235,19629,19671,19840,19877,19898,19983,20041,20229,20397,20552,20577,20644,20672,20742,20749,20847,20858,20860,21167,21360,21371,21546,21847,22061,22164,22679,22746,22814,23074,23369,23411,23781,23906,24038,24130,24221,24422,24505,24587,24622,24644,24712,24991,25116,25363,25468,25519,25628,25990,26157,26159,26166,26219,26631,26673,26780,26800,26893,27040,27059,27119,27302,27345,27394,27502,27689,27738,27746,27773,27822,27828,27838,27855,27925,27956,28250,28279,28342,28373,28406,28413,28566,28572,28745,28991,29024,29049,29221,29621,29634,29707,29714,29739,29779,29785,29829,30307,30585,30788,32047,32257,32876,33978,34074,35033,35113,35393,35467,35569,35575,35620,35767,35768,35808,35857,35930,35989,36046,36062,36175,36268,36375,36386,36567,36742,36872,36907,37068,37163,37200,37303,37304,37370,37385,37415,37612,37630,37748,37762,37854,37863,38108,38193,38232,38290,38485,38569,38608,38817,38821,38998,39102,39124,39374,39556,39615,39819,39865,39873,39920,39934,39941,40427,40841,40872,41721,42169,42181,42364,42389,42656,42715,42727,42763,42824,42853,42859,42870,42940,43014,43078,43424,43756,43813,44157,44169,44296,44451,44578,44705,45092,45118,45153,45174,45321,45378,45496,45548,45620,45643,45774,45926,45937,45939,46018,46466,46861,46983,47181,47727,47930,48000,48181,48447,48840,49114,50157,50273,50276,50315,50323,50345,50444,50507,50508,50608,50658,50738,50821,50873,50938,50964,51329,51358,51443,51476,51583,51756,51763,51777,51788,51802,51898,52020,52058,52151,52342,52350,52364,52380,52549,52554,52594,52661,52671,52738,52839,52845,52854,53027,53055,53106,53148,53229,53322,53335,53369,53537,53597,53680,53682,53759,53802,53821,53901,54160,54169,54220,54612,54964,54989,55010,55021,55061,55089,55133,55197,55220,55264,55286,55293,55385,55493,55655,55668,55669,55815,55849,55885,55896,55918,55940,56127,56150,56360,56371,56456,56520,56657,56735,56858,56918,57014,57061,57342,57412,57698,57906,58042,58085,58130,58158,58250,58275,58286,58359,58439,58460,58514,58569,58664,58679,58755,58901,59034,59065,59115,59118,59205,59361,59372,59449,59524,59579,59616,59656,59960,59970,60068,60300,60319,60396,60408,60411,60422,60428,60429,60489,60513,60702,60736,60767,60903,60984,61026,61154,61364,61575,61587,61621,61662,61692,61974,62150,62289,62310,62329,62417,62647,62675,62762,62784,62804,62860,63063,63248,63407,63439,63446,63536,63555,63672,63760,64073,64258,64303,64308,64348,64385,64456,64669,65293,65302,65480,65581,65688,65727,65730,65847,65960,66026,66032,66152,66204,66276,66319,66342,66496,66555,66683,66811,66893,66898,66978,67200,67395,67419,67543,67578,67580,67697,67714,67801,67822,67858,68020,68050,68095,68260,68312,68315,68373,68576,68615,68755,68783,68788,68832,68983,69052,69081,69106,69143,69263,69279,69305,69317,69496,69554,69671,69714,69874,69991,69996,70189,70359,70469,70504,70571,70612,70967,70995,71038,71148,71164,71169,71473,71490,71582,71693,71712,71814,71854,71905,72041,72060,72097,72157,72166,72202,72268,72288,72340,72368,72495,72565,72582,72668,72764,72770,72909,73030,73043,73149,73227,73256,73462,73655,73914,73947,74066,74076,74111,74112,74218,74254,74276,74291,74471,74529,74599,74672,74839,74942,74968,75232,75258,75393,75407,75531,75534,75605,75631,75791,75795,75812,75889,75937,75972,75996,76087,76146,76227,76233,76247,76379,76409,76507,76632,76713,76777,76869,77039,77069,77138,77189,77280,77366,77476,77530,77537,77570,77682,77805,77889,78022,78025,78027,78184,78261,78268,78484,78620,78984,79032,79198,79204,79330,79366,79456,79581,79674,79893,79971,79979,80050,80066,80224,80241,80284,80566,80580,80620,80666,80673,80732,80785,80838,80880,80898,81022,81119,81177,81272,81341,81383,81498,81531,81679,81736,81738,81886,81977,81978,81999,82008,82073,82445,82480,82555,82605,82654,82660,82709,82768,82849,82868,82954,83012,83292,83457,83606,83758,83765,83772,83859,83888,83899,83939,83956,83991,84036,84297,84531,84546,84591,84781,84988,84995,85053,85296,85357,85393,85415,85564,85581,85758,85839,85846,85869,85930,85959,86029,86160,86183,86190,86273,86430,86447,86510,86706,86721,86864,86877,87080,87240,87262,87295,87350,87364,87396,87406,87418,87422,87607,87670,87678,87791,87913,88206,88312,88334,88370,88497,88561,88565,88595,88783,88828,88935,89254,89280,89352,89683,89875,89912,89923,90175,90264,90463,90545,90601,90684,90688,90834,90897,90906,90910,90916,90940,91041,91190,91287,91344,91469,91477,91498,91539,91575,91920,91922,91952,92052,92109,92282,92423,92575,92601,92680,92699,93143,93230,93261,93466,93492,93514,93599,93610,93664,93738,93918,93962,94017,94025,94035,94057,94077,94158,94182,94190,94199,94243,94330,94448,94551,94769,94774,94840,94843,94910,95000,95142,95143,95220,95486,95600,95625,95646,95702,95741,95762,95916,95919,95991,95996,96126,96282,96473,96549,96609,96727,96830,97274,97359,97513,97610,97667,97786,97909,98178,98607,98611,98960,98967,98972,99013,99026,99047,99256,99322,99397,99449,99463,99490,99608,99611,99722,99729,99829,99988",
          "36,56,112,139,147,192,1715,2141,2148,2245,2364,2620,2728,2777,2791,3011,3045,3163,3239,3673,4213,4232,4485,4509,4974,4975,5877,6013,6591,6643,6852,6863,6932,7161,7247,8019,8130,8168,8244,8510,8643,8681,8928,8930,9023,9049,9080,9099,9141,9175,9194,9346,9384,9460,9469,9498,9593,9612,9650,9688,9707,9819,9820,9821,9878,9935,9943,10049,10057,10165,10212,10220,10239,10258,10269,10328,10429,10467,10469,10512,10581,10600,10636,10654,10664,10752,10809,10820,10885,10961,11015,11021,11047,11113,11170,11188,11208,11303,11455,11512,11518,11521,11551,11576,11599,11645,11721,11759,11877,11892,11899,11911,11993,12025,12109,12215,12234,12253,12291,12310,12386,12481,12500,12557,12566,12671,12747,12804,12899,13089,13211,13214,13260,13279,13298,13374,13469,13492,13621,13640,13678,13697,13754,13773,13921,13944,13963,14020,14058,14061,14077,14153,14191,14229,14267,14305,14381,14457,14533,14590,14628,14663,14685,14818,14896,14913,14918,14985,14989,14998,15034,15141,15179,15227,15236,15255,15350,15377,15407,15464,15597,15635,15692,15730,15749,15806,15825,15844,15920,16034,16053,16063,16110,16148,16167,16186,16205,16224,16243,16281,16300,16319,16376,16433,16483,16490,16509,16661,16737,16813,16870,16927,16940,16943,16965,17022,17091,17117,17212,17231,17307,17345,17364,17368,17437,17459,17478,17497,17592,17610,17611,17687,17720,17744,17782,17801,17839,17896,17934,17953,17954,17972,17979,18048,18105,18124,18181,18200,18276,18314,18352,18409,18447,18479,18593,18618,18694,18765,18770,18789,18846,18903,18960,19084,19150,19188,19230,19264,19302,19340,19359,19416,19419,19435,19453,19682,19740,19834,19872,19891,19910,19948,20005,20062,20074,20119,20138,20157,20178,20195,20233,20271,20309,20347,20423,20438,20461,20594,20613,20898,20993,21021,21069,21077,21145,21164,21188,21202,21219,21274,21278,21335,21380,21411,21430,21468,21482,21563,21582,21715,21734,21753,21791,21810,21829,21867,21905,21924,21943,22019,22073,22076,22228,22266,22342,22399,22418,22437,22456,22513,22533,22589,22647,22684,22703,22735,22832,22840,22945,22950,23010,23028,23083,23178,23197,23235,23255,23273,23329,23330,23347,23368,23425,23463,23499,23501,23539,23558,23577,23605,23623,23634,23710,23748,23862,23881,23946,23976,23995,24052,24150,24194,24204,24223,24242,24250,24337,24432,24451,24470,24527,24565,24584,24755,24857,24869,24888,24941,25002,25040,25073,25078,25094,25097,25135,25211,25287,25302,25325,25381,25382,25477,25534,25553,25565,25667,25714,25781,25819,25838,25876,25914,25952,25971,26123,26142,26161,26218,26237,26351,26408,26427,26484,26503,26636,26687,26705,26763,26875,26902,26997,27116,27162,27203,27244,27250,27295,27377,27383,27418,27434,27593,27618,27670,27719,27736,27909,28080,28099,28113,28194,28295,28536,28574,28593,28762,28897,28954,29201,29226,29227,29292,29439,29562,29638,29752,29809,29860,29879,29934,29985,30027,30056,30143,30172,30230,30259,30317,30346,30375,30404,30462,30491,30520,30607,30636,30665,30723,30752,30781,30810,30839,30897,30926,30955,31013,31071,31100,31129,31158,31187,31245,31274,31290,31361,31448,31506,31535,31651,31680,31725,31767,31796,31854,31912,31970,31999,32086,32115,32144,32173,32231,32260,32318,32347,32376,32392,32405,32463,32492,32579,32666,32695,32724,32840,32869,32927,33014,33043,33072,33101,33130,33188,33217,33304,33333,33362,33478,33507,33536,33594,33623,33681,33710,33855,33884,33913,33971,34029,34116,34174,34232,34261,34290,34348,34522,34609,34638,34696,34725,34754,34783,34812,34870,34928,34957,35027,35116,35142,35279,35324,35353,35359,35427,35546,35664,35796,35855,35867,35985,35992,36004,36080,36107,36112,36167,36191,36226,36284,36297,36356,36415,36605,36669,36728,36787,36858,36976,36980,36988,37048,37166,37172,37230,37293,37299,37348,37392,37419,37488,37537,37548,37674,37791,37850,37914,38039,38288,38352,38411,38470,38528,38541,38600,38659,38731,38981,38993,39161,39226,39248,39292,39533,39592,39781,39912,39971,40101,40130,40188,40246,40275,40333,40391,40507,40536,40565,40623,40676,40681,40710,40739,40768,40797,40884,40913,40942,40971,41000,41174,41319,41464,41493,41522,42054,42083,42112,42172,42231,42260,42337,42406,42499,42574,42603,42620,42649,42736,42771,42811,42845,42886,42973,43016,43048,43116,43155,43210,43248,43275,43278,43283,43285,43347,43360,43447,43488,43574,43597,43641,43655,43684,43759,43814,43899,43909,43921,43996,44056,44071,44119,44129,44134,44158,44190,44308,44334,44360,44395,44470,44500,44525,44530,44583,44603,44632,44661,44669,44684,44707,44743,44782,44823,44841,44944,44975,45010,45019,45140,45254,45332,45389,45425,45519,45576,45634,45706,45763,45821,45848,45854,45893,46030,46098,46105,46111,46161,46364,46381,46444,46508,46591,46619,46735,46742,46762,46969,47103,47175,47180,47216,47234,47250,47317,47340,47346,47378,47539,47567,47587,47605,47660,47765,47850,47940,48069,48130,48136,48306,48329,48409,48499,48519,48543,48566,48589,48610,48624,48644,48704,48739,48880,48945,49026,49028,49071,49228,49311,49361,49477,49535,49574,49646,49706,49811,49813,49828,49831,49855,49856,49940,49945,49973,50004,50088,50108,50138,50199,50266,50271,50287,50629,50649,50995,51037,51053,51077,51243,51251,51256,51332,51381,51450,51691,51728,51852,51886,51943,51945,51988,52060,52101,52357,52382,52589,52763,52764,52862,52886,52905,53171,53184,53438,53511,53622,53631,53660,54118,54178,54263,54383,54461,54512,54587,54625,54630,54737,54868,54939,55000,55107,55157,55311,55330,55368,55380,55397,55517,55554,55638,55672,55724,55754,55956,56039,56111,56114,56248,56291,56450,56474,56582,56603,56684,56693,56901,56911,57136,57237,57308,57416,57503,57514,57570,57572,57688,58019,58059,58095,58231,58300,58399,58422,58536,58771,58786,58787,58918,58968,58997,59004,59040,59054,59308,59321,59371,59424,59578,59597,59608,59784,59814,59828,59835,59884,59923,59982,60001,60030,60136,60140,60142,60162,60221,60235,60615,60719,60777,60944,60958,60964,61069,61211,61342,61435,61496,61687,61857,61912,61921,62131,62169,62231,62248,62274,62363,62561,62769,62855,62974,63084,63137,63278,63412,63566,63809,63914,63969,64055,64056,64135,64172,64204,64273,64310,64357,64367,64375,64404,64666,64775,64779,65119,65215,65219,65270,65421,65559,65599,65650,65693,65850,65915,65923,65976,66094,66159,66254,66281,66477,66540,66628,66681,66729,66750,66756,67035,67063,67084,67153,67155,67275,67393,67411,67458,67465,67565,67917,67930,68006,68008,68011,68065,68080,68211,68213,68216,68488,68500,68514,68536,69019,69024,69142,69233,69273,69345,69514,69657,69684,69693,69723,69749,69833,69834,69862,69899,69927,69958,70050,70237,70275,70423,70461,70599,70845,71034,71078,71084,71129,71131,71320,71356,71496,71564,71585,71613,71624,71671,71692,71694,71772,71901,71907,71957,71988,71993,72052,72088,72121,72313,72390,72391,72509,72622,72742,72744,72880,72910,73023,73460,73498,73547,73572,73719,73735,73800,73955,74022,74056,74226,74265,74320,74338,74451,74518,74619,74620,74625,74658,74715,74792,74844,74849,74851,74961,75029,75075,75088,75099,75340,75374,75813,75874,76010,76042,76107,76150,76177,76193,76439,76453,76748,77012,77071,77195,77306,77310,77396,77477,77551,77568,77581,77700,77704,77928,77954,77995,78172,78340,78468,78859,78945,79043,79205,79383,79430,79433,79712,79793,79847,79916,80037,80437,80456,80574,80774,80981,80988,81139,81163,81179,81253,81364,81404,81809,81845,81889,81904,81907,81947,81963,82045,82052,82084,82097,82137,82364,82401,82459,82473,82816,83059,83207,83295,83366,83468,83521,83663,83673,83686,83701,83821,83915,83928,83973,83996,84129,84187,84189,84268,84320,84356,84373,84395,84719,84812,84835,84890,84941,84971,85016,85069,85102,85157,85216,85219,85288,85383,85586,85652,85656,85804,85945,85988,86100,86187,86218,86307,86337,86368,86480,86483,86627,86784,87030,87085,87090,87144,87317,87436,87493,87529,87590,87800,87812,87822,87916,87945,88042,88309,88337,88482,88521,88555,88724,89077,89227,89321,89340,89386,89387,89403,89526,89616,89626,89652,89781,89815,89893,89922,90019,90267,90346,90467,90639,90804,90816,90857,90928,91037,91066,91087,91285,91305,91327,91604,91632,91642,91692,91725,91808,92062,92220,92306,92410,92440,92484,92562,93428,93442,93566,93803,93812,93897,94130,94138,94249,94445,94478,94620,94807,94815,94826,94946,95020,95178,95282,95369,95412,95461,95526,95548,95827,96067,96531,96703,96781,96782,96881,96884,97038,97151,97186,97219,97234,97366,97475,97505,97848,97849,97867,98239,98497,98982,99027,99092,99150,99172,99224,99331,99353,99438,99495,99498,99504,99511,99513,99528,99541,99544,99625,99773,99805,99869,99882",
          "88,107,174,180,252,718,1627,1665,1752,1763,1910,1921,1928,1946,1956,1960,1973,1989,2000,2005,2021,2032,2040,2071,2073,2084,2097,2098,2102,2104,2115,2124,2130,2145,2171,2175,2187,2220,2234,2236,2254,2262,2285,2315,2403,2407,2444,2460,2489,2503,2520,2552,2558,2597,2615,2639,2650,2675,2687,2718,2726,2729,2750,2752,2760,2762,2776,2787,2796,2840,2856,2881,2888,2898,2906,2917,2945,2963,2964,2978,2991,2993,2994,3002,3019,3024,3041,3047,3048,3076,3079,3128,3138,3167,3170,3176,3187,3191,3236,3254,3263,3266,3267,3269,3283,3292,3307,3365,3397,3423,3426,3436,3444,3445,3449,3465,3488,3489,3494,3505,3514,3555,3559,3583,3590,3599,3626,3652,3666,3669,3678,3697,3705,3712,3725,3788,3806,3808,3809,3819,3844,3847,3848,3850,3866,3872,3892,3894,3904,3909,3957,3962,3982,3992,3993,3995,4002,4021,4039,4045,4065,4082,4085,4103,4120,4143,4150,4193,4196,4219,4222,4224,4228,4238,4254,4288,4292,4298,4301,4305,4307,4313,4326,4332,4339,4345,4365,4421,4446,4463,4468,4478,4503,4510,4518,4520,4543,4556,4564,4598,4615,4632,4638,4664,4673,4689,4740,4744,4756,4758,4776,4792,4794,4823,4825,4832,4856,4861,4871,4874,4889,4909,4932,4937,5463,5583,5598,5655,5676,5921,6005,6104,6298,6472,6481,6487,6568,6586,6593,6605,6657,6701,6720,6741,6789,6838,6904,6909,6965,7090,7118,7181,7232,7234,7265,7293,7324,7349,7359,7510,7551,7582,7619,7699,7802,7854,7881,7913,8041,8060,8147,8163,8200,8250,8375,8616,8686,8750,8909,8992,9014,9031,9089,9146,9210,9244,9292,9551,9729,9733,9754,9813,9928,9954,10008,10010,10115,10256,10290,10301,10302,10419,10437,10454,10589,10594,10784,10904,10994,11086,11089,11141,11187,11289,11390,11420,11610,11680,11716,11865,11926,11936,12027,12080,12179,12362,12401,12402,12544,12675,12714,13000,13097,13288,13538,13552,13583,13721,13753,13761,13806,13815,13861,13877,14015,14125,14174,14281,14392,14393,14482,14550,14562,14576,14582,14825,14903,14948,15065,15115,15288,15345,15399,15718,15735,16087,16195,16308,16384,16544,16723,16810,16860,16968,17265,17316,17376,17429,17558,17675,17730,17915,18019,18067,18151,18178,18271,18328,18474,18623,18638,18641,18671,18691,18711,18714,18732,18776,18788,18798,18836,19113,19165,19295,19425,19491,19655,19712,19778,19857,19901,19943,20136,20211,20399,20426,20479,20491,20514,20543,20584,20667,20800,20931,20947,20976,21035,21123,21147,21258,21416,21500,21530,21556,21840,21959,22005,22069,22118,22281,22379,22492,22695,22857,22940,22958,22979,23047,23122,23215,23216,23401,23408,23436,23609,23835,23846,23847,23980,24005,24120,24179,24244,24634,24742,25376,25704,25860,25891,25892,25985,26005,26054,26105,26135,26165,26362,26365,26571,26572,26592,26646,26869,26953,27071,27117,27177,27179,27311,27334,27338,27494,27540,27560,27600,27633,27651,27744,27899,27992,28010,28035,28053,28063,28090,28096,28117,28119,28135,28136,28159,28300,28463,28476,28499,28508,28525,28546,28548,28579,28609,28639,28648,28655,28656,28838,28899,29182,29541,29593,29605,29674,29675,29690,29854,29896,29917,29942,29992,30015,30044,30073,30102,30131,30160,30189,30218,30247,30276,30305,30334,30363,30392,30421,30450,30479,30508,30537,30566,30595,30624,30653,30682,30740,30769,30798,30827,30856,30885,30914,30943,30972,31001,31030,31059,31088,31117,31146,31175,31204,31233,31262,31291,31320,31349,31378,31407,31436,31465,31494,31523,31552,31581,31610,31639,31668,31697,31726,31755,31784,31813,31842,31871,31900,31929,31958,31987,32016,32045,32103,32132,32161,32190,32219,32248,32277,32306,32335,32364,32393,32422,32451,32480,32509,32538,32567,32596,32625,32654,32683,32712,32741,32770,32799,32828,32857,32886,32915,32944,32973,33002,33031,33060,33089,33118,33147,33176,33205,33234,33263,33292,33321,33350,33379,33408,33437,33466,33524,33553,33582,33611,33640,33669,33698,33727,33756,33785,33814,33843,33872,33901,33959,33988,34017,34046,34075,34104,34133,34162,34191,34220,34249,34278,34307,34336,34365,34394,34423,34452,34481,34510,34539,34568,34597,34626,34655,34684,34713,34742,34771,34800,34829,34858,34887,34916,34945,34974,35015,35046,35059,35073,35104,35127,35155,35178,35214,35240,35267,35341,35415,35440,35486,35534,35593,35614,35652,35725,35784,35843,35864,35914,35973,36017,36032,36056,36069,36095,36155,36190,36214,36251,36285,36344,36376,36403,36475,36497,36534,36558,36593,36621,36626,36657,36716,36732,36733,36739,36775,36846,36905,36964,36979,36984,36997,37036,37095,37126,37154,37218,37235,37245,37277,37336,37407,37446,37466,37505,37525,37550,37582,37656,37677,37685,37715,37733,37779,37797,37838,37897,37938,37968,38027,38086,38102,38158,38182,38217,38237,38276,38340,38360,38399,38443,38458,38477,38529,38588,38622,38647,38719,38738,38778,38796,38799,38837,38901,38935,38960,39019,39090,39149,39208,39280,39339,39398,39424,39429,39430,39462,39502,39521,39537,39544,39580,39604,39651,39677,39710,39769,39784,39786,39900,39959,39977,40112,40141,40170,40228,40257,40286,40315,40344,40373,40402,40431,40489,40518,40576,40605,40634,40663,40692,40721,40750,40779,40808,40837,40866,40895,40924,40953,40982,41011,41040,41069,41098,41127,41156,41185,41214,41243,41272,41301,41330,41359,41388,41417,41446,41475,41504,41533,42004,42033,42062,42091,42120,42156,42180,42270,42298,42324,42345,42386,42420,42477,42507,42534,42551,42562,42582,42657,42744,42819,42837,42894,42931,42954,42965,42981,43056,43077,43094,43095,43108,43109,43131,43243,43244,43293,43368,43455,43530,43605,43692,43744,43767,43818,43842,43929,44004,44065,44079,44096,44098,44117,44130,44150,44166,44218,44241,44264,44316,44333,44358,44403,44429,44432,44441,44478,44498,44527,44553,44640,44715,44776,44790,44838,44868,44877,44931,44938,44952,44983,44985,45006,45027,45051,45063,45066,45087,45100,45133,45148,45209,45267,45289,45294,45303,45340,45397,45455,45527,45560,45584,45642,45670,45714,45771,45829,45922,45940,45958,45983,46028,46090,46123,46173,46202,46278,46287,46288,46325,46352,46372,46380,46383,46397,46483,46511,46544,46582,46654,46658,46758,46769,46808,46825,46828,46933,46981,47105,47194,47204,47223,47337,47358,47409,47451,47485,47521,47531,47570,47572,47580,47600,47620,47687,47710,47805,47856,47876,47878,47890,47934,47943,48032,48059,48068,48110,48133,48137,48144,48175,48231,48240,48328,48357,48362,48405,48424,48428,48432,48441,48532,48537,48539,48556,48582,48606,48640,48641,48651,48671,48678,48722,48742,48831,48930,48938,48996,49016,49052,49076,49096,49097,49138,49140,49153,49155,49162,49167,49193,49209,49260,49273,49294,49327,49335,49359,49375,49385,49433,49454,49459,49504,49537,49551,49559,49565,49567,49594,49602,49618,49659,49666,49677,49690,49734,49764,49802,49826,49906,49922,49964,49975,50022,50066,50290,50324,50385,50402,50478,50558,50570,50574,50703,50839,50911,50950,50953,51006,51069,51079,51108,51114,51121,51268,51353,51449,51454,51547,51574,51644,51661,51707,51795,51817,52091,52097,52162,52276,52399,52413,52465,52482,52681,52735,52747,52936,52995,53006,53012,53250,53270,53333,53550,53559,53582,53591,53652,53805,53838,54020,54156,54212,54231,54306,54638,54761,54777,54802,54815,55060,55079,55120,55284,55556,55608,55733,55735,55741,55782,55793,55947,55951,55986,55991,56010,56021,56041,56090,56148,56227,56288,56411,56647,56675,56805,56854,56871,56882,56932,56958,57283,57343,57352,57401,57472,57520,57883,57978,58008,58051,58068,58074,58090,58162,58419,58716,58764,58791,58833,58949,58989,59016,59062,59113,59181,59198,59338,59366,59528,59763,59903,60076,60090,60238,60263,60315,60342,60436,60483,60510,60543,60596,60624,60765,60809,60884,60906,60943,61104,61147,61159,61188,61209,61221,61229,61230,61281,61536,61573,61748,61784,61804,61834,61997,62343,62420,62509,62544,62629,62667,62789,62924,62984,63045,63107,63253,63255,63477,63527,63546,63552,63565,63704,63813,63859,63996,64045,64278,64409,64416,64638,64674,64873,64931,64958,64987,65035,65049,65168,65253,65341,65419,65451,65524,65818,65831,65861,65867,65908,65921,66085,66120,66173,66440,66511,66608,66625,66653,66654,66793,66818,66908,66955,67043,67139,67223,67226,67257,67408,67607,67634,67657,67698,67902,68057,68126,68444,68513,68629,68656,68674,68694,68733,68759,68876,68911,68964,69021,69080,69090,69129,69155,69249,69281,69304,69415,69416,69432,69582,69637,69709,69797,69941,69942,70089,70103,70172,70255,70297,70331,70335,70421,70732,70786,70914,70932,71061,71236,71243,71258,71454,71535,71558,71586,71640,71737,71816,71852,71897,71904,71964,71976,72137,72280,72348,72425,72521,72638,72786,72820,72932,72966,73127,73295,73321,73343,73621,73641,73697,73782,74003,74115,74157,74426,74470,74517,74546,74645,74664,74802,74805,74863,74879,74920,74958,75042,75281,75328,75336,75423,75449,75484,75741,75822,75861,76085,76289,76486,76686,76693,76709,76804,76912,76955,76958,77023,77056,77104,77108,77208,77308,77398,77455,77558,77595,77651,77736,77786,77894,78003,78009,78029,78055,78058,78060,78095,78296,78313,78346,78673,78773,78778,78947,79081,79082,79100,79113,79277,79317,79354,79363,79462,79518,79529,79594,79676,79722,79738,79786,79906,79948,80014,80046,80096,80256,80342,80394,80401,80644,80764,80849,80896,80993,81220,81225,81230,81353,81381,81382,81516,81638,81764,82016,82041,82051,82098,82371,82420,82436,82511,82526,82649,82794,82874,82992,83119,83172,83224,83349,83503,83523,83596,83658,83809,83817,83873,83914,83992,84032,84048,84372,84550,84611,84613,84715,84832,84902,84963,85038,85097,85133,85225,85317,85552,85854,85968,86076,86143,86520,86564,86609,86618,86633,86638,86654,86897,86998,87031,87096,87225,87245,87286,87434,87455,87480,87709,87818,88080,88271,88283,88353,88418,88441,88511,88648,88708,88756,88816,88959,88987,89163,89212,89213,89222,89243,89265,89374,89423,89520,89590,89592,89649,89650,89674,89684,89696,89701,89742,89884,90000,90091,90099,90156,90224,90254,90284,90544,90642,90693,90698,90806,90919,91036,91120,91139,91156,91246,91284,91299,91358,91422,91524,91657,91700,91872,91885,91965,91984,92179,92215,92328,92348,92444,92558,92643,92646,92787,92857,92938,92988,93122,93170,93185,93306,93340,93347,93472,93618,93682,93751,93857,93985,94011,94160,94595,94630,94734,94902,94928,95017,95094,95248,95312,95313,95336,95543,95592,95609,95620,95886,96088,96123,96143,96408,96446,96568,96599,96639,96716,96772,96785,96818,96868,96869,96877,96975,96977,97232,97381,97511,97632,97797,97820,97841,97884,98097,98250,98302,98335,98386,98499,98653,98675,98711,98749,98781,98973,99048,99136,99143,99178,99282,99393,99447,99509,99707,99974,99995",
          "39,166,173,245,249,1672,1730,1787,1958,1996,2110,2152,2278,2846,3113,3150,3497,3641,3757,3814,3817,3845,4010,4020,4089,4275,4304,4311,4479,4483,4519,4783,4819,4992,4998,5602,5659,8611,9366,9382,9387,9482,9521,9539,9558,9596,9653,9672,9691,9710,9805,9930,10014,10063,10071,10119,10185,10261,10603,10641,10698,10717,10754,10755,10774,10793,10812,10925,10945,10969,11059,11078,11135,11154,11173,11192,11230,11249,11270,11282,11306,11382,11399,11439,11456,11496,11515,11534,11553,11629,11648,11686,11876,11880,11895,11896,11952,11971,12028,12039,12077,12104,12123,12161,12180,12237,12256,12275,12294,12484,12494,12503,12560,12584,12598,12617,12636,12655,12693,12712,12750,12769,12771,12774,12834,12845,12902,12959,12997,13035,13054,13073,13187,13244,13269,13282,13301,13377,13422,13434,13466,13509,13548,13550,13566,13571,13605,13624,13643,13681,13700,13738,13795,13814,13846,13852,13928,13966,14004,14023,14137,14138,14156,14194,14213,14251,14270,14289,14327,14384,14426,14468,14479,14517,14536,14574,14593,14612,14631,14650,14654,14707,14783,14807,14824,14840,14845,14875,15030,15049,15144,15217,15220,15223,15277,15296,15353,15372,15391,15404,15410,15429,15433,15486,15556,15581,15600,15638,15657,15676,15733,15752,15771,15799,15809,15828,15856,15866,15885,15911,15923,15961,15965,15980,15983,16056,16075,16132,16137,16189,16208,16223,16227,16265,16284,16303,16322,16341,16360,16379,16398,16417,16436,16474,16531,16569,16607,16645,16721,16740,16759,16854,16911,16930,16945,16949,16987,17025,17044,17063,17101,17110,17120,17148,17158,17196,17215,17273,17288,17289,17294,17329,17348,17367,17386,17424,17462,17489,17519,17538,17576,17595,17669,17671,17728,17766,17769,17785,17804,17823,17842,17861,17899,17906,17918,17937,18013,18032,18127,18165,18184,18198,18241,18279,18298,18317,18355,18374,18393,18412,18431,18450,18488,18507,18564,18602,18678,18697,18754,18773,18792,18811,18868,18925,18949,18963,18987,19001,19020,19039,19077,19096,19111,19115,19210,19229,19267,19286,19343,19362,19365,19390,19400,19495,19514,19571,19628,19685,19704,19742,19761,19780,19803,19856,19875,19894,19932,19951,19970,20008,20027,20065,20084,20103,20118,20141,20179,20217,20262,20274,20312,20350,20369,20407,20445,20464,20483,20502,20540,20578,20635,20654,20673,20711,20730,20777,20825,20842,20844,20863,20865,20882,20958,20977,20999,21058,21059,21091,21094,21129,21194,21205,21224,21319,21357,21376,21414,21452,21471,21547,21585,21604,21680,21718,21737,21756,21785,21794,21813,21832,21851,21908,21927,21946,21965,21984,22022,22079,22095,22117,22136,22193,22250,22269,22307,22326,22383,22402,22421,22440,22459,22478,22497,22516,22535,22554,22565,22573,22611,22628,22630,22687,22706,22725,22744,22782,22801,22818,22819,22877,22886,22896,22953,22972,23067,23105,23109,23124,23162,23181,23238,23276,23279,23295,23305,23314,23352,23390,23407,23409,23504,23523,23542,23565,23599,23626,23637,23656,23713,23732,23770,23780,23808,23865,23875,23884,23903,23922,23941,23960,23979,23998,24017,24055,24112,24131,24169,24207,24226,24240,24302,24357,24359,24376,24378,24416,24435,24454,24473,24492,24511,24530,24549,24606,24625,24649,24701,24739,24758,24777,24796,24834,24853,24910,25005,25062,25072,25081,25138,25157,25161,25195,25233,25252,25271,25309,25328,25347,25423,25442,25461,25537,25544,25556,25594,25613,25702,25708,25727,25757,25765,25784,25803,25848,25861,25879,25898,25917,25936,25955,25974,25993,26012,26031,26050,26069,26145,26212,26221,26240,26241,26297,26316,26317,26354,26363,26373,26380,26392,26411,26430,26506,26639,26658,26675,26772,27000,27019,27028,27113,27228,27344,27357,27361,27380,27440,27456,27478,27513,27524,27563,27646,27783,27836,27848,27898,27902,27958,27969,28015,28026,28102,28216,28293,28501,28596,28691,28710,28729,28756,28816,28824,28924,29004,29033,29052,29058,29075,29095,29158,29172,29299,29375,29418,29419,29510,29575,29660,29699,29717,29719,29774,29831,29870,29950,30001,30030,30059,30088,30117,30146,30175,30204,30233,30262,30320,30349,30378,30407,30436,30465,30494,30523,30552,30572,30581,30610,30639,30668,30697,30726,30755,30784,30813,30842,30871,30900,30929,30958,30987,31016,31045,31074,31103,31132,31161,31190,31219,31248,31277,31306,31335,31364,31393,31422,31451,31480,31509,31538,31567,31596,31625,31654,31712,31741,31770,31799,31828,31857,31886,31915,31944,31973,32002,32031,32060,32089,32118,32147,32176,32205,32234,32263,32292,32321,32350,32379,32408,32437,32466,32495,32524,32553,32582,32611,32640,32669,32698,32727,32756,32785,32814,32843,32872,32901,32930,32988,33017,33046,33075,33104,33133,33162,33191,33220,33249,33278,33307,33336,33365,33394,33423,33452,33481,33510,33539,33568,33597,33626,33655,33684,33713,33742,33771,33800,33829,33858,33887,33916,33945,33974,34003,34032,34061,34090,34119,34148,34177,34206,34235,34264,34293,34322,34351,34380,34409,34435,34438,34467,34496,34525,34554,34583,34612,34641,34670,34699,34728,34757,34786,34815,34844,34873,34902,34931,34960,34989,35001,35076,35082,35090,35148,35164,35239,35253,35300,35327,35376,35395,35401,35472,35513,35520,35553,35573,35579,35616,35638,35711,35741,35770,35829,35900,35942,35959,36018,36071,36081,36141,36200,36271,36324,36330,36362,36365,36389,36424,36436,36441,36461,36520,36579,36643,36672,36702,36743,36761,36801,36832,36891,36950,37022,37059,37081,37140,37204,37263,37292,37306,37322,37393,37452,37511,37583,37642,37701,37765,37824,37853,37883,37954,38013,38042,38047,38058,38072,38144,38203,38262,38293,38326,38374,38385,38444,38515,38574,38618,38633,38684,38705,38751,38764,38823,38887,38920,38946,38978,39005,39076,39135,39166,39167,39190,39193,39194,39266,39313,39325,39357,39384,39434,39448,39507,39566,39599,39637,39696,39755,39798,39810,39827,39886,39945,39974,40113,40142,40171,40200,40229,40258,40287,40316,40345,40374,40403,40432,40461,40490,40519,40548,40577,40606,40635,40664,40693,40722,40751,40780,40809,40838,40867,40896,40925,40954,40983,41012,41041,41070,41099,41128,41157,41186,41215,41244,41273,41302,41331,41360,41389,41418,41447,41476,41505,41534,41700,41789,41791,41793,41795,41797,41799,41801,41803,41805,41809,41811,41813,41815,41817,41819,41821,41823,41825,41827,41829,41831,41833,41835,41837,41839,41841,41843,41845,41847,41851,41853,41855,41857,41859,41861,41863,41865,41867,41869,41871,41873,41875,41877,41879,41881,41883,41885,41887,41889,41891,41893,41895,41897,41903,41905,41907,41909,41911,41913,41915,41917,41919,41921,41923,41927,41930,41933,41936,41939,41942,41945,41948,41951,41954,41958,41962,41966,41970,41974,41978,41982,41986,41991,42020,42049,42078,42107,42136,42155,42164,42196,42205,42209,42286,42327,42361,42374,42375,42436,42468,42523,42541,42542,42598,42673,42674,42699,42701,42719,42732,42760,42835,42910,42952,42971,43037,43072,43105,43113,43147,43152,43167,43174,43186,43204,43234,43309,43334,43338,43384,43405,43471,43485,43507,43546,43561,43576,43621,43664,43708,43718,43719,43751,43783,43800,43858,43865,43919,43945,43953,43982,44020,44035,44045,44095,44105,44182,44201,44225,44257,44292,44301,44332,44351,44384,44393,44419,44466,44494,44511,44569,44591,44656,44692,44731,44737,44750,44806,44808,44818,44843,44845,44893,44929,44935,44939,44968,45043,45052,45054,45056,45061,45070,45073,45116,45176,45232,45284,45305,45356,45361,45416,45417,45473,45479,45493,45543,45604,45609,45680,45730,45750,45791,45850,45864,45867,45917,45929,45978,46001,46007,46008,46023,46077,46144,46177,46243,46300,46348,46394,46398,46404,46410,46482,46487,46520,46527,46657,46736,46745,46806,46863,46878,46887,46914,46963,46967,46988,47050,47063,47091,47111,47121,47122,47143,47159,47228,47319,47320,47329,47339,47356,47366,47405,47421,47436,47437,47480,47481,47517,47564,47568,47574,47581,47594,47643,47678,47694,47713,47773,47783,47790,47792,47793,47819,47828,47829,47885,47917,47937,47941,47981,48010,48025,48028,48062,48113,48123,48126,48134,48150,48154,48161,48229,48230,48232,48304,48313,48317,48367,48393,48398,48401,48451,48489,48513,48584,48588,48605,48658,48665,48682,48706,48716,48746,48776,48790,48794,48809,48814,48817,48819,48836,48863,48920,49172,49189,49196,49219,49246,49274,49296,49346,49395,49403,49443,49501,49519,49530,49599,49627,49707,49737,49742,49767,49847,49852,49873,49882,49953,49957,49991,49992,50029,50188,50195,50314,50373,50410,50417,50455,50470,50583,50735,50858,50894,50981,51213,51290,51405,51414,51441,51544,51571,51647,51672,51701,51761,51782,51798,52108,52175,52194,52196,52346,52425,52437,52469,52497,52618,52793,52847,52880,52943,52944,52980,53123,53379,53424,53449,53573,53616,53768,53794,53798,53861,53969,54022,54027,54066,54174,54179,54226,54375,54511,54615,54795,54850,54919,54934,54967,55216,55274,55309,55316,55406,55477,55518,55534,55564,55580,55615,55664,55723,55769,55846,55887,55958,56078,56125,56252,56280,56331,56338,56427,56461,56522,56773,57037,57088,57107,57128,57162,57230,57287,57300,57323,57334,57358,57382,57524,57542,57560,57679,57726,57745,57804,57914,58349,58350,58393,58476,58611,58616,58698,58793,58841,58942,59130,59188,59294,59525,59575,59686,59715,59772,59776,59822,59886,59957,59980,60011,60061,60087,60378,60516,60579,60593,60673,60691,60722,60849,61142,61341,61375,61377,61428,61567,61588,61674,61703,61843,61870,62197,62219,62335,62641,62706,62757,62948,63040,63056,63076,63109,63184,63392,63395,63427,63696,63724,63895,64359,64369,64544,64614,64681,64781,64842,64846,64854,64868,65008,65194,65203,65208,65263,65469,65488,65564,65789,65823,65841,65906,65937,65968,65977,66008,66063,66095,66142,66154,66163,66465,66492,66552,66573,66678,66757,66791,66800,66885,67287,67405,67464,67539,67696,67837,67895,68107,68115,68361,68364,68378,68457,68515,68569,68601,68611,68641,68684,68862,68935,69335,69523,69766,69782,69792,70044,70269,70309,70339,70363,70634,70720,70918,70958,70973,71074,71098,71153,71215,71273,71290,71322,71476,71567,71574,71691,71696,71697,71729,71784,71825,71827,71960,71994,72081,72167,72349,72745,72833,72862,72975,73005,73061,73071,73207,73212,73486,73507,73614,73618,73841,73880,74080,74141,74144,74269,74308,74393,74417,74455,74506,74558,74561,74746,74978,75095,75173,75333,75413,75509,75628,75758,75790,75811,75969,75980,76333,76386,76429,76432,76506,76547,76586,76667,76689,76742,76841,76896,77053,77350,77480,77614,77643,77649,77832,77862,77905,78008,78074,78102,78148,78190,78258,78287,78318,78351,78392,78563,78689,78791,78796,78803,78845,78870,79044,79175,79236,79414,79469,79557,79559,79664,79685,79797,79925,79989,80055,80097,80144,80165,80248,80302,80410,80454,80682,80705,80720,80893,80910,80959,81101,81161,81444,81550,81623,81639,81698,81737,81767,81834,81869,81870,81873,81991,82026,82079,82103,82119,82163,82181,82281,82476,82534,82557,82579,82595,82604,82628,82675,82761,82924,83276,83502,83543,83567,83601,83630,83729,83797,83862,83890,83974,83981,84092,84224,84397,84419,84489,84494,84530,84575,84585,84587,84648,84754,84875,85006,85648,85693,85799,85831,85853,85913,85922,86058,86353,86397,86436,86531,86605,86717,86825,87249,87287,87384,87391,87712,88022,88067,88227,88242,88245,88539,88587,88769,88843,88974,89074,89123,89132,89208,89289,89503,89714,89733,89751,89765,89895,90104,90819,91026,91125,91128,91130,91334,91374,91417,91659,91724,91731,91743,91767,91793,91832,91839,91985,92031,92087,92248,92297,92303,92786,93027,93339,93399,93533,93553,93560,93602,93609,93747,93994,94091,94117,94195,94252,94418,94784,94887,94909,95036,95182,95264,95578,95664,95748,95942,95967,96048,96132,96162,96273,96306,96442,96505,96675,96793,96803,96882,97155,97167,97299,97433,97435,97509,97629,97637,97710,97741,97770,97917,97984,98104,98160,98219,98220,98280,98403,98474,98566,98632,98721,98753,98882,98885,98920,98965,99021,99114,99144,99168,99365,99377,99474,99525,99545,99950,99983",
          "12,42,72,145,188,237,248,498,947,1646,1648,1661,1733,1754,1796,1797,1810,1924,1955,1984,2016,2025,2034,2042,2046,2056,2060,2103,2159,2163,2182,2282,2305,2332,2351,2357,2363,2375,2379,2381,2390,2400,2404,2474,2490,2499,2504,2507,2508,2519,2573,2580,2593,2624,2633,2646,2660,2714,2721,2747,2753,2758,2764,2788,2837,2844,2845,2867,2891,2892,2922,2923,2927,2938,2979,2992,3018,3031,3040,3077,3088,3096,3157,3162,3185,3207,3219,3246,3255,3320,3323,3326,3335,3359,3364,3400,3418,3419,3424,3430,3437,3455,3459,3466,3470,3483,3534,3535,3541,3548,3569,3585,3592,3594,3610,3631,3639,3649,3689,3716,3717,3719,3736,3745,3758,3760,3761,3767,3781,3783,3796,3807,3835,3839,3840,3851,3857,3858,3864,3880,3893,3927,3944,3966,3983,3985,3986,4001,4014,4024,4054,4055,4091,4116,4119,4169,4201,4223,4231,4255,4258,4262,4265,4287,4303,4322,4349,4375,4408,4449,4456,4470,4500,4523,4529,4565,4569,4614,4650,4675,4683,4685,4710,4719,4799,4826,4875,4898,4901,4905,4916,4972,4980,4989,5304,5610,5633,5638,6490,6527,6534,6561,6563,6601,6619,6629,6696,6806,6841,6858,6862,6871,6882,6891,6915,6922,6925,6950,6968,6993,7150,7183,7209,7249,7316,7332,7351,7394,7416,7428,7456,7586,7602,7681,7696,7732,7766,7805,7846,7848,7864,7919,7931,7956,7989,8075,8203,8261,8266,8304,8347,8386,8402,8407,8522,8706,8802,8882,8996,9066,9091,9115,9212,9223,9250,9271,9274,9405,9434,9487,9526,9543,9582,9622,9761,9880,10072,10136,10183,10193,10257,10559,10579,10715,10855,10958,11019,11167,11322,11398,11508,11651,11656,11688,11762,11868,11985,12205,12270,12531,12594,12646,12722,12763,12790,12795,12975,13065,13306,13392,13442,13709,13881,13987,14036,14103,14227,14325,14376,14604,14837,14890,14924,15029,15078,15110,15112,15286,15355,15482,15491,15535,15572,15751,16014,16151,16425,16643,16683,16695,16748,16754,16827,16938,17136,17213,17272,17461,17579,17778,17835,17872,18005,18088,18089,18291,18377,18607,18654,18717,19301,19531,19611,19743,19805,19829,19858,19919,20010,20236,20385,20392,20428,20467,20690,20703,20761,20818,20927,21012,21210,21259,21381,21673,21953,21972,22034,22062,22157,22309,22395,22445,22503,22577,22638,22843,22956,23133,23448,23467,23555,23674,23695,23806,23820,23856,23912,23943,24035,24036,24071,24152,24599,24665,24866,24929,25409,25441,25478,25506,25646,25674,25804,25894,26077,26234,26541,26614,26638,26719,26721,26840,26940,27041,27051,27053,27065,27076,27118,27149,27266,27299,27373,27381,27396,27410,27442,27468,27469,27472,27473,27488,27493,27508,27511,27517,27559,27565,27566,27573,27595,27609,27619,27661,27674,27679,27760,27775,27917,27955,27978,27993,27999,28011,28013,28018,28022,28048,28065,28067,28081,28106,28114,28115,28143,28150,28206,28245,28261,28340,28354,28444,28445,28485,28519,28540,28556,28559,28565,28591,28619,28658,28671,28687,28713,28786,28788,28811,28825,28859,28884,28908,28917,28923,28944,28967,28990,28997,29007,29022,29109,29115,29131,29157,29167,29168,29177,29205,29210,29242,29254,29272,29321,29330,29343,29367,29368,29385,29412,29456,29547,29561,29602,29606,29626,29646,29672,29683,29708,29729,29730,29743,29761,29770,29791,29804,29813,29838,29842,29931,29932,29945,29960,29969,29980,30012,30041,30070,30099,30128,30157,30186,30215,30244,30331,30360,30389,30418,30447,30476,30505,30534,30563,30621,30650,30679,30708,30737,30766,30795,30824,30853,30882,30911,30940,30969,30998,31027,31056,31085,31114,31143,31172,31201,31230,31259,31288,31317,31346,31375,31404,31433,31462,31491,31520,31549,31578,31607,31636,31665,31694,31723,31752,31781,31810,31839,31868,31897,31926,31955,31984,32013,32042,32071,32100,32129,32158,32187,32216,32245,32274,32303,32332,32361,32390,32419,32448,32477,32506,32535,32564,32593,32622,32651,32680,32709,32738,32767,32796,32825,32854,32883,32912,32941,32970,32999,33028,33057,33086,33115,33144,33173,33202,33231,33260,33289,33318,33347,33376,33405,33463,33492,33521,33550,33579,33608,33637,33666,33695,33724,33753,33782,33811,33840,33869,33898,33927,33956,34014,34043,34072,34101,34130,34159,34188,34217,34246,34275,34304,34333,34391,34420,34449,34478,34507,34536,34565,34594,34623,34647,34652,34681,34710,34739,34768,34797,34826,34855,34884,34913,34942,34971,35000,35012,35074,35101,35175,35197,35202,35236,35264,35315,35372,35412,35483,35531,35590,35612,35649,35722,35781,35840,35911,35970,35996,36029,36092,36110,36120,36140,36152,36181,36211,36304,36341,36400,36420,36421,36472,36531,36590,36654,36688,36713,36748,36753,36772,36809,36843,36902,36961,37033,37092,37118,37151,37195,37215,37253,37254,37274,37333,37404,37423,37463,37485,37507,37522,37542,37594,37653,37712,37776,37835,37894,37936,37953,37965,38024,38083,38155,38177,38214,38235,38273,38313,38337,38396,38419,38455,38491,38526,38561,38585,38644,38664,38670,38716,38735,38775,38803,38834,38855,38924,38957,39016,39037,39087,39146,39205,39277,39336,39395,39431,39438,39459,39518,39548,39555,39561,39577,39648,39707,39766,39789,39838,39897,39956,40106,40135,40164,40193,40222,40251,40280,40309,40338,40367,40396,40425,40483,40512,40541,40570,40599,40628,40657,40686,40715,40744,40773,40802,40831,40860,40889,40918,40947,40976,41005,41034,41063,41092,41121,41150,41172,41179,41208,41237,41266,41295,41324,41353,41382,41411,41469,41498,41527,41999,42028,42057,42086,42115,42175,42255,42265,42334,42340,42387,42415,42443,42461,42502,42577,42622,42624,42635,42652,42687,42739,42768,42814,42864,42881,42889,42918,42941,42976,43001,43126,43158,43176,43190,43213,43238,43288,43363,43389,43418,43422,43450,43508,43525,43567,43584,43600,43687,43762,43837,43889,43895,43924,43955,43963,43965,43999,44032,44074,44161,44213,44236,44311,44337,44345,44369,44378,44398,44449,44460,44468,44504,44513,44535,44548,44576,44579,44596,44611,44627,44635,44679,44687,44694,44710,44757,44767,44785,44829,44947,45005,45022,45058,45095,45135,45143,45204,45262,45328,45335,45368,45372,45392,45432,45450,45475,45522,45545,45559,45579,45637,45664,45678,45709,45748,45759,45766,45824,45889,45896,45945,45953,45980,46029,46121,46201,46216,46289,46312,46323,46327,46339,46344,46376,46407,46430,46440,46492,46553,46554,46583,46613,46621,46651,46667,46720,46747,46763,46764,46771,46810,46875,46909,46931,46932,46951,46980,46989,46994,47077,47090,47100,47207,47209,47256,47292,47299,47331,47393,47441,47461,47525,47559,47615,47667,47693,47698,47705,47731,47808,47888,47910,47957,47978,47980,47995,48022,48040,48045,48115,48118,48141,48172,48200,48256,48268,48276,48356,48368,48369,48372,48383,48408,48466,48488,48507,48550,48570,48599,48638,48676,48736,48753,48759,48821,48887,48923,48958,49000,49001,49101,49103,49128,49132,49154,49164,49236,49369,49370,49384,49408,49430,49494,49497,49515,49525,49555,49570,49607,49633,49638,49703,49723,49741,49749,49815,49862,50008,50207,50222,50236,50265,50612,50764,50768,50777,50792,50842,51208,51215,51304,51379,51519,51843,52179,52240,52750,52751,52863,52978,53032,53210,53320,53490,53557,53762,53808,53963,54085,54113,54166,54187,54314,54688,54755,54809,54930,54976,55192,55344,55365,55379,55410,55473,55567,55602,55628,55644,55809,55860,55925,55967,56099,56157,56178,56202,56267,56376,56513,56682,56844,56903,56939,56941,57001,57110,57130,57145,57157,57245,57274,57282,57299,57351,57594,57631,57705,57895,58025,58033,58093,58094,58104,58106,58206,58289,58340,58370,58405,58426,58576,58595,58647,58691,58757,58811,58883,58995,59002,59079,59117,59209,59304,59309,59355,59356,59439,59444,59536,59557,59625,59627,59809,59995,60018,60144,60279,60350,60416,60426,60539,60590,60655,60695,60723,60896,60937,61195,61330,61386,61418,61499,61506,61540,61630,61916,61934,61935,61949,62004,62039,62084,62112,62179,62247,62407,62414,62493,62527,62585,62692,62721,62787,62883,62919,62923,62956,62977,63191,63270,63300,63604,63636,63711,63940,63972,64070,64224,64490,64511,64655,64771,64801,64826,64843,64929,65066,65122,65136,65250,65373,65381,65533,65563,65696,65902,66042,66245,66251,66353,66432,66480,66489,66502,66567,66601,66618,66752,66826,66931,67026,67052,67061,67146,67247,67292,67412,67422,67477,67510,67525,67583,67610,67693,67733,67777,67959,68023,68440,68496,68749,68904,69030,69235,69373,69399,69425,69458,69519,69521,69561,69634,69639,69786,69924,69987,70127,70162,70283,70390,70569,70657,70719,70739,70764,70805,70891,70902,70935,71006,71016,71028,71110,71181,71187,71284,71295,71310,71427,71448,71497,71589,71731,71930,72044,72248,72316,72328,72364,72374,72440,72634,72642,72837,72866,72914,72968,72974,73056,73221,73259,73359,73405,73493,73530,73539,73555,73606,73609,73611,73773,73792,73853,73930,73999,74032,74186,74232,74243,74435,74640,74657,74758,74798,74870,74873,74923,75293,75473,75569,75634,75642,75669,75727,75820,76176,76287,76302,76327,76342,76404,76504,76526,76622,76716,76770,76927,77162,77346,77364,77425,77505,77637,77691,77885,77969,78028,78239,78294,78604,78651,78696,78724,78766,78918,78983,79060,79144,79146,79210,79334,79355,79422,79498,79519,79537,79544,79669,79795,79828,79973,80077,80095,80204,80205,80294,80387,80642,80646,80695,80748,80751,80829,80901,80943,80974,81011,81066,81162,81197,81354,81395,81519,81636,81794,82056,82113,82127,82159,82284,82331,82381,82446,82662,82715,82771,82829,82979,83104,83123,83138,83252,83407,83413,83426,83535,83613,83727,84034,84104,84254,84337,84349,84374,84384,84540,84725,84776,84833,84962,85174,85276,85379,85532,85585,85617,85675,85780,85887,86044,86099,86164,86198,86224,86306,86355,86423,86478,86525,86532,86543,86595,86689,86691,86714,86719,86734,86802,86886,86963,86986,87094,87128,87192,87395,87653,87683,87827,87879,87891,87956,88171,88262,88268,88277,88416,88468,88480,88626,88629,88669,88789,88803,88835,89031,89100,89154,89174,89206,89226,89325,89327,89331,89573,89660,89762,89774,89777,90050,90064,90385,90402,90573,90691,90740,90914,91031,91044,91103,91205,91226,91420,91561,91606,91631,91759,91930,91951,91953,92034,92038,92118,92372,92439,92441,92512,92578,92670,92681,92830,92909,93030,93244,93458,93595,93727,94050,94222,94263,94284,94346,94362,94432,94602,94905,94964,94976,94995,95072,95079,95167,95291,95367,95422,95581,95698,95707,95901,96178,96201,96227,96276,96299,96356,96456,96501,96685,96724,96801,96825,96857,96924,97045,97054,97085,97107,97461,97479,97502,97571,97838,97970,98278,98313,98394,98495,98567,98646,98700,98701,98730,98771,98786,98796,98845,98846,99049,99128,99181,99309,99316,99592,99739,99757,99844,99906,99934,99958",
          "176,276,477,608,637,652,665,668,726,738,752,827,897,943,974,1045,1136,1153,1162,1208,1255,1267,1335,1424,1457,1473,1480,1559,1592,1597,1611,1615,1635,1644,1674,1748,1841,2030,2188,2195,2589,2638,2693,2742,2773,2792,2799,2902,2928,3046,3103,3126,3160,3179,3226,3333,3647,3667,3706,3792,3828,4198,4210,4319,4333,4354,4559,4582,4667,4671,4720,4729,4749,4945,5008,5035,5084,5110,5122,5138,5140,5178,5190,5201,5203,5205,5241,5255,5259,5298,5301,5327,5339,5388,5400,5512,5514,5578,5641,5706,5742,5745,5865,5938,5969,6003,6031,6038,6058,6148,6182,6185,6218,6229,6251,6468,6522,6548,6597,6822,7077,7080,7187,7286,7288,7299,7405,7411,7476,7486,7525,7561,7666,7714,7741,7780,7879,7894,7905,7927,8005,8029,8120,8245,8382,8501,8679,8728,8816,9077,9230,9313,9344,9374,9383,9385,9423,9428,9736,9815,9898,10003,10287,10337,10425,10439,10464,10475,10532,10687,10874,10934,11274,11378,11411,11463,11673,11746,11870,12383,12433,12470,12527,13037,13105,13361,13440,13935,13955,14063,14339,14530,14607,14828,15340,15401,15472,15804,15890,16005,16616,16714,16867,16955,17032,17054,17064,17223,17518,17871,18881,19281,19370,19479,19720,20034,20264,20408,20663,20757,20896,21499,21544,21692,21842,22521,22796,23783,23829,24452,24812,25070,25093,25171,26162,26739,26774,26901,26909,27219,27432,27525,27854,28190,28210,28229,28244,28289,28709,28728,28732,29277,29991,30333,30596,30681,31522,31696,32363,32537,33001,33291,33320,33849,34016,34422,35039,35450,35621,35993,36446,36499,37248,37382,37444,37692,37837,38128,38132,38157,38187,38304,38492,38646,38744,38959,39004,39118,39173,39315,39360,39557,39626,39679,39732,39875,39992,41517,42103,42493,42610,42932,43164,43467,43519,43520,43654,44260,44585,45231,45370,45437,45556,45558,45732,45737,45874,45879,45919,45998,46126,46984,47982,48084,49754,49849,49996,50104,50106,50119,50251,50252,50341,50433,50511,50673,50718,50733,50814,50827,50985,51090,51138,51178,51206,51242,51265,51359,51705,51942,51990,52203,52257,52323,52325,52412,52427,52531,52567,52636,52663,52799,53025,53191,53349,53595,53677,53702,54261,54455,54464,54496,54660,54715,54889,54988,55095,55117,55121,55147,55300,55393,55498,55549,55647,55690,55703,55715,55795,56089,56105,56253,56293,56396,56412,56470,56550,56789,57010,57030,57047,57070,57080,57200,57206,57293,57373,57422,57481,57494,57535,57571,57893,57896,57969,57977,58092,58115,58361,58496,58524,58557,58799,58859,58879,58898,58915,58924,59042,59058,59204,59226,59446,59549,59566,59798,59837,59865,59872,59979,60002,60219,60399,60403,60454,60476,60634,60733,60744,60848,60914,61004,61066,61077,61100,61134,61224,61235,61296,61361,61425,61445,61578,61796,61858,62067,62164,62246,62277,62288,62366,62368,62401,62408,62555,62656,62962,62995,63026,63028,63042,63090,63205,63282,63349,63401,63410,63507,63554,63611,63621,63683,63735,63761,63826,63860,63907,63988,64003,64077,64106,64143,64227,64277,64523,64528,64580,64786,64829,64849,64968,64978,64981,65054,65061,65121,65237,65437,65452,65656,65709,65784,65925,66055,66167,66216,66226,66312,66347,66430,66515,66577,66580,66603,66662,66805,66888,67116,67215,67262,67415,67435,67751,67888,67896,67950,68015,68071,68137,68197,68483,68527,68608,68628,68789,68805,68842,68892,68916,69182,69185,69216,69264,69481,69495,69567,69568,69789,69821,69881,69959,70068,70078,70087,70126,70138,70211,70276,70281,70310,70315,70378,70415,70497,70505,70552,70616,70665,70846,70936,70984,70994,70998,71029,71108,71156,71373,71378,71423,71513,71561,71580,71668,71678,71822,71828,71895,72014,72043,72116,72128,72141,72204,72252,72486,72497,72552,72555,72579,72657,72732,72890,73001,73291,73301,73326,73333,73496,73497,73756,73941,73964,74272,74357,74638,74730,74929,75085,75096,75134,75140,75170,75192,75274,75302,75332,75427,75435,75472,75584,75606,75718,75792,75866,75912,75976,76014,76077,76185,76264,76286,76366,76450,76502,76561,76578,76650,76849,76850,76892,76915,77170,77177,77199,77220,77242,77302,77370,77400,77452,77473,77650,77684,77733,77769,78402,78515,78545,78603,78740,78754,78988,79008,79059,79128,79133,79158,79196,79284,79338,79421,79439,79488,79533,79752,79791,79867,79936,79969,80026,80154,80171,80197,80208,80210,80229,80270,80369,80699,80886,80890,81003,81010,81025,81093,81309,81412,81423,81483,81813,81960,82152,82171,82183,82289,82333,82396,82405,82684,82693,82766,82781,82911,82921,82959,83092,83233,83358,83364,83400,83422,83562,83646,83700,83711,83807,83878,83952,84009,84047,84067,84157,84292,84368,84552,84621,84794,84796,84864,85109,85279,85376,85381,85472,85515,85529,85716,85729,85784,85807,85875,85935,86120,86123,86162,86331,86388,86391,86407,86723,86805,86837,86863,86883,86905,86924,86989,87376,87379,87478,87498,87575,87634,87681,87937,88039,88063,88202,88220,88254,88311,88399,88408,88411,88430,88526,88614,88639,88652,88659,88696,88870,89178,89553,89564,89641,89663,89863,90012,90114,90192,90279,90449,90560,90627,90760,90788,90880,90886,91002,91168,91364,91563,91564,91600,91626,91641,91673,91856,91883,91896,91971,92019,92020,92183,92243,92445,92616,92752,92777,92979,93013,93062,93114,93151,93262,93680,93838,94051,94106,94114,94273,94422,94456,94529,94540,94596,94718,94863,95018,95070,95438,95616,95768,95797,95869,95947,95979,95984,96035,96083,96163,96190,96208,96253,96339,96355,96384,96416,96450,96480,96498,96723,96880,97033,97059,97110,97180,97193,97273,97377,97429,97619,97626,97842,97851,98029,98061,98157,98236,98329,98346,98372,98430,98516,98699,98788,98892,98998,99051,99112,99137,99226,99347,99389,99439,99535,99548,99588,99646,99673,99755,99762,99812,99901,99926",
          "13,284,288,351,360,445,453,541,557,612,680,695,737,758,770,775,798,801,855,902,963,964,971,983,1049,1129,1300,1332,1453,1461,1472,1536,1543,1564,1753,1802,1812,1831,1855,1898,1905,1908,2816,3158,3949,4759,5071,5148,5236,5411,5588,5627,5723,5831,5963,6131,6411,6566,6620,6712,6845,6962,7000,7060,7071,7086,7112,7126,7190,7203,7278,7321,7341,7376,7390,7435,7507,7514,7526,7531,7557,7569,7654,7660,7672,7701,7711,7745,7756,7785,7800,7858,7967,8027,8037,8065,8070,8110,8213,8257,8267,8303,8306,8334,8344,8372,8484,8513,8637,8746,8807,8856,8908,8934,9043,9073,9415,9440,9603,9890,9958,10026,10125,10144,10209,10284,10382,10506,10650,10658,10732,10920,11132,11431,11500,11605,11773,11823,11842,11893,12075,12110,12230,12379,12540,12614,12728,12832,12878,12976,13299,13582,13825,13948,14118,14387,14609,14624,14708,15136,15231,15987,16091,16837,17121,17287,17388,18473,18561,18659,19266,20090,20137,20696,20857,20952,21248,21337,21527,21687,22039,22143,22616,22854,22975,23317,23397,23562,23638,24033,24132,24161,24685,24848,25182,25362,25575,25611,25806,26617,26701,26791,26842,26853,27243,27647,27705,27742,27756,27861,28248,28267,28371,28431,29341,29361,29431,29578,29953,32367,34221,35516,36116,36173,36264,36694,36830,36981,36993,37052,37198,37320,37429,37504,38136,38196,38244,38700,39444,39560,39671,39884,42034,44367,45127,45297,45360,45924,45975,48882,50003,50016,50077,50218,50282,50309,50367,50666,50669,50696,50727,50857,50943,50966,50997,51000,51086,51093,51181,51237,51333,51389,51400,51455,51561,51626,51671,51711,51774,51794,51842,51853,51866,51891,52199,52553,52573,52670,52748,52827,52855,52946,52952,53058,53189,53305,53347,53375,53378,53380,53409,53507,53669,53911,54269,54352,54362,54364,54433,54449,54486,54563,54583,54665,54764,54948,55002,55022,55037,55233,55247,55445,55447,55485,55670,55799,55857,55868,55875,56032,56112,56124,56189,56348,56356,56365,56378,56730,56870,56961,56981,57100,57152,57216,57259,57316,57339,57366,57468,57525,57536,57575,57608,57686,57783,57840,57849,57879,57892,57902,58046,58054,58137,58156,58216,58224,58246,58251,58294,58417,58522,58594,58599,58678,58682,58735,58801,58810,58903,59164,59350,59472,59521,59810,59871,59927,59985,60245,60305,60340,60354,60544,60638,60648,60791,60919,60973,61272,61297,61480,61483,61493,61642,61659,61780,61897,61913,61953,62426,62433,62483,62674,62679,62693,62964,63108,63428,63499,63726,63741,63769,63983,63992,64085,64212,64229,64282,64306,64496,64597,64700,64813,64872,64902,64917,65048,65245,65254,65308,65487,65506,65534,65598,65621,65646,65648,65666,65682,65735,65808,65873,65881,66129,66439,66466,66550,66821,66829,66835,66853,66934,67070,67218,67336,67342,67377,67410,67425,67613,67976,67994,68232,68408,68495,68523,68710,68743,68928,69006,69057,69108,69135,69199,69277,69332,69461,69484,69502,69535,69538,69558,69566,69573,69625,69769,69785,69819,69911,69956,70030,70045,70086,70219,70249,70347,70517,70628,70673,70729,70818,71009,71022,71173,71205,71316,71341,71362,71376,71628,71763,72021,72048,72072,72136,72214,72227,72286,72291,72507,72772,72839,72859,72878,72928,72930,72936,73185,73563,73718,73905,73935,73938,73984,74024,74034,74045,74082,74088,74289,74373,74665,74713,74924,74927,74984,75086,75108,75167,75196,75366,75429,75470,75513,75688,75833,75951,76001,76018,76145,76241,76323,76369,76483,76495,76605,76616,76708,76807,76832,76856,76956,76986,77025,77033,77126,77176,77255,77293,77307,77423,77549,77609,77647,77949,78042,78076,78169,78188,78222,78230,78264,78281,78300,78398,78456,78934,79046,79189,79219,79318,79337,79361,79500,79538,79563,79576,79582,79592,79615,79892,79957,79959,79985,80016,80039,80047,80049,80101,80157,80262,80335,80617,80637,80852,80891,81049,81064,81074,81114,81181,81296,81375,81388,81408,81448,81584,81624,81969,82108,82110,82133,82202,82232,82239,82243,82337,82670,82681,82699,82744,82760,82899,83208,83449,83488,83511,83590,83660,83769,83827,83916,83960,84051,84080,84232,84350,84405,84442,84479,84486,84509,84604,84618,84632,84738,84779,84809,84829,84836,84886,85140,85142,85269,85329,85363,85369,85504,85667,85672,85714,85816,85865,85965,86047,86052,86109,86214,86235,86248,86455,86536,86671,86792,86918,86985,87005,87006,87084,87141,87158,87200,87243,87305,87463,87528,87654,87805,87857,87940,88000,88013,88286,88304,88358,88423,88500,88504,88632,88663,88682,88784,88849,88933,89055,89127,89148,89205,89569,89605,89620,89671,89697,89735,89771,89773,89802,89838,89910,89993,90081,90305,90409,90444,90479,90537,90555,90559,90666,90695,90718,90859,91086,91173,91298,91410,91542,91555,91825,91865,91869,91870,91916,91959,92017,92059,92126,92298,92353,92477,92623,92652,92669,92790,92866,92882,92928,93016,93024,93052,93135,93231,93409,93459,93693,93714,93996,93999,94059,94100,94177,94583,94609,94686,94723,94725,94733,94770,94776,94859,94898,94912,94942,95090,95131,95145,95388,95426,95483,95761,95918,95975,95992,96012,96478,96542,96623,96733,96755,96858,97018,97057,97103,97133,97248,97354,97441,97463,97570,97594,97611,97622,97781,97795,97817,97931,97964,98179,98189,98231,98296,98556,98647,98906,98976,99002,99074,99120,99154,99236,99404,99408,99441,99460,99485,99697,99705,99738,99782,99981",
          "61,69,76,90,128,130,1660,1684,1722,1760,1821,1925,2160,2216,2306,2513,2657,2722,2789,2834,2861,2957,2976,3072,3093,3147,3215,3232,3287,3308,3332,3612,3619,3749,3824,4078,4161,4293,4367,4374,4427,4455,4465,4469,4506,4603,4634,4649,4738,4769,4785,4787,5000,5161,6377,6659,7156,7977,8030,8049,8125,8151,8315,8429,8505,8562,8581,8638,8676,9037,9056,9075,9588,9602,9626,9664,9683,9702,9724,9778,9816,9837,9987,9988,10006,10082,10120,10177,10405,10610,10614,10633,10710,10747,10804,10828,10896,10956,10975,11032,11051,11222,11241,11298,11338,11409,11412,11465,11473,11474,11514,11523,11596,11621,11640,11735,11963,12001,12050,12058,12066,12089,12115,12153,12168,12202,12229,12248,12259,12267,12286,12343,12349,12381,12430,12457,12459,12538,12628,12697,12704,12761,12802,12875,12977,12979,12989,13028,13084,13100,13120,13170,13174,13190,13236,13241,13274,13293,13312,13350,13464,13483,13597,13616,13635,13692,13730,13749,13768,13787,13844,13882,13939,13958,13977,14034,14041,14110,14167,14205,14243,14262,14338,14346,14357,14395,14414,14433,14452,14471,14528,14557,14566,14613,14642,14680,14718,14756,14762,14775,14832,14870,14908,14912,14916,14927,14965,15003,15022,15041,15057,15060,15079,15081,15155,15193,15250,15269,15307,15326,15364,15383,15389,15421,15497,15516,15592,15646,15649,15687,15744,15746,15763,15782,15820,15839,15858,15896,15929,15953,15974,15991,16029,16048,16143,16149,16176,16181,16191,16197,16200,16238,16257,16276,16295,16314,16352,16390,16409,16438,16447,16466,16507,16535,16542,16561,16580,16619,16656,16727,16751,16789,16808,16865,16903,16960,17017,17055,17112,17150,17188,17245,17295,17302,17305,17321,17378,17435,17473,17511,17549,17574,17599,17644,17672,17739,17796,17815,17834,17853,17891,17910,17929,17986,18024,18043,18119,18138,18157,18213,18214,18290,18309,18347,18366,18385,18404,18480,18518,18556,18613,18651,18689,18716,18746,18774,18791,18860,18879,18898,18904,18999,19031,19050,19063,19107,19126,19141,19221,19240,19278,19322,19339,19354,19368,19373,19392,19430,19441,19449,19506,19525,19537,19553,19561,19563,19582,19590,19620,19658,19734,19748,19753,19772,19791,19867,19886,19905,19924,20000,20019,20026,20038,20058,20101,20114,20133,20190,20247,20285,20304,20323,20342,20418,20456,20494,20551,20564,20589,20627,20633,20646,20649,20665,20741,20779,20817,20836,20893,20961,20969,21007,21064,21083,21178,21183,21216,21257,21270,21292,21301,21313,21330,21340,21349,21387,21425,21444,21463,21518,21539,21577,21596,21640,21656,21691,21710,21729,21767,21786,21805,21817,21843,21900,21919,21938,21957,21990,21995,22014,22033,22052,22057,22090,22109,22130,22147,22152,22166,22204,22242,22261,22318,22324,22394,22413,22470,22489,22508,22527,22603,22641,22698,22755,22774,22793,22812,22888,22907,22912,22926,22954,23002,23040,23059,23062,23078,23116,23128,23154,23158,23173,23267,23268,23287,23306,23325,23439,23469,23477,23485,23496,23513,23515,23533,23534,23553,23572,23607,23619,23629,23650,23686,23698,23721,23724,23735,23743,23800,23838,23857,23876,23887,23895,23914,23933,23971,23990,24047,24066,24085,24142,24171,24180,24187,24188,24199,24218,24256,24289,24295,24332,24351,24389,24427,24465,24522,24541,24560,24579,24617,24636,24674,24693,24731,24769,24788,24845,24860,24876,24902,24921,24959,24977,24997,25013,25035,25047,25054,25092,25149,25168,25206,25263,25277,25320,25339,25396,25404,25407,25415,25460,25472,25491,25529,25548,25586,25616,25662,25700,25739,25758,25776,25814,25833,25852,25890,25947,26004,26023,26042,26118,26137,26460,26477,26479,26517,26593,26745,26775,26859,26896,26916,26928,26935,26946,27079,27087,27106,27144,27158,27182,27220,27258,27276,27332,27353,27486,27553,27554,27581,27594,27656,27695,27714,27727,27809,27942,27980,28037,28075,28162,28238,28375,28391,28417,28436,28490,28547,28550,28569,28714,28751,28911,29063,29099,29228,29285,29320,29329,29416,29469,29500,29519,29858,29886,29907,29994,30013,30042,30071,30100,30129,30158,30187,30216,30245,30274,30303,30332,30361,30390,30419,30448,30477,30506,30535,30564,30593,30622,30651,30680,30709,30738,30767,30796,30825,30854,30883,30912,30941,30970,30999,31028,31057,31086,31115,31144,31173,31202,31231,31260,31289,31318,31347,31376,31405,31434,31463,31492,31521,31550,31579,31608,31637,31666,31695,31724,31753,31782,31811,31840,31869,31898,31927,31985,32014,32043,32072,32101,32130,32159,32188,32246,32275,32304,32333,32362,32391,32420,32449,32478,32507,32536,32565,32594,32623,32652,32681,32710,32739,32768,32797,32826,32855,32884,32913,32942,32971,33000,33029,33058,33087,33116,33145,33174,33203,33232,33261,33290,33319,33348,33377,33406,33435,33464,33493,33522,33551,33580,33609,33638,33667,33696,33725,33754,33783,33812,33841,33870,33899,33928,33957,33986,34015,34044,34073,34102,34131,34160,34189,34218,34247,34276,34305,34334,34363,34392,34421,34450,34479,34508,34537,34566,34595,34624,34653,34682,34711,34740,34769,34798,34827,34856,34885,34914,34943,34972,35013,35102,35121,35163,35176,35237,35265,35284,35306,35320,35323,35339,35364,35413,35434,35459,35484,35514,35532,35626,35650,35668,35674,35723,35758,35782,35841,35912,35971,36030,36093,36128,36153,36192,36212,36283,36329,36342,36369,36401,36473,36504,36532,36591,36655,36714,36751,36773,36844,36903,36962,36992,37034,37093,37113,37125,37152,37173,37179,37185,37216,37334,37368,37433,37464,37523,37595,37654,37713,37741,37777,37796,37836,37895,37966,38002,38025,38063,38084,38111,38143,38156,38215,38274,38338,38397,38456,38474,38527,38586,38667,38717,38734,38763,38776,38835,38899,38958,38996,39017,39075,39088,39125,39147,39172,39206,39278,39307,39324,39337,39396,39437,39460,39496,39519,39578,39597,39605,39613,39649,39708,39860,39898,39925,39957,39986,40123,40152,40181,40210,40239,40268,40297,40326,40355,40384,40413,40442,40471,40500,40529,40558,40587,40616,40645,40674,40703,40732,40761,40790,40819,40848,40877,40906,40935,40964,40993,41022,41051,41080,41109,41167,41196,41225,41254,41283,41312,41341,41370,41399,41428,41457,41486,41515,41544,42072,42101,42130,42190,42219,42223,42229,42280,42297,42355,42390,42394,42430,42480,42487,42517,42540,42592,42604,42667,42733,42754,42807,42829,42848,42857,42904,42991,43040,43066,43091,43141,43228,43303,43350,43378,43465,43480,43540,43551,43553,43558,43564,43582,43615,43683,43702,43716,43727,43777,43852,43871,43939,43970,43986,44014,44059,44066,44089,44103,44112,44115,44132,44148,44176,44194,44211,44251,44258,44288,44326,44394,44413,44424,44488,44538,44563,44608,44650,44725,44733,44745,44769,44800,44831,44834,44887,44905,44908,44962,45037,45079,45080,45110,45158,45219,45277,45350,45382,45407,45465,45476,45537,45549,45594,45652,45724,45781,45839,45861,45911,45968,45994,46051,46060,46256,46263,46315,46333,46345,46362,46393,46405,46409,46414,46419,46442,46459,46460,46481,46496,46540,46587,46690,46757,46790,46802,46814,46824,46839,46840,46910,46934,46990,46996,47020,47021,47083,47097,47098,47123,47126,47138,47201,47244,47348,47352,47438,47440,47464,47478,47479,47492,47516,47589,47602,47642,47670,47704,47770,47842,47848,47909,47963,47965,47984,48046,48065,48211,48290,48308,48346,48365,48467,48525,48527,48535,48558,48560,48564,48601,48603,48698,48699,48734,48741,48750,48777,48782,48820,48853,48867,48872,48912,48939,48955,48959,48977,48986,49012,49020,49036,49063,49075,49146,49151,49166,49258,49266,49278,49321,49337,49349,49350,49367,49407,49411,49435,49448,49596,49612,49629,49632,49668,49672,49715,49756,49788,49790,49816,49859,49930,49944,49951,49974,49985,50151,50209,50228,50230,50247,50302,50339,50409,50474,50587,50656,50790,51013,51056,51067,51223,51307,51504,51718,51742,52015,52087,52208,52242,52292,52294,52403,52474,52501,52610,52858,52899,52991,53048,53215,53237,53245,53314,53352,53368,53451,53520,53589,53605,53619,53647,53809,53863,53864,53966,54040,54077,54082,54189,54192,54223,54295,54373,54410,54445,54485,54784,54912,54940,55142,55307,55310,55346,55388,55409,55412,55423,55449,55929,55938,55966,56110,56272,56359,56413,56473,56578,56688,56872,56881,56888,56913,56927,56989,57445,57464,57497,57500,57803,57860,57890,57949,58038,58063,58099,58129,58205,58235,58248,58266,58277,58282,58329,58389,58451,58622,58632,58670,58721,58920,58992,59094,59099,59183,59195,59266,59300,59440,59502,59685,59759,59818,59917,60032,60055,60198,60225,60307,60324,60349,60541,60652,60689,60734,60905,60956,60965,61271,61371,61488,61527,61537,61553,61663,61705,61756,61767,62016,62410,62425,62661,62688,62711,62850,62889,62925,63183,63287,63357,63659,63783,63828,63862,63950,63953,64060,64100,64107,64136,64236,64239,64290,64299,64361,64450,64467,64532,64586,64641,64671,64746,64774,64830,64928,64998,65093,65108,65159,65364,65422,65446,65547,65733,65775,65997,66266,66355,66374,66581,66701,66801,66815,66890,66948,67067,67330,67359,67388,67423,67643,67665,67708,67709,67884,67910,67938,67977,67992,68052,68236,68250,68272,68435,68554,68620,68624,68751,68756,68827,68849,68853,68938,68961,69112,69124,69223,69236,69342,69390,69392,69489,69666,69711,69837,69873,69919,69986,70059,70135,70259,70280,70358,70368,70377,70399,70538,70620,70636,70650,70660,70754,70869,70888,70950,70969,71055,71092,71282,71352,71599,71603,71707,71720,71741,71766,71780,71918,71967,72056,72165,72182,72382,72421,72458,72475,72527,72547,72631,72692,72787,72821,72826,72929,72963,73168,73532,73550,73616,73668,73830,73884,73936,74168,74325,74587,74601,74611,74667,74668,74689,74727,74753,74771,74781,74847,74888,75133,75422,75428,75549,75615,75632,75663,75678,75683,75778,75829,75851,75897,75908,75928,76038,76262,76359,76425,76494,76496,76523,76614,76669,76687,76905,76952,76974,77062,77193,77360,77446,77538,77561,77584,77585,77841,77850,77918,77923,77936,78254,78454,78459,78598,78663,78911,79067,79092,79105,79155,79351,79553,79692,79695,79771,79778,79805,79857,79875,79880,80090,80099,80151,80206,80474,80607,80782,80786,80822,81235,81241,81260,81292,81304,81311,81336,81608,81609,81615,81814,81821,81942,82019,82062,82124,82131,82166,82203,82244,82260,82285,82375,82434,82520,82617,82746,82756,82763,82847,82945,82952,82957,82993,83076,83089,83213,83216,83225,83249,83308,83314,83379,83440,83451,83539,83540,83691,83858,84003,84010,84063,84176,84242,84436,84502,84747,84749,84760,84838,84857,84993,85015,85089,85193,85364,85465,85578,85957,86049,86229,86290,86425,86549,86617,86625,86704,86726,86764,86819,86983,86984,87201,87330,87362,87368,87496,87540,87599,87609,87704,87909,88056,88074,88147,88394,88437,88446,88570,88580,88643,88657,88688,88721,88832,88913,88937,88940,88948,89164,89302,89476,89515,89522,89602,89755,89761,89877,89939,89962,89966,89977,90028,90029,90052,90060,90194,90366,90393,90398,90429,90442,90524,90596,90655,90680,90747,90838,90934,90973,90995,91155,91293,91377,91513,91540,91548,91607,91727,91820,91838,91931,92005,92285,92370,92401,92443,92505,92589,92625,92768,92773,92780,92811,92901,92945,92987,93123,93152,93212,93215,93267,93393,93556,93617,93645,93666,93741,93932,94052,94067,94075,94150,94191,94257,94354,94447,94517,94562,94642,94663,94727,94743,94785,95016,95149,95349,95724,95928,95930,96134,96191,96298,96452,96476,96520,96601,96657,96699,96778,96873,96962,97010,97064,97159,97189,97221,97277,97327,97465,97498,97520,97707,97726,97916,98030,98055,98075,98171,98202,98240,98255,98400,98639,98674,98697,98735,98999,99068,99113,99209,99219,99233,99247,99341,99361,99507,99517,99579,99634,99649,99650,99793,99878",
          "35,177,304,312,391,439,448,483,577,601,633,672,674,800,803,835,851,853,871,877,879,884,885,909,913,991,997,1003,1009,1012,1019,1025,1094,1134,1190,1231,1238,1240,1287,1294,1319,1320,1373,1382,1406,1412,1532,1851,1876,1957,2161,2219,2227,2231,2414,2447,2526,2745,2828,2887,2905,2941,3114,3155,3196,4800,5091,5097,5117,5212,5532,5570,5617,5618,5637,5640,5682,5740,5783,5850,5966,5991,5992,6029,6098,6100,6168,6193,6312,6379,6482,6687,6800,7038,7041,7152,7314,7330,7372,7521,7556,7620,7629,7639,7641,7827,8092,8171,8178,8238,8299,8326,8385,8485,8497,8570,8651,8833,9090,9098,9120,9206,9445,9891,9990,10055,10222,10250,10265,10329,10340,10416,10466,10496,10551,10564,10761,10778,10785,10852,10964,11096,11278,11778,11922,12081,12101,12320,13029,13839,13840,13897,14378,14585,14610,14733,15069,15070,15085,15131,15316,15431,15522,15999,16107,16471,16519,16664,16732,16744,16773,16776,17582,18188,18522,18683,18884,18889,19480,19627,20122,20248,20496,21039,21060,21125,21298,22179,23043,23773,24277,24679,24789,24826,24973,25124,25632,25871,25972,26112,26272,26347,26680,26694,27082,27156,27324,27572,28272,28749,28787,29191,29542,31732,33757,35706,35953,36263,36408,36676,36868,37616,37640,37951,38050,38197,38323,38511,38938,38941,39003,39191,39295,39359,39445,39691,39749,39800,39823,42275,42997,45120,45792,45808,45918,47364,50102,50134,50135,50435,50446,50580,50585,50619,50654,50676,50698,50832,50944,51031,51063,51207,51227,51327,51366,51411,51548,51593,51730,51807,51832,51884,51901,51954,51961,51996,51997,52001,52378,52415,52416,52443,52614,52624,52654,52699,52757,52816,52820,52832,52892,52921,53093,53173,53227,53635,53638,53708,53787,53807,53842,53907,53910,53971,54047,54152,54213,54277,54286,54328,54343,54347,54348,54399,54440,54526,54735,54742,54796,54826,54926,54965,55094,55190,55292,55297,55448,55609,55673,55710,55722,55727,55750,55904,55996,56087,56156,56342,56447,56598,56748,56752,56857,56999,57011,57063,57160,57311,57325,57419,57446,57447,57543,57615,57663,57784,57847,57866,57894,57932,57941,58144,58198,58228,58293,58367,58415,58540,58546,58684,58738,58762,58773,58814,58837,58994,59123,59224,59231,59315,59335,59410,59412,59425,59495,59611,59661,59709,59844,60504,60712,60816,60941,61084,61219,61293,61340,61416,61466,61591,61595,61782,61863,61866,61878,61965,61986,62015,62021,62059,62185,62195,62413,62454,62499,62530,62665,62791,62875,62912,62934,62952,63095,63169,63224,63246,63304,63437,63454,63504,64130,64182,64292,64351,64436,64453,64484,64489,64603,64688,64689,64716,64761,64825,64867,65146,65183,65204,65240,65275,65289,65390,65440,65459,65548,65587,65618,65620,65669,65679,65715,65720,65946,66205,66210,66236,66275,66371,66376,66418,66486,66704,66982,67033,67082,67135,67164,67177,67270,67306,67338,67351,67513,67682,67725,67812,67829,67932,67970,68043,68047,68051,68054,68208,68362,68383,68425,68529,68560,68660,68828,68858,68955,68991,69122,69179,69184,69237,69327,69360,69491,69526,69653,69677,69728,69869,69871,69890,69908,69993,70235,70316,70346,70404,70485,70623,70828,70899,71017,71054,71082,71111,71596,71813,71837,71972,72231,72310,72424,72452,72468,72470,72519,72533,72596,72669,72672,72684,72791,72814,72971,73066,73114,73411,73505,73520,73524,73693,73754,73838,73850,73892,73907,73949,74000,74028,74044,74208,74326,74612,74694,74733,75123,75145,75226,75379,75415,75520,75573,75602,75715,75764,75823,75837,75975,76128,76204,76207,76335,76387,76395,76477,76529,76792,76844,76909,76926,76995,77158,77164,77241,77336,77337,77447,77624,77657,77812,78123,78129,78165,78216,78260,78270,78415,78466,78511,78594,78610,78666,78725,78764,78871,78959,79086,79114,79165,79182,79184,79374,79471,79567,79598,79633,79655,79665,79762,79780,79811,79816,79879,79907,79918,79928,79943,79952,79986,80071,80201,80239,80268,80310,80362,80395,80439,80450,80615,80640,80688,80818,80871,80955,81081,81142,81148,81202,81207,81271,81368,81552,81720,81833,81868,81885,81953,82049,82054,82208,82290,82304,82360,82472,82501,82717,82856,82914,83049,83113,83144,83357,83411,83415,83439,83501,83594,83602,83640,83696,83839,83918,83934,84054,84097,84150,84166,84255,84408,84432,84481,84499,84516,84525,84607,84622,84804,84914,84991,84996,85034,85295,85584,85633,85724,85762,85874,85923,85931,85993,86222,86240,86327,86675,86880,87014,87122,87129,87218,87348,87630,87924,88104,88191,88217,88234,88327,88349,88507,88525,88528,88746,88810,88823,88841,88988,89058,89129,89282,89317,89337,89577,89633,89876,89879,89969,90108,90197,90198,90278,90358,90483,90502,90599,90612,90722,90753,90759,90810,90954,91047,91071,91271,91340,91466,91503,91584,91638,91665,91795,91834,91894,91897,91943,91972,92133,92150,92221,92388,92405,92456,92476,92660,92696,92725,92751,92759,92779,92801,92816,92842,92859,92920,92982,93079,93349,93395,93457,93461,93620,93622,93735,93928,93947,94048,94088,94159,94206,94264,94306,94361,94375,94430,94539,94565,94690,94777,94813,94876,94945,95088,95135,95204,95211,95273,95331,95357,95425,95446,95552,95572,95694,95716,95767,95868,95870,96260,96433,96541,96589,96709,96712,96725,96752,96912,96920,97378,97408,97540,97593,97664,97689,97870,97872,98081,98109,98145,98181,98224,98317,98489,98504,98813,98925,99373,99421,99553,99708,99721,99800,99806,99852,99868",
          "78,148,228,229,1649,1731,1736,1769,2035,2094,2155,2156,2230,2415,2429,2628,2679,2727,2746,2829,2852,2870,2983,3193,3214,3425,3479,3517,3574,3793,4314,4328,4342,4517,4797,4851,4858,4986,5080,6052,7524,9479,9488,9597,9711,9749,9806,9920,9977,10262,10604,10642,10680,10756,10767,10832,10889,11003,11022,11117,11193,11212,11231,11242,11250,11264,11342,11364,11459,11498,11516,11585,11592,11649,11801,11858,11915,11941,12048,12086,12124,12143,12162,12219,12238,12257,12276,12295,12313,12314,12341,12371,12390,12428,12485,12514,12523,12580,12599,12694,12713,12741,12827,12881,12960,12998,13017,13036,13043,13045,13074,13093,13131,13169,13206,13207,13239,13245,13264,13278,13302,13321,13393,13394,13397,13416,13428,13435,13518,13530,13568,13587,13644,13663,13689,13739,13758,13777,13796,13830,13910,14005,14024,14078,14081,14092,14119,14157,14176,14195,14233,14252,14271,14290,14309,14328,14343,14366,14385,14404,14441,14499,14556,14575,14661,14669,14670,14727,14784,14841,14879,14917,14943,14955,15012,15031,15050,15088,15107,15126,15164,15183,15219,15221,15297,15335,15354,15373,15378,15411,15430,15468,15506,15525,15582,15620,15639,15658,15696,15711,15734,15772,15810,15829,15848,15867,15886,15943,15962,16000,16019,16037,16076,16133,16152,16171,16209,16228,16247,16266,16285,16304,16323,16342,16361,16380,16418,16456,16475,16513,16532,16589,16601,16608,16627,16665,16703,16741,16760,16763,16779,16830,16855,16874,16893,16950,16969,16988,17039,17041,17045,17093,17140,17197,17216,17290,17292,17330,17349,17387,17406,17425,17444,17463,17501,17520,17596,17615,17691,17710,17729,17748,17805,17824,17843,17862,17881,17900,17919,17926,17957,17995,18033,18071,18128,18185,18195,18204,18242,18260,18261,18299,18318,18356,18394,18413,18432,18489,18508,18527,18603,18679,18736,18751,18755,18793,18803,18812,18850,18869,18907,18945,18964,18983,19002,19040,19097,19116,19135,19204,19249,19306,19325,19344,19382,19420,19458,19496,19534,19567,19572,19610,19648,19667,19686,19705,19762,19781,19800,19819,19876,19895,19914,19933,19952,19990,20047,20085,20104,20123,20142,20161,20171,20180,20237,20275,20294,20313,20332,20351,20370,20389,20427,20446,20465,20484,20499,20529,20541,20548,20560,20572,20598,20616,20617,20636,20655,20660,20712,20731,20750,20786,20788,20864,20902,20921,20978,21016,21054,21092,21111,21168,21187,21206,21221,21263,21264,21282,21339,21358,21395,21415,21434,21453,21491,21510,21516,21529,21567,21586,21662,21700,21738,21757,21795,21814,21852,21871,21909,21928,21947,21966,22004,22023,22042,22050,22080,22099,22110,22194,22209,22213,22270,22289,22327,22365,22375,22384,22403,22441,22460,22479,22498,22536,22555,22574,22631,22650,22688,22726,22745,22802,22859,22878,22897,22916,22969,22973,23030,23049,23068,23125,23144,23163,23199,23240,23258,23277,23315,23334,23349,23353,23372,23410,23423,23429,23505,23541,23543,23581,23600,23610,23648,23657,23676,23733,23752,23771,23809,23828,23904,23961,23999,24018,24037,24056,24094,24107,24113,24151,24170,24208,24227,24237,24246,24303,24315,24328,24331,24341,24360,24398,24436,24455,24474,24493,24512,24531,24550,24569,24588,24607,24626,24645,24664,24683,24721,24757,24759,24774,24778,24835,24854,24873,24949,24968,24987,25000,25061,25063,25082,25120,25158,25196,25221,25253,25272,25310,25348,25367,25386,25405,25462,25481,25538,25557,25595,25614,25652,25688,25709,25747,25785,25823,25842,25880,25893,25937,25956,25975,25994,26032,26051,26070,26089,26108,26127,26146,26203,26222,26260,26279,26298,26355,26374,26387,26412,26431,26450,26469,26526,26545,26554,26621,26678,26697,26735,26754,26773,26811,26925,26926,27020,27068,27134,27198,27210,27248,27267,27362,27387,27431,27552,27652,27704,27780,27799,27806,27837,27894,27977,28017,28103,28122,28181,28198,28312,28331,28616,28635,28871,29117,29241,29471,29498,29499,29604,29620,29630,29697,29711,29724,29765,29794,29908,30002,30031,30060,30089,30118,30147,30176,30205,30234,30263,30292,30321,30350,30379,30408,30437,30466,30495,30524,30553,30582,30611,30640,30669,30698,30727,30756,30785,30814,30843,30872,30901,30930,30959,30988,31017,31046,31075,31104,31133,31162,31191,31220,31249,31278,31307,31336,31365,31394,31423,31452,31481,31510,31539,31568,31597,31626,31655,31684,31713,31742,31771,31800,31829,31858,31887,31916,31945,31974,32003,32032,32061,32090,32119,32148,32177,32206,32235,32264,32293,32322,32351,32380,32409,32438,32467,32496,32525,32554,32583,32612,32641,32670,32699,32728,32757,32786,32815,32844,32873,32902,32931,32960,32989,33018,33047,33076,33105,33134,33163,33192,33221,33250,33279,33308,33337,33366,33395,33424,33453,33482,33511,33540,33569,33598,33627,33656,33685,33714,33743,33772,33801,33830,33859,33888,33917,33946,33975,34004,34033,34062,34091,34120,34149,34178,34207,34236,34265,34294,34323,34352,34381,34410,34439,34468,34497,34526,34555,34584,34613,34642,34671,34700,34729,34758,34787,34816,34845,34874,34903,34932,34961,34990,35002,35060,35063,35091,35119,35143,35160,35165,35232,35254,35310,35328,35382,35402,35445,35452,35458,35473,35521,35580,35625,35639,35680,35688,35712,35771,35824,35830,35876,35901,35960,36019,36048,36082,36127,36142,36201,36239,36272,36302,36331,36374,36390,36462,36521,36580,36633,36644,36703,36762,36791,36792,36833,36892,36951,37023,37082,37141,37205,37264,37323,37353,37394,37453,37481,37486,37512,37546,37584,37643,37702,37746,37766,37825,37884,37955,37993,38014,38056,38073,38145,38173,38179,38204,38263,38299,38327,38386,38445,38516,38575,38624,38634,38706,38765,38793,38824,38886,38888,38947,39006,39053,39077,39107,39136,39174,39195,39235,39242,39267,39326,39385,39449,39508,39567,39596,39602,39638,39668,39695,39697,39756,39828,39887,39926,39946,40117,40146,40175,40204,40233,40262,40291,40320,40349,40378,40407,40436,40465,40494,40523,40552,40581,40610,40639,40668,40697,40726,40755,40784,40813,40842,40871,40900,40929,40958,40987,41016,41045,41074,41103,41132,41161,41190,41219,41248,41277,41306,41335,41364,41393,41422,41451,41480,41509,41538,42008,42037,42066,42095,42124,42163,42184,42274,42307,42349,42365,42372,42424,42440,42449,42511,42586,42629,42661,42698,42714,42717,42729,42748,42779,42823,42898,42913,42985,43060,43135,43222,43297,43372,43390,43443,43459,43486,43504,43534,43568,43609,43636,43666,43696,43722,43771,43787,43829,43846,43859,43880,43914,43933,43967,43971,43974,43993,44008,44064,44083,44107,44109,44110,44170,44228,44245,44277,44320,44335,44372,44407,44482,44557,44615,44644,44681,44719,44753,44760,44794,44865,44881,44922,44930,44956,44977,45031,45082,45085,45104,45152,45213,45271,45293,45308,45344,45401,45459,45531,45553,45588,45618,45646,45718,45744,45775,45816,45833,45905,45962,46037,46096,46165,46168,46190,46194,46200,46211,46226,46280,46282,46286,46291,46295,46309,46424,46568,46618,46641,46644,46671,46672,46674,46719,46753,46835,46859,46883,46895,46924,46929,46949,47019,47101,47108,47114,47141,47164,47211,47233,47240,47241,47251,47273,47275,47300,47355,47360,47381,47399,47453,47474,47512,47519,47534,47544,47550,47683,47730,47766,47772,47780,47838,47859,47868,47871,47881,47916,47951,48009,48067,48092,48109,48186,48218,48223,48319,48334,48342,48392,48446,48469,48478,48479,48503,48518,48547,48554,48572,48581,48604,48612,48655,48691,48711,48725,48733,48843,48847,48855,48866,48902,48936,48998,49021,49077,49094,49100,49157,49200,49211,49251,49261,49287,49354,49372,49394,49429,49471,49541,49554,49571,49606,49615,49617,49622,49651,49680,49700,49739,49762,49783,49857,49863,49865,49880,49971,49982,49983,49989,50010,50043,50312,50405,50485,50517,50591,50811,50979,51048,51139,51183,51301,51413,51555,51568,51600,51683,51776,51812,51906,51928,52115,52340,52410,52488,52525,52577,52643,52649,52749,52891,52938,53022,53231,53340,53483,53513,53603,53712,53741,53891,53902,53952,53974,54129,54133,54201,54402,54437,54441,54458,54500,54547,54548,54621,54667,54687,54825,55052,55181,55191,55278,55306,55700,55851,55856,56109,56301,56322,56554,56566,56590,56616,56634,56656,56681,56780,56852,56988,57069,57576,57598,57600,57693,57939,57948,58015,58029,58069,58082,58193,58243,58262,58371,58378,58388,58427,58493,58598,58602,58690,58706,58723,58832,58857,58987,58991,59001,59176,59471,59478,59568,59587,59637,59644,59842,59908,59922,59942,59983,60015,60092,60120,60133,60157,60178,60363,60412,60471,60608,60679,60780,60784,60907,60977,61046,61059,61061,61168,61186,61202,61344,61434,61507,61635,61685,61798,61817,61824,61825,61828,61845,61851,61899,62177,62306,62312,62423,62453,62460,62466,62599,62600,62618,62640,62826,62979,62994,63011,63049,63144,63227,63302,63330,63373,63419,63514,63671,63782,63873,63989,64122,64275,64495,64793,64936,65053,65160,65176,65276,65299,65331,65436,65447,65514,65717,65872,66021,66450,66522,66694,66974,67039,67108,67118,67194,67272,67285,67324,67434,67461,67675,67875,67886,67933,67995,67999,68038,68207,68246,68310,68414,68465,68505,68574,68577,68582,68698,68703,69078,69096,69310,69321,69362,69405,69456,69595,69608,69696,69708,69754,69793,69891,70083,70090,70098,70199,70246,70252,70261,70424,70457,70489,70549,70605,70653,70682,70684,70686,70833,70880,70940,71040,71127,71160,71456,71485,71487,71646,71702,71727,71747,71762,71875,71888,71950,71958,71962,72173,72330,72361,72523,72602,72606,72690,72698,72891,72990,73220,73224,73287,73319,73573,73588,73770,73787,73798,73873,73881,73971,74103,74104,74285,74302,74383,74523,74541,74627,74677,74778,74834,74867,74937,74957,75063,75128,75132,75208,75269,75455,75506,75835,75957,75963,76162,76199,76393,76528,76573,76697,76730,76774,76889,77020,77081,77129,77269,77348,77376,77421,77471,77592,77611,77681,77820,77844,77906,77907,78118,78119,78130,78304,78406,78440,78510,78578,78599,78776,78993,79039,79222,79327,79346,79436,79508,79530,79562,79589,79612,79631,79641,79699,79773,79978,80058,80314,80322,80403,80476,80484,80638,80687,80691,80696,80746,80879,80905,81040,81085,81193,81203,81269,81339,81384,81717,81725,81835,82035,82184,82200,82272,82329,82394,82564,82644,82680,82858,82859,82935,82961,83034,83135,83136,83139,83159,83193,83315,83412,83656,84091,84098,84147,84162,84211,84407,84524,84639,84659,84772,84787,84824,85106,85202,85233,85320,85375,85427,85545,85562,85563,85755,85915,86065,86180,86194,86201,86207,86330,86370,86375,86450,86475,86568,86602,86670,86799,86809,86835,86839,86876,87023,87092,87168,87248,87264,87385,87537,87647,87700,87728,87730,87758,87779,87850,87868,87901,87954,87968,87978,87994,88003,88038,88093,88129,88221,88447,88486,88594,88611,88770,88796,88814,88817,88852,89045,89067,89311,89432,89608,89609,89706,89824,89917,90062,90074,90088,90242,90343,90437,90957,90999,91064,91158,91268,91419,91430,91509,91592,91735,91934,91964,92064,92068,92071,92147,92363,92754,92845,92996,93076,93189,93199,93284,93388,93491,93505,93523,93539,93572,93785,93814,94028,94095,94251,94308,94489,94519,94524,94556,94622,94674,94703,94794,94814,94841,94907,94957,94971,95054,95237,95363,95436,95488,95503,95587,95665,95684,95688,95904,95909,96002,96018,96108,96215,96316,96485,96646,96864,96906,96932,96978,97060,97204,97287,97428,97519,97731,97869,97900,97962,98058,98156,98218,98233,98254,98420,98439,98687,98709,98812,98954,98975,99091,99125,99174,99269,99295,99410,99419,99494,99699,99744,99853,99896,99908,99951",
          "2,44,119,261,299,368,370,436,447,456,502,741,745,789,837,908,982,1033,1053,1131,1177,1304,1583,1673,1750,1762,1819,1941,2004,2588,2749,2869,3021,3460,3930,4320,5172,5351,5428,5502,5552,5575,5603,5697,6133,6241,6271,6380,6446,6449,6559,6571,6585,6816,6830,6874,6889,6979,7027,7035,7180,7191,7266,7366,7470,7503,7509,7519,7540,7549,7594,7621,7851,7862,7882,7908,7917,8000,8018,8055,8172,8195,8226,8279,8282,8323,8434,8512,8515,8517,8528,8597,8612,8691,8744,8751,8772,8781,8793,8858,8876,8881,8895,8902,8929,9005,9028,9132,9249,9358,9443,9561,9747,9835,9861,9896,10053,10112,10221,10312,10314,10366,10393,10435,10539,10744,11060,11079,11166,11460,11527,11653,11731,11831,11944,11982,12056,12355,12648,12792,12841,12862,13042,13228,13265,13559,13834,13940,14396,14506,14598,14724,14954,14981,15133,15186,15650,15904,16190,16573,16594,16873,16922,17229,17412,17570,17916,17942,18280,18588,19166,19167,19321,19656,19696,20160,20713,20763,20774,20899,21368,21685,21823,21886,22025,22665,23428,23801,24089,24153,25378,25689,25782,26275,26403,26454,26932,26939,26991,27096,27501,27636,27660,27892,28178,28337,28743,28890,28902,29087,29136,29263,29281,29305,29555,29559,29573,29718,29885,29926,29964,29981,30302,34576,35754,35892,35937,36121,36157,36862,37064,37078,37123,37133,37365,37430,37597,37749,37801,37861,37927,38045,38121,38486,38507,38567,38804,38862,38918,39108,39261,39467,39505,39802,39922,39942,41500,43102,45168,45610,45684,45733,45859,47336,47384,48152,50185,50274,50281,50403,50635,50711,50881,51295,51406,51490,51509,51577,51657,51682,51804,52045,52173,52201,52216,52263,52330,52362,52457,52489,52511,52568,52724,52740,52786,52998,53020,53069,53113,53221,53240,53332,53421,53481,53488,53495,53569,53590,53820,53908,54014,54063,54109,54117,54278,54598,54607,54767,54769,54856,54938,54945,54999,55006,55009,55090,55288,55312,55325,55426,55674,55684,55767,55902,55960,56162,56186,56208,56213,56341,56343,56435,56539,56619,56728,56822,56867,57210,57236,57246,57346,57370,57377,57386,57415,57476,57523,57689,57944,58041,58173,58230,58261,58332,58377,58432,58555,58726,58750,58752,58815,58868,58913,59126,59276,59305,59427,59436,59459,59467,59760,60088,60282,60375,60419,60482,60577,60708,60740,60746,60805,60881,60909,60985,61021,61027,61035,61045,61204,61549,61732,61901,62099,62194,62201,62229,62273,62349,62507,62510,62610,62615,62881,62986,62987,63200,63488,63502,63518,63532,63681,63954,64031,64094,64200,64223,64251,64319,64336,64451,64466,64634,64784,64799,64845,64901,65011,65024,65062,65137,65141,65150,65288,65408,65409,65522,65551,65723,65898,65905,65918,65962,66059,66079,66257,66273,66314,66345,66449,66568,66675,66799,66902,67028,67041,67134,67243,67404,67512,67666,67689,67949,67980,68069,68083,68151,68179,68320,68416,68441,68449,68512,68592,68638,68726,68798,69011,69018,69036,69046,69047,69116,69228,69254,69323,69325,69377,69452,69640,69692,70047,70077,70121,70157,70216,70438,70462,70507,70587,70643,70796,70834,70854,70896,71128,71197,71229,71281,71386,71466,71584,71776,71786,72000,72013,72106,72110,72118,72237,72278,72451,72498,72768,72828,73053,73055,73248,73264,73397,73696,73712,73901,73948,73961,74070,74071,74128,74133,74175,74317,74450,74527,74666,74704,74711,74788,74852,74932,75018,75046,75220,75223,75354,75443,75492,75560,75591,75745,75814,75842,75865,75881,75902,75936,76401,76491,76550,76570,76595,76619,76817,76824,76854,76891,77009,77060,77128,77167,77180,77186,77206,77245,77322,77327,77362,77407,77451,77457,77720,77784,77788,78019,78067,78089,78182,78429,78606,78625,78659,78726,78869,78878,79083,79186,79281,79418,79420,79465,79482,79549,79595,79607,79814,79854,79863,79977,79998,80078,80149,80279,80363,80409,80429,80488,80554,80597,80664,80763,80768,80802,80894,81183,81222,81239,81270,81288,81334,81433,81484,81526,81542,81711,81714,81810,81817,81891,81983,81997,82078,82307,82318,82327,82340,82439,82486,82607,82780,83024,83304,83443,83533,83589,83638,83717,83771,83816,83836,83863,83982,84083,84434,84517,84520,84558,84670,84671,84672,84702,84783,85026,85028,85062,85123,85184,85287,85525,85624,85654,85719,85741,85770,85836,85906,85989,86030,86168,86287,86312,86365,86383,86398,86418,86669,86724,86788,86807,86848,86860,86934,86947,87028,87174,87190,87212,87273,87491,87542,87620,87664,87784,87884,88031,88086,88090,88161,88182,88195,88216,88223,88241,88410,88548,88574,88624,88645,88875,88900,88920,88956,89026,89096,89184,89328,89347,89366,89399,89473,89485,89560,89644,89811,89846,89890,90010,90162,90180,90211,90219,90301,90456,90503,90506,90597,90606,90630,90650,90735,90738,90949,91010,91072,91177,91249,91381,91512,91516,91534,91754,91800,91822,91826,92043,92171,92214,92230,92387,92400,92501,92711,92863,92914,92965,93089,93117,93137,93208,93254,93315,93366,93501,93516,93564,93743,93872,93878,93942,93980,94102,94181,94414,94459,94496,94732,94780,94793,94798,94803,94823,94908,94996,95002,95027,95187,95241,95266,95381,95403,95408,95848,95936,95968,96028,96170,96173,96254,96353,96419,96451,96643,96763,96840,97008,97168,97370,97372,97380,97417,97651,97678,97685,97791,97874,97961,97997,98045,98096,98144,98162,98309,98328,98794,99103,99117,99188,99228,99301,99310,99340,99433,99483,99534,99565,99602,99618,99660,99664,99704,99731,99790,99956,99973",
          "54,108,109,140,155,163,179,189,190,238,240,759,1788,1929,1939,1966,1967,1969,1990,2037,2109,2123,2183,2207,2212,2232,2258,2264,2284,2297,2301,2353,2360,2369,2382,2389,2410,2413,2422,2456,2461,2479,2482,2492,2496,2501,2515,2524,2535,2538,2543,2544,2565,2567,2574,2581,2586,2611,2619,2653,2697,2710,2716,2717,2723,2725,2761,2769,2771,2775,2795,2810,2814,2843,2855,2863,2889,2897,2903,2912,2919,2924,2933,2987,3070,3081,3085,3089,3099,3106,3109,3141,3153,3213,3222,3256,3304,3315,3317,3319,3344,3347,3349,3362,3373,3378,3381,3382,3409,3432,3476,3484,3515,3518,3519,3536,3543,3544,3550,3552,3554,3564,3571,3572,3593,3595,3605,3622,3642,3668,3672,3682,3693,3698,3731,3742,3750,3755,3798,3802,3810,3830,3841,3849,3870,3882,3897,3899,3911,3915,3925,3973,3974,3988,3989,4015,4026,4056,4071,4077,4097,4104,4121,4137,4141,4167,4200,4208,4220,4227,4252,4260,4263,4277,4280,4324,4344,4346,4363,4384,4415,4422,4425,4428,4436,4444,4472,4473,4475,4477,4505,4508,4516,4526,4533,4566,4594,4602,4613,4628,4631,4651,4652,4666,4680,4681,4704,4707,4709,4761,4768,4789,4791,4795,4808,4813,4816,4831,4843,4888,4897,4900,4906,4907,4927,4931,4966,4987,4990,5054,5066,5572,6111,6260,6292,6307,6337,6371,6584,6622,6632,6812,6866,6910,6914,6926,6974,6988,7057,7245,7263,7337,7381,7437,7458,7587,7637,7675,7736,7807,7819,7857,7875,7884,7922,7948,7957,8014,8026,8104,8116,8162,8217,8290,8309,8345,8504,8535,8552,8606,8609,8620,8652,8787,8854,8878,8911,9022,9159,9208,9260,9290,9328,9332,9333,9407,9461,9516,9573,9735,9745,9748,9939,10194,10196,10248,10297,10351,10407,10524,10573,10606,10643,10791,10875,10899,10982,11211,11266,11283,11623,11699,11904,11946,12002,12011,12169,12272,12363,12375,12387,12410,12483,12490,12574,12816,12820,12884,12983,13046,13108,13182,13353,13800,13810,13960,14028,14090,14518,14547,14645,14745,14933,15013,15259,15319,15370,15449,15520,15681,15757,15796,16135,16184,16271,16328,16540,16618,16621,16658,16797,16820,17033,17233,17243,17428,17476,17488,17631,17948,17993,18031,18095,18096,18177,18192,18451,18477,18590,18611,18625,18962,18986,19269,19352,19411,19828,19941,19976,20112,20513,20778,20846,20939,20957,20986,21022,21182,21185,21203,21213,21341,21355,21541,21551,21569,21601,21605,21827,21944,22451,22499,22709,22892,23072,23519,23554,23837,23858,23930,24064,24408,24745,24776,24824,24872,24882,24972,25251,25334,25452,25474,25530,25567,26028,26049,26099,26110,26126,26187,26189,26202,26220,26305,26442,26455,26528,26531,26546,26561,26566,26581,26595,26598,26600,26669,26841,26899,27001,27007,27014,27064,27080,27202,27296,27303,27367,27409,27459,27484,27487,27500,27514,27516,27558,27616,27623,27643,27655,27657,27751,27776,27815,28004,28033,28040,28056,28100,28128,28140,28152,28177,28201,28237,28239,28395,28484,28512,28542,28594,28612,28623,28630,28659,28666,28753,28885,28886,28920,28928,29008,29014,29035,29056,29071,29076,29164,29402,29409,29441,29444,29531,29581,29589,29636,29680,29688,29801,29815,29839,29855,29898,29924,29990,30019,30048,30077,30106,30135,30164,30193,30222,30251,30280,30338,30367,30396,30425,30454,30483,30512,30541,30570,30579,30628,30657,30686,30715,30744,30773,30802,30831,30860,30889,30918,30947,31034,31063,31092,31121,31150,31179,31208,31237,31266,31295,31324,31382,31411,31440,31469,31498,31527,31556,31585,31614,31643,31672,31701,31730,31759,31788,31817,31846,31875,31904,31933,31962,31991,32020,32049,32078,32107,32136,32165,32194,32223,32252,32281,32310,32339,32368,32397,32426,32455,32484,32513,32542,32571,32600,32629,32658,32687,32716,32745,32774,32803,32832,32861,32890,32919,32948,32977,33006,33035,33064,33093,33122,33151,33180,33209,33238,33267,33296,33325,33383,33412,33441,33470,33528,33557,33586,33615,33644,33673,33702,33731,33760,33789,33818,33847,33876,33905,33934,33963,33992,34079,34108,34137,34166,34195,34224,34253,34282,34311,34340,34369,34398,34427,34456,34485,34514,34543,34572,34601,34630,34659,34688,34717,34746,34775,34804,34833,34862,34891,34920,34949,34978,35019,35037,35077,35086,35108,35125,35182,35200,35271,35289,35293,35308,35318,35345,35370,35377,35419,35470,35490,35538,35567,35597,35656,35691,35729,35756,35847,35862,35874,35918,35977,35990,36036,36067,36099,36159,36218,36232,36289,36306,36348,36407,36460,36479,36538,36597,36661,36720,36779,36795,36796,36807,36850,36909,36924,36968,37040,37099,37158,37174,37222,37233,37281,37340,37470,37529,37545,37601,37629,37660,37719,37734,37783,37809,37842,37865,37901,37934,37972,38031,38044,38090,38104,38162,38221,38280,38305,38307,38344,38365,38381,38403,38462,38484,38533,38547,38573,38592,38613,38666,38678,38723,38782,38841,38856,38869,38905,38919,38923,38964,38995,39023,39048,39094,39109,39123,39153,39212,39224,39230,39232,39284,39296,39343,39402,39466,39491,39525,39539,39584,39598,39603,39655,39675,39714,39773,39845,39904,39915,39963,40109,40138,40167,40196,40225,40254,40283,40312,40341,40370,40399,40428,40457,40486,40544,40573,40602,40631,40660,40689,40718,40747,40776,40805,40834,40863,40892,40921,40950,40979,41008,41037,41066,41095,41124,41153,41182,41211,41240,41269,41298,41327,41356,41385,41414,41443,41472,41501,41530,41550,41551,41552,41553,41554,41555,41556,41557,41558,41559,41560,41561,41562,41563,41564,41565,41566,41567,41568,41569,41570,41571,41572,41573,41574,41576,41577,41578,41579,41580,41581,41582,41583,41584,41585,41586,41587,41588,41589,41590,41591,41592,41593,41594,41595,41596,41597,41598,41599,41600,41601,41602,41603,41604,41605,41607,41608,41609,41610,41611,41612,41613,41614,41615,41616,41617,41618,41619,41620,41621,41622,41623,41624,41625,41626,41627,41628,41629,41630,41631,41632,41633,41634,41635,41636,41637,41638,41639,41640,41641,41642,41643,41644,41645,41646,41647,41648,41649,41650,41651,41652,41653,41654,41655,41656,41657,41658,41659,41660,41661,41662,41663,41664,41665,41666,41668,41669,41670,41671,41672,41673,41674,41675,41676,41677,41678,41679,41680,41681,41682,41683,41684,41685,41686,41687,41688,41689,41690,41691,41692,41693,41694,41695,41696,41697,41698,41699,41701,41702,41703,41704,41705,41706,41707,41708,41709,41710,41711,41712,41713,41714,41715,41716,41717,41718,41719,41720,41722,41723,41724,41725,41726,41727,41729,41730,41731,41732,41733,41734,41735,41736,41737,41738,41739,41741,41742,41743,41744,41745,41746,41747,41748,41749,41750,41751,41752,41753,41754,41755,41757,41758,41759,41760,41761,41762,41763,41764,41765,41767,41768,41769,41770,41771,41773,41774,41775,41776,41777,41778,41779,41780,41781,41782,41783,41784,41785,41786,41787,41788,41790,41792,41794,41796,41798,41800,41804,41806,41808,41810,41812,41814,41816,41818,41820,41822,41824,41826,41828,41830,41832,41834,41836,41838,41840,41842,41844,41848,41850,41852,41854,41856,41858,41860,41862,41864,41866,41868,41870,41872,41874,41876,41878,41880,41882,41884,41886,41888,41890,41892,41894,41896,41898,41900,41902,41904,41906,41908,41910,41912,41914,41916,41918,41920,41922,41924,41926,41928,41931,41934,41937,41940,41943,41949,41952,41955,41959,41963,41967,41971,41975,41979,41983,41987,41992,42021,42050,42079,42108,42137,42204,42212,42222,42302,42370,42377,42452,42471,42473,42478,42539,42544,42560,42602,42609,42614,42686,42689,42724,42731,42776,42851,42926,43013,43025,43045,43088,43163,43202,43242,43246,43250,43269,43273,43276,43325,43327,43333,43339,43400,43423,43442,43487,43513,43562,43572,43577,43585,43590,43637,43640,43665,43724,43725,43731,43747,43799,43801,43823,43874,43903,43910,43961,43976,44031,44036,44043,44111,44120,44188,44198,44271,44273,44348,44433,44434,44435,44443,44510,44531,44539,44590,44594,44672,44741,44747,44822,44836,44909,44984,45012,45049,45062,45083,45132,45179,45184,45192,45196,45251,45287,45322,45359,45439,45481,45509,45568,45571,45619,45626,45696,45755,45793,45796,45813,45883,45932,45942,46000,46002,46025,46063,46132,46178,46257,46303,46314,46343,46388,46434,46447,46451,46465,46528,46581,46586,46640,46703,46822,46882,46888,46891,46941,46946,46953,46971,47005,47015,47018,47096,47113,47184,47199,47213,47267,47283,47285,47368,47371,47400,47435,47445,47597,47607,47633,47639,47721,47726,47767,47771,47789,47796,47862,47879,48016,48054,48075,48105,48159,48184,48193,48196,48254,48294,48327,48340,48350,48374,48394,48493,48506,48520,48531,48579,48586,48600,48607,48761,48772,48788,48792,48857,48877,48888,48906,48946,48951,48983,49017,49046,49049,49058,49072,49078,49086,49088,49121,49130,49134,49144,49156,49240,49242,49271,49299,49313,49326,49365,49380,49415,49436,49440,49450,49480,49521,49538,49560,49656,49698,49716,49725,49740,49745,49796,49861,49875,49896,49921,49963,50035,50095,50096,50133,50285,50377,50419,50420,50657,50772,50886,50887,51129,51162,51210,51216,51425,51461,51485,51569,51625,51687,51720,51840,52029,52164,52184,52309,52312,52318,52396,52414,52452,52493,52665,52669,52715,52733,52767,52795,52824,52828,52867,53015,53140,53261,53358,53366,53480,53588,53608,53687,53770,53840,53868,53896,54333,54336,54386,54444,54729,54983,54986,54995,55077,55078,55099,55238,55257,55347,55382,55389,55420,55491,55565,55581,55612,55705,55859,55932,55934,55987,56107,56201,56259,56266,56270,56299,56362,56414,56483,56820,56910,56992,57341,57371,57396,57531,57715,57824,57865,57921,57925,57995,58087,58103,58134,58202,58209,58211,58284,58356,58358,58508,58591,58683,58687,58858,58869,58923,58966,59100,59150,59159,59175,59297,59359,59432,59491,59522,59556,59569,59622,59659,59667,59674,59677,59718,59944,59988,60026,60043,60115,60326,60368,60393,60455,60524,60683,60872,60894,60952,61065,61234,61237,61280,61625,61643,61753,61777,61862,61918,61998,62065,62082,62118,62199,62221,62416,62484,62551,62619,62678,62733,62780,62805,62945,62966,63069,63222,63245,63379,63538,63586,63677,63896,63930,64069,64103,64110,64243,64250,64611,64656,64659,64707,64900,64953,65020,65107,65177,65274,65393,65501,65505,65692,65761,65875,65952,66001,66028,66056,66076,66136,66225,66250,66362,66375,66388,66424,66438,66534,66623,66708,66839,66952,66993,67150,67174,67220,67233,67331,67381,67505,67531,67632,67646,67704,67844,67864,67894,68073,68149,68164,68170,68175,68251,68379,68396,68508,68545,68595,68742,68746,68848,69088,69109,69312,69448,69497,69520,69612,69656,69802,70020,70041,70065,70128,70171,70184,70203,70218,70241,70284,70286,70317,70410,70411,70554,70570,70580,70780,70923,70947,71089,71121,71174,71185,71188,71325,71370,71374,71384,71392,71518,71638,71655,71677,71796,71947,71969,72132,72155,72172,72199,72225,72236,72358,72423,72539,72554,72591,72864,72983,73099,73134,73202,73232,73360,73396,73587,73992,73993,74077,74125,74179,74365,74437,74438,74535,74648,74904,75040,75114,75323,75349,75402,75412,75475,75596,75684,75746,75761,75832,76039,76063,76192,76226,76307,76313,76319,76376,76728,76916,77103,77172,77260,77304,77391,77593,77727,77758,77947,77966,77974,78023,78059,78071,78103,78116,78171,78237,78311,78319,78326,78371,78500,78571,78837,78893,78913,78961,79068,79183,79197,79227,79235,79298,79393,79584,79825,80001,80059,80089,80195,80218,80245,80384,80391,80412,80540,80550,80619,80725,80729,80770,80864,80927,80958,81094,81107,81173,81213,81454,81654,81691,81757,81787,81801,81827,81979,82122,82207,82211,82346,82352,82456,82535,82708,82860,82934,83015,83043,83047,83116,83125,83205,83376,83416,83483,83499,83575,83674,83690,83716,83721,83775,83837,83889,83954,84045,84070,84125,84171,84185,84228,84658,84697,84782,84961,84986,85085,85162,85178,85197,85224,85271,85283,85344,85352,85536,85676,85976,86125,86151,86392,86445,86470,86622,86692,86702,86703,86722,86763,86994,87079,87087,87210,87235,87293,87298,87531,87600,87781,87858,88163,88363,88414,88426,88451,88509,88610,88891,88946,88977,89032,89133,89182,89217,89315,89380,89401,89422,89472,89593,89597,89720,89784,89906,90065,90299,90318,90418,90419,90746,90943,90977,90997,91058,91076,91090,91163,91194,91254,91288,91436,91472,91478,91530,91580,91585,91644,91655,91726,91792,91806,91891,91974,92078,92408,92457,92467,92715,92769,92888,92892,92935,92997,92999,93011,93067,93136,93191,93195,93240,93327,93359,93543,93587,93627,93661,93754,93867,93967,94039,94065,94068,94087,94121,94274,94320,94337,94369,94417,94526,94530,94548,94561,94582,94669,95004,95033,95069,95084,95207,95324,95467,95564,95671,95690,96113,96114,96209,96279,96444,96461,96508,96579,96649,96711,96717,96732,96753,96890,96928,97028,97095,97125,97153,97201,97202,97252,97265,97539,97563,97782,97819,97922,97965,97974,98002,98003,98170,98259,98295,98368,98442,98451,98628,98737,99053,99062,99343,99344,99364,99502,99524,99584,99941",
          "305,345,346,357,418,478,497,503,599,631,669,673,854,858,892,1014,1065,1084,1138,1210,1307,1456,1499,1622,1634,1638,1668,1680,1729,1742,1744,2080,2322,3146,3891,3987,4140,4624,5042,5062,5191,5284,5367,5556,5669,5733,5838,5907,6010,6068,6086,6107,6142,6235,6254,6361,6523,6549,6556,6570,6709,6738,6843,6844,7084,7100,7104,7113,7142,7227,7231,7252,7285,7304,7447,7513,7543,7659,7704,7721,7769,8069,8176,8196,8275,8291,8366,8371,8428,8525,8560,8592,8614,8757,8794,8811,8915,9010,9157,9167,9216,9232,9243,9291,9331,9352,9425,9452,9494,9511,9518,9574,9789,9790,9804,9922,9927,10146,10310,10359,10365,10513,10517,10531,10667,10898,11075,11276,11446,11492,11529,11531,11570,11590,11679,11691,11703,11752,11798,11800,11846,11989,12093,12452,12760,12808,12909,13212,13335,13651,13713,14086,14423,14485,14487,14579,14700,15124,15374,15445,15454,15492,15528,15583,15641,15663,15747,15778,15922,16446,16590,16840,17139,17236,17439,17493,17594,17758,17777,18059,18152,18262,18622,18891,18938,19336,19595,19841,20209,20738,20897,20929,21066,21081,22098,22173,22658,22815,23042,23107,23293,23769,24092,24137,24307,24370,25101,25126,25246,25665,25703,25958,26259,26271,26389,26409,26434,26482,26648,26667,26809,27067,27093,27100,27233,27245,27269,27747,27947,28727,28784,28829,28843,28936,29106,29467,29897,32093,35498,35704,35934,36133,36180,36517,36618,37061,37110,37390,37482,37495,37580,37681,37804,38184,38259,38300,38377,38697,39165,39188,39410,39619,39737,39801,39868,40443,43978,45089,45324,45507,49285,50179,50293,50352,50506,50655,50668,50743,50801,50973,50983,51175,51463,51641,51743,51871,51890,51949,52165,52259,52435,52528,52538,52557,52602,52668,52779,52817,52949,52981,53168,53193,53236,53361,53532,53818,53885,53916,54176,54207,54270,54384,54497,54522,54662,54873,54888,54899,54977,55082,55162,55308,55314,55319,55353,55457,55488,56014,56143,56167,56168,56171,56175,56221,56438,56440,56441,56500,56530,56605,56824,56878,56892,56922,56973,57122,57154,57279,57301,57556,57641,57712,57714,57722,57746,57870,57913,58004,58163,58167,58227,58259,58276,58306,58363,58381,58411,58464,58573,58838,58863,58865,59166,59253,59256,59529,59530,59558,59658,59666,59694,59695,59870,60000,60231,60449,60677,60739,60763,60821,60829,60870,60900,60935,61075,61163,61465,61510,61577,61612,61653,61670,61673,61709,61714,61757,61864,61873,61936,62013,62051,62083,62192,62333,62376,62534,62553,62589,62593,62630,62783,62854,62955,63459,63512,63529,63564,63620,63666,63905,63908,63917,64029,64150,64262,64295,64296,64312,64372,64412,64424,64643,64651,64711,64719,64763,64896,64996,65104,65135,65181,65221,65344,65380,65394,65606,65668,65736,65737,65855,65919,65994,66308,66663,66739,66855,67064,67452,67527,67535,67552,67807,67891,68013,68055,68064,68123,68288,68322,68544,68557,68585,68679,68792,69072,69149,69259,69499,69501,69592,69638,69665,69726,69935,69988,70003,70016,70074,70288,70380,70381,70512,70683,70697,70777,70788,70817,70925,70959,71008,71232,71336,71503,71506,71550,71616,71647,71886,72101,72189,72200,72445,72492,72558,72888,73082,73108,73146,73233,73306,73488,73658,73894,73989,74158,74201,74335,74355,74401,74412,74584,74837,74886,74974,75101,75109,75129,75314,75316,75388,75396,75418,75511,75552,75594,75603,75712,75720,75722,75725,75799,75803,75985,76178,76441,76533,76897,76999,77196,77320,77655,77746,78072,78137,78144,78348,78457,78479,78521,79029,79057,79102,79149,79171,79342,79416,79661,79739,79812,79824,79826,79899,80070,80080,80082,80170,80235,80285,80393,80397,80407,80483,80563,80608,80612,80671,80709,80860,81075,81151,81410,81525,81567,81805,82024,82067,82180,82182,82188,82308,82460,82549,82966,83027,83074,83209,83211,83311,83317,83341,83350,83538,83542,83555,83563,83669,83736,83767,83780,83814,83825,83993,84068,84336,84691,84737,84788,84911,84966,85001,85240,85321,85410,85569,85601,85615,85621,85718,85768,85806,85849,85892,85953,85975,86013,86142,86219,86366,86841,86945,87089,87115,87229,87303,87458,87759,87828,87834,87920,88156,88167,88230,88274,88281,88467,88522,88786,88884,88907,89052,89101,89114,89209,89264,89277,89458,89534,89730,89857,89945,89946,90042,90228,90240,90298,90352,90475,90499,90582,90590,90646,90702,90730,90794,90881,90939,90985,91011,91023,91097,91165,91184,91213,91317,91352,91428,91782,91794,91867,92168,92335,92482,92517,92624,92634,92653,92682,92704,92741,92761,92765,92908,92969,93034,93053,93090,93220,93286,93294,93427,93460,93508,93781,93844,93921,94080,94137,94488,94610,94741,94846,94913,94935,94985,95024,95200,95205,95238,95316,95393,95500,95537,95539,95742,95781,96058,96175,96243,96320,96472,96565,96627,96807,96953,97052,97082,97088,97100,97172,97182,97233,97295,97592,97660,97702,97856,97866,97923,97978,98041,98048,98110,98185,98222,98232,98281,98456,98557,98661,98683,98685,98809,98824,98852,98896,98905,98959,99012,99046,99360,99436,99456,99563,99580,99615,99776,99777,99864,99887,99980",
          "77,104,142,151,161,211,225,1662,1681,1683,1700,1738,1741,1757,1776,1782,1795,1798,2074,2082,2091,2487,2634,2654,2712,2782,2830,2886,3253,3406,3753,4018,4132,4148,4235,4257,4476,4644,4696,4767,6166,7208,8225,8806,9376,9395,9481,9523,9547,9604,9661,9680,9699,9718,9740,9775,9791,9832,9984,10022,10060,10649,10703,10749,10858,10877,10953,11002,11029,11119,11200,11238,11257,11352,11447,11485,11580,11607,11618,11637,11694,11903,11960,11998,12055,12061,12112,12131,12226,12264,12283,12299,12302,12333,12340,12353,12397,12436,12442,12473,12519,12530,12568,12592,12663,12709,12766,12777,12801,12830,12835,12853,12872,12920,12978,12986,13005,13015,13044,13081,13082,13096,13103,13158,13271,13290,13309,13480,13485,13491,13632,13638,13670,13693,13708,13746,13765,13784,13803,13822,13898,13909,13917,13956,13993,14069,14091,14098,14107,14183,14259,14278,14297,14316,14335,14354,14411,14563,14601,14615,14636,14639,14760,14772,14829,14857,14867,14883,14905,14962,15000,15038,15076,15114,15171,15182,15190,15213,15228,15247,15285,15304,15323,15361,15380,15456,15513,15574,15589,15608,15630,15637,15684,15722,15741,15817,15836,15855,15874,15893,15926,15931,15950,15969,16026,16045,16102,16121,16178,16216,16235,16254,16292,16311,16330,16349,16368,16463,16482,16485,16501,16558,16577,16634,16653,16669,16672,16691,16710,16805,16824,16881,16919,16957,16995,17014,17049,17052,17109,17111,17128,17147,17204,17280,17299,17375,17394,17413,17451,17477,17527,17546,17584,17603,17616,17629,17660,17679,17701,17722,17755,17764,17774,17788,17831,17850,17869,17907,17912,17964,17998,18002,18021,18040,18047,18065,18097,18135,18154,18166,18211,18230,18249,18306,18325,18363,18382,18401,18420,18458,18486,18496,18534,18553,18610,18629,18686,18705,18743,18762,18800,18819,18838,18857,18876,18952,18990,19019,19047,19066,19104,19106,19142,19199,19218,19275,19294,19313,19332,19408,19446,19465,19473,19541,19579,19598,19617,19636,19688,19693,19731,19769,19788,19826,19845,19883,19902,19921,19940,19959,20016,20028,20035,20072,20073,20092,20146,20149,20187,20206,20214,20225,20263,20339,20358,20377,20396,20415,20434,20472,20510,20565,20586,20605,20643,20662,20681,20776,20814,20833,20845,20852,20890,20926,20966,20985,21000,21004,21009,21023,21042,21055,21061,21099,21128,21135,21175,21186,21232,21251,21308,21346,21365,21377,21403,21441,21460,21467,21498,21515,21536,21555,21574,21612,21631,21632,21701,21712,21726,21764,21765,21783,21802,21821,21859,21878,21897,21935,21954,21973,22011,22030,22049,22087,22101,22106,22144,22163,22181,22220,22239,22277,22296,22334,22353,22429,22448,22505,22524,22562,22581,22600,22676,22690,22733,22752,22771,22828,22866,22885,22904,22942,22961,22966,23018,23037,23056,23065,23075,23091,23094,23113,23132,23227,23265,23284,23322,23338,23341,23360,23398,23417,23474,23507,23512,23550,23569,23588,23645,23664,23683,23740,23744,23759,23775,23797,23816,23873,23892,23911,23949,23968,23987,24006,24025,24044,24063,24074,24082,24139,24158,24196,24215,24272,24299,24310,24329,24367,24424,24443,24459,24462,24519,24538,24576,24595,24614,24652,24728,24747,24785,24804,24823,24842,24899,24918,24931,24937,24956,24975,24994,25079,25108,25112,25119,25146,25165,25222,25241,25259,25260,25298,25317,25374,25379,25431,25437,25450,25488,25507,25561,25564,25602,25621,25640,25697,25735,25754,25773,25787,25830,25849,25868,25887,25906,25925,25930,25963,25982,26001,26039,26058,26096,26115,26191,26229,26245,26286,26324,26343,26400,26457,26495,26514,26552,26634,26647,26666,26704,26742,26793,26856,26913,26952,26968,26970,27008,27027,27217,27293,27407,27419,27504,27555,27578,27635,27703,27730,27748,27749,27768,27788,27901,27918,27932,27965,27996,28028,28087,28091,28110,28186,28205,28257,28332,28357,28369,28400,28471,28692,28775,28813,28828,28832,28834,28965,29003,29041,29060,29064,29098,29139,29193,29212,29250,29261,29300,29326,29364,29383,29411,29459,29478,29497,29668,29687,29702,29782,29814,29891,29916,30009,30038,30067,30096,30125,30154,30183,30212,30241,30299,30328,30357,30386,30415,30444,30473,30502,30531,30560,30589,30618,30647,30676,30705,30734,30763,30792,30821,30850,30879,30908,30937,30966,30995,31024,31053,31082,31111,31140,31169,31198,31227,31256,31285,31314,31343,31372,31401,31430,31459,31488,31517,31546,31575,31604,31633,31662,31691,31720,31749,31778,31807,31836,31865,31894,31923,31952,31981,32010,32039,32068,32097,32126,32155,32184,32213,32242,32271,32300,32329,32358,32387,32416,32445,32503,32532,32561,32590,32619,32648,32677,32706,32735,32764,32793,32822,32851,32880,32909,32938,32967,32996,33025,33054,33083,33112,33141,33170,33199,33228,33257,33286,33315,33344,33373,33402,33431,33460,33489,33518,33547,33576,33605,33634,33663,33692,33721,33750,33779,33808,33837,33866,33895,33924,33953,33982,34011,34040,34069,34098,34127,34156,34185,34214,34243,34272,34301,34330,34359,34388,34417,34475,34504,34533,34562,34591,34620,34649,34678,34707,34736,34765,34794,34823,34852,34881,34910,34939,34968,34997,35009,35054,35098,35136,35172,35261,35314,35357,35409,35447,35455,35456,35480,35528,35587,35646,35670,35683,35719,35751,35753,35765,35778,35811,35816,35837,35908,35946,35967,36026,36089,36149,36174,36208,36279,36303,36307,36338,36397,36437,36528,36564,36587,36642,36651,36710,36769,36813,36840,36871,36882,36899,36935,36958,37030,37089,37139,37148,37212,37271,37313,37330,37367,37401,37460,37519,37591,37650,37673,37709,37730,37757,37764,37773,37832,37866,37882,37891,37962,37985,38021,38080,38115,38152,38211,38270,38334,38384,38393,38452,38473,38523,38582,38641,38713,38772,38831,38895,38954,38983,39013,39050,39084,39143,39202,39241,39243,39274,39333,39356,39392,39414,39415,39456,39493,39515,39554,39574,39645,39669,39670,39688,39704,39763,39835,39894,39953,40104,40133,40162,40191,40220,40249,40278,40307,40336,40365,40394,40423,40452,40481,40510,40539,40568,40597,40626,40655,40713,40742,40771,40800,40829,40858,40887,40916,40945,40974,41003,41032,41061,41090,41119,41148,41177,41206,41235,41264,41293,41351,41380,41409,41438,41467,41496,41525,41929,41932,41935,41938,41941,41944,41947,41950,41953,41960,41964,41968,41972,41976,41980,41984,41989,41994,42023,42052,42081,42110,42162,42170,42215,42233,42245,42294,42320,42329,42400,42410,42484,42557,42625,42647,42677,42678,42713,42721,42794,42876,42877,42884,42929,42950,42958,43002,43017,43031,43121,43195,43197,43247,43268,43312,43328,43344,43351,43353,43358,43432,43483,43496,43505,43583,43595,43669,43680,43723,43736,43742,43746,43788,43791,43793,43811,43832,43868,43906,43913,43920,43979,44047,44054,44061,44069,44101,44124,44139,44143,44216,44306,44380,44387,44428,44447,44453,44514,44532,44537,44543,44617,44657,44690,44695,44764,44780,44854,44862,44903,44913,44917,44919,44926,44927,44973,44988,45009,45011,45017,45057,45077,45090,45129,45138,45199,45257,45283,45330,45373,45387,45445,45477,45483,45491,45517,45546,45574,45632,45704,45754,45761,45819,45846,45891,45948,46005,46033,46064,46097,46108,46155,46159,46298,46305,46349,46377,46378,46389,46415,46420,46634,46636,46642,46666,46679,46744,46773,46795,46842,46846,46898,46903,46919,46974,46975,46998,47040,47124,47160,47183,47198,47202,47252,47291,47342,47357,47363,47403,47417,47430,47446,47452,47491,47514,47547,47556,47675,47714,47717,47720,47725,47735,47814,47824,47839,47884,47893,47906,48146,48153,48203,48351,48355,48426,48434,48436,48580,48626,48631,48689,48690,48697,48714,48764,48766,48781,48795,48818,48829,48862,48870,48908,48926,48988,49027,49039,49044,49056,49065,49066,49150,49176,49180,49181,49213,49241,49305,49353,49363,49371,49377,49422,49439,49467,49499,49542,49598,49623,49628,49643,49702,49784,49801,49827,49835,49838,49917,50001,50085,50248,50254,50370,50390,50407,50466,50554,50617,50756,50823,50942,51203,51205,51270,51314,51316,51323,51378,51538,51585,51598,51617,51872,52036,52212,52213,52280,52383,52562,52581,52619,52652,52653,52924,53103,53109,53456,53563,53567,53613,53634,53692,53785,53846,53853,53934,53953,53970,53983,54028,54074,54144,54161,54204,54322,54392,54504,54532,54570,54579,54590,54646,54686,54750,54783,54822,54970,55072,55111,55279,55407,55455,55475,55537,55679,55739,55803,55817,55980,56139,56229,56373,56408,56434,56442,56464,56536,56651,56668,56672,56706,56711,56744,56841,57269,57278,57398,57410,57413,57432,57595,57702,57706,57711,57899,57964,58253,58490,58504,58614,58714,58749,58856,58981,59067,59133,59208,59214,59228,59327,59420,59562,59583,59646,59761,59897,59975,60119,60192,60323,60391,60490,60521,60633,60713,60718,60842,60865,60916,60936,60970,61000,61073,61189,61190,61225,61233,61261,61309,61354,61395,61414,61468,61639,61650,61710,61751,61877,61880,61924,61966,62092,62136,62154,62156,62269,62537,62897,62929,62992,63165,63192,63260,63525,63719,63773,63793,63829,63882,63911,64124,64154,64180,64207,64211,64293,64333,64449,64452,64486,64587,64589,64599,64759,64886,64911,64944,65055,65059,65326,65441,65461,65484,65503,65610,65641,65657,65843,65885,65912,65935,65938,66060,66227,66291,66514,66529,66556,66657,66882,66949,66975,67044,67051,67095,67104,67112,67152,67290,67389,67463,67500,67641,67660,67701,67728,67975,67979,67997,68063,68101,68152,68192,68324,68349,68541,68614,68752,68796,68971,69008,69016,69063,69156,69287,69329,69336,69387,69394,69435,69442,69591,69593,69889,69898,69943,70143,70177,70182,70398,70541,70559,70582,70637,70652,70659,70687,70716,70728,70926,71057,71096,71152,71194,71212,71324,71377,71554,71573,71724,71756,71779,71823,71832,71961,72122,72220,72410,72480,72530,72557,72780,72832,72852,72996,73079,73085,73123,73171,73431,73492,73562,73667,73690,73738,73828,73893,73928,74198,74214,74361,74596,74610,74631,74634,74724,74825,75036,75077,75121,75259,75480,75490,75703,75706,75804,75925,76024,76065,76066,76157,76236,76362,76394,76620,76762,76827,76837,76846,76864,76979,77027,77074,77105,77109,77211,77235,77323,77351,77440,77458,77818,77985,78156,78342,78365,78390,78436,78442,78514,78529,78713,78970,79073,79093,79148,79208,79213,79220,79245,79252,79262,79283,79302,79512,79520,79526,79536,79694,79750,79870,79883,79896,79924,80015,80027,80069,80083,80127,80219,80379,80428,80503,80626,80648,80749,80781,80875,80976,81016,81072,81129,81322,81370,81414,81485,81486,81509,81535,81555,81729,81754,81792,81838,81950,82022,82153,82216,82276,82316,82326,82335,82339,82410,82481,82530,82678,82688,82745,82800,83077,83126,83196,83218,83238,83298,83329,83340,83371,83435,83532,83704,83760,83791,83997,84094,84163,84213,84278,84307,84440,84569,84688,84803,84818,84887,84968,85023,85107,85250,85268,85289,85405,85413,85457,85703,85918,85979,86116,86182,86231,86267,86279,86419,86476,86497,86527,86585,86626,86753,86765,86766,86956,86982,87041,87162,87197,87322,87365,87398,87401,87416,87649,87674,87771,87890,87907,87914,88041,88050,88089,88126,88322,88419,88445,88463,88481,88517,88621,88824,88846,88999,89010,89017,89116,89244,89355,89361,89373,89516,89624,89705,89858,89864,90130,90347,90361,90421,90462,90480,90486,90529,90568,90681,90687,90772,90814,91277,91393,91573,91633,91666,91734,92097,92173,92191,92249,92506,92523,92537,92584,92775,92851,92876,92877,93002,93043,93323,93334,93376,93390,93441,93509,93557,93641,93749,93977,94094,94133,94141,94350,94466,94601,94627,95113,95305,95358,95361,95409,95414,95509,96019,96176,96249,96321,96512,96826,96918,97023,97024,97031,97043,97087,97229,97285,97445,97456,97500,97597,97627,97670,97951,98099,98187,98228,98345,98384,98534,98633,98690,99232,99307,99413,99428,99527,99538,99653,99655,99761,99877,99910",
          "80,172,332,378,395,404,516,543,731,927,1020,1030,1076,1180,1196,1246,1261,1397,1440,1523,1628,1699,1704,1770,1823,2089,2168,2256,2309,2784,2838,3013,3028,3175,3273,3341,3827,3878,3896,3907,4100,4386,4574,4774,4818,4902,5003,5011,5246,5285,5300,5308,5349,5385,5387,5412,5422,5443,5694,5765,5859,5904,5962,5971,6118,6267,6318,6351,6417,6540,6544,6719,6793,6835,6896,6907,6961,6983,7106,7132,7151,7194,7201,7329,7506,7536,7537,7555,7565,7622,7671,7746,7764,7801,7928,8088,8241,8448,8491,8607,8957,9177,9251,9272,9323,9349,9437,9542,9592,9610,9792,9826,9844,9893,9906,10197,10246,10273,10275,10485,10626,10822,10854,11001,11184,11196,11204,11341,11472,11475,11488,11520,11723,11742,11908,12220,12231,12293,12332,12420,12458,12487,12509,12518,12562,12605,12644,12753,12788,12837,12842,12890,12917,13002,13122,13191,13286,13318,13334,13460,13517,13727,13794,13851,14003,14074,14141,14234,14288,14362,14539,14653,14695,14765,14799,14941,15153,15291,15490,15518,15542,15548,15607,15729,15840,15940,15955,15968,15985,16012,16036,16129,16230,16394,16828,16832,16880,16889,16897,16904,16966,16980,17008,17023,17199,17256,17351,17423,17452,17454,17470,17513,17917,17938,18084,18109,18264,18304,18461,18644,18718,18984,18995,19095,19241,19255,19258,19324,19431,19760,19816,19818,19846,19986,20053,20120,20257,21200,21228,21391,21642,21645,21733,21926,21987,22028,22071,22148,22313,22325,22475,22552,22751,22758,22860,22895,22914,23038,23168,23186,23212,23221,23241,23345,23379,23506,23592,23991,24000,24059,24119,24143,24162,24230,24294,24381,24467,24491,24567,24568,24668,24709,24889,24980,25154,25162,25387,25514,25624,25802,25837,26022,26152,26231,26335,26524,26897,26936,26950,26956,26994,26995,27016,27155,27189,27305,27356,27490,27686,27739,27832,27904,27953,28163,28215,28411,28492,28533,28736,29102,29127,29186,29206,29215,29283,29293,29396,29465,29490,29821,29890,29902,29918,29976,30052,30487,31653,31763,32175,32898,33048,33077,33364,33744,33942,34518,35149,35150,35238,35276,35734,35887,36015,36068,36077,36222,36294,36451,36455,36456,36484,36511,36546,36553,36554,36699,36747,36821,36854,36873,36942,37135,37197,37547,37567,37569,37678,37992,38098,38123,38127,38199,38239,38417,38512,38565,38607,38845,38851,38942,39007,39306,39319,39379,39380,39485,39486,39490,39614,39722,39783,39888,39908,40114,40129,40812,41334,42319,42494,42565,43034,43115,43119,43198,43221,43476,43522,44144,44458,45428,45438,45502,45691,45875,46198,47326,47608,48151,48198,48364,48765,49040,49428,49592,49639,50054,50068,50091,50094,50184,50296,50327,50432,50515,50616,50623,50630,50639,50640,50652,50690,50713,50720,50859,51068,51196,51224,51285,51362,51417,51484,51505,51546,51615,51628,51692,51885,51899,51908,52088,52178,52284,52401,52564,52658,52694,52701,52836,52896,53042,53100,53145,53419,53505,53523,53670,53719,53764,53786,53813,53940,54042,54181,54354,54369,54378,54401,54432,54502,54578,54643,54645,54788,54824,54887,54933,54982,55110,55118,55241,55391,55483,55561,56098,56370,56467,56592,56628,56720,56898,56940,56947,57026,57049,57075,57171,57174,57181,57247,57273,57305,57455,57471,57477,57488,57544,57588,57657,57771,57791,57918,57920,58159,58252,58593,58702,58770,58976,59025,59211,59450,59515,59574,59651,59692,59778,59952,60059,60060,60216,60262,60329,60346,60465,60662,60669,60752,60781,60810,60819,60888,61009,61067,61083,61109,61151,61158,61404,61622,61807,61849,61869,62240,62260,62266,62323,62439,62506,62508,62637,62737,62817,62863,63132,63175,63196,63203,63256,63320,63333,63383,63404,63420,63562,63645,63670,63676,63806,63866,63876,63891,63945,63949,63964,64099,64301,64315,64398,64419,64429,64442,64527,64608,64623,64664,64686,64720,64817,64927,65047,65092,65152,65182,65235,65329,65392,65402,65439,65481,65573,65711,65846,65907,65973,66323,66330,66447,66461,66503,66591,66782,66802,66838,66857,66904,67165,67209,67214,67424,67437,67438,67444,67653,67719,67735,67771,67904,68041,68096,68276,68407,68462,68567,68588,68692,68927,68948,69005,69107,69404,69408,69492,69508,69518,69600,69645,69698,69778,69828,69984,70195,70298,70342,70356,70357,70407,70681,70689,70700,70762,71032,71186,71267,71571,71765,71792,71801,71862,71949,72032,72069,72269,72459,72464,72503,72608,72623,72658,72790,72811,72827,72991,73067,73074,73091,73238,73383,73583,73824,73836,74063,74117,74161,74262,74378,74405,74862,75164,75210,75251,75261,75324,75400,75503,75545,75616,75633,75719,75739,75767,75921,76037,76154,76202,76317,76334,76352,76617,76699,76816,77079,77092,77178,77357,77411,77535,77735,77793,77890,77938,78035,78215,78257,78308,78366,78413,78427,78474,78516,78615,78771,78797,79001,79153,79206,79243,79299,79385,79516,79609,79708,79714,79742,79873,79909,79911,79949,80120,80123,80258,80304,80376,80538,80546,80631,80842,81172,81313,81323,81373,81387,81665,81718,81753,81878,82018,82043,82075,82412,82437,82490,82572,82630,82682,82722,82728,82747,82811,82820,82821,82832,82888,82898,82942,82988,83062,83436,83556,83616,83664,83689,83720,83761,83800,83869,83883,83896,83897,84065,84120,84173,84205,84512,84541,84717,84858,84862,84895,85070,85113,85128,85168,85208,85229,85267,85292,85337,85511,85519,85541,85709,85866,85881,86010,86122,86171,86406,86538,86553,86575,86663,86693,86789,86795,86806,86808,86853,86906,87042,87254,87299,87326,87327,87459,87568,87570,87625,87671,87672,87716,87745,87765,87819,87836,87949,87974,87985,88005,88097,88185,88225,88236,88499,88560,88730,88780,88799,88993,89018,89105,89283,89316,89330,89362,89819,89827,89855,89900,89953,89958,90275,90427,90453,90460,90613,90648,90889,90970,90982,91054,91093,91151,91186,91259,91281,91290,91409,91458,91485,91523,91545,91701,91707,91816,91878,92009,92065,92083,92134,92199,92246,92293,92337,92339,92384,92522,92590,92774,92906,92916,92936,92944,93096,93153,93296,93343,93346,93378,93630,93733,93906,93939,94084,94157,94366,94484,94680,94764,94783,94811,94963,94992,95055,95180,95344,95378,95462,95464,95493,95514,95602,95708,95754,95783,95823,95962,96082,96100,96111,96454,96517,96613,96628,96655,96788,96804,96872,96985,97015,97040,97187,97279,97310,97376,97396,97415,97546,97555,97732,97745,97752,97857,97860,97896,97918,98141,98361,98708,98878,98899,98936,99028,99254,99266,99333,99497,99621,99680,99798,99893,99998",
          "10,38,97,100,158,165,375,583,647,655,657,786,875,1032,1056,1058,1073,1095,1103,1109,1161,1221,1228,1257,1259,1308,1330,1493,1544,1568,1571,1639,1650,1658,1718,1723,1743,2041,2117,2120,2200,2222,2485,2668,2670,2732,2806,2824,3159,3348,3357,3734,3777,3797,4122,4259,4300,4467,4511,4672,4733,4850,4912,4952,5019,5137,5179,5238,5267,5316,5320,5520,5531,5566,5707,5761,5836,5947,6057,6059,6179,6181,6204,6246,6259,6415,6464,6694,6795,6809,6884,7072,7111,7212,7392,7446,7491,7496,7856,7871,7980,7986,7990,8035,8260,8276,8367,8454,8502,8584,8707,8723,8785,9002,9067,9298,9410,9447,9502,9514,9579,9616,9802,9855,10073,10147,10172,10174,10233,10234,10264,10332,10343,10356,10449,10495,10509,10537,10542,10557,10580,10721,10814,10839,11049,11077,11108,11310,11400,11462,11533,11555,11655,11664,11747,11843,12182,12184,12319,12651,12669,12715,12783,12829,12961,13077,13129,13150,13341,13445,13619,13824,13892,13961,14038,14108,14412,14458,14494,14545,14577,14591,14605,14882,14932,15020,15084,15095,15205,15260,15278,15392,15550,15699,15715,15725,15759,15830,16038,16153,16445,16516,16588,16596,16663,16719,16781,16912,16934,17075,17131,17143,17184,17238,17314,17332,17446,17523,17692,17786,17826,17865,18000,18173,18194,18251,18323,18376,18524,18698,18708,18831,18935,18957,18982,19004,19060,19114,19326,19447,19471,19581,19635,19677,19744,19799,19917,20009,20021,20044,20046,20048,20051,20677,20715,20922,20980,21030,21144,21193,21705,21808,21836,21872,21887,21949,21992,22235,22338,22366,22423,22454,22520,22710,23153,23246,23531,23765,24305,24306,24438,24475,24533,24592,24637,24643,24793,24875,24930,25008,25052,25292,25311,25351,25422,25583,25680,25862,26148,26295,26312,26330,26489,26749,26891,26931,27050,27191,27549,27645,27654,27663,27684,27712,27721,27875,27906,27940,27973,28280,28383,28393,28441,28681,28693,28711,28863,28883,28958,29021,29045,29195,29243,29260,29267,29306,29322,29333,29339,29425,29566,29736,29877,30003,30352,30381,30787,31135,31163,31280,31322,31397,31570,31687,31831,32381,32499,32586,32962,32963,33050,33454,33484,33485,33513,33687,34498,34557,34674,34731,34759,34847,34993,35004,35294,35635,35713,35814,36078,36198,36254,36312,36322,36463,36509,36559,36560,36701,36745,36765,36820,36829,37008,37079,37265,37441,37443,37554,37565,37805,37870,37879,37946,37947,38011,38125,38207,38316,38382,38442,38742,39251,39269,39320,39428,39504,39568,39611,39949,39993,40147,40256,40517,41039,41245,41303,41481,42061,42197,42234,42289,42398,42616,42676,42680,42745,42804,42838,42846,42875,42968,43061,43172,43206,43264,43348,43385,43768,43824,44208,44305,44448,44477,44515,44570,44677,44779,44809,44825,44830,44882,45071,45325,45498,45504,45616,45689,45713,45814,45877,45986,45988,46059,46441,46491,46529,46594,46616,46772,46838,47226,48089,48178,48420,48498,48542,49032,49319,49603,49853,50354,50413,50436,50685,50851,51289,51324,51326,51446,51535,51613,51681,51828,51894,51940,52110,52133,52245,52405,52451,52483,52572,52633,52688,52713,52720,52731,52775,52831,52977,53007,53112,53167,53295,53301,53396,53450,53485,53524,53575,53580,53643,53705,53886,53987,53989,53992,53998,54130,54137,54195,54215,54224,54238,54290,54325,54527,54599,54611,54619,54631,54712,54713,54714,54753,54851,54872,54911,55101,55109,55164,55363,55402,55403,55486,55535,55555,55585,55992,56095,56117,56181,56233,56304,56354,56394,56465,56567,56572,56596,56624,56709,56724,56860,57025,57053,57084,57153,57217,57329,57345,57508,57674,57739,57760,57938,57973,58026,58048,58127,58139,58177,58245,58258,58429,58619,58642,59010,59086,59207,59312,59481,59551,59655,59693,59739,59895,59937,60099,60244,60287,60359,60451,60560,60705,60987,61016,61074,61116,61226,61401,61696,61707,61829,61896,62011,62054,62075,62140,62237,62242,62291,62434,62514,62556,62577,62635,62724,62741,62746,62837,62939,63044,63088,63146,63354,63382,63431,63580,63592,63753,63777,63863,63883,63935,63963,64291,64425,64473,64492,64499,64571,64583,64607,64635,64963,64982,65114,65229,65515,65707,65834,65890,65899,65969,66018,66122,66139,66272,66294,66454,66537,66541,66594,66648,66735,66863,66881,67004,67010,67114,67154,67198,67205,67231,67288,67321,67457,67473,67560,67691,67723,67784,67928,68003,68045,68070,68105,68106,68116,68122,68183,68317,68335,68384,68401,68404,68409,68437,68606,68803,68893,68933,68995,69056,69152,69191,69278,69382,69384,69420,69453,69500,69610,69674,69687,69795,69926,69973,70051,70145,70196,70264,70279,70305,70330,70393,70397,70436,70515,70718,70956,71330,71475,71479,71489,71614,71719,71751,71866,71941,71975,72175,72230,72250,72284,72326,72422,72500,72515,72543,72715,72737,72843,72902,72908,72978,73057,73098,73323,73348,73433,73537,73554,73594,73612,73700,73734,73758,73818,73940,74004,74016,74018,74130,74228,74319,74323,74332,74336,74416,74503,74526,74670,74673,74784,74807,75022,75242,75348,75350,75406,75493,75500,75779,75789,75840,75871,75892,76008,76013,76044,76075,76083,76212,76378,76463,76465,76476,76509,76560,76601,76936,76941,77042,77194,77272,77299,77414,77420,77422,77463,77491,77504,77698,77800,77865,77924,77964,78185,78288,78339,78374,78403,78582,78695,78755,78770,78777,78780,78801,78854,78928,79118,79135,79199,79371,79555,79616,79666,79667,79679,79888,79913,80000,80186,80267,80396,80418,80436,80575,80649,80780,80935,80962,80982,81249,81345,81564,81701,81724,81748,81882,81952,81968,82154,82210,82213,82274,82275,82386,82494,82783,82861,82950,82997,83039,83336,83406,83427,83497,83584,83585,83879,83935,83948,84058,84152,84330,84353,84409,84645,84881,85428,85451,85471,85597,85815,85862,86033,86079,86088,86163,86174,86206,86339,86348,86433,86469,86576,86616,86801,86914,86926,86955,87003,87033,87063,87066,87076,87220,87393,87444,87508,87541,87554,87786,87842,87983,88033,88083,88107,88115,88222,88280,88303,88387,88400,88452,88475,88641,88685,89046,89066,89072,89312,89345,89413,89455,89519,89618,89645,89791,89835,89850,89856,89932,89935,90125,90193,90213,90451,90455,90487,90495,90546,90653,90705,90926,90938,91004,91109,91185,91265,91450,91556,91566,91570,91649,91732,91737,91893,92069,92108,92158,92167,92292,92310,92356,92404,92407,92431,92533,92538,92551,92731,92995,93285,93302,93331,93448,93619,93756,93956,93987,94097,94148,94169,94275,94413,94453,94462,94482,94533,94652,94657,94658,94750,94850,94951,95138,95346,95390,95528,95569,95675,95681,95713,96027,96053,96101,96121,96154,96182,96212,96255,96304,96354,96432,96538,96562,96595,96608,96642,96686,96706,96765,96885,97027,97051,97079,97105,97144,97341,97508,97576,97692,97799,97943,97992,97999,98076,98119,98234,98276,98323,98343,98435,98678,98756,98826,98916,98948,98964,99020,99024,99101,99300,99384,99392,99399,99407,99455,99614,99666,99748,99880,99912,99987",
          "25,93,117,154,212,265,281,618,660,776,960,1047,1066,1113,1130,1204,1241,1340,1383,1416,1478,1562,1598,1619,1714,1726,1766,1777,1914,2128,2263,2304,2317,2509,2527,2569,2948,3066,3580,3617,3674,3715,4378,4620,4777,5174,5207,5211,5227,5232,5269,5276,5288,5337,5345,5416,5446,5458,5462,5472,5476,5487,5593,5661,5670,5701,5732,5735,5795,5869,5930,5973,5977,5996,6060,6083,6175,6217,6233,6268,6269,6281,6387,6390,6589,6681,6684,6705,6786,6827,6851,6911,6943,7030,7043,7061,7069,7300,7318,7467,7562,7592,7610,7751,7753,7759,7791,7799,7845,7859,7878,7883,7897,7947,8308,8442,8452,8548,8582,8680,8871,8960,8991,9000,9015,9032,9038,9118,9176,9784,9870,9964,9966,10011,10019,10133,10201,10367,10737,10758,10795,10910,11294,11337,11354,11577,11650,11767,11775,11791,11869,12305,12307,12641,12657,12743,12785,12943,12948,12966,12970,13057,13115,13364,13400,13867,13868,14066,14336,14466,14514,14621,14750,14771,14940,15058,15109,15293,15448,15474,15584,15660,15754,16072,16709,17069,17259,17357,17455,17503,17554,17586,17711,17763,18079,18428,18572,18640,18722,18922,18971,19124,19260,19273,19626,19777,20885,21126,21142,21402,22160,22669,22708,22741,23007,23064,23318,23589,23615,23938,24271,24597,24638,24650,25087,25395,25705,25983,26377,26415,26573,26628,26838,26898,27003,27213,27273,27314,27825,27881,27968,28061,28270,28388,28757,28864,28979,29132,29420,29796,29828,31047,31419,31564,32521,32637,33275,33768,33826,34377,34551,35508,35627,35696,35698,35737,35891,36122,36494,36574,36680,36693,36926,37425,38357,38424,38695,38790,38849,39122,39247,39301,39870,40460,42154,42308,42325,42373,42456,42640,42708,42718,42795,42809,43100,43710,43884,43892,43954,44420,45178,45313,45447,45539,45935,46855,46955,47147,48185,48457,48835,49390,49518,50057,50067,50073,50362,50439,50480,50613,50620,50879,50906,51011,51051,51074,51084,51124,51186,51236,51408,51472,51480,51611,51887,51892,52081,52104,52204,52225,52296,52384,52391,52436,52542,52583,52819,52901,52910,52941,53289,53474,53935,54080,54138,54175,54182,54219,54239,54264,54312,54319,54391,54431,54553,54778,54816,54823,54917,54944,54947,54968,55004,55248,55269,55401,55494,55590,55683,55726,55825,55834,55881,55953,56080,56183,56222,56426,56531,56538,56546,56686,56827,56906,56945,56946,57052,57151,57177,57193,57222,57232,57292,57314,57335,57381,57403,57515,57614,57709,57752,57757,57805,57867,58120,58280,58609,58620,58849,58850,59063,59215,59273,59313,59473,59548,59628,59725,59770,59998,60050,60264,60418,60518,60547,60587,60731,60806,61010,61058,61273,61294,61314,61365,61388,61423,61440,61446,61460,61526,61645,62105,62124,62127,62153,62212,62226,62378,62450,62463,62464,62562,62663,62710,62745,62829,62840,62858,62913,62944,62988,63091,63126,63328,63506,63601,63602,63646,63780,63877,63884,63929,63959,63987,64032,64210,64228,64314,64457,64522,64609,64706,64708,64756,64765,64955,64967,65038,65073,65282,65312,65320,65427,65456,65473,65630,65926,66030,66213,66366,66367,66422,66689,66824,66865,66999,67013,67099,67148,67189,67302,67317,67639,67957,68056,68089,68190,68226,68241,68294,68596,68617,68675,68801,68841,68973,69064,69187,69248,69346,69389,69504,69542,69545,69587,69603,69725,69831,69912,69938,69972,69989,70187,70382,70426,70466,70518,70534,70563,70639,70640,70758,70768,70773,70955,70982,71000,71086,71144,71167,71199,71240,71410,71416,71484,71520,71563,71679,71850,72029,72089,72125,72205,72242,72322,72380,72453,72472,72544,72545,72613,72628,72663,72674,72729,72776,72819,72997,73008,73010,73012,73062,73069,73122,73231,73299,73358,73463,73511,73600,73623,73651,73665,73749,74057,74149,74150,74188,74372,74489,74740,74773,74782,74843,74880,74947,74948,75007,75056,75058,75067,75089,75271,75342,75538,75548,75750,75956,76047,76052,76147,76169,76389,76583,76613,76636,76677,76711,76760,76785,76825,76851,76876,76963,76982,76998,77018,77125,77202,77288,77321,77359,77406,77415,77438,77541,77559,77597,77742,77768,77806,77810,77866,77958,77962,78062,78069,78192,78322,78410,78541,78649,78826,78900,78907,78968,78974,79087,79168,79173,79223,79352,79497,79546,79572,79587,79604,79621,79761,79781,79796,79845,79876,80002,80424,80447,80514,80571,80599,80600,80604,80669,80690,80693,80722,80759,81059,81113,81209,81224,81261,81365,81432,81497,82087,82140,82218,82225,82483,82525,82540,82743,82802,82884,82890,82905,82930,82984,82999,83001,83020,83203,83232,83277,83293,83354,83368,83437,83447,83455,83467,83565,83603,83649,83651,83722,83724,83829,83846,83880,83884,83898,83929,83932,83995,84117,84155,84170,84214,84406,84445,84469,84477,84497,84608,84664,84665,84681,84684,84724,84841,84901,84939,84959,85000,85132,85266,85333,85354,85390,85419,85441,85477,85580,85631,85641,85688,85711,85900,86349,86361,86481,86494,86599,86709,86843,86854,86932,87067,87149,87227,87233,87252,87311,87492,87578,87582,87738,87750,87821,87874,87875,87951,87998,88128,88151,88246,88376,88529,88592,88653,88672,88897,88903,88934,89089,89225,89268,89332,89444,89484,89537,89589,89981,90110,90173,90384,90404,90412,90417,90509,90755,90768,90831,90882,91016,91019,91183,91635,91790,91842,91884,91980,92316,92447,92706,92867,92885,93183,93196,93274,93405,93439,93483,93485,93582,93757,93890,93903,93955,93979,94000,94013,94042,94184,94189,94196,94235,94288,94370,94408,94452,94477,94555,94706,94744,94773,94805,94860,94936,95021,95124,95206,95212,95365,95373,95451,95456,95604,95657,95693,95833,95963,96070,96116,96257,96266,96338,96346,96437,96615,96735,96774,96796,96812,96997,97208,97224,97323,97427,97464,97494,97535,97560,97661,97891,97987,98038,98176,98290,98383,98434,98478,98535,98578,98617,98720,98778,98859,98950,98997,99100,99104,99246,99411,99560,99671,99691,99766,99817,99920",
          "67,255,270,275,306,329,330,349,390,399,485,513,616,676,692,701,736,740,850,1031,1054,1146,1358,1359,1421,1550,1609,1893,2850,3130,3166,3615,3747,4007,4423,4432,4835,4894,5237,5305,5466,5558,5609,5639,5657,5860,5997,6109,6122,6211,6249,6274,6282,6329,6355,6573,6614,6633,6671,6762,6799,6905,6956,7025,7244,7264,7325,7375,7379,7480,7542,7603,7609,7613,7730,7738,7739,7837,7982,8141,8271,8318,8357,8381,8403,8435,8466,8558,8589,8644,8832,8849,8984,9048,9087,9096,9202,9340,9389,9475,9746,9900,10076,10153,10178,10215,10281,10472,10508,10544,10565,10615,10665,10686,10695,10714,10716,10728,11091,11098,11311,11318,11335,11568,11661,11815,11839,11894,11988,12053,12100,12108,12303,12460,12466,12497,12541,12581,12711,13032,13080,13256,13315,13633,14542,14668,14754,14833,15375,15385,15412,15540,15674,15975,16231,16414,16630,16660,16952,17065,17276,17391,17681,18272,18384,18511,18658,18704,18753,18882,19270,19487,19583,19733,19787,19977,20126,20224,20260,20284,20501,20797,20828,20830,21137,22137,22191,22728,22769,22974,24155,24185,24694,24914,25518,25634,25711,25822,26130,26314,26817,26908,27009,27037,27160,27467,27470,27707,27737,27802,28287,28505,28898,28904,28925,29000,29104,29344,29362,29406,29509,29586,29640,29999,30868,30986,35340,35576,35577,35806,35951,35957,36117,36381,36736,36741,36947,37051,37202,37240,37427,37622,37625,37676,37694,37735,37949,37987,38069,38070,38120,38604,38688,38984,39054,39133,39238,39250,39260,39378,39478,39612,40567,41740,42148,43267,43602,45245,45506,48349,50555,50559,50561,50618,50682,50797,50926,50931,50969,51091,51112,51136,51172,51250,51312,51318,51345,51494,51588,51731,51818,51882,51926,51948,52057,52076,52086,52094,52147,52290,52297,52409,52500,52514,52635,52698,52807,52866,52982,52988,53026,53491,53528,53534,53545,53596,53780,53960,54003,54031,54103,54190,54197,54248,54262,54310,54390,54651,54789,54791,54857,54953,55372,55524,55560,55588,55597,55637,55646,55869,55884,56040,56045,56149,56271,56276,56489,56580,56661,56786,56792,56819,56850,56859,57424,57425,57440,57591,57792,57799,57831,57868,57931,58011,58241,58260,58583,58603,58607,58650,58695,58703,58847,59128,59219,59322,59458,59488,59592,59621,59636,59753,59782,59856,59909,59996,60054,60084,60143,60163,60172,60191,60200,60288,60401,60424,60486,60501,60667,60793,60828,60901,60983,61019,61125,61508,61571,61669,61671,61988,62068,62096,62115,62272,62281,62290,62569,62620,62722,62816,62959,63125,63157,63267,63324,63367,63444,63483,63722,63779,63800,63822,63916,63937,64071,64191,64305,64329,64460,64566,64675,64794,64844,64972,65217,65258,65278,65319,65333,65428,65535,65555,65588,65716,65743,65778,66016,66041,66100,66219,66248,66299,66309,66542,66699,66755,66916,66944,67048,67123,67206,67229,67379,67380,67429,67469,67474,67619,67620,67852,68032,68060,68180,68188,68366,68479,68552,68555,68584,68590,68934,68984,69076,69200,69224,69381,69407,69417,69658,69767,70180,70422,70464,70476,70513,70533,70589,70646,70674,70767,70836,71002,71013,71149,71220,71260,71309,71385,71438,71491,71673,71755,71782,71789,71800,71842,71922,71935,72127,72197,72266,72320,72337,72388,72401,72413,73037,73121,73219,73317,73346,73371,73446,73476,73560,73636,73701,73786,73946,73990,74064,74073,74102,74246,74474,74569,74644,74823,74952,75009,75049,75122,75197,75338,75356,75502,75558,75611,75705,75828,75864,75946,76198,76261,76309,76437,76481,76541,76546,76551,76607,76626,76647,76678,77019,77040,77043,77102,77192,77224,77277,77324,77330,77331,77515,77676,77766,77854,77992,78078,78113,78117,78279,78310,78485,78537,78605,78629,78635,78680,78703,78753,78833,78943,78977,79009,79036,79120,79232,79234,79340,79614,79677,79754,79790,79835,79951,79984,80025,80093,80287,80312,80366,80598,80806,80826,81029,81204,81279,81301,81338,81425,81476,81522,81524,81559,81601,82089,82233,82314,82427,82641,82733,82776,82929,82980,83003,83183,83214,83312,83331,83438,83472,83480,83531,83570,83583,83672,83698,84013,84084,84243,84252,84256,84460,84545,84548,84595,84695,84705,84867,84987,85112,85176,85312,85316,85351,85366,85446,85447,85508,85514,85522,85568,85668,85721,85800,85818,85897,85980,86043,86068,86105,86113,86144,86244,86357,86434,86452,86501,86537,86541,86551,86559,86588,86620,86678,86699,86834,86900,86907,86939,86980,87024,87097,87199,87432,87456,87892,87959,88069,88117,88488,88519,88520,88694,88705,88811,88857,89051,89110,89165,89275,89457,89664,89703,89710,89769,89812,90148,90210,90323,90426,90492,90523,90647,90762,90771,90801,90813,90821,90823,90824,90872,90895,90917,90922,90980,91105,91106,91145,91230,91399,91474,91480,91669,91681,91687,91756,91786,91833,91900,91928,91996,92010,92132,92255,92376,92487,92531,92552,92572,92586,92659,92899,92900,93026,93157,93177,93482,93489,93493,93574,93952,94054,94142,94201,94216,94225,94356,94359,94383,94509,94537,94625,94702,94724,95014,95062,95225,95230,95370,95448,95622,95712,95760,95790,95800,95894,96168,96210,96314,96357,96358,96400,96519,96527,96574,96593,96610,96718,96878,96922,96934,97026,97258,97296,97336,97468,97527,97574,97604,97606,97641,97649,97686,97894,97920,98032,98049,98050,98053,98084,98149,98199,98215,98270,98351,98362,98363,98376,98530,98555,98569,98601,98728,98974,99124,99189,99202,99286,99338,99786,99911,99924",
          "33,243,260,437,501,533,538,632,679,694,703,894,896,901,949,954,959,1055,1089,1092,1172,1178,1179,1355,1402,1533,1534,1567,1885,1887,1892,2202,2208,2406,2417,2419,3218,3289,3290,3297,3298,3360,3434,3575,3596,3735,3886,4180,4240,4608,4732,4913,5012,5087,5106,5121,5128,5185,5204,5235,5290,5344,5346,5356,5365,5378,5550,5554,5599,5767,5797,5855,5879,5909,6034,6147,6163,6184,6230,6280,6426,6475,6478,6512,6631,6700,6807,6895,6933,7182,7419,7713,7873,7895,8025,8054,8085,8186,8228,8351,8383,8451,8487,8594,8610,8718,8763,8805,8851,8877,8888,8978,8993,9057,9156,9164,9254,9347,9371,9515,9544,9549,9876,9907,9913,9931,9940,9974,10013,10018,10268,10278,10294,10410,10414,10442,10682,10683,10739,10891,10946,11151,11153,11239,11471,11811,11905,12019,12079,12155,12304,12461,12513,12805,12817,12935,12990,13252,13351,13458,13523,13629,13853,13999,14006,14040,14072,14134,14150,14324,14365,14438,14513,14684,14709,14710,14749,14780,14812,14970,15087,15452,15463,15504,15580,15606,15655,15769,15972,16085,16332,16493,16591,16684,16999,17098,17227,17358,17836,18201,18222,18273,18287,18390,18619,18653,18681,18771,18849,19071,19173,19217,19454,19492,19527,19568,19746,19793,19929,20154,20244,20252,20259,20328,20334,20579,20819,20906,20960,21118,21698,21722,21807,22029,22133,22415,22545,22586,22627,22763,22837,23011,23226,23308,23326,23433,23694,23742,23840,23843,24335,24345,24379,24405,24715,24934,24944,24954,25265,25316,25331,25372,25467,25572,25723,25759,25800,26346,26650,26708,26959,27032,27043,27131,27171,27251,27298,27413,27450,27482,27592,27622,27752,27793,27888,27915,27985,28276,28387,28402,28730,28742,28755,28830,28891,28927,28963,29046,29122,29183,29222,29274,29450,29505,29544,29548,29623,29651,29751,30162,30394,30549,30633,30778,30829,30836,31010,31303,31438,31445,32460,32517,33330,33565,33736,33862,34055,34841,34986,35043,35053,35106,35623,35823,35890,35936,36000,36006,36119,36305,36487,36513,36572,36627,36718,36758,36875,36994,37013,37075,37220,37345,37543,37559,37873,37986,38117,38192,38301,38371,38572,38628,38780,38798,38813,38922,38972,39099,39189,39220,39289,39299,39420,39547,39633,39735,39744,39813,39864,40159,40449,40451,40855,41297,41355,41466,42553,42844,42850,43023,43489,43586,43709,43726,43790,44022,44062,44116,44233,44592,44618,44734,44762,44781,44869,45016,45044,45097,45137,45242,45554,45639,45745,46016,46109,46632,46695,46968,47117,47224,47242,47523,48274,48302,48512,48656,49060,49223,49237,49577,49777,50000,50012,50086,50189,50333,50414,50440,50514,50521,50536,50597,50681,50692,50707,50815,50831,51045,51133,51259,51292,51330,51483,51549,51662,51864,51999,52046,52226,52286,52317,52374,52377,52556,52641,52806,52851,52908,52916,52954,53052,53059,53156,53165,53268,53307,53420,53549,53627,53628,53681,53693,53695,53703,53866,54008,54163,54349,54405,54438,54518,54674,54852,54869,54905,55027,55038,55050,55127,55225,55232,55317,55328,55474,55487,55511,55530,55542,55761,55779,55822,55890,56046,56332,56562,56676,56677,56739,56764,56889,56990,57055,57108,57155,57235,57257,57270,57368,57550,57661,57828,57881,58140,58477,58768,58823,58871,58875,58891,58912,58950,59139,59394,59438,59532,59572,59639,59704,59873,59907,59934,60138,60145,60222,60233,60284,60361,60417,60461,60534,60565,60609,60622,60802,60827,60856,60859,60880,60947,60955,61036,61040,61208,61227,61282,61329,61472,61500,61531,61554,61564,61676,61731,61820,61930,61955,62137,62285,62339,62382,62392,62429,62431,62550,62582,62853,62970,63027,63051,63062,63083,63159,63189,63217,63247,63252,63254,63327,63430,63434,63466,63516,63584,63657,63668,63679,63778,63837,63853,63867,63948,63956,64006,64147,64151,64287,64373,64487,64537,64560,64823,64879,64943,65123,65206,65430,65632,65801,65807,65810,65933,65999,66022,66065,66249,66381,66528,66621,66814,66820,66868,66937,67027,67228,67307,67326,67387,67528,67559,67616,67671,67755,67800,67929,67952,68016,68036,68068,68338,68402,68426,68621,68652,68666,68712,68769,68879,68894,68900,69028,69103,69180,69190,69269,69419,69465,69509,69516,69556,69557,69609,69675,69780,69799,69846,69868,69957,69980,70015,70124,70472,70759,70839,70862,70901,71123,71151,71306,71399,71464,71676,71916,71917,72059,72179,72314,72347,72354,72427,72508,72532,72664,72710,72730,73216,73296,73327,73392,73412,73424,73427,73590,73918,73922,73968,74092,74195,74204,74274,74290,74353,74498,74588,74592,74676,74855,74866,75189,75267,75343,75358,75403,75436,75466,75547,75731,75838,75900,75923,75982,76011,76062,76069,76103,76104,76248,76255,76275,76293,76295,76356,76499,76515,76597,76757,76885,76985,76994,77088,77161,77214,77218,77276,77317,77349,77752,77761,77982,78094,78115,78391,78467,78483,78536,78538,78800,78811,78852,78880,79244,79266,79311,79451,79485,79649,80024,80052,80128,80246,80257,80557,80579,80641,80657,80715,80739,80811,80872,80979,80989,81002,81096,81108,81120,81228,81494,81596,81755,81763,81935,82011,82028,82197,82305,82348,82433,82611,82650,82657,82710,82788,83134,83137,83187,83220,83410,83428,83489,83525,83595,83778,83876,83959,83962,83994,84011,84087,84140,84249,84334,84357,84598,84602,84625,84668,84678,84790,84848,84997,85061,85114,85159,85161,85179,85231,85284,85392,85490,85600,85810,85958,85963,86159,86165,86234,86316,86422,86465,86484,86485,86516,86596,86731,86775,86847,87013,87100,87124,87175,87223,87346,87389,87484,87597,87707,87814,87964,87991,88017,88047,88153,88398,88412,88564,88568,88627,88644,89020,89103,89108,89199,89346,89360,89419,89436,89604,89627,89717,89790,89921,89949,89992,90016,90143,90147,90195,90276,90317,90364,90469,90586,90610,90635,90792,90803,91020,91088,91100,91114,91137,91214,91215,91333,91366,91515,91553,91663,91929,92195,92261,92287,92320,92411,92436,92494,92611,92677,92905,92910,92931,92958,92970,92998,93066,93181,93206,93217,93352,93358,93414,93450,93452,93490,93559,93646,93695,93702,93719,93720,93740,93742,93800,93853,93866,93909,94001,94053,94161,94165,94314,94457,94461,94483,94792,94944,95075,95154,95229,95327,95330,95475,95649,95905,95954,96037,96239,96246,96361,96409,96509,96597,96618,96637,96691,96693,96694,96794,96910,96952,97014,97017,97157,97184,97311,97413,97488,97507,97591,97733,97761,97878,97889,98090,98111,98112,98275,98285,98371,98520,98553,98600,98629,98652,98694,98732,98868,98872,98952,98985,99145,99259,99337,99400,99424,99431,99559,99771,99801,99855,99889,99891,99917,99985",
          "81,113,153,236,246,1233,1981,2058,2096,2135,2136,2204,2300,2412,2452,2587,2809,2879,3007,3054,3190,3204,3228,3291,3683,4052,4215,4441,4524,4548,4715,4801,4809,5684,6350,6638,6655,6814,6997,7185,7205,7255,7529,7548,7830,7945,8002,8004,8087,8118,8128,8137,8156,8175,8232,8270,8294,8319,8346,8416,8424,8650,8662,8669,8840,8891,8897,8975,9039,9100,9134,9201,9238,9296,9397,9480,9619,9657,9695,9728,9753,9759,9828,9847,9942,9965,9986,10075,10081,10094,10113,10208,10236,10303,10360,10411,10455,10521,10588,10645,10656,10763,10835,10930,11036,11158,11216,11234,11285,11291,11386,11413,11480,11575,11614,11652,11709,11918,11938,12034,12097,12203,12222,12241,12384,12388,12393,12450,12479,12488,12602,12698,12716,12724,12735,12794,12849,12887,12963,13016,13020,13039,13172,13192,13267,13343,13419,13436,13534,13590,13609,13647,13856,13924,13932,13989,14008,14071,14121,14179,14216,14217,14236,14255,14274,14383,14388,14424,14440,14521,14540,14616,14619,14635,14691,14703,14787,14920,15015,15039,15053,15091,15121,15148,15262,15281,15300,15357,15369,15396,15471,15475,15566,15604,15623,15642,15661,15694,15737,15756,15832,15851,15870,15889,15984,16003,16098,16193,16250,16288,16326,16345,16383,16402,16440,16442,16450,16459,16497,16592,16649,16668,16706,16782,16839,16877,16896,16915,16991,17067,17141,17253,17257,17334,17352,17447,17466,17485,17580,17637,17694,17713,17732,17789,17827,17846,17884,17922,18017,18036,18131,18219,18225,18226,18245,18283,18302,18359,18378,18397,18530,18568,18606,18663,18682,18758,18796,18815,18872,18910,18929,19176,19237,19252,19271,19309,19347,19385,19442,19499,19588,19605,19651,19727,19765,19860,19879,19936,19955,20012,20031,20107,20115,20145,20152,20164,20240,20273,20373,20411,20430,20449,20468,20487,20525,20639,20753,20810,20821,20829,20843,20848,20886,20981,20998,21095,21171,21190,21293,21304,21323,21342,21456,21475,21494,21570,21589,21627,21665,21684,21912,21969,22045,22083,22121,22178,22216,22254,22273,22311,22349,22368,22406,22539,22558,22599,22615,22634,22653,22672,22729,22748,22783,22786,22881,22900,22919,22921,22957,22993,23014,23071,23108,23147,23166,23185,23261,23337,23348,23356,23375,23413,23432,23489,23546,23603,23641,23660,23679,23755,23774,23869,23888,23926,23945,23964,23970,23983,24002,24021,24040,24116,24135,24265,24268,24439,24572,24629,24648,24705,24743,24800,24805,24819,24821,24935,24971,24990,25104,25105,25123,25180,25199,25332,25369,25427,25465,25471,25484,25560,25579,25606,25626,25712,25769,25788,25789,25807,25902,25921,25997,26016,26073,26092,26213,26235,26370,26472,26510,26529,26535,26649,26871,27023,27270,27480,27840,27846,28068,28179,28285,28309,28333,28410,28448,28562,28600,28733,28771,28773,28901,28922,29051,29073,29170,29189,29398,29458,29648,29816,29875,29949,29987,30063,30121,30179,30208,30237,30295,30353,30382,30411,30440,30469,30527,30556,30643,30672,30701,30730,30846,30933,30991,31020,31049,31107,31136,31165,31194,31223,31252,31281,31484,31513,31542,31571,31600,31629,31658,31745,31803,31832,31919,31977,32035,32151,32267,32383,32412,32470,32528,32557,32644,32673,32731,32818,32847,32934,32992,33021,33137,33166,33195,33224,33253,33311,33340,33369,33427,33514,33543,33601,33630,33688,33775,33804,33833,33891,33949,34065,34094,34123,34181,34210,34239,34268,34297,34355,34384,34413,34442,34471,34529,34587,34703,34732,34761,34790,34848,34906,34935,34964,35005,35031,35047,35056,35067,35094,35168,35241,35331,35365,35430,35436,35476,35517,35524,35642,35774,35815,35833,35904,35963,36022,36085,36145,36236,36275,36334,36393,36428,36498,36583,36706,36755,36800,36954,37026,37085,37267,37294,37326,37364,37515,37587,37646,37705,37769,37887,37958,37988,38017,38148,38183,38241,38389,38432,38479,38566,38637,38709,38768,38827,38891,39009,39129,39134,39198,39227,39233,39270,39369,39388,39452,39511,39570,39700,39831,39890,39978,40227,40285,40314,40343,40372,40430,40488,40546,40575,40604,40633,40691,40749,40807,40894,40952,41068,41126,41155,41213,41271,41329,41387,41416,41445,41474,41503,41532,41846,42032,42119,42159,42244,42253,42295,42379,42419,42458,42532,42581,42735,42743,42790,42799,42852,42893,43004,43055,43084,43177,43207,43208,43217,43282,43311,43319,43494,43501,43604,43623,43673,43691,43766,43841,43870,43902,43928,44003,44078,44122,44207,44217,44220,44240,44267,44276,44315,44402,44464,44639,44768,44789,44858,44876,44951,44980,44994,45026,45099,45147,45208,45266,45339,45396,45454,45478,45526,45583,45641,45770,45828,45900,46017,46039,46092,46136,46163,46185,46193,46208,46292,46423,46588,46697,46715,46733,46832,46884,46930,47106,47253,47258,47276,47370,47383,47389,47396,47433,47509,47546,47573,47644,47740,47781,47854,47901,48013,48043,48077,48148,48155,48335,48423,48504,48596,48623,48692,48702,48707,48752,48844,48885,48971,49038,49064,49068,49116,49136,49159,49187,49199,49217,49402,49457,49491,49509,49516,49548,49699,49718,49804,49805,49825,49867,49884,49886,49913,49926,49965,49994,50123,50217,50234,50249,50257,50267,50328,50469,50495,50545,50638,50683,50701,50712,50793,50808,50849,50987,51092,51127,51240,51325,51496,51621,51697,51746,51781,51800,51863,51914,51956,51975,52025,52234,52298,52418,52468,52472,52479,52898,52917,53041,53090,53179,53344,53376,53412,53422,53471,53478,53525,53585,53655,53696,53709,53749,53851,53856,53944,53967,54004,54091,54268,54435,54465,54520,54568,54576,54813,54870,54900,54924,55073,55178,55218,55430,55497,55598,55756,56052,56097,56116,56155,56204,56214,56254,56333,56357,56419,56430,56454,56547,56659,56678,56698,56717,56934,56998,57124,57159,57240,57248,57289,57354,57480,57577,57668,57750,57769,57872,57979,58112,58279,58643,58669,58722,58910,59013,59029,59046,59087,59167,59242,59342,59385,59442,59464,59640,59682,59779,59841,59843,59878,59914,59953,60193,60261,60302,60444,60581,60693,60703,60939,60946,60950,61064,61133,61167,61172,61400,61469,61525,61624,61689,62045,62130,62389,62424,62430,62546,62560,62666,62680,62846,62848,63003,63421,63443,63572,63588,63614,63662,63798,63846,63893,63993,64036,64063,64149,64156,64217,64365,64392,64406,64510,64552,64628,64661,64672,64699,64734,64739,64834,64855,64882,65006,65025,65170,65178,65179,65234,65334,65349,65432,65617,65629,65710,65833,65837,65839,65888,65928,65957,66003,66378,66592,66674,66703,66788,67042,67075,67163,67188,67197,67314,67348,67433,67466,67489,67730,67955,68010,68109,68327,68403,68427,68850,69050,69121,69338,69551,69575,69733,69885,70061,70130,70214,70308,70313,70323,70445,70491,70610,70619,70712,70742,70800,70887,70895,70904,71093,71219,71299,71522,71523,71637,71674,71802,71836,71924,72005,72082,72087,72095,72096,72203,72290,72302,72381,72462,72478,72562,72595,72644,72673,72722,72747,72846,73036,73128,73179,73197,73269,73316,73328,73387,73395,73453,73467,73589,73615,73847,73890,73950,73958,73995,74165,74181,74194,74293,74304,74310,74376,74384,74507,74522,74573,74653,74799,74842,75073,75159,75241,75318,75321,75362,75444,75454,75504,75638,75780,75806,75992,75995,76031,76035,76058,76100,76108,76161,76224,76298,76330,76355,76447,76629,76691,76826,76845,76862,76917,76931,77014,77119,77184,77382,77444,77579,77654,77669,77718,77953,77957,78105,78124,78162,78178,78236,78349,78411,78441,78504,78507,78556,78884,78979,78989,79094,79115,79129,79306,79309,79335,79360,79407,79524,79603,79760,80243,80244,80269,80286,80354,80537,80567,80656,80704,80716,80791,80827,81062,81067,81090,81103,81232,81620,81668,81726,81853,81974,82115,82130,82177,82362,82398,82566,82661,82721,82812,82834,82865,82913,83000,83093,83133,83141,83359,83377,83477,83544,83614,83615,83715,83818,84153,84168,84193,84383,84465,84511,84564,84612,84624,84627,84945,85169,85191,85220,85340,85356,85367,85385,85388,85510,85517,85555,85717,85843,85857,85896,85985,85995,86011,86040,86061,86106,86136,86176,86369,86385,86512,86577,86583,86741,86928,86978,87043,87110,87132,87152,87260,87280,87325,87329,87433,87467,87479,87500,87566,87660,87687,87692,87701,87846,87922,88095,88316,88396,88465,88477,88536,88788,88826,88885,88989,89005,89187,89370,89538,89670,89691,89916,90023,90138,90146,90166,90237,90290,90314,90315,90501,90507,90508,90749,90965,90971,90996,91049,91135,91300,91414,91456,91596,91636,91661,91686,91729,91846,91926,91987,91989,91999,92085,92156,92189,92218,92346,92418,92496,92756,92826,92921,92953,93020,93074,93099,93187,93247,93293,93350,93479,93548,93679,93840,93851,94588,94592,94611,94662,94710,94711,94767,94779,94967,95037,95249,95335,95508,95598,95682,95743,95758,95818,96000,96050,96075,96112,96142,96247,96362,96365,96387,96445,96525,96548,96566,96587,96651,96800,96891,96907,96939,96941,97030,97042,97074,97117,97156,97217,97276,97283,97300,97340,97411,97729,97765,97871,98069,98241,98294,98300,98310,98577,98837,98957,98983,99005,99067,99116,99121,99183,99437,99452,99500,99526,99607,99642,99683,99710,99733,99745,99770,99822",
          "96,110,134,198,204,323,1654,1675,1786,1940,1953,1976,1978,2007,2008,2022,2050,2057,2069,2126,2140,2154,2176,2178,2180,2186,2210,2257,2267,2313,2329,2336,2342,2354,2362,2378,2405,2411,2430,2432,2435,2436,2437,2462,2466,2469,2505,2514,2562,2571,2583,2585,2621,2631,2642,2651,2672,2677,2698,2754,2757,2783,2786,2805,2822,2827,2862,2914,2929,2944,2982,3034,3049,3086,3102,3105,3118,3120,3134,3156,3164,3180,3184,3192,3249,3250,3264,3274,3321,3342,3368,3380,3433,3452,3462,3477,3485,3487,3493,3496,3503,3510,3526,3532,3560,3562,3597,3603,3637,3656,3677,3686,3690,3694,3720,3724,3744,3784,3843,3853,3883,3906,3918,3924,3932,3951,3954,3956,3961,3967,3978,4023,4079,4092,4107,4109,4125,4149,4172,4176,4182,4187,4197,4203,4221,4226,4253,4267,4299,4312,4396,4412,4442,4459,4570,4571,4584,4599,4605,4607,4616,4660,4663,4668,4670,4702,4735,4743,4760,4811,4844,4870,4886,4963,4981,4982,6253,6257,6262,6381,6382,6389,6435,6546,6579,6583,6599,6640,6787,6833,6837,6859,6901,6967,7125,7167,7172,7218,7274,7388,7487,7502,7607,7633,7645,7690,7697,7788,7960,7984,8003,8097,8112,8248,8295,8296,8412,8446,8511,8572,8590,8621,8626,8645,8655,8665,8666,8688,8690,8779,8873,8893,8923,8940,8958,9006,9016,9027,9054,9062,9076,9145,9253,9279,9471,9507,9578,9583,9752,9777,9858,9859,9860,9949,10088,10099,10181,10217,10372,10403,10516,10536,10552,10789,10831,10864,11058,11101,11155,11269,11426,11450,11476,11538,11701,11718,11734,11763,11914,11991,12018,12044,12134,12281,12289,12324,12331,12522,12575,12780,12855,13024,13179,13199,13202,13283,13317,13389,13395,13405,13411,13478,13519,13535,13695,13720,13737,13780,13901,13936,13965,14265,14322,14341,14739,14781,14782,15168,15233,15524,15569,15761,15818,16114,16371,16502,16767,16774,16844,16918,16924,17078,17113,17335,17573,17625,17818,17866,17880,18289,18357,18411,18442,18499,19046,19243,19389,19666,19887,19909,19913,20413,20437,20537,20815,21195,21237,21392,21486,21978,22059,22244,22288,22320,22474,22486,22517,22617,22686,23102,23182,23230,23471,23866,23916,24093,24205,24251,24466,24518,24603,24827,25127,25188,25213,25306,25546,25591,25745,26029,26053,26250,26284,26336,26381,26449,26508,26516,26518,26522,26525,26551,26558,26564,26576,26583,26584,26586,26609,26612,26698,26764,26785,26801,26829,26881,26910,26951,26966,27049,27086,27089,27115,27148,27192,27232,27285,27292,27294,27337,27370,27405,27433,27443,27449,27455,27495,27498,27503,27522,27534,27536,27548,27567,27587,27598,27601,27602,27603,27605,27621,27626,27634,27650,27676,27692,27723,27797,27820,27827,27833,27866,27867,27895,27910,27950,27957,27979,27986,28008,28019,28027,28039,28041,28046,28060,28064,28097,28131,28134,28160,28165,28191,28212,28219,28231,28247,28282,28288,28305,28401,28421,28483,28520,28529,28532,28538,28560,28581,28598,28602,28613,28614,28620,28628,28657,28668,28680,28684,28706,28707,28720,28725,28726,28796,28821,28855,28861,28887,28916,28937,28953,28995,29013,29028,29038,29072,29074,29086,29091,29101,29103,29125,29147,29149,29150,29160,29220,29304,29315,29317,29355,29356,29360,29369,29374,29384,29393,29394,29410,29434,29443,29468,29476,29521,29564,29568,29595,29624,29637,29643,29649,29655,29673,29713,29787,29797,29865,29905,29909,30022,30051,30080,30109,30138,30167,30196,30225,30254,30283,30312,30341,30370,30399,30428,30457,30486,30515,30544,30573,30602,30631,30660,30689,30718,30747,30776,30805,30834,30863,30892,30921,30950,30979,31008,31037,31066,31095,31124,31153,31182,31211,31240,31269,31298,31327,31356,31385,31414,31443,31472,31501,31530,31559,31588,31617,31646,31675,31704,31733,31762,31791,31820,31849,31878,31907,31936,31965,31994,32023,32052,32081,32110,32139,32168,32197,32255,32284,32313,32342,32371,32400,32429,32458,32487,32516,32545,32574,32603,32632,32661,32690,32719,32748,32777,32806,32835,32864,32893,32922,32951,32980,33009,33038,33067,33096,33125,33154,33183,33212,33241,33270,33299,33328,33357,33386,33415,33444,33473,33502,33531,33560,33589,33618,33647,33676,33705,33763,33792,33821,33850,33879,33908,33937,33966,33995,34053,34082,34111,34140,34169,34198,34227,34256,34285,34314,34343,34372,34401,34430,34459,34488,34517,34546,34575,34604,34633,34662,34691,34720,34749,34778,34807,34836,34865,34894,34923,34952,34981,35022,35045,35111,35130,35134,35139,35185,35221,35223,35274,35283,35285,35287,35296,35313,35348,35422,35433,35464,35493,35541,35600,35659,35677,35732,35791,35850,35921,35945,35980,36039,36102,36134,36162,36194,36221,36252,36292,36351,36366,36410,36419,36482,36541,36566,36600,36608,36616,36617,36664,36723,36782,36853,36884,36912,36938,36971,37043,37102,37161,37170,37178,37225,37284,37343,37378,37414,37422,37473,37532,37604,37663,37722,37786,37815,37845,37876,37904,37975,38034,38093,38165,38224,38283,38347,38406,38418,38465,38493,38536,38553,38558,38595,38654,38671,38673,38726,38785,38844,38853,38859,38871,38908,38967,39026,39097,39156,39215,39287,39346,39405,39469,39528,39541,39546,39587,39658,39717,39741,39776,39848,39907,39966,39975,40128,40157,40186,40215,40244,40273,40302,40331,40360,40389,40418,40447,40476,40505,40534,40563,40592,40621,40650,40679,40708,40737,40766,40795,40824,40853,40882,40911,40940,40969,40998,41027,41056,41085,41114,41143,41201,41230,41259,41288,41317,41346,41375,41404,41433,41462,41491,41520,41549,42019,42048,42077,42106,42135,42144,42157,42195,42230,42285,42296,42326,42330,42360,42369,42384,42404,42435,42460,42522,42597,42634,42637,42672,42759,42778,42783,42834,42872,42909,42972,42996,43018,43030,43071,43146,43233,43308,43383,43412,43430,43470,43510,43514,43545,43559,43620,43630,43707,43753,43782,43857,43875,43944,43956,43988,44019,44025,44039,44094,44099,44153,44155,44181,44256,44295,44298,44331,44365,44418,44439,44493,44499,44502,44517,44519,44520,44568,44631,44655,44702,44703,44730,44742,44748,44761,44805,44827,44892,44967,45042,45115,45163,45194,45224,45282,45299,45355,45412,45470,45514,45542,45599,45657,45677,45729,45786,45844,45857,45916,45973,46010,46024,46038,46049,46107,46110,46135,46187,46213,46246,46248,46395,46463,46467,46473,46486,46547,46584,46585,46606,46615,46628,46635,46659,46721,46723,46759,46811,46813,46819,46849,46867,46962,47004,47165,47166,47218,47282,47341,47349,47369,47372,47379,47543,47588,47622,47636,47645,47650,47680,47703,47709,47722,47809,47844,47903,47912,47913,47931,48049,48058,48076,48078,48170,48180,48225,48227,48315,48339,48348,48433,48440,48455,48458,48484,48544,48551,48562,48576,48628,48629,48653,48751,48780,48830,48881,48934,48980,48981,48994,49006,49030,49084,49090,49095,49120,49125,49126,49208,49226,49356,49389,49461,49589,49597,49620,49635,49650,49693,49705,49730,49768,49776,49799,49818,49854,49864,49881,49919,49923,49946,49955,49958,49959,50036,50191,50241,50306,50331,50376,50404,50535,50573,50577,50677,50732,50779,50955,50965,51052,51137,51194,51392,51430,51435,51486,51584,51675,51680,51695,51801,51821,51888,51921,51934,51972,52008,52051,52328,52462,52464,52503,52585,52651,52678,52692,52700,52741,53037,53060,53111,53118,53207,53235,53345,53489,53561,53667,53685,53704,53755,53775,53803,53904,53912,53928,54071,54154,54585,54609,54785,54895,54979,55014,55045,55161,55193,55214,55256,55357,55405,55440,55576,55632,55643,55650,55654,55907,55998,56058,56382,56428,56529,56569,56631,56703,56801,56890,56925,57004,57184,57207,57336,57362,57393,57423,57426,57443,57491,57628,57637,57795,57912,57999,58022,58188,58382,58395,58479,58604,58677,58680,58759,58889,58939,58955,58964,59003,59092,59332,59388,59398,59431,59523,59722,59875,59881,59968,60224,60273,60293,60357,60365,60522,60649,60769,60850,60889,60930,61022,61105,61232,61380,61580,61632,61654,61657,61744,61805,61923,61931,61951,61954,61994,62020,62052,62108,62141,62249,62256,62262,62315,62356,62488,62650,62653,62716,62738,62842,62942,63020,63114,63127,63138,63274,63291,63334,63445,63456,63473,63857,64196,64231,64285,64344,64350,64379,64388,64393,64578,64602,64625,64741,64773,64836,64853,64956,65016,65134,65371,65376,65405,65448,65465,65491,65602,65683,65748,65760,65766,65772,65953,66158,66244,66298,66417,66472,66475,66566,66619,66635,66686,66698,66734,66773,66792,66797,66867,66878,66966,67078,67094,67130,67149,67255,67328,67356,67384,67490,67545,67836,67843,68174,68223,68280,68311,68332,68677,68721,68829,68847,68972,68976,69067,69147,69166,69183,69294,69313,69354,69369,69459,69513,69552,69579,69713,69759,69870,69905,69917,69944,69990,70021,70073,70104,70109,70191,70212,70244,70314,70521,70548,70856,70964,71072,71100,71125,71239,71251,71254,71272,71307,71375,71442,71541,71598,71611,71621,71723,71846,71884,72067,72112,72160,72350,72581,72815,72879,73094,73206,73214,73361,73617,73626,73692,73731,73742,73762,73784,73899,73944,74020,74099,74182,74258,74359,74368,74387,74440,74444,74621,74749,74860,74895,74941,75035,75081,75168,75270,75298,75376,75536,75566,75650,75753,75781,75834,75841,75852,75887,75911,75942,75955,76078,76089,76112,76168,76259,76277,76374,76592,76593,76603,76688,76747,76789,77003,77132,77135,77316,77690,77692,77703,77706,77910,77919,77948,78036,78174,78267,78338,78412,78622,78692,78810,78865,79071,79293,79345,79386,79392,79419,79506,79651,80265,80344,80614,80800,80929,80936,80949,81070,81190,81191,81330,81362,81376,81411,81482,81489,81514,81523,81539,81644,81669,81808,81864,81867,81916,81990,82093,82134,82146,82147,82258,82366,82370,82431,82656,82727,82822,82875,83098,83153,83167,83206,83215,83275,83302,83310,83322,83365,83383,83429,83485,83604,83605,83670,83693,83841,83926,84000,84053,84082,84130,84286,84439,84642,84662,84748,84786,84792,84834,84912,84992,85021,85308,85313,85349,85414,85416,85521,85884,85974,85998,86016,86021,86039,86115,86140,86245,86513,86601,86613,86662,86668,86710,86725,86885,87007,87008,87070,87184,87193,87221,87313,87426,87509,87550,87564,88030,88073,88251,88320,88660,88686,88727,88842,88844,88905,88924,88950,89106,89181,89233,89246,89392,89698,89718,89831,89950,89997,90020,90069,90135,90160,90236,90641,90671,90743,90936,90998,91141,91172,91180,91196,91232,91235,91324,91404,91413,91432,91441,91448,91486,91660,91664,92051,92055,92093,92178,92478,92530,92912,92977,93007,93058,93130,93228,93465,93469,93518,93638,93654,93673,93677,93784,93880,93969,94033,94038,94155,94188,94294,94322,94327,94469,94534,94544,94637,94665,94714,94800,94842,94897,94986,94988,95052,95096,95111,95122,95214,95224,95321,95332,95339,95465,95470,95850,96041,96181,96275,96294,96391,96633,96751,96808,96846,96926,96976,97019,97099,97243,97289,97317,97573,97631,97783,97787,97831,97834,97844,97938,97966,97988,98159,98312,98340,98409,98467,98558,98742,98921,98971,99089,99109,99268,99284,99516,99567,99612,99627,99654,99726,99840,99961"
        ]
      }
    },
    {
      "pool": {
        "type": "Pool",
        "name": "kShardSelectionHuge.02",
        "servers": [
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312"
        ],
        "shards": [
          "53,97,103,154,251,253,305,320,478,545,595,692,872,998,1040,1180,1181,1528,1611,1631,1733,1827,1881,1939,2000,2101,2196,2229,2253,2257,2285,2355,2431,2449,2508,2521,2669,2738,2876,2904,2981,2982,3007,3044,3143,3262,3263,3389,3436,3439,3499,3533,3559,3571,3629,3642,3676,3682,3842,3846,3863,3932,4033,4101,4161,4178,4209,4297,4381,4432,4453,4460,4477,4672,4713,4761,4772,4865,5065,5114,5555,5689,5801,5860,6032,6066,6246,6304,6412,6446,6530,6652,6745,6830,7090,7151,7155,7174,7327,7418,7456,7458,7509,7812,7814,7906,8095,8205,8295,8714,8727,8745,8866,8901,8930,9015,9024,9036,9227,9235,9340,9421,9475,9482,9493,9553,10071,10156,10176,10181,10409,10416,10436,10456,10554,10574,10583,10599,10683,10714,10718,10750,10805,10824,10899,11090,11100,11177,11249,11286,11405,11546,11583,11618,11704,11785,11789,11834,11921,11955,11966,12123,12134,12226,12240,12260,12366,12411,12477,12602,12686,12734,12796,12797,12836,12888,12913,12925,13017,13076,13112,13128,13221,13236,13272,13291,13381,13467,13498,13549,13733,13782,13807,13860,13949,14002,14133,14161,14202,14212,14282,14285,14299,14421,14439,14534,14542,14612,14681,14724,14747,14851,14873,14883,14953,14970,15196,15269,15276,15340,15451,15491,15555,15568,15601,15635,15790,15804,15809,15915,15941,16023,16039,16085,16344,16462,16464,16504,16586,16645,16688,16761,16834,16840,16847,16880,16957,17016,17026,17033,17103,17175,17287,17332,17431,17467,17548,17634,17781,17794,17830,17891,17922,18157,18158,18235,18358,18367,18456,18460,18469,18489,18523,18556,18566,18579,18603,18749,18812,18962,18973,19031,19081,19096,19245,19254,19277,19341,19417,19447,19476,19489,19495,19511,19597,19611,19633,19650,19658,19706,19764,19831,19862,19962,20176,20298,20365,20386,20426,20511,20552,20685,20687,20832,20871,20918,20936,21161,21529,21542,21726,21793,21847,22000,22063,22115,22209,22222,22408,22426,22489,22528,22575,22578,22599,22803,22867,22888,23044,23124,23281,23306,23411,23441,23577,23636,23670,23796,23817,23857,23871,23901,23943,23974,24034,24084,24206,24274,24319,24331,24435,24493,24501,24537,24619,24744,24751,24782,24853,24961,25061,25229,25299,25441,25448,25458,25460,25498,25503,25560,25577,25585,25674,25821,25832,25834,25928,26062,26093,26111,26184,26213,26373,26423,26454,26468,26546,26586,26589,26751,26792,26874,26907,26986,27061,27242,27251,27315,27342,27413,27473,27549,27567,27609,27628,27772,27816,27838,27904,27973,28000,28024,28033,28139,28293,28356,28396,28411,28680,28702,28718,28746,28894,28895,29011,29091,29160,29174,29226,29263,29305,29398,29845,29856,29893,30112,30123,30234,30287,30338,30458,30603,30732,30783,30905,30943,30961,31225,31407,31424,31433,31495,31547,31571,31612,31628,31642,31671,31679,31852,31913,31918,32166,32205,32220,32233,32279,32282,32497,32552,32719,32747,32947,32975,33027,33071,33184,33197,33245,33323,33385,33407,33462,33516,33586,33615,33648,33658,33662,33686,33775,33796,33850,33872,34025,34112,34122,34125,34165,34183,34223,34225,34255,34316,34399,34452,34581,34585,34607,34678,34716,34758,34789,34849,34998,35248,35322,35409,35741,35788,35981,36037,36048,36147,36187,36324,36365,36412,36581,36641,36692,36860,36907,37141,37143,37167,37246,37406,37481,37500,37517,37688,37747,37952,37977,37984,38063,38153,38168,38202,38268,38303,38375,38526,38570,38615,38655,38676,38683,38691,38731,38751,38808,38922,38985,38997,39085,39121,39129,39437,39465,39674,39682,39786,39840,39877,39942,39947,39992,40073,40157,40209,40518,40712,40714,40725,40753,40782,40792,40833,40909,41008,41216,41253,41271,41318,41496,41499,41598,41650,41692,41704,41816,41830,41956,41961,42301,42310,42327,42367,42376,42633,42696,42748,42820,42844,42847,42987,43154,43200,43290,43428,43466,43555,43642,43731,43749,43857,43880,43913,44020,44023,44122,44231,44279,44432,44506,44617,44679,44717,44755,44765,44775,44849,44880,44936,45077,45140,45191,45294,45313,45391,45441,45455,45506,45540,45599,45813,45876,45967,46045,46106,46168,46258,46272,46383,46444,46459,46492,46609,46718,46760,46855,46915,46920,46950,46955,47118,47127,47169,47183,47200,47279,47411,47836,47881,47901,47927,47953,47957,47986,48028,48061,48089,48099,48150,48173,48179,48237,48257,48322,48438,48452,48515,48876,48962,49114,49116,49227,49252,49409,49442,49521,49572,49642,50092,50098,50157,50172,50236,50355,50368,50422,50425,50487,50509,50664,50679,50737,50802,50859,50914,51074,51142,51194,51280,51318,51530,51555,51631,51633,51646,51697,51764,51796,52105,52136,52167,52265,52301,52367,52378,52381,52401,52527,52599,52630,52669,52678,52774,52804,52810,52837,52859,52890,52957,53007,53043,53155,53162,53279,53288,53338,53486,53555,53621,53734,53755,53895,54023,54088,54107,54189,54208,54231,54236,54255,54269,54357,54480,54485,54565,54610,54688,54779,54793,54824,54867,55021,55211,55260,55349,55354,55377,55492,55712,55764,55836,55985,56077,56265,56368,56410,56628,56657,56767,56783,56923,56972,57046,57056,57167,57219,57228,57351,57379,57399,57436,57461,57553,57645,57679,57864,57907,57914,58032,58042,58215,58361,58415,58422,58434,58637,58918,59033,59052,59106,59153,59235,59240,59400,59442,59568,59874,59951,59965,60259,60326,60447,60471,60480,60507,60576,60723,60730,60910,61047,61085,61172,61212,61637,61665,61670,61716,61729,61756,61854,61989,62042,62291,62298,62389,62390,62431,62441,62454,62523,62527,62644,62800,62832,62875,62912,62926,63045,63211,63339,63350,63509,63511,63530,63578,63585,63719,63855,63925,64156,64168,64193,64259,64274,64325,64332,64500,64501,64503,64560,64561,64711,64718,64774,64802,64888,64965,64976,64992,65051,65297,65363,65410,65495,65542,65606,65674,65708,65827,65843,66024,66060,66063,66115,66137,66142,66198,66202,66206,66294,66313,66363,66437,66460,66465,66469,66490,66496,66635,66892,66932,66974,67053,67129,67227,67339,67483,67493,67520,67582,67671,67803,68181,68274,68281,68313,68376,68541,68762,68884,69067,69105,69245,69312,69341,69476,69523,69567,69576,69624,69643,69724,69793,69880,69950,70036,70161,70214,70216,70292,70302,70314,70363,70493,70643,70775,70807,70837,70885,71090,71140,71190,71249,71269,71304,71365,71380,71404,71588,71662,71684,71730,71735,71777,71779,71834,71947,72050,72109,72164,72215,72293,72405,72594,72743,72870,72969,73048,73329,73358,73361,73374,73663,73898,73973,74009,74352,74486,74712,74739,74850,74922,75028,75113,75147,75163,75170,75213,75291,75303,75390,75691,75717,75748,75776,75844,76045,76298,76393,76422,76431,76464,76507,76535,76566,76568,76573,76648,76666,76683,76684,76763,76880,76905,77027,77049,77078,77173,77251,77274,77314,77437,77493,77506,77649,77738,77807,78077,78237,78252,78255,78263,78275,78327,78384,78388,78389,78408,78453,78460,78531,78594,78833,78886,78896,78930,79055,79141,79270,79289,79303,79397,79489,79493,79615,79680,79731,79846,79863,79902,79947,80017,80110,80120,80154,80252,80354,80473,80495,80606,80641,80717,80758,80797,80807,81120,81183,81202,81246,81366,81385,81404,81457,81463,81522,81553,81673,81788,81836,81841,81842,81898,82057,82192,82309,82313,82386,82426,82516,82558,82613,82640,82661,82819,82904,83067,83097,83246,83275,83318,83358,83364,83412,83424,83538,83540,83627,83640,83770,83845,84022,84109,84139,84170,84178,84480,84540,84542,84559,84584,84588,84693,84738,84808,84875,84879,84891,84970,85074,85151,85156,85256,85297,85363,85456,85557,85743,85973,86228,86331,86349,86351,86354,86357,86397,86465,86515,86520,86531,86546,86625,86779,86794,86811,86855,86963,87079,87083,87255,87265,87315,87333,87385,87404,87604,87632,87737,87805,88006,88117,88134,88159,88177,88184,88268,88297,88426,88546,88620,88701,89330,89359,89383,89392,89397,89429,89478,89491,89845,90265,90287,90382,90629,90713,90848,90978,91213,91255,91383,91387,91407,91481,91527,91596,91611,91675,91776,91790,91793,91894,91984,91986,92006,92086,92117,92195,92199,92287,92406,92407,92414,92469,92702,92866,92982,93031,93286,93405,93445,93620,93636,93734,93736,93740,93822,93846,93875,93912,93967,94088,94136,94151,94300,94310,94328,94436,94664,94721,94729,94812,94866,94896,94959,94982,95047,95104,95171,95219,95337,95447,95745,95926,96048,96063,96070,96110,96196,96292,96324,96485,96592,96622,96635,96866,97026,97044,97155,97202,97207,97239,97288,97306,97444,97462,97484,97489,97492,97535,97634,97635,97662,97850,97885,97934,98032,98068,98218,98235,98272,98359,98434,98443,98476,98639,98837,98885,98923,98963,99095,99169,99254,99269,99277,99306,99435,99596,99654,99657,99666,99713,99729,99834,99898",
          "54,66,108,232,246,359,512,687,746,815,834,968,988,1010,1030,1111,1141,1186,1255,1346,1370,1472,1508,1545,1569,1626,1693,1696,1755,1759,1851,1857,1974,1988,2131,2297,2325,2332,2333,2362,2392,2505,2525,2547,2578,2740,2766,2855,2963,3125,3177,3273,3382,3514,3715,3818,3836,3907,4096,4127,4166,4168,4194,4271,4495,4527,4533,4535,4537,4655,4669,4724,4764,4776,4828,4855,5145,5191,5262,5346,5376,5400,5418,5458,5618,5631,5645,5701,5822,5887,5937,6018,6088,6174,6221,6238,6286,6360,6363,6399,6482,6554,6699,6722,6754,6774,6904,7024,7055,7113,7330,7705,7707,7777,7785,7834,7841,8015,8075,8153,8159,8194,8272,8412,8598,8627,8706,8712,8827,8932,8989,8990,9068,9102,9154,9170,9248,9306,9336,9508,9639,9696,9758,9824,9863,9888,9939,9949,10039,10109,10216,10226,10270,10356,10362,10429,10600,10762,10934,11009,11054,11131,11282,11389,11408,11440,11442,11487,11518,11588,11592,11637,11639,11784,11872,11894,11916,12005,12009,12080,12213,12215,12220,12346,12358,12429,12481,12506,12518,12554,12630,12806,12817,12824,12883,12901,13113,13204,13207,13493,13513,13532,13542,13805,13840,13886,13927,13994,13995,14277,14324,14330,14358,14404,14419,14605,14609,14621,14801,14855,14870,15076,15153,15242,15500,15638,15639,15650,15738,16153,16164,16295,16326,16350,16362,16370,16378,16381,16541,16577,16707,16855,16876,17202,17204,17267,17398,17407,17471,17569,17621,17751,17834,17842,17923,17970,17984,18190,18191,18273,18328,18331,18336,18439,18458,18586,18629,18633,18654,19032,19072,19142,19184,19185,19358,19405,19485,19708,19720,19752,19789,20070,20089,20247,20324,20353,20549,20550,20557,20628,20744,20889,20892,20967,21023,21043,21074,21189,21247,21312,21427,21519,21520,21647,21705,21764,21766,21907,21994,22009,22045,22244,22249,22377,22563,22621,22723,22756,22995,23137,23232,23252,23313,23358,23539,23837,23896,23907,23973,24086,24223,24417,24512,24569,24704,24715,24745,24772,24773,24777,25389,25418,25461,25590,25598,25615,25792,25807,25894,25906,25938,26100,26201,26298,26343,26348,26383,26392,26504,26578,26615,26617,26732,26825,26883,26966,27006,27030,27039,27122,27136,27243,27266,27325,27337,27369,27434,27489,27500,27540,27670,27694,27862,27914,27979,28188,28287,28298,28336,28352,28364,28416,28505,28574,28610,28651,28661,28738,28816,28836,28867,28889,28890,28914,28917,28976,28977,29007,29074,29104,29128,29147,29183,29235,29243,29316,29386,29510,29558,29645,29651,29678,29781,29878,30018,30077,30108,30111,30173,30217,30296,30335,30358,30361,30521,30591,30626,30690,30698,30749,30760,30794,30853,30968,31093,31261,31344,31558,31559,31566,31568,31656,31668,31885,31917,31959,32156,32216,32306,32435,32465,32477,32500,32709,32833,32940,32991,33100,33111,33230,33248,33318,33357,33395,33464,33470,33892,34101,34146,34301,34307,34387,34482,34487,34498,34834,34926,34963,34992,35022,35205,35249,35364,35383,35532,35556,35568,35643,35706,35739,35740,35752,35811,35827,35875,35898,35966,36015,36069,36184,36296,36468,36661,36720,36724,36743,36777,36806,36890,36923,36924,37010,37016,37036,37053,37444,37454,37488,37587,37722,37786,37973,38029,38069,38075,38119,38145,38183,38243,38421,38437,38518,38839,38869,38917,38957,38971,38979,38993,39000,39073,39106,39166,39273,39320,39420,39605,39684,39736,39824,39845,40068,40150,40314,40557,40675,40676,40722,40759,40885,40891,40960,40975,40997,41003,41039,41175,41268,41294,41350,41365,41429,41454,41463,41545,41576,41627,41737,41767,41938,41977,41991,42005,42185,42257,42258,42348,42400,42513,42517,42642,42688,42701,42783,43045,43054,43073,43123,43158,43173,43253,43296,43352,43396,43443,43458,43536,43571,43580,43585,43602,43636,43788,43794,43812,43862,43905,43907,44039,44113,44148,44243,44390,44482,44564,44592,44743,44785,44832,44851,44870,44979,44980,45194,45201,45212,45457,45579,45614,45638,45725,45734,45827,45940,45972,46009,46027,46053,46070,46084,46221,46253,46324,46568,46570,46650,46657,46709,46788,46801,46869,46923,47027,47109,47133,47138,47753,47779,47885,47896,47900,47931,48021,48100,48119,48301,48312,48355,48464,48530,48597,48646,48730,48775,49055,49291,49324,49374,49454,49462,49465,49527,49534,49583,49624,49690,49717,49749,49754,49760,49792,49803,49818,49888,49912,50041,50052,50072,50095,50224,50242,50279,50481,50529,50627,50685,50791,51026,51054,51096,51173,51376,51412,51494,51507,51589,51604,51632,51700,51750,51871,52179,52319,52360,52365,52391,52426,52454,52514,52530,52568,52571,52575,52629,52632,52644,52664,52781,52843,52871,52946,52964,53069,53100,53163,53210,53302,53735,53951,53960,53978,53992,54009,54054,54123,54186,54249,54260,54379,54579,54706,54776,54796,55147,55290,55351,55519,55572,55611,55660,55757,55827,55857,55862,55913,55969,55999,56042,56313,56367,56573,56612,56615,56736,56859,56896,56911,56950,57036,57133,57183,57222,57278,57630,57635,57699,57741,57834,57903,57984,58019,58114,58137,58323,58348,58387,58476,58599,58619,58687,58787,58850,58957,58974,59002,59023,59347,59374,59412,59484,59552,59562,59614,59769,59815,59901,60027,60156,60182,60313,60564,60593,60764,60789,60838,60853,60908,60913,60964,61014,61038,61053,61069,61093,61228,61299,61366,61607,61695,61710,61759,61942,62082,62132,62153,62158,62224,62283,62325,62456,62482,62722,62746,62791,62844,62940,62967,63041,63421,63488,63506,63570,63888,63920,64048,64150,64210,64249,64369,64427,64510,64535,64562,64693,64712,64829,64851,64922,64969,65027,65060,65074,65288,65320,65482,65572,65720,65826,65897,65940,66011,66090,66110,66306,66340,66351,66439,66582,66602,66670,66697,66752,66895,67025,67150,67248,67333,67348,67439,67449,67527,67561,67619,67676,67712,67766,67774,68005,68098,68112,68113,68156,68189,68203,68283,68448,68499,68511,68528,68679,68723,68725,68764,68828,68923,68937,69016,69120,69127,69193,69239,69270,69293,69315,69365,69379,69433,69451,69511,69586,69644,69742,69755,69926,69931,69932,70012,70082,70130,70212,70277,70324,70437,70618,70703,70706,70745,70746,70898,70968,71103,71119,71250,71270,71540,71560,71915,71939,71963,72000,72146,72252,72330,72359,72370,72429,72538,72602,72656,72807,72837,72966,73066,73068,73129,73205,73213,73253,73271,73407,73408,73430,73629,73667,73744,73768,73787,73789,73875,73917,74076,74134,74146,74505,74573,74632,74644,74745,74780,74849,74955,75188,75233,75328,75385,75393,75457,75656,75669,75808,76101,76193,76246,76333,76406,76420,76505,76532,76660,76844,76883,76886,76913,77151,77203,77246,77406,77451,77863,78057,78087,78092,78149,78304,78332,78376,78450,78573,78637,78639,78759,78781,78782,78836,79008,79010,79074,79146,79319,79395,79492,79597,79624,79683,79704,79710,79727,79889,79981,80002,80026,80209,80215,80221,80417,80519,80574,80657,80739,80902,80905,80949,80960,80996,81128,81132,81191,81282,81419,81453,81491,81696,81716,81755,81907,81949,81979,82220,82284,82622,82674,82811,82858,83028,83171,83346,83406,83439,83616,83677,83701,83728,83787,83963,84015,84095,84101,84106,84144,84231,84244,84263,84273,84347,84410,84533,84554,84648,84708,84729,84882,85038,85087,85088,85188,85224,85245,85564,85741,85815,85817,85827,85865,85907,85957,86103,86264,86280,86308,86493,86551,86613,86650,86677,86924,87099,87172,87200,87347,87392,87522,87634,87646,87647,87848,87851,87874,87943,88058,88199,88296,88361,88460,88581,88597,88640,88800,88910,88922,88966,89095,89154,89226,89327,89371,89389,89405,89506,89790,89876,90219,90620,90855,90952,90954,91036,91084,91096,91147,91301,91385,91386,91603,91643,91651,91783,91871,91903,91992,92012,92014,92074,92169,92268,92324,92355,92358,92491,92523,92674,92692,92695,92728,92762,92765,92795,92876,92937,92999,93008,93082,93111,93134,93136,93214,93257,93315,93458,93471,93514,93613,93692,93728,93763,93930,93934,94017,94109,94217,94338,94386,94452,94556,94610,94687,94815,94850,94916,94943,95040,95076,95102,95131,95284,95319,95425,95578,95681,95760,95836,95882,95958,96003,96051,96120,96246,96259,96309,96372,96392,96523,96559,96624,96632,96682,96703,96796,96822,96837,96925,96927,97081,97118,97132,97143,97171,97177,97298,97307,97337,97419,97422,97455,97497,97522,97553,97581,97603,97620,97739,97743,97746,97842,97870,97939,97992,98036,98046,98069,98074,98108,98116,98196,98295,98349,98350,98433,98635,98748,98784,98832,98851,98878,98888,98925,99160,99302,99319,99354,99404,99406,99439,99443,99534,99736,99742,99745,99763,99862,99916,99975",
          "7,187,247,508,535,679,681,741,758,832,910,942,979,984,1289,1314,1411,1413,1437,1519,1614,1641,1782,1846,1849,1897,1903,1963,2030,2033,2259,2274,2308,2329,2384,2522,2561,2564,2606,2736,2789,2800,2802,3035,3060,3093,3145,3152,3185,3437,3459,3512,3579,3605,3628,3651,3769,3814,3833,3952,4103,4154,4172,4184,4269,4405,4413,4419,4614,4682,4743,4757,4924,5102,5135,5142,5433,5545,5574,5583,5637,5722,5807,5833,5905,5987,5991,6003,6131,6173,6379,6512,6603,6635,6672,6691,6921,7049,7167,7214,7235,7259,7443,7459,7492,7502,7547,7674,7722,7732,7744,7757,7796,7889,7893,7921,7932,7951,8112,8123,8142,8307,8587,8628,8734,8839,9108,9187,9247,9293,9318,9412,9526,9560,9665,9668,9781,9851,10019,10483,10668,10885,10890,10967,11204,11235,11325,11332,11392,11454,11701,11885,11933,12049,12085,12237,12336,12362,12368,12391,12440,12608,12676,12725,12766,12868,12894,12983,13149,13215,13235,13305,13385,13409,13455,13476,13531,13534,13563,13727,13809,13936,13951,13954,14073,14409,14496,14637,14679,14848,14876,14944,14951,14980,15027,15049,15057,15123,15150,15299,15372,15513,15563,15621,15625,15651,15724,15815,15866,15911,16031,16418,16426,16459,16518,16585,16668,16685,16734,16887,16951,17006,17059,17169,17222,17402,17404,17449,17652,17675,17734,17766,18078,18119,18180,18312,18329,18379,18397,18432,18529,18544,18738,18827,18936,18970,18996,19006,19201,19300,19332,19350,19959,19960,20005,20065,20082,20210,20331,20416,20446,20665,20723,20755,20771,21032,21137,21179,21192,21235,21440,21484,21525,21643,21678,21831,21833,21861,21962,22075,22146,22190,22252,22317,22521,22624,22706,22730,22787,22862,22933,22966,23086,23162,23262,23408,23459,23482,23676,23677,23844,23893,24045,24126,24152,24179,24411,24424,24591,24706,24812,24846,24988,25105,25153,25347,25569,25603,25672,25735,25883,25885,25916,26231,26263,26276,26300,26338,26369,26480,26502,26513,26644,26837,26899,26989,27027,27241,27278,27367,27392,27593,27664,27759,27836,27910,27938,27956,28044,28123,28418,28494,28517,28527,28613,28815,28881,28952,28994,29085,29093,29115,29165,29297,29384,29714,29721,29786,29886,29986,30040,30065,30242,30268,30276,30389,30405,30406,30582,30586,30719,30781,30846,30926,30944,30972,30990,31145,31150,31203,31212,31262,31281,31285,31315,31377,31384,31553,31562,31577,31626,31719,31722,31747,31800,31833,31911,31947,31989,32011,32023,32118,32169,32181,32201,32264,32369,32671,32672,32910,32983,33014,33031,33047,33053,33089,33240,33242,33507,33549,33556,33642,33678,33753,33827,33837,33894,33925,34038,34040,34118,34132,34236,34530,34603,34735,34920,34951,35002,35012,35104,35262,35340,35356,35374,35384,35388,35400,35451,35615,35641,35644,35772,35813,35816,35844,35931,36011,36132,36165,36255,36293,36660,36663,36773,36793,36934,36972,37313,37366,37378,37470,37499,37573,37656,37693,37738,37837,37905,37980,38026,38074,38194,38223,38342,38386,38653,38664,38675,38838,38950,38970,39048,39088,39192,39230,39246,39321,39441,39486,39576,39580,39592,39673,39698,39801,39807,39820,39985,40136,40186,40191,40228,40229,40238,40255,40400,40431,40905,40964,40988,41140,41194,41202,41436,41486,41680,41691,41975,42113,42418,42429,42561,42589,42754,42813,42856,42934,42989,43069,43193,43256,43338,43535,43566,43750,43842,44078,44098,44099,44235,44532,44549,44699,44744,44747,44766,44837,44859,44902,44967,45010,45053,45117,45134,45161,45224,45237,45381,45404,45612,45648,45652,45667,45875,45938,45968,46113,46137,46372,46422,46430,46677,46818,46874,46969,47022,47023,47026,47073,47121,47142,47234,47270,47313,47347,47442,47455,47640,47660,47758,47791,47816,47856,47872,47904,47990,48297,48338,48475,48494,48510,48525,48580,48651,48654,48667,48718,48768,48850,48880,48883,48931,49018,49020,49041,49138,49245,49276,49334,49397,49436,49722,49736,49766,49807,49816,49838,49848,49929,49941,50006,50028,50061,50159,50357,50378,50445,50449,50457,50533,50585,50599,50813,50853,50900,51034,51048,51079,51233,51418,51424,51535,51786,51872,52073,52171,52200,52295,52574,52623,52671,52722,52749,52821,53088,53090,53096,53149,53554,53599,53785,53795,53802,53830,53964,53972,54004,54028,54078,54199,54314,54429,54472,54577,54890,54892,55066,55232,55263,55414,55447,55670,55701,55858,55864,55882,56065,56150,56198,56215,56261,56287,56310,56315,56371,56394,56444,56474,56475,56543,56674,56775,56817,56881,56966,57091,57139,57149,57169,57190,57192,57211,57277,57309,57329,57537,57859,57992,57994,58254,58400,58423,58457,58657,58720,58742,58808,58896,59059,59270,59286,59325,59348,59461,59553,59599,59669,59794,59897,60048,60403,60440,60595,60612,60641,60656,60672,60733,60835,61136,61207,61226,61395,61510,61640,61738,61776,61893,61896,61898,62255,62406,62426,62641,62792,62834,62993,63056,63091,63127,63224,63306,63449,63599,63924,64082,64109,64112,64135,64152,64375,64498,64511,64547,64610,64629,64816,64906,64962,65046,65059,65201,65413,65435,65638,65741,65764,65937,66037,66042,66127,66357,66365,66434,66463,66596,66677,66699,66827,66896,67222,67239,67423,67607,67746,67889,67934,68063,68118,68151,68273,68442,68503,68564,68605,68697,69084,69205,69207,69210,69279,69345,69382,69457,69466,69477,69524,69741,69743,69881,69997,70065,70140,70175,70366,70397,70405,70417,70556,70565,70573,70695,70704,70786,70886,70950,70975,71185,71191,71373,71523,71643,71749,71750,71942,72076,72180,72327,72354,72366,72537,72546,72616,73051,73140,73316,73645,73793,73806,73833,73985,74203,74338,74377,74418,74440,74519,74544,74556,74606,74607,74807,74840,74851,75060,75101,75226,75298,75312,75355,75376,75398,75664,75694,75838,75904,75947,75962,76037,76130,76515,76530,76575,76582,76613,76813,76876,76961,77236,77284,77373,77475,77602,77768,77787,77893,78120,78288,78356,78422,78435,78462,78467,78538,78545,78669,78816,78843,78877,79025,79030,79041,79161,79181,79200,79467,79484,79707,79726,79734,79776,79779,79958,80121,80222,80365,80368,80485,80528,80544,80562,80613,80633,80742,80878,80900,80951,80966,80987,81261,81353,81435,81540,81616,81706,81723,82435,82497,82510,82511,82567,82671,82865,82899,82908,83008,83032,83041,83272,83343,83356,83582,83586,83763,83993,84061,84062,84190,84208,84346,84594,84786,84827,84869,84967,85091,85332,85439,85632,85689,85698,85809,85845,85869,85948,86020,86059,86122,86288,86414,86475,86644,86830,86833,86929,86933,86941,87006,87285,87447,87611,87621,87633,87703,87724,87844,88090,88223,88244,88259,88293,88341,88523,88598,88658,88669,88683,88689,88926,89071,89072,89189,89279,89281,89283,89286,89294,89306,89410,89426,89427,89451,89697,89731,89761,89766,89895,89925,89954,90043,90237,90263,90295,90309,90315,90324,90400,90431,90689,90742,91212,91218,91240,91282,91312,91365,91484,91638,91733,91751,91781,91806,91875,91904,92058,92097,92330,92356,92703,92722,92771,92788,92923,92995,93067,93119,93482,93773,93860,93976,94250,94346,94407,94608,94651,94702,94875,94879,94909,94928,95128,95161,95173,95303,95489,95573,96004,96078,96129,96182,96387,96662,96745,96820,96895,97000,97222,97309,97424,97456,97507,97587,97590,97594,97712,97808,97819,97825,97943,97954,98020,98073,98085,98121,98154,98200,98297,98418,98496,98525,98845,99029,99058,99171,99194,99219,99223,99238,99259,99521,99524,99566,99679,99895,99930,99978",
          "26,184,208,511,612,633,661,713,759,761,800,906,1037,1093,1144,1196,1197,1235,1292,1294,1303,1321,1424,1513,1520,1625,1648,1680,1683,1802,1864,2145,2182,2185,2377,2439,2490,2816,2868,2955,3272,3278,3291,3342,3424,3502,3789,3903,3927,3938,4105,4157,4213,4220,4261,4309,4412,4583,4762,4771,4831,4862,4886,4895,4997,5080,5143,5164,5297,5506,5542,5723,5955,6028,6036,6062,6140,6144,6353,6429,6447,6507,6531,6539,6609,6706,6785,6887,6894,7001,7133,7338,7419,7676,7890,7960,7998,8069,8173,8190,8423,8480,8507,8509,8634,8776,8790,8877,9091,9142,9229,9323,9369,9398,9437,9446,9450,9472,9487,9499,9505,9510,9514,9524,9643,9671,9823,10023,10095,10229,10377,10384,10453,10614,10709,10722,10725,10789,10816,10873,10925,10944,11011,11021,11189,11435,11555,11581,11607,11672,11812,11857,11887,11923,11943,12077,12101,12353,12413,12421,12505,12540,12706,12723,12765,12911,12932,13186,13315,13391,13427,13636,13644,13863,14319,14331,14445,14473,14494,14546,14552,14777,14778,14835,15044,15102,15267,15471,15506,15619,15801,15814,15918,15938,15967,15970,15998,16130,16149,16234,16261,16348,16556,16736,16757,16846,16985,17080,17140,17141,17244,17299,17428,17441,17571,17810,17817,17900,17949,18049,18145,18203,18229,18321,18371,18452,18464,18602,18622,18801,18806,18810,18892,18969,19093,19297,19314,19345,19651,19776,19803,19982,20025,20333,20398,20436,20444,20491,20531,20668,20851,20910,21075,21228,21232,21280,21287,21308,21363,21409,21672,21698,21737,21774,21822,21843,21852,21900,22118,22159,22198,22341,22520,22603,22762,22792,22863,22871,22912,22932,22943,22980,22985,23015,23063,23096,23145,23209,23210,23222,23307,23505,23525,23542,23596,23693,23718,23732,23765,23769,23800,23826,23873,23914,23932,23960,24092,24101,24148,24177,24251,24466,24694,24785,24823,24831,25111,25149,25184,25385,25471,25534,25554,25783,25805,25856,25866,25981,25987,26012,26014,26098,26141,26206,26247,26308,26568,26632,26722,26847,26906,26956,27015,27051,27127,27182,27183,27220,27382,27387,27457,27460,27486,27519,27611,27728,27931,27934,27937,27993,28028,28049,28058,28078,28179,28239,28324,28346,28399,28406,28450,28514,28601,28622,28668,28793,28828,28965,28996,29137,29180,29211,29325,29457,29562,29602,29709,29917,29963,30156,30252,30438,30475,30509,30516,30534,30677,30733,30757,30818,30856,30865,31182,31209,31233,31317,31421,31545,31826,31832,31889,32186,32395,32467,32523,32897,32927,32985,32998,33223,33412,33510,33550,33565,33744,33773,33776,33910,34110,34249,34256,34317,34376,34401,34479,34517,34537,34752,34824,34871,34892,34947,35175,35209,35281,35424,35548,35625,35732,35775,35796,36049,36199,36243,36303,36312,36322,36349,36407,36409,36453,36594,36652,36684,36731,36744,36761,36776,36939,36996,37009,37012,37145,37189,37345,37385,37411,37585,37668,37930,37931,37942,38006,38158,38339,38341,38361,38411,38530,38538,38553,38642,39069,39092,39203,39426,39459,39466,39582,39622,39841,39892,39899,40015,40061,40072,40121,40124,40213,40233,40447,40532,40635,40754,40840,40929,41057,41176,41264,41299,41324,41523,41594,41653,41741,41794,41964,42036,42048,42137,42350,42405,42590,42591,42842,42881,42979,43053,43161,43184,43242,43414,43463,43515,43543,43596,43653,43761,43806,44126,44265,44283,44427,44455,44545,44716,44745,44763,44816,45208,45331,45346,45406,45559,45753,45937,45973,46026,46090,46147,46162,46236,46266,46421,46435,46714,46727,46770,46823,46938,46949,47062,47111,47182,47275,47335,47368,47459,47500,47502,47532,47533,47568,47775,48015,48067,48077,48468,48478,48536,48546,48610,48688,48692,48708,48840,48847,48855,48875,48915,48917,48991,49000,49165,49188,49426,49501,49502,49551,49657,49764,49798,50069,50201,50299,50351,50386,50432,50527,50557,50576,50588,50712,50844,50856,50961,51015,51047,51152,51265,51278,51330,51433,51462,51596,51638,51802,51810,51822,51987,52042,52058,52133,52187,52500,52512,52591,52655,52703,52841,52915,53008,53014,53053,53059,53231,53448,53541,53551,53583,53736,53788,53814,53981,54030,54062,54074,54132,54423,54530,54549,54572,54633,54747,54818,54828,54837,54883,54901,54931,55002,55004,55048,55117,55208,55229,55413,55458,55610,55704,55740,55787,55788,55842,55978,56003,56071,56091,56384,56419,56498,56545,56738,56843,56860,56977,57275,57373,57456,57586,57726,57770,57774,57785,57800,57881,57883,57993,58090,58105,58795,58906,59044,59057,59284,59295,59298,59489,59569,59655,59706,59925,60038,60102,60105,60373,60396,60427,60579,60688,60754,60865,60915,60963,60971,60972,61040,61166,61188,61411,61537,61568,61594,61641,61697,61963,62041,62206,62229,62474,62583,62728,62879,62988,63016,63044,63110,63223,63251,63358,63403,63482,63555,63625,63707,63708,63985,64482,64530,64595,64669,64796,64879,64883,64981,65042,65101,65147,65280,65294,65359,65391,65496,65502,65525,65571,65593,65732,65882,65964,66087,66117,66149,66151,66329,66398,66480,66515,66716,67048,67166,67205,67224,67253,67267,67334,67436,67728,67756,67866,67886,68010,68027,68055,68108,68114,68220,68356,68360,68480,68483,68707,68821,68895,68942,69117,69416,69430,69462,69467,69582,69622,69667,69704,69706,69982,70025,70035,70066,70227,70386,70488,70503,70527,70712,70741,70857,71401,71464,71493,71494,71511,71700,71711,71773,71827,71845,71887,71945,72152,72257,72283,72669,72678,72698,72823,72848,72964,73088,73124,73633,73771,73798,73820,73868,73897,73941,74012,74110,74112,74118,74197,74260,74690,74707,74796,74814,75065,75095,75228,75231,75498,75526,75555,75568,75670,76355,76392,76650,76676,76823,76869,76885,77095,77404,77433,77440,77601,77608,77719,77727,77732,77855,77978,78005,78147,78206,78212,78322,78328,78378,78515,78686,78714,78841,78874,78904,78992,79063,79092,79132,79140,79447,79480,79611,79721,79793,79799,79843,79931,80112,80283,80332,80416,80638,80873,80938,80994,81082,81148,81211,81255,81741,81973,81982,82036,82137,82171,82369,82423,82571,82590,82738,82962,83001,83063,83085,83186,83233,83307,83347,83372,83498,83553,83566,83742,83775,83854,83889,84180,84269,84408,84551,84562,84582,84624,84719,84892,84895,84901,84923,85008,85064,85161,85193,85209,85212,85253,85280,85412,85464,85478,85531,85555,85620,85680,85842,85871,85924,86033,86034,86165,86220,86317,86382,86385,86501,86527,86536,86539,86580,86611,86655,86771,86816,86920,86989,87185,87198,87349,87359,87509,87614,87749,87758,88035,88133,88157,88373,88385,88464,88509,88521,88635,88671,88791,88812,89025,89311,89574,89637,89730,89874,89907,90126,90197,90462,90540,90774,90854,90881,91006,91063,91156,91224,91381,91483,91531,91561,91665,91913,91934,91943,91973,91998,92078,92202,92240,92305,92371,92435,92547,92587,92724,92826,92905,92908,92919,93175,93393,93403,93461,93517,93579,93679,93704,93717,93746,93791,93936,93938,93969,94007,94072,94078,94168,94212,94240,94288,94417,94524,94604,94778,94807,94808,94942,94993,95376,95660,95773,95784,95791,95833,95837,95875,96047,96225,96230,96245,96350,96542,96585,96591,96853,96889,96891,96901,96970,97056,97059,97090,97124,97266,97287,97429,97523,97748,97807,97828,97924,97988,98037,98038,98095,98383,98388,98408,98602,98621,98750,98761,98769,98811,98894,98945,99043,99055,99116,99187,99342,99400,99570,99579,99700,99977,99981,99991",
          "80,121,192,303,349,351,617,623,627,731,870,880,935,950,969,971,981,1018,1085,1278,1287,1323,1388,1409,1458,1617,1624,1645,1697,1700,1745,1786,1814,1838,1868,2039,2174,2258,2526,2801,2822,2880,2962,3136,3197,3219,3227,3260,3324,3411,3673,3829,3860,3867,3918,4034,4045,4064,4115,4205,4340,4343,4351,4484,4493,4568,4615,4666,4812,4906,4983,5020,5052,5070,5141,5151,5258,5299,5311,5409,5714,5913,5962,5989,6078,6103,6241,6243,6513,6614,6617,6651,6791,6806,6834,7030,7046,7180,7353,7356,7452,7548,7667,7677,7708,7729,7829,7863,7891,7931,8260,8275,8428,8497,8511,8524,8553,8612,8736,8813,8819,8846,8871,8924,9009,9133,9330,9608,9667,9670,9785,9837,9878,10046,10073,10116,10153,10262,10273,10277,10349,10370,10404,10502,10617,10733,10734,11111,11122,11192,11368,11374,11652,11683,11767,11897,11910,12058,12148,12158,12229,12273,12323,12369,12711,12780,12956,13001,13056,13067,13154,13285,13488,13622,13665,13761,13822,13839,13898,13909,14003,14128,14172,14394,14406,14416,14441,14459,14491,14520,14591,14697,14796,14818,14863,14896,14982,14990,15025,15081,15138,15290,15295,15428,15468,15488,15695,15715,15743,15828,15887,15912,15916,16146,16163,16245,16276,16290,16343,16364,16422,16483,16517,16641,16785,16797,16800,16824,16873,16980,17012,17099,17181,17203,17379,17386,17403,17422,17472,17496,17524,17544,17554,17583,17669,17737,17942,17987,18034,18344,18374,18436,18450,18533,18612,18649,18772,18775,18915,18949,18959,18981,19166,19258,19315,19348,19359,19551,19634,19645,19865,19932,19992,20030,20052,20076,20318,20554,20568,20638,20657,20662,20939,20943,21034,21078,21139,21151,21343,21446,21453,21458,21569,21694,21796,21871,21908,21913,21986,21999,22073,22546,22562,22595,22606,22672,22715,22754,22822,22832,23214,23253,23297,23298,23462,23473,23552,23583,23595,23686,23690,23758,23764,23872,23956,23968,24041,24053,24060,24188,24209,24281,24294,24313,24398,24421,24422,24437,24464,24553,24618,24672,24725,24733,25259,25382,25555,25685,25890,25909,25933,25965,25991,26011,26059,26115,26256,26304,26319,26335,26402,26489,26496,26517,26572,26669,26710,26727,26815,26848,26913,26972,26973,27077,27082,27101,27165,27394,27436,27493,27640,27654,27663,27746,27779,27877,27919,27939,27941,27978,28039,28216,28248,28326,28348,28382,28401,28427,28453,28479,28484,28503,28662,28723,28725,28729,28773,28811,28830,29038,29205,29249,29354,29532,29533,29538,29545,29622,29664,29972,29996,30050,30153,30186,30200,30205,30207,30258,30313,30456,30494,30542,30604,30619,30724,30817,30995,31029,31089,31158,31162,31217,31263,31314,31414,31468,31477,31478,31623,31753,31759,31851,31854,31995,32002,32111,32167,32209,32330,32347,32505,32540,32555,32636,32705,32708,32751,32871,32934,32988,33180,33189,33222,33309,33429,33475,33666,33879,33891,33930,34076,34083,34143,34177,34281,34534,34547,34894,35086,35087,35097,35128,35135,35207,35260,35349,35613,35622,35712,35734,35757,35883,35959,36014,36029,36052,36124,36233,36846,36862,37046,37150,37190,37249,37320,37425,37480,37555,37582,37588,37699,37789,37924,37943,38005,38201,38256,38622,38840,39013,39036,39109,39163,39174,39333,39357,39404,39428,39498,39512,39563,39823,39832,39934,39991,40029,40046,40100,40118,40125,40162,40257,40360,40361,40428,40430,40446,40573,40586,40695,40716,40817,40854,40896,40906,40955,40980,41005,41018,41114,41355,41374,41414,41508,41577,41710,41822,41897,41930,41944,41994,42033,42101,42298,42312,42386,42409,42445,42476,42511,42553,42598,42666,42746,42791,42975,43026,43206,43363,43370,43397,43627,43688,43724,43827,44022,44060,44158,44408,44411,44448,44500,44589,44627,44645,44675,44726,44733,44767,44795,44798,44857,44860,44932,45401,45431,45488,45505,45702,45779,45807,45818,46020,46038,46067,46095,46109,46140,46161,46165,46485,46640,46722,46908,46912,47055,47179,47193,47233,47282,47444,47701,47721,47887,47912,47919,47996,48259,48275,48287,48345,48416,48447,48514,48519,48521,48616,48723,48790,48826,48829,48844,48865,48963,49270,49282,49356,49366,49430,49470,49656,49869,49884,49963,49973,49976,50024,50176,50198,50232,50284,50292,50297,50318,50382,50492,50554,50568,50600,50609,50647,50811,50839,50863,50949,50965,51052,51053,51081,51181,51189,51245,51253,51302,51396,51425,51449,51464,51467,51603,51717,51736,51977,51994,52207,52224,52226,52234,52631,52635,52688,52751,52855,52902,52928,52932,52996,53124,53126,53188,53296,53334,53345,53374,53474,53627,53719,53731,53778,53887,53894,53914,53929,54042,54046,54047,54096,54102,54172,54185,54204,54465,54678,54907,54999,55267,55461,55552,55554,55569,55629,55662,55709,55828,55884,55959,55989,56006,56029,56058,56076,56144,56208,56210,56243,56397,56443,56540,56665,56741,56778,57031,57176,57357,57366,57464,57656,57658,57690,57705,57899,57916,57937,57939,57979,58083,58159,58162,58185,58606,58755,59120,59198,59226,59360,59510,59535,59537,59621,59689,59803,59850,59876,59911,59934,59954,60089,60122,60149,60214,60260,60360,60361,60474,60506,60580,60709,60731,60770,60800,60997,61010,61013,61140,61225,61279,61398,61501,61508,61518,61689,61699,61741,61808,61849,61860,61878,61921,61983,62093,62151,62211,62420,62538,62584,62614,62663,62704,62793,63203,63214,63275,63280,63345,63368,63400,63468,63505,63639,63910,63969,64018,64066,64107,64201,64208,64220,64430,64520,64566,64770,64834,64860,64994,65146,65466,65612,65725,65738,65771,65847,65880,65913,65953,65970,66119,66220,66260,66279,66338,66713,66718,66788,66799,66911,66934,66936,66994,67008,67015,67029,67104,67121,67151,67293,67381,67385,67386,67405,67455,67468,67487,67528,67531,67549,67725,67784,68123,68224,68262,68287,68325,68326,68339,68451,68485,68525,68648,68718,68790,68819,68820,68851,69013,69047,69051,69177,69215,69220,69271,69395,69452,69479,69497,69609,69712,69718,69776,69791,69813,69865,69876,70002,70024,70092,70099,70165,70171,70199,70211,70339,70389,70448,70723,70863,70900,70960,71016,71166,71201,71257,71273,71393,71451,71613,71638,71647,71672,71690,71806,71817,71820,71940,72007,72010,72055,72159,72202,72211,72263,72308,72376,72388,72444,72540,72644,72731,72736,72737,72759,72782,72858,72862,72905,72936,73046,73108,73115,73180,73482,73618,73622,73694,73867,73974,74057,74120,74344,74481,74645,74881,75038,75235,75295,75369,75413,75415,75474,75750,75780,75790,75957,76067,76099,76152,76173,76183,76358,76518,76547,76562,76605,76775,76984,76999,77082,77113,77159,77298,77322,77335,77376,77424,77456,77519,77538,77706,77721,77830,77902,77947,78032,78069,78118,78145,78330,78475,78478,78519,78628,78660,78681,78794,78797,78825,78838,78881,78922,79024,79027,79070,79226,79237,79277,79294,79376,79471,79487,79672,79676,79705,79758,79772,79828,79829,79872,79907,80143,80172,80180,80264,80730,80731,80789,80790,80945,81054,81123,81136,81273,81328,81346,81432,81481,81669,82016,82023,82035,82103,82147,82204,82208,82260,82402,82420,82432,82448,82466,82553,82602,82646,82722,82823,82827,82876,83117,83148,83295,83310,83388,83393,83483,83539,83562,83666,83696,83720,83789,83792,83851,83866,83899,83953,84005,84025,84068,84506,84606,84639,84797,84800,84920,84928,85295,85320,85424,85446,85450,85466,85500,85538,85682,85702,85786,85835,85900,85904,85985,85994,86049,86109,86155,86201,86316,86479,86553,86623,86718,86777,86828,86880,86900,86926,87101,87175,87188,87199,87277,87521,87550,87578,87603,87664,87831,87833,88275,88313,88381,88409,88418,88552,88583,88590,88642,88681,88781,88865,88887,88916,88990,89090,89099,89200,89227,89229,89258,89441,89490,89579,89614,89617,89705,89808,89880,89943,89977,89978,90082,90084,90132,90135,90151,90203,90292,90350,90366,90376,90418,90420,90466,90580,90587,90607,90625,90832,90912,90955,91045,91110,91183,91199,91220,91284,91588,91802,91869,91882,91949,91967,92020,92115,92123,92197,92480,92481,92498,92631,92693,92755,92807,92812,92838,92965,93014,93170,93171,93275,93311,93336,93365,93480,93487,93561,93664,93785,93807,93881,93953,93958,93981,93989,94079,94200,94476,94485,94574,94612,94665,94855,94873,94929,94940,94985,95061,95168,95312,95349,95470,95556,95597,95613,95793,95853,95856,95864,95909,96011,96061,96068,96082,96090,96136,96189,96271,96406,96584,96698,96708,96812,96892,96896,96913,97017,97067,97069,97079,97095,97127,97185,97326,97341,97390,97426,97440,97506,97512,97524,97599,97606,97619,97622,97642,97718,97785,97798,97929,97959,98270,98273,98277,98336,98432,98719,98819,98820,98860,98875,99013,99025,99099,99125,99191,99258,99401,99416,99530,99564,99753,99757,99785,99836,99886,99969",
          "10,188,241,248,256,380,433,569,621,768,801,878,888,903,912,940,948,960,991,1189,1209,1213,1221,1253,1282,1445,1461,1543,1571,1577,1579,1586,1609,1722,1780,1905,1940,1967,1997,2087,2120,2263,2267,2272,2323,2387,2493,2747,2790,2937,2957,3005,3023,3047,3062,3215,3258,3292,3413,3457,3701,3845,3991,4000,4085,4151,4152,4165,4232,4310,4478,4541,4842,4953,5004,5060,5254,5529,5586,5697,5719,5731,5741,5793,5918,6026,6185,6209,6311,6356,6444,6454,6464,6558,6578,6671,6744,6801,6868,6912,6931,6989,7101,7533,7537,7590,7611,7706,7752,7797,7850,8068,8359,8492,8529,8874,8895,8912,8919,8955,8974,9364,9570,9681,9778,9855,9974,10102,10113,10128,10245,10269,10520,10568,10610,10643,10924,10945,10973,10992,11495,11570,11636,11689,11829,11839,11927,12011,12038,12063,12129,12174,12431,12510,12777,12792,12819,13183,13198,13295,13341,13358,13362,13402,13465,13581,13594,13611,13859,13891,13972,14124,14199,14327,14357,14383,14466,14489,14610,14658,14719,14815,14849,14963,15084,15162,15239,15312,15315,15559,15566,15579,15735,15881,16014,16072,16217,16272,16313,16317,16319,16354,16559,16885,16964,16994,17015,17111,17253,17261,17300,17320,17491,17501,17715,17775,17820,17935,17959,17978,18012,18258,18289,18388,18389,18478,18754,18823,18929,18932,19029,19048,19073,19131,19179,19299,19334,19518,19526,19629,19712,19727,19744,19822,19840,19928,19938,20081,20097,20150,20154,20223,20448,20476,20509,20525,20878,21046,21218,21426,21460,21468,21493,21495,21497,21902,22121,22256,22425,22478,22517,22684,22793,22833,22835,22918,23117,23157,23294,23404,23457,23464,23626,23714,23751,23753,23754,23782,23793,23986,24046,24098,24122,24214,24270,24310,24326,24361,24377,24617,24675,24684,24695,24699,24728,24813,25113,25130,25275,25380,25487,25489,25602,25639,25670,25698,25705,25713,25777,25833,26246,26297,26331,26421,26919,27068,27072,27081,27094,27237,27303,27307,27314,27495,27617,27835,27887,27924,28150,28308,28343,28373,28609,28669,28753,28878,28902,28962,29050,29338,29409,29426,29497,29633,29681,29716,29809,29873,30015,30237,30248,30302,30355,30442,30445,30488,30525,30628,30671,30740,30779,30829,30847,30896,30898,30992,31031,31143,31208,31213,31739,31767,31898,31952,31983,32117,32127,32427,32442,32502,32688,32766,32889,32979,33114,33163,33306,33307,33352,33423,33503,33704,33741,33794,33807,33995,33997,34005,34237,34270,34271,34326,34346,34450,34594,34599,34681,34724,34751,34817,34845,34908,34948,35029,35136,35151,35210,35221,35554,35617,35708,35737,35842,35897,35952,35994,35995,35996,36262,36435,36473,36771,36785,36822,36896,36932,36986,37018,37076,37080,37163,37235,37242,37270,37326,37343,37348,37459,37472,37506,37520,37641,37711,37860,37865,37926,38030,38092,38217,38235,38264,38389,38490,38514,38563,38689,38796,38928,39021,39034,39066,39067,39082,39098,39127,39172,39292,39490,39558,39577,39627,39691,39728,39889,39893,39921,40119,40141,40308,40317,40340,40472,40549,40560,40563,40665,40751,40852,40882,41151,41172,41237,41354,41390,41713,41729,41744,41910,41912,41926,42112,42318,42381,42406,42425,42446,42477,42606,42726,42839,43017,43137,43219,43220,43245,43482,43551,43643,43693,43945,44102,44119,44305,44307,44343,44346,44393,44450,44463,44524,44582,44667,44721,44804,44892,44949,44960,45211,45271,45314,45378,45624,45628,45668,45729,45951,45959,45996,46265,46267,46387,46469,46482,46486,46573,46606,46619,46716,46734,46758,46873,46924,46954,46974,47010,47025,47060,47129,47175,47217,47241,47281,47796,48024,48036,48042,48063,48317,48412,48455,48467,48623,48685,48858,48978,48980,48996,49043,49066,49299,49306,49376,49463,49468,49557,49625,49663,49880,49904,49972,50065,50084,50088,50116,50118,50229,50411,50448,50482,50497,50534,50580,50598,50620,50731,50751,50795,50948,51037,51102,51125,51180,51262,51263,51307,51344,51374,51405,51578,51660,51720,51814,51823,51883,51939,51980,51988,52075,52153,52275,52312,52486,52633,52684,52811,52903,53030,53112,53186,53316,53379,53549,53709,53881,53997,54001,54066,54076,54138,54155,54211,54301,54373,54456,54490,54731,54750,54860,54977,55039,55068,55251,55310,55384,55445,55485,55515,55558,55678,55741,55744,55826,55871,55880,56039,56056,56078,56089,56222,56242,56296,56541,56589,56620,56707,56936,57061,57132,57202,57203,57242,57401,57443,57465,57471,57574,57601,57620,57729,57756,57791,58197,58252,58391,58537,58540,58562,58584,58730,58835,58959,59009,59129,59268,59340,59370,59818,59940,59975,60057,60129,60134,60137,60221,60225,60385,60406,60457,60473,60515,60658,60922,61003,61044,61257,61283,61332,61394,61423,61424,61463,61530,61547,61792,61912,62252,62466,62491,62550,62554,62629,62688,62814,62897,63051,63100,63342,63349,63399,63438,63524,63531,63582,63816,63849,63995,64104,64170,64371,64406,64439,64578,64624,64625,64628,64698,64852,64900,65243,65251,65401,65588,65656,65711,66111,66155,66186,66349,66411,66467,66504,66728,66842,66898,67069,67084,67332,67628,67899,67994,67997,68004,68219,68233,68296,68302,68357,68390,68489,68571,69002,69008,69099,69114,69121,69126,69130,69154,69187,69250,69414,69435,69470,69506,69520,69661,69840,69921,70112,70254,70267,70280,70295,70506,70772,70963,70974,71101,71153,71288,71376,71473,71627,71814,71874,71888,71984,72031,72034,72075,72242,72269,72404,72433,72581,72619,72809,72860,72951,73097,73105,73131,73158,73306,73330,73536,73729,73736,73773,73794,73882,73955,73993,74125,74263,74378,74470,74788,74800,74866,75145,75186,75208,75279,75567,75610,75714,75723,75736,75806,75858,75862,76002,76005,76035,76417,76595,76617,76640,76704,76727,76898,77001,77012,77035,77057,77219,77467,77728,77744,77788,77891,77898,77950,78058,78093,78426,78513,78527,78777,78779,78892,78932,78953,79119,79126,79158,79178,79192,79384,79385,79547,79636,79693,79702,80041,80067,80069,80082,80093,80394,80418,80444,80511,80538,80559,80620,80986,81029,81159,81188,81416,81518,81576,81606,81704,81816,81866,81959,82079,82133,82259,82385,82391,82496,82498,82595,82601,82649,82675,82710,82724,82728,82766,82794,82849,82852,82917,82935,82958,83053,83189,83293,83476,83486,83867,83915,83954,84018,84225,84248,84279,84467,84614,84700,84745,84848,84907,84955,85089,85097,85328,85472,85979,86018,86071,86083,86371,86437,86518,86631,86770,86790,86878,86910,87056,87081,87170,87219,87287,87300,87313,87326,87425,87485,87526,87753,87798,87813,87890,87998,88086,88097,88194,88206,88231,88255,88505,88536,88559,88584,88733,88823,88850,88903,88956,88997,89000,89216,89288,89323,89391,89500,89513,89555,89557,89561,89602,89608,89820,89900,89935,89969,89976,89992,90136,90222,90282,91054,91061,91075,91124,91238,91314,91417,91426,91461,91470,91521,91539,91706,91711,91926,91957,92007,92061,92165,92177,92186,92303,92306,92360,92364,92365,92636,92809,92994,93131,93180,93259,93494,93502,93527,93572,93833,93852,93862,94002,94051,94087,94160,94236,94295,94354,94356,94394,94453,94621,94741,94920,94926,95461,95493,95532,95560,95580,95874,95997,95999,96023,96211,96297,96399,96778,96784,96995,97209,97211,97296,97370,97435,97571,97588,97670,97688,97716,97769,97879,98420,98438,98456,98517,98541,98752,98869,98922,99204,99411,99440,99441,99519,99537,99543,99637,99705,99765,99840,99868,99918",
          "89,101,143,153,406,523,729,760,861,955,1033,1086,1104,1133,1211,1254,1466,1488,1670,1736,1737,1750,1875,1878,1957,1970,2051,2062,2114,2171,2197,2249,2299,2353,2465,2523,2678,2734,2892,2923,2968,2992,3109,3268,3270,3314,3315,3360,3679,3710,3718,3916,3959,4145,4346,4375,4435,4446,4553,4580,4685,4710,4725,4851,4965,4967,5046,5165,5206,5333,5339,5468,5564,5577,5617,5713,5882,5901,5904,6098,6239,6472,6515,6541,6573,6616,6748,7048,7179,7181,7230,7257,7310,7326,7329,7415,7599,7618,7649,7671,7694,7719,7746,7761,7832,7953,8021,8070,8333,8550,8764,8792,8845,8848,8882,9001,9081,9130,9214,9259,9296,9442,9517,9582,9594,9657,9713,9761,9861,9877,9966,9997,10063,10072,10094,10276,10295,10342,10447,10620,10681,10689,10741,10746,10807,11028,11059,11097,11109,11251,11326,11380,11482,11531,11619,11706,11753,11769,11780,11817,11850,11902,12036,12052,12083,12269,12303,12370,12428,12566,12834,12841,12877,13068,13283,13306,13339,13444,13551,13557,13561,13784,13791,13821,13950,13967,13999,14007,14127,14168,14249,14270,14322,14593,14682,14692,14771,14823,14952,14968,15175,15198,15314,15358,15562,15722,15788,15907,16011,16194,16328,16336,16342,16468,16485,16646,16708,16754,16791,16989,17083,17129,17171,17191,17229,17256,17325,17352,17462,17480,17550,17576,17851,17902,17966,18057,18060,18073,18108,18197,18447,18543,18618,18626,18668,18797,18866,18883,18897,19040,19114,19174,19193,19403,19734,19742,19772,19781,20245,20338,20390,20492,20591,20708,20804,20961,20992,21055,21099,21130,21150,21168,21476,21675,21717,21783,22012,22039,22088,22110,22166,22281,22339,22381,22393,22412,22441,22498,22607,22656,22893,22894,22999,23016,23227,23359,23484,23509,23566,23571,23700,23794,23839,23935,23995,24191,24316,24461,24469,24636,24664,24886,24913,24938,24987,25019,25022,25126,25155,25191,25260,25362,25368,25516,25594,25658,25848,25963,26002,26128,26145,26280,26283,26486,26584,26647,26665,26758,26881,26996,27168,27178,27209,27248,27376,27389,27466,27573,27578,27613,27627,27726,27743,27859,27868,27976,28051,28126,28217,28218,28299,28443,28449,28639,28666,28673,28704,28762,28912,28943,28969,29056,29167,29195,29221,29253,29465,29544,29712,29751,29752,29773,29776,29936,29987,30101,30114,30128,30137,30311,30352,30402,30404,30536,30573,30664,30714,30758,30769,30832,30868,30933,31048,31077,31219,31246,31341,31346,31355,31605,31672,31701,31704,31749,31765,31787,31824,31843,32251,32319,32375,32408,32428,32508,32510,32558,32570,32575,32632,32732,32788,32916,32963,33009,33178,33182,33186,33227,33264,33468,33489,33594,33625,33697,33861,33865,33934,33985,34023,34074,34174,34200,34264,34364,34367,34409,34466,34589,34634,34657,34731,34874,34980,34999,35130,35222,35237,35264,35359,35687,35822,36162,36169,36195,36210,36336,36343,36390,36417,36552,36589,36607,36621,36636,36646,36659,36668,36712,36730,36783,36987,37059,37160,37173,37178,37222,37260,37266,37274,37465,37491,37501,37516,37613,37645,37703,37750,37771,37773,37927,37971,37991,37998,38044,38107,38111,38134,38192,38270,38298,38313,38362,38458,38679,38784,39119,39266,39561,39750,39938,39948,39957,40466,40709,40760,40966,41038,41074,41169,41184,41199,41296,41459,41550,41571,41621,41644,42017,42125,42256,42303,42326,42340,42566,42587,42613,42625,42775,42805,42827,42901,42972,43011,43014,43147,43165,43207,43210,43302,43308,43328,43350,43358,43406,43413,43479,43619,43691,43711,43775,43796,43861,44093,44221,44263,44342,44464,44815,45007,45060,45110,45130,45146,45172,45305,45342,45467,45539,45613,45664,45701,45829,45927,45929,46450,46480,46517,46572,46824,46878,46942,47041,47070,47080,47091,47177,47202,47367,47510,47539,47551,47552,47623,47700,47852,47874,47899,48152,48164,48244,48246,48340,48398,48426,48493,48495,48509,48548,48661,48724,48769,48877,48936,49153,49164,49261,49307,49347,49377,49653,49695,49704,49742,49785,49845,50363,50516,50682,50733,50760,50884,50910,50955,51004,51010,51092,51159,51202,51284,51317,51341,51516,51610,51621,51768,51930,52009,52043,52127,52355,52441,52654,52714,52771,52786,52792,53017,53119,53138,53184,53243,53292,53420,53520,53698,53885,53984,54083,54179,54281,54413,54442,54560,54583,54614,54858,54893,55029,55226,55560,55568,55739,55820,55962,56070,56185,56246,56325,56377,56479,56518,56536,56607,56807,56842,56865,56951,57047,57138,57215,57425,57438,57457,57488,57610,57808,57831,57835,57905,58107,58163,58326,58486,58673,58789,58826,59030,59037,59245,59322,59341,59342,59446,59451,59455,59531,59607,59608,59640,59918,60193,60232,60258,60281,60369,60504,60759,60891,60995,61132,61149,61171,61708,61772,61818,61882,62138,62231,62233,62317,62479,62509,62668,62676,62736,62984,63078,63304,63326,63344,63366,63435,63641,63657,64405,64461,64484,64567,64573,64612,64657,64730,64744,64891,64999,65023,65070,65419,65498,65576,65607,65657,65779,65818,65891,65906,65934,66132,66153,66229,66334,66350,66401,66403,66422,66531,66545,66679,66732,66765,66961,67072,67139,67214,67266,67340,67433,67673,67708,67775,68062,68190,68209,68280,68349,68369,68426,68435,68486,68661,68672,68692,68880,68883,68901,68917,68926,68936,68998,69280,69486,69505,69658,69672,69854,69913,69977,70053,70074,70084,70353,70507,70521,70660,70719,70818,70826,70845,71076,71164,71180,71237,71280,71287,71289,71346,71436,71455,71599,71659,71772,71854,71866,71869,71909,71930,72011,72073,72188,72530,72543,72583,72631,72710,72766,72783,73035,73206,73455,73524,73681,73710,74005,74019,74049,74061,74064,74105,74245,74413,74425,74452,74522,74689,74748,74778,74829,74852,74890,75118,75132,75171,75217,75276,75386,75491,75517,75576,75733,75798,75848,75886,76120,76155,76244,76282,76326,76402,76523,76579,76675,76712,76809,76910,76916,76972,77003,77016,77046,77223,77303,77385,77513,77552,77666,77693,77844,77870,78013,78039,78064,78090,78135,78299,78543,78550,78730,78733,78741,78880,78889,79080,79107,79156,79485,79512,79548,79696,79810,79877,79953,80009,80010,80134,80187,80565,80642,80854,80864,80885,81057,81219,81232,81241,81263,81279,81310,81321,81350,81372,81454,81500,81506,81513,81648,81663,81700,81731,81771,81911,81951,81958,82013,82071,82306,82394,82490,82641,82737,82767,82843,82868,82882,82975,83009,83060,83149,83216,83235,83261,83349,83448,83502,83534,83575,83642,83743,83772,83862,83970,84186,84217,84243,84345,84435,84517,84585,84766,84856,85014,85145,85219,85325,85366,85431,85490,85502,85509,85526,85634,85673,85728,85843,85877,85920,85932,85950,86006,86028,86152,86191,86199,86210,86278,86454,86468,86512,86687,86796,86820,86882,87019,87109,87119,87213,87374,87403,87456,87460,87721,87862,87896,88043,88077,88079,88103,88156,88202,88204,88224,88283,88419,88450,88465,88497,88599,88636,88649,88707,88736,88830,88856,89138,89318,89413,89590,89623,89673,89780,89926,90023,90118,90300,90310,90359,90409,90441,90468,90545,90795,90893,91033,91051,91118,91203,91205,91248,91261,91270,91294,91319,91343,91513,91551,91591,91637,91681,91892,91924,92156,92339,92402,92440,92456,92654,92778,92910,92948,92952,93041,93179,93243,93260,93293,93369,93382,93408,93457,93569,93599,93609,93680,93826,93882,93944,94293,94336,94395,94508,94629,94638,94667,94768,94781,94787,94816,94893,94994,95048,95191,95200,95604,95759,95891,96112,96284,96295,96304,96331,96394,96427,96453,96667,96700,96811,96907,96908,97002,97050,97060,97136,97400,97406,97467,97499,97538,97595,97695,97741,97821,97920,97966,98195,98290,98400,98589,98598,98758,98906,99003,99019,99028,99068,99168,99346,99359,99369,99456,99630,99636,99738,99776,99826,99949,99972",
          "45,122,133,193,220,242,268,647,699,895,927,967,1055,1434,1526,1589,1601,1633,1636,1690,1717,1718,1830,1862,1902,1907,1928,1976,2002,2115,2210,2218,2275,2312,2409,2427,2565,2598,2607,2810,2884,2954,2960,3046,3059,3091,3153,3349,3378,3426,3481,3550,3668,3702,3736,3817,3861,3877,3905,3992,4036,4190,4291,4545,4850,4991,5414,5485,5494,5567,5674,5682,6027,6128,6175,6251,6330,6390,6394,6465,6510,6523,6536,6550,6641,6643,6650,6749,6772,6820,6855,6911,7063,7114,7122,7239,7262,7531,7703,7775,7805,7831,7991,8028,8071,8090,8116,8227,8231,8287,8391,8413,8460,8661,8915,8946,9017,9378,9396,9531,9738,9873,9907,9920,9969,10011,10075,10151,10209,10408,10419,10442,10523,10590,10602,10647,10708,10793,10812,10825,10846,11017,11132,11168,11194,11197,11236,11362,11492,11559,11629,11673,11833,12014,12021,12064,12146,12161,12163,12191,12254,12267,12364,12410,12426,12499,12521,12537,12549,12580,12619,12683,12753,12852,13051,13196,13205,13254,13365,13405,13496,13550,13569,13591,13642,13664,13804,13887,14025,14104,14188,14214,14312,14415,14461,14570,14661,14833,14840,14847,15011,15022,15105,15223,15227,15357,15399,15482,15489,15565,15618,15720,15755,15956,16019,16100,16145,16148,16152,16167,16220,16265,16346,16420,16447,16480,16640,16653,16716,16750,16767,16774,16811,16836,16895,16913,17035,17185,17210,17225,17280,17307,17318,17436,17500,17515,17520,17546,17560,17609,17678,17788,17866,17945,17973,18120,18257,18292,18323,18340,18394,18516,18564,18580,18646,18686,18689,18691,18698,18700,18712,18744,18804,19028,19147,19172,19263,19294,19352,19470,19506,19523,19532,19567,19741,19788,19793,19823,19836,19936,20012,20110,20132,20203,20206,20207,20273,20278,20313,20364,20368,20383,20558,20587,20597,20618,20637,20689,20756,20848,20955,21016,21031,21065,21138,21166,21259,21282,21462,21550,21627,21634,21635,21665,21725,21730,21800,21812,21901,21949,21961,22005,22051,22070,22100,22163,22226,22233,22241,22269,22274,22307,22352,22443,22596,22613,22645,22647,22657,22735,22843,23031,23053,23127,23176,23207,23212,23283,23528,23548,23581,23591,24040,24050,24131,24246,24337,24531,24533,24555,24749,24778,24799,24969,24986,25012,25053,25063,25072,25134,25157,25187,25188,25281,25406,25483,25486,25562,25587,25600,25689,25711,26021,26036,26039,26117,26149,26168,26196,26346,26361,26466,26523,26565,26655,26909,26932,26988,26991,27012,27098,27099,27134,27208,27329,27385,27439,27448,27484,27490,27506,27622,27631,27632,27647,27700,27738,27740,27875,27889,27945,28057,28091,28122,28153,28169,28245,28256,28265,28306,28313,28414,28550,28567,28589,28621,28624,28735,28808,29012,29095,29149,29187,29192,29212,29223,29290,29334,29357,29401,29412,29439,29464,29481,29511,29524,29550,29556,29658,29682,29826,29862,30005,30095,30117,30134,30143,30181,30190,30192,30298,30308,30332,30432,30487,30579,30607,30679,30744,30755,30775,30830,30851,31070,31148,31163,31200,31295,31340,31360,31408,31499,31595,31615,31638,31820,31838,31867,31880,31887,31901,32046,32061,32106,32281,32355,32440,32478,32547,32550,32595,32675,32843,32857,32993,33064,33076,33118,33194,33365,33439,33446,33463,33474,33553,33627,33646,33762,33767,33919,33924,33973,33991,34084,34453,34770,34796,34821,35192,35194,35241,35376,35489,35578,35645,35698,35748,35990,36030,36098,36334,36538,36800,36940,37002,37131,37285,37412,37538,37644,37650,37651,37676,37721,37729,37756,37838,37898,37902,37969,38000,38010,38032,38088,38112,38231,38373,38537,38610,38705,38779,39028,39038,39238,39262,39299,39337,39469,39571,39606,39705,39738,39763,39814,39819,39846,40021,40164,40279,40522,40612,40699,40706,40827,40863,40872,40936,40941,40953,41012,41046,41105,41190,41255,41344,41430,41470,41534,41572,41609,41668,41879,41946,42019,42077,42172,42243,42269,42333,42438,42460,42529,42535,42565,42654,42758,42811,42853,42855,42885,42926,43174,43177,43264,43382,43419,43505,43537,43576,43686,43928,43941,44324,44415,44416,44449,44502,44536,44572,44683,44934,44944,44990,44994,45012,45020,45066,45078,45089,45135,45173,45196,45198,45225,45292,45293,45301,45312,45694,45715,45735,45773,46065,46138,46146,46171,46211,46490,46587,46653,46748,46838,46891,47143,47155,47212,47229,47364,47394,47479,47607,47707,47906,47987,48052,48053,48082,48124,48178,48205,48274,48306,48325,48403,48448,48524,48584,48630,48683,48789,48890,49007,49025,49050,49113,49224,49260,49314,49498,49611,49668,50055,50058,50083,50187,50765,50806,50921,51064,51155,51162,51214,51242,51255,51282,51300,51329,51350,51373,51402,51426,51540,51542,51558,51583,51623,51851,51880,51919,51932,51961,52053,52060,52088,52120,52377,52416,52506,52517,52550,52638,52666,52805,52807,52849,52876,52882,52963,53025,53076,53113,53223,53224,53291,53371,53373,53414,53418,53503,53648,53660,53675,53694,53759,53872,53920,54121,54169,54381,54449,54499,54518,54543,54686,54870,54877,54928,55015,55104,55245,55255,55278,55280,55305,55425,55453,55717,55761,55774,55861,56019,56045,56047,56292,56300,56514,56529,56538,56572,56664,56677,56781,56813,56945,56961,57017,57098,57213,57246,57423,57475,57490,57529,57759,57760,58078,58082,58126,58128,58143,58175,58212,58216,58222,58241,58330,58441,58459,58609,58668,58921,58966,58969,58988,59011,59128,59149,59157,59176,59224,59228,59498,59500,59702,59772,59781,59833,59865,59899,59910,59932,59955,59969,59984,60063,60099,60195,60230,60256,60341,60370,60431,60437,60441,60505,60526,60643,60653,60752,61055,61300,61364,61919,61998,62001,62254,62341,62343,62354,62402,62659,62661,62699,62845,62858,63047,63268,63311,63318,63374,63418,63470,63567,63572,63608,63621,63668,63803,63934,64042,64101,64103,64105,64144,64240,64619,64850,64853,64882,64908,64942,64958,65065,65135,65139,65142,65195,65207,65275,65291,65403,65626,65757,65803,65816,65862,65871,65915,65985,65986,66006,66017,66103,66164,66277,66296,66346,66470,66559,66566,66691,66809,66856,66917,66960,66987,67031,67300,67498,67587,67592,67602,67614,67677,67908,67957,68119,68213,68288,68291,68393,68427,68504,68551,68714,68761,68763,68804,68856,68866,68889,68959,69076,69153,69164,69263,69389,69398,69474,69516,69584,69639,69660,69945,69990,70149,70187,70552,70569,70610,70805,70823,70912,70959,71022,71024,71073,71184,71202,71319,71381,71503,71525,71636,71651,71824,71831,71851,71921,72020,72277,72299,72340,72428,72493,72498,72528,72544,72691,73197,73244,73272,73387,73393,73537,73557,73583,73616,73635,73774,74034,74182,74185,74211,74450,74456,74508,74526,74706,74845,74980,74984,75158,75271,75309,75478,75606,75770,75822,75854,75985,76032,76041,76075,76208,76261,76567,76603,76649,76723,76760,76865,76926,76959,76965,77606,78067,78088,78094,78103,78230,78358,78474,78558,78671,78699,78804,78900,78963,78986,79060,79108,79346,79412,79572,79635,79678,79798,79823,80023,80176,80204,80267,80323,80429,80432,80506,80518,80558,80611,80791,80968,80972,80980,80993,81048,81277,81286,81323,81424,81567,81615,81691,81715,81800,81846,81879,81886,82098,82114,82168,82246,82396,82612,82655,82727,82820,82896,83026,83070,83126,83193,83378,83435,83481,83514,83569,83590,83591,83662,83886,84030,84053,84080,84096,84432,84462,84477,84613,84622,84921,85242,85312,85372,85376,85407,85411,85420,85427,85515,85524,85691,85828,85840,85945,85956,86026,86079,86182,86185,86268,86425,86477,86526,86529,86595,86609,86823,86874,86888,86899,86905,86998,87066,87227,87261,87478,87541,87571,87631,87817,87921,88123,88154,88324,88358,88404,88508,88594,88626,88727,88737,88840,88938,88940,88981,88995,88999,89160,89176,89740,89804,89831,89893,89911,89975,90217,90273,90408,90509,90760,90789,90828,90924,90957,91037,91080,91195,91271,91401,91448,91460,91488,91520,91525,91532,91686,91725,91902,91983,92080,92124,92129,92205,92222,92230,92254,92380,92420,92548,92639,92830,92850,92897,92997,93417,93611,93705,93780,93864,93975,94086,94159,94193,94227,94369,94422,94470,94551,94747,94901,94957,94996,95034,95087,95111,95210,95273,95424,95441,95483,95511,95551,95591,95765,95779,96000,96002,96058,96127,96268,96370,96405,96480,96560,96575,96612,96717,96748,96849,97022,97042,97108,97283,97607,97629,97654,97656,97661,97667,97750,97944,98071,98083,98153,98158,98169,98298,98404,98687,98700,98783,98791,98857,98902,98910,98924,98942,98974,98979,99014,99075,99151,99264,99356,99365,99388,99561,99604,99677,99804",
          "5,276,304,344,386,389,396,400,412,435,437,476,484,636,1245,1327,1386,5115,5171,5173,5186,5196,5232,5244,5246,5251,5313,5326,5386,5401,5408,5491,5513,5530,5568,5621,5695,5834,5869,5876,5899,5931,5992,6112,6177,6179,6183,6294,6329,6347,6387,6504,6517,6599,6764,6776,7071,7092,7165,7466,7592,7664,7678,7873,7916,8233,8383,8520,8851,8911,8948,9272,9942,9956,10059,10114,10210,10496,10562,10595,10775,11065,11124,11141,11217,11218,11349,11468,11659,11783,11821,11859,11986,12256,12347,12719,12885,12912,13176,14227,14410,15151,16308,16375,16473,16481,16572,16809,16963,17133,17489,18293,18912,19152,19225,20118,21645,23350,29471,38980,41067,41587,42609,44454,50186,50364,50459,50528,50881,50987,51206,51237,51299,51480,51562,51585,51645,51648,51962,52160,52219,52235,52255,52266,52278,52333,52339,52423,52515,52534,52535,52570,52647,52676,52746,52754,52866,52907,52939,53057,53125,53146,53359,53369,53560,53672,53820,53957,54002,54091,54219,54427,54482,54533,54539,54570,54819,54823,54891,54982,55063,55124,55152,55175,55356,55360,55396,55417,55491,55493,55532,55563,55565,55615,55667,55724,55845,55991,56031,56032,56126,56141,56149,56152,56274,56299,56553,56828,56845,56956,56995,57011,57082,57096,57152,57212,57232,57268,57274,57299,57330,57384,57392,57519,57595,57665,57686,57824,57908,57931,57961,58095,58214,58313,58358,58370,58411,58558,58695,58794,58876,58949,58991,59121,59363,59402,59519,59637,59638,59644,59665,59683,59774,59779,59821,59843,59991,60005,60244,60247,60249,60317,60349,60391,60589,60611,60617,60697,60699,60821,60833,60881,60885,61071,61120,61148,61190,61239,61604,61625,61815,61816,61907,61965,62010,62112,62131,62176,62236,62267,62304,62400,62599,62626,62653,62692,62766,62785,62913,63228,63255,63369,63455,63734,63825,63923,63940,64016,64037,64057,64098,64256,64260,64273,64408,64438,64617,64658,64759,65012,65425,65643,65752,65873,65973,65980,66016,66287,66367,66487,66560,66600,66616,66653,66655,66720,66742,66878,67017,67058,67143,67200,67241,67250,67287,67349,67414,67485,67525,67534,67557,67600,67640,67796,67836,68096,68110,68127,68168,68223,68317,68403,68615,68657,68715,68787,69053,69537,69641,69721,69735,69739,69781,69912,69929,70200,70414,70418,70470,70537,70570,70606,70841,70862,70896,70992,70998,71216,71229,71313,71435,71469,71520,71596,71752,71776,71828,71882,71931,71949,72028,72103,72149,72209,72256,72424,72521,72524,72595,72630,72828,73119,73298,73519,73523,73641,73782,73847,73912,74088,74103,74277,74464,74479,74531,74566,74585,74626,74681,74834,74954,75006,75075,75206,75324,75370,75399,75454,75479,76015,76115,76142,76224,76283,76382,76447,76451,76556,76638,76651,76659,76671,76805,76946,77004,77013,77051,77065,77087,77105,77128,77140,77161,77199,77347,77466,77634,77659,77690,77724,77778,77781,77936,78012,78036,78130,78192,78213,78232,78305,78410,78798,78893,78941,79014,79072,79221,79581,79834,79951,80150,80235,80290,80296,80322,80424,80632,80948,80965,81002,81062,81227,81235,81245,81329,81392,81524,81582,81626,81793,81858,81875,81876,81878,81952,82007,82087,82118,82139,82149,82194,82312,82392,82393,82407,82528,82632,82730,82742,82777,82991,83312,83322,83329,83438,83461,83471,83524,83702,83729,83858,83888,83913,84004,84066,84143,84270,84528,84564,84672,84758,84792,84858,84945,85036,85220,85334,85360,85398,85409,85603,85606,85657,85753,85793,85853,85885,86068,86117,86144,86215,86237,86259,86345,86431,86581,86675,86694,86758,86850,86865,86893,87052,87120,87210,87259,87329,87393,87434,87441,87461,87666,87726,87783,87837,87901,87955,87975,88049,88100,88213,88299,88318,88467,88495,88498,88715,88766,88866,89050,89215,89259,89373,89409,89543,89609,89828,89933,89990,90083,90296,90573,90575,90633,90684,90882,91029,91076,91163,91336,91715,91808,91969,92084,92179,92288,92386,92463,92482,92492,92582,92648,92717,92759,92779,92789,92799,92837,92847,92893,93112,93222,93258,93279,93508,93541,93629,93663,93689,93753,93945,93947,94102,94104,94133,94192,94246,94248,94315,94382,94399,94406,94505,94514,94683,94753,94811,94953,95054,95095,95144,95189,95325,95345,95391,95399,95490,95510,95607,95611,96102,96143,96168,96334,96366,96484,96524,96534,96549,96608,96627,96683,96854,97049,97397,97475,97554,97609,97890,98077,98377,98386,98463,98545,98772,99103,99133,99245,99252,99253,99262,99391,99491,99608,99621,99714,99731,99750,99790,99850,99863",
          "1,113,190,360,408,458,610,625,738,785,804,896,944,1015,1063,1193,1265,1387,1506,1549,1662,1664,1676,1713,2127,2167,2192,2340,2367,2395,2410,2454,2511,2630,2670,2673,2687,2695,2701,2728,2799,2976,3038,3117,3200,3697,4042,4143,4407,4471,4570,4664,4740,4788,4836,4922,4959,5008,5036,5204,5209,5360,5449,5516,5643,5733,5968,6015,6869,6974,7027,7106,7212,7237,7295,7318,7324,7577,7625,7683,7899,7946,8046,8300,8324,8354,8380,8381,8722,9114,9204,9348,9380,9387,9430,9495,9512,9549,9676,9745,9808,9826,9839,9857,9952,10006,10098,10137,10212,10313,10373,10407,10467,10660,10706,10742,10852,10906,10950,11013,11051,11154,11155,11164,11174,11230,11276,11352,11424,11465,11665,11699,11759,11760,11774,11919,12066,12120,12154,12263,12321,12328,12435,12541,12677,12837,12872,12898,13095,13098,13116,13239,13282,13348,13397,13494,13524,13572,13617,13792,13816,13910,13922,13976,14061,14238,14281,14368,14511,14655,14754,14812,14899,14973,15058,15099,15170,15252,15277,15329,15384,15388,15420,15545,15577,15691,15742,16136,16304,16394,16562,16787,16842,16848,16862,16875,16900,16929,16934,16992,17146,17176,17254,17372,17375,17432,17708,17777,17824,18225,18231,18232,18430,18513,18517,18536,18570,18684,18685,18825,18919,18999,19100,19161,19189,19234,19237,19306,19437,19443,19509,19578,19622,19624,19667,19763,19825,20020,20061,20105,20112,20140,20202,20209,20243,20268,20281,20403,20497,20524,20636,20672,20681,20701,20773,20784,20835,20856,21038,21142,21176,21365,21552,21751,21947,22029,22030,22148,22208,22334,22421,22442,22477,22508,22540,22668,22713,22782,22789,22846,23043,23052,23091,23093,23361,23454,23524,23604,23643,23855,23863,24062,24166,24174,24208,24217,24225,24307,24392,24452,24654,24655,24701,24789,24856,24926,24940,24946,24989,25036,25045,25146,25176,25222,25396,25405,25412,25431,25440,25465,25656,25748,25872,25910,26072,26109,26265,26349,26660,26678,26856,26958,26993,27193,27279,27447,27456,27546,27606,27619,27806,27807,27903,28029,28041,28161,28166,28186,28267,28284,28341,28465,28556,28820,28898,28964,28982,29046,29437,29444,29507,29627,29738,29739,29919,29957,30074,30191,30418,30510,30623,30693,30715,30807,30844,30970,31095,31129,31273,31397,31513,31588,31665,31980,32000,32030,32155,32261,32327,32473,32695,32721,32748,32771,32875,33143,33154,33344,33530,33617,33715,33742,33811,33927,34133,34220,34244,34312,34321,34350,34360,34378,34397,34460,34538,34669,34748,35052,35102,35477,35488,35584,35810,35905,35906,35920,35943,35946,35999,36019,36270,36309,36420,36508,36525,36557,36619,36759,36891,36936,37004,37280,37284,37344,37362,37427,37505,37530,37542,37580,37778,37900,37919,37959,37968,37985,38081,38257,38258,38272,38359,38401,38429,38439,38452,38749,38782,38814,38846,39035,39045,39102,39303,39505,39614,39659,39879,40237,40322,40486,40558,40677,40692,40750,40940,41050,41101,41251,41289,41304,41379,41380,41411,41505,41645,41671,41858,41917,41986,42095,42103,42171,42356,42397,42422,42455,42474,42478,42509,42617,42683,42732,42778,42785,42808,42956,43049,43097,43101,43301,43332,43368,43383,43603,43639,43640,43659,43718,43848,44036,44068,44174,44362,44381,44389,44581,44588,44665,44724,44772,44912,45093,45248,45379,45440,45463,45489,45518,45592,45671,45744,45844,45848,45922,46007,46014,46062,46139,46200,46412,46481,46527,46632,46659,46698,46749,46775,46881,46919,46935,47011,47028,47092,47132,47303,47319,47326,47402,47454,47614,47646,47764,47766,47790,47940,47960,48074,48130,48137,48162,48240,48277,48328,48374,48424,48457,48469,48484,48492,48549,48703,48819,48901,48955,49029,49076,49078,49149,49159,49172,49343,49769,49840,49857,49920,49940,49980,50240,50319,50442,50472,50561,50904,50983,51070,51139,51293,51386,51391,51437,51447,51658,51703,51815,51831,52131,52140,52267,52419,52615,52706,53072,53102,53209,53249,53280,53281,53299,53309,53403,53454,53490,53557,53575,53721,53807,53811,53832,53851,53884,53888,53995,54008,54016,54045,54168,54176,54214,54322,54360,54474,54609,54718,54761,54902,55097,55411,55481,55537,55580,55638,55690,55915,55932,56033,56191,56251,56402,56404,56432,56530,56638,56641,56740,56847,56877,56954,56976,57014,57016,57092,57236,57264,57296,57459,57478,57796,57837,58027,58097,58183,58315,58433,58436,58491,58495,58747,58796,58955,58975,59307,59357,59808,59844,60008,60351,60394,60554,60629,60684,60724,60727,60996,61028,61039,61068,61114,61146,61291,61428,61614,61645,61700,61732,61892,61945,62017,62080,62134,62160,62239,62388,62457,62510,62628,62669,62693,62819,62980,63074,63089,63163,63186,63222,63237,63242,63484,63548,63584,63733,63797,63988,64197,64224,64229,64297,64352,64400,64675,65240,65268,65277,65318,65348,65436,65453,65533,65536,65591,65631,65659,65836,65863,65867,66047,66074,66082,66121,66380,66385,66417,66512,66551,66953,66988,67103,67175,67424,67434,67503,67517,67772,67825,67890,67907,67925,67946,68029,68446,68497,68539,68561,68593,68795,68823,69001,69077,69092,69104,69132,69150,69224,69246,69247,69373,69526,69894,69943,69989,70090,70315,70380,70457,70542,70605,70620,70681,70696,70702,70743,70868,70880,70905,70969,70973,70996,71413,71449,71466,71512,71648,71649,71897,72105,72227,72364,72382,72441,72511,72632,72836,72879,72986,73027,73036,73166,73233,73289,73413,73433,73434,73454,73469,73687,73869,74030,74040,74095,74143,74215,74310,74433,74709,74723,74757,74774,74918,74995,75152,75167,75250,75350,75354,75360,75425,75439,75473,75536,75543,75807,76113,76175,76213,76500,76647,76657,76945,77073,77074,77121,77158,77249,77254,77421,77441,77729,77769,77957,78361,78633,78640,78703,78927,78981,79069,79100,79336,79449,79483,79573,79607,79660,79760,79770,79939,79979,80101,80205,80231,80401,80466,80520,80540,80567,80655,80691,80762,80768,80838,80942,81012,81067,81083,81100,81274,81333,81389,81462,81469,81860,81929,81950,82015,82333,82364,82421,82545,82665,82750,82763,82817,82861,82864,82898,82966,82969,83094,83107,83118,83201,83207,83262,83330,83390,83626,83663,83741,83805,83895,84122,84147,84221,84272,84305,84414,84473,84520,84701,84703,84824,84903,85051,85216,85290,85330,85461,85463,85668,85977,85992,86045,86080,86104,86206,86239,86295,86339,86420,86507,86567,86593,86672,87284,87304,87410,87510,87729,87775,87791,87856,87889,87984,87999,88143,88145,88183,88185,88250,88254,88379,88445,88752,88806,88867,89037,89063,89094,89114,89278,89298,89524,89589,89666,89772,89822,89825,90024,90076,90184,90267,90277,90396,90428,90443,90597,90652,90869,90883,91016,91073,91151,91166,91682,91698,91761,91771,91863,91907,91910,91939,92064,92073,92143,92304,92424,92468,92625,92928,92947,92975,93096,93449,93638,93719,94052,94132,94166,94237,94254,94375,94457,94465,94512,94537,94798,94934,95027,95070,95109,95211,95406,95468,95469,95568,95594,95700,96072,96131,96250,96252,96258,96285,96435,96517,96604,96694,96733,96871,96910,96948,97007,97073,97157,97260,97490,97503,97596,97668,97827,97835,97872,97938,98276,98292,98340,98389,98491,98580,98648,98650,98651,98676,98705,98709,98800,98886,99107,99161,99214,99227,99383,99718,99721,99756,99825,99877,99901,99907",
          "2,21,237,428,460,588,706,734,783,842,860,996,1045,1076,1223,1357,1443,1518,1797,1888,1895,2363,2416,2484,2487,2503,2583,2843,2893,2926,3002,3144,3186,3298,3302,3363,3433,3567,3594,4058,4142,4147,4238,4307,4327,4433,4502,4550,4703,4803,4885,4973,5025,5049,5144,5344,5388,5471,5502,5641,5680,5699,5889,6071,6274,6398,6423,6660,6661,6669,6786,6821,6960,7108,7269,7331,7524,7687,7892,7929,8043,8082,8207,8310,8332,8362,8429,8446,8502,8571,8576,8636,8666,8674,8928,9027,9135,9179,9195,9243,9253,9358,9371,9384,9388,9562,9625,9726,9757,9780,9909,9991,10086,10159,10213,10279,10312,10339,10393,10415,10516,10905,10986,11001,11120,11254,11335,11398,11453,11470,11554,11824,11889,12187,12188,12211,12270,12284,12405,12495,12603,12701,12745,12882,13002,13123,13261,13293,13323,13351,13429,13440,13520,13699,13729,13926,13934,13981,13991,14020,14213,14223,14241,14303,14365,14783,14790,15116,15193,15331,15396,15413,15464,15533,15543,15567,15646,15756,15759,15763,15800,15817,15830,15859,15919,16161,16173,16258,16384,16474,16510,16550,16583,16594,16608,16987,17123,17124,17263,17335,17389,17443,17456,18029,18038,18472,18530,18542,18716,18721,18725,18764,18840,18993,19212,19312,19365,19379,19390,19410,19428,19502,19503,19850,19904,19968,20011,20136,20198,20201,20263,20421,20454,20482,20569,20677,20694,21029,21035,21096,21240,21362,21370,21443,21450,21610,21633,21735,21745,21754,21797,21801,21870,21883,21893,21917,21955,22094,22173,22197,22416,22430,22479,22535,22573,22617,22652,22685,22802,22852,22879,22931,22948,23032,23038,23275,23550,23768,23833,24006,24054,24162,24196,24285,24356,24395,24451,24479,24550,24556,24682,25070,25128,25156,25207,25240,25359,25551,25736,25813,25918,25939,26095,26147,26167,26281,26396,26490,26623,26766,26855,27045,27149,27200,27391,27398,27449,27450,27536,27695,27715,27787,27885,27913,28071,28093,28162,28454,28588,28678,28721,28887,28940,28944,28993,29164,29199,29207,29317,29377,29473,29486,29487,29535,29553,29685,29947,29949,30010,30103,30348,30410,30596,30670,30681,30707,30782,30897,30913,30954,31169,31258,31292,31352,31469,31602,31708,31745,31748,31906,31931,31990,32064,32135,32151,32153,32184,32210,32211,32381,32383,32445,32962,33150,33250,33356,33389,33391,33472,33685,33711,33878,33883,33897,33901,33923,33964,34137,34176,34210,34266,34347,34382,34459,34754,34762,34773,34784,34851,34927,34995,35069,35335,35505,35536,35597,35667,35686,35852,35932,35939,36013,36054,36106,36121,36149,36361,36670,36683,36685,36754,36960,37079,37120,37153,37399,37432,37495,37564,37570,37793,37868,38021,38052,38124,38263,38528,38709,38848,38873,38948,38954,38967,38982,38989,39080,39145,39263,39279,39304,39400,39501,39504,39507,39737,39761,39768,39772,39790,39831,39838,39886,39976,39998,40135,40244,40348,40362,40371,40372,40476,40492,40592,40610,40660,40663,40684,40698,40862,40895,40945,40978,40981,41040,41263,41267,41401,41407,41532,41581,41690,41827,42076,42197,42212,42332,42354,42578,42616,42690,42779,42999,43131,43175,43205,43212,43240,43344,43380,43402,43421,43432,43484,43511,43547,43604,43632,43715,43939,44071,44079,44247,44268,44373,44420,44426,44475,44531,44594,44649,44735,44746,45084,45102,45272,45273,45315,45341,45348,45424,45578,45697,45786,45789,45855,45858,45859,46006,46132,46229,46343,46363,46449,46558,46584,46626,46703,46957,47017,47019,47024,47181,47296,47324,47407,47665,47760,47794,47813,47917,48115,48155,48213,48380,48532,48582,48605,48611,48653,48664,48747,48764,48767,48856,48940,48965,49040,49047,49089,49362,49558,49661,49670,49671,49765,49804,49822,49934,50009,50202,50205,50219,50255,50322,50327,50385,50565,50612,50836,51397,51450,51461,51561,51609,51626,51664,51894,51963,52080,52362,52398,52424,52552,52560,52790,53078,53211,53236,53375,53493,53531,53582,53657,53658,53768,53799,53848,53966,53977,54090,54147,54191,54224,54254,54477,54520,54587,54619,54653,54882,55009,55016,55088,55143,55228,55346,55366,55390,55551,55561,55647,55655,55691,55751,55854,55968,56022,56111,56197,56256,56273,56285,56317,56458,56464,56493,56520,56523,56605,56690,56739,56770,56830,56863,56889,56917,56920,57042,57154,57177,57252,57394,57397,57530,57713,57795,57805,57856,57923,57936,58041,58044,58115,58149,58182,58490,58612,58855,58992,58998,59038,59074,59143,59271,59490,59497,59504,59634,59851,59906,59909,60013,60066,60138,60175,60197,60229,60464,60475,60588,60680,60720,60740,60762,60794,60802,60862,60904,60925,61054,61108,61119,61139,61178,61308,61352,61399,61553,61564,61619,61725,61761,61994,62004,62094,62165,62194,62201,62260,62309,62334,62338,62679,62685,62731,62927,62961,63072,63105,63169,63212,63666,63721,63744,63869,64028,64034,64145,64361,64373,64389,64395,64594,64620,64861,65026,65057,65177,65179,65184,65199,65241,65528,65581,65625,65680,65883,65978,66013,66083,66134,66354,66373,66433,66507,66524,66590,66606,66853,66873,66885,66938,66997,67049,67066,67123,67134,67194,67198,67278,67410,67444,67522,67722,67745,67789,67852,67870,67981,68124,68139,68205,68327,68347,68413,68547,68566,68687,68835,68865,68963,69109,69218,69236,69256,69325,69580,69588,69636,69673,69680,69707,69764,70273,70361,70378,70381,70439,70484,70489,70624,70822,70888,70955,71162,71254,71396,71568,71598,71678,71781,71816,71835,71953,72106,72264,72442,72485,72615,72790,72937,73024,73061,73128,73134,73333,73563,73755,73849,73855,74060,74155,74209,74242,74300,74427,74487,74598,74619,74699,74781,74941,75023,75040,75045,75196,75434,75514,75561,75599,75940,75991,75993,76089,76127,76161,76227,76304,76441,76471,76608,76616,76714,76819,76924,77202,77221,77374,77428,77545,77639,77670,77839,77850,77934,77941,77945,78045,78112,78264,78315,78501,78826,78901,79386,79398,79497,79513,79539,79629,79716,79923,79964,80063,80089,80139,80147,80161,80177,80362,80740,80964,81023,81061,81140,81538,81547,81720,81727,81984,82003,82024,82047,82342,82425,82626,82793,82805,82855,82889,82913,83088,83135,83152,83456,83543,83601,83654,83692,83921,83987,84017,84145,84234,84369,84379,84413,84445,84548,84715,84950,84976,84991,85040,85105,85113,85189,85268,85399,85405,85497,85508,85669,85749,86110,86114,86125,86169,86318,86383,86619,86699,86704,86817,86829,86859,87026,87069,87207,87306,87375,87390,87402,87497,87728,87743,87908,88087,88214,88317,88327,88483,88624,88822,88857,88917,88933,89080,89081,89333,89515,89649,89706,89980,89986,90071,90471,90531,90553,90564,90662,90780,91143,91179,91210,91226,91299,91355,91485,91548,91692,91786,91860,91951,91997,92005,92052,92376,92416,92534,92556,92591,92604,92685,92687,92691,92725,92796,92823,92926,93028,93081,93085,93127,93141,93226,93323,93413,93533,93577,93815,93932,93991,94066,94142,94179,94196,94205,94225,94499,94500,94644,94758,94835,94886,95012,95036,95060,95071,95216,95256,95321,95368,95371,95404,95514,95935,95952,95953,96085,96124,96188,96243,96311,96583,96625,96648,96806,96878,97006,97204,97332,97345,97355,97363,97427,97536,97579,97659,97844,97892,97897,97984,98026,98079,98194,98233,98339,98480,98560,98601,98710,98796,99073,99282,99393,99434,99470,99822,99912,99940",
          "98,177,212,213,312,405,423,461,611,635,645,655,664,717,747,781,881,931,1001,1070,1146,1150,1325,1383,1389,1469,1534,1574,1880,1922,1927,1966,1980,1998,2152,2173,2188,2217,2220,2236,2245,2293,2731,2844,2920,3174,3210,3245,3385,3538,3637,3648,3717,3999,4059,4280,4504,4559,4620,4642,4647,4668,4673,4709,4738,4822,4845,4890,4916,5156,5160,5169,5285,5335,5359,5398,5422,5521,5550,5569,5612,5843,5911,5922,5997,6037,6063,6070,6111,6193,6195,6341,6424,6799,6809,7085,7093,7227,7535,7584,7626,7646,7716,7766,7825,7869,8151,8175,8176,8258,8303,8338,8349,8402,8475,8486,8534,8562,8570,8615,8672,8767,8803,8942,9030,9104,9310,9322,9383,9434,9451,9458,9585,9627,9908,9916,10077,10190,10204,10240,10376,10398,10703,10827,10931,10994,11000,11125,11163,11431,11562,11698,11703,11707,11729,11757,12322,12388,12471,12501,12564,12655,12873,13007,13083,13171,13192,13209,13260,13288,13485,13492,13505,13633,13725,13815,13851,14093,14109,14113,14532,14560,14673,14742,14956,15089,15197,15297,15371,15507,15521,15554,15798,15904,15946,15992,16186,16387,16454,16457,16460,16477,16524,16579,16588,16589,16664,16709,16861,16938,17053,17068,17073,17319,17369,17460,17575,17590,17647,17689,17712,17920,17924,18139,18250,18281,18290,18416,18614,18635,18641,18695,18858,18939,19055,19098,19150,19202,19228,19230,19544,19592,19619,19798,19882,20009,20157,20221,20265,20581,20601,20711,20770,20791,20812,20937,20984,21037,21064,21098,21199,21358,21438,21480,21517,21578,21669,21709,21795,21992,22234,22379,22402,22493,22571,22984,22991,23039,23083,23147,23170,23269,23394,23520,23573,23759,23866,23923,24021,24157,24244,24737,24806,24843,24857,24947,25023,25095,25182,25221,25466,25599,25680,25710,25862,25952,25992,25996,26105,26452,26605,26619,26651,26730,26783,27170,27277,27310,27344,27384,27423,27498,27582,27655,27681,27818,27821,27826,27860,27982,28042,28278,28684,28776,28900,28932,29065,29118,29127,29230,29246,29286,29355,29624,29750,29851,30163,30172,30310,30316,30514,30561,30617,30630,30717,30849,30889,30908,31040,31043,31216,31310,31351,31353,31412,31462,31593,31829,31879,31883,31902,31974,32021,32052,32054,32063,32094,32218,32321,32354,32380,32388,32455,32519,32551,32564,32601,32685,32852,32865,33087,33110,33324,33354,33382,33525,33537,33543,33568,33581,33600,33881,34204,34439,34454,34512,34557,34559,34887,34943,35200,35230,35240,35244,35305,35347,35450,35482,35504,35520,35616,35720,35730,35790,35912,35928,35961,36016,36158,36170,36214,36279,36342,36426,36527,36548,36600,36633,36637,36766,36905,37033,37277,37336,37437,37446,37519,37578,37814,37842,37862,38148,38290,38307,38372,38481,38521,38582,38809,38880,38892,39042,39204,39236,39264,39293,39562,39694,40005,40045,40069,40130,40316,40328,40409,40571,40604,40656,40701,40989,41111,41196,41418,41504,41612,41667,41681,41701,41796,41841,41857,41934,42083,42105,42170,42183,42187,42262,42263,42309,42534,42567,42687,42755,42833,42937,42974,43034,43361,43452,43518,43618,43712,43873,43964,44002,44155,44173,44276,44300,44414,44780,44808,44854,44917,44966,45024,45176,45199,45439,45717,45785,45799,45810,45867,45887,45957,45960,45961,46085,46129,46131,46158,46202,46260,46417,46489,46539,46544,46588,46620,46622,46702,46708,46740,46778,46790,46794,46980,47004,47153,47253,47276,47333,47342,47420,47625,47745,47830,47891,47921,48060,48151,48166,48197,48212,48303,48314,48320,48356,48513,48613,48635,48666,48911,48912,48960,49064,49106,49132,49167,49176,49178,49242,49368,49390,49407,49584,49630,49721,49886,49919,49926,49946,49959,50123,50313,50395,50547,50755,50886,50960,50973,50977,51372,51468,51551,51608,51616,51767,51813,51827,51920,52148,52156,52173,52493,52526,52779,52869,52970,53087,53103,53214,53241,53246,53287,53320,53510,53645,53649,53707,53710,53952,53962,54057,54082,54350,54362,54366,54382,54389,54411,54416,54467,54548,54635,54721,54749,54808,54822,54850,54918,54939,55023,55041,55194,55294,55497,55570,55865,55909,55945,55982,56037,56040,56046,56064,56073,56084,56087,56369,56400,56469,56861,56905,56985,56986,57027,57100,57201,57284,57285,57340,57341,57345,57348,57506,57702,57879,57896,57952,58331,58353,58475,58582,58666,58744,58822,58936,58977,59029,59132,59292,59377,59437,59456,59483,59581,59700,59733,59766,59799,59804,59903,59943,59998,60139,60166,60227,60228,60245,60304,60413,60465,60548,60659,60901,60916,60928,60973,61035,61041,61099,61177,61370,61431,61527,61597,61654,61766,61857,61953,61964,61979,62191,62243,62391,62453,62499,62551,62634,62635,62651,62724,62853,62941,63038,63116,63137,63161,63173,63188,63199,63425,63440,63633,63767,63976,64058,64076,64090,64154,64175,64245,64506,64545,64783,65024,65064,65272,65454,65734,65838,65969,66027,66097,66143,66167,66171,66611,66631,66813,66905,66912,66950,66968,67004,67174,67244,67391,67505,67672,67688,67835,67840,67856,67874,67927,67943,68169,68230,68247,68278,68312,68337,68452,68522,68782,68827,68862,68949,69141,69174,69446,69561,69626,69677,69750,69766,69870,70282,70297,70356,70376,70478,70483,70577,70615,70629,70738,70852,70899,70903,71293,71375,71454,71617,71840,71858,72037,72069,72165,72171,72175,72298,72312,72349,72409,72490,72608,72841,72983,73178,73254,73446,73491,73532,73555,73603,73642,73650,73684,73716,73784,73791,73930,73944,73977,73982,74101,74139,74164,74230,74346,74407,74480,74502,74529,74747,74760,74883,75097,75106,75146,75246,75311,75373,75395,75505,75560,75652,75672,75721,75744,75769,75789,75825,75995,76053,76153,76216,76290,76336,76342,76368,76468,76525,76604,76689,76906,76987,77032,77045,77076,77084,77365,77426,77628,77970,77981,77997,78127,78161,78211,78425,78434,78571,78692,78899,78925,79029,79047,79228,79284,79340,79379,79439,79620,79645,79740,79766,80268,80377,80848,80855,81015,81116,81496,81753,81902,81920,82012,82020,82041,82046,82054,82094,82288,82387,82399,82551,82771,82830,82862,82916,82939,83076,83320,83379,83518,83619,83652,83675,83781,84052,84100,84129,84260,84329,84449,84451,84509,84609,84735,84773,84784,84793,84916,85004,85106,85133,85142,85284,85343,85435,85468,85474,85486,85496,85685,85742,85751,85807,86048,86160,86186,86293,86469,86492,86639,86695,86735,86862,87012,87095,87104,87124,87191,87224,87252,87282,87494,87531,87649,87952,88012,88028,88080,88286,88475,88550,88564,88713,88824,88855,88908,88965,89027,89053,89101,89130,89220,89440,89541,89653,89805,89919,89999,90128,90140,90258,90311,90369,90417,90465,90505,90565,90654,90657,90698,90842,90991,91018,91020,91157,91288,91538,91602,91621,91625,91684,91712,91726,91841,91993,92002,92091,92192,92296,92322,92608,92626,92627,92684,92852,92865,92918,92924,93047,93238,93252,93321,93423,93463,93511,93557,93563,93582,93849,93866,93941,93994,94036,94062,94152,94161,94207,94209,94239,94292,94304,94492,94695,94709,94790,94840,94849,94852,94870,95057,95306,95355,95512,95807,95917,96029,96043,96251,96266,96290,96298,96306,96359,96368,96383,96621,96739,97053,97258,97584,97771,98015,98043,98055,98160,98182,98324,98367,98444,98795,98975,98978,99023,99059,99164,99275,99290,99299,99317,99405,99540,99887,99923",
          "51,233,469,571,584,634,680,722,726,873,897,1053,1060,1108,1147,1208,1301,1305,1341,1352,1544,1573,1595,1652,1751,1837,1887,1956,1992,2069,2163,2369,2429,2486,2603,2625,2634,2942,2946,2971,2985,3056,3074,3083,3084,3217,3223,3239,3442,3447,3453,3580,3623,3659,3662,3663,3705,3707,3874,3888,3901,4002,4055,4520,4622,4636,4641,4659,4675,4705,4752,5116,5213,5387,5487,5511,5515,5538,5539,5663,5775,5818,6055,6154,6155,6168,6277,6365,6436,6459,6576,6584,6631,6673,6778,6782,6793,6847,6849,6873,6876,7005,7038,7217,7264,7268,7313,7446,7581,7613,7718,7740,7754,7843,7859,7866,7883,7911,8106,8220,8230,8232,8274,8346,8368,8397,8401,8405,8504,8544,8611,8757,8972,8999,9003,9018,9144,9237,9263,9497,9516,9605,9806,9897,10002,10024,10249,10278,10444,10525,10536,10598,10758,10809,10814,10892,10977,11170,11244,11308,11313,11367,11403,11448,11452,11471,11526,11558,11651,11653,11813,11879,11957,12008,12019,12056,12216,12225,12277,12302,12307,12381,12519,12621,12694,12700,12710,12756,12856,12860,12875,12936,13104,13127,13195,13203,13395,13425,13486,13556,13682,13695,13764,13847,13907,14016,14163,14185,14198,14243,14263,14374,14586,14705,14721,14726,14782,14882,14936,15186,15251,15296,15343,15412,15526,15557,15644,15656,15748,15880,16022,16054,16061,16095,16150,16243,16257,16335,16406,16441,16496,16597,16599,16647,16729,16770,16850,16892,16898,17132,17272,17316,17343,17387,17468,17561,17601,17787,17853,17884,18042,18159,18185,18339,18500,18547,18761,18872,18964,19063,19155,19186,19218,19270,19296,19415,19559,19595,19659,19729,19771,19782,19940,19954,19987,20130,20167,20196,20330,20376,20377,20382,20394,20413,20496,20622,20811,21033,21072,21081,21083,21163,21165,21169,21244,21345,21352,21391,21491,21518,21530,21596,21690,21733,21792,21810,22022,22151,22152,22349,22365,22405,22423,22482,22610,22665,22748,22900,22917,23013,23079,23190,23352,23399,23512,23625,23684,24019,24027,24114,24189,24227,24261,24276,24372,24383,24442,24510,24600,24674,25029,25081,25121,25162,25346,25348,25421,25432,25477,25511,25579,25626,25844,25847,25959,26097,26119,26192,26204,26233,26388,26437,26465,26482,26579,26894,27060,27066,27123,27162,27163,27210,27233,27255,27323,27432,27451,27678,27888,28022,28095,28116,28120,28160,28246,28370,28402,28460,28528,28602,28605,28645,28649,28850,28927,28960,28984,29156,29161,29163,29319,29385,29423,29539,29632,29695,29775,29789,29823,29829,29922,29952,30142,30169,30195,30204,30443,30546,30657,30768,30819,30911,30914,30935,30957,30996,31126,31349,31496,31517,31524,31530,31576,31622,31685,31700,31711,31793,31819,31861,31977,32083,32132,32227,32298,32309,32374,32397,32411,32441,32471,32524,32585,32592,32712,32737,32783,32810,32868,32909,33033,33176,33326,33369,33387,33400,33509,33570,33611,33641,33677,33707,33826,33854,33880,33938,34009,34109,34114,34144,34184,34208,34289,34314,34365,34422,34526,34577,34617,34801,34828,34869,34889,35168,35179,35285,35399,35430,35517,35642,35722,35724,35902,35984,35992,36126,36242,36488,36512,36575,36601,36715,36799,36887,36993,37057,37119,37136,37297,37408,37563,37595,37724,37761,37767,37872,37914,37928,37953,38046,38066,38071,38084,38114,38210,38323,38345,38430,38473,38488,38597,38947,39040,39057,39183,39190,39207,39227,39291,39383,39393,39399,39446,39550,39572,39608,39647,39666,39669,39729,39755,39766,39778,39783,39884,39905,39952,39974,40058,40103,40183,40349,40408,40452,40503,40539,40588,40607,40626,40775,40812,40846,41013,41085,41129,41409,41445,41458,41495,41512,41610,41622,41657,41672,41805,41807,41814,41876,41905,42046,42079,42123,42211,42218,42227,42272,42339,42380,42408,42432,42472,42562,42845,42976,42983,43022,43062,43079,43213,43351,43399,43416,43424,43487,43608,43660,43678,43774,43778,43808,43892,44043,44047,44215,44440,44443,44512,44551,44605,44633,44673,44935,45031,45032,45059,45068,45091,45244,45278,45418,45493,45508,45527,45545,45560,45575,45651,45655,45714,45781,45788,45856,45988,46086,46111,46123,46175,46247,46288,46425,46557,46565,46592,46686,46783,46905,46906,46945,46989,47015,47029,47036,47079,47089,47239,47340,47426,47433,47515,47571,47612,47651,47687,47723,47867,47877,47951,47967,47977,48011,48026,48104,48108,48156,48189,48239,48293,48341,48358,48387,48390,48439,48511,48560,48608,48766,48803,48863,48939,48953,48982,49015,49039,49079,49196,49219,49222,49262,49330,49379,49392,49666,49837,50018,50140,50173,50323,50393,50541,50822,50924,51062,51199,51208,51268,51290,51401,51528,51716,51745,51833,51978,52010,52018,52050,52241,52287,52510,52548,52581,52604,52624,52628,52858,52969,52975,52991,53009,53207,53376,53458,53593,53874,54079,54089,54108,54174,54283,54426,54440,54510,54513,54620,54664,54671,54768,54770,54772,54820,55030,55216,55315,55323,55347,55535,55574,55627,55636,55665,55680,55771,55790,55927,56618,56825,56831,57020,57070,57081,57260,57306,57458,57692,57722,57940,58014,58047,58075,58224,58249,58334,58571,58698,58771,58810,58917,59034,59199,59242,59253,59259,59327,59332,59343,59349,59353,59359,59589,59718,59729,60023,60053,60174,60250,60252,60262,60330,60518,60669,60722,60761,60918,60977,60983,61118,61143,61202,61245,61263,61270,61388,61419,61430,61466,61575,61783,61786,61833,61843,61874,61887,61955,62035,62083,62180,62219,62448,62451,62487,62522,62533,62563,62587,62607,62627,62646,62660,62667,62738,62795,62841,63071,63181,63390,63450,63491,63501,63569,63587,63600,63630,63725,63756,63857,63947,63998,64314,64326,64372,64402,64524,64581,64615,64723,64753,64814,65079,65197,65225,65293,65423,65546,65684,65775,65888,66028,66057,66081,66135,66177,66280,66282,66308,66341,66369,66410,66580,66597,66680,66698,66747,67073,67089,67189,67282,67400,67409,67427,67507,67667,67702,67714,67743,67785,67798,67846,67868,67883,67916,68042,68084,68089,68199,68206,68354,68454,68473,68565,68706,69060,69113,69162,69171,69196,69299,69411,69481,69578,69612,69619,69631,69646,69785,69800,69804,69818,69862,69956,69961,70019,70022,70154,70210,70244,70311,70502,70640,70787,70820,70844,70846,70859,70944,71031,71041,71042,71080,71235,71272,71277,71286,71305,71307,71377,71456,71467,71625,71669,71785,71889,71898,71918,71941,71985,72080,72134,72184,72206,72344,72372,72400,72476,72643,72667,72687,72734,72813,72864,72934,72940,72996,73091,73136,73161,73225,73264,73280,73295,73299,73321,73356,73398,73466,73493,73606,73657,73708,73767,73843,73900,74153,74266,74284,74289,74305,74397,74428,74484,74511,74559,74570,74577,74777,74802,74820,74828,74864,74876,74948,75119,75236,75335,75512,75574,75638,75657,76084,76094,76124,76171,76212,76366,76404,76457,76501,76522,76527,76549,76561,76588,76612,76636,76655,76664,76708,76716,76804,76996,77165,77309,77324,77384,77429,77551,77572,77590,77594,77795,77854,77862,78113,78133,78197,78261,78333,78447,78482,78525,78708,78723,78769,78783,78786,78870,78884,78888,79173,79231,79261,79335,79353,79431,79446,79463,79535,79556,79610,79668,79670,79879,79915,80084,80236,80248,80327,80329,80395,80411,80449,80451,80513,80524,80659,80793,80813,80819,80857,80899,81007,81034,81170,81210,81260,81265,81400,81444,81477,81489,81645,81667,81732,81845,81895,81932,81943,82060,82072,82107,82165,82191,82347,82348,82371,82485,82494,82584,82594,82668,82669,82707,82754,82922,83116,83151,83215,83232,83279,83280,83296,83370,83397,83420,83533,83557,83676,83704,84094,84590,84676,84820,84914,84961,84975,85001,85010,85013,85057,85181,85249,85259,85285,85374,85568,85570,85666,85679,85720,85725,85954,86022,86090,86138,86158,86273,86343,86403,86409,86458,86574,86587,86605,86657,86709,86710,86951,87025,87087,87088,87195,87243,87256,87364,87379,87538,87949,87962,88142,88294,88334,88782,88868,89029,89143,89149,89255,89313,89334,89422,89512,89572,89611,89688,89693,89699,89774,89809,89833,89924,89955,90033,90086,90199,90204,90284,90340,90523,90706,90727,90796,90811,90901,90923,90974,91089,91169,91262,91351,91541,91589,91597,91630,91703,91759,91857,91972,92042,92125,92229,92245,92378,92458,92506,92528,92545,92563,92570,92602,92642,92646,92730,92777,92803,92814,92821,92889,93048,93107,93230,93379,93426,93473,93665,93745,93770,93775,93811,93859,93903,93950,93996,94268,94341,94447,94564,94582,94619,94637,94711,94770,94825,94862,94961,95119,95122,95160,95165,95184,95202,95223,95271,95358,95361,95584,95666,95702,95715,95816,95849,95861,95889,95914,95971,96176,96222,96264,96361,96551,96593,96596,96637,96654,96656,96676,96678,96887,96962,96979,97080,97203,97289,97392,97508,97725,97728,97863,97888,97895,97898,97902,98007,98008,98102,98204,98222,98311,98464,98467,98612,98669,98727,98788,98833,98887,98901,98920,98947,98948,98967,99006,99008,99047,99111,99136,99166,99193,99256,99322,99448,99478,99479,99523,99592,99620,99634,99650,99653,99659,99691,99720,99725,99798,99816,99873,99937",
          "147,175,330,384,414,416,434,440,472,473,817,867,875,1103,1232,1344,1363,1398,1477,1492,1630,1682,1686,2021,2054,2055,2109,2309,2334,2407,2408,2492,2571,2577,3025,3095,3193,3286,3294,3299,3335,3392,3451,3528,3537,3551,3695,3784,3790,3804,3930,4014,4067,4104,4124,4137,4241,4263,4414,4482,4561,4704,4742,4756,4848,4938,4992,4998,5015,5093,5193,5421,5748,6119,6253,6338,6441,6477,6559,6598,6906,7014,7045,7119,7245,7505,7573,7624,7639,7686,7691,7731,7807,7816,7853,8064,8212,8389,8477,8506,8551,8554,8561,8577,8699,8711,8812,8830,9067,9134,9215,9258,9400,9730,9733,9807,9940,10010,10016,10183,10266,10381,10518,10544,10597,10603,10711,10804,10864,10913,10923,10962,10971,11101,11151,11152,11323,11329,11384,11608,11630,11666,11694,11720,11825,11862,11958,11970,12017,12100,12189,12196,12416,12419,12427,12463,12483,12558,12782,12831,13031,13033,13336,13426,13450,13473,13558,13559,13564,13627,13661,13715,13843,13965,14130,14220,14254,14478,14479,14484,14574,14590,14626,14669,14683,14753,14813,14821,14965,15036,15043,15095,15097,15325,15410,15602,15764,15903,15929,15950,16020,16236,16630,16771,16776,16838,17092,17139,17255,17492,17494,17540,17733,17736,17847,18002,18009,18237,18384,18428,18446,18559,18576,18747,19121,19378,19441,19528,19556,19728,19735,19765,19834,19973,20032,20190,20430,20500,20503,20700,20735,20854,20997,21028,21052,21144,21178,21181,21236,21313,21605,21681,21693,21704,21708,21741,21743,21903,22083,22150,22291,22335,22392,22513,22522,22667,22855,22904,23065,23153,23154,23206,23215,23229,23495,23515,23647,24002,24003,24127,24231,24428,24594,24683,24717,24748,24811,25065,25093,25145,25224,25288,25459,25526,25589,25695,25716,25902,25931,26126,26288,26323,26333,26375,26460,26567,26627,26685,26718,27001,27059,27079,27485,27897,28268,28338,28435,28463,28475,28547,28626,28714,28821,28936,29021,29064,29086,29186,29231,29236,29456,29485,29490,29546,29722,29800,29965,30045,30053,30054,30099,30246,30294,30322,30334,30412,30472,30625,30710,30821,30834,30861,30980,31053,31185,31307,31449,31505,31581,31710,31963,32188,32257,32289,32314,32346,32399,32566,32603,32664,32681,32733,33020,33086,33095,33127,33152,33303,33327,33341,33445,33455,33480,33596,33638,33665,33809,33836,33857,34163,34263,34322,34679,34781,34816,34850,34878,34899,34964,35011,35043,35047,35080,35165,35259,35331,35394,35413,35589,35624,35731,35825,35971,36080,36235,36317,36352,36422,36500,36593,36595,36853,36933,36947,36954,37015,37047,37225,37262,37315,37539,37649,37758,37990,38097,38122,38419,38520,38598,38703,38725,38785,38843,38897,38926,38974,39288,39348,39352,39370,39448,39570,39585,39597,39623,39683,39714,39988,40210,40332,40388,40426,40491,40523,40596,40606,40683,40731,40740,40770,40911,40954,41037,41234,41266,41613,41722,41757,41895,42008,42010,42056,42266,42290,42492,42581,42593,42620,42632,42686,42761,42777,42804,42841,42902,42908,42916,43179,43521,43522,43633,43694,43819,43835,43866,43974,43986,44051,44138,44209,44260,44270,44301,44392,44424,44596,44600,44613,44663,44813,44834,44873,44925,45036,45122,45411,45415,45438,45480,45513,45515,45616,45625,45654,45659,45741,45854,46073,46135,46143,46295,46323,46326,46411,46477,46516,46576,46715,46802,46910,46953,46975,47031,47197,47437,47497,47535,47725,47738,47839,47892,47961,47988,48139,48231,48394,48678,48690,48722,48909,48945,48998,48999,49031,49077,49192,49230,49241,49247,49253,49286,49304,49382,49389,49395,49479,49487,49528,49530,49573,49706,49716,49789,49841,49871,49872,49928,49961,49999,50195,50230,50288,50310,50388,50405,50486,50550,50638,50655,50691,50754,50888,50944,50964,50999,51100,51141,51198,51313,51393,51406,51428,51659,51793,51931,51951,52047,52093,52161,52228,52239,52306,52453,52474,52540,52610,52691,52832,52839,52919,52945,53013,53203,53227,53344,53351,53390,53466,53473,53515,53596,53697,53845,53897,54056,54072,54316,54390,54406,54460,54670,54707,54759,54812,54998,55099,55108,55126,55179,55314,55330,55392,55728,55851,56082,56103,56104,56139,56268,56430,56445,56446,56501,56515,56892,57060,57088,57103,57107,57432,57463,57512,57571,57632,57783,57876,58034,58153,58287,58355,58374,58513,58630,58632,58805,58942,58986,59041,59118,59375,59376,59379,59436,59506,59596,59755,59760,59905,60093,60116,60184,60204,60210,60387,60565,60828,60867,60924,61249,61296,61367,61440,61483,61627,61660,61812,61869,61933,62386,62565,62670,62711,62750,62781,62973,62995,63009,63093,63201,63221,63890,63901,64096,64099,64166,64189,64288,64315,64336,64354,64357,64377,64416,64436,64447,64586,64638,64727,64910,64979,65011,65030,65209,65325,65339,65369,65468,65480,65620,65632,65644,65731,65733,65815,66157,66283,66384,66404,66447,66464,66473,66479,66546,66612,66626,66642,66808,67027,67116,67130,67212,67294,67478,67632,67736,67956,68016,68043,68075,68243,68256,68342,68370,68439,68476,68604,68642,68752,68767,69032,69034,69035,69135,69394,69410,69507,69575,69642,69659,69674,69679,69699,69814,69916,70008,70106,70114,70232,70304,70312,70544,70554,70582,70596,70601,70927,71027,71030,71169,71176,71197,71340,71361,71431,71594,71612,71832,71837,71901,71910,71962,72016,72030,72091,72259,72320,72398,72556,72591,72810,72915,73098,73138,73139,73189,73191,73223,73265,73346,73384,73437,73441,73486,73535,73632,73836,73872,73873,73931,73995,74041,74051,74176,74219,74256,74287,74308,74358,74465,74534,74547,74575,74641,74666,74717,74882,74913,74932,74991,75053,75064,75072,75089,75362,75470,75690,75726,75766,75773,75782,75852,75880,75966,76194,76297,76424,76443,76459,76514,76688,76812,76815,76821,77015,77044,77090,77116,77175,77382,77405,77461,77514,77624,77731,77733,77858,77888,77922,78029,78062,78122,78214,78325,78335,78353,78374,78444,78479,78585,78592,78656,78721,78722,78849,78937,78989,79088,79089,79197,79259,79278,79343,79473,79562,79782,79812,80014,80061,80107,80130,80218,80241,80388,80468,80477,80543,80616,80617,80677,80694,80709,80718,80735,80780,80826,80919,81216,81285,81358,81437,81595,81752,82120,82193,82278,82307,82334,82388,82441,82603,82610,82747,82812,82911,82982,82998,83030,83062,83086,83096,83109,83138,83163,83243,83321,83648,83865,83916,84086,84140,84163,84165,84218,84274,84275,84374,84423,84563,84572,84832,84853,85080,85160,85451,85522,85545,85572,85663,85672,85674,85855,85889,86031,86150,86340,86582,86751,86904,87018,87037,87050,87076,87100,87167,87239,87290,87305,87400,87406,87408,87438,87511,87653,87747,87769,87804,87902,87913,88004,88201,88336,88383,88434,88436,88553,88859,88968,89019,89105,89274,89310,89357,89554,89583,89601,89783,89788,89806,90041,90098,90113,90142,90314,90407,90467,90470,90478,90483,90582,90693,90793,90846,90904,91034,91088,91106,91237,91251,91298,91335,91400,91408,91514,91547,91606,91671,91749,91812,92036,92038,92043,92060,92067,92137,92244,92422,92462,92517,92679,92681,92882,93071,93075,93133,93151,93166,93174,93329,93459,93535,93545,93672,93798,93799,93823,93939,94039,94118,94128,94141,94204,94234,94294,94314,94414,94448,94560,94578,94646,94700,94786,94806,94899,95051,95078,95106,95153,95182,95214,95220,95375,95410,95571,95659,95670,95675,95842,95843,95869,95994,96367,96380,96452,96516,96668,96672,96725,96746,96790,96870,96953,96965,97092,97131,97305,97395,97511,97650,97705,97710,97719,97763,97764,97804,97864,98059,98100,98258,98265,98402,98425,98490,98502,98520,98603,98632,98665,98799,98921,98953,99265,99298,99344,99382,99432,99446,99459,99473,99692,99774,99827,99847,99894",
          "55,76,78,152,171,263,306,332,524,813,856,1402,1532,1578,1605,1741,1795,1801,1876,2050,2150,2183,2193,2234,2237,2405,2425,2444,2480,2507,2520,2604,2637,2737,2746,2797,2817,2974,3127,3205,3212,3246,3275,3283,3297,3350,3359,3401,3423,3470,3486,3577,3714,3751,3754,3777,3794,3809,3868,3872,4009,4113,4169,4189,4265,4328,4365,4411,4475,4513,4546,4688,4731,4777,4793,4802,4896,4980,5019,5064,5077,5129,5273,5331,5589,5704,5794,5872,6235,6518,6587,6625,6721,6852,6946,7028,7032,7244,7397,7420,7528,7557,7648,7659,7800,7973,7975,8035,8066,8192,8376,8377,8432,8471,8664,8689,8773,8909,8939,8954,9053,9075,9163,9239,9250,9568,9619,9662,9700,9796,9894,9958,10040,10108,10175,10330,10433,10539,10591,10593,10707,10720,10731,10886,10897,11289,11337,11370,11421,11567,11714,11852,11939,11942,11963,12060,12082,12173,12238,12285,12644,12679,12690,12726,12851,12889,12895,12917,13151,13158,13194,13232,13319,13369,13371,13373,13437,13458,13472,13621,13628,13660,13663,13747,13755,13892,13894,13912,14047,14052,14071,14186,14187,14276,14284,14306,14396,14424,14463,14639,14709,14727,14824,14928,15041,15047,15060,15211,15303,15377,15574,15590,15663,15699,15854,15885,15905,15947,15995,16046,16079,16120,16140,16282,16287,16383,16438,16446,16505,16551,16627,16642,16644,16666,16759,16943,16981,17104,17232,17344,17408,17527,17529,17547,17552,17555,17589,17592,17595,17660,17682,17686,17688,17726,17739,17792,17811,17865,17873,17885,17916,18061,18137,18154,18303,18325,18382,18383,18391,18507,18549,18730,18965,19005,19105,19127,19311,19331,19337,19398,19429,19448,19555,19580,19591,19643,19646,19665,19691,19743,19887,19900,19906,19979,20060,20138,20250,20335,20380,20424,20438,20494,20513,20546,20582,20594,20617,20623,20658,20748,20858,20886,20947,21007,21086,21157,21385,21414,21439,21454,21466,21538,21556,21565,21600,21782,21815,21926,21929,21933,21942,21953,21968,21989,22010,22178,22219,22316,22321,22327,22362,22436,22447,22515,22560,22570,22666,22704,22804,22868,22885,22890,22997,23040,23049,23160,23230,23376,23438,23439,23440,23448,23536,23565,23592,23593,23633,23651,23652,23719,23728,23746,23818,23854,23981,24017,24037,24093,24132,24187,24202,24219,24221,24338,24366,24381,24447,24535,24566,24576,24628,24653,24666,24698,24838,24867,24923,24927,24939,24991,25003,25049,25076,25116,25274,25283,25303,25331,25546,25625,25631,25708,25717,25743,25811,25825,25829,25846,25859,25868,25920,25924,25949,25988,25990,26064,26084,26137,26235,26238,26240,26262,26301,26324,26382,26426,26506,26563,26636,26689,26821,26923,26955,27029,27050,27083,27167,27203,27217,27258,27264,27326,27440,27471,27503,27529,27680,27720,27770,27799,27808,27960,27962,28031,28048,28070,28141,28145,28156,28198,28220,28258,28274,28362,28420,28473,28521,28558,28612,28617,28648,28691,28716,28756,28818,28824,28848,28883,28888,28973,29006,29105,29170,29194,29259,29275,29296,29379,29383,29419,29430,29495,29529,29576,29613,29635,29842,29875,29966,29981,30001,30046,30086,30106,30115,30118,30162,30180,30261,30264,30272,30317,30370,30371,30420,30454,30515,30539,30640,30871,30880,31033,31087,31109,31193,31486,31487,31509,31533,31543,31572,31738,31847,31855,31908,32077,32312,32361,32504,32589,32669,32939,33022,33025,33147,33247,33249,33257,33316,33367,33453,33559,33607,33620,33652,33696,33782,33947,33958,34108,34291,34366,34370,34407,34473,34687,34744,34812,34843,34852,34939,35055,35058,35081,35095,35121,35250,35326,35350,35352,35439,35467,35474,35487,35490,35515,35576,35675,35707,35729,35845,35903,36021,36070,36093,36134,36160,36267,36269,36314,36339,36404,36532,36535,36653,36697,36725,36734,36792,36873,36875,36919,36937,36948,36988,37007,37130,37213,37252,37253,37316,37331,37350,37376,37382,37386,37393,37407,37422,37528,37594,37605,37818,37822,37849,37867,37916,37960,37994,38108,38135,38144,38171,38175,38423,38426,38487,38632,38654,38736,38837,38888,38988,39010,39041,39159,39170,39202,39216,39235,39272,39276,39375,39416,39452,39458,39586,39652,39719,39725,39813,39829,39881,39908,39923,40037,40065,40067,40070,40087,40098,40131,40158,40159,40268,40325,40438,40440,40443,40481,40658,40659,40671,40685,40735,40864,40920,40944,40972,40992,40993,40994,41020,41048,41077,41094,41143,41157,41240,41257,41369,41474,41482,41560,41708,41714,41721,41753,41773,41826,41840,41844,41864,41869,41871,41874,42000,42004,42085,42140,42174,42279,42314,42383,42454,42551,42582,42700,42728,42764,42812,42819,42936,43016,43072,43088,43156,43190,43244,43318,43330,43360,43460,43495,43523,43539,43578,43611,43623,43670,43703,43821,43933,43956,43972,43999,44042,44059,44189,44195,44281,44385,44459,44479,44544,44580,44599,44689,44698,44774,44820,44863,44865,44882,44945,44964,45061,45112,45129,45145,45270,45419,45503,45574,45590,45598,45617,45691,45730,45769,45806,45835,45838,45860,45895,45962,46128,46130,46180,46186,46315,46457,46575,46672,46713,46723,46808,46857,47040,47300,47370,47387,47710,47859,48008,48010,48023,48138,48201,48215,48299,48339,48456,48462,48559,48693,48739,48891,48925,48926,49011,49049,49197,49297,49350,49391,49420,49518,49520,49588,49770,49974,50021,50099,50131,50607,50633,50715,50728,50772,50779,50874,50907,51259,51458,51554,51587,52023,52170,52211,52225,52289,52296,52495,52587,52675,52731,52785,52834,52838,52842,52870,52913,52935,52954,53001,53117,53133,53232,53250,53270,53283,53444,53505,53519,53553,53607,53625,53677,53994,53999,54246,54274,54377,54394,54493,54603,54630,54778,54791,54862,55052,55089,55111,55119,55161,55272,55282,55331,55339,55434,55527,55607,55637,55700,55737,55767,55931,56020,56276,56316,56457,56610,56647,56676,56679,56729,56764,56774,56812,56854,56900,56906,57035,57079,57238,57288,57361,57742,57807,57814,57909,57951,57989,57995,58002,58016,58084,58096,58106,58279,58311,58321,58342,58488,58548,58681,58803,58884,58895,58898,58932,59015,59202,59257,59386,59391,59444,59546,59642,59742,59753,59827,60049,60054,60171,60203,60254,60333,60492,60561,60841,60955,61025,61088,61094,61206,61436,61521,61601,61702,61749,62032,62054,62122,62270,62365,62512,62546,62547,62585,62709,62945,62954,63015,63055,63075,63098,63122,63171,63308,63331,63340,63348,63352,63445,63458,63632,63659,63712,63886,63981,64122,64137,64148,64151,64294,64300,64302,64321,64333,64334,64346,64558,64703,64735,64789,64799,64809,64866,65137,65185,65215,65350,65362,65678,65763,65833,65886,65901,65994,66040,66078,66361,66420,66673,66702,66774,66795,66834,66897,66928,66943,66957,67034,67042,67047,67265,67299,67354,67358,67403,67411,67442,67574,67579,67646,67686,67707,67726,67744,67855,67910,68321,68371,68398,68405,68612,68776,68792,68914,68922,69137,69144,69158,69180,69184,69298,69387,69489,69503,69546,69608,69758,69875,69879,69892,69934,69985,70118,70183,70224,70248,70287,70344,70362,70413,70455,70551,70576,70678,70804,70813,70979,71011,71081,71114,71143,71161,71173,71246,71324,71399,71626,71668,71707,71873,71927,72025,72043,72064,72135,72214,72296,72395,72582,72663,72713,72745,72846,72897,72959,73060,73086,73104,73234,73447,73733,73763,73801,73893,73906,73988,74037,74073,74311,74396,74408,74441,74528,74724,74726,74762,75054,75116,75131,75230,75243,75306,75525,75642,75738,75775,75809,75868,76176,76237,76247,76268,76311,76365,76386,76454,76461,76484,76495,76565,76615,76696,76791,76877,76902,76933,77137,77155,77290,77312,77340,77363,77364,77366,77458,77462,77701,77803,77868,77943,77986,78144,78392,78568,78586,78606,78627,78642,78755,78765,79043,79093,79102,79138,79186,79276,79387,79407,79418,79434,79515,79617,79619,79864,80042,80052,80103,80113,80276,80288,80294,80403,80428,80470,80471,80529,80533,80692,80810,81181,81379,81447,81505,81562,81764,81799,81827,81832,81900,81939,81966,81980,82019,82417,82449,82489,82513,82524,82538,82581,82653,82693,82711,83051,83108,83268,83408,83532,83700,83723,83755,83893,83908,83985,83996,84085,84257,84339,84354,84552,84638,84865,85023,85045,85175,85314,85633,85745,85759,85836,85844,86001,86003,86019,86038,86128,86148,86156,86219,86270,86298,86462,86464,86517,86599,86690,86757,86765,86784,86813,87014,87159,87162,87228,87254,87263,87437,87455,87473,87482,87546,87628,87865,87885,87895,87898,87904,87910,88105,88155,88241,88266,88270,88310,88367,88429,88440,88530,88819,88939,89166,89173,89420,89446,89481,89492,89547,89558,89638,89650,89652,89694,89958,90116,90214,90220,90332,90334,90335,90469,90614,90648,90726,90870,90936,90946,90976,91027,91058,91078,91185,91186,91257,91302,91326,91341,91348,91360,91469,91495,91509,91568,91653,91670,91850,91941,91954,92114,92209,92210,92220,92325,92412,92486,92614,92666,92688,92744,92840,92862,92921,92929,92976,93217,93244,93255,93291,93462,93469,93532,93566,93594,93600,93678,93708,93737,93817,93828,93830,93858,93956,93993,94031,94171,94229,94242,94291,94312,94320,94430,94474,94528,94640,94685,94767,94913,94919,95259,95456,95662,95707,95737,95776,95780,95808,95873,95991,96001,96103,96113,96122,96154,96155,96210,96301,96395,96421,96442,96498,96539,96548,96689,96728,96768,96890,96969,97064,97173,97224,97248,97265,97301,97401,97510,97567,97692,97697,97905,97919,97936,97975,97987,97997,98063,98138,98146,98184,98254,98314,98351,98497,98688,98736,98746,98827,98839,98918,99056,99061,99145,99146,99233,99251,99316,99360,99372,99572,99618,99655,99671,99761,99842,99884,99980",
          "218,277,285,538,581,596,606,624,643,691,737,807,845,893,1290,1495,1537,1542,1616,1632,1649,1760,1861,1904,1912,1960,1985,2026,2103,2187,2341,2364,2420,2437,2466,2514,2516,2569,2589,2643,2688,2749,2805,2815,2983,2989,3015,3021,3068,3072,3085,3147,3234,3266,3428,3460,3466,3467,3510,3554,3612,3684,3744,3849,3893,3894,3908,3925,3963,3964,4236,4276,4404,4429,4444,4517,4524,4548,4626,4661,4691,4733,4765,4813,4853,4871,4900,4903,4907,5467,5626,5638,5648,5656,5684,5739,5769,5783,5811,6039,6046,6053,6082,6145,6220,6322,6468,6597,6693,6756,6814,6937,7183,7289,7312,7501,7943,7944,7984,8029,8041,8111,8120,8145,8150,8408,8442,8483,8708,8944,9051,9123,9236,9313,9391,9504,9545,9564,9606,9638,9779,9866,10051,10147,10221,10274,10275,10307,10608,10623,10626,10659,10705,10756,10800,10822,10842,10943,10961,10993,11076,11187,11597,11668,11758,11917,11972,12104,12231,12320,12384,12626,12645,12697,12810,12825,12961,13069,13071,13184,13284,13441,13462,13604,13608,13616,13737,13762,13942,13947,14006,14051,14099,14157,14180,14252,14360,14480,14493,14578,14595,14894,15037,15090,15106,15137,15173,15174,15324,15367,15411,15518,15622,15660,15690,15752,15822,15860,15877,16045,16066,16101,16204,16207,16269,16371,16403,16497,16537,16694,16731,16744,16763,16782,16790,16858,16888,16918,16949,16969,17019,17031,17052,17067,17081,17131,17138,17206,17211,17373,17424,17482,17483,17616,17696,17722,17742,17814,17859,17879,17905,18041,18062,18270,18363,18381,18519,18621,18625,18628,18722,18727,18930,18988,19181,19216,19251,19264,19284,19288,19309,19459,19652,19757,19775,19779,19827,19893,19924,19935,20041,20166,20314,20328,20334,20355,20560,20585,20625,20649,20660,20718,20737,20779,20801,20818,20857,20887,20948,21045,21102,21112,21123,21243,21296,21303,21310,21319,21398,21401,21445,21477,21515,21532,21536,21548,21628,21670,21773,21790,21811,21904,21912,21980,22018,22243,22347,22348,22360,22388,22435,22446,22472,22480,22514,22545,22698,22727,22823,22889,22954,22961,22975,22992,23011,23034,23087,23133,23168,23171,23213,23284,23304,23316,23373,23410,23582,23610,23726,23867,23926,23946,23977,23989,24028,24112,24205,24232,24271,24483,24738,24817,24873,24891,24920,24924,25152,25179,25238,25263,25278,25322,25400,25504,25622,25745,25763,25875,25884,26018,26066,26074,26132,26199,26344,26345,26350,26404,26409,26445,26475,26481,26564,26585,26652,26679,26742,26748,26857,26915,27049,27105,27196,27247,27353,27443,27535,27651,27653,27730,27863,27908,27989,28090,28128,28137,28203,28400,28408,28423,28430,28491,28495,28496,28526,28563,28598,28711,28760,28838,28896,29015,29032,29068,29265,29282,29284,29310,29467,29496,29520,29586,29671,29679,29697,29816,29923,29960,29968,30096,30305,30366,30378,30397,30478,30598,30618,30661,30691,30729,30863,30931,31011,31072,31152,31153,31194,31206,31230,31271,31282,31327,31453,31457,31485,31555,31585,31591,31666,31678,31694,31712,31734,31741,31790,31809,31812,31859,31927,31954,32029,32095,32183,32237,32326,32329,32410,32422,32460,32461,32476,32490,32503,32514,32633,32659,32690,32730,32736,32739,32774,33016,33029,33042,33057,33145,33159,33174,33177,33198,33425,33428,33486,33491,33494,33504,33542,33566,33578,33681,33709,33844,33862,33951,33980,34018,34035,34106,34107,34113,34129,34134,34155,34193,34213,34282,34292,34315,34338,34431,34535,34558,34733,34753,34768,34822,34826,34832,34863,34864,34941,34990,34996,35014,35031,35045,35053,35169,35191,35280,35283,35303,35333,35361,35366,35371,35436,35445,35503,35604,35671,35696,35749,35804,35843,35850,35947,35978,35979,35987,36108,36216,36219,36260,36261,36298,36372,36387,36451,36546,36568,36571,36647,36708,36807,36821,36951,37039,37134,37193,37215,37219,37374,37377,37403,37410,37455,37514,37515,37710,37754,37759,37762,37885,37906,37933,37951,37955,37974,38003,38064,38085,38211,38317,38336,38352,38364,38376,38471,38476,38510,38561,38577,38626,38638,38723,38753,38803,38821,38889,38901,38973,39043,39051,39168,39194,39267,39316,39335,39365,39374,39462,39480,39488,39554,39626,39660,39662,39907,39932,39954,39977,40012,40050,40092,40101,40110,40120,40190,40201,40267,40272,40334,40335,40482,40501,40550,40602,40603,40627,40669,40670,40717,40738,40866,40986,41069,41093,41097,41197,41274,41330,41345,41420,41437,41456,41460,41640,41699,41709,41823,41849,41903,41909,41970,42027,42044,42117,42131,42132,42141,42196,42336,42388,42395,42452,42459,42545,42575,42600,42604,42612,42710,42731,42769,43140,43162,43198,43312,43331,43448,43513,43685,43707,43813,43822,43840,43910,43958,43998,44009,44176,44193,44210,44258,44273,44336,44348,44405,44566,44628,44634,44768,44784,44796,44805,44824,44825,44843,44982,45097,45229,45261,45307,45330,45374,45392,45410,45511,45567,45573,45581,45618,45676,45696,45708,45719,45732,45900,45971,45980,46183,46341,46344,46345,46395,46436,46531,46773,46917,47147,47148,47244,47245,47246,47427,47434,47458,47492,47517,47641,47682,47848,47871,47873,47890,48190,48265,48284,48405,48432,48470,48500,48503,48681,48694,48711,48736,48812,48838,48920,49056,49100,49103,49185,49231,49589,49600,49802,49814,49890,49931,50079,50091,50110,50207,50339,50412,50427,50499,50640,50775,50799,50834,50905,51046,51071,51172,51210,51246,51322,51364,51414,51434,51481,51577,51588,51650,51662,51678,51771,51775,51834,51836,51933,52026,52066,52198,52232,52233,52357,52386,52479,52577,52755,52778,52922,52938,52961,52977,52988,53003,53107,53215,53251,53255,53333,53561,53664,53693,53701,53923,54037,54157,54414,54526,54580,54622,54722,54756,54985,55011,55070,55138,55210,55343,55403,55511,55573,55749,55825,55873,55966,55987,56190,56312,56381,56392,56461,56508,56516,56522,56599,56663,56834,56879,57093,57170,57186,57224,57426,57448,57696,57715,57894,58089,58160,58219,58225,58327,58347,58493,58694,58704,58726,58731,58802,58823,58857,58858,59263,59383,59450,59482,59652,59959,59979,60037,60076,60115,60150,60310,60414,60462,60522,60547,60569,60594,60673,60705,61185,61457,61477,61485,61545,61588,61659,61834,61872,61886,61902,61908,61924,61947,61951,62029,62209,62268,62362,62444,62544,62548,62608,62701,62836,62870,62904,62992,63066,63082,63108,63414,63535,63605,63661,63774,63775,63812,63830,63847,63927,63993,64049,64165,64228,64559,64604,64626,64681,64699,64748,64876,64964,64982,65019,65089,65219,65278,65370,65471,65559,65814,65860,66192,66383,66429,66578,66588,66722,66727,66761,66815,66838,66886,66908,66977,67108,67233,67330,67372,67374,67474,67475,67609,67629,67729,67741,67858,67940,68174,68322,68374,68380,68386,68598,68670,68813,68855,68976,69110,69305,69308,69314,69353,69369,69515,69669,69737,69923,70137,70163,70194,70218,70322,70385,70441,70539,70602,70617,70668,70726,70735,70773,71066,71078,71186,71330,71444,71513,71571,71645,71656,71742,71770,71801,71809,71853,71935,71989,72161,72213,72221,72265,72559,72720,72814,72832,72847,72881,72923,72926,73013,73033,73137,73163,73310,73388,73484,73514,73544,73571,73573,73649,73676,73788,74070,74130,74154,74294,74320,74333,74334,74394,74565,74705,74806,74865,74899,74916,74981,75020,75047,75104,75144,75221,75308,75320,75602,75660,75732,75781,75871,76006,76076,76098,76108,76276,76364,76389,76397,76496,76733,76749,76816,76833,77025,77124,77146,77201,77283,77369,77377,77517,77537,77671,77730,77790,77847,77861,77938,77973,78061,78100,78244,78281,78431,78508,78551,78610,78649,78718,78736,78772,79036,79153,79247,79350,79408,79461,79470,79612,79701,79765,79824,79853,80007,80056,80117,80163,80324,80346,80435,80514,80539,80634,80732,80771,80778,80800,80801,80847,80883,80992,81026,81110,81160,81192,81207,81252,81386,81394,81472,81474,81570,81759,81783,81945,82005,82095,82123,82130,82134,82164,82283,82401,82575,82579,82633,82813,82873,83057,83168,83286,83383,83419,83647,83749,83759,83774,83815,83842,83880,83902,83960,83967,84023,84240,84293,84375,84489,84724,84910,84930,85186,85286,85355,85445,85621,85763,85927,85933,86056,86075,86178,86204,86304,86470,86601,86714,86971,87155,87274,87340,87424,87449,87490,87537,87556,87698,87756,87780,87935,87993,88020,88073,88104,88337,88360,88365,88422,88459,88482,88507,88639,88684,88970,89009,89022,89061,89075,89144,89168,89254,89287,89295,89296,89403,89444,89508,89606,89656,89717,89776,89819,89827,89873,89890,89944,89962,90026,90063,90169,90206,90212,90249,90550,90739,90958,90965,91132,91171,91188,91211,91225,91227,91266,91321,91353,91494,91542,91582,91755,91861,92101,92102,92208,92242,92295,92331,92375,92379,92455,92601,92633,92641,92670,92676,92699,92720,92763,92793,92863,92945,92967,93027,93058,93129,93140,93183,93316,93319,93354,93436,93467,93585,93778,93926,93940,94130,94150,94252,94278,94283,94347,94374,94429,94478,94487,94516,94630,94814,94841,94883,95016,95187,95314,95335,95454,95540,95549,95728,95813,95894,95967,96019,96151,96187,96352,96476,96487,96502,96535,96566,96664,96707,96724,96929,96954,97011,97034,97129,97148,97189,97254,97361,97555,97832,97904,97928,97973,98078,98202,98495,98504,98514,98578,98619,98662,98681,98726,98785,99022,99163,99272,99325,99378,99487,99503,99639,99799,99858,99919,99953",
          "71,86,172,226,272,457,559,668,871,939,1007,1043,1116,1308,1360,1399,1463,1550,1556,1711,1812,1891,1899,2073,2119,2129,2194,2321,2322,2343,2452,2572,2595,2602,2745,2806,2850,2919,2961,2967,2980,3041,3069,3100,3123,3163,3203,3296,3323,3398,3431,3452,3455,3462,3603,3615,3631,3644,3917,4053,4240,4539,4627,4722,4747,4755,4962,5033,5180,5210,5226,5404,5444,5463,5520,5640,5655,5658,5717,5718,5735,5757,5776,5795,5816,6061,6091,6138,6170,6181,6194,6282,6397,6403,6761,6831,6924,6966,7103,7169,7311,7317,7379,7391,7457,7483,7597,7733,7959,7964,8012,8023,8052,8063,8083,8127,8209,8214,8302,8384,8454,8656,8662,8670,8797,8899,8970,8978,9044,9061,9065,9079,9105,9113,9374,9938,9960,10020,10036,10092,10106,10142,10201,10248,10310,10331,10357,10401,10418,10499,10611,10641,10740,10765,10965,11079,11083,11190,11271,11293,11327,11354,11355,11458,11467,11505,11782,11811,11876,11890,11984,11997,12031,12112,12170,12183,12200,12304,12451,12460,12575,12668,12937,12951,13075,13089,13096,13329,13335,13399,13418,13446,13601,13641,13675,13721,13828,13873,13932,14037,14082,14121,14426,14628,14701,14740,14743,14784,14837,14900,14955,15070,15156,15165,15308,15374,15403,15493,15540,15797,15839,15841,15842,15855,15863,15865,15878,16024,16026,16040,16047,16065,16176,16230,16244,16388,16433,16434,16650,16724,16726,16802,16844,16945,17000,17086,17087,17155,17217,17239,17296,17337,17350,17645,17653,17700,17890,17944,18010,18200,18211,18222,18275,18304,18440,18575,18637,18645,18652,18667,19103,19386,19439,19501,19553,19664,19687,19713,20161,20349,20460,20519,20609,20924,20930,20935,21005,21022,21042,21088,21121,21122,21124,21267,21318,21408,21418,21487,21540,21590,21807,21818,21906,21940,21952,22008,22017,22202,22263,22506,22519,22532,22582,22691,22710,22764,22775,22834,22945,23059,23067,23121,23126,23251,23268,23282,23354,23445,23585,23600,23763,23774,23785,23916,23920,23921,24161,24298,24340,24353,24357,24358,24412,24609,24647,24794,24901,25032,25039,25080,25124,25250,25285,25374,25614,25776,25796,25830,25905,26031,26037,26056,26079,26139,26186,26282,26362,26378,26544,26549,26566,26674,26746,26747,26778,26827,26840,26981,26992,27005,27021,27032,27064,27117,27166,27169,27191,27232,27317,27331,27427,27487,27511,27595,27636,27756,27805,27878,27965,28084,28103,28112,28114,28206,28270,28456,28468,28614,28631,28759,28772,28790,28805,28812,29106,29499,29595,29676,29730,29735,29791,29815,30068,30097,30158,30197,30203,30241,30424,30505,30531,30606,30627,30636,30651,30852,30965,30982,31073,31078,31117,31221,31234,31243,31266,31322,31334,31347,31354,31376,31635,31693,31768,31792,31940,32016,32048,32090,32143,32163,32217,32334,32344,32421,32436,32612,32767,32906,32953,33113,33132,33134,33135,33226,33259,33270,33313,33347,33363,33484,33513,33515,33526,33529,33963,33992,34001,34149,34181,34187,34206,34241,34284,34295,34306,34357,34394,34435,34528,34532,34591,34696,34861,34987,35140,35163,35212,35227,35462,35525,35527,35530,35628,35710,35747,36178,36217,36292,36306,36360,36461,36530,36644,36719,36797,36802,36879,36958,36962,37019,37058,37151,37180,37290,37372,37421,37435,37440,37460,37550,37556,37845,38166,38396,38596,38630,38698,38730,38787,38791,38927,39259,39398,39403,39409,39484,39517,39519,39529,39530,39538,39560,39620,39631,39677,39690,39712,39759,39872,39909,39958,40010,40013,40022,40056,40082,40280,40379,40463,40484,40561,40734,40764,40803,41023,41108,41155,41286,41297,41298,41340,41503,41549,41552,41643,41837,41855,41888,41942,42151,42169,42204,42230,42450,42457,42586,42837,42850,43009,43025,43029,43037,43052,43135,43163,43183,43203,43211,43237,43310,43320,43329,43373,43469,43530,43574,43717,43793,43920,43971,44082,44168,44182,44266,44326,44510,44615,44742,44748,44833,44959,45017,45170,45230,45335,45486,45537,45597,45663,45747,45792,45833,45897,45977,45990,46134,46210,46270,46437,46645,46646,46688,46706,46725,46956,47077,47106,47159,47218,47431,47518,47633,47650,47716,47718,47765,47770,47844,47963,48003,48030,48039,48078,48111,48168,48256,48581,48627,48647,48717,48905,48994,49048,49211,49239,49328,49537,49564,49602,49707,49832,49844,49945,50119,50253,50373,50410,50435,50560,50593,50635,50723,50750,50770,50789,51095,51143,51200,51213,51332,51367,51438,51514,51553,51634,51753,51860,51911,51917,52033,52059,52177,52238,52400,52434,52525,52648,52668,52674,52697,52820,53148,53218,53341,53436,53484,53637,53774,53840,53843,53910,53946,54019,54025,54087,54162,54257,54287,54369,54478,54489,54545,54799,54987,55158,55276,55286,55387,55462,55684,55967,55998,56211,56239,56250,56323,56510,56524,56581,56873,57078,57262,57398,57473,57544,57615,57730,57819,58043,58164,58362,58449,58529,58679,58723,58741,58762,58979,59036,59075,59122,59479,59545,59814,59942,60418,60435,60598,60638,60639,60707,60708,60788,60888,60976,61096,61135,61145,61154,61583,61762,61891,61899,61941,61988,62026,62227,62312,62413,62418,62443,62468,62569,62591,62849,62979,63039,63143,63258,63316,63473,63617,63732,63753,63921,63952,64004,64021,64026,64118,64143,64162,64337,64609,64785,64843,64875,65002,65005,65116,65305,65319,65384,65751,65755,65841,65889,65921,65929,66030,66086,66219,66278,66345,66474,66614,66750,66874,66920,67026,67043,67168,67331,67392,67861,68038,68054,68141,68222,68252,68263,68320,68654,68781,68846,68910,69014,69082,69096,69275,69380,69401,69495,69599,69727,69830,69866,70013,70037,70176,70189,70236,70300,70440,70482,70501,70560,70572,70600,70663,70800,70876,71097,71106,71181,71234,71414,71476,71481,71502,71559,71637,71675,71708,71833,71916,71925,72087,72118,72303,72345,72365,72381,72387,72413,72439,72494,72526,72666,72690,72850,72866,72970,72974,73228,73318,73459,73510,73520,73700,73840,73877,74003,74017,74067,74104,74195,74297,74298,74423,74445,74602,74627,74657,74683,74758,74761,74808,74893,74896,75069,75211,75259,75330,75331,75462,75534,75674,75680,75722,75787,75810,75912,76064,76260,76399,76510,76673,76719,76743,76750,76853,76864,77072,77214,77273,77334,77336,77345,77391,77410,77583,77638,77656,77664,77673,78076,78298,78343,78493,78522,78565,78693,78839,78882,78933,79215,79263,79381,79409,79456,79474,79551,79605,79946,80000,80058,80183,80298,80446,81056,81117,81187,81201,81292,81332,81397,81415,81504,81604,81713,81822,81825,81844,81899,82268,82359,82431,82437,82480,82565,82625,82735,82769,82891,83112,83205,83231,83311,83335,83351,83429,83451,83513,83550,83568,83667,83760,83852,83873,83911,84074,84090,84148,84209,84247,84362,84404,84405,84573,84575,84698,84717,84813,84818,84986,85231,85339,85370,85492,85653,85687,85779,85879,86002,86012,86051,86052,86107,86129,86181,86217,86283,86643,86799,86922,86956,87187,87286,87397,87423,87466,87685,87832,87939,87976,87994,88057,88069,88136,88280,88400,88417,88437,88438,88510,88511,88709,88712,88814,88827,88902,88913,88943,88978,89380,89424,89459,89494,89931,89959,89966,90147,90286,90374,90411,90498,90549,90632,90728,90873,91107,91149,91191,91356,91399,91502,91522,91540,91610,91683,91718,91821,91877,91937,91952,92028,92057,92076,92147,92219,92223,92285,92367,92431,92490,92588,92624,93022,93050,93137,93149,93235,93280,93360,93415,93455,93571,93645,93754,93782,93813,93877,93893,93986,94010,94271,94325,94454,94713,94775,94859,94906,94990,95025,95085,95116,95458,95526,95631,95638,95727,95804,95865,96025,96093,96172,96237,96287,96310,96426,96540,96587,96609,96645,96765,96816,96821,96963,96978,97200,97379,97458,97460,97627,97717,97793,97840,98177,98247,98323,98355,98363,98626,98697,98734,98816,98844,98870,98935,99203,99292,99363,99392,99420,99548,99605,99651,99667,99904,99951",
          "126,165,198,521,527,619,901,1112,1115,1128,1168,1306,1310,1661,1665,1671,1811,1817,1825,1850,1852,1931,1968,2086,2141,2213,2270,2347,2411,2460,2582,2605,2621,2696,2715,2772,2798,2878,2885,2918,2922,3131,3341,3448,3496,3504,3527,3582,3614,3698,3711,3752,3762,3763,3808,3832,3891,3920,3939,3943,3945,3947,3957,4076,4175,4206,4227,4302,4316,4426,4526,4595,4658,4679,4797,4843,5016,5037,5041,5256,5710,5721,5959,6148,6198,6289,6327,6443,6467,6545,6561,6591,6600,6622,6666,6813,6854,6952,7147,7164,7285,7307,7325,7346,7396,7448,7468,7554,7603,7627,7629,7662,7682,7684,7725,7760,7789,8004,8005,8096,8128,8249,8296,8320,8464,8469,8484,8517,8527,8588,8595,8652,8735,8755,8759,8763,8893,8943,8983,9040,9049,9090,9097,9115,9220,9254,9256,9271,9413,9559,9701,9773,9784,9882,9898,10017,10191,10230,10238,10305,10427,10434,10727,10732,10834,10953,11004,11040,11045,11078,11107,11149,11160,11292,11330,11524,11648,11685,11726,11765,11838,11874,12113,12162,12178,12182,12222,12291,12430,12458,12509,12613,12661,12695,12744,12931,13028,13047,13091,13101,13139,13169,13179,13208,13259,13310,13366,13392,13428,13474,13544,13640,13810,13984,14010,14054,14095,14177,14256,14264,14529,14579,14761,14800,14817,14861,14912,14994,15093,15240,15320,15352,15362,15369,15416,15580,15620,15643,15704,15808,15886,15898,15978,16048,16062,16179,16376,16449,16493,16571,16602,16661,16705,16719,16747,16772,16991,17135,17150,17151,17231,17238,17315,17371,17455,17566,17567,17693,17728,17765,17839,17895,17946,18007,18069,18077,18127,18148,18177,18184,18186,18210,18272,18283,18366,18369,18485,18506,18651,18803,18809,18926,18953,19196,19197,19200,19293,19346,19414,19517,19552,19612,19630,19835,19885,19985,20051,20168,20174,20178,20218,20239,20240,20260,20401,20429,20510,20576,20650,20800,20875,20916,20921,20923,21048,21053,21060,21129,21134,21180,21204,21241,21387,21423,21437,21668,21723,21849,21978,21988,22007,22044,22064,22097,22155,22292,22396,22474,22574,22658,22677,22702,22736,22837,22921,22957,23010,23242,23279,23303,23328,23347,23475,23561,23564,23629,23635,23709,23742,23786,23878,23933,23987,24059,24095,24128,24343,24517,24692,24709,24731,24888,24911,24931,25011,25094,25112,25154,25201,25233,25256,25261,25339,25540,25580,25607,25609,25637,25787,25802,25870,25891,25958,25985,25999,26000,26216,26229,26295,26296,26336,26440,26795,26841,26947,26953,26959,27113,27156,27215,27262,27356,27491,27597,27645,27754,27828,27837,27843,27845,28014,28038,28281,28344,28390,28419,28555,28561,28576,28646,28683,28859,28921,28963,29029,29122,29154,29177,29252,29352,29603,29686,29706,29768,29788,29992,30105,30297,30357,30427,30460,30610,30612,30616,30798,30806,30869,30925,30952,30991,31099,31121,31406,31444,31520,31681,31690,31730,31782,32051,32087,32193,32224,32267,32322,32391,32439,32662,32680,32683,32693,32716,32761,32828,32844,32878,32902,32930,32973,33161,33228,33252,33312,33481,33519,33676,33699,33771,33786,33801,33825,33926,33975,34029,34151,34161,34224,34246,34298,34345,34361,34412,34420,34485,34729,34739,34922,34934,35004,35137,35166,35255,35329,35367,35480,35509,35516,35540,35546,35572,35587,35626,35684,35758,35899,35904,36229,36240,36271,36286,36396,36401,36428,36475,36738,36885,36898,36931,37093,37122,37289,37342,37471,37610,37826,37858,37907,37945,38176,38228,38377,38449,38503,38535,38614,38641,38911,39023,39143,39178,39182,39233,39355,39467,39754,39903,40214,40251,40260,40319,40330,40342,40425,40666,40694,40778,40788,40824,41047,41060,41147,41497,41585,41663,41769,41860,41896,41939,42024,42034,42167,42210,42214,42220,42282,42304,42403,42417,42449,42588,42744,42806,42910,43013,43181,43247,43263,43270,43279,43393,43430,43453,43464,43486,43567,43594,43692,43704,43752,43760,43773,43795,43967,44003,44037,44208,44234,44280,44331,44345,44522,44546,44552,44565,44590,44641,44703,44874,44926,44991,45302,45323,45325,45373,45477,45525,45603,45629,45630,45761,45804,45811,46051,46060,46151,46176,46190,46271,46483,46578,46618,46663,46680,46839,46977,47053,47203,47292,47320,47392,47414,47438,47652,47675,47993,48006,48025,48086,48113,48183,48269,48335,48376,48391,48399,48406,48488,48594,48621,48656,48820,48898,48924,49038,49042,49086,49098,49115,49202,49249,49331,49387,49542,49634,49791,49951,50004,50137,50167,50215,50252,50260,50344,50506,50522,50619,50717,50739,50867,50892,50957,51001,51031,51234,51267,51339,51342,51495,51680,51783,51826,51841,51967,51968,52079,52102,52208,52210,52270,52291,52551,52608,52609,52685,52782,52900,53141,53179,53267,53290,53512,53737,53765,53771,53773,53809,53891,53945,54026,54039,54092,54124,54142,54194,54363,54598,54627,54642,54656,54662,54703,54720,54729,54746,54764,55019,55061,55154,55252,55398,55406,55650,55815,55870,55924,55934,56114,56159,56163,56204,56252,56465,56555,56640,56645,56671,56715,57010,57068,57158,57164,57181,57249,57449,57460,57562,57633,57653,57674,57769,57771,57982,58031,58085,58117,58127,58130,58173,58195,58281,58322,58346,58403,58435,58504,58600,58653,58664,58829,58885,59018,59047,59061,59069,59209,59248,59414,59493,59536,59582,59645,59695,59737,59750,59783,59862,59967,59977,60011,60034,60043,60088,60123,60264,60298,60521,60600,60743,60807,61021,61057,61101,61161,61313,61639,61650,61666,61714,61746,61837,61920,61935,62059,62143,62242,62328,62340,62502,62521,62622,62654,62702,62713,62828,62983,62987,63144,63170,63262,63263,63497,63523,63563,63583,63663,64059,64120,64157,64174,64270,64289,64450,64459,64474,64493,64531,64553,64651,64833,65058,65081,65130,65236,65303,65383,65455,65473,65605,65608,65654,65742,65911,65972,66038,66120,66212,66213,66226,66240,66402,66478,66564,66643,66676,66811,66951,67005,67065,67079,67107,67232,67295,67297,67312,67324,67335,67373,67491,67508,67585,67620,67648,67656,67717,67755,68012,68103,68294,68306,68365,68570,68667,68675,68696,68743,68969,68979,69010,69044,69102,69106,69118,69185,69227,69260,69364,69419,69456,69491,69543,69687,69710,69802,69871,69967,70040,70117,70151,70341,70352,70451,70599,70674,70812,70828,70843,70856,70939,71056,71060,71115,71189,71213,71236,71240,71258,71320,71362,71443,71497,71545,71775,71792,71847,71865,72060,72124,72319,72880,73209,73230,73235,73350,73379,73624,73719,73809,73870,74213,74240,74282,74318,74398,74406,74523,74665,74669,74684,74727,74801,74816,74870,74886,74892,74894,74933,74978,75011,75077,75082,75099,75136,75421,75476,75488,75795,75883,75939,76091,76373,76409,76445,76551,76677,76678,76755,76797,76834,76861,76932,76985,77104,77148,77184,77197,77287,77455,77457,77481,77486,77533,77534,77549,77625,77688,77699,77753,77968,78004,78126,78136,78172,78176,78225,78246,78253,78259,78318,78464,78500,78547,78635,78670,78702,78707,78795,78858,78872,78965,78967,78999,79040,79096,79129,79266,79275,79292,79302,79401,79441,79593,79643,79739,79911,79966,80030,80032,80165,80208,80214,80421,80552,80581,80584,80666,80682,80786,80955,81065,81066,81280,81303,81370,81511,81521,81544,81545,81599,81605,81624,81638,81703,81893,81931,81965,82111,82119,82150,82238,82282,82390,82395,82430,82522,82629,82638,82822,82894,83040,83064,83184,83223,83323,83344,83355,83520,83606,83615,83635,83724,83731,83836,83875,83936,83943,84000,84050,84136,84183,84214,84226,84242,84250,84285,84320,84620,84621,84645,84685,84690,84713,84908,84941,85017,85114,85470,85542,85636,85652,85744,85897,85923,85987,86035,86170,86192,86232,86424,86603,86652,86740,86774,86788,86815,86970,86988,87030,87080,87209,87258,87421,87444,87512,87609,87645,87739,87815,87853,88076,88150,88175,88176,88190,88203,88222,88388,88545,88687,88708,88802,88994,89036,89133,89218,89486,89529,89622,89667,89677,89795,89858,90009,90056,90114,90261,90271,90276,90325,90364,90403,90492,90642,90759,90765,90794,91023,91070,91209,91303,91325,91406,91601,91690,91716,91724,91855,91899,92010,92290,92320,92382,92408,92439,92541,92549,92550,92671,92756,92800,92873,92892,93011,93145,93370,93640,93662,93681,93765,93769,93854,93921,93963,93985,94077,94169,94289,94326,94342,94424,94451,94532,94544,94613,94663,94673,94749,94756,94783,94799,94857,94960,95005,95088,95110,95113,95115,95232,95291,95393,95415,95431,95476,95520,95711,95845,95857,95888,95900,96007,96050,96108,96329,96341,96496,96512,96581,96597,96660,96877,96975,97354,97407,97438,97465,97469,97552,97862,97894,98084,98366,98398,98472,98493,98552,98864,98966,99033,99101,99113,99222,99231,99295,99324,99461,99485,99496,99553,99555,99594,99626,99706,99921,99936,99938",
          "3,33,170,194,291,375,639,656,752,765,814,829,900,923,925,947,1013,1243,1268,1539,1555,1575,1678,1768,1816,1840,1842,2222,2227,2231,2235,2636,2658,2664,2685,2719,2869,2873,2929,3020,3037,3199,3284,3338,3562,3681,3712,3735,3747,3802,3837,3972,3977,3993,4077,4198,4212,4275,4286,4385,4503,4540,4569,4624,4748,4905,5111,5187,5202,5214,5337,5372,5419,5653,5726,5788,5866,5881,6014,6116,6333,6340,6422,6527,6642,6718,6916,6939,6984,6996,7044,7266,7336,7410,7431,7445,7455,7755,7768,7822,7925,7941,8007,8016,8084,8253,8263,8323,8437,8541,8613,8705,8723,9008,9302,9417,9470,9525,9596,9648,9655,9683,9710,9774,9827,9865,9935,10027,10211,10322,10424,10440,10561,10752,10837,10843,10908,10939,11121,11220,11250,11342,11363,11373,11477,11513,11816,11979,12006,12061,12092,12169,12190,12233,12286,12317,12378,12511,12589,12649,12684,12699,12741,12855,12935,13011,13044,13078,13150,13356,13416,13445,13464,13508,13522,13701,13707,13745,13818,13901,13955,13966,14053,14060,14150,14181,14456,14522,14548,14549,14555,14559,14716,14748,14788,14842,14918,14924,14945,14947,15068,15107,15168,15192,15235,15237,15253,15258,15259,15400,15485,15624,15671,15676,15697,15707,15723,15753,15837,15981,16002,16018,16124,16126,16203,16216,16270,16289,16329,16443,16494,16495,16516,16574,16584,16669,16670,16693,16703,16821,16899,16999,17040,17063,17076,17186,17228,17327,17341,17405,17620,17690,17709,17723,17762,17784,17854,17917,17936,18018,18053,18059,18068,18096,18115,18132,18247,18395,18405,18491,18599,18666,18748,18757,18940,18985,19057,19082,19102,19138,19307,19391,19522,19538,19581,19598,19666,19677,19690,19895,19978,20003,20080,20094,20186,20305,20326,20327,20345,20435,20712,20777,20899,20982,21190,21350,21419,21422,21475,21564,21655,21671,21712,21839,21876,21882,21923,21945,22052,22056,22099,22194,22255,22259,22323,22501,22558,22650,22724,22749,22820,22896,22936,22949,22962,22979,23071,23193,23287,23319,23364,23499,23541,23569,23602,23615,23727,23738,23744,24073,24133,24145,24321,24334,24360,24454,24554,24626,24689,24754,24829,24836,24918,24992,25066,25073,25109,25166,25377,25378,25402,25428,25445,25480,25496,25616,25693,25696,25773,25785,25863,25896,25955,26004,26053,26102,26131,26169,26354,26370,26400,26420,26422,26471,26596,26657,26664,26769,26843,26935,27172,27276,27296,27402,27482,27605,27674,27758,27768,27794,27820,27853,27879,27880,27970,28005,28129,28181,28243,28263,28275,28294,28397,28422,28472,28498,28507,28566,28571,28677,28686,28733,28839,28849,28861,28930,28942,28978,28990,29034,29062,29094,29135,29138,29155,29248,29268,29350,29394,29637,29663,29747,29749,29756,29810,29830,29831,29839,29933,30038,30064,30073,30093,30154,30434,30595,30682,30801,30875,30912,31003,31159,31329,31335,31379,31411,31512,31614,31619,31684,31762,31770,31780,31864,32035,32109,32213,32247,32332,32431,32457,32647,32677,32768,32840,32912,32995,33002,33116,33155,33181,33195,33315,33336,33353,33361,33371,33397,33435,33448,33457,33511,33545,33598,33669,33713,33758,33821,33823,33847,33876,33928,34008,34021,34090,34105,34202,34373,34543,34568,34612,34613,34645,34725,34777,34916,34994,35317,35401,35640,35661,35699,35777,35801,35857,35863,35872,35917,35951,35982,36038,36118,36295,36354,36397,36450,36547,36550,36679,36707,36782,36849,36981,36992,37011,37030,37034,37049,37159,37229,37293,37389,37396,37492,37504,37597,37611,37614,37617,37635,37717,37755,37817,37935,38089,38187,38189,38320,38324,38360,38446,38539,38546,38566,38663,38842,38896,38899,38994,39072,39133,39285,39443,39545,39593,39888,40017,40057,40163,40248,40276,40485,40498,40513,40630,40631,40648,40661,40851,40974,40976,40996,41321,41338,41364,41412,41422,41465,41527,41537,41607,41693,41705,41777,41845,41884,41893,42081,42189,42192,42202,42215,42441,42537,42665,42718,42829,42903,42952,42953,43093,43095,43141,43182,43248,43257,43405,43590,43614,43617,43772,43834,43940,43975,44180,44278,44311,44339,44403,44530,44556,44708,44810,45005,45087,45107,45118,45239,45354,45380,45413,45656,45705,45709,45743,45782,45830,45934,45966,46214,46542,46699,46701,46735,46787,46805,46821,46894,47151,47189,47339,47388,47393,47547,47740,47815,47838,47882,47969,48029,48084,48110,48238,48348,48461,48476,48541,48558,48598,48715,48832,48842,48979,48985,49093,49125,49272,49316,49455,49478,49499,49618,49619,49644,49727,49737,49743,49751,49876,49901,49935,49954,49990,50114,50144,50415,50614,50725,50748,50840,50878,51002,51077,51135,51215,51295,51311,51421,51469,51470,51505,51573,51757,51769,51837,51864,51941,51946,51991,52213,52326,52349,52390,52519,52708,52801,52925,52951,52992,53032,53137,53161,53328,53606,53782,53810,54017,54137,54375,54378,54401,54405,54435,54500,54509,54528,54547,54576,54586,54658,54685,54875,54887,55132,55259,55287,55355,55368,55530,55549,55583,55623,55659,55698,55747,55760,55766,55777,55800,55856,55925,55965,56255,56595,56617,56701,56751,56766,56836,56870,56872,56967,57005,57026,57094,57101,57137,57165,57226,57314,57334,57540,57662,57677,57681,57749,57778,57806,57870,57968,58008,58026,58053,58066,58152,58165,58262,58265,58294,58392,58408,58525,58535,58647,58852,59066,59087,59177,59218,59335,59469,59699,59767,59893,59945,59971,59974,60052,60090,60160,60185,60213,60284,60356,60423,60493,60512,60546,60650,60667,60747,60876,60927,60940,60986,61111,61168,61218,61289,61339,61393,61400,61503,61961,62181,62192,62215,62352,62488,62640,62755,62786,62820,62944,62997,63136,63246,63319,63354,63359,63379,63381,63459,63674,63685,63913,64036,64253,64327,64507,64549,64599,64601,64666,64758,64767,64960,65013,65202,65226,65255,65274,65329,65440,65534,65552,65569,65575,65577,65642,65704,65759,65839,66001,66022,66307,66558,66591,66621,66638,66664,66665,66709,66794,67135,67136,67144,67304,67336,67346,67351,67432,67460,67515,67675,67695,67780,67819,67903,67968,68060,68116,68180,68237,68303,68419,68809,68834,69157,69178,69273,69296,69340,69386,69557,69620,69684,69788,69817,69853,70017,70032,70072,70078,70102,70168,70226,70271,70276,70579,70757,70819,70932,70971,71009,71052,71075,71095,71124,71133,71194,71271,71850,71895,72107,72115,72143,72196,72258,72268,72279,72447,72567,72829,72922,72924,72962,73072,73291,73380,73391,73404,73412,73506,73511,73547,73671,73725,73956,74021,74080,74083,74093,74121,74127,74262,74293,74349,74421,74458,74471,74475,74490,74506,74562,74773,74830,74869,74929,74942,74967,75016,75046,75057,75202,75229,75258,75366,75379,75383,75819,75873,75899,75934,75937,75971,76010,76123,76132,76209,76331,76453,76669,76711,76722,76862,77021,77130,77168,77210,77262,77299,77375,77470,77498,77529,77554,77603,77617,77635,77761,77874,77899,78014,78053,78079,78102,78156,78168,78188,78189,78373,78394,78502,78518,78521,78561,78652,78753,78827,78945,79046,79048,79071,79086,79123,79125,79183,79206,79216,79298,79300,79382,79809,80232,80415,80597,80643,80699,80700,80893,80916,81045,81072,81195,81220,81243,81291,81369,81446,81485,81617,81623,81629,81849,81917,81942,81997,82073,82146,82250,82292,82358,82469,82564,82706,82786,82815,82857,82869,82890,82933,83007,83011,83073,83098,83361,83404,83437,83467,83573,83625,83739,83825,83920,83940,83998,84010,84098,84227,84301,84361,84372,84465,84472,84560,84659,84683,84835,84840,84912,85086,85110,85227,85316,85336,85364,85392,85397,85404,85537,85556,85590,85708,85758,85823,85846,85911,85958,85974,85986,86062,86157,86360,86497,86564,86622,86648,86977,87177,87181,87222,87253,87278,87292,87307,87503,87772,87892,87927,87928,87987,88011,88257,88269,88320,88396,88457,88484,88524,88529,88563,88582,88787,88807,88831,88832,88844,88923,89131,89156,89199,89376,89402,89474,89480,89517,89519,89584,89603,89605,89712,89899,89956,90028,90162,90173,90189,90262,90293,90326,90328,90363,90394,90437,90583,90609,90712,90844,90890,90940,90963,91182,91475,91533,91552,91642,91758,91814,91980,92136,92283,92760,92902,92964,93000,93042,93106,93110,93152,93209,93309,93396,93686,93722,93726,93935,93998,94116,94117,94170,94206,94449,94462,94796,94828,94891,95117,95147,95384,95544,95637,95714,95815,95830,95868,95979,96044,96201,96257,96307,96364,96594,96603,96620,96631,96737,96989,97061,97192,97316,97364,97423,97452,97519,97648,97651,97657,97933,98025,98033,98137,98152,98180,98238,98509,98532,98586,98714,98720,98723,98830,98883,99097,99240,99313,99374,99375,99458,99489,99493,99536,99746,99773,99931,99946",
          "73,205,209,281,319,397,574,618,642,839,855,1162,1183,1216,1350,1368,1384,1391,1396,1450,1454,1512,1568,1584,1775,1844,2080,2157,2216,2281,2296,2501,2552,2660,2686,2739,2754,2831,2864,3004,3159,3222,3366,3476,3566,3944,4050,4057,4183,4250,4262,4374,4461,4625,4721,4749,4808,4874,4878,4898,4932,4948,4981,5083,5412,5475,5578,5771,5808,5875,5915,6248,6266,6275,6313,6506,6522,6533,6618,6705,6734,6845,6930,7054,7060,7131,7345,7408,7551,7601,7637,7681,7779,7817,7876,7935,8001,8169,8268,8426,8582,8609,8713,8729,8746,8834,8905,8907,8953,8964,8991,9072,9139,9159,9175,9181,9228,9255,9352,9466,9496,9644,9794,9843,9917,10129,10253,10455,10481,10501,10613,10716,10880,10956,10960,10964,10975,11070,11126,11145,11200,11372,11461,11590,11642,11795,11807,11835,11881,11899,11907,11990,11996,12022,12315,12367,12383,12610,12627,12663,12703,12835,12906,13004,13103,13134,13148,13246,13361,13453,13504,13568,13600,13624,13719,13848,13871,14058,14176,14190,14203,14470,14492,14538,14799,14867,14907,14948,15103,15163,15215,15285,15345,15375,15429,15498,15550,15599,15600,15632,15657,15696,15716,15796,15807,15825,15853,15949,16017,16027,16032,16043,16088,16116,16172,16200,16291,16625,16807,16839,17021,17100,17127,17302,17348,17380,17454,17513,17591,17644,17657,17694,17695,17699,17832,17844,17993,18023,18194,18291,18463,18477,18488,18502,18509,18553,18624,18913,18938,19015,19034,19080,19125,19235,19256,19292,19419,19462,19516,19521,19537,19542,19557,19623,19760,19881,19964,19977,20069,20123,20212,20216,20409,20423,20463,20751,20775,20830,20868,20869,20905,20913,20933,20974,21002,21036,21149,21160,21248,21262,21321,21331,21404,21448,21465,21492,21585,21622,21642,21666,21759,21856,21864,21951,22038,22294,22301,22338,22397,22458,22577,22593,22738,22970,22993,23007,23030,23107,23113,23158,23181,23211,23218,23225,23271,23273,23296,23617,23632,23644,23658,23660,23696,23720,23745,23749,23798,23828,23834,23868,23963,24048,24079,24110,24183,24243,24344,24403,24444,24448,24484,24784,24889,24903,24954,24976,25085,25098,25125,25407,25415,25468,25657,25663,25678,25684,25778,25826,25971,25977,26104,26165,26237,26272,26366,26397,26441,26473,26553,26807,27187,27295,27405,27685,27735,27748,27963,28013,28136,28316,28439,28502,28670,28699,28701,28947,28957,28975,29326,29360,29402,29498,29521,29522,29579,29588,29656,29710,29758,29872,29889,29935,30148,30303,30350,30449,30492,30495,30790,30891,30900,30904,31026,31120,31155,31166,31400,31402,31551,31633,31689,31774,31821,31943,31948,31997,32044,32085,32126,32177,32182,32222,32242,32368,32419,32458,32468,32543,32644,32784,32850,32880,33080,33120,33206,33295,33603,33949,34072,34194,34274,34331,34545,34565,34570,34601,34614,34646,34671,34837,34876,35111,35124,35178,35196,35277,35463,35479,35547,35575,35631,35682,35783,35927,35988,36026,36047,36282,36353,36395,36398,36485,36522,36566,36580,36598,36767,36803,36872,36888,36913,37089,37123,37333,37351,37383,37414,37518,37561,37619,37634,37706,37751,37820,37823,37848,37882,37915,38051,38058,38116,38340,38393,38448,38463,38483,38494,38635,38769,38905,38919,39008,39078,39284,39314,39318,39368,39401,39468,39481,39567,39573,39642,39644,39654,39670,39803,40107,40354,40397,40398,40566,40583,40638,40640,40662,40667,40768,40857,40867,40879,40926,40934,40990,41022,41061,41112,41214,41224,41300,41317,41348,41362,41661,41797,41800,41813,41819,42003,42045,42054,42108,42124,42163,42223,42240,42346,42442,42470,42496,42615,42634,42661,42694,42784,42786,42894,42945,43015,43051,43102,43230,43273,43374,43514,43622,43669,43673,43699,43751,43802,43825,44323,44425,44429,44534,44574,44603,44607,44809,44868,44914,45151,45218,45220,45265,45371,45461,45476,45541,45639,45700,45845,46004,46116,46117,46316,46362,46367,46370,46392,46433,46488,46524,46757,46786,46815,47043,47049,47248,47345,47383,47408,47467,47509,47520,47559,47621,47717,47948,48187,48242,48255,48279,48285,48375,48477,48576,48754,48816,48860,48899,49033,49063,49151,49201,49234,49283,49351,49507,49514,49576,49591,49664,49741,49796,49810,49917,49965,50042,50117,50128,50243,50325,50480,50581,50738,50771,50912,50929,50975,50989,51055,51056,51091,51111,51165,51320,51343,51443,51503,51545,51566,51584,51618,51619,51635,51639,51709,51799,51846,51861,51868,51884,51915,52030,52110,52174,52258,52321,52344,52420,52710,52892,52942,53068,53118,53172,53193,53272,53465,53524,53565,53618,53873,54454,54536,54559,54649,54681,54787,54809,54889,54991,55080,55115,55247,55334,55336,55475,55494,55496,55576,55806,56085,56093,56105,56171,56176,56202,56266,56391,56395,56427,56533,56571,56733,56839,56929,56959,57023,57062,57291,57313,57393,57483,57524,57590,57675,57865,57976,58174,58230,58236,58269,58497,58624,58662,58682,58689,58728,58976,59181,59186,59267,59309,59318,59603,59658,59703,59740,59741,59785,59798,59812,60018,60106,60236,60274,60322,60381,60501,60574,60630,60758,60767,60884,60934,61164,61363,61391,61414,61427,61484,61713,61832,61954,61976,62020,62061,62068,62088,62135,62155,62157,62188,62196,62226,62376,62429,62439,62477,62652,62712,62753,62816,62850,62857,62998,63003,63023,63085,63103,63142,63156,63226,63336,63416,63427,63559,63609,63631,63819,63915,63945,64011,64038,64176,64417,64452,64485,64857,65015,65174,65239,65313,65524,65548,65584,65707,65844,65861,65896,65999,66185,66203,66245,66305,66405,66413,66483,66719,66723,66775,66846,66877,66881,66906,66966,66983,67082,67102,67120,67219,67223,67225,67459,67539,67822,67960,68000,68023,68057,68059,68198,68295,68318,68501,68572,68631,68712,68726,68803,68824,68928,68957,68971,69018,69058,69062,69213,69252,69267,69303,69352,69392,69472,69509,69565,69595,69627,69632,69762,69792,69806,69847,69883,70209,70305,70340,70371,70444,70473,70636,70673,70690,70883,70892,70980,71047,71063,71098,71316,71446,71477,71595,71657,71879,71884,71959,72022,72024,72066,72086,72224,72231,72238,72295,72352,72435,72438,72465,72565,72597,72774,72812,72830,72927,72933,72978,73005,73016,73085,73126,73148,73210,73263,73267,73301,73401,73403,73427,73533,73570,73746,73785,73934,73967,74054,74090,74149,74183,74259,74325,74335,74364,74453,74551,74593,74600,74697,74772,74842,74924,74986,74990,74996,75002,75017,75232,75352,75382,75450,75596,75634,75681,75684,75704,75865,75879,76081,76114,76192,76302,76323,76630,76899,76904,77270,77272,77313,77333,77422,77448,77756,77772,77774,77786,77916,77954,78170,78280,78294,78484,78599,78611,78674,78764,78959,79038,79164,79172,79242,79306,79458,79490,79592,79655,79681,79689,79756,79880,79883,79991,80105,80137,80141,80182,80250,80301,80307,80350,80440,80503,80561,80612,80639,80688,80737,80829,80860,80863,80934,80937,81040,81093,81114,81157,81295,81324,81405,81438,81445,81647,81750,81831,81838,81921,82105,82116,82225,82276,82305,82373,82374,82405,82422,82515,82582,82589,82600,82643,82749,82845,82944,82971,83209,83225,83504,83605,83703,83756,83872,83881,83885,84211,84492,84495,84633,84857,84974,85646,85659,85677,85683,85693,85800,85805,85883,85886,85949,86119,86218,86241,86245,86256,86303,86327,86453,86490,86525,86575,86627,86653,86789,87116,87276,87320,87443,87548,87585,87600,87617,87654,87656,87732,87841,88108,88232,88349,88432,88663,88741,88892,89004,89007,89104,89163,89319,89408,89411,89516,89585,89632,89708,89758,89767,89812,89829,89987,90054,90131,90242,90285,90362,90379,90393,90406,90439,90561,90577,90724,90743,90751,90841,90905,90925,90995,91042,91117,91328,91377,91395,91396,91454,91486,91497,91599,91615,91618,91635,91708,91728,91822,91835,91878,91999,92024,92122,92138,92278,92343,92395,92769,92824,92868,92887,92956,93006,93188,93289,93416,93427,93432,93437,93441,93460,93630,93698,93837,93960,93987,94008,94012,94095,94111,94135,94319,94581,94586,94740,94776,94779,94801,94864,94882,94888,95020,95129,95257,95293,95344,95360,95442,95498,95575,95622,95794,95799,95835,95885,95988,96218,96404,96414,96443,96515,96545,96773,96803,96804,97089,97091,97121,97139,97219,97259,97291,97333,97418,97559,97940,97960,97963,98081,98113,98115,98249,98441,98488,98500,98576,98655,98698,98776,98781,98825,98859,98950,98973,99158,99452,99506,99525,99558,99623,99665,99767,99770,99808,99915,99966,99996",
          "60,83,260,317,421,498,554,557,573,774,907,908,1088,1107,1236,1324,1338,1447,1457,1475,1498,1687,1727,1932,1962,2010,2017,2052,2232,2252,2265,2279,2386,2430,2510,2545,2567,2588,2613,2638,2813,2834,2836,2925,2945,2965,3017,3124,3243,3478,3661,3720,3750,3760,3803,4207,4226,4264,4289,4299,4358,4448,4507,4789,4870,4893,4961,5013,5178,5245,5308,5405,5591,5668,5745,5766,5853,6050,6230,6391,6435,6451,6466,6555,6611,6701,6827,6872,6891,6917,7130,7194,7201,7284,7374,7395,7477,7489,7519,7522,7609,7666,7851,7966,8027,8078,8102,8132,8219,8236,8386,8718,8933,9038,9074,9131,9169,9173,9185,9217,9224,9289,9291,9297,9298,9372,9520,9688,9716,9871,10084,10215,10239,10512,10545,10788,10826,10911,11027,11091,11316,11428,11438,11441,11463,11511,11744,11777,11794,11929,12007,12020,12111,12147,12201,12280,12432,12454,12466,12485,12494,12808,12848,12998,13111,13213,13302,13332,13693,13723,13754,13768,13781,13803,14059,14120,14292,14313,14429,14438,14485,14506,14676,14696,14735,14875,14888,15017,15176,15302,15415,15467,15882,16068,16078,16190,16273,16323,16327,16340,16455,16748,16822,16941,16990,17071,17106,17179,17190,17291,17377,17414,17425,17434,17435,17572,17627,17730,17933,18142,18143,18620,18702,18934,18972,19024,19065,19244,19455,19620,19759,19770,19817,19821,19947,20183,20187,20222,20271,20346,20374,20433,20457,20484,20606,20642,20828,20894,20907,20938,20953,21015,21105,21109,21195,21233,21324,21334,21624,21720,21820,21911,22189,22302,22305,22456,22626,22654,22813,22930,22972,23012,23084,23138,23335,23349,23605,23621,23784,23897,23924,24097,24136,24178,24197,24247,24300,24308,24341,24457,24485,24521,24524,24573,24671,24746,24814,24878,24998,25057,25086,25173,25209,25251,25292,25646,25915,26142,26268,26309,26320,26534,26556,26557,26724,26826,26934,27086,27231,27249,27288,27358,27438,27458,27504,27512,27517,27538,27588,27634,27669,27734,27911,28061,28167,28187,28353,28434,28482,28540,28579,28676,28713,28749,28989,29435,29589,29598,29609,29660,29795,29803,29854,29918,29970,30232,30236,30301,30319,30379,30385,30408,30554,30685,30750,31030,31231,31279,31290,31318,31427,31435,31475,31598,31654,31655,31794,31814,31894,31926,31967,32032,32250,32315,32331,32364,32378,32754,32808,32822,33019,33067,33156,33296,33322,33436,33654,33716,33721,33728,33743,33855,34007,34065,34124,34157,34182,34368,34424,34478,34493,34496,34519,34548,34714,34741,34794,35167,35180,35203,35233,35512,35601,35648,35674,35794,36063,36347,36465,36517,36560,36577,36599,36691,36703,36784,36832,36857,36953,37051,37146,37184,37206,37224,37302,37419,37576,37733,37790,37855,37861,37911,38077,38094,38195,38277,38332,38366,38562,38579,38589,38695,38728,38940,38944,39016,39074,39248,39328,39356,39451,39503,39536,39566,39598,39612,39667,39687,39751,39833,39959,40144,40256,40277,40303,40331,40461,40514,40530,40587,40732,40808,40811,40870,40889,41035,41062,41117,41163,41217,41236,41312,41395,41397,41510,41531,41832,41852,41937,42121,42134,42157,42205,42238,42363,42415,42416,42473,42501,42505,42524,42677,42802,43136,43401,43457,43713,43826,43860,44057,44112,44199,44212,44249,44306,44358,44467,44473,44490,44525,44686,44704,44788,44801,45001,45067,45075,45131,45162,45164,45412,45562,45595,45607,45610,45756,45873,45909,45935,45944,46232,46314,46385,46440,46554,46633,46793,46859,46890,46913,47161,47617,47628,47668,47759,47869,47923,48017,48224,48396,48485,48563,48776,48779,48818,48868,48888,49207,49364,49555,49622,49939,49947,50188,50227,50391,50419,50470,50644,50673,50873,50984,51089,51153,51283,51303,51335,51354,51404,51419,51653,51743,51758,51798,51809,51897,51916,51956,51989,52108,52144,52203,52230,52251,52384,52458,52516,52523,52596,52750,52929,53020,53171,53195,53208,53230,53262,53600,53636,53641,53879,53926,53991,54325,54517,54608,54660,54827,54950,54964,55053,55055,55077,55079,55186,55202,55214,55266,55268,55273,55376,55440,55514,55633,55679,55688,55839,55841,55897,55954,56002,56015,56041,56151,56303,56314,56366,56396,56415,56435,56526,56548,56795,56974,57002,57003,57086,57114,57168,57355,57625,57636,57642,57685,57762,57830,57890,58013,58035,58040,58180,58520,58660,58757,58765,58848,58871,58909,59000,59016,59063,59254,59429,59443,59452,59617,59838,59886,59900,60126,60218,60235,60292,60397,60442,60584,60661,60677,60685,60729,60745,60786,60791,60808,60827,60847,60936,60975,60988,61046,61246,61335,61499,61513,61551,61634,61681,61717,61844,61863,61888,61903,62130,62221,62250,62319,62469,62495,62567,62573,62604,62617,62826,62951,63002,63112,63196,63469,63560,63568,63824,63954,64000,64019,64178,64356,64388,64443,64667,64670,64779,64983,65038,65292,65296,65340,65375,65422,65516,65586,65615,65724,65740,65769,65806,65854,65907,65917,65919,65952,65989,66178,66248,66358,66482,66486,66492,66538,66981,66982,67009,67016,67113,67229,67406,67467,67576,67752,67758,67817,67841,68316,68328,68527,68660,68788,68985,69057,69091,69131,69230,69440,69525,69886,70016,70061,70201,70243,70249,70275,70334,70372,70374,70379,70492,70530,70541,70545,70646,70693,70760,70774,70797,70849,70921,70923,71215,71255,71310,71405,71527,71538,71579,71639,71692,71900,72052,72170,72229,72254,72517,72571,72819,73196,73260,73313,73325,73378,73526,73534,73561,73937,74001,74013,74084,74187,74501,74604,74686,74860,74887,74943,75092,75105,75255,75264,75313,75329,75477,75510,75583,75624,75751,75817,76103,76292,76332,76415,76440,76449,76478,76528,76607,76619,76645,76767,76794,76893,77185,77250,77387,77436,77496,77592,77680,77796,78190,78191,78430,78449,78465,78661,78895,78898,78906,78911,79006,79105,79152,79211,79250,79332,79488,79501,79904,79916,80029,80057,80081,80115,80129,80202,80295,80336,80541,80542,80599,80697,80725,80744,80850,80921,81087,81184,81193,81262,81299,81315,81361,81661,81804,81881,81938,82279,82439,82514,82525,82940,82968,83043,83267,83407,83509,83541,83592,83618,83705,83903,84087,84246,84287,84290,84390,84513,84727,84751,84852,84939,85075,85101,85163,85248,85319,85402,85421,85498,85543,85654,85717,85730,85799,85873,86137,86143,86184,86226,86243,86362,86442,86636,86640,86959,87103,87121,87321,87520,87662,87686,87731,87800,87830,87847,87985,87988,88000,88015,88216,88219,88300,88331,88338,88348,88427,88631,88698,88721,88959,89230,89273,89284,89803,89811,90191,90195,90302,90371,90422,90435,90438,90602,90673,90680,90750,90863,90878,90888,90909,90971,90979,90997,91012,91092,91099,91313,91423,91518,91524,91576,91818,91853,92055,92108,92258,92511,92536,92543,92544,92663,92733,92989,93072,93080,93202,93398,93470,93528,93666,93735,93808,93900,93992,94034,94083,94269,94344,94351,94433,94597,94671,94708,94720,94736,94921,94924,94945,94979,95001,95150,95196,95244,95265,95266,95465,95496,95525,95554,95570,95684,95743,95922,96056,96153,96255,96272,96418,96614,96647,96663,96747,96754,96761,96808,96842,96894,96916,96930,96972,97096,97582,97671,97675,97721,97782,97875,97980,98147,98165,98198,98224,98284,98417,98421,98436,98449,98573,98673,98929,98998,99045,99064,99232,99236,99246,99286,99436,99507,99544,99585,99597,99688,99943",
          "29,112,136,227,371,427,445,565,586,589,592,708,833,869,930,1073,1081,1192,1237,1285,2035,2079,2181,2266,2268,2403,2446,2482,2531,2714,2720,2757,2891,2934,3019,3032,3053,3162,3167,3257,3386,3491,3495,3516,3547,3555,3597,3606,3770,3956,4074,4121,4260,4373,4445,4567,4623,4750,4768,4799,4837,4937,5017,5066,5139,5150,5260,5466,5507,5514,5622,5673,5763,5764,5804,5828,5920,5975,6049,6054,6385,6410,6442,6471,6511,6568,6585,6624,6714,6758,6842,6846,6897,7068,7097,7112,7120,7187,7188,7205,7260,7283,7288,7328,7474,7572,7749,7783,8168,8210,8286,8393,8539,8543,8591,8701,8795,8823,8937,8949,9056,9118,9148,9164,9188,9231,9268,9427,9462,9599,9628,9675,9740,9805,9980,10029,10042,10138,10268,10348,10365,10462,10526,10532,10638,10676,10757,10772,10984,10987,11248,11324,11340,11381,11439,11543,11596,11649,11678,11687,11710,11810,11877,11906,11940,11953,11988,12028,12095,12137,12214,12245,12290,12417,12456,12561,12605,12639,12652,12893,12899,12950,12952,13038,13039,13141,13159,13231,13245,13248,13278,13279,13322,13396,13566,13673,13680,13714,13751,13774,13858,13879,13885,13925,13964,14021,14096,14103,14107,14118,14162,14166,14210,14308,14382,14645,14906,14984,15033,15034,15158,15323,15397,15719,15745,15784,15821,15917,16086,16110,16132,16135,16205,16549,16552,16566,16568,16591,16601,16657,16662,16689,16713,16739,16805,16815,16826,16857,16874,16950,17051,17213,17406,17438,17461,17466,17514,17564,17587,17711,17745,17833,17877,17918,18033,18103,18255,18276,18486,18497,18673,18679,18793,18920,19007,19026,19113,19143,19148,19247,19371,19383,19534,19565,19628,19676,19679,19715,19725,19800,19809,19889,19901,20001,20022,20148,20189,20369,20399,20493,20633,20661,20667,20808,20872,21377,21526,21599,21770,21895,22098,22122,22227,22240,22354,22424,22511,22580,22664,22739,22922,22937,23026,23035,23056,23151,23240,23270,23301,23325,23493,23535,23603,23639,23642,23650,23831,23846,23912,24030,24123,24329,24374,24420,24455,24677,24687,24741,24775,24781,24869,24880,24906,24993,25004,25030,25102,25161,25204,25392,25419,25500,25593,25617,25666,25681,25719,25801,25874,25887,26040,26160,26191,26258,26405,26416,26424,26456,26599,26608,26731,26879,27087,27142,27245,27305,27338,27501,27502,27596,27660,27667,27687,27745,27763,27793,27848,28045,28079,28251,28318,28378,28383,28386,28432,28470,28515,28656,28774,28832,28842,28885,29036,29131,29171,29172,29206,29242,29264,29295,29303,29345,29405,29526,29582,29748,29824,29877,30058,30220,30328,30336,30345,30375,30496,30538,30675,30770,30791,31111,31124,31138,31207,31350,31374,31413,31472,31536,31657,31808,31884,31938,31987,32012,32019,32249,32336,32372,32453,32534,32557,32565,32624,32645,32859,33192,33220,33272,33302,33421,33452,33488,33750,33990,34041,34078,34197,34203,34278,34486,34610,34619,34690,34755,34844,34885,34971,35158,35182,35190,35217,35286,35301,35369,35373,35377,35493,35577,35638,35639,35657,35721,35779,35841,35851,35862,35900,36018,36102,36116,36120,36189,36291,36297,36528,36533,36584,36627,36640,36658,36760,36868,36935,37087,37182,37339,37547,37716,37922,38096,38098,38218,38261,38273,38525,38548,38603,38747,38764,38793,38797,38849,38918,38958,38966,38968,39147,39155,39215,39250,39268,39344,39376,39390,39421,39499,39675,39757,39762,39812,39853,39971,40206,40247,40274,40287,40356,40489,40504,40589,40883,40898,41006,41058,41076,41078,41124,41133,41139,41149,41167,41347,41398,41404,41484,41540,41647,41651,41717,41806,41867,41907,42099,42311,42360,42375,42426,42430,42466,42543,42564,42569,42646,42760,42825,42929,42932,43084,43186,43465,43542,43635,43681,44005,44033,44069,44089,44100,44179,44206,44245,44334,44413,44538,44570,44593,44756,44866,44916,44956,44975,45049,45099,45125,45213,45398,45443,45460,45462,45534,45745,45776,45803,45812,45864,46154,46185,46254,46261,46317,46333,46375,46410,46426,46497,46522,46603,46605,46613,46624,46635,46666,46711,46826,46842,46892,46971,46988,47104,47184,47191,47225,47341,47422,47494,47586,47602,47610,47663,47774,47828,47840,47903,47956,47958,48038,48097,48184,48363,48368,48428,48479,48512,48587,48735,48788,48793,48849,48902,48951,49061,49067,49074,49099,49119,49232,49235,49309,49437,49467,49552,49646,49682,49691,49701,49777,49829,49878,49907,49981,50017,50034,50045,50183,50259,50293,50311,50321,50362,50404,50418,50456,50535,50667,50766,50797,50837,50883,50958,51136,51176,51185,51190,51431,51441,51568,51591,51682,51698,51728,51877,51903,51959,51960,52052,52159,52294,52300,52387,52478,52513,52582,52598,52701,52723,52904,53259,53367,53411,53430,53439,53502,53525,53563,53732,53955,54011,54040,54084,54098,54309,54667,54689,54699,54908,54922,54989,55233,55239,55429,55592,55703,55708,55801,55813,55946,55979,56055,56101,56304,56393,56406,56436,56525,56637,56838,56840,56948,57044,57084,57141,57256,57263,57267,57279,57576,57629,57661,57725,57832,57833,57983,57990,58051,58155,58187,58257,58259,58263,58412,58448,58547,58586,58618,58633,58699,58745,58860,58872,58901,58994,59013,59060,59188,59207,59345,59422,59529,59556,59561,59600,59605,59629,59796,60092,60117,60164,60216,60383,60432,60461,60486,60567,60568,60623,60760,60860,60906,61106,61170,61200,61235,61250,61286,61295,61317,61511,61653,61733,61734,61753,61810,61880,62123,62154,62293,62297,62349,62493,62572,62632,62723,62741,62789,62802,62822,62963,63088,63229,63370,63456,63495,63547,63740,63764,63780,63784,63866,63873,63882,64044,64202,64217,64242,64340,64359,64413,64440,64492,64494,64674,64683,64690,64733,64752,64771,64824,64841,64880,64886,64949,65171,65231,65282,65331,65412,65491,65589,65603,65624,65648,65736,65808,65850,65857,66071,66200,66312,66388,66397,66400,66407,66636,66787,66931,67062,67263,67273,67290,67412,67428,67431,67470,67511,67596,67633,67739,67771,67800,67984,68074,68217,68298,68305,68494,68505,68578,68619,68701,68719,68896,68958,69148,69190,69214,69330,69390,69572,69633,69770,69909,69914,70046,70113,70253,70256,70347,70406,70445,70655,70680,70785,70799,70916,70929,70949,70988,71040,71104,71198,71302,71309,71457,71680,71693,71746,71819,71839,72041,72098,72249,72250,72280,72326,72351,72448,72470,72625,72699,72775,72800,72816,72840,72988,73100,73147,73172,73175,73274,73293,73363,73377,73399,73438,73444,73758,73764,73780,73781,73851,73862,74141,74274,74295,74299,74363,74520,74771,74880,75176,75290,75314,75372,75424,75531,75533,75564,75573,75811,75878,75882,75887,76025,76062,76165,76218,76309,76410,76462,76480,76483,76541,76559,76634,76768,76879,76903,76962,77093,77143,77218,77242,77256,77317,77361,77392,77536,77599,77637,77714,77755,77806,77881,78037,78078,78114,78132,78363,78678,78709,78715,78868,78990,79234,79264,79344,79416,79455,79472,79623,79648,79706,79728,79744,79821,79973,80159,80253,80309,80479,80491,80608,80714,80743,80757,80831,80889,80976,81073,81096,81297,81330,81334,81523,81550,81760,81772,81803,81869,82141,82266,82295,82376,82414,82508,82596,82609,82621,82678,82689,82850,82881,82919,82921,82960,82980,83056,83170,83389,83391,83478,83485,83631,83636,83697,83758,83788,83900,83951,84033,84099,84116,84153,84207,84219,84296,84324,84337,84338,84340,84558,84641,84650,84816,84893,84915,84952,84969,85335,85350,85401,85611,85695,85812,86039,86098,86127,86146,86195,86246,86407,86506,86513,86519,86569,86573,86598,86624,86716,86736,86840,86887,86923,87082,87092,87111,87157,87163,87237,87493,87717,87746,87787,87808,87959,88063,88064,88160,88209,88312,88335,88369,88392,88430,88502,88579,88608,88828,88861,88884,88891,89011,89026,89038,89188,89289,89315,89352,89369,89377,89438,89507,89537,89620,89769,89777,89846,89884,90049,90193,90247,90251,90266,90388,90434,90490,90526,90537,90591,90608,90623,90653,90682,90716,90864,91155,91202,91232,91372,91394,91430,91444,91468,91571,91672,91848,91889,91995,92019,92023,92085,92352,92381,92389,92595,92696,92710,92766,92903,92907,92969,92972,93021,93088,93094,93118,93138,93169,93304,93521,93570,93697,93847,93861,94038,94040,94065,94175,94176,94187,94251,94345,94350,94398,94412,94522,94717,94731,94752,94784,94803,94955,95083,95157,95183,95249,95278,95285,95295,95315,95330,95428,95657,95685,95698,96013,96114,96163,96195,96345,96390,96493,96557,96568,96666,96681,96716,96729,96750,96797,96859,96884,96922,97104,97112,97194,97240,97320,97346,97420,97448,97600,97611,97663,97724,97731,97922,98333,98347,98391,98452,98458,98702,98745,99080,99183,99260,99330,99476,99628,99723",
          "128,145,148,471,479,496,560,590,701,778,779,866,882,933,1009,1039,1274,1435,1482,1515,1522,1585,1701,1779,1961,1977,2047,2155,2190,2372,2491,2533,2541,2667,2814,2871,2875,2877,2932,2940,2998,3108,3128,3198,3321,3482,3788,3899,3915,4012,4030,4048,4150,4230,4329,4387,4618,4629,4656,4698,4701,4758,4785,4818,4840,4879,5047,5092,5110,5161,5341,5442,5644,5738,5767,5791,5810,5896,5942,5956,6001,6164,6188,6228,6384,6404,6483,6487,6743,6841,6858,6900,7127,7314,7334,7348,7366,7465,7562,7630,7717,8074,8184,8208,8343,8367,8417,8461,8556,8668,8675,8725,8868,8934,8975,9109,9280,9346,9556,9867,10033,10037,10057,10225,10264,10284,10299,10301,10315,10347,10380,10551,10559,10570,10662,10664,10697,10771,10860,10903,10979,11057,11086,11106,11241,11242,11260,11262,11265,11305,11319,11350,11401,11459,11485,11501,11715,11766,11786,11871,11891,11954,11994,11995,12001,12124,12253,12297,12329,12356,12465,12532,12555,12583,12597,12623,12631,12746,12959,12974,12982,13064,13164,13210,13269,13338,13357,13377,13378,13739,13769,13773,13878,13957,13959,14048,14067,14196,14248,14266,14391,14455,14465,14589,14641,14776,14858,14916,14949,15184,15288,15494,15725,15765,15847,16112,16127,16159,16177,16260,16333,16604,16704,16804,16860,16915,16939,17024,17043,17055,17216,17221,17223,17284,17288,17336,17383,17512,17563,17623,17807,17901,17931,17990,18026,18195,18495,18508,18598,18699,18946,18986,19041,19160,19162,19208,19465,19675,19698,19731,20055,20068,20282,20432,20439,20518,20559,20577,20716,20895,20906,21222,21250,21292,21392,21481,21482,21501,21504,21523,21570,21652,21696,21724,21841,21960,22002,22111,22147,22186,22221,22245,22270,22284,22344,22375,22450,22567,22591,22732,22753,22951,23000,23048,23097,23111,23342,23374,23444,23446,23508,23575,23606,23731,23905,24106,24269,24273,24459,24489,24525,24557,24665,24792,24893,24895,25005,25190,25254,25334,25364,25372,25474,25485,25647,25652,25729,25771,25903,26327,26581,26594,26691,26752,27002,27016,27201,27306,27396,27403,27516,27624,27642,27699,28080,28109,28355,28376,28388,28442,28467,28489,28620,28700,28817,28941,29139,29280,29359,29374,29458,29478,29611,29623,29640,29705,29863,29885,29920,30211,30466,30529,30621,30637,30696,30745,30857,31038,31141,31269,31294,31452,31484,31644,31646,31702,31776,31914,32027,32033,32105,32241,32255,32293,32310,32349,32631,32640,32673,32745,32820,32869,32908,32949,32956,32990,33054,33083,33099,33104,33153,33160,33301,33418,33557,33575,33576,33672,33691,33769,33903,33911,33913,33966,34195,34205,34254,34305,34389,34416,34443,34522,34655,34819,34835,34873,34949,34957,35010,35015,35091,35234,35269,35318,35330,35351,35362,35438,35743,35805,36002,36122,36201,36330,36457,36464,36540,36721,36740,36742,36746,36882,36927,37006,37099,37107,37126,37171,37195,37308,37479,37599,37602,37642,37671,37700,37752,37884,37988,38020,38039,38068,38241,38385,38447,38523,38532,38536,38745,38851,38934,38935,39113,39136,39367,39508,39535,39655,39706,39811,40242,40262,40382,40619,40673,40858,40874,40933,40937,40987,41030,41220,41229,41310,41432,41553,41559,41590,41617,41635,41669,41711,41739,41818,41995,41997,42102,42260,42274,42552,42556,42838,42851,42891,43066,43078,43224,43225,43229,43251,43300,43339,43923,43924,44170,44230,44286,44290,44370,44376,44394,44508,44548,44563,44647,44893,44962,45048,45137,45254,45357,45464,45623,45627,45738,45793,45883,45905,45975,46054,46058,46104,46141,46144,46177,46224,46467,46535,46665,46717,46728,46772,46899,46909,46911,46947,47033,47047,47207,47226,47227,47346,47405,47483,47528,47609,47639,47680,47697,47731,47875,47930,48068,48085,48260,48266,48290,48315,48473,48480,48570,48637,48944,48950,48989,49001,49016,49022,49223,49289,49353,49414,49422,49509,49547,49574,49711,49801,49825,49930,49953,49975,49984,50068,50150,50204,50317,50331,50348,50372,50668,50677,50767,50803,50858,50865,51006,51512,51533,51765,51882,52001,52008,52130,52220,52285,52316,52382,52399,52447,52547,52586,52637,52651,52659,52728,52827,52836,52853,53023,53247,53297,53432,53471,53535,53538,53568,53682,53718,53741,53792,53858,53890,53907,53922,53943,53967,53976,53985,54117,54188,54302,54376,54498,54602,54934,55006,55112,55131,55317,55388,55468,55517,55528,55559,55628,55900,55901,56099,56217,56248,56352,56414,56441,56517,56560,56576,56592,56619,56731,56897,57085,57087,57403,57418,57427,57451,57507,57598,57622,57740,57765,57852,57893,57934,57935,57960,58036,58103,58108,58184,58260,58316,58333,58383,58468,58480,58539,58580,58608,58656,58724,58806,58987,59154,59269,59320,59505,59541,59554,59567,59590,59610,59693,59800,60080,60108,60125,60186,60282,60332,60366,60424,60433,60529,60535,60625,60696,60739,60769,60859,61012,61027,61051,61064,61232,61330,61386,61404,61612,61709,61730,61775,61881,61910,62013,62043,62074,62139,62147,62156,62200,62207,62222,62241,62303,62348,62464,62472,62478,62529,62580,62621,62623,62684,62734,62777,62801,62888,63243,63770,63793,63904,63930,63974,63986,64094,64110,64116,64130,64233,64418,64592,64664,64872,64878,64901,64934,64935,64950,64955,65020,65279,65387,65402,65428,65433,65540,65565,65709,65756,65768,65810,65830,65925,65942,66067,66102,66194,66270,66315,66370,66573,66585,66717,66740,66839,66890,66984,67044,67190,67277,67416,67560,67601,67696,67703,67719,67875,68520,68606,68844,68920,68986,69041,69055,69073,69095,69134,69266,69424,69448,69485,69530,69665,69959,70049,70086,70093,70285,70420,70426,70739,70767,70840,70889,71035,71064,71072,71142,71253,71268,71458,71607,71712,71812,71849,71872,71914,71934,71996,72294,72492,72685,72768,72785,73022,73069,73177,73262,73354,73370,73448,73450,73465,73478,73504,73654,73885,73927,73971,74044,74220,74402,74468,74510,74533,74567,74694,74753,75027,75148,75204,75205,75256,75336,75401,75532,75623,75646,75648,75840,75877,75910,75965,75970,75997,76319,76463,76487,76498,76513,76516,77097,77193,77228,77399,77476,77574,77578,77789,77805,77962,78003,78017,78139,78239,78271,78528,78589,78768,78983,79009,79135,79193,79249,79442,79524,79585,79594,79601,80034,80070,80096,80126,80179,80185,80269,80334,80349,80353,80404,80637,80667,80668,80879,80930,80947,81031,81367,81368,81440,81534,81640,81794,81805,81821,81868,81909,81954,82026,82034,82049,82234,82698,82851,82877,83004,83103,83195,83326,83345,83452,83470,83500,83545,83554,83578,83600,83807,83961,83984,84011,84026,84069,84079,84121,84303,84387,84483,84524,84671,84796,84806,84978,85021,85134,85191,85258,85278,85298,85338,85362,85386,85503,85583,85601,85645,85660,85750,85780,85861,86087,86222,86309,86390,86394,86498,86508,86521,86563,86819,86907,86908,87000,87273,87433,87450,87558,87748,87762,87879,87964,87989,88013,88119,88165,88229,88304,88328,88399,88574,88607,88614,88675,88779,88790,88984,89085,89179,89205,89224,89235,89262,89300,89351,89368,89452,89533,89544,89660,89736,89945,89964,90201,90202,90228,90320,90440,90446,90473,90538,90584,90631,90884,90941,91146,91161,91267,91384,91500,91534,91629,91640,91677,91803,91930,92029,92069,92176,92214,92217,92236,92243,92342,92447,92454,92732,92912,92966,92977,93013,93023,93037,93154,93219,93381,93476,93504,93723,93750,93777,93797,93908,93910,93974,94054,94115,94266,94390,94464,94552,94583,94595,94813,94861,94900,94907,94932,95017,95132,95181,95342,95369,95427,95588,95671,95697,95735,95736,95787,96017,96028,96100,96144,96506,96510,96574,96731,96763,96775,96789,96833,96855,97115,97119,97161,97223,97227,97282,97541,97641,98011,98096,98123,98144,98166,98302,98345,98409,98499,98575,98593,98633,98647,98674,98691,98804,98810,98991,99032,99309,99514,99527,99528,99567,99578,99584,99694,99768,99906",
          "266,269,282,313,321,341,345,362,366,381,385,394,410,415,477,497,501,529,657,725,836,843,922,954,993,1027,1094,1241,1260,1280,1298,1343,1394,1612,5001,5085,5153,5175,5237,5279,5282,5294,5310,5336,5459,5571,5646,5652,5654,5694,5773,5815,5820,5825,5879,5927,5934,6043,6135,6149,6150,6204,6222,6526,6644,6678,6741,6797,6896,7000,7105,7146,7304,7357,7377,7596,7631,7654,8494,9007,9157,9218,9464,9744,9748,9799,9959,9968,10110,10454,11042,11593,12361,13172,15460,15779,17626,17995,18966,20506,20548,20846,21808,23478,26577,27981,33585,34275,40543,43942,50070,50148,50273,50301,50349,50464,50477,50536,50699,50718,50729,50915,50993,51011,51097,51168,51227,51407,51444,51482,51679,51706,51849,51937,51940,51942,52005,52185,52201,52250,52286,52336,52472,52602,52619,52734,52748,52757,52879,52885,52944,53026,53129,53286,53305,53353,53426,53602,53686,53727,53772,53844,53987,54125,54359,54409,54420,54444,54447,54457,54496,54591,54595,54709,55086,55244,55313,55365,55430,55495,55557,55635,55640,55649,55735,55904,55908,55929,55990,56007,56016,56038,56108,56155,56264,56354,56421,56490,56568,56654,56734,56914,56932,56938,57162,57240,57339,57369,57417,57424,57683,57698,57985,58012,58021,58178,58190,58196,58227,58407,58485,58551,58593,58622,58671,58786,58797,58888,58961,58982,59141,59275,59319,59350,59361,59481,59564,59623,59682,59845,59848,59853,60200,60243,60342,60354,60408,60463,60482,60675,60695,60749,60785,60902,60950,60974,60982,61131,61254,61306,61323,61401,61496,61554,61599,61676,61698,61799,61866,61952,62117,62327,62555,62631,62874,62899,62916,62923,62928,62946,62966,62985,63104,63195,63206,63320,63389,63457,63515,63573,63680,63699,63778,63926,63961,64067,64131,64267,64463,64548,64568,64788,64927,64944,64970,65075,65253,65352,65380,65463,65782,65899,65979,66004,66044,66072,66092,66148,66197,66262,66299,66337,66356,66563,66757,66854,66875,66979,66989,67158,67186,67193,67446,67664,67699,67843,67859,67860,67941,67958,68186,68297,68363,68574,68658,68694,68741,68815,68840,68850,68970,68995,69009,69217,69221,69287,69405,69444,69615,69765,69837,69904,69980,70203,70238,70354,70487,70514,70658,70670,70749,70801,70824,70915,71058,71209,71278,71283,71360,71387,71600,71795,71938,71978,71982,72059,72083,72116,72167,72201,72216,72272,72276,72350,72373,72392,72414,72418,72443,72558,72652,72780,72827,72896,72906,72908,72953,73101,73160,73251,73307,73405,73531,73567,73572,73613,73646,73659,73664,73690,73808,73908,73947,74081,74190,74214,74233,74270,74359,74558,74611,74730,74812,74823,74863,74960,74965,74968,74969,75048,75142,75274,75283,75461,75467,75469,75528,75577,75829,75941,76157,76185,76207,76476,76539,76667,76747,76857,77011,77031,77054,77194,77204,77222,77285,77425,77520,77707,77713,77883,77915,78023,78027,78123,78138,78142,78247,78346,78401,78432,78509,78511,78563,78667,78738,78973,79039,79163,79315,79435,79440,79529,79560,79563,79571,79574,79692,79737,79797,79800,79804,79806,79815,79817,79976,80073,80078,80102,80212,80257,80270,80343,80407,80420,80563,80719,80750,81024,81076,81091,81382,81384,81434,81468,81543,81589,81590,81676,81748,81843,81885,82037,82190,82237,82298,82319,82410,82533,82544,82692,82762,82871,83027,83052,83315,83453,83637,83673,83683,83713,83844,83846,83849,83914,83966,84076,84119,84294,84525,84605,84629,84687,85078,85112,85122,85359,85396,85410,85477,85482,85484,85576,85610,85628,85709,85722,85839,85909,85910,85922,86021,86091,86121,86417,86439,86782,86885,86890,86930,87169,87280,87332,87341,87383,87448,87595,87629,87641,88009,88017,88045,88121,88277,88353,88405,88441,88589,88644,88676,88886,89052,89128,89162,89238,89487,89613,89658,89851,90138,90330,90510,90529,90554,90579,90615,90656,90747,90785,90902,90972,91052,91090,91122,91176,91231,91250,91354,91556,91763,91843,91879,91888,91890,92175,92201,92211,92326,92446,92578,92801,92895,92935,93095,93159,93264,93338,93495,93637,93690,93711,93924,94106,94149,94259,94379,94410,94427,94435,94535,94606,94701,94703,94743,94763,94766,95033,95072,95089,95226,95411,95473,95840,96256,96389,96416,96444,96564,97046,97159,97315,97323,97412,97416,97421,97436,97518,97593,97696,97774,97810,97845,97871,97952,97969,98061,98393,98401,98516,98699,98775,98793,98862,98871,98895,98938,99005,99037,99241,99338,99361,99526,99576,99801,99852,99897,99968",
          "138,162,217,343,456,514,548,555,570,600,601,793,811,904,905,1140,1178,1202,1266,1271,1286,1431,1529,1637,1808,1871,1874,1882,2006,2111,2113,2143,2177,2287,2455,2499,2585,2662,2681,2780,2829,2927,3048,3049,3168,3285,3344,3371,3458,3493,3517,3653,3796,3844,3852,4099,4186,4222,4306,4320,4342,4399,4543,4608,4728,4781,4782,4820,4849,4985,4999,5002,5023,5055,5137,5320,5499,5554,5642,5752,5785,5802,5939,6127,6136,6269,6318,6332,6369,6437,6659,6688,6708,6762,6938,6942,6944,6972,6985,7176,7242,7394,7404,7605,7608,7737,7770,7856,8009,8022,8091,8140,8206,8223,8436,8453,8457,8463,8549,8558,8568,8657,8733,8753,8806,8849,8935,8967,9106,9182,9292,9317,9324,9366,9411,9477,9491,9536,9677,9720,9813,9903,9961,9967,10070,10081,10172,10309,10390,10439,10459,10515,10569,10577,10580,10586,10615,10686,10738,10795,10856,10932,10938,10957,11015,11117,11178,11184,11238,11369,11375,11394,11406,11444,11464,11510,11550,11600,11625,11631,11670,11709,11734,11815,11861,11903,12033,12041,12106,12239,12343,12363,12444,12469,12616,12659,12675,12849,12881,12900,12941,12963,12990,13025,13074,13114,13182,13230,13292,13311,13337,13415,13468,13475,13596,13681,13732,13759,13785,14012,14072,14088,14165,14192,14458,14540,14684,14707,15120,15275,15284,15376,15475,15508,15556,15677,15732,15768,15872,15933,16092,16271,16321,16405,16444,16578,16679,16762,16798,16890,16919,16920,17303,17411,17768,17837,18131,18221,18337,18466,18483,18574,18615,18680,18697,18723,18732,18759,18828,18865,18945,19047,19221,19326,19356,19446,19545,19848,19851,19917,19950,19981,20219,20225,20332,20342,20420,20537,20564,20570,20743,21095,21126,21359,21368,21405,21471,21499,21593,21614,21618,21644,21661,21716,21863,21872,22066,22089,22127,22129,22174,22216,22275,22309,22395,22428,22457,22462,22494,22615,22632,22674,22682,22694,22744,22763,22850,22880,22977,23057,23149,23191,23202,23255,23311,23433,23453,23506,23634,23832,23959,23975,24005,24199,24314,24322,24436,24577,24622,24667,24892,24970,25068,25158,25257,25282,25318,25360,25469,25650,25655,25687,25699,25706,25822,25845,25873,26087,26092,26269,26285,26386,26491,26518,26576,27054,27114,27152,27287,27316,27320,27352,27465,27475,27513,27551,27662,27834,27849,27865,27867,27955,27957,27967,28131,28276,28277,28328,28409,28687,28697,28777,28868,28897,28919,28937,28995,29133,29168,29395,29408,29698,29717,29724,29734,29753,29975,29997,30098,30430,30464,30530,30825,30848,30910,30967,30969,31047,31058,31131,31215,31272,31343,31422,31447,31561,31632,31699,31718,31777,31797,31799,31841,32009,32129,32145,32286,32430,32496,32511,32520,32698,32714,32799,32883,32886,32999,33078,33203,33449,33512,33605,33616,33679,33788,33888,33993,34006,34024,34046,34123,34222,34273,34400,34426,34448,34468,34492,34554,34611,34641,34691,34698,34747,34879,34904,34915,34945,35123,35197,35202,35336,35408,35473,35498,35519,35681,35683,35693,35716,35727,35885,35910,35924,35950,36079,36167,36230,36301,36369,36377,36434,36483,36499,36519,36554,36681,36732,36813,36946,37264,37317,37560,37612,37639,37658,37680,37701,37757,37795,37844,38009,38015,38132,38139,38182,38222,38279,38394,38441,38559,38606,38619,38833,38912,38976,39055,39075,39169,39177,39196,39260,39373,39425,39455,39700,39707,39760,39816,39925,39994,40014,40049,40053,40109,40170,40181,40224,40291,40298,40531,40693,40711,40789,40847,40914,41019,41186,41206,41260,41284,41307,41332,41518,41557,41564,41578,41687,41689,41703,41715,41728,41781,41811,41833,41948,42025,42149,42489,42493,42516,42651,43006,43038,43103,43126,43160,43272,43306,43367,43512,43763,43797,43807,43935,43952,43969,44177,44200,44329,44451,44504,44518,44520,44586,44727,45072,45177,45372,45385,45643,45716,45784,45797,45877,45983,46036,46069,46092,46163,46277,46280,46329,46409,46443,46498,46563,46658,46741,46812,46830,46837,47093,47107,47130,47214,47396,47428,47464,47538,47561,47711,47724,47778,47802,47831,47880,47902,47922,47947,48170,48227,48273,48370,48486,48489,48491,48564,48578,48674,48696,48745,48781,48782,48919,48988,49163,49280,49305,49310,49464,49531,49554,49581,49650,49732,49805,49868,50026,50087,50380,50381,50566,50579,50714,50756,50804,50841,50927,50937,51058,51090,51247,51384,51488,51594,51599,51624,51647,51670,51739,51797,51804,51819,52074,52209,52345,52481,52544,52616,52650,52660,52681,52716,52742,52747,52768,52798,52997,53019,53024,53055,53064,53151,53332,53350,53370,53508,53514,53532,53592,53634,53673,53683,53689,53692,53893,53924,53941,53954,53975,54203,54239,54408,54432,54516,54525,54636,54680,54755,54790,54816,54896,54986,54994,55026,55121,55135,55306,55363,55733,55792,55808,55829,55847,55850,56001,56095,56125,56132,56403,56429,56442,56452,56513,56550,56557,56621,56646,56661,56732,56782,56790,57200,57221,57243,57251,57333,57377,57378,57567,57609,57704,57851,57878,57920,58093,58119,58172,58273,58275,58308,58524,58614,58661,58663,58773,58780,58807,58893,58910,59032,59096,59222,59344,59509,59684,59857,60026,60068,60079,60202,60248,60316,60325,60346,60446,60508,60610,60633,60649,60651,60657,60689,60700,60773,60843,60844,60855,60911,60960,61061,61062,61091,61205,61343,61392,61421,61441,61452,61505,61523,61526,61528,61531,61550,61711,61736,61909,62005,62164,62378,62788,62799,62803,62809,62991,63168,63176,63178,63284,63401,63442,63446,63554,63592,63650,63655,63678,63750,63899,63953,63999,64024,64140,64206,64409,64472,64570,64648,64812,65087,65200,65308,65600,65636,65645,65693,65745,65753,65761,65807,65877,65892,66065,66108,66124,66141,66145,66205,66289,66317,66318,66462,66523,66779,66837,66910,67040,67078,67086,67126,67497,67669,67692,67740,67770,67824,67826,67828,67876,67879,67902,67915,68066,68097,68101,68266,68286,68290,68309,68409,68440,68496,68588,68652,69070,69288,69291,69295,69320,69328,69496,69559,69629,69668,69902,69948,69955,69981,70120,70150,70195,70263,70336,70486,70509,70612,70648,70654,70727,70911,70919,71008,71010,71368,71551,71628,71756,71944,71967,72013,72210,72315,72356,72361,72390,72437,72452,72455,72539,72617,72660,72694,72697,72868,72889,72916,73034,73171,73216,73220,73276,73322,73328,73341,73392,73474,73601,73879,74038,74102,74161,74194,74321,74327,74543,74895,75124,75126,75169,75262,75325,75343,75349,75403,75426,75442,75456,75494,75501,75621,75677,75689,75703,75706,75707,75713,75734,75747,75863,75948,76170,76387,76407,76460,76473,76781,76811,76836,76847,76892,77167,77182,77192,77264,77318,77321,77423,77523,77540,77571,77644,77686,77725,77766,77817,77974,78038,78047,78229,78250,78265,78282,78309,78387,78532,78557,78666,79170,79189,79202,79204,79305,79366,79498,79698,79719,80116,80173,80206,80310,80333,80397,80413,80498,80598,80646,80781,80875,81236,81247,81300,81311,81357,81455,81460,81639,81683,81688,81719,81944,82075,82112,82129,82217,82236,82874,82928,83069,83111,83208,83308,83369,83371,83396,83398,83436,83489,83491,83587,83634,83826,83828,84070,84107,84307,84321,84373,84484,84516,84603,85016,85127,85255,85301,85304,85313,85449,85554,85769,85772,85872,85967,86040,86076,86118,86139,86235,86306,86614,86668,86696,86706,86717,86743,86806,86808,86919,86932,86965,87043,87053,87141,87190,87260,87262,87310,87484,87487,87529,87547,87553,87589,87618,87635,87691,87707,87801,87802,87861,87923,87925,88127,88163,88195,88243,88256,88359,88454,88471,88592,88682,88703,88735,88759,88768,88871,88918,89024,89057,89073,89233,89336,89435,89454,89548,89577,89817,89818,89905,89915,89921,90087,90333,90336,90357,90427,90487,90506,90616,90618,90661,90700,90731,90868,90932,91068,91083,91104,91113,91138,91141,91172,91181,91276,91330,91391,91517,91572,91639,91669,91741,91990,92037,92145,92215,92341,92397,92560,92580,92643,92677,92758,92761,92872,92877,93032,93055,93064,93139,93184,93344,93464,93481,93593,93605,93621,93650,93654,93673,93687,93768,94015,94189,94536,94540,94624,94649,94655,94800,94927,94956,95009,95084,95143,95207,95255,95407,95474,95508,95547,95592,95608,95615,95641,95658,95801,95901,95974,96036,96161,96192,96336,96530,96674,96879,96959,96981,97247,97569,97621,97647,97727,97758,97799,97899,97908,97923,97930,97932,97955,97976,98091,98170,98187,98515,98683,98770,98812,99044,99069,99284,99516,99728,99732,99780,99883,99982",
          "224,236,252,259,374,500,825,826,837,966,1019,1031,1041,1127,1173,1226,1240,1296,1333,1439,1587,1600,1890,2498,2566,2649,2830,2947,3207,3329,3520,3572,3589,3618,3654,3741,3881,3989,4086,4258,4290,4305,4367,4402,4422,4549,4556,4600,4846,4892,4951,4974,5003,5075,5235,5417,5625,5676,5858,5983,5994,6208,6259,6383,6453,6478,6508,6519,6525,6532,6553,6605,6765,6795,6818,6829,7010,7019,7129,7149,7159,7323,7371,7372,7383,7421,7741,7875,7910,7965,8088,8203,8450,8473,8573,8589,8592,8601,8765,8938,8958,8995,9020,9191,9216,9257,9503,9538,9653,9673,9691,9723,9743,9755,9884,9895,9953,9954,10078,10093,10122,10272,10317,10333,10548,10936,10951,10999,11006,11397,11614,11827,11828,11853,11888,11930,11938,12420,12464,12535,12545,12740,12758,12775,12948,12999,13041,13060,13165,13271,13328,13490,13671,13712,13937,14026,14057,14117,14139,14287,14352,14427,14667,15045,15129,15183,15205,15454,15536,15558,15658,15710,15802,15900,15964,16059,16158,16296,16429,16841,16952,16953,17224,17270,17275,17277,17292,17294,17517,17534,17570,17585,17664,17796,17875,17961,18071,18099,18116,18133,18265,18320,18398,18435,18577,18619,18647,18718,18731,18818,18835,18890,18903,19068,19139,19154,19176,19259,19333,19394,19536,19587,19730,19787,19796,19949,19986,20072,20107,20137,20152,20229,20344,20414,20440,20563,20795,20829,21001,21171,21172,21537,21886,22084,22276,22277,22313,22499,22527,22550,22581,22648,22768,22780,22916,22967,22996,23372,23436,23449,24138,24180,24195,24203,24720,24787,25000,25082,25450,25499,25690,25720,25861,25932,25969,26060,26144,26159,26187,26435,26463,26483,26522,26540,26712,26759,26762,26772,26784,26830,26836,27131,27308,27351,27399,27534,27665,27697,27817,27840,28018,28034,28036,28140,28191,28841,28863,28931,28934,28959,29051,29057,29124,29215,29228,29267,29331,29358,29378,29502,29518,29559,29563,29573,29661,29743,29887,29902,30043,30066,30125,30188,30198,30450,30493,30499,30567,30785,30836,30986,31156,31431,31535,31663,31720,31743,31871,31873,32015,32092,32115,32239,32258,32342,32351,32466,32684,32686,32699,32920,32958,32969,32977,33050,33068,33097,33320,33384,33622,34088,34423,34456,34483,34488,34627,34825,34830,34955,35261,35292,35299,35327,35360,35610,35627,35705,35918,35965,35968,36207,36211,36222,36281,36284,36788,36859,36944,37103,37166,37217,37228,37305,37443,37512,37702,37737,37798,37843,37929,37937,38047,38113,38123,38137,38236,38271,38343,38649,38761,39061,39118,39181,39664,39989,40148,40171,40271,40288,40351,40392,40471,40535,40893,40924,40939,40963,41017,41066,41439,41883,41894,41908,42173,42254,42313,42317,42453,42467,42548,42751,42880,42990,43227,43269,43327,43359,43375,43468,43473,43480,43490,43889,44085,44143,44164,44292,44325,44338,44419,44491,44501,44529,44595,44629,44696,44782,44853,44958,45050,45216,45287,45369,45459,45546,45570,45576,45693,45765,45775,45783,45787,45840,45849,45894,46033,46066,46081,46278,46290,46364,46472,46602,46616,46629,46690,46704,46907,47168,47190,47198,47205,47242,47309,47377,47429,47435,47546,47558,47733,47814,47888,47889,47952,47994,48027,48050,48079,48093,48102,48169,48216,48249,48251,48253,48408,48419,48429,48784,48809,48943,48954,49117,49195,49208,49233,49416,49441,49710,49979,50161,50308,50309,50504,50524,50583,50656,50785,50838,50995,51009,51039,51260,51288,51357,51358,51531,51557,51617,51725,51792,51812,51999,52020,52025,52035,52078,52297,52352,52396,52414,52730,52812,52831,52874,52897,52973,52986,53311,53372,53396,53407,53615,53714,53866,54021,54120,54177,54244,54328,54329,54334,54491,54508,54557,54714,54841,54923,54962,54979,55071,55146,55370,55617,55702,55755,55844,55914,55996,56004,56057,56074,56137,56194,56257,56370,56380,56416,56428,56563,56716,56717,56772,56777,56987,57015,57217,57503,57531,57648,57718,57867,57962,58025,58156,58166,58208,58309,58507,58690,58841,58864,58922,58985,59110,59111,59152,59428,59495,59696,59698,59778,59841,60107,60233,60662,60716,60779,60830,60832,60926,61208,61380,61454,61456,61559,61669,61877,61884,61936,61970,62034,62064,62179,62193,62273,62302,62428,62562,62706,62924,62931,63043,63094,63119,63396,63464,63471,63593,63689,63867,63909,63919,63968,64035,64093,64097,64171,64281,64360,64456,64529,64662,64682,64704,64820,64874,64956,65102,65609,65723,65852,65855,65894,65924,65931,65968,66075,66123,66471,66501,66770,66859,66985,67035,67099,67181,67357,67568,67611,67660,67713,67724,67727,67748,67864,67885,67901,68031,68178,68510,68627,68635,68656,68682,68746,68757,68909,69253,69453,69573,69874,69898,69966,70041,70088,70103,70192,70350,70367,70622,70631,70692,70815,70860,71004,71059,71121,71214,71238,71343,71367,71487,71541,71570,71762,71766,71825,72182,72183,72253,72316,72342,72360,72563,73031,73070,73099,73164,73521,73545,73643,73668,73783,73795,73822,73878,73960,73989,74109,74280,74374,74390,74399,74435,74466,74474,74624,74791,74877,74974,75190,75323,75412,75559,75628,75842,75898,75994,76083,76163,76195,76383,76425,76432,76594,76661,76663,76756,76908,77086,77150,77305,77527,77573,77702,77716,77821,77920,78020,78048,78086,78110,78778,78824,79004,79137,79176,79187,79222,79232,79283,79323,79399,79425,79569,79890,79963,80140,80152,80275,80422,80504,80517,80573,80590,80640,80713,80754,80767,80796,80862,81058,81108,81126,81129,81175,81270,81275,81375,81425,81499,81501,81527,81569,81874,81924,81957,81960,81991,82000,82001,82008,82014,82289,82318,82479,82492,82503,82561,82637,82721,82887,82930,83025,83066,83179,83180,83210,83238,83414,83423,83466,83628,83669,83680,83688,83799,83809,83821,83923,83977,83982,84222,84230,84235,84288,84297,84453,84498,84586,84861,85294,85353,85365,85575,85622,85697,85951,86334,86638,86700,86731,86884,86909,87097,87164,87211,87238,87293,87302,87311,87367,87563,87572,87692,87905,88030,88061,88144,88343,88516,88578,88613,88627,88628,88929,89042,89078,89093,89266,89347,89366,89393,89569,89618,89892,89903,90002,90067,90218,90243,90270,90281,90562,90666,90688,90834,90860,90872,90892,91003,91221,91347,91433,91499,91622,91731,91789,91800,91916,92039,92162,92185,92510,92718,92944,93195,93242,93292,93308,93420,93501,93627,93850,93856,93879,93933,94009,94099,94472,94518,94575,94668,94826,94848,94874,95103,95175,95250,95267,95277,95286,95290,95452,95576,95625,95665,95788,95803,95819,95832,95906,96030,96034,96116,96180,96217,96240,96378,96509,96567,96862,97030,97102,97264,97292,97327,97414,97514,97580,97747,97768,97784,97789,97806,97826,97991,97999,98053,98080,98094,98219,98334,98352,98565,98604,98724,98880,99027,99067,99092,99096,99115,99148,99202,99384,99422,99430,99462,99499,99664,99762,99794",
          "124,197,296,310,422,488,578,727,788,803,994,1016,1156,1160,1176,1251,1355,1367,1410,1481,1658,1698,1776,1781,1995,2089,2288,2336,2391,2489,2500,2532,2609,2616,2784,2845,2938,2943,2952,3040,3111,3235,3238,3252,3309,3402,3540,3581,3610,3665,3738,3816,3839,3882,3950,3967,4046,4253,4270,4357,4457,4536,4715,5044,5086,5120,5157,5222,5223,5318,5495,5498,5537,5560,5812,5936,5957,6130,6279,6406,6426,6476,6543,6583,6602,6710,6816,6926,6950,6955,7102,7154,7261,7510,7513,7564,7615,7739,7751,8056,8092,8144,8177,8181,8366,8387,8419,8640,8647,8777,8941,9073,9083,9138,9303,9363,9423,9506,9563,9610,9618,9714,9718,9739,9749,9768,9821,9833,9879,9915,10130,10149,10155,10406,10485,10878,10891,10901,11039,11073,11080,11137,11347,11443,11472,11473,11488,11640,11858,11868,11959,11989,12035,12047,12069,12114,12117,12139,12206,12227,12459,12496,12763,12812,12926,12996,13006,13145,13161,13275,13353,13422,13461,13753,13796,13884,13900,13906,13998,14078,14169,14362,14380,14386,14411,14457,14482,14780,14820,14865,14881,14999,15195,15206,15207,15266,15344,15407,15418,15552,15603,15609,15628,15680,15693,15761,15858,15890,15945,15948,16001,16105,16117,16175,16226,16229,16278,16298,16315,16373,16374,16390,16392,16528,16581,16593,16598,16614,16721,16737,16877,16883,16917,17062,17120,17212,17241,17347,17458,17464,17538,17586,17654,17799,17819,17888,17998,17999,18066,18104,18138,18277,18308,18380,18419,18453,18504,18510,18642,18657,18681,18746,19013,19187,19255,19261,19354,19361,19604,19608,19616,19625,19721,20359,20406,20422,20425,20561,20567,20579,20731,20776,20893,20941,21008,21087,21101,21117,21118,21183,21406,21411,21496,21543,21554,21579,21625,21630,21722,21753,21788,21805,21979,22156,22193,22218,22286,22391,22476,22502,22503,22548,22646,22662,22750,22772,22851,22968,23109,23169,23234,23292,23356,23362,23435,23447,23486,23531,23607,23640,23722,23814,23823,23849,23993,24038,24116,24140,24165,24182,24230,24256,24266,24287,24345,24660,24758,24847,24902,24935,24955,25015,25119,25186,25325,25356,25506,25578,25627,25661,25784,25876,25889,25941,26001,26017,26041,26044,26146,26156,26625,26667,26694,26707,26770,26802,26839,26871,26898,26957,27036,27056,27058,27116,27138,27211,27411,27591,27706,27711,27732,27876,27950,28043,28164,28174,28282,28523,28532,28535,28603,28664,28743,28967,29075,29146,29301,29363,29445,29462,29484,29551,29552,29596,29616,29675,29723,29754,29765,29812,29834,29927,29999,30132,30253,30716,30811,30878,30987,31074,31199,31223,31245,31320,31330,31367,31425,31429,31448,31506,31541,31557,31636,31920,31993,32050,32082,32088,32152,32198,32219,32363,32417,32486,32491,32495,32572,32580,32642,32785,32919,32924,33011,33166,33193,33298,33300,33552,33554,33683,33689,33778,33793,33814,33822,33864,33929,33986,34010,34059,34199,34283,34308,34311,34335,34337,34349,34430,34441,34481,34721,35001,35036,35119,35323,35392,35472,35582,35623,35629,35649,35767,35974,36172,36182,36340,36376,36389,36418,36455,36491,36608,36688,36889,36928,36938,36974,37065,37109,37116,37148,37177,37183,37205,37261,37295,37369,37497,37562,37667,37876,37892,37913,37925,37992,38100,38109,38230,38333,38371,38461,38612,38711,38938,39037,39095,39151,39199,39205,39241,39315,39331,39382,39413,39463,39513,39599,39740,39796,39817,39855,40104,40383,40391,40412,40422,40444,40459,40637,40756,40903,41122,41127,41191,41204,41247,41306,41325,41331,41506,41513,41582,41597,41626,41759,41801,41881,41901,41916,41971,42068,42213,42358,42372,42373,42644,42772,42795,42852,42957,43021,43094,43112,43199,43214,43277,43304,43433,43449,43471,43497,43507,43573,43870,43875,43897,43966,44147,44262,44365,44406,44465,44575,44687,44821,44862,44894,44988,45070,45136,45147,45210,45364,45383,45471,45494,45565,45580,45662,45746,45757,45763,45794,45913,45919,46008,46031,46035,46121,46169,46208,46222,46243,46357,46593,46710,46797,46846,46879,46916,46946,46999,47156,47204,47236,47330,47336,47441,47450,47526,47550,47572,47606,47749,47762,47769,47801,47834,47911,47925,47968,48543,48588,48636,48903,48974,49136,49166,49190,49203,49204,49236,49246,49284,49344,49349,49476,49596,49623,49819,49937,50190,50191,50300,50341,50356,50371,50431,50513,50591,50594,50706,50761,50901,51032,51073,51112,51134,51145,51225,51319,51340,51459,51692,51710,51729,51741,51895,51949,51996,52115,52165,52221,52231,52335,52410,52603,52617,52712,52732,52744,52959,53093,53095,53457,53472,53574,53617,53633,53650,53665,53669,53687,53867,53940,54049,54063,54114,54131,54190,54195,54265,54285,54418,54431,54594,54774,54786,54792,54801,54807,54876,54895,54916,54929,54940,54963,55033,55118,55219,55281,55284,55318,55342,55358,55391,55410,55439,55449,55483,55487,55503,55567,55727,55830,55981,56024,56083,56138,56201,56260,56289,56349,56376,56437,56556,56565,56639,56760,56800,57155,57174,57184,57305,57560,57596,57678,57974,58199,58206,58264,58297,58354,58467,58530,58560,58778,58854,58902,58905,58968,58993,58995,59142,59175,59223,59229,59351,59372,59448,59485,59530,59633,59647,59829,59856,60058,60062,60065,60072,60198,60293,60297,60363,60516,60552,60632,60642,60647,60712,60824,61022,61082,61092,61095,61124,61160,61331,61368,61371,61418,61609,61847,61931,61987,62113,62289,62575,62883,62935,63111,63165,63233,63249,63323,63490,63517,63539,63620,63623,63738,63827,63861,63967,63992,64068,64085,64086,64126,64213,64232,64238,64313,64345,64488,64523,64540,64563,64966,65069,65100,65386,65421,65430,65527,65647,66049,66069,66179,66251,66360,66376,66395,66518,66528,66701,66746,67122,67220,67394,67447,67486,67679,67863,67951,68013,68470,68544,68579,68622,68698,68742,68789,68811,68899,69140,69182,69201,69206,69286,69306,69468,69751,69859,69868,69928,69999,70096,70136,70219,70220,70382,70458,70464,70548,70564,70699,70751,70884,70994,71012,71026,71034,71259,71352,71496,71557,71561,71589,71621,71765,71799,71857,71886,71905,71972,72074,72093,72235,72355,72436,72662,72747,72928,72948,72957,72961,73010,73386,73402,73495,73568,73586,73752,73756,73827,73828,73892,73933,73946,74027,74087,74156,74239,74395,74492,74495,74540,74563,74715,74769,74786,74844,74923,75061,75234,75289,75333,75348,75428,75523,75696,75765,75816,75834,75839,75921,75987,76199,76264,76314,76334,76361,76444,76538,76544,76563,76587,76637,76692,76934,76960,77067,77085,77102,77122,77125,77127,77238,77266,77409,77472,77541,77566,77567,77604,77632,77740,77742,77804,77837,77884,77921,78026,78134,78137,78164,78240,78270,78279,78331,78367,78397,78419,78575,78577,78694,78856,79131,79167,79199,79219,79254,79260,79453,79468,79656,79743,79840,79941,79948,79961,79962,80060,80191,80382,80469,80482,80580,80798,80836,80870,81010,81095,81104,81144,81223,81340,81452,81495,81572,81583,81657,81833,82126,82477,82618,82660,82699,82731,82788,82844,82954,83159,83327,83338,83403,83428,83624,83670,83747,83754,83780,83843,83892,83956,83959,83983,84045,84055,84113,84256,84286,84344,84366,84370,84425,84741,84769,84801,85052,85109,85157,85345,85479,85505,85644,85688,85818,85819,86041,86064,86074,86095,86113,86290,86388,86606,86689,86737,86773,86809,86911,87036,87047,87067,87138,87149,87391,87539,87822,87968,88060,88081,88107,88182,88207,88220,88316,88346,88468,88469,88491,88504,88728,88734,88813,88874,88876,88882,89062,89067,89293,89341,89428,89628,89728,89743,89834,89883,89906,90036,90149,90154,90319,90622,90627,90671,90709,90772,90942,90961,91035,91098,91453,91506,91699,91719,91734,91747,91805,91847,91880,91985,92003,92018,92034,92065,92167,92171,92256,92282,92311,92327,92340,92423,92502,92505,92565,92574,92638,92660,92727,92737,92772,92950,92960,93004,93010,93018,93164,93272,93314,93333,93422,93447,93546,93648,93718,93771,93855,94048,94173,94177,94307,94618,94642,94659,94660,94947,95091,95166,95178,95275,95300,95346,95502,95513,95527,96038,96092,96200,96373,96381,96461,96521,96705,96958,96968,97058,97098,97137,97329,97359,97608,97643,97645,97652,97756,97781,97783,97843,98006,98076,98168,98185,98230,98263,98264,98427,98469,98519,98536,98757,98805,98815,99031,99281,99469,99480,99498,99511,99574",
          "69,140,238,279,302,700,710,886,892,918,946,1083,1157,1210,1231,1300,1339,1505,1607,1623,1666,1735,1771,1937,2121,2228,2251,2573,2596,2761,2778,2779,2812,2856,2870,2991,3051,3133,3209,3365,3374,3563,3592,3633,3678,3820,3822,3856,4004,4331,4345,4355,4408,4565,4566,4598,4613,4651,4665,4814,5131,5397,5451,5609,5664,5780,5784,5845,5854,5933,6041,6064,6163,6237,6250,6290,6323,6362,6386,6493,6498,6656,6670,6763,6823,6856,6884,6898,6998,7031,7200,7411,7438,7471,7669,7693,7759,7837,7900,7908,7938,7939,7947,7958,7970,7990,8008,8010,8060,8257,8357,8370,8394,8407,8516,8679,8732,8751,8772,8807,8829,8843,8908,8923,9063,9087,9242,9307,9319,9431,9489,9724,9751,9786,9934,9948,9999,10043,10049,10121,10259,10289,10311,10363,10535,10543,10557,10671,10724,10760,10786,10877,10893,10963,11058,11096,11303,11309,11481,11483,11499,11537,11613,11674,11688,11820,11971,12015,12115,12165,12198,12350,12448,12468,12486,12641,12708,12743,12805,12809,12861,12874,12891,12980,12985,12994,12997,13009,13022,13027,13050,13180,13241,13326,13404,13447,13495,13503,13560,13583,13738,13758,13916,14009,14100,14222,14233,14343,14364,14395,14450,14453,14617,14659,14662,14718,14749,14756,15009,15143,15386,15473,15481,15519,15549,15665,15705,15731,15758,15829,15845,15894,16004,16070,16123,16267,16268,16294,16314,16355,16361,16508,16636,16882,16908,16955,17088,17144,17200,17476,17738,17759,17856,17887,18027,18040,18047,18141,18212,18326,18386,18422,18455,18546,18611,18640,18658,18705,18736,18951,19249,19460,19512,19543,19702,19956,20058,20077,20120,20170,20224,20235,20434,20487,20515,20627,20763,20796,20977,21108,21111,21271,21326,21488,21582,21780,21802,21834,21867,21996,22272,22318,22355,22467,22470,22505,22552,22712,22743,22908,22927,23020,23103,23185,23238,23406,23428,23523,23526,23546,23547,23562,23748,23825,23944,24011,24175,24222,24382,24387,24536,24602,24791,24824,24932,25079,25202,25215,25280,25287,25321,25357,25463,25482,25610,25651,25731,25765,25816,26085,26374,26385,26406,26436,26485,26612,26705,26737,26744,26755,26764,26866,26884,26952,26962,27155,27194,27207,27357,27371,27380,27383,27459,27523,27533,27749,27788,27823,27833,27871,27920,28066,28085,28154,28158,28185,28266,28283,28305,28369,28587,28737,28807,28809,29052,29197,29804,29988,30000,30089,30214,30295,30351,30384,30403,30477,30480,30500,30659,30747,30761,31017,31267,31440,31482,31492,31546,31669,31750,31840,31872,31907,31929,32041,32065,32133,32168,32232,32248,32304,32360,32393,32432,32484,32532,32578,32692,32725,32801,32806,32966,33004,33021,33093,33122,33271,33343,33372,33373,33417,33501,33659,34030,34033,34057,34095,34103,34148,34323,34405,34406,34418,34501,34606,34847,34882,34982,35003,35018,35050,35059,35067,35077,35337,35437,35458,35466,35526,35552,35570,35585,35688,35853,35895,35935,35997,36003,36136,36145,36238,36424,36513,36590,36618,36662,36675,36733,36756,36763,36855,37014,37042,37137,37179,37218,37245,37387,37404,37574,37633,37682,37941,38016,38133,38193,38304,38306,38314,38321,38380,38395,38425,38436,38477,38601,38678,38694,38773,39126,39223,39253,39301,39313,39522,39710,39873,40004,40008,40030,40085,40102,40122,40126,40166,40297,40310,40315,40320,40326,40370,40429,40649,40657,40681,40766,40793,40806,40835,40836,40958,41045,41189,41243,41262,41356,41423,41551,41616,41630,41649,41698,41723,41817,41831,41887,41915,41962,42181,42382,42448,42462,42482,42793,42906,42940,43035,43096,43127,43172,43201,43218,43260,43342,43409,43435,43533,43553,43564,43595,43600,43708,43784,43816,43828,43854,43871,43894,43957,43988,44076,44088,44121,44192,44267,44379,44477,44620,44640,44670,44725,44753,44800,44987,45047,45069,45178,45232,45251,45333,45538,45569,45646,45704,45707,45722,45724,45834,45890,45928,46098,46166,46237,46293,46312,46330,46451,46471,46682,46904,46930,47007,47039,47095,47135,47271,47360,47462,47560,47619,47679,47855,47928,48121,48235,48331,48520,48609,48729,48884,48935,48938,49044,49094,49109,49124,49135,49170,49210,49221,49354,49398,49438,49453,49484,49593,49674,49849,49864,49879,49986,50156,50244,50254,50256,50330,50335,50466,50469,50589,50651,50702,50773,50798,50823,50835,50930,51203,51285,51297,51338,51538,51761,51825,51844,51847,51867,51876,51973,52036,52065,52099,52119,52163,52164,52244,52313,52324,52392,52395,52432,52614,52795,52802,52873,52949,52972,53031,53257,53294,53340,53377,53521,53775,53857,53877,53965,54218,54220,54252,54266,54558,54581,54611,54645,54647,54655,54785,54871,54906,54959,54996,55027,55191,55197,55316,55529,55531,55582,55603,55613,55722,55810,55846,56136,56158,56363,56433,56705,56710,56809,56816,56849,56891,56901,57119,57156,57189,57280,57331,57400,57421,57476,57491,57495,57538,57647,57651,57747,57754,57880,57964,58068,58246,58304,58318,58421,58478,58575,58708,58953,58978,58981,59014,59017,59051,59095,59197,59208,59272,59416,59533,59540,59680,59723,59789,59973,60006,60032,60073,60231,60386,60488,60500,60544,60799,60917,61031,61275,61287,61355,61468,61506,61628,61712,61781,61948,61992,62152,62232,62346,62364,62380,62392,62404,62520,62681,62703,62772,62876,62891,62911,63035,63064,63150,63417,63646,63647,63658,63760,63997,64106,64257,64268,64278,64320,64487,64590,64652,64665,64967,65144,65365,65372,65417,65477,65532,65637,65672,65675,65727,65870,66234,66355,66375,66378,66520,66627,66755,66947,66980,66993,67038,67146,67185,67237,67301,67337,67510,67524,67540,67583,67734,67830,68036,68076,68146,68184,68355,68433,68441,68458,68509,68513,68559,68825,68916,68955,68973,68974,69011,69029,69079,69133,69161,69199,69311,69343,69362,69514,69554,69779,69784,69786,69834,69946,70028,70310,70370,70513,70635,70651,70653,70729,70734,70737,70926,70952,71050,71109,71130,71136,71160,71172,71265,71281,71328,71479,71500,71555,71727,71764,71876,71885,71968,71977,72035,72067,72121,72192,72301,72368,72391,72449,72479,72489,72512,72529,72659,72859,72910,73009,73052,73558,73779,73804,73925,74065,74216,74264,74419,74442,74482,74584,74661,74751,74775,74795,74853,75015,75024,75141,75384,75397,75437,75487,75499,75627,75632,75705,75756,75771,76039,76048,76078,76134,76151,76256,76258,76266,76312,76439,76442,76497,76552,76626,76724,76788,77010,77142,77411,77512,77623,77682,77745,77825,77865,77879,77885,78081,78167,78286,78418,78473,78494,78504,78572,78590,78698,78905,78915,78942,79045,79053,79114,79190,79510,79520,79538,79638,79742,79841,79935,79954,80031,80088,80158,80245,80265,80273,80306,80364,80400,80610,80690,80733,80852,80880,80935,81248,81342,81349,81354,81420,81451,81636,81735,81744,81828,81887,81890,81894,81940,82078,82253,82269,82271,82311,82323,82328,82436,82482,82483,82568,82745,82764,82773,82789,83132,83137,83175,83241,83269,83271,83278,83291,83352,83377,83405,83464,83560,83630,83796,83924,84006,84031,84115,84130,84258,84444,84455,84530,84568,84838,84990,85071,85275,85443,85777,85794,85961,85989,86077,86223,86286,86311,86312,86344,86348,86365,86373,86393,86451,86541,86572,86698,86726,86742,86745,86768,86791,86800,86802,86854,86945,86974,86987,87220,87337,87381,87463,87496,87593,87642,87770,87809,87849,87869,87877,87914,87960,88032,88084,88171,88186,88370,88424,88453,88654,88797,88930,89178,89277,89477,89505,89648,89722,89725,89913,90075,90231,90329,90413,90491,90555,90570,90798,91071,91105,91184,91222,91316,91358,91414,91493,91523,91609,91646,91833,91942,92000,92015,92111,92188,92388,92533,92559,92655,92673,92753,92845,93083,93100,93158,93233,93251,93282,93353,93505,93510,93652,93694,94113,94121,94145,94226,94273,94408,94707,94881,94887,94941,95022,95063,95203,95262,95363,95366,95539,95562,95649,95656,95758,95774,95866,95870,95887,96094,96177,96312,96413,96422,96438,96536,96569,96611,96670,96697,96732,96764,96787,96788,96904,96974,97313,97325,97437,97457,97482,97680,97734,97759,98107,98161,98232,98286,98376,98405,98419,98435,98440,98670,98716,98782,98853,98960,98971,99021,99046,99053,99289,99368,99376,99517,99577,99609,99662,99681,99683,99695,99779,99784,99913",
          "179,292,324,373,439,694,702,772,799,819,848,911,943,957,1005,1052,1161,1500,1525,1599,1747,1772,1806,1930,2040,2250,2328,2371,2375,2481,2517,2557,2632,2682,2742,2819,2970,3031,3140,3225,3249,3264,3560,3692,3696,3726,3765,3838,3895,3900,4015,4018,4028,4041,4066,4072,4075,4081,4108,4188,4200,4214,4321,4347,4394,4440,4473,4499,4579,4607,4778,4897,4968,4975,5057,5126,5224,5368,5399,5434,5558,5647,5703,5850,5862,5952,6007,6058,6093,6126,6134,6157,6376,6413,6448,6484,6529,6633,6684,6703,6719,6736,6850,6943,6954,6967,7315,7321,7349,7476,7508,7521,7563,7574,7620,7665,7735,7790,7896,8065,8204,8247,8327,8331,8344,8433,8462,8535,8618,8645,8743,8781,8832,8854,8870,8994,9012,9082,9101,9325,9334,9443,9523,9543,9573,9600,9603,9609,9702,9721,9798,9890,10291,10736,10766,10818,10853,10928,11196,11551,11569,11676,11735,11842,11960,12133,12184,12209,12228,12247,12306,12313,12319,12357,12393,12457,12607,12618,12629,12696,12702,12730,12778,12859,12940,12953,13042,13061,13162,13264,13343,13463,13479,13545,13776,13963,13990,14087,14147,14215,14216,14231,14307,14433,14508,14518,14582,14759,14913,15008,15144,15179,15293,15338,15350,15433,15503,15529,15775,15777,15824,15871,15927,16090,16279,16312,16391,16463,16467,16686,16725,16745,16758,16907,16940,16960,16979,17022,17050,17054,17153,17154,17161,17260,17463,17584,17594,17707,17717,17758,17789,17831,17861,17880,17904,17926,17988,17991,18030,18092,18098,18285,18420,18467,18526,18561,18567,18665,18687,18726,18779,18782,18838,18875,18975,19226,19310,19425,19737,19755,19941,19957,20046,20093,20169,20246,20431,20450,20472,20502,20634,20693,20710,20719,20726,20842,20864,20870,20912,21009,21133,21229,21255,21268,21376,21434,21463,21607,21620,21768,21919,21930,21937,22019,22229,22247,22325,22353,22367,22373,22407,22537,22719,22797,22901,22926,22935,23116,23130,23139,23143,23370,23490,23628,23653,23657,23840,23861,23862,23869,24008,24168,24284,24299,24352,24363,24367,24408,24538,24601,24719,24750,24975,25059,25071,25114,25206,25452,25479,25491,25571,25613,25623,25642,25691,25723,25815,25935,25944,25982,26010,26096,26532,26649,26681,26695,26777,26828,26885,27254,27267,27388,27395,27401,27492,27552,27590,27633,27727,27739,27842,27850,27994,28012,28030,28098,28250,28271,28286,28358,28394,28426,28447,28513,28642,28740,28840,28998,29070,29112,29119,29203,29216,29336,29555,29580,29653,29689,29760,29870,29930,29951,30014,30149,30277,30291,30299,30333,30702,30804,30810,30815,30866,30902,30928,30940,31025,31100,31210,31264,31331,31332,31358,31396,31410,31456,31503,31527,31563,31650,31729,31746,31835,31932,32068,32074,32091,32189,32196,32234,32325,32506,32533,32539,32611,32629,32707,32741,32877,33096,33124,33207,33219,33273,33381,33527,33604,33748,33763,33806,33830,33835,33882,33912,33922,33961,34189,34209,34212,34242,34294,34339,34379,34445,34475,34551,34649,34683,34813,34862,34935,35019,35046,35096,35120,35211,35469,35486,35619,35673,35742,35764,35922,36133,36159,36247,36400,36429,36460,36484,36529,36541,36610,36613,36700,36723,36819,36886,36956,37048,37129,37132,37202,37220,37239,37319,37322,37356,37458,37616,37636,37810,37851,37879,38103,38128,38281,38420,38443,38522,38616,38643,38713,38768,38818,38856,38875,38908,38920,38932,38952,38981,39124,39249,39255,39298,39372,39418,39518,39551,39610,39724,39764,39822,39851,39969,39986,39990,39997,40062,40071,40117,40174,40185,40218,40265,40300,40347,40366,40384,40442,40448,40455,40507,40519,40570,40597,40636,40651,40705,40728,40873,40938,41010,41064,41235,41270,41281,41290,41387,41396,41434,41468,41511,41591,41593,41751,41877,41947,42063,42069,42115,42143,42300,42315,42521,42570,42603,42607,42773,42807,43133,43139,43144,43297,43324,43372,43444,43445,43540,43651,43690,43734,43846,43872,43909,43954,43970,44067,44070,44133,44238,44242,44252,44314,44355,44360,44361,44430,44540,44616,44654,44715,44770,44799,44819,44835,44895,44976,45015,45158,45268,45318,45339,45426,45553,45571,45586,45596,45661,45801,45839,45902,45941,46039,46107,46157,46244,46305,46311,46354,46377,46381,46393,46405,46506,46521,46534,46567,46644,46745,46769,46985,47006,47056,47160,47180,47261,47366,47409,47542,47620,47629,47676,47732,47750,47826,47857,47914,48114,48157,48159,48182,48198,48206,48219,48221,48262,48404,48413,48433,48474,48615,48658,48749,48889,49003,49062,49372,49472,49533,49560,49570,49631,49906,49925,50265,50333,50467,50484,50564,50586,50637,50778,51025,51178,51183,51222,51258,51315,51416,51422,51525,51628,51683,51852,51913,51938,51995,52162,52194,52217,52223,52305,52368,52389,52511,52531,52622,52634,52829,53022,53080,53142,53177,53263,53298,53317,53406,53427,53451,53501,53631,53824,54038,54068,54099,54115,54153,54182,54206,54272,54298,54304,54324,54494,54532,54697,54725,54861,55007,55174,55270,55319,55437,55448,55455,55490,55523,55663,55686,55694,55802,55822,55860,55869,55892,55895,56053,56109,56115,56173,56200,56225,56238,56569,56583,56670,56784,56937,57034,57245,57289,57294,57312,57386,57391,57431,57484,57580,57614,57970,58038,58073,58202,58278,58401,58470,58472,58477,58487,58557,58604,58625,58652,58962,58984,59043,59135,59144,59381,59410,59555,59716,59895,59953,59986,60029,60154,60367,60378,60434,60436,60477,60624,60805,60883,60951,60999,61234,61277,61406,61446,61524,61672,61683,61760,61852,61867,61957,62021,62085,62098,62127,62133,62162,62220,62238,62288,62320,62424,62433,62570,62658,62744,62811,62873,63048,63126,63238,63380,63485,63507,63538,63656,63715,63768,63787,63884,63902,63982,64114,64182,64412,64423,64462,64613,64643,64686,64782,64832,64963,65055,65113,65230,65233,65256,65333,65407,65517,65570,65650,65928,66039,66080,66093,66095,66104,66139,66236,66241,66269,66288,66521,66554,66609,66634,66715,66729,66852,66903,66907,66965,67183,67289,67308,67329,67398,67589,67663,67799,67995,68003,68014,68154,68167,68170,68182,68260,68387,68415,68484,68517,68519,68521,68575,68600,68686,68703,68831,68872,68961,68984,68989,69005,69128,69241,69351,69412,69544,69547,69562,69607,69670,69671,69678,69701,69845,69910,70094,70144,70156,70160,70166,70178,70229,70252,70518,70627,70752,70906,70914,70943,71146,71170,71357,71409,71585,71586,71591,71609,71664,71670,71717,72017,72285,72334,72397,72440,72486,72542,72622,72726,72994,72999,73149,73150,73240,73287,73294,73326,73337,73456,73777,73913,73943,74039,74115,74117,74202,74232,74273,74357,74422,74561,74568,74785,74821,74839,74868,74939,75014,75107,75318,75337,75622,75626,75658,75728,76102,76148,76191,76220,76221,76557,76699,76881,76891,76928,76939,76963,76971,77289,77454,77463,77468,77543,77555,77771,77832,77882,77897,78010,78146,78218,78347,78386,78463,78506,78659,78719,79034,79113,79162,79245,79394,79590,79613,79736,80210,80272,80279,80315,80427,80607,80763,80809,80884,80984,81103,81109,81121,81318,81381,81401,81456,81514,81627,81671,81725,81790,81813,81913,81914,81976,82128,82272,82331,82487,82495,82623,82914,82938,82974,83050,83174,83432,83463,83465,83477,83501,83537,83555,83598,83686,83897,83905,83927,83990,84001,84154,84220,84447,84665,84668,84783,84798,84802,84831,84855,84984,84992,85202,85262,85344,85447,85481,85518,85535,85549,85808,85866,85893,85926,86042,86130,86175,86187,86391,86495,86632,86752,86792,87034,87048,87065,87131,87189,87223,87357,87428,87475,87477,87573,87580,87839,87918,87946,87951,87970,88031,88301,88309,88421,88531,88540,88803,88925,89016,89017,89069,89159,89175,89184,89197,89385,89401,89432,89465,89485,89525,89540,89559,89657,89757,89847,89932,89950,89960,89988,90097,90124,90145,90209,90255,90272,90444,90496,90783,90820,90948,91363,91441,91504,91590,91729,91738,91753,91773,91811,91825,91872,91895,91901,92063,92083,92170,92308,92318,92354,92427,92529,92712,92716,92834,92875,92890,93038,93167,93371,93373,93523,93603,93743,93805,93914,93927,93931,93949,94055,94057,94282,94469,94565,94735,94739,94757,94792,94829,94876,94894,95002,95038,95045,95077,95098,95127,95190,95199,95242,95287,95365,95432,95477,95522,95541,95546,95552,95643,95726,95785,95806,95852,95928,95964,95977,96022,96052,96059,96097,96128,96267,96275,96302,96325,96344,96391,96505,96588,96626,96644,96658,96709,96779,96881,96921,96928,96988,96999,97005,97019,97065,97074,97110,97195,97217,97334,97376,97649,97689,98024,98031,98203,98267,98315,98332,98341,98445,98561,98574,98792,98828,98968,99036,99094,99184,99228,99358,99367,99394,99407,99427,99504,99545,99672,99752,99823,99855",
          "84,100,182,221,230,249,338,510,602,677,790,830,928,1050,1056,1089,1121,1230,1234,1313,1318,1342,1407,1423,1448,1557,1594,1703,1715,1896,1965,1989,2061,2166,2282,2397,2515,2684,2710,2782,2905,2949,2958,3001,3027,3120,3139,3157,3211,3331,3340,3346,3399,3405,3435,3536,3607,3683,3850,3889,4005,4062,4109,4266,4301,4396,4401,4523,4551,4766,4792,4854,4891,4987,5000,5067,5096,5480,5536,5563,5678,5705,5916,5960,6034,6089,6261,6392,6570,6575,6581,6639,6676,6733,6833,6839,6881,6918,6977,7012,7074,7141,7236,7252,7291,7484,7486,7559,7570,7579,7607,7713,7772,7793,7887,7907,7913,7987,8050,8265,8315,8317,8358,8430,8467,8552,8569,8608,8623,8625,8720,8799,8811,8929,9189,9368,9453,9481,9539,9571,9589,9649,9846,9883,10064,10193,10235,10303,10346,10452,10457,10469,10522,10534,10601,10635,10657,10828,10849,10857,10871,10888,11043,11087,11093,11105,11123,11270,11306,11383,11391,11409,11658,12029,12032,12048,12502,12553,12557,12647,12660,12671,12716,12748,12799,12867,12897,12929,12979,12981,13063,13325,13350,13451,13527,13603,13669,13704,13708,13918,13946,14062,14123,14226,14646,14647,14713,14722,14805,14825,14845,14909,14976,15039,15075,15115,15209,15319,15326,15463,15469,15530,15553,15857,15932,15976,16050,16055,16080,16096,16182,16357,16437,16456,16498,16520,16569,16609,16717,16795,16823,16859,16944,16986,17018,17065,17142,17214,17237,17328,17409,17442,17445,17459,17487,17510,17611,17658,17713,17756,18065,18207,18249,18306,18424,18437,18476,18482,18498,18554,18677,18856,18876,18944,18960,19126,19229,19266,19268,19336,19400,19450,19563,19575,19610,19686,19693,19746,19778,19990,20211,20227,20248,20319,20347,20378,20447,20478,20514,20603,20682,20691,20787,20822,20827,20879,20909,20917,20972,20973,21182,21299,21447,21553,21584,21680,21762,21813,21888,21966,22076,22104,22246,22265,22343,22350,22372,22406,22413,22492,22659,22766,22809,22842,22891,22898,23006,23050,23060,23108,23161,23180,23189,23293,23382,23418,23530,23589,23598,23760,23806,23884,24049,24089,24107,24164,24201,24249,24254,24324,24362,24371,24406,24458,24539,24604,24621,24646,24659,24710,24776,24807,24821,24830,24832,24907,24933,24950,24965,25037,25089,25101,25267,25326,25416,25430,25433,25529,25557,25635,25697,25704,25715,25907,25979,26023,26063,26124,26143,26181,26357,26417,26477,26607,26634,26698,26767,26781,26891,26892,26905,26970,27063,27088,27093,27147,27199,27225,27334,27404,27602,27668,27691,27801,27822,27917,28054,28073,28074,28201,28375,28492,28497,28533,28573,28578,28583,28590,28692,28946,29077,29181,29202,29245,29247,29337,29381,29414,29509,29564,29566,29600,29604,29659,29787,29818,29825,29833,29916,30070,30079,30151,30292,30314,30337,30343,30421,30470,30526,30544,30545,30648,30706,30780,30874,31021,31046,31319,31383,31560,31618,31651,31731,31781,31972,32121,32204,32362,32384,32516,32556,32573,32634,32794,32813,32830,32965,32980,33028,33034,33101,33109,33262,33276,33279,33288,33291,33370,33383,33482,33592,33647,33656,33739,33770,34017,34060,34136,34214,34440,34444,34508,34509,34584,34623,34656,34718,34726,34734,34969,34991,35110,35198,35220,35270,35294,35345,35454,35470,35478,35830,35907,35909,36096,36224,36272,36329,36388,36496,36501,36515,36542,36582,36586,36602,36682,36780,36841,36871,36945,36976,37017,37073,37165,37272,37429,37586,37673,37697,37883,37897,37964,38249,38269,38331,38358,38413,38424,38475,38543,38644,38696,38714,38819,38841,38870,38999,39110,39138,39300,39338,39438,39541,39600,39703,39770,39788,39809,39878,40002,40009,40025,40080,40113,40134,40153,40199,40202,40203,40293,40435,40449,40538,40600,40621,40721,40899,40973,41116,41128,41136,41646,41720,41795,41835,41920,41972,42080,42347,42359,42427,42433,42458,42559,42619,42675,42712,42716,42982,43018,43055,43168,43202,43246,43325,43446,43462,43582,43629,43847,43978,43991,44030,44107,44150,44254,44308,44374,44480,44626,44643,44680,44730,44845,44846,44905,44906,44915,44951,44957,45026,45044,45063,45132,45214,45238,45389,45435,45583,45710,45726,45759,45768,45802,45853,45861,45898,46001,46028,46108,46198,46251,46348,46432,46465,46500,46851,46888,47008,47312,47421,47465,47570,47574,47785,47946,48046,48191,48248,48415,48437,48466,48504,48557,48579,48618,48687,48823,48828,48913,48969,48976,48984,49021,49156,49186,49287,49295,49336,49370,49425,49508,49522,49544,49673,49677,49723,49865,49882,50120,50221,50340,50398,50542,50666,50684,50763,50781,50824,50832,50935,51063,51101,51128,51235,51240,51305,51310,51388,51420,51473,51486,51508,51513,51570,51721,51730,51772,51866,51893,51986,52082,52222,52282,52341,52406,52465,52537,52554,52626,52663,52828,52848,52860,52877,52995,53000,53051,53185,53198,53238,53289,53322,53364,53365,53544,53823,53953,53969,53971,54095,54215,54237,54312,54341,54365,54393,54453,54473,54524,54535,54650,54830,54865,54943,54967,54976,55094,55103,55140,55149,55215,55240,55326,55480,55602,55674,55773,55799,55877,55899,55903,56106,56157,56213,56431,56459,56642,56930,57040,57143,57235,57336,57390,57395,57396,57492,57707,57789,57857,57863,57866,57912,57973,58048,58151,58181,58192,58290,58350,58461,58474,58590,58655,58719,58782,58784,58883,58940,59137,59190,59303,59304,59492,59543,59566,59736,59757,59947,59995,60152,60173,60286,60542,60570,60787,60923,60978,61034,61060,61105,61221,61256,61573,61584,61589,61674,61879,62170,62244,62247,62401,62465,62597,62714,62774,62835,62907,62953,62981,63008,63059,63070,63139,63160,63247,63256,63257,63266,63291,63297,63314,63330,63341,63351,63373,63394,63461,63503,63580,63611,63627,63727,63728,63991,64063,64133,64200,64264,64537,64557,64564,64700,64766,64951,65008,65172,65300,65437,65476,65505,65558,65702,65809,65834,65938,66021,66133,66275,66291,66303,66327,66408,66450,66502,66637,66658,66659,66683,66773,66814,66913,67095,67096,67228,67328,67341,67413,67415,67581,67637,67854,67862,68064,68147,68264,68267,68276,68353,68436,68478,68591,68603,68669,68857,68858,68888,69056,69155,69168,69202,69317,69388,69399,69490,69570,69574,69616,69634,69703,69798,69819,69824,69850,69903,70184,70235,70261,70390,70546,70581,70583,70717,70762,70778,70848,70999,71091,71134,71183,71196,71416,71418,71480,71710,71788,71802,71808,71892,71950,71971,71976,72026,72108,72157,72251,72273,72675,72676,72739,72791,72886,72888,73084,73089,73133,73165,73184,73192,73290,73453,73634,73638,73747,73841,73880,73929,73961,74010,74050,74136,74189,74234,74271,74272,74459,74488,74583,74642,74735,74752,74764,74767,74787,74862,74910,75025,75406,75518,75589,75712,75958,75967,76011,76200,76245,76405,76576,76685,76914,76951,77075,77139,77169,77224,77269,77306,77389,77547,77598,77737,77843,77859,77907,77937,77949,78183,78285,78459,78574,78582,78587,78598,78605,78675,78835,78851,78924,78968,78988,79002,79084,79117,79198,79345,79448,79457,79536,79546,79622,79703,79789,79814,79905,79919,79933,79997,80148,80149,80240,80366,80509,80545,80582,80631,80747,80814,80815,80871,80926,81030,81059,81075,81158,81290,81322,81326,81516,81533,81777,81870,81969,81993,82089,82145,82232,82258,82670,82685,82694,82733,82787,82816,82947,82961,82972,82977,83017,83305,83362,83457,83521,83544,83561,83613,83811,83869,83929,84009,84091,84134,84278,84313,84322,84382,84415,84416,84443,84468,84597,84610,84612,84637,84655,84704,84709,84723,84736,84747,84779,84829,84876,85050,85059,85063,85090,85119,85123,85289,85395,85553,85558,85574,85813,85916,86032,86058,86207,86221,86301,86363,86364,86405,86411,86438,86446,86592,86600,86635,86680,86804,86825,86838,86864,86894,86917,86938,87074,87132,87186,87202,87215,87508,87766,88040,88092,88118,88137,88242,88371,88395,88401,88593,88641,88685,88808,88935,88952,88953,88992,89001,89194,89246,89280,89348,89472,89567,89573,89578,89710,89723,89778,89862,89871,89917,89957,90027,90105,90361,90392,90455,90718,90775,90817,90825,90880,90922,90935,91007,91101,91164,91217,91244,91304,91410,91492,91577,91856,91868,91968,91971,92095,92134,92196,92248,92257,92315,92337,92496,92557,92592,92629,92649,92662,92846,92906,92914,92953,92970,92991,93079,93261,93266,93452,93472,93498,93590,93660,93702,93747,93758,93819,93853,93907,94011,94041,94097,94129,94178,94185,94258,94311,94404,94443,94554,94658,94698,94981,94987,95026,95035,95167,95318,95390,95398,95409,95466,95471,95521,95535,95577,95579,95770,95802,95883,95890,96009,96123,96156,96317,96339,96340,96528,96693,96756,96805,97008,97013,97094,97149,97152,97236,97417,97472,97477,97547,97859,97942,97947,98098,98374,98478,98508,98537,98562,98774,98866,98879,99185,99242,99243,99293,99294,99366,99415,99447,99606,99648,99684,99707,99789,99876,99905",
          "11,24,95,176,206,255,271,315,424,454,480,503,515,632,709,733,798,808,851,909,932,936,1002,1199,1252,1262,1272,1444,1524,1638,1709,1724,1743,1777,1948,2034,2124,2128,2160,2201,2240,2383,2469,2579,2699,2711,2726,2760,2823,2939,3130,3138,3150,3221,3353,3412,3420,3508,3619,3821,3935,4144,4177,4259,4281,4283,4364,4465,4563,4594,4653,4695,4735,4753,4759,4804,4833,4954,5029,5138,5155,5264,5272,5278,5291,5295,5493,5509,5534,5727,5744,5754,5781,6216,6328,6331,6564,6588,6619,6695,6837,6878,6925,6948,6951,7056,7100,7150,7211,7393,7433,7464,7567,7670,7721,7809,7860,7934,7961,8371,8536,8560,8575,8579,8710,8798,8815,8862,8894,8960,9145,9344,9349,9518,9583,9604,9633,9756,9772,9803,9924,10050,10080,10178,10344,10350,10351,10389,10448,10546,10571,10661,10687,10691,10743,10803,10847,11030,11176,11193,11358,11366,11422,11675,11845,11901,12059,12203,12232,12349,12455,12498,12528,12534,12650,12685,12764,12833,12993,13023,13036,13224,13324,13367,13389,13420,13471,13571,13630,13687,13794,13802,13889,13917,14022,14138,14155,14174,14204,14338,14366,14446,14525,14531,14606,14622,14650,14675,14680,14829,14885,14954,15040,15131,15214,15419,15636,15727,15920,15958,16071,16178,16332,16366,16401,16416,16466,16617,16639,16678,16801,16872,16893,17061,17193,17278,17309,17358,17495,17525,17565,17614,17625,17638,17668,17683,17769,17809,17930,17979,18037,18122,18149,18189,18228,18271,18297,18345,18347,18372,18449,18473,18487,18617,18627,18784,18819,18880,18963,19009,19018,19043,19062,19210,19327,19343,19392,19456,19657,19660,19795,19806,19915,20026,20059,20177,20287,20415,20473,20475,20598,20774,20826,20927,20934,20960,21018,21141,21148,21213,21325,21340,21394,21433,21659,21695,21702,21791,21932,21975,21995,22024,22103,22114,22326,22415,22439,22716,22752,22873,22876,22958,23017,23182,23200,23323,23492,23501,23527,23551,23712,23729,23822,23895,23996,24082,24117,24124,24153,24500,24681,24771,24951,24958,25040,25096,25136,25181,25210,25316,25350,25361,25525,25531,25611,25686,25732,25744,25798,25840,25895,25923,26026,26151,26299,26367,26439,26492,26499,26650,26673,26677,26700,26723,26779,26801,26944,26961,26965,26975,26998,26999,27014,27092,27281,27363,27472,27488,27571,27572,27638,27724,28003,28092,28100,28108,28415,28446,28538,28559,28671,28747,28792,28798,28915,28979,29018,29079,29335,29390,29397,29447,29570,29591,29725,29763,29783,29840,29948,30027,30094,30439,30527,30571,30751,30886,30934,31020,31122,31247,31312,31494,31773,31868,32108,32139,32158,32245,32270,32415,32839,32841,32936,33030,33044,33108,33138,33168,33185,33267,33328,33377,33487,33541,33684,33774,33869,33887,33899,34111,34188,34215,34233,34239,34250,34377,34467,34772,34780,34810,34897,35048,35073,35348,35483,35650,35700,35800,35812,35836,36045,36076,36099,36128,36164,36194,36358,36378,36497,36576,36664,36672,36865,36921,36959,36983,37028,37038,37244,37735,37763,37899,37908,38059,38244,38260,38293,38347,38456,38547,38826,38902,38943,38946,39120,39189,39197,39422,39495,39774,39791,39793,39897,39955,40178,40363,40488,40533,40614,40730,40742,40855,41025,41089,41200,41205,41228,41361,41462,41516,41524,41700,41734,41809,41866,41892,41976,42029,42066,42107,42135,42153,42353,42398,42530,42544,42608,42762,42781,42857,42992,42997,43012,43060,43293,43422,43427,43531,43556,43568,43777,43869,44086,44228,44468,44542,44685,44740,45009,45159,45343,45363,45408,45641,45649,45698,45886,45949,46029,46049,46203,46263,46301,46309,46349,46468,46537,46547,46597,46623,46671,46965,47099,47125,47173,47178,47358,47380,47412,47423,47448,47553,47563,47584,47706,47741,47954,48033,48135,48309,48361,48671,48871,49087,49095,49191,49213,49265,49300,49427,49676,49775,49780,49782,49911,49913,49956,49969,50051,50105,50220,50276,50312,50468,50483,50503,50521,50573,50743,50855,51030,51104,51209,51519,51726,51801,51839,51889,52024,52063,52117,52123,52216,52435,52579,52715,52756,52758,52794,53092,53136,53145,53204,53228,53274,53275,53386,53496,53506,53744,53752,53900,53903,53905,53932,54149,54151,54164,54173,54227,54240,54330,54471,54640,54651,54676,54692,54782,54826,54947,54951,55050,55283,55322,55442,55466,55596,55651,55687,55809,55947,56025,56036,56134,56145,56262,56359,56398,56454,56484,56624,56697,56719,56851,56874,56898,56955,56989,57276,57302,57342,57462,57555,57597,57693,57755,57873,57938,57991,58167,58283,58300,58364,58367,58378,58393,58399,58450,58465,58505,58579,58709,58881,59182,59192,59246,59314,59501,59551,59591,59722,59764,59768,59791,59854,60051,60091,60135,60158,60319,60347,60411,60531,60553,60609,60615,60681,60810,60812,60882,60892,60930,60998,61081,61090,61176,61181,61196,61269,61407,61453,61498,61611,61671,61720,61721,61814,61817,61917,61923,61971,62150,62256,62278,62568,62612,62633,62691,62760,62889,63010,63069,63109,63147,63274,63332,63387,63436,63581,63597,63643,63820,63914,63931,63938,64061,64121,64149,64455,64519,64583,64603,64637,64663,64695,64740,64763,64940,64984,64990,65021,65122,65234,65541,65655,65726,65735,66112,66188,66190,66302,66304,66484,66498,66526,66583,66593,66619,66649,66708,66783,66826,66850,66866,66949,66975,66976,67054,67067,67100,67153,67208,67325,67378,67492,67570,67606,67638,67655,67720,67832,67932,67961,67989,68035,68071,68088,68254,68324,68378,68384,68421,68493,68533,68573,68651,68816,68894,68932,69036,69048,69108,69156,69377,69393,69510,69722,69757,69857,69979,70309,70516,70525,70549,70580,70589,70630,70881,70940,71013,71084,71157,71296,71297,71379,71403,71426,71532,71550,71580,71592,71632,71748,71774,71800,71954,72113,72145,72156,72241,72510,72686,72771,72865,72952,73043,73117,73352,73406,73415,73529,73562,73860,73905,73948,74036,74224,74238,74275,74313,74467,74469,74485,74530,74763,74779,74861,74966,75110,75247,75300,75316,75326,75458,75460,75481,75504,75618,75643,75666,75785,75849,75973,75986,76028,76186,76187,76328,76414,76606,76818,77020,77063,77195,77492,77521,77856,77948,78115,78203,78544,78548,78732,78746,78756,78940,79341,79476,79603,79662,79667,79757,79901,79987,80040,80119,80169,80190,80192,80254,80305,80594,80715,80822,80845,80895,80907,80956,81003,81101,81150,81215,81278,81449,81557,81563,81631,82064,82158,82368,82573,82659,82697,82715,82790,82836,83039,83046,83092,83316,83357,83488,83522,83530,83678,83784,83879,83883,84049,84171,84189,84197,84282,84336,84419,84430,84448,84658,84826,84960,84980,85103,85210,85238,85266,85429,85530,85642,85658,85867,85981,86088,86307,86314,86356,86537,86604,86663,86729,86741,86950,87060,87247,87269,87363,87723,87750,87873,87922,87972,88146,88172,88245,88344,88496,88602,88610,88643,88650,88665,88761,88764,88801,88815,88818,88934,89172,89195,89415,89460,89503,89530,89624,89635,89645,89678,89752,89792,90010,90037,90088,90134,90327,90424,90451,90461,90507,90777,90802,90896,90900,91065,91097,91109,91133,91196,91290,91342,91390,91416,91424,91432,91575,91654,91663,91757,91779,91849,91964,92204,92225,92265,92359,92413,92459,92700,92739,92787,92860,92933,92963,93001,93017,93068,93507,93548,93555,93598,93744,93919,94064,94080,94084,94263,94337,94432,94461,94935,94949,95230,95233,95238,95304,95329,95402,95414,95529,95574,95618,95644,95722,95723,95751,95910,95954,95989,96157,96160,96332,96576,96619,96623,96791,96831,96843,96990,97106,97252,97275,97285,97335,97453,97544,97601,97665,97683,97701,97776,97878,98042,98141,98248,98378,98416,98422,98521,98557,98614,98640,98771,98904,98915,98976,99041,99087,99135,99177,99329,99340,99395,99483,99484,99687,99910,99979",
          "114,135,168,675,689,690,771,796,824,937,1046,1059,1066,1179,1249,1334,1483,1540,1546,1620,1677,1681,1867,1884,1947,2025,2045,2053,2165,2207,2291,2313,2396,2401,2631,2672,2674,2722,2765,2776,2826,2861,2890,2993,3066,3214,3404,3505,3570,3694,3703,3976,3978,3990,4013,4097,4249,4295,4311,4354,4362,4463,4501,4637,4649,4662,4708,4801,4817,4861,4918,4972,5056,5071,5117,5168,5497,5524,5533,5595,5627,5634,5762,6025,6081,6104,6120,6211,6217,6223,6236,6255,6299,6425,6514,6696,6920,7115,7168,7204,7297,7375,7500,7614,7617,7781,7821,7868,7895,7982,8011,8172,8182,8228,8347,8396,8418,8427,8468,8478,8522,8578,8584,8835,8842,9066,9103,9281,9320,9381,9478,9485,9511,9622,9791,10004,10060,10061,10127,10164,10300,10541,10567,10594,10655,10790,10882,10972,11033,11050,11060,11191,11219,11234,11261,11333,11417,11437,11476,11560,11563,11572,11628,11664,11692,11738,12094,12125,12289,12298,12508,12533,12638,12682,12713,12739,13073,13155,13168,13277,13340,13390,13435,13452,13469,13530,13552,13589,13638,13819,13853,13975,14140,14151,14230,14273,14290,14293,14305,14408,14452,14509,14512,14545,14640,14898,14903,14926,14932,14979,15051,15078,15110,15216,15287,15309,15401,15458,15575,15611,15681,15811,15972,16010,16021,16166,16239,16262,16324,16347,16436,16506,16610,16788,16833,16903,16931,16947,17285,17308,17314,17321,17446,17448,17498,17526,17536,17600,17628,17648,17785,17876,17953,17965,17969,17992,18056,18106,18129,18181,18206,18300,18309,18378,18454,18613,18660,18670,18704,18742,18762,18815,18950,19010,19046,19061,19107,19157,19422,19469,19490,19491,19588,19637,19680,19683,19736,19748,19751,19826,19847,19875,19891,19914,20038,20100,20172,20254,20301,20445,20553,20604,20651,20707,20714,20767,20781,20859,20880,20914,20994,21030,21080,21253,21258,21327,21353,21357,21372,21428,21580,21619,21626,21646,21660,21662,21778,21959,22143,22144,22231,22285,22290,22319,22332,22427,22451,22631,22651,22653,22788,22821,22905,22944,23140,23146,23174,23186,23429,23470,23507,23586,23631,23667,23692,23797,23880,23911,23951,23954,23962,24018,24025,24033,24147,24248,24279,24293,24504,24522,24546,24605,24663,24722,24884,24937,25002,25349,25379,25383,25411,25449,25530,25539,25641,25676,25679,25740,25752,25775,25788,25945,25954,25962,26013,26121,26203,26207,26208,26217,26253,26493,26537,26554,26588,26590,26601,26653,26796,26924,26926,26974,27076,27120,27128,27154,27221,27297,27302,27406,27419,27425,27559,27641,27643,27671,27679,27692,27696,27755,27810,27891,27893,27961,28124,28148,28363,28591,28599,28758,28782,28819,28950,28974,29025,29059,29126,29288,29298,29434,29711,29728,29736,29757,29899,29915,29950,29976,30003,30029,30055,30185,30273,30323,30413,30419,30576,30638,30772,30929,30975,30993,31028,31113,31401,31498,31521,31589,31662,31680,31856,31903,31939,31956,32058,32096,32101,32192,32230,32272,32358,32405,32481,32499,32615,32635,32661,32760,32780,33048,33131,33140,33205,33237,33244,33308,33359,33380,33427,33477,33506,33548,33701,33735,33736,33989,34117,34126,34180,34304,34324,34328,34336,34474,34511,34663,34767,34783,34906,34940,34946,35008,35038,35072,35170,35224,35228,35231,35243,35378,35381,35411,35603,35608,35609,35611,35879,35886,35926,35962,36010,36077,36086,36153,36168,36258,36308,36363,36382,36410,36432,36440,36509,36678,36690,36850,36961,36991,37031,37118,37251,37349,37413,37507,37511,37545,37566,37689,37714,37743,37802,37909,38250,38286,38484,38531,38620,38770,38962,38992,39011,39171,39256,39388,39511,39533,39568,39617,39639,39640,39711,39837,40023,40096,40154,40168,40217,40239,40337,40345,40521,40680,40697,40713,40779,40781,41031,41425,41476,41562,41584,41756,41758,41775,42136,42209,42280,42289,42411,42431,42487,42717,42719,42830,42836,42991,43087,43180,43255,43343,43489,43499,43519,43630,43634,43689,43723,43792,43818,43947,44065,44095,44127,44295,44366,44655,44658,44762,44875,44879,44885,44913,44996,45052,45065,45083,45100,45111,45352,45407,45588,45666,45808,45918,45948,45952,46047,46160,46233,46264,46269,46400,46454,46599,46697,46746,46774,46872,46926,46939,47018,47087,47381,47390,47398,47463,47470,47472,47498,47746,47898,47916,47918,47944,48176,48208,48261,48286,48430,48463,48553,48566,48648,48719,48721,48759,48774,48831,48892,48907,48952,48961,48981,49027,49169,49180,49337,49412,49415,49440,49493,49526,49571,49616,49636,49699,49762,49918,50049,50162,50164,50185,50194,50209,50216,50249,50275,50282,50514,50519,50552,50629,50661,50671,50675,51043,51059,51106,51166,51251,51269,51289,51365,51581,51612,51637,51657,51695,51737,51748,52098,52215,52405,52563,52639,52752,52868,52966,52994,53027,53036,53052,53169,53221,53397,53478,53603,53653,53904,53918,53933,54044,54058,54145,54159,54259,54270,54295,54310,54333,54404,54514,54538,54677,54684,54696,54753,54845,54857,55013,55037,55136,55374,55435,55479,55588,55711,55725,55855,55898,55936,55983,56113,56143,56218,56235,56258,56298,56321,56408,56439,56480,56773,57204,57209,57258,57303,57360,57442,57470,57589,57672,57673,57786,57919,57932,57948,57972,58077,58253,58396,58427,58492,58510,58534,58623,58750,58874,58877,58894,58923,59151,59180,59184,59323,59453,59512,59522,59672,59738,59759,59819,59823,59970,60098,60120,60130,60189,60302,60365,60417,60626,60710,60718,60771,60792,61065,61103,61214,61222,61310,61334,61378,61389,61426,61500,61502,61519,61558,61591,61608,61825,61826,62051,62107,62115,62124,62146,62205,62213,62230,62395,62403,62504,62514,62588,62625,62690,62707,62718,62742,62829,63090,63130,63192,63294,63346,63398,63496,63519,63549,63613,63635,63772,63837,63897,63932,64050,64146,64164,64295,64310,64422,64451,64710,64722,64756,64813,64828,64914,65025,65106,65151,65302,65479,65500,65599,65765,65900,65932,65943,66210,66211,66268,66330,66399,66468,66639,66703,66745,66810,66833,66924,66925,66959,66992,67141,67162,67284,67343,67419,67565,67578,67647,67658,67911,67937,68065,68194,68212,68372,68592,68644,68664,68733,68808,68938,68977,69033,69043,69098,69179,69228,69281,69282,69443,69539,69581,69630,69638,69656,69694,69728,69820,69908,69944,69968,70001,70007,70062,70157,70158,70172,70180,70265,70359,70424,70447,70453,70469,70495,70689,70691,70722,70753,70766,70781,70945,71174,71204,71308,71342,71344,71349,71385,71407,71478,71728,71741,71836,71843,71855,71880,71986,72039,72127,72286,72331,72410,72460,72487,72621,72729,72748,72754,72781,72955,73159,73169,73190,73211,73215,73257,73366,73418,73476,73597,73891,73894,73895,73997,74138,74152,74160,74169,74218,74438,74580,74856,74921,74994,75133,75203,75209,75214,75293,75296,75367,75614,75977,75979,76021,76061,76126,76275,76503,76574,76625,76782,76911,76970,77033,77114,77166,77206,77217,77294,77329,77415,77430,77434,77435,77525,77611,77616,77741,77764,77801,77815,77866,77880,77942,77958,78025,78155,78178,78359,78404,78440,78510,78540,78560,78593,78784,78805,78814,78860,78979,78995,79188,79196,79255,79690,79920,79930,79984,79985,80044,80049,80157,80229,80286,80492,80672,80752,80769,80805,80818,81004,81016,81174,81189,81206,81387,81443,81560,81620,81672,81684,81745,81757,81768,81859,82301,82324,82349,82427,82459,82599,82608,82617,82631,82652,82700,82841,82872,82895,83124,83270,83368,83401,83411,83655,83912,83932,83965,84164,84169,84172,84215,84264,84363,84397,84502,84667,84678,84718,84732,84744,84814,84834,84871,84949,84997,85035,85185,85261,85351,85369,85379,85415,85533,85596,85619,85711,85734,85798,85854,85929,86070,86257,86319,86326,86346,86377,86416,86547,86591,86712,86898,86964,87023,87179,87409,87481,87669,87742,87795,87811,87875,87876,88019,88101,88363,88377,88557,88565,88586,88623,88638,88716,88785,88811,88847,88890,88951,89079,89240,89267,89290,89299,89335,89421,89423,89644,89702,89727,89745,89859,89881,90017,90099,90137,90155,90160,90460,90528,90558,90566,90576,90636,90744,90916,90929,90996,91005,91079,91102,91123,91345,91397,91431,91434,91505,91628,91736,91745,91774,91791,91796,91799,91831,92059,92079,92107,92131,92233,92238,92349,92493,92525,92741,92857,92879,92934,93020,93150,93297,93331,93412,93550,93587,93602,93661,93714,93716,93774,93872,93917,94059,94194,94199,94275,94331,94393,94411,94632,94652,94780,94970,94999,95118,95141,95247,95296,95418,95491,95543,95590,95598,95602,95623,95796,95818,95829,95860,95884,95899,95976,95982,96062,96137,96376,96497,96553,96715,96735,96770,96828,96835,96844,96846,96872,96992,96993,97027,97070,97245,97269,97272,97273,97281,97389,97517,97570,97578,97691,97745,97824,97847,98087,98128,98178,98307,98412,98608,98701,98848,98898,98917,99179,99352,99488,99495,99591,99796",
          "32,56,378,620,672,1069,1158,1474,1484,1531,1581,1615,1672,1689,1694,1710,1798,1839,1893,1935,2066,2084,2097,2159,2292,2298,2457,2559,2615,2697,2758,2781,2936,2973,3013,3113,3169,3325,3393,3396,3485,3525,3526,3564,3584,3609,3883,3919,4020,4119,4619,4689,4700,4773,4902,4933,4949,5007,5051,5076,5197,5219,5288,5415,5430,5440,5496,5552,5596,5649,5698,5736,5746,5838,5855,5945,5946,5973,5996,6020,6090,6133,6156,6359,6463,6486,6496,6596,6649,6702,6717,6802,6983,7004,7158,7302,7376,7388,7540,7894,7937,7954,7963,8025,8057,8085,8215,8243,8769,8822,8864,8865,8913,8973,9023,9026,9043,9166,9184,9208,9275,9315,9500,9717,9849,9930,9985,10184,10257,10293,10443,10488,10490,10494,10713,10820,10929,10954,10970,11223,11252,11255,11310,11357,11527,11677,11778,11962,12025,12136,12365,12397,12406,12425,12515,12962,12964,13003,13010,13026,13124,13152,13403,13537,13746,13855,13874,14013,14083,14111,14189,14300,14363,14551,14632,14935,14995,15134,15136,15238,15406,15630,15672,15772,15936,15960,16060,16133,16191,16592,16690,16700,16727,16775,16891,17105,17258,17297,17351,17362,17410,17727,17732,17753,17836,18005,18017,18151,18223,18296,18390,18503,18557,18588,18664,18836,18839,18882,18901,18947,18990,18998,19124,19233,19267,19286,19377,19399,19445,19464,19466,19613,19656,19907,20008,20101,20131,20237,20275,20293,20477,20643,20699,20792,20925,20993,21079,21147,21215,21293,21364,21441,21479,21510,21581,21586,21747,21892,21922,22011,22170,22182,22220,22404,22510,22555,22655,22774,22806,22841,22974,23051,23099,23159,23179,23233,23383,23443,23557,23697,23706,23804,23930,23976,23999,24172,24327,24333,24407,24441,24460,24574,24696,24858,24896,24904,24977,24978,24990,25077,25227,25265,25293,25355,25488,25524,25564,26068,26091,26220,26339,26495,26715,26789,26897,27184,27265,27309,27355,27362,27375,27420,27441,27442,27610,27639,27946,28111,28302,28428,28437,28461,28633,28810,28837,28855,28935,28939,28986,29055,29076,29233,29306,29371,29392,29469,29503,29593,29605,29646,29688,29691,29701,29713,29764,29864,29881,29897,29941,29977,30049,30145,30193,30260,30588,30590,30918,31242,31365,31366,31393,31430,31528,31606,31616,31874,31896,31964,31970,31986,31996,32425,32493,32515,32548,32581,32591,32596,32665,32706,32862,32974,33088,33130,33151,33211,33337,33376,33430,33432,33434,33571,33629,33657,33734,33783,33791,34012,34071,34185,34300,34342,34403,34414,34562,34616,34746,34757,34771,34787,34806,34809,34842,34912,34966,34976,34986,35204,35341,35416,35510,35652,35664,35691,35735,35760,35861,35871,35914,35958,36036,36209,36471,36676,36694,36696,36702,36751,36824,36985,37064,37085,37198,37221,37231,37278,37283,37392,37436,37468,37490,37552,37626,37657,37709,37811,37816,37891,37918,37978,38138,38227,38254,38422,38540,38602,38631,38672,38756,38835,38893,38921,38949,39012,39014,39024,39137,39240,39371,39391,39773,39804,39843,39862,39883,40187,40327,40364,40497,40512,40577,40591,40595,40654,40723,40757,40849,41258,41327,41415,41592,41766,41815,41993,42021,42061,42090,42175,42611,42618,42622,42662,42671,42689,42707,42828,42984,42996,43050,43070,43077,43124,43146,43155,43287,43386,43439,43545,43657,43687,43721,43733,43837,43865,43903,44062,44118,44191,44219,44222,44251,44322,44384,44456,44484,44488,44558,44697,44728,44731,44829,44878,44943,45019,45074,45126,45154,45207,45283,45347,45370,45452,45982,46044,46360,46473,46495,46643,46705,46860,46886,46951,46960,47139,47150,47265,47291,47317,47549,47683,47713,47884,47897,47907,47964,48091,48276,48346,48395,48487,48498,48593,48606,48614,48643,48668,48795,48929,48968,48995,49155,49177,49263,49277,49348,49439,49458,49495,49568,49649,49709,49740,49828,49933,49944,50181,50210,50218,50237,50270,50291,50455,50546,50704,50741,50769,50847,50917,50954,51035,51118,51161,51239,51383,51611,51687,51690,51713,51879,51881,51982,52137,52243,52314,52409,52440,52443,52482,52508,52572,52733,52796,52856,52887,52889,52974,52998,53067,53079,53105,53143,53235,53383,53399,53467,53513,53569,53577,53916,53950,53982,54222,54238,54321,54345,54407,54745,54784,54802,54856,54880,55144,55204,55265,55295,55401,55412,55419,55585,55661,55840,55888,55907,55928,55943,56026,56182,56240,56364,56423,56494,56562,56846,56866,57004,57022,57077,57090,57123,57159,57297,57318,57429,57487,57618,57891,57897,57910,57998,58052,58110,58274,58329,58335,58365,58456,58481,58511,58581,58595,58804,58890,59083,59301,59352,59478,59557,59592,59677,59713,59715,59730,59777,59780,59820,60240,60241,60270,60280,60305,60334,60410,60422,60459,60468,60605,60692,60693,60734,60783,60954,61001,61079,61089,61102,61115,61282,61359,61377,61487,61656,61673,61694,61767,61774,61823,61889,62075,62079,62099,62142,62271,62307,62393,62417,62422,62476,62600,62848,62855,62864,62869,62934,63253,63261,63286,63343,63500,63566,63612,63648,63773,63965,64263,64266,64552,64606,64680,64702,64741,64768,64867,64868,64974,64977,64993,65003,65016,65128,65167,65238,65744,65817,65851,66031,66273,66374,66425,66455,66466,66485,66630,66706,66971,66991,67077,67087,67180,67217,67272,67316,67383,67944,68039,68122,68299,68391,68400,68449,68477,68508,68524,68584,69072,69229,69309,69413,69450,69504,69749,69823,69863,70081,70449,70874,70904,71111,71208,71566,71611,71671,71928,72008,72057,72084,72151,72198,72203,72236,72248,72302,72304,72545,72645,72681,72784,72973,72989,73073,73509,73517,73587,73630,73675,73714,73852,73896,74000,74033,74052,74248,74291,74375,74420,74448,74576,74592,74749,74959,75112,75260,75432,75464,75508,75540,75620,75636,75701,75758,75923,76087,76182,76345,76533,76583,76690,76731,76849,76860,76873,76922,77500,77704,77820,77900,77967,78046,78111,78121,78148,78152,78165,78268,78284,78303,78308,78438,78604,78613,79121,79252,79396,79664,79687,79720,79940,79959,80077,80109,80132,80174,80223,80304,80331,80337,80464,80587,80917,80920,80954,81019,81185,81242,81294,81355,81395,81421,81498,81539,81905,81995,82121,82187,82242,82251,82321,82357,82419,82547,82574,82683,82729,82802,82839,82910,82936,83059,83089,83090,83095,83110,83182,83191,83340,83394,83426,83490,83795,83806,83979,84044,84135,84182,84195,84349,84384,84557,84660,84670,84673,84726,84860,85034,85126,85168,85263,85273,85426,85458,85609,85850,85898,86057,86261,86266,86461,86558,86590,86682,86807,86826,86839,86891,86918,86925,87062,87242,87415,87422,87459,87492,87519,87562,87583,87858,87916,88302,88472,88478,88479,88585,88603,88711,88745,88897,88973,89008,89032,89041,89051,89102,89153,89181,89243,89576,89619,89663,89998,90039,90050,90080,90081,90123,90141,90163,90171,90269,90425,90556,90592,90644,90659,90697,90704,90763,90845,90874,90898,90915,91039,91140,91165,91309,91418,91437,91909,91962,92008,92317,92428,92507,92637,92747,92757,92819,93059,93113,93392,93529,93749,93831,93978,93979,94123,94439,94460,94498,94543,94559,94577,94765,95053,95130,95204,95381,95459,95506,95566,95655,95755,95943,96126,96216,96300,96323,96360,96375,96428,96466,96579,96617,96641,96743,96772,96893,96914,96937,96952,97125,97154,97197,97297,97319,97409,97410,97433,97501,97618,97679,97866,97901,98065,98181,98228,98245,98285,98348,98513,98529,98649,98659,98831,99051,99070,99173,99428,99449,99467,99520,99569,99668,99722,99744,99829,99928",
          "17,240,654,735,786,864,941,976,1072,1097,1369,1397,1706,1738,1885,2041,2072,2144,2162,2204,2445,2554,2592,2683,2716,2828,2996,3076,3195,3333,3370,3380,3472,3519,3596,3639,3724,3827,3831,4026,4027,4163,4179,4335,4348,4573,4610,4628,4657,4687,4770,4809,4815,5021,5048,5130,5154,5221,5543,5588,5666,5679,5688,5691,5708,5883,5917,5963,5984,6180,6206,6225,6319,6565,6567,6653,6789,6794,6815,7035,7082,7098,7142,7173,7192,7193,7300,7399,7424,7450,7585,7651,7675,7804,7806,7845,7878,7903,7962,8051,8186,8196,8525,8533,8546,8563,8619,8728,8785,8987,9155,9299,9308,9314,9452,9522,9727,9811,9830,9875,9904,9926,9957,9973,9975,10048,10119,10135,10223,10280,10319,10335,10340,10438,10486,10654,10755,10783,10802,10845,10851,10898,11378,11486,11498,11528,11576,11622,11638,11656,11669,11746,11776,11802,11911,12051,12074,12090,12108,12141,12330,12434,12611,12667,12802,12884,12965,13016,13125,13243,13331,13345,13401,13448,13509,13580,13775,13801,13827,13941,13992,13997,14044,14064,14182,14183,14251,14255,14403,14550,14642,14725,14765,14767,14773,14942,14977,15013,15055,15083,15142,15191,15241,15313,15327,15487,15614,15718,15721,15834,15996,16035,16108,16125,16144,16147,16252,16293,16345,16359,16431,16450,16484,16545,16573,16628,16695,16881,16968,17044,17094,17162,17189,17219,17395,17488,17539,17692,17813,17903,17955,17957,18031,18091,18093,18302,18392,18401,18433,18474,18501,18551,18714,18733,18767,18834,18845,18854,19086,19260,19590,19636,19669,19700,19745,19767,19777,19783,19797,19845,19897,19925,20108,20114,20184,20191,20214,20354,20402,20437,20544,20624,20705,20861,20949,20975,21056,21100,21116,21154,21249,21289,21354,21467,21534,21707,21711,21732,21756,21761,21781,21838,21954,22037,22059,22060,22119,22125,22179,22223,22346,22559,22565,22597,22602,22604,22616,22620,22633,22636,22784,22849,22914,23021,23046,23135,23187,23257,23261,23265,23295,23324,23412,23415,23456,23609,23613,23645,23780,23820,23949,23985,24061,24065,24134,24241,24350,24384,24394,24509,24552,24564,24606,24632,24643,24711,24743,24779,24864,25033,25069,25115,25120,25291,25297,25308,25332,25443,25495,25512,25522,25528,25897,25947,25986,25993,26050,26055,26120,26150,26154,26202,26212,26251,26257,26277,26510,26575,26613,26633,26708,26738,26757,26886,26925,27019,27038,27048,27126,27139,27192,27454,27521,27562,27576,27581,27635,27815,27847,27906,27987,27990,28117,28119,28135,28165,28228,28325,28466,28508,28541,28569,28595,28604,28703,28717,28783,28928,29071,29200,29232,29241,29376,29629,29667,29767,29796,29806,30019,30051,30071,30136,30141,30155,30182,30218,30748,30793,31005,31014,31107,31134,31177,31192,31201,31226,31291,31389,31418,31470,31522,31548,31556,31714,31717,31837,31888,32059,32119,32144,32254,32288,32433,32462,32479,32560,32607,32660,32722,32804,32823,32932,32986,33039,33263,33283,33325,33379,33651,33895,33905,33941,34094,34201,34207,34219,34228,34240,34290,34340,34372,34592,34667,34720,34730,34818,34854,34952,35020,35099,35105,35116,35389,35417,35425,35446,35545,35549,35750,35754,35769,35923,35953,36032,36035,36196,36241,36249,36265,36275,36305,36328,36371,36379,36686,36729,36737,36753,36790,36796,36893,36911,36949,36982,36999,37250,37259,37265,37335,37434,37448,37462,37478,37482,37510,37577,37620,37720,37749,37886,37896,38381,38544,38613,38781,38820,38945,38964,39005,39269,39280,39377,39385,39424,39445,39526,39548,39590,39611,39650,39656,39709,39839,39887,40066,40208,40221,40230,40253,40295,40299,40508,40674,40688,40748,40845,40878,40880,40918,40923,41070,41090,41419,41569,41725,41732,41742,41904,42060,42073,42082,42129,42236,42316,42331,42488,42527,42640,42667,42670,42672,42673,42681,42685,42713,42815,42835,42873,42884,43231,43233,43238,43268,43276,43305,43423,43426,43554,43697,43736,43745,43898,43927,43932,44056,44106,44128,44175,44207,44303,44367,44439,44610,44630,44656,44693,44806,44888,44923,44978,44992,45022,45103,45104,45142,45163,45257,45329,45332,45356,45458,45482,45496,45720,45870,45930,45955,45985,46156,46249,46286,46319,46321,46346,46398,46511,46525,46532,46552,46651,46664,46689,46700,46707,46756,46849,46937,46986,47063,47171,47174,47452,47589,47627,47644,47786,47788,47810,47985,48058,48129,48195,48527,48531,48540,48672,48720,48737,48780,48835,48843,48897,48997,49008,49110,49137,49278,49325,49396,49525,49586,49595,49683,49746,49862,49894,49955,49960,50077,50169,50354,50444,50621,50648,50690,50777,50861,50898,50959,51086,51296,51327,51333,51506,51966,52040,52081,52095,52206,52361,52413,52439,52543,52573,52658,52677,52765,52824,52923,52936,53016,53260,53276,53349,53459,53571,53591,53638,53746,53794,54014,54154,54183,54364,54371,54392,54506,54555,54616,54632,54981,55034,55095,55220,55329,55375,55383,55672,55685,55796,55949,56010,56122,56127,56133,56230,56271,56282,56326,56342,56386,56491,56509,56625,56695,56998,57253,57286,57321,57337,57474,57497,57528,57639,57734,57822,57963,58028,58055,58086,58100,58301,58320,58366,58404,58431,58447,58454,58515,58565,58594,58642,58658,58753,58811,58837,58907,59062,59067,59090,59194,59210,59220,59231,59316,59433,59558,59601,59612,59625,59654,59659,59816,59870,60047,60081,60095,60103,60409,60530,60545,60674,60686,61029,61142,61216,61565,61595,61635,61677,61686,61934,61981,62022,62087,62100,62121,62337,62516,62564,62582,62798,62818,62859,62882,62970,63050,63083,63132,63172,63183,63190,63322,63384,63428,63431,63529,63651,63694,63704,63829,63834,63848,63874,63937,64009,64180,64282,64338,64483,64621,64762,64800,64885,64909,65050,65110,65183,65322,65332,65366,65681,65712,65789,65923,65941,66051,66136,66154,66182,66184,66232,66272,66544,66681,66817,66829,66945,66946,66970,66996,67007,67117,67362,67407,67430,67456,67590,67595,67615,67652,67659,67687,67791,67811,67842,67847,67894,67928,67954,67973,68137,68196,68392,68471,68549,68554,68709,68750,68778,68805,68853,68892,69024,69028,69181,69186,69200,69313,69323,69367,69385,69478,69494,69603,69648,69716,69731,69753,69763,69973,69976,70077,70293,70348,70416,70498,70529,70538,70568,70604,70694,70701,70744,71020,71068,71390,71483,71486,71515,71660,71725,71771,71803,71932,71994,72049,72068,72187,72234,72271,72281,72309,72318,72324,72341,72375,72384,72432,72458,72547,72600,72601,72776,72794,72805,72821,73015,73047,73087,73270,73275,73312,73368,73512,73538,73569,73620,73759,73835,73920,73938,73999,74028,74063,74111,74322,74351,74503,74581,74633,74818,74837,75029,75117,75129,75138,75166,75225,75239,75245,75317,75321,75361,75446,75497,75527,75585,75590,75818,75836,75906,75983,76023,76079,76085,76138,76202,76545,76558,76641,76653,76769,76827,76828,76938,76991,77138,77247,77278,77507,77589,77694,77715,77736,77877,77890,77955,77982,78428,78621,78630,78658,78729,78734,78818,78871,78912,78950,78978,79067,79150,79209,79256,79368,79465,79507,79508,79531,79606,79626,79685,79746,79749,79978,80133,80363,80367,80373,80527,80564,80585,80619,80658,80676,80710,80761,80825,80974,81042,81250,81254,81364,81448,81655,81665,81767,81829,81987,82053,82077,82102,82211,82222,82241,82274,82356,82363,82447,82481,82605,82607,82657,82755,83010,83072,83074,83080,83114,83136,83176,83183,83250,83409,83446,83511,83527,83668,83718,83733,83740,83761,83855,83969,83994,84056,84206,84299,84534,84596,84778,84862,84898,84968,84972,85025,85049,85060,85079,85247,85566,85723,85736,85767,85915,85991,86011,86099,86198,86211,86296,86429,86503,86509,86626,86651,86722,86727,86733,86957,86958,87049,87054,87144,87173,87291,87377,87382,87439,87513,87559,87688,87794,87842,87903,87917,87981,88008,88016,88037,88055,88135,88228,88263,88265,88276,88390,88606,88710,88804,88852,88958,89074,89083,89098,89219,89304,89307,89326,89462,89691,89724,89732,89816,89830,89937,89993,90029,90093,90096,90205,90391,90395,90567,90590,90612,90723,90738,90815,90861,90886,90960,91041,91252,91515,91620,91644,91735,91824,91914,91965,91982,92232,92266,92385,92396,92442,92452,92478,92581,92600,92603,92630,92808,92870,92988,93060,93146,93324,93330,93372,93492,93500,93506,93531,93565,93619,93682,93733,93818,93884,93898,93951,94060,94186,94215,94360,94366,94588,94598,94802,94822,94863,94897,94971,95042,95108,95272,95297,95324,95370,95400,95478,95612,95620,95624,95905,95950,96005,96067,96077,96159,96254,96270,96369,96439,96543,96638,96685,96695,96781,96813,96899,96956,96984,96986,97093,97126,97317,97338,97550,97674,97706,97722,97775,97814,97816,97918,97967,97972,98016,98171,98253,98287,98294,98296,98310,98316,98360,98413,98498,98641,98711,98742,98818,98824,98836,98865,98874,98951,99205,99323,99343,99414,99542,99550,99586,99663,99703,99737,99739,99902,99973,100003",
          "46,87,150,195,265,326,539,558,682,951,1077,1207,1248,1284,1311,1340,1446,1707,1766,1835,1978,2009,2013,2088,2138,2346,2759,2792,3104,3279,3281,3313,3415,3416,3543,3556,3578,3875,3960,4080,4123,4153,4233,4292,4349,4572,4616,4676,4711,4929,4946,5032,5136,5189,5410,5472,5562,5817,5986,6030,6122,6166,6172,6214,6233,6293,7006,7029,7177,7195,7207,7342,7370,7401,7429,7449,7555,7643,7644,7647,7828,7833,7840,7898,7952,8047,8118,8129,8148,8170,8195,8229,8313,8466,8470,8519,8748,8762,8805,8936,8984,9028,9085,9086,9120,9125,9129,9198,9246,9333,9361,9528,9552,9555,9783,9816,9852,9899,9976,9979,10035,10079,10206,10261,10441,10739,10868,10947,11029,11052,11108,11169,11183,11229,11298,11336,11390,11393,11433,11457,11635,11684,11755,11915,11924,12078,12084,12121,12151,12153,12217,12244,12352,12731,12776,12918,13212,13457,13529,13672,13674,13677,13778,13795,13938,14304,14333,14412,14561,14614,14651,14693,14887,14892,14921,15012,15149,15280,15355,15421,15662,15706,15812,15883,15921,16162,16192,16249,16570,16648,16803,16870,16884,16976,17003,17004,17240,17293,17505,17528,17718,17798,17864,18021,18044,18076,18173,18217,18288,18322,18406,18531,18545,18568,18601,18729,18922,19075,19168,19182,19240,19396,19638,19639,19939,19946,19966,20083,20111,20163,20284,20405,20458,20526,20529,20555,20788,21092,21104,21185,21386,21399,21407,21524,21577,21592,21679,22069,22109,22124,22278,22289,22293,22534,22587,22618,22637,22777,22924,22965,23058,23088,23089,23115,23164,23401,23422,23474,23544,23611,23772,23803,23885,23953,24023,24078,24259,24286,24347,24393,24481,24496,24502,24506,24676,24730,24759,24841,24874,24905,25046,25099,25151,25277,25454,25552,25595,26003,26035,26042,26043,26135,26210,26259,26411,26429,26582,26749,26838,26917,26977,26980,26995,27028,27033,27104,27148,27268,27284,27428,27601,27927,28067,28101,28213,28247,28262,28412,28440,28637,28695,28726,28796,28873,29088,29173,29176,29300,29322,29356,29441,29525,29649,29690,29762,29794,30284,30285,30315,30380,30417,30548,30555,30692,30862,30960,30973,31102,31461,31488,31552,31631,31661,31683,31795,31818,31890,31893,31925,31950,32036,32042,32104,32113,32141,32295,32385,32403,32407,32446,32472,32562,32597,32604,32696,32720,32759,32803,32832,32917,33000,33075,33173,33233,33246,33319,33414,33597,33725,33766,33819,33974,33983,34047,34085,34269,34299,34348,34502,34539,34553,34622,34710,34774,35030,35112,35138,35239,35334,35380,35404,35580,35591,35632,35759,35786,35799,35921,36115,36146,36166,36173,36175,36251,36289,36318,36359,36368,36462,36539,36585,36642,36840,36979,37127,37191,37269,37368,37450,37452,37467,37630,37681,37718,37779,37828,37881,37983,38061,38062,38125,38126,38198,38203,38267,38310,38334,38404,38405,38505,38515,38583,38665,38701,38719,38868,39286,39319,39454,39564,39591,39601,39805,39854,39869,40139,40177,40179,40273,40285,40758,40773,40928,40965,41209,41238,41368,41431,41488,41522,41602,41726,41727,41778,41875,41889,41927,41960,42009,42324,42514,42630,42639,42740,42800,42826,42832,42898,43031,43142,43222,43235,43303,43314,43365,43527,43583,43638,43671,43741,43798,43804,43841,43851,43893,43946,43982,44109,44149,44441,44445,44466,44481,44511,44598,44622,44861,44989,44995,45123,45226,45256,45297,45436,45522,45640,45647,45679,45711,45758,45889,45979,46011,46032,46048,46112,46120,46213,46240,46374,46487,46499,46514,46548,46667,46684,47048,47100,47120,47158,47188,47257,47284,47362,47634,47727,47743,47754,47776,47807,47984,48065,48163,48188,48357,48482,48502,48523,48529,48665,48691,48805,48827,48845,49010,49072,49118,49144,49168,49264,49333,49363,49381,49592,49694,49793,49824,49908,49922,50025,50193,50320,50394,50443,50639,50660,50727,50746,50808,50876,50941,50953,51076,51119,51191,51230,51349,51371,51378,51453,51605,51694,51762,51795,51803,51926,51953,51993,52038,52069,52252,52364,52403,52418,52457,52489,52645,52686,52818,52854,53028,53139,53382,53469,53475,53793,53797,53973,53986,54221,54540,54600,54612,54833,54915,55051,55069,55150,55212,55223,55243,55337,55415,55566,55584,55646,55716,55920,55942,56067,56322,56399,56575,56694,56728,56819,57052,57126,57127,57282,57295,57320,57325,57335,57480,57626,57627,57652,57663,57753,57810,57823,57861,57887,57915,57925,57971,58009,58088,58121,58458,58484,58839,58842,58920,59113,59133,59264,59283,59369,59378,59390,59692,59752,59809,59887,60000,60014,60056,60237,60238,60348,60590,60703,60714,60797,60935,61086,61134,61238,61302,61329,61437,61514,61516,61548,61570,61663,61668,61706,61788,61883,61925,61984,62039,62284,62460,62541,62656,62794,62813,62843,62965,63313,63385,63452,63558,63577,63610,63615,63618,63683,63731,63737,63828,63876,64128,64191,64199,64218,64432,64515,64534,64917,64928,65047,65105,65123,65141,65217,65264,65298,65398,65431,65520,65652,65790,65922,65984,66007,66156,66286,66311,66344,66381,66448,66534,66625,66798,67261,67359,67463,67545,67546,67555,67591,67645,67754,68033,68034,68104,68130,68208,68258,68272,68279,68343,68350,68482,68487,68585,68637,68735,68801,68864,68874,68966,69191,69368,69371,69372,69374,69403,69438,69460,69540,69563,69597,69621,69663,69675,69685,69720,69861,69942,69978,70006,70023,70125,70126,70143,70146,70247,70319,70429,70496,70733,70769,70877,71074,71135,71199,71227,71417,71428,71485,71567,71620,71688,71695,71736,71893,71920,71929,72027,72032,72070,72104,72133,72147,72178,72425,72527,72584,72609,72636,72665,72756,72855,72883,72909,72946,73141,73142,73221,73457,73749,73751,73754,73863,73918,73935,73950,73975,74029,74091,74144,74188,74269,74303,74326,74353,74366,74496,74517,74601,74603,74609,74631,74710,74729,74750,74817,74827,74936,75001,75012,75059,75063,75238,75280,75503,75547,75607,75630,75716,75793,75832,75850,75864,75896,76118,76400,76490,76499,76578,76628,76701,76832,77058,77098,77170,77191,77302,77354,77722,77784,77799,77892,77975,78089,78096,78226,78317,78366,78480,78623,78919,79110,79175,79258,79459,79502,79644,79646,79724,79754,79918,79936,80085,80175,80262,80592,80661,80915,80979,81025,81086,81106,81200,81218,81234,81338,81339,81625,81630,81632,81734,81906,82044,82069,82110,82181,82300,82360,82474,82523,82583,82720,82740,82801,82803,82946,83054,83058,83075,83128,83257,83492,83753,83762,83938,83949,83973,84108,84128,84187,84262,84318,84355,84399,84422,84426,84450,84601,84653,84689,84836,84873,84913,85044,85084,85125,85318,85442,85459,85548,85604,85640,85678,85771,85902,85942,85990,86063,86072,86106,86233,86325,86787,87044,87068,87085,87353,87387,87398,87545,87549,87588,87821,88018,88029,88078,88088,88161,88291,88311,88391,88449,88473,88538,88651,88888,88957,89045,89047,89049,89113,89203,89450,89668,89797,89916,89940,90046,90062,90185,90226,90256,90264,90352,90402,90488,90784,91001,91100,91116,91369,91411,91479,91557,91578,91705,91730,91932,92011,92016,92264,92270,92346,92368,92409,92473,92484,92815,92828,93033,93265,93287,93298,93299,93339,93367,93399,93419,93478,93537,93762,93810,94016,94021,94067,94094,94107,94144,94340,94377,94378,94541,94634,94641,94890,94914,94938,94952,94992,94997,95031,95107,95142,95228,95240,95484,95485,95500,95509,95596,95606,95709,95740,95756,95767,95886,96006,96138,96219,96424,96513,96602,96650,96766,96819,96883,96976,97063,97145,97186,97206,97231,97336,97358,97380,97479,97491,97604,97773,97788,98023,98044,98117,98124,98140,98143,98179,98214,98244,98268,98269,98372,98527,98689,98814,98903,98969,99016,99081,99186,99207,99351,99387,99689,99701,99817",
          "39,161,295,453,631,724,924,952,987,1071,1075,1109,1129,1163,1412,1429,1471,1570,1621,1684,1708,1813,1866,1929,1936,1972,1983,2058,2169,2529,2620,2645,2700,2735,2755,2795,2839,2858,2913,3081,3121,3149,3376,3454,3503,3523,3598,3658,3753,3778,3884,3924,3948,4065,4203,4235,4469,4489,4530,4544,4591,4609,4633,4643,4819,4847,5009,5045,5293,5305,5381,5395,5441,5535,5576,5593,5661,5665,5742,5814,5823,5941,6008,6017,6083,6152,6199,6244,6396,6469,6488,6547,6613,6828,6877,6892,7018,7080,7137,7241,7248,7282,7308,7320,7339,7351,7454,7482,7517,7530,7565,7583,7657,7815,7847,7867,7902,8036,8160,8201,8314,8330,8503,8639,8690,8747,8889,8896,9047,9261,9295,9414,9429,9445,9521,9734,9800,9847,9858,9885,9922,9928,10136,10168,10186,10197,10341,10355,10368,10508,10528,10578,10728,10787,10835,10907,10974,10996,11012,11022,11037,11044,11084,11147,11198,11279,11281,11404,11542,11587,11602,11771,11978,12010,12160,12236,12278,12392,12488,12490,12513,12693,12814,12815,12947,13106,13379,13386,13387,13412,13514,13836,13919,13931,14000,14148,14167,14503,14523,14556,14563,14585,14596,14806,14832,15030,15091,15260,15301,15353,15381,15392,15425,15456,15534,15648,15793,15794,15836,15909,16199,16301,16475,16512,16682,16698,17010,17235,17391,17399,17429,17618,17673,17681,17702,17954,17982,18035,18135,18152,18162,18176,18187,18251,18512,18534,18799,18842,18849,18877,18909,18961,19099,19123,19178,19209,19224,19274,19298,19406,19413,19726,19758,19864,19922,19923,20124,20149,20213,20276,20340,20441,20462,20615,20616,20654,20670,20690,20715,20789,20839,20853,20902,21066,21084,21223,21224,21329,21535,21561,21583,21639,21641,21673,21703,21830,21836,21840,21887,21891,21981,22013,22027,22161,22253,22300,22466,22639,22641,22978,23110,23219,23338,23339,23378,23450,23496,23514,23545,23683,23904,23928,23991,24020,24068,24410,24426,24532,24572,24596,24633,24661,24673,24707,24752,24800,24967,25020,25167,25194,25329,25388,25808,25835,25880,25934,25980,25984,26177,26205,26242,26464,26484,26530,26593,26600,26606,26621,26622,26624,26684,26719,26785,26824,26869,27020,27107,27125,27260,27274,27282,27328,27408,27525,27626,27719,27736,28047,28269,28289,28296,28359,28367,28381,28543,28585,28638,28786,28827,28877,29009,29017,29110,29121,29175,29250,29261,29293,29480,29841,29895,29921,29925,29946,30025,30041,30047,30177,30517,30553,30672,30765,30881,30932,30947,31041,31114,31157,31190,31229,31252,31420,31590,31608,31778,31863,32138,32172,32203,32317,32470,32489,32542,32613,32628,32643,32663,32731,32746,32790,32914,32944,33017,33043,33141,33285,33490,33589,33599,33757,33988,34087,34410,34449,34489,34499,34586,34647,34706,34798,34831,34942,35083,35257,35395,35481,35571,35633,35679,35685,35725,35773,35793,35941,35964,36008,36087,36157,36215,36338,36386,36411,36414,36452,36495,36516,36549,36631,36649,36778,36815,36842,36870,36874,36897,36955,36984,37056,37170,37175,37268,37533,37596,37670,37698,37770,37776,37807,37888,38018,38188,38191,38325,38344,38367,38467,38555,38886,38887,38956,39083,39091,39096,39211,39214,39302,39589,39633,39825,39852,39858,39920,39981,40105,40155,40246,40270,40381,40454,40616,40837,40888,40917,40952,41000,41106,41192,41269,41313,41322,41487,41491,41586,41968,42018,42020,42088,42160,42162,42302,42349,42385,42519,42547,42935,42949,42985,43085,43115,43149,43668,43831,43843,43850,43918,44021,44063,44092,44153,44229,44256,44298,44436,44483,44489,44571,44618,44777,44856,44899,44928,44998,45105,45171,45235,45236,45277,45375,45395,45465,45510,45795,45809,46088,46388,46390,46420,46427,46496,46540,46553,46585,46695,46921,47020,47119,47136,47154,47473,47490,47521,47585,47595,47845,47846,47929,47938,47991,48123,48243,48295,48393,48409,48542,48755,48851,48866,48937,48973,49200,49217,49329,49466,49604,49606,49645,49898,50036,50130,50160,50212,50401,50406,50437,50687,50812,50848,50854,50994,51005,51130,51193,51217,51220,51231,51249,51362,51451,51571,51606,51747,51781,51859,51869,51934,52045,52096,52246,52256,52302,52354,52412,52425,52433,52436,52444,52735,52740,52745,52791,53011,53034,53056,53183,53327,53408,53452,53564,53776,53777,53804,53815,53859,54007,54181,54216,54286,54439,54501,54531,54544,54599,54724,54810,54852,54917,54955,55008,55170,55340,55520,55595,55644,55995,56051,56075,56120,56131,56330,56378,56500,56561,56564,56590,56594,56727,56796,56815,56912,57059,57105,57145,57430,57509,57578,57727,57797,57871,57922,58023,58154,58332,58336,58352,58363,58429,58482,58500,58646,58754,58779,58862,58897,58900,58963,59147,59155,59191,59393,59508,59542,59631,59675,59725,59846,59907,59924,60017,60345,60353,60398,60416,60543,60572,60646,60878,60895,61006,61052,61070,61097,61173,61215,61284,61348,61350,61434,61479,61581,61600,61722,61859,62040,62056,62167,62216,62234,62379,62394,62455,62463,62545,62553,62594,62739,62770,62778,62808,63068,63146,63185,63250,63310,63665,63743,63833,63854,64052,64129,64577,64622,64714,64736,64755,64889,64980,65007,65009,65076,65078,65082,65126,65169,65218,65323,65488,65590,65676,65706,65947,66146,66172,66227,66257,66314,66529,66550,66555,66561,66570,66845,66978,67170,67207,67218,67443,67484,67639,67693,67732,67897,67923,67936,68261,68437,68438,68523,68540,68704,68728,68837,69160,69225,69226,69251,69445,69531,69536,69623,69649,69759,69772,69842,69940,70020,70128,70351,70383,70462,70637,70798,70809,70953,71000,71017,71113,71128,71225,71230,71264,71388,71445,71510,71587,71706,71769,71815,71822,71859,71890,71933,71958,72033,72089,72154,72193,72427,72482,72618,72651,72742,72777,72873,72895,72907,72912,73018,73028,73217,73334,73375,73420,73503,73527,73598,73726,73766,73776,73796,73832,73907,73951,74129,74208,74449,74451,74662,74719,74841,75026,75094,75102,75453,75608,75675,76004,76031,76063,76210,76225,76230,76278,76288,76291,76317,76403,76448,76540,76718,76738,76889,76901,76931,77043,77066,77069,77160,77260,77326,77607,77650,77906,77913,78151,78184,78526,78549,78757,78793,78844,78956,78991,79054,79056,79218,79363,79380,79506,79534,79578,79785,79803,79875,79900,79942,79989,80008,80080,80128,80247,80450,80654,80760,81028,81134,81152,81203,81258,81293,81406,81497,81658,81717,81776,81784,81835,82011,82061,82065,82099,82502,82519,82529,82537,82650,82667,82703,82714,82719,82796,82875,82880,83019,83145,83203,83211,83454,83576,83734,83777,83850,83999,84168,84223,84265,84271,84328,84474,84493,84501,84556,84631,84756,85019,85067,85077,85143,85237,85244,85341,85438,85595,85662,85963,86005,86082,86131,86281,86358,86486,86629,86842,86867,86870,86967,86991,87110,87452,87465,87476,87506,87584,87687,87857,87958,87966,87977,88038,88112,88164,88217,88357,88387,88455,88534,88567,88570,88740,88889,88905,88928,88982,89193,89209,89488,89498,89570,89604,89642,89742,89902,89929,90094,90129,90150,90230,90257,90343,90347,90368,90426,90474,90503,90557,90647,90735,90758,90768,90862,91019,91038,91136,91283,91291,91318,91382,91427,91633,91739,91804,91977,91988,92032,92047,92089,92586,92864,93024,93122,93273,93359,93553,93742,93752,93755,93890,93968,93988,93990,94006,94114,94148,94308,94387,94513,94525,94550,94584,94611,94745,94867,94884,95065,95097,95159,95170,95283,95327,95382,95515,95519,95559,95595,95652,95878,95879,95921,95978,95998,96018,96032,96064,96081,96088,96121,96135,96191,96204,96231,96473,96571,96605,96919,96939,97182,97212,97233,97235,97257,97655,97779,97795,97874,97876,97946,98039,98131,98142,98505,98522,98737,98806,98835,99100,99142,99172,99180,99215,99244,99266,99304,99328,99419,99425,99455,99464,99508,99658,99670,99676,99764,99932",
          "25,91,106,280,369,549,622,707,744,929,972,1038,1099,1151,1198,1218,1239,1374,1382,1476,1509,1675,1691,1714,1720,1730,1810,1847,1873,1911,1919,1951,2024,2044,2134,2233,2241,2316,2390,2458,2478,2502,2551,2706,2857,2874,2881,2899,2906,3255,3310,3337,3347,3397,3488,3640,3756,3773,3859,3962,4049,4061,4106,4129,4135,4136,4155,4223,4237,4317,4400,4431,4470,4485,4552,4577,4677,4706,4712,4769,4800,4869,4971,5253,5330,5659,5662,5907,5977,6013,6099,6107,6171,6226,6297,6316,6373,6608,6751,6784,6810,7042,7064,7079,7162,7270,7359,7469,7507,7653,7758,7764,7786,7826,7888,8339,8373,8410,8415,8448,8500,8501,8514,8626,8630,8683,8693,8858,8881,8977,9093,9167,9171,9232,9300,9345,9444,9457,9479,9547,9613,9623,9685,9715,9815,9828,9970,9994,10134,10479,10493,10503,10511,10524,10624,10656,10673,10675,10692,10723,10749,10839,11061,11068,11165,11232,11268,11297,11491,11601,11680,11697,11705,11723,11847,11851,12026,12067,12093,12149,12156,12339,12340,12382,12414,12590,12653,12749,12779,12785,12811,12907,12921,12966,12988,13005,13034,13043,13092,13160,13181,13191,13202,13220,13265,13333,13567,13676,13734,14004,14049,14314,14475,14488,14530,14533,14660,14744,14826,14857,14937,15032,15169,15177,15180,15232,15246,15291,15317,15333,15336,15592,15653,15741,15754,15757,15805,15867,15891,15952,15969,15986,16005,16131,16156,16193,16212,16275,16540,16656,16749,16849,16863,16928,16962,17025,17036,17075,17136,17184,17205,17286,17452,17518,17562,17779,17806,17878,17910,18054,18123,18214,18400,18412,18426,18481,18484,18582,18662,18711,18786,18869,18870,18885,19059,19069,19083,19129,19144,19248,19301,19494,19868,19963,20045,20091,20102,20142,20303,20348,20400,20461,20470,20571,20608,20632,20727,20754,20765,20850,20863,20888,20932,21027,21071,21113,21162,21184,21220,21284,21288,21295,21297,21356,21393,21395,21461,21572,21601,21715,21727,21771,21776,21880,21898,21972,22142,22160,22306,22324,22410,22487,22518,22644,22728,22897,22902,23136,23278,23332,23574,23655,23665,23674,23735,23788,23795,23918,23942,23964,24010,24155,24280,24370,24391,24404,24409,24471,24497,24708,24788,24826,24876,24894,24984,25170,25319,25343,25351,25472,25490,25542,25549,25797,25877,25900,26127,26136,26164,26173,26279,26302,26387,26399,26447,26515,26531,26642,26666,26683,26696,26703,26790,26793,26937,27252,27479,27539,27607,27733,27751,27902,27943,27947,27964,28020,28125,28211,28223,28337,28374,28395,28469,28582,28606,28744,28770,28780,28862,28891,29151,29166,29185,29191,29214,29239,29256,29393,29404,29415,29432,29442,29631,29655,29708,29906,29924,29932,29945,29985,30030,30107,30130,30146,30189,30226,30255,30265,30392,30411,30457,30498,30532,30558,30578,30584,30721,30728,30792,30916,30959,30971,31054,31098,31167,31173,31455,31459,31471,31510,31579,31653,31713,31785,31937,32013,32043,32079,32214,32223,32268,32450,32518,32583,32623,32653,32700,32743,32770,32777,32792,32816,32824,33269,33282,33582,33705,33718,33784,33843,33884,33950,33970,34032,34121,34135,34286,34313,34371,34413,34477,34527,34540,34702,34808,34975,35288,35296,35441,35443,35444,35484,35564,35680,35692,35820,35896,36183,36206,36225,36316,36346,36463,36498,36505,36523,36534,36698,36755,36758,36883,37026,37062,37100,37108,37310,37328,37477,37583,37646,37675,37690,37806,37995,37997,38067,38172,38178,38184,38206,38397,38450,38508,38607,38681,38688,38733,38822,38828,38847,38853,38865,38900,38961,39002,39003,39015,39060,39132,39135,39160,39221,39225,39311,39324,39483,39555,39584,39629,39632,39732,39868,39972,39987,40011,40052,40198,40232,40263,40292,40479,40526,40562,40582,40585,40776,40819,40860,40890,40907,41072,41141,41148,41150,41226,41337,41399,41424,41469,41507,41528,41891,41933,41950,42013,42051,42074,42246,42259,42305,42469,42522,42699,42787,42818,42824,42859,42876,42879,42920,43044,43074,43092,43169,43197,43417,43425,43450,43454,43477,43491,43510,43599,43607,43647,43716,43801,43809,43811,43876,43973,44054,44090,44116,44125,44190,44294,44330,44401,44472,44496,44519,44576,44601,44632,44674,44773,44791,44836,44921,44941,45153,45183,45215,45242,45300,45303,45322,45421,45434,45484,45501,45819,45943,46087,46223,46242,46252,46325,46336,46397,46518,46692,46724,46731,46831,46958,46979,47076,47102,47287,47288,47353,47415,47495,47825,47866,47886,47997,48004,48134,48177,48194,48272,48307,48343,48359,48545,48583,48625,48660,48702,48712,48807,48821,48848,49014,49111,49129,49150,49194,49500,49585,49615,49720,49860,49881,49924,49967,49987,50075,50250,50433,50606,50618,50758,50788,50925,50939,51218,51475,51484,51509,51656,51770,51774,51789,51828,51854,51945,51957,52254,52445,52521,52607,52656,52814,52833,52955,53071,53115,53120,53326,53357,53533,53534,53572,53622,53704,53738,53805,53898,53942,53993,54085,54250,54332,54353,54438,54443,54486,54521,54648,54682,54716,54738,54781,54817,54949,54956,55025,55082,55096,55217,55254,55302,55474,55608,55881,55890,55940,55955,56164,56353,56434,56455,56627,56666,56687,56824,56885,56931,57151,57255,57257,57270,57380,57441,57518,57542,57585,57655,57743,57811,57821,58059,58305,58312,58424,58453,58546,58555,58764,58814,58818,58836,58866,58916,58997,59001,59008,59026,59058,59064,59217,59255,59572,59747,59788,59847,59881,59908,59990,60162,60169,60207,60226,60255,60267,60272,60571,60592,60606,60666,60823,61112,61182,61325,61373,61491,61630,61667,61773,61795,61861,61950,61969,62018,62057,62063,62065,62235,62286,62322,62540,62665,62674,62768,62780,62812,62838,63053,63073,63087,63095,63097,63177,63462,63562,63589,63776,63807,63906,63933,64100,64343,64370,64596,64653,64794,64862,64897,65061,65220,65237,65271,65555,65629,65717,66161,66162,66175,66207,66218,66223,66293,66342,66348,66359,66389,66409,66430,66605,66771,66822,66870,66888,66921,66958,67110,67167,67210,67302,67366,67529,67624,67630,67700,67778,67839,67884,68086,68138,68142,68211,68315,68456,68460,68463,68766,68780,68921,68927,68930,68956,69078,69147,69172,69212,69223,69310,69417,69533,69592,69613,69831,69937,70039,70052,70121,70124,70167,70188,70206,70474,70522,70607,70861,70902,70928,71003,71054,71252,71312,71406,71472,71547,71601,71608,71784,71881,71891,71937,71983,72163,72323,72325,72328,72406,72642,72677,72778,72874,72882,72956,72984,73012,73040,73277,73300,73551,73615,73678,73765,73797,73817,73886,73910,73968,73981,74047,74151,74158,74168,74198,74217,74222,74379,74405,74429,74473,74572,74639,74738,74810,74900,74905,74949,75033,75174,75356,75513,75539,75600,75639,75719,75753,75833,75885,75922,76262,76279,76315,76327,76341,76381,76430,76436,76491,76643,76694,76841,77103,77147,77200,77263,77348,77397,77591,77662,77775,77816,77908,77963,78002,78043,78085,78128,78207,78454,78499,78653,78687,78739,78796,78848,78917,78939,79007,79019,79065,79174,79180,79230,79246,79253,79333,79392,79415,79450,79454,79540,79553,79609,79788,79858,79974,80083,80160,80164,80201,80348,80602,80615,80625,80628,80649,80653,80680,80684,80779,80817,80823,80911,80983,81085,81099,81173,81305,81331,81336,81423,81502,81577,81797,81823,81904,81974,81998,82082,82125,82135,82170,82189,82205,82339,82377,82456,82566,82577,82860,83037,83078,83087,83102,83141,83157,83187,83192,83213,83222,83474,83484,83563,83620,83699,83790,83863,83894,84063,84292,84427,84436,84595,84674,84677,84753,84987,85342,85436,85454,85485,85495,85598,85875,85962,86027,86154,86194,86196,86197,86240,86251,86274,86279,86347,86359,86402,86448,86533,86658,86748,86767,86775,86793,86902,86962,86993,87249,87272,87297,87298,87378,87431,87436,87451,87534,87845,87911,87929,88022,88272,88284,88340,88354,88444,88487,88609,88668,88677,88717,88738,88754,88767,88948,88988,89010,89155,89367,89407,89434,89457,89497,89514,89549,89553,89560,89630,89753,89784,89800,89854,89882,90229,90232,90252,90275,90306,90448,90508,90520,90690,90781,90877,91067,91069,91081,91085,91114,91130,91137,91198,91297,91305,91449,91458,91530,91536,91691,91815,92106,92323,92362,92585,92664,92844,92900,92957,92961,93044,93069,93285,93290,93401,93466,93515,93518,93556,93559,93586,93688,93730,93876,93886,94020,94070,94299,94333,94401,94479,94614,94681,94686,94706,94904,94964,95124,95348,95373,95386,95555,95593,95679,95746,95826,95872,95916,95970,96080,96086,96412,96447,96458,96501,96633,96818,96888,96903,96940,96982,96996,97020,97076,97183,97229,97244,97274,97302,97330,97356,97378,97525,97556,97597,97687,97815,97841,97849,97911,97921,97986,98109,98213,98234,98288,98309,98534,98630,98813,98919,98992,99154,99213,99350,99438,99522,99580,99595,99724",
          "18,35,110,125,181,297,403,504,533,646,676,821,831,863,975,1057,1145,1152,1206,1331,1381,1438,1440,1473,1647,1660,1729,1790,1791,2031,2059,2063,2081,2205,2494,2597,2614,2693,2847,2897,3101,3242,3322,3357,3445,3544,3620,3664,3968,3979,4167,4211,4251,4282,4410,4424,4490,4560,4606,4732,4852,4860,4926,4928,5073,5179,5321,5361,5416,5602,5685,5737,5809,5857,5976,5990,6040,6094,6146,6348,6417,6445,6566,6579,6690,6700,7021,7070,7143,7175,7224,7303,7354,7610,7622,7679,7711,7824,7980,8254,8256,8326,8335,8526,8559,8610,8635,8659,8789,8800,8838,8841,8876,8950,9006,9019,9062,9095,9168,9252,9357,9896,10083,10115,10188,10294,10332,10359,10411,10474,10550,10651,10799,10806,11130,11231,11356,11410,11430,11490,11585,11695,11865,11886,11926,12098,12258,12276,12348,12394,12526,12586,12598,12633,12715,12724,12728,12823,12853,12919,13024,13126,13266,13273,13364,13419,13659,13752,13877,13921,13979,14234,14244,14245,14432,14474,14557,14695,14711,14737,14786,14819,14997,15210,15432,15435,15437,15438,15480,15538,15782,15876,15971,15984,15988,16012,16307,16539,16733,16784,16831,16843,16902,16906,17082,17112,17274,17361,17365,17382,17479,17556,17827,17848,17896,17927,18079,18094,18199,18350,18409,18490,18560,18563,18644,18674,18781,19021,19091,19117,19279,19372,19510,19561,19584,19621,19945,20024,20056,20231,20295,20350,20392,20639,20683,20823,20838,20855,20979,21004,21230,21355,21403,21431,21473,21513,21546,21573,21621,21623,21734,21819,21846,21931,21967,21987,22003,22034,22199,22464,22497,22731,22751,22767,22800,22882,22884,23100,23166,23266,23468,23494,23659,23699,23776,23870,23881,23894,23966,23969,24000,24004,24075,24272,24309,24311,24418,24419,24680,24762,24802,24828,24934,24945,25131,25404,25473,25517,25892,25972,26015,26065,26313,26372,26725,26741,26806,26835,26985,26987,27047,27073,27096,27124,27143,27348,27444,27508,27520,27563,27592,27623,27765,27884,27916,27959,28016,28065,28077,28321,28421,28444,28462,28663,28712,28866,28997,29010,29201,29321,29341,29370,29388,29455,29542,29547,29585,29822,29852,29871,29901,29903,29942,29969,29973,29978,30002,30175,30201,30557,30655,30802,30828,30953,30963,31090,31133,31186,31197,31309,31583,31727,31865,31899,31909,32244,32297,32449,32586,32608,32610,32682,33023,33032,33090,33098,33224,33231,33311,33404,33406,33431,33551,33663,33731,33785,33962,33971,34081,34159,34226,34402,34666,34684,34792,34858,35017,35161,35172,35195,35236,35263,35307,35429,35494,35496,35523,35544,35837,35839,35888,35954,36017,36287,36307,36416,36437,36442,36603,36710,36775,36835,36854,36876,36975,37104,37117,37140,37144,37176,37474,37568,37677,37772,37946,37999,38036,38129,38197,38247,38288,38390,38418,38444,38702,38786,38811,38827,38866,38894,39039,39176,39209,39297,39308,39434,39574,39876,39895,40036,40111,40160,40188,40441,40465,40499,40581,40625,40641,40687,40729,41098,41113,41211,41285,41358,41604,41605,41638,41772,41821,41859,41885,41928,42122,42147,42261,42287,42343,42365,42498,42629,42727,42927,42942,42944,42960,42970,43086,43089,43150,43345,43415,43472,43502,43517,43529,43624,43739,43758,43799,43810,43832,43937,44011,44061,44081,44167,44217,44218,44349,44354,44597,44709,44754,44790,44848,44855,44886,45106,45124,45160,45202,45403,45422,45521,45658,45742,45796,45920,45942,45945,46407,46456,46507,46625,46742,46806,46889,47059,47243,47273,47301,47305,47334,47432,47469,47513,47522,47638,47655,47693,47860,47926,47992,47999,48066,48302,48440,48577,48675,48725,48757,49057,49069,49104,49187,49220,49243,49315,49341,49445,49549,49556,49612,50197,50302,50416,50429,50438,50605,50703,50956,50976,51131,51370,51429,51477,51500,51661,51704,51811,51885,51888,51969,52002,52116,52202,52263,52337,52369,52466,52699,52813,52817,52880,52960,53312,53325,53422,53438,53485,53495,53507,53578,53581,53584,53628,53655,53695,53786,53835,53880,53958,53989,54000,54167,54248,54391,54488,54519,54592,54672,54769,54825,54840,54868,54881,54984,55003,55056,55172,55178,55200,55269,55320,55418,55421,55467,55477,55562,55668,55758,56116,56117,56286,56350,56470,56582,56614,56623,56655,56768,56793,56988,57058,57080,57298,57323,57466,57517,57587,57641,57691,57695,57817,57898,58004,58015,58030,58072,58293,58395,58438,58543,58545,58605,58636,58768,58912,58930,58937,58999,59050,59251,59261,59302,59447,59473,59594,59650,59732,59748,59761,59980,60001,60024,60033,60045,60061,60083,60144,60339,60499,60537,60678,60766,60909,60921,60931,60968,60981,61000,61267,61315,61356,61442,61445,61582,61632,61685,61703,61791,61805,61911,61993,62055,62109,62163,62183,62357,62397,62525,62689,62748,62851,62880,62952,62982,63060,63123,63145,63276,63288,63376,63700,63822,63989,64244,64318,64424,64428,64466,64521,64618,64645,64737,64821,64884,65066,65124,65129,65266,65377,65408,65539,65549,65670,65679,65784,65797,65858,66000,66085,66310,66326,66391,66396,66508,67105,67114,67262,67318,67356,67389,67877,68125,68259,68282,68311,68341,68402,68423,68534,68737,68832,68900,68940,69019,69290,69318,69404,69449,69529,69692,70234,70278,70638,70931,71096,71158,71284,71384,71491,71492,71622,71654,71744,71780,71970,71991,72012,72054,72132,72412,72480,72564,72580,72702,72798,72806,73059,73076,73112,73383,73443,73445,73489,73591,73623,73674,73805,73821,73986,74031,74128,74175,74204,74246,74250,74323,74342,74350,74499,74560,74734,74811,74836,74935,75030,75076,75165,75193,75222,75378,75416,75430,75471,75480,75592,75611,75655,75692,75698,75831,75956,75969,76117,76238,76349,76391,76423,76534,76585,76695,76762,76942,76955,76974,77064,77068,77118,77164,77181,77213,77248,77301,77459,77503,77553,77679,78000,78044,78072,78175,78290,78292,78324,78370,78385,78393,78591,78612,78625,78646,78820,78859,78998,79443,79527,79542,79598,79618,79898,80065,80071,80076,80092,80124,80131,80225,80230,80255,80338,80339,80439,80546,80576,80745,81000,81050,81063,81105,81147,81239,81351,81482,81697,81751,81796,81820,81992,82032,82056,82367,82380,82552,82620,82691,82696,82759,82792,82903,82931,82964,82965,83012,83133,83421,83441,83602,83709,83719,83752,83884,84036,84083,84173,84229,84454,84457,84649,84680,84725,84867,85000,85062,85154,85222,85452,85608,85655,85706,86024,86173,86227,86258,86485,86602,86612,86659,86683,86697,86702,86795,86973,87041,87130,87299,87323,87587,87612,87615,87627,87789,87792,87803,87870,88054,88098,88200,88247,88258,88333,88462,88562,88690,88697,88783,88841,88893,88980,89006,89180,89196,89394,89436,89448,89466,89607,89719,89733,89749,89760,89861,89875,90013,90091,90139,90190,90216,90345,90370,90387,90502,90658,90674,90679,90859,90899,90985,90989,91025,91086,91095,91131,91142,91175,91197,91246,91269,91366,91413,91419,91439,91473,91478,91697,91767,91770,91832,91854,92066,92554,92621,92658,92767,92813,92822,92880,93334,93337,93446,93454,93543,93601,93709,93786,93809,93918,94153,94267,94303,94533,94571,94594,94605,94677,94689,94754,95056,95155,95177,95213,95252,95389,95462,95488,95501,95505,95603,95789,95795,95858,96076,96140,96197,96199,96282,96327,96382,96471,96532,96792,97105,97191,97230,97342,97387,97548,97615,97616,97682,97690,97744,98075,98125,98151,98176,98239,98261,98395,98503,98569,98590,98646,98661,98838,98849,98861,98882,99020,99140,99288,99541,99702,99781,99993",
          "44,116,163,214,267,323,361,528,599,609,638,650,714,1096,1222,1246,1279,1293,1337,1358,1552,1560,1561,1564,1646,1752,1894,1944,2107,2117,2151,2246,2302,2344,2354,2504,2593,2635,2676,2786,3030,3196,3265,3336,3719,3725,3746,3793,3848,3871,3923,3941,4089,4111,4128,4138,4217,4252,4285,4376,4389,4472,4494,4512,4727,4806,4826,4923,4945,4982,5030,5205,5236,5392,5413,5503,5751,6106,6176,6254,6371,6460,6480,6492,6516,6577,6682,6807,7026,7209,7271,7335,7402,7536,7552,7710,7794,7855,8276,8403,8445,8449,8455,8512,8680,8784,8833,8879,8880,8927,8982,9152,9160,9176,9178,9277,9287,9397,9509,9983,10028,10260,10325,10334,10354,10387,10392,10475,10514,10585,10693,11088,11172,11201,11209,11215,11240,11245,11331,11388,11419,11446,11474,11598,11712,11716,11775,11884,11904,12042,12210,12265,12281,12373,12399,12452,12592,12628,12672,12691,12786,12813,12820,12892,12969,12984,13093,13193,13321,13650,13658,13670,13767,13814,14032,14105,14143,14191,14250,14345,14381,14616,14625,14686,14700,14808,14941,15026,15054,15085,15274,15337,15700,15751,15774,16007,16033,16044,16259,16325,16337,16399,16490,16526,16542,16558,16741,16878,16954,17009,17084,17196,17199,17265,17268,17378,17381,17511,17598,17651,17674,17747,17770,17849,17862,17909,17950,18003,18032,18074,18090,18244,18252,18260,18349,18475,18573,18595,18604,18709,18808,18833,18846,18925,18943,18957,18982,19203,19204,19412,19433,19504,19701,19739,20050,20099,20195,20255,20321,20417,20449,20456,20523,20534,20586,20613,20720,20732,20783,20988,21026,21127,21397,21555,21637,21692,21738,21767,21786,21884,21941,21990,22077,22257,22287,22438,22455,22516,22572,22625,22660,22686,22688,22856,22895,22929,22941,23380,23538,23739,23888,23892,24043,24240,24449,24465,24567,24629,24649,24740,24805,24827,24845,24959,25008,25050,25052,25110,25142,25144,25236,25239,25289,25340,25444,25447,25556,25692,25769,25790,25818,25901,26123,26315,26520,26541,26760,26930,27075,27097,27153,27164,27690,27786,27812,27898,27905,27930,27985,28011,28163,28214,28385,28464,28512,28580,28755,28929,29078,29089,29107,29111,29292,29420,29590,29626,29643,29761,29857,30035,30060,30150,30161,30199,30369,30562,30643,30738,30887,31055,31103,31104,31146,31154,31174,31195,31276,31300,31301,31364,31437,31574,31784,31878,31892,31921,31994,32277,32492,32546,32860,32960,33007,33286,33334,33574,33590,33614,33745,33812,33851,33877,33931,33955,34172,34309,34354,34386,34425,34427,34433,34506,34546,34609,34705,34717,34833,34958,35071,35078,35147,35324,35365,35375,35455,35460,35507,35574,35855,35948,36042,36064,36101,36357,36391,36445,36456,36587,36901,36912,36997,37020,37032,37066,37230,37271,37288,37329,37357,38161,38229,38308,38322,38338,38351,38402,38412,38428,38504,38647,38648,38687,38752,38810,38872,38881,39108,39342,39615,39625,39688,39741,39753,39782,40083,40129,40196,40290,40474,40520,40634,40805,40810,40813,40848,41103,41142,41182,41428,41442,41666,41677,41733,41853,41954,42055,42270,42394,42437,42504,42563,42741,42882,42893,43005,43036,43039,43042,43262,43410,43587,43601,43605,43766,43780,43855,43912,43917,44001,44045,44049,44130,44171,44187,44194,44225,44275,44296,44333,44521,44553,44671,44732,44776,44864,44900,45013,45039,45056,45086,45114,45165,45334,45481,45548,45558,45564,45728,45841,45863,45896,45912,45939,46013,46100,46274,46292,46413,46520,46610,46637,46739,46784,47071,47354,47359,47372,47529,47530,47573,48109,48161,48204,48365,48389,48444,48650,48853,48872,48918,49034,49054,49085,49108,49206,49228,49321,49394,49404,49424,49433,49504,49608,49800,49850,49932,50002,50007,50129,50134,50269,50347,50403,50446,50488,50567,50582,50587,50616,50719,51157,51309,51457,51874,51902,51923,51971,52308,52383,52459,52483,52503,52741,53042,53170,53194,53197,53239,53330,53336,53363,53405,53635,53670,53681,53684,53739,53749,53853,53855,53936,53959,54034,54267,54370,54425,54552,54567,54638,54752,54757,54783,54831,54919,54961,54978,54995,55332,55373,55427,55443,55464,55516,55522,55525,55634,55683,56162,56232,56333,56438,56502,56586,56603,56631,56702,56785,56880,56983,57007,57021,57128,57197,57358,57372,57467,57498,57558,57565,57584,57640,57767,58081,58211,58231,58280,58349,58451,58479,58496,58710,58712,58722,58783,59045,59056,59230,59459,59477,59604,59933,60019,60050,60064,60113,60161,60178,60220,60223,60419,60454,60472,60704,60750,60873,60957,61198,61237,61324,61346,61451,61517,61646,61744,61829,61914,62116,62498,62534,62719,62732,62872,62903,62909,62956,63000,63290,63480,63853,63895,63905,63918,64001,64023,64069,64119,64382,64526,64550,64555,64572,64605,64793,64826,64845,65036,65160,65212,65286,65290,65334,65409,65512,65614,65695,65737,65739,65918,66014,66237,66266,66516,66620,66738,66762,66781,66805,66922,66956,66998,67023,67271,67315,67347,67384,67466,67518,67593,67625,67644,67657,67733,67769,67786,67869,67880,68011,68026,68041,68047,68185,68389,68459,68498,68512,68581,68595,68596,68708,68731,68768,68886,68913,69030,69074,69204,69370,69400,69428,69432,69535,69602,69614,69697,69702,69841,69885,69969,70047,70059,70087,70122,70230,70255,70369,70555,70588,70684,70793,70847,70854,70865,71112,71117,71200,71294,71353,71489,71640,71661,71778,71856,71917,71922,71999,72082,72088,72126,72357,72421,72496,72689,72707,72733,72741,72753,72987,73017,73038,73118,73202,73369,73389,73449,73485,73487,73500,73549,73625,73660,73665,73691,73839,73901,73919,73928,73998,74043,74077,74078,74092,74199,74244,74309,74587,74588,74647,74700,74708,74776,74855,74875,74898,74940,75009,75066,75340,75409,75553,75572,75625,75640,75755,75812,75964,75978,76022,76047,76206,76295,76520,76554,76670,76679,76748,76758,76772,77000,77034,77353,77531,77684,77776,77875,77956,78059,78119,78391,78403,78412,78517,78546,78688,78810,78853,78949,78972,78985,79042,79076,79098,79142,79166,79426,79437,79532,79577,79808,80271,80300,80313,80386,80390,80438,80463,80497,80547,80551,80571,80664,80795,80821,80918,80962,80975,81052,81178,81208,81264,81376,81471,81585,81587,81737,81880,81903,81933,81975,81989,82113,82252,82336,82453,82591,82772,82791,82846,82937,83161,83194,83196,83373,83496,83571,83579,83603,83925,84046,84075,84458,84479,84491,84628,84669,84714,84757,84804,84841,84843,84870,85092,85102,85153,85221,85504,85586,85639,85714,85721,85811,85978,86017,86112,86254,86333,86338,86412,86457,86647,86674,86676,86814,86946,86985,87373,87470,87713,87714,87767,87986,87995,88026,88236,88315,88362,88617,88657,88672,88702,88838,88904,89169,89354,89629,89696,89770,89838,89941,90127,90192,90234,90297,90307,90351,90399,90453,90534,90640,90703,90879,90897,91031,91060,91190,91233,91279,91443,91447,91471,91477,91511,91587,91784,91974,92190,92263,92271,92272,92450,92500,92609,93005,93084,93155,93178,93228,93237,93513,93656,93796,93959,94001,94069,94073,94103,94165,94222,94244,94298,94363,94471,94477,94963,95041,95049,95068,95225,95227,95362,95528,95581,95632,95753,95902,95911,95966,96020,96024,96055,96060,96109,96141,96148,96173,96203,96227,96242,96248,96308,96356,96527,96586,96653,96675,96704,96736,96758,96802,96815,96851,96998,97045,97062,97075,97116,97253,97324,97468,97602,97628,97686,97704,97887,97907,97956,97979,97990,98240,98473,98535,98577,98581,98786,98803,98896,98958,99010,99122,99132,99139,99314,99336,99377,99380,99465,99518,99590,99697,99743,99787,99795,99856,99926",
          "48,93,139,173,516,616,698,720,899,914,1143,1302,1359,1606,1726,1748,1796,1820,2067,2090,2230,2294,2317,2424,2485,2641,2644,2646,2666,2887,2979,3029,3075,3179,3183,3586,3608,3641,3649,3810,3828,3830,4003,4114,4146,4164,4170,4254,4518,4547,4592,4667,4690,4943,4947,4952,4969,5106,5133,5283,5300,5312,5454,5527,5566,5579,5615,5750,5830,5877,5940,6006,6012,6121,6129,6158,6336,6378,6408,6485,6509,6735,6790,6907,6979,7139,7182,7258,7299,7400,7416,7463,7542,7658,7798,7852,7865,7949,8171,8278,8316,8328,8364,8388,8638,8686,8694,8802,8872,8900,9127,9180,9223,9267,9282,9347,9367,9438,9441,9586,9590,9801,9822,9911,9933,9965,10008,10052,10158,10165,10256,10321,10338,10555,10619,10645,10674,10798,11025,11092,11162,11225,11379,11387,11396,11484,11507,11514,11727,11742,11797,11837,11976,11977,11991,12072,12109,12168,12294,12338,12360,12422,12576,12643,12709,12747,12754,12949,12975,13029,13222,13535,13606,13655,13678,13835,13895,14098,14232,14296,14350,14401,14420,14460,14514,14553,14636,14757,14762,14772,14856,15035,15072,15159,15236,15272,15306,15390,15424,15524,15572,15594,15605,15634,15698,15831,15852,15893,16228,16232,16254,16255,16292,16408,16419,16448,16461,16486,16622,16715,16718,16730,16732,16777,16792,16845,16901,16946,16996,17060,17207,17502,17850,18278,18361,18396,18459,18521,18648,18888,18899,19164,19207,19222,19362,19388,19418,19497,19514,19519,19682,19688,19695,19761,19811,19849,19854,19855,19869,19883,19903,19930,20092,20146,20244,20575,20605,20659,20695,20769,20837,20978,21140,21194,21416,21521,21574,21654,21719,21742,21878,22153,22212,22539,22541,22628,22678,22915,22939,22964,23204,23519,23533,23627,23716,23821,23848,23899,23909,23992,23997,24032,24139,24156,24369,24379,24761,24768,24861,24908,24953,24956,25078,25521,25547,25664,25671,25827,25989,26061,26075,26225,26287,26332,26359,26458,26461,26547,26552,26570,26595,26680,26735,26922,26943,26948,27119,27158,27372,27612,27637,27703,27854,28072,28146,28254,28350,28384,28644,28698,28831,28893,29024,29080,29144,29304,29382,29460,29569,29693,29938,29962,30039,30100,30124,30259,30309,30431,30448,30704,30722,30752,30774,30808,30831,31042,31187,31268,31278,31306,31432,31434,31518,31534,31604,31783,31934,32008,32123,32146,32280,32284,32545,32649,32829,32943,33092,33112,33164,33170,33253,33260,33499,33518,33730,33765,33787,33856,33917,33965,34116,34158,34167,34243,34259,34262,34327,34458,34461,34518,34521,34542,34630,34778,34884,34911,34979,35016,35075,35242,35278,35282,35414,35551,35634,35713,35762,35802,36033,36119,36232,36239,36331,36381,36781,36805,36838,36906,36920,37013,37158,37199,37208,37226,37291,37346,37354,37381,37581,37598,37666,37835,37852,37889,37893,37903,38037,38180,38383,38398,38435,38470,38624,38657,38670,38707,38721,38744,38754,38775,38798,38824,38867,39022,39025,39029,39212,39234,39258,39265,39351,39453,39552,39559,39779,39785,39794,39865,39963,39966,40041,40060,40090,40123,40219,40318,40467,40575,40653,40655,40977,40982,41244,41265,41314,41530,41538,41588,41642,41862,41929,42186,42242,42320,42338,42420,42554,42557,42747,42933,43349,43395,43744,43747,43908,43936,43962,44024,44025,44027,44124,44421,44471,44547,44609,44644,44764,44792,44794,44969,44984,45262,45281,45397,45423,45429,45430,45432,45448,45608,45634,45790,45908,46005,46015,46063,46167,46339,46340,46361,46378,46447,46529,46536,46670,46732,46782,46963,46967,46978,46996,47066,47361,47374,47416,47488,47601,47630,47821,47822,47908,48088,48098,48142,48410,48612,49045,49058,49092,49271,49273,49429,49432,49598,49620,49748,49853,49889,49915,49978,50032,50067,50222,50558,50590,50672,50776,50849,50962,51057,51167,51192,51498,51564,51740,51907,52013,52057,52128,52464,52502,52522,52593,52816,52916,52921,52934,53164,53219,53222,53368,53437,53446,53619,53886,53917,54081,54105,54111,54198,54232,54367,54468,54566,54644,54657,54711,54848,54971,55017,55046,55258,55288,55408,55618,55658,55795,55797,55886,55905,55926,56214,56425,56856,56941,56971,56978,57106,57198,57327,57375,57389,57511,57738,57782,57860,57950,57975,58170,58176,58288,58410,58414,58443,58561,58589,58743,58821,58879,58943,58989,59072,59117,59172,59252,59266,59273,59457,59467,59470,59532,59579,59793,59871,59898,59922,60041,60059,60069,60147,60212,60331,60496,60627,60706,60912,60991,60994,61211,61312,61447,61495,61661,61740,61926,61938,62012,62030,62092,62128,62137,62140,62177,62202,62258,62396,62419,62432,62697,62727,62861,62918,63033,63080,63081,63208,63281,63356,63444,63532,63619,63637,63681,63691,63742,63754,64215,64668,64671,64751,64787,64810,64847,64968,65022,65040,65107,65246,65267,65371,65397,65501,65578,65634,65767,65835,65998,66079,66215,66456,66489,66731,66785,66891,66962,67068,67118,67160,67238,67243,67258,67286,67310,67319,67320,67382,67402,67495,67617,67750,67801,67807,67917,67949,67962,68018,68232,68396,68467,68690,68722,69111,69112,69211,69359,69585,69717,69851,69925,69938,70071,70119,70132,70237,70272,70296,70317,70329,70335,70355,70430,70566,70639,70666,70893,70922,71045,71137,71260,71391,71677,71745,71751,71911,72125,72195,72300,72450,72491,72520,72523,72576,72635,72639,72764,72838,73193,73214,73238,73590,73640,73829,73945,73966,74170,74226,74278,74381,74521,74536,74653,74695,74711,74909,75018,75086,75180,75200,75265,75451,75475,75545,75617,75668,75671,75682,75737,75860,75869,75951,76052,76135,76143,76181,76189,76313,76369,76390,76486,76509,76801,76802,76888,76954,77071,77091,77188,77368,77477,77501,77539,77564,77580,77669,78091,78101,78194,78202,78221,78446,78643,78672,78690,78731,79033,79235,79324,79330,79359,79478,79516,79659,79741,79801,79825,79845,79896,80045,80166,80423,80532,80721,80785,80931,80963,80973,81177,81226,81288,81327,81393,81508,81535,81551,81690,81712,81877,81889,82086,82325,82413,82445,82478,82486,82488,82651,82676,82677,82686,82797,82834,82942,83167,83251,83317,83395,83431,83473,83482,83559,83716,83926,84013,84027,84114,84224,84237,84431,84574,84643,84679,84743,84880,85009,85037,85047,85138,85173,85281,85381,85433,85547,85592,85676,85752,85768,85901,86043,86124,86267,86483,86528,86561,86669,86869,86876,86901,86976,87003,87013,87133,87145,87235,87279,87324,87335,87336,87386,87412,87605,87779,87835,87926,87932,87933,88025,88102,88227,88234,88239,88285,88410,88420,88490,88506,88596,88622,88714,88778,88809,88920,89044,89142,89228,89251,89276,89381,89390,89471,89595,89600,89633,89802,89844,89879,89973,90030,90090,90101,90102,90130,90152,90158,90227,90316,90504,90687,90701,90732,90805,90837,90838,90930,90938,90939,90968,91013,91043,91055,91129,91230,91241,91264,91277,91287,91344,91368,91374,91425,91467,91472,91490,91535,91614,91688,91777,91782,91801,91829,91876,91922,91981,92183,92277,92289,92497,92551,92568,92650,92832,92843,92891,92946,92955,92981,92986,93049,93053,93103,93105,93132,93185,93216,93283,93387,93450,93474,93615,93760,93868,93946,93954,94112,94125,94214,94365,94373,94473,94527,94590,94627,94676,94773,94858,94998,95218,95332,95351,95367,95394,95433,95451,95460,95553,95648,95720,95956,95973,96012,96027,96053,96069,96319,96353,96377,96400,96519,96531,96719,96760,96924,96997,97052,97117,97180,97385,97558,97709,97715,97884,97916,97989,98030,98034,98092,98174,98365,98371,98539,98755,98760,99015,99093,99216,99347,99389,99402,99539,99632,99726,99818",
          "151,169,228,314,350,577,585,704,721,791,1000,1074,1227,1264,1332,1362,1373,1480,1563,1783,1834,1898,1938,1945,1946,2046,2147,2366,2385,2394,2474,2689,2753,2764,2833,2964,3164,3175,3182,3348,3352,3430,3497,3553,3689,3733,3743,3795,3805,4091,4095,4204,4257,4380,4403,4437,4464,4714,4795,4884,4978,4995,5050,5519,5683,5715,5729,5813,5847,5950,6141,6312,6358,6393,6407,6439,6726,6752,6860,7099,7163,7286,7407,7414,7470,7499,7503,7504,7558,7628,7849,8086,8098,8143,8174,8308,8342,8350,8443,8567,8599,8717,8844,8891,9037,9206,9294,9343,9410,9448,9572,9645,9687,9693,9782,9842,10096,10369,10371,10375,10431,10521,10531,10712,10874,10985,11007,11115,11156,11213,11227,11300,11321,11322,11609,11623,11650,11679,11840,11912,11967,12097,12140,12268,12310,12326,12334,12398,12437,12446,12524,12527,12593,12662,12742,12927,12958,12991,13252,13268,13290,13298,13363,13483,13538,13588,13749,13897,13962,14246,14472,14638,14644,14794,14854,14987,15092,15202,15233,15292,15321,15431,15531,15587,15670,15786,15851,16091,16165,16185,16389,16452,16652,16781,16799,16905,16958,17281,17588,17763,17771,18050,18128,18160,18172,18256,18264,18316,18402,18585,18608,18669,18814,18863,18910,19036,19119,19269,19318,19382,19408,19416,19520,19589,19614,19756,19773,19912,19955,19967,20106,20162,20362,20410,20536,20592,20619,20852,20954,21070,21145,21173,21336,21338,21413,21503,21752,21927,21971,21976,22014,22033,22093,22136,22141,22336,22351,22414,22568,22600,22699,22747,22798,22814,22881,23288,23330,23491,23517,23521,23578,23702,23713,24069,24077,24100,24146,24348,24429,24472,24634,24690,24693,24925,24974,25174,25183,25208,25220,25273,25295,25301,25311,25315,25523,25537,25667,25726,25779,25804,25838,25849,25967,26153,26291,26561,26690,26717,26775,26834,26976,27235,27261,27294,27474,27518,27541,27791,27923,28023,28149,28340,28448,28597,28600,28696,28720,28922,29042,29069,29092,29157,29272,29328,29512,29513,29516,29543,29554,29617,29621,29684,29696,29703,29928,30215,30231,30243,30293,30374,30440,30446,30482,30577,30580,30660,30742,30858,31013,31198,31265,31356,31428,31479,31531,31627,31807,31834,31853,31962,31968,32076,32086,32110,32137,32266,32337,32339,32398,32480,32501,32650,32703,32826,32893,32911,33035,33059,33085,33172,33208,33217,33281,33388,33546,33726,33829,34169,34186,34230,34234,34257,34261,34333,34417,34494,34507,34618,34625,34642,34745,34972,35035,35040,35057,35155,35314,35357,35390,35412,35506,35555,35581,35723,35763,35770,35785,35998,36012,36025,36319,36518,36765,36966,37115,37121,37154,37201,37203,37301,37361,37438,37463,37493,37527,37549,37866,37961,38086,38152,38305,38379,38391,38438,38465,38469,38524,38645,38652,38727,39006,39101,39111,39117,39237,39282,39470,39661,39695,39983,40033,40289,40346,40386,40414,40541,40555,40643,40678,40816,40876,41007,41053,41187,41227,41245,41366,41385,41509,41521,41648,41906,41979,42007,42012,42016,42042,42071,42237,42323,42480,42579,42692,42756,42768,42890,42963,43023,43194,43258,43278,43321,43504,43546,43754,43883,43993,44013,44111,44117,44246,44340,44452,44783,45028,45249,45260,45337,45453,45677,45767,45822,45881,45924,46052,46152,46318,46419,46589,46631,46822,46833,46883,46898,46900,46928,46941,47032,47065,47075,47137,47192,47262,47269,47344,47373,47545,47596,47613,47654,47781,47864,47943,47976,48145,48203,48508,48603,48904,49006,49037,49068,49075,49146,49181,49193,49285,49322,49357,49388,49516,49519,49529,49569,49697,49718,49874,49952,50246,50324,50353,50389,50617,50887,50889,50919,50951,51108,51140,51144,51248,51380,51455,51466,51641,51722,51899,52212,52293,52569,52583,52721,52888,52920,53054,53253,53559,53656,53716,54146,54273,54311,54415,54542,54624,54637,54920,54992,54993,55507,55513,55605,55656,55782,55837,55951,56023,56119,56168,56216,56297,56499,56593,56704,56753,56908,57053,57207,57315,57496,57514,57561,57657,57895,58000,58005,58069,58134,58147,58270,58271,58338,58388,58616,58638,58776,58952,59007,59077,59503,59586,59626,59756,59859,60004,60022,60077,60140,60668,60711,60751,60837,60961,61180,61337,61349,61422,61533,61536,61572,61748,61763,61838,61845,61949,61995,62189,62263,62330,62496,62603,62737,62833,62860,63026,63131,63227,63325,63372,63553,63586,63805,63817,63964,63971,64012,64051,64227,64385,64390,64539,64593,64659,64661,64742,64815,64830,64913,65014,65180,65187,65295,65321,65374,65381,65438,65538,65686,65703,65849,65905,66043,66113,66295,66510,66759,66796,66869,66894,66930,67215,67309,67345,67380,67599,67604,67608,67747,67797,67991,68015,68024,68269,68314,68329,68410,68662,68720,68729,68987,69101,69165,69331,69361,69484,69534,69653,69664,69812,69832,70097,70179,70536,70659,70755,71125,71334,71554,71575,71614,71667,71687,71818,71878,71907,71998,72129,72177,72270,72311,72348,72402,72417,72462,72471,72661,72786,72792,73110,73185,73360,73371,73400,73458,73507,73647,73703,73705,73739,73799,73819,73902,74007,74059,74178,74312,74316,74389,74591,74635,74740,74819,74985,75004,75365,75511,75522,75601,75679,75739,75742,75952,76020,76068,76074,76086,76090,76270,76351,76418,76506,76884,76967,76977,77205,77225,77243,77323,77445,77502,77691,77718,77759,77798,77871,77989,78066,78204,78476,78556,78600,78608,78655,78876,79017,79111,79155,79205,79213,79244,79419,79511,79717,79748,79852,79895,80050,80233,80284,80357,80442,80671,80675,80686,80881,80995,81146,81149,81163,81244,81726,81947,82092,82154,82383,82442,82527,82535,82751,82956,83042,83121,83142,83173,83281,83319,83325,83444,83493,83497,83564,83658,83736,83779,83877,83957,84205,84302,84599,84644,84682,84774,84812,84983,84999,85027,85058,85118,85140,85167,85174,85311,85487,85528,85896,85937,86078,86285,86366,86452,86473,86589,86665,86756,86845,86928,86937,86979,87039,87160,87230,87369,87371,87528,87564,87581,87850,88005,88329,88513,88522,88691,88744,88756,88834,88872,88907,89046,89106,89119,89152,89338,89417,89551,89927,89948,90143,90530,90736,90746,90767,90773,90803,90806,90827,90843,90910,91002,91135,91180,91373,91392,91474,91501,91550,91652,91659,91700,91970,91996,92150,92212,92253,92314,92372,92508,92593,92711,92783,92831,92835,93040,93045,93098,93356,93375,93391,93443,93444,93479,93519,93544,93558,93612,94014,94253,94270,94385,94420,94423,94486,94501,94643,94699,94759,94821,94833,95164,95253,95437,95457,95640,95674,95721,95775,95792,95825,95957,96315,96357,96472,96518,96606,96680,96692,96824,96856,97012,97097,97178,97263,97408,97413,97470,97589,97613,97755,97820,98000,98047,98225,98533,98570,98743,98767,98912,98987,99001,99235,99239,99426,99505,99809,99810,99880",
          "167,211,298,333,368,390,401,517,685,745,755,757,921,964,1118,1275,1315,1392,1436,1523,1627,1657,1721,1739,1762,1764,1906,1953,1991,2029,2096,2137,2276,2335,2381,2388,2435,2538,2544,2628,2665,2712,2743,2768,2807,3033,3171,3316,3369,3477,3494,3535,3541,3674,3690,3708,3731,3772,3815,3825,3840,3862,3921,3933,3994,4093,4126,4255,4454,4660,4807,4823,4832,4834,4909,4911,5078,5170,5289,5428,5572,5692,5981,5999,6074,6095,6113,6114,6197,6229,6264,6368,6479,6636,6685,6694,6728,6978,7015,7023,7213,7274,7378,7413,7478,7550,7600,7668,7672,7685,7688,7697,7726,7854,7897,7928,7986,8006,8097,8109,8126,8178,8234,8739,8783,8818,9021,9100,9233,9370,9534,9566,9595,9621,9656,9679,9719,9759,9765,9770,9792,9793,9848,9891,9919,9931,10140,10337,10391,10412,10423,10472,10473,10625,10646,10717,10719,10823,10895,10937,11069,11450,11539,11584,11591,11646,11667,11717,11749,11772,11867,11870,11952,12040,12130,12185,12324,12354,12442,12453,12467,12480,12503,12512,12601,12717,12798,12803,12804,12886,12914,13045,13308,13413,13433,13480,13489,13553,13586,13614,13718,13780,14031,14069,14079,14097,14114,14142,14201,14261,14371,14425,14434,14581,14597,14738,14764,15020,15101,15118,15164,15502,15615,15627,15682,15734,15799,15823,15826,15875,15983,16038,16300,16360,16379,16400,16412,16414,16442,16465,16478,16543,16546,16783,16816,16922,17007,17008,17047,17093,17107,17156,17167,17243,17360,17503,17551,17578,17604,17640,17642,17691,17697,17706,17741,17746,17760,17838,17882,17921,17997,18087,18118,18161,18169,18227,18354,18584,18708,18773,18831,18968,19054,19066,19188,19262,19317,19347,19376,19381,19626,19732,20127,20151,20188,20451,20467,20489,20520,20540,20590,20629,20760,20797,20903,21041,21067,21153,21320,21588,21667,21689,21803,21881,21915,21934,21974,22107,22239,22295,22314,22328,22376,22452,22483,22484,22500,22642,22705,22718,22824,22839,22844,22950,23001,23104,23112,23125,23177,23195,23256,23300,23320,23465,23669,23701,23743,23815,23983,24022,24289,24330,24332,24425,24468,24513,24523,24528,24611,24620,24640,24765,24766,24877,24898,25042,25160,25226,25255,25323,25345,25410,25423,25439,25559,25718,25865,25886,25960,25970,26028,26112,26155,26284,26318,26414,26505,26542,26614,26638,26646,26743,27159,27173,27202,27270,27421,27424,27429,27620,27716,27790,28021,28134,28295,28314,28485,28487,28654,28728,28768,28823,28829,28852,28858,28882,28899,28924,28954,29113,29152,29218,29229,29400,29448,29662,29670,29694,29726,29801,29909,30057,30127,30222,30227,30229,30275,30282,30300,30414,30444,30489,30549,30563,30574,30581,30731,30756,30820,30822,30884,30937,30951,30964,31007,31036,31080,31101,31115,31204,31220,31359,31370,31415,31601,31697,31698,31966,32014,32073,32185,32243,32318,32474,32670,32674,32791,32817,32837,32876,32892,32896,32970,32978,33012,33082,33125,33235,33424,33441,33451,33476,33521,33523,33531,33564,33580,33779,33918,33939,33946,34063,34341,34421,34575,34693,34697,34722,34803,35032,35065,35098,35321,35325,35387,35717,35829,35834,35873,35934,36097,36113,36141,36311,36356,36375,36393,36419,36427,36443,36454,36489,36814,36869,36892,36903,37063,37081,37097,37207,37240,37276,37363,37388,37395,37502,37513,37522,37565,37628,37653,37684,37725,37839,37847,37921,37947,38034,38196,38262,38318,38474,38511,38646,38659,38817,38832,38858,38861,38931,39047,39115,39148,39208,39460,39646,39648,39653,39708,39818,39856,39918,39949,40075,40192,40261,40324,40404,40413,40445,40490,40496,40902,40908,41032,41156,41232,41308,41319,41413,41467,41535,41536,41570,41683,41685,41686,41898,41992,42040,42250,42443,42486,42635,42648,42883,43041,43061,43132,43152,43209,43223,43243,43355,43534,43591,43637,43655,43755,43852,43874,43899,43922,44018,44072,44114,44141,44165,44271,44304,44319,44498,44515,44608,44612,44684,44781,45006,45029,45184,45195,45290,45338,45393,45416,45602,45653,45805,45984,46030,46189,46193,46220,46231,46334,46335,46617,46719,46789,46827,46843,46847,46932,46966,47012,47030,47074,47238,47306,47322,47365,47418,47430,47592,47649,47705,47832,47934,47937,47981,47998,48002,48009,48064,48072,48186,48298,48401,48427,48565,48575,48697,48861,49080,49097,49101,49240,49296,49474,49658,49759,49797,49895,49983,50033,50081,50132,50238,50287,50329,50387,50460,50471,50562,50572,50652,50752,50842,50911,50982,51021,51033,51110,51116,51154,51323,51385,51496,51598,51686,51817,51948,51975,52049,52141,52184,52205,52260,52276,52330,52488,52532,52559,52595,52672,52947,53101,53187,53200,53354,53361,53392,53441,53482,53623,53726,53791,53822,53847,53862,53876,53970,54018,54048,54134,54135,54170,54184,54261,54305,54315,54403,54495,54588,54853,54941,54942,55035,55042,55076,55087,55159,55195,55291,55371,55541,55598,55630,55786,55848,55997,56012,56244,56293,56340,56351,56424,56488,56505,56630,56659,56689,56867,56925,56965,57054,57142,57365,57481,57552,57768,57794,58205,58286,58626,58641,58713,58714,58740,58819,58824,58845,58878,58914,59035,59076,59112,59296,59399,59407,59514,59521,59879,59936,60030,60071,60121,60224,60246,60303,60307,60583,60879,61048,61288,61316,61475,61574,61658,61696,61719,61768,61811,61873,61913,62008,62048,62203,62248,62305,62371,62411,62459,62602,62624,62758,62895,62896,63001,63231,63245,63301,63429,63453,63536,63806,63936,63975,63990,64029,64158,64183,64216,64283,64528,64538,64635,64647,64717,64725,64778,64896,64916,65033,65096,65149,65150,65162,65232,65265,65276,65427,65439,65445,65458,65493,65560,65639,65651,65689,65812,65977,66068,66259,66265,66431,66476,66547,66594,66682,66725,66778,66790,66871,66872,66918,67021,67152,67252,67259,67269,67438,67737,67764,67773,67979,67999,68150,68257,68271,68277,68408,68455,68755,68907,68981,69007,69064,69322,69338,69691,69711,69713,69846,70050,70069,70110,70268,70349,70403,70436,70533,70534,70562,70594,70713,70783,71002,71082,71089,71141,71276,71372,71389,71398,71453,71758,71861,71867,72092,72233,72237,72399,72459,72475,72560,72599,72692,72749,72820,72875,73021,73245,73246,73279,73304,73345,73351,73461,73559,73577,73582,73607,73730,73740,73762,73813,73823,73825,73979,74022,74193,74254,74285,74415,74477,74546,74612,74663,74733,74790,74872,74931,74956,74963,74993,75103,75177,75194,75322,75347,75435,75507,75587,75591,75637,75667,75686,75752,75764,75772,75784,75824,75905,76071,76097,76223,76310,76363,76401,76548,76581,76589,76728,76770,76854,76859,76937,76940,76981,77179,77280,77316,77319,77349,77359,77483,77596,77672,77717,77758,77797,78015,78160,78215,78234,78277,78278,78295,78360,78362,78468,78552,78664,78776,78780,78799,78831,78857,79157,79227,79481,79486,79521,79591,79755,79781,79917,80011,80039,80122,80244,80557,80566,80644,80704,80705,80706,80736,80776,80832,80849,80859,80909,80953,81046,81084,81092,81098,81213,81341,81398,81412,81602,81708,81740,81781,81848,81999,82033,82161,82182,82277,82317,82326,82332,82351,82372,82400,82467,82526,82723,82833,82840,82842,82884,82902,82926,83033,83376,83443,83495,83546,83632,83715,83910,83928,83950,84191,84238,84332,84367,84381,84411,84466,84475,84512,84686,84749,84754,84934,84962,85022,85107,85171,85180,85182,85271,85276,85308,85348,85428,85476,85512,85560,85569,85631,85665,85667,85766,85796,85939,86069,86161,86179,86224,86265,86297,86320,86496,86556,86577,86588,86597,86693,86719,86846,86934,86936,86996,87042,87045,87102,87143,87236,87407,87504,87516,87552,87591,87606,87759,87843,87907,87944,88048,88106,88141,88153,88221,88235,88290,88323,88442,88542,88729,88909,88963,89018,89212,89269,89545,89596,89641,89781,90032,90119,90167,90210,90280,90349,90415,90626,90695,90851,91000,91126,91173,91174,91215,91242,91333,91388,91626,91764,91885,92017,92088,92112,92133,92224,92353,92377,92483,92518,92561,92607,92619,92661,92668,92920,92931,92992,93130,93241,93270,93281,93294,93313,93342,93346,93397,93421,93503,93588,93668,93720,93790,93909,94098,94139,94147,94284,94339,94391,94557,94573,94589,94616,94760,94843,94898,94977,95224,95248,95313,95336,95563,95747,95809,95851,95912,95945,95948,96118,96403,96434,96945,97515,97537,97562,97644,97802,97833,97964,98027,98088,98111,98127,98130,98212,98358,98524,98579,98677,98759,98821,98840,99049,99320,99339,99390,99423,99457,99674,99693,99699,99709,99844,99891",
          "85,118,229,328,377,429,575,666,712,782,827,868,1029,1283,1393,1408,1567,1643,1740,1829,1854,1883,2020,2023,2280,2330,2418,2426,2534,3010,3042,3078,3166,3408,3418,3552,3621,3732,3782,3865,3913,4017,4044,4056,4068,4088,4118,4160,4181,4256,4303,4323,4428,4450,4474,4590,4611,4645,4696,4835,4881,4917,4931,5039,5087,5218,5233,5240,5242,5274,5453,5500,5585,5979,5982,6056,6097,6100,6247,6342,6372,6548,6571,6607,6886,6908,6970,7043,7061,7116,7199,7221,7234,7267,7422,7532,7578,7660,8026,8040,8079,8115,8122,8154,8211,8306,8435,8660,8741,8782,8788,8826,8902,9002,9112,9146,9329,9395,9465,9577,9579,9588,9776,9836,9892,10082,10117,10217,10243,10314,10666,10715,10858,11082,11128,11167,11307,11493,11582,11662,11805,11855,11922,11973,12012,12155,12295,12462,12599,12718,12761,12930,13099,13105,13130,13170,13237,13287,13317,13334,13393,13411,13436,13523,13555,13618,13619,13652,13717,13928,13940,13945,14158,14236,14373,14389,14407,14541,14558,14601,14604,14623,14763,14834,14920,14985,15029,15124,15268,15354,15359,15551,15827,15908,15999,16013,16041,16057,16109,16188,16235,16363,16523,16561,16663,16710,16796,16808,16810,16829,16961,17029,17030,17174,17301,17334,17672,17744,17828,17840,17868,17893,17960,18000,18136,18241,18245,18282,18295,18310,18364,18535,18550,18600,18639,18719,19025,19027,19094,19163,19238,19246,19330,19369,19375,19577,19586,19674,19888,20017,20139,20258,20274,20277,20336,20370,20485,20501,20538,20545,20640,20807,20891,21120,21417,21568,21606,21784,21845,22043,22071,22082,22236,22403,22598,22714,22807,22828,23167,23183,23244,23277,23334,23452,23678,23694,23767,23770,23847,23877,23925,24072,24304,24320,24401,24516,24575,24612,24639,24642,24739,24757,24870,24942,25242,25290,25320,25342,25353,25384,25665,25742,25809,25853,25950,26086,26090,26234,26243,26260,26391,26509,26637,26687,26720,26820,26896,27091,27106,27112,27263,27350,27480,27507,27545,27782,27800,27804,27925,28008,28082,28089,28193,28259,28288,28530,28560,28581,28650,28784,28843,28854,28955,29044,29123,29240,29351,29618,29821,29990,30021,30052,30078,30147,30194,30251,30269,30687,30826,30894,30939,31172,31181,31214,31277,31321,31423,31736,31763,31831,31916,31945,31973,31979,32020,32040,32081,32164,32290,32307,32324,32454,32487,32576,32796,32923,33094,33149,33216,33292,33443,33687,33820,33940,34014,34052,34082,34170,34196,34464,34497,34503,34587,34668,34699,34707,34797,34805,34807,34859,34875,34917,35005,35042,35115,35149,35226,35511,35541,35588,35599,35659,35765,35766,35789,36001,36131,36221,36250,36253,36341,36474,36494,36544,36570,36643,36706,36768,36825,36833,36977,37044,37071,37082,37083,37105,37106,37292,37321,37353,37397,37433,37473,37498,37674,37685,37804,37958,38022,38106,38149,38554,38628,38755,38977,39122,39161,39179,39274,39497,39510,39514,39542,39569,39637,39651,39685,39701,39733,39808,39890,40026,40035,40047,40147,40173,40243,40329,40375,40417,40432,40495,40547,40611,40703,40704,40822,40841,40850,40894,40942,40984,41056,41104,41178,41239,41261,41279,41457,41501,41568,41606,41634,41673,41674,41706,41763,41765,41790,41792,41931,41951,42002,42015,42041,42164,42199,42222,42224,42271,42293,42296,42624,42711,42738,42831,42875,43027,43048,43075,43100,43148,43187,43376,43447,43722,43764,43817,43844,43887,43977,44163,44220,44289,44492,44533,44543,44631,44739,44839,44872,45109,45209,45253,45447,45450,45456,45472,45549,45555,45737,45771,45791,46000,46037,46068,46207,46209,46406,46455,46470,46503,46519,46586,46662,46681,46694,46761,46810,46976,46995,47152,47185,47216,47220,47240,47258,47294,47329,47384,47436,47514,47537,47587,47708,47715,47726,47809,47982,48101,48141,48158,48371,48555,48633,48655,48705,48713,48957,49002,49081,49141,49198,49214,49248,49257,49577,49639,49655,49675,50584,50613,50623,50669,50732,50800,50913,50934,50968,51042,51345,51379,51478,51544,51582,51777,51992,52004,52006,52029,52090,52274,52428,52940,52999,53154,53416,53431,53470,53547,53620,53659,53662,53826,53915,54024,54213,54235,54318,54372,54388,54466,54585,54659,54726,54743,54758,54806,54925,55057,55160,55185,55327,55335,55593,55695,55944,55974,56018,56021,56305,56375,56496,56507,56528,56611,56651,56750,56826,56844,56862,56893,56894,56928,56934,57041,57173,57248,57250,57468,57557,57599,57706,57721,57739,57818,57838,57906,57966,57969,58006,58133,58158,58245,58307,58384,58426,58442,58683,58718,58785,58792,58938,59049,59124,59145,59171,59183,59299,59339,59385,59463,59472,59486,59580,59585,59587,59630,59632,59721,59831,59882,59885,59914,59929,59960,60201,60205,60276,60336,60415,60420,60487,60498,60536,60899,61080,61152,61223,61253,61444,61563,61585,61593,61644,61779,62036,62101,62129,62159,62171,62321,62375,62680,62717,62726,62745,62762,62878,62947,63017,63079,63197,63335,63426,63430,63463,63537,63595,63706,63747,63751,63838,63875,63939,63962,63973,63984,64075,64125,64195,64262,64384,64504,64734,64811,65031,65093,65115,65127,65306,65309,65342,65472,65580,65691,65728,65881,65916,66010,66147,66247,66322,66393,66428,66446,66549,66695,66748,66764,67045,67083,67115,67125,67192,67361,67451,67523,67613,67710,68109,68121,68450,68475,68580,68587,68747,68800,68879,69327,69337,69605,69709,69733,69777,69811,69891,70009,70068,70116,70141,70205,70288,70320,70332,70399,70485,70561,70768,70851,70942,70972,71175,71219,71295,71337,71386,71423,71521,71682,71763,71974,72095,72168,72243,72531,72590,72596,72627,72668,72740,72793,72808,72822,73029,73053,73109,73120,73186,73261,73317,73355,73419,73497,73513,73539,73612,73707,73850,74354,74362,74404,74550,74615,74656,74703,74878,75036,75043,75201,75292,75422,75492,75550,75594,75786,75918,75932,75981,75996,76001,76026,76059,76111,76125,76129,76136,76642,76658,76691,76774,76786,76790,76878,76917,77039,77107,77162,77230,77252,77442,77579,77633,77663,77747,77760,77794,77904,77964,78060,78084,78153,78249,78348,78490,78645,78668,78742,78750,78758,78909,79115,79139,79154,79203,79288,79351,79373,79400,79561,79565,79679,79700,79729,79860,80019,80087,80291,80389,80398,80489,80515,80530,80624,80755,80777,80865,80978,81137,81166,81257,81284,81313,81316,81348,81566,81588,81611,81662,81970,82050,82124,82175,82199,82212,82229,82310,82461,82541,82826,82901,82912,82918,82949,82990,83018,83113,83122,83226,83228,83245,83552,83689,83812,83989,84117,84351,84433,84494,84580,84583,84642,84844,84846,84927,84973,85043,85100,85460,85520,85529,85637,85806,85851,85880,85966,86133,86200,86262,86271,86401,86583,86649,86681,86734,86889,87125,87129,87218,87289,87348,87372,87388,87468,87524,87754,87884,87936,88074,88140,88149,88252,88295,88368,88489,88625,88941,88979,89014,89030,89060,89117,89222,89234,89340,89360,89384,89387,89449,89598,89703,89735,89908,89942,90014,90181,90484,90611,90714,91223,91359,91438,91457,91695,91704,91710,91720,91760,91787,91819,91911,91935,91966,92092,92312,92335,92405,92438,92612,92651,92782,92971,92990,93043,93124,93307,93607,93626,93788,93806,93838,94053,94082,94265,94280,94368,94628,94727,94733,94791,94915,94931,95217,95395,95614,95633,95650,95686,95798,96183,96205,96206,96279,96379,96642,96757,96840,96942,97237,97279,97284,97441,97480,97531,97551,97572,97576,97777,97985,98110,98120,98133,98201,98320,98338,98362,98455,98538,98547,98801,98897,99062,99065,99143,99178,99212,99296,99475,99571,99575,99588,99710,99766,99837,99983",
          "107,119,130,629,665,769,874,890,1136,1201,1257,1421,1580,1604,1785,1800,1832,1859,1934,2003,2098,2153,2200,2209,2242,2243,2256,2365,2419,2591,2690,2785,2842,2854,2894,2914,2990,3045,3204,3287,3463,3464,3759,3906,3934,3980,4140,4216,4488,4575,4587,4694,4726,4821,4873,4941,5159,5343,5445,5599,5613,5629,5844,5972,6079,6139,6280,6295,6473,6500,6580,6923,7039,7216,7250,7279,7427,7430,7494,7556,7663,7699,7702,7762,7871,7923,7967,7972,7992,7994,7995,8136,8152,8187,8252,8294,8411,8422,8441,8487,8530,8704,8867,8956,9069,9137,9192,9283,9353,9394,9408,9471,9592,9666,9750,9864,9900,9943,9950,9987,10099,10126,10131,10170,10207,10234,10329,10379,10566,10678,10747,10830,10861,10876,11046,11055,11074,11205,11500,11557,11700,11702,11743,11830,11856,11882,11892,11925,11951,12107,12335,12385,12546,12562,12585,12612,12646,12865,13015,13049,13110,13199,13250,13256,13280,13421,13434,13582,13590,13705,13722,13793,13826,13854,13903,14023,14144,14291,14342,14351,14369,14500,14516,14592,14633,14654,14671,14732,14864,14895,14939,14969,15048,15128,15222,15249,15281,15322,15366,15393,15443,15511,15512,15537,15593,15684,15750,15762,15795,16042,16064,16073,16238,16988,17032,17172,17180,17215,17324,17338,17793,17805,17822,18109,18307,18333,18362,18399,18407,18548,18590,18630,18690,18837,18927,19045,19052,19053,19090,19120,19329,19461,19582,19877,19969,20096,20098,20272,20290,20304,20391,20480,20499,20543,20595,20692,20885,20897,20956,21049,21073,21225,21237,21425,21729,21920,22047,22117,22242,22311,22486,22538,22693,22701,22725,22826,22847,22925,22953,23023,23025,23165,23267,23291,23317,23371,23623,23703,23717,23724,23824,23841,23903,23988,24015,24120,24470,24545,24562,24580,24590,24756,25064,25141,25268,25309,25417,25424,25427,25470,25518,25533,25810,25842,25857,26034,26048,26051,26223,26329,26356,26379,26390,26497,26507,26536,26756,27003,27090,27145,27197,27290,27291,27327,27505,27558,27604,27737,27824,27858,27918,27921,27933,27935,27952,28052,28155,28226,28242,28244,28393,28458,28499,28510,28542,28694,28731,28732,28804,28851,29045,29150,29324,29387,29472,29528,29639,29715,29740,29880,29907,29934,29967,30031,30109,30216,30257,30331,30572,30635,30684,30712,30741,30803,30823,30893,31289,31596,31771,31803,31813,31882,31951,31955,32026,32149,32206,32345,32389,32401,32485,32526,32626,32646,32710,32752,32775,32854,32870,33005,33040,33204,33243,33274,33278,33364,33374,33401,33411,33422,33623,33747,33818,33996,34015,34070,34164,34277,34455,34520,34602,34624,34635,34640,34661,34907,34936,35088,35146,35148,35153,35157,35251,35320,35415,35427,35524,35538,35656,35711,35838,35856,35936,35976,36193,36285,36335,36355,36402,36650,36745,36811,36844,36852,36900,37086,37157,37186,37311,37314,37334,37541,37584,37589,37606,38043,38045,38053,38065,38087,38162,38242,38760,38883,39049,39063,39064,39112,39244,39310,39579,39734,39775,39848,40231,40377,40437,40457,40608,40853,40886,41082,41233,41287,41352,41561,41565,41665,41679,41791,41804,41863,41941,42152,42158,42208,42225,42281,42319,42580,42610,42679,42734,43043,43091,43236,43261,43353,43392,43483,43593,43645,43674,43679,43985,44055,44232,44335,44423,44478,44657,44729,44778,44838,44867,44950,45115,45204,45221,45245,45365,45483,45601,45611,45958,46093,46218,46283,46414,46515,46549,46555,46691,46798,46817,46835,46936,46970,47069,47090,47149,47259,47293,47471,47476,47478,47523,47656,47664,47694,47744,47782,47805,47980,48049,48136,48234,48294,48400,48522,48569,48574,48585,48586,48624,48673,48810,48811,48874,48910,49026,49036,49122,49251,49256,49290,49317,49323,49471,49488,49513,49599,49985,49995,50089,50248,50730,50782,50986,50996,51595,51711,51927,52012,52151,52158,52183,52411,52422,52505,52507,52562,52726,52884,52908,53045,53157,53234,53402,53415,53455,53542,53605,53688,53827,53863,53931,54015,54226,54278,54475,54553,54574,54618,54736,54742,54771,54914,54945,54954,55072,55145,55164,55307,55372,55395,55540,55626,55756,55849,55852,55910,56453,56604,56634,56837,56852,56975,57171,57234,57287,57310,57349,57352,57414,57556,57600,57637,57644,57757,57780,57928,57986,58204,58247,58256,58382,58437,58463,58573,58583,58645,58659,59079,59203,59214,59281,59423,59435,59518,59597,59663,59679,59744,59869,59927,60357,60448,60525,60534,60683,60784,60871,60956,61183,61219,61280,61290,61560,61605,61626,61631,61647,61985,62037,62332,62382,62412,62421,62423,62462,62790,62920,63019,63101,63153,63240,63292,63357,63413,63801,64006,64020,64177,64190,64194,64211,64248,64319,64366,64469,64584,64600,64636,64650,64701,64784,64806,64849,64945,65190,65213,65335,65523,65641,65715,65781,65824,65939,65960,66018,66419,66475,66789,66863,66909,67670,67715,67735,67777,67794,67802,68073,68134,68173,68183,68344,68420,68432,68531,68616,68794,68943,68951,69059,69203,69243,69261,69423,69746,69816,70027,70129,70131,70135,70228,70343,70409,70422,70432,70446,70476,70531,70683,70686,70991,71099,71110,71506,71519,71565,71582,71673,71852,71964,72097,72114,72141,72155,72181,72534,72593,72637,72811,72844,73002,73071,73292,73396,73704,73864,73871,73903,74023,74205,74236,74605,74621,74630,74678,74746,74799,74957,75160,75244,75269,75315,75332,75472,75485,75538,75571,75645,75702,75777,75827,75911,75929,75936,75949,76029,76033,76250,76427,76433,76470,76560,76564,76598,76746,76894,76895,76947,76975,77023,77041,77055,77108,77119,77215,77296,77307,77343,77414,77489,77495,77818,77872,77940,77993,78099,78227,78258,78336,78381,78407,78492,78505,78530,78596,78665,78706,78724,78918,78920,79035,79075,79104,79145,79251,79287,79307,79519,79587,79641,79774,80016,80104,80123,80237,80425,80525,80578,80629,80957,81068,81079,81168,81180,81272,81335,81363,81403,81525,81581,81584,81656,81698,81707,81742,82261,82320,82397,82457,82460,82654,82666,82688,82712,82734,82780,82804,82821,82996,83120,83217,83417,83459,83479,83507,83523,83528,83650,83665,83767,83829,83906,83958,84131,84204,84376,84380,84386,84503,84523,84527,84549,84789,84799,84894,84958,85082,85096,85201,85591,85625,85656,85754,86030,86213,86277,86463,86654,86684,86858,87089,87161,87308,87316,87442,87514,87824,87924,87930,88072,88129,88138,88322,88380,88408,88425,88431,88433,88648,88659,88724,88784,88817,88849,88898,89111,89170,89177,89271,89272,89314,89476,89534,89799,89860,89877,89938,89952,89965,90177,90268,90298,90299,90304,90398,90613,90637,90710,90711,90745,90761,90782,90875,90934,90973,91159,91234,91324,91362,91792,91837,91958,92051,92077,92158,92206,92259,92280,92373,92479,92579,92623,92731,92746,92851,92987,93120,93187,93229,93256,93350,93574,93625,93644,93649,93731,93756,93776,93873,93943,93965,93997,94074,94195,94233,94277,94484,94623,94692,94748,94751,94869,94902,94936,94951,94965,94966,95018,95198,95443,95663,95694,95903,95969,96040,96111,96365,96440,96565,96823,96863,96865,96869,96875,96906,96932,96955,96973,97051,97120,97150,97190,97218,97270,97495,97574,97707,97749,97854,97889,98022,98175,98229,98303,98304,98382,98387,98461,98566,98587,98653,98841,98873,98957,99104,99300,99355,99413,99587,99846,99962,99984,99990",
          "75,109,137,382,383,443,492,540,715,736,740,1008,1185,1190,1228,1297,1468,1470,1496,1510,1548,1757,1788,1824,1860,1879,1943,2130,2168,2176,2180,2215,2225,2378,2470,2537,2713,2717,2752,2756,2763,2933,2986,3011,3064,3096,3098,3132,3208,3271,3343,3434,3546,3588,3599,3858,3981,3982,4079,4083,4224,4298,4538,4780,4894,5134,5152,5217,5280,5456,5474,5559,5594,5730,5778,5835,5890,5894,5895,5906,5961,5966,6186,6301,6350,6427,6470,6490,6502,6520,6582,6637,6712,6919,6980,7007,7033,7081,7292,7602,7661,7734,7742,7784,7808,8031,8105,8163,8382,8518,8564,8740,8816,9000,9054,9092,9122,9140,9354,9386,9404,9405,9548,9575,9672,9706,9802,9893,9910,9927,9932,10074,10139,10241,10246,10255,10296,10318,10386,10618,10633,10637,10784,10912,11143,11199,11214,11296,11385,11416,11423,11447,11616,11690,11790,11864,12044,12053,12186,12205,12472,12474,12687,12787,12863,12876,13053,13058,13066,13070,13135,13190,13255,13270,13327,13372,13456,13466,13518,13541,13578,13587,13605,13690,13710,13736,13744,13811,13813,13852,14042,14122,14171,14289,14344,14349,14354,14387,14543,14631,14670,14690,14734,14739,14760,14862,14866,14891,15024,15189,15203,15204,15224,15231,15234,15373,15408,15462,15470,15532,15686,15832,15874,15901,15906,15914,16180,16195,16213,16253,16318,16351,16356,16377,16600,16753,16764,16778,16997,17122,17259,17417,17475,17535,17559,17599,17721,17804,17826,17858,17974,17994,18014,18100,18193,18209,18572,18792,18807,18874,18923,19078,19110,19141,19191,19211,19220,19278,19289,19305,19380,19453,19475,19648,19704,19714,19722,19801,19832,19965,20044,20063,20088,20113,20125,20238,20252,20294,20322,20325,20357,20512,20621,20684,20740,20746,20959,20964,20995,21089,21188,21198,21273,21383,21390,21591,21649,21821,21855,21860,21943,21984,22035,22086,22158,22196,22331,22422,22488,22592,22676,22854,22872,22906,22911,22955,23002,23078,23119,23247,23264,23554,23656,23671,23827,23889,24013,24029,24047,24056,24151,24176,24317,24346,24446,24462,24487,24662,24721,24760,24820,24863,24943,24948,25025,25038,25100,25213,25217,25398,25492,25501,25570,25601,25633,25648,25724,25760,25762,25789,25823,25852,25869,25983,26019,26025,26067,26069,26113,26179,26365,26408,26467,26503,26629,26648,26853,26954,27024,27140,27205,27253,27349,27397,27688,27714,27773,27832,27944,28105,28110,28178,28184,28235,28292,28365,28436,28627,28682,28688,28778,28803,28847,28874,29114,29140,29142,29285,29501,29534,29592,29746,29865,29958,30004,30032,30048,30133,30157,30249,30256,30263,30289,30365,30398,30485,30711,30718,30746,30753,30776,30842,30845,31116,31128,31137,31175,31196,31236,31259,31308,31372,31387,31405,31473,31542,31599,31624,31675,31755,32089,32263,32299,32382,32531,32577,32641,32915,32967,32994,33026,33305,33368,33440,33483,33535,33562,33902,33908,33979,34027,34042,34051,34075,34115,34168,34216,34232,34253,34293,34381,34549,34564,34896,34902,34929,34931,35068,35118,35225,35420,35550,35636,35654,35776,35866,35937,35949,36022,36046,36117,36263,36477,36555,36562,36567,36605,36671,36818,37023,37102,37243,37257,37424,37442,37524,37569,37579,37726,37769,37788,37799,37873,38033,38038,38232,38291,38512,38576,38591,38634,38734,38765,38806,38831,38857,38877,39206,39228,39247,39329,39381,39419,39496,39531,39588,39663,39665,39672,39720,39860,39896,39953,40006,40115,40189,40250,40312,40323,40393,40434,40450,40534,40548,40579,40594,40733,40785,41014,41110,41154,41305,41541,41637,41865,41981,42011,42028,42053,42155,42278,42351,42374,42413,42468,42546,42743,42817,42866,42912,42919,43164,43340,43575,43581,43725,43726,43740,43742,43800,43906,43916,44044,44066,44186,44201,44284,44313,44351,44356,44434,44526,44537,44850,44929,44963,44968,45079,45167,45241,45259,45296,45328,45470,45499,45542,45680,45681,45911,45987,46040,46050,46122,46153,46178,46194,46225,46226,46255,46382,46399,46418,46423,46462,46594,46696,46766,46768,46777,46841,46884,46940,47005,47050,47146,47163,47230,47363,47491,47667,47669,47714,47803,47819,47910,47965,48022,48045,48120,48167,48252,48334,48366,48384,48420,48441,48442,48465,48483,48572,48604,48670,48778,48785,48791,48837,48932,49173,49293,49400,49523,49540,49545,49562,49672,49688,49891,49958,50180,50234,50369,50384,50498,50545,50969,50974,51014,51103,51156,51158,51243,51257,51274,51304,51356,51547,51556,51569,51829,51845,51870,52037,52071,52129,52145,52237,52292,52325,52356,52460,52555,52693,52759,52777,52952,53012,53132,53144,53556,53680,53699,53818,53934,53983,54119,54349,54354,54385,54417,54476,54615,54957,55151,55225,55271,55444,55781,55883,55939,55994,56034,56346,56373,56382,56467,56483,56489,56648,56652,56662,56678,56748,56933,56960,56973,57089,57102,57273,57385,57415,57534,57579,57581,57788,57848,57885,57977,58011,58020,58070,58188,58234,58268,58371,58405,58432,58483,58516,58621,58654,58769,58827,58846,58946,59005,59025,59174,59178,59250,59260,59362,59460,59462,59475,59734,59762,59817,59852,59891,59963,60009,60104,60288,60622,60670,60715,61004,61020,61066,61100,61113,61372,61375,61549,61606,61620,61731,61787,61806,61894,61978,62038,62095,62199,62240,62301,62779,62890,62892,62929,63052,63129,63134,63220,63283,63383,63534,63545,63782,63823,63978,64072,64080,64414,64415,64518,64713,64721,65034,65085,65095,65098,65399,65462,65531,65557,65562,65583,65628,65649,65777,65800,65811,65831,65956,65966,66140,66517,66645,66651,66668,66797,66806,66929,66990,67127,67165,67179,67276,67327,67476,67500,67504,67618,67631,67768,67977,67998,68175,68308,68366,68382,68444,68453,68518,68552,68732,68754,68769,68992,69031,69183,69189,69558,69579,69695,69714,69736,69754,69782,69794,69838,70003,70174,70307,70331,70387,70391,70427,70442,70515,70625,70672,70685,70697,70750,70777,70788,70835,70867,71021,71267,71291,71400,71501,71507,71650,71821,71946,72005,72096,72228,72478,72499,72566,72586,72592,72613,72641,72684,72843,72911,72972,72995,73007,73079,73143,73155,73269,73473,73499,73542,73610,73679,73702,73712,73732,73824,74024,74253,74347,74348,74417,74436,74457,74548,74654,74732,74871,74902,74950,75010,75137,75184,75486,75515,75546,75551,76145,76169,76178,76188,76214,76344,76537,76654,76709,76729,76730,76918,76968,76982,77052,77187,77190,77198,77279,77485,77752,77824,77889,77895,77896,77919,77944,77961,77991,78070,78205,78219,78238,78243,78248,78256,78310,78433,78559,78569,78619,78650,78735,78875,79128,79136,79210,79212,79273,79295,79326,79528,79541,79786,79811,79816,79868,79986,80074,80100,80189,80385,80441,80448,80596,80702,80772,80773,80792,80802,80828,80869,80914,80929,81064,81142,81302,81409,81486,81494,81510,81532,81555,81608,81710,82009,82081,82136,82162,82281,82335,82379,82500,82569,82634,82879,82909,82978,82984,82986,82993,83155,83198,83290,83353,83468,83494,83499,83542,83551,83574,83577,83808,83830,83860,83981,83986,84007,84041,84065,84102,84133,84160,84310,84765,84785,84864,85046,85076,85094,85322,85371,85390,85480,85493,85494,85563,85727,85747,85761,85803,85997,86132,86135,86149,86216,86234,86299,86341,86445,86504,86596,86607,86618,86851,87078,87096,87135,87156,87192,87204,87342,87417,87500,87620,87640,87712,87716,87734,87919,87942,87965,88218,88281,88568,88577,88673,88730,88914,88971,88977,89002,89035,89087,89198,89470,89591,89771,89839,89848,89865,89912,89991,90051,90072,90074,90157,90240,90244,90289,90348,90385,90770,90786,91077,91254,91379,91462,91604,91750,91866,91897,91923,92191,92275,92313,92391,92394,92444,92558,92567,92606,92610,92657,92714,92729,92770,92775,92998,93065,93128,93153,93210,93223,93247,93355,93374,93410,93418,93578,93584,93751,93812,93865,93883,94358,94376,94442,94547,94607,94682,94817,94856,94950,94975,94980,95015,95052,95186,95299,95383,95387,95492,95572,95605,95629,95653,95703,96015,96142,96229,96261,96263,96280,96303,96358,96437,96503,96533,96578,96628,96718,96850,96868,97140,97251,97277,97310,97381,97493,97653,97800,98021,98190,98210,98326,98346,98407,98437,98494,98542,98549,98550,98611,98672,98679,98756,98773,98964,99038,99040,99126,99137,99182,99188,99196,99201,99226,99308,99515,99560,99603,99629,99778,99963",
          "8,278,289,293,311,355,363,367,399,418,483,770,777,805,938,1114,1244,1259,1326,1376,5012,5053,5068,5100,5103,5162,5265,5277,5353,5389,5557,5667,5732,5756,5761,5789,5799,5846,5873,5900,5912,5919,5923,5935,5964,6051,6052,6085,6086,6298,6343,6388,6461,6475,6544,6711,6773,6862,6863,6932,6949,6961,7047,7057,7136,7157,7186,7319,7365,7680,8269,8465,9177,9230,9584,9682,10030,10044,10413,10838,10930,11210,11400,11545,12242,13163,13703,14315,15219,16114,16740,17964,18520,18605,18606,18763,20381,20970,20989,21207,22440,24351,24505,24511,25027,25307,27256,28628,28800,31012,31135,31805,32056,35328,37679,45665,50043,50145,50303,50463,50496,50570,50670,50862,50868,50891,50950,51121,51221,51238,51287,51381,51446,51465,51832,51853,52021,52051,52083,52178,52455,52509,52518,52561,52806,52850,52910,52943,52982,52987,53015,53199,53216,53256,53588,53595,53733,53748,53947,54005,54027,54043,54093,54229,54242,54247,54263,54319,54422,54512,54527,54607,54629,54639,54754,54821,54836,54866,54958,55064,55139,55169,55182,55230,55292,55775,55816,55930,55970,55971,55972,56028,56102,56107,56129,56288,56450,56598,56672,56797,56841,56864,56895,57073,57247,57343,57536,57717,57869,57996,58113,58146,58210,58226,58386,58440,58523,58774,58775,58790,58798,59221,59225,59326,59432,59458,59465,59653,59656,59714,60021,60136,60382,60453,60549,60578,60628,60842,60845,60962,61084,61121,61191,61194,61472,61497,61856,62071,62223,62279,62294,62377,62408,62492,62559,62650,62764,62796,62865,62893,62968,63138,63278,63295,63298,63504,63508,63802,63818,63870,63887,63922,64010,64088,64277,64355,64441,64602,64607,64726,64877,64907,64985,65052,65084,65094,65168,65284,65336,65592,65602,65729,65787,65794,66091,66105,66281,66386,66505,66541,66618,66663,66804,66955,67006,67020,67051,67169,67254,67311,67445,67454,67457,67496,67643,68002,68053,68157,68200,68335,68351,68364,68457,68608,68700,68749,68797,68861,68869,68906,69061,69249,69304,69356,69397,69555,69591,69893,69954,70083,70109,70159,70326,70395,70421,70716,70718,70764,70901,70910,70967,70970,71007,71206,71222,71248,71336,71468,71471,71605,71624,71658,71663,71739,71743,71790,71807,71810,71902,71903,71906,72018,72077,72090,72139,72142,72153,72207,72282,72367,72385,72473,72497,72522,72575,72655,72723,72757,72767,72857,72992,73000,73157,73212,73309,73397,73515,73734,73810,73939,73992,74025,74247,74252,74276,74371,74386,74410,74538,74569,74670,74718,74754,74848,74906,74944,74976,74979,75123,75140,75338,75418,75455,75563,75794,76018,76110,76316,76321,76335,76352,76353,76395,76475,76489,76814,76953,76992,76993,76994,77018,77123,77126,77177,77344,77360,77370,77490,77613,77643,77846,77873,77887,78257,78683,78704,78727,78740,78852,79020,79165,79171,79233,79509,79886,79934,80003,80012,80021,80144,80325,80371,80374,80380,80396,80412,80748,80753,80759,80894,81115,81133,81196,81204,81256,81378,81607,81873,81963,82062,82108,82219,82424,82472,82484,82616,82704,82758,82828,82988,83003,83254,83450,83612,83623,83646,83730,83768,84014,84064,84071,84104,84126,84312,84396,84521,84522,84635,84711,84761,84811,84833,84847,85026,85214,85217,85417,85510,85525,85561,85670,85770,85795,85834,86066,86067,86073,86089,86202,86231,86248,86249,86381,86395,86421,86456,86571,86744,86778,86841,86863,86868,87331,87426,87474,87693,87733,87760,87784,87790,87834,87840,87880,88251,88303,88548,88580,88854,88863,88937,88950,89191,89250,89324,89398,89400,89447,89468,89502,89575,89659,89674,89681,89782,89798,89813,90008,90015,90058,90103,90211,90239,90279,90331,90516,90585,90593,90766,90771,90776,90829,90911,90959,91004,91017,91062,91139,91253,91268,91274,91376,91403,91459,91498,91519,91612,91678,91810,91852,92009,92026,92160,92161,92200,92231,92237,92363,92400,92433,92441,92513,92524,92816,93087,93192,93335,93343,93700,93801,93977,94026,94093,94100,94127,94156,94261,94364,94567,94620,94645,94680,94730,94746,94871,94948,95030,95032,95192,95320,95340,95533,95749,95941,95965,96087,96158,96220,96296,96328,96600,96629,96699,96711,96767,96944,96991,97084,97134,97170,97220,97348,97466,97498,97534,97624,97713,97732,97735,97805,97926,97949,97950,98018,98305,98321,98644,98657,98668,98686,98962,98970,98972,99276,99335,99453,99562,99652,99832,99893,99911,99941,99950,99959",
          "74,123,129,160,327,346,352,463,507,541,780,841,887,986,1082,1182,1184,1451,1588,1610,1644,1789,1803,1822,1858,1900,2076,2083,2085,2315,2360,2463,2479,2574,2611,2629,2708,2775,2912,2953,2984,3006,3012,3028,3092,3247,3513,3530,3545,3575,3576,3669,3687,3786,3792,3826,3879,3885,3940,4024,4229,4319,4341,4353,4395,4398,4438,4476,4487,4839,4901,4930,5014,5027,5200,5526,5604,5884,5897,6011,6016,6021,6077,6191,6276,6281,6382,6401,6415,6458,6594,6755,6941,6947,7016,7140,7218,7337,7360,7380,7381,7436,7442,7451,7541,7568,7771,7774,7968,7978,8059,8104,8125,8244,8259,8267,8325,8495,8531,8566,8616,8702,8897,8925,8996,9005,9064,9153,9193,9376,9449,9569,9678,9746,9825,9829,9831,9913,9978,9992,10041,10069,10145,10196,10208,10282,10290,10292,10327,10385,10612,10677,10688,10744,10763,10779,10855,10949,11002,11010,11026,11136,11239,11504,11517,11595,11725,11736,11849,11935,12034,12195,12218,12224,12283,12312,12333,12345,12376,12389,12402,12441,12478,12547,12574,12635,12680,12774,12801,12957,13187,13249,13540,13585,13702,13735,13786,13790,13905,13915,13943,13953,13985,14030,14175,14229,14329,14359,14643,14687,14694,14702,14810,14925,15113,15157,15349,15380,15472,15571,15595,15641,15661,15664,15685,15848,15869,15870,15873,15931,15961,16030,16247,16415,16425,16565,16667,16998,17166,17182,17187,17311,17354,17367,17388,17533,17553,17597,17602,17635,17720,17829,17852,17977,18188,18218,18246,18317,18368,18457,18710,18751,18755,18850,18871,18902,18956,18974,19239,19271,19313,19404,19484,19818,19857,19991,20048,20389,20495,20508,20535,20747,20793,20816,20817,20860,20866,20873,20919,20983,21000,21011,21107,21384,21451,21514,21563,21566,21731,21817,22162,22251,22271,22280,22361,22384,22544,22630,22634,22635,22670,22729,22817,22857,22858,23075,23080,23148,23228,23286,23331,23345,23416,23421,23451,23559,23661,23762,23843,23853,23947,23982,24085,24109,24149,24150,24170,24220,24262,24364,24445,24486,24529,24534,24598,24679,24727,24780,24851,24855,24899,25107,25253,25270,25284,25313,25314,25387,25420,25584,25737,25812,25814,25879,25893,26038,26077,26239,26274,26321,26533,26548,26626,26630,26702,26745,26900,26936,27034,27171,27333,27469,27570,27656,27725,27977,28207,28241,28480,28516,28549,28653,28864,28865,28949,29067,29225,29244,29424,29461,29541,29638,29648,29650,29673,29732,29770,29793,29827,29879,30179,30183,30341,30342,30347,30353,30436,30504,30520,30541,30551,30650,30656,30767,30870,30919,30941,31084,31085,31313,31385,31454,31460,31480,31523,31735,31900,31998,32005,32120,32122,32136,32170,32221,32365,32387,32437,32711,32802,32814,32874,32946,33065,33165,33200,33393,33485,33640,33670,33824,33831,33969,33977,34061,34066,34099,34272,34318,34383,34395,34595,34672,34756,34881,34883,34937,35085,35093,35332,35339,35485,35744,35782,35821,35840,35955,35972,36031,36065,36123,36208,36370,36615,36617,36669,36693,36909,36964,37040,37045,37054,37094,37172,37237,37247,37267,37431,37624,37659,37672,37695,37768,37805,37910,37938,37970,38049,38131,38142,38215,38285,38309,38368,38387,38388,38392,38415,38417,38457,38464,38468,38480,38573,38776,38909,38930,38969,39165,39200,39210,39217,39251,39261,39307,39336,39349,39427,39532,39583,39686,39777,39802,39821,39842,39850,39906,39913,39915,39937,39940,39943,40044,40116,40146,40278,40284,40378,40420,40500,40515,40517,40545,40745,40830,41044,41059,41130,41353,41394,41471,41477,41632,41675,41848,41886,41921,41924,42177,42221,42267,42283,42328,42483,42595,42626,42631,42676,42720,42763,42770,42854,42888,42964,42977,43057,43110,43116,43120,43125,43176,43267,43282,43385,43455,43509,43628,43676,43714,43878,43904,43929,43959,44031,44115,44157,44204,44205,44287,44404,44516,44539,44635,44758,44909,44965,44981,45090,45096,45223,45291,45387,45388,45645,45682,45947,45998,46034,46061,46078,46079,46126,46494,46528,46562,46641,46647,46652,46678,46736,46811,46814,46933,46983,47196,47222,47264,47332,47496,47524,47593,47600,47672,47722,47730,47793,47799,48034,48103,48107,48228,48288,48382,48402,48414,48506,48617,48684,48695,48716,48732,48800,48830,48896,48967,49052,49288,49302,49383,49413,49460,49503,49546,49561,49715,49768,49875,49950,50008,50136,50166,50170,50261,50473,50476,50511,50538,50548,50574,50636,50645,50658,50801,50807,50829,51094,51321,51408,51413,51499,51601,51629,51652,51848,52048,52166,52181,52193,52214,52236,52273,52408,52446,52545,52641,52662,52680,52720,52867,53085,53176,53261,53329,53434,53487,53543,53548,53708,53828,53919,53930,54064,54077,54110,54143,54256,54433,54734,55014,55165,55192,55279,55325,55450,55460,55581,55726,55742,55745,55805,56319,56331,56407,56566,56632,56725,56742,56752,56759,56763,56771,56818,57050,57121,57147,57150,57199,57269,57316,57502,57525,57564,57566,57604,57605,57608,57710,57720,57745,57872,57926,57967,58054,58101,58138,58161,58194,58413,58428,58578,58610,58717,58800,58931,59010,59024,59065,59078,59080,59196,59306,59403,59415,59449,59516,59523,59577,59673,59712,59735,59743,59825,59858,60132,60209,60299,60321,60540,60566,60575,60652,60834,60898,60987,61007,61078,61110,61129,61338,61539,61745,61794,61800,61802,61858,61865,61885,61958,62197,62204,62225,62295,62329,62384,62437,62449,62471,62486,62532,62773,62782,62942,63086,63205,63210,63393,63397,63472,63540,63571,63667,63752,63900,64255,64298,64381,64421,64446,64542,64598,64634,64685,64705,64801,64817,64865,65097,65125,65355,65389,65400,65682,65856,65868,65898,66189,66249,66274,66325,66347,66390,66519,66586,66667,66714,66793,66821,66865,66879,66940,67074,67075,67109,67142,67164,67245,67482,67626,67680,67781,67900,67919,67929,68143,68176,68227,68236,68301,68358,68399,68431,68594,68688,68773,68799,68848,68904,68934,69040,69083,69170,69242,69255,69269,69358,69553,69686,69844,69993,70018,70034,70105,70123,70257,70264,70270,70337,70375,70423,70454,70471,70477,70523,70633,70650,70708,70806,71025,71038,71069,71070,71120,71171,71203,71205,71242,71279,71329,71366,71408,71576,71681,71694,71904,71990,72003,72071,72332,72362,72568,72585,72799,72842,72941,72977,73044,73121,73144,73200,73250,73626,73830,73940,73983,74014,74046,74085,74225,74443,74658,74702,74858,74972,75050,75155,75584,75745,75761,75802,75805,75866,75907,75925,76043,76050,76159,76219,76242,76305,76428,76526,76822,76829,77017,77040,77110,77135,77327,77350,77480,77488,77516,77556,77581,77609,77622,77626,77641,77652,77657,77739,77969,78019,78056,78171,78456,78457,78495,78533,78614,78763,78828,78865,78883,78902,78916,79094,79185,79311,79370,79421,79630,79738,79827,79835,79869,79938,80001,80027,80118,80138,80328,80392,80431,80604,80650,80844,80969,80997,81036,81078,81097,81155,81165,81308,81476,81493,81612,81652,81819,81882,81922,81962,82142,82239,82254,82290,82463,82509,82663,82739,82863,83143,83150,83190,83285,83460,83475,83487,83515,83638,83822,83890,84016,84051,84111,84319,84342,84407,84446,84481,84486,84545,84696,84795,84868,85039,85041,85055,85165,85183,85305,85317,85380,85521,85540,85738,85791,85804,85849,85876,85936,85983,86025,86053,86126,86353,86511,86568,86883,87059,87330,87389,87507,87630,87696,87710,87796,87807,88114,88246,88366,88416,88470,88492,88525,88561,88662,88751,88870,88894,88949,88975,89121,89126,89312,89316,89493,89527,89737,89946,90044,90053,90061,90079,90085,90365,90421,90544,90600,90638,90812,90928,91028,91111,91208,91247,91286,91311,91340,91346,91352,91370,91605,91722,92203,92213,92293,92300,92307,92398,92443,92472,92540,92773,92925,93026,93125,93156,93212,93227,93269,93320,93414,93434,93485,93522,93530,93549,93738,94047,94131,94146,94264,94389,94419,94591,94705,94712,94769,94911,94988,95388,95480,95530,95693,95717,95871,95907,95930,95932,96174,96224,96265,96314,96322,96474,96514,96615,96964,97109,97188,97314,97360,97396,97520,97586,97640,97757,98029,98050,98104,98250,98256,98266,98368,98426,98462,98638,98664,98671,98680,98682,98798,98842,98986,99034,99035,99118,99155,99192,99472,99638,99782,99792,99849,99879,99947,100001",
          "6,331,336,340,347,354,356,391,395,404,430,438,442,462,467,494,519,630,684,764,835,849,862,865,917,959,1021,1048,1188,1233,1375,1478,1521,3102,5026,5084,5148,5167,5188,5220,5227,5228,5255,5269,5304,5314,5328,5380,5427,5431,5561,5628,5690,5796,5800,5805,5856,5865,5891,5909,5921,5948,5949,5974,6187,6202,6256,6560,6606,6824,6851,7062,7191,7246,7347,7364,7403,7753,8137,8158,8226,8356,8476,8959,9057,9096,9260,9274,9359,9389,10003,10034,10492,11728,13574,13913,15445,15527,16479,16760,18089,18353,19435,21003,22079,25936,26788,40385,40998,41658,46877,48076,50093,50103,50165,50283,50343,50370,50423,50462,50489,50716,50896,50963,50980,51067,51084,51138,51164,51175,51250,51314,51400,51445,51448,51474,51550,51602,51865,51970,51983,52070,52180,52195,52281,52370,52461,52529,52549,52556,52809,53002,53077,53081,53225,53319,53395,53423,53566,53642,53831,53911,53912,54109,54205,54245,54251,54296,54299,54307,54343,54351,54523,54665,54777,54814,54844,54933,54948,54970,55261,55311,55473,55478,55498,55772,55819,55953,55956,56054,56079,56160,56757,56833,56904,57074,57455,57582,57606,57731,57746,57766,58050,58071,58122,58237,58282,58489,58572,58591,58629,58667,58672,58767,58861,58939,58950,58960,58990,59019,59042,59130,59170,59239,59241,59388,59513,59636,59676,59776,59784,59797,59894,59917,60002,60078,60082,60146,60242,60452,60527,60581,60698,60793,60894,60900,60980,60985,60990,61026,61037,61163,61175,61402,61509,61512,61535,61624,61648,61680,61755,61782,61906,62058,62069,62070,62190,62275,62409,62440,62490,62500,62518,62537,62771,62827,62868,62959,62962,63013,63092,63115,63158,63218,63334,63371,63408,63423,63514,63542,63551,63687,63720,63739,63741,63763,63868,64081,64111,64124,64393,64426,64556,64631,64728,64844,64903,64946,65071,65099,65196,65354,65390,65420,65640,65696,65743,65766,65890,65912,66012,66025,66034,66041,66052,66053,66096,66250,66423,66432,66443,66449,66603,66721,67046,67080,67093,67255,67395,67512,67678,67697,68087,68144,68171,68179,68334,68368,68373,68468,68506,68546,68610,68633,68912,69026,69289,69598,69756,69907,69974,70015,70147,70231,70342,70345,70358,70365,70657,70682,70709,70795,70897,71032,71163,71177,71274,71448,71724,71737,71838,71956,71957,71987,72029,72190,72483,72551,72658,72925,73041,73203,73239,73319,73349,73463,73856,73969,73980,74016,74228,74241,74368,74513,74659,74674,74679,74721,74815,74891,74961,75185,75268,75392,75407,75720,75762,75803,75859,75875,75942,76070,76343,76375,76477,76517,76519,76597,76668,76693,76707,76890,76935,76980,77030,77346,77416,77465,77585,77629,77675,77763,77779,77838,77998,78016,78024,78054,78185,78242,78321,78371,78542,78595,78743,78800,78854,78921,78974,78975,79239,79321,79352,79504,79549,79576,79596,79747,79818,80242,80260,80274,80293,80320,80387,80674,80693,80806,80888,80896,81122,81383,81426,81458,81515,81646,81694,81724,81746,81758,81901,81936,82051,82084,82169,82209,82354,82465,82504,82593,82664,82681,82708,82748,82810,82856,82907,82952,82985,83005,83055,83147,83199,83302,83332,83367,83410,83415,83593,83609,83649,83735,83930,83995,84054,84132,84176,84210,84377,84395,84720,84810,84932,84937,84964,85048,85148,85204,85213,85223,85251,85346,85357,85457,85517,85613,85863,85870,86007,86092,86094,86252,86323,86396,86440,86484,86534,86630,86720,86803,86835,86897,86972,86992,87073,87229,87334,87399,87427,87505,87624,87638,87684,87827,87854,88041,88307,88330,88667,88825,88875,88919,88936,88960,88985,89107,89112,89124,89150,89206,89297,89616,89748,89786,89789,89891,89949,90022,90111,90176,90452,90485,90525,90568,90598,90670,90699,90833,90850,90913,90933,90943,91087,91150,91216,91285,91526,91567,91584,91752,91933,91944,92022,92050,92082,92099,92241,92284,92291,92393,92611,92721,92842,92883,93126,93160,93199,93225,93263,93520,93540,93651,93684,93757,93851,93952,93999,94183,94190,94197,94213,94352,94511,94546,94593,94603,94639,94762,94805,94831,94846,94847,94978,95037,95146,95339,95350,95446,95548,95616,95651,95748,95929,95934,95946,95990,96046,96107,96139,96202,96249,96348,96349,96488,96590,96599,96686,96710,96829,96845,96943,97037,97128,97135,97151,97241,97347,97425,97564,97646,97658,97681,97693,97708,97733,97742,97765,97787,97794,97838,97881,97903,97909,97951,98005,98048,98163,98167,98260,98275,98656,98704,98826,98867,99042,99121,99159,99386,99454,99556,99716,99735",
          "68,553,693,697,730,809,949,963,1065,1134,1148,1291,1320,1329,1361,1428,1441,1553,1758,1804,1863,1892,1909,1914,1926,1969,2154,2244,2248,2319,2402,2473,2617,2659,2733,2825,2852,2907,2995,3176,3202,3276,3414,3461,3468,3469,3489,3542,3611,3636,3671,3691,3834,3841,3955,4035,4131,4171,4393,4449,4505,4564,4578,4646,4920,4925,4927,4956,5267,5286,5327,5362,5481,5541,5650,5720,5779,5806,5878,5938,6084,6273,6572,6595,6680,6864,6870,6958,6965,6990,7232,7247,7265,7273,7277,7280,7367,7398,7511,7621,7638,7640,7778,7861,7882,7917,8054,8147,8261,8409,8489,8643,8649,8780,8801,8825,8890,8952,9013,9052,9161,9201,9279,9436,9469,9634,9664,9722,9874,9887,10045,10144,10252,10403,10471,10530,10630,10884,10902,10909,10914,10918,11318,11445,11571,11779,11836,11866,11918,11936,12071,12081,12116,12249,12266,12522,12559,12595,12762,13100,13121,13376,13410,13631,13685,13800,13817,13841,13939,14008,14045,14237,14240,14467,14476,14510,14539,14615,14714,15003,15031,15100,15109,15185,15286,15409,15450,15501,15517,15596,15668,15968,16184,16215,16274,16432,16502,16519,16603,16658,16743,16886,16930,16970,17027,17066,17119,17165,17209,17353,17477,17541,17670,17703,17975,17976,17985,18025,18105,18130,18155,18243,18266,18429,18462,18539,18578,18707,18717,18753,18776,18942,18955,18980,19014,19101,19177,19232,19357,19424,19473,19478,19505,19533,19539,19541,19596,19601,19607,19631,19671,19733,19780,19861,19863,19894,19919,20259,20279,20297,20299,20315,20329,20452,20459,20541,20562,20583,20696,20962,21063,21106,21115,21221,21656,21750,22050,22116,22304,22356,22363,22398,22579,22759,22831,22907,23033,23036,23074,23203,23226,23272,23340,23368,23488,23721,23940,23958,24051,24074,24154,24204,24268,24355,24378,24456,24488,24597,24651,24713,24783,24815,24879,24887,24910,24971,25054,25062,25084,25138,25219,25235,25276,25455,25545,25725,25843,25998,26107,26226,26227,26228,26261,26431,26455,26472,26479,26620,26846,26852,26902,26950,27041,27069,27530,27547,27569,27675,27707,27760,27811,27936,28075,28183,28233,28303,28307,28334,28368,28387,28471,28511,28577,28611,28672,28708,28834,28835,28871,28926,29090,29143,29184,29190,29260,29291,29348,29367,29375,29561,29741,29759,29779,29869,30166,30208,30326,30433,30507,30564,30589,30599,30605,30649,30654,30662,30667,30764,30777,30835,31083,31170,31240,31248,31249,31489,31515,31544,31617,31766,31779,32102,32150,32174,32262,32300,32406,32563,32667,32851,33001,33046,33061,33066,33251,33314,33630,33635,33688,33690,33752,33845,33915,34002,34048,34288,34319,34375,34728,34790,34888,34895,34901,35064,35213,35533,35542,35651,35738,36005,36007,36051,36060,36061,36072,36074,36155,36213,36220,36226,36385,36413,36425,36504,36620,36626,36695,36762,36858,36863,37005,37070,37090,37133,37181,37192,37236,37304,37384,37451,37529,37567,37715,37742,38213,38238,38296,38513,38558,38567,38594,38599,38690,38712,38715,38774,38778,38904,38955,39140,39156,39164,39184,39270,39515,39634,39635,39657,39713,39826,39864,39891,39900,39912,40027,40133,40509,40511,40552,40613,40672,40832,40904,40951,41029,41049,41051,41252,41339,41382,41441,41446,41475,41623,41678,41748,41761,41786,41847,41932,41973,42092,42179,42366,42378,42379,42419,42428,42558,42706,42708,42771,42789,42899,42924,42994,43064,43080,43114,43228,43313,43362,43364,43475,43588,43720,44026,44048,44154,44216,44233,44236,44239,44407,44444,44625,44681,44691,44705,44898,44927,45014,45138,45152,45179,45217,45384,45479,45547,45591,45755,45814,45865,45874,46074,46077,46080,46173,46196,46216,46291,46322,46327,46441,46795,46834,46840,47072,47084,47086,47298,47299,47302,47350,47401,47481,47507,47618,47645,47684,47829,47858,47863,47989,48019,48037,48117,48143,48174,48202,48220,48232,48436,48472,48497,48535,48628,48652,48733,48797,48886,48949,48959,48966,48990,49009,49059,49120,49212,49215,49274,49298,49320,49345,49386,49405,49456,49481,49494,49536,49651,49684,49725,49977,50111,50268,50289,50474,50478,50479,50544,50683,50694,50696,50697,50749,50820,50946,50979,50990,51017,51163,51184,51316,51359,51375,51502,51541,51665,52142,52249,52431,52649,52665,52679,52899,52912,52931,52979,53010,53099,53180,53181,53202,53245,53295,53417,53424,53712,53756,53783,53996,54013,54158,54264,54436,54470,54625,54693,54710,54739,54835,55001,55078,55341,55344,55379,55385,55420,55470,55548,55666,55935,55984,56094,56169,56233,56301,56307,56329,56361,56374,56512,56745,56779,56787,56798,56899,56953,56969,57013,57067,57109,57153,57368,57781,57787,58171,58223,58417,58462,58494,58514,58532,58634,58644,58676,58703,58751,58958,59108,59136,59146,59187,59313,59409,59434,59646,59912,59962,59999,60046,60188,60328,60340,60374,60380,60550,60660,60665,60721,60896,60933,61116,61133,61231,61311,61342,61347,61384,61566,61578,61649,61675,61754,61809,61820,61895,61991,62045,62285,62299,62501,62556,62560,62576,62595,62596,62677,62710,62787,62885,62948,62955,63065,63235,63388,63420,63624,63628,63777,63786,63865,63950,63959,63963,64055,64077,64083,64311,64392,64449,64486,64692,64707,64732,64738,64919,65004,65068,65083,65186,65261,65347,65485,65490,65494,65630,65666,65951,66159,66196,66276,66372,66438,66604,66613,66786,67014,67090,67101,67204,67408,67441,67481,67516,67635,67933,67935,67938,67969,67971,67992,68022,68250,68268,68275,68465,68479,68530,68536,68638,68649,68676,68702,68721,68734,68736,68852,68905,69065,69149,69264,69471,69518,69583,69610,69618,69683,69705,69769,69833,69839,69890,69984,70064,70085,70139,70169,70181,70269,70284,70591,70687,70730,70784,70839,71019,71148,71251,71301,71546,71558,71644,71726,71729,71731,72212,72554,72555,72561,72671,72849,73014,73181,73183,73518,73602,73651,73653,73669,73826,73837,73858,73962,73965,74123,74257,74288,74391,74414,74432,74447,74595,74636,74638,74889,74945,74999,75074,75154,75179,75199,75219,75281,75286,75420,75724,75841,75884,75944,75999,76056,76133,76251,76632,76726,76739,76754,76800,76837,77026,77036,77331,77332,77371,77438,77515,77582,77700,77709,77851,77911,77977,78040,78307,78436,78497,78601,78647,78751,79194,79329,79555,79713,79773,79861,79888,79943,80047,80106,80259,80289,80359,80376,80381,80486,80556,80698,80794,80891,80904,80913,81022,81131,81253,81517,81556,81622,81659,81668,81714,81722,81810,81872,82043,82067,82157,82243,82280,82308,82765,82768,82878,82924,83034,83106,83153,83178,83274,83359,83510,83643,83656,83690,83750,83800,83819,83918,84118,84138,84213,84253,84276,84281,84295,84306,84341,84360,84365,84400,84417,84460,84510,84526,84845,84918,84946,85056,85270,85643,85684,85703,85705,85729,85782,85895,85921,85925,86188,86263,86399,86441,86560,86586,86746,86818,86994,87122,87267,87358,87576,87582,87644,87674,87828,87888,88197,88233,88262,88282,88319,88352,88384,88393,88514,88618,88655,88858,88915,89100,89125,89147,89157,89158,89241,89350,89536,89672,89779,89832,89835,89840,89896,89914,89920,89971,90020,90047,90073,90410,90524,90604,90814,90871,90944,90981,91008,91044,91564,91661,91673,91834,91961,92040,92120,92187,92347,92417,92434,92449,92564,92719,92754,92797,92805,92854,92878,92942,93054,93092,93198,93211,93351,93378,93560,93635,93646,93699,93840,93857,94004,94033,94172,94321,94529,94794,94905,94995,95043,95090,95096,95112,95279,95354,95416,95495,95630,95677,95691,95730,95810,95822,96066,96228,96384,96407,96507,96511,96573,96652,96857,96861,96926,96960,97077,97184,97196,97228,97261,97303,97304,97352,97367,97372,97430,97447,97700,97812,97893,97993,98173,98217,98280,98289,98308,98397,98429,98530,98563,98591,98600,98616,98618,98751,98854,98856,98959,98984,99149,99206,99263,99305,99398,99501,99535,99583,99601,99642,99771,99831,99878,99925,99935",
          "94,144,158,183,243,542,582,663,847,1117,1187,1328,1365,1425,1432,1642,1650,1663,1746,1778,1818,1836,2005,2015,2077,2093,2112,2116,2142,2206,2260,2358,2382,2413,2438,2441,2560,2581,2652,2661,2837,2841,2849,3034,3119,3160,3330,3394,3407,3444,3569,3625,3721,3722,3767,3929,3965,4022,4084,4134,4192,4379,4447,4456,4519,4597,4632,4671,4692,4816,4913,4960,5252,5355,5462,5465,5479,5517,5565,5575,5635,5700,5824,5831,5951,6029,6067,6072,6108,6110,6196,6265,6270,6418,6431,6450,6732,6750,6796,6867,6964,7072,7447,7586,7715,7743,7745,7940,7993,8020,8155,8197,8200,8221,8273,8372,8398,8513,8607,8631,8687,8688,8731,8884,8904,8910,8997,9124,9222,9278,9385,9390,9425,9576,9631,9661,9699,9708,9711,9812,9854,9856,10025,10090,10182,10194,10199,10271,10298,10324,10519,10529,10604,10616,10636,10650,10813,10817,10867,11020,11038,11053,11056,11089,11206,11216,11224,11267,11402,11449,11621,11644,11756,11761,11898,11913,11974,11987,12054,12086,12138,12144,12311,12433,12514,12544,12581,12800,12866,12869,12896,12946,12978,13131,13242,13244,13289,13297,13314,13408,13517,13597,13598,13689,13696,13716,13726,13823,13865,13933,14015,14050,14091,14239,14311,14392,14436,14444,14495,14497,14580,14649,14710,14974,15016,15018,15038,15071,15160,15248,15265,15273,15370,15427,15535,15612,15694,15703,15740,15787,16139,16141,16170,16202,16286,16407,16500,16527,16530,16554,16634,16751,16779,16851,16916,16965,16972,17020,17157,17245,17279,17306,17370,17390,17419,17437,17557,17665,17687,17841,17889,18019,18080,18084,18107,18156,18166,18269,18294,18411,18431,18442,18496,18591,18696,18720,18765,18860,18862,18886,18893,18935,18979,19020,19070,19079,19115,19128,19159,19223,19364,19430,19546,19568,19571,19655,19753,19794,19838,19872,19890,20019,20042,20047,20121,20141,20158,20200,20393,20527,20573,20580,20813,20833,20849,20945,21006,21260,21472,21775,21787,21865,21868,21889,22120,22164,22589,22673,22689,22771,22816,22886,22892,22987,22990,23008,23092,23199,23274,23276,23318,23384,23390,23405,23425,23537,23664,23687,23723,23900,23931,24052,24070,24159,24173,24212,24235,24336,24376,24473,24503,24558,24585,24586,24595,24625,24816,24822,24949,24960,24979,24994,25074,25139,25205,25223,25335,25337,25507,25514,25520,25532,25588,25621,25976,26024,26083,26264,26266,26273,26418,26433,26587,26639,26682,26845,26864,26882,26916,26983,27285,27313,27359,27373,27481,27542,27555,27676,27689,27783,27870,28177,28202,28227,28279,28565,28594,28705,28766,28856,29053,29072,29130,29179,29255,29270,29342,29417,29425,29433,29577,29606,29615,29666,29718,29785,29874,29910,30009,30023,30056,30081,30129,30144,30196,30283,30452,30455,30609,30730,30762,30800,30855,30974,31056,31211,31224,31260,31325,31392,31476,31491,31525,31565,31609,31660,31706,31801,31839,31869,31965,32055,32107,32131,32229,32494,32528,32605,32627,32729,32807,33013,33121,33212,33258,33268,33280,33299,33345,33458,33609,33645,33653,33737,33740,33797,33799,33833,33959,34128,34130,34141,34396,34740,34760,34856,34870,34898,34924,35154,35385,35405,35453,35566,35593,35831,35854,35877,35893,35901,35929,36104,36143,36228,36234,36362,36561,36714,36837,36847,36904,36930,37077,37241,37307,37398,37484,37531,37534,37647,37655,37744,37785,37809,37874,37934,38335,38406,38414,38416,38492,38557,38574,38586,38650,38656,38693,38716,38767,38780,38789,38800,38855,38871,38913,38942,39001,39033,39046,39050,39114,39191,39296,39520,39619,39680,39693,39721,39771,39835,39911,39917,39946,40051,40086,40108,40220,40236,40269,40389,40394,40502,40528,40544,40696,40741,40791,40839,40912,40959,41021,41125,41357,41450,41483,41500,41544,41614,41676,41873,41925,41943,41949,41958,42245,42299,42355,42361,42368,42384,42404,42463,42471,42499,42715,42753,42801,42809,42814,42950,42954,42966,43153,43283,43315,43319,43357,43456,43549,43550,43648,43824,43914,44064,44096,44132,44181,44213,44240,44269,44321,44382,44611,44639,44710,44883,44907,44911,44952,45055,45121,45187,45386,45557,45566,45605,45626,45685,45815,45868,46148,46351,46355,46368,46373,46408,46463,46478,46598,46614,46630,46655,46685,46743,46845,46853,46866,46875,47101,47267,47297,47443,47504,47516,47525,47536,47588,47636,47673,47704,47729,47751,47789,47804,47843,47879,47966,47979,48112,48140,48210,48214,48392,48422,48445,48600,48710,48727,48743,48893,49065,49140,49157,49225,49259,49269,49318,49340,49617,49637,49761,49806,49859,49902,49948,49982,50016,50101,50121,50203,50213,50367,50490,50500,50530,50628,50659,50676,50709,50768,50825,50852,50902,51016,51049,51123,51127,51150,51207,51216,51224,51286,51355,51479,51534,51592,51600,51651,51688,51705,51780,51790,51935,51936,51955,52054,52152,52155,52385,52448,52480,52491,52702,52770,52815,52819,52840,52894,52990,53006,53044,53213,53269,53282,53307,53366,53400,53410,53460,53494,53536,53614,53715,53723,53729,53790,53901,54101,54106,54275,54352,54397,54398,54399,54582,54735,54741,54851,54898,54899,54927,55065,55155,55162,55187,55190,55206,55438,55488,55489,55642,55645,55699,55783,55838,56008,56044,56060,56170,56219,56226,56278,56281,56348,56456,56462,56472,56534,56547,56686,56749,56820,56855,56876,56886,56949,56957,57117,57187,57223,57271,57300,57383,57404,57450,57532,57533,57700,57737,57809,58029,58239,58337,58398,58460,58574,58643,58781,58929,58965,59012,59022,59054,59071,59093,59100,59165,59279,59336,59527,59576,59583,59602,59754,59813,59855,59873,60314,60399,60426,60428,60455,60466,60467,60478,60483,60599,60607,60798,60852,60979,61019,61036,61126,61195,61261,61303,61379,61592,61652,61664,61770,61842,61876,61927,61968,62081,62108,62145,62149,62173,62272,62276,62296,62367,62531,62616,62919,62964,62990,63107,63213,63273,63361,63424,63448,63502,63594,63670,63679,63796,63896,63903,64025,64040,64214,64231,64285,64331,64374,64460,64477,64495,64677,64678,64750,64798,64831,64854,64870,64932,65072,65077,65158,65208,65426,65461,65556,65567,65579,65661,65692,65904,65935,65936,66061,66114,66169,66239,66309,66392,66650,66705,66726,66737,66777,66807,66819,66824,66914,66933,67036,67111,67148,67155,67191,67251,67288,67307,67360,67388,67472,67558,67605,67653,67906,67931,67947,67983,67986,68025,68048,68072,68148,68160,68226,68394,68401,68445,68516,68577,68582,68878,68931,69097,69302,69408,69501,69538,69611,69651,69796,69821,69855,69872,69888,69983,70133,70177,70222,70241,70434,70460,70511,70587,70779,70825,70981,71023,71061,71155,71187,71338,71345,71475,71516,71583,71641,71716,71768,71789,71952,71965,72036,72042,72128,72131,72174,72205,72225,72261,72339,72426,72461,72598,72706,72755,72938,72939,73179,73421,73426,73435,73472,73553,73575,73742,73757,73778,73876,74086,74328,74355,74383,74384,74446,74460,74557,74951,74975,75051,75114,75182,75192,75284,75301,75345,75380,75411,75459,75468,75549,75609,75687,75796,75976,75984,76164,76243,76265,76388,76466,76531,76710,76789,76803,76820,77088,77117,77163,77275,77597,77645,77780,77810,77924,77933,77987,78117,78323,78338,78355,78383,78396,78523,78936,78957,79028,79061,79064,79241,79267,79290,79318,79372,79422,79475,79552,79589,79628,79671,79842,79894,79955,79999,80024,80068,80111,80145,80263,80297,80454,80461,80516,80521,80548,80588,80626,80738,81005,81038,81319,81344,81380,81388,81399,81609,81633,81916,81972,82038,82179,82294,82521,82614,82709,82725,82806,83165,83263,83289,83341,83380,83413,83440,83508,83595,83726,83766,83814,83837,84420,84529,84657,84661,84695,84722,84823,84850,84933,85011,85053,85152,85172,85179,85323,85423,85541,85607,85614,85821,85838,85890,85980,85996,86111,86163,86269,86535,86628,86715,86844,86916,86940,87008,87027,87061,87072,87193,87344,87480,87488,87651,87829,87838,87953,88128,88215,88237,88332,88350,88476,88515,88526,88543,88656,88678,88862,88931,89129,89171,89247,89253,89275,89339,89501,89520,89594,89646,89685,89704,89841,89857,89936,89961,90007,90104,90168,90170,90341,90355,90373,90414,90493,90494,90539,90574,90675,90705,90737,90752,90809,90998,91187,91265,91329,91389,91405,91456,91529,91592,91595,91664,91713,91766,91830,91886,91921,91955,92045,92068,92110,92276,92294,92437,92448,92527,92577,92584,92590,92617,92640,92701,92748,92853,92958,92973,93090,93109,93116,93189,93190,93215,93303,93486,93567,93581,93589,93634,93703,93707,93863,93970,94061,94068,94120,94140,94302,94415,94647,94679,94774,94853,94865,94903,95079,95137,95158,95246,95270,95565,95639,95689,95719,95754,95763,95859,96184,96363,96462,96541,96589,96780,96807,96825,96832,96898,96917,96947,96949,97039,97066,97179,97299,97375,97527,97583,98097,98106,98192,98226,98396,98479,98695,98965,98981,99181,99189,99221,99297,99334,99364,99399,99418,99512,99589,99617,99622,99645,99800,99917,99934",
          "275,316,392,407,411,425,448,470,486,546,608,613,683,711,742,818,840,854,934,999,1035,1204,1250,5034,5201,5207,5211,5230,5257,5287,5301,5316,5329,5356,5384,5426,5477,5523,5597,5632,5765,5770,5772,5787,5803,5821,5826,5888,5908,5929,5967,5988,6035,6190,6411,6731,6771,6832,6968,7013,7078,7109,7134,7152,7287,7384,7467,7534,7641,7839,8284,8440,8542,8594,8624,8749,8824,8856,8914,9094,9406,9533,9620,9695,9859,10200,10402,10552,10588,11103,11823,11948,12126,12400,12729,13546,14170,15548,17118,18528,19231,20940,20968,24096,26671,29483,32996,41556,42357,50003,50143,50225,50375,50523,50611,50615,50680,50745,50879,50943,51105,51113,51352,51435,51575,51663,51675,51691,51727,51752,51776,52015,52104,52253,52317,52501,52601,52718,52729,52784,52968,53116,53178,53321,53339,53442,53589,53674,53753,53813,53817,53913,54065,54127,54144,54284,54337,54348,54419,54424,54441,54459,54606,54626,54652,54705,54730,54872,54894,54965,54966,54983,55129,55173,55180,55237,55256,55378,55456,55553,55738,55759,55779,55793,55941,56081,56142,56247,56379,56481,56495,56585,56696,56804,56875,56944,57064,57332,57413,57444,57453,57619,57664,57875,57913,58094,58235,58360,58375,58380,58536,58559,58648,58651,58711,58761,58880,59140,59216,59277,59380,59395,59445,59726,59824,59842,59889,59930,60084,60157,60187,60215,60294,60460,60620,60621,60687,60757,60768,60851,60905,60942,61167,61179,61260,61298,61322,61326,61412,61465,61567,61616,61758,61897,62011,62450,62481,62535,62705,62729,62730,62976,62999,63277,63353,63412,63528,63557,63574,63788,63881,63891,63894,63948,63996,64007,64089,64134,64161,64163,64188,64287,64312,64367,64429,64527,64646,64764,64795,64863,65111,65247,65395,65442,65443,65452,65515,65518,65660,65730,65780,65865,65887,66048,66264,66451,66494,66641,66710,67140,67173,67285,67368,67458,67473,67499,67509,67543,67556,67888,68006,68020,68310,68338,68562,68568,68614,68818,68999,69301,69376,69431,69545,69681,69747,69760,70033,70101,70148,70290,70510,70543,70649,70725,70758,70794,70827,70909,70983,71006,71231,71243,71499,71552,71574,71760,71896,71955,71975,72063,72078,72150,72260,72290,72338,72343,72403,72420,72464,72589,72620,72758,72762,72887,72976,73001,73004,73103,73106,73227,73236,73286,73320,73425,73475,73477,73599,73619,73701,73750,73854,73972,74056,74100,74137,74163,74229,74463,74720,74722,74919,75044,75079,75109,75178,75305,75431,75463,75490,75495,75598,75641,75665,75867,75913,76049,76073,76172,76222,76248,76254,76347,76374,76379,76438,76450,76674,76764,76771,76806,76875,76887,76949,77144,77176,77226,77234,77277,77315,77412,77473,77528,77561,77674,77720,77828,77831,77979,78041,78150,78231,78241,78339,78458,78472,78483,78553,78947,78954,79013,79026,79079,79090,79101,79338,79362,79790,79844,80020,80196,80258,80285,80317,80360,80490,80593,80635,80866,80903,80961,80985,81009,81047,81055,81231,81289,81360,81575,81593,81642,81674,81693,81830,81852,81927,81967,82025,82104,82176,82227,82228,82275,82293,82384,82506,82648,82658,82746,82824,82995,83029,83123,83172,83212,83253,83374,83611,83629,83657,83661,83681,83722,83732,83857,83922,84327,84437,84439,84442,84531,84721,84770,84904,85129,85205,85324,85354,85664,85732,85928,85960,86029,86147,86205,86310,86400,86516,86579,86679,86721,86831,86966,87007,87090,87108,87113,87178,87240,87469,87543,87561,87599,87676,87681,88033,88065,88113,88130,88151,88389,88435,88571,88576,88587,88595,88731,88879,89005,89020,89028,89055,89141,89433,89437,89461,89631,89671,89751,89754,89826,90069,90092,90288,90372,90518,90519,90660,90667,90975,90992,91201,91229,91249,91278,91295,91412,91415,91440,91593,91616,91647,91756,91823,91915,91946,92027,92081,92104,92128,92189,92319,92735,92810,92848,93051,93148,93181,93239,93254,93296,93345,93368,93484,93497,93542,93766,93779,93793,94049,94134,94191,94370,94446,94456,94718,94761,94917,94918,95019,95059,95243,95621,95841,95919,96021,96026,96099,96105,96198,96283,96286,96649,96714,96798,96873,96971,97130,97176,97208,97331,97382,97494,97791,97796,97865,97868,97958,98148,98164,98241,98507,98558,98567,98615,98654,98696,98846,98877,98996,99152,99315,99381,99412,99727,99754,99819,99828,99835,99860,99866,99892",
          "72,402,550,653,766,767,795,919,983,1028,1295,1395,1460,1491,1597,1656,1669,1712,1732,1792,1889,1952,1987,2038,2118,2219,2264,2337,2338,2415,2453,2536,2548,2555,2576,2827,2865,2883,2959,3216,3306,3425,3729,3807,3835,4191,4202,4245,4434,4466,4468,4496,4514,4670,4697,4875,4876,4912,5266,5276,5303,5605,5829,6203,6213,6234,6339,6354,6615,6704,6742,6783,6819,6848,6859,6890,6902,6909,7096,7128,7172,7275,7495,7560,7561,7616,7673,7803,7912,7914,7926,7997,8251,8298,8301,8341,8399,8414,8434,8452,8479,8482,8493,8620,8644,8719,8738,8837,8850,8859,9045,9244,9321,9338,9440,9550,9565,9574,9640,9809,9819,10174,10360,10378,10394,10465,10506,10517,10669,10840,11018,11114,11533,11615,11632,11737,11822,11880,11928,11946,12157,12179,12530,12548,12669,12989,13072,13206,13349,13407,13499,13548,13562,13612,13632,13646,13648,13662,13694,13700,13783,13876,13930,13970,13988,14039,14041,14084,14089,14108,14126,14132,14265,14297,14320,14442,14481,14519,14635,14729,14741,14789,14868,14901,14931,15056,15230,15278,15430,15539,15576,15652,15711,15771,15806,15899,15934,16000,16069,16142,16171,16240,16368,16385,16580,16590,16607,16632,16680,16711,16806,16871,16904,17011,17089,17102,17109,17257,17457,17523,17532,17613,17632,17761,18039,18121,18505,18593,18594,18724,18855,18873,18908,19064,19133,19137,19149,19171,19213,19368,19467,19481,19499,19558,19569,19594,19599,19813,19846,19909,19942,19952,20119,20205,20479,20596,20645,20671,20702,20820,20862,20877,20926,20965,20987,21216,21227,21238,21261,21266,21367,21378,21490,21512,21587,21611,21612,21653,21710,21785,21835,21850,21877,21969,21973,21997,22015,22055,22138,22207,22250,22282,22283,22389,22445,22453,22529,22675,22811,22971,23055,23064,23310,23336,23397,23458,23503,23576,23675,23682,23801,24001,24058,24094,24228,24495,24526,24527,24547,24608,24610,24714,24774,24795,24848,24957,25006,25092,25103,25177,25312,25494,25619,25730,25739,25791,25795,25929,26057,26078,26250,26289,26328,26340,26398,26487,26559,26662,26729,26739,26811,26818,26914,26941,27007,27013,27026,27121,27257,27386,27666,27684,27698,27762,27819,27851,27872,27881,27949,27995,28168,28208,28234,28331,28332,28425,28536,28562,28592,28608,28709,28757,28781,28948,28966,29028,29048,29081,29129,29178,29277,29380,29429,29530,29587,29769,29784,29850,29861,29867,29993,30274,30288,30304,30407,30512,30513,30633,30652,30726,30778,30840,30859,30920,30949,31106,31136,31176,31299,31483,31677,31757,31758,31761,31836,31922,31984,32212,32238,32335,32338,32412,32434,32544,32549,32599,32728,32921,32972,33051,33055,33060,33146,33179,33190,33310,33399,33437,33547,33561,33680,33708,33717,33755,33832,33932,34089,34092,34152,34276,34334,34470,34576,34600,34658,34815,34893,35033,35199,35208,35309,35795,36089,36148,36163,36203,36245,36433,36439,36630,36722,36795,36836,36925,36941,36971,37035,37084,37352,37416,37521,37523,37540,37546,37622,37694,37901,38025,38090,38146,38167,38220,38275,38297,38517,38572,38593,38692,38718,38729,38748,38878,39226,39231,39363,39447,39493,39537,39602,39630,39699,39748,39767,39780,39787,40283,40343,40390,40421,40464,40580,40642,40720,40809,40948,41185,41203,41221,41225,41292,41320,41363,41410,41601,41738,41746,41808,41923,41963,41990,41999,42150,42198,42276,42344,42387,42502,42573,42656,42674,42703,42874,43033,43188,43524,43532,43598,43616,43803,43996,44035,44040,44087,44396,44412,44561,44653,44752,44889,45042,45045,45071,45116,45144,45361,45428,45466,45526,45584,45723,45731,45850,45921,45995,46018,46019,46103,46217,46238,46276,46308,46337,46338,46391,46479,46559,46561,46679,46755,46776,46848,46862,46918,47002,47096,47194,47314,47397,47440,47512,47565,47604,47624,47631,47659,47720,47792,47823,47868,47924,47971,48005,48175,48321,48332,48381,48641,48758,48772,48794,48814,48964,48986,49005,49218,49411,49461,49510,49788,49823,49836,49916,49942,49968,50044,50056,50096,50285,50305,50379,50417,50451,50569,50641,50646,50686,50707,50711,50895,51022,51179,51204,51261,51510,51524,51529,51567,51579,51597,51667,51673,51712,51892,52089,52247,52298,52415,52713,52865,52883,52989,53121,53412,53440,53597,53610,53644,53646,53743,53850,54050,54152,54175,54327,54384,54569,54661,54700,54760,54800,54838,54924,55058,55199,55293,55308,55321,55400,55422,55534,55538,55624,55693,55710,55785,56043,56050,56205,56291,56328,56413,56448,56485,56578,56673,56685,56718,56822,56913,57039,57231,57367,57409,57422,57521,57611,57654,57659,57670,57680,57943,57965,58063,58168,58506,58512,58541,58552,58601,58734,58760,58763,58770,58973,59027,59116,59238,59243,59358,59468,59661,59717,59751,59830,59868,59920,59941,60028,60167,60275,60295,60300,60421,60469,60671,60676,60735,60763,60858,60941,60969,61018,61030,61076,61210,61327,61376,61405,61410,61541,61546,61784,61836,62016,62046,62148,62161,62324,62399,62649,62715,62725,62784,62815,62842,62856,63305,63477,63483,63486,63603,63696,63745,63840,63864,63872,64002,64108,64138,64139,64159,64179,64342,64514,64582,64597,64923,65043,65157,65216,65244,65270,65424,65721,65962,66032,66204,66343,66379,66444,66572,66666,66812,66848,67041,67052,67055,67138,67147,67184,67471,67489,67751,67818,67918,68061,68081,68117,68210,68240,68375,68443,68632,68695,68756,68847,68860,68871,68952,68960,69115,69723,69752,70217,70223,70250,70396,70494,70662,70679,70821,70984,71087,71122,71383,71427,71518,71634,71646,71655,71721,71826,72019,72040,72138,72148,72218,72220,72240,72378,72451,72574,72647,72680,72817,73025,73102,73127,73182,73198,73256,73327,73451,73460,73505,73566,73574,73592,73648,73738,73866,73932,73949,73964,73978,74075,74089,74126,74191,74210,74332,74541,74564,74597,74743,74765,74874,74912,75037,75058,75210,75363,75500,75509,75603,75851,75897,75959,76008,76190,76196,76226,76239,76287,76354,76644,76745,76753,76785,76831,76863,77089,77153,77233,77338,77469,77559,77619,77627,77660,77726,77770,77793,77869,77886,77966,78011,78105,78162,78198,78245,78320,78442,78443,78562,78695,78726,78837,78866,78879,78907,79005,79022,79037,79120,79151,79201,79240,79316,79390,79402,79582,79665,79893,79925,79967,80066,80249,80316,80351,80488,80614,80618,80651,80665,80703,80707,80775,80799,80981,81074,81151,81197,81314,81417,81546,81762,82184,82195,82215,82297,82361,82444,82532,82555,82774,82814,82853,83131,83164,83200,83230,83276,83288,83418,83480,83529,83599,83773,83861,83891,84028,84077,84167,84291,84343,84541,84602,84654,84705,84791,84885,84957,84989,85093,85296,85375,85391,85414,85516,85648,86085,86389,86410,86455,86540,86978,87016,87021,87057,87176,87231,87266,87294,87356,87429,87464,87855,87881,87940,87982,88046,88122,88191,88205,88211,88261,88278,88314,88375,88549,88674,88699,88722,88723,88732,88742,88755,88793,89021,89151,89328,89442,89463,89483,89504,89550,89651,89714,89872,89886,89889,90246,90356,90429,90482,90522,90624,90641,90994,91050,91145,91331,91658,91732,91865,91900,91953,92025,92062,92105,92113,92221,92226,92425,92432,92542,92647,92867,92869,92911,92980,93015,93073,93089,93236,93250,93302,93325,93341,93400,93575,93610,93701,93712,93804,93835,93905,93928,94158,94203,94208,94286,94322,94397,94400,94558,94877,94986,95029,95105,95206,95209,95258,95352,95377,95420,95453,95895,95904,96167,96291,96330,96432,96477,96520,96702,96720,96769,97146,97163,97165,97276,97312,97474,97481,97526,97625,97660,98172,98220,98236,98243,98262,98330,98364,98390,98430,98607,98931,98990,99494,99510,99602,99711,99775,99922,99954,99995",
          "19,274,284,348,419,452,455,466,485,506,551,594,614,723,773,797,838,915,992,1194,1379,1403,1535,5081,5181,5268,5275,5307,5309,5315,5319,5323,5364,5391,5402,5483,5580,5619,5792,5797,5849,5947,6124,6262,6495,6524,6542,6592,6601,6632,6698,6746,7002,7041,7083,7117,7341,7362,7423,7545,7575,7576,7604,7650,7724,8080,8149,8528,8590,8847,8968,8979,9251,9736,9921,9951,10089,10177,10527,10572,10627,10983,11019,11085,11102,12252,12551,13082,14129,14160,14173,14378,14435,15383,16227,16702,19276,19866,21456,24375,29644,30676,35173,41519,43341,50023,50179,50208,50217,50281,50454,50502,50643,50818,50872,51012,51160,51232,51276,51724,51749,51838,51887,51928,51952,52034,52097,52101,52154,52315,52363,52451,52661,52776,52787,52789,52917,52983,53018,53127,53191,53308,53489,53500,53624,53654,53700,53750,53801,53875,54148,54271,54290,54336,54522,54623,54641,54859,54897,54904,54990,55122,55167,55248,55300,55361,55404,55482,55834,55894,56179,56220,56236,56277,56339,56362,56669,56762,56884,56922,56964,57157,57347,57549,57593,57602,57793,57828,57855,57917,58131,58317,58498,58566,58592,58649,58715,58951,59031,59160,59276,59310,59338,59438,59496,59534,59649,59806,59896,59983,60101,60219,60375,60601,60702,60775,60836,61455,61621,61737,61827,61840,61868,61871,61960,62050,62084,62090,62105,62350,62366,62372,62416,62517,62611,62675,62733,62854,62887,63054,63062,63135,63415,63492,63556,63711,63746,63879,63956,64127,64147,64185,64279,64307,64350,64425,64471,64517,64588,64731,64836,64839,65032,65112,65164,65170,65204,65349,65357,65358,65449,65568,65653,65698,65926,65957,65974,65997,66238,66301,66488,66542,66569,66576,66678,66744,66760,66867,67019,67085,67203,67317,67342,67390,67462,67641,67689,67851,68058,68102,68126,68131,68234,68293,68385,68469,68550,68640,68645,68685,68740,68849,68972,68990,69173,69268,69332,69346,69436,69464,69771,69924,70014,70021,70115,70402,70404,70411,70584,70810,70838,70907,70913,70934,70978,71127,71144,71241,71285,71318,71440,71474,71593,71714,71805,71846,72199,72219,72288,72422,72431,72481,72495,72746,72839,72954,72968,73092,73249,73335,73677,73914,74066,74221,74286,74494,74594,75364,75443,75678,75974,76007,76177,76751,76776,76787,77061,77080,77101,77115,77257,77403,77413,77439,77518,77705,78049,78201,78267,78296,78297,78329,78461,78580,78588,78616,78701,78717,78760,78790,78832,78864,78961,78971,79001,79068,79091,79309,79339,79389,79403,79651,79661,79813,79871,79994,80064,80281,80282,80445,80560,80695,80839,81021,81352,81488,81548,81574,81730,81733,81761,81928,82004,82203,82257,82285,82314,82475,82642,82702,82717,82915,82945,83000,83022,83146,83273,83375,83449,83570,83621,83712,83813,83907,83962,84020,84081,84088,84152,84181,84196,84314,84325,84392,84712,84849,84943,84963,85005,85042,85243,85383,85465,85565,85573,85593,85630,85671,85788,85826,85919,85959,85982,86120,86225,86255,86292,86422,86427,86491,86610,86661,86692,86772,86953,87038,87117,87128,87134,87142,87225,87396,87499,87601,87622,87708,87971,88115,88147,88167,88230,88423,88517,88632,88680,88773,88877,88900,88932,88954,89248,89260,89538,89654,89679,90019,90110,90318,90527,90586,90610,90663,90949,90990,91192,91273,91307,91367,91429,91455,91491,91562,91648,91742,91851,91928,92048,92098,92100,92118,92193,92273,92328,92332,92445,92467,92531,92683,92707,92781,92818,92874,92927,92993,93078,93108,93277,93278,93310,93366,93389,93390,93632,93677,93729,93892,93922,94029,94181,94241,94602,94653,94824,94872,95050,95094,95434,95464,95503,95673,95678,95682,95850,96008,96014,96133,96134,96169,96190,96346,96463,96657,96665,96687,96839,96935,96941,96967,96977,97156,97349,97762,97790,97996,98002,98062,98129,98278,98394,98481,98492,98512,99127,99144,99176,99301,99311,99678,99717,99948",
          "40,99,111,120,127,132,134,210,223,244,583,659,877,879,980,1113,1400,1456,1493,1619,1807,1942,1954,2048,2094,2110,2189,2198,2226,2254,2339,2373,2428,2434,2640,2692,2725,2783,2804,2808,2821,2838,2931,2950,3043,3052,3080,3103,3191,3293,3295,3355,3368,3429,3487,3534,3616,3622,3650,3785,3806,3876,3975,3988,3998,4210,4231,4273,4322,4324,4361,4418,4430,4480,4492,4500,4510,4515,4576,4717,4736,4827,4857,4914,5011,5035,5095,5296,5317,5385,5484,5486,5639,5712,5840,5902,5998,6080,6102,6184,6249,6308,6535,6686,6713,6723,6780,6781,6988,7011,7058,7076,7272,7386,7481,7487,7582,7634,7765,7782,7858,7942,8202,8420,8472,8496,8498,8605,8653,8730,8817,9121,9249,9264,9311,9492,9541,9731,9901,10068,10251,10254,10395,10430,10460,10480,10558,10640,10695,10710,10780,10791,10991,11075,11094,11129,11139,11412,11520,11529,11565,11579,11826,12207,12308,12407,12447,12520,12594,12596,12666,12674,12688,12707,12736,12755,12832,12945,13035,13081,13117,13173,13223,13251,13307,13316,13346,13443,13519,13656,13961,13978,13996,14024,14040,14070,14110,14137,14200,14267,14269,14336,14402,14431,14462,14526,14528,14583,14600,14608,14685,14704,14758,14843,14871,14879,14966,15000,15062,15094,15122,15126,15141,15148,15155,15182,15208,15262,15264,15510,15604,15655,15659,15736,15856,16218,16237,16397,16402,16423,16499,16509,16529,16544,16769,16817,16995,17001,17041,17078,17168,17195,17266,17322,17329,17349,17385,17396,17401,17416,17470,17484,17537,17568,17603,17655,17666,17679,17764,17835,17943,17948,18051,18114,18170,18238,18239,18254,18263,18334,18355,18479,18527,18540,18688,18728,18750,18756,19095,19132,19140,19252,19287,19324,19374,19525,19640,19696,19768,19859,19980,20015,20053,20064,20117,20173,20253,20261,20396,20453,20490,20528,20611,20641,20729,20757,20762,20819,21039,21061,21263,21276,21302,21317,21361,21452,21459,21506,21551,21560,21597,21602,21638,21746,21824,21854,21938,21944,22031,22057,22090,22154,22204,22214,22264,22418,22434,22473,22526,22530,22679,22734,22758,22769,22903,23003,23005,23142,23178,23231,23236,23344,23367,23388,23471,23476,23483,23688,23730,23755,23819,23883,23887,23955,23957,24118,24121,24181,24236,24275,24288,24312,24335,24519,24542,24561,24742,24914,24921,24983,24996,25056,25067,25178,25192,25197,25228,25456,25478,25561,25572,25586,25766,25855,25881,25913,25974,25975,26030,26129,26157,26193,26215,26218,26219,26271,26317,26363,26450,26476,26597,26659,26692,26750,26753,26797,26942,26994,27010,27023,27108,27180,27374,27381,27499,27544,27575,27589,27594,27629,27701,27718,27761,27830,27873,27890,27892,27907,27929,28007,28130,28170,28180,28209,28230,28317,28339,28347,28459,28506,28586,28869,28953,29013,29031,29099,29132,29153,29162,29209,29289,29311,29339,29369,29389,29391,29407,29443,29475,29557,29594,29610,29668,29733,29745,29797,29894,29905,29939,29954,30020,30042,30059,30063,30067,30184,30235,30247,30262,30280,30324,30340,30473,30663,30665,30673,30689,30737,30956,30978,31000,31001,31018,31060,31097,31184,31237,31345,31386,31394,31445,31532,31549,31550,31569,31723,31732,31788,31798,31806,31822,31844,31866,32047,32124,32161,32176,32179,32231,32265,32274,32287,32296,32305,32308,32328,32357,32371,32373,32379,32444,32448,32475,32553,32587,32600,32618,32648,32679,32702,32715,32769,32831,32861,32894,32903,32922,32941,32952,32961,32964,32968,32992,33008,33117,33187,33191,33229,33232,33236,33329,33340,33350,33466,33498,33567,33577,33720,33768,33839,33867,33870,33916,33972,33984,34022,34251,34320,34344,34351,34353,34429,34457,34560,34626,34766,34814,34900,35006,35039,35049,35184,35188,35218,35271,35346,35391,35423,35428,35521,35531,35562,35595,35612,35621,35753,35833,35991,36081,36084,36192,36197,36204,36256,36266,36274,36283,36304,36344,36448,36467,36536,36543,36556,36569,36574,36612,36624,36651,36655,36711,36748,36764,36787,36791,36823,36851,36861,36877,36908,36910,37021,37060,37095,37101,37142,37185,37258,37275,37282,37296,37298,37299,37306,37341,37359,37415,37453,37637,37678,37704,37745,37764,37829,37887,37894,37981,37993,38004,38083,38136,38141,38181,38226,38259,38265,38266,38282,38283,38301,38433,38453,38506,38551,38590,38627,38640,38686,38697,38717,38742,38790,38823,38830,39020,39062,39076,39146,39186,39195,39218,39327,39414,39439,39461,39477,39485,39553,39595,39596,39624,39668,39744,39847,39894,40007,40040,40081,40089,40225,40286,40307,40309,40339,40358,40418,40427,40477,40506,40553,40574,40639,40774,40777,40783,40834,40930,41055,41132,41165,41208,41210,41277,41336,41381,41386,41433,41529,41664,41684,41731,41750,41762,41782,41785,41825,41842,41843,41850,41851,41922,42026,42038,42052,42065,42091,42156,42188,42232,42268,42306,42335,42362,42370,42377,42390,42435,42503,42510,42541,42605,42621,42704,42721,42752,42765,42798,42803,42823,42868,42921,42959,42993,42995,43001,43068,43071,43111,43113,43157,43226,43286,43294,43311,43378,43411,43420,43492,43500,43508,43520,43644,43664,43771,43814,43881,43926,43949,43950,43961,43995,44000,44029,44142,44169,44197,44257,44320,44344,44347,44387,44409,44462,44560,44567,44585,44666,44676,44707,44712,44737,44828,44842,44847,44877,44910,44922,44974,45030,45139,45141,45143,45189,45243,45282,45351,45400,45437,45485,45509,45604,45749,45800,45828,45831,45862,45869,45872,45906,45976,45992,46145,46201,46501,46526,46538,46730,46800,46832,46925,46961,46972,47013,47054,47081,47083,47085,47144,47224,47263,47406,47410,47424,47449,47457,47482,47527,47541,47548,47608,47622,47648,47695,47835,47842,47849,47974,48054,48263,48300,48324,48347,48360,48644,48700,48879,49030,49160,49375,49408,49512,49628,49659,49667,49685,49755,49858,49897,49936,50247,50272,50274,50345,50350,50439,50440,50452,50549,50559,50678,50695,50713,50722,50843,50909,50933,50940,51060,51072,51115,51177,51195,51228,51291,51306,51436,51497,51755,51901,51912,51921,52022,52062,52112,52122,52269,52487,52538,52584,52690,52692,52788,52872,52886,52895,52971,53061,53063,53075,53229,53278,53404,53409,53419,53468,53504,53537,53545,53839,53902,53939,54032,54033,54069,54130,54133,54141,54293,54347,54396,54458,54683,54712,54854,54969,55106,55163,55386,55616,55639,55824,55911,56121,56180,56280,56390,56531,56532,56660,56909,56980,56999,57019,57055,57259,57266,57364,57374,57434,57594,57607,57666,57711,57953,58018,58140,58145,58179,58189,58213,58299,58341,58430,58517,58639,58640,58670,58707,58716,58867,58904,59164,59227,59237,59297,59392,59421,59627,59687,59958,59985,60110,60211,60318,60327,60511,60514,60679,60742,60804,60816,60817,60992,61186,61403,61425,61557,61643,61785,61962,62028,62089,62097,62169,62246,62361,62549,62579,62581,62589,62619,62666,62752,62977,63031,63204,63309,63329,63404,63510,63543,63629,63941,64013,64031,64074,64234,64317,64397,64470,64587,64632,64684,64892,64971,65090,65189,65205,65245,65248,65283,65346,65376,65432,65563,65594,65677,65718,65792,65840,65842,65878,66008,66106,66144,66180,66209,66214,66298,66328,66331,66336,66472,66577,66801,66864,66948,67326,67364,67370,67513,67537,67550,67642,67760,67809,67845,68078,68093,68132,68135,68136,68149,68153,68244,68323,68345,68367,68668,68677,68817,68854,69003,69093,69222,69360,69513,69552,69571,69655,69884,69911,69965,69972,70010,70262,70279,70301,70316,70325,70590,70609,70740,70754,70816,70832,70842,70918,70941,70990,71132,71139,71150,71159,71210,71224,71422,71509,71553,71604,71606,71701,71782,71842,71919,72062,72117,72119,72172,72223,72310,72358,72411,72674,72683,72824,72867,72872,72899,72981,72990,73194,73222,73394,73410,73428,73452,73479,73490,73627,73673,73721,73727,73815,73816,73890,73963,73996,74174,74279,74292,74498,74504,74589,74599,74610,74668,74673,74731,74847,74911,74997,75073,75195,75218,75272,75285,75294,75394,75402,75440,75649,75828,75870,75889,75900,75946,75953,76009,76240,76257,76267,76285,76301,76357,76372,76465,76570,76571,76732,76765,76773,76780,76839,76866,77005,77081,77156,77189,77211,77227,77235,77358,77408,77570,77749,77757,77762,77829,77876,77929,77939,77990,78095,78109,78173,78209,78260,78357,78398,78514,78541,78618,78651,78960,78994,79081,79112,79262,79280,79297,79317,79337,79564,79694,79715,79792,79819,79892,79899,79921,80266,80406,80457,80472,80475,80522,80537,80549,80550,80569,80621,80701,80749,80944,81135,81320,81439,81651,81678,81695,81754,81818,81855,81896,81897,81946,81953,82144,82201,82207,82210,82299,82322,82337,82434,82471,82476,82546,82559,82662,82761,82795,82920,82923,82925,82927,82987,83031,83309,83381,83402,83447,83583,83585,83791,83816,83848,83975,84038,84047,84057,84125,84162,84233,84383,84391,84488,84539,84615,84748,85066,85073,85137,85199,85225,85234,85236,85267,85408,85444,85491,85584,85597,85719,85824,85944,85984,86010,86100,86115,86159,86284,86355,86370,86413,86557,86562,86645,86656,86739,86761,86981,86999,87028,87035,87077,87212,87246,87355,87432,87523,87560,87565,87683,87704,87715,87818,88010,88067,88169,88198,88208,88364,88481,88499,88527,88533,88686,88750,88758,88772,88775,88896,88906,88974,89034,89116,89237,89482,89496,89542,89686,89762,89801,89814,89887,89994,90233,90235,90360,90384,90386,90397,90436,90463,90617,90694,90894,90918,90970,91010,91162,91204,91310,91327,91338,91350,91361,91421,91559,91687,91788,91817,91828,91846,91893,92075,92164,92172,92178,92392,92457,92494,92569,92894,92943,93061,93201,93249,93349,93363,93647,93761,93870,93896,93901,93916,94037,94046,94163,94188,94201,94272,94274,94381,94418,94489,94561,94656,94737,94771,94880,95000,95081,95099,95101,95133,95309,95310,95331,95356,95385,95438,95667,95729,95732,95812,95834,95987,96130,96145,96170,96226,96233,96274,96288,96316,96337,96459,96481,96595,96601,96634,96646,96985,97101,97147,97238,97339,97461,97529,97563,97585,97617,97730,97761,97910,97945,98013,98221,98301,98890,98944,99011,99018,99090,99124,99218,99348,99349,99463,99474,99599,99900",
          "34,104,231,283,353,495,562,580,604,686,696,962,1058,1064,1119,1267,1288,1335,1372,1406,1465,1667,1763,1787,1856,1915,2004,2082,2161,2423,2472,2626,2709,2862,2889,2898,2901,2921,2941,3058,3116,3137,3151,3165,3172,3241,3358,3403,3406,3549,3638,3857,3914,4008,4032,4037,4182,4279,4334,4339,4388,4423,4863,4877,4882,5101,5184,5192,5510,5525,5606,5608,5681,5696,5716,5724,5861,5944,5970,6395,6400,6540,7009,7087,7118,7202,7229,7278,7439,7538,7619,7632,7704,7727,7728,7776,7885,7919,7976,7989,8072,8081,8189,8474,8565,8621,8637,8654,8685,8716,8721,8828,8853,8916,9071,9158,9382,9415,9581,9658,9741,9817,9818,9841,9902,9945,9955,10148,10358,10366,10564,10582,10596,10694,10726,10735,10859,10863,10921,10946,10981,11095,11228,11263,11269,11272,11290,11376,11386,11494,11611,11626,11748,11875,11905,11982,12057,12449,12479,12637,12678,12732,12750,12790,12791,12826,13020,13046,13079,13177,13211,13214,13234,13344,13482,13543,13620,13647,13668,13684,13757,13770,13833,13837,13845,13890,13911,14068,14076,14316,14348,14370,14397,14454,14468,14505,14565,14588,14611,14630,14652,14770,14775,14831,14915,14922,14988,15080,15088,15132,15426,15629,15709,15902,15910,15923,16053,16075,16106,16223,16233,16380,16382,16525,16649,16675,16692,16697,16765,16819,16864,16948,16983,17002,17038,17108,17116,17226,17236,17333,17522,17619,17743,17886,17929,17934,17996,18028,18097,18117,18167,18174,18178,18219,18262,18274,18356,18408,18414,18569,18632,18739,18740,18826,18914,18977,19033,19056,19111,19173,19328,19397,19486,19615,19627,19670,19692,19738,19762,19785,19804,19810,19837,19976,19997,20007,20087,20153,20155,20156,20197,20215,20311,20385,20408,20686,20717,21110,21114,21125,21186,21201,21231,21285,21315,21402,21410,21420,21449,21485,21676,21684,21794,21806,21814,21890,21910,21928,22096,22130,22181,22195,22230,22268,22370,22385,22390,22394,22523,22557,22583,22609,22781,22801,22960,23042,23082,23132,23263,23321,23329,23333,23365,23386,23419,23426,23511,23518,23540,23555,23556,23568,23624,23750,23902,23952,24091,24113,24328,24453,24613,24648,24650,24652,24747,24840,24875,24885,24890,24952,24962,24973,25031,25035,25075,25169,25218,25298,25462,25481,25493,25509,25548,25550,25707,25746,25793,25800,25836,25858,25966,26005,26073,26108,26130,26138,26175,26178,26183,26195,26255,26292,26355,26462,26580,26688,26800,26804,26928,27017,27103,27141,27144,27311,27354,27514,27537,27603,27657,27705,27774,27775,27827,27852,27922,27986,28037,28076,28086,28088,28144,28151,28190,28225,28309,28371,28457,28474,28544,28554,28596,28623,28806,28826,29022,29026,29047,29061,29116,29217,29302,29329,29449,29482,29515,29523,29560,29565,29614,29802,29808,29940,30007,30016,30122,30174,30239,30267,30409,30518,30601,30611,30613,30658,30708,30799,30809,30841,30864,30879,30942,30998,31057,31123,31202,31255,31275,31323,31342,31417,31516,31620,31673,31686,31703,31846,31849,31876,31891,31904,31961,32006,32031,32034,32062,32080,32148,32226,32517,32657,32763,32849,32901,32925,33062,33420,33497,33532,33593,33637,33650,33798,33815,34003,34031,34258,34265,34267,34296,34332,34388,34446,34510,34541,34580,34582,34686,34692,34872,34933,35079,35464,35592,35630,35745,35746,35781,35806,35864,35865,36055,36062,36181,36492,36521,36609,36774,36827,36864,37025,37147,37430,37625,37627,37705,38127,38294,38316,38354,38552,38578,38660,38684,38685,38792,38795,38836,38895,38998,39017,39099,39271,39392,39436,39440,39450,39645,39704,39718,39747,39927,40074,40106,40344,40355,40576,40624,40652,40689,40755,40790,40823,40831,40868,40925,41052,41086,41131,41145,41152,41370,41371,41427,41444,41485,41517,41542,41548,41633,41662,41803,41828,41957,42086,42100,42207,42216,42251,42352,42434,42518,42678,42788,42848,42871,42947,43067,43369,43400,43429,43528,43552,43746,43787,43791,43863,44110,44202,44371,44713,44751,45095,45149,45157,45190,45267,45274,45280,45600,45852,45866,45916,45936,46256,46350,46379,46676,46712,46759,46767,46968,47042,47057,47172,47223,47231,47255,47278,47280,47493,47575,47643,47677,47678,47686,47691,47755,47768,47818,47915,47939,48012,48018,48226,48291,48310,48350,48351,48354,48418,48538,48620,48822,48839,48941,49017,49088,49133,49373,49378,49477,49550,49563,49566,49705,49779,49783,49795,49899,49909,49997,50030,50053,50149,50171,50182,50258,50400,50413,50420,50532,50634,50736,50742,50774,50866,50926,51019,51394,51565,51681,51756,51821,51824,52064,52072,52087,52288,52328,52347,52366,52374,52417,52456,52496,52497,52566,52578,52585,52725,52863,52896,53226,53331,53362,53388,53445,53492,53499,53608,53611,53676,53760,53852,53868,53935,54029,54192,54207,54487,54497,54505,54511,54663,54679,54690,54691,54842,54936,54944,55005,55043,55130,55236,55502,55571,55587,55731,55960,55963,55975,56092,56112,56186,56308,56417,56539,56559,56587,56712,56713,56821,56915,57066,57113,57136,57166,57193,57196,57293,57326,57344,57437,57494,57575,57624,57689,57694,57703,57736,57882,57888,58057,58076,58139,58220,58238,58310,58357,58576,58752,58772,58833,58851,58925,59097,59103,59139,59168,59206,59293,59337,59418,59487,59511,59616,59670,59705,59888,59916,59931,59946,60010,60170,60190,60222,60287,60323,60344,60376,60490,60585,60634,60818,60889,61109,61251,61281,61309,61382,61636,61821,61918,62072,62282,62290,62331,62358,62507,62708,62767,62776,62825,62914,62932,62971,63133,63162,63167,63217,63230,63239,63252,63300,63364,63410,63447,63638,63675,63758,63766,63966,64065,64290,64404,64435,64536,64687,64746,64761,64803,65138,65156,65181,65191,65210,65299,65311,65404,65464,65486,65489,65504,65530,65646,65822,65845,65859,65945,65990,66005,66050,66073,66193,66199,66252,66297,66441,66442,66647,66730,66847,66900,67137,67421,67694,67805,67821,67823,67850,68030,68056,68177,68193,68249,68300,68304,68377,68653,68716,68759,68859,69004,69069,69237,69442,69482,69690,69734,69748,69774,69877,69896,69964,70038,70055,70127,70170,70313,70346,70400,70459,70532,70634,70661,70677,71088,71290,71317,71348,71524,71581,71713,71979,72045,72169,72179,72197,72230,72501,72605,72640,72709,72727,72751,72831,72876,72945,72997,73039,73054,73093,73258,73315,73552,73689,73723,73728,73846,73916,73952,73954,74094,74099,74251,74361,74370,74430,74524,74618,74736,74768,74927,74958,74962,75039,75249,75277,75400,75417,75419,75445,75743,75820,75881,75895,76042,76088,76119,76131,76289,76337,76577,76687,76706,76810,76852,76919,76927,76979,76989,77111,77208,77268,77379,77380,77449,77499,77560,77576,77612,77631,77711,77800,77811,77852,77988,78098,78104,78316,78375,78379,78416,78427,78455,78537,78567,78680,78684,78791,78803,78823,78948,79168,79195,79236,79281,79334,79495,79567,79633,79669,79674,79795,79839,79848,79874,79885,79944,80256,80402,80458,80595,80648,80652,80723,80774,80910,80924,80982,81017,81124,81145,81217,81225,81391,81490,81509,81618,81692,81718,81786,81856,81867,81892,82039,82286,82429,82556,82588,82847,82848,82897,82906,83047,83083,83154,83242,83442,83607,83672,83751,83797,84241,84378,84388,84393,84438,84607,84623,84691,84716,84825,84842,84887,84935,85085,85146,85195,85432,85635,85686,85707,85731,85775,85969,86105,86177,86378,86500,86514,86552,86616,86620,86673,86708,86903,86997,87033,87084,87093,87137,87241,87283,87457,87568,87607,87610,87668,87675,87764,87782,87785,87886,88014,88062,88099,88166,88189,88192,88196,88397,88448,88463,88547,88664,88704,88720,88753,88805,88967,89066,89092,89115,89201,89207,89264,89292,89303,89337,89399,89425,89521,89750,89898,90183,90241,90313,90405,90472,90729,90804,90822,90969,91046,91112,91409,91507,91580,91608,91727,91797,91881,91896,91989,92182,92246,92286,92316,92419,92430,92495,92535,92622,92628,92644,92672,92839,92968,92983,93003,93062,93196,93246,93301,93358,93386,93491,93597,93696,93710,93748,93781,93913,93915,93961,94025,94075,94090,94126,94230,94235,94245,94257,94445,94459,94494,94497,94744,94974,95007,95010,95145,95251,95347,95378,95557,95626,95688,95741,95783,95824,95862,95867,95931,95949,96049,96117,96194,96212,96338,96402,96475,96495,96508,96537,96582,96598,96630,96688,96759,96933,96961,96980,97078,97256,97399,97471,97605,97778,97883,97935,97941,97995,98017,98028,98136,98242,98257,98293,98356,98384,98460,98486,98489,98510,98718,98730,98763,98908,98936,98989,99229,99403,99429,99627,99635,99640,99740,99769,99807,99965",
          "50,52,201,417,530,579,605,703,787,846,883,894,1022,1042,1092,1165,1224,1281,1299,1385,1420,1459,1507,1530,1551,1719,1916,1923,1984,2007,2012,2199,2211,2314,2379,2509,2762,2773,2794,2872,2944,3022,3290,3409,3449,3507,3601,3626,3647,3660,3783,3843,3847,3853,3946,3997,4070,4173,4333,4344,4382,4491,4588,4601,4639,4674,4678,4680,4856,5022,5038,5128,5203,5241,5324,5325,5420,5423,5424,5425,5504,5672,5687,5832,5943,6302,6309,6314,6346,6562,6683,6724,6753,6767,6882,7051,7052,7065,7340,7344,7587,7701,7877,8103,8130,8322,8337,8485,8585,8596,8629,8770,8779,8922,8945,9041,9084,9150,9290,9301,9327,9342,9426,9428,9447,9455,9488,9616,9923,9929,9936,9984,10107,10203,10237,10263,10396,10466,10505,10549,10563,10609,10684,10690,10704,10821,10879,10883,10915,11343,11496,11502,11532,11620,11724,11773,11863,11900,11937,11941,12000,12159,12250,12300,12424,12587,12620,12843,13012,13118,13227,13257,13374,13406,13570,13643,13657,13730,13969,13983,13987,14081,14217,14259,14321,14367,14422,14428,14515,14575,14619,14656,14816,14880,14902,14905,14930,14957,14975,14989,15001,15004,15015,15021,15220,15334,15422,15486,15584,15631,15769,15778,15785,15791,15849,15884,15892,16074,16219,16241,16251,16305,16534,16755,16756,16794,16923,16926,16977,17034,17045,17163,17208,17440,17593,17641,17786,17803,17906,17908,17928,17980,18004,18111,18280,18538,18616,18682,18766,19236,19253,19283,19421,19472,19593,19812,19828,20095,20185,20291,20372,20443,20468,20507,20669,20697,20768,20890,20990,21174,21217,21264,21291,21307,21348,21349,21424,21489,21589,21608,21777,21851,21905,21991,22108,22206,22215,22254,22459,22525,22740,22840,22853,22860,22870,23037,23250,23487,23516,23558,23567,23588,23638,23906,23990,24055,24190,24234,24297,24301,24365,24498,24530,24582,24645,24735,24809,24837,24909,25034,25122,25129,25148,25164,25262,25709,25978,26016,26511,26514,26538,26583,26603,26631,26643,26878,27009,27130,27188,27198,27239,27445,27644,27729,27882,27900,28087,28219,28260,28291,28417,28529,28643,28870,28981,29004,29014,29399,29451,29453,29500,29506,29814,29832,29835,29844,29855,29896,30250,30327,30376,30537,30569,30678,30725,30850,30906,31063,31144,31165,31297,31311,31463,31481,31504,31507,31587,31981,32004,32097,32142,32194,32302,32316,32343,32370,32498,32559,32655,32666,32825,32989,33041,33128,33402,33460,33465,33544,33572,33612,33760,33875,33943,33987,34138,34166,34179,34198,34362,34462,34500,34615,34738,34782,34956,34970,35275,35316,35354,35514,35558,35583,35666,35983,36028,36154,36177,36320,36479,36510,36628,36736,36902,37078,37238,37254,37746,37760,37784,37853,37890,37904,37920,38035,38199,38284,38287,38353,38501,38533,38611,38661,38680,38813,39090,39180,39232,39281,39359,39380,39464,39474,39506,39641,39722,39758,39797,39800,39904,39933,39979,39996,40016,40097,40140,40193,40306,40436,40451,40554,40682,40727,40747,40749,40765,40769,40842,41041,41146,41295,41311,41359,41376,41472,41478,41492,41659,41980,41983,41987,42058,42062,42097,42321,42439,42643,42659,42794,42938,43117,43129,43366,43394,43467,43606,43615,43768,43776,43785,43839,43900,44017,44052,44145,44272,44318,44375,44435,44494,44550,44706,44719,44858,44977,45051,45082,45200,45206,45269,45359,45367,45409,45433,45512,45633,45686,45820,45823,45851,45989,46164,46182,46287,46294,46820,46861,46863,46880,47021,47037,47045,47068,47250,47318,47605,47773,47870,47876,48031,48041,48071,48083,48185,48247,48296,48407,48421,48517,48533,48550,48607,48659,48698,48709,48761,48833,48869,49130,49134,49174,49254,49294,49303,49401,49410,49448,49492,49776,49809,49813,49893,49923,50104,50124,50168,50251,50602,50649,50650,50753,50821,50857,50875,50880,50931,51040,51361,51390,51395,51463,51492,51593,51677,51719,51738,51858,51863,51914,52068,52076,52091,52146,52277,52311,52348,52402,52528,52533,52567,52613,52621,52653,52739,52911,53005,53318,53342,53360,53385,53387,53433,53528,53609,53678,53722,53781,53784,53838,54193,54479,54563,54589,54767,54832,54849,55020,55120,55142,55289,55352,55380,55459,55506,55544,55546,55609,55643,55673,55803,55923,55938,55976,56062,56072,56118,56189,56468,56486,56521,56675,56684,56688,56780,56806,56926,56996,57099,57208,57237,57311,57616,57634,57660,57668,57687,57708,57801,57941,57946,58045,58120,58650,58680,58799,58868,58892,58908,58972,59114,59115,59167,59282,59287,59328,59333,59364,59507,59560,59681,59690,59701,59782,60151,60172,60257,60283,60404,60608,60790,60967,61008,61058,61077,61153,61494,61540,61556,61576,61657,61679,61687,61688,61890,62062,62141,62172,62257,62385,62503,62513,62552,62686,62749,62769,62938,62958,62972,62989,63014,63027,63179,63338,63382,63653,63979,64192,64344,64481,64623,65223,65312,65343,65441,65497,65499,65587,65601,65716,65793,65848,66208,66284,66353,66418,66657,66693,66939,67057,67091,67195,67296,67305,67344,67350,67661,67878,67895,67972,68085,68115,68204,68319,68383,68412,68462,68495,68532,68613,68727,68796,68875,68933,68953,69045,69046,69050,69100,69285,69498,69556,69700,69715,69852,69889,69905,69927,70357,70435,70467,70490,70508,70519,70558,70619,70833,70938,70948,70964,71093,71116,71126,71188,71228,71299,71332,71355,71544,71633,71747,72014,72085,72110,72137,72176,72500,72572,72606,72679,72787,72854,72965,73077,73107,73296,73357,73417,73462,73501,73609,73680,73696,73699,73859,73888,74079,74159,74181,74393,74434,74651,74698,74725,74770,74897,74983,75078,75197,75212,75223,75310,75357,75377,75381,75521,75644,75688,75699,75725,75759,75908,76215,76233,76318,76322,76469,76662,76735,76792,77077,77134,77229,77231,77291,77836,77894,77909,77910,77996,77999,78334,78486,78581,78634,78677,78951,78966,79184,79314,79331,79354,79367,79388,79429,79438,79462,79503,79657,79695,79796,79838,79993,80059,80167,80200,80280,80436,80508,80536,80572,80765,80923,80939,81020,81035,81138,81161,81164,81249,81276,81461,81531,81591,81670,81826,81865,81915,81996,82085,82140,82226,82389,82428,82433,82540,82619,82832,82955,83013,83045,83101,83125,83134,83214,83331,83342,83512,83567,83674,83693,83771,83833,83841,83859,84058,84166,84174,84175,84177,84194,84255,84487,84500,84518,84578,84618,84728,84822,84911,84977,85136,85187,85211,85274,85361,85394,85400,85519,85567,85810,85820,85858,85964,86054,86061,86287,86342,86494,86705,86776,86786,86812,86853,86955,86961,87118,87312,87574,87623,87763,87893,87996,88059,88110,88173,88249,88271,88382,88519,88660,88661,88792,88820,88845,89012,89086,89325,89364,89382,89467,89597,89662,89713,89923,89947,90006,90153,90180,90245,90323,90383,90541,90551,90569,90595,90645,90707,90721,90722,90753,90852,90857,90967,90982,91022,91040,91094,91119,91177,91260,91263,91380,91476,91566,91689,91908,91947,92072,92537,92562,92784,92804,92881,92941,92959,93203,93377,93623,93764,94030,94043,94071,94108,94220,94704,94804,94818,95212,95281,95440,95664,95683,95695,95705,95718,95724,95750,95820,95828,95992,96054,96084,96234,96239,96294,96343,96386,96563,96669,96730,96920,96923,96936,97031,97043,97048,97068,97187,97255,97476,97575,97823,97880,97931,97994,98058,98149,98399,98568,98588,98592,98645,98713,98754,98829,98847,99050,99057,99108,99195,99199,99285,99318,99321,99333,99353,99373,99468,99549,99614,99615,99616,99675,99712,99857,99885,99967,99986",
          "70,82,239,262,561,673,898,1014,1091,1131,1142,1154,1169,1220,1422,1455,1592,1655,1774,1990,2056,2057,2214,2223,2406,2412,2610,2624,2668,2703,2748,2972,3036,3087,3089,3126,3228,3250,3301,3320,3373,3479,3521,3561,3565,3652,3768,3851,3909,3984,4052,4090,4243,4312,4336,4337,4350,4384,4531,4542,4638,4654,4663,4718,4810,4868,4883,5005,5088,5090,5225,5379,5553,5660,5709,5753,5798,5839,5842,5863,6227,6245,6428,6534,6590,6620,6697,6838,6928,7067,7253,7301,7428,7512,7520,7539,7612,7655,7696,7750,7763,7773,7795,8108,8161,8248,8283,8348,8400,8574,8775,8794,8804,8860,8875,8963,8969,9060,9162,9305,9377,9461,9537,9597,9703,9752,9762,9844,9850,9995,10166,10218,10306,10435,10463,10476,10579,10754,10862,10995,11035,11081,11186,11226,11247,11295,11338,11346,11360,11604,11634,11944,12099,12219,12241,12248,12355,12375,12401,12439,12560,12704,12735,13021,13054,13077,13144,13382,13383,13431,13454,13575,13688,13914,13948,13980,14018,14090,14228,14286,14302,14414,14440,14469,14554,14665,14755,14846,14938,15147,15181,15243,15255,15378,15569,15780,15862,15864,15928,16028,16094,16168,16181,16183,16242,16250,16338,16427,16430,16491,16575,16624,16789,16867,16911,16974,17098,17152,17159,17304,17342,17485,17605,17608,17685,17735,17749,17845,17952,18055,18058,18063,18153,18319,18418,18514,18558,18676,19022,19074,19106,19158,19241,19321,19373,19527,19562,19635,19719,19820,19910,20074,20194,20208,20300,20387,20455,20644,20680,20721,20745,20883,20884,20915,20985,21010,21059,21091,21191,21245,21277,21330,21379,21381,21604,21779,21948,22042,22113,22176,22248,22258,22261,22296,22359,22805,22818,23068,23217,23239,23420,23424,23489,23689,23747,23807,23835,23851,23917,23922,23984,24014,24115,24130,24380,24427,24637,24716,24729,24755,24810,24818,24883,25024,25028,25060,25247,25300,25306,25567,25738,25755,26046,26110,26174,26189,26275,26311,26360,26394,26395,26668,26675,26709,26872,26939,27089,27095,27129,27330,27548,27554,27599,27673,27797,27953,27954,28025,28035,28055,28143,28222,28232,28361,28389,28483,28522,28553,28615,28674,28693,28752,28884,28916,28951,28961,28970,29257,29365,29540,29597,29630,29744,29890,29898,29908,30075,30104,30126,30139,30441,30585,30587,30700,30789,30983,30997,31039,31082,31094,31232,31270,31336,31339,31368,31949,31958,31978,32010,32018,32037,32075,32093,32159,32273,32377,32512,32606,32724,32727,32786,32855,33073,33074,33105,33171,33175,33478,33502,33639,33712,33746,33808,33853,33866,34028,34043,34280,34343,34571,34590,34664,34695,34723,34886,34890,34909,35174,35232,35289,35355,35398,35418,35596,35620,36161,36323,36444,36469,36493,36638,36689,36929,37114,37234,37255,37340,37665,37696,37741,37950,38002,38028,38179,38207,38300,38350,38482,38568,38571,38618,38743,38807,38815,38845,38884,38924,38953,38965,38990,39018,39027,39030,39056,39087,39213,39353,39444,39472,39543,39556,39789,39828,39885,40079,40114,40195,40254,40266,40321,40462,40542,40646,40691,40744,40913,41033,41034,41073,41137,41230,41360,41694,41838,42089,42168,42284,42341,42412,42650,42749,42918,43003,43024,43138,43249,43379,43384,43493,43625,43767,43934,44162,44341,44383,44486,44493,44652,44787,44802,44822,44871,44897,44901,45027,45186,45192,45366,45478,45550,45568,45585,45798,45893,45901,45997,46025,46149,46159,46195,46320,46352,46384,46446,46460,46541,46580,46590,46591,46628,46747,46754,46829,46865,46903,46914,46952,46990,47124,47256,47386,47403,47453,47466,47477,47499,47580,47611,47703,47850,48001,48047,48062,48073,48094,48148,48160,48319,48431,48460,48505,48573,48595,48657,48753,48770,48815,49123,49143,49148,49179,49267,49281,49489,49524,49543,49580,49607,49686,49689,49703,49713,49827,49852,49885,49998,50038,50082,50223,50241,50517,50539,50726,50757,51132,51226,51279,51456,51522,51576,51643,51666,51782,51924,51947,52028,52067,52125,52227,52242,52259,52290,52299,52437,52565,52580,52926,53029,53040,53205,53242,53266,53421,53573,53598,53663,53751,53763,53864,53938,54086,54139,54306,54541,54621,54794,54911,54926,55084,55221,55249,55350,55381,55393,55399,55621,55721,55891,56014,56035,56146,56156,56193,56253,56306,56554,56579,56744,56746,56823,56910,56984,57012,57130,57146,57218,57233,57324,57371,57433,57489,57617,57671,57682,57751,57820,57884,58186,58221,58325,58570,58597,58615,58665,58815,58832,58913,59219,59244,59247,59262,59346,59365,59373,59420,59426,59430,59471,59619,59660,59694,59902,59996,60301,60306,60352,60451,60456,60513,60582,60614,60635,60870,61059,61072,61157,61217,61276,61301,61304,61336,61340,61408,61415,61481,61520,61579,61602,61682,61718,61943,61966,62111,62253,62415,62473,62662,62678,62810,62823,62906,62937,62978,63029,63077,63175,63187,63225,63321,63466,63642,63684,63693,63761,64043,64196,64223,64329,64476,64532,64533,64541,64571,64630,64775,64838,64890,65028,65192,65258,65344,65345,65475,65561,65598,65671,65778,65903,65920,65954,65975,66002,66054,66166,66333,66364,66427,66503,66556,66595,66671,66704,67230,67322,67469,67490,67683,67711,67753,67812,67831,67833,67849,67887,67948,68161,68228,68255,68418,68429,68558,68589,68812,68814,68822,68918,68929,68946,68965,69248,69257,69262,69272,69480,69590,69761,69843,69963,69971,70283,70401,70500,70517,70563,70641,70891,70920,71094,71263,71282,71347,71374,71419,71488,71508,71652,71719,71723,71734,71759,71761,71783,71811,71966,72048,72245,72287,72353,72472,72508,72587,72612,72717,72804,72871,72885,72901,72921,72931,72944,72985,73008,73019,73064,73297,73424,73440,73525,73594,73611,73614,73685,73994,74002,74035,74053,74142,74173,74179,74392,74403,74617,74655,74660,74803,75032,75052,75387,75396,75438,75566,75988,76017,76146,76160,76198,76231,76273,76380,76411,76479,76508,76586,76614,76633,76672,76705,76715,76826,76897,76912,76915,77337,77372,77393,77417,77443,77474,77524,77526,77542,77636,77826,77835,77901,77903,77953,77971,78001,78006,78018,78196,78220,78236,78262,78441,78578,78617,78673,78705,78770,78809,78914,78996,79148,79160,79274,79285,79347,79377,79482,79526,79559,79650,79691,79725,79761,79847,79849,79887,79903,79914,79965,79988,80135,80142,80216,80433,80434,80474,80493,80512,80589,80656,80741,80906,80988,81228,81374,81430,81480,81736,81801,81812,81837,81850,82101,82198,82255,82256,82344,82381,82647,82672,82799,82866,82900,82948,82951,83077,83091,83244,83255,83256,83328,83462,83818,84123,84146,84198,84261,84398,84553,84569,84581,84592,84611,84656,84688,84706,84750,84905,84981,85099,85111,85115,85141,85144,85170,85250,85393,85441,85489,85499,85501,85552,85594,85629,85696,85699,85724,85774,85899,85931,85955,85976,86102,86134,86176,86209,86406,86450,86459,86466,86594,86667,86723,86766,86798,87055,87114,87196,87268,87440,87495,87557,87586,87671,87678,87820,87931,87934,88152,88180,88342,88403,88413,88493,88566,88572,88611,88692,88749,88964,88993,89096,89137,89165,89263,89317,89321,89396,89430,89532,89562,89664,89734,89764,89970,90196,90290,90291,90445,90486,90791,90801,90885,90953,91026,91108,91235,91236,91280,91714,91768,91780,91785,91905,91936,92152,92163,92216,92274,92302,92451,92546,92599,92768,92785,92827,92930,92951,93007,93218,93274,93306,93453,93465,93516,93552,93592,93787,93814,93962,94022,94085,94202,94218,94255,94260,94572,94587,94599,94631,94716,94788,94797,94854,94983,95006,95013,95221,95229,95263,95268,95307,95463,95610,95672,95797,95844,95855,95897,96075,96273,96371,96398,96445,96467,96546,96552,96752,96858,96950,97086,97141,97162,97174,97198,97205,97210,97225,97250,97384,97432,97694,97792,97836,97961,98035,98093,98205,98237,98415,98483,98487,98543,98585,98733,98740,98855,98881,98916,98988,99134,99174,99190,99552,99698,99783,99890,99989",
          "22,178,180,534,607,662,812,816,858,926,1047,1051,1062,1087,1242,1247,1347,1353,1499,1511,1527,1538,1598,1640,1674,1679,1773,1831,1886,1901,1921,1982,1986,2126,2148,2224,2320,2448,2518,2535,2539,2657,2694,2705,2723,2803,2853,2860,2956,3024,3055,3057,3065,3067,3181,3188,3237,3240,3253,3379,3518,3706,3737,3781,3897,3928,4038,4156,4218,4314,4372,4420,4562,4684,4942,4970,4979,5054,5061,5062,5072,5094,5624,5675,5728,5898,6045,6065,6169,6215,6272,6324,6380,6537,6621,6626,6634,6674,6687,6720,6737,6822,6899,6959,6976,7069,7208,7223,7225,7409,7544,7594,7730,7792,7799,7985,8017,8058,8093,8094,8124,8134,8329,8336,8404,8505,8548,8555,8580,8593,8677,8793,8988,9058,9099,9196,9197,9205,9210,9463,9473,9527,9544,9601,9764,9886,9912,9918,10062,10104,10111,10195,10231,10328,10420,10495,10573,10607,10652,10699,10729,10797,10815,10854,10990,10997,11157,11237,11522,11745,11751,11763,11908,11934,12045,12050,12176,12325,12331,12516,12529,12565,12664,12714,12721,12727,12757,12827,12862,12939,12972,13143,13146,13178,13309,13313,13414,13506,13526,13554,13584,13709,13724,13731,13789,13862,13883,13893,13989,14035,14092,14094,14179,14260,14413,14448,14674,14774,14787,14798,14827,14878,14911,14971,14981,15006,15019,15098,15112,15161,15188,15225,15229,15310,15391,15417,15483,15520,15687,15708,15713,15739,15868,15889,15895,15940,15974,15994,16016,16206,16288,16310,16331,16482,16532,16547,16576,16611,16828,16869,16910,16982,17069,17072,17137,17359,17412,17439,17596,17662,17797,17911,17915,18024,18067,18083,18086,18102,18140,18259,18261,18299,18343,18532,18541,18636,18778,18829,18847,18884,18887,18992,18995,19058,19067,19116,19145,19153,19192,19280,19366,19432,19529,19570,19585,19672,19815,20073,20086,20280,20286,20341,20395,20397,20504,20578,20901,21283,21360,21470,21474,21567,21632,21657,21853,22058,22068,22078,22085,22087,22471,22553,22576,22588,22874,22947,22981,22982,23041,23101,23105,23122,23150,23198,23201,23235,23241,23243,23326,23341,23363,23500,23522,23570,23584,23614,23616,23618,23685,23812,23879,23934,24064,24260,24386,24397,24414,24423,24477,24584,24623,24697,24798,24839,24928,25047,25118,25143,25196,25243,25246,25279,25302,25376,25426,25464,25536,25606,25828,25964,26029,26080,26148,26161,26200,26241,26252,26500,26550,26658,26661,26736,26829,26946,26951,26979,27085,27135,27271,27478,27528,27621,27682,27704,27752,27785,27855,27926,27971,27972,27975,27992,28015,28056,28252,28330,28407,28501,28524,28531,28552,28570,28618,28636,28785,28846,29224,29308,29330,29346,29364,29416,29427,29601,29641,29742,29817,29820,29837,29838,29843,29913,30072,30080,30083,30085,30121,30178,30354,30559,30705,30867,30892,31032,31050,31092,31328,31439,31540,31567,31610,31716,31870,31910,31975,32024,32025,32313,32323,32402,32483,32525,32571,32594,32750,32781,32812,32836,32838,32929,32942,32948,32950,33102,33169,33284,33339,33386,33426,33467,33569,33601,33619,33661,33838,33852,33860,33982,34019,34034,34096,34252,34356,34438,34465,34495,34531,34572,34588,34628,34704,34919,34921,35034,35094,35139,35177,35279,35284,35313,35402,35502,35561,35569,35579,35647,35658,35719,35784,35808,35867,35869,36000,36264,36392,36466,36588,36680,36747,36769,36894,36917,37135,37164,37200,37232,37441,37457,37525,37621,37643,37731,37803,37833,38017,38117,38140,38147,38177,38212,38224,38251,38440,38485,38541,38609,38625,38651,38673,38699,38720,38726,38852,38874,38959,39026,39032,39097,39294,39369,39476,39544,40483,40794,40825,40869,40932,40985,41092,41222,41275,41341,41384,41546,41655,41780,41784,41878,41900,42001,42109,42116,42219,42264,42273,42484,42526,42528,42542,42636,42695,42705,42737,42865,42889,43000,43058,43130,43291,43292,43336,43412,43441,43461,43682,43732,43858,43888,43919,43992,44019,44034,44074,44103,44104,44166,44184,44264,44327,44337,44487,44568,44623,44668,44669,44695,44955,45080,45094,45113,45156,45275,45276,45326,45572,45636,45752,45826,45885,45950,46089,46170,46273,46403,46424,46484,46513,46569,46615,46809,46962,47001,47003,47105,47115,47323,47351,47413,47635,47709,47728,47950,48233,48281,48283,48362,48372,48699,48825,49013,49369,49417,49443,49505,49614,49632,49679,49734,49855,49938,49996,50054,50125,50126,50174,50304,50326,50328,50352,50374,50436,50453,50465,50642,50724,50831,51028,51351,51415,51501,51549,51733,51763,51785,51875,51891,51981,52085,52100,52190,52271,52307,52310,52524,52588,52589,52594,52657,52719,52738,52780,52847,52976,53033,53048,53167,53667,53846,53906,53956,54067,54280,54317,54387,54402,54554,54556,54578,54634,54673,54803,54873,54938,54972,55012,55107,55153,55157,55166,55238,55348,55369,55547,55579,55718,55754,55762,55769,55811,55992,56110,56124,56154,56167,56187,56206,56284,56309,56335,56440,56451,56635,56643,56649,56691,56769,56810,56811,56887,56947,56991,57049,57072,57129,57195,57206,57216,57841,58046,58056,58109,58200,58243,58324,58425,58519,58533,58542,58587,58611,58816,58859,58875,59006,59020,59073,59086,59125,59131,59265,59517,59575,59639,59828,59832,60007,60036,60075,60128,60372,60509,60524,60562,60655,60826,60887,60953,61230,61369,61443,61460,61707,61724,61798,61804,61929,61997,61999,62120,62265,62339,62370,62373,62387,62438,62447,62452,62511,62571,62716,62747,62761,62804,62933,63124,63182,63184,63191,63202,63282,63312,63454,63479,63526,63596,63660,63736,63790,63851,63883,63911,63912,63951,64041,64102,64221,64276,64349,64378,64399,64403,64496,64522,64565,64633,64729,64743,64747,64791,64804,64855,64933,65039,65120,65133,65228,65257,65301,65307,65310,65326,65379,65385,65451,65509,65611,65613,65617,65669,65700,65786,65801,65944,65995,66062,66128,66165,66244,66267,66421,66426,66481,66553,66685,66734,66758,66802,66972,67106,67196,67235,67480,67547,67634,67654,67718,67829,67838,68129,68195,68289,68331,68404,68548,68802,68838,68877,69090,69122,69176,69188,69192,69326,69333,69458,69548,69549,69604,69657,69906,69996,70029,70076,70107,70145,70204,70221,70289,70384,70393,70461,70550,70628,70873,70930,70947,70989,70993,71005,71192,71256,71321,71438,71495,71563,71666,71676,72002,72247,72255,72535,72604,72654,72657,72688,72803,72918,72919,72949,73049,73114,73123,73174,73268,73305,73339,73344,73436,73528,73604,73608,73655,73666,73717,73911,73924,73957,73958,74011,74131,74171,74249,74307,74336,74542,74549,74637,74672,74728,74783,74907,74914,74998,75080,75125,75130,75139,75215,75254,75287,75448,75524,75530,75537,75542,75554,75663,75779,75783,75788,75813,75815,75872,75914,75975,76065,76686,76757,76920,76978,77048,77145,77220,77259,77352,77400,77586,77808,77823,77845,77985,78030,78075,78125,78216,78266,78534,78536,78539,78583,78654,78752,78789,78840,78962,78964,78993,79049,79082,79169,79220,79238,79282,79327,79405,79522,79586,79621,79751,79878,79912,80018,80195,80299,80379,80534,80636,80834,81088,81112,81198,81240,81309,81365,81414,81441,81519,81600,81619,81644,81785,81789,81883,81884,81919,81964,81990,82127,82218,82223,82267,82403,82415,82462,82507,82549,82628,82808,82959,82970,83115,83130,83156,83158,83224,83300,83588,83694,83798,83904,83917,83941,83972,84060,84200,84304,84334,84579,84634,84697,84817,84828,84883,84996,85131,85200,85207,85215,85233,85307,85331,85333,85448,85534,85716,85941,85947,86000,86065,86123,86253,86300,86387,86670,86783,86821,86861,86872,86873,86875,86879,86886,86915,86935,86944,87015,87040,87070,87314,87322,87395,87416,87544,87657,87658,87660,87702,87711,87722,87736,87761,87771,87863,87871,87894,88021,88085,88089,88193,88274,88615,88619,88693,88760,88794,88826,88839,88878,89039,89059,89091,89123,89239,89242,89344,89356,89469,89523,89599,89612,89700,89701,89721,89863,90035,90106,90115,90156,90159,90198,90207,90253,90303,90321,90367,90599,90603,90672,90685,90788,90813,90831,90849,90891,91074,91200,91259,91339,91442,91555,91870,91956,92127,92180,92279,92297,92516,92553,92615,92620,92632,92791,92849,92885,93123,93157,93163,93204,93213,93394,93395,93468,93614,93715,93834,93874,93885,93904,93937,94019,94096,94231,94495,94553,94562,94566,94693,94732,94810,94844,94895,94946,94972,95014,95149,95179,95205,95311,95467,95499,95542,95558,96016,96057,96074,96083,96335,96450,96451,96655,96677,96827,96860,96946,97032,97054,97083,97122,97166,97242,97295,97451,97454,97557,97837,97877,97962,98134,98155,98162,98223,98306,98344,98428,98450,98482,98627,98652,98707,98731,98744,98777,98891,98939,98961,99024,99128,99167,99397,99482,99502,99669,99791,99859,99927,99987,99997",
          "61,77,90,96,117,141,200,245,258,308,358,393,489,490,567,626,753,776,820,961,1025,1149,1155,1312,1572,1613,1705,1999,2018,2342,2345,2348,2519,2550,2651,2744,2811,2882,2969,3112,3115,3122,3134,3194,3236,3289,3421,3483,3500,4025,4060,4130,4176,4225,4378,4416,4451,4516,4644,4707,4760,4774,4779,4976,5097,5149,5249,5439,5478,5518,5669,5702,5759,5760,6142,6192,6212,6231,6283,6287,6337,6474,6610,6757,6871,6885,6913,7040,7050,7124,7249,7290,7392,7432,7434,7441,7473,7549,7606,7736,7874,8135,8264,8299,8360,8667,8742,8766,8791,8810,9203,9234,9418,9459,9490,9502,9554,9630,9728,10065,10105,10123,10125,10161,10228,10353,10470,10513,10533,10667,10764,10872,10959,10998,11048,11064,11066,11116,11144,11207,11506,11509,11519,11548,11612,11713,11800,12018,12062,12127,12142,12482,12550,12632,12651,12705,12794,12902,12920,13097,13115,13132,13137,13299,13318,13384,13501,13521,13573,13593,13740,14119,14193,14221,14341,14483,14521,14572,14657,14691,14730,14877,14884,14886,14934,15104,15218,15263,15304,15318,15476,15484,15675,15726,15730,15803,15813,16008,16025,16137,16281,16445,16488,16582,16595,16596,16613,16654,16768,16933,17049,17058,17164,17173,17374,17581,17612,17624,17772,17778,17983,18015,18048,18144,18171,18346,18607,18741,18788,19011,19030,19049,19118,19194,19199,19227,19339,19384,19385,19409,19479,19583,19934,19971,20000,20023,20204,20236,20266,20602,20655,20676,20679,20831,20966,21014,21062,21119,21202,21251,21294,21744,21757,21789,21916,21946,22001,22132,22200,22298,22330,22420,22638,22671,22779,22861,22877,22942,23024,23090,23095,23194,23387,23389,23608,23630,23771,23775,23808,23809,23816,23927,23939,23950,24105,24302,24641,24670,24726,24866,24919,25001,25041,25180,25508,25515,25632,25634,25636,25712,25911,25925,25943,25946,26209,26278,26448,26704,26721,26810,26858,26921,27008,27132,27161,27321,27409,27437,27477,27586,27658,27659,27771,27778,27969,28171,28224,28272,28379,28403,28593,28748,28799,28971,29060,29220,29294,29353,29463,29567,29700,29790,29888,29943,30113,30164,30307,30344,30506,30602,30686,30901,30922,31009,31049,31069,31079,31119,31227,31253,31324,31395,31511,31664,31676,32175,32180,32390,32561,32694,32742,32779,32815,32818,32976,33015,33148,33332,33664,33674,34049,34077,34086,34127,34221,34472,34490,34621,34636,34736,34836,34841,34910,34953,35007,35082,35131,35201,35459,35491,35590,35598,35660,35662,35677,35715,35814,35889,35916,35942,36041,36088,36107,36288,36350,36447,36472,36880,37069,37187,37379,37390,37618,37692,37719,37730,37912,37975,38079,38234,38246,38292,38299,38330,38427,38588,38704,38708,38722,38737,38802,38844,38885,38923,38925,38987,39068,39130,39341,39405,39475,39492,39594,39621,39742,39870,39945,40032,40180,40313,40341,40401,40410,40527,40578,40618,40710,40719,40767,40798,40875,41102,41188,41435,41525,41589,41618,41625,41743,41749,41755,41770,41798,41829,41890,41913,41914,42049,42142,42294,42440,42572,42592,42668,42858,42870,42967,42968,43295,43307,43404,43481,43485,43538,43738,43853,43896,43980,44041,44156,44241,44255,44310,44476,44495,44554,44602,44738,44741,44769,44841,44844,44939,44942,44993,45016,45025,45040,45128,45180,45181,45193,45263,45288,45306,45468,45563,45672,45733,45748,45772,45892,45923,46621,46634,46752,46762,46785,46844,46997,47038,47108,47122,47123,47162,47167,47195,47219,47254,47277,47295,47349,47382,47389,47404,47531,47594,47674,47824,47841,47970,48020,48095,48270,48313,48342,48454,48459,48490,48544,48632,48676,48707,48834,48836,48914,48970,49107,49175,49275,49418,49480,49654,49842,49851,49877,49905,49957,49966,50064,50071,50073,50152,50280,50298,50390,50392,50507,50526,50608,50625,50916,50928,50967,51045,51065,51075,51098,51146,51219,51346,51387,51410,51454,51504,51526,51630,51644,51668,51702,51714,51862,51900,51929,52003,52014,52248,52450,52539,52541,52640,52646,52698,52761,52763,52878,52958,52965,53165,53166,53347,53764,53860,54160,54202,54223,54338,54446,54448,54462,54529,54564,54702,54704,55141,55171,55235,55518,55707,55736,55765,55791,55879,55964,55973,56332,56401,56471,56506,56577,56706,56801,56990,56992,57108,57261,57307,57350,57406,57416,57435,57445,57505,57550,57569,57752,57776,57868,57901,57956,58033,58079,58124,58289,58298,58627,58831,58838,58928,59127,59233,59290,59356,59424,59488,59502,59515,59584,59643,59657,59678,59746,59864,59923,59937,60119,60273,60290,60449,60519,60776,60811,60854,60959,60966,61009,61011,61023,61229,61243,61471,61569,61598,61623,61638,61742,61870,61930,61977,62091,62237,62368,62436,62446,62590,62598,63032,63036,63152,63293,63432,63527,63671,63722,63765,64047,64091,64284,64306,64351,64434,64464,64468,64546,64715,64926,64953,65119,65145,65178,65211,65378,65396,65519,65545,65687,65714,65798,65813,65823,65876,65963,65976,65988,66101,66324,66415,66574,66581,66700,66712,66825,66893,66952,67131,67172,67303,67363,67453,67674,67681,67834,67993,68395,68502,68545,68555,68557,68576,68674,68684,68975,69163,69265,69297,69409,69566,69628,69849,70153,70162,70233,70291,70388,70431,70505,70524,70616,70626,70669,70671,70817,70855,70895,70908,71029,71039,71102,71245,71314,71689,71691,71705,71733,71798,71844,71980,71993,72051,72102,72383,72394,72434,72789,72947,73063,73111,73125,73146,73284,73381,73432,73848,73984,73987,74006,74048,74107,74114,74329,74341,74376,74426,74497,74500,74608,74804,74879,74885,74888,74938,75000,75031,75062,75121,75134,75162,75237,75242,75252,75302,75359,75449,75465,75548,75556,75619,75709,75710,75754,75797,75845,75893,75917,76027,76044,76069,76211,76493,76609,76742,76752,76759,76761,76779,76874,77024,77079,77362,77388,77600,77615,77723,77841,77849,77927,77983,78033,78050,78217,78222,78312,78409,78420,78471,78477,78696,78745,78842,79023,79083,79122,79406,79417,79499,79599,79722,79775,79851,79865,79929,79998,80220,80238,80361,80459,80496,80645,80726,80830,80868,81049,81060,81125,81179,81343,81371,81396,81407,81775,81787,81808,81871,82196,82221,82341,82355,82398,82562,82690,82732,82963,82973,83082,83119,83382,83400,83531,83584,83945,84002,84037,84072,84093,84193,84356,84421,84537,84544,84636,84768,84959,84993,85020,85147,85155,85235,85272,85277,85300,85389,85403,85544,85580,85605,85651,85700,85704,85848,85864,85917,85995,86116,86142,86374,86753,86857,86871,86896,86939,86942,86968,87063,87136,87214,87295,87446,87517,87639,87797,87883,87957,87974,87980,87983,87997,88051,88289,88292,88308,88415,88612,88748,88833,88895,88996,89054,89132,89204,89223,89285,89475,89509,89592,89709,90040,90108,90178,90179,90188,90200,90346,90450,90499,90588,90740,90779,90790,90840,90895,91066,91168,91219,91349,91496,91600,91632,91676,91694,91737,91867,91950,92049,92096,92207,92250,92255,92292,92488,92904,92985,93173,93224,93352,93456,93624,93639,93820,93955,94110,94297,94305,94318,94413,94520,94609,94626,94726,94912,95148,95156,95163,95188,95241,95292,95534,95706,95768,95876,95920,95995,96042,96132,96547,96555,96651,96690,97015,97057,97113,97278,97428,97431,97473,97592,97751,97770,97797,97953,98122,98157,98209,98231,98255,98279,98313,98322,98411,98468,98474,98511,98692,98809,98863,98933,99000,99017,99109,99417,99450,99660,99680,99748,99786,99824,99841,99871",
          "31,215,216,254,420,482,576,670,749,828,850,853,1101,1132,1138,1164,1172,1258,1269,1276,1380,1419,1489,1490,1559,1565,1591,1639,1688,1699,1716,1723,1742,1761,1784,1841,1979,2043,2125,2149,2301,2311,2350,2459,2530,2562,2618,2639,2835,2867,2903,2928,2978,3000,3071,3110,3230,3244,3282,3332,3410,3670,3727,3730,3742,3974,3995,4011,4047,4092,4318,4585,4593,4787,4867,4939,4944,4958,5043,5082,5259,5342,5369,5371,5460,5630,5636,5693,5747,5954,6178,6355,6503,6645,6709,6971,7091,7095,7107,7389,7700,7823,7827,7835,7936,8014,8032,8042,8077,8133,8224,8270,8304,8395,8523,8540,8617,8646,8671,8774,9211,9219,9339,9401,9498,9513,9680,9767,9771,9775,9810,9835,9906,9989,10001,10053,10066,10067,10250,10267,10388,10498,10648,10653,10700,10730,10794,10887,11008,11182,11188,11288,11341,11427,11466,11544,11578,11624,11663,11671,11732,11741,11809,11965,12087,12103,12171,12192,12243,12332,12344,12377,12418,12423,12484,12640,12642,12654,12733,12903,12904,12915,12923,12943,12960,12987,13140,13142,13189,13262,13352,13507,13533,13625,13880,13944,13973,13986,14043,14065,14135,14224,14325,14326,14571,14577,14587,14668,14766,15250,15294,15316,15455,15516,15613,15640,15667,15733,15819,15925,15980,16076,16197,16201,16256,16283,16303,16365,16409,16503,16507,16513,16619,16626,16691,16712,16866,16959,17113,17125,17170,17188,17197,17262,17282,17310,17312,17400,17444,17490,17549,17606,17656,17698,17714,17748,17818,17881,18011,18088,18164,18216,18284,18301,18324,18373,18376,18515,18552,18596,18631,18678,18789,18848,18928,18958,18967,19017,19038,19076,19353,19431,19471,19508,19515,19524,19548,19617,19709,19716,19724,19802,19808,19814,19873,19878,19892,20035,20128,20129,20179,20182,20307,20337,20360,20471,20532,20647,20674,20724,20752,20790,20867,20882,20898,20958,20986,21019,21128,21146,21196,21269,21270,21279,21290,21333,21389,21575,21699,21755,21829,21956,21983,22041,22054,22183,22238,22358,22454,22460,22556,22569,22722,22799,22913,22973,23069,23098,23305,23366,23414,23504,23619,23641,23736,23756,23852,23919,23998,24031,24063,24104,24257,24265,24277,24354,24390,24402,24514,24658,24724,24753,24801,24819,24825,24833,24842,25026,25087,25214,25252,25271,25401,25409,25414,25568,25591,25592,25682,25700,25751,25786,25831,25854,26070,26244,26326,26376,26501,26545,26569,26774,26812,26817,26861,26903,26918,26990,27011,27022,27110,27150,27216,27234,27246,27259,27292,27400,27416,27496,27522,27560,27856,27883,27968,27980,28032,28046,28050,28068,28221,28261,28351,28429,28551,28634,28635,28736,28787,28872,28988,29035,29066,29169,29182,29198,29237,29271,29450,29454,29619,29828,29859,29866,30028,30209,30233,30388,30399,30608,30697,30699,30763,30766,30824,30930,31027,31071,31238,31337,31438,31493,31575,31695,31709,31816,31877,31881,31905,31923,32535,32755,32764,32856,33063,33126,33255,33349,33375,33524,33536,33602,33643,33702,33817,34011,34037,34045,34227,34229,34408,34544,34637,34670,34712,34793,34820,34944,34973,35025,35054,35060,35134,35186,35293,35306,35372,35397,35573,35618,35848,35957,35960,36111,36142,36268,36276,36327,36408,36449,36524,36583,36604,36635,36713,36808,36809,37196,37323,37466,37485,37536,37831,37859,37870,37962,37979,38013,38101,38110,38157,38169,38204,38253,38502,38584,38633,38666,38668,38777,38860,38929,38951,39070,39141,39229,39379,39423,39435,39478,39479,39494,39643,39806,39922,40112,40241,40281,40338,40475,40505,40540,40546,40565,40599,40828,40892,40967,40970,40979,41011,41027,41173,41207,41293,41375,41377,41464,41558,41631,41682,41745,41854,41935,41966,41998,42098,42322,42334,42396,42465,42475,42538,42549,42550,42560,42568,42684,42742,42750,42757,42849,42911,42925,43010,43143,43189,43650,43672,43710,43830,43901,44008,44050,44053,44105,44185,44359,44517,44648,44664,44682,44702,44933,44961,45203,45255,45279,45309,45396,45446,45451,45469,45502,45657,45675,45690,45832,45910,45931,46022,46115,46136,46172,46179,46246,46296,46404,46474,46502,46512,46533,46583,46627,46856,46887,46893,46901,47052,47082,47249,47260,47356,47456,47666,47851,48165,48289,48388,48397,48411,48446,48669,48746,48906,48948,49019,49073,49126,49154,49258,49339,49423,49517,49597,49601,49610,49640,49678,49738,49744,49808,50012,50063,50189,50264,50316,50397,50495,50553,50563,50626,50663,50674,50705,50786,50796,50906,50922,50970,50991,51000,51023,51041,51069,51254,51452,51485,51536,51560,51590,51614,51708,51965,51997,52011,52017,52143,52322,52343,52388,52421,52612,52673,52694,52881,52978,52980,53041,53206,53258,53265,53463,53540,53612,53643,53703,53780,53800,53829,53908,53909,54212,54292,54410,54452,54481,54546,54550,54561,54584,54740,54744,54748,54888,54952,55010,55031,55047,55101,55345,55357,55362,55682,55723,55832,55889,56017,56052,56080,56147,56177,56405,56482,56552,56584,56609,56788,56868,57024,57182,57210,57227,57322,57370,57387,57408,57440,57513,57554,57583,57684,57877,57942,57947,58001,58074,58116,58198,58248,58255,58302,58306,58369,58471,58518,58554,58801,58869,58934,59102,59193,59204,59212,59300,59384,59397,59578,59613,59635,59667,59674,59775,59880,59883,59915,59966,59987,60096,60145,60168,60183,60234,60285,60402,60520,60538,60551,60587,60682,60725,60914,60993,61024,61049,61063,61184,61266,61272,61413,61449,61467,61590,61735,61747,61765,61830,61839,61916,62217,62218,62381,62561,62601,62610,62638,62682,62831,62877,62939,63022,63046,63148,63355,63489,63604,63692,63769,63791,63862,63871,63880,64079,64132,64169,64236,64237,64243,64309,64323,64431,64448,64453,64499,64649,64749,64818,64822,64825,64887,64931,64957,64998,65000,65092,65118,65193,65249,65324,65328,65429,65511,65543,65665,65705,65713,65722,65776,65788,65796,65837,65967,66098,66130,66138,66231,66246,66261,66319,66323,66459,66477,66491,66493,66575,66587,66661,66674,66688,66733,66780,66889,67071,67157,67246,67257,67313,67417,67425,67479,67553,67763,67776,67873,68001,68077,68159,68191,68201,68231,68251,68285,68307,68388,68416,68481,68798,68807,69088,69151,69175,69197,69216,69292,69307,69366,69415,69459,69492,69569,69593,69601,69745,69795,69867,70004,70042,70070,70075,70318,70710,70724,70742,70830,70853,70875,70966,71065,71083,71244,71354,71378,71410,71415,71420,71432,71470,71498,71630,71753,72144,72166,72305,72456,72488,72722,72750,72853,72869,72878,72913,72982,72998,73090,73095,73116,73303,73464,73554,73593,73617,73644,73713,73753,73761,73803,73845,74045,74082,74122,74186,74201,74267,74516,74789,74798,75041,75049,75071,75091,75093,75172,75334,75342,75410,75436,75579,75685,75799,75843,75919,75955,76051,76140,76168,76329,76360,76455,76474,76572,76639,76698,76725,76737,76740,76795,76846,76856,76956,76988,77109,77178,77180,77276,77286,77288,77342,77367,77419,77452,77508,77535,77751,77959,78180,78224,78283,78350,78512,78636,78685,78712,78754,78806,78821,78958,79050,79085,79127,79207,79223,79320,79557,79580,79637,79699,79924,80097,80098,80114,80155,80197,80227,80303,80447,80484,80531,80583,80600,80669,80716,80728,80756,80803,80851,80999,81027,81229,81298,81356,81377,81573,81597,81686,81699,81824,81863,81977,82029,82083,82202,82404,82454,82512,82518,82563,82578,82713,82726,82798,82835,82934,83006,83015,83044,83093,83127,83181,83282,83283,83304,83384,83547,83707,83711,83744,83853,83976,83978,84021,84024,84039,84161,84300,84352,84538,84627,84647,84652,84702,84760,84902,84917,84982,85007,85108,85230,85373,85378,85801,85816,85859,85906,85912,85913,85930,85940,86174,86208,86315,86423,86474,86634,86671,86713,86781,86822,86912,86927,86969,87154,87245,87319,87351,87419,87535,87990,88027,88240,88288,88305,88480,88535,88666,88788,88799,88873,88921,88924,88962,89033,89043,89070,89082,89183,89305,89571,89687,89726,89746,89759,89794,89922,89934,89953,89967,90301,90432,90517,90665,90676,90748,90847,90865,90908,90914,90927,90977,91275,91451,91544,91558,91762,91809,91884,91938,91948,91987,92035,92159,92333,92421,92470,92504,92512,92522,92538,92555,92751,92752,92792,92820,92836,92899,92915,92936,93097,93197,93388,93525,93643,93657,93670,93671,93836,93843,93844,93869,93925,93957,94024,94050,94081,94180,94247,94276,94306,94440,94450,94458,94539,94579,94622,94648,94684,94832,94834,94839,94930,95003,95028,95180,95185,95201,95236,95261,95269,95334,95396,95403,95435,95504,95599,95634,95690,95704,95771,95817,95985,96079,96091,96253,96347,96410,96433,96570,96616,96734,96742,96800,97014,97021,97041,97088,97262,97293,97318,97439,97896,97971,98064,98459,98471,98484,98622,98631,98636,98685,98712,98732,98766,98780,98850,98905,99030,99371,99396,99437,99477,99593,99612,99649,99802,99838,99864,99903,99964",
          "16,63,65,81,475,660,763,891,1004,1006,1032,1084,1102,1130,1203,1270,1322,1366,1414,1442,1558,1828,2008,2070,2075,2133,2221,2349,2351,2443,2584,2859,2895,2987,3014,3061,3090,3206,3213,3226,3356,3364,3558,3634,3801,3869,3873,3954,4246,4277,4386,4425,4486,4630,4889,4915,4950,5121,5234,5354,5366,5407,5473,5522,5548,5601,5677,5768,5848,5910,5965,5978,6109,6123,6137,6210,6307,6420,6455,6456,6462,6505,6521,6586,6629,6654,6759,6766,7008,7020,7111,7233,7238,7316,7322,7426,7488,7543,7738,7756,7801,7880,7930,8061,8117,8218,8318,8352,8421,8510,8581,8614,8658,8707,8756,8940,9050,9202,9212,9225,9304,9356,9529,9941,9988,10192,10236,10345,10833,10919,11314,11344,11516,11657,11681,11787,11832,11969,12076,12164,12202,12262,12272,12274,12450,12531,12636,12689,12692,12768,12830,12854,12890,12971,13030,13037,13122,13167,13296,13312,13525,13539,13607,13637,13666,13799,13935,14075,14102,14106,14268,14487,14498,14504,14535,14566,14996,15063,15073,15167,15221,15368,15598,15616,15692,15744,15975,16034,16093,16264,16410,16553,16563,16766,16865,16868,16925,17283,17355,17430,17519,17677,17705,17729,17855,18043,18064,18179,18387,18444,18511,18581,18832,18889,18937,18941,18989,19071,19165,19273,19282,19363,19513,19749,19750,19784,19786,19805,20021,20407,20565,20620,20698,20706,20904,21013,21090,21170,21197,21205,21305,21557,21629,21648,21700,21714,21825,21862,21939,21985,22006,22026,22310,22409,22551,22611,22612,22700,22776,22790,22878,23004,23047,23085,23173,23216,23299,23377,23392,23437,23466,23472,23725,23865,23886,23937,23994,24012,24129,24200,24413,24559,24565,25017,25140,25175,25231,25234,25338,25393,25442,25446,25576,25638,25683,25701,25770,25942,25995,26027,26052,26307,26377,26535,26555,26686,26713,26754,26768,26867,26895,26904,27102,27181,27213,27426,27431,27574,27652,27809,27988,28010,28094,28107,28138,28210,28215,28240,28452,28534,28640,28679,28706,28707,28715,28741,28876,28958,29040,29063,29188,29193,29254,29411,29422,29494,29514,29527,29537,29548,29647,29737,29883,29944,29982,30165,30306,30356,30491,30524,30594,30695,30703,30948,31019,31034,31035,31125,31256,31443,31474,31667,31756,31775,32215,32260,32276,32630,32687,32723,32762,32842,32887,32891,33129,33142,33188,33239,33241,33342,33366,33538,33694,33706,33789,33813,33968,33999,34093,34190,34415,34659,34685,34877,34923,34938,35109,35142,35273,35370,35565,35567,35668,35803,36050,36075,36100,36109,36129,36254,36476,36537,36572,36639,36654,36726,36867,36922,36943,36957,37074,37155,37256,37312,37358,37371,37543,37544,37551,37591,37774,37780,37966,38024,38099,38150,38208,38248,38349,38384,38604,38662,38669,39089,39408,39429,39433,39482,39549,39578,39628,39849,39875,39901,39930,40249,40311,40668,40702,40726,40761,40884,40947,41042,41081,41087,41115,41118,41181,41246,41248,41254,41707,41872,41902,41918,42133,42148,42201,42288,42295,42364,42410,42508,42533,42698,42776,43098,43105,43167,43266,43284,43334,43371,43437,43476,43559,43701,43867,43915,43981,43983,44214,44226,44460,44621,44636,44660,44749,44827,45021,45076,45228,45445,45770,45871,46043,46083,46206,46228,46230,46257,46279,46298,46303,46332,46369,46376,46442,46523,46566,46604,46673,46753,46825,46929,46948,46981,47112,47176,47187,47201,47211,47375,47508,47555,47556,47562,47603,47657,47692,47783,47795,47812,47933,48013,48016,48035,48040,48057,48122,48386,48450,48760,48857,48881,49071,49184,49308,49419,49431,49475,49483,49490,49784,49846,49866,49883,49892,50031,50133,50200,50450,50508,50624,50792,50809,50816,51066,51088,51133,51187,51264,51273,51334,51336,51348,51360,51856,51984,52189,52229,52272,52469,52564,52683,52711,52800,52962,53038,53046,53175,53301,53337,53391,53425,53511,53639,53640,53671,53691,53728,53803,53889,53944,53998,54052,54116,54253,54340,54358,54374,54492,54504,54562,54643,54654,54695,54846,54874,54910,54973,55045,55328,55364,55397,55521,55524,55586,55798,55937,56178,56254,56270,56327,56527,56613,56726,56747,56755,56799,56927,56979,57043,57120,57229,57241,57281,57319,57510,57527,57621,57988,58343,58372,58389,58549,58553,58603,58865,58954,58971,59109,59156,59368,59382,59801,60016,60368,60390,60603,60645,60813,60829,60850,60897,60937,60958,61159,61162,61294,61307,61318,61387,61433,61610,61769,61797,61967,62126,62187,62274,62287,62308,62318,62326,62405,62435,62470,62524,62536,62577,62636,62687,62696,62720,62806,62807,62866,62943,63084,63307,63315,63437,63729,63809,63844,63852,64078,64172,64235,64280,64291,64478,64479,64608,64616,64654,64694,64745,64797,64986,64995,65056,65063,65088,65134,65544,65667,65668,65701,65747,65758,65914,66109,66126,66335,66454,66540,66615,66763,66935,67010,67149,67171,67187,67236,67306,67536,67552,67563,67651,67730,67742,67867,67913,67920,67926,67952,68045,68187,68215,68246,68424,68609,68626,68833,68870,69027,69094,69116,69235,69319,69437,69439,69447,69508,69512,69568,69778,70048,70281,70465,70611,70632,70667,70782,71044,71123,71129,71223,71421,71463,71534,71740,71757,71988,72136,72217,72321,72371,72469,72509,72549,72588,72718,72725,72772,72930,72993,73055,73170,73201,73373,73631,73636,73881,74097,74184,74227,74319,74345,74365,74461,74525,74682,74691,75115,75275,75319,75389,75444,75558,75588,75661,75693,75767,75972,76012,76105,76109,76293,76306,76543,76620,76680,76681,76702,76858,76929,77106,77265,77300,77420,77444,77450,77765,77783,78181,78187,78208,78287,78529,78554,78576,78615,78663,78716,78802,78850,78873,78926,79058,79328,79420,79496,79632,79735,79783,79866,79960,79972,80015,80025,80038,80091,80099,80170,80171,80239,80312,80399,80487,80500,80568,80627,80678,80720,80808,80833,80858,81013,81130,81143,81154,81312,81410,81466,81467,81565,81596,81643,81681,81983,82021,82213,82214,82302,82412,82440,82443,82493,82636,82716,82837,82854,83365,83526,83536,83639,83714,83746,83776,83803,83810,83817,83824,83835,83901,83935,83997,84309,84358,84364,84394,84664,84762,85061,85169,85197,85198,85260,85356,85511,85624,85675,85778,85825,85852,85935,85953,86023,86302,86332,86336,86895,86947,87152,87168,87171,87361,87570,87718,87937,87948,88095,88162,88386,88443,88447,88452,88466,88560,88605,88725,89015,89110,89164,89270,89345,89362,89510,89511,89626,89684,89787,89823,89867,89897,90031,90125,90182,90224,90308,90317,90377,90678,90692,90734,91048,91158,91228,91300,91402,91510,91569,91702,91826,92135,92151,92552,92573,92589,92596,92616,92749,93016,93115,93147,93305,93385,93477,93562,93669,93691,93706,93839,94076,94182,94184,94238,94262,94343,94384,94576,94688,94710,94842,94860,95011,95208,95231,95405,95517,95561,95645,96065,96071,96221,96420,96522,96577,96639,96753,96762,96785,96836,96876,97340,97393,97463,97516,97560,97568,97669,97803,97831,97867,97978,98012,98014,98126,98283,98329,98343,98361,98375,98477,98643,98913,98983,99048,99085,99153,99170,99410,99551,99644,99661,99690,99874,99945",
          "58,88,202,372,481,658,751,823,844,956,1036,1212,1256,1263,1316,1356,1516,1517,1590,1695,1734,2184,2284,2450,2570,2608,2654,2787,2791,2796,2809,2818,2832,3054,3311,3334,3383,3388,3548,3595,3672,3688,3716,3749,3758,3791,3870,3910,3953,4159,4242,4294,4332,4409,4443,4555,4599,4683,4686,4693,4720,4783,4888,4964,4989,5024,5127,5182,5290,5338,5393,5476,5570,5603,5743,5755,5777,5880,6033,6059,6096,6125,6143,6200,6205,6278,6288,6305,6370,6707,6715,6729,6775,6826,6997,7088,7145,7161,7190,7198,7263,7293,7309,7387,7412,7516,7527,7623,7864,7870,7977,8002,8013,8165,8183,8217,8255,8282,8290,8312,8508,8583,8603,8676,8892,8980,8985,9031,9088,9174,9245,9270,9432,9454,9629,9766,9820,9845,10150,10152,10219,10478,10500,10628,10672,10745,10870,10926,10966,11118,11173,11275,11278,11285,11311,11395,11456,11525,11536,11561,11564,11605,11719,11752,11814,11831,11914,12043,12181,12275,12390,12681,12720,12751,12772,12816,12844,12864,12878,12934,12986,13018,13019,13174,13300,13301,13347,13388,13398,13417,13432,13599,13629,13634,13844,13846,13888,13960,14101,14159,14279,14356,14384,14390,14607,14613,14715,14720,14839,14908,14933,14978,14986,15053,15130,15330,15348,15351,15361,15504,15542,15561,15654,15749,15850,15935,16198,16280,16372,16424,16492,16615,16618,16665,16684,16742,16852,16935,17037,17064,17134,17346,17413,17506,17509,17545,17580,17716,17869,17913,17947,17971,17986,18085,18165,18205,18253,18315,18351,18359,18404,18434,18470,18494,18737,18864,19004,19180,19265,19370,19436,19452,19609,19654,19723,19790,19830,19833,19931,19937,20018,20084,20109,20115,20257,20308,20363,20366,20388,20653,20821,20874,20876,21040,21203,21206,21212,21281,21309,21311,21380,21388,21435,21631,21663,21687,21701,21826,21848,21857,21866,21935,21993,22112,22169,22228,22260,22288,22364,22378,22386,22401,22533,22608,22696,22721,22785,22838,22934,22976,23054,23066,23156,23254,23289,23379,23409,23477,23529,23663,23666,23710,23830,23858,23874,23910,23961,24135,24198,24215,24253,24295,24303,24306,24325,24443,24548,24603,24796,24803,24912,24917,24922,24944,25014,25135,25195,25237,25258,25354,25502,25675,25799,26103,26163,26214,26342,26635,26670,26763,26809,26813,26819,26850,26860,26964,26967,27037,27067,27115,27273,27275,27336,27370,27390,27410,27418,27565,27568,27744,27798,27813,27814,27831,27846,27857,27874,27983,27996,28017,28196,28300,28333,28392,28424,28509,28548,28710,28801,28860,29101,29262,29313,29578,29680,29780,29860,29929,29974,30061,30062,30160,30221,30224,30290,30330,30401,30508,30519,30723,30736,30786,30854,30876,30885,30921,30936,30945,30950,31254,31286,31293,31303,31391,31580,31634,31645,31707,31764,31802,31848,31857,31985,32049,32098,32130,32162,32171,32178,32197,32386,32394,32513,32734,32735,32884,32955,32957,32959,33084,33144,33196,33225,33317,33398,33508,33540,33631,33644,33693,33700,33756,33956,34020,34098,34145,34369,34391,34513,34556,34682,34769,34776,34840,34860,34868,34930,34981,35159,35187,35216,35358,35452,35635,35678,35859,36071,36125,36138,36244,36277,36480,36481,36597,36623,36757,36789,36810,36926,37156,37355,37439,37489,37575,37691,37712,37753,37782,37787,37841,37996,38014,38057,38070,38091,38245,38252,38280,38289,38315,38400,38550,38812,38864,38916,39009,39105,39152,39239,39243,39309,39339,39347,39360,39410,39412,39603,39613,39676,39844,39882,40054,40161,40336,40357,40387,40605,40737,40856,40946,40949,40956,41075,41121,41135,41153,41288,41328,41400,41490,41494,41515,41697,41834,42094,42118,42182,42184,42190,42239,42291,42392,42444,42461,42481,42596,42655,42822,42909,42961,42980,43047,43082,43134,43166,43221,43337,43391,43565,43680,43730,43769,43884,43938,43997,44007,44120,44146,44152,44328,44438,44453,44497,44614,44718,44887,44904,44985,45002,45081,45092,45168,45219,45327,45399,45528,45644,45727,45964,46016,46142,46304,46310,46428,46431,46505,46596,46675,46720,46944,46993,46998,47009,47166,47208,47266,47304,47447,47487,47913,47941,47942,47995,48051,48075,48193,48254,48364,48443,48496,48516,48539,48631,48701,48714,48731,48738,48804,48900,48916,48934,48942,49004,49083,49090,49338,49496,49497,49539,49621,49626,49733,49763,49870,49991,49992,50040,50141,50184,50228,50365,50784,50787,50805,50817,50870,50947,50952,50972,51003,51061,51137,51169,51174,51266,51324,51409,51523,51689,51696,51794,51886,51905,51909,52168,52318,52404,52687,52743,52857,52901,52924,52950,53248,53313,53517,54012,54036,54070,54103,54178,54234,54289,54308,54346,54568,54913,55018,55168,55476,55484,55508,55575,55671,55714,55730,55818,55878,55918,56005,56097,56447,56558,56629,56668,56711,56723,56835,57008,57115,57148,57160,57220,57382,57452,57485,57515,57559,57688,57712,57775,57849,57853,57862,57933,57954,57958,58123,58201,58218,58233,58284,58351,58373,58550,58588,58692,58941,58980,59354,59440,59466,59499,59544,59835,59968,59976,60133,60196,60308,60311,60379,60744,60748,60753,60839,60846,60861,60869,60946,61032,61087,61155,61193,61236,61321,61390,61429,61473,61542,61633,61692,61831,61835,61851,61853,61928,61939,61944,62031,62060,62175,62182,62259,62315,62335,62345,62363,62586,62593,62605,62817,62821,62910,63164,63287,63299,63434,63512,63518,63636,63644,63748,63779,63814,63935,63955,64092,64115,64209,64246,64322,64324,64420,64454,64589,64591,64611,64858,64902,64920,64952,65035,65080,65108,65285,65364,65719,65746,65762,65879,66058,66059,66064,66217,66300,66362,66497,66500,66511,66601,66607,66654,66743,66843,66861,67176,67178,67201,67202,67231,67367,67396,67448,67544,67575,67577,67731,67967,68028,68099,68140,68165,68221,68239,68359,68488,68526,68607,68621,68655,68691,68760,68806,68867,68996,69071,69139,69152,69194,69198,69354,69426,69541,69645,69780,69799,69805,69869,69901,69915,69986,69994,70098,70191,70196,70207,70298,70327,70559,70585,70866,70890,70894,70924,70976,71131,71152,71156,71212,71239,71369,71462,71602,71720,71877,72047,72101,72111,72112,72275,72313,72347,72369,72415,72573,72650,72724,72769,72835,73037,73042,73232,73259,73481,73672,73688,73792,73838,73976,73990,74192,74431,74462,74571,74646,74676,74742,74833,74920,75003,75007,75100,75168,75282,75404,75484,75496,75535,75613,75616,75931,75933,75980,76014,76040,76066,76106,76116,76147,76179,76320,76550,76964,77053,77083,77356,77610,77647,77655,77687,77712,77853,77923,77926,78008,78042,78124,78141,78157,78274,78326,78365,78395,78603,78676,78697,78720,78749,78862,78955,78987,79021,79073,79248,79358,79466,79491,79500,79584,79652,79697,79718,79752,79753,79956,79982,80127,80188,80261,80321,80342,80453,80526,80554,80824,80990,81214,81221,81230,81296,81306,81337,81427,81431,81558,81679,82042,82097,82109,82244,82316,82473,82534,82548,82557,82656,82718,82775,82992,83002,83068,83079,83099,83160,83298,83519,83896,84157,84212,84228,84284,84406,84418,84570,84589,84593,84619,84739,84740,84890,84919,85015,85054,85226,85434,85475,85523,85599,85623,85710,85712,85781,85787,85968,86055,86171,86190,86230,86350,86372,86489,86584,86646,87346,87362,87479,87486,87776,87814,87868,87878,87920,88053,88248,88339,88345,88351,88356,88414,88461,88554,88555,88633,88795,88885,89217,89249,89256,89343,89412,89453,89564,89586,89587,89640,89707,89715,89951,89968,89979,90225,90254,90416,90457,90481,90501,90651,90715,90807,90867,90984,90987,91289,91378,91393,91436,91516,91674,91701,91769,91813,91842,92149,92154,92239,92252,92345,92605,92686,92709,92896,93019,93086,93205,93312,93361,93439,93596,93667,93685,93911,94028,94137,94329,94371,94372,94396,94467,94506,94545,94670,94738,94837,94868,94933,95062,95162,95215,95234,95322,95364,95392,95408,95430,95475,95617,95654,95713,95838,95893,95936,95944,95972,95980,96098,96355,96431,96479,96572,96643,96801,96838,96905,96911,96966,97201,97216,97362,97369,97373,97459,97530,97565,97591,97673,97740,97753,97767,97982,98101,98114,98331,98414,98722,98738,98762,98893,98954,98977,99156,99165,99198,99643,99741,99758,99777,99830,99870,99942",
          "64,131,174,204,270,339,365,450,464,536,641,671,913,1110,1124,1126,1139,1596,1725,1731,1809,2078,2164,2191,2262,2283,2331,2400,2432,2462,2476,2513,2558,2575,2671,2675,2691,2777,2911,3063,3077,3178,3187,3192,3267,3327,3419,3432,3511,3515,3587,3630,3693,3757,3855,3896,3986,4063,4071,4125,4132,4293,4442,4640,4648,4650,4798,4841,4844,4934,4936,4963,4984,4990,5069,5212,5229,5334,5435,5544,5620,5790,6182,6285,6679,6808,6840,6914,7077,7156,7256,7276,7355,7406,7491,7690,7698,7979,8340,8374,8416,8572,8586,8606,8665,8692,8768,8820,8951,8976,9119,9132,9221,9226,9265,9309,9420,9433,9580,9607,9686,9787,9790,9862,9872,9944,9982,10180,10361,10422,10565,10592,10785,10869,10920,11071,11077,11150,11257,11258,11280,11312,11339,11359,11413,11436,11478,11556,11647,11686,11762,11854,11869,11878,11956,12016,12118,12193,12221,12371,12552,12571,12573,12845,13175,13219,13233,13238,13253,13380,13459,13477,13602,13610,13728,13771,13788,13806,14028,14146,14235,14418,14443,14599,14728,14811,14958,14972,15050,15212,15228,15328,15346,15363,15394,15477,15499,15505,15515,15586,15607,15678,15879,16077,16099,16154,16189,16209,16214,16309,16369,16404,16458,16487,16522,16531,16538,16629,16655,16660,16672,16780,16812,16814,16820,16967,16993,17014,17039,17085,17101,17252,17357,17397,17426,17427,17493,17663,17724,17940,17941,17967,17981,18075,18147,18183,18377,18448,18760,18868,18916,19003,19035,19037,19085,19206,19250,19319,19438,19457,19458,19477,19507,19662,19754,19819,19852,19867,19920,19926,19944,19988,19999,20013,20040,20267,20358,20361,20517,20542,20635,20656,20704,20825,20881,20950,20980,20981,21057,21234,21239,21298,21344,21429,21436,21469,21502,21507,21760,21798,21823,21858,21924,21950,21957,21982,22061,22067,22135,22137,22157,22175,22225,22322,22337,22342,22382,22417,22465,22504,22601,22663,22920,22928,22952,23027,23152,23260,23285,23312,23327,23355,23369,23385,23393,23402,23407,23431,23442,23594,23672,23691,23970,24066,24067,24087,24099,24400,24430,24431,24491,24499,24614,24644,24732,24736,24849,24882,24941,24972,24982,24997,25013,25018,25106,25137,25272,25327,25363,25395,25413,25553,25628,25645,25653,25734,25767,25908,25961,26032,26045,26116,26125,26134,26162,26171,26194,26254,26306,26312,26358,26419,26451,26453,26488,26512,26528,26611,26656,26782,26808,26851,26868,27042,27109,27160,27189,27223,27361,27364,27510,27585,27649,27766,28152,28194,28200,28349,28372,28431,28451,28546,28557,28625,28629,28652,28722,28734,28797,28813,28907,29002,29033,29120,29141,29189,29431,29446,29505,29531,29634,29755,29811,29846,29971,29984,30012,30069,30082,30159,30171,30176,30219,30245,30362,30363,30364,30372,30377,30428,30469,30474,30481,30535,30552,30560,30683,30720,30795,30860,30888,30915,30946,30999,31044,31045,31091,31096,31127,31188,31218,31241,31338,31404,31467,31501,31526,31687,31724,31760,31928,31988,32039,32078,32246,32285,32350,32359,32400,32418,32568,32678,32691,32697,32717,32793,32811,32821,32845,32847,32913,33038,33070,33103,33157,33199,33290,33394,33433,33450,33495,33500,33534,33584,33634,33667,33692,33727,33729,33777,33781,33803,33871,33893,34058,34325,34436,34523,34675,34711,34719,34855,34857,34905,35101,35144,35152,35156,35162,35176,35181,35229,35246,35308,35432,35646,35653,35670,35689,35835,35911,35956,36068,36171,36294,36507,36656,36770,36918,36998,37001,37138,37365,37487,37532,37623,37654,37669,37846,37987,38078,38093,38173,38348,38356,38378,38455,38460,38479,38556,38605,38636,38735,38757,38863,39053,39086,39193,39201,39346,39386,39431,39491,39528,39575,39618,39731,39739,39965,40099,40152,40175,40184,40226,40399,40415,40516,40568,40629,40632,40715,40843,40865,40887,40910,40916,40971,41002,41016,41026,41065,41080,41096,41250,41256,41426,41498,41580,41611,41629,41639,41764,41880,41969,42119,42120,42161,42193,42249,42255,42297,42391,42531,42585,42623,42641,42725,42736,42861,42867,42978,43019,43020,43099,43274,43280,43289,43442,43544,43563,43649,43700,43748,43762,43779,43849,43976,43979,44006,44083,44097,44129,44135,44288,44332,44368,44377,44410,44418,44523,44535,44583,44606,44831,44890,44920,44999,45004,45119,45197,45345,45487,45492,45517,45561,45631,45650,45673,45754,45762,45766,45778,45954,45970,46010,46046,46118,46347,46439,46493,46726,46737,46765,46804,46902,47103,47116,47371,47489,47506,47511,47543,47658,47661,47736,47748,47756,47767,48222,48304,48337,48417,48423,48639,48642,48689,48742,48751,48762,48796,48802,48852,48864,48885,48993,49096,49131,49147,49152,49205,49250,49327,49352,49358,49359,49406,49452,49482,49559,49633,49652,49698,49714,49739,49900,49971,50001,50014,50019,50039,50108,50233,50263,50296,50332,50383,50396,50447,50520,50601,50710,50794,50923,50942,51038,51211,51223,51580,51744,51754,51855,51990,52016,52121,52132,52268,52304,52329,52342,52375,52442,52476,52499,52620,52652,52696,52930,52985,53039,53134,53160,53182,53585,53685,53754,53849,53861,53870,53892,53925,53937,54006,54112,54276,54277,54323,54342,54717,54751,54905,54921,55022,55060,55085,55416,55451,55859,55887,55893,55917,55919,55958,56069,56196,56199,56241,56341,56343,56636,56724,56794,56952,56962,57083,57118,57161,57172,57283,57308,57354,57411,57539,57735,57750,57945,58087,58125,58141,58244,58328,58345,58385,58406,58691,58733,58738,58915,58935,59040,59046,59081,59094,59101,59105,59169,59408,59417,59588,59666,59686,59792,59867,60035,60067,60074,60094,60100,60131,60159,60296,60338,60358,60412,60438,60573,60602,60604,60772,60796,60856,60929,61241,61361,61383,61538,61580,61642,61701,61864,61937,62006,62306,62347,62526,62840,62898,62936,63174,63241,63267,63409,63576,63649,63682,63714,63718,63755,63783,63804,63821,63835,63907,63946,64113,64181,64203,64226,64347,64364,64473,64480,64585,64640,64688,64805,64894,64973,64997,65086,65131,65132,65154,65161,65448,65478,65595,65610,65748,65909,65965,65981,66045,66100,66118,66235,66412,66416,66452,66506,66617,66652,66690,66754,66986,67011,67022,67030,67211,67268,67355,67501,67532,67572,67584,67668,67685,67691,67723,67814,67820,67844,67905,67921,67966,67970,67974,68082,68155,68346,68379,68381,68425,68639,68643,68753,68765,68775,68785,68845,68885,68903,68944,69344,69422,69427,69606,69726,69738,69947,70043,70213,70251,70259,70303,70408,70410,70504,70557,70567,70597,70664,70711,70720,70933,71057,71217,71262,71441,71484,71504,71535,71536,71590,71685,71718,71791,71913,72100,72158,72185,72191,72292,72322,72453,72466,72503,73056,73153,73167,73248,73311,73364,73543,73546,73579,73581,73621,73842,73853,73921,74008,74069,74145,74148,74172,74196,74304,74493,74553,74671,74688,74832,74838,75083,75084,75240,75368,75433,75483,75562,75605,75697,75774,76036,76080,76128,76141,76167,76284,76303,76504,76553,76591,76624,76627,76713,76784,76855,76923,77094,77212,77216,77239,77245,77341,77432,77447,77621,77792,77822,77833,77834,77928,78007,78174,78182,78289,78319,78340,78342,78487,78632,78648,78689,78767,78788,78855,78869,78970,78977,79078,79087,79103,79217,79243,79257,79355,79374,79558,79600,79602,79926,79945,79992,80005,80108,80178,80308,80341,80414,80480,80670,80812,80877,80936,80943,80946,81008,81156,81251,81259,81317,81411,81450,81479,81666,81705,81729,81971,82006,82080,82151,82604,82684,82705,82736,82997,83177,83197,83218,83287,83336,83614,83660,83782,83794,83804,83834,83919,84003,84034,84043,84156,84203,84249,84283,84348,84428,84429,84470,84591,84707,84755,84839,84851,84863,84866,84897,84936,84951,85024,85130,85159,85257,85473,85514,85550,85616,85773,85814,85822,85884,85894,86084,86101,86203,86238,86275,86361,86478,86762,86980,87058,87153,87174,87226,87625,87659,87680,87765,87773,87788,87956,88071,88374,88679,88771,88860,88869,88969,89064,89127,89349,89378,89455,89741,89768,89852,89997,90034,90174,90186,90194,90223,90259,90480,90546,90578,90650,90980,91024,91049,91057,91120,91194,91243,91656,91666,91668,91717,91772,91836,91858,91874,92070,92144,92194,92247,92261,92348,92464,92520,92571,92798,92938,92954,93091,93104,93135,93191,93231,93253,93340,93512,93653,93658,93693,93695,93789,93880,93899,93948,94013,94023,94221,94224,94335,94388,94428,94463,94475,94503,94504,94510,94523,94548,94592,94654,94666,94809,94922,95046,95245,95280,95282,95445,95497,95516,95636,95647,95854,95877,95881,95924,95983,96031,96033,96089,96115,96262,96351,96449,96454,96817,96931,97082,97169,97290,97405,97478,97483,97488,97505,97528,97633,97809,97829,97858,98086,98274,98325,98353,98453,98526,98624,98747,98765,98802,98852,98934,98952,99114,99120,99291,99362,99370,99421,99466,99563,99568,99611,99613,99812,99839,99976,99998",
          "62,115,191,357,547,568,597,648,667,688,695,884,1011,1120,1153,1304,1354,1364,1467,1566,1603,1618,1685,1702,1819,1821,1853,1908,2139,2170,2277,2289,2306,2417,2471,2527,2590,2600,2623,2655,2730,2848,2879,3141,3251,3256,3303,3307,3339,3372,3390,3484,3593,3677,3709,3775,3970,4051,4120,4197,4221,4228,4397,4458,4521,4824,4872,4986,5132,5176,5183,5347,5437,5490,5505,5584,5657,5893,5953,6031,6292,6551,6556,6623,6630,6657,6664,6769,6975,7034,7153,7170,7178,7197,7305,7306,7332,7405,7472,7490,7518,7948,8121,8213,8241,8250,8262,8390,8444,8538,8633,8642,8648,8814,8840,8947,8957,8962,9035,9110,9116,9149,9284,9331,9335,9350,9351,9392,9402,9403,9416,9486,9578,9735,9753,9763,9971,9981,10141,10157,10187,10374,10410,10425,10432,10464,10553,10680,10737,10753,10796,10811,10829,10832,10910,10988,11062,11135,11138,11175,11185,11195,11203,11221,11222,11259,11420,11508,11515,11574,11643,11661,11754,11947,12002,12055,12065,12223,12271,12372,12415,12523,12569,12570,12614,12648,12657,12773,12789,12846,12880,12908,12916,12970,13000,13013,13109,13247,13303,13355,13565,13787,13825,13971,14066,14116,14194,14195,14353,14361,14544,14584,14627,14699,14807,14897,15052,15061,15108,15152,15283,15298,15436,15449,15495,15564,15633,15773,15781,15838,15888,15990,16311,16352,16471,16723,16752,16813,16854,16912,17023,17121,17148,17194,17220,17249,17453,17531,17558,17573,17579,17680,17725,17773,17802,17892,17912,18113,18267,18268,18332,18385,18468,18518,18537,18655,18735,18768,18777,18785,18800,18816,18857,18896,18911,19001,19170,19214,19219,19281,19304,19389,19649,19668,19766,19769,19908,20016,20145,20165,20171,20270,20316,20412,20630,20734,20920,21012,21044,21068,21076,21077,21275,21301,21306,21371,21483,21505,21541,21547,21713,21736,21909,21965,22102,22139,22172,22191,22266,22333,22431,22512,22627,22669,22692,22707,22737,22791,22865,22940,22989,23061,23062,23070,23073,23094,23118,23120,23131,23322,23427,23432,23469,23695,23707,23741,23779,23802,23805,23859,23913,23971,24036,24081,24218,24238,24278,24291,24373,24388,24396,24589,24705,24834,24995,25048,25090,25132,25159,25198,25225,25230,25358,25475,25476,25538,25630,25727,25824,26022,26033,26054,26341,26434,26527,26562,26714,26765,26822,26862,26938,27065,27218,27219,27250,27339,27379,27580,27839,27861,27942,27951,28081,28175,28205,28297,28310,28445,28500,28575,28619,28739,28791,28920,28933,28985,29001,29049,29100,29672,29702,29782,29914,29964,30318,30359,30415,30453,30522,30565,30615,30674,30773,30903,30909,30917,30923,30955,30958,30988,30994,31010,31024,31105,31118,31274,31305,31381,31446,31597,31607,31613,31641,31643,31823,31828,31897,31946,31957,31960,32070,32100,32103,32173,32207,32283,32311,32356,32614,32620,32740,32778,32805,32834,32858,32937,33133,33265,33333,33346,33348,33396,33405,33492,33660,33761,33858,33874,33904,33909,33954,34039,34050,34054,34231,34248,34260,34285,34287,34374,34428,34514,34567,34632,34764,34765,34791,34804,34891,34965,34974,34984,34989,35160,35183,35338,35393,35433,35442,35448,35457,35475,35563,35669,35701,35728,35774,35809,35826,35993,36024,36040,36073,36399,36405,36415,36423,36441,36511,36614,36666,36716,36717,36749,37000,37037,37041,37110,37139,37204,37233,37300,37461,37526,37558,37559,37593,37604,37740,37781,37819,37877,37917,37949,37972,38080,38170,38186,38403,38565,38621,38740,38941,39065,39079,39093,39175,39242,39287,39345,39407,39430,39489,39523,39539,39607,39799,40020,40078,40156,40165,40169,40207,40304,40376,40380,40407,40493,40494,40525,40820,40871,40921,40962,41160,41177,41201,41334,41372,41391,41461,41539,41574,41583,41702,41740,41760,41856,42047,42087,42111,42139,42145,42203,42217,42275,42369,42479,42512,42523,42680,42709,42724,42810,42887,42895,42931,43107,43185,43204,43241,43451,43560,43579,43845,43886,43948,43955,43987,43989,44016,44101,44144,44151,44261,44291,44297,44312,44363,44372,44378,44397,44428,44587,44677,44720,44771,44803,44876,44896,44903,44947,45003,45227,45284,45405,45417,45523,45593,45619,45637,45780,45836,45847,45926,46174,46204,46219,46297,46331,46342,46371,46438,46791,46858,46870,46896,46959,47046,47061,47357,47379,47579,47597,47688,47734,47772,47806,47837,47854,47878,48044,48080,48105,48131,48153,48209,48223,48308,48318,48329,48369,48373,48534,48568,48649,48686,48726,48744,48773,48806,48975,49189,49446,49449,49548,49582,49638,49648,49693,49756,49921,50022,50113,50235,50286,50366,50377,50475,50494,50555,50721,50864,50938,51117,51149,51201,51337,51398,51563,51586,51731,51816,51974,51976,52046,52150,52176,52182,52303,52372,52427,52468,52689,52830,52875,53065,53091,53114,53128,53147,53173,53174,53237,53483,53509,53562,53632,53661,53696,53740,53779,53796,53833,53836,53871,53949,54075,54118,54210,54217,54230,54241,54258,54268,54320,54484,54537,54668,54719,54733,54813,54974,55000,55075,55093,55102,55127,55207,55285,55526,55545,55594,55648,55763,55831,55833,55868,55902,55922,56161,56174,56183,56192,56275,56422,56478,56487,56544,56601,56680,56709,56722,56857,56916,56981,57033,57037,57122,57304,57499,57520,57723,57758,57812,57826,57842,57858,58007,58062,58092,58157,58277,58464,58602,58702,58721,58729,58732,58812,58944,59098,59138,59161,59215,59278,59312,59329,59367,59398,59439,59491,59526,59598,59611,59622,59671,59944,59957,59992,60060,60127,60261,60329,60355,60407,60445,60503,60523,60532,60555,60597,60736,60874,60919,61098,61203,61297,61328,61476,61515,61587,61789,61841,62249,62262,62300,62311,62313,62637,62700,62754,62852,62902,62915,62930,62969,62996,63005,63012,63096,63099,63193,63377,63378,63601,63705,63892,63983,64070,64205,64219,64222,64247,64251,64286,64316,64387,64497,64569,64576,64614,64739,64869,64943,65176,65222,65351,65459,65573,65619,65658,65866,65875,66009,66088,66152,66173,66222,66285,66436,66453,66530,66532,66629,66694,66707,66767,66784,66803,66828,66840,66857,66999,67292,67338,67377,67387,67477,67506,67526,67530,67623,67704,67779,67806,68007,68044,68158,68348,68422,68447,68537,68671,68705,68739,68745,68793,68890,68908,68967,68991,69086,69142,69355,69383,69463,69596,69803,69826,69828,70045,70274,70286,70644,70728,70736,70747,70780,71149,71165,71178,71221,71232,71315,71351,71370,71392,71623,71629,71702,71841,71848,71870,71875,71926,72065,72226,72317,72329,72333,72335,72363,72454,72730,72770,72796,72884,72904,72917,72958,72967,73030,73065,73122,73204,73247,73308,73314,73376,73409,73468,73580,73953,74018,74042,74231,74237,74290,74343,74472,74590,74693,74759,74901,75096,75120,75189,75191,75216,75297,75405,75633,75683,75763,75902,75915,75926,75935,75945,75982,75989,76030,76034,76055,76058,76092,76137,76139,76203,76252,76308,76600,76830,76907,76941,77037,77062,77112,77174,77232,77244,77292,77453,77511,77661,77678,77692,77734,77750,77814,77867,78065,78193,78402,78579,78846,79051,79057,79177,79325,79383,79517,79544,79614,79711,79780,79928,79975,79995,80004,80013,80028,80153,80224,80347,80507,80579,80647,80687,80787,80867,80882,80928,80950,80959,80977,81051,81127,81212,81271,81373,81402,81436,81492,81507,81579,81635,81641,81738,81769,81780,81925,81981,82002,82022,82115,82156,82159,82167,82262,82375,82550,82645,82682,82752,83035,83129,83220,83227,83260,83266,83565,83831,83838,83840,83948,83964,83968,83991,84120,84216,84280,84357,84464,84666,84780,84830,84859,84900,84926,84953,84998,85117,85177,85232,85288,85358,85455,85649,85701,85733,85792,85830,85857,85868,85892,85943,85999,86180,86250,86559,86730,86732,86749,86750,86797,86906,87010,87011,87086,87091,87183,87234,87251,87414,87602,87643,87667,87689,87695,87705,87706,87727,87738,87744,87793,87867,87872,87938,87954,87961,87963,87978,87991,88116,88212,88225,88695,88739,88789,88986,89048,89252,89372,89439,89518,89546,89676,89791,89843,89855,89910,90070,90107,90215,90250,90380,90381,90495,90500,90548,90601,90634,90643,90677,90686,90725,90741,90824,91115,91148,91214,91337,91466,91565,91585,91594,91662,91679,91680,91696,91709,91743,91798,91845,91906,91919,92054,92056,92094,92148,92228,92281,92350,92399,92403,92436,92465,92532,92575,92811,92884,92979,93030,93102,93144,93162,93232,93268,93295,93411,93475,93496,93564,93655,93674,93832,93888,93923,94056,94124,94285,94332,94426,94438,94481,94502,94526,94563,94694,94715,94725,94954,95075,95343,95419,95448,95585,95635,95786,95823,95847,95959,96045,96175,96209,96213,96525,96661,96673,96886,96915,96918,96951,97035,97099,97144,97246,97322,97374,97398,97445,97623,97857,97914,97977,97983,98054,98150,98216,98252,98337,98457,98553,98583,98666,98706,98725,98741,98872,98927,99074,99224,99287,99546,99624,99759,99908",
          "57,59,219,294,413,603,920,1044,1105,1167,1345,1351,1462,1583,1634,1692,1770,1793,1843,1877,1924,1993,2036,2068,2140,2208,2295,2368,2389,2594,2653,2721,2741,2820,2930,2948,2966,3148,3190,3233,3261,3308,3345,3384,3474,3480,3539,3574,3646,3819,3892,3969,4073,4133,4208,4369,4406,4417,4441,4528,4571,4652,4825,4858,4859,4866,4994,5158,5185,5195,5446,5611,5993,6005,6092,6151,6219,6252,6326,6349,6375,6389,6457,6552,6663,6668,6677,6716,6740,6779,6929,6933,6986,6993,7144,7210,7222,7231,7361,7369,7497,7589,7593,7633,7689,7769,7811,7872,7933,7950,8030,8222,8237,8242,8288,8291,8293,8378,8439,8451,8459,8600,8696,8857,8869,8993,9186,9273,9288,9332,9337,9460,9480,9483,9641,9712,9754,9869,9876,9914,10013,10032,10047,10120,10146,10189,10214,10308,10383,10397,10461,10482,10575,10622,10649,10663,10781,10819,10941,10976,11034,11072,11134,11140,11179,11317,11351,11414,11469,11489,11530,11553,11568,11594,11633,11711,11733,11747,11893,12096,12167,12208,12351,12387,12563,12582,12625,12769,12829,12910,12944,13087,13090,13217,13228,13400,13502,13515,13595,13653,13824,13829,13902,13904,13977,14283,14295,14301,14490,14501,14524,14547,14562,14706,14803,14830,14929,14960,14961,14967,14998,15010,15028,15074,15077,15087,15139,15311,15379,15404,15442,15478,15479,15496,15669,15688,15689,15702,15766,15810,15835,15913,15944,15963,15977,15979,15985,15987,15993,16058,16067,16210,16222,16330,16398,16453,16620,16633,16659,16676,16735,16832,16837,16896,17057,17096,17158,17289,17615,17630,17639,17684,17791,17867,17883,18095,18110,18236,18240,18330,18352,18417,18461,18597,18643,18703,18824,18891,18917,18918,19012,19122,19402,19426,19492,19572,19574,19661,19841,19860,19880,19921,19951,19958,19970,20027,20062,20135,20147,20269,20310,20418,20589,20600,20738,20739,20815,21082,21152,21211,21339,21531,21571,21576,21595,21617,21658,21748,21816,21897,22046,22205,22303,22461,22463,22531,22543,22629,22720,22794,22830,22848,22875,22986,23072,23196,23391,23396,23637,23876,23898,23941,24057,24141,24144,24169,24186,24210,24211,24389,24399,24405,24480,24764,24862,24865,25117,25248,25317,25370,25397,25408,25644,25673,25750,25956,26008,26088,26188,26211,26352,26381,26389,26428,26443,26524,26525,26592,26701,26740,26803,26816,26859,26887,26889,27043,27111,27185,27227,27286,27293,27300,27340,27341,27347,27414,27532,27566,27616,27661,27710,27731,27915,27991,28002,28118,28173,28176,28273,28319,28413,28441,28478,28488,28545,28632,28719,28730,28892,28904,28910,28945,28972,29098,29136,29222,29238,29273,29344,29361,29410,29476,29575,29608,29652,29654,29836,29847,29995,30006,30008,30036,30076,30088,30120,30152,30212,30279,30416,30461,30523,30570,30709,30787,30927,30976,30984,31051,31052,31067,31112,31178,31222,31298,31399,31570,31573,31640,31682,31737,31811,31924,32022,32057,32154,32625,32668,32797,32938,32981,32997,33238,33330,33461,33473,33517,33539,33633,33649,33673,33714,33780,33846,33849,33935,33936,33945,34004,34097,34119,34160,34162,34211,34579,34620,34708,34749,34802,34823,34846,34914,35410,35435,35500,35560,35605,35607,35676,35878,35963,35985,36006,36053,36103,36110,36135,36486,36490,36622,36667,36735,36798,36830,37067,37075,37188,37216,37360,37375,37428,37449,37476,37603,37608,37629,37648,37739,37801,38042,38060,38076,38274,38276,38355,38408,38410,38489,38493,38592,38766,38771,38799,38801,39107,39116,39149,39245,39325,39361,39384,39471,39502,39525,39919,40039,40077,40212,40282,40353,40359,40405,40458,40468,40700,40707,40724,40802,40815,40900,40915,41004,41068,41158,41242,41291,41326,41342,41466,42031,42039,42206,42231,42247,42308,42325,42371,42423,42500,42506,42536,42896,42905,43119,43171,43285,43288,43389,43586,43612,43646,43684,43756,43994,44077,44237,44253,44259,44317,44350,44369,44557,44624,44734,44840,44946,45035,45188,45286,45320,45394,45449,45497,45594,45678,45751,45774,45857,45882,46042,46133,46199,46259,46458,46509,46550,46577,46595,46639,46656,46771,46885,47044,47064,47088,47141,47186,47228,47237,47446,47460,47626,47833,47936,48000,48014,48323,48471,48551,48589,48592,48677,48740,48808,48824,48846,48859,48927,48977,49162,49342,49450,49451,49485,49486,49511,49687,49774,49817,49820,49835,49949,50127,50399,50525,50701,50827,50932,51018,51027,51122,51126,51271,51277,51308,51417,51493,51746,51784,51788,51906,51908,51958,52019,52032,52039,52149,52558,52592,52727,52753,52773,52783,52993,53037,53074,53083,53212,53220,53293,53348,53394,53488,53651,53705,53706,53717,53720,53745,53747,53769,53770,53787,53854,54010,54022,54150,54291,54331,54613,54694,54780,54879,54885,54975,55083,55123,55250,55275,55333,55469,55472,55486,55550,55577,55591,55612,55657,55743,55875,55896,56013,56338,56345,56426,56466,56633,56692,56721,56754,56832,56940,56943,57038,57069,57214,57239,57265,57428,57508,57638,57732,57790,57804,57927,57944,57955,58065,58144,58379,58469,58620,58696,58919,59039,59068,59166,59249,59256,59406,59539,59688,59710,59711,59840,59904,60118,60392,60405,60458,60868,60920,60939,60952,61033,61043,61201,61255,61416,61459,61461,61562,61603,61651,61743,61757,61764,61828,61915,61996,62000,62073,62310,62434,62542,62642,62664,62671,62673,62683,62740,62830,62847,62901,62921,62960,63140,63324,63367,63386,63419,63475,63544,63652,63702,63841,64014,64062,64095,64155,64230,64252,64258,64303,64304,64444,64475,64516,64656,64777,64991,65418,65444,65508,65521,65554,65697,65699,65785,65795,65948,65993,66003,66020,66150,66187,66242,66243,66368,66382,66535,66536,66648,66662,66830,66904,67092,67182,67226,67323,67420,67494,67535,67567,67573,67705,67790,67792,67793,68032,68120,68145,68162,68330,68414,68464,68472,68507,68586,68623,68630,68717,68744,68842,68915,68935,68947,68964,68988,68993,69020,69021,69124,69129,69169,69219,69240,69294,69347,69363,69455,69465,69502,69551,69652,69822,69856,69941,69988,70152,70240,70323,70468,70571,70592,70977,70985,70986,71051,71085,71138,71151,71247,71327,71434,71548,71556,71564,71631,71829,71830,71908,72058,72122,72314,72416,72474,72505,72570,72653,72696,72704,72716,72738,72763,72834,72979,73188,73218,73229,73331,73332,73340,73540,73724,73786,73844,73874,73909,73970,73991,74324,74369,74388,74439,74537,74552,74692,74756,74792,74926,74947,74964,75034,75108,75157,75278,75358,75635,75729,75826,75876,75909,75968,76054,76104,76150,76204,76205,76229,76274,76286,76330,76359,76371,76376,76524,76610,76783,76824,76948,77007,77019,77050,77070,77133,77141,77271,77394,77464,77630,77698,77767,77914,77932,77952,78022,78074,78129,78200,78272,78301,78337,78351,78382,78411,78437,78498,78607,78631,78819,78863,79016,79109,79179,79208,79271,79279,79537,79658,79675,79686,79708,79732,79764,79777,79909,79950,79970,80022,80054,80086,80168,80213,80226,80326,80452,80591,80630,80722,81001,81033,81037,81044,81081,81107,81169,81269,81459,81528,81840,81853,81910,81941,82031,82117,82148,82155,82240,82248,82327,82505,82572,82587,82592,82606,82679,82785,82809,82932,83023,83024,83185,83284,83313,83392,83430,83445,83622,83737,83765,83783,83974,83992,84035,84199,84330,84490,84546,84576,84896,85032,85033,85158,85306,85329,85377,85418,85488,85650,86044,86172,86291,86329,86335,86337,86505,86542,86637,86641,86642,86686,86701,86805,86848,86892,86975,87017,87029,87071,87208,87327,87354,87501,87577,87594,87596,87616,87900,87945,87979,88036,88047,88178,88267,88306,88326,88520,88798,88853,88946,88987,89174,89186,89301,89329,89331,89563,89625,89669,89821,89850,89930,90016,90055,90064,90078,90120,90133,90213,90339,90404,90412,90442,90454,90702,90836,90839,90903,90906,90919,90921,90983,91015,91125,91508,91549,91560,91579,91581,91707,91925,92013,92132,92357,92361,92653,92656,92669,92736,92780,92886,92932,93435,93448,93488,93573,93580,93617,93767,93784,93842,94032,94092,94157,94219,94296,94309,94323,94367,94483,94596,94755,94772,94878,94892,94962,95058,95152,95176,95374,95422,95742,95848,95915,95981,95996,96101,96164,96305,96401,96429,96556,96793,96834,96848,96912,96934,96994,97164,97215,97442,97446,97496,97702,97711,97738,97780,97848,97869,98049,98070,98090,98156,98183,98246,98259,98271,98370,98410,98439,98451,98454,98555,98609,98610,98629,98822,98889,99077,99082,99086,99119,99123,99130,99208,99217,99237,99255,99270,99345,99409,99460,99793,99956",
          "20,37,102,156,185,186,719,859,876,973,997,1078,1106,1159,1273,1547,1799,1994,2014,2032,2064,2135,2175,2203,2461,2464,2599,2718,2732,2751,2917,2997,3086,3106,3154,3158,3201,3248,3387,3456,3590,3632,3675,3700,3922,3983,4082,4199,4244,4392,4558,4634,4811,4955,5028,5112,5292,5332,5363,5501,5512,5573,5651,5671,5711,5841,5885,5926,6004,6038,6263,6321,6440,6574,6725,6800,6866,6910,6962,7066,7294,7368,7373,7515,7569,7709,7857,8113,8271,8351,8406,8737,8855,8887,9011,9014,9070,9076,9128,9200,9238,9360,9409,9419,9424,9456,9467,9558,9602,9652,9732,9838,9840,9963,9998,10009,10012,10038,10173,10247,10421,10449,10510,10751,10769,10777,10810,10940,11024,11041,11099,11104,11284,11348,11411,11434,11480,11538,11627,11740,11803,11909,11961,11983,12131,12150,12288,12292,12337,12386,12436,12489,12491,12525,12568,12579,12604,12634,12722,12905,13055,13119,13218,13229,13286,13320,13547,13577,13635,13683,13713,13748,13756,13798,13849,13864,13908,13956,14056,14085,14086,14184,14258,14275,14398,14423,14430,14517,14618,14677,14712,14745,14785,15046,15096,15166,15201,15279,15395,15461,15497,15514,15573,15578,15589,15597,15617,15649,15820,15926,16003,16089,16102,16118,16157,16299,16386,16924,16927,16942,16966,16984,17042,17143,17160,17178,17242,17248,17250,17497,17521,17542,17795,17846,17870,18248,18305,18451,18493,18562,18592,18906,18907,19198,19323,19367,19411,19442,19482,19535,19632,19663,19717,19896,19905,19927,19953,19961,20037,20217,20226,20289,20343,20352,20375,20379,20465,20486,20516,20521,20688,20764,20803,20806,20809,20847,20998,21097,21135,21286,21335,21366,21539,21603,21609,21650,21832,21844,21885,21964,21970,22025,22074,22091,22145,22167,22168,22187,22297,22312,22368,22622,22745,22770,22810,22899,22946,23028,23045,23106,23134,23237,23579,23679,23757,23778,23792,23856,24071,24076,24193,24290,24433,24438,24494,24571,24587,24615,24668,24723,24763,24844,24881,25010,25016,25203,25241,25266,25341,25391,25425,25467,25721,25782,25930,25951,26081,26248,26469,26516,26604,26610,26628,26734,26798,26844,26870,26877,26931,27046,27074,27157,27177,27318,27366,27452,27494,27741,27984,28142,28192,28204,28236,28315,28584,28685,28802,28886,28901,29016,29266,29428,29466,29479,29536,29798,29799,29953,29994,30011,30017,30092,30271,30325,30367,30486,30540,30600,30620,30631,30644,31362,31371,31705,31744,31825,32038,32072,32352,32438,32652,32701,32971,33024,33036,33413,33419,33438,33587,33754,33840,33848,33886,33898,33914,33960,33998,34069,34147,34385,34484,34573,34639,34648,34713,34761,34763,34839,34968,35027,35145,35268,35343,35368,35695,35704,35768,36114,36127,36152,36218,36470,36564,36704,36752,36820,36828,36895,37055,37088,37162,37211,37400,37557,37601,37609,37652,37728,37850,37936,37948,37967,38008,38156,38237,38580,38617,38623,38639,38674,38700,38732,38829,38850,38915,39162,39222,39257,39312,39326,39362,39394,39727,39914,39916,39944,40018,40019,40132,40143,40145,40149,40176,40182,40234,40367,40537,40615,40690,40746,40786,40826,41079,41171,41223,41283,41346,41373,41440,41455,41696,41730,41776,41802,41945,41988,42180,42234,42436,42525,42577,42774,43390,43561,43570,43735,43885,43990,44058,44224,44277,44391,44446,44458,44461,44559,44757,44983,45038,45101,45108,45454,45530,45674,45718,45777,45888,45915,45956,46024,46064,46102,46155,46306,46313,46546,46556,46564,46571,46581,46600,46674,46733,46779,46799,46871,46987,47094,47098,47110,47199,47268,47290,47328,47400,47544,47567,47569,47637,47653,47696,47817,47847,47883,47894,48106,48147,48180,48264,48267,48336,48567,48640,48680,48728,48878,48923,48992,49012,49024,49199,49326,49428,49444,49641,49767,49970,49993,50015,50097,50102,50163,50211,50307,50510,50596,50826,50871,50945,51129,51292,51298,51366,51403,51423,51539,51840,51972,52135,52139,52172,52188,52257,52261,52323,52340,52379,52470,52766,52799,52852,53150,53268,53277,53323,53398,53413,53456,53523,53539,53613,53724,53841,54020,54053,54126,54196,54262,54288,54297,54300,54326,54450,54575,54604,54701,54728,54932,54953,55032,55098,55209,55324,55409,55465,55776,55807,56000,56088,56096,56223,56263,56279,56311,56476,56588,56683,56756,56758,56802,56850,56853,57028,57030,57048,57272,57346,57376,57405,57646,57843,57999,58037,58177,58217,58291,58339,58390,58446,58522,58538,58631,58697,58706,58749,59091,59126,59305,59315,59317,59548,59573,59574,59620,59691,59807,59811,59997,60086,60097,60114,60180,60194,60484,60497,60616,60644,60690,60732,60795,60944,60949,61017,61204,61273,61320,61345,61351,61417,61552,61555,61571,61801,61850,61982,62049,62052,62185,62316,62333,62359,62369,62442,62467,62480,62613,62615,62620,62643,62657,62751,62756,62765,62917,62925,62986,63028,63141,63194,63215,63296,63407,63520,63645,63654,63757,63826,64027,64053,64142,64160,64348,64525,64644,64691,64898,64959,65044,65114,65214,65224,65353,65360,65361,65405,65456,65664,65673,65773,65820,65825,65927,65971,66195,66216,66225,66258,66292,66622,66660,66766,66858,66995,67056,67059,67064,67133,67270,67314,67379,67422,67521,67569,67598,67649,67709,67782,67857,67975,68021,68050,68095,68188,68192,68241,68284,68538,68617,68646,68836,68978,69052,69080,69107,69119,69283,69300,69375,69391,69396,69434,69473,69522,69864,69919,69953,69992,70005,70060,70182,70197,70415,70491,70675,70707,70714,70792,71079,71211,71322,71371,71397,71433,71461,71505,71573,71642,71653,71686,71936,71995,72004,72023,72173,72389,72513,72525,72550,72552,72577,72670,72715,72801,72826,72894,72903,72991,73011,73026,73135,73154,73282,73283,73347,73348,73353,73423,73480,73670,73695,73731,73769,74062,74119,74147,74235,74444,74509,74555,74620,74680,74741,74813,74822,75005,75164,75207,75263,75270,75581,75651,75730,75821,75823,75830,76158,76184,76269,76272,76296,76367,76378,76408,76488,76511,76601,76682,76734,76840,76870,77028,77059,77183,77196,77357,77395,77484,77487,77550,77568,77577,77614,77668,77918,77994,77995,78009,78107,78116,78177,78406,78414,78448,78452,78524,78566,78624,78737,78773,78787,78845,78929,78946,78952,79012,79044,79052,79130,79147,79349,79365,79369,79414,79451,79543,79583,79730,79771,79787,79831,79857,79884,79932,79952,80036,80075,80410,80430,80462,80478,80483,80842,80843,80887,80927,80940,81186,81325,81422,81483,81614,81660,81711,81721,81857,81912,81934,81994,82093,82172,82329,82408,82468,82499,82753,83038,83048,83105,83277,83334,83671,83685,83745,83876,83934,83947,84149,84298,84311,84497,84561,84566,84587,84625,85018,85065,85190,85246,85340,85430,85737,85765,85971,85998,86086,86166,86242,86324,86482,86543,86836,87005,87094,87184,87194,87217,87232,87420,87551,87700,87720,87740,87810,87859,87860,88007,88023,88042,88070,88109,88287,88451,88591,88776,88911,88998,89489,89556,89593,89639,89643,89655,89995,90003,90011,90066,90236,90353,90476,90547,90571,90830,91435,91512,91721,91775,91844,92046,92103,92139,92181,92198,92227,92262,92299,92384,92404,92566,92583,92667,92678,92738,92841,92871,92978,93234,93327,93424,93526,93534,93802,93871,93929,93964,94027,94154,94256,94383,94403,94482,94488,94678,94719,94908,94968,95086,95274,95326,95413,95436,95507,95600,95642,95766,95800,95863,95898,95908,95913,96041,96073,96149,96150,96165,96214,96277,96289,96318,96494,96607,96679,96722,96864,97004,97024,97071,97107,97123,97158,97232,97434,97487,97612,97834,97906,97915,97925,97974,97981,98466,98605,98620,98876,99117,99445,99610,99633,99647,99730,99811,99924,99929,99933",
          "67,264,301,436,441,525,532,556,563,628,669,716,756,852,1003,1079,1175,1336,1417,1449,1602,1629,1769,1869,1955,1981,1996,2095,2146,2404,2421,2497,2586,2647,2677,2702,2771,2846,2908,2916,3070,3088,3180,3305,3354,3361,3400,3438,3624,3704,3755,3798,3813,3878,3958,4023,4107,4219,4308,4370,4439,4498,4509,4734,4737,4904,4940,5104,5306,5352,5546,5582,5851,5852,5914,6010,6048,6118,6160,6240,6317,6320,6430,6432,6612,6628,6655,6903,6935,6981,6991,7073,7094,7123,7184,7189,7382,7435,7460,7529,7810,7879,7905,7974,8048,8049,8067,8087,8146,8188,8199,8289,8345,8369,8521,8597,8669,8673,8703,8752,8761,8778,8836,8931,9025,9048,9089,9172,9269,9316,9341,9373,9546,9614,9646,9814,9964,10014,10056,10118,10426,10446,10491,10547,10556,10606,10682,10701,10776,10778,10801,10848,10927,10935,10952,10982,11036,11133,11142,11233,11256,11304,11407,11432,11708,11721,11920,11980,12122,12234,12255,12299,12327,12504,12600,12712,12838,12840,12933,12938,12968,12976,13040,13200,13240,13639,13692,13750,13812,13920,14034,14208,14257,14309,14337,14464,14629,14779,14838,14859,14940,14962,15042,15145,15178,15200,15254,15282,15289,15360,15492,15509,15525,15642,15747,15789,16083,16087,16104,16119,16196,16231,16320,16393,16555,16631,16696,16728,16827,16973,17126,17145,17147,17218,17269,17295,17313,17392,17474,17499,17508,17650,17740,17755,17782,17790,17812,17914,18175,18226,18327,18341,18421,18465,18609,18623,18659,18661,18752,18791,18794,18844,18861,18879,18900,18952,19134,19135,19146,19291,19316,19340,19355,19449,19451,19531,19576,19600,19799,19879,19886,19911,19913,20079,20104,20122,20159,20264,20323,20466,20505,20612,20728,20758,20761,20814,21017,21093,21193,21274,21332,21400,21415,21478,21486,21516,21651,21664,21685,21804,21827,21998,22028,22049,22081,22140,22185,22217,22366,22586,22643,22742,22919,23022,23220,23245,23248,23280,23398,23403,23497,23543,23597,23649,23668,23810,24242,24245,24578,24786,24871,24968,25123,25150,25172,25216,25369,25371,25510,25563,25629,25643,25702,25703,25741,26140,26270,26353,26403,26442,26457,26470,26519,26609,26616,26733,26888,26927,27031,27044,27133,27151,27175,27238,27272,27324,27346,27435,27463,27470,27556,27648,27709,27723,27747,27757,27784,27841,27958,28106,28189,28195,28199,28255,28410,28564,28568,28675,28761,28769,28825,28911,28913,29005,29020,29134,29148,29159,29204,29281,29312,29315,29470,29612,29636,29657,29766,29819,29956,29961,29991,30024,30110,30210,30254,30329,30383,30467,30593,30668,30759,30812,30882,30924,30962,31016,31023,31037,31076,31284,31363,31382,31419,31464,31490,31502,31625,31639,31658,31754,31862,31930,32001,32112,32160,32165,32396,32443,32469,32509,32538,32588,32616,32749,32756,32765,32819,32853,32879,33079,33081,33115,33214,33358,33588,33626,33668,33671,33695,33759,33795,33802,33805,33863,33933,34044,34140,34156,34330,34359,34398,34437,34516,34578,34604,34680,34688,34788,34880,34954,35009,35044,35114,35171,35315,35342,35456,35537,35559,35690,35697,35787,35882,35894,35980,36044,36227,36321,36383,36438,36502,36591,36632,36741,36750,36804,36829,36978,37149,37223,37418,37423,37615,37777,37800,37813,37878,37932,37939,37957,38205,38233,38295,38326,38346,38382,38409,38724,38762,38816,38978,38984,39125,39219,39290,39330,39378,39397,39417,39457,39565,39658,39678,39692,39717,39784,39859,39924,39929,39936,39961,40095,40197,40200,40205,40352,40374,40456,40460,40598,40617,40628,40645,40763,40881,40897,40995,41161,41164,41302,41323,41448,41712,41719,41747,41771,42006,42075,42277,42337,42407,42447,42594,42602,42645,42693,42799,42860,42877,42878,42923,42951,43083,43216,43234,43377,43407,43488,43501,43526,43641,43728,43729,43786,43815,43882,43902,43943,44198,44442,44505,44651,44736,44826,44881,44891,44971,45037,45098,45120,45127,45133,45155,45360,45425,45475,45609,45688,45825,45953,46023,46057,46114,46150,46285,46307,46359,46448,46574,46693,46738,46780,46813,46994,47128,47134,47215,47252,47331,47348,47417,47468,47519,47598,47642,47787,47905,47935,48116,48146,48196,48199,48207,48435,48526,48590,48645,48662,48706,49028,49145,49158,49319,49402,49459,49575,49587,49609,49613,49771,49773,49781,49831,49903,49962,50266,50338,50358,50505,50700,50819,50851,50992,51024,51050,51051,51353,51389,51392,51676,51732,51773,51950,52103,52113,52126,52536,52600,52822,52898,53021,53156,53233,53303,53304,53447,53477,53526,53529,53570,53579,53586,53604,53652,53711,53819,53927,54282,54303,54339,54380,54386,54400,54605,54631,54666,54674,54723,54815,54843,54864,54930,55100,55110,55128,55177,55231,55234,55242,55262,55264,55304,55428,55457,55501,55713,55780,55863,55867,55872,55986,56009,56048,56063,56228,56294,56347,56389,56412,56492,56574,56693,56814,56993,57116,57124,57175,57381,57446,57493,57501,57522,57526,57792,57839,57949,58250,58272,58292,58455,58466,58473,58568,58674,58677,58700,58705,58834,58853,58873,58903,59092,59123,59291,59427,59474,59528,59571,59648,59727,59810,59836,59837,59875,59919,60025,60124,60192,60208,60251,60315,60479,60481,60640,60663,60778,60819,60857,60866,60938,61042,61107,61213,61305,61450,61474,61482,61504,61532,61577,61684,61693,61790,61905,61956,61974,61986,62009,62014,62096,62102,62168,62195,62214,62269,62374,62485,62508,62530,62566,62630,62639,62698,62743,62949,63040,63042,63067,63166,63200,63232,63259,63270,63289,63391,63460,63513,63561,63588,63591,63626,63634,63662,63686,63762,63794,63836,63863,63977,64060,64153,64167,64187,64212,64301,64362,64433,64579,64720,64773,64823,64835,64859,64905,64911,64961,65041,65104,65166,65411,65460,65483,65507,65529,65754,65760,65821,65910,65946,65959,65961,65982,66055,66084,66129,66228,66533,66567,66610,66686,66711,66749,66823,66841,66880,66902,66923,66944,66954,67013,67060,67154,67234,67242,67401,67429,67461,67562,67621,67813,67865,67898,67912,67930,67945,67996,68046,68079,68128,68163,68197,68216,68332,68352,68397,68514,68542,68543,68567,68590,68599,68618,68620,68628,68663,68713,68902,68994,69103,69138,69209,69329,69560,69730,69998,70091,70164,70173,70328,70428,70443,70472,70526,70688,70756,70811,70864,70871,70936,70937,70962,71043,71195,71207,71402,71430,71452,71514,71597,71616,71635,71683,71871,71924,71981,72160,72204,72306,72386,72401,72419,72541,72548,72624,72672,72701,72705,72714,72877,73075,73080,73096,73151,73208,73285,73323,73362,73494,73706,73737,73772,73790,73904,73923,74020,74072,74180,74200,74356,74387,74535,74574,74629,74675,74843,74903,75019,75022,75035,75151,75181,75261,75304,75339,75388,75408,75493,75557,75565,75580,75582,75654,75708,76100,76122,76154,76180,76253,76281,76377,76437,76481,76546,76580,76697,76817,76835,76867,77129,77253,77383,77522,77620,77703,77708,77743,77748,77782,77878,77980,78028,78034,78158,78159,78423,78445,78602,78622,78644,78679,78817,78980,79371,79424,79518,79525,79604,79616,79639,79666,79684,79750,79791,79980,80035,80184,80292,80340,80708,80764,80783,80876,80967,81080,81119,81171,81209,81267,81281,81390,81464,81473,81529,81580,81594,81601,81613,81650,81743,81779,81908,81930,81985,82173,82183,82216,82249,82270,82303,82370,82585,82597,82744,82778,82818,82999,83065,83104,83144,83204,83366,83386,83422,83427,83572,83581,83687,83691,83695,83706,83793,83944,83955,84019,84082,84092,84385,84459,84485,84555,84734,84771,84781,84782,84803,84807,84886,84889,84909,85068,85069,85070,85139,85192,85264,85269,85302,85513,85532,85617,85627,85694,85856,85878,85891,85903,86093,86189,86212,86272,86276,86322,86367,86432,86460,86481,86530,86666,86847,86852,86877,87046,87182,87360,87467,87566,87592,87650,87661,87719,87752,87768,87781,87806,87812,87825,88050,88187,88260,88411,88539,88604,88757,88829,88846,89135,89268,89291,89355,89361,89484,89627,89661,89682,89739,89765,89866,89869,89870,89888,89918,89963,89983,90025,90148,90594,90733,90757,90769,90920,91030,91053,91206,91323,91464,91634,91794,91839,91859,91873,91917,91931,91994,92071,92383,92489,92503,92509,92514,92645,92690,92740,92909,92916,93009,93034,93161,93322,93364,93402,93404,93406,93428,93429,93509,93604,93628,93772,93794,93867,94045,94101,94228,94279,94316,94355,94361,94409,94468,94555,94662,94819,94889,94989,94991,95021,95138,95308,95316,95372,95450,95518,95716,95757,95764,95811,95814,95938,95940,95968,95984,95993,96147,96186,96419,96455,96469,96483,96492,96529,96726,96741,96830,96983,96987,97001,97138,97300,97368,97411,97443,97539,97677,97766,97912,98089,98119,98199,98446,98518,98528,98582,98595,98739,98843,98926,98932,98941,98993,98999,99039,99066,99076,99078,99098,99271,99307,99327,99332,99431,99442,99492,99497,99797,99909,99952,99960",
          "105,159,207,235,364,587,640,743,789,1054,1067,1174,1177,1200,1205,1309,1514,1659,1728,1767,1805,1845,1910,1950,1958,1959,1964,1971,1975,2022,2028,2099,2132,2178,2238,2255,2273,2300,2327,2357,2359,2475,2496,2506,2540,2549,2553,2556,2622,2627,2642,2663,2886,2909,2951,2994,3079,3097,3220,3231,3269,3277,3312,3377,3465,3473,3613,3667,3713,3734,3823,3898,3937,3987,4021,4043,4078,4087,4094,4116,4117,4139,4149,4239,4267,4268,4274,4296,4325,4356,4363,4459,4481,4506,4534,4581,4605,4702,4746,4767,4864,4880,4910,4921,5345,5390,5396,5457,5464,5551,5556,5600,5706,5740,6335,6351,6438,6452,6494,6528,6811,6825,6843,6875,6883,6888,6945,6992,7350,7425,7437,7767,7901,8245,8281,8684,8809,8831,8861,8863,8966,9022,9107,9141,9241,9362,9375,9399,9507,9650,9669,9692,9725,9860,9870,10007,10015,10101,10103,10202,10220,10258,10304,10428,10605,10761,10792,10881,10896,11003,11016,11112,11113,11159,11180,11294,11299,11320,11328,11399,11415,11418,11523,11603,11739,11788,11843,12046,12135,12246,12305,12309,12316,12374,12412,12487,12572,12673,12783,12847,13014,13032,13065,13120,13185,13276,13439,13576,13645,13765,13952,13993,14001,14011,14136,14145,14152,14278,14317,14323,14334,14507,14663,14678,14717,14768,14791,14793,14853,14910,14943,14946,14950,14992,15114,15135,15217,15226,15332,15389,15414,15439,15588,15606,15623,15666,15737,15844,15922,15951,15953,15982,16151,16169,16174,16246,16248,16306,16316,16417,16439,16536,16651,16681,16720,16793,16956,17271,17317,17323,17339,17465,17469,17574,17610,17659,17704,17783,17800,17898,17956,18008,18045,18196,18234,18298,18375,18589,18653,18780,18798,18852,18978,18983,18984,19002,19019,19190,19243,19338,19344,19407,19483,19540,19550,19703,19718,19774,19824,19853,19856,19876,19884,20034,20090,20160,20233,20351,20481,20547,20551,20610,20646,20725,20794,20834,20944,20999,21021,21094,21132,21155,21177,21208,21209,21210,21256,21322,21328,21374,21455,21500,21594,21875,21879,21899,22072,22131,22133,22192,22213,22237,22299,22340,22345,22374,22380,22448,22680,22760,22836,22859,22869,22938,22988,23009,23246,23302,23348,23357,23423,23534,23549,23646,23648,23704,23766,23777,23781,24108,24192,24226,24264,24416,24440,24475,24551,24581,24593,24630,24656,24685,24700,24767,24797,24916,24929,24930,24964,25043,25044,25091,25165,25168,25200,25232,25244,25264,25294,25399,25429,25435,25484,25541,25612,25733,25747,25759,25878,25888,25914,25921,25927,25937,25953,26006,26020,26049,26058,26089,26152,26170,26190,26230,26316,26322,26337,26368,26380,26407,26413,26415,26498,26508,26526,26551,26573,26672,26676,26726,26776,26786,26854,26865,26890,26908,26911,26949,26960,26968,26969,26984,27080,27100,27179,27212,27280,27319,27322,27365,27368,27415,27461,27476,27483,27509,27524,27598,27618,27630,27721,27742,27780,27789,27825,27886,27895,27999,28019,28027,28062,28069,28115,28121,28212,28301,28320,28342,28357,28689,28775,28903,29000,29027,29082,29097,29102,29109,29219,29279,29299,29314,29318,29320,29349,29406,29418,29452,29504,29687,29707,29719,29853,29882,29989,30013,30033,30037,30091,30119,30167,30225,30244,30321,30373,30387,30394,30400,30429,30483,30568,30632,30713,30754,30797,30838,31004,31008,31015,31022,31068,31088,31110,31161,31183,31287,31403,31442,31537,31564,31582,31692,31725,31726,31769,31815,31919,32084,32140,32190,32228,32367,32456,32507,32522,32529,32689,32873,32935,32987,33003,33091,33119,33136,33183,33277,33294,33321,33442,33505,33555,33583,33632,33682,33719,33733,33828,33841,33842,33859,33906,33981,34000,34016,34053,34067,34080,34142,34153,34171,34173,34297,34329,34355,34451,34469,34529,34561,34596,34608,34653,34654,34660,34694,34703,34715,34727,34795,34925,34985,35041,35076,35084,35127,35189,35206,35253,35272,35290,35382,35421,35468,35499,35529,35663,35832,35860,35913,36043,36057,36059,36085,36137,36174,36180,36185,36200,36231,36248,36310,36315,36325,36366,36373,36436,36487,36578,36596,36657,36673,36677,36942,36950,36963,36968,36970,36994,36995,37061,37111,37152,37174,37273,37294,37309,37338,37347,37373,37394,37469,37486,37572,37600,37687,37792,37797,37808,37880,37986,38001,38040,38048,38055,38072,38102,38104,38121,38165,38200,38219,38221,38302,38319,38328,38329,38445,38462,38472,38545,38564,38569,38658,38758,38804,38890,38995,39044,39052,39142,39144,39158,39185,39187,39188,39254,39289,39317,39354,39456,39473,39500,39521,39557,39609,39616,39638,39649,39679,39746,39792,39815,39871,39931,39935,39951,39993,40034,40064,40091,40093,40151,40623,40804,40838,40922,40927,40943,41071,41109,41162,41212,41367,41383,41393,41402,41406,41408,41451,41595,41861,41868,41936,41984,42067,42078,42093,42114,42126,42130,42166,42252,42330,42421,42451,42515,42555,42576,42583,42638,42663,42691,42792,42840,42846,42971,42986,43028,43065,43109,43192,43254,43572,43577,43589,43610,43621,43652,43661,43833,43951,44108,44137,44223,44315,44357,44386,44399,44417,44431,44562,44569,44579,44662,44711,44761,44869,44908,44930,44986,45085,45174,45247,45264,45298,45310,45311,45340,45350,45490,45491,45495,45529,45535,45621,45684,45687,45739,45750,45824,45837,45917,45963,45978,45993,46002,46003,46082,46125,46127,46245,46353,46402,46452,46461,46579,46668,46687,46721,46763,46803,46807,46850,46852,46864,46895,46964,46984,46991,47210,47307,47321,47439,47451,47505,47632,47739,47752,47784,47800,47909,47949,48081,48087,48125,48144,48225,48236,48250,48305,48311,48326,48352,48367,48379,48449,48481,48499,48908,48946,49053,49084,49121,49346,49384,49385,49457,49565,49681,49753,49794,49964,50035,50059,50060,50066,50086,50262,50267,50428,50631,50689,50720,50734,50833,50885,50894,50936,51080,51114,51270,51363,51442,51483,51572,51613,51615,51685,51699,51805,51922,51943,52007,52084,52086,52175,52199,52262,52327,52332,52351,52376,52467,52492,52667,52717,52927,53192,53244,53252,53285,53358,53435,53479,53530,53616,53647,53758,53980,54071,54163,54313,54430,54434,54437,54713,54847,55062,55073,55109,55134,55137,55241,55298,55338,55394,55402,55426,55431,55556,55614,55720,56059,56165,56203,56221,56383,56449,56789,57051,57178,57179,57185,57194,57317,57328,57479,57482,57486,57516,57541,57612,57733,57803,57892,58049,58102,58229,58340,58397,58418,58444,58527,58531,58669,58758,58830,58843,58844,58889,58964,58967,58983,58996,59107,59274,59387,59401,59419,59538,59641,59708,59724,59749,59773,59805,59878,59884,59952,59988,60153,60557,60840,60965,61045,61056,61137,61144,61199,61265,61341,61358,61385,61470,61488,61492,61655,61739,61780,61807,62027,62078,62110,62144,62228,62266,62351,62353,62355,62360,62871,63076,63128,63151,63279,63405,63443,63451,63516,63703,63798,63811,63839,63843,63949,64017,64032,64039,64239,64363,64365,64379,64442,64467,64490,64708,64780,64781,64840,64848,64864,64899,64918,64938,65053,65163,65173,65382,65526,65547,65596,65791,65805,65832,65853,65930,65949,66122,66125,66176,66263,66522,66598,66608,66672,66772,66844,66882,66969,67002,67024,67032,67088,67206,67209,67399,67404,67464,67665,67738,67767,67891,67950,67953,68019,68037,68091,68107,68411,68417,68629,68693,68783,68826,68829,68887,68945,68950,69159,69233,69316,69420,69487,69594,69625,69696,69767,69773,69810,69897,69899,69935,69936,69991,70000,70058,70073,70100,70104,70111,70242,70466,70475,70480,70608,70613,70647,70665,70732,70789,70878,70987,71048,71167,71168,71275,71341,71358,71382,71447,71460,71465,71562,71715,71797,71961,72162,72189,72262,72396,72408,72504,72553,72614,72721,72728,72891,73074,73113,73385,73470,73488,73578,73686,73831,73922,73959,74166,74207,74265,74302,74360,74400,74515,74737,74766,74809,74953,74971,74973,74982,74989,75251,75273,75327,75447,75502,75519,75578,75647,75835,75861,75892,75928,75960,76241,76346,76350,76398,76446,76492,76646,76721,76845,76896,76958,76986,77042,77047,77295,77390,77497,77648,77658,77905,77917,77992,78131,78154,78276,78341,78421,78429,78570,78682,78725,78747,78748,78785,78829,78885,78908,78944,78997,79106,79133,79224,79391,79410,79494,79533,79634,79673,79769,79908,79949,80033,80277,80278,80356,80372,80426,80465,80499,80622,80623,80662,80683,80689,80811,80952,80971,81162,81172,81287,81304,81465,81530,81559,81603,81682,81701,82010,82160,82186,82197,82263,82291,82315,82378,82531,82560,82586,82680,82784,82838,82885,82983,83248,83259,83458,83517,83589,83698,83778,83802,83856,84124,84151,84266,84409,84499,84514,84604,84626,84699,84772,84809,84899,84988,84994,85030,85104,85206,85241,85337,85352,85425,85602,85690,85718,85748,85755,85762,85908,85965,85988,86036,86046,86108,86193,86214,86282,86289,86352,86436,86467,86488,86550,86615,86660,86685,86747,86780,86849,86995,87106,87139,87201,87203,87206,87350,87430,87435,87458,87471,87701,87725,87735,87786,87899,88034,88056,88094,88096,88124,88412,88456,88556,88558,88621,88637,88705,88746,88843,88899,88927,88989,89185,89187,89213,89443,89473,89479,89495,89531,89535,89580,89582,89675,89689,89690,89756,89842,90018,90042,90166,90172,90175,90260,90283,90294,90390,90449,90605,90639,90720,90754,90762,90819,90856,90947,90951,90993,91056,91059,91064,91093,91103,91154,91160,91245,91320,91398,91404,91482,91570,91657,91795,91862,91883,91960,92087,92146,92166,92310,92415,92471,92515,92694,93093,93165,93207,93262,93326,93328,93425,93539,93595,93676,93683,93894,93995,94348,94349,94416,94509,94515,94521,94538,94636,94836,94969,94976,95039,95100,95114,95136,95289,95294,95323,95341,95412,95417,95449,95545,95586,95687,95733,95734,95769,95846,95939,95951,95955,96342,96374,96396,96397,96490,96696,96786,97142,97226,97402,97464,97521,97546,97720,98057,98139,98197,98207,98354,98470,98634,98690,98703,98708,98858,98892,98946,99089,99091,99210,99225,99230,99261,99357,99619,99631,99646,99704,99882,99899,100000",
          "4,14,142,250,307,325,468,499,502,552,593,637,995,1430,1494,1925,1949,2042,2122,2158,2202,2247,2326,2422,2433,2524,2648,2707,2724,2863,3026,3050,3094,3118,3274,3362,3501,3728,3766,3811,3911,3961,4029,4174,4278,4284,4326,4390,4427,4497,4525,4604,4617,4791,4988,5040,5058,5140,5172,5194,5340,5349,5443,5469,5670,5782,5867,5985,6189,6201,6257,6381,6402,6433,6546,6569,6593,6638,6805,6836,6865,6927,6973,6982,6995,7084,7135,7453,7480,7493,7598,7636,7652,7748,7846,7904,8019,8311,8321,8379,8545,8750,8786,8883,9010,9042,9286,9365,9484,9494,9587,9612,9868,10021,10281,10297,10400,10497,10540,10631,10721,10767,10773,10831,10917,11119,11146,11148,11212,11345,11451,11552,11575,11696,11750,11781,11818,12194,12251,12282,12379,12770,12818,12870,12887,12928,12955,12992,13138,13166,13263,13449,13460,13478,13516,13536,13623,13651,13706,13743,13834,13842,13861,13881,14005,14014,14017,14027,14029,14036,14164,14209,14242,14262,14294,14346,14385,14513,14564,14598,14664,14736,14795,14814,14991,15064,15125,15154,15171,15190,15244,15305,15364,15452,15647,15767,15776,15861,15924,16208,16224,16673,16677,16699,16897,17114,17117,17251,17345,17363,17384,17421,17617,17774,17816,17821,17825,17932,17951,18036,18279,18286,18370,18471,18805,18830,19016,19151,19167,19169,19195,19275,19320,19349,19463,19530,19642,19673,19681,19807,19871,19916,19972,20033,20144,20192,20199,20241,20296,20584,20593,20631,20648,20652,20664,20722,20749,20766,20805,21051,21054,21442,21549,21562,21728,22165,22177,22210,22267,22329,22400,22468,22475,22605,22746,22757,22795,22796,22866,22883,22909,22956,22969,22994,23223,23400,23480,23510,23698,23733,23787,23791,23813,24137,24184,24385,24432,24515,24627,24669,24808,24854,24999,25365,25394,25437,25453,25566,25575,25604,25624,25688,25758,25768,25772,25917,25994,26071,26101,26118,26314,26401,26432,26728,26761,26780,26910,26963,27055,27084,27137,27229,27526,27608,27702,27764,27776,27777,27792,27802,27829,28040,28249,28329,28345,28405,28539,28616,28641,28745,28789,28880,28938,28968,28991,29083,29208,29251,29276,29340,29366,29396,29519,29549,29699,29727,29807,30084,30116,30339,30368,30386,30396,30501,30727,30796,31086,31168,31280,31348,31361,31390,31398,31500,31539,31594,31830,32071,32147,32235,32253,32353,32404,32463,32521,32527,32536,32619,32638,32654,32782,32795,32846,32863,32898,32900,32931,33037,33256,33493,33520,33522,33751,33810,33816,33885,33900,34062,34279,34515,34676,34701,34779,34988,35000,35070,35185,35238,35252,35396,35447,35557,35807,35824,35874,35925,36020,36151,36431,36482,36558,36629,36812,36916,37096,37113,37227,37364,37456,37535,37554,37663,37732,37825,37832,37944,37954,38095,38105,38209,38225,38459,38519,38529,38882,38960,39167,39396,39415,39527,39540,39769,39960,39980,39984,39999,40000,40167,40258,40296,40470,40590,40601,40620,40752,40818,40829,40844,40877,40935,41183,41195,41416,41473,41480,41555,41608,41774,41812,41919,41982,42035,42154,42178,42229,42235,42248,42286,42292,42342,42532,42657,42664,42729,42782,42797,42863,42904,42917,42998,43002,43076,43299,43408,43459,43470,43494,43695,43727,43890,43965,44080,44084,44196,44474,44661,44690,44714,44759,44786,44807,44924,44972,45008,45185,45250,45258,45551,45695,45884,45925,45965,46017,46021,46105,46192,46239,46250,46365,46434,46582,46612,46642,46654,46751,46934,46992,47113,47114,47126,47209,47213,47286,47315,47461,47662,47742,47811,47895,47978,48092,48622,48626,48748,48786,48870,48930,48971,49182,49226,49244,49312,49371,49435,49447,49469,49506,49553,49579,49702,49735,49854,49861,50100,50153,50178,50206,50245,50294,50295,50336,50461,50515,50575,50604,50744,50845,50985,51151,51229,51272,51326,51487,51520,51552,51718,51787,51820,51843,51944,52114,52192,52380,52485,52504,52606,52793,52864,53111,53254,53264,53284,53518,53679,53806,53825,53961,54128,54136,54187,54361,54463,54483,54593,54669,54708,54762,54834,54884,55067,55246,55312,55533,55539,55675,55689,55715,56027,56148,56175,56195,56237,56460,56570,56644,56658,56808,56878,57131,57134,57402,57546,57592,57628,57649,57836,57921,57930,57959,57978,57997,58064,58091,58136,58276,58368,58737,58759,58788,58817,58820,58945,59189,59288,59294,59371,59389,59547,59624,59651,59795,59928,60177,60239,60263,60312,60636,60809,60814,60875,61319,61354,61357,61613,61678,61777,61803,61824,61862,61900,61922,61959,61975,62023,62067,62114,62212,62245,62251,62410,62528,62558,62578,62647,62994,63020,63106,63113,63157,63248,63392,63546,63607,63724,63785,63795,63845,63916,63917,64030,64117,64411,64489,64505,64513,64871,64989,64996,65010,65117,65315,65317,65330,65514,65574,65750,65819,65846,65955,66461,66548,66624,66776,66792,66849,66899,67098,67369,67435,67559,67597,67788,67810,67963,68040,68049,68067,68090,68094,68164,68461,68474,68553,68602,68624,68634,68641,68665,68678,68724,68893,68897,68962,69321,69381,69418,69461,69475,69600,69637,69640,69662,69676,69688,69920,69930,69975,70202,70258,70308,70438,70481,70790,70869,70870,71071,71179,71261,71292,71517,71539,71543,71960,72021,72044,72094,72646,72845,72890,72914,72942,73057,73152,73207,73242,73516,73656,73661,73720,73802,73834,73887,74098,74281,74306,74507,74527,74532,74634,74643,74650,74701,74713,74824,74831,74992,75055,75087,75198,75346,75351,75375,75520,75855,75916,76095,76096,76121,76197,76324,76385,76412,76472,76973,76990,76997,77255,77304,77860,77984,78199,78233,78235,78470,78641,78744,78761,78935,79077,79144,79404,79430,79566,79778,79805,79826,80090,80095,80136,80181,80405,80456,80679,80766,80770,80841,80892,80941,80958,81011,81041,81043,81070,81141,81190,81345,81347,81429,81470,81487,81537,81542,81552,81571,81649,81702,81766,81774,81847,81854,81955,82028,82052,82058,82076,82122,82200,82206,82235,82245,82491,82580,82687,82741,82760,82967,83202,83303,83324,83337,83472,83525,83549,83721,83738,83748,83769,83898,83942,83988,84137,84150,84236,84331,84359,84402,84463,84616,84675,84681,84742,84794,84922,84938,84940,84942,84947,84956,84965,85162,85164,85507,85546,85579,85588,85612,85726,85785,85790,85832,85887,85946,86375,86472,86480,86555,86608,86754,86810,86837,86856,86954,86982,87394,87489,87579,87697,87745,87751,87774,87819,88083,88132,88170,88298,88402,88485,88488,88694,88765,88972,89076,89302,89358,89365,89406,89418,89634,89720,89796,89885,90004,90164,90338,90375,90419,90552,90560,90749,90755,90792,90858,90876,90950,91009,91128,91167,91258,91357,91371,91489,91927,92119,92374,92526,92594,92665,92723,92786,92861,93076,93121,93220,93221,93300,93318,93362,93409,93483,93493,93642,93727,93829,93848,93878,93895,94042,94164,94211,94281,94287,94334,94405,94421,94431,94661,94697,94714,94823,95074,95154,95239,95276,95288,95353,95359,95531,95538,95550,95583,95627,95668,95782,95821,96037,96096,96241,96244,96457,96465,96544,96610,96659,96738,96776,96841,96867,97033,97181,97249,97286,97371,97383,97450,97502,97598,97626,97631,97672,97685,97846,97873,97957,97970,98045,98105,98135,98145,98186,98191,98312,98564,98663,98715,98797,98907,98937,99102,99211,99815,99872,99920,99961",
          "23,146,196,222,234,318,474,509,543,651,728,775,965,982,1049,1068,1095,1501,1503,1541,1628,1653,1704,1754,1941,2011,2071,2156,2239,2271,2303,2304,2307,2546,2698,2727,2793,2924,3008,3114,3218,3391,3441,3492,3532,3602,3686,3800,3902,3966,4007,4069,4148,4391,4452,4462,4522,4681,4744,4745,4786,4805,4977,5105,5123,5146,5174,5199,5208,5238,5450,5508,5827,6044,6747,6798,6857,6905,6999,7017,7104,7125,7148,7226,7352,7526,7642,7695,7819,7830,7918,7981,7996,8107,8240,8363,8385,8425,8488,8602,8678,8697,8700,8771,8920,8971,8981,9567,9651,9659,9663,9694,9737,9747,9769,9947,10000,10058,10087,10112,10162,10171,10185,10205,10227,10316,10352,10445,10644,10670,10900,10922,11127,11208,11266,11273,11277,11283,11302,11382,11534,11610,11693,11731,11770,11799,11806,11808,11860,11981,12070,12073,12088,12119,12152,12259,12341,12380,12443,12606,12665,12767,12784,12909,12954,13062,13085,13086,13359,13370,13481,13497,13679,13766,13820,13831,13867,13869,13882,13958,14046,14063,14112,14134,14141,14153,14178,14219,14225,14280,14288,14328,14375,14379,14447,14567,14602,14624,14653,14689,14703,14723,14746,14769,14781,14893,14993,15007,15059,15066,15086,15121,15194,15300,15342,15365,15402,15453,15474,15522,15544,15714,15717,15783,15843,15846,15930,15965,15997,16049,16082,16113,16129,16155,16211,16297,16334,16395,16440,16511,16535,16564,16616,16674,16701,16738,16818,16853,16856,16894,17017,17115,17247,17264,17447,17481,17516,17577,17701,17710,17925,17937,17938,17972,18006,18082,18112,18365,18403,18423,18638,18811,18813,18820,18822,18859,18878,18904,18905,18997,19087,19175,19215,19242,19290,19420,19440,19487,19488,19496,19560,19603,19644,19710,19711,19792,19829,19842,19974,19993,19996,20010,20028,20180,20228,20256,20283,20288,20339,20373,20404,20464,20675,20713,20772,20799,20844,20896,20931,21020,21136,21167,21187,21278,21300,21323,21341,21598,22016,22062,22105,22134,22211,22369,22399,22433,22485,22509,22614,22661,22708,22711,22783,22845,22887,23018,23076,23114,23128,23172,23175,23205,23258,23290,23375,23455,23460,23572,23705,23790,23799,23829,23860,23882,23938,23948,24016,24039,24160,24213,24239,24476,24478,24508,24540,24624,24657,24703,24835,24872,24980,24981,25009,25083,25088,25097,25108,25199,25211,25328,25330,25436,25438,25451,25544,25597,25618,25620,25654,25714,25749,25756,25774,25839,25850,25957,26293,26334,26384,26393,26410,26412,26558,26693,26805,26873,26901,27000,27004,27040,27214,27269,27332,27430,27446,27462,27468,27515,27531,27553,27932,28127,28197,28304,28327,28354,28360,28481,28572,28647,28657,28681,28779,28906,28909,28956,29041,29043,29117,29210,29269,29468,29571,29665,29772,29774,29868,29926,30140,30223,30228,30390,30395,30422,30437,30479,30490,30629,30645,30647,30735,30739,30784,30813,30814,30827,30872,30895,30981,30989,31066,31075,31160,31180,31316,31333,31388,31436,31466,31600,31603,31740,31886,31953,31991,31999,32060,32099,32208,32292,32294,32459,32590,32598,32621,32758,32835,32848,32866,32890,32899,32933,32951,33006,33069,33209,33213,33378,33390,33415,33558,33618,33698,33792,33873,33942,33967,33994,34056,34079,34104,34120,34245,34247,34358,34491,34525,34552,34652,34662,34665,34811,34962,35089,35117,35126,35247,35310,35379,35440,35508,35553,35665,35709,36082,36083,36092,36179,36188,36246,36384,36459,36579,36606,36611,36674,36705,36728,36739,36915,37098,37169,37197,37287,37332,37337,37391,37508,37553,37708,37713,37765,37783,37827,37834,37856,37895,38115,38185,38214,38327,38434,38499,38500,38750,38763,38859,38933,39054,39058,39100,39103,39131,39305,39340,39358,39697,39726,39745,39756,39810,39861,39964,40059,40063,40088,40094,40142,40333,40350,40365,40373,40469,40529,40556,40564,40647,40679,41015,41084,41134,41138,41218,41272,41276,41280,41329,41392,41417,41502,41652,41688,41718,41754,41882,41911,41959,41965,42059,42072,42253,42389,42399,42414,42597,42628,42653,42745,42759,42766,42886,42913,43056,43106,43298,43326,43335,43356,43403,43431,43569,43677,43696,43705,43759,43765,43770,43838,43895,43931,44293,44302,44400,44470,44638,44694,44817,44884,44948,45073,45148,45231,45246,45285,45336,45344,45442,45504,45514,45582,45615,45699,45764,45946,46094,46110,46184,46205,46262,46300,46429,46466,46476,46491,46551,46607,46816,46867,46868,46897,46922,46943,47016,47283,47327,47355,47391,47445,47475,47501,47554,47591,47599,47647,47689,47690,47763,47959,48070,48090,48192,48230,48241,48271,48453,48507,48602,48679,48894,48928,48983,49183,49237,49335,49380,49393,49603,49772,49856,49867,49896,49914,49989,50010,50085,50122,50142,50226,50271,50334,50424,50632,50653,50759,50790,50810,50815,50899,51082,51171,51205,51241,51325,51331,51430,51489,51490,51532,51642,51723,51954,51964,51985,52000,52147,52320,52353,52471,52490,52597,52611,52705,52769,52835,52918,52967,53050,53110,53346,53352,53384,53429,53462,53481,53558,53567,53587,53626,53629,53821,53865,53979,54041,54165,54180,54451,54805,54980,55054,55059,55113,55193,55201,55213,55257,55297,55405,55424,55505,55606,55654,55696,55719,55789,55804,55988,56030,56098,56184,56234,56269,56357,56551,56622,56700,56708,56776,56907,56921,56946,56963,57025,57057,57076,57111,57144,57205,57353,57420,57439,57577,57744,57761,57772,57847,57874,57980,58022,58061,58356,58394,58409,58452,58564,58567,58577,58596,58598,58678,58756,58813,58825,58870,58886,58891,59053,59085,59158,59162,59179,59205,59258,59285,59289,59324,59454,59520,59550,59565,59628,59771,59802,59849,59861,59877,59890,59892,59939,59961,59972,59981,59982,60003,60070,60163,60191,60279,60388,60429,60430,60450,60528,60746,60777,60825,61151,61158,61169,61197,61227,61248,61268,61314,61480,61529,61544,61586,61617,61848,62086,62106,62208,62336,62655,62881,63011,63180,63209,63234,63236,63465,63598,63614,63972,63987,64184,64198,64265,64380,64543,64575,64641,64696,64772,64807,64856,64930,64988,65045,65048,65155,65250,65373,65388,65394,65406,65474,65513,65683,65783,65893,65950,65991,66023,66131,66387,66495,66513,66537,66584,66646,66656,66883,67216,67221,67274,67437,67588,67666,67706,67804,67964,68008,68017,68242,68265,68492,68636,68699,68810,69081,69143,69208,69258,69384,69425,69441,69499,69500,69521,69666,69725,69987,69995,70215,70245,70260,70398,70450,70497,70575,70593,70623,70705,70872,70879,70917,70997,71036,71226,71323,71424,71665,71697,71767,71951,72006,72120,72244,72377,72393,72623,72628,72634,72711,72744,72765,72920,73032,73278,73395,73467,73605,73628,73639,73683,73692,73748,73884,74055,74132,74140,74165,74243,74489,74554,74616,74928,74952,74970,75042,75111,75187,75224,75227,75266,75423,75570,75597,75604,75659,75662,75711,75757,75920,75930,75954,76077,76107,76156,76232,76236,76277,76339,76356,76413,76421,76596,76652,76793,76842,76850,76921,76930,77131,77310,77378,77396,77427,77446,77532,77562,77653,77710,77809,77819,77960,77976,78063,78068,78228,78313,78417,78489,78564,78584,78822,78861,78943,78982,79269,79301,79312,79360,79375,79411,79477,79545,79682,79709,79807,79832,79862,79910,79927,79983,80146,80151,80199,80287,80370,80384,80419,80467,80494,80570,80601,80853,80901,80922,80970,80998,81053,81069,81089,81194,81224,81283,81478,81554,81586,81634,81709,81739,81765,81811,81834,81961,82027,82131,82152,82153,82224,82247,82340,82416,82570,82673,82695,82776,82883,82888,82981,83081,83166,83252,83258,83264,83297,83434,83604,83645,83653,83725,84142,84192,84245,84252,84277,84412,84456,84478,84511,84515,84571,84577,84684,84710,84731,84776,84815,84821,84884,84948,84954,84985,85002,85031,85124,85282,85283,85422,85527,85581,85585,85638,85641,85735,85739,85797,85831,85841,85993,86013,86050,86428,86430,86578,86585,86707,86755,86943,86949,86960,87001,87031,87032,87197,87271,87318,87328,87366,87370,87445,87527,87542,87590,87648,87652,87679,87778,87866,87912,88355,88372,88477,88486,88616,88762,88770,88842,88864,88883,89068,89148,89225,89320,89868,90045,90089,90161,90238,90312,90389,90475,90512,90559,90619,90621,90696,90778,90797,90818,90866,90887,90907,91153,91317,91422,91445,91463,91528,91641,91740,91748,91827,91838,92116,92141,92184,92344,92410,92460,92461,92487,92539,92613,92698,92764,92802,92829,92859,92898,92949,93039,93066,93077,93200,93276,93288,93524,93538,93576,93583,93616,93641,93713,93759,93825,93845,93984,94058,94223,94317,94359,94444,94672,94795,94851,95069,95073,95139,95172,95194,95380,95426,95752,95781,95942,95975,96119,96181,96236,96278,96354,96409,96411,96478,96684,96701,96749,96774,96799,97025,97114,97268,97280,97311,97403,97485,97630,97639,97851,97856,97927,97965,98004,98019,98051,98118,98208,98403,98465,98501,98506,98544,98572,98597,98868,98909,98956,98985,98994,99054,99088,99150,99250,99310,99337,99433,99444,99481,99734,99749,99806,99845,99944",
          "27,42,79,189,287,566,705,739,916,945,978,1061,1170,1214,1219,1377,1593,1673,1920,2001,2186,2278,2286,2318,2370,2376,2440,2468,2477,2563,2824,2999,3105,3184,3189,3351,3417,3506,3761,3774,3887,3890,3951,4039,4110,4185,4201,4247,4248,4352,4455,4508,4511,4582,4589,5125,5243,5247,5350,5351,5406,5432,5531,5532,5547,5549,5607,5610,5758,5868,5886,6009,6232,6284,6310,6405,6421,6501,6557,6667,6787,6812,6817,6861,6969,7003,7053,7121,7132,7166,7185,7196,7444,7479,7525,7922,7927,8099,8131,8156,8185,8266,8424,8641,8754,8873,8926,8965,9039,9077,9080,9117,9151,9183,9240,9519,9540,9626,9996,10031,10132,10288,10343,10372,10658,10782,10904,10955,11005,11047,11098,11158,11301,11353,11460,11796,11819,11931,11998,12003,12024,12079,12089,12102,12172,12287,12301,12404,12500,12624,12658,12788,12922,12995,13059,13133,13216,13258,13354,13438,13528,13649,13720,13830,13856,13870,14033,14207,14211,14298,14537,14708,14869,14917,14959,15014,15119,15172,15447,15465,15523,15528,15547,15683,15746,15760,15937,15991,16160,16277,16284,16367,16514,16560,16587,16635,16683,16978,17070,17097,17149,17177,17183,17192,17198,17273,17330,17364,17507,17643,17719,17776,17860,18101,18224,18342,18415,18445,18715,18745,18770,18790,18843,18948,18987,18991,19097,19130,19205,19401,19579,19618,19678,19685,19791,19902,19929,20075,20220,20230,20292,20427,20533,20572,20588,20730,20786,20802,20836,20841,20971,21085,21316,21346,21498,21613,21721,21765,21925,22048,22053,22092,22201,22279,22315,22320,22383,22429,22619,22640,22683,22690,22765,22864,23019,23077,23308,23481,23612,23662,23673,23681,23773,23783,23838,23980,24035,24080,24283,24292,24296,24434,24467,24568,24631,24793,24868,24963,24966,25127,25381,25558,25662,25867,25922,26047,26082,26180,26222,26232,26236,26310,26446,26591,26799,26814,26863,26880,26997,27071,27377,27422,27433,27467,27557,27561,27579,27672,27693,27767,27899,27998,28006,28083,28366,28433,28486,28665,28833,28875,28879,28987,29019,29039,29227,29343,29373,29574,29625,29642,29876,30034,30202,30382,30463,30471,30533,30575,30624,30938,30979,31002,31140,31151,31171,31179,31205,31409,31451,31529,31752,31969,31992,32017,32069,32333,32447,32464,32537,32569,32584,32617,32772,32787,32888,32945,33045,33072,33107,33275,33403,33573,33772,33804,33868,33889,34036,34055,34363,34380,34392,34536,34598,34629,34759,34800,34838,34918,34961,35051,35062,35074,35150,35214,35235,35254,35297,35449,35736,35791,35792,35798,35818,35849,35969,36078,36091,36095,36176,36191,36212,36223,36280,36345,36394,36430,36563,36565,36625,36727,36848,36878,37027,37092,37279,37401,37590,37736,37775,37791,37854,37869,37989,38027,38216,38312,38357,38534,38542,38587,38677,38903,38914,38996,39123,39153,39752,39857,39968,39978,40084,40240,40395,40644,40708,40739,40771,40968,41107,41159,41179,41349,41443,41479,41547,41563,41573,41575,41654,41846,42064,42128,42176,42401,42584,42682,42735,42862,42914,42962,43040,43128,43145,43208,43232,43309,43322,43548,43562,43658,43666,43667,43719,43781,43953,44004,44012,44028,44134,44160,44161,44211,44250,44299,44398,44469,44499,44591,44700,44811,44823,44937,44938,45011,45034,45169,45317,45353,45355,45377,45382,45390,45427,45635,45660,45712,45816,45903,45904,45986,46076,46187,46197,46750,46792,46882,47145,47164,47221,47310,47337,47480,47486,47503,47534,47582,47590,47747,47862,47983,48043,48126,48171,48181,48211,48217,48278,48292,48537,48547,48599,48619,48682,48704,48750,48813,48895,49060,49142,49578,49590,49629,49647,49665,49708,49728,49758,49830,49834,49873,49887,50135,50196,50199,50315,50361,50556,50698,50882,50981,51029,51170,51244,51472,51476,51511,51537,51546,51620,51672,51759,52092,52107,52186,52204,52309,52407,52484,52494,52576,52605,52642,52861,52893,52933,52953,53086,53098,53135,53140,53315,53461,53476,53516,53590,53630,53702,53834,53837,54073,54171,54200,54209,54243,54445,54534,54732,54775,54829,54886,54937,54997,55092,55114,55203,55299,55309,55564,55597,55734,55770,55778,55812,55853,55916,55952,55961,56123,56128,56130,56172,56181,56212,56231,56267,56302,56511,56735,56888,56939,56997,57292,57338,57412,57477,57500,57613,57701,57904,57924,58112,58150,58191,58261,58416,58503,58563,58809,58856,58924,59070,59394,59480,59786,59787,59790,59863,59872,59949,60085,60199,60362,60774,60806,60863,61050,61127,61130,61224,61262,61274,61360,61396,61397,61462,61534,61796,61846,61940,61972,61973,62015,62019,62066,62136,62281,62342,62484,62519,62759,62884,62900,62905,63049,63121,63272,63365,63395,63406,63476,63709,63717,63781,63813,63832,63842,63878,63980,64003,64005,64054,64328,64383,64396,64410,64716,64757,64975,65017,65148,65175,65182,65229,65259,65327,65356,65416,65487,65537,65622,65633,65685,65749,65804,65908,66019,66066,66163,66254,66414,66445,66527,66543,66632,66675,66696,66735,66818,66915,66916,66919,67076,67112,67264,67281,67291,67465,67519,67542,67594,67616,67636,67684,67816,67853,67924,67978,67988,68245,68253,68430,68673,68738,68748,68777,68786,68891,68939,69000,69006,69167,69234,69469,69577,69836,69860,70134,70193,70246,70540,70621,70656,70776,70935,70956,70961,71053,71298,71356,71363,71394,71411,71526,71528,71530,71549,71572,71618,71787,71883,71912,71997,72038,72046,72284,72477,72579,72629,72712,72760,72815,72852,73062,73082,73302,73709,73718,73735,73745,73807,73865,73899,74108,74283,74331,74339,74380,74478,74491,74578,74623,74825,74835,74854,74859,74908,75067,75090,75128,75175,75183,75220,75248,75253,75482,75569,75731,76016,76024,76060,76093,76255,76307,76590,76611,76665,76703,76741,76799,76808,76843,77002,77006,77149,77207,77267,77308,77754,78055,78210,78251,78413,78535,78638,78710,78847,78887,78913,79000,79191,79214,79310,79361,79364,79413,79433,79663,79763,79822,79876,80037,80079,80193,80198,80243,80311,80345,80358,80481,80501,80673,80712,80729,80837,80925,80991,81094,81362,81408,81592,81687,81756,81770,81778,81791,81923,81926,82096,82409,82530,82644,82756,82781,82825,82829,82941,82957,82979,83219,83234,83236,83247,83339,83363,83385,83516,83558,83801,83878,84089,84110,84188,84201,84259,84389,84401,84434,84452,84507,84567,84608,84640,84662,84787,84888,84924,84925,85081,85135,85176,85194,85218,85239,85240,85310,85413,85483,85551,85578,85647,85681,85757,85783,85847,85972,86016,86305,86328,86330,86369,86523,86532,86549,86725,86738,86801,86881,87098,87296,87301,87338,87384,87453,87518,87536,87598,87699,87836,87915,88179,88181,88321,88503,88518,88551,88706,88743,88955,89003,89084,89103,89120,89145,89192,89211,89395,89568,89716,89763,89807,89849,89856,90012,90048,90057,90068,90144,90221,90401,90536,90542,90543,90606,90669,90800,90889,90926,91011,91021,91032,91152,91207,91375,91420,91480,91487,91537,91563,91619,91667,91685,91912,92157,92173,92218,92249,92260,92267,92453,92939,92996,93142,93245,93347,93633,93659,93725,93795,93906,94035,94138,94301,94530,94531,94580,94674,94722,95082,95120,95237,95254,95301,95482,95523,95669,95692,95731,95762,95805,95947,95960,96185,96193,96238,96326,96385,96388,96425,96468,96486,96500,96538,96740,96814,96902,97016,97234,97344,97350,97366,97509,97676,97699,97714,97726,97801,97968,98072,98189,98282,98328,98475,98678,98729,98749,98753,98779,98794,99002,99009,99084,99141,99175,99268,99274,99283,99598,99656,99715,99867,99958",
          "15,257,273,290,299,329,379,398,431,444,446,449,459,518,537,649,718,748,792,794,1023,1098,1317,1330,1433,3926,5006,5031,5042,5119,5166,5177,5190,5216,5231,5263,5281,5298,5302,5348,5365,5367,5383,5411,5447,5482,5489,5528,5774,5871,5924,5932,6000,6024,6069,6076,6361,6366,6409,6419,6538,6627,6689,6777,6804,6874,6922,7036,7160,7462,7591,7712,7818,7971,8034,8089,8138,8164,8297,8361,8663,8698,8821,9016,9046,9207,9213,9393,9422,9635,10026,10224,12132,12264,12617,12879,13592,15608,15712,16353,16501,17368,17376,18758,21686,22717,24718,26794,30476,32904,35694,50057,50306,50402,50407,50409,50485,50518,50735,50920,51036,51044,51093,51188,51369,51382,51439,51491,51521,51574,51742,51806,52191,52240,52279,52373,52394,52397,52429,52498,52557,52695,52707,52762,52767,52844,52845,52851,52909,52956,52981,53035,53062,53108,53123,53153,53217,53306,53428,53522,53546,53713,53730,53757,53761,53856,53869,53878,53896,53899,54161,54201,54597,54617,54737,54763,54804,55024,55040,55049,55091,55125,55188,55205,55224,55253,55433,55446,55452,55604,55625,55748,55784,55843,55912,55948,56011,56090,56336,56497,56542,56596,56597,56656,56970,57045,57095,57104,57110,57523,57543,57573,57697,57784,57889,58003,58209,58303,58314,58319,58377,58526,58528,58617,58685,58746,58766,58791,58863,58933,59088,59201,59211,59232,59431,59464,59662,59707,59935,59948,59993,60109,60165,60206,60269,60320,60371,60377,60393,60401,60443,60495,60558,60618,60984,61067,61083,61104,61478,61561,61715,61726,61793,61822,62076,62119,62505,62975,63117,63216,63478,63521,63575,63669,63677,63690,63713,63726,63810,63831,63850,63889,63957,64046,64071,64073,64272,64292,64296,64368,64407,64544,64719,64754,64776,64915,64978,65165,65203,65206,65254,65289,65316,65337,65368,65434,65566,65618,65772,65799,65874,65933,66015,66026,66076,66094,66181,66191,66424,66458,66525,66557,66599,66769,66835,66942,67000,67156,67163,67275,67488,67538,67627,67783,67871,67882,67893,67904,67959,67976,68202,68362,68560,68563,68601,68659,68779,68924,68982,69066,69068,69238,69259,69284,69335,69348,69350,69421,69517,69542,69708,69732,69808,69918,69939,69951,70026,70044,70198,70225,70595,70771,70796,70814,70858,71014,71100,71220,71439,71533,71615,71804,71823,71863,72053,72061,72232,72407,72532,72533,72557,72611,72649,72752,72802,72929,72932,73050,73156,73162,73187,73219,73365,73390,73422,73492,73662,73697,74015,74382,74411,74476,74512,74545,74652,74677,74805,74846,74873,74884,75085,75150,75544,75586,75741,75760,75791,75901,75903,75924,76019,76057,76294,76338,76340,76529,76736,76838,76848,76868,76871,76943,76944,77171,77281,77311,77320,77325,77460,77478,77558,77605,77640,77676,77773,77840,77972,78021,78097,78345,78349,78372,78469,78491,78629,78657,78728,78890,78894,78984,79059,79143,79229,79268,79293,79356,79357,79393,79444,79579,79649,79762,79784,79859,79873,79881,79957,80051,80053,80072,80125,80375,80408,80437,80502,80605,80890,80912,81014,81018,81090,81118,81153,81199,81233,81621,81728,81747,81763,81792,81814,81862,81864,81888,82048,82063,82091,82353,82411,82458,82464,82501,82831,82929,82994,83162,83360,83505,83617,83820,84097,84317,84543,84646,84733,84775,84837,84854,84877,84878,84906,84929,85196,85208,85321,85368,85462,85536,85571,85577,85746,85905,86004,86037,86097,86236,86384,86426,86487,86524,86570,86724,86834,86931,87024,87123,87264,87325,87567,87637,87690,87757,87864,88148,88264,88279,88347,88376,88378,88407,88428,88474,88601,88777,88835,88944,88945,89013,89031,89122,89136,89636,89695,89738,89747,89824,89982,89996,90122,90278,90344,90572,90596,90628,90917,90945,91281,91308,91554,91586,91693,92021,92044,92109,92301,92321,92329,92659,92697,92706,92713,92750,92817,92858,93036,93099,93143,93267,93489,93499,93547,93608,93741,93803,93889,93973,94000,94313,94441,94507,94600,94633,94742,94777,94827,94838,95092,95169,95264,95305,95429,95444,95537,95661,95696,95880,95896,95937,96179,96299,96430,96446,96526,96562,96723,97018,97038,97040,97103,97172,97175,97214,97221,97321,97365,97386,97449,97754,97760,97813,97818,97917,97937,98300,98369,98540,98584,98594,98637,98787,98790,98949,99007,99106,99200,99248,99279,99490,99509,99600,99682,99820,99851,99854,99869,99955,99970,99988",
          "92,164,225,376,465,491,544,591,732,802,885,958,1100,1238,1319,1401,1418,1453,1485,1502,1504,1608,1622,1753,1848,1870,1917,2049,2091,2102,2105,2108,2123,2172,2361,2399,2467,2512,2543,2568,2601,2851,3016,3129,3161,3173,3280,3300,3304,3326,3328,3446,3490,3522,3764,3824,3912,3996,4010,4098,4158,4313,4338,4368,4415,4479,4529,4612,4621,4763,4796,5089,5091,5098,5113,5198,5270,5470,5540,5590,5725,5837,5958,6019,6042,6087,6224,6414,6604,6647,6681,6934,6936,6956,7171,7203,7219,7343,7390,7523,7747,7791,7813,7842,7924,7956,8038,8053,8119,8166,8167,8277,8355,8375,8447,8499,8515,8537,8650,8651,8758,9029,9034,9111,9147,9262,9285,9515,9532,9542,9611,9615,9647,9760,9834,9905,9977,10124,10179,10233,10286,10399,10451,10477,10538,10542,10576,10581,10584,10748,10841,10989,11023,11166,11315,11425,11426,11462,11540,11566,11573,11589,11599,11606,11655,11682,11691,11718,11792,11798,11844,11883,11896,11945,11950,12037,12039,12177,12296,12445,12567,12781,12924,12977,13048,13080,13088,13094,13107,13108,13330,13342,13442,13667,13697,13763,13923,13929,13968,14055,14074,14197,14218,14247,14274,14527,14688,14731,14802,14844,15133,15140,15146,15213,15256,15356,15440,15446,15541,15546,15570,15583,15673,15729,15792,15939,15955,15989,16006,16036,16111,16339,16428,16476,16548,16612,16714,16835,17077,17090,17128,17234,17290,17394,17418,17486,17504,17530,17607,17622,17629,17671,17676,17750,17757,17899,17919,18016,18168,18201,18230,18287,18393,18522,18524,18571,18672,18675,18701,18743,18821,18881,18895,18898,18931,18933,19044,19084,19183,19217,19285,19427,19434,19547,19566,19647,19684,19689,19705,19870,20066,20249,20317,20384,20539,20900,20908,20922,20946,21024,21047,21143,21175,21252,21412,21533,21544,21799,21842,22021,22232,22235,22387,22524,22549,22773,22825,22963,23188,23197,23224,23249,23351,23479,23498,23532,23587,23590,23752,23864,24042,24119,24163,24171,24194,24263,24315,24482,24544,24549,24560,24570,24702,24712,24734,24859,24915,25051,25333,25344,25366,25367,25375,25519,25640,25668,25669,25871,25882,26182,26185,26245,26267,26425,26598,26716,26771,26787,26823,26832,26842,26920,26933,26945,26982,27146,27186,27204,27206,27289,27304,27412,27527,27587,27615,27646,27708,27713,27750,27753,27796,27869,27912,27948,27997,28059,28064,28096,28102,28104,28157,28335,28391,28398,28455,28658,28763,28857,28905,28983,28999,29003,29023,29037,29058,29084,29145,29362,29403,29488,29517,29771,29891,29904,29955,29959,30022,30044,30131,30138,30213,30238,30266,30281,30435,30447,30465,30484,30503,30528,30583,30639,30653,30771,30788,30805,31191,31250,31251,31302,31357,31578,31584,31586,31647,31751,31842,31845,31850,32007,32134,32271,32348,32426,32429,32451,32530,32554,32567,32609,32658,32881,32928,33201,33215,33410,33454,33459,33563,33579,33907,33978,34463,34504,34651,34689,34737,34775,34928,34959,34978,35037,35258,35291,35295,35311,35353,35386,35471,35513,35600,35614,35672,35755,35771,35823,35846,35884,35970,36004,36056,36067,36144,36186,36278,36290,36313,36458,36514,36545,36592,36648,36801,36826,36831,36843,36856,36914,36969,36990,37022,37052,37212,37263,37286,37370,37380,37409,37509,37537,37548,37607,37631,37632,37836,37875,38050,38143,38159,38239,38363,38407,38496,38600,38637,38876,38891,38898,38906,38963,39004,39134,39157,39173,39224,39277,39295,39322,39395,39432,39524,39681,39702,39795,39834,39967,39970,40048,40252,40301,40305,40411,40453,40536,40551,40572,40593,40622,40686,40718,40736,40969,40999,41043,41063,41099,41144,41273,41388,41421,41438,41836,41952,41967,41974,42014,42023,42194,42226,42424,42495,42601,42647,42660,42702,42723,42834,42939,42981,43046,43271,43316,43348,43381,43609,43654,43665,43709,43782,43783,43856,43877,43930,43963,44010,44014,44032,44227,44584,44642,44692,44760,44793,44830,44940,45041,45057,45062,45150,45304,45349,45414,45473,45520,45552,45632,45669,45683,45706,45721,45736,45878,45899,45907,45932,45991,45994,46055,46056,46072,46191,46234,46275,46282,46284,46386,46389,46416,46445,46530,46608,46611,46661,46836,46927,47157,47272,47285,47385,47616,47702,47712,47737,47757,47777,47920,47972,47973,48032,48118,48149,48154,48245,48282,48344,48451,48458,48552,48562,48629,48634,48765,48867,49091,49105,49229,49313,49355,49669,49700,49719,49726,49752,50050,50107,50154,50337,50360,50540,50577,50597,50830,50869,50877,50897,50997,51109,51148,51328,51427,51432,51440,51515,51543,51715,51778,51779,51800,51807,51818,51835,52196,52430,52449,52618,52636,52760,52772,52823,52862,53060,53070,53122,53240,53310,53335,53380,53594,53725,53990,54035,54055,54059,54113,54129,54335,54344,54421,54503,54715,54797,54909,54935,54968,55382,55432,55454,55589,55600,55706,55752,56066,56135,56188,56249,56272,56320,56344,56360,56537,56580,56667,56703,56730,56791,56827,56919,56968,57000,57018,57071,57140,57163,57301,57363,57407,57447,57547,57676,57764,57798,57815,57816,57846,57911,58058,58098,58129,58169,58240,58242,58445,58499,58521,58727,58735,58777,59004,59119,59134,59334,59413,59525,59615,59745,59770,59839,59956,60087,60143,60179,60277,60343,60439,60510,60541,60591,60713,60719,60756,60943,61002,61074,61432,61618,61662,61690,61752,61901,61904,62494,62539,62543,62694,62775,62805,62839,62846,63006,63034,63102,63118,63149,63159,63264,63328,63375,63402,63564,63565,63701,63749,63771,63856,63858,63898,63928,64008,64045,64084,64123,64173,64299,64639,64673,64706,64769,64873,64939,64972,64987,65001,65143,65188,65263,65287,65304,65481,65550,65627,65690,65694,65774,65884,65902,65996,66036,66160,66256,66377,66406,66435,66568,66876,66926,66967,67033,67124,67145,67197,67213,67256,67279,67298,67371,67502,67541,67566,67759,67761,67795,67808,67827,67892,67980,67987,68068,68111,68229,68235,68333,68406,68647,68689,68758,68881,68911,68983,68997,69017,69075,69123,69146,69231,69232,69274,69276,69349,69378,69527,69528,69550,69809,69949,70030,70057,70138,70333,70368,70452,70479,70535,70676,70721,70759,70761,70829,70836,71037,71049,71145,71147,71154,71425,71482,71531,71603,71610,71679,71698,71793,71794,71894,71923,71973,72072,72099,72130,72200,72246,72484,72536,72610,72633,72673,72902,72950,72975,73058,73195,73226,73231,73367,73442,73483,73548,73564,73698,73743,73760,73861,74058,74096,74133,74596,74613,74648,74704,74794,74826,74904,75021,75156,75257,75353,75489,75506,75529,75541,75673,75700,75768,75778,75800,75801,75874,76046,76166,76201,76234,76271,76300,76435,76521,76592,76635,76796,76825,76882,76957,77008,77014,77038,77092,77120,77261,77282,77328,77398,77418,77431,77471,77510,77530,77557,77667,77677,77685,77777,77827,77857,78082,78083,78400,78451,78516,78620,78774,78792,78830,78897,78934,78938,79062,79095,79097,79099,79118,79225,79304,79428,79432,79445,79479,79554,79570,79627,79733,79850,79870,80048,80203,80228,80234,80330,80505,80575,80685,80751,80804,80846,80933,80989,81176,81205,81520,81526,81675,81782,81918,82066,82100,82178,82265,82346,82352,82366,82446,82451,82635,82770,82782,82783,82870,83049,83084,83188,83206,83294,83350,83535,83548,83710,83717,83882,84078,84158,84254,84268,84333,84353,84441,84482,84504,84565,84746,84767,84881,85265,85287,85303,85367,85406,85416,85437,85469,85615,85784,85874,85888,86047,86162,86164,86294,86376,86386,86404,86419,86471,86554,86565,86621,86633,86664,86691,86866,86914,86984,86986,87004,87126,87127,87165,87343,87376,87483,87498,87502,87555,87613,87636,87882,87909,87941,88068,88082,88091,88111,88139,88188,88210,88226,88325,88501,88537,88569,88600,88647,88718,88796,88816,88836,88851,88880,89058,89089,89109,89118,89308,89375,89416,89522,89552,89729,89853,89904,89939,89974,89984,89989,90021,90052,90342,90459,90477,90489,90513,90515,90664,91332,91452,91598,91840,91918,91920,91940,92033,92155,92234,92235,92298,92475,92485,92597,92734,92742,92856,93012,93172,93376,93430,93451,93618,93631,93942,94063,94091,94167,94362,94480,94491,94782,95064,95080,95121,95123,95174,95193,95197,95397,95439,95524,95628,95708,95790,95831,96039,96215,96235,96269,96293,96441,96489,96580,96712,96713,96751,96777,96795,96852,96938,97153,97160,97168,97267,97308,97343,97377,97404,97500,97543,97549,97636,97666,97678,97703,97861,97882,97891,97913,98281,98317,98342,98357,98523,98625,98658,98660,98817,98899,98943,99026,99105,99147,99162,99220,99249,99312,99424,99513,99532,99533,99581,99625,99641,99686,99708,99751,99861,99888,99994",
          "28,47,520,522,678,974,1122,1135,1171,1191,1277,1404,1486,1487,1765,1855,1865,1872,2027,2060,2100,2106,2136,2179,2290,2352,2374,2398,2436,2447,2495,2650,2704,2729,2767,2896,2935,3003,3009,3073,3146,3375,3395,3471,3557,3591,3600,3643,3645,3655,3723,3780,3787,3812,3864,3886,3971,4006,4016,4019,4054,4102,4122,4141,4195,4234,4287,4300,4315,4330,4436,4467,4483,4603,4716,4719,4830,4838,4957,5074,5357,5358,5429,5581,6073,6161,6377,6449,6646,6662,6692,6792,6915,6940,6953,7086,7206,7240,7243,7254,7475,7656,7692,7714,7915,7955,7999,8037,8039,8101,8179,8285,8305,8392,8458,8532,8852,8906,8918,8986,9098,9276,9355,9407,9435,9468,9561,9591,9698,9707,9709,9795,9804,9889,10018,10100,10167,10287,10326,10382,10414,10417,10468,10487,10587,10642,10696,10850,10969,10980,11031,11032,11153,11181,11202,11211,11274,11364,11365,11371,11377,11497,11521,11541,11654,11730,11801,11949,11993,12091,12128,12175,12235,12261,12403,12497,12578,12591,12609,12793,12839,12973,13147,13153,13188,13394,13512,13609,13797,13838,13872,14154,14271,14272,14347,14372,14376,14377,14405,14437,14449,14620,14750,14751,14797,14804,14828,14874,15002,15065,15067,15082,15199,15257,15270,15307,15335,15382,15405,15434,15441,15444,15581,15610,15626,15816,15818,15833,15959,16037,16097,16121,16128,16134,16138,16263,16266,16341,16567,16606,16621,16687,16830,16889,16921,16937,17079,17091,17110,17230,17246,17326,17331,17356,17473,17767,17780,17874,17958,18001,18070,18134,18146,18163,18204,18220,18233,18311,18318,18335,18348,18360,18427,18499,18583,18650,18663,18683,18706,18769,18771,18774,18921,18924,19000,19008,19023,19060,19108,19360,19605,19606,19653,19697,19843,19983,19989,20067,20078,20133,20134,20181,20193,20251,20262,20309,20356,20367,20371,20469,20498,20556,20607,20709,20759,20780,20798,20929,20952,20957,21131,21156,21158,21159,21272,21369,21432,21677,21697,21706,21828,21874,21894,21936,22023,22036,22040,22095,22123,22171,22180,22188,22224,22444,22449,22481,22490,22491,22507,22547,22554,22726,22761,22923,23014,23029,23129,23144,23221,23259,23343,23381,23417,23434,23461,23463,23467,23502,23563,23654,23715,23734,23737,23761,23811,23936,23967,23972,23978,24083,24167,24252,24318,24323,24339,24450,24492,24543,24563,24579,24599,24678,24686,24804,24897,25133,25147,25185,25249,25296,25304,25336,25352,25573,25659,25660,25728,25780,25817,25819,25864,25899,25973,26009,26114,26122,26133,26166,26197,26286,26303,26430,26459,26474,26478,26529,26543,26560,26640,26699,26791,26831,26849,26875,26929,27035,27174,27360,27464,27781,27803,27894,28113,28133,28159,28238,28311,28312,28322,28438,28490,28504,28630,28690,28751,28765,28767,28794,28853,28923,28992,29054,29108,29158,29332,29368,29474,29568,29620,29628,29677,29720,29912,29998,30087,30170,30240,30286,30426,30547,30556,30592,30614,30688,30833,30839,31006,31081,31108,31132,31149,31164,31189,31426,31450,31514,31554,31688,31696,31715,31733,31742,31791,31796,31804,31817,31895,31912,31936,31941,32028,32116,32187,32225,32236,32256,32291,32301,32303,32340,32366,32414,32423,32452,32482,32541,32582,32637,32656,32757,32798,32905,33010,33052,33167,33234,33254,33266,33304,33338,33355,33362,33392,33469,33496,33514,33528,33595,33610,33628,33636,33732,33749,33921,33937,33944,34073,34154,34310,34404,34411,34476,34533,34550,34593,34597,34732,34743,34785,34786,34853,34865,34993,34997,35023,35024,35026,35061,35100,35107,35108,35122,35125,35132,35141,35215,35256,35267,35276,35304,35319,35363,35407,35419,35434,35465,35492,35495,35606,35702,35817,35828,35858,35887,35891,35919,35940,35977,36112,36130,36140,36156,36202,36300,36333,36364,36403,36406,36503,36520,36531,36699,36772,36779,36794,36817,36845,36881,36884,36952,36973,37003,37029,37050,37091,37124,37161,37168,37281,37367,37445,37447,37464,37475,37483,37592,37640,37723,37734,37748,37766,38007,38154,38432,38491,38497,38498,38585,38671,38710,38788,38794,38862,38879,39019,39031,39252,39283,39323,39332,39343,39387,39442,39636,39696,39743,39765,39827,39830,39867,39898,39902,39926,39956,39982,40038,40043,40127,40172,40194,40259,40264,40294,40423,40478,40609,40664,40799,40821,40861,40931,40983,41036,41054,41126,41170,41193,41213,41259,41333,41335,41378,41403,41447,41514,41526,41533,41567,41596,41615,41619,41628,41656,41716,41735,41736,41787,41820,41870,41953,41955,42022,42037,42050,42057,42070,42084,42127,42165,42345,42539,42540,42574,42614,42637,42649,42658,42669,42697,42780,42869,42922,42946,42988,43004,43007,43063,43118,43159,43170,43178,43195,43217,43250,43252,43317,43387,43398,43434,43474,43496,43525,43584,43592,43656,43662,43698,43737,43805,43823,43864,43868,43879,43968,44046,44123,44136,44140,44178,44248,44309,44447,44457,44578,44619,44701,44750,44789,44919,45000,45023,45058,45233,45234,45299,45324,45444,45554,45556,45587,45692,45817,45846,45969,45981,45999,46012,46227,46248,46328,46401,46504,46543,46601,46648,46669,46819,46973,47035,47058,47067,47131,47165,47251,47484,47557,47564,47578,47615,47735,47780,47808,47827,47865,47975,48096,48132,48133,48268,48280,48383,48518,48783,48787,48799,49051,49102,49161,49238,49332,49473,49491,49692,49729,49730,49747,49815,49826,49994,50000,50047,50080,50094,50139,50155,50177,50290,50359,50414,50501,50537,50578,50681,50688,50740,50783,50850,50978,51008,51020,51068,51281,51301,51347,51377,51411,51622,51684,51751,51791,51830,51873,51898,51904,51918,52041,52346,52438,52473,52475,52477,52643,53047,53097,53130,53158,53168,53196,53201,53356,53762,53798,53974,54094,54097,54100,54356,54675,54727,54795,54878,55044,55090,55133,55184,55198,55463,55632,55729,55821,55957,55980,56061,56068,56166,56318,56355,56365,56477,56504,56549,56602,56650,56682,56737,56761,56792,56883,56935,57006,57225,57356,57548,57563,57603,57623,57763,57773,57777,57799,57845,57900,57987,58017,58024,58080,58099,58132,58148,58285,58359,58381,58439,58501,58508,58635,58688,58739,58748,58840,58882,58947,58948,59021,59028,59084,59089,59148,59159,59163,59200,59396,59425,59524,59549,59593,59606,59609,59731,59866,59913,59926,59978,59989,60012,60015,60020,60141,60217,60266,60384,60556,60631,60737,60907,61187,61220,61233,61285,61353,61490,61507,61813,61855,61932,61980,62007,62025,62184,62210,62344,62398,62430,62645,62886,63362,63422,63439,63441,63467,63487,63525,63606,63616,63676,63695,63759,63846,63859,63885,63893,63908,63943,63958,64056,64207,64269,64391,64394,64401,64437,64458,64697,64709,64765,64808,64837,65054,65067,65073,65136,65140,65159,65242,65415,65470,65510,65522,65604,65635,65992,66089,66224,66230,66253,66255,66290,66371,66499,66514,66562,66579,66592,66640,66689,66741,66751,66768,66862,66868,66963,67050,67081,67159,67249,67353,67450,67701,67765,67965,68070,68080,68083,68166,68225,68336,68515,68535,68597,68650,68730,68772,68876,68898,69049,69089,69195,69406,69789,69807,69815,69848,69858,69882,69933,69952,70294,70338,70419,70425,70802,70808,70831,70925,71028,71359,71490,71577,71786,71864,71969,71992,72015,72278,72506,72519,72562,72638,72682,72693,72695,72898,73094,73145,73168,73176,73199,73237,73266,73288,73338,73411,73508,73637,73711,73722,73775,73800,73814,74032,74135,74372,74373,74424,74437,74454,74586,74625,74628,74640,74649,74687,74716,74755,74946,75081,75135,75143,75149,75173,75414,75427,75631,75727,75792,75963,76003,76013,76072,76162,76280,76299,76426,76434,76452,76485,76593,76618,76621,76623,76778,76925,76936,76966,76969,76976,76983,76995,77056,77132,77172,77241,77491,77587,77642,77689,77951,77965,78031,78163,78166,78291,78306,78390,78415,78520,78662,78762,78808,78834,78891,78931,79018,79265,79299,79342,79588,79625,79642,79677,79833,79837,79897,79971,79996,80046,80156,80352,80383,80510,80553,80603,80696,80816,80886,80932,81077,81167,81307,81359,81839,81851,81935,82045,82106,82166,82174,82264,82273,82287,82330,82338,82343,82350,82452,82611,82627,83071,83249,83333,83455,83469,83597,83644,83708,83757,83785,83864,83871,83937,83939,84073,84179,84185,84308,84323,84335,84368,84471,84496,84532,84547,84600,84617,84632,84763,84874,84979,85006,85072,85149,85315,85327,85384,85453,85471,85506,85764,85776,85860,85862,85882,85918,85934,86060,86096,86145,86229,86380,86415,86447,86502,86544,86763,86764,86827,86832,86913,86948,86952,87009,87105,87147,87158,87166,87275,87345,87368,87413,87462,87673,87709,87816,87846,87891,87947,88044,88066,88075,88158,88168,88174,88238,88634,88646,88912,89108,89146,89167,89208,89332,89342,89346,89363,89404,89419,89647,89775,90059,90109,90146,90187,90208,90458,90521,90532,90756,90853,90956,91193,91256,91322,91334,91364,91446,91607,91636,91816,91864,91978,91979,92142,92174,92334,92387,92476,92477,92501,92618,92704,92705,92708,92774,92825,93063,93240,93248,93407,93433,93438,93554,93606,93622,93721,93783,93816,93821,93824,93887,93897,94005,94216,94249,94290,94380,94425,94434,94493,94542,94650,94675,94750,94789,94939,94944,94973,95008,95055,95135,95333,95423,95479,95569,95587,95699,95712,95761,95918,95933,95961,96106,96408,96415,96448,96460,96464,96482,96504,96561,96613,96721,96782,96794,96826,97010,97023,97213,97351,97353,97391,97540,97664,97736,97900,98001,98003,98052,98082,98159,98193,98227,98327,98335,98380,98392,98442,98556,98693,98807,98914,98940,98997,99004,99060,99072,99110,99209,99280,99408,99559,99733,99755,99760,99813,99821,99848,99889,99974,99999",
          "38,300,674,784,857,1034,1080,1166,1225,1307,1349,1427,1464,1497,1533,1576,1651,1913,1918,1973,2016,2104,2195,2261,2393,2528,2587,2633,2750,2774,2788,2975,3018,3039,3099,3155,3170,3318,3319,3367,3475,3524,3573,3656,3666,3771,3776,3936,4001,4100,4304,4366,4421,4557,4574,4602,4739,4784,4790,4919,4935,5010,5063,5107,5118,5147,5248,5373,5623,5734,5864,5995,6147,6268,6291,6296,6352,6416,6434,6481,6648,6665,6675,6730,6738,6803,6880,6901,6987,6994,7110,7138,7228,7251,7333,7358,7385,7485,7595,7635,7780,7957,8033,8191,8235,8309,8431,8456,8547,8632,8709,8760,8796,8992,9004,9032,9033,9055,9190,9209,9328,9379,9535,9617,9632,9636,9704,9705,9742,9788,9832,9946,9972,10088,10336,10458,10504,10589,10679,10844,10916,10978,11243,11246,11361,11455,11549,11577,11617,11660,11793,11804,11846,11932,11975,12075,12110,12143,12204,12293,12318,12342,12409,12476,12536,12538,12539,12542,12584,12588,12656,12698,12738,12759,12795,12850,12871,13052,13084,13129,13197,13274,13368,13423,13484,13487,13500,13579,13615,13626,13654,13691,13698,13779,13832,13857,13866,13868,13896,13982,14125,14206,14339,14340,14355,14417,14477,14573,14576,14594,14648,14860,14889,14914,14919,14927,14983,15079,15111,15187,15271,15347,15385,15448,15457,15582,15957,15966,15973,16122,16322,16349,16358,16421,16470,16515,16533,16623,16746,16909,16932,17013,17046,17056,17276,17305,17415,17423,17631,17633,17636,17646,17754,17863,17939,17963,18020,18126,18313,18338,18357,18425,18480,18694,18795,18853,19088,19089,19272,19303,19308,19325,19351,19387,19423,19500,19641,19839,19874,19899,20031,20043,20126,20234,20242,20419,20442,20488,20566,20574,20614,20626,20663,20666,20733,20736,20742,20824,20843,20845,20865,20928,21103,21200,21219,21242,21246,21373,21421,21430,21508,21511,21559,21636,21674,21763,21873,21896,21914,21958,21963,21977,22004,22032,22080,22106,22432,22469,22496,22590,22623,22687,22703,22755,22819,22827,22829,22983,23141,23155,23208,23314,23353,23360,23430,23601,23850,23875,23915,24007,24102,24103,24185,24258,24282,24342,24474,24490,24518,24583,24588,24770,24860,25021,25058,25104,25171,25269,25324,25390,25422,25513,25582,25605,25608,25649,25722,25761,25820,25837,25851,25926,26094,26158,26172,26198,26249,26330,26371,26427,26444,26449,26602,26618,26641,26663,26912,26971,27118,27190,27230,27283,27299,27345,27407,27550,27683,27722,27844,27866,27901,27909,28099,28172,28237,28257,28264,28280,28285,28476,28477,28518,28667,28724,28754,28845,28925,29073,29087,29213,29287,29323,29489,29491,29492,29493,29669,29674,29729,29778,29792,29983,30026,30135,30168,30206,30320,30349,30393,30423,30641,30666,30669,30816,30843,30883,30890,30907,30985,31130,31139,31235,31244,31283,31326,31508,31621,31789,31933,31944,31971,32125,32195,32199,32200,32202,32240,32259,32392,32416,32424,32593,32602,32676,32738,32773,32895,32982,33049,33106,33137,33139,33202,33287,33416,33444,33456,33533,33703,33722,33738,33800,33920,33948,33976,34217,34238,34442,34447,34555,34569,34583,34650,34700,34742,34750,34799,34829,34848,34913,35021,35028,35092,35300,35403,35406,35422,35426,35461,35501,35522,35756,35797,35815,35819,35868,35890,35945,36023,36039,36094,36205,36236,36237,36259,36299,36337,36367,36380,36421,36478,36526,36553,36573,37068,37318,37426,37503,37661,37707,37796,37840,37863,37956,37982,38011,38012,38031,38082,38174,38190,38278,38399,38495,38509,38549,38741,38759,38907,38939,39071,39081,39084,39220,39350,39366,39411,39449,39487,39547,39604,39716,39781,39910,39995,40001,40031,40204,40222,40245,40302,40368,40480,40487,40559,40780,40784,40797,40814,40919,40950,40991,41001,41024,41120,41180,41198,41219,41278,41389,41405,41520,41543,41579,41624,41695,41752,41789,41799,41978,42030,42043,42104,42191,42285,42307,42456,42520,42627,42722,42767,42872,42892,42897,42943,42955,42965,42969,43032,43090,43121,43122,43354,43436,43516,43620,43631,43663,43790,43829,43921,43944,43984,44094,44183,44203,44316,44402,44513,44514,44528,44659,44918,44953,45046,45054,45088,45182,45289,45308,45362,45376,45622,45670,45689,45760,45843,45914,46041,46071,46075,46091,46212,46215,46281,46302,46358,46396,46464,46510,46560,46876,46931,47000,47097,47117,47170,47235,47274,47308,47311,47325,47338,47369,47376,47378,47395,47425,47474,47581,47771,47861,47932,48059,48200,48218,48330,48425,48571,48596,48792,48887,49070,49266,49279,49292,49535,49605,49627,49696,49712,49757,49778,49799,49821,49833,50011,50037,50062,50106,50151,50277,50458,50493,50551,50592,50665,50693,50762,50828,50893,50966,50971,51120,51275,51471,51654,51669,51707,51734,51766,52061,52094,52284,52331,52359,52682,52737,52803,52891,52941,52948,53066,53094,53131,53152,53159,53271,53389,53393,53497,53527,53552,53882,54051,54060,54122,54166,54225,54368,54461,54464,54502,54507,54515,54596,54601,54687,54811,55028,55038,55105,55189,55218,55407,55499,55504,55509,55555,55653,55681,55692,55705,55750,55814,55823,55933,55950,56086,56356,56358,56387,56616,56681,56805,56871,56903,56958,57063,57065,57191,57410,57551,57591,57631,57667,57709,57716,57724,57825,58060,58104,58111,58142,58228,58295,58402,58628,58684,58686,58725,58847,59055,59150,59185,59195,59404,59664,59668,59704,59709,59765,59826,59994,60142,60176,60253,60271,60324,60444,60485,60563,60596,60694,60815,60820,60822,60831,60864,60886,60893,61122,61125,61128,61138,61192,61242,61543,62033,62077,62118,62174,62198,62323,62356,62475,62721,62735,62862,62863,62908,63024,63155,63189,63254,63303,63499,63590,63622,63640,63710,63716,63735,63800,63970,64022,64204,64241,64261,64308,64330,64335,64642,64679,64689,64724,64846,64893,64912,64921,64929,64947,65062,65153,65198,65262,65281,65314,65446,65535,65551,65597,65663,65872,65895,66029,66033,66183,66271,66321,66724,66800,66816,66820,66836,66851,66887,66964,66973,67018,67028,67039,67061,67177,67260,67280,67376,67554,67612,67662,67757,67990,68052,68092,68172,68214,68292,68466,68491,68611,68625,68784,68841,68863,68919,68948,68980,69087,69125,69166,69277,69336,69339,69357,69407,69617,69635,69654,69682,69698,69719,69797,69825,69960,69962,70056,70067,70142,70155,70306,70321,70360,70456,70463,70499,70791,70946,70995,71033,71046,71092,71107,71266,71303,71335,71364,71709,71722,71738,71754,71755,72081,72267,72374,72379,72430,72446,72578,72664,72732,72761,72825,72892,72900,72960,72971,73224,73281,73372,73431,73496,73530,73584,73589,73693,73770,73811,73857,73883,74004,74026,74068,74150,74162,74206,74255,74401,74412,74483,74579,74614,74667,75013,75153,75429,75804,75837,75846,75847,75894,75961,76112,76144,76149,76249,76362,76396,76416,76429,76458,76555,76629,76700,76720,76851,77029,77154,77386,77482,77504,77544,77546,77588,77595,77813,77931,78071,78106,78140,78143,78179,78254,78273,78293,78302,78354,78466,78481,78507,78771,78811,78923,78976,79003,79011,79134,79313,79595,79723,79794,79855,79856,79882,79891,79906,79922,79937,80043,80344,80355,80393,80409,80443,80523,80555,80609,80835,80872,80897,81006,81039,81238,81428,81564,81568,81598,81664,81795,81798,81802,81806,81807,81817,81978,81986,82018,82040,82143,82185,82230,82536,82576,82598,82615,82630,82639,82807,82893,82976,83014,83021,83139,83221,83239,83240,83292,83301,83433,83556,83580,83594,83764,83823,83832,83839,83874,83971,84059,84112,84239,84350,84476,84508,84535,84663,84752,84790,84931,84995,85029,85095,85166,85184,85203,85326,85347,85382,85440,85539,85582,85626,85829,86153,86247,86392,86398,86433,86443,86510,86522,86566,86728,86759,87075,87146,87180,87205,87216,87244,87248,87405,87411,87515,87554,87569,87575,87852,87950,88500,88573,88630,88645,88688,88700,88747,88774,88780,88848,88983,89097,89134,89139,89221,89386,89388,89414,89464,89565,89610,89615,89711,89755,89793,89901,89981,90001,90065,90077,90117,90248,90305,90479,90533,90535,90630,90708,90964,90966,90986,91014,91072,91121,91178,91189,91296,91649,91723,92001,92030,92053,92309,92336,92370,92390,92418,92519,92521,92530,92689,92794,92913,93035,93114,93117,93177,93182,93208,93348,93891,93972,93983,94044,94119,94162,94402,94455,94517,94549,94617,94625,94657,94723,94764,94937,95023,95066,95134,95298,95302,95401,95589,95601,95680,95962,95986,96010,96276,96313,96417,96550,96671,96783,96809,97003,97055,97085,97111,97388,97415,97504,97542,97638,97752,97811,97822,97839,97852,97853,98056,98099,98132,98291,98319,98431,98717,98764,98768,98955,99012,99063,99079,99131,99157,99257,99267,99278,99379,99500,99696,99788,99805,99814,99843,99875,99881",
          "43,322,335,409,426,432,644,652,762,977,990,1348,1635,1744,1794,1933,2019,2065,2092,2212,2305,2324,2414,2488,2769,2770,2888,2900,3107,3135,3142,3156,3259,3317,3381,3440,3450,3583,3617,3627,3685,3699,3799,3866,3904,3973,4031,4371,4584,4635,4699,4723,4741,4794,4829,4908,5284,5375,5377,5394,5436,5438,5592,5616,5707,5749,5819,5892,5930,6023,6115,6153,6300,6344,6364,6374,6563,6589,6658,6853,6879,6963,7215,7298,7461,7546,7553,7566,7580,7588,7788,7802,7844,7884,7909,7920,7983,8018,8044,8076,8114,8141,8157,8238,8239,8279,8490,8491,8681,8695,8724,8726,8787,9143,9165,9312,9326,9439,9557,9624,9654,9689,9729,9777,9880,9881,9986,10055,10097,10133,10232,10367,10405,10450,10509,10768,10866,10875,10889,10933,10958,11049,11110,11171,11253,11264,11287,11291,11535,11547,11586,11645,11764,11848,12030,12105,12230,12473,12492,12517,12556,12670,12771,12828,12857,12858,13201,13267,13294,13470,13511,13741,13742,13777,13808,13850,13899,13974,14080,14131,14310,14318,14393,14399,14471,14536,14733,14792,14852,14904,14964,15005,15023,15127,15339,15341,15423,15459,15560,15637,15645,15674,15701,15728,15942,15943,15954,16009,16015,16081,16084,16103,16489,16643,16936,17028,17074,17095,17366,17433,17637,17808,17823,17857,17872,17897,18022,18124,18413,18443,18525,18565,18656,18693,18713,18734,18787,18867,18971,18994,19042,19092,19104,19109,19112,19156,19335,19444,19480,19498,19554,19694,19699,19707,19816,19898,19918,19948,19975,19995,20006,20036,20071,20175,20232,20285,20428,20522,20599,20673,20678,20750,20785,20810,20963,21025,21226,21257,21304,21337,21342,21444,21464,21509,21527,21528,21615,21616,21640,21682,21740,21749,21758,21809,21837,21918,22065,22128,22149,22419,22584,22681,22709,22733,22808,22812,22815,23123,23337,23413,23560,23740,23789,23945,23965,23979,24090,24125,24142,24143,24158,24207,24233,24237,24267,24305,24349,24368,24439,24507,24607,24638,24769,24852,24900,24985,25193,25245,25305,25386,25403,25434,25505,25535,25583,25781,25803,25841,25860,25898,25904,25912,25968,25997,26224,26347,26351,26438,26539,26574,26645,26697,26711,26833,26940,26978,27025,27053,27062,27070,27176,27195,27222,27224,27226,27228,27236,27343,27393,27417,27583,27614,27795,27940,27974,28060,28063,28097,28229,28253,28323,28380,28404,28493,28537,28660,28727,28771,28795,28844,28908,29103,29125,29196,29234,29283,29333,29347,29413,29440,29459,29477,29508,29581,29583,29599,29607,29704,29731,29813,29849,29911,29931,30090,30102,30278,30468,30497,30502,30622,30634,30642,30694,30977,31062,31064,31065,31257,31288,31304,31375,31378,31380,31416,31465,31497,31611,31652,31810,31827,31858,31860,31875,31935,31942,32003,32053,32128,32157,32252,32269,32341,32376,32413,32753,32776,32800,32885,32907,32954,33077,33123,33158,33162,33210,33218,33221,33261,33289,33297,33335,33351,33408,33409,33479,33591,33608,33655,33675,33710,33723,33724,33834,33953,34013,34068,34100,34139,34175,34191,34302,34384,34390,34393,34432,34434,34471,34480,34505,34524,34643,34673,34674,34827,34903,34950,34967,34977,35013,35066,35133,35245,35287,35535,35543,35637,35703,35714,35726,35870,35880,35881,35892,35986,36009,36027,36058,36105,36190,36252,36332,36348,36506,36559,36645,36665,36709,36718,36816,36834,36839,36965,36989,37008,37043,37112,37128,37194,37209,37327,37330,37402,37420,37571,37660,37662,37664,37727,37794,37871,37963,37965,38019,38056,38118,38130,38155,38255,38311,38365,38370,38431,38466,38516,38527,38560,38682,38706,38746,38805,38937,38991,39094,39139,39154,39198,39275,39278,39509,39581,39671,39689,39730,39776,39863,39880,39975,40003,40042,40137,40215,40235,40396,40402,40403,40416,40439,40473,40510,40524,40567,40650,40787,40800,40801,40901,40957,41009,41088,41091,41119,41166,41174,41215,41241,41249,41282,41301,41351,41453,41481,41554,41599,41600,41603,41620,41636,41641,41824,41985,42096,42110,42146,42241,42244,42265,42402,42485,42491,42507,42571,42599,42714,42730,42739,42796,42816,42821,42900,43059,43151,43275,43281,43347,43388,43438,43440,43558,43597,43613,43675,43683,43702,43706,43753,43789,43820,43859,43891,44131,44159,44172,44352,44353,44364,44388,44395,44433,44437,44503,44509,44541,44646,44723,44779,44797,44812,45166,45175,45205,45222,45240,45295,45321,45402,45474,45507,45533,45589,45642,45713,45842,45880,45891,46096,46181,46241,46475,46545,46660,46744,46796,47034,47051,47232,47247,47289,47316,47343,47540,47576,47670,47685,47699,47761,48048,48055,48172,48229,48258,48316,48327,48333,48349,48353,48385,48434,48528,48554,48561,48741,48771,48801,48817,48841,48854,48882,48922,48933,48956,48958,49082,49127,49139,49216,49255,49311,49360,49421,49541,49567,49660,49787,49812,49927,50076,50078,50090,50257,50342,50346,50531,50610,50630,50814,50860,50908,50988,51013,51124,51197,51236,51252,51559,51607,51625,51640,51760,51842,52027,52031,52111,52118,52124,52138,52218,52350,52542,52553,52670,52709,52808,52905,52937,53073,53082,53084,53106,53109,53300,53314,53491,53498,53601,53742,53767,53808,53812,53883,53921,53963,54031,54233,54412,54698,54788,54855,54946,54988,55036,55156,55176,55183,55423,55510,55619,55753,55835,55921,56049,56153,56259,56295,56337,56385,56388,56411,56420,56535,56606,56608,56698,56699,56829,56858,56902,56982,57075,57125,57180,57244,57454,57469,57504,57568,57572,57643,57714,57719,57748,57813,57844,57918,58010,58135,58203,58207,58251,58258,58296,58344,58569,58585,58607,58613,58675,58693,58793,58849,58911,58970,59003,59234,59311,59321,59405,59441,59494,59618,59685,59719,59720,59763,59938,59950,60148,60181,60265,60278,60335,60337,60350,60491,60517,60533,60559,60560,60654,60691,60726,60782,60803,60848,60890,60945,60970,61005,61016,61117,61123,61141,61189,61247,61259,61278,61374,61420,61438,61439,61469,61486,61489,61705,61990,62002,62044,62047,62125,62277,62280,62414,62445,62458,62557,62574,62606,62609,62618,62672,62783,62797,62950,62957,62974,63057,63114,63198,63260,63317,63333,63347,63360,63433,63474,63541,63552,63579,63672,63697,63799,63960,64015,64293,64341,64376,64676,64790,64819,64904,64925,65006,65049,65109,65152,65221,65252,65260,65273,65392,65623,65688,65802,65829,65958,65983,66046,66070,66099,66170,66316,66339,66366,66457,66623,66644,66669,66692,66739,66831,67012,67063,67070,67132,67418,67426,67440,67514,67533,67571,67603,67721,67909,67914,67955,68666,68774,68968,69012,69037,69039,69244,69493,69532,69587,69589,69744,69783,69829,69917,69922,69957,69958,69970,70011,70186,70373,70394,70412,70520,70547,70553,70642,70645,70748,70954,71108,71306,71325,71326,71442,71569,71619,71703,71704,71813,72056,72194,72208,72266,72297,72346,72380,72423,72463,72502,72607,72648,72700,72708,72735,72818,72833,72856,72861,72863,72893,73045,73078,73173,73324,73359,73382,73429,73498,73541,73550,73565,73889,74106,74113,74212,74261,74314,74315,74317,74514,74539,74664,74685,74784,74857,74934,74977,75068,75088,75159,75161,75344,75374,75575,75653,75676,75695,75749,75856,75890,75938,75950,76000,76384,76502,76512,76602,76766,76950,76952,77022,77060,77186,77240,77293,77401,77479,77548,77563,77593,77654,77735,77791,77864,77930,77935,78035,78052,78073,78169,78186,78223,78269,78314,78344,78405,78609,78626,78691,78766,78807,78969,79066,79124,79291,79469,79523,79530,79653,79654,79759,79767,79802,79854,79968,80219,80251,80302,80318,80335,80378,80663,80727,80782,80788,80856,81071,81139,81182,81237,81484,81628,81637,81685,81815,81937,81968,82055,82059,82074,82132,82231,82362,82365,82438,82450,82520,82779,82886,82950,82953,83016,83100,83229,83354,83399,83596,83651,83659,83664,83679,83684,83727,83827,83847,83887,84008,84127,84155,84267,84469,84505,84694,84764,84872,84966,85012,85254,85293,85419,85467,85559,85740,86009,86014,86015,86151,86167,86183,86499,86785,86983,87051,87064,87115,87148,87303,87525,87619,87663,87677,87755,87826,87887,87969,87992,88001,88003,88024,88052,88131,88253,88273,88394,88458,88494,88588,88652,88769,88991,89023,89040,89214,89244,89379,89456,89539,89581,89692,89698,89773,89785,89815,89836,89928,89972,90095,90165,90358,90430,90447,90464,90563,90681,90821,90931,90937,90962,90988,91127,91272,91293,91315,91428,91450,91465,91543,91631,91645,91650,91746,91929,91991,92041,92090,92153,92351,92366,92474,92499,92572,92652,92682,92776,92790,92922,92962,93046,93194,93206,93317,93357,93442,93536,93675,93827,93971,94089,94122,94143,94174,94198,94232,94243,94330,94392,94437,94490,94519,94569,94601,94635,94690,94691,94696,94734,94830,94910,94923,95151,95455,95472,95536,95582,95609,95646,95676,95701,95778,95827,95839,96152,96166,96247,96423,96499,96558,96636,96640,96727,96744,96874,96880,97009,97029,97199,97243,97394,97486,97632,97637,97855,97998,98188,98381,98385,98551,98559,98571,98623,98642,98675,98982,99303,99331,99486,99531,99554,99719,99914,99957,99971",
          "30,41,149,166,337,447,513,526,564,810,822,889,985,1090,1195,1261,1371,1390,1405,1554,1562,1582,1654,1749,1756,1815,1826,2037,2310,2356,2380,2451,2456,2542,2580,2679,3232,3254,3288,3422,3427,3443,3498,3529,3531,3635,3739,3740,3748,3880,3931,3942,4162,4187,4196,4272,4532,4554,4596,4631,4754,4966,4993,4996,5163,5374,5448,5488,5492,5598,5686,5786,5928,6002,6057,6101,6132,6218,6271,6367,6489,6497,6640,6768,6788,6893,7037,7059,7075,7089,7220,7296,7506,7645,7723,7820,7848,7862,7945,7988,8000,8045,8055,8062,8110,8139,8180,8198,8216,8280,8365,8557,8604,8691,8744,8878,8888,8921,9059,9078,9136,9476,9598,9674,9690,9789,9853,9925,9990,9993,10022,10091,10143,10163,10169,10198,10244,10265,10283,10302,10323,10437,10484,10537,10634,10639,10665,10698,10702,10759,10894,10942,10968,11063,11475,11580,11641,11873,11968,11992,12027,12068,12166,12199,12212,12257,12279,12314,12408,12438,12461,12470,12507,12543,12577,12615,12737,12821,12842,12967,13057,13102,13136,13226,13304,13360,13424,13430,13491,13510,13686,13924,14019,14038,14077,14115,14253,14332,14388,14499,14502,14569,14672,14809,14872,14890,14923,15069,15117,15247,15466,15490,15585,15591,15770,15962,16051,16052,16056,16098,16115,16187,16225,16411,16435,16469,16637,16638,16671,16786,16971,17005,17048,17201,17233,17298,17340,17649,17667,17752,17801,17843,17871,17962,17968,17989,18046,18052,18072,18081,18202,18208,18213,18215,18441,18492,18587,18610,18634,18671,18783,18796,18802,18817,18841,18851,18954,19077,19136,19257,19295,19468,19549,19564,19602,19984,20004,20014,20049,20085,20103,20116,20143,20164,20302,20306,20312,20530,20741,20753,20778,20782,20911,20942,20951,20991,21164,21214,21254,21351,21375,21457,21494,21522,21558,21718,21739,21772,21869,21921,22020,22101,22126,22184,22203,22273,22308,22357,22371,22411,22437,22495,22536,22561,22564,22585,22695,22697,22741,22959,22998,23102,23192,23309,23346,23485,23680,23890,23891,23908,23929,24009,24024,24026,24229,24415,24463,24616,24635,24688,24936,25007,25055,25189,25457,25543,25694,25764,25794,26007,26099,26221,26305,26521,26571,27057,27240,27244,27298,27301,27312,27335,27378,27453,27455,27543,27564,27584,27600,27677,27769,27864,27896,27928,28004,28009,28053,28231,28290,28519,28655,28659,28742,28750,28814,28822,28918,28980,29008,29030,29096,29274,29309,29327,29372,29421,29683,29692,29805,29848,29858,29892,29900,29980,30230,30312,30346,30381,30391,30425,30459,30462,30511,30543,30566,30680,30837,30873,30877,30966,31059,31061,31142,31147,31228,31373,31441,31458,31519,31592,31630,31637,31659,31674,31691,31728,31772,32045,32066,32114,32275,32278,32409,32420,32488,32574,32579,32639,32651,32713,32718,32744,32827,32867,32872,32926,33056,33058,33331,33447,33624,33764,33890,33952,33957,34026,34091,34102,34150,34192,34235,34268,34563,34631,34633,34638,34709,34867,34960,35090,35103,35106,35129,35143,35164,35193,35223,35265,35266,35298,35302,35312,35476,35539,35586,35602,35718,35751,35761,35930,35933,35967,35975,35989,36034,36090,36139,36302,36326,36616,36786,36980,37072,37125,37210,37214,37303,37325,37496,37686,37812,37821,37830,37857,38073,38163,38337,38442,38478,38575,38595,38667,38738,38834,38854,38936,38972,38975,38983,38986,39007,39077,39334,39389,39402,39534,39587,39723,39749,39866,39874,39941,39950,39973,40128,40138,40211,40216,40369,40406,40419,40424,40795,40807,40961,41083,41100,41123,41168,41231,41303,41309,41315,41316,41343,41449,41452,41489,41660,41724,41788,41793,41899,41989,41996,42138,42144,42195,42233,42329,42393,42464,42497,42652,42790,42915,42930,42941,43081,43104,43108,43196,43259,43265,43333,43346,43418,43503,43743,43757,43836,43925,44073,44075,44091,44244,44274,44282,44285,44380,44422,44485,44507,44573,44650,44678,44722,44818,44852,44931,44970,45033,45252,45316,45319,45516,45524,45532,45536,45543,45577,45620,45879,45933,45974,46059,46097,46101,46119,46124,46356,46380,46415,46636,46638,46649,46729,46764,46854,46982,47014,47078,47206,47352,47399,47419,47566,47583,47671,47698,47820,47955,48377,48378,48556,48601,48638,48663,48862,48947,48972,49023,49046,49112,49171,49209,49301,49367,49434,49594,49680,49724,49731,49745,49786,49843,50027,50048,50074,50112,50146,50158,50376,50430,50434,50491,50512,50571,50603,50622,50657,50780,50846,50890,50998,51007,51087,51147,51182,51212,51256,51368,51399,51460,51517,51527,51548,51649,51671,51674,51693,51890,51896,51910,52055,52106,52134,52157,52280,52283,52334,52371,52452,52520,52590,52627,52700,52704,52764,52984,53004,53049,53058,53089,53104,53273,53324,53378,53381,53401,53464,53576,53580,53668,54061,54080,54197,54279,54294,54383,54455,54573,54590,54628,54646,54789,54798,54839,54863,54903,55074,55081,55116,55196,55222,55227,55277,55353,55542,55601,55631,55652,55746,55794,55866,55906,55977,56227,56245,56334,56463,56473,56626,56714,56743,56765,56786,56803,56848,56882,57009,57097,57135,57290,57359,57362,57650,57669,57827,57829,57854,57981,58039,58420,58544,58556,58736,58828,58887,59082,59099,59411,59570,59834,59860,59921,60044,60055,60309,60395,60470,60476,60502,60586,60619,60780,60801,60947,61015,61156,61165,61240,61244,61252,61293,61344,61435,61458,61622,61629,61723,61727,61751,61771,61819,61875,62024,62178,62186,62261,62292,62425,62427,62506,62515,62763,62837,63025,63058,63154,63207,63265,63285,63302,63327,63337,63363,63411,63494,63498,63522,63688,63698,63723,63860,63877,63942,64087,64275,64465,64502,64508,64512,64580,64655,64786,64936,64941,65553,65582,65828,65987,66077,66233,66352,66539,66589,66687,66736,66782,66791,66832,66860,66927,66937,66941,67094,67119,67128,67188,67247,67321,67352,67365,67397,67548,67551,67564,67610,67690,67762,67787,67815,67942,67982,67985,68051,68207,68218,68248,68270,68428,68490,68680,68751,68791,68843,68873,68941,68954,69022,69025,69038,69136,69145,69254,69342,69454,69488,69519,69768,69790,69827,69835,69873,69878,69900,70031,70079,70080,70095,70108,70330,70364,70407,70433,70512,70574,70578,70614,70652,70700,70715,70731,70763,70803,70982,71001,71182,71233,71311,71333,71350,71437,71450,71459,71542,71674,71699,71732,72001,72079,72274,72289,72307,72569,72719,72935,72963,73003,73023,73083,73273,73342,73416,73439,73471,73522,73556,73560,73576,73596,73600,73715,73741,73812,73818,73942,74071,74074,74116,74223,74268,74301,74409,74455,74582,74797,74867,74917,74937,75056,75122,75307,75552,75612,75615,75650,75735,75746,76228,76235,76325,76348,76370,76394,76419,76494,76542,76569,76584,76631,76656,76798,76807,76900,77096,77100,77237,77297,77330,77355,77381,77402,77565,77618,77646,77651,77681,77695,77697,77746,77785,77812,78368,78485,78488,78597,78813,78867,78878,78928,79031,79116,79182,79272,79286,79296,79427,79460,79568,79575,79608,79640,79647,79688,79836,79913,79977,79990,80186,80194,80369,80455,80577,80724,80746,80820,80827,80840,80874,81111,81113,81301,81433,81475,81536,81561,81610,81680,81988,82017,82138,82163,82177,82296,82418,82624,82743,82800,82859,82892,83020,83061,83169,83314,83506,83786,83868,83931,83933,84067,84184,84202,84251,84326,84403,84424,84536,84550,84598,84630,84692,84759,84777,84805,84944,84971,85028,85098,85120,85121,85132,85150,85229,85252,85387,85562,85600,85618,85661,85715,85760,85789,85881,85970,86136,86141,86168,86244,86313,86321,86379,86408,86435,86444,86449,86538,86617,86711,86843,86860,87020,87112,87151,87221,87257,87365,87380,87454,87532,87533,87655,87672,87694,87777,87823,87967,88002,88039,88093,88125,88439,88512,88670,88763,88837,88881,88901,88976,89088,89161,89182,89210,89231,89245,89261,89265,89445,89499,89526,89718,89744,89837,89878,89909,89985,90000,90038,90354,90423,90433,90497,90581,90655,90683,90717,90730,90826,90835,90999,91091,91134,91144,91170,91292,91545,91546,91574,91613,91623,91624,91627,91754,91778,91807,91887,91891,91898,91963,92093,92130,92140,92251,92338,92411,92466,92680,92726,92745,92806,92901,92974,93025,93029,93052,93056,93074,93101,93168,93176,93271,93284,93383,93440,93490,93591,93739,93792,93902,93920,93980,94105,94534,94724,94793,94885,94925,95067,95235,95328,95421,95486,95487,95619,95738,95772,95927,95963,96223,96281,96320,96321,96393,96491,96554,96706,96810,96885,96909,97072,97133,97328,97357,97513,97533,98009,98067,98251,98299,98318,98423,98447,98531,98548,98599,98613,98617,98667,98728,98808,98823,98834,98900,98911,99052,99071,99112,99197,99247,99326,99385,99451,99547,99772,99865,99896,99939,99992",
          "12,36,49,155,157,199,203,261,387,572,750,754,806,902,953,1026,1123,1125,1137,1217,1229,1415,1426,1452,1479,1536,1668,1823,2074,2269,2442,2483,2619,2656,2680,2840,2866,2902,2910,2915,2977,2988,3082,3224,3229,3509,3568,3585,3604,3657,3680,3745,3779,3797,3854,3949,3985,4040,4112,4180,4193,4215,4288,4359,4360,4377,4383,4586,4729,4730,4751,4775,4887,4899,5018,5059,5108,5109,5239,5403,5452,5461,5587,5614,5633,5903,5971,6105,6159,6162,6165,6207,6260,6267,6303,6306,6334,6345,6357,6491,6739,6760,6770,6844,6889,7022,7025,7281,7363,7440,7496,7720,7787,7836,7838,7881,7969,8003,8073,8193,8246,8319,8353,8438,8622,8655,8682,8885,8898,8903,8917,8998,9126,9156,9194,9199,9266,9474,9501,9530,9593,9637,9642,9660,9684,9697,9797,9937,9962,10005,10054,10085,10154,10242,10320,10364,10489,10507,10560,10621,10632,10685,10770,10774,10808,10836,10865,10948,11067,11161,11334,11429,11479,11503,11512,11722,11768,11841,11895,11964,11985,11999,12004,12013,12145,12180,12197,12359,12395,12396,12475,12493,12622,12752,12760,12807,12822,12942,13008,13156,13157,13225,13281,13613,13711,13760,13772,13875,14149,14156,14205,14335,14400,14451,14486,14568,14603,14634,14666,14698,14752,14822,14841,14850,15261,15387,15398,15679,15840,15896,15897,16029,16063,16107,16143,16221,16285,16302,16396,16451,16472,16521,16557,16722,16825,16879,16914,16975,17130,17227,17393,17420,17450,17451,17478,17543,17582,17661,17731,17815,17907,18013,18150,18182,18192,18198,18314,18410,18438,18555,18692,18894,19039,19050,19051,19302,19322,19342,19393,19395,19454,19474,19493,19573,19740,19747,19844,19858,19933,19943,19994,19998,20002,20029,20039,20054,20057,20320,20411,20474,20703,20840,20969,20976,20996,21050,21069,21265,21347,21382,21545,21683,21688,21691,21769,21859,22262,22542,22566,22594,22649,22778,22786,22910,23081,23163,23184,23315,23513,23553,23599,23620,23622,23708,23711,23836,23842,23845,24044,24088,24111,24216,24224,24250,24255,24359,24520,24541,24592,24691,24790,24850,25163,25212,25286,25310,25373,25497,25527,25565,25574,25581,25596,25677,25753,25754,25757,25806,25919,25940,25948,26076,26106,26176,26290,26294,26325,26364,26494,26654,26706,26773,26876,26893,27018,27052,27078,27577,27625,27650,27686,27712,27717,27966,28001,28026,28132,28147,28182,28520,28525,28607,28764,28788,29258,29278,29307,29438,29572,29584,29777,29884,29937,29979,30187,30270,30360,30451,30550,30597,30701,30734,30743,30899,31239,31296,31369,31538,31629,31648,31649,31670,31721,31786,31915,31976,31982,32067,32191,32320,32622,32704,32726,32789,32809,32864,32882,32918,32984,33018,33293,33360,33471,33560,33606,33613,33621,33790,33896,34064,34131,34178,34218,34303,34352,34419,34566,34574,34605,34644,34677,34866,34932,34983,35056,35063,35113,35219,35274,35344,35431,35497,35518,35528,35534,35594,35655,35733,35778,35780,35847,35876,35908,35915,35938,35944,35973,36066,36150,36198,36257,36273,36351,36374,36446,36551,36634,36687,36701,36866,36899,36967,37024,37248,37324,37405,37417,37494,37638,37683,37815,37824,37864,37923,37940,37976,38023,38041,38054,38120,38151,38160,38164,38240,38369,38374,38451,38454,38486,38507,38581,38608,38629,38739,38772,38783,38825,38910,39059,39104,39128,39150,39306,39364,39406,39516,39546,39715,39735,39798,39836,39928,39939,39962,40024,40055,40076,40223,40227,40275,40433,40569,40584,40633,40743,40762,40772,40796,40859,41028,41095,41493,41566,41670,41768,41779,41783,41810,41839,41940,42032,42106,42159,42200,42228,42490,42733,42843,42864,42907,42928,42948,42958,42973,43008,43030,43191,43215,43239,43478,43498,43506,43541,43557,43626,43911,43960,44015,44038,44139,44188,44527,44555,44577,44604,44637,44672,44688,44814,44954,44973,44997,45018,45043,45064,45266,45358,45368,45420,45498,45500,45519,45531,45544,45606,45703,45740,45821,46099,46188,46235,46268,46289,46299,46366,46394,46453,46508,46683,46781,46828,47140,47485,47577,47681,47719,47797,47798,47853,47893,47945,47962,48007,48056,48069,48127,48128,48501,48591,48734,48752,48756,48763,48777,48798,48873,48921,48987,49032,49035,49128,49268,49361,49365,49399,49403,49515,49532,49538,49635,49643,49662,49750,49790,49811,49839,49847,49863,49910,49943,49988,50005,50020,50046,50109,50115,50138,50175,50214,50314,50426,50662,50692,50708,50747,50764,50918,51078,51083,51085,51099,51107,51196,51294,51312,51627,51655,51701,51735,51878,51925,52056,52109,52169,52245,52264,52462,52546,52625,52797,52825,52826,52906,53189,53190,53355,53449,53550,53666,53789,53816,53842,53928,53988,54140,54156,54228,54355,54469,54551,54766,54773,54869,54960,55148,55181,55274,55296,55359,55367,55389,55500,55543,55641,55664,55677,55817,55885,56207,56324,56503,56519,56546,56591,56600,56890,56918,56994,57029,57032,57112,57230,57388,57419,57472,57570,57779,57802,57850,57886,57902,57957,58266,58267,58509,58899,58927,59048,59104,59173,59236,59280,59308,59330,59331,59355,59559,59563,59697,59728,59822,59964,60031,60112,60155,60268,60289,60291,60359,60389,60400,60425,60494,60539,60577,60637,60648,60717,60728,60755,60781,60849,60872,60877,60880,60989,61174,61209,61264,61271,61292,61362,61448,61464,61522,61596,61615,61704,61728,61750,61778,62003,62103,62166,62383,62407,62461,62483,62497,62648,62695,62757,62824,63007,63018,63037,63120,63125,63219,63269,63271,63533,63730,63789,63815,64033,64136,64141,64250,64254,64339,64353,64398,64419,64445,64491,64509,64551,64554,64574,64627,64660,64895,64924,64948,65018,65029,65037,65091,65121,65235,65269,65341,65393,65457,65465,65467,65484,65492,65503,65506,65564,65585,65710,65869,65885,66035,66107,66158,66201,66221,66320,66509,66571,66753,66756,66855,67001,67003,67037,67161,67283,67375,67452,67586,67622,67682,67698,67716,67749,67837,67939,68100,68238,68340,68407,68556,68569,68681,68710,68882,69054,69085,69334,69402,69650,69689,69693,69740,69775,69787,69801,69887,69895,70054,70063,70089,70190,70208,70266,70299,70392,70528,70598,70698,70770,70834,70882,70951,70958,70965,71018,71062,71086,71118,71218,71300,71537,71578,71584,71860,71862,71868,71943,71948,72009,72123,72140,72239,72337,72445,72457,72467,72507,72514,72515,72518,72603,72703,72773,72795,72797,72851,72980,73020,73130,73132,73241,73243,73252,73255,73336,73502,73585,73595,73658,73682,73915,73926,73936,74177,74296,74340,74416,74518,74622,74696,74915,74925,74930,74987,75070,75288,75299,75371,75441,75452,75466,75593,75595,75715,75740,75814,75943,75992,76082,76174,76217,76456,76467,76536,76599,76717,76744,76909,76998,77009,77099,77136,77157,77351,77407,77505,77569,77575,77665,77696,77842,77848,77912,77925,77946,78080,78108,78311,78364,78369,78377,78399,78424,78496,78503,78700,78713,78815,78903,79015,79159,79308,79378,79436,79452,79464,79505,79550,79712,79745,79768,79820,80006,80055,80062,80207,80211,80217,80246,80319,80391,80460,80586,80711,80898,80908,81222,81268,81413,81418,81442,81503,81541,81549,81654,81689,81948,81956,82030,82070,82090,82180,82233,82304,82345,82382,82455,82517,82543,82867,82989,83036,83140,83237,83265,83299,83306,83348,83387,83416,83503,83641,84012,84029,84032,84042,84048,84084,84105,84159,84232,84289,84315,84316,84730,84737,85003,85083,85128,85228,85291,85292,85299,85309,85349,85388,85587,85589,85692,85713,85802,85914,85938,86008,86476,86545,86662,86703,86769,86921,87022,87107,87140,87233,87250,87288,87352,87491,87597,87626,87665,87682,87730,87741,87973,88120,88126,88398,88446,88532,88541,88544,88575,88629,88810,88947,88961,89056,89140,89190,89232,89282,89309,89322,89353,89374,89458,89566,89588,89670,89680,89683,89810,89894,90005,90060,90112,90274,90337,90511,90589,90646,90649,90668,90719,90799,90808,90810,90816,91047,91082,91239,91306,91503,91553,91820,91945,91959,91976,92126,92168,92269,92369,92401,92426,92598,92635,92715,92743,92940,92984,93193,93332,93380,93384,93551,93568,93841,93966,94353,94357,94466,94568,94570,94669,94845,94967,95004,95044,95093,95125,95140,95195,95357,95379,95481,95564,95567,95710,95739,95744,95777,95925,96095,96104,96146,96207,96362,96456,96470,96618,96755,96847,96897,96900,96957,97028,97036,97087,97167,97294,97532,97545,97573,97577,97614,97772,97786,97830,98060,98066,98103,98206,98211,98373,98379,98406,98485,98554,98606,98684,98694,98721,98735,98778,98789,98884,99083,99138,99234,99471,99529,99557,99565,99573,99673,99685,99803,99853,99985,100002",
          "9,13,286,288,309,334,342,370,388,451,487,493,505,531,598,615,970,989,1012,1017,1020,1024,1215,1378,1416,1833,2612,5079,5099,5122,5124,5215,5250,5261,5271,5322,5370,5378,5382,5455,5836,5859,5870,5874,5925,5969,5980,6022,6047,6060,6068,6075,6117,6167,6242,6258,6315,6325,6499,6549,6727,6835,6895,6957,7126,7255,7417,7498,7514,7571,7886,8024,8100,8162,8225,8292,8334,8481,8715,8808,8886,8961,9551,10076,10160,10222,10285,10629,11014,11791,12023,13375,14836,15245,16413,16605,16706,16773,17894,18125,18242,18976,20483,21058,21314,21396,23395,23580,27497,28377,29436,30646,40028,42494,43323,50013,50029,50147,50192,50231,50239,50278,50408,50421,50441,50543,50595,50654,50793,50903,51186,51518,51636,51808,51850,51857,51979,51998,52044,52077,52197,52338,52358,52393,52463,52724,52736,52775,52846,52914,53343,53443,53450,53453,53480,53690,53766,53948,53968,54003,54104,54395,54428,54571,54765,54900,54912,55301,55303,55436,55441,55471,55512,55536,55578,55590,55599,55620,55622,55669,55676,55697,55732,55768,55874,55876,55993,56100,56140,56209,56224,56229,56283,56290,56372,56409,56418,56567,56653,56720,56869,56924,56942,57001,57188,57254,57535,57545,57588,57728,57840,57929,58067,58118,58193,58232,58376,58419,58502,58701,58926,58956,59213,59366,59476,59595,59739,59758,60039,60040,60042,60111,60364,60489,60613,60664,60701,60738,60741,60765,60903,60932,60948,61073,61075,61147,61150,61258,61333,61365,61381,61409,61493,61525,61691,61946,62053,62104,62264,62314,62489,62592,62867,62894,62922,63004,63021,63030,63061,63063,63244,63481,63493,63550,63602,63664,63673,63792,63808,63929,63944,63994,64064,64186,64225,64271,64305,64358,64386,64457,64672,64760,64792,64827,64842,64881,64937,64954,65103,65194,65227,65338,65367,65414,65447,65450,65469,65616,65621,65662,65770,65864,66056,66116,66168,66174,66332,66394,66440,66552,66565,66628,66633,66684,66884,66901,67097,67199,67240,67393,67580,67650,67848,67872,67881,67896,67922,68009,68069,68105,68106,68133,68152,68361,68434,68500,68529,68583,68683,68711,68770,68771,68830,68839,68868,68925,69015,69023,69042,69063,69278,69324,69429,69483,69564,69647,69729,70051,70185,70239,70377,70586,70603,70765,70850,70887,70957,71015,71055,71067,71077,71105,71193,71331,71339,71395,71412,71429,71522,71529,71696,71796,71899,72186,72222,72291,72336,72468,72516,72626,72779,72788,72943,73006,73067,73081,73343,73414,73588,73652,74124,74157,74167,74258,74330,74337,74367,74385,74714,74744,74782,74793,74988,75008,75098,75127,75241,75267,75341,75391,75516,75629,75718,75853,75857,75888,75891,75927,75990,75998,76038,76259,76263,76482,76622,76777,76872,77152,77209,77258,77339,77494,77509,77584,77683,77802,78051,78195,78300,78352,78380,78439,78555,78711,78775,78801,78812,78910,79032,79149,79322,79348,79423,79514,79631,79714,79830,79867,79969,80094,80162,80314,80476,80535,80660,80681,80734,80784,80861,81032,81102,81266,81512,81578,81653,81677,81749,81773,81809,81861,81891,82068,82088,82188,82406,82470,82539,82542,82554,82701,82757,82905,82943,83425,83608,83610,83633,83682,83870,83909,83946,83952,83980,84040,84103,84141,84371,84440,84461,84519,84651,84788,84819,85116,85178,85279,85385,85756,85833,85837,85952,85975,86081,86140,86260,86368,86418,86434,86548,86576,86678,86688,86760,86824,86990,87002,87150,87270,87281,87309,87317,87339,87401,87418,87472,87530,87540,87608,87670,87799,87897,87906,88406,88528,88653,88696,88719,88726,88786,88821,88942,89065,89077,89202,89236,89257,89370,89431,89528,89621,89665,89864,90100,90121,90322,90378,90456,90514,90635,90691,90764,90787,90823,91573,91583,91617,91655,91660,91744,91765,91975,92004,92031,92121,92429,92576,92634,92675,92833,92855,92888,92917,93002,93057,93070,93186,93431,93724,93732,93800,93982,94003,94018,94155,94210,94324,94327,94496,94585,94615,94728,94785,94820,94958,94984,95024,95126,95222,95260,95317,95338,95494,95725,95892,95923,96035,96125,96162,96171,96178,96208,96232,96260,96333,96436,96691,96771,96882,97047,97100,97193,97271,97561,97566,97610,97684,97698,97723,97729,97737,97817,97860,97886,97948,98010,98040,98041,98112,98215,98424,98448,98546,98596,98628,98928,98930,98980,98995,99129,99273,99341,99538,99582,99607,99747,99833"
        ]
      }
    }
  ]
}
)";
const folly::dynamic kEagerShardSelectionHugeJson{
    folly::parseJson(kEagerShardSelectionHuge)};

constexpr folly::StringPiece kEagerShardSelectionLoadBalancerHuge = R"(
{
  "children_type": "LoadBalancerRoute",
  "children_settings" : {},
  "pools": [
    {
      "pool": {
        "type": "Pool",
        "name": "kShardSelectionHuge.LoadBalancer.01",
        "servers": [
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312"
        ],
        "shards": [
          "509,610,640,644,648,659,820,907,972,1096,1148,1199,1377,1417,1420,1529,1565,1576,1587,1588,1589,1632,1713,1761,1830,1832,1860,1882,1942,2075,2093,2142,2246,2249,2253,2276,2361,2542,2596,2739,2812,2841,2865,2868,2916,2950,2973,3125,3334,3338,3379,3732,3752,3800,3816,4115,4124,4131,4360,4494,4606,4626,4772,4821,4822,4827,4877,5015,5046,5154,5177,5229,5296,5426,5467,5510,5536,5540,5542,5546,5647,5664,5665,5681,5757,5793,5800,5814,5828,5847,5853,5954,5976,6021,6023,6028,6064,6238,6370,6422,6510,6675,6745,7018,7148,7248,7261,7287,7309,7360,7401,7466,7472,7478,7499,7512,7937,7959,7991,8043,8204,8210,8214,8395,8399,8421,8490,8516,8627,8647,9017,9123,9297,9317,9319,9363,9401,9424,9433,9505,9541,9786,9897,9937,10128,10167,10169,10240,10243,10306,10491,10540,10692,10729,10798,10803,10881,11097,11160,11164,11344,11506,11536,11625,11693,11909,11994,12030,12049,12201,12277,12279,12321,12347,12545,12554,12634,12748,12773,12866,13159,13324,13417,13457,13462,13551,13683,13704,13832,13876,14025,14047,14160,14209,14219,14371,14405,14551,14614,14626,14678,15093,15122,15175,15226,15242,15500,15549,15579,15591,16364,16400,16536,16552,16571,16799,16843,16976,17071,17155,17310,17490,17622,17642,17664,17700,17903,17956,17981,18162,18416,18538,18782,18787,18888,18908,19119,19154,19161,19263,19488,19533,19715,20049,20094,20168,20344,20354,20559,20658,20707,20808,20827,21017,21169,21243,21306,21413,21581,21620,21649,21797,21845,22085,22251,22306,22493,22625,22674,22767,22822,22864,22908,22922,23201,23213,23449,23490,23766,24057,24108,24325,24363,24589,24646,24936,25028,25085,25089,25185,25216,25217,25262,25308,25406,25592,25605,25710,25797,26071,26139,26199,26242,26267,26394,26441,26563,26630,26718,26867,26882,27047,27073,27102,27315,27375,27464,27700,27808,27883,27984,28164,28459,28673,28747,28791,28814,28841,28847,28915,29030,29065,29110,29246,29249,29571,29847,29925,29928,29965,29971,30264,30960,31018,31453,31656,31783,31861,31948,32091,32102,32787,33262,33309,33572,33599,33842,34208,34326,34846,34991,35177,35689,35697,35708,35902,35950,36007,36136,36440,36447,36804,36823,37014,37111,37324,37357,37386,37562,37699,37940,38109,38116,38233,38504,38576,38806,38836,38898,39041,39058,39062,39253,39258,39264,39310,39321,39630,39667,39924,39938,39984,39987,39991,40317,40720,40781,40865,40984,41111,42092,42262,42344,43000,43169,43189,43251,43355,43387,43406,43454,43631,43948,43949,44016,44284,45323,45364,45386,45440,45818,45943,46219,46365,46801,46809,47178,47955,47979,48087,48112,48463,48858,49002,49558,49976,50005,50027,50371,50418,50437,50458,50556,50562,50734,50741,50872,50895,50925,50948,50988,51018,51134,51168,51460,51468,51543,51634,51649,51670,51688,51806,51822,51844,52056,52073,52174,52282,52433,52555,52600,52616,52677,52812,52877,52900,52983,53183,53252,53388,53463,53587,53721,53772,53915,53962,53991,54307,54345,54436,54466,54569,54881,54958,54974,55057,55258,55302,55345,55415,55496,55503,55659,55751,55820,55841,55880,55911,55935,56027,56047,56122,56223,56327,56366,56405,56635,56642,56756,56948,57002,57104,57112,57427,57566,57584,57589,57643,57669,57874,57980,58117,58157,58196,58308,58325,58503,58575,58829,58853,58876,58933,58947,59138,59145,59260,59275,59311,59346,59367,59390,59435,59505,59511,59527,59683,59743,59749,59794,59854,59879,59891,59977,59999,60017,60102,60106,60108,60169,60334,60379,60526,60847,60885,60959,60969,60988,60991,60996,61102,61174,61199,61252,61262,61383,61394,61408,61430,61442,61528,61664,61681,61815,61943,62055,62098,62129,62191,62351,62353,62362,62371,62422,62511,62516,62578,62652,62775,62877,62879,62935,62940,63010,63139,63150,63233,63315,63353,63377,63384,63418,63447,63490,63606,63705,63710,63913,63920,64109,64140,64188,64214,64240,64252,64288,64317,64444,64476,64553,64572,64595,64687,64760,64808,64887,64930,64933,65029,65210,65247,65450,65531,65575,65674,65695,65930,66051,66128,66223,66230,66377,66444,66527,66682,66784,66877,66980,67158,67250,67540,67579,67795,67828,67993,68030,68075,68209,68321,68397,68609,68645,68745,68784,68816,68930,69105,69140,69160,69208,69222,69441,69574,69594,69641,69663,69773,69798,69886,69955,69967,70027,70054,70133,70301,70351,70539,70662,70726,70844,70996,71039,71126,71209,71353,71406,71474,71501,71734,71754,71798,71804,72100,72117,72134,72159,72183,72206,72253,72275,72324,72335,72384,72573,72652,72691,72725,72884,72887,72941,73048,73120,73281,73372,73494,73533,73802,73855,73896,74309,74327,74447,74528,74583,74593,74699,74786,74890,74954,74972,74994,75015,75179,75183,75190,75537,75733,75760,75809,76002,76179,76337,76349,76361,76382,76475,76625,76771,76853,76922,76992,77059,77120,77165,77289,77401,77498,77583,77600,77617,77630,77648,77750,77847,77892,77898,77981,78048,78256,78290,78328,78336,78350,78407,78518,78546,78714,78734,78929,78963,78975,79112,79211,79246,79249,79263,79267,79434,79490,79491,79568,79585,79601,79618,79709,79871,79981,80184,80223,80271,80411,80584,80678,80844,80887,80914,80964,81071,81349,81358,81393,81434,81467,81521,81574,81611,81841,81928,81936,82038,82237,82254,82267,82311,82330,82356,82387,82479,82508,82542,82694,82707,82851,82885,82925,82940,83087,83391,83546,83642,83661,83676,83744,83824,83910,83922,84062,84172,84229,84234,84251,84533,84539,84553,84640,84699,84880,84946,85014,85047,85050,85056,85099,85213,85237,85243,85387,85486,85659,85687,85830,86055,86264,86272,86304,86333,86499,86552,86579,86894,86903,86927,86954,87112,87156,87237,87486,87572,87659,87684,87695,87736,87775,87776,87851,87882,88044,88102,88178,88219,88338,88612,88662,88802,88836,88847,88908,88938,88985,89183,89197,89262,89306,89418,89453,89549,89566,89591,89614,89707,89722,89744,89746,89756,89808,90095,90391,90423,90505,90538,90699,90729,90779,90825,90874,90876,91118,91370,91392,91521,91618,91679,91739,91877,91890,91892,91914,91923,92124,92210,92294,92373,92414,92619,92633,92822,92930,93014,93100,93193,93273,93338,93397,93408,93805,94083,94090,94258,94295,94371,94552,94604,94684,94745,94804,94915,94941,94981,95028,95067,95106,95130,95252,95301,95351,95352,95354,95489,95555,95667,95686,95722,95846,96089,96165,96258,96490,96554,96563,96660,96888,96919,96931,97039,97070,97171,97269,97320,97355,97358,97368,97371,97416,97638,97665,98065,98135,98225,98242,98267,98291,98488,98527,98545,98656,98712,98716,98851,98875,98907,98963,99003,99131,99146,99221,99249,99346,99482,99540,99561,99573,99667,99804,99823,99851,99949",
          "168,210,1972,2029,2043,2162,2179,2428,2439,2539,2579,2647,2664,2713,2736,3198,3277,3281,3301,3358,3377,3492,3530,3549,3563,3589,3659,3665,3722,3738,4005,4177,4189,4438,4619,4636,4648,4712,4763,4838,5719,6395,7340,7454,7516,9477,9478,9496,9553,9572,9591,9629,9648,9667,9686,9705,9743,9781,9800,9833,9857,10016,10028,10047,10123,10142,10522,10598,10681,10684,10702,10712,10731,10750,10769,10807,10826,10845,10901,10902,10978,11000,11035,11111,11130,11225,11263,11320,11358,11415,11421,11491,11586,11624,11643,11662,11852,11871,11928,11966,12023,12099,12118,12132,12137,12175,12213,12232,12251,12308,12327,12346,12418,12441,12492,12555,12579,12612,12631,12652,12691,12707,12729,12731,12745,12764,12818,12851,12859,12897,12934,12954,13011,13049,13068,13071,13125,13161,13163,13193,13220,13258,13277,13296,13329,13340,13448,13569,13581,13657,13676,13714,13733,13752,13771,13790,13809,13847,13866,13923,13942,14018,14037,14093,14094,14113,14148,14170,14208,14246,14284,14360,14369,14379,14464,14508,14516,14531,14561,14584,14647,14664,14687,14702,14721,14730,14753,14767,14785,14797,14816,14851,14892,14930,14949,14987,15044,15063,15101,15120,15177,15196,15234,15253,15310,15348,15367,15386,15424,15443,15462,15481,15519,15538,15557,15563,15633,15652,15665,15671,15709,15770,15823,15842,15876,15899,15918,15937,15956,15994,16027,16032,16042,16051,16070,16089,16146,16165,16222,16241,16279,16298,16317,16336,16355,16374,16393,16423,16431,16469,16488,16526,16545,16559,16583,16697,16716,16735,16761,16777,16792,16849,16906,16908,16944,17001,17020,17058,17077,17084,17115,17153,17172,17210,17248,17267,17286,17343,17362,17381,17400,17453,17456,17457,17495,17533,17552,17566,17609,17638,17645,17647,17704,17742,17761,17799,17837,17894,17913,17951,17970,18008,18027,18074,18103,18141,18160,18217,18236,18255,18312,18326,18331,18369,18388,18407,18426,18445,18464,18475,18502,18537,18540,18559,18578,18589,18597,18616,18673,18692,18730,18804,18806,18844,18875,18901,18920,18977,19015,19072,19091,19110,19129,19159,19162,19168,19205,19224,19262,19300,19338,19346,19357,19395,19414,19433,19452,19490,19509,19566,19585,19623,19642,19661,19680,19684,19699,19737,19764,19775,19794,19832,19851,19870,19889,19908,19927,19946,19965,19984,20079,20098,20117,20162,20212,20269,20307,20326,20345,20383,20402,20421,20440,20459,20478,20516,20534,20535,20554,20573,20592,20611,20630,20638,20687,20725,20782,20801,20805,20820,20839,20872,20877,20915,20934,20953,20955,20982,20991,21048,21105,21110,21124,21143,21162,21181,21238,21254,21276,21295,21324,21333,21352,21390,21428,21433,21447,21466,21485,21504,21523,21542,21580,21618,21637,21675,21686,21713,21732,21751,21770,21789,21846,21865,21884,21903,21922,21941,21960,21979,21998,22036,22055,22112,22169,22188,22245,22264,22283,22321,22359,22378,22397,22416,22419,22435,22473,22511,22530,22549,22587,22606,22644,22663,22673,22682,22696,22701,22720,22730,22739,22777,22806,22834,22842,22853,22891,22910,22948,22967,22994,23005,23024,23081,23119,23157,23214,23218,23233,23271,23290,23319,23328,23365,23385,23404,23442,23480,23518,23537,23594,23613,23651,23670,23708,23727,23746,23784,23822,23841,23879,23917,23936,23955,23974,23993,24012,24031,24050,24069,24084,24088,24148,24183,24202,24278,24297,24316,24373,24449,24468,24487,24506,24525,24544,24563,24582,24594,24620,24639,24640,24658,24732,24734,24753,24772,24810,24829,24867,24924,24943,24948,24981,25019,25038,25067,25080,25095,25114,25133,25179,25209,25247,25266,25274,25285,25304,25323,25399,25402,25456,25475,25494,25513,25532,25570,25623,25629,25671,25684,25722,25741,25760,25798,25834,25836,25912,25918,25931,25950,25969,25977,26007,26026,26045,26064,26083,26102,26121,26140,26178,26197,26209,26216,26273,26292,26368,26378,26425,26501,26539,26596,26653,26786,26805,26836,26848,26900,26976,27004,27017,27090,27099,27120,27166,27204,27223,27231,27261,27280,27451,27546,27648,27698,27701,27777,27784,27823,27891,27926,28059,28066,28154,28173,28192,28230,28240,28249,28268,28281,28317,28325,28363,28382,28458,28477,28515,28686,28800,28857,28866,28876,28895,28933,28952,29012,29142,29199,29247,29256,29275,29294,29370,29408,29427,29446,29454,29474,29522,29525,29582,29597,29696,29731,29738,29769,29777,29871,29883,29921,29941,29968,29993,29997,30025,30054,30083,30112,30141,30170,30199,30228,30257,30286,30315,30344,30373,30402,30431,30460,30489,30518,30547,30576,30605,30634,30663,30692,30721,30750,30779,30808,30837,30866,30895,30924,30953,30982,31011,31040,31069,31098,31127,31156,31185,31214,31243,31272,31301,31330,31359,31388,31417,31446,31475,31504,31533,31562,31591,31620,31649,31678,31707,31736,31765,31794,31823,31852,31881,31910,31939,31968,31997,32026,32055,32084,32113,32142,32171,32200,32229,32258,32287,32316,32345,32374,32403,32432,32461,32490,32519,32548,32577,32606,32635,32664,32693,32722,32751,32780,32809,32838,32867,32896,32925,32954,32983,33012,33041,33070,33099,33128,33157,33186,33215,33244,33273,33302,33331,33360,33389,33418,33447,33476,33505,33534,33563,33592,33621,33650,33679,33708,33737,33766,33795,33824,33853,33882,33911,33940,33969,33998,34027,34056,34085,34114,34143,34172,34201,34230,34259,34288,34317,34346,34375,34404,34433,34462,34491,34520,34549,34578,34607,34636,34665,34694,34723,34752,34781,34810,34839,34868,34897,34926,34955,34984,35025,35058,35114,35145,35188,35209,35222,35249,35277,35351,35384,35385,35392,35399,35425,35496,35544,35574,35603,35662,35669,35681,35735,35794,35853,35860,35924,35983,36042,36055,36105,36125,36165,36224,36248,36295,36354,36413,36485,36544,36603,36609,36667,36726,36737,36785,36790,36799,36856,36915,36930,36974,36985,37006,37046,37065,37105,37164,37171,37228,37287,37311,37346,37375,37417,37476,37535,37607,37666,37682,37700,37725,37740,37743,37789,37848,37907,37920,37978,37998,38037,38052,38096,38105,38168,38227,38249,38286,38291,38350,38369,38409,38415,38468,38539,38598,38657,38683,38729,38739,38788,38847,38854,38911,38930,38970,39029,39035,39044,39052,39100,39159,39218,39240,39290,39349,39408,39421,39472,39506,39531,39553,39590,39661,39720,39739,39779,39826,39851,39862,39910,39969,39997,40119,40148,40177,40206,40235,40264,40293,40322,40351,40380,40409,40438,40467,40496,40525,40554,40583,40612,40641,40670,40699,40728,40757,40786,40815,40844,40873,40902,40931,40960,40989,41018,41047,41076,41105,41134,41163,41192,41221,41250,41279,41308,41337,41366,41395,41424,41453,41482,41511,41540,42010,42039,42068,42097,42126,42186,42235,42249,42276,42314,42315,42351,42395,42426,42513,42570,42588,42608,42621,42663,42690,42750,42770,42825,42869,42900,42920,42963,42987,43007,43046,43062,43081,43114,43137,43224,43281,43299,43320,43374,43408,43426,43427,43461,43491,43512,43536,43571,43579,43611,43635,43649,43698,43721,43773,43805,43828,43848,43866,43915,43935,43992,43994,44010,44085,44106,44172,44191,44199,44231,44247,44270,44279,44287,44290,44322,44368,44389,44409,44484,44542,44559,44571,44600,44646,44697,44704,44721,44740,44752,44796,44833,44859,44864,44883,44924,44937,44958,44969,45033,45046,45069,45106,45119,45154,45165,45215,45273,45285,45327,45346,45371,45403,45461,45533,45590,45648,45701,45720,45777,45799,45804,45835,45907,45964,46019,46087,46137,46170,46242,46249,46254,46261,46267,46313,46326,46347,46355,46411,46462,46479,46495,46543,46579,46602,46622,46668,46673,46675,46682,46687,46765,46781,46851,46920,46966,46972,47014,47032,47058,47060,47064,47136,47154,47193,47196,47229,47246,47272,47278,47297,47333,47335,47422,47439,47450,47462,47473,47502,47505,47577,47578,47612,47651,47724,47820,47825,47831,47835,47853,47877,47882,47892,47973,48021,48031,48079,48081,48093,48143,48162,48167,48220,48243,48248,48251,48271,48281,48287,48343,48353,48390,48400,48418,48419,48422,48438,48461,48487,48549,48616,48694,48717,48758,48773,48815,48859,48861,48889,48893,48985,49041,49055,49107,49183,49234,49268,49283,49300,49336,49345,49355,49413,49476,49482,49496,49621,49624,49644,49658,49671,49721,49775,49778,49870,49883,49902,49925,49933,49977,49986,50019,50161,50167,50299,50300,50342,50365,50374,50471,50512,50520,50643,50662,50758,50930,51015,51055,51160,51221,51357,51360,51492,51631,51825,51867,51909,51910,51929,52061,52255,52358,52496,52505,52578,52766,52928,52942,52966,53000,53051,53134,53136,53139,53262,53323,53354,53394,53404,53543,53737,53758,53844,53955,54041,54108,54357,54396,54558,54622,54732,54741,54845,54950,55040,55199,55275,55320,55334,55360,55400,55527,55586,55613,55656,55711,55774,55830,55895,55920,55985,56282,56309,56610,56733,56771,56776,57118,57189,57198,57277,57313,57317,57387,57394,57486,57548,57602,57640,57707,57833,58035,58088,58118,58147,58343,58355,58360,58400,58449,58507,58597,58612,58641,58704,58779,58790,58888,59161,59165,59172,59216,59310,59429,59785,59793,59819,60067,60181,60616,60824,60831,60841,60990,61043,61348,61558,61698,61772,61883,61905,61922,61964,62030,62056,62152,62161,62397,62402,62421,62459,62477,62490,62604,62698,62802,62806,62830,62975,63050,63064,63079,63080,63236,63348,63378,63498,63609,63746,63836,64008,64169,64184,64187,64469,64504,64598,64770,64798,64804,64941,65079,65096,65115,65144,65149,65211,65272,65283,65287,65332,65536,65615,65627,65891,65939,65950,65951,66024,66127,66198,66218,66231,66318,66352,66364,66379,66478,66610,66632,66640,66691,66808,66884,66906,66915,66953,66977,66979,66984,67032,67128,67132,67140,67141,67297,67298,67310,67322,67530,67716,67839,67925,67962,67981,68245,68509,68539,68734,68804,68809,68835,68838,68854,68896,68903,68951,68974,69031,69097,69169,69175,69424,69436,69478,69616,69664,69704,69841,69851,69910,69971,69995,70029,70248,70270,70360,70456,70494,70495,70823,70866,70981,70985,71051,71225,71250,71266,71333,71493,71532,71633,71698,71704,71735,71773,71818,71902,71928,71999,72185,72219,72249,72366,72549,72597,72609,72626,72660,72727,72809,72896,72961,73033,73063,73525,73579,73778,73867,73876,74054,74171,74224,74283,74288,74370,74420,74501,74669,74821,74901,74910,74953,74975,74987,75024,75068,75107,75495,75587,75653,75686,75742,75793,75796,76005,76203,76270,76305,76331,76385,76537,76538,76618,76630,76707,76842,76857,76894,77045,77205,77222,77309,77340,77373,77385,77395,77417,77566,77604,77607,77709,77739,77798,77824,77838,77845,77871,77884,77897,78016,78037,78141,78227,78228,78565,78735,78785,78955,78985,79151,79185,79192,79264,79401,79444,79565,79733,79976,80029,80086,80217,80232,80402,80445,80464,80511,80530,80660,80776,80778,80837,81091,81168,81727,81735,81948,81996,82076,82165,82214,82299,82322,82651,82703,82767,82801,82815,83026,83071,83185,83197,83255,83334,83431,83442,83524,83617,83707,83723,83907,83925,83957,83964,83968,84038,84066,84119,84174,84264,84265,84369,84400,84467,84471,84572,84631,84651,84693,84852,84866,84877,84908,84937,85088,85272,85311,85371,85404,85411,85493,85503,85608,85625,85635,85678,85697,85725,85752,85794,85835,85927,86461,86580,86773,86910,86930,87019,87271,87312,87360,87373,87425,87563,87605,87645,87661,88002,88255,88291,88407,88508,88535,88701,88731,88767,88827,88831,88896,88918,88927,88995,89193,89203,89216,89249,89452,89463,89496,89759,89792,89810,89887,89975,90051,90142,90328,90350,90464,90602,90638,90661,91092,91096,91143,91321,91363,91536,91695,91868,91918,92148,92208,92383,92463,92475,92511,92629,92658,92689,92737,92821,92861,92907,92963,93119,93146,93229,93260,93288,93563,93597,93607,93770,93788,93799,93820,93885,93950,94066,94098,94129,94289,94316,94388,94678,94797,94844,94903,95064,95065,95100,95132,95218,95391,95432,95499,95517,95566,95679,95855,95891,95898,95902,95950,96026,96040,96043,96099,96147,96194,96265,96267,96385,96436,96516,96535,96695,96730,96849,96956,97102,97302,97357,97390,97543,97587,97672,97885,97968,98235,98493,98508,98549,98559,98584,98634,98643,98879,98919,98953,99090,99241,99264,99342,99426,99434,99589,99668,99674,99862,99905,99945",
          "22,170,384,620,904,1022,1070,1107,1426,1498,1618,1651,1701,1707,1820,1872,2064,2215,2221,2237,2271,2343,2518,2560,2779,2819,2848,3053,3123,3384,3428,3542,3561,3607,3860,3942,4331,4336,4371,4762,4938,5050,5149,5197,5210,5223,5252,5380,5445,5474,5495,5506,5600,5743,5753,5807,5916,5943,6002,6192,6305,6373,6409,6441,6485,6511,6542,6645,6665,6667,6683,6729,6758,6804,6808,6951,6996,7002,7162,7186,7199,7207,7424,7528,7538,7574,7606,7717,7950,7979,8066,8201,8495,8496,8507,8583,8596,8604,8814,8894,8953,8963,9033,9082,9093,9112,9129,9190,9283,9386,9463,9717,9751,9762,9841,9994,10000,10027,10199,10203,10327,10338,10353,10383,10432,10438,10484,10514,10515,10535,10602,10668,10766,10811,10868,10949,11583,11603,11666,11689,11706,11717,11777,11825,11935,12017,12192,12235,12372,12447,12567,12727,12867,12913,13070,13087,13104,13184,13257,13410,13453,13585,13586,13656,13684,13694,13751,13878,13879,13959,13991,14014,14046,14114,14154,14240,14244,14254,14321,14358,14688,14876,15128,15165,15181,15261,15351,15406,15521,15587,15717,15732,15785,15898,15908,15935,15941,15977,16074,16188,16192,16262,16287,16523,16666,16679,16687,16705,16772,16788,16979,17003,17027,17291,17369,17384,17535,17665,17752,18085,18143,18197,18371,18583,18728,18814,18853,18933,18969,18976,19017,19216,19220,19523,19549,19608,19647,20308,20480,20486,20634,20787,20838,20997,21034,21153,21161,21240,21369,21487,21690,21929,21989,22165,22329,22452,22483,22551,22578,22608,22623,22632,22660,22732,22833,22874,22905,22983,23098,23174,23399,23405,23459,23563,23622,23714,23785,23810,23952,23953,24020,24106,24121,24228,24314,24390,24552,24585,24795,24814,24849,24909,25001,25009,25042,25100,25155,25200,25244,25340,25648,25664,25669,25731,25761,25855,25987,26180,26476,26582,26625,26659,26725,26736,26751,26770,26784,26880,26933,27092,27098,27271,27328,27336,27398,27401,27416,27437,27653,27667,27798,27845,27939,28315,28328,28384,28399,28578,28597,28627,28688,28690,28759,28809,28870,28878,28946,29096,29123,29161,29178,29311,29328,29348,29432,29440,29464,29495,29518,29720,29750,29844,29866,30116,30191,30546,30554,30945,30952,30984,31424,31622,31880,31883,31908,32344,32440,32662,33107,33159,33277,33449,34199,34512,34808,34867,34899,35129,35190,35198,35505,35559,35640,35700,35743,35897,35916,35956,36011,36059,36104,36256,36265,36412,36601,36613,36630,36671,36692,36818,37009,37010,37024,37318,37358,37388,37448,37534,37611,37614,37680,37688,37818,37840,37847,37980,38008,38046,38147,38170,38195,38295,38321,38322,38387,38864,38868,38876,38885,38932,38933,38945,39116,39297,39317,39355,39446,39551,39812,39914,40075,40158,40377,41058,41278,41289,41996,42063,42437,43209,43271,43398,43410,43675,43764,43784,43847,43916,43917,44187,44212,44496,44772,44835,44943,45248,45306,45374,45383,45460,45501,45503,45557,45789,45806,45949,46180,46217,46421,46521,46780,47053,47071,47076,47162,47434,47566,47641,47676,47739,47744,47959,48253,48523,48574,48592,48687,48834,48997,49290,49423,49483,50006,50059,50101,50190,50704,50789,50822,50846,50888,51017,51050,51115,51131,51132,51135,51167,51225,51232,51241,51261,51374,51416,51475,51556,51610,51669,51922,51924,51959,51977,52075,52117,52127,52215,52429,52498,52587,52592,52598,52605,52708,52984,53034,53222,53244,53280,53407,53455,53469,53724,54067,54099,54203,54210,54411,54596,54675,54690,54731,54894,54913,54962,54990,54998,55048,55128,55182,55234,55252,55386,55408,55421,55490,55500,55599,55667,55816,55924,56001,56042,56050,56142,56153,56180,56206,56243,56297,56311,56535,56585,56660,56680,56690,57067,57115,57176,57197,57284,57302,57348,57402,57404,57506,57617,57704,57851,57942,58009,58052,58062,58064,58145,58190,58385,58630,58658,58711,58782,58930,58963,58986,59000,59032,59053,59201,59229,59233,59360,59430,59454,59516,59647,59662,59874,59880,60078,60081,60374,60415,60421,60457,60459,60495,60546,60611,60759,60871,60882,61044,61086,61093,61170,61213,61268,61592,61631,61684,61713,61754,61765,61875,61944,62022,62275,62304,62316,62447,62476,62480,62502,62554,62597,62749,62752,62807,62852,63023,63031,63052,63154,63174,63368,63497,63661,63697,63714,63767,63801,63831,63903,63919,64079,64318,64481,64574,64613,64782,64832,64888,64921,65127,65130,65330,65420,65471,65512,65528,65638,65649,65686,65739,65787,65871,66036,66150,66157,66180,66264,66404,66501,66536,66609,66645,67212,67346,67365,67372,67556,67571,67668,67775,67805,67827,67851,67967,68205,68225,68337,68343,68424,68494,68706,68748,68794,68937,68940,69055,69127,69246,69262,69268,69351,69352,69414,69446,69474,69532,69659,69670,69740,69843,69954,69977,70066,70070,70178,70355,70383,70572,70645,70680,70850,70922,71114,71208,71369,71434,71457,71514,71618,71701,71900,71906,71926,72086,72226,72235,72238,72298,72446,72610,72702,72907,73042,73143,73183,73288,73305,73469,73543,73684,73741,73844,73900,73943,73979,74006,74027,74113,74118,74148,74244,74328,74428,74545,74691,74803,74827,74861,75148,75163,75188,75240,75279,75309,75365,75440,75931,75993,76048,76252,76274,76350,76451,76487,76823,76914,76934,77021,77087,77113,77145,77237,77311,77494,77511,77517,77608,77615,77702,77780,77791,77864,77870,77887,77987,78031,78084,78091,78211,78293,78421,78438,78560,78570,78579,78607,78648,78711,78727,78877,79022,79031,79336,79344,79415,79450,79556,79575,79874,79970,80023,80227,80228,80240,80611,80630,80857,80904,80957,81055,81236,81258,81299,81462,81658,81664,81676,81733,81788,81802,82040,82081,82139,82280,82312,82336,82399,82615,82637,82775,82784,82896,82923,82938,83005,83011,83052,83240,83271,83361,83395,83463,83482,83492,83530,83705,83747,83905,84027,84101,84137,84339,84343,84584,84709,84773,84801,84970,85039,85152,85207,85234,85241,85258,85309,85314,85331,85400,85434,85533,85542,85557,85660,85743,85772,85783,85792,85840,85909,86062,86170,86424,86471,86600,86607,86850,86866,87075,87275,87380,87383,87419,87461,87477,87524,87584,87586,87598,87760,87817,87863,87887,87902,87995,88208,88235,88317,88549,88638,88642,88865,88890,88926,88957,89075,89166,89194,89238,89273,89384,89412,89826,89842,89865,89963,90072,90109,90133,90249,90371,90410,90583,90629,90657,90701,90758,90782,90793,90844,90853,91068,91131,91283,91297,91388,91405,91465,91581,91597,91797,91809,91907,92008,92197,92324,92403,92521,92548,92597,92607,92641,92645,92748,92800,92808,92890,92895,92975,93055,93075,93106,93224,93238,93314,93355,93356,93402,93591,93611,93637,93765,93835,94014,94082,94099,94140,94170,94260,94321,94334,94528,94820,94834,94880,95058,95268,95322,95337,95424,95553,95658,95793,95966,96034,96087,96115,96283,96545,96612,96677,96678,96819,97056,97063,97101,97127,97181,97213,97254,97278,97483,97595,97646,97662,97823,97832,97847,97868,97904,97921,97995,98127,98183,98197,98415,98459,98490,98546,98570,98596,98598,98707,98765,98889,98900,99044,99059,99064,99108,99127,99274,99501,99531,99835,99897,99923,99953",
          "17,19,30,520,548,552,556,571,636,645,664,677,681,689,706,750,785,816,818,859,920,981,1046,1120,1147,1155,1260,1274,1277,1305,1336,1338,1428,1469,1492,1539,1582,1585,1591,1696,1828,1842,1875,2392,4073,5038,5123,5173,5180,5192,5200,5222,5226,5242,5244,5280,5286,5291,5348,5364,5366,5406,5407,5419,5442,5473,5475,5500,5537,5624,5652,5688,5689,5709,5713,5728,5730,5810,5864,5922,5925,5979,6049,6063,6150,6155,6180,6214,6309,6319,6343,6349,6391,6453,6491,6577,6717,6726,6734,6748,6777,7097,7149,7179,7192,7335,7346,7364,7369,7408,7590,7685,7702,7885,7892,7961,7970,7987,8150,8170,8419,8449,8538,8576,8633,8685,8703,8719,8766,8863,8880,8912,9127,9187,9204,9287,9335,9941,10040,10163,10253,10305,10307,10334,10352,10408,10420,10460,10525,10613,10849,10974,10995,11177,11206,11690,11732,11754,11837,12125,14950,15352,16857,19493,22839,23086,23176,24275,24508,24514,24613,24641,24698,24808,25447,26802,28120,28232,29148,29239,29746,30720,31145,31606,35578,36490,38560,41667,42259,42670,42791,45164,45381,46505,48990,50017,50171,50200,50334,50481,50529,50571,50582,50721,50776,50783,50870,50899,50924,51187,51212,51273,51339,51439,51586,51592,51639,51750,51811,51820,51838,51878,52143,52275,52329,52393,52458,52593,52686,52721,52871,53019,53023,53127,53176,53197,53290,53304,53423,53541,53581,53769,53828,53993,54093,54188,54196,54266,54280,54779,55035,55044,55084,55148,55171,55276,55291,55387,55481,55548,55760,55770,55780,55821,55845,56049,56073,56077,56140,56257,56495,56534,56712,56719,56751,56769,56791,56929,56949,57077,57081,57083,57364,57717,57943,57954,57967,58002,58100,58109,58168,58234,58376,58468,58499,58500,58809,58813,58890,58959,59124,59180,59468,59479,59507,59526,59559,59681,60180,60206,60294,60344,60370,60496,60508,60529,60550,60629,60644,60873,61041,61047,61165,61185,61236,61247,61258,61275,61332,61358,61362,61366,61559,61683,61802,61819,61874,61885,61991,62113,62336,62377,62521,62558,62579,62612,62659,62673,62686,62815,62866,62874,62930,63034,63231,63360,63426,63579,63669,63706,63725,63816,63890,64010,64013,64076,64161,64176,64346,64408,64441,64500,64606,64662,64733,64769,64904,64926,64984,65175,65207,65356,65404,65414,65433,65560,65593,65658,65740,65901,65986,66053,66072,66089,66111,66147,66185,66189,66194,66208,66229,66315,66351,66485,66487,66597,66760,66922,66960,66989,67060,67087,67168,67305,67394,67538,67615,67706,67744,67754,67847,67982,68022,68081,68108,68111,68290,68466,68473,68607,68663,68704,68714,68963,68978,68979,69091,69302,69326,69330,69363,69430,69445,69477,69805,69861,69923,69929,70114,70267,70349,70419,70510,70586,70725,70790,70838,70966,71091,71107,71206,71326,71328,71431,71620,71730,71788,71937,71938,71970,72063,72187,72201,72221,72262,72271,72513,72571,72580,72748,72915,72916,72925,73113,73239,73329,73441,73442,73568,73631,73703,73794,73809,74008,74114,74172,74180,74193,74362,74382,74411,74643,74717,74754,74850,74868,74891,74900,74970,75041,75136,75153,75161,75357,75613,75896,76021,76022,76114,76396,76611,76646,76731,76746,76787,76861,76895,76904,77011,77041,77166,77198,77229,77384,77644,77831,78039,78111,78143,78152,78242,78303,78331,78419,78446,78465,78567,78619,78685,78708,78779,78813,78948,79147,79166,79207,79239,79286,79328,79353,79424,79446,79591,79623,79648,79765,79877,80138,80140,80146,80398,80440,80559,80694,80784,80792,81044,81089,81118,81167,81342,81402,81470,81504,81591,81661,81674,81745,81849,81901,81962,82027,82168,82279,82567,82601,82619,82731,82853,82933,83009,83051,83070,83073,83090,83273,83627,83636,83683,83774,83819,83870,83900,84042,84344,84398,84521,84767,84807,84943,84948,85036,85043,85101,85257,85306,85401,85663,85705,85706,85726,85785,85883,86028,86037,86202,86228,86324,86352,86371,86488,86542,86649,86733,86815,86821,86855,86990,87047,87359,87400,87404,87470,87512,87619,87666,87788,87807,87899,88053,88345,88365,88383,88393,88450,88693,88726,88951,88992,89021,89201,89204,89285,89390,89415,89489,89582,89653,89676,89749,89788,89834,89896,89998,90007,90037,90045,90102,90117,90155,90157,90220,90326,90457,90558,90577,90679,90714,90790,90830,91013,91052,91138,91237,91248,91346,91471,91547,91559,91656,91787,91887,91962,92318,92368,92618,92663,92675,92771,92847,92874,93095,93214,93218,93475,93697,93771,93790,93833,93900,94004,94041,94296,94331,94353,94393,94490,94532,94563,94979,95074,95269,95274,95276,95385,95490,95523,95631,95672,95796,95988,96107,96226,96245,96390,96428,96506,96513,96557,96665,96754,96851,96944,96959,96968,97041,97098,97227,97504,97586,97780,98091,98124,98134,98412,98444,98772,98890,99196,99240,99260,99324,99366,99479,99514,99519,99566,99756,99825,99960",
          "277,334,386,406,423,444,465,480,515,553,568,595,621,643,670,723,734,792,817,847,873,989,1021,1043,1050,1068,1141,1175,1327,1353,1363,1386,1404,1432,1437,1448,1477,1563,1636,1791,1833,1845,1903,1954,2943,4639,5144,5234,5254,5330,5338,5700,5849,5867,5918,6040,6135,6139,6154,6158,6197,6265,6410,6533,6813,6941,7121,7202,7216,7222,7229,7238,7339,7453,7571,7579,7644,7687,7743,7783,7787,7914,7992,8036,8072,8140,8190,8392,8519,8546,8635,8653,8682,8742,8885,9040,9065,9111,9121,9122,9275,9284,9351,9453,9862,9874,10184,10205,10241,10417,10424,10623,10657,10805,10907,10931,10948,11073,11090,11126,11336,11566,11632,11714,11769,11835,13724,13865,14089,14096,14372,14535,14611,14705,14802,14915,15330,15442,16130,16263,16599,17264,17436,17568,17588,17667,17966,17975,18122,18337,18594,19494,19681,20266,20296,20674,21050,21639,22612,22830,22850,22932,22984,24280,25132,25485,26422,26919,27025,27242,27491,27497,27791,28158,28273,29371,29405,29485,29733,29737,29996,30511,33930,33985,35844,36673,38746,39820,39905,39979,40677,41300,42336,43120,44121,45936,46999,47254,49604,49874,50058,50117,50128,50180,50193,50288,50317,50443,50497,50534,50631,50632,50744,50796,50863,50889,51047,51361,51421,51432,51444,51514,51551,51612,51623,51630,51636,51694,51754,51796,51845,52009,52119,52150,52269,52319,52347,52360,52404,52448,52461,52485,52535,52601,52606,52657,52730,52888,52913,52972,53146,53164,53232,53328,53334,53381,53466,53750,53814,54010,54018,54052,54255,54296,54420,54478,54719,54834,54877,54952,54985,55007,55394,55513,55525,55718,55746,55928,56230,56320,56321,56369,56436,56517,56553,56558,56602,56707,56740,56919,56966,57042,57072,57095,57173,57694,57748,57935,58108,58379,58406,58473,58631,58895,58919,58937,58970,59023,59295,59306,59319,59745,59832,59885,59902,60062,60083,60103,60171,60311,60351,60509,60582,60760,60846,60931,60934,60981,61057,61127,61266,61303,61305,61405,61586,61620,61668,61702,61882,61952,62261,62322,62324,62436,62539,62557,62573,62586,62646,62729,62750,62758,62810,62882,62911,63122,63151,63187,63209,63311,63355,63364,63591,64049,64245,64268,64340,64732,64751,64791,65034,65110,65163,65256,65359,65372,65384,65496,65561,65694,65782,65791,65879,65983,66011,66118,66263,66354,66445,66474,66504,66607,66652,66778,66933,66961,67133,67181,67232,67234,67237,67542,67577,67724,68002,68138,68301,68323,68367,68411,68485,68516,68580,68616,68634,68702,68860,68871,68874,68880,68881,68908,68946,68975,69032,69053,69061,69073,69178,69364,69565,69808,69907,69921,70019,70093,70161,70325,70531,70575,70635,70872,70882,70883,70970,71011,71087,71271,71286,71293,71443,71516,71577,71631,71669,71770,71882,71899,71953,72024,72193,72195,72211,72407,72444,72541,72592,72621,72731,72743,72905,72987,73116,73162,73196,73423,73725,73726,74392,74554,74680,74761,74928,74960,75000,75037,75144,75171,75175,75303,75395,75460,75499,75604,75691,75696,75713,75774,75821,75903,76088,76113,76143,76530,76580,76606,76676,76683,76754,76803,76863,76949,76960,77002,77016,77134,77238,77274,77338,77355,77433,77489,77528,77688,77785,77908,77946,78014,78186,78229,78259,78292,78422,78669,78763,78861,79016,79063,79089,79247,79339,79370,79406,79468,79504,79539,79580,79644,79645,79662,79784,79915,80020,80121,80176,80214,80296,80321,80415,80526,80573,80635,80692,80799,80851,81104,81175,81625,81635,81688,81721,81899,81913,81998,82006,82058,82162,82248,82323,82466,82505,82560,82723,82900,82965,83008,83110,83237,83330,83396,83404,83559,83593,83655,83679,83728,83783,83840,83986,84079,84197,84290,84291,84388,84390,84430,84603,84758,84925,85004,85073,85086,85300,85323,85394,85407,85455,85497,85674,85707,85781,85850,85863,85999,86022,86268,86490,86500,86591,86623,86712,86748,86798,86824,86870,86999,87045,87062,87102,87511,87580,87646,87748,87782,87870,87967,87986,88009,88159,88214,88318,88440,88518,88527,88637,88765,88833,88867,88964,88996,89054,89084,89161,89357,89367,89395,89417,89550,89587,89638,89986,89994,90017,90080,90233,90310,90432,90465,90531,90715,90849,91161,91250,91312,91323,91337,91418,91601,91749,91853,91871,91888,92021,92084,92103,92193,92229,92367,92429,92434,92455,92491,92500,92638,92721,92738,92829,92854,92926,92966,93056,93097,93213,93281,93301,93382,93423,93500,93506,93570,93696,93704,93755,93813,93920,94241,94343,94557,94573,94837,94882,94969,95005,95009,95039,95071,95089,95109,95137,95161,95190,95407,95497,95506,95619,95627,95719,95729,95778,95808,95899,95973,95976,95994,95997,96071,96130,96244,96263,96284,96301,96348,96402,96406,96434,96671,96817,96909,97179,97239,97353,97386,97414,97477,97618,97679,97950,98043,98051,98054,98092,98130,98168,98253,98358,98457,98469,98703,98777,98869,98884,99097,99099,99306,99329,99350,99352,99374,99448,99657,99706,99820,99879,99900",
          "297,338,341,356,364,486,559,584,585,615,638,656,711,811,890,957,998,1083,1105,1111,1156,1224,1262,1290,1470,1521,1616,1727,1756,1856,1861,2095,2111,2193,2228,2441,2559,2797,2801,2835,2939,2949,3017,3182,3276,3312,3340,3343,3495,3602,3620,3733,3766,3774,4123,4151,4183,4341,4411,4692,4693,4727,4739,4880,4896,5098,5189,5270,5340,5375,5440,5453,5494,5543,5573,5590,5643,5673,5698,5802,5830,5939,5985,6075,6102,6137,6165,6276,6474,6608,6669,6732,6847,7001,7064,7070,7083,7240,7241,7282,7418,7490,7605,7695,8068,8142,8469,8603,8733,8850,8883,8977,9053,9126,9138,9144,9173,9265,9370,9411,9455,9773,10087,10156,10182,10335,10463,10592,10651,10700,10705,10748,10815,10848,10927,10962,10971,11076,11115,11149,11165,11345,11349,11434,11493,11503,11510,11579,11608,11684,11809,11830,12009,12047,12067,12165,12292,12306,12382,12622,12736,12812,12937,13031,13133,13355,13409,13450,13472,13497,13542,13544,13673,13871,13920,14060,14117,14261,14686,14696,14856,14864,15243,15256,15331,15531,16049,16113,16172,16419,16437,16491,16625,16764,16846,16891,17160,17189,17494,17583,17707,17708,18608,18643,18648,18675,18680,18740,18750,18856,19007,19122,19214,19505,19619,19639,19771,19843,20201,20249,20490,20751,20841,20996,21038,21361,21540,21658,21683,21711,21782,21890,22140,22258,22613,22927,22968,22991,23087,23189,23247,23376,23561,23566,23593,23649,23666,23689,23776,23805,24109,24134,24313,24338,24532,24608,24733,25321,25326,25600,25801,25881,25940,26333,26684,26690,26729,26798,26874,26981,27046,27195,27323,27340,27355,27441,27644,27688,27767,27923,28221,28294,28470,28555,28765,28842,28846,28889,28919,28982,29128,29165,29171,29185,29211,29255,29312,29316,29349,29382,29603,29776,29810,29840,29846,30032,30130,30488,30913,30990,32315,32434,32450,32566,33628,34103,34121,34480,34832,35092,35556,35618,35705,35709,35740,35755,35997,36012,36259,36267,36383,36686,36888,37019,37035,37066,37114,37127,37138,37305,37564,37636,37802,37807,37939,38009,38085,38190,38229,38669,38699,38812,38877,38931,38986,39119,39501,39687,39841,39876,39989,39994,40362,40547,40589,40618,40734,40799,40826,40944,41350,41401,42454,42463,42529,42631,42788,42798,43230,43711,43827,43962,44108,44123,44606,44619,44749,44976,45188,45258,45435,45633,45731,45802,46268,46546,47200,47325,47927,47954,48166,48569,48649,49062,49844,50329,50396,50425,50453,50671,50786,50855,50912,50918,50978,50993,51025,51029,51222,51239,51253,51348,51372,51384,51399,51456,51495,51545,51562,51580,51676,52102,52158,52214,52244,52264,52268,52308,52337,52359,52371,52424,52736,52826,52965,53017,53097,53283,53338,53346,53413,53467,53529,53679,53725,53726,53761,53817,53825,53918,54043,54045,54124,54140,54211,54538,54539,54700,54792,54892,54942,55011,55013,55067,55112,55122,55227,55235,55529,55583,55652,55698,55861,55870,55888,55913,55919,55936,55949,55993,56002,56165,56240,56315,56328,56437,56497,56626,56732,56816,56893,56956,56983,57205,57294,57389,57441,57521,57539,57545,57592,57621,57651,57680,57765,57780,57810,57873,58067,58169,58229,58365,58458,58506,58530,58566,58582,58615,58623,58701,58713,58728,58737,58819,58828,58881,59052,59182,59264,59407,59426,59545,59748,59862,59864,60047,60052,60146,60159,60260,60304,60327,60362,60397,60488,60558,60614,60729,60867,61053,61132,61216,61250,61338,61718,61814,61854,61856,61865,61911,62010,62025,62196,62220,62233,62533,62607,62654,62705,62751,62754,62755,62764,62895,62907,62909,62947,63128,63313,63370,63386,63441,63520,63539,63701,63736,63834,64017,64061,64062,64131,64235,64244,64254,64330,64352,64362,64390,64431,64462,64483,64524,64679,64924,65145,65225,65241,65281,65502,65542,65601,65752,65803,65828,65853,65971,66007,66066,66203,66240,66341,66385,66456,66493,66642,66705,66758,66822,66823,66951,67121,67143,67186,67230,67267,67303,67361,67371,67402,67440,67523,67748,67787,67840,67920,67927,68062,68104,68235,68360,68381,68387,68489,68524,68534,68637,68728,68740,68762,68795,68861,68914,69089,69101,69207,69238,69275,69340,69525,69682,69739,69902,70025,70031,70095,70516,70617,70626,70658,70676,70751,70852,70962,71294,71343,71371,71401,71417,71446,71536,71625,71636,71708,71869,71883,71909,71936,71965,72484,72649,72699,72774,72836,72876,72885,73000,73142,73235,73243,73247,73375,73381,73399,73408,73475,73508,73649,73659,73688,73811,73854,73980,73994,74026,74121,74123,74200,74210,74270,74478,74510,74540,74623,74721,74737,74854,74993,75043,75139,75243,75434,75620,75762,75807,75870,76151,76200,76479,76490,76634,76715,77008,77061,77114,77250,77292,77334,77484,77485,77556,77802,77807,77809,77839,77895,77939,77963,78051,78183,78200,78214,78276,78309,78383,78435,78706,78906,78912,78944,79051,79152,79510,79517,79583,79602,79613,79689,79807,79819,80017,80194,80333,80367,80386,80442,80534,80621,80689,80810,80865,80947,80980,81014,81145,81238,81251,81281,81380,81405,81495,81573,81742,81777,81823,81857,82125,82297,82306,82320,82409,82500,82634,82735,82893,82998,83109,83363,83586,83597,83623,83668,83684,83702,84085,84113,84342,84454,84462,84581,84706,84849,84896,84969,85029,85084,85092,85111,85177,85424,85682,85691,85754,85902,86070,86094,86213,86246,86310,86464,86593,86677,86781,86922,86952,86970,86971,86973,87142,87211,87239,87314,87394,87440,87631,87657,87686,87691,87802,87840,87952,88052,88187,88197,88207,88333,88409,88422,88588,88589,88848,89085,89172,89296,89446,89486,89493,89568,89571,89584,89617,89679,89680,89712,89753,89978,90001,90022,90047,90090,90093,90295,90477,90574,90576,90621,90668,90732,90902,90942,91029,91051,91080,91244,91532,91610,91703,91796,91898,91925,91968,91977,92094,92138,92211,92254,92289,92290,92315,92352,92381,92546,92614,92665,92692,92757,92825,92948,92959,92976,92990,93019,93021,93175,93257,93521,93546,93573,93613,93758,93776,93817,93843,93883,93951,93991,94015,94058,94120,94162,94164,94193,94338,94379,94442,94598,94786,94873,94960,95066,95097,95201,95309,95364,95453,95469,95824,96046,96158,96242,96248,96360,96373,96504,96526,96614,96738,96739,96771,96789,96903,97109,97114,97132,97280,97375,97403,97404,97444,97585,97598,97699,97709,97735,97790,97802,97946,97963,97980,97996,98034,98172,98316,98341,98353,98387,98399,98528,98586,98727,98733,98763,98791,98818,98856,98871,98968,99035,99080,99102,99107,99245,99372,99403,99477,99550,99571,99651,99662,99712,99772,99867,99876",
          "87,320,339,374,414,428,641,698,733,903,965,984,995,1072,1100,1110,1149,1166,1198,1211,1236,1239,1275,1291,1310,1345,1410,1513,1542,1548,1572,1590,1593,1608,1698,1867,1868,1923,2087,2092,2608,2699,2719,2800,3107,3223,3355,3370,3701,3888,4087,4145,4217,4359,4530,4716,4806,5018,5023,5043,5045,5065,5079,5083,5145,5181,5196,5215,5224,5231,5265,5273,5274,5283,5311,5314,5395,5464,5496,5523,5538,5687,5741,5778,5801,5845,5899,5903,5949,6051,6080,6173,6202,6277,6285,6331,6393,6433,6442,6494,6498,6507,6553,6607,6648,6769,6817,6920,6982,7063,7169,7396,7423,7457,7495,7522,7535,7779,7808,7907,7929,8001,8197,8237,8311,8336,8427,8494,8642,8767,9140,9327,9441,9462,9473,9669,9700,9866,9963,9968,10038,10285,10529,10619,10952,11103,11221,11464,11524,11542,11634,11844,11872,12139,12505,12874,12931,12936,13156,13165,13176,13648,13729,14658,14736,14835,14848,14904,15150,15483,15545,16059,16804,17274,17983,18086,18630,18645,18779,19034,19200,19405,19483,19824,20746,20789,20928,21359,21561,22010,22226,23058,23486,23511,23596,23672,23767,23957,24356,24901,25021,25249,25300,25775,25895,26453,26562,26744,26861,27252,27341,27372,27425,27879,27975,28376,28573,28675,28723,28741,29078,29442,29486,29663,29929,30694,31477,31709,33739,34087,34319,34406,35083,35605,35682,35788,36184,36744,36867,37001,37070,37130,37687,37754,37944,38250,38648,38689,39120,39179,39180,39627,39632,39663,39853,39931,41029,41145,41232,41435,42683,42841,42921,43809,44229,44347,44921,45500,45692,45752,45931,45941,46336,47088,47775,48314,48416,48664,49469,49669,50032,50044,50048,50349,50381,50426,50461,50463,50487,50605,50665,50726,50917,51158,51195,51214,51282,51284,51303,51377,51427,51431,51540,51710,51877,51912,51998,52000,52085,52200,52256,52569,52695,52808,52833,52989,53072,53078,53208,53294,53536,53602,53630,53665,53843,53909,54050,54084,54094,54237,54256,54692,54794,54806,54831,54955,55012,55062,55069,55143,55155,55201,55246,55343,55351,55361,55399,55414,55547,55550,55557,55639,55701,55772,55963,55976,56012,56135,56340,56666,56671,56899,56942,57141,57164,57204,57250,57261,57307,57569,57587,57593,57664,57776,57797,57857,57971,57986,58151,58344,58601,58846,59083,59088,59131,59380,59381,59629,59724,59740,59744,59777,59861,59876,59928,59959,59994,60116,60122,60150,60272,60290,60376,60387,60568,60792,60812,60875,60998,61018,61049,61054,61300,61306,61397,61447,61477,61793,62037,62210,62265,62326,62626,62794,62859,62905,62996,63057,63096,63100,63106,63234,63244,63299,63307,63319,63406,63449,63494,63495,63600,63616,63689,63729,63747,63768,63879,63915,63994,64582,64728,64755,64857,65039,65077,65100,65118,65296,65328,65435,65570,65799,65815,65827,65848,65883,66048,66202,66255,66261,66482,66532,66564,66588,66647,66650,66862,66869,66879,66907,67007,67056,67086,67190,67268,67325,67495,67497,67581,67667,67756,67873,67912,68256,68279,68300,68345,68413,68519,68713,68736,68780,68856,68870,68919,69015,69115,69148,69229,69256,69258,69403,69470,69753,69771,69847,70106,70120,70197,70453,70544,70555,70556,70557,70667,70717,70778,70814,70892,71007,71015,71024,71031,71133,71140,71221,71226,71248,71264,71291,71339,71365,71426,71433,71510,71615,71649,71688,71713,71815,71908,71919,71940,72039,72406,72414,72477,72594,72648,72696,72700,72718,72753,72783,72797,72804,72850,72993,73160,73191,73240,73265,73267,73385,73635,73643,73664,73869,73965,73987,74052,74340,74468,74496,74614,74703,74808,75047,75152,75212,75262,75382,75441,75448,75539,75564,75608,75658,75772,75932,76050,76109,76134,76196,76209,76225,76280,76315,76637,76694,76744,76751,76802,77154,77231,77264,77312,77553,77659,77677,77685,77968,77977,78201,78206,78238,78265,78284,78525,78638,78720,78850,78858,78938,79020,79024,79095,79176,79179,79258,79287,79400,79593,79675,79686,79702,79748,79803,79858,79938,79946,79954,80030,80249,80283,80325,80430,80468,80513,80529,80588,80634,80701,80741,80789,80832,80951,80992,81020,81028,81038,81083,81150,81185,81243,81274,81316,81427,81478,81499,81513,81633,81670,81719,81766,81934,82160,82195,82474,82507,82512,82762,82803,82804,82941,83142,83164,83299,83388,83522,83536,83552,83738,83838,84154,84231,84261,84321,84346,84441,84586,84594,84701,84707,84795,84813,84826,84883,84922,85080,85121,85232,85444,85460,85535,85565,85791,85859,85954,86318,86448,86522,86550,86751,86774,87057,87104,87157,87169,87238,87266,87269,87282,87315,87502,87637,87638,87762,87832,87848,87906,87908,87917,87979,88094,88307,88550,88668,88691,88706,88785,88882,88883,88949,88960,89016,89088,89150,89179,89301,89326,89333,89382,89429,89454,89474,89576,89656,89675,89708,89797,89840,89854,89944,89988,90078,90122,90149,90245,90251,90348,90375,90468,90723,90820,90892,90905,90931,90961,90968,91025,91302,91313,91329,91440,91454,91464,91589,91622,91623,91658,91746,91866,91927,92048,92130,92502,92515,92539,92583,92622,92685,92955,92973,92993,93010,93222,93287,93379,93394,93436,93507,93528,93633,93643,93665,93725,93750,93830,94168,94256,94265,94789,94956,95003,95083,95202,95242,95317,95410,95463,95547,95582,95629,95689,95701,95774,95838,95987,95995,96001,96060,96155,96228,96372,96415,96443,96572,96619,96624,96682,96689,96838,96913,97152,97261,97418,97493,97510,97567,97589,97940,98230,98314,98393,98406,98461,98472,98603,98648,98815,98840,98876,98886,98961,98995,99065,99096,99155,99158,99208,99257,99412,99669,99670,99743,99749,99780",
          "40,201,206,226,244,296,315,521,626,729,883,888,1029,1128,1318,1325,1348,1503,1652,1678,1749,1846,1859,1866,1937,1961,1970,1977,1980,1985,1991,1994,2002,2017,2036,2051,2052,2053,2068,2070,2106,2157,2166,2174,2185,2213,2238,2252,2273,2281,2328,2331,2340,2345,2370,2376,2383,2384,2385,2401,2416,2431,2471,2473,2510,2540,2545,2601,2618,2625,2652,2655,2663,2682,2731,2756,2847,2859,2866,2874,2878,2904,2920,2931,2946,2988,3027,3055,3116,3122,3124,3212,3284,3293,3316,3385,3386,3391,3392,3393,3435,3438,3456,3469,3486,3501,3502,3506,3511,3521,3531,3533,3540,3547,3644,3648,3654,3670,3708,3762,3778,3821,3871,3912,3958,4011,4028,4029,4068,4072,4146,4165,4178,4181,4229,4261,4286,4364,4381,4439,4447,4451,4471,4481,4486,4499,4525,4531,4534,4536,4542,4544,4549,4568,4578,4583,4586,4596,4621,4622,4694,4713,4714,4721,4736,4748,4750,4798,4828,4836,4839,4855,4865,4891,4917,4954,4978,4985,4993,5058,5370,5427,5535,5551,5690,5716,5756,5764,5782,5857,5927,5970,6065,6081,6120,6132,6206,6221,6234,6311,6358,6412,6530,6575,6595,6682,6697,6747,6752,6768,6825,6828,7022,7042,7098,7175,7237,7348,7353,7430,7459,7469,7518,7534,7560,7705,7797,7803,7836,7839,7900,7911,7963,8024,8167,8273,8284,8293,8320,8322,8337,8353,8394,8465,8585,8799,8898,8904,8925,8927,8959,8966,8986,9205,9356,9373,9377,9399,9738,10102,10204,10272,10313,10387,10561,10575,10685,10727,10783,10802,10976,11005,11346,11665,11696,11789,12207,12312,12389,12437,12454,12951,13023,13336,13778,13869,13884,13891,14053,14300,14306,14367,14477,14489,14623,15080,15248,15485,15503,15562,16115,16141,16154,16246,16494,16852,16866,16910,16925,17152,17230,17322,17328,17397,17806,17856,17947,18026,18098,18269,18423,18642,18706,18735,18864,19190,19312,19607,19618,20151,20181,20261,20475,20599,20826,20994,21040,21101,21141,21261,21311,21366,21459,21594,21681,21948,22174,22328,22490,22722,23060,23250,23333,23455,23762,23909,23925,24014,24484,24586,24986,25030,25184,25234,25282,25291,25505,25638,25932,25998,26013,26061,26062,26215,26544,26553,26556,26575,26577,26591,26599,26611,26620,26679,26747,26789,26954,27035,27183,27358,27453,27476,27505,27543,27569,27589,27611,27685,27729,27931,27964,27982,27994,27995,28005,28007,28025,28032,28043,28049,28051,28052,28079,28105,28108,28129,28142,28145,28147,28227,28405,28420,28450,28469,28472,28497,28500,28502,28514,28577,28589,28604,28608,28610,28625,28632,28637,28669,28807,28970,29011,29050,29085,29207,29308,29327,29358,29389,29523,29526,29530,29534,29572,29591,29608,29617,29625,29652,29671,29677,29678,29721,29772,29832,29957,30020,30049,30078,30107,30136,30165,30194,30223,30252,30281,30310,30339,30368,30397,30426,30455,30484,30513,30542,30571,30600,30629,30658,30687,30716,30745,30774,30803,30832,30861,30890,30919,30948,30977,31006,31035,31064,31093,31122,31151,31180,31209,31238,31267,31296,31325,31354,31370,31383,31412,31441,31470,31499,31528,31557,31586,31615,31644,31673,31702,31731,31760,31789,31818,31847,31876,31905,31934,31963,31992,32021,32050,32079,32108,32137,32166,32195,32224,32282,32311,32340,32369,32398,32427,32456,32485,32514,32543,32572,32601,32630,32659,32688,32717,32746,32775,32804,32833,32862,32891,32920,32949,32978,33007,33036,33065,33094,33123,33152,33181,33210,33239,33268,33297,33326,33355,33384,33413,33442,33471,33500,33529,33558,33587,33616,33645,33674,33703,33732,33761,33790,33819,33848,33877,33906,33935,33964,33993,34022,34051,34080,34109,34138,34167,34196,34225,34254,34283,34312,34341,34370,34399,34428,34457,34486,34544,34573,34602,34631,34660,34689,34718,34747,34776,34805,34834,34863,34892,34921,34950,34979,35007,35055,35065,35079,35081,35109,35158,35183,35211,35216,35272,35299,35321,35346,35383,35420,35462,35491,35539,35598,35657,35730,35746,35789,35848,35873,35919,35978,36037,36064,36072,36100,36160,36219,36290,36319,36349,36425,36480,36539,36598,36662,36721,36738,36780,36803,36851,36880,36910,36920,36969,36989,37041,37100,37116,37159,37223,37241,37282,37341,37377,37412,37471,37530,37602,37661,37689,37720,37745,37747,37784,37843,37857,37902,37930,37973,38000,38032,38091,38163,38222,38281,38345,38404,38423,38427,38463,38534,38593,38619,38652,38724,38783,38842,38858,38906,38916,38965,38975,39024,39061,39095,39154,39213,39285,39305,39344,39403,39482,39526,39585,39656,39715,39736,39774,39846,39964,39985,40124,40153,40182,40240,40269,40298,40327,40356,40414,40472,40501,40530,40559,40588,40617,40646,40675,40704,40733,40762,40791,40820,40849,40878,40936,40965,40994,41023,41052,41081,41110,41139,41168,41197,41226,41255,41284,41313,41342,41371,41400,41429,41458,41487,41516,41545,42015,42044,42073,42102,42131,42191,42221,42240,42281,42335,42356,42431,42446,42448,42518,42593,42639,42668,42703,42755,42801,42830,42842,42856,42871,42905,42922,42928,42935,42962,42992,43020,43067,43103,43107,43142,43180,43229,43261,43304,43330,43342,43352,43379,43391,43466,43474,43484,43541,43556,43566,43616,43659,43677,43703,43778,43853,43881,43897,43952,43983,43984,44015,44024,44053,44090,44127,44128,44146,44147,44177,44206,44252,44275,44327,44414,44438,44467,44489,44564,44577,44602,44624,44625,44651,44678,44689,44726,44732,44739,44770,44801,44840,44888,44923,44963,45003,45038,45064,45111,45121,45125,45159,45191,45220,45278,45351,45408,45426,45443,45466,45482,45538,45595,45653,45665,45674,45725,45743,45782,45797,45840,45912,45969,46004,46014,46031,46048,46062,46072,46093,46124,46138,46141,46148,46228,46234,46264,46276,46319,46367,46408,46426,46429,46436,46452,46550,46563,46612,46664,46670,46693,46696,46712,46728,46730,46739,46770,46799,46834,46858,46900,46911,46937,47016,47044,47056,47066,47079,47099,47110,47135,47151,47156,47158,47168,47269,47288,47322,47328,47338,47353,47410,47429,47522,47548,47558,47569,47582,47586,47635,47666,47685,47696,47716,47733,47769,47791,47801,47817,47818,47861,47869,47928,47964,47975,47999,48015,48080,48149,48192,48205,48215,48221,48235,48244,48278,48283,48320,48322,48389,48444,48474,48502,48516,48627,48701,48723,48726,48744,48754,48760,48785,48822,48846,48871,48928,48968,49007,49118,49141,49255,49263,49264,49309,49320,49331,49383,49441,49544,49557,49664,49673,49678,49719,49771,49841,49842,49848,49891,49899,49901,49939,49960,49972,49995,50146,50235,50392,50399,50491,50689,50730,50762,50847,50896,51008,51049,51153,51341,51693,51791,51854,51923,51971,52211,52281,52379,52439,52463,52480,52532,52759,52813,52829,52834,52953,53150,53206,53287,53648,53697,53742,53752,53754,53800,53884,54016,54078,54272,54367,54499,54684,54805,55034,55080,55237,55471,55663,55704,55716,55720,55762,55979,55984,56236,56549,56725,56758,56814,56885,56957,56985,57223,57275,57420,57449,57527,57604,57697,57786,57836,57982,58014,58146,58298,58327,58455,58521,58592,58638,58730,58740,58951,58999,59132,59141,59156,59177,59291,59292,59318,59334,59376,59452,59484,59576,59609,59657,59663,59701,59811,59882,59905,59910,59969,59973,60048,60085,60128,60149,60195,60230,60259,60367,60400,60696,60725,60732,60761,60796,60975,61001,61145,61196,61313,61320,61360,61396,61407,61497,61519,61627,61694,61734,61925,61929,61961,62135,62241,62481,62520,62535,62581,62712,62768,62868,62922,63161,63371,63452,63455,63496,63642,63796,63825,63962,63980,63998,64233,64506,64535,64831,64863,64877,64889,65105,65106,65195,65201,65337,65398,65545,65584,65667,65849,65851,65954,65982,66052,66113,66135,66155,66215,66268,66348,66408,66544,66766,66774,66832,66909,66998,67192,67242,67397,67407,67520,67590,67614,67645,67679,67713,67737,67831,67954,67988,68098,68131,68253,68273,68289,68308,68455,68470,68502,68517,68530,68685,68732,68741,68754,68813,68844,68952,69029,69071,69293,69428,69689,69748,70273,70291,70324,70371,70473,70613,70621,70685,70824,70957,70993,71311,71508,71641,71700,71835,71873,71877,71984,72006,72015,72105,72124,72152,72260,72308,72360,72412,72853,72870,73117,73200,73208,73222,73254,73284,73307,73648,73752,73777,73804,73808,73817,73843,73931,73942,74065,74145,74275,74330,74536,74551,74556,74576,74662,74681,74783,74894,74918,75010,75039,75082,75093,75151,75227,75263,75277,75287,75307,75326,75394,75450,75630,75670,75693,75704,75839,75899,75948,75953,75960,75999,76000,76074,76201,76228,76245,76290,76338,76344,76371,76408,76454,76480,76572,76759,76871,76903,76911,76930,76938,76990,77049,77075,77127,77256,77275,77374,77402,77403,77602,77634,77646,77724,77779,77796,77914,77955,78088,78208,78244,78387,78389,78420,78503,78627,78654,78737,78798,78807,78873,78981,78982,79023,79274,79307,79548,79606,79630,79710,79930,80125,80155,80199,80346,80361,80544,80545,80549,80777,80816,80889,80978,81275,81293,81463,81473,81503,81556,81575,81594,81656,81799,82023,82042,82117,82176,82294,82418,82533,82544,82574,82691,82716,82857,82866,82883,83023,83068,83170,83381,83470,83545,83612,83685,83746,83908,83987,84116,84380,84455,84458,84527,84537,84551,84711,84936,84982,84994,85019,85020,85060,85100,85200,85282,85319,85461,85478,85583,85650,85749,85949,85978,86119,86129,86515,86533,86555,86573,86621,86683,86752,86767,86814,86845,86908,86942,87061,87133,87361,87591,87656,87854,87869,87898,87900,88036,88076,88085,88106,88260,88273,88533,88630,88707,88764,88944,88965,88981,88994,89059,89122,89248,89250,89294,89338,89383,89408,89428,89556,89662,89849,89861,89907,90126,90145,90178,90260,90339,90445,90471,90532,90571,90781,90817,90855,90960,91124,91247,91264,91357,91408,91455,91482,91587,91594,91768,91778,91835,92026,92063,92239,92374,92392,92399,92402,92422,92580,92587,92651,92666,92860,92940,93009,93180,93272,93319,93894,93911,94055,94071,94104,94115,94269,94318,94367,94736,94748,94806,94864,94970,94991,95053,95068,95114,95246,95280,95372,95419,95576,95621,95882,95971,95981,96004,96079,96084,96343,96371,96429,96555,96666,96759,96760,96915,97096,97115,97118,97382,97495,97538,97581,97688,97728,97730,97825,97840,97883,97937,97957,98059,98177,98214,98377,98407,98512,98693,98931,98991,99063,99170,99200,99229,99383,99457,99520,99594,99639,99791,99814,99848,99938",
          "156,234,259,292,303,319,407,420,446,469,470,600,627,654,709,721,804,809,895,906,1042,1087,1183,1306,1387,1390,1433,1483,1511,1514,1595,1599,1643,1695,2448,2516,2759,3234,3240,3388,3398,4081,4380,4956,5010,5051,5060,5078,5104,5199,5404,5405,5409,5435,5596,5674,5786,5876,5902,6006,6114,6162,6315,6344,6384,6401,6438,6562,6646,6706,6912,7101,7250,7276,7380,7393,7517,7580,7626,7686,7760,7784,7813,8094,8335,8408,8456,8499,8567,8580,8656,8749,8841,8914,8945,9097,9264,9338,9409,9575,9905,9961,9980,10062,10091,10143,10223,10316,10412,10446,10490,10757,10776,10850,10876,11039,11162,11244,11363,11377,11581,11588,11804,11812,11887,11953,11996,12016,12060,12122,12167,12422,12647,12879,13333,13372,13610,14035,14350,14632,14634,14737,14798,14900,14977,15198,15298,15459,15764,15786,15990,16637,16794,17190,17220,17356,17431,17551,17633,17741,17927,18116,18495,18520,18549,18628,18690,18893,18965,19351,19556,19766,20069,20624,21406,22132,22546,22664,22738,22754,23020,23111,23239,23530,23722,23754,23826,24241,24366,24654,24667,24686,24790,24797,24861,25039,25043,25728,26233,26727,26907,26957,27060,27234,27889,27919,28530,28770,28782,28984,29130,29451,29514,29527,29732,29827,29856,29901,35875,36177,36242,36444,36933,37244,37615,37671,37760,38053,38054,38118,38245,38319,38428,38490,38550,38760,38762,39563,39693,39917,39927,40211,40778,40907,42378,43457,45105,45183,45304,45315,46075,47540,50071,50116,50137,50142,50149,50221,50226,50231,50291,50308,50356,50389,50687,50770,50817,51028,51119,51146,51288,51409,51423,51428,51464,51511,51524,51619,51689,51861,51932,52239,52445,52484,52504,52637,52660,52693,52745,52932,52964,52992,53068,53105,53143,53531,53547,53661,53706,53710,53975,54035,54164,54200,54232,54554,54627,54734,54855,54859,54902,54994,55039,55041,55051,55136,55196,55295,55369,55509,55707,55713,55743,55789,55965,56030,56031,56084,56085,56241,56374,56421,56466,56487,56492,56511,56515,56636,56705,56731,56774,56781,56788,56970,56993,57020,57079,57187,57407,57540,57761,57886,58030,58121,58236,58247,58281,58413,58469,58553,58784,58797,58818,58962,58982,59028,59125,59223,59280,59337,59349,59418,59460,59668,59847,59978,60056,60071,60201,60241,60299,60312,60355,60404,60460,60469,60557,60768,60840,60923,61096,61144,61278,61290,61686,61706,61741,61799,61812,61894,62076,62102,62347,62451,62547,62744,62849,62946,63136,63408,63422,63429,63489,63534,63698,63934,63944,64112,64166,64349,64501,64619,64657,64722,64785,64925,65164,65180,65362,65516,65647,65654,65684,65747,65864,65967,66062,66112,66161,66165,66242,66282,66398,66669,66711,66714,66956,67015,67058,67120,67129,67151,67264,67454,67491,67493,67507,67636,67802,67818,68001,68007,68270,68388,68417,68451,68503,68520,68553,68599,68744,68776,68818,68837,68970,69040,69206,69295,69301,69372,69375,69449,69457,69460,69606,69729,69737,69815,69970,70014,70134,70137,70146,70304,70328,70352,70439,70444,70475,70524,70771,70843,71073,71150,71193,71265,71276,71350,71428,71436,71458,71559,71591,71617,71664,71682,71758,71831,71858,72114,72190,72321,72392,72604,72629,72656,72762,72788,72869,72958,72981,72995,73083,73153,73422,73654,73746,73780,73911,73919,73959,73960,73966,74119,74166,74213,74316,74350,74427,74479,74547,74750,74794,74809,75025,75106,75126,75562,75622,75759,75855,75907,75940,75987,76028,76049,76160,76272,76297,76424,76442,76478,76591,76714,76834,76881,76888,76954,76965,77029,77096,77152,77168,77191,77201,77228,77268,77367,77469,77470,77488,77563,77601,77725,77837,77998,78099,78149,78181,78330,78608,78694,78702,78809,78889,79028,79109,79229,79253,79316,79319,79364,79382,79597,79626,79688,79833,80004,80021,80192,80280,80385,80524,80830,80874,80924,81000,81007,81043,81109,81192,81329,81397,81472,81507,81702,81716,81768,81937,82082,82257,82338,82357,82451,82622,82850,83095,83195,83306,83374,83432,83785,83830,83904,83923,83930,83966,83998,84008,84029,84031,84135,84303,84314,84529,84617,84634,84703,84789,84798,84819,84828,84898,84980,85027,85046,85067,85093,85098,85156,85303,85332,85373,85438,85469,85544,85690,85757,85808,85827,85852,85898,85966,86057,86074,86092,86118,86152,86221,86259,86271,86303,86373,86569,86578,86655,86667,86777,86878,87037,87038,87039,87044,87054,87137,87279,87378,87382,87489,87507,87521,87675,87689,87969,88061,88114,88132,88198,88203,88272,88335,88350,88366,88487,88573,88850,88904,88911,89071,89078,89169,89290,89339,89388,89391,89475,89508,89607,89661,89725,89727,89732,89899,89933,89980,89990,90077,90187,90255,90265,90282,90438,90622,90651,90733,90797,90861,90908,90972,91028,91069,91187,91273,91304,91376,91431,91616,91711,91815,91837,92029,92044,92380,92535,92568,92592,92730,92762,92792,92837,93023,93061,93104,93140,93141,93145,93158,93204,93307,93344,93370,93413,93445,93481,93488,93530,93625,93777,93821,93861,93992,94016,94139,94151,94380,94564,94619,94631,94672,94707,94768,94796,94817,94922,94968,94978,95076,95146,95223,95308,95353,95588,95617,95656,95784,95803,95865,95889,95915,95926,96006,96085,96105,96211,96335,96457,96827,96867,96954,97176,97259,97338,97455,97575,97668,97993,98082,98117,98122,98196,98245,98247,98325,98405,98418,98429,98448,98552,98615,98654,98779,98821,98829,98839,98914,98946,99014,99085,99095,99277,99283,99357,99363,99422,99450,99543,99546,99629,99732,99830,99965",
          "83,84,124,195,282,419,540,580,763,771,773,898,912,1064,1097,1174,1184,1244,1248,1250,1450,1516,1557,1570,1614,1781,1901,2269,2280,2557,2570,2701,2883,3131,3194,3374,3416,3609,3728,3775,3789,3855,3939,4199,4348,4358,4429,4490,4521,4538,4920,5094,5136,5165,5184,5343,5444,5468,5539,5581,5752,5787,5792,5794,5839,5863,5934,6000,6024,6170,6245,6405,6480,6555,6690,6715,6810,6876,6877,6919,6969,7037,7085,7174,7223,7272,7344,7473,7483,7500,7583,7591,7668,7820,7903,7958,7974,8209,8239,8269,8287,8331,8457,8544,8632,8765,8773,8782,8906,9003,9046,9125,9203,9211,9334,9449,9489,9742,9768,9825,9827,9889,9911,9979,10127,10145,10238,10404,10423,10578,10661,10770,10782,10817,10836,10918,10973,10981,10989,10996,11043,11144,11329,11435,11437,11466,11469,11546,11561,11578,11604,11683,11705,11707,11788,11792,11796,11917,12054,12467,12515,12565,12611,12682,12752,12822,12864,13033,13079,13487,13499,13631,13941,14140,14331,14348,14498,14583,14926,15172,15252,15271,15334,15336,15428,15489,15555,15632,16126,16381,16492,16530,16557,16563,16685,16753,16905,16917,16923,17019,17340,17775,17787,18099,18121,18126,18300,18330,18781,18833,18870,19101,19225,19305,19489,19540,19558,19573,19801,19980,19985,19998,20002,20029,20110,20199,20282,20288,20395,20401,20521,20531,20581,20618,20851,21134,21244,21245,21321,21344,21389,21513,21525,21706,21771,22337,22347,22350,22529,22575,22868,22964,23012,23100,23211,23364,23402,23443,23536,23653,23665,23872,23910,24122,24166,24247,24322,24327,24401,24619,24779,24864,24896,24942,24984,25358,25520,25531,25550,25786,25835,26384,26406,26440,26451,26766,26872,27114,27318,27675,27716,27759,27786,27792,27807,27863,27916,27951,28034,28259,28296,28336,28345,28359,28435,28718,28812,28818,28910,28999,29089,29276,29392,29426,29447,29453,29491,29692,29694,29807,29888,30017,30024,30206,30420,30510,30815,31126,31931,32286,32424,32569,32598,32605,32917,33396,33860,33976,34295,34483,34628,34933,35231,35360,35551,35571,35601,35812,35821,35886,35895,36014,36041,36111,36216,36368,36496,36536,36550,36595,36750,37076,37249,37300,37312,37381,37508,37553,37618,37812,38101,38119,38130,38383,38814,38825,38874,38880,38883,38977,39323,39426,39441,39479,39509,39582,39609,40253,40654,41002,41216,41274,42053,42165,42228,42237,42363,42535,42545,42654,42840,42895,43354,43549,43810,43843,43867,44581,44599,45187,45237,45240,45302,45337,45358,45550,45566,45662,45676,45981,45982,46162,46274,46435,46869,46921,46961,47617,47661,47712,47991,48004,48023,48124,48191,48749,48845,49655,49785,50451,50588,50594,50596,50667,50829,50853,50939,51005,51033,51040,51059,51094,51099,51149,51190,51247,51305,51309,51334,51376,51499,51919,51986,51991,52092,52093,52302,52311,52316,52742,52850,52853,52884,52912,52935,52976,53137,53247,53286,53454,53644,53715,53733,53849,53855,53979,53997,54098,54229,54456,54693,54981,55606,55657,55665,55702,55848,56065,56091,56225,56614,56687,56837,56866,56923,56968,57105,57165,57391,57417,57444,57568,57597,57758,57845,57905,57908,57947,57989,58086,58497,58574,58577,58765,58775,58905,59005,59336,59386,59493,59673,59762,59804,59945,60208,60220,60464,60519,60525,60789,61138,61347,61392,61449,61520,61548,61655,61760,61984,62049,62078,62158,62286,62501,62523,62541,62594,62776,62933,63018,63068,63117,63149,63168,63190,63314,63318,63458,63475,63479,63523,63615,63694,63818,63823,64075,64160,64168,64311,64326,64387,64530,64533,64818,64970,65070,65078,65173,65264,65365,65416,65612,65675,65700,65713,65728,65788,65811,65991,66133,66172,66237,66283,66473,66476,66495,66630,66680,66690,66706,66866,66895,66923,66972,66986,67029,67073,67179,67208,67227,67248,67251,67327,67431,67567,67633,67796,67990,68017,68091,68154,68298,68352,68356,68581,68597,68687,68770,68815,68885,69176,69289,69333,69490,69569,69617,69705,69747,69779,69858,69884,70040,70064,70069,70173,70215,70222,70274,70289,70402,70417,70430,70526,70581,70595,70598,70654,70761,70798,70900,70951,70953,71035,71115,71301,71340,71465,71500,71515,71771,71811,71945,72192,72255,72522,72614,72643,72659,72711,72803,72927,72956,73034,73073,73148,73170,73181,73251,73278,73304,73449,73464,73610,73673,73861,73865,73875,74051,74061,74287,74552,74572,74581,74629,74674,74877,74908,74930,75206,75236,75250,75253,75257,75408,75420,75437,75526,75614,75676,75770,75783,75815,76004,76051,76126,76130,76189,76311,76549,76588,76651,76794,76868,76902,77353,77394,77481,77521,77527,77606,77662,77747,77986,77991,78033,78034,78049,78104,78112,78269,78291,78385,78426,78741,78758,78789,78817,78863,78872,78995,79156,79248,79261,79408,79501,79577,79627,79628,79638,79697,79706,79779,79831,79862,79995,80067,80102,80119,80188,80266,80339,80355,80426,80523,80728,80809,80836,80885,80897,80997,81325,81431,81487,81551,81617,81673,81678,81709,81773,81839,81939,81986,82187,82194,82209,82270,82295,82470,82492,82497,82556,82612,82736,82854,82926,82971,82975,82987,83029,83072,83085,83105,83124,83163,83177,83199,83228,83229,83262,83263,83286,83355,83448,83452,83641,83739,83757,83790,83802,83804,83812,83886,83906,83920,83924,83943,83946,84050,84160,84199,84555,84679,84734,84742,84780,84814,84855,84869,85011,85022,85032,85048,85196,85215,85226,85429,85560,85671,85679,85774,85832,85847,85926,86036,86089,86138,86243,86440,86453,86474,86544,86705,86708,86846,86965,87051,87121,87143,87146,87216,87230,87323,87485,87497,87571,87595,87662,87668,87693,87696,87747,87749,87960,88025,88048,88133,88184,88205,88252,88444,88513,88524,88571,88620,88697,88709,88812,88943,88952,89012,89060,89061,89310,89440,89492,89547,89585,89770,89800,89832,89883,89941,90049,90118,90312,90858,90907,91164,91174,91199,91292,91490,91590,91629,91788,91814,92056,92072,92145,92300,92311,92520,92627,92840,92850,92919,92922,92934,93102,93111,93219,93308,93391,93657,93762,93896,93910,93915,93983,93995,94153,94175,94317,94336,94386,94716,94809,94833,95015,95057,95080,95116,95169,95176,95209,95226,95405,95439,95574,95810,96122,96441,96448,96667,96690,96761,96792,96866,96892,97053,97175,97210,97294,97516,97613,97617,97623,97864,97956,98001,98015,98116,98327,98349,98391,98423,98463,98517,98619,98624,98722,98738,98764,98802,99009,99086,99446,99512,99537,99547,99558,99598,99609,99695,99837",
          "43,132,254,488,542,544,550,658,722,833,967,1205,1467,1518,1625,1682,1687,1835,1862,1884,1886,2144,2196,2241,2388,2402,2475,2502,3178,3225,3403,3600,3764,3813,3822,3879,4159,4270,4278,4389,4434,4514,4588,4820,4922,5186,5220,5277,5448,5449,5483,5508,5548,5591,5726,5734,5908,5920,5948,6036,6112,6201,6264,6290,6388,6398,6418,6545,6616,6678,6704,6764,6788,6908,6958,6970,7096,7260,7389,7414,7431,7448,7642,7662,7748,7755,7762,7777,7809,7869,7933,8093,8132,8328,8500,8540,8615,8619,8705,8715,8721,8731,8860,8921,9171,9278,9359,9360,9429,9486,9506,9527,9536,9600,9607,9756,10017,10289,10418,10430,10452,10628,10671,10751,10796,10806,11050,11063,11121,11186,11215,11307,11387,11497,11526,11567,11676,11727,11748,11782,11959,12029,12051,12088,12164,12315,12472,12526,12549,12556,12559,12665,12690,12751,12894,12922,12925,13007,13019,13027,13144,13311,13313,13319,13390,13473,13531,13622,13666,13894,14042,14106,14132,14293,14349,14529,14643,14711,14761,14836,14843,14925,14966,14988,15092,15238,15422,15508,15667,15670,15850,15895,15978,16011,16269,16395,16429,16476,16480,16550,16641,16693,17013,17319,17371,17390,17404,17418,17587,17653,17797,17854,17949,17960,18112,18150,18167,18183,18228,18310,18414,18454,18500,18506,18528,18551,18592,18799,18890,18985,19146,19175,19180,19289,19328,19575,19594,20040,20088,20144,20196,20204,20278,20553,20569,20607,20620,20659,20723,20771,20799,20938,20973,21160,21209,21246,21517,21624,21740,21758,21768,22007,22141,22197,22206,22363,22387,22444,22668,22836,22873,23003,23134,23307,23335,23470,23602,23720,23736,23763,23830,23842,24154,24229,24279,24371,24458,24553,24923,25313,25539,25601,25693,25748,25779,25812,25853,25872,25957,26114,26168,26276,26349,26416,26613,26652,26657,26688,26712,26720,26730,26757,26765,26792,26985,27054,27077,27101,27108,27186,27240,27297,27316,27519,27585,27658,27665,27790,28167,28200,28265,28330,28404,28440,28674,28679,28724,28779,28794,28858,28959,28974,28977,29140,29202,29235,29236,29278,29435,29493,29681,29686,29758,29760,29802,29961,29974,29986,29989,30139,30149,30498,30759,30817,31368,31425,32064,32122,32209,32324,32498,32933,33456,33561,33734,33793,33920,34036,34837,35151,35405,35562,35660,35679,35684,35703,35744,35880,35898,35981,36003,36009,36052,36170,36258,36321,36363,36372,36495,36573,36665,36677,36754,36836,36879,37000,37016,37134,37247,37295,37447,37497,37499,37555,37634,37759,37942,37952,38055,38075,38129,38140,38189,38367,38502,38578,38711,38795,38818,38820,38991,38992,39043,39064,39185,39259,39329,39435,39718,39747,39998,40174,40662,41046,41265,41421,41479,42453,42526,42599,42694,42785,42865,42943,42980,43134,43154,43263,43445,43607,43629,43715,43804,44339,44423,44714,44811,45318,45366,45484,45666,45672,45695,45809,45866,45870,46685,46826,47637,47662,48095,48427,48657,48899,49109,49214,49427,49465,49761,50033,50050,50064,50100,50111,50332,50400,50484,50494,50589,50693,50716,50794,50901,50958,50974,50998,51201,51211,51271,51310,51373,51471,51477,51713,51762,51790,51834,51851,51907,51911,51958,51963,52023,52084,52171,52363,52365,52533,52548,52638,52705,52947,53082,53203,53284,53353,53401,53443,53594,53860,53873,53986,54026,54058,54240,54275,54293,54301,54327,54355,54398,54408,54409,54424,54427,54475,54506,54549,54555,54575,54634,54695,54710,54759,54773,54808,54840,54951,54972,55005,55137,55326,55342,55450,55748,55781,55802,55926,56120,56179,56262,56363,56380,56537,56783,56796,56876,56905,56926,56955,57018,57059,57132,57251,57466,57474,57581,57585,57638,57682,57730,57759,57853,57956,58053,58080,58174,58347,58520,58537,58659,58761,58817,58948,59200,59293,59469,59485,59560,59675,59775,59802,60010,60130,60217,60291,60356,60446,60531,60674,60754,60815,60920,61003,61007,61033,61081,61092,61337,61373,61429,61459,61467,61498,61601,61743,61776,61980,62060,62071,62126,62380,62435,62486,62538,62664,62792,62914,63058,63213,63230,63296,63331,63402,63535,63553,63607,63643,63647,63678,63824,64025,64102,64164,64216,64270,64313,64345,64358,64370,64497,64502,64640,64644,64682,64709,64860,64894,65162,65188,65360,65369,65377,65399,65520,65645,65795,65822,65895,66004,66196,66278,66357,66457,66471,66524,66560,66624,66655,66845,66873,66897,67008,67081,67113,67335,67603,68129,68171,68371,68421,68613,68724,68767,68846,68886,68967,69017,69234,69642,69678,69949,69969,70001,70022,70038,70228,70263,70332,70365,70392,70483,70615,70641,70741,70756,70937,71025,71170,71315,71432,71470,71483,71488,71529,71601,71656,71689,71767,71870,71911,71963,71991,72171,72208,72370,72404,72431,72467,72511,72551,72670,72934,72976,73124,73245,73298,73666,73709,73791,73803,73920,73986,74029,74240,74257,74356,74555,74589,74624,74646,74893,75001,75031,75050,75740,75860,75862,76027,76101,76115,76153,76166,76239,76505,76779,76933,77044,77221,77259,77332,77495,77562,77640,77694,77707,77717,77740,77753,77778,77872,78018,78052,78077,78314,78321,78481,78568,78661,78683,78876,78892,78904,78922,79048,79056,79174,79188,79493,79696,79701,79801,79864,79962,80044,80048,80173,80187,80207,80293,80328,80434,80461,80489,80569,80610,80618,80753,80834,80913,80983,81426,81553,81563,81640,81663,81744,82033,82071,82145,82172,82192,82317,82545,82647,82793,82842,82846,82863,82880,83006,83117,83146,83270,83323,83372,83418,83473,83634,83759,83798,83848,84178,84202,84498,84565,84766,84815,84821,84823,84917,84938,85095,85119,85125,85172,85298,85328,85482,85527,85594,85620,85698,85733,85961,85991,86372,86396,86514,86571,86604,86732,86737,86750,86812,86828,87078,87256,87307,87338,87648,87650,87746,87810,87880,87912,87984,88004,88008,88034,88062,88068,88099,88240,88298,88405,88534,88541,88631,88768,88866,88876,88899,89128,89431,89443,89532,89677,89695,89740,89813,90036,90107,90177,90179,90200,90235,90277,90344,90466,90585,90672,90836,90837,90974,90979,90994,91006,91024,91043,91046,91195,91197,91386,91481,91549,91651,91817,91880,91906,92002,92157,92207,92274,92277,92280,92345,92507,92555,92560,92576,93029,93082,93144,93192,93194,93250,93275,93278,93375,93589,93642,93663,93668,93716,93865,93974,94026,94073,94125,94187,94197,94214,94433,94434,94576,94659,94828,94958,94961,95244,95288,95302,95319,95342,95452,95502,95546,95624,95662,95710,95923,95945,95960,96008,96098,96216,96223,96287,96317,96630,96705,96721,96756,96829,96991,97089,97148,97242,97301,97462,97506,97599,97634,97658,97705,97706,97716,97764,97941,98063,98352,98492,98541,98635,98637,98644,98855,98956,98977,99252,99356,99388,99398,99676,99741,99881,99884",
          "75,227,239,512,1368,1667,1694,1705,1740,2009,2333,2685,2860,2953,3080,3200,3248,3270,3303,3396,3529,3618,3726,3730,3947,4004,4042,4061,4102,4234,4325,4347,4433,4625,4657,4687,4846,4884,5381,6113,6603,6623,6854,6949,6960,8051,8127,8133,8146,8165,8556,8640,8659,8683,8697,9178,9305,9343,9362,9438,9533,9552,9628,9647,9666,9685,9737,9799,9881,10061,10065,10070,10083,10118,10141,10198,10298,10350,10374,10379,10388,10597,10616,10659,10679,10723,10768,10825,10827,10842,10844,10960,10970,11064,11110,11148,11205,11243,11262,11281,11319,11395,11401,11441,11490,11507,11509,11547,11635,11685,11737,11973,12003,12065,12117,12150,12181,12193,12212,12250,12269,12273,12345,12408,12440,12535,12552,12573,12593,12604,12649,12680,12706,12730,12744,12826,12858,12896,12901,12906,12991,13018,13063,13094,13111,13124,13143,13162,13200,13238,13295,13369,13382,13486,13580,13675,13770,13789,13808,13903,13922,13998,14013,14112,14169,14226,14264,14302,14397,14454,14473,14492,14511,14568,14606,14701,14720,14758,14815,14853,14891,14929,14967,15005,15036,15062,15125,15138,15195,15199,15201,15290,15309,15363,15423,15499,15537,15575,15651,15727,15765,15841,15917,15924,15936,16046,16050,16069,16088,16145,16202,16240,16259,16260,16278,16297,16316,16335,16354,16392,16468,16487,16525,16582,16603,16604,16696,16702,16715,16791,16796,16825,16829,16841,16864,16886,16981,17000,17057,17209,17228,17247,17304,17361,17380,17432,17475,17532,17585,17655,17760,17779,17798,17817,17874,18064,18108,18207,18234,18349,18368,18387,18406,18463,18476,18485,18501,18563,18634,18710,18712,18748,18767,18786,18824,18862,18900,18919,18926,19090,19128,19147,19223,19261,19280,19292,19299,19318,19394,19413,19565,19596,19622,19641,19660,19679,19812,19822,19831,19964,20059,20071,20097,20116,20135,20173,20185,20192,20230,20268,20306,20363,20382,20458,20477,20515,20591,20629,20686,20743,20895,20914,20933,20971,20979,20988,21028,21047,21104,21180,21256,21294,21351,21446,21465,21484,21522,21579,21617,21636,21674,21693,21769,21788,21864,21883,21902,21921,21940,21997,22016,22065,22111,22149,22175,22187,22240,22263,22282,22301,22358,22396,22434,22491,22510,22514,22605,22624,22643,22681,22700,22757,22778,22795,22871,22883,22909,22915,22928,22946,22978,23023,23052,23194,23232,23289,23339,23373,23384,23419,23460,23498,23559,23574,23631,23669,23707,23726,23878,23973,23985,23992,24011,24030,24049,24220,24239,24249,24258,24296,24334,24448,24486,24543,24657,24676,24695,24735,24741,24752,24771,24783,24809,24847,24885,24906,24961,24999,25037,25056,25170,25208,25303,25341,25398,25455,25469,25512,25569,25588,25645,25683,25778,25816,25873,25904,25911,25949,25968,26006,26025,26044,26082,26101,26158,26177,26196,26253,26291,26329,26367,26386,26424,26443,26462,26481,26500,26519,26538,26728,26819,27147,27165,27222,27255,27392,27625,27849,27851,28058,28262,28370,28552,28590,28647,28793,28799,28808,28831,28854,29059,29229,29268,29388,29436,29503,29619,29654,29833,29841,29863,29910,29940,30082,30111,30140,30227,30297,30314,30372,30430,30459,30517,30575,30662,30691,30749,30807,30981,31068,31155,31184,31213,31271,31300,31387,31416,31503,31561,31590,31619,31648,31677,31706,31793,31851,31909,31996,32025,32054,32112,32141,32170,32199,32228,32402,32431,32472,32489,32518,32547,32576,32663,32721,32779,32837,32924,32953,33011,33069,33098,33156,33214,33243,33272,33301,33359,33388,33475,33504,33591,33620,33649,33678,33765,33794,33823,33852,33939,33997,34009,34026,34113,34171,34229,34241,34258,34316,34345,34403,34461,34490,34519,34548,34577,34606,34635,34693,34780,34838,34925,34954,34983,35024,35038,35141,35156,35159,35161,35243,35319,35350,35387,35424,35446,35495,35543,35602,35661,35769,35793,35852,35869,35881,35923,35982,35995,36058,36060,36164,36223,36353,36426,36543,36602,36611,36666,36725,36784,36815,36855,36914,36973,37104,37227,37286,37416,37475,37560,37606,37665,37724,37732,37736,37738,37788,37906,37977,38036,38048,38049,38095,38167,38226,38325,38408,38467,38478,38538,38563,38597,38656,38728,38846,38860,38910,38937,38969,39028,39158,39229,39348,39364,39407,39471,39530,39589,39719,39778,39850,39871,39909,39968,40132,40161,40190,40248,40277,40306,40335,40364,40393,40625,40683,40712,40741,40770,40828,40857,40886,40915,40973,41031,41060,41118,41147,41176,41205,41263,41292,41321,41379,41408,41437,41495,41524,41961,41969,41973,41977,41990,41995,42024,42082,42171,42246,42252,42385,42402,42411,42481,42491,42537,42569,42613,42648,42662,42761,42767,42780,42810,42868,42885,42930,42959,42966,43035,43041,43047,43122,43184,43277,43359,43437,43521,43596,43639,43758,43803,43808,43995,44232,44307,44366,44381,44524,44544,44597,44664,44771,44810,44855,44900,45018,45117,45139,45200,45227,45236,45331,45384,45388,45518,45575,45679,45705,45738,45788,45820,45863,45892,45921,46050,46052,46079,46085,46114,46220,46238,46271,46277,46284,46332,46374,46448,46502,46514,46556,46601,46711,46746,46812,46833,46845,46894,46947,47024,47025,47048,47059,47130,47189,47284,47312,47324,47397,47408,47432,47457,47524,47538,47560,47728,47748,47794,47798,47926,47958,47967,47987,48001,48006,48027,48035,48194,48325,48363,48402,48412,48476,48511,48705,48770,48774,48811,48932,48954,48957,49102,49169,49253,49301,49312,49328,49498,49527,49545,49590,49600,49679,49682,49809,49821,49890,49907,49912,49949,49954,50041,50049,50078,50110,50112,50113,50115,50121,50205,50460,50523,50579,50625,50633,50645,50661,50731,50757,50784,51038,51044,51142,51191,51248,51294,51311,51447,51523,51526,51536,51539,51597,51614,51664,51767,51850,51862,51951,51987,52207,52335,52473,52477,52510,52770,52909,52996,53205,53350,53397,53406,53435,53486,53548,53572,53858,53917,54086,54116,54146,54257,54259,54309,54341,54393,54507,54543,54592,54653,54658,54680,54718,54790,54799,54843,54922,54971,55024,55221,55265,55268,55463,55480,55838,55893,55937,56060,56128,56188,56197,56281,56285,56308,56377,56564,56576,56727,56836,56842,56853,56868,56894,56994,57021,57024,57188,57267,57268,57356,57499,57528,57775,57843,57856,57983,58020,58049,58066,58142,58184,58212,58310,58446,58478,58621,58668,58681,58685,58967,59235,59379,59567,59736,59906,59981,59986,59997,60035,60165,60183,60215,60258,60280,60309,60383,60595,60635,60921,60960,61166,61410,61513,61833,61927,62189,62270,62475,62545,62657,62727,62910,62985,63035,63265,63322,63335,63415,63573,63630,63703,63713,63723,63897,64043,64054,64117,64141,64339,64509,64534,64551,64642,65087,65088,65261,65571,65625,65681,65865,66034,66074,66119,66121,66311,66363,66631,66649,66827,66880,67184,67261,67349,67514,67618,67673,67694,67868,67905,67953,68087,68132,68193,68372,68623,68653,68657,68665,68670,68737,68779,68852,68988,69020,69119,69141,69284,69318,69483,69585,69607,69644,69686,69783,70132,70210,70354,70373,70602,70678,70744,70924,71070,71075,71180,71288,71329,71357,71581,71663,71855,71959,72092,72130,72169,72207,72229,72317,72355,72487,72569,72771,72800,72845,72889,73016,73210,73215,73217,73257,73283,73458,73468,73503,73519,73531,73674,73683,73842,73923,73924,73925,74098,74238,74296,74321,74333,74351,74452,74466,74509,74550,74582,74652,74687,74702,74875,74905,74921,75006,75028,75051,75080,75150,75199,75311,75392,75432,75471,75491,75555,75567,75668,75729,75801,76034,76163,76170,76195,76253,76615,76743,76776,76847,76900,77037,77326,77328,77555,77596,77623,77626,77661,77760,77774,77813,77853,78002,78063,78108,78320,78394,78414,78473,78495,78601,78640,78645,78757,78816,78827,78919,78965,78973,78991,79000,79070,79157,79160,79164,79215,79320,79332,79380,79417,79479,79941,80060,80142,80211,80226,80236,80308,80441,80518,80519,80547,80645,80670,80683,81080,81170,81214,81231,81399,81477,81549,81683,81690,81694,81780,81854,81896,81929,82273,82419,82447,82531,82585,82589,82700,82799,83014,83152,83154,83321,83420,83510,83618,83659,83740,83741,83826,83834,83857,83881,83885,83977,83979,84006,84078,84100,84103,84110,84145,84198,84215,84272,84323,84325,84623,84650,84673,84825,84888,84935,84967,85068,85078,85204,85421,85439,85496,85618,85720,85735,85817,85889,85955,86008,86166,86266,86270,86358,86413,86523,86574,86608,86697,86698,86873,86902,87111,87117,87130,87167,87178,87261,87545,87673,87778,87809,88020,88024,88075,88228,88247,88340,88401,88655,88739,88858,88914,89036,89125,89324,89377,89478,89533,89723,89891,89995,90004,90053,90079,90206,90218,90321,90356,90390,90430,90549,90674,90739,90829,90856,91014,91089,91101,91260,91272,91295,91338,91341,91706,91828,91944,91950,92030,92252,92557,92603,92657,92807,92827,92932,93022,93045,93086,93205,93245,93322,93404,93418,93429,93522,93535,93603,93604,93824,93829,93849,93869,93892,94021,94122,94298,94485,94615,94666,94771,94845,94869,94886,94950,94974,94998,95085,95172,95267,95277,95395,95457,95567,95597,95606,95677,95715,95728,96069,96081,96104,96117,96141,96180,96262,96280,96289,96291,96307,96382,96418,96447,96510,96539,96550,96588,96744,96768,96859,97073,97093,97205,97434,97522,97712,97925,98000,98056,98094,98161,98485,98573,98589,98739,98754,98934,98993,99066,99076,99122,99243,99248,99272,99348,99359,99371,99375,99406,99445,99470,99821,99940,99955",
          "6,15,157,317,326,427,558,728,742,764,839,900,1036,1116,1119,1150,1317,1401,1423,1474,1552,1581,1607,1800,1808,1816,1840,1865,2065,2108,2133,2201,2349,2398,2409,2457,2591,2658,2703,2770,2954,3082,3100,3227,3238,3258,3551,3601,3606,3635,3684,3704,3913,3914,4083,4317,4753,4786,5009,5225,5228,5260,5281,5490,5504,5648,5679,5685,5799,5889,5894,5896,5898,5900,5923,5928,5946,6053,6157,6171,6242,6348,6359,6394,6445,6451,6459,6477,6516,6519,6532,6618,6692,6759,6821,6831,6892,6899,6923,6964,7050,7235,7258,7322,7358,7520,7553,7559,7650,7744,7923,7951,8010,8048,8129,8202,8219,8259,8313,8316,8423,8461,8481,8796,8892,9024,9061,9092,9158,9509,9713,9769,9854,9879,9934,10052,10085,10296,10505,10510,10520,10523,10534,10548,10773,10830,10863,10906,11044,11106,11487,11522,11548,11692,11704,11795,11992,12052,12446,12590,12595,12642,12670,12782,12791,12865,12871,12918,12956,13154,13166,13175,13188,13226,13348,13352,13391,13471,13510,13595,13627,13716,14065,14490,14493,14826,14844,14980,15043,15152,15371,15398,15610,15860,16615,16644,16755,17200,17395,17569,17658,17706,17794,17921,17941,17982,18327,18514,18545,18699,18707,18709,18809,18834,18873,18928,19397,19512,19574,20265,21074,21223,21250,21354,21378,21384,21397,21458,21492,21615,21781,22051,22103,22171,22207,22295,22377,22719,22736,23106,23303,23586,24068,24214,24413,24421,24751,24955,25086,25139,25228,25254,25377,25459,25637,25813,25882,26040,26043,26301,26344,26579,26750,26752,26753,26776,26790,26818,26878,26969,27062,27126,27153,27206,27235,27301,27309,27457,27641,27693,27830,27839,27872,27911,28409,28473,28748,28935,28975,29036,29107,29144,29169,29324,29429,29455,29516,29565,29861,30201,30275,31841,32057,32218,32439,32845,33349,33436,33533,33871,34145,34664,34712,35048,35374,35509,35518,35570,35748,35766,35825,36001,36094,36261,36262,36379,36506,36569,36774,36808,37056,37117,37196,37296,37301,37309,37406,37626,37632,37778,37881,37943,38015,38110,38133,38194,38201,38317,38605,39176,39234,39249,39322,39358,39559,39634,39804,39918,40172,40415,40444,40937,41377,42016,42646,43175,43663,43670,43831,43912,43960,44070,44317,44501,44774,44932,45422,45686,47093,47220,47266,47367,47496,47692,47701,47761,48060,48297,48442,49329,49663,49732,50026,50103,50124,50204,50261,50366,50379,50538,50549,50622,50678,50708,50771,50834,50844,50960,51087,51089,51147,51179,51198,51356,51402,51895,51981,52012,52047,52118,52183,52262,52621,52930,52931,53043,53169,53319,53351,53377,53453,53477,53816,53903,54101,54114,54127,54244,54324,54350,54429,54516,54633,54745,54754,54807,54812,54849,54866,54880,54898,54954,55019,55028,55210,55212,55254,55322,55395,55476,55572,55579,55600,55619,55620,55744,55858,55954,56306,56472,56524,56716,56722,56768,56778,56840,56856,56950,57060,57127,57448,57675,57676,57740,57773,57985,57992,58187,58319,58330,58398,58423,58428,58433,58494,58505,58874,59057,59066,59074,59106,59192,59247,59251,59326,59357,59483,59513,59546,59580,59699,59728,59731,59750,59949,59962,60004,60069,60072,60095,60104,60111,60188,60395,60456,60494,60512,60588,60641,60714,60918,60951,61193,61402,61415,61437,61489,61570,61794,61975,62003,62087,62111,62370,62489,62587,62704,62796,62801,62838,62908,62926,62998,63087,63210,63317,63343,63435,63471,63515,63664,63763,63799,63838,63889,63918,63921,63931,63968,63997,64097,64101,64177,64238,64421,64470,64471,64545,64565,64633,64695,64702,64712,64764,64827,64957,65017,65074,65083,65097,65131,65138,65156,65214,65403,65486,65521,65562,65800,65805,65903,65948,65963,66201,66442,66518,66671,66713,66719,66830,66958,66995,67040,67045,67202,67216,67313,67320,67370,67400,67503,67544,67548,67757,67792,67880,67885,67951,68039,68042,68165,68261,68318,68399,68430,68432,68532,68654,68814,68821,69012,69168,69193,69421,69473,69590,69630,69633,69650,69685,69694,69724,69756,69875,70007,70071,70107,70150,70290,70372,70470,70551,70560,70608,70651,70867,71071,71163,71190,71234,71268,71393,71398,71461,71659,71743,71759,71805,71891,71942,71987,72002,72011,72022,72150,72181,72287,72430,72489,72525,72536,72693,72697,72773,72838,72860,72882,72913,72959,73049,73341,73459,73596,73740,73895,73933,73939,74046,74084,74190,74256,74282,74311,74360,74480,74539,74568,74574,74606,74757,74797,74835,74892,74990,75023,75084,75102,75143,75172,75512,75541,75609,75610,75656,75689,75884,75962,75984,76071,76111,76187,76231,76399,76471,76484,76524,76665,76673,76720,76732,76806,76818,76843,76939,76997,77073,77148,77171,77247,77345,77354,77378,77390,77409,77509,77674,77763,77929,78024,78075,78158,78286,78332,78397,78401,78519,78614,78672,78792,78920,78939,78958,78964,79052,79099,79103,79104,79137,79191,79195,79259,79273,79395,79403,79405,79492,79564,79657,79670,79767,79809,79889,79935,79960,79988,80107,80122,80141,80147,80320,80341,80360,80462,80508,80532,80602,80622,80686,80866,80942,80946,80963,81024,81030,81033,81294,81390,81566,81972,82002,82126,82199,82262,82296,82313,82504,82562,82583,82643,82777,82877,82886,82909,82981,83057,83145,83166,83176,83280,83441,83484,83573,83578,83708,83742,83917,84134,84191,84257,84328,84399,84431,84491,84656,84657,84667,84669,84844,84981,84998,85033,85147,85183,85399,85520,85619,85760,85885,85929,86026,86298,86309,86487,86504,86526,86539,86562,86567,86584,86632,86635,86644,86716,86735,86756,86849,86958,87131,87187,87191,87198,87242,87283,87352,87522,87523,87677,87705,87881,87888,87982,88179,88180,88190,88201,88257,88263,88331,88453,88461,88546,88658,88753,89102,89149,89195,89237,89448,89461,89481,89497,89500,89613,89713,89809,89814,90027,90182,90291,90325,90472,90510,90520,90608,90615,90862,90877,90904,91133,91444,91468,91493,91613,91694,92086,92101,92116,92140,92149,92163,92271,92344,92377,92378,92417,92464,92468,92481,92536,92654,92688,92758,92870,92902,92913,92986,93078,93124,93167,93246,93268,93324,93325,93353,93478,93529,93534,93542,93552,93561,93662,93938,94096,94124,94207,94220,94223,94271,94390,94412,94450,94547,94655,94675,94685,94715,94752,94939,94947,94948,95019,95105,95112,95192,95247,95292,95314,95375,95384,95415,95444,95516,95540,95643,95651,95777,96047,96109,96146,96164,96366,96389,96392,96462,96704,96797,96832,96833,96845,97020,97161,97165,97313,97316,97432,97476,97559,97643,97708,97763,97775,97886,97890,97930,98086,98173,98304,98337,98366,98561,98582,98655,98740,98808,98942,98947,99041,99162,99206,99227,99238,99303,99369,99405,99549,99587,99630,99693,99863,99943,99971",
          "218,327,343,350,408,487,549,649,653,795,799,918,1037,1051,1117,1133,1216,1218,1309,1342,1527,1843,1853,2077,2088,2147,2217,2449,2522,2952,3032,3078,3136,3288,3633,3785,3818,3920,3994,3997,4168,4233,4633,4656,4790,5146,5216,5309,5321,5368,5420,5469,5501,5530,5567,5653,5704,5749,5754,5950,5988,6213,6227,6237,6272,6332,6374,6437,6513,6662,6693,6766,6784,6801,6939,6994,7014,7259,7268,7305,7723,7768,7863,7868,7880,7904,8021,8174,8301,8369,8455,8631,8671,8689,8713,8748,8812,8866,8944,8970,9013,9045,9055,9182,9259,9262,9267,9402,9448,9676,9779,9807,9843,9852,9912,9959,10029,10033,10059,10129,10252,10488,10501,10676,10857,10869,10963,10984,10999,11033,11088,11268,11321,11326,11353,11368,11502,11711,11728,11756,11807,11857,11878,11888,12046,12200,12242,12268,12288,12429,12480,12564,13008,13058,13215,13243,13254,13429,13521,13541,13797,13833,13915,14248,14390,14503,14587,14723,14726,14794,14834,15061,15068,15494,15551,15792,16007,16282,16386,16611,16642,16898,16982,17103,17191,17219,17234,17262,17474,17598,18081,18107,18191,18263,18650,18729,18993,19058,19136,19195,19438,19461,19604,19643,19728,20523,20691,20692,20695,20729,20735,20795,20876,21071,21079,21326,21332,21502,21509,21549,21587,21606,21695,21696,21800,22088,22167,22189,22285,22439,22824,22949,23013,23048,23099,23136,23222,23468,23597,23617,23716,24149,24182,24206,24264,24477,24781,24989,25140,25151,25370,25418,25429,25551,25655,25717,25910,26080,26104,26296,26357,26366,26505,26676,26689,26716,26726,26737,26895,26906,26960,27018,27026,27033,27129,27159,27208,27224,27319,27343,27458,27588,27631,27677,27710,27812,27900,28045,28071,28111,28170,28195,28211,28246,28313,28379,28455,28523,28534,28689,28701,28764,28785,29027,29048,29070,29092,29262,29390,29404,29422,29583,29709,29722,29766,29800,29850,29975,30043,30409,30612,30614,31039,31293,31310,32088,32584,32905,33059,33149,34657,34692,34741,35023,35123,35555,35762,35832,35952,36013,36051,36243,36274,36533,36577,36812,36817,36827,36987,37062,37153,37180,37369,37391,37478,37574,37577,37621,37816,37821,37856,38051,38135,38138,38200,38234,38260,38285,38426,38616,38630,38737,38761,38797,38982,39139,39175,39303,39316,39373,39474,39487,39510,39624,39641,39794,39817,40073,40274,40478,40988,41985,42248,42710,42937,43130,43578,43644,43738,44259,44738,44957,45166,45246,45264,45573,45985,46477,46501,46681,46870,46913,47001,47743,47880,48380,48459,48585,48995,49129,49446,49487,49765,50045,50139,50292,50351,50450,50496,50566,50670,50694,50774,50807,51081,51116,51140,51182,51279,51434,51506,51537,51698,51744,51749,51773,52007,52141,52202,52218,52241,52283,52326,52385,52394,52958,53054,53061,53131,53220,53234,53266,53298,53539,53610,53717,53735,53781,53898,54292,54494,54514,54528,54673,54720,54782,54829,54966,55015,55023,55339,55340,55634,55728,55810,55844,56406,56482,56565,56606,56803,56830,56851,56896,56953,56965,56997,57038,57039,57244,57321,57340,57350,57460,57509,57530,57763,57812,57934,58111,58154,58166,58203,58283,58366,58440,58447,58570,58613,58777,58796,58927,58969,59017,59033,59111,59152,59155,59197,59384,59434,59437,59461,59648,59691,59738,60127,60161,60213,60317,60320,60433,60484,60491,60564,60730,60772,60820,60892,60929,60966,61119,61175,61253,61264,61307,61324,61629,61733,61801,61826,61860,62064,62205,62217,62298,62355,62386,62391,62442,62524,62632,62753,62799,62819,62823,62901,63006,63186,63249,63375,63399,63413,63557,63599,63613,63627,63655,63685,63750,63792,63985,64000,64048,64272,64284,64338,64493,64542,64567,64590,64639,64705,65174,65313,65483,65529,65691,65734,65777,65896,65981,66190,66232,66406,66423,66484,66726,66957,66970,67047,67100,67210,67416,67420,67504,67780,67817,67860,67924,67966,68037,68186,68304,68307,68353,68380,68639,68650,68834,68868,68957,68960,69082,69215,69217,69221,69251,69378,69422,69455,69578,69852,69887,69953,69979,70285,70350,70413,70418,70579,70638,70661,70670,70703,70738,70820,71122,71134,71165,71166,71202,71279,71391,71395,71396,71440,71481,71555,71587,71590,71878,72020,72126,72299,72377,72426,72433,72535,72574,72587,72785,72817,72835,72938,72969,72972,73007,73136,73315,73354,73355,73363,73393,73556,73679,73760,74107,74205,74206,74487,74494,74538,74756,74906,74985,75005,75013,75124,75141,75245,75265,75296,75351,75404,75486,75487,75516,75577,75579,75660,75716,75775,75914,75958,76036,76061,76140,76211,76249,76345,76508,76640,76679,76681,76733,76966,77004,77054,77121,77190,77226,77279,77315,77368,77430,77443,77487,77678,78006,78138,78145,78147,78275,78297,78430,78458,78531,78539,78806,78821,78840,79378,79399,79521,79716,79737,79740,79822,79992,80079,80198,80278,80324,80347,80655,80737,80760,80765,80779,80788,80928,80941,80956,81100,81155,81252,81255,81256,81300,81331,81445,81457,81465,81582,81588,81612,81855,82055,82092,82179,82196,82463,82532,82536,82676,82772,82785,82813,82837,82964,83002,83097,83274,83316,83405,83633,83844,83860,83990,84024,84076,84123,84184,84222,84227,84248,84260,84270,84410,84446,84778,84831,85126,85151,85163,85494,85502,85566,85730,85734,85778,85937,85983,86041,86054,86060,86107,86256,86393,86420,86641,86818,86832,86857,86884,87021,87108,87414,87473,87501,87513,87536,87608,87621,87676,87729,87903,87943,88077,88224,88306,88313,88385,88403,88596,88623,88728,88734,89157,89160,89207,89234,89528,89535,89574,89586,89606,89622,89689,89711,89789,89803,89825,89852,89886,90084,90313,90329,90489,90493,90519,90564,90636,90697,90935,90941,90969,91017,91147,91389,91446,91602,91719,91913,92012,92060,92088,92122,92128,92185,92192,92233,92242,92406,92454,92509,92559,92637,92640,92903,92939,92978,93004,93031,93070,93211,93290,93363,93503,93515,93855,93870,93904,93945,93963,93984,94024,94113,94230,94415,94471,94515,94628,94799,94853,94966,95044,95047,95236,95250,95399,95983,96157,96377,96514,96515,96578,96584,96776,96843,96852,97012,97086,97108,97160,97215,97271,97305,97306,97514,97523,97533,97588,97739,97784,98005,98011,98193,98471,98560,98623,98669,98814,99094,99192,99211,99215,99330,99577,99605,99774,99899,99966",
          "48,73,85,131,504,527,702,727,1145,1170,1409,1497,1574,1657,1720,1771,1775,1837,2023,2248,2279,2293,2307,2316,2547,2566,2686,2689,2708,2774,2962,2967,3010,3063,3165,3181,3345,3461,3545,3729,3823,3838,3881,3952,4118,4153,4166,4211,4291,4295,4655,4731,4841,4848,4859,4933,4934,4949,5001,5217,5371,5382,5477,5513,5519,5527,5710,5819,5835,5972,6022,6152,6224,6252,6286,6308,6336,6402,6515,6698,6782,6829,6848,6959,7020,7033,7109,7158,7189,7284,7301,7313,7378,7383,7449,7475,7494,7558,7596,7598,7643,7658,7770,7793,7798,7865,7874,7876,7889,7916,7924,8057,8073,8101,8166,8264,8348,8349,8356,8432,8440,8478,8586,8667,8694,8724,8741,8752,8824,8949,9030,9064,9071,9104,9218,9294,9309,9318,9554,9704,9846,9925,9938,10101,10117,10132,10237,10409,10533,10583,10639,10762,10788,10872,10880,10977,10987,11116,11129,11233,11331,11333,11348,11414,11556,11619,11633,11669,11805,11821,12005,12111,12156,12194,12462,12507,12623,12666,12702,12726,12850,12907,12919,13134,13217,13219,13225,13248,13383,13399,13418,13470,13526,13602,13607,13702,13997,14027,14049,14075,14099,14122,14129,14149,14409,14453,14597,14666,14928,14937,14986,15025,15048,15106,15173,15647,15933,16144,16179,16267,16286,16396,16674,16812,16816,16887,16941,16971,17028,17046,17086,17107,17195,17323,17405,17426,17440,17500,17553,17578,17590,17612,17654,17893,18012,18022,18045,18051,18140,18168,18282,18435,18492,18715,18822,18871,19189,19191,19244,19375,19378,19381,19508,19593,19625,19632,19708,19717,19945,19972,20033,20277,20297,20442,20527,20582,20704,20705,20759,20809,20903,20943,21080,21158,21310,21643,21644,21667,21677,21682,21741,21777,21779,21824,21835,22096,22190,22346,22443,22576,22985,22988,23009,23027,23097,23164,23188,23430,23690,23697,23795,23905,23907,24123,24168,24293,24402,24610,24656,24700,24723,24806,24911,25007,25014,25045,25173,25255,25416,25446,25597,25647,25762,25899,25928,26015,26149,26270,26318,26348,26446,26509,26602,26781,26944,26964,26987,27005,27021,27048,27239,27339,28318,28360,28367,28606,28769,28836,28839,28913,29054,29152,29240,29457,29553,29567,29579,29641,29742,29825,29881,29946,30401,30433,30568,30873,30875,30923,31067,31332,31390,31426,31670,31735,31738,31743,31825,32006,32083,32131,32179,32207,32289,32627,32692,32956,32961,32985,33039,33062,33079,33233,33494,33562,33797,34000,34770,34809,35034,35166,35361,35368,35381,35617,35692,36118,36172,36260,36373,36380,36443,36477,36624,36659,36682,36756,36822,36945,37094,37144,37307,37321,37338,37521,37572,37573,37658,37828,38202,38265,38310,38388,38398,38431,38590,38611,38631,38708,38787,38867,39031,39063,39113,39484,39494,39540,39552,39712,39840,40169,40224,40491,40908,41203,41326,41372,41728,41981,42030,42185,42201,42601,42623,42681,42702,42723,42728,42741,42786,42816,42818,42955,43043,43168,43365,43456,43475,43628,43704,43755,43833,43890,44037,44404,44446,44469,44812,44863,45228,45241,45319,45423,45508,45555,45688,45734,45762,45805,45849,46009,46164,46229,46307,46331,46437,46560,46566,46741,46793,46905,47706,47756,47961,48195,48245,48288,48636,48803,48839,49197,49323,49330,49688,49999,50072,50114,50145,50232,50438,50532,50628,50637,50749,50795,50833,50861,50867,50957,51022,51023,51062,51080,51264,51297,51390,51525,51533,51558,51650,51703,51765,51915,52022,52153,52157,52206,52494,52563,52582,52675,52722,52791,52920,52939,52962,53089,53092,53177,53272,53278,53462,53494,53500,53711,53959,54051,54168,54299,54315,54356,54501,54556,54628,54691,54699,54727,54744,54766,54908,54925,55029,55058,55085,55285,55502,55645,55682,55693,55768,55788,55879,55909,55931,56006,56007,56036,56064,56067,56211,56239,56275,56424,56478,56548,56577,56583,56702,56865,56908,56954,57012,57013,57041,57146,57255,57290,57349,57735,57854,57987,58197,58337,58345,58532,58558,58651,58671,58756,59027,59157,59244,59269,59588,59734,59796,59815,59826,59866,59920,59924,59956,60053,60058,60097,60123,60207,60431,60678,60978,61050,61146,61156,61178,61412,61502,61530,61539,61552,61603,61667,62019,62243,62267,62287,62358,62592,62938,63014,63019,63152,63185,63198,63212,63272,63575,63684,63743,63762,64027,64269,64430,64494,64612,64717,64737,64856,64862,64951,64994,65000,65370,65485,65549,65605,65680,65759,65776,65820,65909,65920,65947,65965,65970,66013,66108,66115,66267,66270,66317,66448,66569,66637,66673,66794,66837,67001,67115,67125,67136,67193,67224,67487,67588,67785,67893,67961,68077,68113,68178,68189,68284,68348,68420,68542,68586,68753,68817,68824,68968,69181,69213,69343,69426,69807,69934,69998,70004,70048,70136,70336,70441,70537,70583,70671,70711,70946,70965,71037,71065,71081,71231,71355,71495,71537,71711,71714,71839,71876,71879,71971,72070,72093,72257,72403,72488,72493,72499,72542,72640,72901,72960,72965,72998,73002,73035,73157,73159,73182,73201,73218,73345,73406,73418,73419,73455,73510,73528,73544,73657,73711,73723,73934,74069,74100,74120,74124,74227,74389,74415,74429,74543,74618,74707,74795,74804,74826,74934,75239,75254,75330,75345,75370,75430,75517,75528,75554,75576,75582,75635,75707,75910,76026,76064,76110,76180,76263,76301,76326,76328,76368,76426,76456,76492,76521,76568,76798,76870,76878,77024,77437,77518,77526,77631,77664,77711,77776,77915,78038,78132,78179,78274,78461,78566,78634,78721,78722,78953,79003,79058,79066,79150,79375,79442,79663,79678,80005,80088,80493,80496,80506,80512,80570,80625,80804,81149,81247,81268,81317,81389,81406,81715,81783,81844,81850,81880,82037,82142,82178,82201,82293,82324,82515,82570,82600,82679,82968,83022,83033,83096,83103,83108,83309,83346,83369,83462,83465,83766,83810,83912,84141,84164,84182,84235,84316,84317,84338,84505,84536,84560,84615,84840,84847,84951,85013,85094,85155,85189,85194,85302,85342,85372,85484,85513,85530,85747,85907,86004,86157,86175,86179,86263,86282,86323,86351,86401,86509,86546,86779,86972,86975,87068,87140,87159,87232,87250,87304,87413,87460,87735,87955,88049,88092,88194,88295,88722,88775,88895,88898,88936,89048,89079,89136,89278,89314,89319,89354,89358,89465,89909,89991,90075,90076,90120,90167,90223,90250,90543,90569,90625,90725,90736,90752,90807,90871,90878,90887,91057,91146,91345,91375,91526,91685,91757,91784,91919,92054,92136,92146,92268,92323,92382,92553,92596,92630,92778,92868,92880,92983,93005,93121,93321,93417,93519,93532,93616,93686,93699,93846,93871,93882,93924,93937,93954,93964,94224,94276,94339,94355,94404,94437,94501,94572,94681,94730,94847,94855,94872,94877,94934,94955,94977,95115,95177,95239,95240,95254,95320,95371,95387,95482,95504,95527,95577,95613,95788,95900,96051,96074,96152,96153,96202,96214,96395,96483,96544,96546,96594,96700,96811,96814,96898,96940,97009,97016,97035,97091,97207,97214,97321,97331,97391,97448,97525,97720,97725,97740,97973,98018,98040,98355,98408,98417,98543,98590,98641,98834,98850,98883,98903,98924,99060,99123,99126,99166,99327,99435,99597,99652,99769,99779,99861",
          "24,28,53,199,232,295,311,528,606,1013,1626,1692,1755,1789,1906,2012,2083,2099,2113,2167,2261,2294,2344,2394,2486,2614,2763,2768,2821,3087,3148,3199,3314,3346,3356,3375,3491,3539,3741,3786,3923,3938,3943,3999,4080,4106,4114,4218,4308,4373,4409,4512,4557,4688,4699,4726,4751,4757,4771,4833,4915,5108,5230,5253,5272,5354,5439,5454,5635,5678,5769,5791,5931,6015,6033,6047,6062,6067,6160,6183,6243,6278,6414,6427,6436,6455,6626,6672,6742,6771,6773,6832,6868,6900,6946,7044,7095,7114,7294,7477,7479,7523,7614,7628,7821,7826,7902,8148,8252,8297,8343,8361,8444,8475,8545,8569,8755,8770,8790,8844,8941,8946,9063,9103,9217,9227,9261,9299,9341,9417,9483,9550,9590,9877,9992,10139,10140,10152,10231,10254,10317,10326,10347,10375,10376,10426,10434,10622,10663,10801,10866,10895,10921,10938,11138,11171,11223,11275,11290,11595,11600,11642,11729,11794,11827,11889,12040,12185,12245,12263,12297,12352,12364,12434,12609,12756,12796,12856,12903,12958,12971,13198,13203,13328,13408,13446,13504,13561,13611,13711,13719,13776,13841,13854,13864,14002,14055,14115,14151,14164,14353,14416,14469,14481,14500,14512,14523,14588,14676,14766,15001,15046,15113,15135,15142,15240,15258,15268,15403,15417,15543,15594,15599,15801,15822,15966,15993,16041,16156,16576,16639,16675,16730,16786,16926,17076,17089,17171,17185,17205,17260,17443,17606,17656,17855,17984,17987,18029,18229,18425,18526,18652,18656,18739,18867,19009,19043,19242,19350,19559,19609,19848,19931,19997,20095,20106,20320,20331,20512,20519,20622,20762,20874,20936,21284,21300,21436,21558,21661,22146,22221,22231,22248,22409,22467,22548,22598,22645,22689,22756,22847,23041,23126,23170,23193,23245,23281,23457,23897,24090,24138,24163,24283,24340,24391,24500,24557,24562,24575,24600,24710,24749,24907,24965,24967,25207,25651,25653,25678,25795,25829,25857,26063,26066,26153,26184,26294,26499,26536,26565,26674,26695,26706,26711,26734,26783,26823,26864,26980,27010,27105,27139,27168,27229,27365,27417,27430,27438,27477,27499,27521,27537,27557,27709,27720,27842,27869,27945,28148,28171,28214,28350,28352,28361,28481,28528,28662,28717,28780,28801,28820,28993,28998,29057,29066,29181,29223,29251,29253,29302,29303,29400,29423,29424,29466,29684,29716,29728,29755,29788,29889,29930,30000,30053,30104,30270,30285,30291,30343,30452,30467,30604,30623,30894,31081,31110,31319,31358,31598,31690,31873,31941,31956,31967,32018,32178,32305,32474,32497,32550,32618,33004,33555,33707,33784,34190,34219,34329,34353,34493,34773,34857,34896,35036,35126,35251,35269,35335,35400,35448,35463,35511,35591,35609,35629,35701,35998,36020,36063,36131,36132,36182,36266,36273,36387,36469,36615,36674,36687,36759,36777,36831,36866,36886,36889,36928,37132,37176,37194,37201,37275,37316,37405,37432,37435,37439,37454,37561,37579,37983,38181,38246,38355,38378,38544,38554,38645,38679,38807,38809,38962,38980,39001,39128,39217,39257,39628,39743,39767,39821,39839,39893,39919,40103,40143,40219,40375,40420,40422,40502,40594,40652,40684,40862,40890,41089,41116,41138,41322,41807,41849,41899,41901,41925,41956,41965,42014,42167,42202,42254,42362,42396,42417,42467,42485,42679,42685,42784,42936,43044,43068,43079,43218,43317,43392,43452,43650,43674,43854,43883,43907,43941,43958,44362,44550,44727,44820,44896,45124,45300,45310,45352,45572,45580,45596,45613,45683,45847,45927,45934,45955,46057,46094,46147,46221,46247,46302,46506,46571,46590,46701,46945,47504,47528,47699,47779,47886,48057,48138,48222,48262,48521,48660,48703,48738,48851,48856,48989,49034,49177,49492,49652,49654,49889,49931,50028,50169,50260,50280,50348,50691,50787,50798,50850,50868,51032,51043,51061,51100,51155,51255,51388,51394,51440,51515,51658,51857,51858,51978,52138,52155,52169,52205,52223,52466,52617,52622,52968,52985,53018,53029,53047,53219,53225,53288,53390,53391,53418,53617,53694,53740,53972,54081,54106,54131,54151,54407,54448,54468,54657,54696,54703,54707,54918,55104,55204,55280,55584,55648,55671,55708,56070,56086,56246,56372,56388,56398,56404,56621,56622,56630,56658,56674,56798,56951,57016,57068,57213,57220,57298,57428,57507,57945,58001,58012,58024,58039,58138,58189,58214,58221,58409,58470,58525,58697,58747,58785,58804,58957,58993,59009,59060,59203,59447,59582,59586,59825,59840,59890,59939,59954,59976,60003,60029,60141,60214,60242,60281,60310,60314,60321,60453,60753,60852,60891,60913,60917,61183,61349,61614,61658,61846,61909,62044,62069,62086,62110,62202,62293,62328,62350,62462,62532,62608,62759,62763,62999,63002,63039,63081,63176,63201,63211,63216,63218,63221,63281,63400,63487,63550,63622,63634,63734,63845,63881,63932,64018,64087,64144,64178,64246,64304,64328,64400,64438,64585,64809,64932,64937,64945,64961,65112,65129,65199,65216,65226,65238,65292,65366,65406,65412,65413,65479,65490,65498,65510,65525,65585,65671,65699,65753,65978,65988,66012,66084,66238,66274,66399,66521,66585,66602,66772,66925,67049,67059,67069,67238,67391,67517,67686,67711,67765,67798,68082,68086,68092,68156,68325,68377,68461,68468,68471,68491,68563,68682,68707,68855,69171,69173,69226,69306,69842,69880,70085,70194,70327,70385,70543,70578,71155,71397,71421,71480,71504,71527,71579,71654,71826,71932,72031,72212,72233,72315,72325,72339,72389,72394,72466,72607,72733,72736,72812,72873,73004,73044,73090,73337,73379,73415,73527,73624,73663,73704,73796,73814,73825,73897,74127,74135,74187,74211,74284,74403,74421,74524,74557,74697,74763,74814,74902,74951,74956,74988,75061,75092,75157,75160,75533,75556,75563,75621,75629,75655,75748,75766,75773,75859,75867,75869,75945,75947,76070,76127,76273,76556,76585,76633,76641,76655,76815,76840,76874,76899,77046,77077,77098,77101,77472,77475,77723,77743,77787,77789,77930,77973,78212,78522,78621,78647,78749,78805,78819,78976,79010,79061,79116,79117,79454,79635,79844,79920,79927,79972,80003,80010,80117,80315,80319,80330,80383,80505,80528,80556,80613,80658,80675,80698,80750,80790,80850,81102,81146,81196,81265,81287,81417,81443,81466,81619,81634,81693,81706,81945,81970,82091,82247,82298,82367,82369,82403,82406,82453,82571,82616,82738,82962,83082,83101,83120,83132,83284,83291,83296,83490,83541,83564,83598,83699,83745,83861,83955,83961,83983,84020,84060,84088,84165,84208,84391,84413,84435,84463,84492,84589,85238,85435,85445,85464,85501,85523,85611,85685,85763,85819,85823,85828,85864,85868,85917,85969,86145,86346,86377,86498,86639,86679,86688,86718,86747,86755,86762,86842,86844,86893,86909,86948,87107,87160,87356,87367,87415,87437,87442,87551,87628,87633,87799,87927,87944,88058,88060,88125,88226,88256,88261,88388,88448,88523,88544,88563,88647,88667,88679,88752,88878,89004,89081,89083,89334,89368,89504,89525,89767,89772,89822,89853,90043,90105,90115,90161,90225,90259,90335,90370,90377,90433,90605,90721,90742,90776,90812,90867,90947,91056,91107,91216,91426,91475,91557,91568,91612,91653,91758,91804,91847,91864,91924,92188,92390,92425,92426,92430,92542,92549,92577,92594,92647,92701,92843,92951,92960,93001,93038,93234,93249,93303,93368,93369,93415,93510,93538,93636,93732,93782,93786,94022,94032,94060,94101,94268,94463,94481,94650,94701,94747,94825,94851,94884,94899,94900,94993,94997,95077,95119,95120,95163,95228,95355,95382,95495,95541,95560,95580,95663,95782,95974,95977,96039,96080,96195,96344,96370,96421,96495,96496,96533,96632,96742,96743,96790,96810,96834,96863,96900,96950,96955,96969,97004,97006,97011,97083,97348,97389,97439,97446,97512,97518,97615,97690,97750,97947,98012,98248,98318,98333,98339,98359,98525,98538,98673,98692,98724,98736,98743,98817,98838,98873,98933,99016,99036,99115,99152,99191,99230,99275,99278,99321,99453,99740,99784,99890,99932",
          "79,222,287,309,358,365,373,678,688,802,940,948,1122,1323,1326,1380,1452,1528,1666,1697,1725,1834,2169,2290,2506,2555,2665,2674,2695,2730,2807,2817,2851,2876,2998,3168,3513,3566,3638,3834,4190,4296,4343,4369,4387,4513,4645,4684,4765,4899,4904,4921,4925,5031,5116,5130,5131,5391,5529,5545,5601,5626,5656,5821,5932,5955,6076,6116,6296,6297,6363,6429,6456,6496,6524,6547,6686,6753,6824,6834,6872,6881,6938,6999,7089,7155,7219,7323,7415,7420,7438,7474,7498,7789,7849,7896,7953,7981,7996,8149,8305,8473,8520,8571,8595,8700,8701,8745,8822,8852,8939,9035,9050,9074,9149,9150,9153,9288,9310,9388,9431,9569,9872,9884,9929,10392,10486,10507,10558,10674,10884,10887,10932,11054,11122,11127,11340,11419,11559,11589,11631,11672,11779,11840,11861,11886,11916,11947,12041,12195,12356,12438,12453,12632,12734,12885,13322,13338,13507,13560,13562,13620,13818,13875,13905,13947,13968,13996,14391,14462,14505,14515,14549,14564,14690,14717,14791,14813,14969,14996,15040,15072,15160,15376,15460,15565,15624,15669,15686,15957,16031,16108,16142,16264,16357,16387,16406,16546,16632,16838,17072,17202,17311,17408,17483,17662,17997,18046,18055,18076,18146,18504,18535,18620,18756,18760,18913,18916,19035,19056,19079,19183,19291,19503,19522,20039,20615,20855,20862,20875,20932,21046,21065,21166,21375,21408,21623,21664,21731,21775,21988,22018,22114,22131,22208,22290,22312,22372,22374,22496,22629,22651,22675,22678,22838,22863,22869,22935,23004,23121,23137,23204,23447,23493,23821,24581,24635,24960,24962,24964,25027,25225,25257,25335,25408,25618,25643,25840,26128,26483,26651,26671,26810,26920,26922,26974,27127,27196,27277,27347,27379,27427,27551,27556,27575,27882,27907,27934,27946,28235,28339,28365,28386,28422,28496,28712,28912,28983,28989,29231,29259,29501,29524,29596,29734,29873,29977,30081,30090,30150,30255,30324,30865,30904,31232,31251,31339,32608,32691,32836,33175,34193,34287,34760,35014,35017,35256,35622,35693,35715,35955,36301,36474,36637,36647,36986,37107,37124,37183,37250,37289,37490,37623,37633,37690,38139,38254,38257,38503,38548,38602,38617,38693,39066,39067,39184,39376,39439,39742,39932,40398,40705,40959,41010,42022,42483,42831,42854,43118,43402,43420,43460,43547,43667,43985,44041,44202,44357,45238,45247,45288,45415,45446,45600,45607,45630,45693,45698,45749,45858,45865,46318,46369,47527,48024,48373,48873,49510,49614,50023,50263,50305,50368,50408,50543,50739,50765,50830,50904,50991,51041,51126,51269,51415,51596,51648,51712,51815,51984,52103,52180,52224,52253,52272,52295,52313,52646,52697,52712,52805,52903,52911,52950,53011,53014,53024,53110,53274,53343,53606,53623,53640,53791,53839,53930,53980,54311,54317,54363,54454,54508,54584,54640,54642,54722,54832,54876,54946,54949,55177,55356,55526,55558,55847,56075,56092,56096,56292,56329,56349,56410,56459,56574,56575,56648,56772,56790,56874,56928,56931,57286,57421,57467,57492,57493,57630,57696,57723,57844,57878,57974,58000,58003,58078,58249,58256,58314,58331,58438,58580,58626,58772,58979,59008,59015,59047,59148,59234,59285,59298,59325,59391,59399,59594,59756,59896,59990,60270,60406,60443,60536,60567,60660,60798,60968,61005,61017,61129,61192,61215,61343,61390,61448,61458,61535,61566,61593,61606,61640,61766,61900,61979,62008,62119,62268,62296,62321,62381,62399,62443,62444,62522,62542,62575,62602,62785,62795,63061,63092,63167,63635,63771,63833,63952,64146,64222,64327,64389,64401,64626,64965,64995,64999,65086,65262,65449,65578,65661,65724,65880,65916,65961,66166,66452,66595,66622,66762,66929,67103,67137,67249,67312,67417,67456,67526,67551,67587,67628,67664,67824,68142,68229,68358,68422,68667,68672,68760,68966,69022,69045,69227,69300,69391,69444,69555,69583,69599,69620,69794,69864,69882,69906,70151,70185,70294,70490,70527,70567,70573,70601,70656,70690,70694,70753,70835,70876,71056,71141,71161,71313,71366,71419,71430,71452,71453,71592,71630,71644,71665,71672,71775,71790,71874,71992,72012,72045,72276,72371,72417,72517,72570,72740,72746,72757,72947,72954,73064,73135,73155,73166,73176,73194,73249,73286,73293,73325,73481,73569,73570,73662,73669,73707,73751,73757,73799,73916,73932,74012,74251,74312,74315,74423,74504,74511,74515,74516,74532,74751,74787,75255,75565,75637,75695,75916,75917,75939,75970,76097,76184,76419,76482,76500,76584,76703,76724,76764,76775,76781,76924,76937,76942,77050,77083,77142,77419,77465,77523,77675,77732,77738,77744,77762,77891,77901,77921,77941,77975,77994,78135,78159,78204,78251,78306,78325,78408,78434,78616,78618,78628,78738,78828,78885,78926,78937,79111,79272,79308,79326,79362,79376,79396,79435,79438,79472,79523,79550,79588,79673,79700,79855,79929,80073,80074,80166,80189,80190,80233,80292,80348,80531,80578,80668,80803,80934,81160,81208,81391,81392,81449,81475,81528,81578,81598,81627,81739,81761,81829,82057,82105,82217,82389,82411,82421,82440,82514,82594,82599,82625,82774,82831,82902,82931,82983,83056,83269,83319,83493,83600,83877,83972,84059,84221,84258,84329,84331,84404,84543,84626,84710,84755,84799,84816,84817,84906,84930,84952,84957,85037,85076,85141,85285,85459,85468,85499,85571,85593,85629,85798,85904,85921,85950,86038,86177,86278,86283,86288,86315,86356,86466,86645,86811,86838,86887,86916,86936,87071,87204,87231,87369,87547,87593,87741,87792,88014,88165,88168,88193,88250,88325,88374,88566,88576,88578,88591,88720,88723,88809,89006,89050,89095,89176,89230,89356,89517,89578,89625,89760,89821,89862,89982,90061,90129,90272,90274,90304,90333,90360,90482,90686,90704,90726,90764,90767,90799,90841,90921,90967,91132,91262,91286,91387,91463,91496,91577,91674,91680,91747,91785,91805,91902,91956,92014,92028,92041,92096,92234,92236,92262,92265,92573,92714,92728,92732,92809,92832,92834,92872,92873,93018,93425,93944,94136,94173,94202,94242,94292,94405,94454,94476,94479,94514,94587,94699,94700,94929,94982,95078,95128,95340,95431,95458,95531,95586,95611,95614,95673,95809,95843,96062,96097,96148,96207,96268,96327,96350,96477,96486,96499,96507,96551,96552,96586,96652,96668,96683,96720,96883,97048,97228,97251,97329,97687,97701,97704,97879,98106,98174,98289,98303,98484,98563,98571,98660,98677,98731,98734,98820,98941,99075,99175,99218,99287,99391,99698,99713,99903,99930",
          "5,9,23,26,250,313,506,518,624,628,667,950,1266,1278,1288,1289,1292,1302,1354,1370,1458,1463,1510,1522,1586,1809,1811,1813,1836,1844,1848,1899,2081,2242,2247,2250,2291,2296,2420,2478,2517,2692,3183,3389,3481,3528,3696,3779,3812,3832,3836,3869,4377,4691,4717,4724,4814,4847,4911,4947,5115,5124,5141,5282,5329,5384,5390,5398,5415,5423,5456,5479,5547,5658,5718,5808,5952,5989,6032,6056,6110,6189,6194,6215,6223,6232,6255,6279,6288,6353,6376,6420,6421,6443,6450,6465,6473,6538,6558,6560,6574,6602,6661,6707,6727,6737,6860,7145,7206,7365,7370,7426,7432,7492,7576,7595,7597,7612,7749,7763,7782,7816,7855,7890,8033,8062,8074,8180,8249,8251,8255,8332,8483,8530,8561,8629,8684,8956,8985,9085,9326,9337,9369,9408,9470,9771,9868,9902,9982,9999,10323,10330,10344,10415,10497,10553,10567,10662,10694,10696,10841,10942,11067,11297,11312,11351,11584,11675,11785,11968,11979,12439,12676,12749,12821,12824,12857,12941,13004,13831,13984,14120,14133,14225,14399,14430,14537,14683,14887,14942,14960,15059,15225,15308,15546,15590,15783,15794,15971,16060,16766,16856,17242,17468,17643,17668,17688,17698,18110,18139,18180,18569,18747,18994,19518,19796,19865,19967,20037,20167,20803,21018,21093,21108,21479,21531,21548,21648,22319,22601,22718,22829,22865,23191,23288,23311,23630,23854,24077,24725,24748,24846,25032,25169,25359,25493,25666,26244,26360,26629,26632,26685,26723,26733,26768,26779,26839,26862,26876,26889,27015,27081,27137,27175,27322,27371,27393,27620,27702,27819,28095,28269,28394,28437,28486,28715,28851,28874,29245,29340,29342,29472,29533,29535,29588,29662,29759,29775,29864,30659,31009,31647,31790,31879,32343,32401,33068,33909,35338,35466,35634,35807,36137,36309,36318,36619,36679,37012,37128,37187,37193,37308,37310,37384,37411,37450,37763,37868,37941,38059,38312,38315,38364,38372,38466,38559,38748,39065,39254,39347,39542,39733,39738,39777,39988,40551,40783,42143,42546,42691,42911,42916,42984,43151,43253,43315,43336,43495,43498,43627,43643,43734,43749,43812,44421,44872,45007,45030,45123,45270,45316,45510,45511,45563,45608,45685,45790,46600,47354,47377,49089,49225,50069,50198,50256,50277,50394,50568,50586,50626,50745,51024,51026,51184,51398,51403,51503,51527,51553,51848,51904,52035,52139,52219,52266,52368,52407,52455,52502,52509,52632,52696,52794,52876,52940,53003,53081,53121,53182,53324,53398,53509,53527,53571,53574,53586,53611,53658,53875,54102,54170,54253,54303,54381,54414,54451,54535,54602,54793,54890,54904,54907,55097,55222,55260,55262,55324,55419,55437,55516,55596,55807,56003,56009,56022,56054,56339,56501,56729,56826,56902,57045,57056,57249,57304,57379,57388,57411,57462,57684,57751,57794,57839,58220,58222,58255,58387,58481,59035,59089,59395,59518,59726,59792,59838,59989,60100,60232,60239,60243,60303,60435,60487,60548,60572,60574,60675,60962,61024,61182,61246,61276,61317,61478,61538,61672,61783,61839,61884,61890,62050,62103,62122,62167,62390,62543,62549,62621,62668,62723,62760,62870,62886,63067,63111,63170,63237,63257,63306,63358,63365,63416,63528,63540,63638,63820,63839,63852,63902,63938,64066,64080,64197,64261,64276,64334,64417,64517,64526,64554,64575,64658,64714,64788,64906,64942,64989,65081,65190,65244,65325,65594,65635,65764,65893,65924,65956,66162,66373,66414,66425,66446,66554,66850,67014,67172,67315,67355,67453,67516,67521,67609,67781,67857,67867,68099,68130,68136,68329,68330,68355,68369,68575,68591,69058,69077,69151,69203,69297,69349,69380,69468,69550,69611,69715,69745,69800,69897,69909,69925,69975,70033,70097,70159,70169,70204,70300,70364,70468,70480,70568,70577,70633,70804,70810,70870,71079,71106,71120,71235,71303,71405,71414,71420,71439,71451,71556,71706,71834,71977,71982,72003,72217,72272,72546,72630,72714,72719,72805,73014,73138,73199,73300,73344,73367,73403,73561,73565,73706,73801,74072,74093,74140,74160,74178,74230,74358,74476,74695,74836,74995,75026,75071,75079,75238,75337,75359,75377,75451,75527,75624,75709,75723,75787,75858,75918,76194,76238,76285,76291,76343,76384,76459,76464,76631,76836,76890,76901,76913,76921,76948,76951,77051,77118,77197,77217,77236,77341,77344,77503,77575,77591,77641,77653,77683,77734,77745,77749,77916,78065,78106,78210,78453,78530,78576,78637,78641,78643,78646,78719,78729,78736,78941,78956,79107,79178,79216,79333,79350,79693,79798,79799,79838,79964,79974,79999,80011,80012,80072,80158,80200,80213,80303,80313,80399,80460,80667,80700,80767,80773,80892,80938,81117,81136,81216,81263,81276,81372,81453,81592,81647,81732,82068,82283,82345,82465,82577,82741,83182,83212,83305,83353,83557,83574,83770,83779,83842,83865,84061,84106,84285,84315,84348,84453,84676,84685,84722,84730,84744,84885,84954,85003,85057,85120,85127,85199,85255,85341,85458,85462,85599,85640,85941,86078,86135,86199,86236,86261,86291,86394,86548,86557,86614,86650,86664,86680,86738,86881,86962,87056,87165,87207,87446,87506,87546,87669,87721,87853,87862,87867,87896,87966,88011,88066,88070,88079,88116,88181,88339,88458,88492,88558,88601,88677,88758,88902,88966,88973,89053,89062,89090,89297,89393,89411,89456,89709,89719,89741,89778,89878,89930,89954,89967,90031,90124,90158,90243,90256,90286,90447,90526,90541,90616,90658,90833,90925,90933,91079,91157,91160,91210,91257,91517,91640,91672,91689,91777,91801,91995,92111,92112,92114,92115,92152,92351,92360,92375,92612,92729,92763,92810,92831,92893,92924,93103,93134,93163,93168,93182,93248,93424,93715,93764,93766,93769,93779,93901,93934,93936,94023,94144,94194,94272,94280,94293,94323,94400,94449,94455,94464,94862,94870,95185,95199,95260,95341,95348,95521,95529,95570,95630,95691,95934,95972,96103,96137,96241,96269,96270,96303,96411,96583,96664,96749,96854,97253,97349,97379,97387,97393,97569,97577,97715,97989,98019,98067,98083,98262,98392,98479,98486,98524,98537,98539,98631,98638,98706,98719,98842,98904,99042,99106,99251,99415,99521,99599,99711,99715,99785,99792,99796,99850,99933,99984",
          "123,183,193,202,537,598,623,707,872,891,1077,1135,1447,1601,1690,1739,1806,2067,2131,2194,2274,2372,2434,2450,2459,2575,2781,2820,2875,2936,3217,3309,3703,3905,4164,4205,4256,4284,4385,4402,4431,4448,4515,4573,4637,4734,5004,5057,5068,5243,5271,5294,5503,5521,5611,5696,5759,5772,5804,5861,5941,6019,6061,6073,6176,6187,6365,6399,6431,6434,6444,6463,6552,6576,6580,6739,6840,6869,6918,7046,7198,7283,7407,7427,7434,7533,7664,7678,7692,7811,7891,7920,7930,7999,8032,8040,8067,8080,8089,8122,8223,8278,8324,8413,8450,8486,8608,8630,8735,8792,8801,8859,8890,8907,8924,8980,9052,9199,9231,9248,9301,9368,9375,9398,9422,9426,9444,9504,9529,9585,9731,9797,9923,10242,10355,10370,10377,10413,10433,10595,10660,10672,10724,10775,10818,10847,10862,11016,11062,11105,11181,11194,11235,11287,11300,11314,11389,11440,11565,11682,11700,11745,11873,11954,11975,12032,12074,12084,12145,12177,12216,12404,12520,12563,12723,12742,12809,12840,12846,12876,12932,13001,13009,13010,13055,13072,13075,13147,13160,13218,13287,13396,13426,13439,13498,13511,13836,13896,13952,13970,14009,14084,14224,14555,14662,14681,14704,14714,14773,14880,14978,15294,15426,15611,15774,15779,15803,15973,16055,16537,16711,16724,16876,16977,16983,17104,17144,17165,17445,17597,17723,17749,17829,17950,18144,18466,18541,18546,18696,18784,18790,18795,18797,18805,19016,19080,19177,19185,19250,19279,19337,19670,19689,19784,19853,19993,20004,20176,20200,20253,20544,20563,20575,20734,20781,20831,20868,20974,21057,21225,21398,21418,21421,21512,21603,21609,21616,21699,21776,21853,21962,22053,22776,22823,22852,22870,23021,23031,23209,23249,23257,23269,23422,23527,23677,23678,23692,23850,24104,24136,24417,24663,24720,24886,24900,24966,25036,25066,25076,25176,25187,25193,25330,25389,25412,25612,25724,25767,25820,25845,26052,26111,26156,26289,26353,26696,26700,26903,26965,27052,27055,27286,27532,27762,27782,27860,27903,27927,28234,28277,28297,28334,28372,28474,28509,28676,28790,28792,28914,29067,29213,29273,29282,29415,29484,29647,29710,29798,29820,29911,29995,30592,31032,31329,31764,32265,32753,33246,33968,34324,35088,35301,35380,35501,35550,35565,35630,35632,35702,35801,35879,35888,35988,36002,36047,36185,36188,36246,36449,36570,36634,36640,36811,36816,36819,36828,36934,36939,37004,37057,37129,37351,37379,37387,37551,37568,37575,37668,37691,37761,37867,38064,38186,38255,38362,38629,38632,38690,38936,38940,38985,39056,39070,39121,39168,39245,39246,39255,39427,39433,39443,39536,39549,39565,39606,39740,39751,39809,39814,39815,39879,42150,42210,42303,42459,42475,42490,42495,42549,42559,42617,42789,42879,42944,42960,43150,43270,43279,43284,43478,43560,43563,43676,43908,44021,44135,44221,44286,44294,44340,44350,44386,44512,44521,44536,44609,44685,44746,44756,44816,44849,44850,44866,45081,45091,45505,45528,45547,45561,45661,45663,45845,45888,45890,46534,46625,46908,47533,47925,48014,48233,49466,49546,50009,50051,50070,50155,50182,50238,50294,50498,50819,50915,51104,51238,51257,51263,51298,51517,51530,51591,51609,51655,51725,51823,51893,52028,52043,52306,52515,52524,52545,52684,52859,52864,52915,53038,53062,53117,53125,53157,53230,53408,53415,53448,53508,53656,53776,53793,53795,53799,53862,53942,54053,54273,54302,54374,54422,54457,54503,54597,54663,54721,55025,55145,55332,55436,55468,55482,55813,55837,55867,55901,55903,55999,56131,56195,56218,56226,56429,56604,56670,56862,56995,57029,57094,57149,57211,57221,57288,57610,57612,57754,57826,58339,58342,58407,58608,58618,58699,58780,58960,58971,59107,59143,59400,59463,59475,59690,59790,59821,59833,60023,60049,60132,60364,60498,60503,60580,60617,60653,60749,60762,60834,60869,60999,61002,61012,61042,61259,61411,61511,61560,61638,61693,61886,62128,62211,62251,62403,62611,62617,62624,62694,62695,63113,63215,63323,63338,63476,63533,63561,63699,63841,63847,64175,64307,64322,64568,64632,64758,65028,65057,65102,65111,65246,65538,65572,65829,66212,66325,66392,66507,66559,66613,66644,66707,66780,66831,66997,67083,67102,67259,67373,67799,67897,67907,67941,68090,68135,68214,68215,68227,68297,68394,68436,68568,68859,68925,68944,68953,69110,69260,69272,69367,69397,69541,69668,69721,69755,69839,69894,69963,70125,70233,70429,70432,70433,70434,70479,70523,70769,70781,70812,70831,70851,70864,70884,70948,70961,71062,71083,71101,71425,71455,71568,71648,71715,71736,71889,71912,71978,71995,72064,72144,72158,72303,72309,72346,72540,72713,72801,72903,72933,73022,73051,73068,73163,73280,73290,73390,73622,73630,73691,73785,73789,73882,73902,73952,74007,74062,74154,74307,74512,74530,74693,74701,74714,74722,74734,74775,74829,74955,75083,75105,75310,75346,75368,75618,75657,75662,75735,75950,75964,75966,76053,76131,76405,76457,76554,76574,76608,76609,76659,76685,76790,76801,76813,76928,76959,76983,77000,77028,77065,77080,77085,77116,77377,77441,77450,77671,77842,78015,78107,78114,78170,78299,78577,78596,78633,78664,78782,79015,79053,79091,79251,79255,79373,79617,79654,79656,79681,79718,79849,79903,79934,79955,80009,80116,80273,80382,80435,80455,80475,80533,80907,80912,81082,81130,81428,81436,81446,81512,81597,81667,81730,81779,81807,81956,81988,82123,82157,82228,82252,82286,82382,82629,82633,82669,82685,82711,82724,82737,82833,82872,82919,82946,83066,83179,83221,83227,83241,83385,83709,83852,83947,84025,84111,84159,84362,84438,84578,84791,84891,85263,85476,85539,85801,85820,85893,86020,86050,86067,86155,86225,86322,86350,86435,86443,86467,86556,86695,86754,86790,86823,86917,87059,87065,87101,87185,87194,87228,87251,87297,87337,87377,87499,87588,87768,88199,88296,88302,88352,88442,88543,88590,88703,88821,88863,88963,88979,89143,89256,89464,89579,89694,89817,90068,90172,90712,90795,90800,90890,90894,91148,91167,91192,91240,91242,91245,91251,91270,91289,91359,91380,91427,91484,91578,91579,91628,91821,91938,91957,92190,92203,92329,92413,92458,92471,92954,92992,93051,93069,93165,93200,93298,93300,93581,93644,93653,93837,93841,93852,93887,93946,94030,94176,94178,94180,94348,94358,94402,94475,94536,94577,94634,94938,95056,95081,95110,95216,95235,95286,95289,95454,95480,95559,95575,95605,95644,95654,95820,95872,95883,95970,95989,96011,96090,96131,96337,96359,96404,96534,96645,96746,96775,96777,96786,96848,96911,96943,97113,97174,97362,97503,97553,97561,97614,97659,97855,97893,97926,97927,97953,98024,98129,98166,98258,98265,98287,98324,98342,98507,98909,98996,99029,99195,99205,99207,99539,99613,99696,99752,99827",
          "209,318,366,397,415,417,441,519,564,578,581,607,815,860,869,893,975,990,996,999,1038,1185,1249,1269,1346,1361,1422,1446,1459,1515,1578,1879,1890,3567,4676,5040,5090,5105,5202,5206,5247,5360,5447,5729,5826,5919,6082,6369,6406,6432,6486,6505,6509,6550,6565,6627,6652,6673,6674,6689,6724,6735,6826,7016,7053,7066,7073,7075,7193,7273,7291,7310,7345,7356,7461,7541,7593,7689,7709,7835,7898,7921,8081,8199,8281,8302,8401,8409,8468,8477,8547,8675,8747,8967,8990,9001,9058,9070,9088,9154,9185,9188,9380,9454,9785,9976,10002,10247,10279,10450,10461,10494,10572,10576,10632,10735,10790,10792,10917,10939,10979,11055,11142,11199,11286,11659,11687,11817,11820,11890,11974,11987,12063,12128,12197,12374,12521,12561,12689,12718,12870,12933,13127,13401,13525,13874,14406,14952,15051,15359,15679,15753,16139,16385,16718,16739,16858,16985,17010,17207,17326,17571,17684,18018,18034,18483,18685,18701,18942,19547,20011,20538,20727,21070,21260,21285,21543,21598,21634,21981,22038,22119,22192,23340,23416,23564,24766,24818,24895,25172,25796,26413,26796,26822,26835,26851,26961,26977,27074,27216,27690,28412,28475,28697,28835,29113,29748,29845,36114,36193,36195,36196,36385,36439,36571,36734,37558,37637,37639,37672,37698,37799,38005,38297,38509,38610,38677,38921,39068,39111,39440,39442,39882,45379,45488,45615,45855,45878,49943,50092,50165,50175,50219,50775,50878,50897,51120,51220,51518,51557,51564,51607,51620,51653,51753,51787,51831,51856,52021,52106,52156,52243,52249,52260,52324,52331,52411,52518,52551,52574,52710,52755,52844,53009,53013,53071,53094,53098,53130,53241,53457,53544,53593,53731,53753,54013,54316,54332,54565,54595,54833,54991,55020,55081,55321,55327,55341,55434,55651,55653,55658,55697,55783,55818,55850,55900,56011,56044,56059,56296,56358,56444,56498,56519,56543,56584,56591,56611,56640,56662,56765,56936,57093,57178,57183,57252,57330,57395,57626,57729,57811,57859,57909,58034,58181,58296,58341,58485,58639,58753,58800,58860,58975,59038,59043,59055,59068,59105,59277,59286,59411,59555,59605,59607,59721,59853,59855,60079,60135,60152,60164,60211,60227,60462,60527,60537,60599,60605,60619,60663,60664,60694,60787,61256,61284,61387,61454,61515,61563,61590,61617,61628,61636,61722,61745,61811,61813,61853,62014,62109,62297,62309,62342,62361,62364,62467,62479,62584,62707,62743,62747,62748,62887,63007,63073,63158,63204,63232,63369,63425,63493,63560,63571,63593,63644,63695,63731,63764,63770,63775,63849,63851,64040,64083,64093,64118,64413,64704,64768,64905,64979,64980,65001,65032,65103,65151,65202,65410,65493,65523,65637,65702,65767,65798,65804,65812,66000,66010,66145,66146,66235,66548,66676,66983,67098,67147,67279,67448,67502,67642,67820,67871,67915,68012,68033,68049,68133,68296,68354,68376,68393,68398,68433,68463,68651,68697,68718,68763,68913,68958,68999,69044,69132,69276,69374,69385,69463,69498,69530,69531,69614,69672,69765,69855,70010,70023,70105,70192,70193,70546,70604,70666,70693,70784,70913,70945,70971,70987,71066,71172,71200,71259,71380,71394,71424,71525,72400,72575,72677,72793,72875,72881,73237,73466,73513,73566,73607,73633,73650,73675,73856,73917,73997,74074,74225,74233,74249,74367,74394,74448,74663,74747,74871,74887,74980,75027,75125,75228,75334,75389,75447,75452,75459,75580,75590,75667,75776,75802,75849,75850,75868,75898,76006,76068,76141,76244,76251,76324,76390,76421,76518,76535,76555,76859,76910,76987,77064,77183,77212,77468,77496,77627,77652,77680,77829,77858,78262,78278,78359,78471,78486,78506,78513,78540,78723,79034,79055,79085,79124,79514,79569,79610,79622,79757,80036,80126,80230,80272,80309,80334,80463,80501,80771,80835,80922,81001,81053,81188,81278,81312,81351,81435,81586,81759,81778,81881,81892,81894,82050,82059,82101,82234,82363,82488,82491,82624,82631,83046,83190,83242,83384,83430,83487,83520,83534,83675,83823,83854,83941,84069,84086,84122,84288,84371,84456,84488,84490,84544,84616,84643,84666,84827,84863,84882,85187,85326,85409,85516,85575,85577,85637,85712,85894,86006,86059,86211,86269,86285,86292,86301,86460,86547,86594,86736,86829,86896,86950,87018,87182,87253,87320,87464,87559,87569,87720,87754,87774,87798,87918,87987,87999,88122,88136,88231,88243,88319,88391,88551,88711,89235,89300,89322,89400,89580,89596,89598,89610,89615,89642,89648,89776,89807,89920,89959,90008,90063,90136,90196,90342,90420,90458,90481,90484,90637,90828,90840,90937,91012,91221,91296,91351,91356,91384,91531,91546,91576,91769,91812,91836,91879,91882,92036,92095,92198,92238,92279,92281,92334,92340,92397,92489,92554,92710,92784,92823,92869,92894,93028,93129,93304,93365,93381,93435,93540,93578,93605,93698,93775,93811,93860,93953,94034,94063,94266,94300,94313,94376,94394,94425,94495,94545,94568,94694,94816,94835,94953,95203,95299,95441,95505,95642,95650,95670,95885,95892,95906,95952,96024,96077,96236,96256,96311,96319,96378,96383,96410,96581,96600,96698,96799,96824,96835,96930,97058,97071,97140,97178,97212,97225,97298,97457,97674,97727,97738,97756,97814,97824,97887,97906,97914,97936,97939,98087,98210,98299,98374,98395,98449,98583,98592,98597,98666,98783,98789,98949,98955,98962,99000,99023,99313,99354,99458,99590,99644,99645,99656,99692,99718,99781,99795,99811",
          "4,269,314,394,401,410,484,491,524,574,696,712,720,748,761,765,887,899,925,933,985,1018,1071,1104,1212,1328,1466,1487,1526,1624,1724,1794,1807,1857,1909,1915,2925,2930,2966,3186,3413,3507,3587,3608,3917,4237,4247,5135,5214,5239,5240,5258,5307,5322,5410,5450,5562,5586,5645,5724,5770,5829,5870,5886,5993,6037,6099,6108,6141,6151,6228,6320,6339,6820,6893,6998,7117,7127,7134,7164,7220,7326,7338,7451,7501,7554,7573,7577,7611,7615,7638,7998,8039,8046,8131,8182,8189,8191,8198,8211,8227,8325,8391,8425,8439,8542,8577,8588,8663,8798,8987,9072,9179,9214,9236,9336,9345,9636,9904,10001,10171,10195,10224,10282,10333,10354,10389,10395,10402,10421,10427,10477,10479,10511,10528,10571,10585,10709,10879,10897,11107,11207,11288,11591,11710,11713,11715,11757,11819,11929,12732,12838,12953,13626,15145,15478,15790,15865,16315,17731,17857,18843,19192,19989,20066,21275,22740,22831,22855,22937,23089,24259,24286,24881,25130,25443,25743,26194,26405,27057,27184,27374,27847,28322,28349,28553,28721,30273,32019,32226,34139,35020,35561,36073,36171,36382,36700,37184,37613,39423,40385,42476,44693,45631,48228,48269,49417,49481,49580,50020,50109,50160,50295,50369,50557,50595,50752,50761,50799,50800,50824,50920,50984,51072,51085,51367,51497,51595,51715,51732,51751,51786,51846,51965,51979,51985,52011,52116,52154,52193,52300,52334,52356,52718,52778,52796,52843,52926,52929,52967,53031,53114,53128,53144,53149,53578,53654,53949,54069,54089,54185,54481,54505,54525,54629,54678,54682,55018,55070,55075,55103,55146,55370,55442,55478,55568,55604,55691,55801,56166,56265,56469,56542,56579,56620,56755,56845,57064,57309,57353,57453,57579,57607,57627,57632,57721,57732,57889,57991,58018,58128,58486,58538,58552,58885,59072,59102,59213,59340,59341,59351,59600,59630,59765,59904,59931,59936,59941,59951,60012,60131,60177,60308,60511,60610,60724,60874,60912,61181,61222,61299,61326,61356,61409,61424,61501,61581,61597,61607,61660,61712,61823,61914,61971,62134,62184,62344,62359,62387,62452,62526,62625,62633,62681,62714,62734,62892,62973,63033,63147,63264,63345,63423,63551,63637,63641,63658,63832,63843,63864,63887,63892,63975,64132,64209,64249,64323,64331,64461,64684,64778,64871,64912,64939,65022,65140,65458,65476,65527,65546,65583,65591,65913,65945,65987,66009,66080,66220,66241,66265,66279,66297,66384,66433,66516,66572,66611,66695,66846,67239,67332,67376,67378,67398,67447,67536,67566,67627,67637,67833,67834,67906,67940,67942,67996,68275,68292,68492,68537,68605,68773,68825,69035,69060,69085,69130,69266,69274,69447,69493,69511,69540,69570,69703,69719,69776,69850,69877,69883,69966,69981,70158,70163,70389,70440,70471,70592,70698,70737,70755,70827,70893,70952,70997,71135,71224,71242,71256,71263,71323,71437,71463,71509,71551,71572,71583,71642,71705,71777,71881,72178,72277,72386,72529,72577,72589,72703,72716,72777,72784,72920,73045,73102,73303,73491,73552,73584,73620,73645,73698,73732,73921,74015,74116,74138,74229,74239,74816,74817,74858,75053,75060,75182,75194,75200,75216,75278,75320,75426,75458,75497,75572,75654,75817,75929,76072,76082,76106,76182,76205,76214,76321,76322,76332,76643,76768,77066,77388,77467,77492,77543,77621,77622,77660,77773,77821,77833,77902,78068,78090,78131,78353,78423,78489,78564,78626,78667,78686,78921,79027,79276,79325,79527,79763,79766,79770,79800,79817,79890,79991,80259,80467,80643,80812,80878,80881,80906,80916,80987,81143,81407,81488,81554,81666,81689,81741,81815,81957,82009,82053,82070,82111,82377,82464,82597,82602,82635,82646,82653,82690,83075,83343,83620,83632,83637,83706,83777,83913,84040,84126,84207,84230,84279,84361,84387,84501,84534,84556,84568,84663,84861,84904,84933,84947,85079,85130,85186,85297,85431,85463,85473,85540,85553,85595,85825,85851,86042,86091,86112,86257,86276,86387,86399,86428,86439,86518,86529,86561,86565,86720,86813,86852,87009,87309,87439,87520,87543,87587,87815,87860,87928,88177,88343,88495,88609,88690,88692,88700,88800,88893,89007,89023,89034,89035,89082,89279,89320,89404,89490,89540,89543,89779,89787,89844,89869,89874,89911,90106,90293,90387,90575,90626,90694,91144,91308,91325,91497,91560,91670,91741,91949,91954,92104,92142,92226,92317,92662,92767,92838,92849,92878,93047,93156,93291,93374,93431,93592,93614,93691,93721,93798,93986,94010,94127,94285,94342,94364,94451,94617,94836,94858,95006,95092,95104,95126,95129,95148,95284,95310,95394,95416,95423,95434,95466,95484,95635,95647,95718,95755,95834,95908,95924,96059,96133,96322,96342,96469,96518,96607,96713,96729,96887,96905,96981,96987,97149,97241,97263,97307,97334,97360,97373,97383,97473,97478,97501,97601,97753,97794,97967,98066,98266,98373,98501,98574,98621,98787,98992,99072,99119,99177,99179,99293,99325,99336,99442,99489,99492,99493,99499,99510,99637,99640,99728,99797",
          "92,283,508,517,531,569,629,746,753,755,840,958,1132,1144,1223,1235,1351,1418,1439,1508,1517,1577,1847,1864,1943,2298,2494,2497,2563,2765,2803,2853,2937,2977,3083,3129,3161,3195,3229,3300,3390,3579,3687,4157,4264,4281,4430,4658,4700,4755,4775,4824,4842,5032,5089,5101,5109,5167,5188,5249,5297,5303,5318,5319,5362,5363,5413,5421,5431,5433,5436,5452,5460,5470,5482,5511,5625,5693,5714,5715,5738,5758,5773,5781,5790,5875,5926,5935,5953,5956,5984,5995,6106,6121,6200,6203,6216,6226,6250,6316,6366,6378,6408,6489,6504,6651,6710,6783,6842,6867,6880,6888,6931,6942,7004,7017,7052,7133,7184,7269,7312,7385,7398,7440,7455,7727,7734,7754,7852,7860,7918,8099,8106,8109,8233,8263,8283,8430,8523,8568,8575,8601,8625,8739,9209,9263,9355,9446,9589,9679,9908,9960,10213,10361,10458,10459,10480,10546,10608,10631,10923,10959,11014,11407,11582,11749,11802,11834,12206,12309,12344,12416,12491,12508,12585,12586,12967,13066,13148,13349,13447,14467,14472,14735,15301,15388,15570,15601,15612,15802,16084,16410,17034,17226,17512,17548,17560,17911,18025,18519,18604,18874,19182,19184,19428,19710,19751,20109,20533,20700,20798,21130,22081,22115,23054,23604,23794,23934,25625,25719,25846,25851,26190,26254,26310,26686,27024,27167,27282,27288,27452,27765,27800,27831,28213,28275,28443,28682,28716,28735,28896,29482,29805,29951,30574,30828,31123,31444,32053,32807,33354,33619,33648,35379,35613,35792,35820,35889,36308,36678,37060,37570,37752,38810,38863,39183,39256,39370,39477,39734,39805,39811,39990,40638,41450,41478,41507,42464,42528,42705,42873,43371,43493,43589,43681,44262,44497,44621,45307,45625,45923,46283,46724,46872,47411,49566,50105,50269,50398,50422,50448,50609,50610,50634,50699,50803,50809,50877,50900,50910,51004,51060,51113,51459,51513,51550,51659,51674,51685,51900,52017,52049,52152,52387,52441,52471,52672,52856,52945,53279,53315,53492,53516,53649,53664,53783,53932,53939,54059,54216,54235,54320,54614,54639,54670,54701,54763,54828,54836,54862,54927,55083,55175,55304,55411,55520,55569,55633,55714,55827,55835,55872,56033,56191,56294,56431,56509,56679,56683,56747,56835,56938,57033,57085,57131,57291,57372,57438,57456,57533,57563,57565,57652,57779,57781,57798,57842,57911,57966,58155,58226,58239,58269,58453,58461,58502,58559,58571,58578,58794,58840,58882,58941,59044,59070,59114,59136,59158,59236,59296,59377,59553,59624,59641,59653,59769,59857,59858,59955,60105,60109,60204,60237,60331,60479,60575,60584,60603,60735,60980,61008,61015,61254,61421,61492,61504,61651,61837,61926,62271,62300,62388,62396,62427,62440,62627,62631,62890,62981,63001,63005,63030,63182,63308,63312,63363,63650,63744,63842,63855,63925,64145,64242,64434,64525,64529,64569,64629,64693,64726,64727,64736,64802,64916,65041,65155,65259,65482,65499,65568,65592,65663,65814,65842,65897,65914,65944,65972,66015,66071,66073,66103,66183,66252,66413,66470,66589,66789,67142,67225,67256,67496,67519,67561,67692,67803,67816,67989,68141,68157,68222,68291,68375,68385,68406,68452,68464,68589,68644,68696,68722,68781,68830,68843,68864,68989,68994,68996,69039,69043,69059,69093,69243,69298,69395,69476,69537,69588,69643,69683,69774,70026,70042,70253,70450,70459,70484,70508,70709,70740,70868,70929,70943,71105,71211,71332,71359,71367,71403,71460,71619,71666,71787,71810,71894,72078,72135,72293,72311,72550,72590,72695,72741,72749,72794,72918,72949,73105,73386,73410,73502,73529,73545,73715,73840,73870,74047,74083,74094,74097,74237,74259,74280,74300,74342,74354,74456,74533,74567,74641,74678,74709,74728,74800,74945,75066,75193,75469,75485,75927,75973,76009,76054,76055,76123,76165,76488,76517,76575,76653,76793,76822,76879,76925,77100,77243,77271,77843,77861,78081,78395,78480,78527,78699,78718,78767,78775,78820,78874,78895,78901,79026,79214,79269,79511,79596,79611,79774,79901,80113,80178,80216,80261,80276,80332,80491,80498,80632,80662,80674,80815,80858,80877,80920,80965,81099,81212,81303,81318,81458,81545,81697,81708,81749,81822,81824,81984,82001,82003,82014,82215,82263,82271,82415,82606,82870,82910,82951,83042,83060,83127,83171,83290,83342,83453,83726,83763,83776,83801,83864,83866,83945,84074,84309,84345,84354,84365,84424,84633,84718,84752,84793,84830,84845,84916,85146,85158,85182,85353,85430,85572,85628,85657,85665,85686,85787,85793,85876,86103,86184,86329,86462,86493,86502,86619,86631,86840,86858,86941,87053,87083,87109,87217,87226,87296,87300,87405,87583,87596,87702,87797,87919,87938,87989,88078,88112,88413,88433,88542,88618,88793,88819,88978,89047,89049,89144,89376,89467,89483,89494,89512,89513,89558,89716,89748,90003,90165,90209,90269,90281,90376,90395,90422,90431,90512,90563,90587,90673,90708,90745,90811,90966,91034,91202,91236,91309,91400,91434,91572,91819,92032,92098,92110,92113,92120,92121,92161,92312,92355,92391,92424,92433,92480,92485,92929,92942,92964,93017,93025,93040,93297,93333,93361,93364,93385,93433,93555,93632,93648,93842,93941,94002,94079,94145,94204,94219,94297,94395,94491,94500,94510,94560,94569,94644,94661,94719,94772,94787,95091,95147,95191,95411,95637,95639,95669,95705,95721,95804,95812,95863,95897,96192,96351,96471,96497,96564,96590,96596,96701,96747,96773,96914,97220,97290,97297,97406,97484,97572,97898,97928,97942,98095,98198,98212,98217,98413,98437,98454,98551,98622,98651,98841,98843,98848,98888,98911,98984,99079,99159,99216,99281,99317,99332,99339,99409,99515,99641",
          "63,325,353,429,476,554,588,717,788,857,876,881,930,942,1039,1091,1154,1252,1279,1295,1489,1530,1579,1602,1604,1637,1686,1758,1863,1896,2225,4214,5006,5055,5059,5063,5187,5266,5325,5418,5576,5614,5667,5703,5817,5837,5911,5945,5959,5981,6074,6284,6299,6304,6317,6323,6326,6334,6362,6447,6461,6590,6730,6780,6803,6875,6978,7049,7081,7139,7270,7336,7433,7443,7468,7656,7665,7726,7815,7823,7842,7866,8017,8061,8119,8124,8134,8161,8181,8215,8243,8277,8358,8380,8453,8460,8539,8657,8709,8716,8722,8727,8738,8753,8874,8922,8926,8954,8989,8997,9174,9183,9221,9350,9765,9885,10138,10159,10160,10164,10320,10336,10468,10489,10566,10587,10740,10781,10878,11371,11383,11436,11467,11609,11678,11751,11859,11962,12070,12136,12173,12373,12489,12610,12719,13116,13221,13268,13371,13413,13481,13495,13529,13911,14083,14097,14595,14599,14699,14764,14861,14881,14906,14968,15163,15239,15315,15408,15453,15488,15523,15766,15887,16111,16306,16646,16690,16733,17344,17346,17416,17521,17613,17666,18238,18865,19403,19470,19517,20100,20221,20542,20549,20588,20628,21033,21127,21192,21528,21672,21745,22107,22707,22805,22923,23112,23139,23263,23272,24146,24159,24176,24605,25145,25205,25219,25458,26281,26799,27061,27541,27834,28252,28362,28439,28819,28938,29001,29006,29134,29159,29209,29615,29793,29938,34203,35752,35870,35894,35941,36575,36576,36636,36695,36824,36923,36948,37017,37431,38010,38236,38278,38571,38939,38997,39069,39117,39132,39488,39607,39617,39674,39940,39983,41282,45421,45621,45807,45990,50037,50099,50147,50196,50201,50264,50297,50313,50393,50546,50653,50710,50805,50882,50976,50999,51493,51665,51666,51696,51745,51936,51964,52079,52089,52114,52322,52333,52543,52596,52599,52627,52683,52729,52765,52772,52774,52823,52974,52990,53002,53141,53195,53468,53493,53562,53683,53722,53788,53801,53893,53900,53985,54009,54024,54037,54120,54218,54228,54242,54509,54557,54601,54659,54669,54771,55096,55100,55114,55125,55131,55138,55425,55546,55562,55593,56053,56057,56132,56170,56256,56355,56506,56623,56664,56701,56797,57034,57102,57452,57483,57547,57634,57665,57667,57782,57852,57951,58044,58179,58215,58270,58273,58278,58397,58401,58444,58556,58596,58767,58848,59073,59179,59187,59301,59347,59615,59737,59787,59940,60098,60110,60196,60289,60328,60343,60385,60538,60698,60853,60883,60994,61030,61139,61177,61223,61323,61357,61359,61541,61619,61697,61747,61764,61768,61891,61908,62024,62046,62080,62085,62204,62284,62314,62437,62492,62517,62595,62614,62809,62812,62831,62880,62931,62958,63074,63277,63303,63372,63409,63542,63663,63702,63745,63756,63757,63758,63844,64165,64298,64374,64540,64581,64596,64685,64835,64848,64866,64915,64985,65002,65045,65072,65186,65213,65324,65395,65497,65518,65672,65726,65786,65817,65824,65868,66177,66207,66368,66382,66429,66499,66643,66688,66728,66765,66816,66844,66847,66899,66942,67101,67138,67276,67329,67443,67459,67476,67534,67654,67790,67854,67901,67972,67985,68019,68128,68146,68357,68359,68456,68484,68498,68543,68546,68583,68604,68680,68761,68768,68777,68787,68882,68917,68920,69066,69102,69146,69165,69186,69218,69451,69486,69506,69533,69547,69619,69742,70002,70081,70256,70265,70353,70374,70425,70493,70593,70647,70695,70734,70779,70822,70989,71021,71168,71547,71552,71569,71604,71632,72033,72331,72332,72351,72436,72514,72720,72739,72829,72912,73003,73024,73065,73089,73430,73461,73480,73517,73557,73599,73639,73728,73766,73812,73953,73954,74050,74219,74458,74464,74492,74505,74586,74617,74651,74777,74815,74916,74931,75174,75264,75282,75305,75317,75383,75489,75583,75641,75853,75876,75878,76269,76278,76283,76358,76392,76545,76581,76623,76753,76769,76967,77149,77253,77263,77300,77343,77445,77524,77658,77751,77997,78122,78125,78219,78253,78445,78451,78470,78523,78597,78600,78612,78676,78704,78728,78750,78888,79014,79221,79301,79315,79513,79910,80136,80234,80318,80472,80558,80724,80783,80882,81046,81065,81138,81182,81218,81357,81581,81660,81684,81859,81912,81985,82077,82090,82141,82342,82429,82441,82546,82673,82674,82759,83031,83088,83130,83300,83628,83692,83719,83754,83784,83871,83887,83985,84023,84241,84308,84324,84360,84412,84519,84542,84579,84590,84732,84775,84905,84940,85170,85201,85247,85260,85274,85307,85327,85488,85602,85639,85666,85702,85812,85814,85948,85990,86209,86517,86534,86757,86786,86875,86901,86912,86974,87035,87086,87106,87241,87263,87308,87318,87324,87435,87475,87483,87487,87505,87617,87636,87789,87847,87866,87889,87939,88064,88105,88109,88135,88237,88359,88474,88547,88606,88607,88628,88886,88889,88941,89189,89433,89468,89530,89829,89860,89952,90026,90057,90058,90217,90234,90283,90369,90388,90514,90561,90787,91008,91059,91379,91461,91541,91624,91648,91827,91844,91942,91975,91998,92164,92273,92332,92347,92394,92469,92547,92705,92744,92798,92864,93093,93197,93438,93440,93525,93568,93606,93658,93669,93705,93763,93889,93966,94008,94064,94154,94347,94392,94411,94423,94520,94606,94695,94698,94739,94919,94989,95103,95152,95281,95545,95584,95610,95700,95875,95888,95903,96093,96094,96363,96475,96493,96556,96592,96631,96822,96861,97068,97230,97244,97314,97315,97590,97766,97806,97808,97836,97958,98036,98153,98203,98286,98336,98468,98526,98606,98662,98792,98797,98913,99001,99054,99088,99149,99270,99551,99681,99723,99730,99809,99816,99898,99968,99989",
          "262,592,651,685,769,780,797,821,870,1258,1286,1343,1349,1407,1852,1877,1911,2165,2550,2554,2572,2605,2629,2666,3097,3149,3372,3581,3634,3769,4170,4276,4282,4388,4903,4988,5052,5064,5134,5158,5194,5331,5517,5555,5662,5820,5840,5892,5961,6018,6093,6149,6207,6340,6404,6470,6503,6508,6521,6611,6636,6711,6736,6775,6792,6797,6921,6954,6972,7176,7195,7368,7391,7399,7421,7572,7627,7824,7825,7841,7844,7976,8050,8083,8090,8102,8184,8280,8288,8355,8479,8555,8593,8599,8600,8658,8726,8740,8843,8900,9119,9270,9312,9342,9357,9367,9420,9458,9501,9540,9577,9608,9615,9638,9719,9758,9767,9817,9864,9875,9888,9895,9975,9989,10005,10039,10051,10189,10245,10288,10378,10483,10883,10955,11020,11070,11081,11085,11092,11252,11277,11339,11367,11671,11753,11772,11774,11793,11853,11964,11997,12013,12043,12127,12176,12199,12336,12400,12468,12582,12640,13013,13030,13216,13424,13431,13533,13545,13664,13685,13696,13890,13902,14001,14010,14011,14016,14062,14377,14445,14463,14491,14673,14679,14774,14871,14893,14957,15023,15052,15099,15166,15276,15347,15418,15573,15603,15613,15928,15946,16009,16082,16117,16375,16798,16847,17037,17062,17095,17249,17504,17565,17717,17770,17808,18069,18329,18340,18462,18567,18596,18660,18703,18841,18842,18877,18953,18975,19149,19308,19319,19584,19676,19687,19739,19821,19982,20050,20130,20155,20188,20335,20524,20532,20632,20647,20685,20732,20790,20894,20905,20972,21026,21152,21350,21396,21583,21593,21626,21703,21833,21931,22135,22182,22252,22501,22584,22768,23148,23242,23274,23391,23796,23812,24211,24344,24577,24651,24690,24754,24760,24820,24898,24903,24922,25088,25113,25522,25549,25571,25865,25966,26248,26278,26322,26419,26428,26743,26803,26886,26927,27136,27173,27194,27226,27247,27262,27283,27335,27395,27463,27496,27610,27664,27896,27954,27967,28258,28284,28355,28803,28869,29039,29068,29187,29204,29257,29378,29391,29489,29574,29635,29701,29778,29857,29867,30845,30874,30980,31078,31164,31328,31716,32314,32353,32354,32633,32760,32789,32865,33368,33398,33746,33851,34267,34876,35093,35162,35233,35542,35564,35582,35805,35885,35893,35944,35948,36010,36065,36183,36204,36552,36578,36622,36646,36798,37002,37069,37177,37189,37236,37436,37509,37862,37864,38003,38065,38068,38191,38225,38238,38253,38284,38348,38421,38448,38501,38694,38727,38743,38753,38756,38870,38881,38944,38950,38968,38994,39177,39304,39660,39746,39759,39935,40401,40553,40582,41247,41336,42224,42269,42585,42636,42644,42730,42880,42919,42939,42999,43006,43076,43090,43292,43343,43440,43444,43548,43591,43817,44068,44192,44222,44321,44552,44556,44847,44990,45126,45474,45614,45624,45629,45675,45690,45694,45776,45957,46592,46637,46718,46749,46803,47217,47270,47499,47638,47804,47895,48039,48094,48116,48164,48559,48591,48679,48911,48965,49098,49505,49610,50075,50131,50229,50320,50322,50363,50387,50411,50467,50482,50728,51009,51010,51143,51144,51163,51171,51174,51199,51229,51275,51319,51386,51453,51656,51849,51870,51889,52018,52052,52065,52109,52112,52148,52291,52327,52381,52408,52446,52449,52586,52743,52846,52925,52960,52993,53214,53223,53281,53399,53432,53465,53503,53651,53698,53914,53968,54029,54032,54039,54142,54340,54434,54439,54567,54589,54617,54803,54817,54920,54984,55086,55116,55186,55205,55242,55259,55282,55489,55575,55778,55842,55876,56017,56210,56235,56307,56400,56523,56528,56808,56847,56967,57096,57116,57144,57191,57234,57263,57280,57285,57310,57405,57454,57606,57713,57827,57829,57976,57988,58006,58164,58182,58338,58445,58489,58535,58560,58564,58645,58932,59030,59056,59080,59104,59171,59250,59302,59393,59606,59613,59786,59845,59971,60057,60066,60117,60286,60467,60684,60699,60788,60993,61038,61148,61384,61476,61495,61695,61740,61995,62026,62282,62337,62735,62742,62834,62896,62968,63130,63156,63289,63332,63396,63432,63511,63674,63693,64078,64084,64129,64192,64264,64271,64280,64399,64418,64446,64482,64577,64744,64822,64960,64966,65085,65101,65339,65596,65611,65634,65721,65793,65866,65869,66064,66239,66365,66391,66549,66677,66887,66941,66987,67011,67031,67071,67091,67157,67203,67296,67368,67409,67499,67606,67877,68035,68234,68240,68368,68533,68538,68551,68559,68602,68782,69118,69348,69411,69505,69581,69635,69673,69730,69820,69848,69895,70057,70164,70170,70175,70258,70400,70412,70420,70525,70535,70558,70594,70648,70707,70713,70743,70999,71058,71094,71337,71534,71667,71710,71785,71939,72026,72083,72099,72143,72416,72568,72645,72671,72728,72763,72792,72982,73029,73075,73080,73211,73273,73324,73330,73534,73642,73793,73837,73937,74242,74271,74375,74385,74396,74399,74430,74485,74491,74495,74499,74654,74671,74755,74810,74848,74986,75138,75327,75482,75568,75617,75708,75784,76093,76118,76136,76383,76444,76514,76898,76940,76975,77048,77057,77112,77131,77143,77314,77352,77435,77456,77460,77520,77588,77729,77765,77835,77900,78061,78472,78559,78823,78847,78909,78952,78980,78998,79110,79217,79278,79440,79474,79554,79721,79723,79802,79940,79961,80098,80130,80263,80288,80349,80356,80482,80495,80661,80796,80937,80961,81152,81229,81350,81371,81452,81461,81527,81540,81543,81561,81565,81604,81618,81630,81637,81657,81782,81994,82167,82355,82368,82372,82462,82477,82493,82523,82537,82655,82659,82809,82819,82927,83067,83168,83257,83332,83392,83498,83629,83666,83901,84073,84233,84267,84277,84280,84302,84351,84421,84427,84580,84698,84750,84777,84931,85096,85104,85116,85153,85164,85242,85246,85248,85362,85389,85426,85609,85616,85692,85710,85713,85837,85858,85992,86104,86156,86347,86402,86458,86696,86782,86816,86859,86991,87015,87151,87177,87196,87255,87614,87755,87777,87835,87925,87929,87934,88012,88267,88278,88288,88424,88501,88586,88729,88790,88808,89111,89242,89414,89480,89501,89505,89621,89623,89666,89678,89688,89841,89965,90018,90204,90207,90261,90300,90389,90517,90548,90556,90598,90675,90713,90727,90983,91045,91050,91085,91150,91179,91191,91234,91378,91412,91435,91443,91501,91502,91520,91525,91611,91662,91766,91946,92042,92151,92159,92194,92201,92205,92245,92296,92565,92615,92820,93044,93091,93164,93216,93225,93259,93266,93342,93608,93676,93717,93768,93965,93982,94056,94332,94586,94712,94758,94760,94832,94892,95123,95251,95285,95318,95350,95356,95429,95443,95563,95601,95628,95676,95746,95841,95937,95956,96068,96183,96186,96218,96233,96323,96326,96349,96455,96521,96540,96582,96591,96634,96809,97036,97124,97129,97399,97426,97542,97693,97711,97881,97959,97983,98200,98481,98816,98929,99187,99261,99418,99467,99475,99487,99658,99826,99859,99969,99999",
          "205,264,405,451,561,575,622,661,662,774,790,924,1114,1159,1160,1219,1253,1270,1283,1367,1403,1415,1575,1640,1719,1734,1751,1824,1883,1919,2044,2101,2181,2239,2553,2599,2604,2643,2688,2715,2854,2921,3091,3366,3611,3630,3868,4279,4309,4393,4417,4535,4555,4593,4725,4834,4918,4953,5039,5157,5160,5275,5341,5396,5528,5597,5727,5751,5822,5825,5967,6007,6039,6070,6078,6097,6205,6283,6454,6457,6520,6588,6723,6756,6757,6986,7012,7099,7254,7257,7362,7386,7422,7462,7485,7655,7700,7806,7829,7843,7853,7941,8194,8212,8321,8327,8406,8462,8712,8768,8819,8855,8857,8870,8955,8988,9020,9142,9195,9220,9224,9226,9228,9432,9499,9512,9901,10024,10311,10385,10502,10556,10563,10655,10691,10771,10780,10786,10800,10860,10985,10991,11037,11084,11163,11379,11380,11449,11667,11712,11806,11822,11867,11883,12021,12069,12174,12188,12191,12377,12406,12432,12637,12784,12962,12999,13173,13385,13556,13747,13811,13937,13975,14039,14359,14421,14436,14478,14712,14790,14849,14946,14976,15156,15170,15176,15558,15576,15585,15695,15767,15768,15784,15787,15797,15847,16099,16373,16435,16566,16623,16680,16720,16884,16953,16963,17002,17038,17066,17114,17168,17187,17318,17627,17690,17703,17793,17820,17858,17988,18147,18155,18277,18615,18713,18827,18970,19074,19158,19187,19198,19510,19535,19536,19644,19944,20067,20150,20290,20325,20374,20546,20689,20708,20719,20765,20913,21024,21029,21082,21316,21772,21826,21862,22156,22299,22593,22622,22642,23251,23517,23954,24125,24336,24429,24483,24660,24884,25178,25192,25214,25355,25417,25420,25593,25686,25718,25721,25885,25943,26009,26059,26085,26616,26677,26746,26748,26921,26967,26993,27002,27209,27260,27300,27306,27408,27429,27584,27772,27856,27858,27885,28069,28101,28180,28184,28193,28217,28220,28253,28292,28327,28389,28654,28761,28802,28806,28848,29173,29188,29376,29545,29744,29747,29771,29826,29836,29970,30085,30539,31593,31938,32202,32366,32714,32750,32808,32859,32895,32982,33420,33613,33652,33910,34058,34084,34432,34580,34751,34904,35128,35187,35699,35761,35802,35871,35949,35954,36452,36565,36623,36864,36940,37005,37188,37484,37563,37581,37628,37631,37693,37800,37858,38001,38188,38342,38349,38505,38557,38621,38692,38702,38759,38913,38976,39072,39223,39231,39558,39753,39816,39902,39923,40538,40775,40804,41071,41348,41772,41957,42001,42218,42238,42442,43573,43975,44183,44462,44545,44572,44637,44894,44964,45182,45250,45259,45430,45622,45794,45800,45882,46011,46223,46310,46489,47476,48047,48051,48099,48296,49124,49230,49837,49879,50152,50243,50258,50270,50303,50397,50406,50429,50434,50526,50533,50560,50624,50644,50684,50751,50856,50923,50934,50989,51003,51375,51422,51437,51531,51563,51576,51579,51678,51727,51737,51785,51797,52064,52221,52228,52285,52352,52453,52459,52552,52561,52655,52739,52776,52781,52815,52870,52902,53233,53389,53598,53671,53718,53984,54088,54180,54202,54249,54276,54380,54419,54442,54482,54544,54571,54577,54636,54648,54649,54797,55141,55224,55251,55273,55349,55377,55484,55522,55532,55559,55640,55806,55824,55829,56035,56192,56330,56416,56455,56518,56612,56737,56869,57005,57009,57126,57199,57208,57281,57337,57338,57369,57553,57590,57670,58301,58313,58434,58821,58835,58843,58870,58908,59041,59282,59320,59654,59669,59706,59771,59817,59823,59846,59915,60014,60042,60070,60210,60427,60432,60448,60499,60665,60928,61381,61393,61452,61572,61700,61792,61818,61972,62062,62148,62165,62334,62487,62496,62525,62565,62574,62638,62655,62709,62718,62774,62818,62861,62878,63094,63124,63145,63271,63603,63632,63848,63858,64002,64019,64042,64098,64174,64185,64286,64356,64488,64556,64620,64631,64665,64807,65012,65117,65212,65310,65388,65417,65429,65434,65553,65636,65806,65819,65959,66006,66047,66106,66285,66343,66443,66506,66513,66539,66727,66889,66894,66917,66945,66985,67278,67345,67396,67427,67480,67484,67558,67563,67644,67712,67736,67791,67808,67819,67826,67830,67845,67853,68014,68166,68565,68572,68797,68811,68891,68922,69007,69065,69137,69174,69210,69290,69368,69522,69601,69618,69621,69629,69690,69760,69768,69806,69838,70013,70028,70101,70144,70201,70266,70344,70435,70443,70449,70723,70797,70813,70930,70991,71045,71157,71305,71321,71351,71363,71629,71699,71817,71861,71927,71983,72047,72138,72216,72352,72415,72491,72494,72501,72750,72940,73209,73285,73551,73638,73681,73826,74017,74075,74203,74207,74245,74261,74371,74433,74449,74661,74683,74685,74785,74830,74874,74950,74977,74996,75003,75032,75059,75221,75225,75234,75272,75391,75399,75519,75529,75543,75702,75882,75891,75913,75935,76030,76220,76257,76281,76470,76511,76562,76590,76727,76821,76953,76977,76991,77031,77137,77140,77153,77371,77474,77532,77632,77695,77932,78064,78289,78376,78524,78551,78572,78623,78682,78730,78762,78786,78923,78950,78994,79097,79218,79388,79445,79590,79932,80104,80118,80164,80255,80345,80543,80555,80721,80733,80787,80795,80819,80944,80984,81009,81023,81073,81095,81123,81200,81302,81420,81492,81520,81621,81655,81760,81926,81927,81940,81964,81995,82191,82319,82325,82428,82485,82498,82701,82867,82963,82974,82976,83064,83084,83150,83184,83287,83569,83572,83591,83631,83732,83751,83903,83965,84444,84480,84876,84918,85066,85087,85252,85262,85299,85322,85355,85433,85507,85579,85680,85736,85870,86024,86084,86289,86624,86682,86700,86888,87002,87004,87017,87073,87123,87138,87183,87257,87334,87354,87372,87411,87611,87708,87733,87767,87897,87970,88084,88100,88113,88192,88323,88347,88506,88569,88633,88656,88666,88714,88782,88930,88983,89030,89043,89153,89175,89220,89406,89437,89506,89546,89575,89823,89866,90070,90071,90307,90581,90669,90835,90946,91005,91193,91200,91322,91373,91394,91433,91551,91569,91684,91873,91881,91947,92001,92024,92040,92081,92162,92209,92288,92398,92483,92697,92755,92785,92991,93083,93125,93176,93360,93455,93487,93520,93565,93584,93688,93711,93923,93929,94029,94123,94221,94233,94255,94291,94310,94391,94421,94424,94474,94867,94984,95012,95156,95376,95633,95775,95877,95938,96305,96474,96731,96860,96862,96961,96984,97078,97139,97281,97288,97351,97470,97492,97583,97719,97803,98016,98074,98147,98207,98238,98410,98529,98585,98588,98691,98717,98745,98785,98958,99296,99299,99355,99429,99451,99496,99585,99586,99631,99682,99975,99990,100000",
          "27,118,138,331,625,666,739,743,826,986,1197,1273,1313,1436,1462,1506,1594,2164,2295,2310,2323,2421,2445,2480,2616,2872,3224,3286,3305,3351,3399,3412,3520,3538,3661,3700,3763,4090,4105,4225,4241,4418,4443,4554,4610,4627,4682,4695,4737,4779,4815,5013,5077,5155,5195,5264,5268,5306,5317,5485,5663,5779,5806,5881,5905,5933,5978,6025,6026,6041,6161,6225,6247,6313,6322,6385,6403,6598,6716,6722,6770,6790,6815,6873,6885,6917,6953,7006,7157,7256,7267,7280,7281,7295,7563,7570,7663,7707,7728,7752,7767,7822,7906,7964,8007,8042,8079,8103,8193,8231,8362,8363,8527,8536,8628,8743,8758,8775,8853,8919,9105,9361,9390,9562,9571,9712,9732,9780,9839,9850,9915,9967,10044,10318,10342,10851,10861,10888,10905,10915,10929,10993,11013,11087,11102,11120,11145,11146,11295,11479,11620,11766,11824,11826,12260,12325,12445,12477,12601,12658,12662,12695,12762,12831,12839,12982,12996,13109,13208,13304,13337,13403,13451,13479,13567,13594,13740,13742,13849,13951,13971,14017,14051,14126,14172,14186,14312,14351,14617,14656,14689,14716,14729,14732,14748,14873,14939,15129,15185,15384,15414,15547,15664,15673,15680,15879,16002,16066,16086,16127,16128,16388,16421,16478,16631,16636,16722,16762,16913,16978,16996,17016,17359,17407,17409,17427,17482,17562,17685,17724,17768,17967,18070,18187,18203,18295,18845,18955,19005,19024,19027,19036,19041,19059,19083,19100,19194,19248,19316,19369,19376,19498,19706,19735,19763,19968,20352,20366,20718,20807,20904,21121,21299,21520,21854,21930,21950,22082,22125,22159,22177,22271,22315,22557,22596,22662,22939,22980,23159,23184,23280,23304,23331,23412,23465,23482,23508,23544,23575,23606,23787,23793,23848,23900,24192,24457,24476,24647,24661,24717,24859,24950,25065,25103,25230,25237,25275,25338,25483,25496,25541,25620,25681,25690,25733,25883,25959,25978,26172,26224,26239,26339,26393,26432,26702,26834,27031,27039,27212,27215,27225,27346,27349,27424,27668,27696,27758,27817,27886,27928,28030,28074,28123,28271,28302,28306,28398,28526,28543,28563,28740,28776,28845,28875,28906,28939,28966,28969,28971,29010,29034,29162,29234,29265,29295,29335,29381,29413,29528,29622,29757,29786,29843,29853,29943,29952,29959,29998,30120,30178,30288,30641,30711,30942,31042,31076,31125,31250,31337,31366,31455,31734,31890,32073,32238,32382,32469,32526,32972,32990,33251,33282,33310,33716,33929,33958,34007,34180,34248,34325,34464,34469,34614,34616,34644,34667,34992,35215,35563,35675,35714,35757,35865,36021,36213,36255,36257,36326,36359,36448,36464,36518,36645,36684,36697,36825,36863,36895,36952,37067,37121,37191,37208,37256,37397,37426,37438,37514,37524,37540,37635,37742,37798,37814,37875,37886,38074,38076,38266,38425,38481,38506,38852,39079,39080,39115,39338,39425,39451,39543,39672,39792,39877,39929,40230,40379,40459,40636,40752,40981,41133,41242,41249,42003,42005,42090,42096,42152,42179,42200,42271,42412,42684,42803,42953,43179,43318,43367,43413,43428,43575,43651,43789,44009,44165,44171,44291,44354,44495,44641,44662,44777,44895,44940,45131,45190,45201,45314,45365,45398,45456,45606,45611,45736,45869,45884,45902,46073,46457,46532,46539,46661,46789,46827,46877,47087,47419,47919,48096,48207,48252,48848,48952,48963,49004,49093,49112,49205,49351,49391,49392,49460,49493,49681,49928,50025,50144,50224,50250,50253,50298,50325,50565,50615,50736,51209,51246,51338,51355,51393,51433,51467,51469,51470,51541,51603,51769,51809,51952,52105,52120,52128,52222,52288,52400,52460,52476,52492,52579,52607,52620,52762,52922,53070,53073,53104,53226,53271,53348,53410,53459,53506,53577,53583,53625,53639,53653,53720,53837,53872,53882,53926,53957,54021,54159,54172,54186,54206,54252,54260,54418,54462,54467,54484,54534,54635,54730,54772,54837,54844,54884,54891,55185,55296,55563,55740,55798,55832,55871,55892,55955,56008,56023,56051,56108,56177,56185,56219,56242,56505,56613,56708,56821,56975,57182,57295,57327,57383,57400,57465,57490,57496,57538,57573,57708,57749,57770,57802,57841,57907,58023,58028,58081,58195,58225,58254,58268,58323,58511,58512,58518,58547,58916,58931,58958,59084,59329,59365,59417,59510,59589,59618,59710,59729,59735,59806,59901,60044,60093,60170,60173,60174,60223,60248,60274,60333,60335,60445,60477,60540,60639,60963,60971,61078,61111,61112,61128,61191,61207,61251,61353,61439,61556,61565,61723,61898,61933,62027,62031,62047,62213,62239,62369,62478,62623,62662,62833,62954,63053,63082,63119,63140,63177,63188,63279,63387,63558,63578,63660,63794,63874,63933,63966,63986,64001,64065,64186,64189,64208,64355,64381,64503,64694,64919,65023,65142,65192,65303,65453,65589,65622,65854,65863,65922,65932,65995,66070,66144,66199,66383,66396,66428,66525,66636,66777,66992,67000,67003,67111,67162,67286,67362,67492,67533,67669,67690,67695,67783,67794,67948,68182,68299,68319,68334,68445,68493,68497,68570,68708,68764,68771,68774,68793,68826,68877,69010,69025,69062,69244,69328,69932,69933,70032,70179,70302,70340,70348,70388,70395,70409,70448,70672,70789,70908,70920,70941,71005,71063,71069,71289,71318,71400,71498,71511,71626,71670,71738,71865,72198,72263,72264,72304,72342,72383,72419,72564,72603,72686,72704,72865,72957,73087,73092,73144,73320,73336,73339,73356,73380,73420,73439,73518,73564,73768,73771,73820,73859,73863,73879,73967,73985,74009,74266,74408,74409,74413,74486,74497,74595,74679,74882,75103,75162,75209,75288,75414,75597,75647,75698,76081,76090,76137,76235,76341,76435,76440,76497,76532,76557,76571,76596,76755,76848,76976,77026,77094,77141,77215,77434,77436,77449,77479,77534,77539,77582,77764,77767,77851,77911,77950,77952,78080,78232,78404,78409,78431,78460,78496,78656,78732,78902,78905,78930,78992,79106,79139,79181,79200,79201,79331,79365,79432,79821,79843,79852,79953,79987,80053,80054,80105,80134,80168,80295,80432,80564,80603,80627,80659,80801,80825,80884,80966,81045,81156,81159,81328,81394,81418,81450,81479,81659,81707,81796,81837,81879,81888,81903,81906,81967,81992,82029,82095,82255,82365,82720,82773,82814,82912,83061,83140,83248,83382,83714,83731,83853,83874,84014,84015,84055,84167,84364,84461,84495,84522,84638,84687,84837,84843,84871,84910,84984,85044,85134,85143,85526,85573,85582,85589,85596,85607,85613,85771,85877,85905,85964,85994,86015,86023,86071,86073,86167,86223,86302,86340,86362,86489,86810,86856,86867,86938,87069,87072,87236,87276,87321,87525,87585,87604,87769,87808,87965,87971,87988,88200,88253,88265,88382,88554,88598,88664,88715,88862,88871,88915,89196,89303,89348,89389,89498,89510,89552,89554,89581,89783,89843,89961,90189,90227,90399,90513,90580,90611,90628,90654,90741,90754,90786,90953,91074,91142,91153,91233,91507,91818,92073,92196,92206,92336,92393,92415,92466,92581,92664,92671,92684,92686,92797,92815,92886,93107,93116,93161,93179,93269,93279,93384,93446,93449,93513,93628,93713,93761,93772,93898,93958,94007,94246,94311,94535,94538,94643,94753,94781,94871,95011,95073,95181,95245,95450,95822,95829,95866,95933,95948,95957,95965,95978,96054,96092,96144,96177,96197,96553,96641,96707,96734,96769,96938,96993,97025,97050,97072,97138,97141,97332,97438,97443,97531,97564,97568,97969,97971,98136,98194,98282,98354,98515,98575,98741,98747,98810,98881,98887,98935,98980,99018,99033,99148,99285,99430,99491,99533,99564,99684,99836,99948",
          "45,50,64,106,115,159,169,221,1793,1922,1936,2079,2127,2251,2255,2270,2335,2341,2396,2477,2590,2662,2680,2706,2711,2767,2818,2832,2836,2882,2909,2956,3152,3206,3446,3522,3714,4098,4099,4191,4204,4390,4458,4601,4611,4747,4853,4866,4948,5544,6980,7949,7973,9404,9412,9476,9517,9545,9546,9621,9640,9678,9697,9716,9721,9808,9824,9944,10020,10077,10086,10090,10096,10210,10381,10590,10630,10647,10666,10689,10745,10837,10856,10951,11038,11046,11179,11197,11198,11203,11217,11236,11255,11293,11369,11445,11477,11540,11597,11654,11730,11882,11920,12015,12036,12072,12129,12186,12224,12243,12262,12338,12395,12414,12419,12471,12528,12547,12654,12661,12699,12710,12737,12758,12775,12813,12823,12889,12892,12916,12930,12965,12973,12984,13003,13022,13041,13060,13112,13117,13204,13223,13231,13250,13326,13363,13370,13381,13398,13402,13404,13421,13459,13516,13573,13599,13601,13606,13630,13649,13654,13687,13703,13706,13725,13736,13744,13763,13782,13820,13858,13885,13906,13934,13953,13978,14029,14048,14100,14105,14124,14162,14181,14200,14221,14238,14245,14276,14295,14308,14314,14333,14352,14435,14444,14448,14509,14519,14525,14534,14544,14580,14618,14620,14637,14675,14694,14747,14770,14827,14839,14846,14884,14909,14922,14935,14972,14979,15017,15102,15134,15147,15149,15169,15188,15207,15208,15264,15283,15302,15318,15321,15397,15416,15432,15479,15480,15509,15511,15530,15625,15644,15653,15666,15682,15701,15720,15740,15758,15760,15791,15815,15834,15853,15872,15891,15948,15967,15986,16043,16081,16124,16138,16157,16214,16225,16233,16252,16253,16268,16290,16309,16333,16347,16366,16404,16461,16499,16518,16575,16581,16613,16640,16651,16662,16670,16704,16746,16765,16770,16775,16784,16822,16947,16974,16993,17048,17088,17126,17164,17183,17221,17240,17275,17297,17312,17373,17392,17430,17449,17465,17506,17525,17561,17601,17608,17620,17715,17767,17772,17791,17810,17848,17886,17898,17924,17976,18038,18054,18057,18114,18171,18190,18209,18266,18285,18293,18311,18335,18342,18350,18361,18380,18437,18456,18469,18494,18525,18532,18646,18665,18684,18724,18817,18931,18943,18950,18988,19022,19032,19045,19064,19088,19121,19140,19178,19197,19233,19254,19259,19303,19311,19330,19364,19366,19387,19406,19444,19475,19481,19482,19501,19520,19539,19615,19663,19672,19691,19707,19729,19767,19786,19838,19862,19881,19888,19900,19938,19957,19995,20003,20014,20052,20128,20147,20223,20226,20299,20318,20333,20337,20356,20375,20388,20390,20394,20432,20439,20470,20481,20489,20498,20508,20587,20603,20641,20648,20679,20736,20755,20772,20793,20812,20850,20869,20909,20983,21002,21025,21045,21078,21090,21097,21116,21154,21173,21198,21211,21249,21268,21287,21325,21363,21382,21420,21439,21472,21477,21519,21534,21553,21566,21610,21629,21724,21743,21773,21838,21876,21893,21895,21914,21933,21952,21971,22009,22047,22054,22060,22066,22102,22104,22123,22142,22153,22199,22218,22274,22275,22294,22302,22332,22351,22364,22370,22389,22408,22410,22427,22446,22458,22465,22541,22560,22579,22619,22636,22639,22655,22693,22712,22731,22750,22753,22788,22826,22890,22898,22902,22997,23029,23073,23079,23092,23130,23146,23149,23187,23200,23206,23219,23225,23244,23282,23301,23358,23377,23386,23392,23396,23415,23434,23491,23510,23529,23567,23624,23643,23667,23681,23700,23719,23738,23750,23753,23757,23814,23845,23852,23871,23890,23902,23928,23947,23966,24023,24042,24061,24080,24118,24156,24175,24213,24222,24232,24270,24308,24346,24365,24384,24403,24410,24441,24460,24479,24494,24498,24517,24534,24536,24555,24574,24631,24669,24688,24707,24726,24764,24799,24840,24878,24897,24916,24927,24983,24992,25006,25011,25075,25125,25144,25189,25201,25220,25296,25315,25356,25360,25384,25391,25410,25448,25502,25543,25555,25562,25581,25619,25633,25657,25676,25695,25752,25771,25790,25793,25809,25847,25859,25866,25874,25942,25961,25980,25999,26020,26036,26037,26056,26075,26094,26113,26120,26170,26183,26246,26265,26303,26341,26379,26417,26436,26474,26493,26512,26623,26626,26645,26664,26740,26755,26892,26911,26949,26986,27042,27063,27246,27265,27272,27291,27310,27389,27436,27465,27671,27766,27804,27874,27877,28006,28070,28089,28127,28182,28196,28218,28241,28326,28338,28380,28447,28451,28462,28504,28507,28583,28611,28678,28694,28881,28962,29077,29225,29271,29323,29350,29433,29475,29558,29667,29670,29780,29851,29892,29894,29913,30007,30036,30065,30094,30123,30152,30210,30239,30268,30355,30384,30413,30442,30471,30500,30521,30529,30558,30587,30590,30616,30645,30674,30703,30725,30732,30761,30782,30790,30819,30848,30877,30906,30935,30964,30993,31022,31051,31080,31109,31138,31160,31167,31196,31225,31242,31254,31283,31312,31331,31341,31399,31428,31457,31486,31541,31544,31573,31602,31631,31660,31683,31689,31718,31747,31776,31805,31834,31863,31892,31921,31950,31966,31979,32008,32037,32058,32066,32095,32124,32149,32153,32182,32211,32217,32240,32269,32298,32327,32356,32385,32414,32443,32501,32530,32559,32588,32617,32646,32675,32704,32733,32762,32791,32820,32849,32878,32907,32936,32959,32965,32994,33023,33052,33081,33110,33139,33168,33185,33197,33226,33255,33284,33342,33371,33400,33429,33458,33487,33516,33545,33574,33603,33632,33661,33690,33719,33748,33777,33806,33835,33864,33881,33893,33922,33951,33980,34025,34038,34067,34096,34125,34154,34183,34200,34212,34270,34286,34299,34328,34357,34374,34386,34415,34444,34446,34465,34473,34502,34531,34560,34589,34618,34676,34705,34722,34734,34763,34792,34821,34850,34879,34908,34937,34966,34995,35050,35096,35140,35153,35170,35201,35228,35242,35259,35326,35333,35373,35397,35407,35443,35478,35512,35526,35585,35619,35624,35628,35644,35717,35738,35750,35776,35835,35906,35931,35965,36005,36024,36049,36050,36087,36147,36206,36277,36336,36358,36361,36395,36418,36427,36467,36526,36555,36585,36612,36649,36708,36731,36767,36838,36878,36897,36956,37028,37087,37146,37210,37257,37269,37328,37356,37359,37399,37458,37517,37541,37552,37589,37609,37648,37707,37771,37830,37889,37916,37919,37960,38019,38078,38112,38150,38209,38268,38292,38332,38391,38450,38521,38580,38626,38639,38770,38829,38893,38929,38934,38952,39011,39082,39141,39171,39200,39272,39331,39354,39390,39454,39513,39572,39643,39702,39761,39788,39833,39892,39951,40000,40001,40002,40003,40004,40005,40006,40007,40008,40009,40010,40011,40012,40013,40014,40015,40016,40017,40018,40019,40020,40021,40022,40023,40024,40025,40026,40027,40028,40029,40030,40031,40032,40033,40034,40035,40036,40037,40038,40039,40040,40042,40043,40044,40045,40046,40047,40048,40049,40050,40051,40052,40053,40054,40055,40056,40057,40058,40059,40060,40061,40062,40063,40064,40065,40066,40067,40068,40069,40070,40071,40072,40074,40076,40077,40078,40079,40081,40082,40083,40084,40085,40086,40087,40089,40090,40091,40093,40094,40095,40096,40097,40098,40099,40100,40120,40149,40178,40207,40236,40265,40294,40323,40352,40381,40410,40439,40468,40497,40526,40555,40584,40613,40642,40671,40700,40729,40758,40787,40816,40845,40874,40903,40932,40961,40990,41019,41048,41077,41106,41135,41164,41193,41251,41280,41309,41338,41367,41396,41425,41454,41483,41512,41541,42011,42040,42069,42098,42127,42146,42149,42187,42247,42277,42352,42403,42409,42427,42482,42514,42561,42563,42567,42589,42628,42664,42697,42751,42765,42796,42808,42826,42847,42901,42942,42947,42988,43042,43063,43138,43149,43153,43181,43187,43225,43236,43257,43259,43296,43300,43321,43375,43411,43421,43462,43490,43515,43537,43581,43612,43699,43712,43752,43757,43774,43806,43807,43849,43872,43896,43936,43969,44011,44042,44051,44086,44173,44219,44248,44268,44281,44293,44297,44323,44342,44377,44410,44440,44442,44485,44533,44560,44574,44629,44647,44698,44708,44722,44754,44775,44797,44817,44857,44884,44906,44918,44920,44959,45014,45034,45107,45155,45175,45216,45274,45295,45347,45357,45367,45404,45413,45462,45534,45562,45591,45602,45649,45699,45721,45747,45778,45801,45836,45908,45933,45965,46043,46068,46091,46102,46153,46167,46171,46186,46212,46218,46225,46235,46241,46296,46359,46416,46422,46427,46507,46610,46611,46614,46623,46645,46656,46677,46678,46706,46717,46731,46794,46798,46817,46847,46866,46917,47036,47041,47069,47084,47089,47119,47161,47195,47257,47260,47277,47306,47309,47330,47345,47444,47475,47477,47487,47494,47530,47576,47611,47616,47618,47626,47628,47652,47668,47679,47750,47755,47758,47759,47762,47813,47822,47855,47857,47860,47863,47962,47983,47992,48005,48008,48091,48107,48160,48187,48264,48318,48406,48430,48473,48490,48491,48495,48568,48602,48609,48615,48680,48786,48791,48793,48797,48823,48833,48841,48886,48891,48907,48910,48913,48925,48929,48953,48984,48999,49051,49059,49079,49117,49142,49175,49191,49192,49233,49239,49289,49291,49298,49386,49405,49451,49475,49520,49522,49534,49576,49661,49713,49731,49757,49759,49760,49789,49869,49916,49918,49942,49962,49967,49988,50013,50174,50194,50318,50336,50388,50468,50489,50502,50528,50611,50650,50723,50763,50791,50922,50941,51007,51125,51258,51281,51418,51452,51507,51572,51582,51635,51748,51947,51957,51967,51994,52002,52042,52172,52236,52376,52402,52421,52588,52702,52737,52758,52790,52878,52885,52904,53039,53155,53228,53452,53460,53461,53464,53514,53560,53672,53686,53713,53847,53973,54001,54057,54135,54136,54147,54217,54271,54284,54450,54453,54463,54562,54739,54765,54860,55283,55427,55438,55469,55536,55540,55543,55595,55791,55930,55961,55972,56083,56094,56154,56499,56586,56723,56843,56873,57007,57058,57082,57212,57227,57256,57296,57326,57418,57567,57650,57666,57720,57724,57838,57850,57861,57880,57997,58257,58353,58501,58743,58783,58816,58861,58926,59116,59210,59243,59509,59537,59602,59633,59751,59799,59801,59834,59889,60073,60094,60166,60189,60278,60296,60347,60409,60520,60532,60681,60711,60822,61056,61124,61239,61298,61426,61475,61775,61910,61976,61982,62292,62341,62384,62504,62515,62603,62605,62696,62717,62767,62841,63195,63339,63391,63715,63815,63835,63871,63912,64088,64121,64218,64263,64354,64397,64414,64521,64536,64579,64630,64783,64841,64885,64893,65040,65109,65205,65248,65280,65385,65415,65474,65643,65809,65830,65882,65929,66101,66109,66214,66271,66277,66295,66333,66426,66455,66498,66665,66696,66715,66775,66790,66806,66836,66896,66918,67253,67385,67509,67511,67624,67776,67865,67958,68610,68664,68731,68738,68786,68863,68921,69049,69079,69189,69194,69202,69225,69231,69322,69469,69563,69589,69717,69781,69823,69940,69965,70084,70088,70092,70234,70326,70361,70446,70477,70622,70625,70627,70644,71119,71383,71441,71477,71539,71546,71657,71739,71769,72164,72336,72441,72483,72504,72505,72524,72650,72665,72861,73112,73158,73174,73241,73402,73483,73512,73765,73767,73783,73790,73807,73910,74136,74305,74343,74349,74459,74473,74482,74564,74659,74682,74770,74935,74946,75205,75456,75665,75679,75700,75734,75826,75934,75971,76308,76403,76522,76567,76639,76719,76721,76767,76893,76981,76996,77090,77111,77246,77281,77283,77363,77567,77625,77961,77979,78250,78312,78369,78388,78487,78505,78512,78542,78586,78759,78857,78887,78927,79045,79134,79268,79297,79357,79431,79457,79483,79484,79535,79624,79671,79698,79703,79839,79882,79923,79968,80139,80238,80252,80340,80390,80416,80535,80595,80601,80652,80730,80742,80743,80808,80895,80899,80903,80952,80986,81012,81144,81234,81267,81347,81356,81456,81506,81544,81583,81641,81728,81797,81816,81851,81925,82031,82102,82138,82186,82227,82390,82407,82414,82538,82553,82558,82668,82671,82770,82869,83111,83114,83219,83222,83344,83351,83433,83491,83527,83547,83553,83566,83592,83643,83644,83662,83794,83944,84081,84142,84289,84312,84340,84411,84528,84839,84964,84972,84989,85018,85024,85041,85190,85217,85345,85347,85391,85412,85487,85489,85498,85549,85554,85638,85677,85790,85873,85888,85890,86012,86153,86188,86220,86380,86414,86558,86665,86879,86911,86949,86951,87074,87136,87209,87358,87490,87561,87715,87717,87744,87770,87844,87895,88071,88087,88121,88169,88305,88371,88389,88603,88678,88710,88713,88743,88766,89070,89107,89119,89124,89298,89305,89378,89427,89435,89438,89441,89594,89665,89672,89925,90013,90048,90092,90119,90144,90199,90212,90319,90327,90461,90562,90660,90696,90706,91003,91033,91203,91301,91307,91328,91335,91360,91476,91483,91709,91752,91840,92013,92025,92067,92100,92256,92319,92606,92626,92644,92668,92722,92724,92736,92770,92796,93252,93264,93411,93420,93480,93511,93527,93558,93723,93731,93734,93791,93809,93826,93828,93848,93868,93874,93876,93933,94335,94384,94385,94575,94682,94722,94775,94906,94972,95158,95160,95243,95428,95492,95549,95731,95756,95845,95859,95893,95943,95985,96022,96055,96232,96251,96278,96309,96328,96466,96571,96629,96638,96836,96902,96949,97013,97067,97104,97170,97183,97203,97206,97218,97347,97517,97524,97757,97811,97843,97876,98009,98105,98152,98155,98404,98676,98689,98748,98751,98767,98891,99050,99182,99416,99506,99604,99648,99677,99742,99750,99789,99802,99803,99885,99909,100001",
          "46,65,103,105,122,160,182,186,191,2054,2143,2272,2303,2350,2355,2463,2498,2592,2626,2636,3075,3137,3794,3820,3861,3875,3922,4403,4406,4504,4528,4703,4764,4804,6140,6621,7688,8387,9280,9394,9451,9555,9565,9566,9584,9641,9660,9698,9755,9793,9812,9831,9926,10021,10037,10078,10097,10131,10192,10363,10648,10819,10914,10933,11009,11066,11104,11161,11180,11218,11237,11256,11313,11408,11468,11598,11617,11636,11750,11930,11978,12092,12095,12121,12130,12149,12187,12225,12244,12282,12301,12317,12358,12360,12396,12415,12469,12510,12529,12624,12738,12776,12814,12833,12928,12947,12957,12985,13061,13099,13118,13194,13213,13251,13270,13289,13310,13327,13384,13441,13555,13593,13612,13650,13688,13707,13726,13764,13783,13802,13821,13859,13916,13992,14030,14087,14139,14143,14144,14163,14182,14220,14239,14258,14277,14296,14315,14334,14410,14429,14484,14524,14543,14638,14657,14752,14809,14810,14847,14863,14866,14886,14923,14951,14961,15018,15075,15094,15104,15151,15246,15265,15284,15322,15341,15360,15455,15493,15512,15588,15626,15645,15690,15702,15721,15835,15854,15873,15892,15947,15949,16006,16015,16025,16044,16078,16101,16120,16136,16158,16177,16207,16215,16234,16272,16291,16310,16329,16348,16367,16405,16424,16443,16481,16484,16500,16520,16538,16549,16587,16614,16633,16635,16652,16671,16728,16747,16785,16823,16842,16851,16861,16899,16909,16956,16975,16986,16994,17070,17108,17146,17203,17222,17298,17317,17393,17433,17450,17469,17507,17545,17564,17640,17678,17695,17697,17716,17735,17754,17773,17792,17811,17868,17887,17925,17944,18001,18058,18077,18093,18115,18153,18163,18172,18248,18267,18286,18305,18324,18343,18362,18381,18400,18419,18438,18533,18552,18571,18586,18609,18647,18666,18738,18761,18780,18818,18894,18932,18951,18989,19023,19057,19061,19065,19103,19112,19123,19160,19179,19236,19253,19274,19331,19333,19388,19407,19426,19445,19464,19502,19578,19599,19673,19692,19703,19711,19730,19749,19768,19806,19825,19863,19882,19920,19939,19958,19996,20007,20015,20043,20064,20091,20129,20148,20186,20205,20281,20300,20301,20338,20357,20376,20414,20452,20471,20509,20528,20566,20585,20604,20642,20661,20678,20699,20737,20756,20811,20813,20870,20889,20908,20946,20962,20965,21003,21019,21089,21117,21131,21136,21155,21212,21226,21231,21269,21288,21307,21309,21345,21364,21440,21478,21497,21535,21554,21573,21592,21611,21668,21702,21725,21744,21762,21763,21766,21820,21858,21874,21877,21896,21915,21934,21991,22048,22064,22105,22124,22162,22200,22215,22219,22238,22276,22333,22352,22371,22390,22428,22447,22466,22485,22504,22523,22542,22572,22618,22637,22652,22656,22694,22713,22789,22808,22827,22846,22884,22903,22960,22998,23017,23036,23093,23150,23169,23302,23321,23359,23378,23435,23473,23492,23549,23587,23625,23644,23682,23701,23758,23777,23815,23834,23929,23967,23986,24028,24043,24062,24081,24100,24157,24195,24233,24252,24290,24309,24347,24348,24385,24404,24423,24442,24461,24480,24499,24520,24537,24556,24632,24708,24727,24746,24762,24765,24784,24831,24879,24917,24974,24993,25012,25050,25069,25137,25164,25183,25202,25278,25280,25373,25392,25430,25449,25509,25525,25563,25582,25639,25658,25677,25734,25753,25792,25810,25827,25867,25886,25905,25923,25924,25962,25981,26000,26019,26038,26057,26076,26095,26133,26163,26171,26228,26247,26302,26304,26342,26361,26399,26437,26456,26475,26478,26494,26513,26520,26532,26589,26608,26627,26654,26722,26741,26759,26782,26821,26855,26912,26937,26982,27169,27178,27308,27317,27330,27348,27400,27523,27539,27577,27608,27615,27630,27672,27711,27821,27862,27878,27938,27976,28166,28204,28251,28403,28434,28527,28660,28667,28698,28789,28907,28926,28945,28976,28994,29040,29053,29094,29192,29219,29230,29287,29363,29401,29414,29477,29502,29515,29599,29629,29666,29773,29781,29869,29882,30008,30037,30066,30095,30124,30153,30182,30211,30240,30269,30298,30327,30356,30385,30414,30443,30472,30501,30530,30559,30588,30617,30646,30675,30704,30733,30762,30791,30820,30849,30878,30907,30936,30965,30994,31023,31052,31139,31168,31197,31226,31255,31284,31313,31342,31371,31400,31429,31458,31487,31516,31545,31574,31603,31632,31661,31719,31748,31777,31806,31835,31864,31893,31922,31951,31980,32009,32038,32067,32096,32125,32154,32183,32212,32241,32270,32299,32328,32357,32386,32415,32444,32473,32502,32531,32560,32589,32647,32676,32705,32734,32763,32792,32821,32850,32879,32908,32937,32966,32995,33024,33053,33082,33111,33140,33169,33198,33227,33256,33285,33314,33343,33372,33401,33430,33459,33488,33517,33546,33575,33604,33633,33662,33691,33720,33749,33778,33807,33836,33865,33894,33923,33952,33981,34010,34039,34068,34097,34126,34155,34184,34213,34242,34271,34300,34358,34387,34416,34445,34474,34503,34532,34561,34590,34619,34648,34677,34706,34735,34764,34793,34822,34851,34880,34909,34938,34967,34996,35008,35061,35064,35069,35097,35171,35196,35260,35317,35334,35390,35408,35439,35471,35479,35527,35586,35645,35673,35718,35777,35836,35907,35932,35966,36025,36061,36088,36124,36130,36148,36207,36234,36278,36317,36337,36396,36468,36527,36586,36650,36685,36689,36709,36768,36793,36839,36898,36957,37029,37088,37115,37147,37211,37270,37302,37329,37400,37459,37518,37590,37649,37708,37772,37810,37831,37890,37961,37996,38020,38061,38071,38079,38114,38151,38210,38269,38333,38363,38392,38439,38451,38487,38522,38581,38640,38665,38668,38712,38771,38800,38830,38894,38953,39012,39036,39039,39083,39142,39201,39273,39332,39391,39455,39514,39550,39573,39644,39684,39703,39762,39799,39834,39952,40107,40136,40165,40194,40223,40252,40281,40310,40339,40368,40397,40426,40455,40484,40513,40542,40571,40600,40629,40658,40687,40716,40745,40774,40803,40832,40861,40919,40948,40977,41006,41035,41064,41093,41122,41151,41180,41209,41238,41267,41296,41325,41354,41383,41412,41441,41470,41499,41528,42000,42029,42058,42087,42116,42151,42176,42257,42266,42322,42341,42391,42416,42503,42552,42578,42606,42627,42653,42716,42740,42815,42890,42927,42977,43052,43082,43101,43112,43127,43173,43214,43289,43326,43364,43417,43451,43506,43526,43557,43601,43688,43733,43737,43763,43796,43798,43838,43918,43925,43989,44000,44046,44075,44162,44237,44261,44312,44349,44399,44454,44474,44549,44598,44614,44636,44711,44786,44824,44873,44941,44948,44974,44979,45023,45084,45096,45144,45171,45205,45263,45286,45336,45393,45434,45451,45490,45523,45569,45638,45669,45710,45767,45825,45897,45954,45977,45987,46095,46113,46120,46174,46301,46304,46306,46357,46412,46417,46490,46524,46567,46597,46604,46631,46676,46700,46734,46784,46906,46948,46965,46992,47022,47029,47055,47078,47149,47169,47173,47177,47190,47274,47293,47321,47391,47401,47423,47443,47458,47469,47495,47510,47690,47742,47751,47754,47764,47802,47841,47846,47883,47908,47914,47915,47921,47933,48063,48072,48088,48097,48119,48131,48140,48168,48169,48201,48208,48217,48272,48279,48300,48312,48370,48371,48413,48414,48454,48486,48510,48524,48541,48557,48583,48597,48614,48646,48654,48685,48720,48730,48748,48804,48826,48854,48894,48900,48917,49018,49092,49137,49220,49244,49280,49282,49315,49357,49416,49424,49425,49444,49472,49474,49484,49488,49513,49619,49636,49637,49645,49667,49683,49743,49758,49780,49800,49871,49968,49990,50034,50087,50090,50129,50220,50428,50441,50503,50539,50569,50593,50606,50719,50961,51014,51082,51347,51382,51488,51679,51723,51839,51903,51927,51937,51941,52006,52024,52122,52146,52160,52185,52386,52467,52520,52534,52537,52591,52640,53087,53102,53172,53251,53264,53371,53763,53777,53784,53790,53827,53892,53938,53948,54115,54227,54258,54491,54541,54581,54582,54705,54841,55001,55008,55088,55130,55163,55215,55519,55533,55692,55775,55777,55786,55826,55970,55995,56015,56104,56187,56212,56244,56350,56351,56403,56448,56485,56581,56654,56757,56823,56959,56984,57097,57109,57121,57408,57442,57596,57603,57716,57741,57796,57806,57885,57968,58005,58232,58424,58430,58492,58526,58586,58798,58808,58826,59085,59191,59274,59287,59289,59299,59307,59343,59358,59519,59584,59610,59697,59746,60184,60229,60542,60559,60625,60650,60776,61070,61087,61088,61201,61289,61518,61550,61602,61641,61730,61763,61950,61973,62117,62176,62209,62313,62375,62505,62518,62519,62567,62572,62645,62703,62766,62876,62904,62982,63041,63193,63207,63220,63329,63405,63521,63567,63619,63654,63686,63781,63936,64016,64125,64320,64376,64465,64557,64559,64615,64673,64740,64753,65200,65284,65295,65383,65472,65509,65644,65677,65744,65934,66002,66088,66137,66143,66256,66321,66402,66505,66612,66779,66927,66939,67089,67182,67319,67358,67386,67430,67557,67574,67597,67626,67631,67678,67703,67746,67815,67870,68139,68167,68262,68386,68431,68438,68439,68547,68775,68785,68833,68867,68869,69253,69270,69324,69331,69631,69679,69888,69903,69982,70009,70100,70117,70142,70165,70168,70260,70277,70511,70585,70702,70878,70963,71175,71247,71472,71722,71749,71791,71793,71803,71843,71890,71946,71951,71973,72018,72065,72076,72129,72174,72334,72345,72479,72516,72559,72588,72625,72680,72685,72751,72830,72935,73169,73271,73432,73465,73472,73699,73759,73761,73813,73927,73978,73991,74023,74142,74164,74331,74374,74436,74500,74575,74684,74752,74857,75021,75033,75154,75235,75421,75431,75483,75525,75636,75680,75730,75771,75808,75824,75856,75933,76007,76171,76363,76397,76525,76569,76587,76628,76680,76726,76814,76884,77207,77210,77267,77404,77557,77635,77708,77790,77846,77856,77970,77980,78007,78249,78337,78356,78367,78396,78855,78917,79154,79194,79238,79290,79304,79391,79413,79425,79478,79494,79551,79586,79683,79758,79836,79944,79975,80124,80132,80162,80251,80358,80417,80425,80629,80697,80761,80933,81068,81282,81403,81430,81451,81481,81510,81534,81541,81600,81650,81653,81758,81831,81910,82158,82265,82301,82559,82603,82609,82706,82787,82855,82907,82985,83058,83094,83158,83475,83576,83588,83619,83806,84030,84149,84259,84319,84513,84532,84635,84720,84735,84851,85105,85144,85230,85290,85324,85336,85348,85450,85466,85467,85474,85543,85551,85622,85684,85932,85944,85962,85972,86018,86102,86146,86197,86284,86378,86395,86429,86491,86524,86758,86804,86830,86882,86997,87050,87077,87534,87535,87553,87773,87975,88088,88143,88402,88562,88868,88923,88939,89140,89350,89364,89551,89636,89681,89731,89743,89915,89976,90030,90252,90288,90511,90530,90572,90607,90822,90915,90952,91048,91094,91113,91318,91342,91397,91398,91401,91460,91488,91529,91565,91643,91718,91810,91958,91991,92075,92079,92176,92177,92250,92302,92327,92371,92412,92595,92989,93085,93132,93188,93241,93280,93299,93357,93386,93401,93562,93577,93594,93639,93672,93685,93722,93744,93801,93836,93879,93917,93968,93978,94012,94061,94093,94210,94236,94279,94340,94486,94613,94676,94692,94757,94808,94857,94952,95048,95086,95099,95208,95213,95233,95283,95297,95726,95750,95816,95826,95914,96189,96238,96292,96333,96401,96603,96644,96766,96828,96893,96947,97029,97236,97284,97385,97541,97565,97742,97798,97827,97835,98060,98139,98192,98261,98273,98315,98381,98389,98453,98483,98518,98670,98758,98833,98917,98926,99058,99077,99081,99234,99297,99716,99841,99865,99895,99915,99918,99937,99939",
          "71,129,185,207,224,1735,1778,1949,1963,1982,2039,2151,2226,2244,2275,2314,2339,2617,2811,2913,2989,3016,3051,3139,3140,3177,3202,3247,3325,3339,3352,3407,3662,3790,3867,3972,4008,4034,4139,4243,4334,4612,4630,4647,4919,5120,5526,6955,9474,9563,9601,9620,9639,9658,9674,9675,9677,9696,9772,9810,9829,9924,9962,10095,10114,10722,10741,10760,10912,10950,11254,11292,11406,11442,11482,11539,11558,11615,11657,11881,11900,11957,11980,11995,12014,12033,12090,12147,12166,12204,12223,12261,12280,12318,12365,12413,12451,12588,12603,12717,12755,12888,12945,13021,13059,13078,13098,13102,13138,13230,13249,13263,13325,13344,13415,13420,13477,13496,13515,13536,13574,13591,13600,13667,13686,13705,13743,13762,13781,13819,13857,13887,13933,14085,14104,14180,14199,14218,14237,14256,14275,14294,14313,14389,14408,14427,14522,14541,14655,14659,14667,14674,14677,14693,14788,14860,14902,14921,14959,14990,14997,15016,15054,15073,15111,15130,15206,15215,15244,15263,15272,15282,15287,15320,15339,15358,15415,15434,15510,15515,15529,15539,15561,15567,15586,15643,15662,15700,15738,15776,15795,15833,15852,15871,15939,16004,16023,16061,16064,16080,16118,16175,16194,16199,16213,16232,16251,16289,16327,16346,16365,16403,16422,16455,16460,16479,16489,16498,16555,16574,16593,16612,16650,16707,16726,16745,16802,16821,16850,16859,16869,16878,16935,16954,17011,17030,17068,17087,17106,17119,17163,17182,17201,17258,17277,17279,17296,17315,17353,17410,17448,17467,17517,17524,17543,17547,17563,17600,17619,17623,17657,17676,17714,17771,17781,17790,17809,17828,17847,17885,17904,17923,17943,17961,17994,17999,18037,18090,18132,18189,18208,18227,18257,18259,18265,18284,18294,18322,18341,18360,18379,18398,18417,18436,18455,18512,18531,18550,18570,18584,18626,18664,18702,18721,18733,18759,18766,18778,18816,18835,18892,18911,18917,18930,18968,19006,19025,19082,19120,19139,19164,19196,19215,19234,19256,19268,19272,19276,19348,19349,19367,19372,19386,19424,19432,19443,19462,19476,19519,19538,19557,19587,19614,19652,19690,19709,19747,19785,19804,19807,19823,19842,19861,19880,19899,19918,19937,19956,19975,19994,20032,20070,20127,20165,20184,20208,20213,20241,20279,20298,20317,20336,20393,20412,20431,20450,20469,20507,20526,20545,20570,20583,20602,20621,20640,20682,20697,20716,20754,20773,20792,20849,20883,20887,20925,20944,20963,21020,21096,21115,21138,21172,21191,21229,21267,21286,21305,21327,21343,21419,21438,21457,21464,21476,21533,21552,21571,21590,21628,21666,21704,21723,21761,21780,21799,21818,21875,21894,21913,21932,21951,21970,22008,22027,22046,22084,22126,22198,22223,22236,22293,22303,22331,22369,22388,22407,22422,22426,22464,22502,22540,22559,22654,22692,22711,22717,22749,22764,22787,22800,22825,22844,22882,22901,22911,22920,22977,23015,23034,23053,23110,23129,23205,23224,23243,23262,23283,23300,23309,23357,23362,23395,23414,23452,23509,23528,23580,23591,23642,23661,23696,23706,23737,23756,23813,23832,23853,23863,23870,23889,23908,23927,23965,23984,24003,24022,24041,24060,24079,24098,24117,24174,24201,24212,24231,24288,24326,24364,24383,24440,24478,24497,24516,24554,24573,24611,24630,24687,24689,24706,24744,24763,24782,24801,24807,24839,24858,24877,24953,24976,25018,25029,25048,25143,25181,25276,25295,25314,25337,25352,25371,25425,25428,25466,25499,25504,25533,25542,25580,25599,25659,25675,25679,25694,25713,25740,25751,25770,25808,25831,25869,25903,25922,25941,25944,25960,26017,26055,26074,26093,26125,26150,26188,26207,26264,26283,26359,26397,26435,26473,26492,26511,26568,26587,26644,26663,26682,26761,26807,26854,26887,26948,27138,27141,27176,27197,27249,27290,27312,27366,27404,27445,27518,27694,27708,27803,27841,27857,27921,28050,28088,28107,28202,28324,28335,28446,28449,28454,28487,28582,28677,28754,28880,28943,28951,28987,29002,29354,29399,29421,29537,29551,29570,29627,29665,29705,29725,29741,29817,29862,29893,29912,29988,30006,30035,30064,30093,30122,30151,30180,30209,30238,30267,30296,30325,30354,30383,30412,30441,30470,30499,30528,30557,30586,30615,30644,30673,30702,30731,30760,30789,30818,30847,30876,30905,30934,30963,30992,31021,31050,31079,31108,31137,31166,31195,31224,31253,31282,31311,31340,31369,31398,31427,31456,31485,31514,31543,31572,31601,31630,31659,31688,31717,31746,31775,31804,31833,31862,31891,31920,31949,31978,32007,32036,32065,32094,32123,32152,32181,32210,32239,32268,32297,32326,32355,32384,32413,32442,32471,32500,32529,32558,32587,32616,32645,32674,32703,32732,32761,32790,32819,32848,32877,32906,32935,32964,32993,33022,33051,33080,33109,33138,33167,33196,33225,33254,33283,33312,33341,33370,33399,33428,33457,33486,33515,33544,33573,33602,33631,33660,33689,33718,33747,33776,33805,33834,33863,33892,33921,33950,33979,34008,34037,34066,34095,34124,34153,34182,34211,34240,34269,34298,34327,34356,34385,34414,34443,34472,34501,34530,34559,34588,34617,34646,34675,34704,34733,34762,34791,34820,34849,34878,34907,34936,34965,34994,35006,35095,35157,35169,35199,35245,35258,35290,35292,35322,35332,35386,35398,35406,35477,35525,35554,35584,35637,35643,35676,35716,35775,35834,35866,35905,35940,35964,36023,36057,36066,36086,36146,36199,36205,36240,36247,36276,36335,36394,36466,36525,36584,36648,36707,36766,36837,36870,36896,36955,37027,37054,37086,37145,37209,37268,37327,37398,37437,37457,37516,37588,37647,37706,37755,37756,37770,37829,37888,37923,37959,37995,38018,38077,38149,38176,38208,38267,38331,38390,38420,38449,38494,38520,38579,38638,38710,38769,38828,38892,38951,39010,39081,39131,39140,39187,39199,39237,39271,39312,39314,39330,39389,39453,39498,39512,39571,39622,39642,39701,39725,39760,39832,39891,39933,39944,39950,40110,40139,40168,40197,40226,40255,40284,40313,40342,40371,40400,40429,40458,40487,40516,40545,40574,40603,40632,40661,40690,40719,40748,40777,40806,40835,40864,40893,40922,40951,40980,41009,41038,41067,41096,41125,41154,41183,41212,41241,41270,41299,41328,41357,41386,41415,41444,41473,41502,41531,42002,42031,42060,42089,42118,42166,42178,42251,42268,42293,42328,42343,42418,42450,42466,42496,42505,42554,42571,42580,42655,42693,42711,42742,42772,42775,42802,42806,42817,42849,42892,42979,43019,43054,43093,43110,43129,43161,43162,43201,43216,43241,43274,43291,43366,43409,43453,43482,43492,43528,43552,43580,43588,43603,43622,43642,43690,43748,43765,43822,43825,43840,43860,43894,43927,43951,43981,43987,44002,44023,44029,44038,44077,44097,44100,44126,44138,44149,44154,44164,44200,44226,44239,44272,44300,44303,44314,44343,44353,44401,44476,44505,44509,44551,44582,44588,44638,44663,44675,44682,44713,44788,44832,44842,44875,44950,45025,45098,45130,45146,45169,45207,45230,45265,45338,45395,45433,45453,45525,45582,45640,45712,45769,45827,45862,45899,45956,45976,45989,46020,46034,46071,46133,46146,46150,46227,46251,46269,46272,46273,46320,46341,46346,46386,46392,46399,46400,46468,46472,46474,46494,46517,46523,46548,46564,46569,46650,46776,46779,46783,46786,46796,46818,46820,46836,46893,46935,46936,46939,46942,46954,47000,47008,47065,47094,47112,47157,47197,47227,47280,47343,47359,47394,47427,47448,47466,47467,47501,47503,47561,47583,47631,47700,47738,47763,47816,47821,47837,47851,47867,47872,47897,47923,47924,47960,47974,47976,47996,47998,48002,48042,48061,48066,48101,48135,48171,48204,48224,48261,48273,48336,48344,48378,48397,48417,48425,48462,48470,48485,48567,48594,48617,48622,48632,48659,48663,48672,48673,48681,48732,48747,48755,48801,48852,48860,48879,48905,48949,48982,49003,49073,49135,49185,49202,49215,49216,49232,49247,49256,49333,49366,49388,49397,49409,49421,49437,49502,49507,49529,49540,49564,49583,49591,49720,49727,49755,49766,49779,49817,49885,49888,49892,49948,49987,49997,50127,50177,50211,50344,50372,50476,50483,50493,50504,50505,50578,50740,50782,50927,50949,50992,50994,51054,51064,51291,51349,51380,51554,51708,51736,51747,51778,51783,51931,51966,52069,52192,52235,52271,52314,52366,52787,52789,52955,53005,53016,53049,53057,53186,53194,53239,53253,53285,53326,53341,53444,53604,53689,53792,53919,53927,53946,53961,53994,54141,54291,54412,54446,54531,54574,54725,54762,54906,54996,55223,55228,55337,55354,55374,55392,55454,55589,55641,55729,55831,55852,55905,55933,56048,56102,56401,56673,56746,56760,56811,56976,57015,57091,57142,57463,57555,57690,57767,57832,57862,57863,58036,58288,58316,58392,58394,58448,58541,58633,58672,58696,58911,58965,58974,58977,58983,59014,59082,59222,59259,59323,59451,59564,59752,59869,59912,60045,60276,60369,60407,60535,60598,60601,60774,60926,60945,61068,61241,61615,61728,61803,61821,62058,62093,62151,62234,62469,62498,62601,62691,62731,62797,62820,62827,62847,62893,62903,63181,63206,63292,63297,63453,63465,63543,63802,63865,64044,64091,64137,64259,64505,64555,64561,65007,65013,65052,65354,65407,65444,65475,65508,65530,65577,65597,65705,65825,65858,65900,65955,66020,66179,66280,66293,66300,66335,66337,66358,66441,66508,66546,66557,66586,66641,66709,66718,66720,66722,66736,66786,66954,67062,67178,67204,67299,67573,67648,67684,67718,67731,67764,67804,67850,67876,67918,67968,68000,68026,68029,68044,68048,68112,68127,68162,68333,68336,68390,68400,68535,68600,68655,68739,68931,68932,68981,69001,69002,69094,69205,69482,69527,69602,69661,69757,69764,69976,70000,70008,70167,70207,70293,70322,70386,70403,70416,70668,70801,70802,70861,71019,71154,71177,71191,71203,71218,71249,71404,71597,71623,71635,71746,71764,71807,71809,71859,72027,72133,72142,72147,72184,72186,72232,72241,72254,72341,72442,72449,72473,72886,72922,73017,73189,73310,73318,73362,73366,73440,73578,73580,73603,73652,73708,73721,73819,73885,73886,73926,74110,74183,74216,74235,74380,74388,74404,74442,74463,74525,74765,75070,75184,75211,75217,75256,75301,75416,75557,75749,75810,75857,75997,76016,76043,76057,76167,76372,76377,76411,76416,76548,76657,76739,76783,76858,76866,76946,76957,77160,77248,77254,77261,77262,77412,77656,77840,77855,77903,78217,78463,78624,78799,78842,78903,78951,78987,79453,79522,79552,79865,79904,79997,80156,80247,80250,80343,80509,80521,80539,80596,80623,80680,80707,80717,80971,81042,81079,81290,81359,81369,81415,81440,81606,81642,81643,81649,81803,81846,81883,82004,82021,82190,82221,82391,82435,82458,82541,82551,82640,82725,82751,82810,82841,82916,83303,83386,83474,83665,83713,83832,84028,84090,84121,84158,84195,84218,84236,84282,84392,84473,84507,84644,84714,84740,84757,84805,85188,85214,85420,85479,85731,85789,85987,86063,86095,86131,86150,86169,86191,86230,86281,86296,86412,86482,86554,86590,86647,86651,86759,86768,86776,86780,86787,86923,87036,87046,87088,87176,87246,87375,87386,87441,87577,87855,87961,88057,88059,88434,88532,88538,88716,88732,88825,88830,89022,89191,89240,89371,89570,89601,89682,89758,89764,89868,89908,89913,89926,89938,90152,90215,90230,90322,90386,90446,90476,90765,90796,90885,90950,91134,91238,91310,91349,91453,91605,91696,91771,91775,91843,91895,91967,92015,92107,92237,92341,92510,92525,92529,92621,92673,92733,92740,92802,92803,92937,93054,93064,93071,93239,93471,93504,93537,93541,93567,93634,93729,93808,93810,93960,93997,94009,94069,94103,94105,94309,94566,94597,94795,95050,95134,95173,95196,95311,95359,95397,95445,95478,95491,95530,95615,95757,95776,95792,95871,95873,96007,96145,96188,96205,96281,96367,96470,96708,96726,96762,96916,96927,96935,96942,97061,97097,97126,97195,97249,97472,97537,97562,97639,97673,97697,97801,97805,97815,97875,97901,97908,97948,98229,98396,98397,98425,98462,98612,98663,98793,98853,98915,98951,98981,99025,99304,99315,99319,99523,99596,99685,99694,99758,99993,100002",
          "203,256,286,340,379,400,421,424,425,450,461,468,499,586,590,693,705,724,768,848,862,867,1005,1016,1078,1125,1152,1206,1237,1245,1271,1395,1494,1535,1540,1801,5302,5424,5465,5549,5574,5642,5750,5915,6035,6045,6087,6091,6127,6177,6356,6392,6430,6517,6518,6609,6906,6976,6985,7108,7123,7204,7243,7311,7319,7361,7406,7439,7445,7539,7669,7731,7735,7737,7901,7940,7944,8006,8020,8071,8221,8368,8378,8565,8664,8774,8864,9025,9078,9110,9364,9594,9856,9899,9917,10004,10079,10162,10214,10369,10384,10500,10612,10677,10690,10821,10867,10871,10908,10916,10983,10998,11781,11836,12348,12426,13157,13925,14057,14161,14401,14476,14776,14872,15312,15677,16096,16452,16512,16931,17133,17630,17646,18670,18749,18936,20520,20556,22241,24016,24073,24262,24702,25044,26313,26762,26787,26837,26863,26973,27201,27397,27681,28109,28176,28844,28949,29244,29296,29508,30599,31241,35707,36138,36197,36269,36371,36453,36459,36757,37003,37186,37260,37373,37449,37697,37727,37758,37822,37931,37948,38256,38682,38999,39377,39419,39489,39943,40792,41239,44374,45167,46684,48595,49966,50079,50233,50416,50486,50519,50575,50646,50854,50865,50908,51217,51369,51407,51608,51642,51652,51989,51995,52161,52177,52187,52231,52250,52478,52541,52780,52798,52809,52890,52914,53079,53138,53162,53199,53217,53291,53316,53835,53877,54019,54072,54162,54331,54339,54416,54515,54521,54740,54875,54886,54959,54978,55017,55030,55033,55055,55149,55153,55253,55272,55422,55618,55686,55709,55721,56056,56136,56145,56263,56289,56317,56375,56391,56479,56595,56785,56800,56812,56825,56848,56907,57035,57170,57238,57319,57375,57406,57519,57605,57659,57673,57801,57820,57952,57965,58043,58102,58176,58297,58326,58431,58567,58652,58676,58934,58952,59316,59405,59413,59457,59713,59723,59733,59816,59836,59918,59929,60009,60033,60236,60246,60266,60497,60502,60555,60576,60583,60620,60626,60770,60786,60954,61034,61164,61427,61443,61533,61598,61604,61613,61677,61750,61752,61779,61830,61992,62143,62174,62307,62482,62570,62658,62713,62813,62839,62891,62972,63025,63131,63226,63351,63628,63639,63687,63784,63947,64225,64498,64513,64636,64729,65036,65148,65187,65222,65228,65255,65285,65389,65396,65445,65580,65640,65676,65678,65701,65722,65845,65910,65941,65966,65980,66057,66171,66234,66302,66329,66336,66359,66410,66436,66494,66531,66538,66551,66767,66817,66854,66875,66921,67126,67363,67467,67470,67478,67479,67575,67758,67767,67779,67788,67823,67866,67872,67882,67947,67991,68027,68143,68221,68244,68247,68328,68341,68447,68480,68626,68669,68695,68711,68819,68987,69158,69386,69553,69691,69710,69948,70052,70110,70236,70536,70596,70763,70775,70819,70821,70865,70928,70974,71195,71331,71338,71387,71560,71594,71602,71610,71653,71695,71717,71896,71903,71933,71981,72161,72168,72177,72297,72343,72359,72561,72633,72682,72754,72782,72825,72872,73041,73132,73172,73322,73338,73447,73582,73661,73720,73834,73839,73845,73866,73929,74031,74058,74108,74163,74441,74766,74768,74841,74872,74969,75016,75062,75110,75149,75158,75313,75398,75417,75514,76213,76266,76388,76407,76430,76461,76520,76527,76558,76695,76758,76805,76831,77139,77200,77278,77285,77291,77427,77478,77548,77603,77737,77815,77983,78050,78240,78501,78644,78650,78882,78949,79038,79233,79257,79397,79410,79412,79426,79448,79464,79486,79503,79660,79668,79687,79704,79713,79728,79804,79853,79856,79947,79950,80028,80043,80057,80160,80443,80525,80576,80702,80833,80853,80856,80867,80908,80930,80931,80948,80998,81015,81047,81127,81217,81264,81460,81502,81632,81671,81710,81743,81746,81765,81872,81875,81958,81971,82034,82063,82066,82253,82343,82347,82487,82516,82521,82592,82618,82764,82808,83040,83102,83202,83335,83378,83409,83424,83518,83526,83710,83850,84004,84124,84306,84335,84401,84526,84557,84716,84765,84785,85249,85374,85378,85425,85634,85670,85699,85756,85855,85986,86056,86083,86147,86203,86237,86505,86528,86570,86597,86684,86711,86713,86796,86890,86981,87026,87449,87532,87557,87565,88037,88043,88072,88134,88166,88328,88357,88392,88552,88585,88619,88622,88712,89001,89173,89252,89425,89630,89632,89635,89806,89905,90101,90137,90440,90703,90716,90748,90770,90827,90850,90929,90944,90964,91231,91239,91269,91303,91407,91519,91639,91650,91682,91760,91773,91860,91909,92003,92006,92033,92066,92139,92263,92275,92420,92766,92788,92795,92839,92947,93050,93063,93232,93242,93292,93317,93330,93345,93380,93421,93484,93703,93834,93998,94086,94111,94239,94363,94401,94579,94591,94636,94838,94852,95102,95121,95133,95189,95197,95343,95430,95607,95692,95759,96025,96313,96488,96567,96669,96670,96795,96901,96917,96972,96994,96996,97066,97198,97226,97256,97395,97402,97530,97557,97620,97633,97669,97717,97718,97796,98025,98132,98182,98246,98308,98452,98667,98698,98823,98908,99034,99186,99292,99444,99469,99767,99831,99834,99907,99936",
          "137,208,219,434,931,977,1158,1182,1272,1431,1641,1732,1765,1799,1815,1829,1839,2047,2076,2359,2564,2690,2748,2798,2857,3042,3221,3367,3421,3500,3613,3655,3702,3795,3804,4127,4171,4185,4266,4383,4550,4572,4575,4741,5029,5073,5096,5156,5163,5219,5262,5312,5326,5372,5488,5515,5524,5533,5571,5747,5809,5816,5887,5901,5940,6101,6167,6293,6383,6400,6528,6551,6564,6578,6634,6702,6940,6945,6947,6992,7013,7047,7163,7308,7350,7412,7546,7575,7649,7710,7719,7796,7814,7861,7909,7954,7995,8143,8164,8216,8229,8289,8396,8414,8492,8624,8674,8756,8760,8788,8821,8831,8842,8931,9235,9277,9464,9581,9599,9730,9794,9865,9916,9933,9983,10058,10068,10126,10230,10276,10322,10325,10341,10530,10550,10607,10624,10719,10733,10765,10813,10834,10944,10986,11025,11082,11143,11232,11323,11365,11416,11557,11560,11660,11697,11838,11906,11937,11956,11990,12114,12146,12278,12335,12578,12739,12811,12891,12911,13137,13171,13276,13380,13407,13427,13557,13584,13665,13799,13842,13860,13943,13990,14198,14329,14402,14460,14763,14803,14804,14806,14894,14956,14982,15014,15214,15224,15254,15280,15413,15436,15505,15728,15736,15868,15905,16039,16095,16211,16212,16401,16801,16818,16937,17050,17118,17266,17589,17602,17635,17812,17814,17883,17920,17928,18457,18471,18539,18542,18672,18785,19138,19156,19288,19427,19528,19634,19659,19698,19736,19896,19934,20087,20124,20183,20270,20497,20580,20714,20744,20822,21149,21174,21470,21526,21678,21688,21748,21849,21873,22138,22185,22233,22246,22286,22310,22314,22425,22482,22595,22597,22657,22659,22862,22976,23019,23032,23033,23076,23298,23547,23621,23729,23824,23833,24097,24128,24245,24287,24496,24593,24724,24825,24844,24874,24894,24912,25083,25142,25333,25521,25617,25692,25783,25901,25989,26217,26232,26340,26491,26710,26788,26806,26877,26975,27083,27091,27146,27236,27327,27439,27448,27461,27545,27666,27745,27754,28092,28187,28203,28311,28348,28364,28416,28465,28467,28617,28685,28766,28860,28903,28932,28942,29017,29151,29319,29430,29462,29598,29834,29933,29966,29978,30005,30034,30092,30177,30266,30293,30903,31367,32004,32180,32325,32527,32615,32702,32729,32816,33367,33717,33948,34122,34152,34500,34528,34558,34645,34818,35147,35208,35583,35685,35819,35878,36079,36139,36202,36450,36465,36502,36514,36691,36932,36944,37011,37053,37071,37073,37077,37383,37395,37483,37506,37675,37704,38043,38062,38122,38126,38142,38329,38330,38577,38603,39071,39126,39262,39623,39635,39680,39731,39750,39857,39878,39930,40669,41097,41184,41234,41358,42117,42261,42405,42506,42709,42982,43080,43170,43219,43403,43404,43434,43497,43693,43888,44167,44526,44558,44670,44720,44807,44933,45067,45233,45320,45617,45671,45697,46154,46338,46382,46979,47017,47139,47732,47815,47936,48102,48142,48242,48991,49042,49224,49803,49836,50030,50074,50140,50283,50337,50445,50462,50479,50513,50530,50563,50584,50590,50651,50838,50891,50892,50932,50956,51088,51111,51161,51280,51391,51474,51500,51502,51508,51587,51624,51640,51646,51779,51808,51868,51955,51983,52053,52055,52238,52419,52422,52440,52475,52540,52611,52711,53001,53076,53108,53133,53159,53181,53303,53384,53564,53700,53815,53859,53883,53923,53924,54012,54038,54100,54143,54288,54289,54385,54397,54406,54428,54489,54529,54551,54757,54865,54871,54910,54923,55016,55046,55076,55390,55553,55636,55771,55914,55941,55990,56020,56061,56147,56160,56346,56458,56490,56541,56810,56883,56895,56900,57027,57106,57129,57172,57262,57324,57344,57367,57489,57513,57687,57710,57743,57744,57848,57959,57990,58057,58171,58219,58305,58318,58442,58472,58549,58554,58565,58852,58864,59061,59193,59258,59317,59374,59462,59498,59563,59634,59652,59754,59807,59852,59958,60253,60384,60485,60556,60589,60654,60823,60890,60899,61094,61203,61214,61240,61315,61321,61346,61374,61594,61759,61978,62012,62034,62139,62320,62338,62471,62725,62773,62902,63093,63141,63229,63266,63295,63366,63531,63665,63737,63803,63868,63906,63928,64035,64104,64159,64347,64548,64605,64622,64683,64697,64703,64766,64850,64949,64962,64993,65027,65080,65094,65113,65124,65167,65277,65613,65633,65704,65708,65770,65943,66005,66069,66192,66209,66491,66578,66692,66768,66776,66910,66928,67006,67088,67323,67367,67390,67450,67547,67549,67576,67582,67680,67773,67793,67855,67939,68195,68259,68265,68285,68365,68510,68612,68681,68890,68977,69042,69054,69172,69282,69410,69564,69706,69716,70208,70245,70478,70629,70776,70791,70885,70917,70977,71076,71257,71358,71761,71778,71840,71851,72162,72282,72379,72534,72556,72689,72904,73225,73252,73261,73294,73353,73672,73788,73904,74010,74101,74139,74303,74322,74377,74431,74633,74696,74779,74790,74938,74963,75004,75011,75019,75156,75248,75312,75319,75331,75361,75439,75474,75794,75798,75831,76080,76210,76229,76296,76325,76660,76690,76932,77022,77032,77297,77418,77426,77431,77525,77741,77879,77967,77976,78004,78011,78180,78273,78382,78494,78558,78642,78745,78829,78879,78881,78898,78899,79004,79122,79159,79384,79729,79749,79788,79842,79897,79945,80085,80111,80135,80153,80260,80353,80373,80477,80553,80589,80590,80840,80862,81005,81165,81180,81184,81266,81363,81429,81548,81557,81713,81747,81756,81866,81959,82143,82155,82229,82242,82292,82332,82648,82792,82967,82989,83028,83083,83091,83169,83174,83243,83261,83279,83283,83345,83373,83450,83466,83537,83554,83872,83892,83893,84127,84275,84333,84396,84518,84549,84583,84601,84723,84762,84770,84808,84892,84900,84915,84920,84942,84955,85115,85261,85335,85506,85610,85759,85805,85911,85936,85977,86090,86097,86124,86128,86195,86200,86233,86239,86305,86326,86432,86457,86661,87429,87438,87451,87462,87573,87601,87751,87957,88028,88355,88473,88496,88579,88649,88704,88804,88851,88872,88906,88922,88975,88976,88982,89037,89069,89073,89117,89200,89224,89255,89267,89381,89745,89902,90038,90067,90268,90536,90566,90826,90875,91015,91129,91170,91198,91219,91320,91361,91429,91452,91543,91603,91704,91850,91889,91921,92125,92181,92200,92227,92267,92556,92836,92943,92956,92971,92981,92984,93006,93008,93101,93138,93150,93173,93289,93313,93332,93351,93467,93473,93495,93626,93629,93687,93793,94031,94037,94043,94107,94118,94505,94567,94618,94649,94856,94891,94896,94962,95043,95222,95303,95360,95476,95550,95554,95779,95844,95922,95969,96138,96231,96295,96330,96502,96855,96925,96995,97137,97246,97442,97529,97547,97579,97644,97691,97736,97755,97767,97779,97888,97954,98007,98039,98169,98256,98260,98272,98332,98424,98433,98531,98564,98659,98672,99071,99129,99214,99262,99314,99379,99462,99478,99610,99717,99734,99747,99799,99857,99866,99922,99957,99972",
          "49,68,125,144,241,2072,2325,2556,2609,2630,2649,2908,3044,3059,3074,3090,3173,3873,3900,4174,4404,4461,4579,4617,4641,4746,4845,6879,7676,9568,9587,9644,9663,9682,9701,9715,9739,9796,9834,9840,10043,10708,10746,11240,11373,11392,11430,11554,11563,11601,11639,11658,11848,11921,11924,11943,12057,12076,12152,12209,12228,12247,12266,12285,12298,12361,12380,12399,12456,12465,12475,12551,12570,12627,12683,12684,12703,12770,12779,12798,12807,12836,12893,12912,12988,12993,13026,13064,13083,13121,13140,13178,13197,13232,13235,13255,13273,13292,13330,13358,13365,13368,13406,13444,13463,13482,13494,13503,13520,13539,13558,13575,13577,13596,13613,13615,13634,13653,13672,13710,13741,13767,13786,13805,13843,13862,13900,13904,13938,13957,13973,13976,13995,14007,14033,14052,14067,14109,14128,14166,14185,14204,14223,14280,14299,14318,14337,14356,14394,14413,14431,14432,14451,14470,14501,14527,14546,14565,14603,14622,14660,14698,14755,14789,14793,14808,14817,14831,14850,14868,14869,14888,14898,14945,14964,15002,15021,15097,15116,15154,15158,15192,15212,15230,15249,15306,15325,15382,15420,15439,15477,15534,15553,15629,15705,15743,15819,15838,15857,15914,15952,16028,16040,16047,16104,16159,16161,16180,16182,16218,16237,16239,16244,16256,16275,16294,16313,16351,16389,16408,16411,16465,16522,16560,16598,16602,16617,16655,16677,16689,16750,16769,16778,16845,16883,16902,16997,17007,17035,17073,17092,17130,17206,17225,17263,17282,17301,17320,17339,17377,17415,17434,17510,17529,17567,17605,17661,17702,17719,17738,17753,17757,17776,17795,17833,17852,17890,17909,17985,17989,18004,18020,18039,18042,18061,18080,18102,18118,18134,18137,18156,18175,18270,18308,18346,18365,18399,18403,18418,18422,18441,18498,18517,18536,18555,18605,18621,18631,18669,18726,18821,18897,18954,18992,18996,19011,19014,19030,19049,19068,19087,19108,19125,19163,19277,19315,19391,19429,19467,19486,19524,19543,19562,19580,19600,19602,19638,19695,19714,19752,19790,19809,19847,19866,19885,19904,19923,19942,19961,19999,20018,20056,20075,20170,20189,20227,20228,20246,20280,20303,20322,20324,20359,20360,20364,20379,20398,20417,20436,20451,20493,20547,20550,20623,20626,20645,20664,20683,20702,20717,20721,20740,20816,20835,20873,20911,20930,20949,20968,21006,21010,21032,21063,21120,21122,21139,21177,21215,21234,21235,21253,21272,21315,21329,21348,21367,21386,21424,21443,21462,21481,21595,21614,21633,21671,21709,21728,21747,21839,21861,21880,21882,21911,21918,21937,21956,21975,21994,22013,22032,22070,22089,22108,22127,22150,22168,22203,22222,22257,22279,22300,22317,22336,22355,22393,22412,22431,22450,22469,22488,22500,22507,22526,22561,22564,22602,22621,22640,22661,22697,22716,22773,22792,22811,22817,22849,22906,22925,22963,22982,22986,23039,23096,23115,23131,23138,23172,23210,23248,23286,23324,23343,23381,23400,23438,23476,23495,23514,23524,23571,23590,23628,23647,23662,23685,23704,23745,23761,23799,23818,23894,23913,23932,23951,23989,24004,24024,24027,24046,24065,24103,24114,24141,24164,24200,24217,24236,24274,24312,24350,24369,24388,24407,24411,24426,24428,24430,24445,24464,24502,24503,24509,24521,24540,24559,24578,24616,24670,24673,24680,24692,24711,24730,24768,24863,24920,24958,25015,25034,25051,25053,25057,25060,25091,25106,25110,25111,25148,25167,25190,25224,25243,25281,25319,25342,25357,25361,25411,25414,25433,25490,25528,25547,25587,25598,25604,25642,25661,25699,25720,25737,25756,25772,25794,25832,25870,25889,25908,25920,25927,25946,25965,26041,26060,26079,26098,26117,26119,26136,26155,26174,26193,26269,26288,26345,26383,26438,26588,26846,26870,26894,26915,26930,26972,27095,27124,27181,27190,27281,27415,27428,27542,27580,27599,27606,27627,27683,27713,27732,27864,27943,27971,27998,28000,28014,28036,28093,28169,28283,28321,28378,28397,28466,28511,28537,28549,28587,28644,28704,28739,28744,28758,28774,28872,28929,28941,28986,29005,29062,29081,29138,29154,29176,29233,29309,29347,29366,29395,29480,29556,29610,29613,29693,29698,29727,29784,29822,29914,29955,30011,30040,30069,30098,30127,30156,30181,30185,30214,30243,30256,30272,30301,30330,30359,30417,30475,30504,30533,30562,30591,30620,30649,30678,30707,30736,30765,30794,30823,30852,30881,30910,30939,30968,30997,31026,31055,31084,31113,31142,31171,31200,31229,31287,31316,31345,31374,31403,31432,31461,31490,31519,31548,31577,31635,31664,31693,31722,31751,31780,31809,31838,31867,31896,31925,31954,31983,32012,32041,32070,32099,32128,32157,32186,32215,32244,32273,32302,32331,32360,32389,32418,32447,32476,32505,32534,32563,32592,32621,32650,32679,32708,32737,32766,32795,32824,32853,32882,32911,32940,32969,32998,33027,33056,33085,33114,33143,33172,33201,33230,33259,33288,33313,33317,33346,33375,33404,33433,33462,33491,33520,33549,33578,33607,33636,33665,33694,33723,33752,33781,33810,33839,33868,33897,33926,33955,33984,34013,34042,34071,34100,34112,34129,34158,34187,34216,34245,34274,34303,34332,34361,34390,34412,34419,34448,34477,34506,34535,34564,34593,34622,34651,34680,34709,34738,34767,34796,34825,34854,34883,34912,34941,34970,34999,35011,35100,35124,35174,35224,35263,35298,35337,35369,35371,35375,35411,35435,35482,35504,35510,35530,35568,35589,35648,35721,35780,35803,35839,35872,35910,35969,35999,36028,36070,36091,36151,36210,36281,36320,36333,36340,36399,36471,36501,36530,36589,36653,36712,36771,36842,36881,36901,36925,36960,36990,36998,36999,37021,37032,37091,37150,37190,37214,37251,37273,37332,37371,37374,37403,37462,37487,37501,37593,37624,37652,37711,37751,37753,37775,37811,37834,37859,37893,37913,37918,37921,37933,37964,38006,38023,38082,38154,38213,38261,38272,38314,38336,38361,38395,38433,38435,38454,38480,38525,38562,38584,38636,38643,38680,38715,38774,38805,38833,38875,38897,38956,39015,39057,39086,39145,39170,39178,39204,39244,39276,39308,39335,39363,39394,39458,39492,39517,39576,39618,39647,39673,39689,39706,39765,39793,39795,39837,39896,39955,39981,40080,40088,40092,40105,40134,40163,40192,40221,40250,40279,40308,40337,40366,40395,40424,40453,40482,40511,40540,40569,40598,40627,40656,40685,40714,40743,40772,40801,40830,40859,40888,40917,40946,40975,41004,41033,41062,41091,41120,41149,41178,41207,41222,41236,41294,41323,41352,41381,41410,41439,41468,41497,41526,41998,42027,42056,42085,42114,42174,42199,42203,42207,42225,42242,42264,42290,42309,42339,42367,42392,42414,42501,42525,42531,42576,42600,42605,42632,42651,42738,42764,42793,42813,42836,42855,42888,42925,42975,43050,43092,43104,43125,43157,43182,43212,43235,43287,43329,43341,43362,43407,43435,43436,43449,43473,43481,43500,43524,43554,43599,43653,43682,43686,43730,43761,43785,43795,43816,43836,43923,43950,43972,43973,43998,44073,44114,44160,44186,44196,44235,44283,44310,44346,44356,44397,44445,44472,44503,44547,44573,44634,44674,44691,44699,44709,44759,44766,44784,44844,44871,44907,44946,44971,45000,45021,45074,45094,45122,45142,45186,45203,45255,45261,45334,45391,45449,45471,45480,45494,45521,45564,45578,45605,45623,45636,45681,45682,45708,45758,45765,45810,45817,45823,45868,45895,45952,45991,45997,46006,46058,46081,46086,46169,46179,46197,46207,46232,46237,46244,46252,46262,46265,46285,46322,46356,46361,46366,46432,46458,46480,46509,46526,46542,46562,46626,46643,46660,46662,46704,46760,46788,46816,46821,46862,46864,46902,46986,47006,47072,47085,47109,47131,47132,47148,47179,47188,47248,47249,47263,47265,47295,47298,47311,47375,47425,47468,47471,47472,47483,47529,47537,47614,47619,47621,47647,47648,47653,47657,47674,47697,47702,47734,47776,47832,47874,47875,47889,47891,47939,47952,47966,47969,47972,48018,48037,48038,48082,48083,48103,48120,48157,48179,48209,48259,48293,48333,48379,48391,48396,48399,48456,48468,48483,48500,48534,48545,48561,48565,48642,48666,48709,48713,48731,48740,48865,48868,48883,48919,48942,48947,48967,49013,49043,49104,49168,49174,49188,49227,49248,49252,49254,49279,49316,49325,49340,49342,49396,49418,49438,49458,49468,49526,49556,49595,49625,49634,49773,49829,49846,49981,50148,50153,50154,50202,50216,50340,50343,50346,50384,50452,50540,50547,50642,50748,50813,50828,50898,50936,51036,51078,51095,51107,51192,51260,51262,51522,51552,51566,51589,51716,51740,51816,51920,52026,52067,52077,52098,52123,52595,53028,53045,53053,53086,53152,53209,53218,53238,53518,53546,53615,53637,53729,53823,53833,53941,53995,54011,54017,54064,54234,54342,54353,54365,54545,54747,54780,54787,54804,54842,54867,54878,55105,55184,55552,55745,55764,55917,55982,56076,56103,56193,56198,56228,56754,56833,57044,57113,57320,57390,57414,57475,57502,57552,57633,57635,57736,57835,57891,58098,58132,58160,58204,58208,58391,58404,58457,58519,58581,58589,58640,58657,58851,58878,58897,58936,58953,59031,59048,59137,59239,59240,59249,59397,59409,59470,59477,59535,59711,59714,59716,60021,60121,60176,60190,60226,60275,60341,60366,60423,60493,60507,60562,60748,60862,60893,60897,60910,60948,61051,61079,61115,61126,61205,61312,61444,61735,61797,61867,61868,61999,62041,62138,62187,62299,62332,62438,62465,62503,62536,62639,62873,62900,62963,63103,63239,63414,63457,63513,63612,63712,63716,63717,63742,63878,64024,64082,64092,64108,64332,64364,64472,64701,64977,65018,65042,65116,65307,65342,65582,65600,65832,66090,66092,66097,66098,66110,66310,66533,66596,66920,66964,67092,67122,67175,67266,67421,67472,67568,67649,67658,67707,67766,67768,67810,67863,68046,68103,68158,68181,68269,68374,68482,68487,68506,68635,68647,68649,68671,68683,68689,68717,68831,69026,69033,69201,69311,69433,69700,69761,69803,69827,69896,69950,69974,70017,70118,70152,70303,70379,70590,70630,70806,70875,70916,71012,71030,71036,71348,71444,71612,71650,71721,71742,71863,71915,72108,72123,72146,72296,72363,72789,72848,73027,73060,73084,73226,73234,73270,73275,73369,73495,73504,73595,73677,73781,73878,74132,74253,74348,74418,74454,74563,74801,74813,74878,74911,74917,74979,74982,75207,75275,75353,75464,75542,75645,75915,75944,76094,76188,76191,76267,76320,76417,76602,76662,76906,76923,76964,76973,77035,77086,77133,77146,77188,77223,77399,77416,77500,77544,77546,77576,77696,77795,77943,77999,78109,78134,78189,78225,78226,78363,78493,78535,78553,78752,78774,78795,79033,79394,79461,79542,79600,79643,79707,79912,80051,80056,80065,80143,80492,80583,80807,80859,80985,81048,81058,81106,81126,81158,81176,81242,81291,81332,81672,81819,81884,81909,82036,82185,82224,82238,82241,82341,82455,82596,82598,82882,82977,83129,83147,83157,83181,83191,83217,83256,83260,83268,83289,83611,83653,83712,83781,83805,83822,83971,83978,84102,84115,84177,84294,84322,84535,84619,84700,84708,84736,84884,84909,84927,84949,84973,85009,85108,85270,85398,85491,85742,85782,85842,86007,86080,86141,86185,86205,86300,86535,86648,86656,86690,86729,86772,86920,86960,87081,87155,87181,87267,87272,87370,87403,87431,87517,87622,87680,87711,87726,87871,87893,87953,88001,88174,88175,88346,88390,88421,88545,88597,88671,88675,88684,88738,88755,88798,88829,88869,89152,89177,89198,89223,89266,89272,89344,89375,89397,89488,89511,89700,89880,89984,90184,90331,90355,90434,90435,90618,90677,90778,91218,91314,91367,91438,91595,91647,91772,91783,91848,91940,92007,92039,92127,92153,92222,92357,92593,92599,92620,92631,92635,92656,92676,92678,92683,92707,92898,92952,93133,93160,93198,93233,93326,93456,93726,93789,93847,93863,93907,93914,93922,93971,94062,94128,94389,94410,94420,94439,94754,94761,94890,94911,94924,94943,94980,95150,95293,95513,95542,95590,95785,95799,95801,95839,95931,95944,95949,96044,96139,96151,96213,96234,96324,96413,96547,96575,96616,96622,96654,96741,96945,96948,97094,97119,97143,97188,97309,97346,97412,97551,97676,97746,97776,97785,97821,98093,98102,98191,98211,98213,98364,98580,98599,98602,98979,99142,99231,99440,99562,99591,99619,99623,99628,99661,99702,99714,99727,99763,99788,99888",
          "70,74,143,175,194,242,1737,1934,1999,2001,2112,2190,2377,2427,2476,2602,2694,2915,3189,3275,3468,3557,3576,3671,3751,3890,3926,3990,4013,4032,4070,4136,4163,4212,4330,4370,4462,4600,4796,5093,6103,6287,6635,6731,6977,7547,7661,8117,8136,8155,8554,8573,8649,8668,8673,8687,8808,8972,9107,9143,9147,9181,9293,9430,9580,9618,9637,9642,9656,9694,9867,9903,9998,10066,10074,10093,10155,10188,10207,10226,10321,10397,10549,10586,10644,10777,10894,10967,10990,11024,11083,11100,11118,11147,11157,11195,11214,11280,11301,11309,11347,11356,11384,11423,11461,11499,11564,11594,11708,11765,11803,11879,11955,12012,12031,12035,12098,12107,12126,12183,12221,12240,12316,12394,12411,12421,12449,12506,12536,12542,12587,12620,12625,12677,12696,12701,12789,12905,12908,12915,12924,13038,13095,13152,13168,13209,13222,13247,13266,13285,13323,13456,13475,13532,13578,13589,13646,13722,13760,13779,13798,13817,13931,13950,13988,14064,14178,14197,14235,14273,14292,14330,14368,14400,14425,14449,14480,14504,14520,14682,14805,14862,14901,14938,14995,15011,15066,15071,15204,15299,15327,15333,15337,15356,15394,15451,15470,15527,15554,15622,15698,15755,15831,15869,15888,15907,15945,15964,16097,16119,16249,16325,16344,16363,16420,16439,16496,16515,16553,16554,16564,16568,16572,16610,16648,16667,16734,16800,16811,16819,16895,16914,16933,16990,17047,17085,17180,17218,17237,17284,17313,17370,17389,17522,17541,17617,17634,17636,17674,17677,17712,17750,17807,17845,17864,17902,17940,17959,17978,18092,18111,18149,18206,18244,18301,18320,18339,18358,18396,18472,18491,18510,18562,18574,18624,18662,18700,18757,18852,18947,19042,19099,19118,19137,19251,19287,19297,19327,19384,19422,19448,19460,19515,19555,19612,19631,19669,19725,19745,19783,19802,19859,19878,19916,19935,19954,19992,20030,20086,20125,20140,20163,20193,20239,20315,20372,20391,20410,20429,20448,20503,20505,20562,20600,20619,20733,20752,20866,20871,20901,20923,20942,21075,21113,21132,21170,21227,21265,21379,21417,21474,21493,21550,21588,21721,21750,21759,21816,21892,22044,22120,22139,22158,22196,22234,22253,22272,22348,22386,22424,22462,22481,22519,22538,22633,22671,22747,22785,22804,22861,22899,22918,23051,23070,23127,23203,23220,23260,23336,23355,23393,23431,23437,23488,23556,23583,23640,23646,23659,23786,23790,23792,23819,23944,24001,24039,24058,24096,24115,24172,24210,24248,24267,24343,24362,24400,24495,24571,24609,24628,24666,24761,24780,24791,24837,24913,24970,25160,25198,25236,25293,25312,25350,25388,25426,25445,25464,25523,25524,25540,25578,25654,25730,25749,25825,25844,25939,25996,26034,26072,26129,26164,26186,26195,26205,26243,26319,26326,26338,26376,26395,26433,26452,26471,26604,26642,26660,26693,26699,26756,26760,26832,26984,27152,27174,27304,27307,27342,27350,27390,27412,27421,27725,27763,27764,27991,28029,28390,28429,28545,28561,28618,28752,29018,29047,29055,29093,29279,29487,29511,29756,29872,29948,29967,30004,30062,30091,30207,30236,30265,30294,30323,30326,30410,30439,30497,30555,30584,30613,30671,30700,30729,30758,30816,30932,30961,31019,31048,31077,31106,31193,31222,31309,31338,31396,31454,31483,31512,31599,31628,31657,31686,31715,31744,31773,31802,31860,31889,31918,31947,31976,32005,32034,32063,32092,32121,32150,32208,32237,32266,32295,32411,32556,32585,32614,32643,32672,32701,32730,32759,32788,32817,32846,32875,32904,32991,33020,33049,33078,33136,33165,33194,33223,33252,33281,33339,33397,33426,33455,33542,33571,33600,33629,33658,33745,33774,33803,33832,33861,33890,33919,33977,34006,34035,34064,34151,34238,34354,34383,34441,34470,34499,34586,34615,34673,34702,34789,34905,34934,34963,35052,35080,35167,35250,35303,35330,35388,35404,35475,35523,35549,35641,35747,35773,35903,35938,35962,36084,36144,36203,36392,36508,36523,36582,36632,36705,36760,36764,36794,36835,36876,36894,36931,36953,36982,37025,37080,37084,37143,37207,37246,37262,37266,37315,37325,37396,37455,37586,37645,37684,37768,37827,37872,37878,37917,37926,37957,38016,38113,38206,38242,38429,38434,38447,38497,38514,38518,38570,38767,38826,38890,38949,39008,39138,39164,39197,39265,39328,39361,39372,39387,39569,39636,39640,39699,39758,39787,39889,40176,40205,40234,40263,40321,40350,40408,40437,40466,40495,40524,40611,40640,40698,40727,40756,40785,40843,40901,40930,41017,41075,41104,41162,41220,41307,41365,41394,41423,41440,41452,41510,41539,42009,42038,42067,42125,42142,42208,42301,42331,42350,42425,42462,42479,42512,42587,42641,42749,42899,42934,42986,43005,43008,43021,43036,43074,43099,43136,43156,43200,43223,43298,43331,43373,43393,43516,43535,43550,43565,43610,43697,43934,44050,44084,44131,44152,44193,44246,44265,44385,44408,44481,44483,44528,44589,44645,44666,44676,44686,44735,44795,44821,44852,45013,45045,45075,45076,45214,45272,45345,45402,45485,45532,45589,45647,45719,45787,45834,45853,45906,45925,45963,46041,46044,46181,46182,46184,46196,46205,46390,46446,46531,46580,46609,46663,46692,46722,46755,46785,46787,46843,46993,47002,47010,47023,47030,47095,47116,47171,47172,47235,47239,47308,47323,47402,47415,47493,47592,47695,47830,47865,47899,47920,47990,47994,48064,48106,48125,48258,48265,48331,48435,48445,48700,48719,48796,48824,48842,49054,49111,49178,49257,49281,49292,49302,49310,49399,49455,49463,49539,49562,49649,49662,49689,49691,49711,49712,49807,49887,50015,50083,50098,50284,50326,50697,50862,50902,50996,51118,51370,51482,51605,51734,51814,51879,51897,52252,52303,52369,52522,52529,52615,52648,52727,52732,52746,52881,53216,53306,53312,53382,53439,53441,53632,53797,54148,54233,54247,54387,54460,54483,54487,54492,54517,54616,54708,54716,54756,54914,54943,55043,55134,55213,55219,55362,55435,55444,55685,55737,55797,55916,55946,55988,56063,56126,56134,56194,56200,56205,56269,56279,56476,56512,56665,56718,56741,56777,57006,57048,57050,57076,57114,57192,57202,57215,57331,57434,57551,57681,57685,57777,58047,58110,58152,58364,58412,58462,58542,58543,58627,58635,58646,58654,58661,58663,58675,58760,58812,59021,59036,59134,59140,59149,59162,59163,59202,59482,59496,59671,59672,59742,59892,59921,60040,60134,60636,60779,60785,60838,60877,60886,60938,61013,61063,61099,61108,61136,61157,61197,61319,61334,61509,61544,61666,61699,61725,61738,61844,61919,62038,62057,62089,62224,62227,62278,62345,62357,62531,62568,62676,62740,62843,63015,63029,63037,63105,63381,63393,63472,63492,63545,63585,63608,63682,63927,63957,63981,63995,63999,64015,64126,64127,64171,64337,64439,64520,64653,64690,64810,64821,64884,64891,64903,64975,64997,65019,65191,65294,65300,65336,65358,65411,65423,65431,65478,65586,65607,65614,65626,65642,65738,65769,66246,66296,66535,66604,66745,66809,66841,66903,66914,66932,66962,67289,67374,67572,67611,67945,67960,67983,68061,68177,68185,68200,68808,69212,69334,69388,69400,69401,69427,69549,69577,69654,69732,69758,69790,69825,69994,70056,70080,70181,70268,70292,70296,70367,70458,70487,70603,70688,70696,70772,70857,70859,70960,71207,71342,71360,71459,71542,71684,71685,71768,71774,71948,72036,72050,72090,72176,72191,72329,72372,72418,72567,72586,72735,72775,72892,72924,73107,73119,73349,73428,73485,73549,73775,73821,73888,73969,73976,74001,74085,74091,74212,74241,74267,74344,74445,74460,74462,74514,74615,74626,74764,74824,74876,75012,75185,75299,75386,75410,75446,75570,75589,75905,75954,76133,76144,76148,76186,76222,76380,76635,76749,76809,76873,77070,77156,77163,77175,77386,77464,77580,77697,77731,77904,78335,78341,78432,78462,78475,78497,78583,78591,78693,79006,79018,79275,79341,79381,79427,79460,79642,79650,79785,79894,79908,79990,80013,80274,80466,80469,80487,80536,80572,80736,81004,81019,81097,81125,81306,81396,81413,81422,82149,82468,82478,82519,82580,82667,82672,82718,82740,82807,82903,82982,83017,83112,83272,83389,83401,83403,83478,83694,83718,83725,83856,83988,84043,84089,84451,84466,84577,84628,84652,84686,84694,85010,85058,85154,85286,85294,85406,85669,85845,85910,85970,86226,86258,86260,86364,86477,86492,86652,86658,86681,86797,86961,87001,87027,87060,87258,87331,87387,87407,87510,87576,87581,87737,87739,87794,87843,87852,88082,88464,88583,88602,88839,88947,88958,88962,89003,89040,89056,89167,89170,89307,89308,89469,89747,89805,89882,89894,90098,90134,90176,90248,90345,90372,90381,90436,90818,90854,90981,90993,91039,91227,91267,91330,91425,91442,91528,91574,91620,91677,91733,91764,91829,91859,91901,91970,92074,92117,92187,92272,92276,92278,92350,92516,92602,92604,92617,92806,92828,92856,92950,92994,93057,93105,93112,93118,93120,93166,93270,93406,93476,93678,93724,93752,93760,93884,93902,94081,94166,94245,94253,94303,94307,94377,94407,94472,94487,94504,94543,94646,94759,94821,94921,94925,94937,94983,95174,95183,95256,95374,95459,95519,95618,95695,95733,95805,95821,95876,96015,96102,96128,96405,96420,96422,96656,96728,96791,96806,96850,96865,96889,96982,97459,97556,97747,97932,97975,97981,97986,98017,98028,98062,98108,98142,98165,98195,98249,98252,98379,98446,98475,98511,98744,98766,98897,99199,99239,99250,99265,99267,99370,99423,99466,99557,99568,99572,99575,99643,99775,99942,99991,99997",
          "126,162,167,1647,1663,1986,2078,2100,2172,2484,2899,2986,3171,3209,3259,3645,3664,3782,3811,4038,4057,4093,4111,4184,4202,4274,4350,4353,4362,4395,4718,4923,4983,6199,6324,6916,8370,9472,9524,9548,9567,9586,9605,9624,9643,9662,9681,9757,9774,9795,9985,10042,10080,10673,10688,10699,10707,10742,10764,10968,11011,11220,11258,11372,11391,11458,11486,11505,11549,11552,11562,11638,11784,11847,11866,11885,11910,11933,11942,11961,11965,11999,12006,12037,12085,12113,12148,12151,12170,12198,12208,12214,12227,12246,12265,12284,12398,12405,12417,12476,12493,12512,12550,12553,12569,12607,12626,12664,12740,12757,12778,12854,12868,12873,12886,12940,12944,12949,12968,12987,13025,13062,13101,13106,13139,13177,13196,13233,13234,13253,13386,13443,13474,13553,13576,13614,13652,13671,13690,13728,13766,13785,13823,13895,13899,13979,13982,14127,14146,14184,14203,14222,14241,14260,14279,14298,14317,14355,14364,14374,14428,14507,14526,14602,14697,14811,14944,15024,15035,15077,15096,15103,15119,15191,15229,15267,15305,15362,15400,15419,15437,15457,15461,15476,15495,15514,15564,15571,15609,15628,15685,15704,15723,15742,15745,15762,15780,15837,15875,15894,15951,15970,15989,16008,16010,16033,16065,16083,16103,16116,16122,16155,16160,16198,16217,16236,16255,16274,16312,16331,16350,16369,16407,16426,16464,16517,16521,16556,16578,16597,16600,16673,16692,16712,16749,16768,16787,16806,16879,16901,16907,16939,16951,16959,17015,17053,17129,17167,17186,17224,17281,17300,17338,17342,17396,17398,17403,17414,17509,17528,17540,17604,17651,17670,17680,17682,17693,17718,17736,17737,17756,17830,17832,17851,17870,17889,17908,17932,17935,17946,17965,18003,18010,18060,18073,18078,18083,18117,18136,18193,18199,18212,18231,18253,18288,18307,18345,18364,18383,18402,18421,18434,18459,18478,18516,18529,18544,18548,18554,18573,18595,18649,18668,18687,18744,18763,18801,18820,18832,18858,18863,18896,18906,18934,18948,18972,18991,19008,19010,19013,19026,19048,19067,19078,19086,19105,19143,19144,19169,19181,19219,19257,19314,19358,19371,19409,19466,19485,19504,19542,19633,19637,19675,19694,19789,19808,19827,19864,19884,19903,19922,19960,19969,19979,20017,20036,20055,20093,20131,20169,20207,20243,20245,20283,20321,20340,20353,20378,20416,20435,20454,20473,20492,20511,20530,20568,20606,20610,20625,20701,20720,20739,20834,20853,20891,20910,20948,20967,21005,21043,21062,21068,21102,21119,21148,21156,21157,21176,21196,21214,21233,21252,21271,21322,21328,21347,21385,21404,21423,21442,21461,21480,21537,21572,21575,21651,21669,21670,21689,21708,21746,21803,21822,21841,21860,21879,21898,21917,21936,21955,21974,21993,22012,22031,22063,22145,22183,22202,22230,22259,22278,22292,22316,22335,22345,22354,22373,22392,22411,22430,22487,22506,22544,22550,22563,22582,22614,22620,22677,22734,22772,22791,22848,22867,22880,22924,22929,22941,22962,22981,22992,23000,23057,23061,23095,23114,23152,23167,23171,23190,23228,23285,23342,23361,23370,23380,23418,23475,23494,23532,23551,23570,23608,23684,23691,23703,23741,23760,23764,23772,23798,23817,23855,23860,23874,23893,23931,23950,23969,23988,24007,24026,24045,24051,24083,24099,24102,24140,24191,24193,24197,24216,24234,24235,24253,24254,24261,24266,24273,24292,24311,24330,24368,24387,24406,24425,24463,24482,24501,24539,24558,24596,24615,24653,24672,24691,24729,24767,24786,24802,24843,24862,24919,24938,24957,24995,25017,25033,25071,25090,25109,25128,25166,25204,25242,25261,25299,25327,25353,25375,25394,25413,25432,25451,25470,25527,25545,25584,25603,25607,25622,25660,25698,25736,25755,25774,25850,25888,25907,25926,25964,26002,26003,26021,26078,26097,26116,26154,26192,26230,26249,26268,26306,26325,26364,26382,26398,26401,26420,26439,26496,26515,26637,26672,26692,26724,26795,26858,26914,26971,26990,27038,27066,27085,27104,27123,27135,27172,27237,27274,27351,27360,27368,27384,27423,27435,27485,27579,27728,27789,27826,27852,27897,27933,27959,27970,28016,28054,28073,28130,28149,28243,28263,28278,28320,28358,28377,28392,28396,28415,28468,28567,28624,28643,28777,28795,28833,28852,28909,28947,28985,29037,29118,29174,29194,29214,29232,29270,29437,29460,29517,29552,29594,29612,29631,29633,29650,29745,29783,29795,29935,29937,30010,30039,30068,30097,30126,30155,30184,30213,30242,30271,30300,30322,30329,30358,30387,30416,30445,30474,30503,30532,30561,30619,30648,30677,30706,30735,30764,30793,30822,30851,30880,30909,30938,30967,30996,31025,31054,31083,31112,31141,31170,31199,31228,31257,31286,31315,31344,31373,31402,31431,31460,31489,31518,31547,31576,31605,31634,31663,31692,31721,31750,31779,31808,31837,31866,31895,31924,31953,31982,32011,32040,32069,32098,32127,32156,32185,32214,32243,32272,32301,32330,32359,32388,32417,32446,32475,32504,32533,32562,32591,32620,32649,32678,32707,32736,32765,32794,32823,32852,32881,32910,32939,32968,32997,33026,33055,33084,33113,33142,33171,33200,33229,33258,33287,33316,33345,33374,33403,33432,33461,33490,33519,33548,33577,33606,33635,33664,33693,33722,33751,33780,33809,33838,33867,33896,33925,33954,33983,34012,34041,34070,34099,34128,34157,34186,34215,34244,34273,34302,34331,34360,34389,34418,34447,34476,34505,34534,34563,34592,34621,34650,34679,34708,34737,34766,34795,34824,34853,34882,34911,34940,34969,34998,35010,35035,35057,35072,35084,35089,35099,35133,35173,35219,35220,35244,35262,35304,35336,35410,35437,35481,35529,35588,35615,35647,35672,35720,35779,35838,35859,35909,35935,35968,36027,36053,36090,36150,36187,36209,36253,36270,36280,36339,36398,36430,36470,36505,36529,36588,36625,36652,36711,36746,36770,36841,36874,36877,36900,36959,37031,37090,37149,37213,37272,37331,37361,37402,37424,37461,37520,37592,37651,37710,37774,37833,37860,37892,37963,38022,38081,38153,38178,38212,38271,38294,38318,38335,38359,38394,38414,38453,38499,38524,38546,38583,38642,38714,38741,38773,38832,38857,38896,38955,38979,39014,39085,39144,39181,39182,39203,39275,39334,39393,39413,39416,39457,39516,39538,39575,39646,39685,39705,39728,39764,39790,39796,39836,39895,39954,40121,40150,40179,40208,40237,40266,40295,40324,40353,40382,40411,40440,40469,40498,40527,40556,40585,40614,40643,40672,40701,40730,40759,40788,40817,40846,40875,40904,40933,40962,40991,41020,41049,41078,41107,41136,41165,41194,41223,41252,41281,41310,41339,41368,41397,41426,41455,41484,41513,41542,42012,42041,42070,42099,42128,42147,42188,42213,42220,42278,42305,42353,42388,42428,42515,42548,42568,42590,42615,42619,42665,42688,42696,42752,42797,42827,42839,42902,42912,42917,42951,42970,42989,43003,43010,43064,43096,43139,43178,43192,43226,43252,43301,43314,43376,43463,43509,43511,43538,43592,43613,43700,43775,43794,43850,43900,43937,44012,44033,44052,44055,44087,44113,44141,44174,44249,44324,44361,44375,44411,44431,44486,44529,44561,44648,44683,44723,44765,44798,44860,44885,44960,44972,45035,45108,45136,45156,45217,45239,45275,45290,45298,45348,45405,45463,45489,45535,45565,45567,45592,45627,45650,45659,45667,45722,45779,45837,45909,45966,45979,46012,46013,46021,46036,46040,46070,46103,46134,46139,46140,46145,46215,46236,46255,46260,46297,46353,46360,46368,46373,46413,46425,46431,46461,46478,46498,46504,46510,46513,46552,46557,46593,46596,46608,46630,46669,46726,46782,46800,46807,46829,46889,46892,46899,46907,46927,46944,46950,46997,47012,47028,47038,47042,47081,47082,47102,47104,47120,47187,47203,47225,47237,47247,47279,47313,47318,47334,47344,47365,47373,47385,47412,47431,47447,47449,47490,47584,47596,47640,47658,47673,47746,47757,47760,47786,47800,47806,47852,47858,47905,47922,47947,47950,47953,48100,48129,48210,48275,48291,48299,48323,48332,48366,48388,48407,48448,48450,48522,48538,48553,48555,48620,48661,48708,48728,48743,48838,48940,48962,48973,48974,48992,49031,49047,49099,49149,49201,49210,49295,49360,49378,49381,49401,49426,49445,49464,49490,49581,49587,49613,49641,49684,49794,49812,49893,49910,49952,50061,50126,50604,50660,50746,50840,50852,51021,51034,51066,51123,51141,51152,51197,51200,51202,51276,51462,51618,51706,51768,51813,51859,52030,52031,52032,52167,52278,52353,52530,52559,52728,52777,52838,52869,52897,52971,53077,53154,53158,53267,53373,53374,53517,53526,53533,53618,53626,53641,53736,53744,53771,53879,54122,54285,54287,54305,54313,54334,54338,54443,54472,54474,54827,54897,55167,55287,55298,55336,55338,55355,55592,55617,55649,55877,55981,56151,56480,56493,56502,56563,56644,56649,56736,56828,56861,56879,56904,56916,56943,56977,57439,57450,57642,57649,57662,57671,57718,57768,57823,57926,58031,58056,58083,58105,58113,58178,58213,58238,58287,58375,58396,58402,58425,58548,58550,58617,58719,58731,58745,58746,58899,58945,59037,59268,59324,59490,59497,59585,59612,59620,59626,59687,59767,59789,59888,60212,60249,60269,60339,60420,60523,60640,60682,60741,60857,60863,60986,61114,61242,61249,61288,61569,61576,61726,61727,61737,61739,61770,61785,61822,61859,61893,61959,61977,62029,62157,62200,62257,62308,62418,62455,62563,62697,62779,62871,63085,63223,63235,63269,63276,63451,63460,63467,63470,63526,63549,63576,63605,63656,63970,63990,64004,64047,64113,64395,64443,64745,64776,64777,65026,65090,65091,65165,65267,65279,65327,65352,65467,65468,65698,65703,65852,65876,66025,66126,66153,66164,66286,66360,66389,66405,66570,66598,66749,66770,66860,66943,66947,66971,67344,67382,67439,67498,67501,67508,67625,67661,67759,67789,67811,67846,67879,67916,68078,68088,68172,68184,68201,68263,68266,68287,68412,68419,68474,68531,68556,68561,68676,68730,68791,68812,68887,68901,68907,68941,69014,69023,69120,69261,69280,69358,69450,69487,69512,69702,69853,69866,69920,70198,70205,70224,70506,70691,70705,70710,70830,70934,71052,71238,71244,71347,71390,71492,71609,71819,71847,71848,71856,72017,72049,72054,72091,72256,72376,72396,72456,72512,72666,72675,72988,73047,73096,73111,73198,73297,73376,73416,73474,73653,73660,73852,73906,74067,74199,74390,74548,75131,75363,75409,75462,75501,75571,75574,75581,75592,75626,75732,75736,75800,75818,75830,75845,75848,75989,76045,76159,76206,76271,76702,76835,77072,77303,77499,77564,77672,77689,77722,77927,77956,78057,78191,78223,78234,78247,78282,78492,78550,78632,78674,78700,78843,78967,79177,79347,79389,79541,79579,79735,79830,79861,79887,79939,80275,80336,80368,80438,80451,80718,80813,80883,80960,80972,81018,81122,81201,81315,81326,81437,81455,81537,81593,81629,81648,81840,81847,82099,82107,82222,82240,82528,82719,82838,82848,82972,83048,83055,83086,83128,83231,83265,83500,83528,83647,83657,83749,83755,83902,83931,83940,84112,84114,84183,84220,84225,84274,84377,84433,84478,84637,84726,85002,85135,85251,85377,85382,85442,85524,85803,85833,86087,86173,86251,86360,86363,86437,86495,86507,86687,86715,87012,87040,87270,87340,87344,87556,87624,87627,87641,87838,87911,87963,87980,88091,88130,88172,88377,88379,88384,88702,88717,88747,88892,88967,89008,89011,89134,89323,89405,89462,89687,89692,89766,89796,89837,89957,90024,90131,90132,90413,90485,90609,90617,90645,90663,90955,90975,91136,91166,91208,91225,91228,91339,91385,91390,91571,91614,91646,91678,91736,91886,91978,91994,92027,92037,92077,92186,92354,92409,92459,92474,92527,93126,93243,93348,93494,93550,93649,93854,93881,93949,93961,94040,94108,94213,94262,94302,94315,94431,94460,94508,94512,94594,94599,95046,95170,95198,95261,95345,95460,95538,95608,95764,95806,95840,95853,95955,95982,95998,95999,96030,96184,96290,96403,96532,96558,96837,96989,97135,97264,97318,97325,97363,97424,97436,97453,97467,97743,97800,97804,97839,97892,97903,97952,98014,98022,98079,98088,98378,98441,98544,98548,98671,98769,98854,98894,98910,98912,98939,98940,99017,99038,99082,99084,99134,99140,99167,99171,99204,99255,99381,99672,99787,99833,99883,99904,99925",
          "18,51,55,99,121,231,272,565,836,1165,1664,1703,1880,1900,1902,1920,1945,1951,1965,1971,2003,2014,2015,2038,2049,2066,2118,2146,2203,2223,2311,2327,2346,2365,2371,2442,2454,2467,2468,2470,2481,2488,2523,2537,2541,2600,2603,2610,2623,2669,2724,2734,2743,2849,2871,2895,2955,2969,2970,2972,2975,2990,2999,3000,3003,3030,3036,3056,3058,3061,3071,3133,3142,3144,3169,3201,3203,3231,3257,3262,3265,3272,3279,3313,3337,3354,3387,3420,3448,3450,3457,3467,3475,3480,3482,3512,3553,3573,3627,3643,3675,3699,3709,3803,3854,3889,3895,3902,3910,3931,3935,3945,3963,3971,3976,3977,3980,3991,4003,4006,4009,4016,4022,4030,4033,4035,4053,4060,4062,4074,4075,4086,4117,4126,4129,4147,4154,4244,4245,4285,4310,4327,4337,4394,4407,4413,4435,4453,4492,4551,4589,4635,4677,4701,4766,4807,4860,4872,4879,4914,4930,4935,4944,4957,4958,4959,4965,4967,4973,5036,5112,5256,5619,5649,5980,6027,6077,6124,6129,6156,6321,6338,6345,6364,6367,6529,6582,6668,6688,6750,6846,6855,6864,6883,7032,7040,7146,7154,7292,7352,7444,7552,7564,7585,7599,7623,7670,7680,7733,7758,7795,7804,7833,7834,7912,7925,7936,7983,8078,8123,8160,8205,8206,8230,8242,8256,8340,8365,8437,8464,8543,9011,9086,9155,9162,9276,9314,9414,9491,9654,9655,9723,9787,9909,10098,10186,10219,10324,10373,10541,10555,10609,10706,10726,10846,10865,10870,10882,10926,11068,11071,11140,11246,11602,11761,11780,11972,12083,12091,12142,12172,12210,12378,12482,12638,12678,12692,12759,12981,13128,13461,13543,13792,14012,14175,14202,14257,14419,14455,14578,14648,14792,14919,14984,15055,15216,15266,15338,15438,15568,15617,15739,15812,15877,15880,15901,15912,16079,16162,16302,16382,16444,16624,16629,16826,17060,17149,17159,17161,17174,17374,17508,17542,17992,18044,18246,18333,18336,18497,18741,19003,19404,19421,19423,19577,19606,19674,19718,19849,19897,20013,20060,20453,20474,20724,20920,21053,21067,21297,21400,21401,21608,21652,21694,21727,21792,21819,21916,22003,22021,22155,22308,22356,22766,22887,23363,23444,23526,23595,23627,23823,23851,23885,23942,23982,23994,24075,24321,24386,24392,24655,24792,24933,25010,25084,25141,25147,25229,25235,25294,25434,25480,25589,25979,25995,26134,26266,26282,26480,26540,26542,26557,26559,26580,26603,26615,26681,26769,26813,26815,26929,26989,27075,27084,27211,27259,27326,27422,27426,27481,27506,27510,27526,27528,27529,27535,27570,27638,27936,27974,27989,28001,28009,28084,28086,28124,28132,28144,28156,28264,28418,28430,28480,28544,28576,28580,28640,28651,28661,28750,29258,29338,29438,29494,29532,29560,29577,29664,29676,29723,29806,29849,29859,29972,30016,30045,30074,30103,30132,30161,30190,30219,30248,30277,30306,30335,30364,30393,30422,30451,30480,30509,30538,30567,30625,30654,30683,30712,30741,30770,30799,30857,30886,30915,30944,30973,31002,31031,31060,31089,31118,31147,31176,31205,31234,31263,31292,31321,31350,31379,31408,31437,31466,31495,31524,31553,31582,31611,31640,31669,31698,31727,31756,31785,31814,31843,31872,31901,31930,31959,31988,32017,32046,32075,32104,32133,32162,32191,32220,32249,32278,32307,32336,32365,32394,32423,32452,32481,32510,32539,32568,32597,32626,32655,32684,32713,32742,32771,32800,32829,32858,32887,32916,32945,32974,33003,33032,33061,33090,33119,33148,33177,33206,33235,33264,33293,33322,33351,33380,33409,33438,33467,33496,33554,33583,33612,33641,33670,33699,33728,33786,33815,33844,33873,33902,33931,33960,33989,34018,34047,34076,34105,34134,34163,34192,34250,34279,34308,34337,34366,34395,34424,34453,34482,34511,34540,34569,34598,34627,34656,34685,34714,34743,34772,34801,34830,34859,34888,34917,34946,34975,35066,35105,35144,35179,35195,35268,35309,35342,35356,35367,35389,35416,35431,35461,35487,35507,35535,35566,35594,35653,35726,35785,35810,35884,35915,35933,35974,35991,36033,36096,36115,36156,36215,36235,36244,36249,36286,36345,36370,36388,36404,36476,36535,36594,36658,36717,36776,36847,36861,36906,36965,37037,37055,37096,37155,37219,37278,37337,37408,37467,37503,37510,37526,37598,37641,37657,37716,37750,37780,37839,37898,37912,37915,37969,37989,38028,38087,38159,38218,38248,38277,38306,38341,38400,38438,38459,38476,38530,38589,38623,38720,38749,38779,38838,38879,38902,38961,39020,39049,39091,39150,39209,39225,39281,39340,39399,39417,39463,39481,39497,39522,39581,39621,39652,39683,39686,39711,39770,39842,39901,39960,39976,40127,40156,40185,40214,40243,40272,40301,40330,40359,40388,40417,40446,40475,40504,40533,40562,40591,40620,40649,40678,40707,40736,40765,40794,40823,40852,40881,40910,40939,40968,40997,41026,41055,41084,41113,41142,41171,41200,41229,41258,41287,41316,41345,41374,41403,41432,41461,41490,41519,41548,42018,42047,42076,42105,42134,42194,42217,42284,42313,42359,42380,42434,42441,42474,42486,42521,42533,42596,42642,42671,42675,42700,42758,42766,42769,42773,42777,42833,42878,42908,42945,42949,42961,42995,43027,43070,43073,43087,43097,43145,43148,43185,43203,43205,43232,43258,43260,43307,43322,43382,43395,43419,43429,43469,43544,43594,43619,43626,43645,43660,43706,43754,43781,43797,43856,43863,43882,43893,43943,44018,44027,44093,44180,44209,44255,44278,44282,44330,44390,44391,44417,44426,44436,44444,44492,44567,44604,44610,44622,44654,44673,44729,44751,44763,44804,44815,44891,44902,44966,44992,44997,45001,45060,45072,45088,45114,45162,45180,45181,45223,45229,45281,45354,45375,45411,45469,45541,45598,45656,45728,45785,45803,45843,45851,45886,45915,45938,45972,46026,46078,46080,46106,46116,46143,46149,46195,46203,46209,46233,46250,46253,46266,46329,46342,46354,46456,46499,46512,46516,46559,46595,46599,46624,46638,46646,46648,46653,46698,46738,46777,46791,46823,46879,46952,47013,47031,47034,47039,47047,47057,47061,47127,47129,47134,47144,47163,47167,47219,47221,47222,47264,47271,47302,47303,47307,47310,47314,47361,47374,47386,47392,47428,47484,47579,47627,47634,47649,47655,47708,47715,47736,47782,47812,47834,47849,47873,47911,47938,47948,47970,47997,48017,48029,48050,48053,48132,48158,48239,48241,48246,48260,48289,48309,48360,48377,48395,48465,48501,48540,48577,48613,48637,48645,48669,48677,48710,48737,48771,48779,48784,48813,48816,48828,48850,48875,48898,48904,48916,48922,48972,49005,49009,49011,49035,49037,49050,49080,49085,49106,49113,49133,49143,49152,49170,49222,49231,49243,49259,49288,49297,49324,49334,49339,49364,49393,49404,49420,49434,49456,49549,49593,49686,49728,49744,49770,49795,49808,49850,49877,49900,49904,49914,49937,49961,49998,50065,50355,50423,50516,50541,50767,50884,50968,50970,50986,51046,51128,51130,51157,51180,51189,51322,51342,51458,51481,51560,51638,51667,51700,51717,51721,51970,52016,52048,52066,52137,52140,52145,52321,52481,52597,52613,52689,52814,53360,53383,53387,53393,53431,53530,53599,53834,53857,53990,54005,54119,54254,54297,54300,54359,54404,54608,54624,54689,54770,54800,54858,54879,55170,55194,55230,55424,55470,55492,55507,55508,55514,55614,55630,55642,55662,55677,55680,55694,55712,55747,55773,55840,55910,56000,56118,56138,56159,56445,56561,56700,56762,56779,56782,57031,57133,57185,57253,57260,57374,57409,57522,57656,57700,57887,57901,57927,58013,58072,58124,58170,58299,58334,58335,58384,58390,58456,58480,58587,58707,58708,58712,58744,58795,58836,58972,58988,59039,59194,59396,59406,59445,59503,59540,59741,59747,59900,60007,60034,60065,60205,60348,60358,60380,60381,60473,60517,60773,60835,61037,61103,61131,61135,61285,61302,61327,61441,61505,61517,61600,61690,61888,61967,62042,62081,62252,62319,62405,62456,62457,62461,62494,62590,62598,62720,62777,62798,62803,62943,62957,63022,63047,63290,63342,63361,63463,63522,63574,63690,63880,63894,63909,63923,63942,64089,64309,64440,64515,64516,64539,64668,64680,64725,64820,64851,64852,64864,64874,64899,64914,64986,64988,65010,65037,65082,65098,65099,65291,65338,65537,65616,65685,65768,65904,66087,66160,66206,66339,66340,66420,66459,66558,66712,66743,66746,66828,66871,66872,66905,67018,67020,67191,67196,67244,67383,67532,67599,67762,67778,67856,68028,68085,68117,68119,68196,68198,68242,68344,68405,68507,68558,68631,68690,68822,68899,68947,68982,69009,69048,69150,69376,69437,69622,69854,69922,70062,70129,70231,70250,70391,70502,70542,70632,71003,71020,71095,71117,71145,71409,71512,71687,71725,71757,71849,71892,71955,71980,72051,72102,72111,72153,72170,72295,72455,72481,72724,72759,72778,72900,72919,72926,73021,73077,73095,73103,73129,73228,73331,73368,73634,73689,74109,74152,74159,74162,74432,74744,74760,74818,74976,74981,74997,75030,75038,75111,75202,75249,75424,75540,75619,75625,75672,75744,75747,75769,75880,75922,75959,76155,76234,76284,76304,76375,76391,76422,76670,76675,76698,76795,76819,76883,76918,76969,76978,77010,77122,77124,77155,77284,77295,77408,77497,77719,77797,77960,78203,78283,78345,78482,78520,78636,78717,78742,78849,78960,78966,78978,79077,79126,79356,79496,79794,79815,79829,79982,80031,80064,80163,80237,80372,80486,80504,80562,80581,80628,80672,80731,80757,81008,81017,81050,81116,81128,81137,81221,81289,81490,81614,81675,81887,81919,81920,81989,82017,82088,82116,82266,82450,82539,82677,82835,82947,82970,83045,83165,83226,83259,83398,83460,83515,83608,83621,83835,83895,83921,84139,84293,84341,84359,84428,84496,84538,84899,84976,85005,85117,85124,85175,85449,85561,85606,85626,85645,85653,85658,85694,85737,85769,85788,85813,85943,85947,85982,85984,86035,86051,86082,86196,86210,86389,86431,86446,86459,86560,86760,86769,86833,86898,87113,87294,87618,87651,87740,87841,87885,87935,88018,88055,88103,88259,88329,88471,88493,88512,88531,88540,88640,88794,88838,88984,89113,89192,89211,89410,89495,89527,89557,89563,89565,89892,89971,90123,90238,90338,90351,90353,90405,90728,90737,90766,90930,91075,91083,91154,91207,91220,91355,91491,91588,91627,91691,91702,91762,91937,92076,92099,92416,92437,92446,92723,92743,92805,92855,93041,93042,93065,93128,93276,93575,93601,93684,93794,93806,93931,94192,94237,94277,94341,94406,94493,94523,94861,94916,94923,94926,94959,95034,95155,95165,95166,95272,95298,95300,95401,95435,95568,95727,95946,96049,96096,96106,96150,96161,96217,96340,96375,96435,96523,96561,96653,96663,96714,96758,97081,97106,97388,97430,97449,97625,97748,97769,97818,97902,98033,98077,98098,98126,98128,98140,98150,98154,98226,98322,98532,98609,98613,98630,98770,99043,99087,99110,99165,99198,99320,99401,99508,99552,99555,99620,99679,99686,99689,99764,99870,99886,99894,99952",
          "47,66,267,321,335,392,431,452,459,462,475,511,551,646,878,946,1075,1086,1140,1187,1391,1427,1612,1630,1642,1710,2656,2755,3330,4609,4976,5041,5162,5491,5509,5525,5608,5622,5666,5692,5695,5858,6072,6125,6256,6263,6295,6302,6469,6610,6676,6856,6928,7045,7236,7334,7377,7397,7550,7694,7703,7775,7939,7993,8111,8144,8153,8192,8359,8373,8524,8636,8695,8699,8702,8803,8847,8867,8933,8947,8974,9007,9128,9168,9169,9213,9219,9281,9353,9467,9853,9886,10023,10045,10084,10135,10286,10399,10406,10457,10492,10547,10584,10611,10635,10693,10873,10892,11007,11012,11045,11069,11259,11302,11330,11733,11740,11808,11813,11934,12094,12119,12178,12407,12720,12799,12927,12980,13051,13052,13086,13092,13141,13414,13508,13537,13618,13748,14088,14370,14633,14778,14795,14936,15019,15027,15157,15162,15180,15324,15435,15627,15909,16018,16539,16567,16579,16595,16659,17018,17246,17663,17733,17813,17952,18049,18094,18232,18627,18725,18772,18783,18895,18921,19153,19353,19363,19732,19926,20076,20089,20722,20796,20900,21036,21088,21277,21283,21405,21697,22304,23006,23234,23264,23387,23456,23479,23585,23779,23836,24145,24304,24349,24926,24951,25136,25174,25489,25515,26210,26327,26396,26458,26555,26668,26884,26941,26943,27013,27030,27385,27781,28172,28353,28374,28423,28837,28865,28894,28996,29190,29452,29492,29600,29700,29848,30864,30946,32253,33525,36431,36434,36635,36681,37018,37237,37319,37880,37994,38302,38552,38609,38614,38815,38882,39114,39300,39382,39690,39880,42524,44140,44986,45243,45444,45499,45601,45603,50168,50183,50206,50237,50246,50330,50542,50759,50836,50843,50937,50990,51001,51102,51350,51501,51529,51565,51654,51686,51699,51758,51759,51819,52090,52344,52348,52523,52547,52570,52625,52707,52760,52797,52841,52895,52927,52999,53063,53115,53120,53142,53309,53425,53510,53521,53568,53766,53826,53841,53897,53931,53936,53958,53978,54033,54065,54097,54184,54550,54591,54605,54717,54760,54801,55126,55129,55187,55208,55289,55371,55373,55416,55452,55610,55661,55878,55950,56081,56224,56231,56273,56368,56393,56503,56504,56691,56699,56704,56815,56937,56972,57397,57485,57504,57644,57648,57678,57800,58055,58073,58149,58242,58362,58545,58634,58665,58689,59022,59178,59331,59512,59514,59591,59638,59755,59827,59831,59930,60005,60019,60028,60074,60292,60480,60481,60545,60659,60668,60727,60800,60887,61110,61120,61123,61304,61325,61379,61420,61634,61957,62166,62373,62400,62458,62485,62552,62583,62790,62918,62967,62980,62991,63016,63054,63142,63173,63197,63243,63388,63394,63450,63480,63556,63582,63610,63625,63640,63752,63772,63795,63967,63974,63984,64005,64007,64050,64265,64300,64407,64410,64427,64463,64617,64660,64678,64710,64750,64796,64875,64934,64947,65060,65172,65236,65260,65375,65543,65792,66086,66114,66221,66284,66290,66328,66397,66462,66509,66620,66672,66716,66795,66810,67072,67145,67219,67273,67318,67364,67401,67481,67555,67569,67596,67688,67742,67973,68224,68254,68342,68382,68443,68643,68802,68902,68997,69126,69230,69232,69309,69636,69826,69918,69931,69964,70584,70606,70829,70842,70979,71041,71269,71407,71468,71553,71562,71651,71744,71929,72194,72218,72357,72385,72461,72617,72709,72813,72863,73140,73150,73263,73279,73365,73450,73506,73509,73538,73597,73640,73670,73739,74030,74041,74126,74151,74231,74260,74395,74402,74520,74565,74591,74720,74919,74973,75055,75165,75218,75246,75290,75322,75598,76092,76139,76360,76563,76652,76663,76674,76734,76782,76796,76944,76988,76993,77107,77244,77339,77510,77512,77574,77666,77721,77771,77799,77816,77860,77876,78000,78043,78044,78045,78164,78173,78235,78379,78452,78517,78533,78544,78575,78691,78716,78793,78924,78935,79088,79296,79349,79476,79558,79732,79789,79832,79846,79905,79937,79965,80038,80172,80423,80685,80754,80756,80772,80839,80855,80977,81063,81367,81511,81692,81704,81774,81811,82000,82175,82193,82206,82251,82380,82397,82400,82529,82552,82554,82686,82729,82749,82797,83078,83149,83352,83414,83445,83454,83471,83486,83681,83750,83811,83868,83989,84107,84363,84370,84381,84592,84647,84727,84756,84759,84800,84879,85150,85165,85253,85310,85360,85368,85408,85576,85643,85860,86121,86127,86354,86992,87032,87116,87145,87164,87319,87447,87476,87555,87699,87742,87756,87780,87793,87878,88040,88111,88229,88238,88315,88326,88341,88397,88466,88472,88478,88745,88806,88953,89241,89304,89409,89459,89669,89686,89702,89734,89818,89867,89901,89919,89928,89989,90039,90202,90253,90270,90396,90521,90594,90600,90707,90842,90843,91007,91021,91077,91188,91261,91350,91368,91457,91738,91807,91858,91932,92219,92428,92435,92450,92451,92479,92526,92720,92739,92887,93077,93092,93202,93203,93235,93236,93283,93320,93372,93526,93596,93650,93660,93692,93780,93825,93862,93973,94074,94092,94179,94248,94436,94446,94570,94704,94763,94831,94848,94965,95164,95210,95270,95290,95379,95420,95481,95571,95706,95794,95835,96010,96169,96204,96310,96369,96459,96463,96537,96635,96767,96813,96820,96896,97080,97131,97312,97339,97374,97431,97582,97602,97722,97826,97880,98004,98010,98037,98114,98205,98223,98432,98610,98618,98775,98784,98827,99004,99039,99061,99139,99153,99164,99368,99378,99600,99687,99759,99916,99928,99962,99978",
          "14,29,89,111,116,150,215,233,310,479,905,961,992,1011,1079,1171,1202,1281,1441,1491,1502,1623,1693,1773,1854,1888,1927,1950,1959,1962,1998,2010,2027,2031,2134,2153,2177,2191,2197,2199,2205,2214,2286,2289,2292,2334,2366,2387,2395,2399,2433,2446,2455,2483,2525,2568,2577,2578,2622,2627,2648,2683,2700,2704,2785,2794,2802,2813,2842,2877,2907,2942,2947,2959,2961,2965,2971,2995,3023,3039,3068,3098,3121,3143,3154,3211,3220,3244,3252,3260,3327,3401,3402,3431,3443,3471,3478,3621,3625,3632,3660,3688,3737,3739,3756,3765,3770,3829,3901,3903,3928,3937,3948,3950,3953,4000,4031,4041,4047,4067,4135,4144,4156,4175,4206,4251,4289,4329,4340,4351,4357,4399,4416,4424,4445,4464,4532,4563,4567,4604,4643,4654,4679,4706,4773,4837,4840,4908,4929,4936,4941,4955,4971,4984,5030,5208,5248,5295,5336,5569,5621,5630,5644,5671,5699,5705,5737,5788,5856,5893,5944,5982,6016,6044,6079,6273,6460,6664,6677,6743,6744,6776,6778,6791,6857,6865,7048,7054,7214,7224,7317,7384,7489,7497,7504,7588,7674,7757,8012,8022,8058,8098,8115,8310,8559,8678,8704,8734,8736,8810,8836,8837,9018,9026,9081,9135,9163,9196,9233,9285,9300,9307,9339,9396,9456,9459,9485,9634,9766,9830,9849,9971,10104,10105,10166,10274,10391,10396,10431,10448,10637,10640,10646,10743,10772,10816,11182,11273,11315,11317,11332,11350,11376,11388,11404,11425,11483,11504,11695,11764,11981,12010,12138,12431,12629,12767,12942,12946,13006,13040,13135,13308,13757,13801,13848,14123,14188,14303,14363,14443,14474,14554,14558,14608,14974,15232,15440,15552,15693,15789,16058,16270,16454,16506,16551,16688,16783,16998,17127,17505,17849,17875,18075,18145,18205,18235,18268,18344,18637,18742,18769,18830,18839,19290,19646,19713,19723,19756,19770,19953,20289,20455,20987,21230,21538,21798,22017,22298,22588,22809,22858,23045,23161,23195,23223,23420,23680,23751,24009,24255,24850,24892,25453,25610,25656,25672,25742,26208,26308,26567,26569,26574,26605,26622,26670,26820,26849,26983,27078,27088,27200,27241,27388,27471,27509,27527,27550,27637,27659,27669,27687,27952,27988,27990,28002,28023,28044,28083,28125,28126,28137,28153,28168,28185,28242,28307,28457,28517,28541,28571,28585,28588,28605,28622,28631,28642,28646,28664,28772,28973,29180,29314,29318,29346,29357,29379,29397,29428,29539,29554,29614,29653,29811,29835,29915,30046,30075,30133,30220,30249,30278,30336,30365,30423,30481,30597,30626,30655,30684,30713,30742,30771,30800,30858,30887,30916,30974,31003,31061,31090,31119,31148,31177,31206,31235,31264,31351,31380,31409,31467,31474,31496,31525,31554,31583,31612,31641,31699,31728,31757,31786,31815,31822,31844,31902,31960,31989,32076,32105,32134,32163,32192,32221,32250,32279,32308,32337,32395,32453,32482,32511,32540,32656,32685,32743,32772,32801,32830,32888,32946,32975,33033,33091,33120,33178,33207,33236,33265,33294,33323,33352,33381,33410,33439,33468,33497,33526,33584,33642,33671,33700,33729,33758,33787,33816,33845,33874,33903,33932,33961,33990,34019,34048,34077,34106,34135,34164,34222,34251,34280,34309,34338,34367,34396,34425,34454,34541,34570,34599,34686,34715,34744,34802,34831,34889,34918,34947,34976,35049,35180,35206,35213,35234,35282,35291,35343,35394,35417,35453,35488,35536,35558,35595,35654,35686,35727,35786,35822,35845,35861,35975,36034,36097,36126,36135,36179,36230,36287,36346,36384,36405,36690,36802,36848,36885,36966,36983,36996,37038,37097,37156,37279,37355,37409,37468,37527,37557,37599,37717,37781,37806,37899,37937,37970,38012,38029,38088,38160,38219,38401,38460,38531,38556,38627,38649,38672,38721,38747,38839,38903,38989,39021,39059,39092,39151,39210,39239,39282,39341,39400,39432,39464,39523,39653,39771,39791,39806,39843,39863,39961,39980,39999,40108,40137,40166,40195,40282,40311,40340,40369,40456,40485,40514,40543,40572,40601,40630,40659,40688,40717,40746,40833,40891,40920,40949,40978,41007,41036,41065,41094,41123,41152,41181,41210,41217,41268,41384,41406,41413,41471,41529,42059,42088,42177,42267,42342,42383,42393,42451,42504,42579,42774,42805,42891,42964,42978,43053,43083,43086,43089,43128,43166,43215,43245,43255,43280,43290,43446,43527,43662,43679,43689,43732,43815,43839,43873,43926,44001,44040,44076,44118,44163,44224,44230,44238,44266,44299,44313,44355,44363,44370,44400,44427,44475,44507,44586,44659,44667,44712,44787,44901,44928,44949,44993,44999,45024,45145,45206,45225,45226,45394,45524,45581,45735,45768,45826,45898,46032,46061,46112,46166,46188,46270,46290,46324,46350,46418,46484,46525,46530,46555,46565,46598,46647,46649,46683,46691,46694,46716,46729,46761,46831,46852,46915,46959,46964,47074,47176,47191,47212,47406,47413,47515,47535,47575,47609,47623,47630,47718,47719,47785,47929,47993,48034,48036,48070,48085,48177,48182,48216,48257,48280,48298,48337,48449,48481,48482,48492,48526,48618,48634,48688,48724,48832,48895,48931,48941,48956,49105,49212,49275,49308,49412,49431,49473,49532,49563,49568,49572,49726,49748,49810,49819,49843,49878,49947,50210,50215,50358,50499,50548,50647,50664,50816,50835,50977,51012,51030,51101,51231,51352,51410,51637,51847,51876,51976,51980,52111,52159,52277,52431,52513,52656,52773,52963,53083,53163,53166,53211,53356,53479,53676,53723,53905,53929,54015,54132,54241,54298,54337,54346,54382,54469,54540,54681,54709,54724,54751,54882,54885,55150,55195,55200,55676,55853,55964,55971,56016,56024,56037,56146,56174,56260,56261,56278,56415,56468,56488,56496,56552,56555,56570,56593,56594,56601,56607,56609,56750,56759,56770,56915,57119,57179,57214,57224,57272,57312,57332,57359,57380,57601,57677,57807,58016,58071,58292,58421,58466,58534,58606,58610,58644,58807,58954,59050,59081,59103,59151,59173,59218,59230,59248,59288,59441,59486,59550,59570,59593,59623,59678,59707,59848,60020,60037,60228,60252,60297,60322,60353,60439,60553,60646,60676,60795,60833,60839,60864,61023,61279,61352,61708,61717,61736,61872,61906,61989,61996,62036,62163,62182,62193,62203,62215,62232,62263,62441,62491,62512,62540,62669,62672,62771,62867,62932,63078,63171,63259,63316,63380,63397,63438,63568,63631,63633,63730,63941,63971,63977,64028,64138,64158,64230,64267,64279,64402,64420,64428,64437,64447,64474,64519,64549,64550,64588,64670,64718,64950,65015,65089,65242,65286,65367,65425,65665,65870,66078,66186,66304,66369,66387,66409,66411,66419,66468,66769,66926,66940,67160,67413,67455,67515,67586,67601,67623,67745,67900,67934,68187,68199,68230,68239,68249,68282,68418,68476,68504,68686,68688,68807,68915,68962,69153,69188,69314,69366,69439,69584,69676,69727,69735,69736,69784,69817,69818,69836,69930,70005,70063,70119,70139,70254,70306,70319,70334,70345,70369,70488,70576,70663,70760,70848,70863,71014,71018,71033,71103,71304,71335,71418,71429,71435,71478,71521,71538,71570,71675,71728,71914,71966,71989,71996,72038,72107,72140,72188,72285,72318,72327,72399,72409,72572,72624,72636,72760,72818,72831,72858,72921,72942,72979,73058,73126,73378,73407,73426,73487,73586,73694,73745,73805,73823,73862,74021,74059,74096,74106,74234,74410,74425,74472,74660,74716,74723,74745,74833,75076,75112,75169,75187,75247,75260,75297,75438,75478,75681,75743,75847,76125,76158,76250,76513,76765,76812,76947,77078,77325,77429,77448,77462,77486,77578,77673,77817,78012,78082,78140,78488,78490,78554,78589,78684,78866,78971,79011,79131,79570,79640,79652,79690,79705,79725,79980,80075,80108,80152,80196,80300,80421,80433,80457,80499,80551,80593,80654,80681,80747,80794,80823,80888,80900,80918,81112,81210,81335,81344,81424,81558,81580,81631,81905,81908,81914,82106,82120,82151,82189,82404,82548,82696,82714,82757,82758,82795,82852,82944,82995,83080,83267,83313,83768,83786,83938,84105,84156,84206,84298,84318,84358,84417,84420,84449,84472,84599,84721,84870,84929,85075,85077,85222,85239,85254,85259,85305,85454,85591,85673,85928,85952,86000,86003,86075,86133,86137,86192,86247,86255,86295,86311,86382,86411,86508,86603,86630,86749,86831,86836,87049,87179,87188,87277,87284,87349,87353,87408,87427,87443,87519,87544,87615,87731,87763,87790,87796,87915,87947,87993,88148,88212,88287,88367,88369,88381,88695,88759,88760,89028,89313,89351,89599,89629,89685,89704,89750,89897,89955,90005,90164,90292,90397,90415,90452,90491,90494,90497,90528,90832,91065,91091,91171,91336,91391,91506,91765,91811,91963,92090,92154,92313,92438,92460,92503,92540,92564,92571,92608,92694,92708,92747,92852,92881,92889,92897,93295,93377,93549,93571,93674,93795,93935,93948,94019,94156,94247,94270,94286,94287,94319,94378,94387,94435,94516,94527,94589,94605,94632,94653,94688,94738,94881,94920,94954,94994,95008,95042,95101,95265,95315,95477,95479,95507,95511,95551,95565,95579,95595,95623,95636,95751,95753,95819,95920,95932,95953,96032,96185,96235,96288,96296,96334,96341,96379,96431,96464,96500,96770,96980,97032,97130,97384,97772,97774,97913,97944,98052,98072,98121,98216,98251,98274,98298,98357,98367,98380,98477,98658,98723,98861,98918,98938,99010,99193,99395,99396,99465,99476,99616,99725,99963",
          "200,263,324,344,383,416,438,500,526,757,813,829,932,1040,1052,1082,1093,1163,1192,1229,1280,1296,1297,1314,1560,1606,1613,1804,1827,1889,1974,1988,2288,2299,2386,2493,2546,2606,2644,2661,2766,2831,2864,2940,2951,3005,3006,3322,3394,3408,3463,3623,3624,3740,3780,3787,3842,3863,3940,4048,4352,4728,4812,4852,4890,5025,5027,5125,5159,5164,5198,5287,5299,5350,5352,5373,5425,5629,5634,5722,5815,5841,5913,5965,5975,6126,6164,6188,6301,6439,6483,6497,6536,6763,6798,6935,7092,7197,7297,7333,7482,7604,7616,7618,7867,7887,7946,7978,8009,8082,8188,8360,8482,8518,8711,8725,8964,8971,9136,9242,9257,9286,9484,9492,9535,9576,9625,9951,9955,10036,10100,10200,10228,10244,10394,10441,10487,10493,10617,10718,10980,11042,11072,11125,11183,11424,11724,11949,12339,12424,12524,12597,12621,12645,12952,13076,13149,13320,13347,13378,13425,13501,13524,13659,13669,13929,13954,14080,14340,14375,14819,14907,14958,14991,15082,15409,15602,15706,15793,15861,15915,15930,16013,16134,16412,16427,16654,16916,17006,17102,17251,17337,17472,17530,17581,17696,17721,17969,18210,18802,18859,18878,18966,19098,19148,19245,19296,19437,19511,19755,19810,19850,19869,20202,20219,20220,20316,20657,20670,20892,20940,20959,21189,21218,22092,22122,22217,22532,22995,23104,23151,23179,23254,23270,23520,23699,23827,24078,24269,24489,24671,24855,24880,24939,25121,25163,25240,25279,25284,25322,25380,25436,25585,25673,25716,25811,25817,26173,26226,26293,26337,26444,26497,26633,26640,26833,26883,26905,26945,27133,27193,27268,27743,27785,27961,28161,28197,28225,28298,28408,28482,28521,28601,28615,28804,28950,29325,29590,29679,29823,29899,29954,30101,30114,30169,30188,30198,30507,30536,30652,30884,31029,32903,32943,33088,33391,33407,33581,33657,33802,35226,35312,35391,35592,35783,35842,35939,36008,36044,36315,36316,36454,36551,36917,36921,36936,37007,37072,37131,37137,37181,37242,37389,37465,37544,37619,37695,37877,37929,37991,38057,38240,38243,38252,38311,38416,38440,38568,38612,38766,38811,38866,39192,39318,39375,39500,39562,39616,39625,39629,39730,39921,39937,40270,40304,40480,40596,40647,41227,41261,42025,42111,42282,42706,42734,42863,43085,43123,43479,43786,44026,44344,44706,45177,45317,45687,45746,45798,45974,46714,46748,46871,47664,47843,47968,48012,48347,48621,48712,48878,49008,49272,49470,49820,50040,50056,50186,50278,50537,50679,50820,50885,50971,51177,51204,51230,51363,51520,51604,51616,51668,51714,51880,51881,52062,52071,52126,52142,52307,52428,52536,52550,52560,52571,52609,52631,52639,52682,52691,52709,52771,52785,52861,52879,53080,53192,53242,53475,53476,53487,53538,53540,53601,53609,53674,53675,53746,53756,53832,53867,53876,53956,54034,54104,54191,54279,54283,54351,54533,54572,54768,54775,54810,54846,54956,55031,55092,55217,55451,55504,55539,55541,55717,55906,56026,56028,56055,56082,56158,56173,56237,56325,56387,56417,56432,56527,56637,56689,56721,56831,56849,56982,57065,57195,57218,57242,57303,57526,57537,57625,57692,57808,57864,57876,57919,57922,57957,57984,58017,58125,58172,58324,58369,58436,58459,58465,58491,58516,58774,58820,59018,59051,59120,59246,59263,59328,59387,59404,59489,59494,59504,59533,59547,59554,59565,59617,59650,59684,59791,59830,59919,59938,59946,60075,60137,60139,60148,60277,60371,60405,60410,60630,60721,60790,60803,60844,61180,61184,61187,61206,61749,61948,62007,62018,62168,62183,62214,62228,62276,62303,62327,62352,62383,62393,62445,62529,62836,62920,62965,62971,62976,63283,63337,63462,63482,63709,63740,63827,63885,63922,63991,64046,64052,64219,64274,64433,64547,64812,64897,65031,65063,65133,65143,65350,65355,65378,65391,65604,65697,65745,65857,66178,66181,66576,66629,66687,66798,66864,66969,67009,67107,67167,67291,67294,67295,67483,67546,67584,67594,67683,67685,67782,67835,67913,67943,67974,67986,68025,68326,68350,68395,68472,68549,68678,68799,68865,69034,69037,69163,69192,69296,69398,69443,69985,70006,70024,70186,70239,70384,70501,70715,70793,70837,70858,70968,71046,71068,71278,71327,71593,71985,72025,72068,72098,72163,72223,72224,72490,72632,72667,72707,72723,72758,72795,72807,72808,72868,72964,73018,73026,73040,73059,73072,73147,73246,73260,73282,73309,73340,73601,73627,73686,73970,74060,74122,74177,74294,74301,74347,74366,74434,74636,74712,74731,74845,74999,75335,75640,75643,75692,75763,75788,75816,75930,76119,76216,76254,76294,76354,76365,76539,76671,76718,76735,76738,76740,76752,76887,76935,76972,77052,77093,77379,77536,77569,77668,77783,77935,77937,77984,78151,78153,78163,78194,78213,78243,78344,78400,78448,78476,78574,78602,78657,78739,78743,78784,78788,78846,78856,78962,78990,79079,79226,79265,79367,79473,79573,79734,79810,79966,79983,80092,80177,80254,80291,80311,80317,80364,80388,80520,80522,80582,80586,80605,80651,80734,80744,80939,80990,81034,81140,81157,81187,81314,81327,81532,81607,81628,81776,81842,81893,81898,81918,82047,82135,82269,82277,82467,82550,82569,82575,82591,82683,82765,82782,82928,82949,83079,83118,83192,83496,83635,84175,84266,84385,84389,84588,84745,84975,84977,84979,85042,85054,85055,85118,85145,85386,85440,85556,85605,85655,85696,85732,85908,85925,85956,86034,86077,86293,86317,86381,86598,86739,86851,86979,87022,87163,87292,87333,87397,87457,87468,87515,87562,87567,87606,87616,87639,87643,87710,87820,87824,88007,88127,88292,88368,88425,88494,88581,88681,88754,88771,88787,88834,88860,88945,89063,89087,89093,89131,89159,89232,89236,89284,89287,89471,89507,89521,89541,89572,89628,89657,89690,89763,89871,89898,89927,90183,90273,90416,90534,90644,90649,90700,90751,90896,90951,90991,91149,91280,91294,91332,91369,91445,91487,91751,92047,92241,92299,92342,92472,92495,92727,92817,92846,92961,93035,93060,93169,93311,93463,93930,94203,94396,94465,94468,94502,94554,94580,94671,94713,94737,94818,94866,94927,95025,95026,95035,95195,95304,95366,95440,95471,95473,95533,95655,95739,95766,95769,95878,95890,95921,95935,95958,96072,96076,96095,96125,96149,96156,96200,96224,96237,96300,96426,96430,96458,96640,96688,96692,96821,96967,97077,97162,97196,97356,97545,97635,97656,97682,97695,97793,97828,97850,97905,98021,98064,98078,98101,98118,98133,98143,98164,98292,98326,98502,98506,98893,98989,99141,99163,99203,99349,99530,99633,99768,99954,99977",
          "31,354,376,377,381,387,422,602,650,683,735,747,807,849,889,952,979,994,1004,1024,1048,1081,1112,1142,1194,1227,1232,1243,1263,1264,1371,1419,1504,1677,1772,1878,1912,3073,3692,5022,5037,5047,5069,5081,5099,5107,5111,5176,5315,5332,5394,5432,5459,5480,5486,5577,5582,5592,5595,5615,5675,5784,5798,5852,5868,5888,5910,5994,5998,6014,6069,6143,6210,6219,6231,6244,6270,6289,6335,6354,6372,6407,6543,6600,6604,6628,6649,6699,6779,7019,7026,7036,7051,7410,7417,7488,7625,7667,7729,7915,8016,8038,8063,8441,8445,8541,8618,8634,8641,8661,8789,8839,8899,8903,9008,9109,9113,9130,9139,9215,9225,9246,9289,9946,9947,10056,10227,10499,10625,10838,10935,11114,11768,11790,12158,12525,14214,14878,15716,16449,16984,17082,17709,18176,18848,19329,20111,20256,20576,21455,23035,23313,23483,24353,25077,25329,25354,25487,26465,26709,27221,27538,28900,29163,35896,36310,36432,36639,37638,38296,38703,39825,42556,43098,44473,45032,45173,46281,46476,50018,50062,50089,50136,50239,50359,50430,50447,50465,50488,50524,50603,50686,50880,50909,50945,51019,51156,51218,51249,51336,51371,51445,51581,51599,51660,51726,51810,51865,52037,52054,52100,52343,52450,52495,52629,52664,52784,53153,53200,53202,53372,53497,53501,53554,53621,53659,53730,53782,53871,53874,54062,54073,54123,54167,54358,54360,54366,54425,54430,54447,54488,54513,54546,54586,54752,54883,55063,55154,55159,55209,55250,55290,55350,55364,55383,55431,55577,55582,55678,55696,55784,55814,55969,55989,56115,56119,56386,56449,56508,56652,56912,56971,57057,57071,57099,57117,57254,57461,57561,57564,57646,57737,57830,57900,57937,57996,58058,58084,58264,58312,58322,58333,58386,58410,58441,58443,58452,58539,58585,58605,58873,58884,58922,59045,59129,59252,59330,59333,59416,59421,59506,59571,59829,59860,59911,60147,60203,60316,60390,60450,60515,60570,60642,60647,60657,60766,60783,60818,60826,60837,61020,61097,61162,61255,61369,61431,61464,61482,61542,61585,61679,62009,62073,62091,62159,62188,62218,62305,62398,62473,62566,62786,62915,62921,62927,63101,63104,63284,63285,63298,63321,63733,63774,63811,63817,63899,64030,64325,64576,64618,64649,64715,64735,64754,64767,64811,64824,64920,64990,65128,65161,65257,65401,65554,65746,65750,65790,65794,65797,65836,66033,66061,66093,66132,66156,66169,66222,66327,66421,66464,66553,66575,66615,66697,66733,66740,66771,66819,66833,66886,66891,66913,66996,67025,67080,67117,67166,67221,67436,67589,67591,67647,67726,67732,67752,67761,67887,67892,68034,68084,68217,68228,68528,68630,68636,68725,68766,68942,69003,69100,69136,69162,69211,69239,69316,69488,69546,69763,69824,69857,69859,69893,70043,70058,70116,70156,70183,70202,70394,70704,70752,70795,70871,70877,70942,70980,71064,71285,71334,71413,71565,71575,71595,72001,72075,72120,72131,72246,72319,72338,72356,72429,72476,72822,72851,72895,73213,73230,73351,73401,73436,73437,73516,73548,73676,73682,73702,73733,73795,73816,73827,74090,74105,74247,74352,74364,74406,74553,74566,74580,74650,74762,74796,74865,74907,74962,75014,75252,75266,75280,75286,75306,75375,75465,75623,75677,75737,75981,75994,76232,76303,76312,76412,76565,76589,76700,76855,76882,77117,77257,77424,77516,77565,77589,77619,77686,77772,77896,77933,77971,78093,78193,78199,78329,78368,78443,78499,78548,78593,78671,78825,78830,78915,78932,78997,79005,79035,79138,79142,79162,79231,79993,80035,80137,80209,80337,80377,80465,80470,80478,80653,80677,80824,80845,81037,81039,81166,81171,81348,81377,81438,81560,81646,81785,81793,81806,81860,81902,81954,81961,81975,81982,82136,82205,82226,82268,82522,82547,82568,82590,82726,82734,82779,82796,82864,82969,83239,83375,83390,83504,83560,83607,83695,83756,83796,83851,84204,84247,84394,84452,84474,84609,84728,85149,85365,85509,85547,85612,85708,85715,85779,85796,85880,85895,85924,86053,86114,86232,86319,86334,86342,86379,86417,86587,86634,86761,86862,86871,86915,86995,87310,87549,87734,87905,87931,87933,87946,87948,88027,88213,88354,88439,88559,88584,88625,88680,88880,89019,89068,89139,89151,89221,89247,89258,89281,89737,89889,89940,89964,89987,90032,90033,90111,90112,90154,90285,90454,90459,90614,90667,90893,90903,91104,91121,91201,91406,91533,91715,91722,92131,92169,92182,92322,92349,92362,92364,92453,92497,92690,92794,92915,93048,93073,93335,93419,93498,93612,93774,93792,93827,93975,93990,94046,94049,94227,94278,94301,94345,94438,94458,94494,94503,94578,94608,94640,94670,94709,94721,94729,95030,95326,95333,95593,95704,95717,95732,95910,96003,96135,96179,96221,96453,96489,96503,96573,96831,96842,96897,96963,96970,97022,97075,97128,97262,97447,97482,97534,97636,97647,97724,97734,97760,97792,97807,97919,97976,98035,98103,98131,98293,98305,98311,98416,98542,98562,98650,98668,98715,98718,98774,98819,98895,99056,99083,99098,99273,99376,99380,99394,99471,99581,99636,99663,99735,99778",
          "91,1365,1716,1717,1779,1873,1979,2048,2211,2240,2319,2368,2458,3029,3110,3243,3328,3442,3537,3723,3955,4050,4069,4419,4440,4489,4810,4873,5044,6145,6596,6615,6653,6963,6990,7128,7850,8053,8059,8135,8154,8314,8333,8553,8648,8762,8778,8917,9060,9117,9161,9180,9237,9308,9316,9406,9493,9503,9510,9522,9560,9598,9617,9693,9750,9788,9823,9845,9883,9978,9997,10035,10111,10130,10149,10168,10187,10191,10206,10225,10292,10339,10605,10738,10833,10890,10893,10909,10928,10966,11023,11137,11156,11172,11175,11213,11251,11265,11296,11308,11327,11403,11422,11517,11574,11611,11612,11726,11897,12068,12087,12144,12163,12239,12258,12391,12543,12619,12700,12721,12733,12746,12825,12828,12904,12926,12955,13056,13132,13151,13185,13189,13227,13246,13262,13272,13284,13303,13360,13379,13455,13493,13502,13588,13735,13759,13816,13835,13873,13886,13913,13930,13949,13974,14044,14101,14158,14177,14215,14253,14272,14291,14310,14386,14456,14538,14559,14671,14725,14728,14734,14842,14899,14975,15032,15033,15108,15127,15146,15211,15222,15241,15279,15317,15393,15450,15469,15507,15559,15621,15659,15678,15688,15697,15798,15811,15849,15944,15963,15982,16020,16077,16109,16210,16229,16305,16324,16343,16362,16495,16514,16586,16628,16647,16678,16713,16742,16875,16894,16932,16970,16989,17116,17122,17198,17217,17255,17261,17331,17350,17354,17464,17502,17559,17628,17673,17825,17863,17882,17939,17980,18016,18072,18091,18129,18186,18224,18243,18281,18319,18338,18395,18433,18452,18490,18509,18585,18635,18661,18737,18813,18825,18851,18927,18946,19033,19117,19155,19172,19174,19212,19231,19345,19383,19402,19417,19459,19478,19497,19500,19513,19516,19592,19613,19616,19630,19649,19668,19726,19782,19820,19839,19915,19974,19991,20022,20143,20276,20295,20302,20314,20355,20371,20409,20447,20466,20485,20504,20561,20567,20637,20656,20694,20770,20867,20884,20917,21107,21112,21150,21262,21302,21435,21473,21559,21568,21625,21720,21739,21834,21891,21910,21967,21986,22024,22043,22100,22176,22214,22323,22385,22404,22442,22461,22480,22518,22537,22556,22568,22583,22594,22670,22765,22784,22803,22841,22875,22917,22936,22947,22955,23050,23069,23145,23175,23183,23202,23259,23278,23316,23354,23487,23525,23620,23639,23658,23715,23734,23791,23867,23886,23924,23962,23981,24019,24076,24095,24133,24178,24190,24263,24285,24342,24354,24361,24380,24399,24418,24437,24456,24513,24570,24598,24684,24703,24722,24798,24836,24893,24969,25026,25102,25159,25197,25273,25297,25349,25368,25444,25463,25501,25558,25577,25596,25615,25691,25729,25805,25824,25843,25919,25938,26014,26033,26106,26109,26147,26151,26223,26261,26280,26299,26414,26463,26470,26527,26641,26717,26845,27154,27157,27187,27325,27382,27406,27629,27844,27876,27935,28199,28351,28522,28731,28826,28921,29020,29155,29269,29288,29301,29377,29461,29768,29818,30061,30119,30148,30235,30351,30380,30438,30525,30583,30670,30699,30728,30757,30786,30844,30902,30931,30989,31105,31134,31192,31221,31279,31308,31395,31482,31511,31540,31569,31627,31685,31714,31772,31801,31830,31859,31888,31917,31946,31975,32033,32062,32120,32236,32294,32323,32352,32410,32468,32555,32613,32642,32671,32700,32758,32874,32932,33019,33106,33135,33164,33193,33222,33280,33338,33425,33483,33512,33541,33570,33686,33715,33773,33831,33889,33918,33947,34005,34034,34063,34092,34150,34179,34237,34266,34382,34411,34440,34527,34556,34585,34643,34672,34701,34730,34788,34817,34875,34962,35003,35087,35132,35210,35218,35255,35325,35329,35396,35403,35432,35451,35469,35474,35506,35522,35581,35667,35678,35772,35831,35877,35961,35994,36083,36143,36176,36325,36332,36391,36429,36519,36522,36581,36610,36675,36704,36715,36735,36763,36834,36893,37083,37142,37206,37372,37451,37502,37513,37585,37644,37703,37767,37826,37885,37956,38146,38205,38264,38328,38370,38446,38517,38635,38707,38754,38889,38948,39078,39106,39137,39196,39268,39327,39368,39386,39450,39639,39676,39698,39726,39757,39785,39797,39829,39856,39859,39947,40201,40259,40288,40346,40404,40433,40462,40520,40549,40578,40607,40665,40694,40723,40810,40839,40868,40897,40926,40955,41013,41042,41100,41129,41158,41187,41332,41361,41390,41419,41448,41477,41506,41535,42121,42214,42256,42346,42366,42421,42444,42447,42508,42547,42566,42583,42658,42704,42787,42820,42938,42969,43038,43057,43106,43132,43159,43262,43272,43294,43369,43531,43606,43632,43647,43656,43657,43714,43739,43901,43911,43930,43966,44005,44030,44034,44048,44080,44185,44242,44269,44388,44450,44479,44534,44554,44716,44736,44791,44813,44819,44828,44846,44848,44878,44911,44934,44953,45028,45101,45149,45210,45268,45309,45341,45380,45418,45585,45658,45715,45772,45830,45860,45959,46035,46053,46084,46128,46199,46384,46605,46680,46702,46725,46740,46756,46841,46856,46874,46912,46958,47009,47027,47068,47075,47107,47125,47145,47146,47152,47170,47210,47230,47296,47304,47347,47456,47460,47526,47536,47593,47603,47629,47707,47795,47985,47986,48003,48011,48052,48071,48117,48173,48219,48236,48250,48266,48316,48338,48358,48598,48611,48625,48827,48924,48948,48961,49015,49025,49179,49190,49270,49387,49400,49503,49511,49524,49531,49553,49575,49676,49696,49797,49969,50164,50223,50268,50475,50567,50722,50737,50778,50929,51228,51254,51283,51760,51793,52044,52080,52125,52258,52299,52336,52395,52470,52584,52642,52883,53180,53187,53213,53258,53260,53414,53472,53496,53690,53865,53999,54090,54095,54111,54214,54330,54344,54623,54814,54975,54992,55071,55106,55119,55367,55689,55732,55758,55828,55873,55923,56005,56318,56418,56632,56641,56685,56766,57148,57169,57322,57376,57470,57557,57653,57660,57727,57822,57960,58040,58050,58096,58199,58317,58372,58509,58515,58572,58666,58705,58748,58844,58896,58946,59059,59108,59168,59199,59225,59245,59267,59354,59541,59702,59803,60031,60101,60394,60441,60458,60637,60757,60982,60989,61052,61072,61179,61286,61287,61318,61609,61626,61778,62002,62066,62070,62106,62279,62346,62636,62670,62772,62828,62937,62949,62993,63202,63225,63228,63301,63305,63469,63484,63559,63749,63788,63830,63875,63888,63900,64034,64142,64155,64220,64257,64422,64464,64477,64558,64584,64624,64646,64723,64790,65071,65075,65154,65189,65227,65347,65495,65540,65662,65690,65758,65773,65783,65796,65826,66027,66043,66130,66191,66361,66401,66451,66520,66547,66633,66710,66912,66919,67002,67016,67076,67093,67119,67339,67347,67426,67471,67602,67848,67862,67883,67889,67956,68210,68233,68305,68346,68475,68526,68573,68625,68658,68661,68693,68905,68954,69250,69292,69431,69596,69598,69681,69712,69751,69762,69787,69788,69863,69928,69961,69962,69983,70011,70091,70099,70226,70242,70243,70311,70338,70532,70545,70746,70747,70825,70921,70983,71027,71097,71109,71162,71176,71222,71408,71445,71683,71910,71986,72074,72151,72244,72333,72434,72553,72563,72799,72849,72883,72923,72945,72948,73052,73173,73192,73334,73409,73515,73536,73585,73592,73671,73685,73872,73908,73977,74068,74078,74087,74197,74202,74306,74422,74439,74502,74656,74719,74741,74933,75127,75229,75381,75425,75476,75508,75546,75551,76023,76099,76135,76190,76215,76237,76256,76351,76512,76531,76598,76610,76701,76745,76772,76773,76838,76929,76945,77058,77084,77258,77290,77375,77482,77483,77490,77531,77756,77931,78001,78066,78120,78128,78177,78207,78224,78295,78405,78547,78561,78933,79161,79209,79279,79285,79372,79481,79561,79625,79848,80041,80181,80193,80222,80449,80479,80516,80769,80821,80863,80973,81273,81439,81441,81576,81590,81804,82072,82128,82156,82249,82300,82321,82361,82416,82438,82444,82518,82578,82614,82689,82818,82825,82826,82862,82887,83162,83194,83201,83408,83561,83624,83654,83687,83855,83875,83891,83967,84180,84250,84379,84447,84448,84514,84763,84771,84820,84874,85081,85090,85244,85275,85359,85380,85395,85537,85558,85642,85649,85695,85723,85750,85834,85878,86241,86262,86367,86451,86611,86612,86629,86740,86744,86964,86993,87034,87189,87203,87288,87302,87374,87713,87725,87837,87936,88010,88096,88120,88290,88308,88361,88469,88744,88807,88861,88894,88990,89015,89024,89086,89104,89141,89155,89185,89218,89261,89263,89292,89487,89499,89595,89603,89611,89640,89903,89918,89983,90103,90159,90191,90208,90216,90258,90316,90448,90552,90631,90632,90634,90670,90734,90773,90846,90868,90901,90927,91116,91212,91224,91315,91899,91908,91945,92058,92082,92123,92160,92170,92269,92326,92361,92365,92432,92514,92532,92569,92613,92636,92734,92749,92833,92871,93015,93036,93115,93139,93142,93149,93412,93444,93464,93497,93502,93690,93707,93873,93891,93905,93926,94143,94208,94282,94397,94467,94521,94525,94827,94878,94949,94975,94987,95139,95168,95259,95386,95685,95730,95864,95879,95895,96031,96045,96325,96412,96529,96602,96844,97111,97147,97169,97293,97337,97480,97552,97628,97812,97982,98206,98440,98550,98591,98746,98761,98844,99160,99328,99454,99532,99554,99569,99574,99736,99843",
                "32,127,146,1655,1659,1685,1712,1747,1780,1784,1932,2312,2318,2352,2391,2598,2676,2705,2733,3069,3172,3205,3261,3509,3646,3965,3984,4192,4484,4803,6096,6587,6625,7168,7544,7926,8126,8145,8183,8563,8639,8677,8696,8804,8872,9036,9095,9152,9247,9266,9378,9418,9421,9513,9532,9570,9627,9646,9665,9684,9703,9722,9741,9798,9803,9921,9950,10064,10106,10216,10232,10235,10331,10471,10473,10577,10627,10634,10653,10779,10843,10859,10900,11026,11052,11109,11185,11261,11360,11432,11470,11489,11535,11622,11641,11736,11755,11850,11851,11907,11976,12059,12078,12135,12287,12367,12486,12495,12496,12546,12572,12667,12705,12754,12781,12844,12895,12914,12964,13142,13237,13342,13465,13467,13484,13522,13528,13540,13598,13636,13655,13674,13701,13712,13731,13750,13769,13807,13828,13914,14031,14056,14073,14111,14130,14135,14168,14187,14282,14301,14320,14415,14567,14586,14641,14738,14757,14814,14947,15004,15006,15042,15137,15194,15202,15251,15270,15289,15343,15346,15365,15405,15473,15517,15533,15536,15631,15634,15648,15707,15708,15726,15821,15859,15878,15897,15954,16068,16106,16163,16201,16220,16258,16277,16296,16334,16353,16391,16448,16467,16524,16562,16657,16771,16790,16807,16809,16862,16885,16920,16961,16972,17005,17026,17056,17094,17100,17134,17151,17170,17235,17303,17341,17360,17379,17417,17421,17484,17486,17531,17626,17639,17652,17740,17759,17816,17873,17892,17930,17968,18006,18063,18082,18101,18158,18196,18367,18386,18405,18439,18443,18449,18481,18547,18557,18614,18633,18655,18720,18727,18823,18861,18880,18918,18937,18967,19012,19051,19089,19127,19203,19239,19247,19298,19317,19374,19393,19412,19469,19507,19545,19564,19621,19697,19701,19716,19773,19811,19830,19868,19906,19925,19963,19973,20001,20020,20077,20172,20210,20250,20267,20286,20305,20343,20362,20381,20400,20419,20457,20476,20495,20609,20666,20780,20837,20856,20888,20970,21027,21084,21179,21217,21236,21255,21331,21388,21407,21422,21426,21445,21483,21521,21578,21597,21635,21653,21654,21730,21749,21787,21801,21806,21844,21856,21863,21901,21939,21958,21977,21996,22015,22072,22091,22129,22195,22205,22224,22243,22262,22357,22376,22414,22433,22471,22509,22528,22547,22604,22680,22724,22737,22775,22813,22851,22965,22970,23022,23063,23117,23155,23231,23327,23440,23478,23516,23535,23548,23573,23687,23725,23739,23811,23839,23849,23868,23877,23915,23972,24010,24029,24086,24181,24219,24257,24291,24352,24409,24446,24447,24485,24504,24523,24542,24561,24580,24618,24675,24713,24740,24770,24865,24891,24998,25055,25131,25150,25226,25245,25264,25283,25301,25397,25401,25435,25457,25511,25568,25644,25682,25738,25815,25856,25897,25909,25948,25967,25986,26024,26081,26100,26138,26214,26252,26262,26290,26300,26309,26328,26371,26385,26390,26423,26461,26537,26594,26643,26707,26860,27045,27069,27107,27143,27254,27278,27320,27354,27386,27544,27597,27718,27734,27753,27843,27981,28133,28226,28228,28233,28304,28442,28494,28551,28738,28760,28867,28893,29026,29043,29197,29332,29445,29481,29506,29735,29767,29973,30023,30110,30197,30226,30284,30313,30342,30371,30429,30458,30516,30603,30632,30661,30690,30719,30748,30777,30806,30893,30922,30951,31038,31096,31212,31270,31299,31357,31415,31502,31531,31560,31589,31676,31705,31792,31821,31850,31937,31995,32082,32111,32140,32169,32227,32256,32285,32372,32430,32459,32488,32546,32604,32720,32749,32778,32894,32923,32952,32981,33010,33097,33126,33155,33184,33213,33242,33271,33300,33358,33387,33416,33445,33474,33503,33532,33590,33706,33735,33764,33822,33880,33938,33967,33996,34054,34083,34141,34170,34257,34315,34344,34402,34489,34547,34605,34634,34663,34750,34779,34895,34924,34953,34982,35030,35112,35186,35247,35248,35275,35302,35349,35358,35423,35449,35457,35460,35494,35687,35733,35742,35851,35922,36040,36103,36163,36186,36293,36352,36411,36483,36556,36724,36783,36869,36887,36913,36995,37058,37103,37162,37203,37226,37238,37285,37344,37474,37533,37605,37664,37723,37787,37846,37905,37976,38035,38166,38366,38407,38475,38537,38596,38655,38662,38750,38786,38861,38909,39027,39055,39098,39130,39157,39288,39406,39529,39588,39659,39692,39849,39967,39996,40116,40145,40203,40232,40261,40290,40319,40348,40406,40435,40464,40493,40522,40580,40609,40667,40696,40725,40754,40870,40899,40928,40986,41015,41044,41073,41102,41131,41160,41189,41218,41276,41305,41363,41392,41508,41537,42007,42036,42065,42094,42123,42183,42273,42291,42299,42321,42348,42376,42407,42423,42489,42510,42607,42638,42660,42747,42867,42897,42956,43009,43059,43345,43458,43477,43518,43533,43555,43608,43625,43695,43720,43728,43792,43845,43876,43891,43932,43980,44007,44082,44195,44203,44204,44244,44319,44364,44382,44383,44406,44425,44455,44628,44643,44744,44755,44793,44837,44856,44880,44936,44955,44978,44989,44995,45004,45055,45078,45103,45151,45172,45212,45249,45343,45400,45458,45486,45530,45587,45645,45717,45751,45753,45832,45904,45961,46022,46160,46258,46316,46375,46396,46488,46545,46575,46627,46844,46848,46881,46896,46916,46926,46976,47045,47049,47133,47142,47153,47301,47316,47351,47398,47541,47557,47591,47788,47807,47811,47896,47949,47977,48033,48073,48108,48176,48197,48453,48460,48528,48563,48590,48635,48670,48718,48721,48756,48768,48783,48903,48921,48960,48969,49033,49061,49127,49161,49173,49207,49235,49238,49245,49262,49332,49348,49414,49462,49500,49506,49547,49588,49605,49640,49687,49763,49832,49872,49934,49941,49984,50053,50055,50093,50156,50427,50614,50806,50826,51071,51073,51313,51340,51344,51346,51396,51404,51704,51729,51757,51803,51968,52013,52039,52096,52209,52232,52338,52456,52527,52612,52690,52706,52716,52937,52961,53046,53084,53174,53327,53357,53436,53440,53551,53699,53830,54092,54245,54294,54304,54335,54361,54654,54664,54811,54921,55102,55113,55211,55226,55301,55359,55396,55413,55439,55501,55570,55578,55730,55742,55882,55883,55952,56018,56074,56079,56182,56184,56216,56255,56283,56287,56353,56397,56618,56655,56738,56784,56855,56933,57168,57196,57228,57265,57534,57541,57586,57733,57747,57785,57940,57975,58032,58045,58061,58076,58192,58244,58285,58625,58694,58854,58867,58886,58893,58904,59096,59097,59135,59154,59170,59362,59370,59403,59520,59599,59649,59689,59696,59780,59795,60013,60114,60155,60185,60257,60505,60597,60607,60613,60808,60866,60967,60997,61060,61076,61121,61160,61291,61322,61450,61470,61524,61529,61688,61850,61881,61889,61917,62043,62145,62198,62330,62374,62446,62824,63008,63155,63162,63163,63178,63268,63501,63519,63673,63718,64173,64377,64593,64600,64663,64876,64908,64969,65147,65231,65321,65379,65443,65454,65457,65511,65517,65558,65619,65917,66040,66044,66081,66099,66138,66140,66187,66195,66247,66303,66370,66497,66725,67085,67284,67399,67418,67488,67629,67749,67841,67965,67998,68040,68066,68120,68204,68218,68252,68258,68274,68303,68351,68392,68434,68467,68477,68715,69069,69074,69144,69471,69624,69978,69997,70082,70131,70209,70247,70318,70329,70405,70474,70642,70730,70855,70879,70894,71077,71300,71643,71680,71806,71844,71868,71943,72061,72300,72312,72457,72560,72566,72717,72939,73032,73115,73151,73308,73313,73391,73404,73417,73425,73484,73526,73567,73774,73891,74002,74040,74189,74292,74446,74534,74560,74628,74708,74735,74856,74898,74926,74989,75231,75237,75468,75593,75752,75836,75863,75924,75983,76173,76221,76299,76300,76510,76621,76645,76658,77067,77203,77230,77234,77296,77358,77361,77397,77542,77572,77782,77803,77852,77859,77909,77945,78056,78121,78146,78154,78245,78362,78425,78450,78769,78834,78936,78972,79002,79225,79322,79329,79359,79369,79489,79545,79566,79608,79672,79869,80131,80145,80179,80446,80485,80497,80510,80568,80712,80735,80869,80970,81026,81121,81215,81219,81223,81240,81298,81374,81505,81771,81890,81900,81943,81966,82025,82044,82048,82132,82150,82219,82349,82350,82448,82461,82484,82499,82513,82666,82698,82739,83054,83122,83367,83417,83481,83550,83625,83793,83815,83963,84071,84200,84238,84376,84423,84443,84713,84753,84784,84873,84985,85012,85045,85071,85192,85304,85452,85481,85512,85603,85662,85901,85939,85960,86031,86098,86158,86249,86409,86572,86694,86728,86745,86771,86868,86967,87153,87335,87343,87347,87399,87450,87503,87732,87830,87990,88276,88348,88356,88360,88386,88428,88516,88608,88615,88661,88750,88791,88961,88972,89044,89190,89336,89447,89477,89539,89612,89736,89754,89794,89816,89830,89847,89851,89859,89942,89943,90128,90222,90424,90664,90808,90852,90864,90912,90945,90963,91372,91421,91467,91552,91645,91716,91721,91728,91750,91774,91851,91854,91973,91981,91992,92035,92046,92070,92105,92172,92251,92369,92545,92600,92742,92891,93059,93072,93084,93109,93162,93190,93237,93337,93341,93410,93712,93767,93989,94027,94036,94047,94072,94132,94198,94211,94365,94381,94399,94403,94443,94480,94506,94581,94641,94648,94656,94691,94693,94932,95125,95136,95188,95398,95437,95496,95535,95696,95752,95787,95830,95884,95951,96013,96021,96038,96423,96676,96823,96974,97065,97158,97163,97185,97211,97245,97266,97401,97454,97497,97526,97648,97949,98057,98073,98080,98175,98190,98237,98257,98411,98625,98665,98729,98762,98798,98831,98870,98901,98930,99093,99222,99244,99345,99481,99603,99765,99858",
          "20,82,213,359,362,464,589,605,635,719,754,767,782,783,805,830,843,882,917,921,1027,1085,1088,1126,1143,1176,1191,1209,1214,1247,1282,1284,1434,1460,1538,1541,1547,1556,1584,1688,1814,1818,1874,2170,2218,2235,2707,2825,3119,3237,3296,3707,3874,4294,4356,4382,4400,4437,4580,4585,5007,5014,5114,5133,5151,5171,5263,5397,5414,5441,5455,5534,5564,5677,5739,5748,5762,5768,5780,5811,5827,5842,5958,5990,6220,6266,6360,6440,6541,6554,6581,6754,6811,6819,6927,6957,7005,7007,7024,7055,7094,7137,7138,7226,7315,7343,7354,7413,7441,7617,7693,7720,7725,7740,7818,7870,7872,8056,8064,8100,8158,8240,8258,8272,8307,8587,8730,8845,8950,9042,9252,9295,9303,9436,9623,10007,10054,10116,10137,10267,10364,10476,11031,11065,11094,11191,11396,11417,11448,11519,11572,11810,11984,12073,12082,12133,12337,12376,12589,12819,13167,13387,13423,13570,13918,14373,14434,14865,14963,14994,15090,15203,15344,15381,15719,15992,16092,16248,16541,16729,16900,17142,17208,17491,17607,17958,18041,18348,18513,18719,18723,18745,19201,19211,19355,19450,19591,19653,19792,20068,20113,20134,20255,20488,20806,20945,21001,21008,21103,21207,21281,21383,21796,21825,21920,22058,22068,22184,22201,22889,23165,23584,23896,24238,24394,24419,24535,24704,24803,25492,25746,26204,26375,26827,26917,26988,27109,27185,27253,27264,27403,27444,27741,27750,27835,27850,28189,28763,29031,29878,29936,30835,31154,31183,31386,31473,31618,32198,33677,34228,34373,35137,35695,36113,36129,36542,36563,36568,37243,37442,37571,37813,37869,37874,37945,38094,38251,38430,38496,38802,38808,38988,38990,39252,39367,39381,39418,39936,39982,42304,42333,42611,43075,43199,43441,43678,43770,43820,44898,45311,45312,45742,45885,45920,45995,45999,46775,47046,47595,47799,48163,48648,50107,50125,50143,50159,50214,50316,50473,50659,50715,50781,50812,50982,51103,51245,51252,51278,51308,51383,51419,51466,51542,51559,51575,51645,51766,51824,51833,51835,51905,51950,52267,52301,52305,52339,52454,52506,52517,52666,52674,52860,52865,52934,52957,52986,53064,53067,53101,53170,53318,53370,53411,53716,53767,53773,53829,53870,53951,54087,54230,54321,54473,54552,54593,54641,54668,54685,54733,54861,55036,55132,55160,55172,55180,55244,55294,55323,55333,55381,55429,55446,55460,55495,55566,55591,55736,55765,55785,55866,55943,56196,56215,56336,56361,56392,56409,56420,56443,56471,56571,56795,56802,56846,56875,56887,56944,56991,57054,57092,57190,57264,57546,57616,57618,57645,57809,57910,57950,57994,58077,58153,58175,58237,58352,58354,58416,58450,58483,58579,58649,58653,58709,58741,58776,58842,58902,59147,59262,59270,59271,59352,59415,59423,59448,59456,59604,59645,59808,60063,60091,60107,60125,60218,60472,60530,60563,60628,60643,60742,60807,60825,60843,60878,60927,61014,61025,61055,61098,61231,61555,61583,61616,61711,61774,61809,61841,61892,61915,61941,62000,62035,62121,62146,62160,62245,62559,62591,63017,63036,63066,63070,63118,63179,63310,63359,63589,63649,63840,63961,64058,64074,64133,64139,64281,64316,64324,64342,64405,64455,64478,64570,64604,64648,64795,64814,64819,64922,65046,65076,65171,65223,65230,65251,65266,65346,65608,65762,65779,65802,65821,66035,66134,66174,66224,66228,66320,66458,66583,66639,66759,66936,67024,67034,67065,67068,67074,67124,67333,67340,67592,67677,67699,67705,67813,67898,67931,68058,68100,68147,68264,68458,68459,68603,68622,68648,68839,68897,68924,68926,68965,69083,69095,69161,69177,69267,69413,69539,69576,69615,69660,69695,69738,69952,70076,70102,70122,70123,70588,70624,70699,70724,70766,70807,70886,70890,70897,70954,70986,71116,71132,71138,71158,71171,71189,71237,71262,71361,71368,71379,71482,71507,71566,71634,71748,71760,71920,71956,72004,72019,72046,72196,72239,72283,72448,72601,72651,72661,72694,72712,72871,72962,72970,73242,73352,73443,73452,73478,73489,73522,73535,73559,73647,73695,73730,73831,73945,73972,73981,74215,74236,74248,74298,74379,74579,74598,74600,74748,74793,74828,74859,74869,74884,74912,74915,75098,75118,75135,75166,75355,75419,75523,75595,75786,75825,75888,75941,76046,76132,76292,76316,76340,76346,76434,76462,76577,76648,76736,76750,76766,76828,76833,77001,77038,77047,77513,77573,77577,77639,77687,77811,77814,77912,77913,77951,78047,78305,78343,78534,78815,78894,78942,79030,79062,79069,79090,79127,79202,79240,79528,79619,79632,79684,79720,79746,79792,79841,80022,80034,80061,80106,80329,80357,80413,80444,80500,80552,80606,80841,80868,80917,80940,81284,81297,81378,81442,81501,81530,81546,81602,81731,81740,81852,81976,82112,82245,82291,82395,82432,82527,82712,82823,82836,82873,82891,82917,82918,82991,83063,83115,83156,83210,83251,83253,83318,83338,83671,84019,84039,84049,84108,84212,84244,84299,84300,84305,84415,84459,84504,84515,84562,84614,84630,84636,84654,84729,84769,84810,85031,85167,85203,85418,85436,85748,85879,85899,85996,86045,86064,86154,86212,86277,86336,86400,86410,86415,86438,86615,86674,86778,86783,87029,87127,87345,87381,87514,87579,87632,87642,87719,87856,87859,87864,87894,88101,88141,88176,88404,88613,88654,88733,88840,88928,88954,89009,89039,89228,89396,89402,89567,89637,89667,89693,89836,90141,90201,90332,90428,90443,90547,90683,90744,90769,90774,90798,90986,91182,91255,91276,91362,91402,91473,91518,91535,91538,91683,91803,91990,92050,92137,92217,92508,92570,92574,92642,92799,92814,93033,93039,93547,93576,93640,93651,93797,93818,93919,93940,93959,94020,94045,94112,94244,94373,94429,94497,94559,94616,94645,94647,94756,94778,94790,94854,94901,95031,95087,95140,95151,95175,95258,95263,95307,95421,95562,95680,95703,95723,95770,95772,95817,95861,95980,96020,96129,96240,96261,96347,96414,96479,96482,96569,96620,96636,96658,96661,96921,97122,97145,97392,97409,97423,97499,97666,97759,97813,97924,97960,98027,98137,98264,98421,98636,98645,98865,99111,99132,99176,99197,99210,99291,99312,99334,99402,99518,99583,99635,99807,99842,99845,99913,99914",
          "214,268,634,690,812,936,951,1106,1108,1173,1188,1360,1394,1413,1430,1438,1455,1465,1468,2105,2107,2129,2259,2326,2671,2696,2780,3065,3294,3311,3584,3852,4051,4239,4316,4392,4497,4592,4690,4711,4829,4869,4942,5017,5024,5278,5279,5359,5386,5434,5437,5498,5668,5746,5763,5824,5844,5937,6012,6043,6046,6123,6144,6172,6186,6212,6240,6341,6413,6425,6452,6492,6535,6663,6714,6725,6746,6772,6818,7056,7062,7159,7160,7171,7228,7233,7387,7608,7750,7776,7781,7893,7955,8008,8031,8107,8235,8377,8379,8463,8472,8503,8532,8710,8868,8905,8965,8983,9114,9124,9372,9381,9606,9714,9910,9918,9981,10009,10031,10067,10110,10251,10280,10283,10386,10444,10453,10470,10474,10704,10720,10797,10947,10957,10988,11139,11159,11176,11253,11402,11530,11670,11786,11841,12096,12120,12171,12211,12354,12366,12392,12498,12532,12583,12591,12618,12660,12810,12929,13085,13210,13332,13339,13359,13454,13488,13788,13837,13969,13983,13985,13994,14145,14159,14263,14311,14344,14398,14465,14581,14652,14740,14885,15074,15132,15187,15210,15395,15467,15526,15578,15614,15615,15777,15800,15925,16001,16123,16173,16307,16372,16399,16441,16477,16638,16803,16948,17090,17105,17135,17166,17324,17333,17659,17751,18035,18453,18460,18576,18695,18998,19062,19093,19145,19157,19213,19293,19440,19526,19532,19576,19962,20132,20191,20258,20506,20590,20601,20608,20785,20861,21037,21056,21076,21114,21151,21312,21362,21437,21507,21855,22006,22086,22180,22339,22367,22566,22794,23001,23026,23077,23088,23123,23207,23297,23450,23451,23655,23831,24105,24324,24420,24515,24590,24591,24677,24822,24838,25074,25122,25129,25218,25223,25238,25256,25663,25864,25988,26035,26176,26263,26285,26307,26320,26358,26418,26490,26812,26843,26918,26963,26998,27056,27163,27313,27813,27873,27905,27912,27913,28266,28316,28425,28433,28503,28705,28719,28768,28798,28815,28888,28940,28961,29023,29112,29137,29175,29291,29298,29307,29359,29449,29496,29569,29812,29900,29903,29906,30033,30168,30304,30400,30526,30545,30642,31774,32024,32296,32441,32575,33108,33329,33659,34093,34209,34296,34460,34721,34819,34866,34877,35135,35257,35305,35311,35465,35552,35749,35760,36328,36445,36512,36524,36972,37044,37074,37112,37122,37258,37352,37456,37494,37620,37627,37817,37928,38131,38134,38375,38380,38519,38674,38701,38736,38755,38801,38819,38878,38943,39127,39169,39216,39365,39366,39371,39422,39470,39480,39495,39564,39818,39822,39830,39858,39867,39948,40118,40140,40198,40292,40814,40836,40923,40957,41191,42232,42258,42439,42630,42762,42800,42822,42914,42946,43237,43529,43772,43946,44125,44336,44422,44461,44623,44701,44718,44874,45197,45427,45452,45612,45795,45812,46066,46340,46387,46500,46561,47864,47894,47956,48007,48086,48385,48552,48727,48745,48964,49024,49182,49523,49722,49786,50002,50011,50021,50304,50386,50725,50866,51145,51169,51226,51491,51722,51738,51841,52019,52166,52265,52406,52432,52486,52576,52673,52680,52800,52802,52894,53030,53249,53299,53400,53416,53428,53446,53592,53607,53629,53819,53921,53937,53965,54007,54049,54149,54209,54559,54620,54626,54702,54749,54864,54901,55124,55156,55166,55174,55229,55263,55281,55303,55358,55464,55510,55538,55624,55675,55808,56066,56300,56310,56425,56669,56696,56763,56775,56817,56877,56980,57019,57022,57074,57156,57161,57225,57392,57429,57512,57559,57611,57778,57815,57871,57936,57963,58114,58119,58122,58265,58303,58336,58374,58403,58437,58517,58531,58662,58729,58733,58803,58839,58938,58985,58990,59024,59026,59071,59077,59091,59190,59238,59428,59443,59534,59720,60006,60008,60022,60025,60051,60064,60179,60197,60325,60372,60492,60533,60686,60942,60974,61039,61080,61152,61376,61385,61487,61514,61545,61623,61675,61682,61716,61987,61990,62001,62023,62120,62170,62216,62360,62365,62379,62415,62548,62978,62997,63102,63110,63123,63143,63194,63250,63336,63350,63374,63485,63510,63537,63617,63623,63626,63910,64009,64068,64072,64090,64114,64181,64193,64411,64468,64479,64485,64564,64652,64909,64910,65030,65069,65271,65463,65494,65565,65651,65706,65835,65877,65892,65894,65985,66049,66077,66188,66217,66288,66338,66427,66490,66737,66764,66842,67252,67352,67432,67485,67486,67650,67662,67717,67738,67859,67935,67937,68053,68097,68281,68293,68339,68389,68490,68627,68640,68700,68772,68840,68939,68950,68993,69068,69123,69131,69344,69359,69559,69655,69680,69731,69992,70046,70053,70229,70262,70362,70370,70437,70454,70482,70528,70574,70750,70782,70785,71059,71104,71204,71227,71241,71308,71317,71354,71469,71524,71531,71605,71660,71752,72289,72378,72598,72612,72779,72841,72844,72856,72985,72992,72999,73236,73266,73482,73571,73575,73898,73909,73975,73998,74025,74155,74184,74209,74222,74314,74329,74424,74457,74710,74726,74811,74831,74838,74903,75087,75116,75292,75371,75521,75612,75648,75651,75675,75724,75738,75751,75765,75904,75990,76017,76138,76428,76448,76455,76582,76682,76706,76741,76761,76820,76830,76860,76867,76971,77006,77106,77301,77493,77540,77713,77804,77823,77873,77874,77942,78041,78166,78241,78266,78298,78315,78316,78477,78532,78552,78896,79012,79050,79074,79075,79280,79300,79402,79409,79429,79459,79532,79543,79637,79658,79715,79730,79747,79872,79878,80161,80299,80427,80453,80471,80585,80676,80861,80870,80995,81051,81174,81307,81324,81538,81572,81685,81751,81775,81836,82020,82069,82074,82096,82144,82261,82379,82392,82426,82449,82588,82636,82638,82843,82878,82939,82990,83155,83285,83324,83506,83626,83773,83799,83808,84077,84143,84196,84237,84271,84311,84375,84425,84470,84484,84559,84567,84597,84797,84894,84932,84958,85040,85063,85185,85206,85417,85448,85574,85795,85802,85971,86019,86204,86215,86343,86403,86405,86496,86506,86589,86642,86657,86727,86800,86874,86944,86946,86953,86969,86976,87180,87219,87332,87388,87402,87612,87613,87795,87833,87845,87883,87992,88006,88045,88173,88183,88189,88435,88454,88491,88577,88604,88616,88687,88801,88855,88980,88991,89014,89029,89097,89171,89180,89219,89271,89654,89715,89726,89793,89795,89798,89839,89848,89960,90055,90059,90203,90241,90280,90289,90341,90365,90411,90525,90550,90584,90682,90690,90847,90873,90898,91032,91063,91073,91122,91229,91291,91371,91470,91495,91500,91522,91699,91861,91936,92102,92247,92253,92605,92698,92853,92865,92974,93367,93470,93579,93621,93683,93783,93816,93888,93908,94006,94147,94228,94531,94550,94679,94697,94708,94830,95029,95082,95108,95262,95589,95596,95720,95795,95880,95939,96033,96091,96136,96193,96274,96650,96662,96679,96697,96899,96998,96999,97047,97190,97255,97361,97369,97419,97420,97421,97450,97532,97536,97578,97603,97810,97822,98120,98356,98369,98513,98554,98565,98605,98725,98849,98970,99008,99040,99290,99443,99622,99626,99875,99931,99959,99970,99994",
          "251,293,328,388,403,432,435,523,687,716,777,796,825,1006,1017,1026,1062,1139,1189,1200,1207,1230,1234,1256,1276,1298,1334,1374,1378,1388,1411,1464,1476,1485,1525,1555,1633,1653,1783,1891,2815,3008,3235,3241,3376,4754,4864,5061,5082,5092,5127,5147,5168,5169,5193,5233,5310,5324,5393,5408,5430,5560,5584,5604,5691,5720,5785,5846,5874,5917,5936,5951,5974,6001,6004,6088,6153,6506,6785,6903,6937,7023,7031,7074,7130,7153,7253,7357,7403,7646,7683,7765,7969,8028,8095,8234,8247,8374,8400,8493,8514,8529,8551,8566,8598,8613,8660,8672,8780,8826,8918,8938,8994,8999,9069,9131,9170,9189,9302,9304,9330,9435,9914,10170,10175,10190,10211,10295,10380,10440,10465,10519,10526,10569,11136,11189,11271,11272,11324,11545,11719,11720,11760,12064,15379,15387,15902,15932,15981,17516,20693,23382,23552,24816,24915,25344,26334,26947,28020,28299,28584,28823,28873,28981,29082,29090,29463,29609,30962,30976,32106,34024,34362,34980,35929,36631,36865,37298,37925,38174,39311,39995,41257,41315,42472,42883,45516,45551,45880,46189,46231,46617,48546,49752,50014,50162,50245,50311,50319,50338,50391,50401,50477,50705,50837,50972,51057,51117,51150,51166,51188,51267,51287,51302,51429,51442,51567,51629,51677,51684,51741,51916,51918,52010,52059,52129,52132,52181,52190,52254,52388,52420,52521,52526,52544,52546,52603,52608,52630,52650,52719,52752,52818,52933,52948,52973,52979,53074,53135,53151,53185,53201,53259,53273,53296,53321,53337,53355,53365,53437,53458,53535,53556,53579,53796,53836,53869,54000,54056,54121,54243,54251,54637,54652,54672,54679,54683,54853,54903,54915,54929,54969,54987,55152,55179,55266,55573,55635,55706,55719,55864,55973,56019,56251,56298,56314,56526,56557,56559,56587,56588,56897,56963,57032,57051,57209,57271,57333,57580,57609,57624,57855,57904,58075,58123,58304,58686,58688,58831,58862,58866,58944,58961,59020,59049,59095,59101,59160,59255,59257,59290,59705,59788,59797,59849,59964,60027,60194,60295,60373,60386,60463,60549,60554,60569,60737,60868,60902,61089,61153,61238,61406,61453,61491,61596,61644,61701,61773,61832,61932,61969,62123,62144,62354,62432,62682,62884,63012,63046,63115,63153,63180,63294,63356,63398,63464,63474,63691,63787,63808,63904,63926,63960,63979,64111,64157,64170,64237,64366,64380,64448,64531,64637,64816,64839,65158,65243,65361,65566,65576,65603,65670,65712,65754,65765,65844,66184,66390,66453,66481,66530,66600,66638,66812,66843,66874,66946,66965,67050,67110,67211,67222,67241,67246,67280,67414,67428,67652,67702,67710,67722,67878,67936,68121,68150,68243,68314,68340,68410,68450,68709,68735,68895,68909,69000,69138,69245,69252,69303,69315,69438,69472,69605,69699,69743,69816,69844,69867,70072,70155,70257,70414,70498,70561,70849,70860,70909,71004,71044,71060,71274,71275,71411,71505,71821,72023,72062,72251,72428,72937,73006,73076,73086,73274,73438,73523,73540,73542,73576,73581,73729,73874,73912,73956,74038,74039,74095,74167,74176,74337,74397,74513,74521,74559,74692,74776,74913,75008,75065,75117,75142,75378,75453,75496,75575,75728,75785,75844,75846,75873,76025,76117,76406,76452,76661,76666,76729,77151,77216,77225,77252,77265,77287,77356,77432,77507,77586,77594,77598,77645,77670,77863,77877,77878,77882,78046,78085,78101,78150,78161,78195,78197,78272,78361,78377,78386,78665,78705,78835,79037,79042,79072,79084,79291,79443,79475,79525,79659,79840,79917,79922,79926,79967,80068,80076,80087,80129,80375,80548,80591,80752,80755,80805,80919,81035,81061,81078,81105,81153,81178,81189,81244,81245,81447,81562,81599,81651,81680,81769,81818,81897,81911,82064,82212,82264,82287,82303,82328,82353,82358,82388,82608,82642,82687,82732,82748,82789,82892,82948,82960,82996,83236,83339,83362,83393,83587,83645,83735,83882,83951,83970,84033,84136,84169,84239,84276,84347,84367,84418,84554,84641,84731,84768,84921,84928,84944,85072,85173,85265,85273,85281,85346,85587,85632,85811,85973,86005,86009,86161,86252,86308,86338,86404,86479,86519,86742,86919,86977,87103,87161,87173,87234,87278,87341,87417,87452,87655,87690,87977,87981,88149,88152,88158,88204,88264,88275,88289,88417,88455,88479,88490,88689,88742,88778,88813,88845,89033,89076,89080,89094,89156,89168,89231,89276,89291,89349,89394,89782,89974,89999,90085,90094,90151,90297,90334,90354,90414,90623,90633,91027,91030,91060,91067,91099,91102,91140,91311,91479,91505,91510,91586,91723,92011,92061,92143,92309,92359,92427,92442,92524,92650,92813,93131,93265,93496,93551,93655,93681,93710,93748,93773,93819,93823,93875,93916,94089,94116,94149,94226,94231,94250,94254,94328,94357,94426,94441,94473,94518,94802,94893,95118,95455,95468,95522,95638,95738,95851,95862,95874,95929,96064,96065,96285,96345,96352,96380,96560,96719,96722,96737,96960,97216,97345,97596,97605,97640,97653,97677,97789,97833,97934,97935,97955,98031,98089,98344,98428,98431,98465,98482,98491,98500,98710,99006,99161,99323,99362,99688,99929,99947,99964,99992",
          "62,101,133,257,266,274,285,291,294,466,474,493,522,534,617,793,831,844,1059,1195,1471,1507,1551,1691,1711,1746,2751,3718,4101,4268,4781,5100,5361,5399,5471,5499,5522,5553,5565,5589,5620,5636,5721,5987,6136,6191,6294,6306,6502,6557,6654,6740,6751,6861,7058,7067,7107,7135,7327,7395,7463,7581,7589,7708,7772,8114,8139,8169,8246,8274,8329,8390,8692,8786,8937,9116,9239,9240,9315,9450,9457,9490,9519,9836,9892,9919,9995,10030,10089,10180,10315,10358,10368,10400,10518,10568,10574,10582,10652,10670,10787,10913,10965,10972,10997,11048,11099,11366,11418,11451,11668,11677,11913,11931,11977,12004,12007,12038,12425,12502,12517,12533,12674,12687,12800,12869,12877,13088,13136,13146,13153,13549,13572,13645,13668,13888,13912,13919,13980,14450,14640,14692,14751,14823,14858,15037,15045,15105,15118,15174,15178,15209,15274,15441,15447,15958,16057,16094,16434,16462,16528,16547,16676,16700,16752,16868,16992,17029,17099,17419,17649,17955,18015,18056,18068,18113,18159,18174,18599,18974,18979,19000,19238,19410,19457,19521,19758,20218,20238,20684,20698,20984,21266,21399,21557,21815,22291,22297,22590,22592,22648,22944,23016,23633,23663,23778,23923,24048,24173,24284,24300,24375,24382,25454,25933,26167,26256,26315,26816,26857,26873,27111,27125,27161,27284,27287,27411,27706,27715,27757,27779,27787,27801,27824,27914,27941,28274,28368,28461,28978,29121,29129,29297,29540,35763,35764,36075,36435,36516,36698,37015,37744,37997,38358,38379,38488,38513,38687,38816,38917,39073,39298,39302,39620,39694,39807,45872,50038,50052,50080,50097,50132,50192,50208,50212,50551,50599,50864,50954,51076,51277,51328,51351,51387,51789,51836,51969,52113,52210,52248,52367,52397,52508,52676,52714,52803,52907,52918,53129,53178,53188,53198,53313,53430,53522,53620,53701,53743,53789,53906,53981,53982,54048,54415,54421,54524,54530,54600,54909,54916,55054,55456,55625,55626,55629,55695,55749,55796,55823,55863,55912,55945,55974,56071,56088,56203,56290,56313,56344,56399,56433,56540,56832,56964,57043,57086,57123,57125,57194,57203,57233,57241,57436,57510,57529,57582,57691,57731,57789,57813,57869,57998,58010,58307,58311,58315,58408,58414,58527,58725,58736,58788,58880,58998,59174,59552,59700,59783,59812,59926,60041,60124,60126,60156,60234,60352,60690,60704,60720,60755,60953,61090,61130,61218,61433,61521,61608,61761,61788,61939,61983,62079,62094,62178,62472,62648,62708,62728,62765,62864,63060,63077,63238,63242,63288,63505,63692,63861,63955,64014,64153,64162,64190,64253,64383,64386,64538,64543,64847,64865,64880,65058,65169,65304,65363,65426,65725,65927,65949,66019,66253,66356,66460,66510,66523,66545,66590,66666,66731,66747,66748,66911,66924,67019,67023,67293,67301,67357,67375,67392,68153,68194,68525,68540,68550,68719,68757,68810,68998,69004,69075,69133,69145,69286,69337,69383,69406,69423,69434,69480,69494,69701,69707,69856,69904,69939,69946,70108,70115,70223,70333,70343,70467,70565,70600,70664,70701,70792,70915,70939,71090,71372,71389,71820,71872,71974,71990,71997,72008,72016,72071,72119,72274,72279,72438,72454,72463,72520,72615,72705,72767,72798,72802,72806,72854,72906,72917,72984,73015,73019,73020,73031,73093,73106,73130,73165,73373,73598,73605,73716,73764,73832,73846,73860,73871,74055,74079,74086,74137,74273,74346,74381,74585,74742,74767,74864,74936,74944,75147,75155,75325,75369,75380,75467,75510,75532,75550,75674,75827,75843,75938,76015,76223,76381,76473,76704,76908,76984,77266,77270,77305,77347,77365,77405,77459,77665,77822,77869,77893,77996,78017,78209,78327,78360,78428,78447,78469,78526,78658,78679,78688,78746,78760,78812,78824,78838,78908,79041,79108,79121,79123,79132,79172,79305,79310,79423,79437,79547,79629,79646,79744,79755,79823,79866,80062,80081,80094,80114,80202,80253,80326,80458,80494,80820,80848,80932,81006,81132,81205,81206,81286,81398,81409,81459,81518,81622,81681,81784,81820,81826,81830,81862,81917,81930,82032,82061,82085,82114,82310,82315,82402,82408,82422,82430,82623,82752,82876,82956,83007,83013,83038,83148,83151,83512,83549,83568,83803,83833,83909,84118,84133,84217,84240,84304,84352,84355,84393,84468,84508,84593,84649,84712,84897,84974,85181,85195,85361,85396,85505,85518,85598,85751,85764,85765,85775,85916,85933,86025,86299,86444,86511,86957,87025,87336,87351,87421,87495,87518,87530,87610,87714,87724,88123,88145,88170,88215,88266,88279,88537,88634,88751,88859,88912,88971,88997,89027,89038,89042,89121,89186,89202,89245,89251,89257,89299,89329,89342,89353,89385,89562,89583,89651,89728,89820,89956,90089,90140,90181,90287,90362,90401,90490,90522,90579,90870,90911,90976,90988,91055,91062,91084,91152,91222,91353,91544,91567,91591,91630,91690,91748,91781,91841,91845,91874,91933,91961,91969,91979,92023,92240,92264,92270,92305,92343,92498,92566,92667,92679,92713,92718,92726,92750,92835,92841,92879,93080,93098,93174,93221,93468,93524,93544,93850,93858,93877,94003,94119,94163,94174,94232,94325,94352,94574,94633,94654,94973,95022,95051,95093,95194,95215,95217,95231,95271,95287,95558,95927,96124,96159,96225,96230,96396,96424,96492,96696,96802,96874,96876,97084,97090,97154,97223,97238,97250,97319,97410,97471,97515,97550,97558,97609,97654,97778,97852,97912,97945,98026,98068,98125,98204,98208,98221,98306,98350,98401,98427,98480,98540,98750,98757,98799,98800,98877,98988,99045,99078,99130,99151,99367,99414,99459,99529,99556,99675,99871",
          "7,21,98,216,302,545,560,593,597,614,663,704,714,730,732,751,778,787,794,828,919,938,953,968,980,987,1063,1168,1316,1375,1408,1429,1443,1481,1501,1524,1603,1803,1895,1907,1993,2020,2063,2085,2086,2192,2320,2321,2324,2423,2521,2640,2737,3092,3282,3324,3653,3773,3884,4110,4273,4366,4495,4802,4862,4969,5026,5153,5369,5383,5612,5646,5711,5717,5731,5774,5862,5942,6009,6169,6357,6526,6531,6650,6679,6850,6971,6973,7009,7065,7141,7178,7306,7382,7400,7484,7532,7624,7652,7706,7715,7774,7888,7972,7975,8177,8298,8341,8411,8436,8447,8470,8526,8602,8708,8759,8783,8795,8879,8889,8943,8973,8981,9133,9186,9229,9365,9609,9631,9814,9953,9970,10122,10150,10179,10348,10349,10543,10596,10736,10954,11006,11010,11018,11053,11150,11343,11397,11427,11428,11438,11452,11537,11569,11593,11626,11770,11797,11816,11902,12218,12443,12571,12606,12608,12659,12725,12950,13119,13294,13331,13512,13637,14422,14483,14630,14854,15007,15167,15402,15616,15882,15996,16021,16030,16062,16112,16585,16848,16973,17031,17124,17241,17283,17402,17422,17614,17618,17689,18430,18543,18558,18580,18657,18768,19055,19193,19226,19356,19361,19640,19815,20404,20518,20675,20760,20784,20832,20924,21031,21041,21133,21208,21449,21514,21532,21899,22074,22161,22225,22391,22646,22799,22931,23080,23090,23374,23403,23441,23545,23723,23749,24087,24320,24333,24397,24524,24546,24787,24841,24868,24932,24979,25046,25336,25366,25439,25486,25641,25732,25750,25884,26047,26251,26321,26323,26332,26459,26661,26703,26715,26797,26814,27214,27227,27279,27352,27364,27447,27649,27662,27678,27735,28290,28323,28366,28426,28432,28524,28700,28734,28882,28931,28972,29016,29042,29100,29198,29280,29345,29386,29792,29803,29824,29852,29984,30578,31097,32373,32634,32782,32811,33040,33127,33417,33446,34142,35378,35572,35800,35926,35943,36076,36282,36300,36313,36442,36500,36510,36620,36638,36641,36814,36941,37045,37261,37314,37366,37380,37498,37549,37819,37950,38067,38258,38324,38368,38437,38441,38500,38675,38696,38757,39000,39351,39362,39610,39682,39729,39752,39803,39872,40217,40509,41198,41290,42323,42488,42550,42573,42722,42781,42860,43111,43196,43394,43433,43885,44302,44430,44463,45185,45189,45235,45252,45329,45881,45950,45993,46118,46455,46574,46589,46830,46873,47555,48477,49057,49198,49582,49675,50060,50172,50187,50378,50380,50442,50490,50525,50544,50663,50788,50883,50916,50921,51365,51436,51528,51532,51663,51755,51946,51993,52027,52068,52082,52168,52237,52251,52304,52423,52491,52499,52887,52889,52969,53050,53224,53256,53473,53576,53662,53824,53848,53880,53996,54023,54036,54083,54096,54107,54194,54265,54323,54395,54580,54610,54748,54838,55056,55068,55074,55108,55151,55315,55335,55366,55461,55505,55621,55759,55908,56025,56093,56101,56234,56249,56258,56302,56383,56390,56507,56514,56551,56753,56767,56787,56807,56809,56952,57101,57134,57239,57306,57433,57620,57719,57915,57928,57961,57962,57981,58007,58070,58150,58328,58348,58487,58562,58648,58739,58754,58792,58845,59006,59184,59221,59281,59314,59408,59698,59768,59813,59851,59868,59899,59925,59987,60089,60151,60209,60251,60338,60437,60438,60586,60594,60623,60764,60922,60933,61048,61243,61345,61351,61485,61532,61551,61704,61847,61985,62061,62074,62171,62172,62238,62325,62367,62588,62634,62643,62793,62885,62906,62941,63000,63129,63362,63376,63548,63569,63570,63596,63651,63707,63869,63872,63901,64039,64206,64221,64241,64371,64546,64724,64747,64772,65064,65233,65305,65374,65438,65455,65460,65489,65653,65751,65862,65940,66037,66259,66334,66526,66605,66700,66732,66753,66967,67096,67131,67304,67475,67524,67598,67612,67635,67655,67721,67734,67838,67923,67944,68004,68206,68212,68286,68347,68518,68729,68778,68883,68956,68985,69087,69113,69117,69125,69134,69454,69485,69623,69667,69830,69892,69960,70225,70240,70272,70287,70321,70366,70442,70520,70564,70618,70727,71067,71085,71198,71228,71287,71345,71412,71502,71530,71783,71833,71867,71871,71923,72103,72104,72259,72443,72518,72531,72654,72755,72769,72847,72953,73097,73100,73141,73244,73312,73394,73473,73593,73724,73887,73957,74081,74129,74156,74217,74398,74508,74578,74630,74655,74729,74840,74897,74943,75034,75091,75100,75137,75195,75204,75219,75300,75360,75524,75600,75601,75607,75644,75671,75699,75965,76029,76056,76096,76208,76217,76230,76260,76400,76458,76516,76536,76559,76684,76797,77034,77187,77219,77232,77329,77461,77552,77679,77819,77828,77867,77978,78030,78040,78096,78136,78277,78357,78509,78678,78916,78969,79136,79167,79270,79343,79455,79458,79634,79711,79756,79772,79777,79902,80032,80100,80109,80110,80148,80169,80231,80301,80587,80711,80846,80945,80953,81021,81052,81195,81199,81237,81250,81254,81295,81346,81500,81605,81610,81770,81786,81843,81874,81923,82475,82503,82510,82704,82730,82769,82881,82904,82906,82953,82973,83018,83235,83266,83281,83356,83380,83394,83464,83577,83582,83703,83820,83911,83976,84012,84095,84096,84203,84216,84287,84313,84327,84457,84606,84761,84853,84960,85122,85209,85264,85330,85397,85437,85470,85500,85614,85636,85661,85739,85797,85861,85934,86111,86139,86181,86253,86280,86297,86473,86586,86628,86817,86904,86935,87098,87166,87195,87265,87328,87342,87623,87697,87787,87877,87973,88021,88119,88155,88188,88351,88651,88725,88736,88781,88853,88881,88887,88931,88942,89188,89416,89449,89450,89643,89668,89833,90232,90262,90303,90340,90380,90392,90400,90478,90685,90884,90923,91038,91081,91206,91347,91494,91504,91562,91710,91712,91823,91910,91955,92135,92228,92395,92488,92709,92917,92941,92985,93049,93184,93201,93256,93396,93398,93403,93454,93804,93970,94205,94215,94498,94593,94600,94623,94629,94638,94687,94766,94791,94801,94849,94879,95157,95171,95296,95328,95389,95400,95418,95433,95474,95765,95773,95828,96187,96229,96336,96381,96386,96439,96481,96647,96736,96750,96839,97142,97150,97275,97286,97328,97344,97452,97481,97845,98113,98288,98365,98385,98447,98510,98536,98657,98713,98780,98804,98822,98932,99007,99030,99135,99173,99225,99280,99472,99484,99647,99746,99794,99838,99856,99982",
          "136,184,230,1620,1676,1745,1764,1992,2011,2061,2122,2149,2173,2209,2438,2440,2464,2582,2595,2691,2839,2985,3015,3050,3285,3329,3474,3695,4025,4063,4271,4306,4391,4488,4558,4581,4597,4618,4642,4849,4892,5075,7120,7373,7773,8393,9184,9497,9611,9630,9649,9668,9706,9720,9725,9782,9801,10048,10599,10618,10675,10713,10725,10808,10903,10941,11017,11093,11112,11131,11226,11227,11245,11359,11444,11454,11511,11525,11528,11550,11587,11606,11644,11663,11891,11948,11967,11986,12024,12157,12233,12252,12271,12290,12328,12329,12385,12403,12423,12478,12499,12537,12596,12613,12639,12668,12679,12708,12765,12803,12847,12860,12883,12898,12923,12974,13012,13048,13050,13069,13114,13145,13164,13240,13297,13345,13354,13362,13373,13430,13449,13563,13617,13639,13658,13677,13715,13734,13772,13791,13829,13838,13855,13870,13889,13962,13981,14019,14045,14076,14082,14095,14147,14152,14171,14190,14201,14228,14247,14266,14285,14304,14323,14332,14380,14418,14446,14475,14502,14570,14571,14646,14665,14672,14722,14741,14742,14759,14859,14874,14931,15026,15064,15083,15100,15140,15197,15235,15273,15292,15311,15349,15368,15425,15444,15501,15577,15595,15596,15605,15640,15672,15691,15710,15748,15805,15824,15843,15862,15900,15919,15938,15942,15976,15988,16052,16071,16090,16100,16147,16166,16183,16185,16204,16221,16242,16261,16273,16280,16293,16299,16318,16337,16356,16413,16428,16432,16451,16458,16470,16508,16527,16565,16584,16622,16698,16717,16793,16831,16888,16929,16962,16964,17021,17040,17059,17079,17097,17154,17173,17192,17211,17239,17268,17306,17325,17363,17382,17401,17458,17496,17515,17534,17591,17624,17648,17686,17705,17743,17762,17800,17819,17838,17876,17895,17914,17971,17990,18028,18066,18104,18123,18142,18161,18170,18216,18218,18237,18250,18256,18275,18313,18332,18351,18370,18389,18408,18427,18484,18503,18523,18579,18587,18598,18617,18674,18693,18731,18734,18807,18826,18854,18883,18902,18940,18941,18959,18978,18997,19021,19028,19044,19054,19073,19092,19130,19206,19232,19282,19320,19377,19396,19415,19434,19468,19472,19529,19544,19548,19586,19624,19662,19700,19724,19738,19757,19776,19795,19814,19833,19852,19871,19890,19928,19966,20023,20061,20080,20099,20108,20156,20182,20194,20203,20232,20251,20319,20327,20346,20365,20384,20403,20422,20441,20460,20517,20536,20555,20574,20612,20631,20669,20688,20726,20745,20764,20775,20840,20859,20878,20916,20935,20954,20992,21011,21049,21087,21106,21159,21163,21201,21220,21239,21296,21303,21353,21372,21410,21429,21505,21524,21562,21591,21600,21607,21619,21638,21647,21657,21676,21714,21752,21790,21809,21828,21866,21885,21904,21923,21942,21961,21980,21999,22037,22056,22075,22094,22113,22151,22170,22227,22247,22256,22265,22284,22322,22330,22341,22360,22398,22417,22436,22449,22455,22512,22522,22531,22569,22580,22607,22626,22683,22702,22721,22723,22759,22797,22816,22820,22835,22930,22987,23025,23044,23082,23101,23120,23177,23196,23291,23310,23367,23394,23424,23453,23472,23481,23500,23538,23557,23576,23582,23614,23652,23671,23688,23709,23718,23728,23747,23782,23804,23861,23880,23899,23918,23937,23956,23975,24013,24032,24070,24101,24127,24144,24165,24184,24203,24260,24276,24298,24317,24355,24374,24393,24412,24433,24450,24469,24507,24526,24545,24583,24602,24612,24621,24659,24678,24697,24716,24773,24811,24830,24883,24887,24925,24952,24963,24982,24988,25020,25058,25064,25096,25115,25134,25191,25248,25267,25268,25305,25318,25324,25343,25345,25419,25438,25476,25495,25552,25590,25609,25627,25780,25799,25818,25863,25875,25913,25929,25951,25970,26008,26027,26046,26065,26084,26103,26107,26122,26131,26141,26179,26185,26198,26227,26236,26255,26274,26331,26350,26369,26388,26407,26426,26445,26464,26502,26521,26578,26635,26683,26758,26958,27110,27140,27205,27414,27547,27561,27574,27604,27680,27699,27865,27884,27908,28003,28155,28174,28478,28491,28516,28535,28554,28592,28649,28703,28737,28822,28877,28988,29111,29124,29200,29216,29286,29352,29504,29549,29580,29618,29656,29703,29979,30026,30055,30084,30113,30142,30171,30200,30229,30258,30287,30316,30345,30374,30403,30432,30461,30490,30519,30548,30577,30606,30635,30664,30693,30722,30751,30780,30809,30838,30867,30896,30925,30954,30983,31012,31041,31070,31099,31128,31157,31186,31215,31244,31273,31302,31360,31389,31418,31447,31476,31505,31534,31563,31592,31621,31650,31679,31708,31737,31766,31795,31824,31853,31882,31911,31940,31969,31998,32027,32056,32085,32114,32143,32172,32201,32230,32259,32288,32317,32346,32375,32404,32433,32462,32491,32520,32549,32578,32607,32636,32665,32694,32723,32752,32781,32810,32839,32868,32897,32926,32955,32984,33013,33042,33071,33100,33129,33158,33187,33216,33245,33274,33303,33332,33361,33390,33419,33448,33477,33506,33535,33564,33593,33622,33651,33680,33709,33738,33767,33796,33825,33854,33883,33912,33941,33970,33999,34028,34057,34086,34115,34144,34173,34202,34231,34260,34289,34318,34347,34376,34405,34434,34463,34492,34521,34550,34579,34608,34637,34666,34695,34724,34753,34782,34811,34840,34869,34898,34927,34956,34985,35026,35068,35115,35131,35146,35189,35203,35207,35212,35225,35227,35252,35278,35286,35316,35352,35426,35438,35442,35454,35497,35502,35545,35604,35610,35663,35690,35736,35795,35818,35854,35925,35984,36043,36106,36166,36225,36231,36233,36245,36296,36355,36360,36377,36378,36414,36486,36545,36557,36604,36668,36727,36786,36857,36916,36975,37047,37106,37120,37165,37229,37252,37288,37297,37347,37418,37477,37536,37566,37608,37667,37726,37731,37790,37808,37849,37908,37979,38007,38038,38097,38106,38169,38180,38185,38228,38287,38308,38351,38410,38469,38540,38564,38599,38620,38658,38730,38740,38789,38848,38912,38927,38971,39030,39040,39045,39101,39110,39160,39219,39228,39291,39350,39409,39473,39532,39591,39600,39662,39721,39727,39780,39852,39911,39970,40102,40131,40160,40189,40218,40247,40276,40305,40334,40363,40392,40421,40450,40479,40508,40537,40566,40595,40624,40653,40682,40711,40740,40769,40798,40827,40856,40885,40914,40943,40972,41001,41030,41059,41088,41117,41146,41175,41204,41233,41262,41291,41320,41349,41378,41407,41436,41465,41494,41523,41997,42026,42055,42084,42113,42139,42141,42173,42236,42239,42263,42288,42306,42338,42413,42457,42498,42500,42538,42575,42618,42650,42695,42737,42782,42812,42861,42887,42974,43011,43026,43032,43049,43124,43188,43193,43211,43240,43256,43265,43286,43313,43323,43335,43361,43386,43388,43448,43523,43598,43633,43638,43671,43685,43713,43760,43835,43898,43922,43964,43997,44028,44072,44159,44215,44234,44309,44338,44371,44376,44396,44452,44456,44457,44471,44523,44546,44605,44607,44633,44700,44783,44870,44897,44899,44945,45020,45048,45050,45065,45093,45141,45202,45260,45291,45333,45385,45390,45420,45424,45448,45520,45577,45635,45660,45707,45739,45764,45811,45822,45894,45951,46076,46125,46151,46210,46224,46230,46321,46334,46370,46402,46453,46493,46503,46518,46537,46549,46551,46620,46707,46743,46754,46774,46837,46853,46885,46904,46922,46938,46956,46978,46995,47026,47035,47037,47051,47140,47150,47215,47232,47245,47255,47259,47332,47350,47362,47420,47454,47465,47470,47506,47507,47508,47511,47549,47553,47563,47585,47599,47604,47606,47624,47625,47654,47671,47684,47711,47737,47752,47774,47778,47797,47810,47827,47904,48020,48055,48156,48238,48255,48263,48285,48303,48305,48352,48381,48382,48429,48471,48472,48514,48529,48548,48578,48652,48668,48729,48769,48787,48807,48810,48892,48927,48933,48970,49014,49067,49074,49083,49091,49119,49160,49206,49218,49267,49293,49304,49322,49347,49432,49442,49449,49536,49550,49569,49601,49631,49647,49653,49657,49709,49714,49717,49733,49769,49798,49833,49851,49895,49903,49929,49932,49993,50063,50081,50203,50213,50457,50601,50648,50860,50940,50947,51058,51110,51154,51185,51234,51286,51354,51489,51622,51775,51837,51883,51896,51925,51939,52074,52134,52149,52182,52186,52197,52370,52392,52687,52783,52811,52852,52857,52919,53008,53021,53075,53095,53119,53161,53175,53190,53292,53310,53403,53427,53442,53552,53553,53624,53714,53734,53774,53778,53922,53933,54025,54060,54145,54158,54193,54281,54479,54498,54542,54561,54604,54655,54697,54961,54980,54993,55026,55047,55135,55236,55245,55375,55404,55433,55467,55616,55731,55757,55836,55855,55865,55894,55922,55944,56274,56286,56364,56379,56491,56633,56793,56794,56891,56978,57180,57201,57276,57385,57437,57516,57549,57734,57762,57884,58290,58295,58629,58692,58717,58824,58894,58940,58984,59093,59144,59146,59284,59543,59573,59595,59603,59967,59984,60036,60345,60377,60398,60468,60475,60706,60710,60811,60832,60876,60915,61011,61029,61062,61095,61106,61117,61150,61161,61331,61391,61461,61484,61523,61637,61729,61902,61907,61946,61968,62017,62072,62104,62236,62258,62318,62385,62428,62448,62474,62683,62687,62699,62770,62822,62872,62953,62983,63116,63121,63309,63340,63440,63442,63500,63797,63898,63951,64022,64026,64051,64134,64260,64426,64445,64645,64676,64692,64881,64923,65166,65198,65209,65220,65249,65265,65297,65314,65351,65382,65397,65504,65673,65719,65763,65993,66031,66117,66149,66372,66571,66616,66627,66634,66656,66684,66738,66785,66813,66870,66901,66990,67037,67046,67066,67127,67183,67265,67308,67343,67406,67482,67651,67656,67672,67772,67809,67881,68163,68169,68306,68478,68594,68800,68806,68889,68929,69157,69164,69283,69440,69772,69915,70060,70160,70312,70376,70500,70503,70597,70708,70735,70799,70803,70881,70912,70931,71178,71252,71270,71349,71494,71662,71732,72010,72037,72057,72109,72115,72240,72247,72307,72432,72435,72450,72578,72620,72676,72688,72816,72855,72874,73046,73081,73109,73139,73193,73350,73382,73429,73602,73604,73625,73763,73858,73962,73973,73983,74043,74170,74173,74174,74281,74483,74493,74562,74590,74597,74603,74616,74639,74686,74736,75045,75090,75115,75224,75289,75329,75411,75652,75952,75961,76040,76084,76086,76129,76152,76181,76218,76336,76347,76398,76402,76431,76472,76624,76778,76788,76872,76880,76920,77159,77466,77522,77636,77701,77917,77944,77965,77990,78026,78098,78378,78543,78581,78662,78715,78783,78836,78910,78986,79017,79021,79047,79054,79096,79163,79230,79254,79256,79303,79313,79470,79507,79531,79719,79921,79956,80063,80277,80400,80473,80490,80762,81092,81111,81131,81134,81141,81147,81227,81233,81283,81568,81570,81577,81589,81645,81712,81798,82083,82374,82423,82452,82496,82613,82827,82889,82958,83019,83173,83508,83516,83762,83927,83933,84007,84146,84151,84181,84194,84269,84450,84704,84806,85059,85131,85137,85198,85212,85218,85280,85293,85443,85453,85546,85588,85604,85681,85722,85951,86048,86072,86117,86149,86172,86186,86189,86193,86265,86275,86416,86468,86486,86566,86582,86685,86785,86861,86872,86889,86921,87048,87148,87213,87244,87289,87420,87454,87466,87504,87527,87548,87976,87996,88019,88124,88233,88342,88362,88378,88420,88432,88459,88489,88530,88572,88635,88735,88762,88772,88815,88864,88917,88929,88932,89041,89135,89269,89424,89442,89555,89721,89724,90002,90066,90470,90659,90709,90720,90775,90891,90900,91035,91078,91241,91306,91326,91424,91437,91583,91609,91668,91697,91705,91830,92022,92295,92333,92585,92717,92772,92848,92967,93172,93251,93305,93336,93371,93652,94126,94217,94312,94372,94382,94546,94549,94635,94667,94819,94824,94885,94930,95153,95279,95295,95383,95641,95648,95653,95678,95687,95854,95911,95990,96073,96264,96277,96467,96715,96783,96815,96856,96971,97136,97146,97209,97548,97608,97642,97809,97829,97837,97972,98008,98042,98227,98277,98388,98419,98476,98487,98519,98521,98572,98581,98616,98627,98686,98702,98752,98760,98857,99019,99037,99223,99271,99289,99417,99818,99860",
          "57,95,114,152,171,223,1679,1702,1721,2302,2356,2358,2536,2709,2884,3299,3310,3336,3458,3628,3746,3801,3898,4066,4130,4398,4420,4460,4502,4591,4674,4867,4887,4997,4999,5103,5417,5672,6303,6592,7173,7724,8405,8426,9495,9500,9508,9538,9559,9595,9614,9633,9652,9671,9690,9709,9764,9822,9838,10092,10621,10678,10697,10734,10759,11134,11210,11229,11381,11394,11410,11495,11571,11613,11628,11647,11856,11862,11901,11932,11970,12008,12103,12106,12141,12154,12160,12189,12217,12236,12255,12274,12323,12350,12369,12435,12464,12511,12616,12635,12673,12768,12787,12806,12815,12861,12863,12882,12921,12939,13034,13053,13091,13110,13186,13205,13224,13229,13281,13300,13357,13366,13376,13412,13438,13452,13490,13564,13604,13608,13642,13661,13680,13699,13718,13756,13775,13813,13872,13908,13927,13946,13964,14022,14079,14102,14136,14142,14155,14193,14206,14212,14231,14250,14269,14307,14326,14345,14403,14459,14497,14560,14573,14592,14649,14706,14715,14719,14731,14744,14768,14769,14786,14801,14820,14877,14911,14934,15010,15067,15086,15143,15200,15257,15295,15314,15390,15466,15496,15532,15618,15668,15675,15808,15827,15846,15884,15913,15927,15960,15979,15995,15998,16017,16024,16093,16131,16169,16226,16245,16283,16321,16340,16359,16378,16397,16473,16511,16534,16606,16699,16701,16743,16758,16815,16834,16853,16872,16882,16921,16967,17024,17043,17074,17096,17125,17157,17176,17178,17179,17214,17252,17271,17309,17347,17366,17385,17480,17499,17537,17544,17557,17575,17621,17632,17727,17746,17765,17784,17803,17822,17841,17860,17879,17974,18050,18164,18179,18202,18215,18221,18240,18278,18297,18316,18354,18373,18392,18440,18468,18487,18493,18566,18582,18601,18639,18677,18775,18808,18810,18829,18866,18886,18905,18924,18981,19029,19076,19081,19133,19152,19171,19209,19228,19285,19304,19323,19342,19380,19399,19418,19456,19551,19552,19570,19589,19650,19665,19678,19719,19722,19750,19779,19798,19817,19836,19855,19893,19912,19950,19988,20045,20083,20102,20121,20197,20216,20222,20235,20254,20292,20311,20330,20349,20368,20387,20406,20425,20444,20463,20482,20539,20558,20596,20650,20680,20710,20767,20824,20854,20881,20995,21014,21052,21109,21204,21242,21273,21280,21314,21318,21356,21394,21432,21451,21489,21495,21508,21565,21584,21622,21641,21646,21660,21679,21717,21742,21755,21774,21812,21831,21850,21869,21888,21907,21945,21964,21983,22002,22040,22078,22097,22116,22154,22186,22211,22249,22268,22287,22344,22382,22401,22420,22477,22494,22515,22534,22553,22591,22610,22667,22691,22705,22727,22743,22762,22781,22810,22821,22876,22934,22952,22971,22990,23066,23085,23140,23142,23143,23156,23180,23237,23256,23275,23292,23294,23320,23332,23371,23389,23406,23427,23446,23484,23503,23522,23560,23579,23598,23636,23693,23712,23731,23788,23807,23859,23864,23883,23921,23940,23959,23978,23997,24054,24111,24189,24225,24282,24301,24339,24358,24377,24396,24415,24434,24444,24453,24510,24529,24548,24624,24627,24662,24681,24682,24719,24833,24852,24871,24890,24904,24947,24985,25004,25023,25025,25099,25118,25152,25156,25175,25186,25194,25203,25227,25232,25270,25289,25346,25365,25393,25403,25479,25498,25517,25536,25574,25631,25650,25701,25707,25726,25764,25768,25821,25878,25916,25935,25954,25973,25992,26011,26030,26068,26087,26144,26182,26201,26258,26277,26287,26372,26391,26429,26448,26467,26486,26619,26828,26847,26885,26904,26923,26934,26999,27170,27359,27402,27531,27591,27816,27818,27920,27922,27930,27937,27949,28055,28082,28183,28329,28419,28479,28557,28633,28652,28695,28708,28905,28918,28956,29025,29032,29061,29080,29097,29108,29114,29203,29248,29380,29479,29488,29513,29611,29763,29830,29874,29922,29923,29944,30029,30058,30087,30145,30174,30203,30232,30261,30290,30319,30348,30377,30406,30435,30464,30493,30522,30551,30580,30609,30638,30667,30696,30754,30783,30812,30841,30870,30899,30928,30957,31015,31044,31073,31102,31131,31189,31218,31247,31276,31305,31334,31363,31392,31421,31450,31479,31508,31537,31566,31595,31624,31682,31711,31740,31769,31798,31827,31856,31885,31914,31943,31972,32001,32030,32059,32117,32146,32204,32233,32262,32291,32320,32349,32378,32407,32436,32465,32494,32523,32552,32581,32610,32639,32668,32697,32726,32755,32784,32813,32842,32871,32900,32929,32958,32987,33016,33045,33074,33103,33132,33161,33190,33219,33248,33306,33335,33393,33422,33451,33480,33509,33538,33567,33596,33625,33654,33683,33712,33741,33770,33799,33828,33857,33886,33915,33944,33973,34002,34031,34060,34089,34118,34147,34176,34205,34234,34263,34292,34321,34350,34379,34408,34437,34466,34495,34524,34553,34582,34611,34640,34669,34698,34727,34756,34785,34814,34843,34872,34901,34930,34959,34988,35029,35118,35154,35192,35246,35281,35288,35307,35355,35362,35366,35429,35500,35515,35548,35607,35611,35666,35671,35739,35798,35858,35868,35899,35928,35987,36074,36109,36169,36228,36229,36250,36299,36314,36417,36422,36489,36507,36548,36607,36696,36730,36749,36789,36860,36919,36922,36927,36937,36946,36949,36978,37050,37109,37168,37232,37291,37317,37350,37354,37421,37480,37539,37670,37683,37729,37793,37820,37852,37911,37924,37982,38041,38100,38103,38124,38172,38231,38354,38413,38472,38543,38606,38661,38663,38691,38704,38733,38745,38792,38872,38873,38915,38974,39033,39038,39104,39163,39222,39236,39294,39353,39412,39447,39476,39535,39594,39665,39724,39745,39855,39874,39973,40187,40216,40245,40303,40332,40361,40390,40419,40448,40477,40506,40535,40564,40593,40622,40651,40680,40709,40738,40767,40796,40825,40854,40883,40912,40941,40970,40999,41028,41057,41086,41115,41144,41173,41202,41231,41260,41318,41347,41376,41405,41434,41463,41492,41521,41988,41993,42051,42080,42109,42168,42227,42243,42300,42318,42408,42536,42543,42555,42612,42633,42643,42645,42720,42725,42792,42882,42957,42967,43029,43033,43039,43194,43266,43324,43356,43357,43401,43415,43431,43499,43503,43593,43668,43717,43729,43740,43745,43826,43830,43864,43877,43905,43959,43968,43977,43990,44060,44067,44104,44137,44142,44145,44184,44197,44214,44263,44304,44341,44359,44373,44379,44392,44437,44459,44465,44508,44516,44522,44540,44541,44575,44587,44616,44630,44658,44665,44671,44688,44778,44826,44851,44853,44904,44925,44982,45015,45047,45086,45128,45134,45170,45195,45244,45253,45292,45326,45441,45513,45570,45628,45700,45757,45815,45887,45944,45946,46027,46055,46056,46067,46082,46083,46119,46129,46131,46152,46245,46330,46371,46385,46439,46443,46449,46469,46470,46519,46633,46708,46732,46792,46876,46886,46890,46970,46987,46991,47128,47236,47390,47395,47416,47426,47498,47513,47518,47532,47552,47610,47656,47659,47663,47682,47723,47729,47840,47870,47942,48147,48234,48277,48361,48384,48403,48404,48411,48439,48452,48475,48505,48509,48608,48630,48643,48674,48693,48695,48715,48763,48802,48808,48890,48978,48987,49023,49029,49045,49048,49070,49087,49110,49115,49122,49131,49139,49229,49286,49314,49343,49358,49373,49374,49398,49508,49514,49517,49528,49543,49585,49586,49609,49616,49648,49708,49710,49782,49787,49868,49911,49936,49979,50082,50424,50456,50472,50627,50729,50875,50975,51035,51065,51272,51335,51412,51632,51719,51780,51827,51830,51938,52083,52121,52176,52293,52341,52345,52354,52390,52507,52837,52956,53040,53116,53265,53269,53317,53386,53392,53751,53854,53899,54006,54150,54236,54389,54417,54470,54480,54560,54594,54644,54736,54738,54776,54821,54896,55032,55202,55207,55240,55352,55378,55515,55800,55805,55898,55977,56137,56144,56284,56335,56367,56484,56639,56715,56909,56924,57000,57066,57229,57459,57478,57639,57764,57766,58060,58218,58346,58435,58513,58588,58667,58732,58766,58914,58935,58978,59075,59112,59121,59217,59261,59344,59368,59476,59500,59581,59732,59800,59898,59935,59992,60199,60256,60306,60440,60474,60687,60688,60707,60771,60814,60830,60879,61274,61308,61494,61547,61562,61758,61786,61855,61887,61963,62100,62133,62186,62340,62528,62609,62651,62677,62808,62821,62856,62928,62961,63133,63164,63326,63688,63739,63850,63939,63978,64033,64067,64096,64115,64194,64215,64289,64302,64480,64562,64594,64610,64621,64627,64677,64721,64952,65050,65185,65268,65311,65335,65757,65780,65992,66083,66102,66116,66233,66260,66262,66307,66313,66316,66407,66574,66587,66606,66702,66783,66930,66991,67005,67055,67090,67180,67187,67254,67260,67316,67360,67369,67564,67570,67600,68231,68331,68423,68747,68750,68820,68980,69092,69111,69220,69288,69339,69379,69475,69479,69632,69688,69697,69720,69746,69796,69845,69849,70049,70251,70295,70320,70465,70486,70509,70522,70722,70731,70832,70840,70990,71043,71049,71130,71147,71214,71292,71296,71415,71471,71486,71549,71607,71622,71686,71781,71864,71885,71913,71934,72034,72154,72156,72267,72420,72585,72611,72823,72952,72980,73025,73175,73292,73357,73388,73470,73471,73608,73632,73797,74048,74324,74334,74345,74613,74853,74909,74966,75017,75020,75069,75498,75522,75544,75710,75854,75919,75967,76003,76073,76183,76197,76433,76564,77007,77036,77055,77095,77115,77147,77169,77185,77204,77209,77213,77239,77298,77428,77560,77587,77590,77616,77693,77757,77770,77834,77926,78032,78086,78110,78285,78557,78585,78587,78751,78787,78804,78954,79007,79049,79098,79141,79193,79260,79314,79404,79574,79578,79620,79691,79724,79787,79808,79860,79886,79942,80115,80174,80282,80307,80359,80459,80517,80542,80561,80650,80663,80684,80745,80817,80954,80975,81027,81056,81308,81686,81865,81877,81987,82030,82086,82334,82351,82383,82413,82584,82593,83121,83180,83244,83348,83505,83580,83599,83648,83682,83788,83950,84021,84052,84075,84219,84226,84296,84310,84422,84475,84483,84523,84570,84571,84573,84596,84620,84629,84675,84690,84850,84924,84953,85008,85017,85139,85210,85223,85278,85301,85475,85528,85570,85728,85824,85838,85841,86148,86208,86250,86294,86332,86335,86359,86374,86390,86421,86456,86793,86794,86925,87339,87390,87538,87539,87629,87688,87718,87764,87865,87932,87962,88142,88150,88211,88239,88258,88321,88449,88650,88698,89013,89025,89064,89109,89115,89137,89214,89253,89335,89341,89363,89470,89518,89524,89536,89548,89559,89752,89873,89937,90011,90168,90174,90229,90308,90311,90336,90359,90363,90374,90378,90403,90498,90518,90542,90557,90619,90692,90710,90731,90784,90851,90899,90932,90956,90962,91119,91278,91331,91354,91383,91625,91693,91730,91849,91904,91905,91912,91976,92049,92291,92325,92462,92735,92819,92925,92927,92957,92962,92980,93094,93159,93392,93422,93434,93623,93670,93694,93845,94183,94507,94585,94603,94639,94728,94749,94868,94999,95010,95040,95059,95368,95406,95534,95544,95585,95661,95683,95745,95881,95917,95986,96016,96029,96140,96166,96206,96250,96312,96394,96407,96585,96604,96702,96875,96957,97021,97272,97451,97528,97554,97600,97657,97721,97749,97854,97863,97882,97899,97998,98071,98167,98269,98436,98470,98494,98522,98595,98680,98776,98858,99157,99235,99237,99242,99425,99464,99505,99576,99593,99624,99709,99720,99828,99872,99874,99944,99996",
          "1,279,342,372,411,442,495,529,536,570,604,619,671,682,819,838,935,941,976,1001,1329,1350,1356,1399,1400,1580,1596,1708,1871,2277,2607,3117,3271,3831,3919,4158,5033,5095,5142,5170,5261,5342,5355,5357,5389,5478,5616,5803,5882,6020,6055,6105,6119,6333,6488,6537,6539,6612,6656,6670,6695,6718,6796,6823,6839,6853,6995,7039,7196,7211,7242,7342,7371,7464,7481,7515,7545,7566,7962,7985,8052,8138,8330,8389,8531,8698,8720,8813,8825,9012,9021,9059,9165,9200,9241,9256,9321,9392,9439,9465,9557,9564,9863,9932,9956,9991,10015,10103,10249,10263,10293,10527,10669,10701,10799,10824,10940,11028,11041,11057,11405,11433,11501,11771,11829,11864,11919,12311,12322,12409,12455,12539,12576,12600,12643,12688,12848,12972,12992,13047,13107,13181,13195,13275,13554,13579,13623,13880,14347,14437,14488,14552,14594,14596,14644,14830,14889,14983,14999,15159,15189,15366,16105,16219,16457,16725,17162,17269,17411,17931,18007,18053,18169,18465,18777,18794,19134,19683,19947,20081,20096,20653,20676,20709,20768,20783,21454,21496,21663,22026,22232,22255,22790,22951,23118,23141,23192,23702,24481,24714,24794,24817,24856,24946,24996,25049,25068,25250,25608,25635,26206,26624,26850,26866,26938,27218,27257,27479,27639,27761,27805,27871,27880,28175,28224,28291,28452,28510,28696,28699,28849,28892,29019,29252,29585,29601,29669,29837,29880,29963,31532,32028,32074,35827,36016,36457,36493,36806,36810,37136,37199,37434,37578,37617,37679,37803,38175,38422,38436,38510,38681,38928,39046,39263,39545,39666,39866,39883,45369,45431,45497,45783,45873,45876,45928,45992,50173,50197,50275,50347,50350,50364,50421,50510,50550,50581,50600,50760,50818,50946,51027,51097,51306,51317,51337,51465,51770,51805,51917,52014,52070,52072,52095,52270,52434,52575,52626,52703,52717,52825,52830,52848,52849,52872,52951,52987,53056,53243,53429,53445,53542,53636,53739,53747,53812,53913,53964,54079,54125,54183,54379,54493,54603,54632,54661,54711,54928,54931,54957,54963,55003,55139,55144,55271,55459,55479,55660,55753,55886,55915,56141,56326,56481,56486,56510,56589,57078,57243,57297,57435,57458,57636,57655,57788,57834,58079,58133,58136,58143,58207,58223,58263,58309,58510,58568,58628,58763,59012,59019,59353,59375,59382,59414,59561,59859,60024,60046,60129,60160,60250,60388,60470,60551,60573,60658,60743,60794,60908,60932,60979,61071,61085,61091,61122,61137,61149,61311,61316,61339,61486,61579,61647,61661,61678,61755,61816,61836,62142,62175,62180,62244,62283,62513,62628,62732,62761,62969,63059,63208,63214,63261,63341,63344,63478,63583,63590,63680,63732,64041,64119,64163,64203,64255,64491,64601,64650,64691,64698,64730,64974,65095,65470,65541,65931,65974,66017,66038,66068,66091,66243,66269,66287,66305,66322,66324,66415,66434,66469,66517,66599,66661,66664,66787,66796,66876,67258,67311,67337,67341,67449,67468,67541,67562,67727,67874,67908,67922,68021,68024,68067,68203,68316,68415,68522,68562,68579,68716,68878,68923,68943,69154,69170,69197,69209,69214,69240,69242,69291,69320,69350,69357,69361,69464,69580,69627,69647,69812,69822,69840,69878,69947,70079,70113,70140,70154,70176,70188,70431,70455,70553,70566,70611,70655,70853,70911,70978,71010,71213,71217,71230,71253,71255,71449,71690,71750,72055,72058,72073,72077,72084,72210,72228,72323,72405,72616,72619,72681,72734,72765,72911,72989,73118,73186,73289,73347,73444,73687,73713,73736,73849,73951,73982,73996,74042,74286,74407,74443,74453,74490,74542,74594,74700,74774,74846,75002,75078,75186,75198,75553,75714,75886,75890,75968,75986,76019,76033,76059,76105,76175,76243,76314,76414,76493,76566,76604,76664,76696,76786,76865,76962,76968,77030,77068,77233,77240,77282,77286,77318,77439,77529,77599,77715,77755,77825,77830,78054,78126,78220,78246,78302,78333,78347,78373,78380,78498,78592,78617,78701,78841,78851,78875,78886,79013,79140,79321,79447,79502,79731,79818,79900,79994,80175,80352,80404,80502,80577,80624,81036,81086,81088,81154,81536,81579,81695,81696,81795,81871,81965,82015,82109,82161,82169,82174,82250,82482,82489,82502,82587,82639,82692,82695,82742,82754,82894,83010,83200,83328,83456,83476,83479,83494,83517,83529,83548,83551,83667,83697,83748,83849,83984,84017,84148,84161,84210,84253,84366,84674,84751,84868,84903,85052,85205,85531,85548,85647,85776,85821,85914,86086,86134,86376,86581,86746,86933,86943,87010,87016,87135,87222,87355,87453,87603,87644,87682,87752,87761,87873,87886,88016,88098,88162,88209,88375,88431,88484,88510,88582,88593,88599,88646,88874,89130,89142,89343,89379,89426,89523,89619,89631,89738,89757,89929,90044,90214,90425,90473,90488,90554,90578,90588,90640,90802,90860,90990,91000,91111,91159,91178,91263,91462,91527,91550,91708,91740,91753,91855,91862,91983,92004,92141,92175,92257,92260,92304,92473,92567,92655,92691,92712,92716,92776,92862,92968,92972,93178,93328,93373,93443,93583,93598,93701,94167,94267,94326,94440,94664,94822,94904,95038,95049,95098,95557,95591,95660,95711,95735,95747,95786,95814,95837,95852,95857,95860,96009,96222,96368,96374,96522,96605,96681,96687,96740,96784,96798,96805,96816,96871,96886,96983,97192,97292,97324,97549,97680,97700,97751,97895,98013,98044,98107,98263,98279,98283,98284,98334,98443,98568,98640,98642,98664,98684,98755,98937,99184,99194,99302,99358,99386,99488,99595,99832",
          "102,149,178,217,460,496,1656,1669,1706,1728,1759,1792,1904,1917,1933,1935,1944,1983,1987,1995,2028,2062,2119,2139,2150,2189,2198,2224,2243,2266,2308,2330,2348,2393,2408,2425,2426,2500,2511,2530,2534,2667,2702,2720,2741,2793,2804,2823,2890,2968,3004,3009,3012,3022,3025,3026,3037,3052,3057,3060,3062,3064,3094,3101,3104,3111,3132,3174,3197,3208,3245,3295,3302,3318,3353,3411,3427,3441,3447,3451,3453,3454,3472,3499,3508,3516,3523,3525,3556,3558,3568,3578,3591,3604,3636,3640,3651,3657,3676,3748,3799,3865,3876,3885,3916,3929,3964,3968,3969,3970,3996,4012,4019,4037,4044,4059,4076,4088,4095,4113,4133,4138,4155,4162,4179,4186,4195,4209,4248,4250,4302,4315,4318,4338,4355,4361,4372,4397,4452,4537,4539,4587,4640,4665,4708,4722,4752,4780,4793,4805,4868,4883,4893,4924,4926,4943,4950,4951,4961,4964,4968,5076,5313,5580,5605,5755,5766,6159,6275,6327,6342,6352,6476,6617,6647,6767,6805,6890,7088,7093,7102,7136,7147,7221,7320,7374,7425,7568,7600,7601,7657,7684,7691,7790,7831,7968,8011,8121,8218,8222,8285,8317,8350,8404,8422,8458,8467,8506,8537,8622,8729,8769,8784,8820,8823,8835,8848,8862,8865,8896,8920,8951,8962,8995,9034,9083,9166,9191,9269,9325,9393,9466,9613,9659,9687,9692,9851,9869,9952,9957,9972,10034,10134,10255,10309,10319,10422,10428,10445,10554,11004,11056,11123,11334,11355,11362,11374,11375,11544,11698,12196,12249,12296,12326,12359,12370,12516,12633,12685,12880,12994,13291,13592,13745,14050,14210,14232,14286,14319,14821,15117,15303,15329,15487,15714,15781,15813,15814,15903,15921,16067,16150,16196,16486,16533,16543,16694,16780,17244,17250,17293,17336,17355,17372,17399,17487,17514,17526,17550,17683,17734,17780,17867,17901,17905,17963,18100,18274,18415,18560,18565,18612,18632,18688,18837,18956,19053,19109,19222,19474,19554,19774,19837,19844,20054,20105,20242,20287,20341,20706,20937,20941,20951,21098,21100,21289,21370,21373,21409,21503,21650,21793,21804,21881,22067,22340,22432,22463,22472,22484,22779,22872,22879,22959,22999,23055,23252,23253,23344,23454,23461,23497,23632,23802,23891,23898,23948,24126,24160,24372,24472,24551,24564,24601,24633,24905,24945,25031,25286,25385,25510,25715,25791,25945,25984,26018,26067,26091,26468,26488,26534,26543,26560,26570,26585,26597,26610,26618,26665,26830,26852,26888,26942,26955,27012,27034,27070,27128,27164,27207,27363,27466,27474,27475,27492,27530,27564,27571,27576,27582,27586,27596,27607,27612,27642,27724,27731,27740,27770,27778,27794,27795,27811,27868,27960,27987,28038,28062,28078,28085,28094,28139,28141,28207,28286,28308,28341,28347,28414,28427,28464,28488,28489,28493,28531,28558,28568,28603,28607,28626,28653,28805,28817,28840,28856,28964,28980,28992,29009,29084,29119,29126,29143,29153,29156,29179,29208,29224,29264,29284,29289,29290,29372,29512,29529,29550,29563,29607,29628,29661,29682,29685,29689,29790,29819,29876,29884,29947,30021,30050,30079,30108,30137,30166,30195,30224,30253,30282,30311,30340,30369,30398,30427,30446,30456,30485,30514,30543,30601,30630,30688,30717,30746,30775,30804,30833,30862,30891,30920,30949,30978,31005,31007,31036,31065,31094,31152,31181,31210,31239,31268,31297,31326,31353,31355,31384,31413,31442,31471,31500,31515,31529,31558,31587,31616,31645,31674,31703,31761,31819,31848,31877,31906,31935,31964,31993,32022,32051,32080,32109,32138,32167,32196,32225,32254,32283,32312,32341,32370,32399,32428,32457,32486,32515,32544,32573,32602,32631,32660,32689,32718,32747,32776,32805,32834,32863,32892,32921,32950,32979,33008,33037,33066,33095,33124,33153,33182,33211,33240,33269,33298,33327,33356,33385,33414,33443,33472,33501,33530,33559,33588,33617,33646,33675,33704,33733,33762,33791,33820,33878,33907,33936,33965,33994,34023,34052,34081,34110,34168,34197,34226,34255,34284,34313,34342,34371,34400,34429,34458,34487,34516,34545,34574,34603,34632,34661,34690,34719,34748,34777,34806,34835,34864,34893,34922,34951,35021,35040,35042,35044,35075,35085,35110,35122,35152,35184,35194,35229,35235,35273,35297,35347,35421,35441,35492,35540,35599,35658,35731,35745,35759,35790,35804,35817,35849,35882,35920,35947,35958,35979,36038,36054,36101,36161,36189,36220,36291,36323,36327,36350,36367,36409,36481,36540,36562,36599,36628,36629,36663,36722,36752,36781,36852,36911,36929,36943,36970,37042,37063,37101,37160,37182,37192,37224,37342,37413,37472,37491,37493,37496,37500,37531,37603,37662,37686,37721,37737,37785,37823,37844,37871,37903,37974,37990,37999,38004,38033,38092,38164,38223,38247,38282,38298,38346,38376,38405,38464,38495,38535,38551,38594,38625,38653,38725,38784,38822,38843,38907,38925,38966,39025,39034,39051,39096,39112,39155,39186,39214,39286,39345,39404,39436,39468,39499,39527,39586,39608,39657,39681,39716,39775,39808,39847,39861,39869,39885,39906,39965,40126,40155,40184,40213,40242,40271,40300,40329,40358,40387,40416,40445,40474,40503,40515,40532,40561,40590,40619,40648,40706,40735,40764,40793,40822,40851,40880,40909,40938,40967,40996,41025,41054,41083,41112,41141,41170,41199,41228,41286,41344,41373,41402,41431,41460,41489,41518,41547,41802,42017,42046,42075,42104,42133,42193,42283,42310,42316,42358,42371,42381,42433,42438,42445,42497,42520,42572,42595,42757,42832,42843,42866,42907,42915,42933,42994,43024,43069,43117,43144,43171,43231,43239,43249,43306,43310,43340,43381,43397,43414,43425,43439,43468,43502,43618,43624,43705,43741,43780,43855,43862,43887,43904,43942,44017,44092,44136,44179,44189,44254,44329,44416,44491,44566,44580,44584,44620,44653,44668,44696,44728,44803,44814,44861,44890,44914,44916,44965,44996,45008,45040,45053,45113,45161,45222,45280,45353,45363,45410,45429,45442,45468,45472,45540,45597,45655,45702,45727,45784,45842,45856,45914,45971,45996,46042,46054,46065,46104,46117,46127,46157,46172,46183,46191,46206,46222,46240,46294,46308,46317,46328,46337,46406,46433,46438,46464,46475,46485,46497,46522,46536,46572,46578,46652,46665,46686,46688,46751,46767,46850,46854,46865,46868,46880,46897,47011,47033,47054,47086,47092,47118,47137,47182,47185,47186,47281,47287,47305,47315,47387,47388,47418,47424,47442,47463,47489,47562,47590,47598,47613,47669,47836,47845,47900,47935,47944,47945,47989,48019,48026,48030,48041,48044,48056,48074,48090,48139,48165,48183,48189,48213,48226,48310,48321,48330,48515,48517,48571,48573,48639,48647,48675,48683,48684,48686,48735,48757,48778,48798,48799,48805,48812,48825,48837,48864,48869,48876,48897,48901,48935,48937,48943,48950,48966,48975,48976,49019,49053,49123,49145,49194,49344,49368,49382,49489,49495,49512,49608,49630,49685,49692,49695,49701,49724,49747,49772,49781,49823,49824,49830,49839,49860,49908,49915,49924,49927,49950,50024,50118,50158,50259,50279,50301,50382,50412,50415,50454,50531,50641,50675,50717,50766,51096,51173,51320,51420,51601,51633,51651,51709,51784,51935,52034,52078,52135,52188,52191,52217,52227,52247,52261,52273,52539,52558,52645,52726,52782,52906,53212,53297,53311,53330,53336,53395,53566,53612,53707,54126,54157,54173,54198,54222,54371,54376,54510,54523,54694,54835,54854,55059,55064,55140,55189,55249,55277,55318,55443,55544,55571,55611,55752,55794,55839,55962,56072,56164,56190,56238,56305,56324,56385,56395,56407,56462,56475,56615,56692,56761,56839,56974,56996,57040,57139,57143,57147,57150,57347,57357,57451,57469,57473,57484,57699,57923,57972,58274,58482,58498,58523,58590,58673,58718,58781,58789,58802,58887,58900,58906,58996,59109,59364,59508,59758,59863,60080,60082,60186,60268,60382,60506,60585,60645,60661,60728,60851,61006,61107,61143,61198,61333,61367,61456,61584,61827,62006,62053,62077,62095,62230,62250,62253,62295,62372,62497,62500,62576,62622,62685,62689,62701,62778,62857,62894,62899,63148,63258,63275,63389,63417,63503,63595,63653,63805,63821,64128,64205,64226,64256,64343,64353,64591,64757,64883,64959,65120,65193,65196,65197,65273,65298,65340,65400,65424,65513,65557,65574,65689,65785,65838,65859,65874,65886,65889,65936,66075,66096,66168,66193,66200,66301,66463,66488,66512,66593,66693,66935,66950,66973,67017,67036,67079,67109,67169,67185,67281,67445,67537,67593,67622,67743,67747,67760,67774,67797,67806,67832,67911,68059,68271,68469,68845,68906,69128,69195,69257,69356,69409,69613,69750,70067,70481,70530,70547,70607,70706,70733,70749,70787,70841,70903,70927,70944,70988,71001,71113,71146,71277,71344,71447,71533,71794,71829,71830,71898,72035,72180,72353,72362,72367,72474,72506,72824,72898,72994,73152,73205,73400,73413,73414,73434,73454,73456,73490,73553,73619,73628,73646,73678,73714,73750,73877,73889,74035,74143,74221,74252,74279,74339,74531,74549,74605,74622,74632,74649,74759,74899,74914,74983,75044,75048,75054,75285,75308,75344,75352,75387,75535,75649,75721,75756,75768,75879,75894,76240,76306,76348,76415,76438,76443,76446,76474,76485,76543,77076,77130,77383,77389,77410,77506,77612,77613,77699,77705,77868,78021,78097,78142,78370,78417,78549,78653,78655,78677,78681,79169,79203,79292,79387,79487,79534,79727,79753,79881,79933,80327,80405,80408,80448,80560,80665,80713,80719,80726,80758,80793,80797,80926,80950,80967,80969,80996,81013,81041,81115,81198,81277,81337,81360,81529,81569,81682,81750,81762,81848,81915,81924,81931,81955,82164,82259,82354,82582,82786,83016,83025,83044,83161,83198,83234,83254,83327,83333,83419,83423,83425,83513,83650,83677,83792,83847,83936,84131,84186,84382,84402,84403,84493,84600,84923,85110,85180,85211,85339,85370,85485,85630,85689,85740,85766,85777,85822,85903,85919,85938,85997,86046,86066,86108,86132,86314,86341,86344,86472,86592,86686,87126,87134,87205,87214,87424,87428,87481,87574,87685,87694,88131,88137,88248,88249,88380,88395,88502,88665,88719,88877,88901,88925,89091,89118,89162,89295,89647,89786,89931,89979,89996,90025,90121,90246,90357,90382,90394,90496,90515,90652,90711,90848,90879,90987,90992,91018,91211,91217,91223,91266,91415,91598,91637,91780,91791,91798,91960,92053,92204,92232,92301,92308,92330,92519,92648,92703,92804,92918,93108,93209,93210,93263,93477,93737,93778,93864,93886,93893,93913,93925,94131,94172,94240,94427,94470,94558,94677,94683,94689,94829,94874,94931,95253,95485,95520,95532,95652,95749,95791,95849,96063,96086,96491,96528,96580,96673,96757,96879,96923,96964,96979,97003,97121,97173,97303,97343,97405,97407,97607,97645,97703,97910,97985,98047,98188,98243,98348,98370,98460,98533,98594,98705,98944,98966,99057,99326,99461,99468,99632,99753,99839,99921",
          "41,60,298,301,308,316,369,380,398,454,481,489,539,546,572,573,596,749,762,823,865,880,1067,1099,1102,1118,1151,1301,1303,1396,1486,1553,1774,1897,2024,2206,3743,4576,4970,5072,5085,5182,5461,5561,5628,5680,5843,5866,5884,5890,5897,6090,6130,6236,6314,6448,6569,6606,6613,6749,6774,6781,6897,6930,6989,6991,7003,7011,7015,7034,7144,7262,7402,7450,7527,7567,7635,7647,7653,7679,7712,7886,7935,7943,8013,8086,8108,8152,8207,8236,8338,8339,8670,8838,8846,8936,8948,9051,9101,9151,9234,9245,9282,9320,9379,9468,9809,9945,9948,10025,10032,10154,10158,10277,10345,10362,10593,10853,10919,10922,11008,11124,11178,11224,11316,11357,11573,11787,11832,11833,11845,11849,11939,11940,11945,12020,12022,12351,12772,13113,13433,14026,14189,14589,14600,14713,14953,15089,15446,16626,16871,17051,17232,17536,17783,18062,18252,18521,18855,18899,18912,19283,20174,20198,20293,20597,20668,20769,20794,20802,20879,20907,21015,21290,21338,21448,21630,21655,21736,21778,22635,22714,22845,23299,23351,23366,23568,23601,23705,24067,25153,25559,25566,25636,26356,26410,26777,26794,26879,26962,27094,27329,27460,28303,28702,28934,29120,29407,29543,29576,29868,29983,35041,35608,35636,36241,36491,37175,37234,37363,37696,37739,37855,37909,38137,38303,38320,38549,38758,39503,42160,42998,43051,43834,45871,45947,50042,50120,50255,50286,50289,50492,50518,50598,50672,50702,50706,50754,50755,50780,50785,50804,50845,51002,51039,51105,51368,51395,51424,51594,52003,52130,52144,52170,52198,52332,52447,52761,52804,53044,53085,53088,53107,53132,53302,53447,53502,53504,53570,53678,53745,53757,53810,53878,54134,54377,54426,54490,54573,54728,54758,54798,54874,55093,55158,55176,55183,55198,55329,55398,55687,55688,55862,55978,55983,56043,56068,56169,56319,56402,56452,56525,56617,56645,56663,56799,56838,56864,56917,57089,57226,57360,57361,57501,57505,57837,57924,57933,58116,58148,58201,58418,58420,58475,58544,58551,58674,58715,58834,58872,58917,58956,59078,59110,59127,59453,59501,59544,59614,59676,59774,59932,59961,60077,60096,60247,60301,60336,60500,60600,60672,60680,60750,60801,61173,61220,61378,61382,61471,61516,61649,61656,61665,61724,61835,61842,61904,62223,62235,62255,62317,62394,62449,62649,62715,62736,62782,62814,62898,62916,63009,63021,63038,63097,63135,63530,63581,63624,63720,63804,63946,64202,64423,64592,64616,64913,64940,65084,65132,65462,65544,65774,65975,66029,66054,66124,66258,66331,66519,66561,66667,66679,66804,66938,66988,67021,67217,67283,67605,67786,67814,67903,67914,68079,68134,68173,68176,68220,68267,68295,68363,68454,68499,68501,68521,68548,68564,68587,68618,68642,68646,68758,68823,68857,68872,68875,68945,68959,68992,69114,69139,69285,69517,69536,69741,69770,70055,70213,70230,70408,70460,70679,70748,70898,71023,71050,71053,71112,71124,71184,71192,71261,71298,71519,71528,71543,71576,71645,71716,71745,71812,71998,72148,72149,72215,72270,72273,72678,72721,72756,72955,73070,73078,73088,73156,73190,73258,73277,73541,73829,74011,74255,74268,74461,74484,74577,74675,74698,74738,74789,74883,74964,74967,74992,75213,75230,75294,75885,75909,75926,75988,76318,76460,76552,76553,76576,76599,76600,76627,76654,76722,76980,77174,77413,77519,77633,77638,77972,78053,78073,78079,78100,78167,78202,78205,78255,78364,78562,78609,78687,78698,78761,78790,79065,79170,79212,79250,79323,79745,79764,79806,79868,80008,80018,80019,80091,80220,80306,80323,80639,80706,80909,81032,81135,81401,81595,81687,81858,81921,81938,82198,82246,82359,82384,82385,82471,82561,82664,82697,82753,82871,82937,83175,83178,83278,83459,83495,83579,83609,83678,83743,83795,83949,83999,84044,84246,84503,84661,84764,84856,84913,85035,85064,85103,85129,85148,85228,85315,85343,85384,85422,85550,85590,85644,85856,85920,86242,86313,86427,86463,86521,86940,86959,87055,87064,87093,87125,87139,87154,87186,87285,87371,87423,87472,87474,87640,87703,87723,87743,87825,87876,88023,88139,88330,88373,88438,88457,88553,88718,88761,88888,88910,88919,89000,89398,89420,89451,89482,89529,89845,89948,89968,90014,90021,90056,90083,90096,90113,90127,90231,90368,90504,90535,90570,90750,90984,91162,91253,91451,91593,91676,91776,91966,91993,92166,92224,92283,92419,92461,92465,92499,92591,92672,92896,93032,93223,93271,93447,93545,93588,93624,93635,93709,93831,93912,93972,94018,94070,94185,94261,94349,94351,94368,94419,94428,94660,94731,94894,94914,95127,95144,95221,95227,95294,95347,95396,95413,95417,95472,95501,95556,95599,95659,95807,95831,95847,95858,95912,95913,95964,96036,96167,96376,96543,96576,96674,96870,97112,97166,97177,97270,97326,97330,97352,97521,97675,97713,97737,97744,97768,97773,97862,98020,98115,98123,98148,98390,98398,98414,98438,98547,98704,98768,98847,98862,98863,98880,98987,98994,99180,99542,99678,99846,99854,99979",
          "52,58,120,135,181,322,708,874,1181,1321,1384,1484,1509,1670,1838,1931,1938,1947,1948,1952,2006,2018,2019,2055,2121,2138,2184,2268,2373,2374,2418,2443,2451,2495,2512,2528,2531,2533,2576,2645,2735,2738,2740,2744,2790,2858,2885,2893,2896,2900,2911,2918,2934,2984,3020,3033,3043,3067,3108,3115,3188,3278,3404,3405,3417,3439,3464,3473,3490,3498,3524,3527,3546,3570,3577,3663,3685,3691,3711,3721,3772,3791,3815,3825,3826,3934,3936,3959,3975,3979,3981,3998,4036,4040,4049,4058,4084,4094,4108,4112,4152,4194,4230,4236,4246,4269,4283,4290,4297,4379,4474,4482,4487,4491,4496,4498,4501,4507,4540,4552,4553,4577,4590,4595,4629,4646,4659,4686,4745,4784,4830,4854,4857,4928,4940,5152,5289,5402,5493,5585,5594,5623,5833,5848,5873,5960,5983,5986,6066,6085,6092,6178,6196,6208,6261,6291,6300,6325,6347,6375,6416,6419,6467,6501,6567,6639,6733,6761,6898,6948,6984,7029,7059,7076,7110,7129,7177,7213,7225,7246,7251,7298,7303,7409,7442,7460,7465,7632,7677,7716,7786,7792,7840,7938,7997,8015,8045,8157,8265,8431,8438,8476,8533,8557,8564,8574,8591,8761,8764,8776,8817,8861,9044,9079,9094,9193,9255,9534,9635,9673,9727,9818,9894,9969,10109,10229,10259,10270,10390,10451,10560,11168,11219,11260,11305,11453,11484,11702,11925,11958,12000,12190,12342,12357,12656,13183,13314,13691,13732,13804,13863,14032,14070,14207,14287,14461,14510,14572,14822,14852,14855,15008,15056,15328,15342,15654,15816,16125,16339,16370,16416,16504,16682,16836,16892,16936,17061,17083,17138,17177,17278,17420,17442,17471,17481,17577,17641,17844,17877,17888,18133,18292,18321,18375,18575,18667,18973,19018,19069,19603,19654,19657,19741,19754,19813,19874,19907,20024,20367,20433,20522,20651,20758,20990,21086,21427,21501,21506,21784,21870,21889,21968,21985,22260,22280,22361,22380,22468,22525,23296,23323,23421,23919,23963,24147,24198,24488,24775,25239,25258,25400,25424,25503,25576,25725,26211,26352,26487,26498,26507,26533,26549,26601,26606,26607,26655,26738,26767,26771,26778,26831,26844,26992,27006,27011,27022,27036,27044,27097,27103,27130,27142,27151,27180,27230,27263,27331,27399,27446,27507,27512,27515,27533,27562,27590,27613,27614,27624,27632,27640,27697,27726,27755,27769,27853,27887,27924,27962,27963,27972,27997,28042,28057,28072,28098,28116,28151,28209,28222,28223,28236,28254,28255,28260,28301,28314,28343,28344,28407,28428,28438,28495,28498,28506,28513,28539,28564,28586,28595,28599,28629,28634,28636,28645,28663,28665,28722,28783,28797,28957,28968,29015,29083,29105,29116,29146,29166,29184,29196,29217,29218,29237,29238,29313,29336,29387,29448,29470,29507,29520,29536,29546,29557,29584,29587,29644,29645,29658,29659,29691,29695,29706,29726,29740,29754,29762,29764,29887,29895,29920,30018,30047,30076,30105,30134,30163,30192,30221,30250,30279,30308,30337,30366,30395,30424,30453,30482,30540,30569,30598,30627,30656,30685,30714,30743,30772,30801,30830,30859,30888,30917,30975,31004,31033,31091,31120,31149,31178,31207,31236,31265,31294,31323,31352,31381,31410,31439,31468,31497,31526,31555,31584,31613,31642,31671,31700,31729,31758,31787,31816,31845,31874,31903,31932,31961,31990,32048,32077,32135,32164,32193,32222,32251,32280,32309,32338,32396,32425,32454,32483,32512,32541,32570,32599,32628,32657,32686,32715,32744,32773,32802,32831,32860,32889,32918,32947,32976,33005,33034,33063,33092,33121,33150,33179,33208,33237,33266,33295,33324,33353,33382,33411,33440,33469,33498,33527,33556,33585,33614,33643,33672,33701,33730,33759,33788,33817,33846,33875,33904,33933,33962,33991,34020,34049,34078,34107,34136,34165,34194,34223,34252,34281,34310,34339,34368,34397,34426,34455,34484,34513,34542,34571,34600,34629,34658,34687,34716,34745,34774,34803,34861,34890,34919,34948,34977,35018,35032,35051,35070,35107,35138,35181,35230,35270,35344,35363,35418,35468,35489,35503,35537,35557,35560,35596,35655,35710,35728,35787,35809,35846,35917,35976,36035,36098,36158,36217,36238,36288,36311,36347,36406,36433,36438,36478,36537,36596,36660,36719,36740,36778,36849,36908,36967,37039,37098,37157,37169,37221,37255,37280,37339,37360,37410,37428,37469,37489,37528,37556,37600,37659,37718,37782,37841,37900,37971,38030,38060,38089,38161,38220,38279,38343,38373,38402,38461,38532,38591,38650,38686,38722,38781,38840,38904,38963,39022,39042,39060,39093,39152,39211,39283,39342,39401,39465,39524,39583,39654,39713,39754,39772,39844,39903,39962,40115,40144,40173,40202,40231,40260,40289,40318,40347,40376,40405,40434,40463,40492,40521,40550,40579,40608,40637,40666,40695,40724,40753,40782,40811,40840,40869,40898,40927,40956,40985,41014,41043,41072,41101,41130,41159,41188,41246,41275,41304,41333,41362,41391,41420,41449,41536,42006,42035,42064,42093,42122,42138,42153,42161,42182,42216,42241,42272,42347,42382,42399,42422,42469,42509,42558,42584,42626,42659,42682,42692,42707,42712,42746,42821,42896,42923,42983,43012,43015,43058,43133,43160,43165,43220,43295,43316,43332,43337,43370,43396,43472,43532,43570,43648,43694,43750,43769,43844,43878,43886,43931,43991,44006,44049,44063,44081,44133,44168,44223,44227,44243,44318,44352,44405,44480,44518,44555,44595,44613,44642,44717,44758,44792,44879,44954,45002,45029,45102,45150,45193,45211,45269,45342,45399,45436,45457,45492,45512,45529,45552,45586,45644,45668,45716,45740,45756,45773,45831,45852,45903,45960,46122,46192,46279,46293,46358,46454,46471,46541,46603,46709,46737,46804,46815,46860,46918,46923,46928,46943,46957,46960,47003,47043,47062,47073,47208,47231,47286,47289,47290,47294,47404,47407,47482,47632,47646,47672,47677,47686,47689,47741,47745,47753,47768,47777,47784,47823,47826,47833,47898,47918,47988,48048,48098,48122,48127,48190,48202,48237,48270,48284,48295,48301,48307,48341,48345,48354,48375,48415,48421,48464,48480,48494,48496,48508,48619,48633,48650,48849,48874,48918,48944,49158,49184,49203,49221,49306,49318,49352,49376,49410,49447,49479,49533,49561,49573,49579,49611,49704,49738,49751,49774,49806,49822,49840,49845,49866,49876,49898,49920,49938,49970,49980,50039,50141,50227,50464,50522,50750,50810,50841,50919,50963,51244,51300,51448,51498,51578,51602,51690,51724,51974,52136,52189,52195,52351,52426,52442,52679,52685,52725,52753,52788,52792,52801,52970,53091,53293,53362,53363,53515,53663,53666,53850,53888,53895,53988,54030,54055,54075,54177,54208,54225,54282,54706,54774,54839,54932,54935,54997,55066,55087,55231,55299,55348,55428,55462,55627,55734,55833,55854,55891,55948,55997,56069,56100,56207,56268,56312,56463,56477,56516,56545,56638,56643,56667,56694,56714,56921,56930,57023,57135,57137,57186,57258,57318,57622,57654,57701,57772,57816,57819,57875,57970,58065,58089,58351,58529,58561,58636,58734,58806,59098,59119,59186,59212,59220,59254,59363,59378,59492,59542,59598,59642,59643,59664,59670,59688,59764,59867,59916,59947,59972,60113,60154,60158,60313,60392,60561,60578,60592,60606,60782,60797,60911,60957,61031,61113,61398,61546,61568,61574,61589,61633,61787,61808,61810,61838,61848,61920,62114,62147,62222,62331,62404,62412,62596,62702,62844,62845,62951,62989,63032,63043,63065,63199,63240,63251,63286,63738,63776,63807,64011,64081,64120,64201,64266,64341,64563,64647,64742,64748,64792,64800,64878,64895,64954,64971,64976,65043,65051,65357,65386,65466,65532,65567,65639,65660,65756,65771,65878,65984,66131,66151,66344,66349,66543,66584,66626,66730,66803,66834,66840,66883,67097,67195,67201,67207,67245,67282,67300,67366,67763,67890,68031,68160,68191,68202,68257,68391,68593,68662,68673,68884,68910,69038,69271,69299,69371,69466,69515,69548,69722,69791,69879,69913,70018,70036,70039,70112,70141,70220,70307,70375,70427,70492,70562,70669,70809,70873,70907,71159,71302,71402,71548,71726,71853,71968,72139,72294,72344,72375,72408,72469,72528,72538,72548,72583,72599,72635,72639,72726,72894,73131,73276,73374,73479,73500,73521,73644,73737,73810,73822,73883,73903,73915,74299,74419,74537,74922,74998,75120,75130,75203,75276,75304,75347,75364,75384,75405,75442,75445,75481,75494,75530,75659,75682,75875,76091,76095,76172,76265,76268,76329,76357,76445,76498,76540,76544,76791,76800,76810,76811,76839,76907,76989,77249,77251,77381,77393,77514,77728,77748,77801,77827,77848,77881,77899,78196,78231,78399,78449,78613,78733,78794,78802,78868,78883,78891,78925,78999,79237,79294,79452,79509,79636,79653,79736,79850,79851,79859,79919,79996,80338,80350,80374,80381,80389,80406,80565,80636,80740,80911,80991,80999,81076,81124,81133,81169,81280,81385,81496,81585,81705,81722,81723,81752,81861,82220,82309,82373,82469,82524,82621,82755,82830,82895,83037,83188,83370,83652,83753,83782,83828,83843,83845,83867,83942,84018,84072,84179,84378,84660,84682,84739,84802,85245,85350,85358,85534,85727,85848,85871,85886,86002,86014,86032,86085,86110,86320,86408,86441,86606,86610,86637,86640,86895,86988,87082,87095,87118,87147,87150,87224,87357,87445,87558,87667,87813,87921,87930,88032,88054,88284,88300,88324,88429,88456,88483,88485,88556,88670,88805,88955,88986,89057,89065,89145,89260,89270,89421,89509,89542,89785,89828,89885,89934,89970,90054,90100,90150,90185,90221,90450,90500,90719,90948,91061,91110,91115,91123,91319,91403,91411,91558,91619,91714,91720,91761,91875,91935,91939,91997,92045,92119,92452,92563,92579,92588,92693,92753,92782,92783,92793,92812,92883,92923,92949,93000,93148,93186,93253,93277,93426,93432,93585,93615,93631,93675,93728,93899,94146,94229,94290,94299,94409,94607,94621,94735,94917,94933,95001,95007,95095,95159,95323,95377,95442,95494,95524,95573,95697,95709,95714,95725,95940,96042,96078,96172,96196,96203,96219,96297,96438,96598,96648,96787,96847,96986,97000,97055,97120,97123,97197,97335,97400,97425,97437,97469,97490,97584,97624,97873,97915,98070,98146,98180,98496,98514,98523,98608,98795,98811,98923,99011,99055,99118,99335,99486,99617,99719,99808,99847,99892,99967",
          "37,59,94,164,187,1629,1709,1785,1975,2013,2125,2132,2233,2380,2465,2632,2637,2996,3035,3216,3233,3361,3371,3422,3616,3727,3754,3833,4027,4046,4096,4134,4160,4216,4335,4368,4376,4782,4882,4979,4995,4996,6248,6981,8047,9442,9528,9537,9556,9632,9651,9670,9689,9708,9763,9776,9993,10012,10069,10107,10601,10620,10753,10829,10924,11095,11133,11152,11209,11228,11247,11304,11361,11370,11478,11494,11513,11532,11627,11630,11646,11741,11854,11855,11874,11884,11912,11950,11951,11969,11983,12026,12042,12045,12062,12102,12105,12140,12159,12254,12330,12368,12427,12444,12463,12474,12501,12534,12558,12577,12615,12653,12672,12681,12786,12843,12900,12938,12969,12995,13014,13090,13242,13261,13280,13356,13388,13432,13506,13527,13546,13565,13603,13625,13641,13660,13679,13698,13717,13755,13774,13793,13812,13850,13907,13926,13945,13986,14021,14059,14116,14173,14192,14211,14230,14249,14268,14382,14439,14553,14569,14629,14743,14800,14838,14895,14914,14971,14992,15009,15047,15123,15139,15161,15184,15218,15237,15275,15313,15332,15427,15465,15484,15498,15560,15598,15636,15712,15731,15750,15788,15807,15826,15845,15864,15883,15959,15997,16016,16035,16054,16073,16168,16187,16206,16301,16320,16358,16377,16415,16453,16472,16503,16529,16548,16605,16681,16708,16731,16736,16738,16757,16795,16814,16833,16835,16890,16928,16946,17004,17042,17080,17137,17156,17169,17175,17194,17270,17308,17327,17365,17441,17460,17479,17498,17539,17555,17593,17650,17726,17745,17802,17821,17840,17859,17878,17897,18011,18023,18030,18087,18106,18125,18182,18220,18239,18247,18258,18296,18315,18334,18353,18372,18391,18410,18429,18448,18467,18470,18505,18515,18581,18600,18676,18752,18828,18847,18885,18914,18923,18958,18961,18980,19037,19075,19094,19132,19151,19170,19207,19208,19227,19265,19284,19307,19341,19360,19379,19398,19436,19455,19463,19484,19550,19560,19569,19645,19664,19702,19721,19759,19797,19835,19854,19873,19892,19911,19930,19949,19978,19987,20006,20025,20042,20063,20082,20139,20158,20166,20177,20215,20234,20291,20310,20329,20348,20386,20405,20424,20443,20462,20500,20557,20595,20614,20652,20671,20728,20747,20766,20804,20823,20880,20918,20950,20956,21013,21051,21073,21146,21165,21184,21199,21222,21241,21291,21317,21336,21374,21393,21412,21431,21450,21469,21488,21545,21564,21602,21613,21621,21659,21707,21716,21735,21754,21811,21868,21906,21925,21963,21982,22000,22001,22020,22077,22128,22134,22172,22210,22229,22237,22267,22305,22343,22362,22381,22400,22438,22457,22476,22495,22570,22571,22609,22666,22685,22704,22715,22742,22761,22780,22856,22894,22913,22933,22989,23008,23046,23084,23103,23160,23198,23208,23217,23236,23266,23312,23350,23388,23426,23445,23464,23502,23521,23540,23578,23616,23635,23654,23673,23675,23711,23730,23768,23825,23844,23882,23920,23935,23939,23958,23977,23996,24015,24034,24053,24091,24110,24129,24167,24186,24224,24243,24281,24319,24395,24414,24490,24528,24547,24566,24604,24623,24642,24696,24699,24718,24737,24756,24813,24815,24832,24851,24870,24908,24928,25003,25016,25022,25041,25098,25117,25177,25212,25231,25269,25288,25307,25364,25383,25421,25440,25482,25497,25500,25516,25535,25554,25630,25649,25685,25687,25706,25744,25763,25839,25854,25858,25877,25896,25915,25934,25953,25991,26010,26048,26086,26124,26160,26175,26181,26200,26238,26257,26421,26447,26466,26485,26504,26523,26656,26713,26731,26732,26808,26826,26865,26868,26979,27112,27121,27150,27188,27199,27238,27275,27289,27321,27333,27369,27376,27462,27483,27520,27568,27673,27682,27691,27722,27796,27814,27929,27948,28024,28112,28138,28157,28256,28356,28385,28518,28575,28670,28746,28781,28879,28930,28948,28955,28960,29069,29079,29088,29135,29145,29331,29337,29365,29373,29403,29483,29592,29639,29704,29715,29753,29904,29919,29939,29962,29982,30028,30057,30086,30115,30144,30173,30202,30231,30260,30289,30318,30347,30376,30405,30434,30463,30492,30550,30608,30637,30666,30695,30724,30753,30811,30840,30869,30898,30927,30956,30985,31014,31043,31072,31101,31130,31159,31188,31217,31246,31275,31304,31333,31362,31391,31420,31449,31478,31507,31536,31565,31594,31623,31652,31681,31710,31739,31768,31797,31826,31855,31884,31913,31942,31971,32000,32029,32087,32116,32145,32174,32203,32232,32261,32290,32319,32348,32377,32406,32435,32464,32493,32522,32551,32580,32609,32638,32667,32696,32725,32754,32783,32812,32841,32870,32899,32928,32957,32986,33015,33044,33073,33102,33131,33160,33189,33218,33247,33276,33305,33334,33363,33392,33421,33450,33479,33508,33537,33566,33595,33624,33653,33682,33711,33740,33769,33798,33827,33856,33885,33914,33943,33972,34001,34030,34059,34088,34117,34146,34175,34204,34233,34262,34291,34320,34349,34378,34407,34436,34494,34523,34552,34581,34610,34639,34668,34697,34726,34755,34784,34813,34842,34871,34900,34929,34958,34987,35028,35071,35078,35117,35191,35280,35354,35428,35444,35499,35547,35606,35631,35665,35797,35828,35856,35883,35927,35986,36045,36108,36123,36168,36227,36237,36298,36357,36416,36488,36547,36606,36614,36670,36683,36729,36788,36797,36805,36859,36918,36977,36991,37049,37108,37119,37167,37231,37290,37349,37376,37420,37440,37445,37479,37492,37538,37610,37669,37728,37792,37851,37910,37932,37981,38040,38099,38107,38171,38230,38289,38309,38353,38412,38471,38489,38498,38542,38545,38601,38615,38660,38732,38752,38791,38850,38865,38914,38926,38973,38987,39032,39103,39105,39162,39221,39293,39352,39383,39411,39475,39534,39593,39664,39723,39748,39782,39854,39913,39928,39972,40122,40151,40180,40209,40238,40267,40296,40325,40354,40383,40412,40441,40470,40499,40528,40557,40586,40615,40644,40673,40702,40731,40760,40789,40818,40847,40876,40905,40934,40963,40992,41021,41050,41079,41108,41137,41166,41195,41224,41253,41311,41340,41369,41398,41427,41456,41485,41514,41543,42013,42042,42071,42100,42129,42145,42158,42189,42206,42211,42226,42279,42287,42317,42354,42368,42397,42401,42429,42465,42516,42591,42666,42753,42828,42858,42874,42903,42924,42990,43028,43065,43140,43191,43227,43302,43346,43349,43377,43399,43416,43438,43464,43517,43539,43587,43614,43672,43701,43743,43776,43802,43819,43821,43851,43879,43938,43957,44013,44044,44088,44151,44156,44175,44210,44250,44274,44285,44289,44325,44412,44487,44506,44562,44626,44649,44724,44799,44867,44886,44910,44961,44987,44991,44998,45036,45068,45109,45157,45218,45276,45296,45349,45377,45406,45464,45515,45536,45593,45651,45723,45780,45838,45910,45930,45967,46015,46045,46047,46074,46099,46100,46101,46130,46158,46176,46204,46239,46311,46335,46351,46379,46391,46403,46428,46445,46450,46515,46538,46558,46629,46655,46699,46705,46713,46750,46752,46766,46768,46778,46901,46940,46977,46985,47007,47052,47067,47205,47206,47214,47243,47261,47268,47376,47380,47382,47414,47488,47497,47500,47520,47542,47545,47565,47571,47601,47665,47691,47747,47803,47847,47866,47887,47902,47946,48104,48111,48114,48121,48145,48174,48199,48206,48212,48214,48247,48249,48282,48286,48311,48376,48386,48410,48431,48497,48530,48536,48575,48587,48662,48667,48762,48767,48800,48806,48896,48909,48979,48993,49022,49069,49081,49082,49108,49147,49171,49204,49249,49265,49269,49276,49284,49303,49317,49338,49341,49379,49406,49452,49453,49552,49584,49642,49660,49670,49674,49694,49697,49753,49792,49793,49834,49858,49897,49909,49935,49956,50007,50084,50122,50166,50360,50361,50375,50509,50680,50695,50753,50869,50890,50967,51106,51274,51293,51321,51401,51473,51521,51590,51643,51982,52246,52398,52565,52623,52821,52835,52868,52873,52882,52975,53147,53196,53248,53255,53277,53308,53364,53584,53614,53633,53646,53657,53688,53738,53748,53831,53881,54002,54155,54274,54452,54476,54564,54588,54613,54704,54726,54893,54960,55091,55115,55123,55188,55255,55417,55432,55472,55512,55531,55607,55738,55766,55776,55792,55843,55897,55939,55968,56121,56176,56303,56381,56556,56568,56625,56710,56713,56734,56818,56863,56880,56986,57103,57120,57140,57158,57219,57365,57495,57511,57558,57683,57728,57790,57814,57817,57846,57917,57930,57946,57958,57993,58021,58097,58161,58180,58272,58383,58467,58655,58660,58693,58700,58720,58727,59011,59283,59369,59383,59419,59433,59487,59679,59712,59820,59839,59943,59948,59950,59993,60038,60175,60267,60283,60700,60709,60726,60751,60758,60858,60861,60961,60976,60995,61118,61176,61194,61200,61212,61295,61336,61363,61389,61436,61473,61611,61618,61646,61781,61903,61937,61960,61981,61993,62033,62116,62149,62606,62781,62835,62865,63004,63280,63325,63547,63594,63597,63648,63675,63727,63765,63786,63924,64021,64059,64064,64086,64179,64195,64247,64378,64475,64541,64713,64858,65033,65056,65157,65345,65492,65550,65623,65714,65813,65840,65942,66058,66105,66170,66326,66332,66400,66435,66614,66617,66724,66742,66892,67054,67106,67156,67353,67529,67608,67715,67753,67769,67919,67926,68144,68159,68237,68429,68481,68486,68566,68571,68836,68918,68969,69241,69265,69341,69347,69467,69524,69572,69646,69648,69752,69801,69945,70034,70037,70401,70540,70550,70774,70808,70811,70906,70919,71088,71099,71223,71312,71382,71639,71703,71753,71799,71824,71887,72009,72030,72040,72113,72245,72258,72261,72301,72365,72393,72437,72465,72576,72641,72761,72840,72899,72931,72944,72946,72977,73050,73101,73154,73204,73250,73421,73558,73577,73747,73772,73835,73857,73988,74013,74049,74134,74277,74363,74369,74386,74477,74571,74604,74637,74718,74732,74743,74806,74819,74822,74885,75064,75074,75113,75119,75283,75372,75685,75690,75757,75895,75978,76041,76060,76076,76164,76288,76364,76370,76692,76712,76717,76808,76829,76886,76950,77017,77082,77144,77313,77333,77369,77380,77392,77710,77712,77716,77781,77875,77888,78139,78160,78175,78176,78198,78252,78324,78355,78444,78508,78639,78670,78772,78862,78864,78867,78931,79125,79145,79368,79398,79441,79463,79477,79480,79768,79931,79958,80040,80045,80112,80133,80159,80182,80281,80351,80370,80480,80515,80647,80723,80921,80923,81087,81285,81352,81419,81421,81491,81547,81652,81699,81772,81781,81932,81933,81946,81973,82104,82118,82121,82424,82443,82627,82632,82750,82840,82844,82922,82986,83021,83288,83294,83320,83387,83434,83730,84037,84046,84284,84326,84416,84561,84563,84733,84743,84774,84926,85065,85138,85171,85227,85277,85767,85809,85946,86101,86227,86321,86449,86659,86730,86791,86865,86891,86892,87120,87172,87268,87410,87448,87560,87658,87698,87727,87801,87804,87849,87904,88029,88081,88186,88232,88314,88344,88436,88514,88636,88673,88683,88749,88795,88822,88879,88969,89092,89099,89138,89146,89288,89545,89634,89639,89768,89904,89947,90015,90086,90087,90163,90226,90263,90330,90591,90756,90763,90866,90883,90888,90959,91098,91189,91209,91282,91348,91382,91654,91667,91675,91755,91903,92057,92089,92106,92155,92165,92259,92314,92321,92331,92492,92544,92610,92674,92700,92745,92858,93110,93171,93354,93387,93430,93451,93536,93554,93569,93600,93730,93739,93759,93927,94005,94234,94238,94333,94513,94522,94626,94742,94888,95061,95306,95325,95487,95626,95736,95737,95780,95813,95825,95856,95941,96014,96057,96119,96171,96174,96293,96388,96399,96460,96484,96524,96625,96894,96973,96988,97044,97062,97200,97240,97342,97364,97367,97460,97496,97859,97907,97979,98186,98271,98347,98681,98682,98695,98825,98864,98945,99031,99185,99263,99311,99473,99606,99703,99737",
          "16,253,278,367,371,402,426,455,490,492,532,582,691,697,699,744,756,808,852,886,910,939,945,962,1008,1028,1044,1061,1090,1186,1213,1215,1217,1242,1339,1372,1425,1451,1545,1546,1566,1600,1881,1918,2681,4064,5028,5048,5119,5505,5579,5607,5725,5736,5789,5812,5834,5871,6146,6424,6466,6493,6499,6525,6641,6642,6644,6802,6836,6894,6929,7105,7188,7200,7217,7271,7331,7493,7505,7530,7578,7630,7636,7742,7828,7965,8034,8044,8091,8159,8173,8220,8352,8410,8417,8443,8474,8578,8646,8717,8809,8834,8952,8979,9403,9419,9783,9973,10121,10157,10161,10260,10299,10308,10371,10456,10503,10992,11061,11248,11738,11814,11927,12071,12448,13259,13505,13883,14043,14910,15656,15724,15916,16756,17036,17725,19052,19070,19102,19439,20593,20919,20975,20989,21197,21490,21830,22585,24828,25670,25766,25976,26402,26804,27132,27145,27391,27733,27771,27774,27944,28381,28456,28460,29141,29351,29642,29927,29956,30468,32866,34515,36492,37259,37795,37935,38508,39916,40199,42250,43617,43940,44593,45041,45301,45703,47327,50047,50163,50170,50225,50501,50714,50747,50769,50874,50903,50913,50935,50951,50980,51042,51083,51098,51193,51233,51266,51364,51385,51534,52005,52050,52163,52229,52373,52430,52444,52604,52634,53035,53036,53096,53204,53339,53359,53433,53434,53470,53482,53779,53890,53894,53947,54326,54329,54394,54403,54848,55384,55465,55574,55605,55666,55681,55725,55804,55874,55959,56106,56123,56130,56133,56245,56264,56423,56460,56532,56544,56695,56935,57008,57028,57098,57166,57384,57457,57583,57756,57787,57858,57877,57882,58091,58101,58131,58186,58271,58321,58373,58380,58495,58656,58724,58742,58769,58805,58827,58909,58921,59142,59169,59189,59278,59480,59539,59590,59596,59660,59665,59757,59877,59883,59893,59894,59963,59965,60016,60039,60167,60202,60240,60255,60414,60447,60466,60478,60528,60651,60697,60715,60747,60778,60860,60972,61217,61257,61265,61292,61372,61432,61457,61512,61648,61719,61789,61940,61945,62207,62264,62311,62470,62660,62739,62800,62851,62869,62950,62960,62990,63013,63160,63263,63273,63481,63544,63563,63598,63789,63814,63982,64057,64415,64507,64512,64514,64731,65005,65301,65306,65343,65418,65477,65500,65552,65556,65609,65624,65664,65718,65755,65860,65884,65990,66023,66046,66123,66565,66825,67057,67144,67240,67403,67506,67553,67554,67585,67674,67729,67741,67825,67861,68009,68124,68168,68448,68453,68619,68790,69027,69084,69247,69370,69462,69507,69534,69734,69810,69813,69835,69951,69968,70111,70200,70238,70278,70341,70396,70499,70514,70609,70631,70692,70736,70783,70874,70938,71080,71139,71210,71233,71280,71450,71517,71588,71652,71661,71718,72042,72066,72209,72243,72281,72369,72373,72395,72397,72502,72593,72605,72618,72796,72810,72842,72877,72967,73161,73164,73177,73195,73268,73311,73342,73384,73389,73501,73629,73705,73744,73748,73776,74036,74131,74146,74185,74196,74295,74318,74341,74602,74635,74769,75072,75097,75176,75178,75214,75463,75479,75518,75627,75666,75694,75805,75949,76079,76116,76120,76124,76149,76258,76310,76418,76427,76468,76501,76519,76534,76579,76612,76649,76656,76705,76710,76784,76799,76919,77015,77091,77097,77182,77454,77508,77533,77545,77610,77628,77792,77880,77940,77993,78010,78020,78083,78133,78233,78491,78502,78569,78595,78660,78690,78747,78748,78897,78914,79025,79130,79187,79190,79241,79288,79449,79505,79571,79717,79743,79782,79820,79827,80103,80183,80185,80191,80264,80378,80452,80679,80828,80843,80854,81031,81110,81400,81474,81587,81603,81613,81677,81790,81791,81825,82010,82080,82223,82231,82236,82278,82344,82454,82563,82658,82665,82791,82806,82824,82897,82978,83131,83143,83223,83245,83307,83347,83558,83571,83610,83688,83894,84001,84016,84035,84190,84332,84547,84574,84677,84680,84696,84746,84956,85049,85074,85136,85221,85291,85318,85338,85492,85826,85844,86001,86017,86081,86216,86286,86530,86540,86770,86827,86931,86966,87052,87202,87208,87247,87366,87430,87465,87602,87635,87663,87757,87811,87942,88015,88046,88065,88144,88154,88310,88336,88406,88443,88498,88503,88763,88856,88921,89365,89460,89514,89544,89561,89780,89872,89888,89972,90034,90040,90046,90116,90139,90302,90309,90349,90367,90406,90407,90474,90565,90604,90665,90757,90789,90863,90918,91042,91127,91181,91258,91396,91459,91508,91514,91537,91698,91717,91744,91789,92092,92144,92258,92518,92561,92628,92639,92746,92824,92844,93227,93309,93400,93453,93462,93700,94044,94076,94152,94416,94571,94705,94717,94726,94751,94755,94839,94865,95023,95060,95063,95179,95234,95427,95449,95510,95515,95536,95594,95612,95674,95842,95887,95959,96017,96061,96199,96308,96425,96617,96710,96745,96748,96779,96780,96908,96936,96990,97002,97034,97037,97046,97092,97116,97199,97222,97237,97308,97422,97466,97474,97486,97663,97694,97777,97858,97861,97897,98006,98320,98375,98382,98422,98576,98649,98696,98803,98830,98832,98860,98943,98969,99015,99133,99217,99276,99279,99305,99351,99390,99480,99578,99760,99946",
          "8,11,280,300,336,347,440,449,457,471,547,555,594,603,609,611,639,710,810,824,863,934,955,969,988,1000,1015,1023,1127,1254,1265,1312,1362,1369,1449,1479,1500,1505,1512,1558,1561,1573,1631,2265,2548,3363,3369,3598,3862,4562,4885,4939,4977,5002,5074,5139,5183,5209,5429,5451,5507,5557,5563,5587,5702,5744,5883,5914,5968,6008,6048,6050,6094,6095,6128,6190,6209,6495,6637,6703,6708,6713,6721,6728,6794,6886,6887,6913,6924,7115,7122,7290,7307,7429,7511,7682,7794,7817,7932,7934,8076,8077,8179,8268,8342,8418,8433,8521,8579,8714,8827,8884,8901,8982,8998,9004,9148,9207,9311,9400,9416,9996,10124,10148,10202,10478,10794,10886,10911,11030,11174,11201,11202,11722,11743,11799,11818,15683,17973,19246,19335,21072,23458,23803,25696,25841,25900,26090,26978,27058,27378,27810,27859,28208,28346,28778,28853,29266,29473,31258,33499,34021,35016,35826,38356,38482,40111,41575,41606,41766,41946,42492,42564,43543,43646,45059,45198,45256,47115,49165,49362,49486,50130,50150,50176,50244,50321,50527,50576,50607,50914,50933,51122,51397,51451,51457,51733,51739,51792,51826,51855,51860,51874,52004,52038,52107,52274,52279,52310,52320,52754,52769,52994,53126,53300,53385,53519,53555,53565,53645,53650,53728,53760,53806,53822,53887,53943,53945,54054,54250,54267,54308,54372,54459,54650,54847,54973,55270,55453,55499,55506,55587,55601,55899,55942,56029,56062,56220,56250,56316,56345,56446,56521,56599,56804,56962,57046,57090,57231,57315,57355,57431,57487,57498,57518,57599,57623,57742,57753,57755,57903,58037,58240,58267,58291,58454,58637,58778,59007,59122,59185,59232,59348,59401,59455,59517,59708,59719,59824,60112,60298,60402,60434,60452,60612,60671,60685,60716,60717,60738,60855,60895,60925,60940,61028,61210,61244,61270,61277,61350,61399,61413,61455,61462,61503,61543,61680,61721,61806,61840,61871,61895,61956,62048,62090,62097,62107,62132,62155,62190,62294,62571,62671,62730,62788,62862,62917,63071,63086,63099,63112,63134,63172,63411,63468,63486,63509,63524,63587,63618,63721,63728,63766,63810,63854,64012,64020,64038,64232,64368,64384,64432,64435,64458,64738,64787,64918,65126,65218,65309,65368,65464,65507,65519,65526,65579,65652,65996,66014,66104,66148,66175,66350,66393,66437,66651,66721,66741,66754,66807,66849,66900,66981,66994,67038,67077,67105,67159,67213,67271,67309,67462,67494,67595,67687,67700,67739,67750,67969,67971,68076,68093,68114,68268,68370,68446,68632,68691,68699,68705,68765,69051,69070,69159,69167,69196,69308,69319,69353,69365,69402,69412,69429,69503,69529,69562,69628,69669,69718,69777,69814,69832,69900,69901,69937,70035,70075,70094,70096,70206,70232,70271,70447,70591,70794,70816,70847,70933,71102,71182,71297,71319,71381,71526,71540,71557,71608,71627,71658,71797,71808,71857,71860,71921,71944,71954,72234,72265,72306,72411,72637,72646,72687,72706,72708,72752,72951,72986,73028,73137,73223,73335,73457,73546,73613,73710,73864,73868,73963,73974,74014,74037,74089,74169,74192,74250,74465,74469,74544,74570,74607,74642,74705,74971,75052,75057,75215,75222,75284,75373,75397,75461,75477,75488,75561,75585,75588,75726,75906,75977,76067,76102,76246,76279,76469,76594,76642,76723,76725,76877,76961,77005,77013,77063,77123,77136,77294,77501,77547,77571,77642,77925,77959,77988,78013,78127,78155,78168,78187,78354,78375,78416,78418,78433,78478,78528,78611,78707,78822,78832,78848,78860,79040,79289,79348,79379,79467,79515,79560,79605,79783,79837,80167,80212,80290,80297,80298,80380,80420,80527,80609,80798,80925,80994,81054,81060,81069,81077,81098,81319,81379,81416,81464,81515,81571,81626,81703,81734,81800,81856,81876,81949,81951,82039,82100,82148,82204,82230,82256,82509,82543,82565,82573,82645,82702,82778,82798,82901,82936,82955,82994,83004,83032,83035,83050,83247,83258,83337,83360,83397,83680,83733,83734,83737,83919,83937,84022,84245,84281,84386,84414,84437,84482,84566,84646,84741,84822,84872,84889,84893,84978,85025,85051,85160,85235,85403,85627,85651,85829,85891,85967,86238,86345,86384,86454,86646,86826,86913,87020,87105,87409,87592,87594,87872,87950,88118,88146,88160,88244,88293,88462,88737,88776,88854,89002,89120,89147,89229,89259,89372,89434,89588,89775,89951,90009,90035,90082,90097,90170,90188,90205,90257,90266,90294,90373,90439,90533,90539,90540,90589,90662,90676,90717,90724,90761,90805,90815,90845,90913,91082,91126,91499,91615,91770,91779,91813,91831,91917,92000,92235,92385,92386,92490,92609,92632,92702,92719,92791,93003,93012,93081,93088,93155,93258,93512,93667,93718,93957,93976,93981,93988,94186,94200,94212,94218,94281,94324,94344,94398,94668,94720,95107,95232,95255,95447,95603,95640,95645,95734,95744,95771,95815,95925,95993,96005,96056,96198,96331,96364,96427,96465,96672,97069,97134,97191,97282,97489,97491,97684,97762,97877,97991,97994,98184,98301,98307,98458,98466,98509,98579,98714,98790,98828,98836,98874,98898,99288,99298,99385,99700,99724,99986",
          "3,258,333,352,355,361,363,385,393,412,458,467,494,562,566,579,686,725,779,806,834,914,923,956,1060,1080,1115,1201,1225,1268,1322,1324,1381,1414,1442,1482,1488,1554,1869,2229,2584,3014,3415,4991,5005,5056,5086,5088,5102,5132,5175,5251,5328,5335,5392,5484,5497,5516,5632,5708,5760,5776,5777,5880,5906,5999,6011,6115,6134,6258,6310,6386,6630,6849,6878,6934,6975,7078,7079,7082,7103,7140,7436,7631,7634,7673,7810,7812,7942,8113,8185,8262,8286,8376,8384,8459,8488,8508,8549,8693,8732,8737,8800,8829,8913,9084,9108,9192,9198,9322,9329,9354,9842,9848,9871,9936,10151,10173,10218,10357,10711,10943,11783,12334,12504,13130,14651,14993,15028,15245,17009,17181,18014,18052,18444,18636,18764,19401,19477,19530,19597,20057,20078,20380,20571,20748,21279,21334,21511,21719,22041,22543,22943,23618,24209,24736,24738,25107,26924,27717,28862,29749,29799,29808,29958,31058,31062,32247,34050,35120,36561,37020,39631,39881,40454,42198,44915,45711,45760,45901,45984,49010,50076,50240,50383,50449,50552,50553,50602,50621,50709,50802,50848,50928,51020,51159,51164,51176,51219,51235,51299,51343,51479,51510,51512,51606,51829,51873,51875,52287,52372,52375,52417,52516,52590,52723,52734,52822,52840,53010,53254,53275,53276,53342,53405,53498,53811,53852,54044,54070,54105,54110,54112,54128,54153,54370,54423,54937,55049,55065,55173,55331,55418,55466,55521,55523,55623,55699,55811,55994,56013,56038,56129,56163,56209,56334,56560,56597,56653,56726,56914,57017,57087,57328,57399,57482,57578,57647,57897,57916,57953,58126,58135,58185,58200,58368,58463,58474,58758,58855,59069,59206,59227,59265,59272,59279,59303,59339,59465,59474,59499,59601,59703,59773,59781,59933,60086,60168,60182,60187,60360,60430,60552,60571,60602,60621,60627,60745,60775,60799,60813,60817,60845,60904,60924,60992,61140,61169,61228,61245,61335,61355,61422,61438,61451,61463,61474,61522,61599,61742,61970,62005,62028,62125,62173,62208,62348,62580,62613,62644,62726,62825,63024,63166,63219,63241,63385,63433,63448,63508,63751,63785,63812,63958,63973,63976,64053,64105,64116,64148,64152,64167,64234,64248,64391,64654,64667,64752,64762,64780,64805,64890,64892,64898,64964,64973,65014,65068,65139,65315,65323,65655,65659,65687,65856,66045,66141,66289,66292,66380,66403,66412,66416,66658,66668,66744,66751,66851,66861,66976,67030,67173,67269,67277,67617,67638,67640,67659,67663,67899,67921,67963,67978,68094,68155,68161,68219,68255,68283,68313,68428,68460,68723,68851,68873,68936,68990,69013,69041,69099,69204,69219,69255,69571,69597,69604,69626,69649,69652,69662,69744,69775,69876,69936,70147,70153,70282,70299,70451,70614,70677,70826,70889,70949,70975,71196,71201,71245,71346,71364,71733,71841,71952,72222,72387,72510,72584,72647,72683,72781,72897,73011,73038,73184,73187,73229,73332,73499,73656,73753,73779,73913,74033,74313,74400,74475,74481,74519,74608,74688,74780,74820,74832,74889,74896,74959,74965,75177,75401,75515,75599,75687,75697,75717,75777,75797,75901,75979,75991,76032,76242,76353,76373,76413,76423,76638,76668,77157,77179,77227,77342,77453,77759,77777,77826,77836,77883,78087,78218,78271,78280,78317,78464,78630,78709,78710,78744,78765,78781,78831,79101,79143,79228,79271,79324,79499,79647,79759,79775,79884,79895,80180,80242,80305,80371,80414,80419,80422,80431,80507,80633,80708,80766,80814,80873,80902,81057,81186,81194,81366,81468,81662,81700,81789,81941,81981,82012,82173,82288,82376,82495,82576,82581,82586,82620,82626,82652,82663,82713,82790,82920,82932,82943,83030,83036,83041,83053,83069,83160,83189,83204,83282,83297,83326,83458,83639,83789,83813,83975,84002,84099,84109,84201,84263,84295,84301,84485,84582,84683,84811,84842,84846,84854,84860,84865,84990,84999,85007,85236,85256,85334,85483,85495,85538,85567,85623,85683,85701,85704,85738,85744,85753,85773,85940,86027,86386,86426,86442,86563,86636,86643,86676,86820,86869,87058,87099,87170,87290,87469,87482,87494,87552,87665,87772,87839,87926,87958,88035,88051,88108,88138,88157,88196,88269,88285,88332,88460,88470,88505,88567,88757,88777,88779,88818,88909,89112,89126,89158,89210,89239,89286,89309,89318,89658,89673,89699,89739,89804,89914,89985,90006,90041,90073,90239,90247,90527,90595,90689,90958,91022,91040,91204,91275,91365,91492,91582,91599,91617,91621,91652,91742,91763,91857,91941,91948,92016,92018,92091,92129,92184,92202,92225,92231,92266,92286,92307,92338,92358,92379,92504,92541,92598,92695,92764,92884,92904,92911,92933,93037,93068,93113,93127,93207,93312,93389,93437,93499,93590,93656,93659,93671,93689,93708,93745,93796,94135,94171,94209,94259,94492,94511,94590,94696,94740,94765,94810,94889,95117,95141,95162,95338,95518,95666,95740,95789,95802,95811,95836,96272,96286,96511,96621,96684,96929,96946,96951,96966,97007,97049,97257,97268,97365,97398,97440,97544,97621,97655,97681,97696,97698,97723,97754,97977,97990,98151,98158,98201,98209,98244,98268,98297,98360,98464,98587,98604,98620,98805,98835,98867,98927,99032,99052,99070,99138,99420,99522,99638,99819,99919",
          "141,273,289,348,389,409,413,430,433,472,473,514,530,613,630,700,784,814,832,861,866,868,1010,1069,1074,1121,1167,1193,1222,1331,1333,1496,1520,1617,1621,1671,1894,2090,2635,2778,2826,3242,3650,4788,4910,5049,5067,5213,5218,5257,5292,5293,5334,5353,5374,5401,5518,5613,5650,5651,5660,5712,5818,5823,5912,6017,6071,6117,6138,6239,6471,6479,6755,6902,6936,7008,7087,7091,7143,7170,7230,7275,7277,7289,7302,7347,7367,7584,7698,7718,7722,7761,7910,7988,7994,8105,8253,8292,8617,8654,8777,8815,8869,8886,8887,8968,9029,9160,9222,9760,9811,9882,10271,10291,10462,10538,10570,10591,10638,10730,10823,10936,10937,11169,11393,11429,11744,13826,14442,14496,16620,16958,17438,17747,18009,18223,19451,19601,20231,21085,21247,21560,22035,22093,23346,23462,23611,23668,23717,23901,24008,24072,25290,25668,25826,26143,26311,26404,26691,26714,26825,27870,27890,28424,28868,29044,29310,29538,29616,29789,30496,33434,35204,36549,36826,37283,37984,38066,38198,38651,38676,38794,38884,39768,41442,41756,42455,43652,45376,45414,45673,46576,47262,47551,48128,49791,50031,50178,50242,50335,50353,50459,50572,50688,50773,50825,50905,50907,50962,51109,51148,51331,51426,51487,51573,51627,51673,51764,51902,51930,51933,51953,52041,52519,52580,52662,52704,52744,52756,52810,52874,52959,52997,53004,53033,53065,53122,53257,53325,53367,53402,53484,53558,53600,53642,53684,53691,53925,53950,54068,54076,54139,54165,54199,54221,54246,54471,54477,54566,54666,54676,54723,54781,54786,54819,54820,54830,55098,55203,55313,55458,55528,55603,55631,55755,55763,55957,56152,56199,56217,56277,56323,56389,56422,56451,56573,56697,56743,56829,56920,56987,57111,57430,57479,57725,57929,57955,58107,58320,58471,58488,58563,58584,58751,58877,58892,58925,58929,58943,59064,59153,59196,59373,59402,59538,59631,59727,59766,59805,59913,59991,60153,60254,60265,60271,60442,60514,60566,60701,60836,60854,60898,61101,61248,61263,61283,61328,61368,61403,61419,61691,61715,61771,61790,61852,61861,61876,61879,61947,62040,62101,62254,62280,62406,62419,62468,62642,62684,62690,62719,63048,63075,63089,63347,63403,63424,63461,63819,63856,63886,64095,64198,64199,64213,64363,64382,64394,64454,64459,64518,64833,64907,64938,64992,65003,65004,65009,65044,65065,65067,65153,65269,65290,65316,65317,65318,65322,65387,65442,65628,65732,65742,65816,65911,65979,65998,66039,66082,66125,66182,66197,66562,66579,66659,66670,66723,66852,67161,67199,67235,67274,67334,67441,67446,67550,67630,67740,67842,67946,67984,67987,68074,68102,68248,68701,68727,68866,68912,68949,69086,69104,69198,69355,69396,69510,69543,69804,69811,69860,69865,69914,70148,70149,70174,70190,70221,70428,70452,70463,70529,70714,70721,70765,70815,70992,71042,71118,71462,71499,71578,71600,71681,71838,71880,71931,72079,72085,72145,72398,72460,72482,72485,72537,72653,72679,72701,72738,72834,72943,72950,72973,73054,73104,73125,73145,73203,73255,73262,73314,73364,73370,73398,73435,73591,73717,73727,73743,73769,73833,73851,74019,74263,74264,74467,74647,74690,74812,74925,74939,74991,75191,75268,75339,75341,75433,75505,75559,75578,75661,75701,75711,75755,76276,76282,76367,76410,76436,76449,76503,76756,76943,77110,77150,77173,77181,77319,77502,77554,77618,77629,77663,77754,77775,77794,77857,77920,77934,78263,78301,78358,78424,78668,78697,78756,78768,78808,78946,79064,79078,79080,79119,79180,79411,79599,79726,79751,79891,79898,79914,80007,80203,80215,80481,80703,80714,80738,80831,80915,81084,81211,81226,81246,81248,81262,81321,81340,81508,81517,81533,81812,81863,81895,82013,82065,82517,82610,82828,82845,83250,83301,83325,83399,83421,83444,83469,83764,83969,84138,84209,84262,84464,84576,84692,84859,84907,84965,84983,85325,85423,85432,85646,85746,86069,86126,86130,86178,86503,86545,86653,86707,86743,86803,86929,86968,86987,86996,87114,87281,87291,87363,87412,87516,87533,87652,87679,87753,87803,87829,88140,88164,88294,88301,88476,88575,88674,88748,88797,88820,88873,88916,88970,88998,89359,89369,89491,89600,89799,89924,90153,90171,90186,90190,90306,90383,90516,90567,90592,90593,90624,90656,90678,90783,90839,90865,90909,90978,91095,91274,91316,91395,91423,91447,91449,91511,91608,91634,91745,91863,91911,91982,91988,92174,92216,92284,92366,92396,92449,92486,92534,92550,92781,92818,93046,93147,93154,93255,93310,93383,93531,93580,93647,93746,93753,93815,93943,94109,94110,94134,94444,94614,94624,94651,94746,94762,94788,94895,94918,95041,95186,95257,95334,95362,95634,95699,95798,95867,95896,96023,96052,96110,96160,96252,96315,96332,96398,96417,96487,96626,96680,96764,96904,96937,96958,97005,97231,97247,97260,97304,97458,97487,97566,97683,97758,97816,97846,98046,98163,98319,98321,98330,98473,98626,98688,98801,98902,98922,98928,98978,98986,98990,99147,99156,99213,99308,99427,99432,99503,99536,99582,99690,99815,99849,99873,99927,100003",
          "196,197,235,271,290,396,505,510,525,567,576,675,766,841,846,864,966,973,1041,1123,1157,1203,1251,1315,1347,1352,1357,1364,1366,1385,1389,1392,1393,1445,1475,1490,1610,1645,1790,1805,1817,1826,1858,1926,1964,1968,1997,2033,2045,2114,2116,2158,2283,2287,2337,2338,2347,2367,2397,2424,2453,2472,2529,2532,2549,2551,2561,2594,2641,2678,2684,2772,2808,2833,2873,2894,2901,2910,2926,2932,2935,2958,2960,2974,2980,2981,2997,3001,3038,3127,3135,3145,3151,3230,3251,3268,3306,3331,3350,3395,3410,3414,3429,3440,3504,3565,3582,3586,3588,3614,3658,3679,3680,3710,3713,3759,3768,3771,3846,3856,3859,3877,3887,3908,3921,3933,3946,3960,4017,4043,4128,4142,4173,4207,4249,4272,4405,4426,4450,4454,4466,4480,4493,4522,4527,4541,4545,4546,4547,4560,4561,4653,4661,4662,4669,4678,4697,4698,4705,4730,4742,4817,4863,4876,4881,4895,4946,4960,4962,4994,5021,5118,5129,5250,5438,5631,5775,5805,5832,5851,5878,5885,5929,5957,6030,6042,6089,6174,6195,6198,6328,6330,6396,6462,6500,6624,6660,6685,6944,7116,7452,7471,7640,7747,7771,7847,7899,7952,7966,8023,8187,8224,8300,8312,8364,8420,8480,8489,8754,8818,8875,8910,8916,8935,8969,9041,9102,9106,9258,9268,9324,9348,9413,9520,9525,9531,9645,9726,9873,10046,10498,11027,11267,11385,11758,11860,11863,11923,12412,12630,12650,12910,13067,13346,13489,13514,13628,13682,13845,13967,13972,14420,14447,14486,14532,14625,14746,14779,15098,15458,15502,15541,15544,15713,15775,15881,15906,15934,16022,16203,16338,16510,16942,17556,17572,17936,17962,18303,18446,18591,18939,19310,19334,19546,20153,20159,20175,20272,20420,20912,20964,21044,21140,21576,21599,21760,21837,21857,21976,22405,22453,22567,22649,22699,22770,22798,22807,22893,22938,22996,23135,23229,23466,23612,23789,24124,24177,24318,24323,24471,24750,24940,24978,25059,25215,25390,25473,25508,25526,25573,25777,25828,26088,26132,26169,26225,26530,26547,26548,26550,26590,26662,26890,26996,27072,27122,27256,27420,27454,27489,27583,27617,27628,27893,27966,27983,28012,28021,28031,28047,28076,28077,28104,28118,28121,28146,28188,28319,28453,28570,28621,28638,28641,28650,28672,28767,28810,28827,28850,29029,29133,29334,29417,29632,29657,29712,30014,30072,30159,30217,30246,30309,30362,30388,30391,30449,30478,30565,30594,30710,30739,30768,30797,30826,30855,30971,31000,31087,31116,31174,31203,31216,31261,31348,31377,31406,31435,31464,31493,31551,31580,31609,31638,31667,31754,31812,31870,31899,31928,31957,31986,32015,32044,32160,32189,32276,32334,32421,32479,32508,32595,32624,32653,32682,32711,32740,32769,32798,32827,32856,32885,32914,33030,33117,33146,33204,33378,33465,33523,33552,33610,33639,33668,33697,33726,33755,33813,33900,33987,34045,34132,34161,34277,34306,34335,34364,34393,34431,34451,34509,34538,34567,34596,34625,34654,34683,34799,34828,34860,34886,34915,34944,34973,35062,35103,35193,35205,35217,35266,35295,35414,35485,35519,35533,35633,35651,35694,35724,35799,35813,35863,35913,35972,36031,36154,36178,36343,36364,36402,36423,36592,36656,36845,36883,36890,36904,36963,37217,37239,37276,37335,37362,37596,37655,37714,37896,37922,37967,38026,38216,38275,38339,38457,38483,38555,38587,38685,38718,38777,38900,39018,39047,39074,39089,39148,39207,39279,39397,39461,39483,39520,39579,39595,39601,39650,39709,39899,39958,40041,40125,40154,40183,40212,40241,40299,40328,40357,40386,40473,40531,40560,40763,40821,40850,40879,40966,40995,41024,41053,41082,41140,41169,41256,41285,41314,41343,41430,41459,41488,41546,42043,42045,42074,42132,42140,42192,42292,42311,42312,42332,42357,42432,42470,42519,42527,42530,42594,42669,42726,42756,42862,42906,42948,42993,43022,43143,43183,43254,43305,43380,43542,43569,43658,43661,43735,43779,43861,43869,43947,44057,44058,44091,44102,44178,44205,44253,44280,44328,44415,44490,44565,44601,44612,44652,44660,44680,44773,44802,44839,44889,44912,44942,44970,44981,45039,45112,45160,45221,45234,45279,45362,45409,45467,45487,45654,45726,45741,45841,45913,45970,46003,46046,46069,46088,46089,46115,46142,46156,46175,46214,46259,46275,46299,46363,46401,46533,46535,46570,46573,46577,46607,46639,46689,46710,46727,46797,46805,46857,46925,46973,46982,47070,47080,47155,47174,47192,47455,47459,47486,47554,47681,47688,47749,47787,47907,47932,47971,48188,48267,48292,48324,48326,48359,48387,48437,48443,48533,48593,48696,48775,48789,48884,48914,48915,49148,49163,49186,49195,49250,49277,49307,49419,49485,49578,49626,49665,49729,49735,49736,49746,49750,49814,49894,49905,49978,50046,50272,50307,50395,50500,50742,50871,50876,50952,50959,51016,51070,51151,51170,51296,51315,51702,51752,51962,51973,51992,52040,52063,52099,52220,52233,52315,52438,52487,52490,52512,52566,52628,52644,52647,52659,52667,53329,53673,53765,53889,53920,53954,54046,54061,54171,54368,54388,54413,54495,54537,54606,54618,54671,54677,54698,54746,54818,54863,54941,55042,55168,55239,55261,55267,55594,55790,55812,55819,55921,55975,56004,56034,56113,56161,56295,56337,56352,56384,56439,56453,56457,56494,56608,56627,56629,56646,56742,56745,56749,56806,56834,56884,56886,56960,56979,57003,57073,57138,57163,57167,57266,57363,57378,57517,57554,57574,57658,57672,57738,57818,57821,57825,57888,57898,58027,58141,58165,58191,58194,58217,58302,58484,58528,58533,58600,58624,58710,58822,58825,58830,58928,58980,59076,59090,59345,59389,59422,59577,59632,59680,59850,59887,59966,59974,60118,60318,60330,60332,60389,60425,60591,60604,60618,60632,60666,60670,60692,60756,60804,61032,61082,61141,61155,61267,61310,61370,61490,61534,61582,61610,61746,61791,61795,61800,61928,61938,61942,61958,62032,62181,62206,62225,62259,62409,62564,62700,62811,62888,62936,63072,63098,63120,63346,63390,63491,63517,63577,63667,63700,63708,63754,63755,63790,63791,63965,64123,64283,64335,64396,64403,64573,64696,64789,64803,64815,64828,64859,64861,64946,64948,65184,65224,65232,65348,65353,65569,65595,65729,65887,65958,65964,65989,66067,66107,66211,66306,66346,66386,66394,66431,66479,66483,66582,66685,66761,66763,66858,66959,66968,67263,67350,67354,67442,67518,67522,67604,67621,67670,67720,67770,67821,67869,67909,67964,68018,68110,68125,68278,68302,68309,68442,68633,68720,69307,69393,69528,69872,69999,70217,70337,70387,70406,70649,70675,70745,70757,70770,70972,70976,71047,71048,71136,71137,71143,71183,71216,71314,71388,71545,71979,72028,72080,72292,72471,72526,72627,72662,73009,73013,73110,73133,73178,73180,73253,73272,73445,73451,73477,73514,73680,73722,73755,73806,73848,74053,74153,74220,74223,74297,74725,74772,75094,75146,75201,75273,75295,75315,75385,75586,75646,75673,75754,75819,75872,75877,75998,76020,76098,76122,76420,76467,76780,76970,77089,77099,77387,77442,77550,77605,77667,77714,77808,77849,77886,78005,78157,78323,78381,78384,78393,78555,78573,78580,78588,78652,78675,78712,78731,78853,78890,78940,78957,79019,79242,79295,79312,79639,79682,79741,79769,79834,80033,80042,80084,80150,80221,80289,80331,80616,80710,80727,80775,81164,81257,81259,81305,81310,81333,81361,81386,81480,81493,81832,81944,81980,82005,82046,82060,82094,82235,82282,82378,82457,82506,82817,82879,82908,83081,83100,83230,83246,83507,83514,83519,83581,83622,83752,83831,83980,84005,84026,84041,84056,84093,84128,84132,84144,84188,84223,84283,84476,84500,84510,84605,84653,84655,84878,84919,84950,85082,85480,85700,85745,85786,85867,85872,85912,85942,85981,86093,86217,86325,86672,86822,86899,86937,87091,87119,87215,87259,87301,87306,87392,87488,87526,87785,87806,87816,87823,87910,87923,87941,87997,88026,88218,88270,88282,88299,88427,88557,88600,88617,88741,88773,88792,89098,89215,89274,89407,89439,89466,89479,89655,89659,89729,89801,90169,90244,90271,90296,90320,90324,90603,90620,90777,90785,90869,90920,90924,91001,91070,91108,91112,91117,91175,91256,91489,91671,91688,91713,91799,91802,91824,91876,91915,92180,92213,92223,92448,92470,92513,92528,92582,92649,92687,92789,92875,92946,93226,93318,93362,93474,93593,93706,93787,93856,93859,94078,94085,94283,94304,94329,94360,94553,94584,94782,94875,94883,94940,94990,95032,95219,95329,95392,95402,95404,95498,95512,95525,95561,95632,95763,95832,96118,96120,96302,96318,96393,96397,96449,96530,96559,96570,96606,96659,96853,96933,96965,96992,97164,97235,97267,97322,97397,97612,97630,97652,97671,97714,97788,97830,97865,97929,98023,98085,98100,98138,98402,98445,98455,98498,98503,98505,98614,98726,98773,98782,98866,99069,99105,99169,99258,99294,99601,99754,99810",
          "307,337,443,463,482,507,535,587,642,713,760,772,822,842,845,911,915,922,937,944,970,978,993,1002,1007,1034,1035,1057,1098,1101,1124,1137,1169,1226,1285,1293,1311,1344,1405,1495,1519,1537,1549,1870,1916,2260,2491,4323,4457,5034,5070,5143,5150,5166,5245,5323,5347,5376,5377,5403,5481,5489,5492,5606,5654,5686,5796,5872,5895,5924,6084,6222,6368,6428,6458,6514,6594,6666,6691,6765,6870,6966,7010,7021,7068,7119,7124,7131,7166,7239,7279,7296,7328,7404,7651,7778,7832,7877,7971,8096,8254,8397,8415,8471,8623,8797,8828,8830,8932,8942,8976,9019,9047,9137,9273,9306,9744,9887,10176,10266,10300,10304,10346,10398,10545,10629,10840,11034,11040,11074,11190,11299,11457,11616,11725,11776,12797,13437,13476,13513,13723,14131,14342,14897,14973,15593,15619,15773,17081,17193,17254,17933,17991,18887,19971,19981,20361,20791,21320,21848,22212,22760,23383,24431,25024,25210,26824,27029,27829,28310,28683,29353,33495,36458,36503,36515,37576,37794,38141,38698,39002,39309,39678,39824,39939,41087,43634,45419,45495,45544,47238,49478,50181,50262,50310,50357,50431,50564,50592,50636,50674,50700,50724,50893,51075,51165,51438,51478,51516,51570,51735,51771,51772,51799,51869,51913,51944,51960,52033,52124,52131,52230,52289,52349,52355,52361,52389,52768,52842,52875,52893,52923,53066,53099,53124,53160,53246,53263,53282,53331,53417,53426,53499,53512,53668,53727,53732,53804,53845,53976,53977,54205,54318,54400,54519,54536,54647,54656,54743,54936,55053,55165,55169,55206,55243,55305,55376,55441,55545,55551,55622,55787,55889,55927,56172,56232,56247,56347,56533,56600,56650,56813,56969,57036,57062,57175,57532,57562,57613,57619,57629,57695,57703,57774,57793,58183,58210,58233,58357,58907,58973,59237,59241,59392,59466,59531,59619,59635,59717,59730,60285,60337,60413,60631,60656,60949,61171,61260,61269,61301,61417,61479,61481,61557,61561,61605,61652,61720,61762,61769,61831,61962,62063,62088,62162,62301,62302,62395,62411,62495,62616,62756,62832,63055,63262,63293,63352,63436,63541,63629,63652,63748,63759,63870,63943,64023,64037,64183,64294,64297,64321,64360,64508,64743,64749,64797,64806,64837,64838,64840,64869,64870,64935,64983,64991,65021,65125,65239,65252,65539,65590,65631,65731,65741,65749,65781,66050,66176,66395,66467,66500,66563,66646,66660,66717,66781,66848,66856,66859,66963,67012,67022,67053,67170,67171,67176,67236,67451,67460,67676,67681,67849,68005,68072,68118,68140,68145,68148,68238,68277,68511,68578,68598,68659,68668,68888,68898,68986,69098,69418,69544,69560,69586,69651,69809,69829,69916,70012,70166,70227,70496,70519,70905,70910,71026,71142,71179,71246,71283,71422,71467,71544,71606,71709,71740,71795,71845,71893,71925,72007,72053,72094,72213,72305,72402,72439,72447,72496,72600,72655,72766,72857,72867,72893,73039,73167,73188,73302,73377,73448,73574,73637,73815,74005,74147,74191,74278,74391,74414,74488,74609,74706,74739,74791,74881,74940,74949,75104,75180,75181,75233,75244,75291,75367,75390,75457,75507,75639,75664,75782,75883,75893,75920,75943,75974,76012,76121,76142,76156,76174,76219,76339,76466,76489,76542,76644,76672,76737,76763,76852,76875,77273,77335,77372,77620,77726,77730,77922,77989,78070,78092,78221,78248,78307,78334,78352,78372,78437,78439,78455,78584,78590,78631,78814,78818,78839,78844,78996,79076,79224,79282,79358,79377,79390,79428,79466,79495,79540,79680,79776,79813,79885,79963,80006,80225,80316,80365,80392,80541,80592,80594,80847,80876,80968,81320,81343,81355,81469,81471,81616,81828,81922,81993,82007,82129,82170,82302,82393,82417,82425,82442,82705,82805,82839,82915,83065,83099,83106,83107,83186,83264,83402,83446,83461,83509,83787,83953,83958,84057,84064,84192,84273,84426,84429,84487,84506,84610,84689,84934,85030,85083,85091,85166,85402,85456,85559,85592,85664,85761,85882,86096,86254,86274,86328,86660,86666,86673,86701,87000,87011,87171,87206,87274,87316,87471,87589,87626,87706,87722,87766,87783,87826,87831,87861,87972,88110,88210,88297,88364,88372,88415,88515,88605,88676,88699,88740,88774,88837,88968,89293,89430,89445,89502,89531,89646,89870,89881,89936,89973,90337,90379,90408,90441,90551,90553,90643,90780,90791,90809,90989,91009,91053,91169,91176,91243,91252,91279,91343,91416,91439,91554,91852,91986,92080,92212,92244,92389,92421,92493,92543,92661,92760,93087,93282,93316,93329,93407,93416,93486,93517,93586,93736,93802,93807,93822,93832,93839,93895,93993,94305,94374,94499,94541,94542,94612,94673,94812,95013,95045,95184,95193,95275,95278,95380,95583,95668,95907,95961,96066,96127,96220,96259,96271,96329,96440,96468,96494,96536,96577,96611,96841,96895,97001,97076,97194,97291,97333,97350,97394,97485,97580,97616,97650,97771,97853,97911,97933,98331,98338,98426,98450,98593,98679,98759,98806,98807,99022,99073,99190,99201,99212,99220,99253,99318,99382,99387,99570,99659,99665,99701,99751,99783,99813,99824,99902,99935,99976",
          "34,86,220,247,382,563,591,684,715,781,791,856,916,926,928,929,1164,1220,1299,1337,1341,1376,1379,1398,1435,1444,1454,1531,1569,1605,1689,1767,1768,1822,1825,1849,1850,1913,1930,2026,2059,2137,2612,2613,2659,2673,2880,3084,3095,3112,3210,3280,3383,3629,3681,3776,3805,3837,3941,4188,4242,4321,4401,4410,4414,4623,4723,4770,4778,4878,5016,5020,5053,5113,5126,5221,5333,5358,5379,5457,5541,5559,5568,5683,5771,5813,5854,5891,5964,6054,6346,6397,6423,6484,6572,6658,6680,6760,6952,6987,7028,7165,7210,7215,7355,7363,7508,7648,7838,8084,8208,8354,8388,8398,8498,8509,8534,8550,8605,8771,8791,8961,9009,9068,9172,9197,9391,9427,9530,9734,9770,10041,10050,10108,10401,10436,10443,10447,10481,10482,10504,10562,10810,11080,11128,11279,11284,11325,11328,11443,11481,11541,11543,11674,11681,11739,11828,11875,11898,12116,12300,12548,12686,12793,12852,13123,13126,13155,13180,13201,13305,13307,13316,13367,13375,13468,13500,13547,13662,13827,13893,14000,14054,14068,14165,14196,14242,14283,14361,14407,14417,14495,14548,14627,14777,14796,15689,15703,15863,15910,16140,16164,16170,16174,16430,16505,16570,16609,16686,16817,16863,17012,17123,17132,17145,17285,17383,17492,17699,17945,17977,17996,18120,18130,18148,18233,18254,18424,18482,18577,18840,18909,18915,18944,19038,19085,19131,19186,19202,19235,19629,19671,19840,19877,19898,19983,20041,20229,20397,20552,20577,20644,20672,20742,20749,20847,20858,20860,21167,21360,21371,21546,21847,22061,22164,22679,22746,22814,23074,23369,23411,23781,23906,24038,24130,24221,24422,24505,24587,24622,24644,24712,24991,25116,25363,25468,25519,25628,25990,26157,26159,26166,26219,26631,26673,26780,26800,26893,27040,27059,27119,27302,27345,27394,27502,27689,27738,27746,27773,27822,27828,27838,27855,27925,27956,28250,28279,28342,28373,28406,28413,28566,28572,28745,28991,29024,29049,29221,29621,29634,29707,29714,29739,29779,29785,29829,30307,30585,30788,32047,32257,32876,33978,34074,35033,35113,35393,35467,35569,35575,35620,35767,35768,35808,35857,35930,35989,36046,36062,36175,36268,36375,36386,36567,36742,36872,36907,37068,37163,37200,37303,37304,37370,37385,37415,37612,37630,37748,37762,37854,37863,38108,38193,38232,38290,38485,38569,38608,38817,38821,38998,39102,39124,39374,39556,39615,39819,39865,39873,39920,39934,39941,40427,40841,40872,41721,42169,42181,42364,42389,42656,42715,42727,42763,42824,42853,42859,42870,42940,43014,43078,43424,43756,43813,44157,44169,44296,44451,44578,44705,45092,45118,45153,45174,45321,45378,45496,45548,45620,45643,45774,45926,45937,45939,46018,46466,46861,46983,47181,47727,47930,48000,48181,48447,48840,49114,50157,50273,50276,50315,50323,50345,50444,50507,50508,50608,50658,50738,50821,50873,50938,50964,51329,51358,51443,51476,51583,51756,51763,51777,51788,51802,51898,52020,52058,52151,52342,52350,52364,52380,52549,52554,52594,52661,52671,52738,52839,52845,52854,53027,53055,53106,53148,53229,53322,53335,53369,53537,53597,53680,53682,53759,53802,53821,53901,54160,54169,54220,54612,54964,54989,55010,55021,55061,55089,55133,55197,55220,55264,55286,55293,55385,55493,55655,55668,55669,55815,55849,55885,55896,55918,55940,56127,56150,56360,56371,56456,56520,56657,56735,56858,56918,57014,57061,57342,57412,57698,57906,58042,58085,58130,58158,58250,58275,58286,58359,58439,58460,58514,58569,58664,58679,58755,58901,59034,59065,59115,59118,59205,59361,59372,59449,59524,59579,59616,59656,59960,59970,60068,60300,60319,60396,60408,60411,60422,60428,60429,60489,60513,60702,60736,60767,60903,60984,61026,61154,61364,61575,61587,61621,61662,61692,61974,62150,62289,62310,62329,62417,62647,62675,62762,62784,62804,62860,63063,63248,63407,63439,63446,63536,63555,63672,63760,64073,64258,64303,64308,64348,64385,64456,64669,65293,65302,65480,65581,65688,65727,65730,65847,65960,66026,66032,66152,66204,66276,66319,66342,66496,66555,66683,66811,66893,66898,66978,67200,67395,67419,67543,67578,67580,67697,67714,67801,67822,67858,68020,68050,68095,68260,68312,68315,68373,68576,68615,68755,68783,68788,68832,68983,69052,69081,69106,69143,69263,69279,69305,69317,69496,69554,69671,69714,69874,69991,69996,70189,70359,70469,70504,70571,70612,70967,70995,71038,71148,71164,71169,71473,71490,71582,71693,71712,71814,71854,71905,72041,72060,72097,72157,72166,72202,72268,72288,72340,72368,72495,72565,72582,72668,72764,72770,72909,73030,73043,73149,73227,73256,73462,73655,73914,73947,74066,74076,74111,74112,74218,74254,74276,74291,74471,74529,74599,74672,74839,74942,74968,75232,75258,75393,75407,75531,75534,75605,75631,75791,75795,75812,75889,75937,75972,75996,76087,76146,76227,76233,76247,76379,76409,76507,76632,76713,76777,76869,77039,77069,77138,77189,77280,77366,77476,77530,77537,77570,77682,77805,77889,78022,78025,78027,78184,78261,78268,78484,78620,78984,79032,79198,79204,79330,79366,79456,79581,79674,79893,79971,79979,80050,80066,80224,80241,80284,80566,80580,80620,80666,80673,80732,80785,80838,80880,80898,81022,81119,81177,81272,81341,81383,81498,81531,81679,81736,81738,81886,81977,81978,81999,82008,82073,82445,82480,82555,82605,82654,82660,82709,82768,82849,82868,82954,83012,83292,83457,83606,83758,83765,83772,83859,83888,83899,83939,83956,83991,84036,84297,84531,84546,84591,84781,84988,84995,85053,85296,85357,85393,85415,85564,85581,85758,85839,85846,85869,85930,85959,86029,86160,86183,86190,86273,86430,86447,86510,86706,86721,86864,86877,87080,87240,87262,87295,87350,87364,87396,87406,87418,87422,87607,87670,87678,87791,87913,88206,88312,88334,88370,88497,88561,88565,88595,88783,88828,88935,89254,89280,89352,89683,89875,89912,89923,90175,90264,90463,90545,90601,90684,90688,90834,90897,90906,90910,90916,90940,91041,91190,91287,91344,91469,91477,91498,91539,91575,91920,91922,91952,92052,92109,92282,92423,92575,92601,92680,92699,93143,93230,93261,93466,93492,93514,93599,93610,93664,93738,93918,93962,94017,94025,94035,94057,94077,94158,94182,94190,94199,94243,94330,94448,94551,94769,94774,94840,94843,94910,95000,95142,95143,95220,95486,95600,95625,95646,95702,95741,95762,95916,95919,95991,95996,96126,96282,96473,96549,96609,96727,96830,97274,97359,97513,97610,97667,97786,97909,98178,98607,98611,98960,98967,98972,99013,99026,99047,99256,99322,99397,99449,99463,99490,99608,99611,99722,99729,99829,99988",
          "36,56,112,139,147,192,1715,2141,2148,2245,2364,2620,2728,2777,2791,3011,3045,3163,3239,3673,4213,4232,4485,4509,4974,4975,5877,6013,6591,6643,6852,6863,6932,7161,7247,8019,8130,8168,8244,8510,8643,8681,8928,8930,9023,9049,9080,9099,9141,9175,9194,9346,9384,9460,9469,9498,9593,9612,9650,9688,9707,9819,9820,9821,9878,9935,9943,10049,10057,10165,10212,10220,10239,10258,10269,10328,10429,10467,10469,10512,10581,10600,10636,10654,10664,10752,10809,10820,10885,10961,11015,11021,11047,11113,11170,11188,11208,11303,11455,11512,11518,11521,11551,11576,11599,11645,11721,11759,11877,11892,11899,11911,11993,12025,12109,12215,12234,12253,12291,12310,12386,12481,12500,12557,12566,12671,12747,12804,12899,13089,13211,13214,13260,13279,13298,13374,13469,13492,13621,13640,13678,13697,13754,13773,13921,13944,13963,14020,14058,14061,14077,14153,14191,14229,14267,14305,14381,14457,14533,14590,14628,14663,14685,14818,14896,14913,14918,14985,14989,14998,15034,15141,15179,15227,15236,15255,15350,15377,15407,15464,15597,15635,15692,15730,15749,15806,15825,15844,15920,16034,16053,16063,16110,16148,16167,16186,16205,16224,16243,16281,16300,16319,16376,16433,16483,16490,16509,16661,16737,16813,16870,16927,16940,16943,16965,17022,17091,17117,17212,17231,17307,17345,17364,17368,17437,17459,17478,17497,17592,17610,17611,17687,17720,17744,17782,17801,17839,17896,17934,17953,17954,17972,17979,18048,18105,18124,18181,18200,18276,18314,18352,18409,18447,18479,18593,18618,18694,18765,18770,18789,18846,18903,18960,19084,19150,19188,19230,19264,19302,19340,19359,19416,19419,19435,19453,19682,19740,19834,19872,19891,19910,19948,20005,20062,20074,20119,20138,20157,20178,20195,20233,20271,20309,20347,20423,20438,20461,20594,20613,20898,20993,21021,21069,21077,21145,21164,21188,21202,21219,21274,21278,21335,21380,21411,21430,21468,21482,21563,21582,21715,21734,21753,21791,21810,21829,21867,21905,21924,21943,22019,22073,22076,22228,22266,22342,22399,22418,22437,22456,22513,22533,22589,22647,22684,22703,22735,22832,22840,22945,22950,23010,23028,23083,23178,23197,23235,23255,23273,23329,23330,23347,23368,23425,23463,23499,23501,23539,23558,23577,23605,23623,23634,23710,23748,23862,23881,23946,23976,23995,24052,24150,24194,24204,24223,24242,24250,24337,24432,24451,24470,24527,24565,24584,24755,24857,24869,24888,24941,25002,25040,25073,25078,25094,25097,25135,25211,25287,25302,25325,25381,25382,25477,25534,25553,25565,25667,25714,25781,25819,25838,25876,25914,25952,25971,26123,26142,26161,26218,26237,26351,26408,26427,26484,26503,26636,26687,26705,26763,26875,26902,26997,27116,27162,27203,27244,27250,27295,27377,27383,27418,27434,27593,27618,27670,27719,27736,27909,28080,28099,28113,28194,28295,28536,28574,28593,28762,28897,28954,29201,29226,29227,29292,29439,29562,29638,29752,29809,29860,29879,29934,29985,30027,30056,30143,30172,30230,30259,30317,30346,30375,30404,30462,30491,30520,30607,30636,30665,30723,30752,30781,30810,30839,30897,30926,30955,31013,31071,31100,31129,31158,31187,31245,31274,31290,31361,31448,31506,31535,31651,31680,31725,31767,31796,31854,31912,31970,31999,32086,32115,32144,32173,32231,32260,32318,32347,32376,32392,32405,32463,32492,32579,32666,32695,32724,32840,32869,32927,33014,33043,33072,33101,33130,33188,33217,33304,33333,33362,33478,33507,33536,33594,33623,33681,33710,33855,33884,33913,33971,34029,34116,34174,34232,34261,34290,34348,34522,34609,34638,34696,34725,34754,34783,34812,34870,34928,34957,35027,35116,35142,35279,35324,35353,35359,35427,35546,35664,35796,35855,35867,35985,35992,36004,36080,36107,36112,36167,36191,36226,36284,36297,36356,36415,36605,36669,36728,36787,36858,36976,36980,36988,37048,37166,37172,37230,37293,37299,37348,37392,37419,37488,37537,37548,37674,37791,37850,37914,38039,38288,38352,38411,38470,38528,38541,38600,38659,38731,38981,38993,39161,39226,39248,39292,39533,39592,39781,39912,39971,40101,40130,40188,40246,40275,40333,40391,40507,40536,40565,40623,40676,40681,40710,40739,40768,40797,40884,40913,40942,40971,41000,41174,41319,41464,41493,41522,42054,42083,42112,42172,42231,42260,42337,42406,42499,42574,42603,42620,42649,42736,42771,42811,42845,42886,42973,43016,43048,43116,43155,43210,43248,43275,43278,43283,43285,43347,43360,43447,43488,43574,43597,43641,43655,43684,43759,43814,43899,43909,43921,43996,44056,44071,44119,44129,44134,44158,44190,44308,44334,44360,44395,44470,44500,44525,44530,44583,44603,44632,44661,44669,44684,44707,44743,44782,44823,44841,44944,44975,45010,45019,45140,45254,45332,45389,45425,45519,45576,45634,45706,45763,45821,45848,45854,45893,46030,46098,46105,46111,46161,46364,46381,46444,46508,46591,46619,46735,46742,46762,46969,47103,47175,47180,47216,47234,47250,47317,47340,47346,47378,47539,47567,47587,47605,47660,47765,47850,47940,48069,48130,48136,48306,48329,48409,48499,48519,48543,48566,48589,48610,48624,48644,48704,48739,48880,48945,49026,49028,49071,49228,49311,49361,49477,49535,49574,49646,49706,49811,49813,49828,49831,49855,49856,49940,49945,49973,50004,50088,50108,50138,50199,50266,50271,50287,50629,50649,50995,51037,51053,51077,51243,51251,51256,51332,51381,51450,51691,51728,51852,51886,51943,51945,51988,52060,52101,52357,52382,52589,52763,52764,52862,52886,52905,53171,53184,53438,53511,53622,53631,53660,54118,54178,54263,54383,54461,54512,54587,54625,54630,54737,54868,54939,55000,55107,55157,55311,55330,55368,55380,55397,55517,55554,55638,55672,55724,55754,55956,56039,56111,56114,56248,56291,56450,56474,56582,56603,56684,56693,56901,56911,57136,57237,57308,57416,57503,57514,57570,57572,57688,58019,58059,58095,58231,58300,58399,58422,58536,58771,58786,58787,58918,58968,58997,59004,59040,59054,59308,59321,59371,59424,59578,59597,59608,59784,59814,59828,59835,59884,59923,59982,60001,60030,60136,60140,60142,60162,60221,60235,60615,60719,60777,60944,60958,60964,61069,61211,61342,61435,61496,61687,61857,61912,61921,62131,62169,62231,62248,62274,62363,62561,62769,62855,62974,63084,63137,63278,63412,63566,63809,63914,63969,64055,64056,64135,64172,64204,64273,64310,64357,64367,64375,64404,64666,64775,64779,65119,65215,65219,65270,65421,65559,65599,65650,65693,65850,65915,65923,65976,66094,66159,66254,66281,66477,66540,66628,66681,66729,66750,66756,67035,67063,67084,67153,67155,67275,67393,67411,67458,67465,67565,67917,67930,68006,68008,68011,68065,68080,68211,68213,68216,68488,68500,68514,68536,69019,69024,69142,69233,69273,69345,69514,69657,69684,69693,69723,69749,69833,69834,69862,69899,69927,69958,70050,70237,70275,70423,70461,70599,70845,71034,71078,71084,71129,71131,71320,71356,71496,71564,71585,71613,71624,71671,71692,71694,71772,71901,71907,71957,71988,71993,72052,72088,72121,72313,72390,72391,72509,72622,72742,72744,72880,72910,73023,73460,73498,73547,73572,73719,73735,73800,73955,74022,74056,74226,74265,74320,74338,74451,74518,74619,74620,74625,74658,74715,74792,74844,74849,74851,74961,75029,75075,75088,75099,75340,75374,75813,75874,76010,76042,76107,76150,76177,76193,76439,76453,76748,77012,77071,77195,77306,77310,77396,77477,77551,77568,77581,77700,77704,77928,77954,77995,78172,78340,78468,78859,78945,79043,79205,79383,79430,79433,79712,79793,79847,79916,80037,80437,80456,80574,80774,80981,80988,81139,81163,81179,81253,81364,81404,81809,81845,81889,81904,81907,81947,81963,82045,82052,82084,82097,82137,82364,82401,82459,82473,82816,83059,83207,83295,83366,83468,83521,83663,83673,83686,83701,83821,83915,83928,83973,83996,84129,84187,84189,84268,84320,84356,84373,84395,84719,84812,84835,84890,84941,84971,85016,85069,85102,85157,85216,85219,85288,85383,85586,85652,85656,85804,85945,85988,86100,86187,86218,86307,86337,86368,86480,86483,86627,86784,87030,87085,87090,87144,87317,87436,87493,87529,87590,87800,87812,87822,87916,87945,88042,88309,88337,88482,88521,88555,88724,89077,89227,89321,89340,89386,89387,89403,89526,89616,89626,89652,89781,89815,89893,89922,90019,90267,90346,90467,90639,90804,90816,90857,90928,91037,91066,91087,91285,91305,91327,91604,91632,91642,91692,91725,91808,92062,92220,92306,92410,92440,92484,92562,93428,93442,93566,93803,93812,93897,94130,94138,94249,94445,94478,94620,94807,94815,94826,94946,95020,95178,95282,95369,95412,95461,95526,95548,95827,96067,96531,96703,96781,96782,96881,96884,97038,97151,97186,97219,97234,97366,97475,97505,97848,97849,97867,98239,98497,98982,99027,99092,99150,99172,99224,99331,99353,99438,99495,99498,99504,99511,99513,99528,99541,99544,99625,99773,99805,99869,99882",
          "88,107,174,180,252,718,1627,1665,1752,1763,1910,1921,1928,1946,1956,1960,1973,1989,2000,2005,2021,2032,2040,2071,2073,2084,2097,2098,2102,2104,2115,2124,2130,2145,2171,2175,2187,2220,2234,2236,2254,2262,2285,2315,2403,2407,2444,2460,2489,2503,2520,2552,2558,2597,2615,2639,2650,2675,2687,2718,2726,2729,2750,2752,2760,2762,2776,2787,2796,2840,2856,2881,2888,2898,2906,2917,2945,2963,2964,2978,2991,2993,2994,3002,3019,3024,3041,3047,3048,3076,3079,3128,3138,3167,3170,3176,3187,3191,3236,3254,3263,3266,3267,3269,3283,3292,3307,3365,3397,3423,3426,3436,3444,3445,3449,3465,3488,3489,3494,3505,3514,3555,3559,3583,3590,3599,3626,3652,3666,3669,3678,3697,3705,3712,3725,3788,3806,3808,3809,3819,3844,3847,3848,3850,3866,3872,3892,3894,3904,3909,3957,3962,3982,3992,3993,3995,4002,4021,4039,4045,4065,4082,4085,4103,4120,4143,4150,4193,4196,4219,4222,4224,4228,4238,4254,4288,4292,4298,4301,4305,4307,4313,4326,4332,4339,4345,4365,4421,4446,4463,4468,4478,4503,4510,4518,4520,4543,4556,4564,4598,4615,4632,4638,4664,4673,4689,4740,4744,4756,4758,4776,4792,4794,4823,4825,4832,4856,4861,4871,4874,4889,4909,4932,4937,5463,5583,5598,5655,5676,5921,6005,6104,6298,6472,6481,6487,6568,6586,6593,6605,6657,6701,6720,6741,6789,6838,6904,6909,6965,7090,7118,7181,7232,7234,7265,7293,7324,7349,7359,7510,7551,7582,7619,7699,7802,7854,7881,7913,8041,8060,8147,8163,8200,8250,8375,8616,8686,8750,8909,8992,9014,9031,9089,9146,9210,9244,9292,9551,9729,9733,9754,9813,9928,9954,10008,10010,10115,10256,10290,10301,10302,10419,10437,10454,10589,10594,10784,10904,10994,11086,11089,11141,11187,11289,11390,11420,11610,11680,11716,11865,11926,11936,12027,12080,12179,12362,12401,12402,12544,12675,12714,13000,13097,13288,13538,13552,13583,13721,13753,13761,13806,13815,13861,13877,14015,14125,14174,14281,14392,14393,14482,14550,14562,14576,14582,14825,14903,14948,15065,15115,15288,15345,15399,15718,15735,16087,16195,16308,16384,16544,16723,16810,16860,16968,17265,17316,17376,17429,17558,17675,17730,17915,18019,18067,18151,18178,18271,18328,18474,18623,18638,18641,18671,18691,18711,18714,18732,18776,18788,18798,18836,19113,19165,19295,19425,19491,19655,19712,19778,19857,19901,19943,20136,20211,20399,20426,20479,20491,20514,20543,20584,20667,20800,20931,20947,20976,21035,21123,21147,21258,21416,21500,21530,21556,21840,21959,22005,22069,22118,22281,22379,22492,22695,22857,22940,22958,22979,23047,23122,23215,23216,23401,23408,23436,23609,23835,23846,23847,23980,24005,24120,24179,24244,24634,24742,25376,25704,25860,25891,25892,25985,26005,26054,26105,26135,26165,26362,26365,26571,26572,26592,26646,26869,26953,27071,27117,27177,27179,27311,27334,27338,27494,27540,27560,27600,27633,27651,27744,27899,27992,28010,28035,28053,28063,28090,28096,28117,28119,28135,28136,28159,28300,28463,28476,28499,28508,28525,28546,28548,28579,28609,28639,28648,28655,28656,28838,28899,29182,29541,29593,29605,29674,29675,29690,29854,29896,29917,29942,29992,30015,30044,30073,30102,30131,30160,30189,30218,30247,30276,30305,30334,30363,30392,30421,30450,30479,30508,30537,30566,30595,30624,30653,30682,30740,30769,30798,30827,30856,30885,30914,30943,30972,31001,31030,31059,31088,31117,31146,31175,31204,31233,31262,31291,31320,31349,31378,31407,31436,31465,31494,31523,31552,31581,31610,31639,31668,31697,31726,31755,31784,31813,31842,31871,31900,31929,31958,31987,32016,32045,32103,32132,32161,32190,32219,32248,32277,32306,32335,32364,32393,32422,32451,32480,32509,32538,32567,32596,32625,32654,32683,32712,32741,32770,32799,32828,32857,32886,32915,32944,32973,33002,33031,33060,33089,33118,33147,33176,33205,33234,33263,33292,33321,33350,33379,33408,33437,33466,33524,33553,33582,33611,33640,33669,33698,33727,33756,33785,33814,33843,33872,33901,33959,33988,34017,34046,34075,34104,34133,34162,34191,34220,34249,34278,34307,34336,34365,34394,34423,34452,34481,34510,34539,34568,34597,34626,34655,34684,34713,34742,34771,34800,34829,34858,34887,34916,34945,34974,35015,35046,35059,35073,35104,35127,35155,35178,35214,35240,35267,35341,35415,35440,35486,35534,35593,35614,35652,35725,35784,35843,35864,35914,35973,36017,36032,36056,36069,36095,36155,36190,36214,36251,36285,36344,36376,36403,36475,36497,36534,36558,36593,36621,36626,36657,36716,36732,36733,36739,36775,36846,36905,36964,36979,36984,36997,37036,37095,37126,37154,37218,37235,37245,37277,37336,37407,37446,37466,37505,37525,37550,37582,37656,37677,37685,37715,37733,37779,37797,37838,37897,37938,37968,38027,38086,38102,38158,38182,38217,38237,38276,38340,38360,38399,38443,38458,38477,38529,38588,38622,38647,38719,38738,38778,38796,38799,38837,38901,38935,38960,39019,39090,39149,39208,39280,39339,39398,39424,39429,39430,39462,39502,39521,39537,39544,39580,39604,39651,39677,39710,39769,39784,39786,39900,39959,39977,40112,40141,40170,40228,40257,40286,40315,40344,40373,40402,40431,40489,40518,40576,40605,40634,40663,40692,40721,40750,40779,40808,40837,40866,40895,40924,40953,40982,41011,41040,41069,41098,41127,41156,41185,41214,41243,41272,41301,41330,41359,41388,41417,41446,41475,41504,41533,42004,42033,42062,42091,42120,42156,42180,42270,42298,42324,42345,42386,42420,42477,42507,42534,42551,42562,42582,42657,42744,42819,42837,42894,42931,42954,42965,42981,43056,43077,43094,43095,43108,43109,43131,43243,43244,43293,43368,43455,43530,43605,43692,43744,43767,43818,43842,43929,44004,44065,44079,44096,44098,44117,44130,44150,44166,44218,44241,44264,44316,44333,44358,44403,44429,44432,44441,44478,44498,44527,44553,44640,44715,44776,44790,44838,44868,44877,44931,44938,44952,44983,44985,45006,45027,45051,45063,45066,45087,45100,45133,45148,45209,45267,45289,45294,45303,45340,45397,45455,45527,45560,45584,45642,45670,45714,45771,45829,45922,45940,45958,45983,46028,46090,46123,46173,46202,46278,46287,46288,46325,46352,46372,46380,46383,46397,46483,46511,46544,46582,46654,46658,46758,46769,46808,46825,46828,46933,46981,47105,47194,47204,47223,47337,47358,47409,47451,47485,47521,47531,47570,47572,47580,47600,47620,47687,47710,47805,47856,47876,47878,47890,47934,47943,48032,48059,48068,48110,48133,48137,48144,48175,48231,48240,48328,48357,48362,48405,48424,48428,48432,48441,48532,48537,48539,48556,48582,48606,48640,48641,48651,48671,48678,48722,48742,48831,48930,48938,48996,49016,49052,49076,49096,49097,49138,49140,49153,49155,49162,49167,49193,49209,49260,49273,49294,49327,49335,49359,49375,49385,49433,49454,49459,49504,49537,49551,49559,49565,49567,49594,49602,49618,49659,49666,49677,49690,49734,49764,49802,49826,49906,49922,49964,49975,50022,50066,50290,50324,50385,50402,50478,50558,50570,50574,50703,50839,50911,50950,50953,51006,51069,51079,51108,51114,51121,51268,51353,51449,51454,51547,51574,51644,51661,51707,51795,51817,52091,52097,52162,52276,52399,52413,52465,52482,52681,52735,52747,52936,52995,53006,53012,53250,53270,53333,53550,53559,53582,53591,53652,53805,53838,54020,54156,54212,54231,54306,54638,54761,54777,54802,54815,55060,55079,55120,55284,55556,55608,55733,55735,55741,55782,55793,55947,55951,55986,55991,56010,56021,56041,56090,56148,56227,56288,56411,56647,56675,56805,56854,56871,56882,56932,56958,57283,57343,57352,57401,57472,57520,57883,57978,58008,58051,58068,58074,58090,58162,58419,58716,58764,58791,58833,58949,58989,59016,59062,59113,59181,59198,59338,59366,59528,59763,59903,60076,60090,60238,60263,60315,60342,60436,60483,60510,60543,60596,60624,60765,60809,60884,60906,60943,61104,61147,61159,61188,61209,61221,61229,61230,61281,61536,61573,61748,61784,61804,61834,61997,62343,62420,62509,62544,62629,62667,62789,62924,62984,63045,63107,63253,63255,63477,63527,63546,63552,63565,63704,63813,63859,63996,64045,64278,64409,64416,64638,64674,64873,64931,64958,64987,65035,65049,65168,65253,65341,65419,65451,65524,65818,65831,65861,65867,65908,65921,66085,66120,66173,66440,66511,66608,66625,66653,66654,66793,66818,66908,66955,67043,67139,67223,67226,67257,67408,67607,67634,67657,67698,67902,68057,68126,68444,68513,68629,68656,68674,68694,68733,68759,68876,68911,68964,69021,69080,69090,69129,69155,69249,69281,69304,69415,69416,69432,69582,69637,69709,69797,69941,69942,70089,70103,70172,70255,70297,70331,70335,70421,70732,70786,70914,70932,71061,71236,71243,71258,71454,71535,71558,71586,71640,71737,71816,71852,71897,71904,71964,71976,72137,72280,72348,72425,72521,72638,72786,72820,72932,72966,73127,73295,73321,73343,73621,73641,73697,73782,74003,74115,74157,74426,74470,74517,74546,74645,74664,74802,74805,74863,74879,74920,74958,75042,75281,75328,75336,75423,75449,75484,75741,75822,75861,76085,76289,76486,76686,76693,76709,76804,76912,76955,76958,77023,77056,77104,77108,77208,77308,77398,77455,77558,77595,77651,77736,77786,77894,78003,78009,78029,78055,78058,78060,78095,78296,78313,78346,78673,78773,78778,78947,79081,79082,79100,79113,79277,79317,79354,79363,79462,79518,79529,79594,79676,79722,79738,79786,79906,79948,80014,80046,80096,80256,80342,80394,80401,80644,80764,80849,80896,80993,81220,81225,81230,81353,81381,81382,81516,81638,81764,82016,82041,82051,82098,82371,82420,82436,82511,82526,82649,82794,82874,82992,83119,83172,83224,83349,83503,83523,83596,83658,83809,83817,83873,83914,83992,84032,84048,84372,84550,84611,84613,84715,84832,84902,84963,85038,85097,85133,85225,85317,85552,85854,85968,86076,86143,86520,86564,86609,86618,86633,86638,86654,86897,86998,87031,87096,87225,87245,87286,87434,87455,87480,87709,87818,88080,88271,88283,88353,88418,88441,88511,88648,88708,88756,88816,88959,88987,89163,89212,89213,89222,89243,89265,89374,89423,89520,89590,89592,89649,89650,89674,89684,89696,89701,89742,89884,90000,90091,90099,90156,90224,90254,90284,90544,90642,90693,90698,90806,90919,91036,91120,91139,91156,91246,91284,91299,91358,91422,91524,91657,91700,91872,91885,91965,91984,92179,92215,92328,92348,92444,92558,92643,92646,92787,92857,92938,92988,93122,93170,93185,93306,93340,93347,93472,93618,93682,93751,93857,93985,94011,94160,94595,94630,94734,94902,94928,95017,95094,95248,95312,95313,95336,95543,95592,95609,95620,95886,96088,96123,96143,96408,96446,96568,96599,96639,96716,96772,96785,96818,96868,96869,96877,96975,96977,97232,97381,97511,97632,97797,97820,97841,97884,98097,98250,98302,98335,98386,98499,98653,98675,98711,98749,98781,98973,99048,99136,99143,99178,99282,99393,99447,99509,99707,99974,99995",
          "39,166,173,245,249,1672,1730,1787,1958,1996,2110,2152,2278,2846,3113,3150,3497,3641,3757,3814,3817,3845,4010,4020,4089,4275,4304,4311,4479,4483,4519,4783,4819,4992,4998,5602,5659,8611,9366,9382,9387,9482,9521,9539,9558,9596,9653,9672,9691,9710,9805,9930,10014,10063,10071,10119,10185,10261,10603,10641,10698,10717,10754,10755,10774,10793,10812,10925,10945,10969,11059,11078,11135,11154,11173,11192,11230,11249,11270,11282,11306,11382,11399,11439,11456,11496,11515,11534,11553,11629,11648,11686,11876,11880,11895,11896,11952,11971,12028,12039,12077,12104,12123,12161,12180,12237,12256,12275,12294,12484,12494,12503,12560,12584,12598,12617,12636,12655,12693,12712,12750,12769,12771,12774,12834,12845,12902,12959,12997,13035,13054,13073,13187,13244,13269,13282,13301,13377,13422,13434,13466,13509,13548,13550,13566,13571,13605,13624,13643,13681,13700,13738,13795,13814,13846,13852,13928,13966,14004,14023,14137,14138,14156,14194,14213,14251,14270,14289,14327,14384,14426,14468,14479,14517,14536,14574,14593,14612,14631,14650,14654,14707,14783,14807,14824,14840,14845,14875,15030,15049,15144,15217,15220,15223,15277,15296,15353,15372,15391,15404,15410,15429,15433,15486,15556,15581,15600,15638,15657,15676,15733,15752,15771,15799,15809,15828,15856,15866,15885,15911,15923,15961,15965,15980,15983,16056,16075,16132,16137,16189,16208,16223,16227,16265,16284,16303,16322,16341,16360,16379,16398,16417,16436,16474,16531,16569,16607,16645,16721,16740,16759,16854,16911,16930,16945,16949,16987,17025,17044,17063,17101,17110,17120,17148,17158,17196,17215,17273,17288,17289,17294,17329,17348,17367,17386,17424,17462,17489,17519,17538,17576,17595,17669,17671,17728,17766,17769,17785,17804,17823,17842,17861,17899,17906,17918,17937,18013,18032,18127,18165,18184,18198,18241,18279,18298,18317,18355,18374,18393,18412,18431,18450,18488,18507,18564,18602,18678,18697,18754,18773,18792,18811,18868,18925,18949,18963,18987,19001,19020,19039,19077,19096,19111,19115,19210,19229,19267,19286,19343,19362,19365,19390,19400,19495,19514,19571,19628,19685,19704,19742,19761,19780,19803,19856,19875,19894,19932,19951,19970,20008,20027,20065,20084,20103,20118,20141,20179,20217,20262,20274,20312,20350,20369,20407,20445,20464,20483,20502,20540,20578,20635,20654,20673,20711,20730,20777,20825,20842,20844,20863,20865,20882,20958,20977,20999,21058,21059,21091,21094,21129,21194,21205,21224,21319,21357,21376,21414,21452,21471,21547,21585,21604,21680,21718,21737,21756,21785,21794,21813,21832,21851,21908,21927,21946,21965,21984,22022,22079,22095,22117,22136,22193,22250,22269,22307,22326,22383,22402,22421,22440,22459,22478,22497,22516,22535,22554,22565,22573,22611,22628,22630,22687,22706,22725,22744,22782,22801,22818,22819,22877,22886,22896,22953,22972,23067,23105,23109,23124,23162,23181,23238,23276,23279,23295,23305,23314,23352,23390,23407,23409,23504,23523,23542,23565,23599,23626,23637,23656,23713,23732,23770,23780,23808,23865,23875,23884,23903,23922,23941,23960,23979,23998,24017,24055,24112,24131,24169,24207,24226,24240,24302,24357,24359,24376,24378,24416,24435,24454,24473,24492,24511,24530,24549,24606,24625,24649,24701,24739,24758,24777,24796,24834,24853,24910,25005,25062,25072,25081,25138,25157,25161,25195,25233,25252,25271,25309,25328,25347,25423,25442,25461,25537,25544,25556,25594,25613,25702,25708,25727,25757,25765,25784,25803,25848,25861,25879,25898,25917,25936,25955,25974,25993,26012,26031,26050,26069,26145,26212,26221,26240,26241,26297,26316,26317,26354,26363,26373,26380,26392,26411,26430,26506,26639,26658,26675,26772,27000,27019,27028,27113,27228,27344,27357,27361,27380,27440,27456,27478,27513,27524,27563,27646,27783,27836,27848,27898,27902,27958,27969,28015,28026,28102,28216,28293,28501,28596,28691,28710,28729,28756,28816,28824,28924,29004,29033,29052,29058,29075,29095,29158,29172,29299,29375,29418,29419,29510,29575,29660,29699,29717,29719,29774,29831,29870,29950,30001,30030,30059,30088,30117,30146,30175,30204,30233,30262,30320,30349,30378,30407,30436,30465,30494,30523,30552,30572,30581,30610,30639,30668,30697,30726,30755,30784,30813,30842,30871,30900,30929,30958,30987,31016,31045,31074,31103,31132,31161,31190,31219,31248,31277,31306,31335,31364,31393,31422,31451,31480,31509,31538,31567,31596,31625,31654,31712,31741,31770,31799,31828,31857,31886,31915,31944,31973,32002,32031,32060,32089,32118,32147,32176,32205,32234,32263,32292,32321,32350,32379,32408,32437,32466,32495,32524,32553,32582,32611,32640,32669,32698,32727,32756,32785,32814,32843,32872,32901,32930,32988,33017,33046,33075,33104,33133,33162,33191,33220,33249,33278,33307,33336,33365,33394,33423,33452,33481,33510,33539,33568,33597,33626,33655,33684,33713,33742,33771,33800,33829,33858,33887,33916,33945,33974,34003,34032,34061,34090,34119,34148,34177,34206,34235,34264,34293,34322,34351,34380,34409,34435,34438,34467,34496,34525,34554,34583,34612,34641,34670,34699,34728,34757,34786,34815,34844,34873,34902,34931,34960,34989,35001,35076,35082,35090,35148,35164,35239,35253,35300,35327,35376,35395,35401,35472,35513,35520,35553,35573,35579,35616,35638,35711,35741,35770,35829,35900,35942,35959,36018,36071,36081,36141,36200,36271,36324,36330,36362,36365,36389,36424,36436,36441,36461,36520,36579,36643,36672,36702,36743,36761,36801,36832,36891,36950,37022,37059,37081,37140,37204,37263,37292,37306,37322,37393,37452,37511,37583,37642,37701,37765,37824,37853,37883,37954,38013,38042,38047,38058,38072,38144,38203,38262,38293,38326,38374,38385,38444,38515,38574,38618,38633,38684,38705,38751,38764,38823,38887,38920,38946,38978,39005,39076,39135,39166,39167,39190,39193,39194,39266,39313,39325,39357,39384,39434,39448,39507,39566,39599,39637,39696,39755,39798,39810,39827,39886,39945,39974,40113,40142,40171,40200,40229,40258,40287,40316,40345,40374,40403,40432,40461,40490,40519,40548,40577,40606,40635,40664,40693,40722,40751,40780,40809,40838,40867,40896,40925,40954,40983,41012,41041,41070,41099,41128,41157,41186,41215,41244,41273,41302,41331,41360,41389,41418,41447,41476,41505,41534,41700,41789,41791,41793,41795,41797,41799,41801,41803,41805,41809,41811,41813,41815,41817,41819,41821,41823,41825,41827,41829,41831,41833,41835,41837,41839,41841,41843,41845,41847,41851,41853,41855,41857,41859,41861,41863,41865,41867,41869,41871,41873,41875,41877,41879,41881,41883,41885,41887,41889,41891,41893,41895,41897,41903,41905,41907,41909,41911,41913,41915,41917,41919,41921,41923,41927,41930,41933,41936,41939,41942,41945,41948,41951,41954,41958,41962,41966,41970,41974,41978,41982,41986,41991,42020,42049,42078,42107,42136,42155,42164,42196,42205,42209,42286,42327,42361,42374,42375,42436,42468,42523,42541,42542,42598,42673,42674,42699,42701,42719,42732,42760,42835,42910,42952,42971,43037,43072,43105,43113,43147,43152,43167,43174,43186,43204,43234,43309,43334,43338,43384,43405,43471,43485,43507,43546,43561,43576,43621,43664,43708,43718,43719,43751,43783,43800,43858,43865,43919,43945,43953,43982,44020,44035,44045,44095,44105,44182,44201,44225,44257,44292,44301,44332,44351,44384,44393,44419,44466,44494,44511,44569,44591,44656,44692,44731,44737,44750,44806,44808,44818,44843,44845,44893,44929,44935,44939,44968,45043,45052,45054,45056,45061,45070,45073,45116,45176,45232,45284,45305,45356,45361,45416,45417,45473,45479,45493,45543,45604,45609,45680,45730,45750,45791,45850,45864,45867,45917,45929,45978,46001,46007,46008,46023,46077,46144,46177,46243,46300,46348,46394,46398,46404,46410,46482,46487,46520,46527,46657,46736,46745,46806,46863,46878,46887,46914,46963,46967,46988,47050,47063,47091,47111,47121,47122,47143,47159,47228,47319,47320,47329,47339,47356,47366,47405,47421,47436,47437,47480,47481,47517,47564,47568,47574,47581,47594,47643,47678,47694,47713,47773,47783,47790,47792,47793,47819,47828,47829,47885,47917,47937,47941,47981,48010,48025,48028,48062,48113,48123,48126,48134,48150,48154,48161,48229,48230,48232,48304,48313,48317,48367,48393,48398,48401,48451,48489,48513,48584,48588,48605,48658,48665,48682,48706,48716,48746,48776,48790,48794,48809,48814,48817,48819,48836,48863,48920,49172,49189,49196,49219,49246,49274,49296,49346,49395,49403,49443,49501,49519,49530,49599,49627,49707,49737,49742,49767,49847,49852,49873,49882,49953,49957,49991,49992,50029,50188,50195,50314,50373,50410,50417,50455,50470,50583,50735,50858,50894,50981,51213,51290,51405,51414,51441,51544,51571,51647,51672,51701,51761,51782,51798,52108,52175,52194,52196,52346,52425,52437,52469,52497,52618,52793,52847,52880,52943,52944,52980,53123,53379,53424,53449,53573,53616,53768,53794,53798,53861,53969,54022,54027,54066,54174,54179,54226,54375,54511,54615,54795,54850,54919,54934,54967,55216,55274,55309,55316,55406,55477,55518,55534,55564,55580,55615,55664,55723,55769,55846,55887,55958,56078,56125,56252,56280,56331,56338,56427,56461,56522,56773,57037,57088,57107,57128,57162,57230,57287,57300,57323,57334,57358,57382,57524,57542,57560,57679,57726,57745,57804,57914,58349,58350,58393,58476,58611,58616,58698,58793,58841,58942,59130,59188,59294,59525,59575,59686,59715,59772,59776,59822,59886,59957,59980,60011,60061,60087,60378,60516,60579,60593,60673,60691,60722,60849,61142,61341,61375,61377,61428,61567,61588,61674,61703,61843,61870,62197,62219,62335,62641,62706,62757,62948,63040,63056,63076,63109,63184,63392,63395,63427,63696,63724,63895,64359,64369,64544,64614,64681,64781,64842,64846,64854,64868,65008,65194,65203,65208,65263,65469,65488,65564,65789,65823,65841,65906,65937,65968,65977,66008,66063,66095,66142,66154,66163,66465,66492,66552,66573,66678,66757,66791,66800,66885,67287,67405,67464,67539,67696,67837,67895,68107,68115,68361,68364,68378,68457,68515,68569,68601,68611,68641,68684,68862,68935,69335,69523,69766,69782,69792,70044,70269,70309,70339,70363,70634,70720,70918,70958,70973,71074,71098,71153,71215,71273,71290,71322,71476,71567,71574,71691,71696,71697,71729,71784,71825,71827,71960,71994,72081,72167,72349,72745,72833,72862,72975,73005,73061,73071,73207,73212,73486,73507,73614,73618,73841,73880,74080,74141,74144,74269,74308,74393,74417,74455,74506,74558,74561,74746,74978,75095,75173,75333,75413,75509,75628,75758,75790,75811,75969,75980,76333,76386,76429,76432,76506,76547,76586,76667,76689,76742,76841,76896,77053,77350,77480,77614,77643,77649,77832,77862,77905,78008,78074,78102,78148,78190,78258,78287,78318,78351,78392,78563,78689,78791,78796,78803,78845,78870,79044,79175,79236,79414,79469,79557,79559,79664,79685,79797,79925,79989,80055,80097,80144,80165,80248,80302,80410,80454,80682,80705,80720,80893,80910,80959,81101,81161,81444,81550,81623,81639,81698,81737,81767,81834,81869,81870,81873,81991,82026,82079,82103,82119,82163,82181,82281,82476,82534,82557,82579,82595,82604,82628,82675,82761,82924,83276,83502,83543,83567,83601,83630,83729,83797,83862,83890,83974,83981,84092,84224,84397,84419,84489,84494,84530,84575,84585,84587,84648,84754,84875,85006,85648,85693,85799,85831,85853,85913,85922,86058,86353,86397,86436,86531,86605,86717,86825,87249,87287,87384,87391,87712,88022,88067,88227,88242,88245,88539,88587,88769,88843,88974,89074,89123,89132,89208,89289,89503,89714,89733,89751,89765,89895,90104,90819,91026,91125,91128,91130,91334,91374,91417,91659,91724,91731,91743,91767,91793,91832,91839,91985,92031,92087,92248,92297,92303,92786,93027,93339,93399,93533,93553,93560,93602,93609,93747,93994,94091,94117,94195,94252,94418,94784,94887,94909,95036,95182,95264,95578,95664,95748,95942,95967,96048,96132,96162,96273,96306,96442,96505,96675,96793,96803,96882,97155,97167,97299,97433,97435,97509,97629,97637,97710,97741,97770,97917,97984,98104,98160,98219,98220,98280,98403,98474,98566,98632,98721,98753,98882,98885,98920,98965,99021,99114,99144,99168,99365,99377,99474,99525,99545,99950,99983",
          "12,42,72,145,188,237,248,498,947,1646,1648,1661,1733,1754,1796,1797,1810,1924,1955,1984,2016,2025,2034,2042,2046,2056,2060,2103,2159,2163,2182,2282,2305,2332,2351,2357,2363,2375,2379,2381,2390,2400,2404,2474,2490,2499,2504,2507,2508,2519,2573,2580,2593,2624,2633,2646,2660,2714,2721,2747,2753,2758,2764,2788,2837,2844,2845,2867,2891,2892,2922,2923,2927,2938,2979,2992,3018,3031,3040,3077,3088,3096,3157,3162,3185,3207,3219,3246,3255,3320,3323,3326,3335,3359,3364,3400,3418,3419,3424,3430,3437,3455,3459,3466,3470,3483,3534,3535,3541,3548,3569,3585,3592,3594,3610,3631,3639,3649,3689,3716,3717,3719,3736,3745,3758,3760,3761,3767,3781,3783,3796,3807,3835,3839,3840,3851,3857,3858,3864,3880,3893,3927,3944,3966,3983,3985,3986,4001,4014,4024,4054,4055,4091,4116,4119,4169,4201,4223,4231,4255,4258,4262,4265,4287,4303,4322,4349,4375,4408,4449,4456,4470,4500,4523,4529,4565,4569,4614,4650,4675,4683,4685,4710,4719,4799,4826,4875,4898,4901,4905,4916,4972,4980,4989,5304,5610,5633,5638,6490,6527,6534,6561,6563,6601,6619,6629,6696,6806,6841,6858,6862,6871,6882,6891,6915,6922,6925,6950,6968,6993,7150,7183,7209,7249,7316,7332,7351,7394,7416,7428,7456,7586,7602,7681,7696,7732,7766,7805,7846,7848,7864,7919,7931,7956,7989,8075,8203,8261,8266,8304,8347,8386,8402,8407,8522,8706,8802,8882,8996,9066,9091,9115,9212,9223,9250,9271,9274,9405,9434,9487,9526,9543,9582,9622,9761,9880,10072,10136,10183,10193,10257,10559,10579,10715,10855,10958,11019,11167,11322,11398,11508,11651,11656,11688,11762,11868,11985,12205,12270,12531,12594,12646,12722,12763,12790,12795,12975,13065,13306,13392,13442,13709,13881,13987,14036,14103,14227,14325,14376,14604,14837,14890,14924,15029,15078,15110,15112,15286,15355,15482,15491,15535,15572,15751,16014,16151,16425,16643,16683,16695,16748,16754,16827,16938,17136,17213,17272,17461,17579,17778,17835,17872,18005,18088,18089,18291,18377,18607,18654,18717,19301,19531,19611,19743,19805,19829,19858,19919,20010,20236,20385,20392,20428,20467,20690,20703,20761,20818,20927,21012,21210,21259,21381,21673,21953,21972,22034,22062,22157,22309,22395,22445,22503,22577,22638,22843,22956,23133,23448,23467,23555,23674,23695,23806,23820,23856,23912,23943,24035,24036,24071,24152,24599,24665,24866,24929,25409,25441,25478,25506,25646,25674,25804,25894,26077,26234,26541,26614,26638,26719,26721,26840,26940,27041,27051,27053,27065,27076,27118,27149,27266,27299,27373,27381,27396,27410,27442,27468,27469,27472,27473,27488,27493,27508,27511,27517,27559,27565,27566,27573,27595,27609,27619,27661,27674,27679,27760,27775,27917,27955,27978,27993,27999,28011,28013,28018,28022,28048,28065,28067,28081,28106,28114,28115,28143,28150,28206,28245,28261,28340,28354,28444,28445,28485,28519,28540,28556,28559,28565,28591,28619,28658,28671,28687,28713,28786,28788,28811,28825,28859,28884,28908,28917,28923,28944,28967,28990,28997,29007,29022,29109,29115,29131,29157,29167,29168,29177,29205,29210,29242,29254,29272,29321,29330,29343,29367,29368,29385,29412,29456,29547,29561,29602,29606,29626,29646,29672,29683,29708,29729,29730,29743,29761,29770,29791,29804,29813,29838,29842,29931,29932,29945,29960,29969,29980,30012,30041,30070,30099,30128,30157,30186,30215,30244,30331,30360,30389,30418,30447,30476,30505,30534,30563,30621,30650,30679,30708,30737,30766,30795,30824,30853,30882,30911,30940,30969,30998,31027,31056,31085,31114,31143,31172,31201,31230,31259,31288,31317,31346,31375,31404,31433,31462,31491,31520,31549,31578,31607,31636,31665,31694,31723,31752,31781,31810,31839,31868,31897,31926,31955,31984,32013,32042,32071,32100,32129,32158,32187,32216,32245,32274,32303,32332,32361,32390,32419,32448,32477,32506,32535,32564,32593,32622,32651,32680,32709,32738,32767,32796,32825,32854,32883,32912,32941,32970,32999,33028,33057,33086,33115,33144,33173,33202,33231,33260,33289,33318,33347,33376,33405,33463,33492,33521,33550,33579,33608,33637,33666,33695,33724,33753,33782,33811,33840,33869,33898,33927,33956,34014,34043,34072,34101,34130,34159,34188,34217,34246,34275,34304,34333,34391,34420,34449,34478,34507,34536,34565,34594,34623,34647,34652,34681,34710,34739,34768,34797,34826,34855,34884,34913,34942,34971,35000,35012,35074,35101,35175,35197,35202,35236,35264,35315,35372,35412,35483,35531,35590,35612,35649,35722,35781,35840,35911,35970,35996,36029,36092,36110,36120,36140,36152,36181,36211,36304,36341,36400,36420,36421,36472,36531,36590,36654,36688,36713,36748,36753,36772,36809,36843,36902,36961,37033,37092,37118,37151,37195,37215,37253,37254,37274,37333,37404,37423,37463,37485,37507,37522,37542,37594,37653,37712,37776,37835,37894,37936,37953,37965,38024,38083,38155,38177,38214,38235,38273,38313,38337,38396,38419,38455,38491,38526,38561,38585,38644,38664,38670,38716,38735,38775,38803,38834,38855,38924,38957,39016,39037,39087,39146,39205,39277,39336,39395,39431,39438,39459,39518,39548,39555,39561,39577,39648,39707,39766,39789,39838,39897,39956,40106,40135,40164,40193,40222,40251,40280,40309,40338,40367,40396,40425,40483,40512,40541,40570,40599,40628,40657,40686,40715,40744,40773,40802,40831,40860,40889,40918,40947,40976,41005,41034,41063,41092,41121,41150,41172,41179,41208,41237,41266,41295,41324,41353,41382,41411,41469,41498,41527,41999,42028,42057,42086,42115,42175,42255,42265,42334,42340,42387,42415,42443,42461,42502,42577,42622,42624,42635,42652,42687,42739,42768,42814,42864,42881,42889,42918,42941,42976,43001,43126,43158,43176,43190,43213,43238,43288,43363,43389,43418,43422,43450,43508,43525,43567,43584,43600,43687,43762,43837,43889,43895,43924,43955,43963,43965,43999,44032,44074,44161,44213,44236,44311,44337,44345,44369,44378,44398,44449,44460,44468,44504,44513,44535,44548,44576,44579,44596,44611,44627,44635,44679,44687,44694,44710,44757,44767,44785,44829,44947,45005,45022,45058,45095,45135,45143,45204,45262,45328,45335,45368,45372,45392,45432,45450,45475,45522,45545,45559,45579,45637,45664,45678,45709,45748,45759,45766,45824,45889,45896,45945,45953,45980,46029,46121,46201,46216,46289,46312,46323,46327,46339,46344,46376,46407,46430,46440,46492,46553,46554,46583,46613,46621,46651,46667,46720,46747,46763,46764,46771,46810,46875,46909,46931,46932,46951,46980,46989,46994,47077,47090,47100,47207,47209,47256,47292,47299,47331,47393,47441,47461,47525,47559,47615,47667,47693,47698,47705,47731,47808,47888,47910,47957,47978,47980,47995,48022,48040,48045,48115,48118,48141,48172,48200,48256,48268,48276,48356,48368,48369,48372,48383,48408,48466,48488,48507,48550,48570,48599,48638,48676,48736,48753,48759,48821,48887,48923,48958,49000,49001,49101,49103,49128,49132,49154,49164,49236,49369,49370,49384,49408,49430,49494,49497,49515,49525,49555,49570,49607,49633,49638,49703,49723,49741,49749,49815,49862,50008,50207,50222,50236,50265,50612,50764,50768,50777,50792,50842,51208,51215,51304,51379,51519,51843,52179,52240,52750,52751,52863,52978,53032,53210,53320,53490,53557,53762,53808,53963,54085,54113,54166,54187,54314,54688,54755,54809,54930,54976,55192,55344,55365,55379,55410,55473,55567,55602,55628,55644,55809,55860,55925,55967,56099,56157,56178,56202,56267,56376,56513,56682,56844,56903,56939,56941,57001,57110,57130,57145,57157,57245,57274,57282,57299,57351,57594,57631,57705,57895,58025,58033,58093,58094,58104,58106,58206,58289,58340,58370,58405,58426,58576,58595,58647,58691,58757,58811,58883,58995,59002,59079,59117,59209,59304,59309,59355,59356,59439,59444,59536,59557,59625,59627,59809,59995,60018,60144,60279,60350,60416,60426,60539,60590,60655,60695,60723,60896,60937,61195,61330,61386,61418,61499,61506,61540,61630,61916,61934,61935,61949,62004,62039,62084,62112,62179,62247,62407,62414,62493,62527,62585,62692,62721,62787,62883,62919,62923,62956,62977,63191,63270,63300,63604,63636,63711,63940,63972,64070,64224,64490,64511,64655,64771,64801,64826,64843,64929,65066,65122,65136,65250,65373,65381,65533,65563,65696,65902,66042,66245,66251,66353,66432,66480,66489,66502,66567,66601,66618,66752,66826,66931,67026,67052,67061,67146,67247,67292,67412,67422,67477,67510,67525,67583,67610,67693,67733,67777,67959,68023,68440,68496,68749,68904,69030,69235,69373,69399,69425,69458,69519,69521,69561,69634,69639,69786,69924,69987,70127,70162,70283,70390,70569,70657,70719,70739,70764,70805,70891,70902,70935,71006,71016,71028,71110,71181,71187,71284,71295,71310,71427,71448,71497,71589,71731,71930,72044,72248,72316,72328,72364,72374,72440,72634,72642,72837,72866,72914,72968,72974,73056,73221,73259,73359,73405,73493,73530,73539,73555,73606,73609,73611,73773,73792,73853,73930,73999,74032,74186,74232,74243,74435,74640,74657,74758,74798,74870,74873,74923,75293,75473,75569,75634,75642,75669,75727,75820,76176,76287,76302,76327,76342,76404,76504,76526,76622,76716,76770,76927,77162,77346,77364,77425,77505,77637,77691,77885,77969,78028,78239,78294,78604,78651,78696,78724,78766,78918,78983,79060,79144,79146,79210,79334,79355,79422,79498,79519,79537,79544,79669,79795,79828,79973,80077,80095,80204,80205,80294,80387,80642,80646,80695,80748,80751,80829,80901,80943,80974,81011,81066,81162,81197,81354,81395,81519,81636,81794,82056,82113,82127,82159,82284,82331,82381,82446,82662,82715,82771,82829,82979,83104,83123,83138,83252,83407,83413,83426,83535,83613,83727,84034,84104,84254,84337,84349,84374,84384,84540,84725,84776,84833,84962,85174,85276,85379,85532,85585,85617,85675,85780,85887,86044,86099,86164,86198,86224,86306,86355,86423,86478,86525,86532,86543,86595,86689,86691,86714,86719,86734,86802,86886,86963,86986,87094,87128,87192,87395,87653,87683,87827,87879,87891,87956,88171,88262,88268,88277,88416,88468,88480,88626,88629,88669,88789,88803,88835,89031,89100,89154,89174,89206,89226,89325,89327,89331,89573,89660,89762,89774,89777,90050,90064,90385,90402,90573,90691,90740,90914,91031,91044,91103,91205,91226,91420,91561,91606,91631,91759,91930,91951,91953,92034,92038,92118,92372,92439,92441,92512,92578,92670,92681,92830,92909,93030,93244,93458,93595,93727,94050,94222,94263,94284,94346,94362,94432,94602,94905,94964,94976,94995,95072,95079,95167,95291,95367,95422,95581,95698,95707,95901,96178,96201,96227,96276,96299,96356,96456,96501,96685,96724,96801,96825,96857,96924,97045,97054,97085,97107,97461,97479,97502,97571,97838,97970,98278,98313,98394,98495,98567,98646,98700,98701,98730,98771,98786,98796,98845,98846,99049,99128,99181,99309,99316,99592,99739,99757,99844,99906,99934,99958",
          "176,276,477,608,637,652,665,668,726,738,752,827,897,943,974,1045,1136,1153,1162,1208,1255,1267,1335,1424,1457,1473,1480,1559,1592,1597,1611,1615,1635,1644,1674,1748,1841,2030,2188,2195,2589,2638,2693,2742,2773,2792,2799,2902,2928,3046,3103,3126,3160,3179,3226,3333,3647,3667,3706,3792,3828,4198,4210,4319,4333,4354,4559,4582,4667,4671,4720,4729,4749,4945,5008,5035,5084,5110,5122,5138,5140,5178,5190,5201,5203,5205,5241,5255,5259,5298,5301,5327,5339,5388,5400,5512,5514,5578,5641,5706,5742,5745,5865,5938,5969,6003,6031,6038,6058,6148,6182,6185,6218,6229,6251,6468,6522,6548,6597,6822,7077,7080,7187,7286,7288,7299,7405,7411,7476,7486,7525,7561,7666,7714,7741,7780,7879,7894,7905,7927,8005,8029,8120,8245,8382,8501,8679,8728,8816,9077,9230,9313,9344,9374,9383,9385,9423,9428,9736,9815,9898,10003,10287,10337,10425,10439,10464,10475,10532,10687,10874,10934,11274,11378,11411,11463,11673,11746,11870,12383,12433,12470,12527,13037,13105,13361,13440,13935,13955,14063,14339,14530,14607,14828,15340,15401,15472,15804,15890,16005,16616,16714,16867,16955,17032,17054,17064,17223,17518,17871,18881,19281,19370,19479,19720,20034,20264,20408,20663,20757,20896,21499,21544,21692,21842,22521,22796,23783,23829,24452,24812,25070,25093,25171,26162,26739,26774,26901,26909,27219,27432,27525,27854,28190,28210,28229,28244,28289,28709,28728,28732,29277,29991,30333,30596,30681,31522,31696,32363,32537,33001,33291,33320,33849,34016,34422,35039,35450,35621,35993,36446,36499,37248,37382,37444,37692,37837,38128,38132,38157,38187,38304,38492,38646,38744,38959,39004,39118,39173,39315,39360,39557,39626,39679,39732,39875,39992,41517,42103,42493,42610,42932,43164,43467,43519,43520,43654,44260,44585,45231,45370,45437,45556,45558,45732,45737,45874,45879,45919,45998,46126,46984,47982,48084,49754,49849,49996,50104,50106,50119,50251,50252,50341,50433,50511,50673,50718,50733,50814,50827,50985,51090,51138,51178,51206,51242,51265,51359,51705,51942,51990,52203,52257,52323,52325,52412,52427,52531,52567,52636,52663,52799,53025,53191,53349,53595,53677,53702,54261,54455,54464,54496,54660,54715,54889,54988,55095,55117,55121,55147,55300,55393,55498,55549,55647,55690,55703,55715,55795,56089,56105,56253,56293,56396,56412,56470,56550,56789,57010,57030,57047,57070,57080,57200,57206,57293,57373,57422,57481,57494,57535,57571,57893,57896,57969,57977,58092,58115,58361,58496,58524,58557,58799,58859,58879,58898,58915,58924,59042,59058,59204,59226,59446,59549,59566,59798,59837,59865,59872,59979,60002,60219,60399,60403,60454,60476,60634,60733,60744,60848,60914,61004,61066,61077,61100,61134,61224,61235,61296,61361,61425,61445,61578,61796,61858,62067,62164,62246,62277,62288,62366,62368,62401,62408,62555,62656,62962,62995,63026,63028,63042,63090,63205,63282,63349,63401,63410,63507,63554,63611,63621,63683,63735,63761,63826,63860,63907,63988,64003,64077,64106,64143,64227,64277,64523,64528,64580,64786,64829,64849,64968,64978,64981,65054,65061,65121,65237,65437,65452,65656,65709,65784,65925,66055,66167,66216,66226,66312,66347,66430,66515,66577,66580,66603,66662,66805,66888,67116,67215,67262,67415,67435,67751,67888,67896,67950,68015,68071,68137,68197,68483,68527,68608,68628,68789,68805,68842,68892,68916,69182,69185,69216,69264,69481,69495,69567,69568,69789,69821,69881,69959,70068,70078,70087,70126,70138,70211,70276,70281,70310,70315,70378,70415,70497,70505,70552,70616,70665,70846,70936,70984,70994,70998,71029,71108,71156,71373,71378,71423,71513,71561,71580,71668,71678,71822,71828,71895,72014,72043,72116,72128,72141,72204,72252,72486,72497,72552,72555,72579,72657,72732,72890,73001,73291,73301,73326,73333,73496,73497,73756,73941,73964,74272,74357,74638,74730,74929,75085,75096,75134,75140,75170,75192,75274,75302,75332,75427,75435,75472,75584,75606,75718,75792,75866,75912,75976,76014,76077,76185,76264,76286,76366,76450,76502,76561,76578,76650,76849,76850,76892,76915,77170,77177,77199,77220,77242,77302,77370,77400,77452,77473,77650,77684,77733,77769,78402,78515,78545,78603,78740,78754,78988,79008,79059,79128,79133,79158,79196,79284,79338,79421,79439,79488,79533,79752,79791,79867,79936,79969,80026,80154,80171,80197,80208,80210,80229,80270,80369,80699,80886,80890,81003,81010,81025,81093,81309,81412,81423,81483,81813,81960,82152,82171,82183,82289,82333,82396,82405,82684,82693,82766,82781,82911,82921,82959,83092,83233,83358,83364,83400,83422,83562,83646,83700,83711,83807,83878,83952,84009,84047,84067,84157,84292,84368,84552,84621,84794,84796,84864,85109,85279,85376,85381,85472,85515,85529,85716,85729,85784,85807,85875,85935,86120,86123,86162,86331,86388,86391,86407,86723,86805,86837,86863,86883,86905,86924,86989,87376,87379,87478,87498,87575,87634,87681,87937,88039,88063,88202,88220,88254,88311,88399,88408,88411,88430,88526,88614,88639,88652,88659,88696,88870,89178,89553,89564,89641,89663,89863,90012,90114,90192,90279,90449,90560,90627,90760,90788,90880,90886,91002,91168,91364,91563,91564,91600,91626,91641,91673,91856,91883,91896,91971,92019,92020,92183,92243,92445,92616,92752,92777,92979,93013,93062,93114,93151,93262,93680,93838,94051,94106,94114,94273,94422,94456,94529,94540,94596,94718,94863,95018,95070,95438,95616,95768,95797,95869,95947,95979,95984,96035,96083,96163,96190,96208,96253,96339,96355,96384,96416,96450,96480,96498,96723,96880,97033,97059,97110,97180,97193,97273,97377,97429,97619,97626,97842,97851,98029,98061,98157,98236,98329,98346,98372,98430,98516,98699,98788,98892,98998,99051,99112,99137,99226,99347,99389,99439,99535,99548,99588,99646,99673,99755,99762,99812,99901,99926",
          "13,284,288,351,360,445,453,541,557,612,680,695,737,758,770,775,798,801,855,902,963,964,971,983,1049,1129,1300,1332,1453,1461,1472,1536,1543,1564,1753,1802,1812,1831,1855,1898,1905,1908,2816,3158,3949,4759,5071,5148,5236,5411,5588,5627,5723,5831,5963,6131,6411,6566,6620,6712,6845,6962,7000,7060,7071,7086,7112,7126,7190,7203,7278,7321,7341,7376,7390,7435,7507,7514,7526,7531,7557,7569,7654,7660,7672,7701,7711,7745,7756,7785,7800,7858,7967,8027,8037,8065,8070,8110,8213,8257,8267,8303,8306,8334,8344,8372,8484,8513,8637,8746,8807,8856,8908,8934,9043,9073,9415,9440,9603,9890,9958,10026,10125,10144,10209,10284,10382,10506,10650,10658,10732,10920,11132,11431,11500,11605,11773,11823,11842,11893,12075,12110,12230,12379,12540,12614,12728,12832,12878,12976,13299,13582,13825,13948,14118,14387,14609,14624,14708,15136,15231,15987,16091,16837,17121,17287,17388,18473,18561,18659,19266,20090,20137,20696,20857,20952,21248,21337,21527,21687,22039,22143,22616,22854,22975,23317,23397,23562,23638,24033,24132,24161,24685,24848,25182,25362,25575,25611,25806,26617,26701,26791,26842,26853,27243,27647,27705,27742,27756,27861,28248,28267,28371,28431,29341,29361,29431,29578,29953,32367,34221,35516,36116,36173,36264,36694,36830,36981,36993,37052,37198,37320,37429,37504,38136,38196,38244,38700,39444,39560,39671,39884,42034,44367,45127,45297,45360,45924,45975,48882,50003,50016,50077,50218,50282,50309,50367,50666,50669,50696,50727,50857,50943,50966,50997,51000,51086,51093,51181,51237,51333,51389,51400,51455,51561,51626,51671,51711,51774,51794,51842,51853,51866,51891,52199,52553,52573,52670,52748,52827,52855,52946,52952,53058,53189,53305,53347,53375,53378,53380,53409,53507,53669,53911,54269,54352,54362,54364,54433,54449,54486,54563,54583,54665,54764,54948,55002,55022,55037,55233,55247,55445,55447,55485,55670,55799,55857,55868,55875,56032,56112,56124,56189,56348,56356,56365,56378,56730,56870,56961,56981,57100,57152,57216,57259,57316,57339,57366,57468,57525,57536,57575,57608,57686,57783,57840,57849,57879,57892,57902,58046,58054,58137,58156,58216,58224,58246,58251,58294,58417,58522,58594,58599,58678,58682,58735,58801,58810,58903,59164,59350,59472,59521,59810,59871,59927,59985,60245,60305,60340,60354,60544,60638,60648,60791,60919,60973,61272,61297,61480,61483,61493,61642,61659,61780,61897,61913,61953,62426,62433,62483,62674,62679,62693,62964,63108,63428,63499,63726,63741,63769,63983,63992,64085,64212,64229,64282,64306,64496,64597,64700,64813,64872,64902,64917,65048,65245,65254,65308,65487,65506,65534,65598,65621,65646,65648,65666,65682,65735,65808,65873,65881,66129,66439,66466,66550,66821,66829,66835,66853,66934,67070,67218,67336,67342,67377,67410,67425,67613,67976,67994,68232,68408,68495,68523,68710,68743,68928,69006,69057,69108,69135,69199,69277,69332,69461,69484,69502,69535,69538,69558,69566,69573,69625,69769,69785,69819,69911,69956,70030,70045,70086,70219,70249,70347,70517,70628,70673,70729,70818,71009,71022,71173,71205,71316,71341,71362,71376,71628,71763,72021,72048,72072,72136,72214,72227,72286,72291,72507,72772,72839,72859,72878,72928,72930,72936,73185,73563,73718,73905,73935,73938,73984,74024,74034,74045,74082,74088,74289,74373,74665,74713,74924,74927,74984,75086,75108,75167,75196,75366,75429,75470,75513,75688,75833,75951,76001,76018,76145,76241,76323,76369,76483,76495,76605,76616,76708,76807,76832,76856,76956,76986,77025,77033,77126,77176,77255,77293,77307,77423,77549,77609,77647,77949,78042,78076,78169,78188,78222,78230,78264,78281,78300,78398,78456,78934,79046,79189,79219,79318,79337,79361,79500,79538,79563,79576,79582,79592,79615,79892,79957,79959,79985,80016,80039,80047,80049,80101,80157,80262,80335,80617,80637,80852,80891,81049,81064,81074,81114,81181,81296,81375,81388,81408,81448,81584,81624,81969,82108,82110,82133,82202,82232,82239,82243,82337,82670,82681,82699,82744,82760,82899,83208,83449,83488,83511,83590,83660,83769,83827,83916,83960,84051,84080,84232,84350,84405,84442,84479,84486,84509,84604,84618,84632,84738,84779,84809,84829,84836,84886,85140,85142,85269,85329,85363,85369,85504,85667,85672,85714,85816,85865,85965,86047,86052,86109,86214,86235,86248,86455,86536,86671,86792,86918,86985,87005,87006,87084,87141,87158,87200,87243,87305,87463,87528,87654,87805,87857,87940,88000,88013,88286,88304,88358,88423,88500,88504,88632,88663,88682,88784,88849,88933,89055,89127,89148,89205,89569,89605,89620,89671,89697,89735,89771,89773,89802,89838,89910,89993,90081,90305,90409,90444,90479,90537,90555,90559,90666,90695,90718,90859,91086,91173,91298,91410,91542,91555,91825,91865,91869,91870,91916,91959,92017,92059,92126,92298,92353,92477,92623,92652,92669,92790,92866,92882,92928,93016,93024,93052,93135,93231,93409,93459,93693,93714,93996,93999,94059,94100,94177,94583,94609,94686,94723,94725,94733,94770,94776,94859,94898,94912,94942,95090,95131,95145,95388,95426,95483,95761,95918,95975,95992,96012,96478,96542,96623,96733,96755,96858,97018,97057,97103,97133,97248,97354,97441,97463,97570,97594,97611,97622,97781,97795,97817,97931,97964,98179,98189,98231,98296,98556,98647,98906,98976,99002,99074,99120,99154,99236,99404,99408,99441,99460,99485,99697,99705,99738,99782,99981",
          "61,69,76,90,128,130,1660,1684,1722,1760,1821,1925,2160,2216,2306,2513,2657,2722,2789,2834,2861,2957,2976,3072,3093,3147,3215,3232,3287,3308,3332,3612,3619,3749,3824,4078,4161,4293,4367,4374,4427,4455,4465,4469,4506,4603,4634,4649,4738,4769,4785,4787,5000,5161,6377,6659,7156,7977,8030,8049,8125,8151,8315,8429,8505,8562,8581,8638,8676,9037,9056,9075,9588,9602,9626,9664,9683,9702,9724,9778,9816,9837,9987,9988,10006,10082,10120,10177,10405,10610,10614,10633,10710,10747,10804,10828,10896,10956,10975,11032,11051,11222,11241,11298,11338,11409,11412,11465,11473,11474,11514,11523,11596,11621,11640,11735,11963,12001,12050,12058,12066,12089,12115,12153,12168,12202,12229,12248,12259,12267,12286,12343,12349,12381,12430,12457,12459,12538,12628,12697,12704,12761,12802,12875,12977,12979,12989,13028,13084,13100,13120,13170,13174,13190,13236,13241,13274,13293,13312,13350,13464,13483,13597,13616,13635,13692,13730,13749,13768,13787,13844,13882,13939,13958,13977,14034,14041,14110,14167,14205,14243,14262,14338,14346,14357,14395,14414,14433,14452,14471,14528,14557,14566,14613,14642,14680,14718,14756,14762,14775,14832,14870,14908,14912,14916,14927,14965,15003,15022,15041,15057,15060,15079,15081,15155,15193,15250,15269,15307,15326,15364,15383,15389,15421,15497,15516,15592,15646,15649,15687,15744,15746,15763,15782,15820,15839,15858,15896,15929,15953,15974,15991,16029,16048,16143,16149,16176,16181,16191,16197,16200,16238,16257,16276,16295,16314,16352,16390,16409,16438,16447,16466,16507,16535,16542,16561,16580,16619,16656,16727,16751,16789,16808,16865,16903,16960,17017,17055,17112,17150,17188,17245,17295,17302,17305,17321,17378,17435,17473,17511,17549,17574,17599,17644,17672,17739,17796,17815,17834,17853,17891,17910,17929,17986,18024,18043,18119,18138,18157,18213,18214,18290,18309,18347,18366,18385,18404,18480,18518,18556,18613,18651,18689,18716,18746,18774,18791,18860,18879,18898,18904,18999,19031,19050,19063,19107,19126,19141,19221,19240,19278,19322,19339,19354,19368,19373,19392,19430,19441,19449,19506,19525,19537,19553,19561,19563,19582,19590,19620,19658,19734,19748,19753,19772,19791,19867,19886,19905,19924,20000,20019,20026,20038,20058,20101,20114,20133,20190,20247,20285,20304,20323,20342,20418,20456,20494,20551,20564,20589,20627,20633,20646,20649,20665,20741,20779,20817,20836,20893,20961,20969,21007,21064,21083,21178,21183,21216,21257,21270,21292,21301,21313,21330,21340,21349,21387,21425,21444,21463,21518,21539,21577,21596,21640,21656,21691,21710,21729,21767,21786,21805,21817,21843,21900,21919,21938,21957,21990,21995,22014,22033,22052,22057,22090,22109,22130,22147,22152,22166,22204,22242,22261,22318,22324,22394,22413,22470,22489,22508,22527,22603,22641,22698,22755,22774,22793,22812,22888,22907,22912,22926,22954,23002,23040,23059,23062,23078,23116,23128,23154,23158,23173,23267,23268,23287,23306,23325,23439,23469,23477,23485,23496,23513,23515,23533,23534,23553,23572,23607,23619,23629,23650,23686,23698,23721,23724,23735,23743,23800,23838,23857,23876,23887,23895,23914,23933,23971,23990,24047,24066,24085,24142,24171,24180,24187,24188,24199,24218,24256,24289,24295,24332,24351,24389,24427,24465,24522,24541,24560,24579,24617,24636,24674,24693,24731,24769,24788,24845,24860,24876,24902,24921,24959,24977,24997,25013,25035,25047,25054,25092,25149,25168,25206,25263,25277,25320,25339,25396,25404,25407,25415,25460,25472,25491,25529,25548,25586,25616,25662,25700,25739,25758,25776,25814,25833,25852,25890,25947,26004,26023,26042,26118,26137,26460,26477,26479,26517,26593,26745,26775,26859,26896,26916,26928,26935,26946,27079,27087,27106,27144,27158,27182,27220,27258,27276,27332,27353,27486,27553,27554,27581,27594,27656,27695,27714,27727,27809,27942,27980,28037,28075,28162,28238,28375,28391,28417,28436,28490,28547,28550,28569,28714,28751,28911,29063,29099,29228,29285,29320,29329,29416,29469,29500,29519,29858,29886,29907,29994,30013,30042,30071,30100,30129,30158,30187,30216,30245,30274,30303,30332,30361,30390,30419,30448,30477,30506,30535,30564,30593,30622,30651,30680,30709,30738,30767,30796,30825,30854,30883,30912,30941,30970,30999,31028,31057,31086,31115,31144,31173,31202,31231,31260,31289,31318,31347,31376,31405,31434,31463,31492,31521,31550,31579,31608,31637,31666,31695,31724,31753,31782,31811,31840,31869,31898,31927,31985,32014,32043,32072,32101,32130,32159,32188,32246,32275,32304,32333,32362,32391,32420,32449,32478,32507,32536,32565,32594,32623,32652,32681,32710,32739,32768,32797,32826,32855,32884,32913,32942,32971,33000,33029,33058,33087,33116,33145,33174,33203,33232,33261,33290,33319,33348,33377,33406,33435,33464,33493,33522,33551,33580,33609,33638,33667,33696,33725,33754,33783,33812,33841,33870,33899,33928,33957,33986,34015,34044,34073,34102,34131,34160,34189,34218,34247,34276,34305,34334,34363,34392,34421,34450,34479,34508,34537,34566,34595,34624,34653,34682,34711,34740,34769,34798,34827,34856,34885,34914,34943,34972,35013,35102,35121,35163,35176,35237,35265,35284,35306,35320,35323,35339,35364,35413,35434,35459,35484,35514,35532,35626,35650,35668,35674,35723,35758,35782,35841,35912,35971,36030,36093,36128,36153,36192,36212,36283,36329,36342,36369,36401,36473,36504,36532,36591,36655,36714,36751,36773,36844,36903,36962,36992,37034,37093,37113,37125,37152,37173,37179,37185,37216,37334,37368,37433,37464,37523,37595,37654,37713,37741,37777,37796,37836,37895,37966,38002,38025,38063,38084,38111,38143,38156,38215,38274,38338,38397,38456,38474,38527,38586,38667,38717,38734,38763,38776,38835,38899,38958,38996,39017,39075,39088,39125,39147,39172,39206,39278,39307,39324,39337,39396,39437,39460,39496,39519,39578,39597,39605,39613,39649,39708,39860,39898,39925,39957,39986,40123,40152,40181,40210,40239,40268,40297,40326,40355,40384,40413,40442,40471,40500,40529,40558,40587,40616,40645,40674,40703,40732,40761,40790,40819,40848,40877,40906,40935,40964,40993,41022,41051,41080,41109,41167,41196,41225,41254,41283,41312,41341,41370,41399,41428,41457,41486,41515,41544,42072,42101,42130,42190,42219,42223,42229,42280,42297,42355,42390,42394,42430,42480,42487,42517,42540,42592,42604,42667,42733,42754,42807,42829,42848,42857,42904,42991,43040,43066,43091,43141,43228,43303,43350,43378,43465,43480,43540,43551,43553,43558,43564,43582,43615,43683,43702,43716,43727,43777,43852,43871,43939,43970,43986,44014,44059,44066,44089,44103,44112,44115,44132,44148,44176,44194,44211,44251,44258,44288,44326,44394,44413,44424,44488,44538,44563,44608,44650,44725,44733,44745,44769,44800,44831,44834,44887,44905,44908,44962,45037,45079,45080,45110,45158,45219,45277,45350,45382,45407,45465,45476,45537,45549,45594,45652,45724,45781,45839,45861,45911,45968,45994,46051,46060,46256,46263,46315,46333,46345,46362,46393,46405,46409,46414,46419,46442,46459,46460,46481,46496,46540,46587,46690,46757,46790,46802,46814,46824,46839,46840,46910,46934,46990,46996,47020,47021,47083,47097,47098,47123,47126,47138,47201,47244,47348,47352,47438,47440,47464,47478,47479,47492,47516,47589,47602,47642,47670,47704,47770,47842,47848,47909,47963,47965,47984,48046,48065,48211,48290,48308,48346,48365,48467,48525,48527,48535,48558,48560,48564,48601,48603,48698,48699,48734,48741,48750,48777,48782,48820,48853,48867,48872,48912,48939,48955,48959,48977,48986,49012,49020,49036,49063,49075,49146,49151,49166,49258,49266,49278,49321,49337,49349,49350,49367,49407,49411,49435,49448,49596,49612,49629,49632,49668,49672,49715,49756,49788,49790,49816,49859,49930,49944,49951,49974,49985,50151,50209,50228,50230,50247,50302,50339,50409,50474,50587,50656,50790,51013,51056,51067,51223,51307,51504,51718,51742,52015,52087,52208,52242,52292,52294,52403,52474,52501,52610,52858,52899,52991,53048,53215,53237,53245,53314,53352,53368,53451,53520,53589,53605,53619,53647,53809,53863,53864,53966,54040,54077,54082,54189,54192,54223,54295,54373,54410,54445,54485,54784,54912,54940,55142,55307,55310,55346,55388,55409,55412,55423,55449,55929,55938,55966,56110,56272,56359,56413,56473,56578,56688,56872,56881,56888,56913,56927,56989,57445,57464,57497,57500,57803,57860,57890,57949,58038,58063,58099,58129,58205,58235,58248,58266,58277,58282,58329,58389,58451,58622,58632,58670,58721,58920,58992,59094,59099,59183,59195,59266,59300,59440,59502,59685,59759,59818,59917,60032,60055,60198,60225,60307,60324,60349,60541,60652,60689,60734,60905,60956,60965,61271,61371,61488,61527,61537,61553,61663,61705,61756,61767,62016,62410,62425,62661,62688,62711,62850,62889,62925,63183,63287,63357,63659,63783,63828,63862,63950,63953,64060,64100,64107,64136,64236,64239,64290,64299,64361,64450,64467,64532,64586,64641,64671,64746,64774,64830,64928,64998,65093,65108,65159,65364,65422,65446,65547,65733,65775,65997,66266,66355,66374,66581,66701,66801,66815,66890,66948,67067,67330,67359,67388,67423,67643,67665,67708,67709,67884,67910,67938,67977,67992,68052,68236,68250,68272,68435,68554,68620,68624,68751,68756,68827,68849,68853,68938,68961,69112,69124,69223,69236,69342,69390,69392,69489,69666,69711,69837,69873,69919,69986,70059,70135,70259,70280,70358,70368,70377,70399,70538,70620,70636,70650,70660,70754,70869,70888,70950,70969,71055,71092,71282,71352,71599,71603,71707,71720,71741,71766,71780,71918,71967,72056,72165,72182,72382,72421,72458,72475,72527,72547,72631,72692,72787,72821,72826,72929,72963,73168,73532,73550,73616,73668,73830,73884,73936,74168,74325,74587,74601,74611,74667,74668,74689,74727,74753,74771,74781,74847,74888,75133,75422,75428,75549,75615,75632,75663,75678,75683,75778,75829,75851,75897,75908,75928,76038,76262,76359,76425,76494,76496,76523,76614,76669,76687,76905,76952,76974,77062,77193,77360,77446,77538,77561,77584,77585,77841,77850,77918,77923,77936,78254,78454,78459,78598,78663,78911,79067,79092,79105,79155,79351,79553,79692,79695,79771,79778,79805,79857,79875,79880,80090,80099,80151,80206,80474,80607,80782,80786,80822,81235,81241,81260,81292,81304,81311,81336,81608,81609,81615,81814,81821,81942,82019,82062,82124,82131,82166,82203,82244,82260,82285,82375,82434,82520,82617,82746,82756,82763,82847,82945,82952,82957,82993,83076,83089,83213,83216,83225,83249,83308,83314,83379,83440,83451,83539,83540,83691,83858,84003,84010,84063,84176,84242,84436,84502,84747,84749,84760,84838,84857,84993,85015,85089,85193,85364,85465,85578,85957,86049,86229,86290,86425,86549,86617,86625,86704,86726,86764,86819,86983,86984,87201,87330,87362,87368,87496,87540,87599,87609,87704,87909,88056,88074,88147,88394,88437,88446,88570,88580,88643,88657,88688,88721,88832,88913,88937,88940,88948,89164,89302,89476,89515,89522,89602,89755,89761,89877,89939,89962,89966,89977,90028,90029,90052,90060,90194,90366,90393,90398,90429,90442,90524,90596,90655,90680,90747,90838,90934,90973,90995,91155,91293,91377,91513,91540,91548,91607,91727,91820,91838,91931,92005,92285,92370,92401,92443,92505,92589,92625,92768,92773,92780,92811,92901,92945,92987,93123,93152,93212,93215,93267,93393,93556,93617,93645,93666,93741,93932,94052,94067,94075,94150,94191,94257,94354,94447,94517,94562,94642,94663,94727,94743,94785,95016,95149,95349,95724,95928,95930,96134,96191,96298,96452,96476,96520,96601,96657,96699,96778,96873,96962,97010,97064,97159,97189,97221,97277,97327,97465,97498,97520,97707,97726,97916,98030,98055,98075,98171,98202,98240,98255,98400,98639,98674,98697,98735,98999,99068,99113,99209,99219,99233,99247,99341,99361,99507,99517,99579,99634,99649,99650,99793,99878",
          "35,177,304,312,391,439,448,483,577,601,633,672,674,800,803,835,851,853,871,877,879,884,885,909,913,991,997,1003,1009,1012,1019,1025,1094,1134,1190,1231,1238,1240,1287,1294,1319,1320,1373,1382,1406,1412,1532,1851,1876,1957,2161,2219,2227,2231,2414,2447,2526,2745,2828,2887,2905,2941,3114,3155,3196,4800,5091,5097,5117,5212,5532,5570,5617,5618,5637,5640,5682,5740,5783,5850,5966,5991,5992,6029,6098,6100,6168,6193,6312,6379,6482,6687,6800,7038,7041,7152,7314,7330,7372,7521,7556,7620,7629,7639,7641,7827,8092,8171,8178,8238,8299,8326,8385,8485,8497,8570,8651,8833,9090,9098,9120,9206,9445,9891,9990,10055,10222,10250,10265,10329,10340,10416,10466,10496,10551,10564,10761,10778,10785,10852,10964,11096,11278,11778,11922,12081,12101,12320,13029,13839,13840,13897,14378,14585,14610,14733,15069,15070,15085,15131,15316,15431,15522,15999,16107,16471,16519,16664,16732,16744,16773,16776,17582,18188,18522,18683,18884,18889,19480,19627,20122,20248,20496,21039,21060,21125,21298,22179,23043,23773,24277,24679,24789,24826,24973,25124,25632,25871,25972,26112,26272,26347,26680,26694,27082,27156,27324,27572,28272,28749,28787,29191,29542,31732,33757,35706,35953,36263,36408,36676,36868,37616,37640,37951,38050,38197,38323,38511,38938,38941,39003,39191,39295,39359,39445,39691,39749,39800,39823,42275,42997,45120,45792,45808,45918,47364,50102,50134,50135,50435,50446,50580,50585,50619,50654,50676,50698,50832,50944,51031,51063,51207,51227,51327,51366,51411,51548,51593,51730,51807,51832,51884,51901,51954,51961,51996,51997,52001,52378,52415,52416,52443,52614,52624,52654,52699,52757,52816,52820,52832,52892,52921,53093,53173,53227,53635,53638,53708,53787,53807,53842,53907,53910,53971,54047,54152,54213,54277,54286,54328,54343,54347,54348,54399,54440,54526,54735,54742,54796,54826,54926,54965,55094,55190,55292,55297,55448,55609,55673,55710,55722,55727,55750,55904,55996,56087,56156,56342,56447,56598,56748,56752,56857,56999,57011,57063,57160,57311,57325,57419,57446,57447,57543,57615,57663,57784,57847,57866,57894,57932,57941,58144,58198,58228,58293,58367,58415,58540,58546,58684,58738,58762,58773,58814,58837,58994,59123,59224,59231,59315,59335,59410,59412,59425,59495,59611,59661,59709,59844,60504,60712,60816,60941,61084,61219,61293,61340,61416,61466,61591,61595,61782,61863,61866,61878,61965,61986,62015,62021,62059,62185,62195,62413,62454,62499,62530,62665,62791,62875,62912,62934,62952,63095,63169,63224,63246,63304,63437,63454,63504,64130,64182,64292,64351,64436,64453,64484,64489,64603,64688,64689,64716,64761,64825,64867,65146,65183,65204,65240,65275,65289,65390,65440,65459,65548,65587,65618,65620,65669,65679,65715,65720,65946,66205,66210,66236,66275,66371,66376,66418,66486,66704,66982,67033,67082,67135,67164,67177,67270,67306,67338,67351,67513,67682,67725,67812,67829,67932,67970,68043,68047,68051,68054,68208,68362,68383,68425,68529,68560,68660,68828,68858,68955,68991,69122,69179,69184,69237,69327,69360,69491,69526,69653,69677,69728,69869,69871,69890,69908,69993,70235,70316,70346,70404,70485,70623,70828,70899,71017,71054,71082,71111,71596,71813,71837,71972,72231,72310,72424,72452,72468,72470,72519,72533,72596,72669,72672,72684,72791,72814,72971,73066,73114,73411,73505,73520,73524,73693,73754,73838,73850,73892,73907,73949,74000,74028,74044,74208,74326,74612,74694,74733,75123,75145,75226,75379,75415,75520,75573,75602,75715,75764,75823,75837,75975,76128,76204,76207,76335,76387,76395,76477,76529,76792,76844,76909,76926,76995,77158,77164,77241,77336,77337,77447,77624,77657,77812,78123,78129,78165,78216,78260,78270,78415,78466,78511,78594,78610,78666,78725,78764,78871,78959,79086,79114,79165,79182,79184,79374,79471,79567,79598,79633,79655,79665,79762,79780,79811,79816,79879,79907,79918,79928,79943,79952,79986,80071,80201,80239,80268,80310,80362,80395,80439,80450,80615,80640,80688,80818,80871,80955,81081,81142,81148,81202,81207,81271,81368,81552,81720,81833,81868,81885,81953,82049,82054,82208,82290,82304,82360,82472,82501,82717,82856,82914,83049,83113,83144,83357,83411,83415,83439,83501,83594,83602,83640,83696,83839,83918,83934,84054,84097,84150,84166,84255,84408,84432,84481,84499,84516,84525,84607,84622,84804,84914,84991,84996,85034,85295,85584,85633,85724,85762,85874,85923,85931,85993,86222,86240,86327,86675,86880,87014,87122,87129,87218,87348,87630,87924,88104,88191,88217,88234,88327,88349,88507,88525,88528,88746,88810,88823,88841,88988,89058,89129,89282,89317,89337,89577,89633,89876,89879,89969,90108,90197,90198,90278,90358,90483,90502,90599,90612,90722,90753,90759,90810,90954,91047,91071,91271,91340,91466,91503,91584,91638,91665,91795,91834,91894,91897,91943,91972,92133,92150,92221,92388,92405,92456,92476,92660,92696,92725,92751,92759,92779,92801,92816,92842,92859,92920,92982,93079,93349,93395,93457,93461,93620,93622,93735,93928,93947,94048,94088,94159,94206,94264,94306,94361,94375,94430,94539,94565,94690,94777,94813,94876,94945,95088,95135,95204,95211,95273,95331,95357,95425,95446,95552,95572,95694,95716,95767,95868,95870,96260,96433,96541,96589,96709,96712,96725,96752,96912,96920,97378,97408,97540,97593,97664,97689,97870,97872,98081,98109,98145,98181,98224,98317,98489,98504,98813,98925,99373,99421,99553,99708,99721,99800,99806,99852,99868",
          "78,148,228,229,1649,1731,1736,1769,2035,2094,2155,2156,2230,2415,2429,2628,2679,2727,2746,2829,2852,2870,2983,3193,3214,3425,3479,3517,3574,3793,4314,4328,4342,4517,4797,4851,4858,4986,5080,6052,7524,9479,9488,9597,9711,9749,9806,9920,9977,10262,10604,10642,10680,10756,10767,10832,10889,11003,11022,11117,11193,11212,11231,11242,11250,11264,11342,11364,11459,11498,11516,11585,11592,11649,11801,11858,11915,11941,12048,12086,12124,12143,12162,12219,12238,12257,12276,12295,12313,12314,12341,12371,12390,12428,12485,12514,12523,12580,12599,12694,12713,12741,12827,12881,12960,12998,13017,13036,13043,13045,13074,13093,13131,13169,13206,13207,13239,13245,13264,13278,13302,13321,13393,13394,13397,13416,13428,13435,13518,13530,13568,13587,13644,13663,13689,13739,13758,13777,13796,13830,13910,14005,14024,14078,14081,14092,14119,14157,14176,14195,14233,14252,14271,14290,14309,14328,14343,14366,14385,14404,14441,14499,14556,14575,14661,14669,14670,14727,14784,14841,14879,14917,14943,14955,15012,15031,15050,15088,15107,15126,15164,15183,15219,15221,15297,15335,15354,15373,15378,15411,15430,15468,15506,15525,15582,15620,15639,15658,15696,15711,15734,15772,15810,15829,15848,15867,15886,15943,15962,16000,16019,16037,16076,16133,16152,16171,16209,16228,16247,16266,16285,16304,16323,16342,16361,16380,16418,16456,16475,16513,16532,16589,16601,16608,16627,16665,16703,16741,16760,16763,16779,16830,16855,16874,16893,16950,16969,16988,17039,17041,17045,17093,17140,17197,17216,17290,17292,17330,17349,17387,17406,17425,17444,17463,17501,17520,17596,17615,17691,17710,17729,17748,17805,17824,17843,17862,17881,17900,17919,17926,17957,17995,18033,18071,18128,18185,18195,18204,18242,18260,18261,18299,18318,18356,18394,18413,18432,18489,18508,18527,18603,18679,18736,18751,18755,18793,18803,18812,18850,18869,18907,18945,18964,18983,19002,19040,19097,19116,19135,19204,19249,19306,19325,19344,19382,19420,19458,19496,19534,19567,19572,19610,19648,19667,19686,19705,19762,19781,19800,19819,19876,19895,19914,19933,19952,19990,20047,20085,20104,20123,20142,20161,20171,20180,20237,20275,20294,20313,20332,20351,20370,20389,20427,20446,20465,20484,20499,20529,20541,20548,20560,20572,20598,20616,20617,20636,20655,20660,20712,20731,20750,20786,20788,20864,20902,20921,20978,21016,21054,21092,21111,21168,21187,21206,21221,21263,21264,21282,21339,21358,21395,21415,21434,21453,21491,21510,21516,21529,21567,21586,21662,21700,21738,21757,21795,21814,21852,21871,21909,21928,21947,21966,22004,22023,22042,22050,22080,22099,22110,22194,22209,22213,22270,22289,22327,22365,22375,22384,22403,22441,22460,22479,22498,22536,22555,22574,22631,22650,22688,22726,22745,22802,22859,22878,22897,22916,22969,22973,23030,23049,23068,23125,23144,23163,23199,23240,23258,23277,23315,23334,23349,23353,23372,23410,23423,23429,23505,23541,23543,23581,23600,23610,23648,23657,23676,23733,23752,23771,23809,23828,23904,23961,23999,24018,24037,24056,24094,24107,24113,24151,24170,24208,24227,24237,24246,24303,24315,24328,24331,24341,24360,24398,24436,24455,24474,24493,24512,24531,24550,24569,24588,24607,24626,24645,24664,24683,24721,24757,24759,24774,24778,24835,24854,24873,24949,24968,24987,25000,25061,25063,25082,25120,25158,25196,25221,25253,25272,25310,25348,25367,25386,25405,25462,25481,25538,25557,25595,25614,25652,25688,25709,25747,25785,25823,25842,25880,25893,25937,25956,25975,25994,26032,26051,26070,26089,26108,26127,26146,26203,26222,26260,26279,26298,26355,26374,26387,26412,26431,26450,26469,26526,26545,26554,26621,26678,26697,26735,26754,26773,26811,26925,26926,27020,27068,27134,27198,27210,27248,27267,27362,27387,27431,27552,27652,27704,27780,27799,27806,27837,27894,27977,28017,28103,28122,28181,28198,28312,28331,28616,28635,28871,29117,29241,29471,29498,29499,29604,29620,29630,29697,29711,29724,29765,29794,29908,30002,30031,30060,30089,30118,30147,30176,30205,30234,30263,30292,30321,30350,30379,30408,30437,30466,30495,30524,30553,30582,30611,30640,30669,30698,30727,30756,30785,30814,30843,30872,30901,30930,30959,30988,31017,31046,31075,31104,31133,31162,31191,31220,31249,31278,31307,31336,31365,31394,31423,31452,31481,31510,31539,31568,31597,31626,31655,31684,31713,31742,31771,31800,31829,31858,31887,31916,31945,31974,32003,32032,32061,32090,32119,32148,32177,32206,32235,32264,32293,32322,32351,32380,32409,32438,32467,32496,32525,32554,32583,32612,32641,32670,32699,32728,32757,32786,32815,32844,32873,32902,32931,32960,32989,33018,33047,33076,33105,33134,33163,33192,33221,33250,33279,33308,33337,33366,33395,33424,33453,33482,33511,33540,33569,33598,33627,33656,33685,33714,33743,33772,33801,33830,33859,33888,33917,33946,33975,34004,34033,34062,34091,34120,34149,34178,34207,34236,34265,34294,34323,34352,34381,34410,34439,34468,34497,34526,34555,34584,34613,34642,34671,34700,34729,34758,34787,34816,34845,34874,34903,34932,34961,34990,35002,35060,35063,35091,35119,35143,35160,35165,35232,35254,35310,35328,35382,35402,35445,35452,35458,35473,35521,35580,35625,35639,35680,35688,35712,35771,35824,35830,35876,35901,35960,36019,36048,36082,36127,36142,36201,36239,36272,36302,36331,36374,36390,36462,36521,36580,36633,36644,36703,36762,36791,36792,36833,36892,36951,37023,37082,37141,37205,37264,37323,37353,37394,37453,37481,37486,37512,37546,37584,37643,37702,37746,37766,37825,37884,37955,37993,38014,38056,38073,38145,38173,38179,38204,38263,38299,38327,38386,38445,38516,38575,38624,38634,38706,38765,38793,38824,38886,38888,38947,39006,39053,39077,39107,39136,39174,39195,39235,39242,39267,39326,39385,39449,39508,39567,39596,39602,39638,39668,39695,39697,39756,39828,39887,39926,39946,40117,40146,40175,40204,40233,40262,40291,40320,40349,40378,40407,40436,40465,40494,40523,40552,40581,40610,40639,40668,40697,40726,40755,40784,40813,40842,40871,40900,40929,40958,40987,41016,41045,41074,41103,41132,41161,41190,41219,41248,41277,41306,41335,41364,41393,41422,41451,41480,41509,41538,42008,42037,42066,42095,42124,42163,42184,42274,42307,42349,42365,42372,42424,42440,42449,42511,42586,42629,42661,42698,42714,42717,42729,42748,42779,42823,42898,42913,42985,43060,43135,43222,43297,43372,43390,43443,43459,43486,43504,43534,43568,43609,43636,43666,43696,43722,43771,43787,43829,43846,43859,43880,43914,43933,43967,43971,43974,43993,44008,44064,44083,44107,44109,44110,44170,44228,44245,44277,44320,44335,44372,44407,44482,44557,44615,44644,44681,44719,44753,44760,44794,44865,44881,44922,44930,44956,44977,45031,45082,45085,45104,45152,45213,45271,45293,45308,45344,45401,45459,45531,45553,45588,45618,45646,45718,45744,45775,45816,45833,45905,45962,46037,46096,46165,46168,46190,46194,46200,46211,46226,46280,46282,46286,46291,46295,46309,46424,46568,46618,46641,46644,46671,46672,46674,46719,46753,46835,46859,46883,46895,46924,46929,46949,47019,47101,47108,47114,47141,47164,47211,47233,47240,47241,47251,47273,47275,47300,47355,47360,47381,47399,47453,47474,47512,47519,47534,47544,47550,47683,47730,47766,47772,47780,47838,47859,47868,47871,47881,47916,47951,48009,48067,48092,48109,48186,48218,48223,48319,48334,48342,48392,48446,48469,48478,48479,48503,48518,48547,48554,48572,48581,48604,48612,48655,48691,48711,48725,48733,48843,48847,48855,48866,48902,48936,48998,49021,49077,49094,49100,49157,49200,49211,49251,49261,49287,49354,49372,49394,49429,49471,49541,49554,49571,49606,49615,49617,49622,49651,49680,49700,49739,49762,49783,49857,49863,49865,49880,49971,49982,49983,49989,50010,50043,50312,50405,50485,50517,50591,50811,50979,51048,51139,51183,51301,51413,51555,51568,51600,51683,51776,51812,51906,51928,52115,52340,52410,52488,52525,52577,52643,52649,52749,52891,52938,53022,53231,53340,53483,53513,53603,53712,53741,53891,53902,53952,53974,54129,54133,54201,54402,54437,54441,54458,54500,54547,54548,54621,54667,54687,54825,55052,55181,55191,55278,55306,55700,55851,55856,56109,56301,56322,56554,56566,56590,56616,56634,56656,56681,56780,56852,56988,57069,57576,57598,57600,57693,57939,57948,58015,58029,58069,58082,58193,58243,58262,58371,58378,58388,58427,58493,58598,58602,58690,58706,58723,58832,58857,58987,58991,59001,59176,59471,59478,59568,59587,59637,59644,59842,59908,59922,59942,59983,60015,60092,60120,60133,60157,60178,60363,60412,60471,60608,60679,60780,60784,60907,60977,61046,61059,61061,61168,61186,61202,61344,61434,61507,61635,61685,61798,61817,61824,61825,61828,61845,61851,61899,62177,62306,62312,62423,62453,62460,62466,62599,62600,62618,62640,62826,62979,62994,63011,63049,63144,63227,63302,63330,63373,63419,63514,63671,63782,63873,63989,64122,64275,64495,64793,64936,65053,65160,65176,65276,65299,65331,65436,65447,65514,65717,65872,66021,66450,66522,66694,66974,67039,67108,67118,67194,67272,67285,67324,67434,67461,67675,67875,67886,67933,67995,67999,68038,68207,68246,68310,68414,68465,68505,68574,68577,68582,68698,68703,69078,69096,69310,69321,69362,69405,69456,69595,69608,69696,69708,69754,69793,69891,70083,70090,70098,70199,70246,70252,70261,70424,70457,70489,70549,70605,70653,70682,70684,70686,70833,70880,70940,71040,71127,71160,71456,71485,71487,71646,71702,71727,71747,71762,71875,71888,71950,71958,71962,72173,72330,72361,72523,72602,72606,72690,72698,72891,72990,73220,73224,73287,73319,73573,73588,73770,73787,73798,73873,73881,73971,74103,74104,74285,74302,74383,74523,74541,74627,74677,74778,74834,74867,74937,74957,75063,75128,75132,75208,75269,75455,75506,75835,75957,75963,76162,76199,76393,76528,76573,76697,76730,76774,76889,77020,77081,77129,77269,77348,77376,77421,77471,77592,77611,77681,77820,77844,77906,77907,78118,78119,78130,78304,78406,78440,78510,78578,78599,78776,78993,79039,79222,79327,79346,79436,79508,79530,79562,79589,79612,79631,79641,79699,79773,79978,80058,80314,80322,80403,80476,80484,80638,80687,80691,80696,80746,80879,80905,81040,81085,81193,81203,81269,81339,81384,81717,81725,81835,82035,82184,82200,82272,82329,82394,82564,82644,82680,82858,82859,82935,82961,83034,83135,83136,83139,83159,83193,83315,83412,83656,84091,84098,84147,84162,84211,84407,84524,84639,84659,84772,84787,84824,85106,85202,85233,85320,85375,85427,85545,85562,85563,85755,85915,86065,86180,86194,86201,86207,86330,86370,86375,86450,86475,86568,86602,86670,86799,86809,86835,86839,86876,87023,87092,87168,87248,87264,87385,87537,87647,87700,87728,87730,87758,87779,87850,87868,87901,87954,87968,87978,87994,88003,88038,88093,88129,88221,88447,88486,88594,88611,88770,88796,88814,88817,88852,89045,89067,89311,89432,89608,89609,89706,89824,89917,90062,90074,90088,90242,90343,90437,90957,90999,91064,91158,91268,91419,91430,91509,91592,91735,91934,91964,92064,92068,92071,92147,92363,92754,92845,92996,93076,93189,93199,93284,93388,93491,93505,93523,93539,93572,93785,93814,94028,94095,94251,94308,94489,94519,94524,94556,94622,94674,94703,94794,94814,94841,94907,94957,94971,95054,95237,95363,95436,95488,95503,95587,95665,95684,95688,95904,95909,96002,96018,96108,96215,96316,96485,96646,96864,96906,96932,96978,97060,97204,97287,97428,97519,97731,97869,97900,97962,98058,98156,98218,98233,98254,98420,98439,98687,98709,98812,98954,98975,99091,99125,99174,99269,99295,99410,99419,99494,99699,99744,99853,99896,99908,99951",
          "2,44,119,261,299,368,370,436,447,456,502,741,745,789,837,908,982,1033,1053,1131,1177,1304,1583,1673,1750,1762,1819,1941,2004,2588,2749,2869,3021,3460,3930,4320,5172,5351,5428,5502,5552,5575,5603,5697,6133,6241,6271,6380,6446,6449,6559,6571,6585,6816,6830,6874,6889,6979,7027,7035,7180,7191,7266,7366,7470,7503,7509,7519,7540,7549,7594,7621,7851,7862,7882,7908,7917,8000,8018,8055,8172,8195,8226,8279,8282,8323,8434,8512,8515,8517,8528,8597,8612,8691,8744,8751,8772,8781,8793,8858,8876,8881,8895,8902,8929,9005,9028,9132,9249,9358,9443,9561,9747,9835,9861,9896,10053,10112,10221,10312,10314,10366,10393,10435,10539,10744,11060,11079,11166,11460,11527,11653,11731,11831,11944,11982,12056,12355,12648,12792,12841,12862,13042,13228,13265,13559,13834,13940,14396,14506,14598,14724,14954,14981,15133,15186,15650,15904,16190,16573,16594,16873,16922,17229,17412,17570,17916,17942,18280,18588,19166,19167,19321,19656,19696,20160,20713,20763,20774,20899,21368,21685,21823,21886,22025,22665,23428,23801,24089,24153,25378,25689,25782,26275,26403,26454,26932,26939,26991,27096,27501,27636,27660,27892,28178,28337,28743,28890,28902,29087,29136,29263,29281,29305,29555,29559,29573,29718,29885,29926,29964,29981,30302,34576,35754,35892,35937,36121,36157,36862,37064,37078,37123,37133,37365,37430,37597,37749,37801,37861,37927,38045,38121,38486,38507,38567,38804,38862,38918,39108,39261,39467,39505,39802,39922,39942,41500,43102,45168,45610,45684,45733,45859,47336,47384,48152,50185,50274,50281,50403,50635,50711,50881,51295,51406,51490,51509,51577,51657,51682,51804,52045,52173,52201,52216,52263,52330,52362,52457,52489,52511,52568,52724,52740,52786,52998,53020,53069,53113,53221,53240,53332,53421,53481,53488,53495,53569,53590,53820,53908,54014,54063,54109,54117,54278,54598,54607,54767,54769,54856,54938,54945,54999,55006,55009,55090,55288,55312,55325,55426,55674,55684,55767,55902,55960,56162,56186,56208,56213,56341,56343,56435,56539,56619,56728,56822,56867,57210,57236,57246,57346,57370,57377,57386,57415,57476,57523,57689,57944,58041,58173,58230,58261,58332,58377,58432,58555,58726,58750,58752,58815,58868,58913,59126,59276,59305,59427,59436,59459,59467,59760,60088,60282,60375,60419,60482,60577,60708,60740,60746,60805,60881,60909,60985,61021,61027,61035,61045,61204,61549,61732,61901,62099,62194,62201,62229,62273,62349,62507,62510,62610,62615,62881,62986,62987,63200,63488,63502,63518,63532,63681,63954,64031,64094,64200,64223,64251,64319,64336,64451,64466,64634,64784,64799,64845,64901,65011,65024,65062,65137,65141,65150,65288,65408,65409,65522,65551,65723,65898,65905,65918,65962,66059,66079,66257,66273,66314,66345,66449,66568,66675,66799,66902,67028,67041,67134,67243,67404,67512,67666,67689,67949,67980,68069,68083,68151,68179,68320,68416,68441,68449,68512,68592,68638,68726,68798,69011,69018,69036,69046,69047,69116,69228,69254,69323,69325,69377,69452,69640,69692,70047,70077,70121,70157,70216,70438,70462,70507,70587,70643,70796,70834,70854,70896,71128,71197,71229,71281,71386,71466,71584,71776,71786,72000,72013,72106,72110,72118,72237,72278,72451,72498,72768,72828,73053,73055,73248,73264,73397,73696,73712,73901,73948,73961,74070,74071,74128,74133,74175,74317,74450,74527,74666,74704,74711,74788,74852,74932,75018,75046,75220,75223,75354,75443,75492,75560,75591,75745,75814,75842,75865,75881,75902,75936,76401,76491,76550,76570,76595,76619,76817,76824,76854,76891,77009,77060,77128,77167,77180,77186,77206,77245,77322,77327,77362,77407,77451,77457,77720,77784,77788,78019,78067,78089,78182,78429,78606,78625,78659,78726,78869,78878,79083,79186,79281,79418,79420,79465,79482,79549,79595,79607,79814,79854,79863,79977,79998,80078,80149,80279,80363,80409,80429,80488,80554,80597,80664,80763,80768,80802,80894,81183,81222,81239,81270,81288,81334,81433,81484,81526,81542,81711,81714,81810,81817,81891,81983,81997,82078,82307,82318,82327,82340,82439,82486,82607,82780,83024,83304,83443,83533,83589,83638,83717,83771,83816,83836,83863,83982,84083,84434,84517,84520,84558,84670,84671,84672,84702,84783,85026,85028,85062,85123,85184,85287,85525,85624,85654,85719,85741,85770,85836,85906,85989,86030,86168,86287,86312,86365,86383,86398,86418,86669,86724,86788,86807,86848,86860,86934,86947,87028,87174,87190,87212,87273,87491,87542,87620,87664,87784,87884,88031,88086,88090,88161,88182,88195,88216,88223,88241,88410,88548,88574,88624,88645,88875,88900,88920,88956,89026,89096,89184,89328,89347,89366,89399,89473,89485,89560,89644,89811,89846,89890,90010,90162,90180,90211,90219,90301,90456,90503,90506,90597,90606,90630,90650,90735,90738,90949,91010,91072,91177,91249,91381,91512,91516,91534,91754,91800,91822,91826,92043,92171,92214,92230,92387,92400,92501,92711,92863,92914,92965,93089,93117,93137,93208,93254,93315,93366,93501,93516,93564,93743,93872,93878,93942,93980,94102,94181,94414,94459,94496,94732,94780,94793,94798,94803,94823,94908,94996,95002,95027,95187,95241,95266,95381,95403,95408,95848,95936,95968,96028,96170,96173,96254,96353,96419,96451,96643,96763,96840,97008,97168,97370,97372,97380,97417,97651,97678,97685,97791,97874,97961,97997,98045,98096,98144,98162,98309,98328,98794,99103,99117,99188,99228,99301,99310,99340,99433,99483,99534,99565,99602,99618,99660,99664,99704,99731,99790,99956,99973",
          "54,108,109,140,155,163,179,189,190,238,240,759,1788,1929,1939,1966,1967,1969,1990,2037,2109,2123,2183,2207,2212,2232,2258,2264,2284,2297,2301,2353,2360,2369,2382,2389,2410,2413,2422,2456,2461,2479,2482,2492,2496,2501,2515,2524,2535,2538,2543,2544,2565,2567,2574,2581,2586,2611,2619,2653,2697,2710,2716,2717,2723,2725,2761,2769,2771,2775,2795,2810,2814,2843,2855,2863,2889,2897,2903,2912,2919,2924,2933,2987,3070,3081,3085,3089,3099,3106,3109,3141,3153,3213,3222,3256,3304,3315,3317,3319,3344,3347,3349,3362,3373,3378,3381,3382,3409,3432,3476,3484,3515,3518,3519,3536,3543,3544,3550,3552,3554,3564,3571,3572,3593,3595,3605,3622,3642,3668,3672,3682,3693,3698,3731,3742,3750,3755,3798,3802,3810,3830,3841,3849,3870,3882,3897,3899,3911,3915,3925,3973,3974,3988,3989,4015,4026,4056,4071,4077,4097,4104,4121,4137,4141,4167,4200,4208,4220,4227,4252,4260,4263,4277,4280,4324,4344,4346,4363,4384,4415,4422,4425,4428,4436,4444,4472,4473,4475,4477,4505,4508,4516,4526,4533,4566,4594,4602,4613,4628,4631,4651,4652,4666,4680,4681,4704,4707,4709,4761,4768,4789,4791,4795,4808,4813,4816,4831,4843,4888,4897,4900,4906,4907,4927,4931,4966,4987,4990,5054,5066,5572,6111,6260,6292,6307,6337,6371,6584,6622,6632,6812,6866,6910,6914,6926,6974,6988,7057,7245,7263,7337,7381,7437,7458,7587,7637,7675,7736,7807,7819,7857,7875,7884,7922,7948,7957,8014,8026,8104,8116,8162,8217,8290,8309,8345,8504,8535,8552,8606,8609,8620,8652,8787,8854,8878,8911,9022,9159,9208,9260,9290,9328,9332,9333,9407,9461,9516,9573,9735,9745,9748,9939,10194,10196,10248,10297,10351,10407,10524,10573,10606,10643,10791,10875,10899,10982,11211,11266,11283,11623,11699,11904,11946,12002,12011,12169,12272,12363,12375,12387,12410,12483,12490,12574,12816,12820,12884,12983,13046,13108,13182,13353,13800,13810,13960,14028,14090,14518,14547,14645,14745,14933,15013,15259,15319,15370,15449,15520,15681,15757,15796,16135,16184,16271,16328,16540,16618,16621,16658,16797,16820,17033,17233,17243,17428,17476,17488,17631,17948,17993,18031,18095,18096,18177,18192,18451,18477,18590,18611,18625,18962,18986,19269,19352,19411,19828,19941,19976,20112,20513,20778,20846,20939,20957,20986,21022,21182,21185,21203,21213,21341,21355,21541,21551,21569,21601,21605,21827,21944,22451,22499,22709,22892,23072,23519,23554,23837,23858,23930,24064,24408,24745,24776,24824,24872,24882,24972,25251,25334,25452,25474,25530,25567,26028,26049,26099,26110,26126,26187,26189,26202,26220,26305,26442,26455,26528,26531,26546,26561,26566,26581,26595,26598,26600,26669,26841,26899,27001,27007,27014,27064,27080,27202,27296,27303,27367,27409,27459,27484,27487,27500,27514,27516,27558,27616,27623,27643,27655,27657,27751,27776,27815,28004,28033,28040,28056,28100,28128,28140,28152,28177,28201,28237,28239,28395,28484,28512,28542,28594,28612,28623,28630,28659,28666,28753,28885,28886,28920,28928,29008,29014,29035,29056,29071,29076,29164,29402,29409,29441,29444,29531,29581,29589,29636,29680,29688,29801,29815,29839,29855,29898,29924,29990,30019,30048,30077,30106,30135,30164,30193,30222,30251,30280,30338,30367,30396,30425,30454,30483,30512,30541,30570,30579,30628,30657,30686,30715,30744,30773,30802,30831,30860,30889,30918,30947,31034,31063,31092,31121,31150,31179,31208,31237,31266,31295,31324,31382,31411,31440,31469,31498,31527,31556,31585,31614,31643,31672,31701,31730,31759,31788,31817,31846,31875,31904,31933,31962,31991,32020,32049,32078,32107,32136,32165,32194,32223,32252,32281,32310,32339,32368,32397,32426,32455,32484,32513,32542,32571,32600,32629,32658,32687,32716,32745,32774,32803,32832,32861,32890,32919,32948,32977,33006,33035,33064,33093,33122,33151,33180,33209,33238,33267,33296,33325,33383,33412,33441,33470,33528,33557,33586,33615,33644,33673,33702,33731,33760,33789,33818,33847,33876,33905,33934,33963,33992,34079,34108,34137,34166,34195,34224,34253,34282,34311,34340,34369,34398,34427,34456,34485,34514,34543,34572,34601,34630,34659,34688,34717,34746,34775,34804,34833,34862,34891,34920,34949,34978,35019,35037,35077,35086,35108,35125,35182,35200,35271,35289,35293,35308,35318,35345,35370,35377,35419,35470,35490,35538,35567,35597,35656,35691,35729,35756,35847,35862,35874,35918,35977,35990,36036,36067,36099,36159,36218,36232,36289,36306,36348,36407,36460,36479,36538,36597,36661,36720,36779,36795,36796,36807,36850,36909,36924,36968,37040,37099,37158,37174,37222,37233,37281,37340,37470,37529,37545,37601,37629,37660,37719,37734,37783,37809,37842,37865,37901,37934,37972,38031,38044,38090,38104,38162,38221,38280,38305,38307,38344,38365,38381,38403,38462,38484,38533,38547,38573,38592,38613,38666,38678,38723,38782,38841,38856,38869,38905,38919,38923,38964,38995,39023,39048,39094,39109,39123,39153,39212,39224,39230,39232,39284,39296,39343,39402,39466,39491,39525,39539,39584,39598,39603,39655,39675,39714,39773,39845,39904,39915,39963,40109,40138,40167,40196,40225,40254,40283,40312,40341,40370,40399,40428,40457,40486,40544,40573,40602,40631,40660,40689,40718,40747,40776,40805,40834,40863,40892,40921,40950,40979,41008,41037,41066,41095,41124,41153,41182,41211,41240,41269,41298,41327,41356,41385,41414,41443,41472,41501,41530,41550,41551,41552,41553,41554,41555,41556,41557,41558,41559,41560,41561,41562,41563,41564,41565,41566,41567,41568,41569,41570,41571,41572,41573,41574,41576,41577,41578,41579,41580,41581,41582,41583,41584,41585,41586,41587,41588,41589,41590,41591,41592,41593,41594,41595,41596,41597,41598,41599,41600,41601,41602,41603,41604,41605,41607,41608,41609,41610,41611,41612,41613,41614,41615,41616,41617,41618,41619,41620,41621,41622,41623,41624,41625,41626,41627,41628,41629,41630,41631,41632,41633,41634,41635,41636,41637,41638,41639,41640,41641,41642,41643,41644,41645,41646,41647,41648,41649,41650,41651,41652,41653,41654,41655,41656,41657,41658,41659,41660,41661,41662,41663,41664,41665,41666,41668,41669,41670,41671,41672,41673,41674,41675,41676,41677,41678,41679,41680,41681,41682,41683,41684,41685,41686,41687,41688,41689,41690,41691,41692,41693,41694,41695,41696,41697,41698,41699,41701,41702,41703,41704,41705,41706,41707,41708,41709,41710,41711,41712,41713,41714,41715,41716,41717,41718,41719,41720,41722,41723,41724,41725,41726,41727,41729,41730,41731,41732,41733,41734,41735,41736,41737,41738,41739,41741,41742,41743,41744,41745,41746,41747,41748,41749,41750,41751,41752,41753,41754,41755,41757,41758,41759,41760,41761,41762,41763,41764,41765,41767,41768,41769,41770,41771,41773,41774,41775,41776,41777,41778,41779,41780,41781,41782,41783,41784,41785,41786,41787,41788,41790,41792,41794,41796,41798,41800,41804,41806,41808,41810,41812,41814,41816,41818,41820,41822,41824,41826,41828,41830,41832,41834,41836,41838,41840,41842,41844,41848,41850,41852,41854,41856,41858,41860,41862,41864,41866,41868,41870,41872,41874,41876,41878,41880,41882,41884,41886,41888,41890,41892,41894,41896,41898,41900,41902,41904,41906,41908,41910,41912,41914,41916,41918,41920,41922,41924,41926,41928,41931,41934,41937,41940,41943,41949,41952,41955,41959,41963,41967,41971,41975,41979,41983,41987,41992,42021,42050,42079,42108,42137,42204,42212,42222,42302,42370,42377,42452,42471,42473,42478,42539,42544,42560,42602,42609,42614,42686,42689,42724,42731,42776,42851,42926,43013,43025,43045,43088,43163,43202,43242,43246,43250,43269,43273,43276,43325,43327,43333,43339,43400,43423,43442,43487,43513,43562,43572,43577,43585,43590,43637,43640,43665,43724,43725,43731,43747,43799,43801,43823,43874,43903,43910,43961,43976,44031,44036,44043,44111,44120,44188,44198,44271,44273,44348,44433,44434,44435,44443,44510,44531,44539,44590,44594,44672,44741,44747,44822,44836,44909,44984,45012,45049,45062,45083,45132,45179,45184,45192,45196,45251,45287,45322,45359,45439,45481,45509,45568,45571,45619,45626,45696,45755,45793,45796,45813,45883,45932,45942,46000,46002,46025,46063,46132,46178,46257,46303,46314,46343,46388,46434,46447,46451,46465,46528,46581,46586,46640,46703,46822,46882,46888,46891,46941,46946,46953,46971,47005,47015,47018,47096,47113,47184,47199,47213,47267,47283,47285,47368,47371,47400,47435,47445,47597,47607,47633,47639,47721,47726,47767,47771,47789,47796,47862,47879,48016,48054,48075,48105,48159,48184,48193,48196,48254,48294,48327,48340,48350,48374,48394,48493,48506,48520,48531,48579,48586,48600,48607,48761,48772,48788,48792,48857,48877,48888,48906,48946,48951,48983,49017,49046,49049,49058,49072,49078,49086,49088,49121,49130,49134,49144,49156,49240,49242,49271,49299,49313,49326,49365,49380,49415,49436,49440,49450,49480,49521,49538,49560,49656,49698,49716,49725,49740,49745,49796,49861,49875,49896,49921,49963,50035,50095,50096,50133,50285,50377,50419,50420,50657,50772,50886,50887,51129,51162,51210,51216,51425,51461,51485,51569,51625,51687,51720,51840,52029,52164,52184,52309,52312,52318,52396,52414,52452,52493,52665,52669,52715,52733,52767,52795,52824,52828,52867,53015,53140,53261,53358,53366,53480,53588,53608,53687,53770,53840,53868,53896,54333,54336,54386,54444,54729,54983,54986,54995,55077,55078,55099,55238,55257,55347,55382,55389,55420,55491,55565,55581,55612,55705,55859,55932,55934,55987,56107,56201,56259,56266,56270,56299,56362,56414,56483,56820,56910,56992,57341,57371,57396,57531,57715,57824,57865,57921,57925,57995,58087,58103,58134,58202,58209,58211,58284,58356,58358,58508,58591,58683,58687,58858,58869,58923,58966,59100,59150,59159,59175,59297,59359,59432,59491,59522,59556,59569,59622,59659,59667,59674,59677,59718,59944,59988,60026,60043,60115,60326,60368,60393,60455,60524,60683,60872,60894,60952,61065,61234,61237,61280,61625,61643,61753,61777,61862,61918,61998,62065,62082,62118,62199,62221,62416,62484,62551,62619,62678,62733,62780,62805,62945,62966,63069,63222,63245,63379,63538,63586,63677,63896,63930,64069,64103,64110,64243,64250,64611,64656,64659,64707,64900,64953,65020,65107,65177,65274,65393,65501,65505,65692,65761,65875,65952,66001,66028,66056,66076,66136,66225,66250,66362,66375,66388,66424,66438,66534,66623,66708,66839,66952,66993,67150,67174,67220,67233,67331,67381,67505,67531,67632,67646,67704,67844,67864,67894,68073,68149,68164,68170,68175,68251,68379,68396,68508,68545,68595,68742,68746,68848,69088,69109,69312,69448,69497,69520,69612,69656,69802,70020,70041,70065,70128,70171,70184,70203,70218,70241,70284,70286,70317,70410,70411,70554,70570,70580,70780,70923,70947,71089,71121,71174,71185,71188,71325,71370,71374,71384,71392,71518,71638,71655,71677,71796,71947,71969,72132,72155,72172,72199,72225,72236,72358,72423,72539,72554,72591,72864,72983,73099,73134,73202,73232,73360,73396,73587,73992,73993,74077,74125,74179,74365,74437,74438,74535,74648,74904,75040,75114,75323,75349,75402,75412,75475,75596,75684,75746,75761,75832,76039,76063,76192,76226,76307,76313,76319,76376,76728,76916,77103,77172,77260,77304,77391,77593,77727,77758,77947,77966,77974,78023,78059,78071,78103,78116,78171,78237,78311,78319,78326,78371,78500,78571,78837,78893,78913,78961,79068,79183,79197,79227,79235,79298,79393,79584,79825,80001,80059,80089,80195,80218,80245,80384,80391,80412,80540,80550,80619,80725,80729,80770,80864,80927,80958,81094,81107,81173,81213,81454,81654,81691,81757,81787,81801,81827,81979,82122,82207,82211,82346,82352,82456,82535,82708,82860,82934,83015,83043,83047,83116,83125,83205,83376,83416,83483,83499,83575,83674,83690,83716,83721,83775,83837,83889,83954,84045,84070,84125,84171,84185,84228,84658,84697,84782,84961,84986,85085,85162,85178,85197,85224,85271,85283,85344,85352,85536,85676,85976,86125,86151,86392,86445,86470,86622,86692,86702,86703,86722,86763,86994,87079,87087,87210,87235,87293,87298,87531,87600,87781,87858,88163,88363,88414,88426,88451,88509,88610,88891,88946,88977,89032,89133,89182,89217,89315,89380,89401,89422,89472,89593,89597,89720,89784,89906,90065,90299,90318,90418,90419,90746,90943,90977,90997,91058,91076,91090,91163,91194,91254,91288,91436,91472,91478,91530,91580,91585,91644,91655,91726,91792,91806,91891,91974,92078,92408,92457,92467,92715,92769,92888,92892,92935,92997,92999,93011,93067,93136,93191,93195,93240,93327,93359,93543,93587,93627,93661,93754,93867,93967,94039,94065,94068,94087,94121,94274,94320,94337,94369,94417,94526,94530,94548,94561,94582,94669,95004,95033,95069,95084,95207,95324,95467,95564,95671,95690,96113,96114,96209,96279,96444,96461,96508,96579,96649,96711,96717,96732,96753,96890,96928,97028,97095,97125,97153,97201,97202,97252,97265,97539,97563,97782,97819,97922,97965,97974,98002,98003,98170,98259,98295,98368,98442,98451,98628,98737,99053,99062,99343,99344,99364,99502,99524,99584,99941",
          "305,345,346,357,418,478,497,503,599,631,669,673,854,858,892,1014,1065,1084,1138,1210,1307,1456,1499,1622,1634,1638,1668,1680,1729,1742,1744,2080,2322,3146,3891,3987,4140,4624,5042,5062,5191,5284,5367,5556,5669,5733,5838,5907,6010,6068,6086,6107,6142,6235,6254,6361,6523,6549,6556,6570,6709,6738,6843,6844,7084,7100,7104,7113,7142,7227,7231,7252,7285,7304,7447,7513,7543,7659,7704,7721,7769,8069,8176,8196,8275,8291,8366,8371,8428,8525,8560,8592,8614,8757,8794,8811,8915,9010,9157,9167,9216,9232,9243,9291,9331,9352,9425,9452,9494,9511,9518,9574,9789,9790,9804,9922,9927,10146,10310,10359,10365,10513,10517,10531,10667,10898,11075,11276,11446,11492,11529,11531,11570,11590,11679,11691,11703,11752,11798,11800,11846,11989,12093,12452,12760,12808,12909,13212,13335,13651,13713,14086,14423,14485,14487,14579,14700,15124,15374,15445,15454,15492,15528,15583,15641,15663,15747,15778,15922,16446,16590,16840,17139,17236,17439,17493,17594,17758,17777,18059,18152,18262,18622,18891,18938,19336,19595,19841,20209,20738,20897,20929,21066,21081,22098,22173,22658,22815,23042,23107,23293,23769,24092,24137,24307,24370,25101,25126,25246,25665,25703,25958,26259,26271,26389,26409,26434,26482,26648,26667,26809,27067,27093,27100,27233,27245,27269,27747,27947,28727,28784,28829,28843,28936,29106,29467,29897,32093,35498,35704,35934,36133,36180,36517,36618,37061,37110,37390,37482,37495,37580,37681,37804,38184,38259,38300,38377,38697,39165,39188,39410,39619,39737,39801,39868,40443,43978,45089,45324,45507,49285,50179,50293,50352,50506,50655,50668,50743,50801,50973,50983,51175,51463,51641,51743,51871,51890,51949,52165,52259,52435,52528,52538,52557,52602,52668,52779,52817,52949,52981,53168,53193,53236,53361,53532,53818,53885,53916,54176,54207,54270,54384,54497,54522,54662,54873,54888,54899,54977,55082,55162,55308,55314,55319,55353,55457,55488,56014,56143,56167,56168,56171,56175,56221,56438,56440,56441,56500,56530,56605,56824,56878,56892,56922,56973,57122,57154,57279,57301,57556,57641,57712,57714,57722,57746,57870,57913,58004,58163,58167,58227,58259,58276,58306,58363,58381,58411,58464,58573,58838,58863,58865,59166,59253,59256,59529,59530,59558,59658,59666,59694,59695,59870,60000,60231,60449,60677,60739,60763,60821,60829,60870,60900,60935,61075,61163,61465,61510,61577,61612,61653,61670,61673,61709,61714,61757,61864,61873,61936,62013,62051,62083,62192,62333,62376,62534,62553,62589,62593,62630,62783,62854,62955,63459,63512,63529,63564,63620,63666,63905,63908,63917,64029,64150,64262,64295,64296,64312,64372,64412,64424,64643,64651,64711,64719,64763,64896,64996,65104,65135,65181,65221,65344,65380,65394,65606,65668,65736,65737,65855,65919,65994,66308,66663,66739,66855,67064,67452,67527,67535,67552,67807,67891,68013,68055,68064,68123,68288,68322,68544,68557,68585,68679,68792,69072,69149,69259,69499,69501,69592,69638,69665,69726,69935,69988,70003,70016,70074,70288,70380,70381,70512,70683,70697,70777,70788,70817,70925,70959,71008,71232,71336,71503,71506,71550,71616,71647,71886,72101,72189,72200,72445,72492,72558,72888,73082,73108,73146,73233,73306,73488,73658,73894,73989,74158,74201,74335,74355,74401,74412,74584,74837,74886,74974,75101,75109,75129,75314,75316,75388,75396,75418,75511,75552,75594,75603,75712,75720,75722,75725,75799,75803,75985,76178,76441,76533,76897,76999,77196,77320,77655,77746,78072,78137,78144,78348,78457,78479,78521,79029,79057,79102,79149,79171,79342,79416,79661,79739,79812,79824,79826,79899,80070,80080,80082,80170,80235,80285,80393,80397,80407,80483,80563,80608,80612,80671,80709,80860,81075,81151,81410,81525,81567,81805,82024,82067,82180,82182,82188,82308,82460,82549,82966,83027,83074,83209,83211,83311,83317,83341,83350,83538,83542,83555,83563,83669,83736,83767,83780,83814,83825,83993,84068,84336,84691,84737,84788,84911,84966,85001,85240,85321,85410,85569,85601,85615,85621,85718,85768,85806,85849,85892,85953,85975,86013,86142,86219,86366,86841,86945,87089,87115,87229,87303,87458,87759,87828,87834,87920,88156,88167,88230,88274,88281,88467,88522,88786,88884,88907,89052,89101,89114,89209,89264,89277,89458,89534,89730,89857,89945,89946,90042,90228,90240,90298,90352,90475,90499,90582,90590,90646,90702,90730,90794,90881,90939,90985,91011,91023,91097,91165,91184,91213,91317,91352,91428,91782,91794,91867,92168,92335,92482,92517,92624,92634,92653,92682,92704,92741,92761,92765,92908,92969,93034,93053,93090,93220,93286,93294,93427,93460,93508,93781,93844,93921,94080,94137,94488,94610,94741,94846,94913,94935,94985,95024,95200,95205,95238,95316,95393,95500,95537,95539,95742,95781,96058,96175,96243,96320,96472,96565,96627,96807,96953,97052,97082,97088,97100,97172,97182,97233,97295,97592,97660,97702,97856,97866,97923,97978,98041,98048,98110,98185,98222,98232,98281,98456,98557,98661,98683,98685,98809,98824,98852,98896,98905,98959,99012,99046,99360,99436,99456,99563,99580,99615,99776,99777,99864,99887,99980",
          "77,104,142,151,161,211,225,1662,1681,1683,1700,1738,1741,1757,1776,1782,1795,1798,2074,2082,2091,2487,2634,2654,2712,2782,2830,2886,3253,3406,3753,4018,4132,4148,4235,4257,4476,4644,4696,4767,6166,7208,8225,8806,9376,9395,9481,9523,9547,9604,9661,9680,9699,9718,9740,9775,9791,9832,9984,10022,10060,10649,10703,10749,10858,10877,10953,11002,11029,11119,11200,11238,11257,11352,11447,11485,11580,11607,11618,11637,11694,11903,11960,11998,12055,12061,12112,12131,12226,12264,12283,12299,12302,12333,12340,12353,12397,12436,12442,12473,12519,12530,12568,12592,12663,12709,12766,12777,12801,12830,12835,12853,12872,12920,12978,12986,13005,13015,13044,13081,13082,13096,13103,13158,13271,13290,13309,13480,13485,13491,13632,13638,13670,13693,13708,13746,13765,13784,13803,13822,13898,13909,13917,13956,13993,14069,14091,14098,14107,14183,14259,14278,14297,14316,14335,14354,14411,14563,14601,14615,14636,14639,14760,14772,14829,14857,14867,14883,14905,14962,15000,15038,15076,15114,15171,15182,15190,15213,15228,15247,15285,15304,15323,15361,15380,15456,15513,15574,15589,15608,15630,15637,15684,15722,15741,15817,15836,15855,15874,15893,15926,15931,15950,15969,16026,16045,16102,16121,16178,16216,16235,16254,16292,16311,16330,16349,16368,16463,16482,16485,16501,16558,16577,16634,16653,16669,16672,16691,16710,16805,16824,16881,16919,16957,16995,17014,17049,17052,17109,17111,17128,17147,17204,17280,17299,17375,17394,17413,17451,17477,17527,17546,17584,17603,17616,17629,17660,17679,17701,17722,17755,17764,17774,17788,17831,17850,17869,17907,17912,17964,17998,18002,18021,18040,18047,18065,18097,18135,18154,18166,18211,18230,18249,18306,18325,18363,18382,18401,18420,18458,18486,18496,18534,18553,18610,18629,18686,18705,18743,18762,18800,18819,18838,18857,18876,18952,18990,19019,19047,19066,19104,19106,19142,19199,19218,19275,19294,19313,19332,19408,19446,19465,19473,19541,19579,19598,19617,19636,19688,19693,19731,19769,19788,19826,19845,19883,19902,19921,19940,19959,20016,20028,20035,20072,20073,20092,20146,20149,20187,20206,20214,20225,20263,20339,20358,20377,20396,20415,20434,20472,20510,20565,20586,20605,20643,20662,20681,20776,20814,20833,20845,20852,20890,20926,20966,20985,21000,21004,21009,21023,21042,21055,21061,21099,21128,21135,21175,21186,21232,21251,21308,21346,21365,21377,21403,21441,21460,21467,21498,21515,21536,21555,21574,21612,21631,21632,21701,21712,21726,21764,21765,21783,21802,21821,21859,21878,21897,21935,21954,21973,22011,22030,22049,22087,22101,22106,22144,22163,22181,22220,22239,22277,22296,22334,22353,22429,22448,22505,22524,22562,22581,22600,22676,22690,22733,22752,22771,22828,22866,22885,22904,22942,22961,22966,23018,23037,23056,23065,23075,23091,23094,23113,23132,23227,23265,23284,23322,23338,23341,23360,23398,23417,23474,23507,23512,23550,23569,23588,23645,23664,23683,23740,23744,23759,23775,23797,23816,23873,23892,23911,23949,23968,23987,24006,24025,24044,24063,24074,24082,24139,24158,24196,24215,24272,24299,24310,24329,24367,24424,24443,24459,24462,24519,24538,24576,24595,24614,24652,24728,24747,24785,24804,24823,24842,24899,24918,24931,24937,24956,24975,24994,25079,25108,25112,25119,25146,25165,25222,25241,25259,25260,25298,25317,25374,25379,25431,25437,25450,25488,25507,25561,25564,25602,25621,25640,25697,25735,25754,25773,25787,25830,25849,25868,25887,25906,25925,25930,25963,25982,26001,26039,26058,26096,26115,26191,26229,26245,26286,26324,26343,26400,26457,26495,26514,26552,26634,26647,26666,26704,26742,26793,26856,26913,26952,26968,26970,27008,27027,27217,27293,27407,27419,27504,27555,27578,27635,27703,27730,27748,27749,27768,27788,27901,27918,27932,27965,27996,28028,28087,28091,28110,28186,28205,28257,28332,28357,28369,28400,28471,28692,28775,28813,28828,28832,28834,28965,29003,29041,29060,29064,29098,29139,29193,29212,29250,29261,29300,29326,29364,29383,29411,29459,29478,29497,29668,29687,29702,29782,29814,29891,29916,30009,30038,30067,30096,30125,30154,30183,30212,30241,30299,30328,30357,30386,30415,30444,30473,30502,30531,30560,30589,30618,30647,30676,30705,30734,30763,30792,30821,30850,30879,30908,30937,30966,30995,31024,31053,31082,31111,31140,31169,31198,31227,31256,31285,31314,31343,31372,31401,31430,31459,31488,31517,31546,31575,31604,31633,31662,31691,31720,31749,31778,31807,31836,31865,31894,31923,31952,31981,32010,32039,32068,32097,32126,32155,32184,32213,32242,32271,32300,32329,32358,32387,32416,32445,32503,32532,32561,32590,32619,32648,32677,32706,32735,32764,32793,32822,32851,32880,32909,32938,32967,32996,33025,33054,33083,33112,33141,33170,33199,33228,33257,33286,33315,33344,33373,33402,33431,33460,33489,33518,33547,33576,33605,33634,33663,33692,33721,33750,33779,33808,33837,33866,33895,33924,33953,33982,34011,34040,34069,34098,34127,34156,34185,34214,34243,34272,34301,34330,34359,34388,34417,34475,34504,34533,34562,34591,34620,34649,34678,34707,34736,34765,34794,34823,34852,34881,34910,34939,34968,34997,35009,35054,35098,35136,35172,35261,35314,35357,35409,35447,35455,35456,35480,35528,35587,35646,35670,35683,35719,35751,35753,35765,35778,35811,35816,35837,35908,35946,35967,36026,36089,36149,36174,36208,36279,36303,36307,36338,36397,36437,36528,36564,36587,36642,36651,36710,36769,36813,36840,36871,36882,36899,36935,36958,37030,37089,37139,37148,37212,37271,37313,37330,37367,37401,37460,37519,37591,37650,37673,37709,37730,37757,37764,37773,37832,37866,37882,37891,37962,37985,38021,38080,38115,38152,38211,38270,38334,38384,38393,38452,38473,38523,38582,38641,38713,38772,38831,38895,38954,38983,39013,39050,39084,39143,39202,39241,39243,39274,39333,39356,39392,39414,39415,39456,39493,39515,39554,39574,39645,39669,39670,39688,39704,39763,39835,39894,39953,40104,40133,40162,40191,40220,40249,40278,40307,40336,40365,40394,40423,40452,40481,40510,40539,40568,40597,40626,40655,40713,40742,40771,40800,40829,40858,40887,40916,40945,40974,41003,41032,41061,41090,41119,41148,41177,41206,41235,41264,41293,41351,41380,41409,41438,41467,41496,41525,41929,41932,41935,41938,41941,41944,41947,41950,41953,41960,41964,41968,41972,41976,41980,41984,41989,41994,42023,42052,42081,42110,42162,42170,42215,42233,42245,42294,42320,42329,42400,42410,42484,42557,42625,42647,42677,42678,42713,42721,42794,42876,42877,42884,42929,42950,42958,43002,43017,43031,43121,43195,43197,43247,43268,43312,43328,43344,43351,43353,43358,43432,43483,43496,43505,43583,43595,43669,43680,43723,43736,43742,43746,43788,43791,43793,43811,43832,43868,43906,43913,43920,43979,44047,44054,44061,44069,44101,44124,44139,44143,44216,44306,44380,44387,44428,44447,44453,44514,44532,44537,44543,44617,44657,44690,44695,44764,44780,44854,44862,44903,44913,44917,44919,44926,44927,44973,44988,45009,45011,45017,45057,45077,45090,45129,45138,45199,45257,45283,45330,45373,45387,45445,45477,45483,45491,45517,45546,45574,45632,45704,45754,45761,45819,45846,45891,45948,46005,46033,46064,46097,46108,46155,46159,46298,46305,46349,46377,46378,46389,46415,46420,46634,46636,46642,46666,46679,46744,46773,46795,46842,46846,46898,46903,46919,46974,46975,46998,47040,47124,47160,47183,47198,47202,47252,47291,47342,47357,47363,47403,47417,47430,47446,47452,47491,47514,47547,47556,47675,47714,47717,47720,47725,47735,47814,47824,47839,47884,47893,47906,48146,48153,48203,48351,48355,48426,48434,48436,48580,48626,48631,48689,48690,48697,48714,48764,48766,48781,48795,48818,48829,48862,48870,48908,48926,48988,49027,49039,49044,49056,49065,49066,49150,49176,49180,49181,49213,49241,49305,49353,49363,49371,49377,49422,49439,49467,49499,49542,49598,49623,49628,49643,49702,49784,49801,49827,49835,49838,49917,50001,50085,50248,50254,50370,50390,50407,50466,50554,50617,50756,50823,50942,51203,51205,51270,51314,51316,51323,51378,51538,51585,51598,51617,51872,52036,52212,52213,52280,52383,52562,52581,52619,52652,52653,52924,53103,53109,53456,53563,53567,53613,53634,53692,53785,53846,53853,53934,53953,53970,53983,54028,54074,54144,54161,54204,54322,54392,54504,54532,54570,54579,54590,54646,54686,54750,54783,54822,54970,55072,55111,55279,55407,55455,55475,55537,55679,55739,55803,55817,55980,56139,56229,56373,56408,56434,56442,56464,56536,56651,56668,56672,56706,56711,56744,56841,57269,57278,57398,57410,57413,57432,57595,57702,57706,57711,57899,57964,58253,58490,58504,58614,58714,58749,58856,58981,59067,59133,59208,59214,59228,59327,59420,59562,59583,59646,59761,59897,59975,60119,60192,60323,60391,60490,60521,60633,60713,60718,60842,60865,60916,60936,60970,61000,61073,61189,61190,61225,61233,61261,61309,61354,61395,61414,61468,61639,61650,61710,61751,61877,61880,61924,61966,62092,62136,62154,62156,62269,62537,62897,62929,62992,63165,63192,63260,63525,63719,63773,63793,63829,63882,63911,64124,64154,64180,64207,64211,64293,64333,64449,64452,64486,64587,64589,64599,64759,64886,64911,64944,65055,65059,65326,65441,65461,65484,65503,65610,65641,65657,65843,65885,65912,65935,65938,66060,66227,66291,66514,66529,66556,66657,66882,66949,66975,67044,67051,67095,67104,67112,67152,67290,67389,67463,67500,67641,67660,67701,67728,67975,67979,67997,68063,68101,68152,68192,68324,68349,68541,68614,68752,68796,68971,69008,69016,69063,69156,69287,69329,69336,69387,69394,69435,69442,69591,69593,69889,69898,69943,70143,70177,70182,70398,70541,70559,70582,70637,70652,70659,70687,70716,70728,70926,71057,71096,71152,71194,71212,71324,71377,71554,71573,71724,71756,71779,71823,71832,71961,72122,72220,72410,72480,72530,72557,72780,72832,72852,72996,73079,73085,73123,73171,73431,73492,73562,73667,73690,73738,73828,73893,73928,74198,74214,74361,74596,74610,74631,74634,74724,74825,75036,75077,75121,75259,75480,75490,75703,75706,75804,75925,76024,76065,76066,76157,76236,76362,76394,76620,76762,76827,76837,76846,76864,76979,77027,77074,77105,77109,77211,77235,77323,77351,77440,77458,77818,77985,78156,78342,78365,78390,78436,78442,78514,78529,78713,78970,79073,79093,79148,79208,79213,79220,79245,79252,79262,79283,79302,79512,79520,79526,79536,79694,79750,79870,79883,79896,79924,80015,80027,80069,80083,80127,80219,80379,80428,80503,80626,80648,80749,80781,80875,80976,81016,81072,81129,81322,81370,81414,81485,81486,81509,81535,81555,81729,81754,81792,81838,81950,82022,82153,82216,82276,82316,82326,82335,82339,82410,82481,82530,82678,82688,82745,82800,83077,83126,83196,83218,83238,83298,83329,83340,83371,83435,83532,83704,83760,83791,83997,84094,84163,84213,84278,84307,84440,84569,84688,84803,84818,84887,84968,85023,85107,85250,85268,85289,85405,85413,85457,85703,85918,85979,86116,86182,86231,86267,86279,86419,86476,86497,86527,86585,86626,86753,86765,86766,86956,86982,87041,87162,87197,87322,87365,87398,87401,87416,87649,87674,87771,87890,87907,87914,88041,88050,88089,88126,88322,88419,88445,88463,88481,88517,88621,88824,88846,88999,89010,89017,89116,89244,89355,89361,89373,89516,89624,89705,89858,89864,90130,90347,90361,90421,90462,90480,90486,90529,90568,90681,90687,90772,90814,91277,91393,91573,91633,91666,91734,92097,92173,92191,92249,92506,92523,92537,92584,92775,92851,92876,92877,93002,93043,93323,93334,93376,93390,93441,93509,93557,93641,93749,93977,94094,94133,94141,94350,94466,94601,94627,95113,95305,95358,95361,95409,95414,95509,96019,96176,96249,96321,96512,96826,96918,97023,97024,97031,97043,97087,97229,97285,97445,97456,97500,97597,97627,97670,97951,98099,98187,98228,98345,98384,98534,98633,98690,99232,99307,99413,99428,99527,99538,99653,99655,99761,99877,99910",
          "80,172,332,378,395,404,516,543,731,927,1020,1030,1076,1180,1196,1246,1261,1397,1440,1523,1628,1699,1704,1770,1823,2089,2168,2256,2309,2784,2838,3013,3028,3175,3273,3341,3827,3878,3896,3907,4100,4386,4574,4774,4818,4902,5003,5011,5246,5285,5300,5308,5349,5385,5387,5412,5422,5443,5694,5765,5859,5904,5962,5971,6118,6267,6318,6351,6417,6540,6544,6719,6793,6835,6896,6907,6961,6983,7106,7132,7151,7194,7201,7329,7506,7536,7537,7555,7565,7622,7671,7746,7764,7801,7928,8088,8241,8448,8491,8607,8957,9177,9251,9272,9323,9349,9437,9542,9592,9610,9792,9826,9844,9893,9906,10197,10246,10273,10275,10485,10626,10822,10854,11001,11184,11196,11204,11341,11472,11475,11488,11520,11723,11742,11908,12220,12231,12293,12332,12420,12458,12487,12509,12518,12562,12605,12644,12753,12788,12837,12842,12890,12917,13002,13122,13191,13286,13318,13334,13460,13517,13727,13794,13851,14003,14074,14141,14234,14288,14362,14539,14653,14695,14765,14799,14941,15153,15291,15490,15518,15542,15548,15607,15729,15840,15940,15955,15968,15985,16012,16036,16129,16230,16394,16828,16832,16880,16889,16897,16904,16966,16980,17008,17023,17199,17256,17351,17423,17452,17454,17470,17513,17917,17938,18084,18109,18264,18304,18461,18644,18718,18984,18995,19095,19241,19255,19258,19324,19431,19760,19816,19818,19846,19986,20053,20120,20257,21200,21228,21391,21642,21645,21733,21926,21987,22028,22071,22148,22313,22325,22475,22552,22751,22758,22860,22895,22914,23038,23168,23186,23212,23221,23241,23345,23379,23506,23592,23991,24000,24059,24119,24143,24162,24230,24294,24381,24467,24491,24567,24568,24668,24709,24889,24980,25154,25162,25387,25514,25624,25802,25837,26022,26152,26231,26335,26524,26897,26936,26950,26956,26994,26995,27016,27155,27189,27305,27356,27490,27686,27739,27832,27904,27953,28163,28215,28411,28492,28533,28736,29102,29127,29186,29206,29215,29283,29293,29396,29465,29490,29821,29890,29902,29918,29976,30052,30487,31653,31763,32175,32898,33048,33077,33364,33744,33942,34518,35149,35150,35238,35276,35734,35887,36015,36068,36077,36222,36294,36451,36455,36456,36484,36511,36546,36553,36554,36699,36747,36821,36854,36873,36942,37135,37197,37547,37567,37569,37678,37992,38098,38123,38127,38199,38239,38417,38512,38565,38607,38845,38851,38942,39007,39306,39319,39379,39380,39485,39486,39490,39614,39722,39783,39888,39908,40114,40129,40812,41334,42319,42494,42565,43034,43115,43119,43198,43221,43476,43522,44144,44458,45428,45438,45502,45691,45875,46198,47326,47608,48151,48198,48364,48765,49040,49428,49592,49639,50054,50068,50091,50094,50184,50296,50327,50432,50515,50616,50623,50630,50639,50640,50652,50690,50713,50720,50859,51068,51196,51224,51285,51362,51417,51484,51505,51546,51615,51628,51692,51885,51899,51908,52088,52178,52284,52401,52564,52658,52694,52701,52836,52896,53042,53100,53145,53419,53505,53523,53670,53719,53764,53786,53813,53940,54042,54181,54354,54369,54378,54401,54432,54502,54578,54643,54645,54788,54824,54887,54933,54982,55110,55118,55241,55391,55483,55561,56098,56370,56467,56592,56628,56720,56898,56940,56947,57026,57049,57075,57171,57174,57181,57247,57273,57305,57455,57471,57477,57488,57544,57588,57657,57771,57791,57918,57920,58159,58252,58593,58702,58770,58976,59025,59211,59450,59515,59574,59651,59692,59778,59952,60059,60060,60216,60262,60329,60346,60465,60662,60669,60752,60781,60810,60819,60888,61009,61067,61083,61109,61151,61158,61404,61622,61807,61849,61869,62240,62260,62266,62323,62439,62506,62508,62637,62737,62817,62863,63132,63175,63196,63203,63256,63320,63333,63383,63404,63420,63562,63645,63670,63676,63806,63866,63876,63891,63945,63949,63964,64099,64301,64315,64398,64419,64429,64442,64527,64608,64623,64664,64686,64720,64817,64927,65047,65092,65152,65182,65235,65329,65392,65402,65439,65481,65573,65711,65846,65907,65973,66323,66330,66447,66461,66503,66591,66782,66802,66838,66857,66904,67165,67209,67214,67424,67437,67438,67444,67653,67719,67735,67771,67904,68041,68096,68276,68407,68462,68567,68588,68692,68927,68948,69005,69107,69404,69408,69492,69508,69518,69600,69645,69698,69778,69828,69984,70195,70298,70342,70356,70357,70407,70681,70689,70700,70762,71032,71186,71267,71571,71765,71792,71801,71862,71949,72032,72069,72269,72459,72464,72503,72608,72623,72658,72790,72811,72827,72991,73067,73074,73091,73238,73383,73583,73824,73836,74063,74117,74161,74262,74378,74405,74862,75164,75210,75251,75261,75324,75400,75503,75545,75616,75633,75719,75739,75767,75921,76037,76154,76202,76317,76334,76352,76617,76699,76816,77079,77092,77178,77357,77411,77535,77735,77793,77890,77938,78035,78215,78257,78308,78366,78413,78427,78474,78516,78615,78771,78797,79001,79153,79206,79243,79299,79385,79516,79609,79708,79714,79742,79873,79909,79911,79949,80120,80123,80258,80304,80376,80538,80546,80631,80842,81172,81313,81323,81373,81387,81665,81718,81753,81878,82018,82043,82075,82412,82437,82490,82572,82630,82682,82722,82728,82747,82811,82820,82821,82832,82888,82898,82942,82988,83062,83436,83556,83616,83664,83689,83720,83761,83800,83869,83883,83896,83897,84065,84120,84173,84205,84512,84541,84717,84858,84862,84895,85070,85113,85128,85168,85208,85229,85267,85292,85337,85511,85519,85541,85709,85866,85881,86010,86122,86171,86406,86538,86553,86575,86663,86693,86789,86795,86806,86808,86853,86906,87042,87254,87299,87326,87327,87459,87568,87570,87625,87671,87672,87716,87745,87765,87819,87836,87949,87974,87985,88005,88097,88185,88225,88236,88499,88560,88730,88780,88799,88993,89018,89105,89283,89316,89330,89362,89819,89827,89855,89900,89953,89958,90275,90427,90453,90460,90613,90648,90889,90970,90982,91054,91093,91151,91186,91259,91281,91290,91409,91458,91485,91523,91545,91701,91707,91816,91878,92009,92065,92083,92134,92199,92246,92293,92337,92339,92384,92522,92590,92774,92906,92916,92936,92944,93096,93153,93296,93343,93346,93378,93630,93733,93906,93939,94084,94157,94366,94484,94680,94764,94783,94811,94963,94992,95055,95180,95344,95378,95462,95464,95493,95514,95602,95708,95754,95783,95823,95962,96082,96100,96111,96454,96517,96613,96628,96655,96788,96804,96872,96985,97015,97040,97187,97279,97310,97376,97396,97415,97546,97555,97732,97745,97752,97857,97860,97896,97918,98141,98361,98708,98878,98899,98936,99028,99254,99266,99333,99497,99621,99680,99798,99893,99998",
          "10,38,97,100,158,165,375,583,647,655,657,786,875,1032,1056,1058,1073,1095,1103,1109,1161,1221,1228,1257,1259,1308,1330,1493,1544,1568,1571,1639,1650,1658,1718,1723,1743,2041,2117,2120,2200,2222,2485,2668,2670,2732,2806,2824,3159,3348,3357,3734,3777,3797,4122,4259,4300,4467,4511,4672,4733,4850,4912,4952,5019,5137,5179,5238,5267,5316,5320,5520,5531,5566,5707,5761,5836,5947,6057,6059,6179,6181,6204,6246,6259,6415,6464,6694,6795,6809,6884,7072,7111,7212,7392,7446,7491,7496,7856,7871,7980,7986,7990,8035,8260,8276,8367,8454,8502,8584,8707,8723,8785,9002,9067,9298,9410,9447,9502,9514,9579,9616,9802,9855,10073,10147,10172,10174,10233,10234,10264,10332,10343,10356,10449,10495,10509,10537,10542,10557,10580,10721,10814,10839,11049,11077,11108,11310,11400,11462,11533,11555,11655,11664,11747,11843,12182,12184,12319,12651,12669,12715,12783,12829,12961,13077,13129,13150,13341,13445,13619,13824,13892,13961,14038,14108,14412,14458,14494,14545,14577,14591,14605,14882,14932,15020,15084,15095,15205,15260,15278,15392,15550,15699,15715,15725,15759,15830,16038,16153,16445,16516,16588,16596,16663,16719,16781,16912,16934,17075,17131,17143,17184,17238,17314,17332,17446,17523,17692,17786,17826,17865,18000,18173,18194,18251,18323,18376,18524,18698,18708,18831,18935,18957,18982,19004,19060,19114,19326,19447,19471,19581,19635,19677,19744,19799,19917,20009,20021,20044,20046,20048,20051,20677,20715,20922,20980,21030,21144,21193,21705,21808,21836,21872,21887,21949,21992,22235,22338,22366,22423,22454,22520,22710,23153,23246,23531,23765,24305,24306,24438,24475,24533,24592,24637,24643,24793,24875,24930,25008,25052,25292,25311,25351,25422,25583,25680,25862,26148,26295,26312,26330,26489,26749,26891,26931,27050,27191,27549,27645,27654,27663,27684,27712,27721,27875,27906,27940,27973,28280,28383,28393,28441,28681,28693,28711,28863,28883,28958,29021,29045,29195,29243,29260,29267,29306,29322,29333,29339,29425,29566,29736,29877,30003,30352,30381,30787,31135,31163,31280,31322,31397,31570,31687,31831,32381,32499,32586,32962,32963,33050,33454,33484,33485,33513,33687,34498,34557,34674,34731,34759,34847,34993,35004,35294,35635,35713,35814,36078,36198,36254,36312,36322,36463,36509,36559,36560,36701,36745,36765,36820,36829,37008,37079,37265,37441,37443,37554,37565,37805,37870,37879,37946,37947,38011,38125,38207,38316,38382,38442,38742,39251,39269,39320,39428,39504,39568,39611,39949,39993,40147,40256,40517,41039,41245,41303,41481,42061,42197,42234,42289,42398,42616,42676,42680,42745,42804,42838,42846,42875,42968,43061,43172,43206,43264,43348,43385,43768,43824,44208,44305,44448,44477,44515,44570,44677,44779,44809,44825,44830,44882,45071,45325,45498,45504,45616,45689,45713,45814,45877,45986,45988,46059,46441,46491,46529,46594,46616,46772,46838,47226,48089,48178,48420,48498,48542,49032,49319,49603,49853,50354,50413,50436,50685,50851,51289,51324,51326,51446,51535,51613,51681,51828,51894,51940,52110,52133,52245,52405,52451,52483,52572,52633,52688,52713,52720,52731,52775,52831,52977,53007,53112,53167,53295,53301,53396,53450,53485,53524,53575,53580,53643,53705,53886,53987,53989,53992,53998,54130,54137,54195,54215,54224,54238,54290,54325,54527,54599,54611,54619,54631,54712,54713,54714,54753,54851,54872,54911,55101,55109,55164,55363,55402,55403,55486,55535,55555,55585,55992,56095,56117,56181,56233,56304,56354,56394,56465,56567,56572,56596,56624,56709,56724,56860,57025,57053,57084,57153,57217,57329,57345,57508,57674,57739,57760,57938,57973,58026,58048,58127,58139,58177,58245,58258,58429,58619,58642,59010,59086,59207,59312,59481,59551,59655,59693,59739,59895,59937,60099,60244,60287,60359,60451,60560,60705,60987,61016,61074,61116,61226,61401,61696,61707,61829,61896,62011,62054,62075,62140,62237,62242,62291,62434,62514,62556,62577,62635,62724,62741,62746,62837,62939,63044,63088,63146,63354,63382,63431,63580,63592,63753,63777,63863,63883,63935,63963,64291,64425,64473,64492,64499,64571,64583,64607,64635,64963,64982,65114,65229,65515,65707,65834,65890,65899,65969,66018,66122,66139,66272,66294,66454,66537,66541,66594,66648,66735,66863,66881,67004,67010,67114,67154,67198,67205,67231,67288,67321,67457,67473,67560,67691,67723,67784,67928,68003,68045,68070,68105,68106,68116,68122,68183,68317,68335,68384,68401,68404,68409,68437,68606,68803,68893,68933,68995,69056,69152,69191,69278,69382,69384,69420,69453,69500,69610,69674,69687,69795,69926,69973,70051,70145,70196,70264,70279,70305,70330,70393,70397,70436,70515,70718,70956,71330,71475,71479,71489,71614,71719,71751,71866,71941,71975,72175,72230,72250,72284,72326,72422,72500,72515,72543,72715,72737,72843,72902,72908,72978,73057,73098,73323,73348,73433,73537,73554,73594,73612,73700,73734,73758,73818,73940,74004,74016,74018,74130,74228,74319,74323,74332,74336,74416,74503,74526,74670,74673,74784,74807,75022,75242,75348,75350,75406,75493,75500,75779,75789,75840,75871,75892,76008,76013,76044,76075,76083,76212,76378,76463,76465,76476,76509,76560,76601,76936,76941,77042,77194,77272,77299,77414,77420,77422,77463,77491,77504,77698,77800,77865,77924,77964,78185,78288,78339,78374,78403,78582,78695,78755,78770,78777,78780,78801,78854,78928,79118,79135,79199,79371,79555,79616,79666,79667,79679,79888,79913,80000,80186,80267,80396,80418,80436,80575,80649,80780,80935,80962,80982,81249,81345,81564,81701,81724,81748,81882,81952,81968,82154,82210,82213,82274,82275,82386,82494,82783,82861,82950,82997,83039,83336,83406,83427,83497,83584,83585,83879,83935,83948,84058,84152,84330,84353,84409,84645,84881,85428,85451,85471,85597,85815,85862,86033,86079,86088,86163,86174,86206,86339,86348,86433,86469,86576,86616,86801,86914,86926,86955,87003,87033,87063,87066,87076,87220,87393,87444,87508,87541,87554,87786,87842,87983,88033,88083,88107,88115,88222,88280,88303,88387,88400,88452,88475,88641,88685,89046,89066,89072,89312,89345,89413,89455,89519,89618,89645,89791,89835,89850,89856,89932,89935,90125,90193,90213,90451,90455,90487,90495,90546,90653,90705,90926,90938,91004,91109,91185,91265,91450,91556,91566,91570,91649,91732,91737,91893,92069,92108,92158,92167,92292,92310,92356,92404,92407,92431,92533,92538,92551,92731,92995,93285,93302,93331,93448,93619,93756,93956,93987,94097,94148,94169,94275,94413,94453,94462,94482,94533,94652,94657,94658,94750,94850,94951,95138,95346,95390,95528,95569,95675,95681,95713,96027,96053,96101,96121,96154,96182,96212,96255,96304,96354,96432,96538,96562,96595,96608,96642,96686,96706,96765,96885,97027,97051,97079,97105,97144,97341,97508,97576,97692,97799,97943,97992,97999,98076,98119,98234,98276,98323,98343,98435,98678,98756,98826,98916,98948,98964,99020,99024,99101,99300,99384,99392,99399,99407,99455,99614,99666,99748,99880,99912,99987",
          "25,93,117,154,212,265,281,618,660,776,960,1047,1066,1113,1130,1204,1241,1340,1383,1416,1478,1562,1598,1619,1714,1726,1766,1777,1914,2128,2263,2304,2317,2509,2527,2569,2948,3066,3580,3617,3674,3715,4378,4620,4777,5174,5207,5211,5227,5232,5269,5276,5288,5337,5345,5416,5446,5458,5462,5472,5476,5487,5593,5661,5670,5701,5732,5735,5795,5869,5930,5973,5977,5996,6060,6083,6175,6217,6233,6268,6269,6281,6387,6390,6589,6681,6684,6705,6786,6827,6851,6911,6943,7030,7043,7061,7069,7300,7318,7467,7562,7592,7610,7751,7753,7759,7791,7799,7845,7859,7878,7883,7897,7947,8308,8442,8452,8548,8582,8680,8871,8960,8991,9000,9015,9032,9038,9118,9176,9784,9870,9964,9966,10011,10019,10133,10201,10367,10737,10758,10795,10910,11294,11337,11354,11577,11650,11767,11775,11791,11869,12305,12307,12641,12657,12743,12785,12943,12948,12966,12970,13057,13115,13364,13400,13867,13868,14066,14336,14466,14514,14621,14750,14771,14940,15058,15109,15293,15448,15474,15584,15660,15754,16072,16709,17069,17259,17357,17455,17503,17554,17586,17711,17763,18079,18428,18572,18640,18722,18922,18971,19124,19260,19273,19626,19777,20885,21126,21142,21402,22160,22669,22708,22741,23007,23064,23318,23589,23615,23938,24271,24597,24638,24650,25087,25395,25705,25983,26377,26415,26573,26628,26838,26898,27003,27213,27273,27314,27825,27881,27968,28061,28270,28388,28757,28864,28979,29132,29420,29796,29828,31047,31419,31564,32521,32637,33275,33768,33826,34377,34551,35508,35627,35696,35698,35737,35891,36122,36494,36574,36680,36693,36926,37425,38357,38424,38695,38790,38849,39122,39247,39301,39870,40460,42154,42308,42325,42373,42456,42640,42708,42718,42795,42809,43100,43710,43884,43892,43954,44420,45178,45313,45447,45539,45935,46855,46955,47147,48185,48457,48835,49390,49518,50057,50067,50073,50362,50439,50480,50613,50620,50879,50906,51011,51051,51074,51084,51124,51186,51236,51408,51472,51480,51611,51887,51892,52081,52104,52204,52225,52296,52384,52391,52436,52542,52583,52819,52901,52910,52941,53289,53474,53935,54080,54138,54175,54182,54219,54239,54264,54312,54319,54391,54431,54553,54778,54816,54823,54917,54944,54947,54968,55004,55248,55269,55401,55494,55590,55683,55726,55825,55834,55881,55953,56080,56183,56222,56426,56531,56538,56546,56686,56827,56906,56945,56946,57052,57151,57177,57193,57222,57232,57292,57314,57335,57381,57403,57515,57614,57709,57752,57757,57805,57867,58120,58280,58609,58620,58849,58850,59063,59215,59273,59313,59473,59548,59628,59725,59770,59998,60050,60264,60418,60518,60547,60587,60731,60806,61010,61058,61273,61294,61314,61365,61388,61423,61440,61446,61460,61526,61645,62105,62124,62127,62153,62212,62226,62378,62450,62463,62464,62562,62663,62710,62745,62829,62840,62858,62913,62944,62988,63091,63126,63328,63506,63601,63602,63646,63780,63877,63884,63929,63959,63987,64032,64210,64228,64314,64457,64522,64609,64706,64708,64756,64765,64955,64967,65038,65073,65282,65312,65320,65427,65456,65473,65630,65926,66030,66213,66366,66367,66422,66689,66824,66865,66999,67013,67099,67148,67189,67302,67317,67639,67957,68056,68089,68190,68226,68241,68294,68596,68617,68675,68801,68841,68973,69064,69187,69248,69346,69389,69504,69542,69545,69587,69603,69725,69831,69912,69938,69972,69989,70187,70382,70426,70466,70518,70534,70563,70639,70640,70758,70768,70773,70955,70982,71000,71086,71144,71167,71199,71240,71410,71416,71484,71520,71563,71679,71850,72029,72089,72125,72205,72242,72322,72380,72453,72472,72544,72545,72613,72628,72663,72674,72729,72776,72819,72997,73008,73010,73012,73062,73069,73122,73231,73299,73358,73463,73511,73600,73623,73651,73665,73749,74057,74149,74150,74188,74372,74489,74740,74773,74782,74843,74880,74947,74948,75007,75056,75058,75067,75089,75271,75342,75538,75548,75750,75956,76047,76052,76147,76169,76389,76583,76613,76636,76677,76711,76760,76785,76825,76851,76876,76963,76982,76998,77018,77125,77202,77288,77321,77359,77406,77415,77438,77541,77559,77597,77742,77768,77806,77810,77866,77958,77962,78062,78069,78192,78322,78410,78541,78649,78826,78900,78907,78968,78974,79087,79168,79173,79223,79352,79497,79546,79572,79587,79604,79621,79761,79781,79796,79845,79876,80002,80424,80447,80514,80571,80599,80600,80604,80669,80690,80693,80722,80759,81059,81113,81209,81224,81261,81365,81432,81497,82087,82140,82218,82225,82483,82525,82540,82743,82802,82884,82890,82905,82930,82984,82999,83001,83020,83203,83232,83277,83293,83354,83368,83437,83447,83455,83467,83565,83603,83649,83651,83722,83724,83829,83846,83880,83884,83898,83929,83932,83995,84117,84155,84170,84214,84406,84445,84469,84477,84497,84608,84664,84665,84681,84684,84724,84841,84901,84939,84959,85000,85132,85266,85333,85354,85390,85419,85441,85477,85580,85631,85641,85688,85711,85900,86349,86361,86481,86494,86599,86709,86843,86854,86932,87067,87149,87227,87233,87252,87311,87492,87578,87582,87738,87750,87821,87874,87875,87951,87998,88128,88151,88246,88376,88529,88592,88653,88672,88897,88903,88934,89089,89225,89268,89332,89444,89484,89537,89589,89981,90110,90173,90384,90404,90412,90417,90509,90755,90768,90831,90882,91016,91019,91183,91635,91790,91842,91884,91980,92316,92447,92706,92867,92885,93183,93196,93274,93405,93439,93483,93485,93582,93757,93890,93903,93955,93979,94000,94013,94042,94184,94189,94196,94235,94288,94370,94408,94452,94477,94555,94706,94744,94773,94805,94860,94936,95021,95124,95206,95212,95365,95373,95451,95456,95604,95657,95693,95833,95963,96070,96116,96257,96266,96338,96346,96437,96615,96735,96774,96796,96812,96997,97208,97224,97323,97427,97464,97494,97535,97560,97661,97891,97987,98038,98176,98290,98383,98434,98478,98535,98578,98617,98720,98778,98859,98950,98997,99100,99104,99246,99411,99560,99671,99691,99766,99817,99920",
          "67,255,270,275,306,329,330,349,390,399,485,513,616,676,692,701,736,740,850,1031,1054,1146,1358,1359,1421,1550,1609,1893,2850,3130,3166,3615,3747,4007,4423,4432,4835,4894,5237,5305,5466,5558,5609,5639,5657,5860,5997,6109,6122,6211,6249,6274,6282,6329,6355,6573,6614,6633,6671,6762,6799,6905,6956,7025,7244,7264,7325,7375,7379,7480,7542,7603,7609,7613,7730,7738,7739,7837,7982,8141,8271,8318,8357,8381,8403,8435,8466,8558,8589,8644,8832,8849,8984,9048,9087,9096,9202,9340,9389,9475,9746,9900,10076,10153,10178,10215,10281,10472,10508,10544,10565,10615,10665,10686,10695,10714,10716,10728,11091,11098,11311,11318,11335,11568,11661,11815,11839,11894,11988,12053,12100,12108,12303,12460,12466,12497,12541,12581,12711,13032,13080,13256,13315,13633,14542,14668,14754,14833,15375,15385,15412,15540,15674,15975,16231,16414,16630,16660,16952,17065,17276,17391,17681,18272,18384,18511,18658,18704,18753,18882,19270,19487,19583,19733,19787,19977,20126,20224,20260,20284,20501,20797,20828,20830,21137,22137,22191,22728,22769,22974,24155,24185,24694,24914,25518,25634,25711,25822,26130,26314,26817,26908,27009,27037,27160,27467,27470,27707,27737,27802,28287,28505,28898,28904,28925,29000,29104,29344,29362,29406,29509,29586,29640,29999,30868,30986,35340,35576,35577,35806,35951,35957,36117,36381,36736,36741,36947,37051,37202,37240,37427,37622,37625,37676,37694,37735,37949,37987,38069,38070,38120,38604,38688,38984,39054,39133,39238,39250,39260,39378,39478,39612,40567,41740,42148,43267,43602,45245,45506,48349,50555,50559,50561,50618,50682,50797,50926,50931,50969,51091,51112,51136,51172,51250,51312,51318,51345,51494,51588,51731,51818,51882,51926,51948,52057,52076,52086,52094,52147,52290,52297,52409,52500,52514,52635,52698,52807,52866,52982,52988,53026,53491,53528,53534,53545,53596,53780,53960,54003,54031,54103,54190,54197,54248,54262,54310,54390,54651,54789,54791,54857,54953,55372,55524,55560,55588,55597,55637,55646,55869,55884,56040,56045,56149,56271,56276,56489,56580,56661,56786,56792,56819,56850,56859,57424,57425,57440,57591,57792,57799,57831,57868,57931,58011,58241,58260,58583,58603,58607,58650,58695,58703,58847,59128,59219,59322,59458,59488,59592,59621,59636,59753,59782,59856,59909,59996,60054,60084,60143,60163,60172,60191,60200,60288,60401,60424,60486,60501,60667,60793,60828,60901,60983,61019,61125,61508,61571,61669,61671,61988,62068,62096,62115,62272,62281,62290,62569,62620,62722,62816,62959,63125,63157,63267,63324,63367,63444,63483,63722,63779,63800,63822,63916,63937,64071,64191,64305,64329,64460,64566,64675,64794,64844,64972,65217,65258,65278,65319,65333,65428,65535,65555,65588,65716,65743,65778,66016,66041,66100,66219,66248,66299,66309,66542,66699,66755,66916,66944,67048,67123,67206,67229,67379,67380,67429,67469,67474,67619,67620,67852,68032,68060,68180,68188,68366,68479,68552,68555,68584,68590,68934,68984,69076,69200,69224,69381,69407,69417,69658,69767,70180,70422,70464,70476,70513,70533,70589,70646,70674,70767,70836,71002,71013,71149,71220,71260,71309,71385,71438,71491,71673,71755,71782,71789,71800,71842,71922,71935,72127,72197,72266,72320,72337,72388,72401,72413,73037,73121,73219,73317,73346,73371,73446,73476,73560,73636,73701,73786,73946,73990,74064,74073,74102,74246,74474,74569,74644,74823,74952,75009,75049,75122,75197,75338,75356,75502,75558,75611,75705,75828,75864,75946,76198,76261,76309,76437,76481,76541,76546,76551,76607,76626,76647,76678,77019,77040,77043,77102,77192,77224,77277,77324,77330,77331,77515,77676,77766,77854,77992,78078,78113,78117,78279,78310,78485,78537,78605,78629,78635,78680,78703,78753,78833,78943,78977,79009,79036,79120,79232,79234,79340,79614,79677,79754,79790,79835,79951,79984,80025,80093,80287,80312,80366,80598,80806,80826,81029,81204,81279,81301,81338,81425,81476,81522,81524,81559,81601,82089,82233,82314,82427,82641,82733,82776,82929,82980,83003,83183,83214,83312,83331,83438,83472,83480,83531,83570,83583,83672,83698,84013,84084,84243,84252,84256,84460,84545,84548,84595,84695,84705,84867,84987,85112,85176,85312,85316,85351,85366,85446,85447,85508,85514,85522,85568,85668,85721,85800,85818,85897,85980,86043,86068,86105,86113,86144,86244,86357,86434,86452,86501,86537,86541,86551,86559,86588,86620,86678,86699,86834,86900,86907,86939,86980,87024,87097,87199,87432,87456,87892,87959,88069,88117,88488,88519,88520,88694,88705,88811,88857,89051,89110,89165,89275,89457,89664,89703,89710,89769,89812,90148,90210,90323,90426,90492,90523,90647,90762,90771,90801,90813,90821,90823,90824,90872,90895,90917,90922,90980,91105,91106,91145,91230,91399,91474,91480,91669,91681,91687,91756,91786,91833,91900,91928,91996,92010,92132,92255,92376,92487,92531,92552,92572,92586,92659,92899,92900,93026,93157,93177,93482,93489,93493,93574,93952,94054,94142,94201,94216,94225,94356,94359,94383,94509,94537,94625,94702,94724,95014,95062,95225,95230,95370,95448,95622,95712,95760,95790,95800,95894,96168,96210,96314,96357,96358,96400,96519,96527,96574,96593,96610,96718,96878,96922,96934,97026,97258,97296,97336,97468,97527,97574,97604,97606,97641,97649,97686,97894,97920,98032,98049,98050,98053,98084,98149,98199,98215,98270,98351,98362,98363,98376,98530,98555,98569,98601,98728,98974,99124,99189,99202,99286,99338,99786,99911,99924",
          "33,243,260,437,501,533,538,632,679,694,703,894,896,901,949,954,959,1055,1089,1092,1172,1178,1179,1355,1402,1533,1534,1567,1885,1887,1892,2202,2208,2406,2417,2419,3218,3289,3290,3297,3298,3360,3434,3575,3596,3735,3886,4180,4240,4608,4732,4913,5012,5087,5106,5121,5128,5185,5204,5235,5290,5344,5346,5356,5365,5378,5550,5554,5599,5767,5797,5855,5879,5909,6034,6147,6163,6184,6230,6280,6426,6475,6478,6512,6631,6700,6807,6895,6933,7182,7419,7713,7873,7895,8025,8054,8085,8186,8228,8351,8383,8451,8487,8594,8610,8718,8763,8805,8851,8877,8888,8978,8993,9057,9156,9164,9254,9347,9371,9515,9544,9549,9876,9907,9913,9931,9940,9974,10013,10018,10268,10278,10294,10410,10414,10442,10682,10683,10739,10891,10946,11151,11153,11239,11471,11811,11905,12019,12079,12155,12304,12461,12513,12805,12817,12935,12990,13252,13351,13458,13523,13629,13853,13999,14006,14040,14072,14134,14150,14324,14365,14438,14513,14684,14709,14710,14749,14780,14812,14970,15087,15452,15463,15504,15580,15606,15655,15769,15972,16085,16332,16493,16591,16684,16999,17098,17227,17358,17836,18201,18222,18273,18287,18390,18619,18653,18681,18771,18849,19071,19173,19217,19454,19492,19527,19568,19746,19793,19929,20154,20244,20252,20259,20328,20334,20579,20819,20906,20960,21118,21698,21722,21807,22029,22133,22415,22545,22586,22627,22763,22837,23011,23226,23308,23326,23433,23694,23742,23840,23843,24335,24345,24379,24405,24715,24934,24944,24954,25265,25316,25331,25372,25467,25572,25723,25759,25800,26346,26650,26708,26959,27032,27043,27131,27171,27251,27298,27413,27450,27482,27592,27622,27752,27793,27888,27915,27985,28276,28387,28402,28730,28742,28755,28830,28891,28927,28963,29046,29122,29183,29222,29274,29450,29505,29544,29548,29623,29651,29751,30162,30394,30549,30633,30778,30829,30836,31010,31303,31438,31445,32460,32517,33330,33565,33736,33862,34055,34841,34986,35043,35053,35106,35623,35823,35890,35936,36000,36006,36119,36305,36487,36513,36572,36627,36718,36758,36875,36994,37013,37075,37220,37345,37543,37559,37873,37986,38117,38192,38301,38371,38572,38628,38780,38798,38813,38922,38972,39099,39189,39220,39289,39299,39420,39547,39633,39735,39744,39813,39864,40159,40449,40451,40855,41297,41355,41466,42553,42844,42850,43023,43489,43586,43709,43726,43790,44022,44062,44116,44233,44592,44618,44734,44762,44781,44869,45016,45044,45097,45137,45242,45554,45639,45745,46016,46109,46632,46695,46968,47117,47224,47242,47523,48274,48302,48512,48656,49060,49223,49237,49577,49777,50000,50012,50086,50189,50333,50414,50440,50514,50521,50536,50597,50681,50692,50707,50815,50831,51045,51133,51259,51292,51330,51483,51549,51662,51864,51999,52046,52226,52286,52317,52374,52377,52556,52641,52806,52851,52908,52916,52954,53052,53059,53156,53165,53268,53307,53420,53549,53627,53628,53681,53693,53695,53703,53866,54008,54163,54349,54405,54438,54518,54674,54852,54869,54905,55027,55038,55050,55127,55225,55232,55317,55328,55474,55487,55511,55530,55542,55761,55779,55822,55890,56046,56332,56562,56676,56677,56739,56764,56889,56990,57055,57108,57155,57235,57257,57270,57368,57550,57661,57828,57881,58140,58477,58768,58823,58871,58875,58891,58912,58950,59139,59394,59438,59532,59572,59639,59704,59873,59907,59934,60138,60145,60222,60233,60284,60361,60417,60461,60534,60565,60609,60622,60802,60827,60856,60859,60880,60947,60955,61036,61040,61208,61227,61282,61329,61472,61500,61531,61554,61564,61676,61731,61820,61930,61955,62137,62285,62339,62382,62392,62429,62431,62550,62582,62853,62970,63027,63051,63062,63083,63159,63189,63217,63247,63252,63254,63327,63430,63434,63466,63516,63584,63657,63668,63679,63778,63837,63853,63867,63948,63956,64006,64147,64151,64287,64373,64487,64537,64560,64823,64879,64943,65123,65206,65430,65632,65801,65807,65810,65933,65999,66022,66065,66249,66381,66528,66621,66814,66820,66868,66937,67027,67228,67307,67326,67387,67528,67559,67616,67671,67755,67800,67929,67952,68016,68036,68068,68338,68402,68426,68621,68652,68666,68712,68769,68879,68894,68900,69028,69103,69180,69190,69269,69419,69465,69509,69516,69556,69557,69609,69675,69780,69799,69846,69868,69957,69980,70015,70124,70472,70759,70839,70862,70901,71123,71151,71306,71399,71464,71676,71916,71917,72059,72179,72314,72347,72354,72427,72508,72532,72664,72710,72730,73216,73296,73327,73392,73412,73424,73427,73590,73918,73922,73968,74092,74195,74204,74274,74290,74353,74498,74588,74592,74676,74855,74866,75189,75267,75343,75358,75403,75436,75466,75547,75731,75838,75900,75923,75982,76011,76062,76069,76103,76104,76248,76255,76275,76293,76295,76356,76499,76515,76597,76757,76885,76985,76994,77088,77161,77214,77218,77276,77317,77349,77752,77761,77982,78094,78115,78391,78467,78483,78536,78538,78800,78811,78852,78880,79244,79266,79311,79451,79485,79649,80024,80052,80128,80246,80257,80557,80579,80641,80657,80715,80739,80811,80872,80979,80989,81002,81096,81108,81120,81228,81494,81596,81755,81763,81935,82011,82028,82197,82305,82348,82433,82611,82650,82657,82710,82788,83134,83137,83187,83220,83410,83428,83489,83525,83595,83778,83876,83959,83962,83994,84011,84087,84140,84249,84334,84357,84598,84602,84625,84668,84678,84790,84848,84997,85061,85114,85159,85161,85179,85231,85284,85392,85490,85600,85810,85958,85963,86159,86165,86234,86316,86422,86465,86484,86485,86516,86596,86731,86775,86847,87013,87100,87124,87175,87223,87346,87389,87484,87597,87707,87814,87964,87991,88017,88047,88153,88398,88412,88564,88568,88627,88644,89020,89103,89108,89199,89346,89360,89419,89436,89604,89627,89717,89790,89921,89949,89992,90016,90143,90147,90195,90276,90317,90364,90469,90586,90610,90635,90792,90803,91020,91088,91100,91114,91137,91214,91215,91333,91366,91515,91553,91663,91929,92195,92261,92287,92320,92411,92436,92494,92611,92677,92905,92910,92931,92958,92970,92998,93066,93181,93206,93217,93352,93358,93414,93450,93452,93490,93559,93646,93695,93702,93719,93720,93740,93742,93800,93853,93866,93909,94001,94053,94161,94165,94314,94457,94461,94483,94792,94944,95075,95154,95229,95327,95330,95475,95649,95905,95954,96037,96239,96246,96361,96409,96509,96597,96618,96637,96691,96693,96694,96794,96910,96952,97014,97017,97157,97184,97311,97413,97488,97507,97591,97733,97761,97878,97889,98090,98111,98112,98275,98285,98371,98520,98553,98600,98629,98652,98694,98732,98868,98872,98952,98985,99145,99259,99337,99400,99424,99431,99559,99771,99801,99855,99889,99891,99917,99985",
          "81,113,153,236,246,1233,1981,2058,2096,2135,2136,2204,2300,2412,2452,2587,2809,2879,3007,3054,3190,3204,3228,3291,3683,4052,4215,4441,4524,4548,4715,4801,4809,5684,6350,6638,6655,6814,6997,7185,7205,7255,7529,7548,7830,7945,8002,8004,8087,8118,8128,8137,8156,8175,8232,8270,8294,8319,8346,8416,8424,8650,8662,8669,8840,8891,8897,8975,9039,9100,9134,9201,9238,9296,9397,9480,9619,9657,9695,9728,9753,9759,9828,9847,9942,9965,9986,10075,10081,10094,10113,10208,10236,10303,10360,10411,10455,10521,10588,10645,10656,10763,10835,10930,11036,11158,11216,11234,11285,11291,11386,11413,11480,11575,11614,11652,11709,11918,11938,12034,12097,12203,12222,12241,12384,12388,12393,12450,12479,12488,12602,12698,12716,12724,12735,12794,12849,12887,12963,13016,13020,13039,13172,13192,13267,13343,13419,13436,13534,13590,13609,13647,13856,13924,13932,13989,14008,14071,14121,14179,14216,14217,14236,14255,14274,14383,14388,14424,14440,14521,14540,14616,14619,14635,14691,14703,14787,14920,15015,15039,15053,15091,15121,15148,15262,15281,15300,15357,15369,15396,15471,15475,15566,15604,15623,15642,15661,15694,15737,15756,15832,15851,15870,15889,15984,16003,16098,16193,16250,16288,16326,16345,16383,16402,16440,16442,16450,16459,16497,16592,16649,16668,16706,16782,16839,16877,16896,16915,16991,17067,17141,17253,17257,17334,17352,17447,17466,17485,17580,17637,17694,17713,17732,17789,17827,17846,17884,17922,18017,18036,18131,18219,18225,18226,18245,18283,18302,18359,18378,18397,18530,18568,18606,18663,18682,18758,18796,18815,18872,18910,18929,19176,19237,19252,19271,19309,19347,19385,19442,19499,19588,19605,19651,19727,19765,19860,19879,19936,19955,20012,20031,20107,20115,20145,20152,20164,20240,20273,20373,20411,20430,20449,20468,20487,20525,20639,20753,20810,20821,20829,20843,20848,20886,20981,20998,21095,21171,21190,21293,21304,21323,21342,21456,21475,21494,21570,21589,21627,21665,21684,21912,21969,22045,22083,22121,22178,22216,22254,22273,22311,22349,22368,22406,22539,22558,22599,22615,22634,22653,22672,22729,22748,22783,22786,22881,22900,22919,22921,22957,22993,23014,23071,23108,23147,23166,23185,23261,23337,23348,23356,23375,23413,23432,23489,23546,23603,23641,23660,23679,23755,23774,23869,23888,23926,23945,23964,23970,23983,24002,24021,24040,24116,24135,24265,24268,24439,24572,24629,24648,24705,24743,24800,24805,24819,24821,24935,24971,24990,25104,25105,25123,25180,25199,25332,25369,25427,25465,25471,25484,25560,25579,25606,25626,25712,25769,25788,25789,25807,25902,25921,25997,26016,26073,26092,26213,26235,26370,26472,26510,26529,26535,26649,26871,27023,27270,27480,27840,27846,28068,28179,28285,28309,28333,28410,28448,28562,28600,28733,28771,28773,28901,28922,29051,29073,29170,29189,29398,29458,29648,29816,29875,29949,29987,30063,30121,30179,30208,30237,30295,30353,30382,30411,30440,30469,30527,30556,30643,30672,30701,30730,30846,30933,30991,31020,31049,31107,31136,31165,31194,31223,31252,31281,31484,31513,31542,31571,31600,31629,31658,31745,31803,31832,31919,31977,32035,32151,32267,32383,32412,32470,32528,32557,32644,32673,32731,32818,32847,32934,32992,33021,33137,33166,33195,33224,33253,33311,33340,33369,33427,33514,33543,33601,33630,33688,33775,33804,33833,33891,33949,34065,34094,34123,34181,34210,34239,34268,34297,34355,34384,34413,34442,34471,34529,34587,34703,34732,34761,34790,34848,34906,34935,34964,35005,35031,35047,35056,35067,35094,35168,35241,35331,35365,35430,35436,35476,35517,35524,35642,35774,35815,35833,35904,35963,36022,36085,36145,36236,36275,36334,36393,36428,36498,36583,36706,36755,36800,36954,37026,37085,37267,37294,37326,37364,37515,37587,37646,37705,37769,37887,37958,37988,38017,38148,38183,38241,38389,38432,38479,38566,38637,38709,38768,38827,38891,39009,39129,39134,39198,39227,39233,39270,39369,39388,39452,39511,39570,39700,39831,39890,39978,40227,40285,40314,40343,40372,40430,40488,40546,40575,40604,40633,40691,40749,40807,40894,40952,41068,41126,41155,41213,41271,41329,41387,41416,41445,41474,41503,41532,41846,42032,42119,42159,42244,42253,42295,42379,42419,42458,42532,42581,42735,42743,42790,42799,42852,42893,43004,43055,43084,43177,43207,43208,43217,43282,43311,43319,43494,43501,43604,43623,43673,43691,43766,43841,43870,43902,43928,44003,44078,44122,44207,44217,44220,44240,44267,44276,44315,44402,44464,44639,44768,44789,44858,44876,44951,44980,44994,45026,45099,45147,45208,45266,45339,45396,45454,45478,45526,45583,45641,45770,45828,45900,46017,46039,46092,46136,46163,46185,46193,46208,46292,46423,46588,46697,46715,46733,46832,46884,46930,47106,47253,47258,47276,47370,47383,47389,47396,47433,47509,47546,47573,47644,47740,47781,47854,47901,48013,48043,48077,48148,48155,48335,48423,48504,48596,48623,48692,48702,48707,48752,48844,48885,48971,49038,49064,49068,49116,49136,49159,49187,49199,49217,49402,49457,49491,49509,49516,49548,49699,49718,49804,49805,49825,49867,49884,49886,49913,49926,49965,49994,50123,50217,50234,50249,50257,50267,50328,50469,50495,50545,50638,50683,50701,50712,50793,50808,50849,50987,51092,51127,51240,51325,51496,51621,51697,51746,51781,51800,51863,51914,51956,51975,52025,52234,52298,52418,52468,52472,52479,52898,52917,53041,53090,53179,53344,53376,53412,53422,53471,53478,53525,53585,53655,53696,53709,53749,53851,53856,53944,53967,54004,54091,54268,54435,54465,54520,54568,54576,54813,54870,54900,54924,55073,55178,55218,55430,55497,55598,55756,56052,56097,56116,56155,56204,56214,56254,56333,56357,56419,56430,56454,56547,56659,56678,56698,56717,56934,56998,57124,57159,57240,57248,57289,57354,57480,57577,57668,57750,57769,57872,57979,58112,58279,58643,58669,58722,58910,59013,59029,59046,59087,59167,59242,59342,59385,59442,59464,59640,59682,59779,59841,59843,59878,59914,59953,60193,60261,60302,60444,60581,60693,60703,60939,60946,60950,61064,61133,61167,61172,61400,61469,61525,61624,61689,62045,62130,62389,62424,62430,62546,62560,62666,62680,62846,62848,63003,63421,63443,63572,63588,63614,63662,63798,63846,63893,63993,64036,64063,64149,64156,64217,64365,64392,64406,64510,64552,64628,64661,64672,64699,64734,64739,64834,64855,64882,65006,65025,65170,65178,65179,65234,65334,65349,65432,65617,65629,65710,65833,65837,65839,65888,65928,65957,66003,66378,66592,66674,66703,66788,67042,67075,67163,67188,67197,67314,67348,67433,67466,67489,67730,67955,68010,68109,68327,68403,68427,68850,69050,69121,69338,69551,69575,69733,69885,70061,70130,70214,70308,70313,70323,70445,70491,70610,70619,70712,70742,70800,70887,70895,70904,71093,71219,71299,71522,71523,71637,71674,71802,71836,71924,72005,72082,72087,72095,72096,72203,72290,72302,72381,72462,72478,72562,72595,72644,72673,72722,72747,72846,73036,73128,73179,73197,73269,73316,73328,73387,73395,73453,73467,73589,73615,73847,73890,73950,73958,73995,74165,74181,74194,74293,74304,74310,74376,74384,74507,74522,74573,74653,74799,74842,75073,75159,75241,75318,75321,75362,75444,75454,75504,75638,75780,75806,75992,75995,76031,76035,76058,76100,76108,76161,76224,76298,76330,76355,76447,76629,76691,76826,76845,76862,76917,76931,77014,77119,77184,77382,77444,77579,77654,77669,77718,77953,77957,78105,78124,78162,78178,78236,78349,78411,78441,78504,78507,78556,78884,78979,78989,79094,79115,79129,79306,79309,79335,79360,79407,79524,79603,79760,80243,80244,80269,80286,80354,80537,80567,80656,80704,80716,80791,80827,81062,81067,81090,81103,81232,81620,81668,81726,81853,81974,82115,82130,82177,82362,82398,82566,82661,82721,82812,82834,82865,82913,83000,83093,83133,83141,83359,83377,83477,83544,83614,83615,83715,83818,84153,84168,84193,84383,84465,84511,84564,84612,84624,84627,84945,85169,85191,85220,85340,85356,85367,85385,85388,85510,85517,85555,85717,85843,85857,85896,85985,85995,86011,86040,86061,86106,86136,86176,86369,86385,86512,86577,86583,86741,86928,86978,87043,87110,87132,87152,87260,87280,87325,87329,87433,87467,87479,87500,87566,87660,87687,87692,87701,87846,87922,88095,88316,88396,88465,88477,88536,88788,88826,88885,88989,89005,89187,89370,89538,89670,89691,89916,90023,90138,90146,90166,90237,90290,90314,90315,90501,90507,90508,90749,90965,90971,90996,91049,91135,91300,91414,91456,91596,91636,91661,91686,91729,91846,91926,91987,91989,91999,92085,92156,92189,92218,92346,92418,92496,92756,92826,92921,92953,93020,93074,93099,93187,93247,93293,93350,93479,93548,93679,93840,93851,94588,94592,94611,94662,94710,94711,94767,94779,94967,95037,95249,95335,95508,95598,95682,95743,95758,95818,96000,96050,96075,96112,96142,96247,96362,96365,96387,96445,96525,96548,96566,96587,96651,96800,96891,96907,96939,96941,97030,97042,97074,97117,97156,97217,97276,97283,97300,97340,97411,97729,97765,97871,98069,98241,98294,98300,98310,98577,98837,98957,98983,99005,99067,99116,99121,99183,99437,99452,99500,99526,99607,99642,99683,99710,99733,99745,99770,99822",
          "96,110,134,198,204,323,1654,1675,1786,1940,1953,1976,1978,2007,2008,2022,2050,2057,2069,2126,2140,2154,2176,2178,2180,2186,2210,2257,2267,2313,2329,2336,2342,2354,2362,2378,2405,2411,2430,2432,2435,2436,2437,2462,2466,2469,2505,2514,2562,2571,2583,2585,2621,2631,2642,2651,2672,2677,2698,2754,2757,2783,2786,2805,2822,2827,2862,2914,2929,2944,2982,3034,3049,3086,3102,3105,3118,3120,3134,3156,3164,3180,3184,3192,3249,3250,3264,3274,3321,3342,3368,3380,3433,3452,3462,3477,3485,3487,3493,3496,3503,3510,3526,3532,3560,3562,3597,3603,3637,3656,3677,3686,3690,3694,3720,3724,3744,3784,3843,3853,3883,3906,3918,3924,3932,3951,3954,3956,3961,3967,3978,4023,4079,4092,4107,4109,4125,4149,4172,4176,4182,4187,4197,4203,4221,4226,4253,4267,4299,4312,4396,4412,4442,4459,4570,4571,4584,4599,4605,4607,4616,4660,4663,4668,4670,4702,4735,4743,4760,4811,4844,4870,4886,4963,4981,4982,6253,6257,6262,6381,6382,6389,6435,6546,6579,6583,6599,6640,6787,6833,6837,6859,6901,6967,7125,7167,7172,7218,7274,7388,7487,7502,7607,7633,7645,7690,7697,7788,7960,7984,8003,8097,8112,8248,8295,8296,8412,8446,8511,8572,8590,8621,8626,8645,8655,8665,8666,8688,8690,8779,8873,8893,8923,8940,8958,9006,9016,9027,9054,9062,9076,9145,9253,9279,9471,9507,9578,9583,9752,9777,9858,9859,9860,9949,10088,10099,10181,10217,10372,10403,10516,10536,10552,10789,10831,10864,11058,11101,11155,11269,11426,11450,11476,11538,11701,11718,11734,11763,11914,11991,12018,12044,12134,12281,12289,12324,12331,12522,12575,12780,12855,13024,13179,13199,13202,13283,13317,13389,13395,13405,13411,13478,13519,13535,13695,13720,13737,13780,13901,13936,13965,14265,14322,14341,14739,14781,14782,15168,15233,15524,15569,15761,15818,16114,16371,16502,16767,16774,16844,16918,16924,17078,17113,17335,17573,17625,17818,17866,17880,18289,18357,18411,18442,18499,19046,19243,19389,19666,19887,19909,19913,20413,20437,20537,20815,21195,21237,21392,21486,21978,22059,22244,22288,22320,22474,22486,22517,22617,22686,23102,23182,23230,23471,23866,23916,24093,24205,24251,24466,24518,24603,24827,25127,25188,25213,25306,25546,25591,25745,26029,26053,26250,26284,26336,26381,26449,26508,26516,26518,26522,26525,26551,26558,26564,26576,26583,26584,26586,26609,26612,26698,26764,26785,26801,26829,26881,26910,26951,26966,27049,27086,27089,27115,27148,27192,27232,27285,27292,27294,27337,27370,27405,27433,27443,27449,27455,27495,27498,27503,27522,27534,27536,27548,27567,27587,27598,27601,27602,27603,27605,27621,27626,27634,27650,27676,27692,27723,27797,27820,27827,27833,27866,27867,27895,27910,27950,27957,27979,27986,28008,28019,28027,28039,28041,28046,28060,28064,28097,28131,28134,28160,28165,28191,28212,28219,28231,28247,28282,28288,28305,28401,28421,28483,28520,28529,28532,28538,28560,28581,28598,28602,28613,28614,28620,28628,28657,28668,28680,28684,28706,28707,28720,28725,28726,28796,28821,28855,28861,28887,28916,28937,28953,28995,29013,29028,29038,29072,29074,29086,29091,29101,29103,29125,29147,29149,29150,29160,29220,29304,29315,29317,29355,29356,29360,29369,29374,29384,29393,29394,29410,29434,29443,29468,29476,29521,29564,29568,29595,29624,29637,29643,29649,29655,29673,29713,29787,29797,29865,29905,29909,30022,30051,30080,30109,30138,30167,30196,30225,30254,30283,30312,30341,30370,30399,30428,30457,30486,30515,30544,30573,30602,30631,30660,30689,30718,30747,30776,30805,30834,30863,30892,30921,30950,30979,31008,31037,31066,31095,31124,31153,31182,31211,31240,31269,31298,31327,31356,31385,31414,31443,31472,31501,31530,31559,31588,31617,31646,31675,31704,31733,31762,31791,31820,31849,31878,31907,31936,31965,31994,32023,32052,32081,32110,32139,32168,32197,32255,32284,32313,32342,32371,32400,32429,32458,32487,32516,32545,32574,32603,32632,32661,32690,32719,32748,32777,32806,32835,32864,32893,32922,32951,32980,33009,33038,33067,33096,33125,33154,33183,33212,33241,33270,33299,33328,33357,33386,33415,33444,33473,33502,33531,33560,33589,33618,33647,33676,33705,33763,33792,33821,33850,33879,33908,33937,33966,33995,34053,34082,34111,34140,34169,34198,34227,34256,34285,34314,34343,34372,34401,34430,34459,34488,34517,34546,34575,34604,34633,34662,34691,34720,34749,34778,34807,34836,34865,34894,34923,34952,34981,35022,35045,35111,35130,35134,35139,35185,35221,35223,35274,35283,35285,35287,35296,35313,35348,35422,35433,35464,35493,35541,35600,35659,35677,35732,35791,35850,35921,35945,35980,36039,36102,36134,36162,36194,36221,36252,36292,36351,36366,36410,36419,36482,36541,36566,36600,36608,36616,36617,36664,36723,36782,36853,36884,36912,36938,36971,37043,37102,37161,37170,37178,37225,37284,37343,37378,37414,37422,37473,37532,37604,37663,37722,37786,37815,37845,37876,37904,37975,38034,38093,38165,38224,38283,38347,38406,38418,38465,38493,38536,38553,38558,38595,38654,38671,38673,38726,38785,38844,38853,38859,38871,38908,38967,39026,39097,39156,39215,39287,39346,39405,39469,39528,39541,39546,39587,39658,39717,39741,39776,39848,39907,39966,39975,40128,40157,40186,40215,40244,40273,40302,40331,40360,40389,40418,40447,40476,40505,40534,40563,40592,40621,40650,40679,40708,40737,40766,40795,40824,40853,40882,40911,40940,40969,40998,41027,41056,41085,41114,41143,41201,41230,41259,41288,41317,41346,41375,41404,41433,41462,41491,41520,41549,42019,42048,42077,42106,42135,42144,42157,42195,42230,42285,42296,42326,42330,42360,42369,42384,42404,42435,42460,42522,42597,42634,42637,42672,42759,42778,42783,42834,42872,42909,42972,42996,43018,43030,43071,43146,43233,43308,43383,43412,43430,43470,43510,43514,43545,43559,43620,43630,43707,43753,43782,43857,43875,43944,43956,43988,44019,44025,44039,44094,44099,44153,44155,44181,44256,44295,44298,44331,44365,44418,44439,44493,44499,44502,44517,44519,44520,44568,44631,44655,44702,44703,44730,44742,44748,44761,44805,44827,44892,44967,45042,45115,45163,45194,45224,45282,45299,45355,45412,45470,45514,45542,45599,45657,45677,45729,45786,45844,45857,45916,45973,46010,46024,46038,46049,46107,46110,46135,46187,46213,46246,46248,46395,46463,46467,46473,46486,46547,46584,46585,46606,46615,46628,46635,46659,46721,46723,46759,46811,46813,46819,46849,46867,46962,47004,47165,47166,47218,47282,47341,47349,47369,47372,47379,47543,47588,47622,47636,47645,47650,47680,47703,47709,47722,47809,47844,47903,47912,47913,47931,48049,48058,48076,48078,48170,48180,48225,48227,48315,48339,48348,48433,48440,48455,48458,48484,48544,48551,48562,48576,48628,48629,48653,48751,48780,48830,48881,48934,48980,48981,48994,49006,49030,49084,49090,49095,49120,49125,49126,49208,49226,49356,49389,49461,49589,49597,49620,49635,49650,49693,49705,49730,49768,49776,49799,49818,49854,49864,49881,49919,49923,49946,49955,49958,49959,50036,50191,50241,50306,50331,50376,50404,50535,50573,50577,50677,50732,50779,50955,50965,51052,51137,51194,51392,51430,51435,51486,51584,51675,51680,51695,51801,51821,51888,51921,51934,51972,52008,52051,52328,52462,52464,52503,52585,52651,52678,52692,52700,52741,53037,53060,53111,53118,53207,53235,53345,53489,53561,53667,53685,53704,53755,53775,53803,53904,53912,53928,54071,54154,54585,54609,54785,54895,54979,55014,55045,55161,55193,55214,55256,55357,55405,55440,55576,55632,55643,55650,55654,55907,55998,56058,56382,56428,56529,56569,56631,56703,56801,56890,56925,57004,57184,57207,57336,57362,57393,57423,57426,57443,57491,57628,57637,57795,57912,57999,58022,58188,58382,58395,58479,58604,58677,58680,58759,58889,58939,58955,58964,59003,59092,59332,59388,59398,59431,59523,59722,59875,59881,59968,60224,60273,60293,60357,60365,60522,60649,60769,60850,60889,60930,61022,61105,61232,61380,61580,61632,61654,61657,61744,61805,61923,61931,61951,61954,61994,62020,62052,62108,62141,62249,62256,62262,62315,62356,62488,62650,62653,62716,62738,62842,62942,63020,63114,63127,63138,63274,63291,63334,63445,63456,63473,63857,64196,64231,64285,64344,64350,64379,64388,64393,64578,64602,64625,64741,64773,64836,64853,64956,65016,65134,65371,65376,65405,65448,65465,65491,65602,65683,65748,65760,65766,65772,65953,66158,66244,66298,66417,66472,66475,66566,66619,66635,66686,66698,66734,66773,66792,66797,66867,66878,66966,67078,67094,67130,67149,67255,67328,67356,67384,67490,67545,67836,67843,68174,68223,68280,68311,68332,68677,68721,68829,68847,68972,68976,69067,69147,69166,69183,69294,69313,69354,69369,69459,69513,69552,69579,69713,69759,69870,69905,69917,69944,69990,70021,70073,70104,70109,70191,70212,70244,70314,70521,70548,70856,70964,71072,71100,71125,71239,71251,71254,71272,71307,71375,71442,71541,71598,71611,71621,71723,71846,71884,72067,72112,72160,72350,72581,72815,72879,73094,73206,73214,73361,73617,73626,73692,73731,73742,73762,73784,73899,73944,74020,74099,74182,74258,74359,74368,74387,74440,74444,74621,74749,74860,74895,74941,75035,75081,75168,75270,75298,75376,75536,75566,75650,75753,75781,75834,75841,75852,75887,75911,75942,75955,76078,76089,76112,76168,76259,76277,76374,76592,76593,76603,76688,76747,76789,77003,77132,77135,77316,77690,77692,77703,77706,77910,77919,77948,78036,78174,78267,78338,78412,78622,78692,78810,78865,79071,79293,79345,79386,79392,79419,79506,79651,80265,80344,80614,80800,80929,80936,80949,81070,81190,81191,81330,81362,81376,81411,81482,81489,81514,81523,81539,81644,81669,81808,81864,81867,81916,81990,82093,82134,82146,82147,82258,82366,82370,82431,82656,82727,82822,82875,83098,83153,83167,83206,83215,83275,83302,83310,83322,83365,83383,83429,83485,83604,83605,83670,83693,83841,83926,84000,84053,84082,84130,84286,84439,84642,84662,84748,84786,84792,84834,84912,84992,85021,85308,85313,85349,85414,85416,85521,85884,85974,85998,86016,86021,86039,86115,86140,86245,86513,86601,86613,86662,86668,86710,86725,86885,87007,87008,87070,87184,87193,87221,87313,87426,87509,87550,87564,88030,88073,88251,88320,88660,88686,88727,88842,88844,88905,88924,88950,89106,89181,89233,89246,89392,89698,89718,89831,89950,89997,90020,90069,90135,90160,90236,90641,90671,90743,90936,90998,91141,91172,91180,91196,91232,91235,91324,91404,91413,91432,91441,91448,91486,91660,91664,92051,92055,92093,92178,92478,92530,92912,92977,93007,93058,93130,93228,93465,93469,93518,93638,93654,93673,93677,93784,93880,93969,94033,94038,94155,94188,94294,94322,94327,94469,94534,94544,94637,94665,94714,94800,94842,94897,94986,94988,95052,95096,95111,95122,95214,95224,95321,95332,95339,95465,95470,95850,96041,96181,96275,96294,96391,96633,96751,96808,96846,96926,96976,97019,97099,97243,97289,97317,97573,97631,97783,97787,97831,97834,97844,97938,97966,97988,98159,98312,98340,98409,98467,98558,98742,98921,98971,99089,99109,99268,99284,99516,99567,99612,99627,99654,99726,99840,99961"
        ]
      }
    },
    {
      "pool": {
        "type": "Pool",
        "name": "kShardSelectionHuge.LoadBalancer.02",
        "servers": [
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312",
          "localhost:12312"
        ],
        "shards": [
          "53,97,103,154,251,253,305,320,478,545,595,692,872,998,1040,1180,1181,1528,1611,1631,1733,1827,1881,1939,2000,2101,2196,2229,2253,2257,2285,2355,2431,2449,2508,2521,2669,2738,2876,2904,2981,2982,3007,3044,3143,3262,3263,3389,3436,3439,3499,3533,3559,3571,3629,3642,3676,3682,3842,3846,3863,3932,4033,4101,4161,4178,4209,4297,4381,4432,4453,4460,4477,4672,4713,4761,4772,4865,5065,5114,5555,5689,5801,5860,6032,6066,6246,6304,6412,6446,6530,6652,6745,6830,7090,7151,7155,7174,7327,7418,7456,7458,7509,7812,7814,7906,8095,8205,8295,8714,8727,8745,8866,8901,8930,9015,9024,9036,9227,9235,9340,9421,9475,9482,9493,9553,10071,10156,10176,10181,10409,10416,10436,10456,10554,10574,10583,10599,10683,10714,10718,10750,10805,10824,10899,11090,11100,11177,11249,11286,11405,11546,11583,11618,11704,11785,11789,11834,11921,11955,11966,12123,12134,12226,12240,12260,12366,12411,12477,12602,12686,12734,12796,12797,12836,12888,12913,12925,13017,13076,13112,13128,13221,13236,13272,13291,13381,13467,13498,13549,13733,13782,13807,13860,13949,14002,14133,14161,14202,14212,14282,14285,14299,14421,14439,14534,14542,14612,14681,14724,14747,14851,14873,14883,14953,14970,15196,15269,15276,15340,15451,15491,15555,15568,15601,15635,15790,15804,15809,15915,15941,16023,16039,16085,16344,16462,16464,16504,16586,16645,16688,16761,16834,16840,16847,16880,16957,17016,17026,17033,17103,17175,17287,17332,17431,17467,17548,17634,17781,17794,17830,17891,17922,18157,18158,18235,18358,18367,18456,18460,18469,18489,18523,18556,18566,18579,18603,18749,18812,18962,18973,19031,19081,19096,19245,19254,19277,19341,19417,19447,19476,19489,19495,19511,19597,19611,19633,19650,19658,19706,19764,19831,19862,19962,20176,20298,20365,20386,20426,20511,20552,20685,20687,20832,20871,20918,20936,21161,21529,21542,21726,21793,21847,22000,22063,22115,22209,22222,22408,22426,22489,22528,22575,22578,22599,22803,22867,22888,23044,23124,23281,23306,23411,23441,23577,23636,23670,23796,23817,23857,23871,23901,23943,23974,24034,24084,24206,24274,24319,24331,24435,24493,24501,24537,24619,24744,24751,24782,24853,24961,25061,25229,25299,25441,25448,25458,25460,25498,25503,25560,25577,25585,25674,25821,25832,25834,25928,26062,26093,26111,26184,26213,26373,26423,26454,26468,26546,26586,26589,26751,26792,26874,26907,26986,27061,27242,27251,27315,27342,27413,27473,27549,27567,27609,27628,27772,27816,27838,27904,27973,28000,28024,28033,28139,28293,28356,28396,28411,28680,28702,28718,28746,28894,28895,29011,29091,29160,29174,29226,29263,29305,29398,29845,29856,29893,30112,30123,30234,30287,30338,30458,30603,30732,30783,30905,30943,30961,31225,31407,31424,31433,31495,31547,31571,31612,31628,31642,31671,31679,31852,31913,31918,32166,32205,32220,32233,32279,32282,32497,32552,32719,32747,32947,32975,33027,33071,33184,33197,33245,33323,33385,33407,33462,33516,33586,33615,33648,33658,33662,33686,33775,33796,33850,33872,34025,34112,34122,34125,34165,34183,34223,34225,34255,34316,34399,34452,34581,34585,34607,34678,34716,34758,34789,34849,34998,35248,35322,35409,35741,35788,35981,36037,36048,36147,36187,36324,36365,36412,36581,36641,36692,36860,36907,37141,37143,37167,37246,37406,37481,37500,37517,37688,37747,37952,37977,37984,38063,38153,38168,38202,38268,38303,38375,38526,38570,38615,38655,38676,38683,38691,38731,38751,38808,38922,38985,38997,39085,39121,39129,39437,39465,39674,39682,39786,39840,39877,39942,39947,39992,40073,40157,40209,40518,40712,40714,40725,40753,40782,40792,40833,40909,41008,41216,41253,41271,41318,41496,41499,41598,41650,41692,41704,41816,41830,41956,41961,42301,42310,42327,42367,42376,42633,42696,42748,42820,42844,42847,42987,43154,43200,43290,43428,43466,43555,43642,43731,43749,43857,43880,43913,44020,44023,44122,44231,44279,44432,44506,44617,44679,44717,44755,44765,44775,44849,44880,44936,45077,45140,45191,45294,45313,45391,45441,45455,45506,45540,45599,45813,45876,45967,46045,46106,46168,46258,46272,46383,46444,46459,46492,46609,46718,46760,46855,46915,46920,46950,46955,47118,47127,47169,47183,47200,47279,47411,47836,47881,47901,47927,47953,47957,47986,48028,48061,48089,48099,48150,48173,48179,48237,48257,48322,48438,48452,48515,48876,48962,49114,49116,49227,49252,49409,49442,49521,49572,49642,50092,50098,50157,50172,50236,50355,50368,50422,50425,50487,50509,50664,50679,50737,50802,50859,50914,51074,51142,51194,51280,51318,51530,51555,51631,51633,51646,51697,51764,51796,52105,52136,52167,52265,52301,52367,52378,52381,52401,52527,52599,52630,52669,52678,52774,52804,52810,52837,52859,52890,52957,53007,53043,53155,53162,53279,53288,53338,53486,53555,53621,53734,53755,53895,54023,54088,54107,54189,54208,54231,54236,54255,54269,54357,54480,54485,54565,54610,54688,54779,54793,54824,54867,55021,55211,55260,55349,55354,55377,55492,55712,55764,55836,55985,56077,56265,56368,56410,56628,56657,56767,56783,56923,56972,57046,57056,57167,57219,57228,57351,57379,57399,57436,57461,57553,57645,57679,57864,57907,57914,58032,58042,58215,58361,58415,58422,58434,58637,58918,59033,59052,59106,59153,59235,59240,59400,59442,59568,59874,59951,59965,60259,60326,60447,60471,60480,60507,60576,60723,60730,60910,61047,61085,61172,61212,61637,61665,61670,61716,61729,61756,61854,61989,62042,62291,62298,62389,62390,62431,62441,62454,62523,62527,62644,62800,62832,62875,62912,62926,63045,63211,63339,63350,63509,63511,63530,63578,63585,63719,63855,63925,64156,64168,64193,64259,64274,64325,64332,64500,64501,64503,64560,64561,64711,64718,64774,64802,64888,64965,64976,64992,65051,65297,65363,65410,65495,65542,65606,65674,65708,65827,65843,66024,66060,66063,66115,66137,66142,66198,66202,66206,66294,66313,66363,66437,66460,66465,66469,66490,66496,66635,66892,66932,66974,67053,67129,67227,67339,67483,67493,67520,67582,67671,67803,68181,68274,68281,68313,68376,68541,68762,68884,69067,69105,69245,69312,69341,69476,69523,69567,69576,69624,69643,69724,69793,69880,69950,70036,70161,70214,70216,70292,70302,70314,70363,70493,70643,70775,70807,70837,70885,71090,71140,71190,71249,71269,71304,71365,71380,71404,71588,71662,71684,71730,71735,71777,71779,71834,71947,72050,72109,72164,72215,72293,72405,72594,72743,72870,72969,73048,73329,73358,73361,73374,73663,73898,73973,74009,74352,74486,74712,74739,74850,74922,75028,75113,75147,75163,75170,75213,75291,75303,75390,75691,75717,75748,75776,75844,76045,76298,76393,76422,76431,76464,76507,76535,76566,76568,76573,76648,76666,76683,76684,76763,76880,76905,77027,77049,77078,77173,77251,77274,77314,77437,77493,77506,77649,77738,77807,78077,78237,78252,78255,78263,78275,78327,78384,78388,78389,78408,78453,78460,78531,78594,78833,78886,78896,78930,79055,79141,79270,79289,79303,79397,79489,79493,79615,79680,79731,79846,79863,79902,79947,80017,80110,80120,80154,80252,80354,80473,80495,80606,80641,80717,80758,80797,80807,81120,81183,81202,81246,81366,81385,81404,81457,81463,81522,81553,81673,81788,81836,81841,81842,81898,82057,82192,82309,82313,82386,82426,82516,82558,82613,82640,82661,82819,82904,83067,83097,83246,83275,83318,83358,83364,83412,83424,83538,83540,83627,83640,83770,83845,84022,84109,84139,84170,84178,84480,84540,84542,84559,84584,84588,84693,84738,84808,84875,84879,84891,84970,85074,85151,85156,85256,85297,85363,85456,85557,85743,85973,86228,86331,86349,86351,86354,86357,86397,86465,86515,86520,86531,86546,86625,86779,86794,86811,86855,86963,87079,87083,87255,87265,87315,87333,87385,87404,87604,87632,87737,87805,88006,88117,88134,88159,88177,88184,88268,88297,88426,88546,88620,88701,89330,89359,89383,89392,89397,89429,89478,89491,89845,90265,90287,90382,90629,90713,90848,90978,91213,91255,91383,91387,91407,91481,91527,91596,91611,91675,91776,91790,91793,91894,91984,91986,92006,92086,92117,92195,92199,92287,92406,92407,92414,92469,92702,92866,92982,93031,93286,93405,93445,93620,93636,93734,93736,93740,93822,93846,93875,93912,93967,94088,94136,94151,94300,94310,94328,94436,94664,94721,94729,94812,94866,94896,94959,94982,95047,95104,95171,95219,95337,95447,95745,95926,96048,96063,96070,96110,96196,96292,96324,96485,96592,96622,96635,96866,97026,97044,97155,97202,97207,97239,97288,97306,97444,97462,97484,97489,97492,97535,97634,97635,97662,97850,97885,97934,98032,98068,98218,98235,98272,98359,98434,98443,98476,98639,98837,98885,98923,98963,99095,99169,99254,99269,99277,99306,99435,99596,99654,99657,99666,99713,99729,99834,99898",
          "54,66,108,232,246,359,512,687,746,815,834,968,988,1010,1030,1111,1141,1186,1255,1346,1370,1472,1508,1545,1569,1626,1693,1696,1755,1759,1851,1857,1974,1988,2131,2297,2325,2332,2333,2362,2392,2505,2525,2547,2578,2740,2766,2855,2963,3125,3177,3273,3382,3514,3715,3818,3836,3907,4096,4127,4166,4168,4194,4271,4495,4527,4533,4535,4537,4655,4669,4724,4764,4776,4828,4855,5145,5191,5262,5346,5376,5400,5418,5458,5618,5631,5645,5701,5822,5887,5937,6018,6088,6174,6221,6238,6286,6360,6363,6399,6482,6554,6699,6722,6754,6774,6904,7024,7055,7113,7330,7705,7707,7777,7785,7834,7841,8015,8075,8153,8159,8194,8272,8412,8598,8627,8706,8712,8827,8932,8989,8990,9068,9102,9154,9170,9248,9306,9336,9508,9639,9696,9758,9824,9863,9888,9939,9949,10039,10109,10216,10226,10270,10356,10362,10429,10600,10762,10934,11009,11054,11131,11282,11389,11408,11440,11442,11487,11518,11588,11592,11637,11639,11784,11872,11894,11916,12005,12009,12080,12213,12215,12220,12346,12358,12429,12481,12506,12518,12554,12630,12806,12817,12824,12883,12901,13113,13204,13207,13493,13513,13532,13542,13805,13840,13886,13927,13994,13995,14277,14324,14330,14358,14404,14419,14605,14609,14621,14801,14855,14870,15076,15153,15242,15500,15638,15639,15650,15738,16153,16164,16295,16326,16350,16362,16370,16378,16381,16541,16577,16707,16855,16876,17202,17204,17267,17398,17407,17471,17569,17621,17751,17834,17842,17923,17970,17984,18190,18191,18273,18328,18331,18336,18439,18458,18586,18629,18633,18654,19032,19072,19142,19184,19185,19358,19405,19485,19708,19720,19752,19789,20070,20089,20247,20324,20353,20549,20550,20557,20628,20744,20889,20892,20967,21023,21043,21074,21189,21247,21312,21427,21519,21520,21647,21705,21764,21766,21907,21994,22009,22045,22244,22249,22377,22563,22621,22723,22756,22995,23137,23232,23252,23313,23358,23539,23837,23896,23907,23973,24086,24223,24417,24512,24569,24704,24715,24745,24772,24773,24777,25389,25418,25461,25590,25598,25615,25792,25807,25894,25906,25938,26100,26201,26298,26343,26348,26383,26392,26504,26578,26615,26617,26732,26825,26883,26966,27006,27030,27039,27122,27136,27243,27266,27325,27337,27369,27434,27489,27500,27540,27670,27694,27862,27914,27979,28188,28287,28298,28336,28352,28364,28416,28505,28574,28610,28651,28661,28738,28816,28836,28867,28889,28890,28914,28917,28976,28977,29007,29074,29104,29128,29147,29183,29235,29243,29316,29386,29510,29558,29645,29651,29678,29781,29878,30018,30077,30108,30111,30173,30217,30296,30335,30358,30361,30521,30591,30626,30690,30698,30749,30760,30794,30853,30968,31093,31261,31344,31558,31559,31566,31568,31656,31668,31885,31917,31959,32156,32216,32306,32435,32465,32477,32500,32709,32833,32940,32991,33100,33111,33230,33248,33318,33357,33395,33464,33470,33892,34101,34146,34301,34307,34387,34482,34487,34498,34834,34926,34963,34992,35022,35205,35249,35364,35383,35532,35556,35568,35643,35706,35739,35740,35752,35811,35827,35875,35898,35966,36015,36069,36184,36296,36468,36661,36720,36724,36743,36777,36806,36890,36923,36924,37010,37016,37036,37053,37444,37454,37488,37587,37722,37786,37973,38029,38069,38075,38119,38145,38183,38243,38421,38437,38518,38839,38869,38917,38957,38971,38979,38993,39000,39073,39106,39166,39273,39320,39420,39605,39684,39736,39824,39845,40068,40150,40314,40557,40675,40676,40722,40759,40885,40891,40960,40975,40997,41003,41039,41175,41268,41294,41350,41365,41429,41454,41463,41545,41576,41627,41737,41767,41938,41977,41991,42005,42185,42257,42258,42348,42400,42513,42517,42642,42688,42701,42783,43045,43054,43073,43123,43158,43173,43253,43296,43352,43396,43443,43458,43536,43571,43580,43585,43602,43636,43788,43794,43812,43862,43905,43907,44039,44113,44148,44243,44390,44482,44564,44592,44743,44785,44832,44851,44870,44979,44980,45194,45201,45212,45457,45579,45614,45638,45725,45734,45827,45940,45972,46009,46027,46053,46070,46084,46221,46253,46324,46568,46570,46650,46657,46709,46788,46801,46869,46923,47027,47109,47133,47138,47753,47779,47885,47896,47900,47931,48021,48100,48119,48301,48312,48355,48464,48530,48597,48646,48730,48775,49055,49291,49324,49374,49454,49462,49465,49527,49534,49583,49624,49690,49717,49749,49754,49760,49792,49803,49818,49888,49912,50041,50052,50072,50095,50224,50242,50279,50481,50529,50627,50685,50791,51026,51054,51096,51173,51376,51412,51494,51507,51589,51604,51632,51700,51750,51871,52179,52319,52360,52365,52391,52426,52454,52514,52530,52568,52571,52575,52629,52632,52644,52664,52781,52843,52871,52946,52964,53069,53100,53163,53210,53302,53735,53951,53960,53978,53992,54009,54054,54123,54186,54249,54260,54379,54579,54706,54776,54796,55147,55290,55351,55519,55572,55611,55660,55757,55827,55857,55862,55913,55969,55999,56042,56313,56367,56573,56612,56615,56736,56859,56896,56911,56950,57036,57133,57183,57222,57278,57630,57635,57699,57741,57834,57903,57984,58019,58114,58137,58323,58348,58387,58476,58599,58619,58687,58787,58850,58957,58974,59002,59023,59347,59374,59412,59484,59552,59562,59614,59769,59815,59901,60027,60156,60182,60313,60564,60593,60764,60789,60838,60853,60908,60913,60964,61014,61038,61053,61069,61093,61228,61299,61366,61607,61695,61710,61759,61942,62082,62132,62153,62158,62224,62283,62325,62456,62482,62722,62746,62791,62844,62940,62967,63041,63421,63488,63506,63570,63888,63920,64048,64150,64210,64249,64369,64427,64510,64535,64562,64693,64712,64829,64851,64922,64969,65027,65060,65074,65288,65320,65482,65572,65720,65826,65897,65940,66011,66090,66110,66306,66340,66351,66439,66582,66602,66670,66697,66752,66895,67025,67150,67248,67333,67348,67439,67449,67527,67561,67619,67676,67712,67766,67774,68005,68098,68112,68113,68156,68189,68203,68283,68448,68499,68511,68528,68679,68723,68725,68764,68828,68923,68937,69016,69120,69127,69193,69239,69270,69293,69315,69365,69379,69433,69451,69511,69586,69644,69742,69755,69926,69931,69932,70012,70082,70130,70212,70277,70324,70437,70618,70703,70706,70745,70746,70898,70968,71103,71119,71250,71270,71540,71560,71915,71939,71963,72000,72146,72252,72330,72359,72370,72429,72538,72602,72656,72807,72837,72966,73066,73068,73129,73205,73213,73253,73271,73407,73408,73430,73629,73667,73744,73768,73787,73789,73875,73917,74076,74134,74146,74505,74573,74632,74644,74745,74780,74849,74955,75188,75233,75328,75385,75393,75457,75656,75669,75808,76101,76193,76246,76333,76406,76420,76505,76532,76660,76844,76883,76886,76913,77151,77203,77246,77406,77451,77863,78057,78087,78092,78149,78304,78332,78376,78450,78573,78637,78639,78759,78781,78782,78836,79008,79010,79074,79146,79319,79395,79492,79597,79624,79683,79704,79710,79727,79889,79981,80002,80026,80209,80215,80221,80417,80519,80574,80657,80739,80902,80905,80949,80960,80996,81128,81132,81191,81282,81419,81453,81491,81696,81716,81755,81907,81949,81979,82220,82284,82622,82674,82811,82858,83028,83171,83346,83406,83439,83616,83677,83701,83728,83787,83963,84015,84095,84101,84106,84144,84231,84244,84263,84273,84347,84410,84533,84554,84648,84708,84729,84882,85038,85087,85088,85188,85224,85245,85564,85741,85815,85817,85827,85865,85907,85957,86103,86264,86280,86308,86493,86551,86613,86650,86677,86924,87099,87172,87200,87347,87392,87522,87634,87646,87647,87848,87851,87874,87943,88058,88199,88296,88361,88460,88581,88597,88640,88800,88910,88922,88966,89095,89154,89226,89327,89371,89389,89405,89506,89790,89876,90219,90620,90855,90952,90954,91036,91084,91096,91147,91301,91385,91386,91603,91643,91651,91783,91871,91903,91992,92012,92014,92074,92169,92268,92324,92355,92358,92491,92523,92674,92692,92695,92728,92762,92765,92795,92876,92937,92999,93008,93082,93111,93134,93136,93214,93257,93315,93458,93471,93514,93613,93692,93728,93763,93930,93934,94017,94109,94217,94338,94386,94452,94556,94610,94687,94815,94850,94916,94943,95040,95076,95102,95131,95284,95319,95425,95578,95681,95760,95836,95882,95958,96003,96051,96120,96246,96259,96309,96372,96392,96523,96559,96624,96632,96682,96703,96796,96822,96837,96925,96927,97081,97118,97132,97143,97171,97177,97298,97307,97337,97419,97422,97455,97497,97522,97553,97581,97603,97620,97739,97743,97746,97842,97870,97939,97992,98036,98046,98069,98074,98108,98116,98196,98295,98349,98350,98433,98635,98748,98784,98832,98851,98878,98888,98925,99160,99302,99319,99354,99404,99406,99439,99443,99534,99736,99742,99745,99763,99862,99916,99975",
          "7,187,247,508,535,679,681,741,758,832,910,942,979,984,1289,1314,1411,1413,1437,1519,1614,1641,1782,1846,1849,1897,1903,1963,2030,2033,2259,2274,2308,2329,2384,2522,2561,2564,2606,2736,2789,2800,2802,3035,3060,3093,3145,3152,3185,3437,3459,3512,3579,3605,3628,3651,3769,3814,3833,3952,4103,4154,4172,4184,4269,4405,4413,4419,4614,4682,4743,4757,4924,5102,5135,5142,5433,5545,5574,5583,5637,5722,5807,5833,5905,5987,5991,6003,6131,6173,6379,6512,6603,6635,6672,6691,6921,7049,7167,7214,7235,7259,7443,7459,7492,7502,7547,7674,7722,7732,7744,7757,7796,7889,7893,7921,7932,7951,8112,8123,8142,8307,8587,8628,8734,8839,9108,9187,9247,9293,9318,9412,9526,9560,9665,9668,9781,9851,10019,10483,10668,10885,10890,10967,11204,11235,11325,11332,11392,11454,11701,11885,11933,12049,12085,12237,12336,12362,12368,12391,12440,12608,12676,12725,12766,12868,12894,12983,13149,13215,13235,13305,13385,13409,13455,13476,13531,13534,13563,13727,13809,13936,13951,13954,14073,14409,14496,14637,14679,14848,14876,14944,14951,14980,15027,15049,15057,15123,15150,15299,15372,15513,15563,15621,15625,15651,15724,15815,15866,15911,16031,16418,16426,16459,16518,16585,16668,16685,16734,16887,16951,17006,17059,17169,17222,17402,17404,17449,17652,17675,17734,17766,18078,18119,18180,18312,18329,18379,18397,18432,18529,18544,18738,18827,18936,18970,18996,19006,19201,19300,19332,19350,19959,19960,20005,20065,20082,20210,20331,20416,20446,20665,20723,20755,20771,21032,21137,21179,21192,21235,21440,21484,21525,21643,21678,21831,21833,21861,21962,22075,22146,22190,22252,22317,22521,22624,22706,22730,22787,22862,22933,22966,23086,23162,23262,23408,23459,23482,23676,23677,23844,23893,24045,24126,24152,24179,24411,24424,24591,24706,24812,24846,24988,25105,25153,25347,25569,25603,25672,25735,25883,25885,25916,26231,26263,26276,26300,26338,26369,26480,26502,26513,26644,26837,26899,26989,27027,27241,27278,27367,27392,27593,27664,27759,27836,27910,27938,27956,28044,28123,28418,28494,28517,28527,28613,28815,28881,28952,28994,29085,29093,29115,29165,29297,29384,29714,29721,29786,29886,29986,30040,30065,30242,30268,30276,30389,30405,30406,30582,30586,30719,30781,30846,30926,30944,30972,30990,31145,31150,31203,31212,31262,31281,31285,31315,31377,31384,31553,31562,31577,31626,31719,31722,31747,31800,31833,31911,31947,31989,32011,32023,32118,32169,32181,32201,32264,32369,32671,32672,32910,32983,33014,33031,33047,33053,33089,33240,33242,33507,33549,33556,33642,33678,33753,33827,33837,33894,33925,34038,34040,34118,34132,34236,34530,34603,34735,34920,34951,35002,35012,35104,35262,35340,35356,35374,35384,35388,35400,35451,35615,35641,35644,35772,35813,35816,35844,35931,36011,36132,36165,36255,36293,36660,36663,36773,36793,36934,36972,37313,37366,37378,37470,37499,37573,37656,37693,37738,37837,37905,37980,38026,38074,38194,38223,38342,38386,38653,38664,38675,38838,38950,38970,39048,39088,39192,39230,39246,39321,39441,39486,39576,39580,39592,39673,39698,39801,39807,39820,39985,40136,40186,40191,40228,40229,40238,40255,40400,40431,40905,40964,40988,41140,41194,41202,41436,41486,41680,41691,41975,42113,42418,42429,42561,42589,42754,42813,42856,42934,42989,43069,43193,43256,43338,43535,43566,43750,43842,44078,44098,44099,44235,44532,44549,44699,44744,44747,44766,44837,44859,44902,44967,45010,45053,45117,45134,45161,45224,45237,45381,45404,45612,45648,45652,45667,45875,45938,45968,46113,46137,46372,46422,46430,46677,46818,46874,46969,47022,47023,47026,47073,47121,47142,47234,47270,47313,47347,47442,47455,47640,47660,47758,47791,47816,47856,47872,47904,47990,48297,48338,48475,48494,48510,48525,48580,48651,48654,48667,48718,48768,48850,48880,48883,48931,49018,49020,49041,49138,49245,49276,49334,49397,49436,49722,49736,49766,49807,49816,49838,49848,49929,49941,50006,50028,50061,50159,50357,50378,50445,50449,50457,50533,50585,50599,50813,50853,50900,51034,51048,51079,51233,51418,51424,51535,51786,51872,52073,52171,52200,52295,52574,52623,52671,52722,52749,52821,53088,53090,53096,53149,53554,53599,53785,53795,53802,53830,53964,53972,54004,54028,54078,54199,54314,54429,54472,54577,54890,54892,55066,55232,55263,55414,55447,55670,55701,55858,55864,55882,56065,56150,56198,56215,56261,56287,56310,56315,56371,56394,56444,56474,56475,56543,56674,56775,56817,56881,56966,57091,57139,57149,57169,57190,57192,57211,57277,57309,57329,57537,57859,57992,57994,58254,58400,58423,58457,58657,58720,58742,58808,58896,59059,59270,59286,59325,59348,59461,59553,59599,59669,59794,59897,60048,60403,60440,60595,60612,60641,60656,60672,60733,60835,61136,61207,61226,61395,61510,61640,61738,61776,61893,61896,61898,62255,62406,62426,62641,62792,62834,62993,63056,63091,63127,63224,63306,63449,63599,63924,64082,64109,64112,64135,64152,64375,64498,64511,64547,64610,64629,64816,64906,64962,65046,65059,65201,65413,65435,65638,65741,65764,65937,66037,66042,66127,66357,66365,66434,66463,66596,66677,66699,66827,66896,67222,67239,67423,67607,67746,67889,67934,68063,68118,68151,68273,68442,68503,68564,68605,68697,69084,69205,69207,69210,69279,69345,69382,69457,69466,69477,69524,69741,69743,69881,69997,70065,70140,70175,70366,70397,70405,70417,70556,70565,70573,70695,70704,70786,70886,70950,70975,71185,71191,71373,71523,71643,71749,71750,71942,72076,72180,72327,72354,72366,72537,72546,72616,73051,73140,73316,73645,73793,73806,73833,73985,74203,74338,74377,74418,74440,74519,74544,74556,74606,74607,74807,74840,74851,75060,75101,75226,75298,75312,75355,75376,75398,75664,75694,75838,75904,75947,75962,76037,76130,76515,76530,76575,76582,76613,76813,76876,76961,77236,77284,77373,77475,77602,77768,77787,77893,78120,78288,78356,78422,78435,78462,78467,78538,78545,78669,78816,78843,78877,79025,79030,79041,79161,79181,79200,79467,79484,79707,79726,79734,79776,79779,79958,80121,80222,80365,80368,80485,80528,80544,80562,80613,80633,80742,80878,80900,80951,80966,80987,81261,81353,81435,81540,81616,81706,81723,82435,82497,82510,82511,82567,82671,82865,82899,82908,83008,83032,83041,83272,83343,83356,83582,83586,83763,83993,84061,84062,84190,84208,84346,84594,84786,84827,84869,84967,85091,85332,85439,85632,85689,85698,85809,85845,85869,85948,86020,86059,86122,86288,86414,86475,86644,86830,86833,86929,86933,86941,87006,87285,87447,87611,87621,87633,87703,87724,87844,88090,88223,88244,88259,88293,88341,88523,88598,88658,88669,88683,88689,88926,89071,89072,89189,89279,89281,89283,89286,89294,89306,89410,89426,89427,89451,89697,89731,89761,89766,89895,89925,89954,90043,90237,90263,90295,90309,90315,90324,90400,90431,90689,90742,91212,91218,91240,91282,91312,91365,91484,91638,91733,91751,91781,91806,91875,91904,92058,92097,92330,92356,92703,92722,92771,92788,92923,92995,93067,93119,93482,93773,93860,93976,94250,94346,94407,94608,94651,94702,94875,94879,94909,94928,95128,95161,95173,95303,95489,95573,96004,96078,96129,96182,96387,96662,96745,96820,96895,97000,97222,97309,97424,97456,97507,97587,97590,97594,97712,97808,97819,97825,97943,97954,98020,98073,98085,98121,98154,98200,98297,98418,98496,98525,98845,99029,99058,99171,99194,99219,99223,99238,99259,99521,99524,99566,99679,99895,99930,99978",
          "26,184,208,511,612,633,661,713,759,761,800,906,1037,1093,1144,1196,1197,1235,1292,1294,1303,1321,1424,1513,1520,1625,1648,1680,1683,1802,1864,2145,2182,2185,2377,2439,2490,2816,2868,2955,3272,3278,3291,3342,3424,3502,3789,3903,3927,3938,4105,4157,4213,4220,4261,4309,4412,4583,4762,4771,4831,4862,4886,4895,4997,5080,5143,5164,5297,5506,5542,5723,5955,6028,6036,6062,6140,6144,6353,6429,6447,6507,6531,6539,6609,6706,6785,6887,6894,7001,7133,7338,7419,7676,7890,7960,7998,8069,8173,8190,8423,8480,8507,8509,8634,8776,8790,8877,9091,9142,9229,9323,9369,9398,9437,9446,9450,9472,9487,9499,9505,9510,9514,9524,9643,9671,9823,10023,10095,10229,10377,10384,10453,10614,10709,10722,10725,10789,10816,10873,10925,10944,11011,11021,11189,11435,11555,11581,11607,11672,11812,11857,11887,11923,11943,12077,12101,12353,12413,12421,12505,12540,12706,12723,12765,12911,12932,13186,13315,13391,13427,13636,13644,13863,14319,14331,14445,14473,14494,14546,14552,14777,14778,14835,15044,15102,15267,15471,15506,15619,15801,15814,15918,15938,15967,15970,15998,16130,16149,16234,16261,16348,16556,16736,16757,16846,16985,17080,17140,17141,17244,17299,17428,17441,17571,17810,17817,17900,17949,18049,18145,18203,18229,18321,18371,18452,18464,18602,18622,18801,18806,18810,18892,18969,19093,19297,19314,19345,19651,19776,19803,19982,20025,20333,20398,20436,20444,20491,20531,20668,20851,20910,21075,21228,21232,21280,21287,21308,21363,21409,21672,21698,21737,21774,21822,21843,21852,21900,22118,22159,22198,22341,22520,22603,22762,22792,22863,22871,22912,22932,22943,22980,22985,23015,23063,23096,23145,23209,23210,23222,23307,23505,23525,23542,23596,23693,23718,23732,23765,23769,23800,23826,23873,23914,23932,23960,24092,24101,24148,24177,24251,24466,24694,24785,24823,24831,25111,25149,25184,25385,25471,25534,25554,25783,25805,25856,25866,25981,25987,26012,26014,26098,26141,26206,26247,26308,26568,26632,26722,26847,26906,26956,27015,27051,27127,27182,27183,27220,27382,27387,27457,27460,27486,27519,27611,27728,27931,27934,27937,27993,28028,28049,28058,28078,28179,28239,28324,28346,28399,28406,28450,28514,28601,28622,28668,28793,28828,28965,28996,29137,29180,29211,29325,29457,29562,29602,29709,29917,29963,30156,30252,30438,30475,30509,30516,30534,30677,30733,30757,30818,30856,30865,31182,31209,31233,31317,31421,31545,31826,31832,31889,32186,32395,32467,32523,32897,32927,32985,32998,33223,33412,33510,33550,33565,33744,33773,33776,33910,34110,34249,34256,34317,34376,34401,34479,34517,34537,34752,34824,34871,34892,34947,35175,35209,35281,35424,35548,35625,35732,35775,35796,36049,36199,36243,36303,36312,36322,36349,36407,36409,36453,36594,36652,36684,36731,36744,36761,36776,36939,36996,37009,37012,37145,37189,37345,37385,37411,37585,37668,37930,37931,37942,38006,38158,38339,38341,38361,38411,38530,38538,38553,38642,39069,39092,39203,39426,39459,39466,39582,39622,39841,39892,39899,40015,40061,40072,40121,40124,40213,40233,40447,40532,40635,40754,40840,40929,41057,41176,41264,41299,41324,41523,41594,41653,41741,41794,41964,42036,42048,42137,42350,42405,42590,42591,42842,42881,42979,43053,43161,43184,43242,43414,43463,43515,43543,43596,43653,43761,43806,44126,44265,44283,44427,44455,44545,44716,44745,44763,44816,45208,45331,45346,45406,45559,45753,45937,45973,46026,46090,46147,46162,46236,46266,46421,46435,46714,46727,46770,46823,46938,46949,47062,47111,47182,47275,47335,47368,47459,47500,47502,47532,47533,47568,47775,48015,48067,48077,48468,48478,48536,48546,48610,48688,48692,48708,48840,48847,48855,48875,48915,48917,48991,49000,49165,49188,49426,49501,49502,49551,49657,49764,49798,50069,50201,50299,50351,50386,50432,50527,50557,50576,50588,50712,50844,50856,50961,51015,51047,51152,51265,51278,51330,51433,51462,51596,51638,51802,51810,51822,51987,52042,52058,52133,52187,52500,52512,52591,52655,52703,52841,52915,53008,53014,53053,53059,53231,53448,53541,53551,53583,53736,53788,53814,53981,54030,54062,54074,54132,54423,54530,54549,54572,54633,54747,54818,54828,54837,54883,54901,54931,55002,55004,55048,55117,55208,55229,55413,55458,55610,55704,55740,55787,55788,55842,55978,56003,56071,56091,56384,56419,56498,56545,56738,56843,56860,56977,57275,57373,57456,57586,57726,57770,57774,57785,57800,57881,57883,57993,58090,58105,58795,58906,59044,59057,59284,59295,59298,59489,59569,59655,59706,59925,60038,60102,60105,60373,60396,60427,60579,60688,60754,60865,60915,60963,60971,60972,61040,61166,61188,61411,61537,61568,61594,61641,61697,61963,62041,62206,62229,62474,62583,62728,62879,62988,63016,63044,63110,63223,63251,63358,63403,63482,63555,63625,63707,63708,63985,64482,64530,64595,64669,64796,64879,64883,64981,65042,65101,65147,65280,65294,65359,65391,65496,65502,65525,65571,65593,65732,65882,65964,66087,66117,66149,66151,66329,66398,66480,66515,66716,67048,67166,67205,67224,67253,67267,67334,67436,67728,67756,67866,67886,68010,68027,68055,68108,68114,68220,68356,68360,68480,68483,68707,68821,68895,68942,69117,69416,69430,69462,69467,69582,69622,69667,69704,69706,69982,70025,70035,70066,70227,70386,70488,70503,70527,70712,70741,70857,71401,71464,71493,71494,71511,71700,71711,71773,71827,71845,71887,71945,72152,72257,72283,72669,72678,72698,72823,72848,72964,73088,73124,73633,73771,73798,73820,73868,73897,73941,74012,74110,74112,74118,74197,74260,74690,74707,74796,74814,75065,75095,75228,75231,75498,75526,75555,75568,75670,76355,76392,76650,76676,76823,76869,76885,77095,77404,77433,77440,77601,77608,77719,77727,77732,77855,77978,78005,78147,78206,78212,78322,78328,78378,78515,78686,78714,78841,78874,78904,78992,79063,79092,79132,79140,79447,79480,79611,79721,79793,79799,79843,79931,80112,80283,80332,80416,80638,80873,80938,80994,81082,81148,81211,81255,81741,81973,81982,82036,82137,82171,82369,82423,82571,82590,82738,82962,83001,83063,83085,83186,83233,83307,83347,83372,83498,83553,83566,83742,83775,83854,83889,84180,84269,84408,84551,84562,84582,84624,84719,84892,84895,84901,84923,85008,85064,85161,85193,85209,85212,85253,85280,85412,85464,85478,85531,85555,85620,85680,85842,85871,85924,86033,86034,86165,86220,86317,86382,86385,86501,86527,86536,86539,86580,86611,86655,86771,86816,86920,86989,87185,87198,87349,87359,87509,87614,87749,87758,88035,88133,88157,88373,88385,88464,88509,88521,88635,88671,88791,88812,89025,89311,89574,89637,89730,89874,89907,90126,90197,90462,90540,90774,90854,90881,91006,91063,91156,91224,91381,91483,91531,91561,91665,91913,91934,91943,91973,91998,92078,92202,92240,92305,92371,92435,92547,92587,92724,92826,92905,92908,92919,93175,93393,93403,93461,93517,93579,93679,93704,93717,93746,93791,93936,93938,93969,94007,94072,94078,94168,94212,94240,94288,94417,94524,94604,94778,94807,94808,94942,94993,95376,95660,95773,95784,95791,95833,95837,95875,96047,96225,96230,96245,96350,96542,96585,96591,96853,96889,96891,96901,96970,97056,97059,97090,97124,97266,97287,97429,97523,97748,97807,97828,97924,97988,98037,98038,98095,98383,98388,98408,98602,98621,98750,98761,98769,98811,98894,98945,99043,99055,99116,99187,99342,99400,99570,99579,99700,99977,99981,99991",
          "80,121,192,303,349,351,617,623,627,731,870,880,935,950,969,971,981,1018,1085,1278,1287,1323,1388,1409,1458,1617,1624,1645,1697,1700,1745,1786,1814,1838,1868,2039,2174,2258,2526,2801,2822,2880,2962,3136,3197,3219,3227,3260,3324,3411,3673,3829,3860,3867,3918,4034,4045,4064,4115,4205,4340,4343,4351,4484,4493,4568,4615,4666,4812,4906,4983,5020,5052,5070,5141,5151,5258,5299,5311,5409,5714,5913,5962,5989,6078,6103,6241,6243,6513,6614,6617,6651,6791,6806,6834,7030,7046,7180,7353,7356,7452,7548,7667,7677,7708,7729,7829,7863,7891,7931,8260,8275,8428,8497,8511,8524,8553,8612,8736,8813,8819,8846,8871,8924,9009,9133,9330,9608,9667,9670,9785,9837,9878,10046,10073,10116,10153,10262,10273,10277,10349,10370,10404,10502,10617,10733,10734,11111,11122,11192,11368,11374,11652,11683,11767,11897,11910,12058,12148,12158,12229,12273,12323,12369,12711,12780,12956,13001,13056,13067,13154,13285,13488,13622,13665,13761,13822,13839,13898,13909,14003,14128,14172,14394,14406,14416,14441,14459,14491,14520,14591,14697,14796,14818,14863,14896,14982,14990,15025,15081,15138,15290,15295,15428,15468,15488,15695,15715,15743,15828,15887,15912,15916,16146,16163,16245,16276,16290,16343,16364,16422,16483,16517,16641,16785,16797,16800,16824,16873,16980,17012,17099,17181,17203,17379,17386,17403,17422,17472,17496,17524,17544,17554,17583,17669,17737,17942,17987,18034,18344,18374,18436,18450,18533,18612,18649,18772,18775,18915,18949,18959,18981,19166,19258,19315,19348,19359,19551,19634,19645,19865,19932,19992,20030,20052,20076,20318,20554,20568,20638,20657,20662,20939,20943,21034,21078,21139,21151,21343,21446,21453,21458,21569,21694,21796,21871,21908,21913,21986,21999,22073,22546,22562,22595,22606,22672,22715,22754,22822,22832,23214,23253,23297,23298,23462,23473,23552,23583,23595,23686,23690,23758,23764,23872,23956,23968,24041,24053,24060,24188,24209,24281,24294,24313,24398,24421,24422,24437,24464,24553,24618,24672,24725,24733,25259,25382,25555,25685,25890,25909,25933,25965,25991,26011,26059,26115,26256,26304,26319,26335,26402,26489,26496,26517,26572,26669,26710,26727,26815,26848,26913,26972,26973,27077,27082,27101,27165,27394,27436,27493,27640,27654,27663,27746,27779,27877,27919,27939,27941,27978,28039,28216,28248,28326,28348,28382,28401,28427,28453,28479,28484,28503,28662,28723,28725,28729,28773,28811,28830,29038,29205,29249,29354,29532,29533,29538,29545,29622,29664,29972,29996,30050,30153,30186,30200,30205,30207,30258,30313,30456,30494,30542,30604,30619,30724,30817,30995,31029,31089,31158,31162,31217,31263,31314,31414,31468,31477,31478,31623,31753,31759,31851,31854,31995,32002,32111,32167,32209,32330,32347,32505,32540,32555,32636,32705,32708,32751,32871,32934,32988,33180,33189,33222,33309,33429,33475,33666,33879,33891,33930,34076,34083,34143,34177,34281,34534,34547,34894,35086,35087,35097,35128,35135,35207,35260,35349,35613,35622,35712,35734,35757,35883,35959,36014,36029,36052,36124,36233,36846,36862,37046,37150,37190,37249,37320,37425,37480,37555,37582,37588,37699,37789,37924,37943,38005,38201,38256,38622,38840,39013,39036,39109,39163,39174,39333,39357,39404,39428,39498,39512,39563,39823,39832,39934,39991,40029,40046,40100,40118,40125,40162,40257,40360,40361,40428,40430,40446,40573,40586,40695,40716,40817,40854,40896,40906,40955,40980,41005,41018,41114,41355,41374,41414,41508,41577,41710,41822,41897,41930,41944,41994,42033,42101,42298,42312,42386,42409,42445,42476,42511,42553,42598,42666,42746,42791,42975,43026,43206,43363,43370,43397,43627,43688,43724,43827,44022,44060,44158,44408,44411,44448,44500,44589,44627,44645,44675,44726,44733,44767,44795,44798,44857,44860,44932,45401,45431,45488,45505,45702,45779,45807,45818,46020,46038,46067,46095,46109,46140,46161,46165,46485,46640,46722,46908,46912,47055,47179,47193,47233,47282,47444,47701,47721,47887,47912,47919,47996,48259,48275,48287,48345,48416,48447,48514,48519,48521,48616,48723,48790,48826,48829,48844,48865,48963,49270,49282,49356,49366,49430,49470,49656,49869,49884,49963,49973,49976,50024,50176,50198,50232,50284,50292,50297,50318,50382,50492,50554,50568,50600,50609,50647,50811,50839,50863,50949,50965,51052,51053,51081,51181,51189,51245,51253,51302,51396,51425,51449,51464,51467,51603,51717,51736,51977,51994,52207,52224,52226,52234,52631,52635,52688,52751,52855,52902,52928,52932,52996,53124,53126,53188,53296,53334,53345,53374,53474,53627,53719,53731,53778,53887,53894,53914,53929,54042,54046,54047,54096,54102,54172,54185,54204,54465,54678,54907,54999,55267,55461,55552,55554,55569,55629,55662,55709,55828,55884,55959,55989,56006,56029,56058,56076,56144,56208,56210,56243,56397,56443,56540,56665,56741,56778,57031,57176,57357,57366,57464,57656,57658,57690,57705,57899,57916,57937,57939,57979,58083,58159,58162,58185,58606,58755,59120,59198,59226,59360,59510,59535,59537,59621,59689,59803,59850,59876,59911,59934,59954,60089,60122,60149,60214,60260,60360,60361,60474,60506,60580,60709,60731,60770,60800,60997,61010,61013,61140,61225,61279,61398,61501,61508,61518,61689,61699,61741,61808,61849,61860,61878,61921,61983,62093,62151,62211,62420,62538,62584,62614,62663,62704,62793,63203,63214,63275,63280,63345,63368,63400,63468,63505,63639,63910,63969,64018,64066,64107,64201,64208,64220,64430,64520,64566,64770,64834,64860,64994,65146,65466,65612,65725,65738,65771,65847,65880,65913,65953,65970,66119,66220,66260,66279,66338,66713,66718,66788,66799,66911,66934,66936,66994,67008,67015,67029,67104,67121,67151,67293,67381,67385,67386,67405,67455,67468,67487,67528,67531,67549,67725,67784,68123,68224,68262,68287,68325,68326,68339,68451,68485,68525,68648,68718,68790,68819,68820,68851,69013,69047,69051,69177,69215,69220,69271,69395,69452,69479,69497,69609,69712,69718,69776,69791,69813,69865,69876,70002,70024,70092,70099,70165,70171,70199,70211,70339,70389,70448,70723,70863,70900,70960,71016,71166,71201,71257,71273,71393,71451,71613,71638,71647,71672,71690,71806,71817,71820,71940,72007,72010,72055,72159,72202,72211,72263,72308,72376,72388,72444,72540,72644,72731,72736,72737,72759,72782,72858,72862,72905,72936,73046,73108,73115,73180,73482,73618,73622,73694,73867,73974,74057,74120,74344,74481,74645,74881,75038,75235,75295,75369,75413,75415,75474,75750,75780,75790,75957,76067,76099,76152,76173,76183,76358,76518,76547,76562,76605,76775,76984,76999,77082,77113,77159,77298,77322,77335,77376,77424,77456,77519,77538,77706,77721,77830,77902,77947,78032,78069,78118,78145,78330,78475,78478,78519,78628,78660,78681,78794,78797,78825,78838,78881,78922,79024,79027,79070,79226,79237,79277,79294,79376,79471,79487,79672,79676,79705,79758,79772,79828,79829,79872,79907,80143,80172,80180,80264,80730,80731,80789,80790,80945,81054,81123,81136,81273,81328,81346,81432,81481,81669,82016,82023,82035,82103,82147,82204,82208,82260,82402,82420,82432,82448,82466,82553,82602,82646,82722,82823,82827,82876,83117,83148,83295,83310,83388,83393,83483,83539,83562,83666,83696,83720,83789,83792,83851,83866,83899,83953,84005,84025,84068,84506,84606,84639,84797,84800,84920,84928,85295,85320,85424,85446,85450,85466,85500,85538,85682,85702,85786,85835,85900,85904,85985,85994,86049,86109,86155,86201,86316,86479,86553,86623,86718,86777,86828,86880,86900,86926,87101,87175,87188,87199,87277,87521,87550,87578,87603,87664,87831,87833,88275,88313,88381,88409,88418,88552,88583,88590,88642,88681,88781,88865,88887,88916,88990,89090,89099,89200,89227,89229,89258,89441,89490,89579,89614,89617,89705,89808,89880,89943,89977,89978,90082,90084,90132,90135,90151,90203,90292,90350,90366,90376,90418,90420,90466,90580,90587,90607,90625,90832,90912,90955,91045,91110,91183,91199,91220,91284,91588,91802,91869,91882,91949,91967,92020,92115,92123,92197,92480,92481,92498,92631,92693,92755,92807,92812,92838,92965,93014,93170,93171,93275,93311,93336,93365,93480,93487,93561,93664,93785,93807,93881,93953,93958,93981,93989,94079,94200,94476,94485,94574,94612,94665,94855,94873,94929,94940,94985,95061,95168,95312,95349,95470,95556,95597,95613,95793,95853,95856,95864,95909,96011,96061,96068,96082,96090,96136,96189,96271,96406,96584,96698,96708,96812,96892,96896,96913,97017,97067,97069,97079,97095,97127,97185,97326,97341,97390,97426,97440,97506,97512,97524,97599,97606,97619,97622,97642,97718,97785,97798,97929,97959,98270,98273,98277,98336,98432,98719,98819,98820,98860,98875,99013,99025,99099,99125,99191,99258,99401,99416,99530,99564,99753,99757,99785,99836,99886,99969",
          "10,188,241,248,256,380,433,569,621,768,801,878,888,903,912,940,948,960,991,1189,1209,1213,1221,1253,1282,1445,1461,1543,1571,1577,1579,1586,1609,1722,1780,1905,1940,1967,1997,2087,2120,2263,2267,2272,2323,2387,2493,2747,2790,2937,2957,3005,3023,3047,3062,3215,3258,3292,3413,3457,3701,3845,3991,4000,4085,4151,4152,4165,4232,4310,4478,4541,4842,4953,5004,5060,5254,5529,5586,5697,5719,5731,5741,5793,5918,6026,6185,6209,6311,6356,6444,6454,6464,6558,6578,6671,6744,6801,6868,6912,6931,6989,7101,7533,7537,7590,7611,7706,7752,7797,7850,8068,8359,8492,8529,8874,8895,8912,8919,8955,8974,9364,9570,9681,9778,9855,9974,10102,10113,10128,10245,10269,10520,10568,10610,10643,10924,10945,10973,10992,11495,11570,11636,11689,11829,11839,11927,12011,12038,12063,12129,12174,12431,12510,12777,12792,12819,13183,13198,13295,13341,13358,13362,13402,13465,13581,13594,13611,13859,13891,13972,14124,14199,14327,14357,14383,14466,14489,14610,14658,14719,14815,14849,14963,15084,15162,15239,15312,15315,15559,15566,15579,15735,15881,16014,16072,16217,16272,16313,16317,16319,16354,16559,16885,16964,16994,17015,17111,17253,17261,17300,17320,17491,17501,17715,17775,17820,17935,17959,17978,18012,18258,18289,18388,18389,18478,18754,18823,18929,18932,19029,19048,19073,19131,19179,19299,19334,19518,19526,19629,19712,19727,19744,19822,19840,19928,19938,20081,20097,20150,20154,20223,20448,20476,20509,20525,20878,21046,21218,21426,21460,21468,21493,21495,21497,21902,22121,22256,22425,22478,22517,22684,22793,22833,22835,22918,23117,23157,23294,23404,23457,23464,23626,23714,23751,23753,23754,23782,23793,23986,24046,24098,24122,24214,24270,24310,24326,24361,24377,24617,24675,24684,24695,24699,24728,24813,25113,25130,25275,25380,25487,25489,25602,25639,25670,25698,25705,25713,25777,25833,26246,26297,26331,26421,26919,27068,27072,27081,27094,27237,27303,27307,27314,27495,27617,27835,27887,27924,28150,28308,28343,28373,28609,28669,28753,28878,28902,28962,29050,29338,29409,29426,29497,29633,29681,29716,29809,29873,30015,30237,30248,30302,30355,30442,30445,30488,30525,30628,30671,30740,30779,30829,30847,30896,30898,30992,31031,31143,31208,31213,31739,31767,31898,31952,31983,32117,32127,32427,32442,32502,32688,32766,32889,32979,33114,33163,33306,33307,33352,33423,33503,33704,33741,33794,33807,33995,33997,34005,34237,34270,34271,34326,34346,34450,34594,34599,34681,34724,34751,34817,34845,34908,34948,35029,35136,35151,35210,35221,35554,35617,35708,35737,35842,35897,35952,35994,35995,35996,36262,36435,36473,36771,36785,36822,36896,36932,36986,37018,37076,37080,37163,37235,37242,37270,37326,37343,37348,37459,37472,37506,37520,37641,37711,37860,37865,37926,38030,38092,38217,38235,38264,38389,38490,38514,38563,38689,38796,38928,39021,39034,39066,39067,39082,39098,39127,39172,39292,39490,39558,39577,39627,39691,39728,39889,39893,39921,40119,40141,40308,40317,40340,40472,40549,40560,40563,40665,40751,40852,40882,41151,41172,41237,41354,41390,41713,41729,41744,41910,41912,41926,42112,42318,42381,42406,42425,42446,42477,42606,42726,42839,43017,43137,43219,43220,43245,43482,43551,43643,43693,43945,44102,44119,44305,44307,44343,44346,44393,44450,44463,44524,44582,44667,44721,44804,44892,44949,44960,45211,45271,45314,45378,45624,45628,45668,45729,45951,45959,45996,46265,46267,46387,46469,46482,46486,46573,46606,46619,46716,46734,46758,46873,46924,46954,46974,47010,47025,47060,47129,47175,47217,47241,47281,47796,48024,48036,48042,48063,48317,48412,48455,48467,48623,48685,48858,48978,48980,48996,49043,49066,49299,49306,49376,49463,49468,49557,49625,49663,49880,49904,49972,50065,50084,50088,50116,50118,50229,50411,50448,50482,50497,50534,50580,50598,50620,50731,50751,50795,50948,51037,51102,51125,51180,51262,51263,51307,51344,51374,51405,51578,51660,51720,51814,51823,51883,51939,51980,51988,52075,52153,52275,52312,52486,52633,52684,52811,52903,53030,53112,53186,53316,53379,53549,53709,53881,53997,54001,54066,54076,54138,54155,54211,54301,54373,54456,54490,54731,54750,54860,54977,55039,55068,55251,55310,55384,55445,55485,55515,55558,55678,55741,55744,55826,55871,55880,56039,56056,56078,56089,56222,56242,56296,56541,56589,56620,56707,56936,57061,57132,57202,57203,57242,57401,57443,57465,57471,57574,57601,57620,57729,57756,57791,58197,58252,58391,58537,58540,58562,58584,58730,58835,58959,59009,59129,59268,59340,59370,59818,59940,59975,60057,60129,60134,60137,60221,60225,60385,60406,60457,60473,60515,60658,60922,61003,61044,61257,61283,61332,61394,61423,61424,61463,61530,61547,61792,61912,62252,62466,62491,62550,62554,62629,62688,62814,62897,63051,63100,63342,63349,63399,63438,63524,63531,63582,63816,63849,63995,64104,64170,64371,64406,64439,64578,64624,64625,64628,64698,64852,64900,65243,65251,65401,65588,65656,65711,66111,66155,66186,66349,66411,66467,66504,66728,66842,66898,67069,67084,67332,67628,67899,67994,67997,68004,68219,68233,68296,68302,68357,68390,68489,68571,69002,69008,69099,69114,69121,69126,69130,69154,69187,69250,69414,69435,69470,69506,69520,69661,69840,69921,70112,70254,70267,70280,70295,70506,70772,70963,70974,71101,71153,71288,71376,71473,71627,71814,71874,71888,71984,72031,72034,72075,72242,72269,72404,72433,72581,72619,72809,72860,72951,73097,73105,73131,73158,73306,73330,73536,73729,73736,73773,73794,73882,73955,73993,74125,74263,74378,74470,74788,74800,74866,75145,75186,75208,75279,75567,75610,75714,75723,75736,75806,75858,75862,76002,76005,76035,76417,76595,76617,76640,76704,76727,76898,77001,77012,77035,77057,77219,77467,77728,77744,77788,77891,77898,77950,78058,78093,78426,78513,78527,78777,78779,78892,78932,78953,79119,79126,79158,79178,79192,79384,79385,79547,79636,79693,79702,80041,80067,80069,80082,80093,80394,80418,80444,80511,80538,80559,80620,80986,81029,81159,81188,81416,81518,81576,81606,81704,81816,81866,81959,82079,82133,82259,82385,82391,82496,82498,82595,82601,82649,82675,82710,82724,82728,82766,82794,82849,82852,82917,82935,82958,83053,83189,83293,83476,83486,83867,83915,83954,84018,84225,84248,84279,84467,84614,84700,84745,84848,84907,84955,85089,85097,85328,85472,85979,86018,86071,86083,86371,86437,86518,86631,86770,86790,86878,86910,87056,87081,87170,87219,87287,87300,87313,87326,87425,87485,87526,87753,87798,87813,87890,87998,88086,88097,88194,88206,88231,88255,88505,88536,88559,88584,88733,88823,88850,88903,88956,88997,89000,89216,89288,89323,89391,89500,89513,89555,89557,89561,89602,89608,89820,89900,89935,89969,89976,89992,90136,90222,90282,91054,91061,91075,91124,91238,91314,91417,91426,91461,91470,91521,91539,91706,91711,91926,91957,92007,92061,92165,92177,92186,92303,92306,92360,92364,92365,92636,92809,92994,93131,93180,93259,93494,93502,93527,93572,93833,93852,93862,94002,94051,94087,94160,94236,94295,94354,94356,94394,94453,94621,94741,94920,94926,95461,95493,95532,95560,95580,95874,95997,95999,96023,96211,96297,96399,96778,96784,96995,97209,97211,97296,97370,97435,97571,97588,97670,97688,97716,97769,97879,98420,98438,98456,98517,98541,98752,98869,98922,99204,99411,99440,99441,99519,99537,99543,99637,99705,99765,99840,99868,99918",
          "89,101,143,153,406,523,729,760,861,955,1033,1086,1104,1133,1211,1254,1466,1488,1670,1736,1737,1750,1875,1878,1957,1970,2051,2062,2114,2171,2197,2249,2299,2353,2465,2523,2678,2734,2892,2923,2968,2992,3109,3268,3270,3314,3315,3360,3679,3710,3718,3916,3959,4145,4346,4375,4435,4446,4553,4580,4685,4710,4725,4851,4965,4967,5046,5165,5206,5333,5339,5468,5564,5577,5617,5713,5882,5901,5904,6098,6239,6472,6515,6541,6573,6616,6748,7048,7179,7181,7230,7257,7310,7326,7329,7415,7599,7618,7649,7671,7694,7719,7746,7761,7832,7953,8021,8070,8333,8550,8764,8792,8845,8848,8882,9001,9081,9130,9214,9259,9296,9442,9517,9582,9594,9657,9713,9761,9861,9877,9966,9997,10063,10072,10094,10276,10295,10342,10447,10620,10681,10689,10741,10746,10807,11028,11059,11097,11109,11251,11326,11380,11482,11531,11619,11706,11753,11769,11780,11817,11850,11902,12036,12052,12083,12269,12303,12370,12428,12566,12834,12841,12877,13068,13283,13306,13339,13444,13551,13557,13561,13784,13791,13821,13950,13967,13999,14007,14127,14168,14249,14270,14322,14593,14682,14692,14771,14823,14952,14968,15175,15198,15314,15358,15562,15722,15788,15907,16011,16194,16328,16336,16342,16468,16485,16646,16708,16754,16791,16989,17083,17129,17171,17191,17229,17256,17325,17352,17462,17480,17550,17576,17851,17902,17966,18057,18060,18073,18108,18197,18447,18543,18618,18626,18668,18797,18866,18883,18897,19040,19114,19174,19193,19403,19734,19742,19772,19781,20245,20338,20390,20492,20591,20708,20804,20961,20992,21055,21099,21130,21150,21168,21476,21675,21717,21783,22012,22039,22088,22110,22166,22281,22339,22381,22393,22412,22441,22498,22607,22656,22893,22894,22999,23016,23227,23359,23484,23509,23566,23571,23700,23794,23839,23935,23995,24191,24316,24461,24469,24636,24664,24886,24913,24938,24987,25019,25022,25126,25155,25191,25260,25362,25368,25516,25594,25658,25848,25963,26002,26128,26145,26280,26283,26486,26584,26647,26665,26758,26881,26996,27168,27178,27209,27248,27376,27389,27466,27573,27578,27613,27627,27726,27743,27859,27868,27976,28051,28126,28217,28218,28299,28443,28449,28639,28666,28673,28704,28762,28912,28943,28969,29056,29167,29195,29221,29253,29465,29544,29712,29751,29752,29773,29776,29936,29987,30101,30114,30128,30137,30311,30352,30402,30404,30536,30573,30664,30714,30758,30769,30832,30868,30933,31048,31077,31219,31246,31341,31346,31355,31605,31672,31701,31704,31749,31765,31787,31824,31843,32251,32319,32375,32408,32428,32508,32510,32558,32570,32575,32632,32732,32788,32916,32963,33009,33178,33182,33186,33227,33264,33468,33489,33594,33625,33697,33861,33865,33934,33985,34023,34074,34174,34200,34264,34364,34367,34409,34466,34589,34634,34657,34731,34874,34980,34999,35130,35222,35237,35264,35359,35687,35822,36162,36169,36195,36210,36336,36343,36390,36417,36552,36589,36607,36621,36636,36646,36659,36668,36712,36730,36783,36987,37059,37160,37173,37178,37222,37260,37266,37274,37465,37491,37501,37516,37613,37645,37703,37750,37771,37773,37927,37971,37991,37998,38044,38107,38111,38134,38192,38270,38298,38313,38362,38458,38679,38784,39119,39266,39561,39750,39938,39948,39957,40466,40709,40760,40966,41038,41074,41169,41184,41199,41296,41459,41550,41571,41621,41644,42017,42125,42256,42303,42326,42340,42566,42587,42613,42625,42775,42805,42827,42901,42972,43011,43014,43147,43165,43207,43210,43302,43308,43328,43350,43358,43406,43413,43479,43619,43691,43711,43775,43796,43861,44093,44221,44263,44342,44464,44815,45007,45060,45110,45130,45146,45172,45305,45342,45467,45539,45613,45664,45701,45829,45927,45929,46450,46480,46517,46572,46824,46878,46942,47041,47070,47080,47091,47177,47202,47367,47510,47539,47551,47552,47623,47700,47852,47874,47899,48152,48164,48244,48246,48340,48398,48426,48493,48495,48509,48548,48661,48724,48769,48877,48936,49153,49164,49261,49307,49347,49377,49653,49695,49704,49742,49785,49845,50363,50516,50682,50733,50760,50884,50910,50955,51004,51010,51092,51159,51202,51284,51317,51341,51516,51610,51621,51768,51930,52009,52043,52127,52355,52441,52654,52714,52771,52786,52792,53017,53119,53138,53184,53243,53292,53420,53520,53698,53885,53984,54083,54179,54281,54413,54442,54560,54583,54614,54858,54893,55029,55226,55560,55568,55739,55820,55962,56070,56185,56246,56325,56377,56479,56518,56536,56607,56807,56842,56865,56951,57047,57138,57215,57425,57438,57457,57488,57610,57808,57831,57835,57905,58107,58163,58326,58486,58673,58789,58826,59030,59037,59245,59322,59341,59342,59446,59451,59455,59531,59607,59608,59640,59918,60193,60232,60258,60281,60369,60504,60759,60891,60995,61132,61149,61171,61708,61772,61818,61882,62138,62231,62233,62317,62479,62509,62668,62676,62736,62984,63078,63304,63326,63344,63366,63435,63641,63657,64405,64461,64484,64567,64573,64612,64657,64730,64744,64891,64999,65023,65070,65419,65498,65576,65607,65657,65779,65818,65891,65906,65934,66132,66153,66229,66334,66350,66401,66403,66422,66531,66545,66679,66732,66765,66961,67072,67139,67214,67266,67340,67433,67673,67708,67775,68062,68190,68209,68280,68349,68369,68426,68435,68486,68661,68672,68692,68880,68883,68901,68917,68926,68936,68998,69280,69486,69505,69658,69672,69854,69913,69977,70053,70074,70084,70353,70507,70521,70660,70719,70818,70826,70845,71076,71164,71180,71237,71280,71287,71289,71346,71436,71455,71599,71659,71772,71854,71866,71869,71909,71930,72011,72073,72188,72530,72543,72583,72631,72710,72766,72783,73035,73206,73455,73524,73681,73710,74005,74019,74049,74061,74064,74105,74245,74413,74425,74452,74522,74689,74748,74778,74829,74852,74890,75118,75132,75171,75217,75276,75386,75491,75517,75576,75733,75798,75848,75886,76120,76155,76244,76282,76326,76402,76523,76579,76675,76712,76809,76910,76916,76972,77003,77016,77046,77223,77303,77385,77513,77552,77666,77693,77844,77870,78013,78039,78064,78090,78135,78299,78543,78550,78730,78733,78741,78880,78889,79080,79107,79156,79485,79512,79548,79696,79810,79877,79953,80009,80010,80134,80187,80565,80642,80854,80864,80885,81057,81219,81232,81241,81263,81279,81310,81321,81350,81372,81454,81500,81506,81513,81648,81663,81700,81731,81771,81911,81951,81958,82013,82071,82306,82394,82490,82641,82737,82767,82843,82868,82882,82975,83009,83060,83149,83216,83235,83261,83349,83448,83502,83534,83575,83642,83743,83772,83862,83970,84186,84217,84243,84345,84435,84517,84585,84766,84856,85014,85145,85219,85325,85366,85431,85490,85502,85509,85526,85634,85673,85728,85843,85877,85920,85932,85950,86006,86028,86152,86191,86199,86210,86278,86454,86468,86512,86687,86796,86820,86882,87019,87109,87119,87213,87374,87403,87456,87460,87721,87862,87896,88043,88077,88079,88103,88156,88202,88204,88224,88283,88419,88450,88465,88497,88599,88636,88649,88707,88736,88830,88856,89138,89318,89413,89590,89623,89673,89780,89926,90023,90118,90300,90310,90359,90409,90441,90468,90545,90795,90893,91033,91051,91118,91203,91205,91248,91261,91270,91294,91319,91343,91513,91551,91591,91637,91681,91892,91924,92156,92339,92402,92440,92456,92654,92778,92910,92948,92952,93041,93179,93243,93260,93293,93369,93382,93408,93457,93569,93599,93609,93680,93826,93882,93944,94293,94336,94395,94508,94629,94638,94667,94768,94781,94787,94816,94893,94994,95048,95191,95200,95604,95759,95891,96112,96284,96295,96304,96331,96394,96427,96453,96667,96700,96811,96907,96908,97002,97050,97060,97136,97400,97406,97467,97499,97538,97595,97695,97741,97821,97920,97966,98195,98290,98400,98589,98598,98758,98906,99003,99019,99028,99068,99168,99346,99359,99369,99456,99630,99636,99738,99776,99826,99949,99972",
          "45,122,133,193,220,242,268,647,699,895,927,967,1055,1434,1526,1589,1601,1633,1636,1690,1717,1718,1830,1862,1902,1907,1928,1976,2002,2115,2210,2218,2275,2312,2409,2427,2565,2598,2607,2810,2884,2954,2960,3046,3059,3091,3153,3349,3378,3426,3481,3550,3668,3702,3736,3817,3861,3877,3905,3992,4036,4190,4291,4545,4850,4991,5414,5485,5494,5567,5674,5682,6027,6128,6175,6251,6330,6390,6394,6465,6510,6523,6536,6550,6641,6643,6650,6749,6772,6820,6855,6911,7063,7114,7122,7239,7262,7531,7703,7775,7805,7831,7991,8028,8071,8090,8116,8227,8231,8287,8391,8413,8460,8661,8915,8946,9017,9378,9396,9531,9738,9873,9907,9920,9969,10011,10075,10151,10209,10408,10419,10442,10523,10590,10602,10647,10708,10793,10812,10825,10846,11017,11132,11168,11194,11197,11236,11362,11492,11559,11629,11673,11833,12014,12021,12064,12146,12161,12163,12191,12254,12267,12364,12410,12426,12499,12521,12537,12549,12580,12619,12683,12753,12852,13051,13196,13205,13254,13365,13405,13496,13550,13569,13591,13642,13664,13804,13887,14025,14104,14188,14214,14312,14415,14461,14570,14661,14833,14840,14847,15011,15022,15105,15223,15227,15357,15399,15482,15489,15565,15618,15720,15755,15956,16019,16100,16145,16148,16152,16167,16220,16265,16346,16420,16447,16480,16640,16653,16716,16750,16767,16774,16811,16836,16895,16913,17035,17185,17210,17225,17280,17307,17318,17436,17500,17515,17520,17546,17560,17609,17678,17788,17866,17945,17973,18120,18257,18292,18323,18340,18394,18516,18564,18580,18646,18686,18689,18691,18698,18700,18712,18744,18804,19028,19147,19172,19263,19294,19352,19470,19506,19523,19532,19567,19741,19788,19793,19823,19836,19936,20012,20110,20132,20203,20206,20207,20273,20278,20313,20364,20368,20383,20558,20587,20597,20618,20637,20689,20756,20848,20955,21016,21031,21065,21138,21166,21259,21282,21462,21550,21627,21634,21635,21665,21725,21730,21800,21812,21901,21949,21961,22005,22051,22070,22100,22163,22226,22233,22241,22269,22274,22307,22352,22443,22596,22613,22645,22647,22657,22735,22843,23031,23053,23127,23176,23207,23212,23283,23528,23548,23581,23591,24040,24050,24131,24246,24337,24531,24533,24555,24749,24778,24799,24969,24986,25012,25053,25063,25072,25134,25157,25187,25188,25281,25406,25483,25486,25562,25587,25600,25689,25711,26021,26036,26039,26117,26149,26168,26196,26346,26361,26466,26523,26565,26655,26909,26932,26988,26991,27012,27098,27099,27134,27208,27329,27385,27439,27448,27484,27490,27506,27622,27631,27632,27647,27700,27738,27740,27875,27889,27945,28057,28091,28122,28153,28169,28245,28256,28265,28306,28313,28414,28550,28567,28589,28621,28624,28735,28808,29012,29095,29149,29187,29192,29212,29223,29290,29334,29357,29401,29412,29439,29464,29481,29511,29524,29550,29556,29658,29682,29826,29862,30005,30095,30117,30134,30143,30181,30190,30192,30298,30308,30332,30432,30487,30579,30607,30679,30744,30755,30775,30830,30851,31070,31148,31163,31200,31295,31340,31360,31408,31499,31595,31615,31638,31820,31838,31867,31880,31887,31901,32046,32061,32106,32281,32355,32440,32478,32547,32550,32595,32675,32843,32857,32993,33064,33076,33118,33194,33365,33439,33446,33463,33474,33553,33627,33646,33762,33767,33919,33924,33973,33991,34084,34453,34770,34796,34821,35192,35194,35241,35376,35489,35578,35645,35698,35748,35990,36030,36098,36334,36538,36800,36940,37002,37131,37285,37412,37538,37644,37650,37651,37676,37721,37729,37756,37838,37898,37902,37969,38000,38010,38032,38088,38112,38231,38373,38537,38610,38705,38779,39028,39038,39238,39262,39299,39337,39469,39571,39606,39705,39738,39763,39814,39819,39846,40021,40164,40279,40522,40612,40699,40706,40827,40863,40872,40936,40941,40953,41012,41046,41105,41190,41255,41344,41430,41470,41534,41572,41609,41668,41879,41946,42019,42077,42172,42243,42269,42333,42438,42460,42529,42535,42565,42654,42758,42811,42853,42855,42885,42926,43174,43177,43264,43382,43419,43505,43537,43576,43686,43928,43941,44324,44415,44416,44449,44502,44536,44572,44683,44934,44944,44990,44994,45012,45020,45066,45078,45089,45135,45173,45196,45198,45225,45292,45293,45301,45312,45694,45715,45735,45773,46065,46138,46146,46171,46211,46490,46587,46653,46748,46838,46891,47143,47155,47212,47229,47364,47394,47479,47607,47707,47906,47987,48052,48053,48082,48124,48178,48205,48274,48306,48325,48403,48448,48524,48584,48630,48683,48789,48890,49007,49025,49050,49113,49224,49260,49314,49498,49611,49668,50055,50058,50083,50187,50765,50806,50921,51064,51155,51162,51214,51242,51255,51282,51300,51329,51350,51373,51402,51426,51540,51542,51558,51583,51623,51851,51880,51919,51932,51961,52053,52060,52088,52120,52377,52416,52506,52517,52550,52638,52666,52805,52807,52849,52876,52882,52963,53025,53076,53113,53223,53224,53291,53371,53373,53414,53418,53503,53648,53660,53675,53694,53759,53872,53920,54121,54169,54381,54449,54499,54518,54543,54686,54870,54877,54928,55015,55104,55245,55255,55278,55280,55305,55425,55453,55717,55761,55774,55861,56019,56045,56047,56292,56300,56514,56529,56538,56572,56664,56677,56781,56813,56945,56961,57017,57098,57213,57246,57423,57475,57490,57529,57759,57760,58078,58082,58126,58128,58143,58175,58212,58216,58222,58241,58330,58441,58459,58609,58668,58921,58966,58969,58988,59011,59128,59149,59157,59176,59224,59228,59498,59500,59702,59772,59781,59833,59865,59899,59910,59932,59955,59969,59984,60063,60099,60195,60230,60256,60341,60370,60431,60437,60441,60505,60526,60643,60653,60752,61055,61300,61364,61919,61998,62001,62254,62341,62343,62354,62402,62659,62661,62699,62845,62858,63047,63268,63311,63318,63374,63418,63470,63567,63572,63608,63621,63668,63803,63934,64042,64101,64103,64105,64144,64240,64619,64850,64853,64882,64908,64942,64958,65065,65135,65139,65142,65195,65207,65275,65291,65403,65626,65757,65803,65816,65862,65871,65915,65985,65986,66006,66017,66103,66164,66277,66296,66346,66470,66559,66566,66691,66809,66856,66917,66960,66987,67031,67300,67498,67587,67592,67602,67614,67677,67908,67957,68119,68213,68288,68291,68393,68427,68504,68551,68714,68761,68763,68804,68856,68866,68889,68959,69076,69153,69164,69263,69389,69398,69474,69516,69584,69639,69660,69945,69990,70149,70187,70552,70569,70610,70805,70823,70912,70959,71022,71024,71073,71184,71202,71319,71381,71503,71525,71636,71651,71824,71831,71851,71921,72020,72277,72299,72340,72428,72493,72498,72528,72544,72691,73197,73244,73272,73387,73393,73537,73557,73583,73616,73635,73774,74034,74182,74185,74211,74450,74456,74508,74526,74706,74845,74980,74984,75158,75271,75309,75478,75606,75770,75822,75854,75985,76032,76041,76075,76208,76261,76567,76603,76649,76723,76760,76865,76926,76959,76965,77606,78067,78088,78094,78103,78230,78358,78474,78558,78671,78699,78804,78900,78963,78986,79060,79108,79346,79412,79572,79635,79678,79798,79823,80023,80176,80204,80267,80323,80429,80432,80506,80518,80558,80611,80791,80968,80972,80980,80993,81048,81277,81286,81323,81424,81567,81615,81691,81715,81800,81846,81879,81886,82098,82114,82168,82246,82396,82612,82655,82727,82820,82896,83026,83070,83126,83193,83378,83435,83481,83514,83569,83590,83591,83662,83886,84030,84053,84080,84096,84432,84462,84477,84613,84622,84921,85242,85312,85372,85376,85407,85411,85420,85427,85515,85524,85691,85828,85840,85945,85956,86026,86079,86182,86185,86268,86425,86477,86526,86529,86595,86609,86823,86874,86888,86899,86905,86998,87066,87227,87261,87478,87541,87571,87631,87817,87921,88123,88154,88324,88358,88404,88508,88594,88626,88727,88737,88840,88938,88940,88981,88995,88999,89160,89176,89740,89804,89831,89893,89911,89975,90217,90273,90408,90509,90760,90789,90828,90924,90957,91037,91080,91195,91271,91401,91448,91460,91488,91520,91525,91532,91686,91725,91902,91983,92080,92124,92129,92205,92222,92230,92254,92380,92420,92548,92639,92830,92850,92897,92997,93417,93611,93705,93780,93864,93975,94086,94159,94193,94227,94369,94422,94470,94551,94747,94901,94957,94996,95034,95087,95111,95210,95273,95424,95441,95483,95511,95551,95591,95765,95779,96000,96002,96058,96127,96268,96370,96405,96480,96560,96575,96612,96717,96748,96849,97022,97042,97108,97283,97607,97629,97654,97656,97661,97667,97750,97944,98071,98083,98153,98158,98169,98298,98404,98687,98700,98783,98791,98857,98902,98910,98924,98942,98974,98979,99014,99075,99151,99264,99356,99365,99388,99561,99604,99677,99804",
          "5,276,304,344,386,389,396,400,412,435,437,476,484,636,1245,1327,1386,5115,5171,5173,5186,5196,5232,5244,5246,5251,5313,5326,5386,5401,5408,5491,5513,5530,5568,5621,5695,5834,5869,5876,5899,5931,5992,6112,6177,6179,6183,6294,6329,6347,6387,6504,6517,6599,6764,6776,7071,7092,7165,7466,7592,7664,7678,7873,7916,8233,8383,8520,8851,8911,8948,9272,9942,9956,10059,10114,10210,10496,10562,10595,10775,11065,11124,11141,11217,11218,11349,11468,11659,11783,11821,11859,11986,12256,12347,12719,12885,12912,13176,14227,14410,15151,16308,16375,16473,16481,16572,16809,16963,17133,17489,18293,18912,19152,19225,20118,21645,23350,29471,38980,41067,41587,42609,44454,50186,50364,50459,50528,50881,50987,51206,51237,51299,51480,51562,51585,51645,51648,51962,52160,52219,52235,52255,52266,52278,52333,52339,52423,52515,52534,52535,52570,52647,52676,52746,52754,52866,52907,52939,53057,53125,53146,53359,53369,53560,53672,53820,53957,54002,54091,54219,54427,54482,54533,54539,54570,54819,54823,54891,54982,55063,55124,55152,55175,55356,55360,55396,55417,55491,55493,55532,55563,55565,55615,55667,55724,55845,55991,56031,56032,56126,56141,56149,56152,56274,56299,56553,56828,56845,56956,56995,57011,57082,57096,57152,57212,57232,57268,57274,57299,57330,57384,57392,57519,57595,57665,57686,57824,57908,57931,57961,58095,58214,58313,58358,58370,58411,58558,58695,58794,58876,58949,58991,59121,59363,59402,59519,59637,59638,59644,59665,59683,59774,59779,59821,59843,59991,60005,60244,60247,60249,60317,60349,60391,60589,60611,60617,60697,60699,60821,60833,60881,60885,61071,61120,61148,61190,61239,61604,61625,61815,61816,61907,61965,62010,62112,62131,62176,62236,62267,62304,62400,62599,62626,62653,62692,62766,62785,62913,63228,63255,63369,63455,63734,63825,63923,63940,64016,64037,64057,64098,64256,64260,64273,64408,64438,64617,64658,64759,65012,65425,65643,65752,65873,65973,65980,66016,66287,66367,66487,66560,66600,66616,66653,66655,66720,66742,66878,67017,67058,67143,67200,67241,67250,67287,67349,67414,67485,67525,67534,67557,67600,67640,67796,67836,68096,68110,68127,68168,68223,68317,68403,68615,68657,68715,68787,69053,69537,69641,69721,69735,69739,69781,69912,69929,70200,70414,70418,70470,70537,70570,70606,70841,70862,70896,70992,70998,71216,71229,71313,71435,71469,71520,71596,71752,71776,71828,71882,71931,71949,72028,72103,72149,72209,72256,72424,72521,72524,72595,72630,72828,73119,73298,73519,73523,73641,73782,73847,73912,74088,74103,74277,74464,74479,74531,74566,74585,74626,74681,74834,74954,75006,75075,75206,75324,75370,75399,75454,75479,76015,76115,76142,76224,76283,76382,76447,76451,76556,76638,76651,76659,76671,76805,76946,77004,77013,77051,77065,77087,77105,77128,77140,77161,77199,77347,77466,77634,77659,77690,77724,77778,77781,77936,78012,78036,78130,78192,78213,78232,78305,78410,78798,78893,78941,79014,79072,79221,79581,79834,79951,80150,80235,80290,80296,80322,80424,80632,80948,80965,81002,81062,81227,81235,81245,81329,81392,81524,81582,81626,81793,81858,81875,81876,81878,81952,82007,82087,82118,82139,82149,82194,82312,82392,82393,82407,82528,82632,82730,82742,82777,82991,83312,83322,83329,83438,83461,83471,83524,83702,83729,83858,83888,83913,84004,84066,84143,84270,84528,84564,84672,84758,84792,84858,84945,85036,85220,85334,85360,85398,85409,85603,85606,85657,85753,85793,85853,85885,86068,86117,86144,86215,86237,86259,86345,86431,86581,86675,86694,86758,86850,86865,86893,87052,87120,87210,87259,87329,87393,87434,87441,87461,87666,87726,87783,87837,87901,87955,87975,88049,88100,88213,88299,88318,88467,88495,88498,88715,88766,88866,89050,89215,89259,89373,89409,89543,89609,89828,89933,89990,90083,90296,90573,90575,90633,90684,90882,91029,91076,91163,91336,91715,91808,91969,92084,92179,92288,92386,92463,92482,92492,92582,92648,92717,92759,92779,92789,92799,92837,92847,92893,93112,93222,93258,93279,93508,93541,93629,93663,93689,93753,93945,93947,94102,94104,94133,94192,94246,94248,94315,94382,94399,94406,94505,94514,94683,94753,94811,94953,95054,95095,95144,95189,95325,95345,95391,95399,95490,95510,95607,95611,96102,96143,96168,96334,96366,96484,96524,96534,96549,96608,96627,96683,96854,97049,97397,97475,97554,97609,97890,98077,98377,98386,98463,98545,98772,99103,99133,99245,99252,99253,99262,99391,99491,99608,99621,99714,99731,99750,99790,99850,99863",
          "1,113,190,360,408,458,610,625,738,785,804,896,944,1015,1063,1193,1265,1387,1506,1549,1662,1664,1676,1713,2127,2167,2192,2340,2367,2395,2410,2454,2511,2630,2670,2673,2687,2695,2701,2728,2799,2976,3038,3117,3200,3697,4042,4143,4407,4471,4570,4664,4740,4788,4836,4922,4959,5008,5036,5204,5209,5360,5449,5516,5643,5733,5968,6015,6869,6974,7027,7106,7212,7237,7295,7318,7324,7577,7625,7683,7899,7946,8046,8300,8324,8354,8380,8381,8722,9114,9204,9348,9380,9387,9430,9495,9512,9549,9676,9745,9808,9826,9839,9857,9952,10006,10098,10137,10212,10313,10373,10407,10467,10660,10706,10742,10852,10906,10950,11013,11051,11154,11155,11164,11174,11230,11276,11352,11424,11465,11665,11699,11759,11760,11774,11919,12066,12120,12154,12263,12321,12328,12435,12541,12677,12837,12872,12898,13095,13098,13116,13239,13282,13348,13397,13494,13524,13572,13617,13792,13816,13910,13922,13976,14061,14238,14281,14368,14511,14655,14754,14812,14899,14973,15058,15099,15170,15252,15277,15329,15384,15388,15420,15545,15577,15691,15742,16136,16304,16394,16562,16787,16842,16848,16862,16875,16900,16929,16934,16992,17146,17176,17254,17372,17375,17432,17708,17777,17824,18225,18231,18232,18430,18513,18517,18536,18570,18684,18685,18825,18919,18999,19100,19161,19189,19234,19237,19306,19437,19443,19509,19578,19622,19624,19667,19763,19825,20020,20061,20105,20112,20140,20202,20209,20243,20268,20281,20403,20497,20524,20636,20672,20681,20701,20773,20784,20835,20856,21038,21142,21176,21365,21552,21751,21947,22029,22030,22148,22208,22334,22421,22442,22477,22508,22540,22668,22713,22782,22789,22846,23043,23052,23091,23093,23361,23454,23524,23604,23643,23855,23863,24062,24166,24174,24208,24217,24225,24307,24392,24452,24654,24655,24701,24789,24856,24926,24940,24946,24989,25036,25045,25146,25176,25222,25396,25405,25412,25431,25440,25465,25656,25748,25872,25910,26072,26109,26265,26349,26660,26678,26856,26958,26993,27193,27279,27447,27456,27546,27606,27619,27806,27807,27903,28029,28041,28161,28166,28186,28267,28284,28341,28465,28556,28820,28898,28964,28982,29046,29437,29444,29507,29627,29738,29739,29919,29957,30074,30191,30418,30510,30623,30693,30715,30807,30844,30970,31095,31129,31273,31397,31513,31588,31665,31980,32000,32030,32155,32261,32327,32473,32695,32721,32748,32771,32875,33143,33154,33344,33530,33617,33715,33742,33811,33927,34133,34220,34244,34312,34321,34350,34360,34378,34397,34460,34538,34669,34748,35052,35102,35477,35488,35584,35810,35905,35906,35920,35943,35946,35999,36019,36270,36309,36420,36508,36525,36557,36619,36759,36891,36936,37004,37280,37284,37344,37362,37427,37505,37530,37542,37580,37778,37900,37919,37959,37968,37985,38081,38257,38258,38272,38359,38401,38429,38439,38452,38749,38782,38814,38846,39035,39045,39102,39303,39505,39614,39659,39879,40237,40322,40486,40558,40677,40692,40750,40940,41050,41101,41251,41289,41304,41379,41380,41411,41505,41645,41671,41858,41917,41986,42095,42103,42171,42356,42397,42422,42455,42474,42478,42509,42617,42683,42732,42778,42785,42808,42956,43049,43097,43101,43301,43332,43368,43383,43603,43639,43640,43659,43718,43848,44036,44068,44174,44362,44381,44389,44581,44588,44665,44724,44772,44912,45093,45248,45379,45440,45463,45489,45518,45592,45671,45744,45844,45848,45922,46007,46014,46062,46139,46200,46412,46481,46527,46632,46659,46698,46749,46775,46881,46919,46935,47011,47028,47092,47132,47303,47319,47326,47402,47454,47614,47646,47764,47766,47790,47940,47960,48074,48130,48137,48162,48240,48277,48328,48374,48424,48457,48469,48484,48492,48549,48703,48819,48901,48955,49029,49076,49078,49149,49159,49172,49343,49769,49840,49857,49920,49940,49980,50240,50319,50442,50472,50561,50904,50983,51070,51139,51293,51386,51391,51437,51447,51658,51703,51815,51831,52131,52140,52267,52419,52615,52706,53072,53102,53209,53249,53280,53281,53299,53309,53403,53454,53490,53557,53575,53721,53807,53811,53832,53851,53884,53888,53995,54008,54016,54045,54168,54176,54214,54322,54360,54474,54609,54718,54761,54902,55097,55411,55481,55537,55580,55638,55690,55915,55932,56033,56191,56251,56402,56404,56432,56530,56638,56641,56740,56847,56877,56954,56976,57014,57016,57092,57236,57264,57296,57459,57478,57796,57837,58027,58097,58183,58315,58433,58436,58491,58495,58747,58796,58955,58975,59307,59357,59808,59844,60008,60351,60394,60554,60629,60684,60724,60727,60996,61028,61039,61068,61114,61146,61291,61428,61614,61645,61700,61732,61892,61945,62017,62080,62134,62160,62239,62388,62457,62510,62628,62669,62693,62819,62980,63074,63089,63163,63186,63222,63237,63242,63484,63548,63584,63733,63797,63988,64197,64224,64229,64297,64352,64400,64675,65240,65268,65277,65318,65348,65436,65453,65533,65536,65591,65631,65659,65836,65863,65867,66047,66074,66082,66121,66380,66385,66417,66512,66551,66953,66988,67103,67175,67424,67434,67503,67517,67772,67825,67890,67907,67925,67946,68029,68446,68497,68539,68561,68593,68795,68823,69001,69077,69092,69104,69132,69150,69224,69246,69247,69373,69526,69894,69943,69989,70090,70315,70380,70457,70542,70605,70620,70681,70696,70702,70743,70868,70880,70905,70969,70973,70996,71413,71449,71466,71512,71648,71649,71897,72105,72227,72364,72382,72441,72511,72632,72836,72879,72986,73027,73036,73166,73233,73289,73413,73433,73434,73454,73469,73687,73869,74030,74040,74095,74143,74215,74310,74433,74709,74723,74757,74774,74918,74995,75152,75167,75250,75350,75354,75360,75425,75439,75473,75536,75543,75807,76113,76175,76213,76500,76647,76657,76945,77073,77074,77121,77158,77249,77254,77421,77441,77729,77769,77957,78361,78633,78640,78703,78927,78981,79069,79100,79336,79449,79483,79573,79607,79660,79760,79770,79939,79979,80101,80205,80231,80401,80466,80520,80540,80567,80655,80691,80762,80768,80838,80942,81012,81067,81083,81100,81274,81333,81389,81462,81469,81860,81929,81950,82015,82333,82364,82421,82545,82665,82750,82763,82817,82861,82864,82898,82966,82969,83094,83107,83118,83201,83207,83262,83330,83390,83626,83663,83741,83805,83895,84122,84147,84221,84272,84305,84414,84473,84520,84701,84703,84824,84903,85051,85216,85290,85330,85461,85463,85668,85977,85992,86045,86080,86104,86206,86239,86295,86339,86420,86507,86567,86593,86672,87284,87304,87410,87510,87729,87775,87791,87856,87889,87984,87999,88143,88145,88183,88185,88250,88254,88379,88445,88752,88806,88867,89037,89063,89094,89114,89278,89298,89524,89589,89666,89772,89822,89825,90024,90076,90184,90267,90277,90396,90428,90443,90597,90652,90869,90883,91016,91073,91151,91166,91682,91698,91761,91771,91863,91907,91910,91939,92064,92073,92143,92304,92424,92468,92625,92928,92947,92975,93096,93449,93638,93719,94052,94132,94166,94237,94254,94375,94457,94465,94512,94537,94798,94934,95027,95070,95109,95211,95406,95468,95469,95568,95594,95700,96072,96131,96250,96252,96258,96285,96435,96517,96604,96694,96733,96871,96910,96948,97007,97073,97157,97260,97490,97503,97596,97668,97827,97835,97872,97938,98276,98292,98340,98389,98491,98580,98648,98650,98651,98676,98705,98709,98800,98886,99107,99161,99214,99227,99383,99718,99721,99756,99825,99877,99901,99907",
          "2,21,237,428,460,588,706,734,783,842,860,996,1045,1076,1223,1357,1443,1518,1797,1888,1895,2363,2416,2484,2487,2503,2583,2843,2893,2926,3002,3144,3186,3298,3302,3363,3433,3567,3594,4058,4142,4147,4238,4307,4327,4433,4502,4550,4703,4803,4885,4973,5025,5049,5144,5344,5388,5471,5502,5641,5680,5699,5889,6071,6274,6398,6423,6660,6661,6669,6786,6821,6960,7108,7269,7331,7524,7687,7892,7929,8043,8082,8207,8310,8332,8362,8429,8446,8502,8571,8576,8636,8666,8674,8928,9027,9135,9179,9195,9243,9253,9358,9371,9384,9388,9562,9625,9726,9757,9780,9909,9991,10086,10159,10213,10279,10312,10339,10393,10415,10516,10905,10986,11001,11120,11254,11335,11398,11453,11470,11554,11824,11889,12187,12188,12211,12270,12284,12405,12495,12603,12701,12745,12882,13002,13123,13261,13293,13323,13351,13429,13440,13520,13699,13729,13926,13934,13981,13991,14020,14213,14223,14241,14303,14365,14783,14790,15116,15193,15331,15396,15413,15464,15533,15543,15567,15646,15756,15759,15763,15800,15817,15830,15859,15919,16161,16173,16258,16384,16474,16510,16550,16583,16594,16608,16987,17123,17124,17263,17335,17389,17443,17456,18029,18038,18472,18530,18542,18716,18721,18725,18764,18840,18993,19212,19312,19365,19379,19390,19410,19428,19502,19503,19850,19904,19968,20011,20136,20198,20201,20263,20421,20454,20482,20569,20677,20694,21029,21035,21096,21240,21362,21370,21443,21450,21610,21633,21735,21745,21754,21797,21801,21870,21883,21893,21917,21955,22094,22173,22197,22416,22430,22479,22535,22573,22617,22652,22685,22802,22852,22879,22931,22948,23032,23038,23275,23550,23768,23833,24006,24054,24162,24196,24285,24356,24395,24451,24479,24550,24556,24682,25070,25128,25156,25207,25240,25359,25551,25736,25813,25918,25939,26095,26147,26167,26281,26396,26490,26623,26766,26855,27045,27149,27200,27391,27398,27449,27450,27536,27695,27715,27787,27885,27913,28071,28093,28162,28454,28588,28678,28721,28887,28940,28944,28993,29164,29199,29207,29317,29377,29473,29486,29487,29535,29553,29685,29947,29949,30010,30103,30348,30410,30596,30670,30681,30707,30782,30897,30913,30954,31169,31258,31292,31352,31469,31602,31708,31745,31748,31906,31931,31990,32064,32135,32151,32153,32184,32210,32211,32381,32383,32445,32962,33150,33250,33356,33389,33391,33472,33685,33711,33878,33883,33897,33901,33923,33964,34137,34176,34210,34266,34347,34382,34459,34754,34762,34773,34784,34851,34927,34995,35069,35335,35505,35536,35597,35667,35686,35852,35932,35939,36013,36054,36106,36121,36149,36361,36670,36683,36685,36754,36960,37079,37120,37153,37399,37432,37495,37564,37570,37793,37868,38021,38052,38124,38263,38528,38709,38848,38873,38948,38954,38967,38982,38989,39080,39145,39263,39279,39304,39400,39501,39504,39507,39737,39761,39768,39772,39790,39831,39838,39886,39976,39998,40135,40244,40348,40362,40371,40372,40476,40492,40592,40610,40660,40663,40684,40698,40862,40895,40945,40978,40981,41040,41263,41267,41401,41407,41532,41581,41690,41827,42076,42197,42212,42332,42354,42578,42616,42690,42779,42999,43131,43175,43205,43212,43240,43344,43380,43402,43421,43432,43484,43511,43547,43604,43632,43715,43939,44071,44079,44247,44268,44373,44420,44426,44475,44531,44594,44649,44735,44746,45084,45102,45272,45273,45315,45341,45348,45424,45578,45697,45786,45789,45855,45858,45859,46006,46132,46229,46343,46363,46449,46558,46584,46626,46703,46957,47017,47019,47024,47181,47296,47324,47407,47665,47760,47794,47813,47917,48115,48155,48213,48380,48532,48582,48605,48611,48653,48664,48747,48764,48767,48856,48940,48965,49040,49047,49089,49362,49558,49661,49670,49671,49765,49804,49822,49934,50009,50202,50205,50219,50255,50322,50327,50385,50565,50612,50836,51397,51450,51461,51561,51609,51626,51664,51894,51963,52080,52362,52398,52424,52552,52560,52790,53078,53211,53236,53375,53493,53531,53582,53657,53658,53768,53799,53848,53966,53977,54090,54147,54191,54224,54254,54477,54520,54587,54619,54653,54882,55009,55016,55088,55143,55228,55346,55366,55390,55551,55561,55647,55655,55691,55751,55854,55968,56022,56111,56197,56256,56273,56285,56317,56458,56464,56493,56520,56523,56605,56690,56739,56770,56830,56863,56889,56917,56920,57042,57154,57177,57252,57394,57397,57530,57713,57795,57805,57856,57923,57936,58041,58044,58115,58149,58182,58490,58612,58855,58992,58998,59038,59074,59143,59271,59490,59497,59504,59634,59851,59906,59909,60013,60066,60138,60175,60197,60229,60464,60475,60588,60680,60720,60740,60762,60794,60802,60862,60904,60925,61054,61108,61119,61139,61178,61308,61352,61399,61553,61564,61619,61725,61761,61994,62004,62094,62165,62194,62201,62260,62309,62334,62338,62679,62685,62731,62927,62961,63072,63105,63169,63212,63666,63721,63744,63869,64028,64034,64145,64361,64373,64389,64395,64594,64620,64861,65026,65057,65177,65179,65184,65199,65241,65528,65581,65625,65680,65883,65978,66013,66083,66134,66354,66373,66433,66507,66524,66590,66606,66853,66873,66885,66938,66997,67049,67066,67123,67134,67194,67198,67278,67410,67444,67522,67722,67745,67789,67852,67870,67981,68124,68139,68205,68327,68347,68413,68547,68566,68687,68835,68865,68963,69109,69218,69236,69256,69325,69580,69588,69636,69673,69680,69707,69764,70273,70361,70378,70381,70439,70484,70489,70624,70822,70888,70955,71162,71254,71396,71568,71598,71678,71781,71816,71835,71953,72106,72264,72442,72485,72615,72790,72937,73024,73061,73128,73134,73333,73563,73755,73849,73855,74060,74155,74209,74242,74300,74427,74487,74598,74619,74699,74781,74941,75023,75040,75045,75196,75434,75514,75561,75599,75940,75991,75993,76089,76127,76161,76227,76304,76441,76471,76608,76616,76714,76819,76924,77202,77221,77374,77428,77545,77639,77670,77839,77850,77934,77941,77945,78045,78112,78264,78315,78501,78826,78901,79386,79398,79497,79513,79539,79629,79716,79923,79964,80063,80089,80139,80147,80161,80177,80362,80740,80964,81023,81061,81140,81538,81547,81720,81727,81984,82003,82024,82047,82342,82425,82626,82793,82805,82855,82889,82913,83088,83135,83152,83456,83543,83601,83654,83692,83921,83987,84017,84145,84234,84369,84379,84413,84445,84548,84715,84950,84976,84991,85040,85105,85113,85189,85268,85399,85405,85497,85508,85669,85749,86110,86114,86125,86169,86318,86383,86619,86699,86704,86817,86829,86859,87026,87069,87207,87306,87375,87390,87402,87497,87728,87743,87908,88087,88214,88317,88327,88483,88624,88822,88857,88917,88933,89080,89081,89333,89515,89649,89706,89980,89986,90071,90471,90531,90553,90564,90662,90780,91143,91179,91210,91226,91299,91355,91485,91548,91692,91786,91860,91951,91997,92005,92052,92376,92416,92534,92556,92591,92604,92685,92687,92691,92725,92796,92823,92926,93028,93081,93085,93127,93141,93226,93323,93413,93533,93577,93815,93932,93991,94066,94142,94179,94196,94205,94225,94499,94500,94644,94758,94835,94886,95012,95036,95060,95071,95216,95256,95321,95368,95371,95404,95514,95935,95952,95953,96085,96124,96188,96243,96311,96583,96625,96648,96806,96878,97006,97204,97332,97345,97355,97363,97427,97536,97579,97659,97844,97892,97897,97984,98026,98079,98194,98233,98339,98480,98560,98601,98710,98796,99073,99282,99393,99434,99470,99822,99912,99940",
          "98,177,212,213,312,405,423,461,611,635,645,655,664,717,747,781,881,931,1001,1070,1146,1150,1325,1383,1389,1469,1534,1574,1880,1922,1927,1966,1980,1998,2152,2173,2188,2217,2220,2236,2245,2293,2731,2844,2920,3174,3210,3245,3385,3538,3637,3648,3717,3999,4059,4280,4504,4559,4620,4642,4647,4668,4673,4709,4738,4822,4845,4890,4916,5156,5160,5169,5285,5335,5359,5398,5422,5521,5550,5569,5612,5843,5911,5922,5997,6037,6063,6070,6111,6193,6195,6341,6424,6799,6809,7085,7093,7227,7535,7584,7626,7646,7716,7766,7825,7869,8151,8175,8176,8258,8303,8338,8349,8402,8475,8486,8534,8562,8570,8615,8672,8767,8803,8942,9030,9104,9310,9322,9383,9434,9451,9458,9585,9627,9908,9916,10077,10190,10204,10240,10376,10398,10703,10827,10931,10994,11000,11125,11163,11431,11562,11698,11703,11707,11729,11757,12322,12388,12471,12501,12564,12655,12873,13007,13083,13171,13192,13209,13260,13288,13485,13492,13505,13633,13725,13815,13851,14093,14109,14113,14532,14560,14673,14742,14956,15089,15197,15297,15371,15507,15521,15554,15798,15904,15946,15992,16186,16387,16454,16457,16460,16477,16524,16579,16588,16589,16664,16709,16861,16938,17053,17068,17073,17319,17369,17460,17575,17590,17647,17689,17712,17920,17924,18139,18250,18281,18290,18416,18614,18635,18641,18695,18858,18939,19055,19098,19150,19202,19228,19230,19544,19592,19619,19798,19882,20009,20157,20221,20265,20581,20601,20711,20770,20791,20812,20937,20984,21037,21064,21098,21199,21358,21438,21480,21517,21578,21669,21709,21795,21992,22234,22379,22402,22493,22571,22984,22991,23039,23083,23147,23170,23269,23394,23520,23573,23759,23866,23923,24021,24157,24244,24737,24806,24843,24857,24947,25023,25095,25182,25221,25466,25599,25680,25710,25862,25952,25992,25996,26105,26452,26605,26619,26651,26730,26783,27170,27277,27310,27344,27384,27423,27498,27582,27655,27681,27818,27821,27826,27860,27982,28042,28278,28684,28776,28900,28932,29065,29118,29127,29230,29246,29286,29355,29624,29750,29851,30163,30172,30310,30316,30514,30561,30617,30630,30717,30849,30889,30908,31040,31043,31216,31310,31351,31353,31412,31462,31593,31829,31879,31883,31902,31974,32021,32052,32054,32063,32094,32218,32321,32354,32380,32388,32455,32519,32551,32564,32601,32685,32852,32865,33087,33110,33324,33354,33382,33525,33537,33543,33568,33581,33600,33881,34204,34439,34454,34512,34557,34559,34887,34943,35200,35230,35240,35244,35305,35347,35450,35482,35504,35520,35616,35720,35730,35790,35912,35928,35961,36016,36158,36170,36214,36279,36342,36426,36527,36548,36600,36633,36637,36766,36905,37033,37277,37336,37437,37446,37519,37578,37814,37842,37862,38148,38290,38307,38372,38481,38521,38582,38809,38880,38892,39042,39204,39236,39264,39293,39562,39694,40005,40045,40069,40130,40316,40328,40409,40571,40604,40656,40701,40989,41111,41196,41418,41504,41612,41667,41681,41701,41796,41841,41857,41934,42083,42105,42170,42183,42187,42262,42263,42309,42534,42567,42687,42755,42833,42937,42974,43034,43361,43452,43518,43618,43712,43873,43964,44002,44155,44173,44276,44300,44414,44780,44808,44854,44917,44966,45024,45176,45199,45439,45717,45785,45799,45810,45867,45887,45957,45960,45961,46085,46129,46131,46158,46202,46260,46417,46489,46539,46544,46588,46620,46622,46702,46708,46740,46778,46790,46794,46980,47004,47153,47253,47276,47333,47342,47420,47625,47745,47830,47891,47921,48060,48151,48166,48197,48212,48303,48314,48320,48356,48513,48613,48635,48666,48911,48912,48960,49064,49106,49132,49167,49176,49178,49242,49368,49390,49407,49584,49630,49721,49886,49919,49926,49946,49959,50123,50313,50395,50547,50755,50886,50960,50973,50977,51372,51468,51551,51608,51616,51767,51813,51827,51920,52148,52156,52173,52493,52526,52779,52869,52970,53087,53103,53214,53241,53246,53287,53320,53510,53645,53649,53707,53710,53952,53962,54057,54082,54350,54362,54366,54382,54389,54411,54416,54467,54548,54635,54721,54749,54808,54822,54850,54918,54939,55023,55041,55194,55294,55497,55570,55865,55909,55945,55982,56037,56040,56046,56064,56073,56084,56087,56369,56400,56469,56861,56905,56985,56986,57027,57100,57201,57284,57285,57340,57341,57345,57348,57506,57702,57879,57896,57952,58331,58353,58475,58582,58666,58744,58822,58936,58977,59029,59132,59292,59377,59437,59456,59483,59581,59700,59733,59766,59799,59804,59903,59943,59998,60139,60166,60227,60228,60245,60304,60413,60465,60548,60659,60901,60916,60928,60973,61035,61041,61099,61177,61370,61431,61527,61597,61654,61766,61857,61953,61964,61979,62191,62243,62391,62453,62499,62551,62634,62635,62651,62724,62853,62941,63038,63116,63137,63161,63173,63188,63199,63425,63440,63633,63767,63976,64058,64076,64090,64154,64175,64245,64506,64545,64783,65024,65064,65272,65454,65734,65838,65969,66027,66097,66143,66167,66171,66611,66631,66813,66905,66912,66950,66968,67004,67174,67244,67391,67505,67672,67688,67835,67840,67856,67874,67927,67943,68169,68230,68247,68278,68312,68337,68452,68522,68782,68827,68862,68949,69141,69174,69446,69561,69626,69677,69750,69766,69870,70282,70297,70356,70376,70478,70483,70577,70615,70629,70738,70852,70899,70903,71293,71375,71454,71617,71840,71858,72037,72069,72165,72171,72175,72298,72312,72349,72409,72490,72608,72841,72983,73178,73254,73446,73491,73532,73555,73603,73642,73650,73684,73716,73784,73791,73930,73944,73977,73982,74101,74139,74164,74230,74346,74407,74480,74502,74529,74747,74760,74883,75097,75106,75146,75246,75311,75373,75395,75505,75560,75652,75672,75721,75744,75769,75789,75825,75995,76053,76153,76216,76290,76336,76342,76368,76468,76525,76604,76689,76906,76987,77032,77045,77076,77084,77365,77426,77628,77970,77981,77997,78127,78161,78211,78425,78434,78571,78692,78899,78925,79029,79047,79228,79284,79340,79379,79439,79620,79645,79740,79766,80268,80377,80848,80855,81015,81116,81496,81753,81902,81920,82012,82020,82041,82046,82054,82094,82288,82387,82399,82551,82771,82830,82862,82916,82939,83076,83320,83379,83518,83619,83652,83675,83781,84052,84100,84129,84260,84329,84449,84451,84509,84609,84735,84773,84784,84793,84916,85004,85106,85133,85142,85284,85343,85435,85468,85474,85486,85496,85685,85742,85751,85807,86048,86160,86186,86293,86469,86492,86639,86695,86735,86862,87012,87095,87104,87124,87191,87224,87252,87282,87494,87531,87649,87952,88012,88028,88080,88286,88475,88550,88564,88713,88824,88855,88908,88965,89027,89053,89101,89130,89220,89440,89541,89653,89805,89919,89999,90128,90140,90258,90311,90369,90417,90465,90505,90565,90654,90657,90698,90842,90991,91018,91020,91157,91288,91538,91602,91621,91625,91684,91712,91726,91841,91993,92002,92091,92192,92296,92322,92608,92626,92627,92684,92852,92865,92918,92924,93047,93238,93252,93321,93423,93463,93511,93557,93563,93582,93849,93866,93941,93994,94036,94062,94152,94161,94207,94209,94239,94292,94304,94492,94695,94709,94790,94840,94849,94852,94870,95057,95306,95355,95512,95807,95917,96029,96043,96251,96266,96290,96298,96306,96359,96368,96383,96621,96739,97053,97258,97584,97771,98015,98043,98055,98160,98182,98324,98367,98444,98795,98975,98978,99023,99059,99164,99275,99290,99299,99317,99405,99540,99887,99923",
          "51,233,469,571,584,634,680,722,726,873,897,1053,1060,1108,1147,1208,1301,1305,1341,1352,1544,1573,1595,1652,1751,1837,1887,1956,1992,2069,2163,2369,2429,2486,2603,2625,2634,2942,2946,2971,2985,3056,3074,3083,3084,3217,3223,3239,3442,3447,3453,3580,3623,3659,3662,3663,3705,3707,3874,3888,3901,4002,4055,4520,4622,4636,4641,4659,4675,4705,4752,5116,5213,5387,5487,5511,5515,5538,5539,5663,5775,5818,6055,6154,6155,6168,6277,6365,6436,6459,6576,6584,6631,6673,6778,6782,6793,6847,6849,6873,6876,7005,7038,7217,7264,7268,7313,7446,7581,7613,7718,7740,7754,7843,7859,7866,7883,7911,8106,8220,8230,8232,8274,8346,8368,8397,8401,8405,8504,8544,8611,8757,8972,8999,9003,9018,9144,9237,9263,9497,9516,9605,9806,9897,10002,10024,10249,10278,10444,10525,10536,10598,10758,10809,10814,10892,10977,11170,11244,11308,11313,11367,11403,11448,11452,11471,11526,11558,11651,11653,11813,11879,11957,12008,12019,12056,12216,12225,12277,12302,12307,12381,12519,12621,12694,12700,12710,12756,12856,12860,12875,12936,13104,13127,13195,13203,13395,13425,13486,13556,13682,13695,13764,13847,13907,14016,14163,14185,14198,14243,14263,14374,14586,14705,14721,14726,14782,14882,14936,15186,15251,15296,15343,15412,15526,15557,15644,15656,15748,15880,16022,16054,16061,16095,16150,16243,16257,16335,16406,16441,16496,16597,16599,16647,16729,16770,16850,16892,16898,17132,17272,17316,17343,17387,17468,17561,17601,17787,17853,17884,18042,18159,18185,18339,18500,18547,18761,18872,18964,19063,19155,19186,19218,19270,19296,19415,19559,19595,19659,19729,19771,19782,19940,19954,19987,20130,20167,20196,20330,20376,20377,20382,20394,20413,20496,20622,20811,21033,21072,21081,21083,21163,21165,21169,21244,21345,21352,21391,21491,21518,21530,21596,21690,21733,21792,21810,22022,22151,22152,22349,22365,22405,22423,22482,22610,22665,22748,22900,22917,23013,23079,23190,23352,23399,23512,23625,23684,24019,24027,24114,24189,24227,24261,24276,24372,24383,24442,24510,24600,24674,25029,25081,25121,25162,25346,25348,25421,25432,25477,25511,25579,25626,25844,25847,25959,26097,26119,26192,26204,26233,26388,26437,26465,26482,26579,26894,27060,27066,27123,27162,27163,27210,27233,27255,27323,27432,27451,27678,27888,28022,28095,28116,28120,28160,28246,28370,28402,28460,28528,28602,28605,28645,28649,28850,28927,28960,28984,29156,29161,29163,29319,29385,29423,29539,29632,29695,29775,29789,29823,29829,29922,29952,30142,30169,30195,30204,30443,30546,30657,30768,30819,30911,30914,30935,30957,30996,31126,31349,31496,31517,31524,31530,31576,31622,31685,31700,31711,31793,31819,31861,31977,32083,32132,32227,32298,32309,32374,32397,32411,32441,32471,32524,32585,32592,32712,32737,32783,32810,32868,32909,33033,33176,33326,33369,33387,33400,33509,33570,33611,33641,33677,33707,33826,33854,33880,33938,34009,34109,34114,34144,34184,34208,34289,34314,34365,34422,34526,34577,34617,34801,34828,34869,34889,35168,35179,35285,35399,35430,35517,35642,35722,35724,35902,35984,35992,36126,36242,36488,36512,36575,36601,36715,36799,36887,36993,37057,37119,37136,37297,37408,37563,37595,37724,37761,37767,37872,37914,37928,37953,38046,38066,38071,38084,38114,38210,38323,38345,38430,38473,38488,38597,38947,39040,39057,39183,39190,39207,39227,39291,39383,39393,39399,39446,39550,39572,39608,39647,39666,39669,39729,39755,39766,39778,39783,39884,39905,39952,39974,40058,40103,40183,40349,40408,40452,40503,40539,40588,40607,40626,40775,40812,40846,41013,41085,41129,41409,41445,41458,41495,41512,41610,41622,41657,41672,41805,41807,41814,41876,41905,42046,42079,42123,42211,42218,42227,42272,42339,42380,42408,42432,42472,42562,42845,42976,42983,43022,43062,43079,43213,43351,43399,43416,43424,43487,43608,43660,43678,43774,43778,43808,43892,44043,44047,44215,44440,44443,44512,44551,44605,44633,44673,44935,45031,45032,45059,45068,45091,45244,45278,45418,45493,45508,45527,45545,45560,45575,45651,45655,45714,45781,45788,45856,45988,46086,46111,46123,46175,46247,46288,46425,46557,46565,46592,46686,46783,46905,46906,46945,46989,47015,47029,47036,47079,47089,47239,47340,47426,47433,47515,47571,47612,47651,47687,47723,47867,47877,47951,47967,47977,48011,48026,48104,48108,48156,48189,48239,48293,48341,48358,48387,48390,48439,48511,48560,48608,48766,48803,48863,48939,48953,48982,49015,49039,49079,49196,49219,49222,49262,49330,49379,49392,49666,49837,50018,50140,50173,50323,50393,50541,50822,50924,51062,51199,51208,51268,51290,51401,51528,51716,51745,51833,51978,52010,52018,52050,52241,52287,52510,52548,52581,52604,52624,52628,52858,52969,52975,52991,53009,53207,53376,53458,53593,53874,54079,54089,54108,54174,54283,54426,54440,54510,54513,54620,54664,54671,54768,54770,54772,54820,55030,55216,55315,55323,55347,55535,55574,55627,55636,55665,55680,55771,55790,55927,56618,56825,56831,57020,57070,57081,57260,57306,57458,57692,57722,57940,58014,58047,58075,58224,58249,58334,58571,58698,58771,58810,58917,59034,59199,59242,59253,59259,59327,59332,59343,59349,59353,59359,59589,59718,59729,60023,60053,60174,60250,60252,60262,60330,60518,60669,60722,60761,60918,60977,60983,61118,61143,61202,61245,61263,61270,61388,61419,61430,61466,61575,61783,61786,61833,61843,61874,61887,61955,62035,62083,62180,62219,62448,62451,62487,62522,62533,62563,62587,62607,62627,62646,62660,62667,62738,62795,62841,63071,63181,63390,63450,63491,63501,63569,63587,63600,63630,63725,63756,63857,63947,63998,64314,64326,64372,64402,64524,64581,64615,64723,64753,64814,65079,65197,65225,65293,65423,65546,65684,65775,65888,66028,66057,66081,66135,66177,66280,66282,66308,66341,66369,66410,66580,66597,66680,66698,66747,67073,67089,67189,67282,67400,67409,67427,67507,67667,67702,67714,67743,67785,67798,67846,67868,67883,67916,68042,68084,68089,68199,68206,68354,68454,68473,68565,68706,69060,69113,69162,69171,69196,69299,69411,69481,69578,69612,69619,69631,69646,69785,69800,69804,69818,69862,69956,69961,70019,70022,70154,70210,70244,70311,70502,70640,70787,70820,70844,70846,70859,70944,71031,71041,71042,71080,71235,71272,71277,71286,71305,71307,71377,71456,71467,71625,71669,71785,71889,71898,71918,71941,71985,72080,72134,72184,72206,72344,72372,72400,72476,72643,72667,72687,72734,72813,72864,72934,72940,72996,73091,73136,73161,73225,73264,73280,73295,73299,73321,73356,73398,73466,73493,73606,73657,73708,73767,73843,73900,74153,74266,74284,74289,74305,74397,74428,74484,74511,74559,74570,74577,74777,74802,74820,74828,74864,74876,74948,75119,75236,75335,75512,75574,75638,75657,76084,76094,76124,76171,76212,76366,76404,76457,76501,76522,76527,76549,76561,76588,76612,76636,76655,76664,76708,76716,76804,76996,77165,77309,77324,77384,77429,77551,77572,77590,77594,77795,77854,77862,78113,78133,78197,78261,78333,78447,78482,78525,78708,78723,78769,78783,78786,78870,78884,78888,79173,79231,79261,79335,79353,79431,79446,79463,79535,79556,79610,79668,79670,79879,79915,80084,80236,80248,80327,80329,80395,80411,80449,80451,80513,80524,80659,80793,80813,80819,80857,80899,81007,81034,81170,81210,81260,81265,81400,81444,81477,81489,81645,81667,81732,81845,81895,81932,81943,82060,82072,82107,82165,82191,82347,82348,82371,82485,82494,82584,82594,82668,82669,82707,82754,82922,83116,83151,83215,83232,83279,83280,83296,83370,83397,83420,83533,83557,83676,83704,84094,84590,84676,84820,84914,84961,84975,85001,85010,85013,85057,85181,85249,85259,85285,85374,85568,85570,85666,85679,85720,85725,85954,86022,86090,86138,86158,86273,86343,86403,86409,86458,86574,86587,86605,86657,86709,86710,86951,87025,87087,87088,87195,87243,87256,87364,87379,87538,87949,87962,88142,88294,88334,88782,88868,89029,89143,89149,89255,89313,89334,89422,89512,89572,89611,89688,89693,89699,89774,89809,89833,89924,89955,90033,90086,90199,90204,90284,90340,90523,90706,90727,90796,90811,90901,90923,90974,91089,91169,91262,91351,91541,91589,91597,91630,91703,91759,91857,91972,92042,92125,92229,92245,92378,92458,92506,92528,92545,92563,92570,92602,92642,92646,92730,92777,92803,92814,92821,92889,93048,93107,93230,93379,93426,93473,93665,93745,93770,93775,93811,93859,93903,93950,93996,94268,94341,94447,94564,94582,94619,94637,94711,94770,94825,94862,94961,95119,95122,95160,95165,95184,95202,95223,95271,95358,95361,95584,95666,95702,95715,95816,95849,95861,95889,95914,95971,96176,96222,96264,96361,96551,96593,96596,96637,96654,96656,96676,96678,96887,96962,96979,97080,97203,97289,97392,97508,97725,97728,97863,97888,97895,97898,97902,98007,98008,98102,98204,98222,98311,98464,98467,98612,98669,98727,98788,98833,98887,98901,98920,98947,98948,98967,99006,99008,99047,99111,99136,99166,99193,99256,99322,99448,99478,99479,99523,99592,99620,99634,99650,99653,99659,99691,99720,99725,99798,99816,99873,99937",
          "147,175,330,384,414,416,434,440,472,473,817,867,875,1103,1232,1344,1363,1398,1477,1492,1630,1682,1686,2021,2054,2055,2109,2309,2334,2407,2408,2492,2571,2577,3025,3095,3193,3286,3294,3299,3335,3392,3451,3528,3537,3551,3695,3784,3790,3804,3930,4014,4067,4104,4124,4137,4241,4263,4414,4482,4561,4704,4742,4756,4848,4938,4992,4998,5015,5093,5193,5421,5748,6119,6253,6338,6441,6477,6559,6598,6906,7014,7045,7119,7245,7505,7573,7624,7639,7686,7691,7731,7807,7816,7853,8064,8212,8389,8477,8506,8551,8554,8561,8577,8699,8711,8812,8830,9067,9134,9215,9258,9400,9730,9733,9807,9940,10010,10016,10183,10266,10381,10518,10544,10597,10603,10711,10804,10864,10913,10923,10962,10971,11101,11151,11152,11323,11329,11384,11608,11630,11666,11694,11720,11825,11862,11958,11970,12017,12100,12189,12196,12416,12419,12427,12463,12483,12558,12782,12831,13031,13033,13336,13426,13450,13473,13558,13559,13564,13627,13661,13715,13843,13965,14130,14220,14254,14478,14479,14484,14574,14590,14626,14669,14683,14753,14813,14821,14965,15036,15043,15095,15097,15325,15410,15602,15764,15903,15929,15950,16020,16236,16630,16771,16776,16838,17092,17139,17255,17492,17494,17540,17733,17736,17847,18002,18009,18237,18384,18428,18446,18559,18576,18747,19121,19378,19441,19528,19556,19728,19735,19765,19834,19973,20032,20190,20430,20500,20503,20700,20735,20854,20997,21028,21052,21144,21178,21181,21236,21313,21605,21681,21693,21704,21708,21741,21743,21903,22083,22150,22291,22335,22392,22513,22522,22667,22855,22904,23065,23153,23154,23206,23215,23229,23495,23515,23647,24002,24003,24127,24231,24428,24594,24683,24717,24748,24811,25065,25093,25145,25224,25288,25459,25526,25589,25695,25716,25902,25931,26126,26288,26323,26333,26375,26460,26567,26627,26685,26718,27001,27059,27079,27485,27897,28268,28338,28435,28463,28475,28547,28626,28714,28821,28936,29021,29064,29086,29186,29231,29236,29456,29485,29490,29546,29722,29800,29965,30045,30053,30054,30099,30246,30294,30322,30334,30412,30472,30625,30710,30821,30834,30861,30980,31053,31185,31307,31449,31505,31581,31710,31963,32188,32257,32289,32314,32346,32399,32566,32603,32664,32681,32733,33020,33086,33095,33127,33152,33303,33327,33341,33445,33455,33480,33596,33638,33665,33809,33836,33857,34163,34263,34322,34679,34781,34816,34850,34878,34899,34964,35011,35043,35047,35080,35165,35259,35331,35394,35413,35589,35624,35731,35825,35971,36080,36235,36317,36352,36422,36500,36593,36595,36853,36933,36947,36954,37015,37047,37225,37262,37315,37539,37649,37758,37990,38097,38122,38419,38520,38598,38703,38725,38785,38843,38897,38926,38974,39288,39348,39352,39370,39448,39570,39585,39597,39623,39683,39714,39988,40210,40332,40388,40426,40491,40523,40596,40606,40683,40731,40740,40770,40911,40954,41037,41234,41266,41613,41722,41757,41895,42008,42010,42056,42266,42290,42492,42581,42593,42620,42632,42686,42761,42777,42804,42841,42902,42908,42916,43179,43521,43522,43633,43694,43819,43835,43866,43974,43986,44051,44138,44209,44260,44270,44301,44392,44424,44596,44600,44613,44663,44813,44834,44873,44925,45036,45122,45411,45415,45438,45480,45513,45515,45616,45625,45654,45659,45741,45854,46073,46135,46143,46295,46323,46326,46411,46477,46516,46576,46715,46802,46910,46953,46975,47031,47197,47437,47497,47535,47725,47738,47839,47892,47961,47988,48139,48231,48394,48678,48690,48722,48909,48945,48998,48999,49031,49077,49192,49230,49241,49247,49253,49286,49304,49382,49389,49395,49479,49487,49528,49530,49573,49706,49716,49789,49841,49871,49872,49928,49961,49999,50195,50230,50288,50310,50388,50405,50486,50550,50638,50655,50691,50754,50888,50944,50964,50999,51100,51141,51198,51313,51393,51406,51428,51659,51793,51931,51951,52047,52093,52161,52228,52239,52306,52453,52474,52540,52610,52691,52832,52839,52919,52945,53013,53203,53227,53344,53351,53390,53466,53473,53515,53596,53697,53845,53897,54056,54072,54316,54390,54406,54460,54670,54707,54759,54812,54998,55099,55108,55126,55179,55314,55330,55392,55728,55851,56082,56103,56104,56139,56268,56430,56445,56446,56501,56515,56892,57060,57088,57103,57107,57432,57463,57512,57571,57632,57783,57876,58034,58153,58287,58355,58374,58513,58630,58632,58805,58942,58986,59041,59118,59375,59376,59379,59436,59506,59596,59755,59760,59905,60093,60116,60184,60204,60210,60387,60565,60828,60867,60924,61249,61296,61367,61440,61483,61627,61660,61812,61869,61933,62386,62565,62670,62711,62750,62781,62973,62995,63009,63093,63201,63221,63890,63901,64096,64099,64166,64189,64288,64315,64336,64354,64357,64377,64416,64436,64447,64586,64638,64727,64910,64979,65011,65030,65209,65325,65339,65369,65468,65480,65620,65632,65644,65731,65733,65815,66157,66283,66384,66404,66447,66464,66473,66479,66546,66612,66626,66642,66808,67027,67116,67130,67212,67294,67478,67632,67736,67956,68016,68043,68075,68243,68256,68342,68370,68439,68476,68604,68642,68752,68767,69032,69034,69035,69135,69394,69410,69507,69575,69642,69659,69674,69679,69699,69814,69916,70008,70106,70114,70232,70304,70312,70544,70554,70582,70596,70601,70927,71027,71030,71169,71176,71197,71340,71361,71431,71594,71612,71832,71837,71901,71910,71962,72016,72030,72091,72259,72320,72398,72556,72591,72810,72915,73098,73138,73139,73189,73191,73223,73265,73346,73384,73437,73441,73486,73535,73632,73836,73872,73873,73931,73995,74041,74051,74176,74219,74256,74287,74308,74358,74465,74534,74547,74575,74641,74666,74717,74882,74913,74932,74991,75053,75064,75072,75089,75362,75470,75690,75726,75766,75773,75782,75852,75880,75966,76194,76297,76424,76443,76459,76514,76688,76812,76815,76821,77015,77044,77090,77116,77175,77382,77405,77461,77514,77624,77731,77733,77858,77888,77922,78029,78062,78122,78214,78325,78335,78353,78374,78444,78479,78585,78592,78656,78721,78722,78849,78937,78989,79088,79089,79197,79259,79278,79343,79473,79562,79782,79812,80014,80061,80107,80130,80218,80241,80388,80468,80477,80543,80616,80617,80677,80694,80709,80718,80735,80780,80826,80919,81216,81285,81358,81437,81595,81752,82120,82193,82278,82307,82334,82388,82441,82603,82610,82747,82812,82911,82982,82998,83030,83062,83086,83096,83109,83138,83163,83243,83321,83648,83865,83916,84086,84140,84163,84165,84218,84274,84275,84374,84423,84563,84572,84832,84853,85080,85160,85451,85522,85545,85572,85663,85672,85674,85855,85889,86031,86150,86340,86582,86751,86904,87018,87037,87050,87076,87100,87167,87239,87290,87305,87400,87406,87408,87438,87511,87653,87747,87769,87804,87902,87913,88004,88201,88336,88383,88434,88436,88553,88859,88968,89019,89105,89274,89310,89357,89554,89583,89601,89783,89788,89806,90041,90098,90113,90142,90314,90407,90467,90470,90478,90483,90582,90693,90793,90846,90904,91034,91088,91106,91237,91251,91298,91335,91400,91408,91514,91547,91606,91671,91749,91812,92036,92038,92043,92060,92067,92137,92244,92422,92462,92517,92679,92681,92882,93071,93075,93133,93151,93166,93174,93329,93459,93535,93545,93672,93798,93799,93823,93939,94039,94118,94128,94141,94204,94234,94294,94314,94414,94448,94560,94578,94646,94700,94786,94806,94899,95051,95078,95106,95153,95182,95214,95220,95375,95410,95571,95659,95670,95675,95842,95843,95869,95994,96367,96380,96452,96516,96668,96672,96725,96746,96790,96870,96953,96965,97092,97131,97305,97395,97511,97650,97705,97710,97719,97763,97764,97804,97864,98059,98100,98258,98265,98402,98425,98490,98502,98520,98603,98632,98665,98799,98921,98953,99265,99298,99344,99382,99432,99446,99459,99473,99692,99774,99827,99847,99894",
          "55,76,78,152,171,263,306,332,524,813,856,1402,1532,1578,1605,1741,1795,1801,1876,2050,2150,2183,2193,2234,2237,2405,2425,2444,2480,2507,2520,2604,2637,2737,2746,2797,2817,2974,3127,3205,3212,3246,3275,3283,3297,3350,3359,3401,3423,3470,3486,3577,3714,3751,3754,3777,3794,3809,3868,3872,4009,4113,4169,4189,4265,4328,4365,4411,4475,4513,4546,4688,4731,4777,4793,4802,4896,4980,5019,5064,5077,5129,5273,5331,5589,5704,5794,5872,6235,6518,6587,6625,6721,6852,6946,7028,7032,7244,7397,7420,7528,7557,7648,7659,7800,7973,7975,8035,8066,8192,8376,8377,8432,8471,8664,8689,8773,8909,8939,8954,9053,9075,9163,9239,9250,9568,9619,9662,9700,9796,9894,9958,10040,10108,10175,10330,10433,10539,10591,10593,10707,10720,10731,10886,10897,11289,11337,11370,11421,11567,11714,11852,11939,11942,11963,12060,12082,12173,12238,12285,12644,12679,12690,12726,12851,12889,12895,12917,13151,13158,13194,13232,13319,13369,13371,13373,13437,13458,13472,13621,13628,13660,13663,13747,13755,13892,13894,13912,14047,14052,14071,14186,14187,14276,14284,14306,14396,14424,14463,14639,14709,14727,14824,14928,15041,15047,15060,15211,15303,15377,15574,15590,15663,15699,15854,15885,15905,15947,15995,16046,16079,16120,16140,16282,16287,16383,16438,16446,16505,16551,16627,16642,16644,16666,16759,16943,16981,17104,17232,17344,17408,17527,17529,17547,17552,17555,17589,17592,17595,17660,17682,17686,17688,17726,17739,17792,17811,17865,17873,17885,17916,18061,18137,18154,18303,18325,18382,18383,18391,18507,18549,18730,18965,19005,19105,19127,19311,19331,19337,19398,19429,19448,19555,19580,19591,19643,19646,19665,19691,19743,19887,19900,19906,19979,20060,20138,20250,20335,20380,20424,20438,20494,20513,20546,20582,20594,20617,20623,20658,20748,20858,20886,20947,21007,21086,21157,21385,21414,21439,21454,21466,21538,21556,21565,21600,21782,21815,21926,21929,21933,21942,21953,21968,21989,22010,22178,22219,22316,22321,22327,22362,22436,22447,22515,22560,22570,22666,22704,22804,22868,22885,22890,22997,23040,23049,23160,23230,23376,23438,23439,23440,23448,23536,23565,23592,23593,23633,23651,23652,23719,23728,23746,23818,23854,23981,24017,24037,24093,24132,24187,24202,24219,24221,24338,24366,24381,24447,24535,24566,24576,24628,24653,24666,24698,24838,24867,24923,24927,24939,24991,25003,25049,25076,25116,25274,25283,25303,25331,25546,25625,25631,25708,25717,25743,25811,25825,25829,25846,25859,25868,25920,25924,25949,25988,25990,26064,26084,26137,26235,26238,26240,26262,26301,26324,26382,26426,26506,26563,26636,26689,26821,26923,26955,27029,27050,27083,27167,27203,27217,27258,27264,27326,27440,27471,27503,27529,27680,27720,27770,27799,27808,27960,27962,28031,28048,28070,28141,28145,28156,28198,28220,28258,28274,28362,28420,28473,28521,28558,28612,28617,28648,28691,28716,28756,28818,28824,28848,28883,28888,28973,29006,29105,29170,29194,29259,29275,29296,29379,29383,29419,29430,29495,29529,29576,29613,29635,29842,29875,29966,29981,30001,30046,30086,30106,30115,30118,30162,30180,30261,30264,30272,30317,30370,30371,30420,30454,30515,30539,30640,30871,30880,31033,31087,31109,31193,31486,31487,31509,31533,31543,31572,31738,31847,31855,31908,32077,32312,32361,32504,32589,32669,32939,33022,33025,33147,33247,33249,33257,33316,33367,33453,33559,33607,33620,33652,33696,33782,33947,33958,34108,34291,34366,34370,34407,34473,34687,34744,34812,34843,34852,34939,35055,35058,35081,35095,35121,35250,35326,35350,35352,35439,35467,35474,35487,35490,35515,35576,35675,35707,35729,35845,35903,36021,36070,36093,36134,36160,36267,36269,36314,36339,36404,36532,36535,36653,36697,36725,36734,36792,36873,36875,36919,36937,36948,36988,37007,37130,37213,37252,37253,37316,37331,37350,37376,37382,37386,37393,37407,37422,37528,37594,37605,37818,37822,37849,37867,37916,37960,37994,38108,38135,38144,38171,38175,38423,38426,38487,38632,38654,38736,38837,38888,38988,39010,39041,39159,39170,39202,39216,39235,39272,39276,39375,39416,39452,39458,39586,39652,39719,39725,39813,39829,39881,39908,39923,40037,40065,40067,40070,40087,40098,40131,40158,40159,40268,40325,40438,40440,40443,40481,40658,40659,40671,40685,40735,40864,40920,40944,40972,40992,40993,40994,41020,41048,41077,41094,41143,41157,41240,41257,41369,41474,41482,41560,41708,41714,41721,41753,41773,41826,41840,41844,41864,41869,41871,41874,42000,42004,42085,42140,42174,42279,42314,42383,42454,42551,42582,42700,42728,42764,42812,42819,42936,43016,43072,43088,43156,43190,43244,43318,43330,43360,43460,43495,43523,43539,43578,43611,43623,43670,43703,43821,43933,43956,43972,43999,44042,44059,44189,44195,44281,44385,44459,44479,44544,44580,44599,44689,44698,44774,44820,44863,44865,44882,44945,44964,45061,45112,45129,45145,45270,45419,45503,45574,45590,45598,45617,45691,45730,45769,45806,45835,45838,45860,45895,45962,46128,46130,46180,46186,46315,46457,46575,46672,46713,46723,46808,46857,47040,47300,47370,47387,47710,47859,48008,48010,48023,48138,48201,48215,48299,48339,48456,48462,48559,48693,48739,48891,48925,48926,49011,49049,49197,49297,49350,49391,49420,49518,49520,49588,49770,49974,50021,50099,50131,50607,50633,50715,50728,50772,50779,50874,50907,51259,51458,51554,51587,52023,52170,52211,52225,52289,52296,52495,52587,52675,52731,52785,52834,52838,52842,52870,52913,52935,52954,53001,53117,53133,53232,53250,53270,53283,53444,53505,53519,53553,53607,53625,53677,53994,53999,54246,54274,54377,54394,54493,54603,54630,54778,54791,54862,55052,55089,55111,55119,55161,55272,55282,55331,55339,55434,55527,55607,55637,55700,55737,55767,55931,56020,56276,56316,56457,56610,56647,56676,56679,56729,56764,56774,56812,56854,56900,56906,57035,57079,57238,57288,57361,57742,57807,57814,57909,57951,57989,57995,58002,58016,58084,58096,58106,58279,58311,58321,58342,58488,58548,58681,58803,58884,58895,58898,58932,59015,59202,59257,59386,59391,59444,59546,59642,59742,59753,59827,60049,60054,60171,60203,60254,60333,60492,60561,60841,60955,61025,61088,61094,61206,61436,61521,61601,61702,61749,62032,62054,62122,62270,62365,62512,62546,62547,62585,62709,62945,62954,63015,63055,63075,63098,63122,63171,63308,63331,63340,63348,63352,63445,63458,63632,63659,63712,63886,63981,64122,64137,64148,64151,64294,64300,64302,64321,64333,64334,64346,64558,64703,64735,64789,64799,64809,64866,65137,65185,65215,65350,65362,65678,65763,65833,65886,65901,65994,66040,66078,66361,66420,66673,66702,66774,66795,66834,66897,66928,66943,66957,67034,67042,67047,67265,67299,67354,67358,67403,67411,67442,67574,67579,67646,67686,67707,67726,67744,67855,67910,68321,68371,68398,68405,68612,68776,68792,68914,68922,69137,69144,69158,69180,69184,69298,69387,69489,69503,69546,69608,69758,69875,69879,69892,69934,69985,70118,70183,70224,70248,70287,70344,70362,70413,70455,70551,70576,70678,70804,70813,70979,71011,71081,71114,71143,71161,71173,71246,71324,71399,71626,71668,71707,71873,71927,72025,72043,72064,72135,72214,72296,72395,72582,72663,72713,72745,72846,72897,72959,73060,73086,73104,73234,73447,73733,73763,73801,73893,73906,73988,74037,74073,74311,74396,74408,74441,74528,74724,74726,74762,75054,75116,75131,75230,75243,75306,75525,75642,75738,75775,75809,75868,76176,76237,76247,76268,76311,76365,76386,76454,76461,76484,76495,76565,76615,76696,76791,76877,76902,76933,77137,77155,77290,77312,77340,77363,77364,77366,77458,77462,77701,77803,77868,77943,77986,78144,78392,78568,78586,78606,78627,78642,78755,78765,79043,79093,79102,79138,79186,79276,79387,79407,79418,79434,79515,79617,79619,79864,80042,80052,80103,80113,80276,80288,80294,80403,80428,80470,80471,80529,80533,80692,80810,81181,81379,81447,81505,81562,81764,81799,81827,81832,81900,81939,81966,81980,82019,82417,82449,82489,82513,82524,82538,82581,82653,82693,82711,83051,83108,83268,83408,83532,83700,83723,83755,83893,83908,83985,83996,84085,84257,84339,84354,84552,84638,84865,85023,85045,85175,85314,85633,85745,85759,85836,85844,86001,86003,86019,86038,86128,86148,86156,86219,86270,86298,86462,86464,86517,86599,86690,86757,86765,86784,86813,87014,87159,87162,87228,87254,87263,87437,87455,87473,87482,87546,87628,87865,87885,87895,87898,87904,87910,88105,88155,88241,88266,88270,88310,88367,88429,88440,88530,88819,88939,89166,89173,89420,89446,89481,89492,89547,89558,89638,89650,89652,89694,89958,90116,90214,90220,90332,90334,90335,90469,90614,90648,90726,90870,90936,90946,90976,91027,91058,91078,91185,91186,91257,91302,91326,91341,91348,91360,91469,91495,91509,91568,91653,91670,91850,91941,91954,92114,92209,92210,92220,92325,92412,92486,92614,92666,92688,92744,92840,92862,92921,92929,92976,93217,93244,93255,93291,93462,93469,93532,93566,93594,93600,93678,93708,93737,93817,93828,93830,93858,93956,93993,94031,94171,94229,94242,94291,94312,94320,94430,94474,94528,94640,94685,94767,94913,94919,95259,95456,95662,95707,95737,95776,95780,95808,95873,95991,96001,96103,96113,96122,96154,96155,96210,96301,96395,96421,96442,96498,96539,96548,96689,96728,96768,96890,96969,97064,97173,97224,97248,97265,97301,97401,97510,97567,97692,97697,97905,97919,97936,97975,97987,97997,98063,98138,98146,98184,98254,98314,98351,98497,98688,98736,98746,98827,98839,98918,99056,99061,99145,99146,99233,99251,99316,99360,99372,99572,99618,99655,99671,99761,99842,99884,99980",
          "218,277,285,538,581,596,606,624,643,691,737,807,845,893,1290,1495,1537,1542,1616,1632,1649,1760,1861,1904,1912,1960,1985,2026,2103,2187,2341,2364,2420,2437,2466,2514,2516,2569,2589,2643,2688,2749,2805,2815,2983,2989,3015,3021,3068,3072,3085,3147,3234,3266,3428,3460,3466,3467,3510,3554,3612,3684,3744,3849,3893,3894,3908,3925,3963,3964,4236,4276,4404,4429,4444,4517,4524,4548,4626,4661,4691,4733,4765,4813,4853,4871,4900,4903,4907,5467,5626,5638,5648,5656,5684,5739,5769,5783,5811,6039,6046,6053,6082,6145,6220,6322,6468,6597,6693,6756,6814,6937,7183,7289,7312,7501,7943,7944,7984,8029,8041,8111,8120,8145,8150,8408,8442,8483,8708,8944,9051,9123,9236,9313,9391,9504,9545,9564,9606,9638,9779,9866,10051,10147,10221,10274,10275,10307,10608,10623,10626,10659,10705,10756,10800,10822,10842,10943,10961,10993,11076,11187,11597,11668,11758,11917,11972,12104,12231,12320,12384,12626,12645,12697,12810,12825,12961,13069,13071,13184,13284,13441,13462,13604,13608,13616,13737,13762,13942,13947,14006,14051,14099,14157,14180,14252,14360,14480,14493,14578,14595,14894,15037,15090,15106,15137,15173,15174,15324,15367,15411,15518,15622,15660,15690,15752,15822,15860,15877,16045,16066,16101,16204,16207,16269,16371,16403,16497,16537,16694,16731,16744,16763,16782,16790,16858,16888,16918,16949,16969,17019,17031,17052,17067,17081,17131,17138,17206,17211,17373,17424,17482,17483,17616,17696,17722,17742,17814,17859,17879,17905,18041,18062,18270,18363,18381,18519,18621,18625,18628,18722,18727,18930,18988,19181,19216,19251,19264,19284,19288,19309,19459,19652,19757,19775,19779,19827,19893,19924,19935,20041,20166,20314,20328,20334,20355,20560,20585,20625,20649,20660,20718,20737,20779,20801,20818,20857,20887,20948,21045,21102,21112,21123,21243,21296,21303,21310,21319,21398,21401,21445,21477,21515,21532,21536,21548,21628,21670,21773,21790,21811,21904,21912,21980,22018,22243,22347,22348,22360,22388,22435,22446,22472,22480,22514,22545,22698,22727,22823,22889,22954,22961,22975,22992,23011,23034,23087,23133,23168,23171,23213,23284,23304,23316,23373,23410,23582,23610,23726,23867,23926,23946,23977,23989,24028,24112,24205,24232,24271,24483,24738,24817,24873,24891,24920,24924,25152,25179,25238,25263,25278,25322,25400,25504,25622,25745,25763,25875,25884,26018,26066,26074,26132,26199,26344,26345,26350,26404,26409,26445,26475,26481,26564,26585,26652,26679,26742,26748,26857,26915,27049,27105,27196,27247,27353,27443,27535,27651,27653,27730,27863,27908,27989,28090,28128,28137,28203,28400,28408,28423,28430,28491,28495,28496,28526,28563,28598,28711,28760,28838,28896,29015,29032,29068,29265,29282,29284,29310,29467,29496,29520,29586,29671,29679,29697,29816,29923,29960,29968,30096,30305,30366,30378,30397,30478,30598,30618,30661,30691,30729,30863,30931,31011,31072,31152,31153,31194,31206,31230,31271,31282,31327,31453,31457,31485,31555,31585,31591,31666,31678,31694,31712,31734,31741,31790,31809,31812,31859,31927,31954,32029,32095,32183,32237,32326,32329,32410,32422,32460,32461,32476,32490,32503,32514,32633,32659,32690,32730,32736,32739,32774,33016,33029,33042,33057,33145,33159,33174,33177,33198,33425,33428,33486,33491,33494,33504,33542,33566,33578,33681,33709,33844,33862,33951,33980,34018,34035,34106,34107,34113,34129,34134,34155,34193,34213,34282,34292,34315,34338,34431,34535,34558,34733,34753,34768,34822,34826,34832,34863,34864,34941,34990,34996,35014,35031,35045,35053,35169,35191,35280,35283,35303,35333,35361,35366,35371,35436,35445,35503,35604,35671,35696,35749,35804,35843,35850,35947,35978,35979,35987,36108,36216,36219,36260,36261,36298,36372,36387,36451,36546,36568,36571,36647,36708,36807,36821,36951,37039,37134,37193,37215,37219,37374,37377,37403,37410,37455,37514,37515,37710,37754,37759,37762,37885,37906,37933,37951,37955,37974,38003,38064,38085,38211,38317,38336,38352,38364,38376,38471,38476,38510,38561,38577,38626,38638,38723,38753,38803,38821,38889,38901,38973,39043,39051,39168,39194,39267,39316,39335,39365,39374,39462,39480,39488,39554,39626,39660,39662,39907,39932,39954,39977,40012,40050,40092,40101,40110,40120,40190,40201,40267,40272,40334,40335,40482,40501,40550,40602,40603,40627,40669,40670,40717,40738,40866,40986,41069,41093,41097,41197,41274,41330,41345,41420,41437,41456,41460,41640,41699,41709,41823,41849,41903,41909,41970,42027,42044,42117,42131,42132,42141,42196,42336,42388,42395,42452,42459,42545,42575,42600,42604,42612,42710,42731,42769,43140,43162,43198,43312,43331,43448,43513,43685,43707,43813,43822,43840,43910,43958,43998,44009,44176,44193,44210,44258,44273,44336,44348,44405,44566,44628,44634,44768,44784,44796,44805,44824,44825,44843,44982,45097,45229,45261,45307,45330,45374,45392,45410,45511,45567,45573,45581,45618,45676,45696,45708,45719,45732,45900,45971,45980,46183,46341,46344,46345,46395,46436,46531,46773,46917,47147,47148,47244,47245,47246,47427,47434,47458,47492,47517,47641,47682,47848,47871,47873,47890,48190,48265,48284,48405,48432,48470,48500,48503,48681,48694,48711,48736,48812,48838,48920,49056,49100,49103,49185,49231,49589,49600,49802,49814,49890,49931,50079,50091,50110,50207,50339,50412,50427,50499,50640,50775,50799,50834,50905,51046,51071,51172,51210,51246,51322,51364,51414,51434,51481,51577,51588,51650,51662,51678,51771,51775,51834,51836,51933,52026,52066,52198,52232,52233,52357,52386,52479,52577,52755,52778,52922,52938,52961,52977,52988,53003,53107,53215,53251,53255,53333,53561,53664,53693,53701,53923,54037,54157,54414,54526,54580,54622,54722,54756,54985,55011,55070,55138,55210,55343,55403,55511,55573,55749,55825,55873,55966,55987,56190,56312,56381,56392,56461,56508,56516,56522,56599,56663,56834,56879,57093,57170,57186,57224,57426,57448,57696,57715,57894,58089,58160,58219,58225,58327,58347,58493,58694,58704,58726,58731,58802,58823,58857,58858,59263,59383,59450,59482,59652,59959,59979,60037,60076,60115,60150,60310,60414,60462,60522,60547,60569,60594,60673,60705,61185,61457,61477,61485,61545,61588,61659,61834,61872,61886,61902,61908,61924,61947,61951,62029,62209,62268,62362,62444,62544,62548,62608,62701,62836,62870,62904,62992,63066,63082,63108,63414,63535,63605,63661,63774,63775,63812,63830,63847,63927,63993,64049,64165,64228,64559,64604,64626,64681,64699,64748,64876,64964,64982,65019,65089,65219,65278,65370,65471,65559,65814,65860,66192,66383,66429,66578,66588,66722,66727,66761,66815,66838,66886,66908,66977,67108,67233,67330,67372,67374,67474,67475,67609,67629,67729,67741,67858,67940,68174,68322,68374,68380,68386,68598,68670,68813,68855,68976,69110,69305,69308,69314,69353,69369,69515,69669,69737,69923,70137,70163,70194,70218,70322,70385,70441,70539,70602,70617,70668,70726,70735,70773,71066,71078,71186,71330,71444,71513,71571,71645,71656,71742,71770,71801,71809,71853,71935,71989,72161,72213,72221,72265,72559,72720,72814,72832,72847,72881,72923,72926,73013,73033,73137,73163,73310,73388,73484,73514,73544,73571,73573,73649,73676,73788,74070,74130,74154,74294,74320,74333,74334,74394,74565,74705,74806,74865,74899,74916,74981,75020,75047,75104,75144,75221,75308,75320,75602,75660,75732,75781,75871,76006,76076,76098,76108,76276,76364,76389,76397,76496,76733,76749,76816,76833,77025,77124,77146,77201,77283,77369,77377,77517,77537,77671,77730,77790,77847,77861,77938,77973,78061,78100,78244,78281,78431,78508,78551,78610,78649,78718,78736,78772,79036,79153,79247,79350,79408,79461,79470,79612,79701,79765,79824,79853,80007,80056,80117,80163,80324,80346,80435,80514,80539,80634,80732,80771,80778,80800,80801,80847,80883,80992,81026,81110,81160,81192,81207,81252,81386,81394,81472,81474,81570,81759,81783,81945,82005,82095,82123,82130,82134,82164,82283,82401,82575,82579,82633,82813,82873,83057,83168,83286,83383,83419,83647,83749,83759,83774,83815,83842,83880,83902,83960,83967,84023,84240,84293,84375,84489,84724,84910,84930,85186,85286,85355,85445,85621,85763,85927,85933,86056,86075,86178,86204,86304,86470,86601,86714,86971,87155,87274,87340,87424,87449,87490,87537,87556,87698,87756,87780,87935,87993,88020,88073,88104,88337,88360,88365,88422,88459,88482,88507,88639,88684,88970,89009,89022,89061,89075,89144,89168,89254,89287,89295,89296,89403,89444,89508,89606,89656,89717,89776,89819,89827,89873,89890,89944,89962,90026,90063,90169,90206,90212,90249,90550,90739,90958,90965,91132,91171,91188,91211,91225,91227,91266,91321,91353,91494,91542,91582,91755,91861,92101,92102,92208,92242,92295,92331,92375,92379,92455,92601,92633,92641,92670,92676,92699,92720,92763,92793,92863,92945,92967,93027,93058,93129,93140,93183,93316,93319,93354,93436,93467,93585,93778,93926,93940,94130,94150,94252,94278,94283,94347,94374,94429,94478,94487,94516,94630,94814,94841,94883,95016,95187,95314,95335,95454,95540,95549,95728,95813,95894,95967,96019,96151,96187,96352,96476,96487,96502,96535,96566,96664,96707,96724,96929,96954,97011,97034,97129,97148,97189,97254,97361,97555,97832,97904,97928,97973,98078,98202,98495,98504,98514,98578,98619,98662,98681,98726,98785,99022,99163,99272,99325,99378,99487,99503,99639,99799,99858,99919,99953",
          "71,86,172,226,272,457,559,668,871,939,1007,1043,1116,1308,1360,1399,1463,1550,1556,1711,1812,1891,1899,2073,2119,2129,2194,2321,2322,2343,2452,2572,2595,2602,2745,2806,2850,2919,2961,2967,2980,3041,3069,3100,3123,3163,3203,3296,3323,3398,3431,3452,3455,3462,3603,3615,3631,3644,3917,4053,4240,4539,4627,4722,4747,4755,4962,5033,5180,5210,5226,5404,5444,5463,5520,5640,5655,5658,5717,5718,5735,5757,5776,5795,5816,6061,6091,6138,6170,6181,6194,6282,6397,6403,6761,6831,6924,6966,7103,7169,7311,7317,7379,7391,7457,7483,7597,7733,7959,7964,8012,8023,8052,8063,8083,8127,8209,8214,8302,8384,8454,8656,8662,8670,8797,8899,8970,8978,9044,9061,9065,9079,9105,9113,9374,9938,9960,10020,10036,10092,10106,10142,10201,10248,10310,10331,10357,10401,10418,10499,10611,10641,10740,10765,10965,11079,11083,11190,11271,11293,11327,11354,11355,11458,11467,11505,11782,11811,11876,11890,11984,11997,12031,12112,12170,12183,12200,12304,12451,12460,12575,12668,12937,12951,13075,13089,13096,13329,13335,13399,13418,13446,13601,13641,13675,13721,13828,13873,13932,14037,14082,14121,14426,14628,14701,14740,14743,14784,14837,14900,14955,15070,15156,15165,15308,15374,15403,15493,15540,15797,15839,15841,15842,15855,15863,15865,15878,16024,16026,16040,16047,16065,16176,16230,16244,16388,16433,16434,16650,16724,16726,16802,16844,16945,17000,17086,17087,17155,17217,17239,17296,17337,17350,17645,17653,17700,17890,17944,18010,18200,18211,18222,18275,18304,18440,18575,18637,18645,18652,18667,19103,19386,19439,19501,19553,19664,19687,19713,20161,20349,20460,20519,20609,20924,20930,20935,21005,21022,21042,21088,21121,21122,21124,21267,21318,21408,21418,21487,21540,21590,21807,21818,21906,21940,21952,22008,22017,22202,22263,22506,22519,22532,22582,22691,22710,22764,22775,22834,22945,23059,23067,23121,23126,23251,23268,23282,23354,23445,23585,23600,23763,23774,23785,23916,23920,23921,24161,24298,24340,24353,24357,24358,24412,24609,24647,24794,24901,25032,25039,25080,25124,25250,25285,25374,25614,25776,25796,25830,25905,26031,26037,26056,26079,26139,26186,26282,26362,26378,26544,26549,26566,26674,26746,26747,26778,26827,26840,26981,26992,27005,27021,27032,27064,27117,27166,27169,27191,27232,27317,27331,27427,27487,27511,27595,27636,27756,27805,27878,27965,28084,28103,28112,28114,28206,28270,28456,28468,28614,28631,28759,28772,28790,28805,28812,29106,29499,29595,29676,29730,29735,29791,29815,30068,30097,30158,30197,30203,30241,30424,30505,30531,30606,30627,30636,30651,30852,30965,30982,31073,31078,31117,31221,31234,31243,31266,31322,31334,31347,31354,31376,31635,31693,31768,31792,31940,32016,32048,32090,32143,32163,32217,32334,32344,32421,32436,32612,32767,32906,32953,33113,33132,33134,33135,33226,33259,33270,33313,33347,33363,33484,33513,33515,33526,33529,33963,33992,34001,34149,34181,34187,34206,34241,34284,34295,34306,34357,34394,34435,34528,34532,34591,34696,34861,34987,35140,35163,35212,35227,35462,35525,35527,35530,35628,35710,35747,36178,36217,36292,36306,36360,36461,36530,36644,36719,36797,36802,36879,36958,36962,37019,37058,37151,37180,37290,37372,37421,37435,37440,37460,37550,37556,37845,38166,38396,38596,38630,38698,38730,38787,38791,38927,39259,39398,39403,39409,39484,39517,39519,39529,39530,39538,39560,39620,39631,39677,39690,39712,39759,39872,39909,39958,40010,40013,40022,40056,40082,40280,40379,40463,40484,40561,40734,40764,40803,41023,41108,41155,41286,41297,41298,41340,41503,41549,41552,41643,41837,41855,41888,41942,42151,42169,42204,42230,42450,42457,42586,42837,42850,43009,43025,43029,43037,43052,43135,43163,43183,43203,43211,43237,43310,43320,43329,43373,43469,43530,43574,43717,43793,43920,43971,44082,44168,44182,44266,44326,44510,44615,44742,44748,44833,44959,45017,45170,45230,45335,45486,45537,45597,45663,45747,45792,45833,45897,45977,45990,46134,46210,46270,46437,46645,46646,46688,46706,46725,46956,47077,47106,47159,47218,47431,47518,47633,47650,47716,47718,47765,47770,47844,47963,48003,48030,48039,48078,48111,48168,48256,48581,48627,48647,48717,48905,48994,49048,49211,49239,49328,49537,49564,49602,49707,49832,49844,49945,50119,50253,50373,50410,50435,50560,50593,50635,50723,50750,50770,50789,51095,51143,51200,51213,51332,51367,51438,51514,51553,51634,51753,51860,51911,51917,52033,52059,52177,52238,52400,52434,52525,52648,52668,52674,52697,52820,53148,53218,53341,53436,53484,53637,53774,53840,53843,53910,53946,54019,54025,54087,54162,54257,54287,54369,54478,54489,54545,54799,54987,55158,55276,55286,55387,55462,55684,55967,55998,56211,56239,56250,56323,56510,56524,56581,56873,57078,57262,57398,57473,57544,57615,57730,57819,58043,58164,58362,58449,58529,58679,58723,58741,58762,58979,59036,59075,59122,59479,59545,59814,59942,60418,60435,60598,60638,60639,60707,60708,60788,60888,60976,61096,61135,61145,61154,61583,61762,61891,61899,61941,61988,62026,62227,62312,62413,62418,62443,62468,62569,62591,62849,62979,63039,63143,63258,63316,63473,63617,63732,63753,63921,63952,64004,64021,64026,64118,64143,64162,64337,64609,64785,64843,64875,65002,65005,65116,65305,65319,65384,65751,65755,65841,65889,65921,65929,66030,66086,66219,66278,66345,66474,66614,66750,66874,66920,67026,67043,67168,67331,67392,67861,68038,68054,68141,68222,68252,68263,68320,68654,68781,68846,68910,69014,69082,69096,69275,69380,69401,69495,69599,69727,69830,69866,70013,70037,70176,70189,70236,70300,70440,70482,70501,70560,70572,70600,70663,70800,70876,71097,71106,71181,71234,71414,71476,71481,71502,71559,71637,71675,71708,71833,71916,71925,72087,72118,72303,72345,72365,72381,72387,72413,72439,72494,72526,72666,72690,72850,72866,72970,72974,73228,73318,73459,73510,73520,73700,73840,73877,74003,74017,74067,74104,74195,74297,74298,74423,74445,74602,74627,74657,74683,74758,74761,74808,74893,74896,75069,75211,75259,75330,75331,75462,75534,75674,75680,75722,75787,75810,75912,76064,76260,76399,76510,76673,76719,76743,76750,76853,76864,77072,77214,77273,77334,77336,77345,77391,77410,77583,77638,77656,77664,77673,78076,78298,78343,78493,78522,78565,78693,78839,78882,78933,79215,79263,79381,79409,79456,79474,79551,79605,79946,80000,80058,80183,80298,80446,81056,81117,81187,81201,81292,81332,81397,81415,81504,81604,81713,81822,81825,81844,81899,82268,82359,82431,82437,82480,82565,82625,82735,82769,82891,83112,83205,83231,83311,83335,83351,83429,83451,83513,83550,83568,83667,83760,83852,83873,83911,84074,84090,84148,84209,84247,84362,84404,84405,84573,84575,84698,84717,84813,84818,84986,85231,85339,85370,85492,85653,85687,85779,85879,86002,86012,86051,86052,86107,86129,86181,86217,86283,86643,86799,86922,86956,87187,87286,87397,87423,87466,87685,87832,87939,87976,87994,88057,88069,88136,88280,88400,88417,88437,88438,88510,88511,88709,88712,88814,88827,88902,88913,88943,88978,89380,89424,89459,89494,89931,89959,89966,90147,90286,90374,90411,90498,90549,90632,90728,90873,91107,91149,91191,91356,91399,91502,91522,91540,91610,91683,91718,91821,91877,91937,91952,92028,92057,92076,92147,92219,92223,92285,92367,92431,92490,92588,92624,93022,93050,93137,93149,93235,93280,93360,93415,93455,93571,93645,93754,93782,93813,93877,93893,93986,94010,94271,94325,94454,94713,94775,94859,94906,94990,95025,95085,95116,95458,95526,95631,95638,95727,95804,95865,96025,96093,96172,96237,96287,96310,96426,96540,96587,96609,96645,96765,96816,96821,96963,96978,97200,97379,97458,97460,97627,97717,97793,97840,98177,98247,98323,98355,98363,98626,98697,98734,98816,98844,98870,98935,99203,99292,99363,99392,99420,99548,99605,99651,99667,99904,99951",
          "126,165,198,521,527,619,901,1112,1115,1128,1168,1306,1310,1661,1665,1671,1811,1817,1825,1850,1852,1931,1968,2086,2141,2213,2270,2347,2411,2460,2582,2605,2621,2696,2715,2772,2798,2878,2885,2918,2922,3131,3341,3448,3496,3504,3527,3582,3614,3698,3711,3752,3762,3763,3808,3832,3891,3920,3939,3943,3945,3947,3957,4076,4175,4206,4227,4302,4316,4426,4526,4595,4658,4679,4797,4843,5016,5037,5041,5256,5710,5721,5959,6148,6198,6289,6327,6443,6467,6545,6561,6591,6600,6622,6666,6813,6854,6952,7147,7164,7285,7307,7325,7346,7396,7448,7468,7554,7603,7627,7629,7662,7682,7684,7725,7760,7789,8004,8005,8096,8128,8249,8296,8320,8464,8469,8484,8517,8527,8588,8595,8652,8735,8755,8759,8763,8893,8943,8983,9040,9049,9090,9097,9115,9220,9254,9256,9271,9413,9559,9701,9773,9784,9882,9898,10017,10191,10230,10238,10305,10427,10434,10727,10732,10834,10953,11004,11040,11045,11078,11107,11149,11160,11292,11330,11524,11648,11685,11726,11765,11838,11874,12113,12162,12178,12182,12222,12291,12430,12458,12509,12613,12661,12695,12744,12931,13028,13047,13091,13101,13139,13169,13179,13208,13259,13310,13366,13392,13428,13474,13544,13640,13810,13984,14010,14054,14095,14177,14256,14264,14529,14579,14761,14800,14817,14861,14912,14994,15093,15240,15320,15352,15362,15369,15416,15580,15620,15643,15704,15808,15886,15898,15978,16048,16062,16179,16376,16449,16493,16571,16602,16661,16705,16719,16747,16772,16991,17135,17150,17151,17231,17238,17315,17371,17455,17566,17567,17693,17728,17765,17839,17895,17946,18007,18069,18077,18127,18148,18177,18184,18186,18210,18272,18283,18366,18369,18485,18506,18651,18803,18809,18926,18953,19196,19197,19200,19293,19346,19414,19517,19552,19612,19630,19835,19885,19985,20051,20168,20174,20178,20218,20239,20240,20260,20401,20429,20510,20576,20650,20800,20875,20916,20921,20923,21048,21053,21060,21129,21134,21180,21204,21241,21387,21423,21437,21668,21723,21849,21978,21988,22007,22044,22064,22097,22155,22292,22396,22474,22574,22658,22677,22702,22736,22837,22921,22957,23010,23242,23279,23303,23328,23347,23475,23561,23564,23629,23635,23709,23742,23786,23878,23933,23987,24059,24095,24128,24343,24517,24692,24709,24731,24888,24911,24931,25011,25094,25112,25154,25201,25233,25256,25261,25339,25540,25580,25607,25609,25637,25787,25802,25870,25891,25958,25985,25999,26000,26216,26229,26295,26296,26336,26440,26795,26841,26947,26953,26959,27113,27156,27215,27262,27356,27491,27597,27645,27754,27828,27837,27843,27845,28014,28038,28281,28344,28390,28419,28555,28561,28576,28646,28683,28859,28921,28963,29029,29122,29154,29177,29252,29352,29603,29686,29706,29768,29788,29992,30105,30297,30357,30427,30460,30610,30612,30616,30798,30806,30869,30925,30952,30991,31099,31121,31406,31444,31520,31681,31690,31730,31782,32051,32087,32193,32224,32267,32322,32391,32439,32662,32680,32683,32693,32716,32761,32828,32844,32878,32902,32930,32973,33161,33228,33252,33312,33481,33519,33676,33699,33771,33786,33801,33825,33926,33975,34029,34151,34161,34224,34246,34298,34345,34361,34412,34420,34485,34729,34739,34922,34934,35004,35137,35166,35255,35329,35367,35480,35509,35516,35540,35546,35572,35587,35626,35684,35758,35899,35904,36229,36240,36271,36286,36396,36401,36428,36475,36738,36885,36898,36931,37093,37122,37289,37342,37471,37610,37826,37858,37907,37945,38176,38228,38377,38449,38503,38535,38614,38641,38911,39023,39143,39178,39182,39233,39355,39467,39754,39903,40214,40251,40260,40319,40330,40342,40425,40666,40694,40778,40788,40824,41047,41060,41147,41497,41585,41663,41769,41860,41896,41939,42024,42034,42167,42210,42214,42220,42282,42304,42403,42417,42449,42588,42744,42806,42910,43013,43181,43247,43263,43270,43279,43393,43430,43453,43464,43486,43567,43594,43692,43704,43752,43760,43773,43795,43967,44003,44037,44208,44234,44280,44331,44345,44522,44546,44552,44565,44590,44641,44703,44874,44926,44991,45302,45323,45325,45373,45477,45525,45603,45629,45630,45761,45804,45811,46051,46060,46151,46176,46190,46271,46483,46578,46618,46663,46680,46839,46977,47053,47203,47292,47320,47392,47414,47438,47652,47675,47993,48006,48025,48086,48113,48183,48269,48335,48376,48391,48399,48406,48488,48594,48621,48656,48820,48898,48924,49038,49042,49086,49098,49115,49202,49249,49331,49387,49542,49634,49791,49951,50004,50137,50167,50215,50252,50260,50344,50506,50522,50619,50717,50739,50867,50892,50957,51001,51031,51234,51267,51339,51342,51495,51680,51783,51826,51841,51967,51968,52079,52102,52208,52210,52270,52291,52551,52608,52609,52685,52782,52900,53141,53179,53267,53290,53512,53737,53765,53771,53773,53809,53891,53945,54026,54039,54092,54124,54142,54194,54363,54598,54627,54642,54656,54662,54703,54720,54729,54746,54764,55019,55061,55154,55252,55398,55406,55650,55815,55870,55924,55934,56114,56159,56163,56204,56252,56465,56555,56640,56645,56671,56715,57010,57068,57158,57164,57181,57249,57449,57460,57562,57633,57653,57674,57769,57771,57982,58031,58085,58117,58127,58130,58173,58195,58281,58322,58346,58403,58435,58504,58600,58653,58664,58829,58885,59018,59047,59061,59069,59209,59248,59414,59493,59536,59582,59645,59695,59737,59750,59783,59862,59967,59977,60011,60034,60043,60088,60123,60264,60298,60521,60600,60743,60807,61021,61057,61101,61161,61313,61639,61650,61666,61714,61746,61837,61920,61935,62059,62143,62242,62328,62340,62502,62521,62622,62654,62702,62713,62828,62983,62987,63144,63170,63262,63263,63497,63523,63563,63583,63663,64059,64120,64157,64174,64270,64289,64450,64459,64474,64493,64531,64553,64651,64833,65058,65081,65130,65236,65303,65383,65455,65473,65605,65608,65654,65742,65911,65972,66038,66120,66212,66213,66226,66240,66402,66478,66564,66643,66676,66811,66951,67005,67065,67079,67107,67232,67295,67297,67312,67324,67335,67373,67491,67508,67585,67620,67648,67656,67717,67755,68012,68103,68294,68306,68365,68570,68667,68675,68696,68743,68969,68979,69010,69044,69102,69106,69118,69185,69227,69260,69364,69419,69456,69491,69543,69687,69710,69802,69871,69967,70040,70117,70151,70341,70352,70451,70599,70674,70812,70828,70843,70856,70939,71056,71060,71115,71189,71213,71236,71240,71258,71320,71362,71443,71497,71545,71775,71792,71847,71865,72060,72124,72319,72880,73209,73230,73235,73350,73379,73624,73719,73809,73870,74213,74240,74282,74318,74398,74406,74523,74665,74669,74684,74727,74801,74816,74870,74886,74892,74894,74933,74978,75011,75077,75082,75099,75136,75421,75476,75488,75795,75883,75939,76091,76373,76409,76445,76551,76677,76678,76755,76797,76834,76861,76932,76985,77104,77148,77184,77197,77287,77455,77457,77481,77486,77533,77534,77549,77625,77688,77699,77753,77968,78004,78126,78136,78172,78176,78225,78246,78253,78259,78318,78464,78500,78547,78635,78670,78702,78707,78795,78858,78872,78965,78967,78999,79040,79096,79129,79266,79275,79292,79302,79401,79441,79593,79643,79739,79911,79966,80030,80032,80165,80208,80214,80421,80552,80581,80584,80666,80682,80786,80955,81065,81066,81280,81303,81370,81511,81521,81544,81545,81599,81605,81624,81638,81703,81893,81931,81965,82111,82119,82150,82238,82282,82390,82395,82430,82522,82629,82638,82822,82894,83040,83064,83184,83223,83323,83344,83355,83520,83606,83615,83635,83724,83731,83836,83875,83936,83943,84000,84050,84136,84183,84214,84226,84242,84250,84285,84320,84620,84621,84645,84685,84690,84713,84908,84941,85017,85114,85470,85542,85636,85652,85744,85897,85923,85987,86035,86170,86192,86232,86424,86603,86652,86740,86774,86788,86815,86970,86988,87030,87080,87209,87258,87421,87444,87512,87609,87645,87739,87815,87853,88076,88150,88175,88176,88190,88203,88222,88388,88545,88687,88708,88802,88994,89036,89133,89218,89486,89529,89622,89667,89677,89795,89858,90009,90056,90114,90261,90271,90276,90325,90364,90403,90492,90642,90759,90765,90794,91023,91070,91209,91303,91325,91406,91601,91690,91716,91724,91855,91899,92010,92290,92320,92382,92408,92439,92541,92549,92550,92671,92756,92800,92873,92892,93011,93145,93370,93640,93662,93681,93765,93769,93854,93921,93963,93985,94077,94169,94289,94326,94342,94424,94451,94532,94544,94613,94663,94673,94749,94756,94783,94799,94857,94960,95005,95088,95110,95113,95115,95232,95291,95393,95415,95431,95476,95520,95711,95845,95857,95888,95900,96007,96050,96108,96329,96341,96496,96512,96581,96597,96660,96877,96975,97354,97407,97438,97465,97469,97552,97862,97894,98084,98366,98398,98472,98493,98552,98864,98966,99033,99101,99113,99222,99231,99295,99324,99461,99485,99496,99553,99555,99594,99626,99706,99921,99936,99938",
          "3,33,170,194,291,375,639,656,752,765,814,829,900,923,925,947,1013,1243,1268,1539,1555,1575,1678,1768,1816,1840,1842,2222,2227,2231,2235,2636,2658,2664,2685,2719,2869,2873,2929,3020,3037,3199,3284,3338,3562,3681,3712,3735,3747,3802,3837,3972,3977,3993,4077,4198,4212,4275,4286,4385,4503,4540,4569,4624,4748,4905,5111,5187,5202,5214,5337,5372,5419,5653,5726,5788,5866,5881,6014,6116,6333,6340,6422,6527,6642,6718,6916,6939,6984,6996,7044,7266,7336,7410,7431,7445,7455,7755,7768,7822,7925,7941,8007,8016,8084,8253,8263,8323,8437,8541,8613,8705,8723,9008,9302,9417,9470,9525,9596,9648,9655,9683,9710,9774,9827,9865,9935,10027,10211,10322,10424,10440,10561,10752,10837,10843,10908,10939,11121,11220,11250,11342,11363,11373,11477,11513,11816,11979,12006,12061,12092,12169,12190,12233,12286,12317,12378,12511,12589,12649,12684,12699,12741,12855,12935,13011,13044,13078,13150,13356,13416,13445,13464,13508,13522,13701,13707,13745,13818,13901,13955,13966,14053,14060,14150,14181,14456,14522,14548,14549,14555,14559,14716,14748,14788,14842,14918,14924,14945,14947,15068,15107,15168,15192,15235,15237,15253,15258,15259,15400,15485,15624,15671,15676,15697,15707,15723,15753,15837,15981,16002,16018,16124,16126,16203,16216,16270,16289,16329,16443,16494,16495,16516,16574,16584,16669,16670,16693,16703,16821,16899,16999,17040,17063,17076,17186,17228,17327,17341,17405,17620,17690,17709,17723,17762,17784,17854,17917,17936,18018,18053,18059,18068,18096,18115,18132,18247,18395,18405,18491,18599,18666,18748,18757,18940,18985,19057,19082,19102,19138,19307,19391,19522,19538,19581,19598,19666,19677,19690,19895,19978,20003,20080,20094,20186,20305,20326,20327,20345,20435,20712,20777,20899,20982,21190,21350,21419,21422,21475,21564,21655,21671,21712,21839,21876,21882,21923,21945,22052,22056,22099,22194,22255,22259,22323,22501,22558,22650,22724,22749,22820,22896,22936,22949,22962,22979,23071,23193,23287,23319,23364,23499,23541,23569,23602,23615,23727,23738,23744,24073,24133,24145,24321,24334,24360,24454,24554,24626,24689,24754,24829,24836,24918,24992,25066,25073,25109,25166,25377,25378,25402,25428,25445,25480,25496,25616,25693,25696,25773,25785,25863,25896,25955,26004,26053,26102,26131,26169,26354,26370,26400,26420,26422,26471,26596,26657,26664,26769,26843,26935,27172,27276,27296,27402,27482,27605,27674,27758,27768,27794,27820,27853,27879,27880,27970,28005,28129,28181,28243,28263,28275,28294,28397,28422,28472,28498,28507,28566,28571,28677,28686,28733,28839,28849,28861,28930,28942,28978,28990,29034,29062,29094,29135,29138,29155,29248,29268,29350,29394,29637,29663,29747,29749,29756,29810,29830,29831,29839,29933,30038,30064,30073,30093,30154,30434,30595,30682,30801,30875,30912,31003,31159,31329,31335,31379,31411,31512,31614,31619,31684,31762,31770,31780,31864,32035,32109,32213,32247,32332,32431,32457,32647,32677,32768,32840,32912,32995,33002,33116,33155,33181,33195,33315,33336,33353,33361,33371,33397,33435,33448,33457,33511,33545,33598,33669,33713,33758,33821,33823,33847,33876,33928,34008,34021,34090,34105,34202,34373,34543,34568,34612,34613,34645,34725,34777,34916,34994,35317,35401,35640,35661,35699,35777,35801,35857,35863,35872,35917,35951,35982,36038,36118,36295,36354,36397,36450,36547,36550,36679,36707,36782,36849,36981,36992,37011,37030,37034,37049,37159,37229,37293,37389,37396,37492,37504,37597,37611,37614,37617,37635,37717,37755,37817,37935,38089,38187,38189,38320,38324,38360,38446,38539,38546,38566,38663,38842,38896,38899,38994,39072,39133,39285,39443,39545,39593,39888,40017,40057,40163,40248,40276,40485,40498,40513,40630,40631,40648,40661,40851,40974,40976,40996,41321,41338,41364,41412,41422,41465,41527,41537,41607,41693,41705,41777,41845,41884,41893,42081,42189,42192,42202,42215,42441,42537,42665,42718,42829,42903,42952,42953,43093,43095,43141,43182,43248,43257,43405,43590,43614,43617,43772,43834,43940,43975,44180,44278,44311,44339,44403,44530,44556,44708,44810,45005,45087,45107,45118,45239,45354,45380,45413,45656,45705,45709,45743,45782,45830,45934,45966,46214,46542,46699,46701,46735,46787,46805,46821,46894,47151,47189,47339,47388,47393,47547,47740,47815,47838,47882,47969,48029,48084,48110,48238,48348,48461,48476,48541,48558,48598,48715,48832,48842,48979,48985,49093,49125,49272,49316,49455,49478,49499,49618,49619,49644,49727,49737,49743,49751,49876,49901,49935,49954,49990,50114,50144,50415,50614,50725,50748,50840,50878,51002,51077,51135,51215,51295,51311,51421,51469,51470,51505,51573,51757,51769,51837,51864,51941,51946,51991,52213,52326,52349,52390,52519,52708,52801,52925,52951,52992,53032,53137,53161,53328,53606,53782,53810,54017,54137,54375,54378,54401,54405,54435,54500,54509,54528,54547,54576,54586,54658,54685,54875,54887,55132,55259,55287,55355,55368,55530,55549,55583,55623,55659,55698,55747,55760,55766,55777,55800,55856,55925,55965,56255,56595,56617,56701,56751,56766,56836,56870,56872,56967,57005,57026,57094,57101,57137,57165,57226,57314,57334,57540,57662,57677,57681,57749,57778,57806,57870,57968,58008,58026,58053,58066,58152,58165,58262,58265,58294,58392,58408,58525,58535,58647,58852,59066,59087,59177,59218,59335,59469,59699,59767,59893,59945,59971,59974,60052,60090,60160,60185,60213,60284,60356,60423,60493,60512,60546,60650,60667,60747,60876,60927,60940,60986,61111,61168,61218,61289,61339,61393,61400,61503,61961,62181,62192,62215,62352,62488,62640,62755,62786,62820,62944,62997,63136,63246,63319,63354,63359,63379,63381,63459,63674,63685,63913,64036,64253,64327,64507,64549,64599,64601,64666,64758,64767,64960,65013,65202,65226,65255,65274,65329,65440,65534,65552,65569,65575,65577,65642,65704,65759,65839,66001,66022,66307,66558,66591,66621,66638,66664,66665,66709,66794,67135,67136,67144,67304,67336,67346,67351,67432,67460,67515,67675,67695,67780,67819,67903,67968,68060,68116,68180,68237,68303,68419,68809,68834,69157,69178,69273,69296,69340,69386,69557,69620,69684,69788,69817,69853,70017,70032,70072,70078,70102,70168,70226,70271,70276,70579,70757,70819,70932,70971,71009,71052,71075,71095,71124,71133,71194,71271,71850,71895,72107,72115,72143,72196,72258,72268,72279,72447,72567,72829,72922,72924,72962,73072,73291,73380,73391,73404,73412,73506,73511,73547,73671,73725,73956,74021,74080,74083,74093,74121,74127,74262,74293,74349,74421,74458,74471,74475,74490,74506,74562,74773,74830,74869,74929,74942,74967,75016,75046,75057,75202,75229,75258,75366,75379,75383,75819,75873,75899,75934,75937,75971,76010,76123,76132,76209,76331,76453,76669,76711,76722,76862,77021,77130,77168,77210,77262,77299,77375,77470,77498,77529,77554,77603,77617,77635,77761,77874,77899,78014,78053,78079,78102,78156,78168,78188,78189,78373,78394,78502,78518,78521,78561,78652,78753,78827,78945,79046,79048,79071,79086,79123,79125,79183,79206,79216,79298,79300,79382,79809,80232,80415,80597,80643,80699,80700,80893,80916,81045,81072,81195,81220,81243,81291,81369,81446,81485,81617,81623,81629,81849,81917,81942,81997,82073,82146,82250,82292,82358,82469,82564,82706,82786,82815,82857,82869,82890,82933,83007,83011,83073,83098,83361,83404,83437,83467,83573,83625,83739,83825,83920,83940,83998,84010,84098,84227,84301,84361,84372,84465,84472,84560,84659,84683,84835,84840,84912,85086,85110,85227,85316,85336,85364,85392,85397,85404,85537,85556,85590,85708,85758,85823,85846,85911,85958,85974,85986,86062,86157,86360,86497,86564,86622,86648,86977,87177,87181,87222,87253,87278,87292,87307,87503,87772,87892,87927,87928,87987,88011,88257,88269,88320,88396,88457,88484,88524,88529,88563,88582,88787,88807,88831,88832,88844,88923,89131,89156,89199,89376,89402,89474,89480,89517,89519,89584,89603,89605,89712,89899,89956,90028,90162,90173,90189,90262,90293,90326,90328,90363,90394,90437,90583,90609,90712,90844,90890,90940,90963,91182,91475,91533,91552,91642,91758,91814,91980,92136,92283,92760,92902,92964,93000,93042,93106,93110,93152,93209,93309,93396,93686,93722,93726,93935,93998,94116,94117,94170,94206,94449,94462,94796,94828,94891,95117,95147,95384,95544,95637,95714,95815,95830,95868,95979,96044,96201,96257,96307,96364,96594,96603,96620,96631,96737,96989,97061,97192,97316,97364,97423,97452,97519,97648,97651,97657,97933,98025,98033,98137,98152,98180,98238,98509,98532,98586,98714,98720,98723,98830,98883,99097,99240,99313,99374,99375,99458,99489,99493,99536,99746,99773,99931,99946",
          "73,205,209,281,319,397,574,618,642,839,855,1162,1183,1216,1350,1368,1384,1391,1396,1450,1454,1512,1568,1584,1775,1844,2080,2157,2216,2281,2296,2501,2552,2660,2686,2739,2754,2831,2864,3004,3159,3222,3366,3476,3566,3944,4050,4057,4183,4250,4262,4374,4461,4625,4721,4749,4808,4874,4878,4898,4932,4948,4981,5083,5412,5475,5578,5771,5808,5875,5915,6248,6266,6275,6313,6506,6522,6533,6618,6705,6734,6845,6930,7054,7060,7131,7345,7408,7551,7601,7637,7681,7779,7817,7876,7935,8001,8169,8268,8426,8582,8609,8713,8729,8746,8834,8905,8907,8953,8964,8991,9072,9139,9159,9175,9181,9228,9255,9352,9466,9496,9644,9794,9843,9917,10129,10253,10455,10481,10501,10613,10716,10880,10956,10960,10964,10975,11070,11126,11145,11200,11372,11461,11590,11642,11795,11807,11835,11881,11899,11907,11990,11996,12022,12315,12367,12383,12610,12627,12663,12703,12835,12906,13004,13103,13134,13148,13246,13361,13453,13504,13568,13600,13624,13719,13848,13871,14058,14176,14190,14203,14470,14492,14538,14799,14867,14907,14948,15103,15163,15215,15285,15345,15375,15429,15498,15550,15599,15600,15632,15657,15696,15716,15796,15807,15825,15853,15949,16017,16027,16032,16043,16088,16116,16172,16200,16291,16625,16807,16839,17021,17100,17127,17302,17348,17380,17454,17513,17591,17644,17657,17694,17695,17699,17832,17844,17993,18023,18194,18291,18463,18477,18488,18502,18509,18553,18624,18913,18938,19015,19034,19080,19125,19235,19256,19292,19419,19462,19516,19521,19537,19542,19557,19623,19760,19881,19964,19977,20069,20123,20212,20216,20409,20423,20463,20751,20775,20830,20868,20869,20905,20913,20933,20974,21002,21036,21149,21160,21248,21262,21321,21331,21404,21448,21465,21492,21585,21622,21642,21666,21759,21856,21864,21951,22038,22294,22301,22338,22397,22458,22577,22593,22738,22970,22993,23007,23030,23107,23113,23158,23181,23211,23218,23225,23271,23273,23296,23617,23632,23644,23658,23660,23696,23720,23745,23749,23798,23828,23834,23868,23963,24048,24079,24110,24183,24243,24344,24403,24444,24448,24484,24784,24889,24903,24954,24976,25085,25098,25125,25407,25415,25468,25657,25663,25678,25684,25778,25826,25971,25977,26104,26165,26237,26272,26366,26397,26441,26473,26553,26807,27187,27295,27405,27685,27735,27748,27963,28013,28136,28316,28439,28502,28670,28699,28701,28947,28957,28975,29326,29360,29402,29498,29521,29522,29579,29588,29656,29710,29758,29872,29889,29935,30148,30303,30350,30449,30492,30495,30790,30891,30900,30904,31026,31120,31155,31166,31400,31402,31551,31633,31689,31774,31821,31943,31948,31997,32044,32085,32126,32177,32182,32222,32242,32368,32419,32458,32468,32543,32644,32784,32850,32880,33080,33120,33206,33295,33603,33949,34072,34194,34274,34331,34545,34565,34570,34601,34614,34646,34671,34837,34876,35111,35124,35178,35196,35277,35463,35479,35547,35575,35631,35682,35783,35927,35988,36026,36047,36282,36353,36395,36398,36485,36522,36566,36580,36598,36767,36803,36872,36888,36913,37089,37123,37333,37351,37383,37414,37518,37561,37619,37634,37706,37751,37820,37823,37848,37882,37915,38051,38058,38116,38340,38393,38448,38463,38483,38494,38635,38769,38905,38919,39008,39078,39284,39314,39318,39368,39401,39468,39481,39567,39573,39642,39644,39654,39670,39803,40107,40354,40397,40398,40566,40583,40638,40640,40662,40667,40768,40857,40867,40879,40926,40934,40990,41022,41061,41112,41214,41224,41300,41317,41348,41362,41661,41797,41800,41813,41819,42003,42045,42054,42108,42124,42163,42223,42240,42346,42442,42470,42496,42615,42634,42661,42694,42784,42786,42894,42945,43015,43051,43102,43230,43273,43374,43514,43622,43669,43673,43699,43751,43802,43825,44323,44425,44429,44534,44574,44603,44607,44809,44868,44914,45151,45218,45220,45265,45371,45461,45476,45541,45639,45700,45845,46004,46116,46117,46316,46362,46367,46370,46392,46433,46488,46524,46757,46786,46815,47043,47049,47248,47345,47383,47408,47467,47509,47520,47559,47621,47717,47948,48187,48242,48255,48279,48285,48375,48477,48576,48754,48816,48860,48899,49033,49063,49151,49201,49234,49283,49351,49507,49514,49576,49591,49664,49741,49796,49810,49917,49965,50042,50117,50128,50243,50325,50480,50581,50738,50771,50912,50929,50975,50989,51055,51056,51091,51111,51165,51320,51343,51443,51503,51545,51566,51584,51618,51619,51635,51639,51709,51799,51846,51861,51868,51884,51915,52030,52110,52174,52258,52321,52344,52420,52710,52892,52942,53068,53118,53172,53193,53272,53465,53524,53565,53618,53873,54454,54536,54559,54649,54681,54787,54809,54889,54991,55080,55115,55247,55334,55336,55475,55494,55496,55576,55806,56085,56093,56105,56171,56176,56202,56266,56391,56395,56427,56533,56571,56733,56839,56929,56959,57023,57062,57291,57313,57393,57483,57524,57590,57675,57865,57976,58174,58230,58236,58269,58497,58624,58662,58682,58689,58728,58976,59181,59186,59267,59309,59318,59603,59658,59703,59740,59741,59785,59798,59812,60018,60106,60236,60274,60322,60381,60501,60574,60630,60758,60767,60884,60934,61164,61363,61391,61414,61427,61484,61713,61832,61954,61976,62020,62061,62068,62088,62135,62155,62157,62188,62196,62226,62376,62429,62439,62477,62652,62712,62753,62816,62850,62857,62998,63003,63023,63085,63103,63142,63156,63226,63336,63416,63427,63559,63609,63631,63819,63915,63945,64011,64038,64176,64417,64452,64485,64857,65015,65174,65239,65313,65524,65548,65584,65707,65844,65861,65896,65999,66185,66203,66245,66305,66405,66413,66483,66719,66723,66775,66846,66877,66881,66906,66966,66983,67082,67102,67120,67219,67223,67225,67459,67539,67822,67960,68000,68023,68057,68059,68198,68295,68318,68501,68572,68631,68712,68726,68803,68824,68928,68957,68971,69018,69058,69062,69213,69252,69267,69303,69352,69392,69472,69509,69565,69595,69627,69632,69762,69792,69806,69847,69883,70209,70305,70340,70371,70444,70473,70636,70673,70690,70883,70892,70980,71047,71063,71098,71316,71446,71477,71595,71657,71879,71884,71959,72022,72024,72066,72086,72224,72231,72238,72295,72352,72435,72438,72465,72565,72597,72774,72812,72830,72927,72933,72978,73005,73016,73085,73126,73148,73210,73263,73267,73301,73401,73403,73427,73533,73570,73746,73785,73934,73967,74054,74090,74149,74183,74259,74325,74335,74364,74453,74551,74593,74600,74697,74772,74842,74924,74986,74990,74996,75002,75017,75232,75352,75382,75450,75596,75634,75681,75684,75704,75865,75879,76081,76114,76192,76302,76323,76630,76899,76904,77270,77272,77313,77333,77422,77448,77756,77772,77774,77786,77916,77954,78170,78280,78294,78484,78599,78611,78674,78764,78959,79038,79164,79172,79242,79306,79458,79490,79592,79655,79681,79689,79756,79880,79883,79991,80105,80137,80141,80182,80250,80301,80307,80350,80440,80503,80561,80612,80639,80688,80737,80829,80860,80863,80934,80937,81040,81093,81114,81157,81295,81324,81405,81438,81445,81647,81750,81831,81838,81921,82105,82116,82225,82276,82305,82373,82374,82405,82422,82515,82582,82589,82600,82643,82749,82845,82944,82971,83209,83225,83504,83605,83703,83756,83872,83881,83885,84211,84492,84495,84633,84857,84974,85646,85659,85677,85683,85693,85800,85805,85883,85886,85949,86119,86218,86241,86245,86256,86303,86327,86453,86490,86525,86575,86627,86653,86789,87116,87276,87320,87443,87548,87585,87600,87617,87654,87656,87732,87841,88108,88232,88349,88432,88663,88741,88892,89004,89007,89104,89163,89319,89408,89411,89516,89585,89632,89708,89758,89767,89812,89829,89987,90054,90131,90242,90285,90362,90379,90393,90406,90439,90561,90577,90724,90743,90751,90841,90905,90925,90995,91042,91117,91328,91377,91395,91396,91454,91486,91497,91599,91615,91618,91635,91708,91728,91822,91835,91878,91999,92024,92122,92138,92278,92343,92395,92769,92824,92868,92887,92956,93006,93188,93289,93416,93427,93432,93437,93441,93460,93630,93698,93837,93960,93987,94008,94012,94095,94111,94135,94319,94581,94586,94740,94776,94779,94801,94864,94882,94888,95020,95129,95257,95293,95344,95360,95442,95498,95575,95622,95794,95799,95835,95885,95988,96218,96404,96414,96443,96515,96545,96773,96803,96804,97089,97091,97121,97139,97219,97259,97291,97333,97418,97559,97940,97960,97963,98081,98113,98115,98249,98441,98488,98500,98576,98655,98698,98776,98781,98825,98859,98950,98973,99158,99452,99506,99525,99558,99623,99665,99767,99770,99808,99915,99966,99996",
          "60,83,260,317,421,498,554,557,573,774,907,908,1088,1107,1236,1324,1338,1447,1457,1475,1498,1687,1727,1932,1962,2010,2017,2052,2232,2252,2265,2279,2386,2430,2510,2545,2567,2588,2613,2638,2813,2834,2836,2925,2945,2965,3017,3124,3243,3478,3661,3720,3750,3760,3803,4207,4226,4264,4289,4299,4358,4448,4507,4789,4870,4893,4961,5013,5178,5245,5308,5405,5591,5668,5745,5766,5853,6050,6230,6391,6435,6451,6466,6555,6611,6701,6827,6872,6891,6917,7130,7194,7201,7284,7374,7395,7477,7489,7519,7522,7609,7666,7851,7966,8027,8078,8102,8132,8219,8236,8386,8718,8933,9038,9074,9131,9169,9173,9185,9217,9224,9289,9291,9297,9298,9372,9520,9688,9716,9871,10084,10215,10239,10512,10545,10788,10826,10911,11027,11091,11316,11428,11438,11441,11463,11511,11744,11777,11794,11929,12007,12020,12111,12147,12201,12280,12432,12454,12466,12485,12494,12808,12848,12998,13111,13213,13302,13332,13693,13723,13754,13768,13781,13803,14059,14120,14292,14313,14429,14438,14485,14506,14676,14696,14735,14875,14888,15017,15176,15302,15415,15467,15882,16068,16078,16190,16273,16323,16327,16340,16455,16748,16822,16941,16990,17071,17106,17179,17190,17291,17377,17414,17425,17434,17435,17572,17627,17730,17933,18142,18143,18620,18702,18934,18972,19024,19065,19244,19455,19620,19759,19770,19817,19821,19947,20183,20187,20222,20271,20346,20374,20433,20457,20484,20606,20642,20828,20894,20907,20938,20953,21015,21105,21109,21195,21233,21324,21334,21624,21720,21820,21911,22189,22302,22305,22456,22626,22654,22813,22930,22972,23012,23084,23138,23335,23349,23605,23621,23784,23897,23924,24097,24136,24178,24197,24247,24300,24308,24341,24457,24485,24521,24524,24573,24671,24746,24814,24878,24998,25057,25086,25173,25209,25251,25292,25646,25915,26142,26268,26309,26320,26534,26556,26557,26724,26826,26934,27086,27231,27249,27288,27358,27438,27458,27504,27512,27517,27538,27588,27634,27669,27734,27911,28061,28167,28187,28353,28434,28482,28540,28579,28676,28713,28749,28989,29435,29589,29598,29609,29660,29795,29803,29854,29918,29970,30232,30236,30301,30319,30379,30385,30408,30554,30685,30750,31030,31231,31279,31290,31318,31427,31435,31475,31598,31654,31655,31794,31814,31894,31926,31967,32032,32250,32315,32331,32364,32378,32754,32808,32822,33019,33067,33156,33296,33322,33436,33654,33716,33721,33728,33743,33855,34007,34065,34124,34157,34182,34368,34424,34478,34493,34496,34519,34548,34714,34741,34794,35167,35180,35203,35233,35512,35601,35648,35674,35794,36063,36347,36465,36517,36560,36577,36599,36691,36703,36784,36832,36857,36953,37051,37146,37184,37206,37224,37302,37419,37576,37733,37790,37855,37861,37911,38077,38094,38195,38277,38332,38366,38562,38579,38589,38695,38728,38940,38944,39016,39074,39248,39328,39356,39451,39503,39536,39566,39598,39612,39667,39687,39751,39833,39959,40144,40256,40277,40303,40331,40461,40514,40530,40587,40732,40808,40811,40870,40889,41035,41062,41117,41163,41217,41236,41312,41395,41397,41510,41531,41832,41852,41937,42121,42134,42157,42205,42238,42363,42415,42416,42473,42501,42505,42524,42677,42802,43136,43401,43457,43713,43826,43860,44057,44112,44199,44212,44249,44306,44358,44467,44473,44490,44525,44686,44704,44788,44801,45001,45067,45075,45131,45162,45164,45412,45562,45595,45607,45610,45756,45873,45909,45935,45944,46232,46314,46385,46440,46554,46633,46793,46859,46890,46913,47161,47617,47628,47668,47759,47869,47923,48017,48224,48396,48485,48563,48776,48779,48818,48868,48888,49207,49364,49555,49622,49939,49947,50188,50227,50391,50419,50470,50644,50673,50873,50984,51089,51153,51283,51303,51335,51354,51404,51419,51653,51743,51758,51798,51809,51897,51916,51956,51989,52108,52144,52203,52230,52251,52384,52458,52516,52523,52596,52750,52929,53020,53171,53195,53208,53230,53262,53600,53636,53641,53879,53926,53991,54325,54517,54608,54660,54827,54950,54964,55053,55055,55077,55079,55186,55202,55214,55266,55268,55273,55376,55440,55514,55633,55679,55688,55839,55841,55897,55954,56002,56015,56041,56151,56303,56314,56366,56396,56415,56435,56526,56548,56795,56974,57002,57003,57086,57114,57168,57355,57625,57636,57642,57685,57762,57830,57890,58013,58035,58040,58180,58520,58660,58757,58765,58848,58871,58909,59000,59016,59063,59254,59429,59443,59452,59617,59838,59886,59900,60126,60218,60235,60292,60397,60442,60584,60661,60677,60685,60729,60745,60786,60791,60808,60827,60847,60936,60975,60988,61046,61246,61335,61499,61513,61551,61634,61681,61717,61844,61863,61888,61903,62130,62221,62250,62319,62469,62495,62567,62573,62604,62617,62826,62951,63002,63112,63196,63469,63560,63568,63824,63954,64000,64019,64178,64356,64388,64443,64667,64670,64779,64983,65038,65292,65296,65340,65375,65422,65516,65586,65615,65724,65740,65769,65806,65854,65907,65917,65919,65952,65989,66178,66248,66358,66482,66486,66492,66538,66981,66982,67009,67016,67113,67229,67406,67467,67576,67752,67758,67817,67841,68316,68328,68527,68660,68788,68985,69057,69091,69131,69230,69440,69525,69886,70016,70061,70201,70243,70249,70275,70334,70372,70374,70379,70492,70530,70541,70545,70646,70693,70760,70774,70797,70849,70921,70923,71215,71255,71310,71405,71527,71538,71579,71639,71692,71900,72052,72170,72229,72254,72517,72571,72819,73196,73260,73313,73325,73378,73526,73534,73561,73937,74001,74013,74084,74187,74501,74604,74686,74860,74887,74943,75092,75105,75255,75264,75313,75329,75477,75510,75583,75624,75751,75817,76103,76292,76332,76415,76440,76449,76478,76528,76607,76619,76645,76767,76794,76893,77185,77250,77387,77436,77496,77592,77680,77796,78190,78191,78430,78449,78465,78661,78895,78898,78906,78911,79006,79105,79152,79211,79250,79332,79488,79501,79904,79916,80029,80057,80081,80115,80129,80202,80295,80336,80541,80542,80599,80697,80725,80744,80850,80921,81087,81184,81193,81262,81299,81315,81361,81661,81804,81881,81938,82279,82439,82514,82525,82940,82968,83043,83267,83407,83509,83541,83592,83618,83705,83903,84087,84246,84287,84290,84390,84513,84727,84751,84852,84939,85075,85101,85163,85248,85319,85402,85421,85498,85543,85654,85717,85730,85799,85873,86137,86143,86184,86226,86243,86362,86442,86636,86640,86959,87103,87121,87321,87520,87662,87686,87731,87800,87830,87847,87985,87988,88000,88015,88216,88219,88300,88331,88338,88348,88427,88631,88698,88721,88959,89230,89273,89284,89803,89811,90191,90195,90302,90371,90422,90435,90438,90602,90673,90680,90750,90863,90878,90888,90909,90971,90979,90997,91012,91092,91099,91313,91423,91518,91524,91576,91818,91853,92055,92108,92258,92511,92536,92543,92544,92663,92733,92989,93072,93080,93202,93398,93470,93528,93666,93735,93808,93900,93992,94034,94083,94269,94344,94351,94433,94597,94671,94708,94720,94736,94921,94924,94945,94979,95001,95150,95196,95244,95265,95266,95465,95496,95525,95554,95570,95684,95743,95922,96056,96153,96255,96272,96418,96614,96647,96663,96747,96754,96761,96808,96842,96894,96916,96930,96972,97096,97582,97671,97675,97721,97782,97875,97980,98147,98165,98198,98224,98284,98417,98421,98436,98449,98573,98673,98929,98998,99045,99064,99232,99236,99246,99286,99436,99507,99544,99585,99597,99688,99943",
          "29,112,136,227,371,427,445,565,586,589,592,708,833,869,930,1073,1081,1192,1237,1285,2035,2079,2181,2266,2268,2403,2446,2482,2531,2714,2720,2757,2891,2934,3019,3032,3053,3162,3167,3257,3386,3491,3495,3516,3547,3555,3597,3606,3770,3956,4074,4121,4260,4373,4445,4567,4623,4750,4768,4799,4837,4937,5017,5066,5139,5150,5260,5466,5507,5514,5622,5673,5763,5764,5804,5828,5920,5975,6049,6054,6385,6410,6442,6471,6511,6568,6585,6624,6714,6758,6842,6846,6897,7068,7097,7112,7120,7187,7188,7205,7260,7283,7288,7328,7474,7572,7749,7783,8168,8210,8286,8393,8539,8543,8591,8701,8795,8823,8937,8949,9056,9118,9148,9164,9188,9231,9268,9427,9462,9599,9628,9675,9740,9805,9980,10029,10042,10138,10268,10348,10365,10462,10526,10532,10638,10676,10757,10772,10984,10987,11248,11324,11340,11381,11439,11543,11596,11649,11678,11687,11710,11810,11877,11906,11940,11953,11988,12028,12095,12137,12214,12245,12290,12417,12456,12561,12605,12639,12652,12893,12899,12950,12952,13038,13039,13141,13159,13231,13245,13248,13278,13279,13322,13396,13566,13673,13680,13714,13751,13774,13858,13879,13885,13925,13964,14021,14096,14103,14107,14118,14162,14166,14210,14308,14382,14645,14906,14984,15033,15034,15158,15323,15397,15719,15745,15784,15821,15917,16086,16110,16132,16135,16205,16549,16552,16566,16568,16591,16601,16657,16662,16689,16713,16739,16805,16815,16826,16857,16874,16950,17051,17213,17406,17438,17461,17466,17514,17564,17587,17711,17745,17833,17877,17918,18033,18103,18255,18276,18486,18497,18673,18679,18793,18920,19007,19026,19113,19143,19148,19247,19371,19383,19534,19565,19628,19676,19679,19715,19725,19800,19809,19889,19901,20001,20022,20148,20189,20369,20399,20493,20633,20661,20667,20808,20872,21377,21526,21599,21770,21895,22098,22122,22227,22240,22354,22424,22511,22580,22664,22739,22922,22937,23026,23035,23056,23151,23240,23270,23301,23325,23493,23535,23603,23639,23642,23650,23831,23846,23912,24030,24123,24329,24374,24420,24455,24677,24687,24741,24775,24781,24869,24880,24906,24993,25004,25030,25102,25161,25204,25392,25419,25500,25593,25617,25666,25681,25719,25801,25874,25887,26040,26160,26191,26258,26405,26416,26424,26456,26599,26608,26731,26879,27087,27142,27245,27305,27338,27501,27502,27596,27660,27667,27687,27745,27763,27793,27848,28045,28079,28251,28318,28378,28383,28386,28432,28470,28515,28656,28774,28832,28842,28885,29036,29131,29171,29172,29206,29242,29264,29295,29303,29345,29405,29526,29582,29748,29824,29877,30058,30220,30328,30336,30345,30375,30496,30538,30675,30770,30791,31111,31124,31138,31207,31350,31374,31413,31472,31536,31657,31808,31884,31938,31987,32012,32019,32249,32336,32372,32453,32534,32557,32565,32624,32645,32859,33192,33220,33272,33302,33421,33452,33488,33750,33990,34041,34078,34197,34203,34278,34486,34610,34619,34690,34755,34844,34885,34971,35158,35182,35190,35217,35286,35301,35369,35373,35377,35493,35577,35638,35639,35657,35721,35779,35841,35851,35862,35900,36018,36102,36116,36120,36189,36291,36297,36528,36533,36584,36627,36640,36658,36760,36868,36935,37087,37182,37339,37547,37716,37922,38096,38098,38218,38261,38273,38525,38548,38603,38747,38764,38793,38797,38849,38918,38958,38966,38968,39147,39155,39215,39250,39268,39344,39376,39390,39421,39499,39675,39757,39762,39812,39853,39971,40206,40247,40274,40287,40356,40489,40504,40589,40883,40898,41006,41058,41076,41078,41124,41133,41139,41149,41167,41347,41398,41404,41484,41540,41647,41651,41717,41806,41867,41907,42099,42311,42360,42375,42426,42430,42466,42543,42564,42569,42646,42760,42825,42929,42932,43084,43186,43465,43542,43635,43681,44005,44033,44069,44089,44100,44179,44206,44245,44334,44413,44538,44570,44593,44756,44866,44916,44956,44975,45049,45099,45125,45213,45398,45443,45460,45462,45534,45745,45776,45803,45812,45864,46154,46185,46254,46261,46317,46333,46375,46410,46426,46497,46522,46603,46605,46613,46624,46635,46666,46711,46826,46842,46892,46971,46988,47104,47184,47191,47225,47341,47422,47494,47586,47602,47610,47663,47774,47828,47840,47903,47956,47958,48038,48097,48184,48363,48368,48428,48479,48512,48587,48735,48788,48793,48849,48902,48951,49061,49067,49074,49099,49119,49232,49235,49309,49437,49467,49552,49646,49682,49691,49701,49777,49829,49878,49907,49981,50017,50034,50045,50183,50259,50293,50311,50321,50362,50404,50418,50456,50535,50667,50766,50797,50837,50883,50958,51136,51176,51185,51190,51431,51441,51568,51591,51682,51698,51728,51877,51903,51959,51960,52052,52159,52294,52300,52387,52478,52513,52582,52598,52701,52723,52904,53259,53367,53411,53430,53439,53502,53525,53563,53732,53955,54011,54040,54084,54098,54309,54667,54689,54699,54908,54922,54989,55233,55239,55429,55592,55703,55708,55801,55813,55946,55979,56055,56101,56304,56393,56406,56436,56525,56637,56838,56840,56948,57044,57084,57141,57256,57263,57267,57279,57576,57629,57661,57725,57832,57833,57983,57990,58051,58155,58187,58257,58259,58263,58412,58448,58547,58586,58618,58633,58699,58745,58860,58872,58901,58994,59013,59060,59188,59207,59345,59422,59529,59556,59561,59600,59605,59629,59796,60092,60117,60164,60216,60383,60432,60461,60486,60567,60568,60623,60760,60860,60906,61106,61170,61200,61235,61250,61286,61295,61317,61511,61653,61733,61734,61753,61810,61880,62123,62154,62293,62297,62349,62493,62572,62632,62723,62741,62789,62802,62822,62963,63088,63229,63370,63456,63495,63547,63740,63764,63780,63784,63866,63873,63882,64044,64202,64217,64242,64340,64359,64413,64440,64492,64494,64674,64683,64690,64733,64752,64771,64824,64841,64880,64886,64949,65171,65231,65282,65331,65412,65491,65589,65603,65624,65648,65736,65808,65850,65857,66071,66200,66312,66388,66397,66400,66407,66636,66787,66931,67062,67263,67273,67290,67412,67428,67431,67470,67511,67596,67633,67739,67771,67800,67984,68074,68217,68298,68305,68494,68505,68578,68619,68701,68719,68896,68958,69148,69190,69214,69330,69390,69572,69633,69770,69909,69914,70046,70113,70253,70256,70347,70406,70445,70655,70680,70785,70799,70916,70929,70949,70988,71040,71104,71198,71302,71309,71457,71680,71693,71746,71819,71839,72041,72098,72249,72250,72280,72326,72351,72448,72470,72625,72699,72775,72800,72816,72840,72988,73100,73147,73172,73175,73274,73293,73363,73377,73399,73438,73444,73758,73764,73780,73781,73851,73862,74141,74274,74295,74299,74363,74520,74771,74880,75176,75290,75314,75372,75424,75531,75533,75564,75573,75811,75878,75882,75887,76025,76062,76165,76218,76309,76410,76462,76480,76483,76541,76559,76634,76768,76879,76903,76962,77093,77143,77218,77242,77256,77317,77361,77392,77536,77599,77637,77714,77755,77806,77881,78037,78078,78114,78132,78363,78678,78709,78715,78868,78990,79234,79264,79344,79416,79455,79472,79623,79648,79706,79728,79744,79821,79973,80159,80253,80309,80479,80491,80608,80714,80743,80757,80831,80889,80976,81073,81096,81297,81330,81334,81523,81550,81760,81772,81803,81869,82141,82266,82295,82376,82414,82508,82596,82609,82621,82678,82689,82850,82881,82919,82921,82960,82980,83056,83170,83389,83391,83478,83485,83631,83636,83697,83758,83788,83900,83951,84033,84099,84116,84153,84207,84219,84296,84324,84337,84338,84340,84558,84641,84650,84816,84893,84915,84952,84969,85335,85350,85401,85611,85695,85812,86039,86098,86127,86146,86195,86246,86407,86506,86513,86519,86569,86573,86598,86624,86716,86736,86840,86887,86923,87082,87092,87111,87157,87163,87237,87493,87717,87746,87787,87808,87959,88063,88064,88160,88209,88312,88335,88369,88392,88430,88502,88579,88608,88828,88861,88884,88891,89011,89026,89038,89188,89289,89315,89352,89369,89377,89438,89507,89537,89620,89769,89777,89846,89884,90049,90193,90247,90251,90266,90388,90434,90490,90526,90537,90591,90608,90623,90653,90682,90716,90864,91155,91202,91232,91372,91394,91430,91444,91468,91571,91672,91848,91889,91995,92019,92023,92085,92352,92381,92389,92595,92696,92710,92766,92903,92907,92969,92972,93021,93088,93094,93118,93138,93169,93304,93521,93570,93697,93847,93861,94038,94040,94065,94175,94176,94187,94251,94345,94350,94398,94412,94522,94717,94731,94752,94784,94803,94955,95083,95157,95183,95249,95278,95285,95295,95315,95330,95428,95657,95685,95698,96013,96114,96163,96195,96345,96390,96493,96557,96568,96666,96681,96716,96729,96750,96797,96859,96884,96922,97104,97112,97194,97240,97320,97346,97420,97448,97600,97611,97663,97724,97731,97922,98333,98347,98391,98452,98458,98702,98745,99080,99183,99260,99330,99476,99628,99723",
          "128,145,148,471,479,496,560,590,701,778,779,866,882,933,1009,1039,1274,1435,1482,1515,1522,1585,1701,1779,1961,1977,2047,2155,2190,2372,2491,2533,2541,2667,2814,2871,2875,2877,2932,2940,2998,3108,3128,3198,3321,3482,3788,3899,3915,4012,4030,4048,4150,4230,4329,4387,4618,4629,4656,4698,4701,4758,4785,4818,4840,4879,5047,5092,5110,5161,5341,5442,5644,5738,5767,5791,5810,5896,5942,5956,6001,6164,6188,6228,6384,6404,6483,6487,6743,6841,6858,6900,7127,7314,7334,7348,7366,7465,7562,7630,7717,8074,8184,8208,8343,8367,8417,8461,8556,8668,8675,8725,8868,8934,8975,9109,9280,9346,9556,9867,10033,10037,10057,10225,10264,10284,10299,10301,10315,10347,10380,10551,10559,10570,10662,10664,10697,10771,10860,10903,10979,11057,11086,11106,11241,11242,11260,11262,11265,11305,11319,11350,11401,11459,11485,11501,11715,11766,11786,11871,11891,11954,11994,11995,12001,12124,12253,12297,12329,12356,12465,12532,12555,12583,12597,12623,12631,12746,12959,12974,12982,13064,13164,13210,13269,13338,13357,13377,13378,13739,13769,13773,13878,13957,13959,14048,14067,14196,14248,14266,14391,14455,14465,14589,14641,14776,14858,14916,14949,15184,15288,15494,15725,15765,15847,16112,16127,16159,16177,16260,16333,16604,16704,16804,16860,16915,16939,17024,17043,17055,17216,17221,17223,17284,17288,17336,17383,17512,17563,17623,17807,17901,17931,17990,18026,18195,18495,18508,18598,18699,18946,18986,19041,19160,19162,19208,19465,19675,19698,19731,20055,20068,20282,20432,20439,20518,20559,20577,20716,20895,20906,21222,21250,21292,21392,21481,21482,21501,21504,21523,21570,21652,21696,21724,21841,21960,22002,22111,22147,22186,22221,22245,22270,22284,22344,22375,22450,22567,22591,22732,22753,22951,23000,23048,23097,23111,23342,23374,23444,23446,23508,23575,23606,23731,23905,24106,24269,24273,24459,24489,24525,24557,24665,24792,24893,24895,25005,25190,25254,25334,25364,25372,25474,25485,25647,25652,25729,25771,25903,26327,26581,26594,26691,26752,27002,27016,27201,27306,27396,27403,27516,27624,27642,27699,28080,28109,28355,28376,28388,28442,28467,28489,28620,28700,28817,28941,29139,29280,29359,29374,29458,29478,29611,29623,29640,29705,29863,29885,29920,30211,30466,30529,30621,30637,30696,30745,30857,31038,31141,31269,31294,31452,31484,31644,31646,31702,31776,31914,32027,32033,32105,32241,32255,32293,32310,32349,32631,32640,32673,32745,32820,32869,32908,32949,32956,32990,33054,33083,33099,33104,33153,33160,33301,33418,33557,33575,33576,33672,33691,33769,33903,33911,33913,33966,34195,34205,34254,34305,34389,34416,34443,34522,34655,34819,34835,34873,34949,34957,35010,35015,35091,35234,35269,35318,35330,35351,35362,35438,35743,35805,36002,36122,36201,36330,36457,36464,36540,36721,36740,36742,36746,36882,36927,37006,37099,37107,37126,37171,37195,37308,37479,37599,37602,37642,37671,37700,37752,37884,37988,38020,38039,38068,38241,38385,38447,38523,38532,38536,38745,38851,38934,38935,39113,39136,39367,39508,39535,39655,39706,39811,40242,40262,40382,40619,40673,40858,40874,40933,40937,40987,41030,41220,41229,41310,41432,41553,41559,41590,41617,41635,41669,41711,41739,41818,41995,41997,42102,42260,42274,42552,42556,42838,42851,42891,43066,43078,43224,43225,43229,43251,43300,43339,43923,43924,44170,44230,44286,44290,44370,44376,44394,44508,44548,44563,44647,44893,44962,45048,45137,45254,45357,45464,45623,45627,45738,45793,45883,45905,45975,46054,46058,46104,46141,46144,46177,46224,46467,46535,46665,46717,46728,46772,46899,46909,46911,46947,47033,47047,47207,47226,47227,47346,47405,47483,47528,47609,47639,47680,47697,47731,47875,47930,48068,48085,48260,48266,48290,48315,48473,48480,48570,48637,48944,48950,48989,49001,49016,49022,49223,49289,49353,49414,49422,49509,49547,49574,49711,49801,49825,49930,49953,49975,49984,50068,50150,50204,50317,50331,50348,50372,50668,50677,50767,50803,50858,50865,51006,51512,51533,51765,51882,52001,52008,52130,52220,52285,52316,52382,52399,52447,52547,52586,52637,52651,52659,52728,52827,52836,52853,53023,53247,53297,53432,53471,53535,53538,53568,53682,53718,53741,53792,53858,53890,53907,53922,53943,53967,53976,53985,54117,54188,54302,54376,54498,54602,54934,55006,55112,55131,55317,55388,55468,55517,55528,55559,55628,55900,55901,56099,56217,56248,56352,56414,56441,56517,56560,56576,56592,56619,56731,56897,57085,57087,57403,57418,57427,57451,57507,57598,57622,57740,57765,57852,57893,57934,57935,57960,58036,58103,58108,58184,58260,58316,58333,58383,58468,58480,58539,58580,58608,58656,58724,58806,58987,59154,59269,59320,59505,59541,59554,59567,59590,59610,59693,59800,60080,60108,60125,60186,60282,60332,60366,60424,60433,60529,60535,60625,60696,60739,60769,60859,61012,61027,61051,61064,61232,61330,61386,61404,61612,61709,61730,61775,61881,61910,62013,62043,62074,62139,62147,62156,62200,62207,62222,62241,62303,62348,62464,62472,62478,62529,62580,62621,62623,62684,62734,62777,62801,62888,63243,63770,63793,63904,63930,63974,63986,64094,64110,64116,64130,64233,64418,64592,64664,64872,64878,64901,64934,64935,64950,64955,65020,65279,65387,65402,65428,65433,65540,65565,65709,65756,65768,65810,65830,65925,65942,66067,66102,66194,66270,66315,66370,66573,66585,66717,66740,66839,66890,66984,67044,67190,67277,67416,67560,67601,67696,67703,67719,67875,68520,68606,68844,68920,68986,69041,69055,69073,69095,69134,69266,69424,69448,69485,69530,69665,69959,70049,70086,70093,70285,70420,70426,70739,70767,70840,70889,71035,71064,71072,71142,71253,71268,71458,71607,71712,71812,71849,71872,71914,71934,71996,72294,72492,72685,72768,72785,73022,73069,73177,73262,73354,73370,73448,73450,73465,73478,73504,73654,73885,73927,73971,74044,74220,74402,74468,74510,74533,74567,74694,74753,75027,75148,75204,75205,75256,75336,75401,75532,75623,75646,75648,75840,75877,75910,75965,75970,75997,76319,76463,76487,76498,76513,76516,77097,77193,77228,77399,77476,77574,77578,77789,77805,77962,78003,78017,78139,78239,78271,78528,78589,78768,78983,79009,79135,79193,79249,79442,79524,79585,79594,79601,80034,80070,80096,80126,80179,80185,80269,80334,80349,80353,80404,80637,80667,80668,80879,80930,80947,81031,81367,81368,81440,81534,81640,81794,81805,81821,81868,81909,81954,82026,82034,82049,82234,82698,82851,82877,83004,83103,83195,83326,83345,83452,83470,83500,83545,83554,83578,83600,83807,83961,83984,84011,84026,84069,84079,84121,84303,84387,84483,84524,84671,84796,84806,84978,85021,85134,85191,85258,85278,85298,85338,85362,85386,85503,85583,85601,85645,85660,85750,85780,85861,86087,86222,86309,86390,86394,86498,86508,86521,86563,86819,86907,86908,87000,87273,87433,87450,87558,87748,87762,87879,87964,87989,88013,88119,88165,88229,88304,88328,88399,88574,88607,88614,88675,88779,88790,88984,89085,89179,89205,89224,89235,89262,89300,89351,89368,89452,89533,89544,89660,89736,89945,89964,90201,90202,90228,90320,90440,90446,90473,90538,90584,90631,90884,90941,91146,91161,91267,91384,91500,91534,91629,91640,91677,91803,91930,92029,92069,92176,92214,92217,92236,92243,92342,92447,92454,92732,92912,92966,92977,93013,93023,93037,93154,93219,93381,93476,93504,93723,93750,93777,93797,93908,93910,93974,94054,94115,94266,94390,94464,94552,94583,94595,94813,94861,94900,94907,94932,95017,95132,95181,95342,95369,95427,95588,95671,95697,95735,95736,95787,96017,96028,96100,96144,96506,96510,96574,96731,96763,96775,96789,96833,96855,97115,97119,97161,97223,97227,97282,97541,97641,98011,98096,98123,98144,98166,98302,98345,98409,98499,98575,98593,98633,98647,98674,98691,98804,98810,98991,99032,99309,99514,99527,99528,99567,99578,99584,99694,99768,99906",
          "266,269,282,313,321,341,345,362,366,381,385,394,410,415,477,497,501,529,657,725,836,843,922,954,993,1027,1094,1241,1260,1280,1298,1343,1394,1612,5001,5085,5153,5175,5237,5279,5282,5294,5310,5336,5459,5571,5646,5652,5654,5694,5773,5815,5820,5825,5879,5927,5934,6043,6135,6149,6150,6204,6222,6526,6644,6678,6741,6797,6896,7000,7105,7146,7304,7357,7377,7596,7631,7654,8494,9007,9157,9218,9464,9744,9748,9799,9959,9968,10110,10454,11042,11593,12361,13172,15460,15779,17626,17995,18966,20506,20548,20846,21808,23478,26577,27981,33585,34275,40543,43942,50070,50148,50273,50301,50349,50464,50477,50536,50699,50718,50729,50915,50993,51011,51097,51168,51227,51407,51444,51482,51679,51706,51849,51937,51940,51942,52005,52185,52201,52250,52286,52336,52472,52602,52619,52734,52748,52757,52879,52885,52944,53026,53129,53286,53305,53353,53426,53602,53686,53727,53772,53844,53987,54125,54359,54409,54420,54444,54447,54457,54496,54591,54595,54709,55086,55244,55313,55365,55430,55495,55557,55635,55640,55649,55735,55904,55908,55929,55990,56007,56016,56038,56108,56155,56264,56354,56421,56490,56568,56654,56734,56914,56932,56938,57162,57240,57339,57369,57417,57424,57683,57698,57985,58012,58021,58178,58190,58196,58227,58407,58485,58551,58593,58622,58671,58786,58797,58888,58961,58982,59141,59275,59319,59350,59361,59481,59564,59623,59682,59845,59848,59853,60200,60243,60342,60354,60408,60463,60482,60675,60695,60749,60785,60902,60950,60974,60982,61131,61254,61306,61323,61401,61496,61554,61599,61676,61698,61799,61866,61952,62117,62327,62555,62631,62874,62899,62916,62923,62928,62946,62966,62985,63104,63195,63206,63320,63389,63457,63515,63573,63680,63699,63778,63926,63961,64067,64131,64267,64463,64548,64568,64788,64927,64944,64970,65075,65253,65352,65380,65463,65782,65899,65979,66004,66044,66072,66092,66148,66197,66262,66299,66337,66356,66563,66757,66854,66875,66979,66989,67158,67186,67193,67446,67664,67699,67843,67859,67860,67941,67958,68186,68297,68363,68574,68658,68694,68741,68815,68840,68850,68970,68995,69009,69217,69221,69287,69405,69444,69615,69765,69837,69904,69980,70203,70238,70354,70487,70514,70658,70670,70749,70801,70824,70915,71058,71209,71278,71283,71360,71387,71600,71795,71938,71978,71982,72059,72083,72116,72167,72201,72216,72272,72276,72350,72373,72392,72414,72418,72443,72558,72652,72780,72827,72896,72906,72908,72953,73101,73160,73251,73307,73405,73531,73567,73572,73613,73646,73659,73664,73690,73808,73908,73947,74081,74190,74214,74233,74270,74359,74558,74611,74730,74812,74823,74863,74960,74965,74968,74969,75048,75142,75274,75283,75461,75467,75469,75528,75577,75829,75941,76157,76185,76207,76476,76539,76667,76747,76857,77011,77031,77054,77194,77204,77222,77285,77425,77520,77707,77713,77883,77915,78023,78027,78123,78138,78142,78247,78346,78401,78432,78509,78511,78563,78667,78738,78973,79039,79163,79315,79435,79440,79529,79560,79563,79571,79574,79692,79737,79797,79800,79804,79806,79815,79817,79976,80073,80078,80102,80212,80257,80270,80343,80407,80420,80563,80719,80750,81024,81076,81091,81382,81384,81434,81468,81543,81589,81590,81676,81748,81843,81885,82037,82190,82237,82298,82319,82410,82533,82544,82692,82762,82871,83027,83052,83315,83453,83637,83673,83683,83713,83844,83846,83849,83914,83966,84076,84119,84294,84525,84605,84629,84687,85078,85112,85122,85359,85396,85410,85477,85482,85484,85576,85610,85628,85709,85722,85839,85909,85910,85922,86021,86091,86121,86417,86439,86782,86885,86890,86930,87169,87280,87332,87341,87383,87448,87595,87629,87641,88009,88017,88045,88121,88277,88353,88405,88441,88589,88644,88676,88886,89052,89128,89162,89238,89487,89613,89658,89851,90138,90330,90510,90529,90554,90579,90615,90656,90747,90785,90902,90972,91052,91090,91122,91176,91231,91250,91354,91556,91763,91843,91879,91888,91890,92175,92201,92211,92326,92446,92578,92801,92895,92935,93095,93159,93264,93338,93495,93637,93690,93711,93924,94106,94149,94259,94379,94410,94427,94435,94535,94606,94701,94703,94743,94763,94766,95033,95072,95089,95226,95411,95473,95840,96256,96389,96416,96444,96564,97046,97159,97315,97323,97412,97416,97421,97436,97518,97593,97696,97774,97810,97845,97871,97952,97969,98061,98393,98401,98516,98699,98775,98793,98862,98871,98895,98938,99005,99037,99241,99338,99361,99526,99576,99801,99852,99897,99968",
          "138,162,217,343,456,514,548,555,570,600,601,793,811,904,905,1140,1178,1202,1266,1271,1286,1431,1529,1637,1808,1871,1874,1882,2006,2111,2113,2143,2177,2287,2455,2499,2585,2662,2681,2780,2829,2927,3048,3049,3168,3285,3344,3371,3458,3493,3517,3653,3796,3844,3852,4099,4186,4222,4306,4320,4342,4399,4543,4608,4728,4781,4782,4820,4849,4985,4999,5002,5023,5055,5137,5320,5499,5554,5642,5752,5785,5802,5939,6127,6136,6269,6318,6332,6369,6437,6659,6688,6708,6762,6938,6942,6944,6972,6985,7176,7242,7394,7404,7605,7608,7737,7770,7856,8009,8022,8091,8140,8206,8223,8436,8453,8457,8463,8549,8558,8568,8657,8733,8753,8806,8849,8935,8967,9106,9182,9292,9317,9324,9366,9411,9477,9491,9536,9677,9720,9813,9903,9961,9967,10070,10081,10172,10309,10390,10439,10459,10515,10569,10577,10580,10586,10615,10686,10738,10795,10856,10932,10938,10957,11015,11117,11178,11184,11238,11369,11375,11394,11406,11444,11464,11510,11550,11600,11625,11631,11670,11709,11734,11815,11861,11903,12033,12041,12106,12239,12343,12363,12444,12469,12616,12659,12675,12849,12881,12900,12941,12963,12990,13025,13074,13114,13182,13230,13292,13311,13337,13415,13468,13475,13596,13681,13732,13759,13785,14012,14072,14088,14165,14192,14458,14540,14684,14707,15120,15275,15284,15376,15475,15508,15556,15677,15732,15768,15872,15933,16092,16271,16321,16405,16444,16578,16679,16762,16798,16890,16919,16920,17303,17411,17768,17837,18131,18221,18337,18466,18483,18574,18615,18680,18697,18723,18732,18759,18828,18865,18945,19047,19221,19326,19356,19446,19545,19848,19851,19917,19950,19981,20219,20225,20332,20342,20420,20537,20564,20570,20743,21095,21126,21359,21368,21405,21471,21499,21593,21614,21618,21644,21661,21716,21863,21872,22066,22089,22127,22129,22174,22216,22275,22309,22395,22428,22457,22462,22494,22615,22632,22674,22682,22694,22744,22763,22850,22880,22977,23057,23149,23191,23202,23255,23311,23433,23453,23506,23634,23832,23959,23975,24005,24199,24314,24322,24436,24577,24622,24667,24892,24970,25068,25158,25257,25282,25318,25360,25469,25650,25655,25687,25699,25706,25822,25845,25873,26087,26092,26269,26285,26386,26491,26518,26576,27054,27114,27152,27287,27316,27320,27352,27465,27475,27513,27551,27662,27834,27849,27865,27867,27955,27957,27967,28131,28276,28277,28328,28409,28687,28697,28777,28868,28897,28919,28937,28995,29133,29168,29395,29408,29698,29717,29724,29734,29753,29975,29997,30098,30430,30464,30530,30825,30848,30910,30967,30969,31047,31058,31131,31215,31272,31343,31422,31447,31561,31632,31699,31718,31777,31797,31799,31841,32009,32129,32145,32286,32430,32496,32511,32520,32698,32714,32799,32883,32886,32999,33078,33203,33449,33512,33605,33616,33679,33788,33888,33993,34006,34024,34046,34123,34222,34273,34400,34426,34448,34468,34492,34554,34611,34641,34691,34698,34747,34879,34904,34915,34945,35123,35197,35202,35336,35408,35473,35498,35519,35681,35683,35693,35716,35727,35885,35910,35924,35950,36079,36167,36230,36301,36369,36377,36434,36483,36499,36519,36554,36681,36732,36813,36946,37264,37317,37560,37612,37639,37658,37680,37701,37757,37795,37844,38009,38015,38132,38139,38182,38222,38279,38394,38441,38559,38606,38619,38833,38912,38976,39055,39075,39169,39177,39196,39260,39373,39425,39455,39700,39707,39760,39816,39925,39994,40014,40049,40053,40109,40170,40181,40224,40291,40298,40531,40693,40711,40789,40847,40914,41019,41186,41206,41260,41284,41307,41332,41518,41557,41564,41578,41687,41689,41703,41715,41728,41781,41811,41833,41948,42025,42149,42489,42493,42516,42651,43006,43038,43103,43126,43160,43272,43306,43367,43512,43763,43797,43807,43935,43952,43969,44177,44200,44329,44451,44504,44518,44520,44586,44727,45072,45177,45372,45385,45643,45716,45784,45797,45877,45983,46036,46069,46092,46163,46277,46280,46329,46409,46443,46498,46563,46658,46741,46812,46830,46837,47093,47107,47130,47214,47396,47428,47464,47538,47561,47711,47724,47778,47802,47831,47880,47902,47922,47947,48170,48227,48273,48370,48486,48489,48491,48564,48578,48674,48696,48745,48781,48782,48919,48988,49163,49280,49305,49310,49464,49531,49554,49581,49650,49732,49805,49868,50026,50087,50380,50381,50566,50579,50714,50756,50804,50841,50927,50937,51058,51090,51247,51384,51488,51594,51599,51624,51647,51670,51739,51797,51804,51819,52074,52209,52345,52481,52544,52616,52650,52660,52681,52716,52742,52747,52768,52798,52997,53019,53024,53055,53064,53151,53332,53350,53370,53508,53514,53532,53592,53634,53673,53683,53689,53692,53893,53924,53941,53954,53975,54203,54239,54408,54432,54516,54525,54636,54680,54755,54790,54816,54896,54986,54994,55026,55121,55135,55306,55363,55733,55792,55808,55829,55847,55850,56001,56095,56125,56132,56403,56429,56442,56452,56513,56550,56557,56621,56646,56661,56732,56782,56790,57200,57221,57243,57251,57333,57377,57378,57567,57609,57704,57851,57878,57920,58093,58119,58172,58273,58275,58308,58524,58614,58661,58663,58773,58780,58807,58893,58910,59032,59096,59222,59344,59509,59684,59857,60026,60068,60079,60202,60248,60316,60325,60346,60446,60508,60610,60633,60649,60651,60657,60689,60700,60773,60843,60844,60855,60911,60960,61061,61062,61091,61205,61343,61392,61421,61441,61452,61505,61523,61526,61528,61531,61550,61711,61736,61909,62005,62164,62378,62788,62799,62803,62809,62991,63168,63176,63178,63284,63401,63442,63446,63554,63592,63650,63655,63678,63750,63899,63953,63999,64024,64140,64206,64409,64472,64570,64648,64812,65087,65200,65308,65600,65636,65645,65693,65745,65753,65761,65807,65877,65892,66065,66108,66124,66141,66145,66205,66289,66317,66318,66462,66523,66779,66837,66910,67040,67078,67086,67126,67497,67669,67692,67740,67770,67824,67826,67828,67876,67879,67902,67915,68066,68097,68101,68266,68286,68290,68309,68409,68440,68496,68588,68652,69070,69288,69291,69295,69320,69328,69496,69559,69629,69668,69902,69948,69955,69981,70120,70150,70195,70263,70336,70486,70509,70612,70648,70654,70727,70911,70919,71008,71010,71368,71551,71628,71756,71944,71967,72013,72210,72315,72356,72361,72390,72437,72452,72455,72539,72617,72660,72694,72697,72868,72889,72916,73034,73171,73216,73220,73276,73322,73328,73341,73392,73474,73601,73879,74038,74102,74161,74194,74321,74327,74543,74895,75124,75126,75169,75262,75325,75343,75349,75403,75426,75442,75456,75494,75501,75621,75677,75689,75703,75706,75707,75713,75734,75747,75863,75948,76170,76387,76407,76460,76473,76781,76811,76836,76847,76892,77167,77182,77192,77264,77318,77321,77423,77523,77540,77571,77644,77686,77725,77766,77817,77974,78038,78047,78229,78250,78265,78282,78309,78387,78532,78557,78666,79170,79189,79202,79204,79305,79366,79498,79698,79719,80116,80173,80206,80310,80333,80397,80413,80498,80598,80646,80781,80875,81236,81247,81300,81311,81357,81455,81460,81639,81683,81688,81719,81944,82075,82112,82129,82217,82236,82874,82928,83069,83111,83208,83308,83369,83371,83396,83398,83436,83489,83491,83587,83634,83826,83828,84070,84107,84307,84321,84373,84484,84516,84603,85016,85127,85255,85301,85304,85313,85449,85554,85769,85772,85872,85967,86040,86076,86118,86139,86235,86306,86614,86668,86696,86706,86717,86743,86806,86808,86919,86932,86965,87043,87053,87141,87190,87260,87262,87310,87484,87487,87529,87547,87553,87589,87618,87635,87691,87707,87801,87802,87861,87923,87925,88127,88163,88195,88243,88256,88359,88454,88471,88592,88682,88703,88735,88759,88768,88871,88918,89024,89057,89073,89233,89336,89435,89454,89548,89577,89817,89818,89905,89915,89921,90087,90333,90336,90357,90427,90487,90506,90616,90618,90661,90700,90731,90868,90932,91068,91083,91104,91113,91138,91141,91172,91181,91276,91330,91391,91517,91572,91639,91669,91741,91990,92037,92145,92215,92341,92397,92560,92580,92643,92677,92758,92761,92872,92877,93032,93055,93064,93139,93184,93344,93464,93481,93593,93605,93621,93650,93654,93673,93687,93768,94015,94189,94536,94540,94624,94649,94655,94800,94927,94956,95009,95084,95143,95207,95255,95407,95474,95508,95547,95592,95608,95615,95641,95658,95801,95901,95974,96036,96161,96192,96336,96530,96674,96879,96959,96981,97247,97569,97621,97647,97727,97758,97799,97899,97908,97923,97930,97932,97955,97976,98091,98170,98187,98515,98683,98770,98812,99044,99069,99284,99516,99728,99732,99780,99883,99982",
          "224,236,252,259,374,500,825,826,837,966,1019,1031,1041,1127,1173,1226,1240,1296,1333,1439,1587,1600,1890,2498,2566,2649,2830,2947,3207,3329,3520,3572,3589,3618,3654,3741,3881,3989,4086,4258,4290,4305,4367,4402,4422,4549,4556,4600,4846,4892,4951,4974,5003,5075,5235,5417,5625,5676,5858,5983,5994,6208,6259,6383,6453,6478,6508,6519,6525,6532,6553,6605,6765,6795,6818,6829,7010,7019,7129,7149,7159,7323,7371,7372,7383,7421,7741,7875,7910,7965,8088,8203,8450,8473,8573,8589,8592,8601,8765,8938,8958,8995,9020,9191,9216,9257,9503,9538,9653,9673,9691,9723,9743,9755,9884,9895,9953,9954,10078,10093,10122,10272,10317,10333,10548,10936,10951,10999,11006,11397,11614,11827,11828,11853,11888,11930,11938,12420,12464,12535,12545,12740,12758,12775,12948,12999,13041,13060,13165,13271,13328,13490,13671,13712,13937,14026,14057,14117,14139,14287,14352,14427,14667,15045,15129,15183,15205,15454,15536,15558,15658,15710,15802,15900,15964,16059,16158,16296,16429,16841,16952,16953,17224,17270,17275,17277,17292,17294,17517,17534,17570,17585,17664,17796,17875,17961,18071,18099,18116,18133,18265,18320,18398,18435,18577,18619,18647,18718,18731,18818,18835,18890,18903,19068,19139,19154,19176,19259,19333,19394,19536,19587,19730,19787,19796,19949,19986,20072,20107,20137,20152,20229,20344,20414,20440,20563,20795,20829,21001,21171,21172,21537,21886,22084,22276,22277,22313,22499,22527,22550,22581,22648,22768,22780,22916,22967,22996,23372,23436,23449,24138,24180,24195,24203,24720,24787,25000,25082,25450,25499,25690,25720,25861,25932,25969,26060,26144,26159,26187,26435,26463,26483,26522,26540,26712,26759,26762,26772,26784,26830,26836,27131,27308,27351,27399,27534,27665,27697,27817,27840,28018,28034,28036,28140,28191,28841,28863,28931,28934,28959,29051,29057,29124,29215,29228,29267,29331,29358,29378,29502,29518,29559,29563,29573,29661,29743,29887,29902,30043,30066,30125,30188,30198,30450,30493,30499,30567,30785,30836,30986,31156,31431,31535,31663,31720,31743,31871,31873,32015,32092,32115,32239,32258,32342,32351,32466,32684,32686,32699,32920,32958,32969,32977,33050,33068,33097,33320,33384,33622,34088,34423,34456,34483,34488,34627,34825,34830,34955,35261,35292,35299,35327,35360,35610,35627,35705,35918,35965,35968,36207,36211,36222,36281,36284,36788,36859,36944,37103,37166,37217,37228,37305,37443,37512,37702,37737,37798,37843,37929,37937,38047,38113,38123,38137,38236,38271,38343,38649,38761,39061,39118,39181,39664,39989,40148,40171,40271,40288,40351,40392,40471,40535,40893,40924,40939,40963,41017,41066,41439,41883,41894,41908,42173,42254,42313,42317,42453,42467,42548,42751,42880,42990,43227,43269,43327,43359,43375,43468,43473,43480,43490,43889,44085,44143,44164,44292,44325,44338,44419,44491,44501,44529,44595,44629,44696,44782,44853,44958,45050,45216,45287,45369,45459,45546,45570,45576,45693,45765,45775,45783,45787,45840,45849,45894,46033,46066,46081,46278,46290,46364,46472,46602,46616,46629,46690,46704,46907,47168,47190,47198,47205,47242,47309,47377,47429,47435,47546,47558,47733,47814,47888,47889,47952,47994,48027,48050,48079,48093,48102,48169,48216,48249,48251,48253,48408,48419,48429,48784,48809,48943,48954,49117,49195,49208,49233,49416,49441,49710,49979,50161,50308,50309,50504,50524,50583,50656,50785,50838,50995,51009,51039,51260,51288,51357,51358,51531,51557,51617,51725,51792,51812,51999,52020,52025,52035,52078,52297,52352,52396,52414,52730,52812,52831,52874,52897,52973,52986,53311,53372,53396,53407,53615,53714,53866,54021,54120,54177,54244,54328,54329,54334,54491,54508,54557,54714,54841,54923,54962,54979,55071,55146,55370,55617,55702,55755,55844,55914,55996,56004,56057,56074,56137,56194,56257,56370,56380,56416,56428,56563,56716,56717,56772,56777,56987,57015,57217,57503,57531,57648,57718,57867,57962,58025,58156,58166,58208,58309,58507,58690,58841,58864,58922,58985,59110,59111,59152,59428,59495,59696,59698,59778,59841,60107,60233,60662,60716,60779,60830,60832,60926,61208,61380,61454,61456,61559,61669,61877,61884,61936,61970,62034,62064,62179,62193,62273,62302,62428,62562,62706,62924,62931,63043,63094,63119,63396,63464,63471,63593,63689,63867,63909,63919,63968,64035,64093,64097,64171,64281,64360,64456,64529,64662,64682,64704,64820,64874,64956,65102,65609,65723,65852,65855,65894,65924,65931,65968,66075,66123,66471,66501,66770,66859,66985,67035,67099,67181,67357,67568,67611,67660,67713,67724,67727,67748,67864,67885,67901,68031,68178,68510,68627,68635,68656,68682,68746,68757,68909,69253,69453,69573,69874,69898,69966,70041,70088,70103,70192,70350,70367,70622,70631,70692,70815,70860,71004,71059,71121,71214,71238,71343,71367,71487,71541,71570,71762,71766,71825,72182,72183,72253,72316,72342,72360,72563,73031,73070,73099,73164,73521,73545,73643,73668,73783,73795,73822,73878,73960,73989,74109,74280,74374,74390,74399,74435,74466,74474,74624,74791,74877,74974,75190,75323,75412,75559,75628,75842,75898,75994,76083,76163,76195,76383,76425,76432,76594,76661,76663,76756,76908,77086,77150,77305,77527,77573,77702,77716,77821,77920,78020,78048,78086,78110,78778,78824,79004,79137,79176,79187,79222,79232,79283,79323,79399,79425,79569,79890,79963,80140,80152,80275,80422,80504,80517,80573,80590,80640,80713,80754,80767,80796,80862,81058,81108,81126,81129,81175,81270,81275,81375,81425,81499,81501,81527,81569,81874,81924,81957,81960,81991,82000,82001,82008,82014,82289,82318,82479,82492,82503,82561,82637,82721,82887,82930,83025,83066,83179,83180,83210,83238,83414,83423,83466,83628,83669,83680,83688,83799,83809,83821,83923,83977,83982,84222,84230,84235,84288,84297,84453,84498,84586,84861,85294,85353,85365,85575,85622,85697,85951,86334,86638,86700,86731,86884,86909,87097,87164,87211,87238,87293,87302,87311,87367,87563,87572,87692,87905,88030,88061,88144,88343,88516,88578,88613,88627,88628,88929,89042,89078,89093,89266,89347,89366,89393,89569,89618,89892,89903,90002,90067,90218,90243,90270,90281,90562,90666,90688,90834,90860,90872,90892,91003,91221,91347,91433,91499,91622,91731,91789,91800,91916,92039,92162,92185,92510,92718,92944,93195,93242,93292,93308,93420,93501,93627,93850,93856,93879,93933,94009,94099,94472,94518,94575,94668,94826,94848,94874,95103,95175,95250,95267,95277,95286,95290,95452,95576,95625,95665,95788,95803,95819,95832,95906,96030,96034,96116,96180,96217,96240,96378,96509,96567,96862,97030,97102,97264,97292,97327,97414,97514,97580,97747,97768,97784,97789,97806,97826,97991,97999,98053,98080,98094,98219,98334,98352,98565,98604,98724,98880,99027,99067,99092,99096,99115,99148,99202,99384,99422,99430,99462,99499,99664,99762,99794",
          "124,197,296,310,422,488,578,727,788,803,994,1016,1156,1160,1176,1251,1355,1367,1410,1481,1658,1698,1776,1781,1995,2089,2288,2336,2391,2489,2500,2532,2609,2616,2784,2845,2938,2943,2952,3040,3111,3235,3238,3252,3309,3402,3540,3581,3610,3665,3738,3816,3839,3882,3950,3967,4046,4253,4270,4357,4457,4536,4715,5044,5086,5120,5157,5222,5223,5318,5495,5498,5537,5560,5812,5936,5957,6130,6279,6406,6426,6476,6543,6583,6602,6710,6816,6926,6950,6955,7102,7154,7261,7510,7513,7564,7615,7739,7751,8056,8092,8144,8177,8181,8366,8387,8419,8640,8647,8777,8941,9073,9083,9138,9303,9363,9423,9506,9563,9610,9618,9714,9718,9739,9749,9768,9821,9833,9879,9915,10130,10149,10155,10406,10485,10878,10891,10901,11039,11073,11080,11137,11347,11443,11472,11473,11488,11640,11858,11868,11959,11989,12035,12047,12069,12114,12117,12139,12206,12227,12459,12496,12763,12812,12926,12996,13006,13145,13161,13275,13353,13422,13461,13753,13796,13884,13900,13906,13998,14078,14169,14362,14380,14386,14411,14457,14482,14780,14820,14865,14881,14999,15195,15206,15207,15266,15344,15407,15418,15552,15603,15609,15628,15680,15693,15761,15858,15890,15945,15948,16001,16105,16117,16175,16226,16229,16278,16298,16315,16373,16374,16390,16392,16528,16581,16593,16598,16614,16721,16737,16877,16883,16917,17062,17120,17212,17241,17347,17458,17464,17538,17586,17654,17799,17819,17888,17998,17999,18066,18104,18138,18277,18308,18380,18419,18453,18504,18510,18642,18657,18681,18746,19013,19187,19255,19261,19354,19361,19604,19608,19616,19625,19721,20359,20406,20422,20425,20561,20567,20579,20731,20776,20893,20941,21008,21087,21101,21117,21118,21183,21406,21411,21496,21543,21554,21579,21625,21630,21722,21753,21788,21805,21979,22156,22193,22218,22286,22391,22476,22502,22503,22548,22646,22662,22750,22772,22851,22968,23109,23169,23234,23292,23356,23362,23435,23447,23486,23531,23607,23640,23722,23814,23823,23849,23993,24038,24116,24140,24165,24182,24230,24256,24266,24287,24345,24660,24758,24847,24902,24935,24955,25015,25119,25186,25325,25356,25506,25578,25627,25661,25784,25876,25889,25941,26001,26017,26041,26044,26146,26156,26625,26667,26694,26707,26770,26802,26839,26871,26898,26957,27036,27056,27058,27116,27138,27211,27411,27591,27706,27711,27732,27876,27950,28043,28164,28174,28282,28523,28532,28535,28603,28664,28743,28967,29075,29146,29301,29363,29445,29462,29484,29551,29552,29596,29616,29675,29723,29754,29765,29812,29834,29927,29999,30132,30253,30716,30811,30878,30987,31074,31199,31223,31245,31320,31330,31367,31425,31429,31448,31506,31541,31557,31636,31920,31993,32050,32082,32088,32152,32198,32219,32363,32417,32486,32491,32495,32572,32580,32642,32785,32919,32924,33011,33166,33193,33298,33300,33552,33554,33683,33689,33778,33793,33814,33822,33864,33929,33986,34010,34059,34199,34283,34308,34311,34335,34337,34349,34430,34441,34481,34721,35001,35036,35119,35323,35392,35472,35582,35623,35629,35649,35767,35974,36172,36182,36340,36376,36389,36418,36455,36491,36608,36688,36889,36928,36938,36974,37065,37109,37116,37148,37177,37183,37205,37261,37295,37369,37497,37562,37667,37876,37892,37913,37925,37992,38100,38109,38230,38333,38371,38461,38612,38711,38938,39037,39095,39151,39199,39205,39241,39315,39331,39382,39413,39463,39513,39599,39740,39796,39817,39855,40104,40383,40391,40412,40422,40444,40459,40637,40756,40903,41122,41127,41191,41204,41247,41306,41325,41331,41506,41513,41582,41597,41626,41759,41801,41881,41901,41916,41971,42068,42213,42358,42372,42373,42644,42772,42795,42852,42957,43021,43094,43112,43199,43214,43277,43304,43433,43449,43471,43497,43507,43573,43870,43875,43897,43966,44147,44262,44365,44406,44465,44575,44687,44821,44862,44894,44988,45070,45136,45147,45210,45364,45383,45471,45494,45565,45580,45662,45746,45757,45763,45794,45913,45919,46008,46031,46035,46121,46169,46208,46222,46243,46357,46593,46710,46797,46846,46879,46916,46946,46999,47156,47204,47236,47330,47336,47441,47450,47526,47550,47572,47606,47749,47762,47769,47801,47834,47911,47925,47968,48543,48588,48636,48903,48974,49136,49166,49190,49203,49204,49236,49246,49284,49344,49349,49476,49596,49623,49819,49937,50190,50191,50300,50341,50356,50371,50431,50513,50591,50594,50706,50761,50901,51032,51073,51112,51134,51145,51225,51319,51340,51459,51692,51710,51729,51741,51895,51949,51996,52115,52165,52221,52231,52335,52410,52603,52617,52712,52732,52744,52959,53093,53095,53457,53472,53574,53617,53633,53650,53665,53669,53687,53867,53940,54049,54063,54114,54131,54190,54195,54265,54285,54418,54431,54594,54774,54786,54792,54801,54807,54876,54895,54916,54929,54940,54963,55033,55118,55219,55281,55284,55318,55342,55358,55391,55410,55439,55449,55483,55487,55503,55567,55727,55830,55981,56024,56083,56138,56201,56260,56289,56349,56376,56437,56556,56565,56639,56760,56800,57155,57174,57184,57305,57560,57596,57678,57974,58199,58206,58264,58297,58354,58467,58530,58560,58778,58854,58902,58905,58968,58993,58995,59142,59175,59223,59229,59351,59372,59448,59485,59530,59633,59647,59829,59856,60058,60062,60065,60072,60198,60293,60297,60363,60516,60552,60632,60642,60647,60712,60824,61022,61082,61092,61095,61124,61160,61331,61368,61371,61418,61609,61847,61931,61987,62113,62289,62575,62883,62935,63111,63165,63233,63249,63323,63490,63517,63539,63620,63623,63738,63827,63861,63967,63992,64068,64085,64086,64126,64213,64232,64238,64313,64345,64488,64523,64540,64563,64966,65069,65100,65386,65421,65430,65527,65647,66049,66069,66179,66251,66360,66376,66395,66518,66528,66701,66746,67122,67220,67394,67447,67486,67679,67863,67951,68013,68470,68544,68579,68622,68698,68742,68789,68811,68899,69140,69182,69201,69206,69286,69306,69468,69751,69859,69868,69928,69999,70096,70136,70219,70220,70382,70458,70464,70548,70564,70699,70751,70884,70994,71012,71026,71034,71259,71352,71496,71557,71561,71589,71621,71765,71799,71857,71886,71905,71972,72074,72093,72235,72355,72436,72662,72747,72928,72948,72957,72961,73010,73386,73402,73495,73568,73586,73752,73756,73827,73828,73892,73933,73946,74027,74087,74156,74239,74395,74492,74495,74540,74563,74715,74769,74786,74844,74923,75061,75234,75289,75333,75348,75428,75523,75696,75765,75816,75834,75839,75921,75987,76199,76264,76314,76334,76361,76444,76538,76544,76563,76587,76637,76692,76934,76960,77067,77085,77102,77122,77125,77127,77238,77266,77409,77472,77541,77566,77567,77604,77632,77740,77742,77804,77837,77884,77921,78026,78134,78137,78164,78240,78270,78279,78331,78367,78397,78419,78575,78577,78694,78856,79131,79167,79199,79219,79254,79260,79453,79468,79656,79743,79840,79941,79948,79961,79962,80060,80191,80382,80469,80482,80580,80798,80836,80870,81010,81095,81104,81144,81223,81340,81452,81495,81572,81583,81657,81833,82126,82477,82618,82660,82699,82731,82788,82844,82954,83159,83327,83338,83403,83428,83624,83670,83747,83754,83780,83843,83892,83956,83959,83983,84045,84055,84113,84256,84286,84344,84366,84370,84425,84741,84769,84801,85052,85109,85157,85345,85479,85505,85644,85688,85818,85819,86041,86064,86074,86095,86113,86290,86388,86606,86689,86737,86773,86809,86911,87036,87047,87067,87138,87149,87391,87539,87822,87968,88060,88081,88107,88182,88207,88220,88316,88346,88468,88469,88491,88504,88728,88734,88813,88874,88876,88882,89062,89067,89293,89341,89428,89628,89728,89743,89834,89883,89906,90036,90149,90154,90319,90622,90627,90671,90709,90772,90942,90961,91035,91098,91453,91506,91699,91719,91734,91747,91805,91847,91880,91985,92003,92018,92034,92065,92167,92171,92256,92282,92311,92327,92340,92423,92502,92505,92565,92574,92638,92660,92727,92737,92772,92950,92960,93004,93010,93018,93164,93272,93314,93333,93422,93447,93546,93648,93718,93771,93855,94048,94173,94177,94307,94618,94642,94659,94660,94947,95091,95166,95178,95275,95300,95346,95502,95513,95527,96038,96092,96200,96373,96381,96461,96521,96705,96958,96968,97058,97098,97137,97329,97359,97608,97643,97645,97652,97756,97781,97783,97843,98006,98076,98168,98185,98230,98263,98264,98427,98469,98519,98536,98757,98805,98815,99031,99281,99469,99480,99498,99511,99574",
          "69,140,238,279,302,700,710,886,892,918,946,1083,1157,1210,1231,1300,1339,1505,1607,1623,1666,1735,1771,1937,2121,2228,2251,2573,2596,2761,2778,2779,2812,2856,2870,2991,3051,3133,3209,3365,3374,3563,3592,3633,3678,3820,3822,3856,4004,4331,4345,4355,4408,4565,4566,4598,4613,4651,4665,4814,5131,5397,5451,5609,5664,5780,5784,5845,5854,5933,6041,6064,6163,6237,6250,6290,6323,6362,6386,6493,6498,6656,6670,6763,6823,6856,6884,6898,6998,7031,7200,7411,7438,7471,7669,7693,7759,7837,7900,7908,7938,7939,7947,7958,7970,7990,8008,8010,8060,8257,8357,8370,8394,8407,8516,8679,8732,8751,8772,8807,8829,8843,8908,8923,9063,9087,9242,9307,9319,9431,9489,9724,9751,9786,9934,9948,9999,10043,10049,10121,10259,10289,10311,10363,10535,10543,10557,10671,10724,10760,10786,10877,10893,10963,11058,11096,11303,11309,11481,11483,11499,11537,11613,11674,11688,11820,11971,12015,12115,12165,12198,12350,12448,12468,12486,12641,12708,12743,12805,12809,12861,12874,12891,12980,12985,12994,12997,13009,13022,13027,13050,13180,13241,13326,13404,13447,13495,13503,13560,13583,13738,13758,13916,14009,14100,14222,14233,14343,14364,14395,14450,14453,14617,14659,14662,14718,14749,14756,15009,15143,15386,15473,15481,15519,15549,15665,15705,15731,15758,15829,15845,15894,16004,16070,16123,16267,16268,16294,16314,16355,16361,16508,16636,16882,16908,16955,17088,17144,17200,17476,17738,17759,17856,17887,18027,18040,18047,18141,18212,18326,18386,18422,18455,18546,18611,18640,18658,18705,18736,18951,19249,19460,19512,19543,19702,19956,20058,20077,20120,20170,20224,20235,20434,20487,20515,20627,20763,20796,20977,21108,21111,21271,21326,21488,21582,21780,21802,21834,21867,21996,22272,22318,22355,22467,22470,22505,22552,22712,22743,22908,22927,23020,23103,23185,23238,23406,23428,23523,23526,23546,23547,23562,23748,23825,23944,24011,24175,24222,24382,24387,24536,24602,24791,24824,24932,25079,25202,25215,25280,25287,25321,25357,25463,25482,25610,25651,25731,25765,25816,26085,26374,26385,26406,26436,26485,26612,26705,26737,26744,26755,26764,26866,26884,26952,26962,27155,27194,27207,27357,27371,27380,27383,27459,27523,27533,27749,27788,27823,27833,27871,27920,28066,28085,28154,28158,28185,28266,28283,28305,28369,28587,28737,28807,28809,29052,29197,29804,29988,30000,30089,30214,30295,30351,30384,30403,30477,30480,30500,30659,30747,30761,31017,31267,31440,31482,31492,31546,31669,31750,31840,31872,31907,31929,32041,32065,32133,32168,32232,32248,32304,32360,32393,32432,32484,32532,32578,32692,32725,32801,32806,32966,33004,33021,33093,33122,33271,33343,33372,33373,33417,33501,33659,34030,34033,34057,34095,34103,34148,34323,34405,34406,34418,34501,34606,34847,34882,34982,35003,35018,35050,35059,35067,35077,35337,35437,35458,35466,35526,35552,35570,35585,35688,35853,35895,35935,35997,36003,36136,36145,36238,36424,36513,36590,36618,36662,36675,36733,36756,36763,36855,37014,37042,37137,37179,37218,37245,37387,37404,37574,37633,37682,37941,38016,38133,38193,38304,38306,38314,38321,38380,38395,38425,38436,38477,38601,38678,38694,38773,39126,39223,39253,39301,39313,39522,39710,39873,40004,40008,40030,40085,40102,40122,40126,40166,40297,40310,40315,40320,40326,40370,40429,40649,40657,40681,40766,40793,40806,40835,40836,40958,41045,41189,41243,41262,41356,41423,41551,41616,41630,41649,41698,41723,41817,41831,41887,41915,41962,42181,42382,42448,42462,42482,42793,42906,42940,43035,43096,43127,43172,43201,43218,43260,43342,43409,43435,43533,43553,43564,43595,43600,43708,43784,43816,43828,43854,43871,43894,43957,43988,44076,44088,44121,44192,44267,44379,44477,44620,44640,44670,44725,44753,44800,44987,45047,45069,45178,45232,45251,45333,45538,45569,45646,45704,45707,45722,45724,45834,45890,45928,46098,46166,46237,46293,46312,46330,46451,46471,46682,46904,46930,47007,47039,47095,47135,47271,47360,47462,47560,47619,47679,47855,47928,48121,48235,48331,48520,48609,48729,48884,48935,48938,49044,49094,49109,49124,49135,49170,49210,49221,49354,49398,49438,49453,49484,49593,49674,49849,49864,49879,49986,50156,50244,50254,50256,50330,50335,50466,50469,50589,50651,50702,50773,50798,50823,50835,50930,51203,51285,51297,51338,51538,51761,51825,51844,51847,51867,51876,51973,52036,52065,52099,52119,52163,52164,52244,52313,52324,52392,52395,52432,52614,52795,52802,52873,52949,52972,53031,53257,53294,53340,53377,53521,53775,53857,53877,53965,54218,54220,54252,54266,54558,54581,54611,54645,54647,54655,54785,54871,54906,54959,54996,55027,55191,55197,55316,55529,55531,55582,55603,55613,55722,55810,55846,56136,56158,56363,56433,56705,56710,56809,56816,56849,56891,56901,57119,57156,57189,57280,57331,57400,57421,57476,57491,57495,57538,57647,57651,57747,57754,57880,57964,58068,58246,58304,58318,58421,58478,58575,58708,58953,58978,58981,59014,59017,59051,59095,59197,59208,59272,59416,59533,59540,59680,59723,59789,59973,60006,60032,60073,60231,60386,60488,60500,60544,60799,60917,61031,61275,61287,61355,61468,61506,61628,61712,61781,61948,61992,62152,62232,62346,62364,62380,62392,62404,62520,62681,62703,62772,62876,62891,62911,63035,63064,63150,63417,63646,63647,63658,63760,63997,64106,64257,64268,64278,64320,64487,64590,64652,64665,64967,65144,65365,65372,65417,65477,65532,65637,65672,65675,65727,65870,66234,66355,66375,66378,66520,66627,66755,66947,66980,66993,67038,67146,67185,67237,67301,67337,67510,67524,67540,67583,67734,67830,68036,68076,68146,68184,68355,68433,68441,68458,68509,68513,68559,68825,68916,68955,68973,68974,69011,69029,69079,69133,69161,69199,69311,69343,69362,69514,69554,69779,69784,69786,69834,69946,70028,70310,70370,70513,70635,70651,70653,70729,70734,70737,70926,70952,71050,71109,71130,71136,71160,71172,71265,71281,71328,71479,71500,71555,71727,71764,71876,71885,71968,71977,72035,72067,72121,72192,72301,72368,72391,72449,72479,72489,72512,72529,72659,72859,72910,73009,73052,73558,73779,73804,73925,74065,74216,74264,74419,74442,74482,74584,74661,74751,74775,74795,74853,75015,75024,75141,75384,75397,75437,75487,75499,75627,75632,75705,75756,75771,76039,76048,76078,76134,76151,76256,76258,76266,76312,76439,76442,76497,76552,76626,76724,76788,77010,77142,77411,77512,77623,77682,77745,77825,77865,77879,77885,78081,78167,78286,78418,78473,78494,78504,78572,78590,78698,78905,78915,78942,79045,79053,79114,79190,79510,79520,79538,79638,79742,79841,79935,79954,80031,80088,80158,80245,80265,80273,80306,80364,80400,80610,80690,80733,80852,80880,80935,81248,81342,81349,81354,81420,81451,81636,81735,81744,81828,81887,81890,81894,81940,82078,82253,82269,82271,82311,82323,82328,82436,82482,82483,82568,82745,82764,82773,82789,83132,83137,83175,83241,83269,83271,83278,83291,83352,83377,83405,83464,83560,83630,83796,83924,84006,84031,84115,84130,84258,84444,84455,84530,84568,84838,84990,85071,85275,85443,85777,85794,85961,85989,86077,86223,86286,86311,86312,86344,86348,86365,86373,86393,86451,86541,86572,86698,86726,86742,86745,86768,86791,86800,86802,86854,86945,86974,86987,87220,87337,87381,87463,87496,87593,87642,87770,87809,87849,87869,87877,87914,87960,88032,88084,88171,88186,88370,88424,88453,88654,88797,88930,89178,89277,89477,89505,89648,89722,89725,89913,90075,90231,90329,90413,90491,90555,90570,90798,91071,91105,91184,91222,91316,91358,91414,91493,91523,91609,91646,91833,91942,92000,92015,92111,92188,92388,92533,92559,92655,92673,92753,92845,93083,93100,93158,93233,93251,93282,93353,93505,93510,93652,93694,94113,94121,94145,94226,94273,94408,94707,94881,94887,94941,95022,95063,95203,95262,95363,95366,95539,95562,95649,95656,95758,95774,95866,95870,95887,96094,96177,96312,96413,96422,96438,96536,96569,96611,96670,96697,96732,96764,96787,96788,96904,96974,97313,97325,97437,97457,97482,97680,97734,97759,98107,98161,98232,98286,98376,98405,98419,98435,98440,98670,98716,98782,98853,98960,98971,99021,99046,99053,99289,99368,99376,99517,99577,99609,99662,99681,99683,99695,99779,99784,99913",
          "179,292,324,373,439,694,702,772,799,819,848,911,943,957,1005,1052,1161,1500,1525,1599,1747,1772,1806,1930,2040,2250,2328,2371,2375,2481,2517,2557,2632,2682,2742,2819,2970,3031,3140,3225,3249,3264,3560,3692,3696,3726,3765,3838,3895,3900,4015,4018,4028,4041,4066,4072,4075,4081,4108,4188,4200,4214,4321,4347,4394,4440,4473,4499,4579,4607,4778,4897,4968,4975,5057,5126,5224,5368,5399,5434,5558,5647,5703,5850,5862,5952,6007,6058,6093,6126,6134,6157,6376,6413,6448,6484,6529,6633,6684,6703,6719,6736,6850,6943,6954,6967,7315,7321,7349,7476,7508,7521,7563,7574,7620,7665,7735,7790,7896,8065,8204,8247,8327,8331,8344,8433,8462,8535,8618,8645,8743,8781,8832,8854,8870,8994,9012,9082,9101,9325,9334,9443,9523,9543,9573,9600,9603,9609,9702,9721,9798,9890,10291,10736,10766,10818,10853,10928,11196,11551,11569,11676,11735,11842,11960,12133,12184,12209,12228,12247,12306,12313,12319,12357,12393,12457,12607,12618,12629,12696,12702,12730,12778,12859,12940,12953,13042,13061,13162,13264,13343,13463,13479,13545,13776,13963,13990,14087,14147,14215,14216,14231,14307,14433,14508,14518,14582,14759,14913,15008,15144,15179,15293,15338,15350,15433,15503,15529,15775,15777,15824,15871,15927,16090,16279,16312,16391,16463,16467,16686,16725,16745,16758,16907,16940,16960,16979,17022,17050,17054,17153,17154,17161,17260,17463,17584,17594,17707,17717,17758,17789,17831,17861,17880,17904,17926,17988,17991,18030,18092,18098,18285,18420,18467,18526,18561,18567,18665,18687,18726,18779,18782,18838,18875,18975,19226,19310,19425,19737,19755,19941,19957,20046,20093,20169,20246,20431,20450,20472,20502,20634,20693,20710,20719,20726,20842,20864,20870,20912,21009,21133,21229,21255,21268,21376,21434,21463,21607,21620,21768,21919,21930,21937,22019,22229,22247,22325,22353,22367,22373,22407,22537,22719,22797,22901,22926,22935,23116,23130,23139,23143,23370,23490,23628,23653,23657,23840,23861,23862,23869,24008,24168,24284,24299,24352,24363,24367,24408,24538,24601,24719,24750,24975,25059,25071,25114,25206,25452,25479,25491,25571,25613,25623,25642,25691,25723,25815,25935,25944,25982,26010,26096,26532,26649,26681,26695,26777,26828,26885,27254,27267,27388,27395,27401,27492,27552,27590,27633,27727,27739,27842,27850,27994,28012,28030,28098,28250,28271,28286,28358,28394,28426,28447,28513,28642,28740,28840,28998,29070,29112,29119,29203,29216,29336,29555,29580,29653,29689,29760,29870,29930,29951,30014,30149,30277,30291,30299,30333,30702,30804,30810,30815,30866,30902,30928,30940,31025,31100,31210,31264,31331,31332,31358,31396,31410,31456,31503,31527,31563,31650,31729,31746,31835,31932,32068,32074,32091,32189,32196,32234,32325,32506,32533,32539,32611,32629,32707,32741,32877,33096,33124,33207,33219,33273,33381,33527,33604,33748,33763,33806,33830,33835,33882,33912,33922,33961,34189,34209,34212,34242,34294,34339,34379,34445,34475,34551,34649,34683,34813,34862,34935,35019,35046,35096,35120,35211,35469,35486,35619,35673,35742,35764,35922,36133,36159,36247,36400,36429,36460,36484,36529,36541,36610,36613,36700,36723,36819,36886,36956,37048,37129,37132,37202,37220,37239,37319,37322,37356,37458,37616,37636,37810,37851,37879,38103,38128,38281,38420,38443,38522,38616,38643,38713,38768,38818,38856,38875,38908,38920,38932,38952,38981,39124,39249,39255,39298,39372,39418,39518,39551,39610,39724,39764,39822,39851,39969,39986,39990,39997,40062,40071,40117,40174,40185,40218,40265,40300,40347,40366,40384,40442,40448,40455,40507,40519,40570,40597,40636,40651,40705,40728,40873,40938,41010,41064,41235,41270,41281,41290,41387,41396,41434,41468,41511,41591,41593,41751,41877,41947,42063,42069,42115,42143,42300,42315,42521,42570,42603,42607,42773,42807,43133,43139,43144,43297,43324,43372,43444,43445,43540,43651,43690,43734,43846,43872,43909,43954,43970,44067,44070,44133,44238,44242,44252,44314,44355,44360,44361,44430,44540,44616,44654,44715,44770,44799,44819,44835,44895,44976,45015,45158,45268,45318,45339,45426,45553,45571,45586,45596,45661,45801,45839,45902,45941,46039,46107,46157,46244,46305,46311,46354,46377,46381,46393,46405,46506,46521,46534,46567,46644,46745,46769,46985,47006,47056,47160,47180,47261,47366,47409,47542,47620,47629,47676,47732,47750,47826,47857,47914,48114,48157,48159,48182,48198,48206,48219,48221,48262,48404,48413,48433,48474,48615,48658,48749,48889,49003,49062,49372,49472,49533,49560,49570,49631,49906,49925,50265,50333,50467,50484,50564,50586,50637,50778,51025,51178,51183,51222,51258,51315,51416,51422,51525,51628,51683,51852,51913,51938,51995,52162,52194,52217,52223,52305,52368,52389,52511,52531,52622,52634,52829,53022,53080,53142,53177,53263,53298,53317,53406,53427,53451,53501,53631,53824,54038,54068,54099,54115,54153,54182,54206,54272,54298,54304,54324,54494,54532,54697,54725,54861,55007,55174,55270,55319,55437,55448,55455,55490,55523,55663,55686,55694,55802,55822,55860,55869,55892,55895,56053,56109,56115,56173,56200,56225,56238,56569,56583,56670,56784,56937,57034,57245,57289,57294,57312,57386,57391,57431,57484,57580,57614,57970,58038,58073,58202,58278,58401,58470,58472,58477,58487,58557,58604,58625,58652,58962,58984,59043,59135,59144,59381,59410,59555,59716,59895,59953,59986,60029,60154,60367,60378,60434,60436,60477,60624,60805,60883,60951,60999,61234,61277,61406,61446,61524,61672,61683,61760,61852,61867,61957,62021,62085,62098,62127,62133,62162,62220,62238,62288,62320,62424,62433,62570,62658,62744,62811,62873,63048,63126,63238,63380,63485,63507,63538,63656,63715,63768,63787,63884,63902,63982,64114,64182,64412,64423,64462,64613,64643,64686,64782,64832,64963,65055,65113,65230,65233,65256,65333,65407,65517,65570,65650,65928,66039,66080,66093,66095,66104,66139,66236,66241,66269,66288,66521,66554,66609,66634,66715,66729,66852,66903,66907,66965,67183,67289,67308,67329,67398,67589,67663,67799,67995,68003,68014,68154,68167,68170,68182,68260,68387,68415,68484,68517,68519,68521,68575,68600,68686,68703,68831,68872,68961,68984,68989,69005,69128,69241,69351,69412,69544,69547,69562,69607,69670,69671,69678,69701,69845,69910,70094,70144,70156,70160,70166,70178,70229,70252,70518,70627,70752,70906,70914,70943,71146,71170,71357,71409,71585,71586,71591,71609,71664,71670,71717,72017,72285,72334,72397,72440,72486,72542,72622,72726,72994,72999,73149,73150,73240,73287,73294,73326,73337,73456,73777,73913,73943,74039,74115,74117,74202,74232,74273,74357,74422,74561,74568,74785,74821,74839,74868,74939,75014,75107,75318,75337,75622,75626,75658,75728,76102,76148,76191,76220,76221,76557,76699,76881,76891,76928,76939,76963,76971,77289,77454,77463,77468,77543,77555,77771,77832,77882,77897,78010,78146,78218,78347,78386,78463,78506,78659,78719,79034,79113,79162,79245,79394,79590,79613,79736,80210,80272,80279,80315,80427,80607,80763,80809,80884,80984,81103,81109,81121,81318,81381,81401,81456,81514,81627,81671,81725,81790,81813,81913,81914,81976,82128,82272,82331,82487,82495,82623,82914,82938,82974,83050,83174,83432,83463,83465,83477,83501,83537,83555,83598,83686,83897,83905,83927,83990,84001,84154,84220,84447,84665,84668,84783,84798,84802,84831,84855,84984,84992,85202,85262,85344,85447,85481,85518,85535,85549,85808,85866,85893,85926,86042,86130,86175,86187,86391,86495,86632,86752,86792,87034,87048,87065,87131,87189,87223,87357,87428,87475,87477,87573,87580,87839,87918,87946,87951,87970,88031,88301,88309,88421,88531,88540,88803,88925,89016,89017,89069,89159,89175,89184,89197,89385,89401,89432,89465,89485,89525,89540,89559,89657,89757,89847,89932,89950,89960,89988,90097,90124,90145,90209,90255,90272,90444,90496,90783,90820,90948,91363,91441,91504,91590,91729,91738,91753,91773,91811,91825,91872,91895,91901,92063,92083,92170,92308,92318,92354,92427,92529,92712,92716,92834,92875,92890,93038,93167,93371,93373,93523,93603,93743,93805,93914,93927,93931,93949,94055,94057,94282,94469,94565,94735,94739,94757,94792,94829,94876,94894,95002,95038,95045,95077,95098,95127,95190,95199,95242,95287,95365,95432,95477,95522,95541,95546,95552,95643,95726,95785,95806,95852,95928,95964,95977,96022,96052,96059,96097,96128,96267,96275,96302,96325,96344,96391,96505,96588,96626,96644,96658,96709,96779,96881,96921,96928,96988,96999,97005,97019,97065,97074,97110,97195,97217,97334,97376,97649,97689,98024,98031,98203,98267,98315,98332,98341,98445,98561,98574,98792,98828,98968,99036,99094,99184,99228,99358,99367,99394,99407,99427,99504,99545,99672,99752,99823,99855",
          "84,100,182,221,230,249,338,510,602,677,790,830,928,1050,1056,1089,1121,1230,1234,1313,1318,1342,1407,1423,1448,1557,1594,1703,1715,1896,1965,1989,2061,2166,2282,2397,2515,2684,2710,2782,2905,2949,2958,3001,3027,3120,3139,3157,3211,3331,3340,3346,3399,3405,3435,3536,3607,3683,3850,3889,4005,4062,4109,4266,4301,4396,4401,4523,4551,4766,4792,4854,4891,4987,5000,5067,5096,5480,5536,5563,5678,5705,5916,5960,6034,6089,6261,6392,6570,6575,6581,6639,6676,6733,6833,6839,6881,6918,6977,7012,7074,7141,7236,7252,7291,7484,7486,7559,7570,7579,7607,7713,7772,7793,7887,7907,7913,7987,8050,8265,8315,8317,8358,8430,8467,8552,8569,8608,8623,8625,8720,8799,8811,8929,9189,9368,9453,9481,9539,9571,9589,9649,9846,9883,10064,10193,10235,10303,10346,10452,10457,10469,10522,10534,10601,10635,10657,10828,10849,10857,10871,10888,11043,11087,11093,11105,11123,11270,11306,11383,11391,11409,11658,12029,12032,12048,12502,12553,12557,12647,12660,12671,12716,12748,12799,12867,12897,12929,12979,12981,13063,13325,13350,13451,13527,13603,13669,13704,13708,13918,13946,14062,14123,14226,14646,14647,14713,14722,14805,14825,14845,14909,14976,15039,15075,15115,15209,15319,15326,15463,15469,15530,15553,15857,15932,15976,16050,16055,16080,16096,16182,16357,16437,16456,16498,16520,16569,16609,16717,16795,16823,16859,16944,16986,17018,17065,17142,17214,17237,17328,17409,17442,17445,17459,17487,17510,17611,17658,17713,17756,18065,18207,18249,18306,18424,18437,18476,18482,18498,18554,18677,18856,18876,18944,18960,19126,19229,19266,19268,19336,19400,19450,19563,19575,19610,19686,19693,19746,19778,19990,20211,20227,20248,20319,20347,20378,20447,20478,20514,20603,20682,20691,20787,20822,20827,20879,20909,20917,20972,20973,21182,21299,21447,21553,21584,21680,21762,21813,21888,21966,22076,22104,22246,22265,22343,22350,22372,22406,22413,22492,22659,22766,22809,22842,22891,22898,23006,23050,23060,23108,23161,23180,23189,23293,23382,23418,23530,23589,23598,23760,23806,23884,24049,24089,24107,24164,24201,24249,24254,24324,24362,24371,24406,24458,24539,24604,24621,24646,24659,24710,24776,24807,24821,24830,24832,24907,24933,24950,24965,25037,25089,25101,25267,25326,25416,25430,25433,25529,25557,25635,25697,25704,25715,25907,25979,26023,26063,26124,26143,26181,26357,26417,26477,26607,26634,26698,26767,26781,26891,26892,26905,26970,27063,27088,27093,27147,27199,27225,27334,27404,27602,27668,27691,27801,27822,27917,28054,28073,28074,28201,28375,28492,28497,28533,28573,28578,28583,28590,28692,28946,29077,29181,29202,29245,29247,29337,29381,29414,29509,29564,29566,29600,29604,29659,29787,29818,29825,29833,29916,30070,30079,30151,30292,30314,30337,30343,30421,30470,30526,30544,30545,30648,30706,30780,30874,31021,31046,31319,31383,31560,31618,31651,31731,31781,31972,32121,32204,32362,32384,32516,32556,32573,32634,32794,32813,32830,32965,32980,33028,33034,33101,33109,33262,33276,33279,33288,33291,33370,33383,33482,33592,33647,33656,33739,33770,34017,34060,34136,34214,34440,34444,34508,34509,34584,34623,34656,34718,34726,34734,34969,34991,35110,35198,35220,35270,35294,35345,35454,35470,35478,35830,35907,35909,36096,36224,36272,36329,36388,36496,36501,36515,36542,36582,36586,36602,36682,36780,36841,36871,36945,36976,37017,37073,37165,37272,37429,37586,37673,37697,37883,37897,37964,38249,38269,38331,38358,38413,38424,38475,38543,38644,38696,38714,38819,38841,38870,38999,39110,39138,39300,39338,39438,39541,39600,39703,39770,39788,39809,39878,40002,40009,40025,40080,40113,40134,40153,40199,40202,40203,40293,40435,40449,40538,40600,40621,40721,40899,40973,41116,41128,41136,41646,41720,41795,41835,41920,41972,42080,42347,42359,42427,42433,42458,42559,42619,42675,42712,42716,42982,43018,43055,43168,43202,43246,43325,43446,43462,43582,43629,43847,43978,43991,44030,44107,44150,44254,44308,44374,44480,44626,44643,44680,44730,44845,44846,44905,44906,44915,44951,44957,45026,45044,45063,45132,45214,45238,45389,45435,45583,45710,45726,45759,45768,45802,45853,45861,45898,46001,46028,46108,46198,46251,46348,46432,46465,46500,46851,46888,47008,47312,47421,47465,47570,47574,47785,47946,48046,48191,48248,48415,48437,48466,48504,48557,48579,48618,48687,48823,48828,48913,48969,48976,48984,49021,49156,49186,49287,49295,49336,49370,49425,49508,49522,49544,49673,49677,49723,49865,49882,50120,50221,50340,50398,50542,50666,50684,50763,50781,50824,50832,50935,51063,51101,51128,51235,51240,51305,51310,51388,51420,51473,51486,51508,51513,51570,51721,51730,51772,51866,51893,51986,52082,52222,52282,52341,52406,52465,52537,52554,52626,52663,52828,52848,52860,52877,52995,53000,53051,53185,53198,53238,53289,53322,53364,53365,53544,53823,53953,53969,53971,54095,54215,54237,54312,54341,54365,54393,54453,54473,54524,54535,54650,54830,54865,54943,54967,54976,55094,55103,55140,55149,55215,55240,55326,55480,55602,55674,55773,55799,55877,55899,55903,56106,56157,56213,56431,56459,56642,56930,57040,57143,57235,57336,57390,57395,57396,57492,57707,57789,57857,57863,57866,57912,57973,58048,58151,58181,58192,58290,58350,58461,58474,58590,58655,58719,58782,58784,58883,58940,59137,59190,59303,59304,59492,59543,59566,59736,59757,59947,59995,60152,60173,60286,60542,60570,60787,60923,60978,61034,61060,61105,61221,61256,61573,61584,61589,61674,61879,62170,62244,62247,62401,62465,62597,62714,62774,62835,62907,62953,62981,63008,63059,63070,63139,63160,63247,63256,63257,63266,63291,63297,63314,63330,63341,63351,63373,63394,63461,63503,63580,63611,63627,63727,63728,63991,64063,64133,64200,64264,64537,64557,64564,64700,64766,64951,65008,65172,65300,65437,65476,65505,65558,65702,65809,65834,65938,66021,66133,66275,66291,66303,66327,66408,66450,66502,66637,66658,66659,66683,66773,66814,66913,67095,67096,67228,67328,67341,67413,67415,67581,67637,67854,67862,68064,68147,68264,68267,68276,68353,68436,68478,68591,68603,68669,68857,68858,68888,69056,69155,69168,69202,69317,69388,69399,69490,69570,69574,69616,69634,69703,69798,69819,69824,69850,69903,70184,70235,70261,70390,70546,70581,70583,70717,70762,70778,70848,70999,71091,71134,71183,71196,71416,71418,71480,71710,71788,71802,71808,71892,71950,71971,71976,72026,72108,72157,72251,72273,72675,72676,72739,72791,72886,72888,73084,73089,73133,73165,73184,73192,73290,73453,73634,73638,73747,73841,73880,73929,73961,74010,74050,74136,74189,74234,74271,74272,74459,74488,74583,74642,74735,74752,74764,74767,74787,74862,74910,75025,75406,75518,75589,75712,75958,75967,76011,76200,76245,76405,76576,76685,76914,76951,77075,77139,77169,77224,77269,77306,77389,77547,77598,77737,77843,77859,77907,77937,77949,78183,78285,78459,78574,78582,78587,78598,78605,78675,78835,78851,78924,78968,78988,79002,79084,79117,79198,79345,79448,79457,79536,79546,79622,79703,79789,79814,79905,79919,79933,79997,80148,80149,80240,80366,80509,80545,80582,80631,80747,80814,80815,80871,80926,81030,81059,81075,81158,81290,81322,81326,81516,81533,81777,81870,81969,81993,82089,82145,82232,82258,82670,82685,82694,82733,82787,82816,82947,82961,82972,82977,83017,83305,83362,83457,83521,83544,83561,83613,83811,83869,83929,84009,84091,84134,84278,84313,84322,84382,84415,84416,84443,84468,84597,84610,84612,84637,84655,84704,84709,84723,84736,84747,84779,84829,84876,85050,85059,85063,85090,85119,85123,85289,85395,85553,85558,85574,85813,85916,86032,86058,86207,86221,86301,86363,86364,86405,86411,86438,86446,86592,86600,86635,86680,86804,86825,86838,86864,86894,86917,86938,87074,87132,87186,87202,87215,87508,87766,88040,88092,88118,88137,88242,88371,88395,88401,88593,88641,88685,88808,88935,88952,88953,88992,89001,89194,89246,89280,89348,89472,89567,89573,89578,89710,89723,89778,89862,89871,89917,89957,90027,90105,90361,90392,90455,90718,90775,90817,90825,90880,90922,90935,91007,91101,91164,91217,91244,91304,91410,91492,91577,91856,91868,91968,91971,92095,92134,92196,92248,92257,92315,92337,92496,92557,92592,92629,92649,92662,92846,92906,92914,92953,92970,92991,93079,93261,93266,93452,93472,93498,93590,93660,93702,93747,93758,93819,93853,93907,94011,94041,94097,94129,94178,94185,94258,94311,94404,94443,94554,94658,94698,94981,94987,95026,95035,95167,95318,95390,95398,95409,95466,95471,95521,95535,95577,95579,95770,95802,95883,95890,96009,96123,96156,96317,96339,96340,96528,96693,96756,96805,97008,97013,97094,97149,97152,97236,97417,97472,97477,97547,97859,97942,97947,98098,98374,98478,98508,98537,98562,98774,98866,98879,99185,99242,99243,99293,99294,99366,99415,99447,99606,99648,99684,99707,99789,99876,99905",
          "11,24,95,176,206,255,271,315,424,454,480,503,515,632,709,733,798,808,851,909,932,936,1002,1199,1252,1262,1272,1444,1524,1638,1709,1724,1743,1777,1948,2034,2124,2128,2160,2201,2240,2383,2469,2579,2699,2711,2726,2760,2823,2939,3130,3138,3150,3221,3353,3412,3420,3508,3619,3821,3935,4144,4177,4259,4281,4283,4364,4465,4563,4594,4653,4695,4735,4753,4759,4804,4833,4954,5029,5138,5155,5264,5272,5278,5291,5295,5493,5509,5534,5727,5744,5754,5781,6216,6328,6331,6564,6588,6619,6695,6837,6878,6925,6948,6951,7056,7100,7150,7211,7393,7433,7464,7567,7670,7721,7809,7860,7934,7961,8371,8536,8560,8575,8579,8710,8798,8815,8862,8894,8960,9145,9344,9349,9518,9583,9604,9633,9756,9772,9803,9924,10050,10080,10178,10344,10350,10351,10389,10448,10546,10571,10661,10687,10691,10743,10803,10847,11030,11176,11193,11358,11366,11422,11675,11845,11901,12059,12203,12232,12349,12455,12498,12528,12534,12650,12685,12764,12833,12993,13023,13036,13224,13324,13367,13389,13420,13471,13571,13630,13687,13794,13802,13889,13917,14022,14138,14155,14174,14204,14338,14366,14446,14525,14531,14606,14622,14650,14675,14680,14829,14885,14954,15040,15131,15214,15419,15636,15727,15920,15958,16071,16178,16332,16366,16401,16416,16466,16617,16639,16678,16801,16872,16893,17061,17193,17278,17309,17358,17495,17525,17565,17614,17625,17638,17668,17683,17769,17809,17930,17979,18037,18122,18149,18189,18228,18271,18297,18345,18347,18372,18449,18473,18487,18617,18627,18784,18819,18880,18963,19009,19018,19043,19062,19210,19327,19343,19392,19456,19657,19660,19795,19806,19915,20026,20059,20177,20287,20415,20473,20475,20598,20774,20826,20927,20934,20960,21018,21141,21148,21213,21325,21340,21394,21433,21659,21695,21702,21791,21932,21975,21995,22024,22103,22114,22326,22415,22439,22716,22752,22873,22876,22958,23017,23182,23200,23323,23492,23501,23527,23551,23712,23729,23822,23895,23996,24082,24117,24124,24153,24500,24681,24771,24951,24958,25040,25096,25136,25181,25210,25316,25350,25361,25525,25531,25611,25686,25732,25744,25798,25840,25895,25923,26026,26151,26299,26367,26439,26492,26499,26650,26673,26677,26700,26723,26779,26801,26944,26961,26965,26975,26998,26999,27014,27092,27281,27363,27472,27488,27571,27572,27638,27724,28003,28092,28100,28108,28415,28446,28538,28559,28671,28747,28792,28798,28915,28979,29018,29079,29335,29390,29397,29447,29570,29591,29725,29763,29783,29840,29948,30027,30094,30439,30527,30571,30751,30886,30934,31020,31122,31247,31312,31494,31773,31868,32108,32139,32158,32245,32270,32415,32839,32841,32936,33030,33044,33108,33138,33168,33185,33267,33328,33377,33487,33541,33684,33774,33869,33887,33899,34111,34188,34215,34233,34239,34250,34377,34467,34772,34780,34810,34897,35048,35073,35348,35483,35650,35700,35800,35812,35836,36045,36076,36099,36128,36164,36194,36358,36378,36497,36576,36664,36672,36865,36921,36959,36983,37028,37038,37244,37735,37763,37899,37908,38059,38244,38260,38293,38347,38456,38547,38826,38902,38943,38946,39120,39189,39197,39422,39495,39774,39791,39793,39897,39955,40178,40363,40488,40533,40614,40730,40742,40855,41025,41089,41200,41205,41228,41361,41462,41516,41524,41700,41734,41809,41866,41892,41976,42029,42066,42107,42135,42153,42353,42398,42530,42544,42608,42762,42781,42857,42992,42997,43012,43060,43293,43422,43427,43531,43556,43568,43777,43869,44086,44228,44468,44542,44685,44740,45009,45159,45343,45363,45408,45641,45649,45698,45886,45949,46029,46049,46203,46263,46301,46309,46349,46468,46537,46547,46597,46623,46671,46965,47099,47125,47173,47178,47358,47380,47412,47423,47448,47553,47563,47584,47706,47741,47954,48033,48135,48309,48361,48671,48871,49087,49095,49191,49213,49265,49300,49427,49676,49775,49780,49782,49911,49913,49956,49969,50051,50105,50220,50276,50312,50468,50483,50503,50521,50573,50743,50855,51030,51104,51209,51519,51726,51801,51839,51889,52024,52063,52117,52123,52216,52435,52579,52715,52756,52758,52794,53092,53136,53145,53204,53228,53274,53275,53386,53496,53506,53744,53752,53900,53903,53905,53932,54149,54151,54164,54173,54227,54240,54330,54471,54640,54651,54676,54692,54782,54826,54947,54951,55050,55283,55322,55442,55466,55596,55651,55687,55809,55947,56025,56036,56134,56145,56262,56359,56398,56454,56484,56624,56697,56719,56851,56874,56898,56955,56989,57276,57302,57342,57462,57555,57597,57693,57755,57873,57938,57991,58167,58283,58300,58364,58367,58378,58393,58399,58450,58465,58505,58579,58709,58881,59182,59192,59246,59314,59501,59551,59591,59722,59764,59768,59791,59854,60051,60091,60135,60158,60319,60347,60411,60531,60553,60609,60615,60681,60810,60812,60882,60892,60930,60998,61081,61090,61176,61181,61196,61269,61407,61453,61498,61611,61671,61720,61721,61814,61817,61917,61923,61971,62150,62256,62278,62568,62612,62633,62691,62760,62889,63010,63069,63109,63147,63274,63332,63387,63436,63581,63597,63643,63820,63914,63931,63938,64061,64121,64149,64455,64519,64583,64603,64637,64663,64695,64740,64763,64940,64984,64990,65021,65122,65234,65541,65655,65726,65735,66112,66188,66190,66302,66304,66484,66498,66526,66583,66593,66619,66649,66708,66783,66826,66850,66866,66949,66975,66976,67054,67067,67100,67153,67208,67325,67378,67492,67570,67606,67638,67655,67720,67832,67932,67961,67989,68035,68071,68088,68254,68324,68378,68384,68421,68493,68533,68573,68651,68816,68894,68932,69036,69048,69108,69156,69377,69393,69510,69722,69757,69857,69979,70309,70516,70525,70549,70580,70589,70630,70881,70940,71013,71084,71157,71296,71297,71379,71403,71426,71532,71550,71580,71592,71632,71748,71774,71800,71954,72113,72145,72156,72241,72510,72686,72771,72865,72952,73043,73117,73352,73406,73415,73529,73562,73860,73905,73948,74036,74224,74238,74275,74313,74467,74469,74485,74530,74763,74779,74861,74966,75110,75247,75300,75316,75326,75458,75460,75481,75504,75618,75643,75666,75785,75849,75973,75986,76028,76186,76187,76328,76414,76606,76818,77020,77063,77195,77492,77521,77856,77948,78115,78203,78544,78548,78732,78746,78756,78940,79341,79476,79603,79662,79667,79757,79901,79987,80040,80119,80169,80190,80192,80254,80305,80594,80715,80822,80845,80895,80907,80956,81003,81101,81150,81215,81278,81449,81557,81563,81631,82064,82158,82368,82573,82659,82697,82715,82790,82836,83039,83046,83092,83316,83357,83488,83522,83530,83678,83784,83879,83883,84049,84171,84189,84197,84282,84336,84419,84430,84448,84658,84826,84960,84980,85103,85210,85238,85266,85429,85530,85642,85658,85867,85981,86088,86307,86314,86356,86537,86604,86663,86729,86741,86950,87060,87247,87269,87363,87723,87750,87873,87922,87972,88146,88172,88245,88344,88496,88602,88610,88643,88650,88665,88761,88764,88801,88815,88818,88934,89172,89195,89415,89460,89503,89530,89624,89635,89645,89678,89752,89792,90010,90037,90088,90134,90327,90424,90451,90461,90507,90777,90802,90896,90900,91065,91097,91109,91133,91196,91290,91342,91390,91416,91424,91432,91575,91654,91663,91757,91779,91849,91964,92204,92225,92265,92359,92413,92459,92700,92739,92787,92860,92933,92963,93001,93017,93068,93507,93548,93555,93598,93744,93919,94064,94080,94084,94263,94337,94432,94461,94935,94949,95230,95233,95238,95304,95329,95402,95414,95529,95574,95618,95644,95722,95723,95751,95910,95954,95989,96157,96160,96332,96576,96619,96623,96791,96831,96843,96990,97106,97252,97275,97285,97335,97453,97544,97601,97665,97683,97701,97776,97878,98042,98141,98248,98378,98416,98422,98521,98557,98614,98640,98771,98904,98915,98976,99041,99087,99135,99177,99329,99340,99395,99483,99484,99687,99910,99979",
          "114,135,168,675,689,690,771,796,824,937,1046,1059,1066,1179,1249,1334,1483,1540,1546,1620,1677,1681,1867,1884,1947,2025,2045,2053,2165,2207,2291,2313,2396,2401,2631,2672,2674,2722,2765,2776,2826,2861,2890,2993,3066,3214,3404,3505,3570,3694,3703,3976,3978,3990,4013,4097,4249,4295,4311,4354,4362,4463,4501,4637,4649,4662,4708,4801,4817,4861,4918,4972,5056,5071,5117,5168,5497,5524,5533,5595,5627,5634,5762,6025,6081,6104,6120,6211,6217,6223,6236,6255,6299,6425,6514,6696,6920,7115,7168,7204,7297,7375,7500,7614,7617,7781,7821,7868,7895,7982,8011,8172,8182,8228,8347,8396,8418,8427,8468,8478,8522,8578,8584,8835,8842,9066,9103,9281,9320,9381,9478,9485,9511,9622,9791,10004,10060,10061,10127,10164,10300,10541,10567,10594,10655,10790,10882,10972,11033,11050,11060,11191,11219,11234,11261,11333,11417,11437,11476,11560,11563,11572,11628,11664,11692,11738,12094,12125,12289,12298,12508,12533,12638,12682,12713,12739,13073,13155,13168,13277,13340,13390,13435,13452,13469,13530,13552,13589,13638,13819,13853,13975,14140,14151,14230,14273,14290,14293,14305,14408,14452,14509,14512,14545,14640,14898,14903,14926,14932,14979,15051,15078,15110,15216,15287,15309,15401,15458,15575,15611,15681,15811,15972,16010,16021,16166,16239,16262,16324,16347,16436,16506,16610,16788,16833,16903,16931,16947,17285,17308,17314,17321,17446,17448,17498,17526,17536,17600,17628,17648,17785,17876,17953,17965,17969,17992,18056,18106,18129,18181,18206,18300,18309,18378,18454,18613,18660,18670,18704,18742,18762,18815,18950,19010,19046,19061,19107,19157,19422,19469,19490,19491,19588,19637,19680,19683,19736,19748,19751,19826,19847,19875,19891,19914,20038,20100,20172,20254,20301,20445,20553,20604,20651,20707,20714,20767,20781,20859,20880,20914,20994,21030,21080,21253,21258,21327,21353,21357,21372,21428,21580,21619,21626,21646,21660,21662,21778,21959,22143,22144,22231,22285,22290,22319,22332,22427,22451,22631,22651,22653,22788,22821,22905,22944,23140,23146,23174,23186,23429,23470,23507,23586,23631,23667,23692,23797,23880,23911,23951,23954,23962,24018,24025,24033,24147,24248,24279,24293,24504,24522,24546,24605,24663,24722,24884,24937,25002,25349,25379,25383,25411,25449,25530,25539,25641,25676,25679,25740,25752,25775,25788,25945,25954,25962,26013,26121,26203,26207,26208,26217,26253,26493,26537,26554,26588,26590,26601,26653,26796,26924,26926,26974,27076,27120,27128,27154,27221,27297,27302,27406,27419,27425,27559,27641,27643,27671,27679,27692,27696,27755,27810,27891,27893,27961,28124,28148,28363,28591,28599,28758,28782,28819,28950,28974,29025,29059,29126,29288,29298,29434,29711,29728,29736,29757,29899,29915,29950,29976,30003,30029,30055,30185,30273,30323,30413,30419,30576,30638,30772,30929,30975,30993,31028,31113,31401,31498,31521,31589,31662,31680,31856,31903,31939,31956,32058,32096,32101,32192,32230,32272,32358,32405,32481,32499,32615,32635,32661,32760,32780,33048,33131,33140,33205,33237,33244,33308,33359,33380,33427,33477,33506,33548,33701,33735,33736,33989,34117,34126,34180,34304,34324,34328,34336,34474,34511,34663,34767,34783,34906,34940,34946,35008,35038,35072,35170,35224,35228,35231,35243,35378,35381,35411,35603,35608,35609,35611,35879,35886,35926,35962,36010,36077,36086,36153,36168,36258,36308,36363,36382,36410,36432,36440,36509,36678,36690,36850,36961,36991,37031,37118,37251,37349,37413,37507,37511,37545,37566,37689,37714,37743,37802,37909,38250,38286,38484,38531,38620,38770,38962,38992,39011,39171,39256,39388,39511,39533,39568,39617,39639,39640,39711,39837,40023,40096,40154,40168,40217,40239,40337,40345,40521,40680,40697,40713,40779,40781,41031,41425,41476,41562,41584,41756,41758,41775,42136,42209,42280,42289,42411,42431,42487,42717,42719,42830,42836,42991,43087,43180,43255,43343,43489,43499,43519,43630,43634,43689,43723,43792,43818,43947,44065,44095,44127,44295,44366,44655,44658,44762,44875,44879,44885,44913,44996,45052,45065,45083,45100,45111,45352,45407,45588,45666,45808,45918,45948,45952,46047,46160,46233,46264,46269,46400,46454,46599,46697,46746,46774,46872,46926,46939,47018,47087,47381,47390,47398,47463,47470,47472,47498,47746,47898,47916,47918,47944,48176,48208,48261,48286,48430,48463,48553,48566,48648,48719,48721,48759,48774,48831,48892,48907,48952,48961,48981,49027,49169,49180,49337,49412,49415,49440,49493,49526,49571,49616,49636,49699,49762,49918,50049,50162,50164,50185,50194,50209,50216,50249,50275,50282,50514,50519,50552,50629,50661,50671,50675,51043,51059,51106,51166,51251,51269,51289,51365,51581,51612,51637,51657,51695,51737,51748,52098,52215,52405,52563,52639,52752,52868,52966,52994,53027,53036,53052,53169,53221,53397,53478,53603,53653,53904,53918,53933,54044,54058,54145,54159,54259,54270,54295,54310,54333,54404,54514,54538,54677,54684,54696,54753,54845,54857,55013,55037,55136,55374,55435,55479,55588,55711,55725,55855,55898,55936,55983,56113,56143,56218,56235,56258,56298,56321,56408,56439,56480,56773,57204,57209,57258,57303,57360,57442,57470,57589,57672,57673,57786,57919,57932,57948,57972,58077,58253,58396,58427,58492,58510,58534,58623,58750,58874,58877,58894,58923,59151,59180,59184,59323,59453,59512,59522,59672,59738,59759,59819,59823,59970,60098,60120,60130,60189,60302,60365,60417,60626,60710,60718,60771,60792,61065,61103,61214,61222,61310,61334,61378,61389,61426,61500,61502,61519,61558,61591,61608,61825,61826,62051,62107,62115,62124,62146,62205,62213,62230,62395,62403,62504,62514,62588,62625,62690,62707,62718,62742,62829,63090,63130,63192,63294,63346,63398,63496,63519,63549,63613,63635,63772,63837,63897,63932,64050,64146,64164,64295,64310,64422,64451,64710,64722,64756,64813,64828,64914,65025,65106,65151,65302,65479,65500,65599,65765,65900,65932,65943,66210,66211,66268,66330,66399,66468,66639,66703,66745,66810,66833,66924,66925,66959,66992,67141,67162,67284,67343,67419,67565,67578,67647,67658,67911,67937,68065,68194,68212,68372,68592,68644,68664,68733,68808,68938,68977,69033,69043,69098,69179,69228,69281,69282,69443,69539,69581,69630,69638,69656,69694,69728,69820,69908,69944,69968,70001,70007,70062,70157,70158,70172,70180,70265,70359,70424,70447,70453,70469,70495,70689,70691,70722,70753,70766,70781,70945,71174,71204,71308,71342,71344,71349,71385,71407,71478,71728,71741,71836,71843,71855,71880,71986,72039,72127,72286,72331,72410,72460,72487,72621,72729,72748,72754,72781,72955,73159,73169,73190,73211,73215,73257,73366,73418,73476,73597,73891,73894,73895,73997,74138,74152,74160,74169,74218,74438,74580,74856,74921,74994,75133,75203,75209,75214,75293,75296,75367,75614,75977,75979,76021,76061,76126,76275,76503,76574,76625,76782,76911,76970,77033,77114,77166,77206,77217,77294,77329,77415,77430,77434,77435,77525,77611,77616,77741,77764,77801,77815,77866,77880,77942,77958,78025,78155,78178,78359,78404,78440,78510,78540,78560,78593,78784,78805,78814,78860,78979,78995,79188,79196,79255,79690,79920,79930,79984,79985,80044,80049,80157,80229,80286,80492,80672,80752,80769,80805,80818,81004,81016,81174,81189,81206,81387,81443,81560,81620,81672,81684,81745,81757,81768,81859,82301,82324,82349,82427,82459,82599,82608,82617,82631,82652,82700,82841,82872,82895,83124,83270,83368,83401,83411,83655,83912,83932,83965,84164,84169,84172,84215,84264,84363,84397,84502,84667,84678,84718,84732,84744,84814,84834,84871,84949,84997,85035,85185,85261,85351,85369,85379,85415,85533,85596,85619,85711,85734,85798,85854,85929,86070,86257,86319,86326,86346,86377,86416,86547,86591,86712,86898,86964,87023,87179,87409,87481,87669,87742,87795,87811,87875,87876,88019,88101,88363,88377,88557,88565,88586,88623,88638,88716,88785,88811,88847,88890,88951,89079,89240,89267,89290,89299,89335,89421,89423,89644,89702,89727,89745,89859,89881,90017,90099,90137,90155,90160,90460,90528,90558,90566,90576,90636,90744,90916,90929,90996,91005,91079,91102,91123,91345,91397,91431,91434,91505,91628,91736,91745,91774,91791,91796,91799,91831,92059,92079,92107,92131,92233,92238,92349,92493,92525,92741,92857,92879,92934,93020,93150,93297,93331,93412,93550,93587,93602,93661,93714,93716,93774,93872,93917,94059,94194,94199,94275,94331,94393,94411,94632,94652,94780,94970,94999,95118,95141,95247,95296,95418,95491,95543,95590,95598,95602,95623,95796,95818,95829,95860,95884,95899,95976,95982,96062,96137,96376,96497,96553,96715,96735,96770,96828,96835,96844,96846,96872,96992,96993,97027,97070,97245,97269,97272,97273,97281,97389,97517,97570,97578,97691,97745,97824,97847,98087,98128,98178,98307,98412,98608,98701,98848,98898,98917,99179,99352,99488,99495,99591,99796",
          "32,56,378,620,672,1069,1158,1474,1484,1531,1581,1615,1672,1689,1694,1710,1798,1839,1893,1935,2066,2084,2097,2159,2292,2298,2457,2559,2615,2697,2758,2781,2936,2973,3013,3113,3169,3325,3393,3396,3485,3525,3526,3564,3584,3609,3883,3919,4020,4119,4619,4689,4700,4773,4902,4933,4949,5007,5051,5076,5197,5219,5288,5415,5430,5440,5496,5552,5596,5649,5698,5736,5746,5838,5855,5945,5946,5973,5996,6020,6090,6133,6156,6359,6463,6486,6496,6596,6649,6702,6717,6802,6983,7004,7158,7302,7376,7388,7540,7894,7937,7954,7963,8025,8057,8085,8215,8243,8769,8822,8864,8865,8913,8973,9023,9026,9043,9166,9184,9208,9275,9315,9500,9717,9849,9930,9985,10184,10257,10293,10443,10488,10490,10494,10713,10820,10929,10954,10970,11223,11252,11255,11310,11357,11527,11677,11778,11962,12025,12136,12365,12397,12406,12425,12515,12962,12964,13003,13010,13026,13124,13152,13403,13537,13746,13855,13874,14013,14083,14111,14189,14300,14363,14551,14632,14935,14995,15134,15136,15238,15406,15630,15672,15772,15936,15960,16060,16133,16191,16592,16690,16700,16727,16775,16891,17105,17258,17297,17351,17362,17410,17727,17732,17753,17836,18005,18017,18151,18223,18296,18390,18503,18557,18588,18664,18836,18839,18882,18901,18947,18990,18998,19124,19233,19267,19286,19377,19399,19445,19464,19466,19613,19656,19907,20008,20101,20131,20237,20275,20293,20477,20643,20699,20792,20925,20993,21079,21147,21215,21293,21364,21441,21479,21510,21581,21586,21747,21892,21922,22011,22170,22182,22220,22404,22510,22555,22655,22774,22806,22841,22974,23051,23099,23159,23179,23233,23383,23443,23557,23697,23706,23804,23930,23976,23999,24172,24327,24333,24407,24441,24460,24574,24696,24858,24896,24904,24977,24978,24990,25077,25227,25265,25293,25355,25488,25524,25564,26068,26091,26220,26339,26495,26715,26789,26897,27184,27265,27309,27355,27362,27375,27420,27441,27442,27610,27639,27946,28111,28302,28428,28437,28461,28633,28810,28837,28855,28935,28939,28986,29055,29076,29233,29306,29371,29392,29469,29503,29593,29605,29646,29688,29691,29701,29713,29764,29864,29881,29897,29941,29977,30049,30145,30193,30260,30588,30590,30918,31242,31365,31366,31393,31430,31528,31606,31616,31874,31896,31964,31970,31986,31996,32425,32493,32515,32548,32581,32591,32596,32665,32706,32862,32974,33088,33130,33151,33211,33337,33376,33430,33432,33434,33571,33629,33657,33734,33783,33791,34012,34071,34185,34300,34342,34403,34414,34562,34616,34746,34757,34771,34787,34806,34809,34842,34912,34966,34976,34986,35204,35341,35416,35510,35652,35664,35691,35735,35760,35861,35871,35914,35958,36036,36209,36471,36676,36694,36696,36702,36751,36824,36985,37064,37085,37198,37221,37231,37278,37283,37392,37436,37468,37490,37552,37626,37657,37709,37811,37816,37891,37918,37978,38138,38227,38254,38422,38540,38602,38631,38672,38756,38835,38893,38921,38949,39012,39014,39024,39137,39240,39371,39391,39773,39804,39843,39862,39883,40187,40327,40364,40497,40512,40577,40591,40595,40654,40723,40757,40849,41258,41327,41415,41592,41766,41815,41993,42021,42061,42090,42175,42611,42618,42622,42662,42671,42689,42707,42828,42984,42996,43050,43070,43077,43124,43146,43155,43287,43386,43439,43545,43657,43687,43721,43733,43837,43865,43903,44062,44118,44191,44219,44222,44251,44322,44384,44456,44484,44488,44558,44697,44728,44731,44829,44878,44943,45019,45074,45126,45154,45207,45283,45347,45370,45452,45982,46044,46360,46473,46495,46643,46705,46860,46886,46951,46960,47139,47150,47265,47291,47317,47549,47683,47713,47884,47897,47907,47964,48091,48276,48346,48395,48487,48498,48593,48606,48614,48643,48668,48795,48929,48968,48995,49155,49177,49263,49277,49348,49439,49458,49495,49568,49649,49709,49740,49828,49933,49944,50181,50210,50218,50237,50270,50291,50455,50546,50704,50741,50769,50847,50917,50954,51035,51118,51161,51239,51383,51611,51687,51690,51713,51879,51881,51982,52137,52243,52314,52409,52440,52443,52482,52508,52572,52733,52796,52856,52887,52889,52974,52998,53067,53079,53105,53143,53235,53383,53399,53467,53513,53569,53577,53916,53950,53982,54222,54238,54321,54345,54407,54745,54784,54802,54856,54880,55144,55204,55265,55295,55401,55412,55419,55585,55661,55840,55888,55907,55928,55943,56026,56182,56240,56364,56423,56494,56562,56846,56866,57004,57022,57077,57090,57123,57159,57297,57318,57429,57487,57618,57891,57897,57910,57998,58052,58110,58274,58329,58335,58365,58456,58481,58511,58581,58595,58804,58890,59083,59301,59352,59478,59557,59592,59677,59713,59715,59730,59777,59780,59820,60240,60241,60270,60280,60305,60334,60410,60422,60459,60468,60605,60692,60693,60734,60783,60954,61001,61079,61089,61102,61115,61282,61359,61377,61487,61656,61673,61694,61767,61774,61823,61889,62075,62079,62099,62142,62271,62307,62393,62417,62422,62476,62600,62848,62855,62864,62869,62934,63253,63261,63286,63343,63500,63566,63612,63648,63773,63965,64263,64266,64552,64606,64680,64702,64741,64768,64867,64868,64974,64977,64993,65003,65016,65128,65167,65238,65744,65817,65851,66031,66273,66374,66425,66455,66466,66485,66630,66706,66971,66991,67077,67087,67180,67217,67272,67316,67383,67944,68039,68122,68299,68391,68400,68449,68477,68508,68524,68584,69072,69229,69309,69413,69450,69504,69749,69823,69863,70081,70449,70874,70904,71111,71208,71566,71611,71671,71928,72008,72057,72084,72151,72198,72203,72236,72248,72302,72304,72545,72645,72681,72784,72973,72989,73073,73509,73517,73587,73630,73675,73714,73852,73896,74000,74033,74052,74248,74291,74375,74420,74448,74576,74592,74749,74959,75112,75260,75432,75464,75508,75540,75620,75636,75701,75758,75923,76087,76182,76345,76533,76583,76690,76731,76849,76860,76873,76922,77500,77704,77820,77900,77967,78046,78111,78121,78148,78152,78165,78268,78284,78303,78308,78438,78604,78613,79121,79252,79396,79664,79687,79720,79940,79959,80077,80109,80132,80174,80223,80304,80331,80337,80464,80587,80917,80920,80954,81019,81185,81242,81294,81355,81395,81421,81498,81539,81905,81995,82121,82187,82242,82251,82321,82357,82419,82547,82574,82683,82729,82802,82839,82910,82936,83059,83089,83090,83095,83110,83182,83191,83340,83394,83426,83490,83795,83806,83979,84044,84135,84182,84195,84349,84384,84557,84660,84670,84673,84726,84860,85034,85126,85168,85263,85273,85426,85458,85609,85850,85898,86057,86261,86266,86461,86558,86590,86682,86807,86826,86839,86891,86918,86925,87062,87242,87415,87422,87459,87492,87519,87562,87583,87858,87916,88302,88472,88478,88479,88585,88603,88711,88745,88897,88973,89008,89032,89041,89051,89102,89153,89181,89243,89576,89619,89663,89998,90039,90050,90080,90081,90123,90141,90163,90171,90269,90425,90556,90592,90644,90659,90697,90704,90763,90845,90874,90898,90915,91039,91140,91165,91309,91418,91437,91909,91962,92008,92317,92428,92507,92637,92747,92757,92819,93059,93113,93392,93529,93749,93831,93978,93979,94123,94439,94460,94498,94543,94559,94577,94765,95053,95130,95204,95381,95459,95506,95566,95655,95755,95943,96126,96216,96300,96323,96360,96375,96428,96466,96579,96617,96641,96743,96772,96893,96914,96937,96952,97125,97154,97197,97297,97319,97409,97410,97433,97501,97618,97679,97866,97901,98065,98181,98228,98245,98285,98348,98513,98529,98649,98659,98831,99051,99070,99173,99428,99449,99467,99520,99569,99668,99722,99744,99829,99928",
          "17,240,654,735,786,864,941,976,1072,1097,1369,1397,1706,1738,1885,2041,2072,2144,2162,2204,2445,2554,2592,2683,2716,2828,2996,3076,3195,3333,3370,3380,3472,3519,3596,3639,3724,3827,3831,4026,4027,4163,4179,4335,4348,4573,4610,4628,4657,4687,4770,4809,4815,5021,5048,5130,5154,5221,5543,5588,5666,5679,5688,5691,5708,5883,5917,5963,5984,6180,6206,6225,6319,6565,6567,6653,6789,6794,6815,7035,7082,7098,7142,7173,7192,7193,7300,7399,7424,7450,7585,7651,7675,7804,7806,7845,7878,7903,7962,8051,8186,8196,8525,8533,8546,8563,8619,8728,8785,8987,9155,9299,9308,9314,9452,9522,9727,9811,9830,9875,9904,9926,9957,9973,9975,10048,10119,10135,10223,10280,10319,10335,10340,10438,10486,10654,10755,10783,10802,10845,10851,10898,11378,11486,11498,11528,11576,11622,11638,11656,11669,11746,11776,11802,11911,12051,12074,12090,12108,12141,12330,12434,12611,12667,12802,12884,12965,13016,13125,13243,13331,13345,13401,13448,13509,13580,13775,13801,13827,13941,13992,13997,14044,14064,14182,14183,14251,14255,14403,14550,14642,14725,14765,14767,14773,14942,14977,15013,15055,15083,15142,15191,15241,15313,15327,15487,15614,15718,15721,15834,15996,16035,16108,16125,16144,16147,16252,16293,16345,16359,16431,16450,16484,16545,16573,16628,16695,16881,16968,17044,17094,17162,17189,17219,17395,17488,17539,17692,17813,17903,17955,17957,18031,18091,18093,18302,18392,18401,18433,18474,18501,18551,18714,18733,18767,18834,18845,18854,19086,19260,19590,19636,19669,19700,19745,19767,19777,19783,19797,19845,19897,19925,20108,20114,20184,20191,20214,20354,20402,20437,20544,20624,20705,20861,20949,20975,21056,21100,21116,21154,21249,21289,21354,21467,21534,21707,21711,21732,21756,21761,21781,21838,21954,22037,22059,22060,22119,22125,22179,22223,22346,22559,22565,22597,22602,22604,22616,22620,22633,22636,22784,22849,22914,23021,23046,23135,23187,23257,23261,23265,23295,23324,23412,23415,23456,23609,23613,23645,23780,23820,23949,23985,24061,24065,24134,24241,24350,24384,24394,24509,24552,24564,24606,24632,24643,24711,24743,24779,24864,25033,25069,25115,25120,25291,25297,25308,25332,25443,25495,25512,25522,25528,25897,25947,25986,25993,26050,26055,26120,26150,26154,26202,26212,26251,26257,26277,26510,26575,26613,26633,26708,26738,26757,26886,26925,27019,27038,27048,27126,27139,27192,27454,27521,27562,27576,27581,27635,27815,27847,27906,27987,27990,28117,28119,28135,28165,28228,28325,28466,28508,28541,28569,28595,28604,28703,28717,28783,28928,29071,29200,29232,29241,29376,29629,29667,29767,29796,29806,30019,30051,30071,30136,30141,30155,30182,30218,30748,30793,31005,31014,31107,31134,31177,31192,31201,31226,31291,31389,31418,31470,31522,31548,31556,31714,31717,31837,31888,32059,32119,32144,32254,32288,32433,32462,32479,32560,32607,32660,32722,32804,32823,32932,32986,33039,33263,33283,33325,33379,33651,33895,33905,33941,34094,34201,34207,34219,34228,34240,34290,34340,34372,34592,34667,34720,34730,34818,34854,34952,35020,35099,35105,35116,35389,35417,35425,35446,35545,35549,35750,35754,35769,35923,35953,36032,36035,36196,36241,36249,36265,36275,36305,36328,36371,36379,36686,36729,36737,36753,36790,36796,36893,36911,36949,36982,36999,37250,37259,37265,37335,37434,37448,37462,37478,37482,37510,37577,37620,37720,37749,37886,37896,38381,38544,38613,38781,38820,38945,38964,39005,39269,39280,39377,39385,39424,39445,39526,39548,39590,39611,39650,39656,39709,39839,39887,40066,40208,40221,40230,40253,40295,40299,40508,40674,40688,40748,40845,40878,40880,40918,40923,41070,41090,41419,41569,41725,41732,41742,41904,42060,42073,42082,42129,42236,42316,42331,42488,42527,42640,42667,42670,42672,42673,42681,42685,42713,42815,42835,42873,42884,43231,43233,43238,43268,43276,43305,43423,43426,43554,43697,43736,43745,43898,43927,43932,44056,44106,44128,44175,44207,44303,44367,44439,44610,44630,44656,44693,44806,44888,44923,44978,44992,45022,45103,45104,45142,45163,45257,45329,45332,45356,45458,45482,45496,45720,45870,45930,45955,45985,46156,46249,46286,46319,46321,46346,46398,46511,46525,46532,46552,46651,46664,46689,46700,46707,46756,46849,46937,46986,47063,47171,47174,47452,47589,47627,47644,47786,47788,47810,47985,48058,48129,48195,48527,48531,48540,48672,48720,48737,48780,48835,48843,48897,48997,49008,49110,49137,49278,49325,49396,49525,49586,49595,49683,49746,49862,49894,49955,49960,50077,50169,50354,50444,50621,50648,50690,50777,50861,50898,50959,51086,51296,51327,51333,51506,51966,52040,52081,52095,52206,52361,52413,52439,52543,52573,52658,52677,52765,52824,52923,52936,53016,53260,53276,53349,53459,53571,53591,53638,53746,53794,54014,54154,54183,54364,54371,54392,54506,54555,54616,54632,54981,55034,55095,55220,55329,55375,55383,55672,55685,55796,55949,56010,56122,56127,56133,56230,56271,56282,56326,56342,56386,56491,56509,56625,56695,56998,57253,57286,57321,57337,57474,57497,57528,57639,57734,57822,57963,58028,58055,58086,58100,58301,58320,58366,58404,58431,58447,58454,58515,58565,58594,58642,58658,58753,58811,58837,58907,59062,59067,59090,59194,59210,59220,59231,59316,59433,59558,59601,59612,59625,59654,59659,59816,59870,60047,60081,60095,60103,60409,60530,60545,60674,60686,61029,61142,61216,61565,61595,61635,61677,61686,61934,61981,62022,62087,62100,62121,62337,62516,62564,62582,62798,62818,62859,62882,62970,63050,63083,63132,63172,63183,63190,63322,63384,63428,63431,63529,63651,63694,63704,63829,63834,63848,63874,63937,64009,64180,64282,64338,64483,64621,64762,64800,64885,64909,65050,65110,65183,65322,65332,65366,65681,65712,65789,65923,65941,66051,66136,66154,66182,66184,66232,66272,66544,66681,66817,66829,66945,66946,66970,66996,67007,67117,67362,67407,67430,67456,67590,67595,67615,67652,67659,67687,67791,67811,67842,67847,67894,67928,67954,67973,68137,68196,68392,68471,68549,68554,68709,68750,68778,68805,68853,68892,69024,69028,69181,69186,69200,69313,69323,69367,69385,69478,69494,69603,69648,69716,69731,69753,69763,69973,69976,70077,70293,70348,70416,70498,70529,70538,70568,70604,70694,70701,70744,71020,71068,71390,71483,71486,71515,71660,71725,71771,71803,71932,71994,72049,72068,72187,72234,72271,72281,72309,72318,72324,72341,72375,72384,72432,72458,72547,72600,72601,72776,72794,72805,72821,73015,73047,73087,73270,73275,73312,73368,73512,73538,73569,73620,73759,73835,73920,73938,73999,74028,74063,74111,74322,74351,74503,74581,74633,74818,74837,75029,75117,75129,75138,75166,75225,75239,75245,75317,75321,75361,75446,75497,75527,75585,75590,75818,75836,75906,75983,76023,76079,76085,76138,76202,76545,76558,76641,76653,76769,76827,76828,76938,76991,77138,77247,77278,77507,77589,77694,77715,77736,77877,77890,77955,77982,78428,78621,78630,78658,78729,78734,78818,78871,78912,78950,78978,79067,79150,79209,79256,79368,79465,79507,79508,79531,79606,79626,79685,79746,79749,79978,80133,80363,80367,80373,80527,80564,80585,80619,80658,80676,80710,80761,80825,80974,81042,81250,81254,81364,81448,81655,81665,81767,81829,81987,82053,82077,82102,82211,82222,82241,82274,82356,82363,82447,82481,82605,82607,82657,82755,83010,83072,83074,83080,83114,83136,83176,83183,83250,83409,83446,83511,83527,83668,83718,83733,83740,83761,83855,83969,83994,84056,84206,84299,84534,84596,84778,84862,84898,84968,84972,85025,85049,85060,85079,85247,85566,85723,85736,85767,85915,85991,86011,86099,86198,86211,86296,86429,86503,86509,86626,86651,86722,86727,86733,86957,86958,87049,87054,87144,87173,87291,87377,87382,87439,87513,87559,87688,87794,87842,87903,87917,87981,88008,88016,88037,88055,88135,88228,88263,88265,88276,88390,88606,88710,88804,88852,88958,89074,89083,89098,89219,89304,89307,89326,89462,89691,89724,89732,89816,89830,89937,89993,90029,90093,90096,90205,90391,90395,90567,90590,90612,90723,90738,90815,90861,90886,90960,91041,91252,91515,91620,91644,91735,91824,91914,91965,91982,92232,92266,92385,92396,92442,92452,92478,92581,92600,92603,92630,92808,92870,92988,93060,93146,93324,93330,93372,93492,93500,93506,93531,93565,93619,93682,93733,93818,93884,93898,93951,94060,94186,94215,94360,94366,94588,94598,94802,94822,94863,94897,94971,95042,95108,95272,95297,95324,95370,95400,95478,95612,95620,95624,95905,95950,96005,96067,96077,96159,96254,96270,96369,96439,96543,96638,96685,96695,96781,96813,96899,96956,96984,96986,97093,97126,97317,97338,97550,97674,97706,97722,97775,97814,97816,97918,97967,97972,98016,98171,98253,98287,98294,98296,98310,98316,98360,98413,98498,98641,98711,98742,98818,98824,98836,98865,98874,98951,99205,99323,99343,99414,99542,99550,99586,99663,99703,99737,99739,99902,99973,100003",
          "46,87,150,195,265,326,539,558,682,951,1077,1207,1248,1284,1311,1340,1446,1707,1766,1835,1978,2009,2013,2088,2138,2346,2759,2792,3104,3279,3281,3313,3415,3416,3543,3556,3578,3875,3960,4080,4123,4153,4233,4292,4349,4572,4616,4676,4711,4929,4946,5032,5136,5189,5410,5472,5562,5817,5986,6030,6122,6166,6172,6214,6233,6293,7006,7029,7177,7195,7207,7342,7370,7401,7429,7449,7555,7643,7644,7647,7828,7833,7840,7898,7952,8047,8118,8129,8148,8170,8195,8229,8313,8466,8470,8519,8748,8762,8805,8936,8984,9028,9085,9086,9120,9125,9129,9198,9246,9333,9361,9528,9552,9555,9783,9816,9852,9899,9976,9979,10035,10079,10206,10261,10441,10739,10868,10947,11029,11052,11108,11169,11183,11229,11298,11336,11390,11393,11433,11457,11635,11684,11755,11915,11924,12078,12084,12121,12151,12153,12217,12244,12352,12731,12776,12918,13212,13457,13529,13672,13674,13677,13778,13795,13938,14304,14333,14412,14561,14614,14651,14693,14887,14892,14921,15012,15149,15280,15355,15421,15662,15706,15812,15883,15921,16162,16192,16249,16570,16648,16803,16870,16884,16976,17003,17004,17240,17293,17505,17528,17718,17798,17864,18021,18044,18076,18173,18217,18288,18322,18406,18531,18545,18568,18601,18729,18922,19075,19168,19182,19240,19396,19638,19639,19939,19946,19966,20083,20111,20163,20284,20405,20458,20526,20529,20555,20788,21092,21104,21185,21386,21399,21407,21524,21577,21592,21679,22069,22109,22124,22278,22289,22293,22534,22587,22618,22637,22777,22924,22965,23058,23088,23089,23115,23164,23401,23422,23474,23544,23611,23772,23803,23885,23953,24023,24078,24259,24286,24347,24393,24481,24496,24502,24506,24676,24730,24759,24841,24874,24905,25046,25099,25151,25277,25454,25552,25595,26003,26035,26042,26043,26135,26210,26259,26411,26429,26582,26749,26838,26917,26977,26980,26995,27028,27033,27104,27148,27268,27284,27428,27601,27927,28067,28101,28213,28247,28262,28412,28440,28637,28695,28726,28796,28873,29088,29173,29176,29300,29322,29356,29441,29525,29649,29690,29762,29794,30284,30285,30315,30380,30417,30548,30555,30692,30862,30960,30973,31102,31461,31488,31552,31631,31661,31683,31795,31818,31890,31893,31925,31950,32036,32042,32104,32113,32141,32295,32385,32403,32407,32446,32472,32562,32597,32604,32696,32720,32759,32803,32832,32917,33000,33075,33173,33233,33246,33319,33414,33597,33725,33766,33819,33974,33983,34047,34085,34269,34299,34348,34502,34539,34553,34622,34710,34774,35030,35112,35138,35239,35334,35380,35404,35580,35591,35632,35759,35786,35799,35921,36115,36146,36166,36173,36175,36251,36289,36318,36359,36368,36462,36539,36585,36642,36840,36979,37127,37191,37269,37368,37450,37452,37467,37630,37681,37718,37779,37828,37881,37983,38061,38062,38125,38126,38198,38203,38267,38310,38334,38404,38405,38505,38515,38583,38665,38701,38719,38868,39286,39319,39454,39564,39591,39601,39805,39854,39869,40139,40177,40179,40273,40285,40758,40773,40928,40965,41209,41238,41368,41431,41488,41522,41602,41726,41727,41778,41875,41889,41927,41960,42009,42324,42514,42630,42639,42740,42800,42826,42832,42898,43031,43142,43222,43235,43303,43314,43365,43527,43583,43638,43671,43741,43798,43804,43841,43851,43893,43946,43982,44109,44149,44441,44445,44466,44481,44511,44598,44622,44861,44989,44995,45123,45226,45256,45297,45436,45522,45640,45647,45679,45711,45758,45889,45979,46011,46032,46048,46112,46120,46213,46240,46374,46487,46499,46514,46548,46667,46684,47048,47100,47120,47158,47188,47257,47284,47362,47634,47727,47743,47754,47776,47807,47984,48065,48163,48188,48357,48482,48502,48523,48529,48665,48691,48805,48827,48845,49010,49072,49118,49144,49168,49264,49333,49363,49381,49592,49694,49793,49824,49908,49922,50025,50193,50320,50394,50443,50639,50660,50727,50746,50808,50876,50941,50953,51076,51119,51191,51230,51349,51371,51378,51453,51605,51694,51762,51795,51803,51926,51953,51993,52038,52069,52252,52364,52403,52418,52457,52489,52645,52686,52818,52854,53028,53139,53382,53469,53475,53793,53797,53973,53986,54221,54540,54600,54612,54833,54915,55051,55069,55150,55212,55223,55243,55337,55415,55566,55584,55646,55716,55920,55942,56067,56322,56399,56575,56694,56728,56819,57052,57126,57127,57282,57295,57320,57325,57335,57480,57626,57627,57652,57663,57753,57810,57823,57861,57887,57915,57925,57971,58009,58088,58121,58458,58484,58839,58842,58920,59113,59133,59264,59283,59369,59378,59390,59692,59752,59809,59887,60000,60014,60056,60237,60238,60348,60590,60703,60714,60797,60935,61086,61134,61238,61302,61329,61437,61514,61516,61548,61570,61663,61668,61706,61788,61883,61925,61984,62039,62284,62460,62541,62656,62794,62813,62843,62965,63313,63385,63452,63558,63577,63610,63615,63618,63683,63731,63737,63828,63876,64128,64191,64199,64218,64432,64515,64534,64917,64928,65047,65105,65123,65141,65217,65264,65298,65398,65431,65520,65652,65790,65922,65984,66007,66156,66286,66311,66344,66381,66448,66534,66625,66798,67261,67359,67463,67545,67546,67555,67591,67645,67754,68033,68034,68104,68130,68208,68258,68272,68279,68343,68350,68482,68487,68585,68637,68735,68801,68864,68874,68966,69191,69368,69371,69372,69374,69403,69438,69460,69540,69563,69597,69621,69663,69675,69685,69720,69861,69942,69978,70006,70023,70125,70126,70143,70146,70247,70319,70429,70496,70733,70769,70877,71074,71135,71199,71227,71417,71428,71485,71567,71620,71688,71695,71736,71893,71920,71929,72027,72032,72070,72104,72133,72147,72178,72425,72527,72584,72609,72636,72665,72756,72855,72883,72909,72946,73141,73142,73221,73457,73749,73751,73754,73863,73918,73935,73950,73975,74029,74091,74144,74188,74269,74303,74326,74353,74366,74496,74517,74601,74603,74609,74631,74710,74729,74750,74817,74827,74936,75001,75012,75059,75063,75238,75280,75503,75547,75607,75630,75716,75793,75832,75850,75864,75896,76118,76400,76490,76499,76578,76628,76701,76832,77058,77098,77170,77191,77302,77354,77722,77784,77799,77892,77975,78089,78096,78226,78317,78366,78480,78623,78919,79110,79175,79258,79459,79502,79644,79646,79724,79754,79918,79936,80085,80175,80262,80592,80661,80915,80979,81025,81086,81106,81200,81218,81234,81338,81339,81625,81630,81632,81734,81906,82044,82069,82110,82181,82300,82360,82474,82523,82583,82720,82740,82801,82803,82946,83054,83058,83075,83128,83257,83492,83753,83762,83938,83949,83973,84108,84128,84187,84262,84318,84355,84399,84422,84426,84450,84601,84653,84689,84836,84873,84913,85044,85084,85125,85318,85442,85459,85548,85604,85640,85678,85771,85902,85942,85990,86063,86072,86106,86233,86325,86787,87044,87068,87085,87353,87387,87398,87545,87549,87588,87821,88018,88029,88078,88088,88161,88291,88311,88391,88449,88473,88538,88651,88888,88957,89045,89047,89049,89113,89203,89450,89668,89797,89916,89940,90046,90062,90185,90226,90256,90264,90352,90402,90488,90784,91001,91100,91116,91369,91411,91479,91557,91578,91705,91730,91932,92011,92016,92264,92270,92346,92368,92409,92473,92484,92815,92828,93033,93265,93287,93298,93299,93339,93367,93399,93419,93478,93537,93762,93810,94016,94021,94067,94094,94107,94144,94340,94377,94378,94541,94634,94641,94890,94914,94938,94952,94992,94997,95031,95107,95142,95228,95240,95484,95485,95500,95509,95596,95606,95709,95740,95756,95767,95886,96006,96138,96219,96424,96513,96602,96650,96766,96819,96883,96976,97063,97145,97186,97206,97231,97336,97358,97380,97479,97491,97604,97773,97788,98023,98044,98117,98124,98140,98143,98179,98214,98244,98268,98269,98372,98527,98689,98814,98903,98969,99016,99081,99186,99207,99351,99387,99689,99701,99817",
          "39,161,295,453,631,724,924,952,987,1071,1075,1109,1129,1163,1412,1429,1471,1570,1621,1684,1708,1813,1866,1929,1936,1972,1983,2058,2169,2529,2620,2645,2700,2735,2755,2795,2839,2858,2913,3081,3121,3149,3376,3454,3503,3523,3598,3658,3753,3778,3884,3924,3948,4065,4203,4235,4469,4489,4530,4544,4591,4609,4633,4643,4819,4847,5009,5045,5293,5305,5381,5395,5441,5535,5576,5593,5661,5665,5742,5814,5823,5941,6008,6017,6083,6152,6199,6244,6396,6469,6488,6547,6613,6828,6877,6892,7018,7080,7137,7241,7248,7282,7308,7320,7339,7351,7454,7482,7517,7530,7565,7583,7657,7815,7847,7867,7902,8036,8160,8201,8314,8330,8503,8639,8690,8747,8889,8896,9047,9261,9295,9414,9429,9445,9521,9734,9800,9847,9858,9885,9922,9928,10136,10168,10186,10197,10341,10355,10368,10508,10528,10578,10728,10787,10835,10907,10974,10996,11012,11022,11037,11044,11084,11147,11198,11279,11281,11404,11542,11587,11602,11771,11978,12010,12160,12236,12278,12392,12488,12490,12513,12693,12814,12815,12947,13106,13379,13386,13387,13412,13514,13836,13919,13931,14000,14148,14167,14503,14523,14556,14563,14585,14596,14806,14832,15030,15091,15260,15301,15353,15381,15392,15425,15456,15534,15648,15793,15794,15836,15909,16199,16301,16475,16512,16682,16698,17010,17235,17391,17399,17429,17618,17673,17681,17702,17954,17982,18035,18135,18152,18162,18176,18187,18251,18512,18534,18799,18842,18849,18877,18909,18961,19099,19123,19178,19209,19224,19274,19298,19406,19413,19726,19758,19864,19922,19923,20124,20149,20213,20276,20340,20441,20462,20615,20616,20654,20670,20690,20715,20789,20839,20853,20902,21066,21084,21223,21224,21329,21535,21561,21583,21639,21641,21673,21703,21830,21836,21840,21887,21891,21981,22013,22027,22161,22253,22300,22466,22639,22641,22978,23110,23219,23338,23339,23378,23450,23496,23514,23545,23683,23904,23928,23991,24020,24068,24410,24426,24532,24572,24596,24633,24661,24673,24707,24752,24800,24967,25020,25167,25194,25329,25388,25808,25835,25880,25934,25980,25984,26177,26205,26242,26464,26484,26530,26593,26600,26606,26621,26622,26624,26684,26719,26785,26824,26869,27020,27107,27125,27260,27274,27282,27328,27408,27525,27626,27719,27736,28047,28269,28289,28296,28359,28367,28381,28543,28585,28638,28786,28827,28877,29009,29017,29110,29121,29175,29250,29261,29293,29480,29841,29895,29921,29925,29946,30025,30041,30047,30177,30517,30553,30672,30765,30881,30932,30947,31041,31114,31157,31190,31229,31252,31420,31590,31608,31778,31863,32138,32172,32203,32317,32470,32489,32542,32613,32628,32643,32663,32731,32746,32790,32914,32944,33017,33043,33141,33285,33490,33589,33599,33757,33988,34087,34410,34449,34489,34499,34586,34647,34706,34798,34831,34942,35083,35257,35395,35481,35571,35633,35679,35685,35725,35773,35793,35941,35964,36008,36087,36157,36215,36338,36386,36411,36414,36452,36495,36516,36549,36631,36649,36778,36815,36842,36870,36874,36897,36955,36984,37056,37170,37175,37268,37533,37596,37670,37698,37770,37776,37807,37888,38018,38188,38191,38325,38344,38367,38467,38555,38886,38887,38956,39083,39091,39096,39211,39214,39302,39589,39633,39825,39852,39858,39920,39981,40105,40155,40246,40270,40381,40454,40616,40837,40888,40917,40952,41000,41106,41192,41269,41313,41322,41487,41491,41586,41968,42018,42020,42088,42160,42162,42302,42349,42385,42519,42547,42935,42949,42985,43085,43115,43149,43668,43831,43843,43850,43918,44021,44063,44092,44153,44229,44256,44298,44436,44483,44489,44571,44618,44777,44856,44899,44928,44998,45105,45171,45235,45236,45277,45375,45395,45465,45510,45795,45809,46088,46388,46390,46420,46427,46496,46540,46553,46585,46695,46921,47020,47119,47136,47154,47473,47490,47521,47585,47595,47845,47846,47929,47938,47991,48123,48243,48295,48393,48409,48542,48755,48851,48866,48937,48973,49200,49217,49329,49466,49604,49606,49645,49898,50036,50130,50160,50212,50401,50406,50437,50687,50812,50848,50854,50994,51005,51130,51193,51217,51220,51231,51249,51362,51451,51571,51606,51747,51781,51859,51869,51934,52045,52096,52246,52256,52302,52354,52412,52425,52433,52436,52444,52735,52740,52745,52791,53011,53034,53056,53183,53327,53408,53452,53564,53776,53777,53804,53815,53859,54007,54181,54216,54286,54439,54501,54531,54544,54599,54724,54810,54852,54917,54955,55008,55170,55340,55520,55595,55644,55995,56051,56075,56120,56131,56330,56378,56500,56561,56564,56590,56594,56727,56796,56815,56912,57059,57105,57145,57430,57509,57578,57727,57797,57871,57922,58023,58154,58332,58336,58352,58363,58429,58482,58500,58646,58754,58779,58862,58897,58900,58963,59147,59155,59191,59393,59508,59542,59631,59675,59725,59846,59907,59924,60017,60345,60353,60398,60416,60543,60572,60646,60878,60895,61006,61052,61070,61097,61173,61215,61284,61348,61350,61434,61479,61581,61600,61722,61859,62040,62056,62167,62216,62234,62379,62394,62455,62463,62545,62553,62594,62739,62770,62778,62808,63068,63146,63185,63250,63310,63665,63743,63833,63854,64052,64129,64577,64622,64714,64736,64755,64889,64980,65007,65009,65076,65078,65082,65126,65169,65218,65323,65488,65590,65676,65706,65947,66146,66172,66227,66257,66314,66529,66550,66555,66561,66570,66845,66978,67170,67207,67218,67443,67484,67639,67693,67732,67897,67923,67936,68261,68437,68438,68523,68540,68704,68728,68837,69160,69225,69226,69251,69445,69531,69536,69623,69649,69759,69772,69842,69940,70020,70128,70351,70383,70462,70637,70798,70809,70953,71000,71017,71113,71128,71225,71230,71264,71388,71445,71510,71587,71706,71769,71815,71822,71859,71890,71933,71958,72033,72089,72154,72193,72427,72482,72618,72651,72742,72777,72873,72895,72907,72912,73018,73028,73217,73334,73375,73420,73503,73527,73598,73726,73766,73776,73796,73832,73907,73951,74129,74208,74449,74451,74662,74719,74841,75026,75094,75102,75453,75608,75675,76004,76031,76063,76210,76225,76230,76278,76288,76291,76317,76403,76448,76540,76718,76738,76889,76901,76931,77043,77066,77069,77160,77260,77326,77607,77650,77906,77913,78151,78184,78526,78549,78757,78793,78844,78956,78991,79054,79056,79218,79363,79380,79506,79534,79578,79785,79803,79875,79900,79942,79989,80008,80080,80128,80247,80450,80654,80760,81028,81134,81152,81203,81258,81293,81406,81497,81658,81717,81776,81784,81835,82011,82061,82065,82099,82502,82519,82529,82537,82650,82667,82703,82714,82719,82796,82875,82880,83019,83145,83203,83211,83454,83576,83734,83777,83850,83999,84168,84223,84265,84271,84328,84474,84493,84501,84556,84631,84756,85019,85067,85077,85143,85237,85244,85341,85438,85595,85662,85963,86005,86082,86131,86281,86358,86486,86629,86842,86867,86870,86967,86991,87110,87452,87465,87476,87506,87584,87687,87857,87958,87966,87977,88038,88112,88164,88217,88357,88387,88455,88534,88567,88570,88740,88889,88905,88928,88982,89193,89209,89488,89498,89570,89604,89642,89742,89902,89929,90094,90129,90150,90230,90257,90343,90347,90368,90426,90474,90503,90557,90647,90735,90758,90768,90862,91019,91038,91136,91283,91291,91318,91382,91427,91633,91739,91804,91977,91988,92032,92047,92089,92586,92864,93024,93122,93273,93359,93553,93742,93752,93755,93890,93968,93988,93990,94006,94114,94148,94308,94387,94513,94525,94550,94584,94611,94745,94867,94884,95065,95097,95159,95170,95283,95327,95382,95515,95519,95559,95595,95652,95878,95879,95921,95978,95998,96018,96032,96064,96081,96088,96121,96135,96191,96204,96231,96473,96571,96605,96919,96939,97182,97212,97233,97235,97257,97655,97779,97795,97874,97876,97946,98039,98131,98142,98505,98522,98737,98806,98835,99100,99142,99172,99180,99215,99244,99266,99304,99328,99419,99425,99455,99464,99508,99658,99670,99676,99764,99932",
          "25,91,106,280,369,549,622,707,744,929,972,1038,1099,1151,1198,1218,1239,1374,1382,1476,1509,1675,1691,1714,1720,1730,1810,1847,1873,1911,1919,1951,2024,2044,2134,2233,2241,2316,2390,2458,2478,2502,2551,2706,2857,2874,2881,2899,2906,3255,3310,3337,3347,3397,3488,3640,3756,3773,3859,3962,4049,4061,4106,4129,4135,4136,4155,4223,4237,4317,4400,4431,4470,4485,4552,4577,4677,4706,4712,4769,4800,4869,4971,5253,5330,5659,5662,5907,5977,6013,6099,6107,6171,6226,6297,6316,6373,6608,6751,6784,6810,7042,7064,7079,7162,7270,7359,7469,7507,7653,7758,7764,7786,7826,7888,8339,8373,8410,8415,8448,8500,8501,8514,8626,8630,8683,8693,8858,8881,8977,9093,9167,9171,9232,9300,9345,9444,9457,9479,9547,9613,9623,9685,9715,9815,9828,9970,9994,10134,10479,10493,10503,10511,10524,10624,10656,10673,10675,10692,10723,10749,10839,11061,11068,11165,11232,11268,11297,11491,11601,11680,11697,11705,11723,11847,11851,12026,12067,12093,12149,12156,12339,12340,12382,12414,12590,12653,12749,12779,12785,12811,12907,12921,12966,12988,13005,13034,13043,13092,13160,13181,13191,13202,13220,13265,13333,13567,13676,13734,14004,14049,14314,14475,14488,14530,14533,14660,14744,14826,14857,14937,15032,15169,15177,15180,15232,15246,15291,15317,15333,15336,15592,15653,15741,15754,15757,15805,15867,15891,15952,15969,15986,16005,16131,16156,16193,16212,16275,16540,16656,16749,16849,16863,16928,16962,17025,17036,17075,17136,17184,17205,17286,17452,17518,17562,17779,17806,17878,17910,18054,18123,18214,18400,18412,18426,18481,18484,18582,18662,18711,18786,18869,18870,18885,19059,19069,19083,19129,19144,19248,19301,19494,19868,19963,20045,20091,20102,20142,20303,20348,20400,20461,20470,20571,20608,20632,20727,20754,20765,20850,20863,20888,20932,21027,21071,21113,21162,21184,21220,21284,21288,21295,21297,21356,21393,21395,21461,21572,21601,21715,21727,21771,21776,21880,21898,21972,22142,22160,22306,22324,22410,22487,22518,22644,22728,22897,22902,23136,23278,23332,23574,23655,23665,23674,23735,23788,23795,23918,23942,23964,24010,24155,24280,24370,24391,24404,24409,24471,24497,24708,24788,24826,24876,24894,24984,25170,25319,25343,25351,25472,25490,25542,25549,25797,25877,25900,26127,26136,26164,26173,26279,26302,26387,26399,26447,26515,26531,26642,26666,26683,26696,26703,26790,26793,26937,27252,27479,27539,27607,27733,27751,27902,27943,27947,27964,28020,28125,28211,28223,28337,28374,28395,28469,28582,28606,28744,28770,28780,28862,28891,29151,29166,29185,29191,29214,29239,29256,29393,29404,29415,29432,29442,29631,29655,29708,29906,29924,29932,29945,29985,30030,30107,30130,30146,30189,30226,30255,30265,30392,30411,30457,30498,30532,30558,30578,30584,30721,30728,30792,30916,30959,30971,31054,31098,31167,31173,31455,31459,31471,31510,31579,31653,31713,31785,31937,32013,32043,32079,32214,32223,32268,32450,32518,32583,32623,32653,32700,32743,32770,32777,32792,32816,32824,33269,33282,33582,33705,33718,33784,33843,33884,33950,33970,34032,34121,34135,34286,34313,34371,34413,34477,34527,34540,34702,34808,34975,35288,35296,35441,35443,35444,35484,35564,35680,35692,35820,35896,36183,36206,36225,36316,36346,36463,36498,36505,36523,36534,36698,36755,36758,36883,37026,37062,37100,37108,37310,37328,37477,37583,37646,37675,37690,37806,37995,37997,38067,38172,38178,38184,38206,38397,38450,38508,38607,38681,38688,38733,38822,38828,38847,38853,38865,38900,38961,39002,39003,39015,39060,39132,39135,39160,39221,39225,39311,39324,39483,39555,39584,39629,39632,39732,39868,39972,39987,40011,40052,40198,40232,40263,40292,40479,40526,40562,40582,40585,40776,40819,40860,40890,40907,41072,41141,41148,41150,41226,41337,41399,41424,41469,41507,41528,41891,41933,41950,42013,42051,42074,42246,42259,42305,42469,42522,42699,42787,42818,42824,42859,42876,42879,42920,43044,43074,43092,43169,43197,43417,43425,43450,43454,43477,43491,43510,43599,43607,43647,43716,43801,43809,43811,43876,43973,44054,44090,44116,44125,44190,44294,44330,44401,44472,44496,44519,44576,44601,44632,44674,44773,44791,44836,44921,44941,45153,45183,45215,45242,45300,45303,45322,45421,45434,45484,45501,45819,45943,46087,46223,46242,46252,46325,46336,46397,46518,46692,46724,46731,46831,46958,46979,47076,47102,47287,47288,47353,47415,47495,47825,47866,47886,47997,48004,48134,48177,48194,48272,48307,48343,48359,48545,48583,48625,48660,48702,48712,48807,48821,48848,49014,49111,49129,49150,49194,49500,49585,49615,49720,49860,49881,49924,49967,49987,50075,50250,50433,50606,50618,50758,50788,50925,50939,51218,51475,51484,51509,51656,51770,51774,51789,51828,51854,51945,51957,52254,52445,52521,52607,52656,52814,52833,52955,53071,53115,53120,53326,53357,53533,53534,53572,53622,53704,53738,53805,53898,53942,53993,54085,54250,54332,54353,54438,54443,54486,54521,54648,54682,54716,54738,54781,54817,54949,54956,55025,55082,55096,55217,55254,55302,55474,55608,55881,55890,55940,55955,56164,56353,56434,56455,56627,56666,56687,56824,56885,56931,57151,57255,57257,57270,57380,57441,57518,57542,57585,57655,57743,57811,57821,58059,58305,58312,58424,58453,58546,58555,58764,58814,58818,58836,58866,58916,58997,59001,59008,59026,59058,59064,59217,59255,59572,59747,59788,59847,59881,59908,59990,60162,60169,60207,60226,60255,60267,60272,60571,60592,60606,60666,60823,61112,61182,61325,61373,61491,61630,61667,61773,61795,61861,61950,61969,62018,62057,62063,62065,62235,62286,62322,62540,62665,62674,62768,62780,62812,62838,63053,63073,63087,63095,63097,63177,63462,63562,63589,63776,63807,63906,63933,64100,64343,64370,64596,64653,64794,64862,64897,65061,65220,65237,65271,65555,65629,65717,66161,66162,66175,66207,66218,66223,66293,66342,66348,66359,66389,66409,66430,66605,66771,66822,66870,66888,66921,66958,67110,67167,67210,67302,67366,67529,67624,67630,67700,67778,67839,67884,68086,68138,68142,68211,68315,68456,68460,68463,68766,68780,68921,68927,68930,68956,69078,69147,69172,69212,69223,69310,69417,69533,69592,69613,69831,69937,70039,70052,70121,70124,70167,70188,70206,70474,70522,70607,70861,70902,70928,71003,71054,71252,71312,71406,71472,71547,71601,71608,71784,71881,71891,71937,71983,72163,72323,72325,72328,72406,72642,72677,72778,72874,72882,72956,72984,73012,73040,73277,73300,73551,73615,73678,73765,73797,73817,73886,73910,73968,73981,74047,74151,74158,74168,74198,74217,74222,74379,74405,74429,74473,74572,74639,74738,74810,74900,74905,74949,75033,75174,75356,75513,75539,75600,75639,75719,75753,75833,75885,75922,76262,76279,76315,76327,76341,76381,76430,76436,76491,76643,76694,76841,77103,77147,77200,77263,77348,77397,77591,77662,77775,77816,77908,77963,78002,78043,78085,78128,78207,78454,78499,78653,78687,78739,78796,78848,78917,78939,79007,79019,79065,79174,79180,79230,79246,79253,79333,79392,79415,79450,79454,79540,79553,79609,79788,79858,79974,80083,80160,80164,80201,80348,80602,80615,80625,80628,80649,80653,80680,80684,80779,80817,80823,80911,80983,81085,81099,81173,81305,81331,81336,81423,81502,81577,81797,81823,81904,81974,81998,82082,82125,82135,82170,82189,82205,82339,82377,82456,82566,82577,82860,83037,83078,83087,83102,83141,83157,83187,83192,83213,83222,83474,83484,83563,83620,83699,83790,83863,83894,84063,84292,84427,84436,84595,84674,84677,84753,84987,85342,85436,85454,85485,85495,85598,85875,85962,86027,86154,86194,86196,86197,86240,86251,86274,86279,86347,86359,86402,86448,86533,86658,86748,86767,86775,86793,86902,86962,86993,87249,87272,87297,87298,87378,87431,87436,87451,87534,87845,87911,87929,88022,88272,88284,88340,88354,88444,88487,88609,88668,88677,88717,88738,88754,88767,88948,88988,89010,89155,89367,89407,89434,89457,89497,89514,89549,89553,89560,89630,89753,89784,89800,89854,89882,90229,90232,90252,90275,90306,90448,90508,90520,90690,90781,90877,91067,91069,91081,91085,91114,91130,91137,91198,91297,91305,91449,91458,91530,91536,91691,91815,92106,92323,92362,92585,92664,92844,92900,92957,92961,93044,93069,93285,93290,93401,93466,93515,93518,93556,93559,93586,93688,93730,93876,93886,94020,94070,94299,94333,94401,94479,94614,94681,94686,94706,94904,94964,95124,95348,95373,95386,95555,95593,95679,95746,95826,95872,95916,95970,96080,96086,96412,96447,96458,96501,96633,96818,96888,96903,96940,96982,96996,97020,97076,97183,97229,97244,97274,97302,97330,97356,97378,97525,97556,97597,97687,97815,97841,97849,97911,97921,97986,98109,98213,98234,98288,98309,98534,98630,98813,98919,98992,99154,99213,99350,99438,99522,99580,99595,99724",
          "18,35,110,125,181,297,403,504,533,646,676,821,831,863,975,1057,1145,1152,1206,1331,1381,1438,1440,1473,1647,1660,1729,1790,1791,2031,2059,2063,2081,2205,2494,2597,2614,2693,2847,2897,3101,3242,3322,3357,3445,3544,3620,3664,3968,3979,4167,4211,4251,4282,4410,4424,4490,4560,4606,4732,4852,4860,4926,4928,5073,5179,5321,5361,5416,5602,5685,5737,5809,5857,5976,5990,6040,6094,6146,6348,6417,6445,6566,6579,6690,6700,7021,7070,7143,7175,7224,7303,7354,7610,7622,7679,7711,7824,7980,8254,8256,8326,8335,8526,8559,8610,8635,8659,8789,8800,8838,8841,8876,8950,9006,9019,9062,9095,9168,9252,9357,9896,10083,10115,10188,10294,10332,10359,10411,10474,10550,10651,10799,10806,11130,11231,11356,11410,11430,11490,11585,11695,11865,11886,11926,12098,12258,12276,12348,12394,12526,12586,12598,12633,12715,12724,12728,12823,12853,12919,13024,13126,13266,13273,13364,13419,13659,13752,13877,13921,13979,14234,14244,14245,14432,14474,14557,14695,14711,14737,14786,14819,14997,15210,15432,15435,15437,15438,15480,15538,15782,15876,15971,15984,15988,16012,16307,16539,16733,16784,16831,16843,16902,16906,17082,17112,17274,17361,17365,17382,17479,17556,17827,17848,17896,17927,18079,18094,18199,18350,18409,18490,18560,18563,18644,18674,18781,19021,19091,19117,19279,19372,19510,19561,19584,19621,19945,20024,20056,20231,20295,20350,20392,20639,20683,20823,20838,20855,20979,21004,21230,21355,21403,21431,21473,21513,21546,21573,21621,21623,21734,21819,21846,21931,21967,21987,22003,22034,22199,22464,22497,22731,22751,22767,22800,22882,22884,23100,23166,23266,23468,23494,23659,23699,23776,23870,23881,23894,23966,23969,24000,24004,24075,24272,24309,24311,24418,24419,24680,24762,24802,24828,24934,24945,25131,25404,25473,25517,25892,25972,26015,26065,26313,26372,26725,26741,26806,26835,26985,26987,27047,27073,27096,27124,27143,27348,27444,27508,27520,27563,27592,27623,27765,27884,27916,27959,28016,28065,28077,28321,28421,28444,28462,28663,28712,28866,28997,29010,29201,29321,29341,29370,29388,29455,29542,29547,29585,29822,29852,29871,29901,29903,29942,29969,29973,29978,30002,30175,30201,30557,30655,30802,30828,30953,30963,31090,31133,31186,31197,31309,31583,31727,31865,31899,31909,32244,32297,32449,32586,32608,32610,32682,33023,33032,33090,33098,33224,33231,33311,33404,33406,33431,33551,33663,33731,33785,33962,33971,34081,34159,34226,34402,34666,34684,34792,34858,35017,35161,35172,35195,35236,35263,35307,35429,35494,35496,35523,35544,35837,35839,35888,35954,36017,36287,36307,36416,36437,36442,36603,36710,36775,36835,36854,36876,36975,37104,37117,37140,37144,37176,37474,37568,37677,37772,37946,37999,38036,38129,38197,38247,38288,38390,38418,38444,38702,38786,38811,38827,38866,38894,39039,39176,39209,39297,39308,39434,39574,39876,39895,40036,40111,40160,40188,40441,40465,40499,40581,40625,40641,40687,40729,41098,41113,41211,41285,41358,41604,41605,41638,41772,41821,41859,41885,41928,42122,42147,42261,42287,42343,42365,42498,42629,42727,42927,42942,42944,42960,42970,43086,43089,43150,43345,43415,43472,43502,43517,43529,43624,43739,43758,43799,43810,43832,43937,44011,44061,44081,44167,44217,44218,44349,44354,44597,44709,44754,44790,44848,44855,44886,45106,45124,45160,45202,45403,45422,45521,45658,45742,45796,45920,45942,45945,46407,46456,46507,46625,46742,46806,46889,47059,47243,47273,47301,47305,47334,47432,47469,47513,47522,47638,47655,47693,47860,47926,47992,47999,48066,48302,48440,48577,48675,48725,48757,49057,49069,49104,49187,49220,49243,49315,49341,49445,49549,49556,49612,50197,50302,50416,50429,50438,50605,50703,50956,50976,51131,51370,51429,51477,51500,51661,51704,51811,51885,51888,51969,52002,52116,52202,52263,52337,52369,52466,52699,52813,52817,52880,52960,53312,53325,53422,53438,53485,53495,53507,53578,53581,53584,53628,53655,53695,53786,53835,53880,53958,53989,54000,54167,54248,54391,54488,54519,54592,54672,54769,54825,54840,54868,54881,54984,55003,55056,55172,55178,55200,55269,55320,55418,55421,55467,55477,55562,55668,55758,56116,56117,56286,56350,56470,56582,56614,56623,56655,56768,56793,56988,57058,57080,57298,57323,57466,57517,57587,57641,57691,57695,57817,57898,58004,58015,58030,58072,58293,58395,58438,58543,58545,58605,58636,58768,58912,58930,58937,58999,59050,59251,59261,59302,59447,59473,59594,59650,59732,59748,59761,59980,60001,60024,60033,60045,60061,60083,60144,60339,60499,60537,60678,60766,60909,60921,60931,60968,60981,61000,61267,61315,61356,61442,61445,61582,61632,61685,61703,61791,61805,61911,61993,62055,62109,62163,62183,62357,62397,62525,62689,62748,62851,62880,62952,62982,63060,63123,63145,63276,63288,63376,63700,63822,63989,64244,64318,64424,64428,64466,64521,64618,64645,64737,64821,64884,65066,65124,65129,65266,65377,65408,65539,65549,65670,65679,65784,65797,65858,66000,66085,66310,66326,66391,66396,66508,67105,67114,67262,67318,67356,67389,67877,68125,68259,68282,68311,68341,68402,68423,68534,68737,68832,68900,68940,69019,69290,69318,69404,69449,69529,69692,70234,70278,70638,70931,71096,71158,71284,71384,71491,71492,71622,71654,71744,71780,71970,71991,72012,72054,72132,72412,72480,72564,72580,72702,72798,72806,73059,73076,73112,73383,73443,73445,73489,73591,73623,73674,73805,73821,73986,74031,74128,74175,74204,74246,74250,74323,74342,74350,74499,74560,74734,74811,74836,74935,75030,75076,75165,75193,75222,75378,75416,75430,75471,75480,75592,75611,75655,75692,75698,75831,75956,75969,76117,76238,76349,76391,76423,76534,76585,76695,76762,76942,76955,76974,77064,77068,77118,77164,77181,77213,77248,77301,77459,77503,77553,77679,78000,78044,78072,78175,78290,78292,78324,78370,78385,78393,78591,78612,78625,78646,78820,78859,78998,79443,79527,79542,79598,79618,79898,80065,80071,80076,80092,80124,80131,80225,80230,80255,80338,80339,80439,80546,80576,80745,81000,81050,81063,81105,81147,81239,81351,81482,81697,81751,81796,81820,81992,82032,82056,82367,82380,82552,82620,82691,82696,82759,82792,82903,82931,82964,82965,83012,83133,83421,83441,83602,83709,83719,83752,83884,84036,84083,84173,84229,84454,84457,84649,84680,84725,84867,85000,85062,85154,85222,85452,85608,85655,85706,86024,86173,86227,86258,86485,86602,86612,86659,86683,86697,86702,86795,86973,87041,87130,87299,87323,87587,87612,87615,87627,87789,87792,87803,87870,88054,88098,88200,88247,88258,88333,88462,88562,88690,88697,88783,88841,88893,88980,89006,89180,89196,89394,89436,89448,89466,89607,89719,89733,89749,89760,89861,89875,90013,90091,90139,90190,90216,90345,90370,90387,90502,90658,90674,90679,90859,90899,90985,90989,91025,91086,91095,91131,91142,91175,91197,91246,91269,91366,91413,91419,91439,91473,91478,91697,91767,91770,91832,91854,92066,92554,92621,92658,92767,92813,92822,92880,93334,93337,93446,93454,93543,93601,93709,93786,93809,93918,94153,94267,94303,94533,94571,94594,94605,94677,94689,94754,95056,95155,95177,95213,95252,95389,95462,95488,95501,95505,95603,95789,95795,95858,96076,96140,96197,96199,96282,96327,96382,96471,96532,96792,97105,97191,97230,97342,97387,97548,97615,97616,97682,97690,97744,98075,98125,98151,98176,98239,98261,98395,98503,98569,98590,98646,98661,98838,98849,98861,98882,99020,99140,99288,99541,99702,99781,99993",
          "44,116,163,214,267,323,361,528,599,609,638,650,714,1096,1222,1246,1279,1293,1337,1358,1552,1560,1561,1564,1646,1752,1894,1944,2107,2117,2151,2246,2302,2344,2354,2504,2593,2635,2676,2786,3030,3196,3265,3336,3719,3725,3746,3793,3848,3871,3923,3941,4089,4111,4128,4138,4217,4252,4285,4376,4389,4472,4494,4512,4727,4806,4826,4923,4945,4982,5030,5205,5236,5392,5413,5503,5751,6106,6176,6254,6371,6460,6480,6492,6516,6577,6682,6807,7026,7209,7271,7335,7402,7536,7552,7710,7794,7855,8276,8403,8445,8449,8455,8512,8680,8784,8833,8879,8880,8927,8982,9152,9160,9176,9178,9277,9287,9397,9509,9983,10028,10260,10325,10334,10354,10387,10392,10475,10514,10585,10693,11088,11172,11201,11209,11215,11240,11245,11331,11388,11419,11446,11474,11598,11712,11716,11775,11884,11904,12042,12210,12265,12281,12373,12399,12452,12592,12628,12672,12691,12786,12813,12820,12892,12969,12984,13093,13193,13321,13650,13658,13670,13767,13814,14032,14105,14143,14191,14250,14345,14381,14616,14625,14686,14700,14808,14941,15026,15054,15085,15274,15337,15700,15751,15774,16007,16033,16044,16259,16325,16337,16399,16490,16526,16542,16558,16741,16878,16954,17009,17084,17196,17199,17265,17268,17378,17381,17511,17598,17651,17674,17747,17770,17849,17862,17909,17950,18003,18032,18074,18090,18244,18252,18260,18349,18475,18573,18595,18604,18709,18808,18833,18846,18925,18943,18957,18982,19203,19204,19412,19433,19504,19701,19739,20050,20099,20195,20255,20321,20417,20449,20456,20523,20534,20586,20613,20720,20732,20783,20988,21026,21127,21397,21555,21637,21692,21738,21767,21786,21884,21941,21990,22077,22257,22287,22438,22455,22516,22572,22625,22660,22686,22688,22856,22895,22929,22941,23380,23538,23739,23888,23892,24043,24240,24449,24465,24567,24629,24649,24740,24805,24827,24845,24959,25008,25050,25052,25110,25142,25144,25236,25239,25289,25340,25444,25447,25556,25692,25769,25790,25818,25901,26123,26315,26520,26541,26760,26930,27075,27097,27153,27164,27690,27786,27812,27898,27905,27930,27985,28011,28163,28214,28385,28464,28512,28580,28755,28929,29078,29089,29107,29111,29292,29420,29590,29626,29643,29761,29857,30035,30060,30150,30161,30199,30369,30562,30643,30738,30887,31055,31103,31104,31146,31154,31174,31195,31276,31300,31301,31364,31437,31574,31784,31878,31892,31921,31994,32277,32492,32546,32860,32960,33007,33286,33334,33574,33590,33614,33745,33812,33851,33877,33931,33955,34172,34309,34354,34386,34425,34427,34433,34506,34546,34609,34705,34717,34833,34958,35071,35078,35147,35324,35365,35375,35455,35460,35507,35574,35855,35948,36042,36064,36101,36357,36391,36445,36456,36587,36901,36912,36997,37020,37032,37066,37230,37271,37288,37329,37357,38161,38229,38308,38322,38338,38351,38402,38412,38428,38504,38647,38648,38687,38752,38810,38872,38881,39108,39342,39615,39625,39688,39741,39753,39782,40083,40129,40196,40290,40474,40520,40634,40805,40810,40813,40848,41103,41142,41182,41428,41442,41666,41677,41733,41853,41954,42055,42270,42394,42437,42504,42563,42741,42882,42893,43005,43036,43039,43042,43262,43410,43587,43601,43605,43766,43780,43855,43912,43917,44001,44045,44049,44130,44171,44187,44194,44225,44275,44296,44333,44521,44553,44671,44732,44776,44864,44900,45013,45039,45056,45086,45114,45165,45334,45481,45548,45558,45564,45728,45841,45863,45896,45912,45939,46013,46100,46274,46292,46413,46520,46610,46637,46739,46784,47071,47354,47359,47372,47529,47530,47573,48109,48161,48204,48365,48389,48444,48650,48853,48872,48918,49034,49054,49085,49108,49206,49228,49321,49394,49404,49424,49433,49504,49608,49800,49850,49932,50002,50007,50129,50134,50269,50347,50403,50446,50488,50567,50582,50587,50616,50719,51157,51309,51457,51874,51902,51923,51971,52308,52383,52459,52483,52503,52741,53042,53170,53194,53197,53239,53330,53336,53363,53405,53635,53670,53681,53684,53739,53749,53853,53855,53936,53959,54034,54267,54370,54425,54552,54567,54638,54752,54757,54783,54831,54919,54961,54978,54995,55332,55373,55427,55443,55464,55516,55522,55525,55634,55683,56162,56232,56333,56438,56502,56586,56603,56631,56702,56785,56880,56983,57007,57021,57128,57197,57358,57372,57467,57498,57558,57565,57584,57640,57767,58081,58211,58231,58280,58349,58451,58479,58496,58710,58712,58722,58783,59045,59056,59230,59459,59477,59604,59933,60019,60050,60064,60113,60161,60178,60220,60223,60419,60454,60472,60704,60750,60873,60957,61198,61237,61324,61346,61451,61517,61646,61744,61829,61914,62116,62498,62534,62719,62732,62872,62903,62909,62956,63000,63290,63480,63853,63895,63905,63918,64001,64023,64069,64119,64382,64526,64550,64555,64572,64605,64793,64826,64845,65036,65160,65212,65286,65290,65334,65409,65512,65614,65695,65737,65739,65918,66014,66237,66266,66516,66620,66738,66762,66781,66805,66922,66956,66998,67023,67271,67315,67347,67384,67466,67518,67593,67625,67644,67657,67733,67769,67786,67869,67880,68011,68026,68041,68047,68185,68389,68459,68498,68512,68581,68595,68596,68708,68731,68768,68886,68913,69030,69074,69204,69370,69400,69428,69432,69535,69602,69614,69697,69702,69841,69885,69969,70047,70059,70087,70122,70230,70255,70369,70555,70588,70684,70793,70847,70854,70865,71112,71117,71200,71294,71353,71489,71640,71661,71778,71856,71917,71922,71999,72082,72088,72126,72357,72421,72496,72689,72707,72733,72741,72753,72987,73017,73038,73118,73202,73369,73389,73449,73485,73487,73500,73549,73625,73660,73665,73691,73839,73901,73919,73928,73998,74043,74077,74078,74092,74199,74244,74309,74587,74588,74647,74700,74708,74776,74855,74875,74898,74940,75009,75066,75340,75409,75553,75572,75625,75640,75755,75812,75964,75978,76022,76047,76206,76295,76520,76554,76670,76679,76748,76758,76772,77000,77034,77353,77531,77684,77776,77875,77956,78059,78119,78391,78403,78412,78517,78546,78688,78810,78853,78949,78972,78985,79042,79076,79098,79142,79166,79426,79437,79532,79577,79808,80271,80300,80313,80386,80390,80438,80463,80497,80547,80551,80571,80664,80795,80821,80918,80962,80975,81052,81178,81208,81264,81376,81471,81585,81587,81737,81880,81903,81933,81975,81989,82113,82252,82336,82453,82591,82772,82791,82846,82937,83161,83194,83196,83373,83496,83571,83579,83603,83925,84046,84075,84458,84479,84491,84628,84669,84714,84757,84804,84841,84843,84870,85092,85102,85153,85221,85504,85586,85639,85714,85721,85811,85978,86017,86112,86254,86333,86338,86412,86457,86647,86674,86676,86814,86946,86985,87373,87470,87713,87714,87767,87986,87995,88026,88236,88315,88362,88617,88657,88672,88702,88838,88904,89169,89354,89629,89696,89770,89838,89941,90127,90192,90234,90297,90307,90351,90399,90453,90534,90640,90703,90879,90897,91031,91060,91190,91233,91279,91443,91447,91471,91477,91511,91587,91784,91974,92190,92263,92271,92272,92450,92500,92609,93005,93084,93155,93178,93228,93237,93513,93656,93796,93959,94001,94069,94073,94103,94165,94222,94244,94298,94363,94471,94477,94963,95041,95049,95068,95225,95227,95362,95528,95581,95632,95753,95902,95911,95966,96020,96024,96055,96060,96109,96141,96148,96173,96203,96227,96242,96248,96308,96356,96527,96586,96653,96675,96704,96736,96758,96802,96815,96851,96998,97045,97062,97075,97116,97253,97324,97468,97602,97628,97686,97704,97887,97907,97956,97979,97990,98240,98473,98535,98577,98581,98786,98803,98896,98958,99010,99122,99132,99139,99314,99336,99377,99380,99465,99518,99590,99697,99743,99787,99795,99856,99926",
          "48,93,139,173,516,616,698,720,899,914,1143,1302,1359,1606,1726,1748,1796,1820,2067,2090,2230,2294,2317,2424,2485,2641,2644,2646,2666,2887,2979,3029,3075,3179,3183,3586,3608,3641,3649,3810,3828,3830,4003,4114,4146,4164,4170,4254,4518,4547,4592,4667,4690,4943,4947,4952,4969,5106,5133,5283,5300,5312,5454,5527,5566,5579,5615,5750,5830,5877,5940,6006,6012,6121,6129,6158,6336,6378,6408,6485,6509,6735,6790,6907,6979,7139,7182,7258,7299,7400,7416,7463,7542,7658,7798,7852,7865,7949,8171,8278,8316,8328,8364,8388,8638,8686,8694,8802,8872,8900,9127,9180,9223,9267,9282,9347,9367,9438,9441,9586,9590,9801,9822,9911,9933,9965,10008,10052,10158,10165,10256,10321,10338,10555,10619,10645,10674,10798,11025,11092,11162,11225,11379,11387,11396,11484,11507,11514,11727,11742,11797,11837,11976,11977,11991,12072,12109,12168,12294,12338,12360,12422,12576,12643,12709,12747,12754,12949,12975,13029,13222,13535,13606,13655,13678,13835,13895,14098,14232,14296,14350,14401,14420,14460,14514,14553,14636,14757,14762,14772,14856,15035,15072,15159,15236,15272,15306,15390,15424,15524,15572,15594,15605,15634,15698,15831,15852,15893,16228,16232,16254,16255,16292,16408,16419,16448,16461,16486,16622,16715,16718,16730,16732,16777,16792,16845,16901,16946,16996,17060,17207,17502,17850,18278,18361,18396,18459,18521,18648,18888,18899,19164,19207,19222,19362,19388,19418,19497,19514,19519,19682,19688,19695,19761,19811,19849,19854,19855,19869,19883,19903,19930,20092,20146,20244,20575,20605,20659,20695,20769,20837,20978,21140,21194,21416,21521,21574,21654,21719,21742,21878,22153,22212,22539,22541,22628,22678,22915,22939,22964,23204,23519,23533,23627,23716,23821,23848,23899,23909,23992,23997,24032,24139,24156,24369,24379,24761,24768,24861,24908,24953,24956,25078,25521,25547,25664,25671,25827,25989,26061,26075,26225,26287,26332,26359,26458,26461,26547,26552,26570,26595,26680,26735,26922,26943,26948,27119,27158,27372,27612,27637,27703,27854,28072,28146,28254,28350,28384,28644,28698,28831,28893,29024,29080,29144,29304,29382,29460,29569,29693,29938,29962,30039,30100,30124,30259,30309,30431,30448,30704,30722,30752,30774,30808,30831,31042,31187,31268,31278,31306,31432,31434,31518,31534,31604,31783,31934,32008,32123,32146,32280,32284,32545,32649,32829,32943,33092,33112,33164,33170,33253,33260,33499,33518,33730,33765,33787,33856,33917,33965,34116,34158,34167,34243,34259,34262,34327,34458,34461,34518,34521,34542,34630,34778,34884,34911,34979,35016,35075,35242,35278,35282,35414,35551,35634,35713,35762,35802,36033,36119,36232,36239,36331,36381,36781,36805,36838,36906,36920,37013,37158,37199,37208,37226,37291,37346,37354,37381,37581,37598,37666,37835,37852,37889,37893,37903,38037,38180,38383,38398,38435,38470,38624,38657,38670,38707,38721,38744,38754,38775,38798,38824,38867,39022,39025,39029,39212,39234,39258,39265,39351,39453,39552,39559,39779,39785,39794,39865,39963,39966,40041,40060,40090,40123,40219,40318,40467,40575,40653,40655,40977,40982,41244,41265,41314,41530,41538,41588,41642,41862,41929,42186,42242,42320,42338,42420,42554,42557,42747,42933,43349,43395,43744,43747,43908,43936,43962,44024,44025,44027,44124,44421,44471,44547,44609,44644,44764,44792,44794,44969,44984,45262,45281,45397,45423,45429,45430,45432,45448,45608,45634,45790,45908,46005,46015,46063,46167,46339,46340,46361,46378,46447,46529,46536,46670,46732,46782,46963,46967,46978,46996,47066,47361,47374,47416,47488,47601,47630,47821,47822,47908,48088,48098,48142,48410,48612,49045,49058,49092,49271,49273,49429,49432,49598,49620,49748,49853,49889,49915,49978,50032,50067,50222,50558,50590,50672,50776,50849,50962,51057,51167,51192,51498,51564,51740,51907,52013,52057,52128,52464,52502,52522,52593,52816,52916,52921,52934,53164,53219,53222,53368,53437,53446,53619,53886,53917,54081,54105,54111,54198,54232,54367,54468,54566,54644,54657,54711,54848,54971,55017,55046,55258,55288,55408,55618,55658,55795,55797,55886,55905,55926,56214,56425,56856,56941,56971,56978,57106,57198,57327,57375,57389,57511,57738,57782,57860,57950,57975,58170,58176,58288,58410,58414,58443,58561,58589,58743,58821,58879,58943,58989,59072,59117,59172,59252,59266,59273,59457,59467,59470,59532,59579,59793,59871,59898,59922,60041,60059,60069,60147,60212,60331,60496,60627,60706,60912,60991,60994,61211,61312,61447,61495,61661,61740,61926,61938,62012,62030,62092,62128,62137,62140,62177,62202,62258,62396,62419,62432,62697,62727,62861,62918,63033,63080,63081,63208,63281,63356,63444,63532,63619,63637,63681,63691,63742,63754,64215,64668,64671,64751,64787,64810,64847,64968,65022,65040,65107,65246,65267,65371,65397,65501,65578,65634,65767,65835,65998,66079,66215,66456,66489,66731,66785,66891,66962,67068,67118,67160,67238,67243,67258,67286,67310,67319,67320,67382,67402,67495,67617,67750,67801,67807,67917,67949,67962,68018,68232,68396,68467,68690,68722,69111,69112,69211,69359,69585,69717,69851,69925,69938,70071,70119,70132,70237,70272,70296,70317,70329,70335,70355,70430,70566,70639,70666,70893,70922,71045,71137,71260,71391,71677,71745,71751,71911,72125,72195,72300,72450,72491,72520,72523,72576,72635,72639,72764,72838,73193,73214,73238,73590,73640,73829,73945,73966,74170,74226,74278,74381,74521,74536,74653,74695,74711,74909,75018,75086,75180,75200,75265,75451,75475,75545,75617,75668,75671,75682,75737,75860,75869,75951,76052,76135,76143,76181,76189,76313,76369,76390,76486,76509,76801,76802,76888,76954,77071,77091,77188,77368,77477,77501,77539,77564,77580,77669,78091,78101,78194,78202,78221,78446,78643,78672,78690,78731,79033,79235,79324,79330,79359,79478,79516,79659,79741,79801,79825,79845,79896,80045,80166,80423,80532,80721,80785,80931,80963,80973,81177,81226,81288,81327,81393,81508,81535,81551,81690,81712,81877,81889,82086,82325,82413,82445,82478,82486,82488,82651,82676,82677,82686,82797,82834,82942,83167,83251,83317,83395,83431,83473,83482,83559,83716,83926,84013,84027,84114,84224,84237,84431,84574,84643,84679,84743,84880,85009,85037,85047,85138,85173,85281,85381,85433,85547,85592,85676,85752,85768,85901,86043,86124,86267,86483,86528,86561,86669,86869,86876,86901,86976,87003,87013,87133,87145,87235,87279,87324,87335,87336,87386,87412,87605,87779,87835,87926,87932,87933,88025,88102,88227,88234,88239,88285,88410,88420,88490,88506,88596,88622,88714,88778,88809,88920,89044,89142,89228,89251,89276,89381,89390,89471,89595,89600,89633,89802,89844,89879,89973,90030,90090,90101,90102,90130,90152,90158,90227,90316,90504,90687,90701,90732,90805,90837,90838,90930,90938,90939,90968,91013,91043,91055,91129,91230,91241,91264,91277,91287,91344,91368,91374,91425,91467,91472,91490,91535,91614,91688,91777,91782,91801,91829,91876,91922,91981,92183,92277,92289,92497,92551,92568,92650,92832,92843,92891,92946,92955,92981,92986,93049,93053,93103,93105,93132,93185,93216,93283,93387,93450,93474,93615,93760,93868,93946,93954,94112,94125,94214,94365,94373,94473,94527,94590,94627,94676,94773,94858,94998,95218,95332,95351,95367,95394,95433,95451,95460,95553,95648,95720,95956,95973,96012,96027,96053,96069,96319,96353,96377,96400,96519,96531,96719,96760,96924,96997,97052,97117,97180,97385,97558,97709,97715,97884,97916,97989,98030,98034,98092,98174,98365,98371,98539,98755,98760,99015,99093,99216,99347,99389,99402,99539,99632,99726,99818",
          "151,169,228,314,350,577,585,704,721,791,1000,1074,1227,1264,1332,1362,1373,1480,1563,1783,1834,1898,1938,1945,1946,2046,2147,2366,2385,2394,2474,2689,2753,2764,2833,2964,3164,3175,3182,3348,3352,3430,3497,3553,3689,3733,3743,3795,3805,4091,4095,4204,4257,4380,4403,4437,4464,4714,4795,4884,4978,4995,5050,5519,5683,5715,5729,5813,5847,5950,6141,6312,6358,6393,6407,6439,6726,6752,6860,7099,7163,7286,7407,7414,7470,7499,7503,7504,7558,7628,7849,8086,8098,8143,8174,8308,8342,8350,8443,8567,8599,8717,8844,8891,9037,9206,9294,9343,9410,9448,9572,9645,9687,9693,9782,9842,10096,10369,10371,10375,10431,10521,10531,10712,10874,10985,11007,11115,11156,11213,11227,11300,11321,11322,11609,11623,11650,11679,11840,11912,11967,12097,12140,12268,12310,12326,12334,12398,12437,12446,12524,12527,12593,12662,12742,12927,12958,12991,13252,13268,13290,13298,13363,13483,13538,13588,13749,13897,13962,14246,14472,14638,14644,14794,14854,14987,15092,15202,15233,15292,15321,15431,15531,15587,15670,15786,15851,16091,16165,16185,16389,16452,16652,16781,16799,16905,16958,17281,17588,17763,17771,18050,18128,18160,18172,18256,18264,18316,18402,18585,18608,18669,18814,18863,18910,19036,19119,19269,19318,19382,19408,19416,19520,19589,19614,19756,19773,19912,19955,19967,20106,20162,20362,20410,20536,20592,20619,20852,20954,21070,21145,21173,21336,21338,21413,21503,21752,21927,21971,21976,22014,22033,22093,22136,22141,22336,22351,22414,22568,22600,22699,22747,22798,22814,22881,23288,23330,23491,23517,23521,23578,23702,23713,24069,24077,24100,24146,24348,24429,24472,24634,24690,24693,24925,24974,25174,25183,25208,25220,25273,25295,25301,25311,25315,25523,25537,25667,25726,25779,25804,25838,25849,25967,26153,26291,26561,26690,26717,26775,26834,26976,27235,27261,27294,27474,27518,27541,27791,27923,28023,28149,28340,28448,28597,28600,28696,28720,28922,29042,29069,29092,29157,29272,29328,29512,29513,29516,29543,29554,29617,29621,29684,29696,29703,29928,30215,30231,30243,30293,30374,30440,30446,30482,30577,30580,30660,30742,30858,31013,31198,31265,31356,31428,31479,31531,31627,31807,31834,31853,31962,31968,32076,32086,32110,32137,32266,32337,32339,32398,32480,32501,32650,32703,32826,32893,32911,33035,33059,33085,33172,33208,33217,33281,33388,33546,33726,33829,34169,34186,34230,34234,34257,34261,34333,34417,34494,34507,34618,34625,34642,34745,34972,35035,35040,35057,35155,35314,35357,35390,35412,35506,35555,35581,35723,35763,35770,35785,35998,36012,36025,36319,36518,36765,36966,37115,37121,37154,37201,37203,37301,37361,37438,37463,37493,37527,37549,37866,37961,38086,38152,38305,38379,38391,38438,38465,38469,38524,38645,38652,38727,39006,39101,39111,39117,39237,39282,39470,39661,39695,39983,40033,40289,40346,40386,40414,40541,40555,40643,40678,40816,40876,41007,41053,41187,41227,41245,41366,41385,41509,41521,41648,41906,41979,42007,42012,42016,42042,42071,42237,42323,42480,42579,42692,42756,42768,42890,42963,43023,43194,43258,43278,43321,43504,43546,43754,43883,43993,44013,44111,44117,44246,44340,44452,44783,45028,45249,45260,45337,45453,45677,45767,45822,45881,45924,46052,46152,46318,46419,46589,46631,46822,46833,46883,46898,46900,46928,46941,47032,47065,47075,47137,47192,47262,47269,47344,47373,47545,47596,47613,47654,47781,47864,47943,47976,48145,48203,48508,48603,48904,49006,49037,49068,49075,49146,49181,49193,49285,49322,49357,49388,49516,49519,49529,49569,49697,49718,49874,49952,50246,50324,50353,50389,50617,50887,50889,50919,50951,51108,51140,51144,51248,51380,51455,51466,51641,51722,51899,52212,52293,52569,52583,52721,52888,52920,53054,53253,53559,53656,53716,54146,54273,54311,54415,54542,54624,54637,54920,54992,54993,55507,55513,55605,55656,55782,55837,55951,56023,56119,56168,56216,56297,56499,56593,56704,56753,56908,57053,57207,57315,57496,57514,57561,57657,57895,58000,58005,58069,58134,58147,58270,58271,58338,58388,58616,58638,58776,58952,59007,59077,59503,59586,59626,59756,59859,60004,60022,60077,60140,60668,60711,60751,60837,60961,61180,61337,61349,61422,61533,61536,61572,61748,61763,61838,61845,61949,61995,62189,62263,62330,62496,62603,62737,62833,62860,63026,63131,63227,63325,63372,63553,63586,63805,63817,63964,63971,64012,64051,64227,64385,64390,64539,64593,64659,64661,64742,64815,64830,64913,65014,65180,65187,65295,65321,65374,65381,65438,65538,65686,65703,65849,65905,66043,66113,66295,66510,66759,66796,66869,66894,66930,67215,67309,67345,67380,67599,67604,67608,67747,67797,67991,68015,68024,68269,68314,68329,68410,68662,68720,68729,68987,69101,69165,69331,69361,69484,69534,69653,69664,69812,69832,70097,70179,70536,70659,70755,71125,71334,71554,71575,71614,71667,71687,71818,71878,71907,71998,72129,72177,72270,72311,72348,72402,72417,72462,72471,72661,72786,72792,73110,73185,73360,73371,73400,73458,73507,73647,73703,73705,73739,73799,73819,73902,74007,74059,74178,74312,74316,74389,74591,74635,74740,74819,74985,75004,75365,75511,75522,75601,75679,75739,75742,75952,76020,76068,76074,76086,76090,76270,76351,76418,76506,76884,76967,76977,77205,77225,77243,77323,77445,77502,77691,77718,77759,77798,77871,77989,78066,78204,78476,78556,78600,78608,78655,78876,79017,79111,79155,79205,79213,79244,79419,79511,79717,79748,79852,79895,80050,80233,80284,80357,80442,80671,80675,80686,80881,80995,81146,81149,81163,81244,81726,81947,82092,82154,82383,82442,82527,82535,82751,82956,83042,83121,83142,83173,83281,83319,83325,83444,83493,83497,83564,83658,83736,83779,83877,83957,84205,84302,84599,84644,84682,84774,84812,84983,84999,85027,85058,85118,85140,85167,85174,85311,85487,85528,85896,85937,86078,86285,86366,86452,86473,86589,86665,86756,86845,86928,86937,86979,87039,87160,87230,87369,87371,87528,87564,87581,87850,88005,88329,88513,88522,88691,88744,88756,88834,88872,88907,89046,89106,89119,89152,89338,89417,89551,89927,89948,90143,90530,90736,90746,90767,90773,90803,90806,90827,90843,90910,91002,91135,91180,91373,91392,91474,91501,91550,91652,91659,91700,91970,91996,92150,92212,92253,92314,92372,92508,92593,92711,92783,92831,92835,93040,93045,93098,93356,93375,93391,93443,93444,93479,93519,93544,93558,93612,94014,94253,94270,94385,94420,94423,94486,94501,94643,94699,94759,94821,94833,95164,95253,95437,95457,95640,95674,95721,95775,95792,95825,95957,96315,96357,96472,96518,96606,96680,96692,96824,96856,97012,97097,97178,97263,97408,97413,97470,97589,97613,97755,97820,98000,98047,98225,98533,98570,98743,98767,98912,98987,99001,99235,99239,99426,99505,99809,99810,99880",
          "167,211,298,333,368,390,401,517,685,745,755,757,921,964,1118,1275,1315,1392,1436,1523,1627,1657,1721,1739,1762,1764,1906,1953,1991,2029,2096,2137,2276,2335,2381,2388,2435,2538,2544,2628,2665,2712,2743,2768,2807,3033,3171,3316,3369,3477,3494,3535,3541,3674,3690,3708,3731,3772,3815,3825,3840,3862,3921,3933,3994,4093,4126,4255,4454,4660,4807,4823,4832,4834,4909,4911,5078,5170,5289,5428,5572,5692,5981,5999,6074,6095,6113,6114,6197,6229,6264,6368,6479,6636,6685,6694,6728,6978,7015,7023,7213,7274,7378,7413,7478,7550,7600,7668,7672,7685,7688,7697,7726,7854,7897,7928,7986,8006,8097,8109,8126,8178,8234,8739,8783,8818,9021,9100,9233,9370,9534,9566,9595,9621,9656,9679,9719,9759,9765,9770,9792,9793,9848,9891,9919,9931,10140,10337,10391,10412,10423,10472,10473,10625,10646,10717,10719,10823,10895,10937,11069,11450,11539,11584,11591,11646,11667,11717,11749,11772,11867,11870,11952,12040,12130,12185,12324,12354,12442,12453,12467,12480,12503,12512,12601,12717,12798,12803,12804,12886,12914,13045,13308,13413,13433,13480,13489,13553,13586,13614,13718,13780,14031,14069,14079,14097,14114,14142,14201,14261,14371,14425,14434,14581,14597,14738,14764,15020,15101,15118,15164,15502,15615,15627,15682,15734,15799,15823,15826,15875,15983,16038,16300,16360,16379,16400,16412,16414,16442,16465,16478,16543,16546,16783,16816,16922,17007,17008,17047,17093,17107,17156,17167,17243,17360,17503,17551,17578,17604,17640,17642,17691,17697,17706,17741,17746,17760,17838,17882,17921,17997,18087,18118,18161,18169,18227,18354,18584,18708,18773,18831,18968,19054,19066,19188,19262,19317,19347,19376,19381,19626,19732,20127,20151,20188,20451,20467,20489,20520,20540,20590,20629,20760,20797,20903,21041,21067,21153,21320,21588,21667,21689,21803,21881,21915,21934,21974,22107,22239,22295,22314,22328,22376,22452,22483,22484,22500,22642,22705,22718,22824,22839,22844,22950,23001,23104,23112,23125,23177,23195,23256,23300,23320,23465,23669,23701,23743,23815,23983,24022,24289,24330,24332,24425,24468,24513,24523,24528,24611,24620,24640,24765,24766,24877,24898,25042,25160,25226,25255,25323,25345,25410,25423,25439,25559,25718,25865,25886,25960,25970,26028,26112,26155,26284,26318,26414,26505,26542,26614,26638,26646,26743,27159,27173,27202,27270,27421,27424,27429,27620,27716,27790,28021,28134,28295,28314,28485,28487,28654,28728,28768,28823,28829,28852,28858,28882,28899,28924,28954,29113,29152,29218,29229,29400,29448,29662,29670,29694,29726,29801,29909,30057,30127,30222,30227,30229,30275,30282,30300,30414,30444,30489,30549,30563,30574,30581,30731,30756,30820,30822,30884,30937,30951,30964,31007,31036,31080,31101,31115,31204,31220,31359,31370,31415,31601,31697,31698,31966,32014,32073,32185,32243,32318,32474,32670,32674,32791,32817,32837,32876,32892,32896,32970,32978,33012,33082,33125,33235,33424,33441,33451,33476,33521,33523,33531,33564,33580,33779,33918,33939,33946,34063,34341,34421,34575,34693,34697,34722,34803,35032,35065,35098,35321,35325,35387,35717,35829,35834,35873,35934,36097,36113,36141,36311,36356,36375,36393,36419,36427,36443,36454,36489,36814,36869,36892,36903,37063,37081,37097,37207,37240,37276,37363,37388,37395,37502,37513,37522,37565,37628,37653,37684,37725,37839,37847,37921,37947,38034,38196,38262,38318,38474,38511,38646,38659,38817,38832,38858,38861,38931,39047,39115,39148,39208,39460,39646,39648,39653,39708,39818,39856,39918,39949,40075,40192,40261,40324,40404,40413,40445,40490,40496,40902,40908,41032,41156,41232,41308,41319,41413,41467,41535,41536,41570,41683,41685,41686,41898,41992,42040,42250,42443,42486,42635,42648,42883,43041,43061,43132,43152,43209,43223,43243,43355,43534,43591,43637,43655,43755,43852,43874,43899,43922,44018,44072,44114,44141,44165,44271,44304,44319,44498,44515,44608,44612,44684,44781,45006,45029,45184,45195,45290,45338,45393,45416,45602,45653,45805,45984,46030,46189,46193,46220,46231,46334,46335,46617,46719,46789,46827,46843,46847,46932,46966,47012,47030,47074,47238,47306,47322,47365,47418,47430,47592,47649,47705,47832,47934,47937,47981,47998,48002,48009,48064,48072,48186,48298,48401,48427,48565,48575,48697,48861,49080,49097,49101,49240,49296,49474,49658,49759,49797,49895,49983,50033,50081,50132,50238,50287,50329,50387,50460,50471,50562,50572,50652,50752,50842,50911,50982,51021,51033,51110,51116,51154,51323,51385,51496,51598,51686,51817,51948,51975,52049,52141,52184,52205,52260,52276,52330,52488,52532,52559,52595,52672,52947,53101,53187,53200,53354,53361,53392,53441,53482,53623,53726,53791,53822,53847,53862,53876,53970,54018,54048,54134,54135,54170,54184,54261,54305,54315,54403,54495,54588,54853,54941,54942,55035,55042,55076,55087,55159,55195,55291,55371,55541,55598,55630,55786,55848,55997,56012,56244,56293,56340,56351,56424,56488,56505,56630,56659,56689,56867,56925,56965,57054,57142,57365,57481,57552,57768,57794,58205,58286,58626,58641,58713,58714,58740,58819,58824,58845,58878,58914,59035,59076,59112,59296,59399,59407,59514,59521,59879,59936,60030,60071,60121,60224,60246,60303,60307,60583,60879,61048,61288,61316,61475,61574,61658,61696,61719,61768,61811,61873,61913,62008,62048,62203,62248,62305,62371,62411,62459,62602,62624,62758,62895,62896,63001,63231,63245,63301,63429,63453,63536,63806,63936,63975,63990,64029,64158,64183,64216,64283,64528,64538,64635,64647,64717,64725,64778,64896,64916,65033,65096,65149,65150,65162,65232,65265,65276,65427,65439,65445,65458,65493,65560,65639,65651,65689,65812,65977,66068,66259,66265,66431,66476,66547,66594,66682,66725,66778,66790,66871,66872,66918,67021,67152,67252,67259,67269,67438,67737,67764,67773,67979,67999,68150,68257,68271,68277,68408,68455,68755,68907,68981,69007,69064,69322,69338,69691,69711,69713,69846,70050,70069,70110,70268,70349,70403,70436,70533,70534,70562,70594,70713,70783,71002,71082,71089,71141,71276,71372,71389,71398,71453,71758,71861,71867,72092,72233,72237,72399,72459,72475,72560,72599,72692,72749,72820,72875,73021,73245,73246,73279,73304,73345,73351,73461,73559,73577,73582,73607,73730,73740,73762,73813,73823,73825,73979,74022,74193,74254,74285,74415,74477,74546,74612,74663,74733,74790,74872,74931,74956,74963,74993,75103,75177,75194,75322,75347,75435,75507,75587,75591,75637,75667,75686,75752,75764,75772,75784,75824,75905,76071,76097,76223,76310,76363,76401,76548,76581,76589,76728,76770,76854,76859,76937,76940,76981,77179,77280,77316,77319,77349,77359,77483,77596,77672,77717,77758,77797,78015,78160,78215,78234,78277,78278,78295,78360,78362,78468,78552,78664,78776,78780,78799,78831,78857,79157,79227,79481,79486,79521,79591,79755,79781,79917,80011,80039,80122,80244,80557,80566,80644,80704,80705,80706,80736,80776,80832,80849,80859,80909,80953,81046,81084,81092,81098,81213,81341,81398,81412,81602,81708,81740,81781,81848,81999,82033,82161,82182,82277,82317,82326,82332,82351,82372,82400,82467,82526,82723,82833,82840,82842,82884,82902,82926,83033,83376,83443,83495,83546,83632,83715,83910,83928,83950,84191,84238,84332,84367,84381,84411,84466,84475,84512,84686,84749,84754,84934,84962,85022,85107,85171,85180,85182,85271,85276,85308,85348,85428,85476,85512,85560,85569,85631,85665,85667,85766,85796,85939,86069,86161,86179,86224,86265,86297,86320,86496,86556,86577,86588,86597,86693,86719,86846,86934,86936,86996,87042,87045,87102,87143,87236,87407,87504,87516,87552,87591,87606,87759,87843,87907,87944,88048,88106,88141,88153,88221,88235,88290,88323,88442,88542,88729,88909,88963,89018,89212,89269,89545,89596,89641,89781,90032,90119,90167,90210,90280,90349,90415,90626,90695,90851,91000,91126,91173,91174,91215,91242,91333,91388,91626,91764,91885,92017,92088,92112,92133,92224,92353,92377,92483,92518,92561,92607,92619,92661,92668,92920,92931,92992,93130,93241,93270,93281,93294,93313,93342,93346,93397,93421,93503,93588,93668,93720,93790,93909,94098,94139,94147,94284,94339,94391,94557,94573,94589,94616,94760,94843,94898,94977,95224,95248,95313,95336,95563,95747,95809,95851,95912,95945,95948,96118,96403,96434,96945,97515,97537,97562,97644,97802,97833,97964,98027,98088,98111,98127,98130,98212,98358,98524,98579,98677,98759,98821,98840,99049,99320,99339,99390,99423,99457,99674,99693,99699,99709,99844,99891",
          "85,118,229,328,377,429,575,666,712,782,827,868,1029,1283,1393,1408,1567,1643,1740,1829,1854,1883,2020,2023,2280,2330,2418,2426,2534,3010,3042,3078,3166,3408,3418,3552,3621,3732,3782,3865,3913,4017,4044,4056,4068,4088,4118,4160,4181,4256,4303,4323,4428,4450,4474,4590,4611,4645,4696,4835,4881,4917,4931,5039,5087,5218,5233,5240,5242,5274,5453,5500,5585,5979,5982,6056,6097,6100,6247,6342,6372,6548,6571,6607,6886,6908,6970,7043,7061,7116,7199,7221,7234,7267,7422,7532,7578,7660,8026,8040,8079,8115,8122,8154,8211,8306,8435,8660,8741,8782,8788,8826,8902,9002,9112,9146,9329,9395,9465,9577,9579,9588,9776,9836,9892,10082,10117,10217,10243,10314,10666,10715,10858,11082,11128,11167,11307,11493,11582,11662,11805,11855,11922,11973,12012,12155,12295,12462,12599,12718,12761,12930,13099,13105,13130,13170,13237,13287,13317,13334,13393,13411,13436,13523,13555,13618,13619,13652,13717,13928,13940,13945,14158,14236,14373,14389,14407,14541,14558,14601,14604,14623,14763,14834,14920,14985,15029,15124,15268,15354,15359,15551,15827,15908,15999,16013,16041,16057,16109,16188,16235,16363,16523,16561,16663,16710,16796,16808,16810,16829,16961,17029,17030,17174,17301,17334,17672,17744,17828,17840,17868,17893,17960,18000,18136,18241,18245,18282,18295,18310,18364,18535,18550,18600,18639,18719,19025,19027,19094,19163,19238,19246,19330,19369,19375,19577,19586,19674,19888,20017,20139,20258,20274,20277,20336,20370,20485,20501,20538,20545,20640,20807,20891,21120,21417,21568,21606,21784,21845,22043,22071,22082,22236,22403,22598,22714,22807,22828,23167,23183,23244,23277,23334,23452,23678,23694,23767,23770,23847,23877,23925,24072,24304,24320,24401,24516,24575,24612,24639,24642,24739,24757,24870,24942,25242,25290,25320,25342,25353,25384,25665,25742,25809,25853,25950,26086,26090,26234,26243,26260,26391,26509,26637,26687,26720,26820,26896,27091,27106,27112,27263,27350,27480,27507,27545,27782,27800,27804,27925,28008,28082,28089,28193,28259,28288,28530,28560,28581,28650,28784,28843,28854,28955,29044,29123,29240,29351,29618,29821,29990,30021,30052,30078,30147,30194,30251,30269,30687,30826,30894,30939,31172,31181,31214,31277,31321,31423,31736,31763,31831,31916,31945,31973,31979,32020,32040,32081,32164,32290,32307,32324,32454,32487,32576,32796,32923,33094,33149,33216,33292,33443,33687,33820,33940,34014,34052,34082,34170,34196,34464,34497,34503,34587,34668,34699,34707,34797,34805,34807,34859,34875,34917,35005,35042,35115,35149,35226,35511,35541,35588,35599,35659,35765,35766,35789,36001,36131,36221,36250,36253,36341,36474,36494,36544,36570,36643,36706,36768,36825,36833,36977,37044,37071,37082,37083,37105,37106,37292,37321,37353,37397,37433,37473,37498,37674,37685,37804,37958,38022,38106,38149,38554,38628,38755,38977,39122,39161,39179,39274,39497,39510,39514,39542,39569,39637,39651,39685,39701,39733,39808,39890,40026,40035,40047,40147,40173,40243,40329,40375,40417,40432,40495,40547,40611,40703,40704,40822,40841,40850,40894,40942,40984,41056,41104,41178,41239,41261,41279,41457,41501,41568,41606,41634,41673,41674,41706,41763,41765,41790,41792,41931,41951,42002,42015,42041,42164,42199,42222,42224,42271,42293,42296,42624,42711,42738,42831,42875,43027,43048,43075,43100,43148,43187,43376,43447,43722,43764,43817,43844,43887,43977,44163,44220,44289,44492,44533,44543,44631,44739,44839,44872,45109,45209,45253,45447,45450,45456,45472,45549,45555,45737,45771,45791,46000,46037,46068,46207,46209,46406,46455,46470,46503,46519,46586,46662,46681,46694,46761,46810,46976,46995,47152,47185,47216,47220,47240,47258,47294,47329,47384,47436,47514,47537,47587,47708,47715,47726,47809,47982,48101,48141,48158,48371,48555,48633,48655,48705,48713,48957,49002,49081,49141,49198,49214,49248,49257,49577,49639,49655,49675,50584,50613,50623,50669,50732,50800,50913,50934,50968,51042,51345,51379,51478,51544,51582,51777,51992,52004,52006,52029,52090,52274,52428,52940,52999,53154,53416,53431,53470,53547,53620,53659,53662,53826,53915,54024,54213,54235,54318,54372,54388,54466,54585,54659,54726,54743,54758,54806,54925,55057,55160,55185,55327,55335,55593,55695,55944,55974,56018,56021,56305,56375,56496,56507,56528,56611,56651,56750,56826,56844,56862,56893,56894,56928,56934,57041,57173,57248,57250,57468,57557,57599,57706,57721,57739,57818,57838,57906,57966,57969,58006,58133,58158,58245,58307,58384,58426,58442,58683,58718,58785,58792,58938,59049,59124,59145,59171,59183,59299,59339,59385,59463,59472,59486,59580,59585,59587,59630,59632,59721,59831,59882,59885,59914,59929,59960,60201,60205,60276,60336,60415,60420,60487,60498,60536,60899,61080,61152,61223,61253,61444,61563,61585,61593,61644,61779,62036,62101,62129,62159,62171,62321,62375,62680,62717,62726,62745,62762,62878,62947,63017,63079,63197,63335,63426,63430,63463,63537,63595,63706,63747,63751,63838,63875,63939,63962,63973,63984,64075,64125,64195,64262,64384,64504,64734,64811,65031,65093,65115,65127,65306,65309,65342,65472,65580,65691,65728,65881,65916,66010,66147,66247,66322,66393,66428,66446,66549,66695,66748,66764,67045,67083,67115,67125,67192,67361,67451,67523,67613,67710,68109,68121,68450,68475,68580,68587,68747,68800,68879,69327,69337,69605,69709,69733,69777,69811,69891,70009,70068,70116,70141,70205,70288,70320,70332,70399,70485,70561,70768,70851,70942,70972,71175,71219,71295,71337,71386,71423,71521,71682,71763,71974,72095,72168,72243,72531,72590,72596,72627,72668,72740,72793,72808,72822,73029,73053,73109,73120,73186,73261,73317,73355,73419,73497,73513,73539,73612,73707,73850,74354,74362,74404,74550,74615,74656,74703,74878,75036,75043,75201,75292,75422,75492,75550,75594,75786,75918,75932,75981,75996,76001,76026,76059,76111,76125,76129,76136,76642,76658,76691,76774,76786,76790,76878,76917,77039,77107,77162,77230,77252,77442,77579,77633,77663,77747,77760,77794,77904,77964,78060,78084,78153,78249,78348,78490,78645,78668,78742,78750,78758,78909,79115,79139,79154,79203,79288,79351,79373,79400,79561,79565,79679,79700,79729,79860,80019,80087,80291,80389,80398,80489,80515,80530,80624,80755,80777,80865,80978,81137,81166,81257,81284,81313,81316,81348,81566,81588,81611,81662,81970,82050,82124,82175,82199,82212,82229,82310,82461,82541,82826,82901,82912,82918,82949,82990,83018,83113,83122,83226,83228,83245,83552,83689,83812,83989,84117,84351,84433,84494,84580,84583,84642,84844,84846,84927,84973,85043,85100,85460,85520,85529,85637,85806,85851,85880,85966,86133,86200,86262,86271,86401,86583,86649,86681,86734,86889,87125,87129,87218,87289,87348,87372,87388,87468,87524,87754,87884,87936,88074,88140,88149,88252,88295,88368,88489,88625,88941,88979,89014,89030,89060,89117,89222,89234,89340,89360,89384,89387,89449,89598,89703,89735,89908,89942,90014,90181,90484,90611,90714,91223,91359,91438,91457,91695,91704,91710,91720,91760,91787,91819,91911,91935,91966,92092,92312,92335,92405,92438,92612,92651,92782,92971,92990,93043,93124,93307,93607,93626,93788,93806,93838,94053,94082,94265,94280,94368,94628,94727,94733,94791,94915,94931,95217,95395,95614,95633,95650,95686,95798,96183,96205,96206,96279,96379,96642,96757,96840,96942,97237,97279,97284,97441,97480,97531,97551,97572,97576,97777,97985,98110,98120,98133,98201,98320,98338,98362,98455,98538,98547,98801,98897,99062,99065,99143,99178,99212,99296,99475,99571,99575,99588,99710,99766,99837,99983",
          "107,119,130,629,665,769,874,890,1136,1201,1257,1421,1580,1604,1785,1800,1832,1859,1934,2003,2098,2153,2200,2209,2242,2243,2256,2365,2419,2591,2690,2785,2842,2854,2894,2914,2990,3045,3204,3287,3463,3464,3759,3906,3934,3980,4140,4216,4488,4575,4587,4694,4726,4821,4873,4941,5159,5343,5445,5599,5613,5629,5844,5972,6079,6139,6280,6295,6473,6500,6580,6923,7039,7216,7250,7279,7427,7430,7494,7556,7663,7699,7702,7762,7871,7923,7967,7972,7992,7994,7995,8136,8152,8187,8252,8294,8411,8422,8441,8487,8530,8704,8867,8956,9069,9137,9192,9283,9353,9394,9408,9471,9592,9666,9750,9864,9900,9943,9950,9987,10099,10126,10131,10170,10207,10234,10329,10379,10566,10678,10747,10830,10861,10876,11046,11055,11074,11205,11500,11557,11700,11702,11743,11830,11856,11882,11892,11925,11951,12107,12335,12385,12546,12562,12585,12612,12646,12865,13015,13049,13110,13199,13250,13256,13280,13421,13434,13582,13590,13705,13722,13793,13826,13854,13903,14023,14144,14291,14342,14351,14369,14500,14516,14592,14633,14654,14671,14732,14864,14895,14939,14969,15048,15128,15222,15249,15281,15322,15366,15393,15443,15511,15512,15537,15593,15684,15750,15762,15795,16042,16064,16073,16238,16988,17032,17172,17180,17215,17324,17338,17793,17805,17822,18109,18307,18333,18362,18399,18407,18548,18590,18630,18690,18837,18927,19045,19052,19053,19090,19120,19329,19461,19582,19877,19969,20096,20098,20272,20290,20304,20391,20480,20499,20543,20595,20692,20885,20897,20956,21049,21073,21225,21237,21425,21729,21920,22047,22117,22242,22311,22486,22538,22693,22701,22725,22826,22847,22925,22953,23023,23025,23165,23267,23291,23317,23371,23623,23703,23717,23724,23824,23841,23903,23988,24015,24120,24470,24545,24562,24580,24590,24756,25064,25141,25268,25309,25417,25424,25427,25470,25518,25533,25810,25842,25857,26034,26048,26051,26223,26329,26356,26379,26390,26497,26507,26536,26756,27003,27090,27145,27197,27290,27291,27327,27505,27558,27604,27737,27824,27858,27918,27921,27933,27935,27952,28052,28155,28226,28242,28244,28393,28458,28499,28510,28542,28694,28731,28732,28804,28851,29045,29150,29324,29387,29472,29528,29639,29715,29740,29880,29907,29934,29967,30031,30109,30216,30257,30331,30572,30635,30684,30712,30741,30803,30823,30893,31289,31596,31771,31803,31813,31882,31951,31955,32026,32149,32206,32345,32389,32401,32485,32526,32626,32646,32710,32752,32775,32854,32870,33005,33040,33204,33243,33274,33278,33364,33374,33401,33411,33422,33623,33747,33818,33996,34015,34070,34164,34277,34455,34520,34602,34624,34635,34640,34661,34907,34936,35088,35146,35148,35153,35157,35251,35320,35415,35427,35524,35538,35656,35711,35838,35856,35936,35976,36193,36285,36335,36355,36402,36650,36745,36811,36844,36852,36900,37086,37157,37186,37311,37314,37334,37541,37584,37589,37606,38043,38045,38053,38065,38087,38162,38242,38760,38883,39049,39063,39064,39112,39244,39310,39579,39734,39775,39848,40231,40377,40437,40457,40608,40853,40886,41082,41233,41287,41352,41561,41565,41665,41679,41791,41804,41863,41941,42152,42158,42208,42225,42281,42319,42580,42610,42679,42734,43043,43091,43236,43261,43353,43392,43483,43593,43645,43674,43679,43985,44055,44232,44335,44423,44478,44657,44729,44778,44838,44867,44950,45115,45204,45221,45245,45365,45483,45601,45611,45958,46093,46218,46283,46414,46515,46549,46555,46691,46798,46817,46835,46936,46970,47069,47090,47149,47259,47293,47471,47476,47478,47523,47656,47664,47694,47744,47782,47805,47980,48049,48136,48234,48294,48400,48522,48569,48574,48585,48586,48624,48673,48810,48811,48874,48910,49026,49036,49122,49251,49256,49290,49317,49323,49471,49488,49513,49599,49985,49995,50089,50248,50730,50782,50986,50996,51595,51711,51927,52012,52151,52158,52183,52411,52422,52505,52507,52562,52726,52884,52908,53045,53157,53234,53402,53415,53455,53542,53605,53688,53827,53863,53931,54015,54226,54278,54475,54553,54574,54618,54736,54742,54771,54914,54945,54954,55072,55145,55164,55307,55372,55395,55540,55626,55756,55849,55852,55910,56453,56604,56634,56837,56852,56975,57171,57234,57287,57310,57349,57352,57414,57556,57600,57637,57644,57757,57780,57928,57986,58204,58247,58256,58382,58437,58463,58573,58583,58645,58659,59079,59203,59214,59281,59423,59435,59518,59597,59663,59679,59744,59869,59927,60357,60448,60525,60534,60683,60784,60871,60956,61183,61219,61280,61290,61560,61605,61626,61631,61647,61985,62037,62332,62382,62412,62421,62423,62462,62790,62920,63019,63101,63153,63240,63292,63357,63413,63801,64006,64020,64177,64190,64194,64211,64248,64319,64366,64469,64584,64600,64636,64650,64701,64784,64806,64849,64945,65190,65213,65335,65523,65641,65715,65781,65824,65939,65960,66018,66419,66475,66789,66863,66909,67670,67715,67735,67777,67794,67802,68073,68134,68173,68183,68344,68420,68432,68531,68616,68794,68943,68951,69059,69203,69243,69261,69423,69746,69816,70027,70129,70131,70135,70228,70343,70409,70422,70432,70446,70476,70531,70683,70686,70991,71099,71110,71506,71519,71565,71582,71673,71852,71964,72097,72114,72141,72155,72181,72534,72593,72637,72811,72844,73002,73071,73292,73396,73704,73864,73871,73903,74023,74205,74236,74605,74621,74630,74678,74746,74799,74957,75160,75244,75269,75315,75332,75472,75485,75538,75571,75645,75702,75777,75827,75911,75929,75936,75949,76029,76033,76250,76427,76433,76470,76560,76564,76598,76746,76894,76895,76947,76975,77023,77041,77055,77108,77119,77215,77296,77307,77343,77414,77489,77495,77818,77872,77940,77993,78099,78227,78258,78336,78381,78407,78492,78505,78530,78596,78665,78706,78724,78918,78920,79035,79075,79104,79145,79251,79287,79307,79519,79587,79641,79774,80016,80104,80123,80237,80425,80525,80578,80629,80957,81068,81079,81168,81180,81272,81335,81363,81403,81525,81581,81584,81656,81698,81707,81742,82261,82320,82397,82457,82460,82654,82666,82688,82712,82734,82780,82804,82821,82996,83120,83217,83417,83459,83479,83507,83523,83528,83650,83665,83767,83829,83906,83958,84131,84204,84376,84380,84386,84503,84523,84527,84549,84789,84799,84894,84958,85082,85096,85201,85591,85625,85656,85754,86030,86213,86277,86463,86654,86684,86858,87089,87161,87308,87316,87442,87514,87824,87924,87930,88072,88129,88138,88322,88380,88408,88425,88431,88433,88648,88659,88724,88784,88817,88849,88898,89111,89170,89177,89271,89272,89314,89476,89534,89799,89860,89877,89938,89952,89965,90177,90268,90298,90299,90304,90398,90613,90637,90710,90711,90745,90761,90782,90875,90934,90973,91159,91234,91324,91362,91792,91837,91958,92051,92077,92158,92206,92259,92280,92373,92479,92579,92623,92731,92746,92851,92987,93120,93187,93229,93256,93350,93574,93625,93644,93649,93731,93756,93776,93873,93943,93965,93997,94074,94195,94233,94277,94484,94623,94692,94748,94751,94869,94902,94936,94951,94965,94966,95018,95198,95443,95663,95694,95903,95969,96040,96111,96365,96440,96565,96823,96863,96865,96869,96875,96906,96932,96955,96973,97051,97120,97150,97190,97218,97270,97495,97574,97707,97749,97854,97889,98022,98175,98229,98303,98304,98382,98387,98461,98566,98587,98653,98841,98873,98957,99104,99300,99355,99413,99587,99846,99962,99984,99990",
          "75,109,137,382,383,443,492,540,715,736,740,1008,1185,1190,1228,1297,1468,1470,1496,1510,1548,1757,1788,1824,1860,1879,1943,2130,2168,2176,2180,2215,2225,2378,2470,2537,2713,2717,2752,2756,2763,2933,2986,3011,3064,3096,3098,3132,3208,3271,3343,3434,3546,3588,3599,3858,3981,3982,4079,4083,4224,4298,4538,4780,4894,5134,5152,5217,5280,5456,5474,5559,5594,5730,5778,5835,5890,5894,5895,5906,5961,5966,6186,6301,6350,6427,6470,6490,6502,6520,6582,6637,6712,6919,6980,7007,7033,7081,7292,7602,7661,7734,7742,7784,7808,8031,8105,8163,8382,8518,8564,8740,8816,9000,9054,9092,9122,9140,9354,9386,9404,9405,9548,9575,9672,9706,9802,9893,9910,9927,9932,10074,10139,10241,10246,10255,10296,10318,10386,10618,10633,10637,10784,10912,11143,11199,11214,11296,11385,11416,11423,11447,11616,11690,11790,11864,12044,12053,12186,12205,12472,12474,12687,12787,12863,12876,13053,13058,13066,13070,13135,13190,13255,13270,13327,13372,13456,13466,13518,13541,13578,13587,13605,13690,13710,13736,13744,13811,13813,13852,14042,14122,14171,14289,14344,14349,14354,14387,14543,14631,14670,14690,14734,14739,14760,14862,14866,14891,15024,15189,15203,15204,15224,15231,15234,15373,15408,15462,15470,15532,15686,15832,15874,15901,15906,15914,16180,16195,16213,16253,16318,16351,16356,16377,16600,16753,16764,16778,16997,17122,17259,17417,17475,17535,17559,17599,17721,17804,17826,17858,17974,17994,18014,18100,18193,18209,18572,18792,18807,18874,18923,19078,19110,19141,19191,19211,19220,19278,19289,19305,19380,19453,19475,19648,19704,19714,19722,19801,19832,19965,20044,20063,20088,20113,20125,20238,20252,20294,20322,20325,20357,20512,20621,20684,20740,20746,20959,20964,20995,21089,21188,21198,21273,21383,21390,21591,21649,21821,21855,21860,21943,21984,22035,22086,22158,22196,22331,22422,22488,22592,22676,22854,22872,22906,22911,22955,23002,23078,23119,23247,23264,23554,23656,23671,23827,23889,24013,24029,24047,24056,24151,24176,24317,24346,24446,24462,24487,24662,24721,24760,24820,24863,24943,24948,25025,25038,25100,25213,25217,25398,25492,25501,25570,25601,25633,25648,25724,25760,25762,25789,25823,25852,25869,25983,26019,26025,26067,26069,26113,26179,26365,26408,26467,26503,26629,26648,26853,26954,27024,27140,27205,27253,27349,27397,27688,27714,27773,27832,27944,28105,28110,28178,28184,28235,28292,28365,28436,28627,28682,28688,28778,28803,28847,28874,29114,29140,29142,29285,29501,29534,29592,29746,29865,29958,30004,30032,30048,30133,30157,30249,30256,30263,30289,30365,30398,30485,30711,30718,30746,30753,30776,30842,30845,31116,31128,31137,31175,31196,31236,31259,31308,31372,31387,31405,31473,31542,31599,31624,31675,31755,32089,32263,32299,32382,32531,32577,32641,32915,32967,32994,33026,33305,33368,33440,33483,33535,33562,33902,33908,33979,34027,34042,34051,34075,34115,34168,34216,34232,34253,34293,34381,34549,34564,34896,34902,34929,34931,35068,35118,35225,35420,35550,35636,35654,35776,35866,35937,35949,36022,36046,36117,36263,36477,36555,36562,36567,36605,36671,36818,37023,37102,37243,37257,37424,37442,37524,37569,37579,37726,37769,37788,37799,37873,38033,38038,38232,38291,38512,38576,38591,38634,38734,38765,38806,38831,38857,38877,39206,39228,39247,39329,39381,39419,39496,39531,39588,39663,39665,39672,39720,39860,39896,39953,40006,40115,40189,40250,40312,40323,40393,40434,40450,40534,40548,40579,40594,40733,40785,41014,41110,41154,41305,41541,41637,41865,41981,42011,42028,42053,42155,42278,42351,42374,42413,42468,42546,42743,42817,42866,42912,42919,43164,43340,43575,43581,43725,43726,43740,43742,43800,43906,43916,44044,44066,44186,44201,44284,44313,44351,44356,44434,44526,44537,44850,44929,44963,44968,45079,45167,45241,45259,45296,45328,45470,45499,45542,45680,45681,45911,45987,46040,46050,46122,46153,46178,46194,46225,46226,46255,46382,46399,46418,46423,46462,46594,46696,46766,46768,46777,46841,46884,46940,47005,47050,47146,47163,47230,47363,47491,47667,47669,47714,47803,47819,47910,47965,48022,48045,48120,48167,48252,48334,48366,48384,48420,48441,48442,48465,48483,48572,48604,48670,48778,48785,48791,48837,48932,49173,49293,49400,49523,49540,49545,49562,49672,49688,49891,49958,50180,50234,50369,50384,50498,50545,50969,50974,51014,51103,51156,51158,51243,51257,51274,51304,51356,51547,51556,51569,51829,51845,51870,52037,52071,52129,52145,52237,52292,52325,52356,52460,52555,52693,52759,52777,52952,53012,53132,53144,53556,53680,53699,53818,53934,53983,54119,54349,54354,54385,54417,54476,54615,54957,55151,55225,55271,55444,55781,55883,55939,55994,56034,56346,56373,56382,56467,56483,56489,56648,56652,56662,56678,56748,56933,56960,56973,57089,57102,57273,57385,57415,57534,57579,57581,57788,57848,57885,57977,58011,58020,58070,58188,58234,58268,58371,58405,58432,58483,58516,58621,58654,58769,58827,58846,58946,59005,59025,59174,59178,59250,59260,59362,59460,59462,59475,59734,59762,59817,59852,59891,59963,60009,60104,60288,60622,60670,60715,61004,61020,61066,61100,61113,61372,61375,61549,61606,61620,61731,61787,61806,61894,61978,62038,62095,62199,62240,62301,62779,62890,62892,62929,63052,63129,63134,63220,63283,63383,63534,63545,63782,63823,63978,64072,64080,64414,64415,64518,64713,64721,65034,65085,65095,65098,65399,65462,65531,65557,65562,65583,65628,65649,65777,65800,65811,65831,65956,65966,66140,66517,66645,66651,66668,66797,66806,66929,66990,67127,67165,67179,67276,67327,67476,67500,67504,67618,67631,67768,67977,67998,68175,68308,68366,68382,68444,68453,68518,68552,68732,68754,68769,68992,69031,69183,69189,69558,69579,69695,69714,69736,69754,69782,69794,69838,70003,70174,70307,70331,70387,70391,70427,70442,70515,70625,70672,70685,70697,70750,70777,70788,70835,70867,71021,71267,71291,71400,71501,71507,71650,71821,71946,72005,72096,72228,72478,72499,72566,72586,72592,72613,72641,72684,72843,72911,72972,72995,73007,73079,73143,73155,73269,73473,73499,73542,73610,73679,73702,73712,73732,73824,74024,74253,74347,74348,74417,74436,74457,74548,74654,74732,74871,74902,74950,75010,75137,75184,75486,75515,75546,75551,76145,76169,76178,76188,76214,76344,76537,76654,76709,76729,76730,76918,76968,76982,77052,77187,77190,77198,77279,77485,77752,77824,77889,77895,77896,77919,77944,77961,77991,78070,78205,78219,78238,78243,78248,78256,78310,78433,78559,78569,78619,78650,78735,78875,79128,79136,79210,79212,79273,79295,79326,79528,79541,79786,79811,79816,79868,79986,80074,80100,80189,80385,80441,80448,80596,80702,80772,80773,80792,80802,80828,80869,80914,80929,81064,81142,81302,81409,81486,81494,81510,81532,81555,81608,81710,82009,82081,82136,82162,82281,82335,82379,82500,82569,82634,82879,82909,82978,82984,82986,82993,83155,83198,83290,83353,83468,83494,83499,83542,83551,83574,83577,83808,83830,83860,83981,83986,84007,84041,84065,84102,84133,84160,84310,84765,84785,84864,85046,85076,85094,85322,85371,85390,85480,85493,85494,85563,85727,85747,85761,85803,85997,86132,86135,86149,86216,86234,86299,86341,86445,86504,86596,86607,86618,86851,87078,87096,87135,87156,87192,87204,87342,87417,87500,87620,87640,87712,87716,87734,87919,87942,87965,88218,88281,88568,88577,88673,88730,88914,88971,88977,89002,89035,89087,89198,89470,89591,89771,89839,89848,89865,89912,89991,90051,90072,90074,90157,90240,90244,90289,90348,90385,90770,90786,91077,91254,91379,91462,91604,91750,91866,91897,91923,92191,92275,92313,92391,92394,92444,92558,92567,92606,92610,92657,92714,92729,92770,92775,92998,93065,93128,93153,93210,93223,93247,93355,93374,93410,93418,93578,93584,93751,93812,93865,93883,94358,94376,94442,94547,94607,94682,94817,94856,94950,94975,94980,95015,95052,95186,95299,95383,95387,95492,95572,95605,95629,95653,95703,96015,96142,96229,96261,96263,96280,96303,96358,96437,96503,96533,96578,96628,96718,96850,96868,97140,97251,97277,97310,97381,97493,97653,97800,98021,98190,98210,98326,98346,98407,98437,98494,98542,98549,98550,98611,98672,98679,98756,98773,98964,99038,99040,99126,99137,99182,99188,99196,99201,99226,99308,99515,99560,99603,99629,99778,99963",
          "8,278,289,293,311,355,363,367,399,418,483,770,777,805,938,1114,1244,1259,1326,1376,5012,5053,5068,5100,5103,5162,5265,5277,5353,5389,5557,5667,5732,5756,5761,5789,5799,5846,5873,5900,5912,5919,5923,5935,5964,6051,6052,6085,6086,6298,6343,6388,6461,6475,6544,6711,6773,6862,6863,6932,6949,6961,7047,7057,7136,7157,7186,7319,7365,7680,8269,8465,9177,9230,9584,9682,10030,10044,10413,10838,10930,11210,11400,11545,12242,13163,13703,14315,15219,16114,16740,17964,18520,18605,18606,18763,20381,20970,20989,21207,22440,24351,24505,24511,25027,25307,27256,28628,28800,31012,31135,31805,32056,35328,37679,45665,50043,50145,50303,50463,50496,50570,50670,50862,50868,50891,50950,51121,51221,51238,51287,51381,51446,51465,51832,51853,52021,52051,52083,52178,52455,52509,52518,52561,52806,52850,52910,52943,52982,52987,53015,53199,53216,53256,53588,53595,53733,53748,53947,54005,54027,54043,54093,54229,54242,54247,54263,54319,54422,54512,54527,54607,54629,54639,54754,54821,54836,54866,54958,55064,55139,55169,55182,55230,55292,55775,55816,55930,55970,55971,55972,56028,56102,56107,56129,56288,56450,56598,56672,56797,56841,56864,56895,57073,57247,57343,57536,57717,57869,57996,58113,58146,58210,58226,58386,58440,58523,58774,58775,58790,58798,59221,59225,59326,59432,59458,59465,59653,59656,59714,60021,60136,60382,60453,60549,60578,60628,60842,60845,60962,61084,61121,61191,61194,61472,61497,61856,62071,62223,62279,62294,62377,62408,62492,62559,62650,62764,62796,62865,62893,62968,63138,63278,63295,63298,63504,63508,63802,63818,63870,63887,63922,64010,64088,64277,64355,64441,64602,64607,64726,64877,64907,64985,65052,65084,65094,65168,65284,65336,65592,65602,65729,65787,65794,66091,66105,66281,66386,66505,66541,66618,66663,66804,66955,67006,67020,67051,67169,67254,67311,67445,67454,67457,67496,67643,68002,68053,68157,68200,68335,68351,68364,68457,68608,68700,68749,68797,68861,68869,68906,69061,69249,69304,69356,69397,69555,69591,69893,69954,70083,70109,70159,70326,70395,70421,70716,70718,70764,70901,70910,70967,70970,71007,71206,71222,71248,71336,71468,71471,71605,71624,71658,71663,71739,71743,71790,71807,71810,71902,71903,71906,72018,72077,72090,72139,72142,72153,72207,72282,72367,72385,72473,72497,72522,72575,72655,72723,72757,72767,72857,72992,73000,73157,73212,73309,73397,73515,73734,73810,73939,73992,74025,74247,74252,74276,74371,74386,74410,74538,74569,74670,74718,74754,74848,74906,74944,74976,74979,75123,75140,75338,75418,75455,75563,75794,76018,76110,76316,76321,76335,76352,76353,76395,76475,76489,76814,76953,76992,76993,76994,77018,77123,77126,77177,77344,77360,77370,77490,77613,77643,77846,77873,77887,78257,78683,78704,78727,78740,78852,79020,79165,79171,79233,79509,79886,79934,80003,80012,80021,80144,80325,80371,80374,80380,80396,80412,80748,80753,80759,80894,81115,81133,81196,81204,81256,81378,81607,81873,81963,82062,82108,82219,82424,82472,82484,82616,82704,82758,82828,82988,83003,83254,83450,83612,83623,83646,83730,83768,84014,84064,84071,84104,84126,84312,84396,84521,84522,84635,84711,84761,84811,84833,84847,85026,85214,85217,85417,85510,85525,85561,85670,85770,85795,85834,86066,86067,86073,86089,86202,86231,86248,86249,86381,86395,86421,86456,86571,86744,86778,86841,86863,86868,87331,87426,87474,87693,87733,87760,87784,87790,87834,87840,87880,88251,88303,88548,88580,88854,88863,88937,88950,89191,89250,89324,89398,89400,89447,89468,89502,89575,89659,89674,89681,89782,89798,89813,90008,90015,90058,90103,90211,90239,90279,90331,90516,90585,90593,90766,90771,90776,90829,90911,90959,91004,91017,91062,91139,91253,91268,91274,91376,91403,91459,91498,91519,91612,91678,91810,91852,92009,92026,92160,92161,92200,92231,92237,92363,92400,92433,92441,92513,92524,92816,93087,93192,93335,93343,93700,93801,93977,94026,94093,94100,94127,94156,94261,94364,94567,94620,94645,94680,94730,94746,94871,94948,95030,95032,95192,95320,95340,95533,95749,95941,95965,96087,96158,96220,96296,96328,96600,96629,96699,96711,96767,96944,96991,97084,97134,97170,97220,97348,97466,97498,97534,97624,97713,97732,97735,97805,97926,97949,97950,98018,98305,98321,98644,98657,98668,98686,98962,98970,98972,99276,99335,99453,99562,99652,99832,99893,99911,99941,99950,99959",
          "74,123,129,160,327,346,352,463,507,541,780,841,887,986,1082,1182,1184,1451,1588,1610,1644,1789,1803,1822,1858,1900,2076,2083,2085,2315,2360,2463,2479,2574,2611,2629,2708,2775,2912,2953,2984,3006,3012,3028,3092,3247,3513,3530,3545,3575,3576,3669,3687,3786,3792,3826,3879,3885,3940,4024,4229,4319,4341,4353,4395,4398,4438,4476,4487,4839,4901,4930,5014,5027,5200,5526,5604,5884,5897,6011,6016,6021,6077,6191,6276,6281,6382,6401,6415,6458,6594,6755,6941,6947,7016,7140,7218,7337,7360,7380,7381,7436,7442,7451,7541,7568,7771,7774,7968,7978,8059,8104,8125,8244,8259,8267,8325,8495,8531,8566,8616,8702,8897,8925,8996,9005,9064,9153,9193,9376,9449,9569,9678,9746,9825,9829,9831,9913,9978,9992,10041,10069,10145,10196,10208,10282,10290,10292,10327,10385,10612,10677,10688,10744,10763,10779,10855,10949,11002,11010,11026,11136,11239,11504,11517,11595,11725,11736,11849,11935,12034,12195,12218,12224,12283,12312,12333,12345,12376,12389,12402,12441,12478,12547,12574,12635,12680,12774,12801,12957,13187,13249,13540,13585,13702,13735,13786,13790,13905,13915,13943,13953,13985,14030,14175,14229,14329,14359,14643,14687,14694,14702,14810,14925,15113,15157,15349,15380,15472,15571,15595,15641,15661,15664,15685,15848,15869,15870,15873,15931,15961,16030,16247,16415,16425,16565,16667,16998,17166,17182,17187,17311,17354,17367,17388,17533,17553,17597,17602,17635,17720,17829,17852,17977,18188,18218,18246,18317,18368,18457,18710,18751,18755,18850,18871,18902,18956,18974,19239,19271,19313,19404,19484,19818,19857,19991,20048,20389,20495,20508,20535,20747,20793,20816,20817,20860,20866,20873,20919,20983,21000,21011,21107,21384,21451,21514,21563,21566,21731,21817,22162,22251,22271,22280,22361,22384,22544,22630,22634,22635,22670,22729,22817,22857,22858,23075,23080,23148,23228,23286,23331,23345,23416,23421,23451,23559,23661,23762,23843,23853,23947,23982,24085,24109,24149,24150,24170,24220,24262,24364,24445,24486,24529,24534,24598,24679,24727,24780,24851,24855,24899,25107,25253,25270,25284,25313,25314,25387,25420,25584,25737,25812,25814,25879,25893,26038,26077,26239,26274,26321,26533,26548,26626,26630,26702,26745,26900,26936,27034,27171,27333,27469,27570,27656,27725,27977,28207,28241,28480,28516,28549,28653,28864,28865,28949,29067,29225,29244,29424,29461,29541,29638,29648,29650,29673,29732,29770,29793,29827,29879,30179,30183,30341,30342,30347,30353,30436,30504,30520,30541,30551,30650,30656,30767,30870,30919,30941,31084,31085,31313,31385,31454,31460,31480,31523,31735,31900,31998,32005,32120,32122,32136,32170,32221,32365,32387,32437,32711,32802,32814,32874,32946,33065,33165,33200,33393,33485,33640,33670,33824,33831,33969,33977,34061,34066,34099,34272,34318,34383,34395,34595,34672,34756,34881,34883,34937,35085,35093,35332,35339,35485,35744,35782,35821,35840,35955,35972,36031,36065,36123,36208,36370,36615,36617,36669,36693,36909,36964,37040,37045,37054,37094,37172,37237,37247,37267,37431,37624,37659,37672,37695,37768,37805,37910,37938,37970,38049,38131,38142,38215,38285,38309,38368,38387,38388,38392,38415,38417,38457,38464,38468,38480,38573,38776,38909,38930,38969,39165,39200,39210,39217,39251,39261,39307,39336,39349,39427,39532,39583,39686,39777,39802,39821,39842,39850,39906,39913,39915,39937,39940,39943,40044,40116,40146,40278,40284,40378,40420,40500,40515,40517,40545,40745,40830,41044,41059,41130,41353,41394,41471,41477,41632,41675,41848,41886,41921,41924,42177,42221,42267,42283,42328,42483,42595,42626,42631,42676,42720,42763,42770,42854,42888,42964,42977,43057,43110,43116,43120,43125,43176,43267,43282,43385,43455,43509,43628,43676,43714,43878,43904,43929,43959,44031,44115,44157,44204,44205,44287,44404,44516,44539,44635,44758,44909,44965,44981,45090,45096,45223,45291,45387,45388,45645,45682,45947,45998,46034,46061,46078,46079,46126,46494,46528,46562,46641,46647,46652,46678,46736,46811,46814,46933,46983,47196,47222,47264,47332,47496,47524,47593,47600,47672,47722,47730,47793,47799,48034,48103,48107,48228,48288,48382,48402,48414,48506,48617,48684,48695,48716,48732,48800,48830,48896,48967,49052,49288,49302,49383,49413,49460,49503,49546,49561,49715,49768,49875,49950,50008,50136,50166,50170,50261,50473,50476,50511,50538,50548,50574,50636,50645,50658,50801,50807,50829,51094,51321,51408,51413,51499,51601,51629,51652,51848,52048,52166,52181,52193,52214,52236,52273,52408,52446,52545,52641,52662,52680,52720,52867,53085,53176,53261,53329,53434,53487,53543,53548,53708,53828,53919,53930,54064,54077,54110,54143,54256,54433,54734,55014,55165,55192,55279,55325,55450,55460,55581,55726,55742,55745,55805,56319,56331,56407,56566,56632,56725,56742,56752,56759,56763,56771,56818,57050,57121,57147,57150,57199,57269,57316,57502,57525,57564,57566,57604,57605,57608,57710,57720,57745,57872,57926,57967,58054,58101,58138,58161,58194,58413,58428,58578,58610,58717,58800,58931,59010,59024,59065,59078,59080,59196,59306,59403,59415,59449,59516,59523,59577,59673,59712,59735,59743,59825,59858,60132,60209,60299,60321,60540,60566,60575,60652,60834,60898,60987,61007,61078,61110,61129,61338,61539,61745,61794,61800,61802,61858,61865,61885,61958,62197,62204,62225,62295,62329,62384,62437,62449,62471,62486,62532,62773,62782,62942,63086,63205,63210,63393,63397,63472,63540,63571,63667,63752,63900,64255,64298,64381,64421,64446,64542,64598,64634,64685,64705,64801,64817,64865,65097,65125,65355,65389,65400,65682,65856,65868,65898,66189,66249,66274,66325,66347,66390,66519,66586,66667,66714,66793,66821,66865,66879,66940,67074,67075,67109,67142,67164,67245,67482,67626,67680,67781,67900,67919,67929,68143,68176,68227,68236,68301,68358,68399,68431,68594,68688,68773,68799,68848,68904,68934,69040,69083,69170,69242,69255,69269,69358,69553,69686,69844,69993,70018,70034,70105,70123,70257,70264,70270,70337,70375,70423,70454,70471,70477,70523,70633,70650,70708,70806,71025,71038,71069,71070,71120,71171,71203,71205,71242,71279,71329,71366,71408,71576,71681,71694,71904,71990,72003,72071,72332,72362,72568,72585,72799,72842,72941,72977,73044,73121,73144,73200,73250,73626,73830,73940,73983,74014,74046,74085,74225,74443,74658,74702,74858,74972,75050,75155,75584,75745,75761,75802,75805,75866,75907,75925,76043,76050,76159,76219,76242,76305,76428,76526,76822,76829,77017,77040,77110,77135,77327,77350,77480,77488,77516,77556,77581,77609,77622,77626,77641,77652,77657,77739,77969,78019,78056,78171,78456,78457,78495,78533,78614,78763,78828,78865,78883,78902,78916,79094,79185,79311,79370,79421,79630,79738,79827,79835,79869,79938,80001,80027,80118,80138,80328,80392,80431,80604,80650,80844,80969,80997,81036,81078,81097,81155,81165,81308,81476,81493,81612,81652,81819,81882,81922,81962,82142,82239,82254,82290,82463,82509,82663,82739,82863,83143,83150,83190,83285,83460,83475,83487,83515,83638,83822,83890,84016,84051,84111,84319,84342,84407,84446,84481,84486,84545,84696,84795,84868,85039,85041,85055,85165,85183,85305,85317,85380,85521,85540,85738,85791,85804,85849,85876,85936,85983,86025,86053,86126,86353,86511,86568,86883,87059,87330,87389,87507,87630,87696,87710,87796,87807,88114,88246,88366,88416,88470,88492,88525,88561,88662,88751,88870,88894,88949,88975,89121,89126,89312,89316,89493,89527,89737,89946,90044,90053,90061,90079,90085,90365,90421,90544,90600,90638,90812,90928,91028,91111,91208,91247,91286,91311,91340,91346,91352,91370,91605,91722,92203,92213,92293,92300,92307,92398,92443,92472,92540,92773,92925,93026,93125,93156,93212,93227,93269,93320,93414,93434,93485,93522,93530,93549,93738,94047,94131,94146,94264,94389,94419,94591,94705,94712,94769,94911,94988,95388,95480,95530,95693,95717,95871,95907,95930,95932,96174,96224,96265,96314,96322,96474,96514,96615,96964,97109,97188,97314,97360,97396,97520,97586,97640,97757,98029,98050,98104,98250,98256,98266,98368,98426,98462,98638,98664,98671,98680,98682,98798,98842,98986,99034,99035,99118,99155,99192,99472,99638,99782,99792,99849,99879,99947,100001",
          "6,331,336,340,347,354,356,391,395,404,430,438,442,462,467,494,519,630,684,764,835,849,862,865,917,959,1021,1048,1188,1233,1375,1478,1521,3102,5026,5084,5148,5167,5188,5220,5227,5228,5255,5269,5304,5314,5328,5380,5427,5431,5561,5628,5690,5796,5800,5805,5856,5865,5891,5909,5921,5948,5949,5974,6187,6202,6256,6560,6606,6824,6851,7062,7191,7246,7347,7364,7403,7753,8137,8158,8226,8356,8476,8959,9057,9096,9260,9274,9359,9389,10003,10034,10492,11728,13574,13913,15445,15527,16479,16760,18089,18353,19435,21003,22079,25936,26788,40385,40998,41658,46877,48076,50093,50103,50165,50283,50343,50370,50423,50462,50489,50716,50896,50963,50980,51067,51084,51138,51164,51175,51250,51314,51400,51445,51448,51474,51550,51602,51865,51970,51983,52070,52180,52195,52281,52370,52461,52529,52549,52556,52809,53002,53077,53081,53225,53319,53395,53423,53566,53642,53831,53911,53912,54109,54205,54245,54251,54296,54299,54307,54343,54351,54523,54665,54777,54814,54844,54933,54948,54970,55261,55311,55473,55478,55498,55772,55819,55953,55956,56054,56079,56160,56757,56833,56904,57074,57455,57582,57606,57731,57746,57766,58050,58071,58122,58237,58282,58489,58572,58591,58629,58667,58672,58767,58861,58939,58950,58960,58990,59019,59042,59130,59170,59239,59241,59388,59513,59636,59676,59776,59784,59797,59894,59917,60002,60078,60082,60146,60242,60452,60527,60581,60698,60793,60894,60900,60980,60985,60990,61026,61037,61163,61175,61402,61509,61512,61535,61624,61648,61680,61755,61782,61906,62058,62069,62070,62190,62275,62409,62440,62490,62500,62518,62537,62771,62827,62868,62959,62962,63013,63092,63115,63158,63218,63334,63371,63408,63423,63514,63542,63551,63687,63720,63739,63741,63763,63868,64081,64111,64124,64393,64426,64556,64631,64728,64844,64903,64946,65071,65099,65196,65354,65390,65420,65640,65696,65743,65766,65890,65912,66012,66025,66034,66041,66052,66053,66096,66250,66423,66432,66443,66449,66603,66721,67046,67080,67093,67255,67395,67512,67678,67697,68087,68144,68171,68179,68334,68368,68373,68468,68506,68546,68610,68633,68912,69026,69289,69598,69756,69907,69974,70015,70147,70231,70342,70345,70358,70365,70657,70682,70709,70795,70897,71032,71163,71177,71274,71448,71724,71737,71838,71956,71957,71987,72029,72190,72483,72551,72658,72925,73041,73203,73239,73319,73349,73463,73856,73969,73980,74016,74228,74241,74368,74513,74659,74674,74679,74721,74815,74891,74961,75185,75268,75392,75407,75720,75762,75803,75859,75875,75942,76070,76343,76375,76477,76517,76519,76597,76668,76693,76707,76890,76935,76980,77030,77346,77416,77465,77585,77629,77675,77763,77779,77838,77998,78016,78024,78054,78185,78242,78321,78371,78542,78595,78743,78800,78854,78921,78974,78975,79239,79321,79352,79504,79549,79576,79596,79747,79818,80242,80260,80274,80293,80320,80387,80674,80693,80806,80888,80896,81122,81383,81426,81458,81515,81646,81694,81724,81746,81758,81901,81936,82051,82084,82169,82209,82354,82465,82504,82593,82664,82681,82708,82748,82810,82856,82907,82952,82985,83005,83055,83147,83199,83302,83332,83367,83410,83415,83593,83609,83649,83735,83930,83995,84054,84132,84176,84210,84377,84395,84720,84810,84932,84937,84964,85048,85148,85204,85213,85223,85251,85346,85357,85457,85517,85613,85863,85870,86007,86092,86094,86252,86323,86396,86440,86484,86534,86630,86720,86803,86835,86897,86972,86992,87073,87229,87334,87399,87427,87505,87624,87638,87684,87827,87854,88041,88307,88330,88667,88825,88875,88919,88936,88960,88985,89107,89112,89124,89150,89206,89297,89616,89748,89786,89789,89891,89949,90022,90111,90176,90452,90485,90525,90568,90598,90670,90699,90833,90850,90913,90933,90943,91087,91150,91216,91285,91526,91567,91584,91752,91933,91944,92022,92050,92082,92099,92241,92284,92291,92393,92611,92721,92842,92883,93126,93160,93199,93225,93263,93520,93540,93651,93684,93757,93851,93952,93999,94183,94190,94197,94213,94352,94511,94546,94593,94603,94639,94762,94805,94831,94846,94847,94978,95037,95146,95339,95350,95446,95548,95616,95651,95748,95929,95934,95946,95990,96046,96107,96139,96202,96249,96348,96349,96488,96590,96599,96686,96710,96829,96845,96943,97037,97128,97135,97151,97241,97347,97425,97564,97646,97658,97681,97693,97708,97733,97742,97765,97787,97794,97838,97881,97903,97909,97951,98005,98048,98163,98167,98260,98275,98656,98704,98826,98867,99042,99121,99159,99386,99454,99556,99716,99735",
          "68,553,693,697,730,809,949,963,1065,1134,1148,1291,1320,1329,1361,1428,1441,1553,1758,1804,1863,1892,1909,1914,1926,1969,2154,2244,2248,2319,2402,2473,2617,2659,2733,2825,2852,2907,2995,3176,3202,3276,3414,3461,3468,3469,3489,3542,3611,3636,3671,3691,3834,3841,3955,4035,4131,4171,4393,4449,4505,4564,4578,4646,4920,4925,4927,4956,5267,5286,5327,5362,5481,5541,5650,5720,5779,5806,5878,5938,6084,6273,6572,6595,6680,6864,6870,6958,6965,6990,7232,7247,7265,7273,7277,7280,7367,7398,7511,7621,7638,7640,7778,7861,7882,7917,8054,8147,8261,8409,8489,8643,8649,8780,8801,8825,8890,8952,9013,9052,9161,9201,9279,9436,9469,9634,9664,9722,9874,9887,10045,10144,10252,10403,10471,10530,10630,10884,10902,10909,10914,10918,11318,11445,11571,11779,11836,11866,11918,11936,12071,12081,12116,12249,12266,12522,12559,12595,12762,13100,13121,13376,13410,13631,13685,13800,13817,13841,13939,14008,14045,14237,14240,14467,14476,14510,14539,14615,14714,15003,15031,15100,15109,15185,15286,15409,15450,15501,15517,15596,15668,15968,16184,16215,16274,16432,16502,16519,16603,16658,16743,16886,16930,16970,17027,17066,17119,17165,17209,17353,17477,17541,17670,17703,17975,17976,17985,18025,18105,18130,18155,18243,18266,18429,18462,18539,18578,18707,18717,18753,18776,18942,18955,18980,19014,19101,19177,19232,19357,19424,19473,19478,19505,19533,19539,19541,19596,19601,19607,19631,19671,19733,19780,19861,19863,19894,19919,20259,20279,20297,20299,20315,20329,20452,20459,20541,20562,20583,20696,20962,21063,21106,21115,21221,21656,21750,22050,22116,22304,22356,22363,22398,22579,22759,22831,22907,23033,23036,23074,23203,23226,23272,23340,23368,23488,23721,23940,23958,24051,24074,24154,24204,24268,24355,24378,24456,24488,24597,24651,24713,24783,24815,24879,24887,24910,24971,25054,25062,25084,25138,25219,25235,25276,25455,25545,25725,25843,25998,26107,26226,26227,26228,26261,26431,26455,26472,26479,26620,26846,26852,26902,26950,27041,27069,27530,27547,27569,27675,27707,27760,27811,27936,28075,28183,28233,28303,28307,28334,28368,28387,28471,28511,28577,28611,28672,28708,28834,28835,28871,28926,29090,29143,29184,29190,29260,29291,29348,29367,29375,29561,29741,29759,29779,29869,30166,30208,30326,30433,30507,30564,30589,30599,30605,30649,30654,30662,30667,30764,30777,30835,31083,31170,31240,31248,31249,31489,31515,31544,31617,31766,31779,32102,32150,32174,32262,32300,32406,32563,32667,32851,33001,33046,33061,33066,33251,33314,33630,33635,33688,33690,33752,33845,33915,34002,34048,34288,34319,34375,34728,34790,34888,34895,34901,35064,35213,35533,35542,35651,35738,36005,36007,36051,36060,36061,36072,36074,36155,36213,36220,36226,36385,36413,36425,36504,36620,36626,36695,36762,36858,36863,37005,37070,37090,37133,37181,37192,37236,37304,37384,37451,37529,37567,37715,37742,38213,38238,38296,38513,38558,38567,38594,38599,38690,38712,38715,38774,38778,38904,38955,39140,39156,39164,39184,39270,39515,39634,39635,39657,39713,39826,39864,39891,39900,39912,40027,40133,40509,40511,40552,40613,40672,40832,40904,40951,41029,41049,41051,41252,41339,41382,41441,41446,41475,41623,41678,41748,41761,41786,41847,41932,41973,42092,42179,42366,42378,42379,42419,42428,42558,42706,42708,42771,42789,42899,42924,42994,43064,43080,43114,43228,43313,43362,43364,43475,43588,43720,44026,44048,44154,44216,44233,44236,44239,44407,44444,44625,44681,44691,44705,44898,44927,45014,45138,45152,45179,45217,45384,45479,45547,45591,45755,45814,45865,45874,46074,46077,46080,46173,46196,46216,46291,46322,46327,46441,46795,46834,46840,47072,47084,47086,47298,47299,47302,47350,47401,47481,47507,47618,47645,47684,47829,47858,47863,47989,48019,48037,48117,48143,48174,48202,48220,48232,48436,48472,48497,48535,48628,48652,48733,48797,48886,48949,48959,48966,48990,49009,49059,49120,49212,49215,49274,49298,49320,49345,49386,49405,49456,49481,49494,49536,49651,49684,49725,49977,50111,50268,50289,50474,50478,50479,50544,50683,50694,50696,50697,50749,50820,50946,50979,50990,51017,51163,51184,51316,51359,51375,51502,51541,51665,52142,52249,52431,52649,52665,52679,52899,52912,52931,52979,53010,53099,53180,53181,53202,53245,53295,53417,53424,53712,53756,53783,53996,54013,54158,54264,54436,54470,54625,54693,54710,54739,54835,55001,55078,55341,55344,55379,55385,55420,55470,55548,55666,55935,55984,56094,56169,56233,56301,56307,56329,56361,56374,56512,56745,56779,56787,56798,56899,56953,56969,57013,57067,57109,57153,57368,57781,57787,58171,58223,58417,58462,58494,58514,58532,58634,58644,58676,58703,58751,58958,59108,59136,59146,59187,59313,59409,59434,59646,59912,59962,59999,60046,60188,60328,60340,60374,60380,60550,60660,60665,60721,60896,60933,61116,61133,61231,61311,61342,61347,61384,61566,61578,61649,61675,61754,61809,61820,61895,61991,62045,62285,62299,62501,62556,62560,62576,62595,62596,62677,62710,62787,62885,62948,62955,63065,63235,63388,63420,63624,63628,63777,63786,63865,63950,63959,63963,64055,64077,64083,64311,64392,64449,64486,64692,64707,64732,64738,64919,65004,65068,65083,65186,65261,65347,65485,65490,65494,65630,65666,65951,66159,66196,66276,66372,66438,66604,66613,66786,67014,67090,67101,67204,67408,67441,67481,67516,67635,67933,67935,67938,67969,67971,67992,68022,68250,68268,68275,68465,68479,68530,68536,68638,68649,68676,68702,68721,68734,68736,68852,68905,69065,69149,69264,69471,69518,69583,69610,69618,69683,69705,69769,69833,69839,69890,69984,70064,70085,70139,70169,70181,70269,70284,70591,70687,70730,70784,70839,71019,71148,71251,71301,71546,71558,71644,71726,71729,71731,72212,72554,72555,72561,72671,72849,73014,73181,73183,73518,73602,73651,73653,73669,73826,73837,73858,73962,73965,74123,74257,74288,74391,74414,74432,74447,74595,74636,74638,74889,74945,74999,75074,75154,75179,75199,75219,75281,75286,75420,75724,75841,75884,75944,75999,76056,76133,76251,76632,76726,76739,76754,76800,76837,77026,77036,77331,77332,77371,77438,77515,77582,77700,77709,77851,77911,77977,78040,78307,78436,78497,78601,78647,78751,79194,79329,79555,79713,79773,79861,79888,79943,80047,80106,80259,80289,80359,80376,80381,80486,80556,80698,80794,80891,80904,80913,81022,81131,81253,81517,81556,81622,81659,81668,81714,81722,81810,81872,82043,82067,82157,82243,82280,82308,82765,82768,82878,82924,83034,83106,83153,83178,83274,83359,83510,83643,83656,83690,83750,83800,83819,83918,84118,84138,84213,84253,84276,84281,84295,84306,84341,84360,84365,84400,84417,84460,84510,84526,84845,84918,84946,85056,85270,85643,85684,85703,85705,85729,85782,85895,85921,85925,86188,86263,86399,86441,86560,86586,86746,86818,86994,87122,87267,87358,87576,87582,87644,87674,87828,87888,88197,88233,88262,88282,88319,88352,88384,88393,88514,88618,88655,88858,88915,89100,89125,89147,89157,89158,89241,89350,89536,89672,89779,89832,89835,89840,89896,89914,89920,89971,90020,90047,90073,90410,90524,90604,90814,90871,90944,90981,91008,91044,91564,91661,91673,91834,91961,92040,92120,92187,92347,92417,92434,92449,92564,92719,92754,92797,92805,92854,92878,92942,93054,93092,93198,93211,93351,93378,93560,93635,93646,93699,93840,93857,94004,94033,94172,94321,94529,94794,94905,94995,95043,95090,95096,95112,95279,95354,95416,95495,95630,95677,95691,95730,95810,95822,96066,96228,96384,96407,96507,96511,96573,96652,96857,96861,96926,96960,97077,97184,97196,97228,97261,97303,97304,97352,97367,97372,97430,97447,97700,97812,97893,97993,98173,98217,98280,98289,98308,98397,98429,98530,98563,98591,98600,98616,98618,98751,98854,98856,98959,98984,99149,99206,99263,99305,99398,99501,99535,99583,99601,99642,99771,99831,99878,99925,99935",
          "94,144,158,183,243,542,582,663,847,1117,1187,1328,1365,1425,1432,1642,1650,1663,1746,1778,1818,1836,2005,2015,2077,2093,2112,2116,2142,2206,2260,2358,2382,2413,2438,2441,2560,2581,2652,2661,2837,2841,2849,3034,3119,3160,3330,3394,3407,3444,3569,3625,3721,3722,3767,3929,3965,4022,4084,4134,4192,4379,4447,4456,4519,4597,4632,4671,4692,4816,4913,4960,5252,5355,5462,5465,5479,5517,5565,5575,5635,5700,5824,5831,5951,6029,6067,6072,6108,6110,6196,6265,6270,6418,6431,6450,6732,6750,6796,6867,6964,7072,7447,7586,7715,7743,7745,7940,7993,8020,8155,8197,8200,8221,8273,8372,8398,8513,8607,8631,8687,8688,8731,8884,8904,8910,8997,9124,9222,9278,9385,9390,9425,9576,9631,9661,9699,9708,9711,9812,9854,9856,10025,10090,10182,10194,10199,10271,10298,10324,10519,10529,10604,10616,10636,10650,10813,10817,10867,11020,11038,11053,11056,11089,11206,11216,11224,11267,11402,11449,11621,11644,11756,11761,11898,11913,11974,11987,12054,12086,12138,12144,12311,12433,12514,12544,12581,12800,12866,12869,12896,12946,12978,13131,13242,13244,13289,13297,13314,13408,13517,13597,13598,13689,13696,13716,13726,13823,13865,13933,14015,14050,14091,14239,14311,14392,14436,14444,14495,14497,14580,14649,14710,14974,15016,15018,15038,15071,15160,15248,15265,15273,15370,15427,15535,15612,15694,15703,15740,15787,16139,16141,16170,16202,16286,16407,16500,16527,16530,16554,16634,16751,16779,16851,16916,16965,16972,17020,17157,17245,17279,17306,17370,17390,17419,17437,17557,17665,17687,17841,17889,18019,18080,18084,18107,18156,18166,18269,18294,18411,18431,18442,18496,18591,18696,18720,18765,18860,18862,18886,18893,18935,18979,19020,19070,19079,19115,19128,19159,19223,19364,19430,19546,19568,19571,19655,19753,19794,19838,19872,19890,20019,20042,20047,20121,20141,20158,20200,20393,20527,20573,20580,20813,20833,20849,20945,21006,21260,21472,21775,21787,21865,21868,21889,22120,22164,22589,22673,22689,22771,22816,22886,22892,22987,22990,23008,23092,23199,23274,23276,23318,23384,23390,23405,23425,23537,23664,23687,23723,23900,23931,24052,24070,24159,24173,24212,24235,24336,24376,24473,24503,24558,24585,24586,24595,24625,24816,24822,24949,24960,24979,24994,25074,25139,25205,25223,25335,25337,25507,25514,25520,25532,25588,25621,25976,26024,26083,26264,26266,26273,26418,26433,26587,26639,26682,26845,26864,26882,26916,26983,27285,27313,27359,27373,27481,27542,27555,27676,27689,27783,27870,28177,28202,28227,28279,28565,28594,28705,28766,28856,29053,29072,29130,29179,29255,29270,29342,29417,29425,29433,29577,29606,29615,29666,29718,29785,29874,29910,30009,30023,30056,30081,30129,30144,30196,30283,30452,30455,30609,30730,30762,30800,30855,30974,31056,31211,31224,31260,31325,31392,31476,31491,31525,31565,31609,31660,31706,31801,31839,31869,31965,32055,32107,32131,32229,32494,32528,32605,32627,32729,32807,33013,33121,33212,33258,33268,33280,33299,33345,33458,33609,33645,33653,33737,33740,33797,33799,33833,33959,34128,34130,34141,34396,34740,34760,34856,34870,34898,34924,35154,35385,35405,35453,35566,35593,35831,35854,35877,35893,35901,35929,36104,36143,36228,36234,36362,36561,36714,36837,36847,36904,36930,37077,37241,37307,37398,37484,37531,37534,37647,37655,37744,37785,37809,37874,37934,38335,38406,38414,38416,38492,38557,38574,38586,38650,38656,38693,38716,38767,38780,38789,38800,38855,38871,38913,38942,39001,39033,39046,39050,39114,39191,39296,39520,39619,39680,39693,39721,39771,39835,39911,39917,39946,40051,40086,40108,40220,40236,40269,40389,40394,40502,40528,40544,40696,40741,40791,40839,40912,40959,41021,41125,41357,41450,41483,41500,41544,41614,41676,41873,41925,41943,41949,41958,42245,42299,42355,42361,42368,42384,42404,42463,42471,42499,42715,42753,42801,42809,42814,42950,42954,42966,43153,43283,43315,43319,43357,43456,43549,43550,43648,43824,43914,44064,44096,44132,44181,44213,44240,44269,44321,44382,44611,44639,44710,44883,44907,44911,44952,45055,45121,45187,45386,45557,45566,45605,45626,45685,45815,45868,46148,46351,46355,46368,46373,46408,46463,46478,46598,46614,46630,46655,46685,46743,46845,46853,46866,46875,47101,47267,47297,47443,47504,47516,47525,47536,47588,47636,47673,47704,47729,47751,47789,47804,47843,47879,47966,47979,48112,48140,48210,48214,48392,48422,48445,48600,48710,48727,48743,48893,49065,49140,49157,49225,49259,49269,49318,49340,49617,49637,49761,49806,49859,49902,49948,49982,50016,50101,50121,50203,50213,50367,50490,50500,50530,50628,50659,50676,50709,50768,50825,50852,50902,51016,51049,51123,51127,51150,51207,51216,51224,51286,51355,51479,51534,51592,51600,51651,51688,51705,51780,51790,51935,51936,51955,52054,52152,52155,52385,52448,52480,52491,52702,52770,52815,52819,52840,52894,52990,53006,53044,53213,53269,53282,53307,53366,53400,53410,53460,53494,53536,53614,53715,53723,53729,53790,53901,54101,54106,54275,54352,54397,54398,54399,54582,54735,54741,54851,54898,54899,54927,55065,55155,55162,55187,55190,55206,55438,55488,55489,55642,55645,55699,55783,55838,56008,56044,56060,56170,56219,56226,56278,56281,56348,56456,56462,56472,56534,56547,56686,56749,56820,56855,56876,56886,56949,56957,57117,57187,57223,57271,57300,57383,57404,57450,57532,57533,57700,57737,57809,58029,58239,58337,58398,58460,58574,58643,58781,58929,58965,59012,59022,59054,59071,59093,59100,59165,59279,59336,59527,59576,59583,59602,59754,59813,59855,59873,60314,60399,60426,60428,60455,60466,60467,60478,60483,60599,60607,60798,60852,60979,61019,61036,61126,61195,61261,61303,61379,61592,61652,61664,61770,61842,61876,61927,61968,62081,62108,62145,62149,62173,62272,62276,62296,62367,62531,62616,62919,62964,62990,63107,63213,63273,63361,63424,63448,63502,63594,63670,63679,63796,63896,63903,64025,64040,64214,64231,64285,64331,64374,64460,64477,64495,64677,64678,64750,64798,64831,64854,64870,64932,65072,65077,65158,65208,65426,65461,65556,65567,65579,65661,65692,65904,65935,65936,66061,66114,66169,66239,66309,66392,66650,66705,66726,66737,66777,66807,66819,66824,66914,66933,67036,67111,67148,67155,67191,67251,67288,67307,67360,67388,67472,67558,67605,67653,67906,67931,67947,67983,67986,68025,68048,68072,68148,68160,68226,68394,68401,68445,68516,68577,68582,68878,68931,69097,69302,69408,69501,69538,69611,69651,69796,69821,69855,69872,69888,69983,70133,70177,70222,70241,70434,70460,70511,70587,70779,70825,70981,71023,71061,71155,71187,71338,71345,71475,71516,71583,71641,71716,71768,71789,71952,71965,72036,72042,72128,72131,72174,72205,72225,72261,72339,72426,72461,72598,72706,72755,72938,72939,73179,73421,73426,73435,73472,73553,73575,73742,73757,73778,73876,74086,74328,74355,74383,74384,74446,74460,74557,74951,74975,75051,75114,75182,75192,75284,75301,75345,75380,75411,75459,75468,75549,75609,75687,75796,75976,75984,76164,76243,76265,76388,76466,76531,76710,76789,76803,76820,77088,77117,77163,77275,77597,77645,77780,77810,77924,77933,77987,78117,78323,78338,78355,78383,78396,78523,78936,78957,79028,79061,79064,79241,79267,79290,79318,79372,79422,79475,79552,79589,79628,79671,79842,79894,79955,79999,80024,80068,80111,80145,80263,80297,80454,80461,80516,80521,80548,80588,80626,80738,81005,81038,81319,81344,81380,81388,81399,81609,81633,81916,81972,82038,82179,82294,82521,82614,82709,82725,82806,83165,83263,83289,83341,83380,83413,83440,83508,83595,83726,83766,83814,83837,84420,84529,84657,84661,84695,84722,84823,84850,84933,85011,85053,85152,85172,85179,85323,85423,85541,85607,85614,85821,85838,85890,85980,85996,86111,86163,86269,86535,86628,86715,86844,86916,86940,87008,87027,87061,87072,87193,87344,87480,87488,87651,87829,87838,87953,88128,88215,88237,88332,88350,88476,88515,88526,88543,88656,88678,88862,88931,89129,89171,89247,89253,89275,89339,89501,89520,89594,89646,89685,89704,89841,89857,89936,89961,90007,90104,90168,90170,90341,90355,90373,90414,90493,90494,90539,90574,90675,90705,90737,90752,90809,90998,91187,91265,91329,91389,91405,91456,91529,91592,91595,91664,91713,91766,91830,91886,91921,91955,92045,92068,92110,92276,92294,92437,92448,92527,92577,92584,92590,92617,92640,92701,92748,92853,92958,92973,93090,93109,93116,93189,93190,93215,93303,93486,93567,93581,93589,93634,93703,93707,93863,93970,94061,94068,94120,94140,94302,94415,94647,94679,94774,94853,94865,94903,95079,95137,95158,95246,95270,95565,95639,95689,95719,95754,95763,95859,96184,96363,96462,96541,96589,96780,96807,96825,96832,96898,96917,96947,96949,97039,97066,97179,97299,97375,97527,97583,98097,98106,98192,98226,98396,98479,98695,98965,98981,99181,99189,99221,99297,99334,99364,99399,99418,99512,99589,99617,99622,99645,99800,99917,99934",
          "275,316,392,407,411,425,448,470,486,546,608,613,683,711,742,818,840,854,934,999,1035,1204,1250,5034,5201,5207,5211,5230,5257,5287,5301,5316,5329,5356,5384,5426,5477,5523,5597,5632,5765,5770,5772,5787,5803,5821,5826,5888,5908,5929,5967,5988,6035,6190,6411,6731,6771,6832,6968,7013,7078,7109,7134,7152,7287,7384,7467,7534,7641,7839,8284,8440,8542,8594,8624,8749,8824,8856,8914,9094,9406,9533,9620,9695,9859,10200,10402,10552,10588,11103,11823,11948,12126,12400,12729,13546,14170,15548,17118,18528,19231,20940,20968,24096,26671,29483,32996,41556,42357,50003,50143,50225,50375,50523,50611,50615,50680,50745,50879,50943,51105,51113,51352,51435,51575,51663,51675,51691,51727,51752,51776,52015,52104,52253,52317,52501,52601,52718,52729,52784,52968,53116,53178,53321,53339,53442,53589,53674,53753,53813,53817,53913,54065,54127,54144,54284,54337,54348,54419,54424,54441,54459,54606,54626,54652,54705,54730,54872,54894,54965,54966,54983,55129,55173,55180,55237,55256,55378,55456,55553,55738,55759,55779,55793,55941,56081,56142,56247,56379,56481,56495,56585,56696,56804,56875,56944,57064,57332,57413,57444,57453,57619,57664,57875,57913,58094,58235,58360,58375,58380,58536,58559,58648,58651,58711,58761,58880,59140,59216,59277,59380,59395,59445,59726,59824,59842,59889,59930,60084,60157,60187,60215,60294,60460,60620,60621,60687,60757,60768,60851,60905,60942,61167,61179,61260,61298,61322,61326,61412,61465,61567,61616,61758,61897,62011,62450,62481,62535,62705,62729,62730,62976,62999,63277,63353,63412,63528,63557,63574,63788,63881,63891,63894,63948,63996,64007,64089,64134,64161,64163,64188,64287,64312,64367,64429,64527,64646,64764,64795,64863,65111,65247,65395,65442,65443,65452,65515,65518,65660,65730,65780,65865,65887,66048,66264,66451,66494,66641,66710,67140,67173,67285,67368,67458,67473,67499,67509,67543,67556,67888,68006,68020,68310,68338,68562,68568,68614,68818,68999,69301,69376,69431,69545,69681,69747,69760,70033,70101,70148,70290,70510,70543,70649,70725,70758,70794,70827,70909,70983,71006,71231,71243,71499,71552,71574,71760,71896,71955,71975,72063,72078,72150,72260,72290,72338,72343,72403,72420,72464,72589,72620,72758,72762,72887,72976,73001,73004,73103,73106,73227,73236,73286,73320,73425,73475,73477,73599,73619,73701,73750,73854,73972,74056,74100,74137,74163,74229,74463,74720,74722,74919,75044,75079,75109,75178,75305,75431,75463,75490,75495,75598,75641,75665,75867,75913,76049,76073,76172,76222,76248,76254,76347,76374,76379,76438,76450,76674,76764,76771,76806,76875,76887,76949,77144,77176,77226,77234,77277,77315,77412,77473,77528,77561,77674,77720,77828,77831,77979,78041,78150,78231,78241,78339,78458,78472,78483,78553,78947,78954,79013,79026,79079,79090,79101,79338,79362,79790,79844,80020,80196,80258,80285,80317,80360,80490,80593,80635,80866,80903,80961,80985,81009,81047,81055,81231,81289,81360,81575,81593,81642,81674,81693,81830,81852,81927,81967,82025,82104,82176,82227,82228,82275,82293,82384,82506,82648,82658,82746,82824,82995,83029,83123,83172,83212,83253,83374,83611,83629,83657,83661,83681,83722,83732,83857,83922,84327,84437,84439,84442,84531,84721,84770,84904,85129,85205,85324,85354,85664,85732,85928,85960,86029,86147,86205,86310,86400,86516,86579,86679,86721,86831,86966,87007,87090,87108,87113,87178,87240,87469,87543,87561,87599,87676,87681,88033,88065,88113,88130,88151,88389,88435,88571,88576,88587,88595,88731,88879,89005,89020,89028,89055,89141,89433,89437,89461,89631,89671,89751,89754,89826,90069,90092,90288,90372,90518,90519,90660,90667,90975,90992,91201,91229,91249,91278,91295,91412,91415,91440,91593,91616,91647,91756,91823,91915,91946,92027,92081,92104,92128,92189,92319,92735,92810,92848,93051,93148,93181,93239,93254,93296,93345,93368,93484,93497,93542,93766,93779,93793,94049,94134,94191,94370,94446,94456,94718,94761,94917,94918,95019,95059,95243,95621,95841,95919,96021,96026,96099,96105,96198,96283,96286,96649,96714,96798,96873,96971,97130,97176,97208,97331,97382,97494,97791,97796,97865,97868,97958,98148,98164,98241,98507,98558,98567,98615,98654,98696,98846,98877,98996,99152,99315,99381,99412,99727,99754,99819,99828,99835,99860,99866,99892",
          "72,402,550,653,766,767,795,919,983,1028,1295,1395,1460,1491,1597,1656,1669,1712,1732,1792,1889,1952,1987,2038,2118,2219,2264,2337,2338,2415,2453,2536,2548,2555,2576,2827,2865,2883,2959,3216,3306,3425,3729,3807,3835,4191,4202,4245,4434,4466,4468,4496,4514,4670,4697,4875,4876,4912,5266,5276,5303,5605,5829,6203,6213,6234,6339,6354,6615,6704,6742,6783,6819,6848,6859,6890,6902,6909,7096,7128,7172,7275,7495,7560,7561,7616,7673,7803,7912,7914,7926,7997,8251,8298,8301,8341,8399,8414,8434,8452,8479,8482,8493,8620,8644,8719,8738,8837,8850,8859,9045,9244,9321,9338,9440,9550,9565,9574,9640,9809,9819,10174,10360,10378,10394,10465,10506,10517,10669,10840,11018,11114,11533,11615,11632,11737,11822,11880,11928,11946,12157,12179,12530,12548,12669,12989,13072,13206,13349,13407,13499,13548,13562,13612,13632,13646,13648,13662,13694,13700,13783,13876,13930,13970,13988,14039,14041,14084,14089,14108,14126,14132,14265,14297,14320,14442,14481,14519,14635,14729,14741,14789,14868,14901,14931,15056,15230,15278,15430,15539,15576,15652,15711,15771,15806,15899,15934,16000,16069,16142,16171,16240,16368,16385,16580,16590,16607,16632,16680,16711,16806,16871,16904,17011,17089,17102,17109,17257,17457,17523,17532,17613,17632,17761,18039,18121,18505,18593,18594,18724,18855,18873,18908,19064,19133,19137,19149,19171,19213,19368,19467,19481,19499,19558,19569,19594,19599,19813,19846,19909,19942,19952,20119,20205,20479,20596,20645,20671,20702,20820,20862,20877,20926,20965,20987,21216,21227,21238,21261,21266,21367,21378,21490,21512,21587,21611,21612,21653,21710,21785,21835,21850,21877,21969,21973,21997,22015,22055,22138,22207,22250,22282,22283,22389,22445,22453,22529,22675,22811,22971,23055,23064,23310,23336,23397,23458,23503,23576,23675,23682,23801,24001,24058,24094,24228,24495,24526,24527,24547,24608,24610,24714,24774,24795,24848,24957,25006,25092,25103,25177,25312,25494,25619,25730,25739,25791,25795,25929,26057,26078,26250,26289,26328,26340,26398,26487,26559,26662,26729,26739,26811,26818,26914,26941,27007,27013,27026,27121,27257,27386,27666,27684,27698,27762,27819,27851,27872,27881,27949,27995,28168,28208,28234,28331,28332,28425,28536,28562,28592,28608,28709,28757,28781,28948,28966,29028,29048,29081,29129,29178,29277,29380,29429,29530,29587,29769,29784,29850,29861,29867,29993,30274,30288,30304,30407,30512,30513,30633,30652,30726,30778,30840,30859,30920,30949,31106,31136,31176,31299,31483,31677,31757,31758,31761,31836,31922,31984,32212,32238,32335,32338,32412,32434,32544,32549,32599,32728,32921,32972,33051,33055,33060,33146,33179,33190,33310,33399,33437,33547,33561,33680,33708,33717,33755,33832,33932,34089,34092,34152,34276,34334,34470,34576,34600,34658,34815,34893,35033,35199,35208,35309,35795,36089,36148,36163,36203,36245,36433,36439,36630,36722,36795,36836,36925,36941,36971,37035,37084,37352,37416,37521,37523,37540,37546,37622,37694,37901,38025,38090,38146,38167,38220,38275,38297,38517,38572,38593,38692,38718,38729,38748,38878,39226,39231,39363,39447,39493,39537,39602,39630,39699,39748,39767,39780,39787,40283,40343,40390,40421,40464,40580,40642,40720,40809,40948,41185,41203,41221,41225,41292,41320,41363,41410,41601,41738,41746,41808,41923,41963,41990,41999,42150,42198,42276,42344,42387,42502,42573,42656,42674,42703,42874,43033,43188,43524,43532,43598,43616,43803,43996,44035,44040,44087,44396,44412,44561,44653,44752,44889,45042,45045,45071,45116,45144,45361,45428,45466,45526,45584,45723,45731,45850,45921,45995,46018,46019,46103,46217,46238,46276,46308,46337,46338,46391,46479,46559,46561,46679,46755,46776,46848,46862,46918,47002,47096,47194,47314,47397,47440,47512,47565,47604,47624,47631,47659,47720,47792,47823,47868,47924,47971,48005,48175,48321,48332,48381,48641,48758,48772,48794,48814,48964,48986,49005,49218,49411,49461,49510,49788,49823,49836,49916,49942,49968,50044,50056,50096,50285,50305,50379,50417,50451,50569,50641,50646,50686,50707,50711,50895,51022,51179,51204,51261,51510,51524,51529,51567,51579,51597,51667,51673,51712,51892,52089,52247,52298,52415,52713,52865,52883,52989,53121,53412,53440,53597,53610,53644,53646,53743,53850,54050,54152,54175,54327,54384,54569,54661,54700,54760,54800,54838,54924,55058,55199,55293,55308,55321,55400,55422,55534,55538,55624,55693,55710,55785,56043,56050,56205,56291,56328,56413,56448,56485,56578,56673,56685,56718,56822,56913,57039,57231,57367,57409,57422,57521,57611,57654,57659,57670,57680,57943,57965,58063,58168,58506,58512,58541,58552,58601,58734,58760,58763,58770,58973,59027,59116,59238,59243,59358,59468,59661,59717,59751,59830,59868,59920,59941,60028,60167,60275,60295,60300,60421,60469,60671,60676,60735,60763,60858,60941,60969,61018,61030,61076,61210,61327,61376,61405,61410,61541,61546,61784,61836,62016,62046,62148,62161,62324,62399,62649,62715,62725,62784,62815,62842,62856,63305,63477,63483,63486,63603,63696,63745,63840,63864,63872,64002,64108,64138,64139,64159,64179,64342,64514,64582,64597,64923,65043,65157,65216,65244,65270,65424,65721,65962,66032,66204,66343,66379,66444,66572,66666,66812,66848,67041,67052,67055,67138,67147,67184,67471,67489,67751,67818,67918,68061,68081,68117,68210,68240,68375,68443,68632,68695,68756,68847,68860,68871,68952,68960,69115,69723,69752,70217,70223,70250,70396,70494,70662,70679,70821,70984,71087,71122,71383,71427,71518,71634,71646,71655,71721,71826,72019,72040,72138,72148,72218,72220,72240,72378,72451,72574,72647,72680,72817,73025,73102,73127,73182,73198,73256,73327,73451,73460,73505,73566,73574,73592,73648,73738,73866,73932,73949,73964,73978,74075,74089,74126,74191,74210,74332,74541,74564,74597,74743,74765,74874,74912,75037,75058,75210,75363,75500,75509,75603,75851,75897,75959,76008,76190,76196,76226,76239,76287,76354,76644,76745,76753,76785,76831,76863,77089,77153,77233,77338,77469,77559,77619,77627,77660,77726,77770,77793,77869,77886,77966,78011,78105,78162,78198,78245,78320,78442,78443,78562,78695,78726,78837,78866,78879,78907,79005,79022,79037,79120,79151,79201,79240,79316,79390,79402,79582,79665,79893,79925,79967,80066,80249,80316,80351,80488,80614,80618,80651,80665,80703,80707,80775,80799,80981,81074,81151,81197,81314,81417,81546,81762,82184,82195,82215,82297,82361,82444,82532,82555,82774,82814,82853,83131,83164,83200,83230,83276,83288,83418,83480,83529,83599,83773,83861,83891,84028,84077,84167,84291,84343,84541,84602,84654,84705,84791,84885,84957,84989,85093,85296,85375,85391,85414,85516,85648,86085,86389,86410,86455,86540,86978,87016,87021,87057,87176,87231,87266,87294,87356,87429,87464,87855,87881,87940,87982,88046,88122,88191,88205,88211,88261,88278,88314,88375,88549,88674,88699,88722,88723,88732,88742,88755,88793,89021,89151,89328,89442,89463,89483,89504,89550,89651,89714,89872,89886,89889,90246,90356,90429,90482,90522,90624,90641,90994,91050,91145,91331,91658,91732,91865,91900,91953,92025,92062,92105,92113,92221,92226,92425,92432,92542,92647,92867,92869,92911,92980,93015,93073,93089,93236,93250,93302,93325,93341,93400,93575,93610,93701,93712,93804,93835,93905,93928,94158,94203,94208,94286,94322,94397,94400,94558,94877,94986,95029,95105,95206,95209,95258,95352,95377,95420,95453,95895,95904,96167,96291,96330,96432,96477,96520,96702,96720,96769,97146,97163,97165,97276,97312,97474,97481,97526,97625,97660,98172,98220,98236,98243,98262,98330,98364,98390,98430,98607,98931,98990,99494,99510,99602,99711,99775,99922,99954,99995",
          "19,274,284,348,419,452,455,466,485,506,551,594,614,723,773,797,838,915,992,1194,1379,1403,1535,5081,5181,5268,5275,5307,5309,5315,5319,5323,5364,5391,5402,5483,5580,5619,5792,5797,5849,5947,6124,6262,6495,6524,6542,6592,6601,6632,6698,6746,7002,7041,7083,7117,7341,7362,7423,7545,7575,7576,7604,7650,7724,8080,8149,8528,8590,8847,8968,8979,9251,9736,9921,9951,10089,10177,10527,10572,10627,10983,11019,11085,11102,12252,12551,13082,14129,14160,14173,14378,14435,15383,16227,16702,19276,19866,21456,24375,29644,30676,35173,41519,43341,50023,50179,50208,50217,50281,50454,50502,50643,50818,50872,51012,51160,51232,51276,51724,51749,51838,51887,51928,51952,52034,52097,52101,52154,52315,52363,52451,52661,52776,52787,52789,52917,52983,53018,53127,53191,53308,53489,53500,53624,53654,53700,53750,53801,53875,54148,54271,54290,54336,54522,54623,54641,54859,54897,54904,54990,55122,55167,55248,55300,55361,55404,55482,55834,55894,56179,56220,56236,56277,56339,56362,56669,56762,56884,56922,56964,57157,57347,57549,57593,57602,57793,57828,57855,57917,58131,58317,58498,58566,58592,58649,58715,58951,59031,59160,59276,59310,59338,59438,59496,59534,59649,59806,59896,59983,60101,60219,60375,60601,60702,60775,60836,61455,61621,61737,61827,61840,61868,61871,61960,62050,62084,62090,62105,62350,62366,62372,62416,62517,62611,62675,62733,62854,62887,63054,63062,63135,63415,63492,63556,63711,63746,63879,63956,64127,64147,64185,64279,64307,64350,64425,64471,64517,64588,64731,64836,64839,65032,65112,65164,65170,65204,65349,65357,65358,65449,65568,65653,65698,65926,65957,65974,65997,66238,66301,66488,66542,66569,66576,66678,66744,66760,66867,67019,67085,67203,67317,67342,67390,67462,67641,67689,67851,68058,68102,68126,68131,68234,68293,68385,68469,68550,68640,68645,68685,68740,68849,68972,68990,69173,69268,69332,69346,69436,69464,69771,69924,70014,70021,70115,70402,70404,70411,70584,70810,70838,70907,70913,70934,70978,71127,71144,71241,71285,71318,71440,71474,71593,71714,71805,71846,72199,72219,72288,72422,72431,72481,72495,72746,72839,72954,72968,73092,73249,73335,73677,73914,74066,74221,74286,74494,74594,75364,75443,75678,75974,76007,76177,76751,76776,76787,77061,77080,77101,77115,77257,77403,77413,77439,77518,77705,78049,78201,78267,78296,78297,78329,78461,78580,78588,78616,78701,78717,78760,78790,78832,78864,78961,78971,79001,79068,79091,79309,79339,79389,79403,79651,79661,79813,79871,79994,80064,80281,80282,80445,80560,80695,80839,81021,81352,81488,81548,81574,81730,81733,81761,81928,82004,82203,82257,82285,82314,82475,82642,82702,82717,82915,82945,83000,83022,83146,83273,83375,83449,83570,83621,83712,83813,83907,83962,84020,84081,84088,84152,84181,84196,84314,84325,84392,84712,84849,84943,84963,85005,85042,85243,85383,85465,85565,85573,85593,85630,85671,85788,85826,85919,85959,85982,86120,86225,86255,86292,86422,86427,86491,86610,86661,86692,86772,86953,87038,87117,87128,87134,87142,87225,87396,87499,87601,87622,87708,87971,88115,88147,88167,88230,88423,88517,88632,88680,88773,88877,88900,88932,88954,89248,89260,89538,89654,89679,90019,90110,90318,90527,90586,90610,90663,90949,90990,91192,91273,91307,91367,91429,91455,91491,91562,91648,91742,91851,91928,92048,92098,92100,92118,92193,92273,92328,92332,92445,92467,92531,92683,92707,92781,92818,92874,92927,92993,93078,93108,93277,93278,93310,93366,93389,93390,93632,93677,93729,93892,93922,94029,94181,94241,94602,94653,94824,94872,95050,95094,95434,95464,95503,95673,95678,95682,95850,96008,96014,96133,96134,96169,96190,96346,96463,96657,96665,96687,96839,96935,96941,96967,96977,97156,97349,97762,97790,97996,98002,98062,98129,98278,98394,98481,98492,98512,99127,99144,99176,99301,99311,99678,99717,99948",
          "40,99,111,120,127,132,134,210,223,244,583,659,877,879,980,1113,1400,1456,1493,1619,1807,1942,1954,2048,2094,2110,2189,2198,2226,2254,2339,2373,2428,2434,2640,2692,2725,2783,2804,2808,2821,2838,2931,2950,3043,3052,3080,3103,3191,3293,3295,3355,3368,3429,3487,3534,3616,3622,3650,3785,3806,3876,3975,3988,3998,4210,4231,4273,4322,4324,4361,4418,4430,4480,4492,4500,4510,4515,4576,4717,4736,4827,4857,4914,5011,5035,5095,5296,5317,5385,5484,5486,5639,5712,5840,5902,5998,6080,6102,6184,6249,6308,6535,6686,6713,6723,6780,6781,6988,7011,7058,7076,7272,7386,7481,7487,7582,7634,7765,7782,7858,7942,8202,8420,8472,8496,8498,8605,8653,8730,8817,9121,9249,9264,9311,9492,9541,9731,9901,10068,10251,10254,10395,10430,10460,10480,10558,10640,10695,10710,10780,10791,10991,11075,11094,11129,11139,11412,11520,11529,11565,11579,11826,12207,12308,12407,12447,12520,12594,12596,12666,12674,12688,12707,12736,12755,12832,12945,13035,13081,13117,13173,13223,13251,13307,13316,13346,13443,13519,13656,13961,13978,13996,14024,14040,14070,14110,14137,14200,14267,14269,14336,14402,14431,14462,14526,14528,14583,14600,14608,14685,14704,14758,14843,14871,14879,14966,15000,15062,15094,15122,15126,15141,15148,15155,15182,15208,15262,15264,15510,15604,15655,15659,15736,15856,16218,16237,16397,16402,16423,16499,16509,16529,16544,16769,16817,16995,17001,17041,17078,17168,17195,17266,17322,17329,17349,17385,17396,17401,17416,17470,17484,17537,17568,17603,17655,17666,17679,17764,17835,17943,17948,18051,18114,18170,18238,18239,18254,18263,18334,18355,18479,18527,18540,18688,18728,18750,18756,19095,19132,19140,19252,19287,19324,19374,19525,19640,19696,19768,19859,19980,20015,20053,20064,20117,20173,20253,20261,20396,20453,20490,20528,20611,20641,20729,20757,20762,20819,21039,21061,21263,21276,21302,21317,21361,21452,21459,21506,21551,21560,21597,21602,21638,21746,21824,21854,21938,21944,22031,22057,22090,22154,22204,22214,22264,22418,22434,22473,22526,22530,22679,22734,22758,22769,22903,23003,23005,23142,23178,23231,23236,23344,23367,23388,23471,23476,23483,23688,23730,23755,23819,23883,23887,23955,23957,24118,24121,24181,24236,24275,24288,24312,24335,24519,24542,24561,24742,24914,24921,24983,24996,25056,25067,25178,25192,25197,25228,25456,25478,25561,25572,25586,25766,25855,25881,25913,25974,25975,26030,26129,26157,26193,26215,26218,26219,26271,26317,26363,26450,26476,26597,26659,26692,26750,26753,26797,26942,26994,27010,27023,27108,27180,27374,27381,27499,27544,27575,27589,27594,27629,27701,27718,27761,27830,27873,27890,27892,27907,27929,28007,28130,28170,28180,28209,28230,28317,28339,28347,28459,28506,28586,28869,28953,29013,29031,29099,29132,29153,29162,29209,29289,29311,29339,29369,29389,29391,29407,29443,29475,29557,29594,29610,29668,29733,29745,29797,29894,29905,29939,29954,30020,30042,30059,30063,30067,30184,30235,30247,30262,30280,30324,30340,30473,30663,30665,30673,30689,30737,30956,30978,31000,31001,31018,31060,31097,31184,31237,31345,31386,31394,31445,31532,31549,31550,31569,31723,31732,31788,31798,31806,31822,31844,31866,32047,32124,32161,32176,32179,32231,32265,32274,32287,32296,32305,32308,32328,32357,32371,32373,32379,32444,32448,32475,32553,32587,32600,32618,32648,32679,32702,32715,32769,32831,32861,32894,32903,32922,32941,32952,32961,32964,32968,32992,33008,33117,33187,33191,33229,33232,33236,33329,33340,33350,33466,33498,33567,33577,33720,33768,33839,33867,33870,33916,33972,33984,34022,34251,34320,34344,34351,34353,34429,34457,34560,34626,34766,34814,34900,35006,35039,35049,35184,35188,35218,35271,35346,35391,35423,35428,35521,35531,35562,35595,35612,35621,35753,35833,35991,36081,36084,36192,36197,36204,36256,36266,36274,36283,36304,36344,36448,36467,36536,36543,36556,36569,36574,36612,36624,36651,36655,36711,36748,36764,36787,36791,36823,36851,36861,36877,36908,36910,37021,37060,37095,37101,37142,37185,37258,37275,37282,37296,37298,37299,37306,37341,37359,37415,37453,37637,37678,37704,37745,37764,37829,37887,37894,37981,37993,38004,38083,38136,38141,38181,38226,38259,38265,38266,38282,38283,38301,38433,38453,38506,38551,38590,38627,38640,38686,38697,38717,38742,38790,38823,38830,39020,39062,39076,39146,39186,39195,39218,39327,39414,39439,39461,39477,39485,39553,39595,39596,39624,39668,39744,39847,39894,40007,40040,40081,40089,40225,40286,40307,40309,40339,40358,40418,40427,40477,40506,40553,40574,40639,40774,40777,40783,40834,40930,41055,41132,41165,41208,41210,41277,41336,41381,41386,41433,41529,41664,41684,41731,41750,41762,41782,41785,41825,41842,41843,41850,41851,41922,42026,42038,42052,42065,42091,42156,42188,42232,42268,42306,42335,42362,42370,42377,42390,42435,42503,42510,42541,42605,42621,42704,42721,42752,42765,42798,42803,42823,42868,42921,42959,42993,42995,43001,43068,43071,43111,43113,43157,43226,43286,43294,43311,43378,43411,43420,43492,43500,43508,43520,43644,43664,43771,43814,43881,43926,43949,43950,43961,43995,44000,44029,44142,44169,44197,44257,44320,44344,44347,44387,44409,44462,44560,44567,44585,44666,44676,44707,44712,44737,44828,44842,44847,44877,44910,44922,44974,45030,45139,45141,45143,45189,45243,45282,45351,45400,45437,45485,45509,45604,45749,45800,45828,45831,45862,45869,45872,45906,45976,45992,46145,46201,46501,46526,46538,46730,46800,46832,46925,46961,46972,47013,47054,47081,47083,47085,47144,47224,47263,47406,47410,47424,47449,47457,47482,47527,47541,47548,47608,47622,47648,47695,47835,47842,47849,47974,48054,48263,48300,48324,48347,48360,48644,48700,48879,49030,49160,49375,49408,49512,49628,49659,49667,49685,49755,49858,49897,49936,50247,50272,50274,50345,50350,50439,50440,50452,50549,50559,50678,50695,50713,50722,50843,50909,50933,50940,51060,51072,51115,51177,51195,51228,51291,51306,51436,51497,51755,51901,51912,51921,52022,52062,52112,52122,52269,52487,52538,52584,52690,52692,52788,52872,52886,52895,52971,53061,53063,53075,53229,53278,53404,53409,53419,53468,53504,53537,53545,53839,53902,53939,54032,54033,54069,54130,54133,54141,54293,54347,54396,54458,54683,54712,54854,54969,55106,55163,55386,55616,55639,55824,55911,56121,56180,56280,56390,56531,56532,56660,56909,56980,56999,57019,57055,57259,57266,57364,57374,57434,57594,57607,57666,57711,57953,58018,58140,58145,58179,58189,58213,58299,58341,58430,58517,58639,58640,58670,58707,58716,58867,58904,59164,59227,59237,59297,59392,59421,59627,59687,59958,59985,60110,60211,60318,60327,60511,60514,60679,60742,60804,60816,60817,60992,61186,61403,61425,61557,61643,61785,61962,62028,62089,62097,62169,62246,62361,62549,62579,62581,62589,62619,62666,62752,62977,63031,63204,63309,63329,63404,63510,63543,63629,63941,64013,64031,64074,64234,64317,64397,64470,64587,64632,64684,64892,64971,65090,65189,65205,65245,65248,65283,65346,65376,65432,65563,65594,65677,65718,65792,65840,65842,65878,66008,66106,66144,66180,66209,66214,66298,66328,66331,66336,66472,66577,66801,66864,66948,67326,67364,67370,67513,67537,67550,67642,67760,67809,67845,68078,68093,68132,68135,68136,68149,68153,68244,68323,68345,68367,68668,68677,68817,68854,69003,69093,69222,69360,69513,69552,69571,69655,69884,69911,69965,69972,70010,70262,70279,70301,70316,70325,70590,70609,70740,70754,70816,70832,70842,70918,70941,70990,71132,71139,71150,71159,71210,71224,71422,71509,71553,71604,71606,71701,71782,71842,71919,72062,72117,72119,72172,72223,72310,72358,72411,72674,72683,72824,72867,72872,72899,72981,72990,73194,73222,73394,73410,73428,73452,73479,73490,73627,73673,73721,73727,73815,73816,73890,73963,73996,74174,74279,74292,74498,74504,74589,74599,74610,74668,74673,74731,74847,74911,74997,75073,75195,75218,75272,75285,75294,75394,75402,75440,75649,75828,75870,75889,75900,75946,75953,76009,76240,76257,76267,76285,76301,76357,76372,76465,76570,76571,76732,76765,76773,76780,76839,76866,77005,77081,77156,77189,77211,77227,77235,77358,77408,77570,77749,77757,77762,77829,77876,77929,77939,77990,78095,78109,78173,78209,78260,78357,78398,78514,78541,78618,78651,78960,78994,79081,79112,79262,79280,79297,79317,79337,79564,79694,79715,79792,79819,79892,79899,79921,80266,80406,80457,80472,80475,80522,80537,80549,80550,80569,80621,80701,80749,80944,81135,81320,81439,81651,81678,81695,81754,81818,81855,81896,81897,81946,81953,82144,82201,82207,82210,82299,82322,82337,82434,82471,82476,82546,82559,82662,82761,82795,82920,82923,82925,82927,82987,83031,83309,83381,83402,83447,83583,83585,83791,83816,83848,83975,84038,84047,84057,84125,84162,84233,84383,84391,84488,84539,84615,84748,85066,85073,85137,85199,85225,85234,85236,85267,85408,85444,85491,85584,85597,85719,85824,85944,85984,86010,86100,86115,86159,86284,86355,86370,86413,86557,86562,86645,86656,86739,86761,86981,86999,87028,87035,87077,87212,87246,87355,87432,87523,87560,87565,87683,87704,87715,87818,88010,88067,88169,88198,88208,88364,88481,88499,88527,88533,88686,88750,88758,88772,88775,88896,88906,88974,89034,89116,89237,89482,89496,89542,89686,89762,89801,89814,89887,89994,90233,90235,90360,90384,90386,90397,90436,90463,90617,90694,90894,90918,90970,91010,91162,91204,91310,91327,91338,91350,91361,91421,91559,91687,91788,91817,91828,91846,91893,92075,92164,92172,92178,92392,92457,92494,92569,92894,92943,93061,93201,93249,93349,93363,93647,93761,93870,93896,93901,93916,94037,94046,94163,94188,94201,94272,94274,94381,94418,94489,94561,94656,94737,94771,94880,95000,95081,95099,95101,95133,95309,95310,95331,95356,95385,95438,95667,95729,95732,95812,95834,95987,96130,96145,96170,96226,96233,96274,96288,96316,96337,96459,96481,96595,96601,96634,96646,96985,97101,97147,97238,97339,97461,97529,97563,97585,97617,97730,97761,97910,97945,98013,98221,98301,98890,98944,99011,99018,99090,99124,99218,99348,99349,99463,99474,99599,99900",
          "34,104,231,283,353,495,562,580,604,686,696,962,1058,1064,1119,1267,1288,1335,1372,1406,1465,1667,1763,1787,1856,1915,2004,2082,2161,2423,2472,2626,2709,2862,2889,2898,2901,2921,2941,3058,3116,3137,3151,3165,3172,3241,3358,3403,3406,3549,3638,3857,3914,4008,4032,4037,4182,4279,4334,4339,4388,4423,4863,4877,4882,5101,5184,5192,5510,5525,5606,5608,5681,5696,5716,5724,5861,5944,5970,6395,6400,6540,7009,7087,7118,7202,7229,7278,7439,7538,7619,7632,7704,7727,7728,7776,7885,7919,7976,7989,8072,8081,8189,8474,8565,8621,8637,8654,8685,8716,8721,8828,8853,8916,9071,9158,9382,9415,9581,9658,9741,9817,9818,9841,9902,9945,9955,10148,10358,10366,10564,10582,10596,10694,10726,10735,10859,10863,10921,10946,10981,11095,11228,11263,11269,11272,11290,11376,11386,11494,11611,11626,11748,11875,11905,11982,12057,12449,12479,12637,12678,12732,12750,12790,12791,12826,13020,13046,13079,13177,13211,13214,13234,13344,13482,13543,13620,13647,13668,13684,13757,13770,13833,13837,13845,13890,13911,14068,14076,14316,14348,14370,14397,14454,14468,14505,14565,14588,14611,14630,14652,14770,14775,14831,14915,14922,14988,15080,15088,15132,15426,15629,15709,15902,15910,15923,16053,16075,16106,16223,16233,16380,16382,16525,16649,16675,16692,16697,16765,16819,16864,16948,16983,17002,17038,17108,17116,17226,17236,17333,17522,17619,17743,17886,17929,17934,17996,18028,18097,18117,18167,18174,18178,18219,18262,18274,18356,18408,18414,18569,18632,18739,18740,18826,18914,18977,19033,19056,19111,19173,19328,19397,19486,19615,19627,19670,19692,19738,19762,19785,19804,19810,19837,19976,19997,20007,20087,20153,20155,20156,20197,20215,20311,20385,20408,20686,20717,21110,21114,21125,21186,21201,21231,21285,21315,21402,21410,21420,21449,21485,21676,21684,21794,21806,21814,21890,21910,21928,22096,22130,22181,22195,22230,22268,22370,22385,22390,22394,22523,22557,22583,22609,22781,22801,22960,23042,23082,23132,23263,23321,23329,23333,23365,23386,23419,23426,23511,23518,23540,23555,23556,23568,23624,23750,23902,23952,24091,24113,24328,24453,24613,24648,24650,24652,24747,24840,24875,24885,24890,24952,24962,24973,25031,25035,25075,25169,25218,25298,25462,25481,25493,25509,25548,25550,25707,25746,25793,25800,25836,25858,25966,26005,26073,26108,26130,26138,26175,26178,26183,26195,26255,26292,26355,26462,26580,26688,26800,26804,26928,27017,27103,27141,27144,27311,27354,27514,27537,27603,27657,27705,27774,27775,27827,27852,27922,27986,28037,28076,28086,28088,28144,28151,28190,28225,28309,28371,28457,28474,28544,28554,28596,28623,28806,28826,29022,29026,29047,29061,29116,29217,29302,29329,29449,29482,29515,29523,29560,29565,29614,29802,29808,29940,30007,30016,30122,30174,30239,30267,30409,30518,30601,30611,30613,30658,30708,30799,30809,30841,30864,30879,30942,30998,31057,31123,31202,31255,31275,31323,31342,31417,31516,31620,31673,31686,31703,31846,31849,31876,31891,31904,31961,32006,32031,32034,32062,32080,32148,32226,32517,32657,32763,32849,32901,32925,33062,33420,33497,33532,33593,33637,33650,33798,33815,34003,34031,34258,34265,34267,34296,34332,34388,34446,34510,34541,34580,34582,34686,34692,34872,34933,35079,35464,35592,35630,35745,35746,35781,35806,35864,35865,36055,36062,36181,36492,36521,36609,36774,36827,36864,37025,37147,37430,37625,37627,37705,38127,38294,38316,38354,38552,38578,38660,38684,38685,38792,38795,38836,38895,38998,39017,39099,39271,39392,39436,39440,39450,39645,39704,39718,39747,39927,40074,40106,40344,40355,40576,40624,40652,40689,40755,40790,40823,40831,40868,40925,41052,41086,41131,41145,41152,41370,41371,41427,41444,41485,41517,41542,41548,41633,41662,41803,41828,41957,42086,42100,42207,42216,42251,42352,42434,42518,42678,42788,42848,42871,42947,43067,43369,43400,43429,43528,43552,43746,43787,43791,43863,44110,44202,44371,44713,44751,45095,45149,45157,45190,45267,45274,45280,45600,45852,45866,45916,45936,46256,46350,46379,46676,46712,46759,46767,46968,47042,47057,47172,47223,47231,47255,47278,47280,47493,47575,47643,47677,47678,47686,47691,47755,47768,47818,47915,47939,48012,48018,48226,48291,48310,48350,48351,48354,48418,48538,48620,48822,48839,48941,49017,49088,49133,49373,49378,49477,49550,49563,49566,49705,49779,49783,49795,49899,49909,49997,50030,50053,50149,50171,50182,50258,50400,50413,50420,50532,50634,50736,50742,50774,50866,50926,51019,51394,51565,51681,51756,51821,51824,52064,52072,52087,52288,52328,52347,52366,52374,52417,52456,52496,52497,52566,52578,52585,52725,52863,52896,53226,53331,53362,53388,53445,53492,53499,53608,53611,53676,53760,53852,53868,53935,54029,54192,54207,54487,54497,54505,54511,54663,54679,54690,54691,54842,54936,54944,55005,55043,55130,55236,55502,55571,55587,55731,55960,55963,55975,56092,56112,56186,56308,56417,56539,56559,56587,56712,56713,56821,56915,57066,57113,57136,57166,57193,57196,57293,57326,57344,57437,57494,57575,57624,57689,57694,57703,57736,57882,57888,58057,58076,58139,58220,58238,58310,58357,58576,58752,58772,58833,58851,58925,59097,59103,59139,59168,59206,59293,59337,59418,59487,59511,59616,59670,59705,59888,59916,59931,59946,60010,60170,60190,60222,60287,60323,60344,60376,60490,60585,60634,60818,60889,61109,61251,61281,61309,61382,61636,61821,61918,62072,62282,62290,62331,62358,62507,62708,62767,62776,62825,62914,62932,62971,63133,63162,63167,63217,63230,63239,63252,63300,63364,63410,63447,63638,63675,63758,63766,63966,64065,64290,64404,64435,64536,64687,64746,64761,64803,65138,65156,65181,65191,65210,65299,65311,65404,65464,65486,65489,65504,65530,65646,65822,65845,65859,65945,65990,66005,66050,66073,66193,66199,66252,66297,66441,66442,66647,66730,66847,66900,67137,67421,67694,67805,67821,67823,67850,68030,68056,68177,68193,68249,68300,68304,68377,68653,68716,68759,68859,69004,69069,69237,69442,69482,69690,69734,69748,69774,69877,69896,69964,70038,70055,70127,70170,70313,70346,70400,70459,70532,70634,70661,70677,71088,71290,71317,71348,71524,71581,71713,71979,72045,72169,72179,72197,72230,72501,72605,72640,72709,72727,72751,72831,72876,72945,72997,73039,73054,73093,73258,73315,73552,73689,73723,73728,73846,73916,73952,73954,74094,74099,74251,74361,74370,74430,74524,74618,74736,74768,74927,74958,74962,75039,75249,75277,75400,75417,75419,75445,75743,75820,75881,75895,76042,76088,76119,76131,76289,76337,76577,76687,76706,76810,76852,76919,76927,76979,76989,77111,77208,77268,77379,77380,77449,77499,77560,77576,77612,77631,77711,77800,77811,77852,77988,78098,78104,78316,78375,78379,78416,78427,78455,78537,78567,78680,78684,78791,78803,78823,78948,79168,79195,79236,79281,79334,79495,79567,79633,79669,79674,79795,79839,79848,79874,79885,79944,80256,80402,80458,80595,80648,80652,80723,80774,80910,80924,80982,81017,81124,81145,81217,81225,81391,81490,81509,81618,81692,81718,81786,81856,81867,81892,82039,82286,82429,82556,82588,82847,82848,82897,82906,83047,83083,83154,83242,83442,83607,83672,83751,83797,84241,84378,84388,84393,84438,84607,84623,84691,84716,84825,84842,84887,84935,85085,85146,85195,85432,85635,85686,85707,85731,85775,85969,86105,86177,86378,86500,86514,86552,86616,86620,86673,86708,86903,86997,87033,87084,87093,87137,87241,87283,87457,87568,87607,87610,87668,87675,87764,87782,87785,87886,88014,88062,88099,88166,88189,88192,88196,88397,88448,88463,88547,88664,88704,88720,88753,88805,88967,89066,89092,89115,89201,89207,89264,89292,89303,89337,89399,89425,89521,89750,89898,90183,90241,90313,90405,90472,90729,90804,90822,90969,91046,91112,91409,91507,91580,91608,91727,91797,91881,91896,91989,92182,92246,92286,92316,92419,92430,92495,92535,92622,92628,92644,92672,92839,92968,92983,93003,93062,93196,93246,93301,93358,93386,93491,93597,93696,93710,93748,93781,93913,93915,93961,94025,94075,94090,94126,94230,94235,94245,94257,94445,94459,94494,94497,94744,94974,95007,95010,95145,95251,95347,95378,95557,95626,95688,95741,95783,95824,95862,95867,95931,95949,96049,96117,96194,96212,96338,96402,96475,96495,96508,96537,96582,96598,96630,96688,96759,96933,96961,96980,97078,97256,97399,97471,97605,97778,97883,97935,97941,97995,98017,98028,98136,98242,98257,98293,98356,98384,98460,98486,98489,98510,98718,98730,98763,98908,98936,98989,99229,99403,99429,99627,99635,99640,99740,99769,99807,99965",
          "50,52,201,417,530,579,605,703,787,846,883,894,1022,1042,1092,1165,1224,1281,1299,1385,1420,1459,1507,1530,1551,1719,1916,1923,1984,2007,2012,2199,2211,2314,2379,2509,2762,2773,2794,2872,2944,3022,3290,3409,3449,3507,3601,3626,3647,3660,3783,3843,3847,3853,3946,3997,4070,4173,4333,4344,4382,4491,4588,4601,4639,4674,4678,4680,4856,5022,5038,5128,5203,5241,5324,5325,5420,5423,5424,5425,5504,5672,5687,5832,5943,6302,6309,6314,6346,6562,6683,6724,6753,6767,6882,7051,7052,7065,7340,7344,7587,7701,7877,8103,8130,8322,8337,8485,8585,8596,8629,8770,8779,8922,8945,9041,9084,9150,9290,9301,9327,9342,9426,9428,9447,9455,9488,9616,9923,9929,9936,9984,10107,10203,10237,10263,10396,10466,10505,10549,10563,10609,10684,10690,10704,10821,10879,10883,10915,11343,11496,11502,11532,11620,11724,11773,11863,11900,11937,11941,12000,12159,12250,12300,12424,12587,12620,12843,13012,13118,13227,13257,13374,13406,13570,13643,13657,13730,13969,13983,13987,14081,14217,14259,14321,14367,14422,14428,14515,14575,14619,14656,14816,14880,14902,14905,14930,14957,14975,14989,15001,15004,15015,15021,15220,15334,15422,15486,15584,15631,15769,15778,15785,15791,15849,15884,15892,16074,16219,16241,16251,16305,16534,16755,16756,16794,16923,16926,16977,17034,17045,17163,17208,17440,17593,17641,17786,17803,17906,17908,17928,17980,18004,18111,18280,18538,18616,18682,18766,19236,19253,19283,19421,19472,19593,19812,19828,20095,20185,20291,20372,20443,20468,20507,20669,20697,20768,20890,20990,21174,21217,21264,21291,21307,21348,21349,21424,21489,21589,21608,21777,21851,21905,21991,22108,22206,22215,22254,22459,22525,22740,22840,22853,22860,22870,23037,23250,23487,23516,23558,23567,23588,23638,23906,23990,24055,24190,24234,24297,24301,24365,24498,24530,24582,24645,24735,24809,24837,24909,25034,25122,25129,25148,25164,25262,25709,25978,26016,26511,26514,26538,26583,26603,26631,26643,26878,27009,27130,27188,27198,27239,27445,27644,27729,27882,27900,28087,28219,28260,28291,28417,28529,28643,28870,28981,29004,29014,29399,29451,29453,29500,29506,29814,29832,29835,29844,29855,29896,30250,30327,30376,30537,30569,30678,30725,30850,30906,31063,31144,31165,31297,31311,31463,31481,31504,31507,31587,31981,32004,32097,32142,32194,32302,32316,32343,32370,32498,32559,32655,32666,32825,32989,33041,33128,33402,33460,33465,33544,33572,33612,33760,33875,33943,33987,34138,34166,34179,34198,34362,34462,34500,34615,34738,34782,34956,34970,35275,35316,35354,35514,35558,35583,35666,35983,36028,36154,36177,36320,36479,36510,36628,36736,36902,37078,37238,37254,37746,37760,37784,37853,37890,37904,37920,38035,38199,38284,38287,38353,38501,38533,38611,38661,38680,38813,39090,39180,39232,39281,39359,39380,39464,39474,39506,39641,39722,39758,39797,39800,39904,39933,39979,39996,40016,40097,40140,40193,40306,40436,40451,40554,40682,40727,40747,40749,40765,40769,40842,41041,41146,41295,41311,41359,41376,41472,41478,41492,41659,41980,41983,41987,42058,42062,42097,42321,42439,42643,42659,42794,42938,43117,43129,43366,43394,43467,43606,43615,43768,43776,43785,43839,43900,44017,44052,44145,44272,44318,44375,44435,44494,44550,44706,44719,44858,44977,45051,45082,45200,45206,45269,45359,45367,45409,45433,45512,45633,45686,45820,45823,45851,45989,46164,46182,46287,46294,46820,46861,46863,46880,47021,47037,47045,47068,47250,47318,47605,47773,47870,47876,48031,48041,48071,48083,48185,48247,48296,48407,48421,48517,48533,48550,48607,48659,48698,48709,48761,48833,48869,49130,49134,49174,49254,49294,49303,49401,49410,49448,49492,49776,49809,49813,49893,49923,50104,50124,50168,50251,50602,50649,50650,50753,50821,50857,50875,50880,50931,51040,51361,51390,51395,51463,51492,51593,51677,51719,51738,51858,51863,51914,52068,52076,52091,52146,52277,52311,52348,52402,52528,52533,52567,52613,52621,52653,52739,52911,53005,53318,53342,53360,53385,53387,53433,53528,53609,53678,53722,53781,53784,53838,54193,54479,54563,54589,54767,54832,54849,55020,55120,55142,55289,55352,55380,55459,55506,55544,55546,55609,55643,55673,55803,55923,55938,55976,56062,56072,56118,56189,56468,56486,56521,56675,56684,56688,56780,56806,56926,56996,57099,57208,57237,57311,57616,57634,57660,57668,57687,57708,57801,57941,57946,58045,58120,58650,58680,58799,58868,58892,58908,58972,59114,59115,59167,59282,59287,59328,59333,59364,59507,59560,59681,59690,59701,59782,60151,60172,60257,60283,60404,60608,60790,60967,61008,61058,61077,61153,61494,61540,61556,61576,61657,61679,61687,61688,61890,62062,62141,62172,62257,62385,62503,62513,62552,62686,62749,62769,62938,62958,62972,62989,63014,63027,63179,63338,63382,63653,63979,64192,64344,64481,64623,65223,65312,65343,65441,65497,65499,65587,65601,65716,65793,65848,66208,66284,66353,66418,66657,66693,66939,67057,67091,67195,67296,67305,67344,67350,67661,67878,67895,67972,68085,68115,68204,68319,68383,68412,68462,68495,68532,68613,68727,68796,68875,68933,68953,69045,69046,69050,69100,69285,69498,69556,69700,69715,69852,69889,69905,69927,70357,70435,70467,70490,70508,70519,70558,70619,70833,70938,70948,70964,71093,71116,71126,71188,71228,71299,71332,71355,71544,71633,71747,72014,72085,72110,72137,72176,72500,72572,72606,72679,72787,72854,72965,73077,73107,73296,73357,73417,73462,73501,73609,73680,73696,73699,73859,73888,74079,74159,74181,74393,74434,74651,74698,74725,74770,74897,74983,75078,75197,75212,75223,75310,75357,75377,75381,75521,75644,75688,75699,75725,75759,75908,76215,76233,76318,76322,76469,76662,76735,76792,77077,77134,77229,77231,77291,77836,77894,77909,77910,77996,77999,78334,78486,78581,78634,78677,78951,78966,79184,79314,79331,79354,79367,79388,79429,79438,79462,79503,79657,79695,79796,79838,79993,80059,80167,80200,80280,80436,80508,80536,80572,80765,80923,80939,81020,81035,81138,81161,81164,81249,81276,81461,81531,81591,81670,81826,81865,81915,81996,82085,82140,82226,82389,82428,82433,82540,82619,82832,82955,83013,83045,83101,83125,83134,83214,83331,83342,83512,83567,83674,83693,83771,83833,83841,83859,84058,84166,84174,84175,84177,84194,84255,84487,84500,84518,84578,84618,84728,84822,84911,84977,85136,85187,85211,85274,85361,85394,85400,85519,85567,85810,85820,85858,85964,86054,86061,86287,86342,86494,86705,86776,86786,86812,86853,86955,86961,87118,87312,87574,87623,87763,87893,87996,88059,88110,88173,88249,88271,88382,88519,88660,88661,88792,88820,88845,89012,89086,89325,89364,89382,89467,89597,89662,89713,89923,89947,90006,90153,90180,90245,90323,90383,90541,90551,90569,90595,90645,90707,90721,90722,90753,90852,90857,90967,90982,91022,91040,91094,91119,91177,91260,91263,91380,91476,91566,91689,91908,91947,92072,92537,92562,92784,92804,92881,92941,92959,93203,93377,93623,93764,94030,94043,94071,94108,94220,94704,94804,94818,95212,95281,95440,95664,95683,95695,95705,95718,95724,95750,95820,95828,95992,96054,96084,96234,96239,96294,96343,96386,96563,96669,96730,96920,96923,96936,97031,97043,97048,97068,97187,97255,97476,97575,97823,97880,97931,97994,98058,98149,98399,98568,98588,98592,98645,98713,98754,98829,98847,99050,99057,99108,99195,99199,99285,99318,99321,99333,99353,99373,99468,99549,99614,99615,99616,99675,99712,99857,99885,99967,99986",
          "70,82,239,262,561,673,898,1014,1091,1131,1142,1154,1169,1220,1422,1455,1592,1655,1774,1990,2056,2057,2214,2223,2406,2412,2610,2624,2668,2703,2748,2972,3036,3087,3089,3126,3228,3250,3301,3320,3373,3479,3521,3561,3565,3652,3768,3851,3909,3984,4052,4090,4243,4312,4336,4337,4350,4384,4531,4542,4638,4654,4663,4718,4810,4868,4883,5005,5088,5090,5225,5379,5553,5660,5709,5753,5798,5839,5842,5863,6227,6245,6428,6534,6590,6620,6697,6838,6928,7067,7253,7301,7428,7512,7520,7539,7612,7655,7696,7750,7763,7773,7795,8108,8161,8248,8283,8348,8400,8574,8775,8794,8804,8860,8875,8963,8969,9060,9162,9305,9377,9461,9537,9597,9703,9752,9762,9844,9850,9995,10166,10218,10306,10435,10463,10476,10579,10754,10862,10995,11035,11081,11186,11226,11247,11295,11338,11346,11360,11604,11634,11944,12099,12219,12241,12248,12355,12375,12401,12439,12560,12704,12735,13021,13054,13077,13144,13382,13383,13431,13454,13575,13688,13914,13948,13980,14018,14090,14228,14286,14302,14414,14440,14469,14554,14665,14755,14846,14938,15147,15181,15243,15255,15378,15569,15780,15862,15864,15928,16028,16094,16168,16181,16183,16242,16250,16338,16427,16430,16491,16575,16624,16789,16867,16911,16974,17098,17152,17159,17304,17342,17485,17605,17608,17685,17735,17749,17845,17952,18055,18058,18063,18153,18319,18418,18514,18558,18676,19022,19074,19106,19158,19241,19321,19373,19527,19562,19635,19719,19820,19910,20074,20194,20208,20300,20387,20455,20644,20680,20721,20745,20883,20884,20915,20985,21010,21059,21091,21191,21245,21277,21330,21379,21381,21604,21779,21948,22042,22113,22176,22248,22258,22261,22296,22359,22805,22818,23068,23217,23239,23420,23424,23489,23689,23747,23807,23835,23851,23917,23922,23984,24014,24115,24130,24380,24427,24637,24716,24729,24755,24810,24818,24883,25024,25028,25060,25247,25300,25306,25567,25738,25755,26046,26110,26174,26189,26275,26311,26360,26394,26395,26668,26675,26709,26872,26939,27089,27095,27129,27330,27548,27554,27599,27673,27797,27953,27954,28025,28035,28055,28143,28222,28232,28361,28389,28483,28522,28553,28615,28674,28693,28752,28884,28916,28951,28961,28970,29257,29365,29540,29597,29630,29744,29890,29898,29908,30075,30104,30126,30139,30441,30585,30587,30700,30789,30983,30997,31039,31082,31094,31232,31270,31336,31339,31368,31949,31958,31978,32010,32018,32037,32075,32093,32159,32273,32377,32512,32606,32724,32727,32786,32855,33073,33074,33105,33171,33175,33478,33502,33639,33712,33746,33808,33853,33866,34028,34043,34280,34343,34571,34590,34664,34695,34723,34886,34890,34909,35174,35232,35289,35355,35398,35418,35596,35620,36161,36323,36444,36469,36493,36638,36689,36929,37114,37234,37255,37340,37665,37696,37741,37950,38002,38028,38179,38207,38300,38350,38482,38568,38571,38618,38743,38807,38815,38845,38884,38924,38953,38965,38990,39018,39027,39030,39056,39087,39213,39353,39444,39472,39543,39556,39789,39828,39885,40079,40114,40195,40254,40266,40321,40462,40542,40646,40691,40744,40913,41033,41034,41073,41137,41230,41360,41694,41838,42089,42168,42284,42341,42412,42650,42749,42918,43003,43024,43138,43249,43379,43384,43493,43625,43767,43934,44162,44341,44383,44486,44493,44652,44787,44802,44822,44871,44897,44901,45027,45186,45192,45366,45478,45550,45568,45585,45798,45893,45901,45997,46025,46149,46159,46195,46320,46352,46384,46446,46460,46541,46580,46590,46591,46628,46747,46754,46829,46865,46903,46914,46952,46990,47124,47256,47386,47403,47453,47466,47477,47499,47580,47611,47703,47850,48001,48047,48062,48073,48094,48148,48160,48319,48431,48460,48505,48573,48595,48657,48753,48770,48815,49123,49143,49148,49179,49267,49281,49489,49524,49543,49580,49607,49686,49689,49703,49713,49827,49852,49885,49998,50038,50082,50223,50241,50517,50539,50726,50757,51132,51226,51279,51456,51522,51576,51643,51666,51782,51924,51947,52028,52067,52125,52227,52242,52259,52290,52299,52437,52565,52580,52926,53029,53040,53205,53242,53266,53421,53573,53598,53663,53751,53763,53864,53938,54086,54139,54306,54541,54621,54794,54911,54926,55084,55221,55249,55350,55381,55393,55399,55621,55721,55891,56014,56035,56146,56156,56193,56253,56306,56554,56579,56744,56746,56823,56910,56984,57012,57130,57146,57218,57233,57324,57371,57433,57489,57617,57671,57682,57751,57820,57884,58186,58221,58325,58570,58597,58615,58665,58815,58832,58913,59219,59244,59247,59262,59346,59365,59373,59420,59426,59430,59471,59619,59660,59694,59902,59996,60301,60306,60352,60451,60456,60513,60582,60614,60635,60870,61059,61072,61157,61217,61276,61301,61304,61336,61340,61408,61415,61481,61520,61579,61602,61682,61718,61943,61966,62111,62253,62415,62473,62662,62678,62810,62823,62906,62937,62978,63029,63077,63175,63187,63225,63321,63466,63642,63684,63693,63761,64043,64196,64223,64329,64476,64532,64533,64541,64571,64630,64775,64838,64890,65028,65192,65258,65344,65345,65475,65561,65598,65671,65778,65903,65920,65954,65975,66002,66054,66166,66333,66364,66427,66503,66556,66595,66671,66704,67230,67322,67469,67490,67683,67711,67753,67812,67831,67833,67849,67887,67948,68161,68228,68255,68418,68429,68558,68589,68812,68814,68822,68918,68929,68946,68965,69248,69257,69262,69272,69480,69590,69761,69843,69963,69971,70283,70401,70500,70517,70563,70641,70891,70920,71094,71263,71282,71347,71374,71419,71488,71508,71652,71719,71723,71734,71759,71761,71783,71811,71966,72048,72245,72287,72353,72472,72508,72587,72612,72717,72804,72871,72885,72901,72921,72931,72944,72985,73008,73019,73064,73297,73424,73440,73525,73594,73611,73614,73685,73994,74002,74035,74053,74142,74173,74179,74392,74403,74617,74655,74660,74803,75032,75052,75387,75396,75438,75566,75988,76017,76146,76160,76198,76231,76273,76380,76411,76479,76508,76586,76614,76633,76672,76705,76715,76826,76897,76912,76915,77337,77372,77393,77417,77443,77474,77524,77526,77542,77636,77826,77835,77901,77903,77953,77971,78001,78006,78018,78196,78220,78236,78262,78441,78578,78617,78673,78705,78770,78809,78914,78996,79148,79160,79274,79285,79347,79377,79482,79526,79559,79650,79691,79725,79761,79847,79849,79887,79903,79914,79965,79988,80135,80142,80216,80433,80434,80474,80493,80512,80589,80656,80741,80906,80988,81228,81374,81430,81480,81736,81801,81812,81837,81850,82101,82198,82255,82256,82344,82381,82647,82672,82799,82866,82900,82948,82951,83077,83091,83244,83255,83256,83328,83462,83818,84123,84146,84198,84261,84398,84553,84569,84581,84592,84611,84656,84688,84706,84750,84905,84981,85099,85111,85115,85141,85144,85170,85250,85393,85441,85489,85499,85501,85552,85594,85629,85696,85699,85724,85774,85899,85931,85955,85976,86102,86134,86176,86209,86406,86450,86459,86466,86594,86667,86723,86766,86798,87055,87114,87196,87268,87440,87495,87557,87586,87671,87678,87820,87931,87934,88152,88180,88342,88403,88413,88493,88566,88572,88611,88692,88749,88964,88993,89096,89137,89165,89263,89317,89321,89396,89430,89532,89562,89664,89734,89764,89970,90196,90290,90291,90445,90486,90791,90801,90885,90953,91026,91108,91235,91236,91280,91714,91768,91780,91785,91905,91936,92152,92163,92216,92274,92302,92451,92546,92599,92768,92785,92827,92930,92951,93007,93218,93274,93306,93453,93465,93516,93552,93592,93787,93814,93962,94022,94085,94202,94218,94255,94260,94572,94587,94599,94631,94716,94788,94797,94854,94983,95006,95013,95221,95229,95263,95268,95307,95463,95610,95672,95797,95844,95855,95897,96075,96273,96371,96398,96445,96467,96546,96552,96752,96858,96950,97086,97141,97162,97174,97198,97205,97210,97225,97250,97384,97432,97694,97792,97836,97961,98035,98093,98205,98237,98415,98483,98487,98543,98585,98733,98740,98855,98881,98916,98988,99134,99174,99190,99552,99698,99783,99890,99989",
          "22,178,180,534,607,662,812,816,858,926,1047,1051,1062,1087,1242,1247,1347,1353,1499,1511,1527,1538,1598,1640,1674,1679,1773,1831,1886,1901,1921,1982,1986,2126,2148,2224,2320,2448,2518,2535,2539,2657,2694,2705,2723,2803,2853,2860,2956,3024,3055,3057,3065,3067,3181,3188,3237,3240,3253,3379,3518,3706,3737,3781,3897,3928,4038,4156,4218,4314,4372,4420,4562,4684,4942,4970,4979,5054,5061,5062,5072,5094,5624,5675,5728,5898,6045,6065,6169,6215,6272,6324,6380,6537,6621,6626,6634,6674,6687,6720,6737,6822,6899,6959,6976,7069,7208,7223,7225,7409,7544,7594,7730,7792,7799,7985,8017,8058,8093,8094,8124,8134,8329,8336,8404,8505,8548,8555,8580,8593,8677,8793,8988,9058,9099,9196,9197,9205,9210,9463,9473,9527,9544,9601,9764,9886,9912,9918,10062,10104,10111,10195,10231,10328,10420,10495,10573,10607,10652,10699,10729,10797,10815,10854,10990,10997,11157,11237,11522,11745,11751,11763,11908,11934,12045,12050,12176,12325,12331,12516,12529,12565,12664,12714,12721,12727,12757,12827,12862,12939,12972,13143,13146,13178,13309,13313,13414,13506,13526,13554,13584,13709,13724,13731,13789,13862,13883,13893,13989,14035,14092,14094,14179,14260,14413,14448,14674,14774,14787,14798,14827,14878,14911,14971,14981,15006,15019,15098,15112,15161,15188,15225,15229,15310,15391,15417,15483,15520,15687,15708,15713,15739,15868,15889,15895,15940,15974,15994,16016,16206,16288,16310,16331,16482,16532,16547,16576,16611,16828,16869,16910,16982,17069,17072,17137,17359,17412,17439,17596,17662,17797,17911,17915,18024,18067,18083,18086,18102,18140,18259,18261,18299,18343,18532,18541,18636,18778,18829,18847,18884,18887,18992,18995,19058,19067,19116,19145,19153,19192,19280,19366,19432,19529,19570,19585,19672,19815,20073,20086,20280,20286,20341,20395,20397,20504,20578,20901,21283,21360,21470,21474,21567,21632,21657,21853,22058,22068,22078,22085,22087,22471,22553,22576,22588,22874,22947,22981,22982,23041,23101,23105,23122,23150,23198,23201,23235,23241,23243,23326,23341,23363,23500,23522,23570,23584,23614,23616,23618,23685,23812,23879,23934,24064,24260,24386,24397,24414,24423,24477,24584,24623,24697,24798,24839,24928,25047,25118,25143,25196,25243,25246,25279,25302,25376,25426,25464,25536,25606,25828,25964,26029,26080,26148,26161,26200,26241,26252,26500,26550,26658,26661,26736,26829,26946,26951,26979,27085,27135,27271,27478,27528,27621,27682,27704,27752,27785,27855,27926,27971,27972,27975,27992,28015,28056,28252,28330,28407,28501,28524,28531,28552,28570,28618,28636,28785,28846,29224,29308,29330,29346,29364,29416,29427,29601,29641,29742,29817,29820,29837,29838,29843,29913,30072,30080,30083,30085,30121,30178,30354,30559,30705,30867,30892,31032,31050,31092,31328,31439,31540,31567,31610,31716,31870,31910,31975,32024,32025,32313,32323,32402,32483,32525,32571,32594,32750,32781,32812,32836,32838,32929,32942,32948,32950,33102,33169,33284,33339,33386,33426,33467,33569,33601,33619,33661,33838,33852,33860,33982,34019,34034,34096,34252,34356,34438,34465,34495,34531,34572,34588,34628,34704,34919,34921,35034,35094,35139,35177,35279,35284,35313,35402,35502,35561,35569,35579,35647,35658,35719,35784,35808,35867,35869,36000,36264,36392,36466,36588,36680,36747,36769,36894,36917,37135,37164,37200,37232,37441,37457,37525,37621,37643,37731,37803,37833,38017,38117,38140,38147,38177,38212,38224,38251,38440,38485,38541,38609,38625,38651,38673,38699,38720,38726,38852,38874,38959,39026,39032,39097,39294,39369,39476,39544,40483,40794,40825,40869,40932,40985,41092,41222,41275,41341,41384,41546,41655,41780,41784,41878,41900,42001,42109,42116,42219,42264,42273,42484,42526,42528,42542,42636,42695,42705,42737,42865,42889,43000,43058,43130,43291,43292,43336,43412,43441,43461,43682,43732,43858,43888,43919,43992,44019,44034,44074,44103,44104,44166,44184,44264,44327,44337,44487,44568,44623,44668,44669,44695,44955,45080,45094,45113,45156,45275,45276,45326,45572,45636,45752,45826,45885,45950,46089,46170,46273,46403,46424,46484,46513,46569,46615,46809,46962,47001,47003,47105,47115,47323,47351,47413,47635,47709,47728,47950,48233,48281,48283,48362,48372,48699,48825,49013,49369,49417,49443,49505,49614,49632,49679,49734,49855,49938,49996,50054,50125,50126,50174,50304,50326,50328,50352,50374,50436,50453,50465,50642,50724,50831,51028,51351,51415,51501,51549,51733,51763,51785,51875,51891,51981,52085,52100,52190,52271,52307,52310,52524,52588,52589,52594,52657,52719,52738,52780,52847,52976,53033,53048,53167,53667,53846,53906,53956,54067,54280,54317,54387,54402,54554,54556,54578,54634,54673,54803,54873,54938,54972,55012,55107,55153,55157,55166,55238,55348,55369,55547,55579,55718,55754,55762,55769,55811,55992,56110,56124,56154,56167,56187,56206,56284,56309,56335,56440,56451,56635,56643,56649,56691,56769,56810,56811,56887,56947,56991,57049,57072,57129,57195,57206,57216,57841,58046,58056,58109,58200,58243,58324,58425,58519,58533,58542,58587,58611,58816,58859,58875,59006,59020,59073,59086,59125,59131,59265,59517,59575,59639,59828,59832,60007,60036,60075,60128,60372,60509,60524,60562,60655,60826,60887,60953,61230,61369,61443,61460,61707,61724,61798,61804,61929,61997,61999,62120,62265,62339,62370,62373,62387,62438,62447,62452,62511,62571,62716,62747,62761,62804,62933,63124,63182,63184,63191,63202,63282,63312,63454,63479,63526,63596,63660,63736,63790,63851,63883,63911,63912,63951,64041,64102,64221,64276,64349,64378,64399,64403,64496,64522,64565,64633,64729,64743,64747,64791,64804,64855,64933,65039,65120,65133,65228,65257,65301,65307,65310,65326,65379,65385,65451,65509,65611,65613,65617,65669,65700,65786,65801,65944,65995,66062,66128,66165,66244,66267,66421,66426,66481,66553,66685,66734,66758,66802,66972,67106,67196,67235,67480,67547,67634,67654,67718,67829,67838,68129,68195,68289,68331,68404,68548,68802,68838,68877,69090,69122,69176,69188,69192,69326,69333,69458,69548,69549,69604,69657,69906,69996,70029,70076,70107,70145,70204,70221,70289,70384,70393,70461,70550,70628,70873,70930,70947,70989,70993,71005,71192,71256,71321,71438,71495,71563,71666,71676,72002,72247,72255,72535,72604,72654,72657,72688,72803,72918,72919,72949,73049,73114,73123,73174,73268,73305,73339,73344,73436,73528,73604,73608,73655,73666,73717,73911,73924,73957,73958,74011,74131,74171,74249,74307,74336,74542,74549,74637,74672,74728,74783,74907,74914,74998,75080,75125,75130,75139,75215,75254,75287,75448,75524,75530,75537,75542,75554,75663,75779,75783,75788,75813,75815,75872,75914,75975,76065,76686,76757,76920,76978,77048,77145,77220,77259,77352,77400,77586,77808,77823,77845,77985,78030,78075,78125,78216,78266,78534,78536,78539,78583,78654,78752,78789,78840,78962,78964,78993,79049,79082,79169,79220,79238,79282,79327,79405,79522,79586,79621,79751,79878,79912,80018,80195,80299,80379,80534,80636,80834,81088,81112,81198,81240,81309,81365,81414,81441,81519,81600,81619,81644,81785,81789,81883,81884,81919,81964,81990,82127,82218,82223,82267,82403,82415,82462,82507,82549,82628,82808,82959,82970,83115,83130,83156,83158,83224,83300,83588,83694,83798,83904,83917,83941,83972,84060,84200,84304,84334,84579,84634,84697,84817,84828,84883,84996,85131,85200,85207,85215,85233,85307,85331,85333,85448,85534,85716,85941,85947,86000,86065,86123,86253,86300,86387,86670,86783,86821,86861,86872,86873,86875,86879,86886,86915,86935,86944,87015,87040,87070,87314,87322,87395,87416,87544,87657,87658,87660,87702,87711,87722,87736,87761,87771,87863,87871,87894,88021,88085,88089,88193,88274,88615,88619,88693,88760,88794,88826,88839,88878,89039,89059,89091,89123,89239,89242,89344,89356,89469,89523,89599,89612,89700,89701,89721,89863,90035,90106,90115,90156,90159,90198,90207,90253,90303,90321,90367,90599,90603,90672,90685,90788,90813,90831,90849,90891,91074,91200,91259,91339,91442,91555,91870,91956,92127,92180,92279,92297,92516,92553,92615,92620,92632,92791,92849,92885,93123,93157,93163,93204,93213,93394,93395,93468,93614,93715,93834,93874,93885,93904,93937,94019,94096,94231,94495,94553,94562,94566,94693,94732,94810,94844,94895,94946,94972,95014,95149,95179,95205,95311,95467,95499,95542,95558,96016,96057,96074,96083,96335,96450,96451,96655,96677,96827,96860,96946,97032,97054,97083,97122,97166,97242,97295,97451,97454,97557,97837,97877,97962,98134,98155,98162,98223,98306,98344,98428,98450,98482,98627,98652,98707,98731,98744,98777,98891,98939,98961,99024,99128,99167,99397,99482,99502,99669,99791,99859,99927,99987,99997",
          "61,77,90,96,117,141,200,245,258,308,358,393,489,490,567,626,753,776,820,961,1025,1149,1155,1312,1572,1613,1705,1999,2018,2342,2345,2348,2519,2550,2651,2744,2811,2882,2969,3112,3115,3122,3134,3194,3236,3289,3421,3483,3500,4025,4060,4130,4176,4225,4378,4416,4451,4516,4644,4707,4760,4774,4779,4976,5097,5149,5249,5439,5478,5518,5669,5702,5759,5760,6142,6192,6212,6231,6283,6287,6337,6474,6610,6757,6871,6885,6913,7040,7050,7124,7249,7290,7392,7432,7434,7441,7473,7549,7606,7736,7874,8135,8264,8299,8360,8667,8742,8766,8791,8810,9203,9234,9418,9459,9490,9502,9554,9630,9728,10065,10105,10123,10125,10161,10228,10353,10470,10513,10533,10667,10764,10872,10959,10998,11048,11064,11066,11116,11144,11207,11506,11509,11519,11548,11612,11713,11800,12018,12062,12127,12142,12482,12550,12632,12651,12705,12794,12902,12920,13097,13115,13132,13137,13299,13318,13384,13501,13521,13573,13593,13740,14119,14193,14221,14341,14483,14521,14572,14657,14691,14730,14877,14884,14886,14934,15104,15218,15263,15304,15318,15476,15484,15675,15726,15730,15803,15813,16008,16025,16137,16281,16445,16488,16582,16595,16596,16613,16654,16768,16933,17049,17058,17164,17173,17374,17581,17612,17624,17772,17778,17983,18015,18048,18144,18171,18346,18607,18741,18788,19011,19030,19049,19118,19194,19199,19227,19339,19384,19385,19409,19479,19583,19934,19971,20000,20023,20204,20236,20266,20602,20655,20676,20679,20831,20966,21014,21062,21119,21202,21251,21294,21744,21757,21789,21916,21946,22001,22132,22200,22298,22330,22420,22638,22671,22779,22861,22877,22942,23024,23090,23095,23194,23387,23389,23608,23630,23771,23775,23808,23809,23816,23927,23939,23950,24105,24302,24641,24670,24726,24866,24919,25001,25041,25180,25508,25515,25632,25634,25636,25712,25911,25925,25943,25946,26209,26278,26448,26704,26721,26810,26858,26921,27008,27132,27161,27321,27409,27437,27477,27586,27658,27659,27771,27778,27969,28171,28224,28272,28379,28403,28593,28748,28799,28971,29060,29220,29294,29353,29463,29567,29700,29790,29888,29943,30113,30164,30307,30344,30506,30602,30686,30901,30922,31009,31049,31069,31079,31119,31227,31253,31324,31395,31511,31664,31676,32175,32180,32390,32561,32694,32742,32779,32815,32818,32976,33015,33148,33332,33664,33674,34049,34077,34086,34127,34221,34472,34490,34621,34636,34736,34836,34841,34910,34953,35007,35082,35131,35201,35459,35491,35590,35598,35660,35662,35677,35715,35814,35889,35916,35942,36041,36088,36107,36288,36350,36447,36472,36880,37069,37187,37379,37390,37618,37692,37719,37730,37912,37975,38079,38234,38246,38292,38299,38330,38427,38588,38704,38708,38722,38737,38802,38844,38885,38923,38925,38987,39068,39130,39341,39405,39475,39492,39594,39621,39742,39870,39945,40032,40180,40313,40341,40401,40410,40527,40578,40618,40710,40719,40767,40798,40875,41102,41188,41435,41525,41589,41618,41625,41743,41749,41755,41770,41798,41829,41890,41913,41914,42049,42142,42294,42440,42572,42592,42668,42858,42870,42967,42968,43295,43307,43404,43481,43485,43538,43738,43853,43896,43980,44041,44156,44241,44255,44310,44476,44495,44554,44602,44738,44741,44769,44841,44844,44939,44942,44993,45016,45025,45040,45128,45180,45181,45193,45263,45288,45306,45468,45563,45672,45733,45748,45772,45892,45923,46621,46634,46752,46762,46785,46844,46997,47038,47108,47122,47123,47162,47167,47195,47219,47254,47277,47295,47349,47382,47389,47404,47531,47594,47674,47824,47841,47970,48020,48095,48270,48313,48342,48454,48459,48490,48544,48632,48676,48707,48834,48836,48914,48970,49107,49175,49275,49418,49480,49654,49842,49851,49877,49905,49957,49966,50064,50071,50073,50152,50280,50298,50390,50392,50507,50526,50608,50625,50916,50928,50967,51045,51065,51075,51098,51146,51219,51346,51387,51410,51454,51504,51526,51630,51644,51668,51702,51714,51862,51900,51929,52003,52014,52248,52450,52539,52541,52640,52646,52698,52761,52763,52878,52958,52965,53165,53166,53347,53764,53860,54160,54202,54223,54338,54446,54448,54462,54529,54564,54702,54704,55141,55171,55235,55518,55707,55736,55765,55791,55879,55964,55973,56332,56401,56471,56506,56577,56706,56801,56990,56992,57108,57261,57307,57350,57406,57416,57435,57445,57505,57550,57569,57752,57776,57868,57901,57956,58033,58079,58124,58289,58298,58627,58831,58838,58928,59127,59233,59290,59356,59424,59488,59502,59515,59584,59643,59657,59678,59746,59864,59923,59937,60119,60273,60290,60449,60519,60776,60811,60854,60959,60966,61009,61011,61023,61229,61243,61471,61569,61598,61623,61638,61742,61870,61930,61977,62091,62237,62368,62436,62446,62590,62598,63032,63036,63152,63293,63432,63527,63671,63722,63765,64047,64091,64284,64306,64351,64434,64464,64468,64546,64715,64926,64953,65119,65145,65178,65211,65378,65396,65519,65545,65687,65714,65798,65813,65823,65876,65963,65976,65988,66101,66324,66415,66574,66581,66700,66712,66825,66893,66952,67131,67172,67303,67363,67453,67674,67681,67834,67993,68395,68502,68545,68555,68557,68576,68674,68684,68975,69163,69265,69297,69409,69566,69628,69849,70153,70162,70233,70291,70388,70431,70505,70524,70616,70626,70669,70671,70817,70855,70895,70908,71029,71039,71102,71245,71314,71689,71691,71705,71733,71798,71844,71980,71993,72051,72102,72383,72394,72434,72789,72947,73063,73111,73125,73146,73284,73381,73432,73848,73984,73987,74006,74048,74107,74114,74329,74341,74376,74426,74497,74500,74608,74804,74879,74885,74888,74938,75000,75031,75062,75121,75134,75162,75237,75242,75252,75302,75359,75449,75465,75548,75556,75619,75709,75710,75754,75797,75845,75893,75917,76027,76044,76069,76211,76493,76609,76742,76752,76759,76761,76779,76874,77024,77079,77362,77388,77600,77615,77723,77841,77849,77927,77983,78033,78050,78217,78222,78312,78409,78420,78471,78477,78696,78745,78842,79023,79083,79122,79406,79417,79499,79599,79722,79775,79851,79865,79929,79998,80220,80238,80361,80459,80496,80645,80726,80830,80868,81049,81060,81125,81179,81343,81371,81396,81407,81775,81787,81808,81871,82196,82221,82341,82355,82398,82562,82690,82732,82963,82973,83082,83119,83382,83400,83531,83584,83945,84002,84037,84072,84093,84193,84356,84421,84537,84544,84636,84768,84959,84993,85020,85147,85155,85235,85272,85277,85300,85389,85403,85544,85580,85605,85651,85700,85704,85848,85864,85917,85995,86116,86142,86374,86753,86857,86871,86896,86939,86942,86968,87063,87136,87214,87295,87446,87517,87639,87797,87883,87957,87974,87980,87983,87997,88051,88289,88292,88308,88415,88612,88748,88833,88895,88996,89054,89132,89204,89223,89285,89475,89509,89592,89709,90040,90108,90178,90179,90188,90200,90346,90450,90499,90588,90740,90779,90790,90840,90895,91066,91168,91219,91349,91496,91600,91632,91676,91694,91737,91867,91950,92049,92096,92207,92250,92255,92292,92488,92904,92985,93173,93224,93352,93456,93624,93639,93820,93955,94110,94297,94305,94318,94413,94520,94609,94626,94726,94912,95148,95156,95163,95188,95241,95292,95534,95706,95768,95876,95920,95995,96042,96132,96547,96555,96651,96690,97015,97057,97113,97278,97428,97431,97473,97592,97751,97770,97797,97953,98122,98157,98209,98231,98255,98279,98313,98322,98411,98468,98474,98511,98692,98809,98863,98933,99000,99017,99109,99417,99450,99660,99680,99748,99786,99824,99841,99871",
          "31,215,216,254,420,482,576,670,749,828,850,853,1101,1132,1138,1164,1172,1258,1269,1276,1380,1419,1489,1490,1559,1565,1591,1639,1688,1699,1716,1723,1742,1761,1784,1841,1979,2043,2125,2149,2301,2311,2350,2459,2530,2562,2618,2639,2835,2867,2903,2928,2978,3000,3071,3110,3230,3244,3282,3332,3410,3670,3727,3730,3742,3974,3995,4011,4047,4092,4318,4585,4593,4787,4867,4939,4944,4958,5043,5082,5259,5342,5369,5371,5460,5630,5636,5693,5747,5954,6178,6355,6503,6645,6709,6971,7091,7095,7107,7389,7700,7823,7827,7835,7936,8014,8032,8042,8077,8133,8224,8270,8304,8395,8523,8540,8617,8646,8671,8774,9211,9219,9339,9401,9498,9513,9680,9767,9771,9775,9810,9835,9906,9989,10001,10053,10066,10067,10250,10267,10388,10498,10648,10653,10700,10730,10794,10887,11008,11182,11188,11288,11341,11427,11466,11544,11578,11624,11663,11671,11732,11741,11809,11965,12087,12103,12171,12192,12243,12332,12344,12377,12418,12423,12484,12640,12642,12654,12733,12903,12904,12915,12923,12943,12960,12987,13140,13142,13189,13262,13352,13507,13533,13625,13880,13944,13973,13986,14043,14065,14135,14224,14325,14326,14571,14577,14587,14668,14766,15250,15294,15316,15455,15516,15613,15640,15667,15733,15819,15925,15980,16076,16197,16201,16256,16283,16303,16365,16409,16503,16507,16513,16619,16626,16691,16712,16866,16959,17113,17125,17170,17188,17197,17262,17282,17310,17312,17400,17444,17490,17549,17606,17656,17698,17714,17748,17818,17881,18011,18088,18164,18216,18284,18301,18324,18373,18376,18515,18552,18596,18631,18678,18789,18848,18928,18958,18967,19017,19038,19076,19353,19431,19471,19508,19515,19524,19548,19617,19709,19716,19724,19802,19808,19814,19873,19878,19892,20035,20128,20129,20179,20182,20307,20337,20360,20471,20532,20647,20674,20724,20752,20790,20867,20882,20898,20958,20986,21019,21128,21146,21196,21269,21270,21279,21290,21333,21389,21575,21699,21755,21829,21956,21983,22041,22054,22183,22238,22358,22454,22460,22556,22569,22722,22799,22913,22973,23069,23098,23305,23366,23414,23504,23619,23641,23736,23756,23852,23919,23998,24031,24063,24104,24257,24265,24277,24354,24390,24402,24514,24658,24724,24753,24801,24819,24825,24833,24842,25026,25087,25214,25252,25271,25401,25409,25414,25568,25591,25592,25682,25700,25751,25786,25831,25854,26070,26244,26326,26376,26501,26545,26569,26774,26812,26817,26861,26903,26918,26990,27011,27022,27110,27150,27216,27234,27246,27259,27292,27400,27416,27496,27522,27560,27856,27883,27968,27980,28032,28046,28050,28068,28221,28261,28351,28429,28551,28634,28635,28736,28787,28872,28988,29035,29066,29169,29182,29198,29237,29271,29450,29454,29619,29828,29859,29866,30028,30209,30233,30388,30399,30608,30697,30699,30763,30766,30824,30930,31027,31071,31238,31337,31438,31493,31575,31695,31709,31816,31877,31881,31905,31923,32535,32755,32764,32856,33063,33126,33255,33349,33375,33524,33536,33602,33643,33702,33817,34011,34037,34045,34227,34229,34408,34544,34637,34670,34712,34793,34820,34944,34973,35025,35054,35060,35134,35186,35293,35306,35372,35397,35573,35618,35848,35957,35960,36111,36142,36268,36276,36327,36408,36449,36524,36583,36604,36635,36713,36808,36809,37196,37323,37466,37485,37536,37831,37859,37870,37962,37979,38013,38101,38110,38157,38169,38204,38253,38502,38584,38633,38666,38668,38777,38860,38929,38951,39070,39141,39229,39379,39423,39435,39478,39479,39494,39643,39806,39922,40112,40241,40281,40338,40475,40505,40540,40546,40565,40599,40828,40892,40967,40970,40979,41011,41027,41173,41207,41293,41375,41377,41464,41558,41631,41682,41745,41854,41935,41966,41998,42098,42322,42334,42396,42465,42475,42538,42549,42550,42560,42568,42684,42742,42750,42757,42849,42911,42925,43010,43143,43189,43650,43672,43710,43830,43901,44008,44050,44053,44105,44185,44359,44517,44648,44664,44682,44702,44933,44961,45203,45255,45279,45309,45396,45446,45451,45469,45502,45657,45675,45690,45832,45910,45931,46022,46115,46136,46172,46179,46246,46296,46404,46474,46502,46512,46533,46583,46627,46856,46887,46893,46901,47052,47082,47249,47260,47356,47456,47666,47851,48165,48289,48388,48397,48411,48446,48669,48746,48906,48948,49019,49073,49126,49154,49258,49339,49423,49517,49597,49601,49610,49640,49678,49738,49744,49808,50012,50063,50189,50264,50316,50397,50495,50553,50563,50626,50663,50674,50705,50786,50796,50906,50922,50970,50991,51000,51023,51041,51069,51254,51452,51485,51536,51560,51590,51614,51708,51965,51997,52011,52017,52143,52322,52343,52388,52421,52612,52673,52694,52881,52978,52980,53041,53206,53258,53265,53463,53540,53612,53643,53703,53780,53800,53829,53908,53909,54212,54292,54410,54452,54481,54546,54550,54561,54584,54740,54744,54748,54888,54952,55010,55031,55047,55101,55345,55357,55362,55682,55723,55832,55889,56017,56052,56080,56147,56177,56405,56482,56552,56584,56609,56788,56868,57024,57182,57210,57227,57322,57370,57387,57408,57440,57513,57554,57583,57684,57877,57942,57947,58001,58074,58116,58198,58248,58255,58302,58306,58369,58471,58518,58554,58801,58869,58934,59102,59193,59204,59212,59300,59384,59397,59578,59613,59635,59667,59674,59775,59880,59883,59915,59966,59987,60096,60145,60168,60183,60234,60285,60402,60520,60538,60551,60587,60682,60725,60914,60993,61024,61049,61063,61184,61266,61272,61413,61449,61467,61590,61735,61747,61765,61830,61839,61916,62217,62218,62381,62561,62601,62610,62638,62682,62831,62877,62939,63022,63046,63148,63355,63489,63604,63692,63769,63791,63862,63871,63880,64079,64132,64169,64236,64237,64243,64309,64323,64431,64448,64453,64499,64649,64749,64818,64822,64825,64887,64931,64957,64998,65000,65092,65118,65193,65249,65324,65328,65429,65511,65543,65665,65705,65713,65722,65776,65788,65796,65837,65967,66098,66130,66138,66231,66246,66261,66319,66323,66459,66477,66491,66493,66575,66587,66661,66674,66688,66733,66780,66889,67071,67157,67246,67257,67313,67417,67425,67479,67553,67763,67776,67873,68001,68077,68159,68191,68201,68231,68251,68285,68307,68388,68416,68481,68798,68807,69088,69151,69175,69197,69216,69292,69307,69366,69415,69459,69492,69569,69593,69601,69745,69795,69867,70004,70042,70070,70075,70318,70710,70724,70742,70830,70853,70875,70966,71065,71083,71244,71354,71378,71410,71415,71420,71432,71470,71498,71630,71753,72144,72166,72305,72456,72488,72722,72750,72853,72869,72878,72913,72982,72998,73090,73095,73116,73303,73464,73554,73593,73617,73644,73713,73753,73761,73803,73845,74045,74082,74122,74186,74201,74267,74516,74789,74798,75041,75049,75071,75091,75093,75172,75334,75342,75410,75436,75579,75685,75799,75843,75919,75955,76051,76140,76168,76329,76360,76455,76474,76572,76639,76698,76725,76737,76740,76795,76846,76856,76956,76988,77109,77178,77180,77276,77286,77288,77342,77367,77419,77452,77508,77535,77751,77959,78180,78224,78283,78350,78512,78636,78685,78712,78754,78806,78821,78958,79050,79085,79127,79207,79223,79320,79557,79580,79637,79699,79924,80097,80098,80114,80155,80197,80227,80303,80447,80484,80531,80583,80600,80669,80716,80728,80756,80803,80851,80999,81027,81229,81298,81356,81377,81573,81597,81686,81699,81824,81863,81977,82029,82083,82202,82404,82454,82512,82518,82563,82578,82713,82726,82798,82835,82934,83006,83015,83044,83093,83127,83181,83282,83283,83304,83384,83547,83707,83711,83744,83853,83976,83978,84021,84024,84039,84161,84300,84352,84538,84627,84647,84652,84702,84760,84902,84917,84982,85007,85108,85230,85373,85378,85801,85816,85859,85906,85912,85913,85930,85940,86174,86208,86315,86423,86474,86634,86671,86713,86781,86822,86912,86927,86969,87154,87245,87319,87351,87419,87535,87990,88027,88240,88288,88305,88480,88535,88666,88788,88799,88873,88921,88924,88962,89033,89043,89070,89082,89183,89305,89571,89687,89726,89746,89759,89794,89922,89934,89953,89967,90301,90432,90517,90665,90676,90748,90847,90865,90908,90914,90927,90977,91275,91451,91544,91558,91762,91809,91884,91938,91948,91987,92035,92159,92333,92421,92470,92504,92512,92522,92538,92555,92751,92752,92792,92820,92836,92899,92915,92936,93097,93197,93388,93525,93643,93657,93670,93671,93836,93843,93844,93869,93925,93957,94024,94050,94081,94180,94247,94276,94306,94440,94450,94458,94539,94579,94622,94648,94684,94832,94834,94839,94930,95003,95028,95180,95185,95201,95236,95261,95269,95334,95396,95403,95435,95504,95599,95634,95690,95704,95771,95817,95985,96079,96091,96253,96347,96410,96433,96570,96616,96734,96742,96800,97014,97021,97041,97088,97262,97293,97318,97439,97896,97971,98064,98459,98471,98484,98622,98631,98636,98685,98712,98732,98766,98780,98850,98905,99030,99371,99396,99437,99477,99593,99612,99649,99802,99838,99864,99903,99964",
          "16,63,65,81,475,660,763,891,1004,1006,1032,1084,1102,1130,1203,1270,1322,1366,1414,1442,1558,1828,2008,2070,2075,2133,2221,2349,2351,2443,2584,2859,2895,2987,3014,3061,3090,3206,3213,3226,3356,3364,3558,3634,3801,3869,3873,3954,4246,4277,4386,4425,4486,4630,4889,4915,4950,5121,5234,5354,5366,5407,5473,5522,5548,5601,5677,5768,5848,5910,5965,5978,6109,6123,6137,6210,6307,6420,6455,6456,6462,6505,6521,6586,6629,6654,6759,6766,7008,7020,7111,7233,7238,7316,7322,7426,7488,7543,7738,7756,7801,7880,7930,8061,8117,8218,8318,8352,8421,8510,8581,8614,8658,8707,8756,8940,9050,9202,9212,9225,9304,9356,9529,9941,9988,10192,10236,10345,10833,10919,11314,11344,11516,11657,11681,11787,11832,11969,12076,12164,12202,12262,12272,12274,12450,12531,12636,12689,12692,12768,12830,12854,12890,12971,13030,13037,13122,13167,13296,13312,13525,13539,13607,13637,13666,13799,13935,14075,14102,14106,14268,14487,14498,14504,14535,14566,14996,15063,15073,15167,15221,15368,15598,15616,15692,15744,15975,16034,16093,16264,16410,16553,16563,16766,16865,16868,16925,17283,17355,17430,17519,17677,17705,17729,17855,18043,18064,18179,18387,18444,18511,18581,18832,18889,18937,18941,18989,19071,19165,19273,19282,19363,19513,19749,19750,19784,19786,19805,20021,20407,20565,20620,20698,20706,20904,21013,21090,21170,21197,21205,21305,21557,21629,21648,21700,21714,21825,21862,21939,21985,22006,22026,22310,22409,22551,22611,22612,22700,22776,22790,22878,23004,23047,23085,23173,23216,23299,23377,23392,23437,23466,23472,23725,23865,23886,23937,23994,24012,24129,24200,24413,24559,24565,25017,25140,25175,25231,25234,25338,25393,25442,25446,25576,25638,25683,25701,25770,25942,25995,26027,26052,26307,26377,26535,26555,26686,26713,26754,26768,26867,26895,26904,27102,27181,27213,27426,27431,27574,27652,27809,27988,28010,28094,28107,28138,28210,28215,28240,28452,28534,28640,28679,28706,28707,28715,28741,28876,28958,29040,29063,29188,29193,29254,29411,29422,29494,29514,29527,29537,29548,29647,29737,29883,29944,29982,30165,30306,30356,30491,30524,30594,30695,30703,30948,31019,31034,31035,31125,31256,31443,31474,31667,31756,31775,32215,32260,32276,32630,32687,32723,32762,32842,32887,32891,33129,33142,33188,33239,33241,33342,33366,33538,33694,33706,33789,33813,33968,33999,34093,34190,34415,34659,34685,34877,34923,34938,35109,35142,35273,35370,35565,35567,35668,35803,36050,36075,36100,36109,36129,36254,36476,36537,36572,36639,36654,36726,36867,36922,36943,36957,37074,37155,37256,37312,37358,37371,37543,37544,37551,37591,37774,37780,37966,38024,38099,38150,38208,38248,38349,38384,38604,38662,38669,39089,39408,39429,39433,39482,39549,39578,39628,39849,39875,39901,39930,40249,40311,40668,40702,40726,40761,40884,40947,41042,41081,41087,41115,41118,41181,41246,41248,41254,41707,41872,41902,41918,42133,42148,42201,42288,42295,42364,42410,42508,42533,42698,42776,43098,43105,43167,43266,43284,43334,43371,43437,43476,43559,43701,43867,43915,43981,43983,44214,44226,44460,44621,44636,44660,44749,44827,45021,45076,45228,45445,45770,45871,46043,46083,46206,46228,46230,46257,46279,46298,46303,46332,46369,46376,46442,46523,46566,46604,46673,46753,46825,46929,46948,46981,47112,47176,47187,47201,47211,47375,47508,47555,47556,47562,47603,47657,47692,47783,47795,47812,47933,48013,48016,48035,48040,48057,48122,48386,48450,48760,48857,48881,49071,49184,49308,49419,49431,49475,49483,49490,49784,49846,49866,49883,49892,50031,50133,50200,50450,50508,50624,50792,50809,50816,51066,51088,51133,51187,51264,51273,51334,51336,51348,51360,51856,51984,52189,52229,52272,52469,52564,52683,52711,52800,52962,53038,53046,53175,53301,53337,53391,53425,53511,53639,53640,53671,53691,53728,53803,53889,53944,53998,54052,54116,54253,54340,54358,54374,54492,54504,54562,54643,54654,54695,54846,54874,54910,54973,55045,55328,55364,55397,55521,55524,55586,55798,55937,56178,56254,56270,56327,56527,56613,56726,56747,56755,56799,56927,56979,57043,57120,57229,57241,57281,57319,57510,57527,57621,57988,58343,58372,58389,58549,58553,58603,58865,58954,58971,59109,59156,59368,59382,59801,60016,60368,60390,60603,60645,60813,60829,60850,60897,60937,60958,61159,61162,61294,61307,61318,61387,61433,61610,61769,61797,61967,62126,62187,62274,62287,62308,62318,62326,62405,62435,62470,62524,62536,62577,62636,62687,62696,62720,62806,62807,62866,62943,63084,63307,63315,63437,63729,63809,63844,63852,64078,64172,64235,64280,64291,64478,64479,64608,64616,64654,64694,64745,64797,64986,64995,65056,65063,65088,65134,65544,65667,65668,65701,65747,65758,65914,66109,66126,66335,66454,66540,66615,66763,66935,67010,67149,67171,67187,67236,67306,67536,67552,67563,67651,67730,67742,67867,67913,67920,67926,67952,68045,68187,68215,68246,68424,68609,68626,68833,68870,69027,69094,69116,69235,69319,69437,69439,69447,69508,69512,69568,69778,70048,70281,70465,70611,70632,70667,70782,71044,71123,71129,71223,71421,71463,71534,71740,71757,71988,72136,72217,72321,72371,72469,72509,72549,72588,72718,72725,72772,72930,72993,73055,73170,73201,73373,73631,73636,73881,74097,74184,74227,74319,74345,74365,74461,74525,74682,74691,75115,75275,75319,75389,75444,75558,75588,75661,75693,75767,75972,76012,76105,76109,76293,76306,76543,76620,76680,76681,76702,76858,76929,77106,77265,77300,77420,77444,77450,77765,77783,78181,78187,78208,78287,78529,78554,78576,78615,78663,78716,78802,78850,78873,78926,79058,79328,79420,79496,79632,79735,79783,79866,79960,79972,80015,80025,80038,80091,80099,80170,80171,80239,80312,80399,80487,80500,80568,80627,80678,80720,80808,80833,80858,81013,81130,81143,81154,81312,81410,81466,81467,81565,81596,81643,81681,81983,82021,82213,82214,82302,82412,82440,82443,82493,82636,82716,82837,82854,83365,83526,83536,83639,83714,83746,83776,83803,83810,83817,83824,83835,83901,83935,83997,84309,84358,84364,84394,84664,84762,85061,85169,85197,85198,85260,85356,85511,85624,85675,85778,85825,85852,85935,85953,86023,86302,86332,86336,86895,86947,87152,87168,87171,87361,87570,87718,87937,87948,88095,88162,88386,88443,88447,88452,88466,88560,88605,88725,89015,89110,89164,89270,89345,89362,89510,89511,89626,89684,89787,89823,89867,89897,90031,90125,90182,90224,90308,90317,90377,90678,90692,90734,91048,91158,91228,91300,91402,91510,91569,91702,91826,92135,92151,92552,92573,92589,92596,92616,92749,93016,93115,93147,93305,93385,93477,93562,93669,93691,93706,93839,94076,94182,94184,94238,94262,94343,94384,94576,94688,94710,94842,94860,95011,95208,95231,95405,95517,95561,95645,96065,96071,96221,96420,96522,96577,96639,96753,96762,96785,96836,96876,97340,97393,97463,97516,97560,97568,97669,97803,97831,97867,97978,98012,98014,98126,98283,98329,98343,98361,98375,98477,98643,98913,98983,99048,99085,99153,99170,99410,99551,99644,99661,99690,99874,99945",
          "58,88,202,372,481,658,751,823,844,956,1036,1212,1256,1263,1316,1356,1516,1517,1590,1695,1734,2184,2284,2450,2570,2608,2654,2787,2791,2796,2809,2818,2832,3054,3311,3334,3383,3388,3548,3595,3672,3688,3716,3749,3758,3791,3870,3910,3953,4159,4242,4294,4332,4409,4443,4555,4599,4683,4686,4693,4720,4783,4888,4964,4989,5024,5127,5182,5290,5338,5393,5476,5570,5603,5743,5755,5777,5880,6033,6059,6096,6125,6143,6200,6205,6278,6288,6305,6370,6707,6715,6729,6775,6826,6997,7088,7145,7161,7190,7198,7263,7293,7309,7387,7412,7516,7527,7623,7864,7870,7977,8002,8013,8165,8183,8217,8255,8282,8290,8312,8508,8583,8603,8676,8892,8980,8985,9031,9088,9174,9245,9270,9432,9454,9629,9766,9820,9845,10150,10152,10219,10478,10500,10628,10672,10745,10870,10926,10966,11118,11173,11275,11278,11285,11311,11395,11456,11525,11536,11561,11564,11605,11719,11752,11814,11831,11914,12043,12181,12275,12390,12681,12720,12751,12772,12816,12844,12864,12878,12934,12986,13018,13019,13174,13300,13301,13347,13388,13398,13417,13432,13599,13629,13634,13844,13846,13888,13960,14101,14159,14279,14356,14384,14390,14607,14613,14715,14720,14839,14908,14933,14978,14986,15053,15130,15330,15348,15351,15361,15504,15542,15561,15654,15749,15850,15935,16198,16280,16372,16424,16492,16615,16618,16665,16684,16742,16852,16935,17037,17064,17134,17346,17413,17506,17509,17545,17580,17716,17869,17913,17947,17971,17986,18085,18165,18205,18253,18315,18351,18359,18404,18434,18470,18494,18737,18864,19004,19180,19265,19370,19436,19452,19609,19654,19723,19790,19830,19833,19931,19937,20018,20084,20109,20115,20257,20308,20363,20366,20388,20653,20821,20874,20876,21040,21203,21206,21212,21281,21309,21311,21380,21388,21435,21631,21663,21687,21701,21826,21848,21857,21866,21935,21993,22112,22169,22228,22260,22288,22364,22378,22386,22401,22533,22608,22696,22721,22785,22838,22934,22976,23054,23066,23156,23254,23289,23379,23409,23477,23529,23663,23666,23710,23830,23858,23874,23910,23961,24135,24198,24215,24253,24295,24303,24306,24325,24443,24548,24603,24796,24803,24912,24917,24922,24944,25014,25135,25195,25237,25258,25354,25502,25675,25799,26103,26163,26214,26342,26635,26670,26763,26809,26813,26819,26850,26860,26964,26967,27037,27067,27115,27273,27275,27336,27370,27390,27410,27418,27565,27568,27744,27798,27813,27814,27831,27846,27857,27874,27983,27996,28017,28196,28300,28333,28392,28424,28509,28548,28710,28801,28860,29101,29262,29313,29578,29680,29780,29860,29929,29974,30061,30062,30160,30221,30224,30290,30330,30401,30508,30519,30723,30736,30786,30854,30876,30885,30921,30936,30945,30950,31254,31286,31293,31303,31391,31580,31634,31645,31707,31764,31802,31848,31857,31985,32049,32098,32130,32162,32171,32178,32197,32386,32394,32513,32734,32735,32884,32955,32957,32959,33084,33144,33196,33225,33317,33398,33508,33540,33631,33644,33693,33700,33756,33956,34020,34098,34145,34369,34391,34513,34556,34682,34769,34776,34840,34860,34868,34930,34981,35159,35187,35216,35358,35452,35635,35678,35859,36071,36125,36138,36244,36277,36480,36481,36597,36623,36757,36789,36810,36926,37156,37355,37439,37489,37575,37691,37712,37753,37782,37787,37841,37996,38014,38057,38070,38091,38245,38252,38280,38289,38315,38400,38550,38812,38864,38916,39009,39105,39152,39239,39243,39309,39339,39347,39360,39410,39412,39603,39613,39676,39844,39882,40054,40161,40336,40357,40387,40605,40737,40856,40946,40949,40956,41075,41121,41135,41153,41288,41328,41400,41490,41494,41515,41697,41834,42094,42118,42182,42184,42190,42239,42291,42392,42444,42461,42481,42596,42655,42822,42909,42961,42980,43047,43082,43134,43166,43221,43337,43391,43565,43680,43730,43769,43884,43938,43997,44007,44120,44146,44152,44328,44438,44453,44497,44614,44718,44887,44904,44985,45002,45081,45092,45168,45219,45327,45399,45528,45644,45727,45964,46016,46142,46304,46310,46428,46431,46505,46596,46675,46720,46944,46993,46998,47009,47166,47208,47266,47304,47447,47487,47913,47941,47942,47995,48051,48075,48193,48254,48364,48443,48496,48516,48539,48631,48701,48714,48731,48738,48804,48900,48916,48934,48942,49004,49083,49090,49338,49496,49497,49539,49621,49626,49733,49763,49870,49991,49992,50040,50141,50184,50228,50365,50784,50787,50805,50817,50870,50947,50952,50972,51003,51061,51137,51169,51174,51266,51324,51409,51523,51689,51696,51794,51886,51905,51909,52168,52318,52404,52687,52743,52857,52901,52924,52950,53248,53313,53517,54012,54036,54070,54103,54178,54234,54289,54308,54346,54568,54913,55018,55168,55476,55484,55508,55575,55671,55714,55730,55818,55878,55918,56005,56097,56447,56558,56629,56668,56711,56723,56835,57008,57115,57148,57160,57220,57382,57452,57485,57515,57559,57688,57712,57775,57849,57853,57862,57933,57954,57958,58123,58201,58218,58233,58284,58351,58373,58550,58588,58692,58941,58980,59354,59440,59466,59499,59544,59835,59968,59976,60133,60196,60308,60311,60379,60744,60748,60753,60839,60846,60861,60869,60946,61032,61087,61155,61193,61236,61321,61390,61429,61473,61542,61633,61692,61831,61835,61851,61853,61928,61939,61944,62031,62060,62175,62182,62259,62315,62335,62345,62363,62586,62593,62605,62817,62821,62910,63164,63287,63299,63434,63512,63518,63636,63644,63748,63779,63814,63935,63955,64092,64115,64209,64246,64322,64324,64420,64454,64589,64591,64611,64858,64902,64920,64952,65035,65080,65108,65285,65364,65719,65746,65762,65879,66058,66059,66064,66217,66300,66362,66497,66500,66511,66601,66607,66654,66743,66843,66861,67176,67178,67201,67202,67231,67367,67396,67448,67544,67575,67577,67731,67967,68028,68099,68140,68165,68221,68239,68359,68488,68526,68607,68621,68655,68691,68760,68806,68867,68996,69071,69139,69152,69194,69198,69354,69426,69541,69645,69780,69799,69805,69869,69901,69915,69986,69994,70098,70191,70196,70207,70298,70327,70559,70585,70866,70890,70894,70924,70976,71131,71152,71156,71212,71239,71369,71462,71602,71720,71877,72047,72101,72111,72112,72275,72313,72347,72369,72415,72573,72650,72724,72769,72835,73037,73042,73232,73259,73481,73672,73688,73792,73838,73976,73990,74192,74431,74462,74571,74646,74676,74742,74833,74920,75003,75007,75100,75168,75282,75404,75484,75496,75535,75613,75616,75931,75933,75980,76014,76040,76066,76106,76116,76147,76179,76320,76550,76964,77053,77083,77356,77610,77647,77655,77687,77712,77853,77923,77926,78008,78042,78124,78141,78157,78274,78326,78365,78395,78603,78676,78697,78720,78749,78862,78955,78987,79021,79073,79248,79358,79466,79491,79500,79584,79652,79697,79718,79752,79753,79956,79982,80127,80188,80261,80321,80342,80453,80526,80554,80824,80990,81214,81221,81230,81296,81306,81337,81427,81431,81558,81679,82042,82097,82109,82244,82316,82473,82534,82548,82557,82656,82718,82775,82992,83002,83068,83079,83099,83160,83298,83519,83896,84157,84212,84228,84284,84406,84418,84570,84589,84593,84619,84739,84740,84890,84919,85015,85054,85226,85434,85475,85523,85599,85623,85710,85712,85781,85787,85968,86055,86171,86190,86230,86350,86372,86489,86584,86646,87346,87362,87479,87486,87776,87814,87868,87878,87920,88053,88248,88339,88345,88351,88356,88414,88461,88554,88555,88633,88795,88885,89217,89249,89256,89343,89412,89453,89564,89586,89587,89640,89707,89715,89951,89968,89979,90225,90254,90416,90457,90481,90501,90651,90715,90807,90867,90984,90987,91289,91378,91393,91436,91516,91674,91701,91769,91813,91842,92149,92154,92239,92252,92345,92605,92686,92709,92896,93019,93086,93205,93312,93361,93439,93596,93667,93685,93911,94028,94137,94329,94371,94372,94396,94467,94506,94545,94670,94738,94837,94868,94933,95062,95162,95215,95234,95322,95364,95392,95408,95430,95475,95617,95654,95713,95838,95893,95936,95944,95972,95980,96098,96355,96431,96479,96572,96643,96801,96838,96905,96911,96966,97201,97216,97362,97369,97373,97459,97530,97565,97591,97673,97740,97753,97767,97982,98101,98114,98331,98414,98722,98738,98762,98893,98954,98977,99156,99165,99198,99643,99741,99758,99777,99830,99870,99942",
          "64,131,174,204,270,339,365,450,464,536,641,671,913,1110,1124,1126,1139,1596,1725,1731,1809,2078,2164,2191,2262,2283,2331,2400,2432,2462,2476,2513,2558,2575,2671,2675,2691,2777,2911,3063,3077,3178,3187,3192,3267,3327,3419,3432,3511,3515,3587,3630,3693,3757,3855,3896,3986,4063,4071,4125,4132,4293,4442,4640,4648,4650,4798,4841,4844,4934,4936,4963,4984,4990,5069,5212,5229,5334,5435,5544,5620,5790,6182,6285,6679,6808,6840,6914,7077,7156,7256,7276,7355,7406,7491,7690,7698,7979,8340,8374,8416,8572,8586,8606,8665,8692,8768,8820,8951,8976,9119,9132,9221,9226,9265,9309,9420,9433,9580,9607,9686,9787,9790,9862,9872,9944,9982,10180,10361,10422,10565,10592,10785,10869,10920,11071,11077,11150,11257,11258,11280,11312,11339,11359,11413,11436,11478,11556,11647,11686,11762,11854,11869,11878,11956,12016,12118,12193,12221,12371,12552,12571,12573,12845,13175,13219,13233,13238,13253,13380,13459,13477,13602,13610,13728,13771,13788,13806,14028,14146,14235,14418,14443,14599,14728,14811,14958,14972,15050,15212,15228,15328,15346,15363,15394,15477,15499,15505,15515,15586,15607,15678,15879,16077,16099,16154,16189,16209,16214,16309,16369,16404,16458,16487,16522,16531,16538,16629,16655,16660,16672,16780,16812,16814,16820,16967,16993,17014,17039,17085,17101,17252,17357,17397,17426,17427,17493,17663,17724,17940,17941,17967,17981,18075,18147,18183,18377,18448,18760,18868,18916,19003,19035,19037,19085,19206,19250,19319,19438,19457,19458,19477,19507,19662,19754,19819,19852,19867,19920,19926,19944,19988,19999,20013,20040,20267,20358,20361,20517,20542,20635,20656,20704,20825,20881,20950,20980,20981,21057,21234,21239,21298,21344,21429,21436,21469,21502,21507,21760,21798,21823,21858,21924,21950,21957,21982,22061,22067,22135,22137,22157,22175,22225,22322,22337,22342,22382,22417,22465,22504,22601,22663,22920,22928,22952,23027,23152,23260,23285,23312,23327,23355,23369,23385,23393,23402,23407,23431,23442,23594,23672,23691,23970,24066,24067,24087,24099,24400,24430,24431,24491,24499,24614,24644,24732,24736,24849,24882,24941,24972,24982,24997,25013,25018,25106,25137,25272,25327,25363,25395,25413,25553,25628,25645,25653,25734,25767,25908,25961,26032,26045,26116,26125,26134,26162,26171,26194,26254,26306,26312,26358,26419,26451,26453,26488,26512,26528,26611,26656,26782,26808,26851,26868,27042,27109,27160,27189,27223,27361,27364,27510,27585,27649,27766,28152,28194,28200,28349,28372,28431,28451,28546,28557,28625,28629,28652,28722,28734,28797,28813,28907,29002,29033,29120,29141,29189,29431,29446,29505,29531,29634,29755,29811,29846,29971,29984,30012,30069,30082,30159,30171,30176,30219,30245,30362,30363,30364,30372,30377,30428,30469,30474,30481,30535,30552,30560,30683,30720,30795,30860,30888,30915,30946,30999,31044,31045,31091,31096,31127,31188,31218,31241,31338,31404,31467,31501,31526,31687,31724,31760,31928,31988,32039,32078,32246,32285,32350,32359,32400,32418,32568,32678,32691,32697,32717,32793,32811,32821,32845,32847,32913,33038,33070,33103,33157,33199,33290,33394,33433,33450,33495,33500,33534,33584,33634,33667,33692,33727,33729,33777,33781,33803,33871,33893,34058,34325,34436,34523,34675,34711,34719,34855,34857,34905,35101,35144,35152,35156,35162,35176,35181,35229,35246,35308,35432,35646,35653,35670,35689,35835,35911,35956,36068,36171,36294,36507,36656,36770,36918,36998,37001,37138,37365,37487,37532,37623,37654,37669,37846,37987,38078,38093,38173,38348,38356,38378,38455,38460,38479,38556,38605,38636,38735,38757,38863,39053,39086,39193,39201,39346,39386,39431,39491,39528,39575,39618,39731,39739,39965,40099,40152,40175,40184,40226,40399,40415,40516,40568,40629,40632,40715,40843,40865,40887,40910,40916,40971,41002,41016,41026,41065,41080,41096,41250,41256,41426,41498,41580,41611,41629,41639,41764,41880,41969,42119,42120,42161,42193,42249,42255,42297,42391,42531,42585,42623,42641,42725,42736,42861,42867,42978,43019,43020,43099,43274,43280,43289,43442,43544,43563,43649,43700,43748,43762,43779,43849,43976,43979,44006,44083,44097,44129,44135,44288,44332,44368,44377,44410,44418,44523,44535,44583,44606,44831,44890,44920,44999,45004,45119,45197,45345,45487,45492,45517,45561,45631,45650,45673,45754,45762,45766,45778,45954,45970,46010,46046,46118,46347,46439,46493,46726,46737,46765,46804,46902,47103,47116,47371,47489,47506,47511,47543,47658,47661,47736,47748,47756,47767,48222,48304,48337,48417,48423,48639,48642,48689,48742,48751,48762,48796,48802,48852,48864,48885,48993,49096,49131,49147,49152,49205,49250,49327,49352,49358,49359,49406,49452,49482,49559,49633,49652,49698,49714,49739,49900,49971,50001,50014,50019,50039,50108,50233,50263,50296,50332,50383,50396,50447,50520,50601,50710,50794,50923,50942,51038,51211,51223,51580,51744,51754,51855,51990,52016,52121,52132,52268,52304,52329,52342,52375,52442,52476,52499,52620,52652,52696,52930,52985,53039,53134,53160,53182,53585,53685,53754,53849,53861,53870,53892,53925,53937,54006,54112,54276,54277,54323,54342,54717,54751,54905,54921,55022,55060,55085,55416,55451,55859,55887,55893,55917,55919,55958,56069,56196,56199,56241,56341,56343,56636,56724,56794,56952,56962,57083,57118,57161,57172,57283,57308,57354,57411,57539,57735,57750,57945,58087,58125,58141,58244,58328,58345,58385,58406,58691,58733,58738,58915,58935,59040,59046,59081,59094,59101,59105,59169,59408,59417,59588,59666,59686,59792,59867,60035,60067,60074,60094,60100,60131,60159,60296,60338,60358,60412,60438,60573,60602,60604,60772,60796,60856,60929,61241,61361,61383,61538,61580,61642,61701,61864,61937,62006,62306,62347,62526,62840,62898,62936,63174,63241,63267,63409,63576,63649,63682,63714,63718,63755,63783,63804,63821,63835,63907,63946,64113,64181,64203,64226,64347,64364,64473,64480,64585,64640,64688,64805,64894,64973,64997,65086,65131,65132,65154,65161,65448,65478,65595,65610,65748,65909,65965,65981,66045,66100,66118,66235,66412,66416,66452,66506,66617,66652,66690,66754,66986,67011,67022,67030,67211,67268,67355,67501,67532,67572,67584,67668,67685,67691,67723,67814,67820,67844,67905,67921,67966,67970,67974,68082,68155,68346,68379,68381,68425,68639,68643,68753,68765,68775,68785,68845,68885,68903,68944,69344,69422,69427,69606,69726,69738,69947,70043,70213,70251,70259,70303,70408,70410,70504,70557,70567,70597,70664,70711,70720,70933,71057,71217,71262,71441,71484,71504,71535,71536,71590,71685,71718,71791,71913,72100,72158,72185,72191,72292,72322,72453,72466,72503,73056,73153,73167,73248,73311,73364,73543,73546,73579,73581,73621,73842,73853,73921,74008,74069,74145,74148,74172,74196,74304,74493,74553,74671,74688,74832,74838,75083,75084,75240,75368,75433,75483,75562,75605,75697,75774,76036,76080,76128,76141,76167,76284,76303,76504,76553,76591,76624,76627,76713,76784,76855,76923,77094,77212,77216,77239,77245,77341,77432,77447,77621,77792,77822,77833,77834,77928,78007,78174,78182,78289,78319,78340,78342,78487,78632,78648,78689,78767,78788,78855,78869,78970,78977,79078,79087,79103,79217,79243,79257,79355,79374,79558,79600,79602,79926,79945,79992,80005,80108,80178,80308,80341,80414,80480,80670,80812,80877,80936,80943,80946,81008,81156,81251,81259,81317,81411,81450,81479,81666,81705,81729,81971,82006,82080,82151,82604,82684,82705,82736,82997,83177,83197,83218,83287,83336,83614,83660,83782,83794,83804,83834,83919,84003,84034,84043,84156,84203,84249,84283,84348,84428,84429,84470,84591,84707,84755,84839,84851,84863,84866,84897,84936,84951,85024,85130,85159,85257,85473,85514,85550,85616,85773,85814,85822,85884,85894,86084,86101,86203,86238,86275,86361,86478,86762,86980,87058,87153,87174,87226,87625,87659,87680,87765,87773,87788,87956,88071,88374,88679,88771,88860,88869,88969,89064,89127,89349,89378,89455,89741,89768,89852,89997,90034,90174,90186,90194,90223,90259,90480,90546,90578,90650,90980,91024,91049,91057,91120,91194,91243,91656,91666,91668,91717,91772,91836,91858,91874,92070,92144,92194,92247,92261,92348,92464,92520,92571,92798,92938,92954,93091,93104,93135,93191,93231,93253,93340,93512,93653,93658,93693,93695,93789,93880,93899,93948,94013,94023,94221,94224,94335,94388,94428,94463,94475,94503,94504,94510,94523,94548,94592,94654,94666,94809,94922,95046,95245,95280,95282,95445,95497,95516,95636,95647,95854,95877,95881,95924,95983,96031,96033,96089,96115,96262,96351,96449,96454,96817,96931,97082,97169,97290,97405,97478,97483,97488,97505,97528,97633,97809,97829,97858,98086,98274,98325,98353,98453,98526,98624,98747,98765,98802,98852,98934,98952,99114,99120,99291,99362,99370,99421,99466,99563,99568,99611,99613,99812,99839,99976,99998",
          "62,115,191,357,547,568,597,648,667,688,695,884,1011,1120,1153,1304,1354,1364,1467,1566,1603,1618,1685,1702,1819,1821,1853,1908,2139,2170,2277,2289,2306,2417,2471,2527,2590,2600,2623,2655,2730,2848,2879,3141,3251,3256,3303,3307,3339,3372,3390,3484,3593,3677,3709,3775,3970,4051,4120,4197,4221,4228,4397,4458,4521,4824,4872,4986,5132,5176,5183,5347,5437,5490,5505,5584,5657,5893,5953,6031,6292,6551,6556,6623,6630,6657,6664,6769,6975,7034,7153,7170,7178,7197,7305,7306,7332,7405,7472,7490,7518,7948,8121,8213,8241,8250,8262,8390,8444,8538,8633,8642,8648,8814,8840,8947,8957,8962,9035,9110,9116,9149,9284,9331,9335,9350,9351,9392,9402,9403,9416,9486,9578,9735,9753,9763,9971,9981,10141,10157,10187,10374,10410,10425,10432,10464,10553,10680,10737,10753,10796,10811,10829,10832,10910,10988,11062,11135,11138,11175,11185,11195,11203,11221,11222,11259,11420,11508,11515,11574,11643,11661,11754,11947,12002,12055,12065,12223,12271,12372,12415,12523,12569,12570,12614,12648,12657,12773,12789,12846,12880,12908,12916,12970,13000,13013,13109,13247,13303,13355,13565,13787,13825,13971,14066,14116,14194,14195,14353,14361,14544,14584,14627,14699,14807,14897,15052,15061,15108,15152,15283,15298,15436,15449,15495,15564,15633,15773,15781,15838,15888,15990,16311,16352,16471,16723,16752,16813,16854,16912,17023,17121,17148,17194,17220,17249,17453,17531,17558,17573,17579,17680,17725,17773,17802,17892,17912,18113,18267,18268,18332,18385,18468,18518,18537,18655,18735,18768,18777,18785,18800,18816,18857,18896,18911,19001,19170,19214,19219,19281,19304,19389,19649,19668,19766,19769,19908,20016,20145,20165,20171,20270,20316,20412,20630,20734,20920,21012,21044,21068,21076,21077,21275,21301,21306,21371,21483,21505,21541,21547,21713,21736,21909,21965,22102,22139,22172,22191,22266,22333,22431,22512,22627,22669,22692,22707,22737,22791,22865,22940,22989,23061,23062,23070,23073,23094,23118,23120,23131,23322,23427,23432,23469,23695,23707,23741,23779,23802,23805,23859,23913,23971,24036,24081,24218,24238,24278,24291,24373,24388,24396,24589,24705,24834,24995,25048,25090,25132,25159,25198,25225,25230,25358,25475,25476,25538,25630,25727,25824,26022,26033,26054,26341,26434,26527,26562,26714,26765,26822,26862,26938,27065,27218,27219,27250,27339,27379,27580,27839,27861,27942,27951,28081,28175,28205,28297,28310,28445,28500,28575,28619,28739,28791,28920,28933,28985,29001,29049,29100,29672,29702,29782,29914,29964,30318,30359,30415,30453,30522,30565,30615,30674,30773,30903,30909,30917,30923,30955,30958,30988,30994,31010,31024,31105,31118,31274,31305,31381,31446,31597,31607,31613,31641,31643,31823,31828,31897,31946,31957,31960,32070,32100,32103,32173,32207,32283,32311,32356,32614,32620,32740,32778,32805,32834,32858,32937,33133,33265,33333,33346,33348,33396,33405,33492,33660,33761,33858,33874,33904,33909,33954,34039,34050,34054,34231,34248,34260,34285,34287,34374,34428,34514,34567,34632,34764,34765,34791,34804,34891,34965,34974,34984,34989,35160,35183,35338,35393,35433,35442,35448,35457,35475,35563,35669,35701,35728,35774,35809,35826,35993,36024,36040,36073,36399,36405,36415,36423,36441,36511,36614,36666,36716,36717,36749,37000,37037,37041,37110,37139,37204,37233,37300,37461,37526,37558,37559,37593,37604,37740,37781,37819,37877,37917,37949,37972,38080,38170,38186,38403,38565,38621,38740,38941,39065,39079,39093,39175,39242,39287,39345,39407,39430,39489,39523,39539,39607,39799,40020,40078,40156,40165,40169,40207,40304,40376,40380,40407,40493,40494,40525,40820,40871,40921,40962,41160,41177,41201,41334,41372,41391,41461,41539,41574,41583,41702,41740,41760,41856,42047,42087,42111,42139,42145,42203,42217,42275,42369,42479,42512,42523,42680,42709,42724,42810,42887,42895,42931,43107,43185,43204,43241,43451,43560,43579,43845,43886,43948,43955,43987,43989,44016,44101,44144,44151,44261,44291,44297,44312,44363,44372,44378,44397,44428,44587,44677,44720,44771,44803,44876,44896,44903,44947,45003,45227,45284,45405,45417,45523,45593,45619,45637,45780,45836,45847,45926,46174,46204,46219,46297,46331,46342,46371,46438,46791,46858,46870,46896,46959,47046,47061,47357,47379,47579,47597,47688,47734,47772,47806,47837,47854,47878,48044,48080,48105,48131,48153,48209,48223,48308,48318,48329,48369,48373,48534,48568,48649,48686,48726,48744,48773,48806,48975,49189,49446,49449,49548,49582,49638,49648,49693,49756,49921,50022,50113,50235,50286,50366,50377,50475,50494,50555,50721,50864,50938,51117,51149,51201,51337,51398,51563,51586,51731,51816,51974,51976,52046,52150,52176,52182,52303,52372,52427,52468,52689,52830,52875,53065,53091,53114,53128,53147,53173,53174,53237,53483,53509,53562,53632,53661,53696,53740,53779,53796,53833,53836,53871,53949,54075,54118,54210,54217,54230,54241,54258,54268,54320,54484,54537,54668,54719,54733,54813,54974,55000,55075,55093,55102,55127,55207,55285,55526,55545,55594,55648,55763,55831,55833,55868,55902,55922,56161,56174,56183,56192,56275,56422,56478,56487,56544,56601,56680,56709,56722,56857,56916,56981,57033,57037,57122,57304,57499,57520,57723,57758,57812,57826,57842,57858,58007,58062,58092,58157,58277,58464,58602,58702,58721,58729,58732,58812,58944,59098,59138,59161,59215,59278,59312,59329,59367,59398,59439,59491,59526,59598,59611,59622,59671,59944,59957,59992,60060,60127,60261,60329,60355,60407,60445,60503,60523,60532,60555,60597,60736,60874,60919,61098,61203,61297,61328,61476,61515,61587,61789,61841,62249,62262,62300,62311,62313,62637,62700,62754,62852,62902,62915,62930,62969,62996,63005,63012,63096,63099,63193,63377,63378,63601,63705,63892,63983,64070,64205,64219,64222,64247,64251,64286,64316,64387,64497,64569,64576,64614,64739,64869,64943,65176,65222,65351,65459,65573,65619,65658,65866,65875,66009,66088,66152,66173,66222,66285,66436,66453,66530,66532,66629,66694,66707,66767,66784,66803,66828,66840,66857,66999,67292,67338,67377,67387,67477,67506,67526,67530,67623,67704,67779,67806,68007,68044,68158,68348,68422,68447,68537,68671,68705,68739,68745,68793,68890,68908,68967,68991,69086,69142,69355,69383,69463,69596,69803,69826,69828,70045,70274,70286,70644,70728,70736,70747,70780,71149,71165,71178,71221,71232,71315,71351,71370,71392,71623,71629,71702,71841,71848,71870,71875,71926,72065,72226,72317,72329,72333,72335,72363,72454,72730,72770,72796,72884,72904,72917,72958,72967,73030,73065,73122,73204,73247,73308,73314,73376,73409,73468,73580,73953,74018,74042,74231,74237,74290,74343,74472,74590,74693,74759,74901,75096,75120,75189,75191,75216,75297,75405,75633,75683,75763,75902,75915,75926,75935,75945,75982,75989,76030,76034,76055,76058,76092,76137,76139,76203,76252,76308,76600,76830,76907,76941,77037,77062,77112,77174,77232,77244,77292,77453,77511,77661,77678,77692,77734,77750,77814,77867,78065,78193,78402,78579,78846,79051,79057,79177,79325,79383,79517,79544,79614,79711,79780,79928,79975,79995,80004,80013,80028,80153,80224,80347,80507,80579,80647,80687,80787,80867,80882,80928,80950,80959,80977,81051,81127,81212,81271,81373,81402,81436,81492,81507,81579,81635,81641,81738,81769,81780,81925,81981,82002,82022,82115,82156,82159,82167,82262,82375,82550,82645,82682,82752,83035,83129,83220,83227,83260,83266,83565,83831,83838,83840,83948,83964,83968,83991,84120,84216,84280,84357,84464,84666,84780,84830,84859,84900,84926,84953,84998,85117,85177,85232,85288,85358,85455,85649,85701,85733,85792,85830,85857,85868,85892,85943,85999,86180,86250,86559,86730,86732,86749,86750,86797,86906,87010,87011,87086,87091,87183,87234,87251,87414,87602,87643,87667,87689,87695,87705,87706,87727,87738,87744,87793,87867,87872,87938,87954,87961,87963,87978,87991,88116,88212,88225,88695,88739,88789,88986,89048,89252,89372,89439,89518,89546,89676,89791,89843,89855,89910,90070,90107,90215,90250,90380,90381,90495,90500,90548,90601,90634,90643,90677,90686,90725,90741,90824,91115,91148,91214,91337,91466,91565,91585,91594,91662,91679,91680,91696,91709,91743,91798,91845,91906,91919,92054,92056,92094,92148,92228,92281,92350,92399,92403,92436,92465,92532,92575,92811,92884,92979,93030,93102,93144,93162,93232,93268,93295,93411,93475,93496,93564,93655,93674,93832,93888,93923,94056,94124,94285,94332,94426,94438,94481,94502,94526,94563,94694,94715,94725,94954,95075,95343,95419,95448,95585,95635,95786,95823,95847,95959,96045,96175,96209,96213,96525,96661,96673,96886,96915,96918,96951,97035,97099,97144,97246,97322,97374,97398,97445,97623,97857,97914,97977,97983,98054,98150,98216,98252,98337,98457,98553,98583,98666,98706,98725,98741,98872,98927,99074,99224,99287,99546,99624,99759,99908",
          "57,59,219,294,413,603,920,1044,1105,1167,1345,1351,1462,1583,1634,1692,1770,1793,1843,1877,1924,1993,2036,2068,2140,2208,2295,2368,2389,2594,2653,2721,2741,2820,2930,2948,2966,3148,3190,3233,3261,3308,3345,3384,3474,3480,3539,3574,3646,3819,3892,3969,4073,4133,4208,4369,4406,4417,4441,4528,4571,4652,4825,4858,4859,4866,4994,5158,5185,5195,5446,5611,5993,6005,6092,6151,6219,6252,6326,6349,6375,6389,6457,6552,6663,6668,6677,6716,6740,6779,6929,6933,6986,6993,7144,7210,7222,7231,7361,7369,7497,7589,7593,7633,7689,7769,7811,7872,7933,7950,8030,8222,8237,8242,8288,8291,8293,8378,8439,8451,8459,8600,8696,8857,8869,8993,9186,9273,9288,9332,9337,9460,9480,9483,9641,9712,9754,9869,9876,9914,10013,10032,10047,10120,10146,10189,10214,10308,10383,10397,10461,10482,10575,10622,10649,10663,10781,10819,10941,10976,11034,11072,11134,11140,11179,11317,11351,11414,11469,11489,11530,11553,11568,11594,11633,11711,11733,11747,11893,12096,12167,12208,12351,12387,12563,12582,12625,12769,12829,12910,12944,13087,13090,13217,13228,13400,13502,13515,13595,13653,13824,13829,13902,13904,13977,14283,14295,14301,14490,14501,14524,14547,14562,14706,14803,14830,14929,14960,14961,14967,14998,15010,15028,15074,15077,15087,15139,15311,15379,15404,15442,15478,15479,15496,15669,15688,15689,15702,15766,15810,15835,15913,15944,15963,15977,15979,15985,15987,15993,16058,16067,16210,16222,16330,16398,16453,16620,16633,16659,16676,16735,16832,16837,16896,17057,17096,17158,17289,17615,17630,17639,17684,17791,17867,17883,18095,18110,18236,18240,18330,18352,18417,18461,18597,18643,18703,18824,18891,18917,18918,19012,19122,19402,19426,19492,19572,19574,19661,19841,19860,19880,19921,19951,19958,19970,20027,20062,20135,20147,20269,20310,20418,20589,20600,20738,20739,20815,21082,21152,21211,21339,21531,21571,21576,21595,21617,21658,21748,21816,21897,22046,22205,22303,22461,22463,22531,22543,22629,22720,22794,22830,22848,22875,22986,23072,23196,23391,23396,23637,23876,23898,23941,24057,24141,24144,24169,24186,24210,24211,24389,24399,24405,24480,24764,24862,24865,25117,25248,25317,25370,25397,25408,25644,25673,25750,25956,26008,26088,26188,26211,26352,26381,26389,26428,26443,26524,26525,26592,26701,26740,26803,26816,26859,26887,26889,27043,27111,27185,27227,27286,27293,27300,27340,27341,27347,27414,27532,27566,27616,27661,27710,27731,27915,27991,28002,28118,28173,28176,28273,28319,28413,28441,28478,28488,28545,28632,28719,28730,28892,28904,28910,28945,28972,29098,29136,29222,29238,29273,29344,29361,29410,29476,29575,29608,29652,29654,29836,29847,29995,30006,30008,30036,30076,30088,30120,30152,30212,30279,30416,30461,30523,30570,30709,30787,30927,30976,30984,31051,31052,31067,31112,31178,31222,31298,31399,31570,31573,31640,31682,31737,31811,31924,32022,32057,32154,32625,32668,32797,32938,32981,32997,33238,33330,33461,33473,33517,33539,33633,33649,33673,33714,33780,33846,33849,33935,33936,33945,34004,34097,34119,34160,34162,34211,34579,34620,34708,34749,34802,34823,34846,34914,35410,35435,35500,35560,35605,35607,35676,35878,35963,35985,36006,36053,36103,36110,36135,36486,36490,36622,36667,36735,36798,36830,37067,37075,37188,37216,37360,37375,37428,37449,37476,37603,37608,37629,37648,37739,37801,38042,38060,38076,38274,38276,38355,38408,38410,38489,38493,38592,38766,38771,38799,38801,39107,39116,39149,39245,39325,39361,39384,39471,39502,39525,39919,40039,40077,40212,40282,40353,40359,40405,40458,40468,40700,40707,40724,40802,40815,40900,40915,41004,41068,41158,41242,41291,41326,41342,41466,42031,42039,42206,42231,42247,42308,42325,42371,42423,42500,42506,42536,42896,42905,43119,43171,43285,43288,43389,43586,43612,43646,43684,43756,43994,44077,44237,44253,44259,44317,44350,44369,44557,44624,44734,44840,44946,45035,45188,45286,45320,45394,45449,45497,45594,45678,45751,45774,45857,45882,46042,46133,46199,46259,46458,46509,46550,46577,46595,46639,46656,46771,46885,47044,47064,47088,47141,47186,47228,47237,47446,47460,47626,47833,47936,48000,48014,48323,48471,48551,48589,48592,48677,48740,48808,48824,48846,48859,48927,48977,49162,49342,49450,49451,49485,49486,49511,49687,49774,49817,49820,49835,49949,50127,50399,50525,50701,50827,50932,51018,51027,51122,51126,51271,51277,51308,51417,51493,51746,51784,51788,51906,51908,51958,52019,52032,52039,52149,52558,52592,52727,52753,52773,52783,52993,53037,53074,53083,53212,53220,53293,53348,53394,53488,53651,53705,53706,53717,53720,53745,53747,53769,53770,53787,53854,54010,54022,54150,54291,54331,54613,54694,54780,54879,54885,54975,55083,55123,55250,55275,55333,55469,55472,55486,55550,55577,55591,55612,55657,55743,55875,55896,56013,56338,56345,56426,56466,56633,56692,56721,56754,56832,56940,56943,57038,57069,57214,57239,57265,57428,57508,57638,57732,57790,57804,57927,57944,57955,58065,58144,58379,58469,58620,58696,58919,59039,59068,59166,59249,59256,59406,59539,59688,59710,59711,59840,59904,60118,60392,60405,60458,60868,60920,60939,60952,61033,61043,61201,61255,61416,61459,61461,61562,61603,61651,61743,61757,61764,61828,61915,61996,62000,62073,62310,62434,62542,62642,62664,62671,62673,62683,62740,62830,62847,62901,62921,62960,63140,63324,63367,63386,63419,63475,63544,63652,63702,63841,64014,64062,64095,64155,64230,64252,64258,64303,64304,64444,64475,64516,64656,64777,64991,65418,65444,65508,65521,65554,65697,65699,65785,65795,65948,65993,66003,66020,66150,66187,66242,66243,66368,66382,66535,66536,66648,66662,66830,66904,67092,67182,67226,67323,67420,67494,67535,67567,67573,67705,67790,67792,67793,68032,68120,68145,68162,68330,68414,68464,68472,68507,68586,68623,68630,68717,68744,68842,68915,68935,68947,68964,68988,68993,69020,69021,69124,69129,69169,69219,69240,69294,69347,69363,69455,69465,69502,69551,69652,69822,69856,69941,69988,70152,70240,70323,70468,70571,70592,70977,70985,70986,71051,71085,71138,71151,71247,71327,71434,71548,71556,71564,71631,71829,71830,71908,72058,72122,72314,72416,72474,72505,72570,72653,72696,72704,72716,72738,72763,72834,72979,73188,73218,73229,73331,73332,73340,73540,73724,73786,73844,73874,73909,73970,73991,74324,74369,74388,74439,74537,74552,74692,74756,74792,74926,74947,74964,75034,75108,75157,75278,75358,75635,75729,75826,75876,75909,75968,76054,76104,76150,76204,76205,76229,76274,76286,76330,76359,76371,76376,76524,76610,76783,76824,76948,77007,77019,77050,77070,77133,77141,77271,77394,77464,77630,77698,77767,77914,77932,77952,78022,78074,78129,78200,78272,78301,78337,78351,78382,78411,78437,78498,78607,78631,78819,78863,79016,79109,79179,79208,79271,79279,79537,79658,79675,79686,79708,79732,79764,79777,79909,79950,79970,80022,80054,80086,80168,80213,80226,80326,80452,80591,80630,80722,81001,81033,81037,81044,81081,81107,81169,81269,81459,81528,81840,81853,81910,81941,82031,82117,82148,82155,82240,82248,82327,82505,82572,82587,82592,82606,82679,82785,82809,82932,83023,83024,83185,83284,83313,83392,83430,83445,83622,83737,83765,83783,83974,83992,84035,84199,84330,84490,84546,84576,84896,85032,85033,85158,85306,85329,85377,85418,85488,85650,86044,86172,86291,86329,86335,86337,86505,86542,86637,86641,86642,86686,86701,86805,86848,86892,86975,87017,87029,87071,87208,87327,87354,87501,87577,87594,87596,87616,87900,87945,87979,88036,88047,88178,88267,88306,88326,88520,88798,88853,88946,88987,89174,89186,89301,89329,89331,89563,89625,89669,89821,89850,89930,90016,90055,90064,90078,90120,90133,90213,90339,90404,90412,90442,90454,90702,90836,90839,90903,90906,90919,90921,90983,91015,91125,91508,91549,91560,91579,91581,91707,91925,92013,92132,92357,92361,92653,92656,92669,92736,92780,92886,92932,93435,93448,93488,93573,93580,93617,93767,93784,93842,94032,94092,94157,94219,94296,94309,94323,94367,94483,94596,94755,94772,94878,94892,94962,95058,95152,95176,95374,95422,95742,95848,95915,95981,95996,96101,96164,96305,96401,96429,96556,96793,96834,96848,96912,96934,96994,97164,97215,97442,97446,97496,97702,97711,97738,97780,97848,97869,98049,98070,98090,98156,98183,98246,98259,98271,98370,98410,98439,98451,98454,98555,98609,98610,98629,98822,98889,99077,99082,99086,99119,99123,99130,99208,99217,99237,99255,99270,99345,99409,99460,99793,99956",
          "20,37,102,156,185,186,719,859,876,973,997,1078,1106,1159,1273,1547,1799,1994,2014,2032,2064,2135,2175,2203,2461,2464,2599,2718,2732,2751,2917,2997,3086,3106,3154,3158,3201,3248,3387,3456,3590,3632,3675,3700,3922,3983,4082,4199,4244,4392,4558,4634,4811,4955,5028,5112,5292,5332,5363,5501,5512,5573,5651,5671,5711,5841,5885,5926,6004,6038,6263,6321,6440,6574,6725,6800,6866,6910,6962,7066,7294,7368,7373,7515,7569,7709,7857,8113,8271,8351,8406,8737,8855,8887,9011,9014,9070,9076,9128,9200,9238,9360,9409,9419,9424,9456,9467,9558,9602,9652,9732,9838,9840,9963,9998,10009,10012,10038,10173,10247,10421,10449,10510,10751,10769,10777,10810,10940,11024,11041,11099,11104,11284,11348,11411,11434,11480,11538,11627,11740,11803,11909,11961,11983,12131,12150,12288,12292,12337,12386,12436,12489,12491,12525,12568,12579,12604,12634,12722,12905,13055,13119,13218,13229,13286,13320,13547,13577,13635,13683,13713,13748,13756,13798,13849,13864,13908,13956,14056,14085,14086,14184,14258,14275,14398,14423,14430,14517,14618,14677,14712,14745,14785,15046,15096,15166,15201,15279,15395,15461,15497,15514,15573,15578,15589,15597,15617,15649,15820,15926,16003,16089,16102,16118,16157,16299,16386,16924,16927,16942,16966,16984,17042,17143,17160,17178,17242,17248,17250,17497,17521,17542,17795,17846,17870,18248,18305,18451,18493,18562,18592,18906,18907,19198,19323,19367,19411,19442,19482,19535,19632,19663,19717,19896,19905,19927,19953,19961,20037,20217,20226,20289,20343,20352,20375,20379,20465,20486,20516,20521,20688,20764,20803,20806,20809,20847,20998,21097,21135,21286,21335,21366,21539,21603,21609,21650,21832,21844,21885,21964,21970,22025,22074,22091,22145,22167,22168,22187,22297,22312,22368,22622,22745,22770,22810,22899,22946,23028,23045,23106,23134,23237,23579,23679,23757,23778,23792,23856,24071,24076,24193,24290,24433,24438,24494,24571,24587,24615,24668,24723,24763,24844,24881,25010,25016,25203,25241,25266,25341,25391,25425,25467,25721,25782,25930,25951,26081,26248,26469,26516,26604,26610,26628,26734,26798,26844,26870,26877,26931,27046,27074,27157,27177,27318,27366,27452,27494,27741,27984,28142,28192,28204,28236,28315,28584,28685,28802,28886,28901,29016,29266,29428,29466,29479,29536,29798,29799,29953,29994,30011,30017,30092,30271,30325,30367,30486,30540,30600,30620,30631,30644,31362,31371,31705,31744,31825,32038,32072,32352,32438,32652,32701,32971,33024,33036,33413,33419,33438,33587,33754,33840,33848,33886,33898,33914,33960,33998,34069,34147,34385,34484,34573,34639,34648,34713,34761,34763,34839,34968,35027,35145,35268,35343,35368,35695,35704,35768,36114,36127,36152,36218,36470,36564,36704,36752,36820,36828,36895,37055,37088,37162,37211,37400,37557,37601,37609,37652,37728,37850,37936,37948,37967,38008,38156,38237,38580,38617,38623,38639,38674,38700,38732,38829,38850,38915,39162,39222,39257,39312,39326,39362,39394,39727,39914,39916,39944,40018,40019,40132,40143,40145,40149,40176,40182,40234,40367,40537,40615,40690,40746,40786,40826,41079,41171,41223,41283,41346,41373,41440,41455,41696,41730,41776,41802,41945,41988,42180,42234,42436,42525,42577,42774,43390,43561,43570,43735,43885,43990,44058,44224,44277,44391,44446,44458,44461,44559,44757,44983,45038,45101,45108,45454,45530,45674,45718,45777,45888,45915,45956,46024,46064,46102,46155,46306,46313,46546,46556,46564,46571,46581,46600,46674,46733,46779,46799,46871,46987,47094,47098,47110,47199,47268,47290,47328,47400,47544,47567,47569,47637,47653,47696,47817,47847,47883,47894,48106,48147,48180,48264,48267,48336,48567,48640,48680,48728,48878,48923,48992,49012,49024,49199,49326,49428,49444,49641,49767,49970,49993,50015,50097,50102,50163,50211,50307,50510,50596,50826,50871,50945,51129,51292,51298,51366,51403,51423,51539,51840,51972,52135,52139,52172,52188,52257,52261,52323,52340,52379,52470,52766,52799,52852,53150,53268,53277,53323,53398,53413,53456,53523,53539,53613,53724,53841,54020,54053,54126,54196,54262,54288,54297,54300,54326,54450,54575,54604,54701,54728,54932,54953,55032,55098,55209,55324,55409,55465,55776,55807,56000,56088,56096,56223,56263,56279,56311,56476,56588,56683,56756,56758,56802,56850,56853,57028,57030,57048,57272,57346,57376,57405,57646,57843,57999,58037,58177,58217,58291,58339,58390,58446,58522,58538,58631,58697,58706,58749,59091,59126,59305,59315,59317,59548,59573,59574,59620,59691,59807,59811,59997,60086,60097,60114,60180,60194,60484,60497,60616,60644,60690,60732,60795,60944,60949,61017,61204,61273,61320,61345,61351,61417,61552,61555,61571,61801,61850,61982,62049,62052,62185,62316,62333,62359,62369,62442,62467,62480,62613,62615,62620,62643,62657,62751,62756,62765,62917,62925,62986,63028,63141,63194,63215,63296,63407,63520,63645,63654,63757,63826,64027,64053,64142,64160,64348,64525,64644,64691,64898,64959,65044,65114,65214,65224,65353,65360,65361,65405,65456,65664,65673,65773,65820,65825,65927,65971,66195,66216,66225,66258,66292,66622,66660,66766,66858,66995,67056,67059,67064,67133,67270,67314,67379,67422,67521,67569,67598,67649,67709,67782,67857,67975,68021,68050,68095,68188,68192,68241,68284,68538,68617,68646,68836,68978,69052,69080,69107,69119,69283,69300,69375,69391,69396,69434,69473,69522,69864,69919,69953,69992,70005,70060,70182,70197,70415,70491,70675,70707,70714,70792,71079,71211,71322,71371,71397,71433,71461,71505,71573,71642,71653,71686,71936,71995,72004,72023,72173,72389,72513,72525,72550,72552,72577,72670,72715,72801,72826,72894,72903,72991,73011,73026,73135,73154,73282,73283,73347,73348,73353,73423,73480,73670,73695,73731,73769,74062,74119,74147,74235,74444,74509,74555,74620,74680,74741,74813,74822,75005,75164,75207,75263,75270,75581,75651,75730,75821,75823,75830,76158,76184,76269,76272,76296,76367,76378,76408,76488,76511,76601,76682,76734,76840,76870,77028,77059,77183,77196,77357,77395,77484,77487,77550,77568,77577,77614,77668,77918,77994,77995,78009,78107,78116,78177,78406,78414,78448,78452,78524,78566,78624,78737,78773,78787,78845,78929,78946,78952,79012,79044,79052,79130,79147,79349,79365,79369,79414,79451,79543,79583,79730,79771,79787,79831,79857,79884,79932,79952,80036,80075,80410,80430,80462,80478,80483,80842,80843,80887,80927,80940,81186,81325,81422,81483,81614,81660,81711,81721,81857,81912,81934,81994,82093,82172,82329,82408,82468,82499,82753,83038,83048,83105,83277,83334,83671,83685,83745,83876,83934,83947,84149,84298,84311,84497,84561,84566,84587,84625,85018,85065,85190,85246,85340,85430,85737,85765,85971,85998,86086,86166,86242,86324,86482,86543,86836,87005,87094,87184,87194,87217,87232,87420,87551,87700,87720,87740,87810,87859,87860,88007,88023,88042,88070,88109,88287,88451,88591,88776,88911,88998,89489,89556,89593,89639,89643,89655,89995,90003,90011,90066,90236,90353,90476,90547,90571,90830,91435,91512,91721,91775,91844,92046,92103,92139,92181,92198,92227,92262,92299,92384,92404,92566,92583,92667,92678,92738,92841,92871,92978,93234,93327,93424,93526,93534,93802,93871,93929,93964,94027,94154,94256,94383,94403,94482,94488,94678,94719,94908,94968,95086,95274,95326,95413,95436,95507,95600,95642,95766,95800,95863,95898,95908,95913,96041,96073,96149,96150,96165,96214,96277,96289,96318,96494,96607,96679,96722,96864,97004,97024,97071,97107,97123,97158,97232,97434,97487,97612,97834,97906,97915,97925,97974,97981,98466,98605,98620,98876,99117,99445,99610,99633,99647,99730,99811,99924,99929,99933",
          "67,264,301,436,441,525,532,556,563,628,669,716,756,852,1003,1079,1175,1336,1417,1449,1602,1629,1769,1869,1955,1981,1996,2095,2146,2404,2421,2497,2586,2647,2677,2702,2771,2846,2908,2916,3070,3088,3180,3305,3354,3361,3400,3438,3624,3704,3755,3798,3813,3878,3958,4023,4107,4219,4308,4370,4439,4498,4509,4734,4737,4904,4940,5104,5306,5352,5546,5582,5851,5852,5914,6010,6048,6118,6160,6240,6317,6320,6430,6432,6612,6628,6655,6903,6935,6981,6991,7073,7094,7123,7184,7189,7382,7435,7460,7529,7810,7879,7905,7974,8048,8049,8067,8087,8146,8188,8199,8289,8345,8369,8521,8597,8669,8673,8703,8752,8761,8778,8836,8931,9025,9048,9089,9172,9269,9316,9341,9373,9546,9614,9646,9814,9964,10014,10056,10118,10426,10446,10491,10547,10556,10606,10682,10701,10776,10778,10801,10848,10927,10935,10952,10982,11036,11133,11142,11233,11256,11304,11407,11432,11708,11721,11920,11980,12122,12234,12255,12299,12327,12504,12600,12712,12838,12840,12933,12938,12968,12976,13040,13200,13240,13639,13692,13750,13812,13920,14034,14208,14257,14309,14337,14464,14629,14779,14838,14859,14940,14962,15042,15145,15178,15200,15254,15282,15289,15360,15492,15509,15525,15642,15747,15789,16083,16087,16104,16119,16196,16231,16320,16393,16555,16631,16696,16728,16827,16973,17126,17145,17147,17218,17269,17295,17313,17392,17474,17499,17508,17650,17740,17755,17782,17790,17812,17914,18175,18226,18327,18341,18421,18465,18609,18623,18659,18661,18752,18791,18794,18844,18861,18879,18900,18952,19134,19135,19146,19291,19316,19340,19355,19449,19451,19531,19576,19600,19799,19879,19886,19911,19913,20079,20104,20122,20159,20264,20323,20466,20505,20612,20728,20758,20761,20814,21017,21093,21193,21274,21332,21400,21415,21478,21486,21516,21651,21664,21685,21804,21827,21998,22028,22049,22081,22140,22185,22217,22366,22586,22643,22742,22919,23022,23220,23245,23248,23280,23398,23403,23497,23543,23597,23649,23668,23810,24242,24245,24578,24786,24871,24968,25123,25150,25172,25216,25369,25371,25510,25563,25629,25643,25702,25703,25741,26140,26270,26353,26403,26442,26457,26470,26519,26609,26616,26733,26888,26927,27031,27044,27133,27151,27175,27238,27272,27324,27346,27435,27463,27470,27556,27648,27709,27723,27747,27757,27784,27841,27958,28106,28189,28195,28199,28255,28410,28564,28568,28675,28761,28769,28825,28911,28913,29005,29020,29134,29148,29159,29204,29281,29312,29315,29470,29612,29636,29657,29766,29819,29956,29961,29991,30024,30110,30210,30254,30329,30383,30467,30593,30668,30759,30812,30882,30924,30962,31016,31023,31037,31076,31284,31363,31382,31419,31464,31490,31502,31625,31639,31658,31754,31862,31930,32001,32112,32160,32165,32396,32443,32469,32509,32538,32588,32616,32749,32756,32765,32819,32853,32879,33079,33081,33115,33214,33358,33588,33626,33668,33671,33695,33759,33795,33802,33805,33863,33933,34044,34140,34156,34330,34359,34398,34437,34516,34578,34604,34680,34688,34788,34880,34954,35009,35044,35114,35171,35315,35342,35456,35537,35559,35690,35697,35787,35882,35894,35980,36044,36227,36321,36383,36438,36502,36591,36632,36741,36750,36804,36829,36978,37149,37223,37418,37423,37615,37777,37800,37813,37878,37932,37939,37957,38205,38233,38295,38326,38346,38382,38409,38724,38762,38816,38978,38984,39125,39219,39290,39330,39378,39397,39417,39457,39565,39658,39678,39692,39717,39784,39859,39924,39929,39936,39961,40095,40197,40200,40205,40352,40374,40456,40460,40598,40617,40628,40645,40763,40881,40897,40995,41161,41164,41302,41323,41448,41712,41719,41747,41771,42006,42075,42277,42337,42407,42447,42594,42602,42645,42693,42799,42860,42877,42878,42923,42951,43083,43216,43234,43377,43407,43488,43501,43526,43641,43728,43729,43786,43815,43882,43902,43943,44198,44442,44505,44651,44736,44826,44881,44891,44971,45037,45098,45120,45127,45133,45155,45360,45425,45475,45609,45688,45825,45953,46023,46057,46114,46150,46285,46307,46359,46448,46574,46693,46738,46780,46813,46994,47128,47134,47215,47252,47331,47348,47417,47468,47519,47598,47642,47787,47905,47935,48116,48146,48196,48199,48207,48435,48526,48590,48645,48662,48706,49028,49145,49158,49319,49402,49459,49575,49587,49609,49613,49771,49773,49781,49831,49903,49962,50266,50338,50358,50505,50700,50819,50851,50992,51024,51050,51051,51353,51389,51392,51676,51732,51773,51950,52103,52113,52126,52536,52600,52822,52898,53021,53156,53233,53303,53304,53447,53477,53526,53529,53570,53579,53586,53604,53652,53711,53819,53927,54282,54303,54339,54380,54386,54400,54605,54631,54666,54674,54723,54815,54843,54864,54930,55100,55110,55128,55177,55231,55234,55242,55262,55264,55304,55428,55457,55501,55713,55780,55863,55867,55872,55986,56009,56048,56063,56228,56294,56347,56389,56412,56492,56574,56693,56814,56993,57116,57124,57175,57381,57446,57493,57501,57522,57526,57792,57839,57949,58250,58272,58292,58455,58466,58473,58568,58674,58677,58700,58705,58834,58853,58873,58903,59092,59123,59291,59427,59474,59528,59571,59648,59727,59810,59836,59837,59875,59919,60025,60124,60192,60208,60251,60315,60479,60481,60640,60663,60778,60819,60857,60866,60938,61042,61107,61213,61305,61450,61474,61482,61504,61532,61577,61684,61693,61790,61905,61956,61974,61986,62009,62014,62096,62102,62168,62195,62214,62269,62374,62485,62508,62530,62566,62630,62639,62698,62743,62949,63040,63042,63067,63166,63200,63232,63259,63270,63289,63391,63460,63513,63561,63588,63591,63626,63634,63662,63686,63762,63794,63836,63863,63977,64060,64153,64167,64187,64212,64301,64362,64433,64579,64720,64773,64823,64835,64859,64905,64911,64961,65041,65104,65166,65411,65460,65483,65507,65529,65754,65760,65821,65910,65946,65959,65961,65982,66055,66084,66129,66228,66533,66567,66610,66686,66711,66749,66823,66841,66880,66902,66923,66944,66954,67013,67060,67154,67234,67242,67401,67429,67461,67562,67621,67813,67865,67898,67912,67930,67945,67996,68046,68079,68128,68163,68197,68216,68332,68352,68397,68514,68542,68543,68567,68590,68599,68618,68620,68628,68663,68713,68902,68994,69103,69138,69209,69329,69560,69730,69998,70091,70164,70173,70328,70428,70443,70472,70526,70688,70756,70811,70864,70871,70936,70937,70962,71043,71195,71207,71402,71430,71452,71514,71597,71616,71635,71683,71871,71924,71981,72160,72204,72306,72386,72401,72419,72541,72548,72624,72672,72701,72705,72714,72877,73075,73080,73096,73151,73208,73285,73323,73362,73494,73706,73737,73772,73790,73904,73923,74020,74072,74180,74200,74356,74387,74535,74574,74629,74675,74843,74903,75019,75022,75035,75151,75181,75261,75304,75339,75388,75408,75493,75557,75565,75580,75582,75654,75708,76100,76122,76154,76180,76253,76281,76377,76437,76481,76546,76580,76697,76817,76835,76867,77129,77253,77383,77522,77620,77703,77708,77743,77748,77782,77878,77980,78028,78034,78158,78159,78423,78445,78602,78622,78644,78679,78817,78980,79371,79424,79518,79525,79604,79616,79639,79666,79684,79750,79791,79980,80035,80184,80292,80340,80708,80764,80783,80876,80967,81080,81119,81171,81209,81267,81281,81390,81464,81473,81529,81580,81594,81601,81613,81650,81743,81779,81908,81930,81985,82173,82183,82216,82249,82270,82303,82370,82585,82597,82744,82778,82818,82999,83065,83104,83144,83204,83366,83386,83422,83427,83572,83581,83687,83691,83695,83706,83793,83944,83955,84019,84082,84092,84385,84459,84485,84555,84734,84771,84781,84782,84803,84807,84886,84889,84909,85068,85069,85070,85139,85192,85264,85269,85302,85513,85532,85617,85627,85694,85856,85878,85891,85903,86093,86189,86212,86272,86276,86322,86367,86432,86460,86481,86530,86666,86847,86852,86877,87046,87182,87360,87467,87566,87592,87650,87661,87719,87752,87768,87781,87806,87812,87825,88050,88187,88260,88411,88539,88604,88757,88829,88846,89135,89268,89291,89355,89361,89484,89627,89661,89682,89739,89765,89866,89869,89870,89888,89918,89963,89983,90025,90148,90594,90733,90757,90769,90920,91030,91053,91206,91323,91464,91634,91794,91839,91859,91873,91917,91931,91994,92071,92383,92489,92503,92509,92514,92645,92690,92740,92909,92916,93009,93034,93161,93322,93364,93402,93404,93406,93428,93429,93509,93604,93628,93772,93794,93867,94045,94101,94228,94279,94316,94355,94361,94409,94468,94555,94662,94819,94889,94989,94991,95021,95138,95308,95316,95372,95450,95518,95716,95757,95764,95811,95814,95938,95940,95968,95984,95993,96147,96186,96419,96455,96469,96483,96492,96529,96726,96741,96830,96983,96987,97001,97138,97300,97368,97411,97443,97539,97677,97766,97912,98089,98119,98199,98446,98518,98528,98582,98595,98739,98843,98926,98932,98941,98993,98999,99039,99066,99076,99078,99098,99271,99307,99327,99332,99431,99442,99492,99497,99797,99909,99952,99960",
          "105,159,207,235,364,587,640,743,789,1054,1067,1174,1177,1200,1205,1309,1514,1659,1728,1767,1805,1845,1910,1950,1958,1959,1964,1971,1975,2022,2028,2099,2132,2178,2238,2255,2273,2300,2327,2357,2359,2475,2496,2506,2540,2549,2553,2556,2622,2627,2642,2663,2886,2909,2951,2994,3079,3097,3220,3231,3269,3277,3312,3377,3465,3473,3613,3667,3713,3734,3823,3898,3937,3987,4021,4043,4078,4087,4094,4116,4117,4139,4149,4239,4267,4268,4274,4296,4325,4356,4363,4459,4481,4506,4534,4581,4605,4702,4746,4767,4864,4880,4910,4921,5345,5390,5396,5457,5464,5551,5556,5600,5706,5740,6335,6351,6438,6452,6494,6528,6811,6825,6843,6875,6883,6888,6945,6992,7350,7425,7437,7767,7901,8245,8281,8684,8809,8831,8861,8863,8966,9022,9107,9141,9241,9362,9375,9399,9507,9650,9669,9692,9725,9860,9870,10007,10015,10101,10103,10202,10220,10258,10304,10428,10605,10761,10792,10881,10896,11003,11016,11112,11113,11159,11180,11294,11299,11320,11328,11399,11415,11418,11523,11603,11739,11788,11843,12046,12135,12246,12305,12309,12316,12374,12412,12487,12572,12673,12783,12847,13014,13032,13065,13120,13185,13276,13439,13576,13645,13765,13952,13993,14001,14011,14136,14145,14152,14278,14317,14323,14334,14507,14663,14678,14717,14768,14791,14793,14853,14910,14943,14946,14950,14992,15114,15135,15217,15226,15332,15389,15414,15439,15588,15606,15623,15666,15737,15844,15922,15951,15953,15982,16151,16169,16174,16246,16248,16306,16316,16417,16439,16536,16651,16681,16720,16793,16956,17271,17317,17323,17339,17465,17469,17574,17610,17659,17704,17783,17800,17898,17956,18008,18045,18196,18234,18298,18375,18589,18653,18780,18798,18852,18978,18983,18984,19002,19019,19190,19243,19338,19344,19407,19483,19540,19550,19703,19718,19774,19824,19853,19856,19876,19884,20034,20090,20160,20233,20351,20481,20547,20551,20610,20646,20725,20794,20834,20944,20999,21021,21094,21132,21155,21177,21208,21209,21210,21256,21322,21328,21374,21455,21500,21594,21875,21879,21899,22072,22131,22133,22192,22213,22237,22299,22340,22345,22374,22380,22448,22680,22760,22836,22859,22869,22938,22988,23009,23246,23302,23348,23357,23423,23534,23549,23646,23648,23704,23766,23777,23781,24108,24192,24226,24264,24416,24440,24475,24551,24581,24593,24630,24656,24685,24700,24767,24797,24916,24929,24930,24964,25043,25044,25091,25165,25168,25200,25232,25244,25264,25294,25399,25429,25435,25484,25541,25612,25733,25747,25759,25878,25888,25914,25921,25927,25937,25953,26006,26020,26049,26058,26089,26152,26170,26190,26230,26316,26322,26337,26368,26380,26407,26413,26415,26498,26508,26526,26551,26573,26672,26676,26726,26776,26786,26854,26865,26890,26908,26911,26949,26960,26968,26969,26984,27080,27100,27179,27212,27280,27319,27322,27365,27368,27415,27461,27476,27483,27509,27524,27598,27618,27630,27721,27742,27780,27789,27825,27886,27895,27999,28019,28027,28062,28069,28115,28121,28212,28301,28320,28342,28357,28689,28775,28903,29000,29027,29082,29097,29102,29109,29219,29279,29299,29314,29318,29320,29349,29406,29418,29452,29504,29687,29707,29719,29853,29882,29989,30013,30033,30037,30091,30119,30167,30225,30244,30321,30373,30387,30394,30400,30429,30483,30568,30632,30713,30754,30797,30838,31004,31008,31015,31022,31068,31088,31110,31161,31183,31287,31403,31442,31537,31564,31582,31692,31725,31726,31769,31815,31919,32084,32140,32190,32228,32367,32456,32507,32522,32529,32689,32873,32935,32987,33003,33091,33119,33136,33183,33277,33294,33321,33442,33505,33555,33583,33632,33682,33719,33733,33828,33841,33842,33859,33906,33981,34000,34016,34053,34067,34080,34142,34153,34171,34173,34297,34329,34355,34451,34469,34529,34561,34596,34608,34653,34654,34660,34694,34703,34715,34727,34795,34925,34985,35041,35076,35084,35127,35189,35206,35253,35272,35290,35382,35421,35468,35499,35529,35663,35832,35860,35913,36043,36057,36059,36085,36137,36174,36180,36185,36200,36231,36248,36310,36315,36325,36366,36373,36436,36487,36578,36596,36657,36673,36677,36942,36950,36963,36968,36970,36994,36995,37061,37111,37152,37174,37273,37294,37309,37338,37347,37373,37394,37469,37486,37572,37600,37687,37792,37797,37808,37880,37986,38001,38040,38048,38055,38072,38102,38104,38121,38165,38200,38219,38221,38302,38319,38328,38329,38445,38462,38472,38545,38564,38569,38658,38758,38804,38890,38995,39044,39052,39142,39144,39158,39185,39187,39188,39254,39289,39317,39354,39456,39473,39500,39521,39557,39609,39616,39638,39649,39679,39746,39792,39815,39871,39931,39935,39951,39993,40034,40064,40091,40093,40151,40623,40804,40838,40922,40927,40943,41071,41109,41162,41212,41367,41383,41393,41402,41406,41408,41451,41595,41861,41868,41936,41984,42067,42078,42093,42114,42126,42130,42166,42252,42330,42421,42451,42515,42555,42576,42583,42638,42663,42691,42792,42840,42846,42971,42986,43028,43065,43109,43192,43254,43572,43577,43589,43610,43621,43652,43661,43833,43951,44108,44137,44223,44315,44357,44386,44399,44417,44431,44562,44569,44579,44662,44711,44761,44869,44908,44930,44986,45085,45174,45247,45264,45298,45310,45311,45340,45350,45490,45491,45495,45529,45535,45621,45684,45687,45739,45750,45824,45837,45917,45963,45978,45993,46002,46003,46082,46125,46127,46245,46353,46402,46452,46461,46579,46668,46687,46721,46763,46803,46807,46850,46852,46864,46895,46964,46984,46991,47210,47307,47321,47439,47451,47505,47632,47739,47752,47784,47800,47909,47949,48081,48087,48125,48144,48225,48236,48250,48305,48311,48326,48352,48367,48379,48449,48481,48499,48908,48946,49053,49084,49121,49346,49384,49385,49457,49565,49681,49753,49794,49964,50035,50059,50060,50066,50086,50262,50267,50428,50631,50689,50720,50734,50833,50885,50894,50936,51080,51114,51270,51363,51442,51483,51572,51613,51615,51685,51699,51805,51922,51943,52007,52084,52086,52175,52199,52262,52327,52332,52351,52376,52467,52492,52667,52717,52927,53192,53244,53252,53285,53358,53435,53479,53530,53616,53647,53758,53980,54071,54163,54313,54430,54434,54437,54713,54847,55062,55073,55109,55134,55137,55241,55298,55338,55394,55402,55426,55431,55556,55614,55720,56059,56165,56203,56221,56383,56449,56789,57051,57178,57179,57185,57194,57317,57328,57479,57482,57486,57516,57541,57612,57733,57803,57892,58049,58102,58229,58340,58397,58418,58444,58527,58531,58669,58758,58830,58843,58844,58889,58964,58967,58983,58996,59107,59274,59387,59401,59419,59538,59641,59708,59724,59749,59773,59805,59878,59884,59952,59988,60153,60557,60840,60965,61045,61056,61137,61144,61199,61265,61341,61358,61385,61470,61488,61492,61655,61739,61780,61807,62027,62078,62110,62144,62228,62266,62351,62353,62355,62360,62871,63076,63128,63151,63279,63405,63443,63451,63516,63703,63798,63811,63839,63843,63949,64017,64032,64039,64239,64363,64365,64379,64442,64467,64490,64708,64780,64781,64840,64848,64864,64899,64918,64938,65053,65163,65173,65382,65526,65547,65596,65791,65805,65832,65853,65930,65949,66122,66125,66176,66263,66522,66598,66608,66672,66772,66844,66882,66969,67002,67024,67032,67088,67206,67209,67399,67404,67464,67665,67738,67767,67891,67950,67953,68019,68037,68091,68107,68411,68417,68629,68693,68783,68826,68829,68887,68945,68950,69159,69233,69316,69420,69487,69594,69625,69696,69767,69773,69810,69897,69899,69935,69936,69991,70000,70058,70073,70100,70104,70111,70242,70466,70475,70480,70608,70613,70647,70665,70732,70789,70878,70987,71048,71167,71168,71275,71341,71358,71382,71447,71460,71465,71562,71715,71797,71961,72162,72189,72262,72396,72408,72504,72553,72614,72721,72728,72891,73074,73113,73385,73470,73488,73578,73686,73831,73922,73959,74166,74207,74265,74302,74360,74400,74515,74737,74766,74809,74953,74971,74973,74982,74989,75251,75273,75327,75447,75502,75519,75578,75647,75835,75861,75892,75928,75960,76241,76346,76350,76398,76446,76492,76646,76721,76845,76896,76958,76986,77042,77047,77295,77390,77497,77648,77658,77905,77917,77992,78131,78154,78276,78341,78421,78429,78570,78682,78725,78747,78748,78785,78829,78885,78908,78944,78997,79106,79133,79224,79391,79410,79494,79533,79634,79673,79769,79908,79949,80033,80277,80278,80356,80372,80426,80465,80499,80622,80623,80662,80683,80689,80811,80952,80971,81162,81172,81287,81304,81465,81530,81559,81603,81682,81701,82010,82160,82186,82197,82263,82291,82315,82378,82531,82560,82586,82680,82784,82838,82885,82983,83248,83259,83458,83517,83589,83698,83778,83802,83856,84124,84151,84266,84409,84499,84514,84604,84626,84699,84772,84809,84899,84988,84994,85030,85104,85206,85241,85337,85352,85425,85602,85690,85718,85748,85755,85762,85908,85965,85988,86036,86046,86108,86193,86214,86282,86289,86352,86436,86467,86488,86550,86615,86660,86685,86747,86780,86849,86995,87106,87139,87201,87203,87206,87350,87430,87435,87458,87471,87701,87725,87735,87786,87899,88034,88056,88094,88096,88124,88412,88456,88556,88558,88621,88637,88705,88746,88843,88899,88927,88989,89185,89187,89213,89443,89473,89479,89495,89531,89535,89580,89582,89675,89689,89690,89756,89842,90018,90042,90166,90172,90175,90260,90283,90294,90390,90449,90605,90639,90720,90754,90762,90819,90856,90947,90951,90993,91056,91059,91064,91093,91103,91154,91160,91245,91320,91398,91404,91482,91570,91657,91795,91862,91883,91960,92087,92146,92166,92310,92415,92471,92515,92694,93093,93165,93207,93262,93326,93328,93425,93539,93595,93676,93683,93894,93995,94348,94349,94416,94509,94515,94521,94538,94636,94836,94969,94976,95039,95100,95114,95136,95289,95294,95323,95341,95412,95417,95449,95545,95586,95687,95733,95734,95769,95846,95939,95951,95955,96342,96374,96396,96397,96490,96696,96786,97142,97226,97402,97464,97521,97546,97720,98057,98139,98197,98207,98354,98470,98634,98690,98703,98708,98858,98892,98946,99089,99091,99210,99225,99230,99261,99357,99619,99631,99646,99704,99882,99899,100000",
          "4,14,142,250,307,325,468,499,502,552,593,637,995,1430,1494,1925,1949,2042,2122,2158,2202,2247,2326,2422,2433,2524,2648,2707,2724,2863,3026,3050,3094,3118,3274,3362,3501,3728,3766,3811,3911,3961,4029,4174,4278,4284,4326,4390,4427,4497,4525,4604,4617,4791,4988,5040,5058,5140,5172,5194,5340,5349,5443,5469,5670,5782,5867,5985,6189,6201,6257,6381,6402,6433,6546,6569,6593,6638,6805,6836,6865,6927,6973,6982,6995,7084,7135,7453,7480,7493,7598,7636,7652,7748,7846,7904,8019,8311,8321,8379,8545,8750,8786,8883,9010,9042,9286,9365,9484,9494,9587,9612,9868,10021,10281,10297,10400,10497,10540,10631,10721,10767,10773,10831,10917,11119,11146,11148,11212,11345,11451,11552,11575,11696,11750,11781,11818,12194,12251,12282,12379,12770,12818,12870,12887,12928,12955,12992,13138,13166,13263,13449,13460,13478,13516,13536,13623,13651,13706,13743,13834,13842,13861,13881,14005,14014,14017,14027,14029,14036,14164,14209,14242,14262,14294,14346,14385,14513,14564,14598,14664,14736,14795,14814,14991,15064,15125,15154,15171,15190,15244,15305,15364,15452,15647,15767,15776,15861,15924,16208,16224,16673,16677,16699,16897,17114,17117,17251,17345,17363,17384,17421,17617,17774,17816,17821,17825,17932,17951,18036,18279,18286,18370,18471,18805,18830,19016,19151,19167,19169,19195,19275,19320,19349,19463,19530,19642,19673,19681,19807,19871,19916,19972,20033,20144,20192,20199,20241,20296,20584,20593,20631,20648,20652,20664,20722,20749,20766,20805,21051,21054,21442,21549,21562,21728,22165,22177,22210,22267,22329,22400,22468,22475,22605,22746,22757,22795,22796,22866,22883,22909,22956,22969,22994,23223,23400,23480,23510,23698,23733,23787,23791,23813,24137,24184,24385,24432,24515,24627,24669,24808,24854,24999,25365,25394,25437,25453,25566,25575,25604,25624,25688,25758,25768,25772,25917,25994,26071,26101,26118,26314,26401,26432,26728,26761,26780,26910,26963,27055,27084,27137,27229,27526,27608,27702,27764,27776,27777,27792,27802,27829,28040,28249,28329,28345,28405,28539,28616,28641,28745,28789,28880,28938,28968,28991,29083,29208,29251,29276,29340,29366,29396,29519,29549,29699,29727,29807,30084,30116,30339,30368,30386,30396,30501,30727,30796,31086,31168,31280,31348,31361,31390,31398,31500,31539,31594,31830,32071,32147,32235,32253,32353,32404,32463,32521,32527,32536,32619,32638,32654,32782,32795,32846,32863,32898,32900,32931,33037,33256,33493,33520,33522,33751,33810,33816,33885,33900,34062,34279,34515,34676,34701,34779,34988,35000,35070,35185,35238,35252,35396,35447,35557,35807,35824,35874,35925,36020,36151,36431,36482,36558,36629,36812,36916,37096,37113,37227,37364,37456,37535,37554,37663,37732,37825,37832,37944,37954,38095,38105,38209,38225,38459,38519,38529,38882,38960,39167,39396,39415,39527,39540,39769,39960,39980,39984,39999,40000,40167,40258,40296,40470,40590,40601,40620,40752,40818,40829,40844,40877,40935,41183,41195,41416,41473,41480,41555,41608,41774,41812,41919,41982,42035,42154,42178,42229,42235,42248,42286,42292,42342,42532,42657,42664,42729,42782,42797,42863,42904,42917,42998,43002,43076,43299,43408,43459,43470,43494,43695,43727,43890,43965,44080,44084,44196,44474,44661,44690,44714,44759,44786,44807,44924,44972,45008,45185,45250,45258,45551,45695,45884,45925,45965,46017,46021,46105,46192,46239,46250,46365,46434,46582,46612,46642,46654,46751,46934,46992,47113,47114,47126,47209,47213,47286,47315,47461,47662,47742,47811,47895,47978,48092,48622,48626,48748,48786,48870,48930,48971,49182,49226,49244,49312,49371,49435,49447,49469,49506,49553,49579,49702,49735,49854,49861,50100,50153,50178,50206,50245,50294,50295,50336,50461,50515,50575,50604,50744,50845,50985,51151,51229,51272,51326,51487,51520,51552,51718,51787,51820,51843,51944,52114,52192,52380,52485,52504,52606,52793,52864,53111,53254,53264,53284,53518,53679,53806,53825,53961,54128,54136,54187,54361,54463,54483,54593,54669,54708,54762,54834,54884,55067,55246,55312,55533,55539,55675,55689,55715,56027,56148,56175,56195,56237,56460,56570,56644,56658,56808,56878,57131,57134,57402,57546,57592,57628,57649,57836,57921,57930,57959,57978,57997,58064,58091,58136,58276,58368,58737,58759,58788,58817,58820,58945,59189,59288,59294,59371,59389,59547,59624,59651,59795,59928,60177,60239,60263,60312,60636,60809,60814,60875,61319,61354,61357,61613,61678,61777,61803,61824,61862,61900,61922,61959,61975,62023,62067,62114,62212,62245,62251,62410,62528,62558,62578,62647,62994,63020,63106,63113,63157,63248,63392,63546,63607,63724,63785,63795,63845,63916,63917,64030,64117,64411,64489,64505,64513,64871,64989,64996,65010,65117,65315,65317,65330,65514,65574,65750,65819,65846,65955,66461,66548,66624,66776,66792,66849,66899,67098,67369,67435,67559,67597,67788,67810,67963,68040,68049,68067,68090,68094,68164,68461,68474,68553,68602,68624,68634,68641,68665,68678,68724,68893,68897,68962,69321,69381,69418,69461,69475,69600,69637,69640,69662,69676,69688,69920,69930,69975,70202,70258,70308,70438,70481,70790,70869,70870,71071,71179,71261,71292,71517,71539,71543,71960,72021,72044,72094,72646,72845,72890,72914,72942,73057,73152,73207,73242,73516,73656,73661,73720,73802,73834,73887,74098,74281,74306,74507,74527,74532,74634,74643,74650,74701,74713,74824,74831,74992,75055,75087,75198,75346,75351,75375,75520,75855,75916,76095,76096,76121,76197,76324,76385,76412,76472,76973,76990,76997,77255,77304,77860,77984,78199,78233,78235,78470,78641,78744,78761,78935,79077,79144,79404,79430,79566,79778,79805,79826,80090,80095,80136,80181,80405,80456,80679,80766,80770,80841,80892,80941,80958,81011,81041,81043,81070,81141,81190,81345,81347,81429,81470,81487,81537,81542,81552,81571,81649,81702,81766,81774,81847,81854,81955,82028,82052,82058,82076,82122,82200,82206,82235,82245,82491,82580,82687,82741,82760,82967,83202,83303,83324,83337,83472,83525,83549,83721,83738,83748,83769,83898,83942,83988,84137,84150,84236,84331,84359,84402,84463,84616,84675,84681,84742,84794,84922,84938,84940,84942,84947,84956,84965,85162,85164,85507,85546,85579,85588,85612,85726,85785,85790,85832,85887,85946,86375,86472,86480,86555,86608,86754,86810,86837,86856,86954,86982,87394,87489,87579,87697,87745,87751,87774,87819,88083,88132,88170,88298,88402,88485,88488,88694,88765,88972,89076,89302,89358,89365,89406,89418,89634,89720,89796,89885,90004,90164,90338,90375,90419,90552,90560,90749,90755,90792,90858,90876,90950,91009,91128,91167,91258,91357,91371,91489,91927,92119,92374,92526,92594,92665,92723,92786,92861,93076,93121,93220,93221,93300,93318,93362,93409,93483,93493,93642,93727,93829,93848,93878,93895,94042,94164,94211,94281,94287,94334,94405,94421,94431,94661,94697,94714,94823,95074,95154,95239,95276,95288,95353,95359,95531,95538,95550,95583,95627,95668,95782,95821,96037,96096,96241,96244,96457,96465,96544,96610,96659,96738,96776,96841,96867,97033,97181,97249,97286,97371,97383,97450,97502,97598,97626,97631,97672,97685,97846,97873,97957,97970,98045,98105,98135,98145,98186,98191,98312,98564,98663,98715,98797,98907,98937,99102,99211,99815,99872,99920,99961",
          "23,146,196,222,234,318,474,509,543,651,728,775,965,982,1049,1068,1095,1501,1503,1541,1628,1653,1704,1754,1941,2011,2071,2156,2239,2271,2303,2304,2307,2546,2698,2727,2793,2924,3008,3114,3218,3391,3441,3492,3532,3602,3686,3800,3902,3966,4007,4069,4148,4391,4452,4462,4522,4681,4744,4745,4786,4805,4977,5105,5123,5146,5174,5199,5208,5238,5450,5508,5827,6044,6747,6798,6857,6905,6999,7017,7104,7125,7148,7226,7352,7526,7642,7695,7819,7830,7918,7981,7996,8107,8240,8363,8385,8425,8488,8602,8678,8697,8700,8771,8920,8971,8981,9567,9651,9659,9663,9694,9737,9747,9769,9947,10000,10058,10087,10112,10162,10171,10185,10205,10227,10316,10352,10445,10644,10670,10900,10922,11127,11208,11266,11273,11277,11283,11302,11382,11534,11610,11693,11731,11770,11799,11806,11808,11860,11981,12070,12073,12088,12119,12152,12259,12341,12380,12443,12606,12665,12767,12784,12909,12954,13062,13085,13086,13359,13370,13481,13497,13679,13766,13820,13831,13867,13869,13882,13958,14046,14063,14112,14134,14141,14153,14178,14219,14225,14280,14288,14328,14375,14379,14447,14567,14602,14624,14653,14689,14703,14723,14746,14769,14781,14893,14993,15007,15059,15066,15086,15121,15194,15300,15342,15365,15402,15453,15474,15522,15544,15714,15717,15783,15843,15846,15930,15965,15997,16049,16082,16113,16129,16155,16211,16297,16334,16395,16440,16511,16535,16564,16616,16674,16701,16738,16818,16853,16856,16894,17017,17115,17247,17264,17447,17481,17516,17577,17701,17710,17925,17937,17938,17972,18006,18082,18112,18365,18403,18423,18638,18811,18813,18820,18822,18859,18878,18904,18905,18997,19087,19175,19215,19242,19290,19420,19440,19487,19488,19496,19560,19603,19644,19710,19711,19792,19829,19842,19974,19993,19996,20010,20028,20180,20228,20256,20283,20288,20339,20373,20404,20464,20675,20713,20772,20799,20844,20896,20931,21020,21136,21167,21187,21278,21300,21323,21341,21598,22016,22062,22105,22134,22211,22369,22399,22433,22485,22509,22614,22661,22708,22711,22783,22845,22887,23018,23076,23114,23128,23172,23175,23205,23258,23290,23375,23455,23460,23572,23705,23790,23799,23829,23860,23882,23938,23948,24016,24039,24160,24213,24239,24476,24478,24508,24540,24624,24657,24703,24835,24872,24980,24981,25009,25083,25088,25097,25108,25199,25211,25328,25330,25436,25438,25451,25544,25597,25618,25620,25654,25714,25749,25756,25774,25839,25850,25957,26293,26334,26384,26393,26410,26412,26558,26693,26805,26873,26901,27000,27004,27040,27214,27269,27332,27430,27446,27462,27468,27515,27531,27553,27932,28127,28197,28304,28327,28354,28360,28481,28572,28647,28657,28681,28779,28906,28909,28956,29041,29043,29117,29210,29269,29468,29571,29665,29772,29774,29868,29926,30140,30223,30228,30390,30395,30422,30437,30479,30490,30629,30645,30647,30735,30739,30784,30813,30814,30827,30872,30895,30981,30989,31066,31075,31160,31180,31316,31333,31388,31436,31466,31600,31603,31740,31886,31953,31991,31999,32060,32099,32208,32292,32294,32459,32590,32598,32621,32758,32835,32848,32866,32890,32899,32933,32951,33006,33069,33209,33213,33378,33390,33415,33558,33618,33698,33792,33873,33942,33967,33994,34056,34079,34104,34120,34245,34247,34358,34491,34525,34552,34652,34662,34665,34811,34962,35089,35117,35126,35247,35310,35379,35440,35508,35553,35665,35709,36082,36083,36092,36179,36188,36246,36384,36459,36579,36606,36611,36674,36705,36728,36739,36915,37098,37169,37197,37287,37332,37337,37391,37508,37553,37708,37713,37765,37783,37827,37834,37856,37895,38115,38185,38214,38327,38434,38499,38500,38750,38763,38859,38933,39054,39058,39100,39103,39131,39305,39340,39358,39697,39726,39745,39756,39810,39861,39964,40059,40063,40088,40094,40142,40333,40350,40365,40373,40469,40529,40556,40564,40647,40679,41015,41084,41134,41138,41218,41272,41276,41280,41329,41392,41417,41502,41652,41688,41718,41754,41882,41911,41959,41965,42059,42072,42253,42389,42399,42414,42597,42628,42653,42745,42759,42766,42886,42913,43056,43106,43298,43326,43335,43356,43403,43431,43569,43677,43696,43705,43759,43765,43770,43838,43895,43931,44293,44302,44400,44470,44638,44694,44817,44884,44948,45073,45148,45231,45246,45285,45336,45344,45442,45504,45514,45582,45615,45699,45764,45946,46094,46110,46184,46205,46262,46300,46429,46466,46476,46491,46551,46607,46816,46867,46868,46897,46922,46943,47016,47283,47327,47355,47391,47445,47475,47501,47554,47591,47599,47647,47689,47690,47763,47959,48070,48090,48192,48230,48241,48271,48453,48507,48602,48679,48894,48928,48983,49183,49237,49335,49380,49393,49603,49772,49856,49867,49896,49914,49989,50010,50085,50122,50142,50226,50271,50334,50424,50632,50653,50759,50790,50810,50815,50899,51082,51171,51205,51241,51325,51331,51430,51489,51490,51532,51642,51723,51954,51964,51985,52000,52147,52320,52353,52471,52490,52597,52611,52705,52769,52835,52918,52967,53050,53110,53346,53352,53384,53429,53462,53481,53558,53567,53587,53626,53629,53821,53865,53979,54041,54165,54180,54451,54805,54980,55054,55059,55113,55193,55201,55213,55257,55297,55405,55424,55505,55606,55654,55696,55719,55789,55804,55988,56030,56098,56184,56234,56269,56357,56551,56622,56700,56708,56776,56907,56921,56946,56963,57025,57057,57076,57111,57144,57205,57353,57420,57439,57577,57744,57761,57772,57847,57874,57980,58022,58061,58356,58394,58409,58452,58564,58567,58577,58596,58598,58678,58756,58813,58825,58870,58886,58891,59053,59085,59158,59162,59179,59205,59258,59285,59289,59324,59454,59520,59550,59565,59628,59771,59802,59849,59861,59877,59890,59892,59939,59961,59972,59981,59982,60003,60070,60163,60191,60279,60388,60429,60430,60450,60528,60746,60777,60825,61151,61158,61169,61197,61227,61248,61268,61314,61480,61529,61544,61586,61617,61848,62086,62106,62208,62336,62655,62881,63011,63180,63209,63234,63236,63465,63598,63614,63972,63987,64184,64198,64265,64380,64543,64575,64641,64696,64772,64807,64856,64930,64988,65045,65048,65155,65250,65373,65388,65394,65406,65474,65513,65683,65783,65893,65950,65991,66023,66131,66387,66495,66513,66537,66584,66646,66656,66883,67216,67221,67274,67437,67588,67666,67706,67804,67964,68008,68017,68242,68265,68492,68636,68699,68810,69081,69143,69208,69258,69384,69425,69441,69499,69500,69521,69666,69725,69987,69995,70215,70245,70260,70398,70450,70497,70575,70593,70623,70705,70872,70879,70917,70997,71036,71226,71323,71424,71665,71697,71767,71951,72006,72120,72244,72377,72393,72623,72628,72634,72711,72744,72765,72920,73032,73278,73395,73467,73605,73628,73639,73683,73692,73748,73884,74055,74132,74140,74165,74243,74489,74554,74616,74928,74952,74970,75042,75111,75187,75224,75227,75266,75423,75570,75597,75604,75659,75662,75711,75757,75920,75930,75954,76077,76107,76156,76232,76236,76277,76339,76356,76413,76421,76596,76652,76793,76842,76850,76921,76930,77131,77310,77378,77396,77427,77446,77532,77562,77653,77710,77809,77819,77960,77976,78063,78068,78228,78313,78417,78489,78564,78584,78822,78861,78943,78982,79269,79301,79312,79360,79375,79411,79477,79545,79682,79709,79807,79832,79862,79910,79927,79983,80146,80151,80199,80287,80370,80384,80419,80467,80494,80570,80601,80853,80901,80922,80970,80998,81053,81069,81089,81194,81224,81283,81478,81554,81586,81634,81709,81739,81765,81811,81834,81961,82027,82131,82152,82153,82224,82247,82340,82416,82570,82673,82695,82776,82883,82888,82981,83081,83166,83252,83258,83264,83297,83434,83604,83645,83653,83725,84142,84192,84245,84252,84277,84412,84456,84478,84511,84515,84571,84577,84684,84710,84731,84776,84815,84821,84884,84948,84954,84985,85002,85031,85124,85282,85283,85422,85527,85581,85585,85638,85641,85735,85739,85797,85831,85841,85993,86013,86050,86428,86430,86578,86585,86707,86755,86943,86949,86960,87001,87031,87032,87197,87271,87318,87328,87366,87370,87445,87527,87542,87590,87648,87652,87679,87778,87866,87912,88355,88372,88477,88486,88616,88762,88770,88842,88864,88883,89068,89148,89225,89320,89868,90045,90089,90161,90238,90312,90389,90475,90512,90559,90619,90621,90696,90778,90797,90818,90866,90887,90907,91153,91317,91422,91445,91463,91528,91641,91740,91748,91827,91838,92116,92141,92184,92344,92410,92460,92461,92487,92539,92613,92698,92764,92802,92829,92859,92898,92949,93039,93066,93077,93200,93276,93288,93524,93538,93576,93583,93616,93641,93713,93759,93825,93845,93984,94058,94223,94317,94359,94444,94672,94795,94851,95069,95073,95139,95172,95194,95380,95426,95752,95781,95942,95975,96119,96181,96236,96278,96354,96409,96411,96478,96684,96701,96749,96774,96799,97025,97114,97268,97280,97311,97403,97485,97630,97639,97851,97856,97927,97965,98004,98019,98051,98118,98208,98403,98465,98501,98506,98544,98572,98597,98868,98909,98956,98985,98994,99054,99088,99150,99250,99310,99337,99433,99444,99481,99734,99749,99806,99845,99944",
          "27,42,79,189,287,566,705,739,916,945,978,1061,1170,1214,1219,1377,1593,1673,1920,2001,2186,2278,2286,2318,2370,2376,2440,2468,2477,2563,2824,2999,3105,3184,3189,3351,3417,3506,3761,3774,3887,3890,3951,4039,4110,4185,4201,4247,4248,4352,4455,4508,4511,4582,4589,5125,5243,5247,5350,5351,5406,5432,5531,5532,5547,5549,5607,5610,5758,5868,5886,6009,6232,6284,6310,6405,6421,6501,6557,6667,6787,6812,6817,6861,6969,7003,7053,7121,7132,7166,7185,7196,7444,7479,7525,7922,7927,8099,8131,8156,8185,8266,8424,8641,8754,8873,8926,8965,9039,9077,9080,9117,9151,9183,9240,9519,9540,9626,9996,10031,10132,10288,10343,10372,10658,10782,10904,10955,11005,11047,11098,11158,11301,11353,11460,11796,11819,11931,11998,12003,12024,12079,12089,12102,12172,12287,12301,12404,12500,12624,12658,12788,12922,12995,13059,13133,13216,13258,13354,13438,13528,13649,13720,13830,13856,13870,14033,14207,14211,14298,14537,14708,14869,14917,14959,15014,15119,15172,15447,15465,15523,15528,15547,15683,15746,15760,15937,15991,16160,16277,16284,16367,16514,16560,16587,16635,16683,16978,17070,17097,17149,17177,17183,17192,17198,17273,17330,17364,17507,17643,17719,17776,17860,18101,18224,18342,18415,18445,18715,18745,18770,18790,18843,18948,18987,18991,19097,19130,19205,19401,19579,19618,19678,19685,19791,19902,19929,20075,20220,20230,20292,20427,20533,20572,20588,20730,20786,20802,20836,20841,20971,21085,21316,21346,21498,21613,21721,21765,21925,22048,22053,22092,22201,22279,22315,22320,22383,22429,22619,22640,22683,22690,22765,22864,23019,23077,23308,23481,23612,23662,23673,23681,23773,23783,23838,23980,24035,24080,24283,24292,24296,24434,24467,24568,24631,24793,24868,24963,24966,25127,25381,25558,25662,25867,25922,26047,26082,26180,26222,26232,26236,26310,26446,26591,26799,26814,26863,26880,26997,27071,27377,27422,27433,27467,27557,27561,27579,27672,27693,27767,27899,27998,28006,28083,28366,28433,28486,28665,28833,28875,28879,28987,29019,29039,29227,29343,29373,29574,29625,29642,29876,30034,30202,30382,30463,30471,30533,30575,30624,30938,30979,31002,31140,31151,31171,31179,31205,31409,31451,31529,31752,31969,31992,32017,32069,32333,32447,32464,32537,32569,32584,32617,32772,32787,32888,32945,33045,33072,33107,33275,33403,33573,33772,33804,33868,33889,34036,34055,34363,34380,34392,34536,34598,34629,34759,34800,34838,34918,34961,35051,35062,35074,35150,35214,35235,35254,35297,35449,35736,35791,35792,35798,35818,35849,35969,36078,36091,36095,36176,36191,36212,36223,36280,36345,36394,36430,36563,36565,36625,36727,36848,36878,37027,37092,37279,37401,37590,37736,37775,37791,37854,37869,37989,38027,38216,38312,38357,38534,38542,38587,38677,38903,38914,38996,39123,39153,39752,39857,39968,39978,40084,40240,40395,40644,40708,40739,40771,40968,41107,41159,41179,41349,41443,41479,41547,41563,41573,41575,41654,41846,42064,42128,42176,42401,42584,42682,42735,42862,42914,42962,43040,43128,43145,43208,43232,43309,43322,43548,43562,43658,43666,43667,43719,43781,43953,44004,44012,44028,44134,44160,44161,44211,44250,44299,44398,44469,44499,44591,44700,44811,44823,44937,44938,45011,45034,45169,45317,45353,45355,45377,45382,45390,45427,45635,45660,45712,45816,45903,45904,45986,46076,46187,46197,46750,46792,46882,47145,47164,47221,47310,47337,47480,47486,47503,47534,47582,47590,47747,47862,47983,48043,48126,48171,48181,48211,48217,48278,48292,48537,48547,48599,48619,48682,48704,48750,48813,48895,49060,49142,49578,49590,49629,49647,49665,49708,49728,49758,49830,49834,49873,49887,50135,50196,50199,50315,50361,50556,50698,50882,50981,51029,51170,51244,51472,51476,51511,51537,51546,51620,51672,51759,52092,52107,52186,52204,52309,52407,52484,52494,52576,52605,52642,52861,52893,52933,52953,53086,53098,53135,53140,53315,53461,53476,53516,53590,53630,53702,53834,53837,54073,54171,54200,54209,54243,54445,54534,54732,54775,54829,54886,54937,54997,55092,55114,55203,55299,55309,55564,55597,55734,55770,55778,55812,55853,55916,55952,55961,56123,56128,56130,56172,56181,56212,56231,56267,56302,56511,56735,56888,56939,56997,57292,57338,57412,57477,57500,57613,57701,57904,57924,58112,58150,58191,58261,58416,58503,58563,58809,58856,58924,59070,59394,59480,59786,59787,59790,59863,59872,59949,60085,60199,60362,60774,60806,60863,61050,61127,61130,61224,61262,61274,61360,61396,61397,61462,61534,61796,61846,61940,61972,61973,62015,62019,62066,62136,62281,62342,62484,62519,62759,62884,62900,62905,63049,63121,63272,63365,63395,63406,63476,63709,63717,63781,63813,63832,63842,63878,63980,64003,64005,64054,64328,64383,64396,64410,64716,64757,64975,65017,65148,65175,65182,65229,65259,65327,65356,65416,65487,65537,65622,65633,65685,65749,65804,65908,66019,66066,66163,66254,66414,66445,66527,66543,66632,66675,66696,66735,66818,66915,66916,66919,67076,67112,67264,67281,67291,67465,67519,67542,67594,67616,67636,67684,67816,67853,67924,67978,67988,68245,68253,68430,68673,68738,68748,68777,68786,68891,68939,69000,69006,69167,69234,69469,69577,69836,69860,70134,70193,70246,70540,70621,70656,70776,70935,70956,70961,71053,71298,71356,71363,71394,71411,71526,71528,71530,71549,71572,71618,71787,71883,71912,71997,72038,72046,72284,72477,72579,72629,72712,72760,72815,72852,73062,73082,73302,73709,73718,73735,73745,73807,73865,73899,74108,74283,74331,74339,74380,74478,74491,74578,74623,74825,74835,74854,74859,74908,75067,75090,75128,75175,75183,75220,75248,75253,75482,75569,75731,76016,76024,76060,76093,76255,76307,76590,76611,76665,76703,76741,76799,76808,76843,77002,77006,77149,77207,77267,77308,77754,78055,78210,78251,78413,78535,78638,78710,78847,78887,78913,79000,79191,79214,79310,79361,79364,79413,79433,79663,79763,79822,79876,80037,80079,80193,80198,80243,80311,80345,80358,80481,80501,80673,80712,80729,80837,80925,80991,81094,81362,81408,81592,81687,81756,81770,81778,81791,81923,81926,82096,82409,82530,82644,82756,82781,82825,82829,82941,82957,82979,83219,83234,83236,83247,83339,83363,83385,83516,83558,83801,83878,84089,84110,84188,84201,84259,84389,84401,84434,84452,84507,84567,84608,84640,84662,84787,84888,84924,84925,85081,85135,85176,85194,85218,85239,85240,85310,85413,85483,85551,85578,85647,85681,85757,85783,85847,85972,86016,86305,86328,86330,86369,86523,86532,86549,86725,86738,86801,86881,87098,87296,87301,87338,87384,87453,87518,87536,87598,87699,87836,87915,88179,88181,88321,88503,88518,88551,88706,88743,88955,89003,89084,89103,89120,89145,89192,89211,89395,89568,89716,89763,89807,89849,89856,90012,90048,90057,90068,90144,90221,90401,90536,90542,90543,90606,90669,90800,90889,90926,91011,91021,91032,91152,91207,91375,91420,91480,91487,91537,91563,91619,91667,91685,91912,92157,92173,92218,92249,92260,92267,92453,92939,92996,93142,93245,93347,93633,93659,93725,93795,93906,94035,94138,94301,94530,94531,94580,94674,94722,95082,95120,95237,95254,95301,95482,95523,95669,95692,95731,95762,95805,95947,95960,96185,96193,96238,96326,96385,96388,96425,96468,96486,96500,96538,96740,96814,96902,97016,97234,97344,97350,97366,97509,97676,97699,97714,97726,97801,97968,98072,98189,98282,98328,98475,98678,98729,98749,98753,98779,98794,99002,99009,99084,99141,99175,99268,99274,99283,99598,99656,99715,99867,99958",
          "15,257,273,290,299,329,379,398,431,444,446,449,459,518,537,649,718,748,792,794,1023,1098,1317,1330,1433,3926,5006,5031,5042,5119,5166,5177,5190,5216,5231,5263,5281,5298,5302,5348,5365,5367,5383,5411,5447,5482,5489,5528,5774,5871,5924,5932,6000,6024,6069,6076,6361,6366,6409,6419,6538,6627,6689,6777,6804,6874,6922,7036,7160,7462,7591,7712,7818,7971,8034,8089,8138,8164,8297,8361,8663,8698,8821,9016,9046,9207,9213,9393,9422,9635,10026,10224,12132,12264,12617,12879,13592,15608,15712,16353,16501,17368,17376,18758,21686,22717,24718,26794,30476,32904,35694,50057,50306,50402,50407,50409,50485,50518,50735,50920,51036,51044,51093,51188,51369,51382,51439,51491,51521,51574,51742,51806,52191,52240,52279,52373,52394,52397,52429,52498,52557,52695,52707,52762,52767,52844,52845,52851,52909,52956,52981,53035,53062,53108,53123,53153,53217,53306,53428,53522,53546,53713,53730,53757,53761,53856,53869,53878,53896,53899,54161,54201,54597,54617,54737,54763,54804,55024,55040,55049,55091,55125,55188,55205,55224,55253,55433,55446,55452,55604,55625,55748,55784,55843,55912,55948,56011,56090,56336,56497,56542,56596,56597,56656,56970,57045,57095,57104,57110,57523,57543,57573,57697,57784,57889,58003,58209,58303,58314,58319,58377,58526,58528,58617,58685,58746,58766,58791,58863,58933,59088,59201,59211,59232,59431,59464,59662,59707,59935,59948,59993,60109,60165,60206,60269,60320,60371,60377,60393,60401,60443,60495,60558,60618,60984,61067,61083,61104,61478,61561,61715,61726,61793,61822,62076,62119,62505,62975,63117,63216,63478,63521,63575,63669,63677,63690,63713,63726,63810,63831,63850,63889,63957,64046,64071,64073,64272,64292,64296,64368,64407,64544,64719,64754,64776,64915,64978,65165,65203,65206,65254,65289,65316,65337,65368,65434,65566,65618,65772,65799,65874,65933,66015,66026,66076,66094,66181,66191,66424,66458,66525,66557,66599,66769,66835,66942,67000,67156,67163,67275,67488,67538,67627,67783,67871,67882,67893,67904,67959,67976,68202,68362,68560,68563,68601,68659,68779,68924,68982,69066,69068,69238,69259,69284,69335,69348,69350,69421,69517,69542,69708,69732,69808,69918,69939,69951,70026,70044,70198,70225,70595,70771,70796,70814,70858,71014,71100,71220,71439,71533,71615,71804,71823,71863,72053,72061,72232,72407,72532,72533,72557,72611,72649,72752,72802,72929,72932,73050,73156,73162,73187,73219,73365,73390,73422,73492,73662,73697,74015,74382,74411,74476,74512,74545,74652,74677,74805,74846,74873,74884,75085,75150,75544,75586,75741,75760,75791,75901,75903,75924,76019,76057,76294,76338,76340,76529,76736,76838,76848,76868,76871,76943,76944,77171,77281,77311,77320,77325,77460,77478,77558,77605,77640,77676,77773,77840,77972,78021,78097,78345,78349,78372,78469,78491,78629,78657,78728,78890,78894,78984,79059,79143,79229,79268,79293,79356,79357,79393,79444,79579,79649,79762,79784,79859,79873,79881,79957,80051,80053,80072,80125,80375,80408,80437,80502,80605,80890,80912,81014,81018,81090,81118,81153,81199,81233,81621,81728,81747,81763,81792,81814,81862,81864,81888,82048,82063,82091,82353,82411,82458,82464,82501,82831,82929,82994,83162,83360,83505,83617,83820,84097,84317,84543,84646,84733,84775,84837,84854,84877,84878,84906,84929,85196,85208,85321,85368,85462,85536,85571,85577,85746,85905,86004,86037,86097,86236,86384,86426,86487,86524,86570,86724,86834,86931,87024,87123,87264,87325,87567,87637,87690,87757,87864,88148,88264,88279,88347,88376,88378,88407,88428,88474,88601,88777,88835,88944,88945,89013,89031,89122,89136,89636,89695,89738,89747,89824,89982,89996,90122,90278,90344,90572,90596,90628,90917,90945,91281,91308,91554,91586,91693,92021,92044,92109,92301,92321,92329,92659,92697,92706,92713,92750,92817,92858,93036,93099,93143,93267,93489,93499,93547,93608,93741,93803,93889,93973,94000,94313,94441,94507,94600,94633,94742,94777,94827,94838,95092,95169,95264,95305,95429,95444,95537,95661,95696,95880,95896,95937,96179,96299,96430,96446,96526,96562,96723,97018,97038,97040,97103,97172,97175,97214,97221,97321,97365,97386,97449,97754,97760,97813,97818,97917,97937,98300,98369,98540,98584,98594,98637,98787,98790,98949,99007,99106,99200,99248,99279,99490,99509,99600,99682,99820,99851,99854,99869,99955,99970,99988",
          "92,164,225,376,465,491,544,591,732,802,885,958,1100,1238,1319,1401,1418,1453,1485,1502,1504,1608,1622,1753,1848,1870,1917,2049,2091,2102,2105,2108,2123,2172,2361,2399,2467,2512,2543,2568,2601,2851,3016,3129,3161,3173,3280,3300,3304,3326,3328,3446,3490,3522,3764,3824,3912,3996,4010,4098,4158,4313,4338,4368,4415,4479,4529,4612,4621,4763,4796,5089,5091,5098,5113,5198,5270,5470,5540,5590,5725,5837,5958,6019,6042,6087,6224,6414,6604,6647,6681,6934,6936,6956,7171,7203,7219,7343,7390,7523,7747,7791,7813,7842,7924,7956,8038,8053,8119,8166,8167,8277,8355,8375,8447,8499,8515,8537,8650,8651,8758,9029,9034,9111,9147,9262,9285,9515,9532,9542,9611,9615,9647,9760,9834,9905,9977,10124,10179,10233,10286,10399,10451,10477,10538,10542,10576,10581,10584,10748,10841,10989,11023,11166,11315,11425,11426,11462,11540,11566,11573,11589,11599,11606,11655,11682,11691,11718,11792,11798,11844,11883,11896,11945,11950,12037,12039,12177,12296,12445,12567,12781,12924,12977,13048,13080,13088,13094,13107,13108,13330,13342,13442,13667,13697,13763,13923,13929,13968,14055,14074,14197,14218,14247,14274,14527,14688,14731,14802,14844,15133,15140,15146,15213,15256,15356,15440,15446,15541,15546,15570,15583,15673,15729,15792,15939,15955,15989,16006,16036,16111,16339,16428,16476,16548,16612,16714,16835,17077,17090,17128,17234,17290,17394,17418,17486,17504,17530,17607,17622,17629,17671,17676,17750,17757,17899,17919,18016,18168,18201,18230,18287,18393,18522,18524,18571,18672,18675,18701,18743,18821,18881,18895,18898,18931,18933,19044,19084,19183,19217,19285,19427,19434,19547,19566,19647,19684,19689,19705,19870,20066,20249,20317,20384,20539,20900,20908,20922,20946,21024,21047,21143,21175,21252,21412,21533,21544,21799,21842,22021,22232,22235,22387,22524,22549,22773,22825,22963,23188,23197,23224,23249,23351,23479,23498,23532,23587,23590,23752,23864,24042,24119,24163,24171,24194,24263,24315,24482,24544,24549,24560,24570,24702,24712,24734,24859,24915,25051,25333,25344,25366,25367,25375,25519,25640,25668,25669,25871,25882,26182,26185,26245,26267,26425,26598,26716,26771,26787,26823,26832,26842,26920,26933,26945,26982,27146,27186,27204,27206,27289,27304,27412,27527,27587,27615,27646,27708,27713,27750,27753,27796,27869,27912,27948,27997,28059,28064,28096,28102,28104,28157,28335,28391,28398,28455,28658,28763,28857,28905,28983,28999,29003,29023,29037,29058,29084,29145,29362,29403,29488,29517,29771,29891,29904,29955,29959,30022,30044,30131,30138,30213,30238,30266,30281,30435,30447,30465,30484,30503,30528,30583,30639,30653,30771,30788,30805,31191,31250,31251,31302,31357,31578,31584,31586,31647,31751,31842,31845,31850,32007,32134,32271,32348,32426,32429,32451,32530,32554,32567,32609,32658,32881,32928,33201,33215,33410,33454,33459,33563,33579,33907,33978,34463,34504,34651,34689,34737,34775,34928,34959,34978,35037,35258,35291,35295,35311,35353,35386,35471,35513,35600,35614,35672,35755,35771,35823,35846,35884,35970,36004,36056,36067,36144,36186,36278,36290,36313,36458,36514,36545,36592,36648,36801,36826,36831,36843,36856,36914,36969,36990,37022,37052,37212,37263,37286,37370,37380,37409,37509,37537,37548,37607,37631,37632,37836,37875,38050,38143,38159,38239,38363,38407,38496,38600,38637,38876,38891,38898,38906,38963,39004,39134,39157,39173,39224,39277,39295,39322,39395,39432,39524,39681,39702,39795,39834,39967,39970,40048,40252,40301,40305,40411,40453,40536,40551,40572,40593,40622,40686,40718,40736,40969,40999,41043,41063,41099,41144,41273,41388,41421,41438,41836,41952,41967,41974,42014,42023,42194,42226,42424,42495,42601,42647,42660,42702,42723,42834,42939,42981,43046,43271,43316,43348,43381,43609,43654,43665,43709,43782,43783,43856,43877,43930,43963,44010,44014,44032,44227,44584,44642,44692,44760,44793,44830,44940,45041,45057,45062,45150,45304,45349,45414,45473,45520,45552,45632,45669,45683,45706,45721,45736,45878,45899,45907,45932,45991,45994,46055,46056,46072,46191,46234,46275,46282,46284,46386,46389,46416,46445,46530,46608,46611,46661,46836,46927,47157,47272,47285,47385,47616,47702,47712,47737,47757,47777,47920,47972,47973,48032,48118,48149,48154,48245,48282,48344,48451,48458,48552,48562,48629,48634,48765,48867,49091,49105,49229,49313,49355,49669,49700,49719,49726,49752,50050,50107,50154,50337,50360,50540,50577,50597,50830,50869,50877,50897,50997,51109,51148,51328,51427,51432,51440,51515,51543,51715,51778,51779,51800,51807,51818,51835,52196,52430,52449,52618,52636,52760,52772,52823,52862,53060,53070,53122,53240,53310,53335,53380,53594,53725,53990,54035,54055,54059,54113,54129,54335,54344,54421,54503,54715,54797,54909,54935,54968,55382,55432,55454,55589,55600,55706,55752,56066,56135,56188,56249,56272,56320,56344,56360,56537,56580,56667,56703,56730,56791,56827,56919,56968,57000,57018,57071,57140,57163,57301,57363,57407,57447,57547,57676,57764,57798,57815,57816,57846,57911,58058,58098,58129,58169,58240,58242,58445,58499,58521,58727,58735,58777,59004,59119,59134,59334,59413,59525,59615,59745,59770,59839,59956,60087,60143,60179,60277,60343,60439,60510,60541,60591,60713,60719,60756,60943,61002,61074,61432,61618,61662,61690,61752,61901,61904,62494,62539,62543,62694,62775,62805,62839,62846,63006,63034,63102,63118,63149,63159,63264,63328,63375,63402,63564,63565,63701,63749,63771,63856,63858,63898,63928,64008,64045,64084,64123,64173,64299,64639,64673,64706,64769,64873,64939,64972,64987,65001,65143,65188,65263,65287,65304,65481,65550,65627,65690,65694,65774,65884,65902,65996,66036,66160,66256,66377,66406,66435,66568,66876,66926,66967,67033,67124,67145,67197,67213,67256,67279,67298,67371,67502,67541,67566,67759,67761,67795,67808,67827,67892,67980,67987,68068,68111,68229,68235,68333,68406,68647,68689,68758,68881,68911,68983,68997,69017,69075,69123,69146,69231,69232,69274,69276,69349,69378,69527,69528,69550,69809,69949,70030,70057,70138,70333,70368,70452,70479,70535,70676,70721,70759,70761,70829,70836,71037,71049,71145,71147,71154,71425,71482,71531,71603,71610,71679,71698,71793,71794,71894,71923,71973,72072,72099,72130,72200,72246,72484,72536,72610,72633,72673,72902,72950,72975,73058,73195,73226,73231,73367,73442,73483,73548,73564,73698,73743,73760,73861,74058,74096,74133,74596,74613,74648,74704,74794,74826,74904,75021,75156,75257,75353,75489,75506,75529,75541,75673,75700,75768,75778,75800,75801,75874,76046,76166,76201,76234,76271,76300,76435,76521,76592,76635,76796,76825,76882,76957,77008,77014,77038,77092,77120,77261,77282,77328,77398,77418,77431,77471,77510,77530,77557,77667,77677,77685,77777,77827,77857,78082,78083,78400,78451,78516,78620,78774,78792,78830,78897,78934,78938,79062,79095,79097,79099,79118,79225,79304,79428,79432,79445,79479,79554,79570,79627,79733,79850,79870,80048,80203,80228,80234,80330,80505,80575,80685,80751,80804,80846,80933,80989,81176,81205,81520,81526,81675,81782,81918,82066,82100,82178,82265,82346,82352,82366,82446,82451,82635,82770,82782,82783,82870,83049,83084,83188,83206,83294,83350,83535,83548,83710,83717,83882,84078,84158,84254,84268,84333,84353,84441,84482,84504,84565,84746,84767,84881,85265,85287,85303,85367,85406,85416,85437,85469,85615,85784,85874,85888,86047,86162,86164,86294,86376,86386,86404,86419,86471,86554,86565,86621,86633,86664,86691,86866,86914,86984,86986,87004,87126,87127,87165,87343,87376,87483,87498,87502,87555,87613,87636,87882,87909,87941,88068,88082,88091,88111,88139,88188,88210,88226,88325,88501,88537,88569,88600,88647,88718,88796,88816,88836,88851,88880,89058,89089,89109,89118,89308,89375,89416,89522,89552,89729,89853,89904,89939,89974,89984,89989,90021,90052,90342,90459,90477,90489,90513,90515,90664,91332,91452,91598,91840,91918,91920,91940,92033,92155,92234,92235,92298,92475,92485,92597,92734,92742,92856,93012,93172,93376,93430,93451,93618,93631,93942,94063,94091,94167,94362,94480,94491,94782,95064,95080,95121,95123,95174,95193,95197,95397,95439,95524,95628,95708,95790,95831,96039,96215,96235,96269,96293,96441,96489,96580,96712,96713,96751,96777,96795,96852,96938,97153,97160,97168,97267,97308,97343,97377,97404,97500,97543,97549,97636,97666,97678,97703,97861,97882,97891,97913,98281,98317,98342,98357,98523,98625,98658,98660,98817,98899,98943,99026,99105,99147,99162,99220,99249,99312,99424,99513,99532,99533,99581,99625,99641,99686,99708,99751,99861,99888,99994",
          "28,47,520,522,678,974,1122,1135,1171,1191,1277,1404,1486,1487,1765,1855,1865,1872,2027,2060,2100,2106,2136,2179,2290,2352,2374,2398,2436,2447,2495,2650,2704,2729,2767,2896,2935,3003,3009,3073,3146,3375,3395,3471,3557,3591,3600,3643,3645,3655,3723,3780,3787,3812,3864,3886,3971,4006,4016,4019,4054,4102,4122,4141,4195,4234,4287,4300,4315,4330,4436,4467,4483,4603,4716,4719,4830,4838,4957,5074,5357,5358,5429,5581,6073,6161,6377,6449,6646,6662,6692,6792,6915,6940,6953,7086,7206,7240,7243,7254,7475,7656,7692,7714,7915,7955,7999,8037,8039,8101,8179,8285,8305,8392,8458,8532,8852,8906,8918,8986,9098,9276,9355,9407,9435,9468,9561,9591,9698,9707,9709,9795,9804,9889,10018,10100,10167,10287,10326,10382,10414,10417,10468,10487,10587,10642,10696,10850,10969,10980,11031,11032,11153,11181,11202,11211,11274,11364,11365,11371,11377,11497,11521,11541,11654,11730,11801,11949,11993,12091,12128,12175,12235,12261,12403,12497,12578,12591,12609,12793,12839,12973,13147,13153,13188,13394,13512,13609,13797,13838,13872,14154,14271,14272,14347,14372,14376,14377,14405,14437,14449,14620,14750,14751,14797,14804,14828,14874,15002,15065,15067,15082,15199,15257,15270,15307,15335,15382,15405,15434,15441,15444,15581,15610,15626,15816,15818,15833,15959,16037,16097,16121,16128,16134,16138,16263,16266,16341,16567,16606,16621,16687,16830,16889,16921,16937,17079,17091,17110,17230,17246,17326,17331,17356,17473,17767,17780,17874,17958,18001,18070,18134,18146,18163,18204,18220,18233,18311,18318,18335,18348,18360,18427,18499,18583,18650,18663,18683,18706,18769,18771,18774,18921,18924,19000,19008,19023,19060,19108,19360,19605,19606,19653,19697,19843,19983,19989,20067,20078,20133,20134,20181,20193,20251,20262,20309,20356,20367,20371,20469,20498,20556,20607,20709,20759,20780,20798,20929,20952,20957,21131,21156,21158,21159,21272,21369,21432,21677,21697,21706,21828,21874,21894,21936,22023,22036,22040,22095,22123,22171,22180,22188,22224,22444,22449,22481,22490,22491,22507,22547,22554,22726,22761,22923,23014,23029,23129,23144,23221,23259,23343,23381,23417,23434,23461,23463,23467,23502,23563,23654,23715,23734,23737,23761,23811,23936,23967,23972,23978,24083,24167,24252,24318,24323,24339,24450,24492,24543,24563,24579,24599,24678,24686,24804,24897,25133,25147,25185,25249,25296,25304,25336,25352,25573,25659,25660,25728,25780,25817,25819,25864,25899,25973,26009,26114,26122,26133,26166,26197,26286,26303,26430,26459,26474,26478,26529,26543,26560,26640,26699,26791,26831,26849,26875,26929,27035,27174,27360,27464,27781,27803,27894,28113,28133,28159,28238,28311,28312,28322,28438,28490,28504,28630,28690,28751,28765,28767,28794,28853,28923,28992,29054,29108,29158,29332,29368,29474,29568,29620,29628,29677,29720,29912,29998,30087,30170,30240,30286,30426,30547,30556,30592,30614,30688,30833,30839,31006,31081,31108,31132,31149,31164,31189,31426,31450,31514,31554,31688,31696,31715,31733,31742,31791,31796,31804,31817,31895,31912,31936,31941,32028,32116,32187,32225,32236,32256,32291,32301,32303,32340,32366,32414,32423,32452,32482,32541,32582,32637,32656,32757,32798,32905,33010,33052,33167,33234,33254,33266,33304,33338,33355,33362,33392,33469,33496,33514,33528,33595,33610,33628,33636,33732,33749,33921,33937,33944,34073,34154,34310,34404,34411,34476,34533,34550,34593,34597,34732,34743,34785,34786,34853,34865,34993,34997,35023,35024,35026,35061,35100,35107,35108,35122,35125,35132,35141,35215,35256,35267,35276,35304,35319,35363,35407,35419,35434,35465,35492,35495,35606,35702,35817,35828,35858,35887,35891,35919,35940,35977,36112,36130,36140,36156,36202,36300,36333,36364,36403,36406,36503,36520,36531,36699,36772,36779,36794,36817,36845,36881,36884,36952,36973,37003,37029,37050,37091,37124,37161,37168,37281,37367,37445,37447,37464,37475,37483,37592,37640,37723,37734,37748,37766,38007,38154,38432,38491,38497,38498,38585,38671,38710,38788,38794,38862,38879,39019,39031,39252,39283,39323,39332,39343,39387,39442,39636,39696,39743,39765,39827,39830,39867,39898,39902,39926,39956,39982,40038,40043,40127,40172,40194,40259,40264,40294,40423,40478,40609,40664,40799,40821,40861,40931,40983,41036,41054,41126,41170,41193,41213,41259,41333,41335,41378,41403,41447,41514,41526,41533,41567,41596,41615,41619,41628,41656,41716,41735,41736,41787,41820,41870,41953,41955,42022,42037,42050,42057,42070,42084,42127,42165,42345,42539,42540,42574,42614,42637,42649,42658,42669,42697,42780,42869,42922,42946,42988,43004,43007,43063,43118,43159,43170,43178,43195,43217,43250,43252,43317,43387,43398,43434,43474,43496,43525,43584,43592,43656,43662,43698,43737,43805,43823,43864,43868,43879,43968,44046,44123,44136,44140,44178,44248,44309,44447,44457,44578,44619,44701,44750,44789,44919,45000,45023,45058,45233,45234,45299,45324,45444,45554,45556,45587,45692,45817,45846,45969,45981,45999,46012,46227,46248,46328,46401,46504,46543,46601,46648,46669,46819,46973,47035,47058,47067,47131,47165,47251,47484,47557,47564,47578,47615,47735,47780,47808,47827,47865,47975,48096,48132,48133,48268,48280,48383,48518,48783,48787,48799,49051,49102,49161,49238,49332,49473,49491,49692,49729,49730,49747,49815,49826,49994,50000,50047,50080,50094,50139,50155,50177,50290,50359,50414,50501,50537,50578,50681,50688,50740,50783,50850,50978,51008,51020,51068,51281,51301,51347,51377,51411,51622,51684,51751,51791,51830,51873,51898,51904,51918,52041,52346,52438,52473,52475,52477,52643,53047,53097,53130,53158,53168,53196,53201,53356,53762,53798,53974,54094,54097,54100,54356,54675,54727,54795,54878,55044,55090,55133,55184,55198,55463,55632,55729,55821,55957,55980,56061,56068,56166,56318,56355,56365,56477,56504,56549,56602,56650,56682,56737,56761,56792,56883,56935,57006,57225,57356,57548,57563,57603,57623,57763,57773,57777,57799,57845,57900,57987,58017,58024,58080,58099,58132,58148,58285,58359,58381,58439,58501,58508,58635,58688,58739,58748,58840,58882,58947,58948,59021,59028,59084,59089,59148,59159,59163,59200,59396,59425,59524,59549,59593,59606,59609,59731,59866,59913,59926,59978,59989,60012,60015,60020,60141,60217,60266,60384,60556,60631,60737,60907,61187,61220,61233,61285,61353,61490,61507,61813,61855,61932,61980,62007,62025,62184,62210,62344,62398,62430,62645,62886,63362,63422,63439,63441,63467,63487,63525,63606,63616,63676,63695,63759,63846,63859,63885,63893,63908,63943,63958,64056,64207,64269,64391,64394,64401,64437,64458,64697,64709,64765,64808,64837,65054,65067,65073,65136,65140,65159,65242,65415,65470,65510,65522,65604,65635,65992,66089,66224,66230,66253,66255,66290,66371,66499,66514,66562,66579,66592,66640,66689,66741,66751,66768,66862,66868,66963,67050,67081,67159,67249,67353,67450,67701,67765,67965,68070,68080,68083,68166,68225,68336,68515,68535,68597,68650,68730,68772,68876,68898,69049,69089,69195,69406,69789,69807,69815,69848,69858,69882,69933,69952,70294,70338,70419,70425,70802,70808,70831,70925,71028,71359,71490,71577,71786,71864,71969,71992,72015,72278,72506,72519,72562,72638,72682,72693,72695,72898,73094,73145,73168,73176,73199,73237,73266,73288,73338,73411,73508,73637,73711,73722,73775,73800,73814,74032,74135,74372,74373,74424,74437,74454,74586,74625,74628,74640,74649,74687,74716,74755,74946,75081,75135,75143,75149,75173,75414,75427,75631,75727,75792,75963,76003,76013,76072,76162,76280,76299,76426,76434,76452,76485,76593,76618,76621,76623,76778,76925,76936,76966,76969,76976,76983,76995,77056,77132,77172,77241,77491,77587,77642,77689,77951,77965,78031,78163,78166,78291,78306,78390,78415,78520,78662,78762,78808,78834,78891,78931,79018,79265,79299,79342,79588,79625,79642,79677,79833,79837,79897,79971,79996,80046,80156,80352,80383,80510,80553,80603,80696,80816,80886,80932,81077,81167,81307,81359,81839,81851,81935,82045,82106,82166,82174,82264,82273,82287,82330,82338,82343,82350,82452,82611,82627,83071,83249,83333,83455,83469,83597,83644,83708,83757,83785,83864,83871,83937,83939,84073,84179,84185,84308,84323,84335,84368,84471,84496,84532,84547,84600,84617,84632,84763,84874,84979,85006,85072,85149,85315,85327,85384,85453,85471,85506,85764,85776,85860,85862,85882,85918,85934,86060,86096,86145,86229,86380,86415,86447,86502,86544,86763,86764,86827,86832,86913,86948,86952,87009,87105,87147,87158,87166,87275,87345,87368,87413,87462,87673,87709,87816,87846,87891,87947,88044,88066,88075,88158,88168,88174,88238,88634,88646,88912,89108,89146,89167,89208,89332,89342,89346,89363,89404,89419,89647,89775,90059,90109,90146,90187,90208,90458,90521,90532,90756,90853,90956,91193,91256,91322,91334,91364,91446,91607,91636,91816,91864,91978,91979,92142,92174,92334,92387,92476,92477,92501,92618,92704,92705,92708,92774,92825,93063,93240,93248,93407,93433,93438,93554,93606,93622,93721,93783,93816,93821,93824,93887,93897,94005,94216,94249,94290,94380,94425,94434,94493,94542,94650,94675,94750,94789,94939,94944,94973,95008,95055,95135,95333,95423,95479,95569,95587,95699,95712,95761,95918,95933,95961,96106,96408,96415,96448,96460,96464,96482,96504,96561,96613,96721,96782,96794,96826,97010,97023,97213,97351,97353,97391,97540,97664,97736,97900,98001,98003,98052,98082,98159,98193,98227,98327,98335,98380,98392,98442,98556,98693,98807,98914,98940,98997,99004,99060,99072,99110,99209,99280,99408,99559,99733,99755,99760,99813,99821,99848,99889,99974,99999",
          "38,300,674,784,857,1034,1080,1166,1225,1307,1349,1427,1464,1497,1533,1576,1651,1913,1918,1973,2016,2104,2195,2261,2393,2528,2587,2633,2750,2774,2788,2975,3018,3039,3099,3155,3170,3318,3319,3367,3475,3524,3573,3656,3666,3771,3776,3936,4001,4100,4304,4366,4421,4557,4574,4602,4739,4784,4790,4919,4935,5010,5063,5107,5118,5147,5248,5373,5623,5734,5864,5995,6147,6268,6291,6296,6352,6416,6434,6481,6648,6665,6675,6730,6738,6803,6880,6901,6987,6994,7110,7138,7228,7251,7333,7358,7385,7485,7595,7635,7780,7957,8033,8191,8235,8309,8431,8456,8547,8632,8709,8760,8796,8992,9004,9032,9033,9055,9190,9209,9328,9379,9535,9617,9632,9636,9704,9705,9742,9788,9832,9946,9972,10088,10336,10458,10504,10589,10679,10844,10916,10978,11243,11246,11361,11455,11549,11577,11617,11660,11793,11804,11846,11932,11975,12075,12110,12143,12204,12293,12318,12342,12409,12476,12536,12538,12539,12542,12584,12588,12656,12698,12738,12759,12795,12850,12871,13052,13084,13129,13197,13274,13368,13423,13484,13487,13500,13579,13615,13626,13654,13691,13698,13779,13832,13857,13866,13868,13896,13982,14125,14206,14339,14340,14355,14417,14477,14573,14576,14594,14648,14860,14889,14914,14919,14927,14983,15079,15111,15187,15271,15347,15385,15448,15457,15582,15957,15966,15973,16122,16322,16349,16358,16421,16470,16515,16533,16623,16746,16909,16932,17013,17046,17056,17276,17305,17415,17423,17631,17633,17636,17646,17754,17863,17939,17963,18020,18126,18313,18338,18357,18425,18480,18694,18795,18853,19088,19089,19272,19303,19308,19325,19351,19387,19423,19500,19641,19839,19874,19899,20031,20043,20126,20234,20242,20419,20442,20488,20566,20574,20614,20626,20663,20666,20733,20736,20742,20824,20843,20845,20865,20928,21103,21200,21219,21242,21246,21373,21421,21430,21508,21511,21559,21636,21674,21763,21873,21896,21914,21958,21963,21977,22004,22032,22080,22106,22432,22469,22496,22590,22623,22687,22703,22755,22819,22827,22829,22983,23141,23155,23208,23314,23353,23360,23430,23601,23850,23875,23915,24007,24102,24103,24185,24258,24282,24342,24474,24490,24518,24583,24588,24770,24860,25021,25058,25104,25171,25269,25324,25390,25422,25513,25582,25605,25608,25649,25722,25761,25820,25837,25851,25926,26094,26158,26172,26198,26249,26330,26371,26427,26444,26449,26602,26618,26641,26663,26912,26971,27118,27190,27230,27283,27299,27345,27407,27550,27683,27722,27844,27866,27901,27909,28099,28172,28237,28257,28264,28280,28285,28476,28477,28518,28667,28724,28754,28845,28925,29073,29087,29213,29287,29323,29489,29491,29492,29493,29669,29674,29729,29778,29792,29983,30026,30135,30168,30206,30320,30349,30393,30423,30641,30666,30669,30816,30843,30883,30890,30907,30985,31130,31139,31235,31244,31283,31326,31508,31621,31789,31933,31944,31971,32125,32195,32199,32200,32202,32240,32259,32392,32416,32424,32593,32602,32676,32738,32773,32895,32982,33049,33106,33137,33139,33202,33287,33416,33444,33456,33533,33703,33722,33738,33800,33920,33948,33976,34217,34238,34442,34447,34555,34569,34583,34650,34700,34742,34750,34799,34829,34848,34913,35021,35028,35092,35300,35403,35406,35422,35426,35461,35501,35522,35756,35797,35815,35819,35868,35890,35945,36023,36039,36094,36205,36236,36237,36259,36299,36337,36367,36380,36421,36478,36526,36553,36573,37068,37318,37426,37503,37661,37707,37796,37840,37863,37956,37982,38011,38012,38031,38082,38174,38190,38278,38399,38495,38509,38549,38741,38759,38907,38939,39071,39081,39084,39220,39350,39366,39411,39449,39487,39547,39604,39716,39781,39910,39995,40001,40031,40204,40222,40245,40302,40368,40480,40487,40559,40780,40784,40797,40814,40919,40950,40991,41001,41024,41120,41180,41198,41219,41278,41389,41405,41520,41543,41579,41624,41695,41752,41789,41799,41978,42030,42043,42104,42191,42285,42307,42456,42520,42627,42722,42767,42872,42892,42897,42943,42955,42965,42969,43032,43090,43121,43122,43354,43436,43516,43620,43631,43663,43790,43829,43921,43944,43984,44094,44183,44203,44316,44402,44513,44514,44528,44659,44918,44953,45046,45054,45088,45182,45289,45308,45362,45376,45622,45670,45689,45760,45843,45914,46041,46071,46075,46091,46212,46215,46281,46302,46358,46396,46464,46510,46560,46876,46931,47000,47097,47117,47170,47235,47274,47308,47311,47325,47338,47369,47376,47378,47395,47425,47474,47581,47771,47861,47932,48059,48200,48218,48330,48425,48571,48596,48792,48887,49070,49266,49279,49292,49535,49605,49627,49696,49712,49757,49778,49799,49821,49833,50011,50037,50062,50106,50151,50277,50458,50493,50551,50592,50665,50693,50762,50828,50893,50966,50971,51120,51275,51471,51654,51669,51707,51734,51766,52061,52094,52284,52331,52359,52682,52737,52803,52891,52941,52948,53066,53094,53131,53152,53159,53271,53389,53393,53497,53527,53552,53882,54051,54060,54122,54166,54225,54368,54461,54464,54502,54507,54515,54596,54601,54687,54811,55028,55038,55105,55189,55218,55407,55499,55504,55509,55555,55653,55681,55692,55705,55750,55814,55823,55933,55950,56086,56356,56358,56387,56616,56681,56805,56871,56903,56958,57063,57065,57191,57410,57551,57591,57631,57667,57709,57716,57724,57825,58060,58104,58111,58142,58228,58295,58402,58628,58684,58686,58725,58847,59055,59150,59185,59195,59404,59664,59668,59704,59709,59765,59826,59994,60142,60176,60253,60271,60324,60444,60485,60563,60596,60694,60815,60820,60822,60831,60864,60886,60893,61122,61125,61128,61138,61192,61242,61543,62033,62077,62118,62174,62198,62323,62356,62475,62721,62735,62862,62863,62908,63024,63155,63189,63254,63303,63499,63590,63622,63640,63710,63716,63735,63800,63970,64022,64204,64241,64261,64308,64330,64335,64642,64679,64689,64724,64846,64893,64912,64921,64929,64947,65062,65153,65198,65262,65281,65314,65446,65535,65551,65597,65663,65872,65895,66029,66033,66183,66271,66321,66724,66800,66816,66820,66836,66851,66887,66964,66973,67018,67028,67039,67061,67177,67260,67280,67376,67554,67612,67662,67757,67990,68052,68092,68172,68214,68292,68466,68491,68611,68625,68784,68841,68863,68919,68948,68980,69087,69125,69166,69277,69336,69339,69357,69407,69617,69635,69654,69682,69698,69719,69797,69825,69960,69962,70056,70067,70142,70155,70306,70321,70360,70456,70463,70499,70791,70946,70995,71033,71046,71092,71107,71266,71303,71335,71364,71709,71722,71738,71754,71755,72081,72267,72374,72379,72430,72446,72578,72664,72732,72761,72825,72892,72900,72960,72971,73224,73281,73372,73431,73496,73530,73584,73589,73693,73770,73811,73857,73883,74004,74026,74068,74150,74162,74206,74255,74401,74412,74483,74579,74614,74667,75013,75153,75429,75804,75837,75846,75847,75894,75961,76112,76144,76149,76249,76362,76396,76416,76429,76458,76555,76629,76700,76720,76851,77029,77154,77386,77482,77504,77544,77546,77588,77595,77813,77931,78071,78106,78140,78143,78179,78254,78273,78293,78302,78354,78466,78481,78507,78771,78811,78923,78976,79003,79011,79134,79313,79595,79723,79794,79855,79856,79882,79891,79906,79922,79937,80043,80344,80355,80393,80409,80443,80523,80555,80609,80835,80872,80897,81006,81039,81238,81428,81564,81568,81598,81664,81795,81798,81802,81806,81807,81817,81978,81986,82018,82040,82143,82185,82230,82536,82576,82598,82615,82630,82639,82807,82893,82976,83014,83021,83139,83221,83239,83240,83292,83301,83433,83556,83580,83594,83764,83823,83832,83839,83874,83971,84059,84112,84239,84350,84476,84508,84535,84663,84752,84790,84931,84995,85029,85095,85166,85184,85203,85326,85347,85382,85440,85539,85582,85626,85829,86153,86247,86392,86398,86433,86443,86510,86522,86566,86728,86759,87075,87146,87180,87205,87216,87244,87248,87405,87411,87515,87554,87569,87575,87852,87950,88500,88573,88630,88645,88688,88700,88747,88774,88780,88848,88983,89097,89134,89139,89221,89386,89388,89414,89464,89565,89610,89615,89711,89755,89793,89901,89981,90001,90065,90077,90117,90248,90305,90479,90533,90535,90630,90708,90964,90966,90986,91014,91072,91121,91178,91189,91296,91649,91723,92001,92030,92053,92309,92336,92370,92390,92418,92519,92521,92530,92689,92794,92913,93035,93114,93117,93177,93182,93208,93348,93891,93972,93983,94044,94119,94162,94402,94455,94517,94549,94617,94625,94657,94723,94764,94937,95023,95066,95134,95298,95302,95401,95589,95601,95680,95962,95986,96010,96276,96313,96417,96550,96671,96783,96809,97003,97055,97085,97111,97388,97415,97504,97542,97638,97752,97811,97822,97839,97852,97853,98056,98099,98132,98291,98319,98431,98717,98764,98768,98955,99012,99063,99079,99131,99157,99257,99267,99278,99379,99500,99696,99788,99805,99814,99843,99875,99881",
          "43,322,335,409,426,432,644,652,762,977,990,1348,1635,1744,1794,1933,2019,2065,2092,2212,2305,2324,2414,2488,2769,2770,2888,2900,3107,3135,3142,3156,3259,3317,3381,3440,3450,3583,3617,3627,3685,3699,3799,3866,3904,3973,4031,4371,4584,4635,4699,4723,4741,4794,4829,4908,5284,5375,5377,5394,5436,5438,5592,5616,5707,5749,5819,5892,5930,6023,6115,6153,6300,6344,6364,6374,6563,6589,6658,6853,6879,6963,7215,7298,7461,7546,7553,7566,7580,7588,7788,7802,7844,7884,7909,7920,7983,8018,8044,8076,8114,8141,8157,8238,8239,8279,8490,8491,8681,8695,8724,8726,8787,9143,9165,9312,9326,9439,9557,9624,9654,9689,9729,9777,9880,9881,9986,10055,10097,10133,10232,10367,10405,10450,10509,10768,10866,10875,10889,10933,10958,11049,11110,11171,11253,11264,11287,11291,11535,11547,11586,11645,11764,11848,12030,12105,12230,12473,12492,12517,12556,12670,12771,12828,12857,12858,13201,13267,13294,13470,13511,13741,13742,13777,13808,13850,13899,13974,14080,14131,14310,14318,14393,14399,14471,14536,14733,14792,14852,14904,14964,15005,15023,15127,15339,15341,15423,15459,15560,15637,15645,15674,15701,15728,15942,15943,15954,16009,16015,16081,16084,16103,16489,16643,16936,17028,17074,17095,17366,17433,17637,17808,17823,17857,17872,17897,18022,18124,18413,18443,18525,18565,18656,18693,18713,18734,18787,18867,18971,18994,19042,19092,19104,19109,19112,19156,19335,19444,19480,19498,19554,19694,19699,19707,19816,19898,19918,19948,19975,19995,20006,20036,20071,20175,20232,20285,20428,20522,20599,20673,20678,20750,20785,20810,20963,21025,21226,21257,21304,21337,21342,21444,21464,21509,21527,21528,21615,21616,21640,21682,21740,21749,21758,21809,21837,21918,22065,22128,22149,22419,22584,22681,22709,22733,22808,22812,22815,23123,23337,23413,23560,23740,23789,23945,23965,23979,24090,24125,24142,24143,24158,24207,24233,24237,24267,24305,24349,24368,24439,24507,24607,24638,24769,24852,24900,24985,25193,25245,25305,25386,25403,25434,25505,25535,25583,25781,25803,25841,25860,25898,25904,25912,25968,25997,26224,26347,26351,26438,26539,26574,26645,26697,26711,26833,26940,26978,27025,27053,27062,27070,27176,27195,27222,27224,27226,27228,27236,27343,27393,27417,27583,27614,27795,27940,27974,28060,28063,28097,28229,28253,28323,28380,28404,28493,28537,28660,28727,28771,28795,28844,28908,29103,29125,29196,29234,29283,29333,29347,29413,29440,29459,29477,29508,29581,29583,29599,29607,29704,29731,29813,29849,29911,29931,30090,30102,30278,30468,30497,30502,30622,30634,30642,30694,30977,31062,31064,31065,31257,31288,31304,31375,31378,31380,31416,31465,31497,31611,31652,31810,31827,31858,31860,31875,31935,31942,32003,32053,32128,32157,32252,32269,32341,32376,32413,32753,32776,32800,32885,32907,32954,33077,33123,33158,33162,33210,33218,33221,33261,33289,33297,33335,33351,33408,33409,33479,33591,33608,33655,33675,33710,33723,33724,33834,33953,34013,34068,34100,34139,34175,34191,34302,34384,34390,34393,34432,34434,34471,34480,34505,34524,34643,34673,34674,34827,34903,34950,34967,34977,35013,35066,35133,35245,35287,35535,35543,35637,35703,35714,35726,35870,35880,35881,35892,35986,36009,36027,36058,36105,36190,36252,36332,36348,36506,36559,36645,36665,36709,36718,36816,36834,36839,36965,36989,37008,37043,37112,37128,37194,37209,37327,37330,37402,37420,37571,37660,37662,37664,37727,37794,37871,37963,37965,38019,38056,38118,38130,38155,38255,38311,38365,38370,38431,38466,38516,38527,38560,38682,38706,38746,38805,38937,38991,39094,39139,39154,39198,39275,39278,39509,39581,39671,39689,39730,39776,39863,39880,39975,40003,40042,40137,40215,40235,40396,40402,40403,40416,40439,40473,40510,40524,40567,40650,40787,40800,40801,40901,40957,41009,41088,41091,41119,41166,41174,41215,41241,41249,41282,41301,41351,41453,41481,41554,41599,41600,41603,41620,41636,41641,41824,41985,42096,42110,42146,42241,42244,42265,42402,42485,42491,42507,42571,42599,42714,42730,42739,42796,42816,42821,42900,43059,43151,43275,43281,43347,43388,43438,43440,43558,43597,43613,43675,43683,43702,43706,43753,43789,43820,43859,43891,44131,44159,44172,44352,44353,44364,44388,44395,44433,44437,44503,44509,44541,44646,44723,44779,44797,44812,45166,45175,45205,45222,45240,45295,45321,45402,45474,45507,45533,45589,45642,45713,45842,45880,45891,46096,46181,46241,46475,46545,46660,46744,46796,47034,47051,47232,47247,47289,47316,47343,47540,47576,47670,47685,47699,47761,48048,48055,48172,48229,48258,48316,48327,48333,48349,48353,48385,48434,48528,48554,48561,48741,48771,48801,48817,48841,48854,48882,48922,48933,48956,48958,49082,49127,49139,49216,49255,49311,49360,49421,49541,49567,49660,49787,49812,49927,50076,50078,50090,50257,50342,50346,50531,50610,50630,50814,50860,50908,50988,51013,51124,51197,51236,51252,51559,51607,51625,51640,51760,51842,52027,52031,52111,52118,52124,52138,52218,52350,52542,52553,52670,52709,52808,52905,52937,53073,53082,53084,53106,53109,53300,53314,53491,53498,53601,53742,53767,53808,53812,53883,53921,53963,54031,54233,54412,54698,54788,54855,54946,54988,55036,55156,55176,55183,55423,55510,55619,55753,55835,55921,56049,56153,56259,56295,56337,56385,56388,56411,56420,56535,56606,56608,56698,56699,56829,56858,56902,56982,57075,57125,57180,57244,57454,57469,57504,57568,57572,57643,57714,57719,57748,57813,57844,57918,58010,58135,58203,58207,58251,58258,58296,58344,58569,58585,58607,58613,58675,58693,58793,58849,58911,58970,59003,59234,59311,59321,59405,59441,59494,59618,59685,59719,59720,59763,59938,59950,60148,60181,60265,60278,60335,60337,60350,60491,60517,60533,60559,60560,60654,60691,60726,60782,60803,60848,60890,60945,60970,61005,61016,61117,61123,61141,61189,61247,61259,61278,61374,61420,61438,61439,61469,61486,61489,61705,61990,62002,62044,62047,62125,62277,62280,62414,62445,62458,62557,62574,62606,62609,62618,62672,62783,62797,62950,62957,62974,63057,63114,63198,63260,63317,63333,63347,63360,63433,63474,63541,63552,63579,63672,63697,63799,63960,64015,64293,64341,64376,64676,64790,64819,64904,64925,65006,65049,65109,65152,65221,65252,65260,65273,65392,65623,65688,65802,65829,65958,65983,66046,66070,66099,66170,66316,66339,66366,66457,66623,66644,66669,66692,66739,66831,67012,67063,67070,67132,67418,67426,67440,67514,67533,67571,67603,67721,67909,67914,67955,68666,68774,68968,69012,69037,69039,69244,69493,69532,69587,69589,69744,69783,69829,69917,69922,69957,69958,69970,70011,70186,70373,70394,70412,70520,70547,70553,70642,70645,70748,70954,71108,71306,71325,71326,71442,71569,71619,71703,71704,71813,72056,72194,72208,72266,72297,72346,72380,72423,72463,72502,72607,72648,72700,72708,72735,72818,72833,72856,72861,72863,72893,73045,73078,73173,73324,73359,73382,73429,73498,73541,73550,73565,73889,74106,74113,74212,74261,74314,74315,74317,74514,74539,74664,74685,74784,74857,74934,74977,75068,75088,75159,75161,75344,75374,75575,75653,75676,75695,75749,75856,75890,75938,75950,76000,76384,76502,76512,76602,76766,76950,76952,77022,77060,77186,77240,77293,77401,77479,77548,77563,77593,77654,77735,77791,77864,77930,77935,78035,78052,78073,78169,78186,78223,78269,78314,78344,78405,78609,78626,78691,78766,78807,78969,79066,79124,79291,79469,79523,79530,79653,79654,79759,79767,79802,79854,79968,80219,80251,80302,80318,80335,80378,80663,80727,80782,80788,80856,81071,81139,81182,81237,81484,81628,81637,81685,81815,81937,81968,82055,82059,82074,82132,82231,82362,82365,82438,82450,82520,82779,82886,82950,82953,83016,83100,83229,83354,83399,83596,83651,83659,83664,83679,83684,83727,83827,83847,83887,84008,84127,84155,84267,84469,84505,84694,84764,84872,84966,85012,85254,85293,85419,85467,85559,85740,86009,86014,86015,86151,86167,86183,86499,86785,86983,87051,87064,87115,87148,87303,87525,87619,87663,87677,87755,87826,87887,87969,87992,88001,88003,88024,88052,88131,88253,88273,88394,88458,88494,88588,88652,88769,88991,89023,89040,89214,89244,89379,89456,89539,89581,89692,89698,89773,89785,89815,89836,89928,89972,90095,90165,90358,90430,90447,90464,90563,90681,90821,90931,90937,90962,90988,91127,91272,91293,91315,91428,91450,91465,91543,91631,91645,91650,91746,91929,91991,92041,92090,92153,92351,92366,92474,92499,92572,92652,92682,92776,92790,92922,92962,93046,93194,93206,93317,93357,93442,93536,93675,93827,93971,94089,94122,94143,94174,94198,94232,94243,94330,94392,94437,94490,94519,94569,94601,94635,94690,94691,94696,94734,94830,94910,94923,95151,95455,95472,95536,95582,95609,95646,95676,95701,95778,95827,95839,96152,96166,96247,96423,96499,96558,96636,96640,96727,96744,96874,96880,97009,97029,97199,97243,97394,97486,97632,97637,97855,97998,98188,98381,98385,98551,98559,98571,98623,98642,98675,98982,99303,99331,99486,99531,99554,99719,99914,99957,99971",
          "30,41,149,166,337,447,513,526,564,810,822,889,985,1090,1195,1261,1371,1390,1405,1554,1562,1582,1654,1749,1756,1815,1826,2037,2310,2356,2380,2451,2456,2542,2580,2679,3232,3254,3288,3422,3427,3443,3498,3529,3531,3635,3739,3740,3748,3880,3931,3942,4162,4187,4196,4272,4532,4554,4596,4631,4754,4966,4993,4996,5163,5374,5448,5488,5492,5598,5686,5786,5928,6002,6057,6101,6132,6218,6271,6367,6489,6497,6640,6768,6788,6893,7037,7059,7075,7089,7220,7296,7506,7645,7723,7820,7848,7862,7945,7988,8000,8045,8055,8062,8110,8139,8180,8198,8216,8280,8365,8557,8604,8691,8744,8878,8888,8921,9059,9078,9136,9476,9598,9674,9690,9789,9853,9925,9990,9993,10022,10091,10143,10163,10169,10198,10244,10265,10283,10302,10323,10437,10484,10537,10634,10639,10665,10698,10702,10759,10894,10942,10968,11063,11475,11580,11641,11873,11968,11992,12027,12068,12166,12199,12212,12257,12279,12314,12408,12438,12461,12470,12507,12543,12577,12615,12737,12821,12842,12967,13057,13102,13136,13226,13304,13360,13424,13430,13491,13510,13686,13924,14019,14038,14077,14115,14253,14332,14388,14499,14502,14569,14672,14809,14872,14890,14923,15069,15117,15247,15466,15490,15585,15591,15770,15962,16051,16052,16056,16098,16115,16187,16225,16411,16435,16469,16637,16638,16671,16786,16971,17005,17048,17201,17233,17298,17340,17649,17667,17752,17801,17843,17871,17962,17968,17989,18046,18052,18072,18081,18202,18208,18213,18215,18441,18492,18587,18610,18634,18671,18783,18796,18802,18817,18841,18851,18954,19077,19136,19257,19295,19468,19549,19564,19602,19984,20004,20014,20049,20085,20103,20116,20143,20164,20302,20306,20312,20530,20741,20753,20778,20782,20911,20942,20951,20991,21164,21214,21254,21351,21375,21457,21494,21522,21558,21718,21739,21772,21869,21921,22020,22101,22126,22184,22203,22273,22308,22357,22371,22411,22437,22495,22536,22561,22564,22585,22695,22697,22741,22959,22998,23102,23192,23309,23346,23485,23680,23890,23891,23908,23929,24009,24024,24026,24229,24415,24463,24616,24635,24688,24936,25007,25055,25189,25457,25543,25694,25764,25794,26007,26099,26221,26305,26521,26571,27057,27240,27244,27298,27301,27312,27335,27378,27453,27455,27543,27564,27584,27600,27677,27769,27864,27896,27928,28004,28009,28053,28231,28290,28519,28655,28659,28742,28750,28814,28822,28918,28980,29008,29030,29096,29274,29309,29327,29372,29421,29683,29692,29805,29848,29858,29892,29900,29980,30230,30312,30346,30381,30391,30425,30459,30462,30511,30543,30566,30680,30837,30873,30877,30966,31059,31061,31142,31147,31228,31373,31441,31458,31519,31592,31630,31637,31659,31674,31691,31728,31772,32045,32066,32114,32275,32278,32409,32420,32488,32574,32579,32639,32651,32713,32718,32744,32827,32867,32872,32926,33056,33058,33331,33447,33624,33764,33890,33952,33957,34026,34091,34102,34150,34192,34235,34268,34563,34631,34633,34638,34709,34867,34960,35090,35103,35106,35129,35143,35164,35193,35223,35265,35266,35298,35302,35312,35476,35539,35586,35602,35718,35751,35761,35930,35933,35967,35975,35989,36034,36090,36139,36302,36326,36616,36786,36980,37072,37125,37210,37214,37303,37325,37496,37686,37812,37821,37830,37857,38073,38163,38337,38442,38478,38575,38595,38667,38738,38834,38854,38936,38972,38975,38983,38986,39007,39077,39334,39389,39402,39534,39587,39723,39749,39866,39874,39941,39950,39973,40128,40138,40211,40216,40369,40406,40419,40424,40795,40807,40961,41083,41100,41123,41168,41231,41303,41309,41315,41316,41343,41449,41452,41489,41660,41724,41788,41793,41899,41989,41996,42138,42144,42195,42233,42329,42393,42464,42497,42652,42790,42915,42930,42941,43081,43104,43108,43196,43259,43265,43333,43346,43418,43503,43743,43757,43836,43925,44073,44075,44091,44244,44274,44282,44285,44380,44422,44485,44507,44573,44650,44678,44722,44818,44852,44931,44970,45033,45252,45316,45319,45516,45524,45532,45536,45543,45577,45620,45879,45933,45974,46059,46097,46101,46119,46124,46356,46380,46415,46636,46638,46649,46729,46764,46854,46982,47014,47078,47206,47352,47399,47419,47566,47583,47671,47698,47820,47955,48377,48378,48556,48601,48638,48663,48862,48947,48972,49023,49046,49112,49171,49209,49301,49367,49434,49594,49680,49724,49731,49745,49786,49843,50027,50048,50074,50112,50146,50158,50376,50430,50434,50491,50512,50571,50603,50622,50657,50780,50846,50890,50998,51007,51087,51147,51182,51212,51256,51368,51399,51460,51517,51527,51548,51649,51671,51674,51693,51890,51896,51910,52055,52106,52134,52157,52280,52283,52334,52371,52452,52520,52590,52627,52700,52704,52764,52984,53004,53049,53058,53089,53104,53273,53324,53378,53381,53401,53464,53576,53580,53668,54061,54080,54197,54279,54294,54383,54455,54573,54590,54628,54646,54789,54798,54839,54863,54903,55074,55081,55116,55196,55222,55227,55277,55353,55542,55601,55631,55652,55746,55794,55866,55906,55977,56227,56245,56334,56463,56473,56626,56714,56743,56765,56786,56803,56848,56882,57009,57097,57135,57290,57359,57362,57650,57669,57827,57829,57854,57981,58039,58420,58544,58556,58736,58828,58887,59082,59099,59411,59570,59834,59860,59921,60044,60055,60309,60395,60470,60476,60502,60586,60619,60780,60801,60947,61015,61156,61165,61240,61244,61252,61293,61344,61435,61458,61622,61629,61723,61727,61751,61771,61819,61875,62024,62178,62186,62261,62292,62425,62427,62506,62515,62763,62837,63025,63058,63154,63207,63265,63285,63302,63327,63337,63363,63411,63494,63498,63522,63688,63698,63723,63860,63877,63942,64087,64275,64465,64502,64508,64512,64580,64655,64786,64936,64941,65553,65582,65828,65987,66077,66233,66352,66539,66589,66687,66736,66782,66791,66832,66860,66927,66937,66941,67094,67119,67128,67188,67247,67321,67352,67365,67397,67548,67551,67564,67610,67690,67762,67787,67815,67942,67982,67985,68051,68207,68218,68248,68270,68428,68490,68680,68751,68791,68843,68873,68941,68954,69022,69025,69038,69136,69145,69254,69342,69454,69488,69519,69768,69790,69827,69835,69873,69878,69900,70031,70079,70080,70095,70108,70330,70364,70407,70433,70512,70574,70578,70614,70652,70700,70715,70731,70763,70803,70982,71001,71182,71233,71311,71333,71350,71437,71450,71459,71542,71674,71699,71732,72001,72079,72274,72289,72307,72569,72719,72935,72963,73003,73023,73083,73273,73342,73416,73439,73471,73522,73556,73560,73576,73596,73600,73715,73741,73812,73818,73942,74071,74074,74116,74223,74268,74301,74409,74455,74582,74797,74867,74917,74937,75056,75122,75307,75552,75612,75615,75650,75735,75746,76228,76235,76325,76348,76370,76394,76419,76494,76542,76569,76584,76631,76656,76798,76807,76900,77096,77100,77237,77297,77330,77355,77381,77402,77565,77618,77646,77651,77681,77695,77697,77746,77785,77812,78368,78485,78488,78597,78813,78867,78878,78928,79031,79116,79182,79272,79286,79296,79427,79460,79568,79575,79608,79640,79647,79688,79836,79913,79977,79990,80186,80194,80369,80455,80577,80724,80746,80820,80827,80840,80874,81111,81113,81301,81433,81475,81536,81561,81610,81680,81988,82017,82138,82163,82177,82296,82418,82624,82743,82800,82859,82892,83020,83061,83169,83314,83506,83786,83868,83931,83933,84067,84184,84202,84251,84326,84403,84424,84536,84550,84598,84630,84692,84759,84777,84805,84944,84971,85028,85098,85120,85121,85132,85150,85229,85252,85387,85562,85600,85618,85661,85715,85760,85789,85881,85970,86136,86141,86168,86244,86313,86321,86379,86408,86435,86444,86449,86538,86617,86711,86843,86860,87020,87112,87151,87221,87257,87365,87380,87454,87532,87533,87655,87672,87694,87777,87823,87967,88002,88039,88093,88125,88439,88512,88670,88763,88837,88881,88901,88976,89088,89161,89182,89210,89231,89245,89261,89265,89445,89499,89526,89718,89744,89837,89878,89909,89985,90000,90038,90354,90423,90433,90497,90581,90655,90683,90717,90730,90826,90835,90999,91091,91134,91144,91170,91292,91545,91546,91574,91613,91623,91624,91627,91754,91778,91807,91887,91891,91898,91963,92093,92130,92140,92251,92338,92411,92466,92680,92726,92745,92806,92901,92974,93025,93029,93052,93056,93074,93101,93168,93176,93271,93284,93383,93440,93490,93591,93739,93792,93902,93920,93980,94105,94534,94724,94793,94885,94925,95067,95235,95328,95421,95486,95487,95619,95738,95772,95927,95963,96223,96281,96320,96321,96393,96491,96554,96706,96810,96885,96909,97072,97133,97328,97357,97513,97533,98009,98067,98251,98299,98318,98423,98447,98531,98548,98599,98613,98617,98667,98728,98808,98823,98834,98900,98911,99052,99071,99112,99197,99247,99326,99385,99451,99547,99772,99865,99896,99939,99992",
          "12,36,49,155,157,199,203,261,387,572,750,754,806,902,953,1026,1123,1125,1137,1217,1229,1415,1426,1452,1479,1536,1668,1823,2074,2269,2442,2483,2619,2656,2680,2840,2866,2902,2910,2915,2977,2988,3082,3224,3229,3509,3568,3585,3604,3657,3680,3745,3779,3797,3854,3949,3985,4040,4112,4180,4193,4215,4288,4359,4360,4377,4383,4586,4729,4730,4751,4775,4887,4899,5018,5059,5108,5109,5239,5403,5452,5461,5587,5614,5633,5903,5971,6105,6159,6162,6165,6207,6260,6267,6303,6306,6334,6345,6357,6491,6739,6760,6770,6844,6889,7022,7025,7281,7363,7440,7496,7720,7787,7836,7838,7881,7969,8003,8073,8193,8246,8319,8353,8438,8622,8655,8682,8885,8898,8903,8917,8998,9126,9156,9194,9199,9266,9474,9501,9530,9593,9637,9642,9660,9684,9697,9797,9937,9962,10005,10054,10085,10154,10242,10320,10364,10489,10507,10560,10621,10632,10685,10770,10774,10808,10836,10865,10948,11067,11161,11334,11429,11479,11503,11512,11722,11768,11841,11895,11964,11985,11999,12004,12013,12145,12180,12197,12359,12395,12396,12475,12493,12622,12752,12760,12807,12822,12942,13008,13156,13157,13225,13281,13613,13711,13760,13772,13875,14149,14156,14205,14335,14400,14451,14486,14568,14603,14634,14666,14698,14752,14822,14841,14850,15261,15387,15398,15679,15840,15896,15897,16029,16063,16107,16143,16221,16285,16302,16396,16451,16472,16521,16557,16722,16825,16879,16914,16975,17130,17227,17393,17420,17450,17451,17478,17543,17582,17661,17731,17815,17907,18013,18150,18182,18192,18198,18314,18410,18438,18555,18692,18894,19039,19050,19051,19302,19322,19342,19393,19395,19454,19474,19493,19573,19740,19747,19844,19858,19933,19943,19994,19998,20002,20029,20039,20054,20057,20320,20411,20474,20703,20840,20969,20976,20996,21050,21069,21265,21347,21382,21545,21683,21688,21691,21769,21859,22262,22542,22566,22594,22649,22778,22786,22910,23081,23163,23184,23315,23513,23553,23599,23620,23622,23708,23711,23836,23842,23845,24044,24088,24111,24216,24224,24250,24255,24359,24520,24541,24592,24691,24790,24850,25163,25212,25286,25310,25373,25497,25527,25565,25574,25581,25596,25677,25753,25754,25757,25806,25919,25940,25948,26076,26106,26176,26290,26294,26325,26364,26494,26654,26706,26773,26876,26893,27018,27052,27078,27577,27625,27650,27686,27712,27717,27966,28001,28026,28132,28147,28182,28520,28525,28607,28764,28788,29258,29278,29307,29438,29572,29584,29777,29884,29937,29979,30187,30270,30360,30451,30550,30597,30701,30734,30743,30899,31239,31296,31369,31538,31629,31648,31649,31670,31721,31786,31915,31976,31982,32067,32191,32320,32622,32704,32726,32789,32809,32864,32882,32918,32984,33018,33293,33360,33471,33560,33606,33613,33621,33790,33896,34064,34131,34178,34218,34303,34352,34419,34566,34574,34605,34644,34677,34866,34932,34983,35056,35063,35113,35219,35274,35344,35431,35497,35518,35528,35534,35594,35655,35733,35778,35780,35847,35876,35908,35915,35938,35944,35973,36066,36150,36198,36257,36273,36351,36374,36446,36551,36634,36687,36701,36866,36899,36967,37024,37248,37324,37405,37417,37494,37638,37683,37815,37824,37864,37923,37940,37976,38023,38041,38054,38120,38151,38160,38164,38240,38369,38374,38451,38454,38486,38507,38581,38608,38629,38739,38772,38783,38825,38910,39059,39104,39128,39150,39306,39364,39406,39516,39546,39715,39735,39798,39836,39928,39939,39962,40024,40055,40076,40223,40227,40275,40433,40569,40584,40633,40743,40762,40772,40796,40859,41028,41095,41493,41566,41670,41768,41779,41783,41810,41839,41940,42032,42106,42159,42200,42228,42490,42733,42843,42864,42907,42928,42948,42958,42973,43008,43030,43191,43215,43239,43478,43498,43506,43541,43557,43626,43911,43960,44015,44038,44139,44188,44527,44555,44577,44604,44637,44672,44688,44814,44954,44973,44997,45018,45043,45064,45266,45358,45368,45420,45498,45500,45519,45531,45544,45606,45703,45740,45821,46099,46188,46235,46268,46289,46299,46366,46394,46453,46508,46683,46781,46828,47140,47485,47577,47681,47719,47797,47798,47853,47893,47945,47962,48007,48056,48069,48127,48128,48501,48591,48734,48752,48756,48763,48777,48798,48873,48921,48987,49032,49035,49128,49268,49361,49365,49399,49403,49515,49532,49538,49635,49643,49662,49750,49790,49811,49839,49847,49863,49910,49943,49988,50005,50020,50046,50109,50115,50138,50175,50214,50314,50426,50662,50692,50708,50747,50764,50918,51078,51083,51085,51099,51107,51196,51294,51312,51627,51655,51701,51735,51878,51925,52056,52109,52169,52245,52264,52462,52546,52625,52797,52825,52826,52906,53189,53190,53355,53449,53550,53666,53789,53816,53842,53928,53988,54140,54156,54228,54355,54469,54551,54766,54773,54869,54960,55148,55181,55274,55296,55359,55367,55389,55500,55543,55641,55664,55677,55817,55885,56207,56324,56503,56519,56546,56591,56600,56890,56918,56994,57029,57032,57112,57230,57388,57419,57472,57570,57779,57802,57850,57886,57902,57957,58266,58267,58509,58899,58927,59048,59104,59173,59236,59280,59308,59330,59331,59355,59559,59563,59697,59728,59822,59964,60031,60112,60155,60268,60289,60291,60359,60389,60400,60425,60494,60539,60577,60637,60648,60717,60728,60755,60781,60849,60872,60877,60880,60989,61174,61209,61264,61271,61292,61362,61448,61464,61522,61596,61615,61704,61728,61750,61778,62003,62103,62166,62383,62407,62461,62483,62497,62648,62695,62757,62824,63007,63018,63037,63120,63125,63219,63269,63271,63533,63730,63789,63815,64033,64136,64141,64250,64254,64339,64353,64398,64419,64445,64491,64509,64551,64554,64574,64627,64660,64895,64924,64948,65018,65029,65037,65091,65121,65235,65269,65341,65393,65457,65465,65467,65484,65492,65503,65506,65564,65585,65710,65869,65885,66035,66107,66158,66201,66221,66320,66509,66571,66753,66756,66855,67001,67003,67037,67161,67283,67375,67452,67586,67622,67682,67698,67716,67749,67837,67939,68100,68238,68340,68407,68556,68569,68681,68710,68882,69054,69085,69334,69402,69650,69689,69693,69740,69775,69787,69801,69887,69895,70054,70063,70089,70190,70208,70266,70299,70392,70528,70598,70698,70770,70834,70882,70951,70958,70965,71018,71062,71086,71118,71218,71300,71537,71578,71584,71860,71862,71868,71943,71948,72009,72123,72140,72239,72337,72445,72457,72467,72507,72514,72515,72518,72603,72703,72773,72795,72797,72851,72980,73020,73130,73132,73241,73243,73252,73255,73336,73502,73585,73595,73658,73682,73915,73926,73936,74177,74296,74340,74416,74518,74622,74696,74915,74925,74930,74987,75070,75288,75299,75371,75441,75452,75466,75593,75595,75715,75740,75814,75943,75992,76082,76174,76217,76456,76467,76536,76599,76717,76744,76909,76998,77009,77099,77136,77157,77351,77407,77505,77569,77575,77665,77696,77842,77848,77912,77925,77946,78080,78108,78311,78364,78369,78377,78399,78424,78496,78503,78700,78713,78815,78903,79015,79159,79308,79378,79436,79452,79464,79505,79550,79712,79745,79768,79820,80006,80055,80062,80207,80211,80217,80246,80319,80391,80460,80586,80711,80898,80908,81222,81268,81413,81418,81442,81503,81541,81549,81654,81689,81948,81956,82030,82070,82090,82180,82233,82304,82345,82382,82455,82517,82543,82867,82989,83036,83140,83237,83265,83299,83306,83348,83387,83416,83503,83641,84012,84029,84032,84042,84048,84084,84105,84159,84232,84289,84315,84316,84730,84737,85003,85083,85128,85228,85291,85292,85299,85309,85349,85388,85587,85589,85692,85713,85802,85914,85938,86008,86476,86545,86662,86703,86769,86921,87022,87107,87140,87233,87250,87288,87352,87491,87597,87626,87665,87682,87730,87741,87973,88120,88126,88398,88446,88532,88541,88544,88575,88629,88810,88947,88961,89056,89140,89190,89232,89282,89309,89322,89353,89374,89458,89566,89588,89670,89680,89683,89810,89894,90005,90060,90112,90274,90337,90511,90589,90646,90649,90668,90719,90799,90808,90810,90816,91047,91082,91239,91306,91503,91553,91820,91945,91959,91976,92126,92168,92269,92369,92401,92426,92598,92635,92715,92743,92940,92984,93193,93332,93380,93384,93551,93568,93841,93966,94353,94357,94466,94568,94570,94669,94845,94967,95004,95044,95093,95125,95140,95195,95357,95379,95481,95564,95567,95710,95739,95744,95777,95925,96095,96104,96146,96207,96362,96456,96470,96618,96755,96847,96897,96900,96957,97028,97036,97087,97167,97294,97532,97545,97573,97577,97614,97772,97786,97830,98060,98066,98103,98206,98211,98373,98379,98406,98485,98554,98606,98684,98694,98721,98735,98778,98789,98884,99083,99138,99234,99471,99529,99557,99565,99573,99673,99685,99803,99853,99985,100002",
          "9,13,286,288,309,334,342,370,388,451,487,493,505,531,598,615,970,989,1012,1017,1020,1024,1215,1378,1416,1833,2612,5079,5099,5122,5124,5215,5250,5261,5271,5322,5370,5378,5382,5455,5836,5859,5870,5874,5925,5969,5980,6022,6047,6060,6068,6075,6117,6167,6242,6258,6315,6325,6499,6549,6727,6835,6895,6957,7126,7255,7417,7498,7514,7571,7886,8024,8100,8162,8225,8292,8334,8481,8715,8808,8886,8961,9551,10076,10160,10222,10285,10629,11014,11791,12023,13375,14836,15245,16413,16605,16706,16773,17894,18125,18242,18976,20483,21058,21314,21396,23395,23580,27497,28377,29436,30646,40028,42494,43323,50013,50029,50147,50192,50231,50239,50278,50408,50421,50441,50543,50595,50654,50793,50903,51186,51518,51636,51808,51850,51857,51979,51998,52044,52077,52197,52338,52358,52393,52463,52724,52736,52775,52846,52914,53343,53443,53450,53453,53480,53690,53766,53948,53968,54003,54104,54395,54428,54571,54765,54900,54912,55301,55303,55436,55441,55471,55512,55536,55578,55590,55599,55620,55622,55669,55676,55697,55732,55768,55874,55876,55993,56100,56140,56209,56224,56229,56283,56290,56372,56409,56418,56567,56653,56720,56869,56924,56942,57001,57188,57254,57535,57545,57588,57728,57840,57929,58067,58118,58193,58232,58376,58419,58502,58701,58926,58956,59213,59366,59476,59595,59739,59758,60039,60040,60042,60111,60364,60489,60613,60664,60701,60738,60741,60765,60903,60932,60948,61073,61075,61147,61150,61258,61333,61365,61381,61409,61493,61525,61691,61946,62053,62104,62264,62314,62489,62592,62867,62894,62922,63004,63021,63030,63061,63063,63244,63481,63493,63550,63602,63664,63673,63792,63808,63929,63944,63994,64064,64186,64225,64271,64305,64358,64386,64457,64672,64760,64792,64827,64842,64881,64937,64954,65103,65194,65227,65338,65367,65414,65447,65450,65469,65616,65621,65662,65770,65864,66056,66116,66168,66174,66332,66394,66440,66552,66565,66628,66633,66684,66884,66901,67097,67199,67240,67393,67580,67650,67848,67872,67881,67896,67922,68009,68069,68105,68106,68133,68152,68361,68434,68500,68529,68583,68683,68711,68770,68771,68830,68839,68868,68925,69015,69023,69042,69063,69278,69324,69429,69483,69564,69647,69729,70051,70185,70239,70377,70586,70603,70765,70850,70887,70957,71015,71055,71067,71077,71105,71193,71331,71339,71395,71412,71429,71522,71529,71696,71796,71899,72186,72222,72291,72336,72468,72516,72626,72779,72788,72943,73006,73067,73081,73343,73414,73588,73652,74124,74157,74167,74258,74330,74337,74367,74385,74714,74744,74782,74793,74988,75008,75098,75127,75241,75267,75341,75391,75516,75629,75718,75853,75857,75888,75891,75927,75990,75998,76038,76259,76263,76482,76622,76777,76872,77152,77209,77258,77339,77494,77509,77584,77683,77802,78051,78195,78300,78352,78380,78439,78555,78711,78775,78801,78812,78910,79032,79149,79322,79348,79423,79514,79631,79714,79830,79867,79969,80094,80162,80314,80476,80535,80660,80681,80734,80784,80861,81032,81102,81266,81512,81578,81653,81677,81749,81773,81809,81861,81891,82068,82088,82188,82406,82470,82539,82542,82554,82701,82757,82905,82943,83425,83608,83610,83633,83682,83870,83909,83946,83952,83980,84040,84103,84141,84371,84440,84461,84519,84651,84788,84819,85116,85178,85279,85385,85756,85833,85837,85952,85975,86081,86140,86260,86368,86418,86434,86548,86576,86678,86688,86760,86824,86990,87002,87150,87270,87281,87309,87317,87339,87401,87418,87472,87530,87540,87608,87670,87799,87897,87906,88406,88528,88653,88696,88719,88726,88786,88821,88942,89065,89077,89202,89236,89257,89370,89431,89528,89621,89665,89864,90100,90121,90322,90378,90456,90514,90635,90691,90764,90787,90823,91573,91583,91617,91655,91660,91744,91765,91975,92004,92031,92121,92429,92576,92634,92675,92833,92855,92888,92917,93002,93057,93070,93186,93431,93724,93732,93800,93982,94003,94018,94155,94210,94324,94327,94496,94585,94615,94728,94785,94820,94958,94984,95024,95126,95222,95260,95317,95338,95494,95725,95892,95923,96035,96125,96162,96171,96178,96208,96232,96260,96333,96436,96691,96771,96882,97047,97100,97193,97271,97561,97566,97610,97684,97698,97723,97729,97737,97817,97860,97886,97948,98010,98040,98041,98112,98215,98424,98448,98546,98596,98628,98928,98930,98980,98995,99129,99273,99341,99538,99582,99607,99747,99833"
        ]
      }
    }
  ]
}
)";
const folly::dynamic kEagerShardSelectionLoadBalancerHugeJson{
    folly::parseJson(kEagerShardSelectionLoadBalancerHuge)};

SimpleRouteHandleProvider<HelloGoodbyeRouteHandleIf> gProvider;
RouteHandleFactory<HelloGoodbyeRouteHandleIf> gFactory{gProvider, 0};

HelloGoodbyeRouterInfo::RouteHandlePtr buildShardSelectionRoute(
    const folly::dynamic& json) {
  return facebook::memcache::mcrouter::createShardSelectionRoute<
      HelloGoodbyeRouterInfo,
      BasicShardSelector,
      std::unordered_map<uint32_t, uint32_t>>(gFactory, json);
}

HelloGoodbyeRouterInfo::RouteHandlePtr buildEagerShardSelectionRoute(
    const folly::dynamic& json) {
  return facebook::memcache::mcrouter::createEagerShardSelectionRoute<
      HelloGoodbyeRouterInfo,
      BasicShardSelector,
      std::unordered_map<uint32_t, uint32_t>>(gFactory, json);
}

} // anonymous namespace

BENCHMARK(createShardSelectionRoute_small) {
  auto rh = buildShardSelectionRoute(kShardSelectionSmallJson);
  folly::doNotOptimizeAway(rh);
}

BENCHMARK(createShardSelectionRoute_medium) {
  auto rh = buildShardSelectionRoute(kShardSelectionMediumJson);
  folly::doNotOptimizeAway(rh);
}

BENCHMARK(createShardSelectionRoute_large) {
  auto rh = buildShardSelectionRoute(kShardSelectionLargeJson);
  folly::doNotOptimizeAway(rh);
}

BENCHMARK(createEagerShardSelectionRoute_huge) {
  auto rh = buildEagerShardSelectionRoute(kEagerShardSelectionHugeJson);
  folly::doNotOptimizeAway(rh);
}

BENCHMARK(createEagerShardSelectionRoute_LoadBalancer_huge) {
  auto rh =
      buildEagerShardSelectionRoute(kEagerShardSelectionLoadBalancerHugeJson);
  folly::doNotOptimizeAway(rh);
}

/**
 * BENCHMARK RESULTS (opt mode):
 *
 * ============================================================================
 * ShardSelectionRouteFactoryBenchmark.cpp                   time/iter  iters/s
 * ============================================================================
 * createShardSelectionRoute_small                              7.14us  140.05K
 * createShardSelectionRoute_medium                            29.67us   33.71K
 * createShardSelectionRoute_large                             29.54ms    33.85
 * createEagerShardSelectionRoute_huge                        171.63ms     5.83
 * createEagerShardSelectionRoute_LoadBalancer_hug            140.05ms     7.14
 * ============================================================================
 */
int main(int argc, char** argv) {
  folly::init(&argc, &argv, true /* removeFlags */);
  folly::runBenchmarks();
  return 0;
}
